/*
 * File almlib.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _ALMLIB_H_
#define _ALMLIB_H_

#include "project.def"
#include "tinedefs.h"
#include "srvcore.h"
#include <time.h>

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MAXALARMS
#define MAXALARMS 500
#endif

#define DEFAULT_ALARM_WINDOW 3
#define DEFAULT_ALARM_HEARTBEAT 1200
#define DEFAULT_INSTANT_WINDOW 30
#define DEFAULT_TERMINATION_WINDOW 30
#define DEFAULT_ALARM_DATACHANGE_WINDOW 30
#define DEFAULT_ALARM_SURVEY_INTERVAL 1000

#define ALM_SYSTEM_SYSTEM   5000
#define ALM_SYSTEM_HARDWARE 5001

/* appended to alarmSystem  : */
#define ALM_SYSTEM_HIDDEN   0x00004000 
/* appended to alarm code   : */
#define ALM_APPLYNAME       0x80000000
#define ALM_DISKSPACE       0x40000000 
#define ALM_LINK_ERROR      0x20000000
#define ALM_SYSTEM_DEFAULT  0x10000000
#define ALM_SYSTEM_HIDE     0x08000000
#define ALM_BASECODE(c) ((c) & 0x7ffffff)

#define ALM_SEVERITY_NONE   0
#define ALM_SEVERITY_INFO   3
#define ALM_SEVERITY_WARN   8
#define ALM_SEVERITY_ERR   12
#define ALM_SEVERITY_FATAL 15

#define almNEWALARM     0x01
#define almHEARTBEAT    0x02
#define almOSCILLATION  0x04
#define almDATACHANGE   0x08
#define almINSTANT      0x10
#define almTRANSIENT    0x10
#define almDISABLED     0x20
#define almTERMINATE    0x40
#define almTEST         0x80
#define almSUPPRESS     0x80

#define ALM_OSCILLATION_CUSHION 3
#define ALM_OSCILLATION_MAXIMUM 128
#define ALM_OSCILLATION_WINDOW 8
#define ALM_COLLAPSE_WINDOW 100

#define ALARM_DATA_SHORTSIZE 6
#define ALARM_DATA_SIZE 64

#define ALARM_TEXT_SIZE 64
#define ALARM_TEXT_SHORTSIZE 40
#define ALARM_TEXT_LONGSIZE 128

#define encodeLinkErrorAlarm(c) (ALM_LINK_ERROR + ((c)<<16) + link_error)
#define isLinkErrorAlarm(c) ((c) & ALM_LINK_ERROR)

#define encodeDiskSpaceAlarm(c,a) (ALM_DISKSPACE + ((c)<<16) + (a))
#define isDiskSpaceAlarm(c) ((c) & ALM_DISKSPACE)

#define isHiddenAlarm(c) ((c) & ALM_SYSTEM_HIDE)
#define isSystemDefaultAlarm(c) ((c) & ALM_SYSTEM_DEFAULT)

#define isAlarmApplyName(c) ((c) & ALM_APPLYNAME)

/**
 * \internal
 * \brief alarm snapshot information
 *
 */
typedef struct
{
  short  nalms;  /**< number of active alarms */
  UINT32 mrt;    /**< alarm Most Recent Timestamp */
  UINT32 mrct;   /**< alarm Most Recent Complete Timestamp */
  short  hsv;    /**< alarm highest Severity */
  short  nmrt;   /**< number of alarms at MRT */
  short  nmrct;  /**< number of alarms at MRCT */
  short  nhsv;   /**< number of alarms at HSV */
} AlmInfo;

typedef struct AlarmMsgStruct5
{
  char server[EXPORT_NAME_SHORTSIZE];
  char device[DEVICE_NAME_SHORTSIZE];
  char alarmTag[ALARM_TAG_SHORTSIZE];
  UINT32 alarmCode;
  UINT32 timestamp;
  UINT32 alarmMask;
  BYTE alarmData[ALARM_DATA_SHORTSIZE];
  BYTE alarmDataFormat;
  BYTE alarmDataArraySize;
  BYTE severity;
  BYTE descriptor;
  UINT16 alarmSystem;
} AMS5;                    /* legacy AMS sent out upon request */
#define AMS5_SIZE  72

/**
 * \struct AMS
 * \brief Defines an Alarm Message Structure
 *
 * delivered with calls to stock property "ALARMS"
 */
typedef struct AlarmMsgStruct
{
  char server[EXPORT_NAME_SIZE];     /**< exported server name */
  char device[DEVICE_NAME_SIZE];     /**< device name */
  char alarmTag[ALARM_TAG_SIZE];     /**< short alarm tag */
  UINT32 alarmCode;                  /**< alarm code ID */
  UINT32 timestamp;                  /**< current alarm time (seconds part) */
  UINT32 timestampUSec;              /**< current alarm time (micro-seconds part) */
  UINT32 starttime;                  /**< alarm starttime (seconds part) */
  UINT32 starttimeUSec;              /**< alarm starttime (micro-seconds part) */
  UINT32 alarmMask;                  /**< alarm mask */
  BYTE alarmData[ALARM_DATA_SIZE];   /**< alarm data (if any) */
  BYTE alarmDataFormat;              /**< alarm data format (simple code) */
  BYTE alarmDataArraySize;           /**< alarm data array length */
  BYTE severity;                     /**< alarm severity (0 - 15) */
  BYTE descriptor;                   /**< current alarm descriptor */
  UINT16 alarmSystem;                /**< alarm system ID (0 -> use registered default) */
  BYTE alarmOscillationWindow;       /**< alarm oscillation window (in clear counts) */
  BYTE alarmOscillationWindowPinned; /**< binary: 0 = not pinned */
} AMS; /**< Alarm Message Structure sent out upon request */
#define AMS_SIZE (EXPORT_NAME_SIZE+DEVICE_NAME_SIZE+ALARM_TAG_SIZE+6*4+ALARM_DATA_SIZE+2*4) /* 224 bytes */

#ifndef MAX_AMS_CAPACITY
# define MAX_AMS_CAPACITY 1000
#endif
extern int gMaxAMSCapacity;
TINE_EXPORT void SetAMSCapacity(int value);

typedef struct 
{
  UINT32 timestamp;
  UINT32 timestampUSec;
  UINT32 starttime;
  UINT32 starttimeUSec;
  BYTE descriptor;
  BYTE severity;
  UINT16 nalarms;
} AMSDynSet;

typedef struct ADS5tag    /* Alarm Definition Structure (legacy) */
{
   char alarmTag[ALARM_TAG_SHORTSIZE];
   UINT32 alarmCode;
   UINT32 alarmMask;
   short severity;
   BYTE alarmDataFormat;
   BYTE alarmDataArraySize;
   UINT32 timeRes;
   short hitsChannel;
   short hitsModule;
   char alarmText[ALARM_TEXT_SHORTSIZE];
   char deviceText[ALARM_TEXT_SHORTSIZE-2];
   UINT16 alarmSystem;  /* take 2 bytes from deviceText to hold the system ID */
   char dataText[ALARM_TEXT_SHORTSIZE];
   char url[ALARM_TEXT_SHORTSIZE];
   UINT32 ndefs;
} ADS5;
#define ADS5_SIZE  200

#define ALM_MASK_NONE           0x00000000
#define ALM_MASK_DATA_TO_TAG    0x00000001

/**
 * \struct ADS
 * \brief Defines an Alarm Definition Structure
 *
 * contains largely static information pertaining to a specific alarm code
 */
typedef struct ADStag    /* Alarm Definition Structure */
{
   char alarmTag[ALARM_TAG_SIZE];   /**< short text description of alarm */
   UINT32 alarmCode;                /**< alarm code ID */
   UINT32 alarmMask;                /**< alarm mask */
   UINT16 alarmSystem;              /**< alarm system ID (0 -> use registered default) */
   short  alarmSeverity;            /**< alarm severity (0 - 15) */
   BYTE alarmDataFormat;            /**< alarm data format (simple code) */
   BYTE alarmDataArraySize;         /**< alarm data array length */
   BYTE alarmOscillationWindow;     /**< alarm oscillation window (in clear counts) */
   BYTE alarmOscillationWindowPinned;/**< binary: 0 = not pinned */
   char alarmText[ALARM_TEXT_SIZE]; /**< long text description for alarm tag */
   char deviceText[ALARM_TEXT_SIZE];/**< text description for device type */
   char dataText[ALARM_TEXT_SIZE];  /**< text description for alarm data contents */
   char url[ALARM_TEXT_LONGSIZE];   /**< url link to further information */
} ADS;
#define ADS_SIZE  (ALARM_TAG_SIZE + 2*4 + 4*2 + 3*ALARM_TEXT_SIZE + ALARM_TEXT_LONGSIZE)  /* 368 */

#ifndef ADSType
# define ADSType ADS
#endif
typedef struct ADSListStruct
{
  ADS ads;
# ifndef SMALL_TINE_LIB
  FilterLink *fltr;            /**< assigned filter */
# endif
  struct ADSListStruct *next;
} ADSList;

/**
 * \internal
 * \brief Alarm Definition Structure + 64 char format string used in parsing input from config files
 * 
 */
typedef struct
{
   char alarmTag[ALARM_TAG_SIZE];
   UINT32 alarmCode;  /* alarm code ID */
   UINT32 alarmMask;  /* alarm mask */
   UINT16 alarmSystem;/* alarm system ID */
   short  alarmSeverity;/* alarm severity */
   BYTE alarmDataFormat;
   BYTE alarmDataArraySize;
   short reserved;
   char alarmText[ALARM_TEXT_SIZE];
   char deviceText[ALARM_TEXT_SIZE];
   char dataText[ALARM_TEXT_SIZE];
   char url[ALARM_TEXT_LONGSIZE];
   char alarmDataFormatString[64]; /* for easy reading by csv and xml parsers */
# ifndef SMALL_TINE_LIB
   char filterString[256]; /* supplied filter */
# endif
} ADSIS;

typedef struct AMRBtag
{
  UINT32 almStartTime;
  UINT32 almStopTime;
  UINT32 almMinSeverity;
  UINT32 reserved;
} AMRB; /* Alarm Message Request Block */
#define AMRB_SIZE  16

/**
 * \internal
 * \brief Alarm Settings Structure used in parsing input from config files
 * 
 */
typedef struct
{
  int wnd;
  int trnWnd;
  int dcWnd;
  int trmWnd;
  int oscWnd;
  int clpWnd;
  int hbt;
  int tblSiz;
} ALMSETIS;

#define AssignAlmTimestamp(a,t) {\
  (a)->timestamp = (UINT32)(t);\
  (a)->timestampUSec = (((UINT32)((t) * 1000000)) % 1000000);\
}

#define ALMARM_WATCH_FILE "almwatch.csv"
#define ALARM_MANIFEST "almmf.csv"
#define ALARM_WATCH_MANIFEST "almwatchmf.csv"
#define ALARM_WATCH_MANIFEST_LEGACY "almwatmf.csv"
#define ALM_SETTINGS_FILE "almsettings.csv"

#define ALM_MF_ALL 0
#define ALM_MF_WAT 1

/**
 * \struct AWS
 * \brief Defines an Alarm Watch Structure
 *
 * If an alarm watch table is registered, then it will consist of an array
 * of entries containing the information in this structure, specifying 
 * what and how often the local alarm system should call and check to
 * set or clear alarms automatically.
 *
 * A call to stock property "ALMWATCHTBL" will return an array of AWS entries.
 * Used in conjunction with the linked list AWSLstEntry, which can override the
 * default specifications for the alarm codes and alarm tags used in the 
 * alarm watch table.
 *
 */
typedef struct AlarmWatchStruct
{
  char eqm[EQM_NAME_SIZE];/**< equipment module (eqm) to call to check for alarms */
  char dev[DEVICE_NAME_SIZE];/**< device to use in calling the eqm */ 
  char prp[PROPERTY_NAME_SIZE];/**< property to use in calling the eqm */
  int  siz;               /**< array size to use in calling the eqm */
  int  fmt;               /**< format (long type) to use in calling the eqm */
  int  atyp;              /**< array type */
  int  asys;              /**< alarm system id */
  int  sev;               /**< general highest severity */
  int  hisev;             /**< high severity */
  int  hiwarnsev;         /**< high warning severity */
  int  losev;             /**< low severity */
  int  lowarnsev;         /**< low warning severity */
  int  chkNormal;         /**< != 0 => check vs normal; FALSE = check threshold */
  UINT32 mask;            /**< if fmt is an integer type */
  UINT32 normal;          /**< normal non-alarm state value (after mask) */ 
  int  cnt;               /**< current alarm counter */
  int  cntThreshold;      /**< set alarm when counter > threshold */
  float hi;               /**< high threshold */
  float lo;               /**< low threshold */
  float hiwarn;           /**< high warn threshold */
  float lowarn;           /**< low warn threshold */
} AWS;
#define AWS_SIZE (EQM_NAME_SIZE+DEVICE_NAME_SIZE+PROPERTY_NAME_SIZE+18*4) /* 208 bytes */

/**
 * \struct AWLQS
 * \brief Defines an Alarm Watch List Query Structure for stock property queries
 *
 * Presents the infomration in an AWSLstEntry for querying purposes
 *
 */
typedef struct 
{
  char eqm[EQM_NAME_SIZE];/**< equipment module (eqm) to call to check for alarms */
  char dev[DEVICE_NAME_SIZE];/**< device to use in calling the eqm */
  char prp[PROPERTY_NAME_SIZE];/**< property to use in calling the eqm */
  int  siz;               /**< array size to use in calling the eqm */
  int  fmt;               /**< format (long type) to use in calling the eqm */
  int  atyp;              /**< array type */
  int  asys;              /**< alarm system id */
  int  sev;               /**< general highest severity */
  int  hisev;             /**< high severity */
  int  hiwarnsev;         /**< high warning severity */
  int  losev;             /**< low severity */
  int  lowarnsev;         /**< low warning severity */
  int  chkNormal;         /**< != 0 => check vs normal; FALSE = check threshold */
  UINT32 mask;            /**< if fmt is an integer type */
  UINT32 normal;          /**< normal non-alarm state value (after mask) */
  int  cnt;               /**< current alarm counter */
  int  cntThreshold;      /**< set alarm when counter > threshold */
  float hi;               /**< high threshold */
  float lo;               /**< low threshold */
  float hiwarn;           /**< high warn threshold */
  float lowarn;           /**< low warn threshold */
  int code;                        /**< assigned alarm code for alarms */
  int codeHigh;                    /**< assigned alarm code for 'high' alarms */
  int codeHighWarn;                /**< assigned alarm code for 'high warn' alarms */
  int codeLow;                     /**< assigned alarm code for 'low' alarms */
  int codeLowWarn;                 /**< assigned alarm code for 'low warn' alarms */
  char tagNormal[ALARM_TAG_SIZE];  /**< assigned alarm tag for 'normal' alarms */
  char tagHigh[ALARM_TAG_SIZE];    /**< assigned alarm tag for 'high' alarms */
  char tagHighWarn[ALARM_TAG_SIZE];/**< assigned alarm tag for 'high warn' alarms */
  char tagLow[ALARM_TAG_SIZE];     /**< assigned alarm tag for 'low' alarms */
  char tagLowWarn[ALARM_TAG_SIZE]; /**< assigned alarm tag for 'low warn' alarms */
  char fltr[256];                  /**< assigned filter string */
} AWLQS;
#define AWLQS_SIZE (EQM_NAME_SIZE+DEVICE_NAME_SIZE+PROPERTY_NAME_SIZE+23*4+5*32+256) /* 644 bytes */
#define AWLQS_CPSIZE (EQM_NAME_SIZE+DEVICE_NAME_SIZE+PROPERTY_NAME_SIZE+23*4+5*32) /* 388 bytes */

/**
 * \struct AWSLstEntry
 * \brief Uses an Alarm Watch Structure + extra override information
 *
 * If an alarm watch table is registered, then it will consist of an array
 * of entries containing the information in this structure, specifying 
 * what and how often the local alarm system should call and check to
 * set or clear alarms automatically.  The working linked list used by the
 * local alarm server is of this type.
 *
 *
 */
typedef struct AWSListStruct
{
  AWS aws;                         /**< embedded alarm watch structure */
  int code;                        /**< assigned alarm code for alarms */
  int codeHigh;                    /**< assigned alarm code for 'high' alarms */
  int codeHighWarn;                /**< assigned alarm code for 'high warn' alarms */
  int codeLow;                     /**< assigned alarm code for 'low' alarms */
  int codeLowWarn;                 /**< assigned alarm code for 'low warn' alarms */
  char tagNormal[ALARM_TAG_SIZE];  /**< assigned alarm tag for 'normal' alarms */
  char tagHigh[ALARM_TAG_SIZE];    /**< assigned alarm tag for 'high' alarms */
  char tagHighWarn[ALARM_TAG_SIZE];/**< assigned alarm tag for 'high warn' alarms */
  char tagLow[ALARM_TAG_SIZE];     /**< assigned alarm tag for 'low' alarms */
  char tagLowWarn[ALARM_TAG_SIZE]; /**< assigned alarm tag for 'low warn' alarms */
# ifndef SMALL_TINE_LIB
  FilterLink *fltr;                /**< assigned filter */
# endif
  struct AWSListStruct *nxt;
} AWSLstEntry;

extern AWSLstEntry *almWatchList;
extern int nAlmWatchList;

/**
 * \internal
 * \brief Alarm Watch Structure + additional info following lowarn field used in parsing input from config files
 * 
 */
typedef struct
{
  char eqm[EQM_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  int  siz;
  int  fmt;
  int  atyp;              /* array type */
  int  asys;              /* alarm system id */
  int  sev;               /* general highest severity */
  int  hisev;             /* high severity */
  int  hiwarnsev;         /* high warning severity */
  int  losev;             /* low severity */
  int  lowarnsev;         /* low warning severity */
  int  chkNormal;         /* != 0 => check vs normal; FALSE = check threshold */
  UINT32 mask;            /* if fmt is an integer type */
  UINT32 normal;          /* normal non-alarm state value (after mask) */ 
  int  cnt;
  int  cntThreshold;      /* set alarm when counter > threshold */
  float hi;               /* high threshold */
  float lo;               /* low threshold */
  float hiwarn;           /* high warn threshold */
  float lowarn;           /* low warn threshold */
  int  code;              /* assigned alarm code */
  int  codeHigh;          /* assigned alarm code */
  int  codeHighWarn;      /* assigned alarm code */
  int  codeLow;           /* assigned alarm code */
  int  codeLowWarn;       /* assigned alarm code */
  char tagNormal[ALARM_TAG_SIZE]; /* assigned alarm tag */
  char tagHigh[ALARM_TAG_SIZE]; /* assigned alarm tag */
  char tagHighWarn[ALARM_TAG_SIZE]; /* assigned alarm tag */
  char tagLow[ALARM_TAG_SIZE]; /* assigned alarm tag */
  char tagLowWarn[ALARM_TAG_SIZE]; /* assigned alarm tag */
  char alarmDataFormatString[64]; /* for easy reading by csv and xml parsers */
  char normalValueString[64];
# ifndef SMALL_TINE_LIB
  char filterString[256]; /* supplied filter */
# endif
} AWSIS;

/**
 * \struct ALM_THRESHOLDS
 * \brief Defines a TINE Alarm Watch Threshold object
 *
 * - hi is the high value threshold used for setting value_too_high alarms
 * - hiwarn is the high warning value threshold used for setting warn_too_high alarms
 * - lo is the low value threshold used for setting value_too_low alarms
 * - lowarn is the low warning value threshold used for setting warn_too_low alarms
 */
typedef struct 
{
  int cnt;       /**< number successive threshold exceeded before applying alarm */
  int hisev;     /**< high severity */
  int hiwarnsev; /**< high warning severity */
  int losev;     /**< low severity */
  int lowarnsev; /**< low warning severity */
  float hi;      /**< high threshold */
  float hiwarn;  /**< high threshold warning (approaching high threshold) */
  float lo;      /**< low threshold */
  float lowarn;  /**< low threshold warning (approaching low threshold) */
  UINT32 mask;   /**< if non-zero -> apply mask and check vs. normal value */
  UINT32 normal; /**< normal valid data value */
  int normalIsAlarm; /**< apply a NOT to the 'normal value' */
  int code;      /**< alarm code to apply in lieu of invalid_index */
  int codeHigh;  /**< alarm code to apply in lieu of value_too_high */
  int codeHighWarn;/**< alarm code to apply in lieu of warn_too_high */
  int codeLow;   /**< alarm code to apply in lieu of value too low */
  int codeLowWarn;/**< alarm code to apply in lieu of warn_too_low */
} ALM_THRESHOLDS;

/**
 * \struct ALARM
 * \brief structure used in the local alarm linked list
 *
 * calls to stock property "ALARMS" use this linked list + the static
 * alarm code information contained in the ADS structure to return
 * an array of AMS (alarm message structures) to the caller.
 */
typedef struct ALARMstruct
{
  UINT32 timestamp;       /**< data timestamp */
  UINT32 timestampUSec;   /**< usec fraction of data timestamp */
  UINT32 starttime;       /**< initial data timestamp */
  UINT32 starttimeUSec;   /**< usec fraction of initial data timestamp */
  UINT32 alarmCode;       /**< standardized return code (implies severity)   */
  BYTE   alarmSeverity;   /**< dynamic override of static alarm severity */
  BYTE   alarmStatus;     /**< corresponds to the descriptor */
  BYTE   clrCount;        /**< record of clear requests */
  BYTE   removalBlocked;  /**< flag to keep alarm in table for CAS readout */
  BYTE   alarmData[ALARM_DATA_SIZE]; /**< alarm specific data (database interpreted) */
  AWSLstEntry *wte;       /**< possible reference to alarm watch table entry */
  struct ALARMstruct *next;
} ALARM;

extern ALARM *almList;

/* appendAlarmWatchTable() */
/**
 * \brief Inserts a property to be monitored into the local alarm server's Watch Table
 * 
 * Certain alarms are to be set whenever the value of a property exceeds a defineable 
 * threshold.  Such alarms can be managed automatically by the local alarm server if the
 * alarm criteria are entered into the alarm watch table.  This can be achieved by calling
 * this routine (or supplying a startup configuration file almwatch.csv).
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the property which is to be called by the local alarm server
 * \param dev is the device name associated with the property to be called by the local alarm server
 * \param siz is the data array size to be called by the local alarm server
 * \param fmt is the TINE data format to be called by the local alarm server
 * \param atyp is the TINE data array type to be applied to the property called by the local alarm server
 * \param sev is the severity of the alarm issued when the data returned by the call exceed the
 *  given thresholds.
 * \param sys is the alarm system identifier to be associated with the alarm.
 * \param thr is an ALM_THRESHOLDS object specifying the threshold criteria for setting
 *  the alarm
 *  
 * \note Information can also be entered into the alarm watch table by supplying the startup
 * configuration file almwatch.csv, where the input parameters are given in the relevant columns.
 * This is frequently the preferred way of inputting such alarm information, since if does not
 * involve hard-coding such alarm criteria.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 */
TINE_EXPORT int AppendAlarmWatchTable(char *eqm,char *prp,char *dev,int siz,int fmt,int atyp, int sev,int sys,ALM_THRESHOLDS *thr);
#define appendAlarmWatchTable(a1,a2,a3,a4,a5,a6,a7,a8,a9) AppendAlarmWatchTable(a1,a2,a3,a4,a5,a6,a7,a8,a9)

/* ApplyAlarmWatchFilter() */
/**
 * \brief Applies a filter to an existing local alarm server's Watch Table
 * 
 * A server can filter out automatic watch table alarms by applying a filter condition which must be 
 * satisfied.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the property which is to be called by the local alarm server
 * \param dev is the device name associated with the property to be called by the local alarm server
 * \param parsableFilterString is a parsable filter string defining the filter condition.
 *  This should be of the form /<context>/<server/<device>[<property>]<comparator>\<value\>
 *  where <comparator> is one of '=', '!=', '>', or '<' and \<value\> is the filter's
 *  threshold value.  The filter target address should deliver (or be able to deliver) a single 
 *  numeric or string name value.  If the filter string cannot be parsed of the target address
 *  does not exists, then no filter will be established.
 *  In addition, the filterString can be concatenated with an optional severity, which
 *  should be used when the filter criterion is not met.  The optional serverity
 *  should be specified by appending &sev=\<severity\> at the end of the filter string.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 */
TINE_EXPORT int ApplyAlarmWatchFilter(char *eqm,char *prp,char *dev,char *parsableFilterString);

/* MaxNumAlarmDefinitions */
/**
 * \brief Determines the maximum alarm definition table size when using the appendAlarmInfoTable()
 * API call to register alarm table information.
 * 
 * If an alarm definition configuration file \<local name\>-alarms.csv is used, this variable
 * plays no role.  When the API call appendAlarmInfoTable() is used to register alarm table
 * information, this variable is used to pre-allocate the appropriate alarm definition table
 * size.
 *
 * \b Default: 100
 *  
 * \b Example:

  MaxNumAlarmDefinitions = 500;  

 */
extern TINE_EXPORT int MaxNumAlarmDefinitions;
/* appendAlarmInfoTable() */
/**
 * \brief Inserts an alarm definition into the alarm definition table.
 * 
 * As an alternative to the \<local name\>-alarms.csv configuration file,
 * the front end server can make use of this API call in order to fill in
 * the alarm definition table describing locally generated alarms.
 * This is particularly useful for embedded platforms where there is no 
 * file system, or where a TINE server is used as a translation layer and 
 * needs to map a given alarm system onto the TINE alarm system.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param ads is a pointer to an Alarm Definition Structure (ADS) containing the
 * alarm table information which is to be appended to the alarm definition table.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_appendAlarmInfoTable.c
 * 
 */
TINE_EXPORT int AppendAlarmInfoTable(char *eqm,ADS *ads);
/* appendAlarmInfoTableEx() */
/**
 * \brief Inserts an alarm definition into the alarm definition table. The
 * extended call accepts a parseable filter string.
 * 
 * As an alternative to the \<local name\>-alarms.csv configuration file,
 * the front end server can make use of this API call in order to fill in
 * the alarm definition table describing locally generated alarms.
 * This is particularly useful for embedded platforms where there is no 
 * file system, or where a TINE server is used as a translation layer and 
 * needs to map a given alarm system onto the TINE alarm system.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param ads is a pointer to an Alarm Definition Structure (ADS) containing the
 * alarm table information which is to be appended to the alarm definition table.
 * \param filterString is a parsable filter string defining the filter condition.
 *  This should be of the form /<context>/<server/<device>[<property>]<comparator>\<value\>
 *  where <comparator> is one of '=', '!=', '>', or '<' and \<value\> is the filter's
 *  threshold value.  The filter target address should deliver (or be able to deliver) a single 
 *  numeric or string name value.  If the filter string cannot be parsed of the target address
 *  does not exists, then no filter will be established.  The filter conditions must be
 *  met for the alarm specified in the ADS to be applied via any of the SetAlarm() methods.
 *  Unless, in addition, the filterString has been concatenated with an optional severity, which
 *  should be used when the filter criterion is not met.  The optional serverity
 *  should be specified by appending &sev=\<severity\> at the end of the filter string.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_appendAlarmInfoTableEx.c
 * 
 */
TINE_EXPORT int AppendAlarmInfoTableEx(char *eqm, ADS *ads, char *filterString);

/* SetDeviceAlarm() */
/**
 * \brief Inserts an alarm into the local alarm server table.
 * 
 * Analogous to SetAlarmEx(), this call inserts alarms into the local alarm server table.
 * A server can set an alarm for a specific device by giving its device name.  This is sometime
 * preferred over specifying its device number as in SetAlarm() or SetAlarmEx().
 * For further information, please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devname is the registered device name of the device for which the alarm is to be set.
 * \param almCode is the registered alarm code which defines the alarm
 * \param almData is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.
 * \param almFlags is the optional alarm flag which is used in the alarm descriptor information.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_SetDeviceAlarm.c
 */
TINE_EXPORT int SetDeviceAlarm(char *eqm,char *devname,UINT32 almCode,BYTE *almData,BYTE almFlags);

/* SetAlarmsEx() */
/**
 * \brief Inserts an alarm into the local alarm server table.
 * 
 * This is the extended call corresponding to SetAlarm(), the difference being 
 * the additional argument �flags� which can be used to set the alarm descriptor.
 * A server can set an alarm for a specific device by giving its device number.  
 * For further information, please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which the alarm is to be set.
 * \param almCode is the registered alarm code which defines the alarm
 * \param almData is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.
 * \param almFlags is the optional alarm flag which is used in the alarm descriptor information.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_setAlarmEx.c
 */
TINE_EXPORT int SetAlarmEx(char *eqm,int devNr,UINT32 almCode,BYTE *almData,BYTE almFlags);

/* SetAlarm() */
/**
 * \brief Inserts an alarm into the local alarm server table.
 * 
 * A server can set an alarm for a specific device by giving its device number.  
 * For further information, please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which the alarm is to be set.
 * \param almCode is the registered alarm code which defines the alarm
 * \param almData is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \include eg_SetAlarm.c
 */
TINE_EXPORT int SetAlarm(char *eqm,int devNr,UINT32 almCode,BYTE *almData);

TINE_EXPORT int getAlarmsEx(char *eqmName,AMS *ams,short num,short minSeverity,time_t starttime,time_t stoptime,int startnum);
#define GetAlarmsEx(e,a,n,m,t1,t2,sn) getAlarmsEx(e,a,n,m,t1,t2,sn)
TINE_EXPORT int getAlarms(char *eqmName,AMS *ams,short num,short minSeverity);
#define GetAlarms(e,a,n,m) getAlarms(e,a,n,m)
int numAlarmsEx(char *eqmName,int devNr,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity);
int numAlarms(char *eqmName,UINT32 *almTimeStamp,short *almSeverity);
TINE_EXPORT int GetNumAlarms(char *eqmName,int devNr,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity);

/**
* \brief Instructs the local alarm server to allow or disallow alarms for devices declared offline.
*
* \param value is the desired setting (DEFAULT = FALSE)
*/
TINE_EXPORT void SetAllowOfflineAlarms(int value);
/**
* \brief Gets the local alarm server setting to allow or disallow alarms for devices declared offline.
*
* \return the current setting (DEFAULT = FALSE)
*/
TINE_EXPORT int GetAllowOfflineAlarms(void);

/* ClearAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be cleared.
 * 
 * A server can clear alarms for specific device by giving its device number.  
 * Note that all alarms for the device specified are cleared.  Also note, 
 * that "clearing" an alarm means that its "clear" counter is incremented.  
 * It is not removed immediately from the local alarm list. For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which all alarms is to be cleared.
 *  
 * \include eg_ClearAlarm.c
 */
TINE_EXPORT void ClearAlarm(char *eqm,int devNr);
/* ClearDeviceAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be cleared.
 * 
 * A server can clear alarms for specific device by giving its registered device name.  
 * Note that all alarms for the device specified are cleared.  Also note, 
 * that "clearing" an alarm means that its "clear" counter is incremented.  
 * It is not removed immediately from the local alarm list. For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devname is the registered device name of the device for which all alarms is to be cleared.
 *  
 * \include eg_ClearDeviceAlarm.c
 */
TINE_EXPORT void ClearDeviceAlarm(char *eqm,char *devname);

/* RemoveDeviceAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be marked for removal.
 * 
 * A server can remove alarms for specific device by giving its registered device name.  
 * Note that only the alarm specified for the device specified is removed.  Also 
 * note, that "removing" an alarm means that termination bit will be set.  It is 
 * subsequently marked for removal from the local alarm list.  For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devname is the registered device name of the device for which all alarms is to be removed.
 * \param code is the alarm code, which identifies the specific alarm to be removed
 *
 * \include eg_RemoveDeviceAlarm.c
 */ 
TINE_EXPORT void RemoveDeviceAlarm(char *eqm,char *devname,UINT32 code);

/* RemoveAlarm() */
/**
 * \brief Instructs the local alarm server table that the given alarm is to be marked for removal.
 * 
 * A server can remove alarms for specific device by giving its device number.  
 * Note that only the alarm specified for the device specified is removed.  Also 
 * note, that "removing" an alarm means that termination bit will be set.  It is 
 * subsequently marked for removal from the local alarm list.  For further information, 
 * please see the discussion of the local alarm server.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param devNr is the device number of the device for which all alarms is to be removed.
 * \param code is the alarm code, which identifies the specific alarm to be removed
 *
 * \include eg_RemoveAlarm.c
 */ 
TINE_EXPORT void RemoveAlarm(char *eqm,int devNr,UINT32 code);
short findSeverity(char *name,UINT32 cc);
int RegisterSeverity(int almCode,int almSeverity);
void getAlarmInfo(void);
int getAlarmDefs(char *eqm,ADSList **adslst);
int numAlarmDefs(char *eqm);
ADS *getAlarmDefinition(char *eqm,UINT32 code);
int getAlarmInfoTable(char *eqm);
int initAlarmServer(void);
TINE_EXPORT int GetAlarm(char *eqmName,char *devname,AMS *ams);
int _setAlarm(char *eqm,int devNr,UINT32 almCode,BYTE *almData,BYTE flags,double ts,AWSLstEntry *aws);
int isCASInUserList(char *eqm);
int prepIncomingADS(BYTE *adsstream,ADS *ads);
int prepOutgoingADS(BYTE *adsstream,ADSList *adslst,char *stag,int length);
int prepOutgoingAlarmData(BYTE *datastream,int size,short fmt);
int prepOutgoingAMS(BYTE *amsstream,AMS *amslst,char *stag,int length);
int prepIncomingAMRB(AMRB *amrb,BYTE *amrbstream);
int prepOutgoingAWS(BYTE *awsstream,AWSLstEntry *awslst,char *stag,int length);
int prepIncomingAWS(BYTE *awsstream,void *aws,char *stag,int length);
void checkAlarmsInWatchTable(void);
void clearCASAlarmList(void);
int RemoveAllAlarms(char *eqm);
void freeAlmMemory(void);
void rmvAlarmsFromCAS(char *eqm);
int GetAlarmWatchTable(char *eqm);
int dumpAlmSettings(void);
int getAlmSettings(void);
int dumpAlarmManifest(char *eqm,int to_stdout);
int dumpAlarmWatchManifest(char *eqm,int to_stdout);
void dumpAlmManifest(int mode,int to_stdout);
void getAlarmFilterString(FilterLink *fltr, char *filterString);
FilterLink *getAlarmFilter(char *eqm, UINT32 code);
void setFileErrorAlarm(char *filename);
void clearFileErrorAlarm(void);
void lasClearAlarm(char *eqm,int devNr,UINT32 code,AWSLstEntry *aws);
int prepIncomingAlmFilter(BYTE *filterstream, FilterStruct *flst);
int prepOutgoingAlmFilter(BYTE *filterstream, ADSList *adslst, int length);

extern char *tagList[];
extern int gAlmOscillationWindowPinned;
extern int gAlmOscillationWindow;
extern int gAlmCollapseWindow;
extern int gMaxAMSCapacity;
extern int gAlmMfDumped;

/* AlarmWindow */
/**
 * \brief Determines the minimum time an alarm is to remain in the local alarm table 
 * before it can be removed.
 * 
 * IO routines can sometimes set and clear alarms at such a high frequency that 
 * an alarm could go unnoticed if it appeared and disappeared before a client had 
 * a chance to read it.  The AlarmWindow value determines the amount of time (in seconds) 
 * an alarm must remain in the local alarm list before it can be removed.  The default 
 * is 3 seconds, which is generally fine for a central alarm reader polling at 1 second 
 * intervals.
 *
 * \b Default: 3 sec
 *  
 * \b Example:

  AlarmWindow = 10;  // keep alarms in the table a bit longer

 */
extern TINE_EXPORT int AlarmWindow;

/* AlarmHeartbeat */
/**
 * \brief Determines the time interval at the 'HEARTBEAT' flag will be re-issued 
 * for persistent alarms.
 * 
 * The purpose of an alarm is to alert operators or maintenance staff to a problem.  
 * If an alarm persists and is not dealt with, its timestamp is adjusted to the current 
 * time at intervals given by the AlarmHeartbeat value.  This will 
 * call attention to the alarm.  It also places the alarm at regular intervals in an 
 * archive of alarms.  The developer can set this to any desired value (usually at 
 * initialization time).
 *
 * \b Default: 1200 sec
 *  
 * \b Example:

  AlarmHeartbeat = 3600;  // heartbeat once per hour

 */
extern TINE_EXPORT int AlarmHeartbeat;

/* AlarmInstantWindow */
/**
 * \brief Determines the minimum time an alarm is to remain in the local alarm table 
 * before it can be removed.
 * 
 * Transient alarms (with no duration) are immediately marked for termination as soon 
 * as they are set.  The AlarmInstantWindow determines the amount of time in seconds 
 * an "Instant" Alarm is to remain in the alarm list (default 4 seconds).
 *
 * \b Default: 30 sec
 *  
 * \b Example:

  AlarmInstantWindow = 40;  // keep transient alarms in the table a bit longer

 */
extern TINE_EXPORT int AlarmInstantWindow;

/* TerminationWindow */
/**
 * \brief Determines the minimum time an expired alarm is to remain in the local alarm table. 
 * 
 * Terminated alarms remain in the local alarm table for an amount of time specified by
 * this variable.  Thus a client monitoring the local alarm list will be able tosee that 
 * the alarm was set and has finished.
 *
 * \b Default: 30 sec
 *  
 * \b Example:

  TerminationWindow = 40;  // keep terminated alarms in the table a bit longer

 */
extern TINE_EXPORT int TerminationWindow;


/* AlarmDataChangeWindow */
/**
 * \brief Determines the minimum time a data change alarm is to keep its timestamp in the local alarm table. 
 * 
 * Alarms supplying data can be updating the alarm data quite often.  Each change of alarm data will
 * be represented in the alarm itself.  However the timestamp of the alarm will not be updated unless a
 * time interval larger than the AlarmDataChangeWindow has occured.
 *
 * \b Default: 30 sec
 *  
 * \b Example:

  AlarmDataChangeWindow = 40;  // make the data change window a bit longer

 */
extern TINE_EXPORT int AlarmDataChangeWindow;

extern int NTAGS;
extern int NSEVERITYCODE;

/* MaxNumberAlarms */
/**
 * \brief Determines the maximum number of alarms which can be maintained in the
 * local alarm server's alarm list.
 * 
 * Each registered device can have an associated alarm link list.  The length of this 
 * list will not exceed the value entered for MaxNumAlarms (default 100).  The default 
 * is generally more than enough.  Primarily this variable serves as a safety valve 
 * against sudden alarm avalanches.
 *
 * \b Default: 100
 *  
 * \b Example:

  MaxNumberAlarms = 500;  

 */
extern TINE_EXPORT int MaxNumberAlarms;

/* AlarmTestState */
/**
 * \brief Determines the minimum time an alarm is to remain in the local alarm table 
 * before it can be removed.
 * 
 * When developing the alarm logic for a particular server is frequently useful to 
 * mark the alarms being set with the almTEST bit.  This can either be done on an 
 * individual basis using the SetAlarmEx() call, or one can set the global variable 
 * AlarmTestState to a non-zero value.
 *
 * \b Default: FALSE
 *  
 * \b Example:

  AlarmTestState = TRUE;  // just testing ...

 */
extern int AlarmTestState;
extern int almdbg;
extern int gRequireReadFromCAS;

extern int gIgnoreRegionRules;
TINE_EXPORT void SetIgnoreRegionRules(int value);
extern int gIsCentralAlarmServer;
TINE_EXPORT void isCentralAlarmServer(int value);
TINE_EXPORT int SetAlarmSurveyInterval(UINT32 interval);
TINE_EXPORT int GetAlarmSurveyInterval(void);

/* SetAlarmTableMaximumSize() */
/**
 * \brief Sets the maximum number of alarms managed by the local alarm server
 * 
 * A server keeps a maximum number of alarms in its local table at one time
 * before it begins to discard the older alarms.  The caller can adjust this number
 * with this call. Default is 500.
 *
 * \param tableSize is the maximum number of alarms to be maintained in the local alarm table.
 *
 */ 
TINE_EXPORT void SetAlarmTableMaximumSize(int tableSize);

/* GetAlarmTableMaximumSize() */
/**
 * \brief Gets the maximum number of alarms managed by the local alarm server
 * 
 * A server keeps a maximum number of alarms in its local table at one time
 * before it begins to discard the older alarms.  The caller can retrieve this number
 * with this call. Default is 500.
 *
 * \return the maximum number of alarms to be maintained in the local alarm table.
 *
 */ 
TINE_EXPORT int GetAlarmTableMaximumSize(void);

/* GetAlarmDefinition() */
/**
 * \brief Gets the alarm definition for the alarm code specified.
 * 
 * \param eqm (input) is the 6-character local equipment identifier name, which is internal to server.
 * \param code (input) is the alarm code for which the alarm definition is desired.
 * \param ads (output) is a pointer reference to an alarm definition structure (ADS) to be filled in
 * upon successful completion.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 */ 
TINE_EXPORT int GetAlarmDefinition(char *eqm,UINT32 code,ADS *ads);

/* SetAlarmTerminationWindow() */
/**
 * \brief Sets the size of the termination window in seconds before a terminated alarm is
 * discarded from the local alarm list.
 * 
 * A server keeps keeps a terminated alarm in its list (marked as terminated) for a 
 * length of time specified by the termination window.  This allows the central alarm server
 * and/or other interested parties enough time to discover that the alarm has been terminated.
 * The caller can set this window with this call.
 *
 * \param timeInSeconds is the size of the termination window. (Default: 30)
 *
 */ 
TINE_EXPORT void SetAlarmTerminationWindow(int timeInSeconds);

/* GetAlarmTerminationWindow() */
/**
 * \brief Gets the size of the termination window in seconds before a terminated alarm is
 * discarded from the local alarm list.
 * 
 * A server keeps keeps a terminated alarm in its list (marked as terminated) for a 
 * length of time specified by the termination window.  This allows the central alarm server
 * and/or other interested parties enough time to discover that the alarm has been terminated.
 * Default is 4 seconds.  The caller can discover this time window with this call.
 *
 * \return the size of the termination window.
 *
 */ 
TINE_EXPORT int GetAlarmTerminationWindow(void);

/* SetAlarmTerminationWindow() */
/**
 * \brief Sets the size of the new alarm window in seconds before a new alarm is
 * no longer declared as new.
 * 
 * A server flags an alarm as new when it first appears in its list.
 * After a short time span, the 'new' flag is removed by additional calls to SetAlarm for the
 * same alarm.
 * Default is 3 seconds.  The caller can set this window with this call.
 *
 * \param timeInSeconds is the size of the new alarm window.
 *
 */ 
TINE_EXPORT void SetNewAlarmWindow(int timeInSeconds);

/* GetNewAlarmWindow() */
/**
 * \brief Gets the size of the new alarm window in seconds before a new alarm is
 * no longer declared as new.
 * 
 * A server flags an alarm as new when it first appears in its list.
 * After a short time span, the 'new' flag is removed by additional calls to SetAlarm for the
 * same alarm.
 * Default is 3 seconds.  The caller can discover this time window with this call.
 *
 * \return the size of the new alarm window.
 *
 */ 
TINE_EXPORT int GetNewAlarmWindow(void);

/* SetAlarmDataChangeWindow() */
/**
 * \brief Sets the size of the alarm datachange window in seconds.
 * 
 * When an alarm provides alarm data with the appropriate SetAlarm() call, the
 * alarm will be tagged with the 'DATACHANGE' flag if the supplied data differ from 
 * that provided with the previous call to SetAlarm().  The alarm timestamp 
 * will be adjusted to the current time ONLY if the data change is seen outside of
 * the current AlarmDataChangeWindow (in seconds).
 * The value of the alarm timestamp is relevant in the acquisition of the existing
 * alarm table by the central alarm server.  If this window is too short and the
 * alarm data is constantly changing this can result in very long 'alarm chains' to
 * display and archive.
 * Default is 30 seconds.  The caller can discover this time window with this call.
 *
 * \param timeInSeconds is the size of the alarm data change window. (default = 30).
 *
 */ 
TINE_EXPORT void SetAlarmDataChangeWindow(int timeInSeconds);

/* GetAlarmDataChangeWindow() */
/**
 * \brief Gets the size of the alarm datachange window in seconds.
 * 
 * When an alarm provides alarm data with the appropriate SetAlarm() call, the
 * alarm will be tagged with the 'DATACHANGE' flag if the supplied data differ from 
 * that provided with the previous call to SetAlarm().  The alarm timestamp 
 * will be adjusted to the current time ONLY if the data change is seen outside of
 * the current AlarmDataChangeWindow (in seconds).
 * The value of the alarm timestamp is relevant in the acquisition of the existing
 * alarm table by the central alarm server.  If this window is too short and the
 * alarm data is constantly changing this can result in very long 'alarm chains' to
 * display and archive.
 * Default is 30 seconds.  The caller can discover this time window with this call.
 *
 * \return the size of the alarm data change window.
 *
 */ 
TINE_EXPORT int GetAlarmDataChangeWindow(void);

/* SetAlarmHeartbeat() */
/**
 * \brief Sets the size of the heartbeat alarm window in seconds before the 'heartbeat' flag is
 * applied to an alarm.
 * 
 * A server flags an alarm with a heartbeat when it has persistent in the local alarm list for 
 * a time span equal to this window.
 * Default is 1200 seconds.  The caller can set this window with this call.
 *
 * \param timeInSeconds is the size of the heartbeat window.
 *
 */ 
TINE_EXPORT void SetAlarmHeartbeat(int timeInSeconds);

/* GetAlarmHeartbeat() */
/**
 * \brief Gets the size of the heartbeat alarm window in seconds before the 'heartbeat' flag is
 * applied to an alarm.
 * 
 * A server flags an alarm with a heartbeat when it has persistent in the local alarm list for 
 * a time span equal to this window.
 * Default is 1200 seconds.  The caller can discover this time window with this call.
 *
 * \param timeInSeconds is the size of the heartbeat window.
 *
 */ 
TINE_EXPORT int GetAlarmHeartbeat(void);

/* SetAlarmOscillationWindow() */
/**
 * \brief Sets the size of the alarm oscillation window in clear counts.
 * 
 * A call to a member of the ClearAlarm() family will increment the alarm's 
 * clear counter.  Typically ClearAlarm() is always called at the beginning of
 * an i/o loop and SetAlarm() is called if warranted according to alarm conditions 
 * later in the same loop.  If the clear counter has reached the current oscillation
 * window setting, the alarm is then marked for termination and removal.
 * If the clear counter has been incremented by more than a single count
 * AND a SetAlarm() is then called, the alarm is marked as 'oscillating' as it 
 * appears to the alarm sub system that the alarm conditions are not persistent by
 * appear to come and go.
 * The oscillation window is by default set to 8 counts.  As a terminated alarm will
 * nonetheless reside in the local alarm tables for a time equal to the current 
 * termination window (default = 30 seconds) the oscillation window will automatically
 * 'grow' if an alarm has been marked as terminated but reappears within the termination
 * window.  The maximum oscillation window will however not exceed 128 (corresponding
 * to approximately 2 minutes, assuming a 1 Hz i/o loop is in play).
 * A call to this routine will not only fix the alarm oscillation window but 'pin' it
 * to the value given (i.e. so that it will not automatically grow).
 *
 * \param clearCounts is the size of the alarm oscillation window. (default = 8).
 *
 * \note calling this routine will pin the oscillation window to the given setting
 * so that the value will not 'grow' following the re-setting of a terminated alarm.
 *
 * \note the alarm oscillation window is applied to ALL alarms independent of alarm
 * code or device.  Once called, this will effectively 'pin' the alarm oscillation
 * value for all alarm types to this value.
 * Applying an oscillation window to a specific alarm can be achieved by
 * calling SetAlarmCodeOscillationWindow();
 *
 * \see GetAlarmOscillationWindow(), SetAlarmCodeOscillationWindow()
 */ 
TINE_EXPORT int SetAlarmOscillationWindow(int clearCounts);

/* SetAlarmCodeOscillationWindow() */
/**
 * \brief Sets the size of the alarm oscillation window in clear counts.
 * 
 * A call to a member of the ClearAlarm() family will increment the alarm's 
 * clear counter.  Typically ClearAlarm() is always called at the beginning of
 * an i/o loop and SetAlarm() is called if warranted according to alarm conditions 
 * later in the same loop.  If the clear counter has reached the current oscillation
 * window setting, the alarm is then marked for termination and removal.
 * If the clear counter has been incremented by more than a single count
 * AND a SetAlarm() is then called, the alarm is marked as 'oscillating' as it 
 * appears to the alarm sub system that the alarm conditions are not persistent by
 * appear to come and go.
 * The oscillation window is by default set to 8 counts.  As a terminated alarm will
 * nonetheless reside in the local alarm tables for a time equal to the current 
 * termination window (default = 30 seconds) the oscillation window will automatically
 * 'grow' if an alarm has been marked as terminated but reappears within the termination
 * window.  The maximum oscillation window will however not exceed 128 (corresponding
 * to approximately 2 minutes, assuming a 1 Hz i/o loop is in play).
 * A call to this routine will not only fix the alarm oscillation window but 'pin' it
 * to the value given (i.e. so that it will not automatically grow).
 *
 * Unlike SetAlarmOscillationWindow(), this routine can be used to apply an oscillation
 * window to a specific alarm code, without affecting the behavior of other alarm types.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param code is the alarm code for which the oscillation window is to be applied
 * \param clearCounts is the size of the alarm oscillation window. (default = 8).
 *
 * \note calling this routine will pin the oscillation window to the given setting
 * so that the value will not 'grow' following the re-setting of a terminated alarm.
 *
 *
 * \see SetAlarmOscillationWindow(), GetAlarmCodeOscillationWindow()
 */ 
TINE_EXPORT int SetAlarmCodeOscillationWindow(const char *eqm,int code, int value);

/* GetAlarmOscillationWindow() */
/**
 * \brief Gets the size of the alarm oscillation window in clear counts.
 * 
 * A call to a member of the ClearAlarm() family will increment the alarm's 
 * clear counter.  Typically ClearAlarm() is always called at the beginning of
 * an i/o loop and SetAlarm() is called if warranted according to alarm conditions 
 * later in the same loop.  If the clear counter has reached the current oscillation
 * window setting, the alarm is then marked for termination and removal.
 * If the clear counter has been incremented by more than a single count
 * AND a SetAlarm() is then called, the alarm is marked as 'oscillating' as it 
 * appears to the alarm sub system that the alarm conditions are not persistent by
 * appear to come and go.
 * The oscillation window is by default set to 8 counts.  As a terminated alarm will
 * nonetheless reside in the local alarm tables for a time equal to the current 
 * termination window (default = 30 seconds) the oscillation window will automatically
 * 'grow' if an alarm has been marked as terminated but reappears within the termination
 * window.  The maximum oscillation window will however not exceed 128 (corresponding
 * to approximately 2 minutes, assuming a 1 Hz i/o loop is in play).
 *
 * \return the size of the alarm oscillation window. (default = 8).
 *
 * \see SetAlarmOscillationWindow()
 */ 
TINE_EXPORT int GetAlarmOscillationWindow(void);

/* GetAlarmCodeOscillationWindow() */
/**
 * \brief Gets the size of the alarm oscillation window in clear counts.
 * 
 * A call to a member of the ClearAlarm() family will increment the alarm's 
 * clear counter.  Typically ClearAlarm() is always called at the beginning of
 * an i/o loop and SetAlarm() is called if warranted according to alarm conditions 
 * later in the same loop.  If the clear counter has reached the current oscillation
 * window setting, the alarm is then marked for termination and removal.
 * If the clear counter has been incremented by more than a single count
 * AND a SetAlarm() is then called, the alarm is marked as 'oscillating' as it 
 * appears to the alarm sub system that the alarm conditions are not persistent by
 * appear to come and go.
 * The oscillation window is by default set to 8 counts.  As a terminated alarm will
 * nonetheless reside in the local alarm tables for a time equal to the current 
 * termination window (default = 30 seconds) the oscillation window will automatically
 * 'grow' if an alarm has been marked as terminated but reappears within the termination
 * window.  The maximum oscillation window will however not exceed 128 (corresponding
 * to approximately 2 minutes, assuming a 1 Hz i/o loop is in play).
 *
 * Unlike GetAlarmOscillationWindow(), this routine targets a specific alarm code.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param code is the alarm code for which the oscillation window is to be applied
 *
 * \return the size of the alarm oscillation window (default = 8), or, if negative,
 * the negative value of a TINE return code (e.g. non_existent if the given equipment 
 * module does not exist or the alarm code is unknown).
 *
 * \see SetAlarmCodeOscillationWindow()
 */ 
TINE_EXPORT int GetAlarmCodeOscillationWindow(const char *eqm,int code);

/* SetAlarmCollapseWindow() */
/**
 * \brief Sets the size of the alarm collapse window in counts.
 * 
 * At any given time the prevailing alarm conditions might give rise to an
 * 'alarm storm', where an unusally large number of alarms are seen (due to e.g.
 * hardware modules that are turned off during a shutdown).  Such alarms are generally
 * 'not interesting' and should not create extra network traffic, resource allocation,
 * archiving, etc. if this can be avoided. 
 * As such, if the number of alarms for a given equipment module exceeds the
 * current 'alarm collapse window' setting, the alarm information will be 'compressed'
 * to a single alarm indicating that there are 'more than' the collapse window number
 * of alarms.
 * The value of the collapse window can be set via this call.
 *
 * \param value is the size of the alarm collapse window. (default = 100).
 *
 * \note the alarm collapse window is applied to ALL alarms independent of alarm
 * code or device.
 *
 * \see GetAlarmCollapseWindow()
 */ 
TINE_EXPORT int SetAlarmCollapseWindow(int value);

/* GetAlarmCollapseWindow() */
/**
 * \brief Gets the size of the alarm collapse window in counts.
 * 
 * At any given time the prevailing alarm conditions might give rise to an
 * 'alarm storm', where an unusally large number of alarms are seen (due to e.g.
 * hardware modules that are turned off during a shutdown).  Such alarms are generally
 * 'not interesting' and should not create extra network traffic, resource allocation,
 * archiving, etc. if this can be avoided. 
 * As such, if the number of alarms for a given equipment module exceeds the
 * current 'alarm collapse window' setting, the alarm information will be 'compressed'
 * to a single alarm indicating that there are 'more than' the collapse window number
 * of alarms.
 *
 * \return the size of the alarm collapse window. (default = 100).
 *
 * \see SetAlarmCollapseWindow()
 */ 
TINE_EXPORT int GetAlarmCollapseWindow(void);

/* setAutoLinkErrorAlarms() */
/**
 * \fn TINE_EXPORT void SetAutoLinkErrorAlarms(int value)
 * \brief Sets the autoLinkErrorAlarms flag.
 * 
 * A server can set automatic 'link_error' alarms when a link to another
 * server goes down.  This call sets the value of this flag,
 * (default = TRUE).
 *
 * \param value is the desired setting of this flag (default = TRUE)
 *
 * \see GetAutoLinkErrorAlarms
 */ 
TINE_EXPORT void SetAutoLinkErrorAlarms(int value);
/* GetAutoLinkErrorAlarms() */
/**
 * \fn TINE_EXPORT int GetAutoLinkErrorAlarms(void)
 * \brief Gets the current setting of the autoLinkErrorAlarms flag.
 * 
 * A server can set automatic 'link_error' alarms when a link to another
 * server goes down.  This call retrieves the current value of this setting,
 * (default = TRUE).
 *
 * \return the current setting of the autoLinkErrorAlarms flag.
 *
 * \see SetAutoLinkErrorAlarms
 */ 
TINE_EXPORT int GetAutoLinkErrorAlarms(void);

#define MAX_MINDISKSPACE_TBL_SIZE 10
typedef struct 
{
  char path[128];
  unsigned long minFreeBlks;
  unsigned long availBlks;
} MinDiskSpaceTblEntry;
extern int gMinDiskSpaceTblSize;
extern MinDiskSpaceTblEntry minDiskSpaceTbl[MAX_MINDISKSPACE_TBL_SIZE];

/* SetFreeBlocksAlarmThreshold() */
/**
 * \brief Sets the threshold (in blocks) used in setting 'low disk space' alarms.
 * 
 * A server can set automatic 'low_disk_space' alarms by setting this threshold
 * to a positive non-zero number for the disk path in question.  
 * (default = 0).
 *
 * \param path gives a path defining a disk mount to be monitored
 * \param minFreeBlocks gives the minimum number of (1-Kbyte) blocks which 
 * should be available before an alarm is issued.
 *
 * \return 0 upon success of a tine error code (resources_exhausted if an attempt
 * to monitor more than 10 disks is attempted.
 *
 * \note This routine is only available for those operating systems with disks!
 *
 * \b Example
 * 
 * \include eg_SetFreeBlocksAlarm.c
 */ 
TINE_EXPORT int SetFreeBlocksAlarmThreshold(char *path, UINT32 minFreeBlocks);

/**
 * \brief Sets the threshold (in bytes) used in setting 'low main memory' alarms.
 *
 * A server can set automatic 'low_main_memory' alarms by setting this threshold
 * to a positive non-zero number.
 * (default = 0).
 *
 * \param minAvailBytes gives the minimum number of bytes which
 * should be available main memory before an alarm in is issued.
 *
 */
TINE_EXPORT void SetAvailableMemoryThreshold(UINT32 minAvailBytes);

extern int AlarmServerInitialized;
#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _ALMLIB_H_ */
