/*
 * File hstweb.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "stdlib.h"
#include "math.h"
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include "tine.h"
#include "listener.h"
#include "archive.h"

#ifdef UNIX
char *strcasestr(const char *haystack, const char *needle);
#endif
#ifdef UNIX
pthread_attr_t webThreadAttr, *ptrWebThreadAttr = &webThreadAttr;
pthread_mutex_t webSrvMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hWebSrvMutex = &webSrvMutex;
pthread_mutex_t webRspMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hWebRspMutex = &webRspMutex;
#else
MXHANDLE hWebSrvMutex;
MXHANDLE hWebRspMutex;
#endif

typedef struct 
{
  char name[128];
  int ival;
} NAME128I;

int getArchivedData(char *keyword, int index, int fieldindex, time_t starttime,
  time_t stoptime, int arraysize, BYTE *data, int datafmt, char *datatag, int sample,
  float upr, float lwr, double *ts);
int getIndexFromDeviceName(char *devName, char *devProperty);
int getArchivedKeywordsEx(NAME64 *list, int *listsize, int *arrayType, int *prpSize);
int getArchivedKeywords(NAME64 *list, int *listsize);

extern int nkeywords;
NAME64 keys[2048];
SOCKET startup(UINT16 port);

UINT16 webSrvPort = 8080;
int webDbg = 0;
int get_line(SOCKET sck, char *buf, int size)
{
  int i = 0;
  char c = '\0';
  int n;
  struct pollfd fds[1];

  while ((i < size - 1) && (c != '\n'))
  {
    fds[0].fd = sck; fds[0].events = POLLIN;
    if (!poll(fds, 1, 50))
    {
      if (webDbg) dbglog("timeout waiting on web client socket");
      break;
    }
    n = recv(sck, &c, 1, 0);
    if (n < 0)
    { 
      if (webDbg) dbglog("error on web client socket");
      rprtSckErr("get_line", "recv", sck, 1, NULL);
    }
    if (n > 0)
    {
      if (c == '\r')
      {
        n = recv(sck, &c, 1, MSG_PEEK);
        if ((n > 0) && (c == '\n'))
          recv(sck, &c, 1, 0);
        else
          c = '\n';
      }
      buf[i] = c;
      i++;
    }
    else
    {
      //c = '\n';
      if (webDbg) dbglog("no input on web client socket");
      break;
    }
  }
  buf[i] = '\0';

  return i;
}

#define SERVER_STRING "Server: hsthttpd/0.1.0\r\n"
char *hdr = "<HTTP/1.0 200 OK\r\nServer: hsthttpd/0.1.0\r\nContent-Type: text/html;charset=UTF-8\r\n\r\n";
void headers(SOCKET sck,int isJSON)
{
  char buf[1024];
  char *ct = isJSON ? "application/json" : "text/html;charset=UTF-8";
  //strncpy(buf, hdr, sizeof(buf));
  //send(sck, buf, strlen(buf), 0);
  //return;
  strcpy(buf, "HTTP/1.0 200 OK\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  strcpy(buf, SERVER_STRING);
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "Content-Type: %s\r\n", ct);
  send(sck, buf, (int)strlen(buf), 0);
  strcpy(buf, "\r\n");  // blank line
  send(sck, buf, (int)strlen(buf), 0);
}
void unimplemented(SOCKET sck)
{
  char buf[1024];
  sprintf(buf, "HTTP/1.0 501 Method Not Implemented\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, SERVER_STRING);
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "Content-Type: text/html\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "<HTML><HEAD><TITLE>Method Not Implemented\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "</TITLE></HEAD>\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "<BODY><P>HTTP request method not supported.\r\n");
  send(sck, buf, (int)strlen(buf), 0);
  sprintf(buf, "</BODY></HTML>\r\n");
  send(sck, buf, (int)strlen(buf), 0);
}
void cat(SOCKET sck, FILE *resource)
{
  char buf[1024];

  fgets(buf, sizeof(buf), resource);
  while (!feof(resource))
  {
    send(sck, buf, (int)strlen(buf), 0);
    fgets(buf, sizeof(buf), resource);
  }
}
#define HTTP_REQ_BUFFER_SIZE 65536
#define HTTP_REQ_MAX_LINES 100
char reqbuffer[HTTP_REQ_BUFFER_SIZE];
char *reqlines[HTTP_REQ_MAX_LINES];
char *pagebuffer = NULL;
int pagebufferlen = 32768;
char pagetarget[64];
int pagetargetlen = 0;
DBLDBL trenddata[2000];
NAME64I trendnames[100];

#ifdef WIN32
# define TRENDFMTR "[%g,%I64d]%s"
# define TSFMTR "%I64d"
#else
# define TRENDFMTR "[%g,%lld]%s"
# define TSFMTR "%lld"
#endif
#define JSON_ARRAY_CONT "},\n{\n"
#define JSON_ARRAY_END "}\n]\n"
#define GRAFANA_TIMESERIES 1
#define GRAFANA_TABLE 2
#define SEPTOK '@'
// forget null-pointer checks for the internal routines below ...
int getTrendPoints(char *keywrd, char *dev, time_t start, time_t stop, int isstring, int maxpoints, void *data)
{
  int npts = 0, cc, index = 0;
  char prp[128], tgt[256];
  double ts, dv;
  NAME64 nam;
  DTYPE dout;
  int fmt = isstring ? CF_NAME64 : CF_DOUBLE;
  strncpy(prp, keywrd, 128);
  if (dev == NULL)
  { // grafana style
    if ((dev=strchr(prp, SEPTOK)) != NULL) *dev++ = 0;
  }
  if (dev != NULL)
  {
    index = getIndexFromDeviceName(dev, prp);
    if (index < 0) index = 0;
  }
  else
  {
    dev = "#0";
  }
  sprintf(tgt, "/%.32s/ARCHIVER/%.64s", gDeviceContext, dev);
  DTYPEZERO(dout);
  dout.dFormat = fmt;
  dout.dArrayLength = 1;
  dout.data.vptr = isstring ? (void *)&nam : (void *)&dv;
  cc = alsnExecLinkEx(tgt, prp, &dout, NULL, CA_READ, 500, &ts);
  if (maxpoints > 2000) maxpoints = 2000;
  if (cc != 0 || maxpoints > 1)
  {
    npts = getArchivedData(prp, index, 0, start, stop, maxpoints, (BYTE*)data, fmt, NULL, 0, 0, 0, NULL);
    if (npts < 0) npts = 0;
  }
  if (npts < maxpoints && cc == 0)
  {
    if (isstring)
    {
      strncpy(((NAME64I *)data)[npts].name, nam.name, 64);
      ((NAME64I *)data)[npts].ival = (int)ts;
    }
    else
    {
      ((DBLDBL *)data)[npts].d1val = dv;
      ((DBLDBL*)data)[npts].d2val = ts;
    }
    npts++;
  }
  return npts;
}
void sendJSONTarget(SOCKET sck, char *keywrd,time_t start,time_t stop,int maxpoints)
{
  int i, npts;
  char *sep;
  npts = getTrendPoints(keywrd, NULL, start, stop, FALSE, maxpoints, trenddata);
  sprintf(pagebuffer, "\"target\":\"%.64s\",\n",keywrd);
  send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
  if (webDbg) dbglog("-> %s", pagebuffer);
  strcpy(pagebuffer, "\"datapoints\":[\n");
  send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
  if (webDbg) dbglog("-> %s", pagebuffer);
  for (i = 0; i < npts; i++)
  {
    sep = i < npts - 1 ? ",\n" : "\n";
    sprintf(pagebuffer,TRENDFMTR,trenddata[i].d1val,(SINT64)(trenddata[i].d2val*1000),sep);
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    if (webDbg && i < 3) dbglog("-> %s", pagebuffer);
  }
  strcpy(pagebuffer, "]\n"); 
  send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
  if (webDbg) dbglog("-> %s", pagebuffer);
}
int tz_offset = 0;
time_t getUtcTime(char *buf)
{ // e.g. "2016-10-31T06:33:44.866Z"
  int yy, mo, dd, hh, mm, ss, offset;
  char tmp[32], *c = buf;
  struct tm tm;
  strncpy(tmp, c, 4); tmp[4] = 0;
  yy = atoi(tmp); c += 5;
  strncpy(tmp, c, 2); tmp[2] = 0;
  mo = atoi(tmp); c += 3;
  strncpy(tmp, c, 2); tmp[2] = 0;
  dd = atoi(tmp); c += 3;
  strncpy(tmp, c, 2); tmp[2] = 0;
  hh = atoi(tmp); c += 3;
  strncpy(tmp, c, 2); tmp[2] = 0;
  mm = atoi(tmp); c += 3;
  strncpy(tmp, c, 2); tmp[2] = 0;
  ss = atoi(tmp); c += 3;
  // ends with 'Z' ?
  c += 3;
  offset = (*c == 'Z') ? tz_offset : 0;
  tm.tm_year = yy - 1900;
  tm.tm_mon = mo - 1;
  tm.tm_mday = dd;
  tm.tm_hour = hh;
  tm.tm_min = mm;
  tm.tm_sec = ss;
  tm.tm_isdst = -1;
  return mktime(&tm) + offset;
}
char *getPtrJsonElem(char *buf, char *elem, char stopper)
{
  char *c, sstr[128];
  int len;
  snprintf(sstr,127,"\"%.127s\"",elem);
  len = (int)strlen(sstr);
  if ((c=strstr(buf, sstr)) == NULL) return NULL;
  for (c += len; *c != stopper; c++) if (*c == 0) return NULL;
  c++;
  return c;
}
void getMetricType(char *buf, int *type)
{
  char *c = getPtrJsonElem(buf, "type",':');
  if (c == NULL) return;
  while (*c != '\"') c++;
  *type = !strncmp(++c, "table", 9) ? GRAFANA_TABLE : GRAFANA_TIMESERIES;
}
void getMaxPoints(char *buf, int *pts)
{
  char *c = getPtrJsonElem(buf,"maxDataPoints",':');
  if (c != NULL) *pts = atoi(c);
  return;
}
void getTargets(char *buf, NAME128I *tgts, int *ntgts)
{
  int i = 0;
  char *e, *c, *rs = getPtrJsonElem(buf, "targets", ':');
  if (rs == NULL) return;
  for (c = rs; *c != 0; c++) if (*c == '[') break;
  if (*c == 0) return;
  while (*c != ']')
  {
    if (*c == '{')
    {
      if ((c = getPtrJsonElem(c, "target", ':')) == NULL) break;
      while (*c != '\"') c++;

      strncpy(tgts[i].name, ++c, 128);
      if ((e = strchr(tgts[i].name, '\"')) != 0) *e = 0;
      if ((c = getPtrJsonElem(c, "type", ':')) != NULL)
      {
        while (*c != '\"') c++;
        tgts[i].ival = !strncmp(++c, "table", 9) ? GRAFANA_TABLE : GRAFANA_TIMESERIES;
      }
      if (i++ >= *ntgts) break;
      c = strchr(c, '}');
    }
    c++;
  }
  *ntgts = i;
}
void getRange(char *buf, time_t *start, time_t *stop)
{
  char *c, *rs = getPtrJsonElem(buf, "range",'{');
  if (rs == NULL) return;
  for (c=rs; *c != 0; c++)
  {
    if (*c == '{')
    { // skip "raw" subsection
      c = strchr(c, '}');
      c++;
    }
    if (*c == '\"')
    {
      c++;
      if (!strncmp(c, "from\"", 5))
      { //start time
        c += 5;
        if ((c=strchr(c, '\"')) == NULL) return;
        c++; 
        *start = getUtcTime(c);
      }
      else if (!strncmp(c, "to\"", 3))
      {
        c += 3;
        if ((c = strchr(c, '\"')) == NULL) return;
        c++;
        *stop = getUtcTime(c);
      }
#     ifdef PARSE_RAW_RANGE
      else if (!strncmp(c, "raw\"", 4))
      { // look at the raw time coming in (trap the grafana 1 hour problem)
        c += 4;
        if ((c = strchr(c, '\"')) == NULL) return;
        c++;
        if ((c = strstr(c, "to\"")) == NULL) return;
        c += 3;
        if ((c = strchr(c, '\"')) == NULL) return;
        c++;
        if (!strncmp(c, "now\"",4))
        { // ah ha!
          time_t tnow = time(NULL);
          int delta_t = (int)(tnow - *stop);
          *stop = tnow;
          *start = *start + delta_t;
        }
      }
#     endif
      if ((c = strchr(c, '\"')) == NULL) return;
      c++;
    }
  }

}
//char *page = "<!DOCTYPE HTML PUBLIC \" -//W3C//DTD HTML 4.01 Transitional//EN\"> <html><head><meta http - equiv = \"Content-Type\" content = \"text/html;charset=UTF-8\"><title>ACOP.NET API : Main Page</title><link href = \"tabs.css\" rel = \"stylesheet\" type = \"text/css\"><link href = \"doxygen.css\" rel = \"stylesheet\" type = \"text/css\"></head><body>hello acop</body></html>";
char *page = "<HTML><HEAD><TITLE>Index</TITLE></HEAD><BODY><P>Welcome to Archive Web Services ...</BODY></HTML>\r\n";
THREAD_TYPE accept_request(THREAD_PARAM lpvThreadParam)
{
  char buf[1024];
  int numchars, nreqlines;
  char method[255];
  char url[255];
  int i, j, nbytes;
  int nkeys;
  SOCKET sck = *(SOCKET *)lpvThreadParam;
  char *query_string = NULL, *c;
  char *cts = "Content-Type:", *reqbufptr = reqbuffer;
  FILE *fp = NULL;
  int cc = 0, isJSON = FALSE;
  time_t stopt = time(NULL);
  time_t startt = stopt - 3600 * 6;
  int maxpoints = 200;
  int metricType = GRAFANA_TIMESERIES;

  if (WaitForSystemMutex(hWebRspMutex, -1) != 0) ccerr(mutex_error);
  numchars = get_line(sck, buf, sizeof(buf));
  if (numchars == 0) ccerr(invalid_datarequest);
  i = 0; j = 0;
  while (!isspace(buf[j]) && (i < sizeof(method) - 1))
  {
    method[i] = buf[j];
    i++; j++;
  }
  method[i] = '\0';
  if (webDbg) dbglog("web method: %s",method);

  if (stricmp(method, "GET") && stricmp(method,"POST"))
  {
    unimplemented(sck);
    ccerr(invalid_datarequest);
  }
  i = 0;
  while (isspace(buf[j]) && (j < sizeof(buf))) j++;
  while (!isspace(buf[j]) && (i < sizeof(url) - 1) && (j < sizeof(buf)))
  {
    url[i] = buf[j];
    i++; j++;
  }
  url[i] = '\0';
  if (webDbg) dbglog("url: %s", url);
  if (stricmp(method, "GET") == 0)
  {
    query_string = url;
    while ((*query_string != '?') && (*query_string != '\0'))
      query_string++;
  }
  strcpy(buf, "GET is finished"); // put a tag in the buffer for logic below
  nreqlines = 0; reqbufptr = reqbuffer; nbytes = 0;
  while ((numchars > 0) && strcmp("\n", buf))
  { // read & discard request headers (break on first empty line)
    numchars = get_line(sck, buf, sizeof(buf));
    if (nbytes < HTTP_REQ_BUFFER_SIZE - 256)
    {
      strncpy(reqbufptr, buf, numchars);
      reqlines[nreqlines++] = reqbufptr;
      reqbufptr += numchars; *reqbufptr++ = 0;
      nbytes += numchars + 1;
    }
  }
  for (i = 0; i < nreqlines; i++)
  {
    if (!strnicmp(reqlines[i], cts, strlen(cts)))
    {
      if (strcasestr(reqlines[i], "application/json") != NULL)
        isJSON = TRUE;
    }
    if (webDbg) dbglog("REQ header line: %s", reqlines[i]);
  }
  // check the url
  if (!stricmp(url, "/search"))
  { // grafana search
    if (isJSON)
    {
      nreqlines = 0; reqbufptr = reqbuffer; nbytes = 0;
      numchars = get_line(sck, buf, sizeof(buf));
      while ((numchars > 0)) // && strcmp("\n", buf))
      { // read & discard request headers
        if (nbytes < HTTP_REQ_BUFFER_SIZE - 256)
        {
          strncpy(reqbufptr, buf, numchars);
          reqlines[nreqlines] = reqbufptr;
          reqbufptr += numchars; *reqbufptr++ = 0;
          nbytes += numchars + 1;
          if (webDbg) dbglog("search header line: %s", reqlines[nreqlines]);
          nreqlines++;
        }
        numchars = get_line(sck, buf, sizeof(buf));
      }
      headers(sck,TRUE);
      nkeys = 2048;
      getArchivedKeywords(keys, &nkeys);
      strcpy(pagebuffer, "["); c = &pagebuffer[1];
      for (i = 0; i < nkeys; i++)
      {
#       ifdef GRAFONA_MAP_RESPONSE
        sprintf(c, "{ \"text\": \"%.64s\", \"value\": %d}", keys[i].name, i + 1);
#       else
        sprintf(c, "\"%.64s\"",keys[i].name);
#       endif
        if (i < nkeys - 1) strcat(c, ",");
        if ((int)strlen(pagebuffer) > pagebufferlen - 256)
        {
          send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
          c = pagebuffer;
        }
        else
        {
          c = pagebuffer + (int)strlen(pagebuffer);
        }
      }
      strcat(c, "]");
      send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    }
  }
  if (!stricmp(url, "/query"))
  { // grafana query
    if (isJSON)
    {
      NAME128I targets[100];
      int ntargets = 100;
      nreqlines = 0; reqbufptr = reqbuffer; nbytes = 0;
      if ((numchars=get_line(sck, buf, sizeof(buf))) == 0)
      {
        ntargets = 0;
      }
      while ((numchars > 0)) // && strcmp("\n", buf))
      { // read & discard request headers
        if (nbytes < HTTP_REQ_BUFFER_SIZE - 256)
        {
          strncpy(reqbufptr, buf, numchars);
          reqlines[nreqlines] = reqbufptr;
          reqbufptr += numchars; *reqbufptr++ = 0;
          nbytes += numchars + 1;
#         ifdef WIN32
          if (webDbg) printf("query header line: %s\n", reqlines[nreqlines]);
#         else
          if (webDbg) dbglog("query header line: %s", reqlines[nreqlines]);
#         endif
          // parse out and note the targets ...
          getRange(reqlines[nreqlines], &startt, &stopt);
          getMaxPoints(reqlines[nreqlines], &maxpoints);
          getMetricType(reqlines[nreqlines], &metricType); // one size fits all here!
          getTargets(reqlines[nreqlines], targets, &ntargets);
          nreqlines++;
        }
        numchars = get_line(sck, buf, sizeof(buf));
      }
      headers(sck, TRUE);
      if (ntargets == 0)
      {
        strcpy(pagebuffer, "[]\n");
        send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
      }
      else
      {
        char *ctok = "";
        // timeserie response
        strcpy(pagebuffer, "[\n{\n");
        send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
        if (webDbg) dbglog("-> %s", pagebuffer);
        for (i = 0; i < ntargets; i++)
        {
          sendJSONTarget(sck, targets[i].name, startt, stopt, maxpoints);
          ctok = i < ntargets - 1 ? JSON_ARRAY_CONT : JSON_ARRAY_END;
          strcpy(pagebuffer, ctok);
          send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
          if (webDbg) dbglog("-> %s", pagebuffer);
        }
      }
    /* */
    }
  }
  if (!stricmp(url, "/annotations"))
  { // grafana annotations (from a POST request)
    time_t stopt = time(NULL);
    time_t startt = stopt - 3600 * 6;
    /*
    The JSON request body looks like this:
    {
    "range": {
    "from": "2016-04-15T13:44:39.070Z",
    "to": "2016-04-15T14:44:39.070Z"
    },
    "rangeRaw": {
    "from": "now-1h",
    "to": "now"
    },
    "annotation": {
    "name": "deploy",
    "datasource": "Simple JSON Datasource",
    "iconColor": "rgba(255, 96, 96, 1)",
    "enable": true,
    "query": "#deploy"
    }
    }
    */
    if (isJSON)
    {
      nreqlines = 0; reqbufptr = reqbuffer; nbytes = 0;
      numchars = get_line(sck, buf, sizeof(buf));
      while ((numchars > 0)) // && strcmp("\n", buf))
      { // read & discard request headers
        if (nbytes < HTTP_REQ_BUFFER_SIZE - 256)
        {
          strncpy(reqbufptr, buf, numchars);
          reqlines[nreqlines] = reqbufptr;
          reqbufptr += numchars; *reqbufptr++ = 0;
          nbytes += numchars + 1;
          if (webDbg) dbglog("annotations header line: %s", reqlines[nreqlines]);
          getRange(reqlines[nreqlines], &startt, &stopt);
          nreqlines++;
        }
        numchars = get_line(sck, buf, sizeof(buf));
      }
    }
    /*
    //Grafana expects a response containing an array of annotation objects in the following format:
    [
    {
    annotation: annotation, // The original annotation sent from Grafana.
    time: time, // Time since UNIX Epoch in milliseconds. (required)
    title: title, // The title for the annotation tooltip. (required)
    tags: tags, // Tags for the annotation. (optional)
    text: text // Text for the annotation. (optional)
    }
    ]
    */
    headers(sck, TRUE);
    // Example timeserie response
    strcpy(pagebuffer, "[\n{\n");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    strcpy(pagebuffer, "\"annotation\":\"annotation for keyword\",");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    i = (int)(startt + (stopt - startt) / 2);
    sprintf(pagebuffer, "\"time\": %d000,",i);
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    strcpy(pagebuffer, "\"title\":\"title for keyword\",");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    strcpy(pagebuffer, "\"text\":\"annotation text for keyword\",");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);

    strcpy(pagebuffer, "},\n{\n"); // when there's more coming ...
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    strcpy(pagebuffer, "\"annotation\":\"annotation for keyword\",");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    sprintf(pagebuffer, "\"time\": %d000,",i+100);
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    strcpy(pagebuffer, "\"title\":\"title for keyword\",");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
    strcpy(pagebuffer, "\"text\":\"annotation text for keyword\",");
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);

    strcpy(pagebuffer, "}\n]\n");  // close all brackets ...
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
  }
  // not a grafana url ...
  if (!strnicmp(url, pagetarget, pagetargetlen))
  {
    time_t stopt = time(NULL);
    time_t strtt = stopt - 3600 * 8;
    //DBLDBL data[2000];
    char *keywrd = "curDC";
    char *dev = NULL;
    char *c = &url[pagetargetlen];
    int n, rec, p, index = 0, isString=FALSE;
    int npts = 0;
    int depth = 2000;
    void *data = NULL;
    if (*c == '/')
    {
      keywrd = ++c;
      if ((c=stristr(keywrd, "&POINTS=")) != NULL)
      {
        depth = atoi(&c[8]);
        *c++ = 0;
        if (depth < 1) depth = 1;
        if (depth > 2000) depth = 2000;
      }
      if ((c=strchr(keywrd, '/')) != NULL)
      {
        *c++ = 0;
        dev = c;
      }
    }
    if (webDbg) dbglog("looking for keyword %.64s @ index %d",keywrd,index);

    if ((n=getArchivedEntryFromKeyword(keywrd, &rec, &p)) >= 0)
    {
      isString = isStringFormat(daqTbl[n].daqParam->fmt);
      if (isString && depth > 100) depth = 100;
    }
    data = isString ? (void *)trendnames : (void *)trenddata;
    npts = getTrendPoints(keywrd, dev, strtt, stopt, isString, depth, data);
    //npts = getArchivedData(keywrd, index, 0, strtt, stopt, 200, (BYTE *)data, CF_DOUBLE, NULL, 0, 0, 0, NULL);
    if (npts > 0)
    {
      int i, len;
      c = pagebuffer;
      strcpy(pagebuffer, "<HTML><HEAD><TITLE>Data</TITLE></HEAD><BODY>");
      for (i = 0; i < npts; i++)
      {
        len = (int)strlen(pagebuffer);
        if (len > pagebufferlen - 128) break;
        c = &pagebuffer[len];
        if (isString)
        {
          sprintf(c, "<P>%.32s %.64s", GetDataTimeString((double)trendnames[i].ival, 0), trendnames[i].name);
        }
        else
        {
          sprintf(c, "<P>%.32s %g", GetDataTimeString(trenddata[i].d2val, 0), trenddata[i].d1val);
        }
      }
      strcat(c, "</BODY></HTML>\r\n");
      if (webDbg) dbglog("prepared page of length %d characters",strlen(pagebuffer));
    }
    headers(sck, FALSE);
    send(sck, pagebuffer, (int)strlen(pagebuffer), 0);
  }
  if (!strcmp(url, "/"))
  {
    headers(sck, FALSE);
    strncpy(buf, page, sizeof(buf));
    send(sck, buf, (int)strlen(buf), 0);
  }
  soclose(sck);
err:
  ReleaseSystemMutex(hWebRspMutex);
  return 0;
}
int webSrvTask(void)
{
  THRHANDLE thr = 0;
  SOCKET psck = 0;
  SOCKET sck = -1;
  struct sockaddr_in cln;
  int len = sizeof(cln);
  int cc = 0;
  char *ptr = NULL;
# ifdef WIN32
  DWORD thrId;
# endif
  static int running = FALSE;
  struct tm *lt, *gmt;
  int ltm, lth, gmtm, gmth;
  time_t tnow = time(NULL);
  lt = localtime(&tnow);
  lth = lt->tm_hour; ltm = lt->tm_min;
  gmt = gmtime(&tnow);
  gmth = gmt->tm_hour; gmtm = gmt->tm_min;
  tz_offset = (lth - gmth) * 3600 + (ltm - gmtm) * 60;
  if (running) return already_assigned;
  if ((ptr = getenv("WEB_PORT")) != NULL)
  {
    webSrvPort = atoi(ptr);
  }

  if ((cc = InitMutex(&hWebSrvMutex)) != 0)
  {
    printf("Cannot initialize web server mutex");
    feclog("Cannot initialize web server mutex");
    return cc;
  }
  if ((cc = InitMutex(&hWebRspMutex)) != 0)
  {
    printf("Cannot initialize web data response mutex");
    feclog("Cannot initialize web data response mutex");
    return cc;
  }

  printf("httpd running on port %d\n", webSrvPort);
  feclog("httpd running on port %d", webSrvPort);
  if ((ptr = getenv("WEB_MAXPAGELENGTH")) != NULL)
  {
    pagebufferlen = strtol(ptr, NULL, 0);
  }
  pagebuffer = (char *)calloc(pagebufferlen, 1);
  feclog("maximum page size %d characters", pagebufferlen);

  RegisterUserCommand("WebDebug",NULL,&webDbg,NULL,CA_READ+CA_WRITE);

  sprintf(pagetarget, "/%.32s", gDeviceContext);
  if ((ptr = strchr(pagetarget, '.')) != NULL) *ptr = 0;
  pagetargetlen = (int)strlen(pagetarget);
  psck = startup(webSrvPort);
  while (TRUE)
  {
    if (ServerExitCondition) break;
    if (WaitForSystemMutex(hWebSrvMutex, -1) != 0) break;
    sck = accept(psck,(struct sockaddr *)&cln,&len);
    if (sck == -1)
    {
      rprtSckErr("webSrvTask", "accept", sck, 0, NULL);
      cc = tcp_socket_error;
      break;
    }
#   ifdef UNIX
    pthread_attr_setdetachstate(ptrWebThreadAttr, PTHREAD_CREATE_DETACHED);
    if (pthread_create(&thr, ptrWebThreadAttr, accept_request, &sck) != 0)
      perror("pthread_create");

#   elif defined(WIN32)
    thr = (THRHANDLE)CreateThread(NULL, 0, accept_request, &sck, 0, &thrId);
#   endif
    YieldThread();
    ReleaseSystemMutex(hWebSrvMutex);
  }
  soclose(psck);
  shutdown(psck, 2);
  return cc;
}
SOCKET startup(UINT16 port)
{
  SOCKET httpd = 0;
  struct sockaddr_in name;

  httpd = socket(PF_INET, SOCK_STREAM, 0);
  if (httpd == -1)
  {
    rprtSckErr("startup", "socket", httpd, 0, NULL);
    exit(1);
  }
  memset(&name, 0, sizeof(name));
  name.sin_family = AF_INET;
  name.sin_port = htons(port);
  name.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(httpd, (struct sockaddr *)&name, sizeof(name)) < 0)
  {
    rprtSckErr("startup", "bind", httpd, 0, NULL);
    exit(1);
  }
  if (listen(httpd, 5) < 0)
  {
    rprtSckErr("startup", "listen", httpd, 0, NULL);
    exit(1);
  }
  return(httpd);
}
