/*
 * File websrv.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "app_version.h"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#include "tine.h"
#include "archive.h"

int webSrvTask(void);
void PreSystemInit(void);
void PostSystemInit(void);
extern int needsRestart;
void RestartServer(void);
extern char adbFileName[64];

extern char thisFecName[FEC_NAME_SIZE];
extern char thisContext[CONTEXT_NAME_SIZE];
extern char thisLocation[FEC_LOCATION_SIZE];
extern char thisServerName[EXPORT_NAME_SIZE];
extern char thisFailoverMasterName[EXPORT_NAME_SIZE];
extern char thisFailoverSlaveMasterName[EXPORT_NAME_SIZE];
extern UINT16 thisPortOffset;
extern int startupInputsGiven;
extern int webDbg;
extern UINT16 webSrvPort;

char *usageText = "usage: hstweb [/f=<fecname> /c=<context> /s=<server name> /p=<port offset> /l=<location> /r=<responsible> /d=<debug level> /wp=<web port> /wd=<web debug>";

THREAD_TYPE launchWebSrvThread(THREAD_PARAM lpvThreadParam)
{
  webSrvTask();  
}
void launchWebSrvTask(void)
{
  int cc;
# ifdef WIN32
  DWORD dwThrId = 0;
  THRHANDLE hThr = (THRHANDLE)CreateThread(NULL, 0, launchWebSrvThread, NULL, 0, &dwThrId);
# else
  pthread_t tid;
  cc = pthread_create(&tid,NULL,launchWebSrvThread,0);
# endif
  if (cc != 0) feclog("could not create background thread : %s",strerror(errno));
  return;
}

int main(int argc, char *argv[])
{
  int i,cc,failoverErrorCount=0;
  char *c, scratch[256];

  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'f':
      {
        char *dst = thisFecName;
        int len = FEC_NAME_SIZE;
        switch (tolower(c[1]))
        {
          case 'm': /* failover master specified */
            dst = thisFailoverMasterName;
            len = EXPORT_NAME_SIZE;
            break;
          case 's': /* failover slave master specified */
            dst = thisFailoverSlaveMasterName;
            len = EXPORT_NAME_SIZE;
            break;
          default:
            startupInputsGiven = TRUE;
            break;
        }
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(dst,c,len);
        }
        continue;
      }
      case 'e':
      {
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          failoverErrorCount = atoi(scratch);      
        }
        continue;      
      }
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisContext,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisServerName,c,EXPORT_NAME_SIZE);
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisLocation,c,FEC_LOCATION_SIZE);
        }
        continue;
      case 'p':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          thisPortOffset = atoi(scratch);      
        }
        startupInputsGiven = TRUE;
        continue;
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(adbFileName,c,64);
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case 'w':
      {
        int assignWebPort = FALSE;
        int assignWebDebug = FALSE;
        c++;
        if (tolower(*c) == 'p') assignWebPort = TRUE;
        if (*c == 'd') assignWebDebug = TRUE;
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          if (assignWebPort) webSrvPort = (UINT16)atoi(scratch);
          if (assignWebDebug) webDbg = atoi(scratch);
        }
        continue;
      }
      case '?':
      case 'h':
        printf("%s",usageText);
        exit(0);
      default:
        continue;
    }
  }

  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
  SetPreSystemInitFunction(PreSystemInit);
  SetPostSystemInitFunction(PostSystemInit);
  SetInterpretConsoleCommands(TRUE);
  /* init server: */
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }
  /* call FEC specific initialization: */
  if( failoverErrorCount) 
  {
    SetFailoverThreshold(failoverErrorCount);
  } 
  launchWebSrvTask();
  for(;;)
  {
    SystemCycle(TRUE);
#   ifndef MULTITHREADED
    if (needsRestart) RestartServer();
#   endif
    slave_mode_diagnostics();
  }
  return 0;
}
