/*
 * File devsrv.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "app_version.h"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include "tine.h"
#include "archive.h"

void PostSystemInit(void);
void PreSystemInit(void);

extern int g_isTerminating;
extern UINT32 gMininumFreeDiskSpace;
extern int globalsWatchdog;
extern char globalsMcastAddr[32];
extern char adbFileName[64];

extern char thisFecName[FEC_NAME_SIZE+4];
extern char thisContext[CONTEXT_NAME_SIZE+4];
extern char thisLocation[FEC_LOCATION_SIZE];
extern char thisServerName[EXPORT_NAME_SIZE];
extern char thisFailoverMasterName[EXPORT_NAME_SIZE];
extern char thisFailoverSlaveMasterName[EXPORT_NAME_SIZE];
extern UINT16 thisPortOffset;
extern int startupInputsGiven;
extern int thisWorkAreaSize;
extern int thisSocketRecvBufferSize;
extern int thisSocketSendBufferSize;

char *usageText = "usage: march [/f=<fecname> /c=<context> /s=<server> /p=<port offset> /fm=<failover master name> /fs=<failover slave master> /l=<location> /n=<database file name> /k=<minimumFreeDiskSpaceInKbytes> /w=<workAreaSize /brecv=<socketReceiveBuffers> /bsend=<socketSendBuffers> /d=<debug level>";
int main(int argc, char *argv[])
{
  int i,cc,failoverErrorCount=0;
  char *c, scratch[256];

  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'f':
      {
        char *dst = thisFecName;
        int len = FEC_NAME_SIZE;
        switch (tolower(c[1]))
        {
          case 'm': /* failover master specified */
            dst = thisFailoverMasterName;
            len = EXPORT_NAME_SIZE;
            break;
          case 's': /* failover slave master specified */
            dst = thisFailoverSlaveMasterName;
            len = EXPORT_NAME_SIZE;
            break;
          default:
            startupInputsGiven = TRUE;
            break;
        }
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(dst,c,len);
        }
        continue;
      }
      case 'e':
      {
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          failoverErrorCount = atoi(scratch);      
        }
        continue;      
      }
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisContext,c,CONTEXT_NAME_SIZE);
          startupInputsGiven = TRUE;
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisServerName,c,EXPORT_NAME_SIZE);
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisLocation,c,FEC_LOCATION_SIZE);
        }
        continue;
      case 'p':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          thisPortOffset = atoi(scratch);      
        }
        continue;
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(adbFileName,c,64);
        }
        continue;
      case 'k':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          gMininumFreeDiskSpace = strtol(scratch,NULL,0);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case 'g':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if (stricmp(scratch,"WATCHDOG") == 0)
          {
            globalsWatchdog = TRUE;
          }
        }
        continue;
      case 'w':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          thisWorkAreaSize = strtol(scratch,NULL,0);      
        }
        continue;
      case 'b':
      {
        int *tgt = NULL;
        switch (tolower(c[1]))
        {
          case 'r': /* recv buffer space */
            tgt = &thisSocketRecvBufferSize;
            break;
          case 's': /* failover slave master specified */
            tgt = &thisSocketSendBufferSize;
            break;
          default:
            break;
        }
        if (tgt == NULL) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          *tgt = strtol(scratch, NULL, 0);
        }
        continue;
      }
      case 'r':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          thisSocketRecvBufferSize = strtol(scratch,NULL,0);      
        }
        continue;
      case '?':
      case 'h':
        printf("%s",usageText);
        exit(0);
      default:
        continue;
    }
  }

  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
  SetPreSystemInitFunction(PreSystemInit);
  SetPostSystemInitFunction(PostSystemInit);
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",cc2str(cc)); exit(1); }
  if( failoverErrorCount) 
  {
    SetFailoverThreshold(failoverErrorCount);
  } 
  printf("Server running (debug level %d)\n",tineDebug);
  for(;;)
  {
    if (g_isTerminating) ttyquit(NULL,NULL);
    SystemCycle(TRUE);
    slave_mode_diagnostics();
  }
  return 0;
}
