#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "mstlib.h"

#if defined(MSTBEK_EXPORTS) 
# define LIB_EXPORT __declspec(dllexport)
# else
# define LIB_EXPORT __declspec(dllimport)
# endif

#define CONFIG_FILE "mstbek.csv"

#define CLBR_PATTERN 32641
#define MOVETO_DURATION 600

#define MTR_IDLE   0x0000
#define MTR_MOVE   0x0001
#define MTR_PAUSE  0x0002
#define MTR_REF1   0x0010
#define MTR_REF2   0x0020
#define MTR_REF3   0x0040
#define MTR_END_I  0x0100
#define MTR_END_O  0x0200
#define MTR_ISCLBR 0x0800
#define MTR_UNCLBR 0x1000
#define MTR_ERR    0x8000

#define MTR_CLBR (MTR_REF1|MTR_REF2|MTR_REF3)

NAME32 stsLst[] = {
  "moving",
  "paused",
  "",
  "",
  "calibrating (step1)",
  "calibrating (step2)",
  "calibrating (step3)",
  "",
  "at end switch (inner)",
  "at end switch (outer)",
  "",
  "is calibrating",
  "uncalibrated",
  "",
  "",
  "hardware error",
};

int sizStsLst = sizeof(stsLst)/sizeof(NAME32);

#define MTR_MIN_POSITION -128000000

#define MTR_TYPE_UNKNOWN  0x00
#define MTR_TYPE_LINEAR   0x01
#define MTR_TYPE_ROTATION 0x02

int getMotorType(char *typeStr)
{
  if (typeStr == NULL) return MTR_TYPE_UNKNOWN;
  if (!stricmp(typeStr,"LINEAR")) return MTR_TYPE_LINEAR;
  if (!stricmp(typeStr,"ROTATION")) return MTR_TYPE_ROTATION;
  return MTR_TYPE_UNKNOWN;
}

typedef struct
{
  char name[32];
  char reg[32];
  int type;
  int microStepsPerStep;
  int mstepsPerCycle;
  float cyclesPerMM;
  float mstepsPerMM;
  int posMicroSteps;
  int posMicroStepsRb; // readback position
  int tgtMicroSteps;
  int tgtMicroStepsRb; // readback target position
  int maxPos; // maximum position (micro steps)
  int minPos; // minimum position (micro steps)
  int offPos; // offset position (micro steps)
  int scalePos; // scale factor (usually -1.0 or 1.0)
  int velMean; // normal velocity (msteps / sec)
  int velRef1;
  int velRef2;
  float velRms; // velocity (msteps per second)
  float velLwr;
  float velUpr;
  float accUpr;
  int movCmd;
  int movEtr; // 'estimated time remaining' for move action
  int movPaused;
  int status;
  int cmdSts;
  int lnkStsPos;
  int lnkIdPos; 
  int lnkStsTgt;
  int lnkIdTgt; 
  time_t lastCheckedClbr;
  time_t lastCheckedEndsw;
} BeckoffMotor;

BeckoffMotor *motors = NULL;

char gTarget[128];
int gNumMotors = 1;
int gIsInProcess = FALSE;
void cbTgtPos(int id, int cc);
void startUpdateMotor(int mot);
LIB_EXPORT int moveto(int mot,int cmd);
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd);

int mm2microsteps(int mot,double val)
{ // val in millimeters
  return (int)(val * motors[mot].mstepsPerMM);
}
double microsteps2mm(int mot,int msteps)
{ 
  double val = (double)msteps;
  val /= (double)motors[mot].mstepsPerMM;
  return val;
}

double cycles2mm(int mot,double cycles)
{
  return motors[mot].cyclesPerMM * cycles;
}

typedef struct
{
  char target[128];
  char type[32];
  int motorNr;
  int mstepsStep;
  int mstepsCycle;
  int velNormal;
  int velRef1;
  int velRef2;
  int posMax;
  int posMin;
  int posOff;
  int posScale;
  float mstepsMM;
} MotorInfoStruct;

CSVCOLUMN csvBeckoffConfigColumns[] =
{
  { "TARGET","/localhost/cdi",0,CF_TEXT,128,(BYTE *)OFFSETIN(MotorInfoStruct,target),NULL,0 },
  { "TYPE","LINEAR",0,CF_TEXT,32,(BYTE *)OFFSETIN(MotorInfoStruct,type),NULL,0 },
  { "MOTOR","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,motorNr),NULL,0 },
  { "MSTEPS_MM","12800",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsMM),NULL,0 },
  { "MSTEPS_CYCLE","200",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsCycle),NULL,0 },
  { "MSTEPS_STEP","64",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsStep),NULL,0 },
  { "POSITION_MAXIMUM","768000",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMax),NULL,0 },
  { "POSITION_MINIMUM","0",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMin),NULL,0 },
  { "POSITION_OFFSET","0",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,posOff),NULL,0 },
  { "POSITION_SCALE","1",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,posScale),NULL,0 },
  { "VELOCITY_NORMAL","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,velNormal),NULL,0 },
  { "VELOCITY_REFERENCE_1","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,velRef1),NULL,0 },
  { "VELOCITY_REFERENCE_2","10",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,velRef2),NULL,0 },
};
CSVDB csvBeckoffConfigDb =
{
  csvBeckoffConfigColumns,
  sizeof(csvBeckoffConfigColumns)/sizeof(csvBeckoffConfigColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(MotorInfoStruct),
  0
};

NAME64 *mstNames = NULL;
NAME32 *regNames = NULL;

int getIntegerFromHardware(char *cmd, int *val)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];

  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = val;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"RECV",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendIntegerToHardware(char *cmd, int val)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];

  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  din.data.lptr = &val;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int getFloatFromHardware(char *cmd, float *val)
{
  int cc;
  DTYPE dout;
  char dev[128],errstr[64];
  if (val == NULL) return invalid_parameter;

  memset(&dout,0,sizeof(DTYPE));

  dout.dArrayLength = 1;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = val;

  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"RECV",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendFloatCmdToHardware(int mot,float *tgtVals)
{
  DTYPE dout, din;
  float rbVals[6];
  char dev[128];

  if (mot < 0 || mot > gNumMotors) return illegal_device;

  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = 4;
  din.dFormat = CF_FLOAT;
  din.data.fptr = tgtVals;

  dout.dArrayLength = 6;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = rbVals;

  sprintf(dev,"%s/%s.fltCMD",gTarget,motors[mot].name);
  return ExecLinkEx(dev,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE,200);
}
int sendIntegerCmdToHardware(int mot,int *tgtVals)
{
  int cc, retries=3;
  DTYPE dout, din;
  int rbVals[6];
  char dev[128];

  if (mot < 0 || mot > gNumMotors) return illegal_device;

retrySend:
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.lptr = tgtVals;

  dout.dArrayLength = 6;
  dout.dFormat = CF_INT32;
  dout.data.lptr = rbVals;

  sprintf(dev,"%s/%s.CMD",gTarget,motors[mot].name);
  cc = ExecLink(dev,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE);
  if (cc != 0) return cc;
  if (rbVals[0] != tgtVals[0]) return hardware_error;
  if (rbVals[1] != 0) 
  {
    if (retries-- > 0) goto retrySend;
    return operation_busy;
  }
  if (rbVals[2] != tgtVals[2]) return remitted_data_lost;
  return 0;
}
int sendRecvIntegerCmdToHardware(int mot,int *tgtVals,int *rbVals)
{
  DTYPE dout, din;
  char dev[128];

  if (mot < 0 || mot > gNumMotors) return illegal_device;

  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.lptr = tgtVals;

  dout.dArrayLength = 6;
  dout.dFormat = CF_INT32;
  dout.data.lptr = rbVals;

  sprintf(dev,"%s/%s.CMD",gTarget,motors[mot].name);
  return ExecLink(dev,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE);
}
int sendGroupCmdToHardware(int gid,int *tgtVals)
{
  int cc;
  DTYPE dout, din;
  int rbVals[6];
  char dev[128];

  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.lptr = tgtVals;

  dout.dArrayLength = 6;
  dout.dFormat = CF_INT32;
  dout.data.lptr = rbVals;

  sprintf(dev,"%s/CMDALL",gTarget);
  cc = ExecLink(dev,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE);
  if (cc != 0) return cc;
  if (rbVals[0] != tgtVals[0]) return hardware_error;
  if (rbVals[1] != 0) return operation_busy;
  if (rbVals[2] != tgtVals[2]) return remitted_data_lost;
  return 0;
}

int getRegisteredMotors(NAME64 *names,NAME32 *regnames, int num)
{
  int cc=0;
  DTYPE dout;
  char dev[128],errstr[64];

  if (names == NULL || regnames == NULL || num < 1) return invalid_parameter;

  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = num;
  dout.dFormat = CF_NAME64;
  dout.data.vptr = (NAME64 *)names;

  sprintf(dev,"%s/MOTOR",gTarget);
  if ((cc=ExecLinkEx(dev,"INSTANCES",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s INSTANCES from CDI: %s",dev,GetSystemErrorString((short)cc,errstr));
  }

  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = num;
  dout.dFormat = CF_NAME32;
  dout.data.vptr = (NAME32 *)regnames;
  sprintf(dev,"%s/MOTORREG",gTarget);
  if ((cc=ExecLinkEx(dev,"INSTANCES",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s INSTANCES from CDI: %s",dev,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
static MotorInfoStruct *motorInfo = NULL;
void initialize(void)
{
  int i, cc, n, nmotorsFromFile=0;
  char dev[64],errstr[64];

  if ((cc=csvReadFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb,(void **)&motorInfo)) != 0) 
  {
    feclog("cannot read startup configuration parameters from %s : %s",
      CONFIG_FILE,GetSystemErrorString((short)cc,errstr));
    exit(1);
  }
  nmotorsFromFile = csvBeckoffConfigDb.siz;

  if (nmotorsFromFile < 1)
  {
    feclog("no startup configuration parameters from %s given",CONFIG_FILE);
    exit(1);
  }
  strncpy(gTarget,motorInfo[0].target,128);
  gIsInProcess = (strnicmp(gTarget,"/localhost/cdi",14) == 0) ? TRUE : FALSE;

  if (getIntegerFromHardware("numMotor",&gNumMotors) != 0) exit(1);
  if ((motors=(BeckoffMotor *)calloc(gNumMotors,sizeof(BeckoffMotor))) == NULL)
  {
    feclog("cannot allocate space for %d motors",gNumMotors);
    exit(1);
  }
  if ((mstNames=(NAME64 *)calloc(gNumMotors,sizeof(NAME64))) == NULL)
  {
    feclog("cannot allocate motor name list ! (aborting ...)");
    exit(1);
  }
  if ((regNames=(NAME32 *)calloc(gNumMotors,sizeof(NAME32))) == NULL)
  {
    feclog("cannot allocate motor register name list ! (aborting ...)");
    exit(1);
  }
  if ((cc=getRegisteredMotors(mstNames,regNames,gNumMotors)) != 0)
  {
    feclog("cannot read template instances from CDI ! (aborting ...)");
    exit(1);
  }

  for (i=0; i<gNumMotors; i++)
  {
//    sprintf(mstNames[i].name,"Motor%d",i+1);
//    sprintf(motors[i].name,"Motor%d",i+1);
//    sprintf(motors[i].reg,"MotReg%d",i+1);
    if (strlen(mstNames[i].name) == 0) break;
    strncpy(motors[i].name,mstNames[i].name,32);
    strncpy(motors[i].reg,regNames[i].name,32);
    sprintf(dev,"%s.mSteps",motors[i].name);
    getIntegerFromHardware(dev,&motors[i].microStepsPerStep);
    sprintf(dev,"%s.fSteps",motors[i].name);
    getIntegerFromHardware(dev,&motors[i].mstepsPerCycle);
    motors[i].mstepsPerCycle *= motors[i].microStepsPerStep; // ??
    for (n=0; n<nmotorsFromFile; n++)
    {
      if (motorInfo[n].motorNr == i+1)
      {
        motors[i].maxPos = motorInfo[n].posMax;
        motors[i].minPos = motorInfo[n].posMin;
        motors[i].offPos = motorInfo[n].posOff;
        motors[i].scalePos = motorInfo[n].posScale;
        motors[i].velMean = motorInfo[n].velNormal;
        motors[i].velRef1 = motorInfo[n].velRef1;
        motors[i].velRef2 = motorInfo[n].velRef2;
        motors[i].microStepsPerStep = motorInfo[n].mstepsStep;
        motors[i].mstepsPerCycle = motorInfo[n].mstepsCycle;
        motors[i].mstepsPerMM = motorInfo[n].mstepsMM;
        motors[i].cyclesPerMM = motorInfo[n].mstepsMM/motorInfo[n].mstepsCycle;
        motors[i].type = getMotorType(motorInfo[n].type);
        break;
      }
    }
    sprintf(dev,"%s.Rms",motors[i].name);
    getFloatFromHardware(dev,&motors[i].velRms);
    sprintf(dev,"%s.LwrLmtV",motors[i].name);
    getFloatFromHardware(dev,&motors[i].velLwr);
    sprintf(dev,"%s.UprLmtV",motors[i].name);
    getFloatFromHardware(dev,&motors[i].velUpr);
    sprintf(dev,"%s.UprLmtA",motors[i].name);
    getFloatFromHardware(dev,&motors[i].accUpr);
    startUpdateMotor(i);
  }
  if (i < gNumMotors)
  {
    feclog("hardware claimed %d motors, but only %d configured",gNumMotors,i);
    gNumMotors = i;
  }
}

int SimulateData(int start,int range)
{
  /* This routine is useful for test servers  */
  return start + rand()%(range);
}
void cbPos(int id, int cc)
{
  static int zeroCounter = 0;
  if (id < 0 || id >= gNumMotors) return; // huh?
  motors[id].lnkStsPos = cc;
  if (motors[id].status & MTR_MOVE)
  { /* motor is moving */
    if (motors[id].posMicroStepsRb == 0 && motors[id].tgtMicroSteps != 0)
    { /* beckoff gives a spurious '0' reading from time to time */
      if (zeroCounter++ < 2) return;
    }
  }
  zeroCounter = 0;
  motors[id].posMicroSteps = motors[id].posMicroStepsRb;
  if (motors[id].posMicroSteps == motors[id].tgtMicroSteps)
  {
    if (motors[id].movCmd != PROC_STS_IDLE) moveto(id,PROC_STS_IDLE);
  }
}
void cbTgt(int id, int cc)
{
  if (id < 0 || id >= gNumMotors) return; // huh?
  motors[id].lnkStsTgt = cc;
  if (motors[id].tgtMicroStepsRb != motors[id].tgtMicroSteps)
  {
    feclog("Unexpected jump in target position ! (%d to %d)",
      motors[id].tgtMicroSteps,motors[id].tgtMicroStepsRb);
    motors[id].tgtMicroSteps = motors[id].tgtMicroStepsRb;
  }
}
#define MTR_UPDATE_INTERVAL 200
void startUpdateMotor(int mot)
{
  DTYPE dout;
  char dev[128];
  if (mot < 0 || mot > gNumMotors) return;

  memset(&dout,0,sizeof(DTYPE));

  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &motors[mot].posMicroStepsRb;

  sprintf(dev,"%s/%s.CurPos",gTarget,motors[mot].name);
  motors[mot].lnkIdPos = AttachLinkEx(dev,"RECV",&dout,NULL,CA_READ,MTR_UPDATE_INTERVAL,cbPos,CM_TIMER,mot);

  dout.data.lptr = &motors[mot].tgtMicroStepsRb;
  sprintf(dev,"%s/%s.TgtPos",gTarget,motors[mot].name);
  motors[mot].lnkIdTgt = AttachLinkEx(dev,"RECV",&dout,NULL,CA_READ,MTR_UPDATE_INTERVAL,cbTgt,CM_TIMER,mot);

}
LIB_EXPORT int getPos(int mot,double *val)
{
  int ival;
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = motors[mot].posMicroSteps * motors[mot].scalePos + motors[mot].offPos;
  *val = microsteps2mm(mot,ival);
  return 0;
}
LIB_EXPORT int getPosMin(int mot,double *val)
{
  int ival;
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = motors[mot].minPos * motors[mot].scalePos  + motors[mot].offPos;
  *val = microsteps2mm(mot,ival);
  return 0;
}
LIB_EXPORT int getPosMax(int mot,double *val)
{
  int ival;
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = motors[mot].maxPos * motors[mot].scalePos  + motors[mot].offPos;
  *val = microsteps2mm(mot,ival);
  return 0;
}
LIB_EXPORT int getPosOff(int mot,double *val)
{
  int ival;
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = motors[mot].offPos;
  *val = microsteps2mm(mot,ival);
  return 0;
}
LIB_EXPORT int setPosOff(int mot,double val)
{
  int cc, ival;
  char errstr[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = (int)mm2microsteps(mot,val);
  motors[mot].offPos = ival;
  // save and restore:
  motorInfo[mot].posOff = motors[mot].offPos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getPosScale(int mot,double *val)
{
  int ival;
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = motors[mot].scalePos;
  *val = (double)ival;
  return 0;
}
LIB_EXPORT int setPosScale(int mot,double val)
{
  int cc, ival;
  char errstr[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  ival = (int)val;
  motors[mot].scalePos = ival;
  // save and restore:
  motorInfo[mot].posScale = motors[mot].scalePos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getVel(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  //*val = cycles2mm(mot,motors[mot].velRps);
  *val = microsteps2mm(mot,motors[mot].velMean);
  return 0;
}
LIB_EXPORT int setVel(int mot,double val)
{
  int cc;
  float fval, tgtVals[4];

  if (mot < 0 || mot > gNumMotors) return illegal_device;
  fval = (float)(val * motors[mot].mstepsPerMM / motors[mot].mstepsPerCycle);
  tgtVals[0] = 1; // load target position
  tgtVals[1] = 1; // ??
  tgtVals[2] = (float)(fval * 360.0); // velocity in degrees/sec
  tgtVals[3] = 0;
  cc = sendFloatCmdToHardware(mot,tgtVals);
  if (cc != 0) return cc;
  motors[mot].velRms = (float)(val * motors[mot].mstepsPerMM);
  return 0;
}
LIB_EXPORT int getVelLwr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = cycles2mm(mot,motors[mot].velLwr);
  return 0;
}
LIB_EXPORT int getVelUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = cycles2mm(mot,motors[mot].velUpr);
  return 0;
}
LIB_EXPORT int getAccUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = cycles2mm(mot,motors[mot].accUpr);
  return 0;
}

// TODO:
LIB_EXPORT int getAcc(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val =  SimulateData(5,2);
  return 0;
}

LIB_EXPORT int mstepsPerMM(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].mstepsPerMM;
  return 0;
}

LIB_EXPORT int mstepsPerStep(int mot,int *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].microStepsPerStep;
  return 0;
}

LIB_EXPORT int stepsPerRadian(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].mstepsPerCycle/6.2832;
  return 0;
}

int sendMoveCmd(int mot,int useEndSwitch)
{
  int tgtVals[4];
  tgtVals[0] = 1; // load target position
  tgtVals[1] = 1; // ??
  tgtVals[2] = useEndSwitch ? 0 : 2;// with END Switch control
  tgtVals[3] = 0;
  return motors[mot].cmdSts = sendIntegerCmdToHardware(mot,tgtVals);
}
int stopMoveCmd(int mot)
{
  int tgtVals[4];
  tgtVals[0] = 2;
  tgtVals[1] = 1;
  tgtVals[2] = 0;
  tgtVals[3] = 0;
  return motors[mot].cmdSts = sendIntegerCmdToHardware(mot,tgtVals);
}
int sendGrpMoveCmd(int gid,int useEndSwitch)
{
  int tgtVals[4];
  tgtVals[0] = 1; // load target position
  tgtVals[1] = 1; // ??
  tgtVals[2] = useEndSwitch ? 0 : 1;// with END Switch control
  tgtVals[3] = 0;
  return sendGroupCmdToHardware(gid,tgtVals);
}
int stopGrpMoveCmd(int gid)
{
  int tgtVals[4];
  tgtVals[0] = 2;
  tgtVals[1] = 1;
  tgtVals[2] = 0;
  tgtVals[3] = 0;
  return sendGroupCmdToHardware(gid,tgtVals);
}
LIB_EXPORT int movests(int mot,int *sts)
{
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (sts == NULL) return invalid_parameter;
  *sts = motors[mot].movCmd;
  return 0;
}
LIB_EXPORT int moveprg(int mot,int *prg)
{
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (prg == NULL) return invalid_parameter;
  *prg = motors[mot].movEtr;
  return 0;
}
LIB_EXPORT int moveto(int mot,int cmd)
{
  double etr;
  int sts;
  if ((motors[mot].status & MTR_UNCLBR) != 0) return not_allowed;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
    case PROC_STS_START:
      if (sendMoveCmd(mot,TRUE) == 0)
      {
        etr = fabs(motors[mot].tgtMicroSteps - motors[mot].posMicroSteps);
        etr /= (double)(motors[mot].velMean);
        //etr /= (double)(motors[mot].velRps * motors[mot].mstepsPerCycle);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      break;
    case PROC_STS_RESUME:
      sendMoveCmd(mot,TRUE);
      motors[mot].movPaused = FALSE;
      motors[mot].status &= ~MTR_PAUSE;
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
  }
  return 0;
}
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_START:
      if (sendMoveCmd(mot,FALSE) == 0)
      {
        etr = fabs(motors[mot].tgtMicroSteps - motors[mot].posMicroSteps);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    default:
      return moveto(mot,cmd);
  }
  return 0;
}
typedef struct
{
  int id;
  MotorGroup grp;
} MotGrpTbl;
#define GRPTBL_SIZE 10
MotGrpTbl grps[GRPTBL_SIZE];
int ngrps = 0;

int getGrpTblIndex(MotorGroup *grp) 
{ 
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (memcmp(&grps[i].grp,grp,sizeof(MotorGroup))) continue;
    return i;
  }
  return -1; 
}
int addToGrpTbl(MotorGroup *grp) 
{ 
  int i = getGrpTblIndex(grp);
  if (i >= 0) return i;
  if (ngrps < GRPTBL_SIZE)
  {
    i = ngrps;
    memcpy(&grps[i].grp,grp,sizeof(MotorGroup));
    ngrps++;
    return i;
  }
  return -1; 
}
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp)
{
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    memcpy(grp,&grps[i].grp,sizeof(MotorGroup));
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int setGrpMembers(int grpId,MotorGroup *grp)
{
  int g;
  if ((g=addToGrpTbl(grp)) < 0) return invalid_index;
  grps[g].id = grpId;
  memcpy(&grps[g].grp,grp,sizeof(MotorGroup));
  return 0;
}
LIB_EXPORT int grpmoveto(int grpId,int cmd)
{
  int cc, i, g=-1;
  MotorGroup grp;
  double etr;
  char dev[128];
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors; i++)
  { // don't allow if anybody is being calibrated !
    if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
  }
  for (i=0; i<gNumMotors; i++)
  { // set or remove the sync bit
    sprintf(dev,"%s.SyncFlag",motors[i].name);
    if (IsMotorInGroup(i,&grp))
    {       
      if ((cc=sendIntegerToHardware(dev,1)) != 0)
      {
        motors[i].status |= MTR_ERR;
        continue;
      }
      if (g < 0) g = i;
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
        case PROC_STS_START:
          etr = fabs(motors[i].tgtMicroSteps - motors[i].posMicroSteps);
          etr /= (double)(motors[i].velMean);
          //etr /= (double)(motors[i].velRps * motors[i].mstepsPerCycle);
          motors[i].movEtr = (int)etr;
          motors[i].movPaused = FALSE;
          motors[i].status |= MTR_MOVE;
          break;
        case PROC_STS_PAUSE:
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_RESUME:
          motors[i].movPaused = FALSE;
          motors[i].status &= ~MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
      }
    }
  }
  if (g < 0) return non_existent;
  switch (cmd)
  {
    case PROC_STS_RESUME:
    case PROC_STS_START:
      cc = sendGrpMoveCmd(g,TRUE);
      break;
    case PROC_STS_STOP:
    case PROC_STS_PAUSE:
      cc = stopGrpMoveCmd(g);
      break;
  }
  return cc;
}

LIB_EXPORT int setTgtPosMs(int mot,int val)
{
  int cc, tgtVals[4];  
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (val < motors[mot].minPos || val > motors[mot].maxPos) 
  {
    if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
  }
  tgtVals[0] = 30; // load target position
  tgtVals[1] = 1;  // ??
  tgtVals[2] = val;
  tgtVals[3] = 0;
  cc = motors[mot].cmdSts = sendIntegerCmdToHardware(mot,tgtVals);
  if (cc == 0) motors[mot].tgtMicroSteps = val;  
  return cc;
}
LIB_EXPORT int setTgtPos(int mot,double val)
{
  int ival;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  ival = (int)mm2microsteps(mot,val);
  return setTgtPosMs(mot,ival);
}
LIB_EXPORT int getTgtPosMs(int mot,int *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;

  *val = motors[mot].tgtMicroSteps * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getTgtPos(int mot,double *val)
{
  int ival;
  int cc = getTgtPosMs(mot,&ival);
  if (val == NULL) return invalid_parameter;
  if (cc == 0)
  {
    *val = microsteps2mm(mot,ival);
  }
  return cc;
}
LIB_EXPORT int calibr(int mot,int *sts)
{
  int cc = 0, ival;
  int vals[4];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if ((motors[mot].status & 0xff) != MTR_IDLE) return not_allowed;
  if (motors[mot].type == MTR_TYPE_LINEAR)
  { // motor has end switch (n.b. setTgtPosMs() will reverse the scale and offset)
    motors[mot].status |= MTR_ISCLBR;
    millisleep(200);
    vals[0] = 41; vals[1] = 1; vals[2] = 31; vals[3] = 4661;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 39; vals[3] = motors[mot].velRef1; // ref speed
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    ival = MTR_MIN_POSITION * motors[mot].scalePos + motors[mot].offPos;
    if ((cc=setTgtPosMs(mot,ival)) != 0) return cc;
    if ((cc=sendMoveCmd(mot,TRUE)) != 0) return cc;
    motors[mot].status |= MTR_REF1; // will trigger chkClbrStatus() call
  }
  else
  { // motor has no end switch
    motors[mot].cmdSts = setTgtPosMs(mot,motors[mot].offPos); // tgt position = 0
    vals[0] = 31; vals[1] = 1; vals[2] = 0; vals[3] = 0;  // rb position = 0
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 15; vals[3] = CLBR_PATTERN;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    motors[mot].status = 0; // clear everything 
  }
  if (sts != NULL) *sts = 0;
  return 0;
}
LIB_EXPORT int reset(int mot,int *sts)
{
  int vals[4];

  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (motors[mot].status & MTR_CLBR) return operation_busy;

  vals[0] = 41; vals[1] = 1; vals[2] = 31; vals[3] = 4661;
  motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);

  vals[0] = 41; vals[1] = 1; vals[2] = 7; vals[3] = 0x8000;
  motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);

  motors[mot].status = MTR_UNCLBR;
  if (sts != NULL) *sts = 0;
  calibr(mot,sts);

  return 0;
}
int chkClbrFlag(int mot)
{
  int vals[4], rbvals[6];

  vals[0] = 40; vals[1] = 1; vals[2] = 15; vals[3] = 0;
  motors[mot].cmdSts = sendRecvIntegerCmdToHardware(mot,vals,rbvals);

  if (rbvals[3] == CLBR_PATTERN) 
  {
    if (!(motors[mot].status & MTR_CLBR)) motors[mot].status &= ~MTR_UNCLBR;
    return TRUE;
  }
  motors[mot].status |= MTR_UNCLBR;
  return FALSE;
}
int chkEndStation(int mot,int mask)
{
  char dev[128];
  int reg;
  sprintf(dev,"%s.regExSta",motors[mot].reg);
  getIntegerFromHardware(dev,&reg);
  return (reg & mask) ? TRUE : FALSE;
}
void chkClbrStatus(int mot)
{
  char dev[128];
  int cp, vals[4];
  if (motors[mot].status & MTR_REF1)
  {
    if ((motors[mot].status & MTR_REF2) == 0)
    {
      if (!chkEndStation(mot,0x01)) return;
      motors[mot].status |=  MTR_REF2;
    }  
    sprintf(dev,"%s.CurPos",motors[mot].name);
//    if (getIntegerFromHardware(dev,&cp) != 0) return;
//    cp = cp * motors[mot].scalePos + motors[mot].offPos;
//    if (setTgtPosMs(mot,cp) != 0) return;
//    if (sendMoveCmd(mot,FALSE) != 0) return;
    vals[0] = 30; vals[1] = 1; vals[2] = 0; vals[3] = 0;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 31; vals[1] = 1; vals[2] = 0; vals[3] = 0;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    if (sendMoveCmd(mot,TRUE) != 0) return;
    millisleep(200);
    vals[0] = 41; vals[1] = 1; vals[2] = 31; vals[3] = 4661;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 39; vals[3] = motors[mot].velRef2; // slow speed
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 52; vals[3] = 32768;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 31; vals[3] = 0;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    if (setTgtPosMs(mot,999999*motors[mot].scalePos+motors[mot].offPos) != 0) return;
    if (sendMoveCmd(mot,TRUE) != 0) return;
    motors[mot].status &= ~MTR_REF1; // clear this so we go on to the next check 
  }
  if (motors[mot].status & MTR_REF2)
  {
    if ((motors[mot].status & MTR_REF3) == 0)
    {
      if (!chkEndStation(mot,0x01)) return;
      motors[mot].status |=  MTR_REF3;
    }
    vals[0] = 41; vals[1] = 1; vals[2] = 31; vals[3] = 4661;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    if (setTgtPosMs(mot,motors[mot].offPos) != 0) return;
    vals[0] = 31; vals[1] = 1; vals[2] = 0; vals[3] = 0;   // ist position
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 39; vals[3] = motors[mot].velMean; // normal speed (1000)
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 52; vals[3] = 49152;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    vals[0] = 41; vals[1] = 1; vals[2] = 31; vals[3] = 0;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
    motors[mot].status &= ~MTR_REF2;
    return;
  }
  if (motors[mot].status & MTR_REF3)
  {
    motors[mot].status &= ~MTR_ISCLBR;
    motors[mot].status &= ~MTR_CLBR;
    motors[mot].status &= ~MTR_UNCLBR;
    vals[0] = 41; vals[1] = 1; vals[2] = 15; vals[3] = CLBR_PATTERN;
    motors[mot].cmdSts = sendIntegerCmdToHardware(mot,vals);
  }
  return;
}
LIB_EXPORT int getSts(int mot,int *sts)
{ // is called from the background task @ 10 Hz
  int cc, reg = 0, msts = 0;
  char dev[128];
  time_t t = time(NULL);

  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;

  sprintf(dev,"%s.Status",motors[mot].name);
  if ((cc=getIntegerFromHardware(dev,&msts)) != 0)
  {
    motors[mot].status = MTR_ERR;
  }
  else
  {
    motors[mot].status &= ~MTR_ERR;
    if (msts == 1) motors[mot].status &= ~MTR_MOVE;
    if (msts == 2) motors[mot].status |= MTR_MOVE;
    if (t > motors[mot].lastCheckedEndsw)
    {
      sprintf(dev,"%s.regExSta",motors[mot].reg);
      if ((cc=getIntegerFromHardware(dev,&reg)) == 0)
      {
        if (reg & 1) motors[mot].status |= MTR_END_O; else motors[mot].status &= ~MTR_END_O;
        if (reg & 2) motors[mot].status |= MTR_END_I; else motors[mot].status &= ~MTR_END_I;
      }
      motors[mot].lastCheckedEndsw = t;
    }
  }
  if (!(motors[mot].status & MTR_MOVE) && t > motors[mot].lastCheckedClbr + 5)
  {
    chkClbrFlag(mot);
    motors[mot].lastCheckedClbr = t;
  }
  if (motors[mot].status & MTR_CLBR) chkClbrStatus(mot);
  *sts = motors[mot].status;
  return 0;
}

int startup(void)
{
  initialize();
  SetMotorList(mstNames,gNumMotors); // done
  SetStatusList(stsLst,sizStsLst);   // done 
  SetGetMotorStatusDispatch(getSts); // done
  SetGetMotorVelocityDispatch(getVel); // done
  SetSetMotorVelocityDispatch(setVel); // done
  SetGetMotorAccelerationDispatch(getAcc);
  SetGetMotorPositionDispatch(getPos); // done
  SetGetMotorStepsPerMUDispatch(mstepsPerMM); // done
  SetGetMotorMicroStepsPerStepDispatch(mstepsPerStep); // done
  SetGetMotorStepsPerRadianDispatch(stepsPerRadian); // done
  SetGetMoveToStatusDispatch(movests); // done
  SetGetMoveToProgressDispatch(moveprg); // done
  SetGetGroupMoveStatusDispatch(movests);
  SetGetIncrementMoveStatusDispatch(movests);
  SetGetIncrementGroupMoveStatusDispatch(movests);
  SetSetMoveToStatusDispatch(moveto); // done
  SetSetFreeMoveToStatusDispatch(movetoNoEndSwitch); // done
  SetSetGroupMoveStatusDispatch(grpmoveto); // done ?
  SetSetIncrementGroupMoveStatusDispatch(grpmoveto);
  SetSetIncrementMoveStatusDispatch(moveto); // done (same as MoveTo)
  SetSetMotorTargetPositionDispatch(setTgtPos); // done
  SetGetMotorTargetPositionDispatch(getTgtPos); // done
  SetSetMotorTargetMicroPositionDispatch(setTgtPosMs); // done
  SetGetMotorTargetMicroPositionDispatch(getTgtPosMs); // done
  SetGetGroupMembersDispatch(getGrpMembers); // done ?
  SetSetGroupMembersDispatch(setGrpMembers); // done ?
  SetCalibrateMotorDispatch(calibr);  // done
  SetResetMotorDispatch(reset); // done
  SetGetMotorVelocityLowerLimitDispatch(getVelLwr);
  SetGetMotorVelocityUpperLimitDispatch(getVelUpr);
  SetGetMotorAccelerationUpperLimitDispatch(getAccUpr);
  SetGetMotorPositionMinimumDispatch(getPosMin); // done
  SetGetMotorPositionMaximumDispatch(getPosMax); // done
  SetGetMotorPositionOffsetDispatch(getPosOff); // done
  SetSetMotorPositionOffsetDispatch(setPosOff); // done
  SetGetMotorPositionScaleDispatch(getPosScale); // done
  SetSetMotorPositionScaleDispatch(setPosScale); // done
  return 0;
}

BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      LoadLibrary("cdi32.dll");
      SetInitMotorHardwareDispatch(startup);
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      break;
  }
  return (TRUE);
}
