#ifndef MSTLIB_H_
#define MSTLIB_H_

#include "smpeqm.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#if defined(WIN32) 
# define MST_EXPORT __declspec(dllexport)
#else
# define MST_EXPORT
#endif

/* common motor status bit definitions (property "Status") : */
#define MTR_IDLE   0x0000
#define MTR_MOVE   0x0001
#define MTR_PAUSE  0x0002
#define MTR_STEP   0x0004
#define MTR_REF1   0x0010
#define MTR_REF2   0x0020
#define MTR_REF3   0x0040
#define MTR_ISCLBR 0x0800
#define MTR_UNCLBR 0x1000
#define MTR_ERR    0x8000
 
#define MTR_CLBR (MTR_REF1|MTR_REF2|MTR_REF3)
 
#define MST_GROUP_FILE "motorgroups.csv"
#define MST_MONITOR_FILE "motorscans.csv"

#define MST_DYNAGRP_OFFST 512

typedef struct MonitorInfoStruct
{
  char nam[64]; /**< motor or group device name */
  char dev[132]; /**< full device name of target */
  char prp[64]; /**< target property */
  char fmt[16]; /**< monitor format */
  int len; /**< monitor array length */
  int tmr; /**< monitor timer (msec) */
  int dly; /**< delay (msec) between scan steps */
  int inc; /**< step increment in motor units */
} MonitorInfo;

typedef struct MonitorParamsStruct
{
  char dev[132]; /**< full device name of target */
  char prp[64]; /**< target property */
  int fmt; /**< monitor format */
  int len; /**< monitor array length */
  int tmr; /**< monitor timer (msec) */
  int idx; /**< running index in scan buffer */
  int err; /**< error counts for updates throughout the scan */
  int cle;/**< consecutive link errors for abort scan */
  int dly; /**< delay (msec) between scan steps */
  int inc; /**< step increment in motor units */
  double ts; /**< timestamp of last monitor update */
  BYTE *dat; /**< reference to monitor data */
  struct MonitorParamsStruct *nxt; /**< next pointer */
} MonitorParams;

extern MonitorParams *g_monitorParamsList;

MonitorParams *getMonitorParams(const char *dev,const char *prp,int fmt,int len,int tmr,int dly,int inc);

typedef struct
{
  MonitorParams *params; /**< refers to a table entry */
  double *xAxis; /**< x-axis array for last monitor (from position readback) */
  double *yAxis; /**< y-axis array for last monitor (using 'row length' from monParams) */
  double start; /**< last monitor start time */
  double stop; /**< last monitor stop time */
  double ts; /**< last motor position readback time */
  double tsStep; /**< time last step completed */
  double tgtPos; /**< motor target position */
  double lastPos; /**< last scan position */
  int isStepping; /**< TRUE: step and readback; FALSE: move and readback */
  int capacity; /**< buffer capacity */
  int length; /**< buffer scan length */
  int sts; /**< monitor status (0 if > 50 % readback data OK) */
  int id;  /**< link id */
  int mot; /**< motor table index */
  int grp; /**< group table index */
} MonitorScan;

typedef struct
{
  NAME16 units; /**< natural position units (from motor library) */
  double ztol; /**< zero tolerance according to motor units */
  double spmu; /**< steps per motor unit */
  double spr; /**< steps per revolution */
  double tgtPosition; /**< target position */
  double position; /**< readback position */
  double positionMaximum; /**< maximum position */
  double positionMinimum; /**< minimum position */
  double positionOffset; /**< position offset (software offset) */
  double positionScale; /**< position scale (software scale) */
  double limits[3]; /**< vel lower limit, vel upper limit, acc upper limit */
  double acceleration; /**< current acceleration */
  double velocity; /**< current velocity */
  double current; /**< motor current */
  UINT32 msps; /**< micro-steps per step */
  UINT32 tgtMicroPosition; /**< target position (micro steps) */
  UINT32 rotationmoveallowed; /**< is rotation move allowed (boolean) */
  UINT32 stepcounter; /**< step counter */
  UINT32 microstepcounter; /**< micro-step counter */
  short status; /**< current status */
  short online; /**< online status */
  MonitorScan monScan; /**< monitor information for ScanMonitor */
} Motor;

typedef struct
{
  NAME16 units; /**< natural position units (from motor library) */
  double tgtPosition[6]; /**< target position */
  double position[6]; /**< readback position */
  double positionMaximum[6]; /**< maximum position */
  double positionMinimum[6]; /**< minimum position */
  double positionOffset[6]; /**< position offset (software offset) */
  double positionScale[6]; /**< position scale (software scale) */
  double limits[12]; /**< vel lower limit, vel upper limit, acc upper limit */
  double acceleration; /**< current acceleration */
  double velocity; /**< current velocity */
  double current; /**< motor current */
  short status; /**< current status */
  short online; /**< online status */
} SmarPodMotor;

/**
 * \internal
 * \brief structure for reading in the .csv database info
 */
typedef struct
{
  char name[DEVICE_NAME_SIZE];
  int groupId;
  int motorNr;
  float gearRatio;
  float scale;
  float offset;
  float weight;
} MstGroupInfoStruct;

/**
 * \internal
 * \brief a registered group motor (determined via the database)
 */
typedef struct 
{
  char name[DEVICE_NAME_SIZE]; /**< name of motor group */
  int numberInGroup;   /**< number of motors in group */
  int groupId;         /**< assigned id */
  float gearRatio;     /**< default gear ratio */
  MstGroupInfoStruct *members[NUM_MSTEQM_DEVICES]; /**< group members */
  MonitorScan monScan; /**< monitor information for ScanMonitor */
} MstGroupDevice;

extern MstGroupDevice *mstGroupTbl;
extern int mstGroupTblSize;

#define NUM_MOTORS_MAX NUM_MSTEQM_DEVICES
#define NUM_MOTORGROUPS_MAX 100
#define NUM_MOTOR_FDSETS (NUM_MOTORS_MAX+7)/8

/**
 * \internal
 * \brief cross reference from server to motor plug
 */
typedef struct 
{
  int grpid; /**< group id */
  int grphead; /**< motor id of the group master */
  int nmotors; /**< number motors in group */
  BYTE fd_motors[NUM_MOTOR_FDSETS]; /**< motor descriptor set */
  char grpname[128]; /**< group name */
} MotorGroup;

#define IsMotorInGroup(m,g) ((g)->fd_motors[m/8] & ((0x01)<<(m%8)))

extern SmarPodMotor mstTbl[NUM_MSTEQM_DEVICES];

typedef struct
{
  char motName[64];
  int stsBits;
  int stsAxis;
  int stsCoupled;
  int stsClbr;
  int idxCntlLoop;
  int stateClbr;
  int iSpare7;
  int iSpare8;
  float pos;
  float tgt;
  float lag;
  float vel;
  float load;
  float ovrRide;
  float fSpare7;
  float fSpare8;
} MotorStatus;

typedef struct
{
  char name[64];
  char desc[64];
  char units[16];
  float max;
  float min;
  int access;
} XMA; // Extended Motor Attribute

extern XMA *xmaLst;
extern int xmaLstSize;
extern EQMFCNP xmaDispatch;

#define MST_SEPERATOR_TOKEN ","

#define PROC_STS_IDLE   0x00
#define PROC_STS_START  0x01
#define PROC_STS_PAUSE  0x02
#define PROC_STS_RESUME 0x03
#define PROC_STS_STOP   0x04
#define PROC_STS_ABORT  0x04


typedef float (*GETGMTPFCNP)(MotorGroup *,int,float);
typedef float (*GETGPFCNP)(MotorGroup *);

extern GETGMTPFCNP getGroupMemberTargetPositionFcn;
extern GETGPFCNP getGroupPositionFcn;

typedef int (*CMDFCNP)(void);
typedef int (*GETDFCNP)(int,double *);
typedef int (*GETIFCNP)(int,int *);
typedef int (*SETDFCNP)(int,double);
typedef int (*SETIFCNP)(int,int);
typedef int (*GETPFCNP)(int,int *);
typedef int (*SETPFCNP)(int,int);
typedef int (*GETSFCNP)(int,char *,int);

typedef int (*GETGRPFCNP)(int,MotorGroup *);
typedef int (*SETGRPFCNP)(int,MotorGroup *);

/* scan functions */
typedef int (*SETSPFCNP)(int,int,DTYPE *);
typedef int (*GETSPFCNP)(int,DTYPE *);

typedef int (*GETSTSFCNP)(int,MotorStatus *);
typedef int (*GETERRFCNP)(int,NAME64 *,int *);

typedef int (*GETDSMPFCNP)(int,double *,int);
typedef int (*SETDSMPFCNP)(int,double *,int);


extern CMDFCNP initMotorHdwFcn;
extern GETIFCNP resetMotorFcn;
extern GETIFCNP calibrateMotorFcn;
extern SETDFCNP calibrateToOffsetFcn;
extern GETIFCNP getMotorStatusFcn;
extern GETIFCNP getGroupMotorStatusFcn;
extern GETDFCNP getMotorStepsPerMUFcn;
extern GETDFCNP getMotorStepsPerRadianFcn;
extern GETIFCNP getMotorMicroStepsPerStepFcn;
extern GETDFCNP getMotorVelocityLowerLimitFcn;
extern GETDFCNP getMotorVelocityUpperLimitFcn;
extern GETDFCNP getMotorVelocityFcn;
extern GETDFCNP getMotorAccelerationFcn;
extern GETDFCNP getMotorPositionFcn;
extern GETDFCNP getGroupMotorPositionFcn;
extern GETDFCNP getMotorTargetPositionFcn;
extern SETDFCNP setMotorTargetStepsFcn;
extern GETIFCNP getMotorTargetMicroPositionFcn;
extern GETDFCNP getMotorAccelerationUpperLimitFcn;
extern GETDFCNP getGroupMotorTargetPositionFcn;
extern GETDFCNP getMotorCurrentFcn;
extern SETDFCNP setMotorPositionFcn;
extern SETDFCNP setMotorVelocityFcn;
extern SETDFCNP setMotorAccelerationFcn;
extern SETDFCNP setMotorTargetPositionFcn;
extern SETDFCNP setGroupMotorTargetPositionFcn;
extern SETIFCNP setMotorTargetMicroPositionFcn;
extern SETDFCNP setMotorCurrentFcn;

extern SETPFCNP setIncrementGroupMoveStatusFcn;
extern GETPFCNP getIncrementGroupMoveStatusFcn;
extern SETPFCNP setControlGroupMoveStatusFcn;
extern GETPFCNP getControlGroupMoveStatusFcn;
extern GETPFCNP getControlGroupMoveProgressFcn;
extern GETPFCNP getIncrementGroupMoveProgressFcn;
extern GETPFCNP getStepToProgressFcn;

extern GETPFCNP getIncrementMoveStatusFcn;
extern GETPFCNP getGroupMoveStatusFcn;
extern GETPFCNP getGroupMoveProgressFcn;
extern GETPFCNP getMoveToStatusFcn;
extern GETPFCNP getMoveToProgressFcn;
extern GETPFCNP getStepToStatusFcn;
extern SETPFCNP setStepToStatusFcn;
extern SETPFCNP setIncrementMoveStatusFcn;
extern SETPFCNP setGroupMoveStatusFcn;
extern SETPFCNP setMoveToStatusFcn;
extern SETPFCNP setMicroMoveStatusFcn;
extern SETDFCNP setMotorPositionOffsetFcn;
extern SETDFCNP setMotorPositionScaleFcn;
extern SETPFCNP setFreeMoveToStatusFcn;

extern GETDFCNP getMotorPositionMinimumFcn;
extern GETDFCNP getMotorPositionMaximumFcn;
extern GETDFCNP getMotorPositionOffsetFcn;
extern GETDFCNP getMotorPositionScaleFcn;

extern GETSFCNP getMotorUnitsFcn;

extern SETSPFCNP setScanMoveToStatusFcn;
extern GETSPFCNP getScanMoveToStatusFcn;
extern SETSPFCNP setScanGroupMoveToStatusFcn;
extern GETSPFCNP getScanGroupMoveToStatusFcn;
extern GETPFCNP getScanMoveToProgressFcn;

extern SETIFCNP setMotorOnlineFcn;
extern GETIFCNP getMotorOnlineFcn;

extern GETSTSFCNP getMotorStatusBlockFcn;

extern GETDFCNP getGroupGearRatioFcn;
extern SETDFCNP setGroupGearRatioFcn;

extern GETERRFCNP getLastHdwErrFcn;

extern int gMstSetAlarms;
extern int sizMstStsList;
extern NAME32 *mstStsList;

int initializeMotorHardware(void);
int resetMotor(int motorNr);
int getMotorStatus(int motorNr);
int getGroupMotorStatus(int grpId);
int getMotorUnits(int motorNr,NAME16 *units);
double getMotorVelocityLowerLimit(int motorNr);
double getMotorVelocityUpperLimit(int motorNr);
double getMotorVelocity(int motorNr);
int setMotorVelocity(int motorNr,double newVelocity);
double getMotorAcceleration(int motorNr);
int setMotorAcceleration(int motorNr,double newAcceleration);
double getMotorPosition(int motorNr,double *pos,int len);
double getGroupMotorPosition(int grpId);
int getMotorPositionMinimum(int motorNr,double *pos,int len);
int getMotorPositionMaximum(int motorNr,double *pos,int len);
int getMotorPositionOffset(int motorNr,double *pos,int len);
int getMotorPositionScale(int motorNr,double *pos,int len);
int getMotorTargetPosition(int motorNr,double *pos,int len);
double getGroupGearRatio(int grpId);
double getGroupMotorTargetPosition(int motorNr);
int mstSetGroupHead(int grpId,int motorNr);
int setGroupGearRatio(int grpId,double newRatio);
int setMotorPosition(int motorNr,double *newPosition,int len);
int setMotorTargetPosition(int motorNr,double *newPosition,int len);
int setMotorTargetSteps(int mot,int numSteps);
int setGroupMotorTargetPosition(int motorNr,double newPosition);
int setMotorTargetMicroPosition(int motorNr,int newPosition);
double getMotorAccelerationUpperLimit(int motorNr);
double getMotorCurrent(int motorNr);
int setMotorCurrent(int motorNr,double newCurrent);
void getMotorList(NAME64 **motorNames,int *numMotors);
double getStepsPerMU(int motorNr);
int calibrateMotor(int motorNr);
int calibrateToOffset(int motorNr,double val);
double getStepsPerRadian(int motorNr);
int getMicroStepsPerStep(int motorNr);
int mstGetMotorGroup(const char *tag,int nmotors,int *motors);
int mstGetMotorGroupFromList(const char *motorList,const char *groupName);
int mstGetMotorGroupFromTag(const char *tag);
int setMotorPositionOffset(int motorNr,double *newOffset,int length);
int setMotorPositionScale(int motorNr,double *newScale, int length);
int mstSplitGroupFromList(const char *motorList,int *num,int *motors,char *tag);
void registerStructs(void);
int getLastHardwareError(int motorNr,NAME64 *errstr,int *errcode);
int mstAddPseudoAxisMotorGroup(const char *tag,int grpid);
int mstGetMotorGroupTableIndex(int grpId);

MST_EXPORT int SetXMAList(XMA *list,int listSize,EQMFCNP fcn);
MST_EXPORT int SetMotorList(NAME64 *names,int num);
MST_EXPORT int SetStatusList(NAME32 *lst,int siz);
MST_EXPORT void SetInitMotorHardwareDispatch(CMDFCNP fcn);
MST_EXPORT void SetResetMotorDispatch(GETIFCNP fcn);
MST_EXPORT void SetGetMotorStatusDispatch(GETIFCNP fcn);
MST_EXPORT void SetGetGroupMotorStatusDispatch(GETIFCNP fcn);
MST_EXPORT void SetGetMotorVelocityLowerLimitDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorVelocityUpperLimitDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorVelocityDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorAccelerationDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorPositionDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetGroupMotorPositionDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorTargetPositionDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorTargetStepsDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorTargetMicroPositionDispatch(GETIFCNP fcn);
MST_EXPORT void SetGetMotorAccelerationUpperLimitDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorCurrentDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetGroupMotorTargetPositionDispatch(GETDFCNP fcn);
MST_EXPORT void SetSetMotorPositionDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetMotorVelocityDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetMotorAccelerationDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetMotorTargetPositionDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetMotorTargetStepsDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetGroupMotorTargetPositionDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetMotorTargetMicroPositionDispatch(SETIFCNP fcn);
MST_EXPORT void SetSetMotorCurrentDispatch(SETDFCNP fcn);
MST_EXPORT void SetGetStepToProgressDispatch(GETPFCNP fcn);

MST_EXPORT void SetGetIncrementMoveStatusDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetGroupMoveStatusDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetMoveToStatusDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetStepToStatusDispatch(GETPFCNP fcn);
MST_EXPORT void SetSetStepToStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetSetIncrementMoveStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetSetGroupMoveStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetSetMoveToStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetSetFreeMoveToStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetSetMicroMoveStatusDispatch(SETPFCNP fcn);

MST_EXPORT void SetGetMoveToProgressDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetGroupMoveProgressDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetControlGroupMoveProgressDispatch(GETPFCNP fcn);

MST_EXPORT void SetAllowMstAlarms(int value);
MST_EXPORT void SetGetMotorStepsPerMUDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorMicroStepsPerStepDispatch(GETIFCNP fcn);
MST_EXPORT void SetGetMotorStepsPerRadianDispatch(GETDFCNP fcn);

MST_EXPORT void SetCalibrateMotorDispatch(GETIFCNP fcn);
MST_EXPORT void SetGetGroupMembersDispatch(GETGRPFCNP fcn);
MST_EXPORT void SetSetGroupMembersDispatch(SETGRPFCNP fcn);

MST_EXPORT void SetGetMotorPositionMinimumDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorPositionMaximumDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorPositionOffsetDispatch(GETDFCNP fcn);
MST_EXPORT void SetGetMotorPositionScaleDispatch(GETDFCNP fcn);
MST_EXPORT void SetSetMotorPositionOffsetDispatch(SETDFCNP fcn);
MST_EXPORT void SetSetMotorPositionScaleDispatch(SETDFCNP fcn);

MST_EXPORT void SetSetIncrementGroupMoveStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetSetControlGroupMoveStatusDispatch(SETPFCNP fcn);
MST_EXPORT void SetGetIncrementGroupMoveStatusDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetControlGroupMoveStatusDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetIncrementGroupMoveProgressDispatch(GETPFCNP fcn);

MST_EXPORT void SetCalibrateToOffsetDispatch(SETDFCNP fcn);
MST_EXPORT void SetGetMotorUnitsDispatch(GETSFCNP fcn);

MST_EXPORT void SetGetScanMoveToStatusDispatch(GETSPFCNP fcn);
MST_EXPORT void SetSetScanMoveToStatusDispatch(SETSPFCNP fcn);
MST_EXPORT void SetGetScanGroupMoveToStatusDispatch(GETSPFCNP fcn);
MST_EXPORT void SetSetScanGroupMoveToStatusDispatch(SETSPFCNP fcn);
MST_EXPORT void SetGetScanMoveToProgressDispatch(GETPFCNP fcn);
MST_EXPORT void SetGetMotorOnlineDispatch(GETIFCNP fcn);
MST_EXPORT void SetSetMotorOnlineDispatch(SETIFCNP fcn);

MST_EXPORT void SetGetMotorStatusBlockDispatch(GETSTSFCNP fcn);

MST_EXPORT void SetGetGroupGearRatioDispatch(GETDFCNP fcn);
MST_EXPORT void SetSetGroupGearRatioDispatch(SETDFCNP fcn);

MST_EXPORT void SetGetLastHardwareErrorDispatch(GETERRFCNP fcn);
MST_EXPORT float GetGroupMotorGearRatio(int grpId,int mot);
MST_EXPORT float GetGroupMotorScale(int grpId,int mot);
MST_EXPORT float GetGroupMotorWeight(int grpId,int mot);
MST_EXPORT float GetGroupMotorOffset(int grpId,int mot);

extern GETDSMPFCNP getSmarPodPositionFcn;
extern GETDSMPFCNP getSmarPodTargetPositionFcn;
extern GETDSMPFCNP getSmarPodPositionMinimumFcn;
extern GETDSMPFCNP getSmarPodPositionMaximumFcn;
extern GETDSMPFCNP getSmarPodPositionOffsetFcn;
extern GETDSMPFCNP getSmarPodPositionScaleFcn;
extern GETDSMPFCNP getSmarPodPositionFcn;
extern GETDSMPFCNP getSmarPodPivotFcn;
extern GETIFCNP getSmarPodPivotModeFcn;
extern GETIFCNP getSmarPodSensorModeFcn;

extern SETDSMPFCNP setSmarPodMoveFcn;
extern SETDSMPFCNP setSmarPodPositionFcn;
extern SETDSMPFCNP setSmarPodTargetPositionFcn;
extern SETDSMPFCNP setSmarPodPositionOffsetFcn;
extern SETDSMPFCNP setSmarPodPositionScaleFcn;
extern SETDSMPFCNP setSmarPodPivotFcn;
extern SETIFCNP setSmarPodPivotModeFcn;
extern SETIFCNP setSmarPodSensorModeFcn;

MST_EXPORT void SetSetSmarPodMoveDispatch(SETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodPositionDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetSetSmarPodTargetPositionDispatch(SETDSMPFCNP fcn);
MST_EXPORT void SetSetSmarPodPositionDispatch(SETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodTargetPositionDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodPositionMinimumDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodPositionMaximumDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodPositionOffsetDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodPositionScaleDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetSetSmarPodPositionOffsetDispatch(SETDSMPFCNP fcn);
MST_EXPORT void SetSetSmarPodPositionScaleDispatch(SETDSMPFCNP fcn);

MST_EXPORT void SetSetSmarPodPivotDispatch(GETDSMPFCNP fcn);
MST_EXPORT void SetGetSmarPodPivotDispatch(GETDSMPFCNP fcn);

MST_EXPORT void SetGetSmarPodPivotModeDispatch(GETIFCNP fcn);
MST_EXPORT void SetSetSmarPodPivotModeDispatch(SETIFCNP fcn);
MST_EXPORT void SetGetSmarPodSensorModeDispatch(GETIFCNP fcn);
MST_EXPORT void SetSetSmarPodSensorModeDispatch(SETIFCNP fcn);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* MSTLIB_H_ */
