/*
   Module: plseqm.c  - example TINE pulse generator server, which illustrates
           the basic server API
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "plseqm.h"

#ifndef DOS
# if defined(UNIX) || defined(WIN32)
# define random(r) (r != 0 ? rand()%(r) : 0)
# else
# define random(r) (r != 0 ? (i*(69069))%(r) : 0)
# endif
#endif

#define PRP_PULSE     1
#define PRP_AMPLITUDE 2
#define PRP_FREQUENCY 3
#define PRP_SIGMA     4
#define PRP_NOISE     5
#define PRP_INFO      6
#define PRP_SHIFT     7
#define PRP_VALUE     8
#define PRP_TABLE     9
#define PRP_INTERVAL 10

#define NUM_DEVICES 10
#define NUM_VALUES (1024)

/*
   The code section below illustrates how one registers a structure
   to be exported.  
*/

typedef struct
{
  float amplitude;
  float frequency;
  float noise;
  float sigma;
  float peak;
  float value;
  int interval; // scan interval seconds/advance
} PulseInfo;

typedef struct
{
  int pointer;
  time_t lasttime;
} PulseInfoEx;

#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,amplitude),1,CF_FLOAT,"amplitude")) quit(1);
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,frequency),1,CF_FLOAT,"frequency")) quit(1);
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,noise),1,CF_FLOAT,"noise")) quit(1);
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,sigma),1,CF_FLOAT,"sigma")) quit(1);
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,peak),1,CF_FLOAT,"shift")) quit(1);
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,value),1,CF_FLOAT,"value")) quit(1);
  if (addFieldToStruct("PulseInfo",OFFSETIN(PulseInfo,interval),1,CF_INT32,"interval")) quit(1);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("PulseInfo",sizeof(PulseInfo),NUM_DEVICES)) quit(1);
}

PulseInfo pulseInfoTbl[NUM_DEVICES];
PulseInfoEx pulseInfoTblEx[NUM_DEVICES];
float plsbuf[NUM_DEVICES][NUM_VALUES];

void updatePulse(int devnr)
{ /* used internally -> don't bother with bounds checking */
  int i, k, np, idx;
  float *target = plsbuf[devnr];
  int n = (int)pulseInfoTbl[devnr].noise;
  float p = pulseInfoTbl[devnr].peak;
  float a = pulseInfoTbl[devnr].amplitude;
  float sig = pulseInfoTbl[devnr].sigma;
  time_t t = time(NULL);
  static time_t interim = 0;

  np = pulseInfoTbl[devnr].frequency > 0 ? (int)(NUM_VALUES/pulseInfoTbl[devnr].frequency) : NUM_VALUES;
  for (i=0; i<NUM_VALUES; i++)
  {
    k = i  % np;
    target[i] = random(n) + (float)(exp(-(k-p)*(k-p)/(2*sig*sig)))*a;
  }
  idx = pulseInfoTblEx[devnr].pointer;
  pulseInfoTbl[devnr].value = target[idx];
  if (t < pulseInfoTblEx[devnr].lasttime + pulseInfoTbl[devnr].interval) return;
  pulseInfoTblEx[devnr].lasttime = t;
  pulseInfoTblEx[devnr].pointer = (pulseInfoTblEx[devnr].pointer + 1) % np;
  return;
}
char *plsname[NUM_DEVICES] = {
  "Pulse0","Pulse1","Pulse2","Pulse3","Pulse4",
  "Pulse5","Pulse6","Pulse7","Pulse8","Pulse9" 
};

void plsinit(void)
{
  int i;
  DTYPE dout, din;

  /* some initial values for all devices */
  for (i=0; i<NUM_DEVICES; i++)
  {
    pulseInfoTbl[i].amplitude = 256.0;
    pulseInfoTbl[i].frequency = 1.0;
    pulseInfoTbl[i].sigma = 2.0;
    pulseInfoTbl[i].noise = 30.0;
    pulseInfoTbl[i].peak = 10.0;
    pulseInfoTbl[i].interval = 5;
    pulseInfoTblEx[i].pointer = 0;
    pulseInfoTblEx[i].lasttime = 0;
  }
  registerStructs();

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1024;
  dout.dTag[0] = 0;
  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_NULL;
  din.dArrayLength = 0;
  din.dTag[0] = 0;
  RegisterPropertyInformation(EQM_TAG,"Pulse",&dout,&din,CA_READ,AT_SPECTRUM,1024,"[-1000:1000 V][0:1000 ms]Pulse",PRP_PULSE,NULL);

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 10;
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 1;
  RegisterPropertyInformation(EQM_TAG,"Amplitude",&dout,&din,CA_READ|CA_WRITE|CA_HIST|CA_SAVERESTORE,AT_CHANNEL,10,"[1:1000 V]Pulse Curve Amplitude",PRP_AMPLITUDE,NULL);
  RegisterPropertyInformation(EQM_TAG,"Frequency",&dout,&din,CA_READ|CA_WRITE|CA_HIST|CA_SAVERESTORE,AT_CHANNEL,10,"[1:60]Pulse Curve Frequency",PRP_FREQUENCY,NULL);
  RegisterPropertyInformation(EQM_TAG,"Sigma",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,10,"[0:512]Pulse standard deviation",PRP_SIGMA,NULL);
  RegisterPropertyInformation(EQM_TAG,"Noise",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,10,"[0:100 V]Pulse Noise Level",PRP_NOISE,NULL);
  RegisterPropertyInformation(EQM_TAG,"Shift",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,10,"[0:100 %]Pulse offset",PRP_SHIFT,NULL);

  dout.dFormat = CF_STRUCT;
  strncpy(dout.dTag,"PulseInfo",TAG_NAME_SIZE);
  din.dFormat = CF_STRUCT;
  strncpy(din.dTag,"PulseInfo",TAG_NAME_SIZE);
  RegisterPropertyInformation(EQM_TAG,"PulseInfo",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,10,"Pulse Information",PRP_INFO,NULL);

  DTYPEZERO(dout);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 10;
  DTYPEZERO(din);
  RegisterPropertyInformation(EQM_TAG,"PulseScanValue",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,10,"[0:512 V]Current Pulse Value",PRP_VALUE,NULL);

  DTYPEZERO(dout);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = 10;
  DTYPEZERO(din);
  RegisterPropertyInformation(EQM_TAG,"PulseScanInterval",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,10,"[1:10 sec]scan interval (seconds to advance ptr)",PRP_INTERVAL,NULL);

  /* register device names, bind to device number: */
  for (i=0; i<NUM_DEVICES; i++) RegisterDeviceName(EQM_TAG,plsname[i],i);
}

HistorySpecification myHSpec =
{
  HRS_DEFAULT_POLL_RATE,HRS_DEFAULT_ARCHIVE_RATE,HRS_DEFAULT_DEPTHSHORT,HRS_DEFAULT_DEPTHLONG,
  HRS_DEFAULT_HEARTBEAT,HRS_DEFAULT_REL_TOLERANCE,HRS_DEFAULT_ABS_TOLERANCE,NULL,NULL
};
void plsbkg(void)
{
  int i;
  for (i=0; i<NUM_DEVICES; i++)
  {
    updatePulse(i);
  }
}
int savePropertySettings(char *prpName,float *reference)
{
  int i;
  float fv[NUM_DEVICES];
  for (i=0; i<NUM_DEVICES; i++) 
    fv[i] = *(float *)((BYTE *)reference + i * sizeof(PulseInfo));
  return SavePropertyValues("PLSEQM",prpName,fv,CF_FLOAT,NUM_DEVICES);
}
int plseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,cc, ival, imcarray[NUM_DEVICES];
  float mcarray[NUM_DEVICES], fval;
  PulseInfo *plsf;
  ExportPropertyListStruct prp;

  devnr = GetDeviceNumberEx(EQM_TAG,devName,devProperty);
  if (devnr < 0) devnr = 0;
  if (devnr >= NUM_DEVICES) return illegal_equipment_number;

  if (GetRegisteredPropertyListStruct(EQM_TAG,devProperty,&prp) != 0) return illegal_property;
  prpid = GetPropertyId(EQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_PULSE:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=PutValuesFromFloat(dout,plsbuf[devnr],NUM_VALUES)) != 0) return cc;
        return 0;
    case PRP_AMPLITUDE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(EQM_TAG,devProperty,din,FALSE)) return out_of_range;
          pulseInfoTbl[devnr].amplitude = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = pulseInfoTbl[i].amplitude;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_FREQUENCY:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(EQM_TAG,devProperty,din,FALSE)) return out_of_range;
          pulseInfoTbl[devnr].frequency = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = pulseInfoTbl[i].frequency;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_SIGMA:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(EQM_TAG,devProperty,din,FALSE)) return out_of_range;
          pulseInfoTbl[devnr].sigma = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = pulseInfoTbl[i].sigma;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_NOISE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(EQM_TAG,devProperty,din,FALSE)) return out_of_range;
          pulseInfoTbl[devnr].noise = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = pulseInfoTbl[i].noise;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_SHIFT:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(EQM_TAG,devProperty,din,FALSE)) return out_of_range;
          pulseInfoTbl[devnr].peak = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = pulseInfoTbl[i].peak;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_VALUE:
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = pulseInfoTbl[i].value;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_INTERVAL:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsLong(din,&ival,1)) != 0) return cc;
          if (!AssertRangeValid(EQM_TAG,devProperty,din,FALSE)) return out_of_range;
          pulseInfoTbl[devnr].interval = ival;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            imcarray[i] = pulseInfoTbl[i].interval;
          if ((cc=putValuesFromLongEx(dout,imcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_INFO:
        if (din->dArrayLength)
        { /* allow an atomic set of all relevant values this way : */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"PulseInfo",8)) return invalid_structure_tag;
          plsf = (PulseInfo *)(din->data.vptr);
          if (plsf->amplitude < 1 || plsf->amplitude > 1000) return out_of_range;
          if (plsf->frequency < 1 || plsf->frequency > 100) return out_of_range;
          if (plsf->sigma < 0 || plsf->sigma > 512) return out_of_range;
          if (plsf->noise < 0 || plsf->noise > 100) return out_of_range;
          if (plsf->peak < 0 || plsf->peak > 100) return out_of_range;
          pulseInfoTbl[devnr] = *plsf;
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"PulseInfo",8)) return invalid_structure_tag;
          if (dout->dArrayLength > NUM_DEVICES) dout->dArrayLength = NUM_DEVICES;
          plsf = (PulseInfo *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            plsf[i] = pulseInfoTbl[(i+devnr)%NUM_DEVICES];
          }
        }
        return 0;
    default: 
        return illegal_property;
  }
}
char thisFecName[FEC_NAME_SIZE] = "";
char thisContext[CONTEXT_NAME_SIZE] = "";
char thisServerName[EXPORT_NAME_SIZE] = EQM_EXPORT;
char *thisSubSystem = "TEST";
char thisFecDescription[FEC_DESC_SIZE] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
UINT16 thisPortOffset = 0;

void PreSystemInit(void)
{
//  int cc = FindNameServerOnNetwork();
//  SetStandAlone(TRUE);
//  SetAfInet(AF_INET6);
  SetCycleDelay(0);
  if (strlen(thisContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      strcat(thisFecName,"PLSFEC");
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s pulse server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
}

void PostSystemInit(void)
{
  RegisterEquipmentModule(thisServerName,EQM_TAG,NUM_DEVICES,plseqm,plsinit,plsbkg,200,NULL);
  /* don't let people poll faster than 20 Hz */
  MinPollingRate = MaxPollingRate = 50;

}

