/*
   Module: sineqm.c  - example TINE sine generator server, which illustrates
           the basic server API
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "sineqm.h"

#ifndef DOS
# if defined(UNIX) || defined(WIN32)
# define random(r) (r != 0 ? rand()%(r) : 0)
# else
# define random(r) (r != 0 ? (i*(69069))%(r) : 0)
# endif
#endif

#define PRP_SINE      1
#define PRP_AMPLITUDE 2
#define PRP_FREQUENCY 3
#define PRP_PHASE     4
#define PRP_NOISE     5
#define PRP_INFO      6
#define PRP_DAMPING   7
#define PRP_TABLE     8
#define PRP_TABLE_NAM 9
#define PRP_TABLE_NAM_NAM 10


#define NUM_DEVICES 10
#define NUM_VALUES (1024*8)

/*
   The code section below illustrates how one registers a structure
   to be exported.  
*/

typedef struct
{
  float amplitude;
  float frequency;
  float noise;
  float phase;
  float damping;
  char  description[64];
} SineInfo;

#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,amplitude),1,CF_FLOAT,"amplitude")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,frequency),1,CF_FLOAT,"frequency")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,noise),1,CF_FLOAT,"noise")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,phase),1,CF_FLOAT,"phase")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,damping),1,CF_FLOAT,"damping")) quit(1);
  if (addFieldToStruct("SineInfo",OFFSETIN(SineInfo,description),64,CF_TEXT,"description")) quit(1);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("SineInfo",sizeof(SineInfo),NUM_DEVICES)) quit(1);
}

SineInfo sineInfoTable[NUM_DEVICES];
float sinbuf[NUM_DEVICES][NUM_VALUES];

void updateSine(int devnr)
{ /* used internally -> don't bother with bounds checking */
  int i, k;
  float *target = sinbuf[devnr];
  float f = sineInfoTable[devnr].frequency;
  int n = (int)sineInfoTable[devnr].noise;
  int p = (int)sineInfoTable[devnr].phase;
  float a = sineInfoTable[devnr].amplitude;
  float d = sineInfoTable[devnr].damping/(float)100.0;
  ClearAlarm(SINEQM_TAG, devnr);
  SetAlarmEx(SINEQM_TAG, devnr, 512, NULL, 0);
  for (i=0; i<NUM_VALUES; i++)
  {
    k = (i + p) % NUM_VALUES;
    target[k] = random(n) + (float)(sin(f*i*6.2832/(NUM_VALUES/8))*a);
    if (d > 0.0)
    {
      target[k] *= ((float)1.0 - (d * (float)i)/(float)NUM_VALUES);
    }
  }
  return;
}
char *sinname[NUM_DEVICES] = {
  "SineGen0","SineGen1","SineGen2","SineGen3","SineGen4",
  "SineGen5","SineGen6","SineGen7","SineGen8","SineGen9" 
};

void sininit(void)
{
  int i;
  float fv[NUM_DEVICES];
  DTYPE dout, din;

  /* some initial values for all sine devices */
  for (i=0; i<NUM_DEVICES; i++)
  {
    sineInfoTable[i].amplitude = 256.0;
    sineInfoTable[i].frequency = 1.0;
    sineInfoTable[i].phase = 0.0;
    sineInfoTable[i].noise = 30.0;
    sineInfoTable[i].damping = 0.0;
    sprintf(sineInfoTable[i].description,"Sine Generator %d at your disposal",i);
  }
  registerStructs();

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 8192;
  dout.dTag[0] = 0;
  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_NULL;
  din.dArrayLength = 0;
  din.dTag[0] = 0;
  RegisterPropertyInformation(SINEQM_TAG,"Sine",&dout,&din,CA_READ,AT_SPECTRUM,8192,"[-1000:1000 V][0:1000 ms]Sine Curve",PRP_SINE,NULL);

  dout.dArrayLength = 81920;
  RegisterPropertyInformation(SINEQM_TAG,"SineTable",&dout,&din,CA_READ,AT_SPECTRUM|AT_DOUBLE,8192,"[-1000:1000 V][0:1000 ms]Sine Curve Table",PRP_TABLE,NULL);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 1;
  RegisterPropertyInformation(SINEQM_TAG,"SineTable.NAM",&dout,&din,CA_READ,AT_SCALAR,1,"Sine Curve Table device name",PRP_TABLE_NAM,NULL);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 10;
  RegisterPropertyInformation(SINEQM_TAG,"SineTable.NAM.NAM",&dout,&din,CA_READ,AT_CHANNEL,10,"Sine Curve Table channel names",PRP_TABLE_NAM_NAM,NULL);

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 10;
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 1;
  RegisterPropertyInformation(SINEQM_TAG,"Amplitude",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,10,"[1:1000 V]Sine Curve Amplitude",PRP_AMPLITUDE,NULL);
  if (RestorePropertyValues("SINEQM","Amplitude",fv,CF_FLOAT,NUM_DEVICES) == 0)
  { /* restore property values this way as it's handled as a multi-channel array : */
    for (i=0; i<NUM_DEVICES; i++) sineInfoTable[i].amplitude = fv[i];
  }
  RegisterPropertyInformation(SINEQM_TAG,"Frequency",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,10,"[1:60]Sine Curve Frequency",PRP_FREQUENCY,NULL);
  if (RestorePropertyValues("SINEQM","Frequency",fv,CF_FLOAT,NUM_DEVICES) == 0)
  { /* restore property values this way as it's handled as a multi-channel array : */
    for (i=0; i<NUM_DEVICES; i++) sineInfoTable[i].frequency = fv[i];
  }
  RegisterPropertyInformation(SINEQM_TAG,"Phase",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,10,"[0:512]Sine Curve Phase",PRP_PHASE,NULL);
  if (RestorePropertyValues("SINEQM","Phase",fv,CF_FLOAT,NUM_DEVICES) == 0)
  { /* restore property values this way as it's handled as a multi-channel array : */
    for (i=0; i<NUM_DEVICES; i++) sineInfoTable[i].phase = fv[i];
  }
  RegisterPropertyInformation(SINEQM_TAG,"Noise",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,10,"[0:100 V]Sine Curve Noise Level",PRP_NOISE,"/TEST/SineServer/SineGen4[Noise]");
  if (RestorePropertyValues("SINEQM","Noise",fv,CF_FLOAT,NUM_DEVICES) == 0)
  { /* restore property values this way as it's handled as a multi-channel array : */
    for (i=0; i<NUM_DEVICES; i++) sineInfoTable[i].noise = fv[i];
  }
  RegisterPropertyInformation(SINEQM_TAG,"Damping",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,10,"[0:100 %]Sine Curve Dampling Level",PRP_DAMPING,NULL);
  if (RestorePropertyValues("SINEQM","Damping",fv,CF_FLOAT,NUM_DEVICES) == 0)
  { /* restore property values this way as it's handled as a multi-channel array : */
    for (i=0; i<NUM_DEVICES; i++) sineInfoTable[i].damping = fv[i];
  }

  dout.dFormat = CF_STRUCT;
  strncpy(dout.dTag,"SineInfo",TAG_NAME_SIZE);
  din.dFormat = CF_STRUCT;
  strncpy(din.dTag,"SineInfo",TAG_NAME_SIZE);
  RegisterPropertyInformation(SINEQM_TAG,"SineInfo",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,10,"Sine Generator Information",PRP_INFO,NULL);

  /* register device names, bind to device number: */
  for (i=0; i<NUM_DEVICES; i++) RegisterDeviceName(SINEQM_TAG,sinname[i],i);
}

HistorySpecification myHSpec =
{
  HRS_DEFAULT_POLL_RATE,HRS_DEFAULT_ARCHIVE_RATE,HRS_DEFAULT_DEPTHSHORT,HRS_DEFAULT_DEPTHLONG,
  HRS_DEFAULT_HEARTBEAT,HRS_DEFAULT_REL_TOLERANCE,HRS_DEFAULT_ABS_TOLERANCE,NULL,NULL
};
void sinbkg(void)
{
  static int ptr = 0;
  static int cnt = 0;
  static time_t last_t = 0;
  static int done = FALSE;
  int i;
  time_t t = time(NULL);
  if (last_t == 0) last_t = t;
  /* do IO or something useful here */
  for (i=0; i<NUM_DEVICES; i++)
  {
    updateSine(i);
  }
  if (t < last_t + 5) return;
  last_t = t;
  if (done) return;
  AppendHistoryInformationEx(SINEQM_TAG, "Amplitude", "SineGen0", 10, CF_FLOAT, 5, &myHSpec, 0);
  done = TRUE;
}
int savePropertySettings(char *prpName,float *reference)
{
  int i;
  float fv[NUM_DEVICES];
  for (i=0; i<NUM_DEVICES; i++) 
    fv[i] = *(float *)((BYTE *)reference + i * sizeof(SineInfo));
  return SavePropertyValues("SINEQM",prpName,fv,CF_FLOAT,NUM_DEVICES);
}
int sineqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,cc;
  float mcarray[NUM_DEVICES], fval;
  SineInfo *sinf;
  ExportPropertyListStruct prp;

  devnr = GetDeviceNumberEx(SINEQM_TAG,devName,devProperty);
  if (devnr < 0) devnr = 0;
  if (devnr >= NUM_DEVICES) return illegal_equipment_number;

  if (GetRegisteredPropertyListStruct(SINEQM_TAG,devProperty,&prp) != 0) return illegal_property;
  prpid = GetPropertyId(SINEQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_SINE:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=PutValuesFromFloat(dout,sinbuf[devnr],NUM_VALUES)) != 0) return cc;
        return 0;
    case PRP_TABLE:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=PutValuesFromFloat(dout,sinbuf[0],NUM_VALUES*NUM_DEVICES)) != 0) return cc;
        return 0;
    case PRP_TABLE_NAM:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=PutValuesFromString(dout,"SineTable",CF_TEXT,9,0)) != 0) return cc;
        dout->dArrayLength = dout->dFormat == CF_TEXT ? 9 : 1;
        return 0;
    case PRP_TABLE_NAM_NAM:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=PutValuesFromString(dout,(char *)sinname,CF_STRING,10,devnr)) != 0) return cc;
        return 0;
    case PRP_AMPLITUDE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(SINEQM_TAG,devProperty,din,FALSE)) return out_of_range;
          sineInfoTable[devnr].amplitude = fval;
          savePropertySettings(devProperty,&sineInfoTable[0].amplitude);
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].amplitude;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_FREQUENCY:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(SINEQM_TAG,devProperty,din,FALSE)) return out_of_range;
          sineInfoTable[devnr].frequency = fval;
          savePropertySettings(devProperty,&sineInfoTable[0].frequency);
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].frequency;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_PHASE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(SINEQM_TAG,devProperty,din,FALSE)) return out_of_range;
          sineInfoTable[devnr].phase = fval;
          savePropertySettings(devProperty,&sineInfoTable[0].phase);
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].phase;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_NOISE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(SINEQM_TAG,devProperty,din,FALSE)) return out_of_range;
          sineInfoTable[devnr].noise = fval;
          savePropertySettings(devProperty,&sineInfoTable[0].noise);
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].noise;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_DAMPING:
      if (devnr > 2) return 512+devnr;
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (!AssertRangeValid(SINEQM_TAG,devProperty,din,FALSE)) return out_of_range;
          sineInfoTable[devnr].damping = fval;
          savePropertySettings(devProperty,&sineInfoTable[0].damping);
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].damping;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_INFO:
        if (din->dArrayLength)
        { /* allow an atomic set of all relevant values this way : */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"SineInfo",8)) return invalid_structure_tag;
          sinf = (SineInfo *)(din->data.vptr);
          if (sinf->amplitude < 1 || sinf->amplitude > 1000) return out_of_range;
          if (sinf->frequency < 1 || sinf->frequency > 100) return out_of_range;
          if (sinf->phase < 0 || sinf->phase > 512) return out_of_range;
          if (sinf->noise < 0 || sinf->noise > 100) return out_of_range;
          if (sinf->damping < 0 || sinf->damping > 100) return out_of_range;
          sineInfoTable[devnr] = *sinf;
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"SineInfo",8)) return invalid_structure_tag;
          if (dout->dArrayLength > NUM_DEVICES) dout->dArrayLength = NUM_DEVICES;
          sinf = (SineInfo *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            sinf[i] = sineInfoTable[(i+devnr)%NUM_DEVICES];
          }
        }
        return 0;
    default: 
        return illegal_property;
  }
}
char thisFecName[FEC_NAME_SIZE+4] = "";
char thisContext[CONTEXT_NAME_SIZE+4] = "";
char thisServerName[EXPORT_NAME_SIZE+4] = SINEQM_EXPORT;
char *thisSubSystem = "TEST";
char thisFecDescription[FEC_DESC_SIZE+4] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
UINT16 thisPortOffset = 0;

void PreSystemInit(void)
{
//  int cc = FindNameServerOnNetwork();
//  SetStandAlone(TRUE);
//  SetAfInet(AF_INET6);
  SetCycleDelay(0);
  if (strlen(thisContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      strcat(thisFecName,"SINFEC");
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s sine curve server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
}
void myCleanup(void)
{
  SystemStop(TRUE);
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(thisServerName,SINEQM_TAG,NUM_DEVICES,sineqm,sininit,sinbkg,200,NULL);
  //RegisterEquipmentModule("CosineVS15", COSEQM_TAG, NUM_DEVICES, sineqm, NULL, NULL, 0, NULL);
  /* don't let people poll faster than 20 Hz */
  SetMinimumAllowedPollingInterval(50);
  SetSystemCleanupFunction(myCleanup);
  {
    int i;
    char str[256];
    for (i = 0; i < 20; i++)
    {
      sprintf(str, "just a little error of mine %d", i);
      RegisterErrorCode(512 + i, str, "This little error of mine happens when things go wrong!");
    }
  }
}
