/**
 * \file staeqm.c
 */
#include "stdlib.h"
#include "math.h"
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include "tine.h"
#if defined(UNIX)
#include "dirent.h"
# define DBPATH "../database/"
#elif defined(WIN32)
#include "direct.h"
# define DBPATH "..\\database\\"
#endif

#define STAEQM_TAG "STAEQM"
#define STAEQM_EXPORT "STATE"

#define MAX_NUM_CONFIGS 500
#define MAX_STATES 256
#define MAX_PROCS 100
#define MAX_MACHINE_STATES 50
#define INFO_CUSHION 50

#define admin_rights_required 513

time_t getTimeStampFromLogEntry(char *str); // use this routine from the TINE kernel
int getDeclStateIndex(char *devname);
int getDeclProcIndex(char *devname);
int logproblem(char *text,...);
void PostSystemInit(void);
void PreSystemInit(void);
FILE *createAndOpen(char *fname,char *attr);
int sndTriggerToArchive(char *trg);

BYTE stateRepCounts[MAX_STATES]; // each state must be accounted for once !
char stateValidationErr[1024];
int validateProfileCompleteness(char *profile);

typedef struct StateCorrection
{
  double start;
  double stop;
  int counts;
  int reserved;
  char startStr[64];
  char stopStr[64];
  char fromState[64];
  char toState[64];
  char usr[16];
  char cmt[256];
} StateCorrection;
#define MAX_CORRECTION_PER_DAY 2048

typedef struct
{
  char username[16];
  char useraddr[16];
} CALLER;
CALLER lastCallerToChangeState;

#define PTYPE1 1
#define PTYPE2   2
#define BOTH     3

#define STATEFILE "state.csv"
#define STATEINFOFILE "stateinfo"
#define PROCINFOFILE "procinfo"
#define MACHINFOFILE "machinfo"
#define PROFILESFILE "profiles.csv"
#define CONFIGSFILE "configs.csv"
#define CORRECTIONSFILE "corrections.csv"
#define CORRLOGFILE "corrections.log"

typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char type[16];
  char states[512];
  char color[16];
} OprProfile;

#define PRF_ALWAYS_VISIBLE  0x01
#define PRF_NON_OPERATIONAL 0x02
#define PRF_EXCLUDE_STATS   0x04
#define PRF_PROBLEMS        0x08
#define PRF_NOT_AVAILABLE   0x10

typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char type[16];
  char states[512];
  char color[16];
  int  metric;
  int  flags;
  int  reserved1;
  int  reserved2;
} OprProfileEx;

#define MCA_SINGLE_ENTRY  0
#define MCA_APPEND_ENTRY  1
#define MCA_REVERSE_ENTRY 2

typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char tag[TAG_NAME_SIZE];
  char desc[PROPERTY_DESC_SIZE];
  char shell[64];
  char alias[PROPERTY_NAME_SIZE];
  char partner[PROPERTY_NAME_SIZE];
  char reserved[128];
  int append;
} McaSet;

typedef struct
{
  char stateParams[64];
  char stateProfiles[64];
  char desc[64];
} OprViewFile;

OprViewFile *oprViewFiles = NULL;
int gNumOprViewFiles = 0;

CSVCOLUMN csvOprViewFileColumns[] =
{
  { "PARAMETERS","",0,CF_TEXT,64,(BYTE *)OFFSETIN(OprViewFile,stateParams),NULL,0 },
  { "PROFILES","",0,CF_TEXT,64,(BYTE *)OFFSETIN(OprViewFile,stateProfiles),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(OprViewFile,desc),NULL,0 },
};
CSVDB csvOprViewFileDb =
{
  csvOprViewFileColumns,
  sizeof(csvOprViewFileColumns)/sizeof(csvOprViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(OprViewFile),
  0
};
int readViewerFiles(void)
{
  int i, cc=0;
  char fn[128];
  sprintf(fn,"%.96sCONFIGS%c",DBPATH,FS_DELIMITER);
  if ((cc=csvReadFile(fn,CONFIGSFILE,&csvOprViewFileDb,(void **)&oprViewFiles)) != 0) ccerr(cc);
  gNumOprViewFiles = csvOprViewFileDb.siz;
  for (i=0; i<gNumOprViewFiles; i++)
  {
    validateProfileCompleteness(oprViewFiles[i].desc);
  }
err:
  return cc;
}
int updateViewerFiles(void)
{
  int i, cc=0;
  char fn[128];
  FILE *fp = NULL;
  sprintf(fn,"%.96sCONFIGS%c/%32s",DBPATH,FS_DELIMITER,CONFIGSFILE);

  if (gNumOprViewFiles == 0)
  {
    unlink(fn);
    return 0;
  }
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Parameters, Profiles, Description\n");
  for (i=0; i<gNumOprViewFiles; i++)
  {
    if (!stricmp(oprViewFiles[i].desc,"REMOVE")) continue;
    fprintf(fp,"%.64s, %.64s, %.64s\n",
      oprViewFiles[i].stateParams,oprViewFiles[i].stateProfiles,oprViewFiles[i].desc);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc == 0) readViewerFiles();
  return cc;
}
int appendViewerFile(char *desc,char *filename,int *n,CSVDB *pdb,OprViewFile **popr)
{
  int i,k,cc=0;
  char fn[256],cfgfn[72];
  FILE *fp = NULL;
  OprViewFile *opr = *popr;
  sprintf(fn,"%.96sCONFIGS%c%s",DBPATH,FS_DELIMITER,filename);
  if (*n == 0)
  { /* doesn't exist yet : write the header */
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
    fprintf(fp,"Parameters, Profiles, Description\n");
  }
  else
  {
    if ((fp=fopen(fn,"at")) == NULL) ccerr(file_error);
  }
  for (i=0,k=0; i<56; i++)
  { /* make a new file name */
    if (desc[i] == 0) break;
    if (desc[i] == ' ' || desc[i] == '-' || desc[i] == '_') continue;
    cfgfn[k++] = tolower(desc[i]);
  }
  cfgfn[k] = 0;
  fprintf(fp,"%.56s-param, %.56s-prof, %.64s\n",cfgfn,cfgfn,desc);
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0) return -cc;
  sprintf(fn,"%.96sCONFIGS%c",DBPATH,FS_DELIMITER);
  if ((cc=csvReadFile(fn,filename,pdb,(void **)&opr)) != 0) ccerr(cc);
  *popr = opr;
  *n = pdb->siz;
  return *n;
}
int commitMcaSet(char *fileName,McaSet *mcaset,int setsiz)
{
  FILE *fp=NULL;
  char scratch[256],appnd[64],fn[128];
  int i = 0, cc = 0;

  if (fileName == NULL || mcaset == NULL || setsiz < 1) ccerr(argument_list_error);
  sprintf(fn,"%.96sCONFIGS%c%s",DBPATH,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Context, Server, Property, Tag, Description, Append, Shell, Alias, Partner\n");
  for (i=0; i<setsiz; i++)
  {
    memset(scratch,0,256);
    strncat(scratch,mcaset[i].srv,EXPORT_NAME_SIZE);
    if (strlen(mcaset[i].dev) > 0) sprintf(&scratch[strlen(scratch)],"/%s",mcaset[i].dev);
    strcpy(appnd,mcaset[i].append ? "TRUE" : "FALSE");
    if (mcaset[i].append & MCA_REVERSE_ENTRY) strcat(appnd,"|REVERSE");

    fprintf(fp,"%s, %s, %s, %s, %s, %s, %s, %s, %s\n",
      mcaset[i].ctx,scratch,mcaset[i].prp,mcaset[i].tag,mcaset[i].desc,appnd,mcaset[i].shell,mcaset[i].alias,mcaset[i].partner);
  }

err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int populateMcaSet(char *fileName,McaSet *mcaset,int *setsiz)
{
  FILE *fp=NULL;
  char str[256],scratch[256],fn[128];
  char *c,*srv,*dev;
  int n = 0, done = 0,cc = 0;
  int ctx_col=-1,srv_col=-1,prp_col=-1,dev_col=-1,tag_col=-1,dsc_col=-1;
  int alias_col=-1,partner_col=-1,append_col=-1,shell_col=-1;

  if (fileName == NULL || mcaset == NULL || setsiz == NULL) ccerr(argument_list_error);
  if (*setsiz < 1) ccerr(dimension_error);
  sprintf(fn,"%.96sCONFIGS%c%s",DBPATH,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      ctx_col = findcol(str,"CONTEXT");
      if ((srv_col=findcol(str,"SERVER")) < 0) ccerr(no_such_line);
      if ((prp_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      dev_col = findcol(str,"DEVICE");
      tag_col = findcol(str,"TAG");
      dsc_col = findcol(str,"DESCRIPTION");
      alias_col = findcol(str,"ALIAS");
      partner_col = findcol(str,"PARTNER");
      if ((shell_col=findcol(str,"SHELL")) < 0) shell_col = findcol(str,"HIGH");
      append_col = findcol(str,"APPEND");
      done = 1;
      continue;
    }
    if (n >= *setsiz) break;
    memset(&mcaset[n],0,sizeof(McaSet));
    strncpy(scratch,colptr(srv_col,str),256);
    srv = scratch;
    if ((c=strchr(srv,'/')) != NULL)
    {
      *c++ = 0;
      dev = c;
    }
    else
    {
      dev = NULL;
    }
    if (ctx_col >= 0) 
    {
      strncpy(mcaset[n].ctx,colptr(ctx_col,str),CONTEXT_NAME_SIZE);
    }
    else
    {
      strncpy(mcaset[n].ctx,gDeviceContext,CONTEXT_NAME_SIZE);
    }
    strncpy(mcaset[n].srv,srv,EXPORT_NAME_SIZE);
    if (dev != NULL) strncpy(mcaset[n].dev,dev,DEVICE_NAME_SIZE);
    strncpy(mcaset[n].prp,colptr(prp_col,str),PROPERTY_NAME_SIZE);
    if (dev_col >= 0) 
    { /* if given here, this overrides the above ... */
      strncpy(mcaset[n].dev,colptr(dev_col,str),DEVICE_NAME_SIZE);
    }
    if (tag_col >= 0) 
    {
      strncpy(mcaset[n].tag,colptr(tag_col,str),TAG_NAME_SIZE);
    }
    if (dsc_col >= 0) 
    {
      strncpy(mcaset[n].desc,colptr(dsc_col,str),PROPERTY_DESC_SIZE);
    }
    if (shell_col >= 0) 
    {
      strncpy(mcaset[n].shell,colptr(shell_col,str),64);
    }
    if (alias_col >= 0) 
    {
      strncpy(mcaset[n].alias,colptr(alias_col,str),64);
    }
    if (partner_col >= 0) 
    {
      strncpy(mcaset[n].partner,colptr(partner_col,str),64);
    }
    if (append_col >= 0) 
    {
      strncpy(scratch,colptr(append_col,str),64);
      strupr(scratch);
      if (strstr(scratch,"TRUE") != NULL || strstr(scratch,"YES") != NULL) mcaset[n].append = MCA_APPEND_ENTRY;
      if (strstr(scratch,"REVERSE") != NULL) mcaset[n].append = MCA_APPEND_ENTRY|MCA_REVERSE_ENTRY;
    }
    n++;
  }
  *setsiz = n;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("error reading config file %s : %s",fileName,erlst[cc]);
  }
  return cc;
}
#define GETFIELDP(n,x,f) ((x) ?\
   (BYTE *)(&((OprProfileEx *)set)[n]) + OFFSETIN(OprProfileEx,f) :\
   (BYTE *)(&((OprProfile *)set)[n]) + OFFSETIN(OprProfile,f))
int commitProfileSet(char *fileName,void *set,int setsiz,int extended)
{
  FILE *fp=NULL;
  char fn[128],states[520],color[32],name[72],type[32],*c;
  int i = 0, cc = 0;
  int metric = 0, flags = 0;

  if (fileName == NULL || set == NULL || setsiz < 1) ccerr(argument_list_error);
  sprintf(fn,"%.96sCONFIGS%c%s",DBPATH,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Name, Type, States, Color, Metric, Flags\n");
  for (i=0; i<setsiz; i++)
  {
    states[512] = 0; color[16] = 0;
    c = GETFIELDP(i,extended,states);
    if (strchr(c,',') != NULL) sprintf(states,"\"%.512s\"",c); else strncpy(states,c,512);
    c = GETFIELDP(i,extended,color);
    if (strchr(c,',') != NULL) sprintf(color,"\"%.16s\"",c); else strncpy(color,c,16);
    c = GETFIELDP(i,extended,name);
    strncpy(name,c,64); name[64] = 0;
    while ((c=strchr(name,',')) != NULL) *c = ' '; // blow away any commas in the name !
    c = GETFIELDP(i,extended,type);
    strncpy(type,c,16); type[16] = 0;
    if ((c=strchr(type,',')) != NULL) *c = 0;  // blow away any commas in the type (should never happen) !
    if (extended)
    {
      metric = ((OprProfileEx *)set)[i].metric;
      flags = ((OprProfileEx *)set)[i].flags;
    }
    fprintf(fp,"%.64s, %.16s, %.520s, %.16s, %d, %d\n",
      name,type,states,color,metric,flags);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int getProfileProcList(char *states,NAME64 *lst,int *lstsiz)
{
  int i,n;
  char *c,*ce,*pn,buf[512];
  if (states == NULL || lst == NULL || lstsiz == NULL) return argument_list_error;
  n = *lstsiz;
  strncpy(buf,states,512);
  if ((c=strchr(buf,'"')) != NULL) 
  {
    *c++;
    if ((ce=strchr(c,'"')) != NULL) *ce = 0;
  }
  else
  {
    c = buf;
  }
  for (i=0,pn=strtok(c,","); i<n && pn != NULL; i++, pn=strtok(NULL,","))
  {
    strncpy(lst[i].name,strtrm(pn),64);
  }
  *lstsiz = i;
  return 0;
}
int populateProfileSet(char *fileName,void *set,int *setsiz,int extended)
{
  FILE *fp=NULL;
  char str[1024],fn[128],scratch[64];
  int n = 0, done = 0,cc = 0, itemsize;
  int nam_col=-1,sta_col=-1,typ_col=-1,col_col=-1;
  int met_col=-1,flg_col=-1;

  if (fileName == NULL || set == NULL || setsiz == NULL) ccerr(argument_list_error);
  if (*setsiz < 1) ccerr(dimension_error);
  sprintf(fn,"%.96sCONFIGS%c%s",DBPATH,FS_DELIMITER,fileName);
  if (strnicmp(&fn[strlen(fn)-4],".csv",4)) strcat(fn,".csv");
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  itemsize = extended ? sizeof(OprProfileEx) : sizeof(OprProfile);
  memset(set,0,itemsize * (*setsiz));
  while (fgets(str,1023,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((nam_col=findcol(str,"NAME")) < 0) ccerr(no_such_line);
      if ((sta_col=findcol(str,"STATES")) < 0) ccerr(no_such_line);
      typ_col = findcol(str,"TYPE");
      col_col = findcol(str,"COLOR");
      met_col = findcol(str,"METRIC");
      flg_col = findcol(str,"FLAGS");
      done = 1;
      continue;
    }
    if (n >= *setsiz) break;
    strncpy(GETFIELDP(n,extended,name),colptr(nam_col,str),DEVICE_NAME_SIZE);
    getColumnStr(sta_col,str,GETFIELDP(n,extended,states),512);
    if (typ_col >= 0)
    { 
      strncpy(GETFIELDP(n,extended,type),colptr(typ_col,str),16);
    }
    else
    {
      strncpy(GETFIELDP(n,extended,type),"optional",16);
    }
    if (col_col >= 0)
    { 
      strncpy(GETFIELDP(n,extended,color),colptr(col_col,str),16);
    }
    if (met_col >= 0 && extended)
    { 
      strncpy(scratch,colptr(met_col,str),16);
      ((OprProfileEx *)set)[n].metric = atoi(scratch);
    }
    if (flg_col >= 0 && extended)
    { 
      strncpy(scratch,colptr(flg_col,str),16);
      ((OprProfileEx *)set)[n].flags = atoi(scratch);
    }
    n++;
  }
  *setsiz = n;
err:
  if (fp != NULL) fclose(fp);
  if (cc)
  {
    feclog("error reading config file %s : %s",fileName,erlst[cc]);
  }
  return cc;
}

typedef struct 
{
  char devname[DEVICE_NAME_SIZE];
  char statetag[DEVICE_NAME_SIZE];
  int  machine;     /* OR of machine 1 and machine 2 (if we're colliding) */
  int  evntrg;      /* TRUE: send an event trigger upon state declaration */
  time_t ltd;       /* last time of declaration */
  UINT32 ttd;       /* total time at declared state */
  UINT32 ttd_bc;    /* total time at declared state with beam conditions */
  UINT32 ttd_tc;    /* total time at declared state with test conditions */
  UINT32 ttd_ec;    /* total time at declared state with error conditions (unavailable) */
  UINT32 tnd;       /* total number of state declarations */
} DECLSTATE;

DECLSTATE *declStates = NULL;
int ndeclstates = 0;
int declstatesCapacity = 0;

int gScheduleDeclState = FALSE;
time_t gLastCorrectionTimeStamp = 0;

#define NMACHINETYPES 2
UINT32 declaredState[NMACHINETYPES] = {0,0};  /* undefined */
UINT32 prevDeclaredState[NMACHINETYPES] = {0,0};  /* undefined */

typedef struct 
{
  char devname[DEVICE_NAME_SIZE];
  char statetag[DEVICE_NAME_SIZE];
  int  machine;
  int  reserved;
} DCLSTATE;

typedef struct
{
  char devname[64];
  char structfile[64];
  int  machine;
  int  nstates;
  time_t ltd;       /* last time of declaration */
  UINT32 tnd;       /* total number of procedure declarations */
  NAME64 *statenames;
} DECLPROC;

DECLPROC *declProcs = NULL;
int ndeclprocs = 0;
int declprocsCapacity = 0;

UINT32 declaredProc[NMACHINETYPES] = {0,0}; /* undefined */
UINT32 prevDeclaredProc[NMACHINETYPES] = {0,0}; /* undefined */

typedef struct
{
  char devname[DEVICE_NAME_SIZE];
  char structfile[DEVICE_NAME_SIZE];
  int  machine;
  int  nstates;
} DCLPROC;

#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  if (nadmins == 0) return TRUE;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];
 
  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int nskips = 0;
int *tskips = NULL;
#define SKIPFIRST_FILE "skipfirsttrigger.csv"
int readSkipFirstFile(void)
{
  char fn[128];
  FILE *fp = NULL;
  int n, i, devnr, cc = 0, done = 0;
  int trg_col = -1;
  char str[256], hdr[256], scratch[128];

  if (tskips != NULL)
  {
    free(tskips);
    nskips = 0;
  }
  if ((n = GetDataBaseSize(SKIPFIRST_FILE)) <= 0) ccerr(-n);
  if ((tskips = (int *)calloc(n, sizeof(int))) == NULL) ccerr(out_of_server_memory);
  if ((fp = fopen(fn, "rt")) == NULL) ccerr(no_such_file);
  for (done = 0, i = 0; fgets(str, 255, fp) != NULL && i < nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n", str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr, str, 255);
      if ((trg_col = findcol(hdr, "TRIGGER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch, colptr(trg_col, str), 64);
    devnr = getDeclStateIndex(scratch);
    if (devnr < 0) continue;
    tskips[i] = devnr;
    i++;
  }
  nskips = i;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int triggerInSkipList(int trigger)
{
  int i;
  for (i = 0; i < nskips; i++) if (tskips[i] == trigger) return TRUE;
  return FALSE;
}
int skipFirst(int trigger)
{
  static int lastTrigger = -1;
  static time_t lastSkip = 0;
  time_t tnow = time(NULL);
  if (lastTrigger >= 0 && tnow > lastSkip + 60)
  {
    lastTrigger = -1;
  }
  if (!triggerInSkipList(trigger)) return FALSE;
  if (trigger == lastTrigger) return FALSE;
  lastTrigger = trigger;
  lastSkip = tnow;
  return TRUE;
}
int needsRestart = FALSE;
void RestartServer(void)
{
  static int restartCountdown = 5;
  if (restartCountdown-- > 0) return;
  FlushContractTable();
  SetServerCycleIdle(TRUE);
  FreeEQPModules();
  ServerExitCondition = 0;
  PostSystemInit();
  SetServerCycleIdle(FALSE);
  needsRestart = FALSE;
  restartCountdown = 5;
}
#define quit(i) { feclog("%s: %s",STAEQM_TAG,erlst[i]); exit(1); }
void RegisterStructs(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,ctx),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,tag),TAG_NAME_SIZE,CF_TEXT,"tag")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,desc),PROPERTY_DESC_SIZE,CF_TEXT,"description")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,shell),64,CF_TEXT,"shell")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,alias),PROPERTY_NAME_SIZE,CF_TEXT,"alias")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,partner),PROPERTY_NAME_SIZE,CF_TEXT,"partner")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,reserved),128,CF_TEXT,"reserved")) quit(out_of_local_memory);
  if (addFieldToStruct("McaCfg",OFFSETIN(McaSet,append),1,CF_LONG,"append")) quit(out_of_local_memory);
  if (sealTaggedStruct("McaCfg",sizeof(McaSet),MAX_NUM_CONFIGS)) quit(out_of_local_memory);

  if (addFieldToStruct("PrfCfg",OFFSETIN(OprProfile,name),DEVICE_NAME_SIZE,CF_TEXT,"name")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfg",OFFSETIN(OprProfile,type),16,CF_TEXT,"type")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfg",OFFSETIN(OprProfile,states),512,CF_TEXT,"states")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfg",OFFSETIN(OprProfile,color),16,CF_TEXT,"color")) quit(out_of_local_memory);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("PrfCfg",sizeof(OprProfile),MAX_NUM_CONFIGS)) quit(out_of_local_memory);

  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,name),DEVICE_NAME_SIZE,CF_TEXT,"name")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,type),16,CF_TEXT,"type")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,states),512,CF_TEXT,"states")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,color),16,CF_TEXT,"color")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,metric),1,CF_INT32,"metric")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,flags),1,CF_INT32,"flags")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,reserved1),1,CF_INT32,"reserved1")) quit(out_of_local_memory);
  if (addFieldToStruct("PrfCfgEx",OFFSETIN(OprProfileEx,reserved2),1,CF_INT32,"reserved2")) quit(out_of_local_memory);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("PrfCfgEx",sizeof(OprProfileEx),MAX_NUM_CONFIGS)) quit(out_of_local_memory);

  if (addFieldToStruct("DclState",OFFSETIN(DCLSTATE,devname),DEVICE_NAME_SIZE,CF_TEXT,"name")) quit(out_of_local_memory);
  if (addFieldToStruct("DclState",OFFSETIN(DCLSTATE,statetag),DEVICE_NAME_SIZE,CF_TEXT,"tag")) quit(out_of_local_memory);
  if (addFieldToStruct("DclState",OFFSETIN(DCLSTATE,machine),1,CF_INT32,"machine")) quit(out_of_local_memory);
  if (addFieldToStruct("DclState",OFFSETIN(DCLSTATE,reserved),1,CF_INT32,"reserved")) quit(out_of_local_memory);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("DclState",sizeof(DCLSTATE),MAX_STATES)) quit(out_of_local_memory);

  if (addFieldToStruct("DclProc",OFFSETIN(DCLPROC,devname),DEVICE_NAME_SIZE,CF_TEXT,"name")) quit(out_of_local_memory);
  if (addFieldToStruct("DclProc",OFFSETIN(DCLPROC,structfile),DEVICE_NAME_SIZE,CF_TEXT,"file")) quit(out_of_local_memory);
  if (addFieldToStruct("DclProc",OFFSETIN(DCLPROC,machine),1,CF_INT32,"machine")) quit(out_of_local_memory);
  if (addFieldToStruct("DclProc",OFFSETIN(DCLPROC,nstates),1,CF_INT32,"nstates")) quit(out_of_local_memory);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("DclProc",sizeof(DCLPROC),MAX_PROCS)) quit(out_of_local_memory);

  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,start),1,CF_DOUBLE,"startUtc")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,stop),1,CF_DOUBLE,"stopUtc")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,counts),1,CF_INT32,"counts")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,reserved),1,CF_INT32,"reserved")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,startStr),64,CF_TEXT,"startStr")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,stopStr),64,CF_TEXT,"stopStr")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,fromState),64,CF_TEXT,"fromState")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,toState),64,CF_TEXT,"toState")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,usr),16,CF_TEXT,"user")) quit(out_of_local_memory);
  if (addFieldToStruct("StateCor",OFFSETIN(StateCorrection,cmt),256,CF_TEXT,"comment")) quit(out_of_local_memory);
  if (sealTaggedStruct("StateCor",sizeof(StateCorrection),MAX_CORRECTION_PER_DAY)) quit(out_of_local_memory);
}

typedef struct 
{
  char procname[16];
  char procfile1[32];
  char procfile2[32];
  char rule[16];
  int  nstates1;
  int  nstates2;
  NAME64 *statenames1;
  NAME64 *statenames2;
  time_t ltd;       /* last time of declaration */
  UINT32 ttd;       /* total time at declared state */
  UINT32 ttd_bc;    /* total time at declared state with beam conditions */
  UINT32 ttd_tc;    /* total time at declared state with test conditions */
  UINT32 ttd_ec;    /* total time at declared state with error conditions (unavailable) */
  UINT32 tnd;       /* total number of state declarations */
} MACHSTATE;
// machine states are relevant to multiple beam facilities (e.g. colliders) 
MACHSTATE *machStates = NULL;
int nmachstates = 0;

int gNumParticleTypes = 0;

#define BEAM1CRITERIAFILE "Beam1Criteria.csv"
#define BEAM2CRITERIAFILE "Beam2Criteria.csv"
#define TEST1CRITERIAFILE "Test1Criteria.csv"
#define TEST2CRITERIAFILE "Test2Criteria.csv"

#define STOCK_STRINGLENGTH 128
typedef struct
{
  int configured;
  int nparams;
  int valid;
  int linkstatus;
  char tag[32];
  NAME64 *dserver;
  NAME64 *dproperty;
  DTYPE *param;
  DTYPE *threshold;
} VALIDITYCHECK; 

#define BEAMTYPE 0
#define TESTTYPE 1
#define FINDMTYPE(i) (((i) & 0x000f) - 1)
#define FINDCTYPE(i) ((i) >> 8)
#define MAKETYPEID(c,m) (((c) << 8) | (m+1))

VALIDITYCHECK isValidConditions[2*NMACHINETYPES];
VALIDITYCHECK *isBeamConditions = isValidConditions;
VALIDITYCHECK *isTestConditions = &isValidConditions[NMACHINETYPES];

void checkValidity(VALIDITYCHECK *vc);

int readValidityFile(char *filename, VALIDITYCHECK *vc)
{
  FILE *fp;
  DTYPE *param, *threshold;
  NAME64 *dserver;
  NAME64 *dproperty;
  int cc=0,done=0,prp_col,srv_col,fmt_col,siz_col,thr_col;
  char str[256],hdr[256],scratch[STOCK_STRINGLENGTH],fn[128];
  int n=0,len,fmtsiz;

  vc->configured = FALSE;
  sprintf(fn,"%.96s%s",DBPATH,filename);
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  if ((dserver=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
    return out_of_server_memory;
  if ((dproperty=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
    return out_of_server_memory;
  if ((param=(DTYPE *)calloc(n,sizeof(DTYPE))) == NULL)
    return out_of_server_memory;
  if ((threshold=(DTYPE *)calloc(n,sizeof(DTYPE))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((srv_col=findcol(hdr,"SERVER")) < 0) ccerr(no_such_line);
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(hdr,"SIZE")) < 0) ccerr(no_such_line);
      if ((thr_col=findcol(hdr,"THRESHOLD")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dserver[n].name,colptr(srv_col,str),64);
    strncpy(dproperty[n].name,colptr(prp_col,str),32);
    strncpy(scratch,colptr(fmt_col,str),16);
    param[n].dFormat = ftoi(scratch);
    threshold[n].dFormat = param[n].dFormat;
    fmtsiz = fmtsizeof(param[n].dFormat);
    strncpy(scratch,colptr(siz_col,str),16);
    param[n].dArrayLength = atoi(scratch);
    len =threshold[n].dArrayLength = param[n].dArrayLength;
    if ((param[n].data.vptr=calloc(len,fmtsiz)) == NULL)
      ccerr(out_of_server_memory);
    strncpy(scratch,colptr(thr_col,str),STOCK_STRINGLENGTH);
    switch (threshold[n].dFormat)
    {   
      case CF_FLOAT:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.fptr[0] = (float)strtod(scratch,NULL);
        break;
      case CF_DOUBLE:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.dptr[0] = strtod(scratch,NULL);
        break;
      case CF_LONG:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.lptr[0] = atoi(scratch);
        break;
      case CF_SHORT:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.sptr[0] = (short)atoi(scratch);
        break;
      case CF_TEXT:
      case CF_NAME64:
      case CF_NAME48:
      case CF_NAME32:
      case CF_NAME16:
      case CF_NAME8:
        if ((threshold[n].data.vptr=calloc(STOCK_STRINGLENGTH,1)) == NULL)
          ccerr(out_of_server_memory);
        scratch[STOCK_STRINGLENGTH-1] = 0; strupr(scratch);
        strncpy(threshold[n].data.cptr,scratch,STOCK_STRINGLENGTH);
        break;
      default:
        ccerr(illegal_format);
    }
    n++;
  }
  vc->nparams = n;
  vc->dserver = dserver;
  vc->dproperty = dproperty;
  vc->param = param;
  vc->threshold = threshold;
err:
  fclose(fp);
  return cc;
}
int gIsAvailable = 1;
void cbAvailable(int id,int cc)
{
}
void vcallback(int id,int cc)
{
  VALIDITYCHECK *vc;
  int i;
  if (cc)
  {
    return;
  }
  for (i=0; i<2*NMACHINETYPES; i++) 
  {
    vc = &isValidConditions[i];
    if (vc->configured) checkValidity(vc);
  }
  return;
}
DTYPE *assignValidityParam(VALIDITYCHECK *vc,int id)
{
  int i,j;
  for (i=0; i<2*NMACHINETYPES; i++)
  {
    if (vc == &isValidConditions[i]) break;
    for (j=0; j<isValidConditions[i].nparams; j++)
    {
      if (strncmp(vc->dserver[id].name,isValidConditions[i].dserver[j].name,64))
        continue;
      if (strncmp(vc->dproperty[id].name,isValidConditions[i].dproperty[j].name,32))
        continue;
      if (vc->param[id].dFormat != isValidConditions[i].param[j].dFormat) continue;
      if (vc->param[id].dArrayLength != isValidConditions[i].param[j].dArrayLength) continue;

      return &isValidConditions[i].param[j];
    }
  }
  return NULL;
}
void startAvailabilityLink(void)
{
  int id;
  char dev[128],str[256];
  DTYPE dout;

  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  dout.data.lptr = &gIsAvailable;
  sprintf(dev,"/%.32s/ALARMSTATE/ALL",gDeviceContext);
  id = AttachLink(dev,"ISREADY",&dout,NULL,CA_READ,1000,cbAvailable,CM_DATACHANGE);
  if (id < 0)
  {
    sprintf(str,"Cannot establish availability link to %.128s : %.32s",dev,cc2str(-id));
    feclogEx(TINE_LOGLEVEL_ERROR,str);
    logproblem(str);
  }

}
void startValidityLinks(VALIDITYCHECK *vc,int check_id)
{
  int i,id;
  char str[256];
  DTYPE *param;
  
  vc->configured = TRUE;
  vc->linkstatus = 0; /* hope for the best */
  for (i=0; i<vc->nparams; i++)
  {
    if ((param=assignValidityParam(vc,i)) != NULL)
    {
      vc->param[i] = *param;
      sprintf(str,"%.32s : multiple state criterion link to %.64s %.64s\n",
              vc->tag,vc->dserver[i].name,vc->dproperty[i].name);
      feclog(str);
      dbglog(str);
    }
    else
    {
      id = AttachLinkEx(vc->dserver[i].name,vc->dproperty[i].name,&vc->param[i],NULL,CA_READ,1000,vcallback,CM_POLL,check_id);
      if (id < 0)
      {
        sprintf(str,"%.32s : cannot establish state criterion link to %.64s %.64s : %.32s",
                vc->tag,vc->dserver[i].name,vc->dproperty[i].name,cc2str(-id));
        feclogEx(TINE_LOGLEVEL_ERROR,str);
        logproblem(str);
        vc->linkstatus = -id;
      }
    }
  }
}
char *trimString(char *str)
{
  static char trimBuffer[256];
  char *pstr;
  if (str == NULL)
  {
    trimBuffer[0] = 0;
    goto err;
  }
  for (pstr=str; strchr(" \t\n",*pstr) != NULL; pstr++);
  strncpy(trimBuffer,str,256);
  strtok(trimBuffer," \t\n");
err:
  return trimBuffer;
}
int dumpValidity(int a1,int a2,int a3,int a4)
{
  int i, k;
  VALIDITYCHECK *vc;
  a1 = a1; a2 = a2; a3 = a3; a4 = a4;
  for (i=0; i<2*NMACHINETYPES; i++) 
  {
    vc = &isValidConditions[i];
    for (k=0; k<vc->nparams; k++)
    {
      ttyoutput("%s (%s[%s]) : %s %s (link status %d)",
        vc->tag,vc->dserver[k].name,vc->dproperty[k].name,
        vc->configured ? "configured" : "not configured",
        vc->valid ? "valid" : "not valid",
        vc->linkstatus);
    }
  }
  return 0;
}
void checkValidity(VALIDITYCHECK *vc)
{
  int isvalid = TRUE;
  int i, slen;
  char tmpstr[STOCK_STRINGLENGTH],scratch[128];

  for (i=0; isvalid && i<vc->nparams; i++)
  {
    switch (vc->param[i].dFormat)
    {
      case CF_FLOAT: 
        if (vc->param[i].data.fptr[0] <
            vc->threshold[i].data.fptr[0]) isvalid = FALSE;
        break;
      case CF_DOUBLE: 
        if (vc->param[i].data.dptr[0] <
            vc->threshold[i].data.dptr[0]) isvalid = FALSE;
        break;
      case CF_LONG: 
        if (vc->param[i].data.lptr[0] <
            vc->threshold[i].data.lptr[0]) isvalid = FALSE;
        break;
      case CF_SHORT: 
        if (vc->param[i].data.sptr[0] <
            vc->threshold[i].data.sptr[0]) isvalid = FALSE;
        break;
      case CF_TEXT:
      case CF_NAME64:
      case CF_NAME48:
      case CF_NAME32:
      case CF_NAME16:
      case CF_NAME8:
        slen = vc->param[i].dFormat != CF_TEXT ? fmtsizeof(vc->param[i].dFormat) : vc->param[i].dArrayLength;
        strncpy(scratch,vc->param[i].data.cptr,slen);
        scratch[slen] = 0;
        strupr(scratch);
        strcpy(tmpstr,trimString(scratch));
        if (strlen(tmpstr) == 0 || 
            strstr(vc->threshold[i].data.cptr,tmpstr) == NULL) isvalid = FALSE;
        break;
      default:
        isvalid = FALSE;
    }
  }
  vc->valid = isvalid;
  if (tineDebug > 1) ttyoutput("%s conditions are %s",vc->tag,isvalid ? "valid" : "not valid");
}

int readMachStructureFile(char *fn,NAME64 **nl,int *nn)
{
  FILE *fp;
  int cc=0,done=0,sta_col;
  char str[256],hdr[256];
  int n=0;
  NAME64 *nms = NULL;
 
  if (nl == NULL || nn == NULL) return un_allocated;
  if ((n=GetDataBaseSize(fn)) < 0) return -n;
  if (n == 0) return database_not_loaded;
  if ((nms=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
    return out_of_server_memory;
 
  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((sta_col=findcol(hdr,"STATE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(nms[n].name,colptr(sta_col,str),64);
    n++;
  }
err:
  fclose(fp);
  *nl = nms;
  *nn = n;
  return cc;
}
int readProcStructureFile(char *filename,DECLPROC *dp)
{
  FILE *fp;
  int cc=0,done=0,sta_col;
  char str[256],hdr[256],fn[128];
  int n=0;
  NAME64 *nms = NULL;

  if (dp == NULL) return un_allocated;
  sprintf(fn,"%.96s%s",DBPATH,filename);
  if ((n=GetDataBaseSize(fn)) < 0) return -n;
  if (n == 0) 
  {
    dp->nstates = 0;
    dp->statenames = NULL;
    return 0;
  }
  if ((nms=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((sta_col=findcol(hdr,"STATE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(nms[n].name,colptr(sta_col,str),64);
    n++;
  }
err:
  fclose(fp);
  if (dp ->statenames != NULL) free(dp->statenames);
  dp->statenames = nms;
  dp->nstates = n;
  return cc;
}
int writeProcStructureFile(char *filename,DECLPROC *dp)
{
  FILE *fp;
  int cc=0;
  char fn[128],fnb[256];
  int n=0;
  static time_t backup_done = 0;
  time_t t = time(NULL);

  if (dp == NULL) return un_allocated;
  sprintf(fn,"%.96s%s",DBPATH,filename);
  if (t > backup_done + 86400)
  {
    sprintf(fnb,"%.128s.bak",fn);
    remove(fnb); rename(fn,fnb);
    backup_done = t;
  }
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"STATE\n");
  for (n=0; n<dp->nstates; n++)
  {
    fprintf(fp,"%.64s\n",dp->statenames[n].name);
  }
  fclose(fp);
  return 0;
}

DECLPROC defaultProcs[] = { 
  {"Pnotrunning", "Pnotrunning.csv", 0, 2, 0, 0, NULL },
  {"Pservice", "Pservice.csv", 0, 1, 0, 0, NULL },
  {"Pproblems", "Pproblems.csv", 0, 1, 0, 0, NULL },
  {"Prunning", "Prunning.csv", 0, 3, 0, 0, NULL },
};
int setDefaultDeclProcs(void)
{
  int i;

  ndeclprocs = sizeof(defaultProcs)/sizeof(DECLPROC);
  declprocsCapacity = ndeclprocs + INFO_CUSHION;
  if ((declProcs == NULL) && 
      (declProcs=(DECLPROC *)calloc(declprocsCapacity,sizeof(DECLPROC))) == NULL)
    return out_of_server_memory;
  for (i=0; i<ndeclprocs; i++)
  {
    memcpy(&declProcs[i],&defaultProcs[i],sizeof(DECLPROC));
    declProcs[i].statenames = (NAME64 *)calloc(declProcs[i].nstates,sizeof(NAME64));
  }
  if (declProcs[0].statenames != NULL)
  { /* not running */
    strncpy(declProcs[0].statenames[0].name,"undefined",DEVICE_NAME_SIZE);
    strncpy(declProcs[0].statenames[1].name,"access",DEVICE_NAME_SIZE);
  }
  if (declProcs[1].statenames != NULL)
  { /* service */
    strncpy(declProcs[1].statenames[0].name,"service",DEVICE_NAME_SIZE);
  }
  if (declProcs[2].statenames != NULL)
  { /* problems */
    strncpy(declProcs[2].statenames[0].name,"problems",DEVICE_NAME_SIZE);
  }
  if (declProcs[3].statenames != NULL)
  { /* running */
    strncpy(declProcs[3].statenames[0].name,"running",DEVICE_NAME_SIZE);
    strncpy(declProcs[3].statenames[1].name,"standby",DEVICE_NAME_SIZE);
    strncpy(declProcs[3].statenames[2].name,"studies",DEVICE_NAME_SIZE);
  }
  return 0;
}
int readProcInfoFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,pro_col,mac_col,tnd_col,ltd_col;
  char str[256],hdr[256],scratch[64];
  int i,n=0,na=0,renameit=0;
  char fn[64],fna[64];

  sprintf(fn,"%s.csv",PROCINFOFILE);
  if ((n=GetDataBaseSize(fn)) <= 0) return n;
  if (n == 0) return database_not_loaded;
  sprintf(fna,"%s.add",PROCINFOFILE);
  if ((na=GetDataBaseSize(fna)) > 0) n += na;
  n += INFO_CUSHION;
  if (n > declprocsCapacity && declProcs != NULL) 
  { 
    free(declProcs); declProcs = NULL; 
  }
  if ((declProcs == NULL) && 
      (declProcs=(DECLPROC *)calloc(n,sizeof(DECLPROC))) == NULL)
    return out_of_server_memory;
  declprocsCapacity = n;
  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((pro_col=findcol(hdr,"PROCEDURE")) < 0) ccerr(no_such_line);
      if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
      if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
      if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(declProcs[n].devname,colptr(pro_col,str),64);
    sprintf(declProcs[n].structfile,"%.60s.csv",declProcs[n].devname);
    if ((cc=readProcStructureFile(declProcs[n].structfile,&declProcs[n])) != 0) 
    {
      declProcs[n].nstates = 0;
      declProcs[n].statenames = NULL;
      feclogEx(TINE_LOGLEVEL_ERROR,"Cannot read proc info file %s : %s",declProcs[n].structfile,erlst[cc]);
    }
    strncpy(scratch,colptr(mac_col,str),16);
    declProcs[n].machine = atoi(scratch);
    strncpy(scratch,colptr(ltd_col,str),16);
    declProcs[n].ltd = atoi(scratch);
    strncpy(scratch,colptr(tnd_col,str),16);
    declProcs[n].tnd = atoi(scratch);
    n++;
  }
  if (na > 0)  /* there's an .add file (new states!) */
  {
    fclose(fp); done = 0;
    if ((fp=fopen(fna,"r")) == NULL) ccerr(file_error);
    while ( fgets(str,255,fp) != NULL)
    {
      if (strchr("%;#\n",str[0])) continue;
      if (!done)
      {
        strcpy(hdr,str);
        if ((pro_col=findcol(hdr,"PROCEDURE")) < 0) ccerr(no_such_line);
        if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
        if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
        if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(pro_col,str),64);
      for (i=0; i<n && strncmp(scratch,declProcs[i].devname,64); i++);
      strncpy(declProcs[i].devname,scratch,64);
      sprintf(declProcs[i].structfile,"%.60s.csv",declProcs[i].devname);
      if ((cc=readProcStructureFile(declProcs[i].structfile,&declProcs[i])) != 0)
        continue;
      strncpy(scratch,colptr(mac_col,str),16);
      declProcs[i].machine = atoi(scratch);
      strncpy(scratch,colptr(ltd_col,str),16);
      declProcs[i].ltd = atoi(scratch);
      strncpy(scratch,colptr(tnd_col,str),16);
      declProcs[i].tnd = atoi(scratch);
      if (i == n) n++;
    }
    renameit = TRUE;
  }
err:
  if (fp != NULL) fclose(fp);
  ndeclprocs = n;  
  if (na > 0 && renameit)
  {
    sprintf(fn,"%s.added",PROCINFOFILE);
    rename(fna,fn);
  }
  return cc;
}

int writeProcInfoToFile(void)
{
  FILE *fp;
  int i;
  char fn[64];

  sprintf(fn,"%s.csv",PROCINFOFILE);
  if ((fp=fopen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"Procedure,Machine,LastTime,TotalNumber\n");
  for (i=0; i<ndeclprocs; i++)
  {
    fprintf(fp,"%s,%d,%d,%d\n",
            declProcs[i].devname,declProcs[i].machine,(int)declProcs[i].ltd,declProcs[i].tnd);
  }
  fclose(fp);
  return 0;
}
int writeStatesToFile(void)
{
  FILE *fp;

  if ((fp=fopen(STATEFILE,"w")) == NULL) return file_error;
  if (gNumParticleTypes > 1)
  {
    fprintf(fp,"Particle1State,Particle1Code,Particle2State,Particle2Code\n");
    fprintf(fp,"%s,%d,%s,%d\n",
            declStates[declaredState[0]].devname,declaredState[0],
            declStates[declaredState[1]].devname,declaredState[1]);
  }
  else
  {
    fprintf(fp,"Particle1State,Particle1Code\n");
    fprintf(fp,"%s,%d\n",declStates[declaredState[0]].devname,declaredState[0]);
  }
  fclose(fp);
  return 0;
}
int getNumParticleTypesFromFile(void)
{
  FILE *fp;
  int n = 0;
  char str[256];

  if ((fp=fopen(STATEFILE,"r")) == NULL) return 1;
  while (fgets(str,255,fp) != NULL)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (findcol(str,"PARTICLE1CODE") < 0) break;
    if (findcol(str,"PARTICLE2CODE") < 0) { n = 1; break; }
    n = 2; break;
  }
  if (fp != NULL) fclose(fp);
  return n;
}
int readStatesFromFile(void)
{
  FILE *fp=NULL;
  int cc=0,done=0,p1_col,p2_col;
  int i,j, k;
  char str[256],hdr[256],scratch[16];

  if ((fp=fopen(STATEFILE,"r")) == NULL) 
  {
    gNumParticleTypes = 1;
    declaredState[0] = 0;
    declaredProc[0] = 0;
    return file_error;
  }
  while (fgets(str,255,fp) != NULL)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((p1_col=findcol(hdr,"PARTICLE1CODE")) < 0) ccerr(no_such_line);
      p2_col = findcol(hdr,"PARTICLE2CODE");
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(p1_col,str),16);
    declaredState[0] = atoi(scratch);
    gNumParticleTypes = 1;
    if (p2_col >= 0)
    {
      strncpy(scratch,colptr(p2_col,str),16);
      declaredState[1] = atoi(scratch);
      gNumParticleTypes = 2;
    }
    break;
  }
  for (i=0; i<ndeclprocs; i++) for (j=0; j<declProcs[i].nstates; j++) for (k=0; k<gNumParticleTypes; k++)
  {
    if (!strncmp(declStates[declaredState[k]].devname,declProcs[i].statenames[j].name,64)) 
    {
      declaredProc[k] = i;
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int writeStateInfoToFile(void)
{
  FILE *fp;
  int i;
  char fn[64];

  if (ndeclstates <= 0) return database_not_loaded;
  sprintf(fn,"%s.csv",STATEINFOFILE);
  if ((fp=fopen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"DeviceName,StateTag,Machine,LastTime,TotalTime,TotalTime_bc,TotalTime_tc,TotalTime_ec,TotalNumber\n");
  for (i=0; i<ndeclstates; i++)
  {
    fprintf(fp,"%s,%s,%d,%d,%d,%d,%d,%d,%d\n",
            declStates[i].devname,declStates[i].statetag,
            declStates[i].machine,(int)declStates[i].ltd,
            declStates[i].ttd,declStates[i].ttd_bc,
            declStates[i].ttd_tc,declStates[i].ttd_ec,
            declStates[i].tnd);
  }
  fclose(fp);
  return 0;
}

DECLSTATE defaultStates[] = { 
  {"undefined", "undefined", 1, 0, 0, 0, 0, 0, 0 },
  {"standby", "standby", 1, 0, 0, 0, 0, 0, 0 },
  {"service", "is being serviced", 1, 0, 0, 0, 0, 0, 0 },
  {"problems", "technical problems", 1, 0, 0, 0, 0, 0, 0 },
  {"access", "machine access", 1, 0, 0, 0, 0, 0, 0 },
  {"studies", "machine studies", 1, 0, 0, 0, 0, 0, 0 },
  {"running", "running", 1, 0, 0, 0, 0, 0, 0 },
};
int setDefaultDeclStates(void)
{
  int i;

  ndeclstates = sizeof(defaultStates)/sizeof(DECLSTATE);
  declstatesCapacity = ndeclstates + INFO_CUSHION;
  if ((declStates == NULL) && 
      (declStates=(DECLSTATE *)calloc(declstatesCapacity,sizeof(DECLSTATE))) == NULL)
    return out_of_server_memory;
  for (i=0; i<ndeclstates; i++)
  {
    memcpy(&declStates[i],&defaultStates[i],sizeof(DECLSTATE));
  }
  return 0;
}

CSVCOLUMN csvStateEventTriggersFileColumns[] =
{
  { "STATE","",0,CF_TEXT,64,NULL,NULL,0 }
};
CSVDB csvStateEventTriggersFileDb =
{
  csvStateEventTriggersFileColumns,
  sizeof(csvStateEventTriggersFileColumns)/sizeof(csvStateEventTriggersFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME64),
  0
};
#define EVENT_TRIGGERS_FILE "events.csv"
int readStateTriggerInfoFromFile(void)
{
  int cc = 0;
  NAME64 *evns = NULL;
  if ((cc=csvReadFile(NULL,EVENT_TRIGGERS_FILE,&csvStateEventTriggersFileDb,(void **)&evns)) == 0)
  {
    int idx, i, n = csvStateEventTriggersFileDb.siz;
    for (i=0; i<n; i++)
    {
      if ((idx=getDeclStateIndex(evns[i].name)) < 0) continue;
      declStates[idx].evntrg = TRUE;
      feclog("EVENT Trigger when %.64s declared",evns[i].name);
    }
  }
  if (cc == 0) feclog("Event Triggers file found!");
  return cc;
}
int readStateInfoFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,dev_col,sta_col,mac_col,ltd_col,ttd_col,tnd_col;
  int ttdbc_col=-1,ttdtc_col=-1,ttdec_col=-1;
  char str[256],hdr[256],scratch[64];
  int i,n=0,na=0,renameit=0;
  char fn[64],fna[64];

  sprintf(fn,"%s.csv",STATEINFOFILE);
  if ((n=GetDataBaseSize(fn)) <= 0) return n;
  if (n == 0) return database_not_loaded;
  sprintf(fna,"%s.add",STATEINFOFILE);
  if ((na=GetDataBaseSize(fna)) > 0) n += na;
  n += INFO_CUSHION;
  if (n > declstatesCapacity && declStates != NULL) 
  {
    free(declStates); declStates = NULL;
  }
  if ((declStates == NULL) && 
      (declStates=(DECLSTATE *)calloc(n,sizeof(DECLSTATE))) == NULL)
    return out_of_server_memory;
  declstatesCapacity = n;
  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((dev_col=findcol(hdr,"DEVICENAME")) < 0) ccerr(no_such_line);
      if ((sta_col=findcol(hdr,"STATETAG")) < 0) ccerr(no_such_line);
      if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
      if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
      if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
      ttdbc_col = findcol(hdr,"TOTALTIME_BC");
      ttdtc_col = findcol(hdr,"TOTALTIME_TC");
      ttdec_col = findcol(hdr,"TOTALTIME_EC");
      if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(declStates[n].devname,colptr(dev_col,str),64);
    strncpy(declStates[n].statetag,colptr(sta_col,str),64);
    strncpy(scratch,colptr(mac_col,str),16);
    declStates[n].machine = atoi(scratch);
    strncpy(scratch,colptr(ltd_col,str),16);
    declStates[n].ltd = atoi(scratch);
    strncpy(scratch,colptr(ttd_col,str),16);
    declStates[n].ttd = atoi(scratch);
    if (ttdbc_col >= 0)
    {
      strncpy(scratch,colptr(ttdbc_col,str),16);
      declStates[n].ttd_bc = atoi(scratch);
    }
    if (ttdtc_col >= 0)
    {
      strncpy(scratch,colptr(ttdtc_col,str),16);
      declStates[n].ttd_tc = atoi(scratch);
    }
    if (ttdec_col >= 0)
    {
      strncpy(scratch,colptr(ttdec_col,str),16);
      declStates[n].ttd_ec = atoi(scratch);
    }
    strncpy(scratch,colptr(tnd_col,str),16);
    declStates[n].tnd = atoi(scratch);
    n++;
  }
  if (na > 0)  /* there's an .add file (new states!) */
  {
    fclose(fp); done = 0;
    if ((fp=fopen(fna,"r")) == NULL) ccerr(file_error);
    while ( fgets(str,255,fp) != NULL)
    {
      if (strchr("%;#\n",str[0])) continue;
      if (!done)
      {
        strcpy(hdr,str);
        if ((dev_col=findcol(hdr,"DEVICENAME")) < 0) ccerr(no_such_line);
        if ((sta_col=findcol(hdr,"STATETAG")) < 0) ccerr(no_such_line);
        if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
        if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
        if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
        ttdbc_col = findcol(hdr,"TOTALTIME_BC");
        ttdtc_col = findcol(hdr,"TOTALTIME_TC");
        ttdec_col = findcol(hdr,"TOTALTIME_EC");
        if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(dev_col,str),64);
      for (i=0; i<n && strncmp(scratch,declStates[i].devname,64); i++);
      strncpy(declStates[i].devname,scratch,64);
      strncpy(declStates[i].statetag,colptr(sta_col,str),64);
      strncpy(scratch,colptr(mac_col,str),16);
      declStates[i].machine = atoi(scratch);
      strncpy(scratch,colptr(ltd_col,str),16);
      declStates[i].ltd = atoi(scratch);
      strncpy(scratch,colptr(ttd_col,str),16);
      declStates[i].ttd = atoi(scratch);
      if (ttdbc_col >= 0)
      {
        strncpy(scratch,colptr(ttdbc_col,str),16);
        declStates[i].ttd_bc = atoi(scratch);
      }
      if (ttdtc_col >= 0)
      {
        strncpy(scratch,colptr(ttdtc_col,str),16);
        declStates[i].ttd_tc = atoi(scratch);
      }
      if (ttdec_col >= 0)
      {
        strncpy(scratch,colptr(ttdec_col,str),16);
        declStates[i].ttd_ec = atoi(scratch);
      }
      strncpy(scratch,colptr(tnd_col,str),16);
      declStates[i].tnd = atoi(scratch);
      if (i == n) n++;
    }
    renameit = TRUE;
  }
err:
  if (fp != NULL) fclose(fp);
  ndeclstates = n;  
  if (na > 0 && renameit)
  {
    sprintf(fn,"%s.added",STATEINFOFILE);
    rename(fna,fn);
  }
  return cc;
}
int writeMachineInfoToFile(void)
{
  FILE *fp;
  int i;
  char fn[64];

  sprintf(fn,"%s.csv",MACHINFOFILE);
  if ((fp=fopen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"MachineProcedure,ProcedureFile1,ProcedureFile2,Rule,LastTime,TotalTime,TotalTime_bc,TotalTime_tc,TotalTime_ec,TotalNumber\n");
  for (i=0; i<nmachstates; i++)
  {
    fprintf(fp,"%s,%s,%s,%s,%d,%d,%d,%d,%d,%d\n",
            machStates[i].procname,machStates[i].procfile1,
            machStates[i].procfile2,machStates[i].rule,
            (int)declStates[i].ltd,
            machStates[i].ttd,machStates[i].ttd_bc,
            machStates[i].ttd_tc,machStates[i].ttd_ec,
            machStates[i].tnd);
  }
  fclose(fp);
  return 0;
}
int readMachineInfoFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,proc_col,file1_col,file2_col,rule_col,ltd_col,ttd_col,tnd_col;
  int ttdbc_col=-1,ttdtc_col=-1,ttdec_col=-1;
  char str[256],hdr[256],scratch[64];
  int i,n=0,na=0,renameit=0;
  char fn[64],fna[64];

  sprintf(fn,"%s.csv",MACHINFOFILE);
  if ((n=GetDataBaseSize(fn)) <= 0) return n;
  if (n == 0) return database_not_loaded;
  sprintf(fna,"%s.add",MACHINFOFILE);
  if ((na=GetDataBaseSize(fna)) > 0) n += na;
  if (n > nmachstates && machStates != NULL) free(machStates);
  if ((machStates == NULL) && 
      (machStates=(MACHSTATE *)calloc(n,sizeof(MACHSTATE))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((proc_col=findcol(hdr,"MACHINEPROCEDURE")) < 0) ccerr(no_such_line);
      if ((file1_col=findcol(hdr,"PROCEDUREFILE1")) < 0) ccerr(no_such_line);
      if ((file2_col=findcol(hdr,"PROCEDUREFILE2")) < 0) ccerr(no_such_line);
      if ((rule_col=findcol(hdr,"RULE")) < 0) ccerr(no_such_line);
      if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
      if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
      ttdbc_col = findcol(hdr,"TOTALTIME_BC");
      ttdtc_col = findcol(hdr,"TOTALTIME_TC");
      if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(machStates[n].procname,colptr(proc_col,str),16);
    strncpy(machStates[n].procfile1,colptr(file1_col,str),32);
    if ((cc=readMachStructureFile(machStates[n].procfile1,&machStates[n].statenames1,&machStates[n].nstates1)) != 0) ccerr(cc);
    strncpy(machStates[n].procfile2,colptr(file2_col,str),32);
    if ((cc=readMachStructureFile(machStates[n].procfile2,&machStates[n].statenames2,&machStates[n].nstates2)) != 0) ccerr(cc);
    strncpy(machStates[n].rule,colptr(rule_col,str),16);
    strncpy(scratch,colptr(ltd_col,str),16);
    machStates[n].ltd = atoi(scratch);
    strncpy(scratch,colptr(ttd_col,str),16);
    machStates[n].ttd = atoi(scratch);
    if (ttdbc_col >= 0)
    {
      strncpy(scratch,colptr(ttdbc_col,str),16);
      machStates[n].ttd_bc = atoi(scratch);
    }
    if (ttdtc_col >= 0)
    {
      strncpy(scratch,colptr(ttdtc_col,str),16);
      machStates[n].ttd_tc = atoi(scratch);
    }
    if (ttdec_col >= 0)
    {
      strncpy(scratch,colptr(ttdec_col,str),16);
      machStates[n].ttd_ec = atoi(scratch);
    }
    strncpy(scratch,colptr(tnd_col,str),16);
    machStates[n].tnd = atoi(scratch);
    n++;
  }
  if (na > 0)  /* there's an .add file (new states!) */
  {
    fclose(fp); done = 0;
    if ((fp=fopen(fna,"r")) == NULL) ccerr(file_error);
    while ( fgets(str,255,fp) != NULL)
    {
      if (strchr("%;#\n",str[0])) continue;
      if (!done)
      {
        strcpy(hdr,str);
        if ((proc_col=findcol(hdr,"MACHINEPROCEDURE")) < 0) ccerr(no_such_line);
        if ((file1_col=findcol(hdr,"PROCEDUREFILE1")) < 0) ccerr(no_such_line);
        if ((file2_col=findcol(hdr,"PROCEDUREFILE2")) < 0) ccerr(no_such_line);
        if ((rule_col=findcol(hdr,"RULE")) < 0) ccerr(no_such_line);
        if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
        if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
        ttdbc_col = findcol(hdr,"TOTALTIME_BC");
        ttdtc_col = findcol(hdr,"TOTALTIME_TC");
        ttdec_col = findcol(hdr,"TOTALTIME_EC");
        if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }

      strncpy(scratch,colptr(proc_col,str),16);
      for (i=0; i<n && strncmp(scratch,machStates[i].procname,16); i++);
      strncpy(machStates[i].procname,colptr(proc_col,str),16);
      strncpy(machStates[i].procfile1,colptr(file1_col,str),32);
      strncpy(machStates[i].procfile2,colptr(file2_col,str),32);
      strncpy(machStates[i].rule,colptr(rule_col,str),16);
      strncpy(scratch,colptr(ltd_col,str),16);
      machStates[i].ltd = atoi(scratch);
      strncpy(scratch,colptr(ttd_col,str),16);
      machStates[i].ttd = atoi(scratch);
      if (ttdbc_col >= 0)
      {
        strncpy(scratch,colptr(ttdbc_col,str),16);
        machStates[i].ttd_bc = atoi(scratch);
      }
      if (ttdtc_col >= 0)
      {
        strncpy(scratch,colptr(ttdtc_col,str),16);
        machStates[i].ttd_tc = atoi(scratch);
      }
      if (ttdec_col >= 0)
      {
        strncpy(scratch,colptr(ttdec_col,str),16);
        machStates[i].ttd_ec = atoi(scratch);
      }
      strncpy(scratch,colptr(tnd_col,str),16);
      machStates[i].tnd = atoi(scratch);
      if (i == n) n++;
    }
    renameit = TRUE;
  }
err:
  if (fp != NULL) fclose(fp);
  nmachstates = n;  
  if (na > 0 && renameit)
  {
    sprintf(fn,"%s.added",MACHINFOFILE);
    rename(fna,fn);
  }
  return cc;
}
#define ONEHOUR 3600
#define ONEDAY (ONEHOUR * 24)
#define ONEWEEK (ONEDAY * 7)
#define ONEMONTH (ONEDAY *30)
#define ext_1day "1day"
#define ext_1week "1week"
int dbbackup(char *fn)
{
  struct stat fs, fsb;
  char fnb1[80],fnb2[80];
  int dayfile_exists = FALSE;
  time_t tnow = time(NULL);

  if (stat(fn,&fs) != 0) return no_such_file;
  if (fs.st_size > 0)
  { /* copy only if non-zero length */
    sprintf(fnb1,"%.72s.%s",fn,ext_1day);
    if (stat(fnb1,&fsb) == 0)
    { /* if there: needs to be at least 1 day old */
      dayfile_exists = TRUE;
      if (tnow < fsb.st_mtime + ONEDAY) return 0;
    }
    sprintf(fnb2,"%.72s.%s",fn,ext_1week);
    if (stat(fnb2,&fsb) == 0)
    { /* check this one too !*/
      if (tnow > fsb.st_mtime + ONEWEEK)
      {
        unlink(fnb2); rename(fnb1,fnb2);
      }
    }
    else if (dayfile_exists)
    { /* initial file */
      rename(fnb1,fnb2);
    }
    unlink(fnb1); rename(fn,fnb1);
  }
  return 0;
}
time_t declStateTime = 0;
char gStateChangeTriggerName[DEVICE_NAME_SIZE] = "";
int gStateChangeTrigger = -1;
int gLastStateChangeTigger = -1;
double gStateChangeTriggerTime = 0;

int handleStateChangeTrigger(void)
{
  int i, j, type, newproc = FALSE;
  time_t tt = (time_t)gStateChangeTriggerTime; // time(NULL);
  if (gStateChangeTrigger < 0 || gStateChangeTrigger >= ndeclstates) return invalid_index;
  declStateTime = tt; // post-fix the correct declared time
  type = declStates[gStateChangeTrigger].machine;
  if (type & PTYPE1)
  {
    if (declaredState[0] == (UINT32)gStateChangeTrigger) return 0; /* return already_assigned; */
    prevDeclaredState[0] = declaredState[0];
    declaredState[0] = gStateChangeTrigger;
    gScheduleDeclState = TRUE;
  }
  if (type & PTYPE2)
  {
    if (declaredState[1] == (UINT32)gStateChangeTrigger) return 0; /* return already_assigned; */
    prevDeclaredState[1] = declaredState[1];
    declaredState[1] = gStateChangeTrigger;
    gScheduleDeclState = TRUE;
  }
  declStates[gStateChangeTrigger].ltd = declStateTime;
  declStates[gStateChangeTrigger].tnd++;
  /* use it to trigger archives here ... */
  writeStatesToFile();
  writeStateInfoToFile();
  if (declStates[gStateChangeTrigger].evntrg)
  { /* send event trigger when state is declared */
    sndTriggerToArchive(declStates[gStateChangeTrigger].devname);
  }
  for (i = 0; i < ndeclprocs; i++) for (j = 0; j < declProcs[i].nstates; j++)
  {
    if (strncmp(gStateChangeTriggerName, declProcs[i].statenames[j].name, DEVICE_NAME_SIZE)) continue;
    if (type & PTYPE1)
    {
      if ((int)declaredProc[0] != i)
      {
        prevDeclaredProc[0] = declaredProc[0];
        declaredProc[0] = i;
        newproc = TRUE;
      }
    }
    if (type & PTYPE2)
    {
      if ((int)declaredProc[1] != i)
      {
        prevDeclaredProc[1] = declaredProc[1];
        declaredProc[1] = i;
        newproc = TRUE;
      }
    }
    if (newproc)
    {
      declProcs[i].ltd = declStateTime;
      declProcs[i].tnd++;
      writeProcInfoToFile();
    }
    feclog("State change %.64s processed", declStates[gStateChangeTrigger].statetag);
    return 0;
  }
  return 0;
}
void staexi(void)
{
  writeStateInfoToFile();
  writeProcInfoToFile();
  writeMachineInfoToFile();
}
#define STATE_DURATION_VALID 1.0
void stabkg(void)
{
  int i,j,k,deadband;
  int match1, match2;
  static time_t lasttime = 0;
  double this_t = MakeDataTimeStamp();
  time_t t;
  
  if (gStateChangeTrigger >= 0 && this_t >= gStateChangeTriggerTime + STATE_DURATION_VALID)
  { /* at least 1000 msec at this trigger: accept this as a state change trigger */
    handleStateChangeTrigger();
    gLastStateChangeTigger = gStateChangeTrigger;
    gStateChangeTrigger = -1;
    gStateChangeTriggerName[0] = 0;
  }
  t = (time_t)this_t;
  if (gScheduleDeclState)
  {
    gScheduleDeclState = FALSE;
    SystemScheduleProperty(STAEQM_TAG,"DECLSTATE|DeclaredState");
  }

  deadband = (int)(lasttime > 0 ? (t - lasttime) : 1);
  lasttime = t;
  
  if (deadband == 0) return;
  /* below at 1 Hz */
  ClearAlarm(STAEQM_TAG, 0);
  i = declaredState[0];
  if (declStates == NULL) return;
  declStates[i].ttd += deadband;
  if (isBeamConditions[0].valid) declStates[i].ttd_bc += deadband;
  if (isTestConditions[0].valid) declStates[i].ttd_tc += deadband;
  if (!gIsAvailable) declStates[i].ttd_ec += deadband;
  if ((int)declaredState[1] == i) goto out; /* don't double count */

  if (gNumParticleTypes < 2) goto out; /* single particle machine */

  i = declaredState[1];
  declStates[i].ttd += deadband;
  if (isBeamConditions[1].valid) declStates[i].ttd_bc += deadband;
  if (isTestConditions[1].valid) declStates[i].ttd_tc += deadband;
  if (!gIsAvailable) declStates[i].ttd_ec += deadband;
out:  
  for (i=0; i<nmachstates; i++)
  {
    match1 = FALSE; match2 = FALSE;
    /* check if file1 matches */
    for (j=0; j<machStates[i].nstates1; j++)
    {
      k = declaredState[0];
      if (!strncmp(declStates[k].devname,machStates[i].statenames1[j].name,64)) break;
    }
    if (j < machStates[i].nstates1) match1 = TRUE;

    /* check if file2 matches */
    for (j=0; j<machStates[i].nstates2; j++)
    {
      k = declaredState[1];
      if (!strncmp(declStates[k].devname,machStates[i].statenames2[j].name,64)) break;
    }
    if (j < machStates[i].nstates2) match2 = TRUE;

    /* check Rule (OR or AND) */
    if (!match1 && !match2) continue;
    if (!strncmp(machStates[i].rule,"OR",2))
    { /* count */
      machStates[i].ttd += deadband;
      if (isBeamConditions[1].valid) machStates[i].ttd_bc += deadband;
      if (isTestConditions[1].valid) machStates[i].ttd_tc += deadband;
      if (!gIsAvailable) machStates[i].ttd_ec += deadband;
      continue;
    }
    if (!match1 || !match2) continue;
    if (!strncmp(machStates[i].rule,"AND",3))
    { /* count */
      machStates[i].ttd += deadband;
      if (isBeamConditions[1].valid) machStates[i].ttd_bc += deadband;
      if (isTestConditions[1].valid) machStates[i].ttd_tc += deadband;
      if (!gIsAvailable) machStates[i].ttd_ec += deadband;
      continue;
    }
  }
  if ((t%60) == 0) 
  {
    char fn[128];
    sprintf(fn,"%s.csv",STATEINFOFILE); dbbackup(fn);
    sprintf(fn,"%s.csv",PROCINFOFILE);  dbbackup(fn);
    sprintf(fn,"%s.csv",MACHINFOFILE);  dbbackup(fn);
    writeStateInfoToFile();
    writeProcInfoToFile();
    writeMachineInfoToFile();
  }
  return;  
}

int getDeclaredStateInRange(char *state, time_t start, time_t stop, time_t *nexttime, char *nextstate)
{
  NAME64DBL states[100];
  int i, npts;
  HstRngType rng;
  HstReqType req;

  req.eqm = STAEQM_TAG;
  req.prp = "DeclaredState";
  req.dev = "#0";
  req.index = req.fieldindex = req.sample = 0;
  rng.starttime = (double)(start-900);
  rng.stoptime = (double)(stop+900);
  rng.lwrLimit = rng.uprLimit = 0;
  rng.startstamp = rng.stopstamp = rng.userstamp = 0;
  
  state[0] = 0;
  npts = getLTSData(&req,rng,(BYTE *)states,CF_NAME64DBL,NULL,100);
  if (nextstate != NULL) nextstate[0] = 0;
  if (nexttime != NULL) *nexttime = 0;
    
  if (npts < 0) return -npts;
  for (i=0; i<npts; i++)
  {
    if (states[i].dval < (double)start) continue;
    if (states[i].dval >= (double)stop) break;
    if (strlen(state) == 0)
    { // as yet unknown. this establishes the declared state at the start
      strncpy(state,states[i].name,64);
    }
    else
    { // when are we no longer in the same state?
      if (strnicmp(state,states[i].name,64) != 0)
      { // not the same state!
        if (nextstate != NULL) strncpy(nextstate,states[i].name,64);
        if (nexttime != NULL) *nexttime = (time_t)states[i].dval;
        break;
      }
    }
  }
  return 0;
}

StateCorrection stateCorrection[MAX_CORRECTION_PER_DAY];

CSVCOLUMN csvStateCorrectionsFileColumns[] =
{
  { "Start","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,startStr),NULL,0 },
  { "Stop","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,stopStr),NULL,0 },
  { "From","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,fromState),NULL,0 },
  { "To","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,toState),NULL,0 },
  { "User","",0,CF_TEXT,16,(BYTE *)OFFSETIN(StateCorrection,usr),NULL,0 },
  { "Comment","",0,CF_TEXT,256,(BYTE *)OFFSETIN(StateCorrection,cmt),NULL,0 },
};
CSVDB csvStateCorrectionsFileDb =
{
  csvStateCorrectionsFileColumns,
  sizeof(csvStateCorrectionsFileColumns)/sizeof(csvStateCorrectionsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(StateCorrection),
  0
};
// the property to get the list of stored correction list over a selected time range
int readStateCorrections(time_t start, time_t stop, StateCorrection *list, int *num)
{
  int n=0, i, cc=0, length = *num;
  char fn[128];
  StateCorrection *stcr = stateCorrection;
  time_t t;
  struct tm *lt;
  int dd, mm, yy, yde;
  
  lt = localtime(&stop); yde = lt->tm_yday;
  for (t = start; t < stop+86400; t += 86400)
  { // Loop over the days involved
    lt = localtime(&t);
    dd = lt->tm_mday; mm = lt->tm_mon; yy = lt->tm_year;
    if (yde < lt->tm_yday) break;
    sprintf(fn,"%s%04d/%02d/%02d/",DBPATH,1900+yy,mm+1,dd);
    csvStateCorrectionsFileDb.siz = MAX_CORRECTION_PER_DAY; // pass the size of the storage buffer
    if ((cc=csvReadFile(fn,CORRECTIONSFILE,&csvStateCorrectionsFileDb,(void **)&stcr)) != 0) continue;
    for (i=0; i<csvStateCorrectionsFileDb.siz && n < length; i++)
    { // check > start && < stop
      stcr[i].stop = (double)getTimeStampFromLogEntry(stcr[i].stopStr);
      stcr[i].start = (double)getTimeStampFromLogEntry(stcr[i].startStr);
      stcr[i].counts = (int)(stcr[i].stop - stcr[i].start);
      if (stcr[i].stop < (double)start) continue;
      if (stcr[i].start > (double)stop) continue;
      list[n++] = stcr[i];
    }
    if (n >= length) break;
  }
  *num = n;
  return 0;
}
// the property to get the corrections over a selected time range
#define STATE_CORR_CNTS 0
#define PROC_CORR_CNTS 1
int getCorrectionCounts(time_t start, time_t stop, int *corcnts, int length, int type)
{ // length typically = ndeclstates
  int i, cc=0, counts, to_idx, fr_idx;
  char fn[128];
  StateCorrection *stcr = stateCorrection;
  time_t t;
  struct tm *lt;
  int dd, mm, yy, yde;
    
  lt = localtime(&stop); yde = lt->tm_yday;
  memset(corcnts, 0, length*sizeof(int));
  for (t = start; t < stop+86400; t += 86400)
  { // Loop over the days involved
    lt = localtime(&t);
    dd = lt->tm_mday; mm = lt->tm_mon; yy = lt->tm_year;
    if (yde < lt->tm_yday) break;
    sprintf(fn,"%s%04d/%02d/%02d/",DBPATH,1900+yy,mm+1,dd); // dbpath ?
    csvStateCorrectionsFileDb.siz = MAX_CORRECTION_PER_DAY; // pass the size of the storage buffer
    if ((cc=csvReadFile(fn,CORRECTIONSFILE,&csvStateCorrectionsFileDb,(void **)&stcr)) != 0) continue;
    for (i=0; i<csvStateCorrectionsFileDb.siz; i++)
    { // reduce array to just corrections in given interval
      // fill in UTC values for start, stop
      stcr[i].stop = (double)getTimeStampFromLogEntry(stcr[i].stopStr);
      stcr[i].start = (double)getTimeStampFromLogEntry(stcr[i].startStr);
      stcr[i].counts = (int)(stcr[i].stop - stcr[i].start);
      // if 'this' start/top fully contained then : as is
      if (stcr[i].stop < start || stcr[i].start > stop) 
      { // if fully excluded then skip
        continue;
      }
      if (stcr[i].stop <= (double)stop && stcr[i].start >= (double)start) 
      { // if fully contained then use as is ...
        counts = stcr[i].counts;
      }
      else
      { // need to inperpolate ...
        if (stcr[i].start < (double)start) counts = (int)(stcr[i].stop - start);
        else counts = (int)(stop - stcr[i].start);
      }
      // find each state index from the 'fromState' and 'toState' and accumulate the counts
      if (type == PROC_CORR_CNTS)
      {
        to_idx = getDeclProcIndex(stcr[i].toState);
        fr_idx = getDeclProcIndex(stcr[i].fromState);
      }
      else
      {
        to_idx = getDeclStateIndex(stcr[i].toState);
        fr_idx = getDeclStateIndex(stcr[i].fromState);
      }
      if (to_idx < 0 || fr_idx < 0) continue;
      if (to_idx >= length || fr_idx >= length) continue;
      corcnts[to_idx] += counts;
      corcnts[fr_idx] -= counts;
    }
  }
  return 0;
}
int getProcedureCorrectionCounts(time_t start, time_t stop, int *corcnts, int length)
{ // length typically = ndeclprocs
  int sttcnts[256]; // this should always be enough
  int i, k, sid, cc = 0;
  memset(corcnts, 0, length *sizeof(int));
  if ((cc = getCorrectionCounts(start, stop, sttcnts, 256, STATE_CORR_CNTS)) != 0) return cc;
  for (i=0; i<length; i++) for (k=0; k<declProcs[i].nstates; k++)
  {
    sid = GetDeviceNumber(STAEQM_TAG,declProcs[i].statenames[k].name);
    corcnts[i] += sttcnts[sid]; 
  }
  // any explicit procedure corrections (this should be rare and only applied if a procedure config was changed mid-stream)
  if (getCorrectionCounts(start, stop, sttcnts, length, PROC_CORR_CNTS) == 0)
  {
    for (i=0; i<length; i++) corcnts[i] += sttcnts[i];
  }

  return 0;
}
int writeStateCorrectionToFile(FILE *fp,StateCorrection *newCorrection,time_t start, time_t stop)
{
  char startstr[64], stopstr[64];
  strncpy(startstr,GetDataTimeString((double)start,FALSE),64);
  strncpy(stopstr,GetDataTimeString((double)stop,FALSE),64);
  if (strnicmp(newCorrection->fromState,newCorrection->toState,64))
  { // if from == to then I'm removing a prior correction (back to the way it was)
    fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n",startstr,stopstr,
      newCorrection->fromState,newCorrection->toState,newCorrection->usr,newCorrection->cmt);
    srvlog(CORRLOGFILE,"writeStateCorrectionToFile","correct %.64s to %.64s from %.32s to %.32s by %.16s",
      newCorrection->fromState,newCorrection->toState,startstr,stopstr,newCorrection->usr);
    return 1;
  }
  srvlog(CORRLOGFILE,"writeStateCorrectionToFile","remove correction %.64s %.32s to %.32s by %.16s",
    newCorrection->fromState,startstr,stopstr,newCorrection->usr);
  return 0;
}
// add a new correction:
// note: there is no record of what the declared state was at the start/stop times given
// this needs to be checked at the Op Hist App (you can't stubtract counts from a state you aren't in!)
// you can subtract states from a correction you are now in !
#define assertStopGTstart(t1,t2) ((t2) > (t1) ? TRUE : FALSE)
#define corrhdr "Start, Stop, From, To, User, Comment"  
int addStateCorrections(StateCorrection *newCorrection) 
{
  int i, n, cc=0, nitems, corrWritten=FALSE;
  char fn[128],nfn[128],bfn[132],ofn[128];
  FILE *fp = NULL;
  StateCorrection *stcr = stateCorrection;
  time_t start = (time_t)newCorrection->start, stop = (time_t)newCorrection->stop, t;
  struct tm *lt;
  int dd, mm, yy;
  int dde, mme, yye, yde;
  time_t nexttime;
  char state[64], nextstate[64];
  getDeclaredStateInRange(state,start,stop,&nexttime,nextstate);
  if (strlen(state) > 0)
  { // found a stored value at the start time given
    if (strnicmp(state,newCorrection->fromState,64)) 
    { // wasn't in the 'from state' !
      srvlog(CORRLOGFILE,"addStateCorrections","correction from %.64s at %.24s not allowed (machine state was %.64s at that time)",
        newCorrection->fromState, ctime(&start), state);
      return not_allowed;
    }
  }
  if (nexttime > 0 && stop >= nexttime) stop = nexttime - 1;
  lt = localtime(&stop);
  dde = lt->tm_mday; mme = lt->tm_mon; yye = lt->tm_year; yde = lt->tm_yday;
  for (t = start; t < stop+86400 && start < stop; t += 86400)
  { // Loop over the days involved
    lt = localtime(&t);
    if (lt->tm_yday > yde) break; /* past the end */
    dd = lt->tm_mday; mm = lt->tm_mon; yy = lt->tm_year;
    sprintf(nfn,"%s%04d/%02d/%02d/corrections.nsv",DBPATH,1900+yy,mm+1,dd); // dbpath ?
    if ((fp=createAndOpen(nfn,"w")) == NULL) ccerr(file_error);
    fprintf(fp,"%s\n",corrhdr);
    if (lt->tm_yday < yde)
    { // spans more than 1 day !
      struct tm it;
      it = *lt;
      it.tm_hour = 23; it.tm_min = 59; it.tm_sec = 59;
      stop = mktime(&it);
    }
    nitems = 0; n = 0;
    sprintf(fn,"%s%04d/%02d/%02d/",DBPATH,1900+yy,mm+1,dd); 
    csvStateCorrectionsFileDb.siz = MAX_CORRECTION_PER_DAY; // pass the size of the storage buffer
    if (csvReadFile(fn,CORRECTIONSFILE,&csvStateCorrectionsFileDb,(void **)&stcr) == 0) 
      nitems = csvStateCorrectionsFileDb.siz;
    corrWritten = FALSE;
    for (i=0; i<nitems; i++)
    { // items read from an existing file ...
      stcr[i].stop = (double)getTimeStampFromLogEntry(stcr[i].stopStr);
      stcr[i].start = (double)getTimeStampFromLogEntry(stcr[i].startStr);
      if (stcr[i].stop < (double)start) 
      { // pre-dates current correction: write back to file
        fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
          stcr[i].startStr,stcr[i].stopStr,stcr[i].fromState,stcr[i].toState,stcr[i].usr,stcr[i].cmt);
        n++;
        continue;
      }
      if (start < (time_t)stcr[i].start) 
      { // the new region starts before the old region (possibly overlapping?)
        if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop)) n++;
        corrWritten = TRUE;
        if (stop >= (time_t)stcr[i].start && stop < (time_t)stcr[i].stop)
        { // overlapping: something left from previous correction
          if (assertStopGTstart(stop+1,stcr[i].stop))
          {
            fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
              GetDataTimeString((double)(stop+1),FALSE),stcr[i].stopStr,stcr[i].fromState,
              stcr[i].toState,stcr[i].usr,stcr[i].cmt);
            n++;
          }
        } 
      }
      if (start >= (time_t)stcr[i].start && start <= (time_t)stcr[i].stop) 
      { // overlapping regions(the new region starts after the old region)
        if (start > (time_t)stcr[i].start+1)
        {
          if (assertStopGTstart(stcr[i].start,start-1))
          {
            fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
              stcr[i].startStr,GetDataTimeString((double)(start-1),FALSE),stcr[i].fromState,stcr[i].toState,stcr[i].usr,stcr[i].cmt);
            n++;
          }
        }
        if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop)) n++;
        corrWritten = TRUE;
        if (stop < (time_t)stcr[i].stop-1)
        {
          if (assertStopGTstart(stop+1,stcr[i].stop))
          {
            fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
              GetDataTimeString((double)(stop+1),FALSE),stcr[i].stopStr,stcr[i].fromState,stcr[i].toState,stcr[i].usr,stcr[i].cmt);
            n++;
          }
        }
      }
      if (stcr[i].start > (double)stop) 
      { // post-dates current correction: finish adding new correction and write back to file
        if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop)) n++;
        corrWritten = TRUE;
        fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
          stcr[i].startStr,stcr[i].stopStr,stcr[i].fromState,stcr[i].toState,stcr[i].usr,stcr[i].cmt);    
        n++;
        continue;
      }
    }  
    if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop)) n++;
    fclose(fp); fp = NULL;
    sprintf(ofn,"%.110scorrections.csv",fn);
    if (n >= MAX_CORRECTION_PER_DAY)
    {
      sprintf(stateValidationErr,"%d corrections for %d.%d.%d would exceed capacity",n,dd,mm,yy);
      feclog("%s",stateValidationErr);
      srvlog(CORRLOGFILE,"addStateCorrections","%s",stateValidationErr);
      return at_limit;
    }
    else
    { // okay: update the corrections file ...
      if (nitems > 0)
      {
        strncpy(bfn,ofn,128);
        strcat(bfn,".bak");
        rename(fn,bfn);
      }
      rename(nfn,ofn);
      start = stop + 1;
      stop = (time_t)newCorrection->stop;
    }
  }
err:
  if(fp != NULL) fclose(fp);
  return 0;  
}

ADS cycAlmDef = { "Missing Cycle Number",512,0,0,9,BFMT(CF_INT32),1,8,0,"Missing Cycle Number","undefined state","last cycler number","none" };
int deltaCycleNumber = 0;
int thisCycleNumber = -1;
void cycCb(int cyc,int cc,void *ref)
{
  if (cc != 0)
  {
    if (!isUsingCycleTrigger) return;
    SetAlarm(STAEQM_TAG,0,512,(BYTE *)&thisCycleNumber);
    return;
  }
  if (thisCycleNumber < 0)
  { /* just starting ... */
    thisCycleNumber = cyc - 1;
  }
  deltaCycleNumber = cyc - thisCycleNumber;
  if (deltaCycleNumber > 1) SystemScheduleProperty(STAEQM_TAG,"CycleNumber.Delta");

  thisCycleNumber = cyc;
}
void staini(void)
{
  int i,cc;
  int dbl = tineDebug;
  NAME64 devlst[128];
  remove("problems.log");

  tineDebug = 1;
  stateValidationErr[0] = 0;
  if ((cc=readStateInfoFromFile()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"Cannot read state info file : %s",erlst[cc]);
    setDefaultDeclStates();
  }

  if ((cc=readProcInfoFromFile()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"Cannot read procedure info file : %s",erlst[cc]);
    setDefaultDeclProcs();
  }
  
  if ((cc=readMachineInfoFromFile()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"Cannot read machine info file : %s",erlst[cc]);
  }

  if (ndeclstates > 128) feclogEx(TINE_LOGLEVEL_ERROR,"problem: number of states exceeds 128 elements");
  for (i=0; i<ndeclstates; i++)
  {
    RegisterDeviceName(STAEQM_TAG,declStates[i].devname,i);
    strncpy(devlst[i].name,declStates[i].devname,64);
  }
  AssignDeviceListToProperty(STAEQM_TAG,"CorrectionCounts.States",devlst,ndeclstates);
  for (i=0; i<ndeclprocs; i++)
  {
    RegisterDeviceName(STAEQM_TAG,declProcs[i].devname,i+ndeclstates);    
    strncpy(devlst[i].name,declProcs[i].devname,64);
  }
  AssignDeviceListToProperty(STAEQM_TAG,"CorrectionCounts.Procedures",devlst,ndeclprocs);
  for (i=0; i<nmachstates; i++)
    RegisterDeviceName(STAEQM_TAG,machStates[i].procname,i+ndeclstates+ndeclprocs);    
  AssignNumDevices(STAEQM_TAG,ndeclstates+ndeclprocs+nmachstates);
  

  if ((cc=readStatesFromFile()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_WARN,"Cannot read state file : %s\n",erlst[cc]);
  }
    
  declStateTime = time(NULL);

  if ((cc=readValidityFile(BEAM1CRITERIAFILE, &isBeamConditions[0])) != 0)
  {
    feclog("Cannot read state validity file %s : %s\n",
            BEAM1CRITERIAFILE,erlst[cc]);
  }
  else
  {
    strcpy(isBeamConditions[0].tag,"Beam 1");
    startValidityLinks(&isBeamConditions[0],MAKETYPEID(0,0));
  }
  if ((cc=readValidityFile(TEST1CRITERIAFILE, &isTestConditions[0])) != 0)
  {
    feclog("Cannot read state validity file %s : %s\n",
            TEST1CRITERIAFILE,erlst[cc]);
  }
  else
  {
    strcpy(isTestConditions[0].tag,"Test 1");
    startValidityLinks(&isTestConditions[0],MAKETYPEID(1,0));
  }
  if (gNumParticleTypes > 1)
  {
    if ((cc=readValidityFile(BEAM2CRITERIAFILE, &isBeamConditions[1])) != 0)
    {
      feclog("Cannot read state validity file %s : %s\n",
              BEAM2CRITERIAFILE,erlst[cc]);
    }
    else
    {
      strcpy(isBeamConditions[1].tag,"Beam 2");
      startValidityLinks(&isBeamConditions[1],MAKETYPEID(0,1));
    }
    if ((cc=readValidityFile(TEST2CRITERIAFILE, &isTestConditions[1])) != 0)
    {
      feclog("Cannot read state validity file %s : %s\n",
              TEST2CRITERIAFILE,erlst[cc]);
    }
    else
    {
      strcpy(isTestConditions[1].tag,"Test 2");
      startValidityLinks(&isTestConditions[1],MAKETYPEID(1,1));
    }
  }
  
  startAvailabilityLink();

  AppendAlarmInfoTable(STAEQM_TAG,&cycAlmDef);
  RegisterCycleTriggerFunction(cycCb,NULL,NULL,NULL);

  readViewerFiles();

  RegisterUserCommand("ValidityTable",dumpValidity,NULL,NULL,CA_READ);

  getAdministrators(ADMINS_FILE);

  readStateTriggerInfoFromFile();

  gLastCorrectionTimeStamp = time(NULL);

  // the Reinhard workaround :
  readSkipFirstFile();
  feclog("state server initialized");
  tineDebug = dbl;
}
void trgCallback(int id,int cc)
{
  if (cc && tineDebug > 1) dbglog("TRIGGER : %s",erlst[cc]);
}
int sndTriggerToArchive(char *trg)
{
  DTYPE din;
  int cc;
  short sval = 1;
  char devName[256],str[256];

  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.sptr = (SINT16 *)&sval;
  sprintf(devName,"/%.32s/EVENTSTORE/%.64s",gDeviceContext,trg);
  if ((cc=AttachLink(devName,"TRIGGER",NULL,&din,CA_WRITE,1000,trgCallback,CM_SINGLE)) < 0)
  {
    sprintf(str,"TRIGGER %.128s error : %.32s",devName,erlst[-cc]);
    logproblem(str);
    feclog(str);
  }
  else
  {
    feclog("send trigger %s to archive",trg);
  }
  return cc < 0 ? -cc : 0;
}
int forwardTrigger(char *devsrv,char *trg)
{
  DTYPE din;
  int cc;
  short sval = 1;
  char devName[48],str[256];

  sprintf(devName,"%s/%s",devsrv,trg);
  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.sptr = (SINT16 *)&sval;
  if ((cc=AttachLink(devName,"STATECHANGE",NULL,&din,CA_WRITE,1000,trgCallback,CM_SINGLE)) < 0)
    dbglog("TRIGGER %s error : %s\n",devName,erlst[-cc]);
  sprintf(str,"forward trigger to %s",devsrv);
  feclog(str);
  return cc < 0 ? -cc : 0;
}
int getDeclStateIndex(char *devname)
{
  int i;
  for (i=0; i<ndeclstates; i++)
  {
    if (strnicmp(devname,declStates[i].devname,DEVICE_NAME_SIZE)) continue;
    return i;
  }
  return -1;
}
int addDeclStates(DCLSTATE *src,int len)
{
  int i, idx;
  for (i=0; i<len; i++)
  {
    if ((idx=getDeclStateIndex(src[i].devname)) < 0)
    { /* new! */
      if (ndeclstates == declstatesCapacity) return resources_exhausted;
      idx = ndeclstates++;
      strncpy(declStates[idx].devname,src[i].devname,DEVICE_NAME_SIZE);
    }
    strncpy(declStates[idx].statetag,src[i].statetag,DEVICE_NAME_SIZE);
    declStates[idx].machine = src[i].machine;
  }
  return 0;
}
int getDeclStates(DCLSTATE *dst,int len)
{
  int i;
  for (i=0; i<len; i++)
  {
    strncpy(dst[i].devname,declStates[i].devname,DEVICE_NAME_SIZE);
    strncpy(dst[i].statetag,declStates[i].statetag,DEVICE_NAME_SIZE);
    dst[i].machine = declStates[i].machine;
  }
  return 0;
}
int getDeclProcIndex(char *devname)
{
  int i;
  for (i=0; i<ndeclprocs; i++)
  {
    if (strnicmp(devname,declProcs[i].devname,DEVICE_NAME_SIZE)) continue;
    return i;
  }
  return -1;
}
int addDeclProcs(DCLPROC *src,int len)
{
  int i, idx;
  NAME64 *nms;
  for (i=0; i<len; i++)
  {
    if ((idx=getDeclProcIndex(src[i].devname)) < 0)
    { /* new! */
      if (ndeclprocs == declprocsCapacity) return resources_exhausted;
      if (strlen(src[i].devname) == 0) return invalid_data;
      idx = ndeclprocs++;
      strncpy(declProcs[idx].devname,src[i].devname,DEVICE_NAME_SIZE);
      if ((nms=(NAME64 *)calloc(src[i].nstates,sizeof(NAME64))) == NULL) return out_of_server_memory;
      declProcs[idx].statenames = nms;
      sprintf(declProcs[idx].structfile,"%.60s.csv",src[i].devname);
    }
    else
    {
      if (declProcs[idx].nstates < src[i].nstates)
      { /* more states have been added to the procedure ! */
        nms = (NAME64 *)realloc(declProcs[idx].statenames,src[i].nstates*sizeof(NAME64));
        if (nms != NULL) declProcs[idx].statenames = nms; else src[i].nstates = declProcs[idx].nstates;
      }
    }
    declProcs[idx].nstates = src[i].nstates;
    declProcs[idx].machine = src[i].machine;
  }
  return 0;
}
int getDeclProcs(DCLPROC *dst,int len)
{
  int i;
  for (i=0; i<len; i++)
  {
    strncpy(dst[i].devname,declProcs[i].devname,DEVICE_NAME_SIZE);
    strncpy(dst[i].structfile,declProcs[i].structfile,DEVICE_NAME_SIZE);
    dst[i].machine = declProcs[i].machine;
    dst[i].nstates = declProcs[i].nstates;
  }
  return 0;
}
int staeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc = 0, i, j, k, n, nlen, len, newproc = FALSE;
  int dc = FALSE, bc = FALSE, tc = FALSE, ec = FALSE;
  int retDevName = FALSE;
  short dsize,devnr,fmt;
  NAME16 un;
  char *c, addr[32],usr[USERNAME_SIZE+4];
  BYTE ipxaddr[12];
  UINT32 ipaddr;

  fmt = dout->dFormat;
  dsize = (short)dout->dArrayLength;
  
  devnr = GetDeviceNumberEx(STAEQM_TAG,devName,devProperty);

  n = 1; GetCallerInfo(STAEQM_TAG,&un,ipxaddr,&ipaddr,NULL,&n);
  strncpy(usr,un.name,16); strupr(usr);

  if (!stricmp(devProperty,"CALLER"))  /* deduced machine state(s) */
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength < 1 || dout->dArrayLength > 2) return dimension_error;
    switch (dout->dFormat)
    {
      case CF_NAME16: len = 16; n = dout->dArrayLength; break;
      case CF_NAME32: len = 32; n = dout->dArrayLength; break;
      case CF_TEXT:   if (dout->dArrayLength < 16) return dimension_error;
                      len = 16; n = dout->dArrayLength/16; break;
      default:        return illegal_format;
    }
    strncpy(dout->data.cptr,lastCallerToChangeState.username,16);
    if (n > 1) strncpy(dout->data.cptr+len,lastCallerToChangeState.useraddr,16);
    return 0;
  }
  if (!stricmp(devProperty,"NumParticleTypes"))
  {
    return PutValuesFromLong(dout,&gNumParticleTypes,1);
  }
  if (!stricmp(devProperty,"TIMESTATS"))  /* state time/duration information */
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devnr < 0 || devnr >= ndeclstates) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    switch (dout->dArrayLength)
    {
      case 6:  dout->data.lptr[4] = declStates[devnr].ttd_ec;
      case 5:  dout->data.lptr[4] = declStates[devnr].ttd_tc;
      case 4:  dout->data.lptr[3] = declStates[devnr].ttd_bc;
      case 3:  dout->data.lptr[2] = declStates[devnr].tnd;
      case 2:  dout->data.lptr[1] = (int)declStates[devnr].ltd;
      case 1:  dout->data.lptr[0] = (int)declStates[devnr].ttd;
               break;
      default: return dimension_error;
    }
    return 0;
  }
  if ((dc=!stricmp(devProperty,"TIMECOUNTER")) ||  /* state duration counter */
      (bc=!stricmp(devProperty,"BEAMCOUNTER")) ||
      (tc=!stricmp(devProperty,"TESTCOUNTER")) ||
      (ec=!stricmp(devProperty,"ERRCOUNTER")) )
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devnr < 0 || devnr >= ndeclstates + ndeclprocs + nmachstates) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    if (devnr < ndeclstates)
    {
      for (i=devnr; i<(int)dout->dArrayLength+devnr && i < ndeclstates+devnr; i++)
      {
        if (dc) dout->data.lptr[i-devnr] = declStates[i%ndeclstates].ttd;
        else if (bc) dout->data.lptr[i-devnr] = declStates[i%ndeclstates].ttd_bc;
        else if (tc) dout->data.lptr[i-devnr] = declStates[i%ndeclstates].ttd_tc;
        else if (ec) dout->data.lptr[i-devnr] = declStates[i%ndeclstates].ttd_ec;
      }
    }
    else if (devnr < ndeclstates + ndeclprocs)
    {
      for (i=devnr; i<(int)dout->dArrayLength+devnr && i < ndeclstates+ndeclprocs; i++)
      {
        dout->data.lptr[i-devnr] = 0;
        for (j=0; j<declProcs[i-ndeclstates].nstates; j++)
        {
          k = GetDeviceNumber(STAEQM_TAG,declProcs[i-ndeclstates].statenames[j].name);
          if (k >= 0)
          {
            if (dc) dout->data.lptr[i-devnr] += declStates[k].ttd;
            else if (bc) dout->data.lptr[i-devnr] += declStates[k].ttd_bc;
            else if (tc) dout->data.lptr[i-devnr] += declStates[k].ttd_tc;
            else if (ec) dout->data.lptr[i-devnr] += declStates[k].ttd_ec;
          }
        }
      }
    }
    else if (devnr < ndeclstates + ndeclprocs + nmachstates)
    {
      n = ndeclstates + ndeclprocs;
      for (i=devnr; i<(int)dout->dArrayLength+devnr && i < ndeclstates+ndeclprocs+nmachstates; i++)
      {
        if (dc) dout->data.lptr[i-devnr] = machStates[i%n].ttd;
        else if (bc) dout->data.lptr[i-devnr] = machStates[i%n].ttd_bc;
        else if (tc) dout->data.lptr[i-devnr] = machStates[i%n].ttd_tc;
        else if (ec) dout->data.lptr[i-devnr] = machStates[i%n].ttd_ec;
      }
    }
    dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if ((dc=!stricmp(devProperty,"TIMECOUNTER.Procedure")) ||  /* state procedure duration counter */
      (bc=!stricmp(devProperty,"BEAMCOUNTER.Procedure")) ||
      (tc=!stricmp(devProperty,"TESTCOUNTER.Procedure")) ||
      (ec=!stricmp(devProperty,"ERRCOUNTER.Procedure")) )
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devnr < 0 || devnr >= ndeclprocs) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    for (i=devnr,n=0; i<(int)dout->dArrayLength+devnr && i<ndeclprocs; i++)
    {
      dout->data.lptr[i-devnr] = 0;
      for (j=0; j<declProcs[i].nstates; j++)
      {
        k = GetDeviceNumber(STAEQM_TAG,declProcs[i].statenames[j].name);
        if (k >= 0)
        {
          if (dc) dout->data.lptr[i-devnr] += declStates[k].ttd;
          else if (bc) dout->data.lptr[i-devnr] += declStates[k].ttd_bc;
          else if (tc) dout->data.lptr[i-devnr] += declStates[k].ttd_tc;
          else if (ec) dout->data.lptr[i-devnr] += declStates[k].ttd_ec;
        }
      }
      n++;
    }
    dout->dArrayLength = (UINT32)n;
    return 0;
  }
  if (!stricmp(devProperty,"TIMECOUNTER.Procedure.NAM") ||
      !stricmp(devProperty,"BEAMCOUNTER.Procedure.NAM") ||
      !stricmp(devProperty,"TESTCOUNTER.Procedure.NAM") ||
      !stricmp(devProperty,"EVENTCOUNTER.Procedure.NAM") ||
      !stricmp(devProperty,"ERRCOUNTER.Procedure.NAM") ||
      !stricmp(devProperty,"PROCSTATES.NAM") )
  {
    if (access & CA_WRITE) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength && i<ndeclprocs; i++)
    {
      strncpy(dout->data.cptr + i*nlen,declProcs[i].devname,nlen);
    }
    dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if (!stricmp(devProperty,"EVENTCOUNTER"))  /* state number counter */
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devnr < 0 || devnr >= ndeclstates + ndeclprocs) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    if (devnr < ndeclstates)
    {
      for (i=devnr; i<(int)dout->dArrayLength+devnr; i++)
      {
        dout->data.lptr[i-devnr] = declStates[i%ndeclstates].tnd;
      }
      if ((access&CA_HIST) == 0) setDataTimeStamp((double)declStates[devnr%ndeclstates].ltd);
    }
    else if (devnr < ndeclstates + ndeclprocs)
    {
      for (i=devnr; i<(int)dout->dArrayLength+devnr; i++)
      {
        dout->data.lptr[i-devnr] = declProcs[(i-ndeclstates)%ndeclprocs].tnd;
      }
      if ((access&CA_HIST) == 0) setDataTimeStamp((double)declProcs[(devnr-ndeclstates)%ndeclprocs].ltd);
    }
    return 0;
  }
  if (!stricmp(devProperty,"EVENTCOUNTER.Procedure"))  /* state number counter */
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devnr < 0 || devnr >= ndeclprocs) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    for (i=devnr,n=0; i<(int)dout->dArrayLength; i++)
    {
      dout->data.lptr[i-devnr] = declProcs[i%ndeclprocs].tnd;
      n++;
    }
    if ((access&CA_HIST) == 0) setDataTimeStamp((double)declProcs[devnr%ndeclprocs].ltd);
    dout->dArrayLength = (UINT32)n;
    return 0;
  }
  if (!stricmp(devProperty,"DECLSTATE") || 
      !stricmp(devProperty,"DeclaredState") ||
      !stricmp(devProperty,"LastDeclaredState")
    )
  { /* declared machine state */
    time_t tnow = time(NULL) - 10; /* back off 10 seconds to avoid collision with new state declaration time */
    double ts = (tnow > declStateTime + 300) ? (double)tnow : (double)declStateTime;
    if ((access&CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    if (stricmp(devName,"name") == 0) retDevName = TRUE;
    if ((access & CA_HIST) != 0) retDevName = TRUE;
    if (devProperty[0] == 'L') ts = (double)declStateTime;
    dout->dTimeStamp = ts;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        //memset(dout->data.bptr,0,nlen*dout->dArrayLength);
        c = retDevName ? declStates[declaredState[0]].devname : declStates[declaredState[0]].statetag;
        strncpy(dout->data.cptr,c,nlen);       
        if (dout->dArrayLength > 1) 
        {
          c = retDevName ? declStates[declaredState[1]].devname : declStates[declaredState[1]].statetag;
          strncpy(dout->data.cptr+nlen,c,nlen);       
        }
        return 0;
      case CF_TEXT:
        //memset(dout->data.bptr,0,dout->dArrayLength);
        if (declStates == NULL || declStates[0].ltd >= declStates[1].ltd) i = 0; else i = 1;
        strncpy(dout->data.cptr,declStates[declaredState[i]].statetag,MIN(dout->dArrayLength,64));
        return 0;
      default: 
        return putValuesFromLong(dout,(SINT32 *)declaredState,2);
    }
  }
  if (!stricmp(devProperty,"PreviousState"))
  { /* declared machine state */
    if ((access&CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    if (stricmp(devName,"name") == 0) retDevName = TRUE;
    if ((access & CA_HIST) != 0) retDevName = TRUE;
    dout->dTimeStamp = (double)declStateTime;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        c = retDevName ? declStates[prevDeclaredState[0]].devname : declStates[prevDeclaredState[0]].statetag;
        strncpy(dout->data.cptr,c,nlen);       
        if (dout->dArrayLength > 1) 
        {
          c = retDevName ? declStates[prevDeclaredState[1]].devname : declStates[prevDeclaredState[1]].statetag;
          strncpy(dout->data.cptr+nlen,c,nlen);       
        }
        return 0;
      case CF_TEXT:
        if (declStates[0].ltd >= declStates[1].ltd) i = 0; else i = 1;
        strncpy(dout->data.cptr,declStates[prevDeclaredState[i]].statetag,MIN(dout->dArrayLength,64));
        return 0;
      default: 
        return putValuesFromLong(dout,(SINT32 *)prevDeclaredState,2);
    }
  }
  if (!stricmp(devProperty,"STATELIST"))  /* list machine states */
  {
    if ((access&CA_WRITE)) 
    {
      if (!isAdministrator(usr)) 
      {
        char es[256];
        sprintf(es,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(STAEQM_TAG,es);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"DclState") != 0) return illegal_format;
      cc = addDeclStates((DCLSTATE *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
      writeStateInfoToFile();
    }
    if (dout->dArrayLength > 0)
    {
      if (dout->dArrayLength > (UINT32)ndeclstates) dout->dArrayLength = (UINT32)ndeclstates;
      switch (dout->dFormat)
      {
        case CF_STRUCT:
          if (strcmp(dout->dTag,"DclState")) return invalid_structure_tag;
          return getDeclStates((DCLSTATE *)dout->data.vptr,dout->dArrayLength);
          break;
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          nlen = fmtsizeof(dout->dFormat);
          break;
        default: 
          return illegal_format;
      }
      for (i=0; i<(int)dout->dArrayLength; i++)
      {
        strncpy(dout->data.cptr+nlen*i,declStates[i].statetag,nlen);
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"PROCSTATES"))  
  { /* list machine states in given procedure */
    if (devnr < 0) 
    { /* one last chance through 'added' procs */
      if ((devnr=getDeclProcIndex(devName)) < 0) return illegal_equipment_number;
    }
    n = declProcs[devnr].nstates;
    if ((access&CA_WRITE)) 
    {
      if (din->dArrayLength == 0) return dimension_error;
      if (din->dArrayLength != n)
      { /* mismatch in procinfo and number of states input ! */
        if (din->dArrayLength > (UINT32)n)
        { /* this has to grow ! */
          DCLPROC dp;
          memcpy(&dp, &declProcs[devnr], sizeof(dp));
          dp.nstates = din->dArrayLength;
          cc = addDeclProcs(&dp, n);
        }
        n = declProcs[devnr].nstates = din->dArrayLength;
      }
      cc = GetValuesAsStringEx(din,declProcs[devnr].statenames[0].name,CF_NAME64,n,0);
      if (cc != 0) return cc;
      cc = writeProcStructureFile(declProcs[devnr].structfile,&declProcs[devnr]);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    {
      if (dout->dArrayLength > (UINT32)n) dout->dArrayLength = (UINT32)n;
      cc = PutValuesFromString(dout,declProcs[devnr].statenames[0].name,CF_NAME64,n,0);
    }
    return cc;
  }
  if (!stricmp(devProperty,"StatesInProcedure"))  
  { /* running (dressed) list */
    if ((n=(int)dout->dArrayLength) > 0)
    {
      switch (dout->dFormat)
      {
        case CF_NAME64:
        {
          NAME64 *n64 = (NAME64 *)dout->data.vptr;
          for (i=0,k=0; i<ndeclprocs && k<n; i++) 
          {
            sprintf(n64[k++].name,"PROC: %.48s",declProcs[i].devname);
            for (j=0; j<declProcs[i].nstates && k<n; j++)
            {
              sprintf(n64[k++].name,"STAT: %.48s",declProcs[i].statenames[j].name);
            }
          }
          if (dout->dArrayLength > (UINT32)k) dout->dArrayLength = (UINT32)k;
          return 0;
        }
        case CF_XML:
        {
          char sbuf[128];
          char *str = dout->data.cptr;
          memset(str,0,n);
          for (i=0,k=0; i<ndeclprocs; i++) 
          {
            sprintf(sbuf,"<PROC>\n\t<NAME>%.64s</NAME>\n",declProcs[i].devname);
            if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
            for (j=0; j<declProcs[i].nstates && k<n; j++)
            {
              sprintf(sbuf,"\t<STATE>%.64s</STATE>\n",declProcs[i].statenames[j].name);
              if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
            }
            sprintf(sbuf,"</PROC>\n");
            if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
          }
          if (dout->dArrayLength > (UINT32)k) dout->dArrayLength = (UINT32)k;
          return 0;
        }
        default:
          return illegal_format;
      }
    }
    return dimension_error;
  }
  if (!stricmp(devProperty,"StatesInProfile"))  
  { /* running (dressed) list */
    if (devnr >= gNumOprViewFiles) return invalid_name;
    if ((n=(int)dout->dArrayLength) > 0)
    {
      int p,np=100;
      OprProfile prof[100];
      NAME64 lst[50];
      int lstsiz = 50, pi;
      if ((cc=populateProfileSet(oprViewFiles[devnr].stateProfiles,prof,&np,FALSE)) != 0) return cc;
      if (np > 100) return code_failure;
      switch (dout->dFormat)
      {
        case CF_NAME64:
        {
          NAME64 *n64 = (NAME64 *)dout->data.vptr;
          for (p=0,k=0; p<np; p++)
          {
            sprintf(n64[k++].name,"PROF: %.48s",prof[p].name);
            lstsiz = 50;
            if ((cc=getProfileProcList(prof[p].states,lst,&lstsiz)) != 0) return cc;
            for (i=0; i<lstsiz; i++) 
            {
              if ((pi=getDeclProcIndex(lst[i].name)) < 0) continue;
              for (j=0; j<declProcs[pi].nstates && k<n; j++)
              {
                sprintf(n64[k++].name,"STAT: %.48s",declProcs[pi].statenames[j].name);
              }
            }
          }
          if (dout->dArrayLength > (UINT32)k) dout->dArrayLength = (UINT32)k;
          return 0;
        }
        case CF_XML:
        {
          char sbuf[128];
          char *str = dout->data.cptr;
          memset(str,0,n);
          for (p=0,k=0; p<np; p++)
          {
            sprintf(sbuf,"<PROF>\n\t<NAME>%.64s</NAME>\n",prof[p].name);
            if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
            lstsiz = 50;
            if ((cc=getProfileProcList(prof[p].states,lst,&lstsiz)) != 0) return cc;
            for (i=0; i<lstsiz; i++) 
            {
              if ((pi=getDeclProcIndex(lst[i].name)) < 0) continue;
              for (j=0; j<declProcs[pi].nstates && k<n; j++)
              {
                sprintf(sbuf,"\t<STATE>%.64s</STATE>\n",declProcs[pi].statenames[j].name);
                if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
              }
            }
            sprintf(sbuf,"</PROF>\n");
            if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
          }
          if (dout->dArrayLength > (UINT32)k) dout->dArrayLength = (UINT32)k;
          return 0;
        }
        default:
          return illegal_format;
      }
    }
    return dimension_error;
  }
  if (!stricmp(devProperty,"ProceduresInProfile"))  
  { /* running (dressed) list */
    if (devnr >= gNumOprViewFiles) return invalid_name;
    if ((n=(int)dout->dArrayLength) > 0)
    {
      int p,np=100;
      OprProfile prof[100];
      NAME64 lst[20];
      int lstsiz = 20;
      if ((cc=populateProfileSet(oprViewFiles[devnr].stateProfiles,prof,&np,FALSE)) != 0) return cc;
      if (np > 100) return code_failure;
      switch (dout->dFormat)
      {
        case CF_NAME64:
        {
          NAME64 *n64 = (NAME64 *)dout->data.vptr;
          for (p=0,k=0; p<np; p++)
          {
            sprintf(n64[k++].name,"PROF: %.48s",prof[p].name);
            lstsiz = 50;
            if ((cc=getProfileProcList(prof[p].states,lst,&lstsiz)) != 0) return cc;
            for (i=0; i<lstsiz; i++) 
            {
              sprintf(n64[k++].name,"PROC: %.48s",lst[i].name);
            }
          }
          if (dout->dArrayLength > (UINT32)k) dout->dArrayLength = (UINT32)k;
          return 0;
        }
        case CF_XML:
        {
          char sbuf[128];
          char *str = dout->data.cptr;
          memset(str,0,n);
          for (p=0,k=0; p<np; p++)
          {
            sprintf(sbuf,"<PROF>\n\t<NAME>%.64s</NAME>\n",prof[p].name);
            if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
            lstsiz = 50;
            if ((cc=getProfileProcList(prof[p].states,lst,&lstsiz)) != 0) return cc;
            for (i=0; i<lstsiz; i++) 
            {
              sprintf(sbuf,"\t<PROC>%.64s</PROC>\n",lst[i].name);
              if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
            }
            sprintf(sbuf,"</PROF>\n");
            if ((k += (int)strlen(sbuf)) < n) strcat(str,sbuf); else return buffer_too_small;
          }
          if (dout->dArrayLength > (UINT32)k) dout->dArrayLength = (UINT32)k;
          return 0;
        }
        default:
          return illegal_format;
      }
    }
    return dimension_error;
  }

  if (!stricmp(devProperty,"PROCLIST"))  
  { /* list machine procedures */
    if ((access&CA_WRITE)) 
    {
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"DclProc") != 0) return illegal_format;
      cc = addDeclProcs((DCLPROC *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
      writeProcInfoToFile();
    }
    if (dout->dArrayLength > 0)
    {
      if (dout->dArrayLength > (UINT32)ndeclprocs) dout->dArrayLength = (UINT32)ndeclprocs;
      switch (dout->dFormat)
      {
        case CF_STRUCT:
          if (strcmp(dout->dTag,"DclProc")) return invalid_structure_tag;
          return getDeclProcs((DCLPROC *)dout->data.vptr,dout->dArrayLength);
          break;
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          nlen = fmtsizeof(dout->dFormat);
          break;
        default: 
          return illegal_format;
      }
      for (i=0; i<(int)dout->dArrayLength; i++)
      {
        strncpy(dout->data.cptr+nlen*i,declProcs[i].devname,nlen);
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"DECLPROC") || 
      !stricmp(devProperty,"DeclaredProcedure")) 
  { /* declared machine state */
    if ((access&CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    if (stricmp(devName,"name") == 0 && (access&CA_HIST) == 0) 
    { /* not a history call and not flagged to receive the device name : set the timestamp to time of state change */
      setDataTimeStamp((double)declStateTime);
    }
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        strncpy(dout->data.cptr,declProcs[declaredProc[0]].devname,nlen);       
        if (dout->dArrayLength > 1) 
        {
          strncpy(dout->data.cptr+nlen,declProcs[declaredProc[1]].devname,nlen);       
        }
        return 0;
      case CF_TEXT:
        if (declStates[0].ltd >= declStates[1].ltd) i = 0; else i = 1;
        strncpy(dout->data.cptr,declProcs[declaredProc[i]].devname,MIN(dout->dArrayLength,64));
        return 0;
      default: 
        return putValuesFromLong(dout,(SINT32 *)declaredProc,2);
    }
  }
  if (!stricmp(devProperty,"PreviousProcedure")) 
  { /* declared machine state */
    if ((access&CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    if (stricmp(devName,"name") == 0 && (access&CA_HIST) == 0) 
    { /* not a history call and not flagged to receive the device name : set the timestamp to time of state change */
      setDataTimeStamp((double)declStateTime);
    }
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        strncpy(dout->data.cptr,declProcs[prevDeclaredProc[0]].devname,nlen);       
        if (dout->dArrayLength > 1) 
        {
          strncpy(dout->data.cptr+nlen,declProcs[prevDeclaredProc[1]].devname,nlen);       
        }
        return 0;
      case CF_TEXT:
        if (declStates[0].ltd >= declStates[1].ltd) i = 0; else i = 1;
        strncpy(dout->data.cptr,declProcs[prevDeclaredProc[i]].devname,MIN(dout->dArrayLength,64));
        return 0;
      default: 
        return putValuesFromLong(dout,(SINT32 *)prevDeclaredProc,2);
    }
  }
  if (!stricmp(devProperty,"STATECHANGE"))
  { /* call must set the state (no read-only calls) */
    if (!(access&CA_WRITE)) return illegal_read_write;
    if (devnr < 0 || devnr >= ndeclstates) return illegal_equipment_number;
    if (skipFirst(devnr))
    {
      feclog("State changed to %s by %s [%s] skipped !", declStates[devnr].statetag, un.name, addr);
      return 0;
    }
    if (devnr == gLastStateChangeTigger) return 0; /* not a state CHANGE */
    /* set global variables which are scanned for in background task */
    strncpy(gStateChangeTriggerName, devName, DEVICE_NAME_SIZE);
    gLastStateChangeTigger = devnr;
    gStateChangeTrigger = devnr;
    gStateChangeTriggerTime = MakeDataTimeStamp();
    /* put caller into caller buffer: */
    if (ipaddr != 0) inet_ntoa_b(*(struct in_addr *)&ipaddr, addr);
    strncpy(lastCallerToChangeState.username, un.name, 16);
    strncpy(lastCallerToChangeState.useraddr, addr, 16);
    feclog("State changed to %s by %s [%s]", declStates[devnr].statetag, un.name, addr);
#   ifdef NO_DELAYED_STATECHANGE
    int type = declStates[devnr].machine;
    if (type&PTYPE1) 
    {
      if (declaredState[0] == (UINT32)devnr) return 0; /* return already_assigned; */
      prevDeclaredState[0] = declaredState[0];
      declaredState[0] = devnr;
      gScheduleDeclState = TRUE;
    }
    if (type&PTYPE2) 
    {
      if (declaredState[1] == (UINT32)devnr) return 0; /* return already_assigned; */
      prevDeclaredState[1] = declaredState[1];
      declaredState[1] = devnr;
      gScheduleDeclState = TRUE;
    }
    declStates[devnr].ltd = time(NULL);
    declStates[devnr].tnd++;
    /* use it to trigger archives here ... */
    declStateTime = time(NULL);
    writeStatesToFile();
    writeStateInfoToFile();
    if (declStates[devnr].evntrg)
    { /* send event trigger when state is declared */
      sndTriggerToArchive(declStates[devnr].devname);
    }
    for (i=0; i<ndeclprocs; i++) for (j=0; j<declProcs[i].nstates; j++)
    {
      if (strncmp(devName,declProcs[i].statenames[j].name,DEVICE_NAME_SIZE)) continue;
      if (type&PTYPE1)
      {
        if ((int)declaredProc[0] != i) 
        { 
          prevDeclaredProc[0] = declaredProc[0]; 
          declaredProc[0] = i; 
          newproc = TRUE; 
        }
      }
      if (type&PTYPE2)
      {
        if ((int)declaredProc[1] != i) 
        { 
          prevDeclaredProc[1] = declaredProc[1]; 
          declaredProc[1] = i; 
          newproc = TRUE; 
        }
      }
      if (newproc)
      {
        declProcs[i].ltd = declStateTime;
        declProcs[i].tnd++;
        writeProcInfoToFile();
      }
      gLastStateChangeTigger = gStateChangeTrigger;
      return cc;
    }
#   endif
    if (dout->dArrayLength > 0)
    {
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          nlen = fmtsizeof(dout->dFormat);
          strncpy(dout->data.cptr,declStates[declaredState[0]].statetag,nlen);       
          if (dout->dArrayLength > 1) 
          {
            strncpy(dout->data.cptr+nlen,declStates[declaredState[1]].statetag,nlen);       
          }
          return 0;
        case CF_TEXT:
          if (declStates[0].ltd >= declStates[1].ltd) i = 0; else i = 1;
          strncpy(dout->data.cptr,declStates[declaredState[i]].statetag,MIN(dout->dArrayLength,64));
          return 0;
        default: 
          return putValuesFromLong(dout,(SINT32 *)declaredState,2);
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"CONFIGURATION"))
  {
    if (din->dArrayLength > 0)
    {
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(usr)) 
      {
        char es[256];
        sprintf(es,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(STAEQM_TAG,es);
        return access_denied;
      }
      if (devnr >= gNumOprViewFiles) return invalid_name;
      switch (din->dFormat)
      {
        case CF_STRUCT:
          if (strcmp(din->dTag,"McaCfg")) return invalid_structure_tag;
          if (devnr < 0)
          { /* seems to be brand new ! (n.b. PROFILE and CONFIGURATION share the same configs file */
            if ((n=appendViewerFile(devName,CONFIGSFILE,&gNumOprViewFiles,&csvOprViewFileDb,&oprViewFiles)) < 0) 
              return -n;
            devnr = n - 1; /* appended at the end */
            resetPropertyDeviceList(STAEQM_TAG,"CONFIGURATION");
            resetPropertyDeviceList(STAEQM_TAG,"PROFILE");
          }
          cc = commitMcaSet(oprViewFiles[devnr].stateParams,(McaSet *)din->data.vptr,din->dArrayLength);
          if (cc != 0) return cc;
          break;
        case CF_TEXT:
          if (devnr < 0 || devnr >= gNumOprViewFiles) return invalid_name;
          if (din->dArrayLength < 6) return invalid_data;
          if (!stricmp(din->data.cptr,"REMOVE"))
          {
            if ((cc=updateViewerFiles()) != 0) return cc;
          }
          else
          {
            return not_accepted;
          }
        default:
          return illegal_format;
      }
    }
    if (dout->dArrayLength > 0)
    {
      if (devnr < 0 || devnr >= gNumOprViewFiles) return invalid_name;
      switch (dout->dFormat)
      {
        case CF_STRUCT:
          if (strcmp(dout->dTag,"McaCfg")) return invalid_structure_tag;
          return populateMcaSet(oprViewFiles[devnr].stateParams,(McaSet *)dout->data.vptr,(int *)&dout->dArrayLength);
        default:
          return illegal_format;
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"PROFILE"))
  {
    if (din->dArrayLength > 0)
    {
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(usr)) 
      {
        char es[256];
        sprintf(es,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(STAEQM_TAG,es);
        return access_denied;
      }
      if (devnr >= gNumOprViewFiles) return invalid_name;
      switch (din->dFormat)
      {
        case CF_STRUCT:
          if (strncmp(din->dTag,"PrfCfg",6)) return invalid_structure_tag;
          if (devnr < 0)
          { /* seems to be brand new ! (n.b. PROFILE and CONFIGURATION share the same configs file */
            if ((n=appendViewerFile(devName,CONFIGSFILE,&gNumOprViewFiles,&csvOprViewFileDb,&oprViewFiles)) < 0) 
              return -n;
            devnr = n - 1; /* appended at the end */
            resetPropertyDeviceList(STAEQM_TAG,"CONFIGURATION");
            resetPropertyDeviceList(STAEQM_TAG,"PROFILE");
          }
          if (!strcmp(din->dTag,"PrfCfg"))
            cc = commitProfileSet(oprViewFiles[devnr].stateProfiles,din->data.vptr,din->dArrayLength,FALSE);
          else if (!strcmp(din->dTag,"PrfCfgEx"))
            cc = commitProfileSet(oprViewFiles[devnr].stateProfiles,din->data.vptr,din->dArrayLength,TRUE);
          else
            cc = invalid_structure_tag;
          if (cc != 0) return cc;
          break;
        case CF_TEXT:
          if (devnr < 0 || devnr >= gNumOprViewFiles) return invalid_name;
          if (din->dArrayLength < 6) return invalid_data;
          if (!stricmp(din->data.cptr,"REMOVE"))
          {
            if ((cc=updateViewerFiles()) != 0) return cc;
          }
          else
          {
            return not_accepted;
          }
        default:
          return illegal_format;
      }
      validateProfileCompleteness(devName);
    }
    if (dout->dArrayLength > 0)
    {
      if (devnr < 0 || devnr >= gNumOprViewFiles) return invalid_name;
      switch (dout->dFormat)
      {
        case CF_STRUCT:
          if (!strcmp(dout->dTag,"PrfCfg"))
            return populateProfileSet(oprViewFiles[devnr].stateProfiles,dout->data.vptr,(int *)&dout->dArrayLength,FALSE);
          if (!strcmp(dout->dTag,"PrfCfgEx"))
            return populateProfileSet(oprViewFiles[devnr].stateProfiles,dout->data.vptr,(int *)&dout->dArrayLength,TRUE);
          return invalid_structure_tag;
        default:
          return illegal_format;
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"CONFIGURATION.NAM") || !stricmp(devProperty,"PROFILE.NAM") ||
      !stricmp(devProperty,"StatesInProfile.NAM") || !stricmp(devProperty,"ProceduresInProfile.NAM"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(dout->dFormat);
        break;
      default: 
        return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength && i<gNumOprViewFiles; i++)
    {
      strncpy((char *)(dout->data.bptr+i*len),oprViewFiles[i].desc,len);
    }
    if (i < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)i;
    return 0;
  }  
  if (!stricmp(devProperty,"CycleNumber"))  
  { 
    if (access & CA_WRITE) return illegal_read_write;
    return PutValuesFromLong(dout,&thisCycleNumber,1);
  }
  if (!stricmp(devProperty,"CycleNumber.Delta"))  
  { 
    if (access & CA_WRITE) return illegal_read_write;
    if ((access & CA_HIST) && !isUsingCycleTrigger) return not_applicable;
    return PutValuesFromLong(dout,&deltaCycleNumber,1);
  }
  if (!stricmp(devProperty,"RESTART"))  
  { /* dump database */
    if ((access & CA_WRITE) == 0) return illegal_read_write;
    if (!isAdministrator(usr)) 
    {
      char es[256];
      sprintf(es,"access denied: user %.16s is not an administrator",usr);
      SetEqmCompletion(STAEQM_TAG,es);
      return access_denied;
    }
    needsRestart = TRUE;
    return 0;
  }
  if (!strnicmp(devProperty,"CorrectionCounts",16))
  {
    time_t stop = time(NULL);
    time_t start = stop - 86400; 
    int statescor[256];
    if (dout->dArrayLength < 1) return dimension_error;
    if (din->dArrayLength > 0)
    {
      if (access & CA_WRITE) return illegal_read_write;
      if (din->dArrayLength < 2) return dimension_error;
      switch (din->dFormat)
      {
        case CF_INT32:
          start = (time_t)din->data.lptr[0];
          stop = (time_t)din->data.lptr[1];
          break;
        case CF_DOUBLE:
          start = (time_t)din->data.dptr[0];
          stop = (time_t)din->data.dptr[1];
          break;
        default:
          return illegal_format;
      }
    }
    if (stristr(devProperty,"States") != NULL)
    {
      if ((cc=getCorrectionCounts(start,stop, statescor, ndeclstates, STATE_CORR_CNTS)) != 0) return cc;
      n = ndeclstates;
    }
    else
    {
      if ((cc=getProcedureCorrectionCounts(start,stop, statescor, ndeclprocs)) != 0) return cc;
      n = ndeclprocs;
    }
    if (dout->dArrayLength > (UINT32)n) dout->dArrayLength = (UINT32)n;
    return PutValuesFromLongEx(dout,statescor,n,devnr);
  }
  if (!stricmp(devProperty,"Corrections"))
  {
    time_t stop = time(NULL);
    time_t start = stop - 86400; 
    if (din->dArrayLength > 0)
    {
      if (access & CA_WRITE)
      {
        StateCorrection *sc = (StateCorrection *)din->data.vptr;
        if (din->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(din->dTag,"StateCor")) return invalid_structure_tag;
        if (sc->start < stop - 86400*7)
        { // must be an admin to apply 'old' corrections
          SetEqmCompletion(STAEQM_TAG,"admin rights required for corrections older than 1 week");
          if (!isAdministrator(usr)) return admin_rights_required;
        }
        if (strcmp(din->dTag,"StateCor")) return invalid_structure_tag;
        cc = addStateCorrections((StateCorrection *)din->data.vptr); 
        if (cc == 0) gLastCorrectionTimeStamp = time(NULL);
        return cc;
      }
      if (din->dArrayLength < 2) return dimension_error;
      switch (din->dFormat)
      {
        case CF_INT32:
          start = (time_t)din->data.lptr[0];
          stop = (time_t)din->data.lptr[1];
          break;
        case CF_DOUBLE:
          start = (time_t)din->data.dptr[0];
          stop = (time_t)din->data.dptr[1];
          break;
        default:
          return illegal_format;
      }
    }
    if (dout->dArrayLength > 0)
    {
      int num = dout->dArrayLength;
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag,"StateCor")) return invalid_structure_tag;
      if ((cc=readStateCorrections(start,stop,(StateCorrection *)dout->data.vptr,&num)) != 0) return cc;
      if (num < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)num;
    }
    return 0;
  }
  if (!stricmp(devProperty,"Corrections.NAM"))
  {
    NAME64 n64[1] = { "Keyword" };
    return PutValuesFromString(dout,n64[0].name,CF_NAME64,1,0);
  }
  if (!stricmp(devProperty,"LastCorrectionTimestamp"))
  {
    int val = (int)gLastCorrectionTimeStamp;
    if (IsStringFormat(dout->dFormat))
    {
      return PutValuesFromString(dout, ctime(&gLastCorrectionTimeStamp), CF_TEXT, 24, 0);
    }
    return PutValuesFromLongEx(dout, &val, 1, 0);
  }
  if (!stricmp(devProperty,"LastError"))
  {
    return PutValuesFromString(dout,stateValidationErr,CF_TEXT,1024,0);
  }

  return illegal_property;
}
char thisFecName[FEC_NAME_SIZE+4] = "";
char thisContext[CONTEXT_NAME_SIZE+4] = "";
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE+4] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
UINT16 thisPortOffset = 0;
void PreSystemInit(void)
{
  FeclogDepth = 2000;
  if (strlen(thisContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      strcat(thisFecName,"STAFEC");
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s State server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
}
void PostSystemInit(void)
{
  int cc;
  DTYPE dout, din;
  HistorySpecification hspc;

  gNumParticleTypes = getNumParticleTypesFromFile();
  if (gNumParticleTypes == 0)
  {
    feclog("no particle types in state.csv file (or file not found)!");
    gNumParticleTypes = 1;
  }

  RegisterStructs();

  signal(SIGFPE,SIG_IGN); /* catch floating point domain errors */
  gRequireAcknowledgments = 0; /* turn off ack. requirments for REFRESH mode */
  if ((cc=RegisterEquipmentModule(STAEQM_EXPORT,STAEQM_TAG,MAX_STATES+MAX_PROCS+MAX_MACHINE_STATES,staeqm,staini,stabkg,100,staexi)) != 0)
    feclog("Can't register %s: %s\n",STAEQM_TAG,erlst[cc]);

  RegisterProperty(STAEQM_TAG,"DECLSTATE",gNumParticleTypes,CF_NAME32,CA_READ,"declared state information");
  RegisterProperty(STAEQM_TAG,"DeclaredState",gNumParticleTypes,CF_NAME32,CA_READ,"declared state information");
  RegisterProperty(STAEQM_TAG,"LastDeclaredState", gNumParticleTypes, CF_NAME32, CA_READ, "state information at time of declaration");
  RegisterProperty(STAEQM_TAG,"PreviousState",gNumParticleTypes,CF_NAME32,CA_READ,"previous state information");
  RegisterProperty(STAEQM_TAG,"DECLPROC",gNumParticleTypes,CF_NAME32,CA_READ,"declared procedure information");
  RegisterProperty(STAEQM_TAG,"DeclaredProcedure",gNumParticleTypes,CF_NAME32,CA_READ,"declared procedure information");
  RegisterProperty(STAEQM_TAG,"PreviousProcedure",gNumParticleTypes,CF_NAME32,CA_READ,"previous procedure information");
  RegisterProperty(STAEQM_TAG,"STATECHANGE",1,CF_SHORT,CA_WRITE,"change state information");
  RegisterProperty(STAEQM_TAG,"TIMESTATS",3,CF_LONG,CA_WRITE,"state information");
  RegisterProperty(STAEQM_TAG,"LastError",1024,CF_TEXT,CA_READ,"last state database error");

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_LONG;
  dout.dArrayLength = MAX_STATES;
  RegisterPropertyInformation(STAEQM_TAG,"TIMECOUNTER",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state duration",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"BEAMCOUNTER",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state duration with beam",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"TESTCOUNTER",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state duration test conditions",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"EVENTCOUNTER",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]number of state declaration events",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"ERRCOUNTER",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state duration with errors (unavailable)",-1,NULL);
  dout.dArrayLength = MAX_PROCS;
  RegisterPropertyInformation(STAEQM_TAG,"TIMECOUNTER.Procedure",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state procedure duration",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"BEAMCOUNTER.Procedure",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state procedure duration with beam",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"TESTCOUNTER.Procedure",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state procedure duration test conditions",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"EVENTCOUNTER.Procedure",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]number of procedure declaration events",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"ERRCOUNTER.Procedure",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1E11 sec]state procedure duration with errors (unavailable)",-1,NULL);
  dout.dArrayLength = 1;
  RegisterPropertyInformation(STAEQM_TAG,"NumParticleTypes",
           &dout,NULL,CA_READ,AT_SCALAR,0,"[0:2 ]number of particle types in the machine",-1,NULL);
  
  RegisterProperty(STAEQM_TAG,"TIMECOUNTER.Procedure.NAM",MAX_PROCS,CF_NAME64,CA_READ,"state procedure names");
  RegisterProperty(STAEQM_TAG,"BEAMCOUNTER.Procedure.NAM",MAX_PROCS,CF_NAME64,CA_READ,"state procedure names");
  RegisterProperty(STAEQM_TAG,"TESTCOUNTER.Procedure.NAM",MAX_PROCS,CF_NAME64,CA_READ,"state procedure names");
  RegisterProperty(STAEQM_TAG,"EVENTCOUNTER.Procedure.NAM",MAX_PROCS,CF_NAME64,CA_READ,"state procedure names");
  RegisterProperty(STAEQM_TAG,"ERRCOUNTER.Procedure.NAM",MAX_PROCS,CF_NAME64,CA_READ,"state procedure names");
  RegisterProperty(STAEQM_TAG,"PROCSTATES.NAM",MAX_PROCS,CF_NAME64,CA_READ,"state procedure names");

  RegisterProperty(STAEQM_TAG,"CALLER",2,CF_NAME16,CA_READ,"last caller to change state");

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_NUM_CONFIGS;
  strcpy(dout.dTag,"McaCfg");
  RegisterPropertyInformation(STAEQM_TAG,"CONFIGURATION",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"registered viewer configuration set",-1,NULL);
  RegisterProperty(STAEQM_TAG,"CONFIGURATION.NAM",MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_NUM_CONFIGS;
  strcpy(dout.dTag,"PrfCfg");
  RegisterPropertyInformation(STAEQM_TAG,"PROFILE",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"registered viewer profile set",-1,NULL);
  strcpy(dout.dTag,"PrfCfgEx");
  RegisterPropertyInformation(STAEQM_TAG,"PROFILE",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"registered viewer profile set",-1,NULL);
  RegisterProperty(STAEQM_TAG,"PROFILE.NAM",MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_STATES;
  strcpy(dout.dTag,"DclState");
  RegisterPropertyInformation(STAEQM_TAG,"STATELIST",
           &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"registered states",-1,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_PROCS;
  strcpy(dout.dTag,"DclProc");
  RegisterPropertyInformation(STAEQM_TAG,"PROCLIST",
           &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"registered state procedures",-1,NULL);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 10;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  RegisterPropertyInformation(STAEQM_TAG,"PROCSTATES",
           &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"composite states for registered state procedures",-1,NULL);

  dout.dArrayLength = 256;
  RegisterPropertyInformation(STAEQM_TAG,"StatesInProcedure",
           &dout,&dout,CA_READ,AT_UNKNOWN,0,"dressed array of procedures and the states contained within",-1,NULL);
  dout.dFormat = CF_XML;
  dout.dArrayLength = 2048;
  RegisterPropertyInformation(STAEQM_TAG,"StatesInProcedure",
           &dout,&dout,CA_READ,AT_UNKNOWN,0,"dressed array of procedures and the states contained within",-1,NULL);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 256;
  RegisterPropertyInformation(STAEQM_TAG,"StatesInProfile",
           &dout,&dout,CA_READ,AT_UNKNOWN,0,"dressed array of profiles and the states contained within",-1,NULL);
  dout.dFormat = CF_XML;
  dout.dArrayLength = 2048;
  RegisterPropertyInformation(STAEQM_TAG,"StatesInProfile",
           &dout,&dout,CA_READ,AT_UNKNOWN,0,"dressed array of profiles and the states contained within",-1,NULL);
  RegisterProperty(STAEQM_TAG,"StatesInProfile.NAM",MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 256;
  RegisterPropertyInformation(STAEQM_TAG,"ProceduresInProfile",
           &dout,&dout,CA_READ,AT_UNKNOWN,0,"dressed array of profiles and the procedures contained within",-1,NULL);
  dout.dFormat = CF_XML;
  dout.dArrayLength = 2048;
  RegisterPropertyInformation(STAEQM_TAG,"ProceduresInProfile",
           &dout,&dout,CA_READ,AT_UNKNOWN,0,"dressed array of profiles and the procedures contained within",-1,NULL);
  RegisterProperty(STAEQM_TAG,"ProceduresInProfile.NAM",MAX_NUM_CONFIGS,CF_NAME32,CA_READ,"configuration names");

  RegisterPropertyInformation(STAEQM_TAG,"RESTART",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"restart the server's equipment modules",-1,NULL);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  RegisterPropertyInformation(STAEQM_TAG,"CycleNumber",&dout,NULL,CA_READ,AT_SCALAR,0,"current cycle number for context",-1,NULL);
  RegisterPropertyInformation(STAEQM_TAG,"CycleNumber.Delta",&dout,NULL,CA_READ|CA_HIST,AT_SCALAR,0,"delta (current - last) cycle number for context",-1,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_CORRECTION_PER_DAY;
  strcpy(dout.dTag,"StateCor");
  din.dFormat = CF_INT32;
  din.dArrayLength = 2;
  RegisterPropertyInformation(STAEQM_TAG,"Corrections",
           &dout,&din,CA_READ,AT_UNKNOWN,0,"get state corrections",-1,NULL);

  DTYPEZERO(dout);
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"StateCor");
  RegisterPropertyInformation(STAEQM_TAG,"Corrections",
           &dout,&din,CA_WRITE,AT_UNKNOWN,0,"add new state correction",-1,NULL);

  DTYPEZERO(dout);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  RegisterPropertyInformation(STAEQM_TAG, "LastCorrectionTimestamp",
    &dout, NULL, CA_READ, AT_UNKNOWN, 0, "last correction timestamp as UTC", -1, NULL);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = MAX_STATES;
  DTYPEZERO(din);
  din.dFormat = CF_INT32;
  din.dArrayLength = 2;
  RegisterPropertyInformation(STAEQM_TAG,"CorrectionCounts.States",
           &dout,&din,CA_READ,AT_CHANNEL,0,"get state correction Counts over a range",-1,NULL);

  dout.dArrayLength = MAX_PROCS;
  RegisterPropertyInformation(STAEQM_TAG,"CorrectionCounts.Procedures",
           &dout,&din,CA_READ,AT_CHANNEL,0,"get procedure correction Counts over a range",-1,NULL);

  RegisterProperty(STAEQM_TAG,"Corrections.NAM",1,CF_NAME64,CA_READ,"independent of keyword name");

  memset(&hspc,0,sizeof(hspc));
  hspc.pollingRate = 1000;
  hspc.archiveRate = 1000;
  hspc.heartbeat = 1200;
  hspc.depthShort = 10;
  hspc.depthLong = 3;
  AppendHistoryInformation(STAEQM_TAG,"DeclaredState","#0",1,CF_NAME64,1,&hspc);
} 
int logproblem(char *text,...)
{
  int cc=0;
  double ts = makeDataTimeStamp();
  char str[256];
  FILE *fp = NULL;
  va_list args;
  if (strlen(text) > 80) ccerr(string_too_long);
  va_start(args,text);
  sprintf(str,"%s : ",getDataTimeString(ts,0));
  vsprintf(&str[strlen(str)],text,args);
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  va_end(args);
  if ((fp=fopen("problems.log","a")) == NULL) ccerr(file_error);
  fwrite(str,strlen(str),1,fp);
err:
  if (fp) fclose(fp);
  return cc;
}
FILE *createAndOpen(char *fname,char *attr)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  static char strdel[2] = { FS_DELIMITER , 0};

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,strdel); c != NULL; c=strtok(NULL,strdel))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return fopen(fname,attr);
    }
    strcat(dirname,strdel);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) return NULL;
    }
#   elif defined(WIN32)
    sprintf(fn,"%.120s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return NULL;
    }
#   endif
  }
  return NULL;
}
int getProfileIndexFromName(char *profile)
{ // profile is the official profiles name (the description, which is used at the device name)
  int i;
  for (i=0; i<gNumOprViewFiles; i++)
  {
    if (!strnicmp(oprViewFiles[i].desc,profile,64)) return i;
  }
  return -1;
}
// profile is a view name and each profile/view will contain slices.
// each slice will contain procedures.
// each procedure will contain states.
int validateProfileCompleteness(char *profile)
{
  int si, pi, idx = getProfileIndexFromName(profile);
  int cc=0, i, k, p, np=100, valid = TRUE;
  OprProfile prof[100];
  NAME64 lst[50];
  char txt[64];
  int lstsiz = 50, slen;

  if ((cc=populateProfileSet(oprViewFiles[idx].stateProfiles,prof,&np,FALSE)) != 0) ccerr(file_error);
  memset(stateRepCounts,0,256);
  for (p=0,k=0; p<np; p++)
  {
    lstsiz = 50;
    if ((cc=getProfileProcList(prof[p].states,lst,&lstsiz)) != 0) ccerr(file_error);
    for (i=0; i<lstsiz; i++) 
    {
      if ((pi=getDeclProcIndex(lst[i].name)) < 0) continue;
      for (k=0; k<declProcs[pi].nstates; k++)
      {
        if ((si=getDeclStateIndex(declProcs[pi].statenames[k].name)) < 0) 
        {
          feclogEx(TINE_LOGLEVEL_ERROR,"profile %.64s : procedure %.64s uses state %.64s, which is not in the state database!",
            profile,declProcs[pi].devname,declProcs[pi].statenames[k].name);
          ccerr(code_failure);
        }
        if (si >= MAX_STATES)
        {
          feclogEx(TINE_LOGLEVEL_ERROR,"state %.64s (#%d) exceeds maximum state allocation!",declStates[si].devname,si);
        }
        if (stateRepCounts[si] < 255) stateRepCounts[si]++;
      }
    }
  }
err:
  sprintf(stateValidationErr,"profile %.64s : ",profile);
  slen = (int)strlen(stateValidationErr);
  if (cc)
  {
    sprintf(&stateValidationErr[slen],"%.32s",cc2str(cc));
    return FALSE;
  }
  for (i=0,k=0; i<ndeclstates && i<MAX_STATES; i++)
  {
    if (stateRepCounts[i] != 1)
    {
      if (stateRepCounts[i] == 0) sprintf(txt,"not represented");
      else sprintf(txt,"double counted (%d times)",stateRepCounts[i]);
      feclogEx(TINE_LOGLEVEL_ERROR,"profile %.64s : STATE %.64s is %s",profile,declStates[i].devname,txt);
      if ((slen=(int)strlen(stateValidationErr)) < 960)
      {
        sprintf(&stateValidationErr[slen],"%s%.64s is %s",(k++>0?", ":""),declStates[i].devname,txt);
      }
      valid = FALSE;
    }
  }
  if (valid) strcat(stateValidationErr," is a valid profile");
  return valid;
}

