/*
 * File apceqm.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file APCEQM.C
 *
 *  A legacy archived file contains the following information according the
 *	following protocol pro record :  (assume VERSION = 3)
 *
 *      DeviceServerName - 16 bytes   char string
 *      Property         - 32 bytes   char string
 *      DeviceName       - 16 bytes   char string
 *      size of call     - 2  bytes   short
 *      data format      - 2  bytes   short
 *      scale            - 4  bytes   float
 *      shift            - 4  bytes   float
 *      data buffer - N bytes according to format and size
 *
 *
 *  A modern archived file contains the following information according the
 *	following protocol:
 *
 *     Version		      - 16 bytes   char string (storage version = EVENT STORAGE version)
 * + (pro record):  (VERSION 4.00)
 *     Context          - 32 bytes   char string
 *     ServerName       - 32 bytes   char string
 *     Property         - 64 bytes   char string
 *     DeviceName       - 64 bytes   char string
 *     Tag              - 16 bytes   char string
 *     status buffer    - 32 bytes   char string (short status string)
 *     size of call     - 4  bytes   int32
 *     data format      - 2  bytes   int16
 *     status           - 2  bytes   int16
 *     scale            - 4  bytes   float
 *     shift            - 4  bytes   float
 *     data buffer - N bytes according to format and size
 *      
 * + (pro record):  (VERSION 5.00)
 *     Context          - 32 bytes   char string
 *     ServerName       - 32 bytes   char string
 *     Property         - 64 bytes   char string
 *     DeviceName       - 64 bytes   char string
 *     Tag              - 16 bytes   char string
 *     status buffer    - 32 bytes   char string (short status string)
 *     size of call     - 4  bytes   int32
 *     data format      - 2  bytes   int16
 *     status           - 2  bytes   int16
 *     scale            - 4  bytes   float
 *     shift            - 4  bytes   float
 *     UTC timestamp    - 8  bytes   double
 *     system stamp     - 4  bytes   int32  (system stamp: cycle number/pulse number)
 *     user   stamp     - 4  bytes   int32  (user data stamp: mostly = 0)
 *     data buffer - N bytes according to format and size
 *      
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "tine.h"
#ifdef UNIX
#include <dirent.h>
#endif
#include "arctools.h"
#include "apceqm.h"
#include "arceqm.h"

#define quit(i) { dbglog("Register struct: out of memory"); exit(i); }
void registerAPCBLK(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,sessionID),1,CF_LONG,"sessionID")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,event),1,CF_LONG,"event")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,ext),4,CF_TEXT,"ext")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,expName),EXPORT_NAME_SHORTSIZE,CF_TEXT,"expName")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,devName),DEVICE_NAME_SHORTSIZE,CF_TEXT,"devName")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,devProperty),PROPERTY_NAME_SHORTSIZE,CF_TEXT,"devProperty")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,dSize),1,CF_SHORT,"dSize")) quit(1);
  if (addFieldToStruct("APCBLK",OFFSETIN(APCBLK5,dFormat),1,CF_SHORT,"dFormat")) quit(1);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("APCBLK",sizeof(APCBLK5),1)) quit(1);

  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,sessionID),1,CF_LONG,"sessionID")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,event),1,CF_LONG,"event")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,ext),16,CF_TEXT,"ext")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,ctxName),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,expName),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,devName),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,devProperty),PROPERTY_NAME_SIZE,CF_TEXT,"property")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,dTag),TAG_NAME_SIZE,CF_TEXT,"dTag")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,dSize),1,CF_LONG,"dSize")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,dFormat),1,CF_SHORT,"dFormat")) quit(1);
  if (addFieldToStruct("APCBLKr4",OFFSETIN(APCBLK,reserved),1,CF_SHORT,"reserved")) quit(1);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("APCBLKr4",sizeof(APCBLK),1)) quit(1);
}

#ifdef WIN32
char ArchiveDirectoryPath[128] = "..\\CACHE";
#else
char ArchiveDirectoryPath[128] = "../CACHE";
#endif
ARCHRECORD *ArchRecord = NULL;
int numArchiveTypes = 0;
SESSION session[MAXSESSION];
int numDataSets = 0;
#define session_not_active 513

void set_yday(struct tm *t)
{
  int daysInMonth[12]={31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
  int i;
  t->tm_yday = 0;
  for (i = 0; i < t->tm_mon; i++ ) t->tm_yday += daysInMonth[i];
  if (( t->tm_mon > 1 ) && ( (t->tm_year % 4) == 0 )) t->tm_yday += 1;
  t->tm_yday += t->tm_mday - 1;
}
long tm2time(struct tm * t)
{
  long y, d, s;
  y = t->tm_year - 70;          /* Anzahl der vollen Jahre ab 1.1.1970 */
  d = y * 365 + t->tm_yday;     /* Anzahl der vollen Tage */
  d += (t->tm_year-1) / 4 - 17; /* Schaltjahre ab 1970 */
  s = ((d * 24 + t->tm_hour) * 60 + t->tm_min) * 60 + t->tm_sec;
  return s;
}
long unixtime(int yr,int mm,int da,int hr,int mn,int sc,int dl)
{
  long s;
  struct tm t;

  t.tm_sec = sc;
  t.tm_min = mn;
  t.tm_hour = hr;
  t.tm_mday = da;
  t.tm_mon = mm;
  t.tm_year = yr;
  t.tm_isdst = 0;
  set_yday(&t);

  s  = tm2time(&t);
  if (dl) s -= 3600L;
  s -= 3600L;        /* convert to continental Europe */
  return s;
}
/** 
 * \internal sort list to be in increasing order (earlist file at position 0)
 * files are mostly in order, so this should be quick
 */
void fileSort(NAME8 *fbuf,int start,int num) 
{                                            
  int i;

  if ((num-start) == 1) return;
  for (i=start+1; i<num; i++)
    if ( strncmp(fbuf[i].name,fbuf[start].name,8) < 0 ) fswap(fbuf,start,i);
  fileSort(fbuf,++start,num);
  return;
}
void fswap(NAME8 *fbuf,int i,int j)
{
  static char ftemp[13];
  strncpy(ftemp,fbuf[i].name,8); 
  strncpy(fbuf[i].name,fbuf[j].name,8); 
  strncpy(fbuf[j].name,ftemp,8); 
  return;
}
int getArchiveContracts(void)
{
  FILE *fp = NULL;
  int cc=0,done=0,i=0,n=0;
  int ext_col,trg_col;
  char str[256],hdr[256];

  if ((n=GetDataBaseSize(archiveListFile)) <= 0) ccerr(-n);
  if (n < MAXARCHIVES) n = MAXARCHIVES;
  if (ArchRecord != NULL) free(ArchRecord);
  if ((ArchRecord=(ARCHRECORD *)calloc(n,sizeof(ARCHRECORD))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(archiveListFile,"r")) == NULL) ccerr(no_such_file);
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      done = 1;
      str[255] = 0; strcpy(hdr,str);
      if ((ext_col=findcol(hdr,"EXTENSION")) < 0) ccerr(no_such_column);
      if ((trg_col=findcol(hdr,"TRIGGER")) < 0 && /* preferred header */
          (trg_col=findcol(hdr,"NAME")) < 0 &&    /* first alternative */
          (trg_col=findcol(hdr,"SERVER")) < 0) ccerr(no_such_column);
      continue;
    }
    strncpy(ArchRecord[i].ext,colptr(ext_col,str),16);
    strncpy(ArchRecord[i].trg,colptr(trg_col,str),16);
    if (++i >= n) 
    {
      feclog("Number of archive files is at capacity (%d) !",n);
      break;
    }
  }
  err:
  if (fp) fclose(fp);
  feclog("Archive Contracts file %s : %s",archiveListFile, cc ? erlst[cc] : "found");
  if (i > 0) strncpy(ArchiveDirectoryPath,stdArchiveDirectory,79);
  return i;
}
int getFiles(NAME8 *filebuf,char *fileext,time_t starttime,time_t stoptime,int bufsiz)
{
  int i, n, len, fctr=0, rmvctr=0;
  char dirname[128],fname[32],fullname[192],cmtname[192],fx16[32];
  time_t ftim;
  struct stat fs;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fp;
  long hfp;
  char fn[128];
# endif
  struct tm *t;
  int startyr, startmo, stopyr, stopmo, yr, mo;
  time_t now = time(NULL);
  char str[256];

  if ((t=localtime(&starttime)) == NULL) return 0;
  startyr = t->tm_year + 1900; startmo = t->tm_mon;
  if ((t=localtime(&stoptime)) == NULL) return 0;
  stopyr = t->tm_year + 1900; stopmo = t->tm_mon;
  n = 12 * (stopyr-startyr) + stopmo - startmo + 1;
  strncpy(fx16,fileext,16); fx16[16] = 0;
  if (strlen(fx16) == 0) return 0;
  for (fctr=0,i=0; i<n; i++)
  {
    yr = startyr + (startmo+i)/12;
    mo = (startmo+i)%12 + 1;
#   if defined(UNIX)
    sprintf(dirname,"%.64s/%d/%02d/%.16s",ArchiveDirectoryPath,yr,mo,fx16);
    if ((dp=opendir(dirname)) == NULL) continue;
    while ((dirp=readdir(dp)) != NULL)
    {
      if (strstr(dirp->d_name,fx16) == NULL) continue;
      len = strchr(dirp->d_name,'.') - dirp->d_name;
      strncpy(fname,dirp->d_name,8); fname[8] = 0;
      if (!strcmp(fname,".") || !strcmp(fname,"..")) continue;
      ftim = strtol(fname,0,16);
      if (ftim < starttime || ftim > stoptime) continue;
      if (starttime == 0 && filebuf == NULL && bufsiz == -1) 
      { /* signal for removal ! */
        sprintf(fullname,"%.128s/%.32s",dirname,dirp->d_name);
        strncpy(cmtname,fullname,192);
        len = strlen(cmtname);
        strcpy(&cmtname[len-3],"CMT");
        if (stat(cmtname,&fs) == 0)
        { /* did find a comment file */
          if (fs.st_mtime > ftim + 30) 
          { /* and someone has updated it -> they actually looked at it ! (saved by the bell!) */
            continue;
          }
          remove(cmtname);
        }
        remove(fullname);
        rmvctr++;
        if (rmvctr < 50)
        {
          sprintf(str,"removing un-examined %.16s file from %.24s",fx16,ctime(&ftim));
          AppendHistoryLog(str);
          feclog(str);
        }
      }
      if (filebuf != NULL && fctr < bufsiz) strncpy(filebuf[fctr].name,fname,8);
      fctr++;
    }
    if (rmvctr >= 50)
    {
      sprintf(str,"removed additional %d un-examined event files",rmvctr-50);
      AppendHistoryLog(str);
      feclog(str);
    }
    closedir(dp);
#   elif defined(WIN32)
    sprintf(dirname,"%s\\%d\\%02d\\%.16s",ArchiveDirectoryPath,yr,mo,fx16);
    sprintf(fn,"%s\\*.*",dirname);
    if ((hfp=(long)_findfirst(fn, &fp)) == -1) continue;
    do
    {
      if (strstr(fp.name,fx16) == NULL) continue;
      len = (int)(strchr(fp.name,'.') - fp.name);
      strncpy(fname,fp.name,8); fname[8] = 0;
      ftim = strtol(fname,0,16);
      if (ftim < starttime || ftim > stoptime) continue;
      if (starttime == 0 && filebuf == NULL && bufsiz == -1)
      { /* signal for removal */
        sprintf(fullname,"%.128s\\%.16s",dirname,fp.name);
        len = (int)strlen(cmtname);
        strcpy(&cmtname[len-3],"CMT");
        if (stat(cmtname,&fs) != 0)
        { /* did find a comment file */
          if (fs.st_mtime > ftim + 30) 
          { /* someone has updated it -> they actually looked at it ! (saved by the bell!) */
            continue;
          }
          remove(cmtname);
        }
        remove(fullname);
        rmvctr++;
        if (rmvctr < 50)
        {
          sprintf(str,"removing un-examined %.16s file from %.24s",fx16,ctime(&ftim));
          AppendHistoryLog(str);
          feclog(str);
        }
      }
      if (filebuf != NULL && fctr < bufsiz) strncpy(filebuf[fctr].name,fname,8);
      fctr++;
    } while (_findnext(hfp,&fp) != -1);
    if (rmvctr >= 50)
    {
      sprintf(str,"removed additional %d un-examined event files",rmvctr-50);
      AppendHistoryLog(str);
      feclog(str);
    }
#   endif
  }
  if (filebuf != NULL && fctr > 0) fileSort(filebuf,0,MIN(fctr,bufsiz));
  return fctr;
}
void apcinit(void)
{
  DTYPE dout, din;

  RegisterProperty(APCEQM_TAG,"QUERY.FILES",100,CF_NAME8,CA_READ,"query file names");
  RegisterProperty(APCEQM_TAG,"QUERY.NFILES",1,CF_SHORT,CA_READ,"Query for number of files");
  RegisterProperty(APCEQM_TAG,"QUERY.EQP",100,CF_NAME16,CA_READ,"Query for equipment modules");
  RegisterProperty(APCEQM_TAG,"QUERY.PROPS",100,CF_NAME32,CA_READ,"Query for properties");
  RegisterProperty(APCEQM_TAG,"OPEN",1,CF_SHORT,CA_READ,"open session");
  RegisterProperty(APCEQM_TAG,"OPEN/CMT",1024,CF_TEXT,CA_READ,"read comment");
  RegisterProperty(APCEQM_TAG,"OPEN.CMT",1024,CF_TEXT,CA_READ,"read comment");
  RegisterProperty(APCEQM_TAG,"ANNOTATE",1024,CF_TEXT,CA_WRITE,"write comment file");
  RegisterProperty(APCEQM_TAG,"CLOSE",1,CF_SHORT,CA_READ,"close session");
  RegisterProperty(APCEQM_TAG,"REMOVE",1,CF_SHORT,CA_WRITE,"remove record");

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"APCBLK");
  dout.dFormat = CF_BYTE;
  dout.dArrayLength = 4096;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(APCEQM_TAG,"READDATA",&dout,&din,CA_READ,AT_SCALAR,0,"get archive data",-1,NULL);
  RegisterPropertyInformation(APCEQM_TAG,"READDATA.RAW",&dout,&din,CA_READ,AT_SCALAR,0,"get raw archive data",-1,NULL);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"APCBLKr4");
  dout.dFormat = CF_BYTE;
  dout.dArrayLength = 4096;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(APCEQM_TAG,"READDATA",&dout,&din,CA_READ,AT_SCALAR,0,"get archive data",-1,NULL);
  RegisterPropertyInformation(APCEQM_TAG,"READDATA.RAW",&dout,&din,CA_READ,AT_SCALAR,0,"get raw archive data",-1,NULL);

  RegisterDeviceName(APCEQM_TAG,"keyword",0);

  registerAPCBLK();

  return;
}
void apceqm_exit(void)
{
  freeSessions();
  freeArbTbl();
}
void apctask()
{
  return;
}
void checkSessionActivity(void)
{
  static time_t interim = 0;
  time_t t = time(NULL);
  int i;

  if (t < interim + 3600) return; // check once per hour
  interim = t;
  if (archDebug) dbglog("checking session activity");
  for (i=0; i<MAXSESSION; i++)
  {
    if (session[i].active)
    {
      if (archDebug) dbglog("session %d active since %s",i,ctime(&session[i].active));
      if (t > session[i].active + 1800)
      {
        session[i].count = 0;
        freeSession(i);
      }
    }
  }

  return;
}
void freeApcList(ApcRecord **apcLst)
{
  ApcRecord *apc = *apcLst, *Next;

  if (apcLst == NULL) return;
  for (apc = *apcLst; apc != NULL; apc = Next)
  {
    Next = apc->Next;
    free(apc);
  }
  *apcLst = NULL;
  return;
}
void freeSessions(void)
{
  int i;
  for (i=0; i<MAXSESSION; i++)
  {
    session[i].count = 0;
    freeSession(i);
  }
  return; 
}
int GetLocalSessionID(UINT32 eventNumber,char *ext)
{
  struct tm *t;
  time_t et = (time_t)eventNumber;
  int i, t_year;
  char fullName[256];
    
  if ((t=localtime(&et)) == NULL) return -invalid_data;
  t_year = t->tm_year;
  if (strlen(ext) == 0) return -string_expected;
  sprintf(fullName,"%x.%.16s",eventNumber,ext);
  if (makeFullName(fullName,ext)) return -no_such_file;
  if ((i=loadStoredData(t_year,fullName,ext)) < 0) return i;
  session[i].active = time(NULL);
  return i;
}
int getSessionID(char *fname)
{
  int i;
  time_t t = time(NULL);

  for (i=0; i<MAXSESSION; i++)
    if (session[i].active && !strcmp(session[i].fullName,fname))
    {
      if (t > session[i].active + 3)  /*kluge to trap UDP retries */
        session[i].count++;
      return i;
    }
  return -1;
}
int makeSessionID(char *fname)
{
  int i,idxLeastRecentlyUsed = -1;
  time_t leastRecentlyUsed,t = time(NULL);

  for (i=0; i<MAXSESSION; i++) if (!session[i].active)
  {
    strncpy(session[i].fullName,fname,80);
    session[i].active = session[i].est = time(NULL);
    session[i].count = 1;
    return i;
  }

  leastRecentlyUsed = t;
  for (i=0; i<MAXSESSION; i++)
  {
    if (session[i].active < leastRecentlyUsed) 
    {
      leastRecentlyUsed = session[i].active;
      idxLeastRecentlyUsed = i;
    }
  }
  i = idxLeastRecentlyUsed;
  if (i >= 0 && t > session[i].active + 60)
  {
    session[i].count = 0;
    freeSession(i);
    strncpy(session[i].fullName,fname,80);
    session[i].active = session[i].est = time(NULL);
    session[i].count = 1;
    return i;
  }
  feclog("APC: unable to create session id");
  dumpSessions();
  return -resources_exhausted;
}
void freeSession(int id)
{
  if (id < 0 || id > MAXSESSION) return;
  if (archDebug) dbglog("APC: free session %d (%d)",id,session[id].count);
  if (--session[id].count > 0) return;
  if (session[id].hFile) close(session[id].hFile);
  if (session[id].apc != NULL) freeApcList(&session[id].apc);
  memset(&session[id],0,sizeof(SESSION));
  return;
}
#ifdef UNIX
void itoa(int ival,char *aval,int radix)
{
  switch (radix)
  {
    default:
    case 10: sprintf(aval,"%d",ival); break;
    case 16: sprintf(aval,"%x",ival); break;
  }
}
#endif
int GetEventIdFromFileName(char *fname)
{
  char *c, fn[128], idstr[16];
  int id = 0;
  if (fname == NULL) return 0;
  strncpy(fn,fname,128);
  if ((c=strrchr(fn,'.')) == NULL) return 0;
  *c = 0;
  strncpy(idstr,&fn[strlen(fn)-8],8);
  id = strtol(idstr,NULL,16);
  return id;
}
int makeFullName(char *fname,char *ext)
{
  char dst[128],name[32];
  int i,year,month;
  time_t now = time(NULL);
  struct tm *t;
  time_t ts;

  for (i=0; i<numArchiveTypes; i++) if (!strnicmp(ArchRecord[i].ext,ext,16)) break;
  if (i == numArchiveTypes) 
  {
    dbglog("APC: can't make name for ext %.16s among %d entries",ext,numArchiveTypes);
    return -1;
  }
  strncpy(dst,stdArchiveDirectory,80);
  strncpy(name,fname,8); name[8] = 0;
  ts = (time_t)strtol(name,0,16);
  if ((t=localtime(&ts)) == NULL) 
  {
    dbglog("APC: cannot make local time for %ld ",(long)ts);
    return -1;
  }
  year = t->tm_year + 1900;
  month = t->tm_mon + 1;
  sprintf(fname,"%.128s/%d/%02d/%.16s/%.8s.%.16s",dst,year,month,ext,name,ext);
  return 0;
}
int RemoveEvents(UINT32 starttime, UINT32 stoptime, char *ext)
{
  char fullName[256];
  int i,n,xlen = 3;
  NAME8 evnnames[1024];

  n = getFiles(evnnames,ext,starttime,stoptime,1024);
  for (i=0; i<n; i++)
  {
    sprintf(fullName,"%.8s.%.16s",evnnames[i].name,ext);
    if (makeFullName(fullName,ext)) return -no_such_file;
    if (archDebug) dbglog("APC: remove file %s",fullName);
    remove(fullName);
    /* xlen = strlen(ext); */
    sprintf(&fullName[strlen(fullName)-xlen],"CMT");
    if (archDebug) dbglog("APC: remove file %s",fullName);
    remove(fullName);
  }
  return i;
}

int RemoveEvent(UINT32 evnt, char *ext)
{
  char fullName[256];
  int i=0,xlen = 3;

  sprintf(fullName,"%x.%.16s",evnt,ext);
  if (makeFullName(fullName,ext)) return no_such_file;
  if (archDebug) dbglog("APC: remove file %s\n",fullName);
  if (remove(fullName)) i++;
  /* xlen = strlen(ext); */
  sprintf(&fullName[strlen(fullName)-xlen],"CMT");
  if (archDebug) dbglog("APC: remove file %s\n",fullName);
  if (remove(fullName)) i++;
  return i < 2 ? 0 : file_error;
}
int populateApcRecord(int hFile,ApcRecord *apc,char *ctx,char *srv,int version)
{
  BYTE s[512];
  int ptr = 0, hdrsize;
  UINT16 ssiz;
  strncpy(apc->ctx,ctx,CONTEXT_NAME_SIZE);
  strncpy(apc->srv,srv,EXPORT_NAME_SIZE);
  switch (version)
  {
    default:
    case 3:
#     define FILE_PAYLOAD3_HEADER_SIZE 60
      hdrsize = FILE_PAYLOAD3_HEADER_SIZE;
      read(hFile,s,hdrsize);
      strncpy(apc->prp,(char *)s,PROPERTY_NAME_SHORTSIZE); ptr += PROPERTY_NAME_SHORTSIZE;
      strncpy(apc->dev,(char *)&s[ptr],DEVICE_NAME_SHORTSIZE); ptr += DEVICE_NAME_SHORTSIZE;
      memcpy(&ssiz,&s[ptr],2); ptr += 2;
      apc->siz = ssiz;
      memcpy(&apc->fmt,&s[ptr],2); ptr += 2;
      memcpy(&apc->scale,&s[ptr],4); ptr += 4;
      memcpy(&apc->shift,&s[ptr],4); ptr += 4;
      break;
    case 4:
    case 5:
#     define FILE_PAYLOAD4_HEADER_SIZE 192
#     define FILE_PAYLOAD5_HEADER_SIZE (FILE_PAYLOAD4_HEADER_SIZE+16)
      hdrsize = version == 4 ? FILE_PAYLOAD4_HEADER_SIZE : FILE_PAYLOAD5_HEADER_SIZE;
      read(hFile,s,hdrsize);
      strncpy(apc->prp,(char *)s,PROPERTY_NAME_SIZE); ptr += PROPERTY_NAME_SIZE;
      strncpy(apc->dev,(char *)&s[ptr],DEVICE_NAME_SIZE); ptr += DEVICE_NAME_SIZE;
      strncpy(apc->tag,(char *)&s[ptr],TAG_NAME_SIZE); ptr += TAG_NAME_SIZE;
      strncpy(apc->sts,(char *)&s[ptr],STATUS_SHORTSIZE); ptr += STATUS_SHORTSIZE;
      memcpy(&apc->siz,&s[ptr],4); ptr += 4;
      memcpy(&apc->fmt,&s[ptr],2); ptr += 2;
      memcpy(&apc->err,&s[ptr],2); ptr += 2;
      memcpy(&apc->scale,&s[ptr],4); ptr += 4;
      memcpy(&apc->shift,&s[ptr],4); ptr += 4;
      if (version == 4) break;
      memcpy(&apc->timestamp,&s[ptr],8); ptr += 8;
      memcpy(&apc->sysstamp,&s[ptr],4); ptr += 4;
      memcpy(&apc->usrstamp,&s[ptr],4); ptr += 4;
      break;
  }
  if (ptr != hdrsize) feclog("pointer mismatch (version %d)! : %d vs %d",version,ptr,hdrsize);
  return ptr;
}
int getStorageVersion(char *versionString)
{
  char *c;
  int ver = (int)strtod(versionString,&c);
  switch (ver)
  { /* valid version numbers: */
    case 4:
    case 5:
      return ver;
    default:
      break;
  }
  return 3;
}
/**
 * \internal
 * returns Session id on success
 * returns negative completion code on error 
 */
int loadStoredData(int year,char *fname,char *ext)
{
  int n=0, cc = 0, i, id = -1, ptr, ver = 0, ssize;
  char s[200];
  char expName[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE], *ctx = expName, *srv = &expName[CONTEXT_NAME_SIZE];
  char lastCtxName[CONTEXT_NAME_SIZE], lastSrvName[EXPORT_NAME_SIZE];
  long placeholder = VERSION_LENGTH;
  ApcRecord *apc;
  int expNameLength = CONTEXT_NAME_SIZE + EXPORT_NAME_SIZE;
  int srvNameLength = EXPORT_NAME_SIZE;

  if (archDebug) dbglog("APC: find session for %s (%d)",fname,year);
  if (isArchiveOngoing(fname)) return -operation_busy;
  if ((id = getSessionID(fname)) >= 0)
  { /* session already open */
    if (archDebug) dbglog("APC: %s attaching to session %d",fname,id);
    return id; 
  }
  if ((id = makeSessionID(fname)) < 0) ccerr(-id);
  strncpy(session[id].ext,ext,16);
  
  if ((session[id].hFile=open(fname,O_RDONLY)) == -1 )
    ccerr(no_such_file);

  if (read(session[id].hFile,s,VERSION_LENGTH) <= 0) ccerr(file_error);
  ver = getStorageVersion(s);
  if (ver < 4)
  { /* does not contain  version header (pointing to a server name */
    lseek(session[id].hFile,0,SEEK_SET);
    srvNameLength = expNameLength = EXPORT_NAME_SHORTSIZE;
    ctx = gDeviceContext;
    srv = expName;
    placeholder = 0;
  }
  lastCtxName[0] = lastSrvName[0] = 0;
  for (i=0; read(session[id].hFile,expName,expNameLength) > 0; i++)
  { /* read the context and server name from the top of the record */
    if (archDebug > 1) dbglog("APC: %d stored elements",numDataSets);
    /* Archive storage information is here: */
    if ((apc=(ApcRecord *)calloc(1,sizeof(ApcRecord))) == NULL) ccerr(out_of_local_memory);
    apc->Next = session[id].apc; /* start of current list */
    session[id].apc = apc;

    ptr = populateApcRecord(session[id].hFile,apc,ctx,srv,ver);
    if (ptr < 0) ccerr(-ptr);
    placeholder += ptr + expNameLength;

    if (apc->fmt == CF_STRUCT) 
    {
      if (strlen(apc->tag) == 0) strcpy(apc->tag,"DHS");
      if ((ssize=GetStructSize(apc->tag)) > 0)
      {
        if (strcmp(apc->tag,"AMSr4")) /* already taken into account !? */
          apc->siz *= ssize;
      }
      else
      { /* punt ... */
        apc->siz = sizeof(DHS);
      }
    }
    apc->ptr = placeholder; /* note where the data are */
    placeholder += apc->siz*fmtsizeof(apc->fmt);
    /* skip over the data */
    lseek(session[id].hFile,placeholder,SEEK_SET); 
  }
err:
  if (cc)
  {
    if (archDebug) dbglog("APC: load %s failed - %s",fname,erlst[cc]);
    if (id >= 0) freeSession(id);
    return -cc;
  }
  if (archDebug) dbglog("APC: searched %d records",i);
  return id;
}
int readCMT(char *s,char *fname,int size)
{
  FILE *fp = NULL;
  char cmtname[132],tmp[CMT_SIZE+8];
  int len,bytesread=0,cc = 0;
  short slen;

  strcpy(s,"No comment\n");
  if ((len=(int)strlen(fname)) > 128) ccerr(file_error);
  strncpy(cmtname,fname,128);
  strcpy(&cmtname[len-3],"CMT");
  if ((fp=fopen(cmtname,"rb")) == NULL)	ccerr(no_such_file);
  if (fread((char *)&slen,1,2,fp) == 0) ccerr(0);
  if (slen < size) size = slen;
  if (size > CMT_SIZE) size = CMT_SIZE;
  memset(tmp,0,CMT_SIZE+8);
  if (csvGetLine(tmp,CMT_SIZE,fp) == 0) ccerr(0);
  strncpy(s,tmp,size); s[size-1] = 0;
  for (bytesread=(int)strlen(s); csvGetLine(tmp,CMT_SIZE,fp); bytesread=(int)strlen(s))
  {
    if (strlen(tmp) + (unsigned int)bytesread > (unsigned int)size) break;
    strcat(s,tmp);
  }  
err:
  if (fp != NULL) fclose(fp);
  return cc == no_such_file ? 0 : cc;
}
int writeCMT(char *s,char *fname,int size)
{
  FILE *fp = NULL;
  char cmtname[132],tmp[CMT_SIZE+8];
  short len;
  int cc = 0;

  if ((len=(short)strlen(fname)) > 132) ccerr(file_error);
  strcpy(cmtname,fname);
  strcpy(&cmtname[len-3],"CMT");
  if ((fp=fopen(cmtname,"w")) == NULL)	ccerr(no_such_file);
  if ((len=(short)strlen(s)) > CMT_SIZE)
  {
    len = CMT_SIZE;
    s[CMT_SIZE] = 0;
  }
  memcpy(tmp,&len,2);
  strncpy(&tmp[2],s,len);
  if (fwrite(tmp,len+2,1,fp) != 1) ccerr(file_error);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
                          
int reformatArchiveData(void *dst,short dstFmt,void *src,short srcFmt,int size,float scale,float shift)
{
  int i;

  if (scale == 0) scale = 1;
  if (srcFmt == dstFmt && scale == 1 && shift == 0)
  {
   memcpy(dst,src,size*fmtsizeof(dstFmt));
   return 0;
  }
  if (srcFmt > CF_FLOAT && srcFmt < CF_FLTINT) return illegal_format;
  if (srcFmt > CF_DBLINT && srcFmt != CF_DBLDBL) return illegal_format;
  switch (dstFmt)
  {
    case CF_SHORT:
      for (i=0; i<size; i++) ((short *)dst)[i] = (short)(scale * SRCVAL(src,i,srcFmt) + shift);
      return 0;
    case CF_LONG:
      for (i=0; i<size; i++) ((long *)dst)[i] = (long)(scale * SRCVAL(src,i,srcFmt) + shift);
      return 0;
    case CF_FLOAT:
      for (i=0; i<size; i++) ((float *)dst)[i] = (float)(scale * SRCVAL(src,i,srcFmt) + shift);
      return 0;
    case CF_BYTE:
      for (i=0; i<size; i++) ((BYTE *)dst)[i] = (BYTE)(scale * SRCVAL(src,i,srcFmt) + shift);
      return 0;
    case CF_DOUBLE:
      for (i=0; i<size; i++) ((double *)dst)[i] = (double)(scale * SRCVAL(src,i,srcFmt) + shift);
      return 0;
    default:
      return illegal_format;
  }
}

int reformatArchiveDataEx(void *dst,short dstFmt,void *src,short srcFmt,int index, int size,float scale,float shift)
{
  int i;

  if (srcFmt == CF_IMAGE)
  {
    if (dstFmt != CF_IMAGE) return illegal_format;
    memcpy(dst,src,IMAGE_HDR_SIZE);
    ((DIMAGE *)dst)->frameBuffer = (BYTE *)src + IMAGE_HDR_SIZE;
    ((DIMAGE *)dst)->frameBufferSize = size;
    return 0;
  }
  if (scale == 0) scale = 1;
  if (srcFmt == dstFmt && scale == 1 && shift == 0 && index == 0)
  {
   memcpy(dst,src,size*fmtsizeof(dstFmt));
   return 0;
  }
  if (srcFmt > CF_FLOAT && srcFmt < CF_FLTINT) return illegal_format;
  if (srcFmt > CF_INTINT && srcFmt != CF_DBLDBL) return illegal_format;
  switch (dstFmt)
  {
    case CF_SHORT:
      for (i=0; i<size; i++)
       ((short *)dst)[i] = (short)(scale * SRCVAL(src,i+index,srcFmt) + shift);
       return 0;
    case CF_LONG:
      for (i=0; i<size; i++)
       ((long *)dst)[i] = (long)(scale * SRCVAL(src,i+index,srcFmt) + shift);
       return 0;
    case CF_FLOAT:
      for (i=0; i<size; i++)
       ((float *)dst)[i] = (float)(scale * SRCVAL(src,i+index,srcFmt) + shift);
       return 0;
    case CF_BYTE:
      for (i=0; i<size; i++)
       ((BYTE *)dst)[i] = (BYTE)(scale * SRCVAL(src,i+index,srcFmt) + shift);
       return 0;
    case CF_DOUBLE:
      for (i=0; i<size; i++)
       ((double *)dst)[i] = (double)(scale * SRCVAL(src,i+index,srcFmt) + shift);
       return 0;
    default:
       return illegal_format;
  }
}

int (*filterArchiveData)(char *prp,char *nam,BYTE *buf,int fmt,UINT32 len,int devNr) = NULL;
int apceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int i = 0, cc = 0, j, nlen, n, pass, dsize, ssiz, propFound = 0, ignore_device=FALSE;
  int napcdevs,apcstart,apcstop,apcidx,k,datapos,datalen;
  int t_year, evnId;
  char name[32],ext[32],fullName[256];
  ApcRecord *apc;
  time_t starttime = 0,stoptime = time(NULL);
  short id;
  APCBLK apcblk, *req;
  APCBLK5 *req5;
  int devNr = devName[0] == '#' ? atoi(&devName[1]) : 0;
  struct tm *t;
  time_t ts;
  char *c,*srv,tmpdevnam[DEVICE_NAME_SIZE+8],tmpprpnam[PROPERTY_NAME_SIZE+8],apcdevname[DEVICE_NAME_SIZE+8];
  float shift = 0.0, scale = 1.0;

  if (!stricmp(devProperty,"QUERY.FILES") || /* files of this extension  */
      !stricmp(devProperty,"QUERY.NFILES"))  /* how many files of this extension */
  {
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_LONG: 
           switch (din->dArrayLength)
           {
             case 2: stoptime  = din->data.lptr[1];
             case 1: starttime = din->data.lptr[0];
                     break;
             default: return illegal_read_write;
           }
           break;
        case CF_SHORT:
           switch (din->dArrayLength)
           {
             case 2: stoptime  = unixtime((int)(din->data.sptr[1]%1900),11,31,23,59,59,0);
             case 1: starttime = unixtime((int)(din->data.sptr[0]%1900),0,0,0,0,0,0);
                     break;
             default: return illegal_read_write;
           }
           break;
        default: return illegal_format;
      }
    }
    strncpy(ext,devName,DEVICE_NAME_SHORTSIZE); ext[DEVICE_NAME_SHORTSIZE] = 0;
    if (strlen(ext) == 0) return illegal_device_number;
    if (!stricmp(devProperty,"QUERY.FILES"))
    {
      switch (dout->dFormat)
      {
        case CF_NAME8:
          n = getFiles((NAME8 *)dout->data.vptr,ext,starttime,stoptime,(int)dout->dArrayLength);
          dout->dArrayLength = n;
          break;
        case CF_LONG:
          if (dout->dArrayLength > tmpWorkAreaSize/4) dout->dArrayLength = tmpWorkAreaSize/4;
          n = getFiles((NAME8 *)tmpWorkArea,ext,starttime,stoptime,(int)dout->dArrayLength);
          tmpdevnam[8] = 0;
          for (i=0; i<n && i<(int)dout->dArrayLength; i++)
          {
            strncpy(tmpdevnam,((NAME8 *)tmpWorkArea)[i].name,8);
            dout->data.lptr[i] = strtol(tmpdevnam,NULL,DEVICE_NAME_SHORTSIZE);
          }
          dout->dArrayLength = i;
          break;
        default:
          return illegal_format;
      }
    }
    else
    {
      n = getFiles((NAME8 *)NULL,ext,starttime,stoptime,0);
      putValuesFromLong(dout,(SINT32 *)&n,1);
    }    
    return n < 0 ? -n : 0;
  }
  if (!stricmp(devProperty,"QUERY.EQP")) /* the eqp modules for this session id */
  {
    if (din->dArrayLength != 1) return dimension_error;
    if (din->dFormat != CF_SHORT) return illegal_format;
    if ((id=din->data.sptr[0]-1) < 0 || id > MAXSESSION-1) return out_of_range;
    switch (dout->dFormat)
    {
      case CF_NAME32:
      case CF_NAME16:
        nlen = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    for (i=0,n=0; i<numDataSets; i++)
    {
      for (apc=session[id].apc; apc != NULL && n < (int)dout->dArrayLength; apc=apc->Next)
      {
        srv = dout->data.cptr+n*nlen;
        if (strncmp(srv,apc->srv,nlen)) 
        {
          strncpy(srv,apc->srv,nlen);
          n++;
        }
      }
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"QUERY.PROPS")) /* the properties for this session id and eqp */
  {
    if (din->dArrayLength != 1) return dimension_error;
    if (din->dFormat != CF_SHORT) return illegal_format;
    if ((id=din->data.sptr[0]-1) < 0 || id > MAXSESSION-1) return out_of_range;
    switch (dout->dFormat)
    {
      case CF_NAME64:
      case CF_NAME32:
      case CF_NAME16:
        nlen = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    for (i=0,n=0; i<numDataSets; i++)
    {
      for (apc=session[id].apc; apc != NULL && n < (int)dout->dArrayLength; apc=apc->Next)
      {
        if (strnicmp(devName,apc->srv,EXPORT_NAME_SIZE)) continue;
        strncpy(dout->data.cptr + n*nlen,apc->prp,nlen);
        n++;
      }
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!strnicmp(devProperty,"OPEN",4))
  {
    if (din->dArrayLength != 1) return no_file_space;
    switch (din->dFormat)
    {
      case CF_NAME8:
        strncpy(name,(char *)din->data.vptr,8); break;
      case CF_LONG:
        sprintf(name,"%x",din->data.lptr[0]); break;
      default: return illegal_format;
    }
    name[8] = 0;
    ts = strtol(name,0,16); 
    if ((t=localtime(&ts)) == NULL) return invalid_data;
    t_year = t->tm_year;
    strncpy(ext,devName,16);
    if (strlen(ext) == 0) return string_expected;
    sprintf(fullName,"%.8s.%.16s",name,ext);
    if (archDebug) dbglog("APC: make name for class %.16s",ext);
    if (makeFullName(fullName,ext)) return no_such_file;
    if (strstr(devProperty,"CMT")) return readCMT((char *)dout->data.vptr,fullName,dout->dArrayLength);
    if (archDebug) dbglog("APC: loading %s",fullName);
    if ((i=loadStoredData(t_year,fullName,ext)) < 0) return -i;
    session[i].active = time(NULL);
    if (dout->dFormat != CF_SHORT) return illegal_format;
    switch (dout->dArrayLength)
    {
      case 2: dout->data.sptr[1] = 1;
      case 1: dout->data.sptr[0] = i + 1;
              break;
      default: return illegal_read_write;
    }
    setDataTimeStamp((double)session[i].est);
    return 0;
  }
  if (!stricmp(devProperty,"ANNOTATE"))                       /* CF_STRUCT */
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_TEXT) return illegal_format;
    if (din->dArrayLength < 8) return no_file_space;
    strncpy(name,(char *)din->data.vptr,8);  name[8] = 0;
    strncpy(ext,devName,16);
    if (strlen(ext) == 0) return string_expected;
    sprintf(fullName,"%.8s.%.16s",name,ext);
    if (makeFullName(fullName,ext)) return no_such_file;
    return writeCMT(&((char *)din->data.vptr)[8],fullName,din->dArrayLength-8);
  }
  if (!stricmp(devProperty,"REMOVE"))                       /* CF_STRUCT */
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_CHAR8) return illegal_format;
    if (din->dArrayLength != 1) return no_file_space;
    strncpy(name,(char *)din->data.vptr,8);  name[8] = 0;
    strncpy(ext,devName,16);
    if (strlen(ext) == 0) return string_expected;
    sprintf(fullName,"%.8s.%.16s",name,ext);
    if (makeFullName(fullName,ext)) return no_such_file;
    if (archDebug) dbglog("APC: remove file %.128s\n",fullName);
    if (remove(fullName)) i++;
    sprintf(&fullName[strlen(fullName)-3],"CMT");
    if (archDebug) dbglog("APC: remove file %.128s\n",fullName);
    if (remove(fullName)) i++;
    return i < 2 ? 0 : file_error;
  }
  if (!stricmp(devProperty,"CLOSE"))
  {
    if (din->dArrayLength != 1) return dimension_error;
    if ((cc=getValuesAsShort(din,&id,1)) != 0) return cc;
    id--;
    if (id < 0 || id > MAXSESSION-1) return out_of_range;
    freeSession(id);
    return 0;
  }
  if (!stricmp(devProperty,"SESSION"))
  {
    dumpSessions();
    return 0;
  }
  if (!stricmp(devProperty,"FLUSH"))
  {
    freeSessions();
    return 0;
  }
  if (!stricmp(devProperty,"DEBUG"))
  {
    if (din->dArrayLength == 1 && din->dFormat == CF_SHORT)
      archDebug = din->data.sptr[0];
    else if (dout->dArrayLength == 1 && dout->dFormat == CF_SHORT)
      dout->data.sptr[0] = archDebug;
    else return illegal_read_write;
    return 0;
  }
  if (!strnicmp(devProperty,"READDATA",8))
  {
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (dout->dArrayLength == 0) return illegal_read_write;
    if (!strncmp(din->dTag,"APCBLK",TAG_NAME_SIZE)) 
    { /* legacy request */
      req5 = (APCBLK5 *)din->data.vptr;
      memset(&apcblk,0,sizeof(APCBLK));
      apcblk.sessionID = req5->sessionID;
      apcblk.event = req5->event;
      strncpy(apcblk.ext,req5->ext,4);
      strncpy(apcblk.ctxName,gDeviceContext,CONTEXT_NAME_SIZE);
      strncpy(apcblk.expName,req5->expName,EXPORT_NAME_SHORTSIZE);
      strncpy(apcblk.devName,req5->devName,DEVICE_NAME_SHORTSIZE);
      strncpy(apcblk.devProperty,req5->devProperty,PROPERTY_NAME_SHORTSIZE);
      apcblk.dSize = req5->dSize;
      apcblk.dFormat = req5->dFormat;
      req = &apcblk;
    }
    else if (!strncmp(din->dTag,"APCBLKr4",TAG_NAME_SIZE)) 
    { /* modern request */
      req = (APCBLK *)din->data.vptr;
      if (strlen(req->ctxName) == 0) strncpy(req->ctxName,gDeviceContext,CONTEXT_NAME_SIZE);
    }
    else
    {
      return invalid_structure_tag;
    }
    if (din->dArrayLength != 1) return dimension_error;
    
    if ((id=(short)req->sessionID-1) < 0) ccerr(not_implemented); /* call open session (do later) */
    if (id > MAXSESSION - 1) ccerr(no_file_space);    
err:
    if (cc != 0) /* try to get a new session slot if the request is faulty */
    {
      if (archDebug) dbglog("trying to recover from error %s",erlst[cc]);
      if ((id=GetLocalSessionID(req->event,req->ext)) < 0)
      {
        dbglog("unable to acquire a new session id : %s",erlst[-id]);
        return -id;
      }
    }
    setDataTimeStamp((double)session[id].est);

    if (!session[id].active) 
    {
      SetEqmCompletion(APCEQM_TAG,"session no longer active");
      return session_not_active;
    }
    session[id].active = time(NULL);  /* keep session alive */

    c = GetStoredPropertyFromAlias(req->expName,req->devProperty);
    if (c != NULL) strncpy(tmpprpnam,c,PROPERTY_NAME_SIZE); else tmpprpnam[0] = 0;

    strncpy(apcdevname,req->devName,DEVICE_NAME_SIZE); apcdevname[DEVICE_NAME_SIZE] = 0;
    if (apcdevname[0] != '#') 
    {
      napcdevs = 1;
    }
    else
    {
      apcstart = atoi(&apcdevname[1]);
      if ((c=strchr(&apcdevname[1],'#')) != NULL) apcstop = atoi(&c[1]);
      else apcstop = apcstart;
      napcdevs = apcstop - apcstart + 1;
      if ((c=strchr(apcdevname,':')) != NULL) apcidx = atoi(&c[1]);
      else apcidx = 0;
    }
    if (archDebug) dbglog("looking for %s %s %s",req->expName,req->devName,req->devProperty);
    if (!strcmp(req->devName,DEVICE_IGNORE)) ignore_device = TRUE;
    for (k=0,pass=0; k < napcdevs && pass < napcdevs; pass++)
    {
      if (napcdevs > 1) sprintf(apcdevname,"#%d",k+apcstart);
      for (apc=session[id].apc; apc != NULL; apc=apc->Next)
      {
        if (strnicmp(req->ctxName,apc->ctx,CONTEXT_NAME_SIZE)) continue;
        if (strnicmp(req->expName,apc->srv,EXPORT_NAME_SIZE)) continue;
        if (strnicmp(req->devProperty,apc->prp,PROPERTY_NAME_SIZE) &&
            strnicmp(tmpprpnam,apc->prp,PROPERTY_NAME_SIZE)) continue;
        propFound = 1;
        evnId = GetEventIdFromFileName(session[id].fullName);
        if (!ignore_device)
        { /* normal case: find the right device */
          if (apcdevname[0] == '#')
          {  /* number asked for, prepare alternative devname as name */
             j = atoi(&apcdevname[1]);
             c = GetStoredDeviceName(j,req->expName,evnId);
             if (c == NULL) return illegal_equipment_number; 
             strncpy(tmpdevnam,c,DEVICE_NAME_SIZE); tmpdevnam[DEVICE_NAME_SIZE]=0;
          }
          else 
          {  /* name asked for, prepare alternative devname as number */
             j = GetStoredDeviceNumber((NAME64 *)apcdevname,req->expName,evnId);
             if (j >= 0) sprintf(tmpdevnam,"#%d",j);
             else tmpdevnam[0] = 0;
          }
          if (archDebug > 1) dbglog("alternate device names %s and %s",apcdevname,tmpdevnam);
          if (strlen(apc->dev))
          {
            if (strnicmp(apc->dev,apcdevname,DEVICE_NAME_SIZE) && 
                strnicmp(apc->dev,tmpdevnam,DEVICE_NAME_SIZE)) continue;
          }
        }
        if (archDebug) dbglog("found candidate");
        if (apc->fmt == CF_STRUCT)
        { /* check a couple of things up front */
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          ssiz = GetStructSize(apc->tag);
          if (ssiz == 0) ssiz = 1; /* not registered locally ! */
          /* n.b. apc->siz systematically contains the full size in bytes ! */
          if (strncmp(apc->tag,req->dTag,TAG_NAME_SIZE))
          { /* requested tag is not the same as stored tag */
            if (strlen(apc->tag) > 0 && strlen(req->dTag) > 0) return invalid_structure_tag;
            if (apc->siz > dout->dArrayLength && (apc->siz % dout->dArrayLength) != 0) return invalid_structure_tag;
            if (apc->siz < dout->dArrayLength && (dout->dArrayLength % apc->siz) != 0) return invalid_structure_tag;
          }
        }
        else
        {
          ssiz = 1;
        }
        if (strlen(dout->dTag) == 0)
        { /* actual data request is not tagged ! */
          ssiz = 1; /* set this back, because dArrayLength will carry the size in bytes */
        }
        else 
        {
          if (strncmp(dout->dTag,apc->tag,TAG_NAME_SIZE)) return invalid_structure_tag;
        }
        if ((dsize=fmtsizeof(apc->fmt)*apc->siz) > ARCH_BUF_SIZE) return buffer_too_small;
        if (apc->siz < dout->dArrayLength) dout->dArrayLength = apc->siz;
        datapos = apc->ptr;
        datalen = dout->dArrayLength * ssiz;
        if (napcdevs > 1) 
        {
          dsize = fmtsizeof(apc->fmt);
          datapos += apcidx*dsize;
          datalen = 1 * ssiz;
        }
        if (lseek(session[id].hFile,datapos,SEEK_SET) == -1) return file_error;
        if (read(session[id].hFile,archiveBuffer,dsize) == -1) return file_error;
        if (filterArchiveData != NULL)
        {
          cc = (*filterArchiveData)(req->devProperty,apcdevname,archiveBuffer,apc->fmt,datalen,devNr);
          if (cc != 0) return cc;
        }

        if (strstr(devProperty,"RAW")) { scale = 0; shift = 0; }
        else { scale = apc->scale; shift = apc->shift; }
        cc = reformatArchiveData(dout->data.bptr+k*dsize,dout->dFormat,
                                 archiveBuffer,apc->fmt,datalen,scale,shift);
        if (cc) return cc;
        if (k++ > (int)dout->dArrayLength) return 0;
        pass = 0;
      }
    }
    if (k != 0) return 0;
    if (archDebug) dbglog("can't find %s %s %s",req->expName,req->devName,req->devProperty);
    return un_allocated;
  }
  return illegal_property;
}
