/*
 * File archbase.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "tine.h"
#include "archive.h"
#if defined(UNIX)
# include "errno.h"
# include "dirent.h"
# define ADBPATH "../database/"
# define ADPATH "../DATA"
#elif defined(WIN32)
# include <float.h>
# include "direct.h"
# define ADBPATH "..\\database\\"
# define ADPATH "..\\DATA"
#endif
#if defined(UNIX) && !defined(HAS_CHSIZE)
#define chsize(f,s) ftruncate(f,s)
#endif
#ifdef NEEDS_CHSIZE
extern int chsize(int hFile,SINT32 size);
#endif

char *ArchiveDBfile = NULL;
char *ArchiveDBpath = ADBPATH;
char *ArchiveDatapath = ADPATH; 
char *daqEqmName = NULL;
extern DAQRecord *daqTbl;
extern char cbuff[NETMEX_BUF_SIZE];
extern char cbuff[NETMEX_BUF_SIZE];
extern int archPending,totalNrRecords;
extern int archdebug, ramping, running, beam;
extern int g_isTerminating;
int debugon = 0;
void AppendProblemsLog(char *str);
int OpenDataFile(DAQRecord *daqr,int *hasAlternate);
int makeDirPath(char *fname);
int allocActiveOnly = FALSE;
int isArchiver = FALSE;
DAQParameter *getDaqParameter(char *key);
time_t getTimeStampFromLogEntry(char *str); /* from toolkit.c */
#ifdef WIN32
char *delStr = "\\";
#else
char *delStr = "/";
#endif
char *daqStockFilterNames[] = { "NEVER","ONCE","ALWAYS","FAST","SLOW","FIXTIME","HRT","STATUS", "VOLATILE", "NOPOI" };
char *daqStockFilterDescriptions[] = { "never archive (only repeat)","archive once (per day or restart)","always archive (when conditions warrant)",
  "archive at the given polling interval","archive no more often than once per minute","apply the local timestamp to the data",
  "archive with high resolution timestamp","archive the status code with data", "multi-channel record has volatile names list", 
  "do not keep points-of-interest information" };
int nDaqStockFilterNames = sizeof(daqStockFilterNames)/sizeof(daqStockFilterNames[0]);

DAQFilter *daqFilterTbl = NULL;
int nDaqFilterTblEntries = 0;

int nSubSystems = 0;
SUB_SYS *subSystems = NULL;
int nPrpAlias = 0;
PRP_ALIAS *prpAlias = NULL; /* 'associated' aliases */

UINT32 gValidFilterMask = 0;

int gNumVolatileDaqTblEntries = 0;
int gAllocDataBuffers = TRUE; /* the HISTORY server turns this off */

char daqUpdateErrStr[256] = "";

DaqRecKey *gDaqRecAssoc = NULL;
DaqRecKey *gDaqRecKeyTbl[DAQREC_HASH_SIZE];
DaqRecKey *getDaqRecKey(char *key)
{
  DaqRecKey *ll;
  int idx = (int)(ElfHash((unsigned char *)key) % DAQREC_HASH_SIZE);
  if (key == NULL) return NULL;
  for (ll=gDaqRecKeyTbl[idx]; ll != NULL; ll = ll->nxt)
  {
    if (!strncmp(ll->key,key,256)) return ll;
  }
  return NULL;   
}
void rmvDaqRecKey(char *key)
{
  DaqRecKey *ll, *ll0, *llp;
  int idx = (int)(ElfHash((unsigned char *)key) % DAQREC_HASH_SIZE);

  llp = ll0 = gDaqRecKeyTbl[idx];
  for (ll=ll0; ll != NULL; llp = ll, ll = ll->nxt)
  { /* run thru the list */
    if (!strncmp(ll->key,key,PROPERTY_NAME_SIZE))
    { /* that's the one to remove */
      if (ll == ll0) gDaqRecKeyTbl[idx] = ll->nxt;
      else llp->nxt = ll->nxt;
      free(ll);
      break;
    }
  }
}
void rmvDaqRecKeys(void)
{
  DaqRecKey *ll, *lln=NULL;
  int i;
  
  for (i=0; i<DAQREC_HASH_SIZE; i++)
  {
    for (ll=gDaqRecKeyTbl[i]; ll != NULL; ll=lln)
    {
      lln = ll->nxt;
      free(ll);
    }
    gDaqRecKeyTbl[i] = NULL;
  }
  feclog("DAQ record hash table freed");
  /* now free any associations */
  for (ll=gDaqRecAssoc; ll != NULL; ll=lln)
  {
    lln = ll->nxt;
    free(ll);
  }
  if (gDaqRecAssoc != NULL) feclog("device name associations list freed");
  gDaqRecAssoc = NULL;
}
DaqRecKey *makeDaqRecKey(char *key,int id, int rec,int prm)
{
  DaqRecKey *ll;
  int idx;
  if (key == NULL || id < 0 || rec < 0 || prm < 0) return NULL;
  if ((ll=getDaqRecKey(key)) != NULL) return ll;
  if ((ll=(DaqRecKey *)calloc(1,sizeof(DaqRecKey))) == NULL)
  {
    feclog("no memory to add hash table entry %s !",key);
    return NULL;
  }
  strncpy(ll->key,key,PROPERTY_NAME_SIZE);
  ll->idx = id;
  ll->rec = rec;
  ll->prm = prm;
  idx = (int)(ElfHash((unsigned char *)key) % DAQREC_HASH_SIZE);
  ll->nxt = gDaqRecKeyTbl[idx];
  gDaqRecKeyTbl[idx] = ll;
  return ll;
}
int addDaqRecAssoc(char *assoc,char *param,int id,int rec,int prm)
{
  DaqRecKey *ll;
  if (assoc == NULL || id < 0 || rec < 0 || prm < 0) return argument_list_error;
  for (ll=gDaqRecAssoc; ll != NULL; ll=ll->nxt)
  {
    if (strnicmp(ll->key,assoc,PROPERTY_NAME_SIZE)) continue;
    if (ll->idx == id && ll->rec == rec && ll->prm == prm) return already_assigned;
  }
  if ((ll=(DaqRecKey *)calloc(1,sizeof(DaqRecKey))) == NULL)
  {
    feclog("no memory to add association entry %s !",assoc);
    return out_of_local_memory;
  }
  feclog("associating %.64s with %.64s",assoc,param);
  strncpy(ll->key,assoc,PROPERTY_NAME_SIZE);
  ll->idx = id;
  ll->rec = rec;
  ll->prm = prm;
  ll->nxt = gDaqRecAssoc;
  gDaqRecAssoc = ll;
  return 0;
}

#define MAX_PARAMS 100
#define quit(i) { printf("Register struct: error\n"); exit(i); }
void registerDaqStructs(int maxDaqRecords)
{
  static int done=FALSE;
  if (done) return;
  done = TRUE;
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,key),PROPERTY_NAME_SIZE,CF_TEXT,"key")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,units),UNITS_SIZE,CF_TEXT,"units")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,desc),PROPERTY_DESC_SIZE,CF_TEXT,"desc")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,plot),UNITS_SIZE,CF_TEXT,"plot")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,boundTo),PROPERTY_NAME_SIZE,CF_TEXT,"boundTo")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,sub),SUBSYSTEM_NAME_SIZE,CF_TEXT,"sub")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,assoc),PROPERTY_NAME_SIZE,CF_TEXT,"assoc")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,fmt),1,CF_INT32,"fmt")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,size),1,CF_INT32,"size")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,scaleFactor),1,CF_FLOAT,"scaleFactor")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,scaleOffset),1,CF_FLOAT,"scaleOffset")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,maximum),1,CF_FLOAT,"maximum")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,minimum),1,CF_FLOAT,"minimum")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,aTolerance),1,CF_FLOAT,"aTolerance")) quit(1);
  if (addFieldToStruct(DAQPrmTag,OFFSETIN(DAQPrmInfo,pTolerance),1,CF_FLOAT,"pTolerance")) quit(1);
  if (sealTaggedStruct(DAQPrmTag,sizeof(DAQPrmInfo),MAX_PARAMS)) quit(1);

  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,fn_key),80,CF_TEXT,"keyFile")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,tag),80,CF_TEXT,"tag")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,ctx),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,len),1,CF_INT32,"length")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,fmt),1,CF_INT32,"fmt")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,hbt),1,CF_INT32,"heartbeat")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,tmr),1,CF_INT32,"pollInterval")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,infmt),1,CF_INT32,"inputFmt")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,inlen),1,CF_INT32,"inputLength")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,opt),1,CF_INT32,"filterOptions")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,active),1,CF_INT32,"isActive")) quit(1);
  if (addFieldToStruct(DAQRecTag,OFFSETIN(DAQRecInfo,rec),1,CF_INT32,"recordIndex")) quit(1);
  if (maxDaqRecords < 1000) maxDaqRecords = 1000;
  if (sealTaggedStruct(DAQRecTag,sizeof(DAQRecInfo),maxDaqRecords)) quit(1);

  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,tag),64,CF_TEXT,"tag")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,key),64,CF_TEXT,"key")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,desc),64,CF_TEXT,"description")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,sValid),64,CF_TEXT,"validStr")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,mask),1,CF_INT32,"mask")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,valid),1,CF_INT32,"valid")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,active),1,CF_INT32,"active")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,sValidCmp),1,CF_INT32,"validStrCmp")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,fmtValidMin),1,CF_INT32,"fmtValidMin")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,fmtValidMax),1,CF_INT32,"fmtValidMax")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,fValidMin),1,CF_DOUBLE,"validMin")) quit(1);
  if (addFieldToStruct(DAQFltrTag,OFFSETIN(DAQFilter,fValidMax),1,CF_DOUBLE,"validMax")) quit(1);
  if (sealTaggedStruct(DAQFltrTag,sizeof(DAQFilter),MAX_FILTERS)) quit(1);

}
int getDaqFltrTbl(DAQFilter *dst,int *len)
{
  int i;
  if (dst == NULL || len == NULL || *len == 0) return argument_list_error;
  if (*len > nDaqFilterTblEntries) *len = nDaqFilterTblEntries;

  for (i=0; i<*len; i++)
  {
    memcpy(&dst[i],&daqFilterTbl[i],sizeof(DAQFilter));
  }
  return 0;
}
int addDaqFilters(DAQFilter *src,int len)
{
  int i,k;
  if (src == NULL || len == 0) return 0;
  for (i=0; i<len; i++)
  {
    for (k=0; k<nDaqFilterTblEntries; k++)
    { /* scan thru for edits */
      if (!strncmp(src[i].tag,daqFilterTbl[k].tag,64)) break;
    }
    if (k >= MAX_FILTERS) return resources_exhausted;
    if (k == nDaqFilterTblEntries) nDaqFilterTblEntries++;
    memcpy(&daqFilterTbl[k],&src[i],sizeof(DAQFilter));
  }
  return dumpDaqFltrTbl(ArchiveDBpath,FILTERS_FILENAME);
}
#define fltrhdr "FILTER, KEYWORD, FORMAT, MIN_VALID, MAX_VALID, DESCRIPTION"
int dumpDaqFltrTbl(char *path,char *fn)
{
  FILE *fp=NULL;
  char fullName[256];
  int i, fmt, cc = 0;

  if (path == NULL) ccerr(invalid_parameter);
  if (fn == NULL || strlen(fn) == 0) ccerr(invalid_parameter);
  sprintf(fullName,"%.128s%s",path,fn);
  makeDirPath(fullName);
  if ((fp=fopen(fullName,"w")) == NULL) ccerr(file_error);

  fprintf(fp,"%s\n",fltrhdr);
  for (i=nDaqStockFilterNames; i<nDaqFilterTblEntries; i++)
  {
    fmt = daqFilterTbl[i].fmtValidMax;
    if (fmt == CF_NULL) fmt = daqFilterTbl[i].fmtValidMin;
    if (isStringFormat((short)fmt))
    {
      fprintf(fp,"%.64s, %.64s, %.32s, %.64s, %.64s, %.64s\n",
        daqFilterTbl[i].tag,daqFilterTbl[i].key,itof(fmt),
        daqFilterTbl[i].sValid,daqFilterTbl[i].sValid,daqFilterTbl[i].desc);
    }
    else
    {
      fprintf(fp,"%.64s, %.64s, %.32s, %g, %g, %.64s\n",
        daqFilterTbl[i].tag,daqFilterTbl[i].key,itof(fmt),
        daqFilterTbl[i].fValidMin,daqFilterTbl[i].fValidMax,daqFilterTbl[i].desc);
    }
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR,"cannot dump database : %s",cc2str(cc));
  return cc;
}
void rec2info(DAQRecInfo *dst, DAQRecord *src)
{
  dst->active = src->active;
  strncpy(dst->ctx, src->ctx, CONTEXT_NAME_SIZE);
  strncpy(dst->srv, src->srv, EXPORT_NAME_SIZE);
  strncpy(dst->dev, src->dev, DEVICE_NAME_SIZE);
  strncpy(dst->prp, src->prp, PROPERTY_NAME_SIZE);
  dst->fmt = src->fmt;
  if (dst->fmt == CF_STRUCT)
    strncpy(dst->tag, src->tag, TAG_NAME_SIZE);
  else
    strncpy(dst->tag, src->fn_inp, 80);
  strncpy(dst->fn_key, src->fn_key, 80);
  dst->hbt = src->hbt;
  dst->len = src->len;
  dst->infmt = src->infmt;
  dst->inlen = src->inlen;
  dst->opt = src->option;
  dst->rec = src->rec;
  dst->tmr = src->tmr;
}
int getDaqRecInfo(DAQRecInfo *dst,int *plen)
{
  int i,n,len;
  if (dst == NULL || plen == NULL || (len=*plen) == 0) return argument_list_error;
  if (len > totalNrRecords) len = totalNrRecords;

  for (i=0,n=0; i<totalNrRecords && n<len; i++)
  {
    if (daqTbl[i].rec < 0) continue;
    rec2info(&dst[n], &daqTbl[i]);
    n++;
  }
  *plen = n;
  return 0;
}
int getDaqRecInfoEx(DAQRecInfo *dst,int *plen,char *target)
{
  int i, len, n = 0;
  int blen = 0, elen = 0, mlen = 0;
  char beginToken[64], endToken[64], midToken[64];
  if (dst == NULL || plen == NULL || (len=*plen) == 0) return argument_list_error;
  if (len > totalNrRecords) len = totalNrRecords;
  getWildcardTokens(target,beginToken,endToken,midToken);
  /* calculate these once outside the loop ... */
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);

  for (i=0; i<totalNrRecords && n<len; i++)
  {
    if (daqTbl[i].rec < 0) continue;
    if (!wildcardMatches(daqTbl[i].prp,beginToken,blen,endToken,elen,midToken,mlen)) continue;
    dst[n].active = daqTbl[i].active;
    strncpy(dst[n].ctx,daqTbl[i].ctx,CONTEXT_NAME_SIZE);
    strncpy(dst[n].srv,daqTbl[i].srv,EXPORT_NAME_SIZE);
    strncpy(dst[n].dev,daqTbl[i].dev,DEVICE_NAME_SIZE);
    strncpy(dst[n].prp,daqTbl[i].prp,PROPERTY_NAME_SIZE);
    dst[n].fmt = daqTbl[i].fmt;
    if (dst[n].fmt == CF_STRUCT)
      strncpy(dst[n].tag,daqTbl[i].tag,TAG_NAME_SIZE);
    else
      strncpy(dst[n].tag,daqTbl[i].fn_inp,80);
    strncpy(dst[n].fn_key,daqTbl[i].fn_key,80);
    dst[n].hbt = daqTbl[i].hbt;
    dst[n].len = daqTbl[i].len;
    dst[n].infmt = daqTbl[i].infmt;
    dst[n].inlen = daqTbl[i].inlen;
    dst[n].opt = daqTbl[i].option;
    dst[n].rec = daqTbl[i].rec;
    dst[n].tmr = daqTbl[i].tmr;
    n++;
  }
  *plen = n;
  return 0;
}
typedef struct
{
  char key[PROPERTY_NAME_SIZE]; /* parameter name */
  char units[UNITS_SIZE];       /* units */
  char desc[PROPERTY_DESC_SIZE]; /* description */
  char plot[UNITS_SIZE];        /* plot style (LOG, LIN, etc.) */
  char boundTo[PROPERTY_NAME_SIZE]; /* bound to parameter */
  char sub[SUBSYSTEM_NAME_SIZE]; /* parameter subsystem */
  char assoc[PROPERTY_NAME_SIZE]; /* associate keyword */
  char assocAlt[PROPERTY_NAME_SIZE]; /* alternate associate column header */
  char fmtStr[32];              /* allowable format */
  char tolStr[64];
  int size;             /* data array size assuming given format */
  float scaleFactor;
  float scaleOffset;
  float maximum;       /* maximum value for parameter */
  float minimum;       /* minimun value for parameter */
} DAQPrmInfoFile;

CSVCOLUMN csvDaqPrmFileColumns[] =
{
  { "KEYWORD","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,key),NULL,0 },
  { "UNITS","",0,CF_TEXT,UNITS_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,units),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,PROPERTY_DESC_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,desc),NULL,0 },
  { "PLOT","",0,CF_TEXT,UNITS_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,plot),NULL,0 },
  { "BIND","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,boundTo),NULL,0 },
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,sub),NULL,0 },
  { "ASSOCIATE","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,assoc),NULL,0 },
  { "DEVICES","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(DAQPrmInfoFile,assocAlt),NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(DAQPrmInfoFile,tolStr),NULL,0 },
  { "FORMAT","",0,CF_TEXT,32,(BYTE *)OFFSETIN(DAQPrmInfoFile,fmtStr),NULL,0 },
  { "SIZE","",0,CF_INT32,1,(BYTE *)OFFSETIN(DAQPrmInfoFile,size),NULL,0 },
  { "SCALE","",0,CF_FLOAT,1,(BYTE *)OFFSETIN(DAQPrmInfoFile,scaleFactor),NULL,0 },
  { "OFFSET","",0,CF_FLOAT,1,(BYTE *)OFFSETIN(DAQPrmInfoFile,scaleOffset),NULL,0 },
  { "MAXIMUM","",0,CF_FLOAT,1,(BYTE *)OFFSETIN(DAQPrmInfoFile,maximum),NULL,0 },
  { "MINIMUM","",0,CF_FLOAT,1,(BYTE *)OFFSETIN(DAQPrmInfoFile,minimum),NULL,0 },
};
CSVDB csvDaqPrmFileDb =
{
  csvDaqPrmFileColumns,
  sizeof(csvDaqPrmFileColumns)/sizeof(csvDaqPrmFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(DAQPrmInfoFile)
};

int readDaqPrmFile(char *path,char *fn,DAQPrmInfo *dst,int *siz)
{
  int cc=0,i,n;
  char filename[256], scratch[64], *s1, *s2;
  DAQPrmInfoFile prms[256], *pprms = prms; /* use the stack for quick MMU */
  memset(filename,0,256);
  sprintf(filename,"%.128sinfo%c",ArchiveDBpath,FS_DELIMITER);
  csvDaqPrmFileDb.siz = 256;
  if ((cc=csvReadFile(filename,fn,&csvDaqPrmFileDb,(void **)&pprms)) != 0) ccerr(cc);
  n = csvDaqPrmFileDb.siz;
  if (n > *siz) n = *siz;
  if (n > 256) 
  { /* no one in his right mind will have a parameter breakdown of more than 256 ! */
    feclogEx(TINE_LOGLEVEL_ERROR,"probable stack overwrite in readDaqPrmFile (%d more than the allowed 256 parameters) !",n);
  }
  *siz = n;
  for (i=0; i<n; i++)
  {
    strncpy(dst[i].assoc,prms[i].assoc,PROPERTY_NAME_SIZE);
    if (strlen(dst[i].assoc) == 0) strncpy(dst[i].assoc,prms[i].assocAlt,PROPERTY_NAME_SIZE);
    strncpy(dst[i].boundTo,prms[i].boundTo,PROPERTY_NAME_SIZE);
    strncpy(dst[i].key,prms[i].key,PROPERTY_NAME_SIZE);
    strncpy(dst[i].desc,prms[i].desc,PROPERTY_DESC_SIZE);
    strncpy(dst[i].units,prms[i].units,UNITS_SIZE);
    strncpy(dst[i].plot,prms[i].plot,UNITS_SIZE);
    strncpy(dst[i].sub,prms[i].sub,SUBSYSTEM_NAME_SIZE);
    dst[i].maximum = prms[i].maximum;
    dst[i].minimum = prms[i].minimum;
    dst[i].scaleFactor = prms[i].scaleFactor;
    dst[i].scaleOffset = prms[i].scaleOffset;
    dst[i].size = prms[i].size;
    dst[i].fmt = ftoi(prms[i].fmtStr);
    strncpy(scratch,prms[i].tolStr,64); scratch[63] = 0;
    s1 = scratch;
    if ((s2=strchr(s1,'+')) != NULL) *s2++ = 0;
    if (strchr(s1, '%'))
    {
      dst[i].pTolerance = (float)(atof(s1) / 100.0);
      if (s2 != NULL) dst[i].aTolerance = (float)atof(s2);
    }
    else
    {
      dst[i].aTolerance = (float)atof(s1);
      if (s2 != NULL && strchr(s2, '%')) dst[i].pTolerance = (float)(atof(s2) / 100.0);
    }
  }
err:
  return cc;
}
int getDaqPrmInfo(int rec,DAQPrmInfo *dst,int *len)
{
  int i,n;
  if (dst == NULL || len == NULL || *len == 0) return argument_list_error;
  if ((n=getTableEntryFromRecordNumber(rec)) < 0) return invalid_index;
  if (*len > daqTbl[n].nparam) *len = daqTbl[n].nparam;
  for (i=0; i<*len; i++)
  {
     strncpy(dst[i].key,daqTbl[n].daqParam[i].param,PROPERTY_NAME_SIZE);
     strncpy(dst[i].units,daqTbl[n].daqParam[i].units,UNITS_SIZE);
     strncpy(dst[i].desc,daqTbl[n].daqParam[i].desc,PROPERTY_NAME_SIZE);
     strncpy(dst[i].plot,daqTbl[n].daqParam[i].plot,UNITS_SIZE);
     strncpy(dst[i].boundTo,daqTbl[n].daqParam[i].boundTo,PROPERTY_NAME_SIZE);
     strncpy(dst[i].sub,daqTbl[n].daqParam[i].subsys,SUBSYSTEM_NAME_SIZE);
     strncpy(dst[i].assoc,daqTbl[n].daqParam[i].assoc,PROPERTY_NAME_SIZE);
     dst[i].fmt = daqTbl[n].daqParam[i].fmt;
     dst[i].size = daqTbl[n].daqParam[i].size;
     dst[i].scaleFactor = daqTbl[n].daqParam[i].scaleFactor;
     dst[i].scaleOffset = daqTbl[n].daqParam[i].scaleOffset;
     dst[i].maximum = daqTbl[n].daqParam[i].maximum;
     dst[i].minimum = daqTbl[n].daqParam[i].minimum;
     dst[i].aTolerance = daqTbl[n].daqParam[i].aTolerance;
     dst[i].pTolerance = daqTbl[n].daqParam[i].pTolerance;
  }
  return 0;
}
int getDaqPrmInpt(int rec,DTYPE *dout)
{
  int siz,n;
  if (dout == NULL) return argument_list_error;
  if ((n=getTableEntryFromRecordNumber(rec)) < 0) return invalid_index;
  if (daqTbl[n].inlen < (int)dout->dArrayLength) dout->dArrayLength = daqTbl[n].inlen;
  siz = fmtsizeof(daqTbl[n].infmt);
  if (siz == 0) return 0;
  return PutValuesFromAny(dout,daqTbl[n].indata,daqTbl[n].infmt,1,siz,daqTbl[n].inlen,0);
}

int kwdEndsWith(char *kwd, char *suf)
{
  int klen, slen;
  if (kwd == NULL || suf == NULL) return FALSE;
  if ((klen=(int)strlen(kwd)) > PROPERTY_NAME_SIZE) klen = PROPERTY_NAME_SIZE;
  if ((slen=(int)strlen(suf)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
  if (klen < slen) return FALSE;
  if (strncmp(&kwd[klen-slen],suf,slen)) return FALSE;
  return TRUE;
}
char *toCsvString(char *str)
{
  static char csvStr[256];
  if (strchr(str,',') == NULL) return str;
  sprintf(csvStr,"\"%.128s\"",str);
  return csvStr;
}
char *toFilterString(int option)
{
  int i;
  static char fltr[512];
  fltr[0] = 0;
  for (i=0; i<nDaqFilterTblEntries; i++)
  {
    if (option & daqFilterTbl[i].mask) 
    {
      if (strlen(fltr) > 0) strcat(fltr,"|");
      strncat(fltr,daqFilterTbl[i].tag,16);
    }
  }
  return fltr;
}
int dumpDaqInpt(char *path,char *fn,BYTE *data,int fmt,int len)
{
  FILE *fp=NULL;
  char fullName[256],fmtstr[64];
  int i, fsiz, cc = 0;

  if (fn == NULL || strlen(fn) == 0) ccerr(0);
  if (fmt == CF_NULL || len == 0) ccerr(invalid_data);
  sprintf(fullName,"%.128sinput%s%s",path,delStr,fn);
  makeDirPath(fullName);
  if ((fp=fopen(fullName,"w")) == NULL) ccerr(file_error);
  strncpy(fmtstr,itof(fmt),64);
  fsiz = fmtsizeof(fmt);
  fprintf(fp,"Data, Format\n");
  for (i=0; i<len; i++)
  {
    switch (fmt)
    {
      case CF_INT8:
        fprintf(fp,"%d, %s\n",((BYTE *)data)[i],fmtstr);
        break;
      case CF_INT16:
        fprintf(fp,"%d, %s\n",((SINT16 *)data)[i],fmtstr);
        break;
      case CF_INT32:
        fprintf(fp,"%d, %s\n",((SINT32 *)data)[i],fmtstr);
        break;
      case CF_FLOAT:
        fprintf(fp,"%g, %s\n",((float *)data)[i],fmtstr);
        break;
      case CF_DOUBLE:
        fprintf(fp,"%g, %s\n",((double *)data)[i],fmtstr);
        break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        fprintf(fp,"%.64s, %s\n",(char *)data + i*fsiz,fmtstr);
        break;
    }
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR,"cannot dump database : %s",cc2str(cc));
  return cc;
}
int dumpDaqInptDb(char *path,int idx)
{
  if (idx < 0 || idx >= nrecords) return invalid_parameter;
  if (strlen(daqTbl[idx].fn_inp) == 0) return 0;
  if (daqTbl[idx].infmt == CF_NULL || daqTbl[idx].inlen == 0) return invalid_data;
  return dumpDaqInpt(path,daqTbl[idx].fn_inp,daqTbl[idx].indata,daqTbl[idx].infmt,daqTbl[idx].inlen);
}
int addDaqInpt(char *fn,BYTE *data,int fmt, int len)
{
  int i,n;

  if (fn == NULL || data == NULL) return invalid_parameter;
  if (fn[0] == '#') 
  {
    n = atoi(&fn[1]);
    if ((i=getTableEntryFromRecordNumber(n)) < 0) return invalid_index;
    fn = daqTbl[i].fn_inp;
  }
  else
  {
    i = getEntryFromInputFile(fn,&n);
  }
  return dumpDaqInpt(ArchiveDBpath,fn,data,fmt,len);
}
char *toToleranceStr(float atol, float ptol)
{
  static char tolstr[512];
  tolstr[0] = 0;
  if (atol == 0 && ptol == 0) return "0";
  atol = (float)fabs(atol);
  ptol = (float)fabs(ptol);
  if (atol > 0) sprintf(tolstr,"%g",atol);
  if (ptol > 0) 
  {
    if (strlen(tolstr) > 0) strcat(tolstr," + ");
    sprintf(&tolstr[strlen(tolstr)],"%g%%",ptol*100.0);
  }
  return tolstr;
}
char *getFieldPtr(char *fld,int len)
{
  static char fldBuffer[128];
  if ((int)strlen(fld) < len) return fld;
  strncpy(fldBuffer,fld,len);
  fldBuffer[len] = 0;
  return fldBuffer;
}
int isDaqParameterValid(DAQPrmInfo *prm)
{
  static char *invSet = ", ;'\"\\~*^(){}[]?<>/";
  static char *invAssocSet = ",;'\"\\~*^(){}?<>/";
  char *c, *fld, *txt;
  int cc = 0;
  txt = "keyword"; fld = getFieldPtr(prm->key,PROPERTY_NAME_SIZE);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "boundTo"; fld = getFieldPtr(prm->boundTo,PROPERTY_NAME_SIZE);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "associate"; fld = getFieldPtr(prm->assoc,PROPERTY_NAME_SIZE);
  if ((c=strpbrk(fld,invAssocSet)) != NULL) ccerr(illegal_character);
  return TRUE;
err:
  sprintf(daqUpdateErrStr,"the %s field contains an invalid character: %c",txt,*c);
  return FALSE;
}
int isDaqEntryValid(DAQRecInfo *rec)
{
  static char *invSet = ", ;'\"\\~^(){}[]?<>";
  char *c, *fld, *txt;
  int cc = 0;
  txt = "context"; fld = getFieldPtr(rec->ctx,CONTEXT_NAME_SIZE);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "device"; fld = getFieldPtr(rec->dev,DEVICE_NAME_SIZE);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "property"; fld = getFieldPtr(rec->prp,PROPERTY_NAME_SIZE);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "server"; fld = getFieldPtr(rec->srv,EXPORT_NAME_SIZE);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "filename"; fld = getFieldPtr(rec->fn_key,80);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);
  txt = "tag"; fld = getFieldPtr(rec->tag,80);
  if ((c=strpbrk(fld,invSet)) != NULL) ccerr(illegal_character);  
  return TRUE;
err:
  sprintf(daqUpdateErrStr,"the %s field contains an invalid character: %c",txt,*c);
  return FALSE;
}
int isRestrictedMetaProperty(char *key)
{ /* key will not be NULL !*/
  char *c;
  if ((c=strrchr(key,'.')) == NULL) return FALSE;
  if (strcmp(c,".ONLINE") == 0) return TRUE;
  if (strstr(key,".BIT.") != NULL) return TRUE;
  if (strstr(key,".MASK.") != NULL) return TRUE;
  if (strstr(key,".GATE.") != NULL) return TRUE;
  if (strstr(key,".DMASK.") != NULL) return TRUE;
  return FALSE;
}
int addDaqPrmInfoElem(char *fn,DAQPrmInfo *prm,int len)
{ 
  int i,j,k,n,fsiz,r,reclen;

  if (fn == NULL || prm == NULL) return invalid_parameter;
  if (fn[0] == '#') 
  {
    n = atoi(&fn[1]);
    if ((i=getTableEntryFromRecordNumber(n)) < 0) 
    {
      sprintf(daqUpdateErrStr,"rec %d does not have a table entry",n); 
      feclogEx(TINE_LOGLEVEL_ERROR, daqUpdateErrStr);
      return invalid_index;
    }
    fn = daqTbl[i].fn_key;
  }
  else
  {
    if (fn[0] == '.') return no_such_file;
    i = getEntryFromKeywordFile(fn,&n);
  }
  if (i >= 0)
  { /* editing an existing table entry */
    for (k=0, reclen=0; k<len; k++) 
    {
      fsiz = prm[k].fmt == CF_STRUCT ? GetStructSize(daqTbl[i].tag) : fmtsizeof(prm[k].fmt);
      reclen += prm[k].size * fsiz + getFormatHeaderSize(prm[k].fmt);
    }
    if (isVariableLengthFormat((short)prm[0].fmt))
    { /* variable length => examine prm[0] */
      reclen = 2 * sizeof(size_t);
    }
    if (reclen != daqTbl[i].reclen) 
    {
      sprintf(daqUpdateErrStr,"rec %d has a record length mismatch: %d vs %d",daqTbl[i].rec,daqTbl[i].reclen,reclen);
      feclogEx(TINE_LOGLEVEL_ERROR, daqUpdateErrStr);
      return illegal_data_size;
    }
  }
  else
  { /* new table entry, but check new parameter names */
    char *pkey=NULL;
    for (i=0; i<len; i++)
    {
      if (getDaqParameter(prm[i].key) != NULL) 
      {
        sprintf(daqUpdateErrStr,"name %.64s already exists",prm[i].key); 
        feclogEx(TINE_LOGLEVEL_ERROR, daqUpdateErrStr);
        return name_already_exists;
      }
      if ((pkey=ResolveSystemAlias(prm[i].key)) != NULL)
      {
        if ((j=getArchivedEntryFromKeyword(pkey,&r,&k)) >= 0)
        { /* devName is a keyword property */
          sprintf(daqUpdateErrStr,"name %.64s is already a registered alias for %.64s (rec. %d)",
            prm[i].key,daqTbl[j].daqParam[k].param,r); 
        }
        else
        {
          sprintf(daqUpdateErrStr,"name %.64s is already a registered alias",prm[i].key);
        }
        feclogEx(TINE_LOGLEVEL_ERROR,daqUpdateErrStr);
        return name_already_exists;
      }
      if (isRestrictedMetaProperty(prm[i].key)) 
      {
        sprintf(daqUpdateErrStr,"name %.64s is a restricted meta property",prm[i].key); 
        feclogEx(TINE_LOGLEVEL_ERROR, daqUpdateErrStr);
        return invalid_name;
      }
      if (!isDaqParameterValid(&prm[i])) 
      {
        sprintf(daqUpdateErrStr,"name %.64s is invalid",prm[i].key); 
        feclogEx(TINE_LOGLEVEL_ERROR, daqUpdateErrStr);
        return illegal_character;
      }
    }
  }
  daqUpdateErrStr[0] = 0;
  return dumpDaqPrmInfoElem(ArchiveDBpath,fn,prm,len);
}
#define prmhdr "Keyword,Format,Size,Units,Maximum,Minimum,Tolerance,Plot,Scale,Offset,Description,associate,bind,subsystem"
int dumpDaqPrmInfoElem(char *path,char *fn,DAQPrmInfo *prm,int len)
{
  FILE *fp=NULL;
  char fullName[256];
  int i, cc = 0;

  if (path == NULL || prm == NULL || len < 1) ccerr(invalid_parameter);
  if (fn == NULL || strlen(fn) == 0) ccerr(invalid_parameter);
  sprintf(fullName,"%.128sinfo%s%s",path,delStr,fn);
  makeDirPath(fullName);
  if ((fp=fopen(fullName,"w")) == NULL) ccerr(file_error);

  fprintf(fp,"%s\n",prmhdr);
  for (i=0; i<len; i++)
  {
    fprintf(fp,"%.64s, %.16s, %d, %.32s, %g, %g, %.16s, %.16s, %g, %g, %.64s, %.64s, %.64s, %.32s\n",
      prm[i].key,itof(prm[i].fmt),prm[i].size,
      toCsvString(prm[i].units),prm[i].maximum,prm[i].minimum,
      toToleranceStr(prm[i].aTolerance,prm[i].pTolerance),
      prm[i].plot,prm[i].scaleFactor,prm[i].scaleOffset,
      toCsvString(prm[i].desc),prm[i].assoc,prm[i].boundTo,
      prm[i].sub);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR,"cannot dump database : %s",cc2str(cc));
  return cc;
}
int dumpDaqPrmDb(char *path,int idx)
{
  FILE *fp=NULL;
  char fullName[256];
  int i, cc = 0;

  if (idx < 0 || idx >= nrecords) ccerr(invalid_parameter);
  if (strlen(daqTbl[idx].fn_key) == 0) ccerr(code_failure);
  sprintf(fullName,"%.128sinfo%s%s",path,delStr,daqTbl[idx].fn_key);
  makeDirPath(fullName);
  if ((fp=fopen(fullName,"w")) == NULL) ccerr(file_error);

  fprintf(fp,"%s\n",prmhdr);
  for (i=0; i<daqTbl[idx].nparam; i++)
  {
    fprintf(fp,"%.64s, %.16s, %d, %.32s, %g, %g, %.16s, %.16s, %g, %g, %.64s, %.64s, %.64s, %.32s\n",
      daqTbl[idx].daqParam[i].param,itof(daqTbl[idx].daqParam[i].fmt),daqTbl[idx].daqParam[i].size,
      toCsvString(daqTbl[idx].daqParam[i].units),daqTbl[idx].daqParam[i].maximum,daqTbl[idx].daqParam[i].maximum,
      toToleranceStr(daqTbl[idx].daqParam[i].aTolerance,daqTbl[idx].daqParam[i].pTolerance),
      daqTbl[idx].daqParam[i].plot,daqTbl[idx].daqParam[i].scaleFactor,daqTbl[idx].daqParam[i].scaleOffset,
      toCsvString(daqTbl[idx].daqParam[i].desc),daqTbl[idx].daqParam[i].assoc,daqTbl[idx].daqParam[i].boundTo,
      daqTbl[idx].daqParam[i].subsys);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR,"cannot dump database : %s",cc2str(cc));
  return cc;
}
int recsizfmtmismatch = FALSE;
int reccmp(DAQRecInfo *inf,DAQRecord *rec)
{
  if (inf->fmt < 256) inf->fmt = LFMT(inf->fmt);
  recsizfmtmismatch = FALSE;
  if (inf->rec == rec->rec)
  { /* direct record compare */
    if (inf->len != rec->len || inf->fmt != rec->fmt) 
    { /* but the lengths don't match ! */
      inf->rec = -1; /* force an invalid_reference return code */
      feclogEx(TINE_LOGLEVEL_ERROR,"update record %d refused: size (%d vs %d) or format (%d vs %d) mismatch",
        rec->rec,rec->len,inf->len,rec->fmt,inf->fmt);
      sprintf(daqUpdateErrStr,"update refused for rec %d (size or format mismatch)",rec->rec);
      recsizfmtmismatch = TRUE;
    }
    return 0;
  }
  if (strnicmp(inf->ctx,rec->ctx,CONTEXT_NAME_SIZE)) return -1;
  if (strnicmp(inf->srv,rec->srv,EXPORT_NAME_SIZE)) return -1;
  if (strnicmp(inf->dev,rec->dev,DEVICE_NAME_SIZE)) return -1;
  if (strnicmp(inf->prp,rec->prp,PROPERTY_NAME_SIZE)) return -1;
  if (inf->len != rec->len) return -1;
  if (inf->fmt != rec->fmt) return -1;
  if (inf->fmt == CF_STRUCT)
    if (strnicmp(inf->tag,rec->tag,TAG_NAME_SIZE)) return -1;
  else
    if (strnicmp(inf->tag,rec->fn_inp,80)) return -1;
  return 0;
}
int nextFreeRecord = -1;
int getFreeRecordNumber(void)
{
  int i, rec=nextFreeRecord;
  if (rec < 0)
  { /* find it */
    for (i=0; i<nrecords; i++) if (rec < daqTbl[i].rec) rec = daqTbl[i].rec;
  }
  nextFreeRecord = ++rec;
  return rec;
}
int acqRecordStruct(int idx)
{
  char devnam[132];
  sprintf(devnam,"/%.32s/%.32s",daqTbl[idx].ctx,daqTbl[idx].srv);
  return AcquireAndRegisterStruct(devnam,daqTbl[idx].tag,daqTbl[idx].len);
}
int addDaqRecInfoElements(DAQRecInfo *daqrec,int len)
{
  int i, k, fsiz, nrecs;
  time_t t = time(NULL);
  struct tm *tms;
  char adbBckupPath[256],pathDecor[16];
  static char lastPathDecor[16];

  if (daqrec == NULL) return invalid_parameter;  
  if ((tms=localtime(&t)) != NULL)
  { /* backup current database if first change 'today' */
    sprintf(pathDecor,"-%02d.%02d.%02d%s",tms->tm_mday,tms->tm_mon+1,tms->tm_year%100,delStr);
    if (strncmp(pathDecor,lastPathDecor,16))
    {
      k = (int)strlen(ArchiveDBpath)-1;
      strncpy(adbBckupPath,ArchiveDBpath,k); adbBckupPath[k] = 0;
      strcat(adbBckupPath,pathDecor);
      dumpDaqDb(adbBckupPath,ArchiveDBfile,TRUE);
      strncpy(lastPathDecor,pathDecor,16);
    }
  }
  nrecs = nrecords;
  for (k=0; k<len; k++)
  {
    for (i=0; i<nrecs; i++)
    { /* is already in the table ? (just an edit) */
      if (reccmp(&daqrec[k],&daqTbl[i])) continue;
      if (daqrec[k].rec != daqTbl[i].rec) 
      {
        if (daqrec[k].rec == -1 && daqrec[k].active && !daqTbl[i].active)
        { /* awkward attempt to turn a record back on */
          feclog("addDaqRecInfoElements: re-activate record %d",daqTbl[i].rec);
        }
        else
        {
          if (!recsizfmtmismatch || daqrec[k].rec != -1)
          {
            feclog("addDaqRecInfoElements: editing record %d update refused",daqTbl[i].rec);
            sprintf(daqUpdateErrStr,"new rec %d conflicts with rec %d",daqrec[k].rec,daqTbl[i].rec);
            feclogEx(TINE_LOGLEVEL_ERROR,daqUpdateErrStr);
          }
          return invalid_reference;
        }
      }
      /* record index is the same (this is an edit!) */
      if (daqrec[k].hbt < 5) daqrec[k].hbt = 900; /* sanity check */
      if (daqrec[k].tmr < 10) daqrec[k].tmr = 1000; /* sanity check */
      daqTbl[i].hbt = daqrec[k].hbt;
      daqTbl[i].tmr = daqrec[k].tmr;
      daqTbl[i].option = daqrec[k].opt;
      if (daqrec[k].fmt == CF_STRUCT)
        strncpy(daqTbl[i].tag,daqrec[k].tag,TAG_NAME_SIZE);
      else
        strncpy(daqTbl[i].fn_inp,daqrec[k].tag,80);
      /* consistency will be checked later */
      strncpy(daqTbl[i].fn_key,daqrec[k].fn_key,80);
      strncpy(daqTbl[i].ctx,daqrec[k].ctx,CONTEXT_NAME_SIZE);
      strncpy(daqTbl[i].srv,daqrec[k].srv,EXPORT_NAME_SIZE);
      strncpy(daqTbl[i].dev,daqrec[k].dev,DEVICE_NAME_SIZE);
      strncpy(daqTbl[i].prp,daqrec[k].prp,PROPERTY_NAME_SIZE);
      daqTbl[i].active = daqrec[k].active;
      break;      
    }
    if (i < nrecs) continue;
    /* is new ! */
    if (nrecs >= recordsCapacity) 
    {
      feclogEx(TINE_LOGLEVEL_ERROR,"addDaqRecInfoElements: number of records would exceed buffer capacity (%d)",recordsCapacity);
      sprintf(daqUpdateErrStr,"%d records would exceed capacity (%d)",nrecs,recordsCapacity);
      return resources_exhausted;
    }
    if (!isDaqEntryValid(&daqrec[k])) return illegal_character;
    if (daqrec[k].len < 1) return illegal_data_size;
    if (daqrec[k].rec < 0) daqrec[k].rec = getFreeRecordNumber();
    if (daqrec[k].hbt < 5) daqrec[k].hbt = 900; /* sanity check */
    if (daqrec[k].tmr < 10) daqrec[k].tmr = 1000; /* sanity check */
    daqTbl[i].rec = daqrec[k].rec;
    strncpy(daqTbl[i].ctx,daqrec[k].ctx,CONTEXT_NAME_SIZE);
    strncpy(daqTbl[i].srv,daqrec[k].srv,EXPORT_NAME_SIZE);
    strncpy(daqTbl[i].dev,daqrec[k].dev,DEVICE_NAME_SIZE);
    strncpy(daqTbl[i].prp,daqrec[k].prp,PROPERTY_NAME_SIZE);
    daqTbl[i].len = daqrec[k].len;
    daqTbl[i].fmt = daqrec[k].fmt;
    daqTbl[i].hbt = daqrec[k].hbt;
    daqTbl[i].tmr = daqrec[k].tmr;
    daqTbl[i].mod = CM_TIMER;
    fsiz = fmtsizeof(daqTbl[i].fmt);
    if (daqrec[k].fmt == CF_STRUCT)
      strncpy(daqTbl[i].tag,daqrec[k].tag,TAG_NAME_SIZE);
    else
      strncpy(daqTbl[i].fn_inp,daqrec[k].tag,80);
    if (daqTbl[i].fmt == CF_STRUCT)
    {
      acqRecordStruct(i);
      if ((fsiz=GetStructSize(daqTbl[i].tag)) <= 0) return invalid_structure_tag;
    }
    daqTbl[i].reclen = daqTbl[i].len * fsiz + getFormatHeaderSize(daqTbl[i].fmt);
    daqTbl[i].option = daqrec[k].opt;
    /* consistency will be checked later */
    strncpy(daqTbl[i].fn_key,daqrec[k].fn_key,80);
    daqTbl[i].active = daqrec[k].active;
    nrecs++;
  }
  nrecords = nrecs;
  daqUpdateErrStr[0] = 0;
  /* now dump it to disk at the proper location */
  return dumpDaqDb(ArchiveDBpath,ArchiveDBfile,FALSE);
}
#define daqhdr "Record, Context, Server, Device, Property, Length, Format, Input, Heartbeat, Poll, Mode, Filter, Structure, Active"
int dumpDaqDb(char *path,char *fn,int dumpAll)
{
  FILE *fp=NULL;
  char fullName[256], fmtstr[64];
  int i, cc = 0;

  if (fn == NULL) return invalid_parameter;
  if (path == NULL) path = ArchiveDBpath;
  sprintf(fullName,"%.128s%s",path,fn);
  makeDirPath(path);
  if ((fp=fopen(fullName,"w")) == NULL) ccerr(file_error);
  
  fprintf(fp,"%s\n",daqhdr);
  for (i=0; i<nrecords; i++)
  {
    if (daqTbl[i].rec < 0) continue; /* bad record: ignore */
    strncpy(fmtstr,itof(daqTbl[i].fmt),16);
    if (daqTbl[i].fmt == CF_STRUCT)
    {
      strcat(fmtstr,".");
      strncat(fmtstr,daqTbl[i].tag,TAG_NAME_SIZE);
    }
    fprintf(fp,"%d, %.32s, %.32s, %.64s, %.64s, %d, %.64s, %.80s, %d, %d, %.64s, %.128s, %.80s, %d\n",
      daqTbl[i].rec,daqTbl[i].ctx,daqTbl[i].srv,toCsvString(daqTbl[i].dev),daqTbl[i].prp,daqTbl[i].len, 
      fmtstr,daqTbl[i].fn_inp,daqTbl[i].hbt,daqTbl[i].tmr, 
      GetModeAsString(daqTbl[i].mod), 
      toFilterString(daqTbl[i].option),
      daqTbl[i].fn_key,daqTbl[i].active);
    if (dumpAll && strlen(daqTbl[i].fn_key) > 0) dumpDaqPrmDb(path,i);
    if (dumpAll && strlen(daqTbl[i].fn_inp) > 0) dumpDaqInptDb(path,i);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR,"cannot dump database : %s",cc2str(cc));
  return cc;
}
int getTableEntryFromRecordNumber(int rec)
{
  int i;
  if (rec < 0) return rec;
  for (i=0; i<nrecords; i++)
  {
    if (daqTbl[i].rec == rec) return i;
  }
  return -1;
}
int getEntryFromKeywordFile(char *keywordFile,int *rec)
{
  int i;

  for (i=0; i<nrecords; i++)
  {
    if (!strncmp(daqTbl[i].fn_key,keywordFile,80))
    {
      if (rec != NULL) *rec = daqTbl[i].rec;
      return i;
    }
  }
  return -1;
}
int getEntryFromInputFile(char *inputFile,int *rec)
{
  int i;

  for (i=0; i<nrecords; i++)
  {
    if (!strncmp(daqTbl[i].fn_inp,inputFile,80))
    {
      if (rec != NULL) *rec = daqTbl[i].rec;
      return i;
    }
  }
  return -1;
}
int getEntryFromKeyword(char *keyword,int *rec,int *param)
{
  DAQParameter *daqParam;
  int i,j;
  DaqRecKey *key=getDaqRecKey(keyword);
  
  if (key != NULL)
  { /* quick retrieval via hash table */
    if (rec != NULL) *rec = key->rec;
    if (param != NULL) *param  = key->prm;
    return key->idx;
  }
  for (i=0; i<nrecords; i++)
  { /* not yet hashed ? */
    if (!daqTbl[i].active) continue;
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (!strnicmp(daqParam[j].param,keyword,PROPERTY_NAME_SIZE))
      { /* found it! now hash it! */
        makeDaqRecKey(keyword,i,daqTbl[i].rec,j);
        if (rec != NULL) *rec = daqTbl[i].rec;
        if (param != NULL) *param  = j;
        return i;
      }
    }
  }
  return -1;
}
int getEntryFromAssociation(char *assoc,int start,int *rec,int *param)
{
  DAQParameter *daqParam;
  int i,j;

  if (assoc[0] == '[') return -1; /* used for range information for a waveform keyword */
  for (i=start; i<nrecords; i++)
  {
    if (!daqTbl[i].active) continue;
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (!strnicmp(daqParam[j].assoc,assoc,PROPERTY_NAME_SIZE))
      {
        if (rec != NULL) *rec = daqTbl[i].rec;
        if (param != NULL) *param  = j;
        return i;
      }
    }
  }
  return -1;
}
int getNumLines(FILE *fp)
{
  int i;
  char str[256];

  for (i=0; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    i++;
  }
  rewind(fp);
  return i-1;   /* subtract the header */
}
/**
 * \intern
 * \brief open the targeted data file for reading
 *
 * \param daqr (input) pointer to the data record (with the targeted file
 *  name filled in; (output) the 'hasLegacyStorage' field can be set.
 * \param (input) if non-zero, instructs the routine to explicitly open the
 *  alternative file; (output) indicates if an alternative file was found
 *
 * \note: an alternative file will only exist if the storage type (HRT or not)
 * changed within the month of the targeted file.  Normally there is no 
 * alternative file.
 *
 */
int OpenDataFile(DAQRecord *daqr,int *hasAlternate)
{
  char filenm[256],altname[256],afile[32],altfile[32], *fn;
  int hDataFile = -1, useAltFile = FALSE;
  struct stat sb;
  strncpy(afile,daqr->fname,16);
  strncpy(altfile,daqr->fname,16);
  if (!daqr->useMinimalStorage)
  { /* configured to use HRT */
    altfile[0] = 'a'; altfile[1] = 'r';
  }
  else
  {
    altfile[0] = 't'; altfile[1] = 'a';
  }
  if (!daqr->useMinimalStorage && daqr->hasLegacyStorage)
  { 
    useAltFile = TRUE;
  }
  if (hasAlternate != NULL && *hasAlternate == TRUE) useAltFile = TRUE;
# ifdef WIN32
# define O_readmode O_RDONLY|O_BINARY
  sprintf(filenm,"%s\\%04d\\%02d\\%s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,afile);
  sprintf(altname,"%s\\%04d\\%02d\\%s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,altfile);
# else
# define O_readmode O_RDONLY
  sprintf(filenm,"%s/%04d/%02d/%s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,afile);
  sprintf(altname,"%s/%04d/%02d/%s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,altfile);
# endif
  if (hasAlternate != NULL) *hasAlternate = (stat(altname,&sb) == 0) ? TRUE : FALSE;
retryOpen:
  fn = useAltFile ? altname : filenm;
  hDataFile = open(fn,O_readmode);
  if (hDataFile < 0) 
  {
    if (!daqr->useMinimalStorage && !daqr->hasLegacyStorage)
    { /* maybe was stored another way ? */
      daqr->hasLegacyStorage = TRUE;
      useAltFile = TRUE;
      goto retryOpen;
    }
    if (archdebug > 1) dbglog("error opening file %s for reading : %d",filenm, errno);
  }
  return hDataFile;
}
int inputDataLoad(int ntotal, int offset)
{
  int i, n;
  char str[256],scratch[128];
  int done = 0,cc = 0, fsiz, fmt = CF_INT16;
  int dat_col=-1,fmt_col=-1;
  BYTE *indata;
  FILE *fp;

  for (i=offset; i<ntotal+offset; i++)
  {
    daqTbl[i].infmt = CF_NULL; daqTbl[i].inlen = 0;
    if (strlen(daqTbl[i].fn_inp) == 0) continue;
    sprintf(str,"%sinput/%.80s",ArchiveDBpath,daqTbl[i].fn_inp);
    if ((fp=fopen(str,"rt")) == NULL)
    { // not found ...
      if (daqTbl[i].active)
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "%s : Can't open input file %.72s for active record", daqTbl[i].srv, daqTbl[i].fn_inp);
        sprintf(daqUpdateErrStr, "cannot open input file %.32s for rec %d", daqTbl[i].fn_inp, daqTbl[i].rec);
      }
      memset(daqTbl[i].fn_inp, 0, 80); // blow the reference away (ADBM bug)
      continue;
    }
    if ((daqTbl[i].inlen=getNumLines(fp)) <= 0)
    {
      tineDebug = 1;
      feclogEx(TINE_LOGLEVEL_ERROR,"%s : no records in input file !!",daqTbl[i].srv);
      continue;
    }
    indata = (BYTE *) calloc(daqTbl[i].inlen,64);
    cc = done = 0;
    for (n=0; csvGetLine(str,255,fp) != NULL && n<daqTbl[i].inlen; )
    {
      strtrm(str);
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((dat_col=findcol(str,"DATA")) < 0) ccerr(no_such_line);
        fmt_col = findcol(str,"FORMAT");
        done = 1;
        continue;
      }
      if (n == 0 && fmt_col >= 0)
      {
        strncpy(scratch,colptr(fmt_col,str),64); strupr(scratch);
        if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
      }
      strncpy(scratch,colptr(dat_col,str),64);
      switch (fmt)
      {
        case CF_INT8:
          ((BYTE *)indata)[n] = (BYTE)atoi(scratch);
          break;
        case CF_INT16:
          ((short *)indata)[n] = (short)atoi(scratch);
          break;
        case CF_INT32:
          ((SINT32 *)indata)[n] = (SINT32)atoi(scratch);
          break;
        case CF_FLOAT:
          ((float *)indata)[n] = (float)atof(scratch);
          break;
        case CF_DOUBLE:
          ((double *)indata)[n] = (double)atof(scratch);
          break;
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          fsiz = fmtsizeof(fmt);
          strncpy((char *)indata + n*fsiz,scratch,fsiz);
          break;
      }
      n++;
    }
    daqTbl[i].indata = indata;
    daqTbl[i].infmt = fmt;
    daqTbl[i].inlen = n;
err: 
    if (cc) 
    { 
      feclogEx(TINE_LOGLEVEL_ERROR, "%s (record %d) : %.32s",str,daqTbl[i].rec,cc2str(cc));
      sprintf(daqUpdateErrStr,"rec %d %.32s",daqTbl[i].rec,cc2str(cc));
    }
    fclose(fp);
  }
  return cc;
}
void applyPoiInfo(DAQRecord *daq,DAQParameter *prm)
{
  double rng;
  if (daq->option & NOPOI) return;
  if (prm == NULL) return;
  rng = prm->maximum - prm->minimum;
  if (rng > 0)
  { /* must have a valid range specified to keep points of interest */
    if (prm->aTolerance >= rng * 0.001 || prm->pTolerance > 0)
    { /* tolerance specification look useable */
      prm->keepsPointsOfInterest = CA_USE_POI_TOLERANCE;
    }
    else
    {
      prm->keepsPointsOfInterest = CA_USE_POI_RANGE;
    }
  }
}
void applyRowInfo(DAQParameter *prm,int rowLength)
{
  prm->arrayType |= AT_DOUBLE;
  prm->rowsize = rowLength;
  prm->nrows = prm->size/rowLength;
}
void fillinArrayTypes(int num)
{
  int i,j,n,m,len,rlen;
  char *c,tgt[PROPERTY_NAME_SIZE+4];
  DaqRecKey *dqa = NULL;

  for (i=0,tgt[PROPERTY_NAME_SIZE]=0; i<num; i++)
  {
    for (n=0; n<daqTbl[i].nparam; n++)
    {
      if (daqTbl[i].daqParam == NULL)
      {
        daqTbl[i].active = FALSE;
        feclogEx(TINE_LOGLEVEL_ERROR,"no keyword parameters loaded for record %d",daqTbl[i].rec);
        sprintf(daqUpdateErrStr,"no keyword parameters loaded for rec %d",daqTbl[i].rec);
        continue;
      }
      if (daqTbl[i].daqParam[n].arrayType != AT_UNKNOWN) continue; /* done */
      daqTbl[i].daqParam[n].arrayType = AT_SPECTRUM|AT_SINGLE; /* default */
      daqTbl[i].daqParam[n].rowsize = daqTbl[i].daqParam[n].size;
      if ((len=(int)strlen(daqTbl[i].daqParam[n].assoc)) > 0)
      { /* associated information given */
        if (daqTbl[i].daqParam[n].assoc[0] == '[')
        { /* range information for a SPECTRUM, aka waveform or trace array */
          strncpy(tgt,&daqTbl[i].daqParam[n].assoc[1],PROPERTY_NAME_SIZE-1);
          if ((c=strchr(tgt,']')) != NULL) *c = 0;
          if ((c=strchr(tgt,':')) != NULL)
          { 
            daqTbl[i].daqParam[n].xminimum = (float)strtod(tgt,NULL); c++;
            daqTbl[i].daqParam[n].xmaximum = (float)strtod(c,&c); 
          }
          else 
          {
            c = tgt;
          }
          while (*c == ' ') c++;
          if (*c != 0) strncpy(daqTbl[i].daqParam[n].xunits,c,UNITS_SIZE);
        }
        else if (!strcmp(daqTbl[i].daqParam[n].assoc,".NAM"))
        { /* associating device names -> can only be an MCA */
          daqTbl[i].daqParam[n].arrayType = AT_CHANNEL;
        }
        else if (!strcmp(daqTbl[i].daqParam[n].assoc,".NAM.NAM"))
        { /* associating device names -> can only be an MCA */
          daqTbl[i].daqParam[n].arrayType = AT_CHANNEL;
        }
      }
      if (daqTbl[i].daqParam[n].size == 1) 
      { /* can only be a scalar */
        daqTbl[i].daqParam[n].arrayType = AT_SCALAR;
        applyPoiInfo(&daqTbl[i],&daqTbl[i].daqParam[n]);
      }
      else
      { /* an array of some sort */
        if (daqTbl[i].fmt == CF_STRUCT) 
        { /* peg this as a channel ! */
          daqTbl[i].daqParam[n].arrayType = AT_CHANNEL;
        }
        rlen = 0;
        strncpy(tgt,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE);
        if ((len=(int)strlen(tgt)) > 4 && !strcmp(&tgt[len-4],".NAM"))
        { /* landed on a .NAM keyword ! */
          if (strstr(tgt,".NAM.NAM") != NULL) 
          { /* this is giving sub-device name information : any associations here? */
            for (dqa=gDaqRecAssoc; dqa != NULL; dqa=dqa->nxt)
            {
              if (!strnicmp(dqa->key,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE))
              { /* got one : get the .NAM guy */
                tgt[len-4] = 0;
                if ((j=getEntryFromKeyword(tgt,NULL,&m)) < 0) continue;
                daqTbl[dqa->idx].daqParam[dqa->prm].arrayType = AT_CHANNEL;
                if ((rlen=(int)daqTbl[j].daqParam[m].scaleOffset) > 0)
                { /* .NAM carries a scaleOffset -> partner is a double array */
                  applyRowInfo(&daqTbl[dqa->idx].daqParam[dqa->prm],rlen);
                }
              }
            }
            continue;
          }
          if (!(daqTbl[i].daqParam[n].arrayType & AT_CHANNEL)) daqTbl[i].daqParam[n].arrayType = AT_CHANNEL;
          /* find its partner ... */
          tgt[len-4] = 0;
          if ((j=getEntryFromKeyword(tgt,NULL,&m)) < 0 || daqTbl[j].daqParam == NULL) continue;
          if (!(daqTbl[j].daqParam[m].arrayType & AT_CHANNEL)) daqTbl[j].daqParam[m].arrayType = AT_CHANNEL;
          if ((rlen=(int)daqTbl[i].daqParam[n].scaleOffset) > 0)
          { /* .NAM carries a scaleOffset -> partner is a double array */
            applyRowInfo(&daqTbl[j].daqParam[m],rlen);
          }
          /* CHANNEL arrays are allowed to have points of interest (SPECTRUM arrays aren't!) */
          applyPoiInfo(&daqTbl[j],&daqTbl[j].daqParam[m]);
          /* find any associations ... */
          for (dqa=gDaqRecAssoc; dqa != NULL; dqa=dqa->nxt)
          {
            if (!strnicmp(dqa->key,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE))
            {
              j = dqa->idx; m = dqa->prm;
              if (rlen > 0) applyRowInfo(&daqTbl[j].daqParam[m],rlen);
              applyPoiInfo(&daqTbl[j],&daqTbl[j].daqParam[m]);
            }
          }
        }
        else if (len > 5 && !strcmp(&tgt[len-5],".ENUM"))
        { /* landed on an .ENUM keyword ! */
          daqTbl[i].daqParam[n].arrayType = AT_UNKNOWN;
          if ((rlen=(int)daqTbl[i].daqParam[n].scaleOffset) > 0)
          {
            applyRowInfo(&daqTbl[i].daqParam[n],rlen);
          }
        }
        else 
        { /* this entry is neither a .NAM or .ENUM */
          switch (daqTbl[i].daqParam[n].fmt)
          {
            case CF_TEXT: 
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME48:
            case CF_NAME64:
              daqTbl[i].daqParam[n].arrayType = AT_UNKNOWN;
              break;
            default:
              break;
          }
        }
      }
    }
  }
  for (i=0; i<nPrpAlias; i++)
  {
    strncpy(tgt,prpAlias[i].name,PROPERTY_NAME_SIZE);
    if ((len=(int)strlen(tgt)) > 4 && !strcmp(&tgt[len-4],".NAM"))
    { /* find its partner ... */
      tgt[len-4] = 0;
      if ((j=getEntryFromKeyword(tgt,NULL,&m)) < 0) continue;
      if (daqTbl[j].daqParam == NULL || m >= daqTbl[j].nparam) continue;
      if (!(daqTbl[j].daqParam[m].arrayType & AT_CHANNEL)) daqTbl[j].daqParam[m].arrayType = AT_CHANNEL;
    }
  }
  return;
}
void checkDaqSubSystem(char *subsys)
{
  int i;
  if (subsys == NULL || strlen(subsys) == 0) return;
  for (i=0; i<nSubSystems; i++)
  {
    if (!strncmp(subSystems[i].name,subsys,SUBSYSTEM_NAME_SIZE)) goto checkOut;
  }
  if (nSubSystems == recordsCapacity) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"additional subsystem %.16s would exceed capacity %d",subsys,recordsCapacity);
    return;
  }
  strncpy(subSystems[i].name,subsys,SUBSYSTEM_NAME_SIZE);
  nSubSystems++;
checkOut:
  subSystems[i].number++;
}
int getFilterKeyAndIndex(int length,int tblIdx,char *fltrKey,int *fltrIdx)
{
  int idx = 0;
  char *c;
  if (fltrKey == NULL || fltrIdx == NULL) return argument_list_error;
  if (tblIdx < 0 || tblIdx >= nDaqFilterTblEntries) return dimension_error;
  strncpy(fltrKey,daqFilterTbl[tblIdx].key,64); 
  if ((c=strchr(fltrKey,'/')) != NULL)
  {
    *c++ = 0; 
    if (*c == '#') c++;
    idx = atoi(c);
  }
  *fltrIdx = idx;
  return 0;
}
int getArchivedEntryFromKeyword(char *keyword,int *rec,int *param)
{
  DAQParameter *daqParam;
  int i,j;

  for (i=0; i<totalNrRecords; i++)
  {
    if (!daqTbl[i].active) continue;
    if (daqTbl[i].option&NEVER) continue;
    daqParam = daqTbl[i].daqParam;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (!stricmp(daqParam[j].param,keyword))
      {
        *rec = daqTbl[i].rec;
        *param  = j;
        return i;
      }
    }
  }
  return -illegal_property;
}
int addExtSrcInfo(void)
{
  int i,n,r,j,k,p,num=0;
  char *ext,scratch[256];
  for (i=0; i<totalNrRecords; i++)
  {
    if (!daqTbl[i].active) continue;
    if (daqTbl[i].option&NEVER) continue; /* it's not being archived */
    for (n=0; n<daqTbl[i].nparam; n++)
    {
      if ((ext=strrchr(daqTbl[i].daqParam[n].param,'.')) != NULL && !strcmp(ext,".SRCADDR"))
      { /* a source address keyword : find its partner */
        daqTbl[i].option = ONCE;
        strncpy(scratch,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE); 
        scratch[PROPERTY_NAME_SIZE] = 0;
        p = (int)strlen(scratch) - 8;
        scratch[p] = 0;
        j = getArchivedEntryFromKeyword(scratch,&r,&k);
        if (j >= 0)
        {
          daqTbl[j].daqParam[k].srcIndex = i + 1; /* shuffle +/- 1 (k is always 0) */
          num++;
        }
      }
    }
  }
  return num;
}

#define isNumberFormat(f) (\
  (f) == CF_INT16 ? TRUE :\
  (f) == CF_INT32 ? TRUE :\
  (f) == CF_INT64 ? TRUE :\
  (f) == CF_FLOAT ? TRUE :\
  (f) == CF_DOUBLE ? TRUE : FALSE\
)
void checkFilterFormat(DAQParameter *p)
{
  int i,cc,fltrIdx=0;
  char fltrKey[64];
  for (i=nDaqStockFilterNames; i<nDaqFilterTblEntries; i++)
  {
    if ((cc=getFilterKeyAndIndex(p->size,i,fltrKey,&fltrIdx)) != 0) 
    {
      feclogEx(TINE_LOGLEVEL_ERROR,"filter parameter %.64s not accepted as filter : %.32s",p->param,cc2str(cc));
      continue;
    }
    if (!strnicmp(p->param,fltrKey,64))
    { 
      if (fltrIdx < 0 || fltrIdx >= p->size) 
      {
        feclogEx(TINE_LOGLEVEL_ERROR,"filter parameter %.64s not accepted as filter : invalid index",p->param);
        continue;
      }
      daqFilterTbl[i].active = TRUE;
      if (isStringFormat((short)p->fmt) && isStringFormat((short)daqFilterTbl[i].fmtValidMin)) break;
      if (isNumberFormat(p->fmt) && isNumberFormat(daqFilterTbl[i].fmtValidMin)) break;
      if (p->fmt != daqFilterTbl[i].fmtValidMin)
      {
        feclogEx(TINE_LOGLEVEL_ERROR,"filter parameter %.64s has a possible format mis-match",p->param);
        feclogEx(TINE_LOGLEVEL_ERROR,"archive database : %s, filter table : %s",itof(p->fmt),itof(daqFilterTbl[i].fmtValidMin));
      }
      break;
    }
  }
}
DAQParameter *getDaqParameter(char *key)
{
  int i,n;
  for (i=0; i<nrecords; i++)
  {
    if (!daqTbl[i].active) continue;
    if (daqTbl[i].daqParam == NULL) continue;
    for (n=0; n<daqTbl[i].nparam; n++)
    {
      if (!strnicmp(key,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE))
        return daqTbl[i].daqParam;
    }
  }
  return NULL;
}
/**
 * \internal
 * \param lst is a doublet list of device-alias pairs alternating
 * \param len is the length of the list -> len/2 doublets
 */
int writePriorDevices(char *keyword, NAME64 *lst, int len)
{
  int i, cc = 0;
  char str[256];
  FILE* fp = NULL;
  if (keyword == NULL) return argument_list_error;
  if (lst == NULL || len == 0) ccerr(invalid_parameter);
  sprintf(str, "%sinfo/prior/%.64s-devices.csv", ArchiveDBpath, keyword);
  if ((fp=fopen(str, "wt")) == NULL) ccerr(no_such_file);
  fprintf(fp, "Device, Alias\n");
  for (i = 0; i < len/2; i++)
  {
    fprintf(fp, "%.64s, %.64s\n", lst[2*i].name, lst[2*i+1].name);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0)
  { /* oops! */
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot write prior devices for %.64s : %s", keyword, cc2str(cc));
  }
  else
  { /* who's using this set ? */
    DAQParameter *daqParam;
    int j, k;
    for (i = 0; i < totalNrRecords; i++)
    {
      if (!daqTbl[i].active) continue;
      if (daqTbl[i].option & NEVER) continue;
      daqParam = daqTbl[i].daqParam;
      for (j = 0; j < daqTbl[i].nparam; j++)
      {
        if (daqParam[j].npriors == 0) continue;
        for (k = 0; k < daqParam[j].npriors; k++)
        {
          if (!stricmp(keyword, daqTbl[i].daqParam[j].pr[k].keyword))
          { /* found one: reload 'em */
            loadPriorDevices(&daqParam->pr[k]);
            break;
          }
        }
      }
    }
  }
  return cc;
}
int loadPriorDevices(PriorReference *pr)
{
  int n = 0, cc = 0, done = 0, nd;
  int dev_col=-1, ali_col=-1;
  char str[256], *ali;
  FILE *fp = NULL;

  sprintf(str,"%sinfo/prior/%.64s-devices.csv",ArchiveDBpath,pr->keyword);
  if ((fp=fopen(str,"rt")) == NULL) ccerr(no_such_file);
  if ((nd=getNumLines(fp)) <= 0)
  {
    tineDebug = 1;
    feclogEx(TINE_LOGLEVEL_ERROR,"%s : no entries in priors device file !!",str);
    ccerr(file_error);
  }
  if (pr->devlst != NULL) free(pr->devlst);
  if (pr->aliaslst != NULL) free(pr->aliaslst);
  if ((pr->devlst=(NAME64 *)calloc(nd,sizeof(NAME64))) == NULL) ccerr(out_of_local_memory);
  if ((pr->aliaslst=(NAME64 *)calloc(nd,sizeof(NAME64))) == NULL) ccerr(out_of_local_memory);
  pr->ndevs = nd;
  for (n=0; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((dev_col=findcol(str,"DEVICE")) < 0) ccerr(no_such_line); // the prior device name
      ali_col = findcol(str,"ALIAS");  // should map to the 'current' device name
      done = TRUE;
      continue;
    }
    strncpy(pr->devlst[n].name,colptr(dev_col,str),64);
    ali = colptr(ali_col, str);
    if (strlen(ali) == 0) ali = pr->devlst[n].name;
    strncpy(pr->aliaslst[n].name, ali, 64);
    n++;
  }

err:
  if (fp != NULL) fclose(fp);
  if (cc != 0 && cc != no_such_file)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"cannot open priors file : %s",str);
  }
  return n;
}
#define prHdr "Keyword, Begin, End"
int addDaqPrmPriors(DAQParameter *daqParam, NAME64DBLDBL *priors, int npriors)
{
  int i, cc = 0;
  char str[256],startt[64],stopt[64];
  FILE *fp = NULL;
  sprintf(str, "%sinfo/prior/%.64s.csv", ArchiveDBpath, daqParam->param);
  makeDirPath(str);
  if ((fp=fopen(str, "wt")) == NULL) ccerr(no_such_file);
  daqParam->npriors = 0;
  if (daqParam->pr != NULL) free(daqParam->pr);
  if ((daqParam->pr = (PriorReference *)calloc(npriors, sizeof(PriorReference))) == NULL) ccerr(out_of_local_memory);
  if (fprintf(fp, "%s\n", prHdr) <= 0) ccerr(file_error);
  for (i = 0; i < npriors; i++)
  {
    if (priors[i].d1val > 1)
      strncpy(startt, GetDataTimeString(priors[i].d1val, FALSE), 17);
    else
      startt[0] = 0;
    strncpy(stopt, GetDataTimeString(priors[i].d2val, FALSE), 17);
    if (fprintf(fp, "%.64s, %.17s, %.17s\n", priors[i].name, startt, stopt) <= 0) ccerr(file_error);
    strncpy(daqParam->pr[i].keyword, priors[i].name, 64);
    daqParam->pr[i].begin = (time_t)priors[i].d1val;
    daqParam->pr[i].end = (time_t)priors[i].d2val;
  }
  daqParam->npriors = npriors;
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0 && cc != no_such_file)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"cannot write priors file : %s", str);
  }
  return cc;
}
int getKeywordsWithPriors(NAME64 *keys, int *len)
{
  int cnt = 0, r, p, n = *len;
  char *c, dirname[256], fname[128];
# if defined(UNIX)
  DIR* dp;
  struct dirent* dirp;
# elif defined(WIN32)
  struct _finddata_t fp;
  long hfp;
  char fn[264];
# endif

  sprintf(dirname, "%sinfo/prior/", ArchiveDBpath);
# if defined(UNIX)
  if ((dp=opendir(dirname)) == NULL) goto fin;
  while ((dirp=readdir(dp)) != NULL && cnt < n)
  {
    strncpy(fname, dirp->d_name, 96); fname[96] = 0;
# elif defined(WIN32)
  sprintf(fn, "%s*.*", dirname);
  if ((hfp=_findfirst(fn, &fp)) == -1) goto fin;
  do
  {
    strncpy(fname, fp.name, 96); fname[96] = 0;
# endif
    if (strstr(fname, "-devices.csv") != NULL) continue;
    if ((c=strstr(fname, ".csv")) == NULL) continue;
    *c = 0;
    if (getArchivedEntryFromKeyword(fname, &r, &p) < 0) continue;
    strncpy(keys[cnt++].name, fname, 64);
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp, &fp) != -1 && cnt < n);
  _findclose(hfp);
# endif
fin:
  *len = cnt;
  return 0;
}
int loadDaqParamPriorReferences(DAQParameter *daqParam)
{ // internal: don't check for null arguments ...
  int n = 0, cc = 0, done = 0, np;
  int key_col, beg_col=-1, end_col;
  char str[256],scratch[128];
  FILE *fp = NULL;

  sprintf(str,"%sinfo/prior/%.64s.csv",ArchiveDBpath,daqParam->param);
  if ((fp=fopen(str,"rt")) == NULL) ccerr(no_such_file);
  if ((np=getNumLines(fp)) <= 0)
  {
    tineDebug = 1;
    feclogEx(TINE_LOGLEVEL_ERROR,"%s : no entries in priors file !!",str);
    ccerr(file_error);
  }
  if ((daqParam->pr=(PriorReference *)calloc(np,sizeof(PriorReference))) == NULL) ccerr(out_of_local_memory);
  daqParam->npriors = np;
  for (n=0; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line); // 'param' maps to this old 'keyword'
      if ((end_col=findcol(str,"END")) < 0) ccerr(no_such_line);
      beg_col = findcol(str,"BEGIN");
      done = TRUE;
      continue;
    }
    strncpy(daqParam->pr[n].keyword,colptr(key_col,str),64);
    strncpy(scratch,colptr(end_col,str),64); 
    daqParam->pr[n].end = parseTimeString(scratch);
    strcpy(scratch,"0");
    if (beg_col >= 0)
    {
      strncpy(scratch,colptr(beg_col,str),64); 
    }
    daqParam->pr[n].begin = parseTimeString(scratch);
    // now discover alias list ...
    loadPriorDevices(&daqParam->pr[n]);
    n++;
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0 && cc != no_such_file)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"cannot open priors file : %s",str);
  }
  return n;
}
int loadDaqParamDb(int ntotal, int offset)
{
  int i, n, num, recsize, psize, fsize, fmt;
  char *s1, *s2, *ext, str[256],scratch[128];
  char dnamePrp[PROPERTY_NAME_SIZE+4];
  int done = 0,cc = 0,dbg;
  int key_col,fmt_col,siz_col,uni_col,max_col,min_col,
      tol_col,plt_col,cmt_col,scale_col,offset_col,
      assoc_col,bnd_col,sub_col;
  DAQParameter *daqParam;
  FILE *fp;

  dbg = tineDebug; /* preserve debug state */
  tineDebug = 1; /* turn on to observe errors */
  for (i=offset, num=0; i<ntotal+offset; i++)
  {
    sprintf(str,"%sinfo/%.80s",ArchiveDBpath,daqTbl[i].fn_key);
    if ((fp=fopen(str,"rt")) == NULL)
    {
      char *tag = daqTbl[i].active ? "active" : "inactive";
      int errlev = daqTbl[i].active ? TINE_LOGLEVEL_ERROR : TINE_LOGLEVEL_WARN;
      feclogEx(errlev,"%s (record %s): Can't open parameter file %.72s",daqTbl[i].srv,tag,daqTbl[i].fn_key);
      sprintf(daqUpdateErrStr,"cannot open parameter file %.32s for rec %d",daqTbl[i].fn_key,daqTbl[i].rec);
      daqTbl[i].active = 0;
      continue;
    }
    if ((daqTbl[i].nparam=getNumLines(fp)) <= 0)
    {
      tineDebug = 1;
      feclogEx(TINE_LOGLEVEL_ERROR,"%s : no records in parameter file !!",daqTbl[i].srv);
      sprintf(daqUpdateErrStr,"no records in parameter file %.32s for rec %d",daqTbl[i].fn_key,daqTbl[i].rec);
      return file_error;
    }
    if (daqTbl[i].hasDataPool && daqTbl[i].nparam > 1)
    {
      tineDebug = 1;
      feclogEx(TINE_LOGLEVEL_ERROR,"%s : parameter file for complex type contains more than 1 parameter !!",daqTbl[i].srv);
      sprintf(daqUpdateErrStr,"too many parameters for complex type: file %.80s for rec %d",daqTbl[i].fn_key,daqTbl[i].rec);
      return not_allowed;
    }
    daqParam = (DAQParameter *) calloc(daqTbl[i].nparam,sizeof(DAQParameter));
    recsize = 0; done = 0;
    for (n=0; csvGetLine(str,255,fp) != NULL; )
    {
      strtrm(str);
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
        if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
        if ((siz_col=findcol(str,"SIZE")) < 0) ccerr(no_such_line);
        if ((uni_col=findcol(str,"UNITS")) < 0) ccerr(no_such_line);
        if ((max_col=findcol(str,"MAXIMUM")) < 0) ccerr(no_such_line);
        if ((min_col=findcol(str,"MINIMUM")) < 0) ccerr(no_such_line);
        if ((tol_col=findcol(str,"TOLERANCE")) < 0) ccerr(no_such_line);
        if ((plt_col=findcol(str,"PLOT")) < 0) ccerr(no_such_line);
        if ((cmt_col=findcol(str,"DESCRIPTION")) < 0 &&
            (cmt_col=findcol(str,"COMMENT")) < 0) ccerr(no_such_line);
        scale_col = findcol(str,"SCALE");
        offset_col = findcol(str,"OFFSET");
        assoc_col = findcol(str,"ASSOCIATE");
        if (assoc_col < 0) assoc_col = findcol(str,"DEVICES");
        bnd_col = findcol(str,"BIND");
        sub_col = findcol(str,"SUBSYSTEM");
        done = 1;
        continue;
      }
      strncpy(daqParam[n].param,colptr(key_col,str),PROPERTY_NAME_SIZE);
      strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
      if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
      daqParam[n].fmt = fmt;
      if (daqTbl[i].fmt == CF_STRUCT && fmt != CF_STRUCT) 
      {
        feclogEx(TINE_LOGLEVEL_ERROR,"structure source /%.32s/%.32s[%.64s] mapping to %.64s data type %s not allowed!",
          daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].prp,daqParam[n].param,itof(fmt));
        continue;
      }
      strncpy(scratch,colptr(siz_col,str),16);
      daqParam[n].size = atoi(scratch);
      strncpy(daqParam[n].units,colptr(uni_col,str),UNITS_SIZE);
      strncpy(scratch,colptr(max_col,str),16);
      daqParam[n].maximum = (float)atof(scratch);
      strncpy(scratch,colptr(min_col,str),16);
      daqParam[n].minimum = (float)atof(scratch);
      strncpy(scratch,colptr(tol_col,str),16);
      s1 = scratch;
      if ((s2=strchr(s1, '+')) != NULL) *s2++ = 0;
      if (strchr(s1, '%'))
      {
        daqParam[n].pTolerance = (float)(atof(s1) / 100.0);
        if (s2 != NULL) daqParam[n].aTolerance = (float)atof(s2);
      }
      else
      {
        daqParam[n].aTolerance = (float)atof(s1);
        if (s2 != NULL && strchr(s2,'%')) daqParam[n].pTolerance = (float)(atof(s2) / 100.0);
      }
      strncpy(daqParam[n].desc,colptr(cmt_col,str),PROPERTY_DESC_SIZE);
      strncpy(daqParam[n].plot,colptr(plt_col,str),UNITS_SIZE);
      daqParam[n].offset = recsize;
      fsize = daqParam[n].fmt == CF_STRUCT ? GetStructSize(daqTbl[i].tag) : fmtsizeof(daqParam[n].fmt);
      psize = daqParam[n].size * fsize + getFormatHeaderSize(daqParam[n].fmt);
      if (isVariableLengthFormat((short)daqParam[n].fmt)) psize = 2 * sizeof(size_t);
      recsize += psize;
      daqParam[n].scaleFactor = 1;
      if (scale_col > 0)
      {
        strncpy(scratch,colptr(scale_col,str),16);
        daqParam[n].scaleFactor = (float)atof(scratch);
      }
      if (offset_col > 0)
      {
        strncpy(scratch,colptr(offset_col,str),16);
        daqParam[n].scaleOffset = (float)atof(scratch);
      }
      if (bnd_col > 0)
      {
        strncpy(daqParam[n].boundTo,colptr(bnd_col,str),PROPERTY_NAME_SIZE);
      }
      if (sub_col > 0)
      {
        strncpy(daqParam[n].subsys,colptr(sub_col,str),SUBSYSTEM_NAME_SIZE);
        if (!stricmp("ALL",daqParam[n].subsys)) strcpy(daqParam[n].subsys,"Diagnostics");
      }
      else
      {
        strncpy(daqParam[n].subsys,"Diagnostics",16);
      }
      if (daqTbl[i].active)
      { /* only assign association prp alias for active entries */
        if (assoc_col > 0 && nPrpAlias < totalNrRecords)
        { /* assoc_col -> the 'associate' column 
            in case "ABC" gets its device names from "DEF"
            daqParam[n].param = "ABC", assoc_col -> "DEF.NAM"
            then "ABC.NAM" is formed and matched to the alias "DEF.NAM"
          */
          strncpy(scratch,colptr(assoc_col,str),PROPERTY_NAME_SIZE); scratch[PROPERTY_NAME_SIZE] = 0;
          if (strlen(scratch) > 0)
          { /* associate column present and not empty ! */
            strncpy(daqParam[n].assoc,scratch,PROPERTY_NAME_SIZE);
            memset(dnamePrp,0,PROPERTY_NAME_SIZE+4);
            if (daqParam[n].assoc[0] != '[')
            { /* not a range info association */
              //addDaqRecAssoc(daqParam[n].assoc,daqParam[n].param,i,daqTbl[i].rec,n);
              s1 = scratch; 
              if ((s2=strchr(s1,'+')) != NULL) *s2++ = 0;
              if ((ext=strrchr(s1,'.')) != NULL)
              { /* this should by all accounts be true (ext = ".NAM" or ".ENUM") */
                /* s1 will contain the fully qualified primary association */
                if (strstr(s1,".NAM.NAM") != NULL)
                { /* very special case ! */
                  addDaqRecAssoc(s1, daqParam[n].param, i, daqTbl[i].rec, n);
                  strncpy(prpAlias[nPrpAlias].alias,s1,PROPERTY_NAME_SIZE);
                  strncpy(dnamePrp,daqParam[n].param,PROPERTY_NAME_SIZE-8); 
                  strcat(dnamePrp,".NAM.NAM");
                  strncpy(prpAlias[nPrpAlias].name,dnamePrp,PROPERTY_NAME_SIZE);
                  nPrpAlias++;
                  /* automatically associate the .NAM as well */
                  *ext = 0; ext = ".NAM"; /* the only remaining option */
                }
                addDaqRecAssoc(s1, daqParam[n].param, i, daqTbl[i].rec, n);
                strncpy(prpAlias[nPrpAlias].alias,s1,PROPERTY_NAME_SIZE);
                strncpy(dnamePrp,daqParam[n].param,PROPERTY_NAME_SIZE-5); 
                strcat(dnamePrp,ext);
                strncpy(prpAlias[nPrpAlias].name,dnamePrp,PROPERTY_NAME_SIZE);
                nPrpAlias++;
              }
              if (s2 != NULL && strlen(s2) > 0)
              { /* a double 'association' */
                if (s2[0] == '.')
                { /* can only be the same property */
                  *ext = 0; /* trim former extension from s1 so s1 points to base property */
                  ext = s2; /* and reassign */
                  strncpy(prpAlias[nPrpAlias].alias,s1,PROPERTY_NAME_SIZE);
                  strcat(prpAlias[nPrpAlias].alias,ext);
                }
                else
                { /* some other property given */
                  ext = strrchr(s2,'.');
                  strncpy(prpAlias[nPrpAlias].alias,s2,PROPERTY_NAME_SIZE);
                }
                /* prpAlias[nPrpAlias].alias should now contain the 2nd association */
                addDaqRecAssoc(prpAlias[nPrpAlias].alias,daqParam[n].param,i,daqTbl[i].rec, n);
                if (ext != NULL)
                { /* this should by all accounts be true (ext = ".NAM" or ".ENUM") */
                  strncpy(dnamePrp,daqParam[n].param,PROPERTY_NAME_SIZE-5); 
                  strcat(dnamePrp,ext);
                  strncpy(prpAlias[nPrpAlias].name,dnamePrp,PROPERTY_NAME_SIZE);
                  nPrpAlias++;
                }
              }
            }
          }
        }
        /* only scan the active entries for configured subsystems */
        checkDaqSubSystem(daqParam[n].subsys);
        loadDaqParamPriorReferences(&daqParam[n]);
      }
      checkFilterFormat(&daqParam[n]);
      if (getDaqParameter(daqParam[n].param) != NULL) ccerr(name_already_exists);
      if (daqTbl[i].hasVolatileMca && daqTbl[i].nparam > 1)
      { /* unusual but need to have a correct 'count' */
        gNumVolatileDaqTblEntries += (daqTbl[i].hasVolatileMca && daqTbl[i].nparam - 1);
      }
      n++;
    }
err: 
    fclose(fp);
    if (cc != 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR,"[%d] /%.32s/%.32s/%.64s[%.64s] (%.80s) : %.32s",
        i,daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp,daqTbl[i].fn_key,cc2str(cc)); 
      daqTbl[i].active = 0;
      daqTbl[i].nparam = 0;
      sprintf(daqUpdateErrStr,"rec %d : %.32s",daqTbl[i].rec,cc2str(cc));
      continue;
    }
    if (recsize != (int)daqTbl[i].reclen)
    {
      feclogEx(TINE_LOGLEVEL_ERROR,"[%d] /%.32s/%.32s/%.64s[%.64s] (%.80s) : record length mis-match",
        i,daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp,daqTbl[i].fn_key); 
      daqTbl[i].active = 0;
      daqTbl[i].nparam = 0;
      daqTbl[i].rec = -1;
      sprintf(daqUpdateErrStr,"record length mis-match for rec %d",daqTbl[i].rec);
      continue;
    }
    daqTbl[i].daqParam = daqParam;
    num += n;
  }
  tineDebug = dbg; /* restore debug state */
  return num;
}
int getNumActiveRecords(char *fname)
{
  FILE *fp = NULL;
  char fn[128];
  char str[256],scratch[128];
  int done = 0,i,cc = 0;
  int active_col;

  sprintf(fn,"%s%.64s",ArchiveDBpath,fname);
  if ((fp = fopen(fn,"rt")) == NULL)
  {
    feclog("Can't open data base file : %s ! (starting new?)",fname);
    return 0;
  }

  for (i=0; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((active_col=findcol(str,"ACTIVE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(active_col,str),16);
    if (atoi(scratch)) i++;
  }
err:
  fclose(fp);
  allocActiveOnly = TRUE;
  return i;
}
int loadDaqDb(char *fname, int offset) /* returns number of records allocated */
{
  FILE *fp;
  char fn[128], str[256],scratch[128];
  char *ctx;
  int done = 0,i,fmt,mod,fsiz,rlen,cc = 0;
  int active_col,name_col,property_col,device_col,datalen_col,fmt_col,
      heartbeat_col,poll_col,filter_col,sfile_col,rectype_col,cntxt_col;
  int inp_col = -1, mod_col=-1;

  sprintf(fn,"%s%.64s",ArchiveDBpath,fname);
  if ((fp = fopen(fn,"rt")) == NULL)
  {
    tineDebug = 1;
    feclog("can't open data base file : %s",fname);
    return 0;
  }

  for (i=offset; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((rectype_col=findcol(str,"RECORD")) < 0 &&
          (rectype_col=findcol(str,"RECORD INDEX")) < 0 &&
          (rectype_col=findcol(str,"RECORD TYPE")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(str,"SERVER")) < 0 &&
          (name_col=findcol(str,"NAME")) < 0) ccerr(no_such_line);
      if ((property_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((device_col=findcol(str,"DEVICE")) < 0) ccerr(no_such_line);
      if ((datalen_col=findcol(str,"LENGTH")) < 0 &&
          (datalen_col=findcol(str,"DATA LENGTH")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0 &&
          (fmt_col=findcol(str,"RPC FORMAT")) < 0) ccerr(no_such_line);
      if ((heartbeat_col=findcol(str,"HEARTBEAT")) < 0) ccerr(no_such_line);
      if ((poll_col=findcol(str,"POLL")) < 0 &&
          (poll_col=findcol(str,"POLLING RATE")) < 0) ccerr(no_such_line);
      if ((filter_col=findcol(str,"FILTER")) < 0 &&
          (filter_col=findcol(str,"ARCHIVE")) < 0) ccerr(no_such_line);
      if ((sfile_col=findcol(str,"STRUCTURE")) < 0) ccerr(no_such_line);
      if ((active_col=findcol(str,"ACTIVE")) < 0) ccerr(no_such_line);
      if ((inp_col=findcol(str,"INPUT")) < 0) inp_col = findcol(str,"DATA INPUT");
      cntxt_col = findcol(str,"CONTEXT");
      mod_col = findcol(str,"MODE");
      done = 1;
      continue;
    }
    daqTbl[i].active = atoi(colptr(active_col,str));
    if (allocActiveOnly && !daqTbl[i].active) continue;
    strncpy(scratch,colptr(rectype_col,str),32);
    daqTbl[i].rec = atoi(scratch);
    ctx = cntxt_col < 0 ? gDeviceContext : colptr(cntxt_col,str);
    strncpy(daqTbl[i].ctx,ctx,CONTEXT_NAME_SIZE); 
    strncpy(daqTbl[i].srv,colptr(name_col,str),EXPORT_NAME_SIZE);
    strncpy(daqTbl[i].prp,colptr(property_col,str),PROPERTY_NAME_SIZE);
    strncpy(scratch,colptr(device_col,str),DEVICE_NAME_SIZE);
    strncpy(daqTbl[i].dev,scratch,DEVICE_NAME_SIZE);
    strncpy(scratch,colptr(datalen_col,str),32);
    daqTbl[i].len = atoi(scratch);
    strncpy(scratch,colptr(fmt_col,str),32); 
    if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
    daqTbl[i].fmt = fmt;
    fsiz = fmtsizeof(fmt);
    if (fmt == CF_STRUCT)
    {
      char *c=strchr(scratch,'.');
      if (c != NULL)
      {
        c++;
        strncpy(daqTbl[i].tag,c,TAG_NAME_SIZE);
        acqRecordStruct(i);
        fsiz = GetStructSize(daqTbl[i].tag);
      }
    }
    daqTbl[i].reclen = fsiz*daqTbl[i].len + getFormatHeaderSize(fmt);
    if (isVariableLengthFormat((short)fmt))
    { /* will generally receive less data than asked for */
      daqTbl[i].hasDataPool = TRUE;
      daqTbl[i].hasPoi = FALSE;
      if (daqTbl[i].poolEntryCapacity == 0)
      { /* not yet assigned */
        daqTbl[i].poolEntryCapacity = daqTbl[i].reclen;
        switch (fmt)
        {
          case CF_STRING:
          case CF_KEYVALUE:
            daqTbl[i].poolEntryCapacity *= STRING_STD_CAPACITY;
            break;
          case CF_GSPECTRUM:
          case CF_IMAGE:
            daqTbl[i].poolEntryCapacity += 2 * sizeof(size_t); /* length and pointer */
            break;
          default:
            break;
        }
      }
      /* now re-assign the recordLength */
      daqTbl[i].reclen = 2 * sizeof(size_t);
    }
    rlen = daqTbl[i].reclen;
    if (gAllocDataBuffers)
    { /* the active ARCHIVER needs these buffers */
      if (daqTbl[i].hasDataPool) rlen = sizeof(DataPoolEntry);
      if ((daqTbl[i].data=(char *)calloc(3,rlen)) == NULL)
      {
        feclogEx(TINE_LOGLEVEL_ERROR,"record %d (%s [%s]: no room to allocate 3 X %d bytes",
          daqTbl[i].rec,daqTbl[i].srv,daqTbl[i].prp,daqTbl[i].reclen);
      }
      daqTbl[i].archdata = &daqTbl[i].data[rlen]; /* archive data storage */
      daqTbl[i].lastdata = &daqTbl[i].data[2*rlen]; /* last recent data storage */
      if (daqTbl[i].hasDataPool)
      { /* complex data type */
        DataPoolEntry *dpe = (DataPoolEntry *)daqTbl[i].data;
        int k;
        for (k=0; k<3; k++)
        { /* each of data, archdata, lastdata with its own data block */
          if ((dpe[k].data=(BYTE *)SystemCalloc(1,daqTbl[i].poolEntryCapacity)) == NULL) 
          {
            feclogEx(TINE_LOGLEVEL_ERROR,"record %d (%s [%s]: no room to allocate capacity of %d bytes",
              daqTbl[i].rec,daqTbl[i].srv,daqTbl[i].prp,daqTbl[i].poolEntryCapacity);
          }
          else
          {
            dpe[k].capacity = daqTbl[i].poolEntryCapacity; /* initial capacity */
          }
          if (daqTbl[i].fmt == CF_IMAGE) break; /* don't alloc arch and last buffers */
        }
      }   
    }
    strncpy(scratch,colptr(heartbeat_col,str),32);
    daqTbl[i].hbt = atoi(scratch);
    strncpy(scratch,colptr(poll_col,str),32);
    daqTbl[i].tmr = atoi(scratch);
    if (daqTbl[i].tmr <= 10) daqTbl[i].tmr = 1000;
    strncpy(scratch,colptr(filter_col,str),127); scratch[127] = 0;
    daqTbl[i].option = getFiltersMask(scratch);
    daqTbl[i].useMinimalStorage = (daqTbl[i].option & HSTHDR) ? FALSE : TRUE;
    daqTbl[i].hasSts = (daqTbl[i].option & STATUS) ? TRUE : FALSE;
    daqTbl[i].hasVolatileMca = (daqTbl[i].option & MCAVOLATILE) ? TRUE : FALSE;
    if (daqTbl[i].hasVolatileMca) gNumVolatileDaqTblEntries++;
    strncpy(daqTbl[i].fn_key,colptr(sfile_col,str),80);
    if (inp_col >= 0)
    {
      strncpy(daqTbl[i].fn_inp,colptr(inp_col,str),80);
    }
    mod = CM_TIMER;
    if (mod_col >= 0)
    {
      int m = GetTransferMode(colptr(mod_col,str));
      if (m != 0) mod = m;
      if (BASEMODE(m) == 0) mod = CM_TIMER | mod;
    }
    daqTbl[i].mod = mod;
    if (daqTbl[i].option & ONCE) 
    {
      daqTbl[i].hbt = (UINT16)(30 * 24 * 3600); /* seconds in month */
      daqTbl[i].tmr = 1000;
    }
    i++;
  }
err:
  fclose(fp);
  return (i-offset);
}
#define safefree(p) { if (p != NULL) free(p); }
int freeDaqDb(void)
{
  int i,k,p;
  for (i=0; i<nrecords; i++)
  {
    if (daqTbl[i].daqParam != NULL) 
    {
      for (p=0; p<daqTbl[i].nparam; p++)
      { // look for allocated prior references ...
        if (daqTbl[i].daqParam[p].pr == NULL) continue;
        for (k=0; k<daqTbl[i].daqParam[p].npriors; k++)
        {
          if (daqTbl[i].daqParam[p].pr[k].devlst != NULL) free(daqTbl[i].daqParam[p].pr[k].devlst);
          if (daqTbl[i].daqParam[p].pr[k].aliaslst != NULL) free(daqTbl[i].daqParam[p].pr[k].aliaslst);
        }
        free(daqTbl[i].daqParam[p].pr);
      }
      free(daqTbl[i].daqParam);
    }
    if (daqTbl[i].data != NULL)
    {
      if (daqTbl[i].hasDataPool)
      {
        safefree(((DataPoolEntry *)daqTbl[i].data)->data);
        safefree(((DataPoolEntry *)daqTbl[i].archdata)->data);
        safefree(((DataPoolEntry *)daqTbl[i].lastdata)->data);
      }
      free(daqTbl[i].data);
    }
    if (daqTbl[i].indata != NULL) free(daqTbl[i].indata);
  }
  free(daqTbl); daqTbl = NULL;
  nrecords = 0;
  if (prpAlias != NULL) free(prpAlias); prpAlias = NULL;
  nPrpAlias = 0;
  if (subSystems != NULL) free(subSystems); subSystems = NULL;
  nSubSystems = 0;
  return 0;
}
int dumpArchiveRecord(int a0,int a1,int a2,int a3)
{
  int i, idx=a0;
  a1 = a1; a2 = a2; a3 = a3;
  ttyoutput("Current status of archive record %d:",idx);
  for (i=0; i<nrecords; i++)
  {
    if (daqTbl[i].rec == idx) break;
  }
  if (i == nrecords)
  {
    ttyoutput("record index %d in not a valid entry (%d is largest record index)",idx,daqTbl[nrecords-1].rec);
    return invalid_index;
  }
  idx = i;
  ttyoutput("record source: /%.32s/%.32s/%.64s[%.64s]",
    daqTbl[idx].ctx,daqTbl[idx].srv,daqTbl[idx].dev,daqTbl[idx].prp);
  ttyoutput("record keywords:");
  for (i=0; i<daqTbl[idx].nparam; i++)
  {
    ttyoutput("%.64s",daqTbl[idx].daqParam[i].param);
  }
  ttyoutput("record is %s",daqTbl[idx].active > 0 ? "active" : "not active");
  ttyoutput("record is %s",daqTbl[idx].active ? "accessible" : "not accessible");
  ttyoutput("record last archived: %s",GetDataTimeString(daqTbl[idx].lastarchtime,FALSE));
  ttyoutput("record %s using minimal storage",daqTbl[idx].useMinimalStorage ? "is" : "is not");
  ttyoutput("record %s keeping points of interest",daqTbl[idx].daqParam[0].keepsPointsOfInterest ? "is" : "is not");
  ttyoutput("record %s volatile",daqTbl[idx].hasVolatileMca ? "is" : "is not");
  if (daqTbl[idx].option & NEVER) ttyoutput("record is not being archived");
  if (daqTbl[idx].option & ONCE) ttyoutput("record is archived once per initialization");
  ttyoutput("record filter options: %s",toFilterString(daqTbl[idx].option));
  ttyoutput("record %s passed filter criteria",isPastFilter(daqTbl[idx].option) ? "is" : "is not");
  ttyoutput("record heartbeat : %d seconds",daqTbl[idx].hbt);
  return 0;
}
void logRestart(void)        /* establish file names (1 file per month) */
{
  time_t tnow=time(NULL);
  struct tm *lt=localtime(&tnow);
  int i, mm, yy;
  char fileName[128];

  if (totalNrRecords == 0) 
  {
    AppendHistoryLog("logRestart called with no records");
    feclogEx(TINE_LOGLEVEL_WARN, "logRestart called with no records");
    return;
  }
  mm = lt->tm_mon+1; yy = lt->tm_year;

  for (i=0; i<totalNrRecords; i++)
  {
    getDataFilename(i,yy,mm);
    daqTbl[i].archYear = 1900 + yy;
    daqTbl[i].archMonth = mm;
  }
  /* make sure the directory structure is there ... */
# ifdef WIN32
  sprintf(fileName,"%s\\%04d\\%02d\\%.16s",
    ArchiveDatapath,daqTbl[0].archYear,daqTbl[0].archMonth,daqTbl[0].fname);
# else
  sprintf(fileName,"%s/%04d/%02d/%.16s",
    ArchiveDatapath,daqTbl[0].archYear,daqTbl[0].archMonth,daqTbl[0].fname);
# endif
  makeDirPath(fileName);
}
char *getAnnotationFilePath(int rec,time_t ts,int legacy)
{
  static char fileName[256];
  struct tm *lt=localtime(&ts);
  if (legacy) 
    sprintf(fileName,"%s%c%04d%c%02d%cannotate.%x%c",ArchiveDatapath,
      FS_DELIMITER,1900+lt->tm_year,FS_DELIMITER,lt->tm_mon+1,FS_DELIMITER,rec,FS_DELIMITER);
  else
    sprintf(fileName,"%s%cannotate.%x%c",ArchiveDatapath,FS_DELIMITER,rec,FS_DELIMITER);
  return fileName;
}
int annotateRecord(int rec,time_t ts,char *cmt,int len)
{
  int cc=0, slen, deleteFile=FALSE;
  FILE *fp=NULL;
  char fileName[256], *fpath;

  if (cmt == NULL || strlen(cmt) == 0) deleteFile = TRUE;
  fpath = getAnnotationFilePath(rec,ts,FALSE);
  sprintf(fileName,"%s%lx.%x",fpath,(long)ts,rec);
  makeDirPath(fileName);
  if (deleteFile) 
  {
    unlink(fileName);
    sprintf(fileName,"%s%lx.range",fpath, (long)ts);
    unlink(fileName);
    ccerr(0);
  }
  if ((fp=fopen(fileName,"w")) == NULL) ccerr(file_error);
  if ((slen=(int)strlen(cmt)) < len) len = slen;
  fwrite(cmt,1,len,fp);
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0) feclog("annotateRecord %d : %.32s",rec,cc2str(cc));
  return cc;
}
int getRecordAnnotation(int rec,time_t ts,char *cmt,int *len)
{
  int cc=0, flen;
  FILE *fp=NULL;
  char fileName[256];

  if (cmt == NULL || len == NULL || *len == 0) ccerr(invalid_parameter);
  sprintf(fileName,"%s%lx.%x",getAnnotationFilePath(rec,ts,FALSE),(long)ts,rec);
  if ((fp=fopen(fileName,"r")) == NULL) ccerr(no_such_file);
  flen = *len;
  fread(cmt,flen,1,fp);
  flen = (int)strlen(cmt);
  if (flen < *len) *len = flen;
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0) dbglog("getRecordAnnotation %d : %.32s",rec,cc2str(cc));
  return cc;
}
int probeIsActiveAnnotation(int rec,time_t atime)
{
  char dirname[256], fn[264];
  FILE *fp = NULL;
  time_t t = time(NULL);

  strncpy(dirname,getAnnotationFilePath(rec,atime,FALSE),256);
  sprintf(fn,"%s%lx.range",dirname,(long)atime);
  return file_exists(fn);
}
void aswap(time_t *ft,int i,int j)
{
  time_t tmp;
  tmp = ft[i];
  ft[i] = ft[j];
  ft[j] = tmp;
  return;
}
void asort(time_t *ft,int start,int num) 
{ /* mostly in order or sparse, so this should be quick */
  int i;

  if ((num-start) <= 1) return;
  for (i=start+1; i<num; i++) if (ft[i] < ft[start]) aswap(ft,start,i);
  asort(ft,++start,num);
  return;
}
time_t getActiveAnnotationEnd(int rec,time_t atime)
{
  char *c, dirname[256], fn[264], str[256];
  FILE *fp = NULL;
  time_t t = time(NULL);

  if (atime <= 0) return 0;
  strncpy(dirname,getAnnotationFilePath(rec,atime,FALSE),256);
  sprintf(fn,"%s%lx.range",dirname,(long)atime);
  if ((fp=fopen(fn,"r")) == NULL) return 0;
  /* if the file is there then start off with 'stop = now' */
  while (csvGetLine(str,255,fp) != NULL) 
  {
    strlwr(str);
    if ((c=strstr(str,"stop:")) != NULL)
    {
      c += 5; strtrm(c);
      if (strlen(c) == 0) break;
      t = getTimeStampFromLogEntry(c);
      break;
    }
  }
  if (fp != NULL) fclose(fp);
  return t;
}
#define FTBUF_SIZE 1024
time_t ftbuf[FTBUF_SIZE];
int _fctr = 0;
int applyActiveAnnotation(int rec,time_t atime,int cancelPrevious)
{
  time_t ft = 0;
  char dirname[256], fn[264];
  FILE *fp = NULL;
  if (cancelPrevious)
  {
    getRecordAnnotations(rec,0,atime,NULL);
    if (_fctr == 0) return 0;
    ft = ftbuf[_fctr-1];
    strncpy(dirname,getAnnotationFilePath(rec,ft,FALSE),256);
    sprintf(fn,"%s%lx.range",dirname,(long)ft);
    if (!file_exists(fn)) return 0;
    if ((fp=fopen(fn,"w")) == NULL) return 0;
    fprintf(fp,"start: %s\n",GetDataTimeString((double)ft,FALSE));
    fprintf(fp,"stop: %s\n",GetDataTimeString((double)atime,FALSE));
    fclose(fp);
  }
  else
  {
    time_t tnow = time(NULL);
    getRecordAnnotations(rec,atime,tnow,NULL);
    if (_fctr > 0)
    {
      ft = ftbuf[_fctr-1];
    }
    strncpy(dirname,getAnnotationFilePath(rec,atime,FALSE),256);
    sprintf(fn,"%s%lx.range",dirname,(long)atime);
    if ((fp=fopen(fn,"w")) == NULL) return 0;
    fprintf(fp,"start: %s\n",GetDataTimeString((double)atime,FALSE));
    if (ft > atime) fprintf(fp,"stop: %s\n",GetDataTimeString((double)ft,FALSE));
    else fprintf(fp,"stop:\n");
    fclose(fp);
  }
  return 0;
}
int getRecordAnnotations(int rec,time_t starttime,time_t stoptime,DTYPE *dout)
{
  int i,n,isa,at=0,paat=0,fctr=0,chk_active=FALSE;
  char dirname[256],fname[32],fileext[64];
  int ftim,prvtim=0;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fp;
  long hfp;
  char fn[264];
# endif
  struct tm *t;
  int startyr, startmo, stopyr, stopmo, yy, mm;
  time_t ts, aatim=0;

  _fctr = 0;
  if ((t=localtime(&starttime)) == NULL) return 0;
  startyr = t->tm_year + 1900; startmo = t->tm_mon;
  if ((t=localtime(&stoptime)) == NULL) return 0;
  stopyr = t->tm_year + 1900; stopmo = t->tm_mon;
  n = 12 * (stopyr-startyr) + stopmo - startmo + 1;
  sprintf(fileext,".%x",rec);
  for (fctr=0,i=0; i<n && fctr<FTBUF_SIZE; i++)
  {
    yy = startyr + (startmo+i)/12;
    mm = (startmo+i)%12 + 1;
    ts = starttime;
    strncpy(dirname,getAnnotationFilePath(rec,ts,FALSE),256);
#   if defined(UNIX)
    sprintf(fileext,".%x",rec);
    if ((dp=opendir(dirname)) == NULL) continue;
    while ((dirp=readdir(dp)) != NULL)
    {
      if (strstr(dirp->d_name,fileext) == NULL) continue;
      strncpy(fname,dirp->d_name,8); fname[8] = 0;
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if ((hfp=_findfirst(fn, &fp)) == -1) continue;
    do
    {
      if (strstr(fp.name,fileext) == NULL) continue;
      strncpy(fname,fp.name,8); fname[8] = 0;
#   endif
      ftim = strtol(fname,0,16);
      if (ftim > stoptime) continue;
      if (ftim < starttime) 
      { /* find the most recent old file */
        if (ftim > prvtim) prvtim = ftim;
        continue;
      }
      ftbuf[fctr] = ftim;
      if (++fctr >= FTBUF_SIZE) break; 
#   if defined(UNIX)
    }
    closedir(dp);
#   elif defined(WIN32)
    } while (_findnext(hfp,&fp) != -1);
    _findclose(hfp);
#   endif
  }
  /* need to sort these in ascending time order (just in case there are multiple hits) */
  asort(ftbuf,0,fctr);
  _fctr = fctr;
  if (dout == NULL || dout->dArrayLength <= 0) return dimension_error;
  if (dout->dFormat == CF_INTINT) chk_active = TRUE;
  n = fctr;
  aatim = getActiveAnnotationEnd(rec,(time_t)prvtim);
  if (aatim > starttime)
  { /* an active annotation from earlier time */
    paat = (int)prvtim;
    n++;
  }
  if (n < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)n;
  for (i=0; i<(int)dout->dArrayLength; i++)
  {
    if (paat == 0)
    {
      at = (int)ftbuf[i];
      isa = chk_active ? probeIsActiveAnnotation(rec,ftbuf[i]) : 0;
    }
    else
    {
      if (i == 0)
      {
        isa = TRUE; at = paat;
      }
      else
      {
        at = (int)ftbuf[i-1];
        isa = chk_active ? probeIsActiveAnnotation(rec,ftbuf[i-1]) : 0;
      }
    }
    switch (dout->dFormat)
    {
      case CF_INT32: 
        dout->data.lptr[i] = at; 
        break;
#     ifdef CF_SINT64
      case CF_INT64: 
        dout->data.lptr[i] = (SINT64)at; 
        break;
#     endif
      case CF_DOUBLE: 
        dout->data.dptr[i] = (double)at; 
        break;
      case CF_INTINT:
        ((INTINT *)dout->data.vptr)[i].i1val = at; 
        ((INTINT *)dout->data.vptr)[i].i2val = isa; 
        break;
      default: 
        return illegal_format;
    }
  }
  return 0;
}
int getRecordsWithAnnotations(time_t starttime,time_t stoptime,DTYPE *dout)
{
  int i,n,nlen,rec,fctr=0,isannotated=FALSE;
  char *c,dirname[128],fname[32],fileext[64];
  time_t ftim;
# if defined(UNIX)
  DIR *dp, *dpa;
  struct dirent *dirp, *dirpa;
# elif defined(WIN32)
  struct _finddata_t fp, fpa;
  long hfp, hfpa;
  char fn[128];
# endif
  struct tm *t;
  int startyr, startmo, stopyr, stopmo, yy, mm;
  time_t ts;
  int recbuf[256];

  if (dout == NULL || dout->dArrayLength <= 0) return dimension_error;
  if ((t=localtime(&starttime)) == NULL) return 0;
  startyr = t->tm_year + 1900; startmo = t->tm_mon;
  if ((t=localtime(&stoptime)) == NULL) return 0;
  stopyr = t->tm_year + 1900; stopmo = t->tm_mon;
  n = 12 * (stopyr-startyr) + stopmo - startmo + 1;
  for (fctr=0,i=0; i<n && fctr<(int)dout->dArrayLength; i++)
  {
    yy = startyr + (startmo+i)/12;
    mm = (startmo+i)%12 + 1;
    ts = starttime;
#   if defined(UNIX)
//    sprintf(dirname,"%s/%04d/%02d",ArchiveDatapath,yy,mm);
    sprintf(dirname,"%s",ArchiveDatapath);
    if ((dp=opendir(dirname)) == NULL) continue;
    while ((dirp=readdir(dp)) != NULL)
    {
      // if (dirp->d_type != DT_DIR) continue; // this unfortunately doesn't work !
      // if ((c=strstr(dirp->d_name,"annotate.")) == NULL) continue;
      c = dirp->d_name;
      if (strncmp(c,"annotate.",9)) continue;
      strncpy(fileext,&c[9],16);
      rec = strtol(fileext,NULL,16);
//      sprintf(dirname,"%s/%04d/%02d/annotate.%x",ArchiveDatapath,yy,mm,rec);
      sprintf(dirname,"%s/%.16s",ArchiveDatapath,c);
      sprintf(fileext,".%x",rec);
      isannotated = FALSE;
      if ((dpa=opendir(dirname)) == NULL) continue;
      while ((dirpa=readdir(dpa)) != NULL)
      {
        if (strstr(dirpa->d_name,fileext) == NULL) continue;
        strncpy(fname,dirpa->d_name,8); fname[8] = 0;
        ftim = strtol(fname,0,16);
        if (ftim < starttime || ftim > stoptime) continue;
        isannotated = TRUE;
        break;
      }
      closedir(dpa);
#   elif defined(WIN32)
    sprintf(dirname,"%s\\%d\\%02d",ArchiveDatapath,yy,mm);
    sprintf(fn,"%s\\*.*",dirname);
    if ((hfp=_findfirst(fn, &fp)) == -1) continue;
    do
    {
      if (fp.attrib != _A_SUBDIR) continue;
      if ((c=strstr(fp.name,"annotate.")) == NULL) continue;
      strncpy(fileext,&c[9],16);
      rec = strtol(fileext,NULL,16);
//      sprintf(dirname,"%s\\%d\\%02d\\annotate.%x",ArchiveDatapath,yy,mm,rec);
      sprintf(dirname,"%s\\%.16s",ArchiveDatapath,c);
      sprintf(fn,"%s\\*.*",dirname);
      sprintf(fileext,".%x",rec);
      isannotated = FALSE;
      if ((hfpa=_findfirst(fn, &fpa)) == -1) continue;
      do
      {
        if (strstr(fpa.name,fileext) == NULL) continue;
        strncpy(fname,fpa.name,8); fname[8] = 0;
        ftim = strtol(fname,0,16);
        if (ftim < starttime || ftim > stoptime) continue;
        isannotated = TRUE;
        break;
      } while (_findnext(hfpa,&fpa) != -1);
      _findclose(hfpa);
#   endif
      if (isannotated) 
      {
        recbuf[fctr] = rec;
        if (++fctr >= 256) break; 
      }
#   if defined(UNIX)
    }
    closedir(dp);
#   elif defined(WIN32)
    } while (_findnext(hfp,&fp) != -1);
    _findclose(hfp); 
#   endif
  }
  n = fctr;
  if (n < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)n;
  if (n == 0) return 0;
  if (isStringFormat(dout->dFormat))
  {
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        for (i=0; i<(int)dout->dArrayLength; i++)
        {
          if ((rec=getTableEntryFromRecordNumber(recbuf[i])) < 0 || rec >= totalNrRecords) continue;
          strncpy(dout->data.cptr + i*nlen,daqTbl[rec].daqParam[0].param,nlen);
        }
        return 0;
      default:
        return illegal_format;
    }
  }
  return PutValuesFromLong(dout,recbuf,n);
}
int makeDirPath(char *fname)
{
# if defined(UNIX)
  DIR *dp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
  long hfp;
# endif
  char *c,tmpname[128],dirname[128];

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,delStr); c != NULL; c=strtok(NULL,delStr))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return 0;
    }
    strcat(dirname,delStr);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if ((hfp=_findfirst(fn, &fp)) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
    else
    {
      _findclose(hfp);
    }
#   endif
  }
  return -1;
}
int WriteRecord(DAQRecord *daqr)  /* write record to file */
{
  int hFile, ispoi=FALSE;
  int recsize, hdrsiz, reclen, bwr = 0;
  long fsiz, rsiz;
  char fileName[128],s[256];
  BYTE hdr[HISTORY_HDR_SIZE], *srcptr;
  BYTE cmpDatRec[16];
  double ts;
  SINT32 t,sstmp=0,ustmp=0;
  DAQRecord daqpoi;

  if (g_isTerminating) return 0;
  if (daqr == NULL) return 0; /* should not be possible */
  if (daqr->archPending > 1) ispoi = TRUE;
  daqr->archPending = 0;
  daqr->setArchPending = 0;
  if (archPending > 0) archPending--;
  if (daqr->archtime <= daqr->lastarchtime) return 0; /* no reason */
# ifdef WIN32
  sprintf(fileName,"%s\\%04d\\%02d\\%.16s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,daqr->fname);
# else
  sprintf(fileName,"%s/%04d/%02d/%.16s",
    ArchiveDatapath,daqr->archYear,daqr->archMonth,daqr->fname);
# endif

  srcptr = daqr->archdata;
  reclen = daqr->reclen;
  hdrsiz = getRecordPrefixLength(daqr->useMinimalStorage);
  recsize = reclen + hdrsiz;
# if defined(UNIX)
  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND,S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)) == -1 )
# elif defined(WIN32)
  if ((hFile=open(fileName,O_RDWR|O_APPEND|O_BINARY)) == -1)
  {
    if (archdebug) dbglog("open for writing: %d",errno);
    hFile = open(fileName,O_RDWR|O_CREAT|O_APPEND|O_BINARY,_S_IREAD|_S_IWRITE);
  }
  if (hFile == -1 )
# endif
  {
    if (daqEqmName != NULL) SetAlarm(daqEqmName,0,ALM_APPLYNAME|ALMCODE_FIO,fileName);
    if (archdebug) dbglog("error opening file %s : %d",fileName,errno);
    return -1;
  } 
  fsiz = lseek(hFile,0L,SEEK_END);
  if ((rsiz=(fsiz%recsize)) != 0)
  {
    if (daqEqmName != NULL) SetAlarm(daqEqmName,0,ALM_APPLYNAME|ALMCODE_CORRUPT,fileName);
    sprintf(s,"re-align archive file %.32s %.64s (corrupt? file size %ld, rec size %d -> rem %ld) <recno %d>",
            daqr->srv,daqr->prp,fsiz,recsize,rsiz,daqr->rec);
    AppendProblemsLog(s);
    chsize(hFile,fsiz-rsiz);
  }
  ts = daqr->archtime;
  sstmp = daqr->archsys;
  ustmp = daqr->archusr;
  if (daqr->hasSts) 
  {
    ustmp = daqr->dataStatus;
  }
  if (!daqr->useMinimalStorage)
  {
    memcpy(hdr,&ts,sizeof(double));
    memcpy(&hdr[8],&sstmp,sizeof(SINT32));
    memcpy(&hdr[12],&ustmp,sizeof(SINT32));
  }
  else
  {
    t = (SINT32)ts;
    memcpy(hdr,&t,sizeof(SINT32));
  }
  if (daqr->hasDataPool)
  {
    int hPoolFile=-1,earlyExit=FALSE;
    size_t dp, dlen;
    DataPoolEntry *dpe = (DataPoolEntry *)daqr->data;
#   if defined(UNIX)
    sprintf(fileName,"%s/%04d/%02d/td%.14s",
      ArchiveDatapath,daqr->archYear,daqr->archMonth,&daqr->fname[2]);
    if ((hPoolFile=open(fileName,O_RDWR|O_CREAT|O_APPEND,S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)) == -1 )
#   elif defined(WIN32)
    sprintf(fileName,"%s\\%04d\\%02d\\td%.14s",
      ArchiveDatapath,daqr->archYear,daqr->archMonth,&daqr->fname[2]);
    if ((hPoolFile=open(fileName,O_RDWR|O_CREAT|O_APPEND)) == -1 )
#   endif
    if (hPoolFile == -1) 
    {
      perror(fileName);
      close(hFile);
      feclogEx(TINE_LOGLEVEL_ERROR,"cannot write companion file %.132s",fileName);
      return file_error;
    }
    dp = lseek(hPoolFile,0L,SEEK_END);
    memcpy(cmpDatRec,&dp,sizeof(size_t));
    dlen = dpe->length;
    memcpy(&cmpDatRec[sizeof(size_t)],&dlen,sizeof(size_t));
    srcptr = cmpDatRec;
    if (daqr->fmt == CF_IMAGE)
    {
      if (dlen > (UINT32)dpe->capacity)
      {
        feclogEx(TINE_LOGLEVEL_ERROR,"%.64s/%.64s[%.64s] : incoming image data length %d exceeds capacity %d",
          daqr->srv,daqr->dev,daqr->prp,dlen,dpe->capacity);
        dlen = dpe->capacity;
      }
      if (dlen < IMAGE_HDR_SIZE)
      { /* a weirdness from doocs servers ? */
        feclogEx(TINE_LOGLEVEL_ERROR,"%.64s/%.64s[%.64s] : incoming image data length %d is less than an image header!",
          daqr->srv,daqr->dev,daqr->prp,dlen);
        earlyExit = TRUE;
      }
      else if (write(hPoolFile,(char *)dpe->data,IMAGE_HDR_SIZE) < 0 || 
               (bwr=write(hPoolFile,(char *)&dpe->data[sizeof(DIMAGE)],dlen-IMAGE_HDR_SIZE)) < 0) 
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "%.64s/%.64s[%.64s] : could not write data pool record <%d>",
          daqr->srv, daqr->dev, daqr->prp, bwr, errno);
        earlyExit = TRUE;
      }
      bwr += IMAGE_HDR_SIZE;
    }
    else 
    { /* not an image */
      if (write(hPoolFile, (char *)dpe->data, dlen) < 0)
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "%.64s/%.64s[%.64s] : could not write data pool record <%d>",
          daqr->srv, daqr->dev, daqr->prp, bwr, errno);
        earlyExit = TRUE;
      }
    }
    if (!earlyExit && bwr != dlen)
      feclogEx(TINE_LOGLEVEL_ERROR, "%.64s/%.64s[%.64s] : wrote only %d or %d bytes",
        daqr->srv, daqr->dev, daqr->prp, bwr, dlen);
    close(hPoolFile);
    if (earlyExit)
    {
      close(hFile);
      return no_file_space;
    }
  }
  bwr = write(hFile,hdr,hdrsiz);
  if (bwr != hdrsiz)
    feclogEx(TINE_LOGLEVEL_ERROR, "%.64s/%.64s[%.64s] timestamp header: wrote only %d or %d bytes",
      daqr->srv, daqr->dev, daqr->prp, bwr, hdrsiz);
  bwr = write(hFile,srcptr,reclen);
  if (bwr != reclen)
    feclogEx(TINE_LOGLEVEL_ERROR, "%.64s/%.64s[%.64s] record payload: wrote only %d or %d bytes",
      daqr->srv, daqr->dev, daqr->prp, bwr, reclen);
  close(hFile);
  daqr->lastarchtime = daqr->archtime;
  if (archdebug) dbglog("archiving %s (%g)",daqr->srv,daqr->archtime);
  if (ispoi)
  {
    memcpy(&daqpoi,daqr,sizeof(DAQRecord));
    daqpoi.fname[0] = 'p'; daqpoi.fname[1] = 'i';
    daqpoi.useMinimalStorage = FALSE;
    daqpoi.lastarchtime = 0;
    WriteRecord(&daqpoi);
  }
  return 0;
}
#ifdef ANNOTATION_RECORD_JUNK
int WriteAnnotationRecord(DAQRecord *daqr,double dts)
{
  DAQRecord daqaa;
  int la;

  if (!daqr->active) return not_running;
  if (dts < daqr->lastannote + daqr->hbt) return not_applicable;
  daqr->lastannote = (time_t)dts;
  archPending++;
  la = daqr->annotation;
  memcpy(&daqaa,daqr,sizeof(DAQRecord));
  daqaa.fname[0] = 'a'; daqaa.fname[1] = 'a';
  daqaa.useMinimalStorage = TRUE;
  daqaa.lastarchtime = 0;
  daqaa.archdata = (char *)&la;
  daqaa.reclen = fmtsizeof(CF_INT32);
  return WriteRecord(&daqaa);
}
void setAnnotationRecordFileName(DAQRecord *daqaa,int n,time_t atime)
{
  struct tm *lt=localtime(&atime);
  if (lt == NULL || daqaa == NULL || n < 0) return;
  getDataFilename(n,lt->tm_year,lt->tm_mon+1); /* make sure the proper file name is set */
  memcpy(daqaa->fname,&daqTbl[n].fname,16);
  daqaa->fname[0] = 'a'; daqaa->fname[1] = 'a';
}
// this currently just 'stops' an aa in the past (i.e. removes superfluous records)
// need:
int insertAnnotationRecords(DAQRecord *daqr,time_t starttime, time_t stoptime)
{
  int cc = 0;

  return cc;
}
int removeAnnotationRecords(DAQRecord *daqr,time_t starttime, time_t stoptime)
{ /* remove all archived data from given start to stop */
  DAQRecord daqaa;
  int n, la, cc = 0;
  int yy, mm, yye = 0, mme = -1;
  long pstart, pstop=-1, pend;
  char startstr[32], filenm[256];
  time_t tnow = time(NULL);
  struct tm *lt=localtime(&starttime), *ltnow = localtime(&tnow);
  int hFile = -1, recsiz, hdrsiz = getRecordPrefixLength(TRUE);
  BYTE buffer[64];

  if (starttime >= tnow-5) return not_applicable;
  if (!daqr->active) return not_running;
  n = getTableEntryFromRecordNumber(daqr->rec);
  if (daqr->hbt <= 0) daqr->hbt = 900;
  memcpy(&daqaa,daqr,sizeof(DAQRecord));
  daqaa.fname[0] = 'a'; daqaa.fname[1] = 'a';
  daqaa.useMinimalStorage = TRUE;
  daqaa.lastarchtime = 0;
  daqaa.archdata = (char *)&la;
  daqaa.reclen = fmtsizeof(CF_INT32);
  
  recsiz = hdrsiz + daqaa.reclen;

  strncpy(startstr,ctime(&starttime),24); startstr[24] = 0; /* ctime() uses and returns the same static buffer */
  srvlog(ALF,"removeAnnotationRecords","remove annotation record %.64s after time: %.24s (%d)",
    daqr->daqParam[0].param,startstr,starttime);

  mm = lt->tm_mon; yy = lt->tm_year;
  if (stoptime > starttime)
  { /* indicates an 'insert/replace' action */
    lt = localtime(&stoptime);
    mme = lt->tm_mon; yye = lt->tm_year;
    if (mme == mm && yye == yy) 
    { // all within this one file 
      pstop = locateRecord(&daqaa,(double)stoptime,NULL,FALSE,NULL);
    }
  }
  setAnnotationRecordFileName(&daqaa,n,starttime);
  if ((pstart=locateRecord(&daqaa,(double)starttime,NULL,FALSE,NULL)) >= 0)
  {
    if ((hFile=OpenDataFile(&daqaa,NULL)) == -1) ccerr(no_such_file);
    if (pstop > 0)
    {
      pend = lseek(hFile,0L,SEEK_END);
      while(pstop < pend) 
      {
        lseek(hFile,pstop,SEEK_SET);  // go to the place to read from
        read(hFile,buffer,recsiz);
        lseek(hFile,pstart,SEEK_SET); // go to the place to write to
        write(hFile,buffer,recsiz);
        pstart += recsiz; pstop += recsiz; // advance
      }
    }
    chsize(hFile,pstart);
    pstart = pstop = 0;
    close(hFile); hFile = -1;
  }

  while (yy >= ltnow->tm_year)
  { /* advance 1 month at a time an look for files to remove */
    if (++mm > 11) { mm = 0; yy++; }
    getDataFilename(n,yy,mm+1);
    memcpy(daqaa.fname,&daqTbl[n].fname,16);
    daqaa.fname[0] = 'a'; daqaa.fname[1] = 'a';
    if (mme >= 0 && mm == mme && yy == yye) 
    { /* need to r/w edit this one */
      pstart = 0;
      pstop = locateRecord(&daqaa,(double)stoptime,NULL,FALSE,NULL);
      break; 
    }
    sprintf(filenm,"%s/%c%04d%c%02d%c%s",
      ArchiveDatapath,FS_DELIMITER,1900+yy,FS_DELIMITER,mm+1,FS_DELIMITER,daqaa.fname);
    if (unlink(filenm) != 0) break;
  }

  if (pstop > 0)
  { /* still something left to do: shift records */
    if ((hFile=OpenDataFile(&daqaa,NULL)) == -1) ccerr(no_such_file);
    pend = lseek(hFile,0L,SEEK_END);
    while(pstop < pend) 
    {
      lseek(hFile,pstop,SEEK_SET);  // go to the place to read from
      read(hFile,buffer,recsiz);
      lseek(hFile,pstart,SEEK_SET); // go to the place to write to
      write(hFile,buffer,recsiz);
      pstart += recsiz; pstop += recsiz; // advance
    }
    chsize(hFile,pstart);
  }
err:
  if (hFile != -1) close(hFile);
  srvlog(ALF,"removeAnnotationRecords","remove annotation record %.32s",cc2str(cc));
  return cc;
}
int appendAnnotationRecords(DAQRecord *daqr,time_t starttime,time_t stoptime)
{
  DAQRecord daqaa;
  DTYPE daanxt;
  int n, la = (int)stoptime, atimes[4], cc;
  char startstr[32], stopstr[32];
  time_t t, tnow = time(NULL);

  if (stoptime < starttime) return not_applicable;
  // are there annotation records more recent than the stop time?
  // if yes then they must be for a different annotation than 'starttime'
  // as 'stoptime' gives the boundary in this case
  DTYPEZERO(daanxt);
  daanxt.dArrayLength = 4;
  daanxt.dFormat = CF_INT32;
  daanxt.data.lptr = atimes;
  if ((cc=getRecordAnnotations(daqr->rec,stoptime+1,tnow,&daanxt)) == 0 && daanxt.dArrayLength > 0)
  { /* everything happens in the past ! */
    return insertAnnotationRecords(daqr,starttime,stoptime);
  }
  else
  {
    removeAnnotationRecords(daqr,starttime,0);
  }
  if (!daqr->active) return not_running;
  n = getTableEntryFromRecordNumber(daqr->rec);
  memcpy(&daqaa,daqr,sizeof(DAQRecord));
  daqaa.fname[0] = 'a'; daqaa.fname[1] = 'a';
  daqaa.useMinimalStorage = TRUE;
  daqaa.lastarchtime = 0;
  daqaa.archdata = (char *)&la;
  daqaa.reclen = fmtsizeof(CF_INT32);
  if (daqaa.hbt <= 0) daqaa.hbt = 1800;
  
  strncpy(startstr,ctime(&starttime),24); startstr[24] = 0; /* ctime() uses and returns the same static buffer */
  strncpy(stopstr,ctime(&stoptime),24); stopstr[24] = 0;
  srvlog(ALF,"appendAnnotationRecords","fill in annotation record %.64s from time: %.24s (%d) to %.24s (%d);  heartbeat : %d secs",
    daqr->daqParam[0].param,startstr,starttime,stopstr,stoptime,daqr->hbt);
  for (t=starttime; t <= stoptime; t += daqaa.hbt)
  {
    daqr->lastannote = t;
    archPending++;
    setAnnotationRecordFileName(&daqaa,n,t);
    if (WriteRecord(&daqaa) != 0) return file_error;
  }
  return 0;
}
int ReadLastAnnotationRecord(DAQRecord *daqr,int *data)
{
  DAQRecord daqaa;
  int la;

  if (!daqr->active) return not_running;
  memcpy(&daqaa,daqr,sizeof(DAQRecord));
  daqaa.fname[0] = 'a'; daqaa.fname[1] = 'a';
  daqaa.useMinimalStorage = TRUE;
  daqaa.archdata = (char *)&la;
  daqaa.data = (char *)data; 
  daqaa.reclen = fmtsizeof(CF_INT32);
  ReadRecord(&daqaa);
  return 0; /* a no_such_file return is also a success here */
}
int ReadAnnotationRecord(DAQRecord *daqr,time_t tgttime,int *data,int *ts)
{
  DAQRecord daqaa;
  int hFile = -1;
  int cc, recsize, hdrsiz;
  int la = 0;
  long pstart;
  BYTE hdr[HISTORY_HDR_SIZE];

  if (!daqr->active) return not_running;
  memcpy(&daqaa,daqr,sizeof(DAQRecord));
  daqaa.fname[0] = 'a'; daqaa.fname[1] = 'a';
  daqaa.useMinimalStorage = TRUE;
  daqaa.archdata = (char *)&la;
  daqaa.data = (char *)data; 
  daqaa.reclen = fmtsizeof(CF_INT32);

  if ((pstart=locateRecord(&daqaa,(double)tgttime,NULL,FALSE,NULL)) < 0) ccerr(no_data_in_interval);
  if ((hFile=OpenDataFile(&daqaa,NULL)) == -1) ccerr(no_such_file);
  lseek(hFile,pstart,SEEK_SET);
  hdrsiz = getRecordPrefixLength(TRUE);
  recsize = daqaa.reclen + hdrsiz;

  if (read(hFile,hdr,hdrsiz) != hdrsiz) ccerr(file_error);
  if (read(hFile,&la,sizeof(int)) != (int)daqaa.reclen) ccerr(file_error);
  if (ts != NULL) *ts = *(SINT32 *)hdr;
  if (data != NULL) *data = la;
err:
  if (hFile != -1) close(hFile);
  return 0; /* a no_such_file return is also a success here */
}
#endif // ANNOTATION_RECORD_JUNK
int WriteStatusRecord(DAQRecord *daqr,int id,double dts)
{
  DAQRecord daqsts;
  NAME32I sts;
  char *ccsts = GetCompletionStatusFromCallbackId(id);

  if (ccsts == NULL) return code_failure;
  daqr->archtime = daqr->tstamp = dts;
  daqr->laststs = dts;
  archPending++;
  sts.ival = daqr->dataStatus;
  strncpy(sts.name,ccsts,32);
  memcpy(&daqsts,daqr,sizeof(DAQRecord));
  daqsts.fname[0] = 's'; daqsts.fname[1] = 't';
  daqsts.useMinimalStorage = TRUE;
  daqsts.lastarchtime = 0;
  daqsts.archdata = (char *)&sts;
  daqsts.reclen = fmtsizeof(CF_NAME32I);
  return WriteRecord(&daqsts);
}
int ReadRecord(DAQRecord *daqr)  /* read most recent record #i from file */
{
  int hFile;
  int recsize, hdrsiz, useMinimalStorage;
  long nrecs, p2;
  double ts;
  SINT32 t, sstmp=0, ustmp=0;
  BYTE hdr[HISTORY_HDR_SIZE];

  if (daqr == NULL) return argument_list_error; /* should not be possible */
  daqr->hasLegacyStorage = FALSE;
  if ((hFile=OpenDataFile(daqr,NULL)) == -1) return no_such_file;

  useMinimalStorage = daqr->useMinimalStorage|daqr->hasLegacyStorage;
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqr->reclen + hdrsiz;

  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p2 = (nrecs-1)*recsize;
  lseek(hFile,p2,SEEK_SET);
  read(hFile,hdr,hdrsiz);
  if (!useMinimalStorage) 
  {
    ts = *(double *)hdr;
    sstmp = *(SINT32 *)&hdr[8];
    ustmp = *(SINT32 *)&hdr[12];
  } 
  else 
  { 
    t = *(SINT32 *)hdr;
    ts = (double)t; 
  }
  daqr->tstamp = ts;
  daqr->sysstamp = sstmp;
  daqr->usrstamp = ustmp;
  read(hFile,daqr->archdata,daqr->reclen);
  if (daqr->data != NULL) memcpy(daqr->data,daqr->archdata,daqr->reclen);
  if (daqr->hasDataPool)
  { /* complex data type (must use CF_HISTORY, bufsiz refers to bytes and not elements) */
    DataPoolEntry *ref = (DataPoolEntry *)daqr->archdata;
    DataPoolEntry *dpe = (DataPoolEntry *)daqr->data;
    size_t poolPos, poolLen;
    int hPoolFile = -1, plen = sizeof(size_t);
    char fn[128];
#   ifdef WIN32
    sprintf(fn,"%s\\%04d\\%02d\\td%.14s",ArchiveDatapath,daqr->archYear,daqr->archMonth,&daqr->fname[2]);
#   else
    sprintf(fn,"%s/%04d/%02d/td%.14s",ArchiveDatapath,daqr->archYear,daqr->archMonth,&daqr->fname[2]);
#   endif
    if ((hPoolFile=open(fn,O_RDONLY)) != -1)
    {
      memcpy(&poolPos,daqr->archdata,plen);
      memcpy(&poolLen,&daqr->archdata[plen],plen);
      lseek(hPoolFile,(long)poolPos,SEEK_SET);
      if (read(hPoolFile,(char *)ref->data,(unsigned int)poolLen) == (int)poolLen)
      {
        if (dpe != NULL) memcpy(dpe->data,ref->data,poolLen);
      }
      close(hPoolFile);
    }
  }

  daqr->archtime = daqr->tstamp;
  daqr->archsys = sstmp;
  daqr->archusr = ustmp;

  close(hFile);
  return 0;
}
double getLastArchivedTimeStamp(DAQRecord *daqr)
{
  long nrecs,p;
  SINT32 t;
  double dt = 0;
  int cc, recsize, hdrsiz;
  int hFile, useMinimalStorage;
  BYTE hdr[HISTORY_HDR_SIZE];

  if ( (hFile=OpenDataFile(daqr,NULL)) == -1 ) return (time_t)0;

  useMinimalStorage = daqr->useMinimalStorage|daqr->hasLegacyStorage;
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqr->reclen + hdrsiz;

  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p = (nrecs-1)*recsize;
  
  if (lseek(hFile,p,SEEK_SET) < 0) ccerr(file_error);
  if (read(hFile,hdr,hdrsiz) < 0) ccerr(file_error);
  if (!useMinimalStorage) 
  {
    dt = *(double *)hdr;
  }
  else 
  { 
    t = *(SINT32 *)hdr;
    dt = (double)t;
  }
err:
  close(hFile);
  return dt;
}
long locateRecord(DAQRecord *daqr,double ts, double *rts, int seekFlags, int *hasAlternate)
{
  long nrecs=0,p0=0,p1=0,p2=0;
  SINT32 t0,t1,t2;
  double dt0=0, dt1=0, dt2=0, f;
  int recsize, hdrsiz;
  int hFile, useMinimalStorage, it = 0, useAltFile = FALSE, cc = 0;
  int seekClosestOldRecord = seekFlags == SEEK_CLOSEST_OLD ? TRUE : FALSE;
  BYTE hdr[HISTORY_HDR_SIZE];

  if (hasAlternate != NULL) useAltFile = *hasAlternate;
  if ((hFile=OpenDataFile(daqr,hasAlternate)) == -1) return -no_such_file;

  useMinimalStorage = daqr->useMinimalStorage|daqr->hasLegacyStorage;
  if (useAltFile) useMinimalStorage = !daqr->useMinimalStorage;
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqr->reclen + hdrsiz;

  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p0 = p1 = 0; p2 = (nrecs-1)*recsize;
  
  lseek(hFile,p0,SEEK_SET);
  read(hFile,hdr,hdrsiz); 
  if (!useMinimalStorage) 
  {
    dt0 = *(double *)hdr;
  }
  else 
  { 
    t0 = *(SINT32 *)hdr; dt0 = (double)t0;
  }
  if (ts < dt0) 
  { /* beginning of record file has a more recent timestamp */
    if (seekClosestOldRecord && (daqr->option & MCAVOLATILE)) ccerr(end_of_file);
    dt1 = dt0;
    ccerr(0);
  }
  lseek(hFile,p2,SEEK_SET); 
  read(hFile,hdr,hdrsiz);
  if (!useMinimalStorage) 
  {
    dt2 = *(double *)hdr;
  }
  else 
  { 
    t2 = *(SINT32 *)hdr; dt2 = t2; 
  }

  if (isinf(dt0) || isnan(dt0)) ccerr(file_error);
  if (isinf(dt2) || isnan(dt2)) ccerr(file_error);
  while (p2 > p0 + recsize)
  {
    if (ts > (dt1=dt2)) { p1 = p2; break; }
    while (dt1 != dt0)   
    { /* seek <-- */
      f = ((ts-dt0)/(dt1-dt0))*((double)(p2-p0)/recsize);
      if ((p1=p0+recsize*(long)f) >= p2) break;
      lseek(hFile,p1,SEEK_SET); 
      read(hFile,hdr,hdrsiz); 
      if (!useMinimalStorage) 
      {
        dt1 = *(double *)hdr;
      }
      else 
      { 
        t1 = *(SINT32 *)hdr; dt1 = t1; 
      }
      if (isinf(dt1) || isnan(dt1)) ccerr(file_error);
      if (dt1 < ts) break;
      p2 = p1;
      dt2 = dt1;
      it++;
    }
    dt0 = dt1; p0 = p1;
    while (dt2 != dt1)  
    { /* seek --> */
      f = ((dt2-ts)/(dt2-dt1))*((double)(p2-p0)/recsize);
      if ((p1=p2-recsize*(long)f) <= p0) break;
      lseek(hFile,p1,SEEK_SET); 
      read(hFile,hdr,hdrsiz); 
      if (!useMinimalStorage) 
      {
        dt1 = *(double *)hdr;
      }
      else 
      { 
        t1 = *(SINT32 *)hdr; dt1 = t1; 
      }
      if (isinf(dt1) || isnan(dt1)) ccerr(file_error);
      if (dt1 > ts) break;
      p0 = p1; dt0 = dt1;
      it++;
    }
    p2 = p1; dt2 = dt1;
  }
  if (archdebug > 1) dbglog("record located in %d iterations",it);
err: 
  close(hFile);
  if (dt1 > ts && p1 >= recsize && seekClosestOldRecord) p1 -= recsize;
  if (rts != NULL) *rts = dt1;
  if (cc == file_error) feclogEx(TINE_LOGLEVEL_ERROR, "locateRecord: error reading file %.16s",daqr->fname);
  return cc == 0 ? p1 : -cc;
}
int loadTraceData(BYTE *buffer,int n, int p, time_t tgttime, DTYPE *dout,int index,int fieldindex,int sample)
{
  int cc = 0,useMinimalStorage,seekClosestOldRecord=FALSE,isStrFmt=FALSE,stringsMap=FALSE;
  int hFile = -1, hPoolFile = -1;
  int bufsize, bufbytes, recsize, hdrsiz, rlen;
  long pstart;
  int i, pos, datfmt, datfmtsize, fmtsize, tsize, datfhdrsize, fhdrsize;
  int foffset=0, ffmt=CF_NULL, dfmt, doff;
  SINT32 timestamp, sysstamp = 0, usrstamp = 0;
  double dtimestamp;
  struct tm lt = *localtime((time_t *)&tgttime);
  DTYPE drb;
  BYTE *datbuf, *bptr, hdr[HISTORY_HDR_SIZE];

  if (index > daqTbl[n].daqParam[p].size) return invalid_index;
  if (sample > daqTbl[n].daqParam[p].size) return illegal_delimiter;
  if (index < 0) index = 0;
  fhdrsize = GetFormatHeaderSize(daqTbl[n].daqParam[p].fmt);
  if (daqTbl[n].daqParam[p].fmt == CF_USTRING && isStringFormat(dout->dFormat)) stringsMap = TRUE;
  if (!stringsMap && 
      !isSimpleFormat(daqTbl[n].daqParam[p].fmt) && fhdrsize == 0)
  {
    char dev[132] = "";
    sprintf(dev,"/%.32s/%.32s",daqTbl[n].ctx,daqTbl[n].srv);
    cc = getFieldInFormat(daqTbl[n].daqParam[p].fmt,dev,daqTbl[n].tag,fieldindex,&foffset,&ffmt);
    if (cc != 0) return -cc;
  }
  daqTbl[n].hasLegacyStorage = FALSE;
retryLocateRecord:
  datbuf = dout->data.bptr;  // what is wanted from above
  datfmt = dout->dFormat;
  datfmtsize = fmtsizeof(datfmt);
  datfhdrsize = GetFormatHeaderSize(datfmt);
  bufsize = (int)dout->dArrayLength;
  /* if (daqTbl[n].reclen > TRACE_BUFFER_SIZE) return resources_exhausted; */
  getDataFilename(n,lt.tm_year,lt.tm_mon+1);
  if (daqTbl[n].option & ONCE) seekClosestOldRecord = SEEK_CLOSEST_OLD;
  if ((pstart=locateRecord(&daqTbl[n],(double)tgttime,NULL,seekClosestOldRecord,NULL)) < 0) 
  { /* only returns -1 if file not found */
    if (!daqTbl[n].hasLegacyStorage && !daqTbl[n].useMinimalStorage)
    { /* maybe a transition month ? */
      daqTbl[n].hasLegacyStorage = TRUE;
      goto retryLocateRecord;
    }
    ccerr(pstart == -file_error ? file_error : no_data_in_interval) 
  }
  if ((hFile=OpenDataFile(&daqTbl[n],NULL)) == -1) ccerr(no_such_file);
  lseek(hFile,pstart,SEEK_SET);
  fmtsize = fmtsizeof(daqTbl[n].daqParam[p].fmt);
  if (daqTbl[n].daqParam[p].fmt == CF_STRUCT)
  {
    fmtsize *= GetStructSize(daqTbl[n].tag);
    if (fmtsize <= 0) ccerr(invalid_structure_tag);
  }
  useMinimalStorage = daqTbl[n].useMinimalStorage|daqTbl[n].hasLegacyStorage;
  hdrsiz = getRecordPrefixLength(useMinimalStorage);
  recsize = daqTbl[n].reclen + hdrsiz;

  if (read(hFile,hdr,hdrsiz) != hdrsiz) ccerr(file_error);
  if (!useMinimalStorage) 
  { 
    dtimestamp = *(double *)hdr;
    timestamp = (SINT32)dtimestamp; 
    sysstamp = *(SINT32 *)&hdr[8];
    usrstamp = *(SINT32 *)&hdr[12];
  }
  else 
  { 
    timestamp = *(SINT32 *)hdr;
    dtimestamp = (double)timestamp; 
  }

  if ((rlen=read(hFile,buffer,daqTbl[n].reclen)) != (int)daqTbl[n].reclen) 
  {
    if (archdebug) dbglog("read %d bytes vs. %d bytes attempted",rlen,daqTbl[n].reclen);
    ccerr(file_error);
  }

  if (daqTbl[n].hasDataPool)
  {
    char fn[132];
    size_t poolPos, poolLen;
    int nread, nleft, plen = sizeof(size_t);
    tsize = MIN(daqTbl[n].daqParam[p].size,bufsize);
    rlen = tsize*datfmtsize + datfhdrsize;
#   ifdef WIN32
    sprintf(fn,"%s\\%04d\\%02d\\td%.14s",
      ArchiveDatapath,daqTbl[n].archYear,daqTbl[n].archMonth,&daqTbl[n].fname[2]);
#   else
    sprintf(fn,"%s/%04d/%02d/td%.14s",
      ArchiveDatapath,daqTbl[n].archYear,daqTbl[n].archMonth,&daqTbl[n].fname[2]);
#   endif
    if ((hPoolFile=open(fn,O_RDONLY)) == -1) ccerr(file_error);    
    memcpy(&poolPos,buffer,plen);
    memcpy(&poolLen,&buffer[plen],plen);
    lseek(hPoolFile,(long)poolPos,SEEK_SET);
    if (poolLen < (size_t)rlen) rlen = (int)poolLen;
    nleft = rlen; /* bytes to read */
    plen = 0; /* reset position in buffer */
    while ((nread=read(hPoolFile, &buffer[plen], nleft)) != nleft)
    {
      if (archdebug) dbglog("(data pool file) read %d bytes vs. %d bytes attempted", nread, nleft);
      if (nread < 0)
      {
        perror("read");
        ccerr(file_error);
      }
      nleft -= nread; /* didn't get them all! Now there are this many left */
      plen += nread; /* advance the position in the read buffer */
    }
    bufsize = (rlen - fhdrsize) / fmtsize; /* num elements read : reset this in the case of a data pool */
  }
  
  dout->dTimeStamp = dtimestamp;
  dout->sysStamp = sysstamp;
  dout->dStamp = usrstamp;
  isStrFmt = (datfmt == CF_STRING || daqTbl[n].daqParam[p].fmt == CF_STRING);
  DTYPEZERO(drb);
  if (index == 0 && sample == 0) 
  { /* the whole trace and nothing but the trace ! */
    tsize = MIN(daqTbl[n].daqParam[p].size,bufsize);
    if (dout->dArrayLength > (UINT32)tsize) dout->dArrayLength = (UINT32)tsize;
    bufbytes = bufsize * datfmtsize + datfhdrsize;
    /* but zero out the ends if there are any ... */
    if (bufbytes > rlen && !isStrFmt) memset(&datbuf[rlen*datfmtsize+datfhdrsize],0,bufbytes-rlen);
    if (datfmt == daqTbl[n].daqParam[p].fmt && !isStrFmt)
    { /* equal formats and src and dst are NOT CF_STRING */
      if (datfmt == CF_IMAGE)
      {
        DIMAGE *img = (DIMAGE *)datbuf;
        memcpy(datbuf,buffer,IMAGE_HDR_SIZE);
        img->frameBufferSize = rlen - IMAGE_HDR_SIZE;
        img->frameBuffer = &buffer[IMAGE_HDR_SIZE];
      }
      else if (datfmt == CF_GSPECTRUM)
      { // buffer is what we just read from disc
        DGSPECTRUM *gsp = (DGSPECTRUM *)datbuf;
        memcpy(datbuf, buffer, GSPECTRUM_HDR_SIZE);
        gsp->spectBuffer = (float *)(buffer + sizeof(DGSPECTRUM));
        //memcpy(gsp->spectBuffer, buffer + sizeof(DGSPECTRUM), tsize * fmtsize);
        gsp->spectBufferSize = tsize;
      }
      else
      {
        BYTE *srcptr = buffer+daqTbl[n].daqParam[p].offset; 
        if (datfmt == CF_STRUCT && strncmp(daqTbl[n].tag,dout->dTag,TAG_NAME_SIZE)) 
          ccerr(invalid_structure_tag);
        if (datbuf != srcptr) memcpy(datbuf,srcptr,tsize*fmtsize+fhdrsize);
      }
    }
    else
    { /* requested format is different from stored format (or CF_STRING) */
      dfmt = daqTbl[n].daqParam[p].fmt;
      doff = daqTbl[n].daqParam[p].offset;
      if (ffmt != CF_NULL)
      {
        doff += foffset;
        dfmt = ffmt;
      }
      drb.dArrayLength = bufsize;
      drb.dFormat = datfmt;
      drb.data.vptr = datbuf;
      bptr = buffer + doff;
      switch (dfmt)
      {
        case CF_INT16:
        case CF_INT32:
        case CF_INT64:
        case CF_FLOAT:
        case CF_DOUBLE:
          if ((cc=PutValuesFromAny(&drb,bptr,(short)dfmt,1,fmtsize,tsize,0)) != 0) ccerr(cc);
          break;
        case CF_NAME16I:
        case CF_NAME32I:
        case CF_NAME64I:
        case CF_DBLDBL:
        case CF_INTINT:
        case CF_FLTINT:
        { /* stored as a (data-status) doublet ? */
          short pfmt = getPrimaryFormat(daqTbl[n].daqParam[p].fmt);
          if ((cc=PutValuesFromAny(&drb,(void *)bptr,pfmt,1,fmtsize,tsize,0)) != 0) 
            ccerr(cc);
          break;
        }
        case CF_USTRING:
          /* bptr += 16; <== PutValuesFromString takes ptr to type */
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          if ((cc=PutValuesFromString(&drb,(char *)bptr,daqTbl[n].daqParam[p].fmt,tsize,0)) != 0) ccerr(cc);
          break;
        case CF_GSPECTRUM:
          if (datfmt == CF_FLOAT)
          {
            DGSPECTRUM *gsp = (DGSPECTRUM *)datbuf;
            gsp->spectBuffer = (float *)(buffer + sizeof(DGSPECTRUM));
            memcpy(datbuf, gsp->spectBuffer, tsize*fmtsize);
            break;
          }
          ccerr(illegal_format);
        case CF_SPECTRUM:
          if (datfmt == CF_FLOAT)
          {
            bptr += fhdrsize;
            memcpy(datbuf, bptr,tsize*fmtsize);
            break;
          }
        default:
          if (datfmt != daqTbl[n].daqParam[p].fmt) ccerr(illegal_format);
          memcpy(datbuf,bptr,fhdrsize+tsize*fmtsize);
          break;
      }
    }
  }
  else
  { /* index and/or sample given : jump in the middle and skip around */
    if (sample == 0) sample = 1;
    tsize = MIN(daqTbl[n].daqParam[p].size-index,bufsize);
    if (bufsize > tsize && !isStrFmt) memset(&datbuf[tsize*fmtsize+fhdrsize],0,(bufsize-tsize)*fmtsize);
    for (i=0; i<bufsize; i++)
    {
      drb.dArrayLength = 1;
      drb.dFormat = datfmt;
      drb.data.vptr = &((BYTE *)datbuf)[i*fmtsizeof(datfmt)+fhdrsize];
      if ((pos=(index+sample*i)) >= daqTbl[n].daqParam[p].size) break;
      bptr = buffer + daqTbl[n].daqParam[p].offset + pos*fmtsize;
      switch (daqTbl[n].daqParam[p].fmt)
      {
        case CF_INT16:
          if ((cc=PutValuesFromShort(&drb,(short *)bptr,1)) != 0) ccerr(cc);
          break;
        case CF_INT32:
          if ((cc=PutValuesFromLong(&drb,(SINT32 *)bptr,1)) != 0) ccerr(cc);
          break;
        case CF_INT64:
          if ((cc=PutValuesFromDLong(&drb,(SINT64 *)bptr,1)) != 0) ccerr(cc);
          break;
        case CF_FLOAT:
          if ((cc=PutValuesFromFloat(&drb,(float *)bptr,1)) != 0) ccerr(cc);
          break;
        case CF_DOUBLE:
          if ((cc=PutValuesFromDouble(&drb,(double *)bptr,1)) != 0) ccerr(cc);
          break;
        case CF_USTRING:
          /* bptr += 16; */
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          if ((cc=PutValuesFromString(&drb,(char *)bptr,daqTbl[n].daqParam[p].fmt,1,0)) != 0) ccerr(cc);
          break;          
        default:
          if (datfmt != daqTbl[n].daqParam[p].fmt) ccerr(illegal_format);
          memcpy(&((BYTE *)datbuf)[i*fmtsize+fhdrsize],bptr+fhdrsize,tsize*fmtsize);
          if (fhdrsize > 0) memcpy(&((BYTE *)datbuf)[0],bptr,fhdrsize);
          break;
      }
    }
    dout->dArrayLength = (UINT32)i;
  }
err: 
  if (hFile != -1) close(hFile);
  if (hPoolFile != -1) close(hPoolFile);
  return cc;
}
int getFilter(char *tag)
{
  int i;
  for (i=0; i<nDaqFilterTblEntries && stricmp(tag,daqFilterTbl[i].tag); i++);
  return i < nDaqFilterTblEntries ? i : -1;
}
int getFiltersMask(char *filterString)
{
  int i, mask = 0;
  strupr(filterString);
  for (i=0; i<nDaqFilterTblEntries; i++)
  {
    if (strstr(filterString,daqFilterTbl[i].tag) != NULL) 
      mask |= daqFilterTbl[i].mask;
  }
  return mask;
}
int isPastFilter(UINT32 option)
{
  return ((option & gValidFilterMask) == option) ? TRUE : FALSE;
}
int dumpFilterTable(int a0,int a1,int a2,int a3)
{
  int i;
  a0 = a0; a1 = a1; a2 = a2; a3 = a3;
  ttyoutput("Current Filter Table Status:");
  for (i=0; i<nDaqFilterTblEntries; i++)
  {
    ttyoutput("\t%s : %s (%s)",
      daqFilterTbl[i].tag,
      daqFilterTbl[i].valid ? "VALID" : "NOT VALID",
      daqFilterTbl[i].active ? "active" : "not active");
  }
  return 0;
}
int getFilterTable(void)
{
  int i, n = 0, len, num, fmt, ep, bp;
  char str[256], scratch[128], fn[128];
  int done = 0,cc = 0, fltrFileFound = FALSE;
  int key_col,flt_col,max_col,min_col,fmt_col,dsc_col;
  UINT32 bit;
  FILE *fp = NULL;

  n = nDaqStockFilterNames;
  if (daqFilterTbl != NULL) free(daqFilterTbl); daqFilterTbl = NULL;
  sprintf(fn,"%s%s",ArchiveDBpath,FILTERS_FILENAME);
  if ((num=GetDataBaseSize(fn)) > 0)
  {
    fltrFileFound = TRUE;
    n += num;
  }
  daqFilterTbl = (DAQFilter *)calloc(n,sizeof(DAQFilter));
  if (daqFilterTbl == NULL) ccerr(out_of_local_memory);
  for (n=0; n<nDaqStockFilterNames; n++)
  { /* these are just filter tags (no keyword associations) */
    strncpy(daqFilterTbl[n].tag,daqStockFilterNames[n],64);
    strncpy(daqFilterTbl[n].desc,daqStockFilterDescriptions[n],64);
    bit = (0x0001)<<n;
    daqFilterTbl[n].mask = bit;
    gValidFilterMask |= bit;
    daqFilterTbl[n].active = TRUE;
    daqFilterTbl[n].valid = TRUE;
  }
  if (!fltrFileFound) ccerr(-num);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(file_error);
  done = 0;
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
      if ((flt_col=findcol(str,"FILTER")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
      max_col = findcol(str,"MAX_VALID");
      min_col = findcol(str,"MIN_VALID");
      dsc_col = findcol(str,"DESCRIPTION");
      done = 1;
      continue;
    }
    strncpy(daqFilterTbl[n].tag,colptr(flt_col,str),64);
    for (i=0; i<n && stricmp(daqFilterTbl[n].tag,daqFilterTbl[i].tag); i++);
    if (i < n) continue; /* already in the filter table */
    strncpy(daqFilterTbl[n].key,colptr(key_col,str),64);
    if (dsc_col >= 0)
    {
      strncpy(daqFilterTbl[n].desc,colptr(dsc_col,str),64);
    }
    strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
    if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
#   ifdef HAVE_READ_INFO_DBASE
    /* but the info daq info database keys on the filters ! */
    ridx = getEntryFromKeyword(daqFilterTbl[n].key,NULL,&pidx);
    if (ridx < 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR,"filter key %s not accessible!",daqFilterTbl[n].key);
      continue;
    }
    fmt = daqTbl[ridx].daqParam[pidx].fmt;
#   endif
    memset(scratch,0,128);
    if (min_col >= 0)
    {
      strncpy(scratch,colptr(min_col,str),64); strupr(scratch);
    }
    if ((len=(int)strlen(scratch)) > 0)
    { /* something in the MIN_VALID column ... */
      switch (fmt)
      {
        case CF_BYTE:
        case CF_SHORT:
        case CF_LONG:
        case CF_DLONG:
        case CF_FLOAT:
        case CF_DOUBLE:
          daqFilterTbl[n].fValidMin = strtod(scratch,NULL);
          daqFilterTbl[n].fmtValidMin = CF_DOUBLE;
          break;
        case CF_TEXT:
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          bp = 0;
          if ((ep = (int)strlen(scratch) - 1) > 0)
          {
            if (scratch[0] == '*' && scratch[ep] == '*') 
            {
              bp = 1; scratch[ep] = 0;
              daqFilterTbl[n].sValidCmp = STR_COMP_CONTAINS;
            }
            else if (scratch[0] == '*') 
            {
              bp = 1; 
              daqFilterTbl[n].sValidCmp = STR_COMP_BEGINS;
            }
            else if (scratch[ep] == '*')
            {
              scratch[ep] = 0;
              if (scratch[ep] == '*') daqFilterTbl[n].sValidCmp = STR_COMP_ENDS;
            }
            else daqFilterTbl[n].sValidCmp = STR_COMP_MATCHES;
          }
          strncpy(daqFilterTbl[n].sValid,&scratch[bp],64);
          daqFilterTbl[n].fmtValidMin = CF_NAME64;
          break;
        default:
          feclogEx(TINE_LOGLEVEL_ERROR,"filter key %s does not have the proper format (%s)!",daqFilterTbl[n].key,itof(fmt));
          daqFilterTbl[n].fmtValidMin = CF_NULL;
          continue;
      }
    }
    else
    {
      daqFilterTbl[n].fmtValidMin = CF_NULL;
    }
    memset(scratch,0,128);
    if (max_col >= 0)
    {
      strncpy(scratch,colptr(max_col,str),64); strupr(scratch);
    }
    if ((len=(int)strlen(scratch)) > 0)
    { /* something in the MAX_VALID column ... */
      switch (fmt)
      {
        case CF_BYTE:
        case CF_SHORT:
        case CF_LONG:
        case CF_DLONG:
        case CF_FLOAT:
        case CF_DOUBLE:
          daqFilterTbl[n].fValidMax = strtod(scratch,NULL);
          daqFilterTbl[n].fmtValidMax = CF_DOUBLE;
          break;
        case CF_TEXT:
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          bp = 0;
          if ((ep = (int)strlen(scratch) - 1) > 0)
          {
            if (scratch[0] == '*' && scratch[ep] == '*') 
            {
              bp = 1; scratch[ep] = 0;
              daqFilterTbl[n].sValidCmp = STR_COMP_CONTAINS;
            }
            else if (scratch[0] == '*') 
            {
              bp = 1; 
              daqFilterTbl[n].sValidCmp = STR_COMP_BEGINS;
            }
            else if (scratch[ep] == '*')
            {
              scratch[ep] = 0;
              if (scratch[ep] == '*') daqFilterTbl[n].sValidCmp = STR_COMP_ENDS;
            }
            else daqFilterTbl[n].sValidCmp = STR_COMP_MATCHES;
          }
          strncpy(daqFilterTbl[n].sValid,&scratch[bp],64);
          daqFilterTbl[n].fmtValidMax = CF_NAME64;
          break;
        default:
          feclogEx(TINE_LOGLEVEL_ERROR,"filter key %s does not have the proper format (%s)!",daqFilterTbl[n].key,itof(fmt));
          daqFilterTbl[n].fmtValidMax = CF_NULL;
          continue;
      }
    }
    else
    {
      daqFilterTbl[n].fmtValidMax = CF_NULL;
    }
    daqFilterTbl[n].mask = (0x0001)<<n;
    n++;
  }
err: 
  if (n > MAX_FILTERS) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"too many filters in filter table (%d); truncating to %d",n,MAX_FILTERS);
    n = MAX_FILTERS;
  }
  nDaqFilterTblEntries = n;
  if (cc != 0 && cc != no_such_file) feclogEx(TINE_LOGLEVEL_ERROR,"getFilterTable : %s",erlst[cc]);
  if (fp != NULL) fclose(fp);
  return n;
}
int ntags = 0;
int nkeywords = 0;
NAME64 *keywordList = NULL;

void qnsort(NAME64 **plst,int left, int right)
{
  NAME64 *t;
  int l = left;
  int r = right;
  int p = (left+right)/2;
qsort:
  while (strnicmp(plst[l]->name,plst[p]->name,64) <= 0 && l < p) l++;
  while (strnicmp(plst[r]->name,plst[p]->name,64) > 0 && r > p) r--;
  if (l <= r) 
  { // can swap these
    t = plst[r];
    plst[r--] = plst[l];
    plst[l++] = t;
    if (l <= r) goto qsort;
  }
  // everybody's on the correct side of the pivot !
  if (left < r) qnsort(plst,left,r);
  if (l < right) qnsort(plst,l,right);
  return; 
}
NAME64 *namesort(NAME64 *list,int listsize)
{
  int i;
  NAME64 *newlst = (NAME64 *)calloc(listsize,sizeof(NAME64));
  NAME64 **plst = (NAME64 **)calloc(listsize,sizeof(NAME64 *));
  if (newlst == NULL || plst == NULL) return NULL;

  for (i=0; i<listsize; i++) plst[i] = &list[i];

  qnsort(plst,0,listsize-1);

  for (i=0; i<listsize; i++)
  {
    strncpy(newlst[i].name,plst[i]->name,64);
  }
  free(plst);
  return newlst;
}
/*
 * can use 
 *   int GetDeviceListFromProperty(char *eqm,char *prp,NAME64 *devlst,UINT32 *devlen)
 * to get device list for an MCA property ...
 */
int getArchivedKeywords(NAME64 *list, int *listsize)
{
  int i, j, n, len;
  NAME64 *nlst = NULL;
  if (list == NULL || listsize == NULL) return argument_list_error;
  if ((len = *listsize) < 1) return buffer_too_small;
  if (keywordList != NULL)
  {
    n = MIN(len,nkeywords);
    memcpy(list,keywordList,n*sizeof(NAME64));
    *listsize = n;
    return 0;
  }
  if (ntags < 1) return not_initialized;
  if ((nlst = (NAME64 *)calloc(ntags,sizeof(NAME64))) == NULL) return out_of_local_memory;
  for (i=0,n=0; i<totalNrRecords; i++)
  {
    if (!daqTbl[i].active) continue;
    if (daqTbl[i].option&NEVER) continue; /* it's not being archived */
    if (len < n) break;
    for (j=0; j<daqTbl[i].nparam; j++)
    {
      if (kwdEndsWith(daqTbl[i].daqParam[j].param,".NAM")) continue;
      if (kwdEndsWith(daqTbl[i].daqParam[j].param,".ENUM")) continue;
      strncpy(nlst[n].name,daqTbl[i].daqParam[j].param,64);
      n++;
    }
  }
  nkeywords = n;
  keywordList = namesort(nlst,nkeywords);
  memcpy(list,keywordList,n*sizeof(NAME64));
  if (nlst != NULL) free(nlst);
  *listsize = n;
  return 0;
}
void getSubSysDecoration(char *ssLongName,char *ssDecoration)
{
  char ss[SUBSYSTEM_NAME_SIZE];
  ssDecoration[0] = 0;
  if (strlen(ssLongName) == 0) return;
  strncpy(ss,ssLongName,SUBSYSTEM_NAME_SIZE); strupr(ss);
  if (strstr(ss,"DIAG")) { strcpy(ssDecoration,".DIAG"); return; }
  if (strstr(ss,"RF")) { strcpy(ssDecoration,".RF"); return; }
  if (strstr(ss,"MAG")) { strcpy(ssDecoration,".MAG"); return; }
  if (strstr(ss,"VAC")) { strcpy(ssDecoration,".VAC"); return; }
  if (strstr(ss,"LASER")) { strcpy(ssDecoration,".FEL"); return; }
  strcpy(ssDecoration,".UTIL"); 
  return;
}
void slave_mode_diagnostics_state_name( int state, char * message) {
  if( state == FAILOVER_MASTER) strncat( message, "master", 256-strlen(message)-1);
  else if( state == FAILOVER_SLAVE) strncat( message, "slave", 256-strlen(message)-1);
  else if( state == (FAILOVER_MASTER | FAILOVER_SLAVE)) strncat( message, "slave-as-master", 256-strlen(message)-1);
  else if( state == FAILOVER_NONE) strncat( message, "none", 256-strlen(message)-1);
  else strncat( message, "misconfiguration", 256-strlen(message)-1);
}
int slave_mode_diagnostics_info_file(char *path,char *fn, char *msg)
{
  FILE *fp=NULL;
  char fullName[256];
  int cc = 0;

  if (path == NULL) ccerr(invalid_parameter);
  if (fn == NULL || strlen(fn) == 0) ccerr(invalid_parameter);
  sprintf(fullName,"%.128s%s",path,fn);
  makeDirPath(fullName);
  if ((fp=fopen(fullName,"a")) == NULL) ccerr(file_error);
  fprintf(fp,"%s\n",msg);
err:
  if (fp != NULL) fclose(fp);
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR,"cannot write master/slave failover diagnostics to file: %s",cc2str(cc));
  return cc;
}
void slave_mode_diagnostics( void) {
  static int lastFailoverState = 0;
  struct timeval tv;
  double ts;  
  char message[256];
  int failoverState = GetCurrentFailoverState();
  if( failoverState == FAILOVER_NONE) return;
  if( failoverState != lastFailoverState && lastFailoverState != FAILOVER_NONE) {
    gettimeofday(&tv,(struct timezone *)NULL);
    ts = (double)tv.tv_sec;
    ts += (double)(tv.tv_usec)/1000000.0;
    sprintf(message, "*** %s *** failoverState changed from ", getDataTimeString(ts,FALSE));
    slave_mode_diagnostics_state_name( lastFailoverState, message);
    strncat( message, " to ", 256-strlen(message)-1);
    slave_mode_diagnostics_state_name( failoverState, message);
    feclogEx(TINE_LOGLEVEL_WARN, "%s\n",message);
    slave_mode_diagnostics_info_file(gFeclogPath, "failover.log", message);
    printf("%s\n", message);
  }
  lastFailoverState = failoverState;
}
time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c, day[16] = {0}, mon[64] = {0}, year[16] = {0};
  char hr[64] = {0}, mn[32] = {0}, sec[16] = {0};
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = (int)strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    if (tmts.tm_year < 70) tmts.tm_year += 100;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
