/*
 * File archive.h
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _ARCHIVE_H_
#define _ARCHIVE_H_

#define no_data_in_interval 512
#define invalid_range 513

#define CA_POI_RANGE_FACTOR 0.10
#define CA_POI_TOLERANCE_FACTOR 10.0
#define CA_USE_POI_RANGE 1
#define CA_USE_POI_TOLERANCE 2

#define ARCHIVEDATAPATHDEPTH 3
extern char *ArchiveDBfile;
extern char *ArchiveDBpath;
extern char *ArchiveDatapath;
#define MAX_ADBFILES 20
extern char adbfile[MAX_ADBFILES][64];
extern int nadbfiles;
extern char *delStr;

#define btoa(v) ((v) ? "TRUE" : "FALSE")
#define ALF "annotation.log"

#define ALMCODE_GLB 512
#define ALMCODE_FIO 513
#define ALMCODE_CORRUPT 514
#define ALMCODE_NONEXIST 515
#define ALMCODE_POITOOLARGE 516
extern char *daqEqmName;

#define NDRIVES 4
#define MAX_FECS  50
#define NETMEX_BUF_SIZE 8192
#define RECORD_CAPACITY_BUFFER 100

#define NEVER      0x0001
#define ONCE       0x0002
#define ALWAYS     0x0004
#define FAST       0x0008
#define SLOW       0x0010
#define FIXTIMESTAMP 0x0020
#define HSTHDR     0x0040
#define STATUS     0x0080
#define MCAVOLATILE 0x0100
#define NOPOI      0x0200

#define ARCH_STD_INTERVAL 2
#define ARCH_SLOW_INTERVAL 60

#define ERROR_REACK_MCA_COUNT 2
#define ERROR_APPLY_STATUS_COUNT 5

#define FILTERS_FILENAME "filters.csv"
#define DEFAULT_ADB_FILENAME "datam.csv"
#define MAX_FILTERS 32
#define STR_COMP_MATCHES 0
#define STR_COMP_BEGINS 1
#define STR_COMP_ENDS 2
#define STR_COMP_CONTAINS 3
typedef struct
{
  char tag[64]; /**< the filter tag (option filter database entry) */
  char key[64]; /**< the database parameter keyword providing the data source */
  char desc[64]; /**< description of the filter */
  char sValid[64]; /**< for string parameters: the comparison string */
  int mask; /**< bit position in the validity check */
  int valid; /**< boolean value */
  int active; /**< filter is being used by at least one record */
  int sValidCmp; /**< for string parameters: one of STR_COMP_CONTAINS, STR_COMP_BEGINS, STR_COMP_ENDS, STR_COMP_MATCHES */
  int fmtValidMin; /**< keyword format for min comparison */
  int fmtValidMax; /**< keyword format for max comparison */
  double fValidMin; /**< for number parameters: minimum valid parameter */
  double fValidMax; /**< for number parameters: maximum valid parameter */
} DAQFilter;
#define DAQFltrTag "DAQFltr"

extern DAQFilter *daqFilterTbl;
extern int nDaqFilterTblEntries;
extern char *daqStockFilterNames[];
extern int nDaqStockFilterNames;
extern UINT32 gValidFilterMask;
extern int gAllocDataBuffers;
extern int nkeywords, ntags;
extern NAME64 *keywordList;

int getDaqFltrTbl(DAQFilter *dst,int *len);
int addDaqFilters(DAQFilter *src,int len);
int dumpDaqFltrTbl(char *path,char *fn);
int getFilterTable(void);
int getFiltersMask(char *filterString);
int getFilter(char *tag);
int isPastFilter(UINT32 option);
extern int gIsValidInTransition;
int dumpFilterTable(int a0,int a1,int a2,int a3);
int dumpArchiveRecord(int a0,int a1,int a2,int a3);
int inputDataLoad(int ntotal, int offset);
int annotateRecord(int rec,time_t ts,char *cmt,int len);
int getRecordAnnotation(int rec,time_t ts,char *cmt,int *len);
int getRecordAnnotations(int rec,time_t starttime,time_t stoptime,DTYPE *dout);
int getRecordsWithAnnotations(time_t starttime,time_t stoptime,DTYPE *dout);
int getEntryFromKeyword(char *keyword,int *rec,int *param);
int loadDaqDb(char *fname, int offset);
int loadDaqParamDb(int ntotal, int offset);
void fillinArrayTypes(int num);
int getparam(char *s, char *param, UINT16 *fmt, int *size);
extern int nrecords, nparams, recordsCapacity;
int handler(int deverr,int doserr, int bp,int si);
int addADBtoList(char *fn);
int freeDaqDb(void);

typedef struct
{
  char name[64];
  int recNr;
  int prmNr;
} FoSrc;

typedef struct KeywordFailoverStruct
{
  char keyword[PROPERTY_NAME_SIZE];
  FoSrc foSrc[3];
  int nFailoverSource;
  BYTE *data;
  int fmt;
  int len;
  double dts;
  struct KeywordFailoverStruct *nxt;
} KeywordFailover;
extern KeywordFailover *kwrdFailover;

typedef struct
{
  char key[PROPERTY_NAME_SIZE]; /**< parameter name */
  char units[UNITS_SIZE];       /**< units */
  char desc[PROPERTY_DESC_SIZE]; /**< description */
  char plot[UNITS_SIZE];        /**< plot style (LOG, LIN, etc.) */
  char boundTo[PROPERTY_NAME_SIZE]; /**< bound to property reference */
  char sub[SUBSYSTEM_NAME_SIZE]; /**< parameter subsystem */
  char assoc[PROPERTY_NAME_SIZE]; /**< associate dev names to */
  int fmt;              /**< allowable format */
  int size;             /**< data array size assuming given format */
  float scaleFactor;
  float scaleOffset;
  float maximum;       /**< maximum value for parameter */
  float minimum;       /**< minimun value for parameter */
  float aTolerance;    /**< absolute tolerance in units given */
  float pTolerance;    /**< percent tolerance */
} DAQPrmInfo;
#define DAQPrmTag "DAQPrm"

typedef struct PriorReference
{
  char keyword[64];
  int ndevs;
  time_t begin;
  time_t end;
  NAME64 *devlst;
  NAME64 *aliaslst;
} PriorReference;
#define DAQPriorTag "DAQPrior"

typedef struct
{
  char param[PROPERTY_NAME_SIZE]; /**< parameter name */
  char boundTo[PROPERTY_NAME_SIZE]; /**< bound to parameter */
  char subsys[SUBSYSTEM_NAME_SIZE]; /**< parameter subsystem */
  char assoc[PROPERTY_NAME_SIZE]; /**< associate dev names to */
  UINT16 fmt;          /**< allowable format */
  UINT16 arrayType;    /**< array type */
  int size;            /**< data array size assuming given format */
  int offset;          /**< array starting offset in bytes */
  int rowsize;         /**< row size for double arrays */
  int nrows;           /**< number of rows for double arrays */
  float maximum;       /**< maximum value for parameter */
  float minimum;       /**< minimun value for parameter */
  float aTolerance;    /**< absolute tolerance in units given */
  float pTolerance;    /**< percent tolerance */
  float xmaximum;      /**< x-axis maximum value for spectrum parameter */
  float xminimum;      /**< x-axis minimun value for spectrum parameter */
  char units[UNITS_SIZE]; /**< units */
  char xunits[UNITS_SIZE];/**< x-axis units for spectrum parameter */
  char desc[PROPERTY_DESC_SIZE]; /**< description */
  char plot[UNITS_SIZE]; /**< plot style (LOG, LIN, etc.) */
  time_t timestamp;    /**< most recent timestamp */
  float scaleFactor;   /**< scale factor (applied to stored data) */
  float scaleOffset;   /**< scale offset (applied to stored data) */
  int keepsPointsOfInterest;
  int isglobal;        /**< is being sent as global */
  int globalId;        /**< global id if global */
  int srcIndex;        /**< index to partner recording containing source info (if available) */
  int npriors;         /**< number of prior references */
  PriorReference *pr;  /**< pointer to prior references if any */
} DAQParameter;

typedef struct
{
  char fn_key[80]; /**< keyword structure file name */
  char tag[80]; /**< struct tag name/reserved string space */
  char ctx[CONTEXT_NAME_SIZE]; /**< calling context */
  char srv[EXPORT_NAME_SIZE]; /**< calling equipment module */
  char prp[PROPERTY_NAME_SIZE]; /**< calling property */
  char dev[DEVICE_NAME_SIZE]; /**< calling device name */
  int len; /**< length of call */
  int fmt; /**< call format */
  int hbt; /**< archiving heartbeat (not to be exceeded) */
  int tmr; /**< polling interval */
  int infmt; /**< input data format (if any) */
  int inlen; /**< number of input data points */
  UINT32 opt; /**< option word */
  int  active;/**< is table entry active */
  int rec; /**< entry in record table */
} DAQRecInfo;
#define DAQRecTag "DAQRec"

typedef struct
{
  double tstamp;          /**< timestamp corresponding to data */
  short dataFlg;          /**< 1 = new data; 0 = old data; < 0 = error */
  short rec;              /**< entry in record table */
  UINT16 daqStatus;       /**< 0 = okay; -1 = no response; else hardware status */
  UINT16 dataStatus;      /**< returned status of data acquisition */
  UINT16 errCounter;      /**< number current consecutive errors */
  char ctx[CONTEXT_NAME_SIZE]; /**< calling context */
  char srv[EXPORT_NAME_SIZE]; /**< calling equipment module */
  char prp[PROPERTY_NAME_SIZE]; /**< calling property */
  char dev[DEVICE_NAME_SIZE]; /**< calling device name */
  char tag[TAG_NAME_SIZE];/**< calling tag name (structs) */
  char *data;             /**< pointer to data; */
  char *archdata;         /**< pointer to most recent archived data */
  char *lastdata;         /**< pointer to last most recent data */
  char *rfmtdata;         /**< pointer to reformatted data (for globals?) */
  BYTE *indata;           /**< pointer to input data (if any); */
  short infmt;            /**< input data format (if any) */
  short inlen;            /**< number of input data points */
  int  len;               /**< length of call */
  int  reclen;            /**< total number of bytes in call */
  int  fmt;               /**< call format */
  int  hbt;               /**< archiving heartbeat (not to be exceeded) */
  int  tmr;               /**< polling interval */
  int  mod;               /**< access mode */
  char fn_key[80];        /**< structure file name */
  char fn_inp[80];        /**< input file name */
  char resetMcaPrp[PROPERTY_NAME_SIZE]; /**< flag a reset MCA names property */
  int  nparam;            /**< number of parameters in structure */
  UINT32  option;         /**< option word */
  int  active;            /**< is table entry active */
  int  hasDataPool;       /**< for variable length formats */
  int  poolEntryCapacity; /**< for variable length formats */
  DAQParameter *daqParam; /**< pointer to data structure information */
  KeywordFailover *kwf;   /**< pointer to keyword failover if any */
  double archtime;        /**< timestamp of archived data */
  double lastarchtime;    /**< timestamp of last written archived data */
  double lasttime;        /**< timestamp of last most recent data */
  double laststs;         /**< timestamp of the last status archive */
  time_t lastOnDisk;      /**< last stored archive time for defunct record */
  time_t annotation;      /**< UTC time of active annotation (if any) */
  time_t lastannote;      /**< UTC time of last annotation output (if any) */
  int  sysstamp;          /**< system stamp data */
  int  archsys;           /**< system stamp of archive data */
  int  lastsys;           /**< last system stamp data */
  int  usrstamp;          /**< user stamp data */
  int  archusr;           /**< user stamp of archive data */
  int  lastusr;           /**< last user stamp data */
  int  archPending;       /**< flag to flush archive data to file */
  int  setArchPending;    /**< flag used by bound keywords */
  int  archYear;          /**< the archive year */
  int  archMonth;         /**< the archive month */
  short useMinimalStorage;/**< record is archived with high resolution timestamp */
  short hasLegacyStorage; /**< special flag for backward compatibility reads */
  short hasVolatileMca;   /**< possible name list changes at any point in time */
  short hasSts;           /**< record is archived with status */
  short hasPoi;           /**< record should keep points of interest */
  short isPastFilter;     /**< current data set passes all filter settings */
  short hasBadTimestamp;  /**< flag to indicate validity of timestamp */
  char fname[16];         /**< current archive file name */
} DAQRecord;

extern DAQRecord *daqTbl;

typedef struct daqRecKey
{
  char key[PROPERTY_NAME_SIZE]; /**< keyword */
  int idx;  /**< table index */
  int rec;  /**< record number */
  int prm;  /**< parameter index */
  struct daqRecKey *nxt;
} DaqRecKey;
#define DAQREC_HASH_SIZE 211
extern DaqRecKey *gDaqRecKeyTbl[DAQREC_HASH_SIZE];
extern DaqRecKey *gDaqRecAssoc;

int addDaqRecAssoc(char *assoc,char *param,int id, int rec,int prm);
DaqRecKey *getDaqRecKey(char *key);
void rmvDaqRecKeys(void);
void rmvDaqRecKey(char *key);
DaqRecKey *makeDaqRecKey(char *key,int tbl,int rec,int prm);

#define SEEK_CLOSEST_OLD 0x01
#define SEEK_ROUNDTIMESTAMPS 0x02
long locateRecord(DAQRecord *daqr,double ts, double *rts, int seekFlags, int *hasAlternate);
double getLastArchivedTimeStamp(DAQRecord *daqr);

#define getRecordPrefixLength(ms) ((ms) ? sizeof(SINT32) : (sizeof(double)+2*sizeof(SINT32)))
#define getPrefix(i) (daqTbl[i].useMinimalStorage ? "ar" : "ta")
#define getDataFilename(i,yy,mm) \
    sprintf(daqTbl[i].fname,"%s%02d%02d01.%0x",getPrefix(i),yy,mm,daqTbl[i].rec);\
    daqTbl[i].archYear = 1900 + yy;\
    daqTbl[i].archMonth = mm;

typedef struct
{
  char alias[PROPERTY_NAME_SIZE];
  char name[PROPERTY_NAME_SIZE];
} PRP_ALIAS;
extern int nPrpAlias;
extern PRP_ALIAS *prpAlias;

typedef struct
{
  int number;
  char name[SUBSYSTEM_NAME_SIZE];
} SUB_SYS;
extern int nSubSystems;
extern SUB_SYS *subSystems;
void checkDaqSubSystem(char *subsys);
typedef struct
{
  char dbfile[10][64];
  int  ndbfiles;
  int  acquire;
  int  port;
  char addr[16];
} TSRVINFO;

typedef struct   /* Archive Request Structure */
{
  time_t starttime;
  short  index,
         minutes;
  char   keyword[16];
} ARS;

#define CHK_VALID_MIN 0x01
#define CHK_VALID_MAX 0x02
typedef struct
{
  char name[16];
  char beamtype[16];
  int  fmt;
  void *ptr;
  int fecindex;
  int typeindex;
  float minValid;
  float maxValid;
  int chkValid;
} CHECKNAME;

extern int p_stateindex;
extern int e_stateindex;

extern char gTargetStateServer[64];

typedef struct
{
  char typestr[32];
  char procstr[32];
  char trigstr[32];
} DECLPROC;

extern DECLPROC *eDeclaredProcedures;
extern int nEDeclaredProcedures;
extern DECLPROC *pDeclaredProcedures;
extern int nPDeclaredProcedures;
extern int isArchiver;
extern int nextFreeRecord;
extern char daqUpdateErrStr[256];

void registerDaqStructs(int maxDaqRecords);
void rec2info(DAQRecInfo *dst, DAQRecord *src);
int getDaqRecInfo(DAQRecInfo *dst,int *plen);
int getDaqRecInfoEx(DAQRecInfo *dst,int *plen,char *target);
int getDaqPrmInfo(int rec,DAQPrmInfo *dst,int *len);
int readDaqPrmFile(char *path,char *fn,DAQPrmInfo *dst,int *siz);
int getTableEntryFromRecordNumber(int rec);
int getDaqPrmInpt(int rec,DTYPE *dout);
int dumpDaqDb(char *path,char *fn,int dumpAll);
int dumpDaqPrmInfoElem(char *path,char *fn,DAQPrmInfo *prm,int len);
int getEntryFromKeywordFile(char *keywordFile,int *rec);
int getEntryFromAssociation(char *assoc,int start,int *rec,int *param);
int getEntryFromInputFile(char *inputFile,int *rec);
int addDaqPrmInfoElem(char *fn,DAQPrmInfo *prm,int len);
int addDaqRecInfoElements(DAQRecInfo *daqrec,int len);
int addDaqInpt(char *fn,BYTE *data,int fmt, int len);
int WriteRecord(DAQRecord *daqr);
int WriteStatusRecord(DAQRecord *daqr,int id,double dts);
int WriteAnnotationRecord(DAQRecord *daqr,double dts);
int getFilterKeyAndIndex(int length,int tblIdx,char *fltrKey,int *fltrIdx);
int getArchivedEntryFromKeyword(char *keyword,int *rec,int *param);
int addExtSrcInfo(void);
int getActiveAnnotations(void);
int dumpActiveAnnotations(void);
int ReadRecord(DAQRecord *daqr);
int applyActiveAnnotation(int rec,time_t atime,int cancelPrevious);
int ReadLastAnnotationRecord(DAQRecord *daqr,int *data);
int appendAnnotationRecords(DAQRecord *daqr,time_t starttime,time_t stoptime);
int removeAnnotationRecords(DAQRecord *daqr,time_t starttime, time_t stoptime);
int loadTraceData(BYTE *buffer,int n, int p, time_t tgttime, DTYPE *dout,int index,int fieldindex,int sample);
int ReadAnnotationRecord(DAQRecord *daqr,time_t tgttime,int *data,int *ts);
void getSubSysDecoration(char *ssLongName,char *ssDecoration);
void slave_mode_diagnostics( void);
int getArchivedKeywords(NAME64 *list, int *listsize);
time_t parseTimeString(char *ts);
int addDaqPrmPriors(DAQParameter *daqParam, NAME64DBLDBL *priors, int npriors);
int writePriorDevices(char* keyword, NAME64* lst, int len);
int getKeywordsWithPriors(NAME64* keys, int* len);
int loadPriorDevices(PriorReference* pr);

#endif  /* _ARCHIVE_H_ */

