/*
 * File archlib.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*  =============================================================
   ==
   ==   Dataserver routine: archlib.c
   ==
   ============================================================= */

#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "tine.h"
#include "archive.h"

int nrecords = 0, nparams = 0, recordsCapacity = 0;
extern char cbuff[NETMEX_BUF_SIZE];
extern char dpath[];
extern int archPending,totalNrRecords;
extern int archdebug, ramping, running, beam;
extern UINT16 Eramping,Pramping,Erunning,Prunning,Ebeam,Pbeam;
extern int DclEramping, DclPramping, DclLumi;
extern int DclEInj, DclPInj;
char *_strdate(char *buf);
char *_strtime(char *buf);

char *makedate(char *s)  /* Make a date string in the unusual NETMEX format */
{
  char timebuf[9], datebuf[9];

  _strdate(datebuf);
  sprintf(s,"19YY-MM-DD-%s",_strtime(timebuf));
  strncpy(&s[2],&datebuf[6],2); strncpy(&s[5],&datebuf[0],2);
  strncpy(&s[8],&datebuf[3],2);

  return s;
}
static float daqPoiRangeFactor = (float)CA_POI_RANGE_FACTOR;
static float daqPoiToleranceFactor = (float)CA_POI_TOLERANCE_FACTOR;
int isDaqPointOfInterest(int n, int *isTransition)
{
  int i,j,ispoi=FALSE,useRange;
  char *d,*a;
  double r, t, absv, deltav;
  *isTransition = FALSE;
  if (daqTbl[n].option & ONCE) return FALSE;
  if (daqTbl[n].lastarchtime == 0) return FALSE; /* just starting up ? */
  for (i=0; !ispoi && i<daqTbl[n].nparam; i++)
  {
    /* if (!daqTbl[n].daqParam[i].keepsPointsOfInterest) continue; << don't do this here! */
    d = daqTbl[n].data + daqTbl[n].daqParam[i].offset;
    a = daqTbl[n].archdata + daqTbl[n].daqParam[i].offset;
    r = daqTbl[n].daqParam[i].maximum - daqTbl[n].daqParam[i].minimum;
    useRange = daqTbl[n].daqParam[i].keepsPointsOfInterest == CA_USE_POI_RANGE ? TRUE : FALSE;
    /* useRange = daqTbl[n].daqParam[i].pTolerance > 0 ? FALSE : TRUE; */
    if (useRange && r == 0) return FALSE;
    for (j=0; !ispoi && j<daqTbl[n].daqParam[i].size; j++)
    {
      switch (daqTbl[n].daqParam[i].fmt)
      {
        case CF_DBLDBL:
        case CF_DOUBLE:
          absv = fabs(((double *)a)[j]);
          deltav = fabs(((double *)d)[j] - ((double *)a)[j]);
          break;
        case CF_FLTINT:
        case CF_FLOAT:
          absv = fabs(((float *)a)[j]);
          deltav = fabs(((float *)d)[j] - ((float *)a)[j]);
          break;
        case CF_INT16:
          absv = (double)abs(((short *)a)[j]);
          deltav = (double)abs(((short *)d)[j] - ((short *)a)[j]);
          break;
        case CF_INTINT:
        case CF_INT32:
          absv = (double)labs(((SINT32 *)a)[j]);
          deltav = (double)labs(((SINT32 *)d)[j] - ((SINT32 *)a)[j]);
          break;
        case CF_BYTE:
          absv = (double)abs(((BYTE *)a)[j]);
          deltav = (double)abs(((BYTE *)d)[j] - ((BYTE *)a)[j]);
          break;
        default:
          absv = 0;
          deltav = 0;
          break;
      }
      t = useRange ? r * daqPoiRangeFactor :
          daqPoiToleranceFactor * (daqTbl[n].daqParam[i].aTolerance + daqTbl[n].daqParam[i].pTolerance * absv);
      if (deltav > t) 
      {
        if (archdebug) dbglog("flag point of interest: %.64s -> delta %g allowed tolerance %g",
          daqTbl[n].daqParam[i].param,deltav,t);
        ispoi = TRUE;
        *isTransition = TRUE;
      }
    }
    if (!daqTbl[n].daqParam[i].keepsPointsOfInterest)
    { /* reject it, then ! */
      ispoi = FALSE;
      continue;
    }
  }
  return ispoi;
}
int gIsTrace = FALSE;
char *gTraceKey = "";
#define TOOFARINTHEFUTURE (3600 * 4)
#define BYTE unsigned char
int checkData(int n,int isPastFilter,int isSched) 
{ /* write data to archive file for given daqTbl entry */
  int i,j,k,off=0,advanceAndRecheck=FALSE,istrans=FALSE,isFirstNotPastFilter=FALSE;
  int doarchive=FALSE,istransition=FALSE,apending=FALSE,isHeartbeat=FALSE,filterTransition=FALSE;
  int dsize, rlen = daqTbl[n].reclen;
  UINT16 opt = (UINT16)daqTbl[n].option;
  double dt, t, ts = MakeDataTimeStamp();
  BYTE *rptr = daqTbl[n].archdata;
  BYTE *dptr = daqTbl[n].data;
  char *d,*a;

  if (opt&NEVER) 
  { /* no archive option */
    if (gIsTrace) dbglog("%.192s (rec. %d) rejected! (no archive option)",gTraceKey,daqTbl[n].rec);
    return 0; 
  }
  if (daqTbl[n].tstamp > ts + TOOFARINTHEFUTURE)
  {
    if (archdebug || gIsTrace)
    {
      dbglog("archive timestamp for %.32s/%.64s %.64s is too far in the future !(%g)",
        daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp,daqTbl[n].tstamp);
    }
    if (!daqTbl[n].hasBadTimestamp)
    {
      AppendHistoryLog("ignore %.32s/%.64s[%.64s] as timestamp %.32s is in the future",
        daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp,GetDataTimeString(daqTbl[n].tstamp,FALSE));
      daqTbl[n].hasBadTimestamp = TRUE;
    }
    return 0;     			    
  }
  daqTbl[n].hasBadTimestamp = FALSE;
  dt = daqTbl[n].tstamp - daqTbl[n].archtime;
  if (daqTbl[n].useMinimalStorage) dt = (double)((UINT32)daqTbl[n].tstamp - (UINT32)daqTbl[n].archtime);
  if (dt <= 0) 
  { /* data timestamp hasn't changed */
    if (archdebug || gIsTrace)
    {
      dbglog("archive timestamp unchanged for %.32s/%.64s %.64s (%g vs %g)",
        daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp,daqTbl[n].tstamp,daqTbl[n].archtime);
    }
    return 0;     			    
  }
  if (daqTbl[n].setArchPending)
  {
    doarchive = TRUE;
    if (gIsTrace) dbglog("%.192s (rec. %d) set archive pending",gTraceKey,daqTbl[n].rec);
    goto checkArchiveFlag;
  }
  if (daqTbl[n].archPending)
  { /* archive is still pending */
    if (archdebug || gIsTrace)
    {
      dbglog("archive still pending for %.32s/%.64s %.64s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
      return -1;   	    
    }
  }
  if (ts >= daqTbl[n].archtime + (double)daqTbl[n].hbt)
  { /* archive no matter what */
    if (archdebug || gIsTrace) dbglog("archive heartbeat for %.32s/%.64s %.64s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
    doarchive = TRUE;
    isHeartbeat = TRUE;
  }
  else
  { /* must be past any archive filters ... */
    if (!isPastFilter) 
    {
      if (archdebug || gIsTrace) dbglog("%.32s/%.64s %.64s is not past its archive filter criteria",
        daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
      if (!daqTbl[n].isPastFilter) return 0; /* but store the first data set not past the filter */
      isFirstNotPastFilter = TRUE;
      AppendHistoryLog("%.32s/%.64s %.64s : store initial value not past archive filter", daqTbl[n].srv, daqTbl[n].dev, daqTbl[n].prp);
    }
  }
  if (isPastFilter != (int)daqTbl[n].isPastFilter)
  {
    if (daqTbl[n].lastarchtime > 0)
    { /* otherwise we'll land here the very first time following a server start */
      AppendHistoryLog("%.32s/%.64s %.64s : mark as transition", daqTbl[n].srv, daqTbl[n].dev, daqTbl[n].prp);
      filterTransition = TRUE;
    }
    doarchive = TRUE;
  }
  daqTbl[n].isPastFilter = isPastFilter;
  if (daqTbl[n].hasDataPool)
  {
    DataPoolEntry *dpe = (DataPoolEntry *)daqTbl[n].data;
    DataPoolEntry *ref = (DataPoolEntry *)daqTbl[n].archdata;
    dptr = dpe->data;
    rptr = ref->data;
    rlen = dpe->length;
    switch (daqTbl[n].fmt)
    {
      case CF_KEYVALUE:
      case CF_STRING:
        if (memcmp(dptr,rptr,dpe->length)) doarchive = TRUE;
        break;
      case CF_IMAGE:
        doarchive = TRUE;
        break;
    }
    goto checkArchiveFlag;
  }
  if (doarchive) goto checkArchiveFlag; /* no reason to check tolerance */
  if (archdebug || gIsTrace) 
    dbglog("checking tolerance for %.32s/%.64s %.64s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
  for (i=0; !doarchive && i<daqTbl[n].nparam; i++)
  {
    d = daqTbl[n].data + daqTbl[n].daqParam[i].offset;
    a = daqTbl[n].archdata + daqTbl[n].daqParam[i].offset;
    switch (daqTbl[n].daqParam[i].fmt)
    {
      case CF_DBLDBL:
           advanceAndRecheck = TRUE;
      case CF_DOUBLE:
checkDoubles:
           for (j=0; !doarchive && j<daqTbl[n].daqParam[i].size; j++)
           {
             t = daqTbl[n].daqParam[i].aTolerance +
                 daqTbl[n].daqParam[i].pTolerance * fabs(((double *)a)[j]);
             if ( fabs(((double *)d)[j] - ((double *)a)[j]) > t) doarchive = TRUE;
           }
           if (advanceAndRecheck)
           {
             d += sizeof(double);
             a += sizeof(double);
             advanceAndRecheck = FALSE;
             goto checkDoubles;
           }
           break;
      case CF_SPECTRUM:
           if (memcmp(d, a, SPECTRUM_HDR_SIZE))
           {
             doarchive = TRUE;
             break;
           }
           off = SPECTRUM_HDR_SIZE / sizeof(float);
           goto checkFloats;
      case CF_GSPECTRUM:
           if (memcmp(d, a, GSPECTRUM_HDR_SIZE))
           {
             doarchive = TRUE;
             break;
           }
           off = GSPECTRUM_HDR_SIZE / sizeof(float);
           goto checkFloats;
      case CF_FLTFLT:
           advanceAndRecheck = TRUE;
      case CF_FLTINT:
      case CF_FLOAT:
           off = 0;
checkFloats:
           for (j=off; !doarchive && j<daqTbl[n].daqParam[i].size; j++)
           {
             t = daqTbl[n].daqParam[i].aTolerance +
                 daqTbl[n].daqParam[i].pTolerance * fabs(((float *)a)[j]);
             if ( fabs(((float *)d)[j] - ((float *)a)[j]) > t) doarchive = TRUE;
           }
           if (advanceAndRecheck)
           {
             d += sizeof(float);
             a += sizeof(float);
             advanceAndRecheck = FALSE;
             goto checkFloats;
           }
           break;
      case CF_INT16:
           for (j=0; !doarchive && j<daqTbl[n].daqParam[i].size; j++)
           {
             t = daqTbl[n].daqParam[i].aTolerance +
                 daqTbl[n].daqParam[i].pTolerance * abs(((short *)a)[j]);
             if ( abs(((short *)d)[j] - ((short *)a)[j]) > t) doarchive = TRUE;
           }
           break;
      case CF_INTINT:
           advanceAndRecheck = TRUE;
      case CF_INT32:
checkInts:
           for (j=0; !doarchive && j<daqTbl[n].daqParam[i].size; j++)
           {
             t = daqTbl[n].daqParam[i].aTolerance +
                 daqTbl[n].daqParam[i].pTolerance * labs(((SINT32 *)a)[j]);
             if ( labs(((SINT32 *)d)[j] - ((SINT32 *)a)[j]) > t) doarchive = TRUE;
           }
           if (advanceAndRecheck)
           {
             d += sizeof(SINT32);
             a += sizeof(SINT32);
             advanceAndRecheck = FALSE;
             goto checkInts;
           }
           break;
      case CF_BYTE:
           for (j=0; !doarchive && j<daqTbl[n].daqParam[i].size; j++)
           {
             t = daqTbl[n].daqParam[i].aTolerance +
                 daqTbl[n].daqParam[i].pTolerance * abs(((BYTE *)a)[j]);
             if ( abs(((BYTE *)d)[j] - ((BYTE *)a)[j]) > t) doarchive = TRUE;
           }
           break;
      default:
           dsize = fmtsizeof(daqTbl[n].daqParam[i].fmt)*daqTbl[n].daqParam[i].size;
           dsize += getFormatHeaderSize(daqTbl[n].daqParam[i].fmt);
           if (!doarchive && memcmp(d,a,dsize)) doarchive = TRUE;
           break;
    }
  }
  if (archdebug || gIsTrace) 
    dbglog("tolerance check for %.32s/%.64s %.64s -> %s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp,(doarchive?"out of tolerance":"within tolerance"));
checkArchiveFlag:
  if (doarchive)
  {
    if (archdebug || gIsTrace) dbglog("signaling reason to archive %.32s/%.64s %.64s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
    apending = isDaqPointOfInterest(n,&istrans) ? 2 : 1;
    if (isHeartbeat) 
    { /* don't allow points-of-interest if heartbeat signal (some other filter is in play) */
      apending = 1; 
    }
    daqTbl[n].archPending = apending;
    istransition = gIsValidInTransition;
    if (filterTransition) istransition = TRUE; /* pass thru the filter again: store previous data set */
    if (daqTbl[n].archPending > 1) istransition = TRUE; /* marked as 'interesting' */
    if (istrans) istransition = istrans;
    if (daqTbl[n].hasDataPool) istransition = FALSE; /* not these guys */
    if (istransition && daqTbl[n].archtime != 0 && !daqTbl[n].hasDataPool)
    { /* insert transition record in order to smooth data browsing */
      if (daqTbl[n].archtime < daqTbl[n].lasttime)
      { /* don't have this point already */
        daqTbl[n].archtime = daqTbl[n].lasttime;
        daqTbl[n].archsys = daqTbl[n].lastsys;
        daqTbl[n].archusr = daqTbl[n].lastusr;
        memcpy(daqTbl[n].archdata,daqTbl[n].lastdata,daqTbl[n].reclen);
        WriteRecord(&daqTbl[n]);
        daqTbl[n].lastarchtime = daqTbl[n].lasttime; /* reset this ! */
        daqTbl[n].archPending = apending;
        if (archdebug || gIsTrace) dbglog("inserting transition record for %.32s/%.64s %.64s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
      }
    }
    daqTbl[n].archtime = daqTbl[n].tstamp;
    daqTbl[n].archsys = daqTbl[n].sysstamp;
    daqTbl[n].archusr = daqTbl[n].usrstamp;
    daqTbl[n].setArchPending = 0;
    if (rptr == NULL) goto checkSched;
    memcpy(rptr,dptr,rlen);
    /* is this bound to another keyword ? */
    if (strlen(daqTbl[n].daqParam[0].boundTo) > 0)
    {
      if ((k=getEntryFromKeyword(daqTbl[n].daqParam[0].boundTo,NULL,NULL)) >= 0 &&
          !daqTbl[k].archPending)
      {
        daqTbl[k].archPending = 1;
        daqTbl[k].setArchPending = 1;
        if (daqTbl[k].archtime != daqTbl[k].tstamp) /* it was dropped, put it back */
        {
          daqTbl[k].archtime = daqTbl[k].tstamp;
          daqTbl[k].archsys = daqTbl[k].sysstamp;
          daqTbl[k].archusr = daqTbl[k].usrstamp;
          memcpy(daqTbl[k].archdata,daqTbl[k].data,daqTbl[k].reclen);
        }
      }
    }
  checkSched:
    if (isFirstNotPastFilter)
    {
      AppendHistoryLog("%.32s/%.64s %.64s : signal archive record <pending = %d>", 
        daqTbl[n].srv, daqTbl[n].dev, daqTbl[n].prp,daqTbl[n].archPending);
    }
    if (isSched)
    { /* do this here! otherwise you wait until the background calls outputData() */
      if (archdebug || gIsTrace) 
        dbglog("write scheduled record %.32s/%.64s %.64s",daqTbl[n].srv,daqTbl[n].dev,daqTbl[n].prp);
      WriteRecord(&daqTbl[n]);
    }
    else
    {
      archPending++;
    }
  }
  return 0;
}

