/*
 * File arctools.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
   =============================================================
   ==
   ==   Event archive script reader + GetAndStore()
   ==
   =============================================================
*/
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "tine.h"
#include "threader.h"
#if defined(UNIX)
#include "dirent.h"
#elif defined(WIN32)
#include "direct.h"
#endif
#include "arctools.h"
#include "arceqm.h"
#include "apceqm.h"

char *archiveListFile = ARCHIVE_LIST_FILE;
char *alarmListFile = ALARM_LIST_FILE;
ARB *arbTbl = NULL;
int arbTblSiz = 0;
NAME64 *arbTrgList = NULL;
int useConnectedLinks = FALSE;
BYTE *PreallocatedDataBuffer = NULL;
int PreallocatedDataBufferSize = PREALLOCSIZE;
BYTE qryMemPool[MEM_POOL_SIZE];
UINT32 lastEventNumber = 0;
int EventNotificationPending = FALSE;
int RecordEventPending = FALSE;
int archiveDebugOutput = TRUE;
int GetARBIndex(char *srvName);
int RecordEvent(UINT32 eventNumber,UINT32 eventCode,char *devName);
int SetGlobalEvent(UINT32 eventCode,char *name);
int getFiles(NAME8 *filebuf,char *fileext,time_t starttime,time_t stoptime,int bufsiz);
time_t getTimeStampFromLogEntry(char *str);
int GetAndStore(ARB *arb, int step);
//short ftoi(char *fmt);
void checkSessionActivity(void);
BYTE *archiveBuffer = NULL;
int archiveBufferSize = ARCH_BUF_SIZE;
int archDebug = 0;
char pmArchEqm[EQM_NAME_SIZE] = "";

int getSetNumber(char *trigger)
{
  int i;
  if (trigger == NULL || strlen(trigger) == 0) return -1;
  for (i = 0; i < arbTblSiz && strnicmp(trigger, arbTbl[i].trg, DEVICE_NAME_SIZE); i++);
  return i == arbTblSiz ? -1 : i;
}
int cmdDumpSessions(int arg0,int arg1,int arg2,int arg3)
{
  arg0 = arg1 = arg2 = arg3 = 0;
  dumpSessions();
  return 0;
}
void dumpSessions(void)
{
  int i;
  dbglog("SESSION\tFile\tLast Access\n");
  for (i=0; i<MAXSESSION; i++) if (session[i].active)
  {
    dbglog("%d (%d)\t%s\t%s\n",
          i+1,session[i].count,session[i].fullName,ctime(&session[i].active));
  }
  return;
}
int cmdFreeSessions(int arg0,int arg1,int arg2,int arg3)
{
  arg0 = arg1 = arg2 = arg3 = 0;
  freeSessions();
  return 0;
}
int applyDeviceNames(char *tgtDevName)
{
  if (tgtDevName[0] == '#') return TRUE;
  if (!strcmp(tgtDevName,DEVICE_ALL)) return TRUE;
  if (!strcmp(tgtDevName,DEVICE_IGNORE)) return TRUE;
  return FALSE;
}
int registerDHSAlias(char *stag)
{
  int cc;
  char tag[32];
  if (stag == NULL) return argument_list_error;
  strncpy(tag,stag,TAG_NAME_SIZE); tag[TAG_NAME_SIZE] = 0;
  if (GetStructFromTag(tag) != NULL) return 0; /* already registered */

  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,channelID),32,CF_TEXT,"channelID")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eventcode),1,CF_INT32,"eventcode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,arraysize),1,CF_INT16,"arraysize")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datatype),1,CF_INT16,"datatype")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,samplingrate),1,CF_FLOAT,"samplingrate")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,pretrigger),1,CF_INT32,"pretrigger")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,timeoffset),1,CF_INT32,"timeoffset")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,lowcount),1,CF_INT32,"lowcount")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,highcount),1,CF_INT32,"highcount")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egu),16,CF_TEXT,"egu")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egulow),1,CF_FLOAT,"egulow")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eguhigh),1,CF_FLOAT,"eguhigh")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datalength),1,CF_INT32,"datalength")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,desc),4*64,CF_TEXT,"desc")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(DHS),1)) != 0) ccerr(cc);
err:
  return cc;
}
void registerStructs(void)
{
  int cc;
  char *tag;
  static int done = FALSE;
  if (done) return;
  done = TRUE;

  tag = "AESNOT";
  if ((cc=addFieldToStruct(tag,OFFSETIN(AESnotify,version),1,CF_INT32,"version")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AESnotify,eventNumber),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AESnotify,eventCode),1,CF_INT32,"eventCode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AESnotify,reserved),1,CF_INT32,"reserved")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AESnotify,comment),4*64,CF_TEXT,"comment")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(AESnotify),1)) != 0) ccerr(cc);

  tag = "AESCON";
  if ((cc=addFieldToStruct(tag,OFFSETIN(AEScontext,version),1,CF_INT32,"version")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AEScontext,eventNumber),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AEScontext,eventCode),1,CF_INT32,"eventCode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AEScontext,reserved),1,CF_INT32,"reserved")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AEScontext,deviceName),16,CF_TEXT,"deviceName")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(AEScontext),1)) != 0) ccerr(cc);

  tag = "AES";
  if ((cc=addFieldToStruct(tag,OFFSETIN(AES,eventNumber),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AES,eventCode),1,CF_INT32,"eventCode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AES,comment),4*64,CF_TEXT,"comment")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AES,dsMask),32,CF_BYTE,"dsMask")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(AES),100)) != 0) ccerr(cc);

  tag = "ADRB";
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,eventNumber),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,starttime),1,CF_INT32,"starttime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,stoptime),1,CF_INT32,"stoptime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,devSpecifier.deviceServerName),16,CF_TEXT,"deviceServerName")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,devSpecifier.deviceProperty),16,CF_TEXT,"deviceProperty")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,devSpecifier.deviceName),16,CF_TEXT,"deviceName")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,devSpecifier.dataFormat),1,CF_INT16,"dataFormat")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB5,devSpecifier.dataDimension),1,CF_INT16,"dataDimension")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(ADRB5),1)) != 0) ccerr(cc);

  tag = "ADRBX";
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,eventNumber),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,starttime),1,CF_INT32,"starttime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,stoptime),1,CF_INT32,"stoptime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.deviceContext),16,CF_TEXT,"deviceContext")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.deviceServerName),16,CF_TEXT,"deviceServerName")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.deviceProperty),32,CF_TEXT,"deviceProperty")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.deviceName),32,CF_TEXT,"deviceName")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.dataFormat),1,CF_INT16,"dataFormat")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.reserved),1,CF_INT16,"reserved")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.dataIndex),1,CF_INT32,"dataIndex")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRBX,devSpecifier.dataDimension),1,CF_INT32,"dataDimension")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(ADRBX),1)) != 0) ccerr(cc);

  tag = "ADRBr4";
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,eventNumber),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,starttime),1,CF_INT32,"starttime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,stoptime),1,CF_INT32,"stoptime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceContext),CONTEXT_NAME_SIZE,CF_TEXT,"context")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceServerName),EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceProperty),PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceName),DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataTag),TAG_NAME_SIZE,CF_TEXT,"dataTag")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataFormat),1,CF_INT16,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.reserved),1,CF_INT16,"reserved")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataIndex),1,CF_INT32,"index")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataDimension),1,CF_INT32,"dataDimension")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(ADRB),1)) != 0) ccerr(cc);

  tag = "AIB";
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,numProperties),1,CF_INT16,"numProperties")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,numDevices),1,CF_INT16,"numDevices")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,eventCode),1,CF_INT32,"eventCode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,comment),64,CF_TEXT,"comment")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(AIB),1)) != 0) ccerr(cc);

  tag = "PIB";
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,SamplingRate),1,CF_FLOAT,"SamplingRate")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,numPreTriggerValues),1,CF_INT32,"numPreTriggerValues")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,TimeOffset),1,CF_INT32,"TimeOffset")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,LowCount),1,CF_INT32,"LowCount")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,HighCount),1,CF_INT32,"HighCount")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,Units),16,CF_TEXT,"Units")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,LowEGU),1,CF_DOUBLE,"LowEGU")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,HighEGU),1,CF_DOUBLE,"HighEGU")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,numDevices),1,CF_INT16,"numDevices")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,reserved),1,CF_INT16,"reserved")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(PIB),1)) != 0) ccerr(cc);

  tag = "DHS";
  if (GetStructFromTag(tag) == NULL)
  {
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,channelID),32,CF_TEXT,"channelID")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eventcode),1,CF_INT32,"eventcode")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,arraysize),1,CF_INT16,"arraysize")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datatype),1,CF_INT16,"datatype")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,samplingrate),1,CF_FLOAT,"samplingrate")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,pretrigger),1,CF_INT32,"pretrigger")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,timeoffset),1,CF_INT32,"timeoffset")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,lowcount),1,CF_INT32,"lowcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,highcount),1,CF_INT32,"highcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egu),16,CF_TEXT,"egu")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egulow),1,CF_FLOAT,"egulow")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eguhigh),1,CF_FLOAT,"eguhigh")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datalength),1,CF_INT32,"datalength")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,desc),4*64,CF_TEXT,"desc")) != 0) ccerr(cc);
    if ((cc=sealTaggedStruct(tag,sizeof(DHS),1)) != 0) ccerr(cc);
  }

  tag = "HEADER";
  if (GetStructFromTag(tag) == NULL)
  {
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,channelID),32,CF_TEXT,"channelID")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eventcode),1,CF_INT32,"eventcode")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,arraysize),1,CF_INT16,"arraysize")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datatype),1,CF_INT16,"datatype")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,samplingrate),1,CF_FLOAT,"samplingrate")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,pretrigger),1,CF_INT32,"pretrigger")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,timeoffset),1,CF_INT32,"timeoffset")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,lowcount),1,CF_INT32,"lowcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,highcount),1,CF_INT32,"highcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egu),16,CF_TEXT,"egu")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egulow),1,CF_FLOAT,"egulow")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eguhigh),1,CF_FLOAT,"eguhigh")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datalength),1,CF_INT32,"datalength")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,desc),4*64,CF_TEXT,"desc")) != 0) ccerr(cc);
    if ((cc=sealTaggedStruct(tag,sizeof(DHS),1)) != 0) ccerr(cc);
  }

  tag = "TrgCmp";
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,ctx),32,CF_TEXT,"context")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,srv),32,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,prp),64,CF_TEXT,"trigger")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,dev),64,CF_TEXT,"description")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(TrgChannel),64)) != 0) ccerr(cc);

  tag = "TrgChn";
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,ctx),32,CF_TEXT,"context")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,srv),32,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,prp),64,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgChannel,dev),64,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(TrgChannel),64)) != 0) ccerr(cc);

  tag = "TrgView";
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgViewFile,ctx),32,CF_TEXT,"context")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgViewFile,trg),64,CF_TEXT,"trigger")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgViewFile,desc),64,CF_TEXT,"desc")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgViewFile,file),64,CF_TEXT,"file")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgViewFile,components),1,CF_INT32,"reference")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(TrgViewFile,nComponents),1,CF_INT32,"nComponents")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(TrgViewFile),64)) != 0) ccerr(cc);

  tag = "PHSr4";
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,dsc),PROPERTY_DESC_SIZE,CF_TEXT,"description")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,format),1,CF_INT16,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,arrayType),1,CF_INT16,"arrayType")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,arraySize),1,CF_INT32,"arraySize")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,preTrigger),1,CF_INT32,"preTrigger")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,unitsY),UNITS_SIZE,CF_TEXT,"unitsY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,unitsX),UNITS_SIZE,CF_TEXT,"unitsX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,rangeMinX),1,CF_FLOAT,"rangeMinX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,rangeMaxX),1,CF_FLOAT,"rangeMaxX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,scaleMinY),1,CF_FLOAT,"scaleMinY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PHS,scaleMaxY),1,CF_FLOAT,"scaleMaxY")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(PHS),100)) != 0) ccerr(cc);

  tag = "EVNDHDR";
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,evnCode),1,CF_INT32,"eventCode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,evnNr),1,CF_INT32,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,datSiz),1,CF_INT32,"dataSize")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,datFmt),1,CF_INT16,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,aryTyp),1,CF_INT16,"arrayType")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,preTrg),1,CF_INT32,"preTrigger")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,reserved),1,CF_INT32,"reserved")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,shiftX),1,CF_FLOAT,"shiftX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,scaleX),1,CF_FLOAT,"scaleX)")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,shiftY),1,CF_FLOAT,"shiftY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,scaleY),1,CF_FLOAT,"scaleY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,rangeMinX),1,CF_FLOAT,"rangeMinX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,rangeMaxX),1,CF_FLOAT,"rangeMaxX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,scaleMinY),1,CF_FLOAT,"scaleMinY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,scaleMaxY),1,CF_FLOAT,"scaleMaxY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,unitsX),UNITS_SIZE,CF_TEXT,"unitsX")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,unitsY),UNITS_SIZE,CF_TEXT,"unitsY")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,desc),256,CF_TEXT,"description")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(EVNDHDR,fcnTag),64,CF_TEXT,"function")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(EVNDHDR),1)) != 0) ccerr(cc);

  tag = "ARBQS";
  if ((cc=addFieldToStruct(tag,OFFSETIN(ARBQS,trg),64,CF_TEXT,"eventTrigger")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ARBQS,ext),16,CF_TEXT,"eventExtension")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ARBQS,desc),64,CF_TEXT,"description")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ARBQS,depthInMonths),1,CF_INT32,"purgeDepth")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ARBQS,recordEvents),1,CF_INT16,"recordFlag")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ARBQS,maxQueueSize),1,CF_INT16,"maxQueueSize")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(ARBQS),256)) != 0) ccerr(cc);

  tag = "DASQS";
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,ctx),CONTEXT_NAME_SIZE,CF_TEXT,"context")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,srvArc),EXPORT_NAME_SIZE,CF_TEXT,"serverArchive")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,prpArc),PROPERTY_NAME_SIZE,CF_TEXT,"propertyArchive")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,infile),16,CF_TEXT,"inputFile")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,datTag),TAG_NAME_SIZE,CF_TEXT,"dataTag")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,datTagIn),TAG_NAME_SIZE,CF_TEXT,"dataTagInput")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,datSiz),1,CF_INT32,"dataSize")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,datFmt),1,CF_INT32,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,datFmtIn),1,CF_INT32,"formatInput")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,access),1,CF_INT32,"access")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,wait),1,CF_INT32,"waitTime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,timeout),1,CF_INT32,"timeout")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,value),1,CF_INT32,"value")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,mode),1,CF_INT32,"mode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,monitorRate),1,CF_INT32,"monitorRate")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,monitorTime),1,CF_INT32,"monitorTime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,scale),1,CF_FLOAT,"scale")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(DASQS,shift),1,CF_FLOAT,"shift")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(DASQS),100)) != 0) ccerr(cc);

  tag = "AcqMode";
  if ((cc=OpenBitField(NULL,tag,CF_INT16)) == 0)
  {
    AddFieldToBitField(NULL,tag,0x01,"READ");
    AddFieldToBitField(NULL,tag,0x02,"WRITE");
    AddFieldToBitField(NULL,tag,0x04,"MATCH");
    AddFieldToBitField(NULL,tag,0x08,"WAIT");
    AddFieldToBitField(NULL,tag,0x10,"SETNUM");
    AddFieldToBitField(NULL,tag,0x20,"PREALLOC");
    AddFieldToBitField(NULL,tag,0x40,"FORWARD");
    AddFieldToBitField(NULL,tag,0x80,"SERIES");
    AddFieldToBitField(NULL,tag,0x100,"MONITOR");
    AddFieldToBitField(NULL,tag,0x200,"PERSIST");
    AddFieldToBitField(NULL,tag,0x400,"SETDEV");
    AddFieldToBitField(NULL,tag,0x4000,"DISABLE");
  }
err:
  if (cc) feclog("TaggedStruct %.16s: %.32s",tag,cc2str(cc));
  return;
}
short acctoi(char *mod)
{
  short accessMode = 0;
  if (strstr(strupr(mod),"READ")) accessMode |= READ;
  if (strstr(strupr(mod),"WRITE")) accessMode |= WRITE;
  if (strstr(strupr(mod),"POLL")) accessMode |= MATCH;
  if (strstr(strupr(mod),"MATCH")) accessMode |= MATCH;
  if (strstr(strupr(mod),"WAIT")) accessMode |= WAIT;
  if (strstr(strupr(mod),"SETNUM")) accessMode |= SETNUM;
  if (strstr(strupr(mod),"PREALLOC")) accessMode |= PREALLOC;
  if (strstr(strupr(mod),"FORWARD")) accessMode |= FORWARD;
  if (strstr(strupr(mod),"SERIES")) accessMode |= SERIES;
  if (strstr(strupr(mod),"MONITOR")) accessMode |= MONITOR;
  if (strstr(strupr(mod),"PERSIST")) accessMode |= PERSIST;
  if (strstr(strupr(mod),"SETDEV")) accessMode |= SETDEV;
  if (strstr(strupr(mod),"DISABLE")) accessMode |= DISABLE;
  return accessMode;
}
char *modtoacc(short mode)
{
  static char accstr[128];
  accstr[0] = 0;
  if (mode & READ) strcat(accstr,"READ");
  if (mode & WRITE) strcat(accstr,strlen(accstr)?"|WRITE":"WRITE");
  if (mode & MATCH) strcat(accstr,strlen(accstr)?"|MATCH":"MATCH");
  if (mode & WAIT) strcat(accstr,strlen(accstr)?"|WAIT":"WAIT");
  if (mode & SETNUM) strcat(accstr,strlen(accstr)?"|SETNUM":"SETNUM");
  if (mode & PERSIST) strcat(accstr,strlen(accstr)?"|PERSIST":"PERSIST");
  if (mode & PREALLOC) strcat(accstr,strlen(accstr)?"|PREALLOC":"PREALLOC");
  if (mode & FORWARD) strcat(accstr,strlen(accstr)?"|FORWARD":"FORWARD");
  if (mode & SERIES) strcat(accstr,strlen(accstr)?"|SERIES":"SERIES");
  if (mode & MONITOR) strcat(accstr,strlen(accstr)?"|MONITOR":"MONITOR");
  if (mode & SETDEV) strcat(accstr,strlen(accstr)?"|SETDEV":"SETDEV");
  if (mode & DISABLE) strcat(accstr,strlen(accstr)?"|DISABLE":"DISABLE");
  return accstr;
}

#define logerr(txt) { perror("open"); strcpy(s,txt); goto error; }

FILE *createAndOpen(char *fname,char *attr)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  static char strdel[2] = { FS_DELIMITER , 0};

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,strdel); c != NULL; c=strtok(NULL,strdel))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return fopen(fname,attr);
    }
    strcat(dirname,strdel);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) return NULL;
    }
#   elif defined(WIN32)
    sprintf(fn,"%.120s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return NULL;
    }
#   endif
  }
  return NULL;
}
/* ----------------------------------------------------------------- */
/*  *makefilename                                                    */
/* ----------------------------------------------------------------- */

char *makefilename(char *fname,char *src,char *ext,time_t t)
{
  struct tm *ts;
  int yr, mo;

  ts = localtime(&t);
  yr = ts->tm_year + 1900;
  mo = ts->tm_mon + 1;

  sprintf(fname,"%s%c%d%c%02d%c%.16s%c%lx.%.16s",
    src,FS_DELIMITER,yr,FS_DELIMITER,mo,FS_DELIMITER,ext,FS_DELIMITER,t,ext);
  return fname;
}
int getNumEntries(char *fn)
{
  FILE *fp = NULL;
  char str[256];
  int n=0;

  if ((fp=fopen(fn,"r")) == NULL) return -no_such_file;
  while (csvGetLine(str,255,fp) != NULL)
  {
     strtrm(str);
     if (strlen(str) == 0) continue;            /* carriage return */
     if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
     n++;
  }
  fclose(fp);
  return n-1;  /*  subtract off the header */
}
int getTriggerAlarms(char *fn,char *eqm)
{ //TRIGGER, CODE, SYSTEM, TAG, TEXT, SEVERITY, USER
  FILE *fp = NULL;
  int cc = 0, done = 0;
  int trg_col, cod_col, sys_col, tag_col, txt_col, sev_col, usr_col;
  char str[256], hdr[256], scratch[64];
  int i = 0, n;
  ADS ads;
  if ((n=getNumEntries(fn)) < 0) return n;
  if ((fp = fopen(fn, "r")) == NULL) ccerr(no_such_file);
  while (csvGetLine(str, 255, fp) != NULL && i < n)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n", str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr, str);
      if ((trg_col = findcol(hdr, "TRIGGER")) < 0) ccerr(no_such_column);
      if ((cod_col = findcol(hdr, "ALARM_CODE")) < 0) ccerr(no_such_column);
      if ((sys_col = findcol(hdr, "ALARM_SYSTEM")) < 0) ccerr(no_such_column);
      if ((tag_col = findcol(hdr, "ALARM_TAG")) < 0) ccerr(no_such_column);
      if ((sev_col = findcol(hdr, "ALARM_SEVERITY")) < 0) ccerr(no_such_column);
      txt_col = findcol(hdr, "ALARM_TEXT");
      usr_col = findcol(hdr, "USER");
      done = 1;                              /* read header info */
      continue;
    }
    if ((i=getSetNumber(colptr(trg_col, str))) < 0 || i > arbTblSiz) continue;
    memset(&ads, 0, sizeof(ADS));
    strncpy(scratch, colptr(cod_col, str), 16);
    ads.alarmCode = (UINT32)atoi(scratch);
    // copy alarm code into arbTbl[i]
    arbTbl[i].almCode = ads.alarmCode;
    strncpy(scratch, colptr(sys_col, str), 16);
    ads.alarmSystem = (UINT16)atoi(scratch);
    strncpy(ads.alarmTag, colptr(tag_col, str), 32);
    strncpy(scratch, colptr(sev_col, str), 16);
    ads.alarmSeverity = (short)atoi(scratch);
    if (txt_col >= 0)
    {
      strncpy(ads.alarmTag, colptr(tag_col, str), 32);
    }
    else
    {
      strncpy(ads.alarmTag, ads.alarmTag, 32);
    }
    if (usr_col >= 0)
    {
      // copy allowed alarm trigger setter into arbTbl[i]
      strncpy(arbTbl[i].almTrgUsr, colptr(usr_col, str), 16);
    }
    AppendAlarmInfoTable(eqm, &ads);
  }
err:
  if (fp) fclose(fp);
  feclog("trigger alarms file %s : %.32s", fn, cc ? cc2str(cc) : "found");
  return cc ? -cc : n;
}
int getArchiveSets(char *fn)
{
  FILE *fp = NULL;
  int cc=0,done=0;
  int ext_col,trg_col,rec_col,dep_col,dsc_col,que_col;
  char str[256],hdr[256],scratch[16];
  int i=0,nsets;

  if ((nsets=getNumEntries(fn)) < 0) return nsets;
  if ((arbTbl=(ARB *)calloc(nsets,sizeof(ARB))) == NULL) ccerr(out_of_local_memory);

  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (csvGetLine(str,255,fp) != NULL && i<nsets)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((trg_col=findcol(hdr,"TRIGGER")) < 0 &&
          (trg_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((ext_col=findcol(hdr,"EXTENSION")) < 0) ccerr(no_such_column);
      rec_col = findcol(hdr,"RECORD");
      dep_col = findcol(hdr,"DEPTH");
      dsc_col = findcol(hdr,"DESCRIPTION");
      que_col = findcol(hdr,"QUEUE_SIZE");
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(arbTbl[i].trg,colptr(trg_col,str),DEVICE_NAME_SIZE);
    strncpy(arbTbl[i].ext,colptr(ext_col,str),16);
    if (rec_col >= 0)
    {
      strncpy(scratch,colptr(rec_col,str),16);
      arbTbl[i].recordEvents = atoi(scratch);
    }
    else
    {
      arbTbl[i].recordEvents = 0;
    }
    if (dep_col >= 0)
    {
      strncpy(scratch,colptr(dep_col,str),16);
      arbTbl[i].depthInMonths = atoi(scratch);
    }
    if (dsc_col >= 0)
    {
      strncpy(arbTbl[i].desc,colptr(dsc_col,str),64);
    }
    if (que_col >= 0)
    {
      strncpy(scratch,colptr(que_col,str),16);
      arbTbl[i].maxq = atoi(scratch);
    }
    i++;
  }
err:
  if (fp) fclose(fp);
  feclog("archive sets file %s : %.32s",fn, cc ? cc2str(cc) : "found");
  return cc ? -cc : nsets;
}
int findDAS(DAS *das,int n)
{
  int i;
  
  for (i=0; i<n; i++)
  {
    if (strncmp(das[i].ctx,das[n].ctx,CONTEXT_NAME_SIZE)) continue;
    if (strncmp(das[i].srv,das[n].srv,EXPORT_NAME_SIZE)) continue;
    if (strncmp(das[i].prp,das[n].prp,PROPERTY_NAME_SIZE)) continue;
    if (strncmp(das[i].dev,das[n].dev,DEVICE_NAME_SIZE)) continue;
    if (das[i].dout.dArrayLength != das[n].dout.dArrayLength) continue;
    if (das[i].dout.dFormat != das[n].dout.dFormat) continue;
    if (strncmp(das[i].infile,das[n].infile,16)) continue;
    break;
  }
  return i;
}
int putDASInputData(char *filename,DTYPE *din)
{
  FILE *fp = NULL;
  char fn[256], lfn[64], *c;
  int i,cc=0;

  strncpy(lfn,filename,64);
  if ((c=strstr(lfn,".csv")) != NULL) *c = 0;
  sprintf(fn,"%s%c%s%cDATA%c%.16s.csv",
    stdArchiveDirectory,FS_DELIMITER,TRIGGER_FILES_SUBDIR,FS_DELIMITER,FS_DELIMITER,lfn);
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Data\n");
  for (i=0; i<(int)din->dArrayLength; i++)
  {
    switch (din->dFormat)
    {
      case CF_NAME16: fprintf(fp,"%.16s\n",&din->data.cptr[i*16]); break; 
      case CF_NAME32: fprintf(fp,"%.32s\n",&din->data.cptr[i*32]); break; 
      case CF_NAME48: fprintf(fp,"%.48s\n",&din->data.cptr[i*48]); break; 
      case CF_NAME64: fprintf(fp,"%.64s\n",&din->data.cptr[i*64]); break; 
      case CF_TEXT: fwrite(din->data.cptr,1,din->dArrayLength,fp); break;
      case CF_BYTE: fprintf(fp,"%d\n",din->data.bptr[i]); break;
      case CF_INT16: fprintf(fp,"%d\n",din->data.sptr[i]); break;
      case CF_INT32: fprintf(fp,"%d\n",din->data.lptr[i]); break;
      case CF_FLOAT: fprintf(fp,"%g\n", din->data.fptr[i]); break;
      case CF_DOUBLE: fprintf(fp,"%g\n", din->data.dptr[i]); break;
      default: ccerr(illegal_format);
    }
    if (din->dFormat == CF_TEXT)
    {
      fprintf(fp,"\n");
      break;
    }
  }
err:
  if (fp) fclose(fp);
  if (cc) feclog("Write Data Input File %s : %s",fn,cc2str(cc));
  return cc;
}
int getDASInputFileNames(DTYPE *dout)
{
  char dirname[256], *nptr;
  int i=0,nlen=16,maxfiles;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

  sprintf(dirname,"%s%c%s%cDATA",
    stdArchiveDirectory,FS_DELIMITER,TRIGGER_FILES_SUBDIR,FS_DELIMITER);

  if (dout == NULL) return argument_list_error;
  if (dout->dArrayLength < 1) return dimension_error;
  switch (dout->dFormat)
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      nlen = fmtsizeof(dout->dFormat);
      break;
    default:
      return illegal_format;
  }
  maxfiles = (int)dout->dArrayLength;
  nptr = dout->data.cptr;
# if defined(UNIX)
  if ((dp=opendir(dirname)) == NULL) return no_such_file;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name,".csv") == NULL) continue;
    strncpy(nptr,dirp->d_name,16);
# elif defined(WIN32)
  strcat(dirname,"\\*.csv");
  if ((hfp=_findfirst(dirname, &fd)) == -1) return no_such_file;
  do
  {
    strncpy(nptr,fd.name,16);
# endif
    nptr[16] = 0;
    if (++i == maxfiles) break;
    nptr += nlen;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
# endif
  dout->dArrayLength = i;
  return 0;
}
/**
 * \internal read in the input data for a DAS step
 * \param filename is the input data file name
 * \param din can be unitialized (as used internally at startup)
 *  to fill in the database parameters;
 *  or din can refer to a read query from e.g. the database
 *  manager to acertain the input data for a given file
 */
int getDASInputData(char *filename,DTYPE *din)
{
  FILE *fp = NULL;
  int fsiz,cc=0,done=0;
  int data_col;
  char str[256],hdr[256],scratch[16],fn[256],lfn[64],*c;
  int i=0,nlen,n,maxlen;

  fn[0] = 0;
  if (filename == NULL || strlen(filename) == 0) ccerr(argument_list_error);
  strncpy(lfn,filename,64);
  if ((c=strstr(lfn,".csv")) != NULL) *c = 0;
  sprintf(fn,"%s%c%s%cDATA%c%.16s.csv",
    stdArchiveDirectory,FS_DELIMITER,TRIGGER_FILES_SUBDIR,FS_DELIMITER,FS_DELIMITER,lfn);
  if (din == NULL) ccerr(illegal_read_write);
  if ((fsiz=fmtsizeof(din->dFormat)) <= 0) ccerr(illegal_format);
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(no_such_file);
  if (din->data.vptr == NULL &&
      (din->data.vptr=calloc(n,fsiz)) == NULL) ccerr(out_of_local_memory);
  maxlen = din->dArrayLength;
  if (maxlen == 0) maxlen = n;
  din->dArrayLength = n;

  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;            /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((data_col=findcol(hdr,"DATA")) < 0) ccerr(no_such_line);
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(scratch,colptr(data_col,str),16);
    switch (din->dFormat)
    {
      case CF_NAME16: 
      case CF_NAME32: 
      case CF_NAME48: 
      case CF_NAME64:
        nlen = fmtsizeof(din->dFormat);
        strncpy(&din->data.cptr[i*nlen],str,nlen); 
        break;
      case CF_TEXT: strncpy(din->data.cptr,str,din->dArrayLength); break;
      case CF_BYTE: din->data.bptr[i] = atoi(scratch); break;
      case CF_INT16: din->data.sptr[i] = atoi(scratch); break;
      case CF_INT32: din->data.lptr[i] = atoi(scratch); break;
      case CF_FLOAT: din->data.fptr[i] = (float)strtod(scratch,NULL); break;
      case CF_DOUBLE: din->data.dptr[i] = strtod(scratch,NULL); break;
      default: ccerr(illegal_format);
    }
    if (din->dFormat == CF_TEXT) break;
    if (++i == maxlen) break;
  }
  err:
  if (fp) fclose(fp);
  if (cc) feclog("Data Input File %s : %s",fn,cc2str(cc));
  return cc;
}
void getPathFromFileName(char *fn,char *path)
{
  char *c;
  strncpy(path,fn,80);
  if ((c=strrchr(path,FS_DELIMITER)) == NULL) strcpy(path,".");
  else *c = 0;
}
int getDAS(char *fn,DAS *das,int *isMonitoring)
{
  FILE *fp = NULL;
  int dsiz,ssiz,tsiz,fsiz,cc=0,done=0,rc;
  int ctx_col,srv_col,prp_col,dev_col,srvArc_col,prpArc_col,
      siz_col,fmt_col,acc_col,wait_col,value_col,to_col,
      scale_col,shift_col,file_col = -1,input_format_col = -1,
      rate_col=-1,mode_col=-1,time_col=-1;
  char str[256],hdr[256],scratch[128],path[80];
  char *c,stag[TAG_NAME_SIZE], dev[196], *ctx = gDeviceContext;
  int i=0,j;

  if (das == NULL) ccerr(un_allocated);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  getPathFromFileName(fn,path);
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;            /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      ctx_col = findcol(hdr,"CONTEXT");
      if ((srv_col=findcol(hdr,"SERVER")) < 0 &&
          (srv_col=findcol(hdr,"TAG")) < 0) ccerr(no_such_line);
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((dev_col=findcol(hdr,"DEVICE")) < 0) ccerr(no_such_line);
      if ((srvArc_col=findcol(hdr,"ARCHIVESERVER")) < 0) srvArc_col = findcol(hdr,"ARCHIVETAG");
      prpArc_col = findcol(hdr,"ARCHIVEPROPERTY");
      if ((siz_col=findcol(hdr,"SIZE")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_line);
      acc_col = findcol(hdr,"ACCESS");
      wait_col = findcol(hdr,"WAIT");
      value_col = findcol(hdr,"VALUE");
      to_col = findcol(hdr,"TIMEOUT");
      scale_col = findcol(hdr,"SCALE");
      shift_col = findcol(hdr,"SHIFT");
      file_col = findcol(hdr,"INPUTFILE");
      mode_col = findcol(hdr,"MONITORMODE");
      rate_col = findcol(hdr,"MONITORRATE");
      time_col = findcol(hdr,"MONITORTIME");
      input_format_col = findcol(hdr,"INPUTFORMAT");
      done = 1;                              /* read header info */
      continue;
    }
    if (ctx_col >= 0) ctx = colptr(ctx_col,str);
    strncpy(das[i].ctx,ctx,CONTEXT_NAME_SIZE);
    strncpy(das[i].srv,colptr(srv_col,str),EXPORT_NAME_SIZE);
    strncpy(das[i].prp,colptr(prp_col,str),PROPERTY_NAME_SIZE);
    strncpy(das[i].dev,colptr(dev_col,str),DEVICE_NAME_SIZE);
    strncpy(das[i].srvArc,das[i].srv,EXPORT_NAME_SIZE);
    if (srvArc_col >= 0)
    { /* column given */
      strncpy(scratch,colptr(srvArc_col,str),EXPORT_NAME_SIZE);
      if (strlen(scratch) > 0) strncpy(das[i].srvArc,scratch,EXPORT_NAME_SIZE);
    }
    strncpy(das[i].prpArc,das[i].prp,PROPERTY_NAME_SIZE);
    if (prpArc_col >= 0)
    { /* column given */
      strncpy(scratch,colptr(prpArc_col,str),PROPERTY_NAME_SIZE);
      if (strlen(scratch) > 0) strncpy(das[i].prpArc,scratch,PROPERTY_NAME_SIZE);
    }
    if ((c=strrchr(das[i].prp,'.')) != NULL)
    {
      if (!strcmp(c,".HIST@"))
      {
        das[i].isAcqHist = TRUE;
        if ((das[i].din.data.lptr=(int *)calloc(10,sizeof(int))) == NULL) ccerr(out_of_local_memory);
      }
    }
    /* output DTYPE: */
    strncpy(scratch,colptr(siz_col,str),16);
    das[i].dout.dArrayLength = atoi(scratch);
    strncpy(scratch,colptr(fmt_col,str),64);
    stag[0] = 0;
    if ((c=strchr(scratch,'/')) != NULL)
    {
      *c++ = 0;
      strncpy(stag,c,TAG_NAME_SIZE);
      sprintf(dev,"/%.32s/%.32s",das[i].ctx,das[i].srv);
      rc = AcquireAndRegisterStruct(dev,stag,das[i].dout.dArrayLength);
      if (rc != already_assigned) feclog("AcquireAndRegisterStruct %s from %s: %s",stag,dev,erlst[rc]);
      /* registerDHSAlias(stag); <- not any more ! */
    }
    das[i].dout.dFormat = ftoi(scratch);
    das[i].acc = READ;
    if (acc_col >= 0)
    {
      strncpy(scratch,colptr(acc_col,str),16);
      das[i].acc = acctoi(scratch);
      if (das[i].acc&SETDEV && !(das[i].acc&UNSETDEV))
        das[i].devTgtLst = (DEVTGTLST *)calloc(1,sizeof(DEVTGTLST));
    }
    if (das[i].dout.dFormat == CF_STRUCT && 
        das[i].dout.dArrayLength == 1) /* must be tagged struct */
    { /* if not given: tag = "DHS" */
      if (strlen(stag) == 0) strncpy(stag,"DHS",8);
      strncpy(das[i].dout.dTag,stag,TAG_NAME_SIZE);
    }
    else 
    { 
      strncpy(das[i].dout.dTag,stag,TAG_NAME_SIZE);
    }
    if (file_col >= 0)
    { /* input DTYPE: */
      das[i].din.dFormat = CF_INT32; /* default; */
      if (input_format_col >= 0)
      {
        strncpy(scratch,colptr(input_format_col,str),16);
        if (strlen(scratch) > 0) das[i].din.dFormat = ftoi(scratch);
      }
      strncpy(scratch,colptr(file_col,str),16);
      if (strlen(scratch) > 0)
      {
        strncpy(das[i].infile,scratch,16);
        getDASInputData(das[i].infile,&das[i].din);
        das[i].dinlock = TRUE;
      }
    }
    if (mode_col >= 0)
    { /* monitor mode */
      strncpy(scratch,colptr(mode_col,str),128);
      if (strlen(scratch) > 0) 
      {
        das[i].mode = GetTransferMode(scratch);
        if ((das[i].mode & CM_TIMER) == 0)
        { /* the base mode MUST by TIMER: we need the callbacks coming in ! */
          das[i].mode &= ~(CM_SINGLE|CM_DATACHANGE|CM_EVENT|CM_REGISTER);
          das[i].mode |= CM_TIMER;
        }
      }
    }
    if (rate_col >= 0)
    { /* monitor mode */
      strncpy(scratch,colptr(rate_col,str),16);
      if (strlen(scratch) > 0) das[i].monitorRate = atoi(scratch);
    }
    if (time_col >= 0)
    { /* monitor mode */
      strncpy(scratch,colptr(time_col,str),16);
      if (strlen(scratch) > 0) das[i].monitorTime = atoi(scratch);
    }
    das[i].wait = 0;
    if (wait_col >= 0)
    {
      strncpy(scratch,colptr(wait_col,str),16);
      das[i].wait = atoi(scratch);
    }
    das[i].value = 0;
    if (value_col >= 0)
    {
      strncpy(scratch,colptr(value_col,str),16);
      das[i].value = atoi(scratch);
    }
    if (das[i].acc & SERIES) 
    {
      das[i].numberInSeries = das[i].value;
      das[i].acc &= ~(WAIT|READ);
      if (das[i].wait < 10) das[i].wait *= 1000; /* seconds given ? */
    }
    if (das[i].acc & MONITOR)
    {
      das[i].acc &= ~(WAIT|READ|SERIES);
      if (das[i].monitorTime == 0) das[i].monitorTime = DEFAULT_MONITOR_DURATION;
      if (das[i].monitorRate == 0) das[i].monitorRate = DEFAULT_MONITOR_RATE;
      if (das[i].mode == 0) das[i].mode = DEFAULT_MONITOR_MODE;
      if (BASEMODE(das[i].mode) == CM_SINGLE) das[i].monitorTime = 0;
      *isMonitoring = TRUE;
    }
    else
    {
      *isMonitoring = FALSE;
    }
    das[i].timeout = 0;
    if (to_col >= 0)
    {
      strncpy(scratch,colptr(to_col,str),16);
      das[i].timeout = atoi(scratch);
    }
    if (das[i].acc != MATCH && (j=findDAS(das,i)) == i)
    { /* first occurence */
      fsiz = fmtsizeof(das[i].dout.dFormat);
      dsiz = das[i].dout.dArrayLength;
      if (das[i].dout.dFormat == CF_STRUCT)
      {
        if ((ssiz=GetStructSize(das[i].dout.dTag)) > 0)
          fsiz = ssiz;
        else if (das[i].dout.dArrayLength == 1)
          fsiz = sizeof(DHS);
      }
      tsiz = dsiz * fsiz + getFormatHeaderSize(das[i].dout.dFormat);
      if (das[i].dout.dFormat == CF_IMAGE) tsiz += IMAGE_REF_SIZE;
      if ((tsiz) > PreallocatedDataBufferSize)
      {
        if ((das[i].dout.data.bptr=(BYTE *)calloc(1,tsiz)) == NULL)
          ccerr(out_of_local_memory);
        das[i].freeMemoryOnExit = TRUE;
      }
      else
      { 
         das[i].dout.data.bptr = PreallocatedDataBuffer;
      }
    }
    else 
    { /* found: pass along the data reference */
      das[i].dout.data.vptr = das[j].dout.data.vptr;
    }
    das[i].scale = 1.0;
    if (scale_col >= 0)
    {
      strncpy(scratch,colptr(scale_col,str),16);
      das[i].scale = (float)atof(scratch);
    }
    das[i].shift = 0;
    if (shift_col >= 0)
    {
      strncpy(scratch,colptr(shift_col,str),16);
      das[i].shift = (float)atof(scratch);
    }
    i++;
  }
  err:
  if (fp) fclose(fp);
  if (cc) printf("%s Data Access Structure : %s\n",fn,erlst[cc]);
  return cc ? -cc : i;
}

int readArchiveSets(char *eqm)
{
  char fn[256];
  int i,cc;

  if (PreallocatedDataBuffer == NULL)
  {
    if ((PreallocatedDataBuffer=(BYTE *)calloc(PreallocatedDataBufferSize,1)) == NULL)
    {
      feclog("cannot allocate %d bytes for data buffer!",PreallocatedDataBufferSize);
      return out_of_local_memory;
    }
  }
  if ((arbTblSiz=getArchiveSets(archiveListFile)) < 0) return -arbTblSiz;
  if ((arbTrgList = (NAME64 *)calloc(arbTblSiz, sizeof(NAME64))) == NULL) return -out_of_local_memory;
  strncpy(pmArchEqm, eqm, EQM_NAME_SIZE);
  getTriggerAlarms(alarmListFile,eqm); // usually there isn't one of these ...
  for (i=0; i<arbTblSiz; i++)
  {
    strncpy(arbTrgList[i].name, arbTbl[i].trg, 64);
    sprintf(fn,"%s%c%.64s%c%.64s.csv",
      stdArchiveDirectory,FS_DELIMITER,TRIGGER_FILES_SUBDIR,FS_DELIMITER,arbTbl[i].trg);
    if ((arbTbl[i].nsteps=getNumEntries(fn)) < 0)
    {
      feclog("readArchiveSet %.256s : %.32s\n",fn,erlst[-arbTbl[i].nsteps]);
      continue;
    }
    if ((arbTbl[i].das=(DAS *)calloc(arbTbl[i].nsteps,sizeof(DAS))) == NULL) 
    {
      feclog("cannot allocate %d data access steps for %.64s",arbTbl[i].nsteps,arbTbl[i].trg);
      return out_of_local_memory;
    }
    if ((cc=getDAS(fn,arbTbl[i].das,&arbTbl[i].isMonitoring)) < 0) return -cc;
    FillDeviceLists(&arbTbl[i]);
    FillPropertyAliasList(&arbTbl[i]);
  }
  return 0;
}
void initArchiveTask(void)
{
  if ((numArchiveTypes=getArchiveContracts()) > 0)
  {
    int i;
    for (i=0; i<numArchiveTypes; i++) dbglog("Archive contract %.16s",ArchRecord[i].ext);
  }
  if (archiveBuffer == NULL)
  {
    if ((archiveBuffer=(BYTE *)malloc(archiveBufferSize)) == NULL)
    {
      feclog("serious initialization error : cannot allocate %d bytes for common archive readback buffer",archiveBufferSize);
      exit(1);
    }
    feclog("allocated %d bytes for common archive readback buffer",archiveBufferSize);
  }

  RegisterUserCommand("ArchiveOutput",NULL,&archiveDebugOutput,NULL,CA_WRITE|CA_READ); 
  RegisterUserCommand("ArchiveDebug",NULL,&archDebug,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("sessions",cmdDumpSessions,NULL,NULL,CA_READ);
  RegisterUserCommand("freeSessions",cmdFreeSessions,NULL,NULL,CA_READ);
  
  return;
}
int GetTriggerTable(int index,ARBQS *arb,int arbBufSize)
{
  int i,n;

  for (n=0,i=index; n<arbBufSize && i<arbTblSiz; n++, i++)
  {
    strncpy(arb[n].trg,arbTbl[i].trg,DEVICE_NAME_SIZE);
    strncpy(arb[n].ext,arbTbl[i].ext,16);
    strncpy(arb[n].desc,arbTbl[i].desc,64);
    arb[n].recordEvents = arbTbl[i].recordEvents;
    arb[n].depthInMonths = arbTbl[i].depthInMonths;
    arb[n].maxQueueSize = arbTbl[i].maxq;
  }
  return n;
}

int arbHasSrv(ARB *arb,const char *ctx,const char *srv)
{
  int i;
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;

  if (arb->das == NULL) return FALSE;
  getWildcardTokens((char *)srv,beginToken,endToken,midToken);
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);
  if (strchr(srv,'*') == NULL) 
  { /* no wildcard: result the begin length in this case */
    blen = EXPORT_NAME_SIZE; 
  }
  for (i=0; i<arb->nsteps; i++)
  {
    if (ctx != NULL && strncmp(arb->das[i].ctx,ctx,CONTEXT_NAME_SIZE)) continue;
    if (!wildcardMatches(arb->das[i].srv,beginToken,blen,endToken,elen,midToken,mlen) &&
        !wildcardMatches(arb->das[i].srvArc,beginToken,blen,endToken,elen,midToken,mlen)) continue;
    return TRUE;
  }
  return FALSE;
}
int arbHasPrp(ARB *arb,const char *prp)
{
  int i;
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;

  if (arb->das == NULL) return FALSE;
  getWildcardTokens((char *)prp,beginToken,endToken,midToken);
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);
  if (strchr(prp,'*') == NULL) 
  { /* no wildcard: result the begin length in this case */
    blen = PROPERTY_NAME_SIZE; 
  }
  for (i=0; i<arb->nsteps; i++)
  {
    if (!wildcardMatches(arb->das[i].prp,beginToken,blen,endToken,elen,midToken,mlen) &&
        !wildcardMatches(arb->das[i].prpArc,beginToken,blen,endToken,elen,midToken,mlen)) continue;
    return TRUE;
  }
  return FALSE;
}
int GetTriggersForServer(const char *ctx,const char *srv,const char *prp,NAME64 *lst,int lstSize)
{
  int i, n=0, allSrvs=FALSE, allPrps=FALSE;

  if (srv == NULL || strlen(srv) == 0) allSrvs = TRUE;
  if (prp == NULL || strlen(prp) == 0) allPrps = TRUE;
  for (i=0; i<arbTblSiz && n<lstSize; i++)
  {
    if (!allSrvs && !arbHasSrv(&arbTbl[i],ctx,srv)) continue;
    if (!allPrps && !arbHasPrp(&arbTbl[i],prp)) continue;
    strncpy(lst[n].name,arbTbl[i].trg,64);
    n++;
  }
  return n;
}
int GetTriggerTableEntry(char *trg)
{
  int i;
  if (trg == NULL || strlen(trg) == 0) return -1;
  for (i=0; i<arbTblSiz; i++)
  {
    if (strnicmp(trg,arbTbl[i].trg,DEVICE_NAME_SIZE) == 0) return i;
  }
  return -1;
}
int AppendTriggerTable(ARBQS *arb,int size)
{
  int i,idx,maxq,cc = 0;
  FILE *fp = NULL;
  time_t t = time(NULL);
  struct tm *tms;
  char bckFile[256],fileDecor[16];
  static char lastFileDecor[16];

  if ((tms=localtime(&t)) != NULL)
  { /* backup current database if first change 'today' */
    sprintf(fileDecor,"-%02d.%02d.%02d",tms->tm_mday,tms->tm_mon+1,tms->tm_year%100);
    if (strncmp(fileDecor,lastFileDecor,16))
    {
      strncpy(bckFile,archiveListFile,128); bckFile[128] = 0;
      strcat(bckFile,fileDecor);
      strncpy(lastFileDecor,fileDecor,16);
      rename(archiveListFile,bckFile);
    }
  }
  for (i=0; i<size; i++)
  { /* scan for edits ...*/
    if ((idx=GetTriggerTableEntry(arb[i].trg)) >= 0)
    {
      strncpy(arbTbl[idx].ext,arb[i].ext,16);
      strncpy(arbTbl[idx].desc,arb[i].desc,64);
      arbTbl[idx].depthInMonths = arb[i].depthInMonths;
      arbTbl[idx].recordEvents = arb[i].recordEvents;
      arbTbl[idx].maxq = arb[i].maxQueueSize;
      memset(arb[i].trg,0,DEVICE_NAME_SIZE);
    }
  }
  if ((fp=fopen(archiveListFile,"w")) == NULL &&
      (fp=createAndOpen(archiveListFile,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Trigger, Extension, Record, Depth, Description, QueueSize\n");
  for (i=0; i<arbTblSiz; i++)
  { /* re-write the current list (with edits) */
    if (arbTbl[i].depthInMonths < 0) continue; /* deleted */
    fprintf(fp,"%.64s, %.16s, %d, %d, %.64s, %d\n",
      arbTbl[i].trg,arbTbl[i].ext,arbTbl[i].recordEvents,arbTbl[i].depthInMonths,arbTbl[i].desc,arbTbl[i].maxq);
  }
  for (i=0; i<size; i++)
  { /* append the new guys */
    if (arb[i].depthInMonths < 0) continue; /* deleted */
    if (arb[i].trg[0] == 0) continue; /* was an edit */
    if ((idx=GetTriggerTableEntry(arb[i].trg)) >= 0) maxq = arbTbl[idx].maxq; else maxq = 0;
    fprintf(fp,"%.64s, %.16s, %d, %d, %.64s, %d\n",
      arb[i].trg,arb[i].ext,arb[i].recordEvents,arb[i].depthInMonths,arb[i].desc,maxq);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int GetDASTable(int index,DASQS *das,int dasBufSize)
{
  int n;
  ARB *arb;

  if (index < 0 || index >= arbTblSiz) return -invalid_index;
  arb = &arbTbl[index];
  if (arb->das == NULL) return -code_failure;
  for (n=0; n<arb->nsteps && n<dasBufSize; n++)
  {
    das[n].access = arb->das[n].acc;
    das[n].datFmt = arb->das[n].dout.dFormat;
    das[n].datSiz = arb->das[n].dout.dArrayLength;
    das[n].datFmtIn = arb->das[n].din.dFormat;
    strncpy(das[n].datTag,arb->das[n].dout.dTag,TAG_NAME_SIZE);
    strncpy(das[n].datTagIn,arb->das[n].din.dTag,TAG_NAME_SIZE);
    das[n].mode = arb->das[n].mode;
    das[n].monitorRate = arb->das[n].monitorRate;
    das[n].monitorTime = arb->das[n].monitorTime;
    das[n].scale = arb->das[n].scale;
    das[n].shift = arb->das[n].shift;
    das[n].value = arb->das[n].value;
    das[n].timeout = arb->das[n].timeout;
    das[n].wait = arb->das[n].wait;
    strncpy(das[n].ctx,arb->das[n].ctx,CONTEXT_NAME_SIZE);
    strncpy(das[n].srv,arb->das[n].srv,EXPORT_NAME_SIZE);
    strncpy(das[n].srvArc,arb->das[n].srvArc,EXPORT_NAME_SIZE);
    strncpy(das[n].prp,arb->das[n].prp,PROPERTY_NAME_SIZE);
    strncpy(das[n].prpArc,arb->das[n].prpArc,PROPERTY_NAME_SIZE);
    strncpy(das[n].dev,arb->das[n].dev,DEVICE_NAME_SIZE);
    strncpy(das[n].infile,arb->das[n].infile,16);
  }
  return n;
}
char *getDasFormatString(int fmt,char *tag)
{
  static char fmtstr[64];
  memset(fmtstr,0,64);
  strncpy(fmtstr,GetFormatAsString(fmt),32);
  if (fmt == CF_STRUCT && tag != NULL && strlen(tag) > 0)
  {
    strcat(fmtstr,"/");
    strncat(fmtstr,tag,TAG_NAME_SIZE);
  }
  return fmtstr;
}
int validateDASEntry(DASQS *das)
{
  if (das == NULL) return invalid_parameter;
  if (das->datFmt == CF_NULL || das->datSiz == 0) return 0; // empty step or a wait ...
  if (strlen(das->ctx) == 0) return invalid_parameter;
  if (strlen(das->prp) == 0 || strlen(das->prpArc) == 0) return invalid_property;
  if (strlen(das->srv) == 0 || strlen(das->srvArc) == 0) return invalid_name;
  return 0;
}
int WriteDASTable(char *trg,DASQS *das,int size)
{
  char fn[256], *modstr;
  int i,cc=0;
  FILE *fp = NULL;

  if (trg == NULL || strlen(trg) == 0) return invalid_index;
  for (i=0; i<size; i++) if ((cc=validateDASEntry(&das[i])) != 0) return cc;
  sprintf(fn,"%s%c%s%c%.64s.csv",
    stdArchiveDirectory,FS_DELIMITER,TRIGGER_FILES_SUBDIR,FS_DELIMITER,trg);
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Context, Server, Property, Device, ArchiveServer, ArchiveProperty, Size, Format, ");
  fprintf(fp,"Access, Wait, Value, TimeOut, Scale, Shift, InputFile, InputFormat, ");
  fprintf(fp,"MonitorMode, MonitorRate, MonitorTime\n");

  for (i=0; i<size; i++)
  {
    fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.32s, %.64s, %d, %.32s, ",
      das[i].ctx,das[i].srv,das[i].prp,das[i].dev,das[i].srvArc,das[i].prpArc,das[i].datSiz,
      getDasFormatString(das[i].datFmt,das[i].datTag));
    fprintf(fp,"%.64s, %d, %d, %d, %g, %g, %.64s, %.32s, ",
      modtoacc((short)das[i].access),das[i].wait,das[i].value,das[i].timeout,das[i].scale,das[i].shift,das[i].infile,
      getDasFormatString(das[i].datFmtIn,das[i].datTagIn));
    modstr = das[i].mode != CM_CANCEL ? GetModeAsString(das[i].mode) : "";
    fprintf(fp,"%.64s, %d, %d\n",modstr,das[i].monitorRate,das[i].monitorTime);
  }

err:
  if (fp != NULL) fclose(fp);
  return cc;
}

typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char tag[TAG_NAME_SIZE];
  int siz;
  int fmt;
  float scale;
  float shift;
  BYTE *data;
  FILE *fp;
  int  linkid;
  int  retsiz;
  int  retries;
  int  isAcqHist;
  int  seriesIndex;
  int  numInSeries;
  int  starttime;
  int  lifetime;
  double lasttime;
  int lastsysstamp;
  int lastusrstamp;
  ARB *arb;
  DAS *das;
} DCS; /* Data Callback structure */
#define DCS_CAPACITY 100
typedef struct
{
  int numLinksPending;
  int retries;
  int status;
  DCS dcs[DCS_CAPACITY];
} DCSBLOCK; 

DCS dcs[DCS_CAPACITY];
int ndcs = DCS_CAPACITY;
int ndcs_pending;
int GetAndStoreCallbackError = 0;
DCS dmcs[DCS_CAPACITY];
int ndmcs = DCS_CAPACITY;

typedef struct
{
  ARB *arb;
  int step;
} ARBTriggerStep;

THREAD_TYPE GetAndStoreThreadTask(THREAD_PARAM lpvThreadParam)
{
  ARBTriggerStep *arbstep = (ARBTriggerStep *)lpvThreadParam;
  ARB *arb = arbstep->arb;
  arb->status = GetAndStore(arb, arbstep->step);
  SystemFree(arbstep);
  return 0;
}
THRHANDLE GetAndStoreCreateThread(ARB *arb, int step)
{
# if defined(UNIX)
  pthread_t tid = 0;
  pthread_attr_t gGSThreadAttr, *gPtrGSThreadAttr = &gGSThreadAttr;
# elif defined(WIN32)
  DWORD dwtid = 0;
  HANDLE tid;
# endif
  ARBTriggerStep *arbstep = NULL;
  if ((arbstep = (ARBTriggerStep *)SystemCalloc(1, sizeof(ARBTriggerStep))) == NULL)
  {
    dbglog("PANIC! Cannot create GetAndStoreThread!!!!");
    return 0;
  }
  arbstep->arb = arb;
  arbstep->step = step;
# if defined(UNIX)
  pthread_attr_init(gPtrGSThreadAttr);
  pthread_attr_setdetachstate(gPtrGSThreadAttr, PTHREAD_CREATE_DETACHED);
  pthread_create(&tid, gPtrGSThreadAttr, GetAndStoreThreadTask, (void *)arbstep);
# elif defined(WIN32)
  tid = CreateThread(NULL, 0, GetAndStoreThreadTask, (LPVOID)&arbstep, 0, &dwtid);
# else
  sendAlarmAsMail(ca->email, ca->ams->alarmTag, str);
# endif
  return (THRHANDLE)tid;
}

int getFreeDMCS(void)
{
  int i;
  static int done=FALSE;
  if (!done)
  {
    done = TRUE;
    for (i=0; i<ndmcs; i++) dmcs[i].linkid = -1;
  }
  for (i=0; i<ndmcs; i++) if (dmcs[i].linkid < 0) return i;
  return -1;
}

#define MAX_RETRY_COUNT 3

int writeNumInSeriesToFile(DCS *d)
{
  int cc = 0;
  SINT32 ival;
  char s[512];
  short sval;
  float fval;

  memset(s,0,CONTEXT_NAME_SIZE); strncpy(s,d->ctx,CONTEXT_NAME_SIZE);
  if (fwrite(s,CONTEXT_NAME_SIZE,1,d->fp) != 1) ccerr(file_error);
  memset(s,0,EXPORT_NAME_SIZE); strncpy(s,d->srv,EXPORT_NAME_SIZE);
  if (fwrite(s,EXPORT_NAME_SIZE,1,d->fp) != 1) ccerr(file_error);
  memset(s,0,PROPERTY_NAME_SIZE); strncpy(s,d->prp,PROPERTY_NAME_SIZE);
  sprintf(&s[strlen(s)],".NSEQ");
  if (fwrite(s,PROPERTY_NAME_SIZE,1,d->fp) != 1) ccerr(file_error);
  memset(s,0,DEVICE_NAME_SIZE); strncpy(s,d->dev,DEVICE_NAME_SIZE);
  if (fwrite(s,DEVICE_NAME_SIZE,1,d->fp) != 1) ccerr(file_error);
  memset(s,0,TAG_NAME_SIZE); strncpy(s,d->tag,TAG_NAME_SIZE);
  if (fwrite(s,TAG_NAME_SIZE,1,d->fp) != 1) ccerr(file_error);
  memset(s,0,STATUS_SHORTSIZE); strncpy(s,"0",STATUS_SHORTSIZE);
  if (fwrite(s,STATUS_SHORTSIZE,1,d->fp) != 1) ccerr(file_error);
  ival = 1; memcpy(&s[0],&ival,4);
  sval = CF_INT32; memcpy(&s[4],&sval,2);
  sval = 0; memcpy(&s[6],&sval,2);
  if (fwrite(s,8,1,d->fp) != 1) ccerr(file_error);
  fval = 1; memcpy(&s[0],&fval,4);
  fval = 0; memcpy(&s[4],&fval,4);
  if (fwrite(s,8,1,d->fp) != 1) ccerr(file_error);
  memcpy(&s[0],&d->lasttime,8);
  memcpy(&s[8],&d->lastsysstamp,4);
  memcpy(&s[12],&d->lastusrstamp,4);
  if (fwrite(s,16,1,d->fp) != 1) ccerr(file_error);
  ival = d->seriesIndex; /* n.b. NSEQ = the seriesIndex when we're done */
  if (fwrite(&ival,(size_t)1,(size_t)4,d->fp) != 4) ccerr(file_error);
err:
  return cc;
}
void MonitorAndStoreCallback(int id,int cc)
{ 
  int i=-1,fsiz, dsiz, ssiz, tsiz, hsiz, done=FALSE;
  SINT32 ival;
  short sval;
  float fval;
  double dval;
  char s[512],errstr[64];
  BYTE *bptr;
  time_t t = time(NULL);
  
  if (id < 0 || id >= ndmcs) ccerr(invalid_link);
  i = id;
  if (isTimeout(cc)) ccerr(cc);
  if (dmcs[i].data == NULL) ccerr(out_of_local_memory);
  if (t > dmcs[i].starttime + dmcs[i].lifetime)
  {
    done = TRUE;
  }
  memset(s,0,CONTEXT_NAME_SIZE); strncpy(s,dmcs[i].ctx,CONTEXT_NAME_SIZE);
  if (fwrite(s,CONTEXT_NAME_SIZE,1,dmcs[i].fp) != 1) ccerr(file_error);
  memset(s,0,EXPORT_NAME_SIZE); strncpy(s,dmcs[i].srv,EXPORT_NAME_SIZE);
  if (fwrite(s,EXPORT_NAME_SIZE,1,dmcs[i].fp) != 1) ccerr(file_error);
  memset(s,0,PROPERTY_NAME_SIZE); strncpy(s,dmcs[i].prp,PROPERTY_NAME_SIZE);
  sprintf(&s[strlen(s)],".%d",dmcs[i].seriesIndex);
  dmcs[i].seriesIndex++;
  if (fwrite(s,PROPERTY_NAME_SIZE,1,dmcs[i].fp) != 1) ccerr(file_error);
  memset(s,0,DEVICE_NAME_SIZE); strncpy(s,dmcs[i].dev,DEVICE_NAME_SIZE);
  if (fwrite(s,DEVICE_NAME_SIZE,1,dmcs[i].fp) != 1) ccerr(file_error);
  memset(s,0,TAG_NAME_SIZE); strncpy(s,dmcs[i].tag,TAG_NAME_SIZE);
  if (fwrite(s,TAG_NAME_SIZE,1,dmcs[i].fp) != 1) ccerr(file_error);
  memset(s,0,STATUS_SHORTSIZE); strncpy(s,GetCompletionStatus(id),STATUS_SHORTSIZE);
  if (fwrite(s,STATUS_SHORTSIZE,1,dmcs[i].fp) != 1) ccerr(file_error);
  dsiz = GetCompletionDataSize(dmcs[i].linkid); /* only store what came back */
  hsiz = getFormatHeaderSize(dmcs[i].fmt);
  ival = dsiz+hsiz; memcpy(&s[0],&ival,4);
  sval = dmcs[i].fmt; memcpy(&s[4],&sval,2);
  sval = cc; memcpy(&s[6],&sval,2);
  if (fwrite(s,8,1,dmcs[i].fp) != 1) ccerr(file_error);
  fval = dmcs[i].scale; memcpy(&s[0],&fval,4);
  fval = dmcs[i].shift; memcpy(&s[4],&fval,4);
  if (fwrite(s,8,1,dmcs[i].fp) != 1) ccerr(file_error);
  dmcs[i].lasttime = dval = GetCurrentDataTimeStamp(dmcs[i].linkid);
  memcpy(&s[0],&dval,8);
  dmcs[i].lastsysstamp = ival = GetSystemDataStamp(dmcs[i].linkid);
  memcpy(&s[8],&ival,4);
  dmcs[i].lastusrstamp = ival = GetDataStamp(dmcs[i].linkid);
  memcpy(&s[12],&ival,4);
  if (fwrite(s,16,1,dmcs[i].fp) != 1) ccerr(file_error);
  fsiz = fmtsizeof(dmcs[i].fmt);
  if (dmcs[i].fmt == CF_STRUCT)
  {
    if ((ssiz=GetStructSize(dmcs[i].tag)) > 0) dsiz *= ssiz;
    else if (dsiz == 1) dsiz = sizeof(DHS);
  }
  tsiz = fsiz * dsiz;
  if (tsiz == 0) goto err;
  if (hsiz > 0) 
  {
    bptr = getFormatBodyPtr(dmcs[i].data,dmcs[i].fmt);
  }
  if (bptr != NULL)
  {
    if (fwrite(dmcs[i].data,(size_t)1,(size_t)hsiz,dmcs[i].fp) != (unsigned int)(hsiz)) ccerr(file_error);
    if (fwrite(bptr,(size_t)1,(size_t)tsiz,dmcs[i].fp) != (unsigned int)(tsiz)) ccerr(file_error);
  }
  else
  {
    tsiz += hsiz;
    if (fwrite(dmcs[i].data,(size_t)1,(size_t)tsiz,dmcs[i].fp) != (unsigned int)(tsiz)) ccerr(file_error);
  }

err:
  if (cc)
  {
    done = TRUE;
    GetLastLinkError((short)cc,errstr);
    if (id >= 0 && id < ndmcs)
      AppendHistoryLog("archive: /%.32s/%.32s/%.64s %s :: <%d> %.32s",
        dmcs[i].ctx,dmcs[i].srv,dmcs[i].dev,dmcs[i].prp,cc,errstr); 
    else
      dbglog("MonitorAndStoreCallback: link %d : %.32s",id,errstr);
  }
  if (done && i >= 0)
  {
    CloseLink(dmcs[i].linkid);
    writeNumInSeriesToFile(&dmcs[i]);
    fclose(dmcs[i].fp);
    if (dmcs[i].arb != NULL)
    {
      dmcs[i].arb->fp = NULL;
      AppendHistoryLog("%s : close archive file",dmcs[i].arb->trg);
      dmcs[i].arb->step = 0;
      dmcs[i].arb->resultPending = FALSE;
    }
    dmcs[i].linkid = -1;
  }
}
int rewireAllDeviceName(char *tgt, DAS *das)
{
  if (das->devlst == NULL) return un_allocated;
  if (das->devlst->size < 2)
  {
    strcpy(tgt, "#0");
  }
  else
  {
    sprintf(tgt, "#0-#%d", das->devlst->size - 1);
  }
  return 0;
}
void cbMtGetAndStore(int id, int cc, void *ref)
{
  DCSBLOCK *dcsblk = (DCSBLOCK *)ref;
  DCS *dcs = dcsblk->dcs;
  int lid, fsiz, dsiz, ssiz, tsiz, hsiz;
  int storeNumInSeries = FALSE, no_context = FALSE;
  SINT32 ival;
  short sval;
  float fval;
  char s[512], errstr[64], lnkkey[256];
  BYTE *bptr = NULL;
  int dbg = (archDebug != 0 || tineDebug != 0) ? TRUE : FALSE;

  if (dcsblk->numLinksPending <= 0) return;
  if (cc)
  {
    if (isTimeout(cc) && dcsblk->retries < MAX_RETRY_COUNT) ccerr(cc);
  }
  if (dcs[id].data == NULL) ccerr(out_of_local_memory);
  lid = dcs[id].linkid;
  dsiz = GetCompletionDataSize(lid); /* only store what came back */
  dcs[id].retsiz = dsiz;
  if (dcs[id].isAcqHist)
  {
    dcs[id].seriesIndex++;
    dcs[id].das->seriesIndex = dcs[id].seriesIndex;
    if (dsiz == 0)
    {
      dcs[id].numInSeries = dcs[id].seriesIndex;
      dcs[id].lasttime = GetCurrentDataTimeStamp(lid);
      dcs[id].lastsysstamp = GetSystemDataStamp(lid);
      dcs[id].lastusrstamp = GetDataStamp(lid);
      writeNumInSeriesToFile(&dcs[id]);
      dcs[id].arb->sstStop = 0;
      dcs[id].arb->tsStop = 0;
      ccerr(cc);
    }
  }
  memset(s, 0, CONTEXT_NAME_SIZE); strncpy(s, dcs[id].ctx, CONTEXT_NAME_SIZE);
  if (fwrite(s, CONTEXT_NAME_SIZE, 1, dcs[id].fp) != 1) ccerr(file_error);
  memset(s, 0, EXPORT_NAME_SIZE); strncpy(s, dcs[id].srv, EXPORT_NAME_SIZE);
  if (fwrite(s, EXPORT_NAME_SIZE, 1, dcs[id].fp) != 1) ccerr(file_error);
  memset(s, 0, PROPERTY_NAME_SIZE); strncpy(s, dcs[id].prp, PROPERTY_NAME_SIZE);
  if (dcs[id].seriesIndex > 0)
  {
    sprintf(&s[strlen(s)], ".%d", dcs[id].seriesIndex);
    if (dcs[id].seriesIndex == dcs[id].numInSeries) storeNumInSeries = TRUE;
  }
  if (fwrite(s, PROPERTY_NAME_SIZE, 1, dcs[id].fp) != 1) ccerr(file_error);
  memset(s, 0, DEVICE_NAME_SIZE); strncpy(s, dcs[id].dev, DEVICE_NAME_SIZE);
  if (fwrite(s, DEVICE_NAME_SIZE, 1, dcs[id].fp) != 1) ccerr(file_error);
  memset(s, 0, TAG_NAME_SIZE); strncpy(s, dcs[id].tag, TAG_NAME_SIZE);
  if (fwrite(s, TAG_NAME_SIZE, 1, dcs[id].fp) != 1) ccerr(file_error);
  memset(s, 0, STATUS_SHORTSIZE); strncpy(s, GetCompletionStatus(lid), STATUS_SHORTSIZE);
  if (fwrite(s, STATUS_SHORTSIZE, 1, dcs[id].fp) != 1) ccerr(file_error);
  hsiz = getFormatHeaderSize(dcs[id].fmt);
  ival = dsiz + hsiz; memcpy(&s[0], &ival, 4);
  sval = dcs[id].fmt; memcpy(&s[4], &sval, 2);
  sval = cc; memcpy(&s[6], &sval, 2);
  if (fwrite(s, 8, 1, dcs[id].fp) != 1) ccerr(file_error);
  fval = dcs[id].scale; memcpy(&s[0], &fval, 4);
  fval = dcs[id].shift; memcpy(&s[4], &fval, 4);
  if (fwrite(s, 8, 1, dcs[id].fp) != 1) ccerr(file_error);
  dcs[id].lasttime = GetCurrentDataTimeStamp(lid);
  dcs[id].lastsysstamp = GetSystemDataStamp(lid);
  dcs[id].lastusrstamp = GetDataStamp(lid);
  memcpy(&s[0], &dcs[id].lasttime, 8);
  memcpy(&s[8], &dcs[id].lastsysstamp, 4);
  memcpy(&s[12], &dcs[id].lastusrstamp, 4);
  if (fwrite(s, 16, 1, dcs[id].fp) != 1) ccerr(file_error);
  fsiz = fmtsizeof(dcs[id].fmt);
  if (dcs[id].fmt == CF_STRUCT)
  {
    if ((ssiz = GetStructSize(dcs[id].tag)) > 0) dsiz *= ssiz;
    else if (dsiz == 1) dsiz = sizeof(DHS);
  }
  tsiz = fsiz * dsiz;
  if (tsiz == 0) goto err;
  if (hsiz > 0)
  {
    bptr = getFormatBodyPtr(dcs[id].data, dcs[id].fmt);
  }
  if (bptr != NULL)
  {
    if (fwrite(dcs[id].data, (size_t)1, (size_t)hsiz, dcs[id].fp) != (unsigned int)(hsiz)) ccerr(file_error);
    if (fwrite(bptr, (size_t)1, (size_t)tsiz, dcs[id].fp) != (unsigned int)(tsiz)) ccerr(file_error);
  }
  else
  {
    tsiz += hsiz;
    if (fwrite(dcs[id].data, (size_t)1, (size_t)tsiz, dcs[id].fp) != (unsigned int)(tsiz)) ccerr(file_error);
  }
  if (storeNumInSeries) cc = writeNumInSeriesToFile(&dcs[id]);
  if (dcs[id].arb->sstStart > 0 && dcs[id].lastsysstamp > 0)
  {
    dcs[id].arb->sstStart = dcs[id].lastsysstamp + 1;
  }
  else
  {
    dcs[id].arb->tsStart = dcs[id].lasttime + 0.01;
  }
err:
  if (cc != 0 || dbg) 
    sprintf(lnkkey,"/%.32s/%.32s/%.64s[%.64s]", dcs[id].ctx, dcs[id].srv, dcs[id].dev, dcs[id].prp);
  if (cc)
  {
    GetLastLinkError((short)cc, errstr);
    if (id < DCS_CAPACITY)
    {
      sprintf(s, "archive: %.192s", lnkkey);
    }
    else
    {
      sprintf(s, "archive: link id %d arrived out of context (%d links pending)", id, dcsblk->numLinksPending);
      no_context = TRUE;
      CloseLink(lid);
    }
    sprintf(&s[strlen(s)], " :: <%d> %.32s", cc, errstr);
    AppendHistoryLog(s);
    if (!no_context) dcsblk->status = cc; /* only set upon error */
  }
  dcsblk->numLinksPending--;
  if (dbg) dbglog("%.192s has notified, number links pending now %d",lnkkey, dcsblk->numLinksPending);
}
int mtGetAndStore(ARB *arb, int step)
{
  int dbg = (archDebug != 0 || tineDebug != 0) ? TRUE : FALSE;
  if (arb == NULL) return non_existent;
  if (arb->tid == 0) arb->status = not_running;
  switch (arb->status)
  {
  case not_running:
    arb->status = not_ready;
    arb->tid = GetAndStoreCreateThread(arb, step);
    if (dbg) dbglog("launch step %d from %.64s in its own thread: 0x%lx",
      arb->step, arb->trg, arb->tid);
  case not_ready:
    break;
  default:
    if (dbg) dbglog("step %d from %.64s has finished on thread 0x%lx",
      arb->step, arb->trg, arb->tid);
#     ifdef WIN32
    CloseHandle(arb->tid);
#     endif
    arb->tid = 0;
    break;
  }
  return arb->status;
}
#ifdef MEMORY_LANE
void GetAndStoreCallback(int id, int cc)
{
  int i, fsiz, dsiz, ssiz, tsiz, hsiz;
  int storeNumInSeries = FALSE, no_context = FALSE;
  SINT32 ival;
  short sval;
  float fval;
  char s[512], errstr[64];
  BYTE *bptr = NULL;

  for (i = 0; i < ndcs && dcs[i].linkid != id; i++);
  if (i == ndcs) ccerr(link_error);
  if (cc)
  {
    if (cc == link_timeout && dcs[i].retries < MAX_RETRY_COUNT) ccerr(cc);
    // if (dcs[i].isAcqHist) dcs[i].numInSeries = dcs[i].seriesIndex; << need this?
  }
  if (dcs[i].data == NULL) ccerr(out_of_local_memory);
  dsiz = GetCompletionDataSize(id); /* only store what came back */
  dcs[i].retsiz = dsiz;
  if (dcs[i].isAcqHist)
  {
    dcs[i].seriesIndex++;
    dcs[i].das->seriesIndex = dcs[i].seriesIndex;
    if (dsiz == 0)
    {
      dcs[i].numInSeries = dcs[i].seriesIndex;
      dcs[i].lasttime = GetCurrentDataTimeStamp(id);
      dcs[i].lastsysstamp = GetSystemDataStamp(id);
      dcs[i].lastusrstamp = GetDataStamp(id);
      writeNumInSeriesToFile(&dcs[i]);
      dcs[i].arb->sstStop = 0;
      dcs[i].arb->tsStop = 0;
      ccerr(cc);
    }
  }
  memset(s, 0, CONTEXT_NAME_SIZE); strncpy(s, dcs[i].ctx, CONTEXT_NAME_SIZE);
  if (fwrite(s, CONTEXT_NAME_SIZE, 1, dcs[i].fp) != 1) ccerr(file_error);
  memset(s, 0, EXPORT_NAME_SIZE); strncpy(s, dcs[i].srv, EXPORT_NAME_SIZE);
  if (fwrite(s, EXPORT_NAME_SIZE, 1, dcs[i].fp) != 1) ccerr(file_error);
  memset(s, 0, PROPERTY_NAME_SIZE); strncpy(s, dcs[i].prp, PROPERTY_NAME_SIZE);
  if (dcs[i].seriesIndex > 0)
  {
    sprintf(&s[strlen(s)], ".%d", dcs[i].seriesIndex);
    if (dcs[i].seriesIndex == dcs[i].numInSeries) storeNumInSeries = TRUE;
  }
  if (fwrite(s, PROPERTY_NAME_SIZE, 1, dcs[i].fp) != 1) ccerr(file_error);
  memset(s, 0, DEVICE_NAME_SIZE); strncpy(s, dcs[i].dev, DEVICE_NAME_SIZE);
  if (fwrite(s, DEVICE_NAME_SIZE, 1, dcs[i].fp) != 1) ccerr(file_error);
  memset(s, 0, TAG_NAME_SIZE); strncpy(s, dcs[i].tag, TAG_NAME_SIZE);
  if (fwrite(s, TAG_NAME_SIZE, 1, dcs[i].fp) != 1) ccerr(file_error);
  memset(s, 0, STATUS_SHORTSIZE); strncpy(s, GetCompletionStatus(id), STATUS_SHORTSIZE);
  if (fwrite(s, STATUS_SHORTSIZE, 1, dcs[i].fp) != 1) ccerr(file_error);
  hsiz = getFormatHeaderSize(dcs[i].fmt);
  ival = dsiz + hsiz; memcpy(&s[0], &ival, 4);
  sval = dcs[i].fmt; memcpy(&s[4], &sval, 2);
  sval = cc; memcpy(&s[6], &sval, 2);
  if (fwrite(s, 8, 1, dcs[i].fp) != 1) ccerr(file_error);
  fval = dcs[i].scale; memcpy(&s[0], &fval, 4);
  fval = dcs[i].shift; memcpy(&s[4], &fval, 4);
  if (fwrite(s, 8, 1, dcs[i].fp) != 1) ccerr(file_error);
  dcs[i].lasttime = GetCurrentDataTimeStamp(id);
  dcs[i].lastsysstamp = GetSystemDataStamp(id);
  dcs[i].lastusrstamp = GetDataStamp(id);
  memcpy(&s[0], &dcs[i].lasttime, 8);
  memcpy(&s[8], &dcs[i].lastsysstamp, 4);
  memcpy(&s[12], &dcs[i].lastusrstamp, 4);
  if (fwrite(s, 16, 1, dcs[i].fp) != 1) ccerr(file_error);
  fsiz = fmtsizeof(dcs[i].fmt);
  if (dcs[i].fmt == CF_STRUCT)
  {
    if ((ssiz = GetStructSize(dcs[i].tag)) > 0) dsiz *= ssiz;
    else if (dsiz == 1) dsiz = sizeof(DHS);
  }
  tsiz = fsiz * dsiz;
  if (tsiz == 0) goto err;
  if (hsiz > 0)
  {
    bptr = getFormatBodyPtr(dcs[i].data, dcs[i].fmt);
  }
  if (bptr != NULL)
  {
    if (fwrite(dcs[i].data, (size_t)1, (size_t)hsiz, dcs[i].fp) != (unsigned int)(hsiz)) ccerr(file_error);
    if (fwrite(bptr, (size_t)1, (size_t)tsiz, dcs[i].fp) != (unsigned int)(tsiz)) ccerr(file_error);
  }
  else
  {
    tsiz += hsiz;
    if (fwrite(dcs[i].data, (size_t)1, (size_t)tsiz, dcs[i].fp) != (unsigned int)(tsiz)) ccerr(file_error);
  }
  if (storeNumInSeries) cc = writeNumInSeriesToFile(&dcs[i]);
  if (dcs[i].arb->sstStart > 0 && dcs[i].lastsysstamp > 0)
  {
    dcs[i].arb->sstStart = dcs[i].lastsysstamp + 1;
  }
  else
  {
    dcs[i].arb->tsStart = dcs[i].lasttime + 0.01;
  }
err:
  if (cc)
  {
    GetLastLinkError((short)cc, errstr);
    if (i < ndcs)
    {
      sprintf(s, "archive: /%.32s/%.32s/%.64s %s", dcs[i].ctx, dcs[i].srv, dcs[i].dev, dcs[i].prp);
    }
    else
    {
      sprintf(s, "archive: link id %d arrived out of context (%d links pending)", id, ndcs_pending);
      no_context = TRUE;
      CloseLink(id);
    }
    sprintf(&s[strlen(s)], " :: <%d> %.32s", cc, errstr);
    AppendHistoryLog(s);
    if (!no_context) GetAndStoreCallbackError = cc;
  }
  if (ndcs_pending && !no_context) ndcs_pending--;
}
#define WaitForArchiveLinks(n) {\
  ndcs_pending = n;\
  for (mwCnts=100; ndcs_pending && mwCnts > 0; mwCnts--) SystemDelay(100);\
  n = 0;\
}
int stGetAndStore(ARB *arb,int step)
{
  DAS *das = &arb->das[step];
  FILE *fp = arb->fp;
  int id,cc = 0,first = 1,retries = 0, mwCnts;
  char *substr[5],*c;
  char tmpname[132],devname[DEVICE_NAME_SIZE+8],fulldevname[132],errstr[64];
  DTYPE dout,din,*pdin = NULL;
  int n,lnk = 0,pstart[4],pstop[4],pnow[4],i[4],depth,numlinks;
  int blocksize, ssize, mode = CM_SINGLE|CM_SYNCNOTIFY;
  UINT32 stst[2];
  time_t tentry = time(NULL);

  sprintf(fulldevname, "/%.32s/%.32s/%.64s", das->ctx, das->srv, das->dev); // start it off
  if (archiveDebugOutput)
  {
    AppendHistoryLog("GetAndStore: /%.32s/%.32s/%.64s %s",das->ctx,das->srv,das->dev,das->prp);
  }
  memset(pstart,0,sizeof(pstart)); memset(pstop,0,sizeof(pstop));
  ndcs_pending = 0;
  
  memset(tmpname,0,DEVICE_NAME_SIZE);
  strncpy(tmpname,das->dev,DEVICE_NAME_SIZE);  /* copy to tmpname area for manipulation */
  if (!strcmp(tmpname,DEVICE_ALL))
  { /* prepare new tmpname a la the #0-#100 scheme */
    if (rewireAllDeviceName(tmpname,das) != 0) ccerr(un_allocated);
  }
  if (!strcmp(tmpname,DEVICE_IGNORE))
  { /* fix to "#0" */
    strcpy(tmpname,"#0");
  }
  if ((substr[0]=strtok(tmpname,".")) == NULL) substr[0] = tmpname;
  for (depth=0; substr[depth] != NULL && depth<4; depth++,substr[depth]=strtok(NULL,"."))
  {
    pstart[depth] = pstop[depth] = 0;
    if ((c=strchr(substr[depth],'#')) != NULL) pstart[depth] = atoi(&c[1]);
    if ((c=strrchr(substr[depth],'#')) != NULL) pstop[depth] = atoi(&c[1]);
  }

  if ((blocksize=das->dout.dArrayLength*fmtsizeof(das->dout.dFormat)) <= 0) ccerr(illegal_read_write);
  if (das->dout.dFormat == CF_STRUCT)
  {
    blocksize *= GetStructSize(das->dout.dTag);
    if (blocksize == 0) 
    { /* can only be DHS if this legacy thing is untagged */
      strcpy(das->dout.dTag,"DHS");
      blocksize = sizeof(DHS);
    }
  }
  blocksize += getFormatHeaderSize(das->dout.dFormat);
  ndcs = blocksize > UDP_BUFFER_SIZE ? 1 : UDP_BUFFER_SIZE/blocksize;
  if (ndcs > DCS_CAPACITY) ndcs = DCS_CAPACITY;
  memset(dcs,0,sizeof(DCS)*DCS_CAPACITY); /* clear everything */
top:
  cc = GetAndStoreCallbackError = 0;

  pnow[0] = pstart[0]; pnow[1] = pstart[1]; 
  pnow[2] = pstart[2]; pnow[3] = pstart[3];
  numlinks = 0;

  for (i[0]=pstart[0]; i[0] <= pstop[0]; i[0]++)
    for (i[1]=pstart[1]; i[1] <= pstop[1]; i[1]++)
      for (i[2]=pstart[2]; i[2] <= pstop[2]; i[2]++)
        for (i[3]=pstart[3]; i[3] <= pstop[3]; i[3]++)
  {
    if (strchr(tmpname,'#') == NULL) 
    { /* not using device numbers (normal case) -> take the device name as is */
      strncpy(devname,das->dev,DEVICE_NAME_SIZE);
    }
    else 
    { /* using device numbers and there could be a hierarchy */
      for (n=0,devname[0]=0; n<depth; n++)
        sprintf(&devname[strlen(devname)],n < depth-1 ? "#%d." : "#%d",i[n]);
    }
    if (depth > 1)
    { /* I don't think this nonsense shows up any more (from the HERA era) */
      for (n=0,devname[0]=0; n<depth; n++)
      { /* applies only with a device hierarchy (e.g. #1.#1-#100.#10, etc.) */
        if (substr[n][0] == '#')
          sprintf(&devname[strlen(devname)], n < depth-1 ? "#%d." : "#%d",i[n]);
        else
          sprintf(&devname[strlen(devname)], n < depth-1 ? "%s." : "%s",substr[n]);
      }
    }
    sprintf(fulldevname,"/%.32s/%.32s/%.64s",das->ctx,das->srv,devname);

    memcpy(&dout,&das->dout,sizeof(DTYPE)); /* copy configured 'dout' into the call */
    dout.data.bptr += numlinks*blocksize; /* move pointer along for multiple calls */

    if (das->numberToAcquire)
    { /* only > 0 if SETNUM is in play ! */
      if (das->numberToAcquire < (int)dout.dArrayLength) dout.dArrayLength = das->numberToAcquire;
      blocksize = dout.dArrayLength*fmtsizeof(dout.dFormat) + getFormatHeaderSize(dout.dFormat);
      if (dout.dFormat == CF_STRUCT)
      {
        strncpy(dout.dTag,das->dout.dTag,TAG_NAME_SIZE);
        if ((ssize=GetStructSize(dout.dTag)) == 0)
        { /* untagged !!! */
          dout.dArrayLength = das->dout.dArrayLength * das->numberToAcquire;
          ssize = 1;
        }
        blocksize = ssize * dout.dArrayLength;
      }
      if (blocksize > PreallocatedDataBufferSize) 
      {
        AppendHistoryLog("archive: %.132s %.64s :: number to acquire %d requires %d bytes which exceeds the pre-allocated buffer size of %d bytes",
          fulldevname,das->prp,das->numberToAcquire,blocksize,PreallocatedDataBufferSize);
        lastLnkErrSrc = LNK_ERROR_LCL;
        ccerr(buffer_too_small);
      }
    }
    if (dout.dFormat == CF_IMAGE)
    {
      ((DIMAGE *)dout.data.bptr)->frameBuffer = das->dout.data.bptr + IMAGE_REF_SIZE + IMAGE_HDR_SIZE;
      ((DIMAGE *)dout.data.bptr)->frameBufferSize = das->dout.dArrayLength;
    }
    
    if (das->din.dArrayLength > 0) pdin = &das->din;
    else if (strstr(das->prp,".HIST") != NULL)
    {
      stst[1] = (UINT32)tentry;
      stst[0] = stst[1] - dout.dArrayLength;
      din.dFormat = CF_INT32;
      din.dArrayLength = 2;
      din.data.lptr = (SINT32 *)&stst;
      pdin = &din;
    }
    else pdin = NULL;
    
    dcs[lnk].seriesIndex = das->numberInSeries > 0 ? das->seriesIndex : 0;
    dcs[lnk].numInSeries = das->numberInSeries;
    dcs[lnk].isAcqHist = das->isAcqHist;
    strncpy(dcs[lnk].ctx,das->ctx,CONTEXT_NAME_SIZE);
    strncpy(dcs[lnk].srv,das->srvArc,EXPORT_NAME_SIZE);
    strncpy(dcs[lnk].prp,das->prpArc,PROPERTY_NAME_SIZE);
    strncpy(dcs[lnk].dev,devname,DEVICE_NAME_SIZE);
    strncpy(dcs[lnk].tag,das->dout.dTag,TAG_NAME_SIZE);
    dcs[lnk].siz = (int)das->dout.dArrayLength;
    if (das->dout.dFormat == CF_STRUCT && 
        dcs[lnk].siz == 1 &&
        das->dout.dTag[0] == 0) dcs[lnk].siz = sizeof(DHS);
    dcs[lnk].fmt = das->dout.dFormat;
    dcs[lnk].scale = das->scale;
    dcs[lnk].shift = das->shift;
    dcs[lnk].data = dout.data.bptr;
    dcs[lnk].fp = fp;
    dcs[lnk].arb = arb;
    dcs[lnk].das = das;

    if (useConnectedLinks && (int)das->dout.dArrayLength > UDP_BUFFER_SIZE)
    {
      mode |= CM_CONNECT;
    }
    if ((id=AttachLink(fulldevname,das->prp,&dout,pdin,CA_READ,
            3000,GetAndStoreCallback,mode)) < 0) ccerr(-id);

    dcs[lnk].linkid = id;
    numlinks++;

    if (++lnk >= ndcs)
    { /* passed the number of links in a block ! */
      WaitForArchiveLinks(lnk); /* resets lnk to 0 */
      if (GetAndStoreCallbackError == link_timeout)
      {
        if (dcs[lnk].retries < MAX_RETRY_COUNT)
        {
          if (archiveDebugOutput) AppendHistoryLog("%.132s : GetAndStore: trying again ...",fulldevname);
          dcs[lnk].retries++;
          if (first) goto top;
          i[0] = pnow[0]; i[1] = pnow[1]; i[2] = pnow[2]; i[3] = pnow[3];
        }
        else
        {
          AppendHistoryLog("%.132s : Assuming server down",fulldevname);
          return link_timeout;
        }
      }
      else
      {
        pnow[0] = i[0]; pnow[1] = i[1]; pnow[2] = i[2]; pnow[3] = i[3];
        dcs[lnk].retries = 0;
        cc = GetAndStoreCallbackError;
      }
      first = 0;
      GetAndStoreCallbackError = 0;
      numlinks = 0;
    }
  }
  WaitForArchiveLinks(lnk); /* remaining links */
  if (GetAndStoreCallbackError == link_timeout && dcs[lnk].retries < MAX_RETRY_COUNT)
  {
    if (archiveDebugOutput)
    {
      AppendHistoryLog("%s : GetAndStore: trying again ...",fulldevname);
    }
    if (dcs[lnk].retries == 0)
    {
      pstart[0] = pnow[0]; pstart[1] = pnow[1]; 
      pstart[2] = pnow[2]; pstart[3] = pnow[3];
      if (pstart[3] < pstop[3]) pstart[3]++;
      else if (pstart[2] < pstop[2]) pstart[2]++;
      else if (pstart[1] < pstop[1]) pstart[1]++;
      else if (pstart[0] < pstop[0]) pstart[0]++;
    }
    dcs[lnk].retries++;
    goto top;
  }
  cc = GetAndStoreCallbackError;
err:
  if (cc)
  { /* link error (even after retries) */
    if (archiveDebugOutput)
    {
      AppendHistoryLog("archive: %.132s %.64s :: %.32s",fulldevname,das->prp,GetLastLinkError((short)cc,errstr));
    }
    das->numberToAcquire = 0;
    if (das->acc&SETDEV && das->devTgtLst != NULL) das->devTgtLst->siz = 0;
  }
  else
  { /* all went well */
    if (das->acc&SETNUM) 
    { /* special flag to signal the next call ... */
      switch (dcs[0].fmt)
      {
        case CF_INT16:       
          das->numberToAcquire = *(UINT16 *)dcs[0].data;
          break;
        case CF_INT32:       
          das->numberToAcquire = *(UINT32 *)dcs[0].data;
          break;
        default:
          AppendHistoryLog("%.132s %.64s get number to acquire has irregular format %d !!!",fulldevname,das->prp,dcs[0].fmt);
          das->numberToAcquire = 1;
      }
      if (archiveDebugOutput)
      {
        AppendHistoryLog("%.132s %.64s number to acquire %d",fulldevname,das->prp,das->numberToAcquire);
      }
    }
    else 
    {
      if ((das->acc&PERSIST) == 0) das->numberToAcquire = 0;
    }
    if (das->acc&SETDEV && das->devTgtLst != NULL)
    { /* special flag to make use of the acquired device list in subsequent calls */
      das->devTgtLst->idx = 0;
      if (dcs[0].retsiz < MAX_DEVTGTLST_SIZE)
      { /* the usual case ... */
        dout.dArrayLength = das->devTgtLst->siz = dcs[0].retsiz;
        GetValuesAsStringEx(&dout,das->devTgtLst->lst[0].name,CF_NAME64,MAX_DEVTGTLST_SIZE,0);
      }
      else
      { /* so many?  then get them all */
        das->devTgtLst->siz = 0;
        strncpy(das->dev,DEVICE_ALL,DEVICE_NAME_SIZE);
      }
    }
  }  
  return cc;
}
#endif
int GetAndStore(ARB *arb, int step)
{
  FILE *fp = arb->fp;
  DAS *das = &arb->das[step];
  int id, cc = 0, first = TRUE;
  char *c, tmpname[132], devname[DEVICE_NAME_SIZE + 8], fulldevname[132], errstr[64];
  DTYPE dout, din, *pdin = NULL;
  DCSBLOCK *dcsblk = NULL;
  int lnk=0, pstart=0, pstop=0, pnow, i, numlinks, ndcs;
  int blocksize, ssize, mode = CM_SINGLE | CM_SYNCNOTIFY;
  int dbg = (archDebug != 0 || tineDebug != 0) ? TRUE : FALSE;
  UINT32 stst[2];
  time_t tentry = time(NULL);

  sprintf(fulldevname, "/%.32s/%.32s/%.64s", das->ctx, das->srv, das->dev); // start it off
  if (archiveDebugOutput)
  {
    AppendHistoryLog("GetAndStore: %.128s[%.64s]", fulldevname, das->prp);
  }
  if (dbg) dbglog("GetAndStore: %.128s[%.64s], step %d", fulldevname, das->prp, step);
  memset(tmpname, 0, 132);
  strncpy(tmpname, das->dev, DEVICE_NAME_SIZE);  /* copy to tmpname area for manipulation */
  if (!strcmp(tmpname, DEVICE_ALL))
  { /* prepare new tmpname a la the #0-#100 scheme */
    if (rewireAllDeviceName(tmpname, das) != 0) ccerr(un_allocated);
  }
  if (!strcmp(tmpname, DEVICE_IGNORE))
  { /* fix to "#0" */
    strcpy(tmpname, "#0");
  }
  pstart = pstop = 0;
  if ((c=strchr(tmpname, '#')) != NULL) pstart = atoi(&c[1]);
  if ((c=strrchr(tmpname, '#')) != NULL) pstop = atoi(&c[1]);

  if ((blocksize=das->dout.dArrayLength * fmtsizeof(das->dout.dFormat)) <= 0) ccerr(illegal_read_write);
  if (das->dout.dFormat == CF_STRUCT)
  {
    blocksize *= GetStructSize(das->dout.dTag);
    if (blocksize == 0)
    { /* can only be DHS if this legacy thing is untagged */
      strcpy(das->dout.dTag, "DHS");
      blocksize = sizeof(DHS);
    }
  }
  blocksize += getFormatHeaderSize(das->dout.dFormat);
  ndcs = blocksize > UDP_BUFFER_SIZE ? 1 : UDP_BUFFER_SIZE / blocksize;
  if (ndcs > DCS_CAPACITY) ndcs = DCS_CAPACITY;
  if ((dcsblk=(DCSBLOCK *)SystemCalloc(1, sizeof(DCSBLOCK))) == NULL)
  {
    AppendHistoryLog("GetAndStore: %.128s[%.64s] out of memory!", fulldevname, das->prp);
    ccerr(out_of_local_memory);
  }
top:
  cc = dcsblk->status = 0; /* fresh start ! */
  dcsblk->numLinksPending = 0;
  pnow = pstart;
  numlinks = 0;
  lnk = 0;
  for (i=pstart; i <= pstop; i++)
  {
    if (strchr(tmpname, '#') == NULL)
    { /* not using device numbers (normal case) -> take the device name as is */
      strncpy(devname, das->dev, DEVICE_NAME_SIZE);
    }
    else
    {
      sprintf(devname, "#%d", i);
    }
    sprintf(fulldevname, "/%.32s/%.32s/%.64s", das->ctx, das->srv, devname);
    memcpy(&dout, &das->dout, sizeof(DTYPE)); /* copy configured 'dout' into the call */
    dout.data.bptr += numlinks * blocksize; /* move pointer along for multiple calls */
    if (das->numberToAcquire)
    { /* only > 0 if SETNUM is in play ! */
      if (das->numberToAcquire < (int)dout.dArrayLength) dout.dArrayLength = das->numberToAcquire;
      blocksize = dout.dArrayLength * fmtsizeof(dout.dFormat) + getFormatHeaderSize(dout.dFormat);
      if (dout.dFormat == CF_STRUCT)
      {
        strncpy(dout.dTag, das->dout.dTag, TAG_NAME_SIZE);
        if ((ssize = GetStructSize(dout.dTag)) == 0)
        { /* untagged !!! */
          dout.dArrayLength = das->dout.dArrayLength * das->numberToAcquire;
          ssize = 1;
        }
        blocksize = ssize * dout.dArrayLength;
      }
      if (blocksize > PreallocatedDataBufferSize)
      {
        AppendHistoryLog("archive: %.132s %.64s :: number to acquire %d requires %d bytes which exceeds the pre-allocated buffer size of %d bytes",
          fulldevname, das->prp, das->numberToAcquire, blocksize, PreallocatedDataBufferSize);
        lastLnkErrSrc = LNK_ERROR_LCL;
        ccerr(buffer_too_small);
      }
    }
    if (dout.dFormat == CF_IMAGE)
    {
      ((DIMAGE *)dout.data.bptr)->frameBuffer = das->dout.data.bptr + IMAGE_REF_SIZE + IMAGE_HDR_SIZE;
      ((DIMAGE *)dout.data.bptr)->frameBufferSize = das->dout.dArrayLength;
    }

    if (das->din.dArrayLength > 0) pdin = &das->din;
    else if (strstr(das->prp, ".HIST") != NULL)
    {
      stst[1] = (UINT32)tentry;
      stst[0] = stst[1] - dout.dArrayLength;
      din.dFormat = CF_INT32;
      din.dArrayLength = 2;
      din.data.lptr = (SINT32 *)&stst;
      pdin = &din;
    }
    else pdin = NULL;

    dcsblk->numLinksPending++;
    dcsblk->dcs[lnk].seriesIndex = das->numberInSeries > 0 ? das->seriesIndex : 0;
    dcsblk->dcs[lnk].numInSeries = das->numberInSeries;
    dcsblk->dcs[lnk].isAcqHist = das->isAcqHist;
    strncpy(dcsblk->dcs[lnk].ctx, das->ctx, CONTEXT_NAME_SIZE);
    strncpy(dcsblk->dcs[lnk].srv, das->srvArc, EXPORT_NAME_SIZE);
    strncpy(dcsblk->dcs[lnk].prp, das->prpArc, PROPERTY_NAME_SIZE);
    strncpy(dcsblk->dcs[lnk].dev, devname, DEVICE_NAME_SIZE);
    strncpy(dcsblk->dcs[lnk].tag, das->dout.dTag, TAG_NAME_SIZE);
    dcsblk->dcs[lnk].siz = (int)das->dout.dArrayLength;
    if (das->dout.dFormat == CF_STRUCT &&
      dcsblk->dcs[lnk].siz == 1 &&
      das->dout.dTag[0] == 0) dcsblk->dcs[lnk].siz = sizeof(DHS);
    dcsblk->dcs[lnk].fmt = das->dout.dFormat;
    dcsblk->dcs[lnk].scale = das->scale;
    dcsblk->dcs[lnk].shift = das->shift;
    dcsblk->dcs[lnk].data = dout.data.bptr;
    dcsblk->dcs[lnk].fp = fp;
    dcsblk->dcs[lnk].arb = arb;
    dcsblk->dcs[lnk].das = das;

    if (useConnectedLinks && (int)das->dout.dArrayLength > UDP_BUFFER_SIZE)
    {
      mode |= CM_CONNECT;
    }
    dcsblk->dcs[lnk].linkid = -1;
    if ((id=AttachLinkEx2(fulldevname, das->prp, &dout, pdin, CA_READ,
      3000, cbMtGetAndStore, mode, lnk, (void *)dcsblk)) < 0)
    {
      if (dbg) dbglog("cannot link to %.128s[%.64s] : %.32s", 
        fulldevname, das->prp, cc2str(-id));
      //dcsblk->numLinksPending--;
      //dcsblk->status = -id;
      //continue;
      //ccerr(-id);
    }
    dcsblk->dcs[lnk].linkid = id;
    if (dbg)
      dbglog("link %d %.128s[%.64s] submitted (link id = %d); number links pending now %d", 
        lnk,fulldevname, das->prp, id, dcsblk->numLinksPending);
    numlinks++;

    if (++lnk >= ndcs)
    { /* passed the number of links in a block ! */
      if (dbg && ndcs > 1) dbglog("packed link threashold %d reached (pause and refresh)",ndcs);
      while (dcsblk->numLinksPending > 0)
      {
        millisleep(1);
        if (time(NULL) > tentry + 60)
        { // after 1 minute bail out ...
          AppendHistoryLog("%.132s : %d links still pending after 1 minute", fulldevname, dcsblk->numLinksPending);
          ccerr(not_posted);
        }
      }
      if (isTimeout(dcsblk->status))
      {
        if (dcsblk->retries < MAX_RETRY_COUNT)
        {
          if (archiveDebugOutput) AppendHistoryLog("%.132s : GetAndStore: trying again ...", fulldevname);
          dcsblk->retries++;
          if (first) goto top;
          i = i;
        }
        else
        {
          AppendHistoryLog("%.132s : Assuming server down", fulldevname);
          cc = link_timeout;
          i = pstop;
        }
      }
      else
      {
        pnow = i; 
        dcsblk->retries = 0;
        cc = dcsblk->status;
      }
      first = FALSE;
      dcsblk->status = 0;
      lnk = numlinks = 0;
      pstart = i;
    }
  }
  while (dcsblk->numLinksPending > 0)
  { /* any remaining links ? */
    millisleep(1);
    if (time(NULL) > tentry + 60)
    { // after 1 minute bail out ...
      AppendHistoryLog("%.132s : %d links still pending after 1 minute", fulldevname, dcsblk->numLinksPending);
      ccerr(not_posted);
    }
  }
  if (isTimeout(dcsblk->status) && dcsblk->retries < MAX_RETRY_COUNT)
  {
    if (archiveDebugOutput)
    {
      AppendHistoryLog("%s : GetAndStore: trying again ...", fulldevname);
    }
    if (dcsblk->retries == 0)
    {
      pstart = pnow; 
      if (pstart < pstop) pstart++;
    }
    dcsblk->retries++;
    goto top;
  }
  cc = dcsblk->status;
err:
  if (cc)
  { /* link error (even after retries) */
    if (archiveDebugOutput)
    {
      AppendHistoryLog("archive: %.132s %.64s :: %.32s", fulldevname, das->prp, GetLastLinkError((short)cc, errstr));
    }
    das->numberToAcquire = 0;
    if (das->acc & SETDEV && das->devTgtLst != NULL) das->devTgtLst->siz = 0;
  }
  else
  { /* all went well */
    if (das->acc & SETNUM)
    { /* special flag to signal the next call ... */
      switch (dcsblk->dcs[0].fmt)
      {
      case CF_INT16:
        das->numberToAcquire = *(UINT16 *)dcsblk->dcs[0].data;
        break;
      case CF_INT32:
        das->numberToAcquire = *(UINT32 *)dcsblk->dcs[0].data;
        break;
      default:
        AppendHistoryLog("%.132s %.64s get number to acquire has irregular format %d !!!", fulldevname, das->prp, dcsblk->dcs[0].fmt);
        das->numberToAcquire = 1;
      }
      if (archiveDebugOutput)
      {
        AppendHistoryLog("%.132s %.64s number to acquire %d", fulldevname, das->prp, das->numberToAcquire);
      }
    }
    else
    {
      if ((das->acc & PERSIST) == 0) das->numberToAcquire = 0;
    }
    if (das->acc & SETDEV && das->devTgtLst != NULL)
    { /* special flag to make use of the acquired device list in subsequent calls */
      das->devTgtLst->idx = 0;
      if (dcsblk->dcs[0].retsiz < MAX_DEVTGTLST_SIZE)
      { /* the usual case ... */
        dout.dArrayLength = das->devTgtLst->siz = dcsblk->dcs[0].retsiz;
        GetValuesAsStringEx(&dout, das->devTgtLst->lst[0].name, CF_NAME64, MAX_DEVTGTLST_SIZE, 0);
      }
      else
      { /* so many?  then get them all */
        das->devTgtLst->siz = 0;
        strncpy(das->dev, DEVICE_ALL, DEVICE_NAME_SIZE);
      }
    }
  }
  SystemFree(dcsblk);
  return cc;
}
int MonitorAndStore(ARB *arb,DAS *das)
{
  int id,cc = 0;
  char fulldevname[132],errstr[64];
  DTYPE dout,*pdin = NULL;
  int lnk = 0;
  int dsize;
  FILE *fp = arb->fp;

  sprintf(fulldevname,"/%.32s/%.32s/%.64s",das->ctx,das->srv,das->dev);
  if (archiveDebugOutput)
  {
    AppendHistoryLog("MonitorAndStore: %.132s %.64s",fulldevname,das->prp);
  }
  
  if ((dsize=das->dout.dArrayLength*fmtsizeof(das->dout.dFormat)) <= 0) ccerr(illegal_read_write);

  /* get an dmcs entry */
  if ((lnk=getFreeDMCS()) < 0)
  {
    AppendHistoryLog("MonitorAndStore: link table full!");
    return resources_exhausted;
  }
  
  memcpy(&dout,&das->dout,sizeof(DTYPE));

  if (dout.dFormat == CF_IMAGE)
  {
    ((DIMAGE *)dout.data.bptr)->frameBuffer = dout.data.bptr + IMAGE_REF_SIZE + IMAGE_HDR_SIZE;
    ((DIMAGE *)dout.data.bptr)->frameBufferSize = das->dout.dArrayLength;
  }
  
  dmcs[lnk].seriesIndex = 1;
  dmcs[lnk].numInSeries = das->numberInSeries;
  strncpy(dmcs[lnk].ctx,das->ctx,CONTEXT_NAME_SIZE);
  strncpy(dmcs[lnk].srv,das->srvArc,EXPORT_NAME_SIZE);
  strncpy(dmcs[lnk].prp,das->prpArc,PROPERTY_NAME_SIZE);
  strncpy(dmcs[lnk].dev,das->dev,DEVICE_NAME_SIZE);
  strncpy(dmcs[lnk].tag,das->dout.dTag,TAG_NAME_SIZE);
  dmcs[lnk].siz = (int)dout.dArrayLength;
  dmcs[lnk].fmt = dout.dFormat;
  dmcs[lnk].scale = das->scale;
  dmcs[lnk].shift = das->shift;
  dmcs[lnk].data = dout.data.bptr;
  dmcs[lnk].fp = fp;
  dmcs[lnk].lifetime = das->monitorTime;
  dmcs[lnk].starttime = (int)time(NULL);
  dmcs[lnk].arb = arb;

  if ((id=AttachLinkEx(fulldevname,das->prp,&dout,pdin,CA_READ,
          das->monitorRate,MonitorAndStoreCallback,das->mode,lnk)) < 0) ccerr(-id);
  dmcs[lnk].linkid = id;

err:
  if (cc)
  {
    if (archiveDebugOutput)
    {
      AppendHistoryLog("archive: %.132s %.64s :: %.32s",fulldevname,das->prp,GetLastLinkError((short)cc,errstr));
    }
    das->numberToAcquire = 0;
  }
  return cc;
}

int PollTarget(DAS *das)
{
  int val,cc;
  char devname[256];
  sprintf(devname,"/%.32s/%.32s/%.64s",das->ctx,das->srv,das->dev);
  if ((cc=ExecLink(devname,das->prp,&das->dout,NULL,CA_READ)) != 0) return cc;
  if ((cc=GetValuesAsLong(&das->dout,&val,1)) != 0) return cc;
  if (val == das->value) return 0;
  return not_ready;
}
int SendSignal(DAS *das)
{
  char devname[256],tmpname[128];
  DTYPE din, *dinptr, *doutptr;

  if (das->din.dArrayLength > 0)
  {
    dinptr = &das->din;
    doutptr = &das->dout;
  }
  else
  { /* legacy default case: send 1 int32 value */
    memset(&din,0,sizeof(DTYPE));
    din.dFormat = CF_INT32;
    din.dArrayLength = 1;
    din.data.lptr = &das->value;
    dinptr = &din;
    doutptr = NULL;
  }
  memset(tmpname,0,DEVICE_NAME_SIZE);
  strncpy(tmpname,das->dev,DEVICE_NAME_SIZE);  /* copy to tmpname area for manipulation */
  if (!strcmp(tmpname,DEVICE_ALL))
  { /* prepare new tmpname a la the #0-#100 scheme */
    if (rewireAllDeviceName(tmpname,das) != 0) return un_allocated;
  }
  if (!strcmp(tmpname,DEVICE_IGNORE))
  { /* fix to "#0" */
    strcpy(tmpname,"#0");
  }
  sprintf(devname,"/%.32s/%.32s/%.64s",das->ctx,das->srv,tmpname);
  return ExecLinkEx(devname,das->prp,doutptr,dinptr,CA_WRITE,200);
}
int GetDataForNextStep(DAS *dasIn,DAS *dasOut)
{
  int i,cc=0;
  char devname[132];
  DTYPE *pdin=NULL,dout;

  if (dasIn->dout.dFormat != dasOut->din.dFormat) ccerr(illegal_format);
  if (dasIn->din.dArrayLength > 0) pdin = &dasIn->din;
  memcpy(&dout,&dasIn->dout,sizeof(DTYPE));
  sprintf(devname,"/%.32s/%.32s/%.64s",dasIn->ctx,dasIn->srv,dasOut->dev);
  if ((cc=ExecLinkEx(devname,dasIn->prp,&dout,pdin,CA_READ,3000)) != 0) ccerr(cc);
  for (i=0; i<(int)dasOut->din.dArrayLength && i<(int)dasIn->dout.dArrayLength; i++)
  {
    switch (dasOut->din.dFormat)
    {
      case CF_INT16: 
           dasOut->din.data.sptr[i] = dasIn->dout.data.sptr[i]; break;
      case CF_INT32:  
           dasOut->din.data.lptr[i] = dasIn->dout.data.lptr[i]; break;
      case CF_FLOAT:  
           dasOut->din.data.fptr[i] = dasIn->dout.data.fptr[i]; break;
      case CF_DOUBLE:  
           dasOut->din.data.dptr[i] = dasIn->dout.data.dptr[i]; break;
      case CF_BYTE:  
           dasOut->din.data.bptr[i] = dasIn->dout.data.bptr[i]; break;
      default:
           ccerr(illegal_format);
    }
  }
err:
  return cc;
}
void checkRecordEvent(int cc)
{
  char errstr[64];
  if (cc) ttyoutput("RecordEvent: %.32s",GetLastLinkError((short)cc,errstr));
  RecordEventPending = FALSE;
}
void checkSetEvent(UINT32 eventnumber, int cc)
{
  char errstr[64];
  if (cc == 0) /* everything is okay */
  {
    lastEventNumber = eventnumber;
  }
  EventNotificationPending = FALSE;
  ttyoutput("SetEvent %d : %.32s",eventnumber,GetLastLinkError((short)cc,errstr));
}
int stopARB(ARB *arb)
{
  int i;
  if (arb->step > 0) arb->step = arb->nsteps;
  for (i=0; i<ndmcs; i++) if (dmcs[i].arb == arb) break;
  if (i < ndmcs)
  {
    dmcs[i].lifetime = 0;
  }
  return 0;
}
int isArchiveOngoing(char *filename)
{
  int i;
  char fn[256];

  if (filename == NULL) return FALSE;
  for (i=0; i<arbTblSiz; i++)
  {
    if (arbTbl[i].fp == NULL) continue;
    makefilename(fn,stdArchiveDirectory,arbTbl[i].ext,arbTbl[i].eventtime);
    if (strncmp(filename,fn,128)) continue;
    AppendHistoryLog("attempt to access on-going archive %.256s refused\n",fn);
    return TRUE;
  }
  return FALSE;
}
void freeArbTbl(void)
{
  int i, k;
  DASDEV *devlst, *devnxt;
  DEVLSTTBL *dlt=NULL, *dltnxt;
  for (dlt=devLstTbl; dlt != NULL; dlt=dltnxt)
  {
    dltnxt = dlt->nxt;
    for (devlst = dlt->devlst; devlst != NULL; devlst = devnxt)
    {
      devnxt = devlst->nxt;
      if (devlst->names != NULL) free(devlst->names);
      if (devlst->alias != NULL) free(devlst->alias);
      free(devlst);
    }
  }
  devLstTbl = NULL;
  for (i=0; i<arbTblSiz; i++)
  {
    if (arbTbl[i].das != NULL)
    {
      if (arbTbl[i].das[0].prpAliasLst != NULL)
      { /* all DAS elements will reference the same ... */
        free(arbTbl[i].das[0].prpAliasLst);
      }
      if (arbTbl[i].das[0].xDevLst != NULL)
      { /* all DAS elements will reference the same ... */
        free(arbTbl[i].das[0].xDevLst);
      }
      for (k=0; k<arbTbl[i].nsteps; k++)
      {
        if ((arbTbl[i].das[k].acc&SETDEV) && arbTbl[i].das[k].devTgtLst != NULL)
        { /* some DAS elements will reference the same ... */
          free(arbTbl[i].das[k].devTgtLst);
        }
        if (arbTbl[i].das[k].din.data.vptr != NULL)
        {
          free(arbTbl[i].das[k].din.data.vptr);
        }
        if (arbTbl[i].das[k].freeMemoryOnExit)
        {
          free(arbTbl[i].das[k].dout.data.vptr);
        }
      }
    }
  }
  free(arbTbl); free(arbTrgList);
  arbTrgList = NULL; 
  arbTbl = NULL;
  arbTblSiz = 0;
  return;
}
#define M2SEC(v) (((float)(v))/1000)
int arbTblBusy = FALSE;
void bkgArchiveTask(void)
{
  char fn[256],scratch[128],errstr[64],fulldevname[132];
  int i,step_idx,k,cc,newmonth = FALSE,insize = 0,advanceStep = TRUE;
  time_t tnow=time(NULL),stop;
  double tdtnow = MakeDataTimeStamp();
  struct tm *tmp,tm,tmref;
  static int errCount = 0, mon = -1;

  checkSessionActivity();

  tmp = localtime(&tnow);
  memcpy(&tm,tmp,sizeof(struct tm));
  if (mon != tm.tm_mon)
  {
    newmonth = TRUE;
    mon = tm.tm_mon;
  }
  arbTblBusy = FALSE;
  for (i=0; i<arbTblSiz; i++)
  {
    if (arbTbl[i].das == NULL) continue; /* configuration problems ! */
    if (arbTbl[i].maxq < 0) continue; /* disabled */
    if (arbTbl[i].depthInMonths > 0 && newmonth)
    {
      tmref.tm_mon = (tm.tm_mon - arbTbl[i].depthInMonths + 240) % 12;
      tmref.tm_year = tm.tm_year - arbTbl[i].depthInMonths/12;
      if (tmref.tm_mon > tm.tm_mon) tmref.tm_year--;
      tmref.tm_mday = 1;
      tmref.tm_hour = 0;
      tmref.tm_min = 0;
      tmref.tm_sec = 0;
      tmref.tm_isdst = -1;
      stop = mktime(&tmref);
      /* signal removal: buf = NULL, starttime = 0, bufsiz = -1 */
      getFiles(NULL,arbTbl[i].ext,0,stop,-1);
    }
    if (arbTbl[i].step < 0) 
    { /* pre-step: get trigger time */
      arbTblBusy = TRUE;
      if (!arbTbl[i].resultPending)
      {
        lastEventNumber = (UINT32)time(NULL);
        arbTbl[i].resultPending = TRUE;
        if ((cc=SetGlobalEvent((MARCHSRV_EVENTCODE|i),arbTbl[i].trg)) != 0) 
        {
          AppendHistoryLog("SetGlobalEvent : %s",cc2str(cc));
        }
        else EventNotificationPending = TRUE;
      }
      if (EventNotificationPending) continue;
      arbTbl[i].eventtime = lastEventNumber;
      arbTbl[i].step = 1;
      AppendHistoryLog("%.64s : assign event 0x%x",arbTbl[i].trg,lastEventNumber);
    }
    step_idx = arbTbl[i].step - 1; // off by one in the log report and the internal logic
    if (step_idx < 0) 
    { /* this trigger type is idle */
      if (arbTbl[i].nevnq > 0)
      { /* there is a queue ! */
        EVNQ *pe = NULL, *e = arbTbl[i].evnq;
        while (e->nxt != NULL) 
        {
          pe = e; e = e->nxt;
        }
        arbTbl[i].step = 1;
        arbTbl[i].applyRange = e->applyRange;
        arbTbl[i].maxInRange = e->maxInRange;
        arbTbl[i].tsStart = e->tsStart;
        arbTbl[i].tsStop = e->tsStop;
        arbTbl[i].sstStart = e->sstStart;
        arbTbl[i].sstStop = e->sstStop;
        arbTbl[i].nevnq--;
        if (pe != NULL) pe->nxt = NULL; else arbTbl[i].evnq = NULL;
        if (arbTbl[i].evnq == NULL) arbTbl[i].nevnq = 0; /* sanity check */
      }
      continue;
    }
    arbTblBusy = TRUE;
    if (step_idx >= arbTbl[i].nsteps)
    { /* we're at the end : close up shop */
      scratch[0] = 0;
      if (!arbTbl[i].isMonitoring)
      {
        AppendHistoryLog("%.64s : close archive file",arbTbl[i].trg);
        arbTbl[i].step = 0;
        arbTbl[i].resultPending = FALSE;
        if (arbTbl[i].fp != NULL) fclose(arbTbl[i].fp);
        arbTbl[i].fp = NULL;
      }
      else
      { /* monitor callback will set step = 0 when finished */
        continue;
      }
      if (arbTbl[i].aborted)
      {
        makefilename(fn,stdArchiveDirectory,arbTbl[i].ext,arbTbl[i].eventtime);
        AppendHistoryLog("%.64s : removing archive file %.256s",arbTbl[i].trg,fn);
        remove(fn);
      }
      else if (arbTbl[i].recordEvents > 0)
      { /* register with the SITE event server */
        RecordEventPending = TRUE;
        for (k=0; k<arbTbl[i].nsteps; k++)
        {
          if (strncmp(arbTbl[i].das[k].srvArc,scratch,EXPORT_NAME_SIZE))
          {
            strncpy(scratch,arbTbl[i].das[k].srvArc,EXPORT_NAME_SIZE);
            RecordEvent((UINT32)arbTbl[i].eventtime,MARCHSRV_EVENTCODE|i,scratch);
          }
        }
        /* worry about whether this was successful or not later */
      }
      continue;
    }
    /* All archive procedures get initialized */
    if (step_idx == 0 && arbTbl[i].fp == NULL)
    { /* first things first */
      if (arbTbl[i].isAlarm)
      {
        cc = SetDeviceAlarm(pmArchEqm, arbTbl[i].trg, arbTbl[i].almCode, NULL, almTRANSIENT);
        AppendHistoryLog("SetDeviceAlarm for %.64s : %s", arbTbl[i].trg, cc2str(cc));
        arbTbl[i].isAlarm = FALSE;
      }
      errCount = 0;
      AppendHistoryLog("%s : start archiving",arbTbl[i].trg);
      arbTbl[i].interim = tdtnow; 
      arbTbl[i].start = tnow;
      arbTbl[i].aborted = 0;
      /* use the event number !!! */
      makefilename(fn,stdArchiveDirectory,arbTbl[i].ext,arbTbl[i].eventtime);
      /* check: if this file exists, then I'm already doing or have done this! */
      if ((arbTbl[i].fp = fopen(fn,"wb")) == NULL &&
          (arbTbl[i].fp = createAndOpen(fn,"wb")) == NULL)
      {
        AppendHistoryLog("%.64s : can't open file %.256s",arbTbl[i].trg,fn);
        arbTbl[i].step = arbTbl[i].nsteps+1;
        continue;
      }
      AppendHistoryLog("%s : open file %s",arbTbl[i].trg,fn);
      memset(scratch,0,VERSION_LENGTH); strncpy(scratch,ARCHIVE_STORAGE_VERSION,VERSION_LENGTH);
      if (fwrite(scratch,VERSION_LENGTH,1,arbTbl[i].fp) != 1) 
      {
        AppendHistoryLog("Cannot write version number to file %s !",fn);
      }
    }
    if (arbTbl[i].das[step_idx].acc&DISABLE)
    {
      AppendHistoryLog("%.64s : archive step %d is disabled",arbTbl[i].trg,step_idx);
      arbTbl[i].step++;
      arbTbl[i].interim = tdtnow;
      arbTbl[i].start = tnow;
      continue;
    }
    advanceStep = TRUE; /* start off optimistic ... */
    /* is there an acquired device list ? */
    if (step_idx > 0 && arbTbl[i].das[step_idx-1].devTgtLst != NULL && arbTbl[i].das[step_idx].devTgtLst == NULL)
    { /* forward the device names for subsequent calls */
      arbTbl[i].das[step_idx].devTgtLst = arbTbl[i].das[step_idx-1].devTgtLst;
      AppendHistoryLog("%.64s : forward device target list to acquisition property %.64s",
        arbTbl[i].trg,arbTbl[i].das[step_idx].prp);
    }
    if (arbTbl[i].das[step_idx].acc&UNSETDEV)
    { /* explicit instructions not to use acquired device names */
      arbTbl[i].das[step_idx].devTgtLst = NULL;
      AppendHistoryLog("%.64s : abandon device target list",arbTbl[i].trg);
    }
    if (arbTbl[i].das[step_idx].devTgtLst != NULL && !(arbTbl[i].das[step_idx].acc&SETDEV) &&
        (k=arbTbl[i].das[step_idx].devTgtLst->idx) < arbTbl[i].das[step_idx].devTgtLst->siz)
    { /* only land here if I'm not the step that gets the target device names */
      if (k < 0)
      { /* something weird: bail ! */
        arbTbl[i].das[step_idx].devTgtLst->idx = arbTbl[i].das[step_idx].devTgtLst->siz = 0;
        AppendHistoryLog("%.64s : device index %d not accepted!",arbTbl[i].trg,k);
      }
      else
      { /* prepare current device target */
        strncpy(arbTbl[i].das[step_idx].dev,arbTbl[i].das[step_idx].devTgtLst->lst[k].name,64);
        AppendHistoryLog("%.64s : use device target %.64s (device %d of %d) for property %.64s",
          arbTbl[i].trg,arbTbl[i].das[step_idx].dev,(k+1),
          arbTbl[i].das[step_idx].devTgtLst->siz,arbTbl[i].das[step_idx].prp);
        arbTbl[i].das[step_idx].devTgtLst->idx++; /* next up ... */
        if (arbTbl[i].das[step_idx].devTgtLst->idx < arbTbl[i].das[step_idx].devTgtLst->siz) 
        { /* still more to come: stay on the given step */
          advanceStep = FALSE; 
        }
        else
        { /* last one in the list: prepare for next step */
          arbTbl[i].das[step_idx].devTgtLst->idx = 0;
        }
      }
    }
    sprintf(fulldevname,"/%.32s/%.32s/%.64s",arbTbl[i].das[step_idx].ctx,arbTbl[i].das[step_idx].srv,arbTbl[i].das[step_idx].dev);
    if (arbTbl[i].das[step_idx].acc&MATCH)
    { /* Does the archive step involve polling for a specific return value ? */
      if (tdtnow < arbTbl[i].interim + arbTbl[i].das[step_idx].wait) continue;
      if (arbTbl[i].das[step_idx].wait > 0) AppendHistoryLog("%.64s : waited %d sec prior to this step",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
      AppendHistoryLog("%.64s : polling %.132s %.64s against %d",
        arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp,arbTbl[i].das[step_idx].value);
      switch (cc = PollTarget(&arbTbl[i].das[step_idx]))
      {
        case 0: 
          if (advanceStep) arbTbl[i].step++; /* target met, next step */
          arbTbl[i].start = tnow;
          AppendHistoryLog("%.64s : target met",arbTbl[i].trg);
          break;
        case not_ready:
          if (tnow > arbTbl[i].start + arbTbl[i].das[step_idx].timeout)
          {
            AppendHistoryLog("%.64s : %.132s %.64s timed out",arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp);
            arbTbl[i].step = arbTbl[i].nsteps+1;
            if (archDebug) dbglog("timed out");
          }
          errCount = 0;
          break; /* keep trying */
        default:
          if (errCount++ > 10)  arbTbl[i].step = arbTbl[i].nsteps + 1;
          AppendHistoryLog("%.64s : %.132s %.64s -> %.32s",
            arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp,GetLastLinkError((short)cc,errstr));
          break;
      }
      arbTbl[i].interim = tdtnow;
      continue;
    }
    if (arbTbl[i].das[step_idx].acc&WRITE)
    { /* Does the archive step involve sending a signal to the FEC? */
      if (tdtnow < arbTbl[i].interim + arbTbl[i].das[step_idx].wait) continue;
      if (arbTbl[i].das[step_idx].wait > 0) AppendHistoryLog("%s : waited %d sec prior to this step",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
      AppendHistoryLog("%.64s : Send signal %.132s %.64s",arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp);
      cc = SendSignal(&arbTbl[i].das[step_idx]);
      if (advanceStep) arbTbl[i].step++;
      arbTbl[i].interim = tdtnow;
      arbTbl[i].start = tnow;
      continue;
    }
    if (arbTbl[i].das[step_idx].acc&READ)
    { /* Does the archive step involve getting archive data ? */
      if (tdtnow < arbTbl[i].interim + arbTbl[i].das[step_idx].wait) continue;
      if (arbTbl[i].das[step_idx].wait > 0) AppendHistoryLog("%s : waited %d sec prior to this step",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
      if (step_idx>0 && arbTbl[i].das[step_idx-1].numberToAcquire > 0)
      { /* if set then pass it along ... */
        arbTbl[i].das[step_idx].numberToAcquire = arbTbl[i].das[step_idx-1].numberToAcquire;
      }
      else 
      {
        arbTbl[i].das[step_idx].numberToAcquire = 0;
      }
      AppendHistoryLog("%.64s : get and store %.132s %.64s",arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp);
      if (arbTbl[i].das[step_idx].isAcqHist)
      { /* a .HIST@ thing -> get a range of stored items from the local history */
        if (arbTbl[i].das[step_idx].seriesIndex == 0) arbTbl[i].das[step_idx].seriesIndex = 1;
        arbTbl[i].das[step_idx].numberInSeries = arbTbl[i].das[step_idx].seriesIndex + 1;
        if (arbTbl[i].das[step_idx].din.data.vptr == NULL)
        {
          AppendHistoryLog("%s : no memory for history input",arbTbl[i].trg);
          if (advanceStep) arbTbl[i].step++;
          continue;
        }
        if (arbTbl[i].sstStart > 0)
        { /* get series of snapshots over a system stamp range */
          if (arbTbl[i].sstStart < arbTbl[i].sstStop)
          {
            arbTbl[i].das[step_idx].din.dFormat = CF_INT32;
            if (arbTbl[i].applyRange == APPLY_USRSTAMP_RANGE)
            {
              arbTbl[i].das[step_idx].din.dArrayLength = 6;
              arbTbl[i].das[step_idx].din.data.lptr[5] = 1;
            }
            else
            {
              arbTbl[i].das[step_idx].din.dArrayLength = 5;
            }
            arbTbl[i].das[step_idx].din.data.lptr[0] = (int)arbTbl[i].tsStart;
            arbTbl[i].das[step_idx].din.data.lptr[1] = (int)arbTbl[i].tsStop;
            arbTbl[i].das[step_idx].din.data.lptr[2] = 0;
            arbTbl[i].das[step_idx].din.data.lptr[3] = 0;
            arbTbl[i].das[step_idx].din.data.lptr[4] = arbTbl[i].sstStart;
            cc = mtGetAndStore(&arbTbl[i],step_idx);
            if (cc == 0 || cc == not_ready) continue;
          }
        }
        else if (arbTbl[i].tsStart < arbTbl[i].tsStop)
        { /* get series of snapshots over a time stamp range */
          arbTbl[i].das[step_idx].din.dFormat = CF_DOUBLE;
          arbTbl[i].das[step_idx].din.dArrayLength = 2;
          arbTbl[i].das[step_idx].din.data.dptr[0] = arbTbl[i].tsStart;
          arbTbl[i].das[step_idx].din.data.dptr[1] = arbTbl[i].tsStop;
          cc = mtGetAndStore(&arbTbl[i],step_idx);
          if (cc == 0 || cc == not_ready) continue;
        }
        arbTbl[i].das[step_idx].seriesIndex = 0;
        if (advanceStep) arbTbl[i].step++;
        arbTbl[i].interim = tdtnow;
        arbTbl[i].start = tnow;
        continue;
      }
      cc = mtGetAndStore(&arbTbl[i], step_idx);
      if (cc == not_ready) continue;
      if (advanceStep) arbTbl[i].step++;
      arbTbl[i].interim = tdtnow;
      arbTbl[i].start = tnow;
      continue;
    }
    if (arbTbl[i].das[step_idx].acc&FORWARD)
    { /* Does the archive step involve getting archive data to forward ? */
      if (tdtnow < arbTbl[i].interim + arbTbl[i].das[step_idx].wait) continue;
      if (arbTbl[i].das[step_idx].wait > 0) AppendHistoryLog("%.64s : waited %d sec prior to this step",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
      for (k=step_idx+1; k<arbTbl[i].nsteps; k++)
      { /* look at steps after this one */
        if (k == (step_idx+1))
        { /* copy to the very next step */
          strncpy(scratch,arbTbl[i].das[k].infile,16);
          GetDataForNextStep(&arbTbl[i].das[step_idx],&arbTbl[i].das[k]);
          /* n.b. only simple formats allowed as input */
          insize = arbTbl[i].das[k].din.dArrayLength * fmtsizeof(arbTbl[i].das[k].din.dFormat);
        }
        else 
        { /* forward to all adjacent steps with the same input set */
          if (strncmp(scratch,arbTbl[i].das[k].infile,16)) break;
          memcpy(arbTbl[i].das[k].din.data.bptr,arbTbl[i].das[step_idx+1].din.data.bptr,insize);
        }
      }
      if (advanceStep) arbTbl[i].step++;
      arbTbl[i].interim = tdtnow;
      arbTbl[i].start = tnow;
      continue;
    }
    if (arbTbl[i].das[step_idx].acc&SERIES)
    { /* Does the archive step involve getting an archive data series ? */
      if (arbTbl[i].das[step_idx].seriesIndex <= arbTbl[i].das[step_idx].numberInSeries)
      {
        if (arbTbl[i].das[step_idx].seriesIndex == 0) arbTbl[i].das[step_idx].seriesIndex++; /* start it off */
        if (!arbTbl[i].das[step_idx].isWaiting)
        {
          if (step_idx>0 && arbTbl[i].das[step_idx-1].numberToAcquire > 0)
          { /* if set then pass it along ... */
            arbTbl[i].das[step_idx].numberToAcquire = arbTbl[i].das[step_idx-1].numberToAcquire;
          }
          else
          {
            arbTbl[i].das[step_idx].numberToAcquire = 0;
          }
          AppendHistoryLog("%.64s : get and store %.132s %.64s",arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp);
          cc = mtGetAndStore(&arbTbl[i],step_idx);
          if (cc == not_ready) continue;
          arbTbl[i].interim = tdtnow;
          arbTbl[i].start = tnow;
          arbTbl[i].das[step_idx].isWaiting = TRUE;
        }
        if (tdtnow < arbTbl[i].interim + M2SEC(arbTbl[i].das[step_idx].wait)) continue; /* wait is in msec for SERIES */
        arbTbl[i].interim = tdtnow;
        arbTbl[i].start = tnow;
        arbTbl[i].das[step_idx].isWaiting = FALSE;
        arbTbl[i].das[step_idx].seriesIndex++;
        AppendHistoryLog("%.64s : waited %d msec",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
        continue;
      }
      arbTbl[i].das[step_idx].seriesIndex = 0;
      if (advanceStep) arbTbl[i].step++;
      continue;
    }
    if (arbTbl[i].das[step_idx].acc&MONITOR)
    { /* Does the archive step involve starting an archive data monitor ? */
      if (tdtnow < arbTbl[i].interim + arbTbl[i].das[step_idx].wait) continue;
      if (arbTbl[i].das[step_idx].wait > 0) AppendHistoryLog("%.64s : waited %d sec prior to this step",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
      AppendHistoryLog("%.64s : monitor and store %.132s %.64s",arbTbl[i].trg,fulldevname,arbTbl[i].das[step_idx].prp);
      if (BASEMODE(arbTbl[i].das[step_idx].mode) != CM_CANCEL)
      {
        cc = MonitorAndStore(&arbTbl[i],&arbTbl[i].das[step_idx]);
      }
      arbTbl[i].start = tnow;
      if (advanceStep) arbTbl[i].step++;
      arbTbl[i].nsteps = arbTbl[i].step; /* must be the last script element */
      continue;
    }
    if (arbTbl[i].das[step_idx].acc&WAIT)
    { /* Does the archive step simply involve waiting a specified time? */
      if (tdtnow < arbTbl[i].interim + M2SEC(arbTbl[i].das[step_idx].wait)) continue;
      arbTbl[i].interim = tdtnow;
      arbTbl[i].start = tnow;
      if (advanceStep) arbTbl[i].step++;
      AppendHistoryLog("%.64s : waited %d msec",arbTbl[i].trg,arbTbl[i].das[step_idx].wait);
      continue;
    }
    /* empty step ? */
    AppendHistoryLog("%.64s : empty step",arbTbl[i].trg);
    arbTbl[i].step++;
  }
  if (!arbTblBusy)
  { // not handling anything, so ...
    for (i=0; i<arbTblSiz; i++)
    {
      if (arbTbl[i].tot == 0) arbTbl[i].tot = tnow; // just starting up
      if (tdtnow > (double)(arbTbl[i].tot + 3600))
      { // 1 whole hour of nothing to do with this guy
        AppendHistoryLog("bkgArchiveTask: check trigger %.64s for device updates", arbTbl[i].trg);
        RefreshDeviceLists(&arbTbl[i]);
        arbTbl[i].tot = (time_t)tdtnow;
        break; // refresh device lists one trigger at a time
      }
    }
  }
}
int FillPropertyAliasList(ARB *arb)
{
  char fname[80];
  int i,j,cc=0,done,name_col,alias_col,n;
  char str[256],hdr[256],srv[EXPORT_NAME_SIZE+8],prpname[PROPERTY_NAME_SIZE+8],alias[PROPERTY_NAME_SIZE+8];
  FILE *fp = NULL;
  srv[0] = 0; prpname[0] = 0;
  for (j=0; j<arb->nsteps; j++) if (arb->das[j].acc&READ || arb->das[j].acc&SERIES)
  {
    if (!strncmp(srv,arb->das[j].srvArc,EXPORT_NAME_SIZE)) continue;
    if (!strncmp(prpname,arb->das[j].prp,PROPERTY_NAME_SIZE)) continue;
    strncpy(srv,arb->das[j].srvArc,EXPORT_NAME_SIZE);
    sprintf(fname,"%s-properties.csv",srv);
    if (fp != NULL) fclose(fp);
    if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
scanfile:
    n = 0; done = 0;
    while (csvGetLine(str,255,fp) != NULL)
    {
      strtrm(str);
      if (strchr("%#;\n",str[0])) continue;
      if (!done)
      {
        str[255] = 0; strcpy(hdr,str);
        if ((name_col=findcol(hdr,"PROPERTY_NAME")) < 0) ccerr(no_such_column);
        if ((alias_col=findcol(hdr,"PROPERTY_ALIAS")) < 0) ccerr(no_such_column);
        done = 1;
        continue;
      }
      strncpy(prpname,colptr(name_col,str),PROPERTY_NAME_SIZE);
      if (strncmp(arb->das[j].prp,prpname,PROPERTY_NAME_SIZE)) continue;
      if (arb->das[j].prpAliasLst != NULL)
      {
        strncpy(alias,colptr(alias_col,str),PROPERTY_NAME_SIZE);
        if (n < arb->das[j].nprpAlias)
        {
          strncpy(arb->das[j].prpAliasLst[n].name,alias,PROPERTY_NAME_SIZE);
        }
      }
      n++;
    }
    if (n > 0 && arb->das[j].prpAliasLst == NULL)
    {
      arb->das[j].nprpAlias = n;
      if ((arb->das[j].prpAliasLst=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
        ccerr(out_of_local_memory);
      rewind(fp);
      goto scanfile;
    }
    for (i=j+1; i<arb->nsteps; i++)
      if (arb->das[i].acc&(READ|SERIES) &&
          !strncmp(arb->das[i].srvArc,srv,EXPORT_NAME_SIZE) &&
          !strncmp(arb->das[i].prp,prpname,PROPERTY_NAME_SIZE))
    {
      arb->das[i].prpAliasLst = arb->das[j].prpAliasLst;
      arb->das[i].nprpAlias   = n;
    }
  }
err:
  if (fp) fclose(fp);
  return cc ? -cc : n;
}
typedef struct 
{
  char fname[32];
  time_t maxValid;
  time_t minValid;
} DevsFile;
int getDevicesFileSet(char *path,DevsFile *files,int *nfiles)
{
# define fpat "devices"
  int i=0, k, n, maxfiles;
  long lval;
  char *c,*cstop,dirname[128];
  DevsFile tmp;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

  if (path == NULL || files == NULL || nfiles == NULL) return argument_list_error;
  maxfiles = *nfiles;
  *nfiles = 0;
# if defined(UNIX)
  strncpy(dirname,path,96); dirname[96] = 0;
  if ((dp=opendir(dirname)) == NULL) return no_such_file;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name,fpat) == NULL) continue;
    strncpy(files[i].fname,dirp->d_name,32); files[i].fname[32] = 0;
# elif defined(WIN32)
  sprintf(dirname,"%s\\*.*",path);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return no_such_file;
  do
  {
    if (strstr(fd.name,fpat) == NULL) continue;
    strncpy(files[i].fname,fd.name,32); files[i].fname[32] = 0;
# endif
    if ((c=strchr(files[i].fname,'-')) != NULL)
    {
      c++;
      lval = strtol(c,&cstop,10);
      if (strcmp(cstop,".csv") != 0) continue;
      files[i].maxValid = (time_t)lval;
    }
    else
    {
      if (strcmp(files[i].fname,DEVICES_FILE_NAME) != 0) continue;
      files[i].maxValid = time(NULL);
    }
    if (++i == maxfiles) break;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
# endif
  maxfiles = *nfiles = i;

  /* now sort the files : oldest to newest */
  for (i=0; i<maxfiles; i++)
  {
    for (k=0; k<i; k++)
    {
      if (files[i].maxValid < files[k].maxValid) break;
    }
    if (i > 0 && k < i)
    {
      tmp = files[i];
      for (n=i; n>k; n--) files[n] = files[n-1];
      files[k] = tmp;
    }
  }
  /* now figure out the minValids ... */
  files[0].minValid = 0;
  for (i=1; i<maxfiles; i++)
  {
    files[i].minValid = files[i-1].maxValid;
  }
  return 0;
}
DEVLSTTBL *devLstTbl = NULL;

DEVLSTTBL *getDevLstTblEntry(char *ctx,char *srv,char *srvArc)
{
  DEVLSTTBL *dlt=NULL;
  DASDEV *devlst = NULL;
  for (dlt=devLstTbl; dlt != NULL; dlt=dlt->nxt)
  {
    if (strnicmp(dlt->ctx,ctx,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(dlt->srv,srv,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(dlt->srvArc,srvArc,EXPORT_NAME_SIZE)) continue;
    break;
  }
  return dlt;
}
NAME64 _dlst[1000];
int dumpDevList(char *filename,NAME64 *_dlst,int ndlst)
{
  FILE *fp=NULL;
  int i,cc=0;
  if ((fp=fopen(filename,"w")) == NULL &&
      (fp=createAndOpen(filename,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"DEVICE_NUMBER, DEVICE_NAME\n");
  for (i=0; i<ndlst; i++)
  {
    fprintf(fp," %d, %.64s\n",i,_dlst[i].name);
  }
  feclog("update device list file %.160s\n",filename);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
time_t getServerStartTime(char *ctx, char *srv)
{
  char dev[128];
  DTYPE dout;
  time_t st = 0;
  UINT32 v = 0;
  int cc = 0;
  if (ctx == NULL || srv == NULL) return 0;
  DTYPEZERO(dout);
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.ulptr = &v;
  sprintf(dev, "/%.32s/%.32s/#0", ctx, srv);
  cc = ExecLink(dev, "SRVSTARTTIME", &dout, NULL, CA_READ);
  if (cc == 0)
  {
    st = (time_t)v;
  }
  return st;
}
/*
 * this will be called for each DAS entry so the 'already_assigned' is 
 * important, as it stops the routine from needlessly re-acquiring 
 * device lists, etc.
 */
int addDevLstTblEntry(DAS *das,char *ctx,char *srv,char *srvArc,char *prp)
{
# define ST_ALLOWED_DEADTIME 60
  DASDEV *devlst = das->devlst;
  DEVLSTTBL *dlt=NULL;
  char tgt[256], path[128], filename[160], newname[160];
  int cc = 0, ndlst = 1000;
  int advance = FALSE;
  time_t t = time(NULL);

  if ((dlt=getDevLstTblEntry(ctx, srv, srvArc)) != NULL)
  {
    time_t st = getServerStartTime(ctx, srv);
    if (dlt->devlst == NULL) return invalid_reference;
    st += ST_ALLOWED_DEADTIME;
    if (st <= dlt->devlst->minTimeValid) return already_assigned;
    devlst = dlt->devlst; // should be same as das->devlst
  }
  else
  {
    if ((dlt=(DEVLSTTBL *)calloc(1,sizeof(DEVLSTTBL))) == NULL) return out_of_server_memory;
    strncpy(dlt->ctx, ctx, CONTEXT_NAME_SIZE);
    strncpy(dlt->srv, srv, EXPORT_NAME_SIZE);
    strncpy(dlt->srvArc, srvArc, EXPORT_NAME_SIZE);
    advance = TRUE;
  }
  sprintf(tgt,"/%.32s/%.32s/#0",ctx,srv);
  if (applyDeviceNames(das->dev) && /* need to assign names to numbers */
      (cc=GetSystemDevices(tgt,prp,_dlst,&ndlst)) == 0)
  { /* acquisition successful */
    if (devlst == NULL || devlst->names == NULL || 
        memcmp(_dlst,devlst->names,ndlst*sizeof(NAME64)) != 0)
    { /* server has a newer set of device names than most recently cached ... */
      sprintf(path,"%s%c%s%c%.32s",
        stdArchiveDirectory,FS_DELIMITER,SERVER_FILES_SUBDIR,FS_DELIMITER,srvArc);
      sprintf(filename,"%.128s%c%s",path,FS_DELIMITER,DEVICES_FILE_NAME);
      sprintf(newname,"%.128s%cdevices-%d.csv",path,FS_DELIMITER,(int)t);
      rename(filename,newname);
      dumpDevList(filename,_dlst,ndlst);
      if ((devlst=(DASDEV *)calloc(1,sizeof(DASDEV))) == NULL) ccerr(out_of_local_memory);
      if ((devlst->names=(NAME64 *)calloc(ndlst,sizeof(NAME64))) == NULL) ccerr(out_of_local_memory);
      memcpy(devlst->names,_dlst,ndlst*sizeof(NAME64));
      devlst->size = ndlst;
      devlst->minTimeValid = t;
      devlst->nxt = das->devlst;
      das->devlst = devlst;
    }
    dlt->hasAcquiredCurrentList = TRUE;
    dlt->devlst = devlst;
  }
err:
  if (cc != 0) feclog("addDevLstTblEntry : %.32s %.32s (%.32s) %.64s : %.32s",ctx,srv,srvArc,prp,cc2str(cc));
  if (advance)
  {
    dlt->nxt = devLstTbl;
    devLstTbl = dlt;
  }
  return cc;
}
int assignExtendedDeviceList(DAS *das,char *path,char *filename)
{
  char fn[80];
  int i, cc=0,done,name_col,num_col,n;
  char *c,str[256],hdr[256],scratch[128];
  NAME64 *xlst = NULL;
  FILE *fp = NULL;
  int p = 0,found = 0;

  sprintf(fn,"%s%c%.32s",path,FS_DELIMITER,filename);
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(no_such_file);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  if ((xlst=(NAME64 *)calloc(1,sizeof(NAME64))) == NULL) ccerr(out_of_local_memory);
  done = 0;
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strchr("%#;\n",str[0])) continue;
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((name_col=findcol(hdr,"DEVICE_NAME")) < 0) ccerr(no_such_column);
      if ((num_col=findcol(hdr,"DEVICE_NUMBER")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(num_col,str),16);
    i = atoi(scratch);
    if (i < n)
    {
      if (*(c=colptr(name_col,str)) == '\'') c++;
      strncpy(xlst[i].name,c,DEVICE_NAME_SIZE);
    }
  }
  das->xDevLst = xlst;
  das->nxDevices = n;
  AppendHistoryLog("found extended device list for /%.32s/%.32s [%.64s] (%d entries)",das->ctx,das->srvArc,das->prpArc,n); 

err:
  if (fp) fclose(fp);
  return cc;
}
int assignDeviceList(DAS *das,char *path,char *filename,time_t minValid)
{
  char fn[80];
  int i, cc=0,done,name_col,num_col,alias_col,n;
  char *c,str[256],hdr[256],scratch[128];
  DASDEV *devlst = NULL;
  FILE *fp = NULL;
  int p = 0,found = 0;

  sprintf(fn,"%s%c%.32s",path,FS_DELIMITER,filename);
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(no_such_file);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  if ((devlst=(DASDEV *)calloc(1,sizeof(DASDEV))) == NULL) ccerr(out_of_local_memory);
  devlst->size = n;
  devlst->minTimeValid = minValid;
  if ((devlst->names=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL) ccerr(out_of_local_memory);
  done = 0;
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strchr("%#;\n",str[0])) continue;
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str);
      if ((name_col=findcol(hdr,"DEVICE_NAME")) < 0) ccerr(no_such_column);
      if ((num_col=findcol(hdr,"DEVICE_NUMBER")) < 0) ccerr(no_such_column);
      alias_col = findcol(hdr,"DEVICE_ALIAS");
      if (alias_col >= 0 && (devlst->alias=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
        ccerr(out_of_local_memory);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(num_col,str),16);
    if (devlst != NULL)
    {
      i = atoi(scratch);
      if (i < devlst->size)
      {
        if (*(c=colptr(name_col,str)) == '\'') c++;
        strncpy(devlst->names[i].name,c,DEVICE_NAME_SIZE);
        if (devlst->alias != NULL && alias_col >= 0)
        {
          if ((c=colptr(alias_col,str)) == NULL) continue;
          strncpy(devlst->alias[i].name,c,DEVICE_NAME_SIZE);
        }
      }
    }
  }
  devlst->nxt = das->devlst;
  das->devlst = devlst;
err:
  if (fp) fclose(fp);
  return cc ? -cc : n;
}
int RefreshDeviceLists(ARB *arb)
{
  int i;
  AppendHistoryLog("RefreshDeviceLists: trigger %.64s", arb->trg);
  for (i=0; i < arb->nsteps; i++) if (arb->das[i].acc & (READ | SERIES))
  {
    addDevLstTblEntry(&arb->das[i], arb->das[i].ctx, arb->das[i].srv, arb->das[i].srvArc, arb->das[i].prp);
  }
  return 0;
}
int FillDeviceLists(ARB *arb)
{
  char path[128], srv[EXPORT_NAME_SIZE + 8] = {0}, ctx[CONTEXT_NAME_SIZE + 8] = {0};
  int i,j,nfiles,needsDeviceList=FALSE;
  DevsFile files[200];
  DEVLSTTBL *dlt;

  srv[0] = 0;
  for (j=0; j<arb->nsteps; j++) if (arb->das[j].acc&(READ|SERIES))
  {
    if ((dlt=getDevLstTblEntry(arb->das[j].ctx,arb->das[j].srv,arb->das[j].srvArc)) != NULL)
    { /* already known */
      arb->das[j].devlst = dlt->devlst;
    }
    if (arb->das[j].devlst != NULL) continue; /* already assigned */
    if (!strncmp(srv,arb->das[j].srvArc,EXPORT_NAME_SIZE)) continue;
    strncpy(srv,arb->das[j].srvArc,EXPORT_NAME_SIZE);
    strncpy(ctx,arb->das[j].ctx,CONTEXT_NAME_SIZE);
    /* target server now established: look for static device lists ...*/
    sprintf(path,"%s%c%s%c%.32s",
      stdArchiveDirectory,FS_DELIMITER,SERVER_FILES_SUBDIR,FS_DELIMITER,srv);
    needsDeviceList = applyDeviceNames(arb->das[j].dev);
    nfiles = 200;
    if (getDevicesFileSet(path,files,&nfiles) != 0 || !needsDeviceList) continue;
    for (i=0; i<nfiles; i++)
    { /* in case of multiple device files: newest is at the end of the list */
      assignDeviceList(&arb->das[j],path,files[i].fname,files[i].minValid);
      assignExtendedDeviceList(&arb->das[j],path,XDEVICES_FILE_NAME);
    }
    addDevLstTblEntry(&arb->das[j], arb->das[j].ctx, arb->das[j].srv, arb->das[j].srvArc, arb->das[j].prp);
    if (nfiles == 0)
    { /* no prior device files found : correct the min time valid field */
      if (arb->das[j].devlst != NULL) arb->das[j].devlst->minTimeValid = 0;
    }
    /* finally: scan on the current 'srv' and set the same reference for any 'associates' : */
    for (i=j+1; i<arb->nsteps; i++) 
      if (arb->das[i].acc&(READ|SERIES) && 
          !strncmp(arb->das[i].srvArc,srv,EXPORT_NAME_SIZE) &&
          !strncmp(arb->das[i].ctx,ctx,CONTEXT_NAME_SIZE))
    {
      arb->das[i].devlst = arb->das[j].devlst;
      arb->das[i].xDevLst = arb->das[j].xDevLst;
      arb->das[i].nxDevices = arb->das[j].nxDevices;
    }
  }
  return 0;
}
int GetStoredDevices(NAME64 *devnames,char *srv,char *evnTrg,UINT32 evnTime,int num)
{
  int n=0,i,j,k,found=FALSE,cc=0;
  DAS *das = NULL;
  DASDEV *devlst;
  NAME64 *xlst;
  char tmp[128], *c, *ctx=NULL, *psrv=NULL;

  memcpy(tmp,srv,128);
  if (tmp[0] == '/')
  {
    ctx = &tmp[1];
    if ((c=strchr(ctx,'/')) == NULL) return -non_existent_server;
    *c++ = 0;
    psrv = c;
  }
  else if ((c=strchr(tmp,'/')) != NULL)
  {
    ctx = tmp;
    *c++ = 0;
    psrv = c;
  }
  else
  {
    psrv = tmp;
  }
  if (evnTrg != NULL && strlen(evnTrg) > 0)
  { /* get ARB from evnTrg: */
    if ((k=GetARBIndex(evnTrg)) < 0) return 0;
  }
  else
  {
    for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    {
      if (ctx != NULL && strnicmp(ctx,arbTbl[k].das[j].ctx,CONTEXT_NAME_SIZE)) continue;
      if (!strnicmp(psrv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = TRUE;
    }
    if (!found) return -non_existent_elem;
    k--; /* back up ! */
  }
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  { /* get das from srv : */
    das = &arbTbl[k].das[j];
    if (ctx != NULL && strnicmp(ctx,das->ctx,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(psrv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (devlst=das->devlst; devlst != NULL; devlst=devlst->nxt) if (evnTime >= devlst->minTimeValid) break;
    if (devlst != NULL) 
    { /* a normally acquired device list was found */
      if (devnames != NULL && num > 0)
      {
        memset(devnames,0,num*sizeof(NAME64));
        for (n=0; n<devlst->size && n<num; n++)
        {
          if (devlst->alias != NULL && strlen(devlst->alias[n].name) > 0)
            strncpy(devnames[n].name,devlst->alias[n].name,DEVICE_NAME_SIZE);
          else 
            strncpy(devnames[n].name,devlst->names[n].name,DEVICE_NAME_SIZE);
        }
      }
      n = devlst->size;
      break;
    }
    else
    { /* no device list available: take the given name */
      if (n < num && devnames != NULL) strncpy(devnames[n].name,das->dev,DEVICE_NAME_SIZE);
      n++;
    }
    if ((xlst=das->xDevLst) != NULL)
    { /* an extended device list (for bothersome cases) */
      if (devnames != NULL)
      {
        for (i=0; i<das->nxDevices && i<num-n; i++)
        {
          strncpy(devnames[n+i].name,das->xDevLst[i].name,DEVICE_NAME_SIZE);
        }
      }
      n += das->nxDevices;
      break;
    }
  }
  return n;
}
char *GetStoredDeviceName(int devnr,char *srv,int evnId)
{
  int j,k,p = 0,found = 0;
  DAS *das;
  DASDEV *devlst;
  static char lcldevnam[DEVICE_NAME_SIZE+8];

  sprintf(lcldevnam,"#%d",devnr); /* default */
  /* get das from srv : */
  for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    if (!strnicmp(srv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
  if (!found) return lcldevnam;
  k--; /* back up ! */
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (devlst=das->devlst; devlst != NULL; devlst=devlst->nxt) if (evnId >= devlst->minTimeValid) break;
    if (devlst != NULL) 
    {
      if (devnr < 0 || devnr >= devlst->size) return NULL;
      strncpy(lcldevnam,devlst->names[devnr].name,DEVICE_NAME_SIZE); lcldevnam[DEVICE_NAME_SIZE] = 0;
      break;
    }
  }
  return lcldevnam;
}
char *GetStoredDeviceNameFromAlias(char *alias,char *srv,int evnId)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  DASDEV *devlst;
  static char lcldevnam[DEVICE_NAME_SIZE+8];

  lcldevnam[0] = 0; /* default */
  /* get das from srv : */
  for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    if (!strnicmp(srv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
  if (!found) return lcldevnam;
  k--; /* back up ! */
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (devlst=das->devlst; devlst != NULL; devlst=devlst->nxt) if (evnId >= devlst->minTimeValid) break;
    if (devlst == NULL || devlst->alias == NULL) return NULL;
    for (i=0; i<devlst->size; i++) 
      if (!strnicmp(devlst->alias[i].name,alias,DEVICE_NAME_SIZE)) 
        return devlst->names[i].name;
  }
  return NULL;
}
char *GetStoredDeviceNameFromAliasEx(char *evnTrg,char *alias,char *srv,int evnId)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  DASDEV *devlst;
  static char lcldevnam[DEVICE_NAME_SIZE+8];

  lcldevnam[0] = 0; /* default */
  /* get ARB from evnTrg: */
  if (evnTrg != NULL && strlen(evnTrg) > 0 && evnTrg[0] != '#')
  {
    if ((k=GetARBIndex(evnTrg)) < 0) return NULL;
  }
  else
  {
    for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
      if (!strncmp(srv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
    if (!found) return lcldevnam;
    k--; /* back up ! */
  }
  /* get das from srv : */
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (devlst=das->devlst; devlst != NULL; devlst=devlst->nxt) if (evnId >= devlst->minTimeValid) break;
    if (devlst == NULL || devlst->alias == NULL) return NULL;
    for (i=0; i<devlst->size; i++)
      if (!strnicmp(devlst->alias[i].name,alias,DEVICE_NAME_SIZE))
        return devlst->names[i].name;
  }
  return NULL;
}
int GetStoredDeviceNumber(NAME64 *devname,char *srv,int evnId)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  DASDEV *devlst;

  /* get das from srv : */
  for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    if (!strnicmp(srv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
  if (!found) return -non_existent_elem;
  if (devname->name[0] == '#') return atoi(&devname->name[1]);
  k--; /* back up ! */
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (devlst=das->devlst; devlst != NULL; devlst=devlst->nxt) if (evnId >= devlst->minTimeValid) break;
    if (devlst != NULL) 
    {
      for (i=0; i<devlst->size; i++) if (!strnicmp(devlst->names[i].name,devname->name,64)) break;
      if (i < devlst->size) return i;
    }
  }
  return -1;
}
int GetStoredDeviceNumberEx(char *evnTrg,NAME64 *devname,char *srv,int evnId)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  DASDEV *devlst;
  /* get ARB from evnTrg: */
  if (evnTrg != NULL && strlen(evnTrg) > 0 && evnTrg[0] != '#')
  {
    if ((k=GetARBIndex(evnTrg)) < 0) return -1;
  }
  else
  {
    for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
      if (!strnicmp(srv,arbTbl[k].das[j].srvArc,16)) found = -1;
    if (!found) return -non_existent_elem;
    if (devname->name[0] == '#') return atoi(&devname->name[1]);
    k--; /* back up ! */
  }
  /* get das from srv : */
  if (devname->name[0] == '#') return atoi(&devname->name[1]);
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (devlst=das->devlst; devlst != NULL; devlst=devlst->nxt) if (evnId >= devlst->minTimeValid) break;
    if (devlst != NULL)
    {
      for (i=0; i<devlst->size; i++) if (!strnicmp(devlst->names[i].name,devname->name,DEVICE_NAME_SIZE)) break;
      if (i < devlst->size) return i;
    }
    else 
    {
      if (!strnicmp(devname->name,das->dev,DEVICE_NAME_SIZE)) return 0;
    }
  }
  return -1;
}
int GetStoredTriggers(NAME64 *trgs,char *srv,char *prpname,int num)
{ /* i.e. stored 'archive' contexts*/
  int j,k,p = 0;
  DAS *das;
  NAME64 *trg = trgs;

  if (num == 0) num = 256; /* give it something for a temp buffer */
  if (trg == NULL && (trg=(NAME64 *)alloca(num*sizeof(NAME64))) == NULL) return -out_of_local_memory;
  memset(trg,0,num*sizeof(NAME64));
  for (k=0; k<arbTblSiz; k++)
  {
    for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
    {
      das = &arbTbl[k].das[j];
      if (srv != NULL && strlen(srv) > 0 && 
          strncmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
      if (prpname != NULL && strlen(prpname) > 0 && 
          strnicmp(prpname,das->prpArc,PROPERTY_NAME_SIZE)) continue;
      if (p < num) strncpy(trg[p].name,arbTbl[k].trg,DEVICE_NAME_SIZE);
      p++;
      break;
    }
  }
  return p;
} 
int GetStoredProperties(NAME64 *props,char *srv,char *evnTrg,int num)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  NAME64 *prp = props;
  char *c, *ctx=NULL, *psrv=NULL, tmp[128];

  if (srv == NULL) return -argument_list_error;
  if (num == 0) num = 256; /* give it something for a temp buffer */
  if (prp == NULL && (prp=(NAME64 *)alloca(num*sizeof(NAME64))) == NULL) return -out_of_local_memory;
  memset(prp,0,num*sizeof(NAME64));
  memcpy(tmp,srv,128);
  if (tmp[0] == '/')
  {
    ctx = &tmp[1];
    if ((c=strchr(ctx,'/')) == NULL) return -non_existent_server;
    *c++ = 0;
    psrv = c;
  }
  else if ((c=strchr(tmp,'/')) != NULL)
  {
    ctx = tmp;
    *c++ = 0;
    psrv = c;
  }
  else
  {
    psrv = tmp;
  }
  if (evnTrg != NULL && strlen(evnTrg) > 0 && evnTrg[0] != '#')
  { /* get ARB from evnTrg: */
    if ((k=GetARBIndex(evnTrg)) < 0) return k;
  }
  else
  { /* search for a match in the server name  */
    for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    {
      if (ctx != NULL && strnicmp(ctx,arbTbl[k].das[j].ctx,CONTEXT_NAME_SIZE)) continue;
      if (!strnicmp(psrv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
    }
    if (!found) return -non_existent_elem;
    k--; /* back up ! */
  }
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  { /* get das from srv : */
    das = &arbTbl[k].das[j];
    if (ctx != NULL && strnicmp(ctx,arbTbl[k].das[j].ctx,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(psrv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    /* find all the properties that go with it: */
    for (i=0; i<p; i++) if (!strnicmp(prp[i].name,das->prpArc,PROPERTY_NAME_SIZE)) break;
    if (i < p) continue;
    if (strstr(prp[i].name,".HDR") != NULL) continue;
    if (p < num) strncpy(prp[p].name,das->prpArc,PROPERTY_NAME_SIZE);
    p++;
    if (das->acc&MONITOR || das->numberInSeries > 0)
    {
      if (p < num) sprintf(prp[p].name,"%.59s.NSEQ",das->prpArc);
      p++;
    }
    else if (strstr(das->prp,".HIST@") != NULL || strstr(das->prp,".HST@") != NULL)
    {
      if (p < num) sprintf(prp[p].name,"%.59s.NSEQ",das->prpArc);
      p++;
    }

  }
  return p;
} 
char *GetStoredPropertyFromAlias(char *srv,char *alias)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  static char prp[PROPERTY_NAME_SIZE+8];
 
  /* get das from srv : */
  for (k=0; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    if (!strncmp(srv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
  if (!found) return NULL;
  k--; /* back up ! */
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (das->prpAliasLst == NULL) continue;
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (i=0; i<das->nprpAlias; i++)
      if (!strnicmp(das->prpAliasLst[i].name,alias,PROPERTY_NAME_SIZE))
        return das->prp;
  }
  return NULL;
}
char *GetStoredPropertyFromAliasEx(char *evnTrg,char *srv,char *alias)
{
  int i,j,k,p = 0,found = 0;
  DAS *das;
  static char prp[PROPERTY_NAME_SIZE+8];

  /* get ARB from context: */
  if ((k=GetARBIndex(evnTrg)) < 0) return NULL;
  /* get das from srv : */
  for (j=0; j<arbTbl[k].nsteps; j++) if (isReadable(k,j))
  {
    das = &arbTbl[k].das[j];
    if (das->prpAliasLst == NULL) continue;
    if (strnicmp(srv,das->srvArc,EXPORT_NAME_SIZE)) continue;
    for (i=0; i<das->nprpAlias; i++)
      if (!strnicmp(das->prpAliasLst[i].name,alias,PROPERTY_NAME_SIZE))
        return das->prp;
  }
  return NULL;
}
int parseAnnotationAndEvent(char *str,char **cmt,time_t *evn)
{
  static char tmpstr[1032];
  char *c = NULL;
  time_t tot = 0;

  strncpy(tmpstr,str,1024); tmpstr[1024] = 0;

  if ((c=strrchr(tmpstr,'@')) != NULL)
  {
    if (c == str) return illegal_delimiter;
    *c++ = 0;
    if (strlen(c) < 8) return parsing_error;
    *cmt = tmpstr;
  }  
  else if (tmpstr[0] == '[')
  { /* could also be giving the event number this way */
    if ((c=strchr(tmpstr,']')) == NULL) return illegal_delimiter;
    *c++ = 0; /* advance to the comment part */
    *cmt = c;
    c = &tmpstr[1];
  }
  if (c == NULL) return invalid_parameter;
  if (strlen(c) < 8) return parsing_error;
  if ((tot = atoi(c)) < 100)
  { /* not UTC format */
    tot = getTimeStampFromLogEntry(c);
  }
  *evn = tot;
  return 0;
}
