/*
 * File arctools.h
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _ARCTOOLS_H_
#define _ARCTOOLS_H_

#include "tinetype.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define ARCHIVE_STORAGE_VERSION "5.00"

#define VERSION_LENGTH 16
#define IMAGE_REF_SIZE 16
#ifndef BYTE
# define BYTE unsigned char
#endif

#define MEM_POOL_SIZE (1000 * sizeof(NAME64))
extern BYTE qryMemPool[MEM_POOL_SIZE];
#define N1toN2(n2,n1,n,sz) for (i=0; i<(int)(n); i++) strncpy((n2)[i].name,(n1)[i].name,sz);

#define MARCHSRV_EVENTCODE 0x1000

#define READ     0x01
#define WRITE    0x02
#define MATCH    0x04
#define WAIT     0x08
#define SETNUM   0x10   /* acquire output data size for subsequent call */
#define PREALLOC 0x20
#define FORWARD  0x40   /* acquire input data for subsequent calls */
#define SERIES   0x80
#define MONITOR  0x100
#define PERSIST  0x200  /* applicable in combo with SETNUM */
#define SETDEV   0x400  /* acquire device list and use for subsequent calls */
#define UNSETDEV 0x800  /* clear the setting of the acquired device list */
#define DISABLE  0x4000 /* disable this archive trigger step */

#define PREALLOCSIZE 0x40000
extern BYTE *PreallocatedDataBuffer;
extern int PreallocatedDataBufferSize;

#define ARCH_BUF_SIZE 0x50000
#define CMT_SIZE 1024
extern int archDebug;

#define MAXARCHIVES 256

typedef struct
{
  UINT32 version;
  UINT32 eventNumber;
  UINT32 eventCode;
  UINT32 reserved;
  char comment[4][64];
} AESnotify;

typedef struct
{
  UINT32 version;
  UINT32 eventNumber;
  UINT32 eventCode;
  UINT32 reserved;
  char  deviceName[16];
} AEScontext;

#define AESnotifyStructSize ((sizeof(UINT32)*4)+4*64)
#define AEScontextStructSize ((sizeof(UINT32)*4)+16)

typedef struct AEStag
{
  UINT32 eventNumber;
  UINT32 eventCode;
  char comment[4][64];
  BYTE dsMask[32];
} AES;

#define AESStructSize ((sizeof(UINT32)*2)+(4*64)+(32))

#define MAKECODE(id,cd) (((id)<<16)|(cd))
#define BITSinBYTE 8

/* Note: there was once a scheme to maintain a global event number (utc)
 * provided by an SITE event server and an event code identifying the 
 * category of an event participant.
 * Now: the event code is never used and a specific event is simply 
 * identified by its event number (utc of the event trigger) and the
 * event trigger (name).
 */

#define DEFAULT_MONITOR_DURATION 600
#define DEFAULT_MONITOR_MODE CM_TIMER
#define DEFAULT_MONITOR_RATE 1000


typedef struct DasDevStruct
{
  int size;
  NAME64 *names;
  NAME64 *alias;
  time_t minTimeValid;
  struct DasDevStruct *nxt;
} DASDEV;

typedef struct DevLstTblStruct
{
	DASDEV *devlst;
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char srvArc[EXPORT_NAME_SIZE];
	int hasAcquiredCurrentList;
	struct DevLstTblStruct *nxt;
} DEVLSTTBL;
extern DEVLSTTBL *devLstTbl;

/* device name signal to force ACQ for ALL devices */
#define DEVICE_ALL "__ALL_DEVICES__"
/* device name signal to indicate ignoring the stored device name in readback */
#define DEVICE_IGNORE "__IGNORE_DEVICES__"

/* targeted device list structure */
#define MAX_DEVTGTLST_SIZE 32
typedef struct
{
  NAME64 lst[MAX_DEVTGTLST_SIZE]; /* the list */
  int siz;     /* the list size */
  int idx;     /* current list index */
} DEVTGTLST;

typedef struct DAStag
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char srvArc[EXPORT_NAME_SIZE];
  char prpArc[PROPERTY_NAME_SIZE];
  char infile[16];
  DEVTGTLST *devTgtLst;
  short EqpNoStart;
  short EqpNoStop;
  DTYPE dout;
  DTYPE din;
  short dinlock;
  short freeMemoryOnExit;
  short acc;
  short wait;
  short mode;
  short timeout;
  int value;
  float scale;
  float shift;
  int numberToAcquire;
  int seriesIndex;
  int numberInSeries;
  int monitorRate;
  int monitorTime;
  int isWaiting;
  int isAcqHist;
  DASDEV *devlst;
  int nprpAlias;
  NAME64 *prpAliasLst;
  int nxDevices;
  NAME64 *xDevLst;
} DAS; /* Data Access Structure */

#define APPLY_RANGE_NONE 0
#define APPLY_TIMESTAMP_RANGE 1
#define APPLY_SYSSTAMP_RANGE 2
#define APPLY_USRSTAMP_RANGE 3
#define RANGE_OFFSET 5

typedef struct EvnQ
{
  time_t eventtime; /**< designated time of event */
  int  applyRange; /**< signal to use range values in local history access */
  double tsStart; /**< time range start */
  double tsStop; /**< time range stop */
  int  sstStart; /**< sys stamp range start */
  int  sstStop; /**< sys stamp range stop */
  int  maxInRange; /**< maximum number items in range */
  struct EvnQ *nxt;
} EVNQ;
typedef struct ARBtag
{
  FILE *fp; /**< handle to ongoing trigger file */
  char trg[DEVICE_NAME_SIZE]; /**< trigger name */
  char ext[16]; /**< trigger event file extension (default = 1st 16 chars of trigger name) */
  char desc[64]; /**< event description */
  int  step; /**< current step mailbox */
  int  nsteps; /**< number of steps to fulfill for this trigger type */
  int  depthInMonths; /**< storage depth in months of unviewed event files */
  int  maxq; /**< maximum number of events which can be queued (0 => no queue allowed) */
  int  nevnq; /**< number of triggers in event queue */
  EVNQ *evnq; /**< current event queue (if any) */
  double interim; /**< interim timestamp of data taking */
  time_t start; /**< start time of event data taking */
  time_t eventtime; /**< designated time of event */
  time_t tot; /**< time of trigger */
  int  resultPending; /**< companion flag to event trigger */
  int  recordEvents; /**< binary flag indicating whether the event type should notify the event server when triggered */
  int  recordPending; /**< companion flag to recordEvents */
  int  aborted; /**< on-going active event has been aborted (always reset at start of new trigger) */
  int  status; /**< current status of this ARB activity */
  int  isMonitoring; /**< binary flag indicating a monitor type acquisition */
  int  applyRange; /**< signal to use range values in local history access */
  double tsStart; /**< time range start */
  double tsStop; /**< time range stop */
  int  sstStart; /**< sys stamp range start */
  int  sstStop; /**< sys stamp range stop */
  int  maxInRange; /**< maximum number items in range */
  int almCode; /**< the alarm code to use in SetDeviceAlarm() */
  int isAlarm; /**< indicator of whether to SetDeviceAlarm() or not */
  char almTrgUsr[USERNAME_SIZE]; /**< allowed 'trigger setter' for setting an alarm */
  DAS  *das; /**< Data Access Structure item list (1 item for each of nsteps) */
# ifdef MULTITHREADED  
  THRHANDLE tid; /**< thread handle if GetAndStore is in play */
# endif
} ARB; /**< Archive Reference Block */

extern ARB *arbTbl;
extern int arbTblSiz;
extern NAME64 *arbTrgList;

typedef struct
{
  char trg[DEVICE_NAME_SIZE];
  char ext[16];
  char desc[64];
  int  depthInMonths;
  short recordEvents;
  short maxQueueSize;
} ARBQS;

typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char srvArc[EXPORT_NAME_SIZE];
  char prpArc[PROPERTY_NAME_SIZE];
  char infile[16];
  char datTag[TAG_NAME_SIZE];
  char datTagIn[TAG_NAME_SIZE];
  UINT32 datSiz;
  UINT32 datFmt;
  UINT32 datFmtIn;
  UINT32 access;
  UINT32 wait;
  UINT32 timeout;
  UINT32 value;
  UINT32 mode;
  UINT32 monitorRate; /* milliseconds */
  UINT32 monitorTime; /* seconds */
  float scale;
  float shift;
} DASQS;

typedef struct
{
   char srv[EXPORT_NAME_SIZE];
   char dev[DEVICE_NAME_SIZE];
   char prp[PROPERTY_NAME_SIZE];
   char dsc[PROPERTY_DESC_SIZE];
   UINT16 format;
   UINT16 arrayType;
   UINT32 arraySize;
   UINT32 preTrigger;
   char unitsY[UNITS_SIZE];
   char unitsX[UNITS_SIZE];
   float rangeMinX;
   float rangeMaxX;
   float scaleMinY;
   float scaleMaxY;
} PHS;   /* local Property Header Structure  */

typedef struct LHStag
{
  PHS hdr;
  struct LHStag *next;
} LHS;   /* Local Header Structure  */
int dumpPHS(PHS *phs,int nitems);

#define TRIGGER_FILES_SUBDIR "triggerFiles"
#define SERVER_FILES_SUBDIR "serverFiles"
#define DEVICES_FILE_NAME "devices.csv"
#define XDEVICES_FILE_NAME "xdevices.csv"
#define HEADER_FILE_NAME "header.csv"
#define ARCHIVE_LIST_FILE "pmArchiveList.csv"
#define ALARM_LIST_FILE "pmAlarmList.csv"
#ifdef WIN32
# define stdArchiveDirectory "..\\CACHE"
#else
# define stdArchiveDirectory "../CACHE"
#endif
extern char *archiveListFile;
extern char *alarmListFile;
extern int arbTblBusy;

FILE *createAndOpen(char *fname,char *attr);
void freeArbTbl(void);
int stopARB(ARB *arb);
char *GetStoredDeviceName(int devnr, char *srvname,int evnId);
char *GetStoredDeviceNameEx(char *evnTrg,int devnr, char *srvname,int evnId);
int GetStoredDeviceNumber(NAME64 *devnam, char *srvname,int evnId);
int GetStoredDeviceNumberEx(char *evnTrg,NAME64 *devnam, char *srvname,int evnId);
char *GetStoredDeviceNameFromAlias(char *alias,char *tagname,int evnId);
char *GetStoredDeviceNameFromAliasEx(char *evnTrg,char *alias,char *tagname,int evnId);
char *GetStoredPropertyFromAlias(char *srvname,char *alias);
char *GetStoredPropertyFromAliasEx(char *evnTrg,char *srvname,char *alias);
int GetStoredProperties(NAME64 *props,char *srvname,char *evnTrg,int num);
int GetStoredTriggers(NAME64 *ctxs,char *srvname,char *prpname,int num);
void registerStructs(void);
int GetLocalDeviceServers(char *evnTrg,NAME32 *names,UINT32 start,UINT32 stop,int event,int num);
int GetLocalDeviceServersEx(char *trigger, NAME64 *names, UINT32 start, UINT32 stop, int event, int num);
int GetLocalnumDeviceServers(void);
int GetNumLocalArchivedEvents(char *trigger, UINT32 starttime, UINT32 stoptime, UINT32 eventNumber);
int GetLocalArchivedEvents(char *trigger, UINT32 starttime, UINT32 stoptime, UINT32 *evnlst, UINT32 num, int *sts);
int GetLocalDeviceArchivedEvents(char *trigger, char *srv, char *dev, UINT32 starttime, UINT32 stoptime, AES *aes, UINT32 num, int reverseSort, int *sts);
int GetLocalDeviceArchivedEventList(char *trigger, char *srv, char *dev, UINT32 starttime, UINT32 stoptime, UINT32 *evnlst, UINT32 num, int reverseSort, int *sts);
int GetNumLocalDeviceArchivedEvents(char *devServer, char *devName, UINT32 starttime, UINT32 stoptime);
int GetLocalDevices(NAME64 *names, char *devSrvName, char *trigger, UINT32 evnTime, int num);
int GetLocalDeviceProperties(NAME64 *props, char *devSrvName, char *trigger, int num);
int GetLocalTriggers(NAME64 *names, char *srv, char *prp, int num);
int GetTriggerTable(int index,ARBQS *arb,int arbBufSize);
int GetDASTable(int index,DASQS *das,int dasBufSize);
int AppendTriggerTable(ARBQS *arb,int size);
int WriteDASTable(char *trg,DASQS *das,int size);
int parseAnnotationAndEvent(char *str,char **cmt,time_t *evn);
int getDASInputData(char *fn,DTYPE *din);
int putDASInputData(char *fn,DTYPE *din);
int getDASInputFileNames(DTYPE *dout);
int arbHasSrv(ARB *arb,const char *ctx,const char *srv);
int GetTriggersForServer(const char *ctx,const char *srv,const char *prp,NAME64 *lst,int lstSize);
int getArchiveServers(int evnNr,char *evnTrg,char *ctx,NAME32 *srvs,int *nsrvs);
int getArchiveDevices(int evnNr,char *evnTrg,char *ctx,char *srv,char *prp,NAME64 *devs,int *ndevs);
int getArchiveProperties(int evnNr,char *evnTrg,char *ctx,char *srv,char *dev,NAME64 *prps,int *nprps);
int getSetNumber(char *trigger);
int readArchiveSets(char *eqm);
int getArchiveSets(char *fn);
int getTriggerAlarms(char *fn, char *eqm);
int RefreshDeviceLists(ARB *arb);
int FillPropertyAliasList(ARB *arb);
int FillDeviceLists(ARB *arb);
#define isReadable(a1,a2) (arbTbl[a1].das != NULL && (arbTbl[a1].das[a2].acc&(READ|SERIES|MONITOR)))

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _ARCTOOLS_H_ */
