/*
 * File mexeqm.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include "tine.h"
#include "fecerr.h"
#include "archive.h"
#include "statebase.h"
#include "fecdef.h"
#if defined(UNIX)
# include <dirent.h>
#elif defined(WIN32)
# include <direct.h>
#endif

extern TINE_EXPORT int gScanForNameLists;
int gIsGlobalsServer;
int loadStateInfo(void);
int reacquireStaticDataSet(int i);
void AppendProblemsLog(char *str);
void updateStateInfo(int i,int j,int cc);
void fixFecRepository(void);
int applyMcstAddr(int family, INADR *tgt, INADR *pat);
int overloadEqpmod(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
extern int gAppendProperyAliasNamesToQueries;
extern int MaxRPCAllocation;
DAQRecord *daqTbl;
int g_isTerminating = FALSE;
int checkData(int n,int isPastFilter,int isSched);
void flcopy(long  *dst, long *src, int num);
int getNumActiveRecords(char *fname);
void outputData(void);
void addPartnerToAcls(void);
int pushSelfEntry(DTYPE *din,int idx);
int wrtFlg,archdebug=0;
int totalNrRecords = 0;
UINT32 gMininumFreeDiskSpace = 200000;

char thisFecName[FEC_NAME_SIZE+4] = "";
char thisContext[CONTEXT_NAME_SIZE+4] = "";
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE+4] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
char thisServerName[EXPORT_NAME_SIZE] = MEX_EXPORTNAME;
char thisFailoverMasterName[EXPORT_NAME_SIZE] = "";
char thisFailoverSlaveMasterName[EXPORT_NAME_SIZE] = "";
UINT16 thisPortOffset = 0;
int startupInputsGiven = FALSE;

#define WRITE     -1
#define READ       1
#define MAXFILES   100

/* state variables: */
int ramping = 0, running = 0, beam = 0, lumi = 0;
UINT16 Pramping = 0, Eramping = 0, Pbeam = 0, Ebeam = 0;
UINT16 Prunning = 0, Erunning = 0;
int DclEramping = 0,DclPramping = 0,DclLumi = 0,DclEInj = 0,DclPInj = 0, DclRun = 0;

int checkStates = FALSE;

typedef struct
{
  int id;
  int startPending;
} LINKSTRCT;
LINKSTRCT *recordLinkIdSts;

void outputData(void)
{
  int i, istrace;
  char *key;
  for (i=0; i<totalNrRecords; i++)
  {
    istrace = isTraceLink(recordLinkIdSts[i].id);
    if (istrace) key = conTbl[recordLinkIdSts[i].id]->key;
    if (daqTbl[i].archPending) 
    {
      if (istrace) dbglog("write %.192s (record %d) to archive",key,daqTbl[i].rec);
      WriteRecord(&daqTbl[i]);
    }
    else
    {
      if (istrace) dbglog("no archive pending for %.192s (record %d)",key,daqTbl[i].rec);
    }
  }
}

#define logerr(txt) { perror("open"); strcpy(s,txt); goto error; }

#define FAILOVERS_FILENAME "failovers.csv"
KeywordFailover *kwrdFailover=NULL;
int nfailovers = 0;

KeywordFailover *getKeywordFailover(char *key)
{
  KeywordFailover *kw;
  int i;
  for (kw=kwrdFailover; kw != NULL; kw=kw->nxt)
  {
    if (strnicmp(key,kw->keyword,PROPERTY_NAME_SIZE)) continue;
    return kw;
  }
  if ((kw=(KeywordFailover *)calloc(1,sizeof(KeywordFailover))) == NULL)
  {
    feclog("create keyword failover entry failed: out of memory");
    return NULL;
  }
  strncpy(kw->keyword,key,PROPERTY_NAME_SIZE);
  for (i=0; i<3;i++) kw->foSrc[i].recNr = -1;
  kw->nxt = kwrdFailover;
  kwrdFailover = kw;
  nfailovers++;
  return kwrdFailover;
}
int addFailoverSource(KeywordFailover *kwf,char *src,int idx)
{
  if (idx < 0 || idx > 2) return invalid_index;
  strncpy(kwf->foSrc[idx].name,src,PROPERTY_NAME_SIZE);
  return 0;
}

void freeFailoverInformation(void)
{
  KeywordFailover *kw, *nxtkw;
  for (kw=kwrdFailover; kw != NULL; kw=nxtkw)
  {
    nxtkw = kw->nxt;
    if (kw->data != NULL) free(kw->data);
    free(kw);
  }
  kwrdFailover = NULL;
}

int getFailoverInformation(char *fname)
{
  char fn[80];
  char str[256],scratch[128];
  int done = 0,i,cc = 0,idx;
  int kwd_col=-1,src_col=-1,mtc_col=-1;
  FILE *fp = NULL;
  KeywordFailover *kw;
 
  sprintf(fn,"%.32s",fname);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL; )   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((kwd_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
      if ((src_col=findcol(str,"SOURCE")) < 0) ccerr(no_such_line);
      if ((mtc_col=findcol(str,"METRIC")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(mtc_col,str),16);
    idx = atoi(scratch);
    if (idx < 0 || idx > 2) continue;
    strncpy(scratch,colptr(kwd_col,str),64);
    if ((kw=getKeywordFailover(scratch)) == NULL) return out_of_local_memory;
    strncpy(scratch,colptr(src_col,str),64);
    addFailoverSource(kw,scratch,idx);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
char *dataToString(BYTE *dbuf,int fmt,int siz)
{
  static char sbuf[256];
  int i,p=0;
  sbuf[0] = 0;
  for (i=0; i<siz; i++)
  {
    switch (fmt)
    {
      case CF_INT16:
        sprintf(&sbuf[p],"%d ",((short *)dbuf)[i]);
        break;
      case CF_INT32:
        sprintf(&sbuf[p],"%d ",((int *)dbuf)[i]);
        break;
      case CF_FLOAT:
        sprintf(&sbuf[p],"%g ",((float *)dbuf)[i]);
        break;
      case CF_DOUBLE:
        sprintf(&sbuf[p],"%g ",((double *)dbuf)[i]);
        break;
      case CF_TEXT:
        sprintf(sbuf,"%.255s",dbuf); sbuf[255] = 0;
        return sbuf;
      case CF_NAME16:
        sprintf(&sbuf[p],"%.16s ",(char *)dbuf+i*16);
        break;
      case CF_NAME32:
        sprintf(&sbuf[p],"%.32s ",(char *)dbuf+i*32);
        break;
      case CF_NAME64:
        sprintf(&sbuf[p],"%.64s ",(char *)dbuf+i*64);
        break;
      default:
        sprintf(sbuf,"format %.16s",GetFormatAsString(fmt));
        return sbuf;
    }
    p += (int)strlen(sbuf);
    if (p > 192) break;
  }
  return sbuf;
}
int dumpFailoverStatus(int argc,char *argv[])
{
  KeywordFailover *kw;
  int i,p,n;
  argc = argc; argv = argv;
  for (kw=kwrdFailover; kw != NULL; kw=kw->nxt)
  {
    ttyoutput("failover %.64s : %s",kw->keyword,dataToString(kw->data,kw->fmt,kw->len));
    for (i=0; i<3; i++)
    {
      if (kw->foSrc[i].recNr < 0) continue;
      p = kw->foSrc[i].recNr;
      n = kw->foSrc[i].prmNr;
      ttyoutput("\t  %.64s <%d> : %s",
        daqTbl[p].daqParam[n].param,daqTbl[p].daqStatus,
        dataToString(daqTbl[p].data+daqTbl[p].daqParam[n].offset,daqTbl[p].daqParam[n].fmt,daqTbl[p].daqParam[n].size));
    }
  }
  return 0;
}

/*============= General archiving globals =========================== */

#define GLOBALS_FILENAME "nglobals.csv"
int registerGlobals = 0, brdcstRegistered = FALSE;
int globalsServer = FALSE, globalsWatchdog = FALSE;
int requireNetworkAccess = FALSE;
int archPending = 0;
char globalsMcastAddr[40] = "";
char adbFileName[64] = "";

FILE *wfp, *rfp;
int Hwfp;
void checkDateChange(void);
void logRestart(void);
int readNETMEX(char *fname,int FECoffset, int num, time_t *lasttime);
int mexeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
int GlobalsRegistrationEx(void);

char sbuf[20];
char timebuf[9], datebuf[9];

int fileNo = 0;

#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  if (nadmins == 0) return TRUE;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];
 
  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

GlobalListStruct *globalList;
int ngloballist = 0;

void cbSendGlobal(int i,int cc)
{
}
int getMinFreeDiskSpace(char *fname)
{
  char fn[80];
  char str[256],scratch[128];
  int done = 0,i,cc = 0;
  int col;
  FILE *fp = NULL;
 
  sprintf(fn,"%s%.16s",ArchiveDBpath,fname);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL; )   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((col=findcol(str,"DISK_SPACE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(col,str),64);
    gMininumFreeDiskSpace = strtol(scratch,NULL,0);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int putNetGlobalList(char *fname,NAME64 *keys,int nkeys)
{
  char fn[132];
  int i,cc = 0;
  FILE *fp = NULL;
  sprintf(fn,"%s",fname);
  if ((fp=fopen(fn,"wt")) == NULL) ccerr(no_such_file);
  fprintf(fp,"KEYWORD\n");
  for (i=0; i<nkeys; i++)
  {
    if (strlen(keys[i].name) == 0) continue;
    fprintf(fp,"%.64s\n",keys[i].name);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int getNetGlobalList(char *fname)
{
  char fn[132];
  char str[256],scratch[128];
  int done = 0,i,cc = 0;
  int ctx_col,key_col,fmt_col,siz_col,pmax_col,pmin_col,tol_col,val_col;
  FILE *fp = NULL;
 
  if (globalList != NULL) 
  {
    free(globalList);
    globalList = NULL;
  }
  sprintf(fn,"%s",fname);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    i++;
  }
  if ((globalList=(GlobalListStruct *)calloc(i,sizeof(GlobalListStruct))) == NULL) ccerr(out_of_server_memory);
  rewind(fp);
  for (i=0; csvGetLine(str,255,fp) != NULL; )   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      ctx_col = findcol(str,"CONTEXT");
      if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
      fmt_col = findcol(str,"FORMAT");
      siz_col = findcol(str,"SIZE");
      pmax_col = findcol(str,"PERIOD_MAX");
      pmin_col = findcol(str,"PERIOD_MIN");
      tol_col = findcol(str,"TOLERANCE");
      val_col = findcol(str,"VALUE");
      done = 1;
      continue;
    }
    if (ctx_col >= 0) strncpy(globalList[i].context,colptr(ctx_col,str),CONTEXT_NAME_SIZE);
    strncpy(globalList[i].keyword,colptr(key_col,str),KEYWORD_NAME_SIZE);
    if (siz_col >= 0)
    {
      strncpy(scratch,colptr(siz_col,str),16);
      globalList[i].size = atoi(scratch);
    }
    else
    {
      globalList[i].size = 1;
    }
    if (fmt_col >= 0)
    {
      strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
      globalList[i].format = ftoi(scratch);
    }
    else
    {
      globalList[i].format = CF_FLOAT;
    }
    if (pmax_col >= 0)
    {
      strncpy(scratch,colptr(pmax_col,str),16);
      globalList[i].maxPeriod = atoi(scratch);
    }
    else
    {
      globalList[i].maxPeriod = 1000;
    }
    if (pmin_col >= 0)
    {
      strncpy(scratch,colptr(pmin_col,str),16);
      globalList[i].minPeriod = atoi(scratch);
    }
    else
    {
      globalList[i].minPeriod = 1000;
    }
    if (tol_col >= 0)
    {
      strncpy(scratch,colptr(tol_col,str),16);
      globalList[i].tolerance = strtod(scratch,NULL);
    }
    else
    {
      globalList[i].tolerance = 0;
    }
    if (val_col >= 0)
    {
      int dsiz = globalList[i].size * fmtsizeof(globalList[i].format);
      if (dsiz > 0)
      {
        globalList[i].refdata = (BYTE *)calloc(dsiz,1);
        if (globalList[i].refdata != NULL)
        { /* put default data into index 0 */
          strncpy(scratch,colptr(val_col,str),128);
          stringToDataValue(globalList[i].refdata,scratch,0,globalList[i].format,NULL,globalList[i].size,FALSE);
        }
      }
    }
    i++;
  }
  ngloballist = i;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
void ensureGlobalsProperties(void)
{
  int i, id;
  char desc[128];
  DTYPE d;
  DTYPEZERO(d);
  for (i=0; i<ngloballist; i++)
  {
    if (GetPropertyId(MEX_EQMNAME,globalList[i].keyword) < 0)
    { /* not yet registered => not in normal acquisition database ! */
      sprintf(desc,"global keyword %.64s",globalList[i].keyword);
      d.dFormat = globalList[i].format;
      d.dArrayLength = globalList[i].size;
      d.data.bptr = globalList[i].refdata;
      if ((id=RegisterPropertyInformation(MEX_EQMNAME,globalList[i].keyword,
           &d,NULL,CA_READ,AT_SINGLE,0,desc,-1,NULL)) < 0)
      {
        feclog("Couldn't register globals prp %.64s: %.32s",globalList[i].keyword,erlst[-id]);
      }
      if (registerGlobals) 
      { 
        GlobalsRegistrationEx();
        id = sendNetGlobal(globalList[i].keyword,&d,cbSendGlobal,1000,1000,0);
        feclog("register static global %s : %s",globalList[i].keyword,id < 0 ? erlst[-id] : "success");
      }
    }
  }
}
NAME16 redirfile[10];
NAME64 redirsrv[10];
int nredirfiles = 0;
int getRedirInfo(void)
{
  FILE *fp;
  char str[256];
  int i=0,done = 0,cc = 0;
  int database_col,name_col;

  if ((fp = fopen("redir.csv","rt")) == NULL) return -1;
  while (csvGetLine(str,255,fp) != NULL )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(str,"SERVERNAME")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(redirfile[i].name,colptr(database_col,str),16);
    strncpy(redirsrv[i].name,colptr(name_col,str),64);
    if (++i >= 10) break;
  }
err:
  return i;
}

TSRVINFO srvinfo;
int getStartupInfo(void)
{
  FILE *fp=NULL;
  char str[256],scratch[170],*c;
  int i,done = 0,cc = 0;
  int database_col,method_col,state_col,addr_col;
  INADR ipadr, myadr;

  memcpy(&myadr,getmyipaddr(),gInAdrLen);
  inet_pton(gAfInet,gtMCastAddr,&ipadr);
  applyMcstAddr(gAfInet,&ipadr,&myadr);

  if ((fp=fopen("database.csv","rt")) == NULL && 
      (fp=fopen("startup.csv","rt")) == NULL)
  {
    if (strlen(adbFileName) == 0)
    {
      if (globalsServer || globalsWatchdog)
      {
        strcpy(adbFileName,"dataglob.csv");
      }
      else if (strlen(thisContext) > 0)
      { /* if context known : append it to default name */
        sprintf(adbFileName,"data%.32s.csv",thisContext);
        strlwr(adbFileName);
      }
      else
      {
        strncpy(adbFileName,DEFAULT_ADB_FILENAME,64);
      }
    }
    feclog("did not find data base file name : using database %.64s",adbFileName);
    strncpy(srvinfo.dbfile[0],adbFileName,64);
    srvinfo.ndbfiles = 1;
    globalsWatchdog = FALSE;
    if (strlen(globalsMcastAddr) == 0)
    { /* undetermined */
      inet_ntop(gAfInet,&ipadr,globalsMcastAddr,40);
    }
    return 0;
  }
  while (csvGetLine(str,255,fp) != NULL )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      method_col = findcol(str,"METHOD");
      state_col = findcol(str,"CHECKSTATE");
      addr_col = findcol(str,"ADDRESS");
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(database_col,str),170);
    for (i=0,c=strtok(scratch," "); i<10 && c != NULL; i++,c=strtok(NULL," "))
    {
      strncpy(srvinfo.dbfile[i],c,64);
    }
    srvinfo.ndbfiles = i;
    globalsWatchdog = FALSE;
    if (method_col >= 0)
    {
      strncpy(scratch,colptr(method_col,str),16);
      globalsWatchdog = !strcmp(scratch,"PASSIVE");
    }
    if (state_col >= 0)
    {
      strncpy(scratch,colptr(state_col,str),16);
      strupr(scratch);
      if (strstr(scratch,"TRUE") != NULL) checkStates = TRUE;
    }
    if (addr_col >= 0)
    {
      strncpy(globalsMcastAddr,colptr(addr_col,str),40);
    }
    else
    {
      fixFecRepository();
      inet_ntop(gAfInet,&ipadr,globalsMcastAddr,40);
    }
    done = 2;
    break;
  }      
err:
  if (fp != NULL) fclose(fp);
  return done == 2 ? 0 : -1;
}
void AppendProblemsLog(char *str)
{
  time_t t = time(NULL);
  char s[256];
  FILE *fp = NULL;

  if ((fp=fopen("problems.log","a")) == NULL) logerr("LOG: open");
  s[0] = 0;
  strncpy(s,str,200);
  if (s[strlen(s)-1] == '\n') s[strlen(s)-1] = 0;
  fprintf(fp,"%s : %s",s,ctime(&t));

error: 
  if (fp) fclose(fp);
  return;
}
void flcopy(long *dst, long *src, int num)
{
  while (num--) *dst++ = *src++;
  return;
}
void wrCopy(UINT16 *src,double *dst, int nwords)
{
  while (nwords--) *dst++ = *src++;
  return;
}
extern UINT32 TCPcount,UDPcount,SPXcount,IPXcount;
extern int nconsumer, ncontract;

extern NAME16 allowedNullStates[];
extern int nAllowedNullStates;

int pneedsToDeclare = FALSE;
int eneedsToDeclare = FALSE;
int p_stateindex = 0;
int e_stateindex = 0;

#define NULLVALUE 0.01
#define GRACE_PERIOD 600

int DeclareState(int stateindex, int typeindex)
{
  int i, cc=link_not_open, retries = 3;
  char dn[256];
  static time_t lastdecl = 0;
  time_t t = time(NULL);
  
  sprintf(dn,"%s/%s",gTargetStateServer,
    typeindex == 0 ? eDeclaredProcedures[stateindex].trigstr : 
                     pDeclaredProcedures[stateindex].trigstr );

  for (i=0; i<retries && cc != 0; i++)
  {
    cc = ExecLinkEx(dn,"STATECHANGE",NULL,NULL,CA_WRITE,400);
  }

  AppendHistoryLog("declare state %s : %s",dn,erlst[cc]);
  if (cc == 0)
  {
    lastdecl = t;
  }
  return cc;
}
void checkStateInfo(void)
{
  int i,j,k,nullcnt;
  static int done = 0;
  static int nparam;
  static int p_lastindex = 0, e_lastindex = 0;
  
  /* check null state consistency */
  if (!done)
  {
    done = TRUE;
    if (nCheckedStates == 0) nparam = 0;
    nparam = (nCheckedStates == 0) ? 0 : nCheckedParameters/nCheckedStates;
    for (k=0; k<nCheckedParameters; k++)
    {
      for (i=0; i<nrecords; i++) for (j=0; j<daqTbl[i].nparam; j++)
        if (!stricmp(daqTbl[i].daqParam[j].param,checkedParameters[k].name))
      {
        checkedParameters[k].ptr = daqTbl[i].data + daqTbl[i].daqParam[j].offset;
        checkedParameters[k].fmt = daqTbl[i].daqParam[j].fmt;
        checkedParameters[k].fecindex = i;
        j = daqTbl[i].nparam; i = nrecords; /* found it, get out of inner loops .. */
        break;
      }
    }
    for (k=0; k<nCheckedStates; k++)
    {
      for (i=0; i<nrecords; i++) for (j=0; j<daqTbl[i].nparam; j++)
        if (!stricmp(daqTbl[i].daqParam[j].param,checkedStates[k].name))
      {
        checkedStates[k].ptr = daqTbl[i].data + daqTbl[i].daqParam[j].offset;
        checkedStates[k].fecindex = i; 
        j = daqTbl[i].nparam; i = nrecords; /* found it, get out of inner loops .. */
        break;
      }
    }
  }
  for (i=0; i<nCheckedStates; i++)
  {
    if (daqTbl[checkedStates[i].fecindex].daqStatus != 0) continue;
    if (checkedStates[i].ptr == NULL) continue;
    for (j=0; j<nAllowedNullStates; j++)
    {
      if (!strncmp((char *)checkedStates[i].ptr,allowedNullStates[j].name,16)) break;
    }
    if (j < nAllowedNullStates) continue;  /* I'm in an allowed null state */
    nullcnt = 0;
    for (j=i*nparam; j<i*nparam+nparam; j++)
    {
      if (daqTbl[checkedParameters[j].fecindex].daqStatus != 0) continue;
      switch (checkedParameters[j].fmt)
      {
        case CF_SHORT:
          if (*((short *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        case CF_LONG:
          if (*((SINT32 *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        case CF_FLOAT:
          if (*((float *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        case CF_DOUBLE:
          if (*((double *)checkedParameters[j].ptr) <= NULLVALUE) nullcnt++;
          break;
        default: break;
      }
    }
    if (nCheckedStates > 0 && nullcnt == nparam)
    {
      DeclareState(0,checkedStates[i].typeindex); /* set to undefined */
    }
  }

  /* declare states from inferred machine values */
  if (p_stateindex != p_lastindex || pneedsToDeclare)
  {
    p_lastindex = p_stateindex;
    DeclareState(p_stateindex,1);
    pneedsToDeclare = FALSE;
  }
  if (e_stateindex != e_lastindex || eneedsToDeclare)
  {
    e_lastindex = e_stateindex;
    DeclareState(e_stateindex,0);
    eneedsToDeclare = FALSE;
  }
}

void checkFailoverData(KeywordFailover *kw)
{
  int i, k, n, dptr, dsiz;
  for (k=0; k<3; k++) 
  { 
    i = kw->foSrc[k].recNr; n = kw->foSrc[k].prmNr;
    if (i < 0) continue;
    if (daqTbl[i].daqStatus != 0) continue;
    dptr = daqTbl[i].daqParam[n].offset;
    dsiz = kw->len * fmtsizeof(kw->fmt);
    memcpy(kw->data,(BYTE *)&daqTbl[i].data[dptr],dsiz);
    kw->dts = daqTbl[i].tstamp;     
    break;
  }

}
void checkFailoverList(void)
{
  KeywordFailover *kw;

  for (kw=kwrdFailover; kw != NULL; kw=kw->nxt)
  {
    checkFailoverData(kw);
  }
}

void mexbkg(void)
{
  static int once = FALSE;
  ClearAlarm(daqEqmName,0);
  checkDateChange(); /* check up on things */
  checkFailoverList();
  if (archPending) outputData();
  if (checkStates) checkStateInfo();
  if (!once)
  {
    once = TRUE;
    addPartnerToAcls();
  }
  return;
}
void addOffsetToData(char *data,DAQParameter *f)
{
  int i;
  char *dptr;

  dptr = (char *)(data + f->offset);
  switch (f->fmt)
  {
    case CF_SHORT:
      for (i=0; i<f->size; i++) ((short *)dptr)[i] += (short)f->scaleOffset;
      return;
    case CF_LONG:
      for (i=0; i<f->size; i++) ((UINT32 *)dptr)[i] += (UINT32)f->scaleOffset;
      return;
    case CF_FLOAT:
      for (i=0; i<f->size; i++) ((float *)dptr)[i] += (float)f->scaleOffset;
      return;
    case CF_DOUBLE:
      for (i=0; i<f->size; i++) ((double *)dptr)[i] += (double)f->scaleOffset;
      return;
  }
  return;
}
void scaleFactorToData(char *data,DAQParameter *f)
{
  int i;
  char *dptr;

  dptr = (char *)(data + f->offset);
  switch (f->fmt)
  {
    case CF_SHORT:
      for (i=0; i<f->size; i++) ((short *)dptr)[i] *= (short)f->scaleFactor;
      return;
    case CF_LONG:
      for (i=0; i<f->size; i++) ((UINT32 *)dptr)[i] *= (UINT32)f->scaleFactor;
      return;
    case CF_FLOAT:
      for (i=0; i<f->size; i++) ((float *)dptr)[i] *= (float)f->scaleFactor;
      return;
    case CF_DOUBLE:
      for (i=0; i<f->size; i++) ((double *)dptr)[i] *= (double)f->scaleFactor;
      return;
  }
  return;
}
void prepGlobalData(int i,int j,int cc,DTYPE *din)
{ /* used internally, no need to check args */
  int fmtratio, nlen, len, siz, gid;
  double dval;
  NAME64 nval;
  DBLDBL ddval;
  NAME64DBL ndval;

  din->dFormat = daqTbl[i].daqParam[j].fmt;
  din->dArrayLength = daqTbl[i].daqParam[j].size;
  din->data.vptr = daqTbl[i].data + daqTbl[i].daqParam[j].offset;

  if ((gid=daqTbl[i].daqParam[j].globalId) >= 0 && gcastTbl[gid] != NULL)
  {
    gcastTbl[gid]->status = cc;
    gcastTbl[gid]->sysStamp = daqTbl[i].sysstamp;
  }
  if (daqTbl[i].hasSts && (daqTbl[i].daqParam[j].size == 1 || daqTbl[i].daqParam[j].fmt == CF_TEXT))
  {
    switch (daqTbl[i].daqParam[j].fmt)
    {
      case CF_INT16:
      case CF_INT32:
      case CF_INT64:
      case CF_FLOAT:
      case CF_DOUBLE:
        GetValuesAsDouble(din,&dval,1);
        fmtratio = fmtsizeof(CF_DBLDBL)/fmtsizeof(daqTbl[i].daqParam[j].fmt);
        if (daqTbl[i].rfmtdata == NULL) 
        {
          daqTbl[i].rfmtdata = (char *)calloc(daqTbl[i].len,fmtsizeof(CF_DBLDBL));
          if (daqTbl[i].rfmtdata == NULL) 
          {
            dbglog("cannot allocate memory for global %s",daqTbl[i].daqParam[j].param);
            return;
          }
        }
        din->dFormat = CF_DBLDBL;
        din->dArrayLength = daqTbl[i].daqParam[j].size;
        din->data.vptr = daqTbl[i].rfmtdata + daqTbl[i].daqParam[j].offset*fmtratio;
        ((DBLDBL *)din->data.vptr)->d1val = dval;
        ((DBLDBL *)din->data.vptr)->d2val = (double)cc;
        break;
      case CF_TEXT:
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        if (daqTbl[i].daqParam[j].fmt == CF_TEXT)
        {
          nlen = daqTbl[i].daqParam[j].size;
          siz = len = 1;
        }
        else
        {
          nlen = fmtsizeof(daqTbl[i].daqParam[j].fmt);
          len = daqTbl[i].len;
          siz = daqTbl[i].daqParam[j].size;
        }
        strncpy(nval.name,din->data.cptr,nlen);
        fmtratio = fmtsizeof(CF_NAME64DBL)/nlen;
        if (daqTbl[i].rfmtdata == NULL) 
        {
          daqTbl[i].rfmtdata = (char *)calloc(len,fmtsizeof(CF_NAME64DBL));
          if (daqTbl[i].rfmtdata == NULL) 
          {
            dbglog("cannot allocate memory for global %s",daqTbl[i].daqParam[j].param);
            return;
          }
        }
        din->dFormat = CF_NAME64DBL;
        din->dArrayLength = siz;
        din->data.vptr = daqTbl[i].rfmtdata + daqTbl[i].daqParam[j].offset*fmtratio;
        strncpy(((NAME64DBL *)din->data.vptr)->name,nval.name,64);
        ((NAME64DBL *)din->data.vptr)->dval = (double)cc;
        break;
      case CF_FLTINT:
      case CF_INTINT:
      case CF_DBLDBL:
      case CF_FLTFLT:
        GetValuesAsDBLDBL(din,&ddval,1);
        fmtratio = fmtsizeof(CF_DBLDBL)/fmtsizeof(daqTbl[i].daqParam[j].fmt);
        if (daqTbl[i].rfmtdata == NULL) 
        {
          daqTbl[i].rfmtdata = (char *)calloc(daqTbl[i].len,fmtsizeof(CF_DBLDBL));
          if (daqTbl[i].rfmtdata == NULL) 
          {
            dbglog("cannot allocate memory for global %s",daqTbl[i].daqParam[j].param);
            return;
          }
        }
        din->dFormat = CF_DBLDBL;
        din->dArrayLength = daqTbl[i].daqParam[j].size;
        din->data.vptr = daqTbl[i].rfmtdata + daqTbl[i].daqParam[j].offset*fmtratio;
        ((DBLDBL *)din->data.vptr)->d1val = ddval.d1val;
        ((DBLDBL *)din->data.vptr)->d2val = cc == 0 ? ddval.d2val : (double)cc;
        break;
      case CF_NAME8I:
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME64I:
      case CF_NAME64DBL:
        GetValuesAsNAME64DBL(din,&ndval,1);
        fmtratio = fmtsizeof(CF_NAME64DBL)/fmtsizeof(daqTbl[i].daqParam[j].fmt);
        if (daqTbl[i].rfmtdata == NULL) 
        {
          daqTbl[i].rfmtdata = (char *)calloc(daqTbl[i].len,fmtsizeof(CF_NAME64DBL));
          if (daqTbl[i].rfmtdata == NULL) 
          {
            dbglog("cannot allocate memory for global %s",daqTbl[i].daqParam[j].param);
            return;
          }
        }
        din->dFormat = CF_NAME64DBL;
        din->dArrayLength =  daqTbl[i].daqParam[j].size;
        din->data.vptr = daqTbl[i].rfmtdata + daqTbl[i].daqParam[j].offset*fmtratio;
        strncpy(((NAME64DBL *)din->data.vptr)->name,ndval.name,64);
        ((NAME64DBL *)din->data.vptr)->dval = cc == 0 ? ndval.dval : (double)cc;
        break;
      default: 
        /* keep the given format */
        break;
    }
  }
}
int rewireDaqDataPool(int idx,int *reqCapacity, int *curCapacity)
{
  int i, dsiz=0;
  size_t pptr;
  BYTE *ptr, *dptr;
  DIMAGE *img;
  DGSPECTRUM *gsp;
  DataPoolEntry *dpe, *ref, *lptr;

  dpe = (DataPoolEntry *)daqTbl[idx].data;
  ref = (DataPoolEntry *)daqTbl[idx].archdata;
  lptr = (DataPoolEntry *)daqTbl[idx].lastdata;
  dptr = dpe->data;
  switch (daqTbl[idx].fmt)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      dsiz = daqTbl[idx].len*sizeof(size_t);
      for (i=0; i<daqTbl[idx].len; i++)
      {
        dsiz += (int)strlen(((char **)dptr)[i]) + 1;
      }
      break;
    case CF_IMAGE:
      img = (DIMAGE *)dptr;
      dsiz = img->frameHeader.appendedFrameSize + IMAGE_HDR_SIZE;
      break;
    case CF_GSPECTRUM:
      gsp = (DGSPECTRUM *)dptr;
      // need the extra 12 bytes (size & ptr)
      dsiz = GetCompletionDataSize(recordLinkIdSts[idx].id) + sizeof(DGSPECTRUM); //GSPECTRUM_HDR_SIZE;
      break;
  }
  *reqCapacity = dsiz;
  *curCapacity = dpe->capacity;
  if (dsiz > dpe->capacity)
  { /* must increase capacity */
    if (daqTbl[idx].fmt == CF_IMAGE) return buffer_too_small;
    dsiz += 0x1000; /* a bit of surplus */
    if ((ptr=realloc(ref->data,dsiz)) == NULL) return out_of_server_memory;
    ref->data = ptr;
    ref->capacity = dsiz;
    if ((ptr=realloc(dpe->data,dsiz)) == NULL) return out_of_server_memory;
    dpe->data = ptr;
    dpe->capacity = dsiz;
    dptr = dpe->data;
    if ((ptr=realloc(lptr->data,dsiz)) == NULL) return out_of_server_memory;
    lptr->data = ptr;
    lptr->capacity = dsiz;
  }
  switch (daqTbl[idx].fmt)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      ptr = dptr + daqTbl[idx].len*sizeof(size_t);
      pptr = 0;
      for (i=0; i<daqTbl[idx].len; i++)
      {
        dsiz = (int)strlen(((char **)dptr)[i]) + 1;
        memcpy(ptr,((char **)dptr)[i],dsiz);
        ((char **)dptr)[i] = (char *)NULL + pptr;
        ptr += dsiz;
        pptr += dsiz;
      }
      dpe->length = daqTbl[idx].len*sizeof(size_t) + dsiz;
      break;
    case CF_IMAGE:
      /*
      --- IMAGE data is fixed @ initialization ---- !!!
      img = (DIMAGE *)dptr;
      ptr = dptr + IMAGE_HDR_SIZE;
      memcpy(ptr,img->frameBuffer,img->frameHeader.appendedFrameSize);
      */
      dpe->length = dsiz;
      break;
    case CF_GSPECTRUM:
      dpe->length = dsiz;
      break;
  }
  *reqCapacity = dsiz;
  *curCapacity = dpe->capacity;
  return 0;
}
HistorySpecification glbHSpec = 
{
  HRS_DEFAULT_POLL_RATE,0,1200,-1,0,0,0,NULL,NULL
};
extern int gIsTrace;
extern char *gTraceKey;
void dataCallback(int i,int cc)
{
  double dts, ts = MakeDataTimeStamp();
  SINT32 sstmp = 0, ustmp = 0;
  UINT16 xferReason;
  int rc, j, k, lid, sts, slowDeadband = ARCH_SLOW_INTERVAL;
  int checkArchive=0,reack=FALSE,wrsts=FALSE,isPastFltr=FALSE,isSched=FALSE;
  int reqCap, curCap;
  DTYPE din;
  UINT16 opt;
  static int lastPEnergy;
  static float lastEEnergy;
  /* monitor crucial parameters */
  static int PEpanic = 0, EEpanic = 0,PIpanic = 0; 
  static int STpanic = 0;
  static int globalsRegistered = FALSE;      
  char mcaProperty[PROPERTY_NAME_SIZE+4];
  KeywordFailover *kw;

  i--; /* backup 1 for table entry id */
  if (i < 0 || i >= nrecords) ccerr(invalid_index);
  lid = recordLinkIdSts[i].id;
  gIsTrace = isTraceLink(lid);
  if (gIsTrace) gTraceKey = conTbl[lid]->key;
  sts = GetCurrentLinkStatus(lid);
  if (sts == link_blacklisted || cc == address_unknown || cc == illegal_property)
  {
    if (daqTbl[i].dataStatus != link_blacklisted)
    {
      char s[512],key[256];
      sprintf(key,"/%.32s/%.32s/%.64s[%.64s]",
        daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
      sprintf(s,"%.6s: link to %.256s : %.32s",MEX_EQMNAME,key,erlst[cc]);
      AppendProblemsLog(s);
      SetAlarmEx(MEX_EQMNAME,0,ALMCODE_NONEXIST,key,almTRANSIENT);
    }
    daqTbl[i].dataStatus = link_blacklisted;
    if (gIsTrace) dbglog("%.192s (rec. %d) has been blacklisted",gTraceKey,daqTbl[i].rec);
    return;
  }
  xferReason = GetTransferFlagFromCallbackId(i);
  if (xferReason & CX_EVENT) isSched = TRUE;
  /* dts = GetDataTimeStamp(); */
  dts = GetCurrentDataTimeStamp(recordLinkIdSts[i].id);
  sstmp = GetSystemDataStamp(recordLinkIdSts[i].id);
  if (sstmp == 0) sstmp = gSystemDataStamp; /* maybe this can help */
  ustmp = GetDataStamp(recordLinkIdSts[i].id);
  opt = (UINT16)daqTbl[i].option;
  if (daqTbl[i].tmr < 1000) isSched = TRUE; /* treat exotic fast interval guys as scheduled (background task @1Hz) */
  rc = cc & 0x0fff; /* strip off information bits */
  if (cc & CE_SENDDATA) 
  {
    cc = 0;
    wrsts = TRUE;
  }
  if (opt&ONCE) 
  {
    wrsts = FALSE;
    if (cc == 0) 
    {
      CloseLink(recordLinkIdSts[i].id);
      checkArchive = 1;
      if (strlen(daqTbl[i].resetMcaPrp))
      {
        resetPropertyDeviceList(MEX_EQMNAME,daqTbl[i].resetMcaPrp);
      }
    }
    else 
    { /* keep trying */
      return;
    }
  }
  recordLinkIdSts[i].startPending = FALSE;
  if (archdebug > 1 || gIsTrace)
  {
    dbglog("LINK (%d) %.32s %.64s : %s <%d> (%g)",
                  i,daqTbl[i].srv,daqTbl[i].prp,
                  cc?"not responding":"incoming",cc,dts);
  }
  if (cc == reset_mca_property) 
  { /* explicit signal */
    reack = TRUE;
  }
  if (cc == 0 && daqTbl[i].daqStatus != 0 && 
      daqTbl[i].errCounter > ERROR_REACK_MCA_COUNT) 
  { /* wasn't okay but now it is ... */
    reack = TRUE;
  }
  if (reack)
  { /* check reacquire multi-channel names list */
    char namProperty[PROPERTY_NAME_SIZE+4];
    for (j=0; daqTbl[i].len > 0 && j<daqTbl[i].nparam; j++)
    {
      if (daqTbl[i].daqParam[j].size <= 1) continue;
      if (strstr(daqTbl[i].daqParam[j].param,".NAM") != NULL) continue;
      strncpy(mcaProperty,daqTbl[i].daqParam[j].param,PROPERTY_NAME_SIZE);
      sprintf(namProperty,"%.60s.NAM",mcaProperty);
      if ((k=getEntryFromKeyword(namProperty,NULL,NULL)) >= 0)
      { /* looks like we have a .NAM property with the channel names ... */
        if (reacquireStaticDataSet(k) == 0)
        { /* also trigger a reassignment in the internal association ... */
          strncpy(daqTbl[k].resetMcaPrp,mcaProperty,PROPERTY_NAME_SIZE);
        }
      }
    }
  }
# ifdef THIS_TURNS_OUT_TO_BE_A_GOOD_IDEA
  if (rc == 0 && daqTbl[i].dataStatus != rc) 
  { /* status returned to normal success */
    if (daqTbl[i].laststs > (double)daqTbl[i].hbt) 
    { /* there's been at least 1 status message archived */
      wrsts = TRUE;
    }
  }
# endif
  if (rc != 0 && rc != reset_mca_property)
  { /* error condition */
    if (dts == 0 || rc == link_timeout) dts = makeDataTimeStamp();
    if (daqTbl[i].errCounter > ERROR_APPLY_STATUS_COUNT)
    { /* status changed */
      if (dts > daqTbl[i].laststs + (double)daqTbl[i].hbt) wrsts = TRUE;
    }
  }
  daqTbl[i].dataStatus = rc;
  daqTbl[i].daqStatus = cc;
  daqTbl[i].dataFlg = cc ? -cc : 1;
  if (wrsts && !gIsGlobalsServer)
  {
    WriteStatusRecord(&daqTbl[i],i+1,dts);
    if (gIsTrace) dbglog("%.192s (rec. %d) write status record",gTraceKey,daqTbl[i].rec);
  }
  if (cc) daqTbl[i].errCounter++; else daqTbl[i].errCounter = 0;
  for (j=0; j<daqTbl[i].nparam; j++)
  {
    if (daqTbl[i].daqParam[j].scaleOffset != 0) addOffsetToData(daqTbl[i].data,&daqTbl[i].daqParam[j]);
    if (daqTbl[i].daqParam[j].scaleFactor != 1) scaleFactorToData(daqTbl[i].data,&daqTbl[i].daqParam[j]);
    updateStateInfo(i,j,cc);
    if (daqTbl[i].daqParam[j].isglobal) prepGlobalData(i,j,cc,&din); 
  }
  if (daqTbl[i].option&FIXTIMESTAMP) 
  {
    daqTbl[i].tstamp = ts;
    daqTbl[i].sysstamp = gSystemDataStamp;
    daqTbl[i].usrstamp = 0;
    if (gIsTrace) dbglog("%.192s (rec. %d) fix timestamp",gTraceKey,daqTbl[i].rec);
  }
  else 
  {
    if (dts < 100) 
    {
      if (gIsTrace) dbglog("%.192s (rec. %d) rejected! (invalid data timestamp %g)",gTraceKey,daqTbl[i].rec,dts);
      return;
    }
    daqTbl[i].tstamp = dts;
    daqTbl[i].sysstamp = sstmp;
    daqTbl[i].usrstamp = ustmp;
  }
  /* systematic entries are checked at their polling rates: */
  if (cc) 
  {
    if (gIsTrace) dbglog("%.192s (rec. %d) rejected! (error code %d)",gTraceKey,daqTbl[i].rec,cc);
    return;
  }
  if (daqTbl[i].hasDataPool && (cc=rewireDaqDataPool(i,&reqCap,&curCap)) != 0) 
  {
    DataPoolEntry *dpe = (DataPoolEntry *)daqTbl[i].data;
    if (gIsTrace) dbglog("%.192s (rec. %d) rejected! (could not rewire data pool)",gTraceKey,daqTbl[i].rec);
    feclogEx(TINE_LOGLEVEL_ERROR, "%.64s/%.64s[%.64s] : incoming data %d bytes, data pool capacity %d bytes => %.32s",
      daqTbl[i].srv, daqTbl[i].dev, daqTbl[i].prp, reqCap,curCap,cc2str(cc));
    return;
  }
  if ((ts < (daqTbl[i].archtime + ARCH_STD_INTERVAL)) && !(opt&FAST)) ccerr(0);
  if (daqTbl[i].fmt == CF_IMAGE) slowDeadband = daqTbl[i].hbt;
  if (daqTbl[i].tmr >= ARCH_SLOW_INTERVAL) slowDeadband = daqTbl[i].hbt;
  if ((ts < (daqTbl[i].archtime + slowDeadband)) && (opt&SLOW)) ccerr(0);
  if (isPastFilter(opt)) 
  {
    isPastFltr = TRUE;
    checkArchive = 1;
  }
  else
  { /* 1st time not past and possible hold-over from last update ? */
    if (daqTbl[i].isPastFilter) checkArchive = 1; 
  }
  if (ts > (daqTbl[i].archtime + slowDeadband)) checkArchive = 1;
  if (checkArchive && !gIsGlobalsServer) checkData(i,isPastFltr,isSched); /* decide whether to archive or not */
  cc = 0;
  if (registerGlobals && !globalsRegistered) 
  { /* make sure everything is there */
    for (j=0; j<nrecords; j++) if (recordLinkIdSts[j].startPending) return;
    globalsRegistered = TRUE;
    /* anything extra to do? */
    GlobalsRegistrationEx();
    /* check what I know about: */
    for (i=0; i<nrecords; i++) for (j=0; j<daqTbl[i].nparam; j++)
      for (k=0; k<ngloballist; k++)
       if (!strnicmp(daqTbl[i].daqParam[j].param,globalList[k].keyword,PROPERTY_NAME_SIZE))
    { /* tie it directly to the incoming data: */
      daqTbl[i].daqParam[j].isglobal = TRUE;
      prepGlobalData(i,j,0,&din); /* set up the registration call */
      cc = sendNetGlobal(daqTbl[i].daqParam[j].param,&din,cbSendGlobal,
                        globalList[k].minPeriod,globalList[k].maxPeriod,
                        globalList[k].tolerance);
      feclog("register global %s : %s",daqTbl[i].daqParam[j].param,cc < 0 ? erlst[-cc] : "success");
      daqTbl[i].daqParam[j].globalId = cc;
      /* keep local short-term local history of global parameters */
      AppendHistoryInformation(MEX_EQMNAME,daqTbl[i].daqParam[j].param,"#0",
        daqTbl[i].daqParam[j].size,daqTbl[i].daqParam[j].fmt,k,&glbHSpec);
    } 
    for (kw=kwrdFailover; kw != NULL; kw=kw->nxt) for (k=0; k<ngloballist; k++)
    {
      if (!strnicmp(kw->keyword,globalList[k].keyword,PROPERTY_NAME_SIZE))
      {
        din.dArrayLength = kw->len;
        din.dFormat = kw->fmt;
        din.data.bptr = kw->data;
        cc = sendNetGlobal(kw->keyword,&din,cbSendGlobal,
                        globalList[k].minPeriod,globalList[k].maxPeriod,
                        globalList[k].tolerance);
        feclog("register failover global %s : %s",kw->keyword,cc < 0 ? erlst[-cc] : "success");
        AppendHistoryInformation(MEX_EQMNAME,kw->keyword,"#0",kw->len,kw->fmt,k,&glbHSpec);
      }
    }
  }
err:
  if (cc)
  { /* problems ... */
    if (archdebug || gIsTrace) dbglog("Callback : %s",cc2str(cc));
  }
  else
  { /* copy this data into the lastdata set */
    BYTE *rptr = daqTbl[i].lastdata;
    BYTE *dptr = daqTbl[i].data;
    int rlen = daqTbl[i].reclen;
    if (daqTbl[i].hasDataPool)
    {
      DataPoolEntry *dpe = (DataPoolEntry *)daqTbl[i].data;
      DataPoolEntry *ref = (DataPoolEntry *)daqTbl[i].lastdata;
      dptr = dpe->data;
      rptr = ref->data;
      rlen = dpe->length;

    }
    if (rptr != NULL) memcpy(rptr,dptr,rlen);
    daqTbl[i].lasttime = daqTbl[i].tstamp;
    daqTbl[i].lastsys = daqTbl[i].sysstamp;
    daqTbl[i].lastusr = daqTbl[i].usrstamp;
  }
  if (daqTbl[i].kwf != NULL) checkFailoverData(daqTbl[i].kwf);
  return;
}
int redirDataLink(char *fname,char *srv)
{
  int n, fmt;
  float maxval,minval;
  char str[256],scratch[128],dnamePrp[PROPERTY_NAME_SIZE],desc[128];
  int done = 0,cc = 0;
  int key_col,fmt_col,siz_col,uni_col,max_col,min_col,
      cmt_col;
  FILE *fp;
  DTYPE dout;

  sprintf(str,"%sinfo/%s",ArchiveDBpath,fname);
  if ((fp=fopen(str,"rt")) == NULL)
  {
    feclog("redirDataLink : can't open structure file %s",str);
    return -1;
  }
  for (n=0; csvGetLine(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((key_col=findcol(str,"KEYWORD")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(str,"SIZE")) < 0) ccerr(no_such_line);
      if ((uni_col=findcol(str,"UNITS")) < 0) ccerr(no_such_line);
      if ((max_col=findcol(str,"MAXIMUM")) < 0) ccerr(no_such_line);
      if ((min_col=findcol(str,"MINIMUM")) < 0) ccerr(no_such_line);
      if ((cmt_col=findcol(str,"COMMENT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dnamePrp,colptr(key_col,str),KEYWORD_NAME_SIZE);
    strncpy(scratch,colptr(fmt_col,str),16); strupr(scratch);
    if ((fmt=getFormatType(scratch)) == CF_NULL) fmt = atoi(scratch);
    dout.dFormat = fmt;
    strncpy(scratch,colptr(siz_col,str),16);
    dout.dArrayLength = atoi(scratch);
    strncpy(scratch,colptr(min_col,str),16);
    minval = (float)atof(scratch);
    strncpy(scratch,colptr(max_col,str),16);
    maxval = (float)atof(scratch);
    sprintf(desc,"[%g:%g ",minval,maxval);
    strncat(desc,colptr(uni_col,str),32);
    strcat(desc,"]");
    strncat(desc,colptr(cmt_col,str),48);
    RegisterPropertyInformation(MEX_EQMNAME,dnamePrp,&dout,NULL,CA_READ,AT_UNKNOWN,0,desc,-1,srv);
  }
err:
  fclose(fp);
  return 0;
}
int redirDataFile(char *fname,char *srv)
{
  FILE *fp;
  char fn[80];
  char str[256],scratch[256];
  int done = 0,cc = 0;
  int active_col,sfile_col,rectype_col;

  sprintf(fn,"%s%s",ArchiveDBpath,fname);
  if ((fp = fopen(fn,"rt")) == NULL)
  {
    printf("\n Can't open data base file : %s\n",fname);
    return -1;
  }
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((rectype_col=findcol(str,"RECORD TYPE")) < 0) ccerr(no_such_line);
      if ((sfile_col=findcol(str,"STRUCTURE")) < 0) ccerr(no_such_line);
      if ((active_col=findcol(str,"ACTIVE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    if (atoi(colptr(active_col,str)) == 0) continue;
    strncpy(scratch,colptr(rectype_col,str),32);
    if (atoi(colptr(rectype_col,str)) < 2000) continue;
    redirDataLink(colptr(sfile_col,str),srv);
  }
err:
  fclose(fp);
  return 0;
}
int redirInit(NAME16 *dbfile,NAME64 *dbsrv,int ndbfiles)
{
  FILE *fp = NULL;
  char fn[80];
  int done = 0,i,cc = 0;

  for (i=0; i<ndbfiles; i++)
  {
    sprintf(fn,"%s%s",ArchiveDBpath,dbfile[i].name);
    redirDataFile(fn,dbsrv[i].name);
  }
  return 0;
}
int reacquireStaticDataSet(int i)
{
  DTYPE dout, din, *dptr;
  int id, cc=0;
  int accmode;
  char dev[256],s[256];

  if (strlen(daqTbl[i].ctx) > 0)
    sprintf(dev,"/%.32s/%.32s/%.64s",daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev);
  else
    sprintf(dev,"%.32s/%.64s",daqTbl[i].srv,daqTbl[i].dev);
  if (recordLinkIdSts[i].startPending)
  {
    AppendHistoryLog("reacquire %.128s %.64s already under way",dev,daqTbl[i].prp);
    return 0;
  }
  accmode = CM_TIMER;
  recordLinkIdSts[i].startPending = TRUE;
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = (short)daqTbl[i].fmt;
  dout.dArrayLength = daqTbl[i].len;
  dout.data.vptr = daqTbl[i].data;
  if (daqTbl[i].inlen > 0)
  {
    memset(&din,0,sizeof(DTYPE));
    din.dFormat = daqTbl[i].infmt;
    din.dArrayLength = daqTbl[i].inlen;
    din.data.bptr = daqTbl[i].indata;
    dptr = &din;
  }
  else dptr = NULL;

  if ((id=AttachLinkEx(dev,daqTbl[i].prp,&dout,dptr,CA_READ,
                     daqTbl[i].tmr,dataCallback,accmode,1+i)) < 0)
  {
    sprintf(s,"%.6s: can't attach to %.128s %.64s : %.32s",
      MEX_EQMNAME,dev,daqTbl[i].prp,erlst[-id]);
    AppendProblemsLog(s);
  }
  else
  {
    recordLinkIdSts[i].id = id;
    AppendHistoryLog("reacquire %.128s %.64s",dev,daqTbl[i].prp);
  }
  return cc;
}
extern CSVDB csvExportsInfoDb;
int populateXIS(char *fecname,char *eqmname,DBLNKLIST *fecList,XIS **xis);
XIS *additionalKeywords = NULL;
int nAdditionalKeywords = 0;
int getAdditionalKeywords(void)
{ /* re-scans the exports.csv file (if there) for aliases or redirections */
  int i, sz, cc=0;

  if (isUsingXMLFecDatabase)
  {
    sz = populateXIS(gFecName,MEX_EQMNAME,gFecInfoList,&additionalKeywords);
    if (sz < 0) ccerr(-sz);
  }
  else
  {
    if ((cc=csvReadFile(fecDbPath,"exports.csv",&csvExportsInfoDb,(void **)&additionalKeywords)) != 0) ccerr(cc);
    sz = csvExportsInfoDb.siz;
  }
  nAdditionalKeywords = sz;
  for (i=0; i<nAdditionalKeywords; i++)
  {
    checkDaqSubSystem(additionalKeywords[i].SubSystem);
  }
  feclog("getAdditionalKeywords : added %d additional keywords",sz);
err:
  if (cc != 0 && cc != no_such_file) feclog("getAdditionalKeywords : %s",erlst[cc]);
  return cc;
}
int getRedirectedIndex(char *prp)
{
  int i;
  for (i=0; i<nAdditionalKeywords; i++)
  {
    if (strcmp(additionalKeywords[i].EqmName,MEX_EQMNAME)) continue;
    if (!strnicmp(prp,additionalKeywords[i].PrpName,PROPERTY_NAME_SIZE)) 
    {
      if (strlen(additionalKeywords[i].redirection) > 0) return i;
    }
  }
  return -1;
}
//int useDeviceSpecificRedirection = FALSE;
int linksEstablished = 0;
char primaryDbFile[128];
time_t dbLoadTime = 0;
int daqInit(char *dbfile,int ndbfiles)
{
  DTYPE dout, din, *dptr, dprp;
  int i, k, n, id, cc=0, rlen=0, len, nprps = 0;
  DAQParameter *daqParam;
  KeywordFailover *kw;
  int p, rec;
  int accmode, paccess = CA_READ, isSpectrum = FALSE;
  char dev[256],prp[PROPERTY_NAME_SIZE+8],s[256],hstrdr[256],lclrdr[256],*pRdrStr;

  if (linksEstablished) return 0;
  DTYPEZERO(din);
  srvlog(ALF,"daqInit","--- initializing ---");

  isArchiver = TRUE; /* actively taking and storing data */
  gAppendProperyAliasNamesToQueries = FALSE;

  for (i=0; i<ndbfiles; i++)
  { /* usually only one database set : writing back out collapses to a single database */
    totalNrRecords += getNumActiveRecords(&dbfile[i*64]);
  }
  ArchiveDBfile = primaryDbFile;
  strncpy(ArchiveDBfile,dbfile,64); ArchiveDBfile[64] = 0;
  recordsCapacity = totalNrRecords + RECORD_CAPACITY_BUFFER;

  getFilterTable();

  if ((daqTbl=(DAQRecord *)calloc(recordsCapacity,sizeof(DAQRecord))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(recordsCapacity,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(recordsCapacity,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);
  if ((recordLinkIdSts=(LINKSTRCT *)calloc(recordsCapacity,sizeof(LINKSTRCT))) == NULL) ccerr(out_of_server_memory);

  registerDaqStructs(recordsCapacity*2);

  for (i=0; i<ndbfiles; i++)
  {
    n = loadDaqDb(&dbfile[i*64],nrecords);/* number of recs being archived */
    inputDataLoad(n,nrecords);           /* any input data? */
    nparams += loadDaqParamDb(n,nrecords);    /* number of parameters archived */
    nrecords += n;
  }
  addExtSrcInfo();
  fillinArrayTypes(nrecords);
  logRestart();                          /* establish archive file names */

  for (i=0; i<nrecords; i++)       
  { /* start data links */
    if (daqTbl[i].active <= 0) continue;
    if (!stricmp(daqTbl[i].ctx,"SELF")) continue;
    accmode = daqTbl[i].mod;
    recordLinkIdSts[i].startPending = TRUE;
    dout.dFormat = (short)daqTbl[i].fmt;
    dout.dArrayLength = daqTbl[i].len;
    dout.data.vptr = daqTbl[i].data;
    if (daqTbl[i].fmt == CF_STRUCT) strncpy(dout.dTag,daqTbl[i].tag,TAG_NAME_SIZE); else dout.dTag[0] = 0;
    if (strlen(daqTbl[i].ctx) > 0)
      sprintf(dev,"/%.32s/%.32s/%.64s",daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev);
    else
      sprintf(dev,"%.32s/%.64s",daqTbl[i].srv,daqTbl[i].dev);
    if (daqTbl[i].hasDataPool)
    {
      DataPoolEntry *dpe = (DataPoolEntry *)daqTbl[i].data;
      dout.data.vptr = dpe[0].data;
      if (daqTbl[i].fmt == CF_IMAGE)
      {
        DIMAGE *img = (DIMAGE *)dout.data.vptr;
        img->frameBuffer = dout.data.bptr + sizeof(DIMAGE);
        img->frameBufferSize = dpe[0].capacity - sizeof(DIMAGE);
      }
      if (daqTbl[i].fmt == CF_GSPECTRUM)
      {
        DGSPECTRUM *gsp = (DGSPECTRUM *)dout.data.vptr;
        gsp->spectBuffer = (float *)(dout.data.bptr + sizeof(DGSPECTRUM));
        gsp->spectBufferSize = dpe[0].capacity - sizeof(DGSPECTRUM);
      }
    }
    if (daqTbl[i].inlen > 0)
    {
      din.dFormat = daqTbl[i].infmt;
      din.dArrayLength = daqTbl[i].inlen;
      din.data.bptr = daqTbl[i].indata;
      dptr = &din;
    }
    else dptr = NULL;

    if ((id=AttachLinkEx(dev,daqTbl[i].prp,&dout,dptr,CA_READ,
                       daqTbl[i].tmr,dataCallback,accmode,1+i)) < 0)
    {
      sprintf(s,"%.6s: can't attach to %.128s[%.64s] : %.32s",MEX_EQMNAME,dev,daqTbl[i].prp,erlst[-id]);
      AppendProblemsLog(s);
      if (id == -address_unknown || id == -non_existent) SetAlarmEx(MEX_EQMNAME,0,ALMCODE_NONEXIST,dev,almTRANSIENT);
    }
    else
    {
      recordLinkIdSts[i].id = id;
      sprintf(s,"%.6s: established link %.128s %.64s",MEX_EQMNAME,dev,daqTbl[i].prp);
      daqTbl[i].daqStatus = link_not_open; /* let this be cleared normally */
    }
    feclog(s);
  }
  sprintf(hstrdr,"/%.32s/HISTORY",gDeviceContext); len = (int)strlen(hstrdr);
  if (globalsServer && requireNetworkAccess) 
  {
    paccess |= CA_NETWORK;
    feclog("requiring multicast access of parameters");
  }
  for (i=0; i<nrecords; i++)       /* register properties */
  {
    for (n=0; n<daqTbl[i].nparam; n++)
    {
      hstrdr[len] = 0; /* reset back to the base path */
      isSpectrum = (daqTbl[i].daqParam[n].arrayType & AT_SPECTRUM);
      sprintf(&hstrdr[len],"[%.64s]",daqTbl[i].daqParam[n].param);
      if (isSpectrum)
      {
        sprintf(s,"[%g:%g %.16s][%g:%g %.16s]%.64s",
          daqTbl[i].daqParam[n].minimum,daqTbl[i].daqParam[n].maximum,daqTbl[i].daqParam[n].units,
          daqTbl[i].daqParam[n].xminimum,daqTbl[i].daqParam[n].xmaximum,daqTbl[i].daqParam[n].xunits,
          daqTbl[i].daqParam[n].desc);
      }
      else
      {
        sprintf(s,"[%g:%g %.16s]%.64s",
          daqTbl[i].daqParam[n].minimum,daqTbl[i].daqParam[n].maximum,
          daqTbl[i].daqParam[n].units,daqTbl[i].daqParam[n].desc);
      }
      dprp.dFormat = daqTbl[i].daqParam[n].fmt;
      dprp.dArrayLength = daqTbl[i].daqParam[n].size;
      if (dprp.dFormat == CF_STRUCT) strncpy(dprp.dTag,daqTbl[i].tag,TAG_NAME_SIZE); else dprp.dTag[0] = 0;
      pRdrStr = daqTbl[i].active < 0 ? hstrdr : NULL;
      rlen = (daqTbl[i].daqParam[n].arrayType & AT_DOUBLE) ? daqTbl[i].daqParam[n].rowsize : daqTbl[i].daqParam[n].size;
      if ((id=RegisterPropertyInformation(MEX_EQMNAME,daqTbl[i].daqParam[n].param,
           &dprp,NULL,(short)paccess,daqTbl[i].daqParam[n].arrayType,(UINT16)rlen,s,-1,pRdrStr)) < 0)
      {
        feclog("Couldn't register prp %.64s: %.32s",daqTbl[i].daqParam[n].param,erlst[-id]);
      }
      nprps++;
      if (!strnicmp(daqTbl[i].ctx,thisContext,CONTEXT_NAME_SIZE) &&
           daqTbl[i].nparam == 1 && (daqTbl[i].option & ONCE) == 0)
      { /* this is a one to one acq property and hst parameter and not a names acq */
        if (strlen(daqTbl[i].srv) + strlen(daqTbl[i].prp) < PROPERTY_NAME_SIZE)
        { /* the lengths will fit as an alias */
#         if defined(WIN32) || defined(UNIX)
          snprintf(prp,PROPERTY_NAME_SIZE,"%s.%s",daqTbl[i].srv,daqTbl[i].prp);
#         else
          sprintf(prp,"%.23s.%40s",daqTbl[i].srv,daqTbl[i].prp);
#         endif
          if (strnicmp(prp,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE) != 0 &&
              GetSystemAlias(daqTbl[i].daqParam[n].param) == NULL)
          { /* and this isn't already the property name and not already aliased! */
            RegisterPropertyAlias(MEX_EQMNAME,daqTbl[i].daqParam[n].param,prp);
          }
        }
      }
      if (canSendGlobals) continue; /* don't redirect history calls if I'm the globals server or watchdog */
      /* redirect history calls to the source server or history server */
      strncpy(s,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE); s[PROPERTY_NAME_SIZE] = 0;
      if (strstr(s,".NAM") != NULL) continue;
      strcat(s,".HIST");
      if (daqTbl[i].nparam == 1 && daqTbl[i].active > 0)
      { /* safe to return source */
        int useDeviceSpecificRedirection = isSpectrum;
        if (daqTbl[i].len == 1) useDeviceSpecificRedirection = TRUE;
        if (useDeviceSpecificRedirection)
        { /* need the originating device name */
          sprintf(lclrdr,"/%.32s/%.32s/%.60s#ANY[%.59s.HIST]",
            daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
        }
        else
        { /* go with multi-channel logic */
          sprintf(lclrdr,"/%.32s/%.32s[%.59s.HIST]",daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].prp);
        }
        pRdrStr = lclrdr;
      }
      else
      {
        pRdrStr = hstrdr;
      }
      /* redirect the .HIST calls */
      if ((id=RegisterPropertyInformation(MEX_EQMNAME,s,&dprp,NULL,CA_READ,
           daqTbl[i].daqParam[n].arrayType,(UINT16)rlen,s,-1,pRdrStr)) < 0)
      {
        feclog("Couldn't register prp %.64s: %.32s",s,erlst[-id]);
      }
      /* redirect archive calls to the history server only */
      strncpy(s,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE); s[PROPERTY_NAME_SIZE] = 0;
      strcat(s,".ARCH");
      if ((id=RegisterPropertyInformation(MEX_EQMNAME,s,&dprp,NULL,CA_READ,
           daqTbl[i].daqParam[n].arrayType,(UINT16)rlen,s,-1,hstrdr)) < 0)
      {
        feclog("Couldn't register prp %s: %s",s,erlst[-id]);
      }
    }
  }
  for (n=0; n<nPrpAlias; n++)
  {
    if ((i=getEntryFromKeyword(prpAlias[n].alias,&rec,&p)) < 0) continue;
    daqParam = daqTbl[i].daqParam;
    dprp.dFormat = daqParam[p].fmt;
    dprp.dArrayLength = daqParam[p].size;
    dprp.dTag[0] = 0;
    rlen = (daqParam[p].arrayType & AT_DOUBLE) ? (int)daqParam[p].scaleOffset : daqParam[p].size;
    sprintf(s,"[%g:%g %.16s!%.16s]%.64s",daqParam[p].minimum,
            daqParam[p].maximum,daqParam[p].units,
            daqParam[p].plot,daqParam[p].desc);
    if ((id=RegisterPropertyInformation(MEX_EQMNAME,prpAlias[n].name,
         &dprp,NULL,(short)paccess,daqParam[p].arrayType,(UINT16)rlen,s,-1,NULL)) < 0)
    {
      cc = -id;
      feclog("Couldn't register property %.64s: %.32s",prpAlias[n].name,erlst[cc]);
    }
    nprps++;
  }
  linksEstablished = 1;
  if (nprps > 256) gIgnoreMetaPropertyQueries = TRUE;

  if (getFailoverInformation(FAILOVERS_FILENAME) == 0)
  { /* failover information registered */
    for (kw=kwrdFailover; kw != NULL; kw=kw->nxt)
    { /* run thru the list */
      for (k=0; k<3; k++) for (i=0; i<nrecords; i++) if (daqTbl[i].active)
      { /* find the cross-references into the daq table */
        for (n=0; n<daqTbl[i].nparam; n++)
        {
          if (!strnicmp(daqTbl[i].daqParam[n].param,kw->foSrc[k].name,PROPERTY_NAME_SIZE))
          { /* found this one ! */
            if (kw->len == 0)
            { /* not yet assigned */
              kw->len = daqTbl[i].daqParam[n].size;
              kw->fmt = (int)daqTbl[i].daqParam[n].fmt;
              kw->data = (BYTE *)calloc(kw->len,fmtsizeof(kw->fmt));
              if (kw->data == NULL) feclog("could not allocate failover data area for %.64s",daqTbl[i].daqParam[n].param);
            }
            kw->foSrc[k].recNr = i;
            kw->foSrc[k].prmNr = n;
            sprintf(s,"[%g:%g %.16s]%.64s",
              daqTbl[i].daqParam[n].minimum,daqTbl[i].daqParam[n].maximum,
              daqTbl[i].daqParam[n].units,daqTbl[i].daqParam[n].desc);
            dprp.dFormat = daqTbl[i].daqParam[n].fmt;
            dprp.dArrayLength = daqTbl[i].daqParam[n].size;
            dprp.dTag[0] = 0;
            if ((id=RegisterPropertyInformation(MEX_EQMNAME,kw->keyword,&dprp,NULL,(short)paccess,daqTbl[i].daqParam[n].arrayType,0,s,-1,NULL)) < 0)
            {
              feclog("Couldn't register failover prp %s: %s",kw->keyword,erlst[-id]);
            }            
            daqTbl[i].kwf = kw;
            i = nrecords; /* next 'k' */
            break;
          }
        }
      }
    }
  }
  dbLoadTime = time(NULL);
err:
  if (cc) 
  {
    sprintf(s,"init error: %.32s",erlst[cc]);
    feclog(s);
  }    
  return cc;
}
int daqReload(int argc,char *argv[])
{
  int i;
  ExportListStruct *el = getExportListItem(MEX_EQMNAME);
  argc = argc; argv = argv;
  feclog("reload DAQ database");
  for (i=0; i<nrecords; i++) CloseLink(recordLinkIdSts[i].id);
  linksEstablished = FALSE;
  SystemDelay(1000);
  /* flushStructCache(); */
  freeDaqDb();
  freeFailoverInformation();
  free(recordLinkIdSts); recordLinkIdSts = NULL;
  rmvDaqRecKeys();
  /* we should be back to square one now */
  clearCASAlarmList();
  FlushFecAddrCache();
  daqInit(srvinfo.dbfile[0],srvinfo.ndbfiles);
  if (el != NULL)
  { /* force a re-export in case of 'who's master? problem' */
    el->registered = 0;
    ExportsRegistered = FALSE;
  }
  return 0;
}
void wdCallback(int id,int cc)
{
  static int failcnt = 0;
  if (cc) failcnt++; else failcnt = 0;
  if (failcnt > 2) registerGlobals = TRUE;
} 
void mexexit(void)
{
  freeDaqDb();
  freeFailoverInformation();
  free(recordLinkIdSts); recordLinkIdSts = NULL;
}
void addPartnerToAcls(void)
{
  NAME16 usrs[100];
  int i, nusrs=100;

  if (GetRegisteredUsers(MEX_EQMNAME,usrs,&nusrs) == 0 && nusrs > 0)
  { /* there is a users list ! */
    char *c, fecpartner[64];
    strncpy(fecpartner,GetExportedFecName(),FEC_NAME_SIZE);
    strupr(fecpartner);
    if ((c=strstr(fecpartner,"MARCH")) != NULL)
    {
      *c = 0;
      strcat(fecpartner,"HISTORY");
      for (i=0; i<nusrs; i++) if (strnicmp(usrs[i].name,fecpartner,16) == 0) break;
      if (i == nusrs)
      { /* not found in users list: add it */
        AppendRegisteredUsers(MEX_EQMNAME,(NAME16 *)fecpartner,1);
      }
    }
  }
}
void mexinit(void)
{
  DTYPE dout;
  char *data=NULL;
  int i, j, k, cc=0, nusrs=100;

  SetAutoLinkErrorAlarms(FALSE); /* turn off */
  RegisterDeviceName(MEX_EQMNAME,"keyword",0);

  getAdministrators(ADMINS_FILE);

  remove("problems.log");
  RegisterUserCommand("ArchiveDebug",NULL,&archdebug,NULL,CA_READ+CA_WRITE);
  RegisterUserCommand("FilterTable",dumpFilterTable,NULL,NULL,CA_READ);
  RegisterUserCommand("ArchiveRecord",dumpArchiveRecord,NULL,NULL,CA_READ);
  RegisterUserFunction("reload",daqReload);
  RegisterUserFunction("failovers",dumpFailoverStatus);

  getMinFreeDiskSpace("freespace.csv");
  SetFreeBlocksAlarmThreshold(ArchiveDBpath,gMininumFreeDiskSpace);

  daqInit(srvinfo.dbfile[0],srvinfo.ndbfiles);
  nredirfiles = getRedirInfo();
  redirInit(redirfile,redirsrv,nredirfiles);

  getNetGlobalList(GLOBALS_FILENAME);
  if (ngloballist > 0)
  { /* found a registered globals file */
    feclog("found %d globals from list %s",ngloballist,GLOBALS_FILENAME);
    for (i=0; i<nrecords; i++) for (j=0; j<daqTbl[i].nparam; j++)
      for (k=0; k<ngloballist; k++)
       if (!stricmp(daqTbl[i].daqParam[j].param,globalList[k].keyword))
    { /* fix the format and size */
      globalList[k].format = daqTbl[i].daqParam[j].fmt;
      globalList[k].size = daqTbl[i].daqParam[j].size;
    }
  }
  else if (globalsServer || globalsWatchdog)
  { /* I'm a globals server or watchdog but don't have an explicit globals list => use the database as is */
    if ((globalList=(GlobalListStruct *)calloc(nparams,sizeof(GlobalListStruct))) == NULL) ccerr(out_of_server_memory);
    for (i=0,k=0; i<nrecords; i++) for (j=0; j<daqTbl[i].nparam; j++)
    {
      strncpy(globalList[k].keyword,daqTbl[i].daqParam[j].param,KEYWORD_NAME_SIZE);
      globalList[k].format = daqTbl[i].daqParam[j].fmt;
      globalList[k].size = daqTbl[i].daqParam[j].size;
      globalList[k].maxPeriod = 1000;
      globalList[k].minPeriod = 1000;
      globalList[k].tolerance = 0;
      k++;
    }
    ngloballist = k;
  }
  if (globalsServer) registerGlobals = TRUE;
  if (globalsWatchdog)
  {
    if (ngloballist > 0)
    { /* key on the first in the list */
      dout.dFormat = globalList[0].format;
      dout.dArrayLength = globalList[0].size;
      dout.data.vptr = data;
      if ((cc=recvNetGlobal(globalList[0].keyword,&dout,wdCallback)) < 0)
      {
        feclog("recvNetGlobal  : %s",erlst[-cc]);
      } 
    }
  }
  /* any aliases or redirections ? */
  getAdditionalKeywords();
  ensureGlobalsProperties();
err:
  if (cc != 0) feclog("mexinit : %s",erlst[cc]);
  return;
}
int thisWorkAreaSize = MAX_TRANSPORTSIZE*4;
int thisSocketRecvBufferSize = MAX_RCVBUF;
int thisSocketSendBufferSize = MAX_SNDBUF;
void term_handler(int signum)
{
  g_isTerminating = TRUE;
  feclog("caught SIGTERM");
}
void PreSystemInit(void)
{
  char *ptr;
  if (getStartupInfo()) 
  {
    dbglog("Cannot read configuration database -> abort !");
    exit(2);
  }
  loadStateInfo();
  canSendGlobals = MEX_ISGLOBALSERVER;
  if (canSendGlobals)
  {
    gIsGlobalsServer = TRUE;
    if (strlen(globalsMcastAddr) > 0) SetGCastAddr(globalsMcastAddr);
    GlobalHeartbeat = 3;
    if (!globalsWatchdog) 
    {
      globalsServer = TRUE; 
      if ((ptr=getenv("FORCE_MULTICAST")) != NULL && stricmp(ptr,"TRUE") == 0)
      {
        requireNetworkAccess = TRUE;
      }
    }
  }
  daqEqmName = MEX_EQMNAME;
  gScanForNameLists = FALSE;
  MaxNumClients = 1000;
  MaxNumContracts = 10000; 
  MaxNumConnections = 10000;
  FeclogDepth = 2000;
  SetAllowDependentLinks(FALSE);
  SetWorkAreaSize(thisWorkAreaSize);
  SystemAssignBufferSpaceEx(thisSocketRecvBufferSize,thisSocketRecvBufferSize/2,thisSocketSendBufferSize/2,thisSocketSendBufferSize);
  if (thisSocketSendBufferSize > MAX_SNDBUF) _SystemAssignUdpSendBufferSpace(thisSocketSendBufferSize);
  SetClnRecvQueueDepth(0);
  if (startupInputsGiven)
  {
    int p;
    if (strlen(thisContext) == 0)
    {
      strcpy(thisContext,"TEST");
    }
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      if (gIsGlobalsServer)
        strcat(thisFecName,"GLOBAL");
      else
        strcat(thisFecName,"MARCH");
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s Central archive server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
  else
  {
    strncpy(thisContext,gDeviceContext,CONTEXT_NAME_SIZE);
  }
  return;
}
ADS misGlbAds = 
{
  "Missing Global",
  ALMCODE_GLB,
  0,
  0,
  ALM_SEVERITY_WARN,
  BFMT(CF_TEXT),
  64,
  0, 0,
  "a global keyword is not available",
  "no device",
  "global keyword which is missing",
  ""
};
ADS fioAds =
{
  "archive write error",
  ALMCODE_FIO,
  0,
  0,
  ALM_SEVERITY_WARN,
  BFMT(CF_TEXT),
  64,
  0, 0,
  "cannot write archive data to file",
  "no device",
  "file/path which cannot be written",
  ""
};
ADS fcorruptAds =
{
  "archive file corrupt",
  ALMCODE_CORRUPT,
  0,
  0,
  ALM_SEVERITY_WARN,
  BFMT(CF_TEXT),
  64,
  0, 0,
  "archive data file appears to be corrupt",
  "no device",
  "file/path which is corrupt",
  ""
};
ADS nonexistAds =
{
  "archive source nonexistent",
  ALMCODE_NONEXIST,
  0,
  0,
  ALM_SEVERITY_WARN,
  BFMT(CF_TEXT),
  64,
  0, 0,
  "keyword maps to nonexistent source",
  "no device",
  "keyword which maps to nonexistent source",
  ""
};
void PostSystemInit(void)
{
  int cc;
  DTYPE dprp, din;
  char mexExpName[128];

  signal(SIGTERM,term_handler);
  signal(SIGFPE,SIG_IGN); /* catch floating point domain errors */
  gRequireAcknowledgments = 0; /* turn off ack. requirments for REFRESH mode */
  strncpy(mexExpName,thisServerName,EXPORT_NAME_SIZE);
  if (globalsWatchdog) strcat(mexExpName,".WD");
  if ((cc=RegisterEquipmentModule(mexExpName,MEX_EQMNAME,MEX_NUMDEVICES,mexeqm,mexinit,mexbkg,1000,mexexit)) != 0)
  {
    feclog("Can't register module %s : %s",MEX_EQMNAME,erlst[cc]);
  }
  else
  {
    dprp.dFormat = CF_LONG; dprp.dArrayLength = 1; dprp.dTag[0] = 0;
    RegisterPropertyInformation(MEX_EQMNAME,"NKEYWORDS",&dprp,NULL,CA_READ,AT_SCALAR,0,"number of available machine parameters",-1,NULL);
    RegisterPropertyInformation(MEX_EQMNAME,"NTAGS",&dprp,NULL,CA_READ,AT_SCALAR,0,"number of available machine parameters",-1,NULL);
    dprp.dFormat = CF_NAME32; dprp.dArrayLength = 512; dprp.dTag[0] = 0;
    RegisterPropertyInformation(MEX_EQMNAME,"KEYWORDS",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"list of available machine parameters",-1,NULL);
    RegisterPropertyInformation(MEX_EQMNAME,"TAGS",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"list of available machine parameters",-1,NULL);
    RegisterProperty(MEX_EQMNAME,"FILTERTABLE",32,CF_NAME64I,CA_READ,"read filter table (with validity status)");
    if (MEX_ISGLOBALSERVER)
    {
      dprp.dFormat = CF_LONG; dprp.dArrayLength = 1; dprp.dTag[0] = 0;
      RegisterPropertyInformation(MEX_EQMNAME,"NGLOBALS",&dprp,NULL,CA_READ,AT_SCALAR,0,"number of registered global parameters",-1,NULL);
      dprp.dFormat = CF_NAME64; dprp.dArrayLength = 512; dprp.dTag[0] = 0;
      RegisterPropertyInformation(MEX_EQMNAME,"GLOBALS",&dprp,NULL,CA_READ|CA_WRITE,AT_UNKNOWN,0,"registered global parameters",-1,NULL);
    }
    if ((globalsServer || globalsWatchdog) && !stricmp(gDeviceContext,"SITE"))
    { /* redirect queries to SYSTIME to the time server (in case I'm the SITE globals server) */
      dprp.dFormat = CF_DOUBLE; dprp.dArrayLength = 1; dprp.dTag[0] = 0;
      RegisterPropertyInformation(MEX_EQMNAME,"SYSTIME",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"[1E10:1E11 sec]System Time",-1,"/SITE/TIMESRV");
    }
    /* database access properties */
    dprp.dFormat = CF_STRUCT; dprp.dArrayLength = 1000; strncpy(dprp.dTag,DAQRecTag,TAG_NAME_SIZE);
    RegisterPropertyInformation(MEX_EQMNAME,"RECORD.INFO",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"Record information database",-1,NULL);
    dprp.dFormat = CF_INTINT; dprp.dArrayLength = 1000; dprp.dTag[0] = 0;
    din = dprp; din.dArrayLength = 1;
    RegisterPropertyInformation(MEX_EQMNAME,"RECORD.ANNOTATION",&dprp,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"Record Active Annotations",-1,NULL);
    dprp.dFormat = CF_STRUCT; dprp.dArrayLength = 100; strncpy(dprp.dTag,DAQPrmTag,TAG_NAME_SIZE);
    RegisterPropertyInformation(MEX_EQMNAME,"KEYWORD.INFO",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword information database",-1,NULL);
    dprp.dFormat = CF_INT32; dprp.dArrayLength = 100; memset(dprp.dTag,0,TAG_NAME_SIZE);
    RegisterPropertyInformation(MEX_EQMNAME,"KEYWORD.INPT",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword input",-1,NULL);
    dprp.dFormat = CF_TEXT; dprp.dArrayLength = 256; memset(dprp.dTag,0,TAG_NAME_SIZE);
    RegisterPropertyInformation(MEX_EQMNAME,"KEYWORD.SOURCE",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword source",-1,NULL);
    dprp.dFormat = CF_DBLDBL; dprp.dArrayLength = 1; memset(dprp.dTag,0,TAG_NAME_SIZE);
    RegisterPropertyInformation(MEX_EQMNAME,"KEYWORD.TOLERANCE",&dprp,NULL,CA_READ,AT_UNKNOWN,0,"Record keyword tolerance",-1,NULL);
    RegisterPropertyInformation(MEX_EQMNAME,"DB.RELOAD",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"reload the database",-1,NULL);
    /* just for a test */
    RegisterPropertyInformation(MEX_EQMNAME,"DB.DUMP",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"back out the running database",-1,NULL);

    RegisterProperty(MEX_EQMNAME,"DB.LASTLOADED",1,CF_INT32,CA_READ,"most recent archive database load time (UTC)");
    RegisterProperty(MEX_EQMNAME,"DB.LASTLOADED",32,CF_TEXT,CA_READ,"most recent archive database load time (string)");

    AppendAlarmInfoTable(MEX_EQMNAME,&fioAds);
    AppendAlarmInfoTable(MEX_EQMNAME,&fcorruptAds);
    AppendAlarmInfoTable(MEX_EQMNAME,&nonexistAds);
  }
  if (globalsServer || globalsWatchdog) AppendAlarmInfoTable(MEX_EQMNAME,&misGlbAds);
  feclog("Check incoming state information : %s",checkStates ? "TRUE" : "FALSE");
  feclog("Monitoring %d checked states",nCheckedStates);
  feclog("Monitoring %d checked state parameters",nCheckedParameters);
  feclog("Monitoring %d declared e procedures",nEDeclaredProcedures);
  feclog("Monitoring %d declared p procedures",nPDeclaredProcedures);
  feclog("Is Globals Server : %s", globalsServer ? "TRUE" : "FALSE");
  feclog("Is Globals Watchdog : %s", globalsWatchdog ? "TRUE" : "FALSE");
  feclog("Using %s as targeted State Server",gTargetStateServer);

  if (strlen(thisFailoverMasterName) > 0)
  {
    if (strlen(thisFailoverSlaveMasterName) > 0)
    { /* we're a failover slave ... */
      SetFailoverSlave(MEX_EQMNAME,thisFailoverMasterName,thisFailoverSlaveMasterName);
    }
    else
    { /* we're a failover master ... */
      SetFailoverMaster(MEX_EQMNAME,thisFailoverMasterName);
    }
  }
}
void checkDateChange(void)
{
  int i;
  double ts = MakeDataTimeStamp();
  time_t t = time(NULL); // logRestart() uses this 
  struct tm tm = *localtime(&t);
  static int mday = -1;

  if (mday != tm.tm_mday) /* new day */
  {
    AppendHistoryLog("date change (day %d to day %d",mday,tm.tm_mday);
    mday = tm.tm_mday;
    logRestart();
    for (i=0; i<nrecords; i++)
    {
      if (daqTbl[i].dataFlg < 1) continue;
      if (daqTbl[i].option&ONCE)
      {
        daqTbl[i].archtime = daqTbl[i].tstamp = ts;
        daqTbl[i].archPending = 1;
      }
    }
  }
  return;
}
int getRegisteredKeyword(char *dst,char *src)
{
  int i;
  strncpy(dst,src,PROPERTY_NAME_SIZE);
  for (i=0; i<nPrpAlias; i++)
  {
    if (!strnicmp(prpAlias[i].name,src,PROPERTY_NAME_SIZE))
    {
      strncpy(dst,prpAlias[i].alias,PROPERTY_NAME_SIZE);
      return i;
    }
  }
  return 0;
}
int getIndexFromDeviceName(char *devName,char *devProperty)
{
  DAQParameter *daqParam;
  int i, len, siz, n, p, rec, ptr;
  char namProperty[PROPERTY_NAME_SIZE+4] = "";
 
  if (devName[0] == 0) return 0;
  if (strnicmp(&devProperty[strlen(devProperty)-4],".NAM",4) == 0)
  { /* is a .NAM property */
    return 0;
  }
  if (devName[0] == '#')
  {
    if ((i=atoi(&devName[1])) < 0) i = 0;
    return i;
  }
  strncpy(namProperty,devProperty,PROPERTY_NAME_SIZE-4);
  strcat(namProperty,".NAM");
  for (i=0; i<nPrpAlias; i++)
  {
    if (!strnicmp(prpAlias[i].name,namProperty,PROPERTY_NAME_SIZE))
    {
      strncpy(namProperty,prpAlias[i].alias,PROPERTY_NAME_SIZE);
      break;
    }
  }
  if ((n=getEntryFromKeyword(namProperty,&rec,&p)) < 0) return -non_existent;
  daqParam = daqTbl[n].daqParam;
  siz = daqParam[p].size; 
  ptr = daqTbl[n].daqParam[p].offset;
  switch (daqTbl[n].daqParam[p].fmt)
  {
    case CF_USTRING:
      ptr += 16;
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      len = fmtsizeof(daqTbl[n].daqParam[p].fmt);
      for (i=0; i<siz; i++)
        if (!strnicmp(devName,(char *)&daqTbl[n].data[ptr+len*i],MIN(len,DEVICE_NAME_SIZE))) return i;
      break;
    default:
      return -illegal_format;
  }
  return -illegal_equipment_number;
}
int getDeviceAndSubDevice(char *devName,char *devProperty,int *subindex)
{
  int idx, sidx=-1, rec, i, p, n, rlen=0, slen, len=0;
  DAQParameter *daqParam;
  char *sub, dev[128], prp[76], *dptr;
  
  if ((n=getEntryFromKeyword(devProperty,&rec,&p)) < 0) return -non_existent;

  daqParam = daqTbl[n].daqParam;
  strncpy(dev,devName,128);
  if ((sub=strchr(dev,'/')) != NULL)
  {
    *sub++ = 0;
    len = (int)strlen(sub);
    rlen = daqParam->rowsize;
  }
  if ((idx=getIndexFromDeviceName(dev,devProperty)) < 0) return -idx;
  if (daqParam->arrayType & AT_DOUBLE && len > 0)
  { /* makes sense to look for a companion entry */
    char key[76];
    strncpy(prp,devProperty,56);
    strcat(prp,".NAM.NAM");
    getRegisteredKeyword(key,prp);
    if ((n=getEntryFromKeyword(key,&rec,&p)) >= 0)
    { /* as expected: a companion .NAM.NAM entry */
      daqParam = daqTbl[n].daqParam;
      slen = fmtsizeof(daqParam->fmt);
      dptr = daqTbl[n].data + daqParam->offset;
      for (i=0; i<daqParam->size; i++)
      { /* run thru the list */
        if (strnicmp(sub,dptr + i*slen,len) == 0) 
        { /* got'm */
          sidx = i;
          idx = idx * rlen + sidx;
          break;
        }
      }
    }
  }
  if (subindex != NULL) *subindex = sidx;
  return idx;
}
int mexeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc, i, j, n, r, ival, flen, slen, soff, ioff, doff, bptr,hasInput=FALSE;
  char prop[PROPERTY_NAME_SIZE],dev[DEVICE_NAME_SIZE],keyword[PROPERTY_NAME_SIZE+4];
  char errstr[64], usr[USERNAME_SIZE+4], sstr[256];
  int dsize,index = 0,fmt;
  int fmtsiz, limit, sidx;
  float fvals[2];
  KeywordFailover *kw;
  NAME16 caller;
  BYTE *mcab;

# ifdef NO_ASYNC
  if ((access & CA_FIRST) && (access & CA_LAST)) /* blow synchronous calls out of the water */
  {
    return not_accepted;
  }
# endif
  n = 1; GetCallerInfo(MEX_EQMNAME,&caller,NULL,NULL,NULL,&n);
  strncpy(usr,caller.name,16); usr[16] = 0; strupr(usr);
  /* make local copies of the incoming information ... */
  strncpy(dev,devName,DEVICE_NAME_SIZE);
  if (!stricmp(devName,"keyword")) strcpy(dev,"#0");
  strncpy(prop,devProperty,PROPERTY_NAME_SIZE);
  /* maybe it's an alias property ? */
  for (i=0; i<nPrpAlias; i++)
  {
    if (!strnicmp(prpAlias[i].name,devProperty,PROPERTY_NAME_SIZE))
    {
      strncpy(prop,prpAlias[i].alias,PROPERTY_NAME_SIZE);
      break;
    }
  }

  if ((cc=overloadEqpmod(dev,prop,dout,din,access)) != illegal_property) return cc;
  fmt = dout->dFormat;
  dsize = (int)dout->dArrayLength;
  cc = 0;
  if (!stricmp(prop,"RECORD.INFO"))  
  { /* DAQRecInfo database */
#   ifdef ALLOW_ARCHIVER_ACCESS_TO_DATABASE_UPDATES
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr))
      {
        char es[256];
        sprintf(es,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(MEX_EQMNAME,es);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,DAQRecTag) != 0) return illegal_format;
      cc = addDaqRecInfoElements((DAQRecInfo *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
    else 
#   endif
    if (din->dArrayLength > 0)
    { /* not a WRITE but it does have input specifying particular record(s) */
      hasInput = TRUE;
      cc = GetValuesAsStringEx(din,keyword,CF_TEXT,PROPERTY_NAME_SIZE,0);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    {
      DAQRecInfo *dqr = (DAQRecInfo *)dout->data.vptr;
      if (dout->dArrayLength > (UINT32)totalNrRecords) dout->dArrayLength = totalNrRecords;
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag,DAQRecTag) != 0) return illegal_format;
      if (hasInput) cc = getDaqRecInfoEx(dqr,&dout->dArrayLength,keyword);
      else          cc = getDaqRecInfo(dqr,&dout->dArrayLength);
    }
    return cc;
  }
  if (!stricmp(prop,"KEYWORD.INFO"))  
  { /* DAQRecInfo database */
#   ifdef ALLOW_ARCHIVER_ACCESS_TO_DATABASE_UPDATES
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr))
      {
        char es[256];
        sprintf(es,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(MEX_EQMNAME,es);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,DAQPrmTag) != 0) return illegal_format;
      cc = addDaqPrmInfoElem(devName,(DAQPrmInfo *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
#   endif
    if (dout->dArrayLength > 0)
    {
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag,DAQPrmTag) != 0) return illegal_format;
      n = GetDeviceNumber(MEX_EQMNAME,devName);
      if (n < 0)
      {
        cc = readDaqPrmFile(ArchiveDBpath,devName,(DAQPrmInfo *)dout->data.vptr,(int *)&dout->dArrayLength);
      }
      else
      {
        cc = getDaqPrmInfo(n,(DAQPrmInfo *)dout->data.vptr,(int *)&dout->dArrayLength);
      }
    }
    return cc;
  }
  if (!stricmp(prop,"KEYWORD.INPT"))  
  { /* DAQRecInfo database */
#   ifdef ALLOW_ARCHIVER_ACCESS_TO_DATABASE_UPDATES
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr))
      {
        char es[256];
        sprintf(es,"access denied: user %.16s is not an administrator",usr);
        SetEqmCompletion(MEX_EQMNAME,es);
        return access_denied;
      }
      if (din->dArrayLength == 0) return illegal_read_write;
      cc = addDaqInpt(devName,din->data.bptr,(int)din->dFormat,(int)din->dArrayLength);
      if (cc != 0) return cc;
    }
#   endif
    if (getEntryFromKeywordFile(devName,&n) < 0)
    { /* must then be entered as #<rec number> */
      n = GetDeviceNumber(MEX_EQMNAME,devName);
    }
    cc = getDaqPrmInpt(n,dout);
    return cc;
  }
  if (!stricmp(prop,"KEYWORD.SOURCE"))  
  { /* DAQRecInfo database */
    int fecid, srvid, isDoocsSrc=FALSE;
    strncpy(prop,dev,PROPERTY_NAME_SIZE);
    i = getEntryFromKeyword(prop,&r,&n);
    if (i < 0) 
    {
      if ((i=getRedirectedIndex(prop)) >= 0)
      {
        char *c, *ctx, *srv, *prp;
        strncpy(sstr,additionalKeywords[i].redirection,DEVICE_REDIR_SIZE);
        ctx = sstr;
        if (ctx[0] == '/')
        {
          ctx++;
          if ((srv=strchr(ctx,'/')) == NULL) return non_existent;
          *srv++ = 0;
          if ((prp=strchr(srv,'[')) == NULL) return non_existent;
          *prp++ = 0;
          if ((c=strchr(prp,']')) != NULL) *c = 0;
          if ((c=strchr(srv,'/')) != NULL) *c = 0;
        }
        SetRedirectionParameters(MEX_EQMNAME,ctx,srv,prp,devProperty);
        return server_redirection;
      }
      return non_existent;
    }
    fecid = locateExport(daqTbl[i].srv,daqTbl[i].ctx,daqTbl[i].dev,&srvid);
    if (fecid < 0) return fecid;
    if (!strncmp(SrvTbl[srvid].EqmName,DOOCSEQM_TAG,EQM_NAME_SHORTSIZE)) isDoocsSrc = TRUE;
    if ((j=daqTbl[i].daqParam[n].srcIndex) > 0)
    {
      i = j-1;
      strncpy(prop,daqTbl[i].daqParam[0].param,PROPERTY_NAME_SIZE);
      goto theChase;
    }
    else if (isDoocsSrc) 
    {
      int k;
      char subs[16];
      DUSTRING *us=NULL,*tus=NULL;
      if (!isStringFormat(dout->dFormat) || dout->dFormat == CF_TEXT) return illegal_format;
      if (strlen(prop) > 60) return invalid_keyword;
      if (daqTbl[i].daqParam[j].arrayType & AT_CHANNEL)
      {
        if (strlen(daqTbl[i].daqParam[j].assoc) > 0) 
        {
          strncpy(prop,daqTbl[i].daqParam[j].assoc,PROPERTY_NAME_SIZE);
        }
        else
        {
          strcat(prop,".NAM");
        }
        subs[0] = 0;
        if ((n=getEntryFromKeyword(prop,&r,&j)) < 0) return -n;
        us = (DUSTRING *)(&daqTbl[n].data[daqTbl[n].daqParam[0].offset]);
      }
      getSubSysDecoration(daqTbl[i].daqParam[j].subsys,subs);
      if (us == NULL)
      {
        if (dout->dFormat != CF_TEXT) dout->dArrayLength = 1;
        sprintf(sstr,"/%.32s%s/%.32s/%.64s[%.64s]",
          daqTbl[i].ctx,subs,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
        return PutValuesFromString(dout,sstr,CF_TEXT, (int)strlen(sstr),0);
      }
      if ((tus=(DUSTRING *)alloca(daqTbl[i].daqParam[j].size*fmtsizeof(CF_USTRING))) == NULL)
        return out_of_local_memory;
      for (k=0; k<(int)dout->dArrayLength && k<daqTbl[i].daqParam[j].size; k++)
      {
        if (strlen(us[k].str) == 0) break;
        sprintf(sstr,"/%.32s%s/%.32s/%.64s[%.64s]",
          daqTbl[i].ctx,subs,daqTbl[i].srv,us[k].str,daqTbl[i].prp);
        strncpy(tus[k].str,sstr,80);
      }
      cc = PutValuesFromString(dout,(char *)tus,CF_USTRING,k,0);
      dout->dArrayLength = (UINT32)k;
      return cc;
    }
    else
    {
      sprintf(sstr,"/%.32s/%.32s/%.64s[%.64s]",
        daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
      if (dout->dFormat != CF_TEXT) dout->dArrayLength = 1;
      cc = PutValuesFromString(dout,sstr,CF_TEXT, (int)strlen(sstr),0);
    }
    return cc;
  }
  if (!stricmp(devProperty,"KEYWORD.TOLERANCE"))  
  { /* keyword tolerance (for given interval) */
    if (dout->dArrayLength < 1) return dimension_error;
    strncpy(prop,dev,PROPERTY_NAME_SIZE);
    i = getArchivedEntryFromKeyword(prop,&r,&n);
    if (i < 0) 
    {
      if ((i=getRedirectedIndex(prop)) >= 0)
      {
        char *c, *ctx, *srv, *prp;
        strncpy(sstr,additionalKeywords[i].redirection,DEVICE_REDIR_SIZE);
        ctx = sstr;
        if (ctx[0] == '/')
        {
          ctx++;
          if ((srv=strchr(ctx,'/')) == NULL) return non_existent;
          *srv++ = 0;
          if ((prp=strchr(srv,'[')) == NULL) return non_existent;
          *prp++ = 0;
          if ((c=strchr(prp,']')) != NULL) *c = 0;
          if ((c=strchr(srv,'/')) != NULL) *c = 0;
        }
        SetRedirectionParameters(MEX_EQMNAME,ctx,srv,prp,devProperty);
        return server_redirection;
      }
      return non_existent;
    }
    fvals[0] = daqTbl[i].daqParam[n].aTolerance;
    fvals[1] = daqTbl[i].daqParam[n].pTolerance;
    switch (dout->dFormat)
    {
      case CF_DBLDBL:
        ((DBLDBL *)dout->data.dptr)[0].d1val = (double)fvals[0];
        ((DBLDBL *)dout->data.dptr)[0].d2val = (double)fvals[1];
        dout->dArrayLength = 1;
        return 0;
      case CF_FLTFLT:
        memcpy(dout->data.fptr,fvals,2*sizeof(float));
        dout->dArrayLength = 1;
        return 0;
      default:
        if (dout->dArrayLength > 2) dout->dArrayLength = 2;
        return PutValuesFromFloat(dout,fvals,2);
    }
    return 0;
  }
  if (!stricmp(devProperty,"FILTERTABLE"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    flen = fmtsizeof(dout->dFormat);
    switch (dout->dFormat)
    {
      case CF_NULL:
        dout->dArrayLength = 0;
        return 0;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        /* just a names list */
        slen = flen;
        soff = 0;
        ioff = -1;
        break;
      case CF_NAME8I:
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME48I:
      case CF_NAME64I:
        /* a names list with current validity status */
        slen = flen - 4;
        soff = 0;
        ioff = slen;
        break;
      case CF_STRUCT:
        if (strcmp(dout->dTag,DAQFltrTag)) return invalid_structure_tag;
        return getDaqFltrTbl((DAQFilter *)dout->data.vptr,(int *)&dout->dArrayLength);
      default: 
        return illegal_format;
    }
    if (dout->dArrayLength > (UINT32)nDaqFilterTblEntries) dout->dArrayLength = nDaqFilterTblEntries;
    for (i=0,n=0; i<nDaqFilterTblEntries && n<(int)dout->dArrayLength; i++)
    {
      strncpy((char *)(dout->data.bptr+n*flen),daqFilterTbl[i].tag,slen);
      if (ioff >= 0) 
      { /* send the current valid state the call */
        ival = daqFilterTbl[i].valid;
        memcpy(dout->data.bptr+n*flen + ioff,&ival,4);
      }
      n++;
    }
    return 0;
  }
  if (!stricmp(prop,"DB.RELOAD"))  
  { /* dump database */
    if (!isAdministrator(usr))
    {
      char es[256];
      sprintf(es,"access denied: user %.16s is not an administrator",usr);
      SetEqmCompletion(MEX_EQMNAME,es);
      return access_denied;
    }
    if (access & CA_WRITE) return daqReload(0,NULL);
    return illegal_read_write;
  }
  if (!stricmp(prop,"DB.DUMP"))  
  { /* dump database */
    if (access & CA_WRITE) return dumpDaqDb("./","dbtest.csv",TRUE);
    return illegal_read_write;
  }
  if (!stricmp(devProperty,"DB.LASTLOADED"))
  {
    if ((access & CA_WRITE)) return illegal_read_write;
    if (isNumberFormat(dout->dFormat))
    {
      UINT32 t = (UINT32)dbLoadTime;
      return PutValuesFromLong(dout,&t,1);
    }
    if (isStringFormat(dout->dFormat))
    {
      char dbLoadTimeStr[32];
      sprintf(dbLoadTimeStr,"%.24s",ctime(&dbLoadTime));
      return PutValuesFromString(dout,dbLoadTimeStr,CF_TEXT,32,0);
    }
    return illegal_format;
  }
  if (!stricmp(prop,"STATUS"))
  {
    if (dsize < 1 || dsize > totalNrRecords) return dimension_error;
    if (fmt != CF_SHORT) return illegal_format;
    for (i = 0; i<dsize; i++) (dout->data.sptr)[i] = daqTbl[i].daqStatus;
    return 0;
  }
  if (!stricmp(prop,"NGLOBALS"))
  {
    if (!canSendGlobals) return not_supported;
    putValuesFromLong(dout,(SINT32 *)&ngloballist,1);
    return 0;
  }
  if (!stricmp(prop,"GLOBALS"))
  {
    if (!canSendGlobals) return not_supported;
    if (access & CA_WRITE)
    {
      if (din->dArrayLength == 0) return illegal_read_write;
      if (din->dFormat != CF_NAME64) return illegal_format;
      cc = putNetGlobalList(GLOBALS_FILENAME,(NAME64 *)din->data.vptr,(int)din->dArrayLength);
      if (cc != 0) return cc;
    }
    if (dsize > 0)
    {
      switch (fmt)
      {
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          flen = fmtsizeof(fmt);
          break;
        default:
          return illegal_format;
      }
      if (dout->dArrayLength > (UINT32)ngloballist) dout->dArrayLength = (UINT32)ngloballist;
      for (i=0; i<(int)dout->dArrayLength; i++)
      {
        strncpy(dout->data.cptr + i*flen,globalList[i].keyword,flen);
      }
    }
    return cc;
  }
  if (!stricmp(prop,"NKEYWORDS") || !stricmp(prop,"NTAGS"))
  {
    char *tgtsrv = NULL;
    if (din->dArrayLength > 0)
    {
      if (!isSimpleStringFormat(din->dFormat)) return illegal_format;
      tgtsrv = din->data.cptr;

    }
    if (dsize != 1) return dimension_error;
    for (i=0, n=0; i < nrecords; i++) if (daqTbl[i].active)
    {
      if (tgtsrv != NULL && strnicmp(tgtsrv, daqTbl[i].srv, EXPORT_NAME_SIZE)) continue;
      n += daqTbl[i].nparam;
    }
    if (tgtsrv == NULL) n += nAdditionalKeywords;
    putValuesFromLong(dout,(SINT32 *)&n,1);
    return 0;
  }
  if (!stricmp(prop,"KEYWORDS") || !stricmp(prop,"TAGS")) 
  {
    char *tgtsrv = NULL;
    if (din->dArrayLength > 0)
    {
      if (!isSimpleStringFormat(din->dFormat)) return illegal_format;
      tgtsrv = din->data.cptr;

    }
    flen = fmtsizeof(fmt);
    switch (fmt)
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        /* just a names list */
        slen = flen;
        soff = 0;
        ioff = -1;
        break;
      case CF_NAME8I:
      case CF_NAME16I:
      case CF_NAME32I:
      case CF_NAME48I:
      case CF_NAME64I:
        /* a names list with default format */
        slen = flen - 4;
        soff = 0;
        ioff = slen;
        break;
      case CF_USTRING:
        /* a names list with default format */
        slen = 64; /* properties are not longer than 64 chars (USTRING has 80) */
        soff = 16;
        ioff = 0;
        break;
      default: return illegal_format;
    }
    for (i=0,j=0; j<dsize && i<nrecords; i++) if (daqTbl[i].active)
    {
      if (tgtsrv != NULL && strnicmp(tgtsrv, daqTbl[i].srv, EXPORT_NAME_SIZE)) continue;
      for (n=0; j<dsize && n<daqTbl[i].nparam; n++,j++)
      {
        strncpy((char *)(dout->data.bptr + j*flen),daqTbl[i].daqParam[n].param,slen);
        if (ioff >= 0) 
        { /* send the default format with the call */
          ival = daqTbl[i].daqParam[n].fmt;
          memcpy(dout->data.bptr + j*flen + ioff,&ival,4);
        }
      }
    }
    if (tgtsrv == NULL) for (i=0; i<nAdditionalKeywords; i++)
    {
      if (strcmp(additionalKeywords[i].EqmName,MEX_EQMNAME)) continue;
      if (dout->dArrayLength < (UINT32)j) break;
      strncpy((char *)(dout->data.bptr+j*flen),additionalKeywords[i].PrpName,slen);
      if (ioff >= 0) 
      { /* send the default format with the call */
        ival =  ftoi(additionalKeywords[i].fmtOut);
        memcpy(dout->data.bptr+j*flen + ioff,&ival,4);
      }
      j++;
    }
    dout->dArrayLength = j;
    return 0;
  }
  /* Below: the requested devProperty is a keyword ... */
  for (kw=kwrdFailover; kw != NULL; kw=kw->nxt)
  { /* check the failover list */
    if (!strnicmp(kw->keyword,prop,PROPERTY_NAME_SIZE))
    { /* seems to be registered failover keyword */
      for (j=0; j<3; j++)
      {
        i = kw->foSrc[j].recNr; n = kw->foSrc[j].prmNr;
        if (i < 0) continue;
        if (daqTbl[i].daqStatus != 0) continue;
        strncpy(prop,daqTbl[i].daqParam[n].param,PROPERTY_NAME_SIZE);
        goto theChase;
      }
      if (j == 3) return link_error;
    }
  }
  /* return the current data value requested */
  if ((index=getIndexFromDeviceName(dev,prop)) < 0 &&
      (index=getDeviceAndSubDevice(dev,prop,&sidx)) < 0) return -index;
  for (i=0; i<nrecords; i++) if (daqTbl[i].active)
  {
    for (n=0; n<daqTbl[i].nparam; n++)
    {
theChase:
      if (!stricmp(daqTbl[i].daqParam[n].param,prop))
      {
        if (daqTbl[i].daqStatus != 0 && daqTbl[i].errCounter > ERROR_APPLY_STATUS_COUNT)
        {
          cc = daqTbl[i].daqStatus;
          if (cc < 512) 
          {
            strncpy(errstr,cc2str(cc),32);
            cc += 1000;
          }
          else
          {
            sprintf(errstr,"error %d",cc);
          }
          SetEqmCompletion(MEX_EQMNAME,errstr);
          return cc;
        }
        if (fmt == CF_TEXT)
        {
          switch(daqTbl[i].daqParam[n].fmt)
          {
            case CF_CHAR8:
            case CF_CHAR16:
            case CF_CHAR32:
            case CF_CHAR48:
            case CF_CHAR64:
              if (dsize < fmtsizeof(daqTbl[i].daqParam[n].fmt)) return dimension_error;
              dsize = 1; 
              fmt = daqTbl[i].daqParam[n].fmt;
              break;
            default: 
              break;
          }
        }
        if (dsize < 0) return dimension_error;
        limit = daqTbl[i].daqParam[n].size;
        if (isSimpleFormat(daqTbl[i].daqParam[n].fmt)) limit -= index;
        if (dsize > limit) 
        {
          dsize = dout->dArrayLength = limit;
        }
        dout->dTimeStamp = daqTbl[i].tstamp;
        dout->dStamp = daqTbl[i].usrstamp;
        dout->sysStamp = daqTbl[i].sysstamp;
        if (daqTbl[i].hasDataPool)
        {
          DataPoolEntry *dpe = (DataPoolEntry *)daqTbl[i].data;
          if (dout->dFormat != daqTbl[i].fmt) return illegal_format;
          if (daqTbl[i].fmt == CF_IMAGE)
          {
            dsize = sizeof(DIMAGE);
          }
          if (daqTbl[i].fmt == CF_GSPECTRUM)
          {
            dsize = sizeof(DGSPECTRUM);
          }
          memcpy(dout->data.bptr,dpe->data,dsize);
          dout->dArrayLength = dpe->length;
          return 0;
        }
        bptr = daqTbl[i].daqParam[n].offset; doff = 0;
        if (access & CA_WRITE)
        {
          cc = pushSelfEntry(din,i);
          if (cc != 0) 
          {
            feclog("Couldn't push self entry %.64s : %.32s",daqTbl[i].daqParam[n].param,cc2str(cc));
            return cc;
          }
          if (dsize == 0) return 0;
        }
        switch(daqTbl[i].daqParam[n].fmt)
        {
          case CF_BYTE:
            return PutValuesFromByteEx(dout,(BYTE *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].size,index);
          case CF_INT16:
            return PutValuesFromShortEx(dout,(short *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].size,index);
          case CF_INT32:
            return PutValuesFromLongEx(dout,(SINT32 *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].size,index);
          case CF_INT64:
            return PutValuesFromDLongEx(dout,(SINT64 *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].size,index);
          case CF_FLOAT:
            return PutValuesFromFloatEx(dout,(float *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].size,index);
          case CF_DOUBLE:
            return PutValuesFromDoubleEx(dout,(double *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].size,index);
          case CF_USTRING:
            /* doff = 16; <= PutValuesFromString takes ptr to type */
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64:
            bptr += doff;
            if (dout->dFormat != CF_STRING)
            {
              if (dout->dArrayLength > (UINT32)(daqTbl[i].daqParam[n].size - index)) 
                dout->dArrayLength = (UINT32)(daqTbl[i].daqParam[n].size - index);
            }
            return PutValuesFromString(dout,(char *)&daqTbl[i].data[bptr],daqTbl[i].daqParam[n].fmt,daqTbl[i].daqParam[n].size,index);
          case CF_TEXT:
            switch (fmt)
            {
              case CF_TEXT: break;
              case CF_NAME8:
              case CF_NAME16:
              case CF_NAME32:
              case CF_NAME48:
              case CF_NAME64: dsize = fmtsizeof(fmt); break;
              default: return illegal_format;
            }
            memset(dout->data.cptr,0,dsize);
            if (dsize > daqTbl[i].daqParam[n].size) dsize = daqTbl[i].daqParam[n].size;
            strncpy(dout->data.cptr,(char *)&daqTbl[i].data[bptr],dsize);
            return 0;
          case CF_FLTINT:
          case CF_FLTFLT:
          case CF_DBLDBL:
          case CF_INTINT:
            if (fmt != daqTbl[i].daqParam[n].fmt)
            { /* if not equal fall thru to default case */
              fmtsiz = fmtsizeof(daqTbl[i].daqParam[n].fmt)/2;
#             if defined(INTEL) && !defined(VXWORKS)
              mcab = (BYTE *)alloca(daqTbl[i].daqParam[n].size*fmtsiz);
#             else
              mcab = tmpWorkArea;
#             endif
              for (j=0; j<daqTbl[i].daqParam[n].size; j++) 
                memcpy(mcab+j*fmtsiz,&daqTbl[i].data[bptr],fmtsiz);
              if (daqTbl[i].daqParam[n].fmt == CF_INTINT)
                return PutValuesFromLongEx(dout,(SINT32 *)mcab,daqTbl[i].daqParam[n].size,index);
              if (daqTbl[i].daqParam[n].fmt == CF_FLTINT || daqTbl[i].daqParam[n].fmt == CF_FLTFLT)
                return PutValuesFromFloatEx(dout,(float *)mcab,daqTbl[i].daqParam[n].size,index);
              if (daqTbl[i].daqParam[n].fmt == CF_DBLDBL)
                return PutValuesFromDoubleEx(dout,(double *)mcab,daqTbl[i].daqParam[n].size,index);
            }
          default:
            if (fmt == daqTbl[i].daqParam[n].fmt)
            { /* non-simple format, but easy case to handle */
              int fsize = fmt == CF_STRUCT ? GetStructSize(daqTbl[i].tag) : fmtsizeof(fmt);
              if (dsize > daqTbl[i].daqParam[n].size) dsize = daqTbl[i].daqParam[n].size;
              dsize *= fsize;
              dsize += getFormatHeaderSize(fmt);
              memcpy(dout->data.bptr,&daqTbl[i].data[bptr],dsize);
              return 0;
            }
            return illegal_format;
        }
      }
    }
  }
  for (i=0; i<ngloballist; i++)
  {
    if (!stricmp(globalList[i].keyword,prop) && globalList[i].refdata != NULL)
    {
      return PutValuesFromAny(dout,globalList[i].refdata,globalList[i].format,1,fmtsizeof(globalList[i].format),globalList[i].size,0);
    }
  }
  return illegal_property;
}
/**
 * internal use only: don't bother checking null arguments 
 *
 * primitive arrays allowed: but last 16 bytes are kluged timestamp info ...
 */
int pushSelfEntry(DTYPE *din,int idx)
{
  DAQRecord *dst = &daqTbl[idx];
  double ts = 0, val = 0;
  int cc = 0, sysst = 0, usrst = 0, fsz = fmtsizeof(din->dFormat);
  char s[80];
  DTYPE d;
  struct tm st;
  time_t its;
  BYTE tsbuf[16];

  if (stricmp(dst->ctx,"SELF")) return not_allowed;
  if (dst->nparam != 1) return not_allowed;
  DTYPEZERO(d);
  d.dArrayLength = dst->len;
  switch (din->dFormat)
  {
    case CF_DBLDBL:
    case CF_DBLTIME:
      if (dst->len != 1) return not_allowed;
      val = ((DBLDBL *)din->data.vptr)->d1val;
      ts = ((DBLDBL *)din->data.vptr)->d2val;
      d.dFormat = CF_DOUBLE;
      d.data.dptr = &val;
      break;
    case CF_DBLDBLDBL:
      if (dst->len != 1) return not_allowed;
      val = ((DBLDBLDBL *)din->data.vptr)->d1val;
      ts = ((DBLDBLDBL *)din->data.vptr)->d2val;
      sysst = (int)((DBLDBLDBL *)din->data.vptr)->d3val;
      d.dFormat = CF_DOUBLE;
      d.data.dptr = &val;
      break;
    case CF_NAME64DBL:
    case CF_NAME64TIME:
      if (dst->len != 1) return not_allowed;
      strncpy(s,((NAME64DBL *)din->data.vptr)->name,64); s[64] = 0;
      ts = ((NAME64DBL *)din->data.vptr)->dval;
      d.dArrayLength = (UINT32)strlen(s);
      d.dFormat = CF_TEXT;
      d.data.cptr = s;
      break;
    case CF_NAME64DBLDBL:
      if (dst->len != 1) return not_allowed;
      strncpy(s,((NAME64DBLDBL *)din->data.vptr)->name,64); s[64] = 0;
      ts = ((NAME64DBLDBL *)din->data.vptr)->d1val;
      sysst = (int)((NAME64DBLDBL *)din->data.vptr)->d2val;
      d.dArrayLength = (UINT32)strlen(s);
      d.dFormat = CF_TEXT;
      d.data.cptr = s;
      break;
    case CF_INT16:
      if (din->dArrayLength * fsz <= 16) return dimension_error;
      din->dArrayLength -= 16/fsz;
      memcpy(tsbuf,&din->data.bptr[din->dArrayLength * fsz],16);
      SwapBytes(tsbuf,8,din->dFormat,LTOR); // unswap all bytes as given format
      memcpy(&ts,tsbuf,8);
      DSWAP(ts);
      memcpy(&sysst,&tsbuf[8],4);
      memcpy(&usrst,&tsbuf[12],4);
      ISWAP(sysst);
      ISWAP(usrst);
      memcpy(&d,din,sizeof(DTYPE));
      break;
    case CF_INT32:
    case CF_FLOAT:
      if (din->dArrayLength * fsz <= 16) return dimension_error;
      din->dArrayLength -= 16/fsz;
      memcpy(tsbuf,&din->data.bptr[din->dArrayLength * fsz],16);
      SwapBytes(tsbuf,2,din->dFormat,LTOR); // unswap 1st 8 bytes as given format
      memcpy(&ts,tsbuf,8);
      DSWAP(ts);
      sysst = tsbuf[2];
      usrst = tsbuf[3];
      memcpy(&d,din,sizeof(DTYPE));
      break;
    case CF_INT64:
    case CF_DOUBLE:
      if (din->dArrayLength * fsz <= 16) return dimension_error;
      din->dArrayLength -= 16/fsz;
      memcpy(tsbuf,&din->data.bptr[din->dArrayLength * fsz],16);
      memcpy(&ts,tsbuf,8);
      SwapBytes(&tsbuf[8],1,din->dFormat,LTOR); // unswap as last 8 bytes as given format
      memcpy(&sysst,&tsbuf[8],4);
      memcpy(&usrst,&tsbuf[12],4);
      ISWAP(sysst);
      ISWAP(usrst);
      memcpy(&d,din,sizeof(DTYPE));
      break;    
    default:
      return illegal_format;
  }  
  its = (time_t)ts;
  memcpy(&st,localtime(&its),sizeof(struct tm));
  getDataFilename(idx,st.tm_year,st.tm_mon+1);
  dst->lastarchtime = getLastArchivedTimeStamp(dst);

  if (ts < dst->lastarchtime) return data_stale;
  cc = GetValuesAsAny(&d,dst->data,(short)dst->fmt,fmtsizeof(dst->fmt),d.dArrayLength,0);
  dst->tstamp = ts;
  dst->sysstamp = sysst;
  dst->usrstamp = usrst; 
  dst->archtime = ts;
  memcpy(dst->archdata,dst->data,dst->reclen);

  WriteRecord(dst);
  dst->lastarchtime = ts;
  return 0;
}