/*
* network.h
*
* Provides routines to read the network utilization.
*
*  Created on: Dec 29, 2015
*      Author: Jaka Bobnar
*/
#ifndef NETWORK_H
#define NETWORK_H

#if defined (WIN32)
#include <winnt.h>
#define ERROR_BUFFER_TOO_SMALL ERROR_DS_USER_BUFFER_TO_SMALL
//if build as a static library, EXPORT only as extern "C"
#if defined(NETSTATIC)
#define EXPORT EXTERN_C
#else
//otherwise check if this header is used to build or use the library
//_NETWORK_EXP is only defined when the library is being build
#ifdef _NETWORK_EXP
#undef _NETWORK_EXP
#define _NETWORK_EXP __declspec( dllexport )
#define EXPORT EXTERN_C _NETWORK_EXP
#else 
#define _NETWORK_EXP __declspec( dllimport )
#define EXPORT extern _NETWORK_EXP
#endif
#endif
#elif defined (UNIX)
#define EXPORT
#define ERROR_BUFFER_TOO_SMALL           8309L
#define ERROR_NETWORK_NOT_AVAILABLE      5035L
#endif
/**
* \brief Returns the names of all network interfaces
* Fetches the names of all network interfaces and set them as individual character arrays
* of the names parameter. The names parameter has to be preallocated by the caller. If the
* names array length is too small the ERROR_BUFFER_TOO_SMALL is returned and the
* required buffer size is set on the size parameter. Individual character arrays of the 
* names array are allocated by this routine and have to be freed by the caller when no
* longer needed.
*
* \param names a preallocated array of charater arrays, to which the names of all
* network interfaces will be set
* \param size the size of the names array, to which the number of network interfaces
* is set if the size is too small
* \return 0 if successful or error code
*/
EXPORT int GetNetworkInterfaceNames(char** names, unsigned int *size);
/**
* \brief Returns the friendly names of all network interfaces
* Fetches the names of all network interfaces and set them as individual character arrays
* of the names parameter. The names parameter has to be preallocated by the caller. If the
* names array length is too small the ERROR_BUFFER_TOO_SMALL is returned and the
* required buffer size is set on the size parameter. Individual character arrays of the
* names array are allocated by this routine and have to be freed by the caller when no
* longer needed. The names returned by this routing are the human readable names, such
* as Local Area Connection etc.
*
* \param names a preallocated array of charater arrays, to which the names of all
* network interfaces will be set
* \param size the size of the names array, to which the number of network interfaces
* is set if the size is too small
* \return 0 if successful or error code
*/
EXPORT int GetNetworkInterfaceFriendlyNames(char** names, unsigned int *size);
/**
* \brief Calculates the network interface utilization in percentage and kB/s
* Calculates the average netwok utilization between two consecutive calls to this routine.
* The given utilization parameter length has to match the number of network interfaces and
* the size has to match that as well. The values are given in percentage of the total bandwidth
* currently utilized (percentage parameter) or in kilo bytes per second (absolute parameter).
* The order of elements in the float array is the same as the order of names returned by the
* GetNetworkInterfaceNames routine. If the array is too small ERROR_BUFFER_TOO_SMALL
* is returned.
*
* \param percentage array which will receive the network utilization in percentage for all
* network interfaces
* \param absolute array which will receive the network utilization in kB/s for all
* network interfaces
* \param size the size of the utilization array
* \return 0 if successful or error code
*/
EXPORT int GetNetworkUtilization(float *percentage, float *absolute, unsigned int size);
/**
* \brief Cleanup by freeing all allocated resources
* Free all resources allocated by the GetNetworkInterfaceNames and GetNeworkUtilizationPercentage.
* This routine has to be called when none of the above two methods will never be called again.
*
* \return 0 if successful or error code
*/
EXPORT int NetworkCleanUp(void);
#endif //NETWORK_H
