/*
* parser.c
*
* Implements routines from parser.h
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#include <math.h>
#include "tine.h"
#include "util.h"
#include "parser.h"
#include "process.h"

unsigned int watch_cycle_period = DEFAULT_WATCH_CYCLE_PERIOD; //1 second
boolean print_info_on_each_cycle = DEFAULT_PRINT_INFO_ON_EACH_CYCLE; //do not print
boolean start_server = DEFAULT_START_SERVER; //start the tine server
int suspended_timeout = DEFAULT_SUSPENDED_TIMEOUT; // 10 minutes
boolean allow_monitoring_other_hosts = DEFAULT_ALLOW_MONITORING_OTHER_HOSTS; //do not allow
int pid_refresh_rate = DEFAULT_PID_REFRESH_RATE; //60 seconds
int short_term_history_seconds = DEFAULT_SHORT_TERM_HISTORY;
int long_term_history_months = DEFAULT_LONG_TERM_HISTORY;
int kill_grace_period = DEFAULT_KILL_GRACE_PERIOD;
char **set_environment_variables = NULL;
unsigned int set_environment_variables_length = 0;
char *console_command = '\0';
char *minimum_free_disk_path = '\0';
float minimum_free_disk_mbyte = 0;

const UINT64 THOUSAND = 1024;

void freeTemplates(TEMPLATE *templates, unsigned int templates_count)
{
	unsigned int i,j;
    for (i = 0; i < templates_count; i++)
    {
        free(templates[i].name);
        for (j = 0; j < templates[i].numberOfPairs; j++)
        {
            free(templates[i].keys[j]);
            free(templates[i].values[j]);
        }
        templates[i].numberOfPairs = 0;
    }
}

void freeMacros(MACRO *macros, unsigned int macros_count)
{
	unsigned int i;
    for (i = 0; i < macros_count; i++)
    {
        free(macros[i].key);
        free(macros[i].value);
    }
}

char* findReplaceString(char* text, char* oldWord, char* newWord)
{
    size_t oldLength = strlen(text);
    size_t oldWordLength = strlen(oldWord);
    size_t newWordLength = strlen(newWord);
    size_t newLength = (oldLength + newWordLength + 1); // +1 for \0
    newLength = newLength - oldWordLength; // to avoid going negative
    char* newText = (char*)malloc(newLength * sizeof(char));
    if (newText == NULL) OutOfMemoryExit(3);

    size_t index = strstr(text, oldWord) - text;
    strncpy_s(newText, newLength, text, index);
    strncpy_s(&newText[index], newLength - index, newWord, newWordLength);
    strncpy_s(&newText[index + newWordLength],
        newLength - index - newWordLength,
        &text[index + oldWordLength],
        oldLength - index - oldWordLength);

    return newText;
}

void splitToKeyValue(char* line, char** key, char** value)
{
    char* splitKey = strchr(line, ':');
    if (splitKey == NULL)
    {
        splitKey = strchr(line, '=');
    }
    size_t keySize = splitKey - line + 1;
    size_t startOfValue = trim(&line[keySize]) + keySize; // trim spaces
    size_t valueSize = strlen(line) - startOfValue + 1; // add \0
    *key = malloc(sizeof(char) * keySize);
    *value = malloc(sizeof(char) * valueSize);
    strncpy_s(*key, keySize, line, keySize - 1);
    strncpy_s(*value, valueSize, &line[startOfValue], valueSize - 1);
}

void FindMacros(char** lines, unsigned int num_lines, MACRO macros[], unsigned int* macros_count, boolean parse_all_macros)
{
    *macros_count = 0;
    unsigned int i;
    for (i = 0; i < num_lines; i++)
    {
        if (!parse_all_macros && lines[i][0] == '[') // process or template parsing started, stop macro parsing
            break;

        if (lines[i][0] == '$')
        {
            char* key;
            splitToKeyValue(lines[i], &key, &macros[*macros_count].value);

            // decorate macro
            size_t keySize = strlen(key) + 3; // to add ( ) and \0
            macros[*macros_count].key = malloc(sizeof(char) * keySize);
            strcpy_s(macros[*macros_count].key, keySize, "$(");
            strncpy_s(&macros[*macros_count].key[2], keySize - 2, &key[1], keySize - 3);
            strcat_s(macros[*macros_count].key, keySize, ")");
            free(key);

            (*macros_count)++;
            if (*macros_count >= MAX_MACRO_COUNT)
            {
                printf("Maximum number of macros reached.\n");
                break;
            }
        }
    }
}

void EvaluateMacros(char** lines, unsigned int num_lines, MACRO macros[], unsigned int macros_count)
{
	unsigned int i, j;
    // skip to process definitions
    for (i = 0; i < num_lines; i++)
        if (lines[i][0] == '[')
            break;

    char* startAt = NULL;
    for (; i < num_lines; )
    {
        char* start = startAt == NULL ? strstr(lines[i], "$(") : strstr(startAt,"$(");
        if (start == NULL)
        {
            startAt = NULL;
            i++;
            continue;
        }
        char* end = strstr(start, ")");
        if (end == NULL)
        {
            startAt = NULL;
            i++;
            continue;
        }
        startAt = end;

        // find macro and evaluate it on line
        for (j = 0; j < macros_count; j++)
        {
            if (strstr(start, macros[j].key) == NULL)
                continue;

            char* newLine = findReplaceString(lines[i], macros[j].key, macros[j].value);
            free(lines[i]);
            lines[i] = newLine;
            //break after first occurence and repeat the process with the next
            break;
        }
    }
}

void FindTemplates(char** lines, unsigned int num_lines, TEMPLATE templates[], unsigned int* templates_count)
{
    *templates_count = 0;
    unsigned int i;
    for (i = 0; i < num_lines; i++)
    {
        if (lines[i][0] == '$') // macro
        {
            continue;
        }
        if (compareCaseInsensitive(lines[i], "[template_") != NULL) // template
        {
            size_t nameLength = strchr(lines[i], ']') - lines[i] - 10 + 1; // skip "[template_" and add \0
            templates[*templates_count].name = malloc(sizeof(char)*nameLength);
            if (templates[*templates_count].name == NULL) OutOfMemoryExit(1);
            strncpy_s(templates[*templates_count].name, nameLength, &lines[i][10], nameLength - 1); // skip "[template_"
                                                                                                    // move to next line and start parsing template content
            i++;
            templates[*templates_count].numberOfPairs = 0;
            for (; i < num_lines && lines[i][trim(lines[i])] != '\0'; i++) // until new template or confirugation started
            {
                if (lines[i][0] == '#') // comment
                    continue;

                int numberOfPairs = templates[*templates_count].numberOfPairs;
                if (numberOfPairs >= MAX_PAIR_COUNT)
                {
                    printf("Maximum number of pairs reached.\n");
                    break;
                }

                splitToKeyValue(lines[i], &(templates[*templates_count].keys[numberOfPairs]),
                    &(templates[*templates_count].values[numberOfPairs]));
                templates[*templates_count].numberOfPairs++;
            }

            (*templates_count)++;

            if (*templates_count >= MAX_TEMPLATE_COUNT)
            {
                printf("Maximum number of templates reached.\n");
                break;
            }
        }
    }
}

void EvaluateTemplates(char*** lines_pointer, unsigned int* num_lines, TEMPLATE templates[], unsigned int templates_count)
{
	unsigned int i,j;
    for (i = 0; i < *num_lines; i++)
    {
        char** lines = *lines_pointer;
        if (lines[i][0] == '[' && strchr(lines[i], ']') != NULL && strchr(lines[i], '{') != NULL)
        {
            unsigned int startOfTemplate = i;
            int endOfTemplate = -1;

            // get template name and find if it exists
            size_t idStart = strstr(lines[i], "{") - lines[i] + 1;
            size_t idEnd = strstr(lines[i], "}") - lines[i];
            size_t idLength = idEnd - idStart + 1;
            char* id = malloc(sizeof(char) * idLength);
            if (id == NULL) OutOfMemoryExit(1);
            strncpy_s(id, idLength, &lines[i][idStart], idLength - 1);
            TEMPLATE* theTemplate = NULL;
            for (j = 0; j < templates_count; j++)
            {
                if (strcmp(templates[j].name, id) == 0)
                {
                    theTemplate = &templates[j];
                    break;
                }
            }
            if (theTemplate == NULL)
            {
                printf("Template %s not found.\n", id);
                free(id);
                continue;
            }
            free(id);

            // get config name
            size_t configNameLength = strchr(lines[i], ']') - lines[i] + 2;
            char** rows = malloc(sizeof(char*) * (MAX_PAIR_COUNT + 1));
            if (rows == NULL) OutOfMemoryExit(1);
            rows[0] = malloc(sizeof(char) * (configNameLength));
            if (rows[0] == NULL) OutOfMemoryExit(1);
            strncpy_s(rows[0], configNameLength, lines[i], configNameLength - 1);
            i++;

            // copy template to rows
            for (j = 0; j < theTemplate->numberOfPairs; j++)
            {
                size_t length = (strlen(theTemplate->keys[j]) + strlen(theTemplate->values[j]) + 2);
                rows[j + 1] = malloc(sizeof(char) * length);
                if (rows[j + 1] == NULL) OutOfMemoryExit(1);
                strcpy_s(rows[j + 1], length, theTemplate->keys[j]);
                strcat_s(rows[j + 1], length, ":");
                strcat_s(rows[j + 1], length, theTemplate->values[j]);
            }

            unsigned int rowsCount = theTemplate->numberOfPairs + 1;

            // find if key value already exist in configuration and replace template value
            for (; (i < *num_lines) && lines[i][trim(lines[i])] != '\0'; i++) // until new template or confirugation started
            {
                if (lines[i][0] == '#') // comment
                    continue;

                char* key;
                char* value;
                splitToKeyValue(lines[i], &key, &value);
                boolean inTemplate = FALSE;
                for (j = 0; j < theTemplate->numberOfPairs; j++)
                {
                    if (strcmp(theTemplate->keys[j], key) == 0)
                    {
                        free(rows[j + 1]);
                        size_t length = strlen(lines[i]) + 1;
                        rows[j + 1] = malloc(sizeof(char) * length);
                        if (rows[j + 1] == NULL) OutOfMemoryExit(1);
                        strcpy_s(rows[j + 1], length, lines[i]);
                        inTemplate = TRUE;
                        break;
                    }
                }
                free(key);
                free(value);
                if (!inTemplate)
                {
                    if (rowsCount >= MAX_PAIR_COUNT)
                    {
                        printf("Maximum number of pairs reached.\n");
                        continue;
                    }
                    size_t length = strlen(lines[i]) + 1;
                    rows[rowsCount] = malloc(sizeof(char) * length);
                    if (rows[rowsCount] == NULL) OutOfMemoryExit(1);
                    strcpy_s(rows[rowsCount], length, lines[i]);
                    rowsCount++;
                }
            }

            endOfTemplate = i;

            // evaluate local macros
            MACRO localMacros[MAX_MACRO_COUNT];
            unsigned int localMacrosCount = 0;
            FindMacros(rows, rowsCount, localMacros, &localMacrosCount, TRUE);
            EvaluateMacros(rows, rowsCount, localMacros, localMacrosCount);
            freeMacros(localMacros, localMacrosCount);

            int finalLength = (*num_lines) - (endOfTemplate - startOfTemplate) + rowsCount;  // total lines - lines for template config 
                                                                                             // + new lines added by config
            char** newArray = malloc(sizeof(char*) * finalLength);
            if (newArray == NULL) OutOfMemoryExit(1);
            size_t length = 0;
            for (j = 0; j < startOfTemplate; j++) // before template
            {
                length = strlen(lines[j]) + 1;
                newArray[j] = malloc(sizeof(char) * length);
                if (newArray[j] == NULL) OutOfMemoryExit(1);
                strcpy_s(newArray[j], length, lines[j]);
            }
            for (j = 0; j < rowsCount; j++) // generated template
            {
                length = strlen(rows[j]) + 1;
                newArray[startOfTemplate + j] = malloc(sizeof(char) * length);
                if (newArray[startOfTemplate + j] == NULL) OutOfMemoryExit(1);
                strcpy_s(newArray[startOfTemplate + j], length, rows[j]);
            }
            for (j = 0; j < (*num_lines) - endOfTemplate; j++) // after template
            {
                length = strlen(lines[endOfTemplate + j]) + 1;
                newArray[startOfTemplate + rowsCount + j] = malloc(sizeof(char) * length);
                if (newArray[startOfTemplate + rowsCount + j] == NULL) OutOfMemoryExit(1);
                strcpy_s(newArray[startOfTemplate + rowsCount + j], length, lines[endOfTemplate + j]);
            }

            freeArray(*lines_pointer, *num_lines);
            *lines_pointer = newArray;
            *num_lines = finalLength;
            i = startOfTemplate + rowsCount;
            freeArray(rows, rowsCount);
        }
    }
}

void InitializeDefaults(WATCHED_PROCESS *process)
{
    //static values
    process->name = '\0';
    process->description = '\0';
    process->path_to_executable = '\0';
    process->auto_start = YES;
    process->max_cpu_load = 100;
    process->number_of_allowed_cpu_load_violations = 1;
    process->max_memory = THOUSAND * THOUSAND * THOUSAND * THOUSAND; //1TB
    process->number_of_allowed_memory_violations = 1;
    process->start_after_n_seconds = 10;
    process->number_of_retries = 3;
    process->retry_interval = 20;
    process->preceding_process = NULL;
    process->environment_variables = NULL;
    process->number_of_preceding_processes = 0;
    process->number_of_environment_variables = 0;
    process->working_directory = '\0';
    process->parameters = '\0';
    process->tine_device = '\0';
    process->tine_property = '\0';
    process->context = '\0';
    process->server = '\0';
    process->polling_rate = 1000;
    process->timeout = 10;
    process->timeout_restart = 0;
    process->number_of_parameters = 0;
    process->match_parameters = TRUE;
    process->match_path = TRUE;
    process->caption = '\0';
    process->pause_after_previous_start = 1;
    process->is_service = FALSE;
    process->show_console_window = TRUE;

    //dynamic values
    process->number_of_memory_violations = 0;
    process->number_of_cpu_violations = 0;
    process->determined_not_running_timestamp = 0;
    process->process_requested_to_stop = 0;
    process->last_retry_timestamp = 0;
    process->last_tine_retry = 0;
    process->disabled = FALSE;
    process->stopped_by_user = FALSE;
    process->is_running = FALSE;
    process->is_starting = TRUE;
    process->current_retry = 0;
    process->current_cpu_load = 0;
    process->current_memory_consumption = 0;
    process->pid = -1;
    process->link_id = -1;
    process->property_link_timeout = FALSE;
    process->last_tine_response = 0;
    process->property_disabled = FALSE;
    process->tine_status = -2;
    process->preconditions_started = FALSE;
    process->pid_last_refreshed = 0;
    process->number_of_times_started_by_wd = 0;
    process->number_of_times_stopped_by_wd = 0;
    process->stopped_because_of_cpu = FALSE;
    process->stopped_because_of_memory = FALSE;
    process->stopped_because_of_link = FALSE;
    process->memory_at_stop_time = 0;
    process->cpu_at_stop_time = 0;
    process->num_of_cpu_violations_at_stop_time = 0;
    process->num_of_memory_violations_at_stop_time = 0;
    process->has_alarm = FALSE;
    process->default_property_info_used = FALSE;
    process->dout.data.vptr = NULL;
    process->dout_allocated_by_me = FALSE;
    process->process_restart_requested = 0;
    process->last_restart_reason = NONE;
}   

void PrintProcessInfo(WATCHED_PROCESS *process, unsigned int size)
{
    unsigned int i, j;
    printf("Number of monitored processes: %d\n", size);
    printf("--------------------------------------------------------\n");
    for (i = 0; i < size; i++)
    {
        printf("Process: %s\n", process[i].id);
        printf("\tAlias: %s\n", process[i].alias);
        printf("\tName: %s\n", process[i].name);
        if (process[i].description != NULL)
        {
            printf("\tDescription: %s\n", process[i].description);
        }
        printf("\tService: %s\n", process[i].is_service ? "Yes" : "No");
        printf("\tShow console window: %s\n", process[i].show_console_window ? "Yes" : "No");
        printf("\tPath: %s\n", process[i].path_to_executable);
        printf("\tCommand line parameters: %s\n", process[i].parameters);
        printf("\tMatch path: %s\n", process[i].match_path ? "Yes" : "No");
        printf("\tMatch parameters: %s\n", process[i].match_parameters ? "Yes" : "No");
        printf("\tWorking Directory: %s\n", process[i].working_directory);
        if (process[i].caption == NULL)
        {
            printf("\tProcess caption: None\n");
        }
        else
        {
            printf("\tProcess caption: %s\n", process[i].caption);
        }
        printf("\tPreceding processes (%d): ", process[i].number_of_preceding_processes);
        if (process[i].number_of_preceding_processes > 0)
        {
            for (j = 0; j < process[i].number_of_preceding_processes; j++)
            {
                printf("%s; ", process[i].preceding_process[j]);
            }
            printf("\n");
        }
        else
        {
            printf("None\n");
        }
        printf("\tEnvironment variables (%d): ", process[i].number_of_environment_variables);
        if (process[i].number_of_environment_variables > 0)
        {
            for (j = 0; j < process[i].number_of_environment_variables; j++)
            {
                printf("%s; ", process[i].environment_variables[j]);
            }
            printf("\n");
        }
        else
        {
            printf("None\n");
        }

        printf("\tAuto start: %s\n", process[i].auto_start ?  (process[i].auto_start == YES ? "Yes" : "Once") : "No");
        if (process[i].auto_start == YES)
        {
            printf("\tMonitor after: %d seconds\n", process[i].start_after_n_seconds);
            printf("\tPause after previous process: %d seconds\n", process[i].pause_after_previous_start);
            printf("\tNumber of retries: %d\n", process[i].number_of_retries);
        }
        printf("\tRetry interval: %d seconds\n", process[i].retry_interval);
        printf("\tMax CPU load: %d %%\n", process[i].max_cpu_load);
        printf("\tNumber of allowed CPU violations: %d\n", process[i].number_of_allowed_cpu_load_violations);
        printf("\tMax memory consumption: %3.0f MB\n", process[i].max_memory / (1024.0 * 1024.0));
        printf("\tNumber of allowed memory limit violations: %d\n", process[i].number_of_allowed_memory_violations);

        if (process[i].tine_device == NULL || process[i].tine_property == NULL)
        {
            printf("\tTINE Property: N/A\n");
        }
        else
        {
            printf("\tTINE Property: %s/%s\n", process[i].tine_device, process[i].tine_property);
            printf("\tPolling rate: %d milliseconds\n", process[i].polling_rate);
            printf("\tTimeout: %d seconds\n", process[i].timeout);
            if (process[i].timeout_restart > 0)
            {
                printf("\tRestart timeout: %d seconds\n", process[i].timeout_restart);
            }
            else
            {
                printf("\tRestart timeout: never restart\n");
            }

        }
        printf("--------------------------------------------------------\n");
    }
}

void PrintSettings(WATCHED_PROCESS *processes, unsigned int size, SERVER_INFO *info)
{
    printf("Watch cycle period: %u ms\n", watch_cycle_period);
    printf("Print process info: %s\n", print_info_on_each_cycle ? "Yes" : "No");
    printf("Start server: %s\n", start_server ? "Yes" : "No");
    printf("Allow monitoring other hosts: %s\n", allow_monitoring_other_hosts ? "Yes" : "No");
    printf("Suspended timeout: %d\n", suspended_timeout);
    printf("PID refresh rate: %d\n", pid_refresh_rate);
    if (minimum_free_disk_path) {
        printf("Minimum Free Disk Path: %s\n", minimum_free_disk_path);
        printf("Minimum Free Disk Size: %3.2f\n", minimum_free_disk_mbyte);
    }
    if (console_command) printf("Console command: %s\n", console_command);
    printf("FEC: %s\n", info->fec_name);
    printf("Context: %s\n", info->context);
    printf("Server: %s\n", info->server_name);
    printf("Subsystem: %s\n", info->subsystem);
    printf("Port Offset: %d\n", info->port_offset);
    printf("Description: %s\n", info->description);
    printf("Hardware: %s\n", info->hardware);
    printf("Location: %s\n", info->location);
    printf("Responsible: %s\n\n", info->responsible);
    PrintProcessInfo(processes, size);
}

/**
* \brief Compares two processes and returns -1, 0, or 1, if the first process is less, equal, or greater than the second process.
*
* Compares the processes. Process v1 is less than process v2 if the process v2 depends on the process v1.
* Process that does not depend on anything has the highest priority and is always less than a process
* that depends on something. Two processes are never equal.
*
* \param v1 the first process to compare
* \param v2 the second process to compare
* \return -1 if the first process has priority over the second or 1 vice versa
*/
int processComparator(const void *v1, const void *v2)
{
    unsigned int i;
    const WATCHED_PROCESS *p1 = (WATCHED_PROCESS *)v1;
    const WATCHED_PROCESS *p2 = (WATCHED_PROCESS *)v2;
    if ((p1->number_of_preceding_processes) == 0)
    {
        //if first process does not have a preprocess defined it has priority, regardless of the second process
        return -1;
    }
    else if ((p2->number_of_preceding_processes) == 0)
    {
        //if the first one has a preprocess and the second one doesn't, the second one has priority
        return 1;
    }
    else {
        //if process 1 requires process 2 to start before, the second one has priority
        for (i = 0; i < (p1->number_of_preceding_processes); i++)
        {
            if (strcmp(p1->preceding_process[i], p2->alias) == 0 || strcmp(p1->preceding_process[i], p2->name) == 0)
            {
                return 1;
            }
        }
        //else return 1.
        return -1;
    }
}

WATCHED_PROCESS *ParseFile(char *file_name, unsigned int *size, SERVER_INFO *info)
{
    char **words;
    char **temp;
    int c;
    unsigned int num_lines, j, k, start, n, length;
    boolean errorsInFile;
    WATCHED_PROCESS *processes;
    MACRO *macros;
    TEMPLATE *templates;
    UINT64 th = 1024;

    words = readFile(file_name, &num_lines);
    //count number of defined processes
    c = 0;
    for (j = 0; j < num_lines; j++)
    {
        if (words[j][0] == '[' && (compareCaseInsensitive(words[j], "[template_") == NULL)) c++;
    }
    *size = c;

    //parse global macros
    macros = (MACRO*)malloc(sizeof(MACRO) * MAX_MACRO_COUNT);
    if (macros == NULL) OutOfMemoryExit(1);
    unsigned int macros_count = 0;
    FindMacros(words, num_lines, macros, &macros_count, FALSE);
    
    //parse templates
    templates = (TEMPLATE*)malloc(sizeof(TEMPLATE) * MAX_TEMPLATE_COUNT);
    if (templates == NULL) OutOfMemoryExit(1);
    unsigned int templates_count = 0;
    FindTemplates(words, num_lines, templates, &templates_count);
    
    // evaluate templates and its local macros
    EvaluateTemplates(&words, &num_lines, templates, templates_count);
    freeTemplates(templates, templates_count);

    // evaluate global macros
    EvaluateMacros(words, num_lines, macros, macros_count);
    freeMacros(macros, macros_count);

    processes = (WATCHED_PROCESS *)malloc(sizeof(WATCHED_PROCESS)* c);
    if (processes == NULL) OutOfMemoryExit(3);
    c = -1;
    
    for (j = 0; j < num_lines; j++)
    {
        if (words[j][0] == '#') continue; //comment

        length = (int)strlen(words[j]);
        if (words[j][0] == '[' && (compareCaseInsensitive(words[j], "[template_") == NULL))
        {
            c++;
            //the process definition starts here, alias is given in brackets, which should be removed
            length = (int) (strchr(words[j], ']') - words[j] - 1);
            start = 1;
            processes[c].id = (char*)malloc(sizeof(char)* (length + 1));
            if (processes[c].id)
            {
                memcpy(processes[c].id, &words[j][start], length);
                processes[c].id[length] = '\0';
            }
            processes[c].alias = (char*)malloc(sizeof(char)* (length + 1));
            if (processes[c].alias)
            {
                memcpy(processes[c].alias, &words[j][start], length);
                processes[c].alias[length] = '\0';
            }
            InitializeDefaults(&processes[c]);
            FindMacros(&words[j+1], num_lines - j - 1, macros, &macros_count, FALSE);
            EvaluateMacros(&words[j], num_lines - j, macros, macros_count);
            freeMacros(macros, macros_count);
        }
        else if (c >= 0 && strncmp(words[j], WD_NAME, strlen(WD_NAME)) == 0)
        {
            start = (int)strlen(WD_NAME) + 1;
            processes[c].name = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_ALIAS, strlen(WD_ALIAS)) == 0)
        {
            //alias is defined. First get rid of the old one, then overwrite
            free(processes[c].alias);
            start = (int)strlen(WD_ALIAS) + 1;
            processes[c].alias = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_SERVICE, strlen(WD_SERVICE)) == 0)
        {
            start = (int)strlen(WD_SERVICE) + 1;
            processes[c].is_service = toBoolean(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_SHOW_CONSOLE_WINDOW, strlen(WD_SHOW_CONSOLE_WINDOW)) == 0)
        {
            start = (int)strlen(WD_SHOW_CONSOLE_WINDOW) + 1;
            processes[c].show_console_window = toBoolean(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_DESCRIPTION, strlen(WD_DESCRIPTION)) == 0)
        {
            start = (int)strlen(WD_DESCRIPTION) + 1;
            processes[c].description = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_PATH_TO_EXECUTABLE, strlen(WD_PATH_TO_EXECUTABLE)) == 0)
        {
            start = (int)strlen(WD_PATH_TO_EXECUTABLE) + 1;
            processes[c].path_to_executable = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_PARAMETERS, strlen(WD_PARAMETERS)) == 0)
        {
            start = (int)strlen(WD_PARAMETERS) + 1;
            processes[c].parameters = toChar(&words[j][start], 1);
            processes[c].individual_parameters = split(&words[j][start], &n, ' ', TRUE, TRUE, FALSE);
            processes[c].number_of_parameters = n;
        }
        else if (c >= 0 && strncmp(words[j], WD_MATCH_PARAMETERS, strlen(WD_MATCH_PARAMETERS)) == 0)
        {
            start = (int)strlen(WD_MATCH_PARAMETERS) + 1;
            processes[c].match_parameters = toBoolean(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_MATCH_PATH, strlen(WD_MATCH_PATH)) == 0)
        {
            start = (int)strlen(WD_MATCH_PATH) + 1;
            processes[c].match_path = toBoolean(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_WORKING_DIRECTORY, strlen(WD_WORKING_DIRECTORY)) == 0)
        {
            start = (int)strlen(WD_WORKING_DIRECTORY) + 1;
            processes[c].working_directory = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_PRECEDING_PROCESSES, strlen(WD_PRECEDING_PROCESSES)) == 0)
        {
            start = (int)strlen(WD_PRECEDING_PROCESSES) + 1;
            processes[c].preceding_process = split(&words[j][start], &n, ';', TRUE, FALSE, TRUE);
            processes[c].number_of_preceding_processes = n;
        }
        else if (c >= 0 && strncmp(words[j], WD_ENVIRONMENT_VARIABLES, strlen(WD_ENVIRONMENT_VARIABLES)) == 0)
        {
            start = (int)strlen(WD_ENVIRONMENT_VARIABLES) + 1;
            processes[c].environment_variables = split(&words[j][start], &n, ';', TRUE, TRUE, TRUE);
            processes[c].number_of_environment_variables = n;
        }
        else if (c >= 0 && strncmp(words[j], WD_PROCESS_CAPTION, strlen(WD_PROCESS_CAPTION)) == 0)
        {
            start = (int)strlen(WD_PROCESS_CAPTION) + 1;
            processes[c].caption = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_MONITOR_N_SECONDS_AFTER_START, strlen(WD_MONITOR_N_SECONDS_AFTER_START)) == 0)
        {
            start = (int)strlen(WD_MONITOR_N_SECONDS_AFTER_START) + 1;
            processes[c].start_after_n_seconds = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS, strlen(WD_PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS)) == 0)
        {
            start = (int)strlen(WD_PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS) + 1;
            processes[c].pause_after_previous_start = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_NUMBER_OF_RETRIES, strlen(WD_NUMBER_OF_RETRIES)) == 0)
        {
            start = (int)strlen(WD_NUMBER_OF_RETRIES) + 1;
            processes[c].number_of_retries = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_RETRY_INTERVAL, strlen(WD_RETRY_INTERVAL)) == 0)
        {
            start = (int)strlen(WD_RETRY_INTERVAL) + 1;
            processes[c].retry_interval = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_MAX_CPU_LOAD, strlen(WD_MAX_CPU_LOAD)) == 0)
        {
            start = (int)strlen(WD_MAX_CPU_LOAD) + 1;
            processes[c].max_cpu_load = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS, strlen(WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS)) == 0)
        {
            start = (int)strlen(WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS) + 1;
            processes[c].number_of_allowed_cpu_load_violations = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_AUTOSTART, strlen(WD_AUTOSTART)) == 0)
        {
            start = (int)strlen(WD_AUTOSTART) + 1;
            processes[c].auto_start = toAutoStart(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_MAX_MEMORY, strlen(WD_MAX_MEMORY)) == 0)
        {
            start = (int)strlen(WD_MAX_MEMORY) + 1;
            processes[c].max_memory = th * th * toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS, strlen(WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS)) == 0)
        {
            start = (int)strlen(WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS) + 1;
            processes[c].number_of_allowed_memory_violations = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_TINE_PROPERTY, strlen(WD_TINE_PROPERTY)) == 0)
        {
            start = (int)strlen(WD_TINE_PROPERTY) + 1;
            processes[c].tine_property = toChar(&words[j][start], TRUE);
        }
        else if (c >= 0 && strncmp(words[j], WD_TINE_DEVICE, strlen(WD_TINE_DEVICE)) == 0)
        {
            start = (int)strlen(WD_TINE_DEVICE) + 1;
            processes[c].tine_device = toChar(&words[j][start], TRUE);
            //extract the context and server name from the whole device name
            temp = split(&words[j][start], &n, '/', FALSE, FALSE, TRUE);
            if (n >= 3)
            {
                //if the names starts with a / we get an empty string first
                if (strlen(temp[0]) == 0)
                {
                    processes[c].context = temp[1];
                    processes[c].server = temp[2];
                    k = 3;
                    free(temp[0]);
                }
                else
                {
                    processes[c].context = temp[0];
                    processes[c].server = temp[1];
                    k = 2;
                }
            }
            //free everything that is no longer needed
            for (; k < n; k++)
            {
                free(temp[k]);
            }
            free(temp);
        }
        else if (c >= 0 && strncmp(words[j], WD_POLLING_RATE, strlen(WD_POLLING_RATE)) == 0)
        {
            start = (int)strlen(WD_POLLING_RATE) + 1;
            processes[c].polling_rate = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_TIMEOUT, strlen(WD_TIMEOUT)) == 0)
        {
            start = (int)strlen(WD_TIMEOUT) + 1;
            processes[c].timeout = toNumber(&words[j][start]);
        }
        else if (c >= 0 && strncmp(words[j], WD_RESTART_TIMEOUT, strlen(WD_RESTART_TIMEOUT)) == 0)
        {
            start = (int)strlen(WD_RESTART_TIMEOUT) + 1;
            processes[c].timeout_restart = toNumber(&words[j][start]);
        }
        //watchdog settings
        else if (strncmp(words[j], WD_G_ENVIRONMENT_VARIABLES, strlen(WD_G_ENVIRONMENT_VARIABLES)) == 0)
        {
            start = (int)strlen(WD_G_ENVIRONMENT_VARIABLES) + 1;
            set_environment_variables = split(&words[j][start], &n, ';', TRUE, TRUE, TRUE);
            set_environment_variables_length = n;
        }
        else if (strncmp(words[j], WD_G_PRINT_INFO_ON_EACH_CYCLE, strlen(WD_G_PRINT_INFO_ON_EACH_CYCLE)) == 0)
        {
            start = (int)strlen(WD_G_PRINT_INFO_ON_EACH_CYCLE) + 1;
            print_info_on_each_cycle = toBoolean(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_CPU_MEASURING_PERIOD, strlen(WD_G_CPU_MEASURING_PERIOD)) == 0)
        {
            printf("%s is deprecated. %s is used.\n", WD_G_CPU_MEASURING_PERIOD, WD_G_WATCH_CYCLE_PERIOD);
        }
        else if (strncmp(words[j], WD_G_WATCH_CYCLE_PERIOD, strlen(WD_G_WATCH_CYCLE_PERIOD)) == 0)
        {
            start = (int)strlen(WD_G_WATCH_CYCLE_PERIOD) + 1;
            watch_cycle_period = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_SUSPENDED_TIMEOUT, strlen(WD_G_SUSPENDED_TIMEOUT)) == 0)
        {
            start = (int)strlen(WD_G_SUSPENDED_TIMEOUT) + 1;
            suspended_timeout = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_PID_REFRESH_RATE, strlen(WD_G_PID_REFRESH_RATE)) == 0)
        {
            start = (int)strlen(WD_G_PID_REFRESH_RATE) + 1;
            pid_refresh_rate = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_START_SERVER, strlen(WD_G_START_SERVER)) == 0)
        {
            start = (int)strlen(WD_G_START_SERVER) + 1;
            start_server = toBoolean(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_ALLOW_MONITORING_OTHER_HOSTS, strlen(WD_G_ALLOW_MONITORING_OTHER_HOSTS)) == 0)
        {
            start = (int)strlen(WD_G_ALLOW_MONITORING_OTHER_HOSTS) + 1;
            allow_monitoring_other_hosts = toBoolean(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_CONSOLE_COMMAND, strlen(WD_G_CONSOLE_COMMAND)) == 0)
        {
            start = (int)strlen(WD_G_CONSOLE_COMMAND) + 1;
            console_command = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], WD_G_MINIMUM_FREE_DISK_PATH, strlen(WD_G_MINIMUM_FREE_DISK_PATH)) == 0)
        {
            start = (int)strlen(WD_G_MINIMUM_FREE_DISK_PATH) + 1;
            minimum_free_disk_path = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], WD_G_MINIMUM_FREE_DISK_MBYTE, strlen(WD_G_MINIMUM_FREE_DISK_MBYTE)) == 0)
        {
            start = (int)strlen(WD_G_MINIMUM_FREE_DISK_MBYTE) + 1;
            minimum_free_disk_mbyte = toFloat(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_STOP_ALL_ON_SHUTDOWN, strlen(WD_G_STOP_ALL_ON_SHUTDOWN)) == 0)
        {
            start = (int)strlen(WD_G_STOP_ALL_ON_SHUTDOWN) + 1;
            stop_all_on_shutdown = toBoolean(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_SHORT_TERM_HISTORY, strlen(WD_G_SHORT_TERM_HISTORY)) == 0)
        {
            start = (int)strlen(WD_G_SHORT_TERM_HISTORY) + 1;
            short_term_history_seconds = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_LONG_TERM_HISTORY, strlen(WD_G_LONG_TERM_HISTORY)) == 0)
        {
            start = (int)strlen(WD_G_LONG_TERM_HISTORY) + 1;
            long_term_history_months = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_KILL_GRACE_PERIOD, strlen(WD_G_KILL_GRACE_PERIOD)) == 0)
        {
            start = (int)strlen(WD_G_KILL_GRACE_PERIOD) + 1;
            kill_grace_period = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], WD_G_NETWORK_ENABLED, strlen(WD_G_NETWORK_ENABLED)) == 0)
        {
            start = (int)strlen(WD_G_NETWORK_ENABLED) + 1;
            g_network_enabled = toBoolean(&words[j][start]);
        }
        //fec info settings
        else if (strncmp(words[j], SI_CONTEXT, strlen(SI_CONTEXT)) == 0)
        {
            start = (int)strlen(SI_CONTEXT) + 1;
            info->context = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_SUBSYSTEM, strlen(SI_SUBSYSTEM)) == 0)
        {
            start = (int)strlen(SI_SUBSYSTEM) + 1;
            info->subsystem = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_FEC_NAME, strlen(SI_FEC_NAME)) == 0)
        {
            start = (int)strlen(SI_FEC_NAME) + 1;
            info->fec_name = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_DESCRIPTION, strlen(SI_DESCRIPTION)) == 0)
        {
            start = (int)strlen(SI_DESCRIPTION) + 1;
            info->description = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_HARDWARE, strlen(SI_HARDWARE)) == 0)
        {
            start = (int)strlen(SI_HARDWARE) + 1;
            info->hardware = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_LOCATION, strlen(SI_LOCATION)) == 0)
        {
            start = (int)strlen(SI_LOCATION) + 1;
            info->location = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_RESPONSIBLE, strlen(SI_RESPONSIBLE)) == 0)
        {
            start = (int)strlen(SI_RESPONSIBLE) + 1;
            info->responsible = toChar(&words[j][start], TRUE);
        }
        else if (strncmp(words[j], SI_PORT_OFFSET, strlen(SI_PORT_OFFSET)) == 0)
        {
            start = (int)strlen(SI_PORT_OFFSET) + 1;
            info->port_offset = toNumber(&words[j][start]);
        }
        else if (strncmp(words[j], SI_SERVER_NAME, strlen(SI_SERVER_NAME)) == 0)
        {
            start = (int)strlen(SI_SERVER_NAME) + 1;
            info->server_name = toChar(&words[j][start], TRUE);
        }
    }

    free(templates);
    free(macros);
    
    if (minimum_free_disk_mbyte != 0 || minimum_free_disk_path)
    {
        // Setup defaults
        if (minimum_free_disk_mbyte == 0) 
        {
            minimum_free_disk_mbyte = 10000;
        }
        if (!minimum_free_disk_path)
        {
            minimum_free_disk_path = getWorkingFolder();
        }
    }

    errorsInFile = FALSE;
    for (j = 0; j < *size; j++)
    {
        if (processes[j].path_to_executable == NULL && !processes[j].is_service)
        {
            printf("Process [%s] does not have the path defined.\n", processes[j].alias);
            wlog("Process [%s] does not have the path defined.\n", "Setup", processes[j].alias);
            errorsInFile = TRUE;
        }
        if (processes[j].name == NULL)
        {
            if (processes[j].path_to_executable == NULL)
            {
                printf("Process [%s] does not have the name and path defined.\n", processes[j].alias);
                wlog("Process [%s] does not have the name and path defined.\n", "Setup", processes[j].alias);
                errorsInFile = TRUE;
            }
        }

        if (processes[j].path_to_executable != NULL)
        {
            unsigned int size = 0;
#if defined(WIN32) 
            char c = '\\';
#else
            char c = '/';
#endif
            char** segments = split(processes[j].path_to_executable, &size, c, TRUE, FALSE, FALSE);
            if (size == 0)
            {
                //there are no segments, reverse the file separator
                if (c == '\\') c = '/';
                else c = '\\';
                segments = split(processes[j].path_to_executable, &size, c, TRUE, FALSE, FALSE);
            }
            if (size == 0 && processes[j].name == NULL)
            {
                printf("Process [%s] does not have the name defined and it cannot be deduced from the path.\n", processes[j].alias);
                wlog("Process [%s] does not have the name defined and it cannot be deduced from the path.\n", "Setup", processes[j].alias);
                exit(3);
            }
            size_t byteSize = sizeof(char) * strlen(segments[size - 1]);
            if (processes[j].name == NULL)
            {
                processes[j].name = (char*)malloc(byteSize + 1);
                CheckNullReference(processes[j].name, 3);
                strcpy_s(processes[j].name, byteSize + 1, segments[size - 1]);
                processes[j].name[byteSize] = '\0';
                printf("The name of the process [%s] set to %s.\n", processes[j].alias, processes[j].name);
                wlog("The name of the process [%s] set to %s.\n", "Setup", processes[j].alias, processes[j].name);
            }

            if (_stricmp(processes[j].name, segments[size - 1]) != 0)
            {
                printf("The name of the process [%s] doesn't match the name of the executable %s.\n", processes[j].name, segments[size - 1]);
                wlog("The name of the process [%s] doesn't match the name of the executable %s.\n", "Setup", processes[j].name, segments[size - 1]);
                //errorsInFile = TRUE;
            }
            freeArray(segments, size);
        }

        if (processes[j].timeout_restart < 0)
        {
            processes[j].tine_device = NULL;
            processes[j].tine_property = NULL;
        }
        if (processes[j].tine_device != NULL && processes[j].tine_property != NULL)
        {
            processes[j].tine_status = -1;
        }
        for (n = 0; n < j; n++)
        {
            if (_stricmp(processes[j].alias, processes[n].alias) == 0)
            {
                printf("Process [%s] is duplicated. Please fix it.\n", processes[j].alias);
                errorsInFile = TRUE;
            }
        }
    }

    printf("--------------------------------------------------------\n");
    //sort processes so the ones that are dependent on others are started last
    //qsort(processes, *size, sizeof(WATCHED_PROCESS), processComparator);

    freeArray(words, num_lines);

    if (errorsInFile)
    {
        exit(3);
    }
       
    return processes;
}

int WriteConfigFile(char *file_name, WATCHED_PROCESS *processes, unsigned int num_processes, SERVER_INFO info, SERVER_INFO default_values)
{
    FILE *source, *target;
    char *temp;
    int ch;
    unsigned int i,j;
    size_t length,size;
    WATCHED_PROCESS defaultValuesExample;
    if (fileExists(file_name)) 
    {
        fopen_s(&source, file_name, "r");
        if (source == NULL) return 1;
        length = strlen(file_name) + 4;
        temp = (char*)alloca(sizeof(char)* (length + 1));
        strcpy_s(temp, length + 1, file_name);
        strcat_s(temp, length + 1, ".bck");
        temp[length] = '\0';
        fopen_s(&target, temp, "w");
        if (target == NULL) return 1;
        while ((ch = fgetc(source)) != EOF)
            fputc(ch, target);
        fclose(source);
        fclose(target);
        source = NULL;
    }
    fopen_s(&source, file_name, "w");
    InitializeDefaults(&defaultValuesExample);
    
    if (allow_monitoring_other_hosts != DEFAULT_ALLOW_MONITORING_OTHER_HOSTS) 
        fprintf(source, "%s: YES\n", WD_G_ALLOW_MONITORING_OTHER_HOSTS);
    if (watch_cycle_period != DEFAULT_WATCH_CYCLE_PERIOD)
        fprintf(source, "%s: %u\n", WD_G_WATCH_CYCLE_PERIOD, watch_cycle_period);
    if (print_info_on_each_cycle != DEFAULT_PRINT_INFO_ON_EACH_CYCLE)
        fprintf(source, "%s: YES\n", WD_G_PRINT_INFO_ON_EACH_CYCLE);
    if (start_server != DEFAULT_START_SERVER)
        fprintf(source, "%s: NO\n", WD_G_START_SERVER);
    if (suspended_timeout != DEFAULT_SUSPENDED_TIMEOUT)
        fprintf(source, "%s: %d\n", WD_G_SUSPENDED_TIMEOUT, suspended_timeout);
    if (pid_refresh_rate != DEFAULT_PID_REFRESH_RATE)
        fprintf(source, "%s: %d\n", WD_G_PID_REFRESH_RATE, pid_refresh_rate);
    if (kill_grace_period != DEFAULT_KILL_GRACE_PERIOD)
        fprintf(source, "%s: %d\n", WD_G_KILL_GRACE_PERIOD, kill_grace_period);
    if (short_term_history_seconds != DEFAULT_SHORT_TERM_HISTORY)
        fprintf(source, "%s: %d\n", WD_G_SHORT_TERM_HISTORY, short_term_history_seconds);
    if (long_term_history_months != DEFAULT_LONG_TERM_HISTORY)
        fprintf(source, "%s: %d\n", WD_G_LONG_TERM_HISTORY, long_term_history_months);
    if (stop_all_on_shutdown != DEFAULT_STOP_ALL_ON_SHUTDOWN)
        fprintf(source, "%s: YES\n", WD_G_STOP_ALL_ON_SHUTDOWN);
    if (g_network_enabled != DEFAULT_NETWORK_ENABLED)
        fprintf(source, "%s: NO\n", WD_G_NETWORK_ENABLED);
    if (minimum_free_disk_path) {
        fprintf(source, "%s: %s\n", WD_G_MINIMUM_FREE_DISK_PATH, minimum_free_disk_path);
        fprintf(source, "%s: %f\n", WD_G_MINIMUM_FREE_DISK_MBYTE, minimum_free_disk_mbyte);
    }

    if (console_command != NULL && (strlen(console_command) > 0))
        fprintf(source, "%s: %s\n", WD_G_CONSOLE_COMMAND, console_command);

    fprintf(source,"\n");

    if (_stricmp(info.context,default_values.context) != 0)
        fprintf(source, "%s: %s\n", SI_CONTEXT,info.context);
    if (_stricmp(info.server_name, default_values.server_name) != 0)
        fprintf(source, "%s: %s\n", SI_SERVER_NAME, info.server_name);
    if (_stricmp(info.fec_name, default_values.fec_name) != 0)
        fprintf(source, "%s: %s\n", SI_FEC_NAME, info.fec_name);
    if (info.port_offset != default_values.port_offset)
        fprintf(source, "%s: %d\n", SI_PORT_OFFSET, info.port_offset);
    if (_stricmp(info.subsystem, default_values.subsystem) != 0)
        fprintf(source, "%s: %s\n", SI_SUBSYSTEM, info.subsystem);
    if (_stricmp(info.hardware, default_values.hardware) != 0)
        fprintf(source, "%s: %s\n", SI_HARDWARE, info.hardware);
    if (_stricmp(info.responsible, default_values.responsible) != 0)
        fprintf(source, "%s: %s\n", SI_RESPONSIBLE, info.responsible);
    if (strlen(info.location) > 0)
        fprintf(source, "%s: %s\n", SI_LOCATION, info.location);
    if (info.description != NULL && strlen(info.description) > 0)
        fprintf(source, "%s: %s\n", SI_DESCRIPTION, info.description);

    fprintf(source,"\n");

    for (i = 0; i < num_processes; i++)
    {
        fprintf(source,"[%s]\n", processes[i].id);
        fprintf(source, "%s: %s\n",WD_NAME,processes[i].name);
        if (_stricmp(processes[i].id, processes[i].alias) != 0)
            fprintf(source, "%s: %s\n", WD_ALIAS, processes[i].alias);
        if (!processes[i].is_service)
            fprintf(source, "%s: %s\n", WD_PATH_TO_EXECUTABLE, processes[i].path_to_executable);
        if (processes[i].working_directory != NULL && strlen(processes[i].working_directory) > 0) 
            fprintf(source, "%s: %s\n", WD_WORKING_DIRECTORY, processes[i].working_directory);
        if (processes[i].caption != NULL && strlen(processes[i].caption) > 0)
            fprintf(source, "%s: %s\n", WD_PROCESS_CAPTION, processes[i].caption);
        if (processes[i].parameters != NULL && strlen(processes[i].parameters) > 0)
            fprintf(source, "%s: %s\n", WD_PARAMETERS, processes[i].parameters);
        if (processes[i].description != NULL && strlen(processes[i].description) > 0)
            fprintf(source, "%s: %s\n", WD_DESCRIPTION, processes[i].description);
        if (processes[i].is_service != defaultValuesExample.is_service)
            fprintf(source, "%s: %s\n", WD_SERVICE, processes[i].is_service ? "YES" : "NO");
        if (processes[i].show_console_window != defaultValuesExample.show_console_window)
            fprintf(source, "%s: %s\n", WD_SHOW_CONSOLE_WINDOW, processes[i].show_console_window ? "YES" : "NO");
    
        if (processes[i].number_of_environment_variables > 0)
        {
            length = 3000;
            temp = (char*)alloca(sizeof(char)* length);
            if (temp == NULL) OutOfMemoryExit(6);
            strcpy_s(temp, length, processes[i].environment_variables[0]);
            size = strlen(processes[i].environment_variables[0]);
            for (j = 1; j < processes[i].number_of_environment_variables; j++)
            {
                strcat_s(temp, length, "; ");
                strcat_s(temp, length, processes[i].environment_variables[j]);
                size += strlen(processes[i].environment_variables[j]) + 2;
            }
            temp[size] = '\0';
            fprintf(source, "%s: %s\n", WD_ENVIRONMENT_VARIABLES, temp);
        }

        if (processes[i].number_of_preceding_processes > 0)
        {
            length = 3000;
            temp = (char*)alloca(sizeof(char)* length);
            if (temp == NULL) OutOfMemoryExit(6);
            strcpy_s(temp, length, processes[i].preceding_process[0]);
            size = strlen(processes[i].preceding_process[0]);
            for (j = 1; j < processes[i].number_of_preceding_processes; j++)
            {
                strcat_s(temp, length, "; ");
                strcat_s(temp, length, processes[i].preceding_process[j]);
                size += strlen(processes[i].preceding_process[j]) + 2;
            }
            temp[size] = '\0';
            fprintf(source, "%s: %s\n", WD_PRECEDING_PROCESSES, temp);
        }

        if (processes[i].start_after_n_seconds != defaultValuesExample.start_after_n_seconds)
            fprintf(source, "%s: %d\n", WD_MONITOR_N_SECONDS_AFTER_START, processes[i].start_after_n_seconds);
        if (processes[i].pause_after_previous_start != defaultValuesExample.pause_after_previous_start)
            fprintf(source, "%s: %d\n", WD_PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS, processes[i].pause_after_previous_start);
        if (processes[i].number_of_retries != defaultValuesExample.number_of_retries)
            fprintf(source, "%s: %d\n", WD_NUMBER_OF_RETRIES, processes[i].number_of_retries);
        if (processes[i].retry_interval != defaultValuesExample.retry_interval)
            fprintf(source, "%s: %d\n", WD_RETRY_INTERVAL, processes[i].retry_interval);
        if (processes[i].auto_start != defaultValuesExample.auto_start)
            fprintf(source, "%s: %s\n", WD_AUTOSTART, processes[i].auto_start ? (processes[i].auto_start == YES ? "YES" : "ONCE") : "NO");
        if (processes[i].max_cpu_load != defaultValuesExample.max_cpu_load)
            fprintf(source, "%s: %d\n", WD_MAX_CPU_LOAD, processes[i].max_cpu_load);
        if (processes[i].number_of_allowed_cpu_load_violations != defaultValuesExample.number_of_allowed_cpu_load_violations)
            fprintf(source, "%s: %d\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS, processes[i].number_of_allowed_cpu_load_violations);
        if (processes[i].match_parameters != defaultValuesExample.match_parameters)
            fprintf(source, "%s: %s\n", WD_MATCH_PARAMETERS, processes[i].match_parameters ? "YES" : "NO");
        if (processes[i].match_path != defaultValuesExample.match_path)
            fprintf(source, "%s: %s\n", WD_MATCH_PATH, processes[i].match_path ? "YES" : "NO");
        if (processes[i].max_memory != defaultValuesExample.max_memory)
            fprintf(source, "%s: %d\n", WD_MAX_MEMORY, (int)(processes[i].max_memory/(THOUSAND*THOUSAND)));
        if (processes[i].number_of_allowed_memory_violations != defaultValuesExample.number_of_allowed_memory_violations)
            fprintf(source, "%s: %d\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS, processes[i].number_of_allowed_memory_violations);
        if (processes[i].tine_property != NULL && strlen(processes[i].tine_property) > 0)
            fprintf(source, "%s: %s\n", WD_TINE_PROPERTY, processes[i].tine_property);
        if (processes[i].tine_device != NULL && strlen(processes[i].tine_device) > 0)
            fprintf(source, "%s: %s\n", WD_TINE_DEVICE, processes[i].tine_device);
        if (processes[i].polling_rate != defaultValuesExample.polling_rate)
            fprintf(source, "%s: %d\n", WD_POLLING_RATE, processes[i].polling_rate);
        if (processes[i].timeout != defaultValuesExample.timeout)
            fprintf(source, "%s: %d\n", WD_TIMEOUT, processes[i].timeout);
        if (processes[i].timeout_restart != defaultValuesExample.timeout_restart)
            fprintf(source, "%s: %d\n", WD_RESTART_TIMEOUT, processes[i].timeout_restart);
                
        fprintf(source, "\n");
    }   

    fclose(source);
    return 0;
}

void PrintInstructions(void)
{
    FILE *file;
    char *computer_name;
    printf("Usage: watchdog <configuration_file>\n\n");
    printf("If the <configuration_file> is not provided, [%s] is used.\n", DEFAULT_CONFIG_FILE);
    printf("If the default configuration is also not available, this instructions are printed and example.cfg generated.\n\n");

    printf("The following properties can be configured:\n");
    printf("Global Properties:\n");
    printf("  %s\n    <Yes|No> indicates if the server should be started or\n    not (default: Yes).\n", WD_G_START_SERVER);
    printf("  %s\n    Time in milliseconds between two consecutive checks of\n    all processes (default: 1000 ms).\n", WD_G_WATCH_CYCLE_PERIOD);
    printf("  %s\n    <Yes|No> indicates if a process status should be printed\n    to console in each cycle (default: No).\n", WD_G_PRINT_INFO_ON_EACH_CYCLE);
    printf("  %s\n    Time in seconds how long the watchdog remains suspended\n    before it is automatically resumed (default: 600 s).\n", WD_G_SUSPENDED_TIMEOUT);
    printf("  %s\n    Time in seconds how often the PID of the processes are\n    refreshed. This assures that if a process is killed and\n    PIDs are recycled there is not a different process hiding\n    under the PID associated with a watched process (default: 60 s).\n", WD_G_PID_REFRESH_RATE);
    printf("  %s\n    <Yes|No> indicates if watchdog is allowed to monitor\n    properties from other hosts (default: no).\n", WD_G_ALLOW_MONITORING_OTHER_HOSTS);
    printf("  %s\n    The command which is used to start a process in a\n    separated console on linux (default: none).\n", WD_G_CONSOLE_COMMAND);
    printf("  %s\n    <Yes|No> indicates if all watched processes are sent\n    a stop signal if watchdog terminates gracefully (default: no).\n", WD_G_STOP_ALL_ON_SHUTDOWN);
    printf("  %s\n    Short term history in seconds (default: 300).\n", WD_G_SHORT_TERM_HISTORY);
    printf("  %s\n    Long term history in months (default: 1).\n", WD_G_LONG_TERM_HISTORY);
    printf("  %s\n    Grace period in seconds between term and kill signals\n    sent to a bad process (default: 10 s).\n", WD_G_KILL_GRACE_PERIOD);
    printf("  %s\n    List of environment variables as name=value pairs\n    separated by semi colon.\n", WD_G_ENVIRONMENT_VARIABLES);
    printf("  %s\n    <Yes|No> indicates if a the network utilization is\n    queried or not.\n", WD_G_NETWORK_ENABLED);
    printf("  %s\n    The path that is evaluated if it violates the minimum\n    free disk setting (default: working folder).\n", WD_G_MINIMUM_FREE_DISK_PATH);
    printf("  %s\n    Minimum free disk size in MB. Only used if path is\n    define as well (default: 10000).\n", WD_G_MINIMUM_FREE_DISK_MBYTE);

    printf("\nServer and FEC Properties (standard TINE restrictions apply):\n");
    printf("  %s\n    The name of the server (default: WD_<COMPUTER_NAME>).\n", SI_SERVER_NAME);
    printf("  %s\n    The FEC name identifying the Front End Computer and to\n    which all registered equipment modules are bound\n    (default: <COMPUTER_NAME>.<PORT_OFFSET>).\n", SI_FEC_NAME);
    printf("  %s\n    The subsystem to which attached device servers belong\n    (default: WDOG).\n", SI_SUBSYSTEM);
    printf("  %s\n    32-character string giving the context of all registered\n    equipment modules (default: SERVICE).\n", SI_CONTEXT);
    printf("  %s\n    64-character description of the FECs server duties\n    (default: n/a).\n", SI_DESCRIPTION);
    printf("  %s\n    32-character string giving the physical location of\n    the FEC (default: n/a).\n", SI_LOCATION);
    printf("  %s\n    32-character brief description of the IO hardware found\n    on the FEC (default: None).\n", SI_HARDWARE);
    printf("  %s\n    32-character string listing the developer(s) responsible\n    for the FEC (default: Admin).\n", SI_RESPONSIBLE);
    printf("  %s\n    The Port Offset to be applied to the FEC (default: 20).\n", SI_PORT_OFFSET);

    printf("\nIndividual Process Properties (arrow marks mandatory settings):\n");
    printf("  -> [<process ID>]\n    The process id is always given in rectangular brackets\n    and defines the start of a process configuration. This is\n    the exported device name for the process, when server is started.\n");
    printf("  -> %s\n    Path to the process executable (use quotes if it contains spaces); mandatory if the process is not a service\n", WD_PATH_TO_EXECUTABLE);
    printf("  -> %s\n    Name of the process (the same as it appears in the OS process list).\n    If not provided the last segment of %s is used. Mandatory\n    if service.\n", WD_NAME, WD_PATH_TO_EXECUTABLE);
    printf("  %s\n    An alias for the exported device names (default: same as ID).\n", WD_ALIAS);
    printf("  %s\n    <Yes|No> indicates if this process is a service (Windows only).\n", WD_SERVICE);
    printf("  %s\n    <Yes|No> indicates if the console of this process should be\n    visible or not (Windows only, default: Yes).\n", WD_SHOW_CONSOLE_WINDOW);
    printf("  %s\n    The description of the process\n", WD_DESCRIPTION);
    printf("  %s\n    Working directory where to start the process (use quotes\n    if it contains spaces) (default: current directory).\n", WD_WORKING_DIRECTORY);
    printf("  %s\n    Process startup parameters, separated by spaces.\n    Put quotes around individual parameters if they contain\n    spaces (default: none).\n", WD_PARAMETERS);
    printf("  %s\n    <Yes|No> indicates if parameters listed for this\n    process should match the parameters in the process' command\n    line. Use this setting when several processes with the same\n    name are started using different command line parameters (default: Yes).\n", WD_MATCH_PARAMETERS);
    printf("  %s\n    <Yes|No> indicates if path of this process should\n    match the path to executable listed for this process. Use\n    this setting when several processes with the same name are\n    started from different directories (default: Yes).\n", WD_MATCH_PATH);
    printf("  %s\n    The caption of the process if the process is a\n    console application (default: cmd line).\n", WD_PROCESS_CAPTION);
    printf("  %s\n    Environment variables to use in addition to the default\n    ones. Individual variables should be given in name=value\n    pairs, and should be separated by semicolons. If you want to\n    delete a variable that is set by the parent, set its\n    value to an empty string: VAR=; (default: none).\n", WD_ENVIRONMENT_VARIABLES);
    printf("  %s\n    Processes to start before this process is started,\n    separated by semicolons. The process has to be the ID\n    or alias of one of the other processes managed by watchdog,\n    or a system process (default: none).\n", WD_PRECEDING_PROCESSES);
    printf("  %s\n    Number of seconds to wait after the process was started\n    and before the watchdog starts to watch it. This allows the\n    process to take time for its startup procudere (default: 10 s).\n", WD_MONITOR_N_SECONDS_AFTER_START);
    printf("  %s\n    Number of seconds to wait after the previous process was\n    and before this process is started (default: 1 s).\n", WD_PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS);
    printf("  %s\n    Number of times the process is started before it is\n    disabled (disabled processes are no longer managed by\n    watchdog until started manually) (default: 3).\n", WD_NUMBER_OF_RETRIES);
    printf("  %s\n    Number of seconds to wait before trying to start the\n    process after it was discovered it is not running (default: 20 s).\n", WD_RETRY_INTERVAL);
    printf("  %s\n    <Yes|No|Once> indicates if a non-running process should be\n    started automatically by watchdog. If the value is once\n    the process can only be started with a command to the\n    watchdog (e.g. script execution). The process will not be\n    monitored in that case (default: Yes).\n", WD_AUTOSTART);
    printf("  %s\n    Maximum allowed CPU load in percentage (default: 100 %% ).\n", WD_MAX_CPU_LOAD);
    printf("  %s\n    How many times the max CPU load can be violated, before\n    the process is restarted (default: 1).\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS);
    printf("  %s\n    Maximum memory consumption (commit size) in MB (default: 1 TB).\n", WD_MAX_MEMORY);
    printf("  %s\n    How many times the max memory limit can be violated\n    before the process is restarted (default: 1).\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS);

    printf("  %s\n    TINE property to monitor. If this process is a TINE\n    server, define the property which will be monitored.\n    If no response is received from the property the process\n    is restarted (default: none).\n", WD_TINE_PROPERTY);
    printf("  %s\n    TINE device that owns the monitored property (default: none).\n", WD_TINE_DEVICE);
    printf("  %s\n    The rate in milliseconds at which the TINE property\n    is polled (default: 1000 ms).\n", WD_POLLING_RATE);
    printf("  %s\n    Timeout in seconds at which the process is considered\n    not responding if no update has been received from the\n    property (default: 10 s).\n", WD_TIMEOUT);
    printf("  %s\n    Timeout in seconds at which the process will be restarted\n    if TINE property is not responding. Enter 0 if the\n    process should not be restarted (default: 0 s).\n", WD_RESTART_TIMEOUT);

    fopen_s(&file, "example.cfg", "w");

    if (file == NULL)
    {
        printf("\n******************************************************************************\n");
        printf("*                                                                            *\n");
        printf("* Example file example.cfg could not be generated. Check folder permissions. *\n");
        printf("*                                                                            *\n");
        printf("******************************************************************************\n");
        return;
    }
    fprintf(file, "# Watchdog config\n");
#if defined(UNIX)
    fprintf(file, "%s: gnome-terminal -t <caption> -x sh -c <process>\n", WD_G_CONSOLE_COMMAND);
#endif
    fprintf(file, "%s: Yes\n", WD_G_START_SERVER);
    fprintf(file, "%s: No\n", WD_G_PRINT_INFO_ON_EACH_CYCLE);
    fprintf(file, "%s: 1000\n", WD_G_WATCH_CYCLE_PERIOD);
    fprintf(file, "%s: 1800\n", WD_G_SUSPENDED_TIMEOUT);
    fprintf(file, "%s: No\n", WD_G_ALLOW_MONITORING_OTHER_HOSTS);
    fprintf(file, "%s: No\n", WD_G_STOP_ALL_ON_SHUTDOWN);
    fprintf(file, "%s: 300\n", WD_G_SHORT_TERM_HISTORY);
    fprintf(file, "%s: 1\n\n", WD_G_LONG_TERM_HISTORY);
    fprintf(file, "%s: 10\n\n", WD_G_KILL_GRACE_PERIOD);

    fprintf(file, "# FEC config\n");
    fprintf(file, "# Uncomment the following line to change any of the default parameters\n");
    computer_name = GetMachineName();
    fprintf(file, "#%s: %s.20\n", SI_FEC_NAME, computer_name);
    fprintf(file, "#%s: WD_%s\n", SI_SERVER_NAME, computer_name);
    free(computer_name);
    fprintf(file, "#%s: SERVICE\n", SI_CONTEXT);
    fprintf(file, "#%s: WDOG\n", SI_SUBSYSTEM);
    fprintf(file, "#%s: Watchdog\n", SI_DESCRIPTION);
    fprintf(file, "#%s: \n", SI_LOCATION);
    fprintf(file, "#%s: none\n", SI_HARDWARE);
    fprintf(file, "#%s: Admin\n", SI_RESPONSIBLE);
    fprintf(file, "#%s: 20\n\n", SI_PORT_OFFSET);

    fprintf(file, "# Processes config\n");
    fprintf(file, "# These are example configurations, which may not make 100%% sense. Use them as a reference only.\n");
    fprintf(file, "# Complete list of parameters and their explanation can be obtained by running watchdog without any configuration file.\n");

#if defined(UNIX)
    fprintf(file, "[sinegen]\n");
    fprintf(file, "%s: sinegen\n", WD_NAME);
    fprintf(file, "%s: /home/tine/tine/servers/SineGenerator/bin/sinegen\n", WD_PATH_TO_EXECUTABLE);
    fprintf(file, "%s: /home/tine/tine/servers/SineGenerator/bin\n", WD_WORKING_DIRECTORY);
    fprintf(file, "%s: Yes\n", WD_MATCH_PATH);
    fprintf(file, "%s: No\n", WD_MATCH_PARAMETERS);
    fprintf(file, "%s: 50\n", WD_MAX_CPU_LOAD);
    fprintf(file, "%s: 2\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS);
    fprintf(file, "%s: 100\n", WD_MAX_MEMORY);
    fprintf(file, "%s: 1\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS);
    fprintf(file, "%s: /TEST/WinSineServer/SineGen0\n", WD_TINE_DEVICE);
    fprintf(file, "%s: Amplitude\n", WD_TINE_PROPERTY);
    fprintf(file, "%s: 200\n", WD_POLLING_RATE);
    fprintf(file, "%s: 5\n", WD_TIMEOUT);
    fprintf(file, "%s: 100\n", WD_RESTART_TIMEOUT);

    fprintf(file, "\n[about]\n");
    fprintf(file, "%s: gnome-about\n", WD_NAME);
    fprintf(file, "%s: /usr/bin/gnome-about\n", WD_PATH_TO_EXECUTABLE);
    fprintf(file, "%s: /home\n", WD_WORKING_DIRECTORY);
    fprintf(file, "%s: test\n", WD_PARAMETERS);
    fprintf(file, "%s: TEST=foo; TINE_HOME=C:\\tine\n", WD_ENVIRONMENT_VARIABLES);
    fprintf(file, "%s: No\n", WD_MATCH_PATH);
    fprintf(file, "%s: Yes\n", WD_MATCH_PARAMETERS);
    fprintf(file, "%s: 20\n", WD_MONITOR_N_SECONDS_AFTER_START);
    fprintf(file, "%s: 5\n", WD_NUMBER_OF_RETRIES);
    fprintf(file, "%s: 10\n", WD_RETRY_INTERVAL);
    fprintf(file, "%s: Yes\n", WD_AUTOSTART);
    fprintf(file, "%s: 10\n", WD_MAX_CPU_LOAD);
    fprintf(file, "%s: 2\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS);
    fprintf(file, "%s: 5\n", WD_MAX_MEMORY);
    fprintf(file, "%s: 1\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS);

    fprintf(file, "\n[firefox]\n");
    fprintf(file, "%s: firefox\n", WD_NAME);
    fprintf(file, "%s: /usr/lib64/firefox/firefox\n", WD_PATH_TO_EXECUTABLE);
    fprintf(file, "%s: gnome-about; sinegen; vi\n", WD_PRECEDING_PROCESSES);
    fprintf(file, "%s: www.google.com\n", WD_PARAMETERS);
    fprintf(file, "%s: 30\n", WD_MONITOR_N_SECONDS_AFTER_START);
    fprintf(file, "%s: 5\n", WD_NUMBER_OF_RETRIES);
    fprintf(file, "%s: 20\n", WD_RETRY_INTERVAL);
    fprintf(file, "%s: Yes\n", WD_AUTOSTART);
    fprintf(file, "%s: 30\n", WD_MAX_CPU_LOAD);
    fprintf(file, "%s: 5\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS);
    fprintf(file, "%s: 1000\n", WD_MAX_MEMORY);
    fprintf(file, "%s: 1\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS);

#elif defined(WIN32)
    fprintf(file, "[notepad]\n");
    fprintf(file, "%s: notepad.exe\n", WD_NAME);
    fprintf(file, "%s: C:\\Windows\\notepad.exe\n", WD_PATH_TO_EXECUTABLE);
    fprintf(file, "%s: C:\\\n", WD_WORKING_DIRECTORY);
    fprintf(file, "%s: C:\\test.txt\n", WD_PARAMETERS);
    fprintf(file, "%s: TEST=foo; TINE_HOME=C:\\tine\n", WD_ENVIRONMENT_VARIABLES);
    fprintf(file, "%s: firefox.exe; regedit.exe\n", WD_PRECEDING_PROCESSES);
    fprintf(file, "%s: No\n", WD_MATCH_PATH);
    fprintf(file, "%s: Yes\n", WD_MATCH_PARAMETERS);
    fprintf(file, "%s: 20\n", WD_MONITOR_N_SECONDS_AFTER_START);
    fprintf(file, "%s: 5\n", WD_NUMBER_OF_RETRIES);
    fprintf(file, "%s: 10\n", WD_RETRY_INTERVAL);
    fprintf(file, "%s: Yes\n", WD_AUTOSTART);
    fprintf(file, "%s: 10\n", WD_MAX_CPU_LOAD);
    fprintf(file, "%s: 2\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS);
    fprintf(file, "%s: 5\n", WD_MAX_MEMORY);
    fprintf(file, "%s: 1\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS);
    fprintf(file, "%s: /TEST/WinSineServer/SineGen0\n", WD_TINE_DEVICE);
    fprintf(file, "%s: Amplitude\n", WD_TINE_PROPERTY);
    fprintf(file, "%s: 200\n", WD_POLLING_RATE);
    fprintf(file, "%s: 5\n", WD_TIMEOUT);
    fprintf(file, "%s: 100\n", WD_RESTART_TIMEOUT);

    fprintf(file, "\n[firefox]\n");
    fprintf(file, "%s: firefox.exe\n", WD_NAME);
    fprintf(file, "%s: \"C:\\Program Files (x86)\\Mozilla Firefox\\firefox.exe\"\n", WD_PATH_TO_EXECUTABLE);
    fprintf(file, "%s: 30\n", WD_MONITOR_N_SECONDS_AFTER_START);
    fprintf(file, "%s: 5\n", WD_NUMBER_OF_RETRIES);
    fprintf(file, "%s: 20\n", WD_RETRY_INTERVAL);
    fprintf(file, "%s: Yes\n", WD_AUTOSTART);
    fprintf(file, "%s: 30\n", WD_MAX_CPU_LOAD);
    fprintf(file, "%s: 5\n", WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS);
    fprintf(file, "%s: 500\n", WD_MAX_MEMORY);
    fprintf(file, "%s: 1\n", WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS);
#endif

    fclose(file);

    printf("\n*************************************************************************\n");
    printf("*                                                                       *\n");
    printf("*   Example configuration file has been created for you: example.cfg.   *\n");
    printf("*                                                                       *\n");
    printf("*************************************************************************\n");
}
