/*
* parser.h
*
* Provides routines to parse and print the watchdog configuration file.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#ifndef PARSER_H_
#define PARSER_H_

#include "watchdog.h"

#define MAX_MACRO_COUNT 1000
#define MAX_TEMPLATE_COUNT 20
#define MAX_PAIR_COUNT 50

/* Macro struct, keys should be in format $(NAME) */
typedef struct
{
    char* key;
    char* value;
} MACRO;

/* Template struct */
typedef struct
{
    char* name;
    unsigned int numberOfPairs;
    char* keys[MAX_PAIR_COUNT];
    char* values[MAX_PAIR_COUNT];
} TEMPLATE;

/**
* \brief Search through lines, identify macros and fill them into given array. Number of found
* macros is stored in macros_count.
* \param lines of input text
* \param num_lines number of lines
* \param macros array
* \param macros_count passed by reference returning the number of macros
* \param parse_all_macros boolean if only global macros should be parsed or all macros found in lines
*/
void FindMacros(char** lines, unsigned int num_lines, MACRO macros[], unsigned int* macros_count, boolean parse_all_macros);
/**
* \brief Search through lines and evaluate macro references with appropriate text.
* \param lines of input text
* \param num_lines number of lines
* \param macros array
* \param macros_count number of macros
*/
void EvaluateMacros(char** lines, unsigned int num_lines, MACRO macros[], unsigned int macros_count);

/**
* \brief Search through lines, identify templates and fill them into given array. Number of found
* templates is stored in templates_count.
* \param lines of input text
* \param num_lines number of lines
* \param templates array
* \param templates_count passed by reference returning the number of macros
*/
void FindTemplates(char** lines, unsigned int num_lines, TEMPLATE templates[], unsigned int* templates_count);

/**
* \brief Search through lines and evaluate configurations with templates. For each template copy values
* from template and update it specified configuration values. After that evaluate local template macros.
* \param lines_pointer pointer to lines of input text.
* \param num_lines number of lines passed by reference passing starting number of lines and returning
* end number of lines.
* \param templates array
* \param templates_count number of templates
*/
void EvaluateTemplates(char*** lines_pointer, unsigned int* num_lines, TEMPLATE templates[], unsigned int templates_count);
/**
* The process watching period in milliseconds. The watchdog should check the
* the process on every cycle. The watch_cycle_period defined the time between
* two consecutive checks.
*/
extern unsigned int watch_cycle_period;
/*
* Defines if the process status should be printed out on every watch cycle.
* If 1 the data should be printed if 0 data should not be printed.
*/
extern boolean print_info_on_each_cycle;
/**
* Defines if the server is started, which publishes all the processes' statuses.
*/
extern boolean start_server;
/*
* \brief Initializes the default values of the process.
*
* The routine initializes all the fields of the structure. It is expected that
* all pointers within the structure point to some location and that allow primitive
* fields have some meaningful values.
*
* \param process the process to initialize
*/
void InitializeDefaults(WATCHED_PROCESS *process);
/**
* \brief Prints the information about all the processes.
*
* Prints the static info of the processes. The method should print the name, path,
* working folder, restart criteria etc. in a human readable format.
*
* \param process the pointer to the array of processes
* \param size the number of processes
*/
void PrintProcessInfo(WATCHED_PROCESS *process, unsigned int size);
/**
* \brief Print the settings of this watchdog.
*
* Prints the settings of this watchdog, including all the processes info.
*
* \param processes the pointer to array of processes
* \param size the number of processes
* \param info the server info of this watchdog
*/
void PrintSettings(WATCHED_PROCESS *processes, unsigned int size, SERVER_INFO *info);
/**
* \brief Parse the file and return the pointer to the array of watched processes.
*
* Parses the file given by the parameter file_name. The file is expected to
* follow the following rules:
* - The alias of the process is defined in square brackets []. It is the first info of a process definition.
* - Individual fields are defined as NAME: value pairs. The order is not important.
* - NAME and PATH are required fields
* The routine should return a pointer to the array of processes that contain the static
* info read from the file. In addition the size of the array should be provided in the
* size parameter.
*
* \param file_name the path to the file to parse
* \param size the number of returned processes
* \param info the fec information read from the file
*/
WATCHED_PROCESS *ParseFile(char *file_name, unsigned int *size, SERVER_INFO *info);
/**
* \brief Writes the configuration from the content of the processes
* 
* Writes the settings of all processes to a configuration file. If the file exists
* (and it normally does) it will be overwritten. The old file will be backed out
* to file_name.bak. 
*
* \param file_name the name of the destination file
* \param processes the processes to write to the configuration file
* \param num_processes the number of processes
* \param info the server info to dump into the file
* \param default_info default values for the server info

* \return 0 if the write was successful or the error code otherwise
*/
int WriteConfigFile(char *file_name, WATCHED_PROCESS *processes, unsigned int num_processes, SERVER_INFO info, SERVER_INFO default_info);
/**
* \brief Prints instructions how to start the watchdog
*
* Prints details instructions how to use and start the watchdog.
*/
void PrintInstructions(void);

#endif
