/*
* process.c
*
* Implements routines from the process.h.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#include <time.h>
#include "util.h"
#include "process.h"

static int num_processors = -1;

char **findDeletedVariables(char **env, unsigned int env_size, unsigned int *size)
{
    unsigned int i, m, n;
    char **ret, **temp;
    ret = (char**)malloc(sizeof(char*)* env_size);
    if (ret == NULL) OutOfMemoryExit(5);
    n = 0;
    for (i = 0; i < env_size; i++)
    {
        temp = split(env[i], &m, '=', TRUE, TRUE, FALSE);
        if (m == 1)
        {
            ret[n] = temp[0];
            n++;
            free(temp);
        }
        else
        {
            freeArray(temp, 2);
        }
    }
    *size = n;
    return ret;
}

boolean isDeleted(char *variable, char **del_env, unsigned int delsize)
{
    char **temp;
    unsigned int tempsize, i;
    size_t cmplength;

    temp = split(variable, &tempsize, '=', TRUE, FALSE, FALSE);
    cmplength = strlen(temp[0]);
    boolean found = FALSE;
    for (i = 0; i < delsize; i++)
    {
        if (strncmp(temp[0], del_env[i], cmplength) == 0)
        {
            found = TRUE;
            break;
        }
    }
    freeArray(temp, tempsize);
    return found;
}

boolean isDuplicate(char *variable, char **envs, unsigned int envs_size)
{
    char *env_name;
    unsigned int i;

    unsigned int length = (int)strlen(variable);
    for (i = 0; i < length; i++)
    {
        if (variable[i] == '=')
        {
            length = i;
            break;
        }
    }
    env_name = (char*)alloca(sizeof(char) * length + 2);
    if (env_name == NULL) OutOfMemoryExit(5);
    env_name[length + 1] = '\0';
    memcpy(env_name, variable, length+1);
    for (i = 0; i < envs_size; i++)
    {
        if (strncmp(envs[i], env_name, length+1) == 0)
        {
            return TRUE;
        }
    }
    return FALSE;
}

#if defined(UNIX)

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/sysinfo.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <sys/reboot.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include "watchdog.h"
#include "network.h"

#define MAX_PROC_PATH_SIZE 300
#define MAX_CMD_LINE 2048
#define PROCESS_TAG "<process>"
#define CAPTION_TAG "<caption>"

char process[MAX_PROC_PATH_SIZE];
UINT64 *proc_time;
UINT64 total_active_time, total_time, wProc_time, wTotal_time;
struct sysinfo mem_info;
char **split_console_command;
unsigned int console_command_size;

int GetPid(void)
{
    pid_t pid = getpid();
    return (int)pid;
}

void InitBuffers(unsigned int size)
{
    proc_time = (UINT64 *)malloc(sizeof(UINT64)* size);
    if (proc_time == NULL) OutOfMemoryExit(5);
    total_active_time = 0;
    total_time = 0;    
    wProc_time = 0;
    wTotal_time = 0;
    sysinfo(&mem_info);
    num_processors = get_nprocs();
    if (console_command != NULL)
        split_console_command = split(console_command, &console_command_size, ' ', TRUE, TRUE, FALSE);
}

void FreeBuffers(void)
{
    if (proc_time != NULL) free(proc_time);
    freeArray(split_console_command, console_command_size);
}

UINT64 getMemoryForProcess(int process_id)
{
    FILE* file;
    UINT64 memory;
    static int page_size = -1;
    if (page_size < 0) page_size = getpagesize();

    sprintf(process, "/proc/%d/statm", process_id);
    file = fopen(process, "r");
    if (file == NULL) return 0;
    //we are only interested into 1 parameter: virtual memory
    fscanf(file, "%"PRIu64" %*s %*s %*s %*s %*s %*s", &memory);
    fclose(file);

    memory *= page_size;

    return memory;
}

void GetWatchdogCPULoadAndMemory(int pid, int *cpu_load, UINT64 *memory)
{
    FILE* file;
    UINT64 total, total_time1, total_time2, total_time3, total_time4, user_time, system_time;

    sprintf(process, "/proc/%d/stat", pid);
    file = fopen(process, "r");
    if (file == NULL) OutOfMemoryExit(5);
    //we are only interested into 2 parameters: utime and stime, which 14th and 15th parameter in the row
    fscanf(file, "%*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %"PRIu64" %"PRIu64" %*s %*s %*s %*s %*s %*s %*s %*s",
        &user_time, &system_time);
    fclose(file);

    file = fopen("/proc/stat", "r");
    if (file == NULL) OutOfMemoryExit(5);
    fscanf(file, "cpu %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64, &total_time1, &total_time2, &total_time3, &total_time4);
    fclose(file);
    total = total_time1 + total_time2 + total_time3 + total_time4;

    cpu_load[0] = (int) (100 * (user_time + system_time - wProc_time) /
                               (float)(wTotal_time - total) + 0.5);
    if (cpu_load[0] < 0) cpu_load[0] = 0;
    wProc_time = user_time + system_time;
    wTotal_time = total;
    memory[0] = getMemoryForProcess(pid);
}

void GetCPULoadAndMemory(int *process_ids, unsigned int size, int *cpu_load, UINT64 *memory, int *cpu_total)
{
    double percent;
    FILE* file;
    UINT64 total_time1, total_time2, total_time3, total_time4, total_up, total_bottom,
           user_time, system_time, total_diff;
    unsigned int i;

    file = fopen("/proc/stat", "r");
    if (file == NULL) OutOfMemoryExit(5);
    fscanf(file, "cpu %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64, &total_time1, &total_time2, &total_time3, &total_time4);
    fclose(file);
    total_up = total_time1 + total_time2 + total_time3;
    total_bottom = total_up + total_time4;
    total_diff = total_bottom - total_time;
    
    percent = 100 * (total_up - total_active_time) / total_diff;
    *cpu_total = (int)(percent + 0.5);
    total_active_time = total_up;
    total_time = total_bottom;

    for (i = 0; i < size; i++)
    {
        sprintf(process, "/proc/%d/stat", process_ids[i]);
        file = fopen(process, "r");
        if (file == NULL) continue;
        //we are only interested into 2 parameters: utime and stime, which 14th and 15th parameter in the row
        fscanf(file,
            "%*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %"PRIu64" %"PRIu64" %*s %*s %*s %*s %*s %*s %*s %*s",
            &user_time, &system_time);
        fclose(file);
        cpu_load[i] = (int)(100.0 * (user_time + system_time - proc_time[i]) /
                                    (float)total_diff + 0.5);
        if (cpu_load[i] < 0) cpu_load[i] = 0;
        proc_time[i] = user_time + system_time;
    }  

    for (i = 0; i < size; i++) {
        memory[i] = getMemoryForProcess(process_ids[i]);
    }
}

UINT64 GetAvailableMemory(void)
{
    UINT64 mem;
    mem = mem_info.freeram;
    mem *= mem_info.mem_unit;
    return mem;
}

/*
* Check if the string is purely an integer.
* \param str the string to check
* \return 0 if str is not an integer, or 1 if it is an integer
*/
boolean checkIfNumber(char *str)
{
    int i;
    for (i = 0; str[i] != '\0'; i++)
    {
        if (!isdigit(str[i])) return FALSE;
    }
    return TRUE;
}

char *replace_str(const char *str, const char *old, const char *new)
{
    char *ret, *r;
    const char *p, *q;
    size_t l;
    size_t oldlen = strlen(old);
    size_t count, retlen, newlen = strlen(new);

    if (oldlen != newlen)
    {
        for (count = 0, p = str; (q = strstr(p, old)) != NULL; p = q + oldlen)
            count++;
        /* this is undefined if p - str > PTRDIFF_MAX */
        retlen = p - str + strlen(p) + count * (newlen - oldlen);
    }
    else retlen = strlen(str);

    if ((ret = malloc(retlen + 1)) == NULL)
        OutOfMemoryExit(5);

    for (r = ret, p = str; (q = strstr(p, old)) != NULL; p = q + oldlen)
    {
        /* this is undefined if q - p > PTRDIFF_MAX */
        l = q - p;
        memcpy(r, p, l);
        r += l;
        memcpy(r, new, newlen);
        r += newlen;
    }
    strcpy(r, p);

    return ret;
}

/**
* Reads the command line of the process defined by the pid.
* The method returns the complete command line, together with parameters.
* Individual elements are separated by white spaces.
*
* \param pid the process id for which to read the command line
* \return the complete command line of the process
*/
char *GetCommandLineForProcess(int pid)
{
    char *cmd_line;
    FILE *file;
    char process[MAX_PROC_PATH_SIZE];
    int len, i;

    sprintf(process, "/proc/%d/cmdline", pid);
    file = fopen(process, "r");
    if (file == NULL) return NULL;

    cmd_line = (char *)malloc(sizeof(char)* MAX_CMD_LINE);
    fscanf(file, "%2048s%n", cmd_line, &len);
    fclose(file);

    for (i = 0; i < len - 1; i++)
    {
        if (cmd_line[i] == '\0') cmd_line[i] = ' ';
    }

    return cmd_line;
}

int startProcessDirectly(char *cmd_line, char **parameters, unsigned int par_size,
    char *working_directory, char *caption, char **environment, unsigned int env_size)
{
    pid_t pid;
    char *env[num_environment + env_size + 1];
    char *par[par_size + 2];
    char **del_env;
    char *cmd;
    int i, length, c;
    unsigned int delsize;
    pid = fork();

    if (pid == 0) //child process
    {
        setsid();
        if (env_size > 0)
        {
            del_env = findDeletedVariables(environment, env_size, &delsize);
            c = 0;
            //setup environment if it exists, otherwise just let the startup process to inherit the current one
            for (i = 0; i < env_size; i++)
            {
                //do not add any variable that was deleted
                if (isDeleted(environment[i], del_env, delsize) == 0)
                {
                    length = strlen(environment[i]);
                    env[c] = (char *)malloc(sizeof(char)* (length + 1));
                    if (env[c] == NULL) OutOfMemoryExit(5);
                    memcpy(env[c], environment[i], length);
                    env[c][length] = '\0';
                    c++;
                }
            }
            for (i = 0; i < num_environment; i++)
            {
                if (isDeleted(parent_environment[i], del_env, delsize) == 0 &&
                    isDuplicate(parent_environment[i], environment, env_size) == 0)
                {
                    length = strlen(parent_environment[i]);
                    env[c] = (char *)malloc(sizeof(char)* (length + 1));
                    if (env[c] == NULL) OutOfMemoryExit(5);
                    memcpy(env[c], parent_environment[i], length);
                    env[c][length] = '\0';
                    c++;
                }
            }
            env[c] = NULL;
            freeArray(del_env, delsize);
        }
        //change the working directory of the process
        if (working_directory != NULL) chdir(working_directory);

        //set the command line parameters
        length = strlen(cmd_line);
        par[0] = (char *)malloc(sizeof(char)* (length + 1));
        if (par[0] == NULL) OutOfMemoryExit(5);
        memcpy(par[0], cmd_line, length);
        par[0][length] = '\0';

        for (i = 0; i < par_size; i++)
        {
            length = strlen(parameters[i]);
            par[i + 1] = (char *)malloc(sizeof(char)* (length + 1));
            if (par[i + 1] == NULL) OutOfMemoryExit(5);
            memcpy(par[i + 1], parameters[i], length);
            par[i + 1][length] = '\0';
        }
        par[par_size + 1] = NULL;
        //copy command line to have a fresh part of the memory, so we can get rid of all watchdog resources
        cmd = (char *)malloc(sizeof(char)* (strlen(cmd_line) + 1));
        strcpy(cmd, cmd_line);
        cmd[strlen(cmd_line)] = '\0';
        //start the process
        if (env_size > 0)
            i = execve(cmd, par, env);
        else
            i = execvp(cmd, par);

        if (i != 0)
        {
            i = errno;
            printf("Process %s did not start successfully (%d): %s.\n", process, errno, strerror(errno));
        }
        //on process termination, printout, free memory and exit this process
        if (env_size > 0)
        {
            for (i = 0; i < env_size + num_environment + 1; i++)
            {
                if (env[i] != NULL) free(env[i]);
            }
        }
        for (i = 0; i < par_size + 2; i++)
        {
            if (par[i] != NULL) free(par[i]);
        }
        free(cmd);
        exit(1);
    }
    else if (pid > 0) //parent process
    {
        wlog("Started process [%s] directly. Running under process id (%d).\n", "Process", cmd_line, pid);
        return pid;
    }
    else
    {
        wlog("Could not execute command [%s].\n", "Process", cmd_line);
        return -errno;
    }
}

int StartProcess(char *cmd_line, char **parameters, unsigned int par_size, char *working_directory,
    char *caption, char **environment, unsigned int env_size, boolean is_service, boolean show_console)
{
    pid_t pid;
    char *env[num_environment + env_size + 1];
    char *par[console_command_size + 1];
    char **del_env;
    char *cmd, *process;
    int i, length, c;
    unsigned int delsize;

    //if command is a path to a process verify that the executable really exists
    if (strchr(cmd_line, '/') != NULL)
    {
        FILE *f = fopen(cmd_line, "r");
        if (f == NULL)
        {
            return -ENOENT;
        }
    }

    if (console_command == NULL)
    {
        return startProcessDirectly(cmd_line, parameters, par_size, working_directory, caption,
            environment, env_size);
    }

    pid = fork();

    if (pid == 0) //child process
    {
        setsid();
        if (env_size > 0)
        {
            del_env = findDeletedVariables(environment, env_size, &delsize);
            c = 0;
            //setup environment if it exists, otherwise just let the startup process to inherit the current one
            for (i = 0; i < env_size; i++)
            {
                //do not add any variable that was deleted
                if (isDeleted(environment[i], del_env, delsize) == 0)
                {
                    length = strlen(environment[i]);
                    env[c] = (char *)malloc(sizeof(char)* (length + 1));
                    if (env[c] == NULL) OutOfMemoryExit(5);
                    memcpy(env[c], environment[i], length);
                    env[c][length] = '\0';
                    c++;
                }
            }
            for (i = 0; i < num_environment; i++)
            {
                if (isDeleted(parent_environment[i], del_env, delsize) == 0 &&
                    isDuplicate(parent_environment[i], environment, env_size) == 0)
                {
                    length = strlen(parent_environment[i]);
                    env[c] = (char *)malloc(sizeof(char)* (length + 1));
                    if (env[c] == NULL) OutOfMemoryExit(5);
                    memcpy(env[c], parent_environment[i], length);
                    env[c][length] = '\0';
                    c++;
                }
            }
            env[c] = NULL;
            freeArray(del_env, delsize);
        }
        //change the working directory of the process
        if (working_directory != NULL) chdir(working_directory);

        //create the <process> command
        length = 0;
        for (i = 0; i < par_size; i++)
            length += strlen(parameters[i]) + 1;
        c = strlen(cmd_line) + 1 + length;
        process = (char *)alloca(sizeof(char)* c);
        if (process == NULL) OutOfMemoryExit(5);
        memset(process, ' ', c);
        strcpy(process, cmd_line);
        length = strlen(cmd_line);
        process[length] = ' ';
        for (i = 0; i < par_size; i++)
        {
            strcpy(&process[length + 1], parameters[i]);
            length += strlen(parameters[i]) + 1;
            process[length] = ' ';
        }
        process[c - 1] = '\0';

        for (i = 0; i < console_command_size; i++)
        {
            if (strstr(split_console_command[i], PROCESS_TAG) != NULL)
            {
                //replace <process> with process
                cmd = replace_str(split_console_command[i], PROCESS_TAG, process);
                length = strlen(cmd);
                par[i] = (char *)malloc(sizeof(char)* (length + 1));
                if (par[i] == NULL) OutOfMemoryExit(5);
                memcpy(par[i], cmd, length);
                par[i][length] = '\0';
                free(cmd);
            }
            else if (strstr(split_console_command[i], CAPTION_TAG) != NULL)
            {
                //replace <caption> with caption or process
                if (caption != NULL)
                {
                    cmd = replace_str(split_console_command[i], CAPTION_TAG, caption);
                }
                else
                {
                    cmd = replace_str(split_console_command[i], CAPTION_TAG, process);
                }
                length = strlen(cmd);
                par[i] = (char *)malloc(sizeof(char)* (length + 1));
                if (par[i] == NULL) OutOfMemoryExit(5);
                memcpy(par[i], cmd, length);
                par[i][length] = '\0';
                free(cmd);
            }
            else
            {
                length = strlen(split_console_command[i]);
                par[i] = (char *)malloc(sizeof(char)* (length + 1));
                if (par[i] == NULL) OutOfMemoryExit(5);
                memcpy(par[i], split_console_command[i], length);
                par[i][length] = '\0';
            }
        }
        par[console_command_size] = NULL;

        c = sizeof(char)* (strlen(split_console_command[0]) + 1);
        cmd = (char *)malloc(c);
        if (cmd == NULL) OutOfMemoryExit(5);
        strcpy(cmd, split_console_command[0]);
        cmd[c - 1] = '\0';

        if (env_size > 0)
        {
            i = execve(cmd, par, env);
        }
        else
        {
            i = execvp(cmd, par);
        }

        if (i != 0)
        {
            i = errno;
            printf("Process %s did not start successfully (%d): %s.\n", process, errno, strerror(errno));
        }
        //on process termination free memory and exit this process

        if (env_size > 0)
        {
            for (i = 0; i < env_size + num_environment + 1; i++)
            {
                if (env[i] != NULL) free(env[i]);
            }
        }
        for (i = 0; i < console_command_size + 1; i++)
        {
            if (par[i] != NULL) free(par[i]);
        }
        free(cmd);
        exit(1);
    }
    else if (pid > 0) //parent process
    {
        wlog("Started process [%s]. Running under process id (%d).\n", "Process", cmd_line, pid);
        //in case that the process was started through a console, the "pid" is the PID of the console and the
        //process has a pid that is one higher (likely but not necessarily). The Watchdog will sync the pid with
        //this process later if it is wrong.
        return pid + 1;
    }
    else
    {
        wlog("Could not execute command [%s].\n", "Process", cmd_line);
        return -errno;
    }
}

int StopProcess(int process_id, boolean is_service, char *name, boolean do_kill)
{
    int status;
    int i;
    if (do_kill == 0)
    {
        i = kill(process_id, SIGTERM);
    }
    else
    {
        i = kill(process_id, SIGKILL);
    }

    if (i == 0)
    {
        i = waitpid(process_id, &status, 0);
        return 1;
    }
    return -errno;
}

int GetProcessID(char *process_name, int preferred_pid, char *path, char **parameters,
    unsigned int number_of_parameters, boolean is_service)
{
    DIR *dirp;
    FILE *fp;
    char *cmd;
    char **cmd_param;
    unsigned int num_cmd_param;
    struct dirent *entry;
    char process[MAX_PROC_PATH_SIZE], read_buf[16], status[2];
    int equal, eq;
    unsigned int i, j;
    int pid = -1;

    dirp = opendir("/proc/");
    if (dirp == NULL) return -1;

    while ((entry = readdir(dirp)) != NULL)
    {
        if (checkIfNumber(entry->d_name))
        {
            sprintf(process, "/proc/%s/stat", entry->d_name);
            fp = fopen(process, "r");
            if (fp == NULL) continue;
            //we are only interested into the second and third parameter, which are the process name in brackets and status
            fscanf(fp, "%*s (%s %s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s",
                read_buf, status);
            fclose(fp);
            //read_buf ends with ). Strip the name of this parenthesis
            read_buf[strlen(read_buf) - 1] = '\0';
            if (strncmp(read_buf, process_name, strlen(read_buf)) == 0)
            {
                //we found the process but it is a zombie. ignore it
                if (status[0] == 'Z') continue;
                //the name of the process is a match. Let's check the PID
                pid = strtol(entry->d_name, NULL, 10);
                //we found a process that matches the PID that we are looking for, this is always good enough,
                //because it means we've already done the path and parameters matching once
                if (preferred_pid == pid) break;
                //it is a new pid, let's check the command line

                if (path != NULL || number_of_parameters > 0)
                {
                    cmd = GetCommandLineForProcess(pid);
                    equal = 1;
                    if (cmd != NULL)
                    {
                        cmd_param = split(cmd, &num_cmd_param, ' ', 1, 1, 0);
                        if (path != NULL)
                        {
                            equal = _stricmp(cmd_param[0], path) == 0 ? 1 : 0;
                        }
                        if (number_of_parameters > 0)
                        {
                            equal = equal && (num_cmd_param == number_of_parameters + 1);
                            for (i = 0; i < number_of_parameters && equal; i++)
                            {
                                eq = 0;
                                for (j = 0; j < number_of_parameters && !eq; j++) {
                                    eq = (_stricmp(cmd_param[j+1], parameters[i]) == 0 ? 1 : 0);
                                }
                                equal = equal && eq;
                            }
                        }
                        free(cmd);
                        freeArray(cmd_param, num_cmd_param);
                        if (equal)
                            break; //we found the exact process we are looking for, so terminate and return the pid
                        else pid = -1; //we didn't find the process, so reset the pid and continue
                    }
                    else if (cmd != NULL)
                    {
                        free(cmd);
                    }
                }
                else
                {
                    break;
                }

            }
        }
    }

    closedir(dirp);
    return pid;
}

boolean IsProcessRunning(int pid)
{
    FILE *file;
    char process[MAX_PROC_PATH_SIZE];
    sprintf(process, "/proc/%d/stat", pid);
    file = fopen(process, "r");
    if (file == NULL) return FALSE;
    fclose(file);
    return TRUE;
}

int RestartComputer(void)
{
    sync();
    if (reboot(RB_AUTOBOOT) < 0)
    {
        int err = errno;
        wlog("Error rebooting machine (%d): %s.\n", "Watchdog", errno, strerror(errno));
        return err;
    }
    wlog("Computer will now reboot.\n", "Watchdog");
    return 0;
}

boolean network_not_initialized = TRUE;
unsigned int network_interface_count = 0;
char **network_interfaces = NULL;
int *network_interface_speed = NULL;
UINT64 *previous_total_bytes = NULL;
struct timespec last_network_measure;

int initNetwork()
{
    FILE *file, *speed_file;
    int i,j,length;
    int line_len = 3000;
    UINT64 received, sent;
    char name[64];
    char speed[256];
    char line[line_len];
    file = fopen("/proc/net/dev", "r");
    if (file == NULL) return 1;

    //first count the entries
    fgets(line, line_len - 1, file);
    fgets(line, line_len - 1, file);
    for (i = 0; 1; i++)
    {
        if (fgets(line, line_len - 1, file) == NULL)
            break;
    }
    fclose(file);
    network_interface_count = i;
    network_interfaces = (char**)malloc(sizeof(char*)*network_interface_count);
    if (network_interfaces == NULL) OutOfMemoryExit(2);
    network_interface_speed = (int*)malloc(sizeof(int)*network_interface_count);
    if (network_interface_speed == NULL) OutOfMemoryExit(2);
    previous_total_bytes = (UINT64*)malloc(sizeof(UINT64)*network_interface_count);
    if (previous_total_bytes == NULL) OutOfMemoryExit(2);
    file = fopen("/proc/net/dev", "r");
    if (file == NULL) return 1;
    fgets(line, line_len - 1, file);
    fgets(line, line_len - 1, file);
    for (i = 0; i < network_interface_count; i++)
    {
        fgets(line, line_len - 1, file);
        sscanf(line, " %[^:]s %*s", name);
        network_interfaces[i] = (char*)malloc(sizeof(char)*strlen(name)+1);
        if (network_interfaces[i] == NULL) OutOfMemoryExit(2);
        strcpy(network_interfaces[i],name);
        network_interfaces[i][strlen(name)] = '\0';

        length = strlen(line);
        for (j = strlen(name); j < length && line[j] != ':'; j++)
            ;
        sscanf(&line[j+1], " %"PRIu64" %*s %*s %*s %*s %*s %*s %*s %"PRIu64" %*s", &received, &sent);
        previous_total_bytes[i] = received + sent;

        sprintf(speed, "/sys/class/net/%s/speed", network_interfaces[i]);
        speed_file = fopen(speed, "r");
        //TODO what if there is no speed file, for example wireless interfaces, at the moment default to 1000 MB/s
        if (speed_file == NULL)
            network_interface_speed[i] = 1000;
        else
        {
            if (fgets(line, line_len-1, speed_file) != NULL)
                network_interface_speed[i] = strtol(line, NULL, 10);
            else
                network_interface_speed[i] = 1000;
            fclose(speed_file);
            speed_file = NULL;
        }
    }

    clock_gettime(CLOCK_MONOTONIC, &last_network_measure);
    fclose(file);
    return 0;
}

int GetNetworkInterfaceNames(char** names, unsigned int *size)
{
    int ans = 0;
    if (network_not_initialized)
    {
        ans = initNetwork();
        if (ans != 0)
            return ans;
        network_not_initialized = FALSE;
    }
    if (names == NULL || *size < network_interface_count)
    {
        *size = network_interface_count;
        return ERROR_BUFFER_TOO_SMALL;
    }
    *size = network_interface_count;
    for (ans = 0; ans < network_interface_count; ans++)
    {
        names[ans] = (char*)malloc(sizeof(char)*strlen(network_interfaces[ans])+1);
        strcpy(names[ans], network_interfaces[ans]);
        names[ans][strlen(network_interfaces[ans])] = '\0';
    }
    return 0;
}

int GetNetworkInterfaceFriendlyNames(char** names, unsigned int *size)
{
    return GetNetworkInterfaceNames(names, size);
}

int GetNetworkUtilization(float *percentage, float *absolute, unsigned int size)
{
    FILE *file;
    struct timespec now;
    float diff;
    int i,j,length;
    int line_len = 3000;
    UINT64 received, sent;
    char line[line_len];
    int ans = 0;
    if (network_not_initialized)
    {
        ans = initNetwork();
        if (ans != 0)
            return ans;
        network_not_initialized = FALSE;
    }
    if (percentage == NULL || absolute == NULL || size < network_interface_count)
        return ERROR_BUFFER_TOO_SMALL;

    file = fopen("/proc/net/dev", "r");
    if (file == NULL) return 1;
    fgets(line, line_len - 1, file);
    fgets(line, line_len - 1, file);
    clock_gettime(CLOCK_MONOTONIC, &now);
    diff = (float)(now.tv_sec - last_network_measure.tv_sec + (now.tv_nsec - last_network_measure.tv_nsec)/1e9);
    last_network_measure = now;
    for (i = 0; i < network_interface_count; i++)
    {
        fgets(line, line_len - 1, file);
        length = strlen(line);
        for (j = 0; j < length && line[j] != ':'; j++)
            ;
        sscanf(&line[j+1], " %"PRIu64" %*s %*s %*s %*s %*s %*s %*s %"PRIu64" %*s", &received, &sent);
        absolute[i] = (received + sent - previous_total_bytes[i])/(1024. * diff);
        percentage[i] = ((received + sent - previous_total_bytes[i])*800)/(diff*1000*1000*network_interface_speed[i]);

        previous_total_bytes[i] = received + sent;
    }

    fclose(file);
    return 0;
}

int NetworkCleanUp(void)
{
    if (network_interface_speed != NULL)
    {
        free(network_interface_speed);
        network_interface_speed = NULL;
    }
    if (network_interface_names != NULL)
    {
        freeArray(network_interface_names, network_interface_count);
        network_interface_names = NULL;
    }
    network_interface_count = 0;
    network_not_initialized = TRUE;

    return 0;
}

#elif defined(WIN32)

#include <windows.h>
#include <Winternl.h>
#include <Psapi.h>
#include <TlHelp32.h>
#include <Pdh.h>

#define PIDS_SIZE 1024

HANDLE *hProcess;
FILETIME *kernel_time, *user_time, wKernel_time, wUser_time;
MEMORYSTATUSEX mem_info;
time_t last_measure_time;
PDH_HQUERY cpu_query;
PDH_HCOUNTER cpu_total_h;
struct timespec cpu_measuring_start_time;

UINT64 subtractTime(const FILETIME a, const FILETIME b)
{
    LARGE_INTEGER la, lb;
    la.LowPart = a.dwLowDateTime;
    la.HighPart = a.dwHighDateTime;
    lb.LowPart = b.dwLowDateTime;
    lb.HighPart = b.dwHighDateTime;

    return la.QuadPart - lb.QuadPart;
}

void InitBuffers(unsigned int size)
{
    SYSTEM_INFO sys_info;
    hProcess = (HANDLE*)malloc(sizeof(HANDLE)*size);
    if (hProcess == NULL) OutOfMemoryExit(5);
    kernel_time = (FILETIME*)malloc(sizeof(FILETIME)*size);
    if (kernel_time == NULL) OutOfMemoryExit(5);
    user_time = (FILETIME*)malloc(sizeof(FILETIME)*size);
    if (user_time == NULL) OutOfMemoryExit(5);
    
    PdhOpenQuery(NULL, 0, &cpu_query);
    PdhAddCounter(cpu_query, L"\\Processor(_Total)\\% Processor Time", 0, &cpu_total_h);

    mem_info.dwLength = sizeof(MEMORYSTATUSEX);
    GetSystemInfo(&sys_info);
    num_processors = sys_info.dwNumberOfProcessors;
    last_measure_time = 0;
}

void FreeBuffers(void)
{
    if (hProcess) free(hProcess);
    hProcess = NULL;
    if (kernel_time) free(kernel_time);
    kernel_time = NULL;
    if (user_time) free(user_time);
    user_time = NULL;
    PdhCloseQuery(cpu_query);
}

UINT64 GetAvailableMemory(void)
{
    GlobalMemoryStatusEx(&mem_info);
    return mem_info.ullAvailPhys;
}

void GetWatchdogCPULoadAndMemory(int pid, int *cpu_load, UINT64 *memory)
{
    FILETIME wTime, local_kernel_time, local_user_time;
    ULONGLONG kernel_diff, user_diff;
    PROCESS_MEMORY_COUNTERS pmc;
    HANDLE wProcess;
    int measure_time;
    time_t now;

    now = time(0);
    wProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, pid);

    measure_time = (int)((now - last_measure_time) * 1000);
    if (measure_time == 0) measure_time = 1;
    GetProcessTimes(wProcess, &wTime, &wTime, &local_kernel_time, &local_user_time);
    last_measure_time = time(0);

    kernel_diff = subtractTime(local_kernel_time, wKernel_time);
    user_diff = subtractTime(local_user_time, wUser_time);
    *cpu_load = (int)(100 * (kernel_diff + user_diff) / (measure_time * 10000 * num_processors));

    wKernel_time = local_kernel_time;
    wUser_time = local_user_time;

    if (*cpu_load < 0) *cpu_load = 0;
    if (GetProcessMemoryInfo(wProcess, &pmc, sizeof(pmc)))
    {
        *memory = pmc.PagefileUsage;
    }
    else
    {
        *memory = 0;
    }

    CloseHandle(wProcess);
}

void GetCPULoadAndMemory(int *process_ids, unsigned int size, int *cpu_load, UINT64 *memory, int *cpu_total)
{
    FILETIME time;
    ULONGLONG kernel_diff, user_diff;
    PDH_FMT_COUNTERVALUE counter_val;
    PROCESS_MEMORY_COUNTERS pmc;
    FILETIME pKernel_time, pUser_time;
    struct timespec stop;
    
    int diff;
    unsigned int i;

    //get process handles
    for (i = 0; i < size; i++)
    {
        if (process_ids[i] < 0) continue;
        hProcess[i] = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, process_ids[i]);
    }

    //get start process times for all processes

    clock_gettime(CLOCK_MONOTONIC, &stop);
    diff = (int)((stop.tv_sec - cpu_measuring_start_time.tv_sec) * 1000.0 + (stop.tv_nsec - cpu_measuring_start_time.tv_nsec)/1e6);
    for (i = 0; i < size; i++)
    {
        if (process_ids[i] < 0) 
        {
            cpu_load[i] = 0;
            continue;
        }
        GetProcessTimes(hProcess[i], &time, &time, &pKernel_time, &pUser_time);

        kernel_diff = subtractTime(pKernel_time, kernel_time[i]);
        user_diff = subtractTime(pUser_time, user_time[i]);
        cpu_load[i] = (int)(100 * (kernel_diff + user_diff) / (diff * 10000 * num_processors));
        if (cpu_load[i] < 0) cpu_load[i] = 0; //the process is not running at all
        kernel_time[i] = pKernel_time;
        user_time[i] = pUser_time;
    }       
    cpu_measuring_start_time = stop;

    //read the total cpu
    PdhCollectQueryData(cpu_query);
    PdhGetFormattedCounterValue(cpu_total_h, PDH_FMT_DOUBLE, NULL, &counter_val);
    *cpu_total = (int)counter_val.doubleValue;
    
    //subtract process times for each process, close handles and record the usage
    for (i = 0; i < size; i++)
    {
        if (process_ids[i] < 0)
        {
            memory[i] = 0;
            continue;
        }
        if (GetProcessMemoryInfo(hProcess[i], &pmc, sizeof(pmc)))
            memory[i] = pmc.PagefileUsage;
        else
            memory[i] = 0;
        CloseHandle(hProcess[i]);
    }
}

int StartProcess(char *cmd_line, char **parameters, unsigned int par_size, char *working_directory,
    char *caption, char **environment, unsigned int env_size, boolean is_service, boolean show_console)
{
    char error[ERROR_MESSAGE_SIZE];
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    char *env, *scmd;
    char **del_env;
    wchar_t *cmd, *dir;
    int pid, fl, fe, i, err;
    unsigned int delsize;
    size_t l, length, c;
    SERVICE_STATUS_PROCESS ssi;
    SC_HANDLE scManager, sHandle;
    DWORD bytes_needed;

    if (is_service)
    {
        scManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
        if (scManager == NULL)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to access service manager: %s (%d)\n", "Process", cmd_line, error, err);
            return -err;
        }
        length = strlen(cmd_line) + 1;
        cmd = (wchar_t*)alloca(sizeof(wchar_t)* length);
        if (cmd == NULL) OutOfMemoryExit(5);
        mbstowcs_s(&l, cmd, length, cmd_line, _TRUNCATE);
        cmd[length - 1] = '\0';
        sHandle = OpenService(scManager, cmd, SC_MANAGER_ALL_ACCESS);

        if (sHandle == NULL)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to open the service handle for %s: %s (%d)\n", "Process", cmd_line, error, err);
            CloseServiceHandle(scManager);
            return -err;
        }

        if (StartService(sHandle, 0, NULL))
        {
            if (QueryServiceStatusEx(sHandle, SC_STATUS_PROCESS_INFO, (LPBYTE)&ssi, sizeof(SERVICE_STATUS_PROCESS), &bytes_needed))
            {
                pid = ssi.dwProcessId;
            }
            else
            {
                err = GetLastError();
                strerror_s(error, ERROR_MESSAGE_SIZE, err);
                wlog("Failed to query service info for %s: %s (%d)\n", "Process", cmd_line, error, err);
                pid = -err;
            }
        }
        else
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to start service %s: %s (%d)\n", "Process", cmd_line, error, err);
            pid = -err;
        }

        CloseServiceHandle(scManager);
        CloseServiceHandle(sHandle);

        return pid;
    }
    else
    {
        ZeroMemory(&si, sizeof(si));
        si.cb = sizeof(si);
        ZeroMemory(&pi, sizeof(pi));

        //compose the actual command line as command line + parameters
        fl = (int)strlen(cmd_line);
        if (par_size < 1)
        {
            //if no parameters, just copy the command line to the scmd
            scmd = (char*)alloca(sizeof(char)*(fl + 1));
            if (scmd == NULL) OutOfMemoryExit(5);
            memcpy(scmd, cmd_line, fl);
            scmd[fl] = '\0';
        }
        else
        {
            fe = 0;
            for (l = 0; l < par_size; l++)
            {
                fe += (int)strlen(parameters[l]) + 1;
            }
            scmd = (char*)alloca(sizeof(char)*(fl + fe + 1));
            if (scmd == NULL) OutOfMemoryExit(5);
            memcpy(scmd, cmd_line, fl);
            scmd[fl] = '\0';
            length = fl + 1;
            for (l = 0; l < par_size; l++)
            {
                memcpy(&scmd[length], parameters[l], strlen(parameters[l]));
                if (l < par_size - 1)
                {
                    length += strlen(parameters[l]) + 1;
                    scmd[length - 1] = ' ';
                }
            }

            for (i = 0; i < fl + fe; i++)
            {
                if (scmd[i] == '\0') scmd[i] = ' ';
            }

            scmd[fl + fe] = '\0';
        }

        l = 0;
        length = strlen(scmd) + 1;
        cmd = (wchar_t*)alloca(sizeof(wchar_t)* length);
        if (cmd == NULL) OutOfMemoryExit(5);
        mbstowcs_s(&l, cmd, length, scmd, _TRUNCATE);
        cmd[length - 1] = '\0';

        if (working_directory == NULL)
        {
            dir = NULL;
        }
        else
        {
            length = strlen(working_directory) + 1;
            dir = (wchar_t*)alloca(sizeof(wchar_t)* length);
            if (dir == NULL) OutOfMemoryExit(5);
            mbstowcs_s(&l, dir, length, working_directory, _TRUNCATE);
            dir[length - 1] = '\0';
        }

        if (env_size < 1)
        {
            //if no environment variables given, just use the default one
            env = NULL;
        }
        else
        {
            del_env = findDeletedVariables(environment, env_size, &delsize);

            //otherwise combine the parent environment and the given environment
            length = 0;// the total length of the environment variable string
            for (l = 0; l < env_size; l++)
            {
                //check if the variable is in the deleted ones. If yes (found = 1) than do not add it.
                if (isDeleted(environment[l], del_env, delsize) == 0)
                {
                    length += strlen(environment[l]) + 1; //each variable has to be NULL terminated
                }
            }
            for (l = 0; l < num_environment; l++)
            {
                //check if the variable is in the deleted ones. If yes (found = 1) than do not add it.
                if (isDeleted(parent_environment[l], del_env, delsize) == 0 &&
                    isDuplicate(parent_environment[l], environment, env_size) == 0)
                {
                    length += strlen(parent_environment[l]) + 1; //each variable has to be NULL terminated
                }
            }
            env = (char*)alloca(sizeof(char)*(length + 1));
            if (env == NULL) OutOfMemoryExit(5);
            env[length] = '\0'; //terminate the matrix
            c = 0;
            //copy the parent environment first and the specific environment second
            for (l = 0; l < num_environment; l++)
            {
                if (isDeleted(parent_environment[l], del_env, delsize) == 0 &&
                    isDuplicate(parent_environment[l], environment, env_size) == 0)
                {
                    length = strlen(parent_environment[l]);
                    memcpy(&env[c], parent_environment[l], length);
                    //strcpy_s(&env[c], full-c, parent_environment[l]);
                    c += length + 1;
                    env[c - 1] = '\0';
                }
            }
            for (l = 0; l < env_size; l++)
            {
                if (isDeleted(environment[l], del_env, delsize) == 0)
                {
                    length = strlen(environment[l]);
                    memcpy(&env[c], environment[l], length);
                    //strcpy_s(&env[c], full-c, environment[l]);
                    c += length + 1;
                    env[c - 1] = '\0';
                }
            }

            freeArray(del_env, delsize);
        }

        pid = -1;

        if (caption == NULL)
        {
            length = strlen(scmd) + 1;
            si.lpTitle = (wchar_t*)alloca(sizeof(wchar_t)*length);
            if (si.lpTitle == NULL) OutOfMemoryExit(5);
            mbstowcs_s(&l, si.lpTitle, length, scmd, _TRUNCATE);
            si.lpTitle[length - 1] = '\0';
        }
        else
        {
            length = strlen(caption) + 1;
            si.lpTitle = (wchar_t*)alloca(sizeof(wchar_t)*length);
            if (si.lpTitle == NULL) OutOfMemoryExit(5);
            mbstowcs_s(&l, si.lpTitle, length, caption, _TRUNCATE);
            si.lpTitle[length - 1] = '\0';
        }
        // si.wShowWindow = SW_SHOWNORMAL;
        wlog("Executing: %s\n", "Process", scmd);
        // Start the child process
        if (!CreateProcess(NULL, cmd, NULL, NULL, FALSE, show_console ? CREATE_NEW_CONSOLE : CREATE_NO_WINDOW, env, dir, &si, &pi))
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Could not start process [%s]. Error (%d): %s.\n", "Process", cmd_line, err, error);
            return -err;
        }
        pid = pi.dwProcessId;
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
        return pid;
    }
}

BOOL CALLBACK SendCloseCommand(HWND hwnd, LPARAM lParam)
{
    int *pid_to_close,r;
    DWORD pid;
    DWORD dwThreadId = GetWindowThreadProcessId(hwnd, &pid);
    pid_to_close = (int*)lParam;
    if (pid == pid_to_close[0])
    {
        r = PostMessage(hwnd, WM_CLOSE, 0, 0);
        if (r) SetLastError(ERROR_SUCCESS);
        return FALSE;
    }
    return TRUE;
}

int StopProcess(int process_id, boolean is_service, char *name, boolean do_kill)
{
    BOOL result;
    int err;
    size_t length, l;
    wchar_t *cmd;
    char error[ERROR_MESSAGE_SIZE];
    SC_HANDLE scManager, sHandle;
    SERVICE_STATUS sStatus;

    if (is_service)
    {
        scManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
        if (scManager == NULL)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to access service manager: %s (%d)\n", "Process", name, error, err);
            return -err;
        }
        length = strlen(name) + 1;
        cmd = (wchar_t*)alloca(sizeof(wchar_t)* length);
        if (cmd == NULL) OutOfMemoryExit(5);
        mbstowcs_s(&l, cmd, length, name, _TRUNCATE);
        cmd[length - 1] = '\0';
        sHandle = OpenService(scManager, cmd, SC_MANAGER_ALL_ACCESS);

        if (sHandle == NULL)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to open the service handle for %s: %s (%d)\n", "Process", name, error, err);
            CloseServiceHandle(scManager);
            return -err;
        }

        result = ControlService(sHandle, SERVICE_CONTROL_STOP, &sStatus);
        CloseServiceHandle(scManager);
        CloseServiceHandle(sHandle);
        if (!result)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Error stopping service %s. Error (%d): %s.\n", "Process", name, err, error);
            //we were not successful, will try to kill the process
        }
        else
        {
            return result;
        }
    }
    if (do_kill)
    {
        HANDLE hProcess = OpenProcess(PROCESS_TERMINATE, FALSE, process_id);
        if (hProcess == NULL)
            return FALSE;
        result = TerminateProcess(hProcess, 1);
        err = GetLastError();
        CloseHandle(hProcess);
        if (!result)
        {
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Error terminating process %d. Error (%d): %s.\n", "Process", process_id, err, error);
            return -err;
        }
    }
    else
    {
        SetLastError(ERROR_SUCCESS);
        result = EnumWindows((WNDENUMPROC)SendCloseCommand, (LPARAM)(&process_id));
        if (!result)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            return -err;
        }
    }
    return result;
}

char *GetCommandLineForProcess(int pid)
{
    PROCESS_BASIC_INFORMATION pbi;
    ULONG return_length;
    PPEB ppeb;
    PRTL_USER_PROCESS_PARAMETERS proc_parameters;
    USHORT cmd_len;
    PWSTR command_line;
    char *ret;
    int l, result;
    size_t s;
    NTSTATUS status;
    HANDLE hProcess = INVALID_HANDLE_VALUE;

    hProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, pid);

    if (hProcess == NULL || hProcess == INVALID_HANDLE_VALUE)
    {
        return NULL;
    }

    status = NtQueryInformationProcess(hProcess, ProcessBasicInformation, (PVOID)&pbi, sizeof(pbi), &return_length);

    ppeb = (PPEB)alloca(sizeof(PEB));
    if (ppeb == NULL) OutOfMemoryExit(5);
    result = ReadProcessMemory(hProcess, (PPEB)((PVOID*)&pbi)[1], ppeb, sizeof(PEB), NULL);

    if (result)
    {
        proc_parameters = (PRTL_USER_PROCESS_PARAMETERS)alloca(sizeof(RTL_USER_PROCESS_PARAMETERS));
        if (proc_parameters == NULL) OutOfMemoryExit(5);
        result = ReadProcessMemory(hProcess, ppeb->ProcessParameters, proc_parameters, sizeof(RTL_USER_PROCESS_PARAMETERS), NULL);
        cmd_len = proc_parameters->CommandLine.Length;
        command_line = (PWSTR)alloca(cmd_len);
        if (command_line == NULL) OutOfMemoryExit(5);
        result = ReadProcessMemory(hProcess, proc_parameters->CommandLine.Buffer, command_line, cmd_len, NULL);
    }

    CloseHandle(hProcess);
    if (result)
    {
        l = sizeof(char)*(cmd_len / 2 + 1);
        ret = (char*)malloc(l);
        if (ret == NULL) OutOfMemoryExit(5);
        ret[cmd_len / 2] = '\0';
        errno_t err = wcstombs_s(&s, ret, l, command_line, cmd_len / 2); //the size in bytes is twice the size of buffer
        if (err == EILSEQ)
        {
            // strange character in the process command line.
            for (s = 0; s < (size_t)(cmd_len / 2); s++)
            {
                if (command_line[s] > 255) //replace all too wide characters with ?
                {
                    command_line[s] = '?';
                }
            }
            wcstombs_s(&s, ret, l, command_line, cmd_len / 2);
        }
        if (strlen(ret) == 0)
        {
            free(ret);
            return NULL;
        }
        else
        {
            return ret;
        }
    }
    return NULL;
}

int GetProcessID(char *process_name, int preferred_pid, char *path, char **parameters, unsigned int number_of_parameters, boolean is_service)
{
    PROCESSENTRY32 entry;
    int pid, equal, eq, err;
    unsigned int num_cmd_param;
    unsigned int i, j;
    size_t l, length;
    wchar_t *pn;
    char *cmd, **cmd_param;
    char error[ERROR_MESSAGE_SIZE];
    SC_HANDLE scManager = INVALID_HANDLE_VALUE, sHandle = INVALID_HANDLE_VALUE;
    SERVICE_STATUS_PROCESS ssi;
    DWORD bytes_needed;
    HANDLE snapshot = INVALID_HANDLE_VALUE;
    entry.dwSize = sizeof(PROCESSENTRY32);

    l = 0;
    length = strlen(process_name) + 1;
    pn = (wchar_t*)alloca(sizeof(wchar_t)* (length));
    if (pn == NULL) OutOfMemoryExit(5);
    mbstowcs_s(&l, pn, length, process_name, _TRUNCATE);
    pn[length - 1] = '\0';

    if (is_service)
    {
        scManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
        if (scManager == NULL || scManager == INVALID_HANDLE_VALUE)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to access service manager: %s (%d)\n", "Process", process_name, error, err);
            return -err;
        }

        sHandle = OpenService(scManager, pn, SC_MANAGER_ALL_ACCESS);

        if (sHandle == NULL || sHandle == INVALID_HANDLE_VALUE)
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to open the service handle for %s: %s (%d)\n", "Process", process_name, error, err);
            CloseServiceHandle(scManager);
            return -err;
        }

        if (QueryServiceStatusEx(sHandle, SC_STATUS_PROCESS_INFO, (LPBYTE)&ssi, sizeof(SERVICE_STATUS_PROCESS), &bytes_needed))
        {
            pid = ssi.dwProcessId;
            if (pid == 0) pid = -1;
        }
        else
        {
            err = GetLastError();
            strerror_s(error, ERROR_MESSAGE_SIZE, err);
            wlog("Failed to query service info for %s: %s (%d)\n", "Process", process_name, error, err);
            pid = -err;
        }
        CloseServiceHandle(scManager);
        CloseServiceHandle(sHandle);
        return pid;
    }
    else
    {
        snapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, -1);
        if (snapshot == NULL || snapshot == INVALID_HANDLE_VALUE) return -1;

        pid = -1;
        if (Process32First(snapshot, &entry) == TRUE)
        {
            while (Process32Next(snapshot, &entry) == TRUE)
            {
                if (_wcsicmp(entry.szExeFile, pn) == 0)
                {
                    pid = entry.th32ProcessID;
                    //we found a process that matches the PID that we are looking for, this is always good enough, because it means
                    //we've already done the path and parameters matching once
                    if (preferred_pid == pid) break;

                    if (path != NULL || number_of_parameters > 0)
                    {
                        cmd = GetCommandLineForProcess(pid);
                        equal = 1;
                        if (cmd != NULL && strlen(cmd) > 0)
                        {
                            cmd_param = split(cmd, &num_cmd_param, ' ', TRUE, TRUE, FALSE);
                            if (path != NULL)
                            {
                                equal = _stricmp(cmd_param[0], path) == 0 ? 1 : 0;
                                //strstr(lower_cmd, lower_path) == NULL ? 0 : 1;
                            }

                            if (number_of_parameters > 0)
                            {
                                //the first parameter is the executable path
                                equal = equal & (num_cmd_param == number_of_parameters + 1);
                                for (i = 0; i < number_of_parameters && equal; i++)
                                {
                                    eq = 0;
                                    for (j = 0; j < number_of_parameters && !eq; j++) {
                                        eq = (_stricmp(cmd_param[j+1], parameters[i]) == 0 ? 1 : 0);
                                    }
                                    equal = equal && eq;
                                }
                            }

                            free(cmd);
                            freeArray(cmd_param,num_cmd_param);
                            if (equal) break; //we found the exact process we are looking for, so terminate and return the pid
                            else pid = -1;//we didn't find the process, so reset the pid and continue
                        }
                        else if (cmd != NULL)
                        {
                            free(cmd);
                        }
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        CloseHandle(snapshot);
        return pid;
    }
}

int GetChildrenOfProcessID(int parent_pid, int *pids, unsigned int size)
{
    PROCESSENTRY32 entry;
    HANDLE snapshot = INVALID_HANDLE_VALUE;
    entry.dwSize = sizeof(PROCESSENTRY32);
    unsigned int count = 0;

    snapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, -1);
    if (snapshot == NULL || snapshot == INVALID_HANDLE_VALUE) return -1;

    if (Process32First(snapshot, &entry) == TRUE)
    {
        while (Process32Next(snapshot, &entry) == TRUE)
        {
            if (entry.th32ParentProcessID == parent_pid)
            {
                if (count < size)
                {
                    pids[count] = entry.th32ProcessID;
                }
                count++;
            }
        }
    }
    CloseHandle(snapshot);
    return count;
}

boolean IsProcessRunning(int pid)
{
    DWORD pids[1024];
    DWORD ret;
    int i, c, a;
    i = (int)sizeof(DWORD);
    a = EnumProcesses(pids, i * PIDS_SIZE, &ret);
    c = (int)ret / i;
    for (i = 0; i < c; i++)
    {
        if (pids[i] == pid) return TRUE;
    }
    return FALSE;
}

int RestartComputer(void)
{
    char error[ERROR_MESSAGE_SIZE];
    int err = 0;
    HANDLE hToken;              // handle to process token
    TOKEN_PRIVILEGES tkp;// pointer to token structure

    BOOL fResult;// system shutdown flag

    // Get the current process token handle so we can get shutdown  privilege.

    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
    {
        err = GetLastError();
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        wlog("Could not reboot the computer. The token could not be opened. Error (%d): %s.\n", "Watchdog", err, error);
        return err;
    }

    // Get the LUID for shutdown privilege.

    LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);

    tkp.PrivilegeCount = 1;// one privilege to set
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    // Get shutdown privilege for this process.

    AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0);

    // Cannot test the return value of AdjustTokenPrivileges.

    err = GetLastError();
    if (err != ERROR_SUCCESS)
    {
        CloseHandle(hToken);
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        wlog("Could not reboot the computer. Token privilege adjustment failed. Error (%d): %s.\n", "Watchdog", err, error);
        return err;
    }

    // Display the shutdown dialog box and start the countdown.

    fResult = InitiateSystemShutdown(
        NULL,// shut down local computer
        L"Watchdog requested a system reboot.",// message for user
        30,// time-out period, in seconds
        FALSE,// ask user to close apps
        TRUE);// reboot after shutdown

    err = GetLastError();
    if (!fResult)
    {
        CloseHandle(hToken);
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        wlog("Could not reboot the computer. Error (%d): %s.\n", "Watchdog", err, error);
        return err;
    }
    wlog("Computer will now reboot.\n", "Watchdog", err, error);
    // Disable shutdown privilege.
    tkp.Privileges[0].Attributes = 0;
    fResult = AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0);
    err = GetLastError();
    CloseHandle(hToken);
    if (!fResult)
    {
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        wlog("Could not adjust the token privileges. Error (%d): %s.\n", "Watchdog", err, error);
        return err;
    }
    return 0;
}

int GetPid(void)
{
    DWORD pid = GetCurrentProcessId();
    return (int)pid;
}
#endif
