/*
* process.h
*
* Provides routines to manipulate the process. These routines are all OS dependent.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#ifndef PROCESS_H_
#define PROCESS_H_

#include "project.def"
#include "util.h"

/** The environment variables of the watchdog process. Depending on the
* implementation of the StartProcess routine, these environment variables
* might need to be prepended to the ones provided to the routine as parameters. */
extern char **parent_environment;
/** Number of environment variables */
extern unsigned int num_environment;
/**
* \brief Restarts this computer
*
* Triggers a graceful computer restart and lets the OS make a clean boot.
*
* \return 0 or nothing if the call was successful or the error code if it was not successful
*/
int RestartComputer(void);
/**
* \brief Quickly checks if the PID is running
*
* As quickly as possible checks if the process with the PID is running.
*
* \param pid the process id which is being checked
* \return 1 if the process is running or 0 otherwise
*/
boolean IsProcessRunning(int pid);
/**
* \brief Returns the pid of the process with the given name.
*
* Looks for a process that matches the given name and retrieves the
* process ID for the process if it was found. If the specified process
* is not running -1 is returned. If there are more than one process
* with the same name the pid of one of those processes is returned.
* If the preferred_pid is positive and more than one process with the
* given name exists, the method tries to find the process that matches
* the name and pid. If found the preferred_pid is returned, otherwise
* it returns the pid of one of the process with the given name
*
* \param process_name the name of the process to find (e.g. firefox.exe)
* \param preferred_pid the preferred pid of the process if multiple processes with
*			the same name exist
* \param path the path of the process that we are looking for. If there
*			are multiple processes with the same name, they could be
*			distinguished by different paths.
* \param parameters the parameters of the process. If there are multiple
*			processes with the same name, they could be distinguished
*			by the parameters
* \param number_of_parameters the number of parameters
* \param is_service a flag indicating if the process is a service (1) or not (0)
*
* \return the pid of the process if the process is running or -1 if not running
*/
int GetProcessID(char *process_name, int preferred_pid, char *path,
	char **parameters, unsigned int number_of_parameters, boolean is_service);
/**
* \brief Returns the command line of the process.
*
* Returns the command line used for starting the process identified by the given pid.
*
* \param pid the process id of the process that command line is requested
*
* \return the command line of the process
*/
char *GetCommandLineForProcess(int pid);
/**
* \brief Start a process with the given name and returns the pid of the new process.
*
* Start the process defined by the process name. The process name should be
* the absolute path to the executable. The process name (the running process)
* is then usually just the name of the executable. The directory in which
* the process should be started can also be specified as well as special
* environment settings for the process. If none are given, the default ones
* will be used (the ones that are used for the calling process).
*
* \param cmd_line if process is a service the name of the service, otherwise
*			the absolute path to the process to start
* \param parameters the array of command line parameters
* \param par_size number of parameters
* \param working_directory the absolute path to the folder from which to start the process
* \param caption the caption for the process (applicable for console processes only)
*			if null, the command line should be used for caption
* \param environment the environment variables to set for the started process, the process should also
*			inherit all the parent process variables, which are stored in the environment
* \param env_size number of given environment variables
* \param is_service a flag indicating if the process is a service (1) or not (0)
* \param show_console a flag indicating if the console of the process should be visible (1)
*           or it should not be displayed at all (0)            
*
* \return the process id if the process was started successfully or -error code otherwise
*/
int StartProcess(char *cmd_line, char **parameters, unsigned int par_size,
	char *working_directory, char *caption, char **environment, unsigned int env_size,
	boolean is_service, boolean show_console);

/**
* \brief Terminate the process running under the specified pid.
*
* Terminates the process that is running under the specified pid.
* The process will be terminated abruptly. If the process is a service
* the service should be stopped and only if that fails, the process should
* be stopped.
*
* \param process_id the pid of the process to terminate
* \param is_service a flag indicating if the process is a service (1) or not (0)
* \param the name of the process, required if the process is a service
* \param kill 1 if the kill signal should be sent or 0 if the terminate signal 
* 		 should be sent to the process
*
* \return 1 if the termination was successful or -error code if it failed
*/
int StopProcess(int process_id, boolean is_service, char *name, boolean kill);
/**
* \brief Get CPU usage percentage and memory consumption for several processes.
*
* Measure the CPU use percentage and memory consumption for specified processes. 
* The routine should measure kernel and user times over the specified measure time 
* period. The CPU percentage for each process is stored into the cpu_load. The 
* routine should measure the CPU usage for all process at once and not one by one.
* In addition the total cpu load on this computer should be measured and stored
* into cpu_load. This measurement is guaranteed to be called periodically, therefore
* the measurement from the previous call should be used in the current call to 
* estimate the CPU load.
* The memory consumption of all processes should be checked and stored into
* the memory.
*
* \param process_ids the array of pids of the process, for which the CPU usage should be measured
* \param size the number of processes
* \param cpu_load the pointer to the int array that holds the CPU usage for processes
* \param memory the pointer to the array that holds the memory consumption of processes
* \param total_cpu the total CPU load of this computer
*/
void GetCPULoadAndMemory(int *process_ids, unsigned int size, int *cpu_load, UINT64 *memory,
	int *total_cpu);
/**
* \brief Get the CPU usage and memory consumption of the watchdog process
*
* The routine measures the cpu usage and memory consumption of the watchdog process.
* The routine expects to be called periodically, which allows it to register
* kernel and user times on each call and calculate the average CPU percentage
* over each period.
*
* \param pid the watchdog process id
* \param cpu_load the pointer to store the watchdog cpu usage percentage 
* \param memory the pointer to store the watchdog memory consumption
*/
void GetWatchdogCPULoadAndMemory(int pid, int *cpu_load, UINT64 *memory);
/**
* \brief Reads the available physical memory
*
* Reads and returns the available physical memory on this machine.
*
* \return available physical memory in bytes
*/
UINT64 GetAvailableMemory(void);
/**
* \brief Initialize all the buffers that can be reused
*
* Allocates all the buffers that are used by other routines (mainly 
* GetCPULoadAndMemory) and which can be reused when repeating the calls. 
* Watchdog will call this routine once at startup. It is guaranteed that 
* the other routines will never request data for more processes than the 
* size parameter. Do not forget to call FreeBuffers once they are no
* longer needed.
*
* \param size the size with which the buffers should be allocated
*/
void InitBuffers(unsigned int size);
/**
* \brief Free the buffers allocated by InitBuffers
*
* Free all the buffers. Watchdog will call this routine once just before the exit.
*/
void FreeBuffers(void);
/**
* \brief Returns the PID of the watchdog process
*
* Returns the PID of the process that called this routine (watchdog server itself).
*/
int GetPid(void);
#if defined(WIN32)
/**
* \brief Returns the pids of all processes that are children of the parent_pid
*
* Scans the list of running processes and finds all processes that are direct
* children of the process defined by the parent_pid. The pids of those processes
* are filled into the pids array pointer. The number of found children is returned
* by this routine. If the provided array was too small for all found processes, the
* returned value still provides the number of found processes, but the pids array
* only contain the first "size" of pids.
*
* \param parent_pid the pid of the parent process, whose children we want to find
* \param pids the array that will contain the pids of all children upon completion
* \param size the size of the pids pointer (number of ints)
* \return the number of found children
*
*/
int GetChildrenOfProcessID(int parent_pid, int *pids, unsigned int size);
#endif

#endif