/*
* server.h
*
* Provides routines to handle the TINE related calls, such as open
* and close links to other servers or to start the watchdog TINE
* server.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#ifndef SERVER_H_
#define SERVER_H_

#include "watchdog.h"
/**
* \brief Start the server that offers watchdog data to the clients.
*
* The method should initialize and start the server that offers process information
* and statuses to the clients. The server should periodically call the watchdog's Watch
* method in the background. The Watch task does all the process measurements. The results
* and status can be obtained from the g_process defined in the watchdog.h.
* This method should also block for the duration of the server execution. It should
* not return unless the server was stopped, but note that when the server is stopped
* that will also terminate the watchdog.
*
* \param period the period to use for watch_task execution
* \param min_polling_rate the minimum client polling rate allowed
* \param info pointer to the server and fec information to be used by the server
*/
void StartServer(unsigned int period, unsigned int min_polling_rate, SERVER_INFO info);
/**
* \brief Stop the server
*
* Stop the server from doing any further actions.
*/
void StopServer(void);
/**
* \brief Starts monitoring the TINE property specified by the process
*
* Starts monitoring the property specified by the tine_device and tine_property.
* On successful establishment of the link the link id should be stored into the
* link_id field of the process. The link id is returned. If the link could not
* be established an negative error code may be returned.
* Updates from the property should be received by the provided callback method.
*
* \param process the pointer to the process for which the link should be created
* \param callback the callback method to receive updates from the property
*
* \return the link id if successful, or negative value if not successful
*/
int StartServerMonitor(WATCHED_PROCESS *process, void(*callback)(int, int));
/**
* \brief Closes the link to the property specified by the process
*
* Terminates the link to the remote property specified by the process.
* Upon successful termination 0 is returned, otherwise an error code
* can be returned.
*
* \param process the process for which to terminate the link
*
* \return 0 if termination was successful or an error code otherwise
*/
int CloseServerMonitor(WATCHED_PROCESS *process);

#endif