/*
* util.h
*
* This header is a utility header that provides API to the method
* calls which are suppose to be used on WIN32 architecture, but are
* not available in other (e.g. UNIX). These are mainly routines that
* are related to string and file manipulation.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#ifndef UTIL_H_
#define UTIL_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef LOG_FILE
#define LOG_FILE "serverlog"
#endif

#ifndef HOST_NAME_MAX
 #define HOST_NAME_MAX 64
#endif

#ifndef ERROR_MESSAGE_SIZE
 #define ERROR_MESSAGE_SIZE 100
#endif

typedef int boolean;

#ifdef TRUE
#undef TRUE
#endif

#ifndef TRUE
 #define TRUE 1
#endif

#ifndef FALSE
 #define FALSE 0
#endif

#define CHECK_ACCESS_READ(access) \
do {\
  if (!(access&CA_READ)) \
  {\
    LockMutex(FALSE);\
    return illegal_read_write;\
  }\
} while(0)

#define CHECK_ACCESS_WRITE(access) \
do {\
  if (!(access&CA_WRITE)) \
  {\
    LockMutex(FALSE);\
    return illegal_read_write;\
  }\
} while(0)

/* Debug flag for logging in the console (0 or 1) */
extern int debug;

enum AutoStart {
    NO,
    YES,
    ONCE
};

typedef struct
{
    char *server_name;
    char *fec_name;
    char *subsystem;
    char *context;
    char *description;
    char *location;
    char *hardware;
    char *responsible;
    int port_offset;
} SERVER_INFO;

#define SI_SERVER_NAME "SERVER"
#define SI_FEC_NAME "FEC"
#define SI_SUBSYSTEM "SUBSYSTEM"
#define SI_CONTEXT "CONTEXT"
#define SI_DESCRIPTION "DESCRIPTION"
#define SI_LOCATION "LOCATION"
#define SI_HARDWARE "HARDWARE"
#define SI_RESPONSIBLE "RESPONSIBLE"
#define SI_PORT_OFFSET "PORT_OFFSET"

#ifndef WIN32
#include "prolog.h"
#include <errno.h>
#define strcat_s(d,i,s) strcat(d,s)
#define strcpy_s(d,i,s) strcpy(d,s)
#define strncpy_s(d,i,s,l) strncpy(d,s,l)
#define sprintf_s(d,i,f,s...) sprintf(d,f,s)
#define fopen_s(f,n,c) *f = fopen(n,c)
#define vsnprintf_s(s,d,i,t,a) vsnprintf(s,i,t,a)
#define _stricmp(s,d) stricmp(s,d)
#define ctime_s(b,s,t) sprintf(b,"%s", ctime(t))
#define _snprintf_s(a,b,c,d,e...) snprintf(a,c,d,e)
#define _putenv(s) putenv(s)
#define _strupr_s(s,i) strupr(s)
#ifndef NOERROR
#define NOERROR 0
#endif
#ifndef ERROR_PATH_NOT_FOUND
#define ERROR_PATH_NOT_FOUND 30001L
#endif
#endif /* WIN32 */

/**
* Common exit routine for out of memory errors.
*/
#define OutOfMemoryExit(i) { fprintf(stderr, "Out of memory (%d).\n", i); exit(i); }

#define CheckNullReference(a,i) { if (a == NULL) OutOfMemoryExit(i) }

#ifdef WIN32
#define DEBUG(text,...) wlog(text,"DEBUG",__VA_ARGS__);
#else
#define DEBUG(text,...) wlog(text,"DEBUG",##__VA_ARGS__);
#endif

#ifdef WIN32
#include <time.h>
#define CLOCK_MONOTONIC 1
int clock_gettime(int, struct timespec *spec);
#endif

/**
* Check the parameter for spaces and tabs at the beginning and return
* the index of the first non empty character.
*
* \param words the character array to trim
* \return the index of the first non blank character
*/
int trim(char *words);
/**
* Transform the character array to an integer number.
*
* \param words the character array to parse
* \return the number represented by the parameter
*/
int toNumber(char *words);
/**
* Transform the character array to a floating point number.
*
* \param words the character array to parse
* \return the number represented by the parameter
*/
float toFloat(char* words);
/**
* Transform the character array to an integer representing a boolean value.
* The routine returns 1 if the arrays equals true or yes or any kind of
* a meaningful case format of these two words
*
* \param words the character array to parse
* \return 1 if the words represent true or yes; 0 if they don't
*/
boolean toBoolean(char *words);
/**
* Transform the character array to an integer representing the auto start value
* 0 = NO, 1 = YES, 2 = ONCE.
* 
* \param words the character array to parse
* \return 2 if words match ONCE, 1 if it matches true or yes or 0 othewise
*/
enum AutoStart toAutoStart(char *words);
/**
* Split the character array into several character arrays. It is expected that
* individual strings in the parameter are separated by the given c character.
* The routine returns a pointer to the matrix of characters and stores the
* number of different strings (matrix length) into the size parameter.
*
* \param words the character array to split
* \param size the size of the return value
* \param c the character on which to split
* \param trim_end a flag indicating if the empty spaces at the end of individual
*			parameter should be trimmed
* \param consider_quotes indicates if delimiters that are in quotes should be ignored
* \param escape_percent indicates if percent signs should be escaped with another percent sign
* \return a matrix that contains individual strings from the input parameter
*/
char **split(char *words, unsigned int *size, char c, boolean trim_end, boolean consider_quotes, boolean escape_percent);
/**
* \brief Converts the words array into a nice string
*
* Strips the words array from any empty characters at the beginning and end
* and copies the content into a new character array, which is returned.
*
* \param words the character array to parse
* \param escape_percent indicates if percent signs should be escaped with another percent sign
* \return a pointer to the stripped string
*/
char *toChar(char *words, boolean escape_percent);
/**
* \brief Logs the text into designated log output
*
* Logs the formatted text into the designated watchdog log file and
* into the standard output. The result is the same as calling wlog(1,1,text,...).
*
* \param text the unformatted text to print
* \param tag the tag for the watchdog log
* \param variadic parameters specify the values of macros in the text
*/
void wlog(char *text, char *tag, ...);
/**
* \brief Free the memory allocated by the given array of strings
*
* Free the memory by freeing all the elements in the given array and the array itself.
*
* \param arr the array to free
* \param size the size of the array
*/
void freeArray(char **arr, unsigned int size);
/**
* \brief Blocks the calling thread for the given amount of milliseconds.
*
* Block the calling thread for the given amount of milliseconds.
*
* \param milliseconds the time in milliseconds how long to block
*/
void Wait(unsigned int milliseconds);
/**
* \brief Returns the name of this computer
*
* Returns the name of the computer where this application is running. If the
* name contains the domain name, it is stripped from it.
*
* \return the name of the computer
*/
char *GetMachineName(void);
/**
* \brief Returns the working folder of current process.
* 
* \return the path of the working folder
*/
char* getWorkingFolder(void);
/**
* \brief Transforms a system code to an appropriate tine response code
* 
* Transfrom the code returned by a system call to an appropriate tine
* response code.
*
* \param system_code the system code
* \return tine response code
*/
int systemCodeToTineCode(int system_code);
/**
* \brief Release all fields in the info structure
*
* Release all internal fields of the info object. It is expected that fields have been at least initialized
* if not allocated with malloc.
*
* \param info the info to release
*/
void freeInfo(SERVER_INFO *info);
/**
* \brief Set up default values for the SERVER_INFO
*
* \param info the info structure to check and set up the values for
*/
void setUpDefaultInfo(SERVER_INFO *info, char *context, char *server_prefix, char *server, char *subsystem);
/**
* Reads the contents of a text file and returns it.
* 
* \param file_name the path to the file to read
* \param num_lines the pointer to an int that will contain the number lines in the returned matrix
* \return the array of lines (char*), where each line represents a single line from the file, comments omitted
*/
char** readFile(char *file_name, unsigned int *num_lines);
/**
* \brief Checks if the file exists
*
* Checks if the file exists and returns 1 if it does or 0 if it does not.
*
* \param file_name the absolute path to the file
* \return 0 if the file does not exist or 1 if it does
*/
boolean fileExists(char *file_name);
/**
* \brief Escape any occurence of the percent sign with another percent. The function always returns
* a new array, even if there was nothing to escape.
*
* \param text the text in which to escape te percent character
* \return newly allocated char array with percent characters escaped with another percent character
*/
char* escapePercent(char *text);
/**
* \brief Prepare or release the mutex used for guarding multithreaded access to variables. This
* function is called at startup and during shutdown and should take care of creating the mutex
* and initializing to be able to use it later at any time via the LockMutex function.
* \param prepare true if the mutex should be prepared or false if released
*/
void SetupMutex(boolean prepare);
/**
* \brief Lock or unlock the mutex. If the mutex is currently locked by some thread, the current
* thread is blocked until the lock owner releases the mutex. 
* \param lock true to lock or false to unlock
*/
void LockMutex(boolean lock);
/**
* \brief Same as strstr but case insensitive.
*
* \param str search this string for the occurrence of the token
* \param tok search for this string in the str
* \return a pointer to where the token was found or NULL if not found
*/
char *compareCaseInsensitive(const char *str, const char *tok);
#endif /* UTIL_H_ */
