/*
* utilities.c
*
* Implements routines from the util.h.
*
*  Created on: Sep 26, 2014
*      Author: Jaka Bobnar
*/
#include <string.h>
#include "util.h"
#include "tinetype.h"
#include "errors.h"

int debug = 0;

int systemCodeToTineCode(int system_code)
{
#if defined(WIN32) 
    switch (system_code)
    {
    case ERROR_SUCCESS: return operation_success;
    case ERROR_CURRENT_DIRECTORY:
    case ERROR_INVALID_DRIVE:
    case ERROR_FILE_NOT_FOUND:
    case ERROR_INVALID_NAME:
    case ERROR_BAD_PATHNAME:
    case ERROR_FILENAME_EXCED_RANGE:
    case ERROR_DIRECTORY:
    case ERROR_SERVICE_DOES_NOT_EXIST:
    case ERROR_PATH_NOT_FOUND: return no_such_file;
    case ERROR_WRITE_PROTECT:
    case ERROR_DRIVE_LOCKED:
    case ERROR_SHUTDOWN_IN_PROGRESS:
    case ERROR_SERVICE_DATABASE_LOCKED:
    case ERROR_SERVICE_LOGON_FAILED:
    case ERROR_ACCESS_DENIED: return access_denied;
    case ERROR_OUTOFMEMORY:
    case ERROR_NOT_ENOUGH_MEMORY: return out_of_server_memory;
    case ERROR_BAD_FORMAT: return illegal_format;
    case ERROR_INVALID_PARAMETER:
    case ERROR_BAD_ARGUMENTS:
    case ERROR_INVALID_DATA: return invalid_data;
    case ERROR_ILLEGAL_CHARACTER:
    case ERROR_INVALID_HANDLE:
    case ERROR_INSUFFICIENT_BUFFER:
    case ERROR_INVALID_LEVEL:
    case ERROR_SERVICE_ALREADY_RUNNING:
    case ERROR_SERVICE_DEPENDENCY_DELETED:
    case ERROR_SERVICE_DEPENDENCY_FAIL:
    case ERROR_SERVICE_DISABLED:
    case ERROR_SERVICE_MARKED_FOR_DELETE:
    case ERROR_SERVICE_NO_THREAD:
    case ERROR_BAD_COMMAND: return illegal_command;
    case ERROR_BAD_LENGTH: return illegal_data_size;
    case ERROR_DISK_FULL:
    case ERROR_HANDLE_DISK_FULL: return no_file_space;
    case ERROR_BAD_EXE_FORMAT: return illegal_device;
    case ERROR_SERVICE_REQUEST_TIMEOUT: return link_timeout;
    case ERROR_IO_PENDING: return device_busy;
    }
    return system_code;
#elif defined(UNIX)
    switch (system_code)
    {
    case ERROR_PATH_NOT_FOUND:
    case ENOENT: return no_such_file;
    case EAGAIN:
    case ENOMEM: return out_of_server_memory;
    case ENOSYS: return not_supported;
    case EINVAL: return system_error;
    case EPERM: return unauthorised_action;
    case ESRCH: return device_offline;
    }
    return system_code;
#endif
}

void freeArray(char **arr, unsigned int size)
{
    unsigned int i;
    if (size > 0 && arr != NULL)
    {
        for (i = 0; i < size; i++)
        {
            if (arr[i] != NULL) free(arr[i]);
        }
        free(arr);
    }
}

int trim(char *words)
{
    int j, k;
    int l = (int)strlen(words);
    for (j = 0; j < l && (words[j] == ' ' || words[j] == '\t'); j++)
        ;
    for (k = l - 1; k > j && (words[k] == ' ' || words[k] == '\t' || words[k] == '\n' || words[k] == '\r'); k--)
        ;
    if (k < l - 1)
    {
        words[k + 1] = '\0';
    }
    return j;
}

int trimAndUnQuote(char *words)
{
    int j, k;
    int l = (int)strlen(words);
    for (j = 0; j < l && (words[j] == ' ' || words[j] == '\t' || words[j] == '"'); j++)
        ;
    for (k = l - 1; k > j && (words[k] == ' ' || words[k] == '\t' || words[j] == '"' || words[k] == '\n' || words[k] == '\r'); k--)
        ;
    if (k < l - 1)
    {
        words[k + 1] = '\0';
    }
    return j;
}

int toNumber(char *words)
{
    errno = NOERROR;
    char* temp;
    int start, length;

    start = trim(words);
    length = (int)strlen(words) - start;
    temp = (char*)alloca(sizeof(char)* (length + 1));
    if (temp == NULL) OutOfMemoryExit(3);

    temp[length] = '\0';
    memcpy(temp, &words[start], length);
    char* err;
    int a = strtol(temp, &err, 10);
    if (*err == '\0')
    {
        return a;
    }
    else
    {
        errno = EINVAL;
        return 0;
    }
}

float toFloat(char* words)
{
    errno = NOERROR;
    char* temp;
    int start, length;

    start = trim(words);
    length = (int)strlen(words) - start;
    temp = (char*)alloca(sizeof(char) * (length + 1));
    if (temp == NULL) OutOfMemoryExit(3);

    temp[length] = '\0';
    memcpy(temp, &words[start], length);
    char* err;
    float a = strtof(temp, &err);
    if (*err == '\0')
    {
        return a;
    }
    else
    {
        errno = EINVAL;
        return 0;
    }
}

boolean toBoolean(char *words)
{
    char* temp;
    int start, length;

    start = trim(words);
    length = (int)strlen(words) - start;
    temp = (char*)alloca(sizeof(char)* (length + 1));
    if (temp == NULL) OutOfMemoryExit(3);

    temp[length] = '\0';
    memcpy(temp, &words[start], length);

    return (strncmp(temp, "true", 4) == 0
        || strncmp(temp, "TRUE", 4) == 0
        || strncmp(temp, "True", 4) == 0
        || strncmp(temp, "yes", 3) == 0
        || strncmp(temp, "YES", 3) == 0
        || strncmp(temp, "Yes", 3) == 0) ? TRUE : FALSE;
}

enum AutoStart toAutoStart(char *words)
{
    char* temp;
    int start, length;

    start = trim(words);
    length = (int)strlen(words) - start;
    temp = (char*)alloca(sizeof(char)* (length + 1));
    if (temp == NULL) OutOfMemoryExit(3);

    temp[length] = '\0';
    memcpy(temp, &words[start], length);
    
    if (strncmp(temp, "once", 4) == 0 
        || strncmp(temp, "ONCE", 4) == 0
        || strncmp(temp, "Once", 4) == 0) 
    {
        return ONCE;
    }
    return (strncmp(temp, "true", 4) == 0
        || strncmp(temp, "TRUE", 4) == 0
        || strncmp(temp, "True", 4) == 0
        || strncmp(temp, "yes", 3) == 0
        || strncmp(temp, "YES", 3) == 0
        || strncmp(temp, "Yes", 3) == 0) ? YES : NO;
}

char *toChar(char *words, boolean escape_percent)
{
    char *temp, *temp2;
    int start, length;
    start = trim(words);
    length = (int)strlen(words) - start;
    temp = (char*)malloc(sizeof(char)* (length + 1));
    if (temp == NULL) OutOfMemoryExit(3);
    temp[length] = '\0';
    memcpy(temp, &words[start], length);
    temp2 = escapePercent(temp);
    free(temp);
    return temp2;
}

char **split(char *wordsIn, unsigned int *size, char c, boolean trim_end, boolean consider_quotes, boolean escape_percent)
{
    char** ret;
    unsigned j, i, n, start, length;
    int k;
    boolean quoted;
    char *words;

    words = escape_percent ? escapePercent(wordsIn) : wordsIn;
    start = trim(words);
    length = (int)strlen(words);
    if (length - start <= 0)
    {
        *size = 0;
        return 0;
    }
    if (trim_end)
    {
        for (j = length - 1; j >= start && (words[j] == ' ' || words[j] == '\t' || words[j] == '\r' || words[j] == '\n' || words[j] == c); j--)
            ;
        length = j + 1;
    }
    n = 1;
  
    if (consider_quotes)
    {
        quoted = FALSE;
        for (j = start; j < length - 1; j++)
        {
            if (words[j] == '"')
            {
                quoted = quoted ? FALSE : TRUE;
            }
            if (!quoted && words[j] == c && (j == 0 || words[j - 1] != c))
            {
                n++;
            }
        }
    } 
    else
    {
        for (j = start; j < length - 1; j++)
        {
            if (words[j] == c && (j == 0 || words[j - 1] != c)) n++;
        }
    }
    *size = n;
    ret = (char **)malloc(sizeof(char*)* *size);
    if (ret == NULL) OutOfMemoryExit(3);

    for (j = 0; j < *size; j++)
    {
        i = start;
        //check for half quotes
        quoted = FALSE;
        if (consider_quotes)
        {
            for (; i < length; i++)
            {
                if (words[i] == '"') 
                {
                    quoted = quoted ? FALSE : TRUE;
                }
                if (!quoted && words[i] == c)
                {
                    break;
                }
            }
        } 
        else
        {
            for (; i < length && words[i] != c; i++)
                ;
        }

        if (i == length && words[i] != c)
        {
            i++;
        }
        ret[j] = (char*)malloc(sizeof(char)* (i - start + 1));
        if (ret[j] == NULL) OutOfMemoryExit(3);

        start += (consider_quotes ? trim(&words[start]) : trimAndUnQuote(&words[start]));
        ret[j][(i - start)] = '\0';
        memcpy(ret[j], &words[start], (i - start));
        if (words[i] == c) i++;
        start = i + (consider_quotes ? trim(&words[start]) : trimAndUnQuote(&words[start]));

        for (; start < length && (words[start] == c); start++)
            ;

        if (trim_end)
        {
            /* Get rid of CR or LF at end of line */
            for (k = (int)strlen(ret[j]) - 1; k >= 0 && (ret[j][k] == '\n' || ret[j][k] == '\r' || ret[j][k] == '\0' || ret[j][k] == ' ' || ret[j][k] == '\t'
                || (!consider_quotes && ret[j][k] == '"')); k--)
                ;
            k++;
            ret[j][k] = '\0';
        }
    }
    
    if (escape_percent)
    {
        free(words);
    }
    return ret;
}

char* escapePercent(char *text)
{
    char* newText;
    size_t length, bufferSize, i, j;
    length = strlen(text);
    int count;

    count = 0;
    for (i = 0; i < length; i++)
    {
        if (text[i] == '%') 
        {
            count++;
        }
    }
    bufferSize = length + count + 1;
    newText = (char*)malloc(bufferSize * sizeof(char));
    if (newText == NULL) OutOfMemoryExit(3);
    for (i = 0, j = 0; i < length && j < bufferSize; i++)
    {
        newText[j++] = text[i];
        if (text[i] == '%')
        {
            newText[j++] = '%';
        } 
    }
    newText[j] = '\0';
    if (count > 0)
    {
        printf("WARNING: configuration contains macro, which will not be resolved: %s\n", text);
    }
    return newText;
}

void wlog(char *text, char *tag, ...)
{
    va_list args;
    char buf[512];
    va_start(args, tag);
    vsnprintf_s(buf, 512, 511, text, args);
    va_end(args);
    if (debug)
    {
        printf("%s>", buf);
    }
    srvlog(LOG_FILE, tag, buf);
}

void freeInfo(SERVER_INFO *info)
{
    if (info->context != NULL) free(info->context);
    info->context = NULL;
    if (info->server_name != NULL) free(info->server_name);
    info->server_name = NULL;
    if (info->description != NULL) free(info->description);
    info->description = NULL;
    if (info->fec_name != NULL) free(info->fec_name);
    info->fec_name = NULL;
    if (info->hardware != NULL) free(info->hardware);
    info->hardware = NULL;
    if (info->location != NULL) free(info->location);
    info->location = NULL;
    if (info->responsible != NULL) free(info->responsible);
    info->responsible = NULL;
    if (info->subsystem != NULL) free(info->subsystem);
    info->subsystem = NULL;
}

void setUpDefaultInfo(SERVER_INFO *info, char *context, char *server_prefix, char *server, char *subsystem)
{
    char buf[10];
    int length, size;
    //set up the default server name to be WD_<COMP_NAME>, context to SERVICE, port to 20, fec to <COMP_NAME>.port
    if (info->context == NULL)
    {
        length = (int)strlen(context)+1;
        size = sizeof(char)*length;
        info->context = (char*)malloc(size);
        if (info->context == NULL) OutOfMemoryExit(3);
        strcpy_s(info->context, size, context);
        info->context[length - 1] = '\0';
    }
    if (info->server_name == NULL)
    {
        length = (int)strlen(server) + (int)strlen(server_prefix) + 1;
        size = sizeof(char)*length;
        info->server_name = (char*)malloc(size);
        if (info->server_name == NULL) OutOfMemoryExit(3);
        strcpy_s(info->server_name, size, server_prefix);
        strcat_s(info->server_name, size, server);
        info->server_name[length - 1] = '\0';
    }
    if (info->port_offset == -1)
    {
        info->port_offset = 20;
    }
    if (info->fec_name == NULL)
    {
        sprintf_s(buf, 10, ".%d", info->port_offset);
        length = (int)strlen(server) + (int)strlen(buf) + 1;
        size = sizeof(char)*length;
        info->fec_name = (char*)malloc(size);
        if (info->fec_name == NULL) OutOfMemoryExit(3);
        strcpy_s(info->fec_name, size, server);
        strcat_s(info->fec_name, size, buf);
        info->fec_name[length - 1] = '\0';
    }
    if (info->subsystem == NULL)
    {
        length = (int)strlen(subsystem) + 1;
        size = sizeof(char)*length;
        info->subsystem = (char*)malloc(size);
        if (info->subsystem == NULL) OutOfMemoryExit(3);
        strcpy_s(info->subsystem, size, subsystem);
        info->subsystem[length - 1] = '\0';
    }
    if (info->hardware == NULL)
    {
        length = 5;
        size = sizeof(char)*length;
        info->hardware = (char*)malloc(size);
        if (info->hardware == NULL) OutOfMemoryExit(3);
        strcpy_s(info->hardware, size, "None");
        info->hardware[length - 1] = '\0';
    }
    if (info->location == NULL)
    {
        length = 1;
        size = sizeof(char)*length;
        info->location = (char*)malloc(size);
        if (info->location == NULL) OutOfMemoryExit(3);
        info->location[length - 1] = '\0';
    }
    if (info->responsible == NULL)
    {
        length = 6;
        size = sizeof(char)*length;
        info->responsible = (char*)malloc(size);
        if (info->responsible == NULL) OutOfMemoryExit(3);
        strcpy_s(info->responsible, size, "Admin");
        info->responsible[length - 1] = '\0';
    }

    if (info->description == NULL)
    {
        info->description = '\0';
    }
}

char** readFile(char *file_name, unsigned int *size)
{
    char **words;
    FILE *fp;
    int i, j, lines_allocated, max_line_len;

    lines_allocated = 128;
    max_line_len = 3000;

    fopen_s(&fp, file_name, "r");
    if (fp == NULL)
    {
        printf("Error opening file [%s].\n", file_name);
        wlog("Error opening file [%s].\n", "Setup", file_name);
        exit(3);
    }

    /* Allocate lines of text */
    words = (char **)malloc(sizeof(char*)*lines_allocated);
    if (words == NULL) OutOfMemoryExit(3);

    for (i = 0; 1; i++)
    {
        /* Have we gone over our line allocation? */
        if (i >= lines_allocated)
        {
            int new_size;
            /* Double our allocation and re-allocate */
            new_size = lines_allocated * 2;
            words = (char **)realloc(words, sizeof(char*)*new_size);
            if (words == NULL) OutOfMemoryExit(3);
            lines_allocated = new_size;
        }
        /* Allocate space for the next line */
        words[i] = (char*)malloc(max_line_len);
        if (words[i] == NULL) OutOfMemoryExit(3);

        if (fgets(words[i], max_line_len - 1, fp) == NULL)
        {
            free(words[i]);
            break;
        }

        /* Get rid of CR or LF at end of line */
        for (j = (int)strlen(words[i]) - 1; j >= 0 && (words[i][j] == '\n' || words[i][j] == '\r'); j--)
            ;
        j++;
        words[i][j] = '\0';
    }
    fclose(fp);
    *size = i;
    return words;
}

boolean fileExists(char *file_name)
{
    FILE *file;
    fopen_s(&file, file_name, "r");
    if (file == NULL) return FALSE;
    fclose(file);
    return TRUE;
}

#define CI_SIZE 4096
char *compareCaseInsensitive(const char *str, const char *tok)
{
    char s1[CI_SIZE], s2[CI_SIZE], *c;
    size_t len;
    if (str == NULL || tok == NULL) return NULL;
    strncpy_s(s1, CI_SIZE, str, CI_SIZE - 1);
    strncpy_s(s2, CI_SIZE, tok, CI_SIZE - 1);
    s1[CI_SIZE - 1] = s2[CI_SIZE - 1] = 0;
    _strupr_s(s1, CI_SIZE - 1);
    _strupr_s(s2, CI_SIZE - 1);
    if ((c = strstr(s1, s2)) == NULL) return NULL;
    len = (size_t)c - (size_t)s1;
    return (char *)(str + len);
}

#if defined(UNIX)

#include <pthread.h>
#include <unistd.h>
pthread_mutex_t mutex;

void Wait(unsigned int milliseconds)
{
    usleep(milliseconds * 1000);
}

char* getWorkingFolder(void)
{
    char cwd[1024];
    int length = 1024;
    char* ret;

    if (getcwd(cwd, sizeof(cwd)) != NULL)
    {
        length = strlen(cwd);
        ret = malloc(sizeof(char) * (length + 1));
        if (ret == NULL) OutOfMemoryExit(2);
        ret[length] = '\0';
        memcpy(ret, cwd, length);
    }

    return ret;
}

char *GetMachineName(void)
{
    char name[HOST_NAME_MAX];
    int length = HOST_NAME_MAX;
    char domainname[HOST_NAME_MAX];
    int domainlength = HOST_NAME_MAX;
    FILE *fp;
    char *ret;

    if (gethostname(name, length) == 0)
    {
        length = strlen(name);
        //read the dns domain name and strip the hostname of the domain name if the
        //hostname is set like that
        fp = popen("dnsdomainname", "r");
        if (fp == NULL)
        {
            wlog("Could not read the domain name.", "Setup", SI_SERVER_NAME);
        }
        else
        {
            if (fgets(domainname, HOST_NAME_MAX - 1, fp) == NULL)
            {
                wlog("Could not read the domain name.", "Setup", SI_SERVER_NAME);
            }
            else
            {
                trim(domainname);
                domainlength = strlen(domainname);
                if (length > domainlength) {
                    if (strcmp(name + length - domainlength, domainname) == 0)
                    {
                        length = length - domainlength - 1;
                    }
                }
            }
            fclose(fp);
        }
        ret = malloc(sizeof(char)* (length + 1));
        if (ret == NULL) OutOfMemoryExit(2);
        ret[length] = '\0';
        memcpy(ret, name, length);
    }
    else
    {
        wlog(
            "Could not read the computer name. Please define the server name (%s) in the configuration file.\n",
            "Setup", SI_SERVER_NAME);
        ret = malloc(sizeof(char) * 8);
        if (ret == NULL) OutOfMemoryExit(2);
        ret[7] = '\0';
        memcpy(ret, "UNKNOWN", 7);
    }

    return ret;
}

void SetupMutex(boolean prepare)
{
    if (prepare)
    {
    	if (pthread_mutex_init(&mutex, NULL) != 0)
		{
			wlog("Could not create mutex.\n", "Setup", SI_SERVER_NAME);
			OutOfMemoryExit(2);
		}
    }
    else
    {
    	pthread_mutex_destroy(&mutex);
    }
}

void LockMutex(boolean lock)
{
    if (lock)
    {
    	pthread_mutex_lock(&mutex);
    }
    else
    {
    	pthread_mutex_unlock(&mutex);
    }
}

#elif defined(WIN32)

#include <Windows.h>
HANDLE ghMutex;

void Wait(unsigned int milliseconds)
{
    Sleep(milliseconds);
}

char *getWorkingFolder(void)
{
    char* ret;
    char error[ERROR_MESSAGE_SIZE];
    size_t i, l;
    wchar_t folder[1024];
    DWORD length = 1024;
    length = GetCurrentDirectory(length, folder);
    if (length == 0)
    {
        int err = GetLastError();
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        wlog("Could not read the current working folder. Error (%d): %s.\n", "Setup", err, error);
        return NULL;
    }
    l = sizeof(char) * (length + 1);
    ret = (char*)malloc(l);
    if (ret == NULL) OutOfMemoryExit(5);
    ret[length] = '\0';
    wcstombs_s(&i, ret, l, folder, l);
    return ret;
}

char *GetMachineName(void)
{
    char *ret;
    char error[ERROR_MESSAGE_SIZE];
    size_t i, l;
    wchar_t name[HOST_NAME_MAX];
    DWORD length = HOST_NAME_MAX;
    if (!GetComputerName(name, &length))
    {
        int err = GetLastError();
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        wlog("Could not read the machine name. Error (%d): %s.\n", "Setup", err, error);
        return NULL;
    }
    l = sizeof(char)* (length + 1);
    ret = (char *)malloc(l);
    if (ret == NULL) OutOfMemoryExit(5);
    ret[length] = '\0';
    wcstombs_s(&i, ret, l, name, l);
    return ret;
}

void SetupMutex(boolean prepare)
{
    if (prepare)
    {
        ghMutex = CreateMutex(NULL, FALSE, NULL);
        if (ghMutex == NULL)
        {
            OutOfMemoryExit(1);
        }
    } 
    else
    {
        CloseHandle(ghMutex);
        ghMutex = NULL;
    }
}

void LockMutex(boolean lock)
{
    if (lock)
    {
        WaitForSingleObject(ghMutex, INFINITE);
    }
    else
    {
        ReleaseMutex(ghMutex);
    }
}

int clock_gettime(int clck, struct timespec *spec)
{  
    __int64 wintime;
    GetSystemTimeAsFileTime((FILETIME*)&wintime);
    wintime      -= 116444736000000000i64;  //1jan1601 to 1jan1970
    spec->tv_sec  = wintime / 10000000i64;           //seconds
    spec->tv_nsec = wintime % 10000000i64 *100;      //nano-seconds
    return 0;
}

#endif
