/*
* watchdog.c
*
* Implements routines from watchdog.h
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
//we are linking against the static network library
#define NETSTATIC 
#define USE_NETWORK 1

//#include <vld.h>

#include <math.h>
#include "tine.h"
#include "watchdog.h"
#include "console.h"
#include "process.h"
#include "parser.h"
#include "server.h"
#include "util.h"
#include "network.h"
#include <signal.h>

//environment variables used to start watchdog
char** parent_environment = NULL;
unsigned int num_environment = 0;

//global storage of processes
WATCHED_PROCESS *g_process = NULL;
unsigned int num_processes = 0;
//server information
SERVER_INFO info = { 0 };
//flag indicating if watchdog is running or not
volatile boolean running = TRUE;
boolean g_network_enabled = TRUE;

//flag if the watchdog is suspended or no
boolean suspended = FALSE;
time_t made_suspended;
int remaining_suspended_time = 0;
//watchdog dynamic properties
int total_cpu_usage = 0;
UINT64 available_memory = 0;
int watchdog_cpu_usage = 0;
UINT64 watchdog_memory_consumption = 0;
int watchdog_pid;
char **network_interface_names = NULL;
char **network_interface_friendly_names = NULL;
unsigned int num_network_interfaces = 0;
float *network_utilization_percentage = NULL;
float *network_utilization_absolute = NULL;

// Needed for thread safety
float *network_percentage = NULL;
float *network_absolute = NULL;

//local storage to avoid allocation of memory in every cycle
int *cpu_load = NULL, *process_ids = NULL;
UINT64 *memory = NULL;
time_t last_process_start;
int time_reported = 0;
time_t now;
char *config_file = NULL;
//by default do not stop anything on shutdown
boolean stop_all_on_shutdown = DEFAULT_STOP_ALL_ON_SHUTDOWN;
time_t startup_time;

/**
* Check if the given process is running. If it is running, it is set to not disabled and
* the pid of the process is stored. If it is not running, is_running flag is set to false,
* and other fields that describe the process state are also set appropriately
* (timestamp when determined not running etc.).
*
* \param process the process to check and updated
*/
void checkProcessRunning(WATCHED_PROCESS *process, int only_check)
{
    int pid;
    if (process->auto_start == ONCE)
    {
        return;
    }

    pid = process->pid;
    //if the pid is unknown or there's been too much time since we last checked if the PID still matches
    //our process, refetch the pid from the system.
    //otherwise just quickly check if the pid is still running.
    if (pid < 0 || (now - process->pid_last_refreshed > pid_refresh_rate))
    {
        pid = GetProcessID(process->name,
                           process->pid,
                           process->match_path ? process->path_to_executable : NULL, 
                           process->individual_parameters, 
                           process->match_parameters ? process->number_of_parameters : 0,
                           process->is_service);
        process->pid_last_refreshed = time(0);
    }
    else if (!IsProcessRunning(process->pid))
    {
        process->process_requested_to_stop = FALSE;
        pid = -1;
        process->pid = -1;
    }

    if (pid < 0)
    {
        if (process->is_running)
        {
            process->is_running = FALSE;
            process->determined_not_running_timestamp = time(0);
            //only reset the counter if restart is not guided by the TINE property
            if (process->timeout_restart <= 0 || process->property_disabled)
                process->current_retry = 0;
            process->pid = -1;
        }
    }
    
    if (process->is_running)
    {
        process->stopped_because_of_cpu = FALSE;
        process->stopped_because_of_memory = FALSE;
        process->stopped_because_of_link = FALSE;
    }

    if (only_check) return;
    if (process->process_requested_to_stop > 0)
    {
        //if process_requested_stop is not 0, the process was requested to stop
        //check if its pid is < 0. If yes, the process is stopped, if not, the process
        //is not yet stopped and it should be killed if enough time has passed (5 seconds)
        if (pid < 0)
        {
            process->process_requested_to_stop = 0;
        }
        else
        {
            if (now - process->process_requested_to_stop > kill_grace_period)
            {
                wlog("Process [%s] was requested to stop, but failed to do it gracefully. It will be killed.\n", process->alias, process->alias);
                StopWatchedProcess(process, FALSE, TRUE);
            }
        }
        return;
    }

    if (process->disabled) return;
    if (pid > 0) 
    {
        if (!process->is_running && process->pid != pid)
        {
            wlog("Process [%s] is running on this host under the pid %d.\n", process->alias, process->alias, pid);
        }
        process->is_running = TRUE;
        process->disabled = FALSE;
        process->pid = pid;
        //if TINE property is guiding the process restart, do not reset the process restart counter, until the TINE response is OK
        if (process->tine_device != NULL && process->tine_property != NULL && process->timeout_restart > 0)
        {
            if (process->tine_status == 0 || process->property_disabled)
            {
                process->current_retry = 0;
            }
            else if (process->current_retry == 0)
            {
                process->current_retry = 1;
            }
        }
        else
        {
            process->current_retry = 0;
        }
    }
}

/**
* Check if all preconditions for the current process are running. If they are not and
* they are described by one of the other process, return 0 and wait for the next cycle
* when all preconditions are suppose to be running. Since they are processes they will
* be handled by the watchdog. If the preconditions are not other processes of this
* watchdog check if a precondition is a process name and start it.
*
* \param process the pointer to the array of all processes
* \param size the number of processes
* \param current_process the index of the process that we need to check the preconditions for
*
* \return 0 if the preconditions are not running but might run in the next cycle,
*           1 if all preconditions are running (or there aren't any),
*           -1 if a precondition could not be started (either it failed to start or it is already disabled)
*           -2 if a precondition (that is not a managed processes) was started.
*           In the latter case we should wait until the next cycle before we do anything with the
*           current process, to give time to the started process to really start.
*/
int checkPreconditions(WATCHED_PROCESS *process, unsigned int size, int current_process)
{
    if (process[current_process].number_of_preceding_processes == 0)
    {
        //if there is no preceding processes to start, return successful
        return 1;
    }

    int ret = 1;
    unsigned int i, j;
    for (i = 0; i < process[current_process].number_of_preceding_processes; i++)
    {
        boolean found = FALSE;
        for (j = 0; j < size; j++)
        {
            //theoretically we should only iterate to current_process, because the processes are order, but just to make sure, go through all of them
            if (j == current_process) continue; //skip the checked process
            if (strcmp(process[current_process].preceding_process[i], process[j].alias) == 0
                || strcmp(process[current_process].preceding_process[i], process[j].id) == 0)
            {
                found = TRUE;
                //if one of the preconditioned processes is disabled disable this process and return
                if (process[j].disabled)
                {
                    wlog("Precondition [%s] of process [%s] is disabled.\n", 
                            process[current_process].id, process[current_process].preceding_process[i], process[current_process].id);
                    return -1;
                }
                checkProcessRunning(&process[j], 0);
                if (!process[j].is_running)
                {
                    //if the precondition is not running, wait for the next cycle to start this process.
                    //The precondition is described by another process and will be started then
                    return 0;
                }
                break;
            }
        }
        if (found) continue;
        int pid = GetProcessID(process[current_process].preceding_process[i], -1, NULL, NULL, 0, FALSE);
        if (pid < 0)
        {
            //start precondition
            wlog("Starting process [%s] as a precondition to [%s].\n", 
                    process[current_process].id, process[current_process].preceding_process[i], process[current_process].id);
            pid = StartProcess(process[current_process].preceding_process[i], NULL, 0, process[current_process].working_directory, NULL, NULL, 0, FALSE, TRUE);
            if (pid < 0) {
                return -1;
            }
            ret = -2;
        }
    }
    return ret;
}

/**
* \brief  Measures CPU and Memory consumption for all the processes.
*
* Measures the CPU usage for all the process at once as well as the total CPU.
* The memory consumption for all processes is also checked.
*
* \param processes the pointer to the processes array
* \param cpu_load pointer to an array of ints into which the CPU usage will be written
* \param memory ponter to an array into which the memory consumption will be written
* \param process_ids pointer to the array into which the process_ids are stored when checking the CPU and memory
* \param size the number of processes
* \param cpu_total pointer where the total CPU usage will be stored
* \param memory_total the pointer in which to store the total memory
*/
void measureCPUAndMemory(WATCHED_PROCESS *processes, int *cpu_load, UINT64 *memory, int *process_ids, int size,
    int *cpu_total, UINT64 *memory_total)
{
    int i;
    LockMutex(TRUE);
    for (i = 0; i < size; i++)
    {
        process_ids[i] = processes[i].pid;
    }
    LockMutex(FALSE);

    GetCPULoadAndMemory(process_ids, size, cpu_load, memory, cpu_total);
    
    LockMutex(TRUE);
    for (i = 0; i < size; i++)
    {
        processes[i].current_cpu_load = cpu_load[i];
        //this can ocurr if the process is not running at all
        if (processes[i].current_cpu_load < 0)
        {
            processes[i].current_cpu_load = 0;
        }
        else if (processes[i].current_cpu_load > 101)
        {
            processes[i].current_cpu_load = 0;
        }
        processes[i].current_memory_consumption = memory[i];
    }
    *memory_total = GetAvailableMemory();
    LockMutex(FALSE);
    
    int cpu_percent;
    UINT64 memory_consumption;
    GetWatchdogCPULoadAndMemory(watchdog_pid, &cpu_percent, &memory_consumption);
    
    LockMutex(TRUE);
    watchdog_cpu_usage = cpu_percent;
    watchdog_memory_consumption = memory_consumption;
    LockMutex(FALSE);
}

/**
* \brief Check CPU and Memory parameters and stop the process if necessary
*
* Checks the CPU and memory consumption and compares them to the configured levels. If the levels
* are exceeded it increases the number of allowed limit violations. If the number of limit violations
* exceed the number of allowed violations, the process is stopped. If the limit is not violated the
* counter is reset.
*
* \param process pointer to the process to check
*/
void doCPUAndMemoryCheck(WATCHED_PROCESS *process)
{
    if (process->current_cpu_load > process->max_cpu_load)
    {
        process->number_of_cpu_violations++;
        if ((process->auto_start == YES || process->process_restart_requested) 
             && process->number_of_cpu_violations > process->number_of_allowed_cpu_load_violations)
        {
            process->cpu_at_stop_time = process->current_cpu_load;
            process->num_of_cpu_violations_at_stop_time = process->number_of_cpu_violations;
            if (StopWatchedProcess(process, FALSE, FALSE) == 0)
            {
                process->number_of_times_stopped_by_wd++;
                process->stopped_because_of_cpu = TRUE;
                wlog("Process [%s] stopped. Too many CPU violations. Current CPU load: %d.\n", 
                        process->alias, process->alias, process->current_cpu_load);
            }
            else
            {
                wlog("Failed to stop process [%s]. Too many CPU violations. Current CPU load: %d.\n",
                        process->alias, process->alias, process->current_cpu_load);
            }
        }
    }
    else
    {
        process->number_of_cpu_violations = 0;
    }
    //do the memory check
    if (process->current_memory_consumption > process->max_memory)
    {
        process->number_of_memory_violations++;
        if ((process->auto_start == YES || process->process_restart_requested) 
            && process->number_of_memory_violations > process->number_of_allowed_memory_violations)
        {
            process->memory_at_stop_time = process->current_memory_consumption;
            process->num_of_memory_violations_at_stop_time = process->number_of_memory_violations;
            if (StopWatchedProcess(process, FALSE, FALSE) == 0)
            {
                process->stopped_because_of_memory = TRUE;
                process->number_of_times_stopped_by_wd++;
                wlog("Process [%s] stopped. Too many memory violations. Current memory: %3.2f.\n", 
                        process->alias, process->alias, process->current_memory_consumption / (1024.0*1024.0));
            }
            else
            {
                wlog("Failed to stop process [%s]. Too many memory violations. Current memory: %3.2f.\n", 
                        process->alias, process->alias, process->current_memory_consumption / (1024.0*1024.0));
            }
        }
    }
    else
    {
        process->number_of_memory_violations = 0;
    }
}

/**
* Callback for the TINE property links.
*/
void callback(int link, int cc)
{
    unsigned int i;
    for (i = 0; i < num_processes; i++)
    {
        if (g_process[i].link_id != link) continue;
        if (g_process[i].tine_status != 0 && cc == 0)
        {
            wlog("Connection to %s/%s for process [%s] successful. Link ID: %d.\n", 
                    g_process[i].alias, g_process[i].tine_device, g_process[i].tine_property, g_process[i].alias, link);
        }
        else if (g_process[i].tine_status != cc)
        {
            char errstr[128];
            wlog("Connection to %s/%s for process [%s] error: %s (%d). Link ID: %d.\n", 
                    g_process[i].alias, g_process[i].tine_device, g_process[i].tine_property,
                    g_process[i].alias, GetLastLinkError(cc, errstr), cc, link);
        }
        g_process[i].tine_status = cc;

        if (cc == operation_success)
        {
            g_process[i].last_tine_response = time(0);
            g_process[i].property_link_timeout = FALSE;
        }
        else if (cc == buffer_too_small && g_process[i].default_property_info_used)
        {
            g_process[i].last_tine_response = time(0);
            g_process[i].property_link_timeout = FALSE;
            CloseServerMonitor(&g_process[i]);
        }

        //do not break, because multiple processes can have the same link - not a real usecase, but can happen
    }
}

/**
* \brief Check the status of the TINE property link for the process.
*
* Checks the status of the connection to the TINE property and takes
* appropriate actions. If the connection is active and the response
* was received within the timeout period, the routine doesn't do anything.
* If no response was received for more than timeout, the process is closed.
* If the link is already closed and more time has passed since the last
* retry than specified, method tries to re-attach to the property.
*
* \param process the pointer to the process to check
*/
void checkPropertyLink(WATCHED_PROCESS *process)
{
    if (process->tine_device == NULL || process->tine_property == NULL) 
    {
        process->property_link_timeout = FALSE;
        return;
    }
    if (strlen(process->tine_device) == 0 || strlen(process->tine_property) == 0) 
    {
        process->property_link_timeout = FALSE;
        return;
    }
    if (suspended || process->property_disabled || process->disabled)
    {
        return;
    }

    if ((process->link_id > -1)
        && (process->timeout_restart > 0)
        && (process->is_running)
        && (now - process->last_tine_response >= process->timeout_restart)
        && (now - process->last_tine_retry >= process->timeout_restart)
        && (now - process->last_retry_timestamp > process->timeout_restart + process->start_after_n_seconds))
    {
        //if there was no tine response for timeout_restart, stop the process
        process->property_link_timeout = TRUE;
        if (process->auto_start == YES || process->process_restart_requested)
        {
            if (StopWatchedProcess(process, FALSE, FALSE) == 0)
            {
                process->number_of_times_stopped_by_wd++;
                process->stopped_because_of_link = TRUE;
                wlog("Process [%s] stopped. TINE property (%s/%s) link timeout.\n", 
                        process->alias, process->alias, process->tine_device, process->tine_property);
            }
            else
            {
                wlog("Failed to stop process [%s]. TINE property (%s/%s) link timeout.\n", 
                        process->alias, process->alias, process->tine_device, process->tine_property);
            }
        }

    }
    else if (process->link_id > -1)
    {
        //check when the last response was received and raise the flag if there is a timeout
        if (now - process->last_tine_response > process->timeout)
        {
            if (!process->property_link_timeout)
                wlog("Process [%s] link timeout.\n", process->alias, process->alias);
            process->property_link_timeout = TRUE;
        }
    }
    else if (process->is_running)
    {
        if (process->link_id == -1) //link_id == -1, there is no active link yet
        {
            //start the link
            process->property_disabled = FALSE;
            if (now - process->last_tine_retry >= process->timeout ||
                (process->tine_status == buffer_too_small && process->default_property_info_used))
            {
                StartServerMonitor(process, callback);
            }
        }
        else //link_id < -1, we tried to open a link once but it didn't work (link blacklisted or something like that)
        {
            if ((now - process->last_tine_retry >= process->timeout_restart)
                && (process->auto_start == YES || process->process_restart_requested)
                && (now - process->last_retry_timestamp > process->timeout_restart + process->start_after_n_seconds))
            {
                if (StopWatchedProcess(process, FALSE, FALSE) == 0)
                {
                    process->number_of_times_stopped_by_wd++;
                    process->stopped_because_of_link = TRUE;
                    wlog("Process [%s] stopped. TINE property (%s/%s) link timeout.\n", 
                            process->alias, process->alias, process->tine_device, process->tine_property);
                }
                else
                {
                    wlog("Failed to stop process [%s]. TINE property (%s/%s) link timeout.\n", 
                            process->alias, process->alias, process->tine_device, process->tine_property);
                }
            }
            if (now - process->last_tine_response >= process->timeout)
            {
                process->property_link_timeout = TRUE;
            }
        }
    }
}

/**
* \brief Do the watch magic.
*
* Single try to watch the processes. It measures the CPU and memory, and starts or stops
* processes according to the policies. If the watchdog is suspended, the memory and cpu
* are still measured, but no other action is taken.
*
* \param processes the watched processes
* \param cpu_load the pointer to an array of ints to use when measuring CPU
* \param process_ids the pointer to an array of ints to use when measuring CPU
* \param size the number of processes (the length of all three arrays)
* \param cpu_total the pointer to write the total cpu usage to
* \param memory the pointer to write the total memory usage to
*/
void watchCycle(WATCHED_PROCESS *processes, int *cpu_load, UINT64 *memory, int *process_ids, unsigned int size,
    int *cpu_total, UINT64 *memory_total)
{
    struct timespec start, end;
    clock_gettime(CLOCK_MONOTONIC, &start);

    now = time(0);
    //delay start network utilization measurements
    if (now - startup_time > 60 && num_network_interfaces > 0) 
        GetNetworkUtilization(network_percentage, network_absolute, num_network_interfaces);
    measureCPUAndMemory(processes, cpu_load, memory, process_ids, size, cpu_total, memory_total);

    now = time(0);
    LockMutex(TRUE);
    unsigned int i;
    for (i = 0; i < num_network_interfaces; i++)
    {
        network_utilization_absolute[i] = network_absolute[i];
        network_utilization_percentage[i] = network_percentage[i];
    }

    //if suspended timeout elapsed, resume the watchdog
    if (suspended)
    {
        remaining_suspended_time = (int)(suspended_timeout + made_suspended - now);
        if (remaining_suspended_time <= 0)
        {
            suspended = FALSE;
            remaining_suspended_time = 0;
        }
    }

    for (i = 0; i < size; i++)
    {
        //do not check for running if it was started just recently
        boolean only_check = ((now - processes[i].last_retry_timestamp) < processes[i].start_after_n_seconds);
        processes[i].is_starting = only_check || processes[i].last_retry_timestamp == 0;
        //check if process is running, this will set the is running flag to true or false
        checkProcessRunning(&processes[i], only_check);
        if (only_check)
        {
            continue;
        }
        checkPropertyLink(&processes[i]);
        //ignore all disabled processes and disable all of those that match the criteria (we tried to start them several times but did not succeed)
        if (processes[i].disabled)
        {
            continue;
        }
        if (processes[i].current_retry >= processes[i].number_of_retries)
        {
            wlog("Failed to start process [%s] too many times. Process is now disabled.\n", processes[i].alias, processes[i].alias);
            //On disabling the process stop the process and the link. Disabled process shouldn't occupy any resources
            if (StopWatchedProcess(&processes[i], TRUE, FALSE) == 0)
            {
                processes[i].number_of_times_stopped_by_wd++;
            }            
            processes[i].disabled = TRUE;
            processes[i].current_retry = 0;
            continue;
        }
        if (!suspended)
        {
            if (!processes[i].is_running)
            {
                //if the process is not running and auto_start is true, check start criteria
                if ((processes[i].auto_start == YES || processes[i].process_restart_requested)
                        && (processes[i].pause_after_previous_start <= 0 || (now - last_process_start > processes[i].pause_after_previous_start)))
                {
                    int tryStart = 0;
                    //if eligible for start (enough time has passed), check preconditions
                    if (((now - processes[i].last_retry_timestamp) > processes[i].retry_interval) 
                        && ((now - processes[i].determined_not_running_timestamp) > processes[i].retry_interval))
                    {
                        tryStart = checkPreconditions(processes, size, i);
                        if (tryStart == -2)
                        {
                            //if preconditions were started for the first time, flag the process that its
                            //preconditions are in a startup state
                            if (!processes[i].preconditions_started)
                            {
                                processes[i].preconditions_started = TRUE;
                            }
                            //if preconditions were started for the second time in the row, it means that something
                            //is wrong. In that case flag the process that it could not be started
                            else
                            {
                                processes[i].preconditions_started = FALSE;
                                tryStart = -1;
                            }
                        }
                    }
                    //if -1, it means that one of the preconditions could not be started or one of the preconditions is disabled
                    if (tryStart == -1)
                    {
                        processes[i].last_retry_timestamp = time(0);
                        processes[i].current_retry++;
                    }
                    else if (tryStart == 1)
                    {
                        //try to start the process
                        StartWatchedProcess(&processes[i]);
                        processes[i].number_of_times_started_by_wd++;
                    }
                }
            }
            else
            {
                processes[i].is_starting = FALSE;
                if (processes[i].last_retry_timestamp == 0)
                {
                    processes[i].last_retry_timestamp = now;
                }
                //if TINE is guiding the restart, and TINE property is disabled, than stop the process
                //if (processes[i].timeout_restart > 0 && processes[i].property_disabled)
                //    StopWatchedProcess(&processes[i], 0);

                //if process is running check the CPU and memory conditions
                doCPUAndMemoryCheck(&processes[i]);
            }
        }
    }

    if (print_info_on_each_cycle)
    {
        unsigned int i;
        for (i = 0; i < size; i++)
        {
            PrintProcessStatus(processes[i]);
        }
    }
    LockMutex(FALSE);
    clock_gettime(CLOCK_MONOTONIC, &end);
    int diff = (int)(1000.0 * (end.tv_sec - start.tv_sec) + (end.tv_nsec - start.tv_nsec)/1e6);

    if (time_reported > 15)
    {
        if (diff * 1.5 > watch_cycle_period)
            wlog("Time to execute one cycle (%d ms) is too long for the current watch cycle (%d ms). Increase to at least %d ms is recommended.\n",
            "Watchdog", diff, watch_cycle_period, 2 * diff);
    }
    else
    {
        time_reported = time_reported + 1;
    }
}

/**
* \brief Stores the environment variables received from the main method.
*
* Counts and stores all environment variables. The variables are stored into
* the fields defined by process.h.
*
* \param envp the pointer to the matrix of environment variables as received in the main method
*/
void setupEnvironment(char **envp)
{
    unsigned int i, m, n, actual_numv;
    char **env, **temp, **renv;
    
    for (i = 0; i < set_environment_variables_length; i++)
    {   
        _putenv(set_environment_variables[i]);
    }
#if defined(WIN32)
    wchar_t *name, *value;
    size_t size, l;
    for (i = 0; i < set_environment_variables_length; i++)
    {
        temp = split(set_environment_variables[i], &m, '=', TRUE, TRUE, FALSE);
        size = strlen(temp[0]) + 1;
        name = (wchar_t*)alloca(sizeof(wchar_t)* size);
        if (name == NULL) OutOfMemoryExit(5);
        mbstowcs_s(&l, name, size, temp[0], _TRUNCATE);
        name[size - 1] = '\0';
        if (m == 1)
        {
            value = (wchar_t*)alloca(sizeof(wchar_t)*1);
            if (value == NULL) OutOfMemoryExit(5);
            value[0] = '\0';
        }
        else
        {
            size = strlen(temp[1]) + 1;
            value = (wchar_t*)alloca(sizeof(wchar_t)* size);
            if (value == NULL) OutOfMemoryExit(5);
            mbstowcs_s(&l, value, size, temp[1], _TRUNCATE);
            value[size - 1] = '\0';
        }
        SetEnvironmentVariable(name,value);
        freeArray(temp,m);
    }
#endif

    parent_environment = envp;
    num_environment = 0;
    //count original environment variables
    for (env = parent_environment; *env != 0; env++)
    {
        num_environment++;
    }
    
    if (set_environment_variables_length > 0)
    {
        actual_numv = 0;
        //remove thos environement variables that are overwritten
        parent_environment = (char**)malloc(sizeof(char*)* (num_environment + set_environment_variables_length));
        if (parent_environment == NULL) OutOfMemoryExit(2);
        unsigned int j;
        for (j = 0; j < (int)num_environment; j++)
        {
            renv = split(envp[j], &n, '=', TRUE, TRUE, FALSE);
            boolean found = FALSE;
            for (i = 0; i < set_environment_variables_length; i++)
            {
                temp = split(set_environment_variables[i], &m, '=', TRUE, TRUE, FALSE);
                if (strcmp(temp[0], renv[0]) == 0)
                    found = TRUE;
                freeArray(temp,m);
                if (found) break;
            }
            freeArray(renv,n);
            if (!found)
            {
                m = (int)(sizeof(char)*strlen(envp[j]) + 1);
                parent_environment[actual_numv] = (char*)malloc(m);
                if (parent_environment[actual_numv] == NULL) OutOfMemoryExit(2);
                strcpy_s(parent_environment[actual_numv],m,envp[j]);
                parent_environment[actual_numv][m - 1] = '\0';
                actual_numv++;
            }
        }    
        for (j = 0; j < set_environment_variables_length; j++)
        {
            m = (int)(sizeof(char)*strlen(set_environment_variables[j]) + 1);
            parent_environment[actual_numv] = (char*)malloc(m);
            if (parent_environment[actual_numv] == NULL) OutOfMemoryExit(2);
            strcpy_s(parent_environment[actual_numv], m, set_environment_variables[j]);
            parent_environment[actual_numv][m - 1] = '\0';
            actual_numv++;
        }
        num_environment = actual_numv;   
    }
}

void setUpInfo(SERVER_INFO *info)
{
    char *name;
    name = GetMachineName();
    setUpDefaultInfo(info, "SERVICE", "WD_", name, "WDOG");
    free(name);
}

void setUpConfig(char *file)
{
    printf("Reading configuration file: %s.\n", file);
    wlog("Reading configuration file: %s.\n", "Setup", file);
    g_process = ParseFile(file, &num_processes, &info);
    setUpInfo(&info);
    config_file = toChar(file,FALSE);
    printf("%d processes configured.\n", num_processes);
    wlog("%d processes configured.\n", "Setup", num_processes);
}

/**
* \brief Opens the configuration file and loads the process definitions.
*
* Checks if the configuration file is provided as a startup parameter. If yes,
* the method checks if the file exists and if yes it reads the configuration from it.
* If the configuration file is not provided, the default one is opened if it exists.
* If none of the files exist the instructions are printed.
*
* \param argc number of startup parameters
* \param argv the startup parameters
*/
void openConfigFile(int argc, char **argv)
{
    info.port_offset = -1;
    if (argc < 2)
    {
        if (fileExists(DEFAULT_CONFIG_FILE))
        {
            setUpConfig(DEFAULT_CONFIG_FILE);
            return;
        }
        else if (fileExists(DEFAULT_CONFIG_FILE2))
        {
            setUpConfig(DEFAULT_CONFIG_FILE2);
            return;
        }
    }
    else if (fileExists(argv[1]))
    {
        setUpConfig(argv[1]);
        return;
    }
    //if configuration is not available, print instructions and exit
    PrintInstructions();
    exit(0);
}

/**
* \brief Watch all the processes in an endless loop
*
* Starts an endless loop and monitors all the processes periodically.
* If a process violates any of the settings appropriate actions are
* taken. Regarding functionality this routine is the same as calling
* Watch(void) in an endless loop.
*/
void watchLoop(int wait_period)
{
    //constantly loop over all processes
    while (running)
    {
        Watch();
        Wait(wait_period);
    }
}

void init()
{
    InitBuffers(num_processes);
    cpu_load = (int*)malloc(sizeof(int)*num_processes);
    if (cpu_load == NULL) OutOfMemoryExit(2);
    process_ids = (int*)malloc(sizeof(int)*num_processes);
    if (process_ids == NULL) OutOfMemoryExit(2);
    memory = (UINT64*)malloc(sizeof(UINT64)*num_processes);
    if (memory == NULL) OutOfMemoryExit(2);
    made_suspended = time(0);
    last_process_start = time(0);
    watchdog_pid = GetPid();

#ifdef USE_NETWORK
    if (network_interface_names == NULL && g_network_enabled)
    {
        num_network_interfaces = 0;
        printf("Initialising network counters. This could take a while. Please, stay with me... ");
        int i = GetNetworkInterfaceNames(NULL, &num_network_interfaces);
        if (i == ERROR_BUFFER_TOO_SMALL)
        {
            network_interface_names = (char**)malloc(sizeof(char*)*num_network_interfaces);
            if (network_interface_names == NULL) OutOfMemoryExit(2);
            i = GetNetworkInterfaceNames(network_interface_names, &num_network_interfaces);
            if (i != 0) 
            {
                freeArray(network_interface_names, num_network_interfaces);
                network_interface_names = NULL;
                num_network_interfaces = 0;
            }
            else 
            {
                network_utilization_percentage = (float*)malloc(sizeof(float)*num_network_interfaces);
                if (network_utilization_percentage == NULL) OutOfMemoryExit(2);
                network_utilization_absolute = (float*)malloc(sizeof(float)*num_network_interfaces);
                if (network_utilization_absolute == NULL) OutOfMemoryExit(2);
                
                network_percentage = (float*)malloc(sizeof(float)*num_network_interfaces);
                if (network_percentage == NULL) OutOfMemoryExit(2);
                network_absolute = (float*)malloc(sizeof(float)*num_network_interfaces);
                if (network_absolute == NULL) OutOfMemoryExit(2);

                network_interface_friendly_names = (char**)malloc(sizeof(char*) * num_network_interfaces);
                if (network_interface_friendly_names == NULL) OutOfMemoryExit(2);
                GetNetworkInterfaceFriendlyNames(network_interface_friendly_names, &num_network_interfaces);
                for (i = 0; i < (int)num_network_interfaces; i++)
                {
                    network_utilization_absolute[i] = -1;
                    network_utilization_percentage[i] = -1;
                    network_absolute[i] = -1;
                    network_percentage[i] = -1;
                }
            }
            printf("Done.\n\n");
        } 
        else if (i == ERROR_NETWORK_NOT_AVAILABLE)
        {
            printf("Failed.\n\n");
        } 
        else
        {
            printf("Done.\n\n");
        }
    }
    else if (!g_network_enabled)
    {
        num_network_interfaces = 0;
    }
#endif
}

void releaseProcesses(void)
{
    DEBUG("Releasing resources.\n");
    if (g_process != NULL)
    {
        unsigned int i;
        for (i = 0; i < num_processes; i++)
        {
            if (g_process[i].link_id > -1)
            {
                CloseServerMonitor(&g_process[i]);
            }
            if (g_process[i].id != NULL) free(g_process[i].id);
            if (g_process[i].alias != NULL) free(g_process[i].alias);
            if (g_process[i].name != NULL) free(g_process[i].name);
            if (g_process[i].path_to_executable != NULL) free(g_process[i].path_to_executable);
            if (g_process[i].tine_device != NULL) free(g_process[i].tine_device);
            if (g_process[i].tine_property != NULL) free(g_process[i].tine_property);
            if (g_process[i].parameters != NULL) free(g_process[i].parameters);
            if (g_process[i].working_directory != NULL) free(g_process[i].working_directory);
            freeArray(g_process[i].individual_parameters, g_process[i].number_of_parameters);
            freeArray(g_process[i].preceding_process, g_process[i].number_of_preceding_processes);
            freeArray(g_process[i].environment_variables, g_process[i].number_of_environment_variables);
            if (g_process[i].caption != NULL) free(g_process[i].caption);
            if (g_process[i].context != NULL) free(g_process[i].context);
            if (g_process[i].server != NULL) free(g_process[i].server);
            if (g_process[i].dout.data.vptr != NULL) free(g_process[i].dout.data.vptr);
            InitializeDefaults(&g_process[i]);
        }
        free(g_process);
    }
    g_process = NULL;
    if (cpu_load != NULL) free(cpu_load);
    cpu_load = NULL;
    if (process_ids != NULL) free(process_ids);
    process_ids = NULL;
    if (memory != NULL) free(memory);
    memory = NULL;
    if (console_command != NULL) free(console_command);
    console_command = NULL;
    FreeBuffers();
    DEBUG("Resources released.\n");
}

void Shutdown(void)
{
    static boolean shut_down = FALSE;
    if (shut_down) return;
    EnableCloseWindow();
    StopServer();

    if (stop_all_on_shutdown)
    {
        unsigned int i;
        for (i = 0; i < num_processes; i++)
        {
            StopWatchedProcess(&g_process[i], TRUE, FALSE);
        }
    }
    if (network_interface_names != NULL)
    {
        freeArray(network_interface_names, num_network_interfaces);
        network_interface_names = NULL;
    }
    if (network_interface_friendly_names != NULL)
    {
        freeArray(network_interface_friendly_names, num_network_interfaces);
        network_interface_friendly_names = NULL;
    }
    if (network_utilization_percentage != NULL)
    {
        free(network_utilization_percentage);
        network_utilization_percentage = NULL;
    }
    if (network_utilization_absolute != NULL)
    {
        free(network_utilization_absolute);
        network_utilization_percentage = NULL;
    }
    num_network_interfaces = 0;
    NetworkCleanUp();
    shut_down = TRUE;
    suspended = TRUE;
    releaseProcesses();
    SetupMutex(FALSE);
}

void FreeResources(void)
{
    if (info.context != NULL) free(info.context);
    info.context = NULL;
    if (info.server_name != NULL) free(info.server_name);
    info.server_name = NULL;
    if (info.description != NULL) free(info.description);
    info.description = NULL;
    if (info.fec_name != NULL) free(info.fec_name);
    info.fec_name = NULL;
    if (info.hardware != NULL) free(info.hardware);
    info.hardware = NULL;
    if (info.location != NULL) free(info.location);
    info.location = NULL;
    if (info.responsible != NULL) free(info.responsible);
    info.responsible = NULL;
    if (info.subsystem != NULL) free(info.subsystem);
    info.subsystem = NULL;
    if (config_file != NULL) free(config_file);
    config_file = NULL;
    if (console_command != NULL) free(console_command);
    console_command = NULL;
    freeArray(set_environment_variables, set_environment_variables_length);
    set_environment_variables_length = 0;
    set_environment_variables = NULL;
}

void ReloadProcesses(void)
{
    SERVER_INFO new_info = { 0 };
    releaseProcesses();
    g_process = ParseFile(config_file, &num_processes, &new_info);
    init();
    wlog("Reloaded the configuration file. Now monitoring %d processes\n", "Watchdog", num_processes);
    PrintSettings(g_process, num_processes, &info);
    printf(">");
    freeInfo(&new_info);
}

int RewriteConfiguration(void)
{
    DEBUG("Writing configuration to file.\n");
    SERVER_INFO new_info = { 0 };
    new_info.port_offset = -1;
    setUpInfo(&new_info);
    int ret = WriteConfigFile(config_file, g_process, num_processes, info, new_info);
    freeInfo(&new_info);
    DEBUG("Configuration written.\n");
    return ret;
}

void Watch(void)
{
    watchCycle(g_process, cpu_load, memory, process_ids, num_processes, &total_cpu_usage, &available_memory);
}


int RebootComputer(void)
{
    return RestartComputer();
}

int StartWatchedProcess(WATCHED_PROCESS *process)
{
    DEBUG("Starting process %s.\n",process->alias);
    process->process_requested_to_stop = 0;
    if (process->auto_start == ONCE)
    {
        int pid;
        if (process->is_service)
        {
            pid = StartProcess(process->name,
                               process->individual_parameters, 
                               process->number_of_parameters, 
                               process->working_directory, 
                               process->caption,
                               process->environment_variables, 
                               process->number_of_environment_variables, 
                               process->is_service,
                               process->show_console_window);
        }
        else
        {
            pid = StartProcess(process->path_to_executable, 
                               process->individual_parameters, 
                               process->number_of_parameters, 
                               process->working_directory, 
                               process->caption, 
                               process->environment_variables, 
                               process->number_of_environment_variables, 
                               process->is_service,
                               process->show_console_window);
        }
        if (pid > 0) {
            //refresh the PID
            pid = GetProcessID(process->name, 
                               process->pid, 
                               process->match_path ? process->path_to_executable : NULL,
                               process->individual_parameters, 
                               process->match_parameters ? process->number_of_parameters : 0, 
                               process->is_service);
            DEBUG("Process %s started.\n", process->alias);
            return 0;
        }
        DEBUG("Failed to start process %s: %d.\n",process->alias, -pid);
        //pid is negative in case of an error, so this returns a positive number
        return -pid;
    }

    process->stopped_by_user = 0;
    process->pid = GetProcessID(process->name, 
                                process->pid, 
                                process->match_path ? process->path_to_executable : NULL, 
                                process->individual_parameters, 
                                process->match_parameters ? process->number_of_parameters : 0,
                                process->is_service);

    if (process->pid > 0) {
        process->disabled = 0;
        process->is_running = 1;
        process->process_restart_requested = FALSE;
        process->stopped_because_of_cpu = 0;
        process->stopped_because_of_memory = 0;
        process->stopped_because_of_link = 0;
        DEBUG("Process %s running.\n", process->alias);
        return 0; //process is already running
    }
    
    if (process->is_service)
    {
        process->pid = StartProcess(process->name, 
                                    process->individual_parameters, 
                                    process->number_of_parameters,
                                    process->working_directory, 
                                    process->caption, 
                                    process->environment_variables, 
                                    process->number_of_environment_variables, 
                                    process->is_service, 
                                    process->show_console_window);
    }
    else
    {
        process->pid = StartProcess(process->path_to_executable, 
                                    process->individual_parameters, 
                                    process->number_of_parameters,
                                    process->working_directory, 
                                    process->caption, 
                                    process->environment_variables, 
                                    process->number_of_environment_variables, 
                                    process->is_service, 
                                    process->show_console_window);
    }
    process->has_alarm = FALSE;
    process->disabled = FALSE;
    process->last_retry_timestamp = time(0);
    process->current_retry++;
    last_process_start = time(0);
    //force to refresh the pid
    process->pid_last_refreshed = time(0) - pid_refresh_rate + 2;
    if (process->pid > 0) {
        process->is_running = TRUE;
        process->stopped_because_of_cpu = FALSE;
        process->stopped_because_of_memory = FALSE;
        process->stopped_because_of_link = FALSE;
        //Make sure that we have the correct PID. If the child process is a linux terminal process it can happen that the
        //current pid is the PID of the terminal and not the PID of the process.
        process->pid = GetProcessID(process->name, 
                                    process->pid, 
                                    process->match_path ? process->path_to_executable : NULL, 
                                    process->individual_parameters, 
                                    process->match_parameters ? process->number_of_parameters : 0, 
                                    process->is_service);
        process->process_restart_requested = FALSE;
        if (process->pid > 0)
        {
            DEBUG("Process %s started successfully.\n", process->alias);
            return 0;
        }
        else
        {
            DEBUG("Process '%s' supposedly started, but the name '%s' cannot be found in the list of running processes.\n", process->alias, process->name);
            //this can happen only on linux when the process is still starting up and the proc file descriptor have not been properly set up yet. 
            //There is nothing to worry about, because it will be picked up in the next cycle
#if defined(WIN32) 
            return ERROR_PATH_NOT_FOUND;
#else 
            return 0;
#endif
        }
    }
    // if there was an error the process->pid is negative and it represents the error code of the process start request
    DEBUG("Failed to start process %s: %d.\n", process->alias, -(process->pid));
    return -(process->pid);
}

int StopWatchedProcess(WATCHED_PROCESS *process, boolean stopTineLink, boolean kill)
{
    DEBUG("Stopping process %s.\n", process->alias);
    //if stopping the process, kill the tine property monitor first
    if (stopTineLink)
    {
        CloseServerMonitor(process);
    }
    if (!kill && process->process_requested_to_stop > 0)
    {
        DEBUG("Process %s is already being stopped.\n", process->alias);
        return -1; //already requested to stop gracefully, do not try again, kill will kick in after a few seconds
    }
    int pid = process->pid;
    //whenever a process is stopped, restart the TINE retries
    process->property_disabled = FALSE;
    if (pid < 0)
    {
        pid = GetProcessID(process->name, -1,
                           process->match_path ? process->path_to_executable : NULL,
                           process->individual_parameters,
                           process->match_parameters ? process->number_of_parameters : 0,
                           process->is_service);
    }
    if (pid < 0) 
    {
        process->is_running = FALSE;
        DEBUG("Process %s not running.\n", process->alias);
        return 0; //process already stopped
    }
    if (!kill)
    {
        process->process_requested_to_stop = time(0);
    }
    process->pid = pid;
    int ret = StopProcess(pid, process->is_service, process->name, kill);
    if (ret == 1)
    {
        process->number_of_cpu_violations = 0;
        process->number_of_memory_violations = 0;
        process->pid = -1;
        DEBUG("Process %s stopped.\n", process->alias);
        return 0;
    }
    DEBUG("Failed to stop process %s: %d.\n", process->alias, -ret);
    return -ret;
}

void PrintProcessStatus(WATCHED_PROCESS process)
{
    printf(">>>> Process %s (%d), %s, %s, TINE: %d CPU: %d %%, Memory: %3.2f MB\n>",
        process.alias,
        process.pid,
        process.disabled ? "disabled" : "enabled",
        process.is_running ? "running" : "stopped",
        process.tine_status,
        process.current_cpu_load,
        process.current_memory_consumption / (1024.0*1024.0)
        );
}

void intHandler(int dummy)
{
    SystemKillCycleTimer();
    StopServer();
    //give the server time to terminate gracefully
    Wait(watch_cycle_period * 2);
    Shutdown();
    FreeResources();
}

#ifdef UNIX
#include <sys/wait.h>

void intHandlerChild(int dummy)
{
    while (waitpid((pid_t)(-1), 0, WNOHANG) > 0) {}
}
#endif

int main(int argc, char **argv, char **envp)
{
    char *c;
    int i;
    for (i = 0; i < argc; i++)
    {
        c = argv[i];
        if (*c == '/' || *c == '-')
        {
            switch (tolower(*c))
            {
            case 'h':
            case '?':
                PrintInstructions();
                return 1;
            }
        }
    }
    DisableCloseWindow();

    openConfigFile(argc, argv);
    setupEnvironment(envp);
    startup_time = time(0);
    /*c = getenv("WATCHDOG_RESTARTED");
    if (c == NULL)
    {   
        env = malloc(sizeof(char*)*1);
        env[0] = malloc(sizeof(char)*20+1);
        strcpy_s(env[0], 21, "WATCHDOG_RESTARTED=1");
        env[0][20] = '\0';
        StartProcess(argv[0],&argv[1],argc-1,NULL,"Watchdog",env,1,0);
        return;
    }*/
    PrintSettings(g_process, num_processes, &info);
    SetupMutex(TRUE);
    init();
    if (signal(SIGINT, intHandler) == SIG_ERR)
    {
        printf("Cannot configure interrupt signal.\n");
    }
    if (signal(SIGTERM, intHandler) == SIG_ERR)
    {
        printf("Cannot configure terminate signal.\n");
    }
#if defined(WIN32)
    if (signal(SIGBREAK, intHandler) == SIG_ERR)
    {
        printf("Cannot configure break signal.\n");
    }
#elif defined(UNIX)
    if (signal(SIGCHLD, SIG_IGN) == SIG_ERR)
    {
        printf("Failed to redirect child process interrupt signals to ignore.\n");
        if (signal(SIGCHLD, intHandlerChild) == SIG_ERR)
        {
            printf("Failed to configure child process interrupt signals. Zombie processes will not be cleaned up.\n");
        }
    }
#endif

    if (start_server)
    {
        StartServer(watch_cycle_period, (unsigned int)(1.1*watch_cycle_period), info);
    }
    else
    {
        //if server was not started take care that cycler is started so we can monitor the TINE properties
        SystemStartCycleTimer();
        watchLoop(watch_cycle_period);
    }
    SystemKillCycleTimer();
    Shutdown();
    FreeResources();
    return 0;
}
