/*
* watchdog.h
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#ifndef WATCHDOG_H
#define WATCHDOG_H

#include "project.def"
#include "util.h"

#define WAPPV_MAJOR 2
#define WAPPV_MINOR 8
#define WAPPV_REVISION 0

#define DEFAULT_CONFIG_FILE "watchdog.cfg"
#define DEFAULT_CONFIG_FILE2 "wdgcfg.txt"

/* Definition of configuration parameters for individual process*/
#define WD_NAME "NAME"
#define WD_ALIAS "ALIAS"
#define WD_SERVICE "SERVICE"
#define WD_PROCESS_CAPTION "PROCESS_CAPTION"
#define WD_DESCRIPTION "DESCRIPTION"
#define WD_PATH_TO_EXECUTABLE "PATH_TO_EXECUTABLE"
#define WD_WORKING_DIRECTORY "WORKING_DIRECTORY"
#define WD_MONITOR_N_SECONDS_AFTER_START "MONITOR_N_SECONDS_AFTER_START"
#define WD_PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS "PAUSE_N_SECONDS_AFTER_PREVIOUS_PROCESS"
#define WD_NUMBER_OF_RETRIES "NUMBER_OF_RETRIES"
#define WD_RETRY_INTERVAL "RETRY_INTERVAL"
#define WD_AUTOSTART "AUTO_START"
#define WD_MAX_CPU_LOAD "MAX_CPU_LOAD"
#define WD_NUMBER_OF_ALLOWED_CPU_VIOLATIONS "NUMBER_OF_ALLOWED_CPU_LOAD_VIOLATIONS"
#define WD_PARAMETERS "COMMAND_LINE_PARAMETERS"
#define WD_MATCH_PATH "MATCH_PATH"
#define WD_MATCH_PARAMETERS "MATCH_PARAMETERS"
#define WD_SHOW_CONSOLE_WINDOW "SHOW_CONSOLE_WINDOW"
#define WD_PRECEDING_PROCESSES "PRECEDING_PROCESSES"
#define WD_MAX_MEMORY "MAX_MEMORY"
#define WD_NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS "NUMBER_OF_ALLOWED_MEMORY_VIOLATIONS"
#define WD_ENVIRONMENT_VARIABLES "ENVIRONMENT_VARIABLES"
#define WD_TINE_PROPERTY "T_PROPERTY"
#define WD_TINE_DEVICE "T_DEVICE"
#define WD_POLLING_RATE "T_POLLING_RATE"
#define WD_TIMEOUT "T_TIMEOUT"
#define WD_RESTART_TIMEOUT "T_RESTART_TIMEOUT"

/* Definition of configuration parameters for watchdog */
#define WD_G_CPU_MEASURING_PERIOD "CPU_MEASURING_PERIOD" //deprecated, do not use
#define WD_G_WATCH_CYCLE_PERIOD "WATCH_CYCLE_PERIOD"
#define WD_G_PRINT_INFO_ON_EACH_CYCLE "PRINT_INFO_ON_EACH_CYCLE"
#define WD_G_START_SERVER "START_SERVER"
#define WD_G_SUSPENDED_TIMEOUT "SUSPENDED_TIMEOUT"
#define WD_G_ALLOW_MONITORING_OTHER_HOSTS "ALLOW_MONITORING_OTHER_HOSTS"
#define WD_G_PID_REFRESH_RATE "PID_REFRESH_RATE"
#define WD_G_CONSOLE_COMMAND "CONSOLE_COMMAND"
#define WD_G_STOP_ALL_ON_SHUTDOWN "STOP_ALL_ON_SHUTDOWN"
#define WD_G_SHORT_TERM_HISTORY "SHORT_TERM_HISTORY"
#define WD_G_LONG_TERM_HISTORY "LONG_TERM_HISTORY"
#define WD_G_ENVIRONMENT_VARIABLES "ENVIRONMENT_VARIABLES"
#define WD_G_KILL_GRACE_PERIOD "GRACE_PERIOD_BEFORE_KILL"
#define WD_G_NETWORK_ENABLED "NETWORK_ENABLED"
#define WD_G_MINIMUM_FREE_DISK_PATH "MINIMUM_FREE_DISK_PATH"
#define WD_G_MINIMUM_FREE_DISK_MBYTE "MINIMUM_FREE_DISK_MBYTE"

#define ERR_FORK_NOT_POSSIBLE 1

#define DEFAULT_STOP_ALL_ON_SHUTDOWN         FALSE
#define DEFAULT_SHORT_TERM_HISTORY           300
#define DEFAULT_LONG_TERM_HISTORY            1
#define DEFAULT_KILL_GRACE_PERIOD            10
#define DEFAULT_SUSPENDED_TIMEOUT            600
#define DEFAULT_PID_REFRESH_RATE             60
#define DEFAULT_START_SERVER                 TRUE
#define DEFAULT_PRINT_INFO_ON_EACH_CYCLE     FALSE
#define DEFAULT_WATCH_CYCLE_PERIOD           1000
#define DEFAULT_ALLOW_MONITORING_OTHER_HOSTS FALSE
#define DEFAULT_NETWORK_ENABLED              TRUE

/* Definition of restart reasons - why a particular server was restarted. */
enum LastRestartReason {
    RESTARTED_BY_USER,
    STOPPED_BY_USER,
    CPU_TOO_HIGH,
    MEMORY_TOO_HIGH,
    LINK_TIMEOUT,
    UNKNOWN,
    NONE
};


/* Definition of the process. It contains all process configuration as well as all statuses and measurements of the process. */
typedef struct
{
    char *id; //the id of the process
    char *alias; //the alias of the process
    char *name; //the name of the executable process
    boolean is_service; //flag indicating that this process is a service
    boolean show_console_window; //flag indicating if the console window should be shown or not
    char *description; //the description of the process
    char *path_to_executable; //the path to the executable process
    char *working_directory; // the working directory of the process
    char **individual_parameters; //individual parameters. same as parameters but split into individual parts
    unsigned int number_of_parameters; //number of parameters
    char **preceding_process; //the processes that have to be started before this process can be started
    unsigned int number_of_preceding_processes; //the number of processes
    char **environment_variables; //the environment variables to be used when starting the process
    unsigned int number_of_environment_variables; //the number of environment variables
    int start_after_n_seconds; //the time in seconds to wait before start/continue monitoring the process
    int number_of_retries; //if determined that process is not running try to restart it this many times before giving up
    int retry_interval; //the time in seconds between two consecutive retries to start the process
    enum AutoStart auto_start; //should the process be automatically started by watchdog if it is determined that it is not running
    int max_cpu_load; //the maximum allowed CPU load (percentage)
    int number_of_allowed_cpu_load_violations; //number of allowed CPU limit violations before restarting the process
    UINT64 max_memory; //maximum allowed memory consumption in bytes
    int number_of_allowed_memory_violations;//number of allowed memory limit violations before restarting the process
    char *tine_device; //the name of the TINE device to attach to
    char *tine_property; //the name of the TINE property to attach to
    int polling_rate; //the rate at which to attach
    int timeout; //the timeout in seconds after which the process is flagged as 'fishy' if there is no update from the TINE property
    int timeout_restart;//the timeout in seconds after which the process will be restarted if no update is received from the TINE property
    boolean match_path; //a flag indicating that when searching the processes the path should match besides the name
    boolean match_parameters; //a flag indicating that when searching the processes the parameters should match besides the name
    char *parameters; //command line parameters as they are read
    char *caption; //the title of the console if the process represents a console application
    int pause_after_previous_start; //the pause in seconds after the previous process was started

    char *context; //the context extracted from tine_device
    char *server; //the server extracted from tine_device

    int number_of_times_stopped_by_wd; //number of times watchdog stopped the process
    int number_of_times_started_by_wd; //number of times watchdog started the process
    boolean stopped_because_of_memory; //indicating if the process was stopped due to memory consumption too high
    boolean stopped_because_of_cpu; //indicating if the process was stopped due to CPU too high
    boolean stopped_because_of_link; //indicating if the process was stopped due to link timeout
    boolean stopped_by_user; //indicating if the process was stopped by user
    UINT64 memory_at_stop_time; //memory at the time the process was stopped if stopped because of memory
    int cpu_at_stop_time; //cpu at the time the process was stopped because of CPU
    int num_of_cpu_violations_at_stop_time; //number of times cpu was violated
    int num_of_memory_violations_at_stop_time; //number of times memory consumption was violated
    enum LastRestartReason last_restart_reason;//the reason why the process was restarted

    time_t determined_not_running_timestamp; //the timestamp at which the process was determined that it is not running
    time_t last_retry_timestamp; //the time at which we last attempted to restart the process
    int current_retry; //the number of retries to start that have already been made
    int current_cpu_load; //the current CPU load
    UINT64 current_memory_consumption; //the current memory consumption in bytes
    int number_of_cpu_violations; //number of registered over the limit CPU violations
    int number_of_memory_violations;//number of registered over the limit memory consumption violations
    boolean is_running; //a flag indicating if the process was determined that it is running or not
    boolean is_starting; //flag indicating if the process is currently starting
    boolean disabled; //flag indicating if the process was being restarted but we gave up
    int pid;//current PID of the process
    time_t last_tine_response; //the timestamp of the last response from TINE
    time_t last_tine_retry; //the timestamp of the last retry to connect to TINE
    int tine_status;//the last status of the TINE property link
    int link_id; //the id of the link to TINE property
    boolean property_disabled;//a flag indicating if connection to property is disabled due to connection error
    boolean preconditions_started; //a flag indicating that preconditions were started in the last cycle
    boolean property_link_timeout;//a flag indicating the property link timeout has ocurred
    boolean process_restart_requested;//a flag indicating that a restart of this process was requested via property
    time_t pid_last_refreshed; //the time at which the pid was refreshed for the last time
    time_t process_requested_to_stop; //Time when the process was requested to stop. If 0 it is irrelevant and the stop process has been dealt with already.
    boolean has_alarm; // indicates if the process alarm has been set
    boolean default_property_info_used; //indicates if the default property info was used (CF_DEFAULT + 128 bytes) to retrieve the monitored property info
    DTYPE dout;
    boolean dout_allocated_by_me; //indicates if the dout pointer was allocated by the watchdog in order to know whether the point should be released when process is cleaned up
} WATCHED_PROCESS;

/* Flag indicating if the watchdog is suspended. When suspended it doesn't start or stop any process. But it keeps monitoring them.*/
extern boolean suspended;
/* The timeout in seconds how long is the watchdog suspended before it is switched back to normal state */
extern int suspended_timeout;
/* The time in seconds after which the PID for the processes are refreshed entirely (matching the name, command line, parmaeters) */
extern int pid_refresh_rate;
/* The period between before kill signal is send if term was not successful */
extern int kill_grace_period;
/* Available physical memory on this machine */
extern UINT64 available_memory;
/* Current CPU usage by all process on this machine */
extern int total_cpu_usage;
/* Pointer to the array of all monitored processes */
extern WATCHED_PROCESS *g_process;
/* Number of all process in the g_process array */
extern unsigned int num_processes;
/* Flag indicates if the watchdog is currently running or not. When the running is false, the watchdog should aim to exit gracefully as soon as possible. */
extern volatile boolean running;
/* The time when the watchdog was suspended */
extern time_t made_suspended;
/* The remaining suspended time */
extern int remaining_suspended_time;
/* Is watchdog allowed to create links to properties of servers running on a different host */
extern boolean allow_monitoring_other_hosts;
/* The command to use when starting application on linux via console*/
extern char *console_command;
/* The CPU usage by the watchdog server itself. Be aware that this is the average consumption over the last cycle. */
extern int watchdog_cpu_usage;
/* The memory consumtpion by the watchdog server itself */
extern UINT64 watchdog_memory_consumption;
/* Stop all watched processes when watchdog is stopped gracefully */
extern boolean stop_all_on_shutdown;
/* Short term history in seconds */
extern int short_term_history_seconds;
/* Long term history in months */
extern int long_term_history_months;
/* Environment variables defined in the configuration file */
extern char** set_environment_variables;
/* Number of environment variables defined in the configuration file */
extern unsigned int set_environment_variables_length;
/* Names of all available network interfaces */
extern char **network_interface_names;
/* Friendly names of all available network interfaces */
extern char **network_interface_friendly_names;
/* Number of network interfaces (length of the above array) */
extern unsigned int num_network_interfaces;
/* Network utilization in percentage */
extern float *network_utilization_percentage;
/* Network utilization in kB/s */
extern float *network_utilization_absolute;
/* Watchdog startup time */
extern time_t startup_time;
/* Indicates if network readings are enabled or disabled */
extern boolean g_network_enabled;
/* The path on which the tine alarm will check disk size */
extern char *minimum_free_disk_path;
/* The minimum free disk size at which an alarm will be triggered */
extern float minimum_free_disk_mbyte;
/**
* \brief Starts the watched process and stores the pid and other info
*
* Attempts to start the process and stores relevant information into the process structure.
* If the attempt was successful, the pid has to have a proper value, process should no
* longer be disabled and the last retry timestamp and number of retries should be set
* accordingly.
*
* \param process the process to start
* \return 0 if the process was started successfully or error code otherwise
*/
int StartWatchedProcess(WATCHED_PROCESS *process);
/**
* \brief Stops the watched process
*
* Attempts to stop the process and resets the PID of the process. This method should
* not change any of the properties of the process, except setting the PID to a negative
* value if the stop was successful.
*
* \param process the process to stop
* \param stopTine the flag indicating if the link to the TINE property should be closed (1) or not (0)
* \param kill a flag indicating if the kill or terminate signal should be sent
* \return 0 if process was requested to stop, -1 if not attempted to stop (already stopped or in process of stopping), 
*       system error in case of failed stop attempt        
*/
int StopWatchedProcess(WATCHED_PROCESS *process, boolean stopTineLink, boolean kill);
/**
* \brief Prints the dynamic info of the process.
*
* Prints out all the dynamic data of a process. This includes the CPU consumption,
* memory consumption, running flag etc.
*
* \param process the pointer to the process
*/
void PrintProcessStatus(WATCHED_PROCESS process);
/**
* \brief Gracefully reboots computer
*
* Reboots the local computer and allows a clean boot.
*
* \return 0 or nothing if successful, or error code in the case of an error
*/
int RebootComputer(void);
/**
* \brief Free all allocated memory
*
* Frees all memory allocated by the watchdog. This is mainly the memory used to
* store the process information and status. This method should be called when
* the watchdog is stopped by external command (e.g. through server).
*/
void Shutdown(void);
/**
* \brief Makes a single monitor cycle over all processes.
*
* Checks the properties and behavior of all processes and takes
* appropriate actions (measures CPU and memory, start or stop process, check
* TINE properties etc.). The method does one pass over all the processes
* and returns.
*/
void Watch(void);
/**
 * \brief Reloads all processes from the configuration file
 * 
 * Release all resources allocated by individual process. This means
 * release of all settings and variables. After that reload the settings
 * from the original configuratio file. Before calling this routine,
 * make sure that nobody is still using any of the processes.
 */
void ReloadProcesses(void);
/**
 * \brief Rewrites the configuration using the current process settings
 * 
 * Writes the current process settings and writes them into the configuration file. 
 *
 * \return 0 if the write was successful or the error code otherwise
 */
int RewriteConfiguration(void);
/**
* \brief Free all resources. This routine should be called at the very end.
*/
void FreeResources(void);

#endif /* WATCHDOG_H */
