/*
* wctrl.c
*
* Watchdog client application which provides means to read data
* from or execute commands on any watchdog server.
*
*  Created on: Feb 5, 2014
*      Author: Jaka Bobnar
*/
#define LOG_FILE wctrl
#include <math.h>
#include <inttypes.h>
#include "serverdef.h"
#include "tine.h"
#include "util.h"

#define HOST_NAME_MAX 64
#define DATA_BUFFER_MAX 64
#define DEVICE_SIZE_MAX 132

#define P_LIST "list"
#define P_LIST_SIMPLE "lists"

NAME64 dev_property_list[NUM_PROPERTIES] = {
    { P_N_INFO }, { P_N_CPU }, { P_N_MEMORY }, { P_N_NUM_CPU_VIOLATIONS }, { P_N_NUM_MEMORY_VIOLATIONS },
    { P_N_NOT_RUNNING_SINCE }, { P_N_LAST_RETRY }, { P_N_CURRENT_RETRY }, { P_N_STATUS }, { P_N_START }, { P_N_RESTART },
    { P_N_START_TINE }, { P_N_STOP_TINE }, { P_N_STOP }, { P_N_LAST_TINE_RESPONSE }, { P_N_LAST_TINE_RETRY }, { P_N_TINE_STATUS },
    { P_N_NUM_STOPPED_WD }, { P_N_NUM_RESTARTED } };
NAME64 wd_property_list[NUM_WD_PROPERTIES] = {
    { P_N_SUSPEND }, { P_N_CPU_TOTAL }, { P_N_AVAILABLE_MEMORY }, { P_N_SHUTDOWN }, { P_N_REBOOT }, { P_N_REMAINING_SUSPENDED },
    { P_N_RELOAD }, { P_N_CPU }, { P_N_MEMORY }, { P_N_NETWORK_NAMES }, { P_N_NETWORK_UTILIZATION } };

#define INTERESTING_PROPS_SIZE 6
NAME64 interesting_props[INTERESTING_PROPS_SIZE] = { { P_N_CPU }, { P_N_MEMORY }, { P_N_STATUS }, { P_N_NUM_STOPPED_WD },
{ P_N_NUM_RESTARTED }, { P_N_LAST_RETRY } };

#if defined(UNIX)

/**
* \brief Reads the hostname of this computer, prepends WD_ to it and stores that to the server parameter.
*/
void getServerName(char *server)
{
    char name[HOST_NAME_MAX];
    int length = HOST_NAME_MAX;
    if (gethostname(name, length) == 0)
    {
        length = strlen(name) + 3;
        strcpy(server, "WD_");
        strcpy(&server[3], name);
        server[length] = '\0';
    }
    else
    {
        printf("Could not read the computer name.\n");
        memcpy(server, "UNKNOWN", 7);
        server[7] = '\0';
    }
}

#elif defined(WIN32)

#include <windows.h>

#define ERROR_MESSAGE_SIZE 100

/**
* \brief Reads the hostname of this computer, prepends WD_ to it and stores that to the server parameter.
*/
void getServerName(char *server)
{
    char error[ERROR_MESSAGE_SIZE];
    size_t i, l;
    int err;
    wchar_t name[HOST_NAME_MAX];
    DWORD length = HOST_NAME_MAX;
    if (GetComputerName(name, &length))
    {
        l = sizeof(char)* (length + 4);
        strcpy_s(server, l, "WD_");
        wcstombs_s(&i, &server[3], l, name, l - 3);
        server[l - 1] = '\0';
    }
    else
    {
        err = (int)GetLastError();
        strerror_s(error, ERROR_MESSAGE_SIZE, err);
        printf("Could not read the machine name. Error (%d): %s.\n", err, error);
    }
}
#endif

int isGroupedWatchdog(char *context, char *server,NAME32 *mlist, int *num)
{
  ExpDataStruct srvaddr;
  FecAddrStruct fecaddr;
  if (GetServerAddr(context, server, &srvaddr, &fecaddr) != 0) return FALSE;
  if (strcmp(fecaddr.fecName, "GENS") != 0) return FALSE;
  if (GetSystemGroupMembers(context, server, mlist, num) != 0) return FALSE;
  return TRUE;
}

/**
* \brief Checks if property is one of the watchdog commands and returns 1 if yes, or 0 if read only.
*/
int isProcessWriteProperty(char *prop)
{
    if (_stricmp(prop, P_N_START) == 0 || _stricmp(prop, P_N_STOP) == 0
        || _stricmp(prop, P_N_START_TINE) == 0 || _stricmp(prop, P_N_STOP_TINE) == 0
        || _stricmp(prop, P_N_RESTART) == 0 || _stricmp(prop, P_N_RELOAD) == 0)
        return 1;
    return 0;
}

/**
* \brief Register the WDProcessInfo
*/
void registerStructs(int count)
{
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, alias), 64, CF_TEXT, "alias")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, name), 64, CF_TEXT, "name")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, is_service), 1, CF_SHORT, "isService")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, path), 256, CF_TEXT, "path")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, parameters), 512, CF_TEXT, "parameters")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, match_parameters), 1, CF_SHORT, "matchParameters")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, match_path), 1, CF_SHORT, "matchPath")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, working_directory), 256, CF_TEXT, "workingDir")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, caption), 256, CF_TEXT, "caption")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, preceding_processes), 512, CF_TEXT, "precedingProc")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, environment), 512, CF_TEXT, "envVariables")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, monitor_n_seconds_after_start), 1, CF_SHORT, "monitorAfter")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, pause_n_seconds_after_start), 1, CF_SHORT, "pauseAfter")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, number_of_retries), 1, CF_SHORT, "numOfRetries")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, retry_interval), 1, CF_SHORT, "retryInterval")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, auto_start), 1, CF_SHORT, "autoStart")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, max_cpu_load), 1, CF_SHORT, "maxCPULoad")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, number_of_allowed_cpu_load_violations), 1, CF_SHORT, "nCPUViolations")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, max_memory), 1, CF_FLOAT, "maxMemoryInMB")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, number_of_allowed_memory_violations), 1, CF_SHORT, "nMemViolations")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, tine_device), 132, CF_TEXT, "tineDevice")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, tine_property), 64, CF_TEXT, "tineProperty")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, polling_rate), 1, CF_SHORT, "pollingRate")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, timeout), 1, CF_SHORT, "timeout")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, timeout_restart), 1, CF_INT32, "restartTimeout")) OutOfMemoryExit(2);
    if (sealTaggedStruct(WD_PROCESS_INFO, sizeof(WDProcessInfo), count)) OutOfMemoryExit(2);
}

/**
* \brief Free the memory referenced by the data.
*/
void freeData(DTYPE *data)
{
    switch (data->dFormat)
    {
    case (CF_BITFIELD8) :
        free(data->data.bptr);
        break;
    case (CF_SHORT) :
        free(data->data.sptr);
        break;
    case (CF_DOUBLE) :
        free(data->data.dptr);
        break;
    case (CF_INT64) :
        free(data->data.llptr);
        break;
    case (CF_FLOAT) :
        free(data->data.fptr);
        break;
    case (CF_STRING) :
        free(data->data.strptr);
        break;
    case CF_STRUCT:
        free(data->data.vptr);
        break;
    }
}

/**
* \brief Initialize the data according to its format and array length.
*/
void initData(DTYPE *data)
{
    switch (data->dFormat)
    {
    case (CF_BITFIELD8) :
        data->data.bptr = (UINT8*)malloc(sizeof(char)*(data->dArrayLength));
        break;
    case (CF_SHORT) :
        data->data.sptr = (short*)malloc(sizeof(short)*(data->dArrayLength));
        break;
    case (CF_DOUBLE) :
        data->data.dptr = (double*)malloc(sizeof(double)*(data->dArrayLength));
        break;
    case (CF_FLOAT) :
        data->data.fptr = (float*)malloc(sizeof(float)*(data->dArrayLength));
        break;
    case (CF_INT64) :
        data->data.llptr = (SINT64*)malloc(sizeof(SINT64)*(data->dArrayLength));
        break;
    case (CF_STRING) :
        data->data.strptr = (char**)malloc(sizeof(char*)*(data->dArrayLength));
        break;
    case CF_STRUCT:
        data->data.vptr = (WDProcessInfo*)malloc(sizeof(WDProcessInfo)* (data->dArrayLength));
        strcpy_s(data->dTag, TAG_NAME_SIZE, WD_PROCESS_INFO);
        registerStructs(data->dArrayLength);
        break;
    }
}

/**
* \brief Converts the data to string and stores it into bfr parameter.
*        Method return 0 if the data is already printed, or 1 if it is copied to the buffer.
*/
int dataToStr(char *bfr, DTYPE *data, int idx, char *prop)
{
    int v, i;
    time_t time;
    char errorstr[128];
    char bits[9];
    WDProcessInfo *info;

    switch (data->dFormat)
    {
    case (CF_BITFIELD8) :
        v = data->data.bptr[idx];
        for (i = 0; i < 8; i++)
        {
            bits[i] = '0' + ((v >> i) & 1);
        }
        bits[8] = '\0';
        sprintf_s(bfr, DATA_BUFFER_MAX, "%s", bits);
        return 1;
    case (CF_SHORT) :
        if (_stricmp(prop, P_N_CPU) == 0 || _stricmp(prop, P_N_CPU_TOTAL) == 0)
        {
            sprintf_s(bfr, DATA_BUFFER_MAX, "%d %%", data->data.sptr[idx]);
        }
        else if (_stricmp(prop, P_N_TINE_STATUS) == 0)
        {
            if (data->data.sptr[idx] == -2)
            {
                sprintf_s(bfr, DATA_BUFFER_MAX, "%s", "Not Monitored");
            } 
            else
            {
                GetLastLinkError(data->data.sptr[idx], errorstr);
                sprintf_s(bfr, DATA_BUFFER_MAX, "%s (%d)", errorstr, data->data.sptr[idx]);
            }
        }
        else
        {
            sprintf_s(bfr, DATA_BUFFER_MAX, "%d", data->data.sptr[idx]);
        }
        return 1;
    case (CF_DOUBLE) :
        if (_stricmp(prop, P_N_LAST_RETRY) == 0 ||
            _stricmp(prop, P_N_LAST_TINE_RESPONSE) == 0 ||
            _stricmp(prop, P_N_LAST_TINE_RETRY) == 0 ||
            _stricmp(prop, P_N_NOT_RUNNING_SINCE) == 0)
        {
            time = (time_t)data->data.dptr[idx];
            if (time == 0)
            {
                sprintf_s(bfr, DATA_BUFFER_MAX, "%s", "N/A");
            }
            else
            {
                ctime_s(bfr, DATA_BUFFER_MAX, &time);
                for (v = DATA_BUFFER_MAX - 1; v >= 0; v--)
                {
                    if (bfr[v] == '\r' || bfr[v] == '\n')
                    {
                        bfr[v] = '\0';
                        break;
                    }
                }
                bfr[v - 5] = '\0';
            }
        }
        else
        {
            sprintf_s(bfr, DATA_BUFFER_MAX, "%3.2f", data->data.dptr[idx]);
        }
        return 1;
    case (CF_INT64) :
        if (_stricmp(prop, P_N_MEMORY) == 0 || _stricmp(prop, P_N_AVAILABLE_MEMORY) == 0)
        {
            sprintf_s(bfr, DATA_BUFFER_MAX, "%3.2f MB", data->data.llptr[idx] / 1024. / 1024.);
        }
        else
        {
            sprintf_s(bfr, DATA_BUFFER_MAX, "%"PRId64, data->data.llptr[idx]);
        }
        return 1;
    case (CF_FLOAT) :
        if (_stricmp(prop, P_N_NETWORK_UTILIZATION) == 0)
        {
            for (i = 0; i < (int)data->dArrayLength; i++)
            {
                printf("%d: %3.2f kB/s\n", i, data->data.fptr[i]);
            }
        }
        else if (_stricmp(prop, P_N_NETWORK_RELATIVE) == 0)
        {
            for (i = 0; i < (int)data->dArrayLength; i++)
            {
                printf("%d: %3.2f %%\n", i, data->data.fptr[i]);
            }
        }
        return 0;
    case (CF_STRING) :
        if (_stricmp(prop, P_N_NETWORK_NAMES) == 0)
        {
            for (i = 0; i < (int)data->dArrayLength; i++)
            {
                printf("%d: %s\n", i, data->data.strptr[i]);
            }
        }
        return 0;
    case (CF_STRUCT) :
        info = data->data.vptr;
        printf("Alias:                     %s\n", info[idx].alias);
        printf("Name:                      %s\n", info[idx].name);
        printf("Is Service:                %s\n", info[idx].is_service ? "Yes" : "No");
        printf("Path:                      %s\n", info[idx].path);
        printf("Parameters:                %s\n", info[idx].parameters);
        printf("Match Path:                %s\n", info[idx].match_path ? "Yes" : "No");
        printf("Match Parameters:          %s\n", info[idx].match_parameters ? "Yes" : "No");
        printf("Working Directory:         %s\n", info[idx].working_directory);
        printf("Caption:                   %s\n", info[idx].caption);
        printf("Preceding Processes:       %s\n", info[idx].preceding_processes);
        printf("Environment:               %s\n", info[idx].environment);
        printf("Monitor N Sec After Start: %d s\n", info[idx].monitor_n_seconds_after_start);
        printf("Pause N Sec After Start:   %d s\n", info[idx].pause_n_seconds_after_start);
        printf("Number of Retries:         %d\n", info[idx].number_of_retries);
        printf("Retry Interval:            %d\n", info[idx].retry_interval);
        printf("Auto Start:                %s\n", info[idx].auto_start ? (info[idx].auto_start == 1 ? "Yes" : "Once") : "No");
        printf("Max CPU Load:              %d %%\n", info[idx].max_cpu_load);
        printf("No. Allowed CPU Violat.:   %d\n", info[idx].number_of_allowed_cpu_load_violations);
        printf("Max Memory:                %3.2f MB\n", info[idx].max_memory);
        printf("No. Allowed Mem. Violat.:  %d\n", info[idx].number_of_allowed_memory_violations);
        printf("TINE Device:               %s\n", info[idx].tine_device);
        printf("TINE Property:             %s\n", info[idx].tine_property);
        printf("Polling Rate:              %d ms\n", info[idx].polling_rate);
        printf("Timeout:                   %d s\n", info[idx].timeout);
        printf("Restart Timeout:           %d s\n", info[idx].timeout_restart);
        printf("\n");
        return 0;
    }
    return 0;
}

/**
* \brief Report a TINE error.
*/
void reportError(int cc)
{
    char errstr[128];
    if (cc != 0)
    {
        GetLastLinkError(cc, errstr);
        printf("ERROR: %s (%d)\n", errstr, cc);
    }
}

/**
* \brief Execute read of a single property for one process.
*/
int printProperty(char *device, char *prop)
{
    int cc, fmt[1], siz[1];
    char bfr[DATA_BUFFER_MAX];
    DTYPE dout;

    cc = GetTargetPropertyInformation(device, prop, fmt, siz, NULL);
    if (cc != 0)
    {
        printf("Could not find %s/%s\n", device, prop);
        return -1;
    }
    dout.dFormat = (short)(*fmt);
    if (strcmp(prop, P_N_NETWORK_NAMES) == 0 || strcmp(prop, P_N_NETWORK_UTILIZATION) == 0)
        dout.dArrayLength = siz[0];
    else
        dout.dArrayLength = 1;
    initData(&dout);
    cc = ExecLink(device, prop, &dout, NULL, CA_READ);
    if (cc == 0)
    {
        if (dataToStr(bfr, &dout, 0, prop))
            printf("%s: %s\n", prop, bfr);
    }
    else
    {
        printf("%s", prop);
        reportError(cc);
    }
    freeData(&dout);
    return cc;
}

char** readProperty(char *device, char *prop, int num_processes)
{
    int cc, i, fmt[1], siz[1];
    DTYPE dout;
    char **ret;

    cc = GetTargetPropertyInformation(device, prop, fmt, siz, NULL);
    if (cc != 0)
    {
        printf("Could not find %s/%s\n", device, prop);
        return '\0';
    }
    dout.dFormat = (short)(*fmt);
    dout.dArrayLength = siz[0];
    initData(&dout);
    cc = ExecLink(device, prop, &dout, NULL, CA_READ);
    if (cc == 0)
    {
        ret = (char**)malloc(sizeof(char*)*num_processes);
        if (ret == NULL) OutOfMemoryExit(1);
        for (i = 0; i < num_processes; i++)
        {
            ret[i] = (char*)malloc(sizeof(char)* DATA_BUFFER_MAX);
            if (ret[i] == NULL) OutOfMemoryExit(1);
            dataToStr(ret[i], &dout, i, prop);
        }
    }
    else
    {
        printf("%s", prop);
        reportError(cc);
        ret = '\0';
    }
    freeData(&dout);
    return ret;
}

int getAll(char *context, char *server)
{
    int i, j, k, num, cc;
    char device[DEVICE_SIZE_MAX];
    NAME64 *dList;
    char ***val;

    printf("%s/%s\n", context, server);
    num = 500;
    dList = (NAME64*)calloc(num,sizeof(NAME64));
    if (dList == NULL) OutOfMemoryExit(1);
    sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s", context, server);
    cc = GetSystemDevices(device, NULL, dList, &num);
    val = (char***)malloc(sizeof(char**)* INTERESTING_PROPS_SIZE);
    if (val == NULL) OutOfMemoryExit(1);
    if (cc == 0)
    {
        sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s/%s", context, server, dList[1].name);
        for (j = 0; j < INTERESTING_PROPS_SIZE; j++)
        {
            val[j] = readProperty(device, interesting_props[j].name, num - 1);
        }
    }
    else
    {
        printf("Cannot read the processes from %s/%s. Error: %d\n", context, server, cc);
        free(val);
        free(dList);
        return 0;
    }

    j = 7;
    k = 6;
    for (i = 0; i < num - 1; i++)
    {
        cc = (int)strlen(dList[i + 1].name);
        if (cc > j) j = cc;
        cc = (int)strlen(val[1][i]);
        if (cc > k) k = cc;
    }

    cc = (k % 2 == 1) ? (k - 6) / 2 : (k - 6) / 2;

    if (j > 16)
    {
        j = 16;
        if (num - 1 > 10)
        {
            printf(" #: %*.*s | MON. | RUN. | CPU  | %*s%*s | #SP | #ST | LAST_RETRY \n", -j, j, "PROCESS", cc, "", cc - k, "MEMORY");
        }
        else
        {
            printf("#: %*.*s | MON | RUN | CPU  | %*s%*s | #SP | #ST | LAST ST. RETRY \n", -j, j, "PROCESS", cc, "", cc - k, "MEMORY");
        }

        for (i = 0; i < num - 1; i++)
        {
            if (num - 1 > 10)
            {
                printf("%02d: %*.*s | %s | %s | %*s | %*s | %*s | %*s | %s\n", i + 1,
                    -j, j, dList[i + 1].name, val[2][i][7] == '0' ? "YES" : "NO ",
                    val[2][i][0] == '1' ? "YES" : "NO ", -4, val[0][i], k, val[1][i], 3, val[3][i],
                    3, val[4][i], strlen(val[5][i]) > 4 ? &val[5][i][4] : val[5][i]);
            }
            else
            {
                printf("%d: %*.*s | %s | %s | %*s | %*s | %*s | %*s | %s\n", i + 1,
                    -j, j, dList[i + 1].name, val[2][i][7] == '0' ? "YES" : "NO ",
                    val[2][i][0] == '1' ? "YES" : "NO ", -4, val[0][i], k, val[1][i], 3, val[3][i],
                    3, val[4][i], strlen(val[5][i]) > 4 ? &val[5][i][4] : val[5][i]);
            }
        }
    }
    else
    {
        if (num - 1 > 10)
        {
            printf(" #: %*s | MON. | RUN. | CPU  | %*s%*s | #SP | #ST | LAST_RETRY \n", -j, "PROCESS", cc, "", cc - k, "MEMORY");
        }
        else
        {
            printf("#: %*s | MON | RUN | CPU  | %*s%*s | #SP | #ST | LAST ST. RETRY \n", -j, "PROCESS", cc, "", cc - k, "MEMORY");
        }
        for (i = 0; i < num - 1; i++)
        {
            if (num - 1 > 10)
            {
                printf("%02d: %*s | %s | %s | %*s | %*s | %*s | %*s | %s\n", i + 1,
                    -j, dList[i + 1].name, val[2][i][7] == '0' ? "YES" : "NO ",
                    val[2][i][0] == '1' ? "YES" : "NO ", -4, val[0][i], k, val[1][i], 3, val[3][i],
                    3, val[4][i], strlen(val[5][i]) > 4 ? &val[5][i][4] : val[5][i]);

            }
            else
            {
                printf("%d: %*s | %s | %s | %*s | %*s | %*s | %*s | %s\n", i + 1,
                    -j, dList[i + 1].name, val[2][i][7] == '0' ? "YES" : "NO ",
                    val[2][i][0] == '1' ? "YES" : "NO ", -4, val[0][i], k, val[1][i], 3, val[3][i],
                    3, val[4][i], strlen(val[5][i]) > 4 ? &val[5][i][4] : val[5][i]);

            }
        }
    }


    free(dList);
    for (i = 0; i < INTERESTING_PROPS_SIZE; i++)
    {
        freeArray(val[i], num - 1);
    }
    free(val);
    return 0;
}

/**
* \brief Read all properties of a single process.
*/
int getAllForProcess(char *context, char *server, char *process)
{
    int i, num, cc;
    char device[DEVICE_SIZE_MAX];
    NAME64 *dList;

    NAME32 mlist[10]; /* draw the line at 10 */
    int nsrvs = 10;
    if (isGroupedWatchdog(context, server,mlist,&nsrvs))
    {
      for (i = 0; i < nsrvs; i++)
      {
        getAllForProcess(context, mlist[i].name, process);
      }
      return 0;
    }

    if (_stricmp(process, P_LIST) == 0)
    {
        return getAll(context, server);
    }
    else if (_stricmp(process, P_LIST_SIMPLE) == 0)
    {
        num = 500;
        dList = (NAME64*)malloc(sizeof(NAME64)*num);
        if (dList == NULL) OutOfMemoryExit(1);
        sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s", context, server);
        cc = GetSystemDevices(device, NULL, dList, &num);
        if (cc == 0)
        {
            for (i = 0; i < num; i++)
            {
                printf("%s\n", dList[i].name);
            }
        }
        else
        {
            reportError(cc);
        }
        free(dList);
        return 0;
    }

    sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s/%s", context, server, process);

    if (_stricmp(server, process) == 0)
    {
        for (i = 0; i < NUM_WD_PROPERTIES; i++)
        {
            if (strcmp(wd_property_list[i].name, P_N_SHUTDOWN) == 0 ||
                strcmp(wd_property_list[i].name, P_N_REBOOT) == 0 ||
                strcmp(wd_property_list[i].name, P_N_RELOAD) == 0)
            {
                continue;
            }
            if (printProperty(device, wd_property_list[i].name) == -1)
            {
                return 1;
            }
        }
    }
    else
    {
        for (i = 0; i < NUM_PROPERTIES; i++)
        {
            if (strcmp(dev_property_list[i].name, P_N_INFO) == 0 ||
                strcmp(dev_property_list[i].name, P_N_RESTART) == 0 ||
                strcmp(dev_property_list[i].name, P_N_START) == 0 ||
                strcmp(dev_property_list[i].name, P_N_STOP) == 0 ||
                strcmp(dev_property_list[i].name, P_N_START_TINE) == 0 ||
                strcmp(dev_property_list[i].name, P_N_STOP_TINE) == 0)
            {
                continue;
            }
            if (printProperty(device, dev_property_list[i].name) == -1)
            {
                return 1;
            }
        }
    }

    return 0;
}

/**
* \brief Executed reboot or shutdown commands, which require arming and execution.
*/
int rebootOrShutdown(char *device, char *prop)
{
    int cc, data[1];
    DTYPE dout;

    dout.dArrayLength = 1;
    dout.dFormat = CF_SHORT;
    data[0] = 1;
    dout.data.lptr = data;
    cc = ExecLinkEx(device, prop, NULL, &dout, CA_WRITE, 1000);
    if (cc == 0)
    {
        data[0] = 2;
        cc = ExecLinkEx(device, prop, NULL, &dout, CA_WRITE, 1000);
    }
    return cc;
}

/**
* \brief Read one property for all devices for which it is applicable.
*/
int getAllForProperty(char *context, char *server, char *prop)
{
    int cc, fmt[1], siz[1], num, i, wd_prop;
    char databfr[DATA_BUFFER_MAX];
    DTYPE dout;
    char device[DEVICE_SIZE_MAX], srv[DEVICE_SIZE_MAX];
    NAME16 *dList;

    dList = NULL;
    wd_prop = 0; //1 is watchdog property only, 2 is process property only, 3 is joint property, 0 does not exist
    for (i = 0; i < NUM_WD_PROPERTIES; i++)
    {
        if (_stricmp(wd_property_list[i].name, prop) == 0)
        {
            wd_prop = 1;
            break;
        }
    }
    if (wd_prop == 0)
    {
        //if not a watchdog property, check if it is a device property
        for (i = 0; i < NUM_PROPERTIES; i++)
        {
            if (_stricmp(dev_property_list[i].name, prop) == 0)
            {
                wd_prop = 2;
                break;
            }
        }
    }
    else if (wd_prop == 1)
    {
        for (i = 0; i < NUM_PROPERTIES; i++)
        {
            if (_stricmp(dev_property_list[i].name, prop) == 0)
            {
                wd_prop = 3;
                break;
            }
        }
    }


    if (wd_prop == 0)
    {
        printf("Property %s is not valid.\n", prop);
        return 1;
    }
    else if (wd_prop == 1 || wd_prop == 3)
    {
        sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s/%s", context, server, "#0");
    }
    else if (wd_prop == 2)
    {
        sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s/%s", context, server, "#1");
    }

    if (isProcessWriteProperty(prop))
    {
        sprintf_s(device, 128, "/%s/%s/%s", context, server, "ALL");
        cc = ExecLinkEx(device, prop, NULL, NULL, CA_WRITE, 1000);
    }
    else if (_stricmp(prop, P_N_REBOOT) == 0 || _stricmp(prop, P_N_SHUTDOWN) == 0)
    {
        cc = rebootOrShutdown(device, prop);
    }
    else
    {
        sprintf_s(srv, DEVICE_SIZE_MAX, "/%s/%s", context, server);
        num = 500;
        dList = (NAME16*)malloc(sizeof(NAME16)*num);
        if (dList == NULL) OutOfMemoryExit(1);
        cc = GetDeviceNames(srv, dList, &num);

        GetTargetPropertyInformation(device, prop, fmt, siz, NULL);
        dout.dFormat = (short)(*fmt);
        dout.dArrayLength = siz[0];
        dout.dTag[0] = 0;

        initData(&dout);
        cc = ExecLinkEx(device, prop, &dout, NULL, CA_READ, 1000);
        if (cc == 0)
        {
            if (wd_prop == 1)
            {
                if (dataToStr(databfr, &dout, 0, prop))
                    printf("%s: %s\n", dList[0].name, databfr);
            }
            if (wd_prop == 3)
            {
                for (i = 0; i < num; i++)
                {
                    if (dataToStr(databfr, &dout, i, prop))
                        printf("%s: %s\n", dList[i].name, databfr);
                }
            }
            else if (wd_prop == 2)
            {
                for (i = 1; i < num; i++)
                {
                    if (dataToStr(databfr, &dout, i - 1, prop))
                        printf("%s: %s\n", dList[i].name, databfr);
                }
            }
        }
        freeData(&dout);
    }
    reportError(cc);
    if (dList != NULL) free(dList);
    return 0;
}

/**
* \bried Read one specific property.
*/
int getForProcessAndProperty(char *context, char *server, char *process, char *prop)
{
    int cc;
    char device[DEVICE_SIZE_MAX];
    sprintf_s(device, DEVICE_SIZE_MAX, "/%s/%s/%s", context, server, process);

    if (isProcessWriteProperty(prop))
    {
        cc = ExecLinkEx(device, prop, NULL, NULL, CA_WRITE, 1000);
    }
    else if (_stricmp(prop, P_N_REBOOT) == 0 || _stricmp(prop, P_N_SHUTDOWN) == 0)
    {
        if (_stricmp(process, server) != 0)
        {
            printf("Command %s can only be executed on the server.\n", prop);
            return 1;
        }
        else
        {
            cc = rebootOrShutdown(device, prop);
        }
    }
    else {
        printProperty(device, prop);
        return 0;
    }
    reportError(cc);
    return 0;
}

void printGoodies()
{
    printf("Below is the summary of the most common things that wctrl can do.\n");
    printf("For more detailed instructions type 'wctrl -help'\n\n");

    printf("  wctrl list     \t List all watched processes and basic readings\n");

    printf("  wctrl all stop \t Stop all watched processes\n");
    printf("                 \t (will not be started until requested)\n");
    printf("  wctrl all start \t Start all watched processes\n");
    printf("  wctrl all restart \t Restart all watched processes\n");
    printf("  wctrl <proc> status \t Reports the status of the specified process\n");
    printf("  wctrl <proc> cpu \t CPU usage of the specified processes\n");
    printf("  wctrl /p=memory \t memory consumption of all processes\n");
    printf("  wctrl /p=cputotal \t CPU consumption of the computer running watchdog\n\n");

    printf("Other available properties which can be used either as\n");
    printf("'wctrl <proc> <property>' or 'wctrl /p=<property>' are:\n\n");
    printf("info, cpu, memory, ncpuviolations, nmemoryviolations, notrunningsince,\n");
    printf("lastretry, currentretry, status, lasttineretry, lasttineresponse, tinestatus,\n");
    printf("start, stop, restart, starttinemonitor, stoptinemonitor.\n\n");

    printf("The following properties can only be used with the /p=<property> option:\n");
    printf("cputotal, availablememory, reload.\n");
}

void printHelp()
{
    char server[HOST_NAME_MAX];
    getServerName(server);
    printf("Use this program in the following way:\n\n");
    printf("wctrl.exe [options] [process] [property]\n\n");
    printf("The following options are available:\n");
    printf("  /c=CONTEXT \t The name of the context in which to look for the\n");
    printf("\t\t watchdog server(default: SERVICE)\n");
    printf("  /s=SERVER \t The name of the watchdog server (default: WD_<host> = %s)\n", server);
    printf("  /p=PROPERTY \t The name of the property to read/write (the same\n");
    printf("\t\t property of all devices / processes will be read)\n");
    printf("\t\t If PROPERTY is a process property it will be read for all\n");
    printf("\t\t processes.If it is a server property it will be read\n");
    printf("\t\t for the server only. That same is true for the commands.\n");
    printf("  lists \t\t List all devices of the watchdog server.\n");
    printf("  list \t\t List all devices of the watchdog server with some basic readings.\n");
    printf("  [process] \t The name of the process, which the property will\n\t\t be read for, or command executed on\n");
    printf("  [property] \t the name the property to read or command to execute\n");
    printf("\t\t If [process] is given and [property] is not, all properties\n\t\t (no commands) will be read for that process.\n");
    printf("\t\t If /p=PROPERTY is specified the [process] and [property]\n\t\t parameters are irrelevant and will be ignored.\n");
    printf("\nExamples:\n");
    printf("  wctrl /c=TEST /p=cpu \t Reads the CPU property of all processes\n\t\t\t on the server /TEST/%s\n", server);
    printf("  wctrl /s=WD_Foobar /p=memory \t Reads the Memory property of all\n\t\t\t processes on the server /SERVICE/WD_Foobar\n");
    printf("  wctrl sinegen cpu \t Reads the CPU property of the sinegen process\n\t\t\t on the server /SERVICE/%s\n", server);
    printf("  wctrl sinegen \t Reads all properties of the sinegen process on the\n\t\t\t server /SERVICE/%s\n", server);
    printf("  wctrl /c=TEST /s=WDOG sinegen start \t Executes the start command of\n\t\t\t the sinegen process on /TEST/WDOG\n");
    printf("  wctrl /c=TEST /s=WDOG /p=cpu sinegen memory \t Reads the CPU property\n\t\t\t of all processes on /TEST/WDOG. Sinegen and memory\n\t\t\t parameters are ignored\n");
    printf("  wctrl /c=TEST /s=WDOG list \t List all devices of the server /TEST/WDOG\n");
}

char* getInputLine(void) 
{
    char *ret;
    char line[1024];
    size_t len;
    int i;

    fgets(line, 1024,stdin);
    len = strlen(line);
    for (i = 0; i < (int)len; i++)
    {
        if (line[i] == '\n' || line[i] == '\r')
        {
            line[i] = '\0';
        }
    }
    len = strlen(line);
    ret = (char*)malloc(sizeof(char)*len+1);
    if (ret == NULL) OutOfMemoryExit(5);
    strcpy_s(ret, len+1,line);
    ret[len] = '\0';
    return ret;
}

void printConsoleHelp()
{
    printf(">\n");
    printf("help\n\t\t Print this help\n");
    printf("exit\n\t\t Exit the command mode\n");
    printf("/s=<server>\n\t\t Set a new watchdog server name to connect to\n");
    printf("/c=<context>\n\t\t Set a new context of the watchdog server to connect to\n");
   
    printf("/d=<process>\nprocess=<process>\ndev=<process>\n\t\t Set a new process to connect to (clears the property setting).\n");
    printf("\t\t When a process is set all subsequent commands will be made\n");
    printf("\t\t using this process. When process is set you only have to\n");
    printf("\t\t type the property name to read the value for that process.\n");
    
    printf("/p=<property>\nproperty=<property>\nprop=<property>\n\t\t Set a new property to connect to (clears the process setting).\n");
    printf("\t\t When a property is set all subsequent commands will be made\n");
    printf("\t\t using this property. When property is set you only have\n");
    printf("\t\t to type the process name to read the value of that property.\n");
   
    printf("addr\n\t\t Print the address you are currently connected to\n");
    printf("clear\n\t\t Clear property and process settings\n");
    printf("list\n\t\t List all process and basic readings\n");
    printf("<process> <property>\n\t\t Queries the property or execute command on a specific process.\n\t\t Works only if no process or property is defined yet.\n");
    printf("<process>\n\t\t Reads all properties of the process. Works only if no process\n\t\t or property is defined yet.\n");
    printf("\n");
}

int consoleMode(char *ctx, char *srv)
{
    char *command;
    char **splitCommand;
    int size, ssz, error;
    char *process, *prop, *context, *server;
    size = (int)strlen(ctx);
    context = (char*)malloc(sizeof(char)*size+1);
    strcpy_s(context, size+1, ctx);
    size = (int)strlen(srv);
    server = (char*)malloc(sizeof(char)*size + 1);
    strcpy_s(server, size + 1, srv);
    splitCommand = '\0';
    printf("Entering console mode. Type exit or press CTRL+C to leave.\n");
    printf("Type 'help' to see the list of available commands.\n\n");
    printf("Context: %s\t Server: %s\n",context,server);
    prop = '\0';
    process = '\0';
    while (1)
    {
        printf(">");
        error = 0;
        command = getInputLine();
        if (command == NULL) {
            printf("Something went terribly wrong!\n");
            return 1;
        }
        if (_stricmp(command, "exit") == 0)
        {
            printf("Bye Bye!\n");
            break;
        } 
        else if (_stricmp(command, "help") == 0)
        {
            printConsoleHelp();
        }
        else if (_stricmp(command, "addr") == 0 || _stricmp(command, "a") == 0)
        {   
            if (prop == NULL && process == NULL)
                printf("Context: %s\t Server: %s\n", context, server);
            else if (prop != NULL)
                printf("Context: %s\t Server: %s\t Property (command): %s\n", context, server, prop);
            else if (process != NULL)
                printf("Context: %s\t Server: %s\t Device: %s\n", context, server, process);
        }
        else if (_stricmp(command, "list") == 0)
        {
            error = getAllForProcess(context,server,P_LIST);
        }
        else if (_stricmp(command, "clear") == 0)
        {
            if (prop != NULL)
            {
                free(prop);
                prop = '\0';
            }
            if (process != NULL)
            {
                free(process);
                process = '\0';
            }
        } 
        else
        {
            splitCommand = split(command, &size, '=', 1,0);
            int done = 0;
            if (size == 2)
            {
                if (_stricmp(splitCommand[0], "/p") == 0 || _stricmp(splitCommand[0], "property") == 0 || _stricmp(splitCommand[0], "prop") == 0)
                {
                    if (prop != NULL) free(prop);
                    ssz = (int)(sizeof(char)*(strlen(splitCommand[1]) + 1));
                    prop = malloc(ssz);
                    if (prop == NULL) OutOfMemoryExit(3);
                    strcpy_s(prop, ssz, splitCommand[1]);
                    prop[ssz - 1] = '\0';

                    if (process != NULL)
                    {
                        free(process);
                        process = '\0';
                    }
                    printf("Using property: %s\n", prop);
                    done = 1;
                }
                else if (_stricmp(splitCommand[0], "/d") == 0 || _stricmp(splitCommand[0], "process") == 0 || _stricmp(splitCommand[0], "device") == 0 || _stricmp(splitCommand[0], "dev") == 0)
                {
                    if (process != NULL) free(process);
                    ssz = (int)(sizeof(char)*(strlen(splitCommand[1]) + 1));
                    process = malloc(ssz);
                    if (process == NULL) OutOfMemoryExit(3);
                    strcpy_s(process, ssz, splitCommand[1]);
                    process[ssz - 1] = '\0';
                    if (prop != NULL)
                    {
                        free(prop);
                        prop = '\0';
                    }
                    printf("Using process: %s\n", process);
                    done = 1;
                }
                else if (_stricmp(splitCommand[0], "server") == 0 || _stricmp(splitCommand[0], "/s") == 0)
                {
                    free(server);
                    ssz = (int)(sizeof(char)*(strlen(splitCommand[1]) + 1));
                    server = malloc(ssz);
                    if (server == NULL) OutOfMemoryExit(3);
                    strcpy_s(server, ssz, splitCommand[1]);
                    server[ssz - 1] = '\0';
                    printf("New server is: %s\n", server);
                    done = 1;
                }
                else if (_stricmp(splitCommand[0], "context") == 0 || _stricmp(splitCommand[0], "/c") == 0)
                {
                    free(context);
                    ssz = (int)(sizeof(char)*(strlen(splitCommand[1]) + 1));
                    context = malloc(ssz);
                    if (context == NULL) OutOfMemoryExit(3);
                    strcpy_s(context, ssz, splitCommand[1]);
                    context[ssz - 1] = '\0';
                    printf("New context is: %s\n", context);
                    done = 1;
                }
                else
                {
                    done = 0;
                }
            }
            if (!done)
            {
                freeArray(splitCommand,size);
                splitCommand = split(command, &size, ' ', 1,0);
                if ((prop != NULL || process != NULL) && size == 1 )
                {
                    //one of those is locked
                    if (prop != NULL)
                    {
                        error = getForProcessAndProperty(context, server, splitCommand[0], prop);
                    }
                    else if (process != NULL)
                    {
                        error = getForProcessAndProperty(context, server, process, splitCommand[0]);
                    }
                    else
                    {
                        error = getForProcessAndProperty(context, server, process, prop);
                    }
                }
                else if (size == 2)
                {
                    error = getForProcessAndProperty(context, server, splitCommand[0], splitCommand[1]);
                }
                else if (size == 1)
                {
                    if ((splitCommand[0][0] == 'p' || splitCommand[0][0] == 'P') && splitCommand[0][1] == '=')
                    {
                        error = getAllForProperty(context, server, &splitCommand[0][2]);
                    }
                    else if ((splitCommand[0][0] == 'd' || splitCommand[0][0] == 'D') && splitCommand[0][1] == '=')
                    {
                        error = getAllForProcess(context, server, &splitCommand[0][2]);
                    } 
                    else
                    {
                        error = getAllForProcess(context, server, splitCommand[0]);
                    }
                } 
                else
                {
                    printf("Unknown command\n");
                }
            }
            freeArray(splitCommand,size);
            if (error)
            {
                printf("There was an error during command execution: %d\n",error);
            }
        }
        free(command);
    }
    return 0;
}

//use WCTRL to avoid Eclipse complaining about duplicated main method
#if defined(WCTRL)

int main(int argc, char **argv, char **envp)
{
    char *context, *process, *prop, *shared_prop;
    char server[HOST_NAME_MAX];
    int i, l, error;
    char *c;

    error = 0;
    context = '\0';
    server[0] = '\0';
    process = '\0';
    prop = '\0';
    shared_prop = '\0';
    //start with 1, because 0 is the name of the executable
    for (i = 1; i < argc; i++)
    {
        c = argv[i];
        if (strncmp(c, "/h", 2) == 0 || strncmp(c, "/?", 2) == 0
            || strncmp(c, "-help", 5) == 0 || strncmp(c, "--help", 6) == 0)
        {
            printHelp();
            return 1;
        }
        else if (strncmp(c, "/c=", 3) == 0)
        {
            l = (int)(sizeof(char)*(strlen(&c[3]) + 1));
            context = (char*)malloc(l);
            if (context == NULL) OutOfMemoryExit(2);
            context[l - 1] = '\0';
            strcpy_s(context, l, &c[3]);
        }
        else if (strncmp(c, "/s=", 3) == 0)
        {
            l = (int)(sizeof(char)*(strlen(&c[3]) + 1));
            server[l - 1] = '\0';
            strcpy_s(server, l, &c[3]);
        }
        else if (strncmp(c, "/p=", 3) == 0)
        {
            l = (int)(sizeof(char)*(strlen(&c[3]) + 1));
            shared_prop = (char*)malloc(l);
            if (shared_prop == NULL) OutOfMemoryExit(2);
            shared_prop[l - 1] = '\0';
            strcpy_s(shared_prop, l, &c[3]);
        }
        else
        {
            if (process == NULL)
            {
                l = (int)(sizeof(char)*(strlen(c) + 1));
                process = (char*)malloc(l);
                if (process == NULL) OutOfMemoryExit(2);
                process[l - 1] = '\0';
                strcpy_s(process, l, c);
            }
            else if (prop == NULL)
            {
                l = (int)(sizeof(char)*(strlen(c) + 1));
                prop = (char*)malloc(l);
                if (prop == NULL) OutOfMemoryExit(2);
                prop[l - 1] = '\0';
                strcpy_s(prop, l, c);
            }
        }
    }
    if (server[0] == '\0')
    {
        getServerName(server);
    }
    if (context == NULL)
    {
        int l = sizeof(char)* 8;
        context = (char*)malloc(l);
        if (context == NULL) OutOfMemoryExit(2);
        context[l - 1] = '\0';
        strcpy_s(context, l, "SERVICE");
    }
    if (process == NULL && shared_prop == NULL)
    {
        error = consoleMode(context, server);
    }
    else if (shared_prop != NULL)
    {
        error = getAllForProperty(context, server, shared_prop);
    }
    else if (prop == NULL)
    {
        //process is defined, otherwise the first if would take over
        error = getAllForProcess(context, server, process);
    }
    else
    {
        //process and property are both defined
        error = getForProcessAndProperty(context, server, process, prop);
    }
    if (context != NULL) free(context);
    if (process != NULL) free(process);
    if (prop != NULL) free(prop);
    if (shared_prop != NULL) free(shared_prop);

    if (error)
    {
        printf("\n");
        printGoodies();
    }
    return error;
}
#endif
