/*
 * File almlib.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file almlib.c
 * \brief TINE Local Alarm Server plus API routines
 *
 * The API routines offered here pertain directly to the TINE Local Alarm Server maintained
 * on the running server.  A server can of course run without setting alarms.  However a 
 * control system will suffer if important information relating to the health of the system
 * is surpressed.  One can set a certain category of alarms (threshold alarms) without issuing
 * any API call, but instead supplying a startup configuration file 'almwatch.csv' with 
 * instructions as to which properties to call and into what range the data returned must fall
 * in order to function properly.  Otherwise calls to ClearAlarm() and SetAlarm() for instance
 * should be made inside a server's IO loop where necessary.
 *
 */
#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include "prolog.h"
#include "srvcore.h"
#include "errors.h"
#include "dbtools.h"
#include "almlib.h"
#include "hstlib.h"
#include "tinetype.h"
#include "threader.h"
#include "ttycmd.h"
#include "toolkit.h"
#endif
short ftoi(char *fmt);

TINE_EXPORT int AlarmWindow = DEFAULT_ALARM_WINDOW;
TINE_EXPORT int AlarmHeartbeat = DEFAULT_ALARM_HEARTBEAT;
TINE_EXPORT int AlarmInstantWindow = DEFAULT_INSTANT_WINDOW;
TINE_EXPORT int AlarmDataChangeWindow = DEFAULT_ALARM_DATACHANGE_WINDOW;
TINE_EXPORT int MaxNumberAlarms = MAXALARMS;
TINE_EXPORT int TerminationWindow =  DEFAULT_TERMINATION_WINDOW;
int AlarmServerInitialized = 0;
int AlarmTestState = 0;
int almdbg = 0;
AWSLstEntry *almWatchList = NULL;
int nAlmWatchList = 0;
int gRequireReadFromCAS = FALSE;
int gIsCentralAlarmServer = FALSE;
int gAlmOscillationWindowPinned = FALSE;
int gAlmOscillationWindow = ALM_OSCILLATION_WINDOW;
int gAlmCollapseWindow = ALM_COLLAPSE_WINDOW;
int gMaxAMSCapacity = MAX_AMS_CAPACITY;
void lasClearAlarm(char *eqm,int devNr,UINT32 code,AWSLstEntry *aws);
void touchActiveAlarms(ExportListStruct *el);
#ifndef SMALL_TINE_LIB
int populateADS(char *fecname,char *eqmname,DBLNKLIST *fecList,ADSIS **adsis);
int populateAWS(char *fecname,DBLNKLIST *fecList,AWSIS **awsis);
int populateAlmSettings(char *fecname,DBLNKLIST *fecList,ALMSETIS **asis);
int gAlmMfDumped = FALSE;
#endif

int gAlarmSurveyInterval = DEFAULT_ALARM_SURVEY_INTERVAL;
#define PRE_DUMP_WINDOW 60
#define AMSDynSet_AllowedTerminationPending 20
static int collapsedTerminationPending = FALSE;
static time_t newAlarmBaseTime = 0;

_CSVCOLUMN csvAlarmSettingsColumns[] = 
{
  { "MINIMUM_WINDOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,wnd),NULL,0 },
  { "TRANSIENT_WINDOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,trnWnd),NULL,0 },
  { "DATACHANGE_WINDOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,dcWnd),NULL,0 },
  { "TERMINATE_WINDOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,trmWnd),NULL,0 },
  { "OSCILLATION_WINDOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,oscWnd),NULL,0 },
  { "COLLAPSE_WINDOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,clpWnd),NULL,0 },
  { "HEARTBEAT","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,hbt),NULL,0 },
  { "TABLE_SIZE","",0,CF_INT32,1,(BYTE *)OFFSETIN(ALMSETIS,tblSiz),NULL,0 },
};
CSVDB csvAlarmSettingsDb =
{
  csvAlarmSettingsColumns,
  sizeof(csvAlarmSettingsColumns)/sizeof(csvAlarmSettingsColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ALMSETIS),
  0
};

int getAlmSettings(void)
{
  int cc=0;
# if !defined(VXWORKS)
  ALMSETIS *asis=NULL;

  if (gTineConfigurationCoded) return not_implemented;
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  {
    if ((cc=populateAlmSettings(gFecName,gFecInfoList,&asis)) != 0) ccerr(cc);
  }
  else
# endif
  {
    if ((cc=csvReadFile(fecDbPath,ALM_SETTINGS_FILE,&csvAlarmSettingsDb,(void **)&asis)) != 0) ccerr(cc);
    if (csvAlarmSettingsDb.siz == 0) ccerr(0); 
  }
  if (asis != NULL)
  {
    if (asis->wnd > 0) 
    {
      AlarmWindow = asis->wnd;
      feclog("ALM: minimum window set to %d from file",AlarmWindow);
    }
    if (asis->clpWnd > 0) 
    {
      gAlmCollapseWindow = asis->clpWnd;
      feclog("ALM: collapse window set to %d alarms from file",gAlmCollapseWindow);
    }
    if (asis->dcWnd > 0) 
    {
      AlarmDataChangeWindow = asis->dcWnd;
      feclog("ALM: data-change window set to %d sec from file",AlarmDataChangeWindow);
    }
    if (asis->oscWnd > 0) 
    {
      gAlmOscillationWindow = asis->oscWnd;
      feclog("ALM: oscillation window set to %d clear counts from file",gAlmOscillationWindow);
    }
    if (asis->trmWnd > 0) 
    {
      TerminationWindow = asis->trmWnd;
      feclog("ALM: termination window set to %d sec from file",TerminationWindow);
    }
    if (asis->trnWnd > 0) 
    {
      AlarmInstantWindow = asis->trnWnd;
      feclog("ALM: transient alarm window set to %d sec from file",AlarmInstantWindow);
    }
    if (asis->hbt > 0) 
    {
      AlarmHeartbeat = asis->hbt;
      feclog("ALM: heartbeat set to %d sec from file",AlarmHeartbeat);
    }
    if (asis->tblSiz > 0) 
    {
      MaxNumberAlarms = asis->tblSiz;
      feclog("ALM: max. table size set to %d from file",MaxNumberAlarms);
    }
  }
err:
  SystemFree(asis);
  if (cc == no_such_file) return 0;
  feclog("alarm settings file %s: %.32s",ALM_SETTINGS_FILE, cc ? cc2str(cc) : "found");
# endif
  return cc;
}

int dumpAlmSettings(void)
{
  ttyoutput("Alarm readout from CAS required : %s\n",(gRequireReadFromCAS ? "TRUE" : "FALSE"));
  ttyoutput("Alarm collapse window : %d alarms\n",gAlmCollapseWindow);
  ttyoutput("Alarm osc. window : %d counts\n",gAlmOscillationWindow);
  ttyoutput("Alarm osc. window pinned : %s\n",(gAlmOscillationWindowPinned ? "TRUE" : "FALSE"));
  ttyoutput("Alarm survey interval : %d msec\n",gAlarmSurveyInterval);
  ttyoutput("Alarm new-alarm window : %d sec\n",AlarmWindow);
  ttyoutput("Alarm transient-alarm window : %d sec\n",AlarmInstantWindow);
  ttyoutput("Alarm data-change window : %d sec\n",AlarmDataChangeWindow);
  ttyoutput("Alarm termination window : %d sec\n",TerminationWindow);
  ttyoutput("Alarm heartbeat : %d sec\n",AlarmHeartbeat);
  ttyoutput("Alarm table max. size : %d alarms\n",MaxNumberAlarms);
  ttyoutput("Alarm watch : %d items\n",nAlmWatchList);  
  return 0;
}

#define NEW_ALARM_WINDOW 3

#define idx_too_high      0
#define idx_warn_too_high 1
#define idx_too_low       2
#define idx_warn_too_low  3
#define idx_invalid_data  4

ADSType almErrTblADS[] = 
{
  { "too high",value_too_high,0,0,ALM_SEVERITY_ERR,BFMT(CF_FLOAT),1,ALM_OSCILLATION_WINDOW,0,"measured value is above threshold","","last measured value","" },
  { "near limit",warn_too_high,0,0,ALM_SEVERITY_WARN,BFMT(CF_FLOAT),1,ALM_OSCILLATION_WINDOW,0,"measured value is above warn threshold","","last measured value",""},
  { "too low",value_too_low,0,0,ALM_SEVERITY_ERR,BFMT(CF_FLOAT),1,ALM_OSCILLATION_WINDOW,0,"measured value is below threshold","","last measured value",""},
  { "near limit",warn_too_low,0,0,ALM_SEVERITY_WARN,BFMT(CF_FLOAT),1,ALM_OSCILLATION_WINDOW,0,"measured value is below warn threshold","","last measured value",""},
  { "not valid",invalid_data,0,0,ALM_SEVERITY_WARN,BFMT(CF_INT32),1,ALM_OSCILLATION_WINDOW,0,"measured value not at set point","","last measured value",""},
# ifndef SMALL_TINE_LIB
  { "I/O error",io_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_NULL),0,ALM_OSCILLATION_WINDOW,0,"hardware input/output error","","",""},
  { "File error",file_error,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"file input/output error","","file/path name which caused the alarm",""},
  { "Hardware error",hardware_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"general hardware error","","hardware device description",""},
  { "Network read error",net_read_error,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),0,ALM_OSCILLATION_WINDOW,0,"network read error","","",""},
  { "Network write error",net_write_error,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),0,ALM_OSCILLATION_WINDOW,0,"network write error","","",""},
  { "Link error",link_error,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"link to dependent server is not open","","link which has the problem",""},
  { "Driver not installed",driver_not_installed,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"hardware driver not installed or not initialized","","name of hardware driver",""},
  { "Port not available",port_not_available,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"hardware port not available","","Port Id",""},
  { "Not running",not_running,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_NULL),0,ALM_OSCILLATION_WINDOW,0,"hardware device is not running","","",""},
  { "Sedac error",sedac_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_INT16),3,ALM_OSCILLATION_WINDOW,0,"sedac module has an error","","line/crate/subaddress of defective module",""},
  { "GPIB error",gpib_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"GPIB bus error","","hardware module description",""},
  { "RS232 error",rs232_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"RS232 bus communication error","","hardware module description",""},
  { "Device error",device_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"general device error/device defective","","module description",""},
  { "CAN bus error",canbus_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_INT32),4,ALM_OSCILLATION_WINDOW,0,"CAN module has an error","","address of CAN module",""},
  { "PROFIBUS error",profibus_error,0,ALM_SYSTEM_HARDWARE,ALM_SEVERITY_WARN,BFMT(CF_INT32),4,ALM_OSCILLATION_WINDOW,0,"PROFIBUS module has an error","","address of PROFIBUS module",""},
  { "disk space is low",warn_disk_space,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"disk space is getting low","","disk + available space (blocks)",""},
  { "disk space is very low",low_disk_space,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"disk space is extremely low","","disk + available space (blocks)",""},
  { "low main memory",low_main_memory,0,ALM_SYSTEM_SYSTEM,ALM_SEVERITY_WARN,BFMT(CF_TEXT),64,ALM_OSCILLATION_WINDOW,0,"low available main memory detected","","available vs. threshold (blocks)",""},
# endif
  { "number of alarms too large",max_alarms_exceeded,0,0,ALM_SEVERITY_WARN,BFMT(CF_INT16),32,ALM_OSCILLATION_WINDOW,0,"number of alarms too large for individual reporting","","total number alarms + representative alarm codes",""},
};
int almErrTblADSsize = sizeof(almErrTblADS)/sizeof(almErrTblADS[0]);
ADS *getErrTblADS(short cc)
{
  int i;
  cc = ALM_BASECODE(cc);
  for (i=0; i<almErrTblADSsize; i++)
  {
    if (cc == (short)almErrTblADS[i].alarmCode) return &almErrTblADS[i];
  }
  return NULL;
}
ADS *makeADSFromErrorCode(short cc)
{
  static ADS ads;
  cc = ALM_BASECODE(cc);
  if (cc < 1 || cc >= numErr) return NULL;
  memset(&ads,0,sizeof(ADS));
  ads.alarmCode = cc;
  ads.alarmDataFormat = BFMT(CF_NULL);
  ads.alarmSeverity = ALM_SEVERITY_WARN;
  ads.alarmSystem = 0;
  ads.alarmOscillationWindow = ALM_OSCILLATION_WINDOW;
  strncpy(ads.alarmTag,erlst[cc],ALARM_TAG_SIZE);
  strncpy(ads.alarmText,erlst[cc],ALARM_TAG_SIZE);
  strncpy(ads.deviceText,"registered device",DEVICE_NAME_SIZE);
  return &ads;
}
#ifndef SMALL_TINE_LIB
UINT32 minAvailMemBytes = 0;
TINE_EXPORT void SetAvailableMemoryThreshold(UINT32 minAvailBytes)
{
  minAvailMemBytes = minAvailBytes;
}
void checkAvailMemBytes(void)
{
# if defined(WIN32) || defined(UNIX) || defined(VXWORKS)
  UINT32 memavail;
  ExportListStruct* el = ExportList;
  int code = low_main_memory;
  char txt[64];
  time_t t = time(NULL);
  static time_t interim = 0;

  if (t < interim + 60) return;
  interim = t;
  if (minAvailMemBytes <= 0) return;
  lasClearAlarm(el->EqmName, 0, code, NULL);
  if ((memavail = getAvailableMemoryBytes()) > minAvailMemBytes) return;
  code |= ALM_APPLYNAME;
  sprintf(txt, "%u Kb is less than the %u Kb threshold", memavail/1024, minAvailMemBytes/1024);
  SetAlarmEx(el->EqmName, 0, code, (BYTE*)txt, almNEWALARM);
# endif
}
int gMinDiskSpaceTblSize = 0;
#if defined(WIN32) || defined(UNIX) || defined(MACOS)
MinDiskSpaceTblEntry minDiskSpaceTbl[MAX_MINDISKSPACE_TBL_SIZE];
#endif
TINE_EXPORT int SetFreeBlocksAlarmThreshold(char *path, UINT32 minFreeBlocks)
{
# if defined(WIN32) || defined(UNIX) || defined(MACOS)
  int i;
  for (i=0; i<gMinDiskSpaceTblSize; i++) 
  {
#   ifdef WIN32
    if (!strnicmp(minDiskSpaceTbl[i].path,path,128)) break;
#   else
    if (!strncmp(minDiskSpaceTbl[i].path,path,128)) break;
#   endif
  }
  if (i == gMinDiskSpaceTblSize)
  {
    if (i >= MAX_MINDISKSPACE_TBL_SIZE) return resources_exhausted;
    strncpy(minDiskSpaceTbl[i].path,path,128);
    gMinDiskSpaceTblSize++;
  }
  minDiskSpaceTbl[i].minFreeBlks = minFreeBlocks;
# else
  path = path; minFreeBlocks = minFreeBlocks;
# endif
  return 0;
}
void checkFreeBlocks(void)
{
# if defined(WIN32) || defined(UNIX) || defined(MACOS)
  unsigned long freeblks;
  ExportListStruct *el=ExportList;
  int i, code = warn_disk_space;
  char txt[64];
  time_t t = time(NULL);
  static time_t interim = 0;

  if (t < interim + 60) return;
  interim = t;
  for (i=0; i<gMinDiskSpaceTblSize; i++)
  {
    if (minDiskSpaceTbl[i].minFreeBlks <= 0) continue;
    lasClearAlarm(el->EqmName,0,warn_disk_space,NULL);
    lasClearAlarm(el->EqmName,0,low_disk_space,NULL);
    freeblks = getAvailableDiskSpaceInBlocks(minDiskSpaceTbl[i].path);
    if (freeblks < minDiskSpaceTbl[i].minFreeBlks)
    {
      sprintf(txt,"%.48s %lu kb",minDiskSpaceTbl[i].path,freeblks);
      if (freeblks < minDiskSpaceTbl[i].minFreeBlks/10) code = low_disk_space;
      code = encodeDiskSpaceAlarm(i,code);
      SetAlarmEx(el->EqmName,0,code,(BYTE *)txt,almNEWALARM);
    }
  }
# endif
}
#endif /* !SMALL_TINE_LIB */
int isAllowedAlarmValueFormat(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_BYTE:
    case CF_INT16:
    case CF_INT32:
    case CF_INT64:
    case CF_FLOAT:
    case CF_DOUBLE:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
int isAllowedAlarmDataFormat(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_DEFAULT:
    case CF_STRING:
    case CF_HISTORY:
    case CF_IMAGE: 
    case CF_AIMAGE: 
    case CF_USTRING:
    case CF_SPECTRUM:
    case CF_ASPECTRUM:
    case CF_GSPECTRUM:
    case CF_MDA:
      return FALSE;
    default:
      break;
  }
  return TRUE;
}
_CSVCOLUMN csvAlmWatchFileColumns[] = 
{
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(AWSIS,eqm),NULL, 0 },
  { "DEVICE_NAME","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(AWSIS,dev),NULL, 0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(AWSIS,prp),NULL, 0 },
  { "SIZE","",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,siz),NULL, 0 },
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(AWSIS,alarmDataFormatString),NULL, 0 },
  { "SEVERITY","",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,sev),NULL, 0 },
  { "SEVERITY_HIGH","",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,hisev),NULL, 0 },
  { "SEVERITY_LOW","",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,losev),NULL, 0 },
  { "SEVERITY_HIGHWARN","",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,hiwarnsev),NULL, 0 },
  { "SEVERITY_LOWWARN","",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,lowarnsev),NULL, 0 },
  { "ALARM_SYSTEM","0",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,asys),NULL, 0 },
  { "MASK","0",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,mask),NULL, 0 },
  { "NORMAL","0",0,CF_TEXT,64,(BYTE *)OFFSETIN(AWSIS,normalValueString),NULL, 0 },
  { "COUNT_THRESHOLD","0",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,cntThreshold),NULL, 0 },
  { "HIGH","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,hi),NULL, 0 },
  { "LOW","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,lo),NULL, 0 },
  { "HIGHWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,hiwarn),NULL, 0 },
  { "LOWWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(AWSIS,lowarn),NULL, 0 },
  { "ALARM_CODE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,code),NULL, 0 },
  { "ALARM_CODE_HIGH","0",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,codeHigh),NULL, 0 },
  { "ALARM_CODE_HIGHWARN","0",0,CF_INT32,1,(BYTE*)OFFSETIN(AWSIS,codeHighWarn),NULL, 0 },
  { "ALARM_CODE_LOW","0",0,CF_INT32,1,(BYTE *)OFFSETIN(AWSIS,codeLow),NULL, 0 },
  { "ALARM_CODE_LOWWARN","0",0,CF_INT32,1,(BYTE*)OFFSETIN(AWSIS,codeLowWarn),NULL, 0 },
  { "ALARM_TAG_NORMAL","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(AWSIS,tagNormal),NULL, 0 },
  { "ALARM_TAG_HIGH","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(AWSIS,tagHigh),NULL, 0 },
  { "ALARM_TAG_HIGHWARN","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE*)OFFSETIN(AWSIS,tagHighWarn),NULL, 0 },
  { "ALARM_TAG_LOW","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(AWSIS,tagLow),NULL, 0 },
  { "ALARM_TAG_LOWWARN","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE*)OFFSETIN(AWSIS,tagLowWarn),NULL, 0 },
# ifndef SMALL_TINE_LIB
  { "FILTER","",0,CF_TEXT,256,(BYTE *)OFFSETIN(AWSIS,filterString),NULL, 0 },
# endif
};
CSVDB csvAlmWatchFileDb =
{
  csvAlmWatchFileColumns,
  sizeof(csvAlmWatchFileColumns)/sizeof(csvAlmWatchFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(AWSIS),
  0
};
void parseNormalValue(UINT32 *val,int *chk,char *valstr)
{
  char *c = "";
  int nchk = 1;
  if (strlen(valstr) != 0) c = valstr;
  if (c[0] == '!')
  {
    nchk = -1; c++;
  }
  *val = (UINT32)strtol(c,NULL,0);
  *chk = nchk;
}
int isInWatchTable(AWSIS *aws)
{
  AWSLstEntry *aw;
  for (aw=almWatchList; aw != NULL; aw = aw->nxt)
  {
    if (strncmp(aws->dev,aw->aws.dev,DEVICE_NAME_SIZE)) continue;
    if (strncmp(aws->eqm,aw->aws.eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strncmp(aws->prp,aw->aws.prp,PROPERTY_NAME_SIZE)) continue;
    if (aws->sev != aw->aws.sev) continue;
    if (aws->asys != aw->aws.asys) continue;
    if (aws->code != aw->code) continue;
    if (aws->mask != aw->aws.mask) continue;
    if (aws->siz != aw->aws.siz) continue;
    return TRUE;
  }
  return FALSE;
}
static ADS *cloneADS(ADS *dst, ADS *src, int newCode, char *tag)
{ /* internal : no NULL checking */
  memcpy(dst, src, sizeof(ADS));
  if (newCode >= 512) dst->alarmCode = newCode;
  if (tag != NULL && strlen(tag) > 0) strncpy(dst->alarmTag, tag, ALARM_TAG_SIZE);
  return dst;
}
#define chkAlmTag(t,F) strncpy(lst->t,aws[i].t,ALARM_TAG_SIZE);\
                       if (strlen(lst->t) == 0) snprintf(lst->t, ALARM_TAG_SIZE, F, lst->aws.prp)
int GetAlarmWatchTable(char *eqm)
{
  static int calledOnce = FALSE;
  int i, naw = 0, cc = 0, pass = 0;
  char almfn[64];
  AWSIS *aws = NULL;
  AWSLstEntry *lst = NULL;
  ADS ads;

  if (gTineConfigurationCoded) return not_implemented;
  feclog("ALM: check for watch table for equipment module %.6s",eqm);
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  {
    if (calledOnce) return 0; 
    naw = populateAWS(gFecName,gFecInfoList,&aws);
    if (naw < 0) ccerr(-nAlmWatchList);
  }
  else
# endif
  {
#   ifdef FS_DELIMITER
    sprintf(almfn,"%.6s%c%s",eqm,FS_DELIMITER,ALMARM_WATCH_FILE);
#   else
    strncpy(almfn,ALMARM_WATCH_FILE,64);
    pass++;
#   endif
retryReadAlarmWatch:
    cc = csvReadFile(fecDbPath,almfn,&csvAlmWatchFileDb,(void **)&aws);
    if (pass == 0 && cc == no_such_file && !calledOnce)
    { /* maybe old style ? */
      if (ignoreCommonDbFiles) ccerr(cc);
      strncpy(almfn,ALMARM_WATCH_FILE,64);
      pass++;
      goto retryReadAlarmWatch;
    }
    if (cc != 0) ccerr(cc);
    naw = csvAlmWatchFileDb.siz;
  }
  calledOnce = TRUE;
  for (i=0; i<naw; i++)
  { /* add these to the alarm watch link list */
    if (isInWatchTable(&aws[i])) continue;
    if ((lst=(AWSLstEntry *)SystemCalloc(1,sizeof(AWSLstEntry))) == NULL) ccerr(out_of_server_memory);
    aws[i].atyp = getArrayType(aws[i].alarmDataFormatString);
    aws[i].fmt = BFMT(ftoi(aws[i].alarmDataFormatString));
    if (!isAllowedAlarmValueFormat(aws[i].fmt))
    {
      feclogEx(TINE_LOGLEVEL_WARN, "ALM: watch file : format %.64s not allowed",aws[i].alarmDataFormatString);
      continue;
    }
    if (aws[i].mask != 0) parseNormalValue(&aws[i].normal,&aws[i].chkNormal,aws[i].normalValueString);
    if (aws[i].hisev == 0) aws[i].hisev = aws[i].sev;
    if (aws[i].hiwarnsev == 0) aws[i].hiwarnsev = aws[i].hisev - 2;
    if (aws[i].losev == 0) aws[i].losev = aws[i].sev;
    if (aws[i].lowarnsev == 0) aws[i].lowarnsev = aws[i].losev - 2;
    memcpy(&lst->aws, &aws[i], sizeof(AWS));
    lst->code = aws[i].code;
    lst->codeHigh = aws[i].codeHigh;
    lst->codeHighWarn = aws[i].codeHighWarn;
    lst->codeLow = aws[i].codeLow;
    lst->codeLowWarn = aws[i].codeLowWarn;
    chkAlmTag(tagNormal,"%.23s invalid");
    chkAlmTag(tagHigh, "%.22s too high");
    chkAlmTag(tagHighWarn, "%.20s near limit");
    chkAlmTag(tagLow, "%.23s too low");
    chkAlmTag(tagLowWarn, "%.20s near limit");
#   ifndef SMALL_TINE_LIB
    lst->fltr = getFilterLinkFromParsableString(aws[i].filterString);
    if (lst->code > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads,&almErrTblADS[idx_invalid_data],lst->code,lst->tagNormal), aws[i].filterString);
    if (lst->codeHigh > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads, &almErrTblADS[idx_too_high], lst->codeHigh, lst->tagHigh), aws[i].filterString);
    if (lst->codeHighWarn > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads,&almErrTblADS[idx_warn_too_high],lst->codeHighWarn,lst->tagHighWarn), aws[i].filterString);
    if (lst->codeLow > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads,&almErrTblADS[idx_too_low], lst->codeLow, lst->tagLow), aws[i].filterString);
    if (lst->codeLowWarn > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads,&almErrTblADS[idx_warn_too_low], lst->codeLowWarn, lst->tagLowWarn), aws[i].filterString);
#   endif
    lst->nxt = almWatchList;
    almWatchList = lst;
  }
  nAlmWatchList += naw;
err:
  SystemFree(aws);
  feclogEx(cc?TINE_LOGLEVEL_INFO:TINE_LOGLEVEL_INFO, "ALM: watch file : %.32s", cc ? cc2str(cc) : "found");
  return cc;
}
static ClnHdr almcln;
static ClnLst almcon;
static ContractListStruct almcl;
void clearWatchTableAlarm(AWSLstEntry *aw,int devNr,int except)
{
# define CLR_WTA(c) if (except != (c) && (c) > 0) lasClearAlarm(aw->aws.eqm,devNr,(UINT32)(c),aw)
  CLR_WTA(invalid_data);
  CLR_WTA(value_too_high);
  CLR_WTA(value_too_low);
  CLR_WTA(warn_too_high);
  CLR_WTA(warn_too_low);
  CLR_WTA(aw->code);
  CLR_WTA(aw->codeHigh);
  CLR_WTA(aw->codeHighWarn);
  CLR_WTA(aw->codeLow);
  CLR_WTA(aw->codeLowWarn);
}
void checkAlarmsInWatchTable(void)
{
  AWSLstEntry *aw;
  int i,cc,cnt,alarm_found,no_set_alarm=FALSE,ismca=FALSE;
  int devNr=0, dv=0;
  ExportListStruct *el;
  char sDevName[DEVICE_NAME_SIZE];
  char sPropName[PROPERTY_NAME_SIZE];
  float fval;
  UINT32 ival, code;
  static BYTE almdata[ALARM_DATA_SIZE];
  static DTYPE din,dout;
  static int done = 0;

  if (!done) 
  { /* initialization */
    strncpy(almcln.usr,"LOCAL_ALARM",USERNAME_SIZE);
    almcon.client = &almcln;
    almcl.clnLst = &almcon;
    almcl.nclients = 1;
    memset(&din,0,sizeof(DTYPE));
  }
  done = TRUE;
  if (ServerWaiting) return;
  if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick) != 0) return;
  for (aw=almWatchList; aw != NULL; aw = aw->nxt)
  { /* run thru the watch list ... */
    if ((el=getExportListItem(aw->aws.eqm)) == NULL) continue;
    if (!el->hasPropertyHandler || el->isidle) continue;
    if ((devNr=GetDeviceNumber(aw->aws.eqm,aw->aws.dev)) < 0)
    { /* aw->aws.dev must be an officially registered device ! */
      if (almdbg) dbglog("LAS: %.64s is not a registered device",aw->aws.dev);
      continue;
    }
    if (aw->aws.cntThreshold < 0) continue; /* disabled */
#   ifndef SMALL_TINE_LIB
    no_set_alarm = (aw->fltr != NULL && !aw->fltr->valid && aw->fltr->param1 < 0) ? TRUE : FALSE;
    if (almdbg > 1 && aw->fltr != NULL) 
      dbglog("LAS: %.6s/%.64s[%.64s] : filter is %s",
        aw->aws.eqm,aw->aws.dev,aw->aws.prp,(aw->fltr->valid?"valid":"not valid"));
#   endif
    if (no_set_alarm) goto scanDataSet;
    dout.dArrayLength = aw->aws.siz;
    dout.dFormat = LFMT(aw->aws.fmt);
    dout.data.vptr = tmpWorkArea;
    dout.dTag[0] = 0;
    strncpy(sDevName,aw->aws.dev,DEVICE_NAME_SIZE);
    strncpy(sPropName,aw->aws.prp,PROPERTY_NAME_SIZE);
    almcl.exp = el;
    almcl.compStatus = not_signalled;
    almcl.prp = GetPropertyListStruct(el->EqmName,sPropName,sDevName);
    cc = callEqm(&almcl,sDevName,sPropName,&dout,&din,CA_READ|CA_ALARM);
    if (cc != 0)
    {
      if (almdbg)
      {
        char errstr[64];
        dbglog("LAS: %6s/%.64s[%.64s] : <%d> %.32s", aw->aws.eqm, sDevName, aw->aws.prp, cc, GetErrorText(cc,errstr,32));
      }
      continue;
    }
scanDataSet:
    if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) != 0) continue;
    cnt = 0; /* start fresh */
    ismca = aw->aws.atyp == AT_CHANNEL ? TRUE : /* so registered */
      (aw->aws.atyp != AT_WAVEFORM && aw->aws.siz == el->EqmNumDevices) ? TRUE : /* assumed it to be*/
      FALSE; /* then not ! */
    for (i=0; i<aw->aws.siz; i++)
    { /* loop over all elements in the call */
      alarm_found = FALSE;
      dv = ismca ? devNr + i : devNr; /* the canonical registered device number to use ... */
      if (el->EqmDevLst[dv].isReserved) continue;
      if (no_set_alarm) goto bof;
      switch (LFMT(aw->aws.fmt))
      { /* form an integer for invalid data check and a float for threshold check ... */
        case CF_BYTE: 
          ival = (UINT32)dout.data.bptr[i]; 
          fval = (float)dout.data.bptr[i]; 
          break;
        case CF_INT16:
          ival = (UINT32)dout.data.sptr[i]; 
          fval = (float)dout.data.sptr[i]; 
          break;
        case CF_INT32: 
          ival = (UINT32)dout.data.lptr[i]; 
          fval = (float)dout.data.lptr[i]; 
          break;
#       ifdef SINT64
        case CF_INT64:
          ival = (UINT32)dout.data.llptr[i];
          fval = (float)dout.data.llptr[i];
          break;
#       endif
        case CF_FLOAT:
          ival = (UINT32)dout.data.fptr[i]; 
          fval = (float)dout.data.fptr[i]; 
          break;
        case CF_DOUBLE:
          ival = (UINT32)dout.data.dptr[i]; 
          fval = (float)dout.data.dptr[i]; 
          break;
        default:
          ival = 0; fval = 0;
      }
      if (aw->aws.chkNormal != 0)
      {
        if ((aw->aws.chkNormal == 1  && aw->aws.normal != (ival & aw->aws.mask)) ||
            (aw->aws.chkNormal == -1 && aw->aws.normal == (ival & aw->aws.mask)) )
        {
          alarm_found = TRUE;
          cnt++;
          memcpy(almdata,&ival,4);
          aw->aws.cnt++;
          almErrTblADS[idx_invalid_data].alarmSystem = aw->aws.asys;
          almErrTblADS[idx_invalid_data].alarmSeverity = aw->aws.sev;
          code = aw->code > 0 ? aw->code : invalid_data;
          if (strlen(aw->tagNormal) > 0) 
          { /* specific tag given in watch table config file */
            strncpy(almErrTblADS[idx_invalid_data].alarmTag,aw->tagNormal,ALARM_TAG_SIZE);
          }
          if (aw->aws.cnt > aw->aws.cntThreshold) 
            _setAlarm(aw->aws.eqm,dv,code,almdata,almNEWALARM,makeDataTimeStamp(),aw);
          clearWatchTableAlarm(aw,dv,code);
        }
      }
      else if (fval > aw->aws.hiwarn) 
      {
        alarm_found = TRUE;
        cnt++;
        memcpy(almdata,&fval,4);
        aw->aws.cnt++;
        almErrTblADS[idx_too_high].alarmSystem = aw->aws.asys;
        almErrTblADS[idx_too_high].alarmSeverity = aw->aws.hisev;
        almErrTblADS[idx_warn_too_high].alarmSystem = aw->aws.asys;
        almErrTblADS[idx_warn_too_high].alarmSeverity = aw->aws.hiwarnsev;
        code = aw->code;
        if (aw->codeHighWarn > 0) code = aw->codeHighWarn;
        if (aw->codeHigh > 0 && fval > aw->aws.hi) code = aw->codeHigh;
        if (code == 0) code = fval > aw->aws.hi ? value_too_high : warn_too_high;
        if (strlen(aw->tagHigh) > 0) 
        { /* specific tag given in watch table config file */
          strncpy(almErrTblADS[idx_too_high].alarmTag,aw->tagHigh,ALARM_TAG_SIZE);
          strncpy(almErrTblADS[idx_warn_too_high].alarmTag,aw->tagHighWarn,ALARM_TAG_SIZE);
        }
        if (aw->aws.cnt > aw->aws.cntThreshold)
          _setAlarm(aw->aws.eqm,dv,code,almdata,almNEWALARM,makeDataTimeStamp(),aw);
        clearWatchTableAlarm(aw,dv,code);
      }
      else if (fval < aw->aws.lowarn) 
      {
        alarm_found = TRUE;
        cnt++;
        memcpy(almdata,&fval,4);
        aw->aws.cnt++;
        almErrTblADS[idx_too_low].alarmSystem = aw->aws.asys;
        almErrTblADS[idx_too_low].alarmSeverity = aw->aws.losev;
        almErrTblADS[idx_warn_too_low].alarmSystem = aw->aws.asys;
        almErrTblADS[idx_warn_too_low].alarmSeverity = aw->aws.lowarnsev;
        code = aw->code;
        if (aw->codeLowWarn > 0) code = aw->codeLowWarn;
        if (aw->codeLow > 0 && fval < aw->aws.lo) code = aw->codeLow;
        if (code == 0) code = fval < aw->aws.lo ? value_too_low : warn_too_low;
        if (strlen(aw->tagLow) > 0)
        { /* specific tag given in watch table config file */
          strncpy(almErrTblADS[idx_too_low].alarmTag,aw->tagLow,ALARM_TAG_SIZE);
          strncpy(almErrTblADS[idx_warn_too_low].alarmTag,aw->tagLowWarn,ALARM_TAG_SIZE);
        }
        if (aw->aws.cnt > aw->aws.cntThreshold) 
          _setAlarm(aw->aws.eqm,dv,code,almdata,almNEWALARM,makeDataTimeStamp(),aw);
        clearWatchTableAlarm(aw,dv,code);
      }
bof:  if (!alarm_found)
      { /* no alarm state for this device */
        clearWatchTableAlarm(aw,dv,0);
      }
    }
    ReleaseSystemMutex(hAlmTblMutex);
    /* only clear the watch table counter when a complete pass has no alarms */
    if (cnt == 0) aw->aws.cnt = 0;
  }
  ReleaseSystemMutex(hSystemKernelMutex);
  return;
}
void getAlarmInfo(void)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->adsTable != NULL) continue;
    if (getAlarmInfoTable(el->EqmName) < 0)
    {
      feclogEx(TINE_LOGLEVEL_INFO, "ALM: no alarm information for %.6s",el->EqmName);
    }
  }
}
_CSVCOLUMN csvAlarmInfoColumns[] = 
{
  { "ALARM_TAG","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(ADSIS,alarmTag),NULL,0 },
  { "ALARM_CODE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(ADSIS,alarmCode),NULL,0 },
  { "ALARM_MASK","0",0,CF_INT32,1,(BYTE *)OFFSETIN(ADSIS,alarmMask),NULL,0 },
  { "SEVERITY","1",0,CF_INT16,1,(BYTE *)OFFSETIN(ADSIS,alarmSeverity),NULL,0 },
  { "DATA_FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(ADSIS,alarmDataFormatString),NULL,0 },
  { "DATA_ARRAYSIZE","0",0,CF_BYTE,1,(BYTE *)OFFSETIN(ADSIS,alarmDataArraySize),NULL,0 },
  { "ALARM_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADSIS,alarmText),NULL,0 },
  { "DEVICE_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADSIS,deviceText),NULL,0 },
  { "DATA_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADSIS,dataText),NULL,0 },
  { "URL","not supplied",0,CF_TEXT,ALARM_TEXT_LONGSIZE,(BYTE *)OFFSETIN(ADSIS,url),NULL,0 },
  { "ALARM_SYSTEM","0",0,CF_INT16,1,(BYTE *)OFFSETIN(ADSIS,alarmSystem),NULL,0 },
# ifndef SMALL_TINE_LIB
  { "FILTER","",0,CF_TEXT,256,(BYTE *)OFFSETIN(ADSIS,filterString),NULL, 0 },
# endif
};
CSVDB csvAlarmInfoDb =
{
  csvAlarmInfoColumns,
  sizeof(csvAlarmInfoColumns)/sizeof(csvAlarmInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ADSIS),
  0
};
char *makeAlarmsFileName(char *eqm,int usesubdir,int decorate)
{
  char dname[32];
  if (decorate == 0)
  {
    strcpy(dname,"alarms.csv");
  }
  else
  {
    memset(dname,0,sizeof(dname));
    strncpy(dname,eqm,EQM_NAME_SHORTSIZE);
    strlwr(dname);
#   if !defined(FS_SHORTNAMES)
    strcat(dname,"-alarms.csv");
#   else
    strcat(dname,"-a.csv");
#   endif
  }
  return makeDbFileName(eqm,dname,usesubdir);
}
int validateAlarmDefinition(ADSIS *ads)
{
  int mxsiz, fsiz;
  if (ads->alarmSeverity < 0) ads->alarmSeverity = 0;
  if (ads->alarmSeverity > 15) ads->alarmSeverity = 15;
  if (strlen(ads->alarmTag) == 0) strncpy(ads->alarmTag,ads->alarmText,ALARM_TAG_SIZE);
  if (ads->alarmDataArraySize <= 0)
  { /* guarantee this : */
    ads->alarmDataFormat = BFMT(CF_NULL);
    ads->alarmDataArraySize = 0;
  }
  else 
  {
    ads->alarmDataFormat = BFMT(ftoi(ads->alarmDataFormatString));
    if (!isAllowedAlarmDataFormat(ads->alarmDataFormat))
    {
      ads->alarmDataFormat = BFMT(CF_NULL);
      ads->alarmDataArraySize = 0;
      feclogEx(TINE_LOGLEVEL_WARN, "LAS: alarm data format %.64s not allowed",ads->alarmDataFormatString);
      return FALSE;
    }
    fsiz = fmtsizeof(ads->alarmDataFormat);
    mxsiz = fsiz > 0 ? 64 / fsiz : 0;
    if (ads->alarmDataArraySize > mxsiz) ads->alarmDataArraySize = mxsiz;
  }
  ensureAsciiCharacters(ads->alarmTag, 32);
  ensureAsciiCharacters(ads->alarmText, 64);
  ensureAsciiCharacters(ads->deviceText, 64);
  ensureAsciiCharacters(ads->dataText, 64);
  return TRUE;
}
int getAlarmInfoTable(char *eqm)
{
  int i,k,n,cc=0,hasAlarmSystem=0,decorate=0,usesubdir=TRUE;
  ADSIS *ads = NULL;
  ADSList *lst = NULL;
  char *fname=NULL;
  ExportListStruct *el;

  if (gTineConfigurationCoded) return not_implemented;
  if ((el=getExportListItem(eqm)) == NULL) ccerr(non_existent_elem);
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  {
    n = populateADS(gFecName,eqm,gFecInfoList,&ads);
    if (n < 0) ccerr(-n);
  }
  else
# endif
  {
retryEqmAlarmsFile:
    fname = makeAlarmsFileName(eqm,usesubdir,decorate);
    if ((cc=csvReadFile(fecDbPath,fname,&csvAlarmInfoDb,(void **)&ads)) == no_such_file && decorate++ == 0) 
      goto retryEqmAlarmsFile;
#   ifdef FS_DELIMITER
    if (cc == no_such_file && usesubdir == TRUE)
    { 
      usesubdir = FALSE;
      decorate = 0;
      goto retryEqmAlarmsFile;
    }
#   endif
    if (cc != 0) ccerr(cc);
    feclogEx(TINE_LOGLEVEL_INFO, "alarm definition file %s found!",fname);
    n = csvAlarmInfoDb.siz;
  }
  for (i=0, k=0; i<n; i++) 
  {
    if (ads[i].alarmCode == 0) continue;
    if ((lst=(ADSList *)SystemCalloc(1,sizeof(ADSList))) == NULL) ccerr(out_of_server_memory);
    if (!validateAlarmDefinition(&ads[i])) continue;
    memcpy(&lst->ads,&ads[i],sizeof(ADS));
#   ifndef SMALL_TINE_LIB
    if (strlen(ads[i].filterString) > 0)
    { /* e.g. /ctx/srv/dev[prp]=val&severity=sev */
      lst->fltr = getFilterLinkFromParsableString(ads[i].filterString);
    }
#   endif
    lst->ads.alarmOscillationWindow = ALM_OSCILLATION_WINDOW;
    lst->next = el->adsTable;
    el->adsTable = lst;
    if (ads[i].alarmSystem) hasAlarmSystem = 1;
    k++;
  }
  el->alarmSystem = hasAlarmSystem;
  el->adsTableSize = k;
  feclog("read ALM file for %.6s",eqm);
err:
  if (cc && cc != no_such_file) feclog("getAlarmInfoTable : %.32s",cc2str(cc));
  SystemFree(ads);
  return 0;
}
int isCASInUserList(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  NAME16 usrlst[16];
  int i, n = 16;
  static int casUnknown = FALSE;
  static FecAddrStruct casfec, *fec = NULL;

  if (el == NULL) return FALSE;
  if (casUnknown) return FALSE;
  if (fec == NULL)
  {
    if ((fec=GetFecAddrFromCache(el->EqmContext,"CAS")) == NULL)
    {
      feclogEx(TINE_LOGLEVEL_INFO, "CAS unknown for context %.32s",el->EqmContext);
      casUnknown = TRUE;
      return FALSE;
    }
    memcpy(&casfec,fec,sizeof(FecAddrStruct));
    fec = &casfec; /* reassign to local static contents */
  }
  GetCallerInfo(eqm,usrlst,NULL,NULL,NULL,&n);
  for (i=0; i<16 && i<n; i++) if (!strnicmp(fec->fecName,usrlst[i].name,16)) break;
  return i < n ? TRUE : FALSE;
}
/* TODO:  make this do something */
int RegisterSeverity(int almCode,int almSeverity)
{
  UNUSED(almCode);
  UNUSED(almSeverity);
  return 0;
}
/* TODO:  make this do something */
void enforceAlarm(short cc)
{
 /* method to make sure this alarm goes out */
  UNUSED(cc);
  return;
}
void rmvAlarmsFromCAS(char *eqm)
{
  ExportListStruct *el=getExportListItem(eqm);
  int cc = 0;
  char devName[256], *casctx;
  if (el == NULL) return;
  if (el->almdone) return;
  if (gIgnoreCentralServices) return;
  if (strlen(el->EqmContext) == 0) strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
  if (strlen(el->EqmSubSystem) == 0) strncpy(el->EqmSubSystem,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
  if (!strcmp(el->EqmName,SRVEQM_NAME)) return;
  casctx = getPrimaryContext(el->EqmContext,el->EqmSubSystem);
  sprintf(devName,"/%.32s/%.32s/%.32s",casctx,CAS_SRV,el->EqmExportName);
  cc = ExecLinkEx(devName,"REMOVEALARMS",NULL,NULL,CA_WRITE|CA_RETRY,200);
  if (cc == 0) 
  { /* the CAS deems this FEC to be important */
    gRequireReadFromCAS = TRUE;
    el->almdone = TRUE;
  }
  feclogEx(TINE_LOGLEVEL_INFO,"ALM: clear CAS Alarm List for /%.32s/%.32s: %.32s",el->EqmContext,el->EqmExportName,cc2str(cc));
}
void clearCASAlarmList(void)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next) rmvAlarmsFromCAS(el->EqmName);
}
int initAlarmServer(void)
{
  char *ptr = NULL;
  int v = 0;
  feclog("ALM: initializing local alarm subsystem");
  ptr = getenv("TINE_ALARM_COLLAPSE_WINDOW");
  if (ptr != NULL)
  {
    if ((v=atoi(ptr)) < 0) v = 0;
    feclog("ALM: collapse window set to %d from environment",v);
    SetAlarmCollapseWindow(v);
  }
  ptr = getenv("TINE_ALARM_SURVEY_INTERVAL");
  if (ptr != NULL)
  {
    if ((v = atoi(ptr)) < 0) v = 0;
    feclog("ALM: survey interval set to %d from environment", v);
    SetAlarmSurveyInterval((UINT32)v);
  }
  if (!gIsCentralAlarmServer) clearCASAlarmList();
  AlarmServerInitialized = 1;
  return 0;
}
void freeAlmMemory(void)
{
  AWSLstEntry *wlst=NULL, *nxtw=NULL;
  if (almWatchList != NULL)
  {
    feclog("freeing alarm watch table (%d records)",nAlmWatchList);
    for (wlst=almWatchList; wlst !=NULL; )
    {
      nxtw = wlst->nxt; SystemFree(wlst); wlst = nxtw;
    }
  }
  almWatchList = NULL;
  nAlmWatchList = 0;
  return;
}
int verifyDeviceNr(ExportListStruct *el,int devNr)
{
 /* verify that the device number passed in the API is valid */
  if (el == NULL) return non_existent_elem;
  if (!el->hasRegisteredDevicesNames) return not_initialized;
  if (devNr >= el->EqmNumDevices || devNr < 0) return illegal_device;
  if (el->EqmDevLst != NULL && el->EqmDevLst[devNr].isReserved) return not_registered;
  return 0;
}
int RefreshMRT(UINT32 t,ExportListStruct *e)
{
  if (e == NULL) return FALSE;
  if (t == e->almInfo.mrt) 
  {
    e->almInfo.nmrt++;
  }
  else 
  {
    e->almInfo.mrct = e->almInfo.mrt;
    e->almInfo.nmrct = e->almInfo.nmrt;
    e->almInfo.mrt = t;
    e->almInfo.nmrt = 1;
    return TRUE;
  }
  return FALSE;
}
void RefreshHSV(short s,ExportListStruct *e) 
{
  if (e == NULL) return;
  if (s == e->almInfo.hsv)
  {
    e->almInfo.nhsv++;
  }
  else if (s > e->almInfo.hsv)
  { 
    e->almInfo.hsv = (BYTE)s; 
    e->almInfo.nhsv = 1; 
  }
}
static int gAllowOfflineAlarms = FALSE;
TINE_EXPORT void SetAllowOfflineAlarms(int value)
{
  gAllowOfflineAlarms = value;
}
TINE_EXPORT int GetAllowOfflineAlarms(void)
{
  return gAllowOfflineAlarms;
}
int _setAlarm(char *eqm,int devNr,UINT32 almCode,BYTE *almData,BYTE flags,double ts,AWSLstEntry *aws)
{
  short swpfmt,swpdim,cc = 0,adatasize = 0,keepTimestamp = FALSE;
  ALARM *alm = NULL;
  ALARM *lsa = NULL;  /* least significant alarm */
  short lsc,sc = findSeverity(eqm,almCode),tmpsc = sc;
  int dt, oscWnd = gAlmOscillationWindow, oscPinned = gAlmOscillationWindowPinned, dcWndPassed = FALSE;
  ADS *ads;
  ExportListStruct *el;
  FilterLink *fltr = NULL;
  int altsev = -1; /* 'filter invalid' alternate severity code */
  int sevMatched = FALSE, no_new_alarm = FALSE;

  if (!AlarmServerInitialized) return not_initialized;
  if ((el=getExportListItem(eqm)) == NULL) return non_existent_elem;
  if ((fltr=getAlarmFilter(eqm, almCode)) != NULL && !fltr->valid)
  { /* filter assigned but not valid */
    altsev = fltr->param1;
    tmpsc = sc = altsev; /* assign the 'filter invalid' severity */
    if (altsev < 0) no_new_alarm = TRUE;
    if (almdbg > 1) dbglog("ALM filter: %.6s <%d> not valid, use alternate severity %d", 
      el->EqmName, (int)almCode, altsev);
  }
  if (AlarmTestState) flags |= almTEST; /* declared test state */
  /* transient alarms pickup the TERMINATE signal immediately */
  if (flags & almTRANSIENT) flags |= almTERMINATE;
  /* fill in ALARMStruct for the tag via link list: */
  if (el->isidle) ccerr(server_idle);
  if (verifyDeviceNr(el,devNr)) ccerr(illegal_device_number);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if (el->EqmDevLst[devNr].isReserved) ccerr(not_registered);
  if (!gAllowOfflineAlarms && el->EqmDevLst[devNr].mask & DEVICE_OFFLINE)
  {
    if (almdbg) dbglog("ALM: alarm not set -> %.64s is offline",el->EqmDevLst[devNr].name);
    ccerr(0);
  }
  if ((ads=getAlarmDefinition(eqm,almCode)) != NULL)
  { /* is a registered alarm code (good!) */
    oscWnd = ads->alarmOscillationWindow;
    if (!oscPinned) oscPinned = ads->alarmOscillationWindowPinned;
    if (almData != NULL)
    {
      if ((swpfmt=LFMT(ads->alarmDataFormat)) == CF_BIT)
      {
        swpfmt = CF_BYTE;
        swpdim = MIN((ads->alarmDataArraySize+7)/8,ALARM_DATA_SIZE);
      }
      else 
      {
        swpdim = MIN(ads->alarmDataArraySize,ALARM_DATA_SIZE);
      }
      adatasize = swpdim*getFormatSize(swpfmt);
      if (adatasize > ALARM_DATA_SIZE) adatasize = ALARM_DATA_SIZE;
      prepOutgoingAlarmData(almData,swpdim,swpfmt);
    }
  }
  for (alm=el->EqmDevLst[devNr].almLst; alm != NULL; alm=alm->next)
  {
    if ((lsc=findSeverity(eqm,alm->alarmCode)) < tmpsc) { lsa = alm; tmpsc = lsc; }
    if (almCode == alm->alarmCode && aws == alm->wte)
    { /* it's in the list */
      keepTimestamp = FALSE;
      sevMatched = alm->alarmSeverity == (BYTE)sc ? TRUE : FALSE;
      if (!(alm->alarmStatus & almTERMINATE)) 
      { /* not marked for termination */
        if (!sevMatched)
        { /* validity has changed ? : terminate the current alarm */
          alm->alarmStatus &= ~((BYTE)almNEWALARM);
          alm->alarmStatus |= (BYTE)almTERMINATE;
          alm->clrCount = ads != NULL ? ads->alarmOscillationWindow : ALM_OSCILLATION_WINDOW;
          AssignAlmTimestamp(alm, ts);
          if (almdbg) dbglog("ALM filter: %.6s <%d> validity change, severity change (%d -> %d) => terminate alarm ", 
            el->EqmName, (int)almCode, alm->alarmSeverity, sc);
          return 0; /* let the next setAlarm get the new severity as a new alarm */
        }
        if (alm->alarmStatus & almNEWALARM && 
            ts <= tval2ts(alm->timestamp,alm->timestampUSec) + NEW_ALARM_WINDOW) /* a double SetAlarm() ? */
        { /* reset clear count */
          alm->clrCount = 0;                     
          return 0;
        }
        if (almData != NULL && memcmp(almData,alm->alarmData,adatasize))
        { /* alarm data have changed */
          dcWndPassed = (((UINT32)ts - alm->timestamp) < (UINT32)AlarmDataChangeWindow) ? FALSE : TRUE;
          if (dcWndPassed) alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM only if passed datachange window */
          if (alm->alarmStatus & almDATACHANGE)
          { /* DATACHANGE already set, when was the last one ? */
            if (!dcWndPassed) keepTimestamp = TRUE;
          }
          alm->alarmStatus |= almDATACHANGE;
          if (!keepTimestamp)
          { /* passed the data change window ... */
            AssignAlmTimestamp(alm,ts);
            RefreshMRT(alm->timestamp,el);
          }
          memcpy(alm->alarmData,almData,adatasize); /* copy the new data ! */
          if (almdbg) dbglog("ALM datachange: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
        }
        else
        { /* alarm has been reset */
          if (alm->clrCount > 1  && aws == NULL)
          { /* appears to be oscillating (watch table alarms aren't allowed to oscillate) */
            if ((alm->alarmStatus & almOSCILLATION)) keepTimestamp = TRUE;
            alm->alarmStatus |= almOSCILLATION;  
            if (almdbg) dbglog("ALM oscillation: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus &= ~((BYTE)almDATACHANGE); /* remove DATACHANGE */
            if (!keepTimestamp)
            {
              AssignAlmTimestamp(alm,ts);
              RefreshMRT(alm->timestamp,el);
            }
          }
        }
        alm->clrCount = 0; /* reset clear count      */
        return 0; /* already in list, so return */
      }
      else
      { /* has been marked for termination */
        if (sevMatched && !oscPinned && !(alm->alarmStatus & almTRANSIENT))
        { /* oscillation window is too small ? */
          dt = (int)(alm->timestamp+TerminationWindow) - (int)ts;
          if (dt > 0 && dt < ALM_OSCILLATION_MAXIMUM-oscWnd)
          { /* adjust the oscillation window (assumes a known relationship between clear counts and seconds) */
            oscWnd += dt;
            if (ads != NULL) ads->alarmOscillationWindow = oscWnd; else gAlmOscillationWindow = oscWnd;
            feclog("increase alarm oscillation windows to %d clear counts",oscWnd);
          }
        }
      }
    }
  }
  if (no_new_alarm)
  { /* a filter has been applied but is not valid : no new alarms ! */
    if (almdbg > 1) dbglog("ALM filter: %.6s <%d> not valid, alternate severity < 0 => skip", el->EqmName, (int)almCode);
    return not_applicable;
  }
  /* alarm not in list */
  flags |= almNEWALARM; 
  if (el->almInfo.nalms < MaxNumberAlarms)          
  { /* still room to put the alarm in list */
    if ((alm=(ALARM *)SystemCalloc(1,sizeof(ALARM))) == NULL) ccerr(out_of_server_memory);
    if (gRequireReadFromCAS) alm->removalBlocked = 1; /* must be cleared by CAS readout */
    AssignAlmTimestamp(alm,ts);
    alm->alarmCode = almCode;
    alm->alarmStatus |= flags;
    if (adatasize > 0) memcpy(alm->alarmData,almData,adatasize);
    alm->wte = aws;
    alm->next = (ALARM *)el->EqmDevLst[devNr].almLst;
    el->EqmDevLst[devNr].almLst = alm;
    el->almInfo.nalms++;
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
    alm->alarmSeverity = (BYTE)sc;
    if (almdbg) dbglog("ALM new: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
    alm->starttime = alm->timestamp;
    alm->starttimeUSec = alm->timestampUSec;
    ccerr(0);
  }
  else if (lsa != NULL)  
  { /* replace the least significant alarm */
    alm = lsa;
    if (almdbg) dbglog("ALM replace code: %.6s (%d) with (%d)",el->EqmName,(int)alm->alarmCode,(int)almCode);
    AssignAlmTimestamp(alm,ts);
    alm->alarmCode = almCode;
    alm->alarmStatus = flags;
    alm->starttime = alm->timestamp;
    alm->starttimeUSec = alm->timestampUSec;
    if (adatasize > 0)
    {
      memset(alm->alarmData,0,ALARM_DATA_SIZE);
      memcpy(alm->alarmData,almData,adatasize);
    }
    alm->wte = aws;
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
  }
  cc = max_alarms_exceeded;
err:
  switch (cc)
  {
    case 0: return 0;
    case out_of_server_memory:
    case max_alarms_exceeded:
      enforceAlarm(cc);
      break;
    default:
      SystemFree(alm);
      break;
  }
  return cc;
}
int SetAlarmByDeviceNumber(char *eqm,int devNr,UINT32 almCode,BYTE *almData,BYTE flags,double ts)
{
  return _setAlarm(eqm,devNr,almCode,almData,flags,ts,NULL);
}
/* and in the background... */
int RemoveAllAlarms(char *eqm)
{ /* for 'emergencies' ... */
  ExportListStruct *el=getExportListItem(eqm);
  ALARM *a, *anxt;
  int n;
  if (el == NULL) return non_existent_elem;
  for (n=0; n<el->EqmNumDevices && el->EqmDevLst != NULL; n++)
  {
    for (a=el->EqmDevLst[n].almLst; a != NULL; a=anxt)
    {
      anxt = a->next;
      SystemFree(a);
      el->EqmDevLst[n].almLst = anxt;
    }
    el->EqmDevLst[n].almLst = NULL;
  }
  el->almInfo.nhsv = 0;
  el->almInfo.hsv = 0;
  el->almInfo.nmrt = 0;
  el->almInfo.nmrct = 0;
  el->almInfo.nalms = 0;
  rmvAlarmsFromCAS(eqm);
  return 0;
}
ALARM *RemoveAlarmFromList(ExportListStruct *el, int n, ALARM *alm)
{
  ALARM *a0, *a1, *tmp = alm;

  if (el == NULL || el->EqmDevLst == NULL) return alm->next;
  if (n < 0 || n >= el->EqmNumDevices) return alm->next;
  for (a1=a0=el->EqmDevLst[n].almLst; a1 != NULL; a0=a1,a1=a1->next)
  {
    if (a1 == alm) /* found it */
    {
      if (a1 == a0) el->EqmDevLst[n].almLst = a1->next;
      else a0->next = a1->next;
      tmp = a1->next;
      if (a1->timestamp == el->almInfo.mrt) el->almInfo.nmrt--;
      if (a1->timestamp == el->almInfo.mrct) el->almInfo.nmrct--;
      if (el->almInfo.hsv == findSeverity(el->EqmName,a1->alarmCode))
      {
        if (--(el->almInfo.nhsv) <= 0)
        {
          el->almInfo.nhsv = 0; /* reset it */
          el->almInfo.hsv = 0;
        }
      }
      SystemFree(a1);
      el->almInfo.nalms--;
      break;
    }
  }
  return tmp;    /* return the next link in the list */
}
int prepCollapsedAlarmMessage(AMS *ams,ExportListStruct *el)
{
  char str[256];
  int i, n, k = 1, code = max_alarms_exceeded;
  int region=DEV_REGION_UNKNOWN, multi_regional=FALSE;
  short adata[32];
  ADS *ads;
  memset(adata,0,sizeof(adata));
  memset(ams,0,sizeof(AMS));
  strncpy(ams->server,el->EqmExportName,EXPORT_NAME_SIZE);
  strncpy(ams->device,el->EqmExportName,EXPORT_NAME_SIZE);
  adata[0] = el->almInfo.nalms;
  for (i=0, n=0; i<el->EqmNumDevices; i++)
  { /* run thru the device list */
    if (el->EqmDevLst[i].almLst == NULL) continue;
    if (el->EqmDevLst[i].isReserved) continue;
    if (region == DEV_REGION_UNKNOWN) region = el->EqmDevLst[i].region;
    if (!multi_regional && region != el->EqmDevLst[i].region) multi_regional = TRUE; 
    if (code == max_alarms_exceeded) code = ALM_BASECODE(el->EqmDevLst[i].almLst->alarmCode);
    if (k < 32 && code != adata[k]) adata[k++] = code;
    if (code != (int)ALM_BASECODE(el->EqmDevLst[i].almLst->alarmCode)) break; 
    n++;
  }
  if (n > gAlmCollapseWindow)
  { /* at least this many with the same code */
    if ((ads=getAlarmDefinition(el->EqmName,code)) != NULL)
    {
      sprintf(str,"%d alarms: %.32s",n,ads->alarmTag);
    }
    else
    {
      sprintf(str,"%d alarms: alarm code %d",n,code);
    }
  }
  else
  { /* then use the last known num alarms over threshold */
    n = el->amsDynSet.nalarms;
    sprintf(str,"%d total alarms ( > %d)",n,gAlmCollapseWindow);
  }
  strncpy(ams->alarmTag,str,ALARM_TAG_SIZE);
  ams->alarmCode = max_alarms_exceeded;
  ams->timestamp = el->amsDynSet.timestamp;
  ams->timestampUSec = el->amsDynSet.timestampUSec;
  ams->starttime = el->amsDynSet.starttime;
  ams->starttimeUSec = el->amsDynSet.starttimeUSec;
  ams->alarmMask = multi_regional ? el->region : region;
  memcpy(ams->alarmData,adata,64);
  ams->alarmDataFormat = BFMT(CF_INT16);
  ams->alarmDataArraySize = k;
  ams->severity = el->amsDynSet.severity;
  ams->descriptor = el->amsDynSet.descriptor;
  ams->alarmSystem = 0;
  if (!collapsedTerminationPending)
  { /* this was the termination message, so clear these: */
    el->amsDynSet.nalarms = 0;
    el->amsDynSet.severity = 0;
  }
  return 1;
}
void updateCollapsedAlarmMessage(ExportListStruct *el,struct timeval *tv)
{
  if (gAlmCollapseWindow <= 0) return;
  if (el->almInfo.nalms <= gAlmCollapseWindow)
  { /* no need to collapse */
    if (el->amsDynSet.starttime == 0) return;
    if (el->amsDynSet.descriptor & almTERMINATE) return;
    msglog(0,"collapsed alarms: terminate");
    el->amsDynSet.descriptor = almTERMINATE;
    if (el->amsDynSet.timestamp == (UINT32)tv->tv_sec)
    { /* rapid return to non-collapsed ? */
      el->amsDynSet.timestamp = (UINT32)tv->tv_sec + 1;
      el->amsDynSet.timestampUSec = 0;
    }
    else
    {
      el->amsDynSet.timestamp = (UINT32)tv->tv_sec;
      el->amsDynSet.timestampUSec = (UINT32)tv->tv_usec;
    }
    return;
  }
  if (el->amsDynSet.descriptor == 0 || el->amsDynSet.descriptor & almTERMINATE)
  { /* either never set or last set to terminated -> this is now NEW ! */
    msglog(0,"collapsed alarms: new collapsed alarm!");
    el->amsDynSet.descriptor = almNEWALARM;
    el->amsDynSet.timestamp = (UINT32)tv->tv_sec;
    el->amsDynSet.timestampUSec = (UINT32)tv->tv_usec;
    el->amsDynSet.starttime = (UINT32)tv->tv_sec;
    el->amsDynSet.starttimeUSec = (UINT32)tv->tv_usec;
    el->amsDynSet.severity = (BYTE)el->almInfo.hsv;
    el->amsDynSet.nalarms = (UINT16)el->almInfo.nalms;
    return;
  }
  if (el->amsDynSet.timestamp + AlarmHeartbeat < (UINT32)tv->tv_sec) 
  { /*  set the heartbeat */
    el->amsDynSet.descriptor = almHEARTBEAT;
    el->amsDynSet.timestamp = (UINT32)tv->tv_sec;
    el->amsDynSet.timestampUSec = (UINT32)tv->tv_usec;
    el->amsDynSet.nalarms = (UINT16)el->almInfo.nalms;
    return;
  }
}
void checkFileErrorAlarms(void)
{
  static int hasAlarms = FALSE;
  int hasAlarmsNow = hasLogFileAlarm || hasHstFileAlarm();

  if (hasAlarms && !hasAlarmsNow)
  { /* had file error alarms but not any longer */
    clearFileErrorAlarm();
  }
  hasAlarms = hasAlarmsNow; /* record current state */
  return;
}
void SurveyAlarms(struct timeval *tv)
{
  short sc;
  int n, findHSV, oscWnd, twind = TerminationWindow, rmvBlocked = FALSE;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;
  static struct timeval tv_interim;
  static time_t lasttime = 0;

  if (!AlarmServerInitialized) return;
  if (tv->tv_sec <= newAlarmBaseTime) return;
  /* generally: once per second ... */
  if (MSECS(*tv,tv_interim) < gAlarmSurveyInterval) return;
  if (ServerExitCondition) return;
  gettimeofday(&tv_interim,(struct timezone *)NULL);
# ifndef SMALL_TINE_LIB
  lasttime = thisCycleTime.tv_sec;
  if (!gAlmMfDumped && lasttime > gSystemStartupTime + PRE_DUMP_WINDOW)
  {
    dumpAlmManifest(ALM_MF_ALL,gStartupDebug);
    gAlmMfDumped = TRUE;
  }
# endif
  if (nAlmWatchList > 0) 
  {
    gServerCycleState = CYCLE_STATE_ALMWATCH;
    checkAlarmsInWatchTable();
  }
  gServerCycleState = CYCLE_STATE_ALMPROC;
# ifndef SMALL_TINE_LIB
  if (gMinDiskSpaceTblSize > 0) checkFreeBlocks();
  if (minAvailMemBytes > 0) checkAvailMemBytes();
# endif
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) != 0) return;
  checkFileErrorAlarms();
  if (twind > 3600) twind = 3600; /* draw the line somewhere ! */
  if (collapsedTerminationPending) twind = 0;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->isidle || el->EqmDevLst == NULL) continue;
    updateCollapsedAlarmMessage(el,tv);
    if (tv->tv_sec > (long)el->almInfo.mrt && el->almInfo.mrt > el->almInfo.mrct) 
    { /* adjust most recent complete timestamp */
      el->almInfo.mrct = el->almInfo.mrt;
      el->almInfo.nmrct = el->almInfo.nmrt;
    }
    findHSV = (el->almInfo.hsv == 0) ? TRUE : FALSE;
    for (n=0; el->almInfo.nalms && n<el->EqmNumDevices; n++)
    {
      if (el->EqmDevLst[n].isReserved) continue;
      for (alm=el->EqmDevLst[n].almLst; alm != NULL;)
      {
        if ((alm->alarmStatus & almTERMINATE)) 
        { /*  time to clear the alarm ? */
          if (collapsedTerminationPending && 
              tv->tv_sec > (time_t)(el->amsDynSet.timestamp + AlarmHeartbeat + AMSDynSet_AllowedTerminationPending))
          { /* missing CAS READ ? */
            msglog(0,"collapsed alarms: clear dangling termination pending");
            collapsedTerminationPending = FALSE;
            touchActiveAlarms(el);
          }
          rmvBlocked = collapsedTerminationPending ? FALSE : alm->removalBlocked;
          if (rmvBlocked && alm->timestamp < (UINT32)(tv->tv_sec + DEFAULT_ALARM_HEARTBEAT)) 
          { /* give the CAS a 20 minute chance to read it out */
            alm = alm->next;
            continue;
          }
          if (alm->timestamp + twind < (UINT32)tv->tv_sec) 
          { /* okay to clear the alarm ! */
            if (almdbg) dbglog("ALM clear: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
          else if ((alm->alarmStatus & almTRANSIENT) &&
                    alm->timestamp + AlarmInstantWindow < (UINT32)tv->tv_sec)
          {
            if (almdbg) dbglog("ALM clear transient: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
        }
        else 
        { /* check alarm status */
          oscWnd = gAlmOscillationWindow;
          if ((ads=getAlarmDefinition(el->EqmName,alm->alarmCode)) != NULL) 
            oscWnd = MIN(ads->alarmOscillationWindow,ALM_OSCILLATION_MAXIMUM);
          if ((alm->clrCount > oscWnd - ALM_OSCILLATION_CUSHION) &&
               alm->timestamp + AlarmWindow < (UINT32)tv->tv_sec)
          {
            if (almdbg) dbglog("ALM terminate: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almTERMINATE;
            alm->timestamp = tv->tv_sec;
            alm->timestampUSec = tv->tv_usec;
            RefreshMRT(alm->timestamp,el);
          }
          else if (alm->timestamp + AlarmHeartbeat < (UINT32)tv->tv_sec) 
          { /*  set the heartbeat */
            if (almdbg) dbglog("ALM heartbeat: %.6s <%d> (0x%x)",el->EqmName,(int)alm->alarmCode,alm->alarmStatus);
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus &= ~((BYTE)almDATACHANGE); /* remove DATACHANGE */
            alm->alarmStatus |= almHEARTBEAT;
            alm->timestamp = tv->tv_sec;
            alm->timestampUSec = tv->tv_usec;
            RefreshMRT(alm->timestamp,el);
          }
        }
        if (findHSV)
        {
          if ((sc=findSeverity(el->EqmName,alm->alarmCode)) > el->almInfo.hsv)
          {
            el->almInfo.hsv = sc; el->almInfo.nhsv = 1;
          }
          else if (sc > 0 && sc == el->almInfo.hsv) el->almInfo.nhsv++;
        }
        alm = alm->next;
      }
    }
  }
  ReleaseSystemMutex(hAlmTblMutex);
  return;
}
/* caller subscribes to NALARMS which calls (DATACHANGE mode) */
int numAlarmsEx(char *eqmName,int devNr,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity)
{
  ExportListStruct *el=getExportListItem(eqmName);
  UINT32 ts;
  int nt,sv,nts,nsv;
  ALARM *alm;

  if (el == NULL || el->EqmDevLst == NULL) return 0;
  if (devNr < 0)
  {
    ts = el->almInfo.mrct;
    sv = el->almInfo.hsv;
    nts = el->almInfo.nmrct;
    nsv = el->almInfo.nhsv;
    nt = el->almInfo.nalms;
    if (gAlmCollapseWindow > 0 && (nt > gAlmCollapseWindow || collapsedTerminationPending))
    {
      nsv = nts = nt = 1;
      if (el->amsDynSet.severity == 0) el->amsDynSet.severity = (BYTE)el->almInfo.hsv;
      sv = el->amsDynSet.severity;
      ts = el->amsDynSet.timestamp;
    }
  }
  else
  {
    if (devNr >= el->EqmNumDevices) return 0;
    nt = 0; ts = 0; sv = 0; nts = 0; nsv = 0;
    for (alm=el->EqmDevLst[devNr].almLst; alm != NULL; alm=alm->next)
    {
      if (el->EqmDevLst[devNr].isReserved) continue;
      if (alm->timestamp == ts) nts++;
      if (alm->timestamp > ts) { ts = alm->timestamp; nts = 1; }
      if (alm->alarmSeverity == sv) nsv++;
      if (alm->alarmSeverity > sv) { sv = alm->alarmSeverity; nsv = 1; }
      nt++;
    }
  }
  if (almTimeStamp != NULL) *almTimeStamp = ts;
  if (almSeverity != NULL) *almSeverity  = (short)sv;
  if (NrAtAlmTimeStamp != NULL) *NrAtAlmTimeStamp = (short)nts;
  if (NrAtAlmSeverity != NULL) *NrAtAlmSeverity = (short)nsv;
  return nt;
}
int numAlarms(char *eqmName,UINT32 *almTimeStamp,short *almSeverity)
{
  return numAlarmsEx(eqmName,-1,almTimeStamp,almSeverity,NULL,NULL);
}
TINE_EXPORT int GetNumAlarms(char *eqmName,int devNr,UINT32 *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity)
{
  return numAlarmsEx(eqmName,devNr,almTimeStamp,almSeverity,NrAtAlmTimeStamp,NrAtAlmSeverity);
}
int numAlarmDefs(char *eqm)
{
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) return -non_existent_elem;
  return el->adsTableSize;
}
int getAlarmDefs(char *eqm,ADSList **adslst)
{
  ExportListStruct *el;

  if ((el=getExportListItem(eqm)) == NULL) return 0;
  *adslst = el->adsTable;
  return el->adsTableSize;
}
short findSeverity(char *eqm, UINT32 cc)
{
  ADS *ads;
  ADSList *lst;
  ExportListStruct *start,*stop,*el;

  if (isLinkErrorAlarm(cc)) cc = link_error;
  if (isDiskSpaceAlarm(cc)) cc = (cc & 0xff);
  cc = ALM_BASECODE(cc);
  if (eqm == NULL)
  { /* scan all modules */
    start = ExportList; stop = NULL;
  }
  else
  { /* find the target */
    if ((start=getExportListItem(eqm)) == NULL) return 0;
    stop = start->next;
  }
  for (el=start; el != stop; el=el->next)
  {
    if ((lst=el->adsTable) != NULL) 
    {
      while (lst != NULL && lst->ads.alarmCode != (UINT32)cc) lst = lst->next;
      if (lst != NULL) return lst->ads.alarmSeverity;
    }
  }
  switch (cc)
  { /* wasn't overloaded in the registered ADS table */
    case invalid_data: return almErrTblADS[idx_invalid_data].alarmSeverity;
    case value_too_high: return almErrTblADS[idx_too_high].alarmSeverity;
    case value_too_low: return almErrTblADS[idx_too_low].alarmSeverity;
    case warn_too_high: return almErrTblADS[idx_warn_too_high].alarmSeverity;
    case warn_too_low: return almErrTblADS[idx_warn_too_low].alarmSeverity;
    default: 
      if (cc > 0 && cc < (UINT32)numErr)
      {
        if ((ads=getErrTblADS((short)cc)) != NULL) return (short)ads->alarmSeverity;
        return ALM_SEVERITY_WARN;
      }
      break;
  }
  return 0;
}
TINE_EXPORT int GetAlarmDefinition(char *eqm,UINT32 code,ADS *ads)
{
  ADS *pads;
  if (ads == NULL) return invalid_parameter;
  if ((pads=getAlarmDefinition(eqm,code)) == NULL) return non_existent;
  memcpy(ads,pads,sizeof(ADS));
  return 0;
}
FilterLink *getAlarmFilter(char *eqm, UINT32 code)
{
# ifndef SMALL_TINE_LIB
  ExportListStruct *el=getExportListItem(eqm);
  if (el != NULL)
  { /* equipment module found ! */
    ADSList *lst;
    code = ALM_BASECODE(code);
    for (lst=el->adsTable; lst != NULL; lst=lst->next)
    {
      if ((UINT32)code == lst->ads.alarmCode)
      {
        return lst->fltr;
      }
    }
  }
# else
  eqm = eqm; code = code;
# endif
  return NULL;
}
ADS *getAlarmDefinition(char *eqm,UINT32 code)
{
  ADSList *lst;
  ExportListStruct *el;
  AWSLstEntry *awe;
  if (isLinkErrorAlarm(code)) code = link_error;
  if (isDiskSpaceAlarm(code)) code = code & 0xff;
  if ((el=getExportListItem(eqm)) != NULL)
  {
    code = ALM_BASECODE(code);
    for (lst=el->adsTable; lst != NULL; lst=lst->next) if ((UINT32)code == lst->ads.alarmCode) return &lst->ads;
    switch (code)
    {
      case invalid_data: return &almErrTblADS[idx_invalid_data];
      case value_too_high: return &almErrTblADS[idx_too_high];
      case value_too_low: return &almErrTblADS[idx_too_low];
      case warn_too_high: return &almErrTblADS[idx_warn_too_high];
      case warn_too_low: return &almErrTblADS[idx_warn_too_low];
      default: 
        if (code > 0 && code < (UINT32)numErr)
        {
          ADS *ads;
          if ((ads=getErrTblADS((short)code)) != NULL) return ads;
          return makeADSFromErrorCode((short)code);
        }
        for (awe=almWatchList; awe != NULL; awe=awe->nxt)
        {
          if (awe->codeHighWarn == (int)code) return &almErrTblADS[idx_warn_too_high];
          if (awe->codeHigh == (int)code) return &almErrTblADS[idx_too_high];
          if (awe->codeLowWarn == (int)code) return &almErrTblADS[idx_warn_too_low];
          if (awe->codeLow == (int)code) return &almErrTblADS[idx_too_low];
          if (awe->code == (int)code) return &almErrTblADS[idx_invalid_data];
        }
        break;
    }
  }
  return NULL;
}
/** 
 * \internal used only when collapsed alarms are in play and terminating ...
 */
void touchActiveAlarms(ExportListStruct *el)
{
  ALARM *alm;
  int dv, n = 0;
  UINT32 tt = 0;
  /* move everyone up 1 second (CAS needs to see the collapsed TERMINATE flag) */
  /* i.e. surveyAlarms() should not be allowed to see anything new until the next second rolls over */
  newAlarmBaseTime = thisCycleTime.tv_sec + 1;
  tt = (UINT32)newAlarmBaseTime;
  for (dv=0; el->almInfo.nalms && dv<el->EqmNumDevices; dv++)
  { 
    if (el->EqmDevLst[dv].isReserved) continue;
    for (alm=el->EqmDevLst[dv].almLst; alm != NULL; alm=alm->next)
    {
      if (alm->alarmStatus & almTERMINATE) continue;
      alm->timestamp = tt;
      alm->timestampUSec = 0;
      n++;
    }
  }
  el->almInfo.mrct = tt;
  el->almInfo.mrt = tt;
  el->almInfo.nmrct = n;
}
/**
 * \internal 'exported' but called generally only by stock property ALARMS or ALARMSEXT
 */
TINE_EXPORT int getAlarmsEx(char *eqmName,AMS *ams,short num,short minSeverity,time_t starttime,time_t stoptime,int devNr)
{
  int dv,i=0,n=0,startnum=0,cc=0,isCASRead=FALSE,isSystemDefault=FALSE,isEncoded=FALSE,isHidden=FALSE;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;
  char defaultDeviceName[DEVICE_NAME_SIZE+4],devstr[DEVICE_NAME_SIZE+4],atag[132];
  char scratch[128], *ctx=NULL, *srv=NULL, *dev=NULL, *atagadd=NULL;
  UINT32 baseCode;

  if ((el=getExportListItem(eqmName)) == NULL || el->EqmDevLst == NULL || num <= 0) return 0;
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return -semaphore_busy;
  if (devNr < -1) 
  { /* a backdoor kluge to enable acquisition of a long burst of alarms piecemeal */
    startnum = -devNr; 
  }
  isCASRead = isCASInUserList(eqmName);
  if (devNr == -1)
  { /* request for all alarms: check collapsed alarm first ... */
    if (gAlmCollapseWindow > 0 && el->almInfo.nalms > gAlmCollapseWindow)
    { /* collapse to a single message (if collapse window > 0) */
      for (dv = 0, n = 0; dv < el->EqmNumDevices; dv++)
      {
        if (el->EqmDevLst[dv].isReserved) continue;
        for (alm = el->EqmDevLst[dv].almLst; alm != NULL; alm = alm->next)
        {
          alm->removalBlocked = 0;
          n++;
        }
      }
      if (n < el->almInfo.nalms) el->almInfo.nalms = n; /* consistency check */
      /* signal that the collapsed alarm will need to be terminated */
      /* this is the only location where we set this to TRUE: */
      collapsedTerminationPending = TRUE; 
      ccerr(prepCollapsedAlarmMessage(ams,el));
    }
    if (collapsedTerminationPending)
    { /* we were in a collapsed state, but not anymore: make sure the TERMINATION gets seen by the CAS */
      if (isCASRead || !gRequireReadFromCAS) 
      {
        struct timeval tv;
        gettimeofday(&tv, (struct timezone *)NULL);
        msglog(0,"collapsed alarms: clear collapsed pending");
        collapsedTerminationPending = FALSE;
        updateCollapsedAlarmMessage(el, &tv); /* sets the TERMINATE bit here */
        touchActiveAlarms(el);
      }
      ccerr(prepCollapsedAlarmMessage(ams,el));
    }
  }
  for (dv=0,n=0; el->almInfo.nalms && dv<el->EqmNumDevices; dv++)
  { /* devNr < 0 => all alarms */
    if (devNr >= 0 && dv != devNr) continue;
    if (el->EqmDevLst[dv].isReserved) continue;
    for (alm=el->EqmDevLst[dv].almLst; alm != NULL; alm=alm->next)
    {
      if (isCASRead) alm->removalBlocked = 0;
      isSystemDefault = isSystemDefaultAlarm(alm->alarmCode);
      memset(devstr,0,DEVICE_NAME_SIZE+4); /* reset the device string */
      atagadd = NULL;
      isEncoded = FALSE;
      if (el->EqmDevLst) strncpy(devstr,el->EqmDevLst[dv].name,DEVICE_NAME_SIZE); 
      baseCode = ALM_BASECODE(alm->alarmCode);
      if (isLinkErrorAlarm(alm->alarmCode))
      {
        baseCode = link_error;
        strncpy(devstr,el->EqmExportName,EXPORT_NAME_SIZE);
        isEncoded = TRUE;
        strncpy(scratch,(char *)alm->alarmData,64); scratch[64] = 0;
        if (tokenizeFullName(scratch,&ctx,&srv,&dev) == 0) atagadd = srv;
      }
      if (isDiskSpaceAlarm(alm->alarmCode))
      {
        strncpy(devstr,el->EqmExportName,EXPORT_NAME_SIZE);
        baseCode = baseCode &0xff;
        isEncoded = TRUE;
      }
      if (isAlarmApplyName(alm->alarmCode))
      {
        sprintf(devstr,"%.31s/%.32s",el->EqmContext,el->EqmExportName);
      }
      if (isHiddenAlarm(alm->alarmCode))
      {
        isHidden = TRUE;
      }
      if (alm->alarmCode == file_error) strncpy(devstr,el->EqmExportName,EXPORT_NAME_SIZE);
      if ((time_t)alm->timestamp >= starttime && (time_t)alm->timestamp <= stoptime &&
           alm->alarmSeverity >= minSeverity )
      {
        if (i++ < startnum) continue;
        /* fill in ams[n] here ... */
        strncpy(ams[n].server,el->EqmExportName,EXPORT_NAME_SIZE);
        sprintf(defaultDeviceName,"DEVICE %d",dv);
        strncpy(ams[n].device,devstr,DEVICE_NAME_SIZE);
        if (strlen(ams[n].device) == 0) strncpy(ams[n].device,defaultDeviceName,DEVICE_NAME_SIZE);
        if ((ads=getAlarmDefinition(eqmName,baseCode)) != NULL)
        { /* found a registered alarm definition */
          strncpy(atag,ads->alarmTag,ALARM_TAG_SIZE);
          if (alm->wte != NULL) 
          { /* alarm was set via a watch table entry */
            if (alm->alarmCode < (UINT32)numErr && strlen(ads->alarmTag) < 11)
            { /* likely to be true: fill in the property that triggered the alarm */
              sprintf(atag,"%.21s %.10s",alm->wte->aws.prp,ads->alarmTag);
            }
          }
          else if (atagadd != NULL)
          { /* something to add on (from a link error) */
            atag[64] = 0; /* make sure of this */
            strcat(atag,": ");
            strncat(atag,atagadd,32);
            if (strlen(atag) > ALARM_TAG_SIZE) strcpy(&atag[ALARM_TAG_SIZE-3],"...");
          }
          else if (isSimpleStringFormat((short)LFMT(ads->alarmDataFormat)) &&
                   (ads->alarmMask & ALM_MASK_DATA_TO_TAG) != 0)
          { /* if so designated via the alarmMask */
            atag[64] = 0; /* make sure of this */
            strcat(atag,": ");
            strncat(atag,(char *)alm->alarmData,32);
          }
          strncpy(ams[n].alarmTag,atag,ALARM_TAG_SIZE);
          ams[n].alarmDataFormat = ads->alarmDataFormat;
          ams[n].alarmDataArraySize = ads->alarmDataArraySize;
          ams[n].alarmSystem = isSystemDefault ? 0 : ads->alarmSystem;
          ams[n].alarmOscillationWindow = ads->alarmOscillationWindow;
          ams[n].alarmOscillationWindowPinned = ads->alarmOscillationWindowPinned;
        }
        else
        {
          ams[n].alarmOscillationWindow = gAlmOscillationWindow;
          ams[n].alarmOscillationWindowPinned = gAlmOscillationWindowPinned;
        }
        ams[n].alarmMask = el->EqmDevLst[dv].region;
        if (isHidden) ams[n].alarmSystem |= ALM_SYSTEM_HIDDEN;
        ams[n].timestamp = alm->timestamp;
        ams[n].timestampUSec = alm->timestampUSec;
        ams[n].starttime = alm->starttime;
        ams[n].starttimeUSec = alm->starttimeUSec;
        /* let the encoded alarms keep their 'distinctions' */
        ams[n].alarmCode = isEncoded ? alm->alarmCode : baseCode;
        memcpy(ams[n].alarmData,alm->alarmData,ALARM_DATA_SIZE);
        ams[n].severity = alm->alarmSeverity;
        ams[n].descriptor = alm->alarmStatus;
        if (ams[n].alarmCode < (UINT32)numErr && strlen(ams[n].alarmTag) == 0)
          strncpy(ams[n].alarmTag,erlst[ams[n].alarmCode],ALARM_TAG_SIZE);
        if (++n >= num) ccerr(n);
      }
    }
  }
  cc = n;
err:
  ReleaseSystemMutex(hAlmTblMutex);
  return cc;
}
TINE_EXPORT int getAlarms(char *eqmName,AMS *ams,short num,short minSeverity)
{
  return getAlarmsEx(eqmName,ams,num,minSeverity,0,0x7fffffffL,0);
}
/**
 * \internal 
 * this looks obsolete ! (not used within the library)
 */
TINE_EXPORT int GetAlarm(char *eqmName,char *devname,AMS *ams) 
{ /* get highest alarm for given device */
  int sc,scmax=-1,dv,isSystemDefault=FALSE,isHidden=FALSE;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;
  UINT32 baseCode;
  char devstr[DEVICE_NAME_SIZE+4];

  if ((el=getExportListItem(eqmName)) == NULL) return non_existent_elem;
  if ((dv=GetDeviceNumber(eqmName,devname)) < 0) return illegal_equipment_number;
  if (el->EqmDevLst == NULL) return not_initialized;
  strncpy(devstr, devname, DEVICE_NAME_SIZE); devstr[DEVICE_NAME_SIZE] = 0;
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) != 0) return semaphore_busy;
  for (alm=el->EqmDevLst[dv].almLst; alm != NULL; alm=alm->next)
  {
    isSystemDefault = isSystemDefaultAlarm(alm->alarmCode);
    isHidden = isHiddenAlarm(alm->alarmCode);
    baseCode = ALM_BASECODE(alm->alarmCode);
    if (isLinkErrorAlarm(alm->alarmCode)) 
    {
      strncpy(devstr,el->EqmExportName,EXPORT_NAME_SIZE);
      baseCode = link_error;
    }
    if (isDiskSpaceAlarm(alm->alarmCode))
    {
      baseCode = baseCode & 0xff;
      strncpy(devstr,el->EqmExportName,EXPORT_NAME_SIZE);
    }
    if ((sc=findSeverity(eqmName,baseCode)) > scmax)
    {
      scmax = sc;
      strncpy(ams->server,el->EqmExportName,EXPORT_NAME_SIZE);
      strncpy(ams->device,devstr,DEVICE_NAME_SIZE);
      if ((ads=getAlarmDefinition(eqmName,baseCode)) != NULL)
      {
        strncpy(ams->alarmTag,ads->alarmTag,ALARM_TAG_SIZE);
        ams->alarmMask = el->EqmDevLst[dv].region;
        ams->alarmDataFormat = ads->alarmDataFormat;
        ams->alarmDataArraySize = ads->alarmDataArraySize;
        ams->alarmSystem = isSystemDefault ? 0 : ads->alarmSystem;
      }
      if (isHidden) ams->alarmSystem |= ALM_SYSTEM_HIDDEN;
      ams->timestamp = alm->timestamp;
      ams->timestampUSec = alm->timestampUSec;
      ams->starttime = alm->starttime;
      ams->starttimeUSec = alm->starttimeUSec;
      ams->alarmCode = baseCode;
      memcpy(ams->alarmData,alm->alarmData,ALARM_DATA_SIZE);
      ams->severity = sc;
      ams->descriptor = alm->alarmStatus;
    }
  }
  ReleaseSystemMutex(hAlmTblMutex);
  return 0;
}
int prepIncomingAMRB(AMRB *amrb,BYTE *amrbstream) 
{ /* structure is an array of long ints */
  memcpy(amrb,amrbstream,AMRB_SIZE);
# ifndef INTEL
  SwapBytes((char *)amrb,4,CF_INT32,RTOL);
# endif
  return 0;
}
int prepOutgoingAMS(BYTE *amsstream,AMS *amslst,char *stag,int length)  
{
  AMS5 ams5;
  int i;
  if (!strcmp(stag,"AMSr4"))
  { /* modern alarm definition structure */
#    ifdef INTEL
      if ((void *)amsstream != (void *)amslst) memcpy(amsstream,amslst,(size_t)length*AMS_SIZE);
#    else
     _systemStructSwap(AMSStructStruct,amsstream,(unsigned char *)amslst,LTOR,length,sizeof(AMS));
#    endif
  }
  else
  { /* old style */
    for (i=0; i<length; i++)
    {
      memset(&ams5,0,sizeof(AMS5));
      strncpy(ams5.server,amslst[i].server,EXPORT_NAME_SHORTSIZE);
      strncpy(ams5.device,amslst[i].device,DEVICE_NAME_SHORTSIZE);
      strncpy(ams5.alarmTag,amslst[i].alarmTag,ALARM_TAG_SHORTSIZE);
      ams5.alarmCode = amslst[i].alarmCode;
      ams5.timestamp = amslst[i].timestamp;
      ams5.alarmMask = amslst[i].alarmMask;
      memcpy(ams5.alarmData,amslst[i].alarmData,ALARM_DATA_SHORTSIZE);
      ams5.alarmDataFormat = amslst[i].alarmDataFormat;
      ams5.alarmDataArraySize = amslst[i].alarmDataArraySize;
      ams5.severity = amslst[i].severity;
      ams5.descriptor = amslst[i].descriptor;
      ams5.alarmSystem = amslst[i].alarmSystem;
#     ifdef INTEL
      memcpy(&amsstream[i*AMS5_SIZE],&ams5,AMS5_SIZE);
#     else
      _systemStructSwap(AMS5StructStruct,&amsstream[i*AMS5_SIZE],(unsigned char *)&ams5,LTOR,1,sizeof(AMS5));
#     endif
    }
  }
  return 0;
}
int prepOutgoingAlarmData(BYTE *datastream,int size,short fmt)
{
# ifdef INTEL
  UNUSED(datastream);
  UNUSED(size);
  UNUSED(fmt);
# else
  SwapBytes((char *)datastream,size,fmt,LTOR);
# endif
  return 0;
}
int prepIncomingAWS(BYTE *awsstream,void *awsp,char *stag,int length)
{
  if (!strcmp(stag, "AWSr4"))
  {
    AWS *aws = (AWS *)awsp;
#   ifdef INTEL
    memcpy(aws, awsstream, (size_t)length * AWS_SIZE);
#   else
    _systemStructSwap(AWSStructStruct, awsstream, (unsigned char*)aws, RTOL, length, sizeof(AWS));
#   endif
    return 0;
  }
  if (!strcmp(stag, "AWSr5"))
  {
    AWLQS *aws = (AWLQS *)awsp;
#   ifdef INTEL
    memcpy(aws, awsstream, (size_t)length * AWLQS_SIZE);
#   else
    _systemStructSwap(AWLQSStructStruct, awsstream, (unsigned char*)aws, RTOL, length, sizeof(AWLQS));
#   endif
    return 0;
  }
  return invalid_structure_tag;;
}
int prepOutgoingAWS(BYTE *awsstream,AWSLstEntry *awslst,char *stag,int length)
{
  int i;
  AWSLstEntry *aw;
  if (!strcmp(stag, "AWSr4"))
  {
    for (i = 0, aw = awslst; i < length && aw != NULL; i++, aw = aw->nxt)
    {
#     ifdef INTEL
      memcpy(&awsstream[i*AWS_SIZE], &aw->aws, AWS_SIZE);
#     else
      _systemStructSwap(AWSStructStruct, &awsstream[i*AWS_SIZE], (unsigned char*)&aw->aws, LTOR, 1, sizeof(AWS));
#     endif
    }
    return 0;
  }
  if (!strcmp(stag, "AWSr5"))
  {
    AWLQS aws;
    char fs[512];
    for (i = 0, aw = awslst; i < length && aw != NULL; i++, aw = aw->nxt)
    {
      memcpy(&aws,aw,AWLQS_CPSIZE);
      getAlarmFilterString(aw->fltr, fs);
      memcpy(aws.fltr, fs, 256);
#     ifdef INTEL
      memcpy(&awsstream[i*AWLQS_SIZE], &aws, AWLQS_SIZE);
#     else
      _systemStructSwap(AWSStructStruct, &awsstream[i*AWLQS_SIZE], (unsigned char*)&aw->aws, LTOR, 1, sizeof(AWLQS));
#     endif
    }
    return 0;
  }
  return invalid_structure_tag;
}
int prepIncomingADS(BYTE *adsstream,ADS *ads)
{
# ifdef INTEL
  memcpy(ads,adsstream,ADS_SIZE);
# else
  _systemStructSwap(ADSStructStruct,adsstream,(BYTE *)ads,RTOL,1,sizeof(ADS));
# endif
  return 0;
}
int prepOutgoingADS(BYTE *adsstream,ADSList *adslst,char *stag,int length)
{
  int i;
  ADS5 ads5;
  ADSList *lst;

  if (!strcmp(stag,"ADSr4"))
  { /* modern alarm definition structure */
    for (i=0, lst=adslst; i<length && lst != NULL; i++, lst=lst->next)
    {
#     ifdef INTEL
      memcpy(&adsstream[i*ADS_SIZE],&lst->ads,ADS_SIZE);
#     else
      _systemStructSwap(ADSStructStruct,&adsstream[i*ADS_SIZE],(unsigned char *)&lst->ads,LTOR,1,sizeof(ADS));
#     endif
    }
  }
  else
  { /* old style */
    for (i=0, lst=adslst; i<length && lst != NULL; i++, lst=lst->next)
    {
      memset(&ads5,0,sizeof(ADS5));
      ads5.alarmCode = lst->ads.alarmCode;
      ads5.alarmDataArraySize = lst->ads.alarmDataArraySize;
      ads5.alarmDataFormat = lst->ads.alarmDataFormat;
      ads5.alarmMask = lst->ads.alarmMask;
      ads5.alarmSystem = lst->ads.alarmSystem;
      strncpy(ads5.alarmTag,lst->ads.alarmTag,ALARM_TAG_SHORTSIZE);
      strncpy(ads5.alarmText,lst->ads.alarmText,ALARM_TEXT_SHORTSIZE);
      strncpy(ads5.deviceText,lst->ads.deviceText,ALARM_TEXT_SHORTSIZE-2);
      ads5.severity = lst->ads.alarmSeverity;
      strncpy(ads5.url,lst->ads.url,ALARM_TEXT_SHORTSIZE);
#     ifdef INTEL
      memcpy(&adsstream[i*ADS5_SIZE],&ads5,ADS5_SIZE);
#     else
      _systemStructSwap(ADS5StructStruct,&adsstream[i*ADS5_SIZE],(unsigned char *)&ads5,LTOR,1,sizeof(ADS5));
#     endif
    }
  }
  return 0;
}
#ifndef SMALL_TINE_LIB /* for DOS, etc. ... */
int prepIncomingAlmFilter(BYTE *filterstream, FilterStruct *flst)
{
# ifdef INTEL
  memcpy(flst, filterstream, FILTERSTRUCT_SIZE);
# else
  _systemStructSwap(FilterStructStruct,filterstream,(BYTE *)flst,RTOL,1,sizeof(FilterStruct));
# endif
  return 0;
}
int prepOutgoingAlmFilter(BYTE *filterstream, ADSList *adslst, int length)
{
  int i;
  ADSList *lst;
  FilterStruct flst;

  for (i = 0, lst = adslst; i<length && lst != NULL; i++, lst = lst->next)
  {
    char filterString[512];
    flst.filterIdx = lst->ads.alarmCode;
    getAlarmFilterString(lst->fltr, filterString);
    strncpy(flst.filterLinkStr, filterString, 256);
#   ifdef INTEL
    memcpy(&filterstream[i * FILTERSTRUCT_SIZE], &flst, FILTERSTRUCT_SIZE);
#   else
    _systemStructSwap(FilterStructStruct,&filterstream[i*FILTERSTRUCT_SIZE],(unsigned char *)&flst,LTOR,1,sizeof(FilterStruct));
#   endif
  }
  return 0;
}
/* provide API registration as alternative to -alarms.csv file : */
TINE_EXPORT int AppendAlarmInfoTable(char *eqm, ADS *ads)
{
  return AppendAlarmInfoTableEx(eqm, ads, NULL);
}
TINE_EXPORT int AppendAlarmInfoTableEx(char *eqm,ADS *ads,char *filterString)
{
  ExportListStruct *el;
  ADSList *lst=NULL;
  int cc=0;
  FilterLink *fltr = getFilterLinkFromParsableString(filterString);
  if (eqm == NULL || ads == NULL) return argument_list_error;
  if (ads->alarmCode >= 512 && getAlarmDefinition(eqm, ads->alarmCode) != NULL) ccerr(already_assigned);
  if ((el=getExportListItem(eqm)) == NULL) ccerr(non_existent_elem);
  if ((lst=(ADSList *)SystemCalloc(1,sizeof(ADSList))) == NULL) ccerr(out_of_server_memory);
  if (ads->alarmDataArraySize > 0 && !isAllowedAlarmDataFormat(ads->alarmDataFormat)) ccerr(illegal_format);
  if (ads->alarmOscillationWindow == 0) ads->alarmOscillationWindow = gAlmOscillationWindow;
  if (strlen(ads->alarmTag) == 0) strncpy(ads->alarmTag,ads->alarmText,ALARM_TAG_SIZE);
  memcpy(&lst->ads,ads,sizeof(ADS));
  lst->fltr = fltr;
  lst->ads.alarmOscillationWindowPinned = 0;
  lst->next = el->adsTable;
  el->adsTable = lst;
  el->adsTableSize++;
  if (lst->ads.alarmSystem) el->alarmSystem = 1;
err:
  if (cc != 0) feclog("LAS: alarm info for %.32s (%d) : %.32s",
    ads->alarmTag,ads->alarmCode,cc2str(cc));
  return cc;
}
TINE_EXPORT int ApplyAlarmWatchFilter(char *eqm,char *prp,char *dev,char *parsableFilterString)
{
  AWSLstEntry *lst = NULL;
  AWS aws;
  FilterLink *fltr = getFilterLinkFromParsableString(parsableFilterString);
  int removeFilter=FALSE;
  char fltrStr[512];
  if (parsableFilterString == NULL) removeFilter = TRUE;
  else if (fltr == NULL) return parsing_error;

  for (lst=almWatchList; lst != NULL; lst=lst->nxt)
  { /* find the entry */
    if (strncmp(lst->aws.eqm,aws.eqm,EQM_NAME_SIZE)) continue;
    if (strnicmp(lst->aws.dev,aws.dev,DEVICE_NAME_SIZE)) continue;
    if (strnicmp(lst->aws.prp,aws.prp,PROPERTY_NAME_SIZE)) continue;
    if (removeFilter)
    {
      strcpy(fltrStr,"remove any applied filters");
      if (lst->fltr != NULL)
      {
        rmvFilterLink(lst->fltr);
      }
    }
    else
    {
      sprintf(fltrStr,"apply filter %.192s %.2s %.64s",
        fltr->key,getComparatorString(fltr->fcmp),fltr->fstr);
    }
    lst->fltr = fltr;
    feclog("LAS: %.6s/%.64s[%.64s] -> %s",eqm,dev,prp,fltrStr);
    return 0;

  }
  return un_allocated;
}
TINE_EXPORT int AppendAlarmWatchTable(char *eqm,char *prp,char *dev,int siz,int fmt,int atyp,int sev,int sys,ALM_THRESHOLDS *thr)
{
  int cc = 0;
  AWSLstEntry *aw;
  ADS ads;
  if (eqm == NULL || prp == NULL || dev == NULL || thr == NULL) return argument_list_error;
  if ((aw=(AWSLstEntry *)SystemCalloc(1,sizeof(AWSLstEntry))) == NULL) ccerr(out_of_server_memory);
  if (!isAllowedAlarmValueFormat(fmt)) ccerr(illegal_format);
  strncpy(aw->aws.eqm,eqm,EQM_NAME_SHORTSIZE);
  strncpy(aw->aws.prp,prp,PROPERTY_NAME_SIZE);
  strncpy(aw->aws.dev,dev,DEVICE_NAME_SIZE);
  aw->aws.siz = siz;
  aw->aws.fmt = fmt;
  aw->aws.atyp = atyp;
  aw->aws.sev = sev;
  aw->aws.asys = sys;
  aw->aws.cntThreshold = thr->cnt;
  if (thr->mask != 0)
  {
    aw->aws.mask = thr->mask;
    aw->aws.normal = thr->normal;
    aw->aws.chkNormal = thr->normalIsAlarm ? -1 : 1;
  }
  aw->aws.hi = thr->hi;
  aw->aws.hiwarn = thr->hiwarn;
  aw->aws.lo = thr->lo;
  aw->aws.lowarn = thr->lowarn;
  aw->aws.hisev = thr->hisev > 0 ? thr->hisev : aw->aws.sev;
  aw->aws.hiwarnsev = thr->hiwarnsev > 0 ? thr->hiwarnsev : aw->aws.hisev - 2;
  aw->aws.losev = thr->losev > 0 ? thr->losev : aw->aws.sev;
  aw->aws.lowarnsev = thr->lowarnsev > 0 ? thr->lowarnsev : aw->aws.losev - 2;
  aw->code = thr->code;
  aw->codeHigh = thr->codeHigh;
  aw->codeLow = thr->codeLow;
  aw->codeLowWarn = thr->codeLowWarn;
  snprintf(aw->tagNormal, ALARM_TAG_SIZE, "%.23s invalid", prp);
  snprintf(aw->tagHigh, ALARM_TAG_SIZE, "%.22s too high", prp);
  snprintf(aw->tagHighWarn, ALARM_TAG_SIZE, "%.20s near limit", prp);
  snprintf(aw->tagLow, ALARM_TAG_SIZE, "%.23s too low", prp);
  snprintf(aw->tagLowWarn, ALARM_TAG_SIZE, "%.20s near limit", prp);
  aw->nxt = almWatchList;
  almWatchList = aw;
  nAlmWatchList++;
  if (aw->code > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads, &almErrTblADS[idx_invalid_data], aw->code, aw->tagNormal), NULL);
  if (aw->codeHigh > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads, &almErrTblADS[idx_too_high], aw->codeHigh, aw->tagHigh), NULL);
  if (aw->codeHighWarn > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads, &almErrTblADS[idx_warn_too_high], aw->codeHighWarn, aw->tagHighWarn), NULL);
  if (aw->codeLow > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads, &almErrTblADS[idx_too_low], aw->codeLow,aw->tagLow), NULL);
  if (aw->codeLowWarn > 0) AppendAlarmInfoTableEx(eqm, cloneADS(&ads, &almErrTblADS[idx_warn_too_low], aw->codeLowWarn, aw->tagLowWarn), NULL);

err:
  feclog("LAS: append %.6s/%.64s[%.64s] to alarm watch table : %.32s",eqm,dev,prp,cc2str(cc));
  return cc;
}
#endif /* SMALL_TINE_LIB */
void lasClearAlarm(char *eqm,int devNr,UINT32 code,AWSLstEntry *aws)
{ /* only applies to internally set alarms (mask off any high bits) */
  ALARM *alm;
  ExportListStruct *el;

  if (!AlarmServerInitialized) return;
  if ((el=getExportListItem(eqm)) == NULL ||el->EqmDevLst == NULL) return;
  if (verifyDeviceNr(el,devNr)) return;

  for (alm=el->EqmDevLst[devNr].almLst; alm != NULL; alm=alm->next)
  {
    if (code == file_error)
    { /* mark for immediate removal */
      alm->clrCount = ALM_OSCILLATION_MAXIMUM;
    }
    else if (code == (alm->alarmCode & 0xffff) && aws == alm->wte)
    { /* increment the clear count */
      alm->clrCount = MIN(alm->clrCount+1,ALM_OSCILLATION_MAXIMUM); 
    }
  }
  return;
}
TINE_EXPORT void ClearAlarm(char *eqm,int devNr)
{
  ALARM *alm;
  ExportListStruct *el;
  int start,stop,i;

  if (!AlarmServerInitialized) return;

  if ((el=getExportListItem(eqm)) == NULL || el->EqmDevLst == NULL || el->isidle) return;
  if (devNr == -1)                        /* clear them all */
  {
    start = 0; stop = el->EqmNumDevices;
  }
  else
  {
    if (verifyDeviceNr(el,devNr)) return;
    if (el->EqmDevLst[devNr].isReserved) return;
    start = devNr; stop = devNr + 1;
  }
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) == 0)
  {
    for (i=start; i<stop; i++)
    {
      for (alm=el->EqmDevLst[i].almLst; alm != NULL; alm=alm->next)
      {
        if (alm->alarmStatus & almTRANSIENT) alm->clrCount = ALM_OSCILLATION_MAXIMUM; /* it'll be cleared immediately */
        else alm->clrCount = MIN(alm->clrCount+1,ALM_OSCILLATION_MAXIMUM); /* increment the clear count */
      }
    }
    ReleaseSystemMutex(hAlmTblMutex); 
  }
  return;                                 /* wasn't set */
}
TINE_EXPORT void ClearDeviceAlarm(char *eqm,char *devname)
{
  int devNr = -1;
  if (strcmp(devname,"ALL")) devNr = GetDeviceNumber(eqm,devname);
  ClearAlarm(eqm,devNr);
}
TINE_EXPORT int GetAlarmHeartbeat(void) { return AlarmHeartbeat; }
TINE_EXPORT int GetAlarmTableMaximumSize(void) { return MaxNumberAlarms; }
TINE_EXPORT int GetAlarmTerminationWindow(void) { return TerminationWindow; }
TINE_EXPORT int GetNewAlarmWindow(void) { return AlarmWindow; }
TINE_EXPORT int GetAlarmDataChangeWindow(void) { return AlarmDataChangeWindow; }
TINE_EXPORT void RemoveAlarm(char *eqm,int devNr,UINT32 code)
{ /* clear it immediately */
  ALARM *alm;
  ExportListStruct *el;
  int found = FALSE;

  if (!AlarmServerInitialized) 
  {
    if (almdbg) dbglog("ALM: RemoveAlarm failed : alarm service not initialized");
    goto err;
  }
  if ((el=getExportListItem(eqm)) == NULL || el->isidle) 
  {
    if (almdbg) dbglog("ALM: RemoveAlarm failed : equipment module %s",el == NULL ? "does not exist" : "is idle");
    goto err;
  }
  if (verifyDeviceNr(el,devNr)) 
  {
    if (almdbg) dbglog("ALM: RemoveAlarm failed : %.32s (%.6s) device %d is not valid",
      el->EqmExportName, el->EqmName, devNr);
    goto err;
  }
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) == 0)
  {
    for (alm=el->EqmDevLst[devNr].almLst; alm != NULL; alm=alm->next)
    {
      if (code == alm->alarmCode)
      {
        alm->clrCount = ALM_OSCILLATION_MAXIMUM;
        found = TRUE;
      }
    }
    ReleaseSystemMutex(hAlmTblMutex); 
  }
  if (!found && almdbg > 2) 
    dbglog("ALM: RemoveAlarm %.32s (%.6s) device %d no action : alarm code %d is not an active alarm",
      el->EqmExportName, el->EqmName, devNr, code);
err:
  return; 
}
TINE_EXPORT void RemoveDeviceAlarm(char *eqm,char *devname,UINT32 code)  
{ /* clear it immediately */
  int devNr = GetDeviceNumber(eqm,devname);
  if (eqm != NULL && devNr < 0) 
  {
    ExportListStruct *el=getExportListItem(eqm);
    if (el != NULL && el->inidone)
    { /* spit this out in the log file */
      feclogEx(TINE_LOGLEVEL_WARN, "ALM: RemoveDeviceAlarm failed : device %.64s is not valid for %.6s",devname,eqm);
    }
  }
  RemoveAlarm(eqm,devNr,code);
}
TINE_EXPORT int SetAlarm(char *eqm,int devNr,UINT32 almCode,BYTE *almData)
{
  return SetAlarmEx(eqm,devNr,almCode,almData,almNEWALARM);
}
TINE_EXPORT int SetAlarmEx(char *eqm,int devNr,UINT32 almCode,BYTE *almData,BYTE flags)
{
  double ts = makeDataTimeStamp();
  int cc = semaphore_busy;
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) == 0)
  {
    cc = _setAlarm(eqm,devNr,almCode,almData,flags,ts,NULL);
    ReleaseSystemMutex(hAlmTblMutex); 
  }
  return cc;
}
TINE_EXPORT int SetDeviceAlarm(char *eqm,char *devname,UINT32 almCode,BYTE *almData,BYTE flags)
{
  ExportListStruct *el = getExportListItem(eqm);
  int devNr = GetDeviceNumber(eqm, devname);
  if (el == NULL) return non_existent_elem;
  if (!strnicmp(devname, el->EqmExportName, EXPORT_NAME_SIZE))
  { /* set as a general device server alarm : rides on device #0 */
    devNr = 0;
    almCode |= ALM_APPLYNAME;
  }
  return SetAlarmEx(eqm,devNr,almCode,almData,flags);
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT void isCentralAlarmServer(int value) { gIsCentralAlarmServer = value; }
TINE_EXPORT int SetAlarmSurveyInterval(UINT32 interval)
{
  if (interval < gSystemTick) return out_of_range;
  gAlarmSurveyInterval = interval;
  return 0;
}
TINE_EXPORT int GetAlarmSurveyInterval(void)
{
  return gAlarmSurveyInterval;
}
TINE_EXPORT int SetAlarmOscillationWindow(int value)
{
  if (value < ALM_OSCILLATION_WINDOW) return out_of_range;
  gAlmOscillationWindow = value;
  gAlmOscillationWindowPinned = TRUE;
  feclog("ALM: alarm oscillation window pinned at %d clear counts",gAlmOscillationWindow);
  return 0;
}
TINE_EXPORT int GetAlarmOscillationWindow(void)
{
  return gAlmOscillationWindow;
}
TINE_EXPORT int SetAlarmCodeOscillationWindow(const char *eqm,int code, int value)
{
  ADS *ads = getAlarmDefinition((char *)eqm,code);
  if (ads == NULL) return non_existent;
  if (value < ALM_OSCILLATION_WINDOW) return out_of_range;
  ads->alarmOscillationWindow = (BYTE)value;
  ads->alarmOscillationWindowPinned = (BYTE)0xff;
  feclog("ALM: %.6s 0x<%x> alarm oscillation window pinned at %d clear counts",eqm,code,value);
  return 0;
}
TINE_EXPORT int GetAlarmCodeOscillationWindow(const char *eqm,int code)
{
  ADS *ads = getAlarmDefinition((char *)eqm,code);
  if (ads == NULL) return -non_existent;
  return ads->alarmOscillationWindow;
}
TINE_EXPORT int SetAlarmCollapseWindow(int value)
{
  if (value > 0 && value < 5) return out_of_range;
  if (value <= 0) feclog("ALM: collapse window disabled");
  if (gAlmCollapseWindow <= 0 && value > 0) feclog("ALM: collapse window set to %d",value);
  gAlmCollapseWindow = value;
  return 0;
}
TINE_EXPORT int GetAlarmCollapseWindow(void)
{
  return gAlmCollapseWindow;
}
TINE_EXPORT void SetAlarmHeartbeat(int timeInSeconds) 
{ 
  if (timeInSeconds < 5) return;
  AlarmHeartbeat = timeInSeconds; 
}
TINE_EXPORT void SetAlarmTableMaximumSize(int tableSize) 
{ 
  if (tableSize < 10) return;
  MaxNumberAlarms = tableSize; 
}
TINE_EXPORT void SetAlarmTerminationWindow(int timeInSeconds) 
{ 
  if (timeInSeconds < 5) return;
  if (timeInSeconds > 3600) return;
  TerminationWindow = timeInSeconds; 
}
TINE_EXPORT void SetNewAlarmWindow(int timeInSeconds) 
{ 
  if (timeInSeconds < 1) return;
  AlarmWindow = timeInSeconds; 
}
TINE_EXPORT void SetAlarmDataChangeWindow(int timeInSeconds)
{
  if (timeInSeconds < 1) return;
  AlarmDataChangeWindow = timeInSeconds;
}
TINE_EXPORT void SetAMSCapacity(int value)
{
  gMaxAMSCapacity = value;
}
#endif /* SMALL_TINE_LIB */

#ifndef SMALL_TINE_LIB
int dumpAlarmManifest(char *eqm,int to_stdout)
{
  int cc = 0;
# if defined(FS_DELIMITER) && !defined(FS_RDONLY)
# define almmfhdr "AlarmTag, AlarmCode, AlarmMask, Severity, DataFormat, DataArraysize, AlarmText, DeviceText, Url, AlarmSystem, Filter"
  char fs[512];
  ExportListStruct *el=getExportListItem(eqm);
  ADSList *adl;
  FILE *fp = NULL;
  if (el == NULL) ccerr(non_existent_elem);
  if (el->adsTable == NULL) ccerr(0);
  if ((fp=getMfFile(eqm,"w",ALARM_MANIFEST)) == NULL) ccerr(file_error);
  fprintf(fp, "%s\n", almmfhdr);
  for (adl=el->adsTable; adl != NULL; adl=adl->next)
  {
    getAlarmFilterString(adl->fltr, fs);
    fprintf(fp, "%.32s, %d, %d, %d, %s, %d, %.64s, %.64s, %.128s, %d, %.256s\n",
      adl->ads.alarmTag, adl->ads.alarmCode, adl->ads.alarmMask,
      adl->ads.alarmSeverity, GetFormatAsString(adl->ads.alarmDataFormat), 
      adl->ads.alarmDataArraySize,adl->ads.alarmText, adl->ads.deviceText, 
      adl->ads.url, adl->ads.alarmSystem,fs);
  }
  if (to_stdout) dbglog("local alarm info written");
err:
  if (fp != NULL) fclose(fp);
# endif
  return cc;
}
int dumpAlarmWatchManifest(char *eqm,int to_stdout)
{
  int cc = 0;
# if defined(FS_DELIMITER) && !defined(FS_RDONLY)
# define almwatmfhdr "LocalName, DeviceName, Property, Size, Format, Severity, SeverityHigh, SeverityLow, SeverityHighWarn, SeverityLowWarn, " \
                     "AlarmSystem, Mask, Normal, CountThreshold, High, Low, HighWarn, LowWarn, AlarmCode, AlarmCodeHigh, AlarmCodeHighWarn, " \
                     "AlarmCodeLow, AlarmCodeLowWarn, AlarmTagNormal, AlarmTagHigh, AlarmTagHighWarn, AlarmTagLow, AlarmTagLowWarn, Filter"
  char fs[512];
  AWSLstEntry *awe;
  FILE *fp = NULL;
  if (almWatchList == NULL) ccerr(0);
  for (awe=almWatchList; awe != NULL; awe=awe->nxt)
  {
    if (eqm != NULL && strlen(eqm) > 0 && strncmp(eqm, awe->aws.eqm, EQM_NAME_SHORTSIZE) != 0) continue;
    if (fp == NULL)
    {
      if ((fp=getMfFile(eqm, "w", ALARM_WATCH_MANIFEST)) == NULL) ccerr(file_error);
      fprintf(fp, "%s\n", almwatmfhdr);
    }
    getAlarmFilterString(awe->fltr, fs);
    fprintf(fp, "%.6s, %.64s, %.64s, %d, %s, %d, %d, %d, %d, %d, %d, %d, %d, %d, %f, %f, %f, %f, %d, %d, %d, %d, %d, %.32s, %.32s, %.32s, %.32s, %.32s, %.256s\n", 
      awe->aws.eqm, awe->aws.dev, awe->aws.prp, awe->aws.siz, GetFormatAsString(awe->aws.fmt),
      awe->aws.sev, awe->aws.hisev, awe->aws.losev, awe->aws.hiwarnsev, awe->aws.lowarnsev,
      awe->aws.asys, awe->aws.mask, awe->aws.normal, awe->aws.cntThreshold, awe->aws.hi,
      awe->aws.lo, awe->aws.hiwarn, awe->aws.lowarn, awe->code, awe->codeHigh, awe->codeHighWarn, 
      awe->codeLow, awe->codeLowWarn, awe->tagNormal, awe->tagHigh, awe->tagHighWarn, awe->tagLow, awe->tagLowWarn, fs);
  }
  if (to_stdout) dbglog("local alarm watch info written");
err:
  if (fp != NULL) fclose(fp);
#endif
  return cc;
}
void dumpAlmManifest(int mode,int to_stdout)
{
  ExportListStruct *el;

  feclog("dump alarm and alarm watch manifest");
  for (el = ExportList; el != NULL; el = el->next)
  {
    if(mode == ALM_MF_ALL) dumpAlarmManifest(el-> EqmName,to_stdout);
    if(mode == ALM_MF_WAT || mode == ALM_MF_ALL) dumpAlarmWatchManifest(el-> EqmName,to_stdout);
  }
  return;
}
void getAlarmFilterString(FilterLink *fltr, char *filterString)
{
  if (fltr != NULL)
  {
    int c = fltr->fcmp;
    int altsev = fltr->param1;
    strncpy(filterString, fltr->key, 256); filterString[255] = 0;
    switch (c)
    {
      case fl_eq:
        strcat(filterString, "=");
        break;
      case fl_neq:
        strcat(filterString, "!=");
        break;
      case fl_lt:
        strcat(filterString, "<");
        break;
      case fl_gt:
        strcat(filterString, ">");
        break;
      default:
        break;
    }
    strncat(filterString, fltr->fstr, 64);
    if (altsev >= 0)
    {
      char s[64];
      sprintf(s, "&sev=%d", altsev);
      strcat(filterString, s);
    }
  }
  else
  {
    strcpy(filterString, "NONE");
  }
}

#endif /* SMALL_TINE_LIB */
