/*
 * File clnip.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file clnip.c
 *
 * some client-side IP routines ...
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "prolog.h"
#include "srvcore.h"
#include "srvip.h"
#include "globals.h"
#include "dbtools.h"
#include "ttycmd.h"
#if defined(WIN32) && !defined(NO_TINE_PIPE)
#include "shmalloc.h"
#endif
#ifdef UNIX
# ifdef NOTTY
#   define USE_FGTTY FALSE
# endif
# ifndef NO_SERVICE_PIPE
void chkIpcClnLst(struct pollfd *fds,int nfds);
# endif
#endif /* UNIX */
#ifndef USE_FGTTY
# define USE_FGTTY FALSE
#endif
int foregroundTTY = USE_FGTTY; /* primarily of use in UNIX builds, but passed to SystemCycle() in srvTmrTask() */
SOCKET udpClnSck = 0;
SOCKET mcastClnSck = 0;
int hasClientEvents = 0;
extern int hasServerEvents;
extern INADR fecAddrMcGrp;
#define hasEvents (hasClientEvents || hasServerEvents)
int gListenForMCasts = FALSE;
#ifndef LISTEN_FOR_GLOBALS
# define LISTEN_FOR_GLOBALS TRUE
#endif
#ifndef CLIENT_BURST_LIMIT
# define CLIENT_BURST_LIMIT 1000
#endif
int gListenForGlobals = LISTEN_FOR_GLOBALS;
struct timeval conto;
SOCKET glbClnSck = 0;
int defaultClientTimeout = SYSTEM_TICK;
int clientTimeout = 0;
int clientTimeoutMsec = 0;
int ReturnToClientCycle = 0;
int gInitIPClientDone = FALSE;
int gClientBurstLimit = CLIENT_BURST_LIMIT;
EpAdr gInAdr;
extern int gInitIPServicesDone;
IDLE_CONNECTION *tcpIdleConnectionList = NULL;
#ifdef WIN32
void chkWinPipe(void);
#endif

TINE_EXPORT void SetInterpretConsoleCommands(int value)
{
  foregroundTTY = value;
}
TINE_EXPORT int GetInterpretConsoleCommands(void)
{
  return foregroundTTY;
}
TINE_EXPORT void SetClientBurstLimit(int npackets)
{
  if (npackets >= 1) 
  {
    gClientBurstLimit = npackets;
  }
}
TINE_EXPORT void SetClientTimeout(int timeoutMilliseconds)
{
  if (defaultClientTimeout >= 0) 
  {
    clientTimeout = timeoutMilliseconds/1000;
    clientTimeoutMsec = timeoutMilliseconds%1000;
  }
}
void setupGlobalsSocket(SOCKET *psck,UINT16 p,int isGlbSck)
{
  int rcvbufs = max_clnrcvbuf, sndbufs = max_udpsndbuf;
  char *txt = isGlbSck ? "GLOBALS" : "MCAST";
  if (isGlbSck)
  { /* either a globals port (consumer) or a server port (subscriber) */
    rcvbufs = MAX_RCVBUF;
    sndbufs = MAX_SNDBUF;
  }
  if (setupUdpSocket(psck,gAfInet,p,FALSE,rcvbufs,sndbufs)) 
  {
    *psck = 0;
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot open %s socket",txt);
  }
# ifdef MULTITHREADED
  else
  {
    if (isGlbSck) 
    { /* general globals receiving port */
      hGlbThread = glbRecvCreateThread(0);
    }
    else 
    { /* server multicast subscription port */
      hMcstThread = mcstRecvCreateThread(0);
    }
  }
# endif
}
/**
 * \internal
 * \brief Initialize file descriptor sets and bind sockets for an IP client.
 *
 *
 * Binds sockets for the consumer, MCast, and GCast,
 * and adds these sockets to conset.
 * 
 * Initialize the file descriptor set ext-set.
 *
 * Also adds stdin to conset if stdin is a tty.
 * 
 * \param port consumer port number.
 * \todo describe MCast and GCast.
 * \todo document global fd sets.
 */
int initIPClient(unsigned short port)
{
  int cc = 0;
  int optlvl, opt, optval = 1, optlen = sizeof(int);
  if (defaultClientTimeout <= 0) defaultClientTimeout = getClockTickMilliseconds();
  if (gInitIPClientDone) return 0; 
  gInitIPClientDone = TRUE;
# ifndef DOS
  ipLoaded = TRUE;
# ifdef WIN32
  if (!WinsockLoaded()) ipLoaded = 0;
  MaxPollingRate = MIN_POLLRATE; 
# endif /* WIN32 */
# else /* is DOS : */
  if (!ipLoaded) ccerr(tcp_not_supported);
  defaultClientTimeout = 0;
# endif /* not DOS */
  initIpFromEnv();
  clientTimeout = defaultClientTimeout/1000;
  clientTimeoutMsec = defaultClientTimeout%1000;
  if (MaxPollingRate > 1000) MaxPollingRate = 1000;
  if (gSystemCycleDeadband > 1000) gSystemCycleDeadband = 1000;
  /* set up UDP Consumer Sockets */
  conto.tv_sec = clientTimeout; conto.tv_usec = clientTimeoutMsec*1000;
  if (setupUdpSocket(&udpClnSck,gAfInet,port,TRUE,max_clnrcvbuf,max_udpsndbuf) != 0) ccerr(udp_socket_error);
# ifndef DOS
  if (setsockopt(udpClnSck,SOL_SOCKET,SO_BROADCAST,(char *)&optval,optlen))
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot set broadcast option on consumer socket !");
  }
# endif /* DOS */
# ifndef NOT_MULTICASTING
  {
    int ttl = gtMCastTTL;
    char *optval = (char *)&ttl;
#   ifdef UNIX
    BYTE bttl = gtMCastTTL;
    optval = gAfInet == AF_INET6 ? (char *)&ttl : (char *)&bttl;
    optlen = gAfInet == AF_INET6 ? sizeof(ttl) : sizeof(bttl);
#   endif
    setmcopts(optlvl,opt);
    if (setsockopt(udpClnSck,optlvl,opt,optval,optlen) < 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot set multicast time to live on consumer socket!");
    }
  }
# endif
  fecAddrMcGrp = inadr_zero;
  if (gListenForMCasts && mcastClnSck == 0)
  { /* server multicast socket */
    setupGlobalsSocket(&mcastClnSck,(UINT16)gtMCastPort,FALSE);
    if (mcastClnSck == 0) gListenForMCasts = 0;
  }
  if (gListenForGlobals && glbClnSck == 0) 
  { /* standard globals socket */
    setupGlobalsSocket(&glbClnSck,(UINT16)gtGCastPort,TRUE);
    if (glbClnSck == 0) gListenForGlobals = 0;
  }
err:
  if (cc)
  {
    feclog("Consumer socket : %s",cc2str(cc));
  }
  return cc;
}
int ListenForIPMulticasts(void)
{
  int cc = 0;
  if (gListenForMCasts) ccerr(0); /* is already listening for multicasts */
  gListenForMCasts = TRUE;
  initClient();
  if (!ipLoaded) ccerr(tcp_not_supported);
  if (mcastClnSck == 0) setupGlobalsSocket(&mcastClnSck,(UINT16)gtMCastPort,FALSE);
  if (mcastClnSck == 0) ccerr(tcp_socket_error);
err:
  if (cc) gListenForMCasts = FALSE;
  return cc;
}
int ListenForIPNetGlobals(void)
{
  int cc = 0;
  if (gListenForGlobals) ccerr(0)
  if (!ipLoaded) ccerr(tcp_not_supported);
  setupGlobalsSocket(&glbClnSck,(UINT16)gtGCastPort,TRUE);
  if (glbClnSck == 0) ccerr(tcp_socket_error);
  gListenForGlobals = TRUE;
err:
  if (cc) gListenForGlobals = FALSE;
  return cc;
}
void closeIPClients(void)
{
  int i,k;
  if (!ipLoaded) return;
  if (udpClnSck != 0) soclose(udpClnSck);
  udpClnSck = 0;
  if (gListenForMCasts)
  {
    if (mcastClnSck != 0) soclose(mcastClnSck);
  }
  mcastClnSck = 0;
  if (glbClnSck != 0) soclose(glbClnSck);
  glbClnSck = 0;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    if (conTbl[i]->tcpSck != 0)
    {
      shutdown(conTbl[i]->tcpSck,2);
      soclose(conTbl[i]->tcpSck);
      for (k=i+1; k<nConnectionTableEntries; k++)
        if (conTbl[k]->tcpSck == conTbl[i]->tcpSck) conTbl[k]->tcpSck = 0;
    }
    conTbl[i]->tcpSck = 0;
    feclog("%s link to /%.32s/%.32s/%.64s[%.64s] terminated",
      getProtocolAsString(conTbl[i]->inetProtocol),
      conTbl[i]->ctxName,conTbl[i]->expName,conTbl[i]->devName,conTbl[i]->prpName);
  }
  freeBucketList();
  return;
}

void closeIPConnectionToFec(int index,SOCKET sck)
{
  int i;
  if (!ipLoaded) return;
  if (!(FecTbl[index].inetProtocol&(UDP|UDP6|TCP|TCP6|STREAM|STREAM6))) return;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    if (index != conTbl[i]->fecIdx) continue;
    conTbl[i]->tcpSck = 0; /* signal for new connection */
    if ((conTbl[i]->mode & (CM_CONNECT|CM_STREAM)) == 0) continue;
    feclog("%s connection to /%.32s/%.32s/%.64s[%.64s] terminated",
      getProtocolAsString(conTbl[i]->inetProtocol),
      conTbl[i]->ctxName,conTbl[i]->expName,conTbl[i]->devName,conTbl[i]->prpName);
  }
  if (sck != (SOCKET)0)
  {
    shutdown(sck,2);
    soclose(sck);
  }
  rmvTcpClnSck(sck);
  return;
}
void checkSckBufferSpace(SOCKET sck,int size,int opt,char *tag)
{
  int optval = 1;
  static char last_msg[256] = "";
  _socklen_t optlen = sizeof(_socklen_t);
  if (getsockopt(sck,SOL_SOCKET,opt,(char *)&optval,&optlen) == 0)
  {
    if (optval < size)
    {
      char msg[256];
      snprintf(msg,255,"%s %s: requested %d bytes buffer space, assigned %d bytes !",
        tag, (opt == SO_RCVBUF ? "receive" : "send"), size, optval);
      msg[255] = 0;
      if (strncmp(last_msg, msg, 255))
      { /* only log it if it's a different message */
        feclog(msg);
        strncpy(last_msg, msg, 256);
      }
    }
  }
}
TCPSCKT *tcpClnSckTbl = NULL;
TCPSCKT *getTcpClnSck(SOCKET sck, int index, int type)
{
  int cc = 0;
  TCPSCKT *sckt = NULL;
  SuspendSystemMutex(hLinkLstMutex);
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return NULL;
  sckt = tcpClnSckTbl;
  while (sckt != NULL)
  {
    if (sckt->sck == sck) ccerr(0);
    sckt = sckt->nxt;
  }
  if (index < 0 || type < 0)
  { /* call used to find existing entry came up empty */
    ccerr(0); 
  }
  if ((sckt=SystemCalloc(1, sizeof(TCPSCKT))) != NULL)
  {
    sckt->sck = sck;
    sckt->type = type;
    sckt->idx = index;
    sckt->nxt = tcpClnSckTbl;
    tcpClnSckTbl = sckt;
  }
  else
  {
    cc = out_of_local_memory;
  }
err:
  ReleaseSystemMutex(hLinkLstMutex);
  RestoreSystemMutex(hLinkLstMutex);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot get TCP client socket: %.32s",cc2str(cc));
  return sckt;
}
void rmvTcpClnSck(SOCKET sck)
{
  TCPSCKT *sckt = NULL, *prv = NULL;
  int cc = 0;
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return;
  sckt = tcpClnSckTbl;
  while (sckt != NULL)
  {
    if (sckt->sck == sck)
    {
      if (prv == NULL) tcpClnSckTbl = sckt->nxt; else prv->nxt = sckt->nxt;
      SystemFree(sckt);
      ccerr(0);
    }
    prv = sckt;
    sckt = sckt->nxt;
  }
  err:
  ReleaseSystemMutex(hLinkLstMutex);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot free TCP client socket: %.32s", cc2str(cc));
}
void rmvTcpClnSckList(void)
{
  while (tcpClnSckTbl != NULL) rmvTcpClnSck(tcpClnSckTbl->sck);
}
/** \internal TCP client for connection table entry */
int connectTCP(ConTblEntry *c,int port,int inetProtocol)
{ 
  int fid,cc=0,e;
  int optval = 1, mode = 1;
  int family = getInetFamily(inetProtocol);
  _socklen_t len = getInAddrLen(family);
  struct timeval tv0,tv;
  SCKADR srvadr;
  UINT16 p;

  if (c == NULL) return un_allocated;
  fid = c->fecIdx;
  p = (UINT16)(port + FecTbl[fid].portOffset);
  sckadr_in(&srvadr,(BYTE *)&FecTbl[fid].netAdr,p,family);
  if ((c->tcpSck=(int)socket(family,SOCK_STREAM,0))<0)
  {
    rprtSckErr("connectTCP","socket",0,-1,0,NULL);
    return tcp_socket_error;
  }
  if (setsockopt(c->tcpSck,SOL_SOCKET,SO_KEEPALIVE,(char *)&optval,sizeof(optval))) 
    rprtSckErr("connectTCP","SO_KEEPALIVE",c->tcpSck,-1,0,NULL);
  optval = max_clnrcvbuf;
  if (setsockopt(c->tcpSck,SOL_SOCKET,SO_RCVBUF,(char *)&optval,sizeof(optval))) 
    rprtSckErr("connectTCP","SO_RCVBUF",c->tcpSck,-1,0,NULL);
  checkSckBufferSpace(c->tcpSck,max_clnrcvbuf,SO_RCVBUF,"TCP client");
  optval = max_clnsndbuf;
  if (setsockopt(c->tcpSck,SOL_SOCKET,SO_SNDBUF,(char *)&optval,sizeof(optval))) 
    rprtSckErr("connectTCP","SO_SNDBUF",c->tcpSck,-1,0,NULL);
  checkSckBufferSpace(c->tcpSck,max_clnsndbuf,SO_SNDBUF,"TCP client");
  /* set to non-blocking for the connection attempt : */
  if (soioctl(c->tcpSck,FIONBIO,(char *)&mode)) ccerr(tcp_connect_error);
  gettimeofday(&tv0,(struct timezone *)NULL);  tv = tv0;
  while (connect(c->tcpSck,(struct sockaddr *)&srvadr,len) < 0)
  {
    if ((e=soerrno) == EISCONN) break;
#   ifdef WIN32
    if (e == WSAEISCONN) break; /* Microsoft keeps redefining things from version to version */
#   endif
    millisleep(gSystemTick);
    gettimeofday(&tv,(struct timezone *)NULL);
    if (MSECS(tv,tv0) > DEFAULT_TCP_CONNECTION_TIMEOUT) ccerr(connection_timeout);
  }
  getTcpClnSck(c->tcpSck,fid,inetProtocol);
err:
  if (cc) 
  {
    soclose(c->tcpSck);
    c->tcpSck = 0;
  }
  return cc;
}
int addClientSocketsToSet(struct pollfd *fds,int offset,int maxlen)
{
  int i=0, n=offset;

  memset(&fds[offset],0,sizeof(struct pollfd)*(maxlen-offset));
  ADDSCKTOSET(udpClnSck);
# ifndef MULTITHREADED
  { /* not multi-threaded : add all descriptors to same set */
    if (gListenForMCasts) 
    { /* client is listening for server multicasts */
      ADDSCKTOSET(mcastClnSck);
    }
    ADDSCKTOSET(glbClnSck);
    if (WaitForSystemMutex(hLinkLstMutex, gSystemTick) == 0)
    {
      TCPSCKT *sckt;
      for (sckt=tcpClnSckTbl; sckt != NULL; sckt = sckt->nxt)
        ADDSCKTOSET(sckt->sck);
      ReleaseSystemMutex(hLinkLstMutex);
    }
  }
# else
  /* is multi-threaded : each tcp socket has its own thread */
  UNUSED(i); /* remove at least one warning in case of NO PIPEs */
# endif
# ifdef UNIX
# ifndef NO_SERVICE_PIPE
  if (!gIsRunningAsServer && ipcCmdSck > 0)
  { /* possible client application remote pipe ? */
    ADDSCKTOSET(ipcCmdSck);
    for (i=0; i<ipcClnLstSize; i++) ADDSCKTOSET(IPCfds[i]);
  }
# endif /* NO_SERVICE_PIPE */
# ifndef NO_TINE_PIPE
  for (i=0; !gUseLoopback && i<nClnPipeSck; i++)
  {
    ADDSCKTOSET(gClnPipeSckTbl[i].sck);
  }
# endif /* NO_TINE_PIPE */
  if (!gIsRunningAsServer && foregroundTTY) 
  { 
    if (n >= maxlen) return n; 
    fds[n].fd = STDIN_FILENO; fds[n].events = POLLIN; n++; 
  }
# endif /* UNIX */
# ifdef IPXBSD
  ADDSCKTOSET(ipxClnSck);
  if (gListenForMCasts) ADDSCKTOSET(ipxMcastClnSck);
# endif
  return n;
}
int processTCPDataPacket(SOCKET sck,BYTE *bf,int nread,SCKADR sckadr,int inetprot)
{
  int n = 0, cc = 0, nleft = 0, hprot, hsize;
  TCPBCKT *bckt=findBucket(sck,TRUE);
  if (bckt == NULL) return non_existent;
  hprot = bckt->tineprot;
  hsize = RSPHDR_SIZE; /* worst-case header size */
  nleft = nread; /* the number of bytes in the payload buffer 'bf' */
  while (nleft > 0) 
  { /* examine what just came in ... */
    if (bckt->ptr < hsize) 
    { /* collecting a new header */   
      RspHdr rsphdr;
      int bckptr = bckt->ptr;
      n = nleft > bckt->len - bckt->ptr ? hsize - bckptr : nleft;
      memcpy(&bckt->buf[bckt->ptr], bf, n);
      bckt->ptr += n;
      if (bckt->ptr < hsize)
      { /* still not enough bytes in to form a header */
        nleft = 0; 
        break;
      }
      if (prepIncomingResponseHdr(&rsphdr, NULL, bckt->buf, 0) != hprot) ccerr(illegal_protocol);
      bckt->siz = rsphdr.totalSizeInBytes; /* now we know what size a chunk is ... */
      bckt->ptr = bckptr; /* rewind the stream for processing below ... */
    }
    n = bckt->siz - bckt->ptr;
    if (n <= 0 || n > TCP_BUFFER_SIZE) ccerr(dimension_error); /* corrupt ! */
    if (nleft >= n) 
    {
      memcpy(&bckt->buf[bckt->ptr],bf,n);
      doProcessData(0,bckt->buf,&sckadr,inetprot,NULL);
      bf += n;
      nleft -= n;
      bckt->ptr = 0;
    } 
    else 
    {
      memcpy(&bckt->buf[bckt->ptr],bf,nleft);
      bckt->ptr += nleft;
      nleft = 0;
    }
  }
err:
  if (cc && (tineDebug > 1 || errorDebug)) 
    dbglog("tcp data packet : %s : process %d bytes (%d - %d) : abort",
      cc2str(cc),n,bckt->siz,bckt->ptr);
  ReleaseSystemMutex(hTcpBucketMutex);
  return nread-nleft;
}
#ifdef MULTITHREADED
int reallocBucket(TCPBCKT *bckt)
{ /* increase the bucket buffer size ... */
  int buflen = bckt->siz + TCP_BUFFER_SIZE;
  BYTE *tmpbuf = (BYTE *)realloc(bckt->buf, buflen);
  if (tineDebug || errorDebug)
    dbglog("incoming stream size larger than allocated buffer : %d vs %d", bckt->siz, bckt->len);
  if (tmpbuf == NULL) return out_of_local_memory;
  bckt->buf = tmpbuf;
  return 0;
}
/** 
 * \internal process a CM_STREAM type TCP data stream
 * TODO: this can be consolidated with processTCPDataPacket() above
 * but: careful about passing a '0' in doProcessData() for the data length 
 * when ordinary TCP and careful about when to release hTcpBucketMutex.
 * This can later be simplified once all pre Release 5.xx are no longer active ...
 */
int processTCPDataStream(TCPBCKT **pbckt,int nread)
{
  int n = 0, nleft, hprot, hsize;
  SOCKET sck; 
  BYTE *bf;
  TCPBCKT *bckt=*pbckt;
  SCKADR sckadr;

  if (bckt == NULL) return 0;
  hprot = bckt->tineprot;
  hsize = RSPHDR_SIZE; /* worst-case header size */
  sck = bckt->sck; bf = bckt->clrbuf; 
  if (tineDebug > 2) dbglog("STREAM: received %d bytes on socket %d",nread,sck);
  msglog(0, "Receiving %d STREAM bytes from %.16s", nread, FecTbl[bckt->idx].fecName);
  nleft = nread;
  while (nleft > 0)
  {
    if (bckt->ptr < hsize)
    { /* collecting a new header */
      RspHdr rsphdr;
      int bktptr = bckt->ptr;
      n = nleft > bckt->len - bckt->ptr ? hsize - bktptr : nleft;
      memcpy(&bckt->buf[bckt->ptr], bf, n);
      bckt->ptr += n;
      if (bckt->ptr < hsize)
      { /* still not enough bytes in to form a header */
        nleft = 0;
        break;
      }
      if (prepIncomingResponseHdr(&rsphdr, NULL, bckt->buf, 1) != hprot)
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "protocol mis-match %d vs %d", rsphdr.tineProtocol, hprot);
        bckt->rmvNeeded = TRUE;
        return 0;
      }
      bckt->siz = rsphdr.totalSizeInBytes; /* now we know what size a chunk is ... */
      if (bckt->siz > bckt->len)
      { /* bucket buffer not large enough to hold what's coming ... */
        if (reallocBucket(bckt))
        {
          feclogEx(TINE_LOGLEVEL_ERROR, "cannot reallocate TCP bucket size to %d bytes", bckt->siz);
          bckt->rmvNeeded = TRUE;
          return 0;
        }
      }
      bckt->ptr = bktptr; /* rewind the stream for processing below ... */
    }
    n = bckt->siz - bckt->ptr;
    if (n <= 0) return 0; /* corrupt ! */
    if (nleft >= n) 
    {
      memcpy(&bckt->buf[bckt->ptr],bf,n);
      doProcessData(bckt->siz,bckt->buf,&sckadr,bckt->inetprot,NULL);
      bf += n;
      nleft -= n;
      bckt->ptr = 0;
    } 
    else 
    {
      memcpy(&bckt->buf[bckt->ptr],bf,nleft);
      bckt->ptr += nleft;
      nleft = 0;
    }
  }
  return nread-nleft;
}
#endif /* MULTITHREADED */
void recvIncomingIPDataErr(SOCKET sck,int inetprot,int nread)
{
  rprtSckErr("recvIncomingIPDataErr","read",sck,-1,0,NULL);
  if (nread == 0) feclog("%s read 0 bytes",getProtocolAsString(inetprot));
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (inetprot == PIPE)
  {
    msglog(pipe_error,"reset pipe due to error");
    ClosePipeSocket(sck);
  }
  else
# endif
  {
    soclose(sck); freeBucket(sck,FALSE);
  }
  feclogEx(TINE_LOGLEVEL_ERROR, "%s error",getProtocolAsString(inetprot));
}
int isBadSocket(SOCKET sck)
{
  struct pollfd fds[1];
  int e;
  if (sck == 0) return TRUE;

  if (poll(fds,1,0) < 0 || pollerr(0))
  {
    if ((e=soerrno) != EWOULDBLOCK) return TRUE;
  }
  return FALSE;
}
#ifdef WIN32
void trapBadSockets(struct pollfd *fds,int nfds)
{
  int i;
  for (i=0; i<nfds; i++)
  {
    if (isBadSocket(fds[i].fd))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "socket %d is bad",fds[i].fd);
    }
  }
  if (isBadSocket(tcpSrvSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "tcp server socket is bad"); tcpSrvSck = 0; }
# ifdef MULTITHREADED
  if (isBadSocket(tcpStrmSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "stream server socket is bad"); tcpStrmSck = 0; }
# endif
  if (gAllowRemoteDebugging)
  {
    if (isBadSocket(dbgSrvSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "debug server socket is bad"); dbgSrvSck = 0; }
  }
  if (isBadSocket(udpSrvSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "udp server socket is bad"); udpSrvSck = 0; }
  if (isBadSocket(udpClnSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "udp client socket is bad"); udpClnSck = 0; }
  if (isBadSocket(netSrvSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "net server socket is bad"); netSrvSck = 0; }
  if (isBadSocket(udpSyncSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "sync server socket is bad"); udpSyncSck = 0; }
  if (isBadSocket(glbClnSck)) { feclogEx(TINE_LOGLEVEL_ERROR, "globals server socket is bad"); glbClnSck = 0; }
}
#endif /* WIN32 */
int fdsContains(struct pollfd *fds, int nfds, SOCKET s)
{
  int i;
  if (s == (SOCKET)0) return TRUE; /* not allowed, but just play along */
  for (i = 0; i < nfds; i++) if (fds[i].fd == (int)s) return TRUE;
  return FALSE;
}
int getFdsIdx(struct pollfd *fds,int nfds,SOCKET s)
{
  int i;
  for (i=0; i<nfds && fds[i].fd != (int)s; i++);
  return (i == nfds ? -1 : i);
}
struct pollfd *getFds(struct pollfd *fds,int nfds,SOCKET s)
{
  int i = getFdsIdx(fds,nfds,s);
  if (s == (SOCKET)0) return NULL;
  if (i < 0) return NULL;
  return &fds[i];
}
void recvIpData(struct pollfd *fds,SOCKET *psck,int inetprot)
{
  SOCKET sck = *psck;
  SCKADR sckadr;
  struct sockaddr *sa = (struct sockaddr *)&sckadr;
  int n = 0, nread, family = getInetFamily(inetprot);
  _socklen_t len = getInAddrLen(family);
  char *ptag = NULL;

  if (fds == NULL || (fds->revents & POLLIN) == 0) return;
  if ((nread=readfrom(sck,(char *)clnWorkArea,MIN_WORK_SIZE,0,sa,&len)) > 0)
  { /* read something ... */
    switch (inetprot)
    {
      case PIPE:
        strncpy(gFecAddrBuffer,"local pipe",16);
        memset(&sckadr,0,sizeof(sckadr));
      case TCP:
      case STREAM:
      case TCP6:
      case STREAM6:
        if (inetprot != PIPE)
        { /* falling thru from above */
          if (getpeername(sck,sa,&len) < 0) goto err;
          inetStrAdr(&sckadr,gFecAddrBuffer,FALSE);
        }
        n = processTCPDataPacket(sck,clnWorkArea,nread,sckadr,inetprot);
        break;
      case UDP:
      case UDP6:
      {
        if (sck == mcastClnSck) ptag = "MCAST";
        inetStrAdr(&sckadr, gFecAddrBuffer, FALSE);
        n = (int)doProcessData(0, clnWorkArea, &sckadr, inetprot, ptag);
        break;
      }
    }
  } 
  else 
  { /* select/poll dropped thru and either error or no bytes read when there should have been */
err:if (soerrno != EWOULDBLOCK) 
    { /* serious error*/
      if (inetprot & (PIPE|TCP|STREAM|TCP6|STREAM6)) 
      { /* if a connected socket then shut it down */
        recvIncomingIPDataErr(sck,inetprot,nread);
        *psck = 0;
      }
    }
  }
  if (tineDebug > 3) 
  {
    if (ptag == NULL) ptag = getProtocolAsString(inetprot);
    if (nread <= 0) dbglog("%s: read error",ptag);
    else dbglog("%s: %d bytes read from %d",ptag,nread,n);
  }
}
void doIPConsumer(void)
{
  int i=0,idx=0,k,p;
  int retry=2,nfds=0,to=0,eno=0,nready;
  struct pollfd fds[256];
  SOCKET sck = 0;
# if defined(UNIX)
  char cmdbuf[CMDSIZE];
# endif
# if !defined(MULTITHREADED)
  SCKADR sckadr;
  int nread;
# endif
  static int errlogged=FALSE;
  
  if (!ipLoaded) return;
# ifdef WIN32
  if (!gIsRunningAsServer) chkWinPipe();
# endif
  if (!gRunServerCycleInSeparateThread) nfds = addServerSocketsToSet(fds,nfds,256);
  nfds = addClientSocketsToSet(fds,nfds,256);
  if (nfds == 0)
  { /* no sockets added ! (still initializing?) */
    millisleep(10);
    return;
  }
  if (!ReturnToClientCycle) 
  {
    to = clientTimeout*1000 + clientTimeoutMsec;
  }
  if (nTCPsck) removeFlaggedBuckets(); 
  gWaitingOnSelect = TRUE;
retry_select:
  for (k=0; (nready=poll(fds,nfds,to)) > 0 && k<gClientBurstLimit; k++)
  {
    gWaitingOnSelect = FALSE; to = 0;
    /* check the bread-and-butter UDP socket */
    p = (gAfInet == AF_INET6) ? UDP6 : UDP;
    recvIpData(getFds(fds,nfds,udpClnSck),&udpClnSck,p);
#   ifndef MULTITHREADED
    /* the next few sockets have their own threads in a multi-threaded build */
    if (WaitForSystemMutex(hLinkLstMutex, gSystemTick) == 0)
    {
      TCPSCKT *sckt;
      p = (gAfInet == AF_INET6) ? TCP6 : TCP;
      for (sckt=tcpClnSckTbl; sckt != NULL; sckt = sckt->nxt)
        recvIpData(getFds(fds,nfds,sckt->sck),&sckt->sck,p);
      ReleaseSystemMutex(hLinkLstMutex);
    }
    if (gListenForMCasts) 
    { /* something on the multicast socket ? */
      p = (gAfInet == AF_INET6) ? UDP6 : UDP;
      if (mcastClnSck != 0) recvIpData(getFds(fds,nfds,mcastClnSck),&mcastClnSck,p);
    }
    if (glbClnSck > 0 && fds_isset(glbClnSck))
    { /* something on the globals socket ? */
      _socklen_t len = sizeof(sckadr);
      if ((nread=recvfrom(glbClnSck,(char *)clnWorkArea,GCAST_BUFFER_SIZE,0,(struct sockaddr *)&sckadr,&len)) > 0)
      {
        getGlobalUpdate(&sckadr,clnWorkArea);
      }
    }
#   else
    UNUSED(idx); /* avoid a warning */
#   endif /* MULTITHREADED */
#   ifdef UNIX
    if (foregroundTTY && fds_isset(STDIN_FILENO) && ipcCmdSck == 0 && GetCommandEx(STDIN_FILENO,cmdbuf,CMDSIZE) > 0)
    { /* check application command input */
      InterpretCommand(cmdbuf);
    }
#   ifndef NO_SERVICE_PIPE
    if (!gIsRunningAsServer && ipcCmdSck > 0)
    { /* attached process issuing a command ? */
      chkIpcClnLst(fds,nfds);
    }
#   endif /* NO_SERVICE_PIPE */
#   ifndef NO_TINE_PIPE
    p = PIPE;
    for (i=0; !gUseLoopback && i<nClnPipeSck; i++)
    { /* are there PIPE connections ? */
      sck = (SOCKET)gClnPipeSckTbl[i].sck;
      if (sck > 0) 
      { /* there is a pipe connection ! */
        recvIpData(getFds(fds,nfds,sck),&sck,p);
      }
    }
#   endif /* !NO_TINE_PIPE */
#   endif /* UNIX */
#   ifdef IPXBSD
    if (ipxClnSck && fds_isset(ipxClnSck)) break;
#   endif
    to = 0; /* finish checking all client activity */
    nfds = addClientSocketsToSet(fds,0,256);
  }
  gWaitingOnSelect = FALSE;
  if (k == gClientBurstLimit && tineDebug) 
  {
    dbglog("client burst limit reached");
  }
  if (nready == -1) eno = soerrno;
  for (i=0; i<nfds; i++) if (pollerr(i)) 
  { 
    sck = fds[i].fd; 
    eno = pollerr2errno(fds[i].revents);
    break;
  }
  if (eno != 0 && !ServerExitCondition)
  {
    if (eno != EWOULDBLOCK && eno != EINTR)
    {
      rprtSckErr("doIPConsumer","select",sck,eno,3,NULL);
      if (!errlogged)
      {
        errlogged = TRUE;
        feclogEx(TINE_LOGLEVEL_ERROR, "doIPConsumer : read error on socket <%d>",eno);
#       ifdef WIN32
        trapBadSockets(fds,nfds);
#       endif
      }
    }
    millisleep(1);
    if (eno != EWOULDBLOCK && retry-- > 0) goto retry_select;
  }
  ReturnToClientCycle = (k > 0) ? 1 : 0;
}
#define getFecNameFromTable(i) (((i)<0 || (i)>=numFecTblEntries) ? "???" : FecTbl[i].fecName)
IDLE_CONNECTION *RemoveIdleConnection(IDLE_CONNECTION *ic)
{
  IDLE_CONNECTION *nxt=NULL;
  if (ic == NULL) return ic;
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return ic;
  nxt = ic->nxt;
  if (ic->nxt != NULL) ic->nxt->prv = ic->prv;
  if (ic->prv != NULL) ic->prv->nxt = ic->nxt;
  else if (ic == tcpIdleConnectionList) tcpIdleConnectionList = ic->nxt;
  msglog(0,"%s connection to %s (sck %d) no longer idle",
    getProtocolAsString(ic->trp), getFecNameFromTable(ic->idx),ic->sck);
  SystemFree(ic);
  ReleaseSystemMutex(hLinkLstMutex);
  return nxt;
}
#define TCP_ALLOWED_IDLE_TIME 300000 /* in milliseconds */
static int gTcpAllowedIdleTime = TCP_ALLOWED_IDLE_TIME;
TINE_EXPORT void SetIdleConnectionDuration(int allowedIdleSeconds)
{
  if (allowedIdleSeconds < 1) allowedIdleSeconds = 1;
  if (allowedIdleSeconds > 86400) allowedIdleSeconds = 86400;
  gTcpAllowedIdleTime = allowedIdleSeconds * 1000; /* convert to milliseconds */
  feclog("set TCP idle connection duration to %d seconds",allowedIdleSeconds);
}
IDLE_CONNECTION *AddIdleConnection(SOCKET sck,int idx,int inetProtocol)
{
  IDLE_CONNECTION *ic=NULL;

  if ((ic=GetIdleConnection(idx,inetProtocol)) != NULL) return ic;
  if ((ic=(IDLE_CONNECTION *)SystemCalloc(1,sizeof(IDLE_CONNECTION))) == NULL) return NULL;
  ic->sck = sck;
  ic->idx = idx;
  ic->trp = inetProtocol;
  ic->ttl = gTcpAllowedIdleTime;
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return NULL;
  ic->nxt = tcpIdleConnectionList;
  if (tcpIdleConnectionList != NULL) tcpIdleConnectionList->prv = ic;
  tcpIdleConnectionList = ic;
  msglog(0,"mark %s connection to %.16s (sck %d) as idle",
    getProtocolAsString(inetProtocol), getFecNameFromTable(idx),ic->sck);
  ReleaseSystemMutex(hLinkLstMutex);
  return ic;
}
IDLE_CONNECTION *GetIdleConnection(int idx,int inetProtocol)
{
  IDLE_CONNECTION *ic=NULL;
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return NULL;
  for (ic=tcpIdleConnectionList; ic != NULL; ic=ic->nxt)
  {
    if (ic->idx == idx && ic->trp == inetProtocol) break;
  }
  ReleaseSystemMutex(hLinkLstMutex);
  return ic;
}
IDLE_CONNECTION *GetIdleConnectionBySocket(SOCKET sck,int inetProtocol)
{
  IDLE_CONNECTION *ic=NULL;
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return NULL;
  for (ic=tcpIdleConnectionList; ic != NULL; ic=ic->nxt)
  {
    if (ic->sck == (SOCKET)sck && ic->trp == inetProtocol) break;
  }
  ReleaseSystemMutex(hLinkLstMutex);
  return ic;
}
void rmvIdleConnections(void)
{
  IDLE_CONNECTION *ic=tcpIdleConnectionList;
  while (ic != NULL)
  {
    ic = RemoveIdleConnection(ic); 
  }
  return;
}
