/*
 * File clntools.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include <math.h>
#include "tine.h"
#include "repeater.h"
#ifdef WIN32
#include "process.h"
#endif

#define BLOB_SIZE (1024 * 1024)
//char blob[BLOB_SIZE];
BYTE *blob = NULL;
int blob_size = 0;
PrpQueryStruct prpinfo;
#define _bbval (BYTE *)blob
#define _bsval (short *)blob
#define _blval (UINT32 *)blob
#define _bfval (float *)blob
#define _bdval (double *)blob
#define _bcval (char *)blob
#ifdef SINT64
#define _bllval (SINT64 *)blob
#endif
#define _bn16val (NAME16 *)blob
#define _bn32val (NAME32 *)blob
#define _bn48val (NAME48 *)blob
#define _bn64val (NAME64 *)blob
#define _ustrval (DUSTRING *)blob

void dumpReturnedData(int fmt,int siz,char *tag);
void dumpReturnedSyncData(int length, char *tag);
int tgetFromRepeater(char *dev,char *prp,int fmt,int siz,char *tag,BYTE *data,int fmtIn,int sizIn,int acc);
void assertBlobSizeEx(int fmt, char *tag, int *callsize);

/* TODO: a struct is not considered here (the 1 MB allocation should work most of the time, but ... */
void assertBlobSize(int fmt, int *callsize)
{
  assertBlobSizeEx(fmt, NULL, callsize);
}
void assertBlobSizeEx(int fmt, char *tag, int *callsize)
{ 
  int num = *callsize, fmtsiz = fmtsizeof(fmt), hdrsiz = getFormatHeaderSize(fmt);
  int size = num * fmtsiz + hdrsiz;
  int ssize = GetStructSize(tag);
  if (ssize > 0)
  {
    size = num * ssize;
    fmtsiz = ssize;
  }
  if (size > blob_size)
  { /* need to do something */
    if (blob != NULL) 
    { /* need to reallocate */
      BYTE *ptr = (BYTE *)realloc(blob,size); 
      if (ptr != NULL) 
      { /* no problem */
        blob = ptr; 
        blob_size = size; 
      }
      else
      { /* not accepted */
        *callsize = (blob_size - hdrsiz)/fmtsiz;
      }
    }
    else
    { /* initial allocation */
      if (size < BLOB_SIZE) 
      { /* be generous */
        size = BLOB_SIZE;
      }
      blob = (BYTE *)malloc(size);
      if (blob == NULL)
      { /* serious problem ! */
        *callsize = 0;
      }
      else
      { /* all is well ... */
        blob_size = size;
      }
    }
  }
}
char acqStr[1024] = "";
int displayErrorMessages = TRUE;
int ReportError(int cc)
{
  char *msg;
  if (cc == 0 || (cc&CE_SENDDATA))
  {
    printf("\n");
    return 0;
  }
  if (!displayErrorMessages) return cc;
  if (strlen(acqStr) > 0)
  {
    printf("attempted %.1024s\n\n", acqStr);
  }
  msg = cc < numErr ? cc2longString(cc) : GetCompletionStatus(-1);
  printf("Error: %s\n", msg);
  return cc;
}

typedef struct 
{
  time_t startt;
  int waitt;
  int fmt;
  int errcnt;
  int sizOutputArray;
  int sizTargetArray;
  BYTE *dataOutput;
  BYTE *dataTarget;
} WaitTargetType;

WaitTargetType waitTgtTyp;

int waitTgtOutput = FALSE;

void cbWait4Target(int id,int cc)
{
  int i, fsiz, dsiz, ecode=0;
  time_t t = time(NULL);

  acqStr[0] = 0;
  if (waitTgtTyp.waitt > 0)
  { /* a maximum wait time was specified */
    if (t > waitTgtTyp.startt + waitTgtTyp.waitt)
    { /* and it's been exceeded ! */
      if (waitTgtOutput) printf("target not met within the given timeout\n");
      waitTgtTyp.errcnt = 5;
      ccerr(operation_timeout);
    }
  }
  if (cc) goto err;
  fsiz = fmtsizeof(waitTgtTyp.fmt);
  if (waitTgtTyp.sizOutputArray == waitTgtTyp.sizTargetArray)
  { /* do a compare over the entire set */
    dsiz = waitTgtTyp.sizOutputArray * fsiz;
    if (!memcmp(waitTgtTyp.dataOutput,waitTgtTyp.dataTarget,dsiz)) 
    { /* target met! finish up! */
      if (waitTgtOutput) printf("target met\n");
      ccerr(0);
    }
    return; /* keep on truckin' */
  }
  else
  {
    if (waitTgtTyp.sizOutputArray == 1 && waitTgtTyp.sizTargetArray > 1)
    { /* check against several targets ... */
      for (i=0; i<waitTgtTyp.sizTargetArray; i++)
      {
        if (!memcmp(waitTgtTyp.dataOutput,waitTgtTyp.dataTarget+i*fsiz,fsiz)) 
        { /* target met! finish up! */
          ecode = i;
          if (waitTgtOutput) printf("target %d met\n",i);
          ccerr(0);
        }
      }
      return; /* keep on truckin' */
    }
    else if (waitTgtTyp.sizTargetArray == 1 && waitTgtTyp.sizOutputArray > 1)
    { /* check all output values against the same target */
      for (i=0; i<waitTgtTyp.sizOutputArray; i++)
      {
        if (memcmp(waitTgtTyp.dataTarget,waitTgtTyp.dataOutput+i*fsiz,fsiz)) 
        { /* target not met! keep going! */
          return;
        }
      }
      if (waitTgtOutput) printf("all targets met\n");
      ccerr(0);
    }
    else
    { /* invalid comibination of client and target dimensions ! */
      printf("cannot monitor output array of size %d elements against %d targets\n",
        waitTgtTyp.sizOutputArray,waitTgtTyp.sizTargetArray);
      waitTgtTyp.errcnt = 5;
      ccerr(dimension_error);
    }
  }
err:
  /* negative exit code => stopped on an error */
  if (cc) 
  {
    if (waitTgtTyp.errcnt < 5) return; /* give it a bit of grace */
    if (waitTgtOutput) printf("stop monitoring due to error condition\n");
    ecode = -cc;
  }
  CloseLink(id);
  clientCycle();
  exit(ecode);
}
int _tmonitorWait(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  int id;
  DTYPE dout;

  dout.dFormat = fmt;
  dout.dArrayLength = len;
  dout.dTag[0] = 0;
  dout.data.vptr = buf;
  id = AttachLink(dev,prp,&dout,NULL,CA_READ,tmr,cbWait4Target,CM_POLL);
  return id < 0 ? -id : 0;
}
void tmonitorCallback(int id,int cc)
{
  int fmt;
  int siz;
  printf("\n");
  if (cc) 
  {
    char *src = GetCompletionSource(id) ? "REMOTE" : "LOCAL";
    printf("(%s ERROR: %.32s)\n>", src, GetCompletionStatus(id));
    return;
  }
  fmt = LFMT(prpinfo.prpFormat);
  siz = GetCompletionDataSizeFromCallbackId(id);
  if (siz > (int)prpinfo.prpSize) siz = (int)prpinfo.prpSize;
  dumpReturnedData(fmt,siz,prpinfo.prpTag);
}
int _tmonitor(BYTE *buf,char *dev,char *prp,int fmt,int len,char *tag,int tmr)
{
  int id;
  DTYPE dout;
  acqStr[0] = 0;
  if (fmt == CF_IMAGE)
  {
    DIMAGE *img;
    len -= (sizeof(DIMAGE));
    img = (DIMAGE *)buf;
    img->frameBufferSize = len;
    img->frameBuffer = (BYTE *)&buf[sizeof(DIMAGE)];
  }
  DTYPEZERO(dout);
  dout.dFormat = fmt;
  dout.dArrayLength = len;
  if (tag != NULL) strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  dout.data.vptr = buf;
  id = AttachLink(dev,prp,&dout,NULL,CA_READ,tmr,tmonitorCallback,CM_POLL);
  return id < 0 ? -id : 0;
}
int _tput(BYTE *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  DTYPE din;
  int cc;

  snprintf(acqStr, 1024, "tput: target server = \"%s\" property = \"%.64s\" (format: %s, size: %d)",
    dev == NULL ? "(null)" : dev,
    prp == NULL ? "(null)" : prp,
    GetFormatAsString(fmt), len);
  if (buf == NULL) { fmt = CF_NULL; len = 0; }
  cc = tgetFromRepeater(dev,prp,CF_NULL,0,NULL,(BYTE *)buf,fmt,len,CA_WRITE);
  if (cc == 0) return cc;

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = fmt;
  din.dArrayLength = len;
  din.data.vptr = buf;

  return ExecLinkEx(dev,prp,NULL,&din,CA_WRITE|CA_RETRY,(UINT16)tmr);
}
int _tget(BYTE *buf,char *dev,char *prp,int fmt,int len,char *tag,int tmr)
{
  DTYPE dout;

  if (fmt == CF_IMAGE)
  {
    DIMAGE *img;
    len -= sizeof(DIMAGE);
    img = (DIMAGE *)buf;
    img->frameBufferSize = len;
    img->frameBuffer = (BYTE *)&buf[sizeof(DIMAGE)];
  }
  DTYPEZERO(dout);
  dout.dFormat = fmt;
  dout.dArrayLength = len;
  if (tag != NULL) strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  dout.data.vptr = buf;
  return ExecLinkEx(dev,prp,&dout,NULL,CA_READ|CA_RETRY,(UINT16)tmr);
}
int _tputgetEx(BYTE *buf,char *dev,char *prp,int fmtIn,int lenIn, int fmtOut, int lenOut,int tmr,int acc)
{
  DTYPE din, dout;
  int cc;

  snprintf(acqStr, 1024, "tputget: target server = \"%s\" property = \"%.64s\" (format: %s, size: %d)",
    dev == NULL ? "(null)" : dev,
    prp == NULL ? "(null)" : prp,
    GetFormatAsString(fmtIn), lenIn);
  assertBlobSize(fmtOut,&lenOut);
  if (buf == NULL) { fmtIn = CF_NULL; lenIn = 0; }
  cc = tgetFromRepeater(dev,prp,fmtOut,lenOut,NULL,(BYTE *)buf,fmtIn,lenIn,acc);
  if (cc == 0) return cc;

  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  din.dFormat = fmtIn;
  din.dArrayLength = lenIn;
  din.data.vptr = buf;
  dout.dFormat = fmtOut;
  dout.dArrayLength = lenOut;
  dout.data.vptr = blob;

  return ExecLinkEx(dev,prp,&dout,&din,(short)(acc|CA_RETRY),(UINT16)tmr);
}
int _tputget(BYTE *buf,char *dev,char *prp,int fmt,int len,int tmr,int acc)
{
  return _tputgetEx(buf,dev,prp,fmt,len,fmt,len,tmr,acc);
}
int _tsendrecv(BYTE *buf,char *dev,char *prp,int fmtIn,int lenIn,int tmr,int acc)
{
  DTYPE din, dout;
  int cc, fmtOut = CF_DEFAULT, lenOut = (blob_size > BLOB_SIZE) ? blob_size : BLOB_SIZE;

  snprintf(acqStr, 1024, "tsendrecv: target server = \"%s\" property = \"%.64s\" (format: %s, size: %d)",
    dev == NULL ? "(null)" : dev,
    prp == NULL ? "(null)" : prp,
    GetFormatAsString(fmtIn), lenIn);
  assertBlobSize(fmtOut,&lenOut);
  cc = tgetFromRepeater(dev,prp,fmtOut,lenOut,NULL,(BYTE *)buf,fmtIn,lenIn,acc);
  if (cc == 0) return cc;

  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  din.dFormat = fmtIn;
  din.dArrayLength = lenIn;
  din.data.vptr = buf;
  dout.dFormat = fmtOut;
  dout.dArrayLength = (UINT32)lenOut;
  dout.data.vptr = blob;

  cc = ExecLinkEx(dev,prp,&dout,&din,(short)(acc|CA_RETRY|CA_MUTABLE),(UINT16)tmr);
  if (cc == 0)
  {
    dumpReturnedSyncData((int)dout.dArrayLength,NULL);
  }
  return cc;
}
int GetPropertyInformation(char *srv,char *prp,PrpQueryStruct *srvProps)
{
  DTYPE dout,din;
  int cc = 0;
  DTYPEZERO(dout); DTYPEZERO(din);
  din.dFormat = CF_NAME32;
  din.dArrayLength = 1;
  din.data.vptr = prp;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PRPQUERYSTRUCT_SIZE;
  strcpy(dout.dTag,"PRPQSr4");
  dout.data.vptr = srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,&din,CA_READ|CA_RETRY,500)) != 0) return cc;
  ISWAP(srvProps->prpSize);
  return 0;
}
char *fmtstr(int fmt)
{
  static char s[80];
  switch (LFMT(fmt))
  {
    case CF_BYTE: sprintf(s,"BYTE"); break;
    case CF_SHORT: sprintf(s,"short"); break;
    case CF_LONG: sprintf(s,"UINT32"); break;
    case CF_FLOAT: sprintf(s,"float"); break;
    case CF_DOUBLE: sprintf(s,"double"); break;
    case CF_TEXT: sprintf(s,"char"); break;
    case CF_NAME16: sprintf(s,"NAME16"); break;
    case CF_NAME32: sprintf(s,"NAME32"); break;
    default: sprintf(s,"type <%d>",LFMT(fmt)); break;
  }
  return s;
}
int getCdiPropertyInformation(char *dev,char *prp,PrpQueryStruct *pqs)
{
  int cc, fmt, siz;
  float range[2];
  char ver[64];
  NAME64 dsc;
  NAME16 units;
  DTYPE dout;
  if (strnicmp(prp,"RECV",4) && strnicmp(prp,"SEND",4)) return not_applicable;
  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = 32;
  dout.dFormat = CF_TEXT;
  dout.data.cptr = ver;
  if ((cc=ExecLinkEx(dev,"CDIVERSION",&dout,NULL,CA_READ,500)) != 0) return cc;
  memset(pqs,0,sizeof(PrpQueryStruct));
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &fmt;
  if ((cc=ExecLinkEx(dev,"DISPLAYFORMAT",&dout,NULL,CA_READ,500)) != 0) return cc;
  dout.data.lptr = &siz;
  if ((cc=ExecLinkEx(dev,"DISPLAYSIZE",&dout,NULL,CA_READ,500)) != 0) return cc;
  dout.dArrayLength = 1;
  dout.dFormat = CF_NAME64;
  dout.data.vptr = &dsc;
  if ((cc=ExecLinkEx(dev,"DESCRIPTION",&dout,NULL,CA_READ,500)) != 0) return cc;
  dout.dArrayLength = 1;
  dout.dFormat = CF_NAME16;
  dout.data.vptr = &units;
  if ((cc=ExecLinkEx(dev,"UNITS",&dout,NULL,CA_READ,500)) != 0) return cc;
  dout.dArrayLength = 2;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = range;
  if ((cc=ExecLinkEx(dev,"RANGE",&dout,NULL,CA_READ,500)) != 0) return cc;
  strncpy(pqs->prpName,prp,PROPERTY_NAME_SIZE);
  strncpy(pqs->prpDescription,dsc.name,64);
  strncpy(pqs->prpUnits,units.name,16);
  pqs->prpSize = siz;
  pqs->prpFormat = BFMT(fmt);
  pqs->prpMinValue = range[0];
  pqs->prpMaxValue = range[1];
  return 0;
}
int tinfo(char *dev,char *prp)
{
  int cc, n=1;
  PrpQueryStruct pqs, *ppqs = &pqs;
 
  snprintf(acqStr, 1024, "tinfo: target = \"%s\", property = \"%s\"",
    dev == NULL ? "(null)" : dev,
    prp == NULL ? "(null)" : prp);
  if (getCdiPropertyInformation(dev,prp,&pqs) == 0) goto dumpInfo;
  if ((cc=GetSystemPropertyInformation(dev,prp,&ppqs,&n)) != 0) return cc;
dumpInfo:
  printf("\nDevice : %s\nProperty : %s\nDescription : %s\nSize : %d\nFormat : %s\nMaximum : %f\nMinimum : %f\nUnits : %s\n",
         dev,pqs.prpName,pqs.prpDescription,pqs.prpSize,
         fmtstr(pqs.prpFormat),pqs.prpMaxValue,pqs.prpMinValue,pqs.prpUnits);
  return 0;
}
typedef struct
{ // as defined in ens.c ...
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char fec[FEC_NAME_SIZE];
  char os[FEC_OS_SIZE];
  char sub[SUBSYSTEM_NAME_SIZE];
  char addr[ADDR_SIZE_IPV6];
  char ver[FEC_VERSION_SIZE];
  char resp[FEC_RESP_SIZE];
  int importance;
  int portoffset;
} SrvQueryInfo;
int registerSrvQueryInfoStruct(void)
{
  int cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, ctx),
    CONTEXT_NAME_SIZE, CF_TEXT, "context")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, srv),
    EXPORT_NAME_SIZE, CF_TEXT, "server")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, fec),
    FEC_NAME_SIZE, CF_TEXT, "fec")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, os),
    FEC_OS_SIZE, CF_TEXT, "os")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, sub),
    SUBSYSTEM_NAME_SIZE, CF_TEXT, "subsystem")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, addr),
    ADDR_SIZE_IPV6, CF_TEXT, "ip")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, ver),
    FEC_VERSION_SIZE, CF_TEXT, "version")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, resp),
    FEC_RESP_SIZE, CF_TEXT, "responsible")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, importance),
    1, CF_INT32, "importance")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, portoffset),
    1, CF_INT32, "port_offset")) != 0) return cc;

  if ((cc = sealTaggedStruct("SRVQINF", sizeof(SrvQueryInfo), 1000)) != 0) return cc;
  return 0;
}
char *impstr[] = { "-unassigned-", "Necessary", "Essential", "Critical" };
int tserversEx(char *cnt,char *search,int verbose)
{
  int i, k, cc, imp = 0, fsiz;
  int fmt = verbose ? CF_USTRING : CF_NAME32;
  int n = GetNumberOf("/SITE/ENS/#0", "PROPERTIES");
  NAME32 *srvs;
  DUSTRING *usrvs;
  SrvQueryInfo *sqi = NULL;
  BYTE ipbytes[4];
  char srv[128], ip[64], *psrvname, *tag = "SRVQINF";
  DTYPE dout,din,*dinptr=NULL;
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  snprintf(acqStr, 1024, "tservers: context = \"%s\", search string: %s",
    cnt == NULL ? "(null)" : cnt,
    search == NULL ? "(null)" : search);

  if (verbose)
  {
    registerSrvQueryInfoStruct();
    fmt = CF_STRUCT;
    fsiz = GetStructSize(tag);
  }
  else
  {
    tag = NULL;
    fsiz = fmtsizeof(fmt);
  }
  assertBlobSizeEx(fmt, tag, &n);
  srvs = _bn32val;
  usrvs = _ustrval;
  sqi = (SrvQueryInfo *)blob;
  strcpy(srv,"ENS");
  if (cnt != NULL && strlen(cnt) > 0)
  {
    strcat(srv,"/"); strncat(srv,cnt,CONTEXT_NAME_SIZE);
  }
  if (search != NULL)
  {
    din.dFormat = CF_TEXT;
    din.dArrayLength = (UINT32)strlen(search);
    din.data.cptr = search;
    dinptr = &din;
  }
retry_dout:
  DTYPEZERO(dout);
  dout.dFormat = fmt;
  dout.dArrayLength = n;
  dout.data.vptr = srvs;
  if (tag != NULL) strncpy(dout.dTag, tag, TAG_NAME_SIZE);
  memset(srvs,0,n*fsiz);
  if ((cc=ExecLinkEx(srv, "SERVERS", &dout, dinptr, CA_READ | CA_MUTABLE, 200)) != 0)
  {
    if (cc == illegal_format && fmt == CF_STRUCT)
    {
      tag = NULL;
      fmt = CF_USTRING;
      fsiz = fmtsizeof(fmt);
      goto retry_dout;
    }
    return cc;
  }
  n = (int)dout.dArrayLength;
  printf("Servers in context %s",cnt);
  if (search != NULL) printf(" (filter: %s)",search);
  printf(":\n\n");
  if (fmt == CF_USTRING) printf("%-32s %-16s %-16s %-5s %-8s %s\n","Server","FEC","IP Address","Port","OS","Importance");
  if (fmt == CF_STRUCT) printf("%-16s %-8s %-32s %-16s %-16s %-5s %-8s %-12s %-16s %s\n", "Context", "Subsys.", "Server", "FEC", "IP Address", "Port", "OS", "Version", "Responsible", "Importance");
  for (i=0,k=0; i<n; i++)
  {
    psrvname = fmt == CF_USTRING ? usrvs[i].str : fmt == CF_STRUCT ? sqi[i].srv : srvs[i].name;
    //psrvname = verbose ? usrvs[i].str : srvs[i].name;
    if (strlen(psrvname) == 0) continue;
    if (fmt == CF_USTRING)
    {
      memset(srv, 0, 64);
      strncpy(srv, psrvname, 16);
      memcpy(ipbytes, &usrvs[i].ival, 4);
      if (srv[15] == '&')
      {
        strncpy(&srv[15], &usrvs[i].str[48], 16);
      }
      inet_ntop(gAfInet, &usrvs[i].ival, ip, 40);
      imp = usrvs[i].tm;
      if (imp < 0) imp = 0; if (imp > 3) imp = 3;
      //sprintf(ip, "%d.%d.%d.%d", ipbytes[0], ipbytes[1], ipbytes[2], ipbytes[3]);
      printf("%-32.32s %-16.16s %-16.16s %-5d %-8.8s %s\n", srv, &usrvs[i].str[32], ip, (int)usrvs[i].f1val ,&usrvs[i].str[16],impstr[imp]);
    }
    else if (fmt == CF_STRUCT)
    {
      printf("%-16.16s %-8.8s %-32.32s %-16.16s %-16.16s %-5d %-8.8s %-12.12s %-16.32s %s\n",
        sqi[i].ctx, sqi[i].sub, sqi[i].srv, sqi[i].fec, sqi[i].addr, sqi[i].portoffset,
        sqi[i].os, sqi[i].ver, sqi[i].resp, impstr[sqi[i].importance]);
    }
    else
    {
      printf("%.32s\n", srvs[i].name);
    }
    k++;
  }
  printf("\n%d servers found\n",k);
  return 0;
}
int tservers(char *cnt)
{
  int i,cc,n = 4096;
  NAME32 *srvs;
  snprintf(acqStr, 1024, "tservers: context = \"%s\"",
    cnt == NULL ? "(null)" : cnt);
  assertBlobSize(CF_NAME32, &n);
  srvs = _bn32val;
  if ((cc=GetSystemServers(cnt,srvs,&n)) != 0) return cc;
  printf("Servers in context %s",cnt);
  printf(":\n\n");
  for (i=0; i<n; i++) 
  {
    printf("%s\n",srvs[i].name);
  }
  return 0;
}
int tprops(char *dev)
{
  int i,cc,n=GetNumberOf(dev,"PROPERTIES"),hasQryFcn=FALSE;
  NAME64 *prps;
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  snprintf(acqStr, 1024, "tprops: target = \"%s\"",
    dev == NULL ? "(null)" : dev);
  assertBlobSize(CF_NAME64, &n);
  prps = _bn64val;
  if ((cc=GetSystemProperties(dev,prps,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    cc &= 0x0fff;
  }
  if (cc) return cc;
  printf("%d properties found :\n",n);
  for (i=0; i<n; i++) 
  {
    if (strlen(prps[i].name) == 0) continue;
    printf("%s\n",prps[i].name);
  }
  return 0;
}
int tproplist(char *dev)
{
  int i,cc,n=GetNumberOf(dev,"PROPERTIES"),hasQryFcn=FALSE;
  NAME64 *prps;
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  snprintf(acqStr, 1024, "tproplist: target = \"%s\"",
    dev == NULL ? "(null)" : dev);
  assertBlobSize(CF_NAME64, &n);
  prps = _bn64val;
  if ((cc=GetSystemProperties(dev,prps,&n)) != 0)
  {
    if (cc & has_query_function) 
    {
      hasQryFcn = TRUE;
      printf("Note: server has device query precedence\n");
      printf("      specify the desired device as in 'tproplist /PETRA/Cms.MagnetPs/PCH_OL_7'\n");
    }
    if (cc & CE_SENDDATA) cc = 0;
    cc &= 0x0fff;
  }
  if (cc) return cc;
  printf("Property List for %s (%d properties)\n\n",dev,n);
  for (i=0; i<n; i++) 
  {
    if (strlen(prps[i].name) == 0) continue;
    printf("%s\n",prps[i].name);
  }
  return 0;
}
int tdevsEx(char *dev,char *prp)
{
  int i,cc,n=GetNumberOf(dev,"DEVICES"),hasQryFcn=FALSE;
  NAME64 *devs;
  char devname[DEVICE_NAME_SIZE];
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  snprintf(acqStr, 1024, "tdevs: target = \"%s\" property = \"%s\"",
    dev == NULL ? "(null)" : dev,
    prp == NULL ? "(null)" : prp);
  assertBlobSize(CF_NAME64, &n);
  devs = _bn64val;
  if ((cc=GetSystemDevices(dev,prp,devs,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    cc &= 0x0fff;
  }
  if (cc) return cc;
  if (n == 2 && !strcmp(devs[1].name,"XYZHASQUERYFCN"))
  { // ancient kludge! Still necessary?
    n = 1;
    if (!strcmp(devs[0].name,"#0")) strncpy(devs[0].name,prp,16);
  }
  printf("%d devices found :\n",n);
  for (i=0; i<n; i++) 
  {
    strncpy(devname,devs[i].name,64);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s\n",devname);
  }
  return 0;
}
int tdevs(char *dev)
{
  int i,cc,n=GetNumberOf(dev,"DEVICES"),hasQryFcn=FALSE;
  NAME64 *devs;
  char devname[DEVICE_NAME_SIZE], *pc=NULL, *c;
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  snprintf(acqStr, 1024, "tdevs: target = \"%s\"",
    dev == NULL ? "(null)" : dev);
  assertBlobSize(CF_NAME64, &n);
  devs = _bn64val;
  if ((pc=strchr(dev,'[')) != NULL) 
  {
    *pc++ = 0;
    if ((c=strchr(pc,']')) != NULL) *c = 0;
  }
  if ((cc=GetSystemDevices(dev,pc,devs,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    cc &= 0x0fff;
  }
  if (cc) return cc;
  if (hasQryFcn)
  {
    printf("Device List depends on targeted Property: use tdevsEx()\n");
    printf("      specify the desired property as in 'tdevs /PETRA/ARCHIVER[Orbit.X]'\n");
    return ambiguous;
  }
  printf("%d devices found :\n",n);
  for (i=0; i<n; i++) 
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s\n",devname);
  }
  return 0;
}
int tdevlist(char *dev)
{
  int i,cc,n=GetNumberOf(dev,"DEVICES"),hasQryFcn=FALSE;
  NAME64 *devs;
  char devname[DEVICE_NAME_SIZE], *pc=NULL, *c;
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  snprintf(acqStr, 1024, "tdevlist: target = \"%s\"",
    dev == NULL ? "(null)" : dev);
  assertBlobSize(CF_NAME64, &n);
  devs = _bn64val;
  if ((pc=strchr(dev,'[')) != NULL) 
  {
    *pc++ = 0;
    if ((c=strchr(pc,']')) != NULL) *c = 0;
  }
  if ((cc=GetSystemDevices(dev,pc,devs,&n)) != 0)
  {
    if (cc & has_query_function) 
    {
      hasQryFcn = TRUE;
      printf("Note: server has property query precedence\n");
      printf("      specify the desired property as in 'tdevlist /PETRA/ARCHIVER[Orbit.X]'\n");
    }
    if (cc & CE_SENDDATA) cc = 0;
    cc &= 0x0fff;
  }
  if (cc) return cc;
  printf("DEVICE_NAME,DEVICE_NUMBER\n");
  for (i=0; i<n; i++)
  {
    strncpy(devname,devs[i].name,64);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s,%d\n",devname,i);
  }
  return 0;
}
int tsendstring(char *dev,char *prp,char *str)
{
  if (str == NULL || strlen(str) == 0) return invalid_data;
  return _tput((BYTE *)str,dev,prp,CF_TEXT,(int)strlen(str),1000);
}
BYTE *parseInputDataString(char *str,int *fmt,int *siz)
{
  BYTE *data=NULL;
  char *s, *c, *tok = " ,;\t\n";
  int i, n, len, base = 10;
  if (fmt == NULL || siz == NULL) return NULL;
  if (str == NULL) 
  {
    *fmt = CF_NULL;
    *siz = 0;
    return NULL;
  }
  if (str[0] == '\"')
  {
    if ((c=strchr(&str[1],'\"')) == NULL) return data; /* badly formed */
    *c = 0;
    data = (BYTE *)&str[1];
    *siz = (int)strlen((char *)data);
    if (!isSimpleStringFormat((short)(*fmt))) *fmt = CF_TEXT;
  }
  else
  {
    n = 0;
    len = (int)strlen(str);
    data = (BYTE *)calloc(len,64); /* worst case ? */
    if (!isSimpleFormat(*fmt) && *fmt != CF_NULL)
    {
      char segstr[256];
      int nsegs = GetNumberSegments(*fmt);
      i = 0; segstr[0] = 0;
      for (s=strtok(str, tok); s != NULL; s=strtok(NULL, tok))
      {
        strncat(segstr, s, 64);
        if (++i < nsegs)
        {
          strcat(segstr, " ");
          continue;
        }
        stringToDataValue(data, segstr, n++, *fmt, NULL, 0,FALSE);
        i = 0; segstr[0] = 0;
      }
      *siz = n;
      return data;
    }
    for (s=strtok(str,tok); s != NULL; s=strtok(NULL,tok))
    {
      if (*fmt == CF_NULL)
      {
        len = (int)strlen(s);
        for (i=0; i<len; i++) if (!isdigit(s[i])) break;
        if (i == len) /* all digits ! */
        {
          *fmt = CF_INT32;
          base = 10;
        }
        else if (strncmp(s,"0x",2) == 0 || strncmp(s,"0X",2) == 0)
        {
          *fmt = CF_INT32;
          base = 16;
        }
        else if (s[i] == '.')
        {
          *fmt = CF_DOUBLE;
        }
        else
        {
          if (len < 16) *fmt = CF_NAME16;
          else if (len < 32) *fmt = CF_NAME32;
          else if (len < 64) *fmt = CF_NAME64;
          else *fmt = CF_TEXT;
        }
      }
      switch (*fmt)
      {
        case CF_BYTE:
          ((BYTE *)data)[n] = (BYTE)strtol(s,NULL,base); break;
        case CF_INT16:
          ((UINT16 *)data)[n] = (UINT16)strtol(s,NULL,base); break;
        case CF_INT32:
          ((UINT32 *)data)[n] = strtol(s,NULL,base); break;
        case CF_FLOAT:
          ((float *)data)[n] = (float)strtod(s,NULL); break;
        case CF_DOUBLE:
          ((double *)data)[n] = strtod(s,NULL); break;
        case CF_NAME16:
          strncpy(((NAME16 *)data)[n].name,s,16); break;
        case CF_NAME32:
          strncpy(((NAME32 *)data)[n].name,s,32); break;
        case CF_NAME64:
          strncpy(((NAME64 *)data)[n].name,s,64); break;
        case CF_TEXT:
          strncpy((char *)data,s,len);
          *siz = len;
          return data;
        default:
          break;
      }
      n++;
      if (*siz > 0 && n == *siz) break;
    }
    *siz = n;
  }
  return data;
}
int tput(char *dev,char *prp,char *datastr,int fmt,int siz)
{
  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmt,&siz);
  return _tput(data,dev,prp,fmt,siz,1000);
}
int tputEx(char *dev,char *prp,char *datastr,int fmt,int siz,int tmr)
{
  BYTE *data = NULL;
  if (tmr < 50) tmr = 50;
  data = parseInputDataString(datastr,&fmt,&siz);
  return _tput(data,dev,prp,fmt,siz,tmr);
}
int tputget(char *dev,char *prp,char *datastr,int fmt,int siz,int tmr,int acc)
{
  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmt,&siz);
  return _tputget(data,dev,prp,fmt,siz,tmr,acc);
}
int tputgetEx(char *dev,char *prp,char *datastr,int fmtIn,int sizIn,int fmtOut,int sizOut,int tmr,int acc)
{
  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmtIn,&sizIn);
  return _tputgetEx(data,dev,prp,fmtIn,sizIn,fmtOut,sizOut,tmr,acc);
}
int tsendrecv(char *dev,char *prp,char *datastr,int fmt,int siz,int tmr,int acc)
{
  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmt,&siz);
  return _tsendrecv(data,dev,prp,fmt,siz,tmr,acc);
}
void dumpReturnedSyncData(int length, char *tag)
{
  int fmt,siz;
  fmt = GetCompletionDataType(-1);
  siz = GetCompletionDataSize(-1);
  if (siz > length) siz = length;
  dumpReturnedData(fmt,siz,tag);
}
extern int _useCsvStringLiteral;
int useBracketsInOutput=TRUE;
char outputSeparator[16] = " ";
void dumpReturnedData(int fmt,int siz,char *tag)
{
  int i,cc;
  char s[1024], **strval;
  DIMAGE *img;
  float *fval;
  _useCsvStringLiteral = TRUE;
  switch (fmt)
  {
    case CF_NULL:
      return;
    case CF_STRUCT:
      if ((cc=GetStructureAsString(tag,blob,s,1024)) == 0) printf("%s",s);
      else printf("cannot convert structure %.16s to string : %.32s",tag,cc2str(cc));
      return;
    case CF_IMAGE:
      img = (DIMAGE *)blob;
      printf("frame source header:\n");
      printf("\tbase id: %d\n",img->sourceHeader.baseTag);
      printf("\tport id: %d\n",img->sourceHeader.cameraPortId);
      printf("\tversion: %d\n",img->sourceHeader.versionTag);
      printf("\ttotal len: %d\n",img->sourceHeader.totalLength);
      printf("\ttime (sec): %d\n",img->sourceHeader.timestampSeconds);
      printf("\ttime (usec): %d\n",img->sourceHeader.timestampMicroseconds);
      printf("\tport name: %.64s\n",img->sourceHeader.cameraPortName);
      printf("frame image header:\n");
      printf("\tsrc width: %d\n",img->frameHeader.sourceWidth);
      printf("\tsrc height: %d\n",img->frameHeader.sourceHeight);
      printf("\taoi width: %d\n",img->frameHeader.aoiWidth);
      printf("\taoi height: %d\n",img->frameHeader.aoiHeight);
      printf("\tx start: %d\n",img->frameHeader.xStart);
      printf("\ty start: %d\n",img->frameHeader.yStart);
      printf("\tbytes/pixel: %d\n",img->frameHeader.bytesPerPixel);
      printf("\teff bits/pixel: %d\n",img->frameHeader.effectiveBitsPerPixel);
      printf("\thorz binning: %d\n",img->frameHeader.horizontalBinning);
      printf("\tvert binning: %d\n",img->frameHeader.verticalBinning);
      printf("\tsrc fmt: %d\n",img->frameHeader.sourceFormat);
      printf("\timg fmt: %d\n",img->frameHeader.imageFormat);
      printf("\tframe no.: %d\n",img->frameHeader.frameNumber);
      printf("\tevent no.: %d\n",img->frameHeader.eventNumber);
      printf("\tx scale: %g\n",img->frameHeader.xScale);
      printf("\ty scale: %g\n",img->frameHeader.yScale);
      printf("\timg rot: %g\n",img->frameHeader.imageRotation);
      printf("\treserved: %g\n",img->frameHeader.fspare1);
      printf("\treserved: %g\n",img->frameHeader.fspare2);
      printf("\treserved: %g\n",img->frameHeader.fspare3);
      printf("\timg flags: %d\n",img->frameHeader.imageFlags);
      printf("\treserved: %d\n",img->frameHeader.ispare1);
      printf("\treserved: %d\n",img->frameHeader.ispare2);
      printf("\treserved: %d\n",img->frameHeader.ispare3);
      printf("\treserved: %d\n",img->frameHeader.appendedFrameSize);
      printf("frame buffer size : %d\n",img->frameBufferSize);
#     ifdef OS_64BIT
#     ifdef WIN32
      printf("frame buffer addr : %llu\n",(size_t)img->frameBuffer);
#     else
      printf("frame buffer addr : %lx\n", (size_t)img->frameBuffer);
#     endif
#     else
      printf("frame buffer addr : %x\n",(size_t)img->frameBuffer);
#     endif
      printf("initial 16 bytes of frame:\n");
      for (i=0; i<16 && i<(int)img->frameBufferSize; i++)
      {
        printf(" %2x",img->frameBuffer[i]);
      }
      printf("\n\n");
      break;
    case CF_SPECTRUM:
      dataValueToString(s,blob,0,(short)fmt,NULL,256, FALSE);
      printf("[%s]\n",s);
      fval = (float *)&blob[SPECTRUM_HDR_SIZE];
      for (i=0; i<siz; i++) printf("%g%.16s",fval[i],outputSeparator);
      break;
    case CF_KEYVALUE:
    case CF_STRING:
      strval = (char **)blob;
      for (i=0; i<siz; i++) 
      {
        printf("%s%.16s\n",strval[i],outputSeparator);
      }
      break;
    case CF_TEXT:
      printf("%s\n",(char *)blob);
      break;
    case CF_MDA: /* TODO: handle this properly ... */
    default:
      if (isSimpleFormat((short)fmt) || siz <2) useBracketsInOutput = FALSE;
      for (i=0; i<siz; i++) 
      {
        dataValueToString(s,(BYTE *)blob,i,(short)fmt,NULL,256, FALSE);
        if (useBracketsInOutput) printf("[%s]%.16s",s,outputSeparator);
        else printf("%s%.16s",s,outputSeparator);
      }
      break;
  }
}
int tgetFromRepeater(char *dev,char *prp,int fmt,int siz,char *tag,BYTE *data,int fmtIn,int sizIn,int acc)
{
  int p,cc;
  char rpdev[512],tgt[256],*ptag=tag;
  DTYPE dout,din;
  static int init_done = FALSE, ignoreRptr = FALSE;

  if (ignoreRptr) return not_implemented;
  if (!init_done)
  {
    int ts, tries = 0;
    char *ptr = getenv("TINE_USE_REPEATER");
    init_done = TRUE;
    if (ptr != NULL)
    {
      if (!stricmp(ptr,"FALSE") || !stricmp(ptr,"NO")) 
      {
        ignoreRptr = TRUE;
        return not_implemented;
      }
    }
    p = GetPortOffset(RPTR_LCLFEC);
    appendAddressToCache(RPTR_LCLFEC,gLoopbackAddr,(short)p);
    appendServerToCache(RPTR_LCLFEC,RPTR_LCLCTX,RPTR_LCLSRV,RPTR_EQM);
    sprintf(rpdev,"/%.32s/%.32s/target",RPTR_LCLCTX,RPTR_LCLSRV);
    dout.dFormat = CF_INT32;
    dout.dArrayLength = 1;
    dout.dTag[0] = 0;
    dout.data.lptr = &ts;
probeRepeater:
    cc = ExecLinkEx(rpdev,"SRVSTARTTIME",&dout,NULL,CA_READ|CA_RETRY,100);
    if (cc == connection_timeout || cc == link_timeout)
    {
      if (tries++ == 0)
      {
#       if defined(WIN32)
        flushall();
        spawnlp(_P_DETACH,"tineRepeater","tineRepeater","LOCALHOST",NULL);
#       elif defined(UNIX)
        if (vfork() == 0)
        {
          execlp("tineRepeater","tineRepeater","LOCALHOST",NULL);
          exit(0);
        }
#       endif
        goto probeRepeater;
      }
    }
  }
  if (fmt == CF_DEFAULT)
  { /* have to pre-determine it (CF_DEFAULT local repeater property CALL is CF_NULL) */
    PrpQueryStruct pqs;
    NAME64 pbuf;
    strncpy(pbuf.name,prp,64);
    dout.dFormat = CF_STRUCT;
    dout.dArrayLength = 1;
    strncpy(dout.dTag,"PRPQSr4",TAG_NAME_SIZE);
    dout.data.vptr = (void *)&pqs;
    din.dFormat = CF_NAME64;
    din.dArrayLength = 1;
    din.dTag[0] = 0;
    din.data.cptr = pbuf.name;
    sprintf(tgt,"<%.132s[PROPERTIES]",dev);
    sprintf(rpdev,"/%.32s/%.32s/%.196s",RPTR_LCLCTX,RPTR_LCLSRV,tgt);
    cc = ExecLinkEx(rpdev,"CALL",&dout,&din,CA_READ|CA_RETRY,500);
    if (cc != 0 && !(cc & CE_SENDDATA)) return cc;
    fmt = LFMT(pqs.prpFormat);
    siz = pqs.prpSize;
    ptag = pqs.prpTag;
  }

  sprintf(tgt,"<%.132s[%.64s]",dev,prp);
  sprintf(rpdev,"/%.32s/%.32s/%.196s",RPTR_LCLCTX,RPTR_LCLSRV,tgt);
  DTYPEZERO(din);
  if (data != NULL && fmtIn != CF_NULL && sizIn > 0)
  {
    din.dFormat = fmtIn;
    din.dArrayLength = sizIn;
    din.data.bptr = data;
  }

  //if (siz > sizeof(blob)) siz = sizeof(blob);
  DTYPEZERO(dout);
  if (siz > 0 && fmt != CF_NULL)
  {
    dout.dFormat = fmt;
    dout.dArrayLength = siz;
    if (ptag != NULL) strncpy(dout.dTag,ptag,TAG_NAME_SIZE);
    if (fmt == CF_IMAGE)
    {
      DIMAGE *img;
      siz += sizeof(DIMAGE);
      assertBlobSize(fmt,&siz);
      img = (DIMAGE *)blob;
      img->frameBufferSize = blob_size - (sizeof(DIMAGE));
      img->frameBuffer = (BYTE *)&blob[sizeof(DIMAGE)];
    }
    dout.data.vptr = blob;
  }
  if ((cc=ExecLink(rpdev,"CALL",&dout,&din,(short)acc)) != 0) 
  {
    while (cc == invalid_transport_size || cc == use_stream_transport || cc == work_area_full)
    {
      siz /= 2; 
      if (siz < 1000) return cc;
      dout.dArrayLength = siz;
      if ((cc=ExecLink(rpdev,"CALL",&dout,&din,(short)acc)) == 0) goto get_done;
    }
    return cc;
  }
get_done:
  dumpReturnedSyncData(siz,ptag);
  return 0;
}
int tgetFormatAndSize(char *dev,char *prp,int *fmt,int *siz,char *tag)
{
  int cc;
  if (dev == NULL || prp == NULL || fmt == NULL || siz == NULL) return argument_list_error;
  if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
  *fmt = LFMT(prpinfo.prpFormat);
  *siz = prpinfo.prpSize;
  if (tag != NULL) strncpy(tag,prpinfo.prpTag,TAG_NAME_SIZE);
  return 0;
}
int tgetEx(char *dev,char *prp,int fmt,int siz,char *tag,int tmr)
{
  int cc,retry_cnt=0;
  int fsiz;
  char *ptag=tag,tagb[TAG_NAME_SIZE];

  snprintf(acqStr, 1024, "tget: target server = \"%s\" property = \"%.64s\" (format: %s, size: %d)",
    dev == NULL ? "(null)" : dev,
    prp == NULL ? "(null)" : prp,
    GetFormatAsString(fmt), siz);
  if (dev == NULL || prp == NULL) return argument_list_error;

  if (siz == 0 && fmt == CF_DEFAULT) siz = BLOB_SIZE;
  assertBlobSize(fmt,&siz);
  if (siz == 0) return illegal_data_size;
  cc = tgetFromRepeater(dev,prp,fmt,siz,tag,NULL,CF_NULL,0,CA_READ);
  cc &= ~CE_REDIRECTED;
  if (cc == 0) return cc;
  if (fmt != CF_NULL)
  { /* normally the case (if the repeater didn't respond) */
    fsiz = fmtsizeof(fmt);
    if (siz > blob_size/fsiz) siz = blob_size/fsiz;
get_retry1:
    cc = _tget(blob,dev,prp,fmt,siz,ptag,tmr);
    if (cc == buffer_too_small)
    {
      if (retry_cnt++ == 0)
      {
        if (tgetFormatAndSize(dev,prp,&fmt,&siz,tagb) == 0)
        {
          if (strlen(tagb) > 0) acquireAndRegisterStruct(dev,tagb,siz);
          ptag = tagb;
          goto get_retry1;
        }
      }
    }
    if (cc & CE_SENDDATA) cc = 0;
    if (fmt == CF_DEFAULT && cc == invalid_structure_tag)
    {
      if (tgetFormatAndSize(dev,prp,&fmt,&siz,tagb) == 0 &&
          acquireAndRegisterStruct(dev,tagb,siz) == 0)
      {
        ptag = tagb;
        goto get_retry1;
      }
    }
    if (cc != 0) return cc;
  }
  else
  { /* fmt = NULL => called via tget() */
get_retry2:
    if ((cc=_tget(blob,dev,prp,CF_DEFAULT,siz,ptag,tmr)) != 0) 
    {
      if (cc & CE_SENDDATA) cc = 0;
      while (cc == invalid_transport_size || cc == use_stream_transport || cc == work_area_full)
      {
        siz /= 2; 
        if (siz < 1000) return cc;
        if ((cc=_tget(blob,dev,prp,CF_DEFAULT,siz,ptag,tmr)) == 0) goto get_done;
      }
      if ((cc == link_not_open || cc == connection_timeout) && retry_cnt++ < 2) goto get_retry2;
      return cc;
    }
  }
get_done:
  dumpReturnedSyncData(siz,ptag);
  return 0;
}
int tget(char *dev,char *prp)
{
  return tgetEx(dev,prp,CF_NULL,0,NULL,1000);
}
int tmonitorEx(char *dev,char *prp,int fmt_,int siz_,char *tag,int tmr)
{
  int cc;
  int fmt,fsiz,hsiz,siz;
  char *ptag = tag;
  acqStr[0] = 0;
  if (fmt_ != CF_NULL && siz_ != 0)
  {
    fmt = fmt_;
    siz = siz_;
    prpinfo.prpFormat = BFMT(fmt);
    prpinfo.prpSize = siz;
    if (tag != NULL) strncpy(prpinfo.prpTag,tag,TAG_NAME_SIZE);
  }
  else
  {
    if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
    fmt = LFMT(prpinfo.prpFormat);
    siz = prpinfo.prpSize;
    ptag = prpinfo.prpTag;
  }
  if (fmt == CF_STRUCT && ptag != NULL)
  {
    if ((cc=acquireAndRegisterStruct(dev,ptag,siz)) != 0) return cc;
  }
  fsiz = fmtsizeof(fmt); hsiz = getFormatHeaderSize(fmt);
  assertBlobSize(fmt,&siz);
  if (siz > (blob_size-hsiz)/fsiz) siz = (blob_size-hsiz)/fsiz;
  if ((cc=_tmonitor(blob,dev,prp,fmt,siz,ptag,tmr)) != 0) return cc;
  return 0;
}
int tmonitor(char *dev,char *prp,char *tag)
{
  return tmonitorEx(dev,prp,CF_NULL,0,tag,1000);
}
int tmonitorWait(char *dev,char *prp,char *datastr,int fmt_,int siz_,int tmr)
{
  int cc;
  int fmt,fsiz,hsiz,siz,tsiz;

  if (fmt_ != CF_NULL && siz_ != 0)
  {
    fmt = fmt_;
    siz = siz_;
    prpinfo.prpFormat = BFMT(fmt);
    prpinfo.prpSize = siz;
  }
  else
  {
    if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
    fmt = LFMT(prpinfo.prpFormat);
    siz = prpinfo.prpSize;
  }
  assertBlobSize(fmt,&siz);
  fsiz = fmtsizeof(fmt); hsiz = getFormatHeaderSize(fmt);
  if (siz > (blob_size-hsiz)/fsiz) siz = (blob_size-hsiz)/fsiz;

  waitTgtTyp.dataOutput = _bbval; /* (BYTE *)blob */
  waitTgtTyp.sizOutputArray = siz;
  waitTgtTyp.dataTarget = parseInputDataString(datastr,&fmt,&tsiz);
  waitTgtTyp.sizTargetArray = tsiz;
  waitTgtTyp.fmt = fmt;
  waitTgtTyp.waitt = tmr;
  waitTgtTyp.startt = time(NULL);

  if ((cc=_tmonitorWait(blob,dev,prp,fmt,siz,1000)) != 0) return cc;
  return 0;
}

time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = (int)strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year,"_")) != NULL || (c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
int talarms(char *srv,char *dev)
{
  char devsrv[132], datastr[4096], *pdev = "*";
  DTYPE dout;
  int i, cc, len = BLOB_SIZE;
  double dval;
  AMS *ams;

  if (dev != NULL) pdev = dev;
  snprintf(acqStr, 1024, "talarms: server address = \"%s\" device = \"%.64s\"",
    srv == NULL ? "(null)" : srv, pdev);
  if (srv == NULL) return argument_list_error;

  assertBlobSize(CF_BYTE,&len);
  ams = (AMS *)blob;
  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = blob_size/AMS_SIZE;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"AMSr4");
  dout.data.vptr = (void *)ams;

  sprintf(devsrv,"%.66s/%.64s",srv,pdev);
  cc = ExecLinkEx(devsrv,"ALARMS",&dout,NULL,CA_READ|CA_MUTABLE|CA_RETRY,800);
  switch (cc)
  {
    case 0:
      break;
    case illegal_format:
    case invalid_structure_tag:
      printf("device server %s appears to be pre-release 4.0 and does not support this alarms call!\n",srv);
    default:
      return cc;
  }
  printf("\nFound %d alarms for %s\n\n",dout.dArrayLength,devsrv);
  for (i=0; i<(int)dout.dArrayLength; i++)
  {
    printf("[alarm %d]\n{\n",i);
    printf("\tserver     : %.32s\n",ams[i].server);
    printf("\tdevice     : %.64s\n",ams[i].device);
    printf("\talarmTag   : %.32s\n",ams[i].alarmTag);
    printf("\talarmCode  : %d\n",ams[i].alarmCode);
    dval = (double)ams[i].timestamp + (double)(ams[i].timestampUSec)/1000000.0;
    printf("\ttimestamp  : %10.3f\n",dval);
    dval = (double)ams[i].starttime + (double)(ams[i].starttimeUSec)/1000000.0;
    printf("\tstarttime  : %10.3f\n",dval);
    printf("\talarmMask  : 0x%x\n",ams[i].alarmMask);
    printf("\tseverity   : %d\n",ams[i].severity);
    printf("\tdescriptor : %x\n",ams[i].descriptor);
    printf("\talarmSystem: %d\n",ams[i].alarmSystem);
    if (ams[i].alarmDataArraySize > 0)
    {
      prepDataDump(datastr,(char *)ams[i].alarmData,ams[i].alarmDataArraySize,ams[i].alarmDataFormat,TRUE);
      printf("\talarmData  : %s\n",datastr);
    }
    else
    {
      printf("\talarmData  : none\n");
    }
    printf("}\n");
  }
  return 0;
}
/* #define HST_AS_ANY */
#ifdef HST_AS_ANY
HstHdr dHdr[5000];
#endif
int thistory(char *context,char *keyword,char *devname,char *stoptime,char *depth)
{
  char devsrv[256],srvnam[64], *ctx, *srv;
  time_t stop = time(NULL), start, ts;
  int mf = 1, dp = 0, isLocalHistory = FALSE;
  int i, num = 5000, cc, offset = 0, fmt;
  int isName = FALSE, len = BLOB_SIZE;
  FLTINT *fiDataArray;
  NAME32I *niDataArray;
  float *fDataArray;
  snprintf(acqStr, 1024, "thistory: context = \"%.32s\" property = \"%.64s\" device = \"%.64s\" stop at %s start %s prior",
    context == NULL ? "(null)" : context,
    keyword == NULL ? "(null)" : keyword,
    devname == NULL ? "(null)" : devname,
    stoptime == NULL ? "(null)" : stoptime,
    depth == NULL ? "(null)" : depth);
  if (context == NULL || keyword == NULL) return argument_list_error;
  if (devname == NULL || strlen(devname) == 0) devname = "#0";
  assertBlobSize(CF_BYTE,&len);
  fiDataArray = (FLTINT *)blob;
  niDataArray = (NAME32I *)blob;
  fDataArray = (float *)blob;
  strncpy(devsrv,context,256);
  strupr(devsrv);
  if (devsrv[0] == '/')
  { /* full name specified */
    ctx = &devsrv[1];
    if ((srv=strchr(ctx,'/')) != NULL)
    { /* context and server given (this is normal) */
      srv++;
      if (strstr(srv,"HISTORY") == NULL) isLocalHistory = TRUE; 
    }
    else
    { /* server name not give (?) append it : */
      strcat(devsrv,"/HISTORY");
    }
  }
  else
  { /* just the context given -> assume central archive */
    ctx = devsrv;
    if (strchr(ctx,'/') != NULL) 
    {
      printf("target server %s does not have a valid context",context);
      return argument_list_error;
    }
    sprintf(devsrv,"/%.32s/HISTORY",context);
  }

  strncpy(srvnam,devsrv,64);

  if (devsrv[strlen(devsrv)-1] != '/') strcat(devsrv,"/");
  strcat(devsrv,devname);
  strcat(devsrv,"/");
  strcat(devsrv,keyword);
  if (isLocalHistory && strstr(keyword,".HIST") == NULL) strcat(devsrv,".HIST");
 
  stop = parseTimeString(stoptime);
  isName = strstr(keyword,".NAM") == NULL ? FALSE : TRUE;

  if (!stricmp(depth,"SNAPSHOT"))
  { /* snapshot request ? */
    tgetFormatAndSize(srvnam,keyword,&fmt,&num,NULL);
    if ((cc=GetArchivedTraceDataAsFloat(devsrv,stop,stop,fDataArray,num)) == 0)
    {
      stop = (time_t)gDataTimeStamp;
      printf("snapshot of %s %s/%s at %s",context,keyword,devname,ctime(&stop));
      for (i=0; i<num; i++)
      {
        printf("%g\n",fDataArray[i]);
      }
    }
    return cc;
  }  
  
  strupr(depth);
  if (strstr(depth,"DAY")) mf = 24 * 3600;
  else if (strstr(depth,"HOUR")) mf = 3600;
  else if (strstr(depth,"MINUTE")) mf = 60;
  else if (strstr(depth,"WEEK")) mf = 24 * 3600 * 7;
  dp = mf * atoi(depth);
  start = stop - dp;
  printf("%s/%s, Timestamp\n",keyword,devname);

# ifdef HST_AS_ANY
  if ((cc=GetArchivedDataAsAny(devsrv,start,stop,dHdr,(BYTE *)fDataArray,CF_FLOAT,"",&num)) == 0) 
  {
    for (i=offset; i<num; i++)
    {
      ts = (time_t)dHdr[i].timeStamp;
      printf("%g, %s",fDataArray[i],ctime(&ts));
    }
    return 0;
  }
# endif

  if (isName) goto getText;
  if ((cc=GetArchivedDataAsFloat(devsrv,start,stop,fiDataArray,&num)) == 0) 
  {
    if ((int)fiDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num; i++)
    {
      ts = (time_t)fiDataArray[i].ival;
      printf("%g, %s",fiDataArray[i].fval,ctime(&ts));
    }
    return 0;
  }
  if (cc != illegal_format) return cc;
getText:
  if ((cc=GetArchivedDataAsText(devsrv,start,stop,niDataArray,&num)) == 0) 
  {
    if ((int)niDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num; i++)
    {
      ts = (time_t)niDataArray[i].ival;
      printf("%.32s, %s",niDataArray[i].name,ctime(&ts));
    }
    return 0;
  }
  
  return cc;
}
// from caseqm.h:
typedef struct
{
  char email[128];
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char actionSignalString[128];
  int actionSignalFormat;
  int alarmCode;
} CasActionItem;
#define CASACTIONITEM_TAG "CasAct"
#define XTN_NAME_SIZE 16
typedef struct
{
  char devContext[CONTEXT_NAME_SIZE];
  char devServer[EXPORT_NAME_SIZE];
  char subSystem[SUBSYSTEM_NAME_SIZE];
  char arcExt[XTN_NAME_SIZE];
  int SeverityLevel;
  int ArchiveLevel;
  int Retention;
  int AlarmLevel;
  int AlarmSystem;
  int offline;
} CasDbItem; /* CAS Database Item */
#define CASDBITEM_TAG "CasDb"
#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerCasStructs(void)
{
  static int done = FALSE;
  if (done) return;
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, email), 128, CF_TEXT, "email")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, ctx), CONTEXT_NAME_SIZE, CF_TEXT, "context")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, srv), EXPORT_NAME_SIZE, CF_TEXT, "server")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, dev), DEVICE_NAME_SIZE, CF_TEXT, "device")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, prp), PROPERTY_NAME_SIZE, CF_TEXT, "property")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, actionSignalString), 128, CF_TEXT, "actionData")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, actionSignalFormat), 1, CF_INT32, "actionFormat")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG, OFFSETIN(CasActionItem, alarmCode), 1, CF_INT32, "alarmCode")) quit(1);
  if (sealTaggedStruct(CASACTIONITEM_TAG, sizeof(CasActionItem), 100)) quit(1);

  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, devContext), CONTEXT_NAME_SIZE, CF_TEXT, "context")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, devServer), EXPORT_NAME_SIZE, CF_TEXT, "server")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, subSystem), SUBSYSTEM_NAME_SIZE, CF_TEXT, "subsys")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, arcExt), XTN_NAME_SIZE, CF_TEXT, "ext")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, SeverityLevel), 1, CF_INT32, "severityLevel")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, ArchiveLevel), 1, CF_INT32, "archiveLevel")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, Retention), 1, CF_INT32, "retention")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, AlarmLevel), 1, CF_INT32, "alarmLevel")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, AlarmSystem), 1, CF_INT32, "alarmSystem")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG, OFFSETIN(CasDbItem, offline), 1, CF_INT32, "offline")) quit(1);
  if (sealTaggedStruct(CASDBITEM_TAG, sizeof(CasDbItem), 500)) quit(1);
  done = TRUE;
}
int addtocas(char *context, char *target, char *almsys, int code, char **lst, int lstsiz)
{
  char devsrv[132], tgt[132], *ctx=NULL, *srv=NULL, *c;
  DTYPE dout, din;
  CasDbItem dbItem;
  CasActionItem *actItems = NULL;
  NAME16 n16;
  int i, n, cc, sys;

  if (context == NULL || target == NULL) return argument_list_error;
  DTYPEZERO(dout); DTYPEZERO(din);
  strncpy(tgt, target, 132);
  if (tgt[0] == '/')
  {
    ctx = &tgt[1];
    if ((srv=strchr(ctx, '/')) == NULL) return non_existent_server;
    *srv++ = 0;
  }
  else
  {
    ctx = context;
    srv = tgt;
  }
  if (almsys == NULL || almsys[0] == '0')
  {
    sys = ALM_SYSTEM_SYSTEM;
  }
  else
  {
    sys = atoi(almsys);
    if (sys == 0)
    {
      NAME16I n16i[100];
      sprintf(devsrv, "/%.32s/CAS", context);
      dout.dArrayLength = 100;
      dout.dFormat = CF_NAME16I;
      dout.data.vptr = n16i;
      cc = ExecLinkEx(devsrv, "ALARMSYSTEMS", &dout, NULL, CA_READ, 500);
      if (cc) return cc;
      for (i = 0; i < 100; i++) if (!strnicmp(almsys, n16i[i].name, 16)) break;
      if (i < 100) sys = n16i[i].ival;
    }
  }
  memset(&dbItem, 0, sizeof(dbItem));
  strncpy(dbItem.devContext, ctx, CONTEXT_NAME_SIZE);
  strncpy(dbItem.devServer, srv, EXPORT_NAME_SIZE);
  dbItem.AlarmLevel = 7;
  dbItem.AlarmSystem = sys;
  strncpy(dbItem.arcExt, srv, XTN_NAME_SIZE);
  dbItem.Retention = 3600;
  dbItem.SeverityLevel = 1;
  if (code < 0)
  { // abuse this parameter to signal a removal!
    dbItem.offline = -1;
  }
  sprintf(devsrv, "/SITE/ENS//%.32s/%.32s", ctx, srv);
  dout.dArrayLength = 1;
  dout.dFormat = CF_NAME16;
  dout.data.vptr = &n16;
  if (ExecLinkEx(devsrv, "SUBSYSTEM", &dout, NULL, CA_READ, 500) != 0)
  {
    strcpy(n16.name, "SER");
  }
  strncpy(dbItem.subSystem, n16.name, 16);
  registerCasStructs();
  din.dArrayLength = 1;
  din.dFormat = CF_STRUCT;
  strncpy(din.dTag, CASDBITEM_TAG, TAG_NAME_SIZE);
  din.data.vptr = &dbItem;

  sprintf(devsrv, "/%.32s/CAS", context);
  cc = ExecLink(devsrv, "SERVER.DB", NULL, &din, CA_WRITE);
  if (cc != 0) return cc;
  if (code <= 0 || lst == NULL) lstsiz = 0;
  if (lstsiz == 0) return 0;
  if ((actItems=(CasActionItem*)calloc(lstsiz, sizeof(CasActionItem))) == NULL) return out_of_local_memory;
  for (i=0,n=0; i < lstsiz; i++)
  {
    c = lst[i];
    memset(&actItems[n], 0, sizeof(CasActionItem));
    if (strnicmp(c, "mailto:", 7) == 0)
    {
      c += 7;
      strncpy(actItems[n].email, c, 128);
      actItems[n].alarmCode = code;
      actItems[n].actionSignalFormat = CF_NULL;
      n++;
      continue;
    }
    if (strnicmp(c, "trigger:", 8) == 0)
    {
      c += 8;
      strncpy(actItems[n].ctx, context, CONTEXT_NAME_SIZE);
      strncpy(actItems[n].srv, "EVENTSTORE",EXPORT_NAME_SIZE);
      strncpy(actItems[n].dev, c, DEVICE_NAME_SIZE);
      strncpy(actItems[n].prp, "TRIGGER", PROPERTY_NAME_SIZE);
      actItems[n].alarmCode = code;
      actItems[n].actionSignalFormat = CF_INT32;
      strcpy(actItems[n].actionSignalString, "1");
      n++;
      continue;
    }
    if (strnicmp(c, "annotate:", 9) == 0)
    {
      c += 9;
      strncpy(actItems[n].ctx, context, CONTEXT_NAME_SIZE);
      strncpy(actItems[n].srv, "HISTORY", EXPORT_NAME_SIZE);
      strncpy(actItems[n].prp, c, PROPERTY_NAME_SIZE);
      actItems[n].alarmCode = code;
      actItems[n].actionSignalFormat = CF_TEXT;
      strcpy(actItems[n].actionSignalString, "ANNOTATE");
      n++;
      continue;
    }
    continue;
  }
  strcat(devsrv, "/"); strcat(devsrv, srv);
  din.dArrayLength = n;
  din.dFormat = CF_STRUCT;
  din.data.vptr = actItems;
  strncpy(din.dTag, CASACTIONITEM_TAG, TAG_NAME_SIZE);
  cc = ExecLink(devsrv, "ACTION.DB", NULL, &din, CA_WRITE);
  if (cc != 0) return cc;
  return 0;
}
// from archive.h :
typedef struct
{
  char key[PROPERTY_NAME_SIZE]; /**< parameter name */
  char units[UNITS_SIZE];       /**< units */
  char desc[PROPERTY_DESC_SIZE]; /**< description */
  char plot[UNITS_SIZE];        /**< plot style (LOG, LIN, etc.) */
  char boundTo[PROPERTY_NAME_SIZE]; /**< bound to property reference */
  char sub[SUBSYSTEM_NAME_SIZE]; /**< parameter subsystem */
  char assoc[PROPERTY_NAME_SIZE]; /**< associate dev names to */
  int fmt;              /**< allowable format */
  int size;             /**< data array size assuming given format */
  float scaleFactor;
  float scaleOffset;
  float maximum;       /**< maximum value for parameter */
  float minimum;       /**< minimun value for parameter */
  float aTolerance;    /**< absolute tolerance in units given */
  float pTolerance;    /**< percent tolerance */
} DAQPrmInfo;
#define DAQPrmTag "DAQPrm"
typedef struct
{
  char fn_key[80]; /**< keyword structure file name */
  char tag[80]; /**< struct tag name/reserved string space */
  char ctx[CONTEXT_NAME_SIZE]; /**< calling context */
  char srv[EXPORT_NAME_SIZE]; /**< calling equipment module */
  char prp[PROPERTY_NAME_SIZE]; /**< calling property */
  char dev[DEVICE_NAME_SIZE]; /**< calling device name */
  int len; /**< length of call */
  int fmt; /**< call format */
  int hbt; /**< archiving heartbeat (not to be exceeded) */
  int tmr; /**< polling interval */
  int infmt; /**< input data format (if any) */
  int inlen; /**< number of input data points */
  UINT32 opt; /**< option word */
  int  active;/**< is table entry active */
  int rec; /**< entry in record table */
} DAQRecInfo;
#define DAQRecTag "DAQRec"
void registerDAQStructs(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, key), PROPERTY_NAME_SIZE, CF_TEXT, "key")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, units), UNITS_SIZE, CF_TEXT, "units")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, desc), PROPERTY_DESC_SIZE, CF_TEXT, "desc")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, plot), UNITS_SIZE, CF_TEXT, "plot")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, boundTo), PROPERTY_NAME_SIZE, CF_TEXT, "boundTo")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, sub), SUBSYSTEM_NAME_SIZE, CF_TEXT, "sub")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, assoc), PROPERTY_NAME_SIZE, CF_TEXT, "assoc")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, fmt), 1, CF_INT32, "fmt")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, size), 1, CF_INT32, "size")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, scaleFactor), 1, CF_FLOAT, "scaleFactor")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, scaleOffset), 1, CF_FLOAT, "scaleOffset")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, maximum), 1, CF_FLOAT, "maximum")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, minimum), 1, CF_FLOAT, "minimum")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, aTolerance), 1, CF_FLOAT, "aTolerance")) quit(1);
  if (addFieldToStruct(DAQPrmTag, OFFSETIN(DAQPrmInfo, pTolerance), 1, CF_FLOAT, "pTolerance")) quit(1);
  if (sealTaggedStruct(DAQPrmTag, sizeof(DAQPrmInfo), 10)) quit(1);

  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, fn_key), 80, CF_TEXT, "keyFile")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, tag), 80, CF_TEXT, "tag")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, ctx), CONTEXT_NAME_SIZE, CF_TEXT, "context")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, srv), EXPORT_NAME_SIZE, CF_TEXT, "server")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, prp), PROPERTY_NAME_SIZE, CF_TEXT, "property")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, dev), DEVICE_NAME_SIZE, CF_TEXT, "device")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, len), 1, CF_INT32, "length")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, fmt), 1, CF_INT32, "fmt")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, hbt), 1, CF_INT32, "heartbeat")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, tmr), 1, CF_INT32, "pollInterval")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, infmt), 1, CF_INT32, "inputFmt")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, inlen), 1, CF_INT32, "inputLength")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, opt), 1, CF_INT32, "filterOptions")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, active), 1, CF_INT32, "isActive")) quit(1);
  if (addFieldToStruct(DAQRecTag, OFFSETIN(DAQRecInfo, rec), 1, CF_INT32, "recordIndex")) quit(1);
  if (sealTaggedStruct(DAQRecTag, sizeof(DAQRecInfo), 10)) quit(1);
  return;
}
int addtoarc(char *context, char *target, char *prp, char *kwd, char *subsys, char *fmtstr, int len, char *tolstr)
{
  char devsrv[132], tgt[256], *ctx = NULL, *srv = NULL, *dev = NULL;
  DTYPE dout, din;
  DAQPrmInfo prm;
  DAQRecInfo rec;
  int cc, fmt = CF_NULL, npqs = 1;
  float ptol = 0.0, atol = 0.0;
  PrpQueryStruct pqs[1], *ppqs = pqs;

  if (context == NULL || target == NULL || prp == NULL || kwd == NULL) return argument_list_error;
  DTYPEZERO(dout); DTYPEZERO(din);
  // does keyword already exist? (if yes, then bail)
  dout.dArrayLength = 256;
  dout.dFormat = CF_TEXT;
  dout.data.cptr = tgt;
  sprintf(devsrv, "/%.32s/HISTORY/%64s", context, kwd);
  cc = ExecLink(devsrv, "KEYWORD.SOURCE", &dout, NULL, CA_READ);
  if (cc == 0)
  { // this is bad!
    printf("keyword %.64s already bound to %132s\n", kwd, tgt);
    return already_assigned;
  }
  strncpy(tgt, target, 132);
  if (target[0] == '/')
  {
    ctx = &tgt[1];
    if ((srv = strchr(ctx, '/')) == NULL) return non_existent_server;
    *srv++ = 0;
  }
  else
  {
    ctx = context;
    srv = tgt;
  }
  if ((dev = strchr(srv, '/')) == NULL) dev = "#0"; else *dev++ = 0;
  memset(pqs, 0, sizeof(PrpQueryStruct));
  cc = GetSystemPropertyInformation(target, prp, &ppqs, &npqs);
  if (fmtstr == NULL || strlen(fmtstr) == 0)
  {
    if (cc != 0)
    {
      printf("format not provided in call and property information unavailable from target : %.32s\n",
        cc2str(cc));
      return cc;
    }
    fmt = LFMT(pqs[0].prpFormat);
    if (len <= 0) len = (int)pqs[0].prpSize;
  }
  else
  {
    fmt = GetFormatFromString(fmtstr);
    if (fmt == CF_NULL)
    {
      printf("format %.16s not supported\n", fmtstr);
      return illegal_format;
    }
    if (len <= 0) len = 1;
  }
  registerDAQStructs();
  memset(&rec, 0, sizeof(rec));
  memset(&prm, 0, sizeof(prm));
  // prepare the record entry :
  strncpy(rec.ctx, ctx, CONTEXT_NAME_SIZE);
  strncpy(rec.srv, srv, EXPORT_NAME_SIZE);
  strncpy(rec.prp, prp, PROPERTY_NAME_SIZE);
  strncpy(rec.dev, dev, DEVICE_NAME_SIZE);
  sprintf(rec.fn_key, "%.80s.csv", kwd);
  rec.fmt = fmt;
  rec.len = len;
  rec.infmt = CF_NULL;
  rec.inlen = 0;
  rec.tmr = 1000;
  rec.hbt = 900;
  rec.rec = -1;
  rec.active = 1;
  rec.opt = 0x44;
  // prepare the parameter keyword entry :
  if (tolstr != NULL && strlen(tolstr) != 0)
  {
    if (strchr(tolstr, '%'))
    {
      ptol = (float)(atof(tolstr) / 100.0);
    }
    else
    {
      atol = (float)atof(tolstr);
    }
  }
  strncpy(prm.key, kwd, PROPERTY_NAME_SIZE);
  prm.scaleFactor = 1.0;
  prm.scaleOffset = 0;
  prm.fmt = fmt;
  prm.size = len;
  prm.pTolerance = ptol;
  prm.aTolerance = atol;
  if (subsys != NULL) strncpy(prm.sub, subsys, SUBSYSTEM_NAME_SIZE);
  strncpy(prm.desc, pqs[0].prpDescription, PROPERTY_DESC_SIZE);
  prm.maximum = pqs[0].prpMaxValue;
  prm.minimum = pqs[0].prpMinValue;
  strncpy(prm.units, pqs[0].prpUnits, UNITS_SIZE);

  sprintf(devsrv, "/%.32s/HISTORY/%.64s", context, rec.fn_key);
  din.dArrayLength = 1;
  din.dFormat = CF_STRUCT;
  strncpy(din.dTag, DAQPrmTag, TAG_NAME_SIZE);
  din.data.vptr = &prm;
  cc = ExecLink(devsrv, "KEYWORD.INFO", NULL, &din, CA_WRITE);
  if (cc != 0)
  {
    printf("could not write keyword info for %.32s : %.32s\n", kwd, cc2str(cc));
    return cc;
  }
  sprintf(devsrv, "/%.32s/HISTORY", context);
  strncpy(din.dTag, DAQRecTag, TAG_NAME_SIZE);
  din.data.vptr = &rec;
  cc = ExecLink(devsrv, "RECORD.INFO", NULL, &din, CA_WRITE);
  if (cc != 0)
  {
    printf("could not write record info for keyword %.32s : %.32s\n", kwd, cc2str(cc));
    return cc;
  }
  return 0;
}
int reloadarc(char *context)
{
  char devsrv[132];
  int cc = 0;
  sprintf(devsrv, "/%.32s/HISTORY",context);
  cc = ExecLink(devsrv, "DB.RELOAD", NULL, NULL, CA_WRITE);
  if (cc != 0) ccerr(cc);
  sprintf(devsrv, "/%.32s/ARCHIVER",context);
  cc = ExecLink(devsrv, "DB.RELOAD", NULL, NULL, CA_WRITE);
err:
  return cc;
}
enum
{
  EQ = 0,
  LT = 1,
  LE = 2,
  GT = 3,
  GE = 4
};
int almInfoCheck(char *cnt, char *cmpstr, char *field, int sev, int sys)
{
  int i, k, cc = 0, n = 500, v, item_v, na = 100, cmp = GT;
  int skip = FALSE, banner=FALSE, chkSeverity=TRUE;
  DTYPE dout, din;
  CasDbItem dbitems[500];
  ADS adsitems[100];
  char srv[128];
  registerCasStructs();
  if (cmpstr != NULL)
  {
    if (!strcmp(cmpstr, "=") || !strcmp(cmpstr, ".EQ.") || !strcmp(cmpstr, "EQ")) cmp = EQ;
    else if (!strcmp(cmpstr, "<") || !strcmp(cmpstr, ".LT.") || !strcmp(cmpstr, "LT")) cmp = LT;
    else if (!strcmp(cmpstr, "<=") || !strcmp(cmpstr, ".LE.") || !strcmp(cmpstr, "LE")) cmp = LE;
    else if (!strcmp(cmpstr, ">") || !strcmp(cmpstr, ".GT.") || !strcmp(cmpstr, "GT")) cmp = GT;
    else if (!strcmp(cmpstr, "=>") || !strcmp(cmpstr, ".GE.") || !strcmp(cmpstr, "GE")) cmp = GE;
  }

  if (cnt == NULL || strlen(cnt) == 0) cnt = "TEST";
  if (sev < 0) sev = 0;
  assertBlobSizeEx(CF_STRUCT, "CasDb", &n);
  chkSeverity = strcmp(field, "system") == 0 ? FALSE : TRUE;
  v = chkSeverity ? sev : sys;
  printf("check %.32s for %s %s %d\n", cnt, field, cmpstr, v);
  DTYPEZERO(din); DTYPEZERO(dout);
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 500;
  strcpy(dout.dTag,CASDBITEM_TAG);
  dout.data.vptr = dbitems;
  memset(dbitems, 0, n * sizeof(CasDbItem));
  sprintf(srv, "/%.32s/CAS/#0", cnt);
  if ((cc=ExecLinkEx(srv, "SERVER.DB", &dout, &din, CA_READ|CA_MUTABLE, 200)) != 0) return cc;
  n = (int)dout.dArrayLength;
  for (i = 0; i < n; i++)
  {
    sprintf(srv, "/%.32s/CAS/%.64s", cnt,dbitems[i].devServer);
    dout.dFormat = CF_STRUCT;
    dout.dArrayLength = 100;
    strcpy(dout.dTag, "ADSr4");
    dout.data.vptr = adsitems;
    na = 100;
    memset(adsitems, 0, na * sizeof(CasDbItem));
    if ((cc=ExecLinkEx(srv, "SYSALMDEFS", &dout, &din, CA_READ | CA_MUTABLE, 200)) != 0)
    {
      printf("%.64s : could not get info -> %32s\n", dbitems[i].devServer, cc2str(cc));
    }
    else
    {
      na = (int)dout.dArrayLength;
      printf("--- %.64s : found %d alarm definitions ---\n", dbitems[i].devServer, na);
      banner = TRUE;
      for (k = 0; k < na; k++)
      {
        item_v = chkSeverity ? adsitems[k].alarmSeverity : adsitems[k].alarmSystem;
        switch (cmp)
        {
          case EQ: skip = item_v != v ? TRUE : FALSE; break;
          case LT: skip = item_v >= v ? TRUE : FALSE; break;
          case LE: skip = item_v >  v ? TRUE : FALSE; break;
          case GT: skip = item_v <= v ? TRUE : FALSE; break;
          case GE: skip = item_v <  v ? TRUE : FALSE; break;
          default: skip = FALSE;
        }
        if (skip) continue;
        if (banner)
        {
          printf("\tthose with %s %s %d:\n", field, cmpstr, v);
          banner = FALSE;
        }
        printf("\t%.64s (code = %d, severity = %d, system = %d)\n", 
          adsitems[k].alarmTag, adsitems[k].alarmCode, adsitems[k].alarmSeverity, adsitems[k].alarmSystem);
      }
    }
  }
  return cc;
}
int wdogCheck(char *cnt)
{
  int i, k, p, cc, imp = 0, tries = 0, nprocs, npqs;
  int isgrp = FALSE, fmt = CF_USTRING, n = GetNumberOf("/SITE/ENS/#0", "PROPERTIES");
  NAME32 *srvs;
  DUSTRING *usrvs;
  PrpQueryStruct pqs[10], *ppqs = pqs;
  NAME64 wdprocs[100];
  BYTE ipbytes[4];
  char wdsrv[256], srv[128], ip[64], *psrvname, *search="/sub=WDOG";
  char host[64], hint[128], *msg, *c, *fec;
  DTYPE dout, din;
  if (n < 0) return -n;
  if (n < 1024) n = 1024;
  if (cnt == NULL || strlen(cnt) == 0) cnt = "SERVICE";
  snprintf(acqStr, 1024, "wdogchk: context = \"%s\"", cnt);
  assertBlobSize(fmt, &n);
  srvs = _bn32val;
  usrvs = _ustrval;
  strcpy(srv, "ENS");
  if (cnt != NULL && strlen(cnt) > 0)
  {
    strcat(srv, "/"); 
    strncat(srv, cnt, CONTEXT_NAME_SIZE);
  }
  DTYPEZERO(din); DTYPEZERO(dout);
  din.dFormat = CF_TEXT;
  din.dArrayLength = (UINT32)strlen(search);
  din.data.cptr = search;
  dout.dFormat = fmt;
  dout.dArrayLength = n;
  dout.data.vptr = srvs;
  memset(srvs, 0, n * fmtsizeof(fmt));
  if ((cc=ExecLinkEx(srv, "SERVERS", &dout, &din, CA_READ|CA_MUTABLE, 200)) != 0) return cc;
  n = (int)dout.dArrayLength;
  printf("watchdog servers in context %s", cnt);
  printf(":\n\n");
  printf("%-20s %-16s %-16s %-5s %-8s %-20s %s\n", "Server", "FEC", "IP Address", "Port", "OS", "Host", "Comment");
  DTYPEZERO(dout);
  for (i = 0, k = 0; i < n; i++)
  {
    psrvname = usrvs[i].str;
    if (strlen(psrvname) == 0) continue;
    isgrp = FALSE;
    memset(srv, 0, 64);
    strncpy(srv, psrvname, 16);
    memcpy(ipbytes, &usrvs[i].ival, 4);
    if (srv[15] == '&')
    {
      strncpy(&srv[15], &usrvs[i].str[48], 16);
    }
    inet_ntop(gAfInet, &usrvs[i].ival, ip, 40);
    fec = &usrvs[i].str[32]; // oh, the abuse a USTRING must go thru ...
    //sprintf(ip, "%d.%d.%d.%d", ipbytes[0], ipbytes[1], ipbytes[2], ipbytes[3]);
    if (strncmp(fec,"VXWDOG", 6) == 0) continue; // skip this
    GetHostNameFromAddress(ip, host, 64); host[63] = 0;
    if ((c=strchr(host,'.')) != NULL) *c = 0;
    msg = stricmp(&srv[3], host) ? "WARNING!" : "OK";
    msg = "OK";
    if (stricmp(&srv[3], host))
    {
      if ((c=strchr(&srv[3],'.')) != NULL && isdigit(c[1]))
      {
        msg = "WDOG GROUP MEMBER";
      }
      else
      {
        msg = "WARNING: host name != WD_servername";
      }
      if (!strcmp(fec, "GENS"))
      {
        msg = "WDOG GROUP";
        isgrp = TRUE;
      }
    }
    dout.dFormat = CF_NAME64;
    dout.dArrayLength = 100;
    dout.data.vptr = wdprocs;
    sprintf(wdsrv, "/%.32s/%.32s/#0", cnt, srv);
    if ((cc=ExecLinkEx(wdsrv, "DEVICES", &dout, NULL, CA_READ|CA_MUTABLE, 500)) != 0)
    {
      msg = "WARNING: not running!";
      nprocs = 0;
    }
    else
    {
      nprocs = dout.dArrayLength;
    }
    printf("%-20.32s %-16.16s %-16.16s %-5d %-8.8s %-20s %s\n", srv, fec, ip, (int)usrvs[i].f1val, &usrvs[i].str[16], host, msg);
    if (isgrp) continue; // skip: the processes have alredy been reported
    tries = 0;
    for (p = 1; p < nprocs; p++)
    {
      npqs = 10;
      if ((cc=GetSystemPropertyInformation("/SITE/ENS/#0", wdprocs[p].name, &ppqs, &npqs)) == 0)
      {
        tries = 0;
        c = pqs[0].prpDescription;
        if (strstr(c, "registered FEC") != NULL)
        {
          c = "OK";
        }
        else
        {
          ExpDataStruct srvaddr;
          FecAddrStruct fecaddr;
          NAME16 falias;
          dout.dArrayLength = 1;
          dout.dFormat = CF_NAME16;
          dout.data.vptr = &falias;
          din.dArrayLength = 1;
          din.dFormat = CF_NAME16;
          din.data.vptr = wdprocs[p].name;
          if ((cc=ExecLinkEx("/SITE/ENS/#0", "FECALIASLIST", &dout, &din, CA_READ | CA_MUTABLE, 500)) == 0)
          { // fec name = server name
            c = "OK";
          }
          else if ((cc=GetServerAddr(NULL, wdprocs[p].name, &srvaddr, &fecaddr)) == 0)
          {
            if (!strnicmp(wdprocs[p].name,srvaddr.FecName,16))
            { // fec name = server name
              c = "OK";
            }
            else
            {
              sprintf(hint, "WARNING: server name used (possible FEC = %.16s)", srvaddr.FecName);
              c = hint;
            }
          }
          else if (cc == ambiguous)
          {
            sprintf(hint, "WARNING: server name used and multiple FECS use this name (in different contexts)");
            c = hint;
          }
        }
      }
      else if (cc == connection_timeout || cc == link_timeout)
      {
        if (tries++ < 3)
        { // try again ...
          p--;
          continue;
        }
        c = "WARNING: timed out getting FEC information from ENS!";
      }
      else
      {
        short autostart = 0;
        c = "WARNING: Not a registered FEC!";
        sprintf(wdsrv, "/%.32s/%.32s/%.64s", cnt, srv, wdprocs[p].name);
        dout.dArrayLength = 1;
        dout.dFormat = CF_INT16;
        dout.data.sptr = &autostart;
        if ((cc=ExecLinkEx(wdsrv, "Info.autoStart", &dout, NULL, CA_READ | CA_MUTABLE, 500)) == 0)
        {
          if (!autostart)
          {
            c = "WARNING: Not running and not yet registered!";
          }
        }
      }
      printf(" --- Process %2d ---  %-16.16s %52s %s\n", p, wdprocs[p].name, " ", c);
    }
    k++;
  }
  printf("\n%d servers found\n", k);
  return 0;
}
