/*
 * File cmdcln.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "tine.h"

#define CLN_PATH "/tmp/"
#ifndef CLN_PATH
#  define CLN_PATH "/var/tmp/"
#endif
#define CLN_PERM S_IRWXU

int pipeCln(const char *name)
{
  int fd,len;
  struct sockaddr_un unix_addr;
#ifndef SVR4
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) return -1;

  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  
  sprintf(unix_addr.sun_path, 
          strlen(CLN_PATH) == 9 ? "%s%05d" : "%s%09d", 
          CLN_PATH, getpid());

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if (len != 16) return -5;
# endif

  unlink(unix_addr.sun_path);
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) return -2;
  if (chmod(unix_addr.sun_path, CLN_PERM) < 0) return -3;

  memset(&unix_addr, 0, sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, name);

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
# endif
  if (connect(fd, (struct sockaddr *)&unix_addr,len) < 0) return -4;
#else
  if ((fd=open(name,O_RDWR)) < 0) return -1;
  if (isastream(fd) == 0) return -2;
#endif /* SVR4 */

  return fd;
}

int sckCln(const char *ip,UINT16 port)
{
  int sck,cc=0;
  int optval = 1, mode = 1;
  struct timeval tv0,tv;
  struct sockaddr_in srv_name;
  UINT32 haddr;

  haddr = inet_addr(ip);
  memset(&srv_name,0,sizeof(srv_name));
  memcpy(&srv_name.sin_addr,&haddr,sizeof(struct in_addr));
  srv_name.sin_family = AF_INET;
  srv_name.sin_port = htons(port);

  if ((sck=socket(AF_INET,SOCK_STREAM,0)) < 0) return -1;

  if (setsockopt(sck,SOL_SOCKET,SO_KEEPALIVE,(char *)&optval,sizeof(optval))) perror("SO_KEEPALIVE");
  /* set to non-blocking for the connection attempt : */
  if (ioctl(sck,FIONBIO,(char *)&mode)) perror("ioctl non blocking");
  gettimeofday(&tv0,(struct timezone *)NULL);  tv = tv0;
  while (connect(sck,(struct sockaddr *)&srv_name,sizeof(srv_name)) < 0)
  {
    if (errno == EISCONN) break;
    gettimeofday(&tv,(struct timezone *)NULL);
    if (MSECS(tv,tv0) > 1000) ccerr(-1);
  }
  mode = 0;
  if (ioctl(sck,FIONBIO,(char *)&mode)) perror("ioctl blocking");
err:
  if (cc) 
  {
    close(sck);
    sck = 0;
  }

  return sck;
}

#define MAXLINES 1024

#define quit(txt) { perror(txt); return -1; }
char cmd[256];

int cleanup(void)
{
  static int step = 1;
  
  switch( step )
  {
    case 1: strcpy(cmd,"set debug=0\n"); break;
    case 2: strcpy(cmd,"set logdbg=0\n"); break;
    case 3: strcpy(cmd,"set filter=\n"); break;
    default: return -1;
  }
  return step++;
}
#define signoff(s) {\
  printf("\nThanks for using attachfec !\n");\
  close(csfd);\
  exit(0);\
}

#ifndef PIPE_PORT
# define PIPE_PORT 10503
#endif
     
int sndCommand(char *ipcfn)
{
  char buf[MAXLINES], pipeName[128], *c, *ip, *ctx, *srv;
  int len, cleanupFlag = 0,sendFlag = 0;
  static int csfd = -1;
  static struct timeval to;
  fd_set rset,allset;
  int n,maxfd,nread;
  int cnt = 0;
  UINT16 port;
  FecDataStruct *fec;

  if ((c=strchr(ipcfn,':')) != NULL)
  {
    *c++ = 0;
    ip = ipcfn;
    port = PIPE_PORT + atoi(c);
    if  (csfd < 0 && (csfd=sckCln(ip,port)) < 0)
    {
      printf("sckCln returned %d\n",csfd); 
      quit("CMD client error");
    }
  }
  else if (ipcfn[0] == '/')
  {
    ctx = &ipcfn[1];
    if ((c=strchr(ctx,'/')) != NULL)
    {
      *c++ = 0;
      srv = c;
      initClient();
      fec = GetFecDataFromCache(ctx,srv);
      if (fec == NULL)
      {
        printf("/%s/%s is not a registered server\n",ctx,srv);
        quit("CMD client error");        
      }
      ip = fec->IP;
      port = PIPE_PORT + fec->portOffset;
    }
    else
    {
      printf("%s is not in the form /<context>/<server>\n",ipcfn);
      quit("CMD client error");
    }
    if  (csfd < 0 && (csfd=sckCln(ip,port)) < 0)
    {
      printf("sckCln returned %d\n",csfd); 
      quit("CMD client error");
    }
  }
  else
  {
    sprintf(pipeName,"/tmp/%s.ipc",ipcfn);
    if (csfd < 0 && (csfd=pipeCln(pipeName)) < 0)
    { 
      printf("pipeCln returned %d\n",csfd); 
      quit("CMD client error");
    }
  }
  printf("Remote session established\n>");

  FD_ZERO(&allset);
  
  FD_SET(STDIN_FILENO,&allset);
  FD_SET(csfd,&allset);
  maxfd = csfd;

  for (;;)
  {
    rset = allset;
    to.tv_sec = 1; to.tv_usec = 0;
    if ((n=select(maxfd+1,&rset,NULL,NULL,&to)) < 0) quit("select error");
    if (FD_ISSET(csfd,&rset)) 
    {
      if ((nread = read(csfd,buf,MAXLINES)) <= 0) 
      {
        if (nread < 0) perror("read");
        else printf("server no longer responding\n");
        exit(1);
      }
      else
      {
        buf[nread] = 0;  
        printf("%s",buf);
      }
    }
    if (FD_ISSET(STDIN_FILENO,&rset))
    {
      if ((len=read(STDIN_FILENO,cmd,sizeof(cmd))) < 0) return -1;
      if (cmd[0] == '\n') { printf(">"); continue; }
      sendFlag = 1;
      if (!strncmp(cmd,"help",4))
      {
        printf("attachfec commands: \n>");
        printf("\tquit - exits the current attachfec session\n>");
        printf("\texit - exits the current attachfec session\n>");
        printf("\tkill - terminates the remote server process\n>");
        printf("\n>");
      }
      if (!strncmp(cmd,"quit",4)) cleanupFlag = 1;
      if (!strncmp(cmd,"exit",4)) cleanupFlag = 1;
      if (!strncmp(cmd,"kill",4)) cleanupFlag = 2;
    }
    switch (cleanupFlag)
    {
      case 0: break;
      case 1: sendFlag = 1;
              if (cleanup() < 0) cleanupFlag = 3; 
              break;
      case 2: sendFlag = 1;
              printf("Aborting Server Task\n");
              strcpy(cmd,"quit\n");
              cleanupFlag = 3;
              break;
      case 3: signoff(csfd);
              break;
    }
    if (sendFlag)
    {
      if (write(csfd,cmd,strlen(cmd)) < 0) quit("CMD write error");
      memset(cmd,0,256);
      sendFlag = 0;
    }
  }
  
  return 0;
}
    
