/*
 * File dblib.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file dblib.c
 *
 * \brief general database routines:
 *
 * dblib contains functions which can be used to access CVS files and 
 * extract fields.
 * 
 * It also contains some string and conversion utilities.
 * 
 */
#ifndef NO_DOXYGEN
#include "project.def"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#ifdef WIN32
# include "direct.h"
#endif
#if defined(UNIX) || defined(MACOS)
# include <sys/stat.h>
# include <dirent.h>
# ifndef FREEBSD
# include <alloca.h>
# endif
#endif
#include "prolog.h"
#include "srvcore.h"
#include "srvip.h"
#include "dbtools.h"
#include "hstlib.h"
#include "errors.h"
#include "ttycmd.h"

#ifndef SMALL_TINE_LIB

#define LOINT(k)       ((int)(k))
#define HIINT(k)       ((int)(((SINT32)(k) >> 16) & 0xFFFF))

DBLNKLIST *gFecInfoList = NULL;
int fecXmlDone = FALSE;
TINE_EXPORT int IgnoreFecXmlDatabase(void)
{
  if (fecXmlDone) return already_assigned;
  feclogEx(TINE_LOGLEVEL_INFO, "ignoring fec.xml database");
  fecXmlDone = TRUE;
  return 0;
}
#define MAX_STACK 32
static int dbStackPtr = 0;
void *dbStack[ MAX_STACK ];
int dbpush( void * p )
{
  if( dbStackPtr >= MAX_STACK ) return 0;
  dbStack[dbStackPtr++] = p;
  return 1;
}
int dbpop( void ** p )
{
  if( dbStackPtr <= 0 ) return 0;
  *p = dbStack[--dbStackPtr];
  return 1;
}
TINE_EXPORT DBLNKLIST * XMLListCreate( void )
{
  DBLNKLIST *p = (DBLNKLIST *)SystemCalloc(1,sizeof(DBLNKLIST));
  if( p == NULL ) return NULL;

  p->current = NULL;
  p->first = NULL;
  p->last = NULL;
  p->nElements = 0;

  return p;
}
TINE_EXPORT void XMLListDelete( DBLNKLIST *list )
{
  DBLNKLIST_ITEM *item;
  DBLNKLIST_ITEM *previousItem;
  if (list == NULL) return;
  item = list->last;  
  while( item )
  {
    previousItem = item->previous;
    if( item->dataBufferAllocated ) free( item->data ); 
    free( item );
    item = previousItem;
  }
  free( list );
}
TINE_EXPORT void XMLListAdd( DBLNKLIST * list, void * data )
{
  DBLNKLIST_ITEM * li = (DBLNKLIST_ITEM *)SystemCalloc(1,sizeof( DBLNKLIST_ITEM ));

  if (list == NULL)
  {
    feclogEx(TINE_LOGLEVEL_WARN, "XMLListAdd error: list not yet created");
    return;
  }
  if(li == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "XMLListAdd error: malloc failed");
    return;
  }
  li->dataBufferAllocated = 0;
  if( list->last != NULL ) 
  {
    li->previous = list->last;
    list->last->next =  li;		
  }
  else
  {
    list->first = li;
    li->previous = NULL;
  }
  li->next = NULL;
  li->data = data;
  list->current = li;
  list->last = li;
  ++list->nElements;
}
TINE_EXPORT void XMLListAddCopy( DBLNKLIST * list, void * data, int sizeOfData )
{
  DBLNKLIST_ITEM * li = (DBLNKLIST_ITEM *)SystemCalloc(1,sizeof(DBLNKLIST_ITEM));
  void * dataCopy;

  if( li == NULL )
  {
    puts( "XMLListAddCopy error: malloc failed" );
    return;
  }
  if( (dataCopy = SystemCalloc(1,sizeOfData)) == NULL )
  {
    puts( "XMLListAddCopy error: malloc failed" );
    return;
  }
  memcpy( dataCopy, data, sizeOfData );
  li->dataBufferAllocated = 1;
  if( list->last != NULL ) 
  {
    li->previous = list->last;
    list->last->next =  li;		
  }
  else
  {
    list->first = li;
    li->previous = NULL;
  }
  li->next = NULL;
  li->data = dataCopy;
  list->current = li;
  list->last = li;
  ++list->nElements;
}
TINE_EXPORT void XMLListRemoveCurrent( DBLNKLIST * list )
{
  if( list == NULL || list->current == NULL ) return;

  if( list->current->previous != NULL ) list->current->previous->next = list->current->next;
  else list->first = list->current->next;

  if( list->current->next != NULL ) list->current->next->previous = list->current->previous;
  else list->last = list->current->previous;

  if( list->current->dataBufferAllocated ) free( list->current->data );

  free( list->current );
  --list->nElements;
}
TINE_EXPORT void *XMLListGetNext( DBLNKLIST * list )
{
  void * data;
  if( list != NULL && list->current != NULL )
  {
    data = list->current->data;
    list->current = list->current->next;
    return data;
  }
  return NULL;
}
TINE_EXPORT void *XMLListGetPrevious( DBLNKLIST * list )
{
  void * data;
  if( list != NULL && list->current != NULL )
  {
    data = list->current->data;
    list->current = list->current->previous;
    return data;
  }
  return NULL;
}
TINE_EXPORT int XMLListHasNext( DBLNKLIST * list )
{
  if (list == NULL) return 0;
  return ( list->current->next != NULL );
}
TINE_EXPORT int XMLListHasPrevious( DBLNKLIST * list )
{
  if (list == NULL) return 0;
  return ( list->current->previous != NULL );
}
TINE_EXPORT void XMLListGoToFirst( DBLNKLIST * list )
{
  if (list != NULL) list->current = list->first;
}
TINE_EXPORT void XMLListGoToLast( DBLNKLIST * list )
{
  if (list != NULL) list->current = list->last;
}
void freeFecListMemory( DBLNKLIST *fecList )
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlDEVIS *device;
  xmlPRPIS *property;
  xmlSETIS *set;
  xmlAIS *alias;
  xmlADS *almdef;
  xmlHIS *hst;
  xmlAWIS *alm;
  NAME64 *member;

  if( fecList )
  {
    XMLListGoToFirst( fecList );
    while ((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL)
    {
      if( fec->eqms ) 
      {
        XMLListGoToFirst( fec->eqms );			
        while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL)
        {
          if( eqm->devices )
          {
            XMLListGoToFirst( eqm->devices );
            while((device=(xmlDEVIS*)XMLListGetNext(eqm->devices)) != NULL)
            {
              if (device->sets)
              {
                XMLListGoToFirst(device->sets);
                while((set=(xmlSETIS*)XMLListGetNext(device->sets)) != NULL)
                {
                  if (set->members)
                  {
                    XMLListGoToFirst(set->members);
                    while((member=(NAME64*)XMLListGetNext(set->members)) != NULL) SystemFree( member );
                    XMLListDelete(set->members);
                  }
                  SystemFree(set);
                }
                XMLListDelete(device->sets);
              }
              SystemFree( device );
            }
            XMLListDelete( eqm->devices );
          }
          if( eqm->properties )
          {
            XMLListGoToFirst( eqm->properties );
            while((property=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL) 
            {
              if (property->histories)
              {
                XMLListGoToFirst( property->histories );
                while((hst=(xmlHIS*)XMLListGetNext(property->histories)) != NULL) SystemFree( hst );
                XMLListDelete( property->histories );
              }
              if (property->almwatches)
              {
                XMLListGoToFirst( property->almwatches );
                while((alm=(xmlAWIS*)XMLListGetNext(property->almwatches)) != NULL) SystemFree( alm );
                XMLListDelete( property->almwatches );
              }
              if (property->sets)
              {
                XMLListGoToFirst(property->sets);
                while((set=(xmlSETIS*)XMLListGetNext(property->sets)) != NULL)
                {
                  if (set->members)
                  {
                    XMLListGoToFirst(set->members);
                    while((member=(NAME64*)XMLListGetNext(set->members)) != NULL) SystemFree( member );
                    XMLListDelete(set->members);
                  }
                  SystemFree(set);
                }
                XMLListDelete(property->sets);
              }
              SystemFree( property );
            }
            XMLListDelete( eqm->properties );
          }
          if( eqm->alarmdefs )
          {
            XMLListGoToFirst( eqm->alarmdefs );
            while((almdef=(xmlADS*)XMLListGetNext(eqm->alarmdefs)) != NULL) SystemFree( almdef );
            XMLListDelete( eqm->alarmdefs );
          }
          if (eqm->sets)
          {
            XMLListGoToFirst(eqm->sets);
            while((set=(xmlSETIS*)XMLListGetNext(eqm->sets)) != NULL)
            {
              if (set->members)
              {
                XMLListGoToFirst(set->members);
                while((member=(NAME64*)XMLListGetNext(set->members)) != NULL) SystemFree( member );
                XMLListDelete(set->members);
              }
              SystemFree(set);
            }
            XMLListDelete(eqm->sets);
          }
          SystemFree( eqm );
        }
        XMLListDelete( fec->eqms );
      }
      if (fec->alias)
      {
        XMLListGoToFirst(fec->alias);
        while((alias=(xmlAIS*)XMLListGetNext(fec->alias)) != NULL)
        {
          SystemFree(alias);
        }
        XMLListDelete(fec->alias);
      }
      if (fec->sets)
      {
        XMLListGoToFirst(fec->sets);
        while((set=(xmlSETIS*)XMLListGetNext(fec->sets)) != NULL)
        {
          if (set->members)
          {
            XMLListGoToFirst(set->members);
            while((member=(NAME64*)XMLListGetNext(set->members)) != NULL) SystemFree( member );
            XMLListDelete(set->members);
          }
          SystemFree(set);
        }
        XMLListDelete(fec->sets);
      }
      /* if (fec->almsettings) SystemFree(fec->almsettings); << freed in getAlmSettings() ! */
      SystemFree( fec );
    }
    XMLListDelete( fecList );
  }
}
XML_DBHANDLER *CurrentHandler = NULL;
TINE_EXPORT void XMLSetCurrentHandler( XML_DBHANDLER * xmlHandler )
{
  if( CurrentHandler != NULL ) dbpush( CurrentHandler );
  CurrentHandler = xmlHandler;
}
TINE_EXPORT void XMLRestoreHandler( void )
{
  XML_DBHANDLER * h;
  if( dbpop( (void ** )& h ) ) CurrentHandler = h;
}
static int documentProcessOpeningTag( char *tag )
{
	xmlFIDS * fec;
	
	if( !strcmp( tag, "FEC" ) ) 
	{	/* add new FEC to the list */
		fec = (xmlFIDS *)SystemCalloc(1,sizeof( xmlFIDS ));
    if (fec == NULL) return -1;
		fec->eqms = XMLListCreate(); 
    fec->sets = XMLListCreate();
    fec->alias = XMLListCreate();
    XMLListAdd( (DBLNKLIST*)CurrentHandler->currentDataSection, fec );  
    XMLSetCurrentHandler( &FecHandler );
    CurrentHandler->currentDataSection = fec;
    return 1;
  }
  else if( tag[0] == '?' ) return 1; /* skip header */

  dbglog( "documentProcessOpeningTag( %s ): unknown tag\n", tag );
  return 0;
}
static int documentProcessClosingTag(char *tag )
{
  UNUSED(tag);
  return 1;
}
static int fecProcessOpeningTag(char *tag )
{
  xmlXIS * eqm;
  xmlSETIS *set;
  xmlAIS *alias;
  ALMSETIS *asis;
	
  if( !strcmp( tag, "EQM" ) ) 
  { /* add new EQM to the list */
    eqm = (xmlXIS *)SystemCalloc(1,sizeof( xmlXIS ));
    if (eqm == NULL) return -1;
    eqm->devices = XMLListCreate(); 
    eqm->properties = XMLListCreate();
    eqm->alarmdefs = XMLListCreate();
    eqm->sets = XMLListCreate();
    XMLListAdd( ((xmlFIDS *)CurrentHandler->currentDataSection)->eqms, eqm );
    XMLSetCurrentHandler( &EqmHandler );
    CurrentHandler->currentDataSection = eqm;
    return 1;
  }
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = FEC_NAME_SIZE;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "PORT_OFFSET" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlFIDS *)(CurrentHandler->currentDataSection))->portOffset;
    return 1;
  }
  if( !strcmp( tag, "MTU" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlFIDS *)(CurrentHandler->currentDataSection))->fecMtu;
    return 1;
  }
  if( !strcmp( tag, "BURST_LIMIT" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlFIDS *)(CurrentHandler->currentDataSection))->fecBurstLimit;
    return 1;
  }
  if( !strcmp( tag, "CYCLE_DELAY" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlFIDS *)(CurrentHandler->currentDataSection))->fecCycleDelay;
    return 1;
  }
  if( !strcmp( tag, "WORKAREA_SIZE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlFIDS *)(CurrentHandler->currentDataSection))->fecWorkAreaSize;
    return 1;
  }
  if( !strcmp( tag, "HISTORY_HOME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = HISTORY_HOME_SIZE;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->historyHome;
    return 1;
  }
  if( !strcmp( tag, "NAME_LIST" ) ) 
  { 
    set = (xmlSETIS *)SystemCalloc(1,sizeof( xmlSETIS )); 
    if (set == NULL) return -1;
    XMLListAdd( ((xmlFIDS *)CurrentHandler->currentDataSection)->sets, set );
    XMLSetCurrentHandler( &NameSetHandler );
    CurrentHandler->currentDataSection = set;
    set->members = XMLListCreate(); 
    return 1;
  }
  if( !strcmp( tag, "ALIAS" ) ) 
  { 
    alias = (xmlAIS *)SystemCalloc(1,sizeof( xmlAIS )); 
    if (alias == NULL) return -1;
    XMLListAdd( ((xmlFIDS *)CurrentHandler->currentDataSection)->alias, alias );
    XMLSetCurrentHandler( &AliasHandler );
    CurrentHandler->currentDataSection = alias;
    return 1;
  }
  if( !strcmp( tag, "CONTEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = CONTEXT_NAME_SIZE;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->context;
    return 1;
  }
  if( !strcmp( tag, "SUBSYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = SUBSYSTEM_NAME_SIZE;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->subsystem;
    return 1;
  }
  if( !strcmp( tag, "ALARM_SETTINGS" ) ) 
  {
    asis = (ALMSETIS *)SystemCalloc(1,sizeof( ALMSETIS )); 
    if (asis == NULL) return -1;
    ((xmlFIDS *)(CurrentHandler->currentDataSection))->almsettings = asis;
    XMLSetCurrentHandler( &AlmSettingsHandler );
    CurrentHandler->currentDataSection = asis;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->almsettings;
    return 1;
  }

  dbglog( "fecProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int fecProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "FEC" ) ) XMLRestoreHandler();
  return 1;
}
static int eqmProcessOpeningTag( char *tag )
{
  xmlDEVIS *device;
  xmlPRPIS *property;
  xmlADS *almdefs;
  xmlSETIS *set;

  if( !strcmp( tag, "ALARM_DEFINITION" ) ) 
  { /* add new alarm definition to the list */
    almdefs = (xmlADS *)SystemCalloc(1,sizeof( xmlADS )); 
    if (almdefs == NULL) return -1;
    XMLListAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->alarmdefs, almdefs );
    XMLSetCurrentHandler( &AlmDefHandler );
    CurrentHandler->currentDataSection = almdefs;
    return 1;
  }
  if( !strcmp( tag, "NAME_LIST" ) ) 
  { /* add a new name list to the list */
    set = (xmlSETIS *)SystemCalloc(1,sizeof( xmlSETIS )); 
    if (set == NULL) return -1;
    XMLListAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->sets, set );
    XMLSetCurrentHandler( &NameSetHandler );
    CurrentHandler->currentDataSection = set;
    set->members = XMLListCreate(); 
    return 1;
  }
  if( !strcmp( tag, "DEVICE" ) ) 
  { /* add new (common) DEVICE to the list */
    device = (xmlDEVIS *)SystemCalloc(1,sizeof( xmlDEVIS )); 
    if (device == NULL) return -1;
    device->sets = XMLListCreate();
    XMLListAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->devices, device );
    XMLSetCurrentHandler( &DeviceHandler );
    CurrentHandler->currentDataSection = device;
    return 1;
  }
  if( !strcmp( tag, "DEVICE_SPACE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlXIS*)(CurrentHandler->currentDataSection))->deviceSpace;
    return 1;
  }
  if( !strcmp( tag, "REGION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = TAG_NAME_SIZE;
    CurrentHandler->currentData = &((xmlXIS*)(CurrentHandler->currentDataSection))->region;
    return 1;
  }
  if( !strcmp( tag, "PROPERTY" ) ) 
  { /* add new (common) PROPERTY to the list */
    property = (xmlPRPIS *)SystemCalloc(1,sizeof( xmlPRPIS )); 
    if (property == NULL) return -1;
    property->id = -1; /* default : signal assigning via current number of properties */
    property->histories = XMLListCreate();
    property->almwatches = XMLListCreate();
    property->sets = XMLListCreate();
    XMLListAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->properties, property );
    XMLSetCurrentHandler( &PropertyHandler );
    CurrentHandler->currentDataSection = property;

    return 1;
  }
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = EQM_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "SERVER" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = EXPORT_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverName;
    return 1;
  }
  if( !strcmp( tag, "GROUP" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = EXPORT_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverGroup;
    return 1;
  }
  if( !strcmp( tag, "GROUP_DEVICE_PREFIX" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = XFIX_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverGroupDevPrefix;
    return 1;
  }
  if( !strcmp( tag, "GROUP_DEVICE_POSTFIX" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = XFIX_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverGroupDevPostfix;
    return 1;
  }
  if( !strcmp( tag, "GROUP_INDEX" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlXIS*)(CurrentHandler->currentDataSection))->serverGroupIndex;
    return 1;
  }
  if( !strcmp( tag, "CONTEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = CONTEXT_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->context;
    return 1;
  }
  if( !strcmp( tag, "SUBSYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = SUBSYSTEM_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->subsystem;
    return 1;
  }
  if( !strcmp( tag, "MASTER" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = EXPORT_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverMaster;
    return 1;
  }
  if( !strcmp( tag, "SLAVE_MASTER" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = EXPORT_NAME_SIZE;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->slaveMaster;
    return 1;
  }
  dbglog( "eqmProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int eqmProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "EQM" ) )  XMLRestoreHandler();
  return 1;
}
static int deviceProcessOpeningTag( char *tag )
{
  xmlSETIS *set;
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_NAME_SIZE;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "PROPERTY_SET" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = SET_NAME_SIZE;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->propertySet;
    return 1;
  }
  if( !strcmp( tag, "NAME_LIST" ) ) 
  { 
    set = (xmlSETIS *)SystemCalloc(1,sizeof( xmlSETIS )); 
    if (set == NULL) return -1;
    XMLListAdd( ((xmlDEVIS *)CurrentHandler->currentDataSection)->sets, set );
    XMLSetCurrentHandler( &NameSetHandler );
    CurrentHandler->currentDataSection = set;
    set->members = XMLListCreate(); 
    return 1;
  }
  if( !strcmp( tag, "NUMBER" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlDEVIS*)(CurrentHandler->currentDataSection))->number;
    return 1;
  }
  if( !strcmp( tag, "MASK" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlDEVIS*)(CurrentHandler->currentDataSection))->mask;
    return 1;
  }
  if( !strcmp( tag, "OFFLINE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlDEVIS*)(CurrentHandler->currentDataSection))->offline;
    return 1;
  }
  if( !strcmp( tag, "REGION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = TAG_NAME_SIZE;
    CurrentHandler->currentData = &((xmlDEVIS*)(CurrentHandler->currentDataSection))->region;
    return 1;
  }
  if( !strcmp( tag, "ZPOS" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlDEVIS*)(CurrentHandler->currentDataSection))->zpos;
    return 1;
  }
  if( !strcmp( tag, "REDIRECTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_REDIR_SIZE;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->redirection;
    return 1;
  }
  if( !strcmp( tag, "DESCRIPTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_DESC_SIZE;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->desc;
    return 1;
  }
  if( !strcmp( tag, "LOCATION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_LOCATION_SIZE;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->loc;
    return 1;
  }
  dbglog( "deviceProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int deviceProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "DEVICE" ) )  XMLRestoreHandler();
  return 1;
}
static int defaultProcessData( char *data )
{
  switch( CurrentHandler->currentDataType )
  {
    case CF_TEXT:
      if ((int)strlen(data) > CurrentHandler->currentDataCapacity)
      {
        feclog("truncating xml string data %s to %d characters!",data, CurrentHandler->currentDataCapacity);
      }
      strncpy((char *)CurrentHandler->currentData, data, CurrentHandler->currentDataCapacity);
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_INT32:
      *((SINT32 *)(CurrentHandler->currentData)) = atoi( data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_FLOAT:
      *((float *)(CurrentHandler->currentData)) = (float)atof( data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_DOUBLE:
      *((double*)(CurrentHandler->currentData)) = atof( data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_NULL:
      return 1;
      break;
    default:
      dbglog( "defaultProcessData(%s) : unknown data type: %d", data, CurrentHandler->currentDataType ); 
      return 0;
      break;
  }
  return 1;
}
static int almdefProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "TAG" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 32;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->almTag;
    return 1;
  }
  if( !strcmp( tag, "DATA_FORMAT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->dataFormat;
    return 1;
  }
  if( !strcmp( tag, "ALARM_TEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->almText;
    return 1;
  }
  if( !strcmp( tag, "DEVICE_TEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->devText;
    return 1;
  }
  if( !strcmp( tag, "DATA_TEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->datText;
    return 1;
  }
  if( !strcmp( tag, "URL" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->urlText;
    return 1;
  }
  if( !strcmp( tag, "ALARM_CODE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->code;
    return 1;
  }
  if( !strcmp( tag, "ALARM_MASK" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->mask;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->severity;
    return 1;
  }
  if( !strcmp( tag, "ALARM_SYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->system;
    return 1;
  }
  if( !strcmp( tag, "DATA_SIZE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->dataSize;
    return 1;
  }
  dbglog( "almdefProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int almdefProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALARM_DEFINITION" ) )  XMLRestoreHandler();
  return 1;
}
static int historyProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "DEVICE_NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_NAME_SIZE;
    CurrentHandler->currentData = ((xmlHIS*)(CurrentHandler->currentDataSection))->hstDevName;
    return 1;
  }
  if( !strcmp( tag, "TOLERANCE" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlHIS*)(CurrentHandler->currentDataSection))->hstTolerance;
    return 1;
  }
  if( !strcmp( tag, "REDIRECTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = PROPERTY_REDIR_SIZE;
    CurrentHandler->currentData = ((xmlHIS*)(CurrentHandler->currentDataSection))->hstRemoteServer;
    return 1;
  }
  if( !strcmp( tag, "INDEX" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstIndex;
    return 1;
  }
  if( !strcmp( tag, "HEARTBEAT" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstHeartbeat;
    return 1;
  }
  if( !strcmp( tag, "POLL_RATE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstPollRate;
    return 1;
  }
  if( !strcmp( tag, "ARCHIVE_RATE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstArchRate;
    return 1;
  }
  if( !strcmp( tag, "DEPTH_SHORT" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstDepthShort;
    return 1;
  }
  if( !strcmp( tag, "DEPTH_LONG" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstDepthLong;
    return 1;
  }
  if( !strcmp( tag, "FILTER" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 256;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstFilterString;
    return 1;
  }

  dbglog( "historyProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int alarmProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "DEVICE_NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_NAME_SIZE;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->almDevName;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_HIGH" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevHigh;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_LOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevLow;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_HIGH_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevHighWarn;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_LOW_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevLowWarn;
    return 1;
  }
  if( !strcmp( tag, "SYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSystem;
    return 1;
  }
  if( !strcmp( tag, "VALUE_MASK" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValMask;
    return 1;
  }
  if( !strcmp( tag, "VALUE_NORMAL" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValNormal;
    return 1;
  }
  if( !strcmp( tag, "COUNT_THRESHOLD" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almCntThreshold;
    return 1;
  }
  if( !strcmp( tag, "VALUE_HIGH" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValHigh;
    return 1;
  }
  if( !strcmp( tag, "VALUE_LOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValLow;
    return 1;
  }
  if( !strcmp( tag, "VALUE_HIGH_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValHighWarn;
    return 1;
  }
  if( !strcmp( tag, "VALUE_LOW_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValLowWarn;
    return 1;
  }
  if( !strcmp( tag, "ALARM_CODE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->code;
    return 1;
  }
  if( !strcmp( tag, "ALARM_CODE_HIGH" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->codeHigh;
    return 1;
  }
  if (!strcmp(tag, "ALARM_CODE_HIGHWARN"))
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->codeHighWarn;
    return 1;
  }
  if( !strcmp( tag, "ALARM_CODE_LOW" ) )
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->codeLow;
    return 1;
  }
  if (!strcmp(tag, "ALARM_CODE_LOWWARN"))
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->codeLowWarn;
    return 1;
  }
  if( !strcmp( tag, "ALARM_TAG_NORMAL" ) )
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = ALARM_TAG_SIZE;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->tagNormal;
    return 1;
  }
  if( !strcmp( tag, "ALARM_TAG_HIGH" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = ALARM_TAG_SIZE;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->tagHigh;
    return 1;
  }
  if (!strcmp(tag, "ALARM_TAG_HIGHWARN"))
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = ALARM_TAG_SIZE;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->tagHighWarn;
    return 1;
  }
  if( !strcmp( tag, "ALARM_TAG_LOW" ) )
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = ALARM_TAG_SIZE;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->tagLow;
    return 1;
  }
  if (!strcmp(tag, "ALARM_TAG_LOWWARN"))
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = ALARM_TAG_SIZE;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->tagLowWarn;
    return 1;
  }
  if( !strcmp( tag, "FILTER" ) )
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 256;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->almFilterString;
    return 1;
  }

  dbglog( "alarmProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int propertyProcessOpeningTag( char *tag )
{
  xmlHIS *history;
  xmlAWIS *alarm;
  xmlSETIS *set;

  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = PROPERTY_NAME_SIZE;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "DEVICE_SET" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = SET_NAME_SIZE;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->deviceSet;
    return 1;
  }
  if( !strcmp( tag, "NAME_LIST" ) ) 
  { 
    set = (xmlSETIS *)SystemCalloc(1,sizeof( xmlSETIS )); 
    if (set == NULL) return -1;
    XMLListAdd( ((xmlPRPIS *)CurrentHandler->currentDataSection)->sets, set );
    XMLSetCurrentHandler( &NameSetHandler );
    CurrentHandler->currentDataSection = set;
    set->members = XMLListCreate(); 
    return 1;
  }
  if( !strcmp( tag, "DESCRIPTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 128;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->description;
    return 1;
  }
  if( !strcmp( tag, "EGU" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 16;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->yEgu;
    return 1;
  }
  if( !strcmp( tag, "XEGU" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 16;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->xEgu;
    return 1;
  }
  if( !strcmp( tag, "MAX" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->yMax;
    return 1;
  }
  if( !strcmp( tag, "MIN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->yMin;
    return 1;
  }
  if( !strcmp( tag, "XMAX" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->xMax;
    return 1;
  }
  if( !strcmp( tag, "XMIN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->xMin;
    return 1;
  }
  if( !strcmp( tag, "REDIRECTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = PROPERTY_REDIR_SIZE;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->redirection;
    return 1;
  }
  if( !strcmp( tag, "SIZE_IN" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataInSize;
    return 1;
  }
  if( !strcmp( tag, "SIZE_OUT" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataOutSize;
    return 1;
  }
  if( !strcmp( tag, "DTYPE_IN" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataInType;
    return 1;
  }
  if( !strcmp( tag, "DTYPE_OUT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataOutType;
    return 1;
  }
  if( !strcmp( tag, "ACCESS" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = ACCESS_STR_SIZE;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->access;
    return 1;
  }
  if( !strcmp( tag, "ID" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->id;
    return 1;
  }
  if( !strcmp( tag, "ACCESS_DEADBAND" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->deadband;
    return 1;
  }
  /* alarm watch entries : */
  if( !strcmp( tag, "ALARM" ) ) 
  {
    alarm = (xmlAWIS *)SystemCalloc(1,sizeof( xmlAWIS )); 
    if (alarm == NULL) return -1;
    XMLListAdd( ((xmlPRPIS *)CurrentHandler->currentDataSection)->almwatches, alarm );
    XMLSetCurrentHandler( &AlmWatchHandler );
    CurrentHandler->currentDataSection = alarm;
    return 1;
  }
  /* history entries : */
  if( !strcmp( tag, "HISTORY" ) ) 
  { /* add new PROPERTY HISTORY to the list */
    history = (xmlHIS *)SystemCalloc(1,sizeof( xmlHIS )); 
    if (history == NULL) return -1;
    XMLListAdd( ((xmlPRPIS *)CurrentHandler->currentDataSection)->histories, history );
    XMLSetCurrentHandler( &HistoryHandler );
    CurrentHandler->currentDataSection = history;
    return 1;
  }

  dbglog( "propertyProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int historyProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "HISTORY" ) )  XMLRestoreHandler();
  return 1;
}
static int alarmProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALARM" ) )  XMLRestoreHandler();
  return 1;
}
static int propertyProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "PROPERTY" ) )  XMLRestoreHandler();
  return 1;
}
static int memberProcessOpeningTag( char *tag )
{
  UNUSED(tag);
  return 1;	
}
static int memberProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "MEMBER" ) )  XMLRestoreHandler();
  return 1;
}
static int namesetProcessOpeningTag( char *tag )
{
  NAME64 *member;
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = SET_NAME_SIZE;
    CurrentHandler->currentData = ((xmlSETIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "MEMBER" ) ) 
  { 
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = 64;
    member = (NAME64 *)SystemCalloc(1,sizeof( NAME64 )); 
    if (member == NULL) return -1;
    memcpy(member->name,CurrentHandler->currentDataSection,64);
    XMLListAdd( ((xmlSETIS *)CurrentHandler->currentDataSection)->members, member );
    CurrentHandler->currentData = member;
    return 1;
  }
  dbglog( "namesetProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int namesetProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "NAME_LIST" ) )  XMLRestoreHandler();
  return 1;
}
static int aliasProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_NAME_SIZE;
    CurrentHandler->currentData = ((xmlAIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "TARGET" ) ) 
  { 
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentDataCapacity = DEVICE_NAME_SIZE;
    CurrentHandler->currentData = ((xmlAIS*)(CurrentHandler->currentDataSection))->target;
    return 1;
  }
  dbglog( "aliasProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int aliasProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALIAS" ) )  XMLRestoreHandler();
  return 1;
}
static int almsettingsProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "MINIMUM_WINDOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->wnd;
    return 1;
  }
  if( !strcmp( tag, "TRANSIENT_WINDOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->trnWnd;
    return 1;
  }
  if( !strcmp( tag, "DATACHANGE_WINDOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->dcWnd;
    return 1;
  }
  if( !strcmp( tag, "TERMINATE_WINDOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->trmWnd;
    return 1;
  }
  if( !strcmp( tag, "OSCILLATION_WINDOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->oscWnd;
    return 1;
  }
  if( !strcmp( tag, "COLLAPSE_WINDOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->clpWnd;
    return 1;
  }
  if( !strcmp( tag, "HEARTBEAT" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->hbt;
    return 1;
  }
  if( !strcmp( tag, "TABLE_SIZE" ) ) 
  {
    CurrentHandler->currentDataType = CF_INT32;
    CurrentHandler->currentDataCapacity = 1;
    CurrentHandler->currentData = &((ALMSETIS *)(CurrentHandler->currentDataSection))->tblSiz;
    return 1;
  }
  dbglog( "almsettingsProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int almsettingsProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALARM_SETTINGS" ) )  XMLRestoreHandler();
  return 1;
}

XML_DBHANDLER DocumentHandler = { documentProcessOpeningTag, documentProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER FecHandler = { fecProcessOpeningTag, fecProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER EqmHandler = { eqmProcessOpeningTag, eqmProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER DeviceHandler = { deviceProcessOpeningTag, deviceProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER PropertyHandler = { propertyProcessOpeningTag, propertyProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER AlmDefHandler = { almdefProcessOpeningTag, almdefProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER NameSetHandler = { namesetProcessOpeningTag, namesetProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER MemberHandler = { memberProcessOpeningTag, memberProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER AliasHandler = { aliasProcessOpeningTag, aliasProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER HistoryHandler = { historyProcessOpeningTag, historyProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER AlmWatchHandler = { alarmProcessOpeningTag, alarmProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };
XML_DBHANDLER AlmSettingsHandler = { almsettingsProcessOpeningTag, almsettingsProcessClosingTag, defaultProcessData, NULL, CF_NULL, 0, NULL };

int populateAlmSettings(char *fecname,DBLNKLIST *fecList,ALMSETIS **asis)
{
  xmlFIDS *fec;
  if (fecList == NULL || asis == NULL) return argument_list_error;
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS *)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      *asis = fec->almsettings;
      break;
    }
  }
  return 0;
}
int populateADS(char *fecname,char *eqmname,DBLNKLIST *fecList,ADSIS **adsis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlADS *ads;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || adsis == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        if (eqm->alarmdefs == NULL) continue;
        /* check for common devices */
        XMLListGoToFirst(eqm->alarmdefs);
        while((ads=(xmlADS*)XMLListGetNext(eqm->alarmdefs)) != NULL) lstsiz++;
        if ((*adsis=(ADSIS *)SystemCalloc(lstsiz,sizeof(ADSIS))) == NULL) ccerr(out_of_server_memory);
        for (i=0,XMLListGoToFirst(eqm->alarmdefs); i<lstsiz && (ads=(xmlADS *)XMLListGetNext(eqm->alarmdefs)) != NULL; i++)
        { 
          strncpy((*adsis)[i].alarmTag,ads->almTag,ALARM_TAG_SIZE);
          strncpy((*adsis)[i].alarmText,ads->almText,ALARM_TEXT_SIZE);
          strncpy((*adsis)[i].deviceText,ads->devText,ALARM_TEXT_SIZE);
          strncpy((*adsis)[i].dataText,ads->datText,ALARM_TEXT_SIZE);
          strncpy((*adsis)[i].url,ads->urlText,ALARM_TEXT_LONGSIZE);
          strncpy((*adsis)[i].alarmDataFormatString,ads->dataFormat,64);
          (*adsis)[i].alarmCode = ads->code;
          (*adsis)[i].alarmMask  = ads->mask;
          (*adsis)[i].alarmDataArraySize = ads->dataSize;
          (*adsis)[i].alarmSeverity = ads->severity;
          (*adsis)[i].alarmSystem = ads->system;
        }
      }
      break;
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
int populateAWS(char *fecname,DBLNKLIST *fecList,AWSIS **awsis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlAWIS *alm;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || awsis == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
	while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one, now start counting */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* this is a general list over all embedded eqms */
        XMLListGoToFirst(eqm->properties);
        while((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL) 
        { /* if there are alarm watches associated with the property, count them */
          XMLListGoToFirst(prp->almwatches);
          while ((alm=(xmlAWIS*)XMLListGetNext(prp->almwatches)) != NULL) lstsiz++;
        }
      }
      if ((*awsis=(AWSIS *)SystemCalloc(lstsiz,sizeof(AWSIS))) == NULL) ccerr(out_of_server_memory);
      /* now we know how many, so back up */
      i = 0;
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { 
        XMLListGoToFirst(eqm->properties);
        while((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL) 
        { 
          XMLListGoToFirst(prp->almwatches);
          while ((alm=(xmlAWIS*)XMLListGetNext(prp->almwatches)) != NULL)
          {
            if (i >= lstsiz) break;
            strncpy((*awsis)[i].dev,alm->almDevName,DEVICE_NAME_SIZE); /* TODO: allow sets and '*' */
            strncpy((*awsis)[i].alarmDataFormatString,prp->dataOutType,64);
            strncpy((*awsis)[i].prp,prp->name,PROPERTY_NAME_SIZE);
            strncpy((*awsis)[i].eqm,eqm->name,EQM_NAME_SIZE);
            (*awsis)[i].siz = prp->dataOutSize;
            (*awsis)[i].hisev = alm->almSevHigh;
            (*awsis)[i].hiwarnsev = alm->almSevHighWarn;
            (*awsis)[i].losev = alm->almSevLow;
            (*awsis)[i].lowarnsev = alm->almSevLowWarn;
            (*awsis)[i].cntThreshold = alm->almCntThreshold;
            (*awsis)[i].lo = alm->almValLow;
            (*awsis)[i].lowarn = alm->almValLowWarn;
            (*awsis)[i].hi =  alm->almValHigh;
            (*awsis)[i].hiwarn = alm->almValHighWarn;
            (*awsis)[i].asys = alm->almSystem;
            strncpy((*awsis)[i].normalValueString,alm->almValNormal,64);
            (*awsis)[i].mask = alm->almValMask;
            (*awsis)[i].code = alm->code;
            (*awsis)[i].codeHigh = alm->codeHigh;
            (*awsis)[i].codeHighWarn = alm->codeHighWarn;
            (*awsis)[i].codeLow = alm->codeLow;
            (*awsis)[i].codeLowWarn = alm->codeLowWarn;
            strncpy((*awsis)[i].tagNormal,alm->tagNormal,ALARM_TAG_SIZE);
            strncpy((*awsis)[i].tagHigh,alm->tagHigh,ALARM_TAG_SIZE);
            strncpy((*awsis)[i].tagHighWarn, alm->tagHighWarn, ALARM_TAG_SIZE);
            strncpy((*awsis)[i].tagLow,alm->tagLow,ALARM_TAG_SIZE);
            strncpy((*awsis)[i].tagLowWarn, alm->tagLowWarn, ALARM_TAG_SIZE);
            strncpy((*awsis)[i].filterString,alm->almFilterString,256);
            i++;
          }
        }
      }
      break;
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
int populateHRS(char *fecname,DBLNKLIST *fecList,HRIS **hris)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlHIS *hst;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || hris == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
	while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one, now start counting */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* this is a general list over all embedded eqms */
        XMLListGoToFirst(eqm->properties);
        while((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL) 
        { /* if there are histories associated with the property, count them */
          XMLListGoToFirst(prp->histories);
          while ((hst=(xmlHIS*)XMLListGetNext(prp->histories)) != NULL) lstsiz++;
        }
      }
      if ((*hris=(HRIS *)SystemCalloc(lstsiz,sizeof(HRIS))) == NULL) ccerr(out_of_server_memory);
      /* now we know how many, so back up */
      i = 0;
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { 
        XMLListGoToFirst(eqm->properties);
        while((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL) 
        { /* if there are histories associated with the property, count them */
          XMLListGoToFirst(prp->histories);
          while ((hst=(xmlHIS*)XMLListGetNext(prp->histories)) != NULL) 
          {
            if (i >= lstsiz) break;
            strncpy((*hris)[i].c.EqmDeviceName,hst->hstDevName,DEVICE_NAME_SIZE); /* TODO: allow sets and '*' */
            strncpy((*hris)[i].c.EqmProperty,prp->name,PROPERTY_NAME_SIZE); 
            strncpy((*hris)[i].c.EqmName,eqm->name,EQM_NAME_SIZE);
            (*hris)[i].c.EqmSizeOut = prp->dataOutSize;
            (*hris)[i].c.EqmFormatOut = BFMT(ftoi(prp->dataOutType));
            strncpy((*hris)[i].fmt,prp->dataOutType,64);
            strncpy((*hris)[i].redirection,hst->hstRemoteServer,PROPERTY_REDIR_SIZE);
            strncpy((*hris)[i].tolerance,hst->hstTolerance,64);
            (*hris)[i].pollingRate = hst->hstPollRate;
            (*hris)[i].archiveRate = hst->hstArchRate;
            (*hris)[i].depthShort = hst->hstDepthShort;
            strncpy((*hris)[i].depthLong,hst->hstDepthLong,64);
            (*hris)[i].heartbeat = hst->hstHeartbeat;
            (*hris)[i].recordIndex = hst->hstIndex;
            strncpy((*hris)[i].filterString,hst->hstFilterString,256);
            i++;
          }
        }
      }
      break;
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
int populateFIDS(DBLNKLIST *fecList,FIDS **fids)
{
	xmlFIDS *fec;
  int i, cc = 0, lstsiz = 0;
  if (fecList == NULL || fids == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while(XMLListGetNext(fecList) != NULL) lstsiz++;
  if ((*fids=(FIDS *)SystemCalloc(lstsiz,sizeof(FIDS))) == NULL) ccerr(out_of_server_memory);
  for (i=0,XMLListGoToFirst(fecList); i<lstsiz && (fec=(xmlFIDS *)XMLListGetNext(fecList)) != NULL; i++)
  {
    strncpy((*fids)[i].fecName,fec->name,FEC_NAME_SIZE);
    (*fids)[i].fecPort = fec->portOffset;
    strncpy((*fids)[i].fecContext,fec->context,CONTEXT_NAME_SIZE);
    strncpy((*fids)[i].fecDescription,fec->description,FEC_DESC_SIZE);
    strncpy((*fids)[i].fecHardware,fec->hardware,FEC_HDW_SIZE);
    strncpy((*fids)[i].fecLocation,fec->location,FEC_LOCATION_SIZE);
    strncpy((*fids)[i].fecSubsystem,fec->subsystem,SUBSYSTEM_NAME_SIZE);
    strncpy((*fids)[i].historyHome,fec->historyHome,HISTORY_HOME_SIZE);
    (*fids)[i].fecMtu = fec->fecMtu;
    (*fids)[i].fecBurstLimit = fec->fecBurstLimit;
    (*fids)[i].fecCycleDelay = fec->fecCycleDelay;
    (*fids)[i].fecWorkAreaSize = fec->fecWorkAreaSize;
  }
err:
  if (cc) feclog("populate Fec Info Database from XML : %.32s",cc2str(cc));
  return lstsiz;
}
void xmltoxis(XIS *xis,xmlXIS *eqm,xmlPRPIS *prp)
{
  char xtag[32],ytag[32];
  strncpy(xis->EqmContext,eqm->context,CONTEXT_NAME_SIZE);
  xis->DevSpace = eqm->deviceSpace;
  strncpy(xis->region,eqm->region,TAG_NAME_SIZE);
  strncpy(xis->EqmGroup,eqm->serverGroup,EXPORT_NAME_SIZE);
  strncpy(xis->EqmGroupDevPrefix,eqm->serverGroupDevPrefix,XFIX_SIZE);
  strncpy(xis->EqmGroupDevPostfix,eqm->serverGroupDevPostfix,XFIX_SIZE);
  xis->EqmGroupIdx = eqm->serverGroupIndex;
  strncpy(xis->EqmName,eqm->name,EQM_NAME_SIZE);
  strncpy(xis->ExpName,eqm->serverName,EXPORT_NAME_SIZE);
  strncpy(xis->SubSystem,eqm->subsystem,SUBSYSTEM_NAME_SIZE);
  strncpy(xis->EqmMaster,eqm->serverMaster,EXPORT_NAME_SIZE);
  strncpy(xis->EqmSlaveMaster,eqm->slaveMaster,EXPORT_NAME_SIZE);
  strncpy(xis->access,prp->access,ACCESS_STR_SIZE);
  strncpy(xis->fmtIn,prp->dataInType,FORMAT_STR_SIZE);
  strncpy(xis->fmtOut,prp->dataOutType,FORMAT_STR_SIZE);
  xis->PrpId = prp->id;
  xis->deadband = (int)prp->deadband;
  strncpy(xis->PrpName,prp->name,PROPERTY_NAME_SIZE);
  xis->PrpSizeIn = prp->dataInSize;
  xis->PrpSizeOut = prp->dataOutSize;
  strncpy(xis->units,prp->yEgu,UNITS_SIZE);
  strncpy(xis->xunits,prp->xEgu,UNITS_SIZE);
  xis->max = prp->yMax;
  xis->min = prp->yMin;
  xis->xmax = prp->xMax;
  xis->xmin = prp->xMin;
  strncpy(xis->redirection,prp->redirection,PROPERTY_REDIR_SIZE);
  if (strchr(prp->description,'[') != NULL && strchr(prp->description,']') != NULL)
  { /* take is as is */
    strncpy(xis->description,prp->description,128);
  }
  else
  { /* prepare it */
    sprintf(ytag,"[%.2g:%.2g %.8s]",prp->yMin,prp->yMax,prp->yEgu);
    xtag[0] = 0;
    if (prp->xMin != 0 || prp->yMin != 0) sprintf(xtag,"[%.2g:%.2g %.8s]",prp->xMin,prp->xMax,prp->xEgu);
    sprintf(xis->description,"%.32s%.32s%.64s",ytag,xtag,prp->description);
  }
}
int populateXIS(char *fecname,char *eqmname,DBLNKLIST *fecList,XIS **xis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || xis == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (eqmname == NULL)
        { /* no target name (TODO: fix so that the entire xml tree can work if more that 1 element) */
          eqmname = "(null)";
          if (fec->eqms->nElements != 1) ccerr(not_applicable);
        }
        else 
        { /* target eqm name given */
          if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        }
        /* check for common properties */
        XMLListGoToFirst(eqm->properties);
        while((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL) lstsiz++;
        if ((*xis=(XIS *)SystemCalloc(lstsiz,sizeof(XIS))) == NULL) ccerr(out_of_server_memory);
        for (i=0,XMLListGoToFirst(eqm->properties); i<lstsiz && (prp=(xmlPRPIS *)XMLListGetNext(eqm->properties)) != NULL; i++)
        { /* run through eqm->properties for the common properties */
          xmltoxis(&(*xis)[i],eqm,prp);
        }
        ccerr(0); /* done, break out */
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML %.16s %.6s: %.32s",fecname,eqmname,cc2str(cc));
  return cc ? -cc : lstsiz;
}
int _popNameLst(DBLNKLIST *list,char *listName,NAME64 **lst)
{ /* used locally: no NULL arguments ! */
  xmlSETIS *set;
  NAME64 *member;
  int i, cc = 0, lstsiz = 0;
  XMLListGoToFirst(list);
  while ((set=(xmlSETIS*)XMLListGetNext(list)) != NULL)
  { /* check its sets : */
    if (strnicmp(set->name,listName,SET_NAME_SIZE)) continue;
    /* that's the set !*/
    lstsiz = set->members->nElements;
    XMLListGoToFirst(set->members);
    if ((*lst=(NAME64 *)SystemCalloc(lstsiz,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    for (i=0,XMLListGoToFirst(set->members); i<lstsiz && (member=(NAME64 *)XMLListGetNext(set->members)) != NULL; i++)
    { /* run through eqm->devices for the common devices */
      strncpy((*lst)[i].name,member->name,64);
    }
    break;
  }
err:
  if (cc) feclog("populate %s List from XML : %.32s",listName,cc2str(cc));
  return lstsiz;
}
int populateNameList(DBLNKLIST *fecList,char *fecName,char *eqmName,char *devName,char *prpName,char *listName,NAME64 **lst)
{
	xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlDEVIS *dev;
  int cc = 0, lstsiz = 0;
  int hasEqmTgt=FALSE,hasDevTgt=FALSE,hasPrpTgt=FALSE;

  if (fecList == NULL || fecName == NULL || listName == NULL || lst == NULL) ccerr(argument_list_error);
  if (eqmName != NULL && strlen(eqmName) > 0) hasEqmTgt = TRUE;
  if (devName != NULL && strlen(devName) > 0) hasDevTgt = TRUE;
  if (prpName != NULL && strlen(prpName) > 0) hasPrpTgt = TRUE;
  for (XMLListGoToFirst(fecList);(fec=(xmlFIDS *)XMLListGetNext(fecList)) != NULL;)
  { /* zero in on the FEC : */
    if (strnicmp(fecName,fec->name,FEC_NAME_SIZE) != 0) continue;
    if (!hasEqmTgt)    
    { /* looking for a FEC level name list */
      if (fec->sets == NULL) return 0;
      lstsiz = _popNameLst(fec->sets,listName,lst);
    }
    else
    { /* need to find the target module */
      XMLListGoToFirst(fec->eqms);
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* scan thru the list */
        if (strncmp(eqm->name,eqmName,EQM_NAME_SIZE)) continue;
        if (hasDevTgt)
        { /* looking for a device */
          XMLListGoToFirst(eqm->devices);
          while((dev=(xmlDEVIS*)XMLListGetNext(eqm->devices)) != NULL)
          { /* scan thru the list */
            if (strncmp(dev->name,devName,DEVICE_NAME_SIZE)) continue;
            if (dev->sets == NULL) return 0;
            lstsiz = _popNameLst(dev->sets,listName,lst);
            break;
          }
        }
        else if (hasPrpTgt)
        { /* looking for a property */
          XMLListGoToFirst(eqm->properties);
          while((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL)
          { /* scan thru the list */
            if (strncmp(prp->name,prpName,PROPERTY_NAME_SIZE)) continue;
            if (prp->sets == NULL) return 0;
            lstsiz = _popNameLst(prp->sets,listName,lst);
            break;
          }
        }
        else
        { /* apply at the equipment module level */
          if (eqm->sets == NULL) return 0;
          lstsiz = _popNameLst(eqm->sets,listName,lst);
        }
        break;
      }
    }
    break;
  }
err:
  if (cc) feclog("populate name list from XML : %.32s",cc2str(cc));
  return lstsiz;
}
int populateDEVIS(char *fecname,char *eqmname,DBLNKLIST *fecList,DEVIS **devis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlDEVIS *dev;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || devis == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        /* check for common devices */
        XMLListGoToFirst(eqm->devices);
        while((dev=(xmlDEVIS*)XMLListGetNext(eqm->devices)) != NULL) lstsiz++;
        if ((*devis=(DEVIS *)SystemCalloc(lstsiz,sizeof(DEVIS))) == NULL) ccerr(out_of_server_memory);
        for (i=0,XMLListGoToFirst(eqm->devices); i<lstsiz && (dev=(xmlDEVIS *)XMLListGetNext(eqm->devices)) != NULL; i++)
        { /* run through eqm->devices for the common devices */
          strncpy((*devis)[i].devname,dev->name,DEVICE_NAME_SIZE);
          strncpy((*devis)[i].devdesc,dev->desc,DEVICE_DESC_SIZE);
          strncpy((*devis)[i].devloc,dev->loc,DEVICE_LOCATION_SIZE);
          strncpy((*devis)[i].devrdr,dev->redirection,DEVICE_REDIR_SIZE);
          (*devis)[i].devnum = dev->number;
          (*devis)[i].devmask = dev->mask;
          strncpy((*devis)[i].devregion,dev->region,TAG_NAME_SIZE);
          (*devis)[i].devoffline = dev->offline;
          (*devis)[i].zpos = dev->zpos;
          /* is there a device specific property list ? */
          strncpy((*devis)[i].prplst,dev->propertySet,SET_NAME_SIZE);;
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
int populatePRPDEVIS(char *fecname,char *eqmname,char *prpname,DBLNKLIST *fecList,DEVIS **devis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlSETIS *set;
  NAME64 *member;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || devis == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        XMLListGoToFirst(eqm->properties);
        while ((prp=(xmlPRPIS*)XMLListGetNext(eqm->properties)) != NULL)
        {
          if (strncmp(prp->name,prpname,PROPERTY_NAME_SIZE)) continue;
          /* found the property; does it have a device set ? */
          if (strlen(prp->deviceSet) == 0) ccerr(0);
          /* are there registered sets ? */
          XMLListGoToFirst(eqm->sets);
          while ((set=(xmlSETIS*)XMLListGetNext(eqm->sets)) != NULL)
          {
            if (strncmp(set->name,prp->deviceSet,SET_NAME_SIZE)) continue;
            /* that's the set !*/
            lstsiz = set->members->nElements;
            XMLListGoToFirst(set->members);
            if ((*devis=(DEVIS *)SystemCalloc(lstsiz,sizeof(DEVIS))) == NULL) ccerr(out_of_server_memory);
            for (i=0,XMLListGoToFirst(set->members); i<lstsiz && (member=(NAME64 *)XMLListGetNext(set->members)) != NULL; i++)
            { /* run through eqm->devices for the common devices */
              strncpy((*devis)[i].devname,member->name,DEVICE_NAME_SIZE);
              (*devis)[i].devnum = i;
            }
          }
        }
      }
    }
  }
err:
  if (cc) feclog("populate Property-specific Device List from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
int populateDEVPRPIS(char *fecname,char *eqmname,char *lstname,DBLNKLIST *fecList,NAME64 **prps)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlSETIS *set;
  NAME64 *member;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || prps == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      XMLListGoToFirst( fec->eqms );
      while((eqm=(xmlXIS*)XMLListGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        XMLListGoToFirst(eqm->sets);
        while ((set=(xmlSETIS*)XMLListGetNext(eqm->sets)) != NULL)
        {
          if (strncmp(set->name,lstname,SET_NAME_SIZE)) continue;
          /* that's the set !*/
          lstsiz = set->members->nElements;
          XMLListGoToFirst(set->members);
          if ((*prps=(NAME64 *)SystemCalloc(lstsiz,sizeof(DEVIS))) == NULL) ccerr(out_of_server_memory);
          for (i=0,XMLListGoToFirst(set->members); i<lstsiz && (member=(NAME64 *)XMLListGetNext(set->members)) != NULL; i++)
          { /* run through eqm->devices for the common devices */
            strncpy((*prps)[i].name,member->name,PROPERTY_NAME_SIZE);
          }
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
int populateAIS(char *fecname,DBLNKLIST *fecList,ALIAS_TABLE **ais)
{
  xmlFIDS *fec;
  xmlAIS *alias;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || ais == NULL) ccerr(argument_list_error);
  XMLListGoToFirst(fecList);
  while((fec=(xmlFIDS*)XMLListGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      XMLListGoToFirst( fec->alias );
      while((alias=(xmlAIS*)XMLListGetNext(fec->alias)) != NULL) lstsiz++;
      if ((*ais=(ALIAS_TABLE *)SystemCalloc(lstsiz,sizeof(ALIAS_TABLE))) == NULL) ccerr(out_of_server_memory);
      for (i=0,XMLListGoToFirst(fec->alias); i<lstsiz && (alias=(xmlAIS *)XMLListGetNext(fec->alias)) != NULL; i++)
      { /* run through eqm->devices for the common devices */
        strncpy((*ais)[i].alias,alias->name,DEVICE_NAME_SIZE);
        strncpy((*ais)[i].name,alias->target,DEVICE_NAME_SIZE);
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %.32s",cc2str(cc));
  return cc ? -cc : lstsiz;
}
TINE_EXPORT int XMLReadFile(char *filepath, char *filename, DBLNKLIST *fecList)
{
  FILE *h;
  char c;
  char fn[256], data[256], tag[32], unknownTag[32];
  int  cc = 0, cp, ltp, gtp;
  int  dataCharCnt = 0;
  int  tagCharCnt = 0;
  int  closingTagOpened = 0, openingTagOpened = 0;
  int  dataAllowed = 0;	 /* if the tag is opened <tag> and not closed yet </tag> */
  int  escapeDone = FALSE;

  memset( unknownTag, '\0', 32 );
  sprintf(fn,"%s%s",filepath,filename);
  if( (h = fopen( fn, "r" )) == NULL ) ccerr(no_such_file);

  DocumentHandler.currentDataSection = fecList;
  XMLSetCurrentHandler( &DocumentHandler );

  while( !feof(h) )
  {
    c = getc(h);
    switch( c )
    {
      case '<': 
        data[dataCharCnt] = 0;
        if( dataAllowed ) CurrentHandler->processData( data );
        dataAllowed = 0;
        data[0] = 0;
        dataCharCnt = 0;
        tagCharCnt = 0;
        openingTagOpened = 1;
        if ((c=getc(h)) == '/') 
        { /* closing */       
          openingTagOpened = 0;
          closingTagOpened = 1; 			
        }
        else if (c == '!')
        { /* a comment: scan to termination */
          cp = 0;
          while (!feof(h))
          {
            if ((c=getc(h)) == '-') cp++;
            else if (c == '>' && cp == 2) break; 
            else cp = 0;
          }
          openingTagOpened = 0;
          break;
        }
        else if (c == '?')
        { /* opening query line in a real xml file ? */
          while (!feof(h))
          {
            if ((c=getc(h)) == '>') break;
          }
          openingTagOpened = 0;
          break;
        }
        else 
        { /* opening */
          ungetc(c,h);
        }
        break;
      case '>':
        if( openingTagOpened ) 
        {
          tag[tagCharCnt] = 0;
          if( !unknownTag[0] )
          {
            dataAllowed = CurrentHandler->processOpeningTag( tag );
            if( !dataAllowed ) strncpy( unknownTag, tag, 31 ); 
          }
          tag[0] = 0;
          tagCharCnt = 0;
        }
        else if( closingTagOpened ) 
        {
          tag[tagCharCnt] = 0;
          if( unknownTag[0] )
          {
            if( !strcmp( tag, unknownTag ) ) unknownTag[0] = 0;
          }
          else
          {
            CurrentHandler->processClosingTag( tag );
            tag[0] = 0;
            tagCharCnt = 0;
            dataAllowed = 0;
          }
        }
        openingTagOpened = closingTagOpened = 0;
        break;
      case '&': /* escape sequence : e.g. &lt; or &gt;*/
        ltp = gtp = 0; escapeDone = FALSE;
        while (!feof(h) && !escapeDone)
        {
          switch (getc(h))
          {
            case 'l': ltp++; gtp = 0; break;
            case 'g': gtp++; ltp = 0; break;
            case 't': ltp++, gtp++; break;
            case ';': 
              if (ltp == 2) c = '<';
              if (gtp == 2) c = '>';
              escapeDone = TRUE;
              break;
            default:
              ltp = gtp = 0;
              continue;
          }
        }
        if (feof(h)) break; /* else fall thru */
      default:		 
        if( c >= ' ' && dataCharCnt < 255)
        {
          if( openingTagOpened || closingTagOpened ) tag[tagCharCnt++] = c;
          else if( dataAllowed ) data[dataCharCnt++] = c;
        }		 
    }
  }
err:
  if (h != NULL) fclose(h);
  return cc;
}
#endif /* SMALL_TINE_LIB */

void BytesToHex(BYTE *b,char *s,int n)
{
  int v;

  while (n--)
  {
    *s++ = (v = (*b&0xf0)>>4) < 10 ? '0' + v : 'A' + v - 10;
    *s++ = (v = *b++&0xf) < 10 ? '0' + v : 'A' + v - 10;
  }
  *s = 0;
  strrev(s);
  return;
}
/**
 * \internal
 * \brief Converts a string into a byte array.
 * 
 * Each digit is interpreted as a hexadecimal number.
 * 
 * \note This function also prepends up to 13 '0's to the input string. 
 * 
 * \param s String containing a hexadecimal number (input and output).
 * \param b Pointer to resulting byte array.
 * \param nibs Double length of b and ALSO buffer length (not string length) of s.
 * \todo This is two functions in one: a) Prepend zeros to a number, b) convert to hex
 */
void HexToBytes(char *s,BYTE *b,int nibs)
{
  char buf[20];
  int k,i;
  if (s == NULL || b == NULL) return;
  if (!strnicmp(s,"0x",2)) s += 2;
  if ((k=(int)strlen(s)) < nibs)
  {
    strrev(s); strncat(s,"000000000000",nibs-k);
    strrev(s);
  }
  for (i=0; i<nibs/2; i++)
  {
    strncpy(buf,&s[i*2],2); buf[2] = 0;
    b[i] = (unsigned char)strtol(buf,NULL,16);
  }
}
unsigned int HexToDec(char *str)
{
  int i,d;
  unsigned int off,value;

  off=1;value=0;
  for (i=(int)strlen(str)-1;i>=0;i--)
  {
    if (str[i]>='A') d=str[i]-'A'+10;else d=str[i]-'0';
    value=value+off*d;
    off=off*16;
  }
  return value;
}
#if !defined(UNIX) && !defined(VXWORKS) && !defined(NIOS)
/* provide this posix routine for those OSes which don't have it */
TINE_EXPORT char *strtok_r(char *s, const char *delim, char **ptrptr)
{
  char *str, *c;
  if (delim == NULL || ptrptr == NULL) return NULL;
  if (s != NULL) /* first call */
  {
    /* start at the beginning of the string given */
    *ptrptr = s + strspn(s,delim); /* skip initial delimiters */
  }
  if (*ptrptr == NULL) return NULL; /* I must be at the end */
  str = *ptrptr; /* this is the current starting point */
  if ((c=strpbrk(str,delim)) != NULL) /* move along */
  {
    *ptrptr = c + strspn(c,delim);
    *c = 0;
  }
  else /* no more tokens */
  {
    *ptrptr = NULL;
  }
  return str;
}
#endif

/**
 * \internal
 * \brief Removes whitespace from the beginning and the end of a string.
 * 
 * Also removes newline characters from the end of the string.
 *
 * \note Modifies the input string in situ.
 * 
 * \return Pointer to modified string.
 */
TINE_EXPORT char *strtrm(char *s)
{
  char *start, *end;
  unsigned int len, olen;
# define iswhite(c) ((c)>0 && (c)<127 && isspace(c))
  if (s == NULL) return s;  /* NULL given, give it back */
  if ((len=(int)strlen(s)) == 0) return s; /* empty string given, give it back */
  olen = len; /* note the original string length */
  /* find first non-whitespace */
  for (start=s; iswhite((int)*start) && len > 0; start++, len--); 
  /* find last non-whitespace */
  for (end=start+len-1; iswhite((int)*end) && len > 0; end--, len--) *end = 0;
  if (s == start && len == olen) return s; /* nothing changed */
  /* Move to the left */
  memmove(s, start, len + 1);
  /* return a pointer to the modified string as a curtesy */
  return s;
}
TINE_EXPORT char *strntrm(char *s,int maxlen)
{
  char *start, *end;
  unsigned int len;

  if (s == NULL) return s;  /* NULL given, give it back */
  if ((len=(int)strlen(s)) == 0) return s; /* empty string given, give it back */
  if (len > (unsigned int)maxlen) len = maxlen;
  /* find first non-whitespace */
  for (start=s; isspace((int)*start) && len > 0; start++, len--); 
  /* find last non-whitespace */
  for (end=start+len-1; isspace((int)*end) && len > 0; end--, len--) *end = 0;
  /* Move to the left */
  memmove(s, start, len + 1);
  if (len < (unsigned int)maxlen) s[maxlen-1] = 0;
  /* return a pointer to the modified string as a curtesy */
  return s;
}
TINE_EXPORT int strtobool(const char *str)
{
  char s[8];
  if (str == NULL) return FALSE;
  strncpy(s, str, 8); strupr(s);
  return (!strcmp(s, "TRUE") || !strcmp(s, "YES") || !strcmp(s, "ON"));
}
int isReservedFileName(char *filename)
{
# if defined(WIN32) || defined(DOS)
  char fn[256];
  strncpy(fn,filename,256);
  strupr(fn);
  if (strncmp(fn,"AUX.",4) == 0 || strstr(fn,".AUX.") != NULL || strstr(fn,"\\AUX.") != NULL) return TRUE;
  if (strncmp(fn,"CON.",4) == 0 || strstr(fn,".CON.") != NULL || strstr(fn,"\\CON.") != NULL) return TRUE;
  if (strncmp(fn,"PRN.",4) == 0 || strstr(fn,".PRN.") != NULL || strstr(fn,"\\PRN.") != NULL) return TRUE;
# else
  UNUSED(filename);
# endif
  return FALSE;
}
# define CSV_MAXLINE_LENGTH 350
/**
 * \internal
 * \brief Counts the lines in a text file.
 * 
 * This function assumes that the first line is a header line,
 * which is not counted. It does not count comment lines.
 * Comment lines are lines which start with '%", '#', or ';'.
 * 
 * \return 
 * <ul> 
 * <li>number of lines</li>
 * <li>-1 if file is empty</li>
 * <li>0 if the file contains only a header line</li>
 * <li>-no_such_file if the file could not be openend</li>
 * </ul> 
 */
int GetDataBaseSize(char *filename)
{
  char str[CSV_MAXLINE_LENGTH+8];
  int k, line = -1;
  FILE *fp;

  if (filename == NULL || strlen(filename) == 0) return -argument_list_error;
  if (isReservedFileName(filename)) return -invalid_name;
  if ((fp=fopen(filename,"r")) == NULL) return -no_such_file;
  for (k=0; fgets(str,CSV_MAXLINE_LENGTH+1,fp) != NULL; k++)   /* scan file       */
  {
    if (strchr(str, '\n') == NULL)
    { /* no CR-LF found */
      if (!feof(fp))
      {
        feclogEx(TINE_LOGLEVEL_WARN, "file %s line %d is too long", filename, k);
      }
      while (fgetc(fp) != '\n' && !feof(fp)); /* goto end of line */
    }
    strtrm(str); /* remove trailing white space */
    if (strlen(str) == 0) continue; /* just a carriage return : skip */
    if (strchr("%;#",str[0])) continue; /* comment line(s) : skip */
    line++;
  }
  fclose(fp);
  if (line < 0)
  {
    feclogEx(TINE_LOGLEVEL_WARN, "file %s found, but has no contents",filename);
    line = 0;
  }
  return line;
}
TINE_EXPORT int GetCsvDataBaseSize(const char *filename)
{
  return GetDataBaseSize((char *)filename);
}
int csvNameServerFillPortOffset(char *po);
int csvNameServerFillNetAdr(char *ip);
int csvNameServerFillIpv4Adr(char *ipv4);
int csvNameServerFromHost(char *host);
/**
 * \internal
 * \brief Array of FEC data.
 * This array serves as a local cache for FEC addresses.
 * Each entry in FecTbl is identified by a FEC name and contains
 * the information which is needed to access that FEC.
 * 
 * Declared in srvcore.h, allocated in initFecTable(), deallocated in freeServerMemory().
 * \note Index 0 in this array always refers to the current name server. 
*/
FecDataStruct *FecTblR4 = NULL;

/**
 * \internal
 * \brief The FEC adress table  (release 5 : ipv6 ready).
 * This array serves as a local cache for FEC addresses.
 * Each entry in FecTbl is identified by a FEC name and contains
 * the information which is needed to access that FEC.
 * 
 * Declared in srvcore.h, allocated in initFecTable(), deallocated in freeServerMemory().
 * \note Index 0 in this array always refers to the current name server. 
*/
FecAddrStruct *FecTbl = NULL;

/**
 * \internal
 * \brief Array of Name server data.
 * 
 * This array contains information about available name servers.
 * 
 * Declared in srvcore.h, allocated in SetNameServerAddress(), deallocated in freeServerMemory().
 * 
 * \see numNameServers
 */
FecAddrStruct *NameServer = NULL;

/**
 * \internal
 * \brief Number of valid entries in NameServer.
 * 
 * Declared in dbtools.c, srvcore.h.
 * 
 * \see NameServer
 */
int numNameServers = 0;

/**
 * \internal
 * \brief convert legacy address to modern address struct
 * 
 * used internally: no null pointer checks
 * no IPv6 !
 * 
 */
void toFecAddr(FecAddrStruct *fas, FecDataStruct *fds)
{
  UINT32 haddr;
  memset(fas,0,sizeof(FecAddrStruct));
  strncpy(fas->fecName,fds->fecName,FEC_NAME_SIZE);
  strncpy(fas->strAdr,fds->IP,ADDR_SIZE);
  fas->inetProtocol = fds->inetProtocol;
  fas->tineProtocol = fds->tineProtocol;
  fas->portOffset = fds->portOffset;
  memcpy(&haddr,fds->IPh_addr,4);
  if (haddr == 0) haddr = inet_addr(fds->IP);
  memcpy(fas->netAdr.ipv4,&haddr,4);
}
/**
 * \internal
 * \brief convert ipv4 address to ipv6 mapped address
 * 
 * used internally: no null pointer checks
 * handle in-situ conversions ...
 * 
 */
char *toIpv6Addr(char *ipv4addr,char *ipv6addr)
{
  char adr[ADDR_SIZE_IPV6];
  switch (getInetFamilyFromAddressString(ipv4addr))
  {
    case AF_INET6:
      /* already IPv6 !*/
      if (ipv4addr != ipv6addr)
        strncpy(ipv6addr,ipv4addr,ADDR_SIZE_IPV6);
      break;
    case AF_INET:
      sprintf(adr,"%s%.16s",IPV6_IPV4PREFIX,ipv4addr);
      strncpy(ipv6addr,adr,ADDR_SIZE_IPV6);
      break;
    default:
      break;
  }
  return ipv6addr;
}
/**
 * \internal
 * \brief convert to ipv4 address from ipv6 mapped address
 * 
 * used internally: no null pointer checks
 * handle in-situ conversions ...
 * 
 */
char *toIpv4Addr(char *ipv6addr,char *ipv4addr)
{
  char adr[ADDR_SIZE_IPV6];
  switch (getInetFamilyFromAddressString(ipv6addr))
  {
    case AF_INET:
      /* already IPv4 !*/
      if (ipv4addr != ipv6addr)
        strncpy(ipv4addr,ipv6addr,ADDR_SIZE);
      break;
    case AF_INET6:
    {
      int plen = (int)strlen(IPV6_IPV4PREFIX);
      if (!strncmp(ipv6addr,IPV6_IPV4PREFIX,plen))
      {
        strncpy(adr,&ipv6addr[plen],ADDR_SIZE);
        strncpy(ipv4addr,adr,ADDR_SIZE);
      }
    }
    default:
      break;
  }
  return ipv4addr;
}
int isIpv4on6(char *addr)
{
  int plen = (int)strlen(IPV6_IPV4PREFIX);
  char *c;
  if (strncmp(addr,IPV6_IPV4PREFIX,plen)) return FALSE;
  if ((c=strchr(addr,'.')) == NULL) return FALSE;
  c++;
  if ((c=strchr(c,'.')) == NULL) return FALSE;
  c++;
  if ((c=strchr(c,'.')) == NULL) return FALSE;
  c++;
  if (strlen(c) == 0) return FALSE;
  return TRUE;
}
/**
 * \internal
 * \brief convert modern address to legacy address struct
 * 
 * used internally: no null pointer checks
 * no IPv6 !
 * 
 */
void toFecData(FecDataStruct *fds,FecAddrStruct *fas)
{
  UINT32 haddr = 0;
  int p = getProtocolFromAddressString(fas->strAdr);
  char strAdr[ADDR_SIZE_IPV6];
  toIpv4Addr(fas->strAdr,strAdr);
  strncpy(fds->fecName,fas->fecName,FEC_NAME_SIZE);
  strncpy(fds->IP,strAdr,ADDR_SIZE);
  fds->inetProtocol = fas->inetProtocol;
  fds->tineProtocol = fas->tineProtocol;
  fds->portOffset = fas->portOffset;
  if (p == IPX)
  {
    memcpy(fds->IPXNetwork,fas->netAdr.ipx.IpxNet,16);
  }
  else
  {
    memset(fds->IPXNetwork,0,16);
    memcpy(&haddr,fas->netAdr.ipv4,4);
    if (haddr == 0) haddr = inet_addr(strAdr);
  }
  memcpy(fds->IPh_addr,&haddr,4);
}
static char _ensAddrIpv4_[16] = "";
_CSVCOLUMN csvNameServerColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(FecAddrStruct,fecName),NULL,0 },
  { "IPV4_ADDR","",0,CF_NULL,0,NULL,csvNameServerFillIpv4Adr,0 },
  { "IP_ADDR","",0,CF_TEXT,40,(BYTE *)OFFSETIN(FecAddrStruct,strAdr),csvNameServerFillNetAdr,0 },
  { "PORT_OFFSET","0",0,CF_INT32,1,(BYTE *)OFFSETIN(FecAddrStruct,portOffset),NULL,0 },
  { "FEC_HOST","",0,CF_NULL,0,NULL,csvNameServerFromHost,0 },
};

CSVDB csvNameServerDb =
{
  csvNameServerColumns,
  sizeof(csvNameServerColumns)/sizeof(csvNameServerColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(FecAddrStruct),
  0
};
int csvNameServerFillIpv4Adr(char *ipv4)
{
  if (strlen(ipv4) > 0) strncpy(_ensAddrIpv4_,ipv4,16);
  return 0;
}
int csvNameServerFillNetAdr(char *ip)
{
  int fam = getInetFamilyFromAddressString(ip);
  switch (fam)
  {
    case AF_INET: 
      if (gAfInet == AF_INET6) toIpv6Addr(ip,ip); 
      break;
    case AF_INET6: 
      if (gAfInet == AF_INET) 
      { /* need to use an ipv4 address ! */
        char *ipp = ip;
        if (strlen(_ensAddrIpv4_) > 0) ipp = _ensAddrIpv4_;
        toIpv4Addr(ipp,ip);
        memset(NameServer[csvNameServerDb.cnt].strAdr,0,40);
        strncpy(NameServer[csvNameServerDb.cnt].strAdr,ip,16);
        fam = AF_INET;
      }
      break;
  }
  inet_pton(fam,ip,&NameServer[csvNameServerDb.cnt].netAdr);
  if (tineDebug)
  {
    dbglog("Using Name Server %.16s [%.40s]",
            NameServer[csvNameServerDb.cnt].fecName,
            NameServer[csvNameServerDb.cnt].strAdr);
  }
  return 0;
}
TINE_EXPORT int GetHostNameFromAddress(char *addr, char *host, int length)
{
  struct addrinfo *ai = NULL, hints;
  int cc = 0;
  char *myip = GetMyIpAddress();
# if defined(WIN32)
  /* do this here, just in case this is called prior to SystemInit() ! */
  if (!WinsockLoaded()) return system_error;
# endif
  if (host == NULL || addr == NULL || strlen(addr) == 0) return argument_list_error;
  if (!strcmp(addr, myip))
  {
    strncpy(host, "localhost", length);
    return 0;
  }
  memset(&hints, 0, sizeof(hints)); hints.ai_family = AF_UNSPEC;
  if (getaddrinfo(addr, NULL, &hints, &ai) != 0) ccerr(host_not_resolved);
  if (ai != NULL) getnameinfo(ai->ai_addr, (socklen_t)ai->ai_addrlen, host, (socklen_t)length, NULL, 0, 0);
err:
  if (ai != NULL) freeaddrinfo(ai);
  return cc;

}
int csvNameServerFromHost(char *host)
{
# define MAX_NUMENS_FROMHOSTADDR 3
  struct addrinfo *ai=NULL, hints;
  int cc = 0, len, ptr = csvNameServerDb.cnt;
  BYTE *badr;
# if defined(WIN32)
  /* do this here, just in case this is called prior to SystemInit() ! */
  if (!WinsockLoaded()) return system_error;
# endif
  if (host == NULL || strlen(host) == 0) return argument_list_error;
  if (NameServer == NULL)
  { /* 'NameServer' will be allocated via the call to csvReadFile */
    if ((NameServer=(FecAddrStruct *)SystemCalloc(MAX_NUMENS_FROMHOSTADDR,sizeof(FecAddrStruct))) == NULL) 
      ccerr(out_of_server_memory);
  }
  memset(&hints,0,sizeof(hints)); hints.ai_family = AF_UNSPEC;
  if (getaddrinfo(host,NULL,&hints,&ai) != 0) ccerr(host_not_resolved);
  if (csvNameServerDb.siz == 0)
  { /* not being called from a cshosts.csv read */
    if (numNameServers > MAX_NUMENS_FROMHOSTADDR) ccerr(resources_exhausted);
    ptr = numNameServers++;
  }
  badr = getInAddr((SCKADR *)ai->ai_addr);
  len = (ai->ai_family == AF_INET) ? 4 : 16;
  memcpy(&NameServer[ptr].netAdr,badr,len);
  inet_ntop(ai->ai_family,badr,NameServer[ptr].strAdr,ADDR_SIZE_IPV6);
  NameServer[ptr].inetProtocol = (ai->ai_family == AF_INET) ? UDP : UDP6;
  if (strlen(NameServer[ptr].fecName) == 0)
  {
    sprintf(NameServer[ptr].fecName,"ENS#%d",ptr);
  }
err:
  if (ai != NULL) freeaddrinfo(ai);
  if (cc != 0)
  {
    feclog("csvNameServerFromHost: host %s : %.32s",(host != NULL ? host : "(null)"),cc2str(cc));
  }
  return cc;
}
/**
 * \internal
 * \brief Indicates if NameServer has already been initialized.
 * 
 * Roughly equivalent to NameServerLoaded.
 * \see SetNameServerAddress, GetNameServerAddress
 */
int gotNameServerAddress = 0;

void fixTineHome(void)
{
  char *ptr;
  int i;
  static int done = FALSE;
  if (done) return;
  if (gIsEquipmentNameServer) return; /* don't even look */
  tineHomePath[0] = 0;
  if ((ptr=getenv("TINE_HOME")) != NULL || (ptr=getenv("CONTROLDB")) != NULL) 
    strncpy(tineHomePath,ptr,STD_PATH_LENGTH);
# ifdef FS_DELIMITER
  if (((i=(int)strlen(tineHomePath)) > 0) &&
      (tineHomePath[i-1] != FS_DELIMITER))  /* no delimiter ! */
  {
    tineHomePath[i] = FS_DELIMITER; tineHomePath[i+1] = 0;
  }
# endif
  done = TRUE;
}
/**
 * \internal
 * \brief Initialize Name server information
 * 
 * <ol>
 * <li>If the envronment variables TINE_ENS or CONTROLENS are set, use name
 * server information from there.</li>
 * <li>If not, read name server information from cshosts file which is expected to
 * be in a directory indicated by TINE_HOME or CONTROLDB. Therby changes/sets tineHomePath.</li>
 * </ol> 
 * 
 * Only called by GetAllocSizesFromNameServer()
 * 
 * \see GetAddressFromNameServer()
 * \see SetNameServerAddress()
 * \see tineHomePath
 * 
 * \todo tineHomePath is already initialized in initFecTable.
 */
int GetNameServerAddress(void)
{
  int i, cc;
  char *ptr;
# ifndef SMALL_TINE_LIB
  static int done = FALSE, logged = 0;
  if (!done)
  {
    done = TRUE;
    initIpFromEnv();
    if ((ptr=getenv("TINE_NETWORKADDRESS_RESOLUTION")) != NULL) 
    {
      SetAllowNetworkAddressResolution(strtobool(ptr));
    }
  }
# endif
  if (gIsEquipmentNameServer)
  {
    if (!logged)
    {
      logged = 1;
      feclog("Server is the ENS: not using a name server!");
      feclog("TINE HOME : working directory");
    }
    numNameServers = 0;
    return not_implemented;
  }
  if ((ptr=getenv("TINE_ENS")) != NULL || (ptr=getenv("CONTROLENS")) != NULL)
  {
    SetNameServerAddress(ptr);
    if (!logged)
    {
      logged = 1;
      feclog("ENS: set ENS address from TINE_ENS environment %s", ptr);
    }
  }
  if (gotNameServerAddress) return 0;
  fixTineHome();
  if (logged < 2) feclog("TINE HOME : [%.128s]",tineHomePath);
  logged = 2;
  numNameServers = 0;
  if ((cc=csvReadFile(tineHomePath,CSHOSTS_FILE,&csvNameServerDb,(void **)&NameServer)) == 0)
  {
    gotNameServerAddress = 1;
    numNameServers = csvNameServerDb.siz;
    if (logged < 3) feclog("ENS: set ENS address from %s",CSHOSTS_FILE);
    logged = 3;
  }
  else
  { /* one last try from the DNS */
    feclog("read %s : %.32s",CSHOSTS_FILE,cc2str(cc));
    if (gSystemRunningStandAlone)
    {
      if (logged < 3) feclogEx(TINE_LOGLEVEL_WARN, "running stand-alone: name server address unresolved");
      logged = 3;
      return name_server_unknown;
    }
    if (csvNameServerFromHost("tineens") == 0 || csvNameServerFromHost("tineens1") == 0) 
    { /* look for up to 3 entries */
      if (logged < 3) feclog("ENS: set ENS address from DNS resolution of 'tineens'");
      logged = 3;
      if (csvNameServerFromHost("tineens2") == 0) csvNameServerFromHost("tineens3");
    }
#   ifndef SMALL_TINE_LIB
    if (numNameServers == 0)
    { /* okay, now one really last desperation try from network probe : */
      if (FindNameServerOnNetwork() == 0)
      {
        if (logged < 4) feclog("ENS: set ENS address from NETWORK probe (multicast for services)");
        logged = 4;
      }
    }
#   endif
    if (numNameServers == 0) 
    {
      if (logged < 5) feclogEx(TINE_LOGLEVEL_ERROR, "ENS: name server address unresolved");
      logged = 5;
      cc = name_server_unknown;
    }
    else 
    {
      cc = 0;
      gotNameServerAddress = 1;
    }
  }
  for (i=0; i<numNameServers; i++)
  {
    NameServer[i].tineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
    if (NameServer[i].inetProtocol == 0)
    {
      NameServer[i].inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP;
    }
  }
  return cc;
}

/**
 * \internal
 * \brief Determines the index of a specified column name in a header line.
 * 
 * Some alternate spellings are tolerated.
 * 
 * \return Column index, starting with 0.
 * 
 */
int findcol(char *str,char *tag)
{
  char tagAlt0[96],tagAlt1[96],tagAlt2[96],altc=0,brkc=0;
  char *c;
  int len,i,k;

  if (str == NULL || tag == NULL) return -1;
  /* effect a string-insensitive compare */
  strncpy(tagAlt0,tag,96);
  strupr(tagAlt0);  
  if ((c=strchr(tagAlt0,' ')) != NULL) 
  {
    brkc = ' ';
    altc = '_';
  }
  else if ((c=strchr(tagAlt0,'_')) != NULL) 
  {
    brkc = '_';
    altc = ' ';
  }
  if (brkc != 0) /* prepare alternates */
  {
    for (i=0,k=0; i<(int)strlen(tagAlt0) && i<95; i++)
    {
      if (tagAlt0[i] != brkc)
      {
        tagAlt1[i] = tagAlt0[i];
        tagAlt2[k++] = tagAlt0[i];
      }
      else /* tagAlt0[i] == brkc */
      {
        tagAlt1[i] = altc;
      }
    }
    tagAlt1[i] = 0; tagAlt2[k] = 0;
  }
  for (i=0,c=str; strlen(c); i++,c++)
  {
    while (*c == '\"') { c = strchr(++c,'\"'); if (c++ == NULL) return -1; }
    while (isspace((int)*c) != 0) c++;
    strupr(c);
    len = (int)strcspn(c,",\n"); while (isspace((int)c[len-1]) != 0) len--;
    if (!strncmp(c,tagAlt0,MAX(len,(int)strlen(tagAlt0)))) return i;
    if (brkc != 0) 
    {
      if (!strncmp(c,tagAlt1,MAX(len,(int)strlen(tagAlt1)))) return i;
      if (!strncmp(c,tagAlt2,MAX(len,(int)strlen(tagAlt2)))) return i;
    }
    if ((c=strpbrk(c,",\n")) == NULL) return -1;
  }
  return -1;
}
TINE_EXPORT int GetCsvColumnIndex(const char *csvHeader,const char *columnTag)
{
  return findcol((char *)csvHeader,(char *)columnTag);
}
int csvColumnCorrection = 0;
/**
 * \internal
 * \brief Extracts a partial string from a comma-separated line.
 * 
 * This function scans a comma-separated line and extracts the column
 * which is indexed by col.  The extracted string is copied into
 * the supplied buffer (and is thread safe!)
 * 
 * \param col Column number. Indexing starts at 0.
 * \param str Comma-separated line.
 * 
 * \return 0 upon success or a TINE error code.
 * 
 */
int getColumnStr(int col,char *str,char *buf,int bufsiz)
{
  int colcnt=0, cc = 0;
  char *c;
  int sslen, FindEndOfString = FALSE;

  csvColumnCorrection = 0;
  if (str == NULL || buf == NULL) ccerr(argument_list_error);

  /* find column position 'col' (count the non-embedded commas) : */
  /* c points to current position in string; DO NOT scan past the string length !*/
  for (c=str; strlen(c) > 0 && colcnt < col; c++)
  {
    /* scan through the string ... */
    if (FindEndOfString)
    {
      /* jump ahead to it */
      if ((c=strchr(c,'\"')) == NULL) ccerr(string_expected);
      FindEndOfString = FALSE; /* found it */
    }
    else
    {
      if (*c == ',') colcnt++;
      else if (*c == '\"') FindEndOfString = TRUE; /* quoted string */
    }
  }

  /* did we jump out because strlen(c) == 0 ? */
  if (colcnt < col) ccerr(no_such_column);

  while (isspace((int)(BYTE)*c) != 0) c++; /* trim leading spaces */
  if (*c == '\"') 
  {
    c++;   /* if quoted, jump over initial quote */
    if (strchr(c, '"') == NULL) ccerr(string_expected);
    sslen = (int)strcspn(c,"\""); /* length up to the next \" */
    /* todo: string could contain multiple quotes */
  }
  else
  {
    sslen = (int)strcspn(c,",\n"); /* length up to the next , or \n */  
  }
  while (sslen > 0 && isspace((int)(BYTE)c[sslen-1]) != 0) sslen--; /* Strip trailing spaces */
  if (sslen > bufsiz)  
  {
    /* sslen longer than expected; write log message and continue anyway */
    feclogEx(TINE_LOGLEVEL_ERROR, "sub string \"%s\" too long (%d vs %d)",c,sslen,bufsiz);
    sslen = bufsiz;
    csvColumnCorrection = string_too_long;
  }
  strncpy(buf,c,sslen); /* Copy sub-string to buffer */
  if (sslen < bufsiz) buf[sslen] = 0; /* add \0 if there's room */

err:
  if (cc)
  {
    if (buf != NULL) memset(buf,0,bufsiz);
    c = str == NULL ? "<null string>" : str;
    feclog("csv: column %d in %s -> %.32s",col,c,cc2str(cc));
  }
  return cc;
}

/**
 * \internal
 * \brief Extracts a partial string from a comma-separated line.
 * 
 * This function scans a comma-separated line and extracts the column
 * which is indexed by col.  The extracted string is copied into
 * a static buffer which is returned (not thread-safe!).
 * 
 * A column may me at most 127 bytes long or will be truncated.
 * 
 * \param col Column number. Indexing starts at 0.
 * \param str Comma-separated line.
 * 
 * \return Pointer to static buffer.
 * 
 */
char *colptr(int col,char *str)
{
  static char buf[128];
  buf[0] = 0;
  getColumnStr(col,str,buf,128);
  return buf;
}
TINE_EXPORT char *GetCsvColumnString(int col,const char *str)
{
  return colptr(col,(char *)str);
}
/**
 * \internal
 * \brief Standard Initializer used for most CSV database readers
 * 
 * Allocate internal buffer memory for a CSVDBStruct.
 * This function assumes that a CSVDBStruct is already partially initialized.
 * <ul>  
 * <li>It allocates memory for the buffer db->tgt if db->tgt_siz > 0. 
 * It uses the values of the 
 * fields db->siz, db->tgt_siz, and db->ncols to determine the buffer size.
 * </li>
 * <li> 
 * It also assumes that db->col is already initialized.
 * For each entry in db->col, it replaces the
 * pointer col->start with the previous value of col->ptr and
 * adds the pointer db->tgt as an offset to the pointer col->ptr.
 * </li>
 * 
 * <li> 
 * It sets db->cnt to 0. 
 * </li>
 * 
 * </ul>
 * 
 * \see CSVDBStruct
 * (This is THE standard initializer for most CSV files)
 * (if the csv contents are supposed to be filling a structure line by line, then the
 *  pointer has to be advanced).
 * \bug Do not call csvStdInit twice (UNLESS csvStdExit is called in between) ... 
 */
TINE_EXPORT void csvStdInit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  db->cnt = 0;
  if (!(db->tgt_flgs & CSVDB_PREALLOC) && db->tgt != NULL) 
  { 
    SystemFree(db->tgt); 
  }
  if (db->tgt == NULL) db->tgt = (BYTE *)SystemCalloc(db->siz,db->tgt_siz);
  for (i=0; i<db->ncols; i++) 
  {
    db->col[i].start = db->col[i].ptr; /* reset via csvStdExit() */
    /* set the ptr to the container addr plus the original offset */
    db->col[i].ptr = db->tgt + (size_t)db->col[i].start; 
  }
}
/**
 * \internal
 * \brief Standard Interator for most CSV database files
 *
 * Increments by row (-> advance relevant column pointers)
 */
TINE_EXPORT int csvStdIterator(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return -1;  /* don't iterate */
  for (i=0; i<db->ncols; i++) if (db->col[i].fmt != CF_NULL) db->col[i].ptr += db->tgt_siz;
  db->cnt++;
  return 0;
}
/**
 * \internal
 * \brief Standard Finalizer for most CSV database files.
 * 
 * Restores column pointers to their values before memory allocation.
 * Has to be called before any Memory-reallocation (prior to another call to csvStdInit()).
 * 
 * \see csvStdInit
 */
TINE_EXPORT void csvStdExit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  for (i=0; i<db->ncols; i++) db->col[i].ptr = db->col[i].start;
}
/**
 * \internal
 * \brief This function takes a string, converts it to a data type
 * specified by a format constant, and stores the result at a place
 * pointed to by a void pointer.
 * The signed-ness of an integer type is entirely given by the type of
 * the pointer passed.
 */
TINE_EXPORT int csvAssignValue(void *ptr, int fmt, int siz, char *str)
{
  return _csvAssignValue(ptr, fmt, siz, str, FALSE);
}
int _csvAssignValue(void *ptr,int fmt,int siz,char *str,int internal)
{
  char *pstr = str, *c;
  int i;
  if (ptr == NULL) return 0;         /* intentionally not assigned */
  if (str == NULL) return argument_list_error;
  if (fmt != CF_TEXT && fmt != CF_NULL && strlen(str) == 0) pstr = "0";
  switch (fmt)
  {
    case CF_NULL: break;
    case CF_TEXT: 
      if (siz == 0) *(char *)ptr = 0; else strncpy((char *)ptr,pstr,siz); 
      break;
    case CF_BYTE: 
      if (siz > 1 && siz <= (int)strlen(str)) HexToBytes(pstr,(BYTE *)ptr,siz); 
      else *(BYTE *)ptr = (BYTE)atoi(pstr);
      break;
    case CF_UINT16:
    case CF_INT16:
      for (i=0; i<siz; i++)
      {
        c = strchr(pstr,' ');
        ((SINT16 *)ptr)[i] = (SINT16)strtol(pstr,NULL,0);
        if (c == NULL) break;
        while (*c == ' ') c++;
        pstr = c;
      }
      break;
    case CF_UINT32:
    case CF_INT32:
      for (i=0; i<siz; i++)
      {
        c = strchr(pstr,' ');
        ((SINT32 *)ptr)[i] = strtol(pstr,NULL,0);
        if (c == NULL) break;
        while (*c == ' ') c++;
        pstr = c;
      }
      break;
    case CF_BOOLEAN:
    {
      int bv;
      for (i = 0; i < siz; i++)
      {
        c = strchr(pstr, ' ');
        bv = (!stricmp(pstr, "TRUE") || !stricmp(pstr, "YES")) ? TRUE : FALSE;
        if (internal || !gMarshallBooleanAsByte) ((SINT32 *)ptr)[i] = bv; else ((_Bool *)ptr)[i] = bv;
        if (c == NULL) break;
        while (*c == ' ') c++;
        pstr = c;
      }
      break;
    }
    case CF_FLOAT:
      for (i=0; i<siz; i++)
      {
        c = strchr(pstr,' ');
        ((float *)ptr)[i] = (float)strtod(pstr,NULL);
        if (c == NULL) break;
        while (*c == ' ') c++;
        pstr = c;
      }
      break;
    case CF_DOUBLE: 
      for (i=0; i<siz; i++)
      {
        c = strchr(pstr,' ');
        ((double *)ptr)[i] = strtod(pstr,NULL);
        if (c == NULL) break;
        while (*c == ' ') c++;
        pstr = c;
      }
      break;
    default: return illegal_format;
  }
  return 0;
}
/**
 * \internal
 * \brief gets the next line in an open
 *  csv file and truncates to 'size' characters if necessary
 * 
 * \param s points to a char buffer of at least \em size bytes
 * \param size is the size of the char string buffer to hold the csv line
 * \param stream is the open file stream
 */
TINE_EXPORT char *csvGetLine(char *s, int size, FILE *stream)
{
  char *csvln = fgets(s, size, stream);
  if (csvln != NULL && strchr(s,'\n') == NULL)
  { /* read to the EOL no matter what 'size' is */
    while (fgetc(stream) != '\n' && !feof(stream));
  }
  return csvln;
}
int isInternalCsvDb(CSVDB* db)
{ /* those which make use of BOOLEAN should rreturn TRUE */
  if (db == &csvHstMfDb) return TRUE;
  return FALSE;
}
/**
 * \internal
 * \brief Reads the contents of a CSV file and places the entries in a container.
 * 
 * The contents of parameter target may be NULL or a valid pointer.
 * If target is not NULL, target will be freed and overwritten with 
 * a pointer to the internal memory of the database structure.
 * If db has no "ini", target will be overwritten
 * with a pointer to db's internal memory.
 * If db has an ini function, it reallocates its internal memory. 
 * \param path The directory where the file is
 * \param file The file to open
 * \param db The db structure 
 * \param target pointer reference (pre-assigned or allocated here if need be)
 * \see findcol
 * \note The memory to which target points is freed before re-allocation.
 *
 * \note the only kernel csv file which is occasionally re-read and 
 * actually needs the memory re-allocation is users.csv. Otherwise,
 * making the same csvReadFile() call a second time will have no effect
 * in the end as long as the standard ini, itr, and exi routines are used.
 *
 * \note some CSV File use different (non-standard) routines. e.g. csvExportIterator().
 *
 */
TINE_EXPORT int csvReadFile(char *path,char *file,CSVDB *db,void **target)
{
  FILE *fp=NULL;
  int i,cc=0,done=FALSE,has_hdr=FALSE,siz;
  char *c,str[CSV_MAXLINE_LENGTH],hdr[CSV_MAXLINE_LENGTH],scratch[CSV_MAXLINE_LENGTH],fn[CSV_MAXLINE_LENGTH];

  if (file == NULL || db == NULL) return argument_list_error;
  if (gSystemMutexSetInitialized && WaitForSystemMutex(hCsvMutex,-1) != 0) return semaphore_error;
  if (path == NULL) strncpy(fn,file,128); else sprintf(fn,"%.128s%.96s",path,file);
  siz = db->siz; /* current recorded size */
  db->tgt = NULL; /* NULL this out for each file read (in case target = NULL passed)! */
  /* find database size => effectively the number of rows */
  if ((db->siz=GetDataBaseSize(fn)) < 0) ccerr(-db->siz);
  /* was a pre-allocated target passed ? */
  if (target != NULL) db->tgt = (BYTE *)*target;
  if (db->tgt != NULL && !(db->tgt_flgs & CSVDB_INITIALIZED)) 
  { /* target container was passed and db has not yet been initialized */
    db->tgt_flgs |= CSVDB_PREALLOC;
  }
  _sendToDbg = 1;
  if (siz > 0 && db->siz > siz && db->tgt != NULL)
  { /* a re-read of a db file now sees more elements ! */
    feclog("re-load database file %.32s requires more storage space",file);
    if (db->tgt_flgs & CSVDB_PREALLOC)
    { /* this will most likely lead to a memory overwrite */
      feclogEx(TINE_LOGLEVEL_WARN, "database storage container was pre-allocated: memory overwrite ?");
    }
  }
  /* call init routine if there */
  if (db->ini != NULL) db->ini(db);
  /* flag db as initialzed */
  db->tgt_flgs |= CSVDB_INITIALIZED;
  /* either the target was pre-allocated or the initializer allocated it */
  if (target != NULL) *target = db->tgt;
  /* read through */
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,CSV_MAXLINE_LENGTH,fp) != NULL)
  {
    str[CSV_MAXLINE_LENGTH - 1] = 0;
    if (strchr(str, '\n') == NULL) while (fgetc(fp) != '\n' && !feof(fp));
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carraige return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      if (str[0] == ',') continue; /* an excel special, but if the header ain't read yet ... */
      strncpy(hdr,str,CSV_MAXLINE_LENGTH);
      for (i=0; i<db->ncols; i++) if ((db->col[i].pos=findcol(hdr,db->col[i].nam)) >= 0) has_hdr = TRUE;
      if (!has_hdr) ccerr(header_not_found);
      done = TRUE;
      continue;
    }
    for (i=0; i<db->ncols; i++)
    {
      if (db->col[i].pos < 0) 
      {
        c = db->col[i].def;
      }
      else
      {
        cc = getColumnStr(db->col[i].pos,str,scratch,CSV_MAXLINE_LENGTH-1);
        scratch[CSV_MAXLINE_LENGTH - 1] = 0;
        if (cc)
        {
          sprintf(str, "(column %.32s entry %d) ", db->col[i].nam, i);
          feclog("file %s %s: %.32s", file, str, cc2str(cc));
          cc = 0;
          continue;
        }
        c = scratch;
        if (strlen(scratch) == 0) c = db->col[i].def;
      }
      /* assign value to structure location */
      cc = _csvAssignValue(db->col[i].ptr,db->col[i].fmt,db->col[i].siz,c,isInternalCsvDb(db));
      if (cc)
      {
        sprintf(str, "(column %.32s entry %d) ", db->col[i].nam, i);
        feclog("csvAssignValue %s %s: %.32s", file, str, cc2str(cc));
        cc = 0;
        continue;
      }
      /* call action function if there */
      if (db->col[i].fcn != NULL) if (db->col[i].fcn(c) != 0) break;    
    }
    /* call iterator if there */
    if (db->itr != NULL) if (db->itr(db) != 0) break;
  }
  /* call end routine if there */
  if (db->exi != NULL) db->exi(db);
err:
  if (fp) fclose(fp);
  if (cc) 
  {
    if (db->tgt != NULL && !(db->tgt_flgs & CSVDB_PREALLOC))
    { /* if we've allocated something then free it ! */
      SystemFree(db->tgt);
      db->tgt = NULL;
      if (target != NULL) *target = NULL;
    }
    db->siz = 0;
    if (cc != no_such_file)
    {
      feclog("csvReadFile %s : %.32s",file,cc2str(cc));
    }
  }
  _sendToDbg = 0;
  if (gSystemMutexSetInitialized) ReleaseSystemMutex(hCsvMutex);
  return cc;
}
/**
 * \internal
 * \brief Reads data from memory and converts it to a string.
 * \param ptr Pointer in memory
 * \param fmt TINE CF data type expected at pointer
 * \param siz Maximal string size
 * \param str Destination
 *
 * \note 'str' assumed to be at least 256 characters
 */
TINE_EXPORT int csvGetValue(void *ptr, int fmt, int siz, char *str)
{
  return _csvGetValue(ptr, fmt, siz, str, FALSE);
}
int _csvGetValue(void *ptr,int fmt,int siz,char *str,int internal)
{
#define csvParseValueToString(SF,CF) {\
  int i, len; char *sp = str;\
  for (i=0; i<siz; i++)\
  {\
    if (i > 0) strcat(str," ");\
    len = (int)strlen(str);\
    if (len > 240) break; /* a modicum of safety ... */\
    sp = &str[len];\
    sprintf(sp,SF,((CF *)ptr)[i]);\
  }\
}
  if (ptr == NULL) return 0;         /* intentionally not assigned */
  if (str == NULL) return argument_list_error;
  *str = 0; /* start fresh */
  switch (fmt)
  {
    case CF_TEXT: strncpy(str,(char *)ptr,siz); break;
    case CF_BYTE: BytesToHex((BYTE *)ptr,str,siz); break;
    case CF_BOOLEAN:
    {
      int i, len, bv; 
      char *sp = str;
      for (i=0; i<siz; i++)
      {
        if (i > 0) strcat(str," ");
        len = (int)strlen(str);
        if (len > 240) break; /* a modicum of safety ... */
        sp = &str[len];
        bv = (internal || !gMarshallBooleanAsByte) ? ((SINT32 *)ptr)[i] : ((_Bool *)ptr)[i];
        sprintf(sp,"%s",(bv == 0 ? "FALSE" : "TRUE"));
      }
    }
    case CF_INT16: 
      csvParseValueToString("%d",short);
      break;
    case CF_UINT16:
      csvParseValueToString("%d", UINT16);
      break;
#   ifdef OS_16BIT
    case CF_INT32:
      csvParseValueToString("%ld",SINT32);
      break;
    case CF_UINT32:
      csvParseValueToString("%ld", UINT32);
      break;
#   else
    case CF_INT32:
      csvParseValueToString("%d",SINT32);
      break;
    case CF_UINT32:
      csvParseValueToString("%d", UINT32);
      break;
#   endif
    case CF_FLOAT:
      csvParseValueToString("%g",float);
      break;
    case CF_DOUBLE:
      csvParseValueToString("%g",double);
      break;
    default: 
      return illegal_format;
  }
  return 0;
}
/**
 * \internal
 * \brief Writes data in a CSVDB to a file.
 * \param path Directory
 * \param file File name
 * \param db A CSVDB
 */
TINE_EXPORT int csvWriteFile(char *path,char *file,CSVDB *db)
{
  FILE *fp=NULL;
  int i,j,cc=0;
  char str[512],scratch[256],fn[128];

  if (file == NULL) return argument_list_error;
  if (gSystemMutexSetInitialized && WaitForSystemMutex(hCsvMutex,-1) != 0) return semaphore_error;
  if (path == NULL) strncpy(fn,file,128); else sprintf(fn,"%s%s",path,file);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  for (i=0,str[0]=0; i<db->ncols; i++)
  {
    strcat(str,db->col[i].nam);
    if (i<db->ncols-1) strcat(str,","); 
    if (db->tgt_siz > 0)  /* return pointers to start position */
    {
      db->col[i].ptr = db->tgt + (size_t)db->col[i].start;
    }
  }
  strcat(str,"\n");
  fwrite(str,strlen(str),1,fp);
  for (j=0; j<db->siz; j++)
  {
    str[0] = 0;
    for (i=0; i<db->ncols; i++)
    {
      _csvGetValue(db->col[i].ptr,db->col[i].fmt,db->col[i].siz,scratch,isInternalCsvDb(db));
      strncat(str,scratch,256);
      if (i<db->ncols-1) strcat(str,","); 
    }
    strcat(str,"\n");
    fwrite(str,strlen(str),1,fp);
    /* call iterator if there */
    if (db->itr != NULL) if (db->itr(db) != 0) break;
  }
  /* call end routine if there */
  if (db->exi != NULL) db->exi(db);
err:
  if (fp != NULL) fclose(fp);
  if (gSystemMutexSetInitialized) ReleaseSystemMutex(hCsvMutex);
  return cc;
}
int makeDirectoryPath(char *dname,int log)
{
# ifndef FS_RDONLY
# if defined(UNIX) || defined(WIN32)
# if defined(UNIX)
  DIR *dp;
  char *tok = "/";
# elif defined(WIN32)
  char fn[128];
  char *tok = "\\";
  struct _finddata_t fp;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif
  char *c=dname,*cp=NULL,tmpname[128],dirname[128];
  int cc=0;

  strncpy(tmpname,dname,128); dirname[0] = 0;
  while (*c++ == *tok) strcat(dirname,tok); /* in case it begins with '/' or '\\' */
  for (c=strtok_r(tmpname,tok,&cp); c != NULL; c=strtok_r(NULL,tok,&cp))
  {
    if (strlen(c) == 0) break; /* terminating delimiter gives the last token empty */
    strcat(dirname,c);
    strcat(dirname,tok);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL)
    {
      closedir(dp);
      chmod(dirname,S_DENYNONE);
    }
    else
    {
      if (mkdir(dirname,S_DENYNONE)) ccerr(file_error);
    }
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if ((hfp=_findfirst(fn, &fp)) == -1)
    {
      if (mkdir(dirname)) ccerr(file_error);
    }
    else
    {
      _findclose(hfp);
    }
#   endif
  }
err:
  if (cc != 0)
  {
    if (log) feclogEx(TINE_LOGLEVEL_ERROR, "unable to create directory %.128s : %.128s",dirname,SystemGetLastErrorString());
    return FALSE; 
  }
# endif /* defined(UNIX) || defined(WIN32) */
# else  /* FS_RDONLY not defined */
  dname = dname;
# endif /* FS_RDONLY */
  return TRUE;
}
TINE_EXPORT int MakeDirectoryPath(char *dname,int log)
{
  return makeDirectoryPath(dname,log);
}
#endif /* NO_DOXYGEN */
