/*
 * File dbtools.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file database.c
 * \brief Database routines used for name resolution, etc.
 *      
 */
#include "project.def"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#ifdef WIN32
# include <direct.h>
#endif
#if defined(UNIX) || defined(MACOS)
# include <sys/stat.h>
# include <dirent.h>
#endif
#include "prolog.h"
#include "srvcore.h"
#include "dbtools.h"
#include "srvip.h"
#include "globals.h"

#define ENSADDRPKT4L (RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE)
#define ENSADDRPKT4 (FECDATASTRUCT_SIZE+EXPDATASTRUCT_SIZE)
#define ENSADDRPKT6 (FECADDRSTRUCT_SIZE+EXPDATASTRUCT_SIZE)
int gIsEquipmentNameServer = FALSE;
int gIsGroupNameServer = FALSE;
int gEnsAddrPktSize = ENSADDRPKT6;

TINE_EXPORT void isEquipmentNameServer(int value) { gIsEquipmentNameServer = value; }
TINE_EXPORT void isGroupNameServer(int value) { gIsGroupNameServer = value; }
/** 
 * \internal
 * \brief General client-side database path
 * 
 *  Given by the environment variable setting "TINE_HOME" or (deprecated) 
 *  CONTROLDB.  This specifies the location of the 'cshost.csv' file (among others).
 *
 *  Declared in dbtools.h, initialized in initFecTable().
 *
 * \todo refacter the name to TineHomePath
 *
 */
char tineHomePath[STD_PATH_LENGTH];

/**
 * \internal
 * \brief Number of valid entries in SrvTbl
 * \see SrvTbl
 */
int numSrvTblEntries=0;

/**
 * \internal
 * \brief Number of valid entries in FecTbl
 * \see FecTbl
 */
int numFecTblEntries=0;

/**
 * \internal
 * \brief Maximum size of the SrvTbl table
 *
 * This is more or less hard-wired to the gConTblCapacity.  The
 * original intent was to let this grow dynamically as needed, but
 * there are some unfortunate side-effects on some OSes without good
 * MMUs.  So the value is currently being set up front.
 *
 * \see SrvTbl
 */
int SrvTblSize = 0;

/**
 * \internal
 * \brief Maximum size of the FecTbl table
 *
 * This is more or less hard-wired to the gConTblCapacity.  The
 * original intent was to let this grow dynamically as needed, but
 * there are some unfortunate side-effects on some OSes without good
 * MMUs.  So the value is currently being set up front.
 *
 * \see FecTbl
 */
int FecTblSize=0;

/**
 * \internal
 * \brief Flag whether at least one ENS is known or not.
 */
int NameServerLoaded = 0;

/**
 * \internal
 * \brief Index of the currently active ENS 
 *
 * Array index of NameServer[] array which is currently
 * active.
 */
int curNameServer = 0;

/** 
 * \internal
 * \brief Array of Device Server Structs.
 * 
 * This array is used as a local cache which stores
 * information of known EqmContext/ExportName pairs. 
 * The array is used to map EqmContext/ExportName to
 * Equipment name, FEC name and other data.  
 * 
 * Declared in srvcore.h. 
 * Allocated in initFecTable().
 * 
 * \note The element with index 0 always refers to the current name server.
 */
ExpDataStruct *SrvTbl = NULL;

/** 
 * \internal
 * \brief ExpDataStruct which holds the ENS data
 *
 * This is the static information which is used as the 0th
 * entry of the SrvTbl table when at least one ENS is configured.
 *
 * \note If there is an ENS, then it is the 0th element in the cached
 * address tables AND the 0th element in the connection table.
 *
 */
static ExpDataStruct NameServerData; /**< Name Server data.*/

/** 
 * \internal
 * \brief Connection Table Entry which holds the ENS data
 *
 * This is the static information which is used as the 0th
 * entry of the connection table when at least one ENS is configured.
 *
 * \note If there is an ENS, then it is the 0th element in the cached
 * address tables AND the 0th element in the connection table.
 *
 */
ConTblEntry connNameServer;

/** 
 * \internal
 * \brief Flag which determines whether to use the loopback address when client and
 * server are on the same computer.
 *
 * Some OSes (UNIX, Windows) will use a UNIX socket (pipe) or MMF as a transport
 * between client and server if it is determined that the client is on the same 
 * machine as the server.  This can be over-ridden if gUseLoopback is set to TRUE.
 *
 */
int gUseLoopback = FALSE;
TINE_EXPORT void SetUseLoopback(int value)
{
  gUseLoopback = value;
}
TINE_EXPORT int GetUseLoopback(void)
{
  return gUseLoopback;
}

/** 
 * \internal
 * \brief Flag which determines whether to force the shared-memory address when client and
 * server are on the same computer.
 *
 * Some OSes (UNIX, Windows) will use a UNIX socket (pipe) or MMF as a transport
 * between client and server if it is determined that the client is on the same 
 * machine as the server.  This can be strictly enforced if gNoLoopback is set to TRUE.
 *
 */
int gNoLoopback = FALSE;
TINE_EXPORT void SetForceSharedMemory(int value)
{
  gNoLoopback = value;
}
TINE_EXPORT int GetForceSharedMemory(void)
{
  return gNoLoopback;
}
/**
 * \internal
 * \brief Compares a column of a comma-separated line to a string.
 * 
 * This function scans a comma-separated line and extracts the column which is indexed by col. 
 * 
 * \param name String to compare.
 * \param str Comma-separated line.
 * \param col Column index (starting at 0)
 * \param limit Maximal length of the extracted column (length of result buffer).
 * 
 * \return like strcmp().
 * 
 */
int csvnamcmp(char *name,char *str,int col,int limit)
{
  int n;
  char *scratch, *namescratch;
# if defined(VXWORKS)
  /*  VxWorks does not understand alloca() */
  char scratchBuffer[516], namescratchBuffer[516];
  if (limit < 0 || limit > 512) return -1;
  scratch = scratchBuffer;
  namescratch = namescratchBuffer;
# else
  if (limit < 0) return -1; /* allow a bit of string space past 'limit' */
  if ((scratch=(char *)alloca(limit+8)) == NULL) return -1; /* practically impossible */
  if ((namescratch=(char *)alloca(limit+8)) == NULL) return -1; /* practically impossible */
# endif
  if (name == NULL && str == NULL) return 0; /* NULLs match */
  if (name == NULL && str != NULL) return -1; /* problem with arguments */
  if (str == NULL && name != NULL) return -1;
  if (name != NULL && col < 0) return -1; /* column not present */
  getColumnStr(col,str,scratch,limit); scratch[limit] = 0; 
  strncpy(namescratch,name,limit); namescratch[limit] = 0;
  n = MAX(limit,(int)strcspn(scratch,",\n"));
  return strnicmp(namescratch,scratch,n);
}
int getProtocolFromAddressString(char *strAdr)
{
  if (strAdr == NULL) return (gAfInet == AF_INET6) ? UDP6 : UDP;
  if (strchr(strAdr,':') != NULL)  return UDP6;
  if (strchr(strAdr,'.') != NULL) return UDP;
# if defined(IPXNW) || defined(IPXBSD)
  if (!strncmp(strAdr,"0x",2)) return IPX;
# endif
  return 0;
}
int getInetFamilyFromAddressString(char *strAdr)
{
  if (strAdr == NULL) return gAfInet;
  if (strchr(strAdr,':') != NULL)  return AF_INET6;
  if (strchr(strAdr,'.') != NULL) return AF_INET;
# if defined(IPXNW) || defined(IPXBSD)
  if (!strncmp(strAdr,"0x",2)) return AF_IPX;
# endif
  return 0;
}
int fixInetProtocol(FecAddrStruct *fec,int protocol)
{
  int cc = 0, ipp = (gAfInet == AF_INET6) ? UDP6 : UDP;
  if (ipp == protocol) 
  {
    inet_pton(gAfInet,fec->strAdr,&fec->netAdr);
    return 0;
  }
  if (ipp == UDP6 && protocol == UDP)
  { /* elevate */
    toIpv6Addr(fec->strAdr,fec->strAdr);
    inet_pton(gAfInet,fec->strAdr,&fec->netAdr);
  }
  if (ipp == UDP && protocol == UDP6)
  { /* demote (if possible) */
    if (strstr(fec->strAdr,IPV6_IPV4PREFIX) == NULL) ccerr(not_supported);
    toIpv4Addr(fec->strAdr,fec->strAdr);
    inet_pton(gAfInet,fec->strAdr,&fec->netAdr);
  }
err:
  if (cc) dbglog("fixInetProtocol to %.7s : %.32s",getProtocolAsString(protocol),cc2str(cc));
  return cc;
}
/** 
 * \internal
 * \brief Determine how best to communicate with this FEC
 *
 * Sets the .inetProtocol field to the best available communication protocol.
 *
 * \note If .inetProtocol = 0, the address will again be queried from the ENS.
 * Thus a series of timeouts as seen in the TINE engine will set the .inetProtocol
 * field explicitly to 0 and thereby force an address (re)-resolution.
 *
 */
int fixFecProtocol(int fecidx)
{
  int cc = 0, ipp = (gAfInet == AF_INET6) ? UDP6 : UDP;
  FecAddrStruct *fec = &FecTbl[fecidx];
  int defaultProtocol = getProtocolFromAddressString(fec->strAdr);
# if defined(IPXNW) || defined(IPXBSD)
  if (defaultProtocol == IPX)
  {
    int TransportTime = 0;
#   ifdef IPXNW
    if (IPXGetLocalTarget(fec->netAdr.ipx.IpxNet,&fec->netAdr.ipx.IpxImmAdr,&TransportTime)) 
    {
      cc = ipx_socket_error;
      defaultProtocol = ipp;
    }
#   endif /* IPXNW */
  }
# endif /* defined(IPXNW) || defined(IPXBSD) */
  if ((cc=fixInetProtocol(fec,defaultProtocol)) != 0) ccerr(cc);
  if ((defaultProtocol & (UDP|UDP6)) || /* most often the case : a known IP address */
      !strcmp(fec->fecName,SRVFEC_NAME)) /* a multicast for a service fec -> must be UDP */
  {
    defaultProtocol = ipp;
#   if !defined(NO_TINE_PIPE) && !defined(VXWORKS)
    if (!gUseLoopback) 
    {
      BYTE *myadr = getmyipaddr();
      if (!zerocmp(myadr,SCKADR_LEN)) feclogEx(TINE_LOGLEVEL_ERROR, "cannot determine local ip address!");
      if ((!memcmp(myadr,&fec->netAdr,SCKADR_LEN) || !strncmp(fec->strAdr,gLoopbackAddr,ADDR_SIZE_IPV6)))
      { /* address is on my host ! */
#       if defined(MAKE_TINE32_DLL)
        HWND hwnd = (HWND)FindWindow(fec->fecName, NULL);
        if (hwnd != (HWND)NULL)
        {
          memcpy(&fec->lclAdr.wndMmf,&hwnd,sizeof(HWND)); 
          defaultProtocol = MMF;
        }
        else if (gNoLoopback)
        { /* expect and force shm transfers no matter what */
          defaultProtocol = MMF;
        }
#       elif defined(UNIX)
        if (fec->inetProtocol != PIPE) fec->lclAdr.pipe.sck = 0;
        defaultProtocol = PIPE;
#       endif
        resetOutputBlocks(fecidx);
      }
      fec->addrFlags |= ADDRFLAG_LOCAL_HOST;
      if (gIsRunningAsServer && !strncmp(fec->fecName,gFecName,FEC_NAME_SIZE))
        fec->addrFlags |= ADDRFLAG_LOCAL_PROCESS;
    }
#   endif /* !NO_TINE_PIPE */
  }
  if (fec->tineProtocol == 0 || fec->tineProtocol > DEFAULT_TRANSPORT_PROTOCOL_LEVEL) 
    fec->tineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
  if (tineDebug && fec->inetProtocol != defaultProtocol)
    dbglog("Communicating with %.16s via %s",fec->fecName,getProtocolAsString(defaultProtocol));
  fec->inetProtocol = defaultProtocol;
err:
  if (cc) dbglog("fixFecProtocol: %.32s",cc2str(cc));
  return 0;
}
TINE_EXPORT int gIgnoreVolatileAddressCache = FALSE; /* this will be TRUE for the ENS */
TINE_EXPORT int gIgnoreAddressCacheOnUnknownAddress = FALSE;
int gAllowNetworkAddressResolution = FALSE;
/**
 * \internal
 * \brief Locate an export name and context in the FEC table.
 * 
 * This function tries to locate an entry in the address cache: 
 * FecTbl[] and SrvTbl[] according to the specified export name and context.
 * Fetches FEC data by calling GetAddressFromNameServer() and/or GetAddressFromNameFile(), which 
 * creates or overwrites an entry in FecTbl and/or SrvTbl and return its index.
 * 
 * \param expName Exported device server name
 * \param context EqmContext
 * \param device is passed but only needed for GENS resolution from the local file system (rare)
 * \param srvidx [output] index of corresponding entry in SrvTbl[] or -1. May be set to NULL if no result desired.
 * \return index in FecTbl[] or negative TINE return code.
 * \todo make expName and context const 
 * \see GetAddressFromNameServer 
 * \see GetAddressFromNameFile
 */
int locateExport(char *expName,char *context,char *device,int *srvidx)
{
  char fecName[FEC_NAME_SIZE];
  int i,idx;
# ifndef SMALL_TINE_LIB
  FecAddrStruct fec;
  ExpDataStruct srv;
# endif
  memset(fecName,0,FEC_NAME_SIZE); /* leave fecName in the argument lists below for now */
  if ((i=GetAddressFromNameServer(context,expName,fecName,&idx)) < 0)
  { /* something went wrong ... */
    if (isTimeout(-i))
    { /* ENS timeout out ! */
      if (tineDebug || errorDebug) dbglog("timeout contacting ENS for address");
    }
    else if (context == NULL && i == -ambiguous)
    {
      if (tineDebug || errorDebug) dbglog("no context given and address is ambiguous");
      return i;
    }
    else
    { /* ENS could not resolve the address */
      if (gIgnoreAddressCacheOnUnknownAddress) return i;
    }
#   if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
    if (gIgnoreVolatileAddressCache || (i=GetAddressFromNameFile(gAddrCacheFilePath,context,expName,fecName,device,&idx)) < 0)
#   endif
    if ((i=GetAddressFromNameFile(tineHomePath,context,expName,fecName,device,&idx)) < 0) 
    {
#     ifndef SMALL_TINE_LIB
      if (gAllowNetworkAddressResolution && FindServerOnNetwork(context,NULL,expName,&fec,&srv) == 0)
      { /* last resort: try a network probe */     
        if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return -semaphore_error;
        if (numSrvTblEntries < SrvTblSize) 
        {
          idx = numSrvTblEntries++;
        }
        else 
        {
          ReleaseSystemMutex(hSrvTblMutex); 
          msglog(resources_exhausted,"cannot expand server entry table: capacity reached (%d)",SrvTblSize);
          return -resources_exhausted;
        }
        /* Fill FecTbl entry with data */
        memcpy(&SrvTbl[idx],&srv,sizeof(ExpDataStruct));
        if (context != NULL && strnicmp(context,SrvTbl[idx].EqmContext,CONTEXT_NAME_SIZE) != 0)
        {
          /* context was apparently decorated */
          strncpy(SrvTbl[idx].EqmContext,context,CONTEXT_NAME_SIZE);
        }
        /* Look for FEC name in FecTbl[] */
        for (i=0; i<numFecTblEntries && strnicmp(SrvTbl[idx].FecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
        /* No entry found => create a new one. */
        if (i == numFecTblEntries)
        {
          if (numFecTblEntries < FecTblSize) 
          {
            numFecTblEntries++;
          }
          else 
          {
            ReleaseSystemMutex(hSrvTblMutex); 
            msglog(resources_exhausted,"cannot expand fec entry table: capacity reached (%d)",FecTblSize);
            return -resources_exhausted;
          }
        }      
        memcpy(&FecTbl[i],&fec,sizeof(FecAddrStruct));
        ReleaseSystemMutex(hSrvTblMutex); 
        feclog("ENS: %.32s [FEC %.16s, EQM %.6s] added",SrvTbl[idx].ExportName,SrvTbl[idx].FecName,SrvTbl[idx].EqmName);
        msglog(0,"ENS: %.32s [FEC %.16s, EQM %.6s] added",SrvTbl[idx].ExportName,SrvTbl[idx].FecName,SrvTbl[idx].EqmName);
      }
      else
      { /* all avenues of resolution exhausted ! */
        return -address_unknown;
      }
#     else
      return -address_unknown;
#     endif
    }
  }
  fixFecProtocol(i);
  if (srvidx != NULL) *srvidx = idx;
  return i;
}

/**
 * \internal
 * \brief Retrieves the fec index for the specified FEC name in the FEC table
 *
 */
int getFecIndex(char *fecName)
{
  int i;
  if (strcmp(fecName,ENS_SRV) == 0) return 0;
  if (strcmp(fecName,GENS_FEC) == 0) return 1;
  for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
  if (i < numFecTblEntries) return i;
  return -non_existent_fec;
}
TINE_EXPORT int GetHostFromFecName(char *fecName,char *hostname) 
{
  int i, fam=gAfInet, cc=0;
  FecAddrStruct fec, *pFec;
  SCKADR sckadr;
  _socklen_t len;

  if (fecName == NULL || hostname == NULL) return argument_list_error;
  if ((i=getFecIndex(fecName)) < 0) 
  {
    if ((cc = -i) == non_existent_fec)
    { /* not yet cached ? */
      cc = queryNameServerForFec(fecName,&fec);
    }
    if (cc != 0) ccerr(cc);
    pFec = &fec;    
  }
  else
  {
    pFec = &FecTbl[i];
  }
  len = fam == AF_INET ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);
  if ((sckadr_in(&sckadr,(BYTE *)&pFec->netAdr,(UINT16)gtUDPPort,fam) != 0) ||
      (getnameinfo((const struct sockaddr *)&sckadr,len,hostname,64,NULL,0,NI_NAMEREQD) != 0))
  {
    strncpy(hostname,pFec->strAdr,ADDR_SIZE_IPV6);
  }
err:
  return cc;
}

/**
 * \internal
 * \brief Switches to the next name server.
 * 
 * Increments the global variable curNameServer and
 * copies data from NameServer[curNameServer] to FecTbl[0] and SrvTbl[0].
 * \see curNameServer
 * \see NameServer
 * \see FecTbl
 * \see SrvTbl
 */
TINE_EXPORT int ToggleNameServer(void)
{
  char s[128];
  if (ServerExitCondition) return curNameServer;
  if (NameServer == NULL || FecTbl == NULL) return curNameServer;
  curNameServer = (curNameServer+1) % numNameServers;
  memcpy(&FecTbl[0],&NameServer[curNameServer],sizeof(FecAddrStruct));
  strncpy(SrvTbl[0].FecName,NameServer[curNameServer].fecName,FEC_NAME_SIZE);
  memcpy(&FecTbl[1],&NameServer[curNameServer],sizeof(FecAddrStruct));
  strcpy(FecTbl[1].fecName,GENS_FEC);
  FecTbl[1].portOffset += 101;
  sprintf(s,"ENS: toggle name server to %s",curNameServer == 0 ? "primary" : "secondary");
  if (curNameServer > 0) sprintf(&s[strlen(s)]," no. %d",curNameServer);
  feclog(s); msglog(0,s);
  return curNameServer;
}
void addServerLinksToBlackList(int sid)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    if (conTbl[i]->srvIdx == sid) 
    {
      conTbl[i]->linkStatus = link_blacklisted;
      conTbl[i]->blcounter++;
      appendBlackLnkLst(conTbl[i]->ctxName,conTbl[i]->expName,conTbl[i]->devName,conTbl[i]->prpName,address_unknown);
    }
  }
}
int fecAddrAcqId = -1; /* async link id for addr re-acquisition */
INADR fecAddrMcGrp; /* current assigned multicast group of addr re-acq. */
static int ens_errcnt = 0;
char fecNameTarget[FEC_NAME_SIZE+4];
char exportNameTarget[EXPORT_NAME_SIZE+4];
char exportContextTarget[EXPORT_NAME_SIZE+4];
char exportDeviceTarget[DEVICE_NAME_SIZE+4];
BYTE fecAddressBuf[sizeof(ExpDataStruct)+sizeof(FecAddrStruct)];
BYTE ayncFecAddressBuf[sizeof(ExpDataStruct)+sizeof(FecAddrStruct)];
void fecAddrAcqCb(int lid,int cc)
{
  int i,sid,fid,k,logit=FALSE,updGrp=FALSE,addrChng=FALSE;
  char preCallExp[EXPORT_NAME_SIZE+4],preCallEqm[EQM_NAME_SIZE];
  FecAddrStruct fec;
  ExpDataStruct srv;
  INADR inaddr = inadr_zero;
  ConTblEntry *c = conTbl[lid];
  if (tineDebug > 1) dbglog("asyncRelocate callback for link %d <%d>",lid,cc);
  if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return;
  for (i=0; i<numSrvTblEntries; i++)
  {
    if (strnicmp(exportNameTarget,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) continue;
    if (strlen(exportContextTarget) == 0) break; /* not given */
    if (strnicmp(exportContextTarget,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    break;
  }
  sid = i;
  if (sid == numSrvTblEntries)
  {
    if (numSrvTblEntries >= SrvTblSize) ccerr(resources_exhausted);
    numSrvTblEntries++;
  }
  switch (cc)
  {
    case 0:
      prepIncomingAddress(gEnsAddrPktSize,(char *)ayncFecAddressBuf,&fec,&srv);
      ens_errcnt = 0;
      strncpy(preCallExp,SrvTbl[sid].ExportName,EXPORT_NAME_SIZE); preCallExp[EXPORT_NAME_SIZE] = 0;
      strncpy(preCallEqm,SrvTbl[sid].EqmName,EQM_NAME_SHORTSIZE); preCallEqm[EQM_NAME_SHORTSIZE] = 0;
      memcpy(&SrvTbl[sid],&srv,sizeof(ExpDataStruct));
      if (strnicmp(preCallExp,SrvTbl[sid].ExportName,EXPORT_NAME_SIZE) || strncmp(preCallEqm,SrvTbl[sid].EqmName,EQM_NAME_SHORTSIZE))
      {
        feclog("ENS: name change in assignment %.32s %.6s vs %.32s %.6s\n",
          preCallExp,preCallEqm,SrvTbl[sid].ExportName,SrvTbl[sid].EqmName);
        addrChng = TRUE;
      }
      for (i=0; i<numFecTblEntries && strnicmp(SrvTbl[sid].FecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
      fid = i;
      if (fid == numFecTblEntries)
      {
        if (numFecTblEntries >= FecTblSize) ccerr(resources_exhausted);
        numFecTblEntries++;
      }
      if (c->fecIdx != fid) addrChng = TRUE;
      if (memcmp(FecTbl[fid].strAdr,fec.strAdr,ADDR_SIZE_IPV6) || FecTbl[fid].portOffset != fec.portOffset) addrChng = logit = TRUE;
      memcpy(&FecTbl[fid],&fec,sizeof(FecAddrStruct));
      fixFecProtocol(fid);
      if (memcmp(&fecAddrMcGrp,&inadr_zero,gInAdrLen))
      {
        inaddr = fixMcastHaddr((BYTE *)&fec.netAdr,!(fec.addrFlags & ADDRFLAG_MCAST_MODERN));
        if (memcmp(&fecAddrMcGrp,&inaddr,gInAdrLen))
        { /* new multicast group ! */
          detachPortFromMulticastGroup(mcastClnSck,(UINT16)gtMCastPort,fecAddrMcGrp,"server specific");
          attachPortToMulticastGroup(mcastClnSck,(UINT16)gtMCastPort,inaddr,"server specific");
          updGrp = TRUE;
        }
      }
      for (k=0; k<nConnectionTableEntries; k++)
      {
        if (conTbl[k]->srvIdx == sid)
        { /* assigned to a new FEC ? */
          conTbl[k]->fecIdx = fid;  
        }
        if (!updGrp) continue;
        if (conTbl[k]->fecIdx == fid) conTbl[k]->mcastGrp = inaddr; /* make sure this gets updated */
      }
      if (logit) feclog("ENS: /%.32s/%.32s [FEC %.16s] added",exportContextTarget,exportNameTarget,FecTbl[fid].fecName);
      if (!addrChng && c != NULL && c->rdrId != 0)
      { /* no address change but original link was redirected : revert ! */
        rmvRdrInfo(c->rdrLink,c->linkStatus);
      }
      break;
    case illegal_read_write:
      gEnsAddrPktSize = ENSADDRPKT4L;
      break;
    case illegal_data_size:
      gEnsAddrPktSize = ENSADDRPKT4;
      break;
    default:
      if (++ens_errcnt < 5) if (ToggleNameServer() == 0) ens_errcnt = 5;
      ReleaseSystemMutex(hSrvTblMutex); /* release here due to GetAddressFromNameFile() */
#     if !defined(VXWORKS) && !defined(SMALL_TINE_LIB) && !defined(FS_RDONLY)
      if ((fid=GetAddressFromNameFile(gAddrCacheFilePath,exportContextTarget,exportNameTarget,fecNameTarget,exportDeviceTarget,&sid)) >= 0)
      { /* then look in the local cache again ... */
        fixFecProtocol(fid);
        cc = 0;
      }
#     endif
      if (cc == non_existent)
      { /* put a stop to ENS badgering */
        addServerLinksToBlackList(sid);
      }
      fecAddrMcGrp = inadr_zero;
      fecAddrAcqId = -1;
      return;
  }
err:
  fecAddrMcGrp = inadr_zero;
  fecAddrAcqId = -1; /* release 'semaphore' */
  ReleaseSystemMutex(hSrvTblMutex); 
  return;
}
int asyncRelocateExport(ConTblEntry *c)
{ /* internal use only: no null-pointer checking */
  int cc,siz;
  DTYPE dout;
  char scratch[128],cntxt[CONTEXT_NAME_SIZE+4] = "";
  char *expName = SrvTbl[c->srvIdx].ExportName;
  char *context = SrvTbl[c->srvIdx].EqmContext;
  char *device = c->devName;
  if (!NameServerLoaded || fecAddrAcqId != -1) return -not_ready;

  strncpy(exportNameTarget,expName,EXPORT_NAME_SIZE);
  if (context != NULL) strncpy(cntxt,context,CONTEXT_NAME_SIZE);
  if (device == NULL) device = "";
  strncpy(exportDeviceTarget,device,DEVICE_NAME_SIZE);
  strncpy(exportContextTarget,cntxt,CONTEXT_NAME_SIZE);
  if (tineDebug) dbglog(" ENS: asynchronous query /%.32s/%.32s",cntxt,expName);
  if (strlen(cntxt) == 0) sprintf(cntxt,"#2");
  sprintf(scratch,"%.64s/%.32s",ENS_TAG,cntxt);

  siz = gEnsAddrPktSize;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = (UINT32)siz;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = ayncFecAddressBuf;
  if ((fecAddrAcqId=AttachLinkEx(scratch,expName,&dout,NULL,CA_READ,500,fecAddrAcqCb,CM_SINGLE,c->linkId)) < 0)
  {
    cc = -fecAddrAcqId;
    fecAddrMcGrp = inadr_zero;
    fecAddrAcqId = -1;
    if (tineDebug || errorDebug) dbglog("ENS: error in asynchronous query : %.32s",cc2str(cc));
    return cc;
  }
  return fecAddrAcqId;
}
#ifndef DOS
int appendAddressToCache(char *fecName,char *ip, short portOffset)
{
  int i,cc = 0,fam = gAfInet;
  initClient();
  for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
  if (i < numFecTblEntries && FecTbl[i].inetProtocol) ccerr(0);
  strncpy(FecTbl[i].fecName,fecName,FEC_NAME_SIZE);
  strncpy(FecTbl[i].strAdr,ip,ADDR_SIZE_IPV6);
  inet_pton(fam,FecTbl[i].strAdr,&FecTbl[i].netAdr);
  FecTbl[i].portOffset = portOffset;
  fixFecProtocol(i);
  numFecTblEntries++;
err:
  return cc;
}
/* TODO:  this routine isn't used? */
int appendServerToCache(char *fecName,char *context,char *expName, char *eqmName)
{
  int i,cc = 0;

  if (fecName == NULL || strlen(fecName) == 0) return argument_list_error;
  if (expName == NULL || strlen(expName) == 0) return argument_list_error;
  if (eqmName == NULL || strlen(eqmName) == 0) return argument_list_error;
  initClient();
  if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return semaphore_error;
  for (i=0; i<numSrvTblEntries; i++)
  {
    if (context != NULL && strlen(context) &&
        strnicmp(context,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(expName,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) break;
  }
  if (i < numSrvTblEntries) ccerr(0);
  strncpy(SrvTbl[i].FecName,fecName,FEC_NAME_SIZE);
  if (context != NULL) strncpy(SrvTbl[i].EqmContext,context,CONTEXT_NAME_SIZE);
  strncpy(SrvTbl[i].ExportName,expName,EXPORT_NAME_SIZE);
  strncpy(SrvTbl[i].EqmName,eqmName,EQM_NAME_SHORTSIZE);
  numSrvTblEntries++;
err:
  ReleaseSystemMutex(hSrvTblMutex);
  return cc;
}
TINE_EXPORT void FlushFecAddrCache(void)
{
  if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return;
	numFecTblEntries = 2; /* always leave room for ENS and GENS */
	numSrvTblEntries = 2;
	memset(&FecTbl[numFecTblEntries],0,(FecTblSize-numFecTblEntries)*sizeof(FecAddrStruct));
	memset(&SrvTbl[numSrvTblEntries],0,(SrvTblSize-numSrvTblEntries)*sizeof(ExpDataStruct));
  ReleaseSystemMutex(hSrvTblMutex);
  flushRedirectedLinkList();
  flushReLnkLst();
  flushBlackLnkLst();
  flushWcLst();
}
TINE_EXPORT void flushFecAddrCache(void) { FlushFecAddrCache(); }
TINE_EXPORT int GetENSAddr(char *name,char *addr)
{
  if (NameServerLoaded)
  {
    strncpy(name,NameServer[curNameServer].fecName,FEC_NAME_SIZE);
    strncpy(addr,NameServer[curNameServer].strAdr,ADDR_SIZE_IPV6);
  }
  else
  {
    strncpy(name,"LOCAL FILE",16);
    strncpy(addr,tineHomePath,48);
  }
  return numNameServers;
}
TINE_EXPORT int GetFECAddr(char *fecName,FecDataStruct *fecaddr)
{
  int cc;
  FecAddrStruct fecadr;
  if (fecaddr == NULL || fecName == NULL) return argument_list_error;
  if ((cc=GetFecAddr(fecName,&fecadr)) != 0) return cc;
  toFecData(fecaddr,&fecadr);
  return 0;
}
TINE_EXPORT int GetFecAddr(char *fecName,FecAddrStruct *fecaddr)
{
  int i;
  if (fecaddr == NULL || fecName == NULL) return argument_list_error;
  for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
  if (i == numFecTblEntries) return host_not_resolved;
  memcpy(fecaddr,&FecTbl[i],sizeof(FecAddrStruct));
  return 0;
}
TINE_EXPORT int GetServerAddr(char *context,char *expName,ExpDataStruct *srvaddr,FecAddrStruct *fecaddr)
{
  int sidx,fidx;
  if (expName == NULL) return argument_list_error;
  /* let this call ride with explicitly acquiring the SrvTbl Mutex */
  for (sidx=0; sidx<numSrvTblEntries && sidx <SrvTblSize; sidx++)
  {
    if (context != NULL && strlen(context) &&
        strnicmp(context,SrvTbl[sidx].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(expName,SrvTbl[sidx].ExportName,EXPORT_NAME_SIZE)) break;
  }
  if (sidx == numSrvTblEntries) 
  { /* not yet in cache, then force the issue : */
    initClient();
    fidx = locateExport(expName,context,NULL,&sidx);
    if (fidx < 0) return fidx == -ambiguous ? ambiguous : non_existent_elem;
  }
  else
  { /* in SrvTbl cache, then very strong probability in FecTbl cache */
    for (fidx=0; fidx<numFecTblEntries && strnicmp(SrvTbl[sidx].FecName,FecTbl[fidx].fecName,FEC_NAME_SIZE); fidx++);
    if (fidx == numFecTblEntries) return host_not_resolved;
  }
  if (srvaddr != NULL) memcpy(srvaddr,&SrvTbl[sidx],sizeof(ExpDataStruct));
  if (fecaddr != NULL) memcpy(fecaddr,&FecTbl[fidx],sizeof(FecAddrStruct));
  return 0;
}
TINE_EXPORT int GetServerHostAddress(const char *context,const char *expName,char *host,char *addr,char *fec)
{
  FecAddrStruct fecaddr;
  ExpDataStruct srvaddr;
  int cc = 0;

  if ((cc=GetServerAddr((char *)context,(char *)expName,&srvaddr,&fecaddr)) != 0) ccerr(not_registered);

  if (addr != NULL) sprintf(addr,"%.40s:%d",fecaddr.strAdr,fecaddr.portOffset);
  if (fec != NULL) strncpy(fec,fecaddr.fecName,16);
  if (host != NULL) cc = GetHostFromFecName(fecaddr.fecName,host);

err:
  return cc;
}
/**
 * \todo this is hopefully never used anywhere and can be removes and looks mostly useless!
 * the arguments passed will be overwritten by redirection info so they must be char arrays
 * and not e.g. "PETRA" etc.
 */
TINE_EXPORT int GetConnectionAddr(char *context,char *expName,char *devName,char *prpName)
{
  if (context == NULL || expName == NULL || devName == NULL || prpName == NULL) return 0;
  return redirectLink(context,expName,prpName,devName) ? CE_REDIRECTED : 0;
}
#endif

int prepIncomingAddress(int pktlen,char *addrstream,FecAddrStruct *fec,ExpDataStruct *srv)
{
  RPCFecStruct lfec;
  RPCDataStruct lsrv;
  FecDataStruct lfec4;
  if (fec == NULL) return argument_list_error;
  switch (pktlen)
  {
    case ENSADDRPKT4L:
#     ifdef INTEL
      if (srv != NULL)
      {
        memcpy(&lsrv,addrstream,RPCDATASTRUCT_SIZE);
        memcpy(addrstream,&addrstream[RPCDATASTRUCT_SIZE],RPCFECSTRUCT_SIZE);
      }
      memcpy(&lfec,addrstream,RPCFECSTRUCT_SIZE);
#     else
      if (srv != NULL)
      {
        _structSwap(RPCDataStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)&lsrv,RTOL);
        memcpy(addrstream,&addrstream[RPCDATASTRUCT_SIZE],RPCFECSTRUCT_SIZE);
      }
	    _structSwap(RPCFecStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)&lfec,RTOL);
#   endif
      memset(fec,0,sizeof(FecAddrStruct));
      strncpy(fec->fecName,lfec.fecName,FEC_NAME_SIZE);
      strncpy(fec->strAdr,lfec.IP,ADDR_SIZE);
      memcpy(fec->netAdr.ipv4,lfec.IPh_addr,4);
      fec->portOffset = (int)((lfec.Port[1] << 8) | lfec.Port[0]);
      if (fec->portOffset == 0) fec->portOffset = (int)lfec.portOffset;
      fec->inetProtocol = lfec.Prot;
      fec->tineProtocol = 0;
      if (srv != NULL)
      {
        memset(srv,0,sizeof(ExpDataStruct));
        strncpy(srv->FecName,lsrv.FecName,FEC_NAME_SIZE);
        strncpy(srv->SubSystem,lsrv.SubSystem,SUBSYSTEM_NAME_SHORTSIZE);
        strncpy(srv->ExportName,lsrv.ExportName,EXPORT_NAME_SHORTSIZE);
        strncpy(srv->EqmName,lsrv.EqmName,EQM_NAME_SHORTSIZE+1);
        strncpy(srv->EqmContext,lsrv.EqmContext,CONTEXT_NAME_SIZE);
      }
      break;
    case ENSADDRPKT4:
      if (srv != NULL)
      {
        memcpy(srv,addrstream,EXPDATASTRUCT_SIZE);
        memcpy(addrstream,&addrstream[EXPDATASTRUCT_SIZE],FECDATASTRUCT_SIZE);
      }
#     ifdef INTEL
      memcpy(&lfec4,addrstream,FECDATASTRUCT_SIZE);
#     else
      _structSwap(FDSStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)&lfec4,RTOL);
#     endif
      toFecAddr(fec,&lfec4);
      break;
    case ENSADDRPKT6:
      if (srv != NULL)
      {
        memcpy(srv,addrstream,EXPDATASTRUCT_SIZE);
        memcpy(addrstream,&addrstream[EXPDATASTRUCT_SIZE],FECADDRSTRUCT_SIZE);
      }
#     ifdef INTEL
      memcpy(fec,addrstream,FECADDRSTRUCT_SIZE);
#     else
      _structSwap(FASStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)fec,RTOL);
#     endif
      break;
  }
  return 0;
}
int queryNameServerForFec(char *fecName,FecAddrStruct *fec)
{
  DTYPE dout;
  int cc=0;

  if (fecName == NULL || fec == NULL) return argument_list_error; 
  initClient();
  if (WaitForSystemMutex(hEnsMutex,gSystemTick) != 0) return semaphore_busy;
  DTYPEZERO(dout);
  dout.dFormat = CF_STRUCT;
  isNameServerRequest = TRUE;
retryQuery:
  memset(fec,0,sizeof(FecAddrStruct));
  switch (gEnsAddrPktSize)
  {
    case ENSADDRPKT4L:
      dout.dArrayLength = RPCFECSTRUCT_SIZE;
      dout.data.bptr = fecAddressBuf;
      isNameServerRequest = TRUE;
      cc = tExecLink(ENS_SRV,ENS_TAG,"FEC",fecName,&dout,NULL,CA_READ,500);
      break;
    case ENSADDRPKT4:
      dout.dArrayLength = FECDATASTRUCT_SIZE;
      dout.data.bptr = fecAddressBuf;
      cc = tExecLink(ENS_SRV,ENS_TAG,"FEC",fecName,&dout,NULL,CA_READ,500);
      if (cc == illegal_read_write)
      {
        gEnsAddrPktSize = ENSADDRPKT4L;
        goto retryQuery;
      }
      break;
    case ENSADDRPKT6:
      dout.dArrayLength = FECADDRSTRUCT_SIZE;
      dout.data.bptr = fecAddressBuf;
      cc = tExecLink(ENS_SRV,ENS_TAG,"FEC",fecName,&dout,NULL,CA_READ,500);
      if (cc == illegal_data_size)
      {
        gEnsAddrPktSize = ENSADDRPKT4;
        goto retryQuery;
      }
      break;
  }
  isNameServerRequest = FALSE;      
  if (cc == 0)
  {
    prepIncomingAddress(gEnsAddrPktSize,(char *)fecAddressBuf,fec,NULL);
  }
  ReleaseSystemMutex(hEnsMutex); 
  return cc;
}
int queryNameServerForAddress(char *ctx,char *exp,FecAddrStruct *fec,ExpDataStruct *srv)
{
  DTYPE dout;
  int cc=0;

  if (exp == NULL || fec == NULL || srv == NULL) return argument_list_error; 
  if (gSystemRunningStandAlone && !gAllowNetworkAddressResolution) return not_accepted;
  if (WaitForSystemMutex(hEnsMutex,gSystemTick) != 0) return semaphore_busy;
  if (!stricmp(exp,ENS_SRV) && (strlen(ctx) == 0 || !stricmp(ctx,"SITE") || !stricmp(ctx,"SERVICE")))
  { /* an explicit call to the ENS got thru */
    memcpy(fec,&FecTbl[0],sizeof(FecAddrStruct));
    memcpy(srv,&SrvTbl[0],sizeof(ExpDataStruct));
  }
  DTYPEZERO(dout);
  dout.dFormat = CF_STRUCT;
retryQuery:
  dout.dArrayLength = gEnsAddrPktSize;
  dout.data.bptr = fecAddressBuf;
  isNameServerRequest = TRUE;
  cc = tExecLink(ENS_SRV,ENS_TAG,ctx,exp,&dout,NULL,CA_READ,500);
  switch (gEnsAddrPktSize)
  {
    case ENSADDRPKT4L:
      break;
    case ENSADDRPKT4:
      if (cc == illegal_read_write)
      {
        gEnsAddrPktSize = ENSADDRPKT4L;
        goto retryQuery;
      }
      break;
    case ENSADDRPKT6:
      if (cc == illegal_data_size)
      {
        gEnsAddrPktSize = ENSADDRPKT4;
        goto retryQuery;
      }
      break;
  }
  isNameServerRequest = FALSE;      
  if (cc == 0)
  {
    prepIncomingAddress(gEnsAddrPktSize,(char *)fecAddressBuf,fec,srv);
  }
  ReleaseSystemMutex(hEnsMutex); 
  return cc;
}

TINE_EXPORT FecDataStruct *GetFecDataFromCache(char *context,char *expName)
{
  static FecDataStruct lfec;
  FecAddrStruct *fec = GetFecAddrFromCache(context,expName);
  if (fec == NULL) return NULL;
  toFecData(&lfec,fec);
  return &lfec;
}
TINE_EXPORT FecAddrStruct *GetFecAddrFromCache(char *context,char *expName)
{
  int fecidx,srvidx;
  static FecAddrStruct fec;
  initClient();
  fecidx = locateExport(expName,context,NULL,&srvidx);
  if (fecidx < 0) return NULL;
  /* never return the original to the caller ! */
  memcpy(&fec,&FecTbl[fecidx],sizeof(FecAddrStruct));
  return &fec;
}
#ifndef SMALL_TINE_LIB
int addStandAloneEntryToAddressCache(char *ctxName,char *expName,char *eqmName,int *fecIdx,int *srvIdx)
{
  int i, cc=0;
  SCKADR sckadr;
  InitSystemMutexSet();
  if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return semaphore_error;
  if (srvIdx == NULL || fecIdx == NULL || ctxName == NULL || expName == NULL || eqmName == NULL)
    ccerr(argument_list_error);
  if (numSrvTblEntries == 0)
  { /* I don't have a configured ENS (GENS)! Leave room for them, because I'm probably trying to find it */
    numFecTblEntries++; numSrvTblEntries++; 
  }
  if (numFecTblEntries >= FecTblSize) ccerr(out_of_client_memory);
  if (numSrvTblEntries >= SrvTblSize) ccerr(out_of_client_memory);
  i = numFecTblEntries;
  strncpy(FecTbl[i].fecName,gFecName,FEC_NAME_SIZE);
  strncpy(FecTbl[i].strAdr,gLoopbackAddr,ADDR_SIZE_IPV6);
  FecTbl[i].portOffset = gPortOffset;
  FecTbl[i].tineProtocol = gTineProtocol;
  sckadr.ss_family = (short)gAfInet;
  setInAddrLoopback(&sckadr);
  memcpy(&FecTbl[i].netAdr,getInAddr(&sckadr),gInAdrLen);
  *fecIdx = i;
  i = numSrvTblEntries;	
  strncpy(SrvTbl[i].FecName,gFecName,FEC_NAME_SIZE);
  strncpy(SrvTbl[i].EqmName,eqmName,EQM_NAME_SHORTSIZE);
  strncpy(SrvTbl[i].ExportName,expName,EXPORT_NAME_SIZE);
  strncpy(SrvTbl[i].EqmContext,ctxName,CONTEXT_NAME_SIZE);
  *srvIdx = i;
  numFecTblEntries++; numSrvTblEntries++; 
  feclog("addStandAloneEntryToAddressCache: added /%.32s/%.32s (%.6s @ %.16s) to the FEC table",
    ctxName,expName,eqmName,gFecName);
err:
  ReleaseSystemMutex(hSrvTblMutex);
  if (cc) feclog("addStandAloneEntryToAddressCache: %s",cc2str(cc));
  return cc;
}
int addNetworkEntryToAddressCache(int *fecIdx,int *srvIdx)
{ /* don't need SrvTbl Mutex here */
  int i,cc=0;
  BYTE ipadr[16];
  if (numSrvTblEntries == 0) 
  { /* I don't have a configured ENS (GENS)! Leave room for them, because I'm probably trying to find it */
    numFecTblEntries++; numSrvTblEntries++; 
  }
  if (numFecTblEntries >= FecTblSize) ccerr(out_of_client_memory);
  if (numSrvTblEntries >= SrvTblSize) ccerr(out_of_client_memory);
  i = numFecTblEntries;
  strncpy(FecTbl[i].fecName,SRVFEC_NAME,FEC_NAME_SIZE);
  strncpy(FecTbl[i].strAdr,gtNetSrvAddr,ADDR_SIZE_IPV6);
  inet_pton(gAfInet,gtNetSrvAddr,ipadr);
  memcpy(&FecTbl[i].netAdr,&tNetSrvAddr,sizeof(INADR));
  if (fecIdx != NULL) *fecIdx = i;
  i = numSrvTblEntries;	
  strncpy(SrvTbl[i].FecName,SRVFEC_NAME,FEC_NAME_SIZE);
  strncpy(SrvTbl[i].ExportName,SRVEXP_NAME,EXPORT_NAME_SIZE);
  strncpy(SrvTbl[i].EqmName,SRVEQM_NAME,EQM_NAME_SHORTSIZE);
  if (srvIdx != NULL) *srvIdx = i;
  numFecTblEntries++; /* add "NETWORK" */
  numSrvTblEntries++; 
  feclog("NETWORK: added network services to the FEC table");
  ccerr(0); /* jump out & return 0 */
err:
  if (cc) feclog("addNetworkEntryToAddressCache: %.32s",cc2str(cc));
  return cc;
}
#endif /* !SMALL_TINE_LIB */
/**
 * \internal
 * \brief Retrieve FEC data from name server or the local cache.
 *
 * This function first attempts to find the requested server name in its local
 * address cache.  If an complete entry is found (including transport protocol)
 * the entry's index is returned.
 * If not, the function requests the FEC data from the name server. 
 * If multiple name servers are known, it takes the first one which returns a result.
 * If the FEC is already buffered locally in FecTbl, the buffer is overweritten. 
 * Otherwise a new entry is created. If the FEC is already buffered locally in SrvTbl, 
 * the buffer is overwritten. Otherwise a ne entry is created.  
 * 
 * Special case: If the supplied export name equals "NETWORK", this indicates a (multicast)
 * request for the targeted server to respond directly.  This is for cases where the
 * ENS address is not known (diskless server, with no apriori knowledge of the ENS).
 * Usually, this technique is used to find the ENS itself.  Once known, address resolution
 * can proceed normally.  Note that server must be configured to respond to NETWORK
 * address requests, and that the ENS is usually the only server prepared to do so.
 * Also note that there are in such cases potentially two new entries in FecTbl as well as in SrvTbl. 
 * If the initially request is indeed to find the ENS, then the special slot 0 is filled with the
 * returned address.  Irregardless of this, the next available slot is filled with the full address
 * information of the responding server.
 * 
 * \param context EqmContext 
 * \param expName Export name
 * \param fecName Returns FEC name or empty string. 
 * \param srvidx  Not used if set to NULL. Otherwise, either returns the index of the corresponding 
 * element in SrvTbl[] or -1. 
 * \return positive fec table index or negative of tine error code
 * 
 * \see SrvTbl
 * \see FecTbl
 * 
 * \todo context and expName should be const (not changed by this function).
 * \todo FecTbl has limited memory & is never cleaned => may reach limit if many FECs are used.
 * \todo Same for SrvTbl.
 */
int GetAddressFromNameServer(char *context,char *expName,char *fecName,int *srvidx)
{
# define ccout(o) { cc = o; goto out; }
  int retry = 1;  /* Jump-to-err-then-jump-back-counter; max.number of backjumps for the whole loop */
  int n,i=0,cc=0,sid=-1,fid=0;
  int isNew=FALSE,isServiceRequest=FALSE,incrSrvEntries=FALSE,incrFecEntries=FALSE,loglvl=TINE_LOGLEVEL_INFO;
  char cntxt[CONTEXT_NAME_SIZE+4] = "";
  FecAddrStruct fec;
  ExpDataStruct srv;

  if (ServerExitCondition > 2) return -not_running;
  if (srvidx != NULL) *srvidx = -1;
  if (!strcmp(expName,SRVEXP_NAME)) isServiceRequest = TRUE;
  if (gSystemRunningStandAlone && !gAllowNetworkAddressResolution) return -not_accepted;
  if (!NameServerLoaded && !isServiceRequest && !gSystemRunningStandAlone) return -not_ready;
  if (expName == NULL || strlen(expName) == 0) return -argument_list_error;

  if ((!stricmp(expName,ENS_SRV) || !strcmp(expName,ENS_TAG)) && 
      (context == NULL || strlen(context) == 0 || !stricmp(context,"SITE") || !stricmp(context,"SERVICE")))
  { /* explicit call to ENS */
    if (gSystemRunningStandAlone) return -not_accepted;
    if (srvidx != NULL) *srvidx = 0;
    return 0;
  }
  if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return -semaphore_error;  
  /* Search SrvTbl for matching context + export name  */
  for (i = 0; i < numSrvTblEntries && i < SrvTblSize; i++)
  { /* if context is given, then it has to match; otherwise ignore it */
    if (context != NULL && strlen(context) &&
      strnicmp(context, SrvTbl[i].EqmContext, CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(expName, SrvTbl[i].ExportName, EXPORT_NAME_SIZE)) break;
  }
  sid = i; /* mark the server index */
  if (sid < numSrvTblEntries)
  { /* EQM Entry found: Search FecTbl for matching FEC name */
    for (i = 0; i < numFecTblEntries && strnicmp(SrvTbl[sid].FecName, FecTbl[i].fecName, FEC_NAME_SIZE); i++);
    fid = i; /* mark the fec index: < numFecTblEntries means FEC entry found! */
    if (srvidx != NULL) *srvidx = sid;
    /* Jump out if FEC entry is complete (has protocol information) */
    if (fid < numFecTblEntries && FecTbl[fid].inetProtocol) ccout(0); /* return cached entry */
  }
# ifndef SMALL_TINE_LIB
  if (isServiceRequest)
  { /* Asking the "NETWORK" for the address */
    if (tineDebug) dbglog("using network services as FEC address");
    ccout(addNetworkEntryToAddressCache(&fid, srvidx));
  }
# endif
  if (!NameServerLoaded) ccout(not_accepted);
  for (n=0; n<numNameServers; n++)
  { /* loop over known name servers until address is resolved */
retryRequest:
    if (fecName != NULL) fecName[0] = 0; /* overwrite fecName with null string */
    if (gSystemRunningStandAlone) ccerr(not_accepted);
    /* Buffer context */
    if (context != NULL) strncpy(cntxt,context,CONTEXT_NAME_SIZE);      
    /* Replace context with "#2" if empty. Legacy ENS request (#2 => return both FEC and EQM addresses) */
    if (strlen(cntxt) == 0) sprintf(cntxt,"#2");
          
    cc = queryNameServerForAddress(cntxt,expName,&fec,&srv);

    /* jump to err if not successful */
    if (cc) ccerr(cc);
    /* If no FecTbl in cache, create a new one */
    if (sid == numSrvTblEntries)
    {
      if (numSrvTblEntries < SrvTblSize) 
      {
        incrSrvEntries = TRUE;
        isNew = TRUE;
      }
      else 
      {
        msglog(resources_exhausted,"cannot expand server entry table: capacity reached (%d)",SrvTblSize);
        ccerr(resources_exhausted);
      }
    }
    /* look again for FEC name in FecTbl[] */
    for (i=0; i<numFecTblEntries && strnicmp(srv.FecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
    fid = i; /* mark the fec index */
    /* no entry found => create a new one. */
    if (fid == numFecTblEntries)
    {
      if (numFecTblEntries < FecTblSize) 
      {
        incrFecEntries = TRUE;
        isNew = TRUE;
      }
      else 
      {
        msglog(resources_exhausted,"cannot expand FEC entry table: capacity reached (%d)",FecTblSize);
        ccerr(resources_exhausted);
      }
    }      
    if (WaitForSystemMutex(hMfMutex,-1) == 0)
    { /* fill FecTbl entry with data */
      if (incrSrvEntries) sid = numSrvTblEntries; /* just in case */
      if (incrFecEntries) fid = numFecTblEntries; /* just in case */
      memcpy(&SrvTbl[sid],&srv,sizeof(ExpDataStruct));
      if (context != NULL && strnicmp(context,SrvTbl[sid].EqmContext,CONTEXT_NAME_SIZE) != 0)
      { /* context was apparently decorated */
        strncpy(SrvTbl[sid].EqmContext,context,CONTEXT_NAME_SIZE);
      }
      if (srvidx != NULL) *srvidx = sid;
      /* fid must be > 0! Do not under any circumstances overwrite the ENS or GENS addresses! */
      if (fid > 1) memcpy(&FecTbl[fid],&fec,sizeof(FecAddrStruct));
      if (fecName != NULL) strncpy(fecName,FecTbl[fid].fecName,FEC_NAME_SIZE);
      if (isNew) 
      { /* now increment and log */
        if (incrSrvEntries) numSrvTblEntries++;
        if (incrFecEntries) numFecTblEntries++;
        feclog("ENS: %.32s [FEC %.16s, EQM %.6s] added",SrvTbl[sid].ExportName,SrvTbl[sid].FecName,SrvTbl[sid].EqmName);
        msglog(0,"ENS: %.32s [FEC %.16s, EQM %.6s] added",SrvTbl[sid].ExportName,SrvTbl[sid].FecName,SrvTbl[sid].EqmName);
#       if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
        if (!isAddressInCacheFile(gAddrCacheFilePath,sid,fid)) addAddressToCacheFile(gAddrCacheFilePath,sid,fid);
#       endif
      }
      ReleaseSystemMutex(hMfMutex);
    }
err:
    if (isTimeout(cc) && retry-- > 0) 
    { /* jump back */
      goto retryRequest;  
    }
    if (cc == 0 || cc == not_accepted) 
    { /* finished : jump out */
      ens_errcnt = 0;
      break; 
    }
    loglvl = (cc) ? TINE_LOGLEVEL_ERROR : TINE_LOGLEVEL_INFO;
    if (!stricmp(expName,"CAS") || !stricmp(expName,"CYCLER")) loglvl = TINE_LOGLEVEL_INFO;
    feclogEx(loglvl,"ENS: /%.32s/%.32s [%.16s] : %.32s",
      context != NULL ? context : "",
	    expName != NULL ? expName : "",
      fecName != NULL ? fecName : "",cc2str(cc));
    if (++ens_errcnt < 5) if (ToggleNameServer() == 0) ens_errcnt = 5;
  }
out:
  ReleaseSystemMutex(hSrvTblMutex);
  /* fid is the found/new index in FecTbl[] */
  return cc ? -cc : fid; /* return FEC index */
}
#if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
char gAddrCacheFilePath[128];
char *getTineCacheRoot(void)
{
  char *dptr;
# if defined(WIN32) || defined(DOS)
  if ((dptr=getenv("TINE_CACHE")) != NULL) return dptr;
  if ((dptr=getenv("SystemDrive")) == NULL) dptr = "C:";
# elif defined(UNIX) || defined(MACOS)
  DIR *dp;
  if ((dptr=getenv("TINE_CACHE")) != NULL) return dptr;
  dptr = "/var/lib/tine";
  if ((dp=opendir(dptr)) != NULL) 
  { /* exists ! */
    closedir(dp);
    return "/var/lib";
  }
  dptr = "/var/tmp";
# else
  if ((dptr=getenv("TINE_CACHE")) != NULL) return dptr;
  dptr = ".";
# endif
  return dptr;
}
void establishAddrCacheFilePath(void)
{
  static int done = FALSE;
# ifdef FS_DELIMITER
  char *dptr,fpath[128];
  if (done) return;
  dptr = getTineCacheRoot();
  sprintf(fpath,"%s%ctine%ccache%c",dptr,FS_DELIMITER,FS_DELIMITER,FS_DELIMITER);
  if (makeDirectoryPath(fpath,TRUE)) strncpy(gAddrCacheFilePath,fpath,128);
  else strncpy(gAddrCacheFilePath,dptr,128);
# endif
  done = TRUE;
  return;
}
int isInSrvFile(char *filepath,int srvidx)
{
  FILE *fp = NULL;
  int cc = 0,done = 0,found = FALSE;
  int nam_col=-1,fec_col=-1,eqm_col=-1,ctx_col=-1;
  char filename[256];
  char str[256],hdr[256];
  ExpDataStruct *srv;

  if (WaitForSystemMutex(hDbMutex,-1) != 0) return found;
  if (filepath == NULL) ccerr(argument_list_error);
  if (srvidx < 0 || srvidx >= numSrvTblEntries) ccerr(out_of_range);
  srv = &SrvTbl[srvidx];
  sprintf(filename,"%.240s%s",filepath,EQPDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* scan file */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((ctx_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (csvnamcmp(srv->ExportName,str,nam_col,EXPORT_NAME_SIZE)) continue;
    if (csvnamcmp(srv->EqmContext,str,ctx_col,CONTEXT_NAME_SIZE)) continue;
    if (csvnamcmp(srv->EqmName,str,eqm_col,EQM_NAME_SHORTSIZE)) continue;
    if (csvnamcmp(srv->FecName,str,fec_col,FEC_NAME_SIZE)) continue;
    found = TRUE;
    break;
  }
  fclose(fp); fp = NULL;
err:
  ReleaseSystemMutex(hDbMutex);
  if (cc != 0 && tineDebug) dbglog("check address in file : %.32s",cc2str(cc));
  return found;
}
int isInFecFile(char *filepath,int fecidx)
{
  FILE *fp = NULL;
  int poff = -1,prot = 5,rule=0,cc = 0,done = 0,found = FALSE;
  int fec_col=-1,ip_col=-1,port_col=-1,prot_col=-1,rule_col=-1;
  char filename[256];
  char str[256],hdr[256],scratch[64];
  FecAddrStruct *fec;

  if (WaitForSystemMutex(hDbMutex,-1) != 0) return found;
  if (filepath == NULL) ccerr(argument_list_error);
  if (fecidx < 0 || fecidx >= numFecTblEntries) ccerr(out_of_range);
  fec = &FecTbl[fecidx];
  sprintf(filename,"%.240s%s",filepath,FECDB_FILE);
  if ((fp=fopen(filename,"r")) == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* scan file */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((ip_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      if ((port_col=findcol(hdr,"PORT_OFFSET")) < 0) ccerr(no_such_column);
      if ((prot_col=findcol(hdr,"TINE_PROTOCOL")) < 0) ccerr(0); /* return FALSE: causes update */
      if ((rule_col=findcol(hdr,"MCAST_RULES")) < 0)  ccerr(0); /* return FALSE causes update */
      done = 1;
      continue;
    }
    if (csvnamcmp(fec->fecName,str,fec_col,FEC_NAME_SIZE)) continue;
    if (csvnamcmp(fec->strAdr,str,ip_col,ADDR_SIZE_IPV6)) continue;
    getColumnStr(port_col,str,scratch,16);
    poff = atoi(scratch);
    if (poff != fec->portOffset) continue;
    getColumnStr(prot_col,str,scratch,16);
    prot = atoi(scratch);
    if (prot != fec->tineProtocol) break;
    getColumnStr(rule_col,str,scratch,16);
    if (!stricmp(scratch,"MODERN")) rule = ADDRFLAG_MCAST_MODERN; else rule = atoi(scratch);
    if ((rule & fec->addrFlags) == 0) break;
    found = TRUE;
    break;
  }
err:
  ReleaseSystemMutex(hDbMutex);
  if (fp) fclose(fp);
  if (cc != 0 && tineDebug) dbglog("check address in file : %.32s",cc2str(cc));
  return found;
}
int isAddressInCacheFile(char *filepath,int srvidx,int fecidx)
{
  if (gIsEquipmentNameServer) return FALSE; /* no sense in this */
  if (filepath == NULL) return FALSE;
  /* check names and addresses from the local cache entries for srvidx and fecidx ! */
  if (srvidx < 0 || srvidx >= numSrvTblEntries) return FALSE;
  if (fecidx < 0 || fecidx >= numFecTblEntries) return FALSE;
  if (!isInSrvFile(filepath,srvidx)) return FALSE;
  if (!isInFecFile(filepath,fecidx)) return FALSE;
  /* if we made it this far, it's in the cache ! */
  return TRUE;
}
static char fecdbhdr[] = "FecName,IpxAddr,MacAddr,IPaddr,PortOffset,TineProtocol,McastRules\n";
static char eqpdbhdr[] = "Name,FecName,EqpModule,Context,SubSystem\n";
void writeLineToCacheFile(FILE *fp,int fecidx)
{
  char ipxaddr[32],macaddr[32],ipaddr[48];
  int rules;
  if (FecTbl[fecidx].inetProtocol == IPX)
  {
    IpxAdr ipx = FecTbl[fecidx].netAdr.ipx;
    sprintf(ipxaddr,"%02x%02x%02x%02x",
      ipx.IpxNet[0],ipx.IpxNet[1],ipx.IpxNet[2],ipx.IpxNet[3]);
    sprintf(macaddr,"%02x%02x%02x%02x%02x%02x",
      ipx.IpxNode[0],ipx.IpxNode[1],ipx.IpxNode[2],ipx.IpxNode[3],ipx.IpxNode[4],ipx.IpxNode[5]);
    strcpy(ipaddr,"0"); 
  }
  else
  {
    strcpy(ipxaddr,"0"); strcpy(macaddr,"0");
    if (!zerocmp(FecTbl[fecidx].lclAdr.mac,6)) BytesToHex(FecTbl[fecidx].lclAdr.mac,macaddr,6);
    strncpy(ipaddr,FecTbl[fecidx].strAdr,ADDR_SIZE_IPV6);
  }
  rules = (FecTbl[fecidx].addrFlags & ADDRFLAG_MCAST_MODERN) ? 1 : 0;
  fprintf(fp,"%.16s,%s,%s,%.40s,%d,%d,%d\n",
        FecTbl[fecidx].fecName,ipxaddr,macaddr,ipaddr,
        FecTbl[fecidx].portOffset,FecTbl[fecidx].tineProtocol,rules);
}
extern int csvColumnCorrection;
int addAddressToCacheFile(char *filepath,int srvidx,int fecidx)
{
  FILE *fp = NULL, *fpb = NULL;
  int cc = 0,done = 0,found = FALSE;
  int nam_col=-1,fec_col=-1,eqm_col=-1,ctx_col=-1,ip_col=-1,port_col=-1,prot_col=-1,rule_col=-1;
  int oldmask;
  struct stat sbuf;
  char filename[256],filetemp[256];
  char str[256],hdr[256];
  char *expName,*ctxName,*fecName;

  if (WaitForSystemMutex(hDbMutex,-1) != 0) return mutex_error;
  oldmask = umask(0);
  if (filepath == NULL) ccerr(argument_list_error);
  if (srvidx < 0 || srvidx >= numSrvTblEntries) ccerr(invalid_index);
  if (fecidx < 0 || fecidx >= numFecTblEntries) ccerr(invalid_index);
  ctxName = SrvTbl[srvidx].EqmContext;
  expName = SrvTbl[srvidx].ExportName;
  fecName = SrvTbl[srvidx].FecName;
  if (strcmp(fecName,"FECSINK") == 0) ccerr(0); /* don't put FECSINK entries in local cache */
  sprintf(filetemp,"%.240s%s",filepath,EQPDB_FILE_TEMP);
  if (stat(filetemp,&sbuf) == 0)
  { /* the temp file exists! (other processes updating it?) */
    if (sbuf.st_mtime > (time(NULL) - 300))
    { /* file more recent than the last 5 minutes -> some other client app is updating */
      ccerr(operation_busy);
    }
    /* else: must be something left over from a crash */
  }
  if ((fpb=fopen(filetemp,"wt")) == NULL) ccerr(file_error);
  sprintf(filename,"%.240s%s",filepath,EQPDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) 
  { /* no cache file found ! */
    fprintf(fpb,"%s",eqpdbhdr);
    goto appendSrvDbFile;
  }
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* scan file */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((ctx_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      fprintf(fpb,"%s\n",hdr);
      continue;
    }
    if (!csvnamcmp(expName,str,nam_col,EXPORT_NAME_SIZE) && 
        !csvnamcmp(ctxName,str,ctx_col,CONTEXT_NAME_SIZE)) 
    { /* found the reference, now update the particulars ... */
      found = TRUE;
      fprintf(fpb,"%.32s,%.16s,%.6s,%.32s,%.16s\n",
        SrvTbl[srvidx].ExportName,SrvTbl[srvidx].FecName,SrvTbl[srvidx].EqmName,
        SrvTbl[srvidx].EqmContext,SrvTbl[srvidx].SubSystem); 
    }
    else
    { /* that wasn't it, spit the line back out */
      fprintf(fpb,"%s\n",str);
    }
  }
  fclose(fp); fp = NULL;
appendSrvDbFile:
  if (!found)
  { /* not found : add to file cache */
    fprintf(fpb,"%.32s,%.16s,%.6s,%.32s,%.16s\n",
      SrvTbl[srvidx].ExportName,SrvTbl[srvidx].FecName,SrvTbl[srvidx].EqmName,
      SrvTbl[srvidx].EqmContext,SrvTbl[srvidx].SubSystem); 
  }
  fclose(fpb); fpb = NULL;
  unlink(filename);
  rename(filetemp,filename);
# ifdef UNIX
  chmod(filename, S_ALLWRITE);
# endif
  found = FALSE;
  sprintf(filetemp,"%.240s%s",filepath,FECDB_FILE_TEMP);
  if ((fpb=fopen(filetemp,"wt")) == NULL) ccerr(file_error);
  sprintf(filename,"%s%s",filepath,FECDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) 
  { /* no cache file found ! */
    fprintf(fpb,"%s",fecdbhdr);
    goto appendFecDbFile;
  }
  for (done=0; csvGetLine(str,256,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((ip_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      if ((port_col=findcol(hdr,"PORT_OFFSET")) < 0) ccerr(no_such_column);
      prot_col = findcol(hdr,"TINE_PROTOCOL");
      rule_col = findcol(hdr,"MCAST_RULES");
      done = 1;
      fprintf(fpb,"%s",fecdbhdr);
      continue;
    }
    if (!csvnamcmp(fecName,str,fec_col,FEC_NAME_SIZE)) 
    { /* found it, so update it ! */
      found = TRUE;
      writeLineToCacheFile(fpb,fecidx);
    }
    else
    { /* that wasn't it, so spit the line back out */
      if (csvColumnCorrection == string_too_long) 
      { /* correct a legacy formatting problem */
        writeLineToCacheFile(fpb,fecidx);
      }
      else
      {
        fprintf(fpb,"%s",str);
        if (prot_col < 0) fprintf(fpb,",5"); /* worst case */
        if (rule_col < 0) fprintf(fpb,",0"); 
        fprintf(fpb,"\n");
      }
    }
  }
  if (fp) fclose(fp); 
  fp = NULL;
appendFecDbFile:
  if (!found)
  { /* not found : add to file cache */
    writeLineToCacheFile(fpb,fecidx);
  }
  fclose(fpb); 
  fpb = NULL;
  unlink(filename);
  rename(filetemp,filename);
# ifdef UNIX
  chmod(filename, S_ALLWRITE);
# endif
err:
  if (fp) fclose(fp);
  if (fpb) fclose(fpb);
  oldmask = umask(oldmask);
  ReleaseSystemMutex(hDbMutex);
  return cc;
}
GrpCacheItem *gGrpCacheList = NULL;
time_t gLastGroupCacheAddTime = 0;
static int nGrpCacheListItems = 0;
void addToGroupCacheList(const char *context,const char *group,const char *server,const char *device)
{
  GrpCacheItem *gci=NULL;
  int cc=0;
  if (gIsEquipmentNameServer || gIsGroupNameServer) return;
  if (WaitForSystemMutex(hDbMutex,gSystemTick) != 0) return;
  if ((gci=(GrpCacheItem *)SystemCalloc(1,sizeof(GrpCacheItem))) == NULL) ccerr(out_of_local_memory);
  strncpy(gci->ctx,context,CONTEXT_NAME_SIZE);
  strncpy(gci->grp,group,EXPORT_NAME_SIZE);
  strncpy(gci->srv,server,EXPORT_NAME_SIZE);
  strncpy(gci->dev,device,DEVICE_NAME_SIZE);
  gci->nxt = gGrpCacheList;
  gGrpCacheList = gci;
  gLastGroupCacheAddTime = time(NULL);
  nGrpCacheListItems++;
err:
  if (cc != 0) dbglog("cannot add item to group cache list : %.32s",cc2str(cc));
  ReleaseSystemMutex(hDbMutex);
  return;
}
void flushGroupCacheList(int nitems)
{
  int i=0;
  GrpCacheItem *gci=NULL;
  if (time(NULL) < gLastGroupCacheAddTime + 5) return;
  if (WaitForSystemMutex(hDbMutex,gSystemTick) != 0) return;
  if (nGrpCacheListItems > 0 && tineDebug > 2) 
    dbglog("need to remove %d items from the group cache list",nGrpCacheListItems);
  for (gci=gGrpCacheList; gci != NULL; gci=gGrpCacheList)
  {
    addServerToGroupCacheFile(gAddrCacheFilePath,gci->ctx,gci->grp,gci->srv);
    addDeviceToMemberCacheFile(gAddrCacheFilePath,gci->ctx,gci->srv,gci->dev);
    gGrpCacheList = gci->nxt;
    SystemFree(gci);
    nGrpCacheListItems--;
    i++;
    if (nitems > 0 && i > nitems) break;
  }
  if (gGrpCacheList == NULL)
  {
    if (nGrpCacheListItems != 0) msglog(0,"correct group cache list size from %d to 0",nGrpCacheListItems);
    nGrpCacheListItems = 0;
  }
  if (i > 0) msglog(1,"flushed %d items to group cache list",i);
  ReleaseSystemMutex(hDbMutex);
  return;
}
int addToGroupCacheFile(char *filename,char *column,char *tgt,int tgtsiz)
{
  FILE *fp = NULL, *fpb = NULL;
  int cc = 0,done = 0,found = FALSE;
  int col=-1;
  int oldmask;
  struct stat sbuf;
  char fn[256],tmp[256];
  char *c,str[256],hdr[256];

  if (WaitForSystemMutex(hDbMutex,gSystemTick) != 0) return mutex_error;
  oldmask = umask(0);

  sprintf(tmp,"%.250s.tmp",filename);
  sprintf(fn,"%.250s.csv",filename);
  if (stat(fn,&sbuf) != 0)
  { /* primary group cache file does not yet exist ! */
    strncpy(str,fn,256);
    if ((c=strrchr(str,FS_DELIMITER)) != NULL) *c = 0;
    makeDirectoryPath(str,TRUE);
  }
  if (stat(tmp,&sbuf) == 0)
  { /* the temp file exists! (other processes updating it?) */
    if (sbuf.st_mtime > (time(NULL) - 300))
    { /* file more recent than the last 5 minutes -> some other client app is updating */
      ccerr(operation_busy);
    }
    /* else: must be something left over from a crash */
  }
  if ((fpb=fopen(tmp,"wt")) == NULL) ccerr(file_error);

  if ((fp=fopen(fn,"rt")) == NULL) 
  { /* no cache file found ! */
    fprintf(fpb,"%s\n",column);
    goto appendGrpDbFile;
  }
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* scan file */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((col=findcol(hdr,column)) < 0) ccerr(no_such_column);
      done = 1;
      fprintf(fpb,"%s\n",hdr);
      continue;
    }
    if (!csvnamcmp(tgt,str,col,tgtsiz)) found = TRUE;
    /* always spit the line back out */
    fprintf(fpb,"%s\n",str);
  }
  fclose(fp); fp = NULL;
appendGrpDbFile:
  if (!found)
  { /* not found : add to file cache */
    fprintf(fpb,"%s\n",tgt);
  }
  fclose(fpb); fpb = NULL;
  if (found)
  { /* just remove the temp file */
    unlink(tmp);
  }
  else
  { /* do the shuffle ... */
    unlink(fn);
    rename(tmp,fn);
  }
err:
  if (fp) fclose(fp);
  if (fpb) fclose(fpb);
  oldmask = umask(oldmask);
  ReleaseSystemMutex(hDbMutex);
  return cc;
}
int addServerToGroupCacheFile(char *filepath,char *ctx,char *grp,char *srv)
{
  char filename[256];
  sprintf(filename,"%.128sGROUPS%c%.32s%c%.32s",filepath,FS_DELIMITER,ctx,FS_DELIMITER,grp);
  return addToGroupCacheFile(filename,"Members",srv,EXPORT_NAME_SIZE);
}
int addDeviceToMemberCacheFile(char *filepath,char *ctx,char *srv,char *dev)
{
  char filename[256];
  sprintf(filename,"%.128sGROUPS%c%.32s%c%.32s%cdevices",filepath,FS_DELIMITER,ctx,FS_DELIMITER,srv,FS_DELIMITER);
  return addToGroupCacheFile(filename,"Devices",dev,DEVICE_NAME_SIZE);
}
#endif /* !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB) */
#ifndef SMALL_TINE_LIB
/**
 * \internal
 * \brief Get Group members from the file system.
 * 
 * \note arguments never NULL (no reason to check) 
 * \note used by GetAddressFromNameFile()
 * 
 * \param filepath is the file path
 * \param ctx the context
 * \param srv the server name 
 * \param memLst the resulting member list
 * \param lstSize the list size
 *
 */
int GetGroupMembersFromFile(const char *filepath,const char *ctx,const char *srv,NAME64 **memLst,int *lstSize)
{ 
# ifdef FS_DELIMITER
  FILE *fp = NULL;
  char filename[256];
  char str[256],hdr[256];
  int i = 0,cc = 0,done = FALSE,mem_col=-1,siz=0;
  NAME64 *lst = *memLst;

  sprintf(filename,"%.128sGROUPS%c%.32s%c%.32s.csv",filepath,FS_DELIMITER,ctx,FS_DELIMITER,srv);
  if ((siz=GetDataBaseSize(filename)) < 0) ccerr(-siz);
  if (siz > *lstSize)
  {
    SystemFree(lst);
    if ((lst=(NAME64 *)SystemCalloc(siz,sizeof(NAME64))) == NULL) ccerr(out_of_local_memory);
  }
  *lstSize = siz;
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* always scan the file for the entry */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    { /* Read header. */
      strncpy(hdr,str,255);
      if ((mem_col=findcol(hdr,"MEMBERS")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    getColumnStr(mem_col,str,lst[i++].name,EXPORT_NAME_SIZE);
  }
err:
  if (fp) fclose(fp);
  if (cc) 
  {
    SystemFree(lst);
    *lstSize = 0; lst = NULL;
    feclog("group file %.256s : %.32s",filename,cc2str(cc));
  }
  *memLst = lst;
  return cc;
# else
  filepath = filepath; ctx = ctx; srv = srv; memLst = memLst; lstSize = lstSize;
  return no_such_file;
# endif
}
/**
 * \internal
 * \brief determine whether a group member has the device specified.
 * 
 * \note arguments never NULL (no reason to check) 
 * \note used by GetAddressFromNameFile()
 * 
 * \param filepath is the file path
 * \param ctx the context
 * \param mem the member server name
 * \param tgt the target device
 *
 */
int grpFileHasDevice(const char *filepath,const char *ctx,const char *mem,const char *tgt)
{
# ifdef FS_DELIMITER
  FILE *fp = NULL;
  char filename[256];
  char str[256],hdr[256],dev[DEVICE_NAME_SIZE];
  int cc = 0,done = FALSE,found = FALSE,dev_col = -1;

  sprintf(filename,"%.128sGROUPS%c%.32s%c%.32s%cdevices.csv",filepath,FS_DELIMITER,ctx,FS_DELIMITER,mem,FS_DELIMITER);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* always scan the file for the entry */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    { /* Read header. */
      strncpy(hdr,str,255);
      if ((dev_col=findcol(hdr,"DEVICES")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    getColumnStr(dev_col,str,dev,DEVICE_NAME_SIZE);
    if (strnicmp((char *)tgt,dev,DEVICE_NAME_SIZE) == 0) 
    {
      found = TRUE;
      break;
    }
  }
err:
  if (fp) fclose(fp);
  if (cc) found = FALSE;
  return found;
# else
  filepath = filepath; ctx = ctx; srv = srv; mem = mem; dev = dev;
  return FALSE;
# endif
}
#endif /* !SMALL_TINE_LIB */
#if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
#define structCacheFileHdr "Format, Size, Field"    
int writeStructCacheFile(const char *key,const char *tag)
{
  char fn[256], str[256], keybuf[70], fstr[64];
  char *ctx = keybuf, *srv;
  int i, nfields = 0, cc = 0;
  struct structStruct *ss;
  struct structFormat *sf, **sffa;
  FILE *fp = NULL;
  struct stat sbuf;

  if (key[0] != '/' || strchr(&key[1],'/') == NULL) return invalid_keyword;
  strncpy(ctx,&key[1],70);
  if ((srv=strchr(ctx,'/')) == NULL) return invalid_keyword;
  *srv++ = 0;
  if ((ss=findstructEx((char *)tag,(char *)key)) == NULL) return invalid_structure_tag;
  if (WaitForSystemMutex(hDbMutex,gSystemTick) != 0) return mutex_error;
  snprintf(fn,256,"%s%cSTRUCTS%c%.32s%c%.32s%c%.16s.csv",gAddrCacheFilePath,FS_DELIMITER,FS_DELIMITER,ctx,FS_DELIMITER,srv,FS_DELIMITER,tag);
  if (stat(fn,&sbuf) != 0)
  { /* cache file does not yet exist ! */
    char *c;
    strncpy(str,fn,256);
    if ((c=strrchr(str,FS_DELIMITER)) != NULL) *c = 0;
    makeDirectoryPath(str,TRUE);
  }
  else
  { /* if recently updated then don't bother re-writing ... */
    if (time(NULL) < sbuf.st_mtime + 3600) ccerr(file_already_exists);
  }
  if ((fp=fopen(fn,"wt")) == NULL) ccerr(file_error);
  for (sf=ss->s; sf != NULL; sf=sf->nxt) nfields++; /* count the fields */
  sffa = (struct structFormat **)alloca(nfields*sizeof(size_t));
  if (sffa == NULL) ccerr(out_of_local_memory);
  for (sf=ss->s,i=nfields-1; sf != NULL; sf=sf->nxt,i--)
  { /* the link list is reverse ordered => prepare the proper order */
    sffa[i] = sf;
  }  
  fprintf(fp,"%s\n",structCacheFileHdr);
  for (i=0; i<nfields; i++)
  {
    strncpy(fstr,GetFormatAsString(sffa[i]->fmt),16); fstr[16] = 0;
    if (isTaggableFormat((short)sffa[i]->fmt))
    { /* append the structure/bitfield tag to the format string */
      strcat(fstr,".");
      strncat(fstr,sffa[i]->stag,TAG_NAME_SIZE);
    }
    fprintf(fp,"%s, %d, %.16s\n",fstr,sffa[i]->siz,sffa[i]->field);
  }  
err:
  if (fp) fclose(fp);
  ReleaseSystemMutex(hDbMutex);
  return cc;  
}
typedef struct
{
  char fmtstr[64];
  char fld[16];
  int  siz;
} StructCacheEntry;
_CSVCOLUMN csvStructCacheFileColumns[] = 
{
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StructCacheEntry,fmtstr),NULL,0 },
  { "SIZE","",0,CF_INT32,1,(BYTE *)OFFSETIN(StructCacheEntry,siz),NULL,0 }, 
  { "FIELD","",0,CF_TEXT,16,(BYTE *)OFFSETIN(StructCacheEntry,fld),NULL,0 }
};
CSVDB csvStructCacheFileDb =
{
  csvStructCacheFileColumns,
  sizeof(csvStructCacheFileColumns)/sizeof(csvStructCacheFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(StructCacheEntry),
  0
};
int readStructCacheFile(const char *key, const char *tag, DTYPE *dout)
{
  char fn[256], keybuf[70], fstr[64];
  char *ctx = keybuf, *srv;
  int i, siz = 0, nfields, cc = 0;
  StructCacheEntry *sce = NULL;
  NAME64DBLDBL *ndd;
  if (key == NULL || tag == NULL || dout == NULL) return invalid_parameter;
  if (dout->dFormat != CF_NAME64DBLDBL) return illegal_format;
  if (key[0] != '/' || strchr(&key[1],'/') == NULL) return invalid_keyword;
  strncpy(ctx,&key[1],70);
  if ((srv=strchr(ctx,'/')) == NULL) return invalid_keyword;
  *srv++ = 0;
  snprintf(fn,256,"%s%cSTRUCTS%c%.32s%c%.32s%c%.16s.csv",gAddrCacheFilePath,FS_DELIMITER,FS_DELIMITER,ctx,FS_DELIMITER,srv,FS_DELIMITER,tag);
  if (WaitForSystemMutex(hDbMutex,gSystemTick) != 0) return mutex_error;
  if ((cc=csvReadFile(NULL,fn,&csvStructCacheFileDb,(void **)&sce)) != 0) ccerr(cc);
  nfields = csvStructCacheFileDb.siz;
  if (dout->dArrayLength < (UINT32)(nfields+1)) ccerr(buffer_too_small);
  ndd = (NAME64DBLDBL *)dout->data.vptr;
  for (i=0; i<nfields; i++)
  {
    ndd[i].d1val = (double)sce[i].siz;
    ndd[i].d2val = (double)GetFormatFromString(sce[i].fmtstr);
    if ((int)ndd[i].d2val == CF_STRUCT)
    {
      char *c;
      strncpy(fstr,sce[i].fmtstr,64);     
      if ((c=strchr(fstr,'.')) != NULL) c++; else c = fstr;
      snprintf(ndd[i].name,64,"<%.16s>%.16s",c,sce[i].fld);
    }
    else
    {
      strncpy(ndd[i].name,sce[i].fld,16);
    }
    siz += sce[i].siz * fmtsizeof((short)ndd[i].d2val);
  }
  strncpy(ndd[i].name,tag,16);
  ndd[i].d1val = siz;
  ndd[i].d2val = CF_NULL;
  dout->dArrayLength = nfields + 1;
err:
  SystemFree(sce);
  ReleaseSystemMutex(hDbMutex);
  return cc;  
}
#endif /* !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB) */
/**
 * \brief Get Data and FEC data from FEC and RPC-databases, or only FEC data.
 * 
 *  <ul>
 *  <li>Function 1. Find an entry in the local RPC Database file which corresponds to
 *  a given export name. Overwrite context and fecName. Store this entry in
 *  SrvTbl and write its index to srvidx. Then Function 2.</li>
 *  <li>Function 2. Find an entry in the local FEC Database file which corresponds to
 *  a given fec name.</li>
 *  </ul>
 *  
 * If expName is set to NULL, the function is used only as Function 2. 
 * In this case fecName is an input parameter. otherwise it is used as 
 * Function 1 and Function 2, with fecName as an output parameter.
 * 
 * \param context Function 1: context (output). Function 2: Not used.  
 * \param expName Decides if this function 1 and 2 or only function 2 is used. Function 1: export name. Function 2: NULL.
 * \param fecName Function 1: Fec name (output). Function 2: FEC name (input).
 * \param srvidx Function 1: Index of found entry in SrvTbl. Function 2: Not used.
 * 
 * \see gSrvDataBaseFullName
 * \see gFecDataBaseFullName
 * \see SrvTbl
 * \see FecTbl
 * .
 * \todo Remove gotos and jumps.
 */
int GetAddressFromNameFile(char *filepath,char *context,char *expName,char *fecName,char *devName,int *srvidx)
{
  FILE *fp = NULL;
  int i = -1,k,p,cc = 0,done = 0;
  int nam_col=-1,fec_col=-1,eqm_col=-1,cxt_col=-1,ipx_col=-1,mac_col=-1,csv_col=-1;
  int chkAmbiguous=FALSE, chkContext=FALSE, hasIncremented=FALSE;
  char filename[256];
  char str[256],hdr[256],scratch[64], notedContext[CONTEXT_NAME_SIZE+4], *contxt = "";
  char logtext[512];

  logtext[0] = 0;
  if (ServerExitCondition > 2) return -not_running;
  if (WaitForSystemMutex(hSrvTblMutex,-1) != 0) return -semaphore_error;  
  if (filepath == NULL) ccerr(argument_list_error);
  if (srvidx != NULL) *srvidx = -1;
  if (fecName == NULL) ccerr(non_existent_fec);
  if (expName == NULL) goto acquireFecName;     /* only interested in FEC address*/
  if (strlen(expName) == 0) ccerr(non_existent_elem);
  if (context != NULL && stricmp(context,"DEFAULT")) 
  {
    contxt = context;
    chkContext = TRUE;
  }
  if (gIsEquipmentNameServer) 
  { /* cut to the chase */
    for (i=0; i<numSrvTblEntries; i++)
    {
      if (strnicmp(expName,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) continue;
      if (!strncmp(expName,GLOBALSRV,EXPORT_NAME_SIZE) && strlen(contxt) == 0) break;
      if (strnicmp(contxt,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
      break;
    }
    if (i == numSrvTblEntries) ccerr(non_existent);
    if (srvidx != NULL) *srvidx = i;
    for (k=0; k<numFecTblEntries && strnicmp(SrvTbl[i].FecName,FecTbl[k].fecName,FEC_NAME_SIZE); k++);
    if (k == numFecTblEntries) ccerr(non_existent);
    ReleaseSystemMutex(hSrvTblMutex);
    return k;
  }
  sprintf(filename,"%.240s%s",filepath,EQPDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* always scan the file for the entry */
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    { /* Read header. */
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((cxt_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (!csvnamcmp(expName,str,nam_col,EXPORT_NAME_SIZE))
    { /* found entry in file */
      if (chkContext && csvnamcmp(contxt,str,cxt_col,CONTEXT_NAME_SIZE)) continue;
      if (chkAmbiguous)
      { /* if chkAmbiguous is TRUE then i and notedContext have been assigned */
        getColumnStr(cxt_col,str,scratch,CONTEXT_NAME_SIZE);
        if (strnicmp(notedContext,scratch,CONTEXT_NAME_SIZE))
        { /* same server different context */
          getColumnStr(fec_col,str,scratch,FEC_NAME_SIZE);
          if (strnicmp(SrvTbl[i].FecName,scratch,FEC_NAME_SIZE))
          { /* pointing to different FECs */
            if (hasIncremented) numSrvTblEntries--;
            ccerr(ambiguous);
          }
        }
        continue;
      }
      for (i=0; i<numSrvTblEntries; i++)
      {
        if (strnicmp(expName,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) continue;
        if (!strncmp(expName,GLOBALSRV,EXPORT_NAME_SIZE) && strlen(contxt) == 0) break;
        if (strnicmp(contxt,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
        break;
      }
      if (i == numSrvTblEntries)
      {
        hasIncremented = TRUE;
        if (numSrvTblEntries < SrvTblSize) numSrvTblEntries++;
        else ccerr(buffer_too_small);
      }
      getColumnStr(nam_col,str,SrvTbl[i].ExportName,EXPORT_NAME_SIZE);
      getColumnStr(fec_col,str,SrvTbl[i].FecName,FEC_NAME_SIZE);
      getColumnStr(eqm_col,str,SrvTbl[i].EqmName,EQM_NAME_SHORTSIZE);
      getColumnStr(cxt_col,str,scratch,CONTEXT_NAME_SIZE);
#     ifndef SMALL_TINE_LIB
      if (!gIsEquipmentNameServer && strcmp(SrvTbl[i].FecName,GENS_FEC) == 0 && devName != NULL)
      { /* find GENS info on the local file system */
        NAME64 *memLst = NULL;
        int lstSize = 0, k;
        cc = address_unknown;
        if (GetGroupMembersFromFile(filepath,context,SrvTbl[i].ExportName,&memLst,&lstSize) == 0)
        { /* found local group information */
          for (k=0; k<lstSize; k++)
          { /* scan thru the list */
            if (grpFileHasDevice(filepath,context,memLst[k].name,devName))
            { /* found a group member who has the target device */
              /* incorporate into the redirection tables */
              appendRedirectedLinkList(context,SrvTbl[i].ExportName,"*",devName,context,memLst[k].name,devName,"*",TRUE);
              i = GetAddressFromNameFile(filepath,context,memLst[k].name,fecName,NULL,srvidx);
              cc = (i < 0) ? -i : 0;
              break;
            }
          }
        }
        SystemFree(memLst);
        ccerr(cc);
      }
#     endif
      if (strlen(contxt) > 0)
      {
        if (strnicmp(contxt,scratch,CONTEXT_NAME_SIZE)) 
        { /* not a match */
          if (hasIncremented) numSrvTblEntries--;
          hasIncremented = FALSE;
          i = -1; 
          continue;
        }
        strncpy(SrvTbl[i].EqmContext,scratch,CONTEXT_NAME_SIZE);
      }
      else 
      {
        strncpy(notedContext,scratch,CONTEXT_NAME_SIZE);
        SrvTbl[i].EqmContext[0] = 0;
        chkAmbiguous = TRUE;
      }
      sprintf(logtext,"EFL: %.32s [FEC %.16s, EQM %.8s] added",
	      SrvTbl[i].ExportName,SrvTbl[i].FecName,SrvTbl[i].EqmName);
      if (srvidx != NULL) *srvidx = i;
      if (!chkAmbiguous) break;  /* finished if context given */
    }
  }
  if (i < 0) ccerr(non_existent);
  strncpy(fecName,SrvTbl[i].FecName,FEC_NAME_SIZE);
  fclose(fp); fp = NULL;

acquireFecName:
  i = -1; /* re-initialize negative */
  if (strlen(fecName) == 0) ccerr(non_existent_fec);
  sprintf(filename,"%.240s%s",filepath,FECDB_FILE);
  if ((fp=fopen(filename,"r")) == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,256,fp) != NULL;)
  { /* scan file */
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((csv_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (!csvnamcmp(fecName,str,csv_col,FEC_NAME_SIZE))           
    { /* found entry in file */
      for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].fecName,FEC_NAME_SIZE); i++);
      if (i == numFecTblEntries)
      {
        if (numFecTblEntries < FecTblSize) numFecTblEntries++;
        else ccerr(buffer_too_small);
      }
      getColumnStr(csv_col,str,FecTbl[i].fecName,FEC_NAME_SIZE);
      if ((csv_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,FecTbl[i].strAdr,ADDR_SIZE_IPV6);
      p = getProtocolFromAddressString(FecTbl[i].strAdr);
      if ((cc=fixInetProtocol(&FecTbl[i],p)) != 0) ccerr(cc);
      FecTbl[i].inetProtocol = p;
      if ((ipx_col=findcol(hdr,"IPX_ADDR")) < 0) ipx_col = findcol(hdr,"FEC_NETWORK");
      if ((mac_col=findcol(hdr,"MAC_ADDR")) < 0) mac_col = findcol(hdr,"FEC_NODE");
      if (p != UDP && p != UDP6 && ipx_col >= 0)
      {
        getColumnStr(ipx_col,str,scratch,16);
        HexToBytes(scratch,FecTbl[i].netAdr.ipx.IpxNet,8);
        if (mac_col < 0) ccerr(no_such_column);
        getColumnStr(mac_col,str,scratch,16);
        HexToBytes(scratch,FecTbl[i].netAdr.ipx.IpxNode,12);
        FecTbl[i].inetProtocol = IPX;
        sprintf(FecTbl[i].strAdr,"0x%.12s",scratch);
      }
      sprintf(logtext,"EFL: FEC %.16s [%.40s] added",FecTbl[i].fecName,FecTbl[i].strAdr);
      if ((csv_col=findcol(hdr,"PORT_OFFSET")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,16);
        FecTbl[i].portOffset = atoi(scratch);
      }
      else FecTbl[i].portOffset = 0;
      if ((csv_col=findcol(hdr,"TINE_PROTOCOL")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,16);
        FecTbl[i].tineProtocol = atoi(scratch);
      }
      else if ((csv_col=findcol(hdr,"VERSION")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,16);
        FecTbl[i].tineProtocol = atoi(scratch) == 4 ? 6 : 5;
      }
      else FecTbl[i].tineProtocol = 0;
      if ((csv_col=findcol(hdr,"MCAST_RULES")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,ADDR_SIZE);
        if (strlen(scratch) > 0)
        {
          p = !stricmp(scratch,"MODERN") ? ADDRFLAG_MCAST_MODERN : atoi(scratch);
          if (p != 0) FecTbl[i].addrFlags |= ADDRFLAG_MCAST_MODERN;
        }
      }
      break;
    }
  }
  if (i < 0) 
  {
    cc = non_existent_fec;
    sprintf(logtext,"EFL: FEC %.16s not found in local database %.32s",fecName,FECDB_FILE);
  }
err:
  if (fp) fclose(fp);
# if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
  if (cc && cc != no_such_file && strcmp(gAddrCacheFilePath,filepath) != 0)
  { /* don't record errors from the address file cache */
    if (expName == NULL) expName = "(null)";
    sprintf(logtext,"EFL: /%.32s/%.32s : %.32s",contxt,expName,cc2str(cc));
  }
# endif
  ReleaseSystemMutex(hSrvTblMutex);
  if (!gIgnoreVolatileAddressCache && strlen(logtext) > 0) feclogEx(cc?TINE_LOGLEVEL_WARN:TINE_LOGLEVEL_INFO, logtext);
  return cc ? -cc : i;
}

int dbaseloaded = 0;
int MaxTblAllocationSize = -1;
int reserveFecAddressSpace(void)
{
  int cc = 0;
  if (SrvTbl != NULL && FecTbl != NULL) 
  {
    /* already reserved */
    return 0;
  }
  if (SrvTblSize == 0) SrvTblSize = gConTblCapacity;
  if (FecTblSize == 0) FecTblSize = gConTblCapacity;
  if (MaxTblAllocationSize > 0) SrvTblSize = FecTblSize = MaxTblAllocationSize;

  if ((SrvTbl == NULL) && /* previously allocated ? */
     !(SrvTbl=(ExpDataStruct *)SystemCalloc(SrvTblSize,sizeof(ExpDataStruct))))
    ccerr(out_of_client_memory);

  if ((FecTbl == NULL) && /* previously allocated ? */
      !(FecTbl=(FecAddrStruct *)SystemCalloc(FecTblSize,sizeof(FecAddrStruct))))
    ccerr(out_of_client_memory);

  feclog("Reserved %d client connection link entries",SrvTblSize);
  feclog("Reserved %d client connection addr entries",FecTblSize);

err:
  return cc;
}
/**
 * \internal 
 * \brief Initializes different things relating to name resolution.
 * 
 * <ul>
 * <li>Initializes tineHomePath,gSrvDataBaseFullName,gFecDataBaseFullName</li>
 * <li>Calls GetAllocSizesFromNameServer, initializes FECAllocNum, SrvTblSize</li>
 * <li>Allocates SrvTbl and FecTbl</li>
 * <li>Initialize conTbl with name server entry.</li>
 * </ul>
 * 
 * \see tineHomePath,gSrvDataBaseFullName,gFecDataBaseFullName
 * \see dbaseloaded
 */
int initFecTable(void)
{
  char *ptr;
  int cc = 0;

  if (dbaseloaded) return 0;

  /* Determine gSrvDataBaseFullName and gFecDataBaseFullName from environment variables. */ 
  if (gIsEquipmentNameServer || gIsGroupNameServer)
  {
    feclog("Server is the ENS or GENS: ignore environment settings");
    tineHomePath[0] = 0;
  }
  else
  {
    fixTineHome();
  }
# ifndef SMALL_TINE_LIB
  if ((ptr=getenv("TINE_USE_LOOPBACK")) != NULL ||
      (ptr=getenv("USE_LOOPBACK")) != NULL)
  {
    gUseLoopback = strtobool(ptr);
  }
  if ((ptr=getenv("TINE_FORCE_SHARED_MEMORY")) != NULL ||
      (ptr=getenv("FORCE_SHARED_MEMORY")) != NULL)
  {
    gNoLoopback = strtobool(ptr);
  }
# endif
  if ((ptr=getenv("TINE_LINK_TRACE")) != NULL)
  {
    setTraceLinkKey(ptr);
  }
  if ((ptr=getenv("TINE_CATCH_BREAK")) != NULL)
  {
    gCatchConsoleBreak = strtobool(ptr);
  }  
  if ((cc=reserveFecAddressSpace()) != 0) ccerr(cc);
  numSrvTblEntries = numFecTblEntries = 0;
  if (!gEnsAddrResolutionInProgess) addNameServerToTables();
# if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
  establishAddrCacheFilePath();
# endif
  dbaseloaded = 1;
err:
  if (cc) nConnectionTableEntries = 0;
  return cc;
}

/**
 * \internal 
 * \brief Calls GetNameServerAddress() and Initializes NameServerData and 
 * connNameServer. Sets both output parameters to the gConTblCapacity.
 * 
 * Uses the global variable NameServerLoaded as a lock which ensures
 * that this function is called only once.
 * 
 * \see gConTblCapacity
 * \see NameServerData
 * \see connNameServer
 * \see GetNameServerAddress()
 * 
 */
int addNameServerToTables(void)
{
  int cc=0;

  InitSystemMutexSet();
  if (WaitForSystemMutex(hLinkTblMutex,-1) != 0) ccerr(semaphore_error);
  if (!NameServerLoaded)
  {
    memset(&connNameServer,0,sizeof(ConTblEntry));
    if ((cc=GetNameServerAddress()) != 0) ccerr(cc);
    strncpy(NameServerData.FecName,NameServer[0].fecName,FEC_NAME_SIZE);
    strncpy(NameServerData.ExportName,ENS_TAG,EXPORT_NAME_SIZE);
    strncpy(NameServerData.EqmName,ENS_TAG,EQM_NAME_SHORTSIZE);
    strncpy(connNameServer.expName,ENS_TAG,EXPORT_NAME_SIZE);
    strncpy(connNameServer.devName,"#0",DEVICE_NAME_SIZE);
    strncpy(connNameServer.prpName,"SIZES",PROPERTY_NAME_SIZE);
    connNameServer.formatOut = CF_INT16;
    connNameServer.formatIn = CF_NULL;
    connNameServer.srvIdx = 0;
    connNameServer.sizeOut = 2;
    connNameServer.sizeIn = 0;
    connNameServer.tineProtocol = gTineProtocol;
    connNameServer.mode = CM_REGISTER;
    connNameServer.blknumIn = 1;
    connNameServer.numblksIn = 1;
    connNameServer.nextDataPtr = 0;
    strcpy(connNameServer.key,"ENS");
    connNameServer.tineProtocol = gTineProtocol;
    connNameServer.inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP;
    conTbl[0] = &connNameServer; /* Hardwire first call */
    nConnectionTableEntries = 1;
    if (SrvTbl != NULL && FecTbl != NULL)
    {
      memcpy(&SrvTbl[0],&NameServerData,sizeof(ExpDataStruct));
      memcpy(&FecTbl[0],&NameServer[0],sizeof(FecAddrStruct));
      conTbl[0] = &connNameServer; /* Hardwire first call */
      /* add the GENS information as well ... */
      memcpy(&FecTbl[1],&NameServer[0],sizeof(FecAddrStruct));
      strcpy(FecTbl[1].fecName,GENS_FEC);
      FecTbl[1].portOffset += 101;
      memcpy(&SrvTbl[1],&NameServerData,sizeof(ExpDataStruct));
      strcpy(SrvTbl[1].FecName,GENS_FEC);
      strcpy(SrvTbl[1].EqmName,GENS_TAG);
      strcpy(SrvTbl[1].ExportName,GENS_SRV);
      numSrvTblEntries = numFecTblEntries = 2;
    }
    NameServerLoaded = 1;
  }
err:
  ReleaseSystemMutex(hLinkTblMutex);
  return cc;
}
