/*
 * File eventtools.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include <math.h>
#include "tine.h"
#include "eventtools.h"

void registerStructs(void);

EVNTOOLS_EXPORT int GetArchivedPropertyInformation(const char *context,const char *fullname,PrpQueryStruct *pqs)
{
  DTYPE din, dout;
  char dev[256];

  if (context == NULL || fullname == NULL || pqs == NULL) return argument_list_error;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));

  dout.dArrayLength = 1;
  dout.dFormat = CF_STRUCT;
  strncpy(dout.dTag,"PRPQSr4",TAG_NAME_SIZE);
  dout.data.vptr = (void *)pqs;
  din.dArrayLength = 0;
  din.dFormat = CF_NULL;

  sprintf(dev,"/%.32s/EVENTS/%.192s",context,fullname);
  return ExecLinkEx(dev,"PropertyInformation",&dout,&din,CA_READ,1000);
}
EVNTOOLS_EXPORT int GetArchivedEventList(const char *context,const char *trigger,time_t start,time_t stop,int *evnlst,int *lstsiz)
{
  DTYPE din, dout;
  UINT32 timeRange[2];
  char dev[132];
  int cc;

  if (context == NULL || trigger == NULL || evnlst == NULL || lstsiz == NULL) return argument_list_error;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  timeRange[0] = (UINT32)start;
  timeRange[1] = (UINT32)stop;

  dout.dArrayLength = *lstsiz;
  dout.dFormat = CF_INT32;
  dout.data.lptr = evnlst;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.lptr = (SINT32 *)timeRange;

  sprintf(dev,"/%.32s/EVENTS/%.64s",context,trigger);
  if ((cc=ExecLinkEx(dev,"Events",&dout,&din,CA_READ|CA_MUTABLE,1000)) == 0)
  {
    *lstsiz = dout.dArrayLength;
  }
  return cc;
}
EVNTOOLS_EXPORT int GetArchivedEventData(const char *context,const char *trigger,
                                     const char *devContext,const char *devServer,const char *devProperty,const char *devName,
                                     int dataIndex, int eventNumber, DTYPE *dout)
{
  ADRB adrb;
  DTYPE din;
  char dev[132];

  if (context == NULL || devServer == NULL || devProperty == NULL || dout == NULL)
    return argument_list_error;

  registerStructs();
  if (devContext == NULL) devContext = context;
  if (devName == NULL) devName = "#0";
  if (trigger == NULL) trigger = "";

  memset(&adrb,0,sizeof(ADRB));
  strncpy(adrb.devSpecifier.deviceContext,devContext,CONTEXT_NAME_SIZE);
  strncpy(adrb.devSpecifier.deviceServerName,devServer,EXPORT_NAME_SIZE);
  strncpy(adrb.devSpecifier.deviceProperty,devProperty,PROPERTY_NAME_SIZE);
  strncpy(adrb.devSpecifier.deviceName,devName,DEVICE_NAME_SIZE);
  strncpy(adrb.devSpecifier.dataTag,dout->dTag,TAG_NAME_SIZE);
  adrb.devSpecifier.dataIndex = dataIndex;
  adrb.devSpecifier.dataDimension = dout->dArrayLength;
  adrb.starttime = 0;
  adrb.stoptime = 0;
  adrb.eventNumber = eventNumber;

  memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = 1;
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"ADRBr4");
  din.data.vptr = (void *)&adrb;

  sprintf(dev,"/%.32s/EVENTS/%.64s",context,trigger);
  return ExecLinkEx(dev,"StoredData",dout,&din,CA_READ|CA_MUTABLE,1000);
}

EVNTOOLS_EXPORT char *GetArchiveComment(const char *context,const char *trigger, int eventNumber)
{
  static char cmt[1032];
  DTYPE dout,din;
  char dev[132];
  int cc;

  if (context == NULL || trigger == NULL) return "GetArchiveComment: argument list error";
  sprintf(dev,"/%.32s/EVENTSTORE/%.64s",context,trigger);
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  din.data.lptr = &eventNumber;
  dout.dArrayLength = 1024;
  dout.dFormat = CF_TEXT;
  dout.data.cptr = cmt;

  cc = ExecLinkEx(dev,"COMMENT",&dout,&din,CA_READ|CA_MUTABLE,1000);
  if (cc != 0)
  {
    sprintf(cmt,"cannot get event comment : %.32s",cc2str(cc));
  }
  cmt[dout.dArrayLength] = 0;
  if (strlen(cmt) == 0) sprintf(cmt,"no comment");
  return cmt;
}
EVNTOOLS_EXPORT int SendComment(const char *context,const char *trigger, int eventNumber,const char *cmt)
{
  DTYPE din;
  char dev[132];
  char cmtstr[1032];

  if (context == NULL || trigger == NULL || cmt == NULL) return argument_list_error;
  if (eventNumber > 0) snprintf(cmtstr, 1024, "%.1004s@%d", cmt, eventNumber); else strncpy(cmtstr, cmt, 1024);
  cmtstr[1024] = 0;
  sprintf(dev,"/%.32s/EVENTSTORE/%.64s",context,trigger);
  memset(&din,0,sizeof(DTYPE));

  din.dArrayLength = MIN((UINT32)strlen(cmtstr),1024);
  din.dFormat = CF_TEXT;
  din.data.cptr = cmtstr;

  return ExecLinkEx(dev,"ANNOTATE",NULL,&din,CA_READ|CA_WRITE|CA_MUTABLE,1000);
}
EVNTOOLS_EXPORT int GetArchiveTriggers(const char *context,NAME64 *triggers,int *triggerBufferSize)
{
  char dev[132];

  if (context == NULL || triggers == NULL || triggerBufferSize) return argument_list_error;
  if (*triggerBufferSize < 1) return dimension_error;

  sprintf(dev,"/%.32s/EVENTSTORE",context);
  return GetSystemDevices(dev,NULL,triggers,triggerBufferSize);
}

void registerStructs(void)
{
  int cc;
  char *tag;
  static int done = FALSE;
  if (done) return;
  done = TRUE;

  tag = "ADRBr4";
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,eventNumber),1,CF_LONG,"eventNumber")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,starttime),1,CF_LONG,"starttime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,stoptime),1,CF_LONG,"stoptime")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceContext),CONTEXT_NAME_SIZE,CF_TEXT,"context")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceServerName),EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceProperty),PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.deviceName),DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataTag),TAG_NAME_SIZE,CF_TEXT,"dataTag")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataFormat),1,CF_SHORT,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.reserved),1,CF_SHORT,"reserved")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataIndex),1,CF_LONG,"index")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(ADRB,devSpecifier.dataDimension),1,CF_LONG,"dataDimension")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(ADRB),1)) != 0) ccerr(cc);

  tag = "AIB";
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,numProperties),1,CF_SHORT,"numProperties")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,numDevices),1,CF_SHORT,"numDevices")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,eventCode),1,CF_LONG,"eventCode")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(AIB,comment),64,CF_TEXT,"comment")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(AIB),1)) != 0) ccerr(cc);

  tag = "PIB";
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,SamplingRate),1,CF_FLOAT,"SamplingRate")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,numPreTriggerValues),1,CF_LONG,"numPreTriggerValues")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,TimeOffset),1,CF_LONG,"TimeOffset")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,LowCount),1,CF_LONG,"LowCount")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,HighCount),1,CF_LONG,"HighCount")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,Units),16,CF_TEXT,"Units")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,LowEGU),1,CF_DOUBLE,"LowEGU")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,HighEGU),1,CF_DOUBLE,"HighEGU")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,numDevices),1,CF_SHORT,"numDevices")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(tag,OFFSETIN(PIB,reserved),1,CF_SHORT,"reserved")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(tag,sizeof(PIB),1)) != 0) ccerr(cc);

  tag = "DHS";
  if (GetStructFromTag(tag) == NULL)
  {
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,channelID),32,CF_TEXT,"channelID")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eventcode),1,CF_LONG,"eventcode")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,arraysize),1,CF_SHORT,"arraysize")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datatype),1,CF_SHORT,"datatype")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,samplingrate),1,CF_FLOAT,"samplingrate")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,pretrigger),1,CF_LONG,"pretrigger")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,timeoffset),1,CF_LONG,"timeoffset")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,lowcount),1,CF_LONG,"lowcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,highcount),1,CF_LONG,"highcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egu),16,CF_TEXT,"egu")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egulow),1,CF_FLOAT,"egulow")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eguhigh),1,CF_FLOAT,"eguhigh")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datalength),1,CF_LONG,"datalength")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,desc),4*64,CF_TEXT,"desc")) != 0) ccerr(cc);
    if ((cc=sealTaggedStruct(tag,sizeof(DHS),1)) != 0) ccerr(cc);
  }

  tag = "HEADER";
  if (GetStructFromTag(tag) == NULL)
  {
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,channelID),32,CF_TEXT,"channelID")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eventcode),1,CF_LONG,"eventcode")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,arraysize),1,CF_SHORT,"arraysize")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datatype),1,CF_SHORT,"datatype")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,samplingrate),1,CF_FLOAT,"samplingrate")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,pretrigger),1,CF_LONG,"pretrigger")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,timeoffset),1,CF_LONG,"timeoffset")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,lowcount),1,CF_LONG,"lowcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,highcount),1,CF_LONG,"highcount")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egu),16,CF_TEXT,"egu")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,egulow),1,CF_FLOAT,"egulow")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,eguhigh),1,CF_FLOAT,"eguhigh")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,datalength),1,CF_LONG,"datalength")) != 0) ccerr(cc);
    if ((cc=addFieldToStruct(tag,OFFSETIN(DHS,desc),4*64,CF_TEXT,"desc")) != 0) ccerr(cc);
    if ((cc=sealTaggedStruct(tag,sizeof(DHS),1)) != 0) ccerr(cc);
  }

err:
  if (cc) dbglog("TaggedStruct %s: %s",tag,cc2str(cc));
  return;
}
