/*
 * File ezclient.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file ezclient.c
 * \brief TINE EZ client documentation
 *
 */
#ifndef NO_DOXYGEN
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include <math.h>
#include "tine.h"
#include "eztine.h"
#endif
ezResult *ezResTbl[EZTBL_HASH_SIZE];

/*
 * 'key' is of the form : '/context/server/device[property]<input data>'
 * where 'input data' can contain up to 5 primitive types or 1 compound type
 * and is of the form : 'FMT:<value>,<value>,<value>,<value>,<value>'
 * e.g. 'FLOAT:1.234,55.335' or 'FLINT:1.23,12' or 'TEXT:hello world' or
 *      'NAME64:name1,name2,name3,name4'
 * with the caveat that ~ 128 bytes are available to hold this information as text.
 */
ezResult *getEzResItem(char *key)
{
  ezResult *ezr;
  int idx;

  if (key == NULL) return NULL;
  idx = (int)(ElfHash((unsigned char *)key) % EZTBL_HASH_SIZE);

  for (ezr=ezResTbl[idx]; ezr != NULL; ezr = ezr->nxt)
  {
    if (strnicmp(ezr->key,key,EZTBL_KEY_SIZE)) continue;
    break;
  }
  return ezr;
}
int isSimpleFmt(short fmt)
{
  if (fmt == CF_STRUCT) return FALSE;
  if (fmt <= CF_NAME16) return TRUE;
  if (fmt == CF_NAME32) return TRUE;
  if (fmt == CF_NAME48) return TRUE;
  if (fmt == CF_NAME64) return TRUE;
  return FALSE;
}
DTYPE *getInputDataFromString(char *inptStr)
{
  DTYPE *din = NULL;
  char *c, *fmtstr = inptStr;
  int cc=0,n=0,i,flen;
  char *strValues[10];

  if ((din=(DTYPE *)SystemCalloc(1,sizeof(DTYPE))) == NULL) ccerr(out_of_local_memory);
  if ((c=strchr(inptStr,':')) == NULL) ccerr(illegal_format);
  *c++ = 0;
  din->dFormat = GetFormatFromString(fmtstr);
  for (c=strtok(c,","); c != NULL; c=strtok(NULL,","))
  { /* tokenize ... */
    strValues[n] = c;
    if (++n > 10) break;
  }
  flen = fmtsizeof(din->dFormat);
  if (isSimpleFmt(din->dFormat))
  {
    din->dArrayLength = n;
    din->data.bptr = (BYTE *)SystemCalloc(n,flen);
    if (din->data.bptr == NULL) ccerr(out_of_local_memory);
    switch (din->dFormat)
    {
      case CF_BYTE:
        for (i=0; i<n; i++) din->data.bptr[i] = (BYTE)atoi(strValues[i]); break;
      case CF_INT16:
        for (i=0; i<n; i++) din->data.sptr[i] = (short)atoi(strValues[i]); break;
      case CF_INT32:
        for (i=0; i<n; i++) din->data.lptr[i] = atoi(strValues[i]); break;
      case CF_FLOAT:
        for (i=0; i<n; i++) din->data.fptr[i] = (float)strtod(strValues[i],NULL); break;
      case CF_DOUBLE:
        for (i=0; i<n; i++) din->data.dptr[i] = strtod(strValues[i],NULL); break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        for (i=0; i<n; i++) strncpy(&din->data.cptr[i+flen],strValues[i],flen); break;
      default: ccerr(illegal_format);
    }
  }
  else
  {
    din->dArrayLength = 1;
    din->data.bptr = (BYTE *)SystemCalloc(1,flen);
    switch (din->dFormat)
    {
      case CF_INTINT: if (n < 2) ccerr(invalid_data);
        ((INTINT *)din->data.vptr)->i1val = atoi(strValues[0]);
        ((INTINT *)din->data.vptr)->i2val = atoi(strValues[1]);
        break;
      case CF_FLTINT: if (n < 2) ccerr(invalid_data);
        ((FLTINT *)din->data.vptr)->fval = (float)strtod(strValues[0],NULL);
        ((FLTINT *)din->data.vptr)->ival = atoi(strValues[1]);
        break;
      case CF_DBLDBL: if (n < 2) ccerr(invalid_data);
        ((DBLDBL *)din->data.vptr)->d1val = strtod(strValues[0],NULL);
        ((DBLDBL *)din->data.vptr)->d2val = strtod(strValues[1],NULL);
        break;
      case CF_NAME64I: if (n < 2) ccerr(invalid_data);
        strncpy(((NAME64I *)din->data.vptr)->name,strValues[0],64);
        ((NAME64I *)din->data.vptr)->ival = atoi(strValues[1]);
        break;
      case CF_NAME64DBL: if (n < 2) ccerr(invalid_data);
        strncpy(((NAME64DBL *)din->data.vptr)->name,strValues[0],64);
        ((NAME64DBL *)din->data.vptr)->dval = strtod(strValues[1],NULL);
        break;
      default: ccerr(illegal_format);
    }
  }
err:
  if (cc != 0)
  {
    din->dFormat = CF_NULL;
    din->dArrayLength = 0;
    if (din->data.vptr != NULL) SystemFree(din->data.vptr);
    din->data.vptr = NULL;
  }
  return din;
}

void splitFullNameAndProperty(char *fullNameAndProperty,char **dev,char **prp)
{ /* internal: fullNameAndProperty always points to a 256-char buffer */
  char *c;
  if ((c=strchr(fullNameAndProperty,'[')) != NULL)
  { /* standard TINE form : /context/server/device[property] */
    *c++ = 0;
    *prp = c;
    if ((c=strchr(*prp,']')) != NULL) *c = 0;
    *dev = fullNameAndProperty;
  }
  else
  {
    if ((c=strrchr(fullNameAndProperty,'/')) == NULL)
    { /* no delimiters: a global keyword ? */
      *prp = fullNameAndProperty;
      *dev = &fullNameAndProperty[64];
      strcpy(*dev,"/SITE/GLOBALS");
    }
    else
    { /* straight hierarchy: /context/server/device/property */
      *c++ = 0;
      *prp = c;
      *dev = fullNameAndProperty;
    }
  }
}
int isValidEzNumberFormat(int fmt)
{ /* catch the obvious bad cases here */
  switch (fmt)
  {
    case CF_XML:
    case CF_TEXT:
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_USTRING:
    case CF_UNAME:
    case CF_STRUCT:
    case CF_IMAGE:
      return FALSE;
  }
  return TRUE;
}
int isEzSimpleNumberFormat(short fmt)
{
  if (fmt == CF_CHAR) return FALSE;
  if (fmt < CF_STRUCT) return TRUE;
  return FALSE;
}
int isEzStringFormat(short fmt)
{
  switch (fmt)
  {
    case CF_TEXT:
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_USTRING:
    case CF_UNAME:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
int isAdjFmt(short fmt);
void updateDataSet(ezResult *res)
{
  mexListener *ml;
  int cc = 0, siz, hsiz, fmt;
  char devnam[EZTBL_KEY_SIZE],*c,*dev,*prp;
  strncpy(devnam,res->key,EZTBL_KEY_SIZE);
  if ((c=strchr(devnam,'<')) != NULL) *c = 0;
  splitFullNameAndProperty(devnam,&dev,&prp);
  if (res->idListener < 0) return;
  if ((ml=MexListenerTable[res->idListener]) == NULL) return; /* not listening ! */
  ml->idleTimeInSeconds = 0;
  fmt = ml->dout.dFormat;
  siz = ml->dout.dArrayLength*fmtsizeof(fmt);
  hsiz = getFormatHeaderSize(fmt);
  if (isAdjFmt((short)fmt))
  {
    hsiz = hsiz + sizeof(void *);
    siz *= hsiz;
  }
  else if (fmt == CF_STRUCT)
  { /* must be tagged and registered! */
    int ssz;
    if ((ssz=GetStructSize(ml->dout.dTag)) <= 0) ccerr(invalid_structure_tag);
    siz *= ssz;
  }
  else
  {
    siz += hsiz;
  }
  if (ml->outarr != NULL) 
  { /* adjustable length format type */
    memcpy(res->dout.data.vptr,ml->outarr,siz);
  }
  else 
  {
    BYTE *bptr;
    int bsiz;
    if ((bptr=getFormatBodyPtr(res->dout.data.vptr,fmt)) != NULL)
    { /* e.g CF_IMAGE */
      siz = hsiz;
      bsiz = getFormatBodyBufferSize(ml->outbuf,res->dout.dFormat);
      if (bsiz > 0) memcpy(bptr,&ml->outbuf[hsiz+sizeof(UINT32)],bsiz);
    }
    if (res->dout.data.vptr != ml->outbuf) memcpy(res->dout.data.vptr,ml->outbuf,siz);
  }
err:
  if (cc) dbglog("updateDataSet: %.32s",cc2str(cc));
}
EXPORT_EZ ezDblArray ezGetDblValues(ezResult *res)
{
  ezDblArray eza = {NULL, 0, 0, invalid_parameter};
  double *vals = NULL;
  int cc = 0, sizeInBytes = 0, len;
  time_t ts = 0;
  mexListener *ml;

  if (res == NULL) return eza;
  eza.status = res->status;
  if (res->status != 0 && !(res->status & CE_SENDDATA)) return eza;
  if (!isValidEzNumberFormat(res->dout.dFormat)) ccerr(illegal_format);
  len = res->dout.dArrayLength;
  sizeInBytes = len * sizeof(double);
  if (res->scratch == NULL)
  {
    if ((res->scratch=(BYTE *)SystemCalloc(sizeInBytes,1)) == NULL) ccerr(out_of_local_memory);
    res->scratchCapacity = sizeInBytes;
  }
  else if (sizeInBytes > res->scratchCapacity)
  {
    BYTE *tmp = (BYTE *)realloc(res->scratch,sizeInBytes);
    if (tmp == NULL) ccerr(out_of_local_memory);
    res->scratch = tmp;
    res->scratchCapacity = sizeInBytes;
  }
  updateDataSet(res);
  vals = (double *)res->scratch;
  if ((cc=GetValuesAsDouble(&res->dout,vals,len)) == 0)
  {
    if (res->idListener >= 0)
    {
      if ((ml=MexListenerTable[res->idListener]) != NULL)
      {
        len = ml->lastsize;
        ts = ml->lasttime;
      }
    }
    else
    {
      len = res->lengthUpdate;
      ts = res->timeUpdate;
    }
  }
err:
  if (cc != 0)
  {
    len = 0;
    res->status = cc;
  }
  else if (!isEzSimpleNumberFormat(res->dout.dFormat))
  { /* compound data type converted to double successfully, but ... */
    cc = ambiguous;
  }
  eza.status = cc;
  eza.timeValid = ts;
  eza.length = len;
  eza.values = vals;
  return eza;
}
EXPORT_EZ ezIntArray ezGetIntValues(ezResult *res)
{
  ezIntArray eza = {NULL, 0, 0, invalid_parameter};
  int *vals = NULL;
  int cc = 0, sizeInBytes = 0, len;
  time_t ts = 0;
  mexListener *ml;

  if (res == NULL) return eza;
  eza.status = res->status;
  if (res->status != 0 && !(res->status & CE_SENDDATA)) return eza;
  if (!isValidEzNumberFormat(res->dout.dFormat)) ccerr(illegal_format);
  len = res->dout.dArrayLength;
  sizeInBytes = len * sizeof(double);
  if (res->scratch == NULL)
  {
    if ((res->scratch=(BYTE *)SystemCalloc(sizeInBytes,1)) == NULL) ccerr(out_of_local_memory);
    res->scratchCapacity = sizeInBytes;
  }
  else if (sizeInBytes > res->scratchCapacity)
  {
    BYTE *tmp = (BYTE *)realloc(res->scratch,sizeInBytes);
    if (tmp == NULL) ccerr(out_of_local_memory);
    res->scratch = tmp;
    res->scratchCapacity = sizeInBytes;
  }
  updateDataSet(res);
  vals = (int *)res->scratch;
  if ((cc=GetValuesAsInt32(&res->dout,vals,len)) == 0)
  {
    if (res->idListener >= 0)
    {
      if ((ml=MexListenerTable[res->idListener]) != NULL)
      {
        len = ml->lastsize;
        ts = ml->lasttime;
      }
    }
    else
    {
      len = res->lengthUpdate;
      ts = res->timeUpdate;
    }
  }
err:
  if (cc != 0)
  {
    len = 0;
    res->status = cc;
  }
  else if (!isEzSimpleNumberFormat(res->dout.dFormat))
  { /* compound data type converted to double successfully, but ... */
    cc = ambiguous;
  }
  eza.status = cc;
  eza.timeValid = ts;
  eza.length = len;
  eza.values = vals;
  return eza;
}
EXPORT_EZ ezStrArray ezGetStrValues(ezResult *res)
{
  ezStrArray eza = {NULL, 0, 0, invalid_parameter};
  char **vals = NULL, *sptr;
  int cc = 0, sizeInBytes = 0, len = 0, i, slen, tlen = 0;
  time_t ts = 0;
  mexListener *ml;

  if (res == NULL) return eza;
  updateDataSet(res);
  eza.status = res->status;
  if (res->status != 0 && !(res->status & CE_SENDDATA)) return eza;
  switch (res->dout.dFormat)
  {
    case CF_STRING: 
    case CF_KEYVALUE:
      len = res->dout.dArrayLength;
      sizeInBytes = len * (sizeof(size_t) + 128);
      break;
    case CF_TEXT:
      len = 1;
      sizeInBytes = res->dout.dArrayLength;
      sizeInBytes += ((sizeInBytes+64)%64);
      sizeInBytes += sizeof(size_t);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
    case CF_USTRING:
    case CF_UNAME:
      len = res->dout.dArrayLength;
      sizeInBytes = len * (sizeof(size_t) + fmtsizeof(res->dout.dFormat));
      break;
    case CF_STRUCT:
      ccerr(illegal_format);
    default:
      ccerr(illegal_format);
  }
  if (res->scratch == NULL)
  {
    if ((res->scratch=(BYTE *)SystemCalloc(sizeInBytes,1)) == NULL) ccerr(out_of_local_memory);
    res->scratchCapacity = sizeInBytes;
  }
  else if (sizeInBytes > res->scratchCapacity)
  {
    BYTE *tmp = (BYTE *)realloc(res->scratch,sizeInBytes);
    if (tmp == NULL) ccerr(out_of_local_memory);
    res->scratch = tmp;
    res->scratchCapacity = sizeInBytes;
  }
  vals = (char **)res->scratch;
  if ((cc=GetValuesAsStringEx(&res->dout,(char *)res->scratch,CF_STRING,len,0)) == 0)
  {
    if (res->idListener >= 0)
    {
      if ((ml=MexListenerTable[res->idListener]) != NULL)
      {
        len = ml->lastsize;
        ts = ml->lasttime;
      }
    }
    else
    {
      len = res->lengthUpdate;
      ts = res->timeUpdate;
    }
    tlen = len * sizeof(size_t);
    for (i=0; i<len; i++)
    {
      sptr = (char *)vals[i];
      slen = (int)strlen(sptr);
      vals[i] = (char *)&res->scratch[tlen];
      strcpy(vals[i],sptr);
      *(vals[i]+slen++) = 0;
      tlen += slen;
      if (tlen >= sizeInBytes) break;
    };
  }
err:
  if (cc != 0)
  {
    len = 0;
    res->status = cc;
  }
  eza.status = cc;
  eza.timeValid = ts;
  eza.length = len;
  eza.values = vals;
  return eza;
}
EXPORT_EZ int ezAddMonitor(ezResult *res,void (*nf)(int), int nid)
{
  char *c, devnam[EZTBL_KEY_SIZE];
  mexListener *ml;
  if (res == NULL || nf == NULL) return argument_list_error;
  if (res->idListener < 0) return invalid_link;
  if ((ml=MexListenerTable[res->idListener]) == NULL) return link_not_open;

  strncpy(devnam,res->key,EZTBL_KEY_SIZE);
  if ((c=strchr(devnam,'<')) != NULL) *c = 0;
  return tAttachNotifier(devnam, nf, nid);
}
static char *ezToStringBuffer = NULL;
static int ezToStringBufferCapacity = 0;
EXPORT_EZ char *ezToString(ezResult *res)
{
  int cc = 0, sizeInBytes = 0, len;
  DTYPE d;
  static char errstr[64];

  if (res == NULL) return "[null]";
  if (res->status != 0 && !(res->status & CE_SENDDATA)) ccerr(res->status);
  len = res->dout.dArrayLength;
  if (len == 0) return "[empty]";
  sizeInBytes = len * 64;
  if (sizeInBytes == 64) sizeInBytes += 64;
  if (ezToStringBuffer == NULL)
  {
    if ((ezToStringBuffer=(BYTE *)SystemCalloc(sizeInBytes,1)) == NULL) ccerr(out_of_local_memory);
    ezToStringBufferCapacity = sizeInBytes;
  }
  else if (sizeInBytes > ezToStringBufferCapacity)
  {
    BYTE *tmp = (BYTE *)realloc(ezToStringBuffer,sizeInBytes);
    if (tmp == NULL) ccerr(out_of_local_memory);
    ezToStringBuffer = tmp;
    ezToStringBufferCapacity = sizeInBytes;
  }
  memcpy(&d,&res->dout,sizeof(DTYPE));
  d.dArrayLength = res->lengthUpdate;
  cc = GetValuesAsString(&d,(char *)ezToStringBuffer,(UINT32 *)&sizeInBytes);
err:
  if (cc != 0)
  {
    res->status = cc;
    return GetLastLinkError((short)cc,errstr);
  }
  return (char *)ezToStringBuffer;
}
EXPORT_EZ int ezFreeResult(ezResult *res)
{
  int lid;
  if (res == NULL) return argument_list_error;
  lid = res->idListener;
  if (lid >= 0)
  {
    if (lid >= nMexListeners) return invalid_link;
    stopListener(MexListenerTable[lid]);
    if (res->prv != NULL)
    {
      res->prv = res->nxt;
    }
    else
    {
      int idx = (int)(ElfHash((unsigned char *)res->key) % EZTBL_HASH_SIZE);
      ezResTbl[idx] = res->nxt;
    }
    if (res->dout.data.bptr != NULL) SystemFree(res->dout.data.bptr);
    if (res->devLst != NULL) SystemFree(res->devLst);
    if (res->scratch != NULL) SystemFree(res->scratch);
    if (res->scratchNames != NULL) SystemFree(res->scratchNames);
    SystemFree(res);
  }
  return 0;
}
EXPORT_EZ ezStrArray ezGetChannels(ezResult *res)
{
  ezStrArray eza = {NULL, 0, 0, invalid_parameter};
  char **vals = NULL, *sptr, thisDeviceName[DEVICE_NAME_SIZE];
  DTYPE d;
  time_t ts = 0;
  int cc = 0, sizeInBytes = 0, i, len, slen, tlen = 0;

  if (res == NULL) return eza;
  eza.status = res->status;
  if (res->status != 0 && !(res->status & CE_SENDDATA)) return eza;
  len = MAX(res->dout.dArrayLength,1);
  if (res->devLst == NULL) len = 1;
  sizeInBytes = len * (sizeof(size_t) + fmtsizeof(CF_NAME64));
  if (res->scratchNames == NULL)
  {
    if ((res->scratchNames=(BYTE *)SystemCalloc(sizeInBytes,1)) == NULL) ccerr(out_of_local_memory);
  }
  memset(&d,0,sizeof(DTYPE));
  d.dArrayLength = len;
  d.dFormat = CF_NAME64;
  if (res->devLst != NULL)
  {
    if (res->dout.dArrayLength == 0) ccerr(code_failure);
    d.data.vptr = res->devLst;
  }
  else
  {
    int ps = 0;
    char *c, devnam[EZTBL_KEY_SIZE], *dev, *prp;
    strncpy(devnam,res->key,EZTBL_KEY_SIZE);
    if ((c=strchr(devnam,'<')) != NULL) *c = 0;
    splitFullNameAndProperty(devnam,&dev,&prp);
    if (dev[0] == '/') ps++; /* normally the case */
    if ((c=strchr(&dev[ps],'/')) != NULL && (c=strchr(++c,'/')) != NULL)
    {
      c++; /* we're now pointing at the device name */
      strncpy(thisDeviceName,c,64);
    }
    else
    { /* no device name in key and this isn't an MCA */
      thisDeviceName[0] = 0;
    }
    d.data.vptr = thisDeviceName;
  }
  ts = time(NULL);
  vals = (char **)res->scratchNames;
  if ((cc=GetValuesAsStringEx(&d,(char *)res->scratchNames,CF_STRING,len,0)) == 0)
  {
    tlen = len * sizeof(size_t);
    for (i=0; i<len; i++)
    {
      sptr = (char *)vals[i];
      slen = (int)strlen(sptr);
      vals[i] = (char *)&res->scratchNames[tlen];
      strcpy(vals[i],sptr);
      *(vals[i]+slen++) = 0;
      tlen += slen;
      if (tlen >= sizeInBytes) break;
    };
  }
err:
  if (cc != 0)
  {
    ts = 0;
    len = 0;
    res->status = cc;
  }
  eza.status = cc;
  eza.timeValid = ts;
  eza.length = len;
  eza.values = vals;
  return eza;
}
ezResult *addEzResItem(char *key)
{
  int cc = 0, n = 1, sizeInBytes, idx;
  ezResult *ezr;
  char *c, devnam[EZTBL_KEY_SIZE], *dev, *prp, *inptStr;
  PrpQueryStruct pqs, *ppqs = &pqs;
  DTYPE *din = NULL;

  if ((ezr=getEzResItem(key)) != NULL) return ezr;
  if ((ezr=(ezResult *)SystemCalloc(1,sizeof(ezResult))) == NULL) return NULL;
  strncpy(ezr->key,key,EZTBL_KEY_SIZE);
  strncpy(devnam,key,EZTBL_KEY_SIZE);
  if ((inptStr=strchr(devnam,'<')) != NULL)
  { /* input data ! */
    *inptStr++ = 0; /* remove it from the devnam string */
    if (strchr(inptStr,'@') == NULL)
    { /* 'simple' input data are in the key ! */
      if ((c=strchr(inptStr,'>')) != NULL) *c = 0;
      din = getInputDataFromString(inptStr);
      if (din != NULL) memcpy(&ezr->din,din,sizeof(DTYPE));
    }
  }
  splitFullNameAndProperty(devnam,&dev,&prp);
  cc = GetSystemPropertyInformation(dev,prp,&ppqs,&n);
  if (cc != 0) ccerr(cc);
  ezr->dout.dFormat = LFMT(pqs.prpFormat);
  ezr->dout.dArrayLength = pqs.prpSize;
  strncpy(ezr->dout.dTag,pqs.prpTag,TAG_NAME_SIZE);
  sizeInBytes = pqs.prpSize * fmtsizeof(pqs.prpFormat);
  if (sizeInBytes > 0)
  { /* allocate the necessary output bytes */
    ezr->dout.data.bptr = (BYTE *)SystemCalloc(sizeInBytes,1);
    if (ezr->dout.data.bptr == NULL) ccerr(out_of_local_memory);
  }
  else
  { /* take no chances ... */
    ezr->dout.dFormat = CF_NULL;
    ezr->dout.dArrayLength = 0;
  }
  if (pqs.prpSize > 1 && pqs.prpArrayType & AT_CHANNEL)
  {
    n = pqs.prpSize;
    ezr->devLst = (NAME64 *)SystemCalloc(n,sizeof(NAME64));
    if (ezr->devLst == NULL) ccerr(out_of_local_memory);
    GetSystemDevices(dev,prp,ezr->devLst,&n);
  }
  ezr->idListener = -1; /* until there is one ... */
  idx = (int)(ElfHash((unsigned char *)key) % EZTBL_HASH_SIZE);
  if (ezResTbl[idx] != NULL) ezResTbl[idx]->prv = ezr;
  ezr->nxt = ezResTbl[idx];
  ezResTbl[idx] = ezr;
err:
  ezr->status = cc;
  return ezr;
}
void makeEzResKey(char *key,char *fullNameAndProperty,short access,DTYPE *dt,char *inptTag)
{
  int i;
  char *c;
  strncpy(key,fullNameAndProperty,256);
  if ((c=strchr(key,'<')) != NULL) *c = 0;
  if ((c=strchr(key,'%')) != NULL) *c = 0;
  strtrm(key);
  if (access & CA_WRITE) strcat(key,"+w");
  if (dt == NULL || dt->dFormat == CF_NULL) return; /* easy one ! */
  strcat(key,"<");
  if (inptTag != NULL)
  { /* take the tag given */
    strcat(key,inptTag);
    strcat(key,">");
    return;
  }
  // TODO: need to verify the 'dt' (not all format allowed, etc.)
  if (dt->dArrayLength > 5) dt->dArrayLength = 5;
  key[255] = 0; /* fullNameAndProperty shouldn't be longer than ~192 chars, ever */
  strcat(key,GetFormatAsString(dt->dFormat));
  strcat(key,":");
  switch (dt->dFormat)
  {
    case CF_BYTE:
      for (i=0; i<(int)dt->dArrayLength; i++) 
      {
        sprintf(&key[strlen(key)],"%d",dt->data.bptr[i]);
        if (i < (int)dt->dArrayLength-1) strcat(key,",");
      }
      break;
    case CF_INT16:
      for (i=0; i<(int)dt->dArrayLength; i++) 
      {
        sprintf(&key[strlen(key)],"%d",dt->data.sptr[i]);
        if (i < (int)dt->dArrayLength-1) strcat(key,",");
      }
      break;
    case CF_INT32:
      for (i=0; i<(int)dt->dArrayLength; i++) 
      {
        sprintf(&key[strlen(key)],"%d",dt->data.lptr[i]);
        if (i < (int)dt->dArrayLength-1) strcat(key,",");
      }
      break;
    case CF_FLOAT:
      for (i=0; i<(int)dt->dArrayLength; i++) 
      {
        sprintf(&key[strlen(key)],"%g",dt->data.fptr[i]);
        if (i < (int)dt->dArrayLength-1) strcat(key,",");
      }
      break;
    case CF_DOUBLE:
      for (i=0; i<(int)dt->dArrayLength; i++) 
      {
        sprintf(&key[strlen(key)],"%g",dt->data.dptr[i]);
        if (i < (int)dt->dArrayLength-1) strcat(key,",");
      }
      break;
    case CF_CHAR:
      strncpy(&key[strlen(key)],dt->data.cptr,dt->dArrayLength);
      break;
  }
  strcat(key,">");
}
EXPORT_EZ ezResult *ezSet(char *fullNameAndProperty,...)
{
  va_list argp;
  int cc = 0, n = 0, infmt = CF_NULL;
  short access = CA_WRITE;
  DTYPE *dt = NULL, din;
  int iv, ivals[10];
  float fv, fvals[10];
  char *sv, sval[128];
  char key[1024], *c, *dev, *prp;
  char *inpt = NULL, *intag = NULL, *tagfmt = "";
  ezResult *ezr;

  va_start(argp,fullNameAndProperty);
  if ((inpt=strchr(fullNameAndProperty,'<')) == NULL)
  { /* allow a little bit of slack here ... */
    inpt = strchr(fullNameAndProperty,'%');
  }
  for (c=inpt; c != NULL && *c != 0; c++)
  {
    if (*c != '%') continue;
    switch (*++c)
    {
      case 'd':
        if (infmt == CF_NULL) infmt = CF_INT32;
        if (infmt != CF_INT32) ccerr(argument_list_error);
        iv = va_arg(argp,int);
        if (n < 9) ivals[n++] = iv;
        break;
      case 'f':
      case 'g':
        if (infmt == CF_NULL) infmt = CF_FLOAT;
        if (infmt != CF_FLOAT) ccerr(argument_list_error);
        fv = (float)va_arg(argp,double);
        if (n < 9) fvals[n++] = fv;
        break;
      case 's':
        if (infmt == CF_NULL) infmt = CF_CHAR;
        if (infmt != CF_CHAR) ccerr(argument_list_error);
        for (sv=va_arg(argp,char *); *sv && n<127; sv++) sval[n++] = *sv;
        break;
      case 'D':
        if (infmt != CF_NULL) ccerr(argument_list_error);
#       if defined(WIN32)
#       ifdef OS_64BIT
        dt = (DTYPE *)va_arg(argp,__int64);
        tagfmt = "%.16s:@0x%I64x";
#       else
        dt = (DTYPE *)va_arg(argp,__int32);
        tagfmt = "%.16s:@0x%lx";
#       endif
#       else
        dt = (DTYPE *)va_arg(argp,long);
        tagfmt = "%.16s:@0x%lx";
#       endif
        if (dt != NULL)
        {
          infmt = dt->dFormat;
          n = dt->dArrayLength;
          sprintf(sval, tagfmt, GetFormatAsString(dt->dFormat), (size_t)dt);
          intag = sval;
          access |= CA_SYNC;
        }
        break;
      default:
        break;
    }
  }
  if (dt == NULL)
  {
    din.dArrayLength = n;
    din.dFormat = infmt;
    switch (infmt)
    {
      case CF_INT32: din.data.lptr = ivals; break;
      case CF_FLOAT: din.data.fptr = fvals; break;
      case CF_CHAR:  din.data.cptr = sval; break;
    }
    dt = &din;
  }
err:
  va_end(argp);
  makeEzResKey(key,fullNameAndProperty,access,dt,intag);
  /* addEzResItem only 'adds' when it needs to ... */
  ezr = addEzResItem(key);
  if (cc == 0)
  { /* so far, so good ... */
    if ((c=strchr(key,'<')) != NULL) *c = 0;
    splitFullNameAndProperty(key,&dev,&prp);
    if (dt != NULL) memcpy(&ezr->din,dt,sizeof(DTYPE));
    cc = ExecLinkEx(dev,prp,&ezr->dout,&ezr->din,access,1000);
    if (cc == 0 || (cc & CE_SENDDATA))
    {
      if (ezr->dout.dArrayLength > 0) ezr->lengthUpdate = GetCompletionDataSize(-1);
      ezr->timeUpdate = time(NULL);
    }
  }
  ezr->status = cc;
  return ezr;
}
EXPORT_EZ ezResult *ezGet(char *fullNameAndProperty,...)
{
  va_list argp;
  int cc = 0, n = 0, infmt = CF_NULL;
  short access = CA_READ;
  DTYPE *dt = NULL, din;
  int iv, ivals[10];
  float fv, fvals[10];
  char *sv, sval[128];
  char key[1024], *c, *dev, *prp;
  char *inpt = NULL, *intag = NULL, *tagfmt = "";
  ezResult *ezr;
  mexListener *ml;

  va_start(argp,fullNameAndProperty);
  if ((inpt=strchr(fullNameAndProperty,'<')) == NULL)
  { /* allow a little bit of slack here ... */
    inpt = strchr(fullNameAndProperty,'%');
  }
  for (c=inpt; c != NULL && *c != 0; c++)
  {
    if (*c != '%') continue;
    switch (*++c)
    {
      case 'd':
        if (infmt == CF_NULL) infmt = CF_INT32;
        if (infmt != CF_INT32) ccerr(argument_list_error);
        iv = va_arg(argp,int);
        if (n < 9) ivals[n++] = iv;
        break;
      case 'f':
      case 'g':
        if (infmt == CF_NULL) infmt = CF_FLOAT;
        if (infmt != CF_FLOAT) ccerr(argument_list_error);
        fv = (float)va_arg(argp,double);
        if (n < 9) fvals[n++] = fv;
        break;
      case 's':
        if (infmt == CF_NULL) infmt = CF_CHAR;
        if (infmt != CF_CHAR) ccerr(argument_list_error);
        for (sv=va_arg(argp,char *); *sv && n<127; sv++) sval[n++] = *sv;
        break;
      case 'D':
        if (infmt != CF_NULL) ccerr(argument_list_error);
#       if defined(WIN32)
#       ifdef OS_64BIT
        dt = (DTYPE *)va_arg(argp,__int64);
        tagfmt = "%.16s:@0x%I64x";
#       else
        dt = (DTYPE *)va_arg(argp,__int32);
        tagfmt = "%.16s:@0x%lx";
#       endif
#       else
        dt = (DTYPE *)va_arg(argp,long);
        tagfmt = "%.16s:@0x%lx";
#       endif
        if (dt != NULL)
        {
          infmt = dt->dFormat;
          n = dt->dArrayLength;
          sprintf(sval,tagfmt,GetFormatAsString(dt->dFormat),(size_t)dt);
          intag = sval;
          access |= CA_SYNC;
        }
        break;
      default:
        break;
    }
  }
  if (dt == NULL)
  {
    din.dArrayLength = n;
    din.dFormat = infmt;
    switch (infmt)
    {
      case CF_INT32: din.data.lptr = ivals; break;
      case CF_FLOAT: din.data.fptr = fvals; break;
      case CF_CHAR:  din.data.cptr = sval; break;
    }
    dt = &din;
  }
err:
  va_end(argp);
  makeEzResKey(key,fullNameAndProperty,access,dt,intag);
  /* addEzResItem only 'adds' when it needs to ... */
  ezr = addEzResItem(key);
  if (cc == 0)
  { /* so far, so good ... */
    if ((c=strchr(key,'<')) != NULL) *c = 0;
    splitFullNameAndProperty(key,&dev,&prp);
    if (dt != NULL) memcpy(&ezr->din,dt,sizeof(DTYPE));
    cc = alsnExecLink(dev,prp,&ezr->dout,&ezr->din,access,1000);
    ml = getListener(dev,prp,&ezr->dout,&ezr->din,1000);
    if (ml != NULL) 
    { /* a listener was started */
      ezr->idListener = ml->mlid;
      ezr->lengthUpdate = ml->lastsize;
      ezr->timeUpdate = ml->lasttime;
    }
    else
    { /* the call went synchronously */
      if (cc == 0 || (cc & CE_SENDDATA))
      {
        ezr->lengthUpdate = GetCompletionDataSize(-1);
        ezr->timeUpdate = time(NULL);
      }
    }
  }
  ezr->status = cc;
  return ezr;
}
#if defined(WIN32)
int gBufSrvInitialized = FALSE;
void TineStopCycler(void)
{
  feclog("TineStopCycler has been called");
  // kill it if it's running
  SystemKillCycleTimer();
  gBufSrvInitialized = FALSE;
}
void TineStartCycler(void)
{
  feclog("TineStartCycler has been called");
  // start anew ...
  SystemSetCycleTimer();
  gBufSrvInitialized = TRUE;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      SetConnectionTableCapacity(MexListenerTableCapacity);
      SetClnRecvQueueDepth(0);
      SystemSetCycleTimer();
      gBufSrvInitialized = TRUE;
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      SystemKillCycleTimer();
      break;
  }
  return (TRUE);
}
#endif
