/*
 * File group2csv.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "tine.h"
#ifdef WIN32
# include <direct.h>
# include <lm.h>
#endif
#if defined(UNIX) || defined(MACOS) || defined(VXWORKS)
# include <sys/stat.h>
# include <dirent.h>
#endif

#if defined(WIN32)
int getWinUserGroupMembers(char *grp,char *src,NAME16 **members,int *nmembers)
{
  DWORD ner=0,ne=0,rh=0;
  LPGROUP_USERS_INFO_0 pBuf = NULL, pMem;
  wchar_t tsrv[64];
  wchar_t tgrp[16];
  int i, rc, cc=0, bsiz=10000; /* can accomodate 100 users */
  NAME16 *mems = NULL;

  if (src == NULL) src = "";
  if (*members != NULL) 
  {
    SystemFree(*members); *members = NULL;
  }
  memset(tsrv,0,64*sizeof(wchar_t));
  memset(tgrp,0,64*sizeof(wchar_t));
  mbstowcs(tsrv,src,strlen(src));
  mbstowcs(tgrp,grp,strlen(grp));

retryGetUsers:
  switch (rc=NetGroupGetUsers(tsrv,tgrp,0,(LPBYTE*)&pBuf,bsiz,&ner,&ne,&rh))
  {
    case NERR_Success:
      break;
    case ERROR_ACCESS_DENIED:
      ccerr(access_denied);
    case ERROR_MORE_DATA:
      if (bsiz < 20000) 
      { /* be reasonable */
        bsiz += 2000;
        goto retryGetUsers;
      }
      ccerr(buffer_too_small);
    case NERR_InvalidComputer:
      ccerr(no_such_computer);
    case NERR_UserNotFound:
      ccerr(non_existent);
    default:
      ccerr(code_failure);
  }
  if ((mems=(NAME16 *)SystemCalloc(ner,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
 
  pMem = pBuf;
  for (i=0,pMem=pBuf; i<(int)ner && pMem != NULL; i++,pMem++)
  {
    wcstombs(mems[i].name,pMem->grui0_name,16);
  }
  *members = mems;
err:
  *nmembers = ner;
  if (pBuf != NULL &&!IsBadReadPtr(pBuf->grui0_name,bsiz))
  { 
    NetApiBufferFree(pBuf);
  }
  return cc;
}
#elif defined(UNIX)
#include "grp.h"
int getUnixUserGroupMembers(char *grp,NAME16 **members,int *nmembers)
{
# define NICE_GRPLST_SIZE 256
  int n=0, cc=0;
  NAME16 mems[NICE_GRPLST_SIZE], *pmems;
  struct group *gp;
  char **cp;
  if (grp == NULL || members == NULL || nmembers == NULL) ccerr(argument_list_error);
  setgrent ();
  while ((gp=getgrent()) != NULL)
  { /* interate until we get a match */
    if (strcmp(gp->gr_name,grp)) continue;
    cp = gp->gr_mem;
    while (*cp && n<NICE_GRPLST_SIZE) strncpy(mems[n++].name,*cp++,16);
  }
  endgrent ();
  if ((pmems=(NAME16 *)SystemCalloc(n,sizeof(NAME16))) == NULL) 
    ccerr(out_of_server_memory);
  memcpy(pmems,mems,n*sizeof(NAME16));
  *nmembers = n;
  *members = pmems;
err:
  return cc;
}
#endif /* WIN32 or UNIX */
int mkDirPath(char *dname)
{
# if defined(UNIX)
  DIR *dp;
  char *tok = "/";
# elif defined(WIN32)
  char fn[128];
  char *tok = "\\";
  struct _finddata_t fp;
# endif
  char *c=dname,tmpname[128],dirname[128];

  strncpy(tmpname,dname,128); dirname[0] = 0;
  while (*c++ == *tok) strcat(dirname,tok); /* in case it begins with '/' or '\\' */
  for (c=strtok(tmpname,tok); c != NULL; c=strtok(NULL,tok))
  {
    if (strlen(c) == 0) break; /* terminating delimiter gives the last token empty */
    strcat(dirname,c);
    strcat(dirname,tok);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL)
    {
      closedir(dp);
      chmod(dirname,S_DENYNONE);
    }
    else
    {
      if (mkdir(dirname,S_DENYNONE)) return FALSE;
    }
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return FALSE;
    }
#   endif
  }
  return TRUE;
}
char *usageTxt[] = {
  "\nacquire group members and dump to file\n",
  "\tgroup2csv <[domain:]group name> [destination path]",
};
int main(int argc, char *argv[])
{
  char *grp,*dom;
  int i, cc, nmems;
  NAME16 *mems = NULL;
  FILE *fp = NULL;
  char fn[256],path[128] = ".";
  if (argc < 2)
  {
    for (i=0; i<2; i++) printf("%s\n",usageTxt[i]);
    printf("\n");
    exit(1);
  }
  if (argc > 2)
  {
    strncpy(path,argv[2],128);
  }
  dom = argv[1];
  if ((grp=strchr(argv[1],':')) == NULL)
  {
    grp = argv[1];
    dom = NULL;
  }
  else
  {
    *grp++ = 0;
  }
# ifdef WIN32
  cc = getWinUserGroupMembers(grp,dom,&mems,&nmems);
# else
  cc = getUnixUserGroupMembers(grp,&mems,&nmems);
# endif
  if (cc == 0)
  {
    mkDirPath(path);
    sprintf(fn,"%.128s%c%.32s-members.csv",path,FS_DELIMITER,grp);
    if ((fp=fopen(fn,"w")) == NULL) ccerr(cc);
    fprintf(fp,"USERNAME\n");
    for (i=0; i<nmems; i++)
    {
      fprintf(fp,"%.16s\n",mems[i].name);
    }
  }
err:
  if (cc) printf("group2csv failed: %d\n",cc);
  if (fp != NULL) fclose(fp);
  return 0;
}
