/*
 * File historyUtils.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file historyUtils.c
 * \brief TINE history utilities documentation 
 *
 */
#include "project.def"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include <math.h>
#include "tine.h"
#include "tbufsrv.h"

double blob[8000];
time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[64];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  if (ts == NULL || strlen(ts) == 0) return t;
  hr[0] = 0;
  tnow = localtime(&t);
  len = (int)strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,64);
      if ((c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    { /* no year given, use this year */
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year < 90) tmts.tm_year += 100;
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
EXPORT_B double GetTineTimeStampFromString(char *timestring)
{
  time_t t = parseTimeString(timestring);
  return (double)t;
}
EXPORT_B int thistory(char *server,char *property,char *devname,char *stoptime,char *depth,float *values,time_t *times,int *bufsiz)
{
  char devsrv[256];
  long stop = (long)time(NULL), start, mf = 1, dp = 0;
  int i, num = 5000, cc, offset = 0;
  FLTINT *fiDataArray = (FLTINT *)blob;

  if (values == NULL || bufsiz == NULL) return argument_list_error;
  if (property == NULL) return argument_list_error;
  if (devname == NULL) devname = "#0";

  strncpy(devsrv,server,36); /* use it as it comes in ! */
  strcat(devsrv,"/");
  strcat(devsrv,devname);
  strcat(devsrv,"[");
  strcat(devsrv,property);
  if (strstr(server,"HISTORY") == NULL)  /* not talking to a central archive server => local history */
  { 
    if (strstr(property,".HIST") == NULL && strstr(property,".HST") == NULL)
    { /* append it */
      strcat(devsrv,".HIST");
    }
  }
  strcat(devsrv,"]");
 
  stop = (long)parseTimeString(stoptime);

  if (!stricmp(depth,"SNAPSHOT"))
  { /* snapshot request ? */
    if (times == NULL) return argument_list_error;
    *times = stop;
    num = *bufsiz;
    return GetArchivedDataAsSnapshot(devsrv,times,values,num);
  }

  strupr(depth);
  if (strstr(depth,"DAY")) mf = 24 * 3600;
  else if (strstr(depth,"HOUR")) mf = 3600;
  else if (strstr(depth,"MINUTE")) mf = 60;
  else if (strstr(depth,"WEEK")) mf = 24 * 3600 * 7;
  dp = mf * atoi(depth);
  start = stop - dp;
  printf("%s/%s, Timestamp\n",property,devname);
  if ((cc=GetArchivedDataAsFloat(devsrv,start,stop,fiDataArray,&num)) == 0) 
  {
    if ((int)fiDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num && i<*bufsiz; i++)
    {
      values[i] = fiDataArray[i].fval;
      if (times != NULL) times[i] = fiDataArray[i].ival;
    }
    *bufsiz = num;
    return 0;
  }

  return cc;
}
