/*
 * File iplib.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "prolog.h"
#include "srvcore.h"
#include "srvip.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"

/**
 * \internal
 * \brief UDP server port.
 * 
 */
int gtUDPPort = UDP_PORT;

/**
 * \internal
 * \brief MMF server port.
 * 
 */
int gtStrmPort = STREAM_PORT;

/**
 * \internal .
 * \brief UDP client port.
 */
int gtTCPPort = TCP_PORT;

/**
 * \internal .
 * \brief remote service (pipe) port.
 */
int gtPipePort = PIPE_PORT;                   
/**
 * \internal
 * \brief Address of primary ethernet card.
 */                                     
char gtEthrAddr[64]; /* INADDR_ANY if not given */

/**
  *\internal
 * \brief Port for multicast network services.
 */
int gtNetSrvPort = NETSRV_PORT;
          
/**
 * \internal
 * \brief Client UDP port for globals multicast.
 */
int gtGCastPort = GCAST_PORT;

/**
 * \internal
 * \brief Client UDP multicast port.
 */
int gtMCastPort = MCAST_PORT;

/**
 * \internal
 * \brief UDP client port.
 * 
 **/
int gtUDPClnPort = UDP_CLN_PORT;

/**
 * \internal
 * \brief Address of network services.
 */                                     
char *gtNetSrvAddr = NETSRV_ADDR;

/**
 * \internal
 * \brief Local broadcast address.
 */
char *gtNetLclBcast = NETBCST_ADDR;

/**
 * \internal
 * \brief Multicast address
 */
char *gtMCastAddr = MCAST_ADDR;
char *gtMCastAddrLegacy = MCAST_ADDR_LEGACY;
char gtMCastAddrBuffer[64];

/**
 * \internal
 * \brief Tine Engine operation mode
 *
 * Single process OSes (DOS, NIOS) use 'TRUE', where in some sense, TINE becomes the OS.  
 * Multi-process OSes (UNIX, WINDOWS, VxWorks, VMS, etc.) use 'FALSE', where the
 * TINE engines only wakes up if 'select()' fires (incoming data) or if the 
 * SystemPolling rate so dictates.
 */
int gIsCyclePollingMode = SERVER_POLLING_MODE;
/**
 * \internal
 * \brief fixes systematics of multicast address formulation
 */
int gMultiCastUseLegacyRules = FALSE;
int gDcsMultiCastUseLegacyRules = TRUE;
TINE_EXPORT void SetUseMCastLegacyRules(int value)
{
  gMultiCastUseLegacyRules = value;
  return;
}
TINE_EXPORT int GetUseMCastLegacyRules(void)
{
  return gMultiCastUseLegacyRules;
}
/**
 * \internal
 * \brief assign multicast address/pattern list
 */
MCastAddrMap *gtMCastAddrLst = NULL;
int nMCastAddrLstEntries = 0;

/**
 * \internal
 * \brief Time to live for multicasts (no. of router hops)
 */
int gtMCastTTL = MCAST_TTL;
UINT32 sendtoErrcnt = 0;
UINT32 selectErrcnt = 0;
UINT32 selectFailcnt = 0;

char *gLoopbackAddr = LOOPBACKADDR_IPV4;
/**
 * \internal
 * \brief Canonical address family for this build
 */
int gAfInet = AF_INET;
int gInAdrLen = sizeof(struct in_addr);
TINE_EXPORT int SetAfInet(int family)
{
  if (family != AF_INET6) 
  { /* then it's AF_INET no matter what */
    family = AF_INET;
    gInAdrLen = sizeof(struct in_addr);
    gLoopbackAddr = LOOPBACKADDR_IPV4;
  }
  else 
  { /* is AF_INET6 ! */
    gInAdrLen = sizeof(struct in6_addr);
    gLoopbackAddr = LOOPBACKADDR_IPV6;
    gtNetSrvAddr = NETSRV_ADDR6;
    gtMCastAddr = MCAST_ADDR6;
  }
  feclog("INET family set to %s",family == AF_INET6 ? "IPv6" : "IPv4");
  gAfInet = family;
  return gAfInet;
}
TINE_EXPORT int GetAfInet(void)
{
  return gAfInet;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int SetSrvUDPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtUDPPort = port;
  return 0;
}
TINE_EXPORT int GetSrvUDPPort(void)
{
  return gtUDPPort;
}
TINE_EXPORT int SetSrvTCPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtTCPPort = port;
  return 0;
}
TINE_EXPORT int GetSrvTCPPort(void)
{
  return gtTCPPort;
}
TINE_EXPORT int SetGCastPort(int port)
{ 
  if (port < 1024) return not_allowed;
  gtGCastPort = port;
  return 0;
}
TINE_EXPORT int GetGCastPort(void)
{
  return gtGCastPort;
}
TINE_EXPORT int SetMCastPort(int port)
{
  if (port < 1024) return not_allowed;
  gtMCastPort = port;
  return 0;
}
TINE_EXPORT int GetMCastPort(void)
{
  return gtMCastPort;
}
TINE_EXPORT int SetClnUDPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtUDPClnPort = port;
  return 0;
}
TINE_EXPORT int GetClnUDPPort(void)
{
  return gtUDPClnPort;
}
int isInvalidIpV4Addr(char *addr)
{ /* must be of the form a.b.c.d */
  int i,n=0;
  char *c;
  if (addr == NULL || strlen(addr) == 0) return argument_list_error;
  for (i=0,c=addr; i<16 && *c != 0; i++, c++)
  {
    switch (*c)
    {
      case ':': 
      case '|': i = 16; break; /* added port at the end? */
      case '.': n++; break;
      default: if (!isdigit(*c)) return not_allowed;
    }
  }
  return n == 3 ? 0 : not_allowed;
}
int isInvalidIpV6Addr(char *addr)
{ /* must be of general form a:b:c:d::e:f:g:h */
  int i,n=0,abbr=FALSE;
  char *c,ch;
  if (addr == NULL || strlen(addr) == 0) return argument_list_error;
  if (isIpv4on6(addr)) return 0; /* mapped ipv4 address is okay */
  if (strstr(addr,"::") != NULL) abbr = TRUE;
  for (i=0,c=addr; i<40 && *c != 0; i++, c++)
  {
    switch (ch = toupper(*c))
    {
      case '|': i = 40; break; /* added port at the end? */
      case ':': n++; break;
      default: 
        if (isdigit(ch)) break;
        if (ch >= 'A' && ch <= 'F') break;
        return not_allowed;
    }
  }
  if (!abbr && n < 8) return not_allowed;
  return 0;
}
TINE_EXPORT int SetNetSrvAddr(char *addr)
{
  static char mcaddr[64];
  int cc = isInvalidIpV4Addr(addr);
  if (cc != 0) cc = isInvalidIpV6Addr(addr);
  if (cc != 0) return cc;
  strncpy(mcaddr,addr,40);
  gtNetSrvAddr = mcaddr;
  return 0;
}
TINE_EXPORT char *GetNetSrvAddr(void)
{
  return gtNetSrvAddr;
}
int _setMcstAdr(char *addr,char *buf,int islegacy)
{
  int cc = isInvalidIpV4Addr(addr);
  if (cc != 0) cc = isInvalidIpV6Addr(addr);
  if (cc != 0) return cc;
  strncpy(buf,addr,40);
  if (islegacy) gtMCastAddrLegacy = buf; else gtMCastAddr = buf;
  feclog("assign %sserver multicast address to %s",(islegacy ? "legacy " : ""),addr);
  return 0;
}
TINE_EXPORT int SetMCastAddr(char *addr)
{ /* sets the generic gcast address */
  return _setMcstAdr(addr,gtMCastAddrBuffer,FALSE);;
}
TINE_EXPORT char *GetMCastAddr(void)
{
  return gtMCastAddr;
}
TINE_EXPORT int SetMCastAddrLegacy(char *addr)
{ /* sets the generic gcast address */
  static char mcladr[64];
  return _setMcstAdr(addr,mcladr,TRUE);
}
TINE_EXPORT char *GetMCastAddrLegacy(void)
{
  return gtMCastAddrLegacy;
}
TINE_EXPORT int SetMCastMask(char *addr)
{ /* accept addr in the form of 'mask1|pattern1,mask2|pattern2,...' */
  char buf[1030];
  char *mm,*mp,*c,*cp,*tok = ",";
  int n=1;
  if (addr == NULL || strlen(addr) == 0) return argument_list_error;
  if (strlen(addr) > 1024) return string_too_long;
  strncpy(buf,addr,1024); buf[1024] = 0;
  for (c=buf; c != NULL; c++)
  { /* count the commas */
    if ((c=strchr(c,',')) == NULL) break;
    n++;
  }
  SystemFree(gtMCastAddrLst); nMCastAddrLstEntries = 0;
  if ((gtMCastAddrLst=(MCastAddrMap *)SystemCalloc(n,sizeof(MCastAddrMap))) == NULL) return out_of_local_memory;
  for (n=0,mm=strtok_r(buf,tok,&cp); mm != NULL; mm=strtok_r(NULL,tok,&cp))
  {
    if ((c=strchr(mm,'|')) != NULL)
    {
      *c++ = 0;
      mp = c;
      if (isInvalidIpV4Addr(mp) && isInvalidIpV6Addr(mp))
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "assign multicast address pattern %.40s is invalid",mp);
        continue;
      }
    }
    else 
    {
      mp = getmyipaddrstr();
    }
    if (isInvalidIpV4Addr(mm) && isInvalidIpV6Addr(mm))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "assign multicast address %.40s is invalid",mm);
      continue;
    }
    strncpy(gtMCastAddrLst[n].mcStrAdr,mm,ADDR_SIZE_IPV6);
    strncpy(gtMCastAddrLst[n].patStrAdr,mp,ADDR_SIZE_IPV6);    
    inet_pton(gAfInet,gtMCastAddrLst[n].mcStrAdr,&gtMCastAddrLst[n].mcInAdr);
    inet_pton(gAfInet,gtMCastAddrLst[n].patStrAdr,&gtMCastAddrLst[n].patInAdr);
    feclog("set server multicast address mask for pattern %.40s to %.40s",
      gtMCastAddrLst[n].mcStrAdr,gtMCastAddrLst[n].patStrAdr);
    n++;
  }
  nMCastAddrLstEntries = n;
  return 0;
}
TINE_EXPORT char *GetMCastMask(char *pattern)
{
  int i;
  INADR inadr;
  BYTE *madr,*padr;
  if (pattern == NULL || nMCastAddrLstEntries == 0) goto err;
  inet_pton(gAfInet,pattern,&inadr); padr = (BYTE *)&inadr;
  for (i=0; i<nMCastAddrLstEntries; i++)
  {
    madr = (BYTE *)&gtMCastAddrLst[i].mcInAdr;
    if (padr[0] != madr[0]) continue;
    if (gMultiCastUseLegacyRules && padr[1] != madr[1]) continue;
    return gtMCastAddrLst[i].mcStrAdr;
  }
err:
  return GetMCastAddr();
}
TINE_EXPORT int SetGCastAddr(char *addr)
{ /* synonymous with SetMCastAddr() */
  return SetMCastAddr(addr);
}
TINE_EXPORT int SetGCastMask(char *addr)
{
  return SetMCastMask(addr);
}
TINE_EXPORT char *GetGCastAddr(void)
{
  return GetMCastAddr();
}
TINE_EXPORT char *GetGCastMask(char *addr)
{
  return GetMCastMask(addr);
}
TINE_EXPORT int SetMCastTTL(int ttl)
{
  if (ttl < 1) return not_allowed;
  gtMCastTTL = ttl;
  feclog("multicast TTL set to %d",ttl);
  return 0;
}
TINE_EXPORT int GetMCastTTL(void)
{
  return gtMCastTTL;
}
INADR fixMcastHaddr(BYTE *bAdr,int legacyRules)
{ /* internal: params never NULL */
  int i;
  INADR inadr, mcAdr;
  BYTE hAdr[16], *padr;
  char *dAdr = legacyRules ?  gtMCastAddrLegacy : gtMCastAddr;
  memcpy(hAdr,bAdr,gInAdrLen);
  inet_pton(gAfInet,dAdr,&mcAdr);
  for (i=0; i<nMCastAddrLstEntries; i++)
  { /* run thru the list of configured multicast address mapping */
    padr = (BYTE *)&gtMCastAddrLst[i].patInAdr;
    if (padr[0] != hAdr[0]) continue;
    if (legacyRules && padr[1] != hAdr[1]) continue;
    mcAdr = gtMCastAddrLst[i].mcInAdr;
    break;
  }
  hAdr[0] = ((BYTE *)&mcAdr)[0]; /* e.g. 239 or 238 or 0xff */
  if (legacyRules) hAdr[1] = ((BYTE *)&mcAdr)[1]; /* e.g. 1 */
  if (gInAdrLen == 16) memcpy(hAdr,&mcAdr,8); /* ipv6 : e.g. ff05:: */
  memcpy(&inadr,hAdr,gInAdrLen); /* "a.b.c.d" -> "239.b.c.d" or "a.b.c.d" -> "238.1.c.d" */
  return inadr;
}
void initAddrPortsFromCsv(void);
void initAddrPortsFromEnv(void)
{
  char *ptr=NULL;
  if ((ptr=getenv("TINE_MCAST_ADDR")) != NULL) SetMCastAddr(ptr);
  if ((ptr=getenv("TINE_GCAST_ADDR")) != NULL) SetGCastAddr(ptr);

  if ((ptr=getenv("TINE_MCAST_MASK")) != NULL) SetMCastMask(ptr);
  if ((ptr=getenv("TINE_GCAST_MASK")) != NULL) SetGCastMask(ptr);
  return;
}
int dumpMCastAddrList(void)
{
  int i;
  ttyoutput("multicast address mappings (%s conditions)",(gMultiCastUseLegacyRules ? "legacy" : "modern"));
  ttyoutput("default: %.40s -> %40s",getmyipaddrstr(),GetMCastAddr());
  for (i=0; i<nMCastAddrLstEntries; i++)
  {
    ttyoutput(" [%d]   : %.40s -> %40s",i,gtMCastAddrLst[i].patStrAdr,gtMCastAddrLst[i].mcStrAdr);
  }
  return 0;
}
void initSockBufferSizeFromEnv(void)
{
  char *ptr=NULL;
  UINT32 rcvsiz = MAX_RCVBUF, sndsiz = MAX_SNDBUF;
  int assign = FALSE;
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  if (max_clnrcvbuf != MAX_RCVBUF) 
  { /* already set via API: use the API setting */
    rcvsiz = max_clnrcvbuf;
  }
  else if ((ptr=getenv("TINE_RECV_BUFFER_SIZE")) != NULL) 
  {
    rcvsiz = (UINT32)strtol(ptr,NULL,0);
    assign = TRUE;
  }
  if (max_srvsndbuf != MAX_SNDBUF) 
  { /* already set via API: use the API setting */
    sndsiz = max_srvsndbuf;
  }
  else if ((ptr=getenv("TINE_SEND_BUFFER_SIZE")) != NULL) 
  {
    sndsiz = (UINT32)strtol(ptr,NULL,0);
    assign = TRUE;
  }
  if (assign)
  {
    _SystemAssignBufferSpace(rcvsiz,sndsiz);
    if (sndsiz < (UINT32)max_udpsndbuf) _SystemAssignUdpSendBufferSpace(sndsiz);
  }
}
void dumpSckSettings(SOCKET sck)
{
  int optval = 1;
  _socklen_t optlen = sizeof(_socklen_t);
  if (getsockopt(sck,SOL_SOCKET,SO_RCVBUF,(char *)&optval,&optlen) == 0)
  {
    ttyoutput("\tRCV buffer space : %d bytes\n",optval);
  }
  if (getsockopt(sck,SOL_SOCKET,SO_SNDBUF,(char *)&optval,&optlen) == 0)
  {
    ttyoutput("\tSND buffer space : %d bytes\n",optval);
  }
}
char *getSckStr(SOCKET sck)
{
  static char sckstr[64];
# if defined(WIN32) && defined(OS_64BIT)
  sprintf(sckstr,"%Id",sck);
# else
  sprintf(sckstr,"%d",sck);
# endif
  return sckstr;
}
int dumpSocketSettings(void)
{
  int i;
  TCPSCKT *sckt;
  ttyoutput("Server Socket Settings (%s) :\n",(gAfInet == AF_INET6 ? "IPv6" : "IPv4"));
  if (!gIsRunningAsServer) ttyoutput("NOT running as server!\n");
  ttyoutput("UDP server socket (udpSrvSck) : %s\n",getSckStr(udpSrvSck));
  if (udpSrvSck != 0) dumpSckSettings(udpSrvSck);
  ttyoutput("UDP network address request socket (netSrvSck) : %s\n",getSckStr(netSrvSck));
  if (netSrvSck != 0) dumpSckSettings(netSrvSck);
  ttyoutput("TCP server socket (tcpSrvSck) : %s\n",getSckStr(tcpSrvSck));  
  for (i=0; i<nTCPsck; i++)
  {
    if (tcpSckTbl[i].type&(TCP|TCP6))
    {
      ttyoutput("TCP server daughter socket : %s\n",getSckStr(tcpSckTbl[i].sck));
      if (tcpSckTbl[i].sck != 0) dumpSckSettings(tcpSckTbl[i].sck);
    }
  }
  ttyoutput("TCP server stream socket (tcpStrmSck) : %s\n",getSckStr(tcpStrmSck));
  for (i=0; i<nTCPsck; i++)
  {
    if (tcpSckTbl[i].type&(STREAM|STREAM6))
    {
      ttyoutput("STREAM server daughter socket : %s\n",getSckStr(tcpSckTbl[i].sck));
      if (tcpSckTbl[i].sck != 0) dumpSckSettings(tcpSckTbl[i].sck);
    }
  }
  ttyoutput("TCP server debug socket (dbgSrvSck) : %s\n",getSckStr(dbgSrvSck));
  for (i=0; i<nTCPsck; i++)
  {
    if (tcpSckTbl[i].type&(DBGPIPE|DBGPIPE6))
    {
      ttyoutput("server debug daughter socket : %s\n",getSckStr(tcpSckTbl[i].sck));
      if (tcpSckTbl[i].sck != 0) dumpSckSettings(tcpSckTbl[i].sck);
    }
  } 
  ttyoutput("Client Socket Settings      :\n");
  ttyoutput("UDP client socket (udpClnSck) : %s\n",getSckStr(udpClnSck));
  if (udpClnSck != 0) dumpSckSettings(udpClnSck);
  ttyoutput("UDP client multicast socket (mcastClnSck) : %s\n",getSckStr(mcastClnSck));
  if (mcastClnSck != 0) dumpSckSettings(mcastClnSck);
  ttyoutput("UDP client globals socket (glbClnSck) : %s\n",getSckStr(glbClnSck));
  if (glbClnSck != 0) dumpSckSettings(glbClnSck);
  ttyoutput("UDP client socket for trivial synchronous calls (udpSyncSck) : %s\n",getSckStr(udpSyncSck));
  if (udpSyncSck != 0) dumpSckSettings(udpSyncSck);

  for (sckt = tcpClnSckTbl; sckt != NULL; sckt = sckt->nxt)
  {
    ttyoutput("TCP/STREAM client socket : %s\n", getSckStr(sckt->sck));
    dumpSckSettings(sckt->sck);
  }
  ttyoutput("TCP send transmission errors :\n");
  ttyoutput("total select errors on write set : %u\n",selectFailcnt);
  ttyoutput("current select errors on write set : %u\n",selectErrcnt);
  ttyoutput("total send errors : %u\n",sendtoErrcnt);

  return 0;
}
#endif /* not SMALL_TINE_LIB */

UINT32 TCPcount = 0, UDPcount = 0,
	     SPXcount = 0, IPXcount = 0,
	     BRDcount = 0;
char gMyipAddrBuffer[64];
int ipLoaded = FALSE, ipxLoaded = FALSE;
int gPacketMtu = UDP_BUFFER_SIZE;
TINE_EXPORT int SetPacketMTU(int mtu)
{
  if (mtu > UDP_MAXBUFFER_SIZE) return not_allowed;
  if (mtu <= 0) mtu = 0;
  else if (mtu < 512) mtu = 512;
  feclog("packet MTU set to %d",mtu);
  gPacketMtu = mtu;
  if (!gInitIPClientDone)
  {
    if (gPacketMtu > max_udpsndbuf/2) 
      _SystemAssignUdpSendBufferSpace((UINT32)gPacketMtu*2);
  }
  return 0;
}
TINE_EXPORT int GetPacketMTU(void)
{
  return gPacketMtu;
}
struct pollfd *ext_pollfds = NULL;
int next_pollfds = 0;
#ifdef EXT_FDSET
int externalFdSets = -1;
#else
int externalFdSets = 0;
#endif
void FD_MERGE(fd_set fdsrc,fd_set *fddst)
{
  int i;
  for (i=0; i<(int)sizeof(fd_set); i++)
    ((BYTE *)fddst)[i] |= ((BYTE *)&fdsrc)[i];
}
#if defined(WIN32)
int isWindowsNT = TRUE;
int WSAinitialized = -1;
int WinsockLoaded(void)
{
  static WSADATA wsaData;
  static WORD wVersionRequested;
  char *ptr;
  if((ptr=getenv("OS")) != NULL && !stricmp(ptr,"Windows_NT"))
  {
    wVersionRequested = MAKEWORD( 2, 2 );
  }
  else
  {
    wVersionRequested = MAKEWORD( 1, 1 );
    isWindowsNT = FALSE;
  } 
	if (WSAinitialized == -1) 
  {
    int rc = WSAStartup(wVersionRequested,&wsaData);
    WSAinitialized = (rc == 0) ? 1 : 0;
    if (gIsRunningAsServer)
    {
      if( WSAinitialized) feclogEx(TINE_LOGLEVEL_INFO, "Windows Socket startup success");
      else feclogEx(TINE_LOGLEVEL_ERROR, "Windows Socket startup failed, err : %x",rc);
    }
    /* feclog("Windows Socket startup %s",WSAinitialized == 0 ? "success" : "failed"); */
  }
  return WSAinitialized;
}
int GetErrno(void)
{
  int err = WSAGetLastError();
  switch (err)
  { /* map some common errors to BSD */
    case WSAEINTR: return EINTR;
    case WSAEWOULDBLOCK: return EWOULDBLOCK;
    case WSAEISCONN: return EISCONN;
    case WSAENOBUFS: return ENOBUFS;
  }
  return err;
}
#define isSckInvalid(s) ((s) == INVALID_SOCKET)
#elif defined(NEED_POLL)
#define isSckInvalid(s) ((s) < 0 || (s) >= FD_SETSIZE)
#else
#define isSckInvalid(s) ((s) < 0)
#endif /* WIN32 */

/**
 * \internal
 * \brief Maximal number of receive buffers for socket options.
 */
int max_clnrcvbuf = MAX_RCVBUF;
int max_srvrcvbuf = MAX_RCVBUF;

/**
 * \internal
 * \brief Maximal number of send buffers for socket options.
 */
int max_clnsndbuf = MAX_SNDBUF;
int max_srvsndbuf = MAX_SNDBUF;
int max_udpsndbuf = MAX_SNDBUF;

TINE_EXPORT int _SystemAssignUdpSendBufferSpace(UINT32 sndBufferSpace)
{
  if (sndBufferSpace < 0x1000) sndBufferSpace = 0x1000;
  max_udpsndbuf = (int)sndBufferSpace;
  feclog("udp client socket send buffer space set to %d bytes",max_udpsndbuf);
  return 0;
}
TINE_EXPORT int _SystemAssignBufferSpace(UINT32 rcvBufferSpace,UINT32 sndBufferSpace)
{
  return _SystemAssignBufferSpaceEx(rcvBufferSpace,0,0,sndBufferSpace);
}
TINE_EXPORT int _SystemAssignBufferSpaceEx(UINT32 clnRcvBufferSpace,UINT32 srvRcvBufferSpace,UINT32 clnSndBufferSpace,UINT32 srvSndBufferSpace)
{
  if (clnRcvBufferSpace > 0 || srvRcvBufferSpace > 0)
  {
    if (clnRcvBufferSpace == 0) clnRcvBufferSpace = max_clnrcvbuf;
    if (clnRcvBufferSpace < 0x1000) clnRcvBufferSpace = 0x1000; 
    max_clnrcvbuf = (int)clnRcvBufferSpace;
    if (srvRcvBufferSpace == 0) 
    { /* assume that a generic server will not be 'sent' large payloads of data */
      srvRcvBufferSpace = (UINT32)max_clnrcvbuf;
      if (srvRcvBufferSpace > MAX_RCVBUF*2) srvRcvBufferSpace /= 2;
    }
    if (srvRcvBufferSpace < 0x1000) srvRcvBufferSpace = 0x1000; 
    max_srvrcvbuf = (int)srvRcvBufferSpace;
    feclog("client socket recv buffer space set to %d bytes",max_clnrcvbuf);
    feclog("server socket recv buffer space set to %d bytes",max_srvrcvbuf);
  }
  if (clnSndBufferSpace > 0 || srvSndBufferSpace > 0)
  {
    if (srvSndBufferSpace == 0) srvSndBufferSpace = max_srvsndbuf;
    if (srvSndBufferSpace < 0x1000) srvSndBufferSpace = 0x1000; 
    max_srvsndbuf = (int)srvSndBufferSpace;
    if (clnSndBufferSpace == 0)
    { /* assume that a generic client will not be 'sending' large payloads of data */
      clnSndBufferSpace = (UINT32)max_srvsndbuf;
      if (clnSndBufferSpace > MAX_SNDBUF*2) clnSndBufferSpace /= 2;
    }
    max_clnsndbuf = (int)clnSndBufferSpace;
    feclog("client socket send buffer space set to %d bytes",max_clnsndbuf);
    feclog("server socket send buffer space set to %d bytes",max_srvsndbuf);
  }
  return 0;
}
int gSufferInSilence = FALSE; /* for special cases that need to suppress output */
TINE_EXPORT void SetSufferInSilence(int value)
{
  gSufferInSilence = value;
}
/**
 * \internal
 * \param dbgbuf is a optional user-supplied buffer: if != NULL then suppress
 *  anything that might lead to a call to dbglog() as there is liable to be a mutex 
 *  concurrency issue !
 */
void rprtSckErr(const char *src,const char *txt,SOCKET sck,int err, int debug,char *dbgbuf)
{
  char msgbuf[256], errbuf[256] = "", scktxt[64] = "", *pbuf;
  static char last_txt[64] = "";
  static int last_errno = 0;
  static time_t last_logged = 0;
  time_t t = time(NULL);
  int logit = FALSE;
  int dbgit = (tineDebug >= debug || errorDebug);

# if defined(WIN32)
  LPVOID lpMsgBuf;
  DWORD lastErr = (DWORD)(err < 0 ? GetErrno() : err);
  if (ServerExitCondition) return;
  if (lastErr != (DWORD)last_errno) logit = TRUE;
  /* un map BSD back to WSA */
  switch (lastErr)
  {
    case EINTR: lastErr = WSAEINTR; break;
    case ECONNABORTED: lastErr = WSAECONNABORTED;  break;
    case EWOULDBLOCK: lastErr = WSAEWOULDBLOCK; break;
    case EISCONN: lastErr = WSAEISCONN; break;
    case ENOBUFS: lastErr = WSAENOBUFS; break;
    case EIO: lastErr = WSAGetLastError(); break;
    default: break;
  }
  if (FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,lastErr,
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), 
        (LPTSTR) &lpMsgBuf,0,NULL) > 0)
  {
    snprintf(errbuf,256,"%s : %s",txt,(char *)lpMsgBuf);
    LocalFree( lpMsgBuf );
  }
  else
  { 
    snprintf(errbuf,256,"%s : %d",txt,lastErr);
  }
  if (lastErr == 0 && !strcmp(txt,"select")) snprintf(errbuf,256,"%s : timeout or hangup",txt);
  last_errno = (int)lastErr;
  if (sck != (SOCKET)0)
  {
#   ifdef VS_LEGACY_BUILD
    sprintf(scktxt, "(socket: %d)", sck);
#   else
    sprintf(scktxt, "(socket: %Id)", sck);
#   endif
  }
# elif defined(UNIX) || defined(VXWORKS)
  int eno = err < 0 ? errno : err;
  if (ServerExitCondition) return;
  if (eno == EIO) err = errno;
  if (eno != last_errno) logit = TRUE;
  last_errno = eno;
  snprintf(errbuf,256,"%s : %s",txt,strerror(eno));
  if (eno == 0 && !strcmp(txt,"select")) snprintf(errbuf,256,"%s : timeout or hangup",txt);
  if (sck != (SOCKET)0) sprintf(scktxt,"(socket: %d)",sck);
# else
  if (ServerExitCondition) return;
  if (eno != last_errno) logit = TRUE;
  last_errno = eno;
  snprintf(errbuf,256,"%s : %d",txt,eno);
  if (sck != (SOCKET)0) sprintf(scktxt,"(socket: %d)",sck);
# endif
  if (strncmp(txt,last_txt,64)) logit = TRUE;
  strncpy(last_txt,txt,64);
  if (logit == FALSE && t > last_logged) logit = TRUE;
  pbuf = (dbgbuf != NULL) ? dbgbuf : msgbuf;
  snprintf(pbuf,255,"in %.16s %.32s: %.192s",src,scktxt,errbuf);
  if (logit) 
  {
    int restore = FALSE;
    if (_sendToDbg == 0) 
    { /* output to console or not ? */
      _sendToDbg = (dbgbuf != NULL) ? -1 : tineDebug;
      restore = TRUE;
    }
    feclogEx(TINE_LOGLEVEL_ERROR,pbuf);
    if (restore) _sendToDbg = 0;
    last_logged = t;
    dbgit = FALSE; /* already output to console via feclog */
  }
  if (gSufferInSilence) return;
# ifndef DOS
  if (dbgit && dbgbuf == NULL) dbglog("%s : %s",getDataTimeString(makeDataTimeStamp(),0),pbuf);
# else
  if (dbgit) soperror(txt);
# endif
}
TCPBCKT *bcktList = NULL;
TCPBCKT *findBucket(SOCKET sck,int lock)
{
  TCPBCKT *bckt=NULL;
  if (sck == 0) return NULL;
  if (WaitForSystemMutex(hTcpBucketMutex,-1) != 0) return NULL;
  for (bckt=bcktList; bckt != NULL && bckt->sck != (SOCKET)sck; bckt=bckt->nxt);
  if (!lock || bckt == NULL) ReleaseSystemMutex(hTcpBucketMutex);
  return bckt;
}
void removeFlaggedBuckets(void)
{
# ifdef MULTITHREADED
  TCPBCKT *bckt=NULL;
  if (WaitForSystemMutex(hTcpBucketMutex,-1) != 0) return;
  for (bckt=bcktList; bckt != NULL; bckt=bckt->nxt)
  {
    if (bckt->rmvNeeded && bckt->sdThrState != SD_THREAD_STOPPED) 
    { /* CloseConnectionSocket() frees bckt => return and come back later ! */
      ReleaseSystemMutex(hTcpBucketMutex);
      CloseConnectionSocket(bckt->sck,bckt->inetprot,TRUE);
      return;
    }
  }
  ReleaseSystemMutex(hTcpBucketMutex);
# endif
  return;
}
/**
 * \internal
 * \param sck the tcp socket
 * \param buflen the desired buffer length to hold the collected results
 * \param clrlen the collector buffer length to hold incoming traffic
 */
TCPBCKT *getBucket(SOCKET sck,int buflen,int clrlen,int inetprot,int tineprot,int source)
{
  TCPBCKT *bckt=NULL;
  int cc = 0;
# ifdef MULTITHREADED
  int needRcvThread=FALSE, needSndThread=FALSE;
  SEMHANDLE sem = NULL;
  if (!(inetprot&(STREAM|STREAM6|DBGPIPE|DBGPIPE6)))
  {
    if (source == SRV_RSP_)
    { /* a server-side bucket */
      needSndThread = gRunTransportInSeparateThread; 
    }
    else
    { /* a client-side bucket */
      needRcvThread = TRUE;
    }
  }
# else
  UNUSED(source);
# endif
  if (gClientIsClosing) return NULL;
  if (WaitForSystemMutex(hTcpBucketMutex,-1) != 0) return NULL;
  for (bckt=bcktList; bckt != NULL; bckt=bckt->nxt) if (bckt->sck == sck)
  {
    msglog(0, "attaching to TCP bucket for socket %d", sck);
    ccerr(0);
  }
  if ((bckt=(TCPBCKT *)SystemCalloc(1,sizeof(TCPBCKT))) != NULL)
  {
    bckt->buf = (BYTE *)SystemCalloc(1, buflen);
    bckt->len = bckt->buf != NULL ? buflen : 0;
    bckt->clrbuf = (BYTE *)SystemCalloc(1, clrlen);
    bckt->clrlen = bckt->buf != NULL ? clrlen : 0;
    bckt->sck = sck;
    bckt->inetprot = inetprot;
    bckt->tineprot = tineprot;
#   ifdef MULTITHREADED
    bckt->sdThrState = SD_THREAD_STOPPED;
    if (needRcvThread)
    {
      bckt->sdThrState = SD_THREAD_CREATED;
      bckt->rcvThr = sckRecvCreateThread(sck,inetprot);
    }
    if (needSndThread)
    {
      bckt->sdTblSize = 2 * srvWorkAreaSize/buflen;
      bckt->sdTbl = (StreamDataTable *)SystemCalloc(bckt->sdTblSize,sizeof(StreamDataTable));
      if (bckt->sdTbl == NULL)
      {
        msglog(out_of_local_memory, "out of memory for bucket for socket %d", sck);
        SystemFree(bckt);
        ccerr(out_of_local_memory);
      }
      bckt->sdThrState = SD_THREAD_CREATED;
      bckt->sndThr = sckSendCreateThread(sck,inetprot,&sem);
      bckt->sndSem = sem;
    }
#   endif
    bckt->nxt = bcktList;
    bcktList = bckt;
    msglog(0, "new TCP bucket for socket %d", sck);
  }
err:
  if (cc) bckt = NULL;
  ReleaseSystemMutex(hTcpBucketMutex);
  return bckt;
}
#ifdef MULTITHREADED
int threadExists(THRHANDLE t, int timeout)
{
# if defined(WIN32)
  DWORD result = WaitForSingleObject(t, (DWORD)timeout);
  if (result == WAIT_TIMEOUT) 
  {
    return TRUE;
  }
# elif defined(UNIX)
  /* not the best solution, but it should always work for us */
  UNUSED(timeout);
  if (pthread_kill(t, 0) == 0)
  {
    return TRUE;
  }
# endif
  return FALSE;
}
#endif
void freeBucket(SOCKET sck,int yield)
{
  TCPBCKT *bckt, *b;
  int cc=0, hasThread=FALSE;
  if ((bckt=findBucket(sck,TRUE)) == NULL) return;
  if (bckt == bcktList) bcktList = bckt->nxt;
  else
  {
    for (b=bcktList; b != NULL && b->nxt != bckt; b = b->nxt);
    if (b == NULL || b->nxt == NULL) ccerr(non_existent);
    b->nxt = bckt->nxt;
  }
# ifdef MULTITHREADED
  hasThread = ((bckt->sndThr != (THRHANDLE)0) || (bckt->rcvThr != (THRHANDLE)0)) ? TRUE : FALSE;
  if (yield && hasThread && bckt->sdThrState != SD_THREAD_STOPPED)
  { /* sck thread needs to be stopped */
    bckt->sdThrState = SD_THREAD_STOPPING;
    if (bckt->sndSem != (SEMHANDLE)0) ReleaseSystemSemaphore(bckt->sndSem);
  }
# else
  UNUSED(yield);
  UNUSED(hasThread);
  SystemFree(bckt);
# endif
err:
  ReleaseSystemMutex(hTcpBucketMutex);
  if (cc != 0 && tineDebug) dbglog("free TCP socket bucket: %.32s",cc2str(cc));
}
void freeBucketList(void)
{
  TCPBCKT *bckt=NULL;
  if (WaitForSystemMutex(hTcpBucketMutex,-1) != 0) return;
  msglog(0, "freeing TCP bucket list");
  bckt = bcktList;
  while (bckt != NULL)
  {
    bcktList = bckt->nxt;
#   ifdef MULTITHREADED
    if (bckt->sdThrState != SD_THREAD_STOPPED)
    {
      bckt->sdThrState = SD_THREAD_STOPPING;
      soclose(bckt->sck);
      if (bckt->sndSem != (SEMHANDLE)0) ReleaseSystemSemaphore(bckt->sndSem);
    }
#   else
    SystemFree(bckt);
#   endif
    bckt = bcktList;
  }
  bcktList = NULL;
  ReleaseSystemMutex(hTcpBucketMutex);
  return;
}
int dumpBucketList(void)
{
  TCPBCKT *b;
# ifdef MULTITHREADED
  int i, p;
  char scratch[128];
  ttyoutput("Queue overwrite : %s",streamQueueOverwriteEager ? "EAGER" : "LAST RESORT");
# endif
  ttyoutput("Current TCP/IP bucket list :\n");
  if (bcktList == NULL)
  {
    ttyoutput("no TCP/STREAM connections are active");
    return 0;
  }
  if (WaitForSystemMutex(hTcpBucketMutex,-1) != 0) return mutex_error;
  for (b=bcktList; b != NULL; b=b->nxt)
  {
    ttyoutput("socket : %d",b->sck);
    ttyoutput("buffer length : %d bytes",b->len);
    ttyoutput("collector length : %d bytes",b->clrlen);
#   ifdef MULTITHREADED
    ttyoutput("queue head : %d",b->sdTblHead);
    ttyoutput("queue tail : %d",b->sdTblTail);
    ttyoutput("queue mark : %d",b->sdTblMark - 1);
    ttyoutput("queue size : %d segments",b->sdTblSize);
    ttyoutput("queue segment size : %d bytes",TCP_BUFFER_SIZE);
    ttyoutput("queue is waiting for mark : %s",b->sdTblWaitMark ? "TRUE" : "FALSE");
    if (b->sdTbl == NULL) continue;
    ttyoutput("queue segments to send:\n");
    for (i=0, p = b->sdTblTail; i < b->sdTblSize && p != b->sdTblHead; i++, p = (p+1) % b->sdTblSize)
    {
      if (!b->sdTbl[p].isbusy) continue;
      inetStrAdr(&b->sdTbl[p].dst.adr.ip,scratch,FALSE);
      ttyoutput("segment %d : %d bytes to %.16s %.40s (%s)",
        p,b->sdTbl[p].msgsize,b->sdTbl[p].dst.usr,scratch,
        getProtocolAsString(b->sdTbl[p].dst.inetProtocol));
    }
#   endif
  }
  ReleaseSystemMutex(hTcpBucketMutex);
  return 0;
}
#if defined(UNIX) && !defined(NO_TINE_PIPE)
void ClosePipeSocket(int sck)
{
  int i, isSrvSck=FALSE;
  char pipename[32];

  close(sck);
  freeBucket(sck,FALSE);

  pipename[0] = 0;
  for (i=0; i<nIPCsck; i++) 
  { /* am I a server sending data to a client ? */
    if (ipcSckTbl[i].sck == sck)
    {
      isSrvSck = TRUE;
      strncpy(pipename,ipcSckTbl[i].name,16);
      ipcSckTbl[i].sck = 0; ipcSckTbl[i].type = 0;
      nIPCsck--;
      break;
    }
  }
  if (!isSrvSck) goto chkClnSck;
  for (; i<nIPCsck; i++) ipcSckTbl[i] = ipcSckTbl[i+1];
  for (i=0; i<nconsumers; i++) 
  { /* flush any cross-references in the client table */
    if (ClnTbl[i]->ep.sck == sck) 
    { 
      msglog(0,"closing pipe connection %.16s to %.16s",pipename,ClnTbl[i]->usr);
      ClnTbl[i]->ep.sck = 0; 
    }
  }
  return; /* finished */
chkClnSck:
  for (i=0; i<numFecTblEntries; i++)
  { /* am I a client trying to contact a server ? */
    if (sck == FecTbl[i].lclAdr.pipe.sck)
    {
      msglog(0,"closing pipe connection to %.16s",FecTbl[i].fecName);
      FecTbl[i].inetProtocol = 0;
      FecTbl[i].lclAdr.pipe.sck = 0;
      rmvClnPipeSck(sck);
      break;
    }
  }
}
ClnPipeSck gClnPipeSckTbl[PIPE_TBL_CAPACITY];
int nClnPipeSck = 0;
int addClnPipeSck(int sck,int idx)
{
  int i, slot = -1;
  if (sck == 0 || idx < 0) return invalid_parameter;
  for (i=0; i<nClnPipeSck; i++)
  {
    if (gClnPipeSckTbl[i].sck == sck) 
    {
      gClnPipeSckTbl[i].fecIdx = idx;
      return 0;
    }
    if (slot < 0 && gClnPipeSckTbl[i].sck == 0) slot = i;
  }
  if (slot < 0) slot = nClnPipeSck;
  if (slot == PIPE_TBL_CAPACITY) return resources_exhausted;
  gClnPipeSckTbl[slot].sck = sck;
  gClnPipeSckTbl[slot].fecIdx = idx;
  if (slot == nClnPipeSck) nClnPipeSck++;
  return 0;
}
int rmvClnPipeSck(int sck)
{
  int i;
  for (i=0; i<nClnPipeSck; i++)
  {
    if (gClnPipeSckTbl[i].sck == sck)
    {
      gClnPipeSckTbl[i].sck = 0;
      gClnPipeSckTbl[i].fecIdx = -1;
      return 0;
    }
  }
  return non_existent;
}
#endif
void removeBucketFromList(TCPBCKT *bckt)
{
  TCPBCKT *b;
  int i, cc = 0;
  SOCKET sck;
  if (bckt == NULL) return;
  sck = bckt->sck;
  if (WaitForSystemMutex(hTcpBucketMutex, -1) != 0) return;
  if (bckt == bcktList) bcktList = bckt->nxt;
  else
  {
    for (b = bcktList; b != NULL && b->nxt != bckt; b = b->nxt);
    if (b == NULL || b->nxt == NULL) ccerr(code_failure);
    b->nxt = bckt->nxt;
  }
err:
  ReleaseSystemMutex(hTcpBucketMutex);
  if (cc != 0) return;
  for (i=0; i < nTCPsck; i++)
  { /* socket on the server side ? */
    if (sck == tcpSckTbl[i].sck)
    { /* found it ! */
      msglog(0, "closing connection to daughter socket %d", sck);
      if (sck != 0) soclose(sck);
      tcpSckTbl[i].type = 0;
      tcpSckTbl[i].sck = 0;
      if (nTCPsck > 0) nTCPsck--;
      break;
    }
  }
  if (nTCPsck < 0) nTCPsck = 0; /* sanity check */
  for (; i < nTCPsck; i++) tcpSckTbl[i] = tcpSckTbl[i + 1];
  return;
}
void CloseConnectionSocket(SOCKET sck,int inetProtocol,int yield)
{
  int i, k;
  IDLE_CONNECTION *ic;
  ContractListStruct *cl;
  ClnLst *clnLst;
  TCPSCKT *sckt;

  freeBucket(sck,yield);

  if (inetProtocol == DBGPIPE)
  {
    for (i=0; i<nDBGsck; i++)              
    { /* socket on the server side ? */
      if (sck == dbgSckTbl[i].sck)
      {
        char ipNum[40] = "";
        if (sck != 0) 
        {
          SCKADR sckadr;
          _socklen_t len = getInAddrLen(inetProtocol);
          if (getpeername(sck,(struct sockaddr *)&sckadr,&len) == 0) inetStrAdr(&sckadr, ipNum, FALSE);
          soclose(sck);
        }
        else
        {
          sprintf(ipNum,"invalid socket");
        }
#       if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
        srvlog("commands.log", "ATTACH_FEC", "remote access closed from: %s", ipNum);
#       endif
        dbgSckTbl[i].type = 0;
        dbgSckTbl[i].sck = 0;
        nDBGsck--;
        break;
      }
    }
    for (; i<nDBGsck; i++) dbgSckTbl[i] = dbgSckTbl[i+1];
  }
  for (i=0; i<nTCPsck; i++)              
  { /* socket on the server side ? */
    if (sck == tcpSckTbl[i].sck)
    { /* found it ! */
      msglog(0, "closing connection to daughter socket %d", sck);
      if (sck != 0) soclose(sck);
      tcpSckTbl[i].type = 0;
      tcpSckTbl[i].sck = 0;
      if (nTCPsck > 0) nTCPsck--;
      break;
    }
  }
  if (nTCPsck < 0) nTCPsck = 0; /* sanity check */
  for (; i<nTCPsck; i++) tcpSckTbl[i] = tcpSckTbl[i+1];
  for (i=0; i<nconsumers; i++) if (ClnTbl[i]->ep.sck == sck) 
  { /* socket on the server side */
    ClnTbl[i]->ep.sck = 0;
    for (k=0; k<ncontracts; k++)
    {
      if ((cl=ContractList[k]) == NULL) continue;
      if ((clnLst=cl->clnLst) == NULL) continue;
      if (clnLst->client == ClnTbl[i])
      {
        cl->deliveryPending = 0;
        cl->clnLst->status.counter = 0;
      }
    }
  }
  if ((sckt=getTcpClnSck(sck, -1, -1)) != NULL)
  { /* socket on the client side */
    msglog(0, "closing connection to FEC %.16s on socket %d", FecTbl[sckt->idx].fecName, sck);
    closeIPConnectionToFec(sckt->idx, sck);
    while ((ic=GetIdleConnectionBySocket(sck, inetProtocol)) != NULL)
    { /* socket is now closed : it can't be idle ! */
      RemoveIdleConnection(ic);
      break;
    }
  }
  for (i=0; i < nConnectionTableEntries; i++)
  { /* socket on the client side ? => remove any trace of it */
    if (conTbl[i] == NULL) continue;
    if (conTbl[i]->tcpSck == sck) conTbl[i]->tcpSck = 0; 
  }
}
#define MAX_SELECT_ERRCNT 100
#define SEND_RETRY_COUNT 3
#define POLL_RETRY_COUNT 5
#define SEND_RETRY_TIMEOUT 20
int sendIpData(ClnHdr *cln,BYTE *data,UINT32 msgsize,int msgtype,int inetProtocol,int yield,int canFreeOnError)
{
  static int last_cc = 0, errcnt = 0;
  struct pollfd fds[1];
  SOCKET sck;
  TCPBCKT *bckt = NULL;
  int cc = 0, retry = 0, ns, msize;
  int to=0, nready, nsent=0, npkts=1;
  int needSndMutex = TRUE;
  char *errmsg = "sendIpData -> select", *strprot = getProtocolAsString(cln->inetProtocol);
  char dbgerr[256] = "";

  if (!ipLoaded) return tcp_not_supported;
  if (inetProtocol == UDP && msgsize > (UINT32)gPacketMtu) return use_stream_transport;
# ifndef VXWORKS
  if (msgtype == SRV_RSP_) needSndMutex = FALSE;
# endif
  if ((sck=cln->ep.sck) <= 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"sendIPData: trying to send on unassigned socket!");
    return tcp_socket_error;
  }
  msize = (int)msgsize;
  to = gSystemTick > 1 ? (int)gSystemTick : 1;
  if (needSndMutex && (cc=WaitForSystemMutex(hSndMutex,100)) != 0) 
  {
    if (last_cc == semaphore_busy) errcnt++; else errcnt = 0;
    last_cc = semaphore_busy;
    if (errcnt == 0) feclogEx(TINE_LOGLEVEL_WARN, "sendIpData: could not get mutex <%d> (waited 100 msec)!", cc);
    msglog(last_cc,"sendIpData: could not get mutex : count = %d",errcnt);
    return semaphore_busy;
  }
  _sendToDbg = -1;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (isUnixSocket(sck)) goto SendIPretry;
# endif
SelectIPretry:
  dbgerr[0] = 0;
  fds[0].fd = sck; fds[0].events = POLLOUT; fds[0].revents = 0;
  if ((nready=poll(fds,1,to)) != 1 || pollerr(0))
  {
    errmsg = "select";
    if (pollhup(0)) ccerr(net_write_error);
    if (nready < 0)
    {
      rprtSckErr("sendIpData", errmsg, sck, -1, 4, dbgerr);
      selectErrcnt++;
    }
    if (retry++ > POLL_RETRY_COUNT)
    {
      msglog(cannot_select,"%s socket %d not available for writing! payload: %d, send buffers: %d",strprot,sck,msgsize,max_srvsndbuf);
      cc = (nready < 0 && soerrno != EINTR) ? net_write_error : cannot_select;
      selectErrcnt++;
      selectFailcnt++;
      ccerr(cc);
    }
    to = SEND_RETRY_TIMEOUT;
    goto SelectIPretry;
  }
  selectErrcnt = 0; /* reset if no select error */
  retry = 0; /* reset this, as most retries occur in the poll loop above */
  while (msize > 0)
  {
SendIPretry:
    if ((ns=(int)writeto(sck,(char *)&data[nsent],msize,MSG_NOSIGNAL,
                      (struct sockaddr *)&cln->adr.ip,getInAddrLen(cln->adr.ip.ss_family))) <= 0)
    {
      if (soerrno == EPIPE) retry = SEND_RETRY_COUNT;
      errmsg = "sendto";
      rprtSckErr("sendIpData",errmsg,sck,-1,4,dbgerr);
      sendtoErrcnt++;
      if (retry++ > SEND_RETRY_COUNT) ccerr(net_write_error);
      goto SendIPretry;
    }
    nsent = ns;
    msize -= nsent;
    if (msize > 0) npkts++;
  }
err:
  if (tineDebug > 2 && cc == 0)
  {
    char addr[64];
    inetStrAdr(&cln->adr.ip,addr,FALSE);
    sprintf(dbgerr,"%s: %d bytes (%d packets) to %.16s [%.40s]",strprot,msgsize,npkts,cln->usr,addr);
  }
  if (cc != 0)
  {
    if (cc == cannot_select)
    {
#     ifdef WIN32
      feclog("in sendIpData %.64s (socket %Id) : unable to get write resources", errmsg, sck);
#     else
      feclog("in sendIpData %.64s (socket %d) : unable to get write resources", errmsg, sck);
#     endif
    }
    else
    {
      rprtSckErr("sendIpData", errmsg, sck, -1, 1, dbgerr);
      if (errcnt == 0) feclog("sendIpData: %.32s", cc2str(cc));
    }
    if (cln->inetProtocol & (TCP|TCP6|STREAM|STREAM6)) 
    { /* a peer-to-peer socket connection */
      int mutexReleaseNeeded = TRUE;
      if ((bckt=findBucket(sck,TRUE)) == NULL)
      { /* this would be very unusual ... */
        msglog(cc,"no bucket found for TCP socket %d !",sck);
        cc = tcp_socket_error;
        mutexReleaseNeeded = FALSE;
        canFreeOnError = TRUE;
      }
      else
      { /* augment this sockets error count and check ... */
        bckt->errcnt += selectErrcnt;
        if (bckt->errcnt > MAX_SELECT_ERRCNT) cc = tcp_socket_error;
      }
      if (cc != cannot_select) 
      { /* not (or no longer seen as) a select error */
        selectErrcnt = 0;
        msglog(cc,"close connection on socket %d due to error",sck);
        if (canFreeOnError)
        {
          if (mutexReleaseNeeded) ReleaseSystemMutex(hTcpBucketMutex), mutexReleaseNeeded = FALSE;
          CloseConnectionSocket(sck,cln->inetProtocol,yield);
          bckt = NULL; /* CloseConnectionSocket frees the bucket memory ! */
        }
        else
        {
          if (bckt != NULL) bckt->rmvNeeded = TRUE;
        }
      }
      if (mutexReleaseNeeded) ReleaseSystemMutex(hTcpBucketMutex);
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    if (cln->inetProtocol == PIPE) 
    {
      ClosePipeSocket(sck);
      gUseLoopback = TRUE;
      msglog(cc,"cannot send data on pipe: set loopback flag to TRUE");
    }
#   endif
  }
  if (last_cc == cc && cc != 0) 
  { /* same error as last time */
    errcnt++;
  }
  else 
  {
    if (errcnt > 0) 
      msglog(cc,"last error %.32s changed to %.32s after %d successive failures",cc2str(last_cc),cc2str(cc),errcnt);
    errcnt = 0;
  }
  last_cc = cc;
  if (selectErrcnt == 0 && bckt != NULL) bckt->errcnt = 0;
  if (needSndMutex) ReleaseSystemMutex(hSndMutex);
  _sendToDbg = 0;
  if (strlen(dbgerr)) dbglog(dbgerr);
  return cc;
}
int gMaxTcpMsgSize = DEFAULT_MAX_TCPMSG_SIZE;
TINE_EXPORT void SetMaxTcpMessageSize(int value)
{
  if (value > 1472) gMaxTcpMsgSize = value;
  feclog("maximum tcp message size set to %d",gMaxTcpMsgSize);
  return;
}
TINE_EXPORT int GetMaxTcpMessageSize(void)
{
  return gMaxTcpMsgSize;
}
int sendStreamData(ClnHdr *cln,BYTE *data,int msgsize)
{
# ifndef MULTITHREADED
  UNUSED(cln);
  UNUSED(data);
  UNUSED(msgsize);
  return not_implemented;
# else
  struct pollfd fds[1];
  int to=0, nready=0, nsent=0;
  SOCKET sck;
  int n, cc = 0, retry = 0, sizeOK = FALSE;
# ifdef WIN32
  int msize = 0, sndsize, maxsize;
# else
  size_t msize = 0, sndsize, maxsize;
# endif
  char *strprot = getProtocolAsString(cln->inetProtocol);
  int terminate = FALSE;
  
  if (!ipLoaded) return tcp_not_supported;
  if ((sck=cln->ep.sck) <= 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "sendStreamData: trying to send on unassigned socket!");
    return tcp_socket_error;
  }
  to = gSystemTick * (1+msgsize/8192);
  if (to > 500) to = 500; else if (to < 0) to = 1;
SelectIPretry:
  fds[0].fd = sck; fds[0].events = POLLOUT;
  if ((nready=poll(fds,1,to)) != 1 || pollerr(0))
  {
    if (pollhup(0)) ccerr(net_write_error);
    rprtSckErr("sendStreamData","select",sck,-1,1,NULL);
    selectErrcnt++;
    selectFailcnt++;
    if (retry++ > 3) ccerr(cannot_select);
    goto SelectIPretry;
  }
  if (tineDebug > 2)
  {
    char addr[64];
    inetStrAdr(&cln->adr.ip,addr,FALSE);
    dbglog("%s: %d bytes to %.16s [%.30s]",strprot,msgsize,cln->usr,addr);
  }
  msize = msgsize;
  maxsize = gMaxTcpMsgSize;
  while (msize > 0)
  {
    sndsize = msize < maxsize ? msize : maxsize;
    if (WaitForSystemMutex(hSndMutex,-1) != 0) return semaphore_busy;
    if ((n=(int)send(sck,(char *)&data[nsent],sndsize,0)) <= 0)
    { /* couldn't send anything ... */
      ReleaseSystemMutex(hSndMutex);
      if (soerrno == ENOBUFS)
      { /* server or client does not have the buffer space ?! */
        if (!sizeOK) maxsize /= 2;
        if (maxsize < TCP_BUFFER_SIZE) ccerr(net_write_error);
        millisleep(gSystemTick);
      }
      else
      { /* something serious has happened */
        terminate = TRUE;
        sendtoErrcnt++;
        ccerr(net_write_error);
      }
    }
    else
    { /* the max tcp message size is known to work for this peer */
      ReleaseSystemMutex(hSndMutex);
      sizeOK = TRUE;
      nsent += n;
      msize -= n;
    }
    if (msize > 0)
    {
      if (tineDebug) dbglog("%s : sent %d of %d bytes",strprot,nsent,msgsize);
    }
  }
err:
  if (cc != 0)
  {
    int p = (gAfInet == AF_INET6) ? STREAM6 : STREAM;
    rprtSckErr("sendStreamData","send",sck,-1,0,NULL);
    if (terminate) CloseConnectionSocket(sck,p,TRUE);
  }
  return cc;
# endif
}
INADR tMCastAddr;
INADR tMCastAddrLegacy;
INADR tGCastAddrLegacy;
INADR tGCastTgtAddr;
INADR tNetSrvAddr;
SCKADR tGCastSckAddr;
INADR tEthrAddr; /* designated interface address (if multiple cards) */

int detachPortFromMulticastGroup(SOCKET sck,UINT16 port,INADR grpAddr,char *grpType)
{
# ifndef NOT_MULTICASTING
  int i,n=0,cc=0;
  UNUSED(port);
  if (!zerocmp(&grpAddr,gInAdrLen)) 
  { /* not a valid group */
    return invalid_mcast_group;
  }
  if (sck == mcastClnSck)
  {
    for (i=0; i<nConnectionTableEntries; i++)
    { /* count the links in the link table with this group */
      if (conTbl[i] == NULL) continue;
      if (!adrcmp(&conTbl[i]->mcastGrp,&grpAddr)) n++;
    }
  }
  else if (sck == glbClnSck)
  {
    for (i=0; i<nglobals; i++) 
    { /* count the globals connected to this group */
      if (!adrcmp(&glbTbl[i]->mcastGrp,&grpAddr)) n++;
    }
  }
  if (n > 1)
  { /* there are other links attached to this group -> don't drop out ! */
    return 0;
  }
  if (gAfInet == AF_INET)
  { /* ipv4 */
    struct ip_mreq mreq;
    struct in_addr *in4 = (struct in_addr *)&grpAddr;
    struct in_addr *et4 = (struct in_addr *)&tEthrAddr;
    mreq.imr_multiaddr.s_addr = in4->s_addr;
    mreq.imr_interface.s_addr = strlen(gtEthrAddr) > 0 ? et4->s_addr : htonl(INADDR_ANY);
    cc = setsockopt(sck,IPPROTO_IP,IP_DROP_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
  }
  else if (gAfInet == AF_INET6)
  { /* ipv4 */
    struct ipv6_mreq mreq;
    struct in6_addr *in6 = (struct in6_addr *)&grpAddr;
    mreq.ipv6mr_multiaddr = *in6;
    mreq.ipv6mr_interface = 0;
    cc = setsockopt(sck,IPPROTO_IPV6,IPV6_DROP_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
  }
  feclog("detaching from %s multicast group : %s",grpType,cc ? "failed" : "success");
  return cc ? mcast_init_error : 0;
# else
  sck = sck;
  port = port;
  grpAddr = grpAddr;
  grpType = grpType;
  return not_implemented;
# endif
}
int attachPortToMulticastGroup(SOCKET sck,UINT16 port,INADR grpAddr,char *grpType)
{
# ifndef NOT_MULTICASTING
  int i,n=0,cc=0;
  UNUSED(port);
  if (!zerocmp(&grpAddr,gInAdrLen)) 
  { /* not a valid group */
    return invalid_mcast_group;
  }
  if (sck == mcastClnSck)
  {
    for (i=0; i<nConnectionTableEntries; i++)
    { /* count the links in the link table with this group */
      if (conTbl[i] == NULL) continue;
      if (!adrcmp(&conTbl[i]->mcastGrp,&grpAddr)) n++;
    }
  }
  else if (sck == glbClnSck)
  {
    for (i=0; i<nglobals; i++) 
    { /* count the globals connected to this group */
      if (glbTbl[i]->disabled) continue;
      if (!adrcmp(&glbTbl[i]->mcastGrp,&grpAddr)) n++;
    }
  }
  if (n > 0)
  { /* there are other links attached to this group -> no need to join ! */
    return 0;
  }
  if (grpType == NULL) grpType = "unknown";
  if (gAfInet == AF_INET)
  { /* ipv4 */
    struct ip_mreq mreq;
    struct in_addr *in4 = (struct in_addr *)&grpAddr;
    struct in_addr *et4 = (struct in_addr *)&tEthrAddr;
    mreq.imr_multiaddr.s_addr = in4->s_addr;
    mreq.imr_interface.s_addr = strlen(gtEthrAddr) > 0 ? et4->s_addr : htonl(INADDR_ANY);
    cc = setsockopt(sck,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
  }
  else if (gAfInet == AF_INET6)
  { /* ipv6 */
    struct ipv6_mreq mreq;
    struct in6_addr *in6 = (struct in6_addr *)&grpAddr;
    mreq.ipv6mr_multiaddr = *in6;
    mreq.ipv6mr_interface = 0;
    cc = setsockopt(sck,IPPROTO_IPV6,IPV6_ADD_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
  }
  if (cc != 0) rprtSckErr("attachPortToMulticastGroup","setsockopt",sck,-1,0,NULL);
  feclog("joining %s multicast group : %s",grpType,cc ? "failed" : "success");
  return cc ? mcast_init_error : 0;
# else
  sck = sck;
  port = port;
  grpAddr = grpAddr;
  grpType = grpType;
  return not_implemented;
# endif
}
#define MCAST_REATTACH_INTERVAL 300
int inhibitReattachToServiceGroup = TRUE;
void reAttachToServiceGroup(void)
{
  static time_t interim = 0;
  time_t t;
  if (inhibitReattachToServiceGroup) return;
  t = time(NULL);
  if (t < interim + MCAST_REATTACH_INTERVAL) return;
  attachPortToMulticastGroup(netSrvSck,(UINT16)gtNetSrvPort,tNetSrvAddr,"services");
  interim = t;
  return;
}
int gInitAddrPortsDone = FALSE;
int isMCastPort(UINT16 port)
{
  if (port == gtMCastPort) return TRUE;
  if (port == gtGCastPort) return TRUE;
  if (port == gtNetSrvPort) return TRUE;
  return FALSE;
}
int applyBcstAddr(int family,INADR *src,INADR *dst)
{
  *dst = *src;
  if (family == AF_INET)
  {
    struct in_addr *in4 = (struct in_addr *)dst;
#   if !defined(VXWORKS) && !defined(VMS) && !defined(NIOS) && !defined(DOS)
    in4->s_addr |= 0xff000000;
#   else
    in4->s_addr = 0xffffffff;
#   endif
    return 0;
  }
  return not_supported; /* broadcast not supported in ipv6 */
}
int applyMcstAddr(int family,INADR *tgt,INADR *pat)
{
  if (family == AF_INET)
  {
    ((BYTE *)tgt)[3] = ((BYTE *)pat)[3];
    ((BYTE *)tgt)[2] = ((BYTE *)pat)[2];
    if (!gMultiCastUseLegacyRules) ((BYTE *)tgt)[1] = ((BYTE *)pat)[1];
    return 0;
  }
  if (family == AF_INET6)
  { /* TODO: establish rules here! */
    BYTE b = ((BYTE *)tgt)[0];
    memcpy(tgt,pat,16);
    ((BYTE *)tgt)[0] = b;
    return 0;
  }
  return not_supported; 
}
static int gInitIpFromEnvDone = FALSE;
void initIpFromEnv(void)
{
  char *ptr;
  if (gInitIpFromEnvDone) return;
  gInitIpFromEnvDone = TRUE;
  if ((ptr=getenv("TINE_STANDALONE")) != NULL)
  {
    gSystemRunningStandAlone = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
  }
  if ((ptr=getenv("TINE_LEGACY_MULTICAST")) != NULL)
  {
    if (strtobool(ptr)) SetUseMCastLegacyRules(TRUE);
  }
  if ((ptr=getenv("TINE_IPV6")) != NULL)
  {
    if (strtobool(ptr)) SetAfInet(AF_INET6);
  }
  if ((ptr=getenv("TINE_TCP_IDLESECONDS")) != NULL)
  {
    SetIdleConnectionDuration(atoi(ptr));
  }
# ifndef SMALL_TINE_LIB
  initSockBufferSizeFromEnv();
# endif
# if !defined(VXWORKS) && !defined(VMS) && !defined(NIOS) && !defined(DOS) && !defined(SMALL_TINE_LIB)
  if (!gInitAddrPortsDone)
  { /* allow these to be fixed via general environment or csv files */
    initAddrPortsFromCsv();
    initAddrPortsFromEnv();
    gInitAddrPortsDone = TRUE;
  }
# endif
}
/**
 * \internal
 * \brief Binds an IP socket for UDP Communication and initializes some globals.
 *
 * Only binds a socket. 
 * If port equals one of the reserved port numbers gtMCastPort,
 * gtGCastPort, or gtNetSrvPort, the socket is added to a multicast group.
 * On UNIX and Win32 Systems, the send and receive buffers are set according
 * to the global variables max_sndbuf and max_rcvbuf.
 * 
 * Initializes the following globals when called for the first time:
 * <ul>
 * <li>tMCastAddr</li>
 * <li>tGCastAddr</li>
 * <li>tNetSrvAddr</li>
 * <li>tGCastSckAddr</li>
 * </ul>
 *
 * \param s Socket number (output)
 * \param port Port number. If set to 0, take default TINE client port.
 *
 */
int setupUdpSocket(SOCKET *s,int family,unsigned short port,int initMultiCast,int rcvbuf,int sndbuf)
{
  SCKADR sckadr;
  int mode = 1,optval,cc = 0;
  char txt[50];
  unsigned short udpport;
  static UINT16 nextFreePort = 0;
  static int lInitDone = FALSE;
  INADR ipadr, myadr;
  _socklen_t optlen;
# if defined(DOS) || defined(VXWORKS)
  int reuseFlg = TRUE;
# else
  int reuseFlg = FALSE;
# endif

  initIpFromEnv();
  if (!lInitDone && initMultiCast) 
  { /* always fill this in at initialization time for use by server code */
    lInitDone = TRUE;
    memcpy(&myadr,getmyipaddr(),gInAdrLen);
    feclog("determined local host ip address to be : %.40s",getmyipaddrstr());
#   ifdef NOT_MULTICASTING
    /* service port address */
    applyBcstAddr(family,&myadr,&tNetSrvAddr);
#   else
    if (family == AF_INET)
    { /* legacy canonical producer-subscriber address (in case it's needed) */
      inet_pton(family,MCAST_ADDR_LEGACY,&ipadr);
      memcpy(&tMCastAddrLegacy,&ipadr,sizeof(INADR));
    }
    /* specific producer-subscriber address */
    ipadr = fixMcastHaddr((BYTE *)&myadr,gMultiCastUseLegacyRules);
    if (gtMCastAddr != gtMCastAddrBuffer || gMultiCastUseLegacyRules)
    { /* has not been pre-assigned: reassign the proper IP string representation */
      inet_ntop(family,&ipadr,gtMCastAddrBuffer,40);
      gtMCastAddr = gtMCastAddrBuffer;
    }
    memcpy(&tMCastAddr,&ipadr,sizeof(INADR));
    if (family == AF_INET)
    { /* legacy canonical producer-consumer address (in case it's needed) */
      inet_pton(family,GCAST_ADDR_LEGACY,&ipadr);
      memcpy(&tGCastAddrLegacy,&ipadr,sizeof(INADR));  
    }
    /* prepare return socket for gcasts from a server (same mcast addr, differernt port) */
    sckadr_in(&tGCastSckAddr,(BYTE *)&tMCastAddr,(UINT16)gtGCastPort,family);
    /* service port address */
    inet_pton(family,gtNetSrvAddr,&ipadr);
    memcpy(&tNetSrvAddr,&ipadr,sizeof(INADR));
    if (strlen(gtEthrAddr) > 0)
    {
      inet_pton(family,gtEthrAddr,&ipadr);
      memcpy(&tEthrAddr,&ipadr,sizeof(INADR));  /* service port */
    }
#   endif
  }
  if (s == NULL) ccerr(argument_list_error);
  if (nextFreePort == 0) nextFreePort = gtUDPClnPort;
  sprintf(txt,"UDP PORT %d: ",port);
  udpport = port ? port : nextFreePort;

bindport:
  *s = (SOCKET)socket(family,SOCK_DGRAM,0);
  if (isSckInvalid(*s)) soerr("socket");
# if defined(UNIX)
  optval = fcntl(*s,F_GETFD,0); optval |= FD_CLOEXEC; fcntl(*s, F_SETFD, optval);
# endif
# if defined(FIX_IPV6_V6ONLY)
  if (family == AF_INET6)
  {
    optval = 0; optlen = sizeof(optval);
    if (setsockopt(*s,IPPROTO_IPV6,IPV6_V6ONLY,(char*)&optval,optlen)) soerr("IPV6_V6ONLY");
  }
# endif
  if (isMCastPort(port)) reuseFlg = TRUE;
  if (reuseFlg)
  {
    optval = 1; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) soerr("REUSEADDR");
#   ifdef MACOS
    optval = 1; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_REUSEPORT,(char *)&optval,optlen)) soerr("REUSEPORT");
#   endif
  }
# if defined(UNIX) || defined(WIN32)
  if (rcvbuf > 0)
  {
    optval = rcvbuf; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_RCVBUF,(char *)&optval,optlen)) rprtSckErr("setupUdpSocket","RCVBUF",*s,-1,0,NULL);
    checkSckBufferSpace(*s,rcvbuf,SO_RCVBUF,"UDP");
  }
  if (sndbuf)
  {
    optval = sndbuf; optlen = sizeof(optval);
    if (setsockopt(*s,SOL_SOCKET,SO_SNDBUF,(char *)&optval,optlen)) rprtSckErr("setupUdpSocket","SNDBUF",*s,-1,0,NULL);
    checkSckBufferSpace(*s,sndbuf,SO_SNDBUF,"UDP");
  }
# endif
  if (udpport == 0) goto portbound;
  sckadr_in(&sckadr,NULL,udpport,family);
  if (bind(*s,(struct sockaddr *)&sckadr,getInAddrLen(family)))
  { /* n.b. VISTA requires closing and opening a new socket */
    soclose(*s);
    if (port == 0 && udpport++ < gtUDPClnPort + UDP_CLN_RANGE) goto bindport;
    soerr("cannot bind");
  }
  if (port == 0) nextFreePort = udpport + 1;
portbound:
  /* attach to any relevant multicast groups (if possible) */
  if (port == gtNetSrvPort) attachPortToMulticastGroup(*s,port,tNetSrvAddr,"services");
  /* blocking or non-blocking ? */
  if (gIsCyclePollingMode && soioctl(*s,FIONBIO,(char *)&mode)) soerr("NONBLOCKING");
  feclog("%s bound to %d (socket %d %sblocking)",
          txt,inetPort(&sckadr),*s,(gIsCyclePollingMode ? " non-" : " "));
err: 
  if (cc != 0)
  {
    char scktxt[32];
    SOCKET sck = s != NULL ? *s : 0;
    rprtSckErr("setupUdpSocket",txt,sck,-1,0,NULL);
    if (s != NULL)
    {
      if (*s > 0) 
      {
#       ifdef NEED_POLL
        if (*s >= FD_SETSIZE) strcpy(txt,"out of range");
#       endif
        soclose(*s);
      }
      *s = 0;
    }
    else
    { /* not really possible */
      strcpy(scktxt,"(null entry)");
    }
  }
  return cc;
}
#if defined(UNIX) && !defined(FREEBSD) && !defined(SOLARIS)
int getThisFecMacAddress(BYTE *addr)
{
  struct ifreq ifbuf;
  SOCKET s = socket(AF_INET, SOCK_DGRAM, 0);
  int cc = 0;
  if (s == (SOCKET)0) return code_failure;
  bzero(&ifbuf,sizeof(ifbuf));
  strcpy(ifbuf.ifr_name,"eth0");
  if (ioctl(s,SIOCGIFHWADDR,&ifbuf)) ccerr(udp_socket_error);
  memcpy(addr,ifbuf.ifr_hwaddr.sa_data,6);
err:
  close(s);
  return cc;
}
#endif
#if !defined(VMS) && !defined(NIOS) && !defined(DOS) && !defined(SMALL_TINE_LIB)
static char gtIpAddr[40]; 
CSVCOLUMN csvSrvIPColumns[] =
{
  { "ADDRESS","",0,CF_TEXT,ADDR_SIZE_IPV6,(BYTE *)gtIpAddr,NULL,0 },
};
CSVDB csvSrvIPDb =
{
  csvSrvIPColumns,
  sizeof(csvSrvIPColumns)/sizeof(csvSrvIPColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  0,
  0
};
CSVCOLUMN csvIPMaskColumns[] =
{
  { "ADDRESS","",0,CF_TEXT,ADDR_SIZE_IPV6,(BYTE *)OFFSETIN(MCastAddrMap,mcStrAdr),NULL,0 },
  { "PATTERN","",0,CF_TEXT,ADDR_SIZE_IPV6,(BYTE *)OFFSETIN(MCastAddrMap,patStrAdr),NULL,0 },
};
CSVDB csvIPMaskDb =
{
  csvIPMaskColumns,
  sizeof(csvIPMaskColumns)/sizeof(csvIPMaskColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(MCastAddrMap),
  0
};
void initAddrPortsFromCsv(void)
{
  fixTineHome();
  memset(gtIpAddr,0,40);
  if (csvReadFile(fecDbPath,MCAST_ADDRESS_FILE,&csvSrvIPDb,NULL) == 0) SetMCastAddr(gtIpAddr);
  memset(gtIpAddr,0,40);
  if (csvReadFile(fecDbPath,GCAST_ADDRESS_FILE,&csvSrvIPDb,NULL) == 0) SetGCastAddr(gtIpAddr);
  memset(gtIpAddr,0,40);
  if (csvReadFile(fecDbPath,MCAST_MASK_FILE,&csvIPMaskDb,(void **)&gtMCastAddrLst)    == 0 ||
      csvReadFile(tineHomePath,MCAST_MASK_FILE,&csvIPMaskDb,(void **)&gtMCastAddrLst) == 0 ||
      csvReadFile(fecDbPath,GCAST_MASK_FILE,&csvSrvIPDb,(void **)&gtMCastAddrLst)     == 0 ||
      csvReadFile(tineHomePath,GCAST_MASK_FILE,&csvSrvIPDb,(void **)&gtMCastAddrLst)  == 0) 
  {
    int i;
    char *pattern;
    nMCastAddrLstEntries = csvIPMaskDb.siz;
    for (i=0; i<nMCastAddrLstEntries; i++)
    { 
      if (isInvalidIpV4Addr(gtMCastAddrLst[i].mcStrAdr) && isInvalidIpV6Addr(gtMCastAddrLst[i].mcStrAdr))
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "assigned multicast address %.40s is invalid!",gtMCastAddrLst[i].mcStrAdr);
        continue;
      }
      inet_pton(gAfInet,gtMCastAddrLst[i].mcStrAdr,&gtMCastAddrLst[i].mcInAdr);
      pattern = gtMCastAddrLst[i].patStrAdr;
      if (strlen(pattern) == 0) pattern = getmyipaddrstr();
      if (isInvalidIpV4Addr(pattern) && isInvalidIpV6Addr(pattern))
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "assign multicast address pattern %.40s is invalid!",pattern);
        gtMCastAddrLst[i].mcInAdr = inadr_zero;
        continue;
      }
      inet_pton(gAfInet,pattern,&gtMCastAddrLst[i].patInAdr);
    }
  }
}
# if !defined(VXWORKS)
int getSrvIPAddr(void)
{
  char *ptr;
  int mtu;
  static int cc = 0;
  static int done = FALSE;

  if (done) return cc;
  if ((ptr=getenv("FEC_GCAST_ADDR")) != NULL)
  {
    SetGCastAddr(ptr);
    feclog("set fec globals multicast address from environment : %s",ptr);
    done = TRUE;
  }
  if ((ptr=getenv("FEC_PACKETMTU")) != NULL)
  {
    mtu = atoi(ptr);
    if (mtu > 512) 
    {
      feclog("set packet mtu from environment : %s",ptr);
      SetPacketMTU(mtu);
    }
    done = TRUE;
  }
  gtEthrAddr[0] = 0; gtIpAddr[0] = 0;
  if ((cc=csvReadFile(fecDbPath,MY_ADDRESS_FILE,&csvSrvIPDb,NULL)) == 0) 
  {
    strncpy(gtEthrAddr,gtIpAddr,32);
    feclog("set fec address from file : %s",gtEthrAddr);
    done = TRUE;
    return cc;
  }
  if ((ptr=getenv("FEC_ADDRESS")) != NULL)
  {
    strncpy(gtEthrAddr,ptr,16);
    feclog("set fec address from environment : %s",gtEthrAddr);
    done = TRUE;
    return cc = 0;
  }
  cc = un_allocated;
  return cc;
}
BYTE *getmyipaddr(void)
{
  static char hname[30];
  static SCKADR stlclAddr;
  static SCKADR strmtAddr;
  static int done = 0;
  BYTE *myadr, ipadr[16];
  int sck;
  if (done) return getInAddr(&stlclAddr);
# if defined(WIN32)
  /* do this here, just in case this is called prior to SystemInit() ! */
  if (!WinsockLoaded()) 
  {
    memset(&stlclAddr,0,sizeof(SCKADR));
    return getInAddr(&stlclAddr);
  }
# endif
  stlclAddr.ss_family = (short)gAfInet;
  if (gSystemRunningStandAlone) 
  { /* fix my ip to be the loopback address */
    done = TRUE;
    setInAddrLoopback(&stlclAddr);
    return getInAddr(&stlclAddr);
  }
  getSrvIPAddr();
  if (strlen(gtEthrAddr) > 0)
  { /* explicit my address info given in myaddr.csv */
    done = TRUE;
    gAfInet = getInetFamilyFromAddressString(gtEthrAddr);
    myadr = getInAddr(&stlclAddr);
    inet_pton(gAfInet,gtEthrAddr,myadr);
    return myadr;
  }
  setInAddrAny(&stlclAddr);
  if (gethostname(hname,30) != -1)
  {
    struct addrinfo *ai, hints;
    memset(&hints,0,sizeof(hints));
    hints.ai_family = gAfInet;
    if (getaddrinfo(hname,NULL,&hints,&ai) == 0)
    { /* try the simple way first */
      int len = gAfInet == AF_INET6 ? sizeof(SCKADR) : sizeof(struct sockaddr_in);
      memcpy(&stlclAddr,ai->ai_addr,len);
      freeaddrinfo(ai);
    }
  }
  if (!isInAddrAny(&stlclAddr)) goto addr_done;
  if (gAfInet == AF_INET6)
  {
    FecAddrStruct fec;
    memset(&fec,0,sizeof(fec));
    if (GetThisFecAddresses(&fec) == 0)
    {
      inet_pton(AF_INET6,fec.strAdr,ipadr);
      sckadr_in(&stlclAddr,ipadr,0,AF_INET6);
      goto addr_done;
    }
  }
  if (isInAddrAny(&stlclAddr) || /* simple didn't work */
      isInAddrLoopback(&stlclAddr))   /* or came back with the loopback */
  {
    if ((sck=(int)socket(gAfInet, SOCK_DGRAM, 0)) > 0)
    {
      _socklen_t nRet=(_socklen_t)sizeof(SCKADR);
      if (gAfInet == AF_INET) inet_pton(AF_INET,"128.127.50.1",ipadr);
      ip4adr(&strmtAddr,ipadr,IPPORT_ECHO);
      if (connect(sck,(struct sockaddr *)&strmtAddr,nRet) != -1)
      {
        getsockname(sck,(struct sockaddr *)&stlclAddr,&nRet);
      }
      soclose(sck);
    }
  }
addr_done:
  done = TRUE;
  return getInAddr(&stlclAddr);
}
static int isLoopback(BYTE *adr,int fam)
{
  switch (fam)
  {
    case AF_INET:
      if (adr[0] != 0x7f) break;
      if (adr[1] != 0 || adr[2] != 0) break;
      if (adr[3] != 0x01) break;
      return TRUE;
    case AF_INET6:
    {
      int i;
      for (i=0; i<15; i++) if (adr[i] != 0) return FALSE;
      if (adr[15] != 0x01) break;
      return TRUE;
    }  
    default: break;
  }
  return FALSE;
}
int GetThisFecAddresses(FecAddrStruct *fec)
{
  static int cc = not_supported;
# ifdef USE_IPV6
  static int done = FALSE;
  static char mac[8];
  static char ipv6[40];
  static char ipv4[16];
  static int fam, ip6active = FALSE, ip4active = FALSE, addressesFound = FALSE;
  char addr[64];
  BYTE *inaddr;
# if defined(WIN32)
  UINT32 rc, size;
  PIP_ADAPTER_ADDRESSES aaLst, aa;
  PIP_ADAPTER_UNICAST_ADDRESS ua;
# elif defined(UNIX)
  struct ifaddrs *ifLst = NULL, *ifa;
# endif
  if (done)
  {
    memcpy(fec->lclAdr.mac,mac,6);
    if (ip4active)
    {
      BYTE _inadr[4];
      inet_pton(AF_INET,ipv4,_inadr);
      memcpy(fec->netAdr.ipv4,_inadr,4);
    }
    if (ip6active)
    {
      strncpy(fec->strAdr,ipv6,40);
    }
    else if (addressesFound)
    {
      strncpy(fec->strAdr,ipv4,16);
    }
    return cc;
  }
# if defined(WIN32)
  if (GetAdaptersAddresses(AF_UNSPEC,GAA_FLAG_INCLUDE_PREFIX,NULL,NULL,&size) != ERROR_BUFFER_OVERFLOW)
    return code_failure;
  aaLst = (PIP_ADAPTER_ADDRESSES)calloc(1,size);
  rc = GetAdaptersAddresses(AF_UNSPEC,GAA_FLAG_INCLUDE_PREFIX,NULL,aaLst,&size);
  for (aa=aaLst; aa != NULL; aa=aa->Next)
  {
    if (zerocmp(aa->PhysicalAddress,6) == 0) continue;
    memcpy(mac,aa->PhysicalAddress,6);
    memcpy(fec->lclAdr.mac,aa->PhysicalAddress,6);
    for (ua = aa->FirstUnicastAddress; ua != NULL; ua = ua->Next)
    {
      fam = ua->Address.lpSockaddr->sa_family;
      if (fam != AF_INET && fam != AF_INET6) continue;
      addressesFound = TRUE;
      inaddr = getInAddr((SCKADR *)ua->Address.lpSockaddr);
      inet_ntop(fam,inaddr,addr,40);
      if (fam == AF_INET) 
      {
        memcpy(fec->netAdr.ipv4,inaddr,4);
        strncpy(ipv4,addr,16);
        ip4active = TRUE;
      }
      if (fam == AF_INET6) 
      {
        strncpy(fec->strAdr,addr,40);
        strncpy(ipv6,addr,40);
        ip6active = TRUE;
      }
    }
    if (addressesFound) 
    {
      char ifadr[16], adrs[256];
      int slen = 0;
      bzero(adrs,256);
      BytesToHex(fec->lclAdr.mac,ifadr,6);
      if (ip4active) sprintf(adrs,"%.16s",ipv4);
      slen = (int)strlen(adrs);
      if (ip6active)
      {
        if (slen > 0) strcat(adrs," and ");
        strncat(adrs,fec->strAdr,40);
      }
      else
      {
        strncpy(fec->strAdr,addr,40);
      }
      feclog("interface %16s bound to %.64s",ifadr,adrs);
      cc = 0;
      break;
    }
  }
  free(aaLst);
  done = TRUE;
# elif defined(UNIX) && !defined(SOLARIS)
  if (getifaddrs(&ifLst) != 0) return code_failure;
  for (ifa=ifLst; ifa != NULL; ifa=ifa->ifa_next)
  {
    if (ifa->ifa_addr == NULL) continue;
    if (!(ifa->ifa_flags & IFF_UP)) continue;
    if (!strncmp(ifa->ifa_name,"lo",2)) continue;
    fam = ifa->ifa_addr->sa_family;
#   ifdef FREEBSD
    if (fam == AF_LINK)
    {
      BYTE *mp = LLADDR((struct sockaddr_dl *)ifa->ifa_addr);
      if (strncmp(ifa->ifa_name,"en0",3)) continue;
      memcpy(mac,mp,6);
      memcpy(fec->lclAdr.mac,mac,6);
    }
#   endif
    if (fam != AF_INET && fam != AF_INET6) continue;
    inaddr = getInAddr((SCKADR *)ifa->ifa_addr);
    inet_ntop(fam,inaddr,addr,40);
    if (fam == AF_INET) 
    {
      if (ip4active) continue; /* already got the primary address */
      if (isLoopback(inaddr,AF_INET)) continue;
      strncpy(ipv4,addr,16);
      memcpy(fec->netAdr.ipv4,inaddr,4);
      ip4active = TRUE;
    }
    if (fam == AF_INET6) 
    {
      if (ip6active) continue; /* already got the primary address */
      if (isLoopback(inaddr,AF_INET6)) continue;
      strncpy(fec->strAdr,addr,40);
      strncpy(ipv6,addr,40);
      ip6active = TRUE;
    }
    addressesFound = TRUE;
  }
  if (addressesFound) 
  {
    char ifadr[16], adrs[256];
    int slen = 0;
    bzero(adrs,256);
#   ifndef FREEBSD
    getThisFecMacAddress(fec->lclAdr.mac);
    memcpy(mac,fec->lclAdr.mac,6);
#   endif
    BytesToHex(fec->lclAdr.mac,ifadr,6);
    if (ip4active) sprintf(adrs,"%.16s",ipv4);
    slen = (int)strlen(adrs);
    if (ip6active)
    {
      if (slen > 0) strcat(adrs," and ");
      strncat(adrs,fec->strAdr,40);
    }
    else
    {
      strncpy(fec->strAdr,addr,40);
    }
    feclog("interface %16s bound to %.64s",ifadr,adrs);
    cc = 0;
  }
  freeifaddrs(ifLst);
  done = TRUE;
# endif /* WIN32 or UNIX */
# endif /* USE_IPV6 */
  return cc;
}
# else /* VXWORKS is #defined */
int getSrvIPAddr(void) { return not_implemented; }
# if (_WRS_VXWORKS_MAJOR < 6)
char *gtEthrIf = "ln0";
BYTE *getmyipaddr(void)
{
  char addr[64];
  int found = TRUE; /* be optimistic */
  static BYTE myadr[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  static int done = FALSE;
# ifdef BOOT_STRING_TO_STRUCT
  BOOT_PARAMS cfgStruct;
# endif

  if (done) return myadr;

  if (strlen(gtEthrAddr) > 0)
  {
    done = TRUE;
    inet_pton(gAfInet,gtEthrAddr,myadr);
    return myadr;
  }
# ifdef BOOT_STRING_TO_STRUCT
  bootStringToStruct(BOOT_LINE_ADRS,&cfgStruct);
# endif

  if (ifAddrGet(gtEthrIf,addr) == OK) goto out;
  gtEthrIf = "ei0";
  if (ifAddrGet(gtEthrIf,addr) == OK) goto out;
  gtEthrIf = "fei0";
  if (ifAddrGet(gtEthrIf,addr) == OK) goto out;
  gtEthrIf = "gei0";
  if (ifAddrGet(gtEthrIf,addr) == OK) goto out;
  found = FALSE;
out:
  if (found) 
  {
    dbglog("my ip address (interface %s) : %s",gtEthrIf,addr);
    inet_pton(gAfInet,gtEthrAddr,myadr);
    done = TRUE;
  }
  else
  {
    dbglog("cannot determine my ip address (interface name not one of \"ei0\", \"fei0\", \"gei0\", \"ln0\"");
  }
  return myadr;
}
# else /* (_WRS_VXWORKS_MAJOR >= 6)*/
BYTE *getmyipaddr(void)
{
  static char hname[30];
  static SCKADR stlclAddr;
  static SCKADR strmtAddr;
  static int done = 0;
  static BYTE myadr[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int sck;
  if (done) return myadr;
  stlclAddr.ss_family = (short)gAfInet;
  setInAddrAny(&stlclAddr);
  if (gethostname(hname,30) != -1)
  {
    struct addrinfo *ai, hints;
    memset(&hints,0,sizeof(hints)); hints.ai_family = gAfInet;
    if (getaddrinfo(hname,NULL,&hints,&ai) == 0)
    { /* try the simple way first */
      int len = gAfInet == AF_INET6 ? sizeof(SCKADR) : sizeof(struct sockaddr_in);
      memcpy(&stlclAddr,ai->ai_addr,len);
      freeaddrinfo(ai);
    }
  }
  if (isInAddrAny(&stlclAddr) || /* simple didn't work */
      isInAddrLoopback(&stlclAddr))   /* or came back with the loopback */
  {
    if ((sck=(int)socket(gAfInet, SOCK_DGRAM, 0)) > 0)
    {
      BYTE ipadr[16];
      _socklen_t nRet=(_socklen_t)sizeof(SCKADR);
      if (gAfInet == AF_INET) inet_pton(AF_INET,"128.127.50.1",ipadr);
      ip4adr(&strmtAddr,ipadr,IPPORT_ECHO);
      if (connect(sck,(struct sockaddr *)&strmtAddr,nRet) != -1)
      {
        getsockname(sck,(struct sockaddr *)&stlclAddr,(int *)&nRet);
      }
      soclose(sck);
    }
  }
  done = TRUE;
  return getInAddr(&stlclAddr);
}
# endif /* VXWORKS version check */
int GetThisFecAddresses(FecAddrStruct *fec)
{
  return not_implemented;
}
# endif /* !defined(VWWORKS) */
#else  /* !defined(VMS) && !defined(NIOS) && !defined(DOS) && !defined(SMALL_TINE_LIB) */
int getSrvIPAddr(void) { return not_implemented; }
# if defined(NIOS) || defined(VMS)
BYTE *getmyipaddr(void)
{
  static BYTE myadr[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  return myadr;
}
# endif
#endif /* !defined(VXWORKS) && !defined(VMS) && !defined(NIOS) && !defined(DOS) && !defined(SMALL_TINE_LIB) */
char *getmyipaddrstr(void)
{
  static int done=FALSE;
  if (!done)
  {
    inet_ntop(gAfInet,getmyipaddr(),gMyipAddrBuffer,40);
    done = TRUE;
  }
  return gMyipAddrBuffer;
}
TINE_EXPORT char *GetMyIpAddress(void)
{
  static char ip[64];
  memcpy(ip,getmyipaddrstr(),64);
  return ip;
}
/* trivial RPC call: */
int tUDPCall(int fecindex,BYTE *buf,UINT32 msgsize,int timout)
{
  struct pollfd fds[1];
  static SCKADR sckadr;
  static ClnHdr cln;
  static UINT16 ltid = 0xffff;
  RspHdr rsphdr;
  RspSpecs specs;
  int cc,n;
  int p = (gAfInet == AF_INET6) ? UDP6 : UDP;
  _socklen_t len;
  BYTE lclbuf[IPX_BUFFER_SIZE];
  FecAddrStruct *fec = fecindex ? &FecTbl[fecindex] : &NameServer[curNameServer];

  if (udpSyncSck == 0)
  { 
    if (setupUdpSocket(&udpSyncSck,gAfInet,0,FALSE,MIN_RCVBUF,MIN_SNDBUF) != 0) 
    {
      rprtSckErr("tUDPCall","setupUdpSocket",udpSyncSck,-1,0,NULL);
      return udp_socket_error;
    }
  }
  fds[0].fd = udpSyncSck; fds[0].events = POLLIN; fds[0].revents = 0;
  while (poll(fds,1,0) > 0 && !pollerr(0))
  { /* flush any input bytes that are somehow there ... */
    len = sizeof(sckadr);
    recvfrom(udpSyncSck,(char *)lclbuf,IPX_BUFFER_SIZE,0,(struct sockaddr *)&sckadr,&len);
  }
  memset(&cln,0,sizeof(ClnHdr));
  strncpy(cln.usr,fec->fecName,USERNAME_SIZE);
  if (!zerocmp(&fec->netAdr,sizeof(INADR))) return address_unknown; 
  sckadr_in(&sckadr,(BYTE *)&fec->netAdr,(UINT16)(gtUDPPort + fec->portOffset),gAfInet);
  memcpy(&cln.adr.ip,&sckadr,sizeof(sckadr));
  cln.ep.sck = udpSyncSck;
  cln.inetProtocol = UDP;
  if ((cc=sendIpData(&cln,buf,msgsize,CLN_REQ_,p,TRUE,TRUE)) != 0) return cc;
lsnForDataGram:
  len = sizeof(sckadr);
  fds[0].revents = 0;
  switch (poll(fds,1,timout))
  {
    case -1: 
      rprtSckErr("tUDPCall","select",udpSyncSck,-1,1,NULL);
      return udp_socket_error;
    case 0:  
      return link_timeout;
    default: 
      if (pollerr(0))
      {
        rprtSckErr("tUDPCall","select",udpSyncSck,-1,1,NULL);
        return udp_socket_error;
      }
      n = recvfrom(udpSyncSck,(char *)buf,IPX_BUFFER_SIZE,0,(struct sockaddr *)&sckadr,&len);
  }
  prepIncomingResponseHdr(&rsphdr, &specs, buf, 0);
  if (n != (int)rsphdr.totalSizeInBytes) return remitted_data_lost;
  if (specs.cc == illegal_protocol) return specs.cc;
  if (specs.id == ltid)
  { /* phantom block ! */
    goto lsnForDataGram;
  }
  ltid = specs.id;
  return 0;
}
#ifndef RDATE_TIMEOUT
# define RDATE_TIMEOUT 1
#endif
#define IPPORT_TIMESERVER	37              /* well known internet port number for rdate	*/
#define SEVENTY_YEARS		2208988800UL     /* midnight of 1.1.1900 to midnight of 1.1.1970	*/
char TimeServerHost[64];
#if defined(VXWORKS) && (_WRS_VXWORKS_MAJOR < 6)
# define NEED_GETHOSTBYNAME
#endif
#ifdef NEED_GETHOSTBYNAME
struct hostent *gethostbyname(const char *hstname)
{
  static struct hostent hst;
  static char addr[4];
  static char *paddr[2] = {NULL, NULL};
  int haddr;

  if ((haddr = hostGetByName((char *)hstname)) == ERROR) return NULL;

  hst.h_name = (char *)hstname;
  hst.h_aliases = 0;
  hst.h_addrtype = AF_INET;
  hst.h_length = strlen(hstname);
  memcpy(addr,&haddr,4);
  hst.h_addr_list = paddr;
  hst.h_addr = addr;

  return &hst;
}
struct hostent *gethostbyaddr(const char *hstaddr, int hstlen, int hsttype)
{
  static struct hostent hst;
  static char haddr[4];
  static char *addr; 
  char *c, hstaddrBuffer[32];
  int i = 0;

  strncpy(hstaddrBuffer,hstaddr,16); hstaddrBuffer[16] = 0;
  hst.h_name = (char *)hstaddr;
  hst.h_aliases = 0;
  hst.h_addrtype = hsttype;
  hst.h_length = hstlen;

  for (c=strtok(hstaddrBuffer,"."); c != NULL; c=strtok((char *)NULL,".")) 
  {
    haddr[i++] = atoi(c);
  }
  addr = (char *) &(haddr[0]);
  hst.h_addr = addr;

  return &hst;
}
#endif
time_t getUnixTime(char *tsHost)
{
  int sockfd=-1; /* socket fd's */
  static SCKADR svrAdr;
  UINT32 tmptime;
  static struct hostent *host = NULL;
  static int firstCallLogged = 0;
  int cc = 0,mode = 1,done = 0, e;
  time_t interim;

  if (firstCallLogged == 0)
  {
    firstCallLogged++;
    feclog("using TIME Server %s", tsHost == NULL ? "local host" : tsHost);
  }
  if (tsHost == NULL) return time(NULL);
  if (host == NULL)
  {
    char pstr[32];
    struct addrinfo hints, *res;
    sprintf(pstr,"%d",IPPORT_TIMESERVER);
    memset(&hints,0,sizeof(struct addrinfo));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    if (getaddrinfo(tsHost,pstr,&hints,&res) == 0)
    {
      svrAdr = *(SCKADR *)res->ai_addr;
      freeaddrinfo(res);
    }
  }
  if ((sockfd=(int)socket(AF_INET,SOCK_STREAM,0)) < 0 )
  {
    rprtSckErr("getUnixTime","socket",0,-1,1,NULL);
    ccerr(tcp_socket_error);
  }
  if (soioctl(sockfd,FIONBIO,(char *)&mode))
  {
    rprtSckErr("getUnixTime","soioctl",sockfd,-1,1,NULL);
    ccerr(tcp_socket_error);
  }
  for (time(&interim);
       (connect(sockfd,(struct sockaddr *)&svrAdr,sizeof(svrAdr))<0);)
  {
    if ((e=soerrno) == EISCONN) break;
#   ifdef WIN32
    if (e == WSAEISCONN) break; /* Microsoft keeps redefining things from version to version */
#   endif
    if (time(NULL) > interim + RDATE_TIMEOUT)
    {
      rprtSckErr("getUnixTime","connect",sockfd,-1,1,NULL);
      ccerr(tcp_connect_error);
    }
  }
  while (time(NULL) < interim + RDATE_TIMEOUT && !done)
  {
    if (soread(sockfd,(char *)&tmptime,sizeof(UINT32)) == sizeof(UINT32)) done = 1;
  }
  if (!done)
  {
    rprtSckErr("getUnixTime","soread",sockfd,-1,2,NULL);
    ccerr(connection_timeout);
  }
  if (firstCallLogged == 1)
  {
    firstCallLogged++;
    feclog("got time from %s",tsHost);
  }
err:
  if (sockfd > 0) soclose(sockfd);
  if (cc)
  {
    if (firstCallLogged < 3)
    {
      firstCallLogged++;
      feclog("get time from %s : %.32s",tsHost,cc2str(cc));
    }
    return 0;
  }
  return (ntohl(tmptime) - SEVENTY_YEARS);  /* convert to unix time	*/
}
int ip4adr(SCKADR *sckadr,BYTE *adr,UINT16 port)
{
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  memset(sa4,0,sizeof(struct sockaddr_in));
  if (adr != NULL) memcpy(&sa4->sin_addr,adr,sizeof(struct in_addr));
  else sa4->sin_addr.s_addr = INADDR_ANY;
  sa4->sin_port = htons(port);
  sa4->sin_family = AF_INET;
  if (sa4->sin_addr.s_addr == 0) return address_unknown;
  return 0;
}
int ip6adr(SCKADR *sckadr,BYTE *adr,UINT16 port)
{
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  memset(sa6,0,sizeof(struct sockaddr_in6));
  if (adr != NULL) memcpy(&sa6->sin6_addr,adr,sizeof(struct in6_addr));
  else sa6->sin6_addr = in6addr_any;
  sa6->sin6_port = htons(port);
  sa6->sin6_family = AF_INET6;
  if (sa6->sin6_addr.s6_addr[0] == 0 && sa6->sin6_addr.s6_addr[15] == 0 ) return address_unknown;
  return 0;
}
int sckadr_in(SCKADR *sckadr,BYTE *adr,UINT16 port,int family)
{
  if (family == AF_INET) return ip4adr(sckadr,adr,port);
  if (family == AF_INET6) return ip6adr(sckadr,adr,port);
  return not_supported;
}
BYTE *getInAddr(SCKADR *sckadr)
{
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  return sckadr->ss_family == AF_INET ? (BYTE *)&sa4->sin_addr.s_addr : sa6->sin6_addr.s6_addr;
}
int getInAddrLen(int family)
{
  switch (family)
  {
    case AF_INET6: return sizeof(struct sockaddr_in6);
    case AF_INET: return sizeof(struct sockaddr_in);
#   ifdef UNIX
    case AF_UNIX: return sizeof(struct sockaddr_un);
#   endif
    default: break;
  }
  return sizeof(struct sockaddr_in);
}
int sckadrcmp(SCKADR *adr1,SCKADR *adr2)
{
  if (adr1->ss_family == AF_INET)
  { /* IPv4 */
    struct sockaddr_in *sa1 = (struct sockaddr_in *)adr1;
    struct sockaddr_in *sa2 = (struct sockaddr_in *)adr2;
    if (sa1->sin_port != sa2->sin_port) return -1;
    return memcmp(&sa1->sin_addr,&sa2->sin_addr,4);
  }
  if (adr1->ss_family == AF_INET6)
  {
    struct sockaddr_in6 *sa1 = (struct sockaddr_in6 *)adr1;
    struct sockaddr_in6 *sa2 = (struct sockaddr_in6 *)adr2;
    if (sa1->sin6_port != sa2->sin6_port) return -1;
    return memcmp(&sa1->sin6_addr,&sa2->sin6_addr,16);
  }
  return -1;
}
/**
 * \internal
 * used internally only. 
 * no null pointer checks and dst must be long enough
 */
int inetStrAdr(SCKADR *sckadr,char *dst,int addport)
{
  int p, fam = sckadr->ss_family;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  void *padr;
  dst[0] = 0; /* truncate to begin with ... */
  if (fam == AF_INET) 
  {
    padr = &sa4->sin_addr;
    p = ntohs(sa4->sin_port);
  }
  else 
  {
    padr = &sa6->sin6_addr;
    p = ntohs(sa6->sin6_port);
  }
  if (inet_ntop(fam,padr,dst,40) == NULL) return non_existent;
  if (addport) sprintf(&dst[strlen(dst)],":%d",p);
  return 0;
}
TINE_EXPORT int GetInetAddress(SCKADR *sckadr, char *addr)
{
  if (sckadr == NULL || addr == NULL) return argument_list_error;
  return inetStrAdr(sckadr, addr, TRUE);
}
UINT16 inetPort(SCKADR *sckadr)
{
  int fam = sckadr->ss_family;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  if (sckadr == NULL) return 0;
  if (fam == AF_INET) return ntohs(sa4->sin_port);
  return ntohs(sa6->sin6_port);
}
void setInAddrAny(SCKADR *sckadr)
{
  int fam = sckadr->ss_family;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  if (sckadr == NULL) return;
  if (fam == AF_INET) sa4->sin_addr.s_addr = INADDR_ANY;
  else sa6->sin6_addr = in6addr_any;
  return;
}
void setInAddrLoopback(SCKADR *sckadr)
{
  int fam = sckadr->ss_family;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  if (sckadr == NULL) return;
  if (fam == AF_INET) sa4->sin_addr.s_addr = INADDR_LOOPBACK;
  else sa6->sin6_addr = in6addr_loopback;
  return;
}
int isInAddrAny(SCKADR *sckadr)
{
  int fam = sckadr->ss_family;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  if (sckadr == NULL) return FALSE;
  if (fam == AF_INET) return sa4->sin_addr.s_addr == INADDR_ANY;
  return !memcmp(&sa6->sin6_addr,&in6addr_any,16);
}
int isInAddrLoopback(SCKADR *sckadr)
{
  int fam = sckadr->ss_family;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)sckadr;
  struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)sckadr;
  if (sckadr == NULL) return FALSE;
  if (fam == AF_INET) return sa4->sin_addr.s_addr == INADDR_LOOPBACK;
  return !memcmp(&sa6->sin6_addr,&in6addr_loopback,16);
}
#ifndef USE_IPV6
IN6_ADDR in6addr_any = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
IN6_ADDR in6addr_loopback = { 0x7f, 0xff, 0xff, 0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x01 };
#ifdef VXWORKS
INADR inadr_zero = { 0 };
#else
INADR inadr_zero = { 0, 0, 0, 0 };
#endif
int inet_pton(int family,const char *src,void *dst)
{
  DWORD haddr;
  if (src == NULL || dst == NULL) return argument_list_error;
  if (family != AF_INET) return not_supported;
# ifdef VXWORKS
  haddr = inet_addr((char *)src);
# else
  haddr = inet_addr(src);
# endif
  memcpy(dst,&haddr,4);
  return 0;
}
char *inet_ntop(int family,const void *src,char *dst,size_t cnt)
{
  if (src == NULL || dst == NULL) return NULL;
  if (family != AF_INET) return NULL;
  if (cnt < 16) return NULL;
  inet_ntoa_b(*(struct in_addr *)src,dst);
  return dst;
}
#define NEED_GETNAMEINFO
#define NEED_GETADDRINFO
#else
INADR inadr_zero = {{{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }}};
#endif
/* VxWorks sometimes needs these extra (depending on the build) */
#ifdef NEED_IN6ADDR_ANY
const struct in6_addr in6addr_any = {{{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }}};
#endif
#ifdef NEED_IN6ADDR_LOOPBACK
const struct in6_addr in6addr_loopback = {{{ 0x7f, 0xff, 0xff, 0xff, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x01 }}};
#endif
#ifdef NEED_GETADDRINFO
int getaddrinfo(const char *hostname,
  const char *service,
  const struct addrinfo *hints,
  struct addrinfo **res)
{
  struct hostent *hent;
  static struct addrinfo ai;
  static struct sockaddr_in sa;
  UINT16 p = 0;
  if (hostname == NULL) return -1;
  if (strchr(hostname, '.') != NULL && isdigit(hostname[0]))
  { /* string ip address given */
#   ifdef VXWORKS
    UINT32 adr = inet_addr((char *)hostname);
#   else
    UINT32 adr = inet_addr(hostname);
#   endif
    hent = gethostbyaddr((const char *)&adr, 4, AF_INET);
  }
  else
  {
    hent = gethostbyname(hostname);
  }
  if (hent == NULL) return -1;
  memset(&ai, 0, sizeof(ai));
  ai.ai_family = hent->h_addrtype;
  ai.ai_protocol = IPPROTO_IP;
  ai.ai_addrlen = hent->h_length;
  ai.ai_canonname = hent->h_name;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = hent->h_addrtype;
  memcpy(&sa.sin_addr, hent->h_addr, 4);
  if (service != NULL) p = (UINT16)atoi(service);
  sa.sin_port = htons(p);
  ai.ai_addr = (struct sockaddr *)&sa;
  if (res != NULL) *res = &ai;
  return 0;
}
void freeaddrinfo(struct addrinfo *ai)
{ /* just return */
  ai = ai;
  return;
}
#endif

#ifdef NEED_GETNAMEINFO
int getnameinfo(const struct sockaddr *sa, _socklen_t salen,
                       char *host, size_t hostlen,
                       char *serv, size_t servlen, int flags)
{
  struct hostent *hent;
  serv = serv;
  servlen = servlen;
  flags = flags;
  if ((hent=gethostbyaddr((const char *)sa,4,AF_INET)) == NULL) return -1;
  if (salen < 1) return -1;
  if (host == NULL) return -1;
  strncpy(host,hent->h_name,hostlen);
  return 0;
}
#endif
#ifdef NEED_POLL
int poll(struct pollfd *fds, int nfds, int timeout)
{
  fd_set rdfds, wrfds, erfds;
  fd_set *prdfds=NULL, *pwrfds=NULL, *perfds=NULL;
  struct timeval to;
  int i, nready, fdmax = 0;

  FD_ZERO(&rdfds); FD_ZERO(&wrfds); FD_ZERO(&erfds);

  for (i=0; i<nfds; i++)
  {
    if (fds[i].fd > fdmax) fdmax = fds[i].fd;
    if (fds[i].events & POLLIN) { FD_SET(fds[i].fd,&rdfds); prdfds = &rdfds; }
    if (fds[i].events & POLLOUT) { FD_SET(fds[i].fd,&wrfds); pwrfds = &wrfds; }
    FD_SET(fds[i].fd,&erfds); 
    perfds = &erfds;
    fds[i].revents = 0;
  }
  to.tv_sec = timeout/1000; to.tv_usec = (timeout%1000)*1000;
  nready = select(fdmax+1,prdfds,pwrfds,perfds,&to);
  if (nready > 0)
  {
    for (i=0; i<nfds; i++)
    {
      if (prdfds != NULL && fds[i].events & POLLIN && FD_ISSET(fds[i].fd,prdfds))
      {
        fds[i].revents |= POLLIN;
      }
      if (pwrfds != NULL && fds[i].events & POLLOUT && FD_ISSET(fds[i].fd,pwrfds))
      {
        fds[i].revents |= POLLOUT;
      }
      if (perfds != NULL && FD_ISSET(fds[i].fd,perfds))
      {
        fds[i].revents |= POLLERR;
      }
    }
  }
  return nready;
}
#endif /* NEED_POLL */
