/*
 * File listener.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file listener.c
 * \brief TINE static listener documentation 
 *
 */
#ifndef NO_DOXYGEN
#include "project.def"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include <math.h>
#include "tine.h"
#include "listener.h"
#ifdef MULTITHREADED
#include "threader.h"
#endif

extern int gBufSrvInitialized;
TINE_EXPORT int IsParentLink(int linkId);
#ifdef _USRDLL
void TineStartCycler(void);
#ifdef WIN32
int heapCheck(void)
{
  static int bad = FALSE;
  if (_heapchk() != _HEAPOK && !bad) 
  {
    bad = TRUE;
  }
  return bad;
}
#endif
#endif

#ifdef MULTITHREADED
# if defined(UNIX)
# if !defined(PTHREAD_MUTEX_RECURSIVE) && defined(PTHREAD_MUTEX_RECURSIVE_NP)
# define PTHREAD_MUTEX_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP
# endif
  pthread_mutex_t gLsnTblMutex = PTHREAD_MUTEX_INITIALIZER;
  MXHANDLE hLsnTblMutex = &gLsnTblMutex;
  int lsnInitMutex(MXHANDLE *mx)
  { /* don't really need a recursive mutex any more and the older linuxes have problems here */
#   ifdef OS_64BIT
    pthread_mutexattr_t Attr;
    pthread_mutexattr_init(&Attr);
    pthread_mutexattr_settype(&Attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(*mx, &Attr);
#   else
    pthread_mutex_init(*mx, NULL);
#   endif
    return 0;
  }
# elif defined(WIN32)
  MXHANDLE hLsnTblMutex;
  int lsnInitMutex(MXHANDLE *mx)
  {
    if (mx == NULL) return -1;
    *mx = CreateMutex(NULL,FALSE,NULL);
    if (*mx == NULL) return -2;
    return 0;
  }
# ifdef _USRDLL
  /* not exported out of the windows DLL */
  int WaitForMutex(MXHANDLE mx,int to)
  {
    int cc;
    if (mx == (MXHANDLE)0) return 0;
    cc = WaitForSingleObject(mx,(to) < 0 ? INFINITE : to);
    if (cc == WAIT_FAILED) cc = GetLastError();
    return cc == WAIT_OBJECT_0 ? 0 : cc;
  }
# endif
# elif defined(VXWORKS)
  MXHANDLE hLsnTblMutex;
# endif
  int ReleaseLsnMutex(MXHANDLE mx) 
  { 
    int cc = 0;
    if (mx == (MXHANDLE)0) return 0; /* already freed (?) */
#   if defined(UNIX)
    cc = ReleaseMutex(mx);
#   elif defined(VXWORKS)
    if ((cc=ReleaseMutex(mx)) == OK) cc = 0;
#   elif defined(WIN32)
    if (ReleaseMutex(mx) == 0) cc = GetLastError();
#   else
    ReleaseMutex(mx);
#   endif
    return cc;
  }
#else
int hLsnTblMutex = 0;
#define ReleaseLsnMutex(m) (m = 0)
#endif

void initLsnMutex(void)
{
# ifdef MULTITHREADED
  lsnInitMutex(&hLsnTblMutex);
# else
  hLsnTblMutex = 0;
# endif
}

int getlsnrs(int i1,int i2,int i3,int i4)
{
  mexListener *ml=NULL;
  int i;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    ttyoutput("%.128s[%.64s] %d %.16s value(s) <%d> (remaining idle time: %d sec, link id %d)",
      ml->devName,ml->devProperty,ml->dout.dArrayLength,
      GetFormatAsString(ml->dout.dFormat),ml->status,
      (ml->allowedIdleTime - ml->idleTimeInSeconds),
      ml->lid);
  }
  return 0;
}
void initSettings(void)
{
  char *ptr=getenv("TINE_LISTENER_IDLE_TIME");
  if (ptr != NULL) 
  {
    SetAllowedIdleTime(atoi(ptr));
  }
  RegisterUserCommand("listeners",getlsnrs,NULL,NULL,CA_READ);
}
LSNR_LINK *gLsnrLnkTbl[LSNRTBL_HASH_SIZE];
char *makeLsnrLnkKey(char *devName,char *devProperty,DTYPE *dout,DTYPE *din)
{
  static char key[256];
  char outkey[64],inkey[64];

  sprintf(key,"%.132s[%.64s]",devName,devProperty);
  strupr(key);
  if (dout != NULL) 
  {
    sprintf(outkey,"%d%.16s",dout->dArrayLength,GetFormatAsString(dout->dFormat));
  }
  else
  {
    strcpy(outkey,"[NULL]");
  }
  if (din != NULL && din->dArrayLength > 0) 
  { /* this could have collisions!  trap din->dArrayLength > 0 earlier and don't hash ! */
    sprintf(inkey,"%d%.16s",din->dArrayLength,GetFormatAsString(din->dFormat));
  }
  else
  {
    strcpy(inkey,"[NULL]");
  }
  strncat(key,outkey,64);
# ifdef USE_HASH_INKEY
  strncat(key,inkey,64);
# endif
  return key;
}
LSNR_LINK *findLsnrLnk(char *key)
{
  LSNR_LINK *ll;
  int idx = (int)(ElfHash((unsigned char *)key) % LSNRTBL_HASH_SIZE);

  for (ll=gLsnrLnkTbl[idx]; ll != NULL; ll = ll->nxt)
  {
    if (!strncmp(ll->key,key,256)) return ll;
  }
  return NULL;   
}
void rmvLsnrLnk(char *key)
{
  LSNR_LINK *ll, *ll0, *llp;
  int idx = (int)(ElfHash((unsigned char *)key) % LSNRTBL_HASH_SIZE);

  llp = ll0 = gLsnrLnkTbl[idx];
  for (ll=ll0; ll != NULL; llp = ll, ll = ll->nxt)
  { /* run thru the list */
    if (!strncmp(ll->key,key,256))
    { /* that's the one to remove */
      if (ll == ll0) gLsnrLnkTbl[idx] = ll->nxt;
      else llp->nxt = ll->nxt;
      free(ll);
      break;
    }
  }
}
LSNR_LINK *getLsnrLnk(char *devName,char *devProperty,DTYPE *dout,DTYPE *din)
{
  char key[256];
  LSNR_LINK *ll;
  int idx;
  if (dout == NULL && din == NULL) 
  { /* don't hash these calls (e.g. from getPropertyInfoFromListenerEx()) */
    return NULL;
  }
  if (din != NULL && din->dArrayLength > 0) 
  { /* don't hash if there is input data */
    return NULL;
  }
  strncpy(key,makeLsnrLnkKey(devName,devProperty,dout,din),256);
  if ((ll=findLsnrLnk(key)) != NULL) return ll;
  if ((ll=(LSNR_LINK *)calloc(1,sizeof(LSNR_LINK))) == NULL)
  {
    feclog("no memory to add hash table entry %s !",key);
    return NULL;
  }
  strncpy(ll->key,key,256);
  ll->mlid = -1;
  idx = (int)(ElfHash((unsigned char *)key) % LSNRTBL_HASH_SIZE);
  ll->nxt = gLsnrLnkTbl[idx];
  gLsnrLnkTbl[idx] = ll;
  return ll;
}

mexListener **MexListenerTable = NULL;
int MexListenerTableCapacity = MAX_NUM_LISTENERS;
int defaultPollingMode = CM_POLL;
int gListenerTableError = 0;
int gWaitingForLink = -1;
int gAllowedIdleTime = ALLOWED_IDLE_TIME;
EXPORT_B void SetAllowedIdleTime(int idleTime)
{
  if (idleTime < 5) idleTime = 5;
  gAllowedIdleTime = idleTime;
}
EXPORT_B void SetListenerTableCapacity(int capacity)
{
  if (capacity < 10) capacity = 10;
  MexListenerTableCapacity = capacity;
}
EXPORT_B int GetListenerTableCapacity(void)
{
  return MexListenerTableCapacity;
}
EXPORT_B int getListenerTableCapacity(void)
{
  return MexListenerTableCapacity;
}
/* borrow this (un-exported) helper routine from tine lib */
int isAdjFmt(short fmt)
{
  switch (fmt)
  {
    case CF_STRING:
    case CF_AIMAGE:
    case CF_ASPECTRUM:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
void listenerCb(int id,int cc,void *ref)
{
  int i, n, siz, slen, mlid = id - 1, firstcall = FALSE;
  mexListener *ml = (mexListener *)ref;
  BYTE *tmp;
  time_t this_t = time(NULL);
  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return;
  if (ml == NULL)
  { // reference == NULL ?
    if (mlid < 0 || mlid >= nMexListeners) goto err; /* already free()ed ? */
    if ((ml=MexListenerTable[mlid]) == NULL) goto err;
  }
  if (ml->lid < 0)
  { /* called during the attachlink wait ? */
    ml->lastsize = ml->dout.dArrayLength;
    firstcall = TRUE;
  }
  else
  {
    ml->lastsize = ml->isGlobal ? (int)ml->dout.dArrayLength : GetCompletionDataSize(ml->lid);
  }
  if (cc == link_timeout)
  { /* some people get really irritated at infrequent timeout messages */
    if (ml->to_count++ < 2 && !firstcall) goto err;
  }
  else
  {
    ml->to_count = 0;
  }
  if (ml->lastsize > (int)ml->dout.dArrayLength )
  { // should not occur !
    ml->lastsize = (int)ml->dout.dArrayLength;
  }
  switch (ml->dout.dFormat)
  {
    case CF_STRING:
    {
      char *pstr=NULL;
      int zsiz = 0;
      n = ml->lastsize;
      siz = 0;
      for (i=0; i<n; i++) 
      { /* find the longest string */
        if ((pstr=ml->dout.data.strptr[i]) == NULL) break;
        if ((slen=(int)strlen(pstr)) > siz) siz = slen;
      }
      if (siz > ml->capacity)
      { /* adjust them all */
        siz += STRING_STD_CAPACITY;
        zsiz = siz - ml->capacity;
        for (i=0; i<n; i++) 
        {
          if ((tmp=(BYTE *)realloc(ml->outarr[i],siz)) == NULL) break;
          memset(&tmp[ml->capacity], 0, zsiz);
          ml->outarr[i] = tmp;
        }
        if (i == n) ml->capacity = siz;
      }
      for (i=0; i<n; i++)
      {
        if ((pstr=ml->dout.data.strptr[i]) == NULL) break;
        strncpy((char *)ml->outarr[i],ml->dout.data.strptr[i],ml->capacity);
      }
      if (i < n) ml->lastsize = i;
      break;
    }
    default:
      break;
  }
  ml->status = cc;
  if (cc != 0)
  {
    strncpy(ml->statusString, GetCompletionStatus(ml->lid), STATUS_SIZE);
  }
  else
  {
    ml->statusString[0] = 0;
  }
  if (ml->lasttime == 0) ml->lasttime = this_t;
  if (!ml->isGlobal)
  {
    ml->idleTimeInSeconds += (int)(this_t - ml->lasttime);
    if (ml->allowedIdleTime < 5) ml->allowedIdleTime = gAllowedIdleTime;
    if (ml->idleTimeInSeconds > ml->allowedIdleTime) 
    {
      stopListener(ml);
      goto err;
    }
    ml->timestamp = GetCurrentDataTimeStamp(ml->lid);
    if (ml->timestamp < 100) 
    {
      ml->timestamp = MakeDataTimeStamp(); /* just starting up ? */
      ml->sysstamp = GetSystemDataStampFromCallbackId(id);
      ml->usrstamp = GetDataStampFromCallbackId(id);
    }
    else
    {
      ml->sysstamp = GetSystemDataStamp(ml->lid);
      ml->usrstamp = GetDataStamp(ml->lid);
    }
  }
  else
  {
#   ifndef TINE32_DLL
    ml->timestamp = PutDataTimeStampU(0,glbTbl[ml->lid]->time,glbTbl[ml->lid]->timeUSEC);
    ml->sysstamp = glbTbl[ml->lid]->sysStamp;
    ml->usrstamp = glbTbl[ml->lid]->dataStamp;
#   else
    ml->timestamp = GetGlobalsTimeStamp(ml->lid);
    ml->sysstamp = GetGlobalsSystemStamp(ml->lid);
    ml->usrstamp = GetGlobalsDataStamp(ml->lid);
#   endif
  }
# ifndef MULTITHREADED
  hLsnTblMutex = 0; /* need a recursive mutex in case the notifier 'calls something'  */
# endif
  if (ml->nf != NULL) ml->nf(ml->nid);
  ml->lasttime = this_t;
err:
  ReleaseLsnMutex(hLsnTblMutex);
  return;
}
int getValidIdleTime(UINT32 payload, int idletime)
{
  if (payload > XLARGE_PAYLOAD && idletime > 10) return 10;
  if (payload > LARGE_PAYLOAD && idletime > 20) return 20;
  if (payload > MEDIUM_PAYLOAD && idletime > 30) return 30;
  return idletime;
}
mexListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int modeflags)
{ // does NOT have hLsnTblMutex when called !
  int ctc = GetConnectionTableCapacity();
  int i,cc=0;
  int siz,hsiz,asiz=0,sizIn,ssz,mode=defaultPollingMode|CM_WAIT,hasIncremented=FALSE;
  mexListener *ml=NULL;
  char *c, glbkey[128];

  if (MexListenerTable == NULL)
  {
    SetClnRecvQueueDepth(0);
    if (MexListenerTableCapacity > ctc) SetConnectionTableCapacity(MexListenerTableCapacity);
    if ((MexListenerTable=(mexListener **)SystemCalloc(MexListenerTableCapacity,sizeof(mexListener *))) == NULL)
    {
      gListenerTableError = out_of_local_memory;
      feclog("Cannot allocate listener table !");
      return NULL;
    }
  }
  siz = dout->dArrayLength * fmtsizeof(dout->dFormat);
  hsiz = getFormatHeaderSize(dout->dFormat);
  switch (dout->dFormat)
  {
    case CF_IMAGE:
    case CF_GSPECTRUM:
      siz += hsiz + sizeof(UINT32) + sizeof(BYTE *);
      break;
    case CF_MDA:
      siz = dout->dArrayLength * fmtsizeof(GetFormatFromString(dout->dTag));
      siz += hsiz + sizeof(BYTE *);
      break;
    case CF_STRUCT:
      if (!StructIsBlittable(dout->dTag))
      {
        gListenerTableError = use_stream_transport; /* special signal */
        return NULL;
      }
      if ((ssz = GetStructSize(dout->dTag)) <= 0)
      {
        gListenerTableError = invalid_structure_tag;
        return NULL;
      }
      siz *= ssz;
      break;
    default:
      if (isAdjFmt(dout->dFormat))
      {
        hsiz = hsiz + sizeof(void *);
        siz *= hsiz;
        asiz = getFormatStandardCapacity(dout->dFormat);
      }
      else
      {
        siz += hsiz;
      }
      break;
  }
  if ((ml=(mexListener *)calloc(1,sizeof(mexListener))) == NULL) ccerr(out_of_local_memory);
  ml->lid = -not_initialized;
  ml->allowedIdleTime = gAllowedIdleTime;
  /* ml->status = link_timeout; <- dependent links won't WAIT for a callback !! */ 
  if ((ml->outbuf=(BYTE *)calloc(siz,1)) == NULL) ccerr(out_of_local_memory);
  if (asiz > 0)
  { /* adjustable format size */
    if ((ml->outarr=(BYTE **)calloc(dout->dArrayLength,sizeof(BYTE *))) == NULL) ccerr(out_of_local_memory);
    ml->capacity = asiz;
    asiz *= getFormatTypeSize(dout->dFormat);
    for (i=0; i<(int)dout->dArrayLength; i++)
    {
      if ((ml->outarr[i]=(BYTE *)calloc(asiz,1)) == NULL) ccerr(dout->dArrayLength);
    }
  }
  ml->dout.dArrayLength = dout->dArrayLength;
  ml->dout.dFormat = dout->dFormat;
  memcpy(ml->dout.dTag,dout->dTag,TAG_NAME_SIZE);
  ml->dout.data.vptr = ml->outbuf;
  switch (dout->dFormat)
  {
    case CF_IMAGE:
      ((DIMAGE *)ml->dout.data.bptr)->frameBuffer = ml->dout.data.bptr + hsiz + sizeof(UINT32) + sizeof(BYTE*);
      ((DIMAGE *)ml->dout.data.bptr)->frameBufferSize = dout->dArrayLength;
      mode |= CM_SYNCNOTIFY;
      break;
    case CF_GSPECTRUM:
      ((DGSPECTRUM *)ml->dout.data.bptr)->spectBuffer = (float *)(ml->dout.data.bptr + hsiz + sizeof(UINT32) + sizeof(BYTE*));
      ((DGSPECTRUM *)ml->dout.data.bptr)->spectBufferSize = dout->dArrayLength;
      mode |= CM_SYNCNOTIFY;
      break;
    case CF_MDA:
      ((DMDA *)ml->dout.data.bptr)->data.bptr = ml->dout.data.bptr + MDA_HDR_SIZE + sizeof(BYTE*);
      mode |= CM_SYNCNOTIFY;
      break;
    default:
      break;
  }
/*  if (dout->dArrayLength > XLARGE_PAYLOAD && ml->allowedIdleTime > 10) ml->allowedIdleTime = 10; */
  ml->allowedIdleTime = getValidIdleTime(dout->dArrayLength,ml->allowedIdleTime);
  if (din != NULL && (sizIn=din->dArrayLength * fmtsizeof(din->dFormat)) > 0)
  {
    hsiz = getFormatHeaderSize(din->dFormat);
    if (isAdjFmt(din->dFormat))
    {
      hsiz = hsiz + sizeof(void *);
      sizIn *= hsiz;
    }
    else if (din->dFormat == CF_STRUCT)
    { /* must be tagged and registered! */
      if (!StructIsBlittable(din->dTag))
      {
        gListenerTableError = use_stream_transport; /* special signal */
        return NULL;
      }
      if ((ssz=GetStructSize(din->dTag)) <= 0) ccerr(invalid_structure_tag);
      sizIn *= ssz;
    }
    else
    {
      sizIn += hsiz;
    }
    if ((ml->inbuf=(BYTE *)calloc(sizIn,1)) == NULL) ccerr(out_of_local_memory);
    ml->din.data.bptr = ml->inbuf;
    memcpy(ml->din.data.vptr,din->data.vptr,sizIn);
    ml->din.dArrayLength = din->dArrayLength;
    ml->din.dFormat = din->dFormat;
    memcpy(ml->din.dTag,din->dTag,TAG_NAME_SIZE);
  }
  else
  {
    ml->din.dFormat = CF_NULL;
    ml->din.dArrayLength = 0;
    ml->din.data.vptr = NULL;
    memset(ml->din.dTag,0,TAG_NAME_SIZE);
  }
  strncpy(ml->devName,devName,132);
  strncpy(ml->devProperty,devProperty,PROPERTY_NAME_SIZE);

  if (WaitForSystemMutex(hLsnTblMutex,-1) == 0)
  { // traverse the lsnr table => get the mutex
    for (i=0; i<MexListenerTableCapacity && MexListenerTable[i] != NULL; i++);
    if (i < MexListenerTableCapacity) MexListenerTable[i] = ml;
    ReleaseLsnMutex(hLsnTblMutex);
  }
  else
  { // this should never happen
    ml->lid = -mutex_error;
    ccerr(mutex_error);
  }
  if (i == MexListenerTableCapacity) 
  {
    ml = NULL;
    ccerr(resources_exhausted);
  }
  ml->mlid = i;
  if (isGlobal(devName))
  {
    memset(glbkey,0,128);
    if (devName[0] == '/')
    { /* context given ! */
      strncpy(glbkey,devName,64);
      if ((c=strchr(&glbkey[1],'/')) != NULL) 
      { /* server given : strip off device name if present ! */
        c++;
        if ((c=strchr(c,'/')) != NULL) 
        { /* has no relevance for a 'global' ! */
          *c = 0;
        }
      }
      strcat(glbkey,"/");
    }
    strncat(glbkey,devProperty,PROPERTY_NAME_SIZE);
    ml->lid = recvNetGlobalEx2(glbkey,&ml->dout,listenerCb,ml->mlid + 1,ml);
    ml->isGlobal = TRUE;
  }
  else
  {
    if (modeflags > 0) mode |= modeflags;
    ml->lid = AttachLinkEx2(devName,devProperty,&ml->dout,&ml->din,CA_READ,pollrate,listenerCb,mode,ml->mlid + 1,ml);
    ml->isGlobal = FALSE;
  }
err:
  if (ml != NULL && ml->lid < 0)
  {
    MexListenerTable[ml->mlid] = NULL;
    gListenerTableError = cc ? cc : -ml->lid;
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    if (ml->outarr != NULL)
    {
      for (i=0; i<(int)dout->dArrayLength; i++) if (ml->outarr[i] != NULL) free(ml->outarr[i]);
      free(ml->outarr);
    }
    free(ml);
    if (hasIncremented) nMexListeners--;
    ml = NULL;
  }
  if (ml != NULL)
  {
    if (ml->dout.dArrayLength != 0)
    {
      if (dout->dFormat == CF_IMAGE) siz = hsiz;
      if (dout->dFormat == CF_GSPECTRUM) siz = hsiz;
      if (dout->dFormat == CF_MDA) siz = hsiz;
      memcpy(dout->data.vptr,ml->outbuf,siz);
    }
    ml->pollingRate = pollrate;
  }
  gListenerTableError = cc; 
  return ml;
}
int stopListener(mexListener *ml)
{
  int i,mlid,cc=0;
  char key[256];
  if (ml == NULL) return non_existent_elem;
  if (IsParentLink(ml->lid)) 
  { /* this listener has somehow gotten to be a parent ! */
    ml->idleTimeInSeconds = 0;
    ccerr(not_accepted);
  }
  if (ml->isGlobal)
  {
    CloseGlobalLink(ml->lid);
  }
  else
  {
    CloseLink(ml->lid);
  }
  strncpy(key,makeLsnrLnkKey(ml->devName,ml->devProperty,&ml->dout,&ml->din),256);
  rmvLsnrLnk(key);
  mlid = ml->mlid;
  if (MexListenerTable[mlid] == NULL) ccerr(not_running);
  if (ml->outbuf != NULL) free(ml->outbuf);
  if (ml->inbuf != NULL) free(ml->inbuf);
  if (ml->outarr != NULL)
  {
    for (i=0; i<(int)ml->dout.dArrayLength; i++) if (ml->outarr[i] != NULL) free(ml->outarr[i]);
    free(ml->outarr);
  }
  free(ml);
  MexListenerTable[mlid] = NULL;
  if (mlid == nMexListeners-1) nMexListeners--;
err:
  return cc;
}
void stopAllListeners(void)
{
  int i;
  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return;
  for (i=0; i<nMexListeners; i++)
  {
    stopListener(MexListenerTable[i]);
  }
  nMexListeners = 0;
  ReleaseLsnMutex(hLsnTblMutex);
}
void freeMexListenerTable(void)
{
  if (MexListenerTable == NULL) return;
  stopAllListeners();
  free(MexListenerTable);
  MexListenerTable = NULL;
}
int nMexListeners = 0;
mexListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int rate)
{
  mexListener *ml=NULL;
  UINT32 dsiz;
  LSNR_LINK *ll;
  int i = -1;

  if (MexListenerTable == NULL) return NULL;

  if ((ll=getLsnrLnk(devName,devProperty,dout,din)) != NULL)
  { /* got a hash table entry */
    if (ll->mlid >= 0) 
    { /* seems to be fully known */
      i = ll->mlid;
      if ((ml=MexListenerTable[i]) == NULL)
      { /* now this is unsual ! */
        rmvLsnrLnk(ll->key);
        i = -1;
      }
    }
  }
  if (i < 0)
  { /* not yet known to hash list */
    for (i=0; i<nMexListeners; i++)
    {
      if ((ml=MexListenerTable[i]) == NULL) continue;
      if (strnicmp(devName,ml->devName,132)) continue;
      if (strlen(devProperty) > 0 && strnicmp(devProperty,ml->devProperty,PROPERTY_NAME_SIZE)) continue;
      if (dout != NULL)
      {
        if (dout->dFormat != ml->dout.dFormat) continue;
        if (dout->dArrayLength != ml->dout.dArrayLength) continue;
        if (strncmp(dout->dTag,ml->dout.dTag,TAG_NAME_SIZE)) continue;
      }
      if (din != NULL && din->dArrayLength != 0)
      {
        if (din->dFormat != ml->din.dFormat) continue;
        if (din->dArrayLength != ml->din.dArrayLength) continue;
        if (strncmp(din->dTag,ml->din.dTag,TAG_NAME_SIZE)) continue;
        dsiz = ml->din.dArrayLength*fmtsizeof(ml->din.dFormat);
        if (din->dFormat == CF_SPECTRUM) dsiz += 96;
        if (memcmp(din->data.vptr,ml->inbuf,dsiz)) continue;
      }
      break;
    }
  }
  if (i < nMexListeners)
  {
    if (ml != NULL && rate != 0 && rate < ml->pollingRate)
    {  /* force re-start of listener at the faster rate */
       stopListener(ml); 
       ml = NULL;
    }
    if (ll != NULL) ll->mlid = i;
  }
  else
  {
    ml = NULL;
  }
  return ml;  
}
int lastMexListenerLid = -1;
char *lastMexListenerStatusString = NULL;
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int accessflags)
{
  int cc = 0, modeflags = 0, i;
  mexListener *ml;
  BYTE *bptr;
  UINT32 siz,ssz,hsiz,bsiz;

  if (devName == NULL || devProperty == NULL || dout == NULL) return argument_list_error;
  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
  gListenerTableError = 0;

  if (accessflags & CA_NETWORK) modeflags = CM_NETWORK;
  else if (accessflags & CA_CONNECT) modeflags = CM_CONNECT;

  if ((ml=getListener(devName,devProperty,dout,din,rate)) == NULL)
  { /* not in the current table */
    ReleaseLsnMutex(hLsnTblMutex); /* give up the table mutex */
    if ((ml=startListener(devName,devProperty,dout,din,rate,modeflags)) == NULL )
    { /* oops! */
      if (gListenerTableError == 0) gListenerTableError = un_allocated;
      return gListenerTableError;
    }
    if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
    i = ml->mlid;
    MexListenerTable[i] = ml;
    if (i == nMexListeners) nMexListeners++;
  }  
  if (dout->dFormat != ml->dout.dFormat) ccerr(illegal_format);
  ml->idleTimeInSeconds = 0;
  if (ml->timestamp < 100) ml->timestamp = MakeDataTimeStamp(); /* starting up ? */
  dout->dTimeStamp = ml->timestamp;
  dout->sysStamp = ml->sysstamp;
  dout->dStamp = ml->usrstamp;
  siz = MIN(dout->dArrayLength,ml->dout.dArrayLength)*fmtsizeof(dout->dFormat);
  hsiz = getFormatHeaderSize(dout->dFormat);
  if (isAdjFmt(dout->dFormat))
  {
    hsiz = hsiz + sizeof(void *);
    siz *= hsiz;
  }
  else if (dout->dFormat == CF_STRUCT)
  { /* must be tagged and registered! */
    if ((ssz=GetStructSize(dout->dTag)) <= 0) 
    {
      ccerr(invalid_structure_tag);
    }
    siz *= ssz;
  }
  else
  {
    siz += hsiz;
  }
  LockLinkData();
  if (ml->outarr != NULL) 
  { /* adjustable length format type */
    memcpy(dout->data.vptr,ml->outarr,siz);
  }
  else 
  { /* make sure lastdata size consistent with data set independent of callback */
    ml->lastsize = ml->isGlobal ? (int)ml->dout.dArrayLength : GetCompletionDataSize(ml->lid);
    if ((bptr=GetFormatBodyPtr(dout->data.vptr,dout->dFormat)) != NULL)
    { /* e.g CF_IMAGE */
      siz = hsiz; /* fix to just the header bytes for memcpy below */
      bsiz = GetFormatBodyBufferSize(dout->data.vptr,dout->dFormat);
      if (bsiz < (UINT32)ml->lastsize) 
      {
        UnlockLinkData();
        ccerr(buffer_too_small);
      }
      if (ml->lastsize > 0) memcpy(bptr,&ml->outbuf[hsiz+sizeof(UINT32)+sizeof(BYTE *)],ml->lastsize * fmtsizeof(dout->dFormat));
    }
    if (dout->dFormat == CF_MDA)
    {
      siz = hsiz; /* fix to just the header bytes for memcpy below */
      bptr = ((DMDA *)dout->data.vptr)->data.bptr;
      bsiz = ml->lastsize * fmtsizeof(GetFormatFromString(dout->dTag));
      if (ml->lastsize > 0) memcpy(bptr,&ml->outbuf[hsiz+sizeof(BYTE *)],bsiz);
    }
    if (dout->data.vptr != ml->outbuf) memcpy(dout->data.vptr,ml->outbuf,siz);
    if (accessflags & CA_MUTABLE) dout->dArrayLength = ml->lastsize;
  }
  if (dTime != NULL) *dTime = dout->dTimeStamp;
  UnlockLinkData();
  cc = ml->status;
  lastMexListenerStatusString = ml->statusString;
  lastMexListenerLid = ml->lid;
  switch (cc)
  {
    case access_denied:
    case illegal_property:
    case illegal_device:
    case illegal_format:
    case name_unknown:
      stopListener(ml);
      break;
    default:
      break;
  }
err:
  ReleaseLsnMutex(hLsnTblMutex);
  return cc;
}
int nerrors = -1;
char *errsrc[2] = { "LCL", "RMT" };
EXPORT_B char *alsnGetLastLinkError(short cc, char *errstr)
{
  static char s[64];
  int srcid = 0;

  if (errstr == NULL) return "[null error string]";
  if (lastMexListenerLid >= 0)
  {
    if (nerrors < 0) nerrors = GetErrorListSize();
    srcid = GetCompletionSource(lastMexListenerLid);
    if (srcid < 0 || srcid > 1) srcid = 0;
    if (cc < 0) cc = (short)(-cc);
    cc = (short)(cc & 0x0fff);
    if (cc > nerrors) sprintf(s,"%s: (ERR %d undefined)",errsrc[srcid],cc);
    else sprintf(s,"%s: %.32s",errsrc[srcid],cc2str(cc));
    strncpy(errstr,s,56);
    return errstr;
  }
  return GetLastLinkError(cc,errstr);
}
EXPORT_B int getPropertyInfoFromListener(char *dev, char *prp, int *siz, int *fmt)
{
  return getPropertyInfoFromListenerEx(dev,prp,siz,fmt,NULL);
}
EXPORT_B int getPropertyInfoFromListenerEx(char *dev, char *prp, int *siz, int *fmt, char *tag)
{
  mexListener *ml;
  int cc = un_allocated;
  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
  if ((ml=getListener(dev,prp,NULL,NULL,0)) != NULL)
  {
    if (siz != NULL) *siz = ml->dout.dArrayLength;
    if (fmt != NULL) *fmt = ml->dout.dFormat;
    if (tag != NULL) strncpy(tag,ml->dout.dTag,TAG_NAME_SIZE);
    ccerr(0);
  }
err:
  ReleaseLsnMutex(hLsnTblMutex);
  return cc;
}
int isGlobal(char *devName)
{
  char devNameBuffer[128], *ds = devNameBuffer, *c;
  strncpy(devNameBuffer,devName,128);
  if (devNameBuffer[0] == '/') 
  {
    if ((ds=strchr(&devNameBuffer[1],'/')) == NULL) return 0;
    ds++;
    if ((c=strchr(ds,'/')) != NULL) *c = 0;
    if ((c=strchr(ds,'[')) != NULL) *c = 0;
  }
  
  if (!stricmp(ds,"GLOBQRY")) return -1;
  if (!stricmp(ds,GLOBALSRV)) return -1;
  if (!stricmp(ds,"CYCLER")) return -1;
  return 0;
}
int isMetaCall(char *devName,char *devProperty)
{
  if (strstr(devName,"HISTORY") != NULL) return -1;
  if (strstr(devProperty,".HIST") != NULL) return -1;
  if (strstr(devProperty, ".HST") != NULL) return -1;
  if (strstr(devProperty,".EGU") != NULL) return -1;
  if (strstr(devProperty,".NAM") != NULL) return -1;
  if (strstr(devProperty,".DESC") != NULL) return -1;
  return 0;
}
int isQueryProperty(char *devProperty)
{
  if (strstr(devProperty,"PROPS") != NULL) return -1;
  if (strstr(devProperty,"PROPERTIES") != NULL) return -1;
  if (strstr(devProperty,"DEVICES") != NULL) return -1;
  if (strstr(devProperty,"TAGS") != NULL) return -1;
  if (strstr(devProperty,"CONTEXTS") != NULL) return -1;
  if (strstr(devProperty,"SUBSYSTEMS") != NULL) return -1;
  if (strstr(devProperty,"USERS") != NULL) return -1;
  if (strstr(devProperty,"IPNETS") != NULL) return -1;
  if (strstr(devProperty,"CLIENTS") != NULL) return -1;
  if (strstr(devProperty,"CONTRACTS") != NULL) return -1;
  if (strstr(devProperty,"HISTORIES") != NULL) return -1;
  if (strstr(devProperty,"LOGFILE") != NULL) return -1;
  if (strstr(devProperty,"LOGDEPTH") != NULL) return -1;
  if (strstr(devProperty,"STRUCTFORMAT") != NULL) return -1;
  if (strncmp(devProperty,"SRV",3) == 0) return -1;
  if (strstr(devProperty,"APPDATE") != NULL) return -1;
  if (strstr(devProperty,"APPVERSION") != NULL) return -1;
  if (strstr(devProperty,"ALARMS") != NULL) return -1;
  if (strstr(devProperty,"ALMDEFS") != NULL) return -1;
  return 0;
}
#endif /* NO_DOXYGEN */
EXPORT_B int setDefaultListenerMode(int mode)
{
  return SetDefaultListenerMode(mode);
}
EXPORT_B int SetDefaultListenerMode(int mode)
{
  if (mode > 0) defaultPollingMode = mode;
  return defaultPollingMode;
}

EXPORT_B int alsnStopLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din)
{
  int cc;
  mexListener *ml;
  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
  if ((ml=getListener(devName,devProperty,dout,din,0)) == NULL) ccerr(not_running);
  cc = stopListener(ml);
err:
  ReleaseLsnMutex(hLsnTblMutex);
  return cc;
}
EXPORT_B void alsnStopAllLinks(void)
{
  stopAllListeners();
}
EXPORT_B int alsnSetSubscriptionRenewalThreshold(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int thresholdInPercent)
{
  int cc;
  mexListener *ml;
  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
  if ((ml=getListener(devName,devProperty,dout,din,0)) == NULL) ccerr(not_running);
  cc = SetSubscriptionRenewalThreshold(ml->lid,thresholdInPercent);
err:
  ReleaseLsnMutex(hLsnTblMutex);
  return cc;
}

/* alsnExecLinkEx() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 * \param dTimeStamp will be filled with the data time stamp corresponding to the requested data.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see DTYPE
 */
EXPORT_B int alsnExecLinkEx(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long timeout,double *dTimeStamp)
{
  int cc=0,retry=2,len;
  DTYPE lcldin, *dinptr;

  static int initialized = FALSE;
  if (!initialized) 
  {
    initialized = TRUE;
    initLsnMutex();
    initSettings();
  }

# ifdef _USRDLL
  if (!gBufSrvInitialized) TineStartCycler();
# endif
  if (din != NULL) 
  {
    dinptr = din;
  }
  else
  {
    memset(&lcldin,0,sizeof(DTYPE));
    lcldin.dFormat = CF_NULL;
    dinptr = &lcldin;
  }
  if (isGlobal(devName))
  { /* n.b. This 'cut-to-the-chase' assumes CF_FLOAT with some CF_NAME64 exceptions */
    if (dout == NULL) return argument_list_error;
    if (!strcmp(devProperty,"NPROPS") || !strcmp(devProperty,"NPROPERTIES"))
    {
      dout->data.sptr[0] = 1;
      return 0;
    }
    if (!strcmp(devProperty,"PROPS") || !strcmp(devProperty,"PROPERTIES"))
    {
      char s[64];
      int bfmt = BFMT(CF_FLOAT);
      strncpy(s,dinptr->data.cptr,64); strupr(s);
      if (strstr(s,"OPTIC") || strstr(s,"STATE") || strstr(s,"TEXT") || strstr(s,"FILE"))
      {
        bfmt = BFMT(CF_NAME64);
      }
      if (!strcmp(dout->dTag,"PRPQSr4"))
      {
        PrpQueryStruct pqs;
        memset(&pqs,0,sizeof(PrpQueryStruct));
        strncpy(pqs.prpName,dinptr->data.cptr,PROPERTY_NAME_SIZE);
        pqs.prpSize = 1;
        pqs.prpFormat = bfmt;
        memcpy(dout->data.vptr,&pqs,sizeof(PrpQueryStruct));
      }
      else if (!strcmp(dout->dTag,"PQS"))
      {
        PropertyQueryStruct svrProp;
        if ((len=dinptr->dArrayLength) > PROPERTY_NAME_SHORTSIZE) len = PROPERTY_NAME_SHORTSIZE; 
        strncpy(svrProp.prpName,dinptr->data.cptr,len);
        svrProp.prpSize = 1;
        svrProp.prpFormat = bfmt;
        strncpy(s,dinptr->data.cptr,len); strupr(s);
        memcpy(dout->data.vptr,&svrProp,sizeof(PropertyQueryStruct));
      }
      return 0;
    }
    /* a real global is handled within getDataFromListener */
  }
  if (timeout < 1) 
  { /* the old way of forcing synchronous calls */
    access |= CA_SYNC;
    timeout = 1000;
  }
  if (timeout < 10) 
  { /* thinking in seconds */
    timeout *= 1000;
  }
  if (timeout < 100) 
  { /* oh no you don't ! */
    timeout = 100;
  }
  if (!(access&CA_WRITE) && !(access&CA_SYNC) && !isQueryProperty(devProperty) && 
      /* !isAdjFmt(dout->dFormat) && !isAdjFmt(dinptr->dFormat) && */
      !isMetaCall(devName,devProperty))
  {
    cc = getDataFromListener(devName,devProperty,dout,dinptr,dTimeStamp,(int)timeout,access);
    if (cc != use_stream_transport) ccerr(cc);
  }
  while (retry--)
  {
    if (access & CA_SYNC) access &= ~(CA_SYNC);
    cc = ExecLinkEx(devName,devProperty,dout,dinptr,access,(UINT16)timeout);
    if (dTimeStamp != NULL) *dTimeStamp = getDataTimeStamp();
    lastMexListenerLid = -1;
    if (cc & CE_SENDDATA) cc = 0;
    if (cc == 0) break;
    if (timeout >= 1000) break;
  }
err:
  return cc;
}
/* alsnExecLink() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see DTYPE
 */
EXPORT_B int alsnExecLink(char *devName,char *devProperty,
                          DTYPE *dout,DTYPE *din,short access,long timeout)
{
  return alsnExecLinkEx(devName,devProperty,dout,din,access,timeout,NULL);
}

EXPORT_B short alsnCall(char *fullDeviceNameAndProperty,DTYPE *dout,DTYPE *din,short access,double *dTimeStamp)
{
  char *c;
  char sbuf[256],dev[132],prp[PROPERTY_NAME_SIZE+4];  // was static : why ?
  int cc, i, rate = 1000;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    strupr(c);
    if (strstr(c,"NETWORK") != NULL || strstr(c,"MCAST") != NULL || strstr(c,"MULTICAST") != NULL)
    {
      access |= CA_NETWORK;
    }
    else if (strstr(c,"CONNECT") != NULL) 
    {
      access |= CA_CONNECT;
    }
    if (strstr(c,"SYNC") != NULL) 
    {
      access |= CA_SYNC;
    }
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) ccerr(illegal_property);
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,PROPERTY_NAME_SIZE);  prp[PROPERTY_NAME_SIZE] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=(int)strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,132);
  for (i=(int)strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks
  
  cc = alsnExecLinkEx(dev,prp,dout,din,access,rate,dTimeStamp);
err:
  return (short)cc;
}
/**
 * \brief Executes a synchronous call within an asynchronous static listener
 * for any type of array input/output
 * 
 * Asynchronous/Synchronous data exchange with input and output data of the
 * data size, type, and tag specified.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param access is the requested data access (e.g. CA_READ and/or CA_WRITE).
 * \param putdatafmt is the input data format
 * \param putdata is a reference to the input data
 * \param putsiz is the input data size
 * \param puttag is the input data tag (relevant if the input data is a tagged structure or
 *  bitfield).
 * \param getdatafmt is the output data format
 * \param getdata is a reference to the buffer which is to receive the output data
 * \param getsiz is the output data size
 * \param gettag is the output data tag (relevant if the input data is a tagged structure or
 *  bitfield).
 * \param dTimeStamp is a reference to a double value to receive the data timestamp.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink()
 */
EXPORT_B short tputgetAnyArray(char *fullDeviceNameAndProperty, short access,
                               short putdatafmt, void *putdata, int putsiz, char *puttag,
                               short getdatafmt, void *getdata, int getsiz, char *gettag,
                               double *dTimeStamp)
{
  DTYPE dout, din;

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = putdatafmt;
  din.dArrayLength = putsiz;
  if (puttag != NULL) strncpy(din.dTag,puttag,TAG_NAME_SIZE);
  din.data.bptr = putdata;

  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = getdatafmt;
  dout.dArrayLength = getsiz;
  if (gettag != NULL) strncpy(dout.dTag,gettag,TAG_NAME_SIZE);
  dout.data.bptr = getdata;
  return alsnCall(fullDeviceNameAndProperty,&dout,&din,access,dTimeStamp);
}
/**
 * \brief Executes a synchronous call within an asynchronous static listener
 * for array input/output (extended call).
 * 
 * Asynchronous/Synchronous data exchange with input and output data of the
 * data size and type specified.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param access is the requested data access (e.g. CA_READ and/or CA_WRITE).
 * \param putdatafmt is the input data format
 * \param putdata is a reference to the input data
 * \param putsiz is the input data size
 * \param getdatafmt is the output data format
 * \param getdata is a reference to the buffer which is to receive the output data
 * \param getsiz is the output data size
 * \param dTimeStamp is a reference to a double value to receive the data timestamp.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink()
 */
EXPORT_B short tputgetArrayEx(char *fullDeviceNameAndProperty, short access,
                              short putdatafmt,void *putdata,int putsiz,
                              short getdatafmt, void *getdata, int getsiz,double *dTimeStamp)
{
  DTYPE dout, din;

  din.dFormat = putdatafmt;
  din.dArrayLength = putsiz;
  memset(din.dTag,0,TAG_NAME_SIZE);
  din.data.vptr = putdata;

  dout.dFormat = getdatafmt;
  dout.dArrayLength = getsiz;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = getdata;

  return alsnCall(fullDeviceNameAndProperty,&dout,&din,access,dTimeStamp);
}
EXPORT_B short tputgetArray(char *fullDeviceNameAndProperty, short access,
                                                short putdatafmt,void *putdata,int putsiz,
                                                short getdatafmt, void *getdata, int getsiz)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,access,putdatafmt,putdata,putsiz,getdatafmt,getdata,getsiz,NULL);
}
EXPORT_B short tputString(char *fullDeviceNameAndProperty,char *putdata)
{
  char *c,sbuf[256];
  int len = 0;

  if (fullDeviceNameAndProperty == NULL || putdata == NULL) return argument_list_error;
  if ((len=(int)strlen(putdata)) == 0) return argument_list_error;
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
  return tputgetArray(sbuf,CA_WRITE|CA_READ,CF_TEXT,putdata,len+1,CF_NULL,NULL,0);
}
EXPORT_B short tputStringArray(char *fullDeviceNameAndProperty,char *putdata, int buffersize)
{	
  char *c,sbuf[256];
  int len = 0;
  char *putdataaux;  
  int i = 0;
  int j = 0;
  int localindex = 0;
  int globalindex = 0;	
  int stringsize = 0;

  if (fullDeviceNameAndProperty == NULL || putdata == NULL || buffersize <=0) return argument_list_error;
  if ((len=(int)strlen(putdata)) == 0) return argument_list_error;
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
 
  //Reserves memory for the array to put
  putdataaux = calloc(buffersize,32); //reserves space for the data  
  
  for (i=0;i<buffersize;i++)
  {	//Move localindex the size of the string
    stringsize = 0;
    while (putdata[j] != '\n') 
    {
      stringsize++;
      j++;
    }
    memcpy(putdataaux+globalindex,putdata+localindex,MIN(stringsize,32));
    localindex = localindex + stringsize + 1;
    j++; //move also the \n chars
    globalindex += 32;
  }
  return tputgetArray(sbuf,CA_WRITE|CA_READ,CF_NAME32,putdataaux,buffersize,CF_NULL,NULL,0);    
}
EXPORT_B short tputDoubleArray(char *fullDeviceNameAndProperty,double *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_DOUBLE,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B short tputSingleArray(char *fullDeviceNameAndProperty,float *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_FLOAT,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B short tputLongArray(char *fullDeviceNameAndProperty,int *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_LONG,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B short tputShortArray(char *fullDeviceNameAndProperty,short *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_SHORT,putdata,putsiz,outfmt,outdata,outsiz);
}
EXPORT_B void tDetachNotifier(void (*nf)(int))
{
  mexListener *ml;
  int i;
  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    if (ml->nf == nf) ml->nf = NULL;
  }
}
EXPORT_B short tAttachNotifier(char *fullDeviceNameAndProperty, void (*nf)(int), int nid)
{
  char *c,sbuf[256],dev[132],prp[PROPERTY_NAME_SIZE+4];
  int i, cc = 0;
  mexListener *ml=NULL;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,PROPERTY_NAME_SIZE);  prp[PROPERTY_NAME_SIZE] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=(int)strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,132);
  for (i=(int)strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks

  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    if (strncmp(dev,ml->devName,132)) continue;
    if (strlen(prp) > 0 && strncmp(prp,ml->devProperty,PROPERTY_NAME_SIZE)) continue;
    break;
  }
  if (ml == NULL || i == nMexListeners) ccerr(link_timeout);
  ml->nf = nf;
  ml->nid = nid;
err:
  ReleaseLsnMutex(hLsnTblMutex);
  return (short)cc;
}
/* the extended call will start the listener if it needs to and if it can */
EXPORT_B short tAttachNotifierEx(char *fullDeviceNameAndProperty,short datafmt,int buffersiz,void (*nf)(int), int nid)
{
  char *c,sbuf[256],dev[132],prp[PROPERTY_NAME_SIZE+4];
  int i, accessflags = 0, access = CA_READ, rate = 1000, cc = 0;
  mexListener *ml=NULL;
  DTYPE dout;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    strupr(c);
    if (strstr(c,"NETWORK") != NULL || strstr(c,"MCAST") != NULL || strstr(c,"MULTICAST") != NULL)
    {
      access |= CA_NETWORK;
    }
    else if (strstr(c,"CONNECT") != NULL)
    {
      access |= CA_CONNECT;
    }
    if (strstr(c,"SYNC") != NULL) access |= CA_SYNC;
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,PROPERTY_NAME_SIZE);  prp[PROPERTY_NAME_SIZE] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=(int)strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,132);
  for (i=(int)strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks

  if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
  for (i=0; i<nMexListeners; i++)
  {
    if ((ml=MexListenerTable[i]) == NULL) continue;
    if (strncmp(dev,ml->devName,132)) continue;
    if (strlen(prp) > 0 && strncmp(prp,ml->devProperty,PROPERTY_NAME_SIZE)) continue;
    break;
  }
  if (ml == NULL || i == nMexListeners) 
  { /* not yet in the table */
    ReleaseLsnMutex(hLsnTblMutex); /* release the mutex */
    memset(&dout,0,sizeof(DTYPE));
    dout.dFormat = datafmt;
    dout.dArrayLength = buffersiz;
    if (access & CA_NETWORK) accessflags = CM_NETWORK;
    if (access & CA_CONNECT) accessflags = CM_CONNECT;
    ml = startListener(dev,prp,&dout,NULL,rate,accessflags);
    if (ml == NULL) return link_timeout;
    if (WaitForSystemMutex(hLsnTblMutex,-1) != 0) return mutex_error;
    i = ml->mlid;
    MexListenerTable[i] = ml;
    if (i == nMexListeners) nMexListeners++;
  }
  ml->nf = nf;
  ml->nid = nid;
  ReleaseLsnMutex(hLsnTblMutex);
  return (short)cc;
}
EXPORT_B short tgetArray(char *fullDeviceNameAndProperty,short datafmt,void *dataout,int buffersiz)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_NULL,NULL,0,datafmt,dataout,buffersiz,NULL);
}
/* tgetSingleArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of Single (float) data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the float array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the float array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink()
 */
EXPORT_B short tgetSingleArray(char *fullDeviceNameAndProperty,float *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_FLOAT,dataout,buffersiz);
}
EXPORT_B short tputgetSingleArray(char *fullDeviceNameAndProperty,float *datain,int insize,float *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_FLOAT,datain,insize,CF_FLOAT,dataout,outsize,datatime);
}
/* tgetDoubleArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of double data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the double array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink()
 */
EXPORT_B short tgetDoubleArray(char *fullDeviceNameAndProperty,double *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_DOUBLE,dataout,buffersiz);
}
EXPORT_B short tputgetDoubleArray(char *fullDeviceNameAndProperty,double *datain,int insize,double *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_DOUBLE,datain,insize,CF_DOUBLE,dataout,outsize,datatime);
}
/* tgetLongArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of long data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the long integer array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink().
 */
EXPORT_B short tgetLongArray(char *fullDeviceNameAndProperty,long *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_LONG,dataout,buffersiz);
}
EXPORT_B short tputgetLongArray(char *fullDeviceNameAndProperty,long *datain,int insize,long *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_LONG,datain,insize,CF_LONG,dataout,outsize,datatime);
}
/* tgetShortArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of short integer data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the short integer array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink()
 */
EXPORT_B short tgetShortArray(char *fullDeviceNameAndProperty,short *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_SHORT,dataout,buffersiz);
}
EXPORT_B short tputgetShortArray(char *fullDeviceNameAndProperty,short *datain,int insize,short *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_SHORT,datain,insize,CF_SHORT,dataout,outsize,datatime);
}
/* tgetStringArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Synchronous data exchange. This is a convenience
 * routine which accesses ExecLink() explicitly for an array of NAME32 data type
 * without data input.  This is useful for retrieving a list of properties or devices from a 
 * server for instance.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]'.
 * For example: "/HERA/BPM/WL167[PROPERTIES]".
 * \param dataout is a reference to the NAME32 array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the NAME32 array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \see alsnExecLink()
 */
EXPORT_B short tgetStringArray(char *fullDeviceNameAndProperty,char *dataout,int buffersiz)
{
  int cc, len, i, j;
  char *c,sbuf[256];

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;

  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tgetArray(sbuf,CF_NAME32,c,buffersiz/32)) != 0) return (short)cc;

  for (j=0, i=0; i<buffersiz && j<buffersiz; j += 32)
  {
    len = MIN((int)strlen(&c[j]),32);
    strncpy(&dataout[i],&c[j],len);
    i += len;
    dataout[i++] = '\n';
  }
  return 0;
}
EXPORT_B short tgetString(char *fullDeviceNameAndProperty,char *dataout,int buffersiz)
{
  int cc;
  char *c,sbuf[256];

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;

  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tgetArray(sbuf,CF_TEXT,c,buffersiz)) != 0) return (short)cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
EXPORT_B short tputgetString(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz)
{
  int cc, inputlength = 0;
  char *c,sbuf[256],*cin = NULL;

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;
  if (datain != NULL) 
  {
    cin = datain;
    inputlength = (int)strlen(cin);
  }
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tputgetArray(sbuf,CA_READ,CF_TEXT,cin,inputlength,CF_TEXT,c,buffersiz)) != 0) return (short)cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
EXPORT_B short tputgetStringEx(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz,double *datatime)
{
  int cc, inputlength = 0;
  char *c,sbuf[256],*cin = NULL;

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;
  if (datain != NULL) 
  {
    cin = datain;
    inputlength = (int)strlen(cin);
  }
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tputgetArrayEx(sbuf,CA_READ,CF_TEXT,cin,inputlength,CF_TEXT,c,buffersiz,datatime)) != 0) return (short)cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
