/*
 * File mkhstfiles.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
int tgetEx(char *dev,char *prp,int fmt,int siz,int tmr);
int tgetFromRepeater(char *dev,char *prp,int fmt_,int siz_);
void fixFecRepository(void);
void fixHistoryRepository(void);

char *usageTxt[] = {
  "\nConfigures 'standard' local history random access files according to contents of history.csv\n",
  "mkhstfiles [/m=<eqm name> /s=<server name> /p=<history home> /l=<legacy home> /b=<spillover buffer size>] /t=<monthly> /x=<true>\n",
  "e.g. mkhstfiles /m=BPMEQM /l=x:\\histories\\BPM\n",
  "-will look for history.csv in the BPMEQM subdirectory",
  "of the FEC_HOME area",
  "-will assume that HISTORY_HOME is given by an environment",
  "variable or is a HISTORY directory adjacent to FEC_HOME",
  "-will search for legacy history files in x:\\histories\\BPM",
  "and convert the contents of any files found\n",
  "e.g. mkhstfiles /s=BPM\n",
  "-will look for history.csv in the FEC_HOME area and attempt to",
  "match the contents of the EXPORT_NAME column with the server 'BPM'",
  "-will assume that HISTORY_HOME is given by an environment variable",
  "or is a HISTORY directory adjacent to FEC_HOME",
  "-will assume that any legacy files are located in the same directory",
  "or in a 'SAVED' subdirectory of HISTORY_HOME\n",
  "use the '-x' switch to examine the activity of mkhstfiles without",
  "actually creating the standard file set",
  "e.g. mkhstfiles /m=BPMEQM /l=x:\\histories\\BPM /x=true",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))
extern int _stdFileSetProbe_;
int main(int argc, char *argv[])
{
  int i,sz;
  char *c,scratch[128],eqm[8],srv[64],path[132],spath[132];

  if (argc < 2) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }

  memset(eqm,0,8); memset(srv,0,64);
  memset(path,0,132); memset(spath,0,132);
  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'm':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(eqm,c,8);
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(srv,c,32);
        }
        continue;
      case 'p':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(path,c,128);
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(spath,c,128);
        }
        continue;
      case 'b':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if ((sz=atoi(scratch)) >= 0) SetStandardSpillOverSize(sz);
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if (stricmp(scratch,"monthly") == 0) useMonthlyHistoryFiles = TRUE;
        }
        continue;
      case 'x':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if (stricmp(scratch,"true") == 0) _stdFileSetProbe_ = TRUE;
        }
        continue;
      default:
        continue;
    }
  }
  gListenForGlobals = gListenForMCasts = FALSE;
  nofeclog = TRUE;
  gIsRunningAsServer = TRUE;
  fixFecRepository();
  fixHistoryRepository();
  if (strlen(path) > 0) SetHistoryFilesRepository(path);
  if (strlen(spath) > 0) SetHistoryStaticFilesRepository(spath);
  SetUseStandardHistoryFiles(TRUE);
  if (strlen(eqm) > 0) 
  {
    if (LoadHistoryInformation(HISTORY_FILENAME,NULL,eqm) != 0 &&
        LoadHistoryInformation(HISTORY_MANIFEST,NULL,eqm) != 0)
      printf("no history configuration found!\n");
  }
  else if (strlen(srv) > 0) 
  {
    if (LoadHistoryInformation(HISTORY_FILENAME,srv,NULL) != 0 &&
        LoadHistoryInformation(HISTORY_MANIFEST,srv,NULL) != 0)
      printf("no history configuration found!\n");
  }
# ifdef WIN32
  if (!_stdFileSetProbe_)
  {
    printf("\nyou may wish to run the 'contig' utility in order to ensure that all\n");
    printf("standard files are seen as continguous by the file system\n");
  }
# endif
  printf("\n");
  return 0;
}
