/*
 * File prolog.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef PROLOG_H_
#define PROLOG_H_

#include "project.def"
#include "tinedefs.h"
#include "errors.h"
#include "swaplib.h"
#include "tinetype.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \brief Allocate memory for an array. Synonym for calloc() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param nr Number of elements.
 * \param sz Size of an element in bytes.
 */
#ifndef SystemCalloc
# define SystemCalloc(nr,sz) calloc(nr,sz)
#endif

/**
 * \brief Free allocated memory. Synonym for free() under normal circumstances.
 *
 * \param ptr pointer to memory area to be free()ed.
 */
#ifndef SystemFree
# define SystemFree(ptr) if ((ptr) != NULL) free(ptr), (ptr) = NULL
#endif

#ifndef FEC_TIMEOUT
#define FEC_TIMEOUT      1   /* seconds */
#endif
#ifndef SERVER_TIMEOUT
#define SERVER_TIMEOUT   1   /* seconds */
#endif
#ifndef PORTOFFSET
# define PORTOFFSET 0
#endif
#ifndef WATCHDOG
# define WATCHDOG 0
#endif

#define UNUSED(x) (void)(x)

/* OS-specific definitions ... */
#if defined(MACOS) /* ---------------- MAC OS ------------ */
#  define UNIX     /*       default to UNIX settings       */
#  define FREEBSD
#  define OS_TAG "MACOS"
#  define s6_addr16 __u6_addr.__u6_addr16
#  define s6_addr32 __u6_addr.__u6_addr32
#endif
#if defined(DOS)  /* ------------------ DOS -------------- */
#  define OS_TAG "DOS"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define SERVER_POLLING_MODE -1
#  define millisleep(ms) delay(ms)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#  define RDATE
#  define FEC_SYNCHRONIZE_CLOCK TRUE
   typedef int _socklen_t;
   typedef int socklen_t;
#elif defined(UNIX) /* ---------------- UNIX (any variety) --- */
#  if defined(__GNUC__) && __GNUC__ >= 7
#  pragma GCC diagnostic ignored "-Wformat-truncation"
#  pragma GCC diagnostic ignored "-Wimplicit-fallthrough"
#  pragma GCC diagnostic ignored "-Wstringop-truncation"
#  pragma GCC diagnostic ignored "-Wcast-function-type"
#  endif
#  if !defined(USE_IPV6) && !defined(USE_IPV4)
#    define USE_IPV6
#  endif
#  ifndef OS_TAG
#    define OS_TAG "UNIX"
#  endif
#  define FS_DELIMITER '/'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  ifndef _REENTRANT
#    define _REENTRANT
#  endif
#  define SERVER_POLLING_MODE 0
#  ifdef SOLARIS
#    define s6_addr32 _S6_un._S6_u32
#    ifndef isinf
#      include <ieeefp.h>
#      define isinf(x) (!finite((x)) && (x)==(x))
#    endif   
#  endif
#  define _socklen_t socklen_t
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(char *)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
#  define MIN_WORK_SIZE 0x10000
#  if !defined(SINT64) && !defined(NO_SINT64)
#    ifdef OS_64BIT
#      define SINT64 long
#      define UINT64 unsigned long
#    else
#      define SINT64 long long
#      define UINT64 unsigned long long
#    endif
#  endif
#elif defined(VXWORKS)  /* ------------ VxWorks ----------- */
#  define OS_TAG "VXWORKS"
#  define FS_DELIMITER '/'
#  define SCAN_FOR_NAME_LISTS FALSE
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  ifndef HAVE_POLL
#    define NEED_POLL
#  endif
#  define SERVER_POLLING_MODE 0
   typedef int _socklen_t; /* and we have WindRiver to thank for this absurdity */
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(int)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
   void *alloca(size_t size);
#  define MIN_WORK_SIZE 2000
#  define RDATE
#  define NO_SERVICE_PIPE
#  ifndef VXWTSK_FLAGS
#    define VXWTSK_FLAGS VX_FP_TASK
#  endif
#  ifndef VXWTSK_STACK
#    define VXWTSK_STACK 10000
#  endif
#  ifndef VXWTSK_PRIORITY
#    define VXWTSK_PRIORITY 100
#  endif
#  define NO_IPV6_V6ONLY
#elif defined(VMS)   /* -----------------  VMS  ------------ */
#  define OS_TAG "VMS"
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_POLLING_MODE 0
#  ifndef HAVE_POLL
#    define NEED_POLL
#  endif
   typedef int _socklen_t; 
   typedef int socklen_t; 
#  define soioctl(fd,REQ,buf) 0
#  define unlink(nm) remove(nm)
#  define millisleep(ms) ;
#  define soclose(i) close(i)
#  define shutdown(s,t) close(s)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#elif defined(WIN32)  /* ---------------  Win32 ------------ */
#  define OS_TAG "WIN32"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  define soerrno GetErrno()
#  define SERVER_POLLING_MODE 0
#  define soioctl(fd,REQ,buf) ioctlsocket(fd,REQ,(u_long *)buf)
#  define millisleep(ms) Sleep(ms)
#  define soclose(i) closesocket(i)
#  ifdef NEED_TIMEZONE
   struct timezone { int dummy; };
#  endif
#  define MIN_WORK_SIZE 0x10000
#  ifndef USE_IPV6
   typedef int _socklen_t;
   typedef int socklen_t;
#  else
#  define _socklen_t socklen_t
#  endif
#  ifdef NEED_STRTOLL
     SINT64 _strtoi64(const char *nptr, char **endptr, int base);
#  endif
   /* a few common functions from the unix world ... */
#  if !defined(strtoll) && !defined(MINGW)
#    define strtoll _strtoi64
#  endif
#  ifndef bzero
#    define bzero(a,l) memset(a,0,l)
#  endif
#  ifndef bcmp
#    define bcmp(a,b,l) memcmp(a,b,l)
#  endif
#  if _MSC_VER < 1900 
#  ifndef snprintf
#    define snprintf _snprintf
#  endif
#  if !defined(vsnprintf) && !defined(MINGW)
#    define vsnprintf _vsnprintf
#  endif
#  endif
#  if !defined(trunc) && defined(NEED_TRUNC)
#    define trunc(d) ((double)((int)d))
#  endif
#  ifndef strcasestr
#    define strcasestr stristr
#  endif
#  ifdef NEED_ISNAN
#    include <float.h>
#    define isinf(v) (!_finite(v))
#    define isnan(v) _isnan(v)
#  endif
#elif defined(NIOS)     /* ------------ Altera NIOS --------- */
#  define OS_TAG "NIOS"
#  define FS_DELIMITER '/'
#  define FS_RDONLY
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define ForceStructAlignment -1
#  define SERVER_POLLING_MODE -1
#  ifndef HAVE_POLL
#    define NEED_POLL
#  endif
#  define millisleep(ms) sleep(ms)
   typedef int _socklen_t;
   typedef int socklen_t;
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(u_long *)buf)
#  define soread(fd,buf,nbyte)  recv(fd,(char *)buf,nbyte,0)
#  define sowrite(fd,buf,nbyte) send(fd,(char *)buf,nbyte,0)
   struct timezone { int dummy; };
#  define MIN_WORK_SIZE 2000
#  define RDATE
#  define FEC_SYNCHRONIZE_CLOCK TRUE
#  ifndef INCLUDE_IN_LOGFILE
#    define INCLUDE_IN_LOGFILE FALSE
#  endif
#else                  /* ------------  ????? --------------- */ 
#  ifndef OS_TAG
#    define OS_TAG "UNKNOWN"
#  endif
   struct timezone { int dummy; };
   typedef int _socklen_t;
   typedef int socklen_t;
#  define MIN_WORK_SIZE 2000
#endif

#ifdef NEED_BOOL
   typedef unsigned char _Bool;
#endif

#if defined(USE_IPV6) && !defined(NO_IPV6_V6ONLY)
# define FIX_IPV6_V6ONLY
#endif
#ifndef INCLUDE_IN_LOGFILE
# define INCLUDE_IN_LOGFILE TRUE
#endif

#ifdef FS_RDONLY
# define NO_LTS
#endif

#ifndef SCAN_FOR_DEVICE_LISTS
# define SCAN_FOR_DEVICE_LISTS TRUE
#endif

#ifndef SCAN_FOR_NAME_LISTS
# define SCAN_FOR_NAME_LISTS TRUE
#endif

#ifndef ForceStructAlignment
# define ForceStructAlignment 0
#endif

#ifndef soerrno
#  define soerrno errno
#endif

#ifndef SERVER_POLLING_MODE
#  define SERVER_POLLING_MODE 0
#endif

/* 'us-or-them' definitions : */
#ifdef DOS   
#  include "dos.h"
#  define malloc_t void *
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT; }
#  define srvWorkArea     WorkArea
#  define srvWorkAreaSize WorkAreaSize
#  define clnWorkArea     WorkArea
#  define clnWorkAreaSize WorkAreaSize
#  define whoami(c) ;
#else
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT+size/5000; }
#  ifndef soread
#    define soread(fd,buf,nbyte)  read(fd,(char *)buf,nbyte)
#  endif
#  ifndef sowrite
#    define sowrite(fd,buf,nbyte) write(fd,(char *)buf,nbyte)
#  endif
#  if !defined(SUNOS)
#    define malloc_t size_t
#  endif
#  if !defined(WIN32)
   char *strrev(char *s);
#  endif
#  if !defined(WIN32) && !defined(NIOS)
   int stricmp(const char *s1,const  char *s2); /* from the DOS world */
   int strnicmp(const char *s1,const  char *s2,int len); /* from the DOS world */
   char *strupr(char *s);
   char *strlwr(char *s);
#  endif
   extern BYTE *srvWorkArea;
   extern UINT32 srvWorkAreaSize;
   extern BYTE *clnWorkArea;
   extern UINT32 clnWorkAreaSize;
   char *whoami(char *user);
   void ClientDelay(int sec);  /* delay(0) never returns */
#endif                         /* DOS: definitions */

#ifndef VXWORKS
#  define inet_ntoa_b(n,b) strcpy(b,inet_ntoa(n))
#else
# if (_WRS_VXWORKS_MAJOR < 6)
  struct hostent
  {
    char *h_name;
    char **h_aliases;
    int  h_addrtype;
    int  h_length;
    char *h_addr;
  };
# elif (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MINOR > 8)
# define HAVE_GETTIMEOFDAY
# endif
#endif

#if !defined(UNIX) && !defined(HAVE_GETTIMEOFDAY)
  TINE_EXPORT struct timeval *gettimeofday(struct timeval *t,struct timezone *tz);
#endif

#ifndef WIN32
  typedef int SOCKET;
#endif

#ifdef MULTITHREADED
# if defined(WIN32)
#   define WINTHREAD
# elif defined (UNIX)
#   define PTHREAD
# endif
# include "threader.h"
# define NUM_DEFAULT_DATAOUT_BUFFERS 2
#else
# define NUM_DEFAULT_DATAOUT_BUFFERS 1
#endif

#ifndef CLNHDR_STRUCT_INITIALIZER
# define CLNHDR_STRUCT_INITIALIZER { 0 }
#endif

void closeIPClients(void);

#ifndef IPXNW
typedef struct  /* an important entry from nwipxspx.h */
{
   BYTE network[4];
   BYTE node[6];
   BYTE socket[2];
} IPXAddress;
#endif

#define IsMemberIPControlNets(e,a) IsMemberIPNets((e)->aclLst.nNets,(e)->aclLst.nets,a)
#define IsMemberIPBcastNets(a) IsMemberIPNets(sizIpBcstLst,IpBcstLst,a)

#if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
# define doShmConsumer(n) DoClientEvents(n)
# define AcceptShm(n) DoServerEvents(n)
#else
# define doShmConsumer(n) ;
# define AcceptShm(n) ;
#endif
#if defined(IPXNW) || defined(IPXBSD)
#  include "srvipx.h"
#  define IsMemberIPXControlNets(addr) IsMemberIPXNets(NrOfIPXNets,IPXNetList,addr)
#  define IsMemberIPXBcastNets(addr) IsMemberIPXNets(NrOfIPXBcastNets,IPXBcastList,addr)
#  define sendGCast(buf,size)  sendIPgcast(buf,size); sendIPXgcast(buf,size)
#  define AcceptRequest(tv) { AcceptShm(100); AcceptIP(tv); AcceptIPX(tv); }
#  define CleanupTransfers(tv) { CleanupIPX(tv); CleanupIP(tv); CleanupStrandedDatagrams();}
#  define ntoa(c,b) (\
     (c)->inetProtocol == IPX ?\
     strcpy(b,ipxnet_ntoa((c)->adr.ipx.IpxNode)) :\
     inet_ntop(AF_INET,getInAddr(&(c)->adr.ip),b,16))
#  define tCall(i,buf,sz,to) (ipLoaded ? tUDPCall(i,buf,sz,to) : tIPXCall(i,buf,sz,to))
#  define IsMemberControlNets(e,c) ((c)->inetProtocol == IPX ?\
				   IsMemberIPXControlNets(&(c)->adr.ipx) :\
				   IsMemberIPControlNets((e),&(c)->adr.ip))
#  define IsMemberBcastNets(c)   ((c)->inetProtocol == IPX ?\
				   IsMemberIPXBcastNets(&(c)->adr.ipx) :\
				   IsMemberIPBcastNets(&(c)->adr.ip)    )
   void closeIPXClients(void);
   int initIPXClient(int socket);
   int sendSPXData(WORD ID,int socket);
#  define ShutdownClients(i) {\
     closeIPXClients();\
     closeIPClients();\
   }
#  define CheckConsumerData(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle++;}\
     doShmConsumer(gBurstLimit);\
     doIPXConsumer();\
     doIPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#else
#  define sendGCast(buf,size) sendIPgcast(buf,size)
#  define AcceptRequest(tv) { AcceptShm(100); AcceptIP(tv); }
#  define CleanupTransfers(tv) { CleanupIP(tv); CleanupStrandedDatagrams(); }
#  define ntoa(c,b) inet_ntop(AF_INET,getInAddr(&(c)->adr.ip),b,16)
#  define tCall(i,buf,sz,to) tUDPCall(i,buf,sz,to)
#  define IsMemberControlNets(e,c) IsMemberIPControlNets((e),&(c)->adr.ip)
#  define IsMemberBcastNets(c) IsMemberIPBcastNets(&(c)->adr.ip)
#  define sendSPXData(i,j) ipx_not_supported
#  define ShutdownClients(i) {\
	   closeIPClients();\
   }
#  define CheckConsumerData(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle++;}\
     doShmConsumer(gBurstLimit);\
     doIPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#endif         /* ! IPXNW || IPXBSD */

extern TINE_EXPORT int tineDebug;
extern int NGdebug;
extern int errorDebug;
extern int _sendToDbg;
int serverCycle(void);
int initTineResources(void);

#ifdef INTEL  /* Data Preparation MACROS: */

# define PrepSubData(pD,sz,f,dir) ;

# define PrepOutgoingSrvSettings(s,buf) {if ((void *)buf != (void *)s) memcpy(buf,s,SERVERSETTINGSSTRUCT_SIZE); }
# define PrepOutgoingAMS(amsstruct,buf,n) {if ((void *)buf != (void *)amsstruct) memcpy(buf,amsstruct,n*AMS_SIZE); }
# define PrepOutgoingAQS(a,buf) {if ((void *)buf != (void *)a) memcpy(buf,a,ACTIVITYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCTEX_SIZE); }
# define PrepOutgoingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*XPROPERTYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPRPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PRPQUERYSTRUCT_SIZE); }
# define PrepOutgoingCONQS(c) ;
# define PrepOutgoingConTblInfo(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*CONTBL_QUERY_SIZE); };
# define PrepOutgoingHRS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*HRS_SIZE); };
# define PrepIncomingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCT_SIZE); }
# define PrepIncomingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCTEX_SIZE); }
# define PrepIncomingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*XPROPERTYQUERYSTRUCT_SIZE); }
# define PrepIncomingPRPPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PRPQUERYSTRUCT_SIZE); }
# define PrepIncomingHRS(rb,buf) memcpy(rb,buf,HRS_SIZE);

#else     /* non-INTEL world must SWAP ! */

# define PrepSubData(pD,sz,f,dir) {\
    SwapBytes((char *)pD,sz,LFMT(f),dir);\
}
# define PrepOutgoingSrvSettings(s,buf) {\
    _structSwap(ServerSettingsStructStruct,(unsigned char *)buf,0,NULL,(unsigned char *)s,LTOR);\
}
# define PrepOutgoingAMS(amsstruct,buf,n)  {\
    _systemStructSwap(AMSStructStruct,(unsigned char *)buf,(unsigned char *)amsstruct,LTOR,n,sizeof(AMS));\
}
# define PrepOutgoingAQS(a,buf) {\
    _structSwap(ActivityQueryStructStruct,(unsigned char *)buf,0,NULL,(unsigned char *)a,LTOR);\
}
# define PrepOutgoingPQS(p,buf,n) {\
    _systemStructSwap(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStruct));\
}
# define PrepOutgoingPQSX(p,buf,n) {\
    _systemStructSwap(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStructEx));\
}
# define PrepOutgoingXPQS(p,buf,n) {\
    _systemStructSwap(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(XPropertyQueryStruct));\
}
# define PrepOutgoingPRPQS(p,buf,n) {\
    _systemStructSwap(PrpQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PrpQueryStruct));\
}
# define PrepOutgoingCONQS(c) {\
    ISWAP(c.contract.EqmSizeIn);\
    ISWAP(c.contract.EqmSizeOut);\
    ISWAP(c.pollingInterval);\
    ISWAP(c.nclients);\
  }
# define PrepOutgoingConTblInfo(p,buf,n) {\
    _systemStructSwap(ConTblInfoStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,CONTBL_QUERY_SIZE);\
}
# define PrepOutgoingHRS(p,buf,n) {\
    _systemStructSwap(HRSStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(HRS));\
  }
# define PrepIncomingPQS(p,buf,n) {\
    _systemStructSwap(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStruct));\
  }
# define PrepIncomingPQSX(p,buf,n) {\
    _systemStructSwap(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStructEx));\
  }
# define PrepIncomingXPQS(p,buf,n) {\
    _systemStructSwap(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(XPropertyQueryStruct));\
  }
# define PrepIncomingPRPQS(p,buf,n) {\
    _systemStructSwap(PrpQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PrpQueryStruct));\
  }
# define PrepIncomingHRS(rb,buf) {\
    _systemStructSwap(HRSStructStruct,(unsigned char *)buf,(unsigned char *)rb,RTOL,1,sizeof(HRS));\
  }

#endif  /* INTEL (data prepartion MACROS) */

/* client-side macros : */
#define IncrementConnections(n) (n)++

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* PROLOG_H_ */
