/*
 * File pytine.cpp
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// pytine.cpp : Defines the exported functions for the DLL application.
//

#ifdef WIN32
#include "stdafx.h"
#endif
#ifdef _DEBUG
# define DBG_BUILD_WINDOWS // the python distribution on windows doesn't have debug libraries!
#endif
#ifdef DBG_BUILD_WINDOWS
# undef _DEBUG //Prevent linking debug build of python in windows ...
#endif
#define PY_SSIZE_T_CLEAN
#include <Python.h>
#ifdef DBG_BUILD_WINDOWS
# define _DEBUG 1
#endif

#undef TINE_DECORATED_CONSTANTS
#undef tine_decorated_constants
#if PY_MINOR_VERSION > 4
#ifndef HAVE_TIMEZONE
#define HAVE_TIMEZONE
#endif
#ifndef HAVE_GETTIMEOFDAY
#define HAVE_GETTIMEOFDAY
#endif
#endif
#include "tine.h"
#include "tbufsrv.h"
#include "threader.h"
#include "eventtools.h"
#include "hstlib.h"
//#include "revision.h"

#define PYTINE_VERSION "1.2.1"
static int array_to_python(PyObject **o, void *data, int size, int format, char *tag, int use_bytearray);
static int acqInfo(int size,int format,char *typestr,char *tag,char *srv);
static void *array_from_python(PyObject *o, int *size, int format, char *tag);
static char *python_to_string(PyObject *pyo);
static int array_length_from_python(PyObject *o);
static PyObject *makeStructDictionary(char *tag, BYTE *data);
static PyObject *pytine_getExcObject(int cc);
static PyObject *thrownTineException(PyObject *ex, int cc, const char *fcn, const char *arg0, const char *arg1, const char *arg2);
static int format_from_python(PyObject* o);
static void _datacallback_handler(int id, int cc, void *p);
static void _groupcallback_handler(int id, int cc, void *p);
static int tine_attach(const char *server, const char *property,
  DTYPE *dout, DTYPE *din, int access, int mode, int interval,
  void (*callback)(int, int, void *), void *p, int id);

struct callback
{
  PyObject *callable;
  void *data;
  size_t size;
  int format;
  int mode;
  int use_bytearray;
  char tag[TAG_NAME_SIZE];
};

static time_t parseTimeString(char *ts)
{
  char tsWorkArea[64];
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  strncpy( tsWorkArea, ts, 64);
  hr[0] = 0;
  tnow = localtime(&t);
  len = (int)strlen(tsWorkArea);
  strupr(tsWorkArea);
  for (i=0; i<len && isdigit(tsWorkArea[i]); i++);
  if (i == len) t = atoi(tsWorkArea);
  else if (!strcmp(tsWorkArea,"NOW")) t = time(NULL);
  else if ((c=strstr(tsWorkArea,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,tsWorkArea,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year,"_")) != NULL || (c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
	  if (tmts.tm_year < 100) tmts.tm_year += 2000;
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
static time_t parseDepthSting(char *ds)
{
  char *c=NULL;
  long depth = strtol(ds,&c,10);
  switch (tolower(*c))
  {
    case 'd': depth *= 86400; 
      break;
    case 'h': depth *= 3600; 
      break;
    case 'm':
      if (strnicmp(c,"min",3) == 0)
      {
        depth *= 60;
      }
      else if (strnicmp(c,"mon",3) == 0)
      {
        depth *= 86400 * 31;
      }
      break;
    case 'w':
      depth *= 86400 * 7;
      break;
    default:
    case 's': 
      break;
  }
  return (time_t)depth;
}
static int getArchivedData(char *dev,char *prp,time_t start,time_t *stop,int access,int sample,int index,int format,void *data,int *num,int timeout)
{
  DTYPE din,dout;
  UINT32 startstopArray[4];
  char *c, asrv[STATUS_SIZE],aprp[PROPERTY_NAME_SIZE];
  int cc, needtag=FALSE,isSnapshot=FALSE;
  if (num == NULL || *num <= 0) return buffer_too_small;
  strncpy(asrv,dev,STATUS_SIZE); 
  strncpy(aprp,prp,PROPERTY_NAME_SIZE);
  if (strstr(asrv,"/HISTORY") == NULL)
  { // going for the local history
    needtag = TRUE;
    if ((c=strrchr(aprp,'.')) != NULL)
    {
      if (stricmp(c,".HIST") == 0 || stricmp(c,".HST") == 0) needtag = FALSE;
    }
    if (needtag && strlen(aprp) < 58) strcat(aprp,".HIST");
  }
  if (start == *stop) isSnapshot = TRUE;
  DTYPEZERO(dout); DTYPEZERO(din);
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)*stop;
  startstopArray[2] = (UINT32)index;
  startstopArray[3] = (UINT32)sample;
  dout.dArrayLength = *num;
  dout.dFormat = format;
  dout.data.vptr = data;
  din.dArrayLength = isSnapshot ? 2 : 4;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;
  if ((cc=ExecLinkEx(asrv,aprp,&dout,&din,(short)access,timeout)) != 0) 
  {
    *num = 0;
  }
  else 
  {
    *num = (int)dout.dArrayLength;
    *stop = (UINT32)dout.dTimeStamp;
  }
  return cc;
}
/*---------------------------------------------------------------------
 * event archive stuff
 * 
 * yet missing:
 *   static int events_getArchiveTriggers(const char *context,NAME64 *triggers,int *triggerBufferSize)
 * 
 * duplicated from eventtools:
 *   static char *events_getArchiveComment(const char *context,const char *trigger, int eventNumber)
 *   static int events_sendComment(const char *context,const char *trigger, int eventNumber,const char *cmt)
 *   static int events_getArchivedPropertyInformation(const char *context,const char *fullname,PrpQueryStruct *pqs)
 *   static int events_getArchivedEventList(const char *context,const char *trigger,time_t start,time_t stop,int *evnlst,int *lstsiz)
 *   static void events_registerStructs(void)
 *   static int events_getArchivedEventData(const char *context,const char *trigger,
 *                              const char *devContext,const char *devServer,const char *devProperty,const char *devName,
 *                              int dataIndex, int eventNumber, DTYPE *dout) 
 *
 * pytine API functions:
 *   static PyObject* pytine_getEventArchiveComment(PyObject *self, PyObject *args, PyObject *kwargs) 
 *   static PyObject* pytine_setEventArchiveComment(PyObject *self, PyObject *args, PyObject *kwargs) 
 *   static PyObject* pytine_triggerEvent(PyObject *self, PyObject *args, PyObject *kwargs) 
 *   static PyObject* pytine_getEventList(PyObject *self, PyObject *args, PyObject *kwargs) 
 *   static PyObject* pytine_getArchivedEventData(PyObject *self, PyObject *args, PyObject *kwargs) 
 */
static int getEventIdFromPython(PyObject* o)
{
  int eventId = 0;
  int rc = -1;
# if PY_MAJOR_VERSION < 3
  if (PyInt_Check(o)) PyArg_ParseTuple(o, "i", &eventId);
# else
  if (PyLong_Check(o))
  {
    eventId = PyLong_AsLong(o);
  }
# endif
  else if (PyUnicode_Check(o))
  {
    char* ts = python_to_string(o);
    eventId = (int)parseTimeString(ts);
  }
  return eventId;
}
static PyObject* pytine_getEventArchiveComment(PyObject *self, PyObject *args, PyObject *kwargs)
{
  int eventId = 0;
  PyObject *evn = NULL;
  char *comment = (char*)"", *context = (char*)"", *trigger = (char*)"";
  static const char *kwlist[] = {
        "context",
        "event",
        "eventId",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ssO", (char**)kwlist,
                                   &context, &trigger,&evn)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.getEventArchiveComment(context='str',event='str',eventId=val)");    
  }
  eventId = getEventIdFromPython(evn);
  if (eventId <= 0) return PyErr_Format(PyExc_TypeError, "invalid event id object or value");
  comment = GetArchiveComment(context,trigger,eventId); // can't be NULL !
  char *sts = NULL, *scmt = NULL, *ucmt = NULL, *c=NULL;
  if ((c=stristr(comment, "[STATUS:")) != NULL) sts = c + 8;
  if ((c=stristr(comment, "[STATIC:")) != NULL) scmt = c + 8;
  if ((c=stristr(comment, "[USER:")) != NULL) ucmt = c + 6;
  if (sts != NULL) { if ((c = strchr(sts, ']')) != NULL) *c = 0; }  else sts = (char*)"NONE";
  if (scmt != NULL) { if ((c = strchr(scmt, ']')) != NULL) *c = 0; }  else scmt = (char*)"";
  if (ucmt != NULL) { if ((c = strchr(ucmt, ']')) != NULL) *c = 0; }  else ucmt = (char*)"";
  if (ucmt == NULL) ucmt = comment;
  PyObject *pdo = PyDict_New();
//  PyDict_SetItemString(pdo, "status", Py_BuildValue("s", sts));
//  PyDict_SetItemString(pdo, "static", Py_BuildValue("s", scmt));
//  PyDict_SetItemString(pdo, "user", Py_BuildValue("s", ucmt));
  PyDict_SetItemString(pdo, "status", PyUnicode_Decode(sts, strlen(sts), "latin-1", "ignore"));
  PyDict_SetItemString(pdo, "static", PyUnicode_Decode(scmt, strlen(scmt), "latin-1", "ignore"));
  PyDict_SetItemString(pdo, "user", PyUnicode_Decode(ucmt, strlen(ucmt), "latin-1", "ignore"));
  //printf("debug pytine_getEventArchiveComment: %s\n", comment);
  return Py_BuildValue("{s:s,s:s,s:i,s:O}","context",context,"event",trigger,"eventId",eventId,"comment",pdo);
}
/*
 * [STATUS:SAVE][STATIC:some descriptive comment][USER:]
 * [STATUS:DELETE][STATIC:some descriptive comment][USER:]
 * [STATUS:NONE][STATIC:some descriptive comment][USER:]
 */
static PyObject *pytine_setEventArchiveComment(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  PyObject *value = NULL, *evn = NULL;
  int tineResultCode = 0, eventId = 0;
  char *comment = (char*)"", *context = (char*)"", *trigger = (char*)"";
  static const char *kwlist[] = {
        "context",
        "event",
        "eventId",
        "comment",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ssOO", (char**)kwlist,
                                   &context, &trigger,&evn,&value)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.setEventArchiveComment(context='str',event='str',eventId=val,comment='str')");    
  }
  if (value == NULL) return PyErr_Format(PyExc_TypeError, "invalid comment object");
  eventId = getEventIdFromPython(evn);
  if (eventId <= 0) return PyErr_Format(PyExc_TypeError, "invalid event id object or value");

  char cmt[1032], *sts = NULL, *scmt = NULL, *ucmt = NULL, *s;
  if (PyDict_Check(value))
  {
    Py_ssize_t dp = 0;
    PyObject *key, *val;
    int p = 0;
    while (PyDict_Next(value, &dp, &key, &val))
    {
      s = python_to_string(key);
      if (s == NULL) continue;
      if (!stricmp(s, "STATUS")) sts = python_to_string(val);
      if (!stricmp(s, "STATIC")) scmt = python_to_string(val);
      if (!stricmp(s, "USER")) ucmt = python_to_string(val);
    }
    memset(cmt, 0, 1032);
    if (sts != NULL) snprintf(cmt, 16, "[STATUS:%s]", sts);
    p = (int)strlen(cmt);
    if (scmt != NULL) snprintf(&cmt[p], 256, "[STATIC:%s]",scmt);
    p = (int)strlen(cmt);
    if (ucmt != NULL) snprintf(&cmt[p], 775, "[USER:%s]", ucmt);
  }
  if (PyUnicode_Check(value))
  {
    s = python_to_string(value);
    strncpy(cmt, s, 1024); 
  }
  cmt[1024] = 0;
//  tineResultCode = SendComment(context,trigger,eventId,comment);
  tineResultCode = SendComment(context, trigger, eventId, cmt);
  return Py_BuildValue("i",tineResultCode);
}
static PyObject* pytine_triggerEvent(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  int tineResultCode = 0;
  char *context, *event;
  char name[132], *comment = (char*)"";
  int triggerLevel = 1, triggerTime = 0, rangeStart = 0, rangeStop = 0, rangeMax = -1, options = 0;
  static const char *kwlist[] = {
        "context",
        "event",
        "comment",
        "triggerLevel",
        "triggerTime",
        "rangeStart",
        "rangeStop",
        "rangeMax",
        "options",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ss|siiiiii", (char**)kwlist,
                                   &context, &event, &comment,&triggerLevel,&triggerTime,
                                   &rangeStart,&rangeStop,&rangeMax,&options)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.triggerEvent(context='str',event='str',[comment='str',triggerLevel=val,triggerTime=val,rangeStart=val,rangeStop=val,rangeMax=val,options=val])");    
  }
  snprintf(name, 128, "/%.32s/EVENTSTORE/%.64s", context, event);
  tineResultCode = SendEventTriggerEx(name,comment,triggerLevel,triggerTime,rangeStart,rangeStop,rangeMax,options);
  return Py_BuildValue("i",tineResultCode);
}
static PyObject* pytine_getEventList(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  int tineResultCode = 0;
  char *context = NULL, *trigger = NULL;
  char *startTimeString = NULL, *stopTimeString = NULL;
  int maxEventsToRetrieve = 1;
  time_t startTime, stopTime;
  int *listOfEvents;
  PyObject *value = NULL;

  static const char *kwlist[] = {
        "context",
        "event",
        "startTime",
        "stopTime",
        "limit",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ss|ssi", (char**)kwlist,
                                   &context,&trigger,&startTimeString,&stopTimeString,&maxEventsToRetrieve)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.getEventList(context='str',event='str',[startTime='str',stopTime='str',limit=val])");    
  }
  struct tm  *timeStruct;
  if (startTimeString != NULL) {
    startTime = parseTimeString(startTimeString);
  } else {
    startTime = time(NULL);
    timeStruct = localtime(&startTime);
    timeStruct->tm_mon--;
    startTime = mktime(timeStruct);
    /*startTime = 0;*/
  }
  if (stopTimeString != NULL) {
    stopTime = parseTimeString(stopTimeString);
  } else {
    stopTime = time(NULL);
  }
  if (maxEventsToRetrieve < 1) {
    maxEventsToRetrieve = 1;
  } else if (maxEventsToRetrieve > 1024) {
    maxEventsToRetrieve = 1024;
  }
  if ((listOfEvents=(int*)SystemCalloc(maxEventsToRetrieve,sizeof(int))) == NULL) return PyErr_NoMemory();
  tineResultCode = GetArchivedEventList(context,trigger,startTime,stopTime,listOfEvents,&maxEventsToRetrieve);
  if (tineResultCode) 
  {
    PyObject *ex = pytine_getExcObject(tineResultCode);
    if (ex != NULL) return thrownTineException(ex, tineResultCode, "getEventList", context, trigger, NULL);
  }
  if (tineResultCode == 0) array_to_python(&value, listOfEvents, maxEventsToRetrieve, CF_INT32, NULL, FALSE);
  double ts;
  char *tss;
  PyObject *values_o = PyList_New(maxEventsToRetrieve);
  PyObject *item;
  for (int i = 0; i < maxEventsToRetrieve; i++)
  {
    ts = (double)listOfEvents[i];
    tss = GetDataTimeString(ts, FALSE);
    item = Py_BuildValue("{s:i,s:s}", "id", listOfEvents[i], "time", tss);
    PyList_SetItem(values_o, i, item);
  }
  SystemFree(listOfEvents);
  PyObject *dtype = Py_BuildValue("{s:s,s:s,s:O}",
    "context", context, "event", trigger, "events", values_o); 
  Py_CLEAR(value);
  Py_CLEAR(values_o);
  return dtype;
}
static PyObject* pytine_getArchivedEventData(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  int tineResultCode = 0;
  char *context = NULL, *trigger = NULL, *formatString = NULL;
  char *eventContext = NULL, *eventServer = NULL, *eventDevice = NULL, *eventProperty = NULL;
  char *eventTarget = NULL;
  char fullEventPropertyAdress[1024], addrbuf[1024];
  int eventId = 0, size = 0, formatVal = 0;
  void *eventValueBuffer;
  PrpQueryStruct prpQueryStruct;
  DTYPE dout;
  PyObject *value = NULL, *evn=NULL;

  static const char *kwlist[] = {
        "context",
        "event",
        "eventId",
        "channel",
        "eventServer",
        "eventDevice",
        "eventProperty",
        "eventContext",
        "format",
        "size",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ssO|ssssssi", (char**)kwlist,
                                   &context,&trigger,&evn,&eventTarget,&eventServer,&eventDevice,
                                   &eventProperty,&eventContext,&formatString,&size)) 
  {
     return PyErr_Format(PyExc_SyntaxError, 
       "PyTine.getArchivedEventData(context='str',event='str',eventId=val[,channel='str',eventServer='str',eventDevice='str',eventProperty='str',eventContext='str',format='str',size=val])");    
  }
  eventId = getEventIdFromPython(evn);
  if (eventId <= 0) return PyErr_Format(PyExc_TypeError, "invalid event id object or value");
  if (eventTarget == NULL && (eventServer == NULL || eventDevice == NULL || eventProperty == NULL))
    return PyErr_Format(PyExc_TypeError, "the event channel either via eventTarget or eventServer, eventDevice, and eventProperty must be specified");
  if (eventTarget != NULL)
  {
    char *ctx = NULL, *srv = NULL, *dev = NULL, *prp = NULL, *c;
    strncpy(addrbuf, eventTarget, 1024); fullEventPropertyAdress[1023] = 0;
    if ((prp=strchr(addrbuf, '[')) != NULL)
    { // normal tine representation
      *prp++ = 0;
      if ((c = strchr(prp, ']')) != NULL) *c = 0;
    }
    else
    { // someone who likes to mix up where and what ...
      if ((prp=strrchr(addrbuf, '/')) != NULL) *prp++ = 0;
    }
    tokenizeFullName(addrbuf, &ctx, &srv, &dev);
    if (strcmp(ctx, "DEFAULT")) eventContext = ctx;
    if (strlen(dev) == 0 || prp == NULL)
    {
      return PyErr_Format(PyExc_TypeError, "the event target is missing a device and/or property");
    }
    eventServer = srv;
    eventDevice = dev;
    eventProperty = prp;
  }
  if (eventContext == NULL) 
  {
    eventContext = context;
  }
  sprintf(fullEventPropertyAdress,"/%.32s/%.32s/%.64s[%.64s]",eventContext,eventServer,eventDevice,eventProperty);
  tineResultCode = GetArchivedPropertyInformation(context,fullEventPropertyAdress,&prpQueryStruct);
  if (formatString != NULL) {
    formatVal = (int)GetFormatFromString(formatString);
  } else {
    formatVal = LFMT(prpQueryStruct.prpFormat);
  }
  if( size == 0) {
    size = prpQueryStruct.prpSize;
  }
  if ((eventValueBuffer=SystemCalloc(size,fmtsizeof(formatVal))) == NULL) return PyErr_NoMemory();
  DTYPEZERO(dout);
  dout.dFormat = formatVal;
  dout.dArrayLength = size;
  dout.data.vptr = eventValueBuffer;
  strncpy(dout.dTag,prpQueryStruct.prpTag,TAG_NAME_SIZE);
  tineResultCode = GetArchivedEventData( context,trigger,eventContext,eventServer,eventProperty,eventDevice,0,eventId,&dout);
  if (tineResultCode) 
  {
    PyObject *ex = pytine_getExcObject(tineResultCode);
    if (ex != NULL) return thrownTineException(ex, tineResultCode, "pytine_getArchivedEventData", context, trigger, NULL);
  }
  if (tineResultCode == 0) array_to_python(&value, eventValueBuffer, size, formatVal, NULL, FALSE);
  SystemFree(eventValueBuffer);
  double ts = dout.dTimeStamp;
  PyObject *dtype = Py_BuildValue("{s:s,s:s,s:i,s:s,s:i,s:d,s:i,s:i,s:s,s:O}",
    "context",context,"event",trigger,"eventId",eventId,"channel",fullEventPropertyAdress,
    "status", tineResultCode, "timestamp", ts, "sysstamp", dout.sysStamp, "usrstamp", dout.dStamp, 
    "timestring", GetDataTimeString(ts, FALSE), "data", value);
  Py_CLEAR(value);
  return dtype;
}
//TODO:  hash this ...
typedef struct PrpInfoItem
{
  char srv[72];
  char prp[64];
  DTYPE dout;
  DTYPE din;
  struct PrpInfoItem *nxt;
} PrpInfoItem;
PrpInfoItem *prpLst = NULL;

void addPrpItem(char *srv, char *prp, DTYPE *dout, DTYPE *din)
{
  PrpInfoItem pii, *p;
  char *c;
  if (srv == NULL || prp == NULL) return;
  if (srv[0] != '/') return; // must be fully qualified
  memset(&pii,0,72+64);
  DTYPEZERO(pii.dout); DTYPEZERO(pii.din);
  strncpy(pii.srv,srv,72);
  if ((c=strchr(&pii.srv[1],'/')) == NULL) return; // must contain a server
  c++;
  if ((c=strchr(c,'/')) != NULL) *c = 0; // lop off the device name
  strncpy(pii.prp,prp,64);
  if (dout != NULL) memcpy(&pii.dout,dout,sizeof(DTYPE)); // never called with dout = NULL !
  if (din != NULL) memcpy(&pii.din,din,sizeof(DTYPE));

  for (p=prpLst; p != NULL; p = p->nxt)
  {
    if (strnicmp(p->srv,pii.srv,66)) continue;
    if (strnicmp(p->prp,pii.prp,64)) continue;
    return; // already there !
  }
  if ((p=(PrpInfoItem *)calloc(1,sizeof(PrpInfoItem))) == NULL) return;
  memcpy(p,&pii,sizeof(PrpInfoItem));
  p->nxt = prpLst;
  prpLst = p;

  switch (dout->dFormat)
  {
    case CF_STRUCT:
      if (strlen(dout->dTag) > 0) AcquireAndRegisterStruct(srv,dout->dTag,dout->dArrayLength);
      break;
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      if (strlen(dout->dTag) > 0) AcquireAndRegisterBitfield(srv,dout->dTag);
      break;
    default:
      break;
  }
  if (din != NULL) switch (din->dFormat)
  {
    case CF_STRUCT:
      if (strlen(din->dTag) > 0 && strncmp(din->dTag,dout->dTag,TAG_NAME_SIZE) != 0) 
        AcquireAndRegisterStruct(srv,din->dTag,dout->dArrayLength);
      break;
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      if (strlen(din->dTag) > 0 && strncmp(din->dTag,dout->dTag,TAG_NAME_SIZE) != 0) 
        AcquireAndRegisterBitfield(srv,din->dTag);
      break;
    default:
      break;
  }
}
void addUsrPrpItem(char *srv, char *prp, int size, int fmt, char *tag,int access)
{
  DTYPE dout, din; 
  DTYPEZERO(dout); DTYPEZERO(din);
  if (access & CA_READ)
  {
    dout.dArrayLength = (UINT32)size;
    dout.dFormat = fmt;
    if (tag != NULL) strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  }
  if (access & CA_WRITE)
  {
    din.dArrayLength = (UINT32)size;
    din.dFormat = fmt;
    if (tag != NULL) strncpy(din.dTag,tag,TAG_NAME_SIZE);
  }
  addPrpItem(srv,prp,&dout,&din);
}
void addUsrPrpItemEx(char *srv, char *prp, int sizeOut, int fmtOut, char *tagOut, int sizeIn, int fmtIn, char *tagIn)
{
  DTYPE dout, din; 
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = (UINT32)sizeOut;
  dout.dFormat = fmtOut;
  if (tagOut != NULL) strncpy(dout.dTag,tagOut,TAG_NAME_SIZE);
  din.dArrayLength = (UINT32)sizeIn;
  din.dFormat = fmtIn;
  if (tagIn != NULL) strncpy(din.dTag,tagIn,TAG_NAME_SIZE);
  addPrpItem(srv,prp,&dout,&din);
}
PrpInfoItem *getPrpItem(char *srv, char *prp)
{
  PrpInfoItem *p;
  char *c, dev[72];
  if (srv == NULL || prp == NULL) return NULL;
  memcpy(dev,srv,72);
  if (dev[0] != '/') return NULL; // must be fully qualified
  if ((c=strchr(&dev[1],'/')) == NULL) return NULL; // must contain a server
  c++; if ((c=strchr(c,'/')) != NULL) *c = 0; // lop off the device name
  for (p=prpLst; p != NULL; p = p->nxt)
  {
    if (strnicmp(p->srv,dev,66)) continue;
    if (strnicmp(p->prp,prp,64)) continue;
    return p; // found it !
  }
  return NULL;
}
void freePrpItemList(void)
{
  PrpInfoItem *p, *pnxt;
  for (p=prpLst; p != NULL; p = pnxt)
  {
    pnxt = p->nxt;
    free(p);
  }
  prpLst = NULL;
}
static int tine_getpropertyinformation(char *server, char *property, int access, int *format, int *size, char *tag) 
{
  PrpInfoItem *pii;
  PrpQueryStruct p[10], *pqs = p;
  int pn = 10;
  char errstr[128],ltag[64];
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);

  if ((pii=getPrpItem(server,property)) != NULL)
  { // already known ?
    if (access == CA_WRITE)
    {
      *format = (int)pii->din.dFormat;
      *size = (int)pii->din.dArrayLength;
      if (tag != NULL) strncpy(tag,pii->din.dTag,TAG_NAME_SIZE);
    }
    else
    {
      *format = (int)pii->dout.dFormat;
      *size = (int)pii->dout.dArrayLength;
      if (tag != NULL) strncpy(tag,pii->dout.dTag,TAG_NAME_SIZE);
    }
    return 0;
  }
  if (access == CA_READ && getPropertyInfoFromListenerEx(server,property,size,format,ltag) == 0)
  { // is it known to an underlying listener ?
    dout.dArrayLength = (UINT32)*size;
    dout.dFormat = (short)*format;
    strncpy(dout.dTag,ltag,TAG_NAME_SIZE);
    addPrpItem(server,property,&dout,NULL);
    if (tag != NULL) strncpy(tag,ltag,TAG_NAME_SIZE);
    return 0; // we're done
  }
  int res = invalid_property;
  Py_BEGIN_ALLOW_THREADS;
  res = GetSystemPropertyInformation(server, property, &pqs, &pn);
  Py_END_ALLOW_THREADS;
  if (res == 0 && pn > 0) 
  { // info obtained from the target server
    int index = 0;
    int i = 0;
    if (access == CA_WRITE)
    {
      for (i = 0; i < pn; i++)
      {
        if (LFMT(p[i].prpFormatIn) != CF_NULL)
        {
          index = i;
          break;
        }
      }
      *format = (int)LFMT(p[index].prpFormatIn);
      if (*size == 0 && *format <= CF_FLOAT) *size = p[index].prpSizeIn;
    }
    else
    {
      *format = (int)LFMT(p[index].prpFormat);
      if (*size == 0) *size = p[index].prpSize;
    }
    dout.dArrayLength = (UINT32)p[index].prpSize;
    dout.dFormat = (short)LFMT(p[index].prpFormat);
    strncpy(dout.dTag,p[index].prpTag,TAG_NAME_SIZE);
    din.dArrayLength = (UINT32)p[index].prpSizeIn;
    din.dFormat = (short)LFMT(p[index].prpFormatIn);
    strncpy(din.dTag,p[index].prpTagIn,TAG_NAME_SIZE);
    addPrpItem(server,property,&dout,&din);
    if (tag != NULL) strncpy(tag,dout.dTag,TAG_NAME_SIZE);
    return 0;
  } 
  else 
  {
    PyErr_Format(PyExc_RuntimeError, "tine: '%s': failed to obtain information for property '%s' (%s)",
                                     server, property, GetLastLinkError(res,errstr));
    return -1;
  }
}

static int tine_read(const char *server, const char *property, void *data, int *size, int format, int access, int timeout,DTYPE *res) 
{
  DTYPE dout;
  double ts;
  int cc, acc=CA_READ|CA_MUTABLE;
  PrpInfoItem *pii;
  DTYPEZERO(dout);
  dout.data.vptr    = data;
  dout.dArrayLength = *size;
  dout.dFormat      = format;
  switch (format)
  {
    case CF_STRUCT:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      pii = getPrpItem((char *)server,(char *)property);
      if (pii != NULL) strncpy(dout.dTag,pii->dout.dTag,TAG_NAME_SIZE);
      break;
    case CF_MDA:
      {
        pii = getPrpItem((char *)server,(char *)property);
        if (pii != NULL) strncpy(dout.dTag,pii->dout.dTag,TAG_NAME_SIZE);
        DMDA *mda = (DMDA *)data;
        mda->data.bptr = dout.data.bptr + MDA_HDR_SIZE + sizeof(size_t);
      }
      break;
    case CF_GSPECTRUM:
    {
      DGSPECTRUM* gsp = (DGSPECTRUM*)data;
      gsp->spectBuffer = (float*)(dout.data.bptr + GSPECTRUM_HDR_SIZE + sizeof(int) + sizeof(size_t));
      gsp->spectBufferSize = dout.dArrayLength;
      break;
    }
    default:
      break;
  }
  acc |= access;
  cc = alsnExecLinkEx((char*)server, (char*)property, &dout, NULL, acc, timeout, &ts);
  if (cc == 0)
  {
    int sz = (int)dout.dArrayLength;
    if (dout.dFormat == CF_TEXT || dout.dFormat == CF_XML)
    { // special checking (doocs servers might reduce the returned data length)
      if (*size > sz) 
      { // make sure we're null-terminated ...
        dout.data.cptr[sz] = 0;
        if (dout.data.cptr[sz-1] != 0) sz++;
      }
    }
    *size = sz;
  }
  if (res != NULL) memcpy(res,&dout,sizeof(DTYPE));
  return cc;
}
static PyObject* pytine_stop(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *server, *property, *typeOutStr = NULL, *typeInStr = NULL;
  int sz = 0;
  PyObject *value = NULL;
  void *dataIn = NULL;
  int sizeOut = 0, sizeIn = 0, sizeBytes;
  int formatOut = CF_NULL, formatIn = CF_NULL, fmt, acc = CA_READ;
  int res;
  int use_bytearray = FALSE, use_inptbytearray = FALSE;
  char *pc, tagOut[64] = "", tagIn[64] = "";
  const char *pdtag, *pdintag;
  static const char *kwlist[] = {
    "address",
    "property",
    "input",
    "format",
    "size",
    "inputformat",
    "inputsize",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|sOsisi", (char**)kwlist, &server, &property,
    &value, &typeOutStr, &sizeOut, &typeInStr, &sizeIn))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.stop_get(address='str',property='str'[,input=obj,format='str',size=val,inputformat='str',inputsize=val)");
  }
  if ((pc = strchr(server, '[')) != NULL) *pc++ = 0;
  if (property == NULL)
  {
    if (pc == NULL) return PyErr_Format(PyExc_ValueError,
      "tine: address %s does not contain a property and no property provided as argument", server);
    property = pc;
    if ((pc = strchr(property, ']')) != NULL) *pc = 0;
  }
  if (typeOutStr != NULL)
  {
    formatOut = (int)GetFormatFromString(typeOutStr);
    if (stristr(typeOutStr, ".BYTES") != NULL) use_bytearray = TRUE;
    if (formatOut == CF_STRUCT && sizeOut == 0) sizeOut = 1;
  }
  if (typeInStr != NULL)
  {
    formatIn = (int)GetFormatFromString(typeInStr);
    if (stristr(typeInStr, ".BYTES") != NULL) use_inptbytearray = TRUE;
    if (formatIn == CF_STRUCT && sizeIn == 0) sizeIn = 1;
  }
  if (value)
  { // value given ...
    if (typeInStr == NULL) typeInStr = typeOutStr;
    if (acqInfo(sizeIn, formatIn, typeInStr, tagIn, server))
    {
      if (sizeIn == 0) sizeIn = array_length_from_python(value);
      sz = sizeIn;
      res = tine_getpropertyinformation(server, property, acc, &formatIn, &sz, tagIn);
      if (res != 0) return NULL;
      if (sizeIn == 0 || formatIn > CF_FLOAT) sizeIn = sz;
    }
    dataIn = array_from_python(value, &sizeIn, formatIn, tagIn);
    if (!dataIn) return NULL;
    if (formatIn == CF_STRUCT)
    {
      int ssiz = GetStructSize(tagIn);
      if (ssiz > 1) sizeBytes = ssiz * sizeIn;
      if (use_inptbytearray) sizeIn = sizeBytes;
    }
  }
  if (acqInfo(sizeOut, formatOut, typeOutStr, tagOut, server))
  {
    sz = sizeOut;
    res = tine_getpropertyinformation(server, property, CA_READ, &formatOut, &sz, tagOut);
    if (res != 0) return NULL;
    if (sizeOut == 0) sizeOut = sz;
  }
  pdintag = use_inptbytearray ? "" : tagIn;
  pdtag = use_bytearray ? "" : tagOut;
  if (getPrpItem(server, property) == NULL) addUsrPrpItemEx(server, property, sizeOut, formatOut, (char*)pdtag, sizeIn, formatIn, (char*)pdintag);
  fmt = (formatOut == CF_MDA) ? GetFormatFromString(tagOut) : formatOut;
  sizeBytes = fmtsizeof(fmt) * sizeOut;
  sizeBytes += getFormatHeaderSize(formatOut);
  if (formatOut == CF_STRUCT)
  {
    int ssiz = GetStructSize(tagOut);
    if (ssiz > 1) sizeBytes = ssiz * sizeOut;
    if (use_bytearray) sizeOut = sizeBytes;
  }
  if (formatOut == CF_MDA) sizeBytes += sizeof(size_t);
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = (UINT32)sizeOut; dout.dFormat = formatOut;
  strncpy(dout.dTag, pdtag, TAG_NAME_SIZE);
  dout.data.vptr = NULL;
  din.dArrayLength = (UINT32)sizeIn; din.dFormat = formatIn;
  strncpy(din.dTag, pdintag, TAG_NAME_SIZE);
  din.data.vptr = dataIn;
  Py_BEGIN_ALLOW_THREADS;
  res = alsnStopLink((char *)server, (char *)property, &dout, &din);
  Py_END_ALLOW_THREADS;
  PyObject *ex = pytine_getExcObject(res);
  if (ex != NULL) return thrownTineException(ex, res, "stop", server, property, NULL);
  Py_RETURN_NONE;
}
static PyObject *thrownTineException(PyObject *ex,int cc,const char *fcn,const char *arg0, const char *arg1, const char *arg2)
{
  char s[1024],errstr[128];
  sprintf(s, "tine: %s(",fcn);
  if (arg0 != NULL) strcat(s,arg0);
  if (arg1 != NULL) 
  { 
    strcat(s, ", "); 
    strcat(s, arg1); 
    if (arg2)
    {
      strcat(s, ", ");
      strcat(s, arg2);
    }
  }
  strcat(s, "): ");
  strcat(s, GetLastLinkError(cc, errstr));
  PyObject *eo = Py_BuildValue("(si)", s, cc);
  PyErr_SetObject(ex, eo);
  return NULL;
}
static int tine_write(const char *server, const char *property, void *data, int size, int format, char *tag, int access, int timeout, PyObject *callable)
{
  int acc = CA_READ|CA_WRITE;
  DTYPE din, dout;
  DTYPEZERO(din); DTYPEZERO(dout);
  if (format == CF_STRING)
    din.data.strptr = (char **)&data;
  else
    din.data.vptr    = data;
  din.dArrayLength = size;
  din.dFormat      = format;
  if (tag != NULL) strncpy(din.dTag,tag,TAG_NAME_SIZE);

  acc |= access;
  // Some servers generate an error if one does not read back a value.
  // Thus dout and CA_READ was added, although, not really necessary.
  if (callable != NULL)
  {
    struct callback *cb;
    if ((cb = (struct callback *)malloc(sizeof(struct callback))) == NULL)
      return out_of_local_memory;
    cb->callable = callable;
    cb->data = NULL;
    cb->size = 0;
    cb->format = CF_NULL;
    cb->mode = CM_SINGLE;
    cb->use_bytearray = false;
    memset(cb->tag, 0, TAG_NAME_SIZE);
    Py_INCREF(cb->callable);
    int lid = tine_attach(server, property, &dout, &din, acc, cb->mode, timeout, _datacallback_handler, cb, UNASSIGNED_CALLBACKID);
    return (lid < 0) ? -lid : 0;
  }
  return ExecLinkEx((char*)server, (char*)property, &dout, &din, (short)acc, (unsigned short)timeout);
}

static int tine_acq(const char *server, const char *property, DTYPE *dout, DTYPE *din, int access, int timeout) 
{
  double ts;
  int cc;
  switch (din->dFormat)
  {
    case CF_MDA:
    case CF_IMAGE:
      ccerr(not_supported);
    default:
      break;
  }
  switch (dout->dFormat)
  {
    case CF_MDA:
      {
        DMDA *mda = (DMDA *)dout->data.vptr;
        mda->data.bptr = dout->data.bptr + MDA_HDR_SIZE + sizeof(size_t);
      }
    default:
      break;
  }
  cc = alsnExecLinkEx((char*)server, (char*)property, dout, din, access|CA_MUTABLE, timeout, &ts);
err:
  return cc;
}

static int tine_query(const char *server, const char *property,
                      void *data_in, int size_in, int format_in,
            		      void *data_out, int *size_out, int *format_out,
		                  int timeout) 
{
  DTYPE din;
  DTYPE dout;
  int result;

  memset(&din, 0, sizeof(din));
  memset(&dout, 0, sizeof(dout));

  din.data.vptr    = data_in;
  din.dArrayLength = size_in;
  din.dFormat      = format_in;

  dout.data.vptr    = data_out;
  dout.dArrayLength = *size_out;
  dout.dFormat      = *format_out;

  result = ExecLinkEx((char*)server, (char*)property,
		      &dout, &din, CA_READ | CA_WRITE | CA_MUTABLE, timeout);

  *size_out = dout.dArrayLength;
  *format_out = dout.dFormat; 

  return result;
}
static int tine_attach(const char *server, const char *property,
                       DTYPE *dout, DTYPE *din, int acc, int mode, int interval,
                       void (*callback)(int, int, void*), void *p, int id) 
{
  return AttachLinkEx2((char*)server, (char*)property, dout, din,
                       acc, interval, callback, mode, id, p);
}

static int tine_detach(int linkid) 
{
  return CloseLink(linkid);
}

// Py_BuildValue("s", (char*)data); expects a null-terminated string
static PyObject* data_to_python(void *data, int slen, int format, char *tag) 
{
  int cap = slen;
  switch (format) 
  {
    case CF_BOOLEAN:
      return PyBool_FromLong(*((int *)data));
    case CF_BYTE:
      return Py_BuildValue("B", *((BYTE *)data));
    case CF_UINT16:
      return Py_BuildValue("H", *((UINT16 *)data));
    case CF_INT16:
      return Py_BuildValue("h", *((short *)data));
    case CF_UINT32:
      return Py_BuildValue("I", *((UINT32 *)data));
    case CF_INT32:
      return Py_BuildValue("i", *((int *)data));
#   ifdef SINT64
    case CF_UINT64:
      return Py_BuildValue("k", *((UINT64 *)data));
    case CF_INT64:
      return Py_BuildValue("l", *((SINT64 *)data));
#   endif
    case CF_INTINT:
      {
        int *d = (int *)data;
        return Py_BuildValue("(ii)", d[0], d[1]);
      }
    case CF_INTINTINT:
      {
        int *d = (int *)data;
        return Py_BuildValue("(iii)", d[0], d[1], d[2]);
      }
    case CF_INTINTINTINT:
      {
        int *d = (int *)data;
        return Py_BuildValue("(iiii)", d[0], d[1], d[2], d[3]);
      }
    case CF_FLOAT:
      return Py_BuildValue("f", *((float *)data));
    case CF_FLTFLT:
      {
        float *d = (float *)data;
        return Py_BuildValue("(ff)", d[0], d[1]);
      }
    case CF_FLTINT:
      {
        float f = *(float *)data;
        int i = ((int *)data)[1];
        return Py_BuildValue("(fi)", f, i);
      }
    case CF_INTFLTINT:
      {
        float f = ((float *)data)[1];
        int *i = (int *)data;
        return Py_BuildValue("(ifi)", i[0], f, i[2]);
      }
    case CF_FLTFLTINT:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        return Py_BuildValue("(ffi)", f[0], f[1], i[2]);
      }
    case CF_FLTINTINT:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        return Py_BuildValue("(fii)", f[0], i[1], i[2]);
      }
    case CF_INTFLTFLTFLT:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        return Py_BuildValue("(ifff)", i[0], f[1], f[2], f[3]);
      }
    case CF_INTINTFLTFLT:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        return Py_BuildValue("(ifff)", i[0], i[1], f[2], f[3]);
      }
    case CF_FLTINTFLTINT:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        return Py_BuildValue("(fifi)", f[0], i[1], f[2], i[3]);
      }
    case CF_DOUBLE:
      return Py_BuildValue("d", *((double *)data));
    case CF_DBLDBL:
      {
        double *d = (double *)data;
        return Py_BuildValue("(dd)", d[0], d[1]);
      }
    case CF_DBLDBLDBL:
      {
        double *d = (double *)data;
        return Py_BuildValue("(ddd)", d[0], d[1], d[2]);
      }
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_XML:
    case CF_TEXT:
      slen = (int)strlen((char *)data);
      if (cap > 0 && slen > cap) slen = cap;
      return PyUnicode_Decode((char *)data, slen, "latin-1", "ignore");
      //return Py_BuildValue("s#", (char *)data, slen);
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      cap = fmtsizeof(format);
      slen = (int)MIN(cap, (int)strlen((char *)data));
      return PyUnicode_Decode((char *)data, slen, "latin-1", "ignore");
//      return Py_BuildValue("s#", (char *)data, slen);
    case CF_USTRING:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        char s[84];
        strncpy(s,(char *)data + 16,80); s[80] = 0;
        return Py_BuildValue("iffis",i[0],f[1],f[2],i[3],s);
      }
    case CF_UNAME:
      {
        float *f = (float *)data;
        int *i = (int *)data;
        char s[20];
        strncpy(s,(char *)data + 16,20); s[16] = 0;
        return Py_BuildValue("ifffs",i[0],f[1],f[2],f[3],s);
      }
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      {
        int flen = fmtsizeof(format), ival;
        char s[72]; memset(s,0,72);
        slen = flen - 4;
        memcpy(&ival,(BYTE *)data+slen,4);
        strncpy(s,(char *)data,slen);
        slen = (int)strlen(s);
        return Py_BuildValue("s#i",s,slen,ival);
      }
    case CF_NAME16FI:
      {
        char *s = (char *)data;
        float f = *(float *)(s + 16);
        int i = *(int *)(s + 20);
        slen = (int)MIN(strlen(s), 16);
        return Py_BuildValue("s#fi",s,slen,f,i);
      }
    case CF_NAME16II:
      {
        char *s = (char *)data;
        int *i = (int *)(s + 16);
        slen = (int)MIN(strlen(s), 16);
        return Py_BuildValue("s#ii", s,slen,i[0],i[1]);
      }
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      {
        int flen = fmtsizeof(format);
        char *s = (char *)data;
        slen = flen - 16;
        double *d = (double *)(s + slen);
        slen = MIN((int)strlen(s),slen);
        return Py_BuildValue("s#dd",s,slen,d[0],d[1]);
      }
    case CF_NAME64DBLDBLDBL:
      {
        char *s = (char *)data;
        double *d = (double *)(s + 64);
        slen = (int)MIN(strlen(s), 64);
        return Py_BuildValue("s#ddd",s,slen,d[0],d[1],d[2]);
      }
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      {
        DBITFIELD *bf = findBitField(NULL,tag); // NULL out the server dependency
        struct bitfieldFormat *bff;
        if (bf == NULL) return PyErr_Format(PyExc_TypeError, "bitfield %s is not registered",tag);
        PyObject *pdo = PyDict_New();
        PyObject *item;
        int ival;
        for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
        {
          ival = GetFieldFromBitfield(NULL,tag,bff->field,*(int *)data);
          item = data_to_python(&ival,1,CF_INT32,NULL);
          if (item == NULL) return NULL;
          PyDict_SetItemString(pdo,bff->field,item);
        }
        return pdo;
      }
    case CF_STRUCT:
      {
        struct structStruct *ss = GetStructFromTag(tag);
        if (ss == NULL) return PyErr_Format(PyExc_TypeError, "structure %s is not registered",tag);
        struct structFormat *sf;
        int i, nfields = 0;
        for (sf=ss->s; sf != NULL; sf=sf->nxt) nfields++;
        PyObject *pdo = PyDict_New();
        PyObject *item;
        BYTE *p;
        for (sf=ss->s, i=0; sf != NULL && i<nfields; sf=sf->nxt, i++) 
        {
          if (sf->fmt == CF_NULL) continue;
          p = (BYTE *)data + sf->addr;
          if (sf->siz == 1 || sf->fmt == CF_TEXT) 
          {
            item = data_to_python(p,sf->siz,sf->fmt,sf->stag);
          }
          else 
          {
            item = NULL;
            array_to_python(&item,p,sf->siz,sf->fmt,sf->stag, FALSE);
          }
          if (item == NULL) return NULL;
          PyDict_SetItemString(pdo,sf->field,item);
        }
        return pdo;
      }
    //case CF_IMAGE, CF_SPECTRUM, CF_MDA: // handled in array_to_python
    default:
      return NULL;
  }
}
void getMdaIndicies(int *ia,int ival,DMDA *mda)
{
  int i;
  for (i=mda->ndim-1; i>=0; i--)
  {
    ia[i] = ival % (mda->axis[i].len);
    ival /= mda->axis[i].len;
  }
}
int getAppendedImageWidth(const DIMAGE * const apImage)
{
  if (apImage->frameHeader.aoiWidth > 0)
    return apImage->frameHeader.aoiWidth;
  else
    return apImage->frameHeader.sourceWidth;
}
int getAppendedImageHeight(const DIMAGE * const apImage)
{
  if (apImage->frameHeader.aoiHeight > 0)
    return apImage->frameHeader.aoiHeight;
  else
    return apImage->frameHeader.sourceHeight;
}
static BYTE *getRawImageArray(const DIMAGE * const apImage, int *numBytes)
{
  if (apImage->frameHeader.appendedFrameSize <= 0) return NULL;
  if (numBytes != NULL) *numBytes = apImage->frameHeader.appendedFrameSize;
  return apImage->frameBuffer;
}
static int array_to_python(PyObject **o, void *data, int size, int format, char *tag, int use_bytearray) 
{
  if (size <= 0) 
  {
    *o = Py_BuildValue("");
    return -1;
  } 
  PyObject *item;
  BYTE *p;
  char *bd;
  int i;
  switch (format)
  {
    case CF_TEXT:
    case CF_XML:
      bd = (char *)data;
      if (bd[size-1] != 0) 
      {
        if ((bd=(char *)malloc(size+8)) == NULL)
        { // oops!  reset it and lop off the end
          bd = (char *)data;
          size--;
        }
        else
        {
          memcpy(bd,data,size);
        }
        bd[size] = 0;
      }
      *o = data_to_python(bd, size, format, tag);
      if (bd != data) free(bd);
      break;
    case CF_SPECTRUM:
      {
        hdrSPECTRUM *hdr = (hdrSPECTRUM *)data;
        PyObject *spect = PyList_New(size);
        p = (BYTE *)data + sizeof(hdrSPECTRUM);
        for (i=0; i < size; i++) 
        {
          item = data_to_python(p, 1, CF_FLOAT, tag);
          PyList_SetItem(spect, i, item);
          p += sizeof(float);
        }
        *o = Py_BuildValue("{s:s,s:i,s:f,s:f,s:i,s:O}", 
          "comment",hdr->comment,"utc",hdr->tm,"start",hdr->s_start,"inc",hdr->s_inc,"status",hdr->status,
          "data",spect);
        Py_CLEAR(spect);
      }
      break;
    case CF_GSPECTRUM:
    {
      DGSPECTRUM *gsp = (DGSPECTRUM *)data;
      PyObject *spect = PyList_New(size);
      PyObject *stats_o = PyList_New(24);
      PyObject *tparam_o = PyList_New(8);
      for (i = 0; i < size; i++)
      {
        item = data_to_python(&gsp->spectBuffer[i], 1, CF_FLOAT, NULL);
        PyList_SetItem(spect, i, item);
      }
      for (i = 0; i < 24; i++)
      {
        item = data_to_python(&gsp->statistics[i], 1, CF_FLOAT, NULL);
        PyList_SetItem(stats_o, i, item);
      }
      for (i = 0; i < 8; i++)
      {
        item = data_to_python(&gsp->timing_param[i], 1, CF_INT32, NULL);
        PyList_SetItem(tparam_o, i, item);
      }
      *o = Py_BuildValue("{s:s,s:i,s:i,s:f,s:f,s:i,s:i,s:i,s:f,s:i,s:O,s:O,s:O}",
        "comment", gsp->comment, "id", gsp->id,"utc", gsp->tm,"start", gsp->s_start, "inc", gsp->s_inc,
        "status", gsp->status,"groups",gsp->groups,"group_size",gsp->group_size,"group_inc",gsp->group_inc,
        "stat_mask",gsp->stat_mask,"statistics",stats_o,"timing_param",tparam_o,
        "data", spect);
      Py_CLEAR(spect);
    }
    break;
    case CF_IMAGE:
      {
        DIMAGE *img = (DIMAGE *)data;
        p = img->frameBuffer;
        PyObject *srcHdr;
        PyObject *frmHdr;
        srcHdr = Py_BuildValue("{s:i,s:i,s:i,s:i,s:i,s:i,s:s}",
          "baseTag",img->sourceHeader.baseTag,
          "cameraPortId",img->sourceHeader.cameraPortId,
          "versionTag",img->sourceHeader.versionTag,
          "totalLength",img->sourceHeader.totalLength,
          "timestampSeconds",img->sourceHeader.timestampSeconds,
          "timestampMicroseconds",img->sourceHeader.timestampMicroseconds,
          "cameraPortName",img->sourceHeader.cameraPortName);
        frmHdr = Py_BuildValue("{s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:i,s:f,s:f,s:f,s:f,s:f,s:f,s:i,s:i,s:i,s:i,s:i}",
          "sourceWidth",img->frameHeader.sourceWidth,
          "sourceHeight",img->frameHeader.sourceHeight,
          "aoiWidth",img->frameHeader.aoiWidth,
          "aoiHeight",img->frameHeader.aoiHeight,
          "xStart",img->frameHeader.xStart,
          "yStart",img->frameHeader.yStart,
          "bytesPerPixel",img->frameHeader.bytesPerPixel,
          "effectiveBitsPerPixel",img->frameHeader.effectiveBitsPerPixel,
          "horizontalBinning",img->frameHeader.horizontalBinning,
          "verticalBinning",img->frameHeader.verticalBinning,
          "sourceFormat",img->frameHeader.sourceFormat,
          "imageFormat",img->frameHeader.imageFormat,
          "frameNumber",img->frameHeader.frameNumber,
          "eventNumber",img->frameHeader.eventNumber,
          "xScale",img->frameHeader.xScale,
          "yScale",img->frameHeader.yScale,
          "imageRotation",img->frameHeader.imageRotation,
          "fspare1",img->frameHeader.fspare1,
          "fspare2",img->frameHeader.fspare2,
          "fspare3",img->frameHeader.fspare3,
          "imageFlags",img->frameHeader.imageFlags,
          "ispare1",img->frameHeader.ispare1,
          "ispare2",img->frameHeader.ispare2,
          "ispare3",img->frameHeader.ispare3,
          "appendedFrameSize",img->frameHeader.appendedFrameSize);
        // make the image from img->frameBuffer (appendedFrameSize bytes < img->frameBufferSize bytes)
        int nb = 0;
        BYTE *imgBytesArr = getRawImageArray(img, &nb);
        if (imgBytesArr == NULL) 
        {
          PyErr_SetString(PyExc_TypeError, "No image bytes returned");
          return -1;
        }
        //*o = Py_BuildValue("{s:O,s:O,s:y#}",
        //  "sourceHeader", srcHdr, "frameHeader", frmHdr, "imageBytes", imgBytesArr, nb);
        PyObject *imgBytes = Py_BuildValue("y#",imgBytesArr,nb);
        *o = Py_BuildValue("{s:O,s:O,s:O}",
            "sourceHeader", srcHdr, "frameHeader", frmHdr, "imageBytes",imgBytes);
        Py_CLEAR(srcHdr);
        Py_CLEAR(frmHdr);
        Py_CLEAR(imgBytes);
      }
      break;
    case CF_MDA:
      { // tag will contain the canonical type (e.g. 'FLOAT')
        DMDA *mda = (DMDA *)data;
        PyObject *pdo = PyDict_New();
        PyObject *item;
        PyObject *oa[MAX_MDA_DIMS];
        int slen = (int)MIN(80,strlen(mda->cmt));
        item = PyUnicode_Decode((char *)mda->cmt, slen, "latin-1", "ignore");
        PyDict_SetItemString(pdo,"description",item);
        int fmt = GetFormatFromString(tag);
        int siz = 1, fsiz = fmtsizeof(fmt);
        int ia[MAX_MDA_DIMS];
        for (i=0; i<mda->ndim; i++)
        { // the axis information ...
          item = Py_BuildValue("{s:i,s:i,s:s,s:f,s:f}",
            "dim",i,
            "len",mda->axis[i].len,
            "units",mda->axis[i].units,
            "min",mda->axis[i].min,
            "max",mda->axis[i].max);
          PyDict_SetItemString(pdo,mda->axis[i].label,item);
          siz *= mda->axis[i].len;
          oa[i] = NULL;
        }
        int in;
        p = mda->data.bptr;
        for (i=0; i<siz; i++)
        {
          getMdaIndicies(ia,i,mda);
          item = data_to_python(p, 1, fmt, tag);
          p += fsiz;
          in = mda->ndim - 1;
          if (oa[in] == NULL) oa[in] = PyList_New(mda->axis[in].len);
          PyList_SetItem(oa[in], ia[in], item);
          while (in > 0 && ia[in] == mda->axis[in].len-1)
          {
            if (oa[in-1] == NULL) oa[in-1] = PyList_New(mda->axis[in-1].len);
            PyList_SetItem(oa[in-1], ia[in-1], oa[in]);
            oa[in] = NULL;
            in--;
          }
        }
        PyDict_SetItemString(pdo,"data",oa[0]);
        *o = pdo;
      }
      break;
    case CF_KEYVALUE:
    case CF_STRING:
      {
        char **s = (char **)data;
        int len = (int)strlen(s[0]) + 1;
        if (size == 1) *o = data_to_python(s[0],len,format,tag);
        else
        {
          *o = PyList_New(size);
          for (i=0; i<size; i++)
          {
            len = (int)strlen(s[i]) + 1;
            item = data_to_python(s[i],len,format,tag);
            PyList_SetItem(*o, i, item);
          }
        }
        break;
      }
    default:
      int fsiz = fmtsizeof(format);
      if (format == CF_STRUCT)
      {
        fsiz = GetStructSize(tag);
        if (use_bytearray)
        {
          if (fsiz > 0)
          {
            *o = makeStructDictionary(tag, ((BYTE *)data));
            break;
          }
          fsiz = fmtsizeof(CF_STRUCT);
          format = CF_BYTE;
          use_bytearray = FALSE;
        }
      }
      if (size == 1)
      {
        *o = data_to_python(data, 0, format, tag);
      }
      else if (use_bytearray)
      {
        int nb = size * fsiz;
        *o = Py_BuildValue("y#", ((BYTE *)data), nb);
      }
      else
      {
        *o = PyList_New(size);
        for (i=0; i<size; i++) 
        {
          p = (BYTE *)data + i*fsiz;
          item = data_to_python(p, 1, format, tag);
          PyList_SetItem(*o, i, item);
        }
        break;
      }
  }
  if (*o == NULL)
  {
    PyErr_SetString(PyExc_TypeError, "Could not make a python object (non-utf8 char?)");
    return -1;
  }
  return 0;
}

static int data_from_python(PyObject *o, void *data, int format, char *tag)
{
  char *tmp = NULL;
  int slen = 16;
  PyObject *tuple = Py_BuildValue("(O)", o);
  switch (format) 
  {
    case CF_BYTE:
      PyArg_ParseTuple(tuple, "B", (BYTE *)data);
      break;
    case CF_UINT16:
      PyArg_ParseTuple(tuple, "H", (UINT16 *)data);
    case CF_INT16:
      PyArg_ParseTuple(tuple, "h", (short *)data);
      break;
    case CF_BOOLEAN:
    case CF_UINT32:
      PyArg_ParseTuple(tuple, "I", (UINT32 *)data);
    case CF_INT32:
      PyArg_ParseTuple(tuple, "i", (int *)data);
      break;
    case CF_UINT64:
      PyArg_ParseTuple(tuple, "k", (UINT32 *)data);
    case CF_INT64:
      PyArg_ParseTuple(tuple, "l", (int *)data);
      break;
    case CF_INTINT:
      {
        int *d = (int *)data;
        PyArg_ParseTuple(tuple, "(ii)", d, &d[1]);
        break;
      }
    case CF_INTINTINT:
      {
        int *d = (int *)data;
        PyArg_ParseTuple(tuple, "(iii)", d, &d[1], &d[2]);
        break;
      }
    case CF_INTINTINTINT:
      {
        int *d = (int *)data;
        PyArg_ParseTuple(tuple, "(iiii)", d, &d[1], &d[2], &d[3]);
        break;
      }
    case CF_FLOAT:
      PyArg_ParseTuple(tuple, "f", (float*)data);
      break;
    case CF_FLTFLT:
      {
        float *d = (float *)data;
        PyArg_ParseTuple(tuple, "(ff)", d, &d[1]);
        break;
      }
    case CF_FLTINT:
      {
        FLTINT *d = (FLTINT *)data;
        PyArg_ParseTuple(tuple,"(fi)", &d->fval, &d->ival);
        break;
      }
    case CF_INTFLTINT:
      {
        INTFLTINT *d = (INTFLTINT *)data;
        PyArg_ParseTuple(tuple,"(ifi)",&d->i1val,&d->fval,&d->i2val);
        break;
      }
    case CF_FLTFLTINT:
      {
        FLTFLTINT *d = (FLTFLTINT *)data;
        PyArg_ParseTuple(tuple,"(ffi)",&d->f1val,&d->f2val,&d->ival);
        break;
      }
    case CF_FLTINTINT:
      {
        FLTINTINT *d = (FLTINTINT *)data;
        PyArg_ParseTuple(tuple,"(fii)",&d->fval,&d->i1val,&d->i2val);
        break;
      }
    case CF_INTFLTFLTFLT:
      {
        DFILTER *d = (DFILTER *)data;
        PyArg_ParseTuple(tuple,"(ifff)",&d->ival,&d->f1val,&d->f2val,&d->f3val);
        break;
      }
    case CF_INTINTFLTFLT:
      {
        FWINDOW *d = (FWINDOW *)data;
        PyArg_ParseTuple(tuple,"(iiff)",&d->i1val,&d->i2val,&d->f1val,&d->f2val);
        break;
      }
    case CF_FLTINTFLTINT:
      {
        FLTINTFLTINT *d = (FLTINTFLTINT *)data;
        PyArg_ParseTuple(tuple,"(fifi)",&d->f1val,&d->i1val,&d->f2val,&d->i2val);
        break;
      }
    case CF_DOUBLE:
      PyArg_ParseTuple(tuple, "d", (double*)data);
      break;
    case CF_DBLDBL:
      {
        double *d = (double *)data;
        PyArg_ParseTuple(tuple, "(dd)", d, &d[1]);
        break;
      }
    case CF_DBLDBLDBL:
      {
        double *d = (double *)data;
        PyArg_ParseTuple(tuple, "(ddd)", d, &d[1], &d[2]);
        break;
      }
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_XML:
    case CF_TEXT:
      slen = (int)PyObject_Length(o);
      if (PyArg_ParseTuple(tuple, "es", "latin-1",&tmp))
      {
        strcpy((char *)data, tmp);
        PyMem_Free(tmp);
      }
      break;
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      /* Python pre-allocates memory for the string. */
      slen = fmtsizeof(format);
      if (PyArg_ParseTuple(tuple, "es", "latin-1", &tmp))
      {
        strncpy((char *)data, tmp, slen);
        PyMem_Free(tmp);
      }
      break;
    case CF_USTRING:
      {
        DUSTRING *d = (DUSTRING *)data;
        if (PyArg_ParseTuple(tuple,"(iffis)",&d->ival,&d->f1val,&d->f2val,&d->tm,&tmp))
          strncpy(d->str,tmp,80);
        break;
      }
    case CF_UNAME:
      {
        DUNAME *d = (DUNAME *)data;
        if (PyArg_ParseTuple(tuple,"(ifffs)",&d->ival,&d->f1val,&d->f2val,&d->f3val,&tmp))
          strncpy(d->name,tmp,16);
        break;
      }
    case CF_SPECTRUM:
    {
      PyObject *fo = NULL;
      hdrSPECTRUM *d = (hdrSPECTRUM *)data;
      float *fa = NULL, *sa = NULL;
      int sz = 0;
      if (PyArg_ParseTuple(tuple, "(siffiO)", &tmp, &d->tm, &d->s_start, &d->s_inc, &d->status, &fo))
      {
        strncpy(d->comment, tmp, 80);
        sa = (float *)((BYTE *)data + SPECTRUM_HDR_SIZE);
        sz = (int)PyList_Size(fo);
        fa = (float *)array_from_python(fo, &sz, CF_FLOAT, NULL);
        memcpy(sa, fa, sz * sizeof(float));
        free(fa);
      }
      break;
    }
    case CF_GSPECTRUM:
    {
      PyObject *fo = NULL, *stat_o=NULL, *tparam_o=NULL;
      DGSPECTRUM *d = (DGSPECTRUM *)data;
      float *fa = NULL, *sa = NULL;
      int sz = 0, *ia = NULL;
      if (PyArg_ParseTuple(tuple, "(siiffiiifiOOO)", &tmp, 
        &d->id,&d->tm,&d->s_start, &d->s_inc, &d->status, &d->groups, &d->group_size, 
        &d->group_inc,&d->stat_mask,&stat_o,&tparam_o, &fo))
      {
        strncpy(d->comment, tmp, 80);
        sz = (int)PyList_Size(fo);
        d->spectBuffer = (float*)array_from_python(fo, &sz, CF_FLOAT, NULL);
        d->spectBufferSize = sz;
        sz = MAX((int)PyList_Size(stat_o),24);
        fa = (float*)array_from_python(stat_o, &sz, CF_FLOAT, NULL);
        memcpy(d->statistics, fa, sz*sizeof(float));
        sz = MAX((int)PyList_Size(tparam_o), 8);
        ia = (int *)array_from_python(tparam_o, &sz, CF_INT32, NULL);
        memcpy(d->timing_param, ia, sz * sizeof(int));
        free(fa);
        free(ia);
      }
      break;
    }
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      {
        int flen = fmtsizeof(format), ival;
        slen = flen - 4;
        if (PyArg_ParseTuple(tuple,"(si)",&tmp,&ival))
        {
          memcpy((BYTE *)data+slen,&ival,sizeof(SINT32));
          slen = (int)MIN(slen, (int)strlen(tmp));
          strncpy((char *)data,tmp,slen);
        }
        break;
      }
    case CF_NAME16FI:
      {
        float fval;
        int ival;
        if (PyArg_ParseTuple(tuple,"(sfi)",&tmp,&fval,&ival))
        {
          memcpy((BYTE *)data+16,&fval,sizeof(float));
          memcpy((BYTE *)data+20,&ival,sizeof(SINT32));
          slen = (int)MIN(slen, (int)strlen(tmp));
          strncpy((char *)data,tmp,slen);
        }
        break;
      }
    case CF_NAME16II:
      {
        int i[2];
        if (PyArg_ParseTuple(tuple,"(sii)",&tmp,&i[0],&i[1]))
        {
          memcpy((BYTE *)data+16,i,2*sizeof(SINT32));
          slen = (int)MIN(slen, (int)strlen(tmp));
          strncpy((char *)data,tmp,slen);
        }
        break;
      }
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      {
        int flen = fmtsizeof(format);
        slen = flen - 16;
        double d[2];
        if (PyArg_ParseTuple(tuple,"(sdd)",&tmp,&d[0],&d[1]))
        {
          memcpy((BYTE *)data+slen,d,2*sizeof(double));
          slen = (int)MIN(slen, (int)strlen(tmp));
          strncpy((char *)data,tmp,slen);
        }
        break;
      }
    case CF_NAME64DBLDBLDBL:
      {
        double d[3];
        if (PyArg_ParseTuple(tuple,"(sddd)",&tmp,&d[0],&d[1],&d[2]))
        {
          slen = 64;
          memcpy((BYTE *)data+slen,d,3*sizeof(double));
          slen = (int)MIN(slen, (int)strlen(tmp));
          strncpy((char *)data,tmp,slen);
        }
        break;
      }
    case CF_STRUCT:
      {
        if (PyBytes_Check(o))
        { // untagged structure (Frank's optics server?)
          PyArg_ParseTuple(tuple, "s*", (BYTE *)data);
//          PyArg_ParseTuple(tuple, "B", (BYTE *)data);
          break;
        }
        if (!PyDict_Check(o))
        {
          PyErr_SetString(PyExc_TypeError, "dictionary expected for structure type");
          break;
        }
        PyObject *typ, *val;
        Py_ssize_t dp;
        PyObject *item;
        char s[128];
        struct structStruct *ss = GetStructFromTag(tag);
        if (ss == NULL) 
        {
          sprintf(s,"structure %s is not registered",tag);
          PyErr_SetString(PyExc_TypeError, s);
          return -1;
        }
        BYTE *p;
        struct structFormat *sf;
        int nfields = 0;
        char fld[32]; memset(fld,0,32);
        for (sf=ss->s; sf != NULL; sf=sf->nxt) 
        {
          p = (BYTE *)data + sf->addr;
          strncpy(fld,sf->field,16); 
          item = PyDict_GetItemString(o, fld);
          if (item == NULL)
          {
            sprintf(s,"dictionary missing key: %.16s",fld);
            PyErr_SetString(PyExc_TypeError, s);
            break;
          }
          if (PyDict_Check(item))
          { // dictionary item contains formatting info
            dp = 0;
            PyDict_Next(item, &dp, &typ, &val);
            item = val;
          }
          if (sf->siz > 1)
          {
            int sz = sf->siz;
            array_from_python(item, &sz, sf->fmt, sf->stag);
          }
          else
          {
            data_from_python(item, p, sf->fmt, sf->stag);
          }
          nfields++;
        }
        break;
      }
    default:
      PyErr_SetString(PyExc_TypeError, "attempt to convert unhandled type (single element is not a supported type)");
      break;
  }
  Py_CLEAR(tuple);
  PyObject *err = PyErr_Occurred();
# ifdef DEBUG
  if (err != NULL)
  {
    PyObject *ptype, *pvalue, *ptraceback;
    char *str;
    PyErr_Fetch(&ptype, &pvalue, &ptraceback);
    str = python_to_string(pvalue);
    PyErr_SetString(PyExc_TypeError, str);
  }
# endif
  return err == NULL ? 0 : -1;
}
static char *python_to_string(PyObject *pyo)
{
  PyObject *pystr = PyObject_Str(pyo);
  char *str;
# if PY_MAJOR_VERSION >= 3
  str = (char *)_PyUnicode_AsString(pystr);
# else
  str = PyString_AsString(pystr);
# endif
  return str;
}
static PyObject *pytine_getExcObject(int cc)
{
  switch (cc)
  {
    case 0: 
      return NULL;
#   if PY_MAJOR_VERSION >= 3
    case link_timeout:
    case connection_timeout:
      return PyExc_TimeoutError;
      break;
    case access_denied:
    case unauthorised_action:
      return PyExc_PermissionError;
      break;
    case out_of_range:
      return PyExc_ValueError;
      break;
    case illegal_format:
      return PyExc_TypeError;
      break;
    case illegal_read_write:
      return PyExc_AttributeError;
      break;
    case non_existent:
    case non_existent_elem:
    case host_not_resolved:
    case non_existent_server:
    case non_existent_fec:
      return PyExc_LookupError;
      break;
#   endif
    default:
      break;
  }
  return PyExc_Exception;
}
char gSrvEqm[16] = "";
static PyObject *pytine_allowedusers(PyObject* self, PyObject* args, PyObject* kwargs)
{
  PyObject *lstadd = NULL, *lstrmv = NULL;
  static const char* kwlist[] = {
    "add",
    "remove",
    NULL
  };
  int cc = 0;
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OO", (char**)kwlist, &lstadd, &lstrmv))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.allowed_users([add='list',remove='list'])");
  }
  if (strlen(gSrvEqm) == 0)
  {
    return PyErr_Format(PyExc_NameError, "Server does not yet have a module name (not initialized?");
  }
  if (lstadd != NULL)
  { // users to add
    int len = 0;
    NAME16 *lst = (NAME16 *)array_from_python(lstadd, &len, CF_NAME16, NULL);
    cc = AppendRegisteredUsers(gSrvEqm, lst, len);
  }
  if (lstrmv != NULL)
  { // users to remove
    int len = 0;
    NAME16 *lst = (NAME16 *)array_from_python(lstrmv, &len, CF_NAME16, NULL);
    cc = RemoveRegisteredUser(gSrvEqm, lst, len);
  }
  // return a list of current registered users
  int len = 128;
  NAME16 lst[128];
  cc = GetRegisteredUsers(gSrvEqm, lst, &len);
  PyObject *o = NULL;
  array_to_python(&o, lst, len, CF_NAME16, NULL, FALSE);
  return o;
}
static PyObject *pytine_allowedaddresses(PyObject* self, PyObject* args, PyObject* kwargs)
{
  PyObject* lstadd = NULL, * lstrmv = NULL;
  static const char* kwlist[] = {
    "add",
    "remove",
    NULL
  };
  int cc = 0;
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OO", (char**)kwlist, &lstadd, &lstrmv))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.allowed_addresses([add='list',remove='list'])");
  }
  if (strlen(gSrvEqm) == 0)
  {
    return PyErr_Format(PyExc_NameError, "Server does not yet have a module name (not initialized?");
  }
  if (lstadd != NULL)
  { // users to add
    int len = 0;
    NAME64* lst = (NAME64 *)array_from_python(lstadd, &len, CF_NAME64, NULL);
    cc = AppendRegisteredNetsList(gSrvEqm, lst, len);
  }
  if (lstrmv != NULL)
  { // users to remove
    int len = 0;
    NAME64 *lst = (NAME64 *)array_from_python(lstrmv, &len, CF_NAME64, NULL);
    cc = RemoveRegisteredNetsList(gSrvEqm, lst, len);
  }
  // return a list of current registered users
  int len = 128;
  NAME64 lst[128];
  cc = GetRegisteredNetworks(gSrvEqm, lst, &len);
  PyObject* o = NULL;
  array_to_python(&o, lst, len, CF_NAME64, NULL, FALSE);
  return o;
}
static PyObject *pytine_register_errorcode(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *msg = NULL, *msgLong = NULL;
  int code = 0;
  static const char* kwlist[] = {
    "code",
    "text",
    "longtext",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "is|s", (char**)kwlist, &code, &msg, &msgLong))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.register_error_code(code=value,text='str',longtext='str')");
  }
  if (msg == NULL)
  {
    return PyErr_Format(PyExc_SyntaxError, "missing error text");
  }
  if (msgLong == NULL) msgLong = msg;
  int cc = RegisterErrorCode(code, msg, msgLong);
  if (cc != 0)
  {
    char s[128];
    sprintf(s, "cannot register error code : %.32s", cc2str(cc));
    PyErr_SetString(PyExc_TypeError, s);
    return PyErr_Occurred();
  }
  Py_RETURN_NONE;
}

static PyObject *pytine_register_type(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *tag = NULL;
  PyObject *o = NULL;
  static const char *kwlist[] = {
    "tag",
    "struct",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO", (char**)kwlist, &tag, &o))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.registerType(tag='str',struct=obj)");
  }
  if (!PyDict_Check(o))
  {
    return PyErr_Format(PyExc_TypeError, "dictionary expected for structure type");
  }
  PyObject *key, *value;
  Py_ssize_t pos = 0;
  int addr = 0, nelems = 1, cc = 0;
  while (PyDict_Next(o, &pos, &key, &value)) 
  {
    char *fld = python_to_string(key);
    nelems = 1;
    if (PyList_Check(value))
    { // check arrays ....
      nelems = (int)PyList_Size(value);
      value = PyList_GetItem(value, 0);
    }
    if (PyDict_Check(value))
    { // formatting instructions ...
      PyObject *typ, *val;
      Py_ssize_t p = 0;
      PyDict_Next(value, &p, &typ, &val);
      char *c, *pstr = python_to_string(typ);
      char typstr[72];
      strncpy(typstr, pstr, 64);
      //Py_DECREF(pstr);
      if ((c=strchr(typstr, '*')) != NULL)
      {
        *c++ = 0;
        nelems = atoi(c);
      }
      int fmt = GetFormatFromString(typstr);
      switch (fmt)
      {
        case CF_BYTE:
        case CF_CHAR:
          break;
        case CF_UINT16:
        case CF_INT16:
        case CF_DOUBLE:
        case CF_UINT32:
        case CF_INT32:
        case CF_FLOAT:
          if (PyList_Check(val)) nelems = (int)PyList_Size(val);
          break;
        default:
          char msg[256];
          sprintf(msg, "field format %s not supported", typstr);
          //Py_DECREF(fld);
          return PyErr_Format(PyExc_TypeError, msg);
      }
      AddFieldToStruct(tag, addr, nelems, fmt, fld);
      addr += fmtsizeof(fmt) * nelems;
      //Py_DECREF(fld);
      continue;
    }
    if (PyFloat_Check(value))
    {
      AddFieldToStruct(tag, addr, nelems, CF_FLOAT, fld);
      addr += sizeof(float) * nelems;
    }
    if (PyLong_Check(value))
    {
      AddFieldToStruct(tag, addr, nelems, CF_INT32, fld);
      addr += sizeof(int)* nelems;
    }
    if (PyBytes_Check(value))
    {
      AddFieldToStruct(tag, addr, nelems, CF_BYTE, fld);
      addr += nelems;
    }
    if (PyUnicode_Check(value))
    {
      char *c = python_to_string(value);
      nelems = (int)strlen(c);
      if (nelems > 0)
      {
        if (nelems % 2) nelems++;
        AddFieldToStruct(tag, addr, nelems, CF_CHAR, fld);
        addr += nelems;
      }
    }
    //Py_DECREF(fld);
  }
  cc = SealTaggedStruct(tag, addr, 1);
  return Py_BuildValue("i", cc);
}
static PyObject *structure_to_bytes(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *tag = NULL;
  PyObject *o = NULL;
  static const char *kwlist[] = {
    "tag",
    "struct",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO", (char**)kwlist, &tag, &o))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.registerType(tag='str',struct=obj)");
  }
  if (!PyDict_Check(o))
  {
    return PyErr_Format(PyExc_TypeError, "dictionary expected for structure type");
  }

  PyObject *item;
  char s[128];
  struct structStruct *ss = GetStructFromTag(tag);
  if (ss == NULL)
  {
    sprintf(s, "structure %s is not registered", tag);
    PyErr_SetString(PyExc_TypeError, s);
    return PyErr_Format(PyExc_TypeError, s);
  }
  int siz = GetStructSize(tag);
  BYTE *p, *fdata, *data = (BYTE *)alloca(siz);
  struct structFormat *sf;
  int dsiz, nfields = 0;
  char fld[32]; memset(fld, 0, 32);
  for (sf = ss->s; sf != NULL; sf = sf->nxt)
  {
    p = (BYTE *)data + sf->addr;
    strncpy(fld, sf->field, 16);
    item = PyDict_GetItemString(o, fld);
    if (item == NULL)
    {
      sprintf(s, "dictionary missing key: %.16s", fld);
      return PyErr_Format(PyExc_KeyError, s);
    }
    if (PyDict_Check(item))
    { // value also contains formatting instructions
      PyObject *typ, *val;
      Py_ssize_t pi = 0;
      PyDict_Next(item, &pi, &typ, &val);
      item = val;
    }
    siz = sf->siz;
    fdata = (BYTE*)array_from_python(item, &siz, sf->fmt, sf->stag);
    dsiz = siz * GetFormatSize(sf->fmt); // why doesn't: siz *= GetFormatSize(sf->fmt); work?
    if (fdata != NULL) memcpy(p, fdata, dsiz);
    nfields++;
  }
  PyObject *b = NULL;
  array_to_python(&b, data, ss->size, CF_BYTE, (char*)"", FALSE);
  return b;
  // return a 'bytes' object :
  //return Py_BuildValue("y#", data, ss->size);
}
// get forwarded ordered list for easy assignments ...
static int getStructFieldArray(char *tag, struct structFormat **flds, int *nfields)
{
  struct structStruct *ss = GetStructFromTag(tag);
  struct structFormat *sf;
  int i, nflds = 0;
  if (flds == NULL || nfields == NULL) return argument_list_error;
  if (ss == NULL)
  {
    *nfields = 0;
    return invalid_structure_tag;
  }
  for (sf = ss->s; sf != NULL; sf = sf->nxt)
  {
    if (sf->fmt == CF_NULL) break;
    nflds++;
  }
  if (nflds > *nfields) return buffer_too_small;
  for (sf=ss->s, i=nflds-1; sf != NULL && i >= 0; sf = sf->nxt, i--)
  {
    flds[i] = sf;
    if (sf->fmt == CF_NULL) break;
  }
  *nfields = nflds;
  return 0;
}
static PyObject *makeStructDictionary(char *tag, BYTE *data)
{
  struct structStruct *ss = GetStructFromTag(tag);
  if (ss == NULL) return PyErr_Format(PyExc_TypeError, "structure %s is not registered", tag);
  struct structFormat *sf;
  int cc, i, nfields = 512;
  struct structFormat *flds[512];
  if (data == NULL)
    return PyErr_Format(PyExc_AssertionError, "no structure data for dictionary");
  if ((cc=getStructFieldArray(tag, flds, &nfields)) != 0)
    return PyErr_Format(PyExc_TypeError, "get structure array for %s : %.32s", tag, cc2str(cc));
  PyObject *pdo = PyDict_New();
  PyObject *item;
  BYTE *p;
  for (i = 0; i < nfields; i++)
  {
    sf = flds[i];
    if (sf->fmt == CF_NULL) continue;
    p = (BYTE *)data + sf->addr;
    if (sf->siz == 1 || sf->fmt == CF_TEXT)
    {
      item = data_to_python(p, sf->siz, sf->fmt, sf->stag);
    }
    else
    {
      item = NULL;
      array_to_python(&item, p, sf->siz, sf->fmt, sf->stag, FALSE);
    }
    if (item == NULL) return NULL;
    PyDict_SetItemString(pdo, sf->field, item);
  }
  return pdo;
}
static PyObject *bytes_to_structure(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *tag = NULL;
  PyObject *o = NULL;
  static const char *kwlist[] = {
    "tag",
    "bytes",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO", (char**)kwlist, &tag, &o))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.registerType(tag='str',bytes=obj)");
  }
  if (!PyBytes_Check(o) && !PyList_Check(o))
  {
    return PyErr_Format(PyExc_TypeError, "bytes object expected for bytes type");
  }
  char *data = NULL;
  if (PyBytes_Check(o)) data = PyBytes_AsString(o);
  else
  {
    int siz = array_length_from_python(o);
    data = (char *)array_from_python(o, &siz, CF_BYTE, NULL);
  }
  return makeStructDictionary(tag, (BYTE *)data);
}
static int array_length_from_python(PyObject *o)
{ // used to determine input length when not explicitly given
  if (o == NULL) return 0;
  if (PyBytes_Check(o)) return (int)PyBytes_Size(o);
  if (PyUnicode_Check(o)) return 1;  // 1 string 
  if (PyList_Check(o)) return (int)PyList_Size(o);
  return 1;
}
static int format_from_python(PyObject *o)
{
  if (o == NULL) return CF_NULL;
  if (PyBytes_Check(o)) return CF_BYTE;
  if (PyUnicode_Check(o)) return CF_STRING;
  if (PyLong_Check(o)) return CF_INT32;
  if (PyFloat_Check(o)) return CF_FLOAT;
# if PY_MAJOR_VERSION < 3
  if (PyInt_Check(o)) return CF_INT32;
# endif
  if (PyList_Check(o))
  {
    PyObject *io = PyList_GetItem(o, 0);
    return format_from_python(io);
  }
  return CF_NULL;
}
static int getPythonStringLength(PyObject *o)
{
# if PY_MAJOR_VERSION >= 3
  return (int)PyUnicode_GetLength(o);
# else
  return (int)PyUnicode_GetSize(o);
# endif
}
static void *array_from_python(PyObject *o, int *size, int format, char *tag) 
{
  int string_size = 0;
  int bytes_size = 0;
  int hdr_size = 0;
  if (PyDict_Check(o) && (tag == NULL || strlen(tag) == 0))
  { // a dictionary passed but it doesn't appear to be a structure
    if (format == CF_SPECTRUM)
    {
      char cmt[80] = "";
      int cc, tm = 0, sts = 0;
      float start = 0.0, inc = 1.0;
      PyObject *tpl = PyDict_GetItemString(o, "comment");
      if (tpl != NULL) cc = data_from_python(tpl, cmt, CF_TEXT, NULL);
      tpl = PyDict_GetItemString(o, "utc");
      if (tpl != NULL) cc = data_from_python(tpl, &tm, CF_INT32, NULL);
      tpl = PyDict_GetItemString(o, "start");
      if (tpl != NULL) cc = data_from_python(tpl, &start, CF_FLOAT, NULL);
      tpl = PyDict_GetItemString(o, "inc");
      if (tpl != NULL) cc = data_from_python(tpl, &inc, CF_FLOAT, NULL);
      tpl = PyDict_GetItemString(o, "status");
      if (tpl != NULL) cc = data_from_python(tpl, &sts, CF_INT32, NULL);
      o = Py_BuildValue("(siffiO)", cmt, tm, start, inc, sts, PyDict_GetItemString(o, "data"));
    }
    else if (format == CF_GSPECTRUM)
    {
      char cmt[80] = "";
      int cc, id = 0, tm = 0, sts = 0, grps = 1, grp_siz = 1, sts_msk = 0;
      float start = 0.0, inc = 1.0, grp_inc = 1.0;
      PyObject* tpl = PyDict_GetItemString(o, "comment");
      if (tpl != NULL) cc = data_from_python(tpl, cmt, CF_TEXT, NULL);
      tpl = PyDict_GetItemString(o, "id");
      if (tpl != NULL) cc = data_from_python(tpl, &id, CF_INT32, NULL);
      tpl = PyDict_GetItemString(o, "utc");
      if (tpl != NULL) cc = data_from_python(tpl, &tm, CF_INT32, NULL);
      tpl = PyDict_GetItemString(o, "start");
      if (tpl != NULL) cc = data_from_python(tpl, &start, CF_FLOAT, NULL);
      tpl = PyDict_GetItemString(o, "inc");
      if (tpl != NULL) cc = data_from_python(tpl, &inc, CF_FLOAT, NULL);
      tpl = PyDict_GetItemString(o, "status");
      if (tpl != NULL) cc = data_from_python(tpl, &sts, CF_INT32, NULL);
      tpl = PyDict_GetItemString(o, "groups");
      if (tpl != NULL) cc = data_from_python(tpl, &grps, CF_INT32, NULL);
      tpl = PyDict_GetItemString(o, "group_size");
      if (tpl != NULL) cc = data_from_python(tpl, &grp_siz, CF_INT32, NULL);
      tpl = PyDict_GetItemString(o, "group_inc");
      if (tpl != NULL) cc = data_from_python(tpl, &grp_inc, CF_FLOAT, NULL);
      tpl = PyDict_GetItemString(o, "stat_mask");
      if (tpl != NULL) cc = data_from_python(tpl, &sts_msk, CF_INT32, NULL);
      o = Py_BuildValue("(siiffiiifiOOO)", cmt, id, tm, start, inc, sts, grps, grp_siz, grp_inc, sts_msk,
        PyDict_GetItemString(o, "statistics"),
        PyDict_GetItemString(o, "timing_param"),
        PyDict_GetItemString(o, "data"));
    }
    else
    {
      o = PyDict_GetItemString(o, "data");
    }
    if (o == NULL) 
    {
      PyErr_SetString(PyExc_TypeError, "no recognizable data passed !");
      return NULL;
    }
  }
  if (PyBytes_Check(o))
  {
    bytes_size = (int)PyBytes_Size(o);
    if (*size == 1) *size = bytes_size;
  }
  if (PyUnicode_Check(o))
  { // 1 string item passed
    string_size = getPythonStringLength(o);
    if (fmtsizeof(format) == 1)
    { // either CF_TEXT or CF_STRING
      if (format == CF_TEXT)
      {
        if (*size < string_size) *size = string_size;
        else string_size = *size;
      }
    }
    else
    { // input likely 1 NAME64, 1 USTRING, etc.
      if (*size < 1) *size = 1;
    }
  }
# if PY_MAJOR_VERSION < 3
  if (PyInt_Check(o) || PyLong_Check(o) || PyFloat_Check(o) || PyBytes_Check(o) || PyDict_Check(o) || PyUnicode_Check(o) || PyTuple_Check(o))
# else
  if (PyLong_Check(o) || PyFloat_Check(o) || PyBytes_Check(o) || PyDict_Check(o) || PyUnicode_Check(o) || PyTuple_Check(o))
# endif
  {
    int total_size;
    int elem_size = string_size > 0  ? string_size : *size;
    total_size = elem_size * fmtsizeof(format) + hdr_size;
    if (format == CF_TEXT || format == CF_STRING) total_size++;
    if (format == CF_SPECTRUM)
    {
      PyObject* fo = PyTuple_GetItem(o, 5);
      int sz = (int)PyList_Size(fo);
      if (sz > *size) *size = sz;
      total_size = *size * sizeof(float) + SPECTRUM_HDR_SIZE;
    }
    if (format == CF_GSPECTRUM)
    {
      PyObject* fo = PyTuple_GetItem(o, 12);
      int sz = (int)PyList_Size(fo);
      if (sz > *size) *size = sz;
      total_size = *size * sizeof(float) + GSPECTRUM_HDR_SIZE;
    }
    if (format == CF_STRUCT)
    {
      int ssiz = GetStructSize(tag);
      if (ssiz != 0) total_size *= ssiz;
      else total_size = bytes_size;
    }
    void *data = malloc(total_size);
    memset(data, 0, total_size);    
    if (data_from_python(o, data, format, tag) != 0) 
    {
      free(data);
      return NULL;
    }
    return data;
  } 
  if (PyList_Check(o)) 
  { // an array of something ...
    void *data, *p;
    PyObject *item;
    int i, dsize, imax = (int)PyList_Size(o);
    if (imax > *size) *size = imax;
    if (format == CF_STRUCT && (tag == NULL || strlen(tag) == 0)) format = CF_BYTE;
    if (format == CF_STRING)
    { // special considerations ...
      dsize = 0;
      for (i = 0; i < imax; i++)
      {
        item = PyList_GetItem(o, i);
        if (PyUnicode_Check(item))
        { // this should be TRUE if format == CF_STRING!
          dsize += getPythonStringLength(item) + 1;
        }
        else
        {
          PyErr_SetString(PyExc_TypeError, "an array of strings was not passed!");
          return NULL;
        }
      }
    }
    else
    {
      dsize = imax * fmtsizeof(format);
    }
    if (dsize == 0)
    {
      PyErr_SetString(PyExc_TypeError, "attempt to convert to a null type (either data size or format is empty or unknown)");
      return NULL;
    }
    data = malloc(dsize);
    memset(data, 0, dsize);
    int ptr = 0;
    for (i=0; i<imax; i++) 
    {
      //p = (char *)data + i * fmtsizeof(format);
      p = (char *)data + ptr;
      item = PyList_GetItem(o, i);
      ptr += (format == CF_STRING ? (getPythonStringLength(item) + 1) : fmtsizeof(format));
      if (data_from_python(item, p, format, tag)) 
      {
        free(data);
        return NULL;
      }
    }
    *size = imax;
    return data;
  } 
  else 
  {
    PyErr_SetString(PyExc_TypeError, "attempt to convert unhandled type (not a number, string, byte array, dictionary, or list)");
    return NULL;
  }
}
#define MAXNUM_SYSATTRIBUTES 100
static int gNumSysAttrbutes = MAXNUM_SYSATTRIBUTES;
static NAME64I gSysAttributesList[MAXNUM_SYSATTRIBUTES];
static void getSysAttributesList(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  Py_BEGIN_ALLOW_THREADS;
  if (GetSystemAttributeList(gSysAttributesList, &gNumSysAttrbutes) != 0) gNumSysAttrbutes = 0;
  Py_END_ALLOW_THREADS;
}
static int getSysAttributeFormat(char *attr)
{
  if (attr == NULL) return CF_INT32;
  getSysAttributesList();
  for (int i = 0; i < gNumSysAttrbutes; i++)
  {
    if (!stricmp(attr, gSysAttributesList[i].name))
    {
      return gSysAttributesList[i].ival;
    }
  }
  return CF_INT32;
}
static PyObject* pytine_setAttribute(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *attr, *typestr = NULL;
  PyObject *value = NULL;
  void *data;
  int format = CF_INT32;
  int size=1,res;
  static const char *kwlist[] = {
        "attribute",
        "value",
        "format",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO|s", (char**)kwlist, &attr, &value, &typestr)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.setAttribute(attribute='str',value=obj[,format='str')");    
  }
  format = typestr == NULL ? getSysAttributeFormat(attr) : (int)GetFormatFromString(typestr);
  data = array_from_python(value, &size, format, NULL);
  if (!data) return NULL;
  Py_BEGIN_ALLOW_THREADS;
  res = SetSystemAttribute(attr,data,format);
  Py_END_ALLOW_THREADS;
  free(data);
  PyObject *ex = pytine_getExcObject(res);
  if (ex != NULL) return thrownTineException(ex, res, "setAttribute", attr,NULL,NULL);
  Py_RETURN_NONE;
}
static PyObject* pytine_getAttribute(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *attr, *typestr = NULL;
  PyObject *value = NULL;
  BYTE data[64];
  int format = CF_INT32;
  int size=1,res;
  static const char *kwlist[] = {
        "attribute",
        "format",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|s", (char**)kwlist, &attr, &typestr)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.setAttribute(attribute='str'[,format='str')");    
  }
  format = typestr == NULL ? getSysAttributeFormat(attr) : (int)GetFormatFromString(typestr);
  Py_BEGIN_ALLOW_THREADS;
  res = GetSystemAttribute(attr,data,format);
  if (res == 0) array_to_python(&value, data, size, format, NULL, FALSE);
  Py_END_ALLOW_THREADS;
  PyObject *ex = pytine_getExcObject(res);
  if (ex != NULL) return thrownTineException(ex, res, "getAttribute", attr, NULL, NULL);
  return value;
}

#define BLOBSIZE 65536
BYTE dataBlob[BLOBSIZE];
static PyObject* pytine_get(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *server, *property = NULL, *typestr = NULL, *modestr = NULL;
  int timeout = 1000, sz=0, use_bytearray=FALSE;
  int mode = CM_TIMER, access = CA_READ;
  PyObject *value = NULL;
  PyObject *dtype = NULL;
  static const char *kwlist[] = {
        "address",
        "property",
        "format",
        "size",
        "timeout",
        "mode",
        NULL
    };

  void *data;
  int size = 0, sizeBytes;
  int format = CF_NULL, fmt;
  int res;
  char *pc, tag[64];

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ssiis", (char**)kwlist, 
    &server, &property, &typestr, &size, &timeout, &modestr)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.get(address='str',property='str'[,format='str',size=val,timeout=val,mode='str')");    
  }
  if ((pc=strchr(server,'[')) != NULL) *pc++ = 0;
  if (property == NULL)
  {
    if (pc == NULL) return PyErr_Format(PyExc_ValueError, 
      "tine: address %s does not contain a property and no property provided as argument",server);
    property = pc;
    if ((pc=strchr(property,']')) != NULL) *pc = 0;
  }
  if (typestr != NULL) 
  {
    format = (int)GetFormatFromString(typestr);
    if (stristr(typestr,".BYTES") != NULL) use_bytearray = TRUE;
  }
  if (modestr != NULL) 
  {
    char mod[128];
    mode = (int)GetTransferMode(modestr); // just check CONNECT or NETWORK
    if (mode & CM_CONNECT) access |= CA_CONNECT;
    if (mode & CM_STREAM) access |= CA_STREAM;
    if (mode & CM_NETWORK) access |= CA_NETWORK;
    strncpy(mod,modestr,128); strupr(mod);
    if (strstr(mod,"WRITE") != NULL) access |= CA_WRITE;
    if (strstr(mod,"ASYNC") == NULL && strstr(mod,"SYNC") != NULL) access |= CA_SYNC;
  }
  if (acqInfo(size,format,typestr,tag,server))
  {
    sz = size;
    res = tine_getpropertyinformation(server, property, CA_READ, &format, &sz, tag);
    if (res != 0) return NULL;
    if (size == 0) size = sz;
  }
  else
  {
    const char *pdtag = use_bytearray ? "" : tag;
    if (getPrpItem(server,property) == NULL) addUsrPrpItem(server,property,size,format,(char *)pdtag,CA_READ);
  }
  fmt = (format == CF_MDA) ? GetFormatFromString(tag) : format;
  sizeBytes = fmtsizeof(fmt) * size;
  sizeBytes += getFormatHeaderSize(format);
  switch (format)
  {
     case CF_STRUCT:
     {
       int ssiz = GetStructSize(tag);
       if (ssiz > 0) sizeBytes = ssiz * size;
       break;
     }
     case CF_MDA:
       sizeBytes += sizeof(size_t);
       break;
     case CF_IMAGE:
     case CF_GSPECTRUM:
       sizeBytes += sizeof(int) + sizeof(size_t);
       break;
     case CF_STRING:
       sizeBytes = size * sizeof(char*);
       break;
  }
  data = calloc(1,sizeBytes);
  if (data == NULL)
  {
     return PyErr_Format(PyExc_MemoryError, "cannot allocate %d bytes",sizeBytes);
  }
  if (format == CF_IMAGE)
  { // make sure these parts of the 'embedded' DIMAGE are set properly
    ((DIMAGE *)data)->frameBufferSize = size;
    ((DIMAGE *)data)->frameBuffer = ((BYTE *)data) + getFormatHeaderSize(CF_IMAGE) + sizeof(int) + sizeof(size_t);
  }
  DTYPE d; DTYPEZERO(d);
  Py_BEGIN_ALLOW_THREADS;
  res = tine_read(server, property, data, &size, format, access, timeout, &d);
  Py_END_ALLOW_THREADS;
  int isErr = FALSE;
  if (HAS_DATA(res)) array_to_python(&value, data, size, format, d.dTag, use_bytearray); else isErr = TRUE;
  free(data);
  if (isErr)
  {
    PyObject *ex = pytine_getExcObject(res);
    if (ex != NULL) return thrownTineException(ex, res, "get", server, property, NULL);
  }
  double ts = d.dTimeStamp;
  dtype = Py_BuildValue("{s:i,s:d,s:i,s:i,s:s,s:O}",
    "status",res,"timestamp",ts,"sysstamp",d.sysStamp,"usrstamp",d.dStamp,"timestring",GetDataTimeString(ts,FALSE),"data",value);
  Py_CLEAR(value);
  return dtype;
}

static PyObject* pytine_set(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *server, *property, *typestr = NULL, *modestr = NULL;
  PyObject *callable = NULL, *value = NULL;
  int timeout = 1000;
  void *data;
  int size = 0;
  int format = CF_NULL, access = CA_WRITE;
  int res;
  char *pc, tag[64];
  static const char *kwlist[] = {
        "address",
        "property",
        "input",
        "format",
        "size",
        "timeout",
        "mode",
        "callback",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|sOsiisO",  (char**)kwlist, &server, &property, &value, &typestr, &size, &timeout, &modestr, &callable)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.set(address='str',property='str'[,input=obj,format='str',size=val,timeout=val,mode='str')");    
  }
  if ((pc=strchr(server,'[')) != NULL) *pc++ = 0;
  if (property == NULL)
  {
    if (pc == NULL) return PyErr_Format(PyExc_ValueError, 
      "tine: address %s does not contain a property and no property provided as argument",server);
    property = pc;
    if ((pc=strchr(property,']')) != NULL) *pc = 0;
  }
  if (typestr != NULL) format = (int)GetFormatFromString(typestr);
  if (modestr != NULL) 
  {
    int m = (int)GetTransferMode(modestr); // just check CONNECT or NETWORK
    if (m & CM_CONNECT) access |= CA_CONNECT;
    if (m & CM_STREAM) access |= CA_STREAM;
  }
  if (callable != NULL && !PyCallable_Check(callable))
  {
    return PyErr_Format(PyExc_ValueError, "tine: asynchronous set : %s", "callback object not callable");
  }
  if (value) 
  { // value given ...
    if (acqInfo(size,format,typestr,tag,server))
    { // TRUE => info not supplied: need to acquire the size and format from the server
      res = tine_getpropertyinformation(server, property, CA_WRITE, &format, &size, tag);
      if (res) return NULL;
      if (size == 0)
      { // not registered as a WRITEable with input
        size = array_length_from_python(value);
      }
      if (format == CF_NULL)
      { // not registered as a WRITEable with input: go with the READ settings
        int sz = size;
        tine_getpropertyinformation(server, property, CA_READ, &format, &sz, tag);
        if (format == CF_NULL)
        { // still came up empty; last chance:
          format = format_from_python(value);
        }
      }
    }
    else
    {
      if (getPrpItem(server,property) == NULL) addUsrPrpItem(server,property,size,format,tag,CA_WRITE);
    }
    /* After this call, size may be smaller as before, e.g. if a list
       contains fewer elements than the server accepts. */
    data = array_from_python(value, &size, format, tag);
    if (!data) return NULL;
    Py_BEGIN_ALLOW_THREADS;
    res = tine_write(server, property, data, size, format, tag, access, timeout,callable);
    Py_END_ALLOW_THREADS;
    free(data);
  } 
  else
  { // If no value was given, this is a trigger without data.
    res = tine_write(server, property, NULL, 0, CF_NULL, NULL, access, timeout,callable);
  }
  PyObject *ex = pytine_getExcObject(res);
  if (ex != NULL) return thrownTineException(ex, res, "set", server, property, NULL);
  Py_RETURN_NONE;
}

static PyObject* pytine_call(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *server, *property, *typeOutStr=NULL, *typeInStr=NULL, *modestr=NULL;
  int timeout = 1000, sz = 0;
  PyObject *value = NULL;
  PyObject *dtype = NULL;
  void *dataOut, *dataIn = NULL;
  int sizeOut = 0, sizeIn = 0, sizeBytes;
  int formatOut = CF_NULL, formatIn = CF_NULL, fmt, acc = CA_READ;
  int res, mode = CM_TIMER, access = CA_READ;
  int use_bytearray = FALSE, use_inptbytearray = FALSE;
  char *pc, tagOut[64] = "", tagIn[64] = "", mod[128];
  const char *pdtag, *pdintag;
  static const char *kwlist[] = {
        "address",
        "property",
        "input",
        "mode",
        "format",
        "size",
        "inputformat",
        "inputsize",
        "timeout",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|sOssisii", (char**)kwlist, &server, &property, 
    &value, &modestr, &typeOutStr, &sizeOut, &typeInStr, &sizeIn, &timeout)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.call(address='str',property='str'[,input=obj,format='str',size=val,inputformat='str',inputsize=val,timeout=val,mode='str')");    
  }
  if ((pc=strchr(server,'[')) != NULL) *pc++ = 0;
  if (property == NULL)
  {
    if (pc == NULL) return PyErr_Format(PyExc_ValueError, 
      "tine: address %s does not contain a property and no property provided as argument",server);
    property = pc;
    if ((pc=strchr(property,']')) != NULL) *pc = 0;
  }
  if (typeOutStr != NULL) 
  {
    formatOut = (int)GetFormatFromString(typeOutStr);
    if (stristr(typeOutStr,".BYTES") != NULL) use_bytearray = TRUE;
    if (formatOut == CF_STRUCT && sizeOut == 0) sizeOut = 1;
  }
  if (typeInStr != NULL)
  {
    formatIn = (int)GetFormatFromString(typeInStr);
    if (stristr(typeInStr, ".BYTES") != NULL) use_inptbytearray = TRUE;
    if (formatIn == CF_STRUCT && sizeIn == 0) sizeIn = 1;
  }
  if (modestr != NULL) 
  {
    mode = (int)GetTransferMode(modestr); // just check CONNECT or NETWORK
    if (mode & CM_CONNECT) access |= CA_CONNECT;
    if (mode & CM_STREAM) access |= CA_STREAM;
    if (mode & CM_NETWORK) access |= CA_NETWORK;
    strncpy(mod,modestr,128); strupr(mod);
    if (strstr(mod, "WRITE") != NULL)
    {
      access |= CA_WRITE;
      acc = CA_WRITE;
    }
    if (strstr(mod,"SYNC") != NULL) access |= CA_SYNC;
  }
  if (value)
  { // value given ...
    if (typeInStr == NULL) typeInStr = typeOutStr;
    if (acqInfo(sizeIn,formatIn,typeInStr,tagIn,server))
    {
      if (sizeIn == 0) sizeIn = array_length_from_python(value);
      sz = sizeIn;
      res = tine_getpropertyinformation(server, property, acc, &formatIn, &sz, tagIn);
      if (res != 0) return NULL;
      if (sizeIn == 0 || formatIn > CF_FLOAT) sizeIn = sz;
    }
    dataIn = array_from_python(value, &sizeIn, formatIn, tagIn);
    if (!dataIn) return NULL;
    if (formatIn == CF_STRUCT)
    {
      int ssiz = GetStructSize(tagIn);
      sizeBytes = ssiz > 1 ? ssiz * sizeIn : sizeIn;
      if (use_inptbytearray) sizeIn = sizeBytes;
    }
  }
  if (acqInfo(sizeOut,formatOut,typeOutStr,tagOut,server))
  {
    sz = sizeOut;
    res = tine_getpropertyinformation(server, property, CA_READ, &formatOut, &sz, tagOut);
    if (res != 0) return NULL;
    if (sizeOut == 0) sizeOut = sz;
  }
  pdintag = use_inptbytearray ? "" : tagIn;
  pdtag = use_bytearray ? "" : tagOut;
  if (getPrpItem(server,property) == NULL) addUsrPrpItemEx(server,property,sizeOut,formatOut,(char *)pdtag,sizeIn,formatIn,(char *)pdintag);
  fmt = (formatOut == CF_MDA) ? GetFormatFromString(tagOut) : formatOut;
  sizeBytes = fmtsizeof(fmt) * sizeOut;
  sizeBytes += getFormatHeaderSize(formatOut);
  if (formatOut == CF_STRUCT)
  {
    int ssiz = GetStructSize(tagOut);
    if (ssiz > 1) sizeBytes = ssiz * sizeOut;
    if (use_bytearray) sizeOut = sizeBytes;
  }
  if (formatOut == CF_MDA) sizeBytes += sizeof(size_t);
  dataOut = malloc(sizeBytes);
  if (dataOut == NULL)
  {
     return PyErr_Format(PyExc_MemoryError, "cannot allocate %d bytes",sizeBytes);
  }
  DTYPE dout, din; 
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = (UINT32)sizeOut; dout.dFormat = formatOut;
  strncpy(dout.dTag,pdtag,TAG_NAME_SIZE);
  dout.data.vptr = dataOut;
  din.dArrayLength = (UINT32)sizeIn; din.dFormat = formatIn;
  strncpy(din.dTag,pdintag,TAG_NAME_SIZE);
  din.data.vptr = dataIn;
  Py_BEGIN_ALLOW_THREADS;
  res = tine_acq(server, property, &dout, &din, access, timeout);
  Py_END_ALLOW_THREADS;
  int isErr = FALSE;
  if (HAS_DATA(res)) array_to_python(&value, dataOut, (int)dout.dArrayLength, formatOut, tagOut, use_bytearray); else isErr = TRUE;
  free(dataOut);
  if (isErr)
  {
    PyObject *ex = pytine_getExcObject(res);
    if (ex != NULL) return thrownTineException(ex, res, "call", server, property, NULL);
  }
  double ts = dout.dTimeStamp;
  dtype = Py_BuildValue("{s:i,s:d,s:i,s:i,s:s,s:O}",
    "status",res,"timestamp",ts,"sysstamp",dout.sysStamp,"usrstamp",dout.dStamp,"timestring",GetDataTimeString(ts,FALSE),"data",value);
  Py_CLEAR(value);
  return dtype;
}

static PyObject* pytine_hist(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *server, *property, *stop = NULL, *start = NULL, *depth = NULL, *flags = NULL;
  int cc, timeout = 1000, depthsecs = 1000;
  PyObject *value = NULL;
  void *data;
  PyObject *dtype = NULL;
  int num = 0, npoints = 1000, smpl = 0, indx = 0;
  int format = CF_NULL, fmt;
  int access = CA_READ|CA_RETRY|CA_MUTABLE;
  int acqSysStamp=FALSE, acqSnapshot=FALSE;
  time_t stopt=time(NULL), startt = stopt - 86400;
  char *pc, flgstr[128];
  static const char *kwlist[] = {
        "address",
        "property",
        "start",
        "stop",
        "depth",
        "flags",
        "timeout",
        "sample",
        "numberPoints",
        NULL
        };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|sssssiii", (char**)kwlist, &server, &property, 
    &start, &stop, &depth, &flags, &timeout,&smpl,&npoints)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.history(address='str',property='str',start='str',stop='str',depth='str',flags='str',timeout=val,sample=val,numberPoints=val)");    
  }
  if ((pc=strchr(server,'[')) != NULL) *pc++ = 0;
  if (property == NULL)
  {
    if (pc == NULL) return PyErr_Format(PyExc_ValueError, 
      "tine: address %s does not contain a property and no property provided as argument",server);
    property = pc;
    if ((pc=strchr(property,']')) != NULL) *pc = 0;
  }
  if (start != NULL) startt = parseTimeString(start);
  if (stop != NULL) stopt = parseTimeString(stop);
  if (depth != NULL) 
  {
    if (strstr(depth,"SNAP") != NULL || strstr(depth,"snap") != NULL)
    {
      acqSnapshot = TRUE;
      startt = stopt;
    }
    else
    {
      depthsecs = (int)parseDepthSting(depth);
      if (start == NULL) startt = stopt - depthsecs;
      else if (stop != NULL) stopt = startt + depthsecs;
    }
  }

  if (flags != NULL)
  {
    strncpy(flgstr,flags,128); strupr(flgstr);
    if (strstr(flgstr, "CONNECT")) access |= CA_CONNECT;
    else if (strstr(flgstr, "TCP")) access |= CA_CONNECT;
    if (strstr(flgstr,"SYSSTAMP") != NULL) acqSysStamp = TRUE;
    else if (strstr(flgstr,"SYSTEMSTAMP") != NULL) acqSysStamp = TRUE;
    else if (strstr(flgstr,"CYCLENUMBER") != NULL) acqSysStamp = TRUE;
    else if (strstr(flgstr,"PULSENUMBER") != NULL) acqSysStamp = TRUE;
    else if (strstr(flgstr,"EVENTNUMBER") != NULL) acqSysStamp = TRUE;
  }
  if (acqSnapshot)
  {
    cc = tine_getpropertyinformation(server, property, CA_READ, &format, &num, NULL);
    if (cc != 0) return NULL;
    data = calloc(1, num * fmtsizeof(format) + GetFormatHeaderSize(format));
    fmt = format;
  }
  else if (acqSysStamp)
  {
    num = npoints;
    data = calloc(npoints,sizeof(DBLDBLDBL));
    fmt = CF_DBLDBLDBL;
  }
  else
  {
    num = npoints;
    data = calloc(npoints,sizeof(DBLDBL));
    fmt = CF_DBLDBL;
  }
  if (data == NULL)
  {
     return PyErr_Format(PyExc_MemoryError, "cannot allocate memory for history call");
  }
  if (timeout < 100) timeout = 100;
  Py_BEGIN_ALLOW_THREADS;
  cc = getArchivedData(server,property,startt,&stopt,access,smpl,indx,fmt,data,&num,timeout);
  Py_END_ALLOW_THREADS;
  if (cc == 0) array_to_python(&value, data, num, fmt, NULL, FALSE);
  free(data);
  if (cc) 
  {
    PyObject *ex = pytine_getExcObject(cc);
    if (ex != NULL) return thrownTineException(ex, cc, "hist", server, property, NULL);
  }
  if (acqSnapshot)
  {
    dtype = Py_BuildValue("{s:s,s:d,s:O}","timestamp",GetDataTimeString((double)stopt,FALSE),"utc",(double)stopt,"data",value);
    Py_CLEAR(value);
  }
  else
  {
    dtype = value;
  }

  return dtype;
}

static PyObject* pytine_errorlist(PyObject* self, PyObject* args, PyObject* kwargs)
{
  char *text = NULL;
  PyObject *item = NULL;
  int code = -1;
  int len;
  char *c, errstr[64];
  static const char* kwlist[] = {
        "code",
        "text",
        NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|is", (char**)kwlist, &code, &text)) return NULL;
  int nerr = GetErrorListSize();
  PyObject* pdo;
  if (code >= 0)
  { // specific error requested
    if (code < nerr)
    {
      pdo = PyDict_New();
      strncpy(errstr, erlst[code], 32);
      for (c = errstr; *c != 0; c++) if (*c == ' ') *c = '_';
      len = (int)strlen(errstr) + 1;
      item = data_to_python(errstr, len, CF_STRING, NULL);
      PyDict_SetItemString(pdo, "text", item);
      PyDict_SetItemString(pdo, "code", Py_BuildValue("i", code));
      return pdo;
    }
    return PyErr_Format(PyExc_IOError, "code %d not registered", code);
  }
  if (text != NULL)
  {
    // specific error requested
    strncpy(errstr, text, 32);
    for (c = errstr; *c != 0; c++) if (*c == '_') *c = ' ';
    for (int i = 0; i < nerr; i++) if (strnicmp(errstr, erlst[i], 32) == 0)
    {
      pdo = PyDict_New();
      len = (int)strlen(text) + 1;
      item = data_to_python(text, len, CF_STRING, NULL);
      PyDict_SetItemString(pdo, "text", item);
      PyDict_SetItemString(pdo, "code", Py_BuildValue("i", i));
      return pdo;
    }
    return PyErr_Format(PyExc_IOError, "code for %s not found", text);
  }

  PyObject *o = PyList_New(nerr);
  for (int i = 0; i < nerr; i++)
  {
      strncpy(errstr, erlst[i], 32);
      for (c = errstr; *c != 0; c++) if (*c == ' ') *c = '_';
      len = (int)strlen(errstr) + 1;
      item = data_to_python(errstr, len, CF_STRING, NULL);
      pdo = PyDict_New();
      PyDict_SetItemString(pdo, "text", item);
      PyDict_SetItemString(pdo, "code", Py_BuildValue("i", i));
      PyList_SetItem(o, i, pdo);
  }
  return o;
}
char *sevlst[] = {(char*)"INFO",(char*)"WARN",(char*)"ERROR",(char*)"FATAL"};
static PyObject *pytine_log(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *msg = NULL, *fname = NULL, *tag = NULL;
  int cc = 0, sev = 0;
  static const char *kwlist[] = {
        "message",
        "file",
        "severity",
        "tag",
        NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|sis", (char **)kwlist, &msg, &fname, &sev, &tag)) return NULL;
  if (sev < 0) sev = 0;
  if (sev > 3) sev = 3;
  if (fname == NULL)
  {
    cc = feclogEx(sev,(char*)"%s",msg);
  }
  else
  {
    if (tag == NULL) tag = sevlst[sev];
    cc = srvlog(fname, tag, msg);
  }
  return Py_BuildValue("i", cc);
}
static PyObject *pytine_list(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *context=NULL, *server=NULL, *property=NULL, *device=NULL, *structtag=NULL, dev[132], tag[64];
  const int LABEL_SIZE = 16;
  char lbl1[LABEL_SIZE+1] = "", lbl2[LABEL_SIZE+1] = "";
  PyObject *value = NULL, *value2 = NULL;
  PyObject *dtype = NULL;
  NAME64 *n64 = NULL;
  NAME32 *n32 = NULL;
  int num;
  int cc, nlists = 1;
  static const char *kwlist[] = {
        "context",
        "server",
        "device",
        "property",
        "structure",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|sssss", (char**)kwlist, &context, &server, &device, &property, &structtag)) return NULL;
  if (context == NULL || strlen(context) == 0) 
  { // no context given -> get the available contexts
    num = 1024;
    if ((n32=(NAME32 *)calloc(num,32)) == NULL) return NULL;
    strncpy( lbl1, "contexts", LABEL_SIZE);
    Py_BEGIN_ALLOW_THREADS;
    cc = GetSystemContexts(n32,&num);
    Py_END_ALLOW_THREADS;
    if (!HAS_DATA(cc))
    {
      PyObject *ex = pytine_getExcObject(cc);
      if (ex != NULL) return thrownTineException(ex, cc, "list", NULL, NULL, NULL);
    }
    array_to_python(&value, n32, num, CF_NAME32, NULL, FALSE);
  }
  else if (server == NULL || strlen(server) == 0)
  { // no server given -> get the available servers for the context given
    if (!stricmp(context, "attributes"))
    { // special query for system attributes list
      getSysAttributesList();
      if ((n32 = (NAME32 *)calloc(gNumSysAttrbutes, 32)) == NULL) return NULL;
      strncpy(lbl1, "attributes", LABEL_SIZE);
      for (int i = 0; i < gNumSysAttrbutes; i++)
      {
        strncpy(n32[i].name, gSysAttributesList[i].name, 32);
      }
      array_to_python(&value, n32, gNumSysAttrbutes, CF_NAME32, NULL, FALSE);
      goto done;
    }
    num = 1024;
    if ((n32=(NAME32 *)calloc(num,32)) == NULL) return NULL;
    strncpy( lbl1, "servers", LABEL_SIZE);
    Py_BEGIN_ALLOW_THREADS;
    cc = GetSystemServers(context,n32,&num);
    Py_END_ALLOW_THREADS;
    if (!HAS_DATA(cc))
    {
      PyObject *ex = pytine_getExcObject(cc);
      if (ex != NULL) return thrownTineException(ex, cc, "list", context, NULL, NULL);
    }
    array_to_python(&value, n32, num, CF_NAME32, NULL, FALSE);
  }
  else
  { // directed query for properties or devices
    sprintf(dev,"/%.32s/%.32s",context,server);
    if (structtag != NULL && strlen(structtag) > 0)
    { // here a query for structure fields
      if (GetStructFromTag(structtag) == NULL) AcquireAndRegisterStruct(dev,structtag,10);
      goto struct_query;
    }
    num = 2048;
    if ((n64=(NAME64 *)calloc(num,64)) == NULL) return NULL;
    if (device != NULL && strlen(device) > 0) 
    { // a device was specified -> get the targeted properties
      strcat(dev,"/"); strcat(dev,device);
      goto prps_query;
    }
    Py_BEGIN_ALLOW_THREADS;
    cc = GetSystemDevices(dev,property,n64,&num);
    Py_END_ALLOW_THREADS;
    if (!HAS_DATA(cc)) 
    {
      PyObject *ex = pytine_getExcObject(cc);
      if (ex != NULL) return thrownTineException(ex, cc, "list", context, server, property);
    }
    if (cc != 0)
    { // is property oriented
      goto prps_query;
    }
    else
    {
      strncpy( lbl1, "devices", LABEL_SIZE);
      array_to_python(&value, n64, num, CF_NAME64, NULL, FALSE);
    }
prps_query:
    if (property != NULL && strlen(property) > 0) 
    { // a property was specified
      if (strlen(lbl1) == 0)
      { // and a device was also specified !
        int format = CF_NULL;
        cc = tine_getpropertyinformation(server, property, CA_READ, &format, &num, tag);
        if (format == CF_STRUCT)
        {
          structtag = tag;
          goto struct_query;
        }
      }
      goto done; 
    }
    num = 2048;
    Py_BEGIN_ALLOW_THREADS;
    cc = GetSystemProperties(dev,n64,&num);
    Py_END_ALLOW_THREADS;
    if (!HAS_DATA(cc))
    {
      PyObject *ex = pytine_getExcObject(cc);
      if (ex != NULL) return thrownTineException(ex, cc, "list", context, server, device);
    }
    if (strlen(lbl1) == 0)
    { // no device list
      array_to_python(&value, n64, num, CF_NAME64, NULL, FALSE);
      strncpy( lbl1, "properties", LABEL_SIZE);
    }
    else 
    { // there is a device list 
      if (cc != 0) goto done; // property query function-> just dump the devices ...
      array_to_python(&value2, n64, num, CF_NAME64, NULL, FALSE);
      strncpy( lbl2, "properties", LABEL_SIZE);
      nlists = 2;
    }
    goto done;
struct_query:
    struct structStruct *ss = GetStructFromTag(structtag);
    if (ss == NULL) return PyErr_Format(PyExc_TypeError, "structure %s is not registered",structtag);
    struct structFormat *sf;
    int i, nfields = 0;
    for (sf=ss->s; sf != NULL; sf=sf->nxt) if (sf->fmt != CF_NULL) nfields++;
    if ((n64=(NAME64 *)calloc(nfields,64)) == NULL)
    {
      return PyErr_Format(PyExc_MemoryError, "cannot allocated structure space");
    }
    for (sf=ss->s, i=0; sf != NULL && i<nfields; sf=sf->nxt, i++) 
    {
      if (sf->fmt == CF_NULL) continue;
      strncpy(n64[i].name,sf->field,FIELD_NAME_SIZE);
    }
    array_to_python(&value, n64, nfields, CF_NAME64, NULL, FALSE);
    strncpy( lbl1, "fields", LABEL_SIZE);
  }
done:
  if (n32 != NULL) free(n32);
  if (n64 != NULL) free(n64);
  if (nlists == 1 && strlen(lbl1) != 0)
  {
    dtype = Py_BuildValue("{s:O}",lbl1,value);
    Py_CLEAR(value);
  }
  else if (strlen(lbl1) != 0 && strlen(lbl2) != 0)
  {
    dtype = Py_BuildValue("{s:O,s:O}",lbl1,value,lbl2,value2);
    Py_CLEAR(value); Py_CLEAR(value2);
  }
  if (dtype == NULL)
    return PyErr_Format(PyExc_AttributeError, "nothing to list (arguments fully specified)!");
  return dtype;
}

static PyObject* pytine_tty(PyObject *self, PyObject *args) 
{
  char *cmd;
  if (!PyArg_ParseTuple(args, "s", &cmd)) return NULL;

  if (strlen(cmd) > 0) SystemInterpretConsoleCommand(cmd);

  Py_RETURN_NONE;
}
/**
 * import tine
 * valuelist = tine.query(server, property, value, timeout = 10000)
 */
static PyObject* pytine_query(PyObject *self, PyObject *args) 
{
  char *server, *property;
  PyObject *value_in = NULL;
  PyObject *value_out = NULL;
  long timeout = 1000;

  void *data_in;
  int size_in;
  int format_in;
  void *data_out = NULL;
  int size_out;
  int format_out;
  int res;
  char tagOut[64], tagIn[64];

  if (!PyArg_ParseTuple(args, "ssO|l", &server, &property, &value_in, &timeout)) return NULL;

  res = tine_getpropertyinformation(server, property, CA_WRITE, &format_in, &size_in, tagIn);
  if (res) return NULL;

  res = tine_getpropertyinformation(server, property, CA_READ, &format_out, &size_out, tagOut);
  if (res) return NULL;

  /* After this call, size may be smaller as before, e.g. if a list
     contains less elements than the server accepts. */
  data_in = array_from_python(value_in, &size_in, format_in, NULL);
  if (!data_in) return NULL;

  data_out = malloc(fmtsizeof(format_out) * size_out);

  Py_BEGIN_ALLOW_THREADS;
  res = tine_query(server, property, data_in, size_in, format_in, 
		               data_out, &size_out, &format_out, timeout);
  Py_END_ALLOW_THREADS;

  if (res != 0) 
  {
    free(data_in);
    free(data_out);
    PyObject *ex = pytine_getExcObject(res);
    if (ex != NULL) return thrownTineException(ex, res, "query", server, property, NULL);
  }

  array_to_python(&value_out, data_out, size_out, format_out, tagOut, FALSE);

  free(data_in);
  free(data_out);

  return value_out;
}

typedef struct pyMonLnk
{
  char key[200];
  int lid; // link id
  int cbid; // callback id
  struct callback *cb;
  struct pyMonLnk *nxt;
} PyMonLnk; // Python monitor link
PyMonLnk *monLnks = NULL;

PyMonLnk *getMonLnk(int lid)
{
  PyMonLnk *lnk;
  for (lnk=monLnks; lnk != NULL; lnk=lnk->nxt)
  {
    if (lnk->lid == lid) return lnk;
  }
  return NULL;
}

void addMonLnk(int lid,int cbid,struct callback *cb,char *key)
{
  PyMonLnk *lnk;
  for (lnk=monLnks; lnk != NULL; lnk=lnk->nxt)
  {
    if (lnk->lid == lid) return; // already in list
  }
  if ((lnk=(PyMonLnk *)calloc(1,sizeof(PyMonLnk))) == NULL) return;
  lnk->lid = lid;
  lnk->cbid = cbid;
  lnk->cb = cb;
  strncpy(lnk->key,key,200);
  lnk->nxt = monLnks;
  monLnks = lnk;
  return;
}
void rmvMonLnk(int lid)
{
  PyMonLnk *lnk, *lnkp=NULL, *lnkn=NULL;
  for (lnk=monLnks; lnk != NULL; lnk=lnk->nxt)
  {
    lnkn = lnk->nxt;
    if (lnk->lid == lid)
    { // remove this one
      if (lnk == monLnks) monLnks = lnkn; else lnkp->nxt = lnkn;
      if (lnk->cb != NULL)
      {
        if (lnk->cb->data != NULL) free(lnk->cb->data);
        Py_XDECREF(lnk->cb->callable);
        free(lnk->cb);
        lnk->cb = NULL;
      }
      break;
    }
    lnkp = lnk;
  }
}
/*
static void _callback_handler(int id, int cc, void *p) 
{
  PyGILState_STATE gstate;

  struct callback *cb = (struct callback*)p;
  PyObject *data = NULL, *result = NULL;
  PyMonLnk *lnk = getMonLnk(id);
  int cbid = lnk != NULL ? lnk->cbid : id;
  gstate = PyGILState_Ensure();

  array_to_python(&data, cb->data, (int)cb->size, cb->format, cb->tag, FALSE);

	PyObject *args = Py_BuildValue("(iiN)", cbid, cc, data);
	PyObject *kwargs = Py_BuildValue("",  NULL);
	result = PyObject_Call(cb->callable, args, kwargs);

	Py_XDECREF(args);
	Py_XDECREF(kwargs);
	Py_XDECREF(data);
	if (result != NULL) Py_XDECREF(result);

  PyGILState_Release(gstate);
}
*/
static void _datacallback_handler(int id, int cc, void *p)
{
  PyGILState_STATE gstate;
  gstate = PyGILState_Ensure();
  struct callback *cb = (struct callback *)p;
  PyObject *data = NULL, *dtype = NULL, *result = NULL;
  PyMonLnk *lnk = getMonLnk(id);
  int cbid = lnk != NULL ? lnk->cbid : id;
  DTYPE d; DTYPEZERO(d);
  UpdateDataStampsFromLinkId(id, &d);
  double ts = d.dTimeStamp;
  //gstate = PyGILState_Ensure();
  if (cb->data != NULL && cb->format != CF_NULL)
  {
    int csiz = GetCompletionDataSize(id);
    if (csiz <= 0) csiz = (int)cb->size;
    array_to_python(&data, cb->data, csiz, cb->format, cb->tag, cb->use_bytearray); // always fill in regardless of cc
  }
  else
  {
    data = Py_BuildValue("");
  }
  dtype = Py_BuildValue("{s:i,s:d,s:i,s:i,s:s,s:O}",
    "status",cc,"timestamp",ts,"sysstamp",d.sysStamp,"usrstamp",d.dStamp,"timestring",GetDataTimeString(ts,FALSE),"data",data);

  PyObject *args = Py_BuildValue("(iiN)", cbid, cc, dtype);
  result = PyObject_Call(cb->callable, args, NULL);
	Py_XDECREF(args);
	Py_XDECREF(data);
	Py_XDECREF(dtype);
	if (result != NULL) Py_XDECREF(result);  else PyErr_SetString(PyExc_TypeError, "unable to call data callback handler");
  if (lnk == NULL)
  { // not monitored : a one-shot asynchronous transaction ...
    if (cb->data != NULL) free(cb->data);
    Py_XDECREF(cb->callable);
    free(cb);
  }
  PyGILState_Release(gstate);
}
static void _groupcallback_handler(int id, int cc, void *p) 
{
  PyGILState_STATE gstate;
  PyObject *data = NULL, *dtype = NULL, *result = NULL;
  GrpTblEntry *g=GetCallbackGroup((size_t)_groupcallback_handler);
  GrpMember *mbr;
  PyMonLnk *lnk;
  struct callback *cb = (struct callback*)p;
  int lid, csiz, mcc, n;
  double ts;
  DTYPE d; DTYPEZERO(d);
  if (g == NULL)
  { //TODO: throw an exception?  This shouldn't be possible
    return;
  }
  g->members = GetGroupMemberList(g);
  gstate = PyGILState_Ensure();
  PyObject *gtype = PyList_New(g->numInGrp);
  for (mbr=g->members,n=0; mbr != NULL && n < g->numInGrp; mbr=mbr->nxt, n++)
  {
    lid = mbr->conTblId;
    if ((lnk=getMonLnk(lid)) == NULL) continue;
    cb = lnk->cb;
    csiz = GetCompletionDataSize(lid);
    mcc = GetCurrentLinkStatus(lid);
    if (HAS_DATA(mcc)) array_to_python(&data, cb->data, csiz, cb->format, cb->tag, cb->use_bytearray);
    UpdateDataStampsFromLinkId(lid,&d);
    ts = d.dTimeStamp;
    dtype = Py_BuildValue("{s:i,s:s,s:i,s:d,s:i,s:i,s:O}",
      "id",lnk->cbid,"key",lnk->key,"status",cc,"timestamp",ts,
      "sysstamp",d.sysStamp,"usrstamp",d.dStamp,"data",data);
    PyList_SetItem(gtype, n, dtype);
  }
  PyObject *pdo = PyDict_New();
  PyDict_SetItemString(pdo,"members",gtype);
  PyDict_SetItemString(pdo,"synchronized",Py_BuildValue("i",(g->grpSyncLevel == GRP_SYNC_INSYNC?1:0)));
  PyDict_SetItemString(pdo,"timestring", Py_BuildValue("s",GetDataTimeString(ts,FALSE)));
  PyDict_SetItemString(pdo,"timestamp", Py_BuildValue("d",ts));
  PyDict_SetItemString(pdo,"systemstamp", Py_BuildValue("i",g->grpSysStamp));
	PyObject *args = Py_BuildValue("(iiN)", g->grpId, g->grpCc, pdo);
	result = PyObject_Call(cb->callable, args, NULL);

	Py_XDECREF(args);
	Py_XDECREF(data);
	Py_XDECREF(dtype);
	if (result != NULL) Py_XDECREF(result);  else PyErr_SetString(PyExc_TypeError, "unable to call group callback handler");

  PyGILState_Release(gstate);
}

static int acqInfo(int size,int format,char *typestr,char *tag,char *srv)
{
  if (size == 0) return TRUE;
  switch (format)
  {
    case CF_NULL:
    case CF_DEFAULT:
      return TRUE;
      break;
    case CF_STRUCT:
      if (stristr(typestr, ".BYTES") != NULL)
      {
        char *c = strchr(typestr, '.');
        if (stricmp(c, ".BYTES"))
        { // then it's of the form "STRUCT.tag.BYTES"
          char tmp[TAG_NAME_SIZE+4];
          strncpy(tmp, ++c, TAG_NAME_SIZE);
          if ((c = strchr(tmp, '.')) != NULL) *c = 0;
          if (GetStructFromTag(tmp) != NULL)
          { // and this tag is registered !
            strncpy(tag, tmp, TAG_NAME_SIZE);
          }
        }
        break;
      }
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      {
        char *tagstr = strchr(typestr,'.');
        if (tagstr != NULL) strncpy(tag,++tagstr,TAG_NAME_SIZE); else return TRUE;
        if (GetStructFromTag(tag) == NULL) AcquireAndRegisterStruct(srv,tag,size);
        break;
      }
    default:
      break;
  }
  return FALSE;
}
static PyObject* pytine_attach(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *server, *property, *modestr = NULL, *tolstr = NULL, *accstr = NULL;
  PyObject *callable, *value = NULL;
  int id = 0, interval = 1000, sz = 0, access = CA_READ;
  struct callback *cb;
  int formatOut = CF_NULL, sizeOut = 0, mode = CM_TIMER, use_bytearray = FALSE;
  int formatIn = CF_NULL, sizeIn = 0;
  char *pc, *typestrOut = NULL, *typestrIn = NULL, tagOut[64], tagIn[64] = "";
  void *dataIn = NULL;
  int cc=0, lid;
  static const char *kwlist[] = {
        "address",
        "property",
        "callback",
        "input",
        "format",
        "size",
        "inputformat",
        "inputsize",
        "interval",
        "mode",
        "id",
        "tolerance",
        "access",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ssO|Osisiisiss", (char**)kwlist,
    &server, &property, &callable, 
    &value, &typestrOut, &sizeOut, &typestrIn, &sizeIn, &interval, &modestr, &id, &tolstr, &accstr)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.attach(address='str',property='str',callback=obj[,input=obj,format='str',size=val,inputformat='str',inputsize=val,interval=val,mode='str',id=val,tolerance='str')");    
  }
  if ((pc=strchr(server,'[')) != NULL) *pc++ = 0;
  if (property == NULL)
  {
    if (pc == NULL) return PyErr_Format(PyExc_ValueError, 
      "tine: address %s does not contain a property and no property provided as argument",server);
    property = pc;
    if ((pc=strchr(property,']')) != NULL) *pc = 0;
  }
  if (!PyCallable_Check(callable))
    return PyErr_Format(PyExc_ValueError, "%s: attach('%s', '%s'): %s",
                        "tine", server, property, "object not callable");
  if (value) 
  { // value given ...
    if (typestrIn == NULL) typestrIn = typestrOut;
    if (acqInfo(sizeIn,formatIn,typestrIn,tagIn,server))
    {
      sz = sizeIn;
      cc = tine_getpropertyinformation(server, property, CA_WRITE, &formatIn, &sz, tagIn);
      if (cc != 0) return NULL;
      if (sizeIn == 0) sizeIn = sz;
    }
    if (sizeIn == 0) 
    { // value give but info query didn't deliver ....
      sizeIn = 1; formatIn = CF_INT32;
    }
    dataIn = array_from_python(value, &sizeIn, formatIn, tagIn);
    if (!dataIn) return NULL;
  }
  if (typestrOut != NULL) 
  {
    formatOut = (int)GetFormatFromString(typestrOut);
    if (stristr(typestrOut,".BYTES") != NULL) use_bytearray = TRUE;
  }
  if (modestr != NULL) 
  {
    mode = (int)GetTransferMode(modestr);
  }
  if (accstr != NULL)
  {
    access |= (int)GetAccessMode(accstr);
    if (access & CA_WRITE)
    {
      mode |= CM_SINGLE;
      mode &= ~(CM_TIMER|CM_DATACHANGE|CM_EVENT);
    }
  }
  if (tolstr != NULL && strlen(tolstr) > 0) mode |= CM_NODUPLICATES;
  int isGrouped = (mode & CM_GROUPED || mode & CM_SYNCGROUP) ? TRUE : FALSE;
  if (acqInfo(sizeOut,formatOut,typestrOut,tagOut,server))
  {
    sz = sizeOut;
    cc = tine_getpropertyinformation(server, property, CA_READ, &formatOut, &sz, tagOut);
    if (cc != 0) return NULL;
    if (sizeOut == 0) sizeOut = sz;
  }
  else
  {
    if (getPrpItem(server,property) == NULL) addUsrPrpItem(server,property,sizeOut,formatOut,tagOut,CA_READ);
  }
  int sizeBytes = fmtsizeof(formatOut) * sizeOut;
  sizeBytes += getFormatHeaderSize(formatOut);
  if (formatOut == CF_STRUCT)
  {
    int ssiz = GetStructSize(tagOut);
    if (ssiz > 0) sizeBytes = ssiz * sizeOut;
  }
  if (formatOut == CF_IMAGE) sizeBytes += sizeof(int) + sizeof(size_t);

  if ((cb=(struct callback *)malloc(sizeof(struct callback))) == NULL) 
    return PyErr_Format(PyExc_MemoryError, "cannot allocate %d bytes for callback",sizeof(struct callback));
  cb->callable = callable;
  cb->data     = malloc(sizeBytes);
  cb->size     = sizeOut;
  cb->format   = formatOut;
  cb->mode     = mode;
  cb->use_bytearray = use_bytearray;
  strncpy(cb->tag,tagOut,TAG_NAME_SIZE);

  if (cb->data == NULL) 
  {
    free(cb);
    return PyErr_Format(PyExc_MemoryError, "cannot allocate %d bytes for data",sizeBytes);
  }
  if (formatOut == CF_IMAGE)
  { // make sure these parts of the 'embedded' DIMAGE are set properly
    ((DIMAGE *)cb->data)->frameBufferSize = sizeOut;
    ((DIMAGE *)cb->data)->frameBuffer = ((BYTE *)cb->data) + getFormatHeaderSize(CF_IMAGE) + sizeof(int) + sizeof(size_t);
  }
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = (UINT32)sizeOut;
  dout.dFormat = formatOut;
  strncpy(dout.dTag,tagOut,TAG_NAME_SIZE);
  dout.data.vptr = cb->data;
  din.dArrayLength = sizeIn;
  din.dFormat = formatIn;
  strncpy(din.dTag,tagIn,TAG_NAME_SIZE);
  din.data.vptr = dataIn;
  /* Increase the refernce counter to let the Python know,
     that we need to keep this around. */
  Py_INCREF(cb->callable); // decrement this when we remove the monitor link
  XTCBFCNP cbfcn = isGrouped ? _groupcallback_handler : _datacallback_handler;
  Py_BEGIN_ALLOW_THREADS;
  lid = tine_attach(server, property, &dout, &din, access, cb->mode, interval,
                    cbfcn, cb, UNASSIGNED_CALLBACKID);
  Py_END_ALLOW_THREADS;

  if (lid < 0) 
  {
    cc = -lid;
#   if PY_MAJOR_VERSION >= 3
    PyErr_Format(PyExc_ConnectionError, "cannot attach to /%s[%s] (%.32s)",server,property,cc2str(cc));
#   else
    PyErr_Format(PyExc_IOError, "cannot attach to /%s[%s] (%.32s)", server, property, cc2str(cc));
#   endif
  }
  else if (tolstr != NULL)
  {
    char tolStr[128];
    float tolAbs = 0, tolRel = 0, fv;
    strncpy(tolStr,tolstr,128); tolStr[127] = 0;
    if ((pc=strchr(tolStr,'+')) != NULL)
    {
      *pc++ = 0;
      fv = (float)strtod(pc,NULL);
      if (strchr(pc,'%') != NULL) tolRel = (float)(fv/100.0); else tolAbs = fv;
    }
    fv = (float)strtod(tolStr,NULL);
    if (strchr(tolStr,'%') != NULL) tolRel = (float)(fv/100.0); else tolAbs = fv;
    SetNotificationTolerance(lid, tolAbs, tolRel);
  }
  char key[200];
  sprintf(key,"%.132s[%.64s]",server,property);
  addMonLnk(lid,id,cb,key);
  return Py_BuildValue("i", lid);
}

/**
 * import tine
 * linkid = tine.attach(server, property, callable, id = 0, timeout = 10000)
 * # ...
 * tine.detach(linkid)
 */
static PyObject* pytine_detach(PyObject *self, PyObject *args) 
{
  int id;
  char errstr[128];

  if (!PyArg_ParseTuple(args, "i", &id)) return NULL;
  int res = tine_detach(id);
  rmvMonLnk(id);
  if (res != 0) return PyErr_Format(PyExc_IOError, "%s: detach(%d): %s",
                        "tine", id, GetLastLinkError(res,errstr));
  Py_RETURN_NONE;
}

static PyObject* pytine_strerror(PyObject *self, PyObject *args) 
{
  int cc;
  char errstr[128];
  if (!PyArg_ParseTuple(args, "i", &cc)) return NULL;
  return Py_BuildValue("s",  GetLastLinkError((short)cc,errstr));
}
static PyObject* pytine_version(PyObject *self, PyObject *args) 
{
  return PyBytes_FromFormat("TINE %s, PyTine %s", GetSystemVersionString(), PYTINE_VERSION);
}

static PyObject* pytine_register_fec(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *fec=NULL, *dsc=NULL, *loc=NULL, *rsp=NULL;
  const char *sub="MISC", *ctx="TEST", *hdw="none";
  int po = -1;
  int cc=0;
  static const char *kwlist[] = {
        "name",
        "subsystem",
        "context",
        "description",
        "location",
        "hardware",
        "responsible",
        "port",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ssssssi", (char**)kwlist, &fec, &sub, &ctx, &dsc, &loc, &hdw, &rsp, &po)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.register_fec(name='str'[,subsystem='str',context='str',descripton='str',location='str',hardware='str',responsible='str',port=val])");    
  }
  if (po < 0) po = GetPortOffset(fec);
  cc = RegisterFecInformation(fec,(char*)sub,(char*)ctx,dsc,loc,(char*)hdw,rsp,po);
  return Py_BuildValue("i", cc);
}
static PyObject *pytine_clearAlarm(PyObject *self, PyObject *args, PyObject *kwargs)
{
  int cc = 0, nr = -1, code = 0;
  char *dev=NULL;
  static const char *kwlist[] = {
        "code",
        "device",
        "devicenumber",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|isi", (char**)kwlist, &code, &dev, &nr)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.clearalarm(code=val[,device='str',devicenumber=val,options='str'])");    
  }
  if (nr == -1 && dev != NULL) nr = GetDeviceNumber(gSrvEqm,dev);
  if (dev != NULL && stricmp(dev, "ALL") == 0) nr = -1;
  if (code != 0 && nr < 0) ccerr(invalid_parameter);
  if (code == 0) ClearAlarm(gSrvEqm,nr); else RemoveAlarm(gSrvEqm,nr,code);
err:
  if (cc != 0)
  {
    char errstr[128];
    return PyErr_Format(PyExc_RuntimeError,"PyTine.clearalarm() cannot clear specific alarm for non-specified device! : %s",GetLastLinkError(cc,errstr));
  }
  return Py_BuildValue("i", cc);
}
static PyObject *pytine_setAlarm(PyObject *self, PyObject *args, PyObject *kwargs)
{
  PyObject *value = NULL;
  char *dev=NULL, *optstr=NULL;
  int cc = 0, nr = 0, code = 0, flags = almNEWALARM;
  static const char *kwlist[] = {
        "code",
        "device",
        "devicenumber",
        "data",
        "options",
        NULL
    };
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|siOs", (char**)kwlist, &code, &dev, &nr, &value, &optstr)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.setalarm(code=val[,device='str',devicenumber=val,data=Object,options='str'])");    
  }
  BYTE *data = NULL;
  if (value != NULL)
  {
    ADS ads;
    if ((cc=GetAlarmDefinition(gSrvEqm,code,&ads)) == 0)
    {
      int sz = (int)ads.alarmDataArraySize;
      int fmt = (int)LFMT(ads.alarmDataFormat);
      data = (BYTE *)array_from_python(value, &sz, fmt, NULL);
      if (!data) ccerr(invalid_data);
    }
  }
  if (nr == 0 && dev != NULL) nr = GetDeviceNumber(gSrvEqm,dev);
  if (optstr != NULL)
  {
    char s[128];
    strncpy(s,(const char *)optstr,128);
    if (strstr((const char *)s,"TRANSIENT") != NULL) flags |= almTRANSIENT;
  }
  SetAlarmEx(gSrvEqm,nr,code,data,flags);
err:
  if (cc != 0)
  {
    char errstr[128];
    return PyErr_Format(PyExc_RuntimeError,"PyTine.setalarm() : %s",GetLastLinkError(cc,errstr));
  }
  return Py_BuildValue("i", cc);
}
static PyObject* pytine_register_server(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *c, *srv=NULL, *fec=NULL, expstr[128], fecstr[64];
  const char *ctx="TEST", *eqm="PYEQM", *sub="MISC", *hdw = "none";
  char *dsc = NULL, *loc = NULL, *rsp = NULL;
  int sz = 1000, po = 0;
  int cc=0;
  static const char *kwlist[] = {
        "name",
        "context",
        "eqm",
        "capacity",
        "fec",
        "subsystem",
        "description",
        "location",
        "hardware",
        "responsible",
        "port",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ssissssssi", 
       (char**)kwlist, &srv, &ctx, &eqm, &sz, &fec, &sub, &dsc, &loc, &hdw, &rsp, &po)) 
  {
     return PyErr_Format(PyExc_SyntaxError, 
       "PyTine.register_server(name='str'[,context='str',eqm='str',capacity=val,fec='str',subsystem='str',description='str',location='str',hardware='str',responsible='str',port=val])");    
  }
  if (!strcmp(GetExportedFecName(),FECNAME_UNKNOWN))
  { // fec name not yet known ...
    if (fec == NULL)
    { // and not specified ...
      memset(fecstr,0,64);
      strncpy(fecstr,ctx,2);
      if ((c=strchr((char*)ctx,'.')) != NULL) strncat(fecstr,++c,2);
      else if ((c=strchr((char*)ctx,'_')) != NULL) strncat(fecstr,++c,1);
      strncat(fecstr,srv,32);
      while ((c=strchr(fecstr,'.')) != NULL || (c=strchr(fecstr,'_')) != NULL || (c=strchr(fecstr,' ')) != NULL)
      { // remove any funny characters ...
        *c++ = 0;
        strcat(fecstr,c);
      }
      fecstr[16] = 0;
      fec = fecstr;
    }
    if (po <= 0) po = GetPortOffset(fec);
    cc = RegisterFecInformation(fec,(char*)sub,(char*)ctx,dsc,loc,(char*)hdw,rsp,po);
  }
  sprintf(expstr,"%.32s/INIT=SRV|KEEP",srv);
  cc = AttachServerEx(expstr,(char*)eqm,sz,NULL,0);
  strncpy(gSrvEqm,eqm,EQM_NAME_SHORTSIZE);
  return Py_BuildValue("i", cc);
}
static PyObject* pytine_register_device(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *nam=NULL, *dsc=NULL, *loc=NULL, *rdr=NULL;
  PyObject *prpLst = NULL;
  float zpos = 0;
  int nr = 0;
  int cc=0;
  static const char *kwlist[] = {
        "name",
        "number",
        "description",
        "location",
        "zposition",
        "redirection",
        "propertylist",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "si|ssfsO", (char**)kwlist, &nam, &nr, &dsc, &loc, &zpos, &rdr, &prpLst)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.register_device(name='str',number=val[,description='str',location='str',zposition=val,redirection='str',propertylist=Dict])");    
  }
  cc = RegisterDeviceName(gSrvEqm,nam,nr);
  if (cc == 0)
  {
     if (dsc != NULL) RegisterDeviceDescription(gSrvEqm,nam,dsc);
     if (loc != NULL) SetDeviceLocation(gSrvEqm,nam,loc);
     if (zpos != 0) SetDeviceZPosition(gSrvEqm,nam,zpos);
     if (prpLst != NULL)
     {
       if (!PyDict_Check(prpLst))
       {
         return PyErr_Format(PyExc_TypeError, "dictionary expected for property list!");
       }
       if (PyDict_Size(prpLst) < 1)
       {
         return PyErr_Format(PyExc_TypeError, "property list dictionary is empty !");
       }
       PyObject *key, *value;
       Py_ssize_t pos = 0;
       PyDict_Next(prpLst, &pos, &key, &value);
       void *data;
       int lsz = 1;
       if ((data=array_from_python(value, &lsz, CF_NAME64, NULL)) != NULL)
       {
         char keyStr[512];
         data_from_python(key, (void *)keyStr, CF_TEXT, NULL);
         cc = AssignPropertyList(gSrvEqm, nam, keyStr, lsz, (NAME64 *)data);
       }
     }
  }
  return Py_BuildValue("i", cc);
}
static PyObject* pytine_register_property(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *nam=NULL, *tag=NULL, *dsc=NULL, *egu=NULL, *modstr=NULL, *fmtstr=NULL, *fmtstrIn=NULL;
  PyObject *devLst = NULL;
  float max = 0, min = 0;
  int sz=0, szIn=0, fmt=CF_NULL, fmtIn=CF_NULL, acc=CA_READ;
  int cc=0;
  static const char *kwlist[] = {
        "name",
        "size",
        "format",
        "size_input",
        "format_input",
        "mode",
        "min",
        "max",
        "units",
        "description",
        "devicelist",
    NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sis|issffssO", (char**)kwlist, &nam, &sz, &fmtstr, &szIn, &fmtstrIn, &modstr, &min, &max, &egu, &dsc, &devLst)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.register_property(name='str',size=val,format='str'[,size_input=val,format_input='str',mode='str,min=val,max=val,units='str',description='str',devicelist=Dict])");    
  }
  int atyp = AT_UNKNOWN;
  if (fmtstr != NULL) 
  {
    fmt = (int)GetFormatFromString(fmtstr);
    if (strchr(fmtstr,'.') != NULL) atyp = GetArrayTypeFromString(fmtstr);
  }
  if (fmtstrIn != NULL)
  {
    fmtIn = (int)GetFormatFromString(fmtstrIn);
  }
  if (atyp == AT_UNKNOWN) atyp = AT_SPECTRUM;
  if ((tag=strchr(fmtstr,'.')) != NULL) tag++;
  if (modstr != NULL) 
  {
    char mod[128];
    strncpy(mod,modstr,128); strupr(mod);
    if (strstr(mod,"WRITE") != NULL) acc |= CA_WRITE;
    if (strstr(mod,"NETWORK") != NULL) acc |= CA_NETWORK;
    if (strstr(mod,"HIST") != NULL) acc |= CA_HIST;
  }
  if (acc & CA_WRITE && sz > 0 && szIn == 0) 
  {
    szIn = atyp == AT_CHANNEL ? 1 : sz;
  }
  if (szIn > 0 && fmtIn == CF_NULL) fmtIn = fmt;
  cc = RegisterBufferedPropertyEx2(nam,szIn,fmtIn,tag,sz,fmt,tag,max,min,egu,acc,dsc,-1,atyp,sz); 
  if (devLst != NULL)
  {
    if (!PyDict_Check(devLst))
    {
      return PyErr_Format(PyExc_TypeError, "dictionary expected for device list!");
    }
    if (PyDict_Size(devLst) < 1)
    {
      return PyErr_Format(PyExc_TypeError, "device list dictionary is empty !");
    }
    PyObject *key, *value;
    Py_ssize_t pos = 0;
    PyDict_Next(devLst, &pos, &key, &value);
    void *data;
    int lsz = 1;
    if ((data=array_from_python(value, &lsz, CF_NAME64, NULL)) != NULL)
    {
      cc = AssignDeviceListToProperty(gSrvEqm, nam, (NAME64 *)data, lsz);
    }
  }

  return Py_BuildValue("i", cc);
}
static PyObject* pytine_attach_server(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *srv=NULL, *eqm=NULL;
  int sz = 0;
  int cc=0;
  static const char *kwlist[] = {
        "server",
        "eqm",
        "capacity",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ssi", (char**)kwlist, &srv, &eqm, &sz)) 
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.attach_server(server='str',eqm='str',capacity=val)");    
  }
  if (srv == NULL && eqm == NULL)
  {
    cc = AttachServerFromFecXml();
  }
  else
  {
    char srvStr[64], eqmStr[64], *psrv=NULL, *peqm=NULL;
    if (srv != NULL)
    {
      sprintf(srvStr,"%.32s/INIT=KEEP",srv);
      psrv = srvStr;
    }
    else if (eqm != NULL)
    {
      sprintf(eqmStr,"%.6s",eqm);
      if (srv == NULL) strcat(eqmStr,"/INIT=KEEP");
      peqm = eqmStr;
    }
    cc = AttachServerEx(psrv,peqm,sz,NULL,0); 
  }
  strncpy(gSrvEqm,GetEquipmentModuleName(),EQM_NAME_SHORTSIZE);
  return Py_BuildValue("i", cc);
}
static PyObject* pytine_pushdata(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  PyObject *value = NULL;
  char *c, *prp=NULL, *dev=NULL;
  int nr = 0, sz = 1, lsz = 1, schd = 0, syss = 0, usrs = 0;
  double ts = 0;
  int cc=0, szOut, fmt;
  char devName[64], fmtStr[128], tag[64];
  BYTE *data, *pdata;
  char **sa = NULL;
  static const char *kwlist[] = {
        "property",
        "data",
        "device",
        "devicenumber",
        "size",
        "scheduled",
        "timestamp",
        "systemstamp",
        "userstamp",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO|siiidii", (char**)kwlist, &prp, &value, &dev, &nr, &sz, &schd, &ts, &syss, &usrs)) 
  {
     return PyErr_Format(PyExc_SyntaxError, 
       "PyTine.pushdata(property='str',data=object,device='str',devicenumber=val,size=val,scheduled=val,timestamp=val)");    
  }

  cc = GetRegisteredPropertyInformation(gSrvEqm,prp,NULL,NULL,NULL,&szOut,NULL,fmtStr,NULL,NULL);
  if (cc != 0) goto err;
  fmt = GetFormatFromString(fmtStr);
  if ((c=strchr(fmtStr,'.')) != 0) strncpy(tag,++c,TAG_NAME_SIZE); else tag[0] = 0;
  if (GetFormatHeaderSize(fmt) > 0)
  { // get the allocation right !
    if (sz == 1) sz = szOut;
    lsz = sz;
  }
  data = (BYTE *)array_from_python(value, &lsz, fmt, tag);
  if (sz == 1) sz = lsz;
  if (!data) ccerr(invalid_data);
  // set the 'ceiling' 
  if (sz < szOut)
  { // pushing fewer than registered
    SetBufferedDataSize(prp, sz);
  }
  else
  { // set the ceiling to 0, so that array calls of single elements work MCA-like
    SetBufferedDataSize(prp, 0);
  }
  if (nr == 0 && dev != NULL) nr = GetDeviceNumberEx(gSrvEqm,dev,prp);
  sprintf(devName,"#%d",nr);
  if (ts > 1000 && schd == 0) schd = (int)ts;
  pdata = data;
  if (fmt == CF_STRING)
  { // gotta find the strings ...
    int p = 0;
    sa = (char **)alloca(sz * sizeof(char *));
    if (sa == NULL) ccerr(out_of_local_memory);
    for (int i = 0; i < sz; i++)
    {
      sa[i] = (char *)&data[p];
      p += (int)(strlen(sa[i])+1);
    }
    pdata = (BYTE *)sa;
  }
  Py_BEGIN_ALLOW_THREADS;
  cc = pushBufferedDataEx(prp,devName,pdata,sz,schd,ts,syss,usrs);
  Py_END_ALLOW_THREADS;
  free(data);
err:
  if (cc != 0)
  {
    char errstr[128];
    return PyErr_Format(PyExc_RuntimeError,"PyTine.pushdata(%s) : %s",prp,GetLastLinkError(cc,errstr));
  }
  return Py_BuildValue("i", cc);
}

int prpHdlr(char *prp,char *dev, void *ref)
{
  PyGILState_STATE gstate;
  PyObject *callable = (PyObject *)ref;
  PyObject *data = Py_None, *dtype = NULL;

  gstate = PyGILState_Ensure();
  int szIn=0;
  char *c, fmtStr[128], tag[64];
  int cc = GetRegisteredPropertyInformation(gSrvEqm,prp,NULL,NULL,NULL,NULL,&szIn,NULL,fmtStr,NULL);
  if (cc != 0) ccerr(cc);
  if (szIn > 0)
  {
    int fmt = GetFormatFromString(fmtStr);
    if ((c=strchr(fmtStr,'.')) != 0) strncpy(tag,++c,TAG_NAME_SIZE);
    int fsiz = fmtsizeof(fmt);
    int isString = FALSE;
    if (fmt == CF_STRUCT)
    {
      fsiz = GetStructSize(tag);
    }
    else if (fmt == CF_STRING || fmt == CF_KEYVALUE)
    {
      fsiz = sizeof(char*);
      isString = TRUE;
    }
    else if (IsVariableLengthFormat(fmt))
    {
      fsiz = GetFormatStandardCapacity(fmt);
    }
    fsiz *= szIn;
    fsiz += getFormatHeaderSize(fmt);
    BYTE *rbdata = (BYTE *)malloc(fsiz);
    if (!rbdata) return out_of_server_memory;
    cc = pullBufferedData(prp,dev,rbdata,szIn);
    if (cc == 0) array_to_python(&data,rbdata,szIn,fmt,tag, FALSE);
    free(rbdata);
    if (szIn > 0 && data == NULL)
    {
      data = Py_None;
      cc = invalid_data;
    }
  }
  else
  {
    data = Py_None;
  }
  if (cc == 0)
  {
    int nr = GetDeviceNumberEx(gSrvEqm,dev,prp);
    int nc = 1;
    ClnInfo cinf;
    char caddr[128];
    if (GetCallerInformation(gSrvEqm, &cinf, &nc) == 0)
    {
      GetInetAddress(&cinf.addr, caddr);
      c = cinf.userName;
    }
    else
    {
      strcpy(caddr, "unknown");
      c = caddr;
    }
    char ip[128], host[128], *cip;
    strncpy(ip, caddr, 128);
    if ((cip = strchr(ip, ':')) != NULL) *cip = 0;
    if (GetHostNameFromAddress(ip, host, 128) != 0)
    {
      strcpy(host, "unknown");
    }
    dtype = Py_BuildValue("{s:s,s:s,s:i,s:s,s:s,s:s,s:O}","property",prp,"device",dev,"devicenumber",nr,"caller",c,"caller_address",caddr,"caller_host",host,"data",data);

	  PyObject *args = Py_BuildValue("(N)", dtype);
	  PyObject *result = PyObject_Call(callable, args, NULL);
    data_from_python(result,&cc,CF_INT32,NULL);

	  Py_XDECREF(args);
	  Py_XDECREF(data);
	  Py_XDECREF(dtype);
    if (result != NULL) Py_XDECREF(result); else PyErr_SetString(PyExc_TypeError, "unable to call property handler");
  }
err:
  PyGILState_Release(gstate);
  return cc;
}
static PyObject* pytine_attach_handler(PyObject *self, PyObject *args, PyObject *kwargs) 
{
  char *property;
  PyObject *handler;
  int cc=0;
  static const char *kwlist[] = {
        "property",
        "handler",
        NULL
    };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO", (char**)kwlist, &property, &handler))
  {
     return PyErr_Format(PyExc_SyntaxError, "PyTine.attach(property='str',handler=obj)");    
  }
  if (!PyCallable_Check(handler))
    return PyErr_Format(PyExc_ValueError, "tine: attach_handler('%s'): %s",
                        property, "object not callable");

  /* Increase the refernce counter to let the Python know,
     that we need to keep this around. */
  Py_INCREF(handler);
  Py_BEGIN_ALLOW_THREADS;
  cc = RegisterServerCallbackEx(property, prpHdlr, handler);
  Py_END_ALLOW_THREADS;

  return Py_BuildValue("i", cc);
}
static PyObject *deathHandler = NULL;
static PyObject *pytine_register_death_handler(PyObject *self, PyObject *args, PyObject *kwargs)
{
  PyObject *handler;
  int cc = 0;
  static const char* kwlist[] = {
        "handler",
        NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O", (char**)kwlist, &handler))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.register_death_handler(handler=obj)");
  }
  if (!PyCallable_Check(handler))
    return PyErr_Format(PyExc_ValueError, "tine: register_death_handler(): %s","object not callable");
  deathHandler = handler;

  return Py_BuildValue("i", cc);
}

static PyObject *pytine_add_history_record(PyObject *self, PyObject *args, PyObject *kwargs)
{
  char *prp = NULL, *dev = (char*)"#0", fmtStr[128];
  int cc = 0;
  int len = -1, frm = -1, idx = -1;
  int monint = HRS_DEFAULT_POLL_RATE, archint = HRS_DEFAULT_ARCHIVE_RATE;
  int dshort = HRS_DEFAULT_DEPTHSHORT, dlong = HRS_DEFAULT_DEPTHLONG;
  int hb = HRS_DEFAULT_HEARTBEAT, flgs = USE_POI_NONE;
  float ptol = HRS_DEFAULT_REL_TOLERANCE, atol = HRS_DEFAULT_ABS_TOLERANCE;

  static const char* kwlist[] = {
    "property",
    "device",
    "length",
    "format",
    "index",
    "monitor_interval",
    "archive_interval",
    "depth_short",
    "depth_long",
    "heartbeat",
    "tolerance_percent",
    "tolerance_absolute",
    NULL
  };

  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|siiiiiiiiff", (char**)kwlist, &prp, &dev, &len, &frm, &idx, &monint, &archint, &dshort, &dlong, &hb, &ptol, &atol))
  {
    return PyErr_Format(PyExc_SyntaxError, "PyTine.add_history_record(property='str'[,device='str',length=val,format=val,index=val,monitor_interval=val,archive_interval=val,depth_short=val,depth_long=val,heartbeat=val,tolerance_percent=val,tolerance_absolute=val)");
  }

  if (strlen(gSrvEqm) == 0)
  {
    return PyErr_Format(PyExc_NameError, "Server does not yet have a module name (not initialized?)");
  }
  if (prp == NULL || strlen(prp) == 0)
  {
    return PyErr_Format(PyExc_NameError, "Property can not be empty.");
  }
  if (frm == -1 || len == -1)
  {
    int lentmp = 0;
    cc = GetRegisteredPropertyInformation(gSrvEqm, prp, NULL, NULL, NULL, &lentmp, NULL, fmtStr, NULL, NULL);
    if (cc != 0)
    {
      char errstr[128];
      return PyErr_Format(PyExc_RuntimeError, "PyTine.add_history_record(...) : %s", GetLastLinkError(cc, errstr));
    }
    if (frm == -1)
    {
      frm = GetFormatFromString(fmtStr);
    }
    if (len == -1)
    {
      len = lentmp;
    }
  }

  HistorySpecification h;
  memset(&h, 0, sizeof(HistorySpecification));
  h.pollingRate = monint;
  h.archiveRate = archint;
  h.depthShort = dshort;
  h.depthLong = dlong;
  h.heartbeat = hb;
  h.pTolerance = ptol;
  h.aTolerance = atol;

  if ((cc = AppendHistoryInformationEx(gSrvEqm, prp, dev, len, frm, idx, &h, flgs)) != 0)
  {
    PyObject* ex = pytine_getExcObject(cc);
    if (ex != NULL) return thrownTineException(ex, cc, "append_history", gSrvEqm, prp, NULL);
  }
  return Py_BuildValue("i", cc);
}

struct module_state {
    PyObject *error;
};
static PyMethodDef tine_methods[] = {
  { "get_attribute", (PyCFunction)pytine_getAttribute, METH_VARARGS|METH_KEYWORDS, "'attribute'[,'format']" },
  { "set_attribute", (PyCFunction)pytine_setAttribute, METH_VARARGS|METH_KEYWORDS, "'attribute',value[,'format']" },
  { "get", (PyCFunction)pytine_get, METH_VARARGS|METH_KEYWORDS, "'address','property'[,'format',size,timeout,'mode']" },
  { "set", (PyCFunction)pytine_set, METH_VARARGS|METH_KEYWORDS, "'address','property'[,input,'format',size,timeout,'mode',callback]" },
  { "call",(PyCFunction)pytine_call,METH_VARARGS|METH_KEYWORDS, "'address','property'[,input,'mode','format',size,'inputformat',inputsize,timeout]" },
  { "stop_get",(PyCFunction)pytine_stop,METH_VARARGS | METH_KEYWORDS, "'address','property'[,input,'format',size,'inputformat',inputsize]" },
  { "query",    pytine_query,    METH_VARARGS, "" },
  { "attach", (PyCFunction)pytine_attach,METH_VARARGS|METH_KEYWORDS,"'address','property',callback[,input,'format',size,'inputformat',inputsize,interval,'mode',id,'tolerance']" },
  { "detach",   pytine_detach,   METH_VARARGS, "linkid" },
  { "list", (PyCFunction)pytine_list, METH_VARARGS|METH_KEYWORDS, "[context,server,device,property,structure]" },
  { "history",(PyCFunction)pytine_hist,METH_VARARGS|METH_KEYWORDS, "address,property[,'stop','depth','flags',timeout,sample,numberPoints]" },
  { "attach_server", (PyCFunction)pytine_attach_server,METH_VARARGS|METH_KEYWORDS,"['server','eqm',capacity]" },
  { "pushdata", (PyCFunction)pytine_pushdata,METH_VARARGS|METH_KEYWORDS,"property, data, (device, devicenumber) [,size, scheduled, timestamp]" },
  { "attach_handler", (PyCFunction)pytine_attach_handler,METH_VARARGS|METH_KEYWORDS,"'property',handler" },
  { "register_fec", (PyCFunction)pytine_register_fec,METH_VARARGS|METH_KEYWORDS,"'name',['subsystem','context','description','location','hardware','responsible',port]" },
  { "register_server", (PyCFunction)pytine_register_server,METH_VARARGS|METH_KEYWORDS,"'name'[,'context','eqm',capacity,'fec','subsystem']" },
  { "register_device", (PyCFunction)pytine_register_device,METH_VARARGS|METH_KEYWORDS,"'name',number,['description','location','hardware','responsible',zposition,'redirection',propertylist]" },
  { "register_property", (PyCFunction)pytine_register_property,METH_VARARGS|METH_KEYWORDS,"'name',size,'format',[size_input,'format_input','mode',min,max,'units','description',devicelist]" },
  { "register_type", (PyCFunction)pytine_register_type,METH_VARARGS | METH_KEYWORDS,"'tag',struct" },
  { "structure_to_bytes", (PyCFunction)structure_to_bytes,METH_VARARGS | METH_KEYWORDS,"'tag',struct" },
  { "bytes_to_structure", (PyCFunction)bytes_to_structure,METH_VARARGS | METH_KEYWORDS,"'tag',bytes" },
  { "setalarm", (PyCFunction)pytine_setAlarm,METH_VARARGS|METH_KEYWORDS,"'code'[,('device',devicenumber),data,'options']" },
  { "clearalarm", (PyCFunction)pytine_clearAlarm,METH_VARARGS|METH_KEYWORDS,"['code',('device',devicenumber)]" },
  { "triggerEvent", (PyCFunction)pytine_triggerEvent,METH_VARARGS|METH_KEYWORDS,"'context','event'[,'comment','triggerLevel','triggerTime','rangeStart','rangeStop','rangeMax','options']"},
  { "getEventList", (PyCFunction)pytine_getEventList,METH_VARARGS|METH_KEYWORDS,"'context','event'[,'startTime','stopTime','limit']"},
  { "getArchivedEventData", (PyCFunction)pytine_getArchivedEventData,METH_VARARGS|METH_KEYWORDS,"'context','event','eventId',['eventItem','eventServer','eventDevice','eventProperty','eventContext','format','size']"},
  { "getEventArchiveComment", (PyCFunction)pytine_getEventArchiveComment,METH_VARARGS|METH_KEYWORDS,"'context','event','eventId'"},
  { "setEventArchiveComment", (PyCFunction)pytine_setEventArchiveComment,METH_VARARGS|METH_KEYWORDS,"'context','event','eventId','comment'"},
  { "register_death_handler", (PyCFunction)pytine_register_death_handler,METH_VARARGS|METH_KEYWORDS,"handler"},
  { "errorlist", (PyCFunction)pytine_errorlist,METH_VARARGS|METH_KEYWORDS,"list of error codes"},
  { "register_error_code", (PyCFunction)pytine_register_errorcode,METH_VARARGS|METH_KEYWORDS,"'code','message'[,'long message']"},
  { "allowed_users", (PyCFunction)pytine_allowedusers,METH_VARARGS|METH_KEYWORDS,"['add','remove']"},
  { "allowed_addresses", (PyCFunction)pytine_allowedaddresses,METH_VARARGS|METH_KEYWORDS,"['add','remove']"},
  { "add_history_record", (PyCFunction)pytine_add_history_record,METH_VARARGS|METH_KEYWORDS,"'property','device','length','format','index','monitor_interval','archive_interval','depth_short','depth_long','heartbeat','tolerance_percent','tolerance_absolute'"},
  { "log", (PyCFunction)pytine_log, METH_VARARGS|METH_KEYWORDS, "'message'[,file,severity,tag]"},
  { "strerror", pytine_strerror, METH_VARARGS, "error code" },
  { "version",  pytine_version,  METH_NOARGS,  "" },
  { "debug",    pytine_tty,      METH_VARARGS, "debug command"},
  { NULL,       NULL,      0,            NULL }
};
// morituri te salutamus ...
void tineDeathHandler(const char *msg)
{ 
  PyGILState_STATE gstate;
  gstate = PyGILState_Ensure();
  if (deathHandler != NULL)
  {
    SetDieAnotherDay(TRUE);
    PyObject *args = Py_BuildValue("(s)", msg);
    PyObject *result = PyObject_Call(deathHandler, args, NULL);
    Py_XDECREF(args);
    //if (result != NULL) Py_XDECREF(result); else PyErr_SetString(PyExc_TypeError, "unable to call death handler");
  }
  else
  {
    char s[256];
    sprintf(s, "print('%.128s : process will be terminated!')", msg);
    PyRun_SimpleString(s);
  }
  PyGILState_Release(gstate);
}
#if PY_MAJOR_VERSION >= 3
#define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
#else
#define GETSTATE(m) (&_state)
static struct module_state _state;
#endif

static int myextension_traverse(PyObject *m, visitproc visit, void *arg) 
{
    Py_VISIT(GETSTATE(m)->error);
    return 0;
}

static int myextension_clear(PyObject *m) 
{
    Py_CLEAR(GETSTATE(m)->error);
    return 0;
}

void registerErrorCodes(PyObject *module)
{
  char *c, errstr[64];
  int nerr = GetErrorListSize();
  for (int i=0; i<nerr; i++)
  { 
    strncpy(errstr, erlst[i], 32);
    for (c = errstr; *c != 0; c++) if (*c == ' ') *c = '_';
    PyModule_AddIntConstant(module, errstr, i);
  }
}

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef tine_module = {
        PyModuleDef_HEAD_INIT,
        "tine",
        NULL,
        sizeof(struct module_state),
        tine_methods,
        NULL,
        myextension_traverse,
        myextension_clear,
        NULL
};
#define INITERROR return NULL
PyMODINIT_FUNC PyInit_PyTine (void)
#else
#define INITERROR return
PyMODINIT_FUNC initPyTine(void)
#endif
{
  //Py_Initialize();
  PyEval_InitThreads();

# if PY_MAJOR_VERSION >= 3
  PyObject *module = PyModule_Create(&tine_module);
# else
  PyObject *module = Py_InitModule("PyTine", tine_methods);
# endif
  if (module == NULL) INITERROR;

  registerErrorCodes(module);
  gStartupDebug = FALSE;
  SetConnectionTableCapacity(5000);
  SetUserType("PyTine");
  SetDieFunction(tineDeathHandler);
  SetSufferInSilence(TRUE); /* keep the underlying output off the screen */
  SystemStartCycleTimer();
  Py_AtExit(SystemKillCycleTimer);
  #if PY_MAJOR_VERSION >= 3
  return module;
# endif
}



