/*
 * File repeater.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "listener.h"
#if defined(UNIX)
#include "dirent.h"
#elif defined(WIN32)
#include "direct.h"
#endif
#include "repeater.h"

NAME64 schPrpLst[MAX_SCHEDULED_PROPERTIES];
int nSchPrps = 0;
int gTruncateDeviceNames = 0;
int hasExclusiveRead = FALSE;
int ignoreListeningErrors = FALSE;
int repeatHistories = FALSE;
int rpPollingRate = RPTR_POLLRATE;
NAME64 *rpDeviceNames = NULL;
PropertyQueryStruct *rpPrpsTrunc = NULL;
PrpQueryStruct *rpPrps = NULL;
int rpPortOffset = RPTR_DEFAULT_PORT;
int rpteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
int dumpDevices(char *tgtContext,char *tgtServer);
int dumpStructure(char *tgtContext,char *tgtServer,char *tag);
int dumpPropertyList(char *tgtContext,char *tgtServer,int idx);
int dumpProperties(char *tgtContext,char *tgtServer);
int getDevicesFromCache(char *tgtContext,char *tgtServer);
int getPropertiesFromCache(char *tgtContext,char *tgtServer);
int getStructFromCache(char *tgtContext,char *tgtServer,char *tag,int siz);
void rptini(void);
void rptbkg(void);
void rptexi(void);

char thisFec[FEC_NAME_SIZE+4], thisContext[CONTEXT_NAME_SIZE+4], thisServer[EXPORT_NAME_SIZE+4];
char thisDescription[128], thisHost[32], thisUser[32];
char gTargetDeviceServer[256];
char thisDeviceServer[32];
char thisFailoverMaster[EXPORT_NAME_SIZE+4], thisSlaveMaster[EXPORT_NAME_SIZE+4];
int rptGetNumberOf(char *srv,char *stkprp)
{
  DTYPE dout;
  int cc = 0;
  short n;
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = &n;
  if ((cc=ExecLinkEx(srv,stkprp,&dout,NULL,CA_READ,800)) != 0 && !(cc & CE_SENDDATA)) return -cc;
  return n;
}
void makeFecName(char *myContext,char *myServer)
{
  char *c, ctx[16],srv[16];
  int i;
  ctx[0] = (char)toupper(myContext[0]); // 'P'
  ctx[1] = (char)tolower(myContext[1]); // 'e'
  c = myServer;
  memset(srv,0,16);
  srv[0] = (char)toupper(*c++);
  for (i=1; i<16 && *c != 0; i++)
  {
    srv[i] = (char)tolower(*c++);
    if (srv[i] == '_') srv[i] = (char)toupper(*c++);
  }
  if ((c=strstr(srv, ".rpt")) != NULL) *c = 0; // chop this off the end
  if (srv[11] == '.') srv[11] = 0; // don't allow a '..<port>' in the name
  sprintf(thisFec, "Rpt%.2s%.7s.%d",ctx,srv,rpPortOffset);
}
int rpPreSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer) 
{
  int cc=0;
  char *ptr = NULL;
  if (!isLocalRepeater)
  {
    if ((ptr=getenv("TINE_PORT")) != NULL) rpPortOffset = (UINT16)atoi(ptr);
    if ((ptr=getenv("TINE_CONTEXT")) != NULL) 
    {
      strncpy(thisContext,ptr,CONTEXT_NAME_SIZE);
    }
    else
    {
      strncpy(thisContext,myContext,CONTEXT_NAME_SIZE); 
    }
    thisContext[CONTEXT_NAME_SIZE] = 0;
    if ((ptr=getenv("TINE_SERVER")) != NULL) 
    {
      strncpy(thisServer,ptr,EXPORT_NAME_SIZE);
    }
    else
    {
      strncpy(thisServer,myServer,EXPORT_NAME_SIZE);
    }
    thisServer[EXPORT_NAME_SIZE] = 0;
    if (strlen(thisFec) == 0)
    {
      makeFecName(myContext,myServer);
    }
  }
  if ((ptr=getenv("TINE_REPEATER")) != NULL && !stricmp(ptr,"SYNCHRONOUS")) isSyncRepeater = TRUE;
  if ((ptr=getenv("HOSTNAME")) == NULL && (ptr=getenv("COMPUTERNAME")) == NULL) ptr = "local.host";
  strncpy(thisHost,ptr,32);
  sprintf(thisDescription,"/%.32s/%.32s repeater",tgtContext,tgtServer);
  if ((ptr=getenv("USER")) == NULL && (ptr=getenv("USERNAME")) == NULL) ptr = "admin";
  strncpy(thisUser,ptr,16); 
  thisUser[16] = 0;

  if ((cc=RegisterFecInformation(thisFec,"RPT",thisContext,thisDescription,thisHost,"none",thisUser,(UINT16)rpPortOffset)) != 0)
  {
    feclog("could not register repeater FEC name : %s",cc2str(cc));
  }
  if (!ignoreListeningErrors) SetHistoryDeactivateOnError(TRUE);
  return 0;
}
NAME64 *hstProperties = NULL;
int numHstProperties = 0;
int acqHistoryInfo(char *tgt, int *nhstprps, NAME64 *hstprps)
{
  DTYPE dout;
  int cc = 0;

  if (tgt == NULL || nhstprps == NULL || hstprps == NULL) return argument_list_error;
  if (*nhstprps < 1) return 0;
  dout.dFormat = CF_NAME64;
  dout.dArrayLength = (UINT32)(*nhstprps);
  dout.data.vptr = hstprps;
  if ((cc=ExecLinkEx(tgt, "HISTORIES", &dout, NULL, CA_READ | CA_MUTABLE, 500)) == 0)
  {
    *nhstprps = (int)dout.dArrayLength;
  }
  return cc;
}
int probeDevQryPrecedence(char *tgt,int *nprps)
{
  DTYPE dout;
  int cc = 0, fmt = CF_NAME64;
  NAME64 plst[100];

  if (tgt == NULL) return FALSE;
  dout.dFormat = (short)fmt;
  dout.dArrayLength = 100;
  dout.data.cptr = (char *)plst;
  if ((cc=ExecLinkEx(tgt,"PROPERTIES",&dout,NULL,CA_READ|CA_RETRY|CA_MUTABLE,500)) & CE_SENDDATA)
  {
    if ((cc & 0xfff) == has_query_function) return TRUE;
  }
  if (cc == illegal_property) return TRUE; /* e.g. tango or doocs server */
  if (dout.dArrayLength > (UINT32)*nprps) *nprps = dout.dArrayLength;
  return FALSE;
}
NAME64 plstbuf[1000];
typedef struct
{
  NAME64 *plst;
  PrpQueryStruct *pqslst;
  int lstsiz;
} PLST;
PLST *plst = NULL;
int isGensTarget = FALSE;
int hasDevQryPrecedence = FALSE;
int isSyncRepeater = FALSE;
int isLocalRepeater = FALSE;
int ndevices = 0,nproperties = 0;
int rpPostSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer)
{
  int cc=0,i,p,n,np,dbg;
  int dumpResults = TRUE, targetUnavailable = FALSE;
  short acc;
  char tgt[256], dsc[256], dname[DEVICE_NAME_SIZE+4];
  DTYPE dout, din;
  FecAddrStruct *fa;
  ADS ads;

  myServer = myServer;  /* avoid warnings ... */
  myContext = myContext;
  if (isLocalRepeater)
  {
    if ((cc=RegisterEquipmentModule(RPTR_LCLSRV,RPTR_EQM,1,rpteqm,rptini,rptbkg,1000,rptexi)) != 0)
    {
      feclog("could not register repeater equipment module : %.32s",cc2str(cc));
    }

    cc = RegisterPropertyInformation(RPTR_EQM,"CALL",NULL,NULL,
                                       CA_READ|CA_WRITE,AT_UNKNOWN,
                                       1,"local repeater calling property",1,NULL);
    if (cc < 0) return -cc;
    cc = RegisterDeviceName(RPTR_EQM,"target",0);
    return cc;
  }

  sprintf(tgt,"/%.32s/%.32s/#0",tgtContext,tgtServer);
  sprintf(gTargetDeviceServer,"/%.32s/%.32s/",tgtContext,tgtServer);
  strncpy(thisDeviceServer, tgtServer, 32);
# ifdef TEST_DB_READ
  targetUnavailable = TRUE;
  goto infoRegistration;
# endif
  dbg = tineDebug;
  if ((ndevices=rptGetNumberOf(tgt,"NDEVICES")) < 0)
  {
    cc = -ndevices;
    tineDebug = 1;
    feclog("cannot get number of devices from target %.256s: %.32s",tgt,cc2str(cc));
    tineDebug = dbg;
    targetUnavailable = TRUE;
    goto infoRegistration;
  }
  if ((rpDeviceNames=(NAME64 *)calloc(ndevices,sizeof(NAME64))) == NULL)
  {
    feclog("cannot allocate space for device names!");
    exit(1);
  }
  n = ndevices;
  if ((cc=GetSystemDevices(tgt,NULL,rpDeviceNames,&n)) != 0 && !(cc & CE_SENDDATA))
  {
    tineDebug = 1;
    feclog("cannot get device list from target %.256s: %.32s",tgt,cc2str(cc));
    tineDebug = dbg;
    targetUnavailable = TRUE;
    goto infoRegistration;
  }
  feclog("obtained device list (%d devices) from target %.256s",ndevices,tgt);
  if ((fa=GetFecAddrFromCache(tgtContext,tgtServer)) != NULL)
  {
    isGensTarget = stricmp(fa->fecName,"GENS") == 0 ? TRUE : FALSE;
  }
  cc = 0;
  sprintf(tgt,"/%.32s/%.32s/*",tgtContext,tgtServer);
  if ((nproperties=rptGetNumberOf(tgt,"NPROPS")) < 0)
  {
    cc = -nproperties;
    tineDebug = 1;
    feclog("cannot get number of properties from target %.256s: %.32s",tgt,cc2str(cc));
    tineDebug = dbg;
    targetUnavailable = TRUE;
    goto infoRegistration;
  }
  sprintf(tgt,"/%.32s/%.32s/#0",tgtContext,tgtServer);
  if (probeDevQryPrecedence(tgt,&nproperties))
  {
    tineDebug = 1;
    feclog("target %.256s has device query precedence", tgt);
    tineDebug = dbg;
    hasDevQryPrecedence = TRUE;
    if ((plst=(PLST *)calloc(ndevices,sizeof(PLST))) == NULL)
    {
      tineDebug = 1;
      feclog("cannot get alloc property list space");
      exit(1);
    }
    nproperties = 0;
    for (i=0; i<ndevices; i++)
    {
      np = 1000;
      sprintf(tgt,"/%.32s/%.32s/%.64s",tgtContext,tgtServer,rpDeviceNames[i].name);
      if ((cc=GetSystemProperties(tgt,plstbuf,&np)) == 0 || cc & CE_SENDDATA)
      {
        if ((plst[i].plst=(NAME64 *)calloc(np,sizeof(NAME64))) == NULL)
        {
          tineDebug = 1;
          feclog("cannot alloc property list space for %d properties",np);
          exit(1);
        }
        memcpy(plst[i].plst,plstbuf,np*sizeof(NAME64));
        plst[i].lstsiz = np;
      }
      cc = 0;
      nproperties += np;
    }
    if (ndevices > 100 && nproperties > 10000) nproperties = 5000; /* a sensible cap */
    sprintf(tgt,"/%.32s/%.32s/*",tgtContext,tgtServer);
  }
  feclog("obtained property list (%d properties) from target %.256s",nproperties,tgt);
  n = nproperties;
  if ((rpPrps=(PrpQueryStruct *)calloc(nproperties,sizeof(PrpQueryStruct))) == NULL)
  {
    tineDebug = 1;
    feclog("cannot allocate space for property list!");
    exit(1);
  }
  if (hasDevQryPrecedence) 
  {
    ExpDataStruct srvaddr;
    FecAddrStruct fecaddr;
    if (GetServerAddr(tgtContext,tgtServer,&srvaddr,&fecaddr) == 0)
    { /* this should always be successful here ! */
      if (!strncmp(srvaddr.EqmName,DOOCSEQM_TAG,EQM_NAME_SHORTSIZE)) goto oneAtATime;
      if (!strncmp(srvaddr.FecName, GENS_FEC, FEC_NAME_SIZE)) goto oneAtATime;
    }
  }
  if (n < 600)
  {
    if ((cc=GetSystemPropertyInformation(tgt,NULL,&rpPrps,&n)) != 0 && !(cc & CE_SENDDATA))
    {
      if (hasDevQryPrecedence && (cc == illegal_property || cc == illegal_device))
      { /* e.g. tango or doocs */
        int retry;
oneAtATime:
        for (i=0,p=0; i<ndevices; i++)
        {
          n = plst[i].lstsiz;
          sprintf(tgt,"/%.32s/%.32s/%.64s",tgtContext,tgtServer,rpDeviceNames[i].name);
          plst[i].pqslst = &rpPrps[p];
          for (retry=10; retry>=0; retry--)
          {
            if ((cc=GetSystemPropertyInformation(tgt,NULL,&plst[i].pqslst,&n)) != 0 && !(cc & CE_SENDDATA))
            {
              dbglog("cannot get property list from target %.256s: %.32s",tgt,cc2str(cc));
              if (retry == 0) 
              {
                targetUnavailable = TRUE;
                goto infoRegistration;
              }
            }
            retry = 0;
          }
          p += plst[i].lstsiz;
          if (p > nproperties) break;
        }
        goto infoRegistration;
      }
      tineDebug = 1;
      feclog("cannot get property list from target %.256s: %.32s",tgt,cc2str(cc));
      tineDebug = dbg;
      targetUnavailable = TRUE;
      goto infoRegistration;
    }
    cc = 0;
  }
  else
  {
    if ((rpPrpsTrunc=(PropertyQueryStruct *)calloc(n,sizeof(PropertyQueryStruct))) == NULL)
    {
      tineDebug = 1;
      feclog("cannot allocate space for property list!");
      exit(1);
    }
    if ((cc=GetDevicePropertyInformation(tgt,rpPrpsTrunc,&n)) != 0 && !(cc & CE_SENDDATA))
    {
      tineDebug = 1;
      feclog("cannot get property list from target %.256s: %.32s",tgt,erlst[cc]);
      tineDebug = dbg;
      targetUnavailable = TRUE;
      goto infoRegistration;
    }
    cc = 0;
    for (i=0; i<n; i++)
    {
      strncpy(rpPrps[i].prpName,rpPrpsTrunc[i].prpName,PROPERTY_NAME_SHORTSIZE);
      strncpy(rpPrps[i].prpDescription,rpPrpsTrunc[i].prpDescription,32);
      rpPrps[i].prpSize = rpPrpsTrunc[i].prpSize;
      rpPrps[i].prpFormat = rpPrpsTrunc[i].prpFormat;
      rpPrps[i].prpAccess = rpPrpsTrunc[i].prpAccess;
    }
  }
infoRegistration:
  if (targetUnavailable)
  {
    if ((cc=getDevicesFromCache(tgtContext,tgtServer)) != 0) 
    {
      tineDebug = 1;
      feclog("device information for /%.32s/%.32s not available from cache : %s",tgtContext,tgtServer,cc2str(cc));
      exit(1);
    }
    feclog("obtained device list (%d devices) from cache", ndevices);
    if ((cc=getPropertiesFromCache(tgtContext,tgtServer)) != 0)
    {
      tineDebug = 1;
      feclog("property information for /%.32s/%.32s not available from cache : %s",tgtContext,tgtServer,cc2str(cc));
      exit(1);
    }
    feclog("obtained property list (%d properties) from cache", nproperties);
    dumpResults = FALSE;
  }
  if (gTruncateDeviceNames > 0) ndevices = gTruncateDeviceNames;
  n = ndevices;
  if ((cc=RegisterEquipmentModule(thisServer,RPTR_EQM,n,rpteqm,rptini,rptbkg,1000,rptexi)) != 0)
  {
    tineDebug = 1;
    feclog("could not register repeater equipment module : %.32s",cc2str(cc));
    tineDebug = dbg;
  }
  dname[0] = 0;
  for (i=0,n=0; i<ndevices; i++)
  {
    strncat(dname,rpDeviceNames[i].name,DEVICE_NAME_SIZE);
    if (strlen(dname) == 0) continue;
    if ((cc=RegisterDeviceName(RPTR_EQM,dname,n++)) != 0)
    {
      dbglog("cannot register device name %.64s : %.32s",dname,cc2str(cc));
    }
    dname[0] = 0;
  }
  for (i=0; i<nproperties; i++)
  {
    if (strstr(rpPrps[i].prpName,".HIST") != NULL) continue;
    if (strstr(rpPrps[i].prpName,".ARCH") != NULL) continue;
    if (isGensTarget && rpPrps[i].prpArrayType & AT_CHANNEL)
    { /* a GENS target can't forward the MCA info ! */
      rpPrps[i].prpArrayType &= ~(AT_CHANNEL);
      rpPrps[i].prpSize = 1;
    }
    memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
    dout.dFormat = (short)LFMT(rpPrps[i].prpFormat);
    dout.dArrayLength = rpPrps[i].prpSize;
    strncpy(dout.dTag,rpPrps[i].prpTag,TAG_NAME_SIZE);
    din.dFormat = (short)LFMT(rpPrps[i].prpFormatIn);
    din.dArrayLength = rpPrps[i].prpSizeIn;
    strncpy(din.dTag,rpPrps[i].prpTagIn,TAG_NAME_SIZE);
    sprintf(dsc,"[%g:%g %.16s]",rpPrps[i].prpMinValue,rpPrps[i].prpMaxValue,rpPrps[i].prpUnits);
    if (rpPrps[i].prpArrayType & AT_SPECTRUM)
    {
      sprintf(&dsc[strlen(dsc)],"[%g:%g %.16s]",rpPrps[i].rngMinValue,rpPrps[i].rngMaxValue,rpPrps[i].rngUnits);
    }
    sprintf(&dsc[strlen(dsc)],"%.64s",rpPrps[i].prpDescription);
    acc = rpPrps[i].prpAccess;
    if (hasExclusiveRead) 
    {
      acc &= ~(CA_READ);
      acc |= CA_XREAD;
    }
    cc = RegisterPropertyInformation(RPTR_EQM,rpPrps[i].prpName,&dout,&din,
                                       acc,rpPrps[i].prpArrayType,
                                       rpPrps[i].rowSize,dsc,i,rpPrps[i].prpRedirection);
    if (cc < 0)
    {
      tineDebug = 1;
      feclog("cannot register property name %.64s : %.32s",rpPrps[i].prpName,cc2str(-cc));
      tineDebug = dbg;
    }
    if (dout.dFormat == CF_STRUCT && strlen(dout.dTag) > 0)
    {
      if (targetUnavailable)
      {
        cc = getStructFromCache(tgtContext,tgtServer,dout.dTag,dout.dArrayLength);
      }
      else
      {
        cc = AcquireAndRegisterStruct(tgt,dout.dTag,dout.dArrayLength);
      }
      if (cc != 0 && cc != already_assigned) feclog("cannot register structure : %.16s",dout.dTag);
    }
    if (din.dFormat == CF_STRUCT && strlen(din.dTag) > 0)
    {
      if (targetUnavailable)
      {
        cc = getStructFromCache(tgtContext,tgtServer,din.dTag,din.dArrayLength);
      }
      else
      {
        cc = AcquireAndRegisterStruct(tgt,din.dTag,din.dArrayLength);
      }
      if (cc != 0 && cc != already_assigned) feclog("cannot register structure : %.16s",din.dTag);
    }
  }  
  if (hasDevQryPrecedence)
  {
    for (i=0; i<ndevices; i++)
    {
      AssignPropertyList(RPTR_EQM,rpDeviceNames[i].name,rpDeviceNames[i].name,plst[i].lstsiz,plst[i].plst);
    }
  }
  if (repeatHistories)
  { /* acquire history information */
    if ((numHstProperties=rptGetNumberOf(tgt, "NHISTORIES")) < 0)
    {
      numHstProperties = nproperties;
    }
    sprintf(tgt, "/%.32s/%.32s/#0", tgtContext, tgtServer);
    hstProperties = (NAME64 *)calloc(numHstProperties, sizeof(NAME64));
    cc = acqHistoryInfo(tgt,&numHstProperties,hstProperties);
    if (cc != 0)
    {
      tineDebug = 1;
      feclogEx(TINE_LOGLEVEL_ERROR, "could not acquire history information : %.32s", cc2str(cc));
      tineDebug = dbg;
    }
    else
    { /* register the overloaded .HIST meta-properties */
      int id;
      char hstPrpName[72];
      DTYPEZERO(dout);
      for (i = 0; i < numHstProperties; i++)
      {
        id = GetPropertyId(RPTR_EQM, hstProperties[i].name);
        if (id < 0)
        {
          tineDebug = 1;
          feclogEx(TINE_LOGLEVEL_ERROR, "%.64s not a repeated property !", hstProperties[i].name);
          tineDebug = dbg;
          continue;
        }
        sprintf(dsc, "[%g:%g %.16s]", rpPrps[id].prpMinValue, rpPrps[id].prpMaxValue, rpPrps[id].prpUnits);
        if (rpPrps[id].prpArrayType & AT_SPECTRUM)
        {
          sprintf(&dsc[strlen(dsc)], "[%g:%g %.16s]", rpPrps[id].rngMinValue, rpPrps[id].rngMaxValue, rpPrps[id].rngUnits);
        }
        sprintf(&dsc[strlen(dsc)], "%.64s", rpPrps[id].prpDescription);
        sprintf(hstPrpName, "%.56s.HIST", hstProperties[i].name);
        dout.dFormat = CF_DBLDBL;
        dout.dArrayLength = 1000;
        cc = RegisterPropertyInformation(RPTR_EQM, hstPrpName, &dout, NULL,
          CA_READ, rpPrps[id].prpArrayType,
          rpPrps[id].rowSize, dsc, nproperties+i, rpPrps[id].prpRedirection);
        if (cc < 0)
        {
          tineDebug = 1;
          feclog("cannot register property name %.64s : %.32s", hstPrpName, cc2str(-cc));
          tineDebug = dbg;
        }
        sprintf(hstPrpName, "%.56s.ARCH", hstProperties[i].name);
        cc = RegisterPropertyInformation(RPTR_EQM, hstPrpName, &dout, NULL,
          CA_READ, rpPrps[id].prpArrayType,
          rpPrps[id].rowSize, dsc, nproperties + numHstProperties + i, rpPrps[id].prpRedirection);
        if (cc < 0)
        {
          tineDebug = 1;
          feclog("cannot register property name %.64s : %.32s", hstPrpName, cc2str(-cc));
          tineDebug = dbg;
        }
        sprintf(hstPrpName, "%.56s.HIST@", hstProperties[i].name);
        dout.dFormat = rpPrps[id].prpFormat;
        dout.dArrayLength = rpPrps[id].prpSize;
        cc = RegisterPropertyInformation(RPTR_EQM, hstPrpName, &dout, NULL,
          CA_READ, rpPrps[id].prpArrayType,
          rpPrps[id].rowSize, dsc, nproperties + 2*numHstProperties + i, rpPrps[id].prpRedirection);
        if (cc < 0)
        {
          tineDebug = 1;
          feclog("cannot register property name %.64s : %.32s", hstPrpName, cc2str(-cc));
          tineDebug = dbg;
        }
      }
    }
  }
  memset(&ads, 0, sizeof(ADS));
  strncpy(ads.alarmTag, "Target Value unavailable", 32);
  ads.alarmCode = target_value_unavailable;
  ads.alarmMask = 0xff;
  ads.alarmSeverity = 7;
  ads.alarmDataFormat = BFMT(CF_TEXT);
  ads.alarmDataArraySize = 64;
  ads.alarmSystem = 0;
  strncpy(ads.alarmText, "Target Value unavailable to Repeater", 64);
  strncpy(ads.dataText, "Current link status", 64);
  if ((cc=AppendAlarmInfoTable(RPTR_EQM, &ads)) != 0)
  {
    feclog("appendAlarmInfoTable : %s", cc2str(cc));
  }
  memset(&ads, 0, sizeof(ADS));
  strncpy(ads.alarmTag, "Repeater not in sync", 32);
  ads.alarmCode = repeater_not_synchronized;
  ads.alarmMask = 0xff;
  ads.alarmSeverity = 7;
  ads.alarmDataFormat = BFMT(CF_NULL);
  ads.alarmDataArraySize = 0;
  ads.alarmSystem = 0;
  strncpy(ads.alarmText, "Repeater no longer in sync with target", 64);
  if ((cc=AppendAlarmInfoTable(RPTR_EQM, &ads)) != 0)
  {
    feclog("appendAlarmInfoTable : %s", cc2str(cc));
  }

  if (dumpResults)
  {
    dumpDevices(tgtContext,tgtServer);
    dumpProperties(tgtContext,tgtServer);
  }
  if (strlen(thisFailoverMaster) > 0)
  {
    if (strlen(thisSlaveMaster) > 0)
    { /* designated a failover slave */
      feclog("%.32s is designated a %.32s failover SLAVE, monitoring MASTER %.32s ", 
        thisServer, thisFailoverMaster, thisSlaveMaster);
      SetFailoverSlave(RPTR_EQM, thisFailoverMaster, thisSlaveMaster);
    }
    else
    { /* designated a failover master */
      feclog("%.32s is designated a %.32s failover MASTER",
        thisServer, thisFailoverMaster);
      SetFailoverMaster(RPTR_EQM, thisFailoverMaster);
    }
  }
  monitorTarget();
  tineDebug = dbg;
  return 0;
}
int isRptInSyncWithTarget(void)
{
  char tgt[256];
  NAME64 *lst = NULL;
  int nd, np, cc, res=TRUE;
  sprintf(tgt, "%.72s#0", gTargetDeviceServer);

  nd = rptGetNumberOf(tgt, "NDEVICES");
  np = rptGetNumberOf(tgt, "NPROPS");
  if (nd < 0 || np < 0) return TRUE; // until further notice ...
  if (nd != ndevices) return FALSE;
  if (!hasDevQryPrecedence && np != nproperties) return FALSE;
  if ((lst=(NAME64 *)calloc(nd, sizeof(NAME64))) == NULL)
  {
    return TRUE; // in this case, just get out ...
  }
  if ((cc=GetSystemDevices(tgt, NULL, lst, &nd)) != 0 && !(cc & CE_SENDDATA))
  {
    res = TRUE; // in this case, just get out ...
    goto done;
  }
  if (memcmp(rpDeviceNames, lst, ndevices*sizeof(NAME64)))
  { // the lists differ !
    res = FALSE;
    goto done;
  }
done:
  free(lst);
  return res;
}
int gRptStartTime = 0;
int gTgtStartTime = 0;
int needRestart = FALSE;
void cbStartTime(int id, int cc)
{
  if (cc != 0) return;
  if (gRptStartTime == 0) gRptStartTime = gTgtStartTime;
  if (gTgtStartTime > gRptStartTime)
  {
    feclogEx(TINE_LOGLEVEL_INFO, "Target server has been restarted!");
    if (!isRptInSyncWithTarget())
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "Repeater is no longer in sync with Target server!");
      needRestart = TRUE;
    }
    gRptStartTime = gTgtStartTime;
  }
  if (needRestart)
  { // this stays on until the repeater is re-started ...
    SetDeviceAlarm(RPTR_EQM, thisDeviceServer, repeater_not_synchronized, NULL, almNEWALARM);
  }
}
void monitorTarget(void)
{
  char tgt[256];
  int lid;
  DTYPE dout;
  DTYPEZERO(dout);
  sprintf(tgt,"%.72s#0",gTargetDeviceServer);
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &gTgtStartTime;
  lid = AttachLink(tgt, "SRVSTARTTIME", &dout, NULL, CA_READ, 1000, cbStartTime, CM_TIMER);
  if (lid < 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot start monitoring link: %.32s!",cc2str(-lid));
  }
}
void lsnrCb(int id)
{
  SystemScheduleProperty(RPTR_EQM, schPrpLst[id].name);
}
int rpteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char *c,lclTargetDeviceName[256],lclTargetProperty[PROPERTY_NAME_SIZE+4];
  DTYPE *dinptr=NULL;
  int cc = 0;
  char *sts = NULL, stsBuffer[256];
  if (*devName == '<')
  { /* local repeater getting directions */
    strncpy(lclTargetDeviceName,&devName[1],256);
    if ((c=strchr(lclTargetDeviceName,'[')) == NULL)
    { /* badly formed */
      return invalid_name;
    }
    *c++ = 0;
    strncpy(lclTargetProperty,c,PROPERTY_NAME_SIZE);
    if ((c=strchr(lclTargetProperty,']')) != NULL) *c = 0;
    SetSystemUserName(GetCaller(RPTR_EQM));
  }
  else
  { /* targeted repeater */
    if (!isLocalRepeater)
    { /* this should always be true, if I'm here ... */
      if (GetPropertyId(RPTR_EQM,devProperty) < 0) 
      { /* a bad 'history.csv' can sneak past the isIllegalProperty() check ! */
        return illegal_property;
      }
    }
    strncpy(lclTargetDeviceName,gTargetDeviceServer,128);
    strncat(lclTargetDeviceName,devName,DEVICE_NAME_SIZE);
    strncpy(lclTargetProperty,devProperty,PROPERTY_NAME_SIZE);
  }
  if (din != NULL && din->dArrayLength > 0) dinptr = din;
  access |= CA_MUTABLE;
  if (isSyncRepeater || (access & CA_WRITE))
  {
    cc = ExecLinkEx(lclTargetDeviceName,lclTargetProperty,dout,din,access,(UINT16)(rpPollingRate/2));
    sts = GetLastLinkError(cc, stsBuffer);
  }
  else
  {
    cc = alsnExecLink(lclTargetDeviceName,lclTargetProperty,dout,dinptr,access,rpPollingRate);
    sts = lastMexListenerStatusString;
    if (HAS_DATA(cc) && nSchPrps > 0)
    {
      char fullDeviceNameAndProperty[256];
      int p;
      for (p=0; p < nSchPrps && strncmp(schPrpLst[p].name,lclTargetProperty,PROPERTY_NAME_SIZE); p++);
      if (p < nSchPrps)
      {
        sprintf(fullDeviceNameAndProperty, "%.72s[%.64s]", lclTargetDeviceName, lclTargetProperty);
        tAttachNotifier(fullDeviceNameAndProperty, lsnrCb, p);
      }
    }
  }
  if (cc != 0 && (cc & CE_SENDDATA) == 0) 
  {
    SetEqmCompletion(RPTR_EQM,sts);
    if ((access & CA_HIST) == 0) cc += 1000;
    if (!isLocalRepeater && cc != link_timeout && cc != connection_timeout)
    {
      SetDeviceAlarm(RPTR_EQM, devName, target_value_unavailable, sts, almNEWALARM);
    }
  }
  return cc;
}
void rptini(void)
{
}
void rptbkg(void)
{
  if (!isLocalRepeater)
  {
    ClearAlarm(RPTR_EQM, -1);
  }
}
void rptexi(void)
{
}
FILE *createAndOpen(char *fname,char *attr)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  static char strdel[2] = { FS_DELIMITER , 0};

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,strdel); c != NULL; c=strtok(NULL,strdel))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return fopen(fname,attr);
    }
    strcat(dirname,strdel);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) return NULL;
    }
#   elif defined(WIN32)
    sprintf(fn,"%.120s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return NULL;
    }
#   endif
  }
  return NULL;
}
char *makeFormatString(int fmt,char *tag, int atype)
{
  static char fmtstr[128];
  strcpy(fmtstr,GetFormatAsString(fmt));
  if (isTaggableFormat((short)fmt) && strlen(tag) > 0)
  {
    strcat(fmtstr,".");
    strncat(fmtstr,tag,TAG_NAME_SIZE);
  }
  else
  {
    if (atype & AT_CHANNEL)
    {
      strcat(fmtstr,".CHANNEL");
    }
    else if (atype & AT_SPECTRUM)
    { 
      strcat(fmtstr,".SPECTRUM");
    }
  }
  return fmtstr;
}
int dumpStructure(char *tgtContext,char *tgtServer,char *tag)
{
  int i, n, fmt, cc = 0;
  FILE *fp = NULL;
  char fn[256], field[64], stag[64], *c, *fmtstr;
  NAME64DBLDBL ndd[256];
  static char lasttag[64] = "";

  if (tgtContext == NULL || tgtServer == NULL) ccerr(argument_list_error);
  if (tag == NULL || strlen(tag) == 0) ccerr(invalid_structure_tag);
  if (strncmp(tag,lasttag,TAG_NAME_SIZE) == 0) ccerr(0); /* nothing to do */
  n = 256;
  if ((cc=queryStructFormat(tag,(BYTE *)ndd,CF_NAME64DBLDBL,&n)) != 0) ccerr(cc);
  sprintf(fn,"CACHE%c%.32s%c%.32s%cSTRUCTURES%c%.16s-fields.csv",
    FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER,FS_DELIMITER,tag);
  if ((fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);

  fprintf(fp,"Field, Format, Size\n");

  for (i=0; i<n; i++)
  {
    strncpy(field,ndd[i].name,64);
    fmt = (int)ndd[i].d2val;
    if (fmt == CF_NULL) break;
    if (fmt == CF_STRUCT)
    {
      if ((c=strchr(field,'<')) != NULL)
      {
        c++;
        strncpy(stag,c,64);
        if ((c=strchr(stag,'>')) != NULL) *c++ = 0;
        strncpy(field,c,16);
        dumpStructure(tgtContext,tgtServer,stag);
      }
    }
    else
    {
      stag[0] = 0;
    }
    fmtstr = makeFormatString(fmt,stag,AT_SCALAR);
    fprintf(fp,"%.16s, %.64s, %d\n",field, fmtstr, (int)ndd[i].d1val);
  }
  strncpy(lasttag,tag,TAG_NAME_SIZE);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int dumpPropertyList(char *tgtContext,char *tgtServer,int idx)
{
  int i, cc = 0;
  FILE *fp = NULL;
  char fn[256];
  if (tgtContext == NULL || tgtServer == NULL) ccerr(argument_list_error);
  sprintf(fn,"CACHE%c%.32s%c%.32s%cPROPERTIES%c%.64s-properties.csv",
    FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER,FS_DELIMITER,rpDeviceNames[idx].name);
  if ((fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);

  fprintf(fp,"Property\n");
  for (i=0; i<plst[idx].lstsiz; i++)
  {
    fprintf(fp,"%.64s\n",plst[idx].plst[i].name);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int dumpDevices(char *tgtContext,char *tgtServer)
{
  int i, cc = 0;
  FILE *fp = NULL;
  char fn[256];
  if (tgtContext == NULL || tgtServer == NULL) ccerr(argument_list_error);
  sprintf(fn,"CACHE%c%.32s%c%.32s%crp-devices.csv",
    FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER);
  if ((fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  if (plst != NULL)
  {
    fprintf(fp,"Number, Name, PropertyList\n");
    for (i=0; i<ndevices; i++)
    {
      sprintf(fn,"%.64s-properties.csv",rpDeviceNames[i].name);
      fprintf(fp,"%d, %.64s, %s\n",i,rpDeviceNames[i].name,fn);
      dumpPropertyList(tgtContext,tgtServer,i);
    }
  }
  else
  {
    fprintf(fp,"Number, Name\n");
    for (i=0; i<ndevices; i++)
    {
      fprintf(fp,"%d, %.64s\n",i,rpDeviceNames[i].name);
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int dumpProperties(char *tgtContext,char *tgtServer)
{
  int i, fmt, fmtIn, cc = 0;
  FILE *fp = NULL;
  char fn[256];
  if (tgtContext == NULL || tgtServer == NULL) ccerr(argument_list_error);
  sprintf(fn,"CACHE%c%.32s%c%.32s%crp-properties.csv",
    FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER);
  if ((fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Property, Size, Format, SizeIn, FormatIn, Access, Description, Max, Min, Units, XMax, XMin, XUnits\n");
  for (i=0; i<nproperties; i++)
  {
    fmt = LFMT(rpPrps[i].prpFormat); fmtIn = LFMT(rpPrps[i].prpFormatIn);
    fprintf(fp,"%.64s, %d, %.16s, %d, %.16s, %.64s, %.64s, %g, %g, %.16s, %g, %g, %.16s\n",
      rpPrps[i].prpName,rpPrps[i].prpSize,
      makeFormatString(fmt,rpPrps[i].prpTag,rpPrps[i].prpArrayType),
      rpPrps[i].prpSizeIn,
      makeFormatString(fmtIn,rpPrps[i].prpTagIn,AT_SCALAR),
      GetAccessAsString((int)rpPrps[i].prpAccess),rpPrps[i].prpDescription,
      rpPrps[i].prpMaxValue,rpPrps[i].prpMinValue,rpPrps[i].prpUnits,
      rpPrps[i].rngMaxValue,rpPrps[i].rngMinValue,rpPrps[i].rngUnits);
    if (fmt == CF_STRUCT) dumpStructure(tgtContext,tgtServer,rpPrps[i].prpTag);
    if (fmtIn == CF_STRUCT) dumpStructure(tgtContext,tgtServer,rpPrps[i].prpTagIn);
  }

err:
  if (fp != NULL) fclose(fp);
  return cc;
}
typedef struct
{
  int number;
  char name[DEVICE_NAME_SIZE];
  char plst[PROPERTY_NAME_SIZE+16];
} RptDevice;

_CSVCOLUMN csvRpDevsColumns[] =
{
  { "NUMBER","0",0,CF_INT32,1,(BYTE *)OFFSETIN(RptDevice,number),NULL,0 },
  { "NAME","0",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(RptDevice,name),NULL,0 },
  { "PROPERTY_LIST","",0,CF_TEXT,PROPERTY_NAME_SIZE+16,(BYTE *)OFFSETIN(RptDevice,plst),NULL,0 },
};
CSVDB csvRpDevsDb =
{
  csvRpDevsColumns,
  sizeof(csvRpDevsColumns)/sizeof(csvRpDevsColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(RptDevice),
  0
};

_CSVCOLUMN csvRpPlstColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,NULL,NULL,0 },
};
CSVDB csvRpPlstDb =
{
  csvRpPlstColumns,
  sizeof(csvRpPlstColumns)/sizeof(csvRpPlstColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME64),
  0
};

typedef struct
{
  char prp[PROPERTY_NAME_SIZE];
  char fmtStr[64];
  char fmtStrIn[64];
  char accStr[64];
  char dsc[PROPERTY_DESC_SIZE];
  char egu[UNITS_SIZE];
  char xegu[UNITS_SIZE];
  int siz;
  int sizIn;
  float max;
  float min;
  float xmax;
  float xmin;
} RptProperty;

_CSVCOLUMN csvRpPrpsColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(RptProperty,prp),NULL,0 },
  { "SIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(RptProperty,siz),NULL,0 },
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(RptProperty,fmtStr),NULL,0 },
  { "SIZE_IN","0",0,CF_INT32,1,(BYTE *)OFFSETIN(RptProperty,sizIn),NULL,0 },
  { "FORMAT_IN","",0,CF_TEXT,64,(BYTE *)OFFSETIN(RptProperty,fmtStrIn),NULL,0 },
  { "ACCESS","READ",0,CF_TEXT,64,(BYTE *)OFFSETIN(RptProperty,accStr),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,PROPERTY_DESC_SIZE,(BYTE *)OFFSETIN(RptProperty,dsc),NULL,0 },
  { "MAX","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(RptProperty,max),NULL,0 },
  { "MIN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(RptProperty,min),NULL,0 },
  { "UNITS","",0,CF_TEXT,UNITS_SIZE,(BYTE *)OFFSETIN(RptProperty,egu),NULL,0 },
  { "XMAX","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(RptProperty,xmax),NULL,0 },
  { "XMIN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(RptProperty,xmin),NULL,0 },
  { "XUNITS","",0,CF_TEXT,UNITS_SIZE,(BYTE *)OFFSETIN(RptProperty,xegu),NULL,0 },
};
CSVDB csvRpPrpsDb =
{
  csvRpPrpsColumns,
  sizeof(csvRpPrpsColumns)/sizeof(csvRpPrpsColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(RptProperty),
  0
};

typedef struct
{
  char fld[16];
  int siz;
  char fmtStr[64];
} RptStruct;

_CSVCOLUMN csvRpStructColumns[] =
{
  { "FIELD","0",0,CF_TEXT,16,(BYTE *)OFFSETIN(RptStruct,fld),NULL,0 },
  { "SIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(RptStruct,siz),NULL,0 },
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(RptStruct,fmtStr),NULL,0 },
};
CSVDB csvRpStructDb =
{
  csvRpStructColumns,
  sizeof(csvRpStructColumns)/sizeof(csvRpStructColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(RptStruct),
  0
};

int getDevicesFromCache(char *tgtContext,char *tgtServer)
{
  int i, cc = 0;
  char path[256], fn[128];
  RptDevice *devs = NULL;

  if (tgtContext == NULL || tgtServer == NULL) return argument_list_error;
  sprintf(path,"CACHE%c%.32s%c%.32s%c",FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER);
  if ((cc=csvReadFile(path,"rp-devices.csv",&csvRpDevsDb,(void **)&devs)) != 0) return cc;
  ndevices = csvRpDevsDb.siz;
  if (ndevices <= 0) return file_error;
  if ((rpDeviceNames=(NAME64 *)calloc(ndevices,sizeof(NAME64))) == NULL) return out_of_local_memory;
  sprintf(path,"CACHE%c%.32s%c%.32s%cPROPERTIES%c",
    FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER,FS_DELIMITER);
  for (i=0; i<ndevices; i++) 
  {
    if (i != devs[i].number)
    {
      dbglog("device database corrupt: %.64s claims number %d and should be %d",
        devs[i].name,devs[i].number,i);
      return database_not_loaded;
    }
    strncpy(rpDeviceNames[i].name,devs[i].name,64);
    if (strlen(devs[i].plst) > 0)
    {
      if (plst == NULL)
      {
        if ((plst=(PLST *)calloc(ndevices,sizeof(PLST))) == NULL) return out_of_local_memory;
      }
      sprintf(fn,"%.64s-properties.csv",rpDeviceNames[i].name);
      if ((cc=csvReadFile(path,fn,&csvRpPlstDb,(void **)&plst[i].plst)) != 0) return cc;
      plst[i].lstsiz = csvRpPlstDb.siz;
    }
  }
  free(devs);
  return cc;
}
int getPropertiesFromCache(char *tgtContext,char *tgtServer)
{
  int i, cc = 0;
  char *c, path[256];
  RptProperty *prps = NULL;

  if (tgtContext == NULL || tgtServer == NULL) return argument_list_error;
  sprintf(path,"CACHE%c%.32s%c%.32s%c",FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER);
  if ((cc=csvReadFile(path,"rp-properties.csv",&csvRpPrpsDb,(void **)&prps)) != 0) return cc;
  nproperties = csvRpPrpsDb.siz;
  if (nproperties <= 0) return file_error;
  if ((rpPrps=(PrpQueryStruct *)calloc(nproperties,sizeof(PrpQueryStruct))) == NULL) return out_of_local_memory;
  for (i=0; i<nproperties; i++)
  {
    strncpy(rpPrps[i].prpName,prps[i].prp,PROPERTY_NAME_SIZE);
    strncpy(rpPrps[i].prpDescription,prps[i].dsc,PROPERTY_DESC_SIZE);
    rpPrps[i].prpSize = prps[i].siz;
    rpPrps[i].prpSizeIn = prps[i].sizIn;
    rpPrps[i].prpMaxValue = prps[i].max;
    rpPrps[i].prpMinValue = prps[i].min;
    rpPrps[i].rngMaxValue = prps[i].xmax;
    rpPrps[i].rngMinValue = prps[i].xmin;
    strncpy(rpPrps[i].prpUnits,prps[i].egu,UNITS_SIZE);
    strncpy(rpPrps[i].rngUnits,prps[i].xegu,UNITS_SIZE);

    rpPrps[i].prpAccess = (BYTE)GetAccessMode(prps[i].accStr);

    rpPrps[i].prpFormat = (BYTE)BFMT(GetFormatFromString(prps[i].fmtStr));
    rpPrps[i].prpFormatIn = (BYTE)BFMT(GetFormatFromString(prps[i].fmtStrIn));
    if (LFMT(rpPrps[i].prpFormat) == CF_STRUCT)
    {
      if ((c=strchr(prps[i].fmtStr,'.')) != NULL)
      {
        c++;
        strncpy(rpPrps[i].prpTag,c,TAG_NAME_SIZE);
      }
    }
    else
    {
      rpPrps[i].prpArrayType = GetArrayTypeFromString(prps[i].fmtStr);
    }
    if (LFMT(rpPrps[i].prpFormatIn) == CF_STRUCT)
    {
      if ((c=strchr(prps[i].fmtStrIn,'.')) != NULL)
      {
        c++;
        strncpy(rpPrps[i].prpTagIn,c,TAG_NAME_SIZE);
      }
    }
  }
  free(prps);
  return cc;
}
int getStructFromCache(char *tgtContext,char *tgtServer,char *tag,int siz)
{
  int cc = 0, i, offset = 0, fmt, nfields;
  char *c, path[256], fn[128];
  RptStruct *fields = NULL;

  if (tgtContext == NULL || tgtServer == NULL) return argument_list_error;
  sprintf(path,"CACHE%c%.32s%c%.32s%cSTRUCTURES%c",FS_DELIMITER,tgtContext,FS_DELIMITER,tgtServer,FS_DELIMITER,FS_DELIMITER);
  sprintf(fn,"%.16s-fields.csv",tag);
  if ((cc=csvReadFile(path,fn,&csvRpStructDb,(void **)&fields)) != 0) return cc;
  nfields = csvRpStructDb.siz;
  if (nfields <= 0) return file_error;
  for (i=0; i<nfields; i++)
  {
    fmt = GetFormatFromString(fields[i].fmtStr);
    addFieldToStruct(tag,offset,fields[i].siz,fmt,fields[i].fld);
    offset += fields[i].siz * fmtsizeof(fmt);
    if (fmt == CF_STRUCT)
    {
      if ((c=strchr(fields[i].fmtStr,'.')) != NULL)
      {
        c++;
        if ((cc=getStructFromCache(tgtContext,tgtServer,c,fields[i].siz)) != 0)
        {
          feclog("couldn't read nested structure %.16s from cache",c);
        }
      }
    }
  }
  if ((cc=sealTaggedStruct(tag,offset,siz)) != 0) 
  {
    feclog("couldn't seal tagged structure %.16s : %.32s",tag,cc2str(cc));
  }
  free(fields);
  return cc;
}
