/*
 * File shmalloc.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file shmalloc.c
 * \brief shared memory routines for windows
 *
 */
#ifndef NO_TINE_PIPE
#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <time.h>
#include "prolog.h"
#include "srvcore.h"
#include "dbtools.h"
#include "srvip.h"
#include "shmalloc.h"
#ifdef MULTITHREADED
#include "threader.h"
HANDLE shmCreateThread(void);
DWORD dwShmThreadId;
HANDLE hShmThread = 0;
#endif

#ifdef VS_LEGACY_BUILD
typedef __int64 LONG_PTR;
#endif
SECURITY_ATTRIBUTES mmfSA;
SECURITY_DESCRIPTOR mmfSD;
HWND hShmWnd = 0;
HINSTANCE hParent = 0;
int gForceShmDatagrams = TRUE;
int gMmfInitDone = FALSE;
int gStopShmMessageLoop = FALSE;
UINT32 gShmBufferSize = SHM_BUFFER_SIZE;
TINE_EXPORT void SetMMFMessageSize(int sizeInBytes)
{
  if (hShmWnd != 0)
  {
    feclogEx(TINE_LOGLEVEL_WARN,"late attempt to set shared memory buffer size refused");
    return;
  }
  if (sizeInBytes < 8192) sizeInBytes = 8192;
  gShmBufferSize = sizeInBytes;
  feclog("shared memory buffer size set to %d bytes",gShmBufferSize);
  if (gShmBufferSize > SHM_BUFFER_SIZE)
  {
    gUseMmfJumbos = TRUE;
    feclog("shared memory now using jumbo messages");
  }
}
TINE_EXPORT int GetMMFMessageSize(void)
{
  return gShmBufferSize;
}
static char shmWndName[32];

__declspec(dllexport) LRESULT CALLBACK shmWndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
  char shmTag[64];
  shmDataGram *dg;
  HANDLE hDG;
  int siz;
  HWND hClnWnd;
  ClnHdr *cln;
  switch (iMessage)
  {
    case WM_REQ_LCL: /* server getting a request */
      if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick*100) == 0)
      {
        hClnWnd = (HWND)(LONG_PTR)lParam;
        if ((cln=LocateConsumerInList(NULL,NULL,(BYTE *)&hClnWnd)) == NULL) 
        {
          ReleaseSystemMutex(hSystemKernelMutex);
          break;
        }
        memset(shmTag,0,64);
#       ifdef OS_64BIT
        _snprintf(shmTag,16,"_shmTCon:%I64x", (LONG_PTR)hShmWnd);
#       else
        _snprintf(shmTag,16,"_shmTCon:%x",(UINT32)hShmWnd);
#       endif
        cln->inetProtocol = MMF;
        cln->ep.wndMmf = (HWND)(LONG_PTR)lParam;
        siz = (int)wParam;
        dg = (shmDataGram *)shmGetDataGram(shmTag,siz,&hDG);
        if (dg != NULL)
        { /* serious problem if this is NULL !*/
          if (tineDebug) sprintf(dbgbuf,"SHM: [LOCAL] ");
          strncpy(cln->usr,dg->dataTag,16);
          getConsumerRequest(cln,shmPayload(dg),dg->size);
          dg->dirty = 0;
          shmFreeDataGram(dg,hDG);
        }
#       ifndef MULTITHREADED
        ReturnToServerCycle++;
#       endif
        ReleaseSystemMutex(hSystemKernelMutex);
      }
      break;
    case WM_RCV_LCL: /* client getting a reply */
      memset(shmTag,0,64);
#     ifdef OS_64BIT
      _snprintf(shmTag,16,"_shmTDat:%I64x", (LONG_PTR)hShmWnd);
#     else
      _snprintf(shmTag,16,"_shmTDat:%x",(UINT32)hShmWnd);
#     endif
      siz = (int)wParam;
      dg = (shmDataGram *)shmGetDataGram(shmTag,siz,&hDG);
      if (dg != NULL)
      { /* serious problem if this is NULL !*/
        //TODO: protocol 7
        UINT32 psiz = gUseMmfJumbos ? siz : 0;
        gInAdr.wndMmf = (HWND)(LONG_PTR)lParam;
        strncpy(gFecAddrBuffer,"local pipe",16);
        doProcessData(psiz,shmPayload(dg),&lParam,MMF,"MMF");
        dg->dirty = 0;
        shmFreeDataGram(dg,hDG);
      }
#     ifndef MULTITHREADED
      ReturnToClientCycle++;
#     endif
      break;
    case WM_STOP_LCL:
      gStopShmMessageLoop = TRUE;
      break;
    case WM_NCCREATE:  /* known work-around if CreateWindow fails */
      return 1;
      break;
    default:
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
static int shmCreateWindowDone = FALSE;
void DestroyShmWindow(void)
{
  if (hShmWnd != 0 && strlen(shmWndName) > 0)
  {
    feclog("closing down shared memory communication");
    if (DestroyWindow(hShmWnd) == FALSE) feclog("destroy shm window: 0x%x",GetLastError());
    if (UnregisterClass(shmWndName,hParent) == FALSE) feclog("unregister shm window class: 0x%x",GetLastError());
  }
  else
  {
    feclog("shared memory communication not active");
  }
  hShmWnd = 0;
  shmCreateWindowDone = FALSE;
  gMmfInitDone = FALSE;
}
extern int gIsWindowsConsole;

void OnCreateShmFailure(void)
{
  if (gNoLoopback)
  {
    char *msg = "client is enforcing shared memory access but cannot open the necessary shared memory window!";
    dbglog(msg);
    if (dieFunction != NULL) dieFunction(msg);
    if (gDieAnotherDay) { gDieAnotherDay = FALSE; return; }
    dbglog("aborting program!");
    exit(1);
  }
  gUseLoopback = TRUE;
}
#ifndef HWND_MESSAGE
# define HWND_MESSAGE 0
#endif
int CreateShmWindow(void)
{
  static int shmCc = -1;
  static WNDCLASS wndclass;
  DWORD hConsole = 0;
  int err, isConsole = FALSE;
  char lpConsoleTitle[256];

  if (hParent == 0) return not_applicable;
  if (shmCreateWindowDone) return shmCc;
  if ((hConsole=GetConsoleTitle(lpConsoleTitle,256)) != 0) 
  {
    isConsole = TRUE;
  }
  else
  { /* default was TRUE; set this to FALSE : */
    gIsWindowsConsole = FALSE;
  }
  if (!strcmp(gFecName,FECNAME_UNKNOWN))
  { /* not a FEC ? */
#   ifdef OS_64BIT
    sprintf(shmWndName,"%I64x", (LONG_PTR)hParent);
#   else
    sprintf(shmWndName,"%x",(UINT32)hParent);
#   endif
    //if (isConsole) return shmCc;
  }
  else
  { /* definitely a FEC */
    strncpy(shmWndName,gFecName,16); 
  }
  shmWndName[16] = 0;

  shmCreateWindowDone = TRUE;
  feclog("starting shared memory communication (window name : %.32s, instance %x)",shmWndName,hParent);
  if (hShmWnd != 0 && FindWindow(shmWndName,NULL) != 0) 
  { /* still active */
    feclog("shared memory communication still active : take no action !");
    return shmCc = 0; 
  }
  
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS; 
  wndclass.lpfnWndProc   = (WNDPROC) shmWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hParent;
  wndclass.hIcon         = LoadIcon (0, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (0, IDC_ARROW);
  wndclass.hbrBackground = GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = 0;
  wndclass.lpszClassName = shmWndName;

  feclog("register shared memory windows class");
  if (!RegisterClass(&wndclass)) 
  {
    if ((err=GetLastError()) != ERROR_CLASS_ALREADY_EXISTS)
    {
      feclog("cannot register shared memory window class : %d",err);
      OnCreateShmFailure();
      return -1;
    }
  }

  feclog("create shared memory window");
  if (!(hShmWnd = CreateWindow (
			shmWndName, shmWndName,               
			WS_BORDER|WS_CAPTION,         
			CW_USEDEFAULT, CW_USEDEFAULT, 
			CW_USEDEFAULT, CW_USEDEFAULT, 
			HWND_MESSAGE, 0,                 
			hParent,               
			(LPSTR)""                     
      ) ) ) 
  {
    err = GetLastError();
    feclog("cannot create shared memory window : %d",err);
    OnCreateShmFailure();
    return -1;
  }

  return shmCc = 0;
}

int shmStart(void)
{
  int cc = 0, wait_cnts = 5;
  char *ptr = getenv("TINE_MMF_MESSAGE_SIZE");
  if (ptr != NULL) SetMMFMessageSize(atoi(ptr));
# ifdef MULTITHREADED
  if (hShmThread == NULL) hShmThread = shmCreateThread();
  cc = hShmThread == NULL ? -1 : 0;
# else
  cc = CreateShmWindow();
# endif
  while (hShmWnd == 0 && cc == 0) 
  { // make sure I've got a window listening for the reply
    Sleep(gSystemTick);
    if (wait_cnts-- == 0) cc = -1;
  }
  return cc;
}
int shmStop(void)
{
  int cc, err;
  if (hShmWnd == NULL) return 0;
# if !defined(USE_SENDMESSAGE_TIMEOUT) && defined(MULTITHREADED)
  /* there is NO way to get this thread to stop gracefully : */
  feclog("stopping shared memory communication");
  DestroyShmWindow();
  err = TerminateThread(hShmThread,0);
# else
  err = SendMessageTimeout(hShmWnd,WM_STOP_LCL,0,0,SMTO_NORMAL,gSystemTick*2,NULL);  // <- this doesn't work !
# endif
  cc = (err == 0 ? GetLastError() : 0);
# ifdef MULTITHREADED
  if (cc == 0) 
  {
    hShmThread = NULL;
  }
  else
  {
    feclog("cannot close shm message loop");
  }
# endif
# ifndef USE_SENDMESSAGE_TIMEOUT
  DestroyShmWindow();
# endif
  hShmWnd = 0;
  return cc;
}
int sendShmData(ClnHdr *cln,BYTE *data,int size,UINT iMessage)
{
  char shmTag[64];
  HANDLE hShm = 0;
  if (size > (int)(gShmBufferSize-sizeof(shmDataGram))) return use_stream_transport;
  memset(shmTag,0,64);
  _snprintf(shmTag,16,"%s:%x",iMessage == WM_RCV_LCL ? "_shmTDat" : "_shmTCon"  ,(LONG)(LONG_PTR)cln->ep.wndMmf);
  return shmSndDataGram(cln,shmTag,data,size,&hShm,iMessage);
}
int mmfInit(void)
{
  InitializeSecurityDescriptor(&mmfSD,SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(&mmfSD,TRUE,NULL,FALSE);
  mmfSA.nLength = sizeof(mmfSA);
  mmfSA.lpSecurityDescriptor = &mmfSD;
  mmfSA.bInheritHandle = TRUE;
  return 0;
}
HANDLE mmfalloc(char *tag,int SizeInBytes,BYTE **lpView,int *mmfFlag)
{
  HANDLE hFileMap;
  int MMFattached = 0;

  if (!gMmfInitDone)
  {
    gMmfInitDone = TRUE;
    mmfInit();
  }
  SizeInBytes = (int)gShmBufferSize; // maximum allowed at the moment
  if (tag == NULL || strlen(tag) == 0) return NULL;
  if ((hFileMap=CreateFileMapping(hSHARE,&mmfSA,PAGE_READWRITE,0,SizeInBytes,tag)) != NULL)
  {
    MMFattached = (GetLastError() == ERROR_ALREADY_EXISTS);
    if ((*lpView=(BYTE *)MapViewOfFile(hFileMap,FILE_MAP_WRITE,0,0,0)) == NULL)
    {
      char dbgerr[256] = "";
      rprtSckErr("mmfalloc",tag,0,errno,0,dbgerr);
      CloseHandle(hFileMap);
      return NULL;
    }
  }
  if (mmfFlag) *mmfFlag = MMFattached;
  return hFileMap;
}
int shmSndDataGram(ClnHdr *cln,char *tag,BYTE *data,int size,HANDLE *h,UINT iMessage)
{
  shmDataGram *dg = NULL;
  int cc = connection_timeout,rsize, MMFattached = 0, err = -1;
  char *c;
  HWND hWnd;
  LPVOID lpMsgBuf;
# ifdef OS_64BIT
  DWORD_PTR *pdw = NULL;
# endif
  if ((c=strchr(tag,':')) == NULL) return out_of_shared_memory; c++;
  hWnd = (HWND)(LONG_PTR)strtol(c,NULL,16);
  rsize = sizeof(shmDataGram) + size;

  if ((*h=mmfalloc(tag,rsize,(BYTE **)&dg,&MMFattached)) == NULL)
  {
    dbglog("shmSndDataGram error: %s (%x)",tag,GetLastError());
    *h = 0;
    return out_of_shared_memory;
  }
  memcpy(shmPayload(dg),data,size);
  dg->size = size;
  dg->dirty = TRUE;
  memcpy(dg->dataTag,gUserName,16);

# ifdef OS_64BIT
  err = (int)SendMessageTimeout(hWnd,iMessage,size,(long)(LONG_PTR)hShmWnd,SMTO_NORMAL,gSystemTick*10,(PDWORD_PTR)&pdw);
  cc = (err != 0 && pdw != NULL) ? (int)(*pdw) : 0;
# else
  err = (int)SendMessageTimeout(hWnd,iMessage,size,(long)hShmWnd,SMTO_NORMAL,gSystemTick*10,(LPDWORD)&cc);
# endif
  if (MMFattached) shmFreeDataGram(dg,*h);
  if (err == 0 || cc < 0 || cc > numErr)
  {
    cc = connection_timeout;
    if (err == 0) 
    {
      err = GetLastError();
      FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,err,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,0,NULL);
      if (tineDebug > 1 || errorDebug) dbglog("SendMessageTimeout: %s",lpMsgBuf);
    }
    if (tineDebug > 1) dbglog("shmSndDataGram: %d (message returned %d)",err,cc);
  }
  return cc;
}
shmDataGram *shmGetDataGram(char *tag,int size,HANDLE *h)
{
  shmDataGram *dg = NULL;
  int rsize,MMFattached=0;

  rsize = sizeof(shmDataGram) + size;

  if ((*h=mmfalloc(tag,rsize,(BYTE **)&dg,&MMFattached)) == NULL)
  {
    dbglog("shmGetDataGram shm error: %s (%x)",tag,GetLastError());
    *h = 0;
    return NULL;
  }
  return dg;
}
void shmFreeDataGram(shmDataGram *dg,HANDLE h)
{
  if (h == 0) return;
  UnmapViewOfFile(dg);
  CloseHandle(h);
}
BYTE *shmDataBlockAlloc(char *tag,int size,HANDLE *hBlock)
{
  BYTE *d = NULL;
  int MMFattached=0;

  if ((*hBlock=mmfalloc(tag,size,&d,&MMFattached)) == NULL)
  {
	 dbglog("shm error: %s",tag);
	 *hBlock = 0;
	 return NULL;
  }
  return d;
}
void shmDataBlockFree(BYTE *d,HANDLE hBlock)
{
  if (hBlock == 0) return;
  UnmapViewOfFile(d);
  CloseHandle(hBlock);
}
TINE_EXPORT void SetForceShmDatagrams(int value)
{
  gForceShmDatagrams = value;
}
extern int hasClientEvents;
extern int hasServerEvents;
void DoClientEvents(int maxevents)
{
# ifndef MULTITHREADED
  MSG msg;
  int i = 0;
  while (PeekMessage(&msg,hShmWnd,WM_RCV_LCL,WM_RCV_LCL,PM_REMOVE))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    if (maxevents > 0 && i++ > maxevents) break;
  }
  hasClientEvents = i;
# else
  maxevents = maxevents;
# endif
}
void DoServerEvents(int maxevents)
{
# ifndef MULTITHREADED
  MSG msg;
  int i = 0;
  while (PeekMessage(&msg,hShmWnd,WM_REQ_LCL,WM_REQ_LCL,PM_REMOVE))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    if (maxevents > 0 && i++ > maxevents) break;
  }
  hasServerEvents = i;
# else
  maxevents = maxevents;
# endif
}
#ifdef MULTITHREADED
THREAD_TYPE shmTaskThread(THREAD_PARAM lpvThreadParam)
{
  MSG msg;

  if (CreateShmWindow() != 0) return -1;
  SetThreadName("ShmTask");
  //while (GetMessage(&msg,hShmWnd,WM_REQ_LCL,WM_STOP_LCL) > 0)
  while (GetMessage(&msg,NULL,0,0) > 0)
  { /* below is useless: those Microsoft engineers have seen to it that this NEVER drops through ! */
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    if (msg.message == WM_STOP_LCL) break;
    if (msg.message == WM_QUIT) break;
    if (gStopShmMessageLoop) break;
    if (ServerExitCondition) break;
  }
  DestroyShmWindow();
  return 0;
}
HANDLE shmCreateThread(void)
{
  HANDLE hThread = 0;
  int threadPriority = -1;
  char *ptr = getenv("TINE_SHM_THREAD_PRIORITY");

  hThread = CreateThread(NULL,0,shmTaskThread,0,0,&dwShmThreadId);
  if (ptr != NULL) threadPriority = atoi(ptr);
  if (threadPriority > 0 && hThread != NULL) 
  {
    feclog("shared memory thread priority set to %d via environment",threadPriority);
    SetThreadPriority(hThread,threadPriority);
  }
  return hThread;
}
#endif /* MULTITHREADED*/
#endif /* !NO_TINE_PIPE */