/*
 * File srvip.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file srvip.c
 * \brief TCP/UDP routines and definitions
 * \internal
 *
 */
#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "prolog.h"
#include "srvcore.h"
#include "srvip.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"
#include "ttycmd.h"
#if defined(WIN32) && !defined(NO_TINE_PIPE)
#include "shmalloc.h"
#endif
#endif /* NO_DOXYGEN */
#ifdef UNIX
int ipcSrvSck = 0;
int nIPCsck = 0;
char gIpcTinePipe[64];
IPCSCKT *ipcSckTbl = NULL;
int ipcConTblSize = MAX_IPCCONNECTIONS;
#endif

UINT16 gPortOffset = PORTOFFSET;
int ReturnToServerCycle = 0;
TINE_EXPORT int gRespondToServiceRequests = RESPOND_TO_SERVICE_REQUESTS;
TINE_EXPORT void SetRespondToServiceRequests(int value)
{
  gRespondToServiceRequests = value;
}
TINE_EXPORT int GetRespondToServiceRequests(void)
{
  return gRespondToServiceRequests;
}
int gInitIPServicesDone = FALSE;
int gWaitingOnSelect = FALSE;
extern int nBusyContractProgress;
int removeClientTableEntry(int i);
char *netsListFile[2] = { IPNETS_FILE, IPBCASTNETS_FILE };
char *netsListTag[2] = { NETS_ALLOWED_SETNAME, NETS_NETWORK_SETNAME};
char gSrvPipeName[32] = { 0 };
SOCKET tcpSrvSck = 0, tcpStrmSck = 0; /* TCP sockets */
SOCKET udpSrvSck = 0, udpSyncSck = 0; /* UDP sockets */
int hasServerEvents = 0;
extern int hasClientEvents;
extern int gServerCycleNeedsScheduler;
extern UINT32 gTimeFdIdle;
#define hasEvents (hasClientEvents || hasServerEvents)
int gAllowRemoteDebugging = TRUE;
TINE_EXPORT void SetAllowRemoteDebugging(int value)
{
  gAllowRemoteDebugging = value;
}
TINE_EXPORT int GetAllowRemoteDebugging(void)
{
  return gAllowRemoteDebugging;
}
SOCKET dbgSrvSck = 0;
TCPSCKT *dbgSckTbl = NULL;
int dbgConTblSize = MAX_DBGCONNECTIONS;
int nDBGsck = 0;
SOCKET netSrvSck = 0;
TCPSCKT *tcpSckTbl = NULL;
int nTCPsck = 0;
int tcpConTblSize = MAX_TCPCONNECTIONS;
TINE_EXPORT void SetTcpConnectionTableSize(int value)
{
  if (value < 8) value = 8;
  tcpConTblSize = value;
  feclog("set tcp connection table size to %d",tcpConTblSize);
}
TINE_EXPORT int GetTcpConnectionTableSize(void)
{
  return tcpConTblSize;
}
void (*socketAcceptEntryFcn)(void) = NULL;
int loadIP(void)
{
# ifdef DOS
  char *ptr=NULL;
  int i,s=0;
  if (!loaded()) return system_error;
  if ((ptr=getenv("KEEPSOCKETS")) != NULL)
  {
    s = atoi(ptr);
  }
  for (i=63; i > s; i--)
  {
    soclose(i);
    shutdown(i,2); /* shutdown any open ports */
  }
# endif
# ifdef WIN32
  if (!WinsockLoaded()) return system_error;
  MaxPollingRate = MIN_POLLRATE;
# endif
  return 0;
}
SCKACL *IpBcstLst = NULL;
int sizIpBcstLst = 0;

int MaxNumAllowedIPNets = MAX_NUM_IPNETS;
TINE_EXPORT int AppendRegisteredIPNetsList(NAME16 *ipaddr,int addsiz)
{
  UNUSED(ipaddr);
  UNUSED(addsiz);
  dbglog("AppendRegisteredIPNetsList() is deprecated; please use AppendRegisteredNetsList() instead");
  return function_deprecated;
}
TINE_EXPORT int RemoveRegisteredIPNetsList(NAME16 *ipaddr,int rmvsiz)
{
  UNUSED(ipaddr);
  UNUSED(rmvsiz);
  dbglog("AppendRegisteredIPNetsList() is deprecated; please use AppendRegisteredNetsList() instead");
  return function_deprecated;
}
void dumpRegisteredNetsFile(NAME64 *iplst,SCKACL *addrlst,int lstsiz,FILE *fp)
{
  int i;
  char ipstr[64];

  for (i=0; i<lstsiz; i++) 
  {
    if (addrlst != NULL)
    {
      if (addrlst[i].addr.ss_family == 0xee) continue;
      inetStrAdr(&addrlst[i].addr,ipstr,FALSE);
      addIpSubStr(ipstr,addrlst[i]);
      fprintf(fp,"%.64s\n",ipstr);
    }
    else if (iplst != NULL)
    {
      if (iplst[i].name[0] == 0) continue;
      fprintf(fp,"%.64s\n",iplst[i].name);
    }
  }
}
int idxInNetsLst(char *ipsa,SCKACL *lst,int siz)
{
  int i;
  char *c;
  BYTE adr[16];
  int fam = gAfInet;
  int clen = fam == AF_INET ? 4 : 16;
  int msk = clen * 8; /* default: must match */
  if ((c=strchr(ipsa,'/')) != NULL) 
  {
    *c++ = 0; msk = atoi(c);
  }
  inet_pton(fam,ipsa,adr);
  for (i=0; i<siz; i++) 
  {
    if (!memcmp(getInAddr(&lst[i].addr),adr,clen)) 
    { /* address in the list, now check the CIDR mask */
      if (lst[i].mask == msk) return i;
    }
  }
  return -1;
}
int addRegNetsLst(const char *filename,int type,int *nr,SCKACL **list,NAME64 *ipaddr,int addsiz)
{
  int cc=0, k, numToAdd=0, nlist = *nr;
  SCKACL *lst = *list;
  char ipsa[64];
# if !defined(FS_RDONLY) && !defined(FS_NONE)
  /* normal case: there is a file system */
  FILE *fp = NULL;
  char netsfilename[256], *fn;
  if (type < 0 || type > 1) ccerr(no_such_file);
  if (isUsingXMLFecDatabase) ccerr(not_implemented);
  memset(&gLastAllowedCln,0,sizeof(ClnHdr)); /* reset this */
  fn = filename == NULL ? netsListFile[type] : (char *)filename;
  if (ipaddr != NULL && addsiz > 0 && strlen(ipaddr[0].name) > 0)
  {
    for (k=0; k<addsiz; k++) 
    {
      strntrm(ipaddr[k].name,64);
      strncpy(ipsa,ipaddr[k].name,64);
      if (idxInNetsLst(ipsa,lst,nlist) < 0) numToAdd++; else ipaddr[k].name[0] = 0;
    }
    if (numToAdd == 0) return 0; /* already in list */
    sprintf(netsfilename,"%.128s%.32s",fecDbPath,fn);
    if ((fp=fopen(netsfilename,"w")) == NULL) ccerr(no_such_file);
    fprintf(fp,"SUBNET\n");
    dumpRegisteredNetsFile(NULL,lst,nlist,fp);
    dumpRegisteredNetsFile(ipaddr,NULL,addsiz,fp);
    fclose(fp);
  }
  getRegNetsFromFile(NULL,NULL,NULL,filename,type,nr,list);
# else
  int msk = 0, fam = gAfInet, len;
  BYTE ipadr[16];
  UINT16 *p;
  char *c;
  if (lst == NULL)
  {
    if ((lst=(SCKACL *)SystemCalloc(MaxNumAllowedIPNets,sizeof(SCKACL))) == NULL)
      ccerr(out_of_server_memory);
    *list = lst;
    nlist = 0;
  }
  if (ipaddr == NULL || addsiz < 1 || strlen(ipaddr[0].name) == 0) ccerr(argument_list_error);
  for (k=0; k<addsiz; k++) 
  {
    strntrm(ipaddr[k].name,64);
    strncpy(ipsa,ipaddr[k].name,64);
    if (idxInNetsLst(ipsa,lst,nlist) < 0) numToAdd++; else ipaddr[k].name[0] = 0;
  }
  if (numToAdd == 0) return 0; /* already in list */
  if (nlist == MaxNumAllowedIPNets) ccerr(resources_exhausted);
  for (k=0; k<addsiz; k++)
  {
    if (ipaddr[k].name[0] == 0) continue;
    strncpy(ipsa,ipaddr[k].name,64);
    if ((c=strchr(ipsa,'/')) != NULL) *c++ = 0;
    msk = fam == AF_INET ? 32 : 128;
    inet_pton(fam,ipsa,ipadr);
    if (c != NULL) 
    {
      msk = atoi(c);
    }
    else if ((c=strrchr(ipsa,'.')) != NULL)
    {
      c++;
      if (!strcmp(c,"255")) msk = 24;
      if (!strcmp(c,"0") && strncmp(ipsa,"0.",2)) msk = 24;
    }
    if (msk < 16) msk = 16;
    lst[nlist].mask = (BYTE)msk;
    lst[nlist].addr.ss_family = fam;
    p = fam == AF_INET ? (UINT16 *)&((struct sockaddr_in *)&lst[nlist].addr)->sin_port :
                         (UINT16 *)&((struct sockaddr_in6 *)&lst[nlist].addr)->sin6_port;
    *p = htons((UINT16)gtGCastPort); /* potential broadcasting */
    len = fam == AF_INET ? 4 : 16;
    memcpy(&lst[nlist].addr,ipadr,len);
    if (++nlist == MaxNumAllowedIPNets) break;
  }
  *nr = nlist;
# endif
err:
  return cc;
}
int rmvRegNetsLst(const char *filename,int type,int *nr,SCKACL **list,NAME64 *ipaddr,int rmvsiz)
{
  int cc=0, i, k, numToRemove=0, nlist = *nr;
  SCKACL *lst = *list;
  char ipsa[64];
# if !defined(FS_RDONLY) && !defined(FS_NONE)
  /* normal case: there is a file system */
  FILE *fp = NULL;
  char netsfilename[256], *fn;
  if (type < 0 || type > 1) ccerr(no_such_file);
  if (isUsingXMLFecDatabase) ccerr(not_implemented);
  memset(&gLastAllowedCln,0,sizeof(ClnHdr)); /* reset this */
  fn = filename == NULL ? netsListFile[type] : (char *)filename;
  if (ipaddr != NULL && rmvsiz > 0 && strlen(ipaddr[0].name) > 0)
  {
    for (k=0; k<rmvsiz; k++) 
    {
      strntrm(ipaddr[k].name,64);
      strncpy(ipsa,ipaddr[k].name,64);
      if ((i=idxInNetsLst(ipsa,lst,nlist)) >= 0) 
      {
        lst[i].addr.ss_family = 0xee;
        numToRemove++; 
      }
    }
    if (numToRemove == 0) return 0; /* was not in list anyway */
    sprintf(netsfilename,"%.128s%.32s",fecDbPath,fn);
    if (nlist == numToRemove)
    { /* remove the whole file */
      unlink(netsfilename);
      return 0;
    }
    if ((fp=fopen(netsfilename,"w")) == NULL) ccerr(no_such_file);
    fprintf(fp,"SUBNET\n");
    dumpRegisteredNetsFile(NULL,lst,nlist,fp);
    fclose(fp);
  }
  getRegNetsFromFile(NULL,NULL,NULL,filename,type,nr,list);
# else
  if (lst == NULL) return 0; /* no list anyway */
  if (ipaddr == NULL || rmvsiz < 1 || strlen(ipaddr[0].name) == 0) ccerr(argument_list_error);
  for (k=0; k<rmvsiz; k++) 
  {
    strntrm(ipaddr[k].name,64);
    strncpy(ipsa,ipaddr[k].name,64);
    if ((i=idxInNetsLst(ipsa,lst,nlist)) >= 0) 
    {
      lst[i].addr.ss_family = 0xee;
      numToRemove++; 
    }
  }
  if (numToRemove == 0) return 0; /* was not in list anyway */
  for (i=0, k=0; i<nlist; i++) 
  {
    if (lst[i].addr.ss_family != 0xee)
    {
      if (i != k) memcpy(&lst[k++],&lst[i],sizeof(SCKACL));
    }
  }
  nlist = k;
  *nr = nlist;
# endif
err:
  return cc;
}
_CSVCOLUMN csvIpNetsFileColumns[] = 
{
  { "SUBNET","",0,CF_TEXT,64,NULL,NULL,0 }
};
CSVDB csvIpNetsFileDb =
{
  csvIpNetsFileColumns,
  sizeof(csvIpNetsFileColumns)/sizeof(csvIpNetsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME64),
  0
};
int getAllowedNets(const char *eqm)
{
  int cc = 0, i;
  char ipstr[32];
  ExportListStruct *el=getExportListItem((char *)eqm);
# if defined(UNIX) || defined(WIN32)
  struct stat sbuf;
  char pfn[256], fn[64];
# endif
  if (gTineConfigurationCoded) return not_implemented;
  if (el == NULL) return argument_list_error;
  if (strlen(el->netsFileName) == 0)
  {
    strncpy(el->netsFileName,IPNETS_FILE,32);
#   if defined(UNIX) || defined(WIN32)
    sprintf(fn,"%.6s%c%.32s",eqm,FS_DELIMITER,IPNETS_FILE);
    sprintf(pfn,"%.128s%.64s",fecDbPath,fn);
    if (stat(pfn,&sbuf) == 0) 
    { /* subdirectory file exists: use it! */
      strncpy(el->netsFileName,fn,32);
    }
#   endif
  }
  cc = getRegNetsFromFile(eqm,NULL,NULL,el->netsFileName,NETS_ACCESS,&el->aclLst.nNets,&el->aclLst.nets);
  if (cc == 0) 
  { /* print configuration info to log file */
    el->chkAcl |= EQM_CHECK_NET;
    if (el->aclLst.nets == 0) 
    {
      feclogEx(TINE_LOGLEVEL_WARN, "%.6s: NO IP address has WRITE access",eqm);
      dbglog("WARNING: empty file %32s found => NO IP address has WRITE access to %.6s",el->netsFileName,eqm);
    }
    else 
    {
      for (i=0; i<el->aclLst.nNets; i++) 
      {
        inetStrAdr(&el->aclLst.nets[i].addr,ipstr,FALSE);
        addIpSubStr(ipstr,el->aclLst.nets[i]);
        feclog("%.6s: address %.32s has WRITE access",eqm,ipstr);
      }
    }
  }
  else
  {
    feclog("%.6s: ALL IP addresses have WRITE access",eqm);
#   if defined(UNIX) || defined(WIN32)
    if (cc == no_such_file)
    { /* then reset the prefered file location */
      sprintf(fn, "%.6s%c%.32s", eqm, FS_DELIMITER, IPNETS_FILE);
      strncpy(el->netsFileName,fn, 32);
    }
#   endif
  }
  return cc;
}
int getRegNetsFromFile(const char *eqm,const char *dev,const char *prp,const char *filename,int type,int *nr,SCKACL **list)
{
  int size=0,n,fam=gAfInet,msk=0;
  int cc=no_such_file;
  SCKACL *lst=NULL;
  NAME64 *nets = NULL;
  char *fn, *c, ipsa[64];
  BYTE badr[16];
  
  if (gTineConfigurationCoded) return not_implemented;
  if (type < 0 || type > 1) ccerr(no_such_file); 
  if ((lst=*list) != NULL) SystemFree(lst);
  *list = NULL;
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  { /* okay, give this a try */
    fn = netsListTag[type];
    size = populateNameList(gFecInfoList,gFecName,(char *)eqm,(char *)dev,(char *)prp,fn,&nets);
    cc = (size < 0) ? -size : (size == 0) ? not_defined : 0;
  }
# endif
  if (cc != 0)
  { /* try the csv file alternative */
    fn = filename == NULL ? netsListFile[type] : (char *)filename;
    cc = csvReadFile(fecDbPath,fn,&csvIpNetsFileDb,(void **)&nets);
    size = csvIpNetsFileDb.siz;
  }
  if (cc != 0 || size == 0) ccerr(cc);

  lst = (SCKACL *)SystemCalloc(MaxNumAllowedIPNets,sizeof(SCKACL));
  if (lst == NULL) return out_of_server_memory;
  for (n=0; n<size; n++)
  {
    strncpy(ipsa,nets[n].name,64);
    if ((c=strchr(ipsa,'/')) != NULL) *c++ = 0;
    inet_pton(fam,ipsa,badr);
    sckadr_in(&lst[n].addr,badr,(UINT16)gtGCastPort,fam);
    msk = fam == AF_INET ? 32 : 128;
    if (c != NULL) 
    {
      msk = atoi(c);
    }
    else if ((c=strrchr(ipsa,'.')) != NULL)
    {
      c++;
      if (!strcmp(c,"255")) msk = 24;
      if (!strcmp(c,"0") && strncmp(ipsa,"0.",2)) msk = 24;
    }
    if (msk < 16) msk = 16;
    lst[n].mask = (BYTE)msk;
  }
err:
  SystemFree(nets);
  if (nr != NULL) *nr = size;
  if (list != NULL) *list = lst;
  return cc;
}
int setupTcpServerSocket(int family,int port)
{
  SCKADR sckadr;
  int optval = 1, optlen = sizeof(int), mode, cc = 0;
  int sck = 0;
  struct linger to;
  char txt[50];

  mode = 1;
  sckadr_in(&sckadr,NULL,(UINT16)(port+gPortOffset),family);
  sprintf(txt,"TCP PORT %d: ",port+gPortOffset);

  if ((sck=(int)socket(family,SOCK_STREAM,0)) < 0) soerr("socket");
# if defined(UNIX)
  optval = fcntl(sck,F_GETFD,0); optval |= FD_CLOEXEC; fcntl(sck, F_SETFD, optval);
# endif  
# if defined(FIX_IPV6_V6ONLY)
  if (family == AF_INET6)
  {
    optval = 0;
    if (setsockopt(sck,IPPROTO_IPV6,IPV6_V6ONLY,(char*)&optval,optlen)) soerr("IPV6_V6ONLY");
  }
# endif
  optval = 1;
  if (setsockopt(sck,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) soerr("REUSEADDR");
  to.l_onoff = 0; to.l_linger = 0;
  if (setsockopt(sck,SOL_SOCKET,SO_LINGER,(char *)&to,sizeof(to))) soerr("LINGER");
  optval = 1;
  if (setsockopt(sck,SOL_SOCKET,SO_KEEPALIVE,(char *)&optval,sizeof(optval))) soerr("KEEPALIVE");
  if (bind(sck,(struct sockaddr *)&sckadr,getInAddrLen(family))) soerr("bind");
  if (listen(sck,NumTCPChan) < 0) soerr("listen");
  if (gIsCyclePollingMode && soioctl(sck,FIONBIO,(char *)&mode)) soerr("NONBLOCKING");

  feclog("%s initialized",txt);
err:
  if (cc != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s socket error",txt);
    rprtSckErr("setupTcpServerSocket",txt,sck,-1,0,NULL);
    if (sck > 0) soclose(sck);
    sck = 0;
  }
  return sck;
}
#if defined(UNIX) || defined(WIN32) || defined(MACOS)
TINE_EXPORT int OpenIpcPipe(char *pipeName)
{
  char ipcfn[128];
  if (ipcCmdSck != 0) return file_already_exists;
  sprintf(ipcfn,"%s%.32s.ipc",IPC_PATH,pipeName);
  if ((ipcCmdSck=pipeListen(ipcfn)) == 0) return file_error;
  strncpy(gSrvPipeName,pipeName,32);
  return 0;
}
TINE_EXPORT int openIPCPipe(char *pipeName)
{
  return OpenIpcPipe(pipeName);
}
TINE_EXPORT int CloseIpcPipe(char *pipeName)
{
# ifdef UNIX
  char ipcfn[128], *pn = pipeName;
  if (pn == NULL) pn = gSrvPipeName; /* passing NULL -> use the stored name */
  if (pn == NULL || strlen(pn) == 0) return invalid_parameter;
  if (ipcCmdSck == 0) return 0;
  if (gIsRunningAsServer)
  {
    if (strstr(pn,gFecName) != NULL) return not_accepted;
  }
  sprintf(ipcfn,"%s%.32s.ipc",IPC_PATH,pn);
  unlink(ipcfn);
  if (soclose(ipcCmdSck) != 0) return unix_socket_error; 
  ipcCmdSck = 0;
# elif defined(WIN32)
# if !defined(NO_SERVICE_PIPE)
  pipeName = pipeName;
  if (ipcCmdSck == 0) return 0;
  CloseHandle(((PIPEINST *)ipcCmdSck)->hPipeInst);
  ipcCmdSck = 0;
# endif
# else
  pipeName = pipeName;
# endif
  return 0;
}
TINE_EXPORT int GetIpcPipeName(char *pipeName)
{
  if (ipcCmdSck == 0) return not_initialized;
  strncpy(pipeName,gSrvPipeName,32);
  return 0;
}
#endif
void chkIpcClnLst(struct pollfd *fds,int nfds)
{
# if defined(UNIX) && !defined(NO_SERVICE_PIPE)
  int len,i,k,idx,close_it=FALSE;
  char cmdbuf[CMDSIZE];
  if (ipcCmdSck > 0 && fds_isset(ipcCmdSck)) addIPCcln();
  for (i=0; i<ipcClnLstSize; i++)
  {
    close_it = FALSE;
    if (fds_isset(IPCfds[i]))
    {
      if ((len=GetCommandEx(IPCfds[i],cmdbuf,CMDSIZE)) > 0) InterpretCommand(cmdbuf);
      if (len == 0) close_it = TRUE;
    }
    if (fds_iserr(IPCfds[i])) close_it = TRUE;
    if (close_it)
    {
      close(IPCfds[i]);
      fds[idx].events = fds[idx].revents = 0;
      for (k=i; k<ipcClnLstSize-1; k++) IPCfds[k] = IPCfds[k+1];
      i--; ipcClnLstSize--;
      msglog(0,"Remote session closed");
#     if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
      srvlog("commands.log", "DETACH PIPE", "session terminated");
#     endif
    }
  }
# else
  fds = fds; nfds = nfds;
# endif
}
void chkDbgClnLst(struct pollfd *fds,int nfds)
{
  int len,i,idx,close_it=FALSE;
  char cmdbuf[CMDSIZE];
  for (i=0; i<dbgConTblSize; i++)
  {
    close_it = FALSE;
    if (fds_isset(dbgSckTbl[i].sck))
    {
      if ((len=GetCommandFromStream(dbgSckTbl[i].sck,cmdbuf,CMDSIZE)) > 0) InterpretCommand(cmdbuf);
      if (len == 0) close_it = TRUE;
    }
    if (fds_iserr(dbgSckTbl[i].sck)) close_it = TRUE;
    if (close_it)
    { /* don't spin if there are descriptor problems ! */
      fds[idx].events = fds[idx].revents = 0;
      CloseConnectionSocket(dbgSckTbl[i].sck,DBGPIPE,TRUE);
      msglog(0,"Remote session closed");
      i--; /* CloseConnectionSocket will decrease the table size */
    }
  }
}
#ifdef WIN32
void chkWinPipe(void)
{
# if !defined(NO_SERVICE_PIPE)
  char cmdbuf[CMDSIZE];
  if (pipeAccept((PIPEINST *)ipcCmdSck) != 0)
  {
    if (GetCommandEx(ipcCmdSck,cmdbuf,CMDSIZE) > 0) InterpretCommand(cmdbuf);
  }
# endif
  return;
}
#endif
int initIPServer(void)
{
  int optlvl, opt, optval = 1, optlen = sizeof(int), cc = 0;
  UINT16 p;
  char txt[64], *ptr=NULL;
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_SERVICE_PIPE)
  char ipcfn[32];
# endif

  if (gInitIPServicesDone) return 0; 
  if (WaitForSystemMutex(hSystemKernelMutex,-1) != 0) return mutex_error; 
  if ((cc=loadIP()) != 0) ccerr(cc);
  strcpy(txt,"initIPServer() : ");
  feclog("Using Port Offset %d",gPortOffset);
  tcpSrvSck = setupTcpServerSocket(gAfInet,gtTCPPort);
# ifdef MULTITHREADED
  tcpStrmSck = setupTcpServerSocket(gAfInet,gtStrmPort);
# endif
  if (gAllowRemoteDebugging)
  {
    dbgSrvSck = setupTcpServerSocket(gAfInet,gtPipePort);
  }
  p = (UINT16)(gtUDPPort + gPortOffset);
  if (setupUdpSocket(&udpSrvSck,gAfInet,p,TRUE,max_srvrcvbuf,max_srvsndbuf) != 0) soerr("init udpSrvSck");
# ifndef DOS
  if (setsockopt(udpSrvSck,SOL_SOCKET,SO_BROADCAST,(char *)&optval,optlen)) soerr("opt BROADCAST");
# endif /* DOS */
# ifndef NOT_MULTICASTING
  {
    int ttl = gtMCastTTL;
    char *optval = (char *)&ttl;
#   ifdef UNIX
    BYTE bttl = gtMCastTTL;
    optval = gAfInet == AF_INET6 ? (char *)&ttl : (char *)&bttl;
    optlen = gAfInet == AF_INET6 ? sizeof(ttl) : sizeof(bttl);
#   endif
    setmcopts(optlvl,opt);
    if (setsockopt(udpSrvSck,optlvl,opt,optval,optlen)<0) soerr("opt MULTICAST");
  }
# endif /* NOT_MULTICASTING */
  if (gRespondToServiceRequests)
  {
    p = (UINT16)(gtNetSrvPort);
    if (setupUdpSocket(&netSrvSck,gAfInet,p,TRUE,MIN_RCVBUF,MIN_SNDBUF) != 0) soerr("init netSrvSck");
  }
  p = 0;
  if (setupUdpSocket(&udpSyncSck,gAfInet,0,FALSE,MIN_RCVBUF,MIN_SNDBUF) != 0) soerr("init udpSyncSck");
  getRegNetsFromFile(NULL,NULL,NULL,NULL,NETS_NETWORK,&sizIpBcstLst,&IpBcstLst);
  if ((ptr=getenv("TINE_GLOBALS_BROADCAST_ONLY")) != NULL)
  {
    SetSendGlobalsAsBroadcastOnly(strtobool(ptr));
  }
  ipLoaded = TRUE;
# if defined(WIN32) && !defined(NO_SERVICE_PIPE)
  sprintf(ipcfn,"%s%s.ipc",IPC_PATH,gFecName);
  if ((ipcCmdSck=pipeListen(ipcfn)) == 0) 
  { 
    feclogEx(TINE_LOGLEVEL_ERROR, "Cannot start attachfec service");
  }
  else 
  {
    feclog("IPC server listening on %s",ipcfn);
    strncpy(gSrvPipeName,gFecName,32);
  }
# endif /* WIN32 */
# ifdef UNIX
  signal(SIGPIPE, SIG_IGN); /* ignore any broken pipe signals */
  signal(SIGTTIN, SIG_IGN); /* ignore any tty in signals */
  signal(SIGTTOU, SIG_IGN); /* ignore any tty out signals */
# ifndef NOTTY
  if (foregroundTTY && (foregroundTTY=isatty(STDIN_FILENO)) != 0)
  {
    if (gStartupDebug) printf("running in foreground\n");
  }
# else
  foregroundTTY = FALSE;
# endif /* NOTTY */
# ifndef NO_SERVICE_PIPE
  sprintf(ipcfn,"%s%s.ipc",IPC_PATH,gFecName);
  if ((ipcCmdSck=pipeListen(ipcfn)) < 0) 
  { 
    perror("IPC listen"); msglog(-ipcCmdSck,"pipe listener returned %d",ipcCmdSck); 
    feclogEx(TINE_LOGLEVEL_ERROR, "Cannot start IPC attachfec Service (error %d)",ipcCmdSck);
  }
  else 
  {
    feclog("IPC server listening on %.32s (fd %d)",ipcfn,ipcCmdSck);
    strncpy(gSrvPipeName,gFecName,32);
  }
  /* if (bsdsignal(SIGPIPE, sig_pipe) == SIG_ERR) printf("signal error\n"); */
  if (bsdsignal(SIGHUP, sig_pipe) == SIG_ERR) printf("signal error\n");
# endif /* NO_SERVICE_PIPE */
# ifndef NO_TINE_PIPE
  sprintf(gIpcTinePipe,"%s%s-server.ipc",IPC_PATH,gFecName);
  if ((ipcSrvSck=pipeListen(gIpcTinePipe)) < 0)
  {
    perror("IPC Producer Socket"); printf("return %d\n",ipcSrvSck);
    feclogEx(TINE_LOGLEVEL_ERROR, "Cannot start IPC Server (error %d)",ipcSrvSck);
  }
  else
  {
    feclog("IPC Producer socket listening on %.64s (fd %d)",gIpcTinePipe,ipcSrvSck);
  }
  if ((ipcSckTbl=(IPCSCKT *)SystemCalloc(ipcConTblSize,sizeof(IPCSCKT))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "IPC connection buffer: couldn't allocate %d connections",ipcConTblSize);
  }
# endif /* NO_TINE_PIPE */
# endif /* UNIX */
  /* use tcpSckTbl as a table: */
  if ((tcpSckTbl=(TCPSCKT *)SystemCalloc(tcpConTblSize,sizeof(TCPSCKT))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "TCP connection buffer: couldn't allocate %d connections",tcpConTblSize);
  }
  /* use dbgSckTbl as a table: */
  if ((dbgSckTbl=(TCPSCKT *)SystemCalloc(dbgConTblSize,sizeof(TCPSCKT))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "DBG connection buffer: couldn't allocate %d connections",dbgConTblSize);
  }
err:
  ReleaseSystemMutex(hSystemKernelMutex);
  if (cc != 0)
  {
    rprtSckErr("initIPServer", txt, 0, -1, 0, NULL);
    sprintf(dbgbuf, "%s FATAL socket error (%d) with port offset %d (port %d)",
      txt, soerrno, gPortOffset, p);
    feclogEx(TINE_LOGLEVEL_ERROR, dbgbuf);
    if (dieFunction != NULL) dieFunction(dbgbuf);
    if (gDieAnotherDay) { gDieAnotherDay = FALSE; return port_not_available; }
    if (!gSufferInSilence) dbglog(dbgbuf);
#   if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
    if (soerrno == ERROR_ALREADY_EXISTS)
    {
      char hint[128];
      sprintf(hint, "\nPort %d is being used by another application on this PC.", p);
      strcat(dbgbuf, hint);
      sprintf(hint,"\nIs a server with the same port already running?");
      strcat(dbgbuf, hint);
    }
    MessageBox(NULL, dbgbuf, gFecName, IDOK);
#   endif
    exit(1); /* this is serious! don't try to start */
  }
  gInitIPServicesDone = TRUE;
  return cc;
}
int dbgPipeAllowedAccess(SCKADR addr)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el = el->next)
  {
    if (!(el->chkAcl & EQM_CHECK_NET)) continue;
    if (!IsMemberIPControlNets(el,&addr)) return FALSE;
  }
  return TRUE;
}
int AssignConnectionSocket(SOCKET parent,int inetProtocol,int *nsck,TCPSCKT *tbl,int capacity)
{
  int sck, optval, n, cc = 0;
  SCKADR addr;
  char ipstr[64];
  TCPBCKT *bckt;
  _socklen_t len = sizeof(addr);
  if (parent == 0) 
  {
    dbglog("attempt to assign a connection socket with no parent socket!");
    return -1;
  }
  if ((sck=(int)accept(parent,(struct sockaddr *)&addr,&len)) > 0)
  {
    if (tineDebug) dbglog("accepted %s daughter socket %d on parent %d",getProtocolAsString(inetProtocol),sck,parent);
    optval = max_srvsndbuf;
    if (setsockopt(sck,SOL_SOCKET,SO_SNDBUF,(char *)&optval,sizeof(optval))) rprtSckErr("AssignConnectionSocket","SO_SNDBUF",sck,-1,0,NULL);
    checkSckBufferSpace(sck,max_srvsndbuf,SO_SNDBUF,"TCP server");
    optval = max_srvrcvbuf;
    if (setsockopt(sck,SOL_SOCKET,SO_RCVBUF,(char *)&optval,sizeof(optval))) rprtSckErr("AssignConnectionSocket","SO_RCVBUF",sck,-1,0,NULL);
    checkSckBufferSpace(sck,max_srvrcvbuf,SO_RCVBUF,"TCP server");
    n = *nsck;
    if (n < 0) 
    {
      dbglog("invalid tcp socket table entry %d !",n); 
      ccerr(code_failure);
    }
    if (n < capacity)
    {
      if (inetProtocol&(DBGPIPE|DBGPIPE6))
      { /* avoid a warning on the exotic compiler used in doocs builds */
        SCKADR sa;
        memcpy(&sa,&addr,sizeof(SCKADR));
        if (!dbgPipeAllowedAccess(addr)) ccerr(access_denied);
        inetStrAdr(&sa,ipstr,FALSE);
        msglog(0,"remote debug access granted to %.40s",ipstr);
#       if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
        srvlog("commands.log", "ATTACH_FEC", "remote access from: %.40s", ipstr);
#       endif
      }
      if ((bckt=getBucket(sck,TCP_BUFFER_SIZE,0,inetProtocol,0,SRV_RSP_)) != NULL)
      {
        bckt->idx = -1;
        tbl[n].sck = sck;
        tbl[n].type = inetProtocol;
        if (tineDebug) dbglog("assigned tcp bucket %d (from %d) to persistent connection",n,capacity);
        (*nsck)++;
        return n;
      }
      msglog(0,"tcp bucket table cannot allocate bucket");
      cc = out_of_local_memory;
    }
    else
    {
      cc = resources_exhausted;
      msglog(0,"tcp bucket table capacity %d exhausted",capacity);
    }
err:
    msglog(cc,"closing daughter socket %d : %.32s",sck,cc2str(cc));
    soclose(sck);
  }
  return -1;
}
ClnHdr *getNetworkClientFromList(void)
{
  int i;
  for (i=0; i<nconsumers; i++)
  {
    if (strcmp(ClnTbl[i]->usr,"NETWORK")) continue;
    if  (inetPort(&ClnTbl[i]->adr.ip) != (UINT16)gtMCastPort) continue;
    return ClnTbl[i];
  }
  return NULL;
}
ClnHdr *LocateConsumerInList(SCKADR *ip,IpxAdr *nwAddr,void *hwnd)
{
  int i;
  HWND hWnd = 0;
# if !defined(IPXNW) && !defined(IPXBSD)
  UNUSED(nwAddr); /* avoid warning */
# endif
  if (hwnd != NULL) 
  {
#   ifdef MAKE_TINE32_DLL
    hWnd = *(HWND *)hwnd; 
#   endif
  }
  if (WaitForSystemMutex(hClientTableMutex,-1) != 0) return NULL;
  for (i=0; i<nconsumers; i++)
  { /* is client in client list ? */
    if (ClnTbl[i]->ncontracts == 0 && 
        nBusyContractProgress == 0 && 
        ClnTbl[i]->inetProtocol != PIPE)
    { /* remove any list entries with no contracts (PIPES exempted here) */
      removeClientTableEntry(i--);
      continue;
    }
    if (ip != NULL && sckadr_eq(ip,&ClnTbl[i]->adr.ip)) break;
#   if defined(IPXNW) || defined(IPXBSD)
    if (nwAddr != NULL && !memcmp(nwAddr,&ClnTbl[i]->adr.ipx,sizeof(IpxAdr))) break;
#   endif
    if (hWnd != 0 && ClnTbl[i]->ep.wndMmf == hWnd) break;
  }
  if (i < nconsumers) goto locateDone;
  if (i >= ClientListCapacity ||
     (ClnTbl[i]=(ClnHdr *)SystemCalloc(1,sizeof(ClnHdr))) == NULL)
  {
    if (tineDebug || errorDebug) 
    {
      if (i >= ClientListCapacity) dbglog("cannot add client to list: capacity %d has been reached",ClientListCapacity);
      else if (tineDebug) dbglog("cannot add client to list: out of memory");
    }
    ReleaseSystemMutex(hClientTableMutex);
    return NULL;
  }
  nconsumers++;
  if (ip != NULL) 
  {
    memcpy(&ClnTbl[i]->adr.ip,ip,sizeof(SCKADR));
  }
# if defined(IPXNW) || defined(IPXBSD)
  if (nwAddr != NULL) memcpy(&ClnTbl[i]->adr.ipx,nwAddr,sizeof(IpxAdr));
# endif
  if (hwnd != NULL) ClnTbl[i]->ep.wndMmf = hWnd;
locateDone:
  ReleaseSystemMutex(hClientTableMutex);
  return ClnTbl[i];
}
#ifdef UNIX
SCKADR *whichUnixAddr(int sck)
{
  int i;
  for (i=0; i<nIPCsck; i++)
  {
    if (ipcSckTbl[i].sck == sck)
      return &ipcSckTbl[i].ipAddr;      
  }
  return NULL;
}
int isUnixSocket(int sck)
{
  int i;
  if (sck == ipcSrvSck) return TRUE;
  for (i=0; i<nIPCsck; i++)
  {
    if (ipcSckTbl[i].sck == sck) return TRUE;       
  }
  for (i=0; i<numFecTblEntries; i++) if (FecTbl[i].inetProtocol == PIPE) 
  {
    if (FecTbl[i].lclAdr.pipe.sck == sck) return TRUE;
  }
  return FALSE;
}
int AcceptIPC(int sck)
{ /* TODO:  ipv6 code ? Unix sockets are ipv4-like */
  short cc = not_implemented;
# ifndef NO_TINE_PIPE
  TCPBCKT *bckt;
  ClnHdr *cln = NULL;
  SCKADR sckadr;
  struct sockaddr_in *sa4 = (struct sockaddr_in *)&sckadr;
  int s = 0;
  _socklen_t len = sizeof(struct sockaddr_in);
  char pipename[32];
# ifndef SVR4
  struct sockaddr_un unAddr;
  struct stat statbuf;
# else
  struct strrecvfd rfd;
# endif /* SVR4 */
  if (sck == ipcSrvSck) 
  {
#   ifndef SVR4
    len = sizeof(unAddr);
    if ((s=accept(sck,(struct sockaddr *)&unAddr,&len)) < 0) ccerr(not_accepted);
    strncpy(pipename,unAddr.sun_path,14); pipename[14] = 0;
    if (stat(pipename,&statbuf) < 0) ccerr(net_read_error);
    unlink(pipename);
#   else
    strncpy(pipename,"SVR4 pipe",32);
    if (ioctl(sck,I_RECVFD,&rfd) < 0) ccerr(-1);
    s = rfd.fd;
#   endif /* SVR4 */
    memset(&sckadr,0,sizeof(SCKADR));
    memcpy(&sa4->sin_addr,getmyipaddr(),4);
    sa4->sin_family = AF_UNIX;
    sa4->sin_port = (unsigned short)s;
    if ((cln=LocateConsumerInList(&sckadr,NULL,NULL)) == NULL) ccerr(out_of_server_memory);
    cln->ep.sck = s;
    cln->inetProtocol = PIPE;
    if ((bckt=getBucket(s,TCP_BUFFER_SIZE,0,PIPE,0,SRV_RSP_)) != NULL) bckt->idx = -1;
    ipcSckTbl[nIPCsck].sck = s;
    ipcSckTbl[nIPCsck].type = PIPE;
    memcpy(&ipcSckTbl[nIPCsck].ipAddr,&sckadr,sizeof(SCKADR));
    strncpy(ipcSckTbl[nIPCsck].name,pipename,16);
    msglog(0,"add ipc %.32s bucket %d (from %d) to persistent connection",pipename,nIPCsck,ipcConTblSize);
    nIPCsck++;
    cc = 0;
  }
err:
  if (cc)
  {
    if (tineDebug || errorDebug) dbglog("IPC ERR: %.32s (sck %d)",cc2str(cc),sck);
  }
# else
  UNUSED(sck);
# endif /* NO_TINE_PIPE */
  return cc;
}
#endif
int getInetFamily(int inetProtocol)
{
  switch (inetProtocol)
  {
    case UDP6:
    case TCP6:
    case STREAM6:
    case DBGPIPE6:
      return AF_INET6;
    case UDP:
    case TCP:
    case STREAM:
    case DBGPIPE:
      return AF_INET;
    case PIPE:
    case MMF:
      return AF_UNIX;
    case IPX:
      return AF_IPX;
    default:
      break;
  }
  return 0;
}
int disallowSingleDelivery = FALSE;
ClnHdr *getConsumerIPAddress(SOCKET sck,int inetProtocol)
{
  ClnHdr *cln = NULL;
  SCKADR ipAddr;
  int retry=2, family = getInetFamily(inetProtocol);
  int nread = 0,e,cc = -1;
  char *strprot = getProtocolAsString(inetProtocol);
  char strbuf[512],addr[64];
  struct pollfd fds[1];
  TCPBCKT *bckt;
  BYTE *bufptr=srvWorkArea;
  _socklen_t len = getInAddrLen(family);
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  SCKADR *pIpAddr=NULL;
# endif
  if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return NULL;
  /* don't block if there are stale data : POLLING mode bypasses getNextCycleTimeout() ! */
retry_select:
  fds[0].fd = sck; fds[0].events = POLLIN;
  if (poll(fds,1,0) == -1 || pollerr(0))
  {
    if ((e=soerrno) != EWOULDBLOCK && tineDebug) rprtSckErr("getConsumerIPAddress","select",sck,e,0,NULL);
    if (retry-- > 0) goto retry_select;
    ccerr(tcp_socket_error);
  }
  if ((fds[0].revents & POLLIN) == 0) goto err;
  cc = 0;
  if (inetProtocol == UDP) UDPcount++;
  if ((nread=readfrom(sck,(char *)srvWorkArea,srvWorkAreaSize,0,
             (struct sockaddr *)&ipAddr,&len)) <= 0) ccerr(net_read_error);
  switch (inetProtocol)
  {
#   if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
    case MMF:
      strncpy(addr,"localhost",ADDR_SIZE_IPV6);
      ipAddr.ss_family = family;
#     ifndef MULTITHREADED
      ReturnToServerCycle++;
#     endif
      if (tineDebug) sprintf(dbgbuf,"%s: [%s] ",strprot,addr);
      ccerr(0);
      break;
#   endif
    case UDP:
    case UDP6:
      UDPcount++;
      inetStrAdr(&ipAddr,addr,TRUE);
      ipAddr.ss_family = family;
      break;
    case STREAM:
    case STREAM6:
    case TCP: 
    case TCP6: 
      TCPcount++;      
      len = sizeof(ipAddr);
      if (getpeername(sck,(struct sockaddr *)&ipAddr,&len) < 0) ccerr(tcp_socket_error);
      inetStrAdr(&ipAddr,addr,TRUE);
      ipAddr.ss_family = family;
      break;
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    case PIPE:
      strncpy(addr,"localhost",ADDR_SIZE_IPV6);
      if ((pIpAddr=whichUnixAddr(sck)) == NULL) ccerr(net_read_error);
      ipAddr = *pIpAddr;
      ipAddr.ss_family = AF_UNIX;
      break;
#   endif      
  }
  if (tineDebug) sprintf(dbgbuf,"%s: [%s] ",strprot,addr);
  if (inetProtocol & (TCP|PIPE|STREAM|TCP6|STREAM6))
  { /* TODO: consider consolidation with processTCPDataStream() */
    int hsize = REQHDR_SIZE;
    int n, nleft = nread;
    if ((bckt=findBucket(sck,FALSE)) == NULL) ccerr(tcp_socket_error);
    while (nleft > 0)
    {
      if (bckt->ptr < hsize)
      { /* collecting a new header */
        ReqHdr reqhdr;
        int bckptr = bckt->ptr;
        int bckbytes = nleft > bckt->len ? bckt->len : nleft;
        memcpy(&bckt->buf[bckt->ptr],bufptr, bckbytes);
        bckt->ptr += bckbytes;
        if (bckt->ptr < hsize)
        { /* still not enought bytes in to form a header */
          nleft = 0;
          break;
        }
        prepIncomingRequestHdr(&reqhdr, bckt->buf);
        bckt->siz = reqhdr.totalSizeInBytes; /* now we know what size a chunk is ... */
        bckt->ptr = bckptr; /* rewind the stream for processing below ... */
      }
      if ((n=bckt->siz-bckt->ptr) <= 0) 
      {
        msglog(illegal_data_size,
          "tcp bucket size %d, ptr %d -> %d bytes but nbytes = %d (from %d)!",
          bckt->siz,bckt->ptr,n,nleft,nread);
        ccerr(illegal_data_size);
      }
      if (nleft >= n)
      { /* full bucket: time to interpret */
        memcpy(&bckt->buf[bckt->ptr],bufptr,n);
        if ((cln=LocateConsumerInList(&ipAddr,NULL,NULL)) == NULL) 
          ccerr(out_of_server_memory);
        cln->ep.sck = sck;  cln->inetProtocol = inetProtocol;
        disallowSingleDelivery = (nleft > n) ? TRUE : FALSE;
        getConsumerRequest(cln,bckt->buf,UDP_BUFFER_SIZE);
        bufptr += n;
        nleft -= n;
        bckt->siz = bckt->ptr = 0; /* reset these */
      }
      else
      { /* continue collecting */
        memcpy(&bckt->buf[bckt->ptr],bufptr,nleft);
        bckt->ptr += nleft;
        nleft = 0;
      }
    }
    ccerr(0);
  }
  if ((cln=LocateConsumerInList(&ipAddr,NULL,NULL)) == NULL) ccerr(out_of_server_memory);
  cln->ep.sck = sck;
  cln->inetProtocol = inetProtocol;
  getConsumerRequest(cln,srvWorkArea,UDP_BUFFER_SIZE);
err:
  ReleaseSystemMutex(hSystemKernelMutex);
  if (cc > 0)
  {
    if ((inetProtocol & (TCP|STREAM|TCP6|STREAM6)) && (!gIsCyclePollingMode || soerrno != EWOULDBLOCK))
    {
      msglog(cc,"closing connection socket %d due to error",sck);
      CloseConnectionSocket(sck,inetProtocol,TRUE);
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    if ((inetProtocol & PIPE) && (!gIsCyclePollingMode || soerrno != EWOULDBLOCK))
    {
      ClosePipeSocket(sck);
      return NULL; /* skip the debug ouput */
    }
#   endif
    if (tineDebug) 
    {
      sprintf(strbuf,"%s ERR: %.32s <errno: %d>",strprot,cc2str(cc),soerrno);
      rprtSckErr("getConsumerIPAddress",strbuf,sck,errno,0,NULL);
    }
  }
  return cln;
}
void dumpIPNets(void)
{
  int i, idx;
  char ipstr[64];
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  for (el=ExportList; el != NULL; el = el->next)
  {
    dbglog("Equipment Module : %.6s\n",el->EqmName);
    if (!el->chkAcl)
    {
      dbglog("\tWrite Access open to all networks\n");
      continue;
    }
    if (el->chkAcl & EQM_CHECK_NET)
    {
      dbglog("\tWrite Access open to :\n");
      for (i=0; i<el->aclLst.nNets; i++)
      {
        inetStrAdr(&el->aclLst.nets[i].addr,ipstr,FALSE);
        addIpSubStr(ipstr,el->aclLst.nets[i]);
        dbglog("\t  %.32s\n",ipstr);
      }
    }
#   ifndef SMALL_TINE_LIB
    if (el->chkAcl & PRP_CHECK_NET)
    {
      for (idx=0; idx<PROPERTY_HASH_SIZE; idx++)
        for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
      {
        if (prp->isReserved) continue;
        if (prp->aclLst.nNets > 0)
        {
          dbglog("    Property : %.64s\n",prp->prpName);
          dbglog("\tWrite Access open to :\n");
          for (i=0; i<prp->aclLst.nNets; i++)
          {
            inetStrAdr(&prp->aclLst.nets[i].addr,ipstr,FALSE);
            addIpSubStr(ipstr,prp->aclLst.nets[i]);
            dbglog("\t  %.64s\n",ipstr);
          }
        }
      }
    }
    if (el->chkAcl & DEV_CHECK_NET)
    {
      for (idx=0; idx<el->EqmNumDevices && el->EqmDevLst != NULL; idx++)
      {
        if (el->EqmDevLst[idx].aclLst == NULL) continue;
        if (el->EqmDevLst[idx].aclLst->nNets > 0)
        {
          dbglog("    Device : %.64s\n",el->EqmDevLst[idx].name);
          dbglog("\tWrite Access open to :\n");
          for (i=0; i<el->EqmDevLst[idx].aclLst->nNets; i++)
          {
            inetStrAdr(&el->EqmDevLst[idx].aclLst->nets[i].addr,ipstr,FALSE);
            addIpSubStr(ipstr,el->EqmDevLst[idx].aclLst->nets[i]);
            dbglog("\t  %.32s\n",ipstr);
          }
        }
      }
    }
#   else
    prp = NULL;
    idx = 0;
#   endif
  }
}
int isSubNet(int mask,BYTE *tgt,BYTE *src)
{
  int i, k, bit, nb = mask/8, nr = mask%8;
  for (i=0; i<nb; i++)
  { /* these bytes must match */
    if (tgt[i] != src[i]) return FALSE;    
  }
  for (k=0; k<nr; k++)
  { /* these bits must match */
    bit = (1 << (7-k));
    if ((tgt[i] & bit) != (src[i] & bit)) return FALSE;
  }
  return TRUE;
}
int IsMemberIPNets(int num,SCKACL *lst,SCKADR *ipaddr)
{
  int n;
  int off = ipaddr->ss_family == AF_INET ? OFFSETIN(struct sockaddr_in,sin_addr) : OFFSETIN(struct sockaddr_in6,sin6_addr);
  BYTE *ipadr, *lstadr;

  if (ipaddr == NULL) return FALSE;
  for (n=0; n<num; n++)
  {
    lstadr = (BYTE *)(&lst[n].addr) + off;
    ipadr = (BYTE *)ipaddr + off;
    if (isSubNet(lst[n].mask,lstadr,ipadr)) return TRUE;
  }
  return FALSE;
}
int gSndIpBroadcastsOnly = FALSE;
TINE_EXPORT void SetSendGlobalsAsBroadcastOnly(int value)
{
  feclog("GLB: setting SendGlobalsAsBroadcastOnly to %s",value ? "TRUE" : "FALSE");
  gSndIpBroadcastsOnly = value;
  if (!value) return;
  if (sizIpBcstLst == 0)
  {
    INADR ipadr, myadr;
    if (IpBcstLst == NULL) IpBcstLst = (SCKACL *)SystemCalloc(1,sizeof(SCKACL));
    if (IpBcstLst == NULL) return;
    sizIpBcstLst = 1;
    memcpy(&myadr,getmyipaddr(),gInAdrLen);
    applyBcstAddr(gAfInet,&myadr,&ipadr);
    sckadr_in(&IpBcstLst->addr,(BYTE *)&ipadr,(UINT16)gtGCastPort,gAfInet);
    feclog("GLB: using local subnet as broadcast net");
  }
}
int sendIPgcast(BYTE *buf,int size)
{
  int cc = 0, n, ncasts;
  struct pollfd fds[1];
  SCKADR *psckadr;
  char *tag = NULL, dbgerr[256];

  if (!ipLoaded) return tcp_not_supported;
  ncasts = gSndIpBroadcastsOnly ? sizIpBcstLst : sizIpBcstLst + 1;
  if (ncasts == 0) ccerr(configuration_error);
  if (WaitForSystemMutex(hSndMutex,100) != 0) return semaphore_busy;
  dbgerr[0] = 0;
  for (n=0; n<ncasts; n++)
  {
    fds[0].fd = udpSrvSck; fds[0].events = POLLOUT; fds[0].revents = 0;
    switch (poll(fds,1,0))
    {
      case 1: if (!pollerr(0)) break;
      case 0: 
        if (tineDebug > 1 || errorDebug) sprintf(dbgerr,"socket %s: write set not ready",getSckStr(udpSrvSck));
        ccerr(udp_socket_error);
      default:
        rprtSckErr("sendIPgcast","select",udpSrvSck,-1,2,NULL);
        ccerr(udp_socket_error);
    }
    if (n < sizIpBcstLst)
    { /* there are configured broadcast nets ! */
      psckadr = &IpBcstLst[n].addr;
      tag = "broadcast";
    }
    else
    { /* normal: just the multicast socket */
      psckadr = &tGCastSckAddr;
      tag = "multicast";
    }
    if (sendto(udpSrvSck,(char *)buf,size,0,(struct sockaddr *)psckadr,getInAddrLen(psckadr->ss_family)) < 0)
    {
      if (tineDebug || errorDebug) sprintf(dbgerr,"UDP: %s failed",tag);
      ccerr(tcp_socket_error);
    }
    else if (tineDebug > 1)
    {
      char addr[64];
      inetStrAdr(psckadr,addr,FALSE);
      sprintf(dbgerr,"UDP %s : %.30s ",tag,addr);
    }
  }
err:
  ReleaseSystemMutex(hSndMutex);
  if (strlen(dbgerr)) dbglog(dbgerr);
  return cc;
}
void closeIPServer(void)
{
  if (!ipLoaded) return;
  feclog("closing IP server subsystem");
  if (tcpSrvSck)
  {
    shutdown(tcpSrvSck,2);
    soclose(tcpSrvSck);
  }
  tcpSrvSck = 0;
  if (tcpStrmSck)
  {
    shutdown(tcpStrmSck,2);
    soclose(tcpStrmSck);
  }
  tcpStrmSck = 0;
  if (dbgSrvSck)
  {
    shutdown(dbgSrvSck,2);
    soclose(dbgSrvSck);
  }
  dbgSrvSck = 0;
  if (udpSrvSck)
  {
    shutdown(udpSrvSck,2);
    soclose(udpSrvSck);
  }
  udpSrvSck = 0;
  if (netSrvSck)
  {
    shutdown(netSrvSck,2);
    soclose(netSrvSck);
  }
  netSrvSck = 0;
  if (udpSyncSck) 
  {
    shutdown(udpSyncSck,2);
    soclose(udpSyncSck);
  }
  udpSyncSck = 0;
  return;
}
int gSelectOnSysPoll = TRUE;
int gSystemCycleIdleTime = 0;
struct timeval *getNextCycleTimeout(struct timeval *tv)
{ /* don't block if there are stale data : */
  int needToPoll = (StaleData || ReturnToServerCycle || hasEvents);
  unsigned int deadtime;
  static short firstCycle = 1;
  static struct timeval lastCycleTime;
  static struct timeval to;

  if (!gRunServerCycleInSeparateThread && ReturnToClientCycle) needToPoll = TRUE;
  if (firstCycle) { lastCycleTime = *tv; firstCycle = 0; }
  if (needToPoll)
  { /* no delay ... */
    to.tv_sec = 0;
    to.tv_usec = 0;
    ReturnToServerCycle = 0;
    return &to;
  }
  deadtime = MSECS(*tv,lastCycleTime);
  if (gSelectOnSysPoll) deadtime = 0; /* TRUE if nothing on the fd set */
  lastCycleTime = *tv;
  if (deadtime >= (UINT32)gSystemCycleDeadband) deadtime = gSystemCycleDeadband;
  gSystemCycleIdleTime = gSystemCycleDeadband - deadtime;
  if ((to.tv_sec=(gSystemCycleIdleTime)/1000) < 0)
  { /* a really long delay if this is secs < 0 !*/
    to.tv_sec = 0;
    to.tv_usec = 1000;
  }
  else 
  { /* normal case: get the correct usecs */
    to.tv_usec = ((gSystemCycleIdleTime)%1000)*1000;
  }
  return &to;
}
int IsInForeground(void)
{
# ifdef UNIX
  struct pollfd fds[1];

  if (gIsCyclePollingMode) return -1;
  if (!foregroundTTY) return 0;
  fds[0].fd = STDIN_FILENO; fds[0].events = POLLIN;
  if (poll(fds,1,0) < 0 || pollerr(0)) return 0;
# endif
  return -1;
}
int isServerSocket(SOCKET sck)
{
  int i;
  if (sck == 0) return FALSE;
  if (sck == udpSrvSck) return TRUE;
  if (sck == tcpSrvSck) return TRUE;
  if (sck == tcpStrmSck) return TRUE;
  if (sck == dbgSrvSck) return TRUE;
  if (sck == netSrvSck) return TRUE;
  for (i=0; i<nTCPsck; i++) 
  { /* either TCP or STREAM handles requests in the same way */
    if (sck == tcpSckTbl[i].sck) return TRUE;
  }
  for (i=0; i<nDBGsck; i++) 
  { /* debug pipe requests */
    if (sck == dbgSckTbl[i].sck) return TRUE;
  }
  return FALSE;
}
int addServerSocketsToSet(struct pollfd *fds,int offset,int maxlen)
{
  int i, n=offset;
  memset(&fds[offset],0,sizeof(struct pollfd)*(maxlen-offset));
  if (!gServerCycleIdle)
  {
    ADDSCKTOSET(tcpSrvSck);
    ADDSCKTOSET(tcpStrmSck);
    for (i=0; i<nTCPsck; i++) 
    { /* either TCP or STREAM handles requests in the same way */
      ADDSCKTOSET(tcpSckTbl[i].sck);
    }
    ADDSCKTOSET(udpSrvSck);
    ADDSCKTOSET(netSrvSck);
  }
  ADDSCKTOSET(dbgSrvSck);
  for (i=0; i<nDBGsck; i++) 
  { /* debug pipe requests */
    ADDSCKTOSET(dbgSckTbl[i].sck);
  }
# ifdef UNIX
  /* stdin is a normal fd in the unix world :*/
  if (foregroundTTY) { if (n >= maxlen) return n; fds[n].fd = STDIN_FILENO; fds[n].events = POLLIN; n++; }
# ifndef NO_SERVICE_PIPE
  /* let remote process attach to a server running in the background */
  ADDSCKTOSET(ipcCmdSck);
  for (i=0; i<ipcClnLstSize; i++) ADDSCKTOSET(IPCfds[i]);
# endif /* NO_SERVICE_PIPE */
  if (gServerCycleIdle) return n;
# ifndef NO_TINE_PIPE
  ADDSCKTOSET(ipcSrvSck);
  for (i=0; i<nIPCsck; i++) 
  { /* client on the same machine as the server ? */
    ADDSCKTOSET(ipcSckTbl[i].sck);
  }
# endif /* NO_TINE_PIPE */
# endif /* UNIX */
  if (gServerCycleIdle) return n;
# ifdef IPXBSD
  /* anybody still using IPX ? */
  ADDSCKTOSET(ipxSrvSck);
# endif
  /* allow external socket 'hooks' beyond our control */
  if (ext_pollfds != NULL) for (i=0; i<next_pollfds && n < maxlen; i++) 
  {
    fds[n++] = ext_pollfds[i];
  }
  return n;
}
#ifdef MULTITHREADED
# ifndef DEFAULT_SERVERCYCLE_THREAD 
#   define DEFAULT_SERVERCYCLE_THREAD TRUE
# endif
/* serverCycle() and sockets on separate thread from clientCycle() */
int gRunServerCycleInSeparateThread = DEFAULT_SERVERCYCLE_THREAD;
TINE_EXPORT void SetRunServerCycleInSeparateThread(int value)
{
  gRunServerCycleInSeparateThread = value;
}
TINE_EXPORT int GetRunServerCycleInSeparateThread(void)
{
  return gRunServerCycleInSeparateThread;
}
int gNeedToStartSrvCycleThread = FALSE;
int gNeedToStartClnCycleThread = FALSE;
# ifndef DEFAULT_TRANSPORT_THREAD
#   if defined(VXWORKS) || defined(SMALL_TINE_LIB)
#     define DEFAULT_TRANSPORT_THREAD FALSE
#   else 
#     define DEFAULT_TRANSPORT_THREAD TRUE
#   endif
# endif
/* affects tcp transport only : */
int gRunTransportInSeparateThread = DEFAULT_TRANSPORT_THREAD;
TINE_EXPORT void SetRunTransportInSeparateThread(int value)
{
  gRunTransportInSeparateThread = value;
}
TINE_EXPORT int GetRunTransportInSeparateThread(void)
{
  return gRunTransportInSeparateThread;
}
#endif /* MULTITHREADED */
TINE_EXPORT int PollExternalDescriptor(SOCKET s,int events)
{
  int i;
  if (ext_pollfds == NULL)
  {
    if ((ext_pollfds=(struct pollfd *)SystemCalloc(EXT_POLLFDS_CAPACITY,sizeof(struct pollfd))) == NULL)
      return out_of_local_memory;
    next_pollfds = 0;
  }
  for (i=0; i<next_pollfds; i++) if (ext_pollfds[i].fd == (int)s)
  {
    ext_pollfds[i].events = (short)events;
    return 0;
  }
  if (isServerSocket(s)) return not_accepted;
  ext_pollfds[next_pollfds].events = (short)events;
  next_pollfds++;
  return 0;
}
void AcceptIP(struct timeval *tv)
{
# define MAX_POLLED_FDS 256
  struct pollfd fds[MAX_POLLED_FDS];
  int cc = 0,retry=2,i,ierrsck=-1,p,idx,nfds=0,pto,nready,eno=0,rpterr=FALSE;
  struct timeval *to, ltv;
  static int neterrcnt = 0;
# if defined(UNIX)
  char cmdbuf[CMDSIZE];
# endif

  if (!ipLoaded) return;

  if (!gIsCyclePollingMode)
  { /* fill in the rw select()ion set : */
    ltv = *tv;
    nfds = addServerSocketsToSet(fds,nfds,MAX_POLLED_FDS);
    if (!gRunServerCycleInSeparateThread) nfds = addClientSocketsToSet(fds,nfds,MAX_POLLED_FDS);
    to = getNextCycleTimeout(tv);
    if (nfds == 0)
    { /* no sockets added ! (still initializing ?) */
      millisleep(10);
      return;
    }
#   if defined(WIN32)
    /* Windows Pipes are not select()able, but handle them here anyway */
    chkWinPipe();
#   endif 
retry_select:
    /* anything fall through select() ? */
    if (!gRunServerCycleInSeparateThread) gWaitingOnSelect = TRUE;
    pto = to->tv_sec * 1000 + to->tv_usec/1000;
    nready = poll(fds,nfds,pto);
    if (nready < 0) eno = soerrno;
    if (nready < 0 && eno == EINTR && retry-- > 0)
    {
      to->tv_sec = 0; to->tv_usec = 0;
      goto retry_select;
    }
    if (!gRunServerCycleInSeparateThread) gWaitingOnSelect = FALSE;
    gSelectOnSysPoll = TRUE;
    gettimeofday(tv,(struct timezone *)NULL);
    gTimeFdIdle = MSECS(*tv,ltv);
    /* is there a hook ? */
    if (socketAcceptEntryFcn) (socketAcceptEntryFcn)();
    if (nready < 0) 
    { 
      if (eno == EINTR) return; /* pass on an interrupted i/o */
      rpterr = TRUE;
      neterrcnt++; 
      ccerr(nready); /* serious low-level socket error (errno will be set) */
    }
    if (nready == 0) 
    { /* empty pass : call the scheduler (else trust the consumer logic) */
      gServerCycleNeedsScheduler = TRUE;
      if (tineDebug > 5) dbglog("null request -> call scheduler");
      ccerr(0);
    }
    gSelectOnSysPoll = FALSE;
#   ifdef UNIX
#   ifndef NO_TINE_PIPE
    /* local client ? */
    if (ipcSrvSck > 0 && fds_isset(ipcSrvSck)) AcceptIPC(ipcSrvSck);
    for (i=0; i<nIPCsck; i++) 
    {
      if (ipcSckTbl[i].sck > 0 && fds_isset(ipcSckTbl[i].sck)) getConsumerIPAddress(ipcSckTbl[i].sck,PIPE);
    }
#   endif /* NO_TINE_PIPE */
    /* stdin ? */
    if (foregroundTTY && fds_isset(STDIN_FILENO) && GetCommandEx(STDIN_FILENO,cmdbuf,CMDSIZE) > 0) 
      InterpretCommand(cmdbuf);
#   ifndef NO_SERVICE_PIPE
    /* attached process issuing a command ? */
    chkIpcClnLst(fds,nfds);
#   endif /* NO_SERVICE_PIPE */
#   endif /* UNIX */
    for (i=0; i<nfds; i++) if (pollerr(i)) 
    { 
      ierrsck = i;
      eno = pollerr2errno(fds[i].revents);
      rpterr = eno == ECONNABORTED ? FALSE : TRUE;
      neterrcnt++;
      ccerr(-1); /* low-level socket error, e.g. TCP hangup (errno will NOT be set) */
    }
    neterrcnt = 0; /* reset if we've made it this far (the normal case) */
  }
  /* is someone trying to get a connection ? */
  if (gIsCyclePollingMode || fds_isset(tcpSrvSck))
  {
    p = (gAfInet == AF_INET6) ? TCP6 : TCP;
    if ((i=AssignConnectionSocket(tcpSrvSck,p,&nTCPsck,tcpSckTbl,tcpConTblSize)) >= 0) 
    {
      fds[nfds].fd = tcpSckTbl[i].sck; fds[nfds].events = POLLIN; nfds++;
    }
  }
# ifdef MULTITHREADED
  if (gIsCyclePollingMode || fds_isset(tcpStrmSck))
  {
    p = (gAfInet == AF_INET6) ? STREAM6 : STREAM;
    if ((i=AssignConnectionSocket(tcpStrmSck,p,&nTCPsck,tcpSckTbl,tcpConTblSize)) >= 0) 
    { 
      fds[nfds].fd = tcpSckTbl[i].sck; fds[nfds].events = POLLIN; nfds++;
    }
  }
# endif
  if (gAllowRemoteDebugging)
  {
    if (gIsCyclePollingMode || fds_isset(dbgSrvSck))
    {
      p = (gAfInet == AF_INET6) ? DBGPIPE6 : DBGPIPE;
      if ((i=AssignConnectionSocket(dbgSrvSck,p,&nDBGsck,dbgSckTbl,dbgConTblSize)) >= 0) 
      { 
        fds[nfds].fd = dbgSckTbl[i].sck; fds[nfds].events = POLLIN; nfds++;
      }
    }
  }
  /* any TCP cleanup work necessary ? */
  if (nTCPsck) removeFlaggedBuckets(); 
  /* has something come in on a connected socket ? */
  for (i=0; i<nTCPsck; i++)
  {
    if (gIsCyclePollingMode || fds_isset(tcpSckTbl[i].sck)) 
      getConsumerIPAddress(tcpSckTbl[i].sck,tcpSckTbl[i].type);
  }
  /* has something has come in on a UDP socket ? */
  p = (gAfInet == AF_INET6) ? UDP6 : UDP;
  if (gIsCyclePollingMode || fds_isset(udpSrvSck)) getConsumerIPAddress(udpSrvSck,p);
  /* has something has come in on the service socket ? */
  if (netSrvSck != 0)
  {
    if (gIsCyclePollingMode || fds_isset(netSrvSck)) getConsumerIPAddress(netSrvSck,p);
  }
  for (i=0; i<nDBGsck; i++)
  {
    if (gIsCyclePollingMode || fds_isset(dbgSckTbl[i].sck)) chkDbgClnLst(fds,nfds);
  }
err:
  if (cc < 0) 
  { /* socket errors ! */
    int close_it=FALSE;
    char *errtxt = neterrcnt > 100 ? "abort" : "select";
    SOCKET errsck = ierrsck >= 0 ? fds[ierrsck].fd : (SOCKET)0;
    if (rpterr) rprtSckErr("AcceptIP",errtxt,errsck,eno,1,NULL);
    for (i=0; i<nTCPsck; i++)
    { /* maybe the connection to a daughter socket went down ? */
      close_it = (neterrcnt && fds_isset(tcpSckTbl[i].sck)) ? TRUE :FALSE;
      if (fds_iserr(tcpSckTbl[i].sck))
      {
        neterrcnt = 0;
        close_it = TRUE;
        if (tineDebug || errorDebug) dbglog("closing TCP socket %d due to error",tcpSckTbl[i].sck);
      }
      if (close_it) CloseConnectionSocket(tcpSckTbl[i].sck,tcpSckTbl[i].type,TRUE);
    }
    for (i=0; i<nDBGsck; i++)
    { /* maybe the connection to a daughter socket went down ? */
      close_it = (neterrcnt && fds_isset(dbgSckTbl[i].sck)) ? TRUE :FALSE;
      if (fds_iserr(dbgSckTbl[i].sck)) 
      {
        neterrcnt = 0;
        close_it = TRUE;
        if (rpterr) feclogEx(TINE_LOGLEVEL_WARN, "closing debug session (socket %d)", dbgSckTbl[i].sck);
      }
      if (close_it) CloseConnectionSocket(dbgSckTbl[i].sck,DBGPIPE,FALSE);
    }
    if (!ServerExitCondition && neterrcnt > 100)
    { /* this is serious : abort the process */
      feclogEx(TINE_LOGLEVEL_FATAL, "serious networking problem with the current socket descriptor set");
      feclogEx(TINE_LOGLEVEL_FATAL, "please check your networking environment, including your ethernet card!");
      feclogEx(TINE_LOGLEVEL_FATAL, "nfds = %d, dump of fds array:", nfds);
      for (i=0; i < MAX_POLLED_FDS; i++)
      {
        if (fds[i].fd > 0)
        {
          feclogEx(TINE_LOGLEVEL_FATAL, "[%d] fd = %d, events = %d, revents = %d", i, fds[i].fd, fds[i].events, fds[i].revents);
        }
      }
      if (gDieOnSocketErrors)
      {
        if (dieFunction != NULL) dieFunction("Serious Socket errors! Aborting the server process.");
        if (gDieAnotherDay) { gDieAnotherDay = FALSE; return; }
        dieFunction = NULL; /* avoid multiple calls */
        feclogEx(TINE_LOGLEVEL_FATAL, "aborting the server process ...");
        ExitEQPModules();
        exit(1);
      }
      neterrcnt = 0; /* start counting again ... */
    }
    millisleep(10); /* give the system a breather ... */
  }
  gettimeofday(tv,(struct timezone *)NULL);
  return;
}
void CleanupIP(struct timeval *tv)
{
  UNUSED(tv);
  return;
}
