/*
 * File srvipx.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file srvipx.c
 * \brief IPX routines and definitions
 * \internal
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "prolog.h"
#include "srvipx.h"
#include "globals.h"
#include "dbtools.h"
#include "errors.h"
#ifdef IPXBSD
#include <netipx/ipx.h>
#include "srvip.h"
#endif

extern UINT16 gPortOffset;
int GetSPXStream(WORD ID,ClnHdr *cln);
void CloseSPXStream(WORD ID);
int sendSPXData(WORD ID,int socket);
void addIPXNet(BYTE *net,BYTE *mac,BYTE *node);
void dumpIPXNets(void);

struct IPXNet *IPXNets = NULL;
extern IPXAddress IPXFecAddress;

BYTE spxSrvSckB[2] = {SPX_SRVSCK,SPX_SRVSCK};
const BYTE ipxClnSckB[2] = {BRDCST_SOCKET,BRDCST_SOCKET};
BYTE ipxSrvSckB[2] = {IPX_SRVSCK,IPX_SRVSCK};
const BYTE ipxMcastSckB[2] = {MCAST_SOCKET,MCAST_SOCKET};

int ipxSrvSck = 0;
int ipxGlbSck = 0;
#ifdef IPXNW

IPXECB  *brdcstECB,
        *glbECB[NUM_GLB_ECB],
        *regECB[NUM_REG_ECB],
        *PCECB;                  /* Producer/Consumer IPX Send ECB */

SPXECB  *rcvECB[NUM_RCV_ECB],
        *lsnECB[NUM_LSN_ECB],
        *sndECB,
        *conECB;

WORD ISwap(WORD w) { return (w >> 8) | (w << 8); }

IPXECB *CreateIPXECB(int socket)
{
  IPXECB *ecb;

  if ((ecb=(IPXECB *)SystemCalloc(1,sizeof(IPXECB))) != NULL)
    initIPXECB(ecb,socket);
  return ecb;
}
SPXECB *CreateSPXECB(int socket)
{
  SPXECB *ecb;

  if ((ecb=(SPXECB *)SystemCalloc(1,sizeof(SPXECB))) != NULL)
    initSPXECB(ecb,socket);
  return ecb;
}
WORD PCSocket = 0;
#define nwAddr (ecb->hdr.source)
#define immAddr (ecb->ecb.immediateAddress)
#else
fd_set ipxset;

int ipxCreateSocket(UINT32 network, short port)
{
  int s,optval,optlen = sizeof(int),rc,nlen;
  struct sockaddr_ipx ipxaddr;
  char str[80];

  sprintf(str,"IPX: ERR socket init");
  if ((s=socket(AF_IPX, SOCK_DGRAM, 0)) < 0)
  {
    perror(str);
    feclog(str);
  }
  else
  {
    optval = 1;
    sprintf(str,"IPX: ERR setsockopt");
    if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) 
    {
      perror(str);
      feclog(str);
    }
    ipxaddr.sipx_family = AF_IPX;
    ipxaddr.sipx_network = htonl(network);
    ipxaddr.sipx_port = htons(port);
    ipxaddr.sipx_type = IPX_FRAME_ETHERII;
    sprintf(str,"IPX: ERR bind");
    if ((rc=bind(s, (struct sockaddr *)&ipxaddr,sizeof(ipxaddr))) < 0)
    {
      perror(str);
      feclog(str);
      close(s);
      s = -1;
    }
  }
  if (s >= 0)
  {
    sprintf(str,"IPX: open socket %d",s);
    feclog(str);
  }
  return s;
}
#endif

BYTE IPXbrdcstNode[6] = {0xff,0xff,0xff,0xff,0xff,0xff};

IPXNetListStruct *IPXNetList = NULL;
int NrOfIPXNets = 0;
IPXNetListStruct *IPXBcastList = NULL;
int NrOfIPXBcastNets = 0;

int getIPXControlNets(char *fn,int *nr,IPXNetListStruct **list)
{
  int i,j,n=0,size;
  char ipxnetsfilename[128];
  IPXNetListStruct *lst = NULL;
  FILE *fp = NULL;
  int cc=0,done=0;
  int net_col,nod_col,rou_col;
  char c,str[256],hdr[256],scratch[16],hexstr[64];

  n = 0; lst = *list;
  SystemFree(lst);

  sprintf(ipxnetsfilename,"%s%s",fecDbPath,fn);
  if ((size=GetDataBaseSize(ipxnetsfilename)) <= 0)
  {
    feclog("No registered IPX control nets\n>");
    ccerr(0);
  }
  if ((lst=(IPXNetListStruct *)SystemCalloc(sizeof(IPXNetListStruct),size)) == NULL)
      return out_of_server_memory;
  if ((fp=fopen(ipxnetsfilename,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((net_col=findcol(hdr,"NETWORK")) < 0) ccerr(no_such_column);
      if ((nod_col=findcol(hdr,"NODE")) < 0) ccerr(no_such_column);
      if ((rou_col=findcol(hdr,"ROUTER")) < 0) ccerr(no_such_column);
      done = 1;                              /* read header info */
      continue;
    }
    getColumnStr(net_col,str,scratch,16); 
    strupr(scratch); 
    for (i=0,j=0; i<4 && j<strlen(scratch); i++) 
    {
      if (i<4-strlen(scratch)/2) lst[n].net[i] = 0;
      else
      {
        sprintf(hexstr,"0x%c%c",scratch[j],scratch[j+1]); j += 2;
        lst[n].net[i] = (BYTE)strtol(hexstr,NULL,16);
      }
    }
    getColumnStr(nod_col,str,scratch,16); 
    strupr(scratch); 
    for (i=0,j=0; i<6 && j<strlen(scratch); i++) 
    {
      if (i<6-strlen(scratch)/2) lst[n].mac[i] = 0;
      else
      {
        sprintf(hexstr,"0x%c%c",scratch[j],scratch[j+1]); j += 2;
        lst[n].mac[i] = (BYTE)strtol(hexstr,NULL,16);
      }
    }
    getColumnStr(rou_col,str,scratch,16); 
    strupr(scratch); 
    for (i=0,j=0; i<6 && j<strlen(scratch); i++) 
    {
      if (i<6-strlen(scratch)/2) lst[n].node[i] = 0;
      else
      {
        sprintf(hexstr,"0x%c%c",scratch[j],scratch[j+1]); j += 2;
        lst[n].node[i] = (BYTE)strtol(hexstr,NULL,16);
      }
    }
    n++;
  }
err:
  if (cc) feclog("%s : %s",fn,cc2str(cc));
  *nr = n; *list = lst;
  return cc;
}
int setupIPX(void)
{
# ifdef IPXNW
  BYTE major, minor;
  WORD maxCon, availCon;

  if (IPXInitialize()) return -1;
  if (!SPXInitialize(&major,&minor,&maxCon,&availCon)) return -1;
# endif
  return 0;
}
static ClnHdr IncomingSPXClient;
int initIPX(void)
{
  WORD socket;
  int i, cc = 0;

  if (setupIPX()) ccerr(ipx_not_supported);

  memset(&IncomingSPXClient,0,sizeof(ClnHdr));

  ipxSrvSckB[0] = IPX_SRVSCK + gPortOffset;
  ipxSrvSckB[1] = IPX_SRVSCK + gPortOffset;
  spxSrvSckB[0] = SPX_SRVSCK + gPortOffset;
  spxSrvSckB[1] = SPX_SRVSCK + gPortOffset;

# ifdef IPXNW
  IPXGetInternetworkAddress((BYTE *)&IPXFecAddress);

  /* Support consumer subscriptions:
     Producer/Consumer send ECB & Consumer Receive ECB
     (get immediated address from caller)
  */
  PCSocket = 0;
  if (IPXOpenSocket((BYTE *)&PCSocket,0)) ccerr(ipx_socket_error);
  if ((PCECB=CreateIPXECB(PCSocket)) == NULL) ccerr(out_of_server_memory);

  socket = *(int *)&ipxSrvSckB[0];
  if (IPXOpenSocket((BYTE *)&socket,0)) ccerr(ipx_socket_error);
  for (i=0; i<NUM_REG_ECB; i++)  /*/ create subscription registration ECBs */
  {
    if ((regECB[i]=CreateIPXECB(socket)) == NULL) ccerr(out_of_server_memory);
    IPXListenForPacket(&regECB[i]->ecb);
  }
# else  /* not IPXNW but IPXBSD */
  ipxSrvSck = ipxCreateSocket(0L,*(short *)ipxSrvSckB);
  if (ipxSrvSck < 0) ipxSrvSck = 0;
# endif

  /* look for broadcasting information: */
  socket = 0;
  getIPXControlNets(IPXNETS_FILE,&NrOfIPXNets,&IPXNetList);
  getIPXControlNets(IPXBCASTNETS_FILE,&NrOfIPXBcastNets,&IPXBcastList);
  if (canSendGlobals)
  {
    socket = 0;
#   ifdef IPXNW
    if (IPXOpenSocket((BYTE *)&socket,0)) ccerr(ipx_socket_error);
    if ((brdcstECB=CreateIPXECB(socket)) == NULL) ccerr(out_of_server_memory);
    memset(brdcstECB->hdr.destination.node,0xff,6);      /*/ IPX broadcast */
    memset(brdcstECB->ecb.immediateAddress,0xff,6);
    memcpy(brdcstECB->hdr.destination.network,IPXFecAddress.network,4);
    *(WORD *)brdcstECB->hdr.destination.socket = *(WORD *)&ipxClnSckB[0];
#   else  /* not IPXNW but IPXBSD */
    ipxGlbSck = ipxCreateSocket(0L,socket);
    if (ipxGlbSck < 0) ipxGlbSck = 0;
#   endif
    for (i=0; i<NrOfIPXBcastNets; i++)  /* add them all to broadcast list */
      addIPXNet(IPXBcastList[i].net,IPXBcastList[i].mac,IPXBcastList[i].node);
    if (StartupDebug) dumpIPXNets();
  }
  else if (StartupDebug) printf("No IPX broadcasting from this FEC");
err:
  if (cc)
  {
    feclog("IPX failed to initialize : %s",cc2str(cc));
  }
  else
  {
    feclog("IPX initialized");
    ipxLoaded = TRUE;
  }
  return cc;
}
void closeIPXServer(void)
{
  if (!ipxLoaded) return;
# ifdef IPXNW
  IPXCloseSocket(lsnECB[0]->ecb.socketNumber);
  IPXCloseSocket(sndECB->ecb.socketNumber);
  IPXCloseSocket(PCSocket);
# else
  if (ipxSrvSck) close(ipxSrvSck);
  if (ipxGlbSck) close(ipxGlbSck);
# endif
  return;
}
void AcceptIPX(struct timeval *tv)
{
# ifndef IPXNW
  struct pollfd fds[256];
  int nfds=0, idx, n, maxlen=256;
  int retry = 2,nready;
# endif
  if (!ipxLoaded) return;

# ifdef IPXNW
  while (getConsumerIPXAddress() != NULL);
# else
  if (ipxSrvSck == 0) return;
  ADDSCKTOSET(ipxSrvSck);

retry_select:
  if ((nready=poll(fds,nfds,0)) == 0) goto err;
  if (nready < 0)
  {
    rprtSckErr("AcceptIPX","select",ipxSrvSck,2,NULL);
    if (retry-- > 0) goto retry_select;
    goto err;
  }
  getConsumerIPXAddress();
# endif
err:
  gettimeofday(tv, (struct timezone *) NULL);
  return;
}

ClnHdr* getConsumerIPXAddress(void)
{
  int i, cc = 0;
  ClnHdr *cln = NULL;
# ifdef IPXNW
  IPXECB *ecb;
  static int ecbcnt = -1;
# else
  int timeout = 1;
  int index;
  int bytes;
  IpxAdr ipxadr;
  struct sockaddr_ipx sckadr;
  int len = sizeof(sckadr);
# endif

# ifdef IPXNW
  ecbcnt = (++ecbcnt)%NUM_REG_ECB;
  for (i=0; i<NUM_REG_ECB; i++)  /* Cycle through listening ECBs*/
    if (!(ecb=regECB[(i+ecbcnt)%NUM_REG_ECB])->ecb.inUseFlag) break;

  if ((ecbcnt=i) == NUM_REG_ECB) return (ClnHdr *)NULL;
  if (ecb->ecb.completionCode) ccerr(ipx_socket_error);
  memcpy(srvWorkArea,ecb->buf,IPX_BUFFER_SIZE);
# else
  if (ipxSrvSck == 0) return (ClnHdr *)NULL;
  if ((i=recvfrom(ipxSrvSck,(char *)srvWorkArea,srvWorkAreaSize,0,
          (struct sockaddr *)&sckadr,&len)) < 0) ccerr(ipx_socket_error);
  if (i != PeekWord(*(UINT16 *)srvWorkArea)) ccerr(net_read_error);

  sckadr.sipx_family = AF_IPX;
  memcpy(ipxadr.IpxNode,sckadr.sipx_node,6);
  memcpy(ipxadr.IpxNet,&sckadr.sipx_network,4);
  cln->ep.ipx.sckRmt = ntohs(sckadr.sipx_port);
# endif
  IPXcount++;
  if (tineDebug)
  {
    char addr[64];
    sprintf(addr,ipxnet_ntoa(ipxadr.IpxNode));
    sprintf(dbgbuf,"IPX: [%s] ",addr);
  }
  if ((cln=LocateConsumerInList(NULL,&ipxadr,NULL)) == NULL)
    ccerr(out_of_server_memory);
  cln->ep.ipx.sckLcl = ipxSrvSck;
  cln->inetProtocol = IPX;

err: 
  if (cc && (tineDebug || errorDebug)) dbglog("ERR: %s",cc2str(cc));
# ifdef IPXNW
  ecb->ecb.socketNumber = *(int *)&ipxSrvSckB[0];
  IPXListenForPacket(&ecb->ecb);
# else
  FD_SET(ipxSrvSck,&ipxset);
# endif
  if (cln != NULL) getConsumerRequest(cln,srvWorkArea,IPX_BUFFER_SIZE);
  return cln;
}

int sendIPXData(ClnHdr *cln, BYTE *data)
{
  short cc=0;
  unsigned short msgsize = PeekWord(*(UINT16 *)data);
# ifndef IPXNW
  struct pollfd fds[1];
  int retry = 2,IPXSocket,bytesSent;
  static struct sockaddr_ipx adrCln;
# endif

# ifdef IPXNW
  if (msgsize > IPX_BUFFER_SIZE) return use_stream_transport;
  memcpy(&PCECB->hdr.destination.network,&cln->ipx,12);
  memcpy(&PCECB->ecb.immediateAddress,cln->ipx.IpxImmAdr,6);
  memcpy(PCECB->buf,data,msgsize);
  PCECB->ecb.fragmentDescriptor[1].size = msgsize;
  IPXSendPacket(&PCECB->ecb);
  while (PCECB->ecb.inUseFlag && !PCECB->ecb.completionCode);
  if ((cc=PCECB->ecb.completionCode) != 0)
  {
    if (tineDebug || errorDebug) printf("IPX: Failed packet delivery (error %x)",cc);
    return ipx_socket_error;
  }
# else
  memcpy(adrCln.sipx_node,cln->ipx.IpxNode,6);
  memcpy(&adrCln.sipx_network,cln->ipx.IpxNet,4);
  adrCln.sipx_port = htons(*(UINT16 *)cln->ep.ipx.sckRmt);
  adrCln.sipx_family = AF_IPX;
  IPXSocket = cln->sck;
  if (msgsize > IPX_BUFFER_SIZE) return invalid_transport_size;
  fds[0].fd = IPXSocket; fds[0].events = POLLOUT;
retry_select:
  switch (poll(fds,1,0))
  {
    case 1: 
      break;
    case 0: 
      if (tineDebug > 1 || errorDebug) dbglog("IPX: socket %d: write set not ready",IPXSocket);
	    return 0;
    default:
      rprtSckErr("sendIPXData","select",IPXSocket,2,NULL);
      if (retry-- > 0) goto retry_select;
	    return 0;
  }
  if (WaitForSystemMutex(hSndMutex,100) != 0) return semaphore_busy;
  if (sendto(IPXSocket,(char *)data,msgsize,0,
             (struct sockaddr *)&adrCln,sizeof(adrCln)) < 0)
  {
    rprtSckErr ("sendIPXData","sendto",IPXSocket,0,NULL);
    ccerr(ipx_socket_error);
  }
# endif
  if (tineDebug > 1)
  {
    char addr[64];
    strcpy(addr,ipxnet_ntoa(cln->ipx.IpxNode));
    dbglog("IPX: %d bytes to %s [%s]",msgsize,cln->usr,addr);
  }
err:
  ReleaseSystemMutex(hSndMutex);
  return cc;
}
#pragma argsused
void CleanupIPX(struct timeval *tv)
{
# ifdef IPXNW
  int i;
  for (i=0; i<NUM_RCV_ECB; i++)
  {
    if (rcvECB[i]->hdr.dataStreamType == 0xfe) clearStrandedPackets(rcvECB[i]);
  }
# endif
  return;
}
void addIPXNet(BYTE *net,BYTE *mac,BYTE *node)
{
  struct IPXNet *n;
  if ((n=SystemCalloc(1,sizeof(struct IPXNet))) == NULL) return;
  memcpy(n->network,net,4);
  memcpy(n->node,mac,6);
  memcpy(n->ImmediateAddress,node,6);
  n->nextNet = IPXNets;
  IPXNets = n;
  return;
}
void dumpIPXNets(void)
{
  int i,n;
  char s[40];

  dbglog("Registered IPX Subnets (sending globals %s):\n",
          canSendGlobals ? "enabled" : "disabled" );
  for (n=0; n<NrOfIPXNets; n++)
  {
    for (i=0;i<4;i++) sprintf(&s[i*2],"%02X",IPXNetList[n].net[i]);
    sprintf(dbgbuf,"\t[%s] ",s);
    for (i=0;i<6;i++) sprintf(&s[i*2],"%02X",IPXNetList[n].mac[i]);
    sprintf(&dbgbuf[strlen(dbgbuf)],"[%s]",s);
    dbglog(dbgbuf);
  }
}
int IsMemberIPXNets(int num,IPXNetListStruct *lst,IpxAdr *ipxaddr)
{
  int n;

  if (num == 0) return TRUE;  /* everything's open */
  if (ipxaddr == NULL) return FALSE;
  for (n=0; n<num; n++)
  {
    if (memcmp(lst[n].net,ipxaddr->IpxNet,4)) continue;
    if (!memcmp(lst[n].mac,IPXbrdcstNode,6)) return TRUE;
    if (!memcmp(lst[n].mac,ipxaddr->node,6)) return TRUE;
  }
  return FALSE;
}
int sendIPXgcast(BYTE *buf,int size)
{
  int i;
  struct IPXNet *n;
# ifdef IPXNW
  memcpy(brdcstECB->buf,buf,size);
  for (i=0,n=IPXNets; i<NrOfIPXBcastNets && n != NULL; i++,n=n->nextNet)
  {
    if (tineDebug > 1 && n == IPXNets)
      dbglog(dbgbuf,"IPX: broadcast on %d subnets",NrOfIPXBcastNets);
    memcpy(brdcstECB->hdr.destination.node,n->node,6);
    memcpy(brdcstECB->ecb.immediateAddress,n->ImmediateAddress,6);
    memcpy(brdcstECB->hdr.destination.network,n->network,4);

    IPXSendPacket(&brdcstECB->ecb);
    while (brdcstECB->ecb.inUseFlag);
    if (brdcstECB->ecb.completionCode && tineDebug > 1)
      dbglog("broadcast failed: %x",brdcstECB->ecb.completionCode);
  }
# else
  struct pollfd fds[1];
  int cc, IPXSocket,bytesSent;
  struct sockaddr_ipx adrCln;

  if (WaitForSystemMutex(hSndMutex,100) != 0) return semaphore_busy;
  for (i=0,n=IPXNets; i<NrOfIPXBcastNets && n != NULL; i++,n=n->nextNet)
  {
    if (tineDebug > 1 && n == IPXNets)
      dbglog("IPX: broadcast on %d subnets",NrOfIPXBcastNets);
    memcpy(adrCln.sipx_node,n->node,6);
    memcpy(&adrCln.sipx_network,n->network,4);
    adrCln.sipx_port = htons(*(UINT16 *)ipxClnSckB);
    adrCln.sipx_family = AF_IPX;
    IPXSocket = ipxGlbSck;
    if (size > IPX_BUFFER_SIZE) ccerr(use_stream_transport);
    fds[0].fd = IPXSocket; fds[0].events = POLLOUT;
    switch (poll(fds,1,0))
    {
      case 1: 
        break;
      case 0: 
        if (tineDebug > 1 || errorDebug) dbglog("IPX: socket %d: write set not ready",IPXSocket);
	      ccerr(ipx_socket_error);
      default:
        rprtSckErr("sendIPXgcast","select",IPXSocket,2,NULL);
	      ccerr(ipx_socket_error);
    }
    if (sendto(IPXSocket,(char *)buf,size,0,
               (struct sockaddr *)&adrCln,sizeof(adrCln)) < 0)
    {
      rprtSckErr ("sendIPXgcast","sendto",IPXSocket,1,NULL);
      ccerr(ipx_socket_error);
    }
  }
err:
  ReleaseSystemMutex(hSndMutex);
# endif
  return 0;
}

/* trivial RPC call: */
int tIPXCall(int fecindex,BYTE *buf,int timout)
{
# ifdef IPXNW
  static ClnHdr cln;
  static BYTE ipxSck[2];
  int i,cc;
  time_t t0 = time(NULL);
  FecAddrStruct *rpcfec = fecindex ? &FecTbl[fecindex] : &NameServer[curNameServer];

  ipxSck[0] = IPX_SRVSCK + rpcfec->portOffset;
  ipxSck[1] = IPX_SRVSCK + rpcfec->portOffset;

  memset(&cln,0,sizeof(ClnHdr));
  strncpy(cln.usr,rpcfec->fecName,FEC_NAME_SIZE);
  memcpy(&cln.ipx,&rpcfec->netAdr.ipx,16);
  memcpy(cln.ep.ipx.sckRmt,ipxSck,2);

  if ((cc=sendIPXData(&cln,buf)) != 0) return cc;

  timout /= 1000;
  if (timout < 1) timout = 1;
  for (i=0; time(NULL) <= t0 + timout; i = (++i)%NUM_CNS_ECB)
    if (!cnsECB[i]->ecb.inUseFlag && !cnsECB[i]->ecb.completionCode)
    {
      if (((PrdrHdr *)&cnsECB[i]->buf[2])->subId == 1)
      {
        memcpy(buf,cnsECB[i]->buf,IPX_BUFFER_SIZE);
        IPXListenForPacket(&cnsECB[i]->ecb);
        return 0;
      }
    }
  return link_not_open;
# else
  return not_implemented;
# endif
}
