/*
 * File srvwizard.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include "tine.h"

/*

  C Server Wizard :
  
  does device file exist ? -> read device information
  links.csv -> start any client links, globals (HW links ?)
  BITFIELDS
  tag generation areas
  1st pass : as is
  subsequent : read modules and append sections
*/
char targetPath[256];
int putProcessFunction(FILE *fp,char *procname, int nsteps);
char *strudr(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c == '.' || *c == '-') *c = '_';
  return s;
}
#ifdef WIN32
unsigned long PASCAL FAR inet_addr (const char FAR * cp) { return 0; }
#endif
char tagNameFilter[64];
char *gettype(int fmt, char *fstr);
#define EXPORT_FILE_BAK "exports.bak"
#define MAKEFILE_NAME "fec.mak"

typedef struct ExportPropertyListEx
{
  char  prpName[PROPERTY_NAME_SIZE];       /* property name */
  int  prpId;              /* property id */
  UINT32 prpSize;          /* maximum supported array size */
  UINT32 prpSizeIn;        /* maximum supported input array size */
  short prpFormat;         /* native format */
  short prpFormatIn;       /* native input format */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* property array type if array */
  short numRows;           /* num rows if array */
  short rowSize;           /* row size if array */
  char  prpDescription[PROPERTY_DESC_SIZE];/* short description of property */
  PrpEgu egu;   /* EGU information for display */
  PrpEgu xegu;  /* x-axis EGU information for spectrum display */
  short numProcessSteps; 
  struct ExportPropertyListEx *next;
} ExportPropertyListStructEx;

typedef struct EQPstruct
{
  char tagname[EXPORT_NAME_SIZE];
  char lclname[EQM_NAME_SIZE];
  int numdevs;
  int numprps;
  ExportPropertyListStructEx *prp;
  struct EQPstruct *next;
} EQP;

EQP *eqplist = NULL;

void ParsePropertyExDescription(char *str,ExportPropertyListStructEx *prp)
{
  char *c=NULL, *c1, tmp[128], *s = str;

  if (str == NULL) return;
  if ((c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)               /* 1st token: egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->egu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)               /* prp egu given */
      {
        c1++;
        prp->egu.min = (float)strtod(c,&c); c = c1;
        prp->egu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->egu.units,tmp,16);
        }
      }
    }
  }
  if (str != NULL && (c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)             /* 2nd token: x egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->xegu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)                 /* x egu given */
      {
        c1++;
        prp->xegu.min = (float)strtod(c,&c); c = c1;
        prp->xegu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->xegu.units,tmp,16);
        }
      }
    }
  }
  if (str == NULL) str = s;
  strncpy(prp->prpDescription,str,PROPERTY_DESC_SIZE);  /* 3rd token: description part */
}

int putExportInformation(void)
{
  int cc = 0;
  FILE *fp=NULL;
  ExportPropertyListStructEx *prp;
  EQP *el;
  char desc[128],mode[128];
  char type[64],cftype[64];
  char fullFileName[256],fullBackupName[256];

  if (strlen(targetPath) > 0) 
  {
    sprintf(fullFileName,"%.128s%s",targetPath,EXPORT_FILE);	 
    sprintf(fullBackupName,"%.128s%s",targetPath,EXPORT_FILE_BAK);	 
  }
  else
  {
    sprintf(fullFileName,"%s",EXPORT_FILE);	 
    sprintf(fullBackupName,"%s",EXPORT_FILE_BAK);	 
  }
  cc = rename(fullFileName,fullBackupName);  /* back it up ! */
  if (cc != 0)
  {
    perror("rename");
    cc = 0;
  }

  if ((fp=fopen(fullFileName,"w")) == NULL) ccerr(no_such_file);
  fprintf(fp,"EXPORT_NAME,LOCAL_NAME,NUM_MODULES,PROPERTY,PROPERTY_SIZE,PROPERTY_ID,FORMAT,ACCESS,DESCRIPTION\n");
  for (el=eqplist; el != NULL; el=el->next)  // go through the list
  {
    for (prp=el->prp; prp != NULL; prp = prp->next)
    {
      strcpy(mode,"CA_READ");
      if (prp->prpAccessMode&CA_WRITE) strcat(mode,"|CA_WRITE");
        strcpy(type,gettype(prp->prpFormat,cftype));
      if (prp->prpFormat == CF_STRUCT) strcpy(type,"struct");
      fprintf(fp,"%s,%s,%d,",el->tagname,el->lclname,el->numdevs);
      fprintf(fp,"%s,%d,%d,%s,%s,",prp->prpName,prp->prpSize,prp->prpId,type,mode);
      if (prp->egu.min == 0 && prp->egu.max == 0 && strlen(prp->egu.units) == 0)
	      sprintf(desc,"%s",prp->prpDescription);
      else
	      sprintf(desc,"[%g:%g %s]%s",prp->egu.min,prp->egu.max,prp->egu.units,prp->prpDescription);
      fprintf(fp,"%s\n",desc);
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
char fecname[17] = "";
RPCFecExStruct RPCFecEx;
UINT16 PortOffset = 0;

void getBestGuessFecInfo(void)
{
  char *c,tmp[64];
  if (((c=getenv("HOSTNAME")) == NULL) &&
      ((c=getenv("HOST")) == NULL) &&
      ((c=getenv("COMPUTERNAME")) == NULL))
  {
	  c = "TESTSERVER";
  }
  strncpy(tmp,c,12); tmp[12] = 0;
  sprintf(&tmp[strlen(tmp)],".%d",PortOffset);
  strncpy(fecname,tmp,FEC_NAME_SIZE);
  if (((c=getenv("USER")) == NULL) &&
      ((c=getenv("USERNAME")) == NULL) &&
      ((c=getenv("LOGINNAME")) == NULL))
  {
	  c = "ME";
  }
  strcpy(RPCFecEx.os,OS_TAG);
  sprintf(RPCFecEx.desc,"%s's Test FEC",c);
  strcpy(RPCFecEx.hdw,"none");
  sprintf(RPCFecEx.loc,"%s's office",c);
  strncpy(RPCFecEx.resp,c,FEC_RESP_SIZE);
}
int getFecInformation(void)
{
  FILE *fp = NULL;
  int cc=0,done=0;
  int name_col,desc_col,loc_col,hdw_col,resp_col,poff_col,con_col,sub_col;
  char str[256],hdr[256],fecfilename[256],scratch[64];

  if (strlen(targetPath) > 0) 
  {
	 sprintf(fecfilename,"%.128s%s",targetPath,FECID_FILE);	 
  }
  else
  {
	 sprintf(fecfilename,"%s",FECID_FILE);	 
  }
  if ((fp=fopen(fecfilename,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((name_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      poff_col = findcol(hdr,"PORT_OFFSET");
      con_col = findcol(hdr,"CONTEXT");
      sub_col = findcol(hdr,"SUBSYSTEM");
      if ((desc_col=findcol(hdr,"DESCRIPTION")) < 0) ccerr(no_such_column);
      if ((loc_col=findcol(hdr,"LOCATION")) < 0) ccerr(no_such_column);
      if ((hdw_col=findcol(hdr,"HARDWARE")) < 0) ccerr(no_such_column);
      if ((resp_col=findcol(hdr,"RESPONSIBLE")) < 0) ccerr(no_such_column);
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(fecname,colptr(name_col,str),FEC_NAME_SIZE);
    strncpy(RPCFecEx.os,OS_TAG,12);
    if (poff_col >= 0)
    {
      strncpy(scratch,colptr(poff_col,str),16);
      PortOffset = atoi(scratch);
    }
    if (con_col >= 0)
    {
      strncpy(gDeviceContext,colptr(con_col,str),CONTEXT_NAME_SIZE);
      strupr(gDeviceContext);
    }
    if (sub_col >= 0)
    {
      strncpy(gDeviceSubSystem,colptr(sub_col,str),SUBSYSTEM_NAME_SIZE);
      strupr(gDeviceSubSystem);
    }
    strncpy(RPCFecEx.desc,colptr(desc_col,str),FEC_DESC_SIZE);
    strncpy(RPCFecEx.loc,colptr(loc_col,str),FEC_LOCATION_SIZE);
    strncpy(RPCFecEx.hdw,colptr(hdw_col,str),FEC_HDW_SIZE);
    strncpy(RPCFecEx.resp,colptr(resp_col,str),FEC_RESP_SIZE);
    break;
  }
err:
  if (fp) fclose(fp);
  if (cc)
  {
	getBestGuessFecInfo();
  }
  return cc;
}

int dumpExportsFiles = 0;
int getExportInformation(void)
{
  FILE *fp;
  int cc=0,done=0;
  int export_name_col,local_name_col,property_col,property_size_col,
      num_modules_col,format_col,access_col,description_col,
      property_id_col,insize_col,informat_col,nsteps_col;
  char str[256],hdr[256],scratch[64],fullFileName[256];
  ExportPropertyListStructEx *prp, *p;
  EQP *el;
  long prpId;

  if (strlen(targetPath) > 0) 
  {
    sprintf(fullFileName,"%.128s%s",targetPath,EXPORT_FILE);	 
  }
  else
  {
    sprintf(fullFileName,"%s",EXPORT_FILE);	 
  }
  if ((fp=fopen(fullFileName,"r")) == NULL) ccerr(no_such_file);

  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file */
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((export_name_col=findcol(hdr,"EXPORT_NAME")) < 0) ccerr(no_such_column);
      if ((local_name_col=findcol(hdr,"LOCAL_NAME")) < 0) ccerr(no_such_column);
      if ((num_modules_col=findcol(hdr,"NUM_MODULES")) < 0) ccerr(no_such_column);
      if ((property_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_column);
      if ((property_size_col=findcol(hdr,"PROPERTY_SIZE")) < 0) ccerr(no_such_column);
      if ((property_id_col=findcol(hdr,"PROPERTY_ID")) < 0) prpId = -1;
      if ((format_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_column);
      if ((access_col=findcol(hdr,"ACCESS")) < 0) ccerr(no_such_column);
      if ((description_col=findcol(hdr,"DESCRIPTION")) < 0) ccerr(no_such_column);
      if ((informat_col=findcol(hdr,"INFORMAT")) < 0) informat_col = -1;
      if ((insize_col=findcol(hdr,"PROPERTY_INSIZE")) < 0) insize_col = -1;
      if ((nsteps_col=findcol(hdr,"NUM_STEPS")) < 0) nsteps_col = -1;
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(scratch,colptr(export_name_col,str),EXPORT_NAME_SIZE);
    for (el=eqplist; el != NULL; el=el->next) if (!strncmp(scratch,el->tagname,EXPORT_NAME_SIZE)) break;
    if (el == NULL)               /* new eqp type */
    {
      el = (EQP *)calloc(1,sizeof(EQP));
      if (el == NULL) ccerr(out_of_server_memory);
      strncpy(el->tagname,scratch,EXPORT_NAME_SIZE);
      strncpy(el->lclname,colptr(local_name_col,str),EQM_NAME_SIZE);
      strncpy(scratch,colptr(num_modules_col,str),16);
      el->numdevs = atoi(scratch);
      if (el->numdevs == 0) el->numdevs = 1;
      el->numprps = 0;
      el->next = eqplist;
      eqplist = el;
    }	
    if ((prp=(ExportPropertyListStructEx *)calloc(1,sizeof(ExportPropertyListStructEx))) == NULL) ccerr(out_of_server_memory);
    strncpy(prp->prpName,colptr(property_col,str),PROPERTY_NAME_SIZE);
    strncpy(scratch,colptr(property_size_col,str),16);
    prp->prpSize = atoi(scratch);
    if (insize_col != -1)
    {
      strncpy(scratch,colptr(insize_col,str),16);
      prp->prpSizeIn = atoi(scratch);
    }
    else prp->prpSizeIn = prp->prpSize;
    strncpy(scratch,colptr(format_col,str),16);
    prp->prpFormat = ftoi(scratch);
    prp->prpArrayType = getArrayType(scratch);
    if (informat_col != -1)
    {
      strncpy(scratch,colptr(informat_col,str),16);
      prp->prpFormatIn = ftoi(scratch);
    }
    else prp->prpFormatIn = prp->prpFormat;
    if (prp->prpFormatIn == CF_NULL) prp->prpSizeIn = 0;
    if (prp->prpSizeIn > 0 && prp->prpSize == 0)
    { /* make sure the variables get created ! */
      prp->prpSize = prp->prpSizeIn;
      prp->prpFormat = prp->prpFormatIn;
    }
    strncpy(scratch,colptr(access_col,str),16);
    prp->prpAccessMode = mtoi(scratch);
    ParsePropertyExDescription(colptr(description_col,str),prp);
    if (property_id_col >= 0)
    {
      strncpy(scratch,colptr(property_id_col,str),16);
      prp->prpId = atoi(scratch);
    }
    else 
    {
      prp->prpId = el->numprps + 1;
      dumpExportsFiles = -1;
    }
    if (nsteps_col >= 0)
    {
      strncpy(scratch,colptr(nsteps_col,str),16);
      prp->numProcessSteps = atoi(scratch);
    }
    else 
    {
  	  prp->numProcessSteps = 0;
    }
    for (p=el->prp; p != NULL; p=p->next)
    {
      if (p->prpId == prp->prpId)
      {
        printf("WARNING: Duplicate property id (%d) for %s and %s\n",
		         prp->prpId,p->prpName,prp->prpName);
        printf("         The resulting code will not compile !\n");
      }
    }
    if (prp->numProcessSteps > 0)
    {
      if (prp->prpFormat != CF_SHORT && prp->prpFormat != CF_LONG)
      {
        prp->prpFormat = CF_SHORT;
      }
    }
    prp->next = el->prp;
    el->prp = prp;
    el->numprps++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
char *getatype(int atyp, char *astr)
{
  static char str[64],atstr[64];
  switch (atyp)
  {
    default:
    case AT_UNKNOWN: strcpy(str,""); strcpy(atstr,"AT_UNKNOWN"); break;
    case AT_SCALAR: strcpy(str,"SCALAR"); strcpy(atstr,"AT_SCALAR"); break;
    case AT_DOUBLE: strcpy(str,"DOUBLE"); strcpy(atstr,"AT_DOUBLE"); break;
    case AT_CHANNEL: strcpy(str,"CHANNEL"); strcpy(atstr,"AT_CHANNEL"); break;
    case AT_SPECTRUM: strcpy(str,"TRACE"); strcpy(atstr,"AT_SPECTRUM"); break;
    case AT_COLLECTION: strcpy(str,"COLLECTION"); strcpy(atstr,"AT_COLLECTION"); break;
    case AT_IMAGE: strcpy(str,"IMAGE"); strcpy(atstr,"AT_IMAGE"); break;
    case AT_JPEG: strcpy(str,"JPEG"); strcpy(atstr,"AT_JPEG"); break;
    case AT_MPEG: strcpy(str,"MPEG"); strcpy(atstr,"AT_MPEG"); break;
    case AT_BMP: strcpy(str,"BMP"); strcpy(atstr,"AT_BMP"); break;
    case AT_GIF: strcpy(str,"GIF"); strcpy(atstr,"AT_GIF"); break;
    case AT_WAV: strcpy(str,"WAV"); strcpy(atstr,"AT_WAV"); break;
    case AT_MP3: strcpy(str,"MP3"); strcpy(atstr,"AT_MP3"); break;
  }
  if (astr != NULL) strcpy(astr,atstr);
  return str;
}
char *gettype(int fmt, char *fstr)
{
  static char str[64], fmtstr[64];
  switch(fmt)
  {
    default:
    case CF_STRUCT: strcpy(str,"BYTE"); strcpy(fmtstr,"CF_STRUCT"); break;
    case CF_BYTE: strcpy(str,"BYTE"); strcpy(fmtstr,"CF_BYTE"); break;
    case CF_SHORT: strcpy(str,"short"); strcpy(fmtstr,"CF_SHORT"); break;
    case CF_LONG: strcpy(str,"UINT32"); strcpy(fmtstr,"CF_LONG"); break;
    case CF_FLOAT: strcpy(str,"float"); strcpy(fmtstr,"CF_FLOAT"); break;
    case CF_DOUBLE: strcpy(str,"double"); strcpy(fmtstr,"CF_DOUBLE"); break;
    case CF_TEXT: strcpy(str,"char"); strcpy(fmtstr,"CF_TEXT"); break;
    case CF_NAME8: strcpy(str,"NAME8"); strcpy(fmtstr,"CF_NAME8"); break;
    case CF_NAME8I: strcpy(str,"NAME8I"); strcpy(fmtstr,"CF_NAME8I"); break;
    case CF_NAME16: strcpy(str,"NAME16"); strcpy(fmtstr,"CF_NAME16"); break;
    case CF_NAME16I: strcpy(str,"NAME16I"); strcpy(fmtstr,"CF_NAME16I"); break;
    case CF_NAME32: strcpy(str,"NAME32"); strcpy(fmtstr,"CF_NAME32"); break;
    case CF_NAME32I: strcpy(str,"NAME32I"); strcpy(fmtstr,"CF_NAME32I"); break;
    case CF_NAME48: strcpy(str,"NAME48"); strcpy(fmtstr,"CF_NAME48"); break;
    case CF_NAME48I: strcpy(str,"NAME48I"); strcpy(fmtstr,"CF_NAME48I"); break;
    case CF_NAME64: strcpy(str,"NAME64"); strcpy(fmtstr,"CF_NAME64"); break;
    case CF_NAME64I: strcpy(str,"NAME64I"); strcpy(fmtstr,"CF_NAME64I"); break;
    case CF_NAME16FI: strcpy(str,"NAME16FI"); strcpy(fmtstr,"CF_NAME16FI"); break;
    case CF_NAME16II: strcpy(str,"NAME16II"); strcpy(fmtstr,"CF_NAME16II"); break;
    case CF_NAME16DBLDBL: strcpy(str,"NAME16DBLDBL"); strcpy(fmtstr,"CF_NAME16DBLDBL"); break;
    case CF_NAME32DBLDBL: strcpy(str,"NAME32DBLDBL"); strcpy(fmtstr,"CF_NAME32DBLDBL"); break;
    case CF_NAME64DBLDBL: strcpy(str,"NAME64DBLDBL"); strcpy(fmtstr,"CF_NAME64DBLDBL"); break;
    case CF_FLTINT: strcpy(str,"FLTINT"); strcpy(fmtstr,"CF_FLTINT"); break;
    case CF_LNGINT: strcpy(str,"INTINT"); strcpy(fmtstr,"CF_LNGINT"); break;
    case CF_DBLINT: strcpy(str,"DBLINT"); strcpy(fmtstr,"CF_DBLINT"); break;
    case CF_SHTINT: strcpy(str,"SHTINT"); strcpy(fmtstr,"CF_SHTINT"); break;
    case CF_TDS: strcpy(str,"INTFLTINT"); strcpy(fmtstr,"CF_TDS"); break;
    case CF_USTRING: strcpy(str,"DUSTRING"); strcpy(fmtstr,"CF_USTRING"); break;
    case CF_XYS: strcpy(str,"FLTFLTINT"); strcpy(fmtstr,"CF_XYS"); break;
    case CF_FLTINTINT: strcpy(str,"FLTINTINT"); strcpy(fmtstr,"CF_FLTINTINT"); break;
    case CF_INTINTINT: strcpy(str,"INTINTINT"); strcpy(fmtstr,"CF_INTINTINT"); break;
    case CF_FILTER: strcpy(str,"DFILTER"); strcpy(fmtstr,"CF_FILTER"); break;
    case CF_SPECTRUM: strcpy(str,"DSPECTRUM"); strcpy(fmtstr,"CF_SPECTRUM"); break;
    case CF_ADDRESS: strcpy(str,"DADDRESS"); strcpy(fmtstr,"CF_ADDRESS"); break;
    case CF_WINDOW: strcpy(str,"DWINDOW"); strcpy(fmtstr,"CF_WINDOW"); break;
    case CF_POINT: strcpy(str,"DPOINT"); strcpy(fmtstr,"CF_POINT"); break;
  }
  if (fstr != NULL) strcpy(fstr,fmtstr);
  return str;
}
int isComplexProperty(ExportPropertyListStructEx *prp)
{
   if (prp->prpSizeIn == 0 || prp->prpSize == 0) return 0;
   if (prp->prpFormat == prp->prpFormatIn &&
	   prp->prpSize == prp->prpSizeIn) return 0;
   return -1;
}
int writeThisHdrFile(EQP *el)
{
  FILE *fp = NULL;
  char fname[64];
  char tag[64],lcltag[64],bigtag[64],fullFileName[256];
  char idleName[128],levelName[128],stepsName[128],procname[64],*c;
  char type[64],typeIn[64];
  char ndevstag[64];
  ExportPropertyListStructEx *prp;
  int cc=0;
  double mx,mn;
  time_t t = time(NULL);

  sprintf(ndevstag,"NUM_%s_DEVICES",el->lclname); strupr(ndevstag);

  sprintf(tag,"%.6s",el->lclname); strlwr(tag); strudr(tag);
  sprintf(fname,"%.6s.h",tag);
  sprintf(bigtag,"%.6s",tag); strupr(bigtag);

  if (strlen(targetPath) > 0) 
  {
    sprintf(fullFileName,"%.128s%s",targetPath,fname);	 
  }
  else
  {
    sprintf(fullFileName,"%s",fname);
  }
  if ((fp=fopen(fullFileName,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"/*\n * %s.h    REVISION HISTORY:\n",el->lclname);
  fprintf(fp," * Generated by SRVWIZARD on %s*/\n\n",ctime(&t));

  fprintf(fp,"#ifndef %s_H_\n",bigtag);
  fprintf(fp,"#define %s_H_\n\n",bigtag);
  
  fprintf(fp,"#define %s_TAG \"%s\"\n\n",bigtag,el->lclname);
  fprintf(fp,"#define %s_RATE %d\n\n",bigtag,1000);
  fprintf(fp,"#define %s %d\n\n",ndevstag,el->numdevs);
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    fprintf(fp,"#define PRP_%s %d\n",bigtag,prp->prpId);
    fprintf(fp,"#define PRP_%s_SIZE %d\n",bigtag,prp->prpSize);
    if (isComplexProperty(prp))
    {
      fprintf(fp,"#define PRP_%s_INSIZE %d\n",bigtag,prp->prpSizeIn);
    }
    if (prp->prpAccessMode & CA_WRITE)
    {
      mx = prp->egu.max;
      mn = prp->egu.min;
      fprintf(fp,"#define PRP_%s_UPR_LIMIT %g\n",bigtag,mx);
      fprintf(fp,"#define PRP_%s_LWR_LIMIT %g\n",bigtag,mn);
    }
  }
  fprintf(fp,"\n");
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    strcpy(type,gettype(prp->prpFormat,NULL));
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    sprintf(lcltag,"%s",prp->prpName); strlwr(lcltag); strudr(lcltag);
    if (prp->prpFormat == CF_STRUCT)
    {
      fprintf(fp,"\n/*TODO: change the following datatype to your structure :*/\n");
    }
    if (prp->prpSize == 1)
    {
      fprintf(fp,"extern %s g_%s;\n",type,lcltag);
    }
    else if (prp->prpSize > 1)
    {
      if (prp->prpArrayType == AT_SPECTRUM)
        fprintf(fp,"extern %s g_%sBuffer[%s][PRP_%s_SIZE];\n",type,lcltag,ndevstag,bigtag);
      else
        fprintf(fp,"extern %s g_%sBuffer[PRP_%s_SIZE];\n",type,lcltag,bigtag);
    }
    if (isComplexProperty(prp))
    {
      if (prp->prpFormat == CF_STRUCT)
      {
	      fprintf(fp,"\n/*TODO: change the following datatype to your structure :*/\n");
      }
	    strcpy(typeIn,gettype(prp->prpFormatIn,NULL));
      if (prp->prpSizeIn == 1)
      {
	      fprintf(fp,"extern %s g_%sIn;\n",typeIn,lcltag);
      }
      else if (prp->prpSizeIn > 1)
      {
	      fprintf(fp,"extern %s g_%sInBuffer[PRP_%s_INSIZE];\n",typeIn,lcltag,bigtag);
      }
    }
  }
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
	  if (prp->numProcessSteps > 0)
	  {
      strncpy(procname,prp->prpName,PROPERTY_NAME_SIZE);
      if ((c=strstr(procname,".START")) == NULL) continue;
      *c = 0;
      sprintf(idleName,"g_%sIdleCondition",procname);
      sprintf(levelName,"g_%sProcessLevel",procname);
      sprintf(stepsName,"g_%sNumProcessSteps",procname);
      fprintf(fp,"extern short %s;\n",idleName);
      fprintf(fp,"extern short %s;\n",levelName);
      fprintf(fp,"extern short %s;\n",stepsName);
      fprintf(fp,"int %sProcessFunction(void);\n",procname);
	  }
  }

  sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag);
  fprintf(fp,"\nint %s(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);\n",tag);
  fprintf(fp,"void %s_bkg(void);\n",tag);
  fprintf(fp,"void %s_ini(void);\n",tag);
  fprintf(fp,"void %s_exi(void);\n",tag);
  fprintf(fp,"\nint SimulateData(int start,int range);\n");
  fprintf(fp,"\nchar *SimulateStringData(int length);\n");

  fprintf(fp,"\n#endif\n");
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
void putGeneralHdrInfo(FILE *fp)
{
  fprintf(fp,"#include <stdio.h>\n");
  fprintf(fp,"#include <stdlib.h>\n");
  fprintf(fp,"#include <string.h>\n");
  fprintf(fp,"#include <math.h>\n");
  fprintf(fp,"#include <time.h>\n");
  fprintf(fp,"#include \"tine.h\"\n");
}
void putThisHdrInfo(FILE *fp,EQP *el)
{
  char hdr[64];
  sprintf(hdr,"%s",el->lclname); strlwr(hdr); strudr(hdr);
  fprintf(fp,"#include \"%s.h\"\n",hdr);
}
void putGeneralInfo(FILE *fp)
{
  char tag[64],lcltag[64];
  EQP *el;

  fprintf(fp,"\nint SimulateData(int start,int range)\n");
  fprintf(fp,"{\n");
  fprintf(fp,"  /* This routine is useful for test servers  */\n");
  if (!strcmp(OS_TAG,"DOS"))
  {
     fprintf(fp,"  return start + random(range);\n");
  }
  else 
  {
     fprintf(fp,"  return start + rand()"); fputc('%',fp); fprintf(fp,"(range);\n");
  }
  fprintf(fp,"}\n");

  fprintf(fp,"\nchar *SimulateStringData(int length)\n");
  fprintf(fp,"{\n  static char lclStringBuffer[512];\n");
  fprintf(fp,"  time_t t = time(NULL);\n  static int counter = 1;\n");
  fprintf(fp,"  /* This routine is useful for test servers  */\n");
  fprintf(fp,"  sprintf(lclStringBuffer,\""); fputc('%',fp); 
  fprintf(fp,"d : Simulated String generated "); fputc('%',fp); 
  fprintf(fp,"s\",counter++,ctime(&t));\n");
  fprintf(fp,"  if (length < 0) length = 0;\n");
  fprintf(fp,"  if (length < 512) lclStringBuffer[length] = 0;\n");
  fprintf(fp,"  return lclStringBuffer;\n");
  fprintf(fp,"}\n");

  fprintf(fp,"\nvoid PreSystemInit(void)\n");
  fprintf(fp,"{\n");
  fprintf(fp,"  /* Some System Settings with default values (as an example) */\n");
  fprintf(fp,"  /* Please see the server API documentation for other system settings */\n");
  fprintf(fp,"  SetConnectionTableCapacity(1000);\n");
  fprintf(fp,"  SetContractListCapacity(1000);\n");
  fprintf(fp,"  SetClientListCapacity(100);\n");
  fprintf(fp,"  SetWorkAreaSize(65535);\n\n");
  fprintf(fp,"  /* If you plan to use FEC Name registration via API call rather than */\n");
  if (strlen(fecname) > 0)
  {
    fprintf(fp,"  /* the fecid.csv startup file, then uncomment the following line */\n");
    fprintf(fp,"  /*\n\n  RegisterFecInformation(\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",%d);\n\n  */\n\n",
		    fecname,gDeviceSubSystem,gDeviceContext,RPCFecEx.desc,RPCFecEx.loc,RPCFecEx.hdw,RPCFecEx.resp,PortOffset);
  }
  else
  {
    fprintf(fp,"  /* the fecid.csv startup file, then you will need to supply the information */\n");
    fprintf(fp,"  /* in the following (commented) API call : */\n\n");
    fprintf(fp,"  /* RegisterFecInformation(fecname,subsystem,context,fecdesc,loc,hdw,resp,PortOffset); */\n\n");
  }
  fprintf(fp,"}\n\n");

  fprintf(fp,"void PostSystemInit(void)\n");
  fprintf(fp,"{\n");
  fprintf(fp,"  /* register equipment modules here: */\n");
  for (el=eqplist; el != NULL; el=el->next)  // go through the list
  {
    sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag);
    sprintf(lcltag,"%s",tag); strupr(lcltag);

    fprintf(fp,"  RegisterEquipmentModule(NULL,%s_TAG,0,%s,%s_ini,%s_bkg,%s_RATE,%s_exi);\n\n",
     	      lcltag,tag,tag,tag,lcltag,tag);
    fprintf(fp,"  /* Alternative:  you can forgo the 'exports.csv' file by registering all */\n");
    fprintf(fp,"  /* properties via API call and supplying the exported device server name as */\n");
    fprintf(fp,"  /*\n");
    fprintf(fp,"  RegisterEquipmentModule(\"%s\",%s_TAG,NUM_%s_DEVICES,%s,%s_ini,%s_bkg,%s_RATE,%s_exi);\n */\n",
     	      el->tagname,lcltag,lcltag,tag,tag,tag,lcltag,tag);
  }
  fprintf(fp,"  /* TODO: add any remaining general initialization code here */\n");
  fprintf(fp,"}\n");
}

void putMain(FILE *fp)
{
# ifndef VXWORKS
  fprintf(fp,"int main(int argc, char *argv[])\n");
  fprintf(fp,"{\n");
  fprintf(fp,"  int cc;\n");
  fprintf(fp,"  /* TODO: modify command line input to suit your own needs. */\n");
  fprintf(fp,"  /* Here: the debug level is set according to initial input */\n");
  fprintf(fp,"  switch (argc)\n");
  fprintf(fp,"  {\n");
  fprintf(fp,"    case 2:\n");
  fprintf(fp,"      tineDebug = atoi(argv[1]);\n");
  fprintf(fp,"      break;\n");
  fprintf(fp,"    default:\n");
  fprintf(fp,"      tineDebug = 0;\n");
  fprintf(fp,"  }\n");
# else
  fprintf(fp,"int tineServer(void)\n");
  fprintf(fp,"{\n");
  fprintf(fp,"  int cc;\n\n");
# endif
  fprintf(fp,"  /* To deviate from default settings, initialize */\n");
  fprintf(fp,"  /* any system variables in PreSystemInit() */\n\n");
  fprintf(fp,"  /* Make all other registrations in PostSystemInit() */\n\n");
  fprintf(fp,"  SetPreSystemInitFunction(PreSystemInit);\n");
  fprintf(fp,"  SetPostSystemInitFunction(PostSystemInit);\n");
  fprintf(fp,"  /* initialize tine server: */\n\n");
  fprintf(fp,"  if ((cc=SystemInit(TRUE)) != 0)\n");
  fprintf(fp,"  {\n");
  fprintf(fp,"    printf(cc2str(cc));\n");
  fprintf(fp,"    exit(1);\n");
  fprintf(fp,"  }\n\n");
  fprintf(fp,"  for(;;)\n");
  fprintf(fp,"  {\n");
  fprintf(fp,"    /* TODO: add any additional non-TINE processing here */\n");
  fprintf(fp,"    /*       being careful not to block operations */\n\n");
  fprintf(fp,"    SystemCycle(TRUE);\n");
  fprintf(fp,"  }\n");
  fprintf(fp,"  return 0;\n");
  fprintf(fp,"}\n");
}

void putThisBkgInfo(FILE *fp,EQP *el)
{
  char tag[64],bigtag[64],lcltag[64],type[64],eqtag[64],procname[64],*c;
  char ndevstag[64];
  ExportPropertyListStructEx *prp;
  int mx,mn;

  sprintf(ndevstag,"NUM_%s_DEVICES",el->lclname); strupr(ndevstag);

  fprintf(fp,"\n");
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    strcpy(type,gettype(prp->prpFormat,NULL));
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    sprintf(lcltag,"%s",prp->prpName); strlwr(lcltag); strudr(lcltag);
    if (prp->prpFormat == CF_STRUCT)
    {
      fprintf(fp,"\n/* TODO: change the following datatype to your structure : */\n");
    }
    if (prp->numProcessSteps > 0)
    {
      strncpy(procname,prp->prpName,PROPERTY_NAME_SIZE);
      if ((c=strstr(procname,".START")) == NULL) continue;
      *c = 0;
      fprintf(fp,"short g_%sIdleCondition;\n",procname);
      fprintf(fp,"short g_%sProcessLevel;\n",procname);
      fprintf(fp,"short g_%sNumProcessSteps = %d;\n",procname,prp->numProcessSteps);
    }
    else if (prp->prpSize == 1)
    {
      fprintf(fp,"%s g_%s;\n",type,lcltag);
    }
    else if (prp->prpSize > 1)
    {
      if (prp->prpArrayType == AT_SPECTRUM)
        fprintf(fp,"%s g_%sBuffer[%s][PRP_%s_SIZE];\n",type,lcltag,ndevstag,bigtag);
      else
        fprintf(fp,"%s g_%sBuffer[PRP_%s_SIZE];\n",type,lcltag,bigtag);
    }
    if (isComplexProperty(prp))
    {
	    strcpy(type,gettype(prp->prpFormatIn,NULL));
      if (prp->prpFormatIn == CF_STRUCT)
      {
        fprintf(fp,"\n/* TODO: change the following datatype to your structure : */\n");
      }
      if (prp->prpSizeIn == 1)
      {
        fprintf(fp,"%s g_%sIn;\n",type,lcltag);
      }
      else if (prp->prpSizeIn > 1)
      {
        fprintf(fp,"%s g_%sInBuffer[PRP_%s_INSIZE];\n",type,lcltag,bigtag);
      }
    }
  }

  sprintf(tag,"%.6s",el->lclname); strlwr(tag); strudr(tag);
  sprintf(eqtag,"%.6s_TAG",tag); strupr(eqtag);
  strcat(tag,"_bkg");
  fprintf(fp,"\nvoid %s(void)\n",tag);
  fprintf(fp,"{\n  int i, n;\n");
  fprintf(fp,"  /* TODO: put your IO or other background activity here */\n\n");
  fprintf(fp,"  /* Clear alarms at start of IO task (see if they come back) */\n");
  fprintf(fp,"  ClearAlarm(%s,-1);\n\n",eqtag);

  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    if (prp->prpAccessMode & CA_WRITE) continue;
    if (prp->numProcessSteps > 0) continue;
    strcpy(type,gettype(prp->prpFormat,NULL));
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    sprintf(lcltag,"%s",prp->prpName); strlwr(lcltag); strudr(lcltag);
    mx = (int)prp->egu.max;
    mn = (int)prp->egu.min;
    if (mx == mn) mx = mn + 100;
    switch( prp->prpFormat)
    {
      case CF_SHORT:
	    case CF_LONG:
	    case CF_FLOAT:
	    case CF_DOUBLE:
	    case CF_BYTE:
        fprintf(fp,"  /* TODO: replace SimulateData() with your own data acquisition */\n");
        fprintf(fp,"  /*       and manipulation routines */\n");
        if (prp->prpSize == 1)
        {
          fprintf(fp,"  g_%s = (%s)SimulateData(%d,%d);\n",lcltag,type,mn,mx-mn);
        }
        else if (prp->prpSize > 1)
        {
          if (prp->prpArrayType == AT_SPECTRUM)
          {
            fprintf(fp,"  for (n=0; n<%s; n++)\n",ndevstag);
            fprintf(fp,"    for (i=0; i<PRP_%s_SIZE; i++)\n",bigtag);
            fprintf(fp,"  {\n");
            fprintf(fp,"     g_%sBuffer[n][i] = (%s)SimulateData(%d,%d);\n",lcltag,type,mn,mx-mn);
            fprintf(fp,"  }\n");
          }
          else
          {
            fprintf(fp,"  for (i=0; i<PRP_%s_SIZE; i++)\n",bigtag);
            fprintf(fp,"  {\n");
            fprintf(fp,"     g_%sBuffer[i] = (%s)SimulateData(%d,%d);\n",lcltag,type,mn,mx-mn);
            fprintf(fp,"  }\n");
          }
        }
        break;
  	  case CF_TEXT:
        if (prp->prpSize > 1)
        {
          fprintf(fp,"  /* TODO: replace SimulateStringData() with your own data acquisition */\n");
          fprintf(fp,"  /*       and manipulation routines */\n");
          fprintf(fp,"  strncpy(g_%sBuffer,SimulateStringData(PRP_%s_SIZE),PRP_%s_SIZE);\n",
          lcltag,bigtag,bigtag);
        }
      default:
        break;
    }
  }

  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    if (prp->numProcessSteps > 0)
    {
      strncpy(procname,prp->prpName,PROPERTY_NAME_SIZE);
      if ((c=strstr(procname,".START")) == NULL) continue;
      *c = 0;
      fprintf(fp,"\n  /* call process function  */\n");
      fprintf(fp,"  %sProcessFunction();\n",procname);
    }
  }
  
  fprintf(fp,"\n  /* make use of SetAlarm() as needed :\n\n");
  fprintf(fp,"  SetAlarm(%s,devNr,almCode,almFlag);\n\n",eqtag);
  fprintf(fp,"  */\n");
  fprintf(fp,"}\n");
}
int putProcessFunction(FILE *fp,char *procname, int nsteps)
{
  char fcnName[128],idleName[128],exitName[128],levelName[128],stepsName[128];
  int i;

  sprintf(exitName,"%sProcessFunctionExit",procname);
  sprintf(fcnName,"%sProcessFunction",procname);
  sprintf(idleName,"g_%sIdleCondition",procname);
  sprintf(levelName,"g_%sProcessLevel",procname);
  sprintf(stepsName,"g_%sNumProcessSteps",procname);
  fprintf(fp,"int %s(void)\n",fcnName);
  fprintf(fp,"{\n");
  fprintf(fp,"  static int duration = 0;\n");
  fprintf(fp,"  static time_t start = 0;\n");
  fprintf(fp,"  time_t interim = time(NULL);\n\n");
  fprintf(fp,"  if (%s) goto %s;\n",idleName,exitName);

  fprintf(fp,"  /* TODO: fill in the activities to be performed at each step       */\n");
  fprintf(fp,"  /*       The following example code simply waits for 3 seconds at  */\n");
  fprintf(fp,"  /*       each process level.  You may want to start IO readouts    */\n");
  fprintf(fp,"  /*       at one level and poll for a ready flag in another, etc.   */\n");
  fprintf(fp,"  if (%s > 0 && duration == 0)\n",levelName);
  fprintf(fp,"  {\n");
  fprintf(fp,"    start = interim;\n");
  fprintf(fp,"    duration = 3;\n");
  fprintf(fp,"  }\n");
  fprintf(fp,"  switch (%s)\n",levelName);
  fprintf(fp,"  {\n");
  for (i=1; i<=nsteps; i++)
  {
    fprintf(fp,"    case %d:\n",i);
    fprintf(fp,"	  if (interim < start + duration) return 0;\n");
    fprintf(fp,"	  duration = 0;\n");
    fprintf(fp,"      break;\n");
  }
  fprintf(fp,"    default:\n");
  fprintf(fp,"      %s = 0;\n",levelName);
  fprintf(fp,"      break;\n");
  fprintf(fp,"  }\n");
  fprintf(fp,"  if (%s > 0) %s = (%s + 1) ",levelName,levelName,levelName);
  fputc('%',fp); 
  fprintf(fp," (%s + 1);\n",stepsName);
  fprintf(fp,"%s:\n",exitName);
  fprintf(fp,"  return 0;\n");
  fprintf(fp,"}\n");
  return 0;
}
void putRegisterPropertyInfo(FILE *fp, EQP *el)
{
  int cc = 0;
  ExportPropertyListStructEx *prp;
  char desc[256],mode[128];
  char tag[64],lcltag[64],bigtag[64],type[64],cftype[64],hungtag[64];
  char atype[64];

  sprintf(tag,"%.6s",el->lclname); strlwr(tag); strudr(tag); strncpy(hungtag,tag,64);
  sprintf(lcltag,"%.6s_TAG",tag); strupr(lcltag); 
  hungtag[0] = toupper(hungtag[0]);
  fprintf(fp,"\n/* Alternative to 'exports.csv': You can call 'Register%sProperties()' */\n",hungtag);
  fprintf(fp,"/*   -> This might be preferable in cases where the server does not have a */\n");
  fprintf(fp,"/*   -> file system.  You should also call RegisterServerName() in lieu of */\n");
  fprintf(fp,"/*   -> the 'fecid.csv' file.  */\n");
  fprintf(fp,"\nint Register%sProperties(void)\n",hungtag);
  fprintf(fp,"{\n  int cc = 0;\n");
  fprintf(fp,"  DTYPE dout,din;\n\n");
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    strcpy(type,gettype(prp->prpFormat,cftype));
    if (prp->egu.min == 0 && prp->egu.max == 0 && strlen(prp->egu.units) == 0)
	    sprintf(desc,"%s",prp->prpDescription);
	  else if (prp->xegu.min == 0 && prp->xegu.max == 0)
	    sprintf(desc,"[%g:%g %s]%s",prp->egu.min,prp->egu.max,prp->egu.units,prp->prpDescription);
    else
	    sprintf(desc,"[%g:%g %.16s][%g:%g %.16s]%.64s",
        prp->egu.min,prp->egu.max,prp->egu.units,
        prp->xegu.min,prp->xegu.max,prp->xegu.units,
        prp->prpDescription);

	  strcpy(mode,"CA_READ");
	  if (prp->prpAccessMode&CA_WRITE) strcat(mode,"|CA_WRITE");

    getatype(prp->prpArrayType,atype);
	  fprintf(fp,"  dout.dArrayLength = PRP_%s_SIZE;\n",bigtag);
	  fprintf(fp,"  dout.dFormat = %s;\n",cftype);

    if (prp->prpSizeIn > 0)
    {
      strcpy(type,gettype(prp->prpFormatIn,cftype));
      if (isComplexProperty(prp))
	      fprintf(fp,"  din.dArrayLength = PRP_%s_INSIZE;\n",bigtag);
      else
	      fprintf(fp,"  din.dArrayLength = PRP_%s_SIZE;\n",bigtag);
  	  fprintf(fp,"  din.dFormat = %s;\n",cftype);
    }
    else
    {
      fprintf(fp,"  din.dArrayLength = 0;\n");
	    fprintf(fp,"  din.dFormat = CF_NULL;\n");
    }

	fprintf(fp,"  if ((cc=RegisterPropertyInformation(%s,\"%s\",&dout,&din,%s,%s,0,\"%s\",PRP_%s,\"\")) < 0) goto err;\n\n",
            lcltag,prp->prpName,mode,atype,desc,bigtag);

  }
  fprintf(fp,"err:\n");
  fprintf(fp,"  if (cc < 0)\n");
  fprintf(fp,"  {\n");
  fprintf(fp,"    printf(\"Can't register property : ");
  fputc('%',fp); fprintf(fp,"s\\n>\",erlst[-cc]);\n");
  fprintf(fp,"  }\n");
  fprintf(fp,"  return cc;\n");
  fprintf(fp,"}\n\n");
  return;
}
void putThisIniInfo(FILE *fp,EQP *el)
{
  char tag[64],lcltag[64],hungtag[64],ndevstag[64];
  sprintf(ndevstag,"NUM_%s_DEVICES",el->lclname); strupr(ndevstag);
  putRegisterPropertyInfo(fp,el);
  sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag); strncpy(hungtag,tag,64);
  sprintf(lcltag,"%.6s_TAG",tag); strupr(lcltag);
  strcat(tag,"_ini");
  fprintf(fp,"\nvoid %s(void)\n",tag);
  fprintf(fp,"{\n  char devnam[DEVICE_NAME_SIZE];\n  int i;\n");
  fprintf(fp,"  /* TODO: put your initialization here */\n");
  fprintf(fp,"  /* If you plan to use Property registration via API call rather than */\n");
  fprintf(fp,"  /* the exports.csv startup file, uncomment the following line: */\n\n");
  hungtag[0] = toupper(hungtag[0]);
  fprintf(fp,"  /* Register%sProperties(); */\n\n",hungtag);
  fprintf(fp,"  /* TODO: register any device names here */\n");
  fprintf(fp,"  /*       below is only an example ...   */\n");
  fprintf(fp,"  for (i=0; i<%s; i++)\n",ndevstag);
  fprintf(fp,"  {\n");
  fprintf(fp,"    sprintf(devnam,\"device_"); fputc('%',fp); fprintf(fp,"d\",i);\n");
  fprintf(fp,"    RegisterDeviceName(%s,devnam,i);\n",lcltag);
  fprintf(fp,"  }\n");
  fprintf(fp,"\n  /* TODO: register any global receive routines as in this example */\n");
  fprintf(fp,"  /*\n\n");
  fprintf(fp,"  glbID = recvNetGlobal(\"HPMAGEN\",&dout,rcvGlobal);\n");
  fprintf(fp,"\n  */\n\n  /* TODO: start any links you might need as in this example */\n");
  fprintf(fp,"  /*\n\n");
  fprintf(fp,"  lnkID = AttachLinkEx(\"/HERA/BPM/WL197X\",\"POSITIONS.X\",&dout,&din,CA_READ,1000,showlink,CM_POLL,myid);\n");
  fprintf(fp,"\n  */\n");
  fprintf(fp,"}\n");
}
void putThisExiInfo(FILE *fp,EQP *el)
{
  char tag[64];
  sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag);
  strcat(tag,"_exi");
  fprintf(fp,"\nvoid %s(void)\n",tag);
  fprintf(fp,"{\n");
  fprintf(fp,"  /* TODO: put your shutdown routines here */\n");
  fprintf(fp,"}\n");
}
void putThisEqmInfo(FILE *fp,EQP *el)
{
  char tag[72],lcltag[72],bigtag[72],arrtag[72],type[72],cftype[72],procname[72],*c;
  char slcltag[72],sarrtag[72],prpsiztag[72];
  char slcltagIn[72],lcltagIn[72];
  ExportPropertyListStructEx *prp;

  sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag);
  sprintf(lcltag,"%.6s_TAG",tag); strupr(lcltag);
  fprintf(fp,"\nint %s(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)\n",tag);
  fprintf(fp,"{\n  int devnr,prpid,i,cc;\n");
  /* local variables for write commands */
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    if (prp->prpSizeIn == 0) continue;
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    strcpy(type,gettype(prp->prpFormatIn,cftype));
    sprintf(arrtag,"l_%.64s",prp->prpName); strlwr(arrtag); strudr(arrtag);
    if (prp->prpSizeIn > 1) 
	  {
      if (isComplexProperty(prp))
        sprintf(&arrtag[strlen(arrtag)],"[PRP_%s_INSIZE]",bigtag);
      else
        sprintf(&arrtag[strlen(arrtag)],"[PRP_%s_SIZE]",bigtag);
	  }
    fprintf(fp,"  %s %s;\n",type,arrtag);  
  }

  fprintf(fp,"\n");

  fprintf(fp,"  /* get device number from device name */\n");
  fprintf(fp,"  devnr = GetDeviceNumberEx(%s,devName,devProperty);\n",lcltag);
  fprintf(fp,"  /* you may want to make the following check: */\n");
  fprintf(fp,"  if (devnr < 0) return illegal_equipment_number;\n\n");  

  fprintf(fp,"  /* TODO: If READ properties take input data, include code to examine the contents of din. */\n");
  fprintf(fp,"  /*       If different actions need to be taken at the start or end of a link, examine the */\n");
  fprintf(fp,"  /*       'access' parameter against CA_FIRST or CA_LAST. */\n");
  fprintf(fp,"  /*       If format overloading allowed (you return different data according to the request */\n");
  fprintf(fp,"  /*       format), then replace calls to PutDataFromShort() etc with the desired code. */\n\n");
  
  fprintf(fp,"  prpid = GetPropertyId(%s,devProperty);\n\n",lcltag);
  fprintf(fp,"  switch (prpid)\n");
  fprintf(fp,"  {\n");
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    strncpy(procname,prp->prpName,PROPERTY_NAME_SIZE);
    sprintf(bigtag,"%s",prp->prpName); strupr(bigtag); strudr(bigtag);
    fprintf(fp,"    case PRP_%s:\n",bigtag);
    if (prp->prpSize == 0 && prp->prpSizeIn == 0 && prp->prpAccessMode & CA_WRITE)
    {
      fprintf(fp,"        if (access&CA_WRITE)\n");
      fprintf(fp,"        {\n");

      if (prp->numProcessSteps > 0 && (c=strstr(procname,".PAUSE")) != NULL)
      {
        *c = 0;
        fprintf(fp,"           g_%sIdleCondition = TRUE;\n",procname);
      }
      else if (prp->numProcessSteps > 0 && (c=strstr(procname,".RESUME")) != NULL)
      {
        *c = 0;
        fprintf(fp,"           g_%sIdleCondition = FALSE;\n",procname);
      }
      else if (prp->numProcessSteps > 0 && (c=strstr(procname,".STOP")) != NULL)
      {
        *c = 0;
        fprintf(fp,"           g_%sProcessLevel = 0;\n",procname);
        fprintf(fp,"           g_%sIdleCondition = FALSE;\n",procname);
      }
      else 
      {
        fprintf(fp,"           if (din->dArrayLength > 0) return illegal_read_write;\n");
        fprintf(fp,"           /* TODO: Take action for property %s here */\n",bigtag);
      }
      fprintf(fp,"        }\n");
    }
    if (prp->prpSize > 0 || prp->prpSizeIn > 0)
    {
      strcpy(type,gettype(prp->prpFormat,cftype));
      sprintf(lcltag,"&g_%s",prp->prpName); strlwr(lcltag); strudr(lcltag);
      sprintf(slcltag,"g_%s",prp->prpName); strlwr(slcltag); strudr(slcltag);
      sprintf(sarrtag,"l_%s",prp->prpName); strlwr(sarrtag); strudr(sarrtag);
      if (prp->prpSize > 1) 
      {
        sprintf(lcltag,"g_%s",prp->prpName); strlwr(lcltag); strudr(lcltag);
        strcat(lcltag,"Buffer");
      }
      if (isComplexProperty(prp) && prp->prpSizeIn > 0) 
      {
        sprintf(lcltagIn,"&g_%s",prp->prpName); strlwr(lcltagIn); strudr(lcltagIn);
        sprintf(slcltagIn,"g_%s",prp->prpName); strlwr(slcltagIn); strudr(slcltagIn);
        strcat(lcltagIn,"In"); strcat(slcltagIn,"In");
        if (prp->prpSizeIn > 1) 
        {
          sprintf(lcltagIn,"g_%s",prp->prpName); strlwr(lcltagIn); strudr(lcltagIn);
          strcat(lcltagIn,"InBuffer");
        }
      }

      if (!(prp->prpAccessMode & CA_WRITE))
      {
        fprintf(fp,"        if (access&CA_WRITE) return illegal_read_write;\n");
      }
      else   // CA_WRITE
      {
        if (prp->prpSizeIn > 1) sprintf(arrtag,"l_%s",prp->prpName);
        else sprintf(arrtag,"&l_%s",prp->prpName);

        strlwr(arrtag); strudr(arrtag);
        fprintf(fp,"        if (access&CA_WRITE)\n");
        fprintf(fp,"        {\n");

        if (prp->prpSizeIn > 0)
        {
          fprintf(fp,"          if (din->dArrayLength > 0)\n");
          fprintf(fp,"          {\n");

          if (isComplexProperty(prp)) sprintf(prpsiztag,"PRP_%.6s_INSIZE",bigtag);
          else sprintf(prpsiztag,"PRP_%.6s_SIZE",bigtag);

          fprintf(fp,"            if (din->dArrayLength > %s) return dimension_error;\n",prpsiztag);


          if (prp->numProcessSteps > 0 && (c=strstr(procname,".START")) != NULL)
          {
            *c = 0;
            fprintf(fp,"            /* Do range checking: */\n");
            switch( prp->prpFormatIn)
            {
              case CF_SHORT:
                fprintf(fp,"            if ((cc=GetValuesAsShort(din,%s,%s)) != 0) return cc;\n",arrtag,prpsiztag);
                break;
              case CF_LONG:
                fprintf(fp,"            if ((cc=GetValuesAsLong(din,%s,%s)) != 0) return cc;\n",arrtag,prpsiztag);
                break;
            }
            fprintf(fp,"            if (%s > g_%sNumProcessSteps)\n",sarrtag,procname);
            fprintf(fp,"            {\n");
            fprintf(fp,"              return out_of_range;\n");
            fprintf(fp,"            }\n");
            fprintf(fp,"            if (%s < 0)\n",arrtag);
            fprintf(fp,"            {\n");
            fprintf(fp,"              return out_of_range;\n");
            fprintf(fp,"            }\n");
            fprintf(fp,"            /* Copy incoming data into global variable */\n");
            fprintf(fp,"            g_%sProcessLevel = %s;\n",procname,sarrtag);
            fprintf(fp,"            g_%sIdleCondition = FALSE;\n",procname);
          }
          else
          {			
            switch( prp->prpFormatIn)
            {
              case CF_SHORT:
                fprintf(fp,"            if ((cc=GetValuesAsShort(din,%s,%s)) != 0) return cc;\n",arrtag,prpsiztag);
                break;
              case CF_LONG:
                fprintf(fp,"            if ((cc=GetValuesAsLong(din,%s,%s)) != 0) return cc;\n",arrtag,prpsiztag);
                break;
              case CF_FLOAT:
                fprintf(fp,"            if ((cc=GetValuesAsFloat(din,%s,%s)) != 0) return cc;\n",arrtag,prpsiztag);
                break;
              case CF_DOUBLE:
                fprintf(fp,"            if ((cc=GetValuesAsDouble(din,%s,%s)) != 0) return cc;\n",arrtag,prpsiztag);
                break;
              case CF_TEXT:
                fprintf(fp,"            if (din->dFormat != %s) return illegal_format;\n",cftype);
                fprintf(fp,"            strncpy(%s,din->data.cptr,din->dArrayLength);\n",arrtag);
                break;
              default:
                strupr(type);
                fprintf(fp,"            if (din->dFormat != %s) return illegal_format;\n",cftype);
                fprintf(fp,"            memcpy(%s,din->data.vptr,din->dArrayLength*fmtsizeof(din->dFormat));\n",arrtag);
                break;
            }
            switch( prp->prpFormatIn)
            {
              case CF_BYTE:
              case CF_SHORT:
              case CF_LONG:
              case CF_FLOAT:
              case CF_DOUBLE:
                if (prp->prpSizeIn > 1) 
                {
                  fprintf(fp,"            for (i=0; i<(int)din->dArrayLength; i++)\n");
                  fprintf(fp,"            {\n");
                  fprintf(fp,"              if (%s[i] > PRP_%s_UPR_LIMIT) return out_of_range;\n",arrtag,bigtag);
                  fprintf(fp,"              if (%s[i] < PRP_%s_LWR_LIMIT) return out_of_range;\n",arrtag,bigtag);
                  fprintf(fp,"            }\n");
                  if (isComplexProperty(prp)) 
                    fprintf(fp,"            memcpy(%s,%s,din->dArrayLength*fmtsizeof(din->dFormat));\n",lcltagIn,arrtag);
                  else
                    fprintf(fp,"            memcpy(%s,%s,din->dArrayLength*fmtsizeof(din->dFormat));\n",lcltag,arrtag);
                }
                if (prp->prpSizeIn == 1)
                {
                  fprintf(fp,"            if (%s > PRP_%s_UPR_LIMIT) return out_of_range;\n",sarrtag,bigtag);
                  fprintf(fp,"            if (%s < PRP_%s_LWR_LIMIT) return out_of_range;\n",sarrtag,bigtag);
                  if (isComplexProperty(prp)) 
                    fprintf(fp,"            %s = %s;\n",slcltagIn,sarrtag);
                  else
                    fprintf(fp,"            %s = %s;\n",slcltag,sarrtag);
                }
                break;
              case CF_TEXT:
                fprintf(fp,"            /* TODO:  Insert any necessary checking on the incoming string */\n");
                if (isComplexProperty(prp)) 
                  fprintf(fp,"            strncpy(%s,%s,din->dArrayLength);\n",lcltagIn,arrtag);
                else
                  fprintf(fp,"            strncpy(%s,%s,din->dArrayLength);\n",lcltag,arrtag);
                break;
              default:
                fprintf(fp,"            /* TODO:  Insert any necessary bounds checking on the incoming data */\n");
                break;
            }

          }
          fprintf(fp,"          }\n");
        }
        else
        {
          fprintf(fp,"          if (din->dArrayLength > 0) return illegal_read_write;\n");
        }
        fprintf(fp,"        }\n");
 	    }

      if (prp->prpSize > 1) sprintf(arrtag,"l_%s",prp->prpName);
      else sprintf(arrtag,"&l_%s",prp->prpName);
      strncpy(procname,prp->prpName,PROPERTY_NAME_SIZE);

      fprintf(fp,"        if (dout->dArrayLength > 0)\n");
      fprintf(fp,"        {\n");
      fprintf(fp,"          if (dout->dArrayLength > PRP_%s_SIZE) return dimension_error;\n",bigtag);

      if (prp->numProcessSteps > 0 && (c=strstr(procname,".START")) != NULL)
      {
        *c = 0;
        fprintf(fp,"          if ((cc=PutValuesFromShort(dout,&g_%sProcessLevel,1)) != 0) return cc;\n",procname);
 	    }
      else if (prp->numProcessSteps > 0 && (c=strstr(procname,".STATUS")) != NULL)
      {
        *c = 0;
        fprintf(fp,"          if ((cc=PutValuesFromShort(dout,&g_%sProcessLevel,1)) != 0) return cc;\n",procname);
      }
      else
      {
        if (prp->prpSize > 0)
        {
          switch( prp->prpFormat)
          {
            case CF_SHORT:
              if (prp->prpSize > 1 && (int)prp->prpSize == el->numdevs && prp->prpArrayType == AT_CHANNEL)
              {
                fprintf(fp,"          if ((cc=PutValuesFromShortEx(dout,%s,PRP_%s_SIZE,devnr)) != 0) return cc;\n",lcltag,bigtag);
              }
              else if (prp->prpArrayType == AT_SPECTRUM)
              {
                fprintf(fp,"          if ((cc=PutValuesFromShort(dout,%s[devnr],PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              else
              {
                fprintf(fp,"          if ((cc=PutValuesFromShort(dout,%s,PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              break;
            case CF_LONG:
              if (prp->prpSize > 1 && (int)prp->prpSize == el->numdevs && prp->prpArrayType == AT_CHANNEL)
              {
                fprintf(fp,"          if ((cc=PutValuesFromLongEx(dout,%s,PRP_%s_SIZE,devnr)) != 0) return cc;\n",lcltag,bigtag);
              }
              else if (prp->prpArrayType == AT_SPECTRUM)
              {
                fprintf(fp,"          if ((cc=PutValuesFromLong(dout,%s[devnr],PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              else
              {
                fprintf(fp,"          if ((cc=PutValuesFromLong(dout,%s,PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              break;
            case CF_FLOAT:
              if (prp->prpSize > 1 && (int)prp->prpSize == el->numdevs && prp->prpArrayType == AT_CHANNEL)
              {
                fprintf(fp,"          if ((cc=PutValuesFromFloatEx(dout,%s,PRP_%s_SIZE,devnr)) != 0) return cc;\n",lcltag,bigtag);
              }
              else if (prp->prpArrayType == AT_SPECTRUM)
              {
                fprintf(fp,"          if ((cc=PutValuesFromFloat(dout,%s[devnr],PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              else
              {
                fprintf(fp,"          if ((cc=PutValuesFromFloat(dout,%s,PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              break;
            case CF_DOUBLE:
              if (prp->prpSize > 1 && (int)prp->prpSize == el->numdevs && prp->prpArrayType == AT_CHANNEL)
              {
                fprintf(fp,"          if ((cc=PutValuesFromDoubleEx(dout,%s,PRP_%s_SIZE,devnr)) != 0) return cc;\n",lcltag,bigtag);
              }
              else if (prp->prpArrayType == AT_SPECTRUM)
              {
                fprintf(fp,"          if ((cc=PutValuesFromDouble(dout,%s[devnr],PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              else
              {
                fprintf(fp,"          if ((cc=PutValuesFromDouble(dout,%s,PRP_%s_SIZE)) != 0) return cc;\n",lcltag,bigtag);
              }
              break;
            case CF_TEXT:
              fprintf(fp,"          if (dout->dFormat != %s) return illegal_format;\n",cftype);
              fprintf(fp,"          strncpy(dout->data.cptr,%s,dout->dArrayLength);\n",lcltag);
              break;
            default:
              strupr(type);
              fprintf(fp,"          if (dout->dFormat != %s) return illegal_format;\n",cftype);
              fprintf(fp,"          memcpy(dout->data.vptr,%s,dout->dArrayLength*fmtsizeof(dout->dFormat));\n",lcltag);
              break;
          }
        }
      }
      fprintf(fp,"        }\n");
 	  }	
    fprintf(fp,"        return 0;\n");
  }
  fprintf(fp,"    default:\n");
  fprintf(fp,"        return illegal_property;\n");
  fprintf(fp,"  }\n");
  fprintf(fp,"}\n");
}
int putEQPTemplate(EQP *el)
{
  FILE *fp = NULL;
  char fname[64],fullFileName[256],tag[64],procname[64],*c;
  int cc=0;
  time_t t = time(NULL);
  ExportPropertyListStructEx *prp;

  if ((cc=writeThisHdrFile(el)) != 0) ccerr(cc);
  sprintf(tag,"%.6s",el->lclname); strlwr(tag); strudr(tag);
  sprintf(fname,"%.6s.c",tag);
  if (strlen(targetPath) > 0) 
  {
    sprintf(fullFileName,"%.128s%s",targetPath,fname);	 
  }
  else
  {
    sprintf(fullFileName,"%s",fname);
  }
  if ((fp=fopen(fullFileName,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"/*\n * %s.c    REVISION HISTORY:\n",tag);
  fprintf(fp," * Generated by SRVWIZARD on %s*/\n\n",ctime(&t));
  putGeneralHdrInfo(fp);
  putThisHdrInfo(fp,el);
  putThisBkgInfo(fp,el);
  for (prp=el->prp; prp != NULL; prp = prp->next)
  {
    if (prp->numProcessSteps > 0)
    {
      strncpy(procname,prp->prpName,PROPERTY_NAME_SIZE);
      if ((c=strstr(procname,".START")) == NULL) continue;
      *c = 0;
      putProcessFunction(fp,procname,prp->numProcessSteps);
    }
  }
  putThisIniInfo(fp,el);
  putThisExiInfo(fp,el);
  putThisEqmInfo(fp,el);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int putDevServerTemplate(char *fn)
{
  FILE *fp = NULL;
  EQP *el;
  char fname[64],fullFileName[256];
  int cc=0;
  time_t t = time(NULL);

  sprintf(fname,"%s.c",fn);
  if (strlen(targetPath) > 0) 
  {
	 sprintf(fullFileName,"%.128s%s",targetPath,fname);	 
  }
  else
  {
	 sprintf(fullFileName,"%s",fname);
  }
  if ((fp=fopen(fullFileName,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"/*\n * %s    REVISION HISTORY:\n",fn);
  fprintf(fp," * Generated by SRVWIZARD on %s*/\n\n",ctime(&t));
  putGeneralHdrInfo(fp);
  for (el=eqplist; el != NULL; el=el->next) putThisHdrInfo(fp,el);
  putGeneralInfo(fp);
  putMain(fp);

err:
  if (fp != NULL) fclose(fp);
  return cc;
}


int putMakeFile(char *ostype)
{
  FILE *fp = NULL;
  int cc=0;
  EQP *el;
  char fullFileName[256],tag[64],modlist[256],os[64],*c;
  time_t t = time(NULL);

  if (ostype != NULL && strlen(ostype) > 0) strncpy(os,ostype,64);
  else
  {
    if (!strcmp(OS_TAG,"DOS") || !strcmp(OS_TAG,"WIN32")) return 0;
#   ifdef _HPUX_SOURCE
    strcpy(os,"HP");
#   else
    if ((c=getenv("OSTYPE")) != NULL) strncpy(os,c,64); else strcpy(os,"LINUX");
#   endif
    strupr(os);
  }

  if (strlen(targetPath) > 0) 
  {
	 sprintf(fullFileName,"%.128s%s",targetPath,MAKEFILE_NAME);	 
  }
  else
  {
	 sprintf(fullFileName,"%s",MAKEFILE_NAME);
  }
  if ((fp=fopen(fullFileName,"w")) == NULL) ccerr(file_error);

  fprintf(fp,"# Makefile for TINE server generated on %s",ctime(&t));
  fprintf(fp,"# SUN:                  CC = CC         CCL = CC         LM = \n");
  fprintf(fp,"# SGI,OSF,LINUX,FreeBSD CC = cc         CCL = cc         LM = -lm\n");
  fprintf(fp,"# ULT                   CC = gcc -c -g  CCL = gcc -g     LM = \n");
  fprintf(fp,"# HP:                   CC = cc -Aa     CCL = cc         LM = -lM\n");
  fprintf(fp,"# SOLARIS               CC = CC         CCL = CC         LM = -lsocket -lnsl\n");
  fprintf(fp,"# MACOSX                CC = clang -m64 -fPIC -w -c\n");
  fprintf(fp,"# MACOSX                CCL = clang -m64 -fPIC -w -undefined suppress\n");
  fprintf(fp,"# MACOSX                LM = -lm -lpthread\n");
  fprintf(fp,"#\n");
  fprintf(fp,"# Multi-threaded builds: use libtinemt.a (or libtimemt.so) and add -DMULTITHREADED to CC definition\n");
  fprintf(fp,"#\n");
  fprintf(fp,"\nINCL = /usr/include/tine\n");
  fprintf(fp,"TINEEXE = devserv\n");
  fprintf(fp,"TINELIB = libtine.a\n");
  fprintf(fp,"LIBS = /usr/lib\n");
  fprintf(fp,"HDRS = ${INCL}/project.def \\\n");
  fprintf(fp,"	${INCL}/prolog.h \\\n");
  fprintf(fp,"	${INCL}/srvcore.h \\\n");
  fprintf(fp,"	${INCL}/srvip.h \\\n");
  fprintf(fp,"	${INCL}/errors.h \\\n");
  fprintf(fp,"	${INCL}/tinetype.h \\\n");
  fprintf(fp,"	${INCL}/tinedefs.h \\\n");
  fprintf(fp,"	${INCL}/dbtools.h \\\n");
  fprintf(fp,"	${INCL}/swaplib.h \\\n");
  fprintf(fp,"	${INCL}/toolkit.h \\\n");
  fprintf(fp,"	${INCL}/almlib.h\n");
  if (!strcmp(os,"SUNOS"))
  {
    fprintf(fp,"\nCPP = CC -I${INCL}\n");
    fprintf(fp,"CC = CC -c -I${INCL}\n");
    fprintf(fp,"CCL = CC\n");
    fprintf(fp,"LM = \n");
  }
  else if (!strcmp(os,"ULTRIX"))
  {
    fprintf(fp,"\nCPP = g++ -I${INCL}\n");
    fprintf(fp,"CC = gcc -c -g -I${INCL}\n");
    fprintf(fp,"CCL = gcc -g\n");
    fprintf(fp,"LM = \n");
  }
  else if (!strcmp(os,"HP"))
  {
    fprintf(fp,"\nCPP = g++ -I${INCL}\n");
    fprintf(fp,"CC = cc -c -Aa -I${INCL}\n");
    fprintf(fp,"CCL = cc\n");
    fprintf(fp,"LM = -lM\n");
  }
  else if (!strcmp(os,"SOLARIS"))
  {
    fprintf(fp,"\nCPP = CC -compat -I${INCL}\n");
    fprintf(fp,"CC = cc -c -I${INCL}\n");
    fprintf(fp,"CCL = cc\n");
    fprintf(fp,"LM = -lm -lsocket -lnsl -lrt\n");
  }
  else if (!strcmp(os,"MACOS"))
  {
    fprintf(fp,"CC = clang -m64 -fPIC -w -c -I${INCL}\n");
    fprintf(fp,"CCL = clang -m64 -fPIC -w -undefined suppress\n");
    fprintf(fp,"LM = -lm -lrt\n");
  }
  else
  {
    fprintf(fp,"\nCPP = g++ -I${INCL}\n");
    fprintf(fp,"CC = cc -c -g -I${INCL}\n");
    fprintf(fp,"CCL = cc -g\n");
    fprintf(fp,"LM = -lm -lrt\n");
  }

  modlist[0] = 0;
  for (el=eqplist; el != NULL; el=el->next) 
  {
    sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag);
    sprintf(&modlist[strlen(modlist)],"%s.o ",tag);
  }

  fprintf(fp,"\n${TINEEXE}: devsrv.o %s\n",modlist);
  fprintf(fp,"\t${CCL} -o ${TINEEXE} devsrv.o %s ${LIBS}/${TINELIB} ${LM}\n",modlist);
  fprintf(fp,"\tchmod a+s ${TINEEXE}\n");
  fprintf(fp,"\ndevsrv.o: devsrv.c ${HDRS}\n");
  fprintf(fp,"\t${CC} devsrv.c\n\n");

  for (el=eqplist; el != NULL; el=el->next) 
  {
    sprintf(tag,"%s",el->lclname); strlwr(tag); strudr(tag);
    fprintf(fp,"%s.o: %s.c %s.h ${HDRS}\n",tag,tag,tag);
    fprintf(fp,"\t${CC} %s.c\n\n",tag);
  }
  fprintf(fp,"clean:\n");
  fprintf(fp,"\trm *.o\n");

err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int main(int argc, char *argv[])
{
  EQP *el;
  int cc,len;
  char os[64],c;

  os[0] = 0;
  switch (argc)
  {           
    case 3:
	  strncpy(os,argv[2],64); os[63] = 0;
    case 2:
	  strncpy(targetPath,argv[1],128); targetPath[128] = 0;
	  if ((len=strlen(targetPath)) > 0)
	  {
		if (strchr(targetPath,'\\') != NULL) c = '\\'; else c = '/';
	    if (targetPath[len-1] != c) targetPath[len] = c; targetPath[len+1] = 0;
	  }
      break;
    default:
	  targetPath[0] = 0;
  }

  getFecInformation();

  if ((cc=getExportInformation()) != 0)
  {
	printf("error : %s\n",cc2str(cc));
	exit(1);
  }

  if (dumpExportsFiles) putExportInformation();

  if ((cc=putDevServerTemplate("devsrv")) != 0)
  {
	printf("devsrv error : %s\n",cc2str(cc));
	exit(1);
  }
  for (el=eqplist; el != NULL; el=el->next)  // go through the list
  {
     if ((cc=putEQPTemplate(el)) != 0)  // make a .c module for this one
	 {
		printf("%s file error : %s\n",el->lclname,cc2str(cc));
		exit(1);
	 }
  }
  putMakeFile(os);
}
