/*
 * File swaplib.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * \internal
 * \file swaplib.c
 * \brief Byte-Swapping and Structures
 * 
 * This module contains
 * <ul>
 * <li>functions for byte-swapping</li>
 * <li>functions which handle structure descriptions and tagged structures</li>
 * <li>an initializer for globally used structure definitions</li>
 * </ul>
 * 
 *    swaplib.c is also to be used when machines either do not use
 *    INTEL byte order or use INTEL byte order but have non-BYTE
 *    alignment (such as OSF)
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "prolog.h"
#include "srvcore.h"
#include "globals.h"
#include "swaplib.h"
#include "hstlib.h"
#include "ttycmd.h"

/*----------------------------------------------------------------------------*
 *  Switches between Intel and Motorola byte order conventions 15.09.93 swh 
 *  Special KLUGE for VMS - deliver Mot order for float,double 10.03.94 swh
 *  Let VMS compile using IEEE format pass thru SwapBytes      10.09.2006 pdd
 *----------------------------------------------------------------------------*/

void memrev(void *d,int len)
{
  register BYTE *p1, *p2;
  for (p1=(BYTE *)d, p2=p1+len-1; p2 > p1; ++p1, --p2)
  {
    *p1 ^= *p2;
    *p2 ^= *p1;
    *p1 ^= *p2;
  }
  return;
}

int _swapBytes(char *dat,int size,int fmt,int dir)
{ /* use MEMREV macro instead of memrev() here -> size could be large! */
  register int i;
  register BYTE *p1, *p2; /* for MEMREV */
  int hsiz,len;
# ifdef VMS
  int j, a = (dir==RTOL?0:1),b = (a+1)%2;
  char c;
# endif

  switch(fmt)
  {
    case CF_NULL: 
      return 0;
    case CF_TEXT:
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_BYTE:
    case CF_BITFIELD8:
    case CF_XML:
    case CF_STRING8:
    case CF_STRING16:
    case CF_STRING32:
    case CF_STRING48:
    case CF_STRUCT:
    case CF_HISTORY:
      return 0;
    case CF_BITFIELD16:
    case CF_INT16:
    case CF_UINT16:
#     if !defined(INTEL) && !defined(VMS)
      for (i=0; i<size*2; i+=2) MEMREV(&dat[i], 2);
#     endif
      return 0;
    case CF_BIT:
    case CF_BOOLEAN:
    case CF_BITFIELD32:
    case CF_INT32:
    case CF_UINT32:
#     ifdef VMS
      return 0;
#     endif
    case CF_FLOAT:
#     ifndef VMS
      for (i=0; i<size*4; i+=4) MEMREV(&dat[i],4);
#     else
      for (i=0; i<size; i++) if (((int *)dat)[i]) for (j=0; j<2; j++)
      {
        c = dat[i*4+j+2];
        dat[i*4+j+2] = dat[i*4+j] - a*j;
        dat[i*4+j] = c + b*j;
      }
#     endif
      return 0;
    case CF_INT64:
    case CF_DOUBLE:
#     ifndef VMS
      for (i=0; i<size*8; i+=8) MEMREV(&dat[i],8);
#     else
      for (i=0; i<size; i+=8) for (j=0; j<4; j++)
      {
        c= *(dat+i+8+j-2-(j/2));
        *(dat+i+8+j-2-(j/2))= *(dat+i+j);         /*not for VMS serv yet */
        *(dat+i+j)=c + b*(j ? 0 : 0x20);
      }
#     endif
      return 0;
#   ifndef VMS
    case CF_DBLTIME:
    case CF_DBLDBL:
      for (i=0; i<size*16; i+=16) {MEMREV(&dat[i],8); MEMREV(&dat[i+8], 8);}
      return 0;
    case CF_NAMEDBLDBL:
      for (i=0; i<size*32; i+=32) {MEMREV(&dat[16+i], 8); MEMREV(&dat[16+i+8], 8);}
      return 0;
    case CF_NAME32DBLDBL:
      for (i=0; i<size*48; i+=48) {MEMREV(&dat[32+i], 8); MEMREV(&dat[32+i+8], 8);}
      return 0;
    case CF_NAME64TIME:
    case CF_NAME64DBL:
      for (i=0; i<size*72; i+=72) {MEMREV(&dat[64+i], 8);}
      return 0;
    case CF_NAME64DBLDBL:
      for (i=0; i<size*80; i+=80) {MEMREV(&dat[64+i], 8); MEMREV(&dat[64+i+8], 8);}
      return 0;
    case CF_NAME64DBLDBLDBL:
      for (i=0; i<size*88; i+=88) {MEMREV(&dat[64+i], 8); MEMREV(&dat[64+i+8], 8); MEMREV(&dat[64+i+16], 8);}
      return 0;
    case CF_NAME16FI:
    case CF_NAME16II:
      for (i=0; i<size*24; i+=24) {MEMREV(&dat[i+16],4); MEMREV(&dat[i+20],4);}
      return 0;
    case CF_NAME8I:
      for (i=0; i<size*12; i+=12) {MEMREV(&dat[i+8],4);}
      return 0;
    case CF_NAME16I:
      for (i=0; i<size*20; i+=20) {MEMREV(&dat[i+16],4);}
      return 0;
    case CF_NAME32I:
      for (i=0; i<size*36; i+=36) {MEMREV(&dat[i+32],4);}
      return 0;
    case CF_NAME48I:
      for (i=0; i<size*52; i+=52) {MEMREV(&dat[i+48],4);}
      return 0;
    case CF_NAME64I:
      for (i=0; i<size*68; i+=68) {MEMREV(&dat[i+64],4);}
      return 0;
    case CF_FF:
    case CF_FI:
    case CF_II:
      for (i=0; i<size*8; i+=8) {MEMREV(&dat[i],4); MEMREV(&dat[i+4],4);}
      return 0;
    case CF_III:
    case CF_FII:
    case CF_FFI:
    case CF_TDS:
      for (i=0; i<size*12; i+=12) {MEMREV(&dat[i],4); MEMREV(&dat[i+4],4); MEMREV(&dat[i+8],4);}
      return 0;
    case CF_DBLDBLDBL:
      for (i=0; i<size*24; i+=24) {MEMREV(&dat[i], 8); MEMREV(&dat[i+8], 8); MEMREV(&dat[i+16], 8);}
      return 0;
    case CF_USTRING:
      for (i=0; i<size*96; i+=96)
      {
        MEMREV(&dat[i],4); MEMREV(&dat[i+4],4); MEMREV(&dat[i+8],4); MEMREV(&dat[i+12],4);
      }
      return 0;
    case CF_UNAME:
      for (i=0; i<size*32; i+=32)
      {
        MEMREV(&dat[i],4); MEMREV(&dat[i+4],4); MEMREV(&dat[i+8],4); MEMREV(&dat[i+12],4);
      }
      return 0;
    case CF_TTII:
    case CF_IIII:
    case CF_IIFF:
    case CF_FIFI:
    case CF_IFFF:
      for (i=0; i<size*16; i+=16)
      {
        MEMREV(&dat[i],4); MEMREV(&dat[i+4],4); MEMREV(&dat[i+8],4); MEMREV(&dat[i+12],4);
      }
      return 0;
    case CF_ASPECTRUM:
    {
      hdrSPECTRUM shs;
      hsiz = SPECTRUM_HDR_SIZE + 2 * sizeof(SINT32);
      for (i = 0; i < size; i++)
      {
        memcpy(&shs, &dat[i*hsiz], sizeof(hdrSPECTRUM));
        _structSwap(SpectrumHeaderStruct, (BYTE *)dat, i*hsiz, NULL, (BYTE *)&shs, dir);
        MEMREV(&dat[i*hsiz + SPECTRUM_HDR_SIZE], 4); /* offset */
        MEMREV(&dat[i*hsiz + SPECTRUM_HDR_SIZE + 4], 4); /* size */
      }
      return 0;
    }
    case CF_SPECTRUM: 
    { /* 'size' is kluged to pretend that the header is 24 floats up front (and we skip over these) */
      hdrSPECTRUM shs;
      memcpy(&shs, dat, sizeof(hdrSPECTRUM));
      _structSwap(SpectrumHeaderStruct, (BYTE *)dat, 0, NULL, (BYTE *)&shs, dir);
      for (i = SPECTRUM_HDR_SIZE; i < SPECTRUM_HDR_SIZE + size * 4; i += 4) MEMREV(&dat[i], 4);
      return 0;
    }
    case CF_GSPECTRUM:
    {  /* size parameter is in bytes in this case */
      hdrGSPECTRUM shs;
      memcpy(&shs, dat, sizeof(hdrGSPECTRUM));
      _structSwap(GSpectrumHeaderStruct, (BYTE *)dat, 0, NULL, (BYTE *)&shs, dir);
      for (i = GSPECTRUM_HDR_SIZE; i < size; i += 4) MEMREV(&dat[i], 4);
      return 0;
    }
    case CF_AIMAGE:
    {
      hdrIMAGE ihs;
      hsiz = IMAGE_HDR_SIZE + 2 * sizeof(SINT32);
      for (i = 0; i < size; i++)
      {
        memcpy(&ihs, &dat[i*hsiz], sizeof(hdrIMAGE));
        _structSwap(ImageHeaderStruct, (BYTE *)dat, i*hsiz, NULL, (BYTE *)&ihs, dir);
        MEMREV(&dat[i*hsiz + IMAGE_HDR_SIZE], 4); MEMREV(&dat[i*hsiz + IMAGE_HDR_SIZE + 4], 4);
      }
      return 0;
    }
    case CF_IMAGE: /* does not use the 'size' parameter at all */
    {
      hdrIMAGE ihs;
      memcpy(&ihs, dat, sizeof(hdrIMAGE));
      _structSwap(ImageHeaderStruct, (BYTE *)dat, 0, NULL, (BYTE *)&ihs, dir);
      return 0;
    }
    case CF_MDA: /* does not use the 'size' parameter at all */
    {
      DMDA mda;
      memcpy(&mda, dat, MDA_HDR_SIZE); /* a redundant set of the same bytes (needed if dir == LTOR) */
      _structSwap(MDAHeaderStruct, (BYTE *)dat, 0, NULL, (BYTE *)&mda, dir);
      for (i = 0, len = 1; i < mda.ndim; i++) if (mda.axis[i].len > 0) len *= mda.axis[i].len;
      SwapBytes((char *)&dat[MDA_HDR_SIZE], len, mda.fmt, dir);
      return 0;
    }
#   endif /* !defined VMS */
    default: if (tineDebug > 1 || errorDebug) dbglog("_swapBytes : format %d is unknown",fmt);
  }
  return illegal_format;
}
#if !defined(INTEL) && !defined(NONALIGNED) && !defined(VMS_IEEE)
int cfHistorySwap(char *tag,BYTE *src,BYTE *dst,int dir,int siz)
{
  register BYTE *p1, *p2;
  int fmt = ftoi(tag), dsiz, i, num;
  struct structStruct *s;
  if ((dsiz=getCarriedFormatSize(tag)) <= 0) return illegal_format;
  dsiz += HISTORY_HDR_SIZE;
  num = siz / dsiz;
  if (dst != src) memcpy(dst,src,siz);
  switch (fmt)
  {
    case CF_NULL:
      if ((s=findstruct(tag)) == NULL) return invalid_structure_tag;
      for (i=0; i<num; i++)
      { /* note this assumes a proper byte alignment : need to enforce this ! */
        MEMREV(&dst[i*dsiz],8); MEMREV(&dst[8+i*dsiz],4); MEMREV(&dst[12+i*dsiz],4);
        _taggedStructSwap(tag,NULL,src,dst,dir,1,s->size);
      }
      return 0;
    case CF_STRUCT:
    case CF_HISTORY:
      break;
    case CF_IMAGE:
    case CF_STRING:
    case CF_KEYVALUE:
      num = 1; /* swapped one entry at a time */
    default:
      for (i=0; i<num; i++)
      {
        MEMREV(&dst[i*dsiz],8); MEMREV(&dst[8+i*dsiz],4); MEMREV(&dst[12+i*dsiz],4);
        SwapBytes((char *)&dst[16+i*dsiz],1,fmt,dir);
      }
      return 0;
  }
  return illegal_format;
}
#else
int cfHistorySwap(char *tag,BYTE *src,BYTE *dst,int dir,int siz)
{
  UNUSED(dir);
  switch (ftoi(tag))
  {
    case CF_NULL:
      if (findstruct(tag) == NULL) return invalid_structure_tag;
      break;
    case CF_STRUCT:
    case CF_HISTORY:
      return illegal_format;
    default:
      break;
  }
  if (dst != src) memcpy(dst,src,siz);
  return 0;
}
#endif  /* !INTEL && !NONALIGNED && !VMS_IEEE */

static int gControlStructInitialized = FALSE;
struct structStruct *structList = NULL;

int isInvalidStructureTag(char *tag)
{
  int i, len = 0;
  if (tag == NULL) return TRUE;
  if ((len=(int)strlen(tag)) == 0) return TRUE;
  if (len > TAG_NAME_SIZE) len = TAG_NAME_SIZE;
  if (ftoi(tag) != CF_NULL) return TRUE;
  if (strcmp(tag,STRUCT_LIST_MARKER) == 0) return TRUE;
  for (i = 0; i < len; i++) if ((BYTE)tag[i] > 127) return TRUE;
  return FALSE;
}

int containsStructKey(struct srvKey *lst,char *key)
{
  struct srvKey *k;
  if (lst == NULL || key == NULL) return TRUE;
  for (k=lst; k != NULL; k=k->nxt)
    if (strnicmp(k->key,key,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+2) == 0) return TRUE;
  return FALSE;
}
/**
 * \internal
 * \brief Finds a tagged structure.
 */
struct structStruct *findstructEx(char *tag, char *key, ...)
{
  struct structStruct *s=NULL, *sstart=structList;
  char *srv=NULL, keystr[128];
  int cc=0;
  va_list argp;

  if (tag == NULL || strlen(tag) == 0) return NULL;
  va_start(argp,key);
  if (key != NULL && strlen(key) > 0 && key[0] != '/')
  { /* args using: context and server separately OR marker + next */
    if (strcmp(key,STRUCT_LIST_MARKER) == 0)
    {
      if ((sstart=va_arg(argp,struct structStruct *)) == NULL) ccerr(0);
      key = NULL;
    }
    else if ((srv=va_arg(argp,char *)) != NULL)
    { /* context and server given individually */
      sprintf(keystr,"/%.32s/%.32s",key,srv);
      key = keystr;
    }
  }
  for (s=sstart; s != NULL; s=s->next)
  {
    if (strncmp(s->name,tag,TAG_NAME_SIZE)) continue;
    if (s->key != NULL && key != NULL && strlen(key) > 0 && !containsStructKey(s->key,key)) continue;
    break;
  }
  if (s != NULL && s->rawlen == 0)
  {
    if (tineDebug || errorDebug) dbglog("findstruct : struct %.16s not properly terminated",tag);
    s = NULL;
    ccerr(invalid_structure_size);
  }
err:
  va_end(argp);
  if (cc != 0) dbglog("findstructEx: %.32s",cc2str(cc));
  return s;
}
struct structStruct *findstruct(char *tag)
{
  return findstructEx(tag,NULL);
}
TINE_EXPORT struct structStruct* FindStruct(char* tag) {
	return findstruct(tag);
}
int getFieldFormat(struct structStruct *s,char *field)
{
  struct structFormat *sf;
  if (s == NULL || field == NULL) return CF_NULL;
  for (sf=s->s; sf != NULL; sf=sf->nxt)
  {
    if (!strnicmp(sf->field,field,FIELD_NAME_SIZE)) return sf->fmt;
  }
  return CF_NULL;
}
int getFieldSize(struct structStruct *s,char *field)
{
  struct structFormat *sf;
  if (s == NULL || field == NULL) return 0;
  for (sf=s->s; sf != NULL; sf=sf->nxt)
  {
    if (!strnicmp(sf->field,field,FIELD_NAME_SIZE)) return sf->siz;
  }
  return 0;
}
/**
 * \internal
 * \brief Finds or allocates a tagged structure.
 */
struct structStruct *getTaggedStruct(char *tag)
{
  struct structStruct *s;

  if ((s=findstruct(tag)) != NULL) return s; /* found it */
  if ((s=(struct structStruct *)SystemCalloc(1,sizeof(struct structStruct))) != NULL)
  { /* didn't find it -> need to allocated it */
    strncpy(s->name,tag,TAG_NAME_SIZE);
    s->rawlen = 0;
    s->next = structList;
    structList = s;
    if (tineDebug) dbglog("getTaggedStruct : New structure tag registered: %s",tag);
  }
  return s;
}
static struct structStruct *structHdrMark = NULL;
void markStructRegistry(void)
{
  structHdrMark = structList;
}
int queryStructFormat(char *tag,BYTE *sfq,int fmt,int *sfqsiz)
{
  int i,n,fsz,sz=0,nq,cc=0;
  char field[64];
  struct structStruct *s, *ss;
  struct structFormat *sf;
  NAME64DBLDBL *ndd = (NAME64DBLDBL *)sfq;
  NAME16II *nii = (NAME16II *)sfq;
  INTINT *ii = (INTINT *)sfq;
  
  if (sfqsiz == NULL) ccerr(argument_list_error);
  if ((nq = *sfqsiz) < 1) ccerr(dimension_error);
  if ((s=findstruct(tag)) == NULL) ccerr(invalid_structure_tag);
  for (sf=s->s,n=0; sf != NULL; sf=sf->nxt,n++); /* count them first (reverse ordered!) */
  for (i=0,sf=s->s; i<nq && i<n && sf != NULL; i++, sf=sf->nxt)
  {
    memset(field,0,64);
    if (sf->fmt == CF_STRUCT)
    {
      if ((ss=findstruct(sf->stag)) == NULL) return invalid_structure_tag;
      fsz = ss->rawlen;
      sprintf(field,"<%.16s>%.16s",sf->stag,sf->field);
    }
    else
    {
      switch (sf->fmt)
      {
        case CF_STRING:
          fsz = sizeof(size_t);
          break;
        default:
          fsz = getFormatTypeSize(sf->fmt);
          break;
      }
      strncpy(field,sf->field,16);
    }
    sz += sf->siz * fsz;
    switch (fmt)
    {
      case CF_INTINT:
        ii[n-i-1].i1val = sf->siz;
        ii[n-i-1].i2val = sf->fmt;
        break;
        break;
      case CF_NAME16II:
        strncpy(nii[n-i-1].name,field,16);
        nii[n-i-1].i1val = sf->siz;
        nii[n-i-1].i2val = sf->fmt;
        break;
      case CF_NAME64DBLDBL:
        strncpy(ndd[n-i-1].name,field,64);
        ndd[n-i-1].d1val = (double)sf->siz;
        ndd[n-i-1].d2val = (double)sf->fmt;
        break;
      default: return illegal_format;
    }
  }
  if (i == nq) i--;
  if (i < 0) ccerr(code_failure);
  switch (fmt)
  {
    case CF_INTINT:
      ii[i].i1val = sz;
      ii[i].i2val = CF_NULL;
      break;
      break;
    case CF_NAME16II:
      strncpy(nii[i].name,tag,16);
      nii[i].i1val = sz;
      nii[i].i2val = CF_NULL;
      break;
    case CF_NAME64DBLDBL:
      strncpy(ndd[i].name,tag,16);
      ndd[i].d1val = (double)sz;
      ndd[i].d2val = (double)CF_NULL;
      break;
    default: 
      return illegal_format;
  }
  *sfqsiz = n+1;
err:
  return cc;
}
TINE_EXPORT int GetStructFormat(char *tag,NAME64DBLDBL *sfn64, int *n)
{
  return queryStructFormat(tag,(BYTE *)sfn64,CF_NAME64DBLDBL,n);
}
int isAllowedStructFormat(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_DEFAULT:
    case CF_HISTORY:
      return FALSE;
    default:
      break;
  }
  return TRUE;
}
/**
 * \brief Prepend a structFormat to a structFomat list.
 * 
 * This method creates a new structFormat and sets it in
 * the first element of a linked list of structFormats.
 * A structFormat may describe a scalar or an array.
 * 
 * 
 * \param target void pointer which is interpreted as a double pointer to a structFormat. Insert the new struct format before this.
 * \param addr Address offset of this field in the structure. 
 * \param size Array length if the described field. (not the size in bytes).
 * \param fmt TINE CF_ format of the new struct element
 * \param subtag is a possible structure tag for the case fmt == CF_STRUCT
 * \param field is the corresponding registered field name
 * 
 * \todo target should be structFormat double pointer.
 */ 
int addtostruct(void *target,int addr,int size, int fmt, char *subtag, char *field)
{
  struct structStruct *ss;
  struct structFormat *tmp,**s = (struct structFormat **)target;
  int offset = (*s) ? (*s)->off + (*s)->siz*getFormatSize((*s)->fmt) + getFormatHeaderSize((*s)->fmt) : 0;

  if (!isAllowedStructFormat(fmt)) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "struct format %s is not allowed",itof(fmt));
    return illegal_format;
  }
  if ((*s) != NULL && (*s)->fmt == CF_STRUCT)
  { /* embedded struct -> find the proper offset */
    if ((ss=findstruct((*s)->stag)) == NULL) return illegal_format;
    offset = (*s)->off + (*s)->siz*ss->rawlen;
  }
  if ((*s) != NULL && isVariableLengthFormat((short)(*s)->fmt))
  { /* pointer(s) in structure => include the offset and length in extended space  */
    offset = (*s)->off + (*s)->siz*(getFormatHeaderSize((*s)->fmt) + 2*fmtsizeof(CF_INT32));
  }
  tmp = (struct structFormat *)SystemCalloc(1,sizeof(struct structFormat));
  if (tmp == NULL) return out_of_local_memory;
  tmp->off = offset;
  tmp->fmt = fmt;
  tmp->siz = size;
  tmp->addr = addr;
  if (subtag != NULL) strncpy(tmp->stag, subtag, FIELD_NAME_SIZE);
  if (strlen(field) > FIELD_NAME_SIZE)
    feclogEx(TINE_LOGLEVEL_WARN, "struct %.16s field %s truncated to %d characters",tmp->stag,field,FIELD_NAME_SIZE);
  strncpy(tmp->field,field,FIELD_NAME_SIZE);
  /* Add tmp to list */
  tmp->nxt = *s;  
  *s = tmp;
  return 0;
}
int addKeyToStruct(struct structStruct *s,char *key)
{
  struct srvKey *pkey = NULL;
  if (s == NULL || key == NULL) return invalid_parameter;
  if (s->key != NULL && containsStructKey(s->key,key)) return already_assigned;
  if ((pkey=(struct srvKey *)SystemCalloc(1,sizeof(struct srvKey))) == NULL) return out_of_local_memory;
  strncpy(pkey->key,key,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+2);
  pkey->nxt = s->key;
  s->key = pkey;
  return 0;
}
/**
 * \internal
 * \brief Frees a structFormat list
 * \param sf Pointer to the first list element
 */
void freeStruct(struct structFormat *sf)
{
  struct structFormat *tmp,*tmp2;
  for (tmp=sf; tmp != NULL; tmp=tmp2)
  {
    tmp2 = tmp->nxt;
    SystemFree(tmp); 
  } 
}
void freeStructKeys(struct srvKey *sk)
{
  struct srvKey *k, *k2;
  for (k=sk; k != NULL; k=k2)
  {
    k2 = k->nxt;
    SystemFree(k);
  }
}
void freeStructList(int freeMarkedList)
{
  struct structStruct *s=structList,*snxt;
  while (s != NULL)
  {
    if (!freeMarkedList && s == structHdrMark) break;
#   ifndef DOS
    SystemFree(s->sdata);
#   endif
    freeStruct(s->s);
    freeStructKeys(s->key);
    snxt = s->next;
    SystemFree(s);
    s = snxt;
  }
  structList = s;
}
TINE_EXPORT void flushStructCache(void) { freeStructList(FALSE); }
/**
 * \internal
 * call with tags = NULL to get the number of structs
 */
int getRegisteredStructNames(char *tags,int *ntags,int fmt,int includeKeys,int includeSystem)
{
  struct structStruct *s=NULL, *sstop;
  int n=0,flen=0;
  if (ntags == NULL) return argument_list_error;
  if (tags != NULL)
  { /* not just a number query */
    if (!isSimpleStringFormat((short)fmt) || fmt == CF_TEXT) return illegal_format;
    flen = fmtsizeof(fmt);
    memset(tags,0,(*ntags)*flen);
  }
  sstop = includeSystem ? NULL : structHdrMark;
  for (s=structList; s != NULL; s=s->next)
  {
    if (s == sstop) break;
    if (tags != NULL)
    { /* fill in the struct tag names */
      if (n >= *ntags) break;
      if (includeKeys && s->key != NULL) strncpy(tags+n*flen,s->key->key,47);
      strncat(tags+n*flen,s->name,16);
    }
    n++;
  }
  *ntags = n;
  return 0;
}
int dumpRegisteredStrucures(void)
{
  struct structStruct *s=NULL, *sstart=structList;
  int isSysStruct= FALSE;
  char *tag = "(local)";
  ttyoutput("structure registry : user-defined structures");
  for (s=sstart; s != NULL; s=s->next)
  {
    if (s == structHdrMark)
    {
      ttyoutput("structure registry : system-defined structures");
      isSysStruct = TRUE;
      tag = "(system)";
    }
    if (!isSysStruct) tag = s->key == NULL ? "(local)" : s->key->key;
    ttyoutput("  %s : %16s",tag,s->name);
  }
  return 0;
}
/**
 * \deprecated
 * \brief Adds a field description to a tagged structure.
 * 
 * This function has two meanings:
 * <ol>
 * <li>Adds a field to a tagged structure</li>
 * <li>Sets the size and the array size of a tagged structure</li>
 * </ol>
 * 
 * \param tag Tag name
 * \param addr <ol><li>Address offset of this field in the structure.</li>
 *  <li>Size of the element in bytes</li></ol>
 * \param size <ol><li>Array length if the described field. (not the size in bytes).</li>
 *  <li>Number of elements</li></ol>
 * \param fmt TINE CF_ format of the new struct element
 * 
 * 
 * \return 0 or out_of_local_memory if no struct could be found/allocated for tag.
 * 
 * \see StructStruct
 */ 
TINE_EXPORT int fillstruct(char *tag,int addr,int size,int fmt)
{
  struct structStruct *s;

  if (!gControlStructInitialized)
  { /* if this legacy call is used, make sure the control structs are registered properly */
    initControlStructs();
  }
  if (isInvalidStructureTag(tag)) return invalid_structure_tag;
  if ((s=getTaggedStruct(tag)) == NULL) return out_of_local_memory;
  if (s->sealed) return struct_sealed;
  if (fmt == CF_NULL)
  {
    s->num = size;
    s->size = addr;
#   ifndef DOS
    if ((s->sdata=(BYTE *)SystemCalloc(s->num,s->size)) == NULL) return out_of_local_memory;
#   ifdef INTEL
    if (s->rawlen == 0) s->rawlen = s->size; /* a quick and dirty register from an intel machine */
#   endif
#   else /* is DOS (CF_STRING not allowed ?) */
    s->rawlen = s->size;
#   endif
    s->sealed = TRUE;
    return 0;
  }
  if (!s->sealed) s->rawlen += size*getFormatSize(fmt);
  return addtostruct(&s->s,addr,size,fmt,NULL,"");
}
TINE_EXPORT int GetStructSize(char *tag)
{
  struct structStruct *s=findstruct(tag);
  return s == NULL ? 0 : s->rawlen;
}
TINE_EXPORT int StructIsBlittable(char *tag)
{
  struct structStruct *s=findstruct(tag);
  return s == NULL ? FALSE : !s->hasExtSpace;
}
TINE_EXPORT struct structStruct *GetStructFromTag(char *tag)
{
  return findstruct(tag);
}
TINE_EXPORT struct structStruct *GetStructFromTagEx(char *tag,char *key)
{
  if (key == NULL || key[0] == '/') return findstructEx(tag,key);
  return findstruct(tag);
}
TINE_EXPORT int GetStructStatus(char *tag)
{
  struct structStruct *s = findstruct(tag);
  if (s == NULL) return not_registered;
  if (!s->sealed) return not_initialized;
  return 0;
}
/**
 * \internal
 * \return the number of bytes swapped
 */
int _structSwap(struct structFormat *s,BYTE *stream,int offset,int *tail,BYTE *struc,int dir)
{
  int n,k,dsiz=0,hsiz,hoff,tsiz,extfmt,extoff,extsiz, *sptr;
  struct structStruct *ss;
  BYTE *extptr, *hptr, **xptr;
  BYTE *dat = &stream[offset];

  if (dir == RTOL)
  { /* in-coming (stream = src; struc = dst) */
    for (n=0; s != NULL && s->siz; n += dsiz)
    {
      if (s->fmt == CF_STRUCT)
      { /* nested struct swap */
        if ((ss=findstruct(s->stag)) == NULL) return invalid_structure_tag;
        for (k=0; k<s->siz; k++)
        {
          /* _structSwap(ss->s,&dat[s->off+k*ss->rawlen],&struc[s->addr+k*ss->size],RTOL); */
          _structSwap(ss->s,stream,offset+s->off+k*ss->rawlen,tail,&struc[s->addr+k*ss->size],RTOL);
        }
        dsiz = s->siz * ss->rawlen;
      }
      else if (s->fmt == CF_BOOLEAN && gMarshallBooleanAsByte)
      {
        for (k = 0; k < s->siz; k++) struc[s->addr + k] = (_Bool)((SINT32 *)&dat[s->off])[k];
      }
      else if (!isVariableLengthFormat((short)s->fmt))
      { /* 'normal' format type : joy! */
        SwapBytes((char *)&dat[s->off],s->siz,s->fmt,RTOL);
        dsiz = getFormatHeaderSize(s->fmt) + s->siz*getFormatSize(s->fmt);
        memcpy(&struc[s->addr],&dat[s->off],dsiz);
      }
      else
      { /* variable length format type */
        dsiz = 0;
        hoff = s->off;
        hsiz = getFormatHeaderSize(s->fmt);
        for (k=0; k<s->siz; k++)
        {
          if (hsiz > 0)
          { /* first the header (if any) */
            SwapBytes((char *)&dat[hoff],1,s->fmt,RTOL);
            memcpy(&struc[s->addr+k*hsiz],&dat[hoff],hsiz);
            hoff += hsiz;
          }
          else
          { /* CF_STRING: swap the reference to the variable part */
            SwapBytes((char *)&dat[hoff],2,CF_INT32,RTOL);
          }
          memcpy(&extoff,&dat[hoff],4); hoff += 4;
          memcpy(&extsiz,&dat[hoff],4); hoff += 4;
          extfmt = getFormatDataType(s->fmt);
          if (extfmt == CF_NULL) dbglog("_structSwap: trapped a NULL format in structure"); /* avoid a warning */
          /* swap the extended bytes to proper byte order */
          SwapBytes((char *)&stream[extoff],extsiz,extfmt,RTOL);
          /* now do the assignments */
          extptr = (BYTE *)&stream[extoff]; 
          dsiz += hsiz + 2*sizeof(SINT32) + extsiz;
          if (hsiz > 0) 
          { /* CF_IMAGE, CF_AIMAGE, CF_ASPECTRUM */
            tsiz = GetFormatTypeSize(s->fmt);
            xptr = (BYTE **)GetFormatBodyPtrAddr(&struc[s->addr+k*tsiz],s->fmt);
            sptr = (int *)GetFormatBodyBufferSizeAddr(&struc[s->addr+k*tsiz],s->fmt);
            if (xptr == NULL || sptr == NULL) continue;
            if (isAdjustableLengthFormat((short)s->fmt))
            {
              *xptr = (BYTE *)(size_t)extptr;
              *sptr = extsiz;
            }
            else
            {
              if (extsiz > *sptr) extsiz = *sptr; /* truncate to fit */
              memcpy(xptr,extptr,extsiz);
            }
          }
          else 
          { /* CF_STRING */
            ((size_t *)&struc[s->addr])[k] = (size_t)extptr;
          }
        }
      }
      s = s->nxt;
    }
  }
  else
  { /* out-going (struc = src; stream = dst) */
    /* point to 'end' of raw structure length, in case we need it */
    extoff = tail != NULL ? *tail : 0; 
    for (n=0; s != NULL && s->siz; n += dsiz)
    {
      if (s->fmt == CF_STRUCT)
      { /* nested struct swap */
        if ((ss=findstruct(s->stag)) == NULL) return invalid_structure_tag;
        for (k=0; k<s->siz; k++)
        {
          _structSwap(ss->s,stream,offset+s->off+k*ss->rawlen,tail,&struc[s->addr+k*ss->size],LTOR);
        }
        dsiz = s->siz * ss->rawlen;
      }
      else if (s->fmt == CF_BOOLEAN && gMarshallBooleanAsByte)
      {
        for (k = 0; k < s->siz; k++) ((SINT32*)&dat[s->off])[k] = (SINT32)struc[s->addr + k];
      }
      else if (!isVariableLengthFormat((short)s->fmt))
      { /* 'normal' format type */
        dsiz = getFormatHeaderSize(s->fmt) + s->siz*getFormatSize(s->fmt);
        memcpy(&dat[s->off],&struc[s->addr],dsiz);
        SwapBytes((char *)&dat[s->off],s->siz,s->fmt,LTOR);
      }
      else if (s->owner != NULL && s->owner->hasExtSpace)
      { /* variable length format type */
        /* s->addr will point to the proper location of the real struct contents! */
        extsiz = 0;
        hoff = s->off;
        hsiz = getFormatHeaderSize(s->fmt);
        dsiz = 0;
        for (k=0; k<s->siz; k++)
        {
          if (hsiz > 0) 
          { /* CF_AIMAGE, CF_ASPECTRUM */
            hptr = (BYTE *)&struc[s->addr+k*getFormatTypeSize(s->fmt)];
            extptr = getFormatBodyPtr(hptr,s->fmt);
            extsiz = getFormatBodyBufferSize(hptr,s->fmt)*fmtsizeof(s->fmt);
            memcpy(&dat[hoff],&struc[s->addr],hsiz);
            memcpy(&dat[hoff+hsiz],&extoff,4);
            memcpy(&dat[hoff+hsiz+4],&extsiz,4);
            SwapBytes((char *)&dat[hoff],1,s->fmt,LTOR);
            dsiz += hsiz;
            hoff += hsiz;
          }
          else 
          { /* CF_STRING */
            extptr = (BYTE *)((size_t *)&struc[s->addr])[k];
            extsiz = (int)strlen((char *)extptr)+1;
            memcpy(&dat[hoff],&extoff,4);
            memcpy(&dat[hoff+4],&extsiz,4);
            SwapBytes((char *)&dat[hoff],2,CF_INT32,LTOR);
          }
          hoff += 2*sizeof(SINT32);
          memcpy(&stream[extoff],extptr,extsiz);
          extfmt = getFormatDataType(s->fmt);
          SwapBytes((char *)&stream[extoff],extsiz,extfmt,LTOR);
          dsiz += 2*sizeof(SINT32) + extsiz;
          extoff += extsiz;
        }
      }
      s = s->nxt;
    }
    if (tail != NULL) *tail = extoff;
  }
  return n;  /* number of bytes swapped */
}
/**
 * \internal used for swapping registered (user defined) tagged structures
 * \return the number of structs swapped
 */
int _taggedStructSwap(char *tag,char *key,BYTE *dat,BYTE *struc,int dir,int num,int structsiz)
{
  int i,nb=0,tail=0,nbchk=FALSE,maxnum;
  struct structStruct *ss=findstructEx(tag,key);
# ifdef INTEL
  int domemcpy;
# endif

  if (ss == NULL || ss->rawlen == 0) return -un_allocated;
  if (num <1 || structsiz < 1) return -out_of_range;
# ifdef INTEL
  domemcpy = ss->s == NULL ? TRUE : FALSE;
# ifdef MSDOS
  domemcpy = TRUE;
# endif
  if (domemcpy)
  { /* tag registered without fields (not a good thing to do, but it should still work) */
    if (dir == RTOL) memcpy(struc,dat,ss->rawlen*num);
    else memcpy(dat,struc,ss->rawlen*num);
    return num;
  }
# endif
  switch (dir)
  {
    case RTOL: nbchk = TRUE; break;
    case LTOL: dir = RTOL; break; /* special comparison circumstances */
  }
  if (dir == LTOR) tail = ss->hasExtSpace ? num*ss->rawlen : 0;
  maxnum = num;
  if (ss->hasExtSpace && ss->num < num) maxnum = ss->num;
  for (i=0; i<maxnum; i++)
  {
    nb += _structSwap(ss->s,dat,i*ss->rawlen,&tail,&struc[i*structsiz],dir);
    if (nbchk && nb >= num*ss->rawlen) break;
  }
  return i+1;
}
/**
 * \internal used with kernel system registered structures (for stock properties)
 * \return the number of structs swapped
 */
int _systemStructSwap(struct structFormat *sf,BYTE *dat,BYTE *struc,int dir,int num,int structsiz)
{
  int len = 0, fsiz, rawlen = num*structsiz, maxnum;
  struct structFormat *s = sf;
  struct structStruct *ss = NULL;
  int i,nb,tail=0,nbchk=FALSE;
# ifdef INTEL
  int domemcpy = (s == NULL) ? TRUE : FALSE;
# endif

  if (num < 1 || structsiz < 1) return -out_of_range;
  while (s != NULL && s->siz)
  { /* run thru the fields */
    fsiz = getFormatSize(s->fmt);
    if (s->fmt == CF_STRUCT)
    { /* embedded struct */
      if ((ss=findstruct(s->stag)) == NULL) return invalid_structure_tag;
      fsiz = ss->rawlen;
    }
    len += s->siz*fsiz;
    s = s->nxt;
  }
# ifdef INTEL
# ifdef MSDOS
  domemcpy = TRUE;
# endif
  if (domemcpy)
  { /* tag registered without fields (not a good thing to do, but it should still work) */
    if (dir == RTOL) memcpy(struc,dat,len*num);
    else memcpy(dat,struc,len*num);
    return num;
  }
# endif
  switch (dir)
  {
    case RTOL: nbchk = TRUE; break;
    case LTOL: dir = RTOL; break; /* special comparison circumstances */
  }
  if (sf != NULL)
  {
    s = sf;
    ss = sf->owner;
    rawlen = ss == NULL ? num*len : ss->rawlen;
    if (dir == LTOR && ss != NULL) tail = ss->hasExtSpace ? num*rawlen : 0;
  }
  if (ss == NULL) return 0;
  maxnum = num;
  if (ss->hasExtSpace && ss->num < num) maxnum = ss->num;
  for (i=0,nb=0; i<maxnum; i++)
  {
    nb += _structSwap(sf,dat,i*rawlen,&tail,&struc[i*structsiz],dir);
    if (nbchk && nb >= num*rawlen) break;
  }
  return i+1;
}

/**
* \internal
* \brief Structure description for legacy PktHdr
* \see PktHdr
*/
struct structFormat *PktHdrStructStruct = NULL;

/**
* \internal
* \brief Structure description for ReqHdr
* \see ReqHdr
*/
struct structFormat *ReqHdrStructStruct = NULL;

/**
* \internal
* \brief Structure description for RspHdr
* \see RspHdr
*/
struct structFormat *RspHdrStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for CONTRACT
 * \see CONTRACT
 */
struct structFormat *CONTRACTStruct = NULL;
struct structFormat *CONTRACT5Struct = NULL;

/**
 * \internal
 * \brief Structure description for legacy PrdrHdr
 * \see PrdrHdr
 */
struct structFormat *PrdrHdrStruct = NULL;
struct structFormat *PrdrHdr5Struct = NULL;

/**
* \internal
* \brief Structure description for SubRspHdr
* \see SubRspHdr
*/
struct structFormat *SubRspHdrStruct = NULL;

/**
* \internal
* \brief Structure description for SubReqHdr
* \see SubReqHdr
*/
struct structFormat *SubReqHdrStruct = NULL;

/**
 * \internal
 * \brief legacy structure descriptions for SubInfoPkt and SubInfoPkt5
 * \see SubInfoPkt
 */
struct structFormat *SubInfoPktStruct = NULL;
struct structFormat *SubInfoPkt5Struct = NULL;

/**
 * \internal
 * \brief Structure description for RPCDataStruct
 * \see RPCDataStruct
 */
struct structFormat *RPCDataStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for RPCFecStruct
 * \see RPCFecStruct
 */
struct structFormat *RPCFecStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for AMSStruct
 * \see AMSStruct
 */
struct structFormat *AMSStructStruct = NULL;

/**
 * \internal
 * \brief Legacy Structure description for AMSStruct
 * \see AMSStruct
 */
struct structFormat *AMS5StructStruct = NULL;

/**
 * \internal
 * \brief Structure description for ADSStruct
 * \see ADSStruct
 */
struct structFormat *ADSStructStruct = NULL;

/**
 * \internal
 * \brief Legacy Structure description for ADSStruct
 * \see ADSStruct
 */
struct structFormat *ADS5StructStruct = NULL;

/**
 * \internal
 * \brief Structure description for AlarmWatchStruct
 * \see AlarmWatchStruct
 */
struct structFormat *AWSStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for AlarmWatchStruct
 * \see AlarmWatchStruct
 */
struct structFormat *AWLQSStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for PropertyQueryStruct
 * \see PropertyQueryStruct
 */
struct structFormat *PropertyQueryStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for PropertyQueryStructEx
 * \see PropertyQueryStructEx
 */
struct structFormat *PropertyQueryStructExStruct = NULL;

/**
\internal
\brief Structure description for XPropertyQueryStruct
\see XPropertyQueryStruct
*/
struct structFormat *XPropertyQueryStructStruct = NULL;

/**
\internal
\brief Structure description for PrpQueryStruct
\see PrpQueryStruct
*/
struct structFormat *PrpQueryStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for DeviceInfoStruct
 * \see DeviceInfoStruct
 */
struct structFormat *DeviceQueryStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for ActivityQueryStruct
 * \see ActivityQueryStruct
 */
struct structFormat *ActivityQueryStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for ServerStateStruct
 * \see ServerStateStruct
 */
struct structFormat *ServerStateStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for GlbDataHdr3
 * \see GlbDataHdr3
 */
struct structFormat *GlbDataHdr3Struct = NULL;

/**
 * \internal
 * \brief Structure description for GlbDataHdr4
 * \see GlbDataHdr4
 */
struct structFormat *GlbDataHdr4Struct = NULL;

/**
* \internal
* \brief Structure description for GlbDataHdr
* \see GlbDataHdr
*/
struct structFormat *GlbDataHdrStruct = NULL;

/**
 * \internal
 * \brief Structure description for ClientQueryStruct (legacy)
 * \see ClientQueryStruct
*/
struct structFormat *LClientQueryStructStruct = NULL;

/**
* \internal
* \brief Structure description for ClnInfoQueryStruct (modern)
* \see ClnIntoQueryStruct
*/
struct structFormat *ClnInfoQueryStructStruct = NULL;
/**
 * \internal
 * \brief Structure description for SpectrumHeader
 * \see hdrSPECTRUM
 */
struct structFormat *SpectrumHeaderStruct = NULL;
/**
* \internal
* \brief Structure description for GSpectrumHeader
* \see hdrGSPECTRUM
*/
struct structFormat *GSpectrumHeaderStruct = NULL;

/**
 * \internal
 * \brief Structure description for ImageHeader
 * \see hdrIMAGE
 */
struct structFormat *ImageHeaderStruct = NULL;

/**
 * \internal
 * \brief Structure description for type MDA (multi-dimensional array)
 * \see DMBA
 */
struct structFormat *MDAHeaderStruct = NULL;

/**
 * \internal
 * \brief Structure description for HRSStruct
 * \see HRSStruct
 */
struct structFormat *HRSStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for DSUMMARY
 * \see DSUMMARY
 */
struct structFormat *SummaryStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for FDSStruct
 * \see FecDataStruct
 */
struct structFormat *FDSStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for FecAddrStruct
 * \see FecDataStruct
 */
struct structFormat *FASStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for ConTblInfo
 * \see ConTblInfo
 */
struct structFormat *ConTblInfoStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for ConTblInfo
 * \see ConTblInfo
 */
struct structFormat *WriteAccessQueryStruct = NULL;

/**
 * \internal
 * \brief Structure description for ServerSettingsStruct
 * \see ServerSettingsStruct
 */
struct structFormat *ServerSettingsStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for FilterStruct
 * \see FilterStruct
 */
struct structFormat *FilterStructStruct = NULL;

/**
 * \internal
 * \brief Structure description for the central logger entry structure: CLOGStruct
 * \see CLOG
 */
struct structFormat *CLOGStructStruct = NULL;
int logStructureEntries = TRUE;
/**
 * \internal
 * \brief Initializes and registers pre-defined tagged structures.
 * 
 * \see PktHdrStructStruct 
 * \see SubInfoPktStruct
 * \see PrdrHdrStruct
 * \see ServerStateStructStruct
 * \see RPCDataStructStruct
 * \see RPCFecStructStruct
 * \see ServerStateStructStruct
 * \see GlbDataHdrStruct
 * \see SpectrumHeaderStruct
 * \see ImageHeaderStruct
 * \see MDAHeaderStruct
 * \see PropertyQueryStruct
 * \see PropertyQueryStructEx
 * \see XPropertyQueryStruct
 * \see ActivityQueryStruct
 * \see ContractQueryStruct
 * \see ClientQueryStruct
 * \see AMSStructStruct
 * \see ADSStructStruct
 */
int initControlStructs(void)
{
  struct structStruct *ss = NULL;
  int i, cc;
  char tag[16];

  if (gControlStructInitialized) return 0;
  gControlStructInitialized = TRUE;
  logStructureEntries = FALSE;

# ifndef DOS
  if ((cc=addtostruct(&PktHdrStructStruct,
          (int)OFFSETIN(PktHdr,totalSize),
          1,CF_INT16,NULL,"totalSize")) != 0) return cc;
  if ((cc=addtostruct(&PktHdrStructStruct,
          (int)OFFSETIN(PktHdr,userName),
          USERNAME_SIZE,CF_TEXT,NULL,"userName")) != 0) return cc;
  if ((cc=addtostruct(&PktHdrStructStruct,
          (int)OFFSETIN(PktHdr,tineProtocol),
          1,CF_INT16,NULL,"payloadType")) != 0) return cc;
  if ((cc=addtostruct(&PktHdrStructStruct,
          (int)OFFSETIN(PktHdr,number),
          1,CF_INT16,NULL,"numParts")) != 0) return cc;

  if ((cc=addtostruct(&ReqHdrStructStruct,
      (int)OFFSETIN(ReqHdr, totalSizeInBytes),
      1, CF_UINT32, NULL, "totalSizeBytes")) != 0) return cc;
  if ((cc=addtostruct(&ReqHdrStructStruct,
      (int)OFFSETIN(ReqHdr, headerSizeInBytes),
      1, CF_UINT32, NULL, "headerSizeBytes")) != 0) return cc;
  if ((cc = addtostruct(&ReqHdrStructStruct,
    (int)OFFSETIN(ReqHdr, pid),
    1, CF_UINT32, NULL, "pid")) != 0) return cc;
  if ((cc = addtostruct(&ReqHdrStructStruct,
    (int)OFFSETIN(ReqHdr, tineVersion),
    1, CF_INT32, NULL, "tineVersion")) != 0) return cc;
  if ((cc = addtostruct(&ReqHdrStructStruct,
    (int)OFFSETIN(ReqHdr, numberSegments),
    1, CF_INT16, NULL, "numberSegments")) != 0) return cc;
  if ((cc = addtostruct(&ReqHdrStructStruct,
    (int)OFFSETIN(ReqHdr, tineProtocol),
    1, CF_INT16, NULL, "tineProtocol")) != 0) return cc;
  if ((cc=addtostruct(&ReqHdrStructStruct,
      (int)OFFSETIN(ReqHdr, endianness),
      1, CF_UINT16, NULL, "endianness")) != 0) return cc;
  if ((cc=addtostruct(&ReqHdrStructStruct,
      (int)OFFSETIN(ReqHdr, encoding),
      1, CF_UINT16, NULL, "encoding")) != 0) return cc;
  if ((cc=addtostruct(&ReqHdrStructStruct,
      (int)OFFSETIN(ReqHdr, userName),
      USERNAME_SIZE, CF_TEXT, NULL, "userName")) != 0) return cc;
  if ((cc=addtostruct(&ReqHdrStructStruct,
      (int)OFFSETIN(ReqHdr, userType),
      USERTYPE_SIZE, CF_TEXT, NULL, "userType")) != 0) return cc;

  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, tineProtocol),
    1, CF_INT32, NULL, "tineProtocol")) != 0) return cc;
  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, totalSizeInBytes),
    1, CF_UINT32, NULL, "totalSizeBytes")) != 0) return cc;
  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, headerSizeInBytes),
    1, CF_UINT32, NULL, "headerSizeBytes")) != 0) return cc;
  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, numberSegments),
    1, CF_INT16, NULL, "numberSegments")) != 0) return cc;
  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, endianness),
    1, CF_UINT8, NULL, "endianness")) != 0) return cc;
  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, encoding),
    1, CF_UINT8, NULL, "encoding")) != 0) return cc;
  if ((cc = addtostruct(&RspHdrStructStruct,
    (int)OFFSETIN(RspHdr, fecName),
    USERNAME_SIZE, CF_TEXT, NULL, "fecName")) != 0) return cc;

  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, msgsize),
    1, CF_INT32, NULL, "msgSize")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, extsize),
    1, CF_INT32, NULL, "extSize")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, mtu),
    1, CF_INT32, NULL, "mtu")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, blkid),
    1, CF_INT32, NULL, "blkId")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, numblks),
    1, CF_INT16, NULL, "numBlks")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, blknum),
    1, CF_INT16, NULL, "blkNum")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, id),
    1, CF_INT16, NULL, "id")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, mode),
    1, CF_INT16, NULL, "mode")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, pollingInterval),
    1, CF_INT32, NULL, "interval")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, starttime),
    1, CF_INT32, NULL, "startTime")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmProperty),
    PROPERTY_NAME_SIZE, CF_TEXT, NULL, "property")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmDeviceName),
    DEVICE_NAME_SIZE, CF_TEXT, NULL, "deviceName")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmName),
    EQM_NAME_SIZE, CF_TEXT, NULL, "eqm")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmSizeIn),
    1, CF_INT32, NULL, "sizeIn")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmSizeOut),
    1, CF_INT32, NULL, "sizeOut")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.hEqmName),
    1, CF_BYTE, NULL, "handle")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmAccess),
    1, CF_BYTE, NULL, "access")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmFormatIn),
    1, CF_BYTE, NULL, "formatIn")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.EqmFormatOut),
    1, CF_BYTE, NULL, "formatOut")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.strTagIn),
    TAG_NAME_SIZE, CF_TEXT, NULL, "tagIn")) != 0) return cc;
  if ((cc = addtostruct(&SubReqHdrStruct,
    (int)OFFSETIN(SubReqHdr, contract.strTagOut),
    TAG_NAME_SIZE, CF_TEXT, NULL, "tagOut")) != 0) return cc;

  if ((cc=addtostruct(&SubInfoPktStruct,
          (int)OFFSETIN(SubInfoPkt,msgsize),
          1,CF_INT16,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
          (int)OFFSETIN(SubInfoPkt,extsize),
          1,CF_INT16,NULL,"extSize")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,mtu),
          1,CF_INT16,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
          (int)OFFSETIN(SubInfoPkt,numblks),
          1,CF_INT16,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
          (int)OFFSETIN(SubInfoPkt,blknum),
          1,CF_INT16,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,id),
           1,CF_INT16,NULL,"id")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,mode),
           1,CF_INT16,NULL,"mode")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,blkid),
           1,CF_INT16,NULL,"blkId")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,pollingInterval),
           1,CF_INT32,NULL,"pollingRate")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,starttime),
           1,CF_INT32,NULL,"startTime")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmProperty),
           PROPERTY_NAME_SIZE,CF_TEXT,NULL,"property")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmDeviceName),
           DEVICE_NAME_SIZE,CF_TEXT,NULL,"deviceName")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmName),
           EQM_NAME_SIZE,CF_TEXT,NULL,"eqm")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmSizeIn),
           1,CF_INT32,NULL,"sizeIn")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmSizeOut),
           1,CF_INT32,NULL,"sizeOut")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.hEqmName),
           1,CF_BYTE,NULL,"handle")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmAccess),
           1,CF_BYTE,NULL,"access")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmFormatIn),
           1,CF_BYTE,NULL,"formatIn")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.EqmFormatOut),
           1,CF_BYTE,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.strTagIn),
           TAG_NAME_SIZE,CF_TEXT,NULL,"tagIn")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPktStruct,
           (int)OFFSETIN(SubInfoPkt,contract.strTagOut),
           TAG_NAME_SIZE,CF_TEXT,NULL,"tagOut")) != 0) return cc;

  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,msgsize),
           1,CF_INT16,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,mtu),
           1,CF_INT16,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,numblks),
           1,CF_BYTE,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,blknum),
           1,CF_BYTE,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,id),
           1,CF_INT16,NULL,"id")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,mode),
           1,CF_INT16,NULL,"mode")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,blkid),
           1,CF_INT16,NULL,"blkId")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,pollingInterval),
           1,CF_INT32,NULL,"pollingRate")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,starttime),
           1,CF_INT32,NULL,"startTime")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmProperty),
           PROPERTY_NAME_SHORTSIZE,CF_TEXT,NULL,"property")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmDeviceName),
           DEVICE_NAME_SHORTSIZE,CF_TEXT,NULL,"deviceName")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmName),
           EQM_NAME_SIZE,CF_TEXT,NULL,"eqm")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmSizeIn),
           1,CF_INT32,NULL,"sizeIn")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmSizeOut),
           1,CF_INT32,NULL,"sizeOut")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.hEqmName),
           1,CF_BYTE,NULL,"handle")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmAccess),
           1,CF_BYTE,NULL,"access")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmFormatIn),
           1,CF_BYTE,NULL,"formatIn")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.EqmFormatOut),
           1,CF_BYTE,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.strTagIn),
           TAG_NAME_SHORTSIZE,CF_TEXT,NULL,"tagIn")) != 0) return cc;
  if ((cc=addtostruct(&SubInfoPkt5Struct,
           (int)OFFSETIN(SubInfoPkt5,contract.strTagOut),
           TAG_NAME_SHORTSIZE,CF_TEXT,NULL,"tagOut")) != 0) return cc;

  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,msgsize),
           1,CF_INT16,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,subId),
           1,CF_INT16,NULL,"subId")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,retCode),
           1,CF_INT16,NULL,"compCode")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,numblks),
           1,CF_INT16,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,blknum),
           1,CF_INT16,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,mtu),
           1,CF_INT16,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,timestampMSEC),
           1,CF_INT16,NULL,"timestampMsec")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,EqmFormat),
           1,CF_BYTE,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,counter),
           1,CF_BYTE,NULL,"counter")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,timestamp),
           1,CF_INT32,NULL,"timestampSec")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdr5Struct,
           (int)OFFSETIN(PrdrHdr5,ClnStarttime),
           1,CF_INT32,NULL,"startTimeTag")) != 0) return cc;

  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,msgsize),
           1,CF_INT16,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,subId),
           1,CF_INT16,NULL,"subId")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,retCode),
           1,CF_INT16,NULL,"compCode")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,numblks),
           1,CF_INT16,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,blknum),
           1,CF_INT16,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,mtu),
           1,CF_INT16,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,EqmFormat),
           1,CF_INT16,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,counter),
           1,CF_INT16,NULL,"counter")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,tineProtocol),
           1,CF_INT16,NULL,"tineProtocol")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,xferReason),
           1,CF_INT16,NULL,"xferReason")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,ClnStarttime),
           1,CF_INT32,NULL,"startTimeTag")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,stssize),
           1,CF_INT16,NULL,"statusSize")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,stscode),
           1,CF_INT16,NULL,"statusCode")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,timestamp),
           1,CF_INT32,NULL,"tv_sec")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,timestampUSec),
           1,CF_INT32,NULL,"tv_usec")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,userstamp),
           1,CF_INT32,NULL,"userStamp")) != 0) return cc;
  if ((cc=addtostruct(&PrdrHdrStruct,
           (int)OFFSETIN(PrdrHdr,sysstamp),
           1,CF_INT32,NULL,"sysStamp")) != 0) return cc;

  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, msgSize),
    1, CF_INT32, NULL, "msgSize")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, subId),
    1, CF_INT16, NULL, "subId")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, retCode),
    1, CF_INT16, NULL, "compCode")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, numblks),
    1, CF_INT16, NULL, "numBlks")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, blknum),
    1, CF_INT16, NULL, "blkNum")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, mtu),
    1, CF_INT32, NULL, "mtu")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, dataFormat),
    1, CF_INT16, NULL, "dataFormat")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, counter),
    1, CF_INT16, NULL, "counter")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, tineProtocol),
    1, CF_INT16, NULL, "tineProtocol")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, xferReason),
    1, CF_INT16, NULL, "xferReason")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, starttime),
    1, CF_INT32, NULL, "starttime")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, stssize),
    1, CF_INT16, NULL, "statusSize")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, stscode),
    1, CF_INT16, NULL, "statusCode")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, timestamp),
    1, CF_INT32, NULL, "tv_sec")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, timestampUSec),
    1, CF_INT32, NULL, "tv_usec")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, userstamp),
    1, CF_INT32, NULL, "userStamp")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, sysstamp),
    1, CF_INT32, NULL, "sysStamp")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, dataSize),
    1, CF_INT32, NULL, "dataSize")) != 0) return cc;
  if ((cc = addtostruct(&SubRspHdrStruct,
    (int)OFFSETIN(SubRspHdr, msgSizeInElements),
    1, CF_INT32, NULL, "sizeElements")) != 0) return cc;

  if ((cc=addtostruct(&RPCDataStructStruct,
           (int)OFFSETIN(RPCDataStruct,FecName),
           FEC_NAME_SIZE,CF_TEXT,NULL,"fecName")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
           (int)OFFSETIN(RPCDataStruct,SubSystem),
           5,CF_TEXT,NULL,"subSystem")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
           (int)OFFSETIN(RPCDataStruct,ExportName),
           EXPORT_NAME_SHORTSIZE,CF_TEXT,NULL,"exportName")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
           (int)OFFSETIN(RPCDataStruct,EqmName),
           EQM_NAME_SHORTSIZE+1,CF_TEXT,NULL,"eqmName")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
           (int)OFFSETIN(RPCDataStruct,EqmContext),
           CONTEXT_NAME_SHORTSIZE,CF_TEXT,NULL,"context")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
           (int)OFFSETIN(RPCDataStruct,alarmSystem),
           1,CF_INT16,NULL,"almSystem")) != 0) return cc;

  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,fecName),
           FEC_NAME_SIZE,CF_TEXT,NULL,"fecName")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,reserved),
           5,CF_BYTE,NULL,"reserved")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,Network),
           4,CF_BYTE,NULL,"ipxNet")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,Node),
           6,CF_BYTE,NULL,"ipxNode")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,ImmediateAddress),
           6,CF_BYTE,NULL,"ipxImmAddr")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,IP),
           ADDR_SIZE,CF_TEXT,NULL,"ipAddr")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,portOffset),
           1,CF_BYTE,NULL,"portOffset")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,IPh_addr),
           4,CF_BYTE,NULL,"ip_haddr")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
           (int)OFFSETIN(RPCFecStruct,Prot),
           2,CF_BYTE,NULL,"transportType")) != 0) return cc;

  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,desc),
           64,CF_TEXT,NULL,"desc")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,dataFmt),
           1,CF_BYTE,NULL,"dataFmt")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,dataSiz),
           1,CF_BYTE,NULL,"dataSiz")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,data),
           64,CF_BYTE,NULL,"data")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,timeEntered),
           1,CF_INT32,NULL,"timeEntered")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,codeNow),
           1,CF_INT32,NULL,"codeNow")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,codePrv),
           1,CF_INT32,NULL,"codePrv")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
           (int)OFFSETIN(ServerStateStruct,codeNxt),
           1,CF_INT32,NULL,"codeNxt")) != 0) return cc;

  if ((cc=addtostruct(&GlbDataHdr3Struct,
           (int)OFFSETIN(GlbDataHdr3,name),
           KEYWORD_NAME_SHORTSIZE,CF_TEXT,NULL,"keyword")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr3Struct,
           (int)OFFSETIN(GlbDataHdr3,size),
           1,CF_INT16,NULL,"size")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr3Struct,
           (int)OFFSETIN(GlbDataHdr3,format),
           1,CF_INT16,NULL,"format")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr3Struct,
           (int)OFFSETIN(GlbDataHdr3,time),
           1,CF_INT32,NULL,"utc")) != 0) return cc;

  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,name),
           KEYWORD_NAME_SIZE,CF_TEXT,NULL,"keyword")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,size),
           1,CF_INT16,NULL,"size")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,format),
           1,CF_INT16,NULL,"format")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,status),
           1,CF_INT16,NULL,"status")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,dTimeStampUSEC),
           1,CF_INT16,NULL,"timeStampUSEC")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,dStamp),
           1,CF_INT32,NULL,"dataStamp")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,sysStamp),
           1,CF_INT32,NULL,"sysStamp")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,dTimeStamp),
           1,CF_INT32,NULL,"timeStamp")) != 0) return cc;
  if ((cc=addtostruct(&GlbDataHdr4Struct,
           (int)OFFSETIN(GlbDataHdr4,dTimeStampMSEC),
           1,CF_INT32,NULL,"timeStampMSec")) != 0) return cc;

  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, name),
    KEYWORD_NAME_SIZE, CF_TEXT, NULL, "keyword")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, size),
    1, CF_INT16, NULL, "size")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, format),
    1, CF_INT16, NULL, "format")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, status),
    1, CF_INT16, NULL, "status")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, interval),
    1, CF_INT16, NULL, "interval")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, dStamp),
    1, CF_INT32, NULL, "dataStamp")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, sysStamp),
    1, CF_INT32, NULL, "sysStamp")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, dTimeStamp),
    1, CF_INT32, NULL, "timeStamp")) != 0) return cc;
  if ((cc = addtostruct(&GlbDataHdrStruct,
    (int)OFFSETIN(GlbDataHdr, dTimeStampUSEC),
    1, CF_INT32, NULL, "timeStampUSec")) != 0) return cc;

  if ((cc=addtostruct(&SpectrumHeaderStruct,
           (int)OFFSETIN(hdrSPECTRUM,comment),
           80,CF_TEXT,NULL,"comment")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
           (int)OFFSETIN(hdrSPECTRUM,tm),
           1,CF_INT32,NULL,"tm")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
           (int)OFFSETIN(hdrSPECTRUM,s_start),
           1,CF_FLOAT,NULL,"start")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
           (int)OFFSETIN(hdrSPECTRUM,s_inc),
           1,CF_FLOAT,NULL,"inc")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
           (int)OFFSETIN(hdrSPECTRUM,status),
           1,CF_INT32,NULL,"status")) != 0) return cc;

  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, comment),
    80, CF_TEXT, NULL, "comment")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, id),
    1, CF_INT32, NULL, "id")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, tm),
    1, CF_INT32, NULL, "tm")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, s_start),
    1, CF_FLOAT, NULL, "start")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, s_inc),
    1, CF_FLOAT, NULL, "inc")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, status),
    1, CF_INT32, NULL, "status")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, groups),
    1, CF_INT32, NULL, "groups")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, group_size),
    1, CF_INT32, NULL, "group_size")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, group_inc),
    1, CF_FLOAT, NULL, "group_inc")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, stat_mask),
    1, CF_INT32, NULL, "stat_mask")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, statistics),
    GSP_STSZ, CF_FLOAT, NULL, "statistics")) != 0) return cc;
  if ((cc = addtostruct(&GSpectrumHeaderStruct,
    (int)OFFSETIN(hdrGSPECTRUM, timing_param),
    GSP_TISZ, CF_INT32, NULL, "timing_param")) != 0) return cc;

  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.baseTag),
           1,CF_INT32,NULL,"baseTag")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.cameraPortId),
           1,CF_INT32,NULL,"cameraPortId")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.versionTag),
           1,CF_INT32,NULL,"version")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.totalLength),
           1,CF_INT32,NULL,"totalLength")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.timestampSeconds),
           1,CF_INT32,NULL,"timestampSec")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.timestampMicroseconds),
           1,CF_INT32,NULL,"timestampUSec")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,sourceHeader.cameraPortName),
           64,CF_TEXT,NULL,"cameraPort")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.sourceWidth),
           1,CF_INT32,NULL,"sourceWidth")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.sourceHeight),
           1,CF_INT32,NULL,"sourceHeight")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.aoiWidth),
           1,CF_INT32,NULL,"aoiWidth")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.aoiHeight),
           1,CF_INT32,NULL,"aoiHeight")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.xStart),
           1,CF_INT32,NULL,"xStart")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.yStart),
           1,CF_INT32,NULL,"yStart")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.bytesPerPixel),
           1,CF_INT32,NULL,"bytesPerPixel")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.effectiveBitsPerPixel),
           1,CF_INT32,NULL,"effBitsPerPixel")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.horizontalBinning),
           1,CF_INT32,NULL,"horzBinning")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.verticalBinning),
           1,CF_INT32,NULL,"vertBinning")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.sourceFormat),
           1,CF_INT32,NULL,"sourceFormat")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.imageFormat),
           1,CF_INT32,NULL,"imageFormat")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.frameNumber),
           1,CF_INT32,NULL,"frameNumber")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.eventNumber),
           1,CF_INT32,NULL,"evnNum")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.xScale),
           1,CF_FLOAT,NULL,"xScale")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.yScale),
           1,CF_FLOAT,NULL,"yScale")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.imageRotation),
           1,CF_FLOAT,NULL,"imageRotation")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.fspare1),
           1,CF_FLOAT,NULL,"fspare1")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.fspare2),
           1,CF_FLOAT,NULL,"fspare2")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.fspare3),
           1,CF_FLOAT,NULL,"fspare3")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.imageFlags),
           1,CF_INT32,NULL,"imageFlags")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.ispare1),
           1,CF_INT32,NULL,"ispare1")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.ispare2),
           1,CF_INT32,NULL,"ispare2")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.ispare3),
           1,CF_INT32,NULL,"ispare3")) != 0) return cc;
  if ((cc=addtostruct(&ImageHeaderStruct,
           (int)OFFSETIN(hdrIMAGE,frameHeader.appendedFrameSize),
           1,CF_INT32,NULL,"appFrameSize")) != 0) return cc;

  if ((cc=addtostruct(&MDAHeaderStruct,
           (int)OFFSETIN(DMDA,cmt),
           80,CF_TEXT,NULL,"cmt")) != 0) return cc;
  if ((cc=addtostruct(&MDAHeaderStruct,
           (int)OFFSETIN(DMDA,ndim),
           1,CF_INT32,NULL,"ndim")) != 0) return cc;
  if ((cc=addtostruct(&MDAHeaderStruct,
           (int)OFFSETIN(DMDA,fmt),
           1,CF_INT32,NULL,"fmt")) != 0) return cc;
  for (i=0; i<MAX_MDA_DIMS; i++)
  {
    sprintf(tag,"len[%d]",i);
    if ((cc=addtostruct(&MDAHeaderStruct,(int)OFFSETIN(DMDA,axis[i].len),1,CF_INT32,NULL,tag)) != 0) return cc;
    sprintf(tag,"reserved[%d]",i);
    if ((cc=addtostruct(&MDAHeaderStruct,(int)OFFSETIN(DMDA,axis[i].reserved),1,CF_INT32,NULL,tag)) != 0) return cc;
    sprintf(tag,"label[%d]",i);
    if ((cc=addtostruct(&MDAHeaderStruct,(int)OFFSETIN(DMDA,axis[i].label),16,CF_TEXT,NULL,tag)) != 0) return cc;
    sprintf(tag,"units[%d]",i);
    if ((cc=addtostruct(&MDAHeaderStruct,(int)OFFSETIN(DMDA,axis[i].units),16,CF_TEXT,NULL,tag)) != 0) return cc;
    sprintf(tag,"min[%d]",i);
    if ((cc=addtostruct(&MDAHeaderStruct,(int)OFFSETIN(DMDA,axis[i].min),1,CF_FLOAT,NULL,tag)) != 0) return cc;
    sprintf(tag,"max[%d]",i);
    if ((cc=addtostruct(&MDAHeaderStruct,(int)OFFSETIN(DMDA,axis[i].max),1,CF_FLOAT,NULL,tag)) != 0) return cc;
  }

  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvWorkArea),
           1,CF_INT32,"srvWorkArea")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,cycDeadBand),
           1,CF_INT32,"cycDeadBand")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,minPollInterval),
           1,CF_INT32,"minPollInterval")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,conTblCapacity),
           1,CF_INT32,"conTblCapacity")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,conRenewalLen),
           1,CF_INT32,"conRenewalLen")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,clnTblCapacity),
           1,CF_INT32,"clnTblCapacity")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,ackOnChange),
           1,CF_INT32,"ackOnChange")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvBurstLimit),
           1,CF_INT32,"srvBurstLimit")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvBurstCycDly),
           1,CF_INT32,"srvBurstCycDly")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvPacketMTU),
           1,CF_INT32,"srvPacketMTU")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,tcpMaxMsgSize),
           1,CF_INT32,"tcpMaxMsgSize")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,tcpTblCapacity),
           1,CF_INT32,"tcpTblCapacity")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvSendBuffers),
           1,CF_INT32,"srvSendBuffers")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvRecvBuffers),
           1,CF_INT32,"srvRecvBuffers")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvLazySched),
           1,CF_INT32,"srvLazySched")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvBkgTskRentr),
           1,CF_INT32,"srvBkgTskRentr")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvCycInSepThrd),
           1,CF_INT32,"srvCycInSepThrd")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,tcpTraInSepThrd),
           1,CF_INT32,"tcpTraInSepThrd")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,clnWorkArea),
           1,CF_INT32,"clnWorkArea")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,lnkTblCapacity),
           1,CF_INT32,"lnkTblCapacity")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,clnBurstLimit),
           1,CF_INT32,"clnBurstLimit")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,clnSendBuffers),
           1,CF_INT32,"clnSendBuffers")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,clnRecvBuffers),
           1,CF_INT32,"clnRecvBuffers")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,clnRecvQueue),
           1,CF_INT32,"clnRecvQueue")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,useLoopbackAddr),
           1,CF_INT32,"useLoopbackAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,useWdogLinks),
           1,CF_INT32,"useWdogLinks")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,allowCommonLnks),
           1,CF_INT32,"allowCommonLnks")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,retryOnTimeout),
           1,CF_INT32,"retryOnTimeout")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvSchedIntvl),
           1,CF_INT32,"srvSchedIntvl")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvRetrdRemovl),
           1,CF_INT32,"srvRetrdRemovl")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,srvAllowRemote),
           1,CF_INT32,"srvAllowRemote")) != 0) return cc;
  if ((cc=addFieldToStruct("SrvSetQy",
           (int)OFFSETIN(ServerSettingsStruct,reserved4),
           1,CF_INT32,"reserved4")) != 0) return cc;
  
  if ((cc=sealTaggedStruct("SrvSetQy",sizeof(ServerSettingsStruct),1)) != 0) return cc;

  if ((ss=findstruct("SrvSetQy")) == NULL) return out_of_server_memory;
  ServerSettingsStructStruct = ss->s;

# endif /* !defined DOS */

  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,fecName),
           FEC_NAME_SIZE,CF_TEXT,"Name")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,IPXNetwork),
           4,CF_BYTE,"IPXNetwork")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,IPXNode),
           6,CF_BYTE,"IPXNode")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,IPXImmediateAddress),
           6,CF_BYTE,"IPXImmAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,IP),
           ADDR_SIZE,CF_TEXT,"IP")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,IPh_addr),
           4,CF_TEXT,"IPh_addr")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,portOffset),
           1,CF_INT32,"portOffset")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,inetProtocol),
           1,CF_INT32,"TransProt")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
           (int)OFFSETIN(FecDataStruct,tineProtocol),
           1,CF_INT32,"tineProtocol")) != 0) return cc;
  if ((cc=sealTaggedStruct("FDSr4",sizeof(FecDataStruct),1)) != 0) return cc;

  if ((ss=findstruct("FDSr4")) == NULL) return out_of_server_memory;
  FDSStructStruct = ss->s;

  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,fecName),
           FEC_NAME_SIZE,CF_TEXT,"Name")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,netAdr),
           16,CF_BYTE,"netAdr")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,lclAdr),
           8,CF_BYTE,"lclAdr")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,strAdr),
           ADDR_SIZE_IPV6,CF_TEXT,"strAdr")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,portOffset),
           1,CF_INT32,"portOffset")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,inetProtocol),
           1,CF_INT32,"TransProt")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,tineProtocol),
           1,CF_INT32,"tineProtocol")) != 0) return cc;
  if ((cc=addFieldToStruct("FASr5",
           (int)OFFSETIN(FecAddrStruct,addrFlags),
           1,CF_INT32,"addrFlags")) != 0) return cc;
  if ((cc=sealTaggedStruct("FASr5",sizeof(FecAddrStruct),1)) != 0) return cc;

  if ((ss=findstruct("FASr5")) == NULL) return out_of_server_memory;
  FASStructStruct = ss->s;

  if ((cc=addFieldToStruct("PQS",
           (int)OFFSETIN(PropertyQueryStruct,prpName),
           PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
           (int)OFFSETIN(PropertyQueryStruct,prpDescription),
           32,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
           (int)OFFSETIN(PropertyQueryStruct,prpSize),
           1,CF_INT16,"size")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
           (int)OFFSETIN(PropertyQueryStruct,prpFormat),
           1,CF_BYTE,"format")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
           (int)OFFSETIN(PropertyQueryStruct,prpAccess),
           1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=sealTaggedStruct("PQS",sizeof(PropertyQueryStruct),100)) != 0) return cc;

  if ((ss=findstruct("PQS")) == NULL) return out_of_server_memory;
  PropertyQueryStructStruct = ss->s;

# ifndef SMALL_TINE_LIB
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpName),
           PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpDescription),
           64,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpRedirection),
           32,CF_TEXT,"redirection")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpTag),
           TAG_NAME_SHORTSIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpTagIn),
           TAG_NAME_SHORTSIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpUnits),
           UNITS_SIZE,CF_TEXT,"units")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpMinValue),
           1,CF_FLOAT,"min")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpMaxValue),
           1,CF_FLOAT,"max")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpSize),
           1,CF_INT32,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpSizeIn),
           1,CF_INT32,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpNumOverloads),
           1,CF_INT32,"overloads")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpHistoryDepthShort),
           1,CF_INT16,"historyShort")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpHistoryDepthLong),
           1,CF_INT16,"historyLong")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
           (int)OFFSETIN(PropertyQueryStructEx,prpFormat),
           4,CF_BYTE,"formats")) != 0) return cc;
  if ((cc=sealTaggedStruct("PQSX",sizeof(PropertyQueryStructEx),100)) != 0) return cc;

  if ((ss=findstruct("PQSX")) == NULL) return out_of_server_memory;
  PropertyQueryStructExStruct = ss->s;

  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpName),
           PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpDescription),
           64,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpRedirection),
           32,CF_TEXT,"redirection")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpTag),
           TAG_NAME_SHORTSIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpTagIn),
           TAG_NAME_SHORTSIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpUnits),
           UNITS_SIZE,CF_TEXT,"units")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpMinValue),
           1,CF_FLOAT,"min")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpMaxValue),
           1,CF_FLOAT,"max")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpSize),
           1,CF_INT32,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpSizeIn),
           1,CF_INT32,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpNumOverloads),
           1,CF_INT32,"overloads")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpHistoryDepthShort),
           1,CF_INT16,"historyShort")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpHistoryDepthLong),
           1,CF_INT16,"historyLong")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpFormat),
           1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpFormatIn),
           1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpAccess),
           1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpGraphType),
           1,CF_BYTE,"graphType")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,rngUnits),
           UNITS_SIZE,CF_TEXT,"rangeUnits")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,rngMinValue),
           1,CF_FLOAT,"rangeMin")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,rngMaxValue),
           1,CF_FLOAT,"rangeMax")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,numRows),
           1,CF_INT16,"numRows")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,rowSize),
           1,CF_INT16,"rowSize")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,prpArrayType),
           1,CF_INT16,"arrayType")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
           (int)OFFSETIN(XPropertyQueryStruct,reserved),
           3,CF_INT16,"reserved")) != 0) return cc;
  if ((cc=sealTaggedStruct("XPQS",sizeof(XPropertyQueryStruct),10)) != 0) return cc;

  if ((ss=findstruct("XPQS")) == NULL) return out_of_server_memory;
  XPropertyQueryStructStruct = ss->s;
# endif

  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpName),
           PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpDescription),
           PROPERTY_DESC_SIZE,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpRedirection),
           PROPERTY_REDIR_SIZE,CF_TEXT,"redirection")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpTag),
           TAG_NAME_SIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpTagIn),
           TAG_NAME_SIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpUnits),
           UNITS_SIZE,CF_TEXT,"units")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpMinValue),
           1,CF_FLOAT,"min")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpMaxValue),
           1,CF_FLOAT,"max")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpSize),
           1,CF_INT32,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpSizeIn),
           1,CF_INT32,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpNumOverloads),
           1,CF_INT32,"overloads")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpHistoryDepthShort),
           1,CF_INT16,"historyShort")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpHistoryDepthLong),
           1,CF_INT16,"historyLong")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpFormat),
           1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpFormatIn),
           1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpAccess),
           1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpGraphType),
           1,CF_BYTE,"graphType")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,rngUnits),
           UNITS_SIZE,CF_TEXT,"rangeUnits")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,rngMinValue),
           1,CF_FLOAT,"rangeMin")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,rngMaxValue),
           1,CF_FLOAT,"rangeMax")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,numRows),
           1,CF_INT16,"numRows")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,rowSize),
           1,CF_INT16,"rowSize")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,prpArrayType),
           1,CF_INT16,"arrayType")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
           (int)OFFSETIN(PrpQueryStruct,reserved),
           3,CF_INT16,"reserved")) != 0) return cc;
  if ((cc=sealTaggedStruct("PRPQSr4",sizeof(PrpQueryStruct),10)) != 0) return cc;

  if ((ss=findstruct("PRPQSr4")) == NULL) return out_of_server_memory;
  PrpQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, ctx),
    CONTEXT_NAME_SIZE, CF_TEXT, "context")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, srv),
    EXPORT_NAME_SIZE, CF_TEXT, "server")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, name),
    DEVICE_NAME_SIZE, CF_TEXT, "name")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, desc),
    DEVICE_DESC_SIZE, CF_TEXT, "description")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, loc),
    DEVICE_LOCATION_SIZE, CF_TEXT, "location")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, zpos),
    1, CF_FLOAT, "z-position")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, mask),
    1, CF_INT32, "mask")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, region),
    1, CF_INT32, "region")) != 0) return cc;
  if ((cc=addFieldToStruct("DEVQSr5",
    (int)OFFSETIN(DeviceInfoStruct, number),
    1, CF_INT32, "number")) != 0) return cc;
  if ((cc=sealTaggedStruct("DEVQSr5", sizeof(DeviceInfoStruct), 1)) != 0) return cc;

  if ((ss=findstruct("DEVQSr5")) == NULL) return out_of_server_memory;
  DeviceQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,FecName),
           FEC_NAME_SIZE,CF_TEXT,"fecName")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numCalls),
           1,CF_INT32,"numCalls")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,localtime),
           1,CF_INT32,"localtime")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,starttime),
           1,CF_INT32,"starttime")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,systemPollingRate),
           1,CF_INT32,"sysPollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numBkgTsks),
           1,CF_INT16,"numBkgTasks")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numTotalContracts),
           1,CF_INT16,"numContracts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numTotalClients),
           1,CF_INT16,"numClients")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numTargetContracts),
           1,CF_INT16,"numTgtCont")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numTargetClients),
           1,CF_INT16,"numTgtCln")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numConnections),
           1,CF_INT16,"numConnections")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numConnectionTimeouts),
           1,CF_INT32,"numTimeouts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numConnectionArrivals),
           1,CF_INT32,"numArrivals")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numUDPpkts),
           1,CF_INT32,"numUDPPkts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numTCPpkts),
           1,CF_INT32,"numTCPPkts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numIPXpkts),
           1,CF_INT32,"numIPXPkts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
           (int)OFFSETIN(ActivityQueryStruct,numSPXpkts),
           1,CF_INT32,"numSPSPkts")) != 0) return cc;
  if ((cc=sealTaggedStruct("AQS",sizeof(ActivityQueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("AQS")) == NULL) return out_of_server_memory;
  ActivityQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmProperty),
           PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmDeviceName),
           DEVICE_NAME_SHORTSIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmName),
           EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmSizeIn),
           1,CF_INT32,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmSizeOut),
           1,CF_INT32,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.hEqmName),
           1,CF_BYTE,"hEqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmAccess),
           1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmFormatIn),
           1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.EqmFormatOut),
           1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.strTagIn),
           TAG_NAME_SHORTSIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,contract.strTagOut),
           TAG_NAME_SHORTSIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,pollingInterval),
           1,CF_INT32,"pollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
           (int)OFFSETIN(Contract5QueryStruct,nclients),
           1,CF_INT32,"nClients")) != 0) return cc;
  if ((cc=sealTaggedStruct("CONQS",sizeof(Contract5QueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("CONQS")) == NULL) return out_of_server_memory;
  CONTRACT5Struct = ss->s;

  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmProperty),
           PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmDeviceName),
           DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmName),
           EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmSizeIn),
           1,CF_INT32,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmSizeOut),
           1,CF_INT32,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.hEqmName),
           1,CF_BYTE,"hEqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmAccess),
           1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmFormatIn),
           1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.EqmFormatOut),
           1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.strTagIn),
           TAG_NAME_SIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,contract.strTagOut),
           TAG_NAME_SIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,pollingInterval),
           1,CF_INT32,"pollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
           (int)OFFSETIN(ContractQueryStruct,nclients),
           1,CF_INT32,"nClients")) != 0) return cc;
  if ((cc=sealTaggedStruct("CTQSr4",sizeof(ContractQueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("CTQSr4")) == NULL) return out_of_server_memory;
  CONTRACTStruct = ss->s;

  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.userName),
           USERNAME_SIZE,CF_TEXT,"username")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.IPXaddress),
           12,CF_BYTE,"ipxAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.node),
           6,CF_BYTE,"ipxNode")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_family),
           1,CF_INT16,"ipFamily")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_port),
           1,CF_INT16,"ipPort")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_addr.s_addr),
           4,CF_BYTE,"ipAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_zero),
           8,CF_BYTE,"ipZero")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.ncontracts),
           1,CF_INT16,"numContracts")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
           (int)OFFSETIN(ClientQueryStruct,cln.tineProtocol),
           1,CF_INT16,"inetProtocol")) != 0) return cc;
  if ((cc=sealTaggedStruct("CLNQS",sizeof(ClientQueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("CLNQS")) == NULL) return out_of_server_memory;
  LClientQueryStructStruct = ss->s;

  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, userName),
    USERNAME_SIZE, CF_TEXT, "username")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, userAddr),
    ADDR_SIZE_IPV6, CF_TEXT, "address")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, userType),
    USERTYPE_SIZE, CF_TEXT, "appType")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, pid),
    1, CF_INT32, "pid")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, port),
    1, CF_UINT16, "port")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, ncontracts),
    1, CF_INT16, "numContracts")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, tineProtocol),
    1, CF_INT16, "tineProtocol")) != 0) return cc;
  if ((cc = addFieldToStruct("ClnInfoQS",
    (int)OFFSETIN(ClnInfoQueryStruct, inetProtocol),
    1, CF_INT16, "inetProtocol")) != 0) return cc;

  if ((ss=findstruct("ClnInfoQS")) == NULL) return out_of_server_memory;
  ClnInfoQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,server),
           EXPORT_NAME_SHORTSIZE,CF_TEXT,"server")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,device),
           DEVICE_NAME_SHORTSIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmTag),
           ALARM_TAG_SHORTSIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmCode),
           1,CF_INT32,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,timestamp),
           1,CF_INT32,"almTime")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmMask),
           1,CF_INT32,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmData),
           ALARM_DATA_SHORTSIZE,CF_BYTE,"almData")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmDataFormat),
           1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmDataArraySize),
           1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,severity),
           1,CF_BYTE,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,descriptor),
           1,CF_BYTE,"almDescriptor")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
           (int)OFFSETIN(AMS5,alarmSystem),
           1,CF_INT16,"almSystem")) != 0) return cc;
  if ((cc=sealTaggedStruct("AMS",sizeof(AMS5),gMaxAMSCapacity)) != 0) return cc;

  if ((ss=findstruct("AMS")) == NULL) return out_of_server_memory;
  AMS5StructStruct = ss->s;
    
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,server),
           EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,device),
           DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmTag),
           ALARM_TAG_SIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmCode),
           1,CF_INT32,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,timestamp),
           1,CF_INT32,"almTime")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,timestampUSec),
           1,CF_INT32,"almTimeUSec")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,starttime),
           1,CF_INT32,"almStart")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,starttimeUSec),
           1,CF_INT32,"almStartUSec")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmMask),
           1,CF_INT32,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmData),
           ALARM_DATA_SIZE,CF_BYTE,"almData")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmDataFormat),
           1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmDataArraySize),
           1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,severity),
           1,CF_BYTE,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,descriptor),
           1,CF_BYTE,"almDescriptor")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmSystem),
           1,CF_INT16,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmOscillationWindow),
           1,CF_BYTE,"almOscWindow")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
           (int)OFFSETIN(AMS,alarmOscillationWindowPinned),
           1,CF_BYTE,"almOscPinned")) != 0) return cc;
  if ((cc=sealTaggedStruct("AMSr4",sizeof(AMS),gMaxAMSCapacity)) != 0) return cc;

  if ((ss=findstruct("AMSr4")) == NULL) return out_of_server_memory;
  AMSStructStruct = ss->s;

  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmTag),
           ALARM_TAG_SHORTSIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmCode),
           1,CF_INT32,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmMask),
           1,CF_INT32,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,severity),
           1,CF_INT16,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmDataFormat),
           1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmDataArraySize),
           1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,timeRes),
           1,CF_INT32,"timeRes")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,hitsChannel),
           1,CF_INT16,"timeChan")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,hitsModule),
           1,CF_INT16,"timeModule")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmText),
           ALARM_TEXT_SHORTSIZE,CF_TEXT,"almText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,deviceText),
           ALARM_TEXT_SHORTSIZE-2,CF_TEXT,"devText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,alarmSystem),
           1,CF_INT16,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,dataText),
           ALARM_TEXT_SHORTSIZE,CF_TEXT,"dataText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,url),
           ALARM_TEXT_SHORTSIZE,CF_TEXT,"url")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
           (int)OFFSETIN(ADS5,ndefs),
           1,CF_INT32,"numDefs")) != 0) return cc;
  if ((cc=sealTaggedStruct("ADS",sizeof(ADS5),100)) != 0) return cc;

  if ((ss=findstruct("ADS")) == NULL) return out_of_server_memory;
  ADS5StructStruct = ss->s;

  if ((cc=addFieldToStruct("Filt",
           (int)OFFSETIN(FilterStruct,filterIdx),
           1,CF_INT32,"FilterIdx")) != 0) return cc;
  if ((cc=addFieldToStruct("Filt",
           (int)OFFSETIN(FilterStruct,filterLinkStr),
           256,CF_TEXT,"FilterLink")) != 0) return cc;
  if ((cc=sealTaggedStruct("Filt",sizeof(FilterStruct),256)) != 0) return cc;

  if ((ss=findstruct("Filt")) == NULL) return out_of_server_memory;
  FilterStructStruct = ss->s;

  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmTag),
           ALARM_TAG_SIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmCode),
           1,CF_INT32,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmMask),
           1,CF_INT32,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmSystem),
           1,CF_INT16,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmSeverity),
           1,CF_INT16,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmDataFormat),
           1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmDataArraySize),
           1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmOscillationWindow),
           1,CF_BYTE,"almOscWindow")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmOscillationWindowPinned),
           1,CF_BYTE,"almOscPinned")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,alarmText),
           ALARM_TEXT_SIZE,CF_TEXT,"almText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,deviceText),
           ALARM_TEXT_SIZE,CF_TEXT,"devText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,dataText),
           ALARM_TEXT_SIZE,CF_TEXT,"dataText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
           (int)OFFSETIN(ADS,url),
           ALARM_TEXT_LONGSIZE,CF_TEXT,"url")) != 0) return cc;
  if ((cc=sealTaggedStruct("ADSr4",sizeof(ADS),256)) != 0) return cc;

  if ((ss=findstruct("ADSr4")) == NULL) return out_of_server_memory;
  ADSStructStruct = ss->s;

  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,eqm),
           EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,dev),
           DEVICE_NAME_SIZE,CF_TEXT,"dev")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,prp),
           PROPERTY_NAME_SIZE,CF_TEXT,"prp")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,siz),
           1,CF_INT32,"len")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,fmt),
           1,CF_INT32,"fmt")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,atyp),
           1,CF_INT32,"arrayType")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,asys),
           1,CF_INT32,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,sev),
           1,CF_INT32,"genSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,hisev),
           1,CF_INT32,"hiSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,hiwarnsev),
           1,CF_INT32,"hiwarnSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,losev),
           1,CF_INT32,"loSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,lowarnsev),
           1,CF_INT32,"lowarnSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,chkNormal),
           1,CF_INT32,"valueCheck")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,mask),
           1,CF_INT32,"valueMask")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,normal),
           1,CF_INT32,"valueNormal")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,cnt),
           1,CF_INT32,"cnts")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,cntThreshold),
           1,CF_INT32,"cntsThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,hi),
           1,CF_FLOAT,"hiThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,lo),
           1,CF_FLOAT,"loThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,hiwarn),
           1,CF_FLOAT,"hiwarnThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
           (int)OFFSETIN(AWS,lowarn),
           1,CF_FLOAT,"lowarnThreshold")) != 0) return cc;
  if ((cc=sealTaggedStruct("AWSr4",sizeof(AWS),100)) != 0) return cc;

  if ((ss=findstruct("AWSr4")) == NULL) return out_of_server_memory;
  AWSStructStruct = ss->s;


  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, eqm),
    EQM_NAME_SIZE, CF_TEXT, "eqm")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, dev),
    DEVICE_NAME_SIZE, CF_TEXT, "dev")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, prp),
    PROPERTY_NAME_SIZE, CF_TEXT, "prp")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, siz),
    1, CF_INT32, "len")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, fmt),
    1, CF_INT32, "fmt")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, atyp),
    1, CF_INT32, "arrayType")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, asys),
    1, CF_INT32, "almSystem")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, sev),
    1, CF_INT32, "genSeverity")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, hisev),
    1, CF_INT32, "hiSeverity")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, hiwarnsev),
    1, CF_INT32, "hiwarnSeverity")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, losev),
    1, CF_INT32, "loSeverity")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, lowarnsev),
    1, CF_INT32, "lowarnSeverity")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, chkNormal),
    1, CF_INT32, "valueCheck")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, mask),
    1, CF_INT32, "valueMask")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, normal),
    1, CF_INT32, "valueNormal")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, cnt),
    1, CF_INT32, "cnts")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, cntThreshold),
    1, CF_INT32, "cntsThreshold")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, hi),
    1, CF_FLOAT, "hiThreshold")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, lo),
    1, CF_FLOAT, "loThreshold")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, hiwarn),
    1, CF_FLOAT, "hiwarnThreshold")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, lowarn),
    1, CF_FLOAT, "lowarnThreshold")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, code),
    1, CF_INT32, "code")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, codeHigh),
    1, CF_INT32, "codeHigh")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, codeHighWarn),
    1, CF_INT32, "codeHighWarn")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, codeLow),
    1, CF_INT32, "codeLow")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, codeLowWarn),
    1, CF_INT32, "codeLowWarn")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, tagNormal),
    ALARM_TAG_SIZE, CF_TEXT, "tagNormal")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, tagHigh),
    ALARM_TAG_SIZE, CF_TEXT, "tagHigh")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, tagHighWarn),
    ALARM_TAG_SIZE, CF_TEXT, "tagHighWarn")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, tagLow),
    ALARM_TAG_SIZE, CF_TEXT, "tagLow")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, tagLowWarn),
    ALARM_TAG_SIZE, CF_TEXT, "tagLowWarn")) != 0) return cc;
  if ((cc = addFieldToStruct("AWSr5",
    (int)OFFSETIN(AWLQS, fltr),
    256, CF_TEXT, "filter")) != 0) return cc;

  if ((cc = sealTaggedStruct("AWSr5", sizeof(AWLQS), 100)) != 0) return cc;

  if ((ss = findstruct("AWSr4")) == NULL) return out_of_server_memory;
  AWLQSStructStruct = ss->s;

  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,property),
           PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,device),
           DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,datasiz),
           1,CF_INT32,"dataSize")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,datafmt),
           1,CF_INT32,"dataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,pollingRate),
           1,CF_INT32,"pollingRate")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,archiveRate),
           1,CF_INT32,"archiveRate")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,depthShort),
           1,CF_INT32,"depthShort")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,depthLong),
           1,CF_INT32,"depthLong")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,heartbeat),
           1,CF_INT32,"heartbeat")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,pTolerance),
           1,CF_FLOAT,"tolerancePct")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
           (int)OFFSETIN(HRS,aTolerance),
           1,CF_FLOAT,"toleranceAbs")) != 0) return cc;
  if ((cc=sealTaggedStruct("HRSr4",sizeof(HRS),1)) != 0) return cc;

  if ((ss=findstruct("HRSr4")) == NULL) return out_of_server_memory;
  HRSStructStruct = ss->s;

  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,ave),
           1,CF_DOUBLE,"average")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,dev),
           1,CF_DOUBLE,"deviation")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,max),
           1,CF_DOUBLE,"maximum")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,min),
           1,CF_DOUBLE,"minimum")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,ts),
           1,CF_INT32,"timeAve")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,tMax),
           1,CF_INT32,"timeMax")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,tMin),
           1,CF_INT32,"timeMin")) != 0) return cc;
  if ((cc=addFieldToStruct("DSUMMARY",
           (int)OFFSETIN(DSUMMARY,cnt),
           1,CF_INT32,"counts")) != 0) return cc;
  if ((cc=sealTaggedStruct("DSUMMARY",sizeof(DSUMMARY),1)) != 0) return cc;

  if ((ss=findstruct("DSUMMARY")) == NULL) return out_of_server_memory;
  SummaryStructStruct = ss->s;

  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,context),
           32,CF_TEXT,"context")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,tag),
           32,CF_TEXT,"tag")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,logger),
           32,CF_TEXT,"logger")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,text),
           256,CF_TEXT,"text")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,priority),
           1,CF_INT32,"priority")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,status),
           1,CF_INT32,"status")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,timeSent),
           1,CF_DOUBLE,"timeSent")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
           (int)OFFSETIN(CLOG,timeLogged),
           1,CF_DOUBLE,"timeLogged")) != 0) return cc;
  if ((cc=sealTaggedStruct("CLOG",sizeof(CLOG),1000)) != 0) return cc;

  if ((ss=findstruct("CLOG")) == NULL) return out_of_server_memory;
  CLOGStructStruct = ss->s;


  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,expName),
           EXPORT_NAME_SIZE,CF_TEXT,"expName")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,prpName),
           PROPERTY_NAME_SIZE,CF_TEXT,"prpName")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,devName),
           DEVICE_NAME_SIZE,CF_TEXT,"devName")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,context),
           CONTEXT_NAME_SIZE,CF_TEXT,"context")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,eqm),
           EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,mode),
           1,CF_INT16,"mode")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,access),
           1,CF_INT16,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,formatIn),
           1,CF_INT16,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,formatOut),
           1,CF_INT16,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,sizeIn),
           1,CF_INT32,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,sizeOut),
           1,CF_INT32,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,strTagIn),
           TAG_NAME_SIZE,CF_TEXT,"strTagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,strTagOut),
           TAG_NAME_SIZE,CF_TEXT,"strTagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,pollingInterval),
           1,CF_INT32,"pollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,heartbeat),
           1,CF_INT32,"heartbeat")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,starttime),
           1,CF_INT32,"starttime")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,dtimestamp),
           1,CF_INT32,"dtimestamp")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,dtimestampMSEC),
           1,CF_INT32,"dtimestampMSEC")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,counter),
           1,CF_INT32,"counter")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,linkStatus),
           1,CF_INT16,"linkStatus")) != 0) return cc;
  if ((cc=addFieldToStruct("CONTBLr4",
           (int)OFFSETIN(ConTblInfo,reserved),
           1,CF_INT16,"reserved")) != 0) return cc;

  if ((cc=sealTaggedStruct("CONTBLr4",sizeof(ConTblInfo),1000)) != 0) return cc;

  if ((ss=findstruct("CONTBLr4")) == NULL) return out_of_server_memory;
  ConTblInfoStructStruct = ss->s;


  if ((cc=addFieldToStruct("WRACCTBL",
           (int)OFFSETIN(WriteAccessQuery,user),
           USERNAME_SIZE,CF_TEXT,"user")) != 0) return cc;
  if ((cc=addFieldToStruct("WRACCTBL",
           (int)OFFSETIN(WriteAccessQuery,addr),
           ADDR_SIZE,CF_TEXT,"addr")) != 0) return cc;
  if ((cc=addFieldToStruct("WRACCTBL",
           (int)OFFSETIN(WriteAccessQuery,prp),
           PROPERTY_NAME_SIZE,CF_TEXT,"prp")) != 0) return cc;
  if ((cc=addFieldToStruct("WRACCTBL",
           (int)OFFSETIN(WriteAccessQuery,dev),
           DEVICE_NAME_SIZE,CF_TEXT,"dev")) != 0) return cc;
  if ((cc=addFieldToStruct("WRACCTBL",
           (int)OFFSETIN(WriteAccessQuery,timestamp),
           1,CF_DOUBLE,"timestamp")) != 0) return cc;

  if ((cc=sealTaggedStruct("WRACCTBL",sizeof(WriteAccessQuery),1000)) != 0) return cc;

  if ((ss=findstruct("WRACCTBL")) == NULL) return out_of_server_memory;
  WriteAccessQueryStruct = ss->s;

  if ((cc=addFieldToStruct("RgnRule",
           (int)OFFSETIN(DeviceRegionItem,dev),
           DEVICE_NAME_SIZE,CF_TEXT,"pattern")) != 0) return cc;
  if ((cc=addFieldToStruct("RgnRule",
           (int)OFFSETIN(DeviceRegionItem,reg),
           TAG_NAME_SIZE,CF_TEXT,"region")) != 0) return cc;
  if ((cc=addFieldToStruct("RgnRule",
           (int)OFFSETIN(DeviceRegionItem,code),
           1,CF_INT32,"code")) != 0) return cc;

  if ((cc=sealTaggedStruct("RgnRule",sizeof(DeviceRegionItem),100)) != 0) return cc;

  /* fix the current header as marked as persistent */
  markStructRegistry();

  feclog("Control Structures Swap information registered");
  logStructureEntries = TRUE;
  return 0;
}
 
