/*
 * File taddtoarc.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
int addtoarc(char *context, char *target, char *prp, char *kwd, char *subsys, char *fmtstr, int len, char *tolstr);
int reloadarc(char* context);
int ReportError(int cc);

char* usageTxt[] = {
  "\nAdds a server to the indicated central archive server\n",
  "taddtoarc <context> <target address> <target property> [/k=<archive keyword> /s=<archive system> [/f=<format> /l=<data length> /t=<tolerance>]",
  "\n\t<target address> can either be prefixed with the server's context or not, in which case the ARCHIVE server",
  "context will be assumed. In lieu of a 'target address' you can specify 'reload' as the second parameter.",
  "In such a case, all other input is ignored and a 'DB.RELOAD' command is sent to the central archive.",
  "The should be the final call to 'addtoarc' in any script which is designed to add entries to a central archiver."
  "The 'keyword' option can (and should) be given to specify the desired archive keyword.  If omitted, a best-guess of <server.property> will be assumed.",
  "The 'system' option can be given to specify which archive system the added server should belong to",
  "The 'format' option can be used to specify a storage format, if omitted, then the registered property datatype will be used",
  "The 'length' option can be used to specify a data length, if omitted length=1 will be assumed (unless the format option is also omitted)",
  "The 'tolerance' option can be used to specify an archive tolerance, append a '%' to indicate percent tolerance"
  "\ne.g.\n\ttaddtoarc TEST /PETRA/BPM Orbit.X",
  "\nor\n\ttaddtoarc TEST /PETRA/BPM Orbit.X /k=PetraOrbit.X",
  "\n\nFinally: make sure to issue a 'reload' command so as to instruct the central archive to re-read its database.",
  "\ne.g.\n\ttaddtoarc TEST reload",
  "\nAlso note that you must be a recognized ARCHIVER administrator to successfully use this tool.",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char* argv[])
{
  int i, len = 0, n = 0;
  char *c, *ctx, *tgt, *prp, *kwd = NULL, *sub = NULL, *fmt = NULL, *tol=NULL;
  char scratch[32], stdkey[256];
  if (argc < 3 || (argc < 4 && stricmp(argv[2],"reload")))
  {
    for (i = 0; i < nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n", usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 200;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  printf("\n");
  ctx = argv[1];
  tgt = argv[2];
  if (!stricmp(tgt, "reload"))
  {
    return ReportError(reloadarc(ctx));
  }
  prp = argv[3];
  for (i = 4; i < argc; i++)
  { /* options given */
    c = argv[i];
    if (*c == '/') c++;
    switch (tolower(*c))
    {
      case 'k': // keyword
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          kwd = c;
        }
        continue;
      case 's': // system
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          sub = c;
        }
        continue;
      case 'd':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          tineDebug = atoi(scratch);
          foregroundTTY = TRUE;
        }
        continue;
      case 'f': // format
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          fmt = c;
        }
        continue;
      case 't': // format
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          tol = c;
        }
        continue;
      case 'l':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          len = atoi(scratch);
        }
        continue;
      default:
        continue;
    }
  }
  if (kwd == NULL)
  {
    char srv[64];
    if (tgt[0] == '/')
    {
      if ((c=strchr(&tgt[1], '/')) != NULL) c++;
    }
    else
    {
      c = tgt;
    }
    if (c != NULL)
    {
      strncpy(srv, c, 32); srv[32] = 0;
      if ((c = strchr(srv, '/')) != NULL) *c = 0;
      sprintf(stdkey, "%.32s.%32s", srv, prp);
      kwd = stdkey;
    }
  }
  //int addtoarc(char* context, char* target, char* prp, char* kwd, char* subsys, char* fmtstr, int len, char* tolstr);
  return ReportError(addtoarc(ctx, tgt, prp, kwd, sub, fmt, len, tol));
}
