/*
 * File taddtocas.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
int addtocas(char *context, char *target, char *almsys, int code, char **lst, int lstsiz);
int ReportError(int cc);

char* usageTxt[] = {
  "\nAdds a server to the indicated central alarm server\n",
  "taddtocas <context> <device server> /system=<alarm system> [/action=mailto:<address> /action=trigger:<event> /action=annotate:<keyword> /code=<alarm code> /remove=<TRUE>]",
  "\n\t<device server> can either be prefixed with the server's context or not, in which case the CAS",
  "context will be assumed.",
  "The 'system' option must be given to specify which alarm system the added server should belong to",
  "You can add multiple 'action' options (e.g. multiple emails and/or events or annotations)",
  "However, when actions are provided, then the 'code' option must also be provided so as to", 
  "indicate which alarm code should trigger the specifed action(s)",
  "An 'action' option must be prefixed with one of 'mailto:', 'trigger:', or 'annotate:'",
  "The 'remove' option (set = TRUE) can be used to removed a server from the CAS.  Note that the", 
  "'system' option must also be specified so as to properly identify the server entry in the CAS database.",
  "\ne.g.\n\ttaddtocas SITE /SERVICE/WD_ACCLXSERVER1 /sys=Infastructure",
  "\nor\n\ttaddtocas SITE /SERVICE/WD_ACCLXSERVER1 /sys=Infastructure /action=mailto:Donald.Trump@Skidrow.com /code=512",
  "\nor\n\ttaddtocas SITE ARCHIVER /sys=Infastructure /action=mailto:Donald.Trump@Skidrow.com /code=999",
  "\n\nFinally: make sure to issue a 'RESET' command to the CAS so as to instruct it to re-read its database.",
  "\ne.g.\n\ttsend /SITE/CAS RESET",
  "\nAlso note that you must be a recognized CAS administrator to successfully use this tool.",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char* argv[])
{
  int i, code=0, n = 0;
  char *c, *ctx, *tgt, *sys=NULL, **lst;
  char scratch[32];
  if (argc < 4)
  {
    for (i = 0; i < nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n", usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 200;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  printf("\n");
  ctx = argv[1];
  tgt = argv[2];
  lst = (char**)alloca((argc - 4) * sizeof(char*));
  for (i = 3; i < argc; i++)
  { /* options given */
    c = argv[i];
    if (*c == '/')
    { // expected ...
      c++;
    }
    else
    { // a quick 'just add the server' without option switches 
      if (i == 3) sys = c;
      continue;
    }
    switch (tolower(*c))
    {
    case 's': // system
      if ((c=strchr(c, '=')) != NULL)
      {
        c++;
        sys = c;
      }
      continue;
    case 'd':
      if ((c = strchr(c, '=')) != NULL)
      {
        c++;
        strncpy(scratch, c, 32);
        tineDebug = atoi(scratch);
        foregroundTTY = TRUE;
      }
      continue;
    case 'a': // action item
      if ((c=strchr(c, '=')) != NULL)
      {
        c++;
        lst[n++] = c;
      }
      continue;
    case 'c':
      if ((c = strchr(c, '=')) != NULL)
      {
        c++;
        strncpy(scratch, c, 32);
        code = atoi(scratch);
      }
      continue;
    case 'r':
      if ((c = strchr(c, '=')) != NULL)
      {
        c++;
        strncpy(scratch, c, 32);
        code = strtobool(scratch) == TRUE ? -1 : 0;
      }
      continue;
    default:
      continue;
    }
  }
  return ReportError(addtocas(ctx,tgt,sys,code,lst,n));
}
