/*
 * File tbufsrv.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file bufsrv.c
 * \brief TINE buffered server documentation 
 *
 */
#include "project.def"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include <math.h>
#ifdef UNIX
#include <dlfcn.h>
#endif
#include "tine.h"
#include "tbufsrv.h"
#include "listener.h"

#ifdef _USRDLL
void TineStartCycler(void);
void TineStopCycler(void);
#endif

#if defined(UNIX) && !defined(SOLARIS)
__attribute__((destructor)) void TBufsrvEpilog(void)
{
  freeMexListenerTable();
}
#endif

#define MAX_BUFFERED_PROPERTIES 512
#define NOTREADY_DEADBAND 500
char gServerWindow[64];
char gServerName[64];
char gEqmName[16];
NAME16 gLastWriter;
UINT32 gLastWriterIp = 0;
int gNumDevices = 0;
int gBufSrvInitialized = FALSE;
int unpushedDataCompletionCode = not_initialized;
typedef struct
{
  double dataTimeStamp;
  int sysStamp;
  int usrStamp;
} DEVSTAMPS;
typedef struct
{
  char eqm[EQM_NAME_SIZE];
  char nam[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  DTYPE dout;
  DTYPE din;
  DTYPE blk;
  float max;
  float min;
  DEVSTAMPS *devStamps;
  double dataTimeStamp;
  int sysStamp;
  int usrStamp;
  int synId; /* property synonym id + 1 */
  int inputChanged;
  int manualArchive;
  int interimCompletionCode;
  int pushedCompletionCode;
  int arrayType;
  int arrayLengthCeiling;
  int acc; /* registered access */
  HNDLRCBP callback;
  NTFRCBP notifier;
  void *cbReference;
  BYTE **stringSpace;
  int *stringSpaceCapacity;
  int notifierId;
  int notifierCalled;
  int needToNotify;
  int notreadyCount;
  int dinStructSize;
  int doutStructSize;
} BufferedProperty;

BufferedProperty bPrps[MAX_BUFFERED_PROPERTIES];
int nBufPrps = 0;
int gNotReadyDeadBand = NOTREADY_DEADBAND;
NAME64 regPrps[MAX_BUFFERED_PROPERTIES];
void (*gBufferedNotifier)(int) = NULL;
int gBufferedNotifierId = 0;
char gPropertyNofified[PROPERTY_NAME_SIZE+4];
char gDeviceNotified[DEVICE_NAME_SIZE+4];

int allocBufferedStorage(BufferedProperty *bp);

int bsMemoryInitialized = FALSE;
int initBufferedServerMemory(void)
{
  if (bsMemoryInitialized) return 0;
  bsMemoryInitialized = TRUE;
  memset(bPrps,0,sizeof(BufferedProperty)*MAX_BUFFERED_PROPERTIES);
  nBufPrps = 0;
  return 0;
}

EXPORT_B void SetNotifierDeadband(int timeInMilliseconds)
{
# define MSEC_PER_CLKTICK 10
  if (timeInMilliseconds < 50) timeInMilliseconds = 50;
  gNotReadyDeadBand = timeInMilliseconds / MSEC_PER_CLKTICK;
}
int areValuesOutOfBounds(int id,BYTE *inpt,int len)
{
  int i;
# define CHECKRANGE(t) {\
    t *vals = (t *)inpt;\
    for (i = 0; i < (int)bPrps[id].din.dArrayLength && i < len; i++)  {\
      if ((vals[i] < bPrps[id].min) || (vals[i] > bPrps[id].max)) return out_of_range;\
    }\
  }
  if (bPrps[id].min == 0 && bPrps[id].max == 0) return 0;
  switch (bPrps[id].din.dFormat)
  {
    case CF_BYTE: 
      CHECKRANGE(BYTE);
      break;
    case CF_UINT16:
      CHECKRANGE(unsigned short);
      break;
    case CF_INT16:
      CHECKRANGE(short);
      break;
    case CF_UINT32: 
      CHECKRANGE(unsigned int);
      break;
    case CF_INT32:
      CHECKRANGE(int);
      break;
    case CF_FLOAT:
      CHECKRANGE(float);
      break;
    case CF_DOUBLE: 
      CHECKRANGE(double);
      break;
  }
  return 0;
}
int hasInputHandler(int pid)
{
  if (gBufferedNotifier != NULL) return TRUE;
  if (pid < 0 || pid >= MAX_BUFFERED_PROPERTIES) return FALSE;
  if (bPrps[pid].callback != NULL) return TRUE;
  if (bPrps[pid].notifier != NULL) return TRUE;
  return FALSE;
}
#define RWACC(a) ((a)&0x03)
int currentPropertyId = -1;
int currentDeviceNumber = 0;
int gNotifierCalled = 0;
int gRequireAccessMatchOnInput = TRUE;
int bufsrveqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,iin,i,aidx=0,iout,p,id,cc=0,dsiz,ssiz,hsiz,doutlen,dinlen,num=1;

  if ((id=GetPropertyId(gEqmName,devProperty)) < 0) return illegal_property;
  if (devName == NULL || strlen(devName) == 0) devnr = 0;
  else 
  {
get_devnr:
    if ((devnr=GetDeviceNumberEx(gEqmName,devName,devProperty)) < 0) 
    {
      char *c = strrchr(&devName[2],'#');
      if (c != NULL)
      {
        *c++ = 0;
        aidx = atoi(c);
        goto get_devnr;
      }
      return illegal_equipment_number;
    }
  }
  if (access & CA_WRITE)
  {
     GetCallerInfo(gEqmName,&gLastWriter,NULL,&gLastWriterIp,NULL,&num);
  }
  if (bPrps[id].synId > 0) 
  { /* property is a synonym, remap to target */
    id = bPrps[id].synId - 1; 
  }
  if ((access & CA_HIST) && bPrps[id].manualArchive) return not_applicable;
  if (bPrps[id].dout.dFormat != CF_TEXT &&
      !(bPrps[id].arrayType & AT_SPECTRUM) &&
      dout->dArrayLength > bPrps[id].dout.dArrayLength - devnr) 
     return dimension_error;
  if (dout->dFormat == CF_STRUCT)
  { /* catch struct formatting mistakes early */
    if (bPrps[id].dout.dFormat != CF_STRUCT) return illegal_format;
    if (strlen(bPrps[id].dout.dTag) > 0 && 
        strncmp(bPrps[id].dout.dTag,dout->dTag,TAG_NAME_SIZE))
    { /* registered with a tag and the outgoing tag is not the same */
      return invalid_structure_tag;
    }
  }
  if (devName[0] == '#' && (bPrps[id].arrayType & AT_SPECTRUM))
  { /* check for specific index */
    aidx = devnr; /* note the array index */
    devnr = 0; /* fix */

  }
  if (din->dArrayLength > 0)
  {
    int bSizeInBytes = 0; /* stored buffer input size in bytes for this item */
    int hasBufferedHandler = FALSE;
    BYTE *inpt = din->data.bptr; /* reference to input data */
    int caccess = RWACC(access | CA_READ);
    int raccess = RWACC(bPrps[id].acc|CA_READ);
    /* if (!(access & CA_WRITE)) return illegal_read_write; */
    if (caccess != raccess)
    {
      feclog("%.64s : caller access %d != registered access %d", bPrps[id].nam, caccess, raccess);
      if (gRequireAccessMatchOnInput) return illegal_read_write;
    }
    //if (RWACC(bPrps[id].acc|CA_READ) != RWACC(access|CA_READ)) return illegal_read_write;
    if (bPrps[id].din.dArrayLength == 0) return illegal_data_size;
    if (!hasInputHandler(id)) return not_implemented;
    /* a fairly stringent condition : */
    if (din->dFormat != bPrps[id].din.dFormat)
    { /* not the format registered */
      if (IsNumberFormat(bPrps[id].din.dFormat))
      { /* maybe we can cast ? */
        int fsiz = GetFormatSize(bPrps[id].din.dFormat);
        if ((inpt=alloca(din->dArrayLength*fsiz)) == NULL) return out_of_local_memory;
        cc = GetValuesAsAny(din, inpt, bPrps[id].din.dFormat, fsiz, din->dArrayLength, 0);
        if (cc != 0) return cc;
      }
      else
      {
        return illegal_format;
      }
    }
    if (strlen(bPrps[id].din.dTag) > 0 && 
        strncmp(bPrps[id].din.dTag,din->dTag,TAG_NAME_SIZE))
    { /* registered with a tag and the incoming tag is not the same */
      if (din->dFormat == CF_STRUCT) return invalid_structure_tag;
    }
    hsiz = getFormatHeaderSize(bPrps[id].din.dFormat);
    if (bPrps[id].din.dFormat == CF_IMAGE ||
        bPrps[id].din.dFormat == CF_GSPECTRUM) hsiz += sizeof(UINT32) + sizeof(BYTE*);
    iin = devnr;
    if (bPrps[id].arrayType & AT_SPECTRUM || hsiz > 0) 
    {
      iin *= bPrps[id].din.dArrayLength;
      bSizeInBytes = (int)bPrps[id].din.dArrayLength * fmtsizeof(bPrps[id].din.dFormat);
      bSizeInBytes += hsiz;
    }
    else
    {
      bSizeInBytes = fmtsizeof(bPrps[id].din.dFormat);
      if (bPrps[id].din.dArrayLength == 1) iin = 0;
    }
    /* do range checking */
    if (areValuesOutOfBounds(id,inpt,din->dArrayLength)) return out_of_range;
    if (access & (CA_FIRST|CA_SAVERESTORE)) 
    { /* this contract is coming in */
      if (bPrps[id].inputChanged || bPrps[id].notifierCalled || gNotifierCalled)
      { /* a new contract setting the same property which is already in progress (completion not yet set) */
        return operation_busy;
      }
      bPrps[id].interimCompletionCode = not_ready;
      bPrps[id].notreadyCount = 0;
    }
    else 
    { /* either not_ready or set by SetBufferedCompletion() */
      if (bPrps[id].interimCompletionCode == not_ready) bPrps[id].notreadyCount++;
      if (bPrps[id].notreadyCount > gNotReadyDeadBand) 
      {
        feclog("Write call to Property %s appears to be hung : RESET",devProperty);
        SetBufferedCompletion(not_posted);
      }
      ccerr(bPrps[id].interimCompletionCode);
    }
    
    dinlen = din->dArrayLength;
    if (dinlen > (int)bPrps[id].din.dArrayLength) dinlen = (int)bPrps[id].din.dArrayLength;
    dsiz = dinlen * fmtsizeof(bPrps[id].din.dFormat);
    dsiz += hsiz;
   
    p = bSizeInBytes * devnr;

    if (bPrps[id].din.dFormat == CF_STRUCT) 
    { /* registered as a structure */
      if ((ssiz=bPrps[id].dinStructSize) <= 0)
      { /* not yet known */
        if ((cc=allocBufferedStorage(&bPrps[id])) != 0) ccerr(cc);
        ssiz = bPrps[id].dinStructSize;
      }
      p *= ssiz; /* adjust the buffer boundary */
      dsiz *= ssiz; /* adjust the data size request */
    }
    else if (bPrps[id].din.dFormat == CF_STRING || bPrps[id].din.dFormat == CF_KEYVALUE)
    {
      dsiz = sizeof(char*);
      p = dsiz * devnr;
      dsiz *= dinlen;
    }
    bPrps[id].inputChanged = 1;
    if (memcmp(&bPrps[id].din.data.bptr[p],inpt,dsiz)) 
    {
      bPrps[id].inputChanged = -1;
      if (access & CA_SAVERESTORE) bPrps[id].needToNotify = TRUE;
    }
    if (bPrps[id].din.dFormat == CF_IMAGE)
    {
      DIMAGE *src = (DIMAGE *)din->data.vptr;
      DIMAGE *tgt = (DIMAGE *)&bPrps[id].din.data.bptr[p];
      hsiz = getFormatHeaderSize(CF_IMAGE);
      memcpy(tgt,src,hsiz);
      tgt->frameBufferSize = bPrps[id].din.dArrayLength;
      tgt->frameBuffer = &bPrps[id].din.data.bptr[p+hsiz+sizeof(UINT32)+sizeof(BYTE *)];
      if (src->frameBufferSize > tgt->frameBufferSize) src->frameBufferSize = tgt->frameBufferSize;
      memcpy(tgt->frameBuffer,src->frameBuffer,src->frameBufferSize);
    }
    else if (bPrps[id].din.dFormat == CF_GSPECTRUM)
    {
      DGSPECTRUM *src = (DGSPECTRUM *)din->data.vptr;
      DGSPECTRUM *tgt = (DGSPECTRUM *)&bPrps[id].din.data.bptr[p];
      int fsiz = fmtsizeof(CF_GSPECTRUM);
      hsiz = getFormatHeaderSize(CF_GSPECTRUM);
      memcpy(tgt, src, hsiz);
      tgt->spectBufferSize = bPrps[id].din.dArrayLength;
      tgt->spectBuffer = (float *)(&bPrps[id].din.data.bptr[p + hsiz + sizeof(UINT32) + sizeof(BYTE*)]);
      if (src->spectBufferSize > tgt->spectBufferSize) src->spectBufferSize = tgt->spectBufferSize;
      memcpy(tgt->spectBuffer, src->spectBuffer, src->spectBufferSize * fsiz);
    }
    else
    { /* not an image: copy the incoming bytes into input buffer */
      memset(&bPrps[id].din.data.bptr[p],0,bPrps[id].din.dArrayLength);
      memcpy(&bPrps[id].din.data.bptr[p],inpt,dsiz);
    }
    currentPropertyId = id;
    currentDeviceNumber = devnr;
    if (bPrps[id].notifier != NULL)
    { /* a property specific notifier registered (e.g. labview) */
      hasBufferedHandler = TRUE;
      strncpy(bPrps[id].dev,devName,DEVICE_NAME_SIZE);
      bPrps[id].notifierCalled = -1; /* cleared in SetBufferedCompletion() */
      bPrps[id].notifier(bPrps[id].notifierId);
      return not_ready;
    }
    if (gBufferedNotifier != NULL)
    { /* general notifier */
      hasBufferedHandler = TRUE;
      gNotifierCalled = -1;
      strncpy(gPropertyNofified,devProperty,PROPERTY_NAME_SIZE);
      strncpy(gDeviceNotified,devName,DEVICE_NAME_SIZE);
      gBufferedNotifier(gBufferedNotifierId);
      return not_ready;
    }
    if (bPrps[id].callback != NULL) 
    {
      hasBufferedHandler = TRUE;
      strncpy(bPrps[id].dev,devName,DEVICE_NAME_SIZE);
      bPrps[id].notifierCalled = -1;
      if (bPrps[id].cbReference != NULL)
        cc = ((XHNDLRCBP)bPrps[id].callback)(devProperty,devName,bPrps[id].cbReference);
      else
        cc = bPrps[id].callback();
      bPrps[id].notifierCalled = 0;
      bPrps[id].interimCompletionCode = 0;
    }
    bPrps[id].inputChanged = 0; /* reset here no matter what */
    if (!hasBufferedHandler) cc = not_initialized;
  } /* din->dArrayLength > 0 branch */
err:
  if (cc) return cc;
  if (dout->dArrayLength > 0)
  {
    int doutbytes;
    if (bPrps[id].pushedCompletionCode != 0) return bPrps[id].pushedCompletionCode;
    if (bPrps[id].arrayLengthCeiling > 0 && 
        bPrps[id].arrayLengthCeiling < (int)dout->dArrayLength)
      dout->dArrayLength = bPrps[id].arrayLengthCeiling;
    doutlen = bPrps[id].dout.dArrayLength;
    hsiz = getFormatHeaderSize(bPrps[id].dout.dFormat); /* format header size (if any) */
    if (bPrps[id].dout.dFormat == CF_IMAGE ||
        bPrps[id].dout.dFormat == CF_GSPECTRUM) hsiz += sizeof(UINT32) + sizeof(BYTE*);
    doutbytes = doutlen * fmtsizeof(bPrps[id].dout.dFormat);
    doutbytes += hsiz;
    if (bPrps[id].dout.dFormat == CF_STRING) doutbytes = doutlen * sizeof(char *);
    iout = i = devnr;
    dout->dTimeStamp = bPrps[id].dataTimeStamp;
    dout->sysStamp = bPrps[id].sysStamp;
    dout->dStamp = bPrps[id].usrStamp;
    if (bPrps[id].arrayType & AT_SPECTRUM || hsiz > 0 || bPrps[id].dout.dFormat == CF_TEXT) 
    { /* set the device specific data stamps */
      dout->dTimeStamp = bPrps[id].devStamps[i].dataTimeStamp;
      dout->sysStamp = bPrps[id].devStamps[i].sysStamp;
      dout->dStamp = bPrps[id].devStamps[i].usrStamp;
      /* establish the correct 'start' index */
      i *= doutlen;
      doutlen *= gNumDevices;
      if (aidx > 0) i = aidx; /* specific index requested */
    }
    else
    { /* MCAs have different storage as SPECTRUM arrays (and there won't be a header) */
      doutbytes = fmtsizeof(bPrps[id].dout.dFormat);
    }
    switch (bPrps[id].dout.dFormat)
    {
      case CF_SHORT:
        if ((cc=PutValuesFromShortEx(dout,bPrps[id].dout.data.sptr,doutlen,i)) != 0) return cc;
        break;
      case CF_LONG:
        if ((cc=PutValuesFromLongEx(dout,bPrps[id].dout.data.lptr,doutlen,i)) != 0) return cc;
        break;
      case CF_FLOAT:
        if ((cc=PutValuesFromFloatEx(dout,bPrps[id].dout.data.fptr,doutlen,i)) != 0) return cc;
        break;
      case CF_DOUBLE:
        if ((cc=PutValuesFromDoubleEx(dout,bPrps[id].dout.data.dptr,doutlen,i)) != 0) return cc;
        break;
      default:
        if (dout->dFormat != bPrps[id].dout.dFormat) return illegal_format;
        ssiz = 1; /* structure size (see below) */
        if (bPrps[id].dout.dFormat == CF_STRUCT) 
        { /* need to get the structure size */
          if ((ssiz=bPrps[id].doutStructSize) <= 0)
          { /* not yet known */
            if ((cc=allocBufferedStorage(&bPrps[id])) != 0) return cc;
            ssiz = bPrps[id].doutStructSize;
          }
        }
        /* doutbytes = bytes per element in the buffered storage */
        i = iout * doutbytes * ssiz; /* we're looking at bytes here ... */
        if (dout->dArrayLength > (UINT32)bPrps[id].dout.dArrayLength)
        { /* check request versus storage */
          dout->dArrayLength = (UINT32)bPrps[id].dout.dArrayLength;
        }
        /* now get the right amount to copy */
        doutbytes = dout->dArrayLength * fmtsizeof(bPrps[id].dout.dFormat);
        doutbytes += hsiz;
        if (bPrps[id].dout.dFormat == CF_STRING)
        {
          doutbytes = dout->dArrayLength * sizeof(char *);
        }
        if (bPrps[id].dout.dFormat == CF_IMAGE || bPrps[id].dout.dFormat == CF_GSPECTRUM)
        { /* special case: for IMAGE just memcpy the header parts */
          doutbytes = hsiz;
        }
        memcpy(dout->data.bptr,&bPrps[id].dout.data.bptr[i],doutbytes*ssiz);
        break;
    }
  }
  return 0;
}
int isScheduleDecorated(char *prpName)
{
  char *c, pname[PROPERTY_NAME_SIZE+4];
  if (prpName == NULL) return FALSE;
  strncpy(pname, prpName, PROPERTY_NAME_SIZE); pname[PROPERTY_NAME_SIZE] = 0;
  strupr(pname);
  if ((c=strrchr(pname,'.')) == NULL) return FALSE;
  return strstr(c,".SCH") == NULL ? FALSE : TRUE;
}
/** 
 * \brief Manually supplies data to the local history subsystem for the given property
 * 
 * The data buffer given is taken to represent multiple data segments which are appended
 * to the local history sequentially in time intervals specified by the 'timeIntvl' 
 * parameter (see below).  In addition the 'shot number' interval (e.g. pulse number, 
 * or cycle number interval) can also be specified in order to properly tag the data
 * being supplied to the local history system.  If the supplied data size exactly 
 * matches the registered property length then a single data segment will be added to
 * the local history.  Otherwise the property size given must be a multiple of the
 * registered data size of the property specified.
 *
 * \param prpName is the property name from whose input buffer the associated data are to be read.
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer from which the associated data are to be read.
 * \param prpData is a reference to the data buffer from which data are read and passed on (pushed)
 *        to the local history subsystem.
 * \param prpSiz is the number of elements to be read from the given data buffer.  A positive value
 *        indicates cronologically ascending array elements (oldest to most recent).  A negative
 *        value indicates cronologically descending array elements (most recent first).
 * \param timeIntvl is the time interval in milliseconds which is used to separate the data 
 *        segments found in the given data buffer.
 * \param shotIntvl is the shot interval which is used to separate the data segments found in the
 *        given data buffer.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:
 *
 * \include eg_pushHistoryData.c
 */
EXPORT_B int pushHistoryData(char *prpName,char *devName,BYTE *prpData,int prpSiz,double timeIntvl,int shotIntvl)
{
  int cc=0, id, i, k, fsiz, nloops, cronAscending=TRUE;
  DTYPE dout;
  double ts = MakeDataTimeStamp();
  int ss = GetSystemDataStamp(-1);

  if (prpData == NULL) return buffer_too_small;
  if (prpSiz == 0) prpSiz = 1;
  if (prpSiz < 0) 
  { /* most recent 1st, oldest last */
    cronAscending = FALSE;
    prpSiz = -prpSiz;
  }
  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return illegal_property;
  if ((i=GetDeviceNumberEx(gEqmName,devName,prpName)) < 0) return -i;
  if (bPrps[id].synId > 0) 
  { /* property is a synonym, remap to target */
    id = bPrps[id].synId - 1; 
  }
  if ((prpSiz % bPrps[id].dout.dArrayLength) != 0) return dimension_error;
  bPrps[id].manualArchive = TRUE;

  DTYPEZERO(dout);
  dout.dFormat = bPrps[id].dout.dFormat;
  dout.dArrayLength = bPrps[id].dout.dArrayLength;

  nloops = prpSiz/bPrps[id].dout.dArrayLength;
  fsiz = bPrps[id].dout.dArrayLength * GetFormatSize(bPrps[id].dout.dFormat);
  if (timeIntvl <= 0 || shotIntvl <= 0) nloops = 1;
 
  timeIntvl /= 1000.0;
  ts -= (nloops - 1) * timeIntvl;
  ss -= (nloops - 1) * shotIntvl;

  for (k=0; k<nloops; k++)
  {
    i = cronAscending ? k : (nloops-1-k);
    dout.data.bptr = &prpData[i*fsiz];
    dout.dTimeStamp = ts + k * timeIntvl;
    dout.sysStamp = ss + k * shotIntvl;
    if ((cc=AppendHistoryData(gEqmName,prpName,devName,&dout)) != 0) break;
  }
  return cc;
}
EXPORT_B int SetBufferedPropertyCompletion(char *prpName, int compCode)
{
  int id = -1;
  if (prpName != NULL && strlen(prpName) > 0) id = GetPropertyId(gEqmName,prpName);
  if (id < 0) id = currentPropertyId;
  gNotifierCalled = 0;
  if (id < 0) 
  {
    return illegal_property;
  }
  bPrps[id].interimCompletionCode = compCode;
  bPrps[id].inputChanged = 0;
  bPrps[id].notifierCalled = 0;
  return 0;
}
EXPORT_B int SetBufferedCompletion(int compCode)
{
  return SetBufferedPropertyCompletion(NULL,compCode);
}
/**
 * \brief Sets the completion code to be applied for unpushed property data for the given property.
 * 
 * \param prpName is the targeted property name
 * \param compCode is the completion code to apply when a caller accesses the target property prior
 *  to any data being pushed into the property's buffer.
 *
 * \return 0 upon success or illegal_property when the target property is not found in the registry.
 */
EXPORT_B int SetBufferedPropertyStatus(char *prpName, int compCode)
{
  int id = -1;
  if (prpName != NULL && strlen(prpName) > 0) id = GetPropertyId(gEqmName,prpName);
  if (id < 0) id = currentPropertyId;
  if (id < 0)  return illegal_property;
  bPrps[id].pushedCompletionCode = compCode;
  return 0;
}
/**
 * \brief Sets the completion code to be applied for any unpushed data (default: not_initialized)
 * 
 * \param compCode is the completion code to apply when a caller accesses a property prior
 *  to any data being pushed into its buffer.
 *
 */
EXPORT_B void SetUnpushedDataCompletionCode(int compCode)
{
  if (compCode < 0) return;
  unpushedDataCompletionCode = compCode;
}
/**
 * \brief Gets the completion code to be applied for any unpushed data (default: not_initialized)
 * 
 * \return the completion code to be applied for any unpushed data.
 */
EXPORT_B int GetUnpushedDataCompletionCode(void)
{
  return unpushedDataCompletionCode;
}
static int exitOnInitFailure = FALSE;
/** 
 * \brief Determines whether the server process issues a hard exit upon an initialiation error
 *
 * A value of TRUE will only be accepted if he server has not yet been initialized
 *
 * \param value is the desired setting
 *
 * \return the current setting.
 */
EXPORT_B int SetExitOnInitFailure(int value)
{
  if (!gBufSrvInitialized)
  {
    exitOnInitFailure = value;
  }
  return exitOnInitFailure;
}
static int hasInitializedOnce = FALSE;
/** 
 * \brief Attaches the TINE server according to the input given
 *
 * \param srvExportName is the desired export name.  If this is omitted (i.e. a zero-length 
 *        string is passed, then the export name must be obtained from an associated 'exports.csv'
 *        in the local FEC database repository.
 * \param srvEqmName is the desired local equipment module name.  If this is omitted (i.e. a zero-length
 *        string is passed, then the local equipment module name is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 * \param ndevices is the number of devices (instances) associated with this device server.
 *        If this is omitted (i.e. 0 is passed), then the number of devices is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 * \param tmr is a backgound task which should be called by the TINE engine on a regular basis
 *        (timer interval determined by the 5th input parameter).  The task function should have
 *        the prototype void (*tmr)(void);
 * \param tmrInterval is the timer interval in milliseconds which determines how often the timer
 *        backgound task will be called.
 *
 * A call to AttachServerEx() or AttachServer() serves to initialize the device server.  
 * Theoretically, you can call AttachServer() with all 'NULL'
 * parameters, in which case all registration information is obtained from the 'exports.csv' 
 * file (see <a href="csvFiles.html"><b><i>Server Startup csv-Files</i></b></a>.
 * Note however, that if you take this strategy, you should take care to avoid a lock-out
 * situation regarding the name resolution.  Namely, if all initialization names are to reside
 * in 'fecid.csv' (FEC name) and 'exports.csv' (device server name, local equipment module name,
 * and properties) then the fecid.csv should not make use of an 'Export_Name' column or a 
 * 'Local_Name' column.  That is, the fecid.csv file should only contain one entry, referring
 * to the server process in question.  In cases where more than one FEC process are running
 * on the same computer, each process should have its own fecid.csv in its working directory
 * when the export.csv file cannot make use of an 'Export_Name' column in order to latch
 * the appropriate device server.  Of course, if AttachServer() provides an 'Export Name' in
 * the first parameter, then there is no ambiguity.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:  Attach to server and set the server's export name (requires exports.csv file to
 *              set the local equipment module name and number of modules):
 *
 * \include eg_AttachServer0.c
 *
 * \b Example:  Attach to server and set the server's export name, local equipment module name and
 *              number of modules (does not read exports.csv file!):
 *
 * \include eg_AttachServer1.c
 */
EXPORT_B int AttachServerEx(char *srvExportName,char *srvEqmName,int ndevices,void (*tmr)(void),int tmrInterval) 
{
  int i, n, id, sid, mxid=1, cc, initfec=TRUE, keepAlive=FALSE, acqEqmName=FALSE, useXMLdb=FALSE;
  int nprps = MAX_BUFFERED_PROPERTIES, hasprplst = FALSE, hasdevlst = FALSE;
  char *lclName = NULL;
  ExportPropertyListStruct pls;
  char *c, *ctx=NULL, *srv=NULL, prpBaseName[PROPERTY_NAME_SIZE+4];
  char *opts=NULL, sExportName[128],sEqmName[128];
  NAME8 eqmname;
  int useFgTTY = GetInterpretConsoleCommands();  

  SetRunAsServer();
  SetSystemLogging(TRUE); sExportName[0] = sEqmName[0] = 0;
  feclog("Attach Server: export name %s, eqm name %s, device capacity %d (process id: %d)",
          (srvExportName == NULL ? "(null)" : srvExportName), 
          (srvEqmName == NULL ? "(null)" : srvEqmName), 
          ndevices,SystemGetProcessId());
  initBufferedServerMemory();
  if (srvExportName != NULL && strlen(srvExportName) != 0)
  { /* looking for an explicit export name */
    strncpy(sExportName,srvExportName,128);
    if (sExportName[0] == '/')
    { /* context given */
      ctx = &sExportName[1];
      if ((srv=strchr(ctx,'/')) != NULL) *srv++ = 0;
    }
    else
    {
      srv = sExportName;
    }
    if ((c=strchr(srv,'/')) != NULL)
    {
      *c++ = 0;
      opts = c;
    }
    feclog("Attach Server to exported server name %.32s", srv);
    if (srvEqmName != NULL) 
    {
      memset(sEqmName,0,EQM_NAME_SIZE);
      strncpy(sEqmName,srvEqmName,EQM_NAME_SHORTSIZE);
    }
  }
  else if (srvEqmName != NULL && strlen(srvEqmName) != 0)
  { /* looking for an explicit EQM name */
    strncpy(sEqmName, srvEqmName, 128); sEqmName[127] = 0;
    if ((c=strchr(sEqmName,'/')) != NULL)
    {
      *c++ = 0;
      opts = c;
    }
    feclog("Attach Server to local equipment module %.6s", sEqmName);
  }
  else
  { /* empty arguments: must allow a fec.xml readout */
    acqEqmName = TRUE;
    keepAlive = TRUE;
    useXMLdb = TRUE;
  }
  SetAppVersion(bufsrvMajor, bufrvMinor, bufsrvRevision);
  if (ctx != NULL) SetExportedContext(srvEqmName,ctx);
  if (opts != NULL)
  { /* were options given ? */
    strupr(opts);
    if (strstr(opts,"INIT") != NULL && (c=strchr(opts,'=')) != NULL)
    {
       if (strstr(c,"SRV") != NULL || strstr(c,"SVR") != NULL) initfec = FALSE;
       if (strstr(c,"FEC") != NULL) initfec = TRUE;
       if (strstr(c,"KEEP") != NULL) keepAlive = TRUE;
    }
  }
  if (!hasInitializedOnce || initfec)
  { /* a LabView restart doesn't unload the library ! */
    char *ptr = getenv("TINE_UNPUSHED_DATA_STATUS");
    if (ptr != NULL)
    {
      SetUnpushedDataCompletionCode(atoi(ptr));
    }
    ptr = getenv("TINE_INPUT_ACCESS_MUST_MATCH");
    if (ptr != NULL)
    {
      gRequireAccessMatchOnInput = !stricmp(ptr, "TRUE");
    }
    if (!keepAlive)
    {
      stopAllListeners();
#     ifdef _USRDLL
      TineStopCycler();
#     endif
      FreeEQPModules();
      gBufSrvInitialized = FALSE;
    }
    MinPollingRate = 20;
    gNotifierCalled = FALSE;
    if (!useXMLdb) IgnoreFecXmlDatabase();
    if ((cc=SystemInit(useFgTTY)) != 0) 
    { 
#     ifndef TBUFSRVDLL_EXPORTS
      char errstr[64];
      tineDebug = 1;
      feclog("init error: %.32s\n",GetLastLinkError((short)cc,errstr)); 
#     endif
      if (exitOnInitFailure) exit(1);
      return cc;
    }
    hasInitializedOnce = TRUE;
  }
  if (acqEqmName)
  {
    n = 1;
    GetEquipmentModuleList(&eqmname,&n);
    if (n < 1)
    {
      feclog("Attach Server to (null entry)");
      return argument_list_error;
    }
    strncpy(sEqmName,eqmname.name,6);
  }
  resetLastMinuteInitialization();

  memset(gServerName,0,64); memset(gEqmName,0,16);
  if (srv != NULL) strncpy(gServerName,srv,EXPORT_NAME_SIZE);
  if (strlen(sEqmName) > 0) strncpy(gEqmName,sEqmName,EQM_NAME_SHORTSIZE);

  if (strlen(gServerName) > 0 && strlen(gEqmName) > 0)
  { /* all arguments passed */
    gNumDevices = ndevices;
    cc = RegisterEquipmentModule(gServerName,gEqmName,ndevices,bufsrveqm,NULL,tmr,tmrInterval,NULL);
    feclog("register %.32s (%.6s) %d devices : %.32s",gServerName,gEqmName,ndevices,cc2str(cc));
  }
  else
  { 
    if (strlen(gServerName) > 0)
    { /* just the export name was passed, find the local name */
      if ((lclName=getLocalName(gServerName)) == NULL) return non_existent_elem;
      strncpy(gEqmName, lclName, EQM_NAME_SHORTSIZE) ;
    }
    if (strlen(gEqmName) > 0)
    {
      cc = RegisterEquipmentModule(NULL,gEqmName,0,bufsrveqm,NULL,tmr,tmrInterval,NULL);
      feclog("register (%.6s) : %.32s",gEqmName,cc2str(cc));
    }
    else
    {
      feclog("export and local names unknown : cannot register equipment module !");
      return argument_list_error;
    }
  }
  hasprplst = hasDevicePropertyList(gEqmName,NULL);
  if ((cc=GetRegisteredPropertyList(gEqmName,regPrps,&nprps)) != 0) 
  {
    feclog("GetRegisteredPropertyList : %.32s",cc2str(cc));
    return cc;
  }
  if ((n=GetNumberRegisteredDevices(gEqmName)) < 0) n = ndevices > 0 ? ndevices : 1;
  gNumDevices = n;
  if (nprps > 0)
  { /* note: regPrps is just a list in 'some' order => property 'id' is not sequential! */
    int idm = -1;
    for (i=0; i < nprps; i++)
    { /* find the maximum assigned property id */
      idm = GetPropertyId(gEqmName, regPrps[i].name);
      if (idm > mxid) mxid = idm;
    }
    mxid++;
  }
  feclog("determined %d registered properties, %d registered devices",nprps,gNumDevices);
  for (i=0; i<nprps; i++)
  { /* note: regPrps is just a list in 'some' order => property 'id' is not sequential! */
    if ((id=GetPropertyId(gEqmName,regPrps[i].name)) < 0) 
    {
      feclog("property %.64s not found in registry for (%.6s)",regPrps[i].name,gEqmName);
      continue;
    }
    if ((cc=GetRegisteredPropertyListStruct(gEqmName,regPrps[i].name,&pls)) != 0) 
    {
      feclog("property %.64s not found in registry for (%.6s)",regPrps[i].name,gEqmName);
      continue;
    }
    hasdevlst = hasPropertyDeviceList(gEqmName,regPrps[i].name);
    if (hasprplst) 
    { /* force separate storage for each device  */
      pls.prpArrayType &= ~AT_CHANNEL; /* can't be a channel array */
      pls.prpArrayType |= AT_SPECTRUM; 
    }
    if (hasdevlst && !(pls.prpArrayType & AT_CHANNEL)) pls.prpArrayType |= AT_SPECTRUM; 

    strncpy(bPrps[id].nam,pls.prpName,PROPERTY_NAME_SIZE);
    bPrps[id].max = (float)pls.egu.max;
    bPrps[id].min = (float)pls.egu.min;
    bPrps[id].dout.dFormat = pls.prpFormat;
    bPrps[id].dout.dArrayLength = pls.prpSize;
    strncpy(bPrps[id].dout.dTag,pls.prpFormatTag.name,TAG_NAME_SIZE);
    bPrps[id].din.dFormat = pls.prpFormatIn;
    bPrps[id].din.dArrayLength = pls.prpSizeIn;
    strncpy(bPrps[id].din.dTag,pls.prpFormatTagIn.name,16);
    bPrps[id].arrayType = pls.prpArrayType;
    bPrps[id].dataTimeStamp = 0;
    bPrps[id].inputChanged = 0;
    bPrps[id].pushedCompletionCode = unpushedDataCompletionCode;
    bPrps[id].acc = pls.prpAccessMode;
    allocBufferedStorage(&bPrps[id]);

    nBufPrps++;
 
    if (isScheduleDecorated(pls.prpName))
    { /* has been decorated as a 'scheduled' property */
      strncpy(prpBaseName, pls.prpName, PROPERTY_NAME_SIZE); prpBaseName[PROPERTY_NAME_SIZE] = 0;
      if ((c=strrchr(prpBaseName,'.')) != NULL)
      { /* this had better be true ! */
        *c = 0;
        if ((sid=GetPropertyId(gEqmName,prpBaseName)) < 0)
        { /* not yet registered => assign new max property id */
          sid = RegisterBufferedPropertyEx(prpBaseName,pls.prpSizeIn,pls.prpFormatIn,pls.prpSize,pls.prpFormat,
                                           (float)pls.egu.max,(float)pls.egu.min,pls.egu.units,
                                           pls.prpAccessMode,pls.prpDescription,mxid);
          if (sid != mxid)
          {
            feclog("error in synonym property registration !");
          }
          mxid++;
        }
        bPrps[sid].synId = id + 1;
        feclog("property %.64s (%d) is a synonym for property %.64s (%d): data will be mapped",
          prpBaseName,sid,pls.prpName,id);
      }
    }  
  }
  feclog("registered %d buffered properties",nBufPrps);
# ifdef _USRDLL
  TineStartCycler();
# endif
  gBufSrvInitialized = TRUE;
  return 0;
}
int allocBufferedStorage(BufferedProperty *bp)
{
  static int last_cc=0;
  static char last_prp[PROPERTY_NAME_SIZE] = "";
  int cc = 0, dsiz, nelems;
  if (bp->dout.data.bptr != NULL) goto allocInput;
  nelems = bp->arrayType & AT_SPECTRUM ? gNumDevices : 1;
  dsiz = bp->dout.dArrayLength * fmtsizeof(bp->dout.dFormat);
  dsiz += getFormatHeaderSize(bp->dout.dFormat);
  switch (bp->dout.dFormat)
  { /* some special cases */
    case CF_TEXT: /* looks like a SPECTRUM array type */
      nelems = gNumDevices;
      break;
    case CF_STRUCT: /* output struct data */
      if ((bp->doutStructSize=GetStructSize(bp->dout.dTag)) > 0)
      { /* MUST be registered! */
        dsiz *= bp->doutStructSize;
      }
      else 
        ccerr(invalid_structure_tag);
      break;
    case CF_IMAGE:
    case CF_GSPECTRUM:
      dsiz += sizeof(UINT32) + sizeof(BYTE*);
      break;
    case CF_STRING:
    { /* allocate space for the string pointers */
      int i;
      dsiz = bp->dout.dArrayLength * sizeof(char*);
      /* and an initial string space capacities */
      bp->stringSpace = (BYTE**)calloc(nelems * bp->dout.dArrayLength, dsiz);
      bp->stringSpaceCapacity = (int*)calloc(nelems * bp->dout.dArrayLength, sizeof(int));
      if (bp->stringSpace == NULL) ccerr(out_of_server_memory);
      for (i = 0; i < (int)(nelems * bp->dout.dArrayLength); i++)
      {
        bp->stringSpace[i] = (BYTE*)calloc(nelems, STRING_STD_CAPACITY);
        bp->stringSpaceCapacity[i] = STRING_STD_CAPACITY;
      }
    }
    break;
  }
  if (dsiz > 0 && (bp->dout.data.bptr=calloc(nelems,dsiz)) == NULL)
    ccerr(out_of_server_memory);
  if ((bp->devStamps = (DEVSTAMPS *)calloc(nelems,sizeof(DEVSTAMPS))) == NULL)
    ccerr(out_of_server_memory);
allocInput:
  if (bp->din.data.bptr) ccerr(0); /* already done */
  if (bp->arrayType & AT_CHANNEL && bp->din.dArrayLength == 1) 
  { /* allow individual storage for the most recent input for each device */
    nelems = bp->dout.dArrayLength;
  }
  dsiz = bp->din.dArrayLength * fmtsizeof(bp->din.dFormat);
  dsiz += getFormatHeaderSize(bp->din.dFormat);
  switch (bp->din.dFormat)
  { /* special cases*/
    case CF_STRUCT:
      if ((bp->dinStructSize = GetStructSize(bp->din.dTag)) > 0)
      {
        dsiz *= bp->dinStructSize;
      }
      else 
        ccerr(invalid_structure_tag);
      break;
    case CF_IMAGE:
    case CF_GSPECTRUM:
      dsiz += sizeof(UINT32) + sizeof(BYTE*);
      break;
  }
  if (dsiz > 0 && (bp->din.data.bptr=calloc(nelems,dsiz)) == NULL)
    ccerr(out_of_server_memory);
err:
  if (cc && cc != last_cc && strncmp(last_prp,bp->nam,PROPERTY_NAME_SIZE)) 
  {
    char fstr[64];
    memset(fstr,0,64);
    sprintf(fstr,"%.32s",GetFormatAsString(bp->dout.dFormat));
    if (bp->dout.dFormat == CF_STRUCT) 
    {
      strcat(fstr,"."); strncat(fstr,bp->dout.dTag,TAG_NAME_SIZE);
    }
    feclog("allocBufferedStorage %.64s %d %s values : %32s",bp->nam,bp->dout.dArrayLength,fstr,cc2str(cc));
    last_cc = cc;
    strncpy(last_prp,bp->nam,PROPERTY_NAME_SIZE);
  }
  return cc;
}
/** 
 * \brief Attaches the TINE server according to the input given
 *
 *
 * \param srvExportName is the desired export name.  If this is omitted (i.e. a zero-length 
 *        string is passed, then the export name must be obtained from an associated 'exports.csv'
 *        in the local FEC database repository.
 * \param srvEqmName is the desired local equipment module name.  If this is omitted (i.e. a zero-length
 *        string is passed, then the local equipment module name is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 * \param ndevices is the number of devices (instances) associated with this device server.
 *        If this is omitted (i.e. 0 is passed), then the number of devices is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 *
 * A call to AttachServerEx() or AttachServer() serves to initialize the device server.  It is 
 * analogous to setting the server ActiveX control's enabled property to 'true', where the
 * ActiveX interface is concerned.  Theoretically, you can call AttachServer() with all 'NULL'
 * parameters, in which case all registration information is obtained from the 'exports.csv' 
 * file (see <a href="csvFiles.html"><b><i>Server Startup csv-Files</i></b></a>.
 * Note however, that if you take this strategy, you should take care to avoid a lock-out
 * situation regarding the name resolution.  Namely, if all initialization names are to reside
 * in 'fecid.csv' (FEC name) and 'exports.csv' (device server name, local equipment module name,
 * and properties) then the fecid.csv should not make use of an 'Export_Name' column or a 
 * 'Local_Name' column.  That is, the fecid.csv file should only contain one entry, referring
 * to the server process in question.  In cases where more than one FEC process are running
 * on the same computer, each process should have its own fecid.csv in its working directory
 * when the export.csv file cannot make use of an 'Export_Name' column in order to latch
 * the appropriate device server.  Of course, if AttachServer() provides an 'Export Name' in
 * the first parameter, then there is no ambiguity.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:  Attach to server and set the server's export name (requires exports.csv file to
 *              set the local equipment module name and number of modules):
 *
 * \include eg_AttachServer0.c
 *
 * \b Example:  Attach to server and set the server's export name, local equipment module name and
 *              number of modules (does not read exports.csv file!):
 *
 * \include eg_AttachServer1.c
 */
EXPORT_B int AttachServer(char *srvExportName,char *srvEqmName,int ndevices) 
{
  return AttachServerEx(srvExportName,srvEqmName,ndevices,NULL,0);
}
EXPORT_B int AttachServerFromFecXml(void) 
{
  return AttachServerEx(NULL,NULL,0,NULL,0);
}
EXPORT_B void TerminateServer(void) 
{
  stopAllListeners();
# ifdef _USRDLL
  TineStopCycler();
# endif
  FreeEQPModules();
  gNotifierCalled = FALSE;
  hasInitializedOnce = FALSE;
  gBufSrvInitialized = FALSE;
}

EXPORT_B int GetRegisteredProperties(NAME32 *prps, short *nprps)
{
  int i, n;
  if (prps == NULL || nprps == NULL) return -1;
  for (i=0, n=0; i<nBufPrps && n<*nprps; i++)
  {
    if (strlen(bPrps[i].nam) == 0) continue;
    strncpy(prps[n++].name,bPrps[i].nam,32);
  }
  *nprps = n;
  return nBufPrps;
}
int taggableFormat(short fmt)
{
  switch (fmt)
  {
    case CF_STRUCT:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
    case CF_HISTORY:
    case CF_STRING:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
EXPORT_B int GetRegisteredPropertyInformation
(
  char *eqm, char *prp, /* input */
  float *max, float *min,char *units,
  int *sizeOut, int *sizeIn, char *fmtOut, char *fmtIn,
  char *description
)
{
  int cc;
  ExportPropertyListStruct xpls;
  if (eqm == NULL) eqm = gEqmName;

  if ((cc=GetRegisteredPropertyListStruct(eqm,prp,&xpls)) == 0)
  {
    if (max != NULL) *max = xpls.egu.max;
    if (min != NULL) *min = xpls.egu.min;
    if (units != NULL) strncpy(units,xpls.egu.units,16);
    if (sizeOut != NULL) *sizeOut = xpls.prpSize;
    if (sizeIn != NULL) *sizeIn = xpls.prpSizeIn;
    if (fmtOut != NULL)
    {
      strncpy(fmtOut, GetFormatAsString(xpls.prpFormat), 16); fmtOut[15] = 0;
      if (taggableFormat(xpls.prpFormat))
      {
        strcat(fmtOut,".");
        strncat(fmtOut,xpls.prpFormatTag.name,TAG_NAME_SIZE);
      }
    }
    if (fmtIn != NULL)
    {
      strncpy(fmtIn, GetFormatAsString(xpls.prpFormatIn), 16); fmtIn[15] = 0;
      if (taggableFormat(xpls.prpFormatIn))
      {
        strcat(fmtIn,".");
        strncat(fmtIn,xpls.prpFormatTagIn.name,TAG_NAME_SIZE);
      }
    }
    if (description != NULL) strncpy(description,xpls.prpDescription,PROPERTY_DESC_SIZE);    
  }
  return cc;
}
EXPORT_B int GetLastCaller(char *un,char *ip)
{
  int b1,b2,b3,b4;
  if (un != NULL) strncpy(un,gLastWriter.name,16);
  if (ip != NULL)
  {
    b1 = (gLastWriterIp & 0xff);
    b2 = (gLastWriterIp & 0xff00) >> 8;
    b3 = (gLastWriterIp & 0xff0000) >> 16;
    b4 = (gLastWriterIp & 0xff000000) >> 24;
    sprintf(ip,"%d.%d.%d.%d",b1,b2,b3,b4);
  }
  return 0;
}

EXPORT_B int StartRuntimeDebugging(void)
{
# ifdef WIN32
  int cc = 0, pid = GetCurrentProcessId();
  char pipeName[32],cmd[128];
  STARTUPINFO sInfo;
  PROCESS_INFORMATION pInfo; 
  memset(&sInfo,0,sizeof(sInfo));
  if (GetIpcPipeName(pipeName) != 0)
  {
    sprintf(pipeName,"labview-%d",pid);
  }
  OpenIpcPipe(pipeName);
  sprintf(cmd,"AttachFec %s",pipeName);
  CreateProcess(NULL,cmd,NULL,NULL,0,NORMAL_PRIORITY_CLASS,NULL,NULL,&sInfo,&pInfo);
# endif
	return 0;
}

/** 
 * \brief Establishes the maximum returned array length for the target property
 * 
 * A called property will return up to the registered array length of array elements,
 * regardless of the number of values actually 'pushed' into the array.  This call
 * can be used to influence the actual number of elements returned to the caller.
 * 
 * A caller can always ask for to receive fewer data elements than registered. 
 * However asking for more elements will result in a truncated call, where the number
 * of elements returned is given by either by the registered output data length for the
 * designated property or the buffered data size registered with this call.
 *
 * Subsequent calls to PushBufferedData will automatically adjust the buffered data size
 * according to the property size given IF the property size is greater than 1 AND there
 * is already an existing buffered data size registered.
 *
 * Setting the buffered data size to '0' will revert to the initial conditions
 * where the registered property size determines the maximum data size to be returned.
 *
 * \note This call can only by applied to properties which are registered to return
 * an array of type AT_SPECTRUM.
 * 
 * \param prpName is the property name into whose buffer size is to be adjusted.
 * \param dataSiz is the maximum number of elements to be returned to a caller.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 */
EXPORT_B int SetBufferedDataSize(char *prpName,int dataSiz)
{
  int id;
  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return illegal_property;
  if (bPrps[id].synId > 0) 
  { /* property is a synonym, remap to target */
    id = bPrps[id].synId - 1; 
  }
  if ((bPrps[id].arrayType & AT_SPECTRUM) == 0) return not_accepted;
  if (dataSiz > (int)bPrps[id].dout.dArrayLength)
    dataSiz = (int)bPrps[id].dout.dArrayLength;
  bPrps[id].arrayLengthCeiling = dataSiz;
  return 0;
}
/** 
 * \brief Returns the maximum returned array length for the target property
 * 
 * \param prpName is the property name into whose buffer size is to be returned.
 *
 * \return buffer size if successful, otherwise the negative of a TINE completion code 
 * which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int GetBufferedDataSize(char *prpName)
{
  int id;
  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return -illegal_property;
  if (bPrps[id].synId > 0) 
  { /* property is a synonym, remap to target */
    id = bPrps[id].synId - 1; 
  }
  return bPrps[id].arrayLengthCeiling;
}
/** 
 * \brief Refreshes the contents of the image data associated with the given property
 * 
 * \param prpName is the property name into whose buffer the associated data are to be written
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer into which the associated data are to be written
 * \param sourceHeader is a reference to the video frame source header.  This contains informational
 *        content and a NULL pointer can be passed here without affecting the image frame to be transfered.
 * \param frameHeader is a reference to the video frame header.  This contains important information
 *        pertaining to the transfered video frame and cannot be NULL.
 * \param frameBytes is a reference to the image data to be written (pushed) into the property's data
 *        buffer.
 * \param frameSizeBytes is the number of bytes contained in the image frame to be transfered.
 * \param prpSchedule is the schedule 'flag' determining whether the property is to be
 *        scheduled immediately ('true') or not. If prpSchedule is >= 0x10 (16) and < 0x1000
 *        then it is assumed to provide the specific scheduling 'scope' (some combination of 
 *        CA_ALARM - 0x10, CA_HIST - 0x20, CA_NETWORK - 0x200).  If prpSchedule < 0x10 then it
 *        treated as a simple boolean which schedules the property(ies) for all subsystems.  
 *        If prpSchedule > 0x1000 it is assumed to provide a user-assigned UTC timestamp to 
 *        the data in question (and the property is not scheduled in this case).
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 */
EXPORT_B int pushImageData(char *prpName,char *devName,TImageSourceHeader *sourceHeader,TImageFrameHeader *frameHeader,BYTE *frameBytes,int frameSizeBytes,int prpSchedule)
{
  DIMAGE img;

  if (frameHeader == NULL || frameBytes == NULL) return invalid_parameter;
  memcpy(&img.frameHeader,frameHeader,sizeof(TImageFrameHeader));
  if (sourceHeader != NULL) 
  {
    memcpy(&img.sourceHeader,sourceHeader,sizeof(TImageSourceHeader));
  }
  else
  {
    memset(&img.sourceHeader,0,sizeof(TImageSourceHeader));
  }
  img.sourceHeader.totalLength = sizeof(TImageSourceHeader) + sizeof(TImageFrameHeader) + frameHeader->appendedFrameSize;
  img.frameBufferSize = frameSizeBytes;
  img.frameBuffer = frameBytes;
  return pushBufferedData(prpName,devName,(BYTE *)&img,frameSizeBytes,prpSchedule);
}
/** 
 * \brief Refreshes the contents of the data buffer associated with the given property
 * 
 * \param prpName is the property name into whose buffer the associated data are to be written
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer into which the associated data are to be written
 * \param prpData is a reference to the data to be written (pushed) into the property's data
 *        buffer.
 * \param prpSiz is the number of elements to be written into the property's data buffer.
 * \param prpSchedule is the schedule 'flag' determining whether the property is to be
 *        scheduled immediately ('true') or not. If prpSchedule is >= 0x10 (16) and < 0x1000
 *        then it is assumed to provide the specific scheduling 'scope' (some combination of 
 *        CA_ALARM - 0x10, CA_HIST - 0x20, CA_NETWORK - 0x200).  If prpSchedule < 0x10 then it
 *        treated as a simple boolean which schedules the property(ies) for all subsystems.  
 *        If prpSchedule > 0x1000 it is assumed to provide a user-assigned UTC timestamp to 
 *        the data in question (and the property is not scheduled in this case).
 *        A negative value will indicate that the property should only be scheduled for
 *        contracts using the devName given.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:
 *
 * \include eg_pushBufferedData.c
 */
EXPORT_B int pushBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule)
{
  return pushBufferedDataEx(prpName,devName,prpData,prpSiz,prpSchedule,0,0,0);
}
/** 
 * \brief Refreshes the contents of the data buffer associated with the given property 
 * (extended form).
 * 
 *
 * \param prpName is the property name into whose buffer the associated data are to be written
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer into which the associated data are to be written
 * \param prpData is a reference to the data to be written (pushed) into the property's data
 *        buffer.
 * \param prpSiz is the number of elements to be written into the property's data buffer.
 * \param prpSchedule is the schedule 'flag' determining whether the property is to be
 *        scheduled immediately ('true') or not. If prpSchedule is >= 0x10 (16) and < 0x1000
 *        then it is assumed to provide the specific scheduling 'scope' (some combination of 
 *        CA_ALARM - 0x10, CA_HIST - 0x20, CA_NETWORK - 0x200).  
 *        A negative value will indicate that the property should only be scheduled for
 *        contracts using the devName given.
 * \param timeStamp is a UTC double precision timestamp to be applied to the data (in lieu of the
 *        current system timestamp.
 * \param sysStamp is an integer 'system stamp' to be applied to the data (in lieu of the current
 *        system stamp.
 * \param usrStamp is an integer 'user defined stamp' to be applied to the data.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:
 *
 * \include eg_pushBufferedData.c
 */
EXPORT_B int pushBufferedDataEx(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule,double timeStamp,int sysStamp,int usrStamp)
{
  int cc, id, i, p, fsz, hsz, fmt;
  int ssz = 1, ds;
  double ts;
  char prpStr[256];

  if (prpData == NULL) return buffer_too_small;
  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return illegal_property;
  if ((i=GetDeviceNumberEx(gEqmName,devName,prpName)) < 0) return -i;
  if (bPrps[id].synId > 0) 
  { /* property is a synonym, remap to target */
    id = bPrps[id].synId - 1; 
  }
  if (bPrps[id].arrayLengthCeiling > 0  && prpSiz > 1)
  { /* if a ceiling exists update it with the push call */
    bPrps[id].arrayLengthCeiling = prpSiz;
  }
  fmt = bPrps[id].dout.dFormat; /* simplify by assignment ... */
  if (fmt == CF_STRUCT) 
  { /* get the structure size */
    if ((ssz=bPrps[id].doutStructSize) <= 0)
    { /* not yet known */
      if ((cc=allocBufferedStorage(&bPrps[id])) != 0) return cc;
      ssz = bPrps[id].doutStructSize;
    }
  }
  ts = timeStamp > 0x1000 ? timeStamp : makeDataTimeStamp();
  ds = sysStamp > 0 ? sysStamp : GetSystemDataStamp(-1);
  bPrps[id].pushedCompletionCode = 0;
  if (prpSiz > i + (int)bPrps[id].dout.dArrayLength) 
  {
    prpSiz = (int)bPrps[id].dout.dArrayLength;
    if (bPrps[id].arrayType != AT_SPECTRUM && i < prpSiz) prpSiz -= i;
  }
  fsz = ssz * fmtsizeof(fmt);
  hsz = getFormatHeaderSize(fmt);
  switch (fmt)
  { /* some last minute checks ... */
    case CF_TEXT:
      p = bPrps[id].dout.dArrayLength * i;
      memset(&bPrps[id].dout.data.cptr[p],0,bPrps[id].dout.dArrayLength);
      strncpy(&bPrps[id].dout.data.cptr[p],(const char *)prpData,prpSiz*fsz);
      bPrps[id].devStamps[i].dataTimeStamp = ts;
      bPrps[id].devStamps[i].sysStamp = ds;
      bPrps[id].devStamps[i].usrStamp = usrStamp;
      break;
    case CF_IMAGE:
      {
        DIMAGE *tgt, *src = (DIMAGE *)prpData;
        int xsz = sizeof(UINT32) + sizeof(BYTE *);
        p = (bPrps[id].dout.dArrayLength + hsz + xsz) * i;
        tgt = (DIMAGE *)&bPrps[id].dout.data.bptr[p];
        memcpy(tgt,src,hsz);
        tgt->frameBufferSize = src->frameBufferSize;
        tgt->frameBuffer = &((BYTE *)tgt)[hsz+xsz];
        memcpy(tgt->frameBuffer,src->frameBuffer,src->frameBufferSize);
      }
      break;
    case CF_GSPECTRUM:
      {
        DGSPECTRUM *tgt, *src = (DGSPECTRUM *)prpData;
        int xsz = sizeof(UINT32) + sizeof(BYTE *);
        p = (bPrps[id].dout.dArrayLength + hsz + xsz) * i;
        tgt = (DGSPECTRUM *)&bPrps[id].dout.data.bptr[p];
        memcpy(tgt, src, hsz);
        tgt->spectBufferSize = src->spectBufferSize;
        tgt->spectBuffer = (float *)(&((BYTE*)tgt)[hsz + xsz]);
        memcpy(tgt->spectBuffer, src->spectBuffer, src->spectBufferSize * fsz);
      }
      break;
    case CF_STRING:
    case CF_KEYVALUE: 
    {
      char **sasrc = (char **)prpData;
      int k, len;
      p = i;
      if (bPrps[id].arrayType & AT_SPECTRUM) p = bPrps[id].dout.dArrayLength * i;
      for (k = 0; k < prpSiz; k++)
      {
        len = (int)strlen(sasrc[k]);
        if (len > bPrps[id].stringSpaceCapacity[p+k])
        { // re-alloc
          BYTE *ptr = NULL;
          int siz = len + STRING_STD_CAPACITY;
          ptr = (BYTE *)realloc(bPrps[id].stringSpace[i], siz);
          if (ptr != NULL)
          {
            bPrps[id].stringSpace[i] = ptr;
            bPrps[id].stringSpaceCapacity[p + k] = siz;
          }
          else
          {
            len = bPrps[id].stringSpaceCapacity[p + k];
          }
        }
        bPrps[id].dout.data.strptr[p+k] = (char *)bPrps[id].stringSpace[p+k];
        strncpy((char *)bPrps[id].stringSpace[p+k], sasrc[k], len);
      }
      break;
    }
    default:
      if (bPrps[id].arrayType & AT_SPECTRUM || hsz > 0)
      {
        p = (bPrps[id].dout.dArrayLength * fsz + hsz) * i;
        memset(&bPrps[id].dout.data.bptr[p],0,bPrps[id].dout.dArrayLength*fsz);
        /* and the device specific data stamps */
        bPrps[id].devStamps[i].dataTimeStamp = ts;
        bPrps[id].devStamps[i].sysStamp = ds;
        bPrps[id].devStamps[i].usrStamp = usrStamp;
      }
      else
      {
        p = fsz * i;
      }
      memcpy(&bPrps[id].dout.data.bptr[p],prpData,prpSiz*fsz+hsz);
      break;
  }
  bPrps[id].dataTimeStamp = ts;
  bPrps[id].sysStamp = ds;
  bPrps[id].usrStamp = usrStamp;
  memset(prpStr, 0, 256);
  if (prpSchedule < 0)
  {
    sprintf(prpStr, "%.64s<%.64s>", prpName, devName);
    prpSchedule = -prpSchedule;
  }
  else
  {
    strncpy(prpStr, prpName, PROPERTY_NAME_SIZE);
  }
  if (prpSchedule > 0)
  { /* non-zero schedule flag */
    if (prpSchedule < 0x10) 
    { /* treat as boolean */
      SystemScheduleProperty(gEqmName,prpStr);
    }
    else if (prpSchedule < 0x1000)
    { /* treat as input scheduling scope */
      SystemSchedulePropertyEx(gEqmName,prpStr,prpSchedule);
    }
    else
    { /* treat as assigned UTC timestamp */
      if (timeStamp > 0x1000)
      { /* stay with it and just schedule */
        SystemScheduleProperty(gEqmName,prpStr);
      }
      else
      {
        bPrps[id].dataTimeStamp = (double)prpSchedule;
      }
    }
  }
  return 0;
}
/** 
 * \brief Retrieves the device (buffer) index associated with the input parameters
 * 
 *
 * \param prpName is the property name associated with the device name passed
 * \param devName is the device name giving the location with the property buffer and 
 *  hence the buffer index.
 *
 * \return a positive index number if successful, otherwise the negative of a TINE 
 *  completion code which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int GetBufferIndex(char *prpName,char *devName)
{
  return GetDeviceNumberEx(gEqmName,devName,prpName);
}
/** 
 * \brief Retrieves the contents of the input data buffer associated with the given property
 * 
 *
 * \param prpName is the property name from whose input buffer the associated data are to be read.
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer from which the associated data are to be read.
 * \param prpData is a reference to the data buffer into which the incoming data should be read 
 *        (pulled).
 * \param prpSiz is the number of elements to be read into the property's data buffer.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:
 *
 * \include eg_pullBufferedData.c
 */
EXPORT_B int pullBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz)
{
  int id, cc, i, p, fsz, hsz, ssz = 1, isText = FALSE, isString = FALSE, isTrace = FALSE;

  if (prpData == NULL) return buffer_too_small;
  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return illegal_property;
  if ((i=GetDeviceNumberEx(gEqmName,devName,prpName)) < 0) return -i;
  if (bPrps[id].synId > 0) 
  { /* property is a synonym, remap to target */
    id = bPrps[id].synId - 1; 
  }
  hsz = getFormatHeaderSize(bPrps[id].din.dFormat);
  if (bPrps[id].arrayType & AT_SPECTRUM || hsz > 0) isTrace = TRUE;
  if (bPrps[id].din.dFormat == CF_TEXT) isText = TRUE;
  else if (bPrps[id].din.dFormat == CF_STRING) isString = TRUE;
  else if (bPrps[id].din.dFormat == CF_KEYVALUE) isString = TRUE;
  else if (bPrps[id].din.dFormat == CF_STRUCT)
  { /* get the structure size */
    if ((ssz=bPrps[id].dinStructSize) <= 0)
    { /* not yet known */
      if ((cc=allocBufferedStorage(&bPrps[id])) != 0) return cc;
      ssz = bPrps[id].dinStructSize;
    }
  }
  if (prpSiz > i + (int)bPrps[id].din.dArrayLength) 
  {
    prpSiz = (int)bPrps[id].din.dArrayLength;
    if (bPrps[id].arrayType != AT_SPECTRUM) prpSiz -= i;
  }
  fsz = ssz * fmtsizeof(bPrps[id].din.dFormat);
  if (bPrps[id].din.dFormat == CF_IMAGE ||
      bPrps[id].din.dFormat == CF_GSPECTRUM) hsz += sizeof(UINT32) + sizeof(BYTE*);
  p = fsz * i;

  if (isTrace) 
  {
    p = (bPrps[id].din.dArrayLength * fsz + hsz) * i;
  }
  else if (bPrps[id].din.dArrayLength == 1 && !(bPrps[id].arrayType & AT_CHANNEL))
  { /* registered to only accept 1 element */
    p = 0;
  }
  if (isText)
  {
    strncpy((char *)prpData,(const char *)&bPrps[id].din.data.cptr[p],prpSiz*fsz);
  }
  else if (isString)
  {
    fsz = sizeof(char*);
    p = (bPrps[id].din.dArrayLength * fsz) * i;
    memcpy(prpData, &bPrps[id].din.data.bptr[p], prpSiz * fsz);
  }
  else
  {
    memcpy(prpData,&bPrps[id].din.data.bptr[p],hsz+prpSiz*fsz);
  }
  bPrps[id].inputChanged = 0;
  return 0;
}
/** 
 * \brief Retrieves the property which caused the notifier to be called
 * 
 *
 * \param prpName (output) is the property name which caused the notifer to be called
 *
 * \return 0 upon success otherwise a TINE return code
 *
 */
EXPORT_B int getNotifiedProperty(char *prpName)
{
  int id = currentPropertyId;
  if (id < 0 && prpName != NULL && strlen(prpName) > 0) id = GetPropertyId(gEqmName,prpName);
  if (id < 0) return illegal_property;
  if (bPrps[id].notifierCalled)
  {
    if (prpName != NULL) strncpy(prpName,bPrps[id].nam,PROPERTY_NAME_SIZE);
  }
  else
  { // this is obsolete !
    if (!gNotifierCalled) return not_signalled;
    if (prpName != NULL) strncpy(prpName,gPropertyNofified,PROPERTY_NAME_SIZE);
  }
  return 0;
}
/** 
 * \brief Retrieves the property which caused the notifier to be called
 * 
 *
 * \param prpName (output) is the property name which caused the notifer to be called
 *
 * \return 0 upon success otherwise a TINE return code
 *
 */
EXPORT_B int getNotifiedPropertyAndDevice(char *prpName,char *devName)
{
  int id = currentPropertyId;
  if (id < 0 && prpName != NULL && strlen(prpName) > 0) id = GetPropertyId(gEqmName,prpName);
  if (id < 0) return illegal_property;
  if (bPrps[id].notifierCalled)
  {
    if (prpName != NULL) strncpy(prpName,bPrps[id].nam,PROPERTY_NAME_SIZE);
    if (devName != NULL) strncpy(devName,bPrps[id].dev,DEVICE_NAME_SIZE);
    //if (prpName != NULL) strncpy(prpName,bPrps[id].nam,PROPERTY_NAME_SHORTSIZE);
    //if (devName != NULL) strncpy(devName,bPrps[id].dev,DEVICE_NAME_SHORTSIZE);
  }
  else
  { // this is obsolete !
    if (!gNotifierCalled) return not_signalled;
    if (prpName != NULL) strncpy(prpName,gPropertyNofified,PROPERTY_NAME_SIZE);
    if (devName != NULL) strncpy(devName,gDeviceNotified,DEVICE_NAME_SIZE);
    //if (prpName != NULL) strncpy(prpName,gPropertyNofified,PROPERTY_NAME_SHORTSIZE);
    //if (devName != NULL) strncpy(devName,gDeviceNotified,DEVICE_NAME_SHORTSIZE);
  }
  return 0;
}
/** 
 * \brief Checks whether there are new input data for the given property
 * 
 *
 * \param prpName is the property name whose input data buffer is to be checked for changes.
 *
 * \return 0 (False) if the input data have not changed or -1 (True) if there are new input
 * data to be acquired by calling PullBufferedData().
 *
 * \b Example:
 *
 * \include eg_pullBufferedData.c
 */
EXPORT_B int hasInputChanged(char *prpName)
{
  int id;

  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return 0;
  return bPrps[id].inputChanged == -1 ? TRUE : FALSE;
}
/** 
 * \brief Returns the device number associated with the WRITE call.
 * 
 * This call can be made inside the server's callback routine along with
 * 'hasInputChanged' in order to ascertain which device was 'set'. 
 *
 * \return The device number (obtained from the registered device name) used
 * in the call.
 *
 * \b Example:
 *
 * \include eg_GetInputDeviceNumber.c
 */
EXPORT_B int GetInputDeviceNumber(void)
{
  return currentDeviceNumber;
}
/** 
 * \brief Registers a callback routine to be called when a WRITE access property is
 * called.
 * 
 * \param prpName
 * \param cb
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:
 *
 * \include eg_RegisterServerCallback.c
 */
EXPORT_B int RegisterServerCallback(char *prpName, HNDLRCBP cb)
{
  int id;

  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return illegal_property;
  bPrps[id].callback = cb;
  bPrps[id].cbReference = NULL;
  return 0;
}
/** 
 * \brief Registers a property handler callback routine to be called when a WRITE 
 * access property is called. This extended callback will provide reference pointer 
 * back to the callback routine when it is fired.
 * 
 * \param prpName
 * \param cb
 * \param ref is a user-defined reference pointer to be returned when the property
 *  handler callback is called.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 */
EXPORT_B int RegisterServerCallbackEx(char *prpName, XHNDLRCBP cb, void *ref)
{
  int id;
  if ((id=GetPropertyId(gEqmName,prpName)) < 0) return illegal_property;
  bPrps[id].callback = (HNDLRCBP)cb;
  bPrps[id].cbReference = ref;
  if (bPrps[id].synId > 0)
  { /* same handler for him too! */
    id = bPrps[id].synId - 1;
    bPrps[id].callback = (HNDLRCBP)cb;
    bPrps[id].cbReference = ref;
  }
  /* if SAVE/RESTORE registered we need to do this here : */
  ApplyStoredPropertyValues(gEqmName,prpName);
  return 0;
}
/** 
 * \brief Registers a Notifier routine to be called when a WRITE access property is
 * called.
 * 
 * The difference between a callback and a notifier is that it is assumed that the callback 
 * completes all activity associated with the transaction while inside the callback routine.
 * A notififier will call the notifier routine and expect the caller to call the 
 * SetBufferedCompletion() routine when the transaction is complete.  This scenario is useful 
 * for development platforms such as LabView.
 *
 * \param prpName the Property name for which the notifier is in effect
 * \param nf the notifier routine
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 * \b Example:
 *
 * \include eg_RegisterServerNotifier.c
 */
EXPORT_B int RegisterServerNotifier(char *prpName, NTFRCBP nf)
{
  return RegisterServerNotifierEx(prpName,nf,0);
}
/** 
 * \brief Registers a Notifier routine to be called when a WRITE access property is
 * called (extended call).
 * 
 * The difference between a callback and a notifier is that it is assumed that the callback 
 * completes all activity associated with the transaction while inside the callback routine.
 * A notififier will call the notifier routine and expect the caller to call the 
 * SetBufferedCompletion() routine when the transaction is complete.  This scenario is useful 
 * for development platforms such as LabView.
 *
 * \param prpName the Property name for which the notifier is in effect
 * \param nf the notifier routine
 * \param nid is the desired notifier id to be returned in the notifier supplied notifier routine.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetLastLinkError().
 *
 */
EXPORT_B int RegisterServerNotifierEx(char *prpName, void (*nf)(int), int nid)
{
  int id, i, cc = 0;
  char *pn, sbuf[1024];
  static char *tok = ", ";

  if (prpName == NULL || strlen(prpName) == 0)
  {
    gBufferedNotifier = nf;
    gBufferedNotifierId = nid;
  }
  else
  {
    strncpy(sbuf, prpName, 1024); sbuf[1023] = 0;
    for (pn = strtok(sbuf,tok); pn != NULL; pn = strtok(NULL,tok))
    {
      strtrm(pn);
      if ((id=GetPropertyId(gEqmName,pn)) >= 0)
      {
        bPrps[id].notifier = nf;
        bPrps[id].notifierId = nid;
        if (bPrps[id].needToNotify)
        { /* late save-restore */
          int n = GetNumberRegisteredDevices(gEqmName);
          for (i=0; i<n; i++)
          { /* traverse the device list */
            strncpy(bPrps[id].dev,"#%d",i);
            bPrps[id].notifierCalled = -1; /* cleared in SetBufferedCompletion() */
            bPrps[id].notifier(bPrps[id].notifierId);
          }
          bPrps[id].needToNotify = FALSE;
        }
      }
      else
      {
        cc = illegal_property;
      }
    }
  }
  return cc;
}
// TODO: atyp, rowlength, egu ...
/** 
 * \brief Registers a property with the current device server.
 * 
 * It many cases it is easiest to supply local configuration files with
 * all of the property and device information relevant for a buffered
 * server. In such cases there are remarkably few API calls necessary to
 * provide a fully functioning device server.  
 *
 * On the other hand, if it is necessary to register additional properties
 * this registration call can be used.
 *
 * \param prpName is the propery name (up to 64 characters, but try to use
 *  32 or less).
 * \param prpInSiz is the maximum input data size accepted by the property
 * \param prpInFmt is the data format of the input data
 * \param prpOutSiz is the maximum output data size accepted by the property
 * \param prpOutFmt is the data format of the output data
 * \param prpMax is the nominal maximum value of the property data
 * \param prpMin is the nominal minimum value of the property data
 * \param prpEgu is the engineering unit of the property data
 * \param access is the access accepted by the property (e.g. CA_READ and/or CA_WRITE).
 * \param prpDsc is the description of the property.
 *
 * \return a positive property ID if successful, otherwise the negative of a TINE completion 
 * code which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int RegisterBufferedProperty
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
) 
{
  return RegisterBufferedPropertyEx(prpName,prpInSiz,prpInFmt,prpOutSiz,prpOutFmt,
                                    prpMax,prpMin,prpEgu,access,prpDsc,-1);
}
/** 
 * \brief Registers a property with the current device server.
 * (extended call).
 * 
 * It many cases it is easiest to supply local configuration files with
 * all of the property and device information relevant for a buffered
 * server. In such cases there are remarkably few API calls necessary to
 * provide a fully functioning device server.  
 *
 * On the other hand, if it is necessary to register additional properties
 * this registration call can be used.
 *
 * \param prpName is the propery name (up to 64 characters, but try to use
 *  32 or less).
 * \param prpInSiz is the maximum input data size accepted by the property
 * \param prpInFmt is the data format of the input data
 * \param prpOutSiz is the maximum output data size accepted by the property
 * \param prpOutFmt is the data format of the output data
 * \param prpMax is the nominal maximum value of the property data
 * \param prpMin is the nominal minimum value of the property data
 * \param prpEgu is the engineering unit of the property data
 * \param access is the access accepted by the property (e.g. CA_READ and/or CA_WRITE).
 * \param prpDsc is the description of the property.
 * \param prpId is the desired property ID (returned in a call to GetPropertyId()).
 *
 * \return a positive property ID if successful, otherwise the negative of a TINE completion 
 * code which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int RegisterBufferedPropertyEx
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc, int prpId
) 
{
  int atyp = AT_UNKNOWN, rowlength = 0;
  char localDescription[128],fmtstr[64];
  char *c, *atypptr=fmtstr, *dscptr=localDescription;

  sprintf(localDescription,"%.96s",prpDsc);

  if ((c=strchr(localDescription,'[')) != NULL)
  {
    c++;
    if (strchr(c,'.') != NULL) 
    {
      atypptr = c;
    }
    else
    {
      sprintf(fmtstr,"%.16s.%.16s",GetFormatAsString(prpOutFmt),c);
      atypptr = fmtstr;
    }
    if ((c=strchr(c,']')) != NULL)
    {
      c++; dscptr = c;
    }
  }
  atyp = GetArrayTypeFromString(atypptr);
  rowlength = prpOutSiz;
  return RegisterBufferedPropertyEx2(prpName,prpInSiz,prpInFmt,NULL,prpOutSiz,prpOutFmt,NULL,
                                     prpMax,prpMin,prpEgu,access,dscptr,prpId,atyp,rowlength);
}
/** 
 * \brief Registers a property with the current device server.
 * (doubly extended call).
 * 
 * It many cases it is easiest to supply local configuration files with
 * all of the property and device information relevant for a buffered
 * server. In such cases there are remarkably few API calls necessary to
 * provide a fully functioning device server.  
 *
 * On the other hand, if it is necessary to register additional properties
 * this registration call can be used.
 *
 * \param prpName is the propery name (up to 64 characters, but try to use
 *  32 or less).
 * \param prpInSiz is the maximum input data size accepted by the property
 * \param prpInFmt is the data format of the input data
 * \param prpOutSiz is the maximum output data size accepted by the property
 * \param prpOutFmt is the data format of the output data
 * \param prpMax is the nominal maximum value of the property data
 * \param prpMin is the nominal minimum value of the property data
 * \param prpEgu is the engineering unit of the property data
 * \param access is the access accepted by the property (e.g. CA_READ and/or CA_WRITE).
 * \param prpDsc is the description of the property.
 * \param prpId is the desired property ID (returned in a call to GetPropertyId()).
 * \param arrayType is the array type (e.g. AT_CHANNEL, AT_TRACE) of the property
 * \param rowLength is the row length of the property (relevant if the property delivers
 * a double array).
 *
 * \return a positive property ID if successful, otherwise the negative of a TINE completion 
 * code which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int RegisterBufferedPropertyEx2
(
  char *prpName, long prpInSiz, short prpInFmt, char *prpInTag, long prpOutSiz, short prpOutFmt, char *prpOutTag,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc, int prpId, int arrayType, int rowLength
) 
{
  int id,sid,cc;
  char localEqpname[16],localProperty[PROPERTY_NAME_SIZE+4],localDescription[256];
  char *c, prpBaseName[PROPERTY_NAME_SIZE+4];
  DTYPE dout, din;

  initBufferedServerMemory();
  if (prpName == NULL) return argument_list_error;
  strncpy(localEqpname,gEqmName,EQM_NAME_SHORTSIZE);
  strncpy(localProperty,prpName,PROPERTY_NAME_SIZE);
  if (prpDsc == NULL) prpDsc = "no description available";
  if (strchr(prpDsc,'[') != NULL)
  {
    strncpy(localDescription,prpDsc,256);
  }
  else
  {
    if (prpEgu == NULL) prpEgu = "none";
    sprintf(localDescription,"[vscale=%g:%g %.32s][desc=%.64s]",prpMin,prpMax,prpEgu,prpDsc);
  }
  
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  dout.dFormat = prpOutFmt;
  dout.dArrayLength = prpOutSiz;
  if (prpOutTag != NULL && IsTaggableFormat(prpOutFmt)) strncpy(dout.dTag,prpOutTag,TAG_NAME_SIZE);
  din.dFormat = prpInFmt;
  din.dArrayLength = prpInSiz;
  if (prpInTag != NULL && IsTaggableFormat(prpInFmt)) strncpy(din.dTag,prpInTag,TAG_NAME_SIZE);

  if (arrayType == AT_UNKNOWN && prpOutFmt == CF_SPECTRUM)
  {
    arrayType = AT_SINGLE|AT_SPECTRUM;
  }
  if (rowLength == 0) rowLength = prpOutSiz;
  if (access == 0) access = CA_READ;

  id = RegisterPropertyInformation(localEqpname,localProperty,&dout,&din,
         access,(short)arrayType,(UINT16)rowLength,prpDsc,prpId,NULL); 

  if (id < 0 || id >= MAX_BUFFERED_PROPERTIES) return -illegal_property;

  strncpy(bPrps[id].nam,prpName,PROPERTY_NAME_SIZE);
  bPrps[id].max = prpMax;
  bPrps[id].min = prpMin;
  bPrps[id].dout.dFormat = prpOutFmt;
  bPrps[id].dout.dArrayLength = prpOutSiz;
  if (prpOutTag != NULL && IsTaggableFormat(prpOutFmt)) strncpy(bPrps[id].dout.dTag,prpOutTag,TAG_NAME_SIZE);
  bPrps[id].din.dFormat= prpInFmt;
  bPrps[id].din.dArrayLength = prpInSiz;
  if (prpInTag != NULL && IsTaggableFormat(prpInFmt)) strncpy(bPrps[id].din.dTag,prpInTag,TAG_NAME_SIZE);
  bPrps[id].arrayType = arrayType;
  bPrps[id].pushedCompletionCode = unpushedDataCompletionCode;
  bPrps[id].acc = access;
  if (isScheduleDecorated(prpName))
  { /* has been decorated as a 'scheduled' property */
    strncpy(prpBaseName, prpName, PROPERTY_NAME_SIZE); prpBaseName[PROPERTY_NAME_SIZE] = 0;
    if ((c=strrchr(prpBaseName,'.')) != NULL)
    { /* this had better be true ! */
      *c = 0;
      if ((sid=GetPropertyId(localEqpname,prpBaseName)) < 0)
      { /* not yet registered  */
        sid = RegisterBufferedPropertyEx2(prpBaseName,prpInSiz,prpInFmt,prpInTag,
                                         prpOutSiz,prpOutFmt,prpOutTag,
                                         prpMax,prpMin,prpEgu,
                                         access,prpDsc,-1,arrayType,rowLength);
      }
      bPrps[sid].synId = id + 1;
    }
  }  

  if ((cc=allocBufferedStorage(&bPrps[id])) != 0) return -cc;

  nBufPrps++;
  return id;
}
/** 
 * \brief Registers a device with the current device server.
 * 
 * It many cases it is easiest to supply local configuration files with
 * all of the property and device information relevant for a buffered
 * server. In such cases there are remarkably few API calls necessary to
 * provide a fully functioning device server.  
 *
 * On the other hand, if it is necessary to register additional devices
 * this registration call can be used.
 *
 * \param devName is the device name (up to 64 characters)
 * \param devNr is the associated device number
 * \param devRdr is an optional redirection string (if the device is to be
 * redirected to another server
 * \param devDesc is an optional description of the device.
 *
 * \return 0 if successful, otherwise a TINE completion 
 * code which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int RegisterBufferedDeviceName(char *devName,int devNr,char *devRdr,char *devDesc)
{
  int cc = RegisterDeviceName(gEqmName,devName,devNr);
  if (cc == 0)
  {
    if (devRdr != NULL && strlen(devRdr) > 0) RedirectDeviceName(gEqmName,devName,devRdr);
    if (devDesc != NULL && strlen(devDesc) > 0) RegisterDeviceDescription(gEqmName,devName,devDesc);
  }
  return cc;
}
/** 
 * \brief Registers a device with the current device server.
 * (extended call).
 * 
 * It many cases it is easiest to supply local configuration files with
 * all of the property and device information relevant for a buffered
 * server. In such cases there are remarkably few API calls necessary to
 * provide a fully functioning device server.  
 *
 * On the other hand, if it is necessary to register additional devices
 * this registration call can be used.
 *
 * \param devName is the device name (up to 64 characters)
 * \param devNr is the associated device number
 * \param devMask is the associated device mask
 * \param zPos is the associated Z or longitudinal position of the device
 * \param devRdr is an optional redirection string (if the device is to be
 * redirected to another server
 * \param devDesc is an optional description of the device.
 * \param devLocation is an optional text specifying the physical locatio of the device.
 *
 * \return 0 if successful, otherwise a TINE completion 
 * code which can be interpreted by a call to GetLastLinkError().
 *
 */
EXPORT_B int RegisterBufferedDeviceNameEx(char *devName,int devNr,int devMask,float zPos,char *devRdr,char *devDesc,char *devLocation)
{
  int cc = RegisterDeviceName(gEqmName,devName,devNr);
  if (cc == 0)
  {
    if (devRdr != NULL && strlen(devRdr) > 0) RedirectDeviceName(gEqmName,devName,devRdr);
    if (devMask != 0) SetDeviceMask(gEqmName,devName,devMask);
    if (zPos != 0) SetDeviceZPosition(gEqmName,devName,zPos);
    if (devDesc != NULL && strlen(devDesc) > 0) SetDeviceDescription(gEqmName,devName,devDesc);
    if (devLocation != NULL && strlen(devLocation) > 0) SetDeviceLocation(gEqmName,devName,devLocation);
  }
  return cc;
}

EXPORT_B int TriggerEvent(char *devname,char *cmt,short triggerLevel)
{
  return sendEventTrigger(devname,cmt,triggerLevel);
}

void (*gTriggerNotifier)(int) = NULL;
int gTriggerNotifierId = 0;
int gTiggerCycleNumber = 0;
int gTriggerCycleStatus = 0;
void cbBufferedTrigger(int cycleNumber,int cycleStatus,void *ref)
{
  ref = ref;
  gTiggerCycleNumber = cycleNumber;
  gTriggerCycleStatus = cycleStatus;
  gTriggerNotifier(gTriggerNotifierId);
}
EXPORT_B int GetTriggerCycle(int *cycleNumber,int *cycleStatus)
{
  if (gTriggerNotifier == NULL) return un_allocated;
  if (cycleNumber != NULL) *cycleNumber = gTiggerCycleNumber;
  if (cycleStatus != NULL) *cycleStatus = gTriggerCycleStatus;
  return 0;
}
EXPORT_B int RegisterCycleTriggerNotifier(void (*nf)(int), int nid)
{
  int cc;
  static char localEqpname[16];
  strncpy(localEqpname,gEqmName,EQM_NAME_SHORTSIZE);
  if ((cc=RegisterCycleTriggerFunction(cbBufferedTrigger,localEqpname,NULL,NULL)) == 0)
  {
    gTriggerNotifier = nf;
    gTriggerNotifierId = nid;
  }
  if (cc == already_assigned) cc = 0;
  return cc;
}
EXPORT_B char *GetEquipmentModuleName(void)
{
  return gEqmName;
}
