/*
 * File tcycler.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#ifdef UNIX
#include <dlfcn.h>
#endif
#include "tine.h"
#include "threader.h"

int gSystemInitialized=FALSE;
int cycle_busy = FALSE;
int cycleTerminated = FALSE;
unsigned int idCycleTimer = 0;
int userNameSet = FALSE;
int gSystemDelay = 20;

int gUseWindowsTimer = FALSE;
int gWindowsTimerInUse = FALSE;
int gThreadTimerInUse = TRUE;
int cycleStarted = FALSE;
int gIsClosingDynamic = FALSE;
int gTmrThreadDetached = FALSE;
THRHANDLE hTmrThread = (THRHANDLE)0;

#if defined(WIN32)
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
#elif defined(UNIX)
pthread_attr_t gCycleThreadAttr, *gPtrCycleThreadAttr = &gCycleThreadAttr;
pthread_mutex_t cycleStartMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hCyclerStartMutex = &cycleStartMutex;
pthread_mutex_t cycleStopMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hCyclerStopMutex = &cycleStopMutex;
#elif defined(VXWORKS)
#define MULTITHREADED
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
extern int gVxwDefaultTaskOptions;
extern int gVxwDefaultTaskPriority;
extern int gVxwDefaultTaskStack;
#endif

#ifndef VXWORKS
/* use for debugging purposes ... */
void cyclog(char *txt,...)
{
  static char fn[64] = "cycler.log";
  FILE *fp=NULL;
  char str[1024];
  long flen;
  int cc=0;
  va_list args;

  va_start(args,txt);
  vsnprintf(str,512,txt,args); str[512] = 0;
  va_end(args);
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");

  if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  fseek(fp, 0L, SEEK_END); flen = ftell(fp);
  fwrite(str,strlen(str),1,fp);
err:
  UNUSED(cc);
  UNUSED(flen);
  if (fp != NULL) fclose(fp);
  return;
}
#endif
void InitCycleMutexSet(void);
void TineStopCycler(void)
{
  if (cycleTerminated) return;
  SystemKillCycleTimer();
  millisleep(20);
}
#if defined(UNIX) && !defined(SOLARIS)
#ifndef TINE_START_CYCLER
# define TINE_START_CYCLER FALSE
#endif
__attribute__((constructor)) void TineLoadCycler(void)
{
# ifndef TINE_NO_DYNAMIC_LOAD
  int startCycler = TINE_START_CYCLER;
  char *ptr = getenv("TINE_START_CYCLER");
  gIsDynamicLoad = TRUE;
  if (ptr != NULL)
  {
    startCycler = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
  }
  if (startCycler)
  { /* start cycler immediately */
    SystemSetCycleTimer();
  }
  else
  { /* let the app manage the cycling */
    InitCycleMutexSet();
  }
# endif
}
__attribute__((destructor)) void TineStopCyclerEpilog(void)
{
# ifndef TINE_NO_DYNAMIC_LOAD
  gIsClosingDynamic = TRUE;
  if (cycleTerminated) return;
  TineStopCycler();
  freeServerMemory();
# endif
}
#endif
void InitCycleMutexSet(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  InitMutex(&hCyclerStartMutex);
  InitMutex(&hCyclerStopMutex);
# if defined(UNIX)
  pthread_mutex_lock(hCyclerStartMutex);
  pthread_mutex_lock(hCyclerStopMutex);
# endif
  ReleaseSystemMutex(hCyclerStartMutex);
  ReleaseSystemMutex(hCyclerStopMutex);
}

void cycleTmrTask(void)
{
  if (!gSystemInitialized) 
  { /* don't spin in not initialized yet! */
    millisleep(gSystemDelay);
    return;
  }
  if (cycleTerminated) return;
  if (cycle_busy) return;
  cycle_busy = TRUE;
  SystemCycle(foregroundTTY);
  cycle_busy = FALSE;
  if (ServerExitCondition > 3) cycleTerminated = TRUE;
}

void cycleInitTask(int isServer)
{
  if (isServer) 
  { 
    SystemInit(FALSE);
  }
  else
  {
    initClient();
  }
  gSystemInitialized = TRUE;
  MaxPollingRate = MinPollingRate = 20;
}

THREAD_TYPE cycleTmrThread(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;

  SetThreadName("tcycler");
  if (WaitForSystemMutex(hCyclerStartMutex,-1) != 0) 
  {
    feclog("cycleTmrThread: start cycler mutex unavailable !");
    return 0;
  }
  UseMultiThreadedCycler(TRUE);
  while (!cycleTerminated)
  {
    cycleTmrTask();
  }
  cycleStarted = FALSE;
  CloseServers(TRUE);
  FreeEQPModules();
# ifdef UNIX
  pthread_detach(hTmrThread);
# endif
  hTmrThread = (THRHANDLE)0;
  gTmrThreadDetached = TRUE;
  ReleaseSystemMutex(hCyclerStartMutex);
  return 0;
}
#if defined(WIN32)
SECURITY_ATTRIBUTES tmrSA;
SECURITY_DESCRIPTOR tmrSD;
DWORD dwTmrThreadId;
THRHANDLE cycleCreateThread(void)
{
  THRHANDLE hThread = 0;
  int threadPriority = -1;
  char *ptr = getenv("TINE_CYCLE_THREAD_PRIORITY");

  InitializeSecurityDescriptor(&tmrSD,SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(&tmrSD,TRUE,NULL,TRUE);
  tmrSA.nLength = sizeof(tmrSA);
  tmrSA.lpSecurityDescriptor = &tmrSD;
  tmrSA.bInheritHandle = TRUE;

  cycleTerminated = FALSE;
  hThread = (THRHANDLE)CreateThread(&tmrSA,0,cycleTmrThread,0,0,&dwTmrThreadId);
  if (ptr != NULL) threadPriority = atoi(ptr);
  if (threadPriority > 0) 
  {
    feclog("local cycle thread priority set to %d via environment",threadPriority);
    SetThreadPriority(hThread,threadPriority);
  }
  return hThread;
}
#elif defined(UNIX)
THRHANDLE cycleCreateThread(void)
{
  int cc;
  pthread_t tid;

  pthread_attr_init(gPtrCycleThreadAttr);
# ifdef FORCE_PTHREAD_SCOPE_SYSTEM
  if ((cc=pthread_attr_setscope(gPtrCycleThreadAttr,PTHREAD_SCOPE_SYSTEM)) != 0) 
  {
    feclog("Cannot set thread scope !\n");
  }
# endif
  cycleTerminated = FALSE;
  cc = pthread_create(&tid,gPtrCycleThreadAttr,cycleTmrThread,NULL);
  return cc ? 0 : (THRHANDLE)tid;
}
#elif defined(VXWORKS)
THRHANDLE cycleCreateThread(void)
{
  static int cnt=0;
  char tname[32];
  int tskid = taskIdSelf();
  int tskp = gVxwDefaultTaskPriority, tsko = gVxwDefaultTaskOptions;
  int tcks = sysClkRateGet()/100;
  cnt++;
  strncpy(tname,"tcycler",8); tname[7] = 0;
  taskPriorityGet(tskid,&tskp);
  taskOptionsGet(tskid,&tsko);
  if (tskp == 1) 
  { /* in case started from the shell prompt */
    tskp = gVxwDefaultTaskPriority;
    tsko = gVxwDefaultTaskOptions;
  }
  if (tcks < 1) tcks = 1;
  kernelTimeSlice(tcks);
  cycleTerminated = FALSE;
  return (THRHANDLE)taskSpawn(tname,tskp,tsko,gVxwDefaultTaskStack,(FUNCPTR)cycleTmrThread,0,0,0,0,0,0,0,0,0,0);
}
#endif
TINE_EXPORT int LaunchRuntimeDebugWindow(void)
{
# ifdef WIN32
  int pid = GetCurrentProcessId();
  char *c, *exe=NULL, pipeName[32],cmd[128],app[256];
  STARTUPINFO sInfo;
  PROCESS_INFORMATION pInfo; 

  memset(&sInfo,0,sizeof(sInfo));
  if (GetIpcPipeName(pipeName) != 0)
  {
    strncpy(app,GetCommandLine(),255); app[255] = 0;
    strlwr(app);
    if ((c=strstr(app,".exe")) != NULL)
    {
      *c = 0;
      if ((exe=strrchr(app,'\\')) != NULL) exe++;
    }
    if (exe == NULL || strlen(exe) == 0)
    {
      exe = "TineApp";
    }
    sprintf(pipeName,"%.20s-%d",exe,pid);
  }
  if (strlen(pipeName) == 0) return -1;
  OpenIpcPipe(pipeName);
  sprintf(cmd,"AttachFec %s",pipeName);
  CreateProcess(NULL,cmd,NULL,NULL,0,NORMAL_PRIORITY_CLASS,NULL,NULL,&sInfo,&pInfo);
# endif
	return 0;
}
TINE_EXPORT void killCycleTimer(void)
{
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    feclog("SystemKillCycleTimer: stop cycler mutex unavailable !");
    return;
  }
  if ((!cycleStarted && !gIsClosingDynamic) || cycleTerminated)
  {
    feclog("attempt to kill timer refused : cycle %s",cycleStarted ? "already terminated" : "not started");
    ReleaseSystemMutex(hCyclerStopMutex);
    return;
  }
  CloseClientLinks();
  cycleTerminated = TRUE; /* causes the cycle thread to stop => issues a SystemReset() */
  feclog("terminating local cycle");
  millisleep(gSystemDelay*2);
  ReleaseSystemMutex(hCyclerStopMutex);
  if (WaitForSystemMutex(hCyclerStartMutex,gSystemTick) != 0) 
  {
    return;
  }
  CloseClients(TRUE);
  CloseClientMutexSet();
  CloseSystemMutexSet();
  ReleaseSystemMutex(hCyclerStartMutex);
}
TINE_EXPORT void SystemWaitCycleTimer(void)
{
  if (gTmrThreadDetached) return;
  if (hTmrThread != (THRHANDLE)0) WaitForThreadTermination(hTmrThread);
  gTmrThreadDetached = TRUE;
}
TINE_EXPORT void SystemKillCycleTimer(void)
{
  killCycleTimer();
}
void TineStartCycler(void)
{
  SystemSetCycleTimer();
}
TINE_EXPORT void SystemStartCycleTimer(void)
{
  SystemSetCycleTimer();
}
void SystemSetCycleTimer(void)
{
  int cc = 0;
  InitSystemMutexSet();
  InitCycleMutexSet();
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    if (gIsRunningAsServer) feclog("SystemSetCycleTimer: stop cycler mutex unavailable !");
    return;
  }
  if (cycleStarted) ccerr(-1);
  cycleStarted = TRUE;
  cycleInitTask(FALSE);
  hTmrThread = cycleCreateThread();
  millisleep(20);
err:
  UNUSED(cc);
  ReleaseSystemMutex(hCyclerStopMutex);
  return;
}
