/*
 * File teventdata.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "project.def"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"
#include "eventtools.h"

char *usageTxt[] = {
  "\nSynchronous call: gets the specified stored event data\n",
  "teventdata <event trigger> <device name> <property> <event no.> [/c=<event server context> /i=<index> /s=<size> /f=<format> /t=<timeout>]",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))
void DataDump(char *txt,char *data,int siz,short fmt,short dir);

int main(int argc, char *argv[])
{
  int i, cc;
  int siz = 0, hsiz = 0, didx=0, tmr=1000, fmt=CF_NULL;
  UINT32 evnnr = 0;
  char *c, scratch[256];
  char actx[CONTEXT_NAME_SIZE],ctx[CONTEXT_NAME_SIZE],srv[EXPORT_NAME_SIZE],
       dev[DEVICE_NAME_SIZE],prp[PROPERTY_NAME_SIZE],trg[DEVICE_NAME_SIZE];
  DTYPE dout;
  DIMAGE img;
  PrpQueryStruct pqs;

  if (argc < 5) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  strncpy(trg,argv[1],DEVICE_NAME_SIZE);
  strncpy(scratch,argv[2],256);
  if (scratch[0] != '/')
  {
    printf("device name is not in the form /<context>/<server>/<device> !\n");
    exit(1);
  }
  c = strtok(scratch,"/");
  strncpy(ctx,c,CONTEXT_NAME_SIZE);
  strncpy(actx,c,CONTEXT_NAME_SIZE);
  if ((c=strtok(NULL,"/")) == NULL)
  {
    printf("device name is not in the form /<context>/<server>/<device> !\n");
    exit(1);
  }
  strncpy(srv,c,EXPORT_NAME_SIZE);
  if ((c=strtok(NULL,"/")) == NULL)
  {
    strcpy(dev,"#0");
  }
  else
  {
    strncpy(dev,c,DEVICE_NAME_SIZE);
  }
  strncpy(prp,argv[3],PROPERTY_NAME_SIZE);
  evnnr = strtol(argv[4],NULL,0);
  for (i=5; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(actx,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 'i':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          didx = atoi(scratch);      
        }
        continue;
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          fmt = ftoi(scratch);      
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          siz = atoi(scratch);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tmr = atoi(scratch); 
        }
        continue;
      default:
        continue;
    }
  }
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  memset(&dout,0,sizeof(DTYPE));
  sprintf(scratch,"/%.32s/%.32s/%.64s[%.64s]",ctx,srv,dev,prp);
  if ((cc=GetArchivedPropertyInformation(actx,scratch,&pqs)) != 0) return cc;
  if (siz == 0) siz = pqs.prpSize;
  if (fmt == CF_NULL) fmt = LFMT(pqs.prpFormat);
  strncpy(dout.dTag,pqs.prpTagIn,TAG_NAME_SIZE);
  dout.dArrayLength = siz;
  dout.dFormat = fmt;
  hsiz = getFormatHeaderSize(fmt);
  if ((dout.data.vptr=SystemCalloc(siz,fmtsizeof(fmt))) == NULL) return out_of_local_memory;
  if (fmt == CF_IMAGE)
  {
    img.frameBufferSize = siz;
    img.frameBuffer = dout.data.bptr;
    dout.data.vptr = &img;
  }
  if ((cc=GetArchivedEventData(actx,trg,ctx,srv,prp,dev,didx,evnnr,&dout)) == 0)
  {
    DataDump("event data",dout.data.cptr,dout.dArrayLength,dout.dFormat,RTOL);
  }
  else
  {
    printf("error: %.32s\n",cc2str(cc));
  }
  return cc;
}
