/*
 * File tglobals.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

typedef struct
{
  int lid;
  char dev[64];
  BYTE *data;
  int fmt;
  int siz;
} MONITOR;

MONITOR *globals = NULL;
char gContext[32];

void cbMonitor(int id, int cc)
{
  if (id == 0) printf("\n--- %.32s Global Keywords @ %s ---\n", gContext, GetDataTimeString((double)time(NULL), FALSE));
  if (cc != 0)
  {
    char* src = (cc & CE_SENDCODE) ? "REMOTE" : "LOCAL";
    printf("%.64s : (%s ERROR: %.32s)\n",globals[id].dev,src,cc2str(cc));
    return;
  }
  switch (globals[id].fmt)
  {
    case CF_DOUBLE:
      printf("%s : %g\n",globals[id].dev,*(double *)globals[id].data);
      break;
    case CF_FLOAT:
      printf("%s : %g\n",globals[id].dev,*(float *)globals[id].data);
      break;
    case CF_LONG:
      printf("%s : %d\n",globals[id].dev,*(SINT32 *)globals[id].data);
      break;
    case CF_SHORT:
      printf("%s : %d\n",globals[id].dev,*(short *)globals[id].data);
      break;
    case CF_TEXT:
    case CF_NAME64:
    case CF_NAME32:
    case CF_NAME16:
      printf("%s : %s\n",globals[id].dev,(char *)globals[id].data);
      break;
    default:
      printf(" %s : unexpected format %.32s\n",globals[id].dev, GetFormatAsString(globals[id].fmt));
      break;
  }
}
int getGlobalsList(char *context,NAME64DBLDBL *list,int *listSize)
{
  DTYPE dout;
  char dev[128];
  int cc;

  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_NAME64DBLDBL;
  dout.dArrayLength = *listSize;
  dout.data.vptr = list;

  sprintf(dev,"/%s/GLOBALS",context);
  cc = ExecLink(dev,"SRVGLOBALS",&dout,NULL,CA_READ|CA_MUTABLE);
  *listSize = dout.dArrayLength;

  return cc;
}
char *usageTxt[] = {
  "\nMonitors the global parameters from the designated globals server\n",
  "tglobals <context> [/d=<debug level>] or",
  "tglobals [/d=<debug level>] (default context = 'SITE')",
  "e.g. tglobals PETRA",
};
#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))
int main(int argc, char *argv[])
{
  char *c, scratch[256];
  char ctx[32],dev[132];
  NAME64DBLDBL ndd[100];
  int i, cc, siz=100;
  DTYPE dout;

  memset(ctx,0,32);
  if (argc == 1) strcpy(ctx,"SITE");
  foregroundTTY = TRUE;
  SetUserType("CMDLINE");
  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (i == 1 && *c++ != '/') 
    { /* legacy input */
      strncpy(ctx,argv[1],32);
      continue;
    }
    switch (tolower(*c))
    {
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
          foregroundTTY = TRUE;
        }
        continue;
      case 'c':
        strncpy(ctx,c,32);
        continue;
      case '?':
        for (i=0; i<nUsageLines; i++)
        {
          printf(i == 1 ? "Usage : " : "        ");
          printf("%s\n",usageTxt[i]);
        }
        printf("\n");
        exit(1);
      default:
        continue;
    }
  }  
  if (strlen(ctx) == 0) strcpy(ctx,"SITE"); // in case of simple /d=1 args
  cc = getGlobalsList(ctx,ndd,&siz);

  if (cc != 0)
  {
    printf("cannot get globals list : %.32s\n",cc2str(cc));
    exit(1);
  }
  strncpy(gContext, ctx, 32);
  globals = (MONITOR *)calloc(siz,sizeof(MONITOR));
  memset(&dout,0,sizeof(DTYPE));
  for (i=0; i<siz; i++)
  {
    strncpy(globals[i].dev,ndd[i].name,64);
    globals[i].fmt = (int)ndd[i].d1val;
    globals[i].siz = (int)ndd[i].d2val;
    globals[i].data = (BYTE *)calloc(globals[i].siz,fmtsizeof(globals[i].fmt));
    dout.dFormat = globals[i].fmt;
    dout.dArrayLength = globals[i].siz;
    dout.data.bptr = globals[i].data;
    sprintf(dev,"/%s/%s",ctx,globals[i].dev);
    globals[i].lid = recvNetGlobalEx(dev,&dout,cbMonitor,i);
  }

  /* keep the process alive */
  for (;;)
  {
    SystemCycle(TRUE);
  }
  return 0;
}

