/*
 * File thistory.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
int tget(char *dev,char *prp);
int thistory(char *context,char *keyword,char *devname,char *stoptime,char *depth);
int ReportError(int cc);
char *usageTxt[] = {
  "\nRetrieves a stored archive data for the keyword and device specfied",
  "Data can be obtained from a central archive server or from a local history,",
  "depending on how the context and keyword are specified.\n",
  "A simple keyword along with a context will contact the central archive",
  "server for the context.",
  "\n        e.g. thistory HERA HPDCCur now 1hour\n",
  "The history depth can be specified in hours, days, weeks, or months and",
  "is parsed as",
  "\n        <number><time unit>\n",
  "The history target time can be 'now', a valid unix timestamp, or a date",
  "and time string in the form",
  "\n        <day>.<month>.<year>_<hour>.<minute>.<second>\n",
  "where all entries except <day> are optional from right to left.",
  "\n        e.g. thistory HERA HPDCCur 31.05.2006 1day",
  "          or thistory HERA HPDCCur 31.05.2006_12:00:00 1hour\n",
  "A local history can be obtained by specifying the target device server",
  "along with the context (with a leading \"/\")",
  "\n        e.g. thistory /HERA/HEEKOLLI SOLLWERT HESL66i 31.05.2006 1day\n",
  "A history snapshot of an array record at a specific time can be obtained",
  "by specifying 'snapshot' as the depth parameter.",
  "\n        e.g. thistory /HERA/HEPBPM ORBIT.X #1 7.06.2006_12:00:00 snapshot",
  "\nreturns the array data record stored at or after noon on June 7th 2006.",
  "The timestamp of the record found is always displayed along with the data.\n",
  "Usage : thistory <context> <keyword> <device name> <stop time> <depth>"
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int i;
  char *context=NULL, *keyword=NULL, *device=NULL, *stoptime="NOW", *depth="1 DAY";
  char tmpdev[256];
  switch (argc)
  {
    case 2:
      keyword = argv[1];
      break;
    case 3:
      keyword = argv[1];
      depth = argv[2];
      break;
    case 4:
      keyword = argv[1];
      stoptime = argv[2];
      depth = argv[3];
      break;
    case 5:
      context = argv[1];
      keyword = argv[2];
      stoptime = argv[3];
      depth = argv[4];
      break;
    case 6:
      context = argv[1];
      keyword = argv[2];
      device = argv[3];
      stoptime = argv[4];
      depth = argv[5];
      for (i=0; i<(int)strlen(device) && isdigit(device[i]); i++);
      if (i == (int)strlen(device)) 
      {
        sprintf(tmpdev,"#%.32s",device);
        device = tmpdev;
      }
      break;
    default:
      for (i=0; i<nUsageLines; i++)
      {
        printf("%s\n",usageTxt[i]);
      }
      printf("\n");
      exit(1);
  }
  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  //  printf("\n");
  return ReportError(thistory(context, keyword, device, stoptime, depth));
}
