/*
 * File tine32.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file tine32.c
 * \brief TINE Win32 DLL API routines:
 *
 */
#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "shmalloc.h"
#ifdef MULTITHREADED
#include "threader.h"
#else
int CloseSystemMutexSet(void);
int CloseClientMutexSet(void);
#endif
#endif /* NO_DOXYGEN */

extern int gSelectOnSysPoll;
UINT_PTR idCycleTimer = 0;
int cycleTerminated = FALSE;
int cycleStarted = FALSE;
int cycle_busy = 0;
int gSystemDelay = 20;
#ifndef MULTITHREADED
char gTineCycleWindowName[256];
#endif
HWND hTineCycleWnd = 0;
#define hNULL 0
void DoEvents(int maxevents);
void flushContractTable(void);
int hasMessages(void);

void srvTmrTask(void)
{
  if (cycleTerminated) return;
  if (cycle_busy++) return;
  SystemCycle(foregroundTTY);
  cycle_busy = 0;
  if (ServerExitCondition > 3) cycleTerminated = TRUE;
}
#ifdef MULTITHREADED
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
DWORD dwTmrThreadId;
HANDLE hTmrThread = 0;
void afun(void)
{
  static int cnt = 0;
  cnt++;
  return;
}
THREAD_TYPE tmrTaskThread(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;
  SetThreadName("CycleTask");
  if (WaitForSystemMutex(hCyclerStartMutex,-1) != 0)
  {
    feclog("tmrTaskThread: start local cycler mutex unavailable !");
    return semaphore_error;
  }
  feclog("timer thread starting");
  while (!cycleTerminated)
  { /* if nobody's at home then pause a bit to spare the CPU */
    if (!ServerInitialized && nConnectionTableEntries == 0) millisleep(gSystemDelay);
    /* run thru SystemCycle(): should block on select with timeout of gSystemDelay */
    srvTmrTask();
#   ifndef NO_TINE_PIPE
    if (!cycleTerminated) 
    { /* let windows do its thing ...    */
      DoEvents(50);
    }
#   endif
  }
  feclog("timer thread ending");
  feclog("flushing client and contract lists");
  flushContractTable();
  feclog("shutting down server ports");
  CloseServers(TRUE);
  FreeEQPModules();
  //SystemStop(FALSE);
  cycleStarted = FALSE;
  ReleaseSystemMutex(hCyclerStartMutex);
  return 0;
}

HANDLE tmrCreateThread(void)
{
  HANDLE hThread = 0;
  int threadPriority = -1;
  char *ptr = getenv("TINE_CYCLE_THREAD_PRIORITY");
  SECURITY_ATTRIBUTES tmrSA;
# ifdef USE_SECURITY_DESCRIPTOR
  SECURITY_DESCRIPTOR tmrSD;
  InitializeSecurityDescriptor(&tmrSD,SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(&tmrSD,TRUE,NULL,TRUE);
  tmrSA.lpSecurityDescriptor = &tmrSD;
# else
  tmrSA.lpSecurityDescriptor = NULL;
# endif
  tmrSA.nLength = sizeof(tmrSA);
  tmrSA.bInheritHandle = TRUE;

  cycleTerminated = FALSE;
  hThread = CreateThread(&tmrSA,0,tmrTaskThread,0,0,&dwTmrThreadId);
  if (ptr != NULL) threadPriority = atoi(ptr);
  if (threadPriority > 0 && hThread != NULL) 
  {
    feclog("local cycle thread priority set to %d via environment",threadPriority);
    SetThreadPriority(hThread,threadPriority);
  }
  return hThread;
}
#else /* not MULTITHREADED */
int hCyclerStartMutex = 0;
int hCyclerStopMutex = 0;
__declspec(dllexport) LRESULT CALLBACK TineCyleWndProc(HWND hWnd, WORD iMessage, WORD wParam, LONG lParam)
{
  switch (iMessage)
  {
    case WM_TIMER:       
      srvTmrTask(); 
      if (!gSelectOnSysPoll ||   /* select fell through */
          ReturnToClientCycle || /* client loop needs attention */
          ReturnToServerCycle)   /* server loop needs attention */
        PostMessage(hTineCycleWnd,WM_TIMER,0,0);
      break;
    case WM_DESTROY:     
      KillTimer(hWnd,1);
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
int CreateTineCycleWindow(void)
{
  static WNDCLASS wndclass;
  static char  lpszWN[264];

  strncpy(lpszWN,gTineCycleWindowName,256); lpszWN[256] = 0;
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS;  /* essential style!! */
  wndclass.lpfnWndProc   = (WNDPROC) TineCyleWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hParent;
  wndclass.hIcon         = LoadIcon (hNULL, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (hNULL, IDC_ARROW);
  wndclass.hbrBackground = (HBRUSH)GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = hNULL;
  wndclass.lpszClassName = lpszWN;

  if (!RegisterClass(&wndclass)) return -1;

  if (!(hTineCycleWnd = CreateWindow (
			lpszWN, lpszWN,               /* lpClassName, lpWindowName */
			WS_BORDER|WS_CAPTION,         /* dwStyle */
			CW_USEDEFAULT, CW_USEDEFAULT, /* X, Y */
			CW_USEDEFAULT, CW_USEDEFAULT, /* nWidth, nHeight */
			hNULL, hNULL,                 /* hWndParent, hMenu */
			hParent,                      /* hInstance */
			(LPSTR)""                     /* createstruct */
      ) ) ) return -1;

  feclog("created timer Message Loop");
  return 0;
}
#endif /* MULTITHREADED */

VOID CALLBACK srvTimerProc(
  HWND hwnd,         /* handle to window */
  UINT uMsg,         /* WM_TIMER message */
  UINT_PTR idEvent,  /* timer identifier */
  DWORD dwTime       /* current system time */
)
{
  hwnd = hwnd;
  uMsg = uMsg;
  idEvent = idEvent;
  dwTime = dwTime;
  srvTmrTask();
}

TINE_EXPORT void SystemSetCycleTimer(void)
{
  SetServerExitCondition(0);
  InitSystemMutexSet();
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    feclog("SystemSetCycleTimer: stop local cycler mutex unavailable !");
    return;
  }
  if (cycleStarted) 
  {
    if (gIsRunningAsServer) feclog("attempt to start local cycler while it is still in use");
    goto err;
  }
  cycleStarted = TRUE;
  if (gIsRunningAsServer) feclog("starting local cycler");
# ifndef MULTITHREADED
  CreateTineCycleWindow();
  idCycleTimer = SetTimer(hTineCycleWnd,1,gSystemDelay,NULL);
  if (idCycleTimer == 0) 
  {
    if (gIsRunningAsServer) feclog("failed to create local cycle timer : %d",GetLastError());
  }
# else
  if (gIsRunningAsServer) feclog("creating local cycler thread");
  UseMultiThreadedCycler(TRUE);
  hTmrThread = tmrCreateThread();
  millisleep(gSystemDelay);
# endif
err:
  ReleaseSystemMutex(hCyclerStopMutex);
}
TINE_EXPORT void SystemWaitCycleTimer(void)
{
# ifdef MULTITHREADED
  WaitForThreadTermination(hTmrThread);
# endif
}
TINE_EXPORT void SystemStartCycleTimer(void)
{
  SystemSetCycleTimer();
}
TINE_EXPORT void SystemKillCycleTimer(void)
{
  int cc = 0;
# ifdef MULTITHREADED
  DWORD threadExitCode;
# endif
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    feclog("SystemKillCycleTimer: stop local cycler mutex unavailable !");
    return;
  }
  if (!cycleStarted || cycleTerminated) 
  {
    if (!cycleTerminated) feclog("attempt to kill timer refused : local cycle not started");
    ReleaseSystemMutex(hCyclerStopMutex);
    return;
  }
  feclog("terminating local cycle");
  cycleTerminated = TRUE;
  CloseClientLinks();
# ifdef MULTITHREADED
  millisleep(gSystemDelay*2);
  if (cycleStarted)
  { /* cycle thread is still active ! */
    WaitForSingleObject(hTmrThread,2000);
    cc = GetExitCodeThread(hTmrThread,&threadExitCode);
    if (cc != 0 && threadExitCode == STILL_ACTIVE)
    { /* serious problem :  */
      feclog("force timer thread termination");
      cc = TerminateThread(hTmrThread,-1);
      cycleStarted = FALSE;
      feclog("shutting down server ports");
      CloseServers(TRUE);
      FreeEQPModules();
      freeUserCommands();
      ReleaseSystemMutex(hCyclerStartMutex); /* this, of course, probably won't work ! */
    }
    else
    {
      feclog("timer thread exited normally during wait");
    }
  }
  else
  {
    feclog("timer thread exited normally prior to wait");
  }
  /* n.b. AttachServer() calls SystemKillCycleTimer() -> can't close clients !!!! */
  if (!gIsRunningAsServer)
  {
#   ifdef TINE64_EXPORTS
    CloseClientsEx(TRUE, TRUE);
#   else
    CloseClients(TRUE);
#   endif
    CloseClientMutexSet();
  }

  if (hTmrThread != 0) 
  {
    feclog("closing timer thread handle");
    CloseHandle(hTmrThread);
  }
  hTmrThread = 0;
# endif
  feclog("closing system mutex set");
  CloseSystemMutexSet();
  ReleaseSystemMutex(hCyclerStopMutex);
# ifdef THIS_IS_NEEDED
  if (WaitForSystemMutex(hCyclerStartMutex,-1) != 0) 
  {
    return;
  }
  ReleaseSystemMutex(hCyclerStartMutex);
# endif
}
TINE_EXPORT int SetSystemDelay(int msecs)
{
  if (msecs >= 10) gSystemDelay = msecs;
  if (gSystemDelay < MaxPollingRate)
  {
    MaxPollingRate = MinPollingRate = gSystemDelay;
  }
  return gSystemDelay;
}
int hasMessages(void)
{
  DWORD dw = GetQueueStatus(QS_SENDMESSAGE|QS_POSTMESSAGE);
  return dw == 0 ? FALSE : TRUE;
}
#ifndef NO_TINE_PIPE
void DoEvents(int maxevents)
{
  MSG msg;
  int i = 0;
  while (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    if (maxevents > 0 && i++ > maxevents) break;
  }
}
#endif
TINE_EXPORT int LaunchRuntimeDebugWindow(void)
{
# ifdef WIN32
  int pid = GetCurrentProcessId();
  char *c, *exe=NULL, pipeName[32],cmd[128],app[256];
  STARTUPINFO sInfo;
  PROCESS_INFORMATION pInfo; 

  memset(&sInfo,0,sizeof(sInfo));
  if (GetIpcPipeName(pipeName) != 0)
  {
    strncpy(app,GetCommandLine(),255); app[255] = 0;
    strlwr(app);
    if ((c=strstr(app,".exe")) != NULL)
    {
      *c = 0;
      if ((exe=strrchr(app,'\\')) != NULL) exe++;
    }
    if (exe == NULL || strlen(exe) == 0)
    {
      exe = "TineApp";
    }
    sprintf(pipeName,"%.20s-%d",exe,pid);
  }
  if (strlen(pipeName) == 0) return -1;
  OpenIpcPipe(pipeName);
  sprintf(cmd,"AttachFec %s",pipeName);
  CreateProcess(NULL,cmd,NULL,NULL,0,NORMAL_PRIORITY_CLASS,NULL,NULL,&sInfo,&pInfo);
# endif
	return 0;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  char *ptr;
  lpvReserved = lpvReserved;
  switch (fdwReason)
  {
  case DLL_PROCESS_ATTACH:  /* DLL being loaded */
      gIsDynamicLoad = TRUE;
#     ifdef MULTITHREADED
      InitMutex(&hCyclerStartMutex);
      InitMutex(&hCyclerStopMutex);
      if ((ptr = getenv("TINE_DBG_PIPE")) != NULL) OpenIpcPipe(ptr);
#     else
      {
        struct _timeb tb;
        _ftime(&tb);
        sprintf(gTineCycleWindowName, "TineCycle @ %hu:%hu", (int)tb.time, tb.millitm);
      }
#     endif
#     ifndef NO_TINE_PIPE
      hParent = hInstDll;
#     endif
      if ((ptr=getenv("TINE_START_CYCLER")) != NULL &&
           stricmp(ptr,"TRUE") == 0) SystemSetCycleTimer();
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  /* DLL being unloaded */
      if (lpvReserved)
      { // process is also terminating ...
      }
      // no matter whether UnloadLibrary or process termination
      // we want to do the following ...
      feclog("library unloading");
      SetServerExitCondition(1); // make sure all the EQMs' exit routines are called
#     ifndef MULTITHREADED
      if (hTineCycleWnd != NULL)
      {
        DestroyWindow (hTineCycleWnd);
        UnregisterClass(gTineCycleWindowName,hParent);
      }
#     endif
      SystemKillCycleTimer(); // make sure any links close gracefully ...
      CloseClientMutexSet();
#     ifndef NO_TINE_PIPE
      hParent = 0;
#     endif
      break;
  }
  return (TRUE);
}
