/*
 * File tineRepeater.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define APPV_MAJOR 2
#define APPV_MINOR 2
#define APPV_REVISION 0 

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"
#include "listener.h"
#include "repeater.h"

char *usageTxt[] = {
  "\ntine repeater: acquire and re-register and export information",
  "from the specified TINE device server\n",
  "tineRepeater <context> <device server> [/c=<new context> /s=<new server name> /f=<fec name> /p=<port offset> /r=<polling interval> /m=<polling mode> /l=<listener table capacity> /x=TRUE /i=TRUE /h=TRUE /t=<idle time> /sched=<p1,p2,p3,...>/d=<debug level>]\n",
  "if no new context or server are specified, the repeater server will",
  "have the same context as the target server, and a server name given",
  "by the target server name appended with '.RPT'",
  "e.g.",
  "\ttineRepeater HERA BPM\n\n",
  "will produce a server called BPM.RPT in context HERA\n",
  "\ttineRepeater PETRA undulator /c=PETRA.EXT /s=undulator\n\n",
  "will produce a server called undulator in context PETRA.EXT",
  "a port can be specified with the /p switch (default = 101)",
  "a polling interval can be specified with the /r switch (default = 1000 ms)",
  "(note: the polling interval determines the listener refresh rate for any repeated properties)",
  "the listener table capacity can be adjusted with the /l switch (default = 5000 entries)",
  "a polling mode can be specified with the /m switch, e.g. /m=DATACHANGE (default = TIMER)",
  "a specific FEC name can be given with the /f switch (default = Rpt<context%2><server%7>.<port>)",
  "exclusive read can be specified with /x=TRUE",
  "the listening idle time (default = 300 seconds) can be set with the /t switch",
  "  (e.g. /t=86400 will continue listening for 1 day even when the target value is not being read)",
  "ignore listening errors (e.g. via history configuration) can be specified with /i=TRUE",
  "remote local histories will be repeated if /h=TRUE (default = FALSE => maintain true local histories)"
  "specific properties can be monitored and scheduled on change using the /sched=<property list> switch",
  "  the property list should comma separated and contain no spaces",
};

#define nUsageLines sizeof(usageTxt)/sizeof(char *)
int rpPreSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer);
int rpPostSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer);
UINT32 wkAreaSize = 0x100000;
void dumpUsage(void)
{
  int i;
  for (i=0; i<nUsageLines; i++)
  {
    printf(i == 2 ? "Usage : " : "        ");
    printf("%s\n",usageTxt[i]);
  }
  printf("\n");
  exit(1);
}
int main(int argc, char *argv[])
{
  int i, m, p, cc;
  char *c,scratch[64],cntxt[CONTEXT_NAME_SIZE+4],devsrv[EXPORT_NAME_SIZE+4];
  if (argc > 1 && !stricmp(argv[1],RPTR_LCLCTX)) isLocalRepeater = TRUE;
  if (argc < 3 && !isLocalRepeater) dumpUsage();
  memset(thisFec,0,FEC_NAME_SIZE+4);
  memset(cntxt,0,CONTEXT_NAME_SIZE+4);
  memset(devsrv,0,EXPORT_NAME_SIZE+4);
  memset(thisFailoverMaster,0,EXPORT_NAME_SIZE+4);
  memset(thisSlaveMaster,0,EXPORT_NAME_SIZE+4);
  SetUserType("REPEATER");
  if (isLocalRepeater)
  {
    strncpy(cntxt,RPTR_LCLCTX,CONTEXT_NAME_SIZE);
    strncpy(devsrv,RPTR_LCLSRV,EXPORT_NAME_SIZE);
    strncpy(thisFec,RPTR_LCLFEC,FEC_NAME_SIZE);
    strncpy(thisContext,RPTR_LCLCTX,CONTEXT_NAME_SIZE);
    if ((p=GetPortOffset(thisFec)) < RPTR_DEFAULT_PORT)
    { /* then I must have gotten a free port -> let's try my favorite! */
      p = RPTR_DEFAULT_PORT;
    }
    rpPortOffset = p;
    gIgnoreCentralServices = TRUE;
    gTineConfigurationCoded = TRUE;
    nofeclog = TRUE;
    StartupDebug = FALSE;
  }
  else
  {
    strncpy(cntxt,argv[1],CONTEXT_NAME_SIZE);
    strncpy(devsrv,argv[2],EXPORT_NAME_SIZE-4);
    strcat(devsrv,".RPT");
  }
  for (i=3; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'c':
        if (isLocalRepeater) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(cntxt,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 's':
      {
        int isslave = FALSE;
        int isschd = FALSE;
        if (isLocalRepeater) continue;
        if (!strnicmp(c, "slave", 5))
        { 
          isslave = TRUE;
        }
        if (!strnicmp(c, "sched", 5))
        {
          isschd = TRUE;
        }
        if ((c=strchr(c, '=')) != NULL)
        {
          c++;
          if (isschd)
          {
            char *cp = NULL;
            int n = 0;
            for (n = 0, cp = strtok(c, ","); n < MAX_SCHEDULED_PROPERTIES && cp != NULL; cp = strtok(NULL, ","), n++)
            {
              strncpy(schPrpLst[n].name, cp, PROPERTY_NAME_SIZE);
            }
            nSchPrps = n;
            if (n > 0) setDefaultListenerMode(CM_DATACHANGE);
          }
          else if (isslave)
          {
            strncpy(thisSlaveMaster, c, EXPORT_NAME_SIZE);
          }
          else
          { // server name option
            strncpy(devsrv, c, EXPORT_NAME_SIZE);
          }
        }
        continue;
      }
      case 'f':
      {
        int isfec = FALSE;
        if (isLocalRepeater) continue;
        if (strnicmp(c, "failover", 8))
        { /* no match: assume this is a fecname option and not a failover_master */
          isfec = TRUE;
        }
        if ((c=strchr(c, '=')) != NULL)
        {
          c++;
          if (isfec)
            strncpy(thisFec, c, FEC_NAME_SIZE);
          else
            strncpy(thisFailoverMaster, c, EXPORT_NAME_SIZE);
        }
        continue;
      }
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case 'x':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          hasExclusiveRead = stricmp(scratch,"TRUE") == 0 ? TRUE : FALSE;
        }
        continue;
      case 'i':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          ignoreListeningErrors = stricmp(scratch,"TRUE") == 0 ? TRUE : FALSE;
        }
        continue;
      case 'h':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          repeatHistories = stricmp(scratch, "TRUE") == 0 ? TRUE : FALSE;
        }
        continue;
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        { /* truncate the number of device names to input value */
          c++;
          strncpy(scratch,c,32);
          gTruncateDeviceNames = atoi(scratch);      
        }
        continue;
      case 'r':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          rpPollingRate = atoi(scratch);
          if (rpPollingRate < 100) rpPollingRate = 100;
        }
        continue;
      case 'p':
        if (isLocalRepeater) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          rpPortOffset = atoi(scratch); 
        }
        continue;
      case 'w':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          wkAreaSize = atoi(scratch); 
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        { /* listener table size */
          c++;
          strncpy(scratch,c,32);
          m = atoi(scratch);
          if (m > 10) MexListenerTableCapacity = m;
        }
        continue;
      case 'm':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          m = GetTransferMode(scratch);
          if (m == CM_NULL) m = CM_TIMER;
          setDefaultListenerMode(m);
          m = BASEMODE(m);
          if (m != CM_TIMER) SetSuppressHeartbeatNotification(FALSE);
          if (m == CM_SINGLE) isSyncRepeater = TRUE;
        }
        continue;
      case 't':
        if ((c = strchr(c, '=')) != NULL)
        { /* idle time */
          int it;
          c++;
          strncpy(scratch, c, 32);
          it = atoi(scratch);
          SetAllowedIdleTime(it);
        }
        continue;
      default:
        continue;
    }
  }
  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
  if (!isLocalRepeater) printf("Server running (debug level %d)\n",tineDebug);
  SetConnectionTableCapacity(MexListenerTableCapacity);
  SetWorkAreaSize(wkAreaSize);
  FeclogDepth = 500;
# ifdef MULTITHREADED
  useMultiThreadedEquipmentFunctions = FALSE;
  useMultiThreadedBackgroundTasks = FALSE;
# endif
  rpPreSystemInit(argv[1],argv[2],cntxt,devsrv);
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }
  rpPostSystemInit(argv[1],argv[2],cntxt,devsrv);
  if (isLocalRepeater) 
  {
    gSystemRunningStandAlone = TRUE;
    sndRegisteredExports(&thisCycleTime);
    gSystemRunningStandAlone = FALSE;
  }
  for (;;)
  {
    SystemCycle(TRUE);
  }
  return 0;
}
