/*
 * File tputget.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
extern int displayErrorMessages; 
int tputget(char *dev,char *prp,char *datastr,int fmt,int siz,int tmr,int acc);
int tputgetEx(char *dev,char *prp,char *datastr,int fmtIn,int sizIn,int fmtOut,int sizOut,int tmr,int acc);
void dumpReturnedSyncData(int length, char *tag);
int ReportError(int cc);
char *usageTxt[] = {
  "\nSynchronous call: sends the input data to the specified TINE device",
  "server and property, and acquires the output. Sets the WRITE access bit by default.\n",
  "This call should be used for attribute style properties where the format and size are",
  "the same for both input and output data",
  "A single string can be sent by enclosing the data in quotes \"\".",
  "A data array can be separated by blanks, commas, or semi-colons.",
  "A string data array will be sent as either NAME32 or NAME16 elements.",
  "Do not enclose string arrays with quotes!\n",
  "You can specify a format by using the /f option switch and you can",
  "specify a specific access (default= READ|WRITE) by using the /a option switch and you can",
  "specify an explicit size by using the /s option switch (cannot exceed the number of elements!)\n",
  "A timeout other than the default 1000 msec can be specified via the /t option switch\n",
  "Usage : tputget <device name> <property> [/f=format /s=size /t=timeout] <data1> [<data2> <data3> ...]\n",
  "        e.g. tputget /HERA/HEPBLM/WL105K1 MODE 1",
  "        or   tputget /HERA/HEPBPM/WL105K1 MESSAGE \"hello\"",
  "\nUse quotes \"\" around device names with spaces\n",
  "        e.g. tputget \"/HERA/HEPBPM/WL197 MX\" MODE 1",
  "        or   tputget \"/HERA/HEPBPM/WL197 MX\" MESSAGE \"hello\"",
  "\nUse option /e=TRUE or FALSE to display error messages when the call is not successfull (default=TRUE)",
  "Use option /d=<debug level> to display verbose debugging information",
  "\nSee also: 'tmonitor', 'tget' for receiving data, 'tsend' for sending data.",
  "        and 'tsendrecv' for sending and receiving data with different data types and sizes",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))
int isNumPrefix(char c)
{
  if (c == '.') return TRUE;
  if (c == '+') return TRUE;
  if (c == '-') return TRUE;
  return FALSE;
}
int main(int argc, char *argv[])
{
  int i, len, cc, dargc, fmt=CF_NULL, siz = -1, tmr=1000, acc=CA_READ|CA_WRITE;
  char *s, datastr[4096];
  char *c, scratch[256], **dargv;
  int isText = FALSE;

  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
  SetUserType("CMDLINE");

  if (argc < 4) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  dargv = (char **)malloc((argc-3)*sizeof(char *));
  dargc = 0;
  for (i=3; i<argc; i++)
  { /* check for options and data */
    c = argv[i];
    if (*c++ != '/') 
    { /* a data input entry */
      dargv[dargc++] = argv[i];
      continue;
    }
    switch (tolower(*c))
    {
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          fmt = ftoi(scratch);      
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          siz = atoi(scratch);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
          foregroundTTY = TRUE;
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tmr = atoi(scratch); 
        }
        continue;
      case 'a':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          acc = getAccessType(scratch);
        }
        continue;
      case 'e':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          displayErrorMessages = strtobool(scratch);
        }
        continue;
      default:
        continue;
    }
  }
  if (dargc == 1 && fmt != CF_BOOLEAN)
  {
    s = dargv[0];
    len = (int)strlen(s);
    for (i=0; i<len; i++) if (!isdigit(s[i]) && !isNumPrefix(s[i])) break;
    if (fmt == CF_NULL && (i < len) && strncmp(s,"0x",2) && strncmp(s,"0X",2))
    {
      isText = TRUE;
    }
  }
  datastr[0] = 0;
  if (isText) strcat(datastr,"\"");
  for (i=0; i<dargc; i++) 
  {
    strcat(datastr,dargv[i]);
    if (i < dargc - 1) strcat(datastr," ");
  }
  if (isText) strcat(datastr,"\"");
  if (siz < 0) siz = dargc;
  SetUserType("CMDLINE");
  return ReportError(tputget(argv[1], argv[2], datastr, fmt, siz, tmr, acc));
}
