/*
 * File tservers.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int displayErrorMessages;
int tservers(char *context);
int tserversEx(char *cnt,char *sub,int verbose);
int ReportError(int cc);
char *usageTxt[] = {
  "\nRetrieves a list of registered servers for the given context and filter criteria\n",
  "tservers <context> [/sub=<subsystem> /os=<OS> /eqm=<equipment module> /ip=<ip address> /resp=<responsible> /imp=<importance> /v=<TRUE for vervose output>]",
  "\nsearch string can be abbreviated and occur multiple times",
  "e.g. tservers PETRA",
  "     tservers PETRA /s=RF /s=VAC /o=UNIX",
  "     tservers PETRA /eqm=CDI /ip=192.168.",
  "\nSee also: 'tproplist', 'tdevlist' for retrieving information",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int i,istart=1,cc,len,slen=0,dumpUsage=FALSE,ctxFound=FALSE,verbose=FALSE;
  char search[1024], *ctx, *c, *psearch = NULL;
  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 200;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  printf("\n");
  switch (argc)
  {
    case 1:
      dumpUsage = TRUE;
      ccerr(invalid_parameter);
    default:
      search[0] = 0;
      if (argv[1][0] != '/') 
      {
        ctx = argv[1];
        istart = 2;
        ctxFound = TRUE;
      }
      for (i=istart; i<argc; i++)
      {
        if (argv[i][0] != '/') continue;
        if (!ctxFound && tolower(argv[i][1]) == 'c')
        {
          ctxFound = TRUE;
          if ((c=strchr(argv[i],'=')) == NULL) 
          {
            dumpUsage = TRUE;
            ccerr(invalid_parameter);
          }
          c++;
          ctx = strtrm(c);
        }
        else if (argv[i][1] == 'v')
        {
          if ((c = strchr(argv[i], '=')) == NULL)
          {
            dumpUsage = TRUE;
            ccerr(invalid_parameter);
          }
          c++;
          verbose = strtobool(c);
        }
        else
        {
          slen = (int)strlen(search);
          len = (int)strlen(argv[i]);
          if (slen + len + 1 < 1024) sprintf(&search[slen],"%s ",argv[i]);
        }
      }
      if (!ctxFound)
      {
        dumpUsage = TRUE;
        ccerr(invalid_parameter);
      }
      slen = (int)strlen(search);
      if (slen > 0) psearch = search;
      break;
  }

  return ReportError(tserversEx(ctx,psearch,verbose));
  
err:
  if (dumpUsage)
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  if (cc != 0) printf("error : %s\n",erlst[cc]);
  printf("\n");
  return cc;
}
