/*
 * File ttycmd.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file ttycmd.c
 * \brief TINE console command line API
 *
 * The API routines offered here allow the developer to register his own command-line 
 * calling routines to be used for interacting with a running server at the command line.
 *
 */

#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stddef.h>
#include <math.h>
#include "prolog.h"
#include "srvcore.h"
#include "dbtools.h"
#include "srvip.h"
#include "globals.h"
#include "hstlib.h"
#include "ttycmd.h"
#include "threader.h"

/* command relevant entries: */
void ExitEQPModules(void);
static char ttyFilterKeep[64] = "";
static char ttyFilterHide[64] = "";
char traceKey[FULL_NAME_SIZE];

#define CMDBUFSIZE 20
extern int gClnCommunicationLockout;
extern int ttySplitLinesInLists;
extern int ttySkipCanceledLinks;
#ifdef VXWORKS
int suppressPrintf = FALSE;
#endif
#if defined(WIN32)
#ifndef UNDER_CE
#include <conio.h>
#endif

size_t ipcCmdSck = 0;
int gIsWindowsConsole = TRUE;
void stopPipeListener(void)
{
# ifndef NO_SERVICE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)ipcCmdSck;
  if (pipeInst == NULL) return;
  feclogEx(TINE_LOGLEVEL_INFO, "shutting down service pipe");
  DisconnectNamedPipe(pipeInst->hPipeInst);
  ipcCmdSck = 0; 
# endif
}
void resetPipeListener(PIPEINST *pipeInst)
{
# ifndef NO_SERVICE_PIPE
  if (pipeInst == NULL) return;
  DisconnectNamedPipe(pipeInst->hPipeInst);
  pipeInst->oOverlap.hEvent = pipeInst->hPipeConnectEvent;
  ConnectNamedPipe(pipeInst->hPipeInst,&pipeInst->oOverlap);
  pipeInst->isConnected = FALSE;
  srvlog("commands.log", "DETACH PIPE", "session terminated");
# else
  pipeInst = pipeInst;
# endif
}
int ipcprintf(char *str)
{
# ifndef NO_SERVICE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)ipcCmdSck;
  int i,rc,slen=(int)strlen(str),wlen=0;
  char *sptr = str;
  if (ipcCmdSck != 0 && pipeInst->isConnected) 
  {
    for (i=0; i<3; i++)
    {
      rc = WriteFile(pipeInst->hPipeInst,sptr,(DWORD)slen,(LPDWORD)&wlen,NULL);
      if (rc == 0) 
      {
        rc = GetLastError();
        if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
        break;
      }
      if (wlen < slen) 
      {
        sptr = &str[wlen];
        slen -= wlen;
      }
      else
      {  
        break;
      }
    }
  }
# endif
  return 0;
}
void GetCommand(void)
{
  static char prvcmd[CMDBUFSIZE][CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen[CMDBUFSIZE], clen = 0, prvptr = 0, curptr = 0;
  int i;
  char c;
  while (kbhit())
  {
    switch (c=getch())
    {
      case -32: /* extended (cursor movement) */
        switch (c=getch())
        {
          case 'M': /* right */
                    if (clen > len)
                    {
                      putch(cmdbuf[len++]);
                    }
                    break;
          case 'S': /* Delete */
          case 'K': /* left */
                    if (len > 0)
                    {
                      putch('\r'); putch('>'); 
                      for (i=0; i<len; i++) putch(' ');
                      cmdbuf[len--] = 0;
                      putch('\r'); putch('>'); 
                      for (i=0; i<len; i++) putch(cmdbuf[i]);
                    }
                    break;
          case 'H': /* up */
                    putch('\r'); putch('>'); 
                    for (i=0; i<len; i++) putch(' ');
                    strncpy(cmdbuf,prvcmd[prvptr],len=prvlen[prvptr]);
                    cmdbuf[len] = 0;
                    putch('\r'); putch('>'); 
                    printf(cmdbuf);
                    prvptr = (CMDBUFSIZE + prvptr - 1) % CMDBUFSIZE;
                    break;
          case 'P': /* down */
                    if (clen > len)
                    {
                      cmdbuf[clen] = 0;
                      putch('\r'); putch('>'); 
                      printf(&cmdbuf[len]);
                      len = clen;
                    }
                    break;
        }

        break;
      case 9: /* tab */
        break;
      case 8: /* backspace */
        if (len > 0)
        {
          putch('\r'); putch('>'); 
          for (i=0; i<len; i++) putch(' ');
          cmdbuf[len--] = 0;
          putch('\r'); putch('>'); 
          for (i=0; i<len; i++) putch(cmdbuf[i]);
        }
        break;
      case 10:
      case 13:
        ttyoutput("");
        if (len == 0) break;
        cmdbuf[len] = 0;
        InterpretCommand(cmdbuf);
        CMDcount++;
        curptr = (curptr + 1) % CMDBUFSIZE;
        prvptr = curptr;
        strncpy(prvcmd[prvptr],cmdbuf,len);
        prvlen[prvptr] = len;
        len = 0;
        break;
      default:
        putch(c);
        cmdbuf[len++] = c;
        if (len > CMDSIZE-2) len = 0;
        clen = MAX(clen,len);
        break;
    }
  }
  return;
}
int GetCommandEx(size_t fd,char *buf,int bufsiz)
{
# ifndef NO_SERVICE_PIPE
  int rc;
  UINT32 siz, len=0;
  PIPEINST *pipeInst = (PIPEINST *)fd;

  if (!pipeInst->isConnected) return 0;
  if ((rc=PeekNamedPipe(pipeInst->hPipeInst,NULL,0,NULL,(LPDWORD)&siz,NULL)) == 0)
  {
    resetPipeListener(pipeInst);
    return -1;
  }
  siz = MIN(siz,(UINT32)bufsiz);
  if ((rc=ReadFile(pipeInst->hPipeInst,pipeInst->chBuf,(DWORD)siz,(LPDWORD)&len,&pipeInst->oOverlap)) == 0) 
  {
    rc = GetLastError();
    if (rc == ERROR_IO_PENDING) return 0;
    resetPipeListener(pipeInst);
    return -1;
  }
  strncpy(buf,pipeInst->chBuf,len);
  memset(pipeInst->chBuf,0,bufsiz);
  buf[len] = 0;
  if (len == 0) 
  {
    return 0;
  }
  CMDcount++;
  return len;
# else
  fd = fd; buf = buf; bufsiz = bufsiz;
  return 0;
# endif
}
extern int isWindowsNT;
size_t pipeListen(const char *name)
{
# ifndef NO_SERVICE_PIPE
  PIPEINST *pipeInst=NULL; 
  HANDLE hPipeConnectEvent,hPipeReadEvent;
  int rc, cc = 0;
  int openMode = PIPE_ACCESS_DUPLEX|FILE_FLAG_OVERLAPPED;
  int pipeMode = PIPE_TYPE_MESSAGE|PIPE_READMODE_MESSAGE|PIPE_WAIT;
  char *ptr = getenv("TINE_REMOTE_PIPE");

  if (!isWindowsNT)
  {
    feclogEx(TINE_LOGLEVEL_INFO, "Remote PIPE disabled (not a Windows NT OS)");
    return 0;
  }
  if (ptr != NULL && !stricmp(ptr,"NO"))
  {
    feclog("Remote PIPE disabled via environment");
    return 0;
  }

  /* no security attribute, manual-reset event, initial state = signaled, unnamed event object */
  hPipeConnectEvent = CreateEvent(NULL,TRUE,TRUE,NULL);
  if (hPipeConnectEvent == NULL) ccerr(code_failure); 
  hPipeReadEvent = CreateEvent(NULL,TRUE,TRUE,NULL);
  if (hPipeReadEvent == NULL) ccerr(code_failure); 
  if ((pipeInst=(PIPEINST *)calloc(1,sizeof(PIPEINST))) == NULL) ccerr(out_of_server_memory);
  pipeInst->hPipeConnectEvent = hPipeConnectEvent; 
  pipeInst->hPipeReadEvent = hPipeReadEvent;     
  pipeInst->oOverlap.hEvent = pipeInst->hPipeConnectEvent; 
  pipeInst->hPipeInst = 
    CreateNamedPipe(name,openMode,pipeMode,MAXIPCCLIENTS,IPCBUFSIZE,IPCBUFSIZE,1000,NULL);
  if (pipeInst->hPipeInst == INVALID_HANDLE_VALUE) ccerr(code_failure);
  if (ConnectNamedPipe(pipeInst->hPipeInst,&pipeInst->oOverlap) != 0) ccerr(code_failure);
  pipeInst->isConnected = FALSE; 
err:
  if (cc != 0) 
  {
    rc = GetLastError();
    feclogEx(TINE_LOGLEVEL_ERROR, "CMD listen : %.32s (windows err %d)",cc2str(cc),rc);
  }
  return (size_t)pipeInst;
# else
  name = name;
  return 0;
# endif
}
size_t pipeAccept(PIPEINST *pipeInst)
{
  if (pipeInst == NULL) return 0;
  /* just test the wait status */
  if (WaitForSingleObject(pipeInst->oOverlap.hEvent,0) != 0) return 0;
  if (pipeInst->isConnected == FALSE)
  {
    srvlog("commands.log", "ATTACH PIPE", "connected");
    pipeInst->isConnected = TRUE;
    pipeInst->oOverlap.hEvent = pipeInst->hPipeReadEvent;     
  }
  return (size_t)pipeInst;
}
#elif defined(UNIX)
#include <sys/un.h>
int  ipcfd = -1;
int  ipcCmdSck = 0;
int  IPCfds[MAXIPCCLIENTS],ipcClnLstSize = 0;
int ipcprintf(char *str)
{
# ifndef NO_SERVICE_PIPE
  int i;
  for (i=0; i<ipcClnLstSize; i++) if (write(IPCfds[i],str,strlen(str)) < 0) perror("write");
# endif
  return 0;
}
int GetCommandEx(int fd,char *buf,int bufsiz)
{
  static int len=0;
  static char prvcmd[CMDSIZE];

  if ((len=read(fd,buf,bufsiz)) <= 0)
  {
    /* don't let it spin in the background ! */
    if (fd == STDIN_FILENO) foregroundTTY = FALSE;
    return len;
  }
  buf[len] = 0;
  if (strcmp(buf,"!")) 
  { /* make a note of this command */
    strncpy(prvcmd,buf,CMDSIZE);
  }
  else
  { /* re-issue the previous command */
    strncpy(buf,prvcmd,CMDSIZE);
  }
  if (fd != STDIN_FILENO) { ipcfd = fd; printf("%s",buf); }
  else ipcfd = -1;
  CMDcount++;
  return len;
}
int pipeListen(const char *name)
{
  int fd,len;
  struct sockaddr_un unix_addr;
# ifdef SVR4
  int fdarr[2];
# endif

  unlink(name);

# ifndef SVR4
  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path,name);
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) return -1;
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) return -2;
  if (chmod(unix_addr.sun_path, IPC_PERM) < 0) return -3;
  if (listen(fd, 5) < 0) return -4;
# else
  if ((fd=creat(name,S_DENYNONE)) < 0) return -1;
  if (close(fd) < 0) return -2;
  if (pipe(fdarr) < 0) return -3;
  if (ioctl(fdarr[1],I_PUSH,"connld") < 0) return -4;
  if (fattach(fdarr[1],name) < 0) return -5;
  fd = fdarr[0];
# endif /* SVR4 */

  return fd;
}
int pipeConnect(const char *name)
{
  int cc = 0, fd = 0,len;
  struct sockaddr_un unix_addr;
# ifndef SVR4
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) ccerr(-1);
  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  sprintf(unix_addr.sun_path,
    strlen(IPC_PATH) == 9 ? "%s%05d" : "%s%09d", IPC_PATH, getpid());

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if (len != 16) ccerr(-5);
# endif
  unlink(unix_addr.sun_path);
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) ccerr(-2);
  if (chmod(unix_addr.sun_path, IPC_PERM) < 0) ccerr(-3);
  memset(&unix_addr, 0, sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, name);
# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
# endif
  if (connect(fd, (struct sockaddr *)&unix_addr,len) < 0) ccerr(-4);
#else
  if ((fd=open(name,O_RDWR)) < 0) ccerr(-1);
  if (isastream(fd) == 0) 
  {
    close(fd); 
    fd = 0;
    ccerr(-2);
  }
#endif /* SVR4 */
err:
  if (cc)
  {
    if (fd > 0) close(fd);
    fd = 0;
  }
  return fd;
}

#define STALE 30
int pipeAccept(int lsnfd, uid_t *uidptr)
{
  int cc=0,clnfd,len;
  time_t staletime;
  char pipename[32];
# ifndef SVR4
  struct sockaddr_un unix_addr;
  struct stat statbuf;
# else
  struct strrecvfd recvfd;
# endif

# ifndef SVR4
  len = sizeof(unix_addr);
  if ((clnfd=accept(lsnfd, (struct sockaddr *)&unix_addr,(_socklen_t *)&len)) < 0) ccerr(-1);
  len -= sizeof(unix_addr.sun_family);

  strncpy(pipename,unix_addr.sun_path,14); pipename[14] = 0;
  if (stat(pipename,&statbuf) < 0) ccerr(-2);

# ifdef S_ISSOCK
  if (S_ISSOCK(statbuf.st_mode) == 0) ccerr(-3);
# endif

  if ((statbuf.st_mode & (S_IRWXG | S_IRWXO)) ||
      (statbuf.st_mode &  S_IRWXU) != S_IRWXU) ccerr(-4);

  staletime = time(NULL) - STALE;
  if (statbuf.st_atime < staletime ||
      statbuf.st_ctime < staletime ||
      statbuf.st_mtime < staletime) ccerr(-5);

  if (uidptr != NULL) *uidptr = statbuf.st_uid;

  unlink(pipename);
# else
  if (ioctl(lsnfd,I_RECVFD,&recvfd) < 0) ccerr(-1);
  if (uidptr != NULL) *uidptr = recvfd.uid;
  clnfd = recvfd.fd;
# endif /* SVR4 */
  return clnfd;
err:
  perror("pipeAccept");
  if (cc < -1) msglog(cc,"pipe accept error : name %.32s len %d",pipename,len);
  return cc;
}

#define IPCquit(txt) { perror(txt); printf("return %d\n",clnfd); return; }
void addIPCcln(void)
{
  int clnfd;
  uid_t uid;
  char strbuf[CMDSIZE];

  if ((clnfd = pipeAccept(ipcCmdSck,&uid)) < 0) IPCquit("accept error");
  if (ipcClnLstSize < MAXIPCCLIENTS)
  {
    IPCfds[ipcClnLstSize] = clnfd;
    ipcClnLstSize++;
    msglog(0,"Remote Session granted to uid %d",uid);
    srvlog("commands.log", "ATTACH PIPE", "connected to uid %d",uid);
  }
  else
  {
    sprintf(strbuf,"Maximum %d clients exceeded\n",MAXIPCCLIENTS);
    if (write(clnfd,strbuf,strlen(strbuf)) < 0) perror("write");
  }
}
#ifndef BSDSIGNAL
Sigfunc *bsdsignal(int signo, Sigfunc *disp)
{
  struct sigaction act;

  act.sa_handler = (void(*)())disp;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;

  if (signo != SIGALRM) act.sa_flags |= SA_RESTART;
  return sigaction(signo,&act,NULL) < 0 ? SIG_ERR : 0;
}
#endif

void sig_pipe(int signo)
{
  UNUSED(signo);
  printf("attachfec session terminated\n");
#ifdef INTEL
  if (bsdsignal(SIGPIPE, sig_pipe) == SIG_ERR) printf("signal error\n");
#endif
}
void sig_break(int signo)
{
  UNUSED(signo);
  printf("received break signal: closing down\n");
  feclog("received break signal: closing down");
  ttyquit(NULL,NULL);
}

#elif defined(DOS)
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
void GetCommand(void)
{
  static char prvcmd[CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen = 0, clen = 0;
  int i,x,y;
  char c;

  if (kbhit())
  {
    switch (c=getch())
    {
      case 0: /* extended (cursor movement) */
        switch (c=getch())
        {
          case 'M': /* right */
                    if (clen > len)
                    {
                      putch(cmdbuf[len++]);
                    }
                    break;
          case 'K': /* left */
                    if (len > 0)
                    {
                      x = wherex(); y = wherey();
                      gotoxy(x-1,y); putch(' ');
                      gotoxy(x-1,y);
                      len--;
                    }
                    break;
          case 'H': /* up */
                    x = wherex(); y = wherey();
                    for (i=len-1; i>=0; i--) { gotoxy(i,y); putch(' '); }
                    gotoxy(0,y);
                    strncpy(cmdbuf,prvcmd,len=prvlen);
                    cmdbuf[len] = 0;
                    printf(cmdbuf);
                    break;
          case 'P': /* down */
                    if (clen > len)
                    {
                      cmdbuf[clen] = 0;
                      printf(&cmdbuf[len]);
                      len = clen;
                    }
                    break;
        }
        break;
      case 9: /* tab */
        break;
      case 8: /* backspace */
        if (len > 0)
        {
          x = wherex(); y = wherey();
          gotoxy(x-1,y); putch(' ');
          gotoxy(x-1,y);
          len--;
        }
        break;
      case 10:
      case 13:
        ttyoutput("");
        cmdbuf[len] = 0;
        InterpretCommand(cmdbuf);
        CMDcount++;
        strncpy(prvcmd,cmdbuf,len);
        prvlen = len;
        len = 0;
        break;
      default:
        putch(c);
        cmdbuf[len++] = c;
        if (len > CMDSIZE-2) len = 0;
        clen = MAX(clen,len);
        break;
    }
  }
  return;
}

#elif defined(VXWORKS)
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
int GetCommandEx(int fd,char *buf,int bufsiz)
{
  static int len=0;
  static char prvcmd[CMDSIZE];

  if ((len=read(fd,buf,bufsiz)) < 0) return -1;
  buf[len] = 0;
  if (len == 0) 
  {
    /* don't let it spin in the background ! */
    return 0;
  }
  if (strcmp(buf,"!")) 
  { /* make a note of this command */
    strncpy(prvcmd,buf,CMDSIZE);
  }
  else
  { /* re-issue the previous command */
    strncpy(buf,prvcmd,CMDSIZE);
  }
  printf("%s",buf);
  CMDcount++;
  return len;
}
#elif defined(NIOS)
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
int putch(int c)
{
# ifdef NIOS_I
  nr_txchar(c);
# endif
  return c;
}
int getch(void)
{
# ifdef NIOS_I
  return nr_rxchar();
# else
  return 0;
# endif
}
void GetCommand(void)
{
  static char prvcmd[CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen = 0, clen = 0;
  int c;

  switch (c=getch())
  {
    case 27:
    case 0:
      if ((c=getch()) != '[') break;
      switch (c=getch())
      {
        case 'C': /* right */
                  if (clen > len)
                  {
                    putch(cmdbuf[len++]);
                  }
                  break;
        case 'D': /* left */
                  if (len > 0)
                  {
                    len--;
                  }
                  break;
        case 'A': /* up */
                  strncpy(cmdbuf,prvcmd,len=prvlen);
                  cmdbuf[len] = 0;
                  printf(cmdbuf);
                  break;
        case 'B': /* down */
                  if (clen > len)
                  {
                    cmdbuf[clen] = 0;
                    printf(&cmdbuf[len]);
                    len = clen;
                  }
                  break;
      }
      break;
    case 9: /* tab */
      break;
    case 8: /* backspace */
      if (len > 0)
      {
        len--;
      }
      break;
    case 10:
    case 13:
      ttyoutput("");
      cmdbuf[len] = 0;
      InterpretCommand(cmdbuf);
      CMDcount++;
      strncpy(prvcmd,cmdbuf,len);
      prvlen = len;
      len = 0;
      break;
    default:
      if (c < 0) return;
      putch(c);
      cmdbuf[len++] = c;
      if (len > CMDSIZE-2) len = 0;
      clen = MAX(clen,len);
      break;
  }
  return;
}

#else
int ipcprintf(char *str)
{
  str = str;
  return 0;
}
void execCmd(char *cmdbuf)
{
  InterpretCommand(cmdbuf);
}
#endif

int GetCommandFromStream(SOCKET fd,char *buf,int bufsiz)
{
  int len;

  if ((len=recvfrom(fd,buf,bufsiz,0,NULL,NULL)) < 0) return -1;
  buf[len] = 0;
  if (len == 0) 
  { /* don't let it spin in the background ! */
    return 0;
  }
  CMDcount++;
  return len;
}

int dbgprintf(char *str,int pstdout)
{
  int i;
  if (WaitForSystemMutex(hSndMutex,gSystemTick) != 0) return semaphore_busy;
# ifdef VXWORKS
  if (suppressPrintf) pstdout = FALSE;
# endif
  if (pstdout) printf("%s",str);
  for (i=0; i<nDBGsck; i++) send(dbgSckTbl[i].sck,str,(int)strlen(str),0);
  ReleaseSystemMutex(hSndMutex);
  return 0;
}

struct CommandListStruct {
  char cmd[10];
  int (*fcn)(char *,char *);
};
struct CommandListStruct CommandList[] = {
  { "quit",ttyquit},
  { "exit",ttyquit},
  { "help",ttyhelp},
  { "set",ttysetval},
  { "get",ttygetval},
  { "dump",ttydump},
  { "which",ttywhich},
  { "flush",ttyflush}
};
int NumCommands = sizeof(CommandList)/sizeof(struct CommandListStruct);
int CommandLevel = 0;

USRCMD *gUsrCmdLst = NULL;
int gUsrCmdLstSiz = 0;

USRFCN *gUsrFcnLst = NULL;
int gUsrFcnLstSiz = 0;

TINE_EXPORT int SystemInterpretConsoleCommand(const char *cmd)
{
  return InterpretCommand((char *)cmd);
}
int InterpretCommand(char *buf)
{
  char cmd[CMDSIZE];
  char *c,*ce=NULL,*cp=NULL,*arg1=NULL, *arg2=NULL;
  int cc=0,i,len=0;
  int argc;
  char *argv[10],*vstr,sstr[128];
  USRFCN *uf;

  if (buf == NULL) ccerr(0);
  while (*buf == ' ') buf++;
  if (strlen(buf) <= 1) ccerr(0);
  strncpy(cmd,buf,CMDSIZE-1); cmd[CMDSIZE-1] = 0;
  for (i=0; i<NumCommands; i++)
  {
    len = (int)strlen(CommandList[i].cmd);
    if (!strncmp(cmd,CommandList[i].cmd,len)) break;
  }
  if (i < NumCommands) 
  {
    if ((c=strchr(cmd,'=')) != NULL)
    {
      arg2 = c+1;
      *c = 0;
    }
    arg1 = &cmd[len];
    if (arg2 != NULL && *arg1 != ' ') ccerr(-1);
    while (*arg1 == ' ') arg1++;
    ccerr(CommandList[i].fcn(arg1,arg2));
  }
  for (i=0; i<32 && cmd[i] == ' '; i++);
  vstr = &cmd[i];
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c++ = 0;
  for (uf=gUsrFcnLst; uf != NULL; uf=uf->next)
  {
    if (!strncmp(vstr,uf->name,32))
    {
      argc = 0;
      len = (int)strlen(uf->name);
      if (c != NULL && (ce=strchr(c,')')) != NULL)
      {
        *ce = 0; strncpy(sstr,c,62); sstr[63] = 0;
        for (i=0,c=strtok_r(sstr,",",&cp); c && i<10; i++,c=strtok_r(NULL,",",&cp))
        {
          argv[i] = c;
        }
        argc = i;
      }
      i = uf->fcn(argc,argv);
      ttyoutput("result is %d",i);
      ccerr(0);  
    }
  }
err: 
  if (cc) ttyoutput("\n>Syntax: set name = value or get name");
  ttyoutput("");
  return cc;
}
#ifdef SMALL_TINE_LIB
char *ttycmdlst[] = {
  "quit","get modules","get properties(<eqm>)","get devices(<eqm>)","histories(<eqm>)","get alarmwatch","get clients",
  "get contracts","get contract(#)","flush contracts","flush  clients","get/set ServerIdle",
  "get globals", "get connections", "get connection(#)", "get messages","get/set BurstLimit",
  "get/set CycleDelay", "get memory", "get time", "get version", "get nets", "get stats", "get/set trace",
  "get semaphores", "get structures", "get users", "get/set filter", "get/set debug", "get/set logdbg", "which", "help"
};
static int nttycmds = sizeof(ttycmdlst)/sizeof(ttycmdlst[0]);
#endif
int ttyhelp(char *s1,char *s2)
{
  USRCMD *uc = NULL;
  USRFCN *uf = NULL;
  char sstr[64];
# ifdef SMALL_TINE_LIB
  int i;
# endif

  UNUSED(s1); 
  UNUSED(s2);
  ttyoutput("Currently available commands:");
# ifdef SMALL_TINE_LIB
  for (i=0; i<nttycmds; i++)
  {
    ttyoutput("\t%s",ttycmdlst[i]);
  }
# else
  ttyoutput("\tquit                   - terminates process (foreground) or service viewer (background)");
  ttyoutput("\tkill                   - terminates process as well as service viewer");
  if (gEqmListSize)
  {
    ttyoutput("\tget address            - displays the registered FEC address information");
    ttyoutput("\tget modules            - displays a list of registered equipment modules");
    ttyoutput("\tget properties(<eqm>)  - displays the registered properties for equipment module <eqm>");
    ttyoutput("\tget devices(<eqm>)     - displays the registered devices for equipment module <eqm>");
    ttyoutput("\tget devicegroups(<eqm>) - displays any registered device groups and members for equipment module <eqm>");
    ttyoutput("\tget histories(<eqm>)   - displays the registered local histories for equipment module <eqm>");
    ttyoutput("\tget alarmwatch(<eqm>)  - displays the registered alarm watch table for equipment module <eqm>");
    ttyoutput("\tget alarmsettings      - displays the current alarm settings");
    ttyoutput("\tget almoscillation     - displays the current alarm oscillation window");
    ttyoutput("\tset almoscillation     - sets the current alarm oscillation window");
    ttyoutput("\tget almcollapse        - displays the current alarm collapse window");
    ttyoutput("\tset almcollapse        - sets the current alarm collapse window");
    ttyoutput("\tget clients            - displays the current consumer list");
    ttyoutput("\tget followers(<eqm>)   - displays the current followers of equipment module <eqm>");
    ttyoutput("\tget contracts          - displays the current contract list");
    ttyoutput("\tget contract(#)        - displays contract Nr. <#> information");
    ttyoutput("\tflush contracts        - removes all attached clients and their contracts");
    ttyoutput("\tflush clients          - removes all attached clients and their contracts");
    ttyoutput("\tflush filterlinks      - removes all filter links");
    ttyoutput("\tflush alarms           - removes all alarms");
    ttyoutput("\tflush <address>        - reacquires the given link address </context/server>");
    ttyoutput("\tget serveridle         - displays server cycle idle state");
    ttyoutput("\tdump histories(<eqm>)  - writes local history configuration into a manifest file");
    ttyoutput("\tdump alarminfo         - writes local alarm and alarmwatch configuration into a manifest file");
    ttyoutput("\tget deadbands          - displays the current access deadband list");
    ttyoutput("\tset cycletrigger       - sets the system cycle monitor to TRUE or FALSE");
    ttyoutput("\tget cycletrigger       - displays the current system cycle stamp (if a monitoring a cycle trigger)");
    ttyoutput("\tset timesync           - sets TINE time synchronization to TRUE or FALSE");
    ttyoutput("\tget timesync           - displays the current setting for TINE time sychronization");
    ttyoutput("\tset serveridle = TRUE/FALSE - flags the server cycle as idle if TRUE");
  }
  if (nglobals)
  {
    ttyoutput("\tget globals            - displays the current globals list");
  }
  if (nConnectionTableEntries)
  {
    ttyoutput("\tget connections        - displays the current connection list");
    ttyoutput("\tget connection(#)      - displays connection Nr. <#> information");
    ttyoutput("\tget redirections       - displays the current redirection list");
    ttyoutput("\tget relinks            - displays the current relink (coercion) list");
    ttyoutput("\tget wildcards          - displays the current wildcard link list");
    ttyoutput("\tflush cache            - flushes the client-side address cache");
  }
  ttyoutput("\tget messages           - displays the current system message table");
  ttyoutput("\tget BurstLimit         - displays Burst Limit in packets");
  ttyoutput("\tset BurstLimit         - sets Burst Limit to num packets input");
  ttyoutput("\tget CycleDelay         - displays Cycle Delay in msec");
  ttyoutput("\tset CycleDelay         - sets Cycle Delay to msec input");
  ttyoutput("\tget MicroDelay         - displays Cycle Micro Delay in usec");
  ttyoutput("\tset MicroDelay         - sets Cycle Micro Delay to usec input");
# ifdef DOS
  ttyoutput("\tget memory             - displays available memory");
  ttyoutput("\tget watchdog           - displays current watchdog settings");
  ttyoutput("\tset watchdog           - enables (non zero value)/disables (zero) watchdog");
# endif
# if !defined(DOS) && !defined(NIOS) && !defined(VMS)
  ttyoutput("\tget mcastmask          - displays multicast mask settings");
# endif
  ttyoutput("\tget time               - displays local time");
  ttyoutput("\tget offset             - displays current timestamp offset");
  ttyoutput("\tset offset             - sets timestamp offset");
  ttyoutput("\tget version            - displays TINE version number");
  ttyoutput("\tget build              - displays build information for this library");
  ttyoutput("\tget users              - displays WRITE access user list");
  ttyoutput("\tget nets               - displays WRITE access net list");
  ttyoutput("\tget admins             - displays registered administrators");
  ttyoutput("\tget stats              - displays operational statistics");
  ttyoutput("\tget settings           - displays performance settings");
  ttyoutput("\tget memory             - displays current memory usage");
  ttyoutput("\tget env                - displays relevant environment variables");
  ttyoutput("\tget semaphores         - displays current semaphore status");
  ttyoutput("\tget sockets            - displays sockets and their settings");
  ttyoutput("\tset badpackets         - sets bad packet count");
  ttyoutput("\tget badpackets         - displays current bad packet count");
  ttyoutput("\tget structures         - displays current structure registry");
  ttyoutput("\tget threads            - displays thread priorities");
  ttyoutput("\tget filter             - displays current debug output filter string");
  ttyoutput("\tset filter             - sets debug output filter string");
  ttyoutput("\tget filterlinks        - displays assigned filter links (local history/alarm subsystems)");
  ttyoutput("\tget splitlines         - displays whether 'get connections' uses 1 or 2 lines for each table entry");
  ttyoutput("\tset splitlines         - sets whether 'get connections' uses 1 or 2 lines for each table entry");
  ttyoutput("\tget skipcanceled       - displays whether 'get connections' includes canceled links or not");
  ttyoutput("\tset skipcanceled       - sets whether 'get connections' includes canceled links or not");
  ttyoutput("\tget suspendcallbacks   - displays whether link callbacks are suspended or not");
  ttyoutput("\tset suspendcallbacks   - sets whether link callbacks are suspended or not");
  ttyoutput("\tset debug = 0,1,2,3,4  - sets debug level (0 = no debug printing)");
  ttyoutput("\tset errdbg = 0,1       - sets error debug output off/on");
  ttyoutput("\tset logdbg = 0,1       - turns debug logging off/on");
  ttyoutput("\tset almdbg = 0,1       - turns debug local alarm system off/on");
  ttyoutput("\tset hstdbg = 0,1       - turns debug local history system off/on");
  ttyoutput("\tset errloglvl = 0,1,2  - set threshold for error.log output:INFO,WARN(default),ERROR");
  ttyoutput("\tset lockout = TRUE/FALSE - simulates complete timeout conditions if TRUE");
  ttyoutput("\tset clientidle = TRUE/FALSE - flags the client cycle as idle if TRUE");
  ttyoutput("\tset probe = N          - turns on a time probe for N cycles");
  ttyoutput("\twhich <addr>           - display address information of target (e.g. 'which ENS')");
  ttyoutput("\tset trace              - trace link");
  ttyoutput("\thelp                   - display this list");
# endif /* SMALL_TINE_LIB */
  ttyoutput("\n>Extra commands:");
  for (uc=gUsrCmdLst; uc != NULL; uc=uc->next)
  {
    sstr[0] = 0;
    if (uc->access & CA_READ) strcpy(sstr,"get ");
    if (uc->access & CA_WRITE)
    {
      if (strlen(sstr)) strcat(sstr," or set ");
      else strcpy(sstr,"set ");
    }
    if (uc->iparam != NULL) strcat(sstr,"integer value");
    else if (uc->fparam != NULL) strcat(sstr,"float value");
    else if (uc->fcn != NULL) strcat(sstr,"function value");
    ttyoutput("\t%.32s\t - %s",uc->cmd,sstr);
  }
  for (uf=gUsrFcnLst; uf != NULL; uf=uf->next)
  {
    sstr[0] = 0;
    ttyoutput("\t%.32s()",uf->name);
  }
  return 0;
}

void setTraceLinkKey(char *fullname)
{
  if (fullname == NULL) return;
  strtrm(fullname);
  if (fullname[0] == '\0')
  {
    strcpy(traceKey, "\0");
    return;
  }
  if (strchr(fullname, '['))
  {
    strncpy(traceKey, fullname, FULL_NAME_SIZE);
  }
  else
  {
    size_t lengthFull;
    size_t lengthProp;

    char *prop;
    char tmpFullName[FULL_NAME_SIZE];
    traceKey[0] = 0;
    if ((prop = strrchr(fullname, '/')) == NULL) return;
    lengthProp = MIN(PROPERTY_NAME_SIZE,strlen(prop));
    lengthFull = MIN(FULL_NAME_SIZE,strlen(fullname));
    
    strncpy(tmpFullName, fullname, lengthFull - lengthProp);
    tmpFullName[lengthFull - lengthProp] = '\0';
    prop[0] = '[';
    strncat(tmpFullName, prop, lengthProp);
    strncat(tmpFullName, "]\0", 2);

    strncpy(traceKey, tmpFullName, FULL_NAME_SIZE);
  }
}
static int freeSrvMemoryOnQuit = FALSE;
TINE_EXPORT void SetFreeServerMemoryOnQuit(int value)
{
  freeSrvMemoryOnQuit = value;
}
TINE_EXPORT int ttyquit(char *s1,char *s2)
{
  UNUSED(s1);
  UNUSED(s2);
  dumpMsgTableToFile();
  quitServer();
  if (nConnectionTableEntries || nglobals) CloseClients(TRUE);
  ExitEQPModules();
  if (freeSrvMemoryOnQuit) freeServerMemory();
# ifdef DOS
  disableWatchDog();
# endif
  if (_postSysCycFcn != NULL) _postSysCycFcn();
  feclogEx(TINE_LOGLEVEL_INFO, "stopped");
  exit(0);
  return 0;
}
int ttyflush(char *var,char *s2)
{
  int len;
  UNUSED(s2);
  if (var == NULL) return -1;
  if ((len=(int)strlen(var)) <= 1) return argument_list_error;
  if (var[0] == '/')
  {
    int i=0;
    if (var[len-1] == '\n') len--;
    for (i=0; i<nConnectionTableEntries; i++)
    {
      if (conTbl[i] == NULL) continue;
      if (strnicmp(conTbl[i]->key,var,len) == 0)
      {
        if (fecAddrAcqId != -1) return semaphore_busy;
        reAcqAddr(conTbl[i]);
        return 0;
      }
    }
    return invalid_link;
  }
  if (strstr(var,"contracts"))
  {
    ttyoutput("removing all contracts and clients from tables");
    flushContractTable();
    return 0;
  }
  if (strstr(var,"clients"))
  {
    ttyoutput("removing all contracts and clients from tables");
    flushContractTable();
    return 0;
  }
  if (strstr(var,"alarms"))
  {
    ExportListStruct *el;
    ttyoutput("removing all alarms from devices");
    for (el=ExportList; el != NULL; el=el->next) RemoveAllAlarms(el->EqmName);
    return 0;
  }
# ifndef SMALL_TINE_LIB
  if (strstr(var,"filterlinks"))
  {
    ttyoutput("removing all filter links");
    flushFilterLinks();
    return 0;
  }
# endif
  if (strstr(var,"cache"))
  {
    FlushFecAddrCache();
    return 0;
  }
  return 0;
}
int ttywhich(char *var,char *s)
{
  char *c,*ctx=NULL,*srv=NULL,vstr[128];
  int i, cc=0;
  FecAddrStruct fecaddr;
# ifndef SMALL_TINE_LIB
  ExpDataStruct srvaddr;
# endif
  UNUSED(s);
  if (var == NULL) return -1;
  for (i=0; i<128 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],128);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if (strlen(vstr) == 0) ccerr(argument_list_error);
  if (stricmp(vstr,ENS_FEC) == 0 || stricmp(vstr,GENS_FEC) == 0)
  {
    if (numNameServers < 1) ccerr(name_server_unknown);
    memcpy(&fecaddr,&FecTbl[0],sizeof(FecAddrStruct));
    ttyoutput("%.16s -> %.40s @ port offset %d",fecaddr.fecName,fecaddr.strAdr,fecaddr.portOffset);
    memcpy(&fecaddr,&FecTbl[1],sizeof(FecAddrStruct));
    ttyoutput("%.16s -> %.40s @ port offset %d",fecaddr.fecName,fecaddr.strAdr,fecaddr.portOffset);
    ccerr(0);
  }
# ifndef SMALL_TINE_LIB
  if (vstr[0] == '/')
  {
    ctx = &vstr[1];
    if ((c=strchr(ctx,'/')) == NULL) ccerr(argument_list_error);
    *c++ = 0;
    srv = c;
    if ((c=strchr(srv,'/')) != NULL) *c = 0;
    if ((cc=GetServerAddr(ctx,srv,&srvaddr,&fecaddr)) != 0) ccerr(not_registered);
    ttyoutput("/%.32s/%.32s -> %.6s @ %.16s",ctx,srv,srvaddr.EqmName,srvaddr.FecName);
    ttyoutput("%.16s -> %.40s @ port offset: %d",fecaddr.fecName,fecaddr.strAdr,fecaddr.portOffset);
    ccerr(0);
  }
  if ((cc=GetFecAddr(vstr,&fecaddr)) != 0) ccerr(not_registered);
  ttyoutput("%.16s -> %.40s port %d",fecaddr.fecName,fecaddr.strAdr,fecaddr.portOffset);
# endif
err:
  if (cc != 0) dbglog("input : %.32s",cc2str(cc));
  return 0;
}
int ttydump(char *var,char *s)
{
  char *c,*arg=NULL,vstr[32];
  int i;
  UNUSED(s);
  if (var == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL || (c=strchr(vstr,'[')) != NULL) 
  {
    *c++ = 0;
    arg = c;
    if ((c=strchr(arg,')')) != NULL || (c=strchr(arg,']')) != NULL) 
    { /* well formed */
      *c = 0;
      if ((c=strchr(arg,'"')) != NULL)
      { /* explicitly entered as a string */
        arg++;
        if ((c=strchr(arg,'"')) != NULL)
        { /* well formed string */
          *c = 0;
        }
      }
    }
    strtrm(arg);
  }

  if (stricmp(vstr,"histories") == 0)
  {
    return dumpHistoryManifest(arg,TRUE);
  }
  if (stricmp(vstr, "alarminfo") == 0)
  {
    dumpAlmManifest(ALM_MF_ALL,TRUE);
    return 0;
  }
  if (stricmp(vstr, "messages") == 0)
  {
    dumpMsgTableToFile();
    return 0;
  }
  dbglog("dump %.32s is unvailable",vstr);
  return 0;
}
int getBinaryValueFromString(char *val)
{
  if (isdigit(val[0]))
  {
    return atoi(val);
  }
  else
  {
    strlwr(val);
    if (strstr(val,"true") != NULL) return TRUE;
    if (strstr(val,"yes") != NULL) return TRUE;
    if (strstr(val,"on") != NULL) return TRUE;
  }
  return FALSE;
}
int ttygetval(char *var,char *s)
{
  USRCMD *uc;
  char *c,*arg=NULL,*cp=NULL,sstr[128],vstr[64];
  int i,index=-1,len,ival[4];

  if (var == NULL) return -1;
  for (i=0; i<64 && var[i] == ' '; i++);
  if ((c=strchr(var,'\n')) != NULL) *c = 0;
  strncpy(vstr,&var[i],64);
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL || (c=strchr(vstr,'[')) != NULL) 
  {
    *c++ = 0;
    index = atoi(c);
    arg = c;
    if ((c=strchr(arg,')')) != NULL || (c=strchr(arg,']')) != NULL) 
    { /* well formed */
      *c = 0;
      if ((c=strchr(arg,'"')) != NULL)
      { /* explicitly entered as a string */
        arg++;
        if ((c=strchr(arg,'"')) != NULL)
        { /* well formed string */
          *c = 0;
        }
      }
    }
    strtrm(arg);
  }

  for (uc=gUsrCmdLst; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_READ)
    {
      if (uc->iparam != NULL)
      {
        ttyoutput("%s is %d",uc->cmd,*uc->iparam);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        ttyoutput("%s is %g",uc->cmd,*uc->fparam);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = (int)strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        ttyoutput("result is %d",i);
        return 0;  
      }
    }
  }
  strlwr(var);
# ifndef SMALL_TINE_LIB
  if (strstr(var,"filterlinks")) return dumpFilterLinks();
# endif
  if (strstr(var,"filter"))
  {
    ttyoutput("debug text filter : %.64s",ttyFilterKeep);
    ttyoutput("debug negative text filter : %.64s",ttyFilterHide);
    return 0;
  }
  if (strstr(var,"version")) return dumpVersion();
  if (strstr(var,"build")) return dumpBuildInfo();
  if (strstr(var,"semaphores")) return dumpSemaphores();
  if (strstr(var,"structures")) return dumpRegisteredStrucures();
# if !defined(DOS) && !defined(NIOS) && !defined(VMS)
  if (strstr(var,"mcastmask")) return dumpMCastAddrList();
# endif
# ifdef MULTITHREADED
  if (strstr(var,"threads")) return dumpThreadPriorities();
# endif
  if (strstr(var,"probe"))
  {
    ttyoutput("probe %d",ttimeprobe);
    return 0;
  }
  if (strstr(var,"debug"))
  {
    ttyoutput("Debug level %d",tineDebug);
    return 0;
  }
  if (strstr(var,"errdbg"))
  {
    ttyoutput("Error debug level %d",errorDebug);
    return 0;
  }
  if (strstr(var,"almdbg"))
  {
    ttyoutput("Alarm debug level %d",almdbg);
    return 0;
  }
  if (strstr(var,"almcollapse"))
  {
    ttyoutput("Alarm Collapse window %d",gAlmCollapseWindow);
    return 0;
  }
  if (strstr(var,"almoscillation"))
  {
    ttyoutput("Alarm Oscillation window %d",gAlmOscillationWindow);
    return 0;
  }
  if (strstr(var,"hstdbg"))
  {
    ttyoutput("History Debug level %d",hstdbg);
    return 0;
  }
  if (strstr(var,"lockout"))
  {
    ttyoutput("Communication lockout %s",(gClnCommunicationLockout ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"splitlines"))
  {
    ttyoutput("split lines in output lists %s",(ttySplitLinesInLists ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"skipcanceled"))
  {
    ttyoutput("skip canceled links in output lists %s",(ttySkipCanceledLinks ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"suspendcallbacks"))
  {
    ttyoutput("suspend callbacks set to %s",(GetSuspendCallbacks() ? "TRUE" : "FALSE"));
    return 0;
  }
# ifdef DOS
  if (strstr(var,"watchdog"))
  {
    ttyoutput("hardware watchdog %s",
            !watchdogBaseAddress ? "not configured" :
            watchdogEnabled ? "enabled" : "disabled");
    return 0;
  }
# endif
  if (!strcmp(var,"time")) return dumpServerTime();
  if (strstr(var,"users")) { dumpRegisteredUsers(); return 0; }
  if (strstr(var,"usergroups")) { dumpRegisteredGroups(); return 0; }
  if (strstr(var,"nets")) { dumpIPNets(); return 0; }
  if (strstr(var,"admins")) { dumpRegisteredAdmins(); return 0; }
  if (strstr(var,"globals")) { dumpGlobals(); return 0; }
  if (strstr(var,"badpackets"))
  {
     ttyoutput("bad packet counts: %d",gBadPacketCnts);
     return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     ttyoutput("Command Level %d entered",CommandLevel);
     return 0;
  }
  if (strstr(var,"messages")) return dumpMessageList();
  if (gEqmListSize)
  {
    if (strstr(var,"clients")) return dumpConsumerList(); 
    if (strstr(var,"client"))
    {
      if (index > 0) i = index;
      else if (s != NULL) i = atoi(s);
      else i = 0;
      dumpConsumerStats(i);
      return 0;
    }
    if (strstr(var,"contracts")) return dumpContractList();
    if (strstr(var,"contract"))
    {
      if (index > 0) i = index;
      else if (s != NULL) i = atoi(s);
      else i = 0;
      dumpContractStats(i);
      return 0;
    }
    if (strstr(var,"followers")) return dumpFollowerList(arg);
    if (strstr(var,"address")) return dumpFecAddr();
    if (strstr(var,"modules")) return dumpEqmList();
    if (strstr(var,"properties")) return dumpEqmProperties(arg);
    if (strstr(var,"devices")) return dumpEqmDevices(arg);
    if (strstr(var,"devicegroups")) return dumpMcaDeviceMembers(arg);
    if (strstr(var,"histories")) return dumpEqmHistories(arg);
    if (strstr(var,"alarmwatch")) return dumpEqmAlmWatch(arg);
    if (strstr(var,"alarmsettings")) return dumpAlmSettings();
  }
  if (nConnectionTableEntries)
  {
    if (strstr(var,"connections"))
    {
      dumpConnectionList();
      ttyoutput("\n");
      return 0;
    }
    if (strstr(var,"connection"))
    {
      if (index > 0) i = index;
      else if (s != NULL) i = atoi(s);
      else i = 0;
      dumpConnectionStats(i);      
      return 0;
    }
#   ifndef SMALL_TINE_LIB
    if (strstr(var,"groups"))
    {
      dumpConnectionGroups(); 
      ttyoutput("\n");
      return 0;
    }
    if (strstr(var,"redirections"))
    {
      dumpRedirectionList();
      ttyoutput("\n");
      return 0;
    }
    if (strstr(var,"wildcards"))
    {
      dumpWcList();
      ttyoutput("\n");
      return 0;
    }
    if (strstr(var,"deadbands"))
    {    
      dumpPrpDbaList();
      ttyoutput("\n");
      return 0;
    }
    if (strstr(var,"relinks") || strstr(var,"coercions"))
    {    
      dumpRelinkTable();
      ttyoutput("\n");
      return 0;
    }
#   endif
  }
# ifndef SMALL_TINE_LIB
  if (strstr(var,"settings")) return dumpSettings();
  if (strstr(var,"sockets")) return dumpSocketSettings();
  if (strstr(var,"buckets")) return dumpBucketList();
  if (strstr(var,"env")) return dumpEnv();
# endif
  if (strstr(var,"burstlimit"))
  {
    ttyoutput("Burst Limit set at %d",gBurstLimit);
    return 0;
  }
  if (strstr(var,"cycledelay"))
  {
    ttyoutput("Cycle Delay set at %d",gCycleDelay);
    return 0;
  }
  if (strstr(var,"microdelay"))
  {
    ttyoutput("Cycle Micro Delay set at %d",gCycleMicroDelay);
    return 0;
  }
  if (strstr(var,"serveridle"))
  {
    ttyoutput("Server Cycle Idle : %s",gServerCycleIdle ? "TRUE" : "FALSE");
    return 0;
  }
  if (strstr(var,"clientidle"))
  {
    ttyoutput("Client Cycle Idle : %s",gClientCycleIdle ? "TRUE" : "FALSE");
    return 0;
  }
  if (strstr(var,"cycletrigger"))
  {
    ttyoutput("cycle trigger active : %s",isUsingCycleTrigger ? "TRUE" : "FALSE");
    if (isUsingCycleTrigger)
    {
      ttyoutput("current cycle : %d",gSystemDataStampRb);
      ttyoutput("cycle offset  : %d",gCycleOffset);
    }
    return 0;
  }
  if (strstr(var,"timesync"))
  {
    ttyoutput("Using time synchronization : %s",useGlobalSynchronization ? "TRUE" : "FALSE");
    if (useGlobalSynchronization)
    {
      ttyoutput("Current Data Timestamp Offset : %g",gDataTimeStampOffset);
    }
    return 0;
  }
  if (strstr(var,"offset"))
  {
    ttyoutput("Current Data Timestamp Offset : %g",gDataTimeStampOffset);
    return 0;
  }
  if (strstr(var, "trace"))
  {
    ttyoutput("current link trace: %s", traceKey);
    return 0;
  }
  if (strstr(var,"stats")) return dumpServerStats();
  if (strstr(var, "memory"))
  {
    double m = ((double)SystemGetMemory()) / (1024 * 1024);
    ttyoutput("Current memory usage : %.3f MB", m);
    return 0;
  }
  ttyoutput("Variable unavailable: %s",var); 
  return -1;
}
int ttysetval(char *var,char *val)
{
  USRCMD *uc = NULL;
  char *c,*cp=NULL,sstr[64],vstr[64];
  int i,len,ival[4];
  int v;
  double dv;

  if (var == NULL || val == NULL) return -1;
  for (i=0; i<64 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],64);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=gUsrCmdLst; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_WRITE)
    {
      if (uc->iparam != NULL) 
      {
        *uc->iparam = atoi(val);
        ttyoutput("%s set to %d",uc->cmd,*uc->iparam);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        *uc->fparam = (float)strtod(val,NULL);
        ttyoutput("%s is %g",uc->cmd,*uc->fparam);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = (int)strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
          if (i<4) ival[i] = atoi(val);
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        ttyoutput("result is %d",i);
        return 0;
      }
    }
  }
  strlwr(var);
  if (strstr(var,"burstlimit"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    v = SetBurstLimit(v);
    ttyoutput("Burst Limit set to %d",gBurstLimit);
    setInCommandLog(var, val);
    return 0;
  }
  if (strstr(var,"cycledelay"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    v = SetCycleDelay(v);
    ttyoutput("Cycle Delay set to %d",gCycleDelay);
    setInCommandLog(var, val);
    return 0;
  }
  if (strstr(var,"microdelay"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    v = SetCycleMicroDelay(v);
    ttyoutput("Cycle Micro Delay set to %d",gCycleMicroDelay);
    setInCommandLog(var, val);
    return 0;
  }
  if (strstr(var,"offset"))
  {
    dv = atof(val);
    ttyoutput("Data Timestamp offset set from %g to %g",gDataTimeStampOffset,dv);
    gDataTimeStampOffset = dv;
    setInCommandLog(var, val);
    return 0;
  }
# ifdef DOS
  if (strstr(var,"watchdog"))
  {
     if ((v=atoi(val)) == 0) disableWatchDog();
     else enableWatchDog();
     return 0;
  }
# endif
  if (strstr(var,"probe"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    if (!ttimeprobe && v) SystemTimeProbe("probestart");
    ttimeprobe = v;    
    return 0;
  }
  if (strstr(var,"debug"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     tineDebug = v;
     ttyoutput("Debug level %d",tineDebug);
     if (v == 0)
     { /* turning debug OFF: turn them all OFF */
       if (hstdbg > 0) ttyoutput("History Debug level %d",hstdbg);
       hstdbg = 0;
       if (almdbg > 0) ttyoutput("Alarm Debug level %d",almdbg);
       almdbg = 0;
       if (errorDebug > 0) ttyoutput("Error Debug level %d",errorDebug);
       errorDebug = 0;
     }
     return 0;
  }
  if (strstr(var,"errloglvl"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     if (v>2) v = 2;
     gErrorLogLevel = v;
     ttyoutput("Error log level %d [%s]",gErrorLogLevel, logLevel2str(gErrorLogLevel));
     return 0;
  }
  if (strstr(var,"logdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     if (v > 0)
     {
       if ((dbgfp=fopen("debug.log","w")) == NULL)
         ttyoutput("Couldn't open debug log file");
       else ttyoutput("debug logging ON");
     }
     else
     {
       if (dbgfp) fclose(dbgfp);
       dbgfp = NULL;
       ttyoutput("debug logging OFF");
     }
     return 0;
  }
  if (strstr(var,"errdbg"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    errorDebug = v;
    ttyoutput("Error debug level %d",errorDebug);
    return 0;
  }
  if (strstr(var,"almdbg"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    almdbg = v;
    ttyoutput("Alarm Debug level %d",almdbg);
    return 0;
  }
  if (strstr(var,"almcollapse"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    ttyoutput("Alarm Collapse window %d",v);
    return SetAlarmCollapseWindow(v);
  }
  if (strstr(var,"almoscillation"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    ttyoutput("Alarm Oscillation window %d",v);
    return SetAlarmOscillationWindow(v);
  }
  if (strstr(var,"hstdbg"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    hstdbg = v;
    ttyoutput("History Debug level %d",hstdbg);
    return 0;
  }
  if (strstr(var,"lockout"))
  {
    gClnCommunicationLockout = getBinaryValueFromString(val);
    ttyoutput("Communication lockout set to %s",(gClnCommunicationLockout ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"splitlines"))
  {
    ttySplitLinesInLists = getBinaryValueFromString(val);
    ttyoutput("split lines in output lists set to %s",(ttySplitLinesInLists ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"skipcanceled"))
  {
    ttySkipCanceledLinks = getBinaryValueFromString(val);
    ttyoutput("skip canceled links in output lists set to %s",(ttySkipCanceledLinks ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"suspendcallbacks"))
  {
    v = getBinaryValueFromString(val);
    if (v) SetSuspendCallbacks(v); else ResetSuspendCallbacks();
    ttyoutput("suspend callbacks set to %s",(v ? "TRUE" : "FALSE"));
    return 0;
  }
  if (strstr(var,"commandlevel"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    CommandLevel = v;
    ttyoutput("Command Level %d entered",CommandLevel);
    return 0;
  }
  if (strstr(var,"badpackets"))
  {
    if ((v=atoi(val)) < 0) v = 0;
    gBadPacketCnts = v;
    ttyoutput("bad packet counts set to %d",gBadPacketCnts);
    return 0;
  }
  if (strstr(var,"filter"))
  {
    for (i=0; i<64 && val[i] == ' '; i++);
    strncpy(vstr,&val[i],64);
    if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
    if (strlen(vstr) > 0)
    {
      c = vstr;
      if (*c == '-')
      {
        memset(ttyFilterHide,0,64);
        c++; strtrm(c);
        strncpy(ttyFilterHide,c,63);
        ttyoutput("debug negative text filter %s entered",ttyFilterHide);
      }
      else
      {
        memset(ttyFilterKeep,0,64);
        strncpy(ttyFilterKeep,vstr,63);
        ttyoutput("debug text filter %s entered",ttyFilterKeep);
      }
    }
    else
    {
      memset(ttyFilterHide,0,64);
      memset(ttyFilterKeep,0,64);
      ttyoutput("all debug text filters cleared");
    }
    return 0;
  }
  if (strstr(var,"serveridle"))
  {
    gServerCycleIdle = getBinaryValueFromString(val);
    ttyoutput("Server Cycle Idle : %s",gServerCycleIdle ? "TRUE" : "FALSE");
    setInCommandLog(var, val);
    return 0;
  }
  if (strstr(var,"clientidle"))
  {
    gClientCycleIdle = getBinaryValueFromString(val);
    ttyoutput("Client Cycle Idle : %s",gClientCycleIdle ? "TRUE" : "FALSE");
    setInCommandLog(var, val);
    return 0;
  }
  if (strstr(var,"cycletrigger"))
  {
    int cc = 0, v = getBinaryValueFromString(val);
    ttyoutput("set cycle trigger to : %s",v ? "TRUE" : "FALSE");
    if (v) cc = SystemStartCycleTrigger();
    if (cc == 0) isUsingCycleTrigger = v;
    return cc;
  }
  if (strstr(var,"timesync"))
  {
    int cc = 0, v = getBinaryValueFromString(val);
    ttyoutput("set use time synchronization to : %s",v ? "TRUE" : "FALSE");
    if (v) cc = SystemStartGlobalSynchronization();
    if (cc == 0) useGlobalSynchronization = v;
    if (useGlobalSynchronization == 0) gDataTimeStampOffset = 0;
    return cc;
  }
  if (strstr(var, "trace"))
  {
    setTraceLinkKey(val);
    ttyoutput("current link trace: %s", traceKey);
    return 0;
  }
  ttyoutput("Variable %s unavailable",var);
  return -1;
}
int dbglog(char *text,...)
{
# ifndef NO_SERVICE_PIPE
# if defined(WIN32)
  PIPEINST *pipeInst = (PIPEINST *)ipcCmdSck;
  int rc,len=0;
# elif defined(UNIX)
  int i;
# endif
# endif
  static int done = FALSE;
  int cc=0;
  int use_printf = TRUE, addcr = FALSE;
  char str[512], *strs = str;
  va_list args;

  if (!done)
  {
    done = TRUE;
    InitSystemMutexSet();
  }
  if (WaitForSystemMutex(hLogMutex,-1) != 0) return semaphore_error;
  if (text[strlen(text)-1] != '\n')
  { /* use date/time prefix */
    struct timeval tv;
    double ts;
    gettimeofday(&tv,(struct timezone *)NULL);
    ts = (double)tv.tv_sec;
    ts += (double)(tv.tv_usec)/1000000.0;
    sprintf(str,"%s ",getDataTimeString(ts,FALSE));
    strs = str + strlen(str);
    addcr = TRUE;
  }
# if defined(UNIX)
  if (!foregroundTTY) use_printf = FALSE;
  va_start(args,text);
  vsnprintf(strs,256,text,args); strs[256] = 0;
  va_end(args);
# elif defined(WIN32)
  if (!gIsWindowsConsole) use_printf = FALSE;
  va_start(args,text);
  _vsnprintf(strs,256,text,args); strs[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) ccerr(string_too_long);
  va_start(args,text);
  vsprintf(strs,text,args);
  va_end(args);
# endif
  strcat(strs,addcr ? "\n>" : ">");
  if (strlen(ttyFilterKeep) && strstr(strs,ttyFilterKeep) == NULL) ccerr(0);
  if (strlen(ttyFilterHide) && strstr(strs,ttyFilterHide) != NULL) ccerr(0);
  if (dbgfp != NULL) fprintf(dbgfp,"%s",str);
# ifndef NO_SERVICE_PIPE
  if (strstr(str,"PIPE") != NULL) ccerr(0);
# if defined(WIN32)
  if (ipcCmdSck != 0 && pipeInst->isConnected) 
  {
    rc = WriteFile(pipeInst->hPipeInst,str,(UINT32)strlen(str),(LPDWORD)&len,NULL);
    if (rc == 0) 
    {
      rc = GetLastError();
      if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
    }
  }
# elif defined(UNIX)
  for (i=0; i<ipcClnLstSize; i++) if (write(IPCfds[i],str,strlen(str)) < 0) perror("write");
# endif
# endif
  dbgprintf(str,use_printf);
err:
  ReleaseSystemMutex(hLogMutex);
  return cc;
}

TINE_EXPORT int ttyoutput(char *text,...)
{
  static int done=FALSE;
  int cc=0, len = 0, use_printf = TRUE;
  char str[512];
  va_list args;

  if (!done)
  {
    done = TRUE;
    InitSystemMutexSet();
  }
  if (WaitForSystemMutex(hLogMutex,-1) != 0) return semaphore_error;
# if defined(UNIX)
  if (!foregroundTTY) use_printf = FALSE;
  va_start(args,text);
  vsnprintf(str,256,text,args);
  va_end(args);
# elif defined(WIN32)
  if (!gIsWindowsConsole) use_printf = FALSE;
  va_start(args,text);
  _vsnprintf(str,256,text,args);
  va_end(args);
# else
  if (strlen(text) < 128)
  {
    va_start(args,text);
    vsprintf(str,text,args);
    va_end(args);
  }
  else
  {
    strncpy(str,text,256);
  }
# endif
  str[256] = 0; /* make sure string length <= 256 here */
  if ((len=(int)strlen(str)) > 0) len--; /* last char in string */
  if (len >= 0 && len < 500)
  { /* someone once managed an absurd string length ... */
    if (str[len] != '\n') strcat(str, "\n"); /* make sure there's a CR-LF */
  }
  strcat(str,">"); /* append prompt */

  ipcprintf(str); 
  dbgprintf(str,use_printf);
  ReleaseSystemMutex(hLogMutex);
  return cc;
}
void setInCommandLog(char *command, char *val) 
{
# if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
  srvlog("commands.log", "LOCAL_COMMAND", "set %s = %s", command, val);
# else
  command = command;
  val = val;
# endif
}
USRCMD *getUsrCmd(char *cmd)
{
  USRCMD *uc;
  if (cmd == NULL) return NULL;
  for (uc=gUsrCmdLst; uc != NULL; uc=uc->next)
  {
    if (!strncmp(cmd,uc->cmd,32)) break;
  }
  return uc;
}
USRFCN *getUsrFcn(char *name)
{
  USRFCN *uf;
  if (name == NULL) return NULL;
  for (uf=gUsrFcnLst; uf != NULL; uf=uf->next)
  {
    if (!strncmp(name,uf->name,32)) break;
  }
  return uf;
}

#endif /* NO_DOXYGEN */
TINE_EXPORT int RegisterUserCommand(
    char *cmd,
    int (*fcn)(int,int,int,int),
    int *iparam,
    float *fparam,
    int access)
{
  USRCMD *uc;
  if (cmd == NULL) return argument_list_error;
  if (getUsrCmd(cmd) != NULL) return already_assigned;
  if ((uc=(USRCMD *)SystemCalloc(1,sizeof(USRCMD))) == NULL) return out_of_local_memory;
  if (access == 0) access = CA_READ;
  strncpy(uc->cmd,cmd,32);
  uc->fcn = fcn;
  uc->iparam = iparam;
  uc->fparam = fparam;
  uc->access = access;
  uc->next = gUsrCmdLst;
  gUsrCmdLst = uc;
  gUsrCmdLstSiz++;
  return 0;
}
TINE_EXPORT int RegisterUserFunction(
    char *name,
    int (*fcn)(int argc,char *argv[]))
{
  USRFCN *uf;
  if (name == NULL) return argument_list_error;
  if (getUsrFcn(name) != NULL) return already_assigned;
  if ((uf=(USRFCN *)SystemCalloc(1,sizeof(USRFCN))) == NULL) return out_of_local_memory;
  strncpy(uf->name,name,32);
  uf->fcn = fcn;
  uf->next = gUsrFcnLst;
  gUsrFcnLst = uf;
  gUsrFcnLstSiz++;
  return 0;
}
void freeUserCommands(void)
{
  USRCMD *uc, *ucnxt;
  USRFCN *uf, *ufnxt;
  for (uc=gUsrCmdLst; uc != NULL; uc=ucnxt)
  {
    ucnxt = uc->next;
    SystemFree(uc);
  }
  gUsrCmdLst = NULL; gUsrCmdLstSiz = 0;
  for (uf=gUsrFcnLst; uf != NULL; uf=ufnxt)
  {
    ufnxt = uf->next;
    SystemFree(uf);
  }
  gUsrFcnLst = NULL; gUsrFcnLstSiz = 0;
}
