/*
 * File twait4target.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int waitTgtOutput;
int tmonitorWait(char *dev,char *prp,char *datastr,int fmt_,int siz_,int tmr);

char *usageTxt[] = {
  "\nAsynchronous call : Monitors the specified TINE device server property for", 
  "the given target values and stops when the target is met or an error",
  "condition arises\n",
  "twait4target <device name> <property> [/s=<size> /f=<format> /t=<timout> <target1> [<target2> <target3> ...]\n",
  "A single string can be sent by enclosing the data in quotes \"\".",
  "A target array can be separated by blanks, commas, or semi-colons.",
  "A string data array will be sent as NAME64, NAME32 or NAME16 elements.",
  "Do not enclose string arrays with quotes!\n",
  "You can specify a format by using the /f option switch and you can",
  "specify an explicit size of the monitored data (default = 1) by using", 
  "the /s option switch (cannot exceed the number of elements!)\n",
  "A maximum wait timeout can be specified via the /t option switch", 
  "(default : does not timeout)\n",
  "twait4target /PETRA/MotorSteering/Motor1 Move.STATUS 0",
  "twait4target /PETRA/MotorSteering/Motor1 Move.STATUS /s=1 /f=int32 0\n",
  "Use quotes \"\" around device names with spaces\n",
  "twait4target \"/PETRA/Collimator/Outer Jaw 300M\" Position 300\n",
  "If the target value(s) is(are) omitted, a default target with the", 
  "single value of '0' is assumed. Multiple targets can also be supplied.", 
  "If thedata size is '1' then any supplied target will cause the monitor",
  "to stop in which case the exit code will specify which target caused", 
  "the monitor to halt. If the target array size is '1' (a single target)", 
  "and the data size is > '1' then all elements of the output data array", 
  "must match the target in order to stop the monitor. If both the target", 
  "array and data size array are > '1', then the monitor will halt",
  "immediately if they do not have the same dimension, otherwise each", 
  "element is compared element by element to determine if the targets are", 
  "met.",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int i, cc, len, dargc;
  int siz = 0, fmt = CF_NULL, tmr=0;
  char *s, datastr[4096];
  char *c, scratch[256], **dargv;
  int isText = FALSE;

  if (argc < 3) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 3 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  dargv = (char **)malloc((argc-3)*sizeof(char *));
  dargc = 0;
  for (i=3; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') 
    { /* a data input entry */
      dargv[dargc++] = argv[i];
      continue;
    }
    switch (tolower(*c))
    {
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          fmt = ftoi(scratch);      
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          siz = atoi(scratch);      
        }
        continue;
      case 'o':
        waitTgtOutput = TRUE;
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);
          waitTgtOutput = tineDebug;
          foregroundTTY = TRUE;
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tmr = atoi(scratch); 
        }
        continue;
      default:
        continue;
    }
  }
  if (dargc == 1)
  {
    s = dargv[0];
    len = (int)strlen(s);
    for (i=0; i<len; i++) if (!isdigit(s[i]) && (s[i] != '.')) break;
    if ((i < len) && strncmp(s,"0x",2) && strncmp(s,"0X",2))
    {
      isText = TRUE;
    }
  }
  datastr[0] = 0;
  if (isText) strcat(datastr,"\"");
  for (i=0; i<dargc; i++) 
  {
    strcat(datastr,dargv[i]);
    if (i < dargc - 1) strcat(datastr," ");
  }
  if (isText) strcat(datastr,"\"");

  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 200;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  if ((cc=tmonitorWait(argv[1],argv[2],datastr,fmt,siz,tmr)) != 0) printf("error : %s\n",erlst[cc&0xfff]);

  for (;;) SystemCycle(TRUE);
  return cc;
}
