/*
 * File client.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file client.c
 * \brief TINE Client-side API Routines
 *      
 * Below are the principal client-side API routines which allow the user to obtain data from
 * elements in the control system.  Data acquistion can either be synchronous (e.g. ExecLink()) 
 * where execution is blocked until a call completes, or asynchronous (e.g. AttachLink()) where
 * the results of a call are given to a callback routine upon completion.  Data can also be 
 * 'monitored' asynchronously either at a supplied polling rate or upon data change.  
 *
 */
#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "prolog.h"
#include "srvcore.h"
#include "globals.h"
#include "dbtools.h"
#include "srvip.h"
#include "toolkit.h"
#include "ttycmd.h"
#if defined(IPXNW) || defined(IPXBSD)
#  include "srvipx.h"
#endif
#if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
#  include "shmalloc.h"
#endif
#endif /* NO_DOXYGEN */

void _cbFltrLnk(int id, int cc,void *ref);
void wcSingleCallback(int id,int cc,void *ref);
void wcGroupCallback(int id,int cc,void *ref);
void _relinkXrefCb(int id,int cc);
void _cbLnkLsn(int id, int cc);
int initIPClient(unsigned short port);
void doIPConsumer(void);
void DataDump(char *txt,char *data,int siz,short fmt,short dir);
int addDependentLink(int linkId);
int attachLwdLink(int linkId);
int isGlobalsParameter(char *ctx,char *srv,char *kwd);
int glbcpy(DTYPE *d,GlobalListStruct *g);
void freeConTblEntryMemory(int idx);
int putMcaValues(McaTblEntry *mca,ConTblEntry *cmca,ConTblEntry *c);
void rmvMcaLinkList(void);
int isENSCall(char *devName);

short lastLnkErrSrc = 0;
short lastLnkErr = 0;
int allowCommonLinks = TRUE;
int gLastGlobalDataSize = 0;
int useMultiThreadedCycler = FALSE;
int useConnectedSockets = FALSE;
int needToAcknowledge = FALSE;
int gAlwaysRetry = TRUE;
int gAutoLinkWatchdogs = TRUE;
int gUseSynchronousNotification = FALSE;
int gAutoLinkErrorAlarms = TRUE;
int gNoMsgLog = FALSE;
int _gIsInsideNofityClient = FALSE;
int gBadPacketCnts = 0;
#define BADPCKT_CEILING 1000
static int _gSuspendCallbacks = FALSE;
static int _gSuspendCallbacksCounter = 0;
static char gUserType[USERTYPE_SIZE] = "R5CLIENT";
TINE_EXPORT void SetUserType(const char *type)
{
  if (type != NULL) strncpy(gUserType, type, USERTYPE_SIZE);
}
TINE_EXPORT char *GetUserType(void)
{
  return gUserType;
}
TINE_EXPORT void SetDefaultTransportMode(int value)
{
  useConnectedSockets = (value == TCP || value == TCP6) ? TRUE : FALSE;
}
TINE_EXPORT int GetDefaultTransportMode(void)
{
  if (gAfInet == AF_INET6) return useConnectedSockets ? TCP6 : UDP6;
  return useConnectedSockets ? TCP : UDP;
}
TINE_EXPORT char gLastStatusString[STATUS_SIZE];

TINE_EXPORT void SetAllowDependentLinks(int value)
{
  allowCommonLinks = value;
}
TINE_EXPORT int GetAllowDependentLinks(void)
{
  return allowCommonLinks;
}
/**
 * \internal
 * \brief Number of entries in conTbl
 * 
 * \see conTbl
 * \note Incremented only in IncrementConnections() 
 */
int nConnectionTableEntries = 0;
/**
 * \internal
 * \brief Number of entries in wcTable
 * 
 * \see wcTable
 * \note Incremented only in getWildCardTableId() 
 */
int nWildCardTableEntries = 0;
UINT32 nConnectionTimeouts = 0,nArrivals = 0;
int SystemLocalDataTimeSynchronization = FALSE;
int SystemSuppressHeartbeatNotification = TRUE;
char gFecAddrBuffer[64];

/**
 * \internal
 * \brief Client connection table
 */
ConTblEntry **conTbl = NULL;
          
GrpTblEntry *grpTbl = NULL;

ConTblEntry *safeGetConTblEntryPtr(int id)
{
  ConTblEntry *c = NULL;
  if (id < 0 || id > gConTblCapacity) return NULL;
  if (WaitForSystemMutex(hLinkTblMutex, -1) != 0) return NULL;
  c = conTbl[id];
  ReleaseSystemMutex(hLinkTblMutex);
  return c;
}
#define getConTblEntryPtr(id) (id >= 0 && id < nConnectionTableEntries ? conTbl[id] : NULL)
/**
 * \internal 
*/
void makeOutputDTYPEFromConTblEntry(ConTblEntry *c,DTYPE *dout)
{ 
# ifndef SMALL_TINE_LIB
  ReLnkTblEntry* r;
# endif
  if (c == NULL || dout == NULL) return;
  dout->dArrayLength = c->lastDataSize;
# ifndef SMALL_TINE_LIB
  if ((r=c->rlkEntry) != NULL)
  { /* has been re-linked */
    dout->dFormat = r->lnkFmt;
    if (r->sts == invalid_datarequest) dout->dArrayLength = r->lnkSiz;
    dout->dTag[0] = 0;
    dout->data.vptr = r->dat;
  }
  else
# endif
  { /* normal case */
    dout->dFormat = c->formatOut;
    strncpy(dout->dTag, c->strTagOut, TAG_NAME_SIZE);
    dout->data.vptr = c->dataMarshalled != NULL ? c->dataMarshalled : c->dataPtr;
  }
  dout->dTimeStamp = (double)c->dtimestamp + ((double)c->dtimestampUSEC) / 1000000.0;
  dout->dStamp = c->datastamp;
  dout->sysStamp = c->systemstamp;
  dout->xferReason = c->xferReason;
  return;
}
int isTraceLink(int id)
{
  if (!isValidConTblEntry(id)) return FALSE;
  if (strlen(traceKey) == 0) return FALSE;
  if (!strnicmp(conTbl[id]->key, traceKey, FULL_NAME_SIZE)) return TRUE;
  return FALSE;
}

void traceLinkOutput(ConTblEntry *c, char *tag, char *msg)
{
  ttyoutput("[%s (lid: %d)] %s %s", tag, c->linkId, c->key, msg);
}

int dtypecmp(DTYPE *d1,DTYPE *d2,int compareContents)
{
  int dsiz;
  if (d1->dArrayLength != d2->dArrayLength) return FALSE;
  if (d1->dFormat != d2->dFormat) return FALSE;
  if (strcmp(d1->dTag,d2->dTag)) return FALSE;
  if (!compareContents) return TRUE;
  dsiz = d1->dArrayLength * fmtsizeof(d1->dFormat) + getFormatHeaderSize(d1->dFormat);
  if (memcmp(d1->data.bptr,d2->data.bptr,dsiz)) return FALSE;
  return TRUE;
}
#ifndef SMALL_TINE_LIB
void clearDependentLinks(int linkId);
void removeDependentLink(int linkId);
void wcMembersCallback(int id,int cc,void *ref);
WcTblEntry *wcTable[LNKTBL_HASH_SIZE];
WcTblEntry *getWcLstItemWithKey(char *ctx, char *srv, char *dev, char *prp)
{
  WcTblEntry *wc;
  char key[FULL_NAME_SIZE];
  int idx;

  if (nWildCardTableEntries == 0) return NULL; /* no need to look */
  makeLinkKey(key,ctx,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);

  for (wc=wcTable[idx]; wc != NULL; wc = wc->nxt)
  {
    if (strnicmp(wc->key,key,FULL_NAME_SIZE)) continue;
    break;
  }
  return wc; 
}
WcTblEntry *getWcLstItem(char *ctx, char *srv, char *dev, char *prp,DTYPE *dout,DTYPE *din)
{
  WcTblEntry *wc;
  char key[FULL_NAME_SIZE];
  int idx;

  if (nWildCardTableEntries == 0) return NULL; /* no need to look */
  makeLinkKey(key,ctx,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);

  for (wc=wcTable[idx]; wc != NULL; wc = wc->nxt)
  {
    if (strnicmp(wc->key,key,FULL_NAME_SIZE)) continue;
    if (!dtypecmp(&wc->dout,dout,FALSE)) continue;
    if (!dtypecmp(&wc->din,din,TRUE)) continue;
    break;
  }
  return wc; 
}
WcTblEntry *appendWcLst(char *ctx, char *srv, char *dev, char *prp,DTYPE *dout,DTYPE *din)
{
  int idx;
  char key[FULL_NAME_SIZE];
  WcTblEntry *wc=getWcLstItem(ctx,srv,dev,prp,dout,din);
  if (wc != NULL) return wc;
  if ((wc=(WcTblEntry *)SystemCalloc(1,sizeof(WcTblEntry))) == NULL)
  {
    msglog(out_of_local_memory,"appendWcLst: out of memory");
    return NULL;
  }
  makeLinkKey(key,ctx,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  strncpy(wc->key,key,FULL_NAME_SIZE);
  memcpy(&wc->dout,dout,sizeof(DTYPE));
  memcpy(&wc->din,din,sizeof(DTYPE));
  wc->nxt = wcTable[idx];
  wcTable[idx] = wc;
  nWildCardTableEntries++;
  msglog(0,"append %.192s to the wildcard link list",key);
  if (gIsRunningAsServer) feclog("append %.192s to the wildcard link list",key);
  return wc;
}
void flushWcLst(void)
{
  WcTblEntry *wc, *wcn;
  int i, k;

  if (nWildCardTableEntries == 0) return;
  if (tineDebug) dbglog("flush the wilcdard link list (%d entries)",nWildCardTableEntries);
  for (i=0; i<LNKTBL_HASH_SIZE; i++)
  {
    for (wc=wcTable[i]; wc != NULL; wc = wcn)
    {
      if (wc->hasGensParent) 
      {
        for (k=0; k<wc->length; k++) _closeLink(wc->listId[k]);
      }
      if (wc->linkId > 0 && wc->linkId < nConnectionTableEntries)
      { /* revert */
        conTbl[wc->linkId]->cbfcn = wc->cbfcn;
        conTbl[wc->linkId]->cbRef = wc->cbRef;
        conTbl[wc->linkId]->hasReference = wc->hasReference;
        conTbl[wc->linkId]->wcLink = NULL;
        if (wc->hasGensParent) conTbl[wc->linkId]->dataPtr = wc->dout.data.cptr;
      }
      wcn = wc->nxt;
      SystemFree(wc->list);
      SystemFree(wc->status);
      SystemFree(wc->listId);
      if (wc->hasGensParent) SystemFree(wc->data);
      SystemFree(wc);
    }
    wcTable[i] = NULL;
  }
  nWildCardTableEntries = 0;
  for (i=0; i<nConnectionTableEntries; i++) conTbl[i]->wcLink = NULL; 
}
int dumpWcList(void)
{
  WcTblEntry *wc=NULL;
  int i;
  ttyoutput(" Current Wildcard Table\n");
  for (i=0; i<LNKTBL_HASH_SIZE; i++) for (wc=wcTable[i]; wc != NULL; wc = wc->nxt)
  {
    if (wc->cbfcn == (TCBFCNP)wcMembersCallback) continue;
    ttyoutput("%.192s (output %d %s, input %d %s) : %s",
      wc->key,wc->dout.dArrayLength,GetFormatAsString(wc->dout.dFormat),
      wc->din.dArrayLength,GetFormatAsString(wc->din.dFormat),(wc->active ? "active" : "not active"));
    if (wc->length > 0)
    {
      int k;
      if (wc->hasGensParent)
      {
        ttyoutput("  members:");
        for (k=0; k<wc->length; k++) ttyoutput("  %.32s",wc->list[k].name);
      }
      else
      {
        ttyoutput("  distributed devices:");
        for (k=0; k<wc->length; k++) ttyoutput("  %.64s",wc->list[k].name);
      }
    }
  }
  return 0;
}
int assignWildcardLink(int lid)
{
  int keepTgtLink;
  WcTblEntry *wc, *pwc;
  ConTblEntry *c=getConTblEntryPtr(lid);
  DTYPE dout, din;
  if (c == NULL) return invalid_link;
  if (c->mode & CM_GROUPED) return illegal_mode;
  pwc = getWcLstItemWithKey(c->ctxName,c->expName,c->devName,c->prpName);
  c->mode &= ~(CM_SYNCNOTIFY); /* this can't be turned on for wildcard links */
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = c->formatOut; dout.dArrayLength = c->sizeOut; dout.data.cptr = c->dataPtr;
  din.dFormat = c->formatIn; din.dArrayLength = c->sizeIn; din.data.cptr = c->dataIn;
  wc = getWcLstItem(c->ctxName,c->expName,c->devName,c->prpName,&dout,&din);
  if (wc == NULL) wc = appendWcLst(c->ctxName,c->expName,c->devName,c->prpName,&dout,&din);
  if (wc == NULL) return out_of_local_memory;
  keepTgtLink = (c->cbfcn == (TCBFCNP)wcMembersCallback);
  if (!strcmp(FecTbl[c->fecIdx].fecName,"GENS")) wc->hasGensParent = TRUE;
  wc->mode = c->mode;
  if (wc->hasGensParent)
  {
    int i, mode = c->mode;
    int dsize = c->sizeOut * fmtsizeof(c->formatOut);
    DTYPE dout,din;
    char dev[128];
    c->tocounter = -1; /* wait signal */
    c->mode = CM_REGISTER;
    c->heartbeat = HBT_REPLY_PENDING;
    keepTgtLink = TRUE;
    if (wc->length == 0)
    { /* members as yet unknown */
      NAME32 mems[64] = {{{ 0 }}};
      NAME64 *pmems = NULL;
      char *pnam;
      int n = -1;
      if (pwc != NULL)
      { /* but we've already obtained this information !*/
        n = pwc->length;
        pmems = pwc->list;
      }
      else
      { /* need to acquire */
        n = 64;
        if (GetSystemGroupMembers(c->ctxName,c->expName,mems,&n) != 0) n = -1;
      }
      if (n > 0)
      {
        if ((wc->list=(NAME64 *)SystemCalloc(n,sizeof(NAME64))) == NULL) return out_of_local_memory;
        if ((wc->status=(int *)SystemCalloc(n,sizeof(int))) == NULL) return out_of_local_memory;
        if ((wc->listId=(int *)SystemCalloc(n,sizeof(int))) == NULL) return out_of_local_memory;
        if ((wc->data=(BYTE *)SystemCalloc(n,dsize)) == NULL) return out_of_local_memory;
        wc->length = n;
        for (i=0; i<n; i++)
        {
          pnam = (pmems == NULL) ? mems[i].name : pmems[i].name;
          strncpy(wc->list[i].name,pnam,32);
        }
      }
      else
      {
        keepTgtLink = FALSE;
        wc->hasGensParent = FALSE;
        goto done;
      }
    }
    DTYPEZERO(dout); DTYPEZERO(din);
    din.dArrayLength = c->sizeIn; din.dFormat = c->formatIn; din.data.bptr = (BYTE *)c->dataIn;
    dout.dArrayLength = c->sizeOut; dout.dFormat = c->formatOut;
    wc->numActive = wc->numPending = wc->length;
    for (i=0; i<wc->length; i++)
    { /* start the individual links */
      wc->status[i] = not_ready;
      dout.data.bptr = wc->data + (size_t)i*dsize;
      sprintf(dev,"/%.32s/%.32s/*",c->ctxName,wc->list[i].name);
      wc->listId[i] = _attachLink(dev,c->prpName,&dout,&din,c->access,
                        c->pollingInterval,(void *)wcMembersCallback,mode,i,wc,UNASSIGNED_INDEX);
      if (wc->listId[i] < 0) wc->status[i] = link_error;
    }
  }
done:
  wc->active = TRUE;
  wc->linkId = lid;
  wc->cbId = c->cbId;
  wc->cbfcn = (TCBFCNP)c->cbfcn; /* re-assign below */
  wc->hasReference = (BYTE)c->hasReference;
  wc->cbRef = c->cbRef;
  if (!keepTgtLink)
  { /* reset these to the wildcard callback  */
    c->cbId = lid;
    c->cbRef = wc;
    c->cbfcn = (TCBFCNP)wcSingleCallback;
  }
  c->hasReference = TRUE;
  c->isWildcardLink = TRUE;
  c->wcLink = wc;
  return 0;
}

#define IsMcaParentLink(c) ((c)->cbfcn == (TCBFCNP)_cbMcaXRef)
#endif /* SMALL_TINE_LIB */

cbXRefTblEntry **cbXRefTbl = NULL;
int nCbXRefTblEntries = 0;

extern ConTblEntry connNameServer;
#ifndef clnWorkArea
/**
 * \internal 
 * \brief Buffer for incoming and outgoing data.
 * \todo Overwritten with a macro (DOS  hack) 
 */
BYTE *clnWorkArea = NULL;
BYTE *mcstWorkArea = NULL;
#endif
/**
 * \internal
 * \brief Size of clnWorkArea in bytes.
 * \todo Overwritten with a macro (DOS hack)
 */
#ifndef clnWorkAreaSize 
UINT32 clnWorkAreaSize = MIN_WORK_SIZE;
#endif
BYTE *clientbuf;

char *modetoa[] = { "CANCEL","SINGLE","DATACHANGE","TIMER","REGISTER","RECEIVE","EVENT","INVALID" };

GlobalListStruct **glbTbl;
int GlobalHeartbeat = GLOBALHEARTBEAT;
AccessLockListItem *accessLockLst = NULL;

UINT32 LastCompletionDataSize = 0;
int LastCompletionDataType = CF_NULL;
int LastIncomingLinkId = -1;

/* 
 * General Purpose Routines:
 */
#ifdef WIN32
void timeSinceLastCall(char *text,...)
{  	
  char str[512];
  va_list args;
  struct timeb tb = { 0, 0 };
  static struct timeb tb0 = { 0, 0 };
  int dt;
  va_start(args,text);
  _vsnprintf(str,256,text,args); str[256] = 0;
  va_end(args);

  ftime(&tb);
  if (tb0.time == 0) tb0 = tb;
  dt = (int)((tb.time-tb0.time)*1000 + (tb.millitm-tb0.millitm));
  if (dt > 16) printf("%s : %d msec\n",str,dt);
  tb0 = tb;
}
BOOL winBreakHndlr(DWORD fdwCtrlType)
{
  switch( fdwCtrlType ) 
  { 
    case CTRL_C_EVENT: 
    case CTRL_CLOSE_EVENT: 
    case CTRL_BREAK_EVENT:
      ttyquit(NULL,NULL);
      return( TRUE ); 
    default: 
      return FALSE; 
  } 
}
#endif
char *gStatusSourceString[2] = { "LCL", "RMT" };
void GetErrorString(int cc,char *str)
{
  if (cc < numErr) strncpy(str,cc2str(cc),32);
  else sprintf(str,"error %d",cc);
}
/**
 * \internal 
 * \brief makes a link table key
 *
 * \param key is a passed buffer to hold the key (thread safe) or 
 * NULL (not thread safe) if the local static buffer is to be used
 * \param context 
 * \param server
 * \param device
 * \param property
 *
 * \return a pointer to the link key (either the pointer passed or the internal buffer)
 */
char *makeLinkKey(char *key,char *context,char *server,char *device,char *property)
{ 
  static char lclKeyBuf[256];
  if (key == NULL) key = lclKeyBuf;
  snprintf(key,FULL_NAME_SIZE,"/%.32s/%.32s/%.64s[%.64s]",context,server,device,property);
  return key;
}
char *getLinkKey(ConTblEntry *c)
{
  if (c == NULL) return "--non existent--";
  if (strlen(c->key) != 0) return c->key;
  return makeLinkKey(c->key, c->ctxName, c->expName, c->devName, c->prpName);
}
TINE_EXPORT int GetConnectionTable(ConTblInfo *tbl,int *tblSize)
{
  int k, n = 0;
  if (tbl == NULL || tblSize == NULL) return argument_list_error;
  if (WaitForSystemMutex(hLinkTblMutex,-1) != 0) return mutex_error;
  for (k=0; k<nConnectionTableEntries; k++)
  { /* only the active monitored links ... */
    if (conTbl[k]->mode == CM_CANCEL) continue;
    if (conTbl[k]->mode == CM_REGISTER) continue;
    if (conTbl[k]->mode == CM_SINGLE) continue;
    if (n >= *tblSize) break;
    memcpy(&tbl[n],conTbl[k],CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE+PROPERTY_NAME_SIZE); 
    memset(tbl[n].eqm,0,EQM_NAME_SIZE);
    strncpy(tbl[n].eqm,SrvTbl[conTbl[k]->srvIdx].EqmName,EQM_NAME_SHORTSIZE);
    tbl[n].mode = conTbl[k]->mode;
    tbl[n].access = conTbl[k]->access;
    tbl[n].formatIn = conTbl[k]->formatIn;
    tbl[n].formatOut = conTbl[k]->formatOut;
    tbl[n].sizeIn = conTbl[k]->sizeIn;
    tbl[n].sizeOut = conTbl[k]->sizeOut;
    memcpy(tbl[n].strTagIn,conTbl[k]->strTagIn,TAG_NAME_SIZE);
    memcpy(tbl[n].strTagOut,conTbl[k]->strTagOut,TAG_NAME_SIZE);
    tbl[n].pollingInterval = conTbl[k]->pollingInterval;
    tbl[n].heartbeat = conTbl[k]->heartbeat;
    tbl[n].starttime = conTbl[k]->starttime;
    tbl[n].dtimestamp = conTbl[k]->dtimestamp;
    tbl[n].dtimestampMSEC = conTbl[k]->dtimestampUSEC/1000;
    tbl[n].counter = conTbl[k]->counter;
    tbl[n].linkStatus = conTbl[k]->linkStatus;
    n++;
  }
  *tblSize = n;
  ReleaseSystemMutex(hLinkTblMutex);
  return 0;
}
int ttySplitLinesInLists = FALSE;
int ttySkipCanceledLinks = TRUE;
/**
 * \intern used internally: no pointer checking ...
 * \param lnkId is the link id
 * \param strSts should be at least 256 char
 */
int getLinkOutputStatus(int lnkId,char *strSts,int maxKeyLength,int splitlines)
{
  ConTblEntry *c;
  char sts[256],scratch[256],fmtstr[96];
  char *septok = splitlines ? "\n\t" : " : ";
  int baseMode;
  if (lnkId < 0 || lnkId >= nConnectionTableEntries) return invalid_index;
  if ((c=conTbl[lnkId]) == NULL) return invalid_link;
  baseMode = BASEMODE(c->mode);
  if (c->boundToId > 0)
  {
    sprintf(sts,"BOUND to link [%d]",c->boundToId);
    if (c->mcaIndex > 0) 
    {
      sprintf(scratch," as MCA element %d",c->mcaIndex-1);
      strcat(sts,scratch);
    }
  }
  else
  {
    strcpy(sts, baseMode < CM_SINGLE ? "-" : c->tocounter ? "DOWN" : "UP");
  }
  if (maxKeyLength < 1) maxKeyLength = 40;
  sprintf(fmtstr,"[%%4d] %%-%d.196s %%s %%6d value(s) @ %%5d msec (%%8s -%%s-) (cbId: %%10d) - %%s",maxKeyLength);
  snprintf(strSts,256,fmtstr,
    lnkId,c->key,septok,(int)c->sizeOut,(int)c->pollingInterval,GetModeAsString(c->mode),
    getProtocolAsString(c->inetProtocol),c->cbId,sts);
  return 0;
}
int getMaxLinkKeyStrLen(void)
{
  int k, klen, len = 0;
  for (k=0; k<nConnectionTableEntries; k++)
  {
    if (conTbl[k] == NULL) continue;
    if (ttySkipCanceledLinks && conTbl[k]->mode == CM_CANCEL) continue;
    if ((klen=(int)strlen(conTbl[k]->key)) > len) len = klen;
  }
  return len;
}
TINE_EXPORT int GetConnectionList(char *lstbuf,int bufsize)
{
  int k, n, nbytes=0, keylen;
  char linkstatus[256];

  if (lstbuf == NULL || bufsize < 256) return argument_list_error;
  n = bufsize / 256;
  if (n < 1) return buffer_too_small;
  lstbuf[0] = 0;
  keylen = getMaxLinkKeyStrLen();
  if (WaitForSystemMutex(hLinkTblMutex,-1) != 0) return mutex_error;
  for (k=0; k<nConnectionTableEntries; k++)
  {
    if (nbytes > bufsize - 256) break;
    if (conTbl[k] != NULL && conTbl[k]->mode == CM_CANCEL) continue;
    if (getLinkOutputStatus(k,linkstatus,keylen,FALSE) != 0) continue;
    strcat(lstbuf,linkstatus); strcat(lstbuf,"\n");
    nbytes = (int)strlen(lstbuf);
  }
  ReleaseSystemMutex(hLinkTblMutex);
  return 0;
}
int outputConnectionList(void)
{
  int k, keylen;
  char linkstatus[256];

  keylen = getMaxLinkKeyStrLen();
  ttyoutput(" Current Connection Table\n");
  for (k=0; k<nConnectionTableEntries; k++)
  {
    if (ttySkipCanceledLinks && conTbl[k] != NULL && conTbl[k]->mode == CM_CANCEL) continue;
    if (getLinkOutputStatus(k,linkstatus,keylen,ttySplitLinesInLists) != 0) continue;
    ttyoutput("%.256s\n",linkstatus);
  }
  return 0;
}
#ifndef SMALL_TINE_LIB
int outputRedirectionList(void)
{
  RdrTblEntry *r=NULL;
  int i;

  ttyoutput(" Current Redirection Table\n");

  for (i=0; i<REDIR_HASH_SIZE; i++)
  {
    for (r=rdrLnkTbl[i]; r != NULL; r=r->nxt)
    {
      ttyoutput("/%.32s/%.32s/%.64s[%.64s] -> /%.32s/%.32s/%.64s[%.64s]",
          r->srcContext,r->srcServerName,r->srcDeviceName,r->srcProperty,
          r->dstContext,r->dstServerName,r->dstDeviceName,r->dstProperty);
    }

  }
  return 0;
}
char *grpSyncLevelStr[] = { 
  "not synchronized", "sychronization startup", "synchronize event notification", 
  "synchronize cycle offset", "synchronize effective update interval", "is synchronized" };
int outputConnectionGroups(void)
{
  double ts;
  int i,k,off;
  char *tag;
  ConTblEntry *c;
  GlobalListStruct *glb;
  GrpTblEntry *g;
  GrpMember *gm;

  ttyoutput(" Current Group Table\n");
  for (g=grpTbl,k=0; g != NULL; g=g->next,k++)
  {
    ttyoutput(" Group %d Members :\n",k);
    for (gm=GetGroupMemberList(g); gm != NULL; gm=gm->nxt)
    {
      off = 0; tag = "";
      if ((i=gm->conTblId) >= 0 && (c=conTbl[i]) != NULL)
      {
        if (g->grpHeadLnkId == i) tag = " (*head*)";
        off = c->syncOffset;
      }
      else if ((i=gm->glbTblId) >= 0 && (glb=glbTbl[i]) != NULL)
      {
        if (g->grpHeadLnkId == i) tag = " (*head*)";
        off = glb->syncOffset;
      }
      ttyoutput("   %.192s + %d cnts %s\n",gm->key,off,tag);
    }
    ttyoutput(" number in group : %d\n",g->numInGrp);
    ttyoutput(" number pending : %d\n",g->numPending);
    ttyoutput(" current group cycle stamp : %d\n",g->grpSysStamp);
    ttyoutput(" last group cycle stamp : %d\n",g->grpLastStamp);
    ttyoutput(" current group cycle dispersion : %d counts\n",g->grpBndWdthC);
    ttyoutput(" current group time dispersion : %d msec\n",g->grpBndWdthT);
    ttyoutput(" current group synchronization : %s\n",grpSyncLevelStr[g->grpSyncLevel]);
    ttyoutput(" effective group update interval : %d msec\n",g->grpUpdateInterval);
    ttyoutput(" group updating monotonically : %s\n",g->grpIncMono ? "TRUE" : "FALSE");
    ts = PutDataTimeStampU(0,g->grpTimeUpdate.tv_sec,g->grpTimeUpdate.tv_usec);
    ttyoutput(" most recent update : %s\n",getDataTimeString(ts,0));
    ttyoutput(" current group status code : %d\n",g->grpCc);
  }
  return 0;
}
#endif /* !SMALL_TINE_LIB) */
int outputConnectionStats(int k)
{
  int baseMode;
  char linkstatus[5],flgStr[132];
  ConTblEntry *c;
  time_t ts;

  if (k < 0 || k >= nConnectionTableEntries)
  {
    ttyoutput("connection %d is not available",k);
    ttyoutput("available connections are 0 to %d",nConnectionTableEntries-1);
    return 0;
  }
  c = conTbl[k];
  baseMode = BASEMODE(c->mode);
  strcpy(linkstatus, baseMode < CM_SINGLE ? "-" : c->tocounter ? "DOWN" : "UP");
  ttyoutput(" Link Mode: %.128s",GetModeAsString(c->mode));
  ttyoutput(" Front End: %.16s",FecTbl[c->fecIdx].fecName);
  ttyoutput(" Front End Addr: %.40s (port %d)",FecTbl[c->fecIdx].strAdr,FecTbl[c->fecIdx].portOffset);
  if (c->mode & CM_NETWORK)
  {
    inet_ntop(gAfInet,&c->mcastGrp,flgStr,40);
    ttyoutput(" Multicast group: %.40s",flgStr);
  }
  ttyoutput(" Protocol: %s",getProtocolAsString(c->inetProtocol));
  ttyoutput(" Protocol Level: %d",c->tineProtocol);
  ttyoutput(" Contract: [%d] %.192s %d msec (%s) - %s",
    k,c->key,(int)c->pollingInterval,GetModeAsString(c->mode),linkstatus);
  ttyoutput(" Last Link Status: %d (%.32s)", c->linkStatus,cc2str(c->linkStatus));
# ifdef OS_16BIT
  ttyoutput(" Output: %ld %s element(s) (tag: %.16s)",c->sizeOut,itof(c->formatOut),c->strTagOut);
  ttyoutput(" Output: client data is %s",c->allocatedBytes ? "unbound" : "bound");
  if (c->allocatedBytes) ttyoutput(" Output: %ld bytes allocated for payload",c->allocatedBytes);
  ttyoutput(" Input : %ld %s element(s) (tag: %.16s)",c->sizeIn,itof(c->formatIn),c->strTagIn);
  ttyoutput(" Link Starttime: %.24s (%ld)",ctime((time_t *)&c->starttime),c->starttime);
  ttyoutput(" Last Arrived: %.24s (%ld)",ctime((time_t *)&c->timereq),c->timereq);
  ttyoutput(" Last Timestamp: %.24s (%ld)",ctime((time_t *)&c->dtimestamp),c->dtimestamp);
  ttyoutput(" Last System stamp: %ld",c->systemstamp);
  ttyoutput(" Last Datastamp: %ld",c->datastamp);
# else
  ttyoutput(" Output: %d %s element(s) (tag: %.16s)",c->sizeOut,itof(c->formatOut),c->strTagOut);
  ttyoutput(" Output: client data is %s",c->allocatedBytes ? "unbound" : "bound");
  if (c->allocatedBytes) ttyoutput(" Output: %d bytes allocated for payload",c->allocatedBytes);
  ttyoutput(" Input : %d %s element(s) (tag: %.16s)",c->sizeIn,itof(c->formatIn),c->strTagIn);
  ts = (time_t)c->starttime; ttyoutput(" Link Starttime: %.24s (%d)",ctime(&ts),c->starttime);
  ts = (time_t)c->timeLastSync; ttyoutput(" Last Arrived: %.24s (%d)",ctime(&ts),c->timeLastSync);
  ts = (time_t)c->dtimestamp; ttyoutput(" Last Timestamp: %.24s (%d)",ctime(&ts),c->dtimestamp);
  ttyoutput(" Last System stamp: %d",c->systemstamp);
  ttyoutput(" Last Datastamp: %d",c->datastamp);
# endif
  flgStr[0] = 0;
  if (c->rbModFlags & RB_USE_TIMESTAMP) strcpy(flgStr,"FIXTIMESTAMP");
  if (c->rbModFlags & RB_USE_ERRORVALUE) 
  {
    DTYPE d;
    char *p;
    UINT32 len = c->formatOut == CF_TEXT ? c->sizeOut : 1;
    DTYPEZERO(d);
    d.dFormat = c->formatOut;
    d.dArrayLength = len;
    d.data.bptr = c->errValue;
    if (strlen(flgStr) > 0) strcat(flgStr,", ");
    strcat(flgStr," ERRORVALUE: ");
    len = (UINT32)strlen(flgStr); p = &flgStr[len];
    len = 132 - len;
    _GetValuesArrayAsString(&d,p,&len,NULL,TRUE);
  }
  if (strlen(flgStr) == 0) strcpy(flgStr,"NONE");
  ttyoutput(" Flags: %s",flgStr);
  ttyoutput(" Is Pending: %s",c->notifyPending ? "YES" : "NO");
  ttyoutput(" Counter: %d",c->counter);
  ttyoutput(" Timeouts: %d",c->tocounter);

  ttyoutput(" Link Heartbeat: %d",(int)c->heartbeat);
  ttyoutput(" Num Blocks: %d",c->numblks);
  ttyoutput(" Is Grouped: %s",c->isGrouped ? "YES" : "NO");
  if (c->isGrouped)
  {
    ttyoutput(" Latency within Group: %d msec",c->grpLatency);
    ttyoutput(" Synchronization offset: %d",c->syncOffset);
  }
  ttyoutput(" Is Parent: %s",c->dpdLink != NULL ? "YES" : "NO");
  ttyoutput(" Is Dependent: %s (parent = connection %d)",(c->boundToId > 0 ? "YES" : "NO"),c->boundToId);
# ifndef SMALL_TINE_LIB
  ttyoutput(" Is MCA Link: %s",c->mcaLink != NULL ? "YES" : "NO");
  if (c->mcaLink != NULL) ttyoutput(" Is MCA Parent Format Elevated: %s",c->mcaLink->isElevatedFormat ? "YES" : "NO");
  ttyoutput(" Is MCA Parent: %s",IsMcaParentLink(c) ? "YES" : "NO");
# endif
  ttyoutput(" Is Listening link: %s",(c->cbfcn == _cbLnkLsn) ? "YES" : "NO");
  if (c->idleCount > 0) ttyoutput(" Listening idle count: %d",c->idleCount);
  ttyoutput(" Queue Length: %d",c->queueLen);
  ttyoutput(" Callback Id: %d",c->cbId);
  ttyoutput(" Needs Notification: %s",c->needsNotification ? "YES" : "NO");
  if (_gSuspendCallbacks) ttyoutput(" callbacks are currently suspended: YES");
  ttyoutput("\n");
  return 0;
}
int dumpConnectionList(void)
{
  return outputConnectionList();
}
int dumpConnectionStats(int i)
{
  return outputConnectionStats(i);
}
#ifndef SMALL_TINE_LIB
int dumpConnectionGroups(void)
{
  return outputConnectionGroups();
}
int dumpRedirectionList(void)
{
  return outputRedirectionList();
}
#endif
int gResetUseLoopback = FALSE;
int initClnWorkArea(UINT32 size)
{
  if ((clnWorkArea=(BYTE *)SystemCalloc(1,size)) == NULL) return out_of_client_memory;
# ifdef MULTITHREADED
  if ((mcstWorkArea=(BYTE *)SystemCalloc(1,size)) == NULL) return out_of_client_memory;
# endif
  clnWorkAreaSize = size;
  return 0;
}
int initClientTransport(int port)
{
  int cc = 0;
  char *ptr;
  if ((ptr=getenv("TINE_STANDALONE")) != NULL)
  {
    gSystemRunningStandAlone = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
  }
# if !defined(DOS)
  if ((cc=initControlStructs()) != 0) 
  { /* log the error by don't stop initializing */
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot initialize control structs : %.32s",cc2str(cc));
    cc = 0;
  }
  if ((cc=initClnWorkArea(clnWorkAreaSize)) != 0) return cc;
# endif
  if ((cc=initIPClient((UINT16)port)) != 0 && cc != tcp_not_supported) return cc;
# if defined(IPXBSD) || defined(IPXNW)
  if ((cc=initIPXClient(port)) != 0 && cc != ipx_not_supported) return cc;
# endif
# if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
  if (!gIsRunningAsServer && shmStart()) 
  {
    if (!gUseLoopback) gResetUseLoopback = TRUE;
    gUseLoopback = TRUE;
  }
# endif
  return cc;
}
int gInitClientDone = FALSE;
int gClientIsClosing = FALSE;
/**
 * \internal
 * \brief Initializes the TINE client.
 *
 * initClient has an internal flag which ensures that it is
 * called only once.
 *
 * This function
 * <ul>
 * <li>Finds the user name and stores it in gUserName</li>
 * <li>Calls initFecTable()</li>
 * <li> calls initTransport() </li>
 * </ul>
 * \see gUserName
 * 
 */
int initClient(void)
{
  int cc = 0;
# ifndef SMALL_TINE_LIB
  int val;
  char *ptr = NULL;
# endif
  if (gInitClientDone) return 0;
  InitSystemMutexSet();
  if (WaitForSystemMutex(hSystemClientMutex,-1) != 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot initialize client resources : wait for client mutex failed!");
    return semaphore_error;
  }
  if (gInitClientDone) ccerr(0);
  gClientIsClosing = FALSE;
  feclog("initializing client resources");
# ifndef SMALL_TINE_LIB
  if ((ptr=getenv("TINE_CONSOLE")) != NULL)
  {
    foregroundTTY = strtobool(ptr);
  }
  if ((ptr=getenv("TINE_TRANSPORT")) != NULL)
  {
    if (!strnicmp(ptr,"TCP",3)) useConnectedSockets = TRUE;
  }
  if ((ptr=getenv("TINE_DEBUG")) != NULL)
  {
    int dbg = atoi(ptr);
    tineDebug = dbg > 0 ? dbg : 0;
    if (tineDebug) foregroundTTY = TRUE;
  }
# endif
  getSrvIPAddr();
  if (gSystemStartupTime == 0) gSystemStartupTime = time(NULL);
  whoami(gUserName);
  /* alloc room for the principal client connection table */
  if (conTbl == NULL && (conTbl=(ConTblEntry **)SystemCalloc(gConTblCapacity,sizeof(ConTblEntry *))) == NULL)
  {
    ccerr(out_of_client_memory);
  }
  else
  {
    feclog("Client Connection Table has allocated space for %d entries",gConTblCapacity);
  }
# ifndef SMALL_TINE_LIB
  if ((ptr=getenv("TINE_QUEUEDEPTH")) != NULL)
  {
    if ((val=atoi(ptr)) < 0) val = 0;
    SetClnRecvQueueDepth(val);
  } 
# endif
  if ((cc=initFecTable()) != 0) ccerr(cc); /* first call initializes the Fec Table */
  if ((cc=initClientTransport(0)) != 0) ccerr(cc);
  if (clnWorkArea == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "Client Work Space not allocated !!");
    ccerr(un_allocated);
  }
  if (gCatchConsoleBreak)
  {
    feclog("catch interrupt and terminate signals");
#   if defined(WIN32)
    if (SetConsoleCtrlHandler((PHANDLER_ROUTINE)winBreakHndlr,TRUE ) == 0) feclogEx(TINE_LOGLEVEL_WARN, "cannot catch console signals");
#   elif defined(UNIX)
    if (bsdsignal(SIGINT, sig_break) == SIG_ERR) feclog("cannot catch SIGINT");
    if (bsdsignal(SIGTERM, sig_break) == SIG_ERR) feclog("cannot catch SIGTERM");
#   endif
  }
  gInitClientDone = TRUE;
err:
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR, "initializing client resources failed : %.32s",cc2str(cc));
  ReleaseSystemMutex(hSystemClientMutex);
  return cc;
}
TINE_EXPORT void CloseClientLinks(void)
{ /* just close all of the links (don't touch any sockets) */
  int n;
  for (n=0; n<nConnectionTableEntries; n++) 
  {
    if (conTbl == NULL) break; /* how ? */
    if (conTbl[n] == NULL) continue; /* how ? */
    if (conTbl[n]->glbTblId > 0) continue; /* don't worry about these */
    if (conTbl[n]->tineProtocol > 4) _closeLink(n);
  }
  for (n=0; n<nConnectionTableEntries; n++)
  {
    if (conTbl == NULL) break; /* how ? */
    checkSubscriptions(FALSE);
    if (ReturnToClientCycle)
    { /* back up */
      if (ReturnToClientCycle%2) hardDelay(20);
      ReturnToClientCycle = 0;
      if (--n < 0) break;
    }
  }
}
TINE_EXPORT void CloseClientsEx(int closesockets,int quickExit)
{ /* For packing efficiency do this in two steps: */
  int n;

  gClientIsClosing = TRUE;
# ifdef MULTITHREADED
  /* help the termination of globals threads along ... */
  stopGlobalsThread(glbClnSck);
  stopGlobalsThread(mcastClnSck);
# endif
  if (nConnectionTableEntries > 0) dumpMsgTableToFile();
  WaitForSystemMutex(hSystemGlobalsUpdateMutex,-1);
  if (quickExit) goto check_done;
  if (conTbl == NULL)
  { /* n.b. : a SystemReset() will reset nConnectionTableEntries = 1 if a NameServer was loaded ... */
    nConnectionTableEntries = 0;
  }
  for (n=0; n<nConnectionTableEntries; n++) 
  {
    if (conTbl == NULL) break; /* how ? */
    if (conTbl[n] == NULL) continue; /* how ? */
    if (conTbl[n]->tineProtocol > 4) _closeLink(n);
  }
  for (n=0; n<nConnectionTableEntries; n++)
  {
    if (conTbl == NULL) break; /* how ? */
    checkSubscriptions(FALSE);
    if (ReturnToClientCycle)
    { /* back up */
      if (ReturnToClientCycle%2) hardDelay(20);
      ReturnToClientCycle = 0;
      if (--n < 0) break;
    }
  }
  flushGlobalsTable();
  rmvMcaLinkList();
check_done:
  if (closesockets) 
  {
    ShutdownClients(TRUE);
    if (!quickExit && conTbl != NULL)
    { /* safe to remove the connection table */
      for (n=1; n<gConTblCapacity; n++)
      {
        freeConTblEntryMemory(n);
        SystemFree(conTbl[n]);
      }
      SystemFree(conTbl);
    }
    rmvIdleConnections();
  }
  nConnectionTableEntries = 0;
# if defined(UNIX) || defined(WIN32) || defined(MACOS)
  CloseIpcPipe(NULL);
# endif
  gInitClientDone = FALSE;
# ifdef MULTITHREADED
  if (hNotificationSemaphore != (SEMHANDLE)0) ReleaseSystemSemaphore(hNotificationSemaphore);
# endif
  ReleaseSystemMutex(hSystemGlobalsUpdateMutex);
}
TINE_EXPORT void CloseClients(int closesockets)
{
  CloseClientsEx(closesockets,FALSE);
}
TINE_EXPORT GrpTblEntry *GetCallbackGroup(size_t id)
{
  GrpTblEntry *gt;

  if (id == 0) return NULL;
  for (gt=grpTbl; gt && gt->grpId != id; gt=gt->next);
  return gt;
}
TINE_EXPORT GrpMember *GetGroupMemberList(GrpTblEntry *grp)
{
  int i,cc=0;
  ConTblEntry *c;
  GlobalListStruct *glb;
  GrpMember *grpMmbr=NULL;
  if (grp == NULL) return NULL;
  if (grp->members == NULL)
  { /* not yet allocated */
    for (i=0; i<nConnectionTableEntries; i++)
    {
      if ((c=conTbl[i]) == NULL) continue;
      if (c->grp != grp) continue;
      if ((grpMmbr=(GrpMember *)calloc(1,sizeof(GrpMember))) == NULL) ccerr(out_of_local_memory);
      grpMmbr->conTblId = i; grpMmbr->glbTblId = -1;
      sprintf(grpMmbr->key,"%.192s",c->key);
      grpMmbr->dout.dFormat = c->formatOut;
      grpMmbr->dout.dArrayLength = c->sizeOut;
      strncpy(grpMmbr->dout.dTag,c->strTagOut,TAG_NAME_SIZE);
      grpMmbr->dout.dTimeStamp = PutDataTimeStampU(0,c->dtimestamp,c->dtimestampUSEC);
      grpMmbr->dout.sysStamp = c->systemstamp;
      grpMmbr->dout.dStamp = c->datastamp;
      grpMmbr->dout.xferReason = (short)c->xferReason;
      grpMmbr->dout.data.cptr = c->dataPtr;
      grpMmbr->nxt = grp->members;
      grp->members = grpMmbr;
    }
    for (i=0; i<nglobals; i++)
    {
      if ((glb=glbTbl[i]) == NULL) continue;
      if (glb->grp != grp) continue;
      if ((grpMmbr=(GrpMember *)calloc(1,sizeof(GrpMember))) == NULL) ccerr(out_of_local_memory);
      grpMmbr->glbTblId = i; grpMmbr->conTblId = -1;
      sprintf(grpMmbr->key,"/%.32s/%.32s/keyword[%.64s]",glb->context,glb->server,glb->keyword);
      grpMmbr->dout.dFormat = glb->format;
      grpMmbr->dout.dArrayLength = glb->size;
      grpMmbr->dout.dTimeStamp = PutDataTimeStampU(0,glb->time,glb->timeUSEC);
      grpMmbr->dout.sysStamp = glb->sysStamp;
      grpMmbr->dout.dStamp = glb->dataStamp;
      grpMmbr->dout.xferReason = (short)CX_EVENT;
      grpMmbr->dout.data.cptr = (char *)glb->data;
      grpMmbr->nxt = grp->members;
      grp->members = grpMmbr;
    }
  }
  else
  { /* list allocated: update the relevant fields */
    for (grpMmbr=grp->members; grpMmbr != NULL; grpMmbr=grpMmbr->nxt)
    { 
      if (grpMmbr->conTblId >= 0)
      { /* is a normal connection link */
        if ((c=conTbl[grpMmbr->conTblId]) != NULL)
        {
          grpMmbr->dout.dTimeStamp = PutDataTimeStampU(0,c->dtimestamp,c->dtimestampUSEC);
          grpMmbr->dout.sysStamp = c->systemstamp;
          grpMmbr->dout.dStamp = c->datastamp;
          grpMmbr->dout.xferReason = (short)c->xferReason;
        }
      }
      else if (grpMmbr->glbTblId >= 0)
      { /* is a globals link */
        if ((glb=glbTbl[grpMmbr->glbTblId]) != NULL)
        {
          grpMmbr->dout.dTimeStamp = PutDataTimeStampU(0,glb->time,glb->timeUSEC);
          grpMmbr->dout.sysStamp = glb->sysStamp;
          grpMmbr->dout.dStamp = glb->dataStamp;
        }
      }
    }
  }
err:
  if (cc != 0)
  {
    feclog("GetGroupMemberList: %.32s",cc2str(cc));
    return NULL;
  }
  return grp->members;
}
GrpTblEntry *getGrpTblEntry(size_t id)
{ /* id is always a callback address */
  GrpTblEntry *gt;

  if (id == 0) return NULL;
  for (gt=grpTbl; gt && gt->grpId != id; gt=gt->next);

  if (!gt)
  {
    if ((gt=(GrpTblEntry *)SystemCalloc(1,sizeof(GrpTblEntry))) == NULL) return NULL;
    gt->grpId = id;
    gt->next = grpTbl;
    grpTbl = gt;
  }
  return gt;
}
int IsMetaToHistory(char *srcprp,char *tgtsrv)
{
  if (strstr(srcprp,".HIST") == NULL) return FALSE;
  if (stricmp(tgtsrv,"HISTORY") != 0) return FALSE;
  return TRUE;
}
int isNameServerRequest = FALSE;
int DataTagging = FALSE;
RdrTblEntry *rdrLnkTbl[REDIR_HASH_SIZE] = { NULL };
RdrTblEntry *appendRedirectedLinkList(char *con,char *srv,char *prp,char *dev,char *dstctx,char *dstsrv,char *dstdev,char *dstprp,int isGrpMem)
{
  RdrTblEntry *r;
  char *bpstr, *mpstr;
  char key[FULL_NAME_MAXSIZE+8];
  int idx;

  if ((r=(RdrTblEntry *)SystemCalloc(1,sizeof(RdrTblEntry))) == NULL) return NULL;
  strncpy(r->srcContext,con,CONTEXT_NAME_SIZE);
  strncpy(r->srcServerName,srv,EXPORT_NAME_SIZE);
  strncpy(r->srcProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(r->srcDeviceName,dev,DEVICE_NAME_SIZE);
  strncpy(r->dstContext,dstctx,CONTEXT_NAME_SIZE);
  strncpy(r->dstServerName,dstsrv,EXPORT_NAME_SIZE);
  strncpy(r->dstDeviceName,dstdev,DEVICE_NAME_SIZE);
  if (strlen(dstprp) > 0) 
  {
    strncpy(r->dstProperty,dstprp,PROPERTY_NAME_SIZE);
    if (strlen(dstprp) <= PROPERTY_NAME_SHORTSIZE && 
        !IsMetaProperty(dstprp,NULL,NULL,TRUE) && 
        IsMetaProperty(prp,&bpstr,&mpstr,TRUE) &&
        !IsMetaToHistory(prp,dstsrv))
    {
      strcat(r->dstProperty,".");
      strncat(r->dstProperty,mpstr,15);
    }
  }
  else 
  {
    strncpy(r->dstProperty,prp,PROPERTY_NAME_SIZE);
  }
  r->isGroupMember = isGrpMem;
  makeLinkKey(key,con,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % REDIR_HASH_SIZE);
  r->nxt = rdrLnkTbl[idx];
  if (rdrLnkTbl[idx] != NULL) rdrLnkTbl[idx]->prv = r;
  rdrLnkTbl[idx] = r;

  msglog(0,"append /%.32s/%.32s/%.64s[%.64s] -> /%.32s/%.32s/%.64s[%.64s] to redirection table",
    con,srv,dev,prp,dstctx,dstsrv,dstdev,r->dstProperty);
  return r;
}
TINE_EXPORT void flushRedirectedLinkList(void)
{
  RdrTblEntry *r, *rnxt=NULL;
  int i;
  for (i=0; i<REDIR_HASH_SIZE; i++)
  {
    for (r=rdrLnkTbl[i]; r != NULL; r=rnxt)
    {
      if (gIsRunningAsServer && ServerExitCondition)
      { /* just do it */
        rnxt = r->nxt;
        free(r);
      }
      else
      {
        rmvRdrInfo(r,0);
      }
    }
    rdrLnkTbl[i] = NULL;
  }
}

/**
 * \internal
 * \brief Removes a redirected link from a doubly linked list
 * \param r List Element
 * \param cc unused
 * \todo Remove unused parameter
 */
int removeRedirectedLink(RdrTblEntry *r,int cc)
{
  char key[FULL_NAME_MAXSIZE+8];
  int idx;
  UNUSED(cc);
  if (r == NULL) return argument_list_error;
  msglog(0,"removing /%.32s/%.32s/%.64s/%.64s from redirection table",
    r->dstContext,r->dstServerName,r->dstProperty,r->dstDeviceName);
  if (r->nxt != NULL) r->nxt->prv = r->prv;
  if (r->prv != NULL) r->prv->nxt = r->nxt;  
  makeLinkKey(key,r->srcContext,r->srcServerName,r->srcDeviceName,r->srcProperty);
  idx = (int)(ElfHash((unsigned char *)key) % REDIR_HASH_SIZE);
  if (r == rdrLnkTbl[idx]) rdrLnkTbl[idx] = r->nxt;
  SystemFree(r);
  return 0;
}

/**
 * \internal
 * \brief Get data from redirected linked list
 *
 * \param con EqmContext
 * \param srv Server Name 
 * \param prp Property
 * \param dev Device name
 *
 * This function looks for an entry in the global Redirection table
 * whose corresponding fields match the four parameters.
 * If an entry is found, the parameters are overwritten with the entry's values.
 * 
 * \return TRUE if an entry was found. FALSE if no entry was found.
 * 
 * \see rdrLnkTbl
 * 
 */
int redirectLink(char *con,char *srv,char *prp,char *dev)
{
  RdrTblEntry *r=NULL;
  char key[FULL_NAME_MAXSIZE+8];
  int idx;
  /* routine is not exported and none of the parameters is NULL */
  makeLinkKey(key,con,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % REDIR_HASH_SIZE);

  for (r=rdrLnkTbl[idx]; r != NULL; r=r->nxt)
  {
    if (strnicmp(r->srcContext,con,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(r->srcServerName,srv,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(r->srcProperty,prp,PROPERTY_NAME_SIZE) && strcmp(r->srcProperty,"*")) continue;
    if (strnicmp(r->srcDeviceName,dev,DEVICE_NAME_SIZE) && strcmp(r->srcDeviceName,"*")) continue;
    strncpy(con,r->dstContext,CONTEXT_NAME_SIZE);
    strncpy(srv,r->dstServerName,EXPORT_NAME_SIZE);
    if (strcmp(r->dstDeviceName,"*")) strncpy(dev,r->dstDeviceName,DEVICE_NAME_SIZE);
    if (strcmp(r->dstProperty,"*")) strncpy(prp,r->dstProperty,PROPERTY_NAME_SIZE);
    msglog(0,"redirecting /%.32s/%.32s/%.64s[%.64s] -> /%.32s/%.32s/%.64s[%.64s] from redirection table",
      r->srcContext,r->srcServerName,r->srcDeviceName,r->srcProperty,
      r->dstContext,r->dstServerName,r->dstDeviceName,r->dstProperty);
    return TRUE;
  }
  return FALSE;
}

McaTblEntry *mcaLnkLst = NULL;
McaTblEntry *getMcaLinkEntry(char *con,char *srv,char *prp,char *dev)
{ /* find a match */
  McaTblEntry *m=NULL;
  /* routine is not exported and none of the parameters is NULL */
  for (m=mcaLnkLst; m != NULL; m=m->nxt)
  {
    if (strnicmp(m->mcaContext,con,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(m->mcaServerName,srv,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(m->mcaProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(m->mcaDevice,dev,DEVICE_NAME_SIZE)) continue;
    if (m->linkId < 0) continue;
    return m;
  }
  return NULL;
}
/* TODO:  add the index; worry about incrementing nlinks, finding ndevices, etc. */
McaTblEntry *addMcaLinkEntry(char *con,char *srv,char *prp,char *dev,int size,int isElvFmt)
{
  McaTblEntry *m = getMcaLinkEntry(con,srv,prp,dev);

  if (m == NULL)
  { /* must be new ! */
    if ((m=(McaTblEntry *)SystemCalloc(1,sizeof(McaTblEntry))) == NULL) return NULL;
    strncpy(m->mcaContext,con,CONTEXT_NAME_SIZE);
    strncpy(m->mcaServerName,srv,EXPORT_NAME_SIZE);
    strncpy(m->mcaProperty,prp,PROPERTY_NAME_SIZE);
    strncpy(m->mcaDevice,dev,DEVICE_NAME_SIZE);
    m->nxt = mcaLnkLst;
    if (mcaLnkLst != NULL) mcaLnkLst->prv = m;
    mcaLnkLst = m;
    m->ndevices = size;
    m->isElevatedFormat = isElvFmt;
    msglog(0,"add /%.32s/%.32s/%.64s[%.64s] to the mca link table",
      m->mcaContext,m->mcaServerName,m->mcaDevice,m->mcaProperty);
  }

  return m;
}
void addDependentLinkToMca(McaTblEntry *mca,int lnkId)
{
  McaLnkItem *item, *it;
  int bid = -1;
  if (mca == NULL) return;
  if ((item=(McaLnkItem *)SystemCalloc(1,sizeof(McaLnkItem))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "no memory to allocate next mca link item !");
    return;
  }
  for (it=mca->links; it != NULL; it=it->nxt) 
  {
    if (it->linkId == lnkId) 
    {
      it->isCanceled = FALSE;
      return;
    }
  }
  item->linkId = lnkId;
  bid = conTbl[lnkId]->boundToId; /* make a note in case of 'back out' */
  if (bid > 0 && bid != mca->linkId) item->boundToId = bid;
  item->nxt = mca->links;
  mca->links = item;
  mca->nlinks++;
  msglog(0, "add %s dependent %s [%d] to MCA %s [%d]", item->boundToId > 0 ? "bound" : "unbound",
    conTbl[item->linkId]->key, item->linkId, conTbl[mca->linkId]->key, mca->linkId);
}
void cancelDependentLinkInMca(McaTblEntry *mca,int lnkId)
{
  McaLnkItem *item;
  for (item=mca->links; item != NULL; item=item->nxt)
  { /* move thru the list */
    if (item->linkId == lnkId)
    { /* found it */
      item->isCanceled = TRUE;
      return;
    }
  }
}
void rmvDependentLinkFromMca(McaTblEntry *mca,int lnkId)
{
  McaLnkItem *item, *lastitem=mca->links;
  for (item=mca->links; item != NULL; item=item->nxt)
  { /* move thru the list */
    if (item->linkId == lnkId)
    { /* found it */
      if (item == mca->links)
      { /* must be at the beginning */
        mca->links = item->nxt;
      }
      else
      { /* splice the ends */
        lastitem->nxt = item->nxt;
      }
      msglog(0, "remove dependent %s [%d] from MCA %s [%d]", 
        conTbl[item->linkId]->key, item->linkId, conTbl[mca->linkId]->key, mca->linkId);
      SystemFree(item);
      mca->nlinks--;
      return;
    }
    lastitem = item;
  }
}
int removeMcaLinkEntry(McaTblEntry *m)
{
  int i;
  msglog(0,"removing /%.32s/%.32s/%.64s[%.64s] from the mca link table",
    m->mcaContext,m->mcaServerName,m->mcaDevice,m->mcaProperty);
  if (m->nxt != NULL) m->nxt->prv = m->prv;
  if (m->prv != NULL) m->prv->nxt = m->nxt;
  if (m == mcaLnkLst) mcaLnkLst = m->nxt;
  for (i=0; i<nConnectionTableEntries; i++)
  { /* run thru the list and cross-check */
    if (conTbl[i] == NULL) continue; /* should not be possible */
    if (conTbl[i]->mcaLink == m)
    { /* this shouldn't happen, but just in case ... */
      msglog(code_failure,"late removal of mca link %d (element %d)",i,conTbl[i]->mcaIndex-1);
      conTbl[i]->mcaLink = NULL;
      conTbl[i]->mcaIndex = 0;
    }
  }
  SystemFree(m);
  return 0;
}
void rmvMcaLinkList(void)
{
  McaTblEntry *m=mcaLnkLst, *mn=NULL;
  while (m != NULL)
  {
    mn = m->nxt;
    free(m);
    m = mn;
  }
  mcaLnkLst = NULL;
}
void movDpdToMca(ConTblEntry *c,McaTblEntry *mca)
{ 
  if (WaitForSystemMutex(hDpdLinkMutex,-1) == 0)
  {
    if (c->dpdLink != NULL)
    {
      DpdTblEntry *dte, *ndte=NULL;
      int bid;
      for (dte=c->dpdLink; dte != NULL; dte=ndte)
      {
        bid = dte->linkId;
        ndte = dte->nxt;
        addDependentLinkToMca(mca,bid);
        conTbl[bid]->boundToId = mca->linkId;
        conTbl[bid]->mcaIndex = c->mcaIndex;
        conTbl[bid]->mcaLink = mca;
        free(dte);
      }
      c->dpdLink = NULL;
    }
    ReleaseSystemMutex(hDpdLinkMutex);
  }
}
void checkTcpIdleConnections(void)
{
  TCPSCKT *sckt = NULL;
  int i;
  for (sckt=tcpClnSckTbl; sckt != NULL; sckt = sckt->nxt)
  {
    if (sckt->sck == (SOCKET)0) continue;
    for (i=0; i < nConnectionTableEntries; i++)
    { /* anybody using this socket ? */
      if (conTbl[i] == NULL) continue; /* should not be possible */
      if (conTbl[i]->tcpSck == sckt->sck) break; /* yes => it's not idle */
    }
    if (i == nConnectionTableEntries)
    { /* no one's using it: mark it as idle */
      AddIdleConnection(sckt->sck, sckt->idx, sckt->type);
    }
  }
  return;
}

int ConTblRdrCmp(ConTblEntry *cte,char *cxt,char *grp,char *dev,char *prp)
{
  RdrTblEntry *r = cte->rdrLink;
  if (r == NULL) return -1;
  if (cte->mode == CM_CANCEL) return -1;
  if (cte->rdrId <= 0) return -1;
  if (strnicmp(r->srcContext,cxt,CONTEXT_NAME_SIZE)) return -1;
  if (strnicmp(r->srcServerName,grp,EXPORT_NAME_SIZE)) return -1;
  if (strnicmp(r->srcProperty,prp,PROPERTY_NAME_SIZE)) return -1;
  if (strnicmp(r->srcDeviceName,dev,DEVICE_NAME_SIZE)) return -1;
  return 0;
}
int assignConTblMemory(int i)
{
  if (i < 0) return invalid_index;
  if (i >= gConTblCapacity)
  {
#   ifndef SMALL_TINE_LIB
    if (gAllowAutoGrowConnectionTable)
    {
#     define conTblCapacityBuffer 500
      int cc = SetConnectionTableCapacity(gConTblCapacity + conTblCapacityBuffer);
      if (cc != 0) return cc;
    }
    else
#   endif
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "connection table capacity has been reached (%d)", gConTblCapacity);
      msglog(resources_exhausted, "connection table capacity has been reached (%d)", gConTblCapacity);
      return resources_exhausted;
    }
  }
  if ((conTbl[i]=(ConTblEntry *)SystemCalloc(1,sizeof(ConTblEntry))) == NULL) return out_of_client_memory;
  return 0;
}

void prepNameServerEntry(DTYPE *dout,DTYPE *din,char *devctx,char *devgrp,char *devnam,char *devprp) 
{
  UNUSED(devctx);
  UNUSED(devgrp);
  conTbl[0] = &connNameServer;
  conTbl[0]->tineProtocol = gTineProtocol;
  conTbl[0]->inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP;
  strcpy(conTbl[0]->devName,devnam);
  strcpy(conTbl[0]->prpName,devprp);
  strcpy(conTbl[0]->key, "ENS");
  if (dout != NULL) 
  {
   conTbl[0]->sizeOut = dout->dArrayLength;
   conTbl[0]->formatOut = dout->dFormat;
   conTbl[0]->dataPtr = (char *)dout->data.vptr;
  } 
  else 
  {
   conTbl[0]->sizeOut = 0;
   conTbl[0]->formatOut = CF_NULL;
   conTbl[0]->dataPtr = NULL;
  }
  if (din != NULL) 
  {
   conTbl[0]->sizeIn = din->dArrayLength;
   conTbl[0]->formatIn = din->dFormat;
   conTbl[0]->dataIn = (char *)din->data.vptr;
   conTbl[0]->bytesTotal = conTbl[0]->sizeIn*fmtsizeof(conTbl[0]->formatIn);
   conTbl[0]->datamtu = UDP_BUFFER_SIZE-SUBREQHDR_SIZE-REQHDR_SIZE;
   if (conTbl[0]->bytesTotal < conTbl[0]->datamtu) 
   {
     conTbl[0]->numblksIn = 1;
     conTbl[0]->bytesRemaining = conTbl[0]->bytesTotal;
     conTbl[0]->nextDataSize = conTbl[0]->bytesTotal;
   } 
   else 
   { 
     conTbl[0]->bytesTotal = 0; 
   }
  } 
  else 
  {
   conTbl[0]->sizeIn = 0;
   conTbl[0]->formatIn = CF_NULL;
   conTbl[0]->dataIn = NULL;
  }
  conTbl[0]->notifyPending = 1;
  conTbl[0]->hasNotifiedOnce = FALSE;
  conTbl[0]->cancelPending = FALSE;
  conTbl[0]->needsNotification = FALSE;
  conTbl[0]->isWaitingSyncNotification = FALSE;
  conTbl[0]->isWildcardLink = FALSE;
  SystemFree(conTbl[0]->bufOut);
  SystemFree(conTbl[0]->bufOutBlks);
  SystemFree(conTbl[0]->bufCmp);
  SystemFree(conTbl[0]->bufVtl);
  SystemFree(conTbl[0]->dtf);
}
BYTE *getStructPtr(struct structStruct *s,char *key,DTYPE *d,BYTE *ref)
{ /* used internally -> arguments never NULL */
# ifdef INTEL
  UNUSED(s); 
  UNUSED(d);
  UNUSED(key);
  return ref;
# else
  _taggedStructSwap(d->dTag,key,ref,s->sdata,LTOL,d->dArrayLength,s->size);
  return s->sdata;
# endif
}
/**
 * \internal
 * \brief find the total number of input bytes for a struct with extended space
 */
int fixStructExtendedSpace(struct structStruct *ss,DTYPE *d,char *ctx,char *srv)
{
  int i,k,nbytes=0,dsiz=0,hsiz=0,extsiz=0,cc=0;
  struct structFormat *s=ss->s;
  struct structStruct *sse;
  BYTE *src = d->data.bptr, *hptr;
  DTYPE ssd;
  char *strptr;
  for (i=0; i<(int)d->dArrayLength; i++)
  { /* have to collect the string space for all elements of a structure array */
    for (s=ss->s; s != NULL && s->siz; s = s->nxt)
    {
      switch (s->fmt)
      {
        case CF_STRUCT:
          if ((sse=findstructEx(s->stag,ctx,srv)) == NULL) ccerr(invalid_structure_tag);
          if (sse->hasExtSpace)
          { /* prepare for re-entry ... */
            ssd.dArrayLength = s->siz;
            ssd.data.bptr = (BYTE *)&src[s->addr];
            dsiz = fixStructExtendedSpace(sse,&ssd,ctx,srv);
          }
          else
          {
            dsiz = s->siz * sse->rawlen;
          }
          break;
        case CF_KEYVALUE:
        case CF_STRING:
          for (k=0; k<s->siz; k++)
          {
            if ((strptr=(char *)((size_t *)(src+s->addr))[k]) == NULL) ccerr(invalid_data);
            extsiz += (int)strlen(strptr) + 1; /* add room for terminating 0 */
          }
          dsiz = s->siz * (2*sizeof(SINT32));
          break;
        case CF_AIMAGE:
        case CF_IMAGE:
        case CF_ASPECTRUM:
          hsiz = getFormatHeaderSize(s->fmt);
          for (k=0; k<s->siz; k++)
          {
            hptr = (BYTE *)&src[s->addr+k*getFormatTypeSize(s->fmt)];
            extsiz += getFormatBodyBufferSize(hptr,s->fmt)*fmtsizeof(s->fmt);
          }
          dsiz = s->siz * (hsiz+2*sizeof(SINT32));
          break;
        default:
          dsiz=s->siz*getFormatSize(s->fmt);
          break;
      }
      nbytes += dsiz;
    }
  }
err:
  if (cc) feclog("fixStructExtendedSpace: structure %.16s -> %.32s",s->stag,cc2str(cc));
  return cc ? 0 : nbytes + extsiz;
}
/**
 * \internal
 * \brief copy/prepare outbound (input) data from a client call for delivery
 */
void cpyInputDataBytes(ConTblEntry *cte,DTYPE *d,size_t siz) 
{
  BYTE *dst=(BYTE *)cte->dataIn,*src=d->data.bptr;
  BYTE *bptr = GetFormatBodyPtr(d->data.vptr, d->dFormat);
  struct structStruct *s;
  char key[128];
  if (d->dFormat == CF_STRUCT && (s=findstructEx(d->dTag,cte->ctxName,cte->expName)) != NULL)
  { /* is a registered tagged structure */
    sprintf(key,"/%.32s/%.32s",cte->ctxName,cte->expName);
    _taggedStructSwap(d->dTag,key,dst,src,LTOR,d->dArrayLength,s->size);
  }
  else if (d->dFormat == CF_MDA)
  {
    DMDA *mda = (DMDA *)src;
    memcpy(dst,mda,MDA_HDR_SIZE);
    memcpy(&dst[MDA_HDR_SIZE],mda->data.vptr,(size_t)d->dArrayLength*fmtsizeof(mda->fmt));
  }
  else if (bptr != NULL)
  { /* e.g. DIMAGE or DGSPECTRUM */
    int hsiz = GetFormatHeaderSize(d->dFormat);
    memcpy(dst, src, hsiz);
    memcpy(&dst[hsiz], bptr, siz-hsiz);
  }
  else
  { /* most cases: just straight-up copy */
    memcpy(dst,src,siz);
  }
}
/**
 * \internal
 * \brief compare the contents of struct with given data 
 */
int structcmp(struct structStruct *s,char *key,DTYPE *d,BYTE *ref,int nbytes)
{
  struct structFormat *sf=s->s;
  struct structStruct *sse;
  char *strptr;
  BYTE *src = d->data.bptr, *dataPtr=NULL, *hptr, *extptr;
  int i,k,nb,dsiz,extsiz=0,hsiz;
  if (s->hasExtSpace)
  { /* tagged struct with extended space */
    nb = 0;
    dataPtr = getStructPtr(s,key,d,ref);
    for (i=0; i<(int)d->dArrayLength; i++)
    { /* scan through the string space for all elements of a structure array */
      nb += s->rawlen;
      for (sf=s->s; sf != NULL && sf->siz; sf = sf->nxt)
      {
        switch (sf->fmt)
        {
          case CF_STRUCT:
            if ((sse=findstructEx(sf->stag,key)) == NULL) return TRUE;
            /* TODO: this also needs to re-enter and check for string space */
            break;
          case CF_KEYVALUE:
          case CF_STRING: /* compare the strings */
            for (k=0; k<sf->siz; k++)
            {
              if ((strptr=(char *)((size_t *)(src+sf->addr))[k]) == NULL) return TRUE;
              extsiz = (int)strlen(strptr) + 1; /* add room for terminating 0 */
              if (memcmp(&ref[nb],strptr,extsiz) != 0) return TRUE;
              nb += extsiz;
            }
            break;
          case CF_AIMAGE:
          case CF_IMAGE:
          case CF_ASPECTRUM:
            hsiz = getFormatHeaderSize(sf->fmt);
            for (k=0; k<sf->siz; k++)
            {
              hptr = (BYTE *)&src[sf->addr+k*getFormatTypeSize(sf->fmt)];
              if (memcmp(&ref[nb],hptr,hsiz) != 0) return TRUE;
              extsiz += GetFormatBodyBufferSize(hptr,sf->fmt)*fmtsizeof(sf->fmt);
              extptr = GetFormatBodyPtr(hptr,sf->fmt);
              if (memcmp(&ref[nb],extptr,extsiz) != 0) return TRUE;
              nb += extsiz;
            }
            break;
          default: /* compare the data contents */
            dsiz = fmtsizeof(sf->fmt) * sf->siz;
            if (memcmp(dataPtr,src+sf->addr,dsiz) != 0) return TRUE;
            dataPtr += dsiz;
            break;
        }
      }
      src += s->size;
    }
  }
  else
  { /* normal case : just compare the bytes straight-up */
    if (nbytes != (int)d->dArrayLength*s->rawlen) return TRUE;
    dataPtr = getStructPtr(s,key,d,ref);
    if (memcmp(d->data.vptr,dataPtr,(size_t)d->dArrayLength*s->size)) return TRUE;
  }
  return FALSE; /* structs do not differ */
}
#define TO_THRESHOLD 10
#define TO_RETRY_THRESHOLD 2
#define TO_TABLE_SIZE 10
INTINT FecToTbl[TO_TABLE_SIZE];

void setConTblSize(UINT32 *siz,DTYPE *d,char *ctx,char *srv)
{
  int sz = 0;
  struct structStruct *s;
  if (d == NULL) goto out;
  sz = d->dArrayLength;
  if (d->dFormat == CF_STRUCT && (s=findstructEx(d->dTag,ctx,srv)) != NULL)
  { /* registered tagged structure (reference to link calling parameters!) */
    sz = d->dArrayLength*s->rawlen;
  }
out:
  *siz = sz;
}
UINT32 getAssignedDataLength(ConTblEntry *c,int isInpt)
{
  UINT32 dsiz = isInpt ? c->sizeIn : c->sizeOut;
  char *dtag = isInpt ? c->strTagIn : c->strTagOut;
  int dfmt = isInpt ? c->formatIn : c->formatOut;
  struct structStruct *s;
  switch (dfmt)
  {
    case CF_STRUCT:
      if ((s=findstructEx(dtag,c->ctxName,c->expName)) != NULL && s->rawlen > 0)
      {
        dsiz /= s->rawlen;
      }
      break;
    default:
      break;
  }
  return dsiz;
}
void freeConTblEntryMemory(int idx)
{
  if (conTbl == NULL) return;
  if (idx < 1 || idx >= gConTblCapacity) return;
  if (conTbl[idx] == NULL) return;
  if (conTbl[idx]->allocatedBytes) SystemFree(conTbl[idx]->dataPtr);
  SystemFree(conTbl[idx]->dataIn);
  SystemFree(conTbl[idx]->bufOut);
  SystemFree(conTbl[idx]->bufOutBlks);
  SystemFree(conTbl[idx]->bufCmp);
  SystemFree(conTbl[idx]->bufVtl);
  SystemFree(conTbl[idx]->dtf);
  SystemFree(conTbl[idx]->extStringspace);
  return;
}
int feccmp(int id1,int id2)
{
  if (FecTbl == NULL) return FALSE;
  if (id1 == id2) return TRUE;
  if (id1 >= numFecTblEntries || id2 >= numFecTblEntries) return FALSE;
  if (id1 < 0 || id2 < 0) return FALSE;
  if (!strnicmp(FecTbl[id1].fecName,FecTbl[id2].fecName,FEC_NAME_SIZE)) return TRUE;
  if (memcmp(&FecTbl[id1].netAdr,&FecTbl[id2].netAdr,sizeof(INADR))) return FALSE;
  if (FecTbl[id1].portOffset != FecTbl[id2].portOffset) return FALSE;
  return TRUE;
}
int eqmcmp(int id1,int id2)
{
  if (SrvTbl == NULL) return FALSE;
  if (id1 == id2) return TRUE;
  if (id1 >= numSrvTblEntries || id2 >= numSrvTblEntries) return FALSE;
  if (!strncmp(SrvTbl[id1].EqmName,SrvTbl[id2].EqmName,EQM_NAME_SHORTSIZE)) return TRUE;
  return FALSE;
}
/**
 * \internal
 * \brief if address is reacquired, we could go from local MMF to loopback UDP and these
 *  have different data MTUs !
 * 
 */
void resetOutputBlocks(int fecidx)
{
  int i, datamtu = PMTU(0,FecTbl[fecidx].inetProtocol,MINIMUM_REQHDR_PROTOCOL); /* worst-case: small */
  ConTblEntry *c;
  UINT32 nb;
  if (datamtu <= 0) return; /* not possible ! */
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if ((c=conTbl[i]) == NULL) continue; /* should be impossible ! */
    if (c->fecIdx != fecidx) continue; /* attached to different FEC */
    nb = (c->sizeBytesOut+STATUS_SIZE)/datamtu + ((c->sizeBytesOut+STATUS_SIZE)%datamtu ? 1 : 0);
    nb *= 3; /* client could be using a small MTU ! */
    if (nb == c->numblks) continue; /* nothing to do */
    if (nb > 0)
    {
      nb += (8 - nb%8); 
      SystemFree(c->bufOutBlks);
      c->bufOutBlks = (BYTE *)SystemCalloc(nb,1);
    }
  }  
  return;
}
#define _DG_PROT(p) ((p & IPX) ? IPX : (gAfInet == AF_INET6) ? UDP6 : UDP) 
/**
 * \internal
 * \brief Gets or finds a free slot in the client connection table.
 *
 * Returns an existing link if the input parameters match an entry in the
 * table
 *
 * This function tries to find an inactive link in the connection
 * table which has the same context, device name, device group, property,
 * and input/ouput data formats.
 *
 * \param dout Pointer to the output data buffer
 * \param din Pointer to the input data buffer 
 * \param devctx EqmContext name
 * \param devgrp Device group
 * \param devnam Device name
 * \param property Property
 * 
 * \todo Move static int initialized to initClient().
 * \todo nConnectionTableEntries: connect incrementation with memory allocation.
 * 
 */
int getConTblEntry(DTYPE *dout,DTYPE *din,char *devctx,char *devgrp,char *devnam,char *devprp,short access,int reuseIndex)
{
  static int initialized = 0;
  UINT32 doutsize,dinsize,reqsize,datamtu,nb;
  int srvidx = -1,fecidx = -non_existent_fec,slot = 0,boundToId = 0, breakOnSlotFound = FALSE, depOk = TRUE;
  int cc=0,i,fmtsizein,devnamlen,devnampad,dsiz,fsiz,reqfmt;
  struct structStruct *s;
  ReLnkTblEntry *rlkEntry = NULL;
  BYTE *dataPtr=NULL;
  char keyStr[128],*key=NULL;
  char ldevgrp[EXPORT_NAME_SIZE+4],ldevprp[PROPERTY_NAME_SIZE+4];
  char ldevnam[DEVICE_NAME_MAXSIZE+4],ldevctx[CONTEXT_NAME_SIZE];
  time_t t = time(NULL);
# ifndef SMALL_TINE_LIB
  BLACKLISTED_LINK *bll;
# endif
  if (tineDebug) dbglog("connect to /%.32s/%.32s/%.64s[%.64s]",devctx==NULL?"(null)":devctx,devgrp,devnam,devprp);
  if (!initialized)
  {
    if ((cc=initClient()) != 0) return -cc;
    initialized = 1;
  }
  if (NameServerLoaded && !strcmp(devgrp,ENS_TAG)) /* Name Server request */
  { /* NOTE: direct ENS_TAG requests do not handle input data sets (no swapping, etc) */
    prepNameServerEntry(dout,din,devctx,devgrp,devnam,devprp);
    return 0;
  }  
  /* has this been redirected or relinked ? */
  reqfmt = dout == NULL ? CF_NULL : dout->dFormat;
  reqsize = dout == NULL ? 0 : dout->dArrayLength;
  ldevctx[0] = 0;
  strncpy(ldevgrp,devgrp,EXPORT_NAME_SIZE);
  strncpy(ldevprp,devprp,PROPERTY_NAME_SIZE);
  strncpy(ldevnam,devnam,DEVICE_NAME_MAXSIZE);
  if (devctx != NULL) strncpy(ldevctx,devctx,CONTEXT_NAME_SIZE);
# ifndef SMALL_TINE_LIB
  if ((bll=getBlackLnkLstItem(ldevctx,ldevgrp,ldevnam,ldevprp)) != NULL) 
  {
    msglog(bll->status,"link %.192s has been blacklisted due to return code %d",bll->key,bll->status);
    return -link_blacklisted;
  }
# endif
  redirectLink(ldevctx,ldevgrp,ldevprp,ldevnam);
# if !defined(SMALL_TINE_LIB)
  rlkEntry = relinkLink(ldevctx,ldevgrp,ldevprp,ldevnam,&key,dout);
# endif
  /* resolve FEC name and EQM name: */
  if ((fecidx=locateExport(ldevgrp,ldevctx,devnam,&srvidx)) < 0) 
  {
#   ifndef SMALL_TINE_LIB
    appendBlackLnkLst(ldevctx,ldevgrp,ldevnam,ldevprp,address_unknown);
#   endif
    return -address_unknown;
  }
  if (srvidx < 0) return -non_existent;
  
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return -semaphore_busy;
  if (nConnectionTableEntries == 0 && NameServerLoaded) 
  { /* in case all active connections were closed and we're starting over ? */
    nConnectionTableEntries = 1;
  }
  /* check to see if in connection Table */
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue; /* should be impossible ! */
    if (i > 0 && slot == 0 &&
      conTbl[i]->heartbeat < HBT_SUB_PENDING &&
      conTbl[i]->needsNotification == FALSE &&
      (time_t)conTbl[i]->starttime < (t - 2) && /* should be covered by the next check */
      (time_t)conTbl[i]->timeLastSync < (t - 2) && /* must be canceled for at least 2 secs */
      BASEMODE(conTbl[i]->mode) == CM_CANCEL)
    {
      slot = i;     /* slot is available */
      if (isTraceLink(i)) traceLinkOutput(conTbl[i], "checkLink", "canceled link is being reused");
    }
    if (breakOnSlotFound) 
    { /* find an available slot if possible */
      if (slot > 0) break; else continue;
    }
    if (strnicmp(ldevnam,conTbl[i]->devName,DEVICE_NAME_SIZE)) 
    { /* not equal at normal length */
      if (conTbl[i]->devName[DEVICE_NAME_SIZE-1] == '&' && conTbl[i]->extStringspace != NULL)
      { /* are the long names equal ? */
        if (strnicmp(&ldevnam[DEVICE_NAME_SIZE-1],conTbl[i]->extStringspace,conTbl[i]->extStringsize))
        { /* not equal at long length -> keep looking */
          continue;
        }
        /* else they're equal -> so fall thru */
      }
      else 
      { /* connection entry is not a long name -> keep looking */
        continue;
      }
    }
    if (ldevctx[0] != 0 && strnicmp(ldevctx,conTbl[i]->ctxName,CONTEXT_NAME_SIZE)) continue;
    if ((feccmp(fecidx,conTbl[i]->fecIdx) && eqmcmp(srvidx,conTbl[i]->srvIdx) &&
         !strnicmp(ldevprp,conTbl[i]->prpName,PROPERTY_NAME_SIZE)) ||
         !ConTblRdrCmp(conTbl[i],ldevctx,devgrp,devnam,devprp))
    { /* check for an existing table entry ... */
      if (conTbl[i]->boundToId > 0 && conTbl[i]->mcaIndex <= 0) 
      { /* already bound to a parent link */
        continue; 
      }
      if (dout && dout->dArrayLength > 0)
      {
        if (dout->dFormat == CF_DEFAULT) 
        { /* a default request has found a 'possible' match : use it */
#         ifndef SMALL_TINE_LIB
          if (IsMcaParentLink(conTbl[i])) continue; /* an MCA link head -> don't use this one ! */
#         endif
          dout->dFormat = conTbl[i]->formatOut;
          fsiz = fmtsizeof(dout->dFormat);
          if (fsiz > 0) dout->dArrayLength /= fsiz;
          strncpy(dout->dTag,conTbl[i]->strTagOut,TAG_NAME_SIZE);
          dsiz = conTbl[i]->sizeOut;
          if (dout->dFormat == CF_STRUCT)
          { /* best possible tack */
            if ((s=findstructEx(dout->dTag,conTbl[i]->ctxName,conTbl[i]->expName)) != NULL && s->rawlen > 0)
            {
              dsiz = (int)conTbl[i]->sizeOut/s->rawlen;
              dout->dArrayLength /= (UINT32)s->rawlen;
            }
          }
          if (conTbl[i]->mcaLink != NULL) dsiz = 1; /* this is an MCA element ! */
          if (dout->dArrayLength > (UINT32)dsiz) dout->dArrayLength = (UINT32)dsiz;
        }
        if (dout->dFormat != conTbl[i]->formatOut) continue;
        if (isTaggableFormat(dout->dFormat) && strncmp(dout->dTag,conTbl[i]->strTagOut,TAG_NAME_SIZE)) continue;
        if (!strncmp(dout->dTag,LWD_DTAG,TAG_NAME_SIZE) && strncmp(dout->dTag,conTbl[i]->strTagOut,TAG_NAME_SIZE)) continue;
        if (dout->dFormat == CF_STRUCT && (s=findstructEx(dout->dTag,conTbl[i]->ctxName,conTbl[i]->expName)) != NULL)
        {
          if (dout->dArrayLength*s->rawlen != conTbl[i]->sizeOut) continue;
        }
        else if (dout->dFormat == CF_HISTORY)
        {
          if ((fsiz=getCarriedFormatSize(dout->dTag)) < 0) continue;
          doutsize = (UINT32)fsiz + HISTORY_HDR_SIZE;
          if (dout->dArrayLength*doutsize != conTbl[i]->sizeOut) continue;
        }
        else if (dout->dFormat == CF_MDA)
        {
          if (strnicmp(conTbl[i]->strTagOut,dout->dTag,TAG_NAME_SIZE)) continue;
          if (dout->dArrayLength != conTbl[i]->sizeOut) continue;
        }
        else if (isAdjustableLengthFormat(dout->dFormat))
        { /* array length carried in the tag */
          if (strncmp(dout->dTag,conTbl[i]->strTagOut,TAG_NAME_SIZE)) continue;
        }
        else
        {
          if (dout->dArrayLength != conTbl[i]->sizeOut) continue;
        }
      }
      else if (conTbl[i]->sizeOut) continue;
      if (din && din->dArrayLength > 0)
      {
        if (conTbl[i]->dataIn == NULL) continue;
        if (din->dFormat != conTbl[i]->formatIn) continue;
        if (isTaggableFormat(din->dFormat) && strncmp(din->dTag,conTbl[i]->strTagIn,TAG_NAME_SIZE)) continue;
        if (din->dFormat == CF_STRUCT && (s=findstructEx(din->dTag,conTbl[i]->ctxName,conTbl[i]->expName)) != NULL)
        {
          sprintf(keyStr,"/%.32s/%.32s",conTbl[i]->ctxName,conTbl[i]->expName);
          if (structcmp(s,keyStr,din,(BYTE *)conTbl[i]->dataIn,conTbl[i]->sizeIn)) continue;
        }
        else if (din->dFormat == CF_HISTORY)
        {
          if ((fsiz=getCarriedFormatSize(din->dTag)) < 0) continue;
          dinsize = (UINT32)fsiz + HISTORY_HDR_SIZE;
          if (din->dArrayLength*dinsize != conTbl[i]->sizeIn) continue;
          if (memcmp(din->data.vptr,dataPtr,(size_t)din->dArrayLength*dinsize)) continue;
        }
        else if (din->dFormat == CF_MDA)
        {
          DMDA *mda = (DMDA *)din->data.vptr;
          if (din->dArrayLength != conTbl[i]->sizeIn) continue;
          if ((fsiz=getCarriedFormatSize(conTbl[i]->strTagIn)) < 0) continue;
          dinsize = conTbl[i]->sizeIn*fsiz;
          if (memcmp(din->data.vptr,conTbl[i]->dataIn,MDA_HDR_SIZE)) continue;
          if (memcmp(mda->data.vptr,&conTbl[i]->dataIn[MDA_HDR_SIZE],dinsize)) continue;
        }
        else
        {
          int hsiz = GetFormatHeaderSize(conTbl[i]->formatIn);
          if (din->dArrayLength != conTbl[i]->sizeIn) continue;
          if ((fsiz=fmtsizeof(conTbl[i]->formatIn)) < 0) continue;
          dinsize = conTbl[i]->sizeIn*fsiz;
          if (hsiz == 0)
          { /* most data types */
            if (memcmp(din->data.vptr, conTbl[i]->dataIn, (size_t)dinsize)) continue;
          }
          else
          {
            BYTE *bptr = GetFormatBodyPtr(din->data.vptr, din->dFormat);
            if (bptr == NULL)
            { /* e.g. DSPECTRUM */
              dinsize += hsiz; /* data follow the header directly */
              if (memcmp(din->data.vptr, conTbl[i]->dataIn, (size_t)dinsize)) continue;
            }
            else
            { /* e.g. DGSPECTURM, DIMAGE */
              if (memcmp(din->data.vptr, conTbl[i]->dataIn, (size_t)hsiz)) continue;
              if (memcmp(bptr, &conTbl[i]->dataIn[hsiz], dinsize)) continue;
            }
          }
        }
      }
      else if (conTbl[i]->sizeIn) continue;
      /* check the WRITE access bit as well */
      if ((access & CA_WRITE) != (conTbl[i]->access & CA_WRITE)) continue;
      if (dout != NULL && dout->dFormat != CF_NULL)
      {
        if (BASEMODE(conTbl[i]->mode) != CM_CANCEL)  
        { /* active link */
          if (isTraceLink(i)) traceLinkOutput(conTbl[i], "getConTblEntry", "bind link to existing entry");
          if (tineDebug) dbglog("latch onto active %s link",modetoa[BASEMODE(conTbl[i]->mode)]);
          boundToId = i;
          dout->xferReason = i;
          conTbl[i]->idleCount = 0; /* reset in case parent is a listening link */
          depOk = allowCommonLinks;
          if (!depOk && (conTbl[i]->relinked & RELINKED_STRUCT))
          {
            char lclkey[FULL_NAME_MAXSIZE+8];
            rlkEntry = getReLnkTblEntry(conTbl[i]);
            makeLinkKey(lclkey,devctx,devgrp,devnam,devprp);
            if (rlkEntry != NULL) if (strnicmp(rlkEntry->key,lclkey,FULL_NAME_MAXSIZE+8)) depOk = TRUE;
          }
          if (depOk) 
          { /* add to dependency list (below) */
            if (slot > 0) break; /* done looking */
            breakOnSlotFound = TRUE; /* see if there's an available slot */
            continue;
          }
          ccerr(link_exists); /* redirect output data */
        }
        else                                            
        { /* not active link */
          if (conTbl[i]->dataMarshalled != NULL && dout->data.vptr != conTbl[i]->dataMarshalled)
          {
            continue;
          }
          else if (dout->data.vptr != conTbl[i]->dataPtr)
          { /* different stored buffer */
            if (dout->data.vptr == NULL && conTbl[i]->allocatedBytes)
            { /* if we're here, we can just reuse the already allocated area */
              dout->data.vptr = conTbl[i]->dataPtr;
            }
            else
            { /* keep looking */
              continue;
            }
          }
        }
      }
      if (tineDebug) dbglog("re-use inactive link");
      conTbl[i]->sizeReq = reqsize;
      conTbl[i]->formatReq = reqfmt;
      conTbl[i]->starttime = (UINT32)t;
      conTbl[i]->dtimestamp = conTbl[i]->dtimestampUSEC = 0;
      conTbl[i]->mcastGrp = inadr_zero;
      conTbl[i]->rdrLink = NULL;
#     ifndef SMALL_TINE_LIB
      clearDependentLinks(i);
#     endif
      conTbl[i]->dpdLink = NULL;
      conTbl[i]->mcaLink = NULL;
      conTbl[i]->mcaEntry = NULL;
      conTbl[i]->mcaIndex = 0;
      conTbl[i]->rdrId = 0;
      conTbl[i]->cbId = 0;
      conTbl[i]->hasReference = FALSE;
      conTbl[i]->cbRef = NULL;
      conTbl[i]->rbModFlags = 0;
      conTbl[i]->dataStatus = not_posted;
      conTbl[i]->linkStatus = notification_pending;
      conTbl[i]->tocounter = 0;
      conTbl[i]->hasNotifiedOnce = FALSE;
      conTbl[i]->cancelPending = FALSE;
      conTbl[i]->notifyPending = 0;
      conTbl[i]->renewalMultiplier = 1;
      conTbl[i]->renewalThreshold = 0;
      conTbl[i]->needsNotification = FALSE;
      conTbl[i]->isWaitingSyncNotification = FALSE;
      conTbl[i]->isWildcardLink = FALSE;
      conTbl[i]->subId = 0;
      conTbl[i]->blksin = 0;
      conTbl[i]->bytesin = 0;
      conTbl[i]->queueLen = 0;
      conTbl[i]->xferReason = CX_NULL;
      conTbl[i]->lastDataSize = 0;
      conTbl[i]->boundToId = 0;
      conTbl[i]->glbTblId  = 0;
      conTbl[i]->tcpSck = 0;
      if (conTbl[i]->tocounter)
      { /* if a SINGLE link comes back, keep the current counter, but put it on the right boundary */
        conTbl[i]->tocounter += (TO_RETRY_THRESHOLD - conTbl[i]->tocounter%TO_RETRY_THRESHOLD);
      }
      conTbl[i]->inetProtocol = _DG_PROT(FecTbl[fecidx].inetProtocol);
      conTbl[i]->rlkEntry = rlkEntry;
      strncpy(conTbl[i]->key, "\0", FULL_NAME_SIZE);
      if (conTbl[i]->bufOutBlks) memset(conTbl[i]->bufOutBlks,0,conTbl[i]->numblks);
      ReleaseSystemMutex(hLinkTblMutex);
      return i;  /* re-use inactive link */
    } /* if contract is in table */
  }  /* for all connection table entries */
  /* brace ourselves for a new entry: */
  if (reuseIndex > 0 && reuseIndex < nConnectionTableEntries)
  {
    slot = reuseIndex;
    conTbl[slot]->rdrLink = NULL;
#   ifndef SMALL_TINE_LIB
    clearDependentLinks(slot);
#   endif
    conTbl[slot]->dpdLink = NULL;
    conTbl[slot]->boundToId = 0;
    conTbl[slot]->mcaLink = NULL;
    conTbl[slot]->mcaEntry = NULL;
    conTbl[slot]->mcaIndex = 0;
    conTbl[slot]->rlkEntry = rlkEntry;
  }
  else if (slot == 0)
  {
    slot = nConnectionTableEntries;
    if ((cc=assignConTblMemory(nConnectionTableEntries)) != 0) ccerr(cc);
  }
  else
  {
    if (conTbl[slot]->allocatedBytes) SystemFree(conTbl[slot]->dataPtr);
    SystemFree(conTbl[slot]->dataIn);
    SystemFree(conTbl[slot]->bufOut);
    SystemFree(conTbl[slot]->bufOutBlks);
    SystemFree(conTbl[slot]->bufCmp);
    SystemFree(conTbl[slot]->bufVtl);
    SystemFree(conTbl[slot]->dtf);
    SystemFree(conTbl[slot]->extStringspace);
    memset(conTbl[slot],0,sizeof(ConTblEntry));
  }
  /* assign data indices and other parameters in connection Table: */
  if (ldevctx[0] != 0)
    strncpy(conTbl[slot]->ctxName,(char *)ldevctx,CONTEXT_NAME_SIZE);
  else conTbl[slot]->ctxName[0] = 0;
  strncpy(conTbl[slot]->expName,(char *)ldevgrp,EXPORT_NAME_SIZE);
  strncpy(conTbl[slot]->devName,(char *)ldevnam,DEVICE_NAME_SIZE);
  if ((devnamlen=(int)strlen(ldevnam)) > DEVICE_NAME_SIZE)
  { /* long device name */
    conTbl[slot]->devName[DEVICE_NAME_SIZE-1] = '&';
    devnamlen -= DEVICE_NAME_SIZE-1;
    devnampad = devnamlen % 2; /* make sure it's an even number of extra bytes */
    if ((conTbl[slot]->extStringspace=(char *)SystemCalloc(1,(size_t)(devnamlen+devnampad))) == NULL)
      ccerr(out_of_client_memory);
    conTbl[slot]->extStringsize = devnamlen + devnampad;
    strncpy(conTbl[slot]->extStringspace,(char *)&ldevnam[DEVICE_NAME_SIZE-1],devnamlen);
    if (devnampad > 0) conTbl[slot]->extStringspace[devnamlen] = 0;
  }
  strncpy(conTbl[slot]->prpName,(char *)ldevprp,PROPERTY_NAME_SIZE);
  conTbl[slot]->formatIn = din ? din->dFormat : CF_NULL;
  conTbl[slot]->formatOut = dout ? dout->dFormat : CF_NULL;
  if (din && isTaggableFormat(din->dFormat))
    strncpy(conTbl[slot]->strTagIn,din->dTag,TAG_NAME_SIZE);
  if (dout && (isTaggableFormat(dout->dFormat) || !strcmp(dout->dTag,LWD_DTAG)))
    strncpy(conTbl[slot]->strTagOut,dout->dTag,TAG_NAME_SIZE);
  conTbl[slot]->srvIdx = srvidx;
  conTbl[slot]->fecIdx = fecidx;
  setConTblSize(&conTbl[slot]->sizeIn,din,devctx,devgrp);
  setConTblSize(&conTbl[slot]->sizeOut,dout,devctx,devgrp);
  conTbl[slot]->sizeReq = reqsize;
  conTbl[slot]->formatReq = reqfmt;
  conTbl[slot]->tineProtocol = FecTbl[fecidx].tineProtocol;
  conTbl[slot]->mode = CM_REGISTER;
  conTbl[slot]->blknumIn  = 1;
  conTbl[slot]->numblksIn = 1;
  conTbl[slot]->nextDataPtr = 0;
  conTbl[slot]->blkidIn = 0;
  strncpy(conTbl[slot]->key, "\0", FULL_NAME_SIZE);
  /* define the input data mtu : */
  conTbl[slot]->datamtu = SUBMTU(FecTbl[fecidx].inetProtocol,DEFAULT_TRANSPORT_PROTOCOL_LEVEL); 
  conTbl[slot]->inetProtocol = _DG_PROT(FecTbl[fecidx].inetProtocol);
  if ((time_t)conTbl[slot]->starttime == t) conTbl[slot]->starttime++;
  else conTbl[slot]->starttime = (UINT32)t;
  conTbl[slot]->dtimestamp = 0;
  conTbl[slot]->dtimestampUSEC = 0;
  conTbl[slot]->rbModFlags = 0;
  conTbl[slot]->dataStatus = not_posted;
  conTbl[slot]->linkStatus = notification_pending;
  conTbl[slot]->tocounter = 0;
  conTbl[slot]->renewalMultiplier = 1;
  conTbl[slot]->renewalThreshold = 0;
  conTbl[slot]->hasNotifiedOnce = FALSE;
  conTbl[slot]->cancelPending = FALSE;
  conTbl[slot]->notifyPending = 0;
  conTbl[slot]->needsNotification = FALSE;
  conTbl[slot]->isWildcardLink = FALSE;
  conTbl[slot]->queueDepth = gClnRecvQueueDepth;
  conTbl[slot]->subId = 0;
  conTbl[slot]->blksin = 0;
  conTbl[slot]->bytesin = 0;
  conTbl[slot]->queueLen = 0;
  conTbl[slot]->lastDataSize = 0;
  conTbl[slot]->tcpSck = 0;
  conTbl[slot]->tocounter = 0;
  conTbl[slot]->xferReason = CX_NULL;
  conTbl[slot]->boundToId = 0;
  conTbl[slot]->rdrLink = NULL;
# ifndef SMALL_TINE_LIB
  clearDependentLinks(slot);
# endif
  conTbl[slot]->dpdLink = NULL;
  conTbl[slot]->cbId = 0;
  conTbl[slot]->hasReference = FALSE;
  conTbl[slot]->cbRef = NULL;
  if (reuseIndex > 0 && reuseIndex < nConnectionTableEntries)
  {
    reuseIndex = -1;
    ReleaseSystemMutex(hLinkTblMutex);
    return slot;  /* return here and don't mess with the memory ...*/
  }
  conTbl[slot]->isWaitingSyncNotification = FALSE;
  if (din && din->data.vptr != NULL)
  { /* an input data object was passed */
    if (din->dFormat == CF_STRUCT && (s=findstructEx(din->dTag,devctx,devgrp)) != NULL)
    { /* registered tagged structure (reference to link calling parameters!) */
      conTbl[slot]->sizeIn = din->dArrayLength*s->rawlen;
      if (s->hasExtSpace) 
      { 
        conTbl[slot]->sizeIn = fixStructExtendedSpace(s,din,devctx,devgrp);
      }
    }
    else if (din->dFormat == CF_HISTORY)
    { /* a HISTORY type special */
      if ((dsiz=getCarriedFormatSize(din->dTag)) < 0) ccerr(-dsiz);
      dsiz += HISTORY_HDR_SIZE;
      conTbl[slot]->sizeIn = din->dArrayLength*dsiz;
    }
    fmtsizein = fmtsizeof(din->dFormat);
    if (din->dFormat == CF_MDA) fmtsizein = getCarriedFormatSize(din->dTag);
    if (fmtsizein < 0) ccerr(illegal_format);
    if ((dinsize=conTbl[slot]->sizeIn*fmtsizein) == 0)
    { /* object was passed, but nothing to send */
      conTbl[slot]->dataIn = NULL;
    }
    else 
    { /* got something to send! */
      dinsize += GetFormatHeaderSize(din->dFormat);
      if (isAdjustableLengthFormat(din->dFormat))
      { /* buffer was already allocated and prepared : use it */
        conTbl[slot]->dataIn = din->data.cptr;
      }
      else 
      { /* make a copy of the given input data */
        if ((conTbl[slot]->dataIn=(char *)SystemCalloc(1,(size_t)dinsize)) == NULL)
          ccerr(out_of_client_memory);
      }
      if (fmtsizein > 0) conTbl[slot]->datamtu -= conTbl[slot]->datamtu % fmtsizein;
      cpyInputDataBytes(conTbl[slot],din,(size_t)dinsize);
      conTbl[slot]->bytesTotal = dinsize;
      conTbl[slot]->numblksIn = 
        dinsize/conTbl[slot]->datamtu + (dinsize%conTbl[slot]->datamtu ? 1 : 0);
      conTbl[slot]->bytesRemaining = conTbl[slot]->bytesTotal;
      conTbl[slot]->nextDataSize = MIN(conTbl[slot]->datamtu,conTbl[slot]->bytesRemaining);
      if (conTbl[slot]->extStringsize > 0)
      { /* need to adjust the above since the extended string space will eat up the initial data payload space */
        conTbl[slot]->nextDataSize -= conTbl[slot]->extStringsize; 
      }
    }
  }
  else 
  {
    conTbl[slot]->dataIn = NULL;
  }
  if (dout)
  {
    if (dout->dFormat == CF_STRUCT && (s=findstructEx(dout->dTag,devctx,devgrp)) != NULL)
    { /* registered tagged structure (reference to link calling parameters!) */
      conTbl[slot]->sizeOut = dout->dArrayLength*s->rawlen;
      doutsize = dout->dArrayLength*s->size;
      if (s->rawlen < s->size)
      { /* cannot allow a queue in this case ! */
        conTbl[slot]->queueDepth = 0;
      }
    }
    else if (isAdjustableLengthFormat(dout->dFormat))
    {
      doutsize = dout->dArrayLength;
    }
    else if (dout->dFormat == CF_HISTORY)
    { /* special kind of struct: con->sizeOut is bytes not elements */
      if ((dsiz=(UINT32)getCarriedFormatSize(dout->dTag)) < 0) ccerr(-dsiz);
      dsiz += HISTORY_HDR_SIZE;
      doutsize = dsiz*dout->dArrayLength;
      conTbl[slot]->sizeOut = doutsize;
    }
    else
    {
      fsiz = fmtsizeof(dout->dFormat);
      if (dout->dFormat == CF_MDA) fsiz = getCarriedFormatSize(dout->dTag);
      if (fsiz < 0) ccerr(illegal_format);
      doutsize = conTbl[slot]->sizeOut*fsiz;
      doutsize += getFormatHeaderSize(dout->dFormat);
    }
    if (dout->data.vptr == NULL) 
    { /* unassigned pointer => unbound data => allocate memory here ! */
      UINT32 xsz = 0;
      switch (dout->dFormat)
      {
        case CF_MDA: 
          xsz = sizeof(DUNION); 
          break;
        case CF_IMAGE: 
        case CF_GSPECTRUM:
          xsz = sizeof(UINT32) + sizeof(size_t);
          break;
      }
      if ((conTbl[slot]->dataPtr =(char *)SystemCalloc(1,(size_t)(doutsize+xsz))) == NULL)
        ccerr(out_of_client_memory);
      dout->data.vptr = conTbl[slot]->dataPtr;
      conTbl[slot]->allocatedBytes = doutsize + xsz;
      switch (dout->dFormat)
      {
        case CF_MDA:
          ((DMDA *)dout->data.vptr)->data.vptr = &conTbl[slot]->dataPtr[MDA_HDR_SIZE];
          break;
        case CF_IMAGE:
          ((DIMAGE *)dout->data.vptr)->frameBufferSize = doutsize - IMAGE_HDR_SIZE;
          ((DIMAGE *)dout->data.vptr)->frameBuffer = (BYTE *)&conTbl[slot]->dataPtr[IMAGE_HDR_SIZE + xsz];
          break;
        case CF_GSPECTRUM:
          ((DGSPECTRUM *)dout->data.vptr)->spectBufferSize= doutsize - GSPECTRUM_HDR_SIZE;
          ((DGSPECTRUM *)dout->data.vptr)->spectBuffer = (float *)&conTbl[slot]->dataPtr[GSPECTRUM_HDR_SIZE + xsz];
          break;
      }
    }
    else 
    { /* data bound to given reference pointer */
      conTbl[slot]->dataPtr = (char *)dout->data.vptr;
    }
    /* prepare output data mtu */
    datamtu = PMTU(0,FecTbl[fecidx].inetProtocol,MINIMUM_REQHDR_PROTOCOL); /* worst-case: small */
    /* worst-case: could send data + status */
    conTbl[slot]->numblks = (doutsize+STATUS_SIZE)/datamtu + ((doutsize+STATUS_SIZE)%datamtu ? 1 : 0);
    conTbl[slot]->numblks *= 3; /* client could be using a small MTU ! */
    if (conTbl[slot]->numblks)
    {
      nb = conTbl[slot]->numblks;
      nb += (8 - nb%8); 
      conTbl[slot]->bufOut = (char *)SystemCalloc(1,(size_t)doutsize+STATUS_SIZE);
      conTbl[slot]->bufOutBlks = (BYTE *)SystemCalloc(nb,1);
    }
  }
  else 
  { /* STATUS will come back in the payload no matter what */
    conTbl[slot]->dataPtr = NULL;
    conTbl[slot]->numblks = 1;
    conTbl[slot]->bufOutBlks = (BYTE *)SystemCalloc(8,1);
  }
  conTbl[slot]->rlkEntry = rlkEntry;
# ifndef SMALL_TINE_LIB
  if (boundToId > 0)
  { /* this link depends on another */
    if (slot != boundToId)
    { /* don't let something crazy happen */
      conTbl[slot]->boundToId = boundToId;
      if ((cc=addDependentLink(slot)) != 0) ccerr(cc);
    }
  }
# endif /* SMALL_TINE_LIB */
err:
  if (cc != 0)
  { /* allocation error ! */
    ReleaseSystemMutex(hLinkTblMutex);
    return -cc;
  }
  /* better together with allocation */
  if (slot == nConnectionTableEntries) IncrementConnections(nConnectionTableEntries);
  ReleaseSystemMutex(hLinkTblMutex);
  return slot; /* everythings OK, return list No. */
}
/**
 * \internal return influences whether WaitForSyncLink should call clientCycle() or not
 */
void ensureSendRequest(void)
{
# ifdef MULTITHREADED
  if (!gWaitingOnSelect) return; /* no need to do anything here */
  checkSubscriptions(FALSE);
# endif
}
void waitClientNotify(int id,int timeout)
{
# ifdef MULTITHREADED
  int i;
  if (ServerExitCondition) return;
  if (WaitForSemaphore(hNotificationSemaphore,timeout) == 0)
  { /* got the semaphore */
    if (conTbl[id]->isWaitingSyncNotification)
    { /* but still waiting; does somebody else want it ? */
      for (i=0; i<nConnectionTableEntries; i++)
      {
        if (i == id) continue;
        if (conTbl[i] == NULL) continue;
        if (conTbl[i]->isWaitingSyncNotification)
        { /* maybe this one !*/
          ReleaseSystemSemaphore(hNotificationSemaphore);
          return;
        }
      }
    }
  }
# else
  UNUSED(id);
  UNUSED(timeout);
# endif
}
void _cbExecLink(int id,int cc)
{
  UNUSED(cc);
  if (!isValidConTblEntry(id)) return;
  conTbl[id]->linkStatus = cc; /* in case of redirection: reset this */
  conTbl[id]->isWaitingSyncNotification = FALSE;
# ifdef MULTITHREADED
  ReleaseSystemSemaphore(hNotificationSemaphore);
# endif
}
int WaitForSyncLink(int id, int to)
{
  struct timeval tv0,tv;
  int td = 0, cc;
  gettimeofday(&tv0,(struct timezone *)NULL);
  tv = tv0;
  conTbl[id]->isWaitingSyncNotification = TRUE;
  ensureSendRequest();
  while (conTbl[id]->isWaitingSyncNotification)
  {
    if (!gWaitingOnSelect) clientCycle(); else waitClientNotify(id,gSystemTick);
    gettimeofday(&tv,(struct timezone *)NULL);
    if (ServerExitCondition) return is_terminating;
    td = tvDeltaMs(tv, tv0);
    if (to >= 0 && td >= to) break;
    if (conTbl[id]->dpdLink != NULL || conTbl[id]->mcaIndex > 0)
    { /* the parent will have set this */
      if (conTbl[id]->hasNotifiedOnce)
        conTbl[id]->isWaitingSyncNotification = FALSE;
    }
  }
  waitClientNotify(id,0);
  if (tineDebug > 3) dbglog("synchronous link : %d msec",td);
  cc = conTbl[id]->isWaitingSyncNotification ? -1 : 0;
  conTbl[id]->isWaitingSyncNotification = FALSE;
  return cc;
}
TINE_EXPORT int PutLnkTimeStamp(int i)
{
  if (i >= nConnectionTableEntries || i < 0) return argument_list_error;
  if (conTbl[i] == NULL || conTbl[i]->mode == CM_CANCEL) return invalid_link;
  conTbl[i]->rbModFlags |= RB_USE_TIMESTAMP;
  return 0;
}
TINE_EXPORT int PutLnkErrorValue(int i, void *errValue)
{
  if (i >= nConnectionTableEntries || i < 0) return argument_list_error;
  if (conTbl[i] == NULL || conTbl[i]->mode == CM_CANCEL) return invalid_link;
  memcpy(conTbl[i]->errValue,errValue,64);
  conTbl[i]->rbModFlags |= RB_USE_ERRORVALUE;
  return 0;
}
int prepIncomingSubResponseHdr(int tineProtocol, SubRspHdr *phdr,BYTE *stream)
{
  switch (tineProtocol)
  {
    case 5:
    {
      PrdrHdr5 PHdr5, *pPHdr5 = &PHdr5;
#     ifdef INTEL
      memcpy(pPHdr5, stream, sizeof(PrdrHdr5));
#     else
      memset(pPHdr5, 0, sizeof(PrdrHdr5));
      _structSwap(PrdrHdr5Struct, stream, 0, NULL, (BYTE *)pPHdr5, RTOL);
#     endif
      phdr->msgSize = pPHdr5->msgsize;
      phdr->subId = pPHdr5->subId;
      phdr->retCode = pPHdr5->retCode;
      phdr->stscode = pPHdr5->retCode;
      phdr->numblks = pPHdr5->numblks;
      phdr->blknum = pPHdr5->blknum;
      phdr->mtu = pPHdr5->mtu;
      phdr->timestampUSec = pPHdr5->timestampMSEC * 1000;
      phdr->dataFormat = pPHdr5->EqmFormat;
      phdr->counter = pPHdr5->counter;
      phdr->timestamp = pPHdr5->timestamp;
      phdr->starttime = pPHdr5->ClnStarttime;
      phdr->stssize = pPHdr5->retCode == 0 ? 0 : STATUS_SHORTSIZE;
      phdr->tineProtocol = 5;
      break;
    }
    case 6:
    {
      PrdrHdr PHdr, *pPHdr = &PHdr;
#     ifdef INTEL
      memcpy(pPHdr, stream, sizeof(PrdrHdr));
#     else
      memset(pPHdr, 0, sizeof(PrdrHdr5));
      _structSwap(PrdrHdrStruct, stream, 0, NULL, (BYTE *)pPHdr, RTOL);
#     endif
      phdr->msgSize = pPHdr->msgsize;
      phdr->subId = pPHdr->subId;
      phdr->retCode = pPHdr->retCode;
      phdr->stscode = pPHdr->stscode;
      phdr->numblks = pPHdr->numblks;
      phdr->blknum = pPHdr->blknum;
      phdr->mtu = pPHdr->mtu;
      phdr->timestampUSec = pPHdr->timestampUSec;
      phdr->dataFormat = pPHdr->EqmFormat;
      phdr->counter = pPHdr->counter;
      phdr->timestamp = pPHdr->timestamp;
      phdr->timestampUSec = pPHdr->timestampUSec;
      phdr->starttime = pPHdr->ClnStarttime;
      phdr->stssize = pPHdr->stssize;
      phdr->tineProtocol = pPHdr->tineProtocol;
      phdr->xferReason = pPHdr->xferReason;
      phdr->userstamp = pPHdr->userstamp;
      phdr->sysstamp = pPHdr->sysstamp;
      phdr->dataSize = 0; /* could make a guess, but ... */
      phdr->msgSizeInElements = 0; /* could make a guess, but ... */
      break;
    }
    case 7:
#     ifdef INTEL
      memcpy(phdr, stream, sizeof(SubRspHdr));
#     else
      memset(phdr, 0, sizeof(SubRspHdr));
      _structSwap(SubRspHdrStruct, stream, 0, NULL, (BYTE *)phdr, RTOL);
#     endif
      break;
    default:
       return illegal_protocol;
  }
  return 0;
}
int prepOutgoingRequestHdr(int protocol, UINT32 totalSizeBytes, int segments, const char *usr, const char *typ, BYTE *stream)
{ /* fill in the request header */
  if (protocol < MINIMUM_REQHDR_PROTOCOL)
  {
    PktHdr h;
    h.totalSize = (UINT16)totalSizeBytes;
    memcpy(h.userName, usr, USERNAME_SIZE);
    h.tineProtocol = (short)protocol;
    /* legacy globals use protocol = 4 and fill in the number of contained globals as 'number' */
    h.number = protocol < 5 ? (UINT16)segments : (UINT16)REVISIONID(99, 99); /* so as to interface with any R4 server */
#   ifdef INTEL
    memcpy(stream, &h, PKTHDR_SIZE);
#   else
    _structSwap(PktHdrStructStruct, stream, 0, NULL, (BYTE *)&h, LTOR);
#   endif
  }
  else
  {
    ReqHdr rh;
    rh.totalSizeInBytes = totalSizeBytes;
    rh.headerSizeInBytes = (UINT32)sizeof(rh);
    rh.numberSegments = segments;
    rh.tineProtocol = (SINT16)protocol;
    rh.tineVersion = VERSIONID(MAJOR_VERSION, MINOR_VERSION, VERSION_REVISION);
    rh.encoding = gCharEncoding;
    rh.endianness = gEndianness;
    strncpy(rh.userName,usr,USERNAME_SIZE);
    strncpy(rh.userType,typ,USERTYPE_SIZE);
    rh.pid = SystemGetProcessId();
#   ifdef INTEL
    memcpy(stream, &rh, sizeof(ReqHdr));
#   else
    _structSwap(ReqHdrStructStruct, stream, 0, NULL, (BYTE *)&rh, LTOR);
#   endif
  }
  return 0;
}
int prepOutgoingSubRequestHdr(int tineProtocol,ConTblEntry *c,char *eqm,int datasize,int subid,BYTE *pSubstream)
{
  switch (tineProtocol)
  {
    case 5:
    { /* release 3.xx era */
      SubInfoPkt5 sub5, *pSub5 = &sub5;
      memset(pSub5, 0, sizeof(SubInfoPkt5));
      pSub5->msgsize = SUBINFO5_SIZE + datasize;
      pSub5->mtu = c->datamtu;
      pSub5->numblks = (BYTE)c->numblksIn;
      pSub5->blknum = (BYTE)c->blknumIn;
      pSub5->id = subid;
      pSub5->mode = (short)TARGETMODE(c->mode);
      pSub5->blkid = c->blkidIn;
      pSub5->pollingInterval = c->pollingInterval;
      pSub5->starttime = c->starttime;
      strncpy(pSub5->contract.EqmName, eqm, EQM_NAME_SIZE);
      strncpy(pSub5->contract.EqmProperty, c->prpName, PROPERTY_NAME_SHORTSIZE);
      strncpy(pSub5->contract.EqmDeviceName, c->devName, DEVICE_NAME_SHORTSIZE);
      if (strlen(c->devName) > DEVICE_NAME_SHORTSIZE && strlen(c->prpName) <= DEVICE_NAME_SHORTSIZE)
      {
        pSub5->contract.EqmDeviceName[DEVICE_NAME_SHORTSIZE - 1] = '&';
        pSub5->contract.EqmProperty[DEVICE_NAME_SHORTSIZE] = 0;
        strncpy(&pSub5->contract.EqmProperty[DEVICE_NAME_SHORTSIZE + 1], &c->devName[DEVICE_NAME_SHORTSIZE - 1], DEVICE_NAME_SHORTSIZE - 1);
      }
      pSub5->contract.EqmSizeOut = c->sizeOut;
      pSub5->contract.EqmSizeIn = c->sizeIn;
      pSub5->contract.EqmAccess = (BYTE)c->access;
      pSub5->contract.hEqmName = 0;
      pSub5->contract.EqmFormatOut = (BYTE)(BFMT(c->formatOut));
      pSub5->contract.EqmFormatIn = (BYTE)(BFMT(c->formatIn));
      strncpy(pSub5->contract.strTagIn, c->strTagIn, TAG_NAME_SHORTSIZE);
      strncpy(pSub5->contract.strTagOut, c->strTagOut, TAG_NAME_SHORTSIZE);
#     ifdef INTEL
      memcpy(pSubstream, pSub5, sizeof(SubInfoPkt5));
#     else
      memset(pSubstream, 0, sizeof(SubInfoPkt5));
      _structSwap(SubInfoPkt5Struct, pSubstream, 0, NULL, (BYTE *)pSub5, LTOR);
#     endif
      break;
    }
    case 6:
    { /* release 4.xx era */
      SubInfoPkt sub, *pSub = &sub;
      memset(pSub, 0, sizeof(SubInfoPkt));
      pSub->msgsize = SUBINFO_SIZE + datasize;
      pSub->pollingInterval = c->pollingInterval;
      pSub->id = subid;
      pSub->blknum = c->blknumIn;
      pSub->numblks = (UINT16)c->numblksIn;
      pSub->blkid = c->blkidIn;
      pSub->mtu = c->datamtu;
      pSub->mode = (short)TARGETMODE(c->mode);
      pSub->starttime = c->starttime;
      strncpy(pSub->contract.EqmName, eqm, EQM_NAME_SIZE);
      strncpy(pSub->contract.EqmProperty, c->prpName, PROPERTY_NAME_SIZE);
      strncpy(pSub->contract.EqmDeviceName, c->devName, DEVICE_NAME_SIZE);
      pSub->contract.EqmSizeOut = c->sizeOut;
      pSub->contract.EqmSizeIn = c->sizeIn;
      pSub->contract.EqmAccess = (BYTE)c->access;
      pSub->contract.hEqmName = 0;
      pSub->contract.EqmFormatOut = (BYTE)(BFMT(c->formatOut));
      pSub->contract.EqmFormatIn = (BYTE)(BFMT(c->formatIn));
      strncpy(pSub->contract.strTagIn, c->strTagIn, TAG_NAME_SIZE);
      strncpy(pSub->contract.strTagOut, c->strTagOut, TAG_NAME_SIZE);
      pSub->extsize = c->extStringspace == NULL ? 0 : c->extStringsize;
      pSub->msgsize += pSub->extsize;
#     ifdef INTEL
      if ((void *)pSubstream != (void *)pSub) memcpy(pSubstream, pSub, sizeof(SubInfoPkt));
#     else
      memset(pSubstream, 0, sizeof(SubInfoPkt));
      _structSwap(SubInfoPktStruct, pSubstream, 0, NULL, (BYTE *)pSub, LTOR);
#     endif
      break;
    }
    case 7:
    {
      SubReqHdr sub, *pSub = &sub;
      memset(pSub, 0, sizeof(SubReqHdr));
      pSub->msgsize = SUBREQHDR_SIZE + datasize;
      pSub->pollingInterval = c->pollingInterval;
      pSub->id = subid;
      pSub->blknum = (UINT16)c->blknumIn;
      pSub->numblks = (UINT16)c->numblksIn;
      pSub->blkid = c->blkidIn;
      pSub->mtu = c->datamtu;
      pSub->mode = (short)TARGETMODE(c->mode);
      pSub->starttime = c->starttime;
      strncpy(pSub->contract.EqmName, eqm, EQM_NAME_SIZE);
      strncpy(pSub->contract.EqmProperty, c->prpName, PROPERTY_NAME_SIZE);
      strncpy(pSub->contract.EqmDeviceName, c->devName, DEVICE_NAME_SIZE);
      pSub->contract.EqmSizeOut = c->sizeOut;
      pSub->contract.EqmSizeIn = c->sizeIn;
      pSub->contract.EqmAccess = (BYTE)c->access;
      pSub->contract.hEqmName = 0;
      pSub->contract.EqmFormatOut = (BYTE)(BFMT(c->formatOut));
      pSub->contract.EqmFormatIn = (BYTE)(BFMT(c->formatIn));
      strncpy(pSub->contract.strTagIn, c->strTagIn, TAG_NAME_SIZE);
      strncpy(pSub->contract.strTagOut, c->strTagOut, TAG_NAME_SIZE);
      pSub->extsize = c->extStringspace == NULL ? 0 : c->extStringsize;
      pSub->msgsize += pSub->extsize;
#     ifdef INTEL
      if ((void *)pSubstream != (void *)pSub) memcpy(pSubstream, pSub, sizeof(SubReqHdr));
#     else
      memset(pSubstream, 0, sizeof(SubReqHdr));
      _structSwap(SubReqHdrStruct, pSubstream, 0, NULL, (BYTE *)pSub, LTOR);
#     endif
      break;
    }
    default:
       return illegal_protocol;
  }
  c->mode &= ~CM_RETRY; /* remove this here */
  return 0;
}
int prepIncomingResponseHdr(RspHdr *hdr, RspSpecs *spc, BYTE *stream, UINT32 totalSizeInBytes)
{
  UINT32 p = IPEEK(*(UINT32 *)stream);
  if (p > DEFAULT_TRANSPORT_PROTOCOL_LEVEL)
  { /* indicative of legacy system (pre Release 5); 1st 2 bytes contain total size */
    p = SPEEK(*(UINT16 *)&stream[PHDR_PROT_OFFSET+2]) == 6 ? 6 : 5;
  }
  if (p < MINIMUM_REQHDR_PROTOCOL)
  { /* map to modern request header */
    ConTblEntry *c = NULL;
    int id = -1;
    UINT16 hsiz = 2;
    hdr->tineProtocol = p;

    if (totalSizeInBytes == 0)
    { /* normal (reassembly) case */
      totalSizeInBytes = (UINT32)SPEEK(*(UINT16 *)stream);
    }
    else
    { /* stream case (it's all in) */
      UINT32 msgsiz = IPEEK(*(UINT32 *)stream);
      if (msgsiz > totalSizeInBytes) totalSizeInBytes = msgsiz;
      hsiz = 2; /* this is artifically still 2 ! */
    }
    hdr->totalSizeInBytes = totalSizeInBytes;
    hdr->endianness = BO_LITTLE_ENDIAN;
    hdr->encoding = CE_ASCII;
    hdr->headerSizeInBytes = hsiz;
    hdr->numberSegments = 0;
    id = (int)SPEEK(*(SINT16 *)&stream[PHDR_SUBID_OFFSET+2]);
    if (id >= 0 && id < nConnectionTableEntries)
    {
      if ((c = conTbl[id]) != NULL && c->fecIdx >= 0)
      {
        strncpy(hdr->fecName, FecTbl[c->fecIdx].fecName, FEC_NAME_SIZE);
      }
    }
    /* response specifications from initial subscription : */
    if (spc == NULL) return p;
    spc->prot = p;
    spc->id = (UINT16)id;
    spc->cc = SPEEK(*(UINT16 *)&stream[PHDR_CC_OFFSET + 2]);
    spc->fmt = SPEEK(*(UINT16 *)&stream[PHDR_FMT_OFFSET + 2]);
    spc->stssiz = SPEEK(*(UINT16 *)&stream[PHDR_STSSIZ_OFFSET + 2]);
    spc->starttime = IPEEK(*(UINT32 *)&stream[PHDR_STTM_OFFSET + 2]);
    spc->sysstamp = IPEEK(*(UINT32 *)&stream[PDHR_SYST_OFFSET + 2]);
    spc->userstamp = IPEEK(*(UINT32 *)&stream[PDHR_URST_OFFSET + 2]);
  }
  else
  {
#   ifdef INTEL
    memcpy(hdr, stream, sizeof(RspHdr));
#   else
    _structSwap(RspHdrStructStruct, stream, 0, NULL, (BYTE *)hdr, RTOL);
#   endif
    /* TODO: fill in response specs */
    if (spc == NULL) return p;
    spc->prot = p;
    spc->id = SPEEK(*(UINT16 *)&stream[SUBRSPHDR_SUBID_OFFSET + RSPHDR_SIZE]);
    spc->cc = SPEEK(*(UINT16 *)&stream[SUBRSPHDR_CC_OFFSET + RSPHDR_SIZE]);
    spc->fmt = SPEEK(*(UINT16 *)&stream[SUBRSPHDR_FMT_OFFSET + RSPHDR_SIZE]);
    spc->stssiz = SPEEK(*(UINT16 *)&stream[SUBRSPHDR_STSSIZ_OFFSET + RSPHDR_SIZE]);
    spc->starttime = IPEEK(*(UINT32 *)&stream[SUBRSPHDR_STTM_OFFSET + RSPHDR_SIZE]);
    spc->sysstamp = IPEEK(*(UINT32 *)&stream[SUBRSPHDR_SYST_OFFSET + RSPHDR_SIZE]);
    spc->userstamp = IPEEK(*(UINT32 *)&stream[SUBRSPHDR_URST_OFFSET + RSPHDR_SIZE]);
  }
  return p;
}

static UINT32 assignHeartbeat(ConTblEntry *c)
{ /* not exported : don't check null pointers */
  UINT32 hbt;
  int bmode = BASEMODE(c->mode);
  if (bmode == CM_CANCEL)
  {
    hbt = HBT_NONE;
  }
  else if ((bmode == CM_DATACHANGE) || (bmode == CM_EVENT))
  { /* standard 'isalive' heartbeat */
    hbt = (UINT32)(2+HBT_MAXIMUM);
  }
  else
  { /* heatbeat: 10 minute maximum, 1 sec minimum  */
    hbt = MAX((UINT32)HBT_MINIMUM,c->pollingInterval/1000);
    if (hbt > 600) hbt = 600;
  }
  if (c->linkStatus == (short)link_timeout ||
      c->linkStatus == (short)notification_pending) hbt = (UINT32)HBT_MINIMUM;
  if ((c->linkStatus & 0x0fff) == information_static)
  { /* close the link remotely but keep a link table entry locally */
    c->mode = CM_REGISTER;
    c->linkStatus = 0;
    hbt = HBT_NONE;
  }
  return hbt;
}
void prepOutgoingData(BYTE *d,BYTE *s,UINT32 siz,short fmt)
{
  memcpy(d,s,siz);
# ifdef INTEL
  UNUSED(fmt);
# else
  if (GetFormatBodyPtr(d, fmt) == NULL) siz /= getFormatSize(fmt);
  SwapBytes((char *)d,siz,fmt,LTOR);
# endif
}
SOCKET getTcpConTblSocket(ConTblEntry *c,int inetProtocol)
{ /* used internally : no check against NULL pointer necessary */
  IDLE_CONNECTION *ic=NULL;
  TCPBCKT *bckt;
  int port=gtTCPPort, buflen=TCP_BUFFER_SIZE, clrlen=0;
  char lnkn[256], *lnkp = getProtocolAsString(inetProtocol);
# ifdef MULTITHREADED
  int mode=0;
# endif
  if (gClientIsClosing) return c->tcpSck;
  makeLinkKey(lnkn, c->ctxName, c->expName, c->devName, c->prpName);
  SuspendSystemMutex(hLinkLstMutex);
  if (c->tcpSck == 0)
  { /* not yet connected ? */
    TCPSCKT *sckt = NULL;
    if (WaitForSystemMutex(hLinkLstMutex, -1) == 0)
    {
      for (sckt=tcpClnSckTbl; sckt != NULL; sckt = sckt->nxt)
      {
        if (c->fecIdx == sckt->idx && sckt->type == inetProtocol)
        { /* but there is a connection to the FEC */
          c->tcpSck = sckt->sck;
          c->inetProtocol = inetProtocol;
          msglog(0, "joining %s connection to %.256s", lnkp, lnkn);
          break;
        }
      }
      ReleaseSystemMutex(hLinkLstMutex);
    }
    if ((ic=GetIdleConnection(c->fecIdx, inetProtocol)) != NULL)
    { /* no longer idle */
      RemoveIdleConnection(ic);
    }
  }
  if (c->tcpSck == 0)
  { /* still not connected */
#   ifdef MULTITHREADED
    clrlen = TCP_BUFFER_SIZE;
    buflen = c->sizeOut * fmtsizeof(c->formatOut) + TCP_BUFFER_SIZE; 
    if (inetProtocol&(STREAM|STREAM6)) port = gtStrmPort; 
#   endif
    if (connectTCP(c,port,inetProtocol) == 0)
    { /* success */
      msglog(0,"new %s connection to %.256s",lnkp,lnkn);
      if ((bckt=getBucket(c->tcpSck,buflen,clrlen,inetProtocol,c->tineProtocol,CLN_REQ_)) != NULL)
      { /* got a tcp bucket */
        bckt->idx = c->fecIdx;
        /* FecTbl[c->fecIdx].inetProtocol = inetProtocol; <- don't force all connections to FEC to use TCP or STREAM ! */
        c->inetProtocol = inetProtocol;
#       ifdef MULTITHREADED
        if (soioctl(c->tcpSck,FIONBIO,(char *)&mode)) 
          msglog(0,"could not reset %s socket to blocking mode",lnkp);
        if (inetProtocol&(STREAM|STREAM6)) recvStreamCreateThread(c->tcpSck,SRV_RSP_);
#       endif
      }
      else 
      { /* oops! problems -> close the socket */
        msglog(0,"could not get TCP bucket for %.256s",lnkn);
        CloseConnectionSocket(c->tcpSck,inetProtocol,TRUE);
        c->tcpSck = 0;
      }
    }
  }
  RestoreSystemMutex(hLinkLstMutex);
  return c->tcpSck;
}
void checkSubscriptions(int isENS)
{
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) == 0)
  {
    while (prepSubRequest(isENS) >= 0);
    ReleaseSystemMutex(hLinkTblMutex);
  }
  else
  {
    if (tineDebug > 1) dbglog("checkSubscriptions: couldn't get link table mutex within %d msec",gSystemTick);
    if (!gClientIsClosing) ReturnToClientCycle++;
  }
}
char gDoocsUserName[USERNAME_SIZE+8] = "";
char *checkUserName(char *usr,char *eqm)
{
  char *pdu;
  static int done=FALSE;
  if (gIsCentralAlarmServer) return usr;
  if (!done)
  {
    done = TRUE;
    pdu = GetDoocsUserName();
    if (strlen(pdu) == 0 && (pdu=getenv("DOOCS_USER")) != NULL)
    {
      SetDoocsUserName(pdu);
    }
  }
  if (!strcmp(eqm,DOOCSEQM_TAG) && strlen(gDoocsUserName) > 0)
  {
    return gDoocsUserName;
  }
  return usr;
}
TINE_EXPORT int SetSubscriptionRenewalThreshold(int linkId,int thresholdInPercent)
{
  ConTblEntry *c=NULL;
  int t = 0, rm = 0;
  if (linkId < 0 || linkId >= nConnectionTableEntries) return invalid_index;
  if ((c=conTbl[linkId]) == NULL) return invalid_link;
  if (thresholdInPercent < 10 || thresholdInPercent > 90) return out_of_range;
  rm = getEffectiveRenewalMultiplier(c->renewalMultiplier,c->pollingInterval);
  if (c->tineProtocol < 6 && rm > 4) rm = 4; /* legacy server */
  t = (CTR_RENEWAL*rm)*thresholdInPercent/100;
  if (t < CTR_ACK_REQUEST) t = CTR_ACK_REQUEST;
  c->renewalThreshold = t;
  feclog("renewal threshold for %.192s set to %d percent (%d remaining transfers)",c->key,thresholdInPercent,t);
  return 0;
}
TINE_EXPORT int GetSubscriptionRenewalThreshold(int linkId,int *thresholdInPercent)
{
  ConTblEntry *c=NULL;
  if (linkId < 0 || linkId >= nConnectionTableEntries) return invalid_index;
  if ((c=conTbl[linkId]) == NULL) return invalid_link;
  if (thresholdInPercent == NULL) return argument_list_error;
  *thresholdInPercent = c->renewalThreshold == 0 ? (CTR_ACK_REQUEST * 100)/ CTR_RENEWAL : c->renewalThreshold;
  return 0;
}
static int getSubsRenewalCutoff(ConTblEntry *c)
{
  int rm = getEffectiveRenewalMultiplier(c->renewalMultiplier,c->pollingInterval);
  if (c->renewalThreshold > 0) return c->renewalThreshold;
  if (c->tineProtocol < 6 && rm > 4) rm = 4; /* legacy server */
  return CTR_ACK_REQUEST * rm;
}
int prepSubRequest(int isENS)
{ /* has hLinkTblMutex */
  BYTE data[REQ_BUFFER_SIZE];
  short fecIdx = -1, i, fid, nreq = 0;
  UINT32 dsize, dpadding, extsize, pktlen = 0, stepsize = 0, totalsize = 0;
  short prot=0, bmode;
  static short scanStart = 0; /* scan start thru the connection table */
  static UINT16 blkid = 0; /* tag the input data blocks */
  ConTblEntry *c=NULL;
  struct timeval tv_now;
  char tin[TAG_NAME_SIZE+4],tout[TAG_NAME_SIZE+4];
  char *usr=gUserName; /* default user to appear in the request header */
  int cc,p,inetProtocol=0,fecIsLocal=FALSE,hasXferMutex=FALSE,renewalCutoff;
  SOCKET sck=0;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  TCPBCKT *bckt;
  char ipcpipe[64];
# endif

  gettimeofday(&tv_now, NULL);
  for (i=scanStart; i<nConnectionTableEntries; i++)
  {
    if (i > 0 && isENS) break;
    if ((c=conTbl[i]) == NULL) continue;
    if (c->tineProtocol < 5) continue;
    if (c->boundToId > 0) continue;
    if (c->heartbeat < HBT_SUB_PENDING) continue;
    if (c->linkStatus == link_blacklisted) continue;
    if (nreq == 0)
    { /* fecIdx not yet assigned */
      scanStart = i + 1; /* start at the next one in the list next time */ 
      prot = c->tineProtocol;
      totalsize = PAYLOADHDR_SIZE(prot);
      fecIdx = c->fecIdx;
      pktlen = FecTbl[fecIdx].inetProtocol == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE;
      memset(data,0,pktlen);
      stepsize = SUBREQ_STEPSIZE(prot);
      if (FecTbl[fecIdx].inetProtocol == PIPE ||
          FecTbl[fecIdx].inetProtocol == MMF)
        fecIsLocal = TRUE;
      if (FecTbl[fecIdx].inetProtocol == 0 || strlen(SrvTbl[c->srvIdx].EqmName) == 0)
      { /* still don't have an address ? was flushed ? */
        if (c->tocounter > 0)
        { /* and checkConnections() is busy resolving somebody else's address ... */
          if (tineDebug > 0) dbglog("last minute address re-acquisition for %.16s",FecTbl[fecIdx].fecName);
          if ((fid=locateExport(c->expName,c->ctxName,c->devName,&c->srvIdx)) < 0)
          { /* address is no longer valid ? */
#           ifndef SMALL_TINE_LIB
            appendBlackLnkLst(c->ctxName,c->expName,c->devName,c->prpName,address_unknown);
#           endif
            ReleaseSystemMutex(hLinkTblMutex); /* never know what someone might do inside a callback ... */
            NotifyClient(i, -fid);
            if (WaitForSystemMutex(hLinkTblMutex, gSystemTick) != 0) ccerr(mutex_error);
          }
          else
          { /* re-assign the fec table index */
            c->fecIdx = fid;
          }
        }
        continue;
      }
    }
    else 
    { /* fecIdx has been assigned: must match or continue */
      if (fecIdx != c->fecIdx) 
      { /* not a match */
        continue;
      }
    }
    if (c->mode & CM_CONNECT && !fecIsLocal)
    {
      if (inetProtocol & (UDP|UDP6|IPX)) continue;
      if (c->tcpSck == 0 && c->timeLastSync == (UINT32)tv_now.tv_sec) continue; /* let a second go by between attempts */
      p = (gAfInet == AF_INET6) ? TCP6 : TCP;
      sck = getTcpConTblSocket(c,p);
      inetProtocol = p;
    }
#   if defined(MULTITHREADED)
    else if (c->mode & CM_STREAM && !fecIsLocal)
    {
      if (inetProtocol & (UDP|UDP6|IPX)) continue;
      p = (gAfInet == AF_INET6) ? STREAM6 : STREAM;
      sck = getTcpConTblSocket(c,p);
      inetProtocol = p;
    }
#   endif
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    else if (FecTbl[fecIdx].inetProtocol == PIPE) 
    {
      sck = FecTbl[fecIdx].lclAdr.pipe.sck;
      if (sck == 0)
      { /* pipe socket not yet established ... */
        sprintf(ipcpipe,"%s%.16s-server.ipc",IPC_PATH,FecTbl[fecIdx].fecName);
        if ((sck=pipeConnect(ipcpipe)) <= 0) 
        { /* couldn't establish a connection ! */
          perror("IPC Socket");
          sck = 0;
        }
        else
        { /* looks good so far ... */
          addClnPipeSck(sck,fecIdx);
          FecTbl[fecIdx].lclAdr.pipe.sck = sck;
          if (tineDebug) dbglog("FEC %.16s: attach pipe",FecTbl[fecIdx].fecName);
          if ((bckt=getBucket(sck,TCP_BUFFER_SIZE,0,PIPE,c->tineProtocol,CLN_REQ_)) != NULL)
          { /* got a bucket */
            bckt->idx = fecIdx;
          }
          else 
          { /* something went wrong !*/
            msglog(code_failure,"could not acquire ipc pipe bucket to %.16s",FecTbl[fecIdx].fecName);
            ClosePipeSocket(sck);
            sck = 0;
          }
        }
      }
      if (sck == 0)
      { /* still no pipe socket -> resort to loopback for this one */
        msglog(0,"pipe to %.16s not available: use UDP loopback",FecTbl[fecIdx].fecName);
        inetProtocol = FecTbl[fecIdx].inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP;
        fecIsLocal = FALSE;
      }
      else
      { /* looks good ... */
        inetProtocol = PIPE;
      }
    }
#   endif
    else
    {
      if (inetProtocol&(TCP|TCP6|STREAM|STREAM6|PIPE)) continue;
      sck = 0;  /* let sendSubRequest figure it out */
      inetProtocol = FecTbl[fecIdx].inetProtocol;
#     if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
      if (inetProtocol == MMF && gUseMmfJumbos && c->sizeBytesOut > gShmBufferSize)
      {
        dbglog("requested output (%d bytes) > shared memory block : use UDP loopback !",c->sizeBytesOut);
        inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP;
      }
#     endif
    }
    if (c->extStringspace != NULL && c->extStringsize > 0 && c->nextDataPtr == 0)
    { /* initial packet carries the extended string space (if any) */
      extsize = c->extStringsize;
      memcpy(&data[totalsize + stepsize],c->extStringspace,extsize);
    }
    else
    {
      extsize = 0;
    }
    if (c->sizeIn && c->dataIn)
    { /* input data size to be sent to the server */
      short fmt = c->nextDataPtr == 0 ? c->formatIn : GetFormatDataType(c->formatIn);
      dsize = (short)c->nextDataSize;
      if ((totalsize + stepsize + dsize) > pktlen) break;
      dpadding = dsize % 2; /* make sure of even boundaries */
      prepOutgoingData((BYTE *)&data[totalsize+stepsize+extsize],(BYTE *)&c->dataIn[(int)c->nextDataPtr],(UINT32)dsize,fmt);
    }
    else dpadding = dsize = 0;
    if (!hasXferMutex) 
    {
      if (WaitForSystemMutex(hXferMutex, gSystemTick) != 0)
      {
        fecIdx = -1;
        goto err;
      }
      hasXferMutex = TRUE;
    }
    if (tineDebug > 1) dbglog("prepare to request %.192s (hbt %d, mode %s)",c->key,c->heartbeat,modetoa[BASEMODE(c->mode)]);
    prepOutgoingSubRequestHdr(prot,c,SrvTbl[c->srvIdx].EqmName,dsize+dpadding,i,&data[totalsize]);
    usr = checkUserName(usr,SrvTbl[c->srvIdx].EqmName);
    /* this fecIdx is going out, so start counting ... */
    nreq++;
    totalsize += stepsize + extsize + dsize + dpadding;
    if (tineDebug)
    {
      memset(tin,0,TAG_NAME_SIZE+4);
      if (c->strTagIn[0] != 0) strncpy(tin,c->strTagIn,TAG_NAME_SIZE);
      memset(tout,0,TAG_NAME_SIZE+4);
      if (c->strTagOut[0] != 0) strncpy(tout,c->strTagOut,TAG_NAME_SIZE);
      dbglog("subs %.192s (%d %d %.16s %d %d %.16s %d of %d) [%.16s] (%d ms): %s",
              c->key,(int)c->sizeOut,c->formatOut,tout,(int)c->sizeIn,c->formatIn,tin,
              c->blknumIn,c->numblksIn,
              FecTbl[fecIdx].fecName,(int)c->pollingInterval,modetoa[BASEMODE(c->mode)]);
    }
    if (tineDebug > 3 && c->dataIn) DataDump("data sent: ",c->dataIn,(int)c->sizeIn,c->formatIn,LTOR);
    ReturnToClientCycle = 0;
    if (c->blknumIn == c->numblksIn)
    {
      int bmode = BASEMODE(c->mode);
      if (c->renewalMultiplier < 1) c->renewalMultiplier = 1;
      renewalCutoff = getSubsRenewalCutoff(c);
      c->blknumIn = 1;
      c->nextDataPtr = 0;
      c->blkidIn = (++blkid)%1000;
      c->bytesRemaining = c->bytesTotal;
      c->heartbeat = assignHeartbeat(c);
      if (c->counter < (renewalCutoff/2) || 
          c->counter == renewalCutoff || 
          bmode == CM_DATACHANGE || bmode == CM_EVENT) c->counter = -1;
      c->timeLastSync = (UINT32)tv_now.tv_sec;
      c->timeMsecLastSync = (UINT32)tv_now.tv_usec/1000;
      c->hasAcknowledged = FALSE;
    }
    else if (c->dataIn && c->numblksIn > c->blknumIn)
    {
      ReturnToClientCycle = c->numblksIn - c->blknumIn;
      c->blknumIn++;
      c->nextDataPtr += c->nextDataSize;
      c->bytesRemaining = c->bytesTotal - c->nextDataPtr;
    }
    c->nextDataSize = MIN(c->datamtu,c->bytesRemaining);
    if (isTraceLink(i)) 
    {
      char msg[64];
      sprintf(msg,"send %.16s subscription request",modetoa[BASEMODE(c->mode)]);
      traceLinkOutput(c, "prepSubRequest", msg);
    }
    if ((totalsize + stepsize) > pktlen) break;  /* no more packing possible */
  } /* end of scan loop */
  if (fecIdx >= 0) 
  { /* requests to send out have been found ! */
    prepOutgoingRequestHdr(prot, totalsize, nreq, usr, gUserType, data); /* fill in the request header */
    if (!strncmp(FecTbl[fecIdx].strAdr,gtNetSrvAddr,ADDR_SIZE_IPV6)) sck = NETSRV_SOCKET;
    if (inetProtocol == 0)
    { /* can happen due to a TCP link having been CANCELed */
      fecIdx = -1;  /* get out of the loop ! */
    }
    else if ((cc=sendSubRequest(fecIdx,sck,inetProtocol,data,totalsize)) != 0)
    { /* send unsuccessful ! */
      int isConnected = inetProtocol & (TCP|TCP6|STREAM|STREAM6|PIPE);
      for (i=0; i<nConnectionTableEntries; i++) 
      { /* try, try, again ... */
        if ((c=conTbl[i]) == NULL) continue;
        if (c->fecIdx != fecIdx) continue;
        if (c->inetProtocol != inetProtocol) continue;
        if (tineDebug) dbglog("resubmit %.192s due to error <%d>",c->key,cc);
        if (cc != not_implemented && inetProtocol != PIPE && inetProtocol != MMF && BASEMODE(c->mode) > CM_SINGLE)
        { /* set it back */
          c->heartbeat = HBT_SUB_PENDING;  
        }
        if (isConnected && getTcpClnSck(sck, -1, -1) == NULL)
        { /* ensure this */
          c->tcpSck = 0;
        }
        bmode = BASEMODE(c->mode);
        if (bmode == CM_CANCEL) c->mode = CM_CANCEL; /* strip off e.g. CM_NETWORK */
        if ((sck == 0 || bmode == CM_CANCEL) && isConnected)
        { /* in this case, don't try again ... */
          c->suppressCallback = FALSE;
          c->needsNotification = TRUE;
          c->heartbeat = HBT_MINIMUM;
          /* c->mode = CM_CANCEL; */
          c->linkStatus = cc;
          lastLnkErr = cc;
          strncpy(gLastStatusString,cc2str(cc),STATUS_SHORTSIZE);
          strncpy(c->lastStatusString,cc2str(cc),STATUS_SHORTSIZE);
        }
      }
      fecIdx = -1;  /* get out of the loop ! */
    }
    else
    { /* send successful (hurray!) */
      for (i=0; i<nConnectionTableEntries; i++) 
      {
        if ((c=conTbl[i]) == NULL) continue;
        if (BASEMODE(c->mode) == CM_CANCEL)
        {
          if (inetProtocol & (TCP|TCP6|STREAM|STREAM6|PIPE)) c->heartbeat = HBT_NONE;
        }
      }
    }
  }
  else
  { /* nothing to do: can run thru the list from the beginning next time */
    scanStart = 0;
  }
err:
  if (hasXferMutex) ReleaseSystemMutex(hXferMutex);
  return fecIdx;
}
int sendSubRequest(int fecIdx,SOCKET sck,int inetProtocol,BYTE *data,UINT32 sizeInBytes)
{
  int cc, family = getInetFamily(inetProtocol);
  UINT16 poff, p;
  SCKADR sckadr;
  BYTE *badr;
  static ClnHdr cln;
# if defined(IPXNW) || defined(IPXBSD)
  static BYTE ipxSck[2];
# endif

  memset(&cln,0,sizeof(ClnHdr));
  strncpy(cln.usr,FecTbl[fecIdx].fecName,USERNAME_SIZE);
  cln.inetProtocol = inetProtocol;
  poff = FecTbl[fecIdx].portOffset;
  memset(&sckadr,0,sizeof(sckadr));
  switch (inetProtocol)
  {
    case IPX:
      if (!ipxLoaded) return ipx_not_supported;
#     if defined(IPXNW) || defined(IPXBSD)
      ipxSck[0] = IPX_SRVSCK + (BYTE)poff;
      ipxSck[1] = IPX_SRVSCK + (BYTE)poff;
      memcpy(&cln.adr.ipx.IpxNet,FecTbl[fecIdx].netAdr.ipx.IpxNet,16);
      memcpy(&cln.ep.ipx.sckRmt,ipxSck,2);
      cln.ep.ipx.sckLcl = ipxClnSck;
#     endif
      break;
    case UDP6:
    case UDP:
      if (!ipLoaded) return tcp_not_supported;
      if (sck == NETSRV_SOCKET)
      {
        p = (UINT16)gtNetSrvPort; 
        badr = (BYTE *)&tNetSrvAddr;
      }
      else
      {
        p = (UINT16)(gtUDPPort + poff); 
        badr = (BYTE *)&FecTbl[fecIdx].netAdr;
      }
      if (sckadr_in(&sckadr,badr,p,family) != 0) return address_unknown;
      memcpy(&cln.adr.ip,&sckadr,sizeof(sckadr));
      cln.ep.sck = udpClnSck;
      break;
    case TCP6:
    case TCP:
      if (!ipLoaded) return tcp_not_supported;
      p = (UINT16)(gtTCPPort + poff);
      badr = (BYTE *)&FecTbl[fecIdx].netAdr;
      if (sckadr_in(&sckadr,badr,p,family) != 0) return address_unknown;
      memcpy(&cln.adr.ip,&sckadr,sizeof(sckadr));
      cln.ep.sck = sck;
      break;
#   ifdef MULTITHREADED
    case STREAM6:
    case STREAM:
      if (!ipLoaded) return tcp_not_supported;
      p = (UINT16)(gtStrmPort + poff);
      badr = (BYTE *)&FecTbl[fecIdx].netAdr;
      if (sckadr_in(&sckadr,badr,p,family) != 0) return address_unknown;
      memcpy(&cln.adr.ip,&sckadr,sizeof(sckadr));
      cln.ep.sck = sck;
      break;
#   endif
#   if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
    case MMF:
      if (!ipLoaded) return tcp_not_supported;
      cln.ep.wndMmf = FecTbl[fecIdx].lclAdr.wndMmf;
      break;
#   endif
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    case PIPE:
      cln.ep.pipe.sck = sck;
      memcpy(&cln.adr.ip,getmyipaddr(),SCKADR_LEN);
      break;
#   endif
    case   0:
    default:
      if (tineDebug > 1) dbglog("FEC %.16s: acquire protocol",FecTbl[fecIdx].fecName);
      return invalid_transport;
  }
  if (inetProtocol == STREAM) msglog(0, "Sending STREAM Subscription to %.16s", cln.usr);
  if ((cc=sendToPeer(&cln,data,sizeInBytes,CLN_REQ_,FALSE)) != 0)
  { /* problems ... */
    char addr[64];
    getPeerAddrStr(addr,&cln);
    msglog(cc,"Send Subscription to %s (%s) : %.32s",addr,getProtocolAsString(inetProtocol),cc2str(cc));
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    if (inetProtocol == PIPE) 
    {
      ClosePipeSocket(sck);
      if (++(FecTbl[fecIdx].lclAdr.pipe.err) > 10)
      {
        gUseLoopback = TRUE;
        msglog(cc,"FEC %.16s : > 10 pipe errors on send, set use loopback flag to TRUE",FecTbl[fecIdx].fecName);
        FecTbl[fecIdx].lclAdr.pipe.err = 0;
      }
    }
#   endif
    return cc;
  }
  if (tineDebug > 1) dbglog("FEC %.16s: sent subscription request",FecTbl[fecIdx].fecName);
  return 0;
}
void getPeerAddrStr(char *addr,ClnHdr *cln)
{
  switch (cln->inetProtocol)
  {
    case UDP:
    case TCP:
    case STREAM:
    case UDP6:
    case TCP6:
    case STREAM6:
      inetStrAdr(&cln->adr.ip,addr,TRUE);
      break;
    case IPX:
#     if defined(IPXNW) || defined(IPXBSD)
      sprintf(addr,"%.30s",ipxnet_ntoa(cln->adr.ipx.IpxNode));
#     else
      sprintf(addr,"ipx addr illegal");
#     endif
      break;
    case MMF:
#     if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
#     ifdef OS_64BIT
      _snprintf(addr,30,"hWnd = 0x%I64x",(UINT64)cln->ep.wndMmf);
#     else
      _snprintf(addr, 30, "hWnd = 0x%x", (int)cln->ep.wndMmf);
#     endif
#     else
      sprintf(addr,"mmf addr illegal");
#     endif
      break;
    case PIPE:
#     if defined(UNIX) && !defined(NO_TINE_PIPE)
      sprintf(addr,"%.16s-server.ipc",cln->usr);
#     else
      sprintf(addr,"pipe addr illegal");
      break;
#     endif
    default:
      sprintf(addr,"address unknown");
      break;
  }
  return;
}
int getIncomingFecIndex(void *param,int inetprot)
{
  int i,idx;
  UINT16 poff;

  if (param == NULL) return -1;

  for (i=0; i<nConnectionTableEntries; i++) if (conTbl[i] != NULL && conTbl[i]->inetProtocol == inetprot)
  { /* scan thru my connection table until the incoming address matches */
    idx = conTbl[i]->fecIdx;
    poff = FecTbl[idx].portOffset;
    switch (inetprot)
    {
      case UDP:
      case TCP:
        {
          struct sockaddr_in *sa4 = (struct sockaddr_in *)param;
          if (inetprot == UDP && sa4->sin_port != htons((UINT16)(gtUDPPort + poff))) continue;
          if (!memcmp(&sa4->sin_addr.s_addr,&FecTbl[idx].netAdr.ipv4,4)) return idx;
          break;
        }
      case UDP6:
      case TCP6:
        {
          struct sockaddr_in6 *sa6 = (struct sockaddr_in6 *)param;
          if (inetprot == UDP6 && sa6->sin6_port != htons((UINT16)(gtUDPPort + poff))) continue;
          if (!memcmp(&sa6->sin6_addr.s6_addr,&FecTbl[idx].netAdr.ipv6,16)) return idx;
          break;
        }
#     if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
      case MMF:
        {
          HWND hwnd = *(HWND *)param;
          if (hwnd == FecTbl[idx].lclAdr.wndMmf) return idx;
          break;
        }
#     endif
      case IPX:
          idx = conTbl[i]->fecIdx;
          if (!memcmp(&FecTbl[idx].netAdr.ipx,&gInAdr.ipx,10)) return idx;
          break;
    }
  }
  return -1;
}
#if !defined(SMALL_TINE_LIB)
BLACKLISTED_LINK *gBlackLnkTbl[LNKTBL_HASH_SIZE];
int gBlackLnkTblSize = 0;
BLACKLISTED_LINK *getBlackLnkLstItem(char *ctx, char *srv, char *dev, char *prp)
{
  BLACKLISTED_LINK *bl;
  char key[FULL_NAME_MAXSIZE+8];
  int idx;

  if (gBlackLnkTblSize == 0) return NULL; /* no need to look */
  makeLinkKey(key,ctx,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);

  for (bl=gBlackLnkTbl[idx]; bl != NULL; bl = bl->nxt)
  {
    if (strnicmp(bl->key,key,FULL_NAME_MAXSIZE+8)) continue;
    break;
  }
  return bl; 
}
void appendBlackLnkLst(char *ctx, char *srv, char *dev, char *prp,int status)
{
  int idx;
  char key[FULL_NAME_MAXSIZE+8];
  BLACKLISTED_LINK *bl=getBlackLnkLstItem(ctx,srv,dev,prp);
  if (bl != NULL) return;
  if ((bl=(BLACKLISTED_LINK *)SystemCalloc(1,sizeof(BLACKLISTED_LINK))) == NULL)
  {
    msglog(out_of_local_memory,"appendBlackLnkLst: out of memory");
    return;
  }
  makeLinkKey(key,ctx,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  strncpy(bl->key,key,FULL_NAME_MAXSIZE+8);
  bl->status = status;
  bl->nxt = gBlackLnkTbl[idx];
  gBlackLnkTbl[idx] = bl;
  gBlackLnkTblSize++;
  msglog(0,"append %.192s to the link black list",key);
  if (gIsRunningAsServer) feclog("append %.192s to the link black list",key);
}
void flushBlackLnkLst(void)
{
  BLACKLISTED_LINK *bl, *bln;
  int i;

  if (gBlackLnkTblSize == 0) return;
  if (tineDebug) dbglog("flush the link black list (%d entries)",gBlackLnkTblSize);
  for (i=0; i<LNKTBL_HASH_SIZE; i++)
  {
    for (bl=gBlackLnkTbl[i]; bl != NULL; bl = bln)
    {
      bln = bl->nxt;
      SystemFree(bl);
    }
    gBlackLnkTbl[i] = NULL;
  }
  gBlackLnkTblSize = 0;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    conTbl[i]->blcounter = 0; 
    if (conTbl[i]->linkStatus == link_blacklisted) conTbl[i]->linkStatus = link_timeout;
  }
}
ReLnkTblEntry *gReLnkTbl[LNKTBL_HASH_SIZE];
int gReLnkTblSize = 0;
ReLnkTblEntry *getReLnkTblEntry(ConTblEntry *c)
{
  return c->rlkEntry;
}
ReLnkTblEntry *getReLnkLstItem(char *ctx,char *srv,char *dev,char *prp,int siz,int fmt)
{ /* siz and fmt refer to original requested siz and fmt */
  ReLnkTblEntry *r;
  char key[FULL_NAME_MAXSIZE+8];
  int idx;

  if (gReLnkTblSize == 0) return NULL; /* no need to look */
  makeLinkKey(key,ctx,srv,dev,prp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  for (r=gReLnkTbl[idx]; r != NULL; r = r->nxt)
  {
    if (strnicmp(r->key,key,FULL_NAME_MAXSIZE+8)) continue;
    if (r->sts == invalid_datarequest)
    { /* then I don't want an exact match of the original calling conditions */
      if (r->siz == siz && r->fmt == fmt) continue;
      if (r->lnkSiz == siz && r->lnkFmt == fmt) break;
    }
    if (r->siz != siz) continue;
    break;
  }
  return r; 
}
void appendReLnkCb(ConTblEntry *c)
{
  ReLnkCbEntry *cbLst;
  ReLnkTblEntry *r;
  if (c == NULL) return;
  if (c->cbfcn == _relinkXrefCb) return;
  if ((r=c->rlkEntry) == NULL) return;
  for (cbLst=r->cbLst; cbLst != NULL; cbLst=cbLst->nxt)
  {
    if (cbLst->cbfcn != c->cbfcn) continue;
    if (cbLst->cbId != c->cbId) continue;
    if (cbLst->cbRef != c->cbRef) continue;
    if (cbLst->hasReference != c->hasReference) continue;
    cbLst->lnkId = c->linkId;
    return; /* already in the list */
  }
  cbLst = (ReLnkCbEntry *)SystemCalloc(1,sizeof(ReLnkCbEntry));
  if (cbLst == NULL) return;
  cbLst->cbfcn = c->cbfcn;
  cbLst->cbId = c->cbId;
  cbLst->cbRef = c->cbRef;
  cbLst->hasReference = c->hasReference;
  cbLst->lnkId = c->linkId;
  cbLst->nxt = r->cbLst;
  r->cbLst = cbLst;
  return;
}
ConTblEntry *getReLnkParent(ConTblEntry *c)
{
  ReLnkCbEntry *cb;
  ReLnkTblEntry *r;
  ConTblEntry *cr;
  int i,dsiz;

  for (i=0; i<LNKTBL_HASH_SIZE; i++)
  {
    for (r=gReLnkTbl[i]; r != NULL; r=r->nxt)
    {
      for (cb=r->cbLst; cb != NULL; cb=cb->nxt)
      {
        if (c->linkId == cb->lnkId) continue;
        if ((cr=conTbl[cb->lnkId]) == NULL) continue;
        if (cr->boundToId) continue; /* not a parent */
        if (strnicmp(cr->ctxName,c->ctxName,CONTEXT_NAME_SIZE)) continue;
        if (strnicmp(cr->expName,c->expName,EXPORT_NAME_SIZE)) continue;
        if (strnicmp(cr->devName,c->devName,DEVICE_NAME_SIZE)) continue;
        if (strnicmp(cr->prpName,c->prpName,PROPERTY_NAME_SIZE)) continue;
        if (strnicmp(cr->strTagOut,c->strTagOut,TAG_NAME_SIZE)) continue;
        if (cr->sizeOut != c->sizeOut) continue;
        if (cr->formatOut!= c->formatOut) continue;
        if (cr->sizeIn != c->sizeIn) continue;
        if (c->sizeIn > 0)
        { /* input data ! */
          if (cr->formatIn != c->formatIn) continue;
          dsiz = c->sizeIn*fmtsizeof(c->formatIn);
          if (memcmp(cr->dataIn,c->dataIn,(size_t)dsiz)) continue;
        }
        return cr;
      }
    }
  }
  return NULL; 
}
int dumpRelinkTable(void)
{
  ReLnkTblEntry *r;
  int i;
  char str[256];
  for (i=0; i<LNKTBL_HASH_SIZE; i++)
  {
    for (r=gReLnkTbl[i]; r != NULL; r=r->nxt)
    {
      if (strlen(r->tag) > 0)
      {
        if (r->sts == has_structure_tag)
        {
          sprintf(str,"field %.16s to structure %.16s",r->field,r->tag);
        }
        else
        {
          sprintf(str,"field %.16s to bitfield %.16s",r->field,r->tag);
        }
      }
      else
      {
        sprintf(str,"%d %s(s) to %d %s(s)",r->lnkSiz,itof(r->lnkFmt),r->siz,itof(r->fmt));
      }
      ttyoutput("%.96s relinks %s",r->key,str);
    }
  }
  return 0;
}
/** \internal
 *  \brief called in a direct response to 
 *  'invalid_datarequest','has_structure_tag', 'has_bitfield_tag' 
 *  or pre-emptively when a known bitfield is beiing accessed.
 */
void appendReLnkLst(ConTblEntry *c,char *tag,int fmt,int siz,int sts)
{ /* relink to fmt, siz, and tag;  'c' remains unchanged here */
  int i,p,idx,chkDeviceDecorations=FALSE;
  char key[FULL_NAME_MAXSIZE+8],prp[PROPERTY_NAME_SIZE+4],dev[DEVICE_NAME_SIZE+4], *fld=NULL;
  ReLnkTblEntry *r=getReLnkLstItem(c->ctxName,c->expName,c->devName,c->prpName,c->sizeOut,c->formatOut);
  if (r != NULL) 
  { /* already in the list !*/
    return;
  }
  if ((r=(ReLnkTblEntry *)SystemCalloc(1,sizeof(ReLnkTblEntry))) == NULL)
  {
    msglog(out_of_local_memory,"appendReLnkLst: out of memory");
    return;
  }
  r->sts = sts;
  makeLinkKey(key,c->ctxName,c->expName,c->devName,c->prpName);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  strncpy(r->key,key,FULL_NAME_MAXSIZE+8);
  strncpy(r->tag,tag,TAG_NAME_SIZE);
  r->lnkFmt = c->formatOut;
  r->lnkSiz = c->sizeOut;
  r->fmt = fmt;
  r->siz = siz;
  if (sts != invalid_datarequest)
  { /* then either a struct field or a bitfield field */
    if (isBitfieldFormat(fmt)) 
    {
      r->dat = (BYTE *)c->dataPtr;
      chkDeviceDecorations = TRUE;
    }
    strncpy(prp,c->prpName,PROPERTY_NAME_SIZE);
    p = (int)strlen(prp);
    p = MIN(p,PROPERTY_NAME_SIZE) - 1;
    for (i=p; i>0; i--) if (prp[i] == '.') break;
    if (i > 0 && i < p)
    { /* is the property field-decorated ? */
      prp[i] = 0;
      fld = &prp[i+1];
      strncpy(r->field,fld,FIELD_NAME_SIZE);
    }
    if (fld == NULL  && chkDeviceDecorations)
    { /* is the device field-decorated  (CDI does this) ? */
      strncpy(dev,c->devName,DEVICE_NAME_SIZE);
      p = (int)strlen(dev);
      p = MIN(p,DEVICE_NAME_SIZE) - 1;
      for (i=p; i>0; i--) if (dev[i] == '.') break;
      if (i > 0 && i < p)
      { 
        dev[i] = 0;
        fld = &dev[i+1];
        strncpy(r->field,fld,FIELD_NAME_SIZE);
      }
    }
  }
  c->rlkEntry = r;
  appendReLnkCb(c);
  c->cbfcn = (TCBFCNP)_relinkXrefCb;
  c->cbId = c->linkId;
  c->hasReference = FALSE;
  r->nxt = gReLnkTbl[idx];
  gReLnkTbl[idx] = r;
  gReLnkTblSize++;
  msglog(0,"%.192s will be re-linked (field %.16s)",r->key,r->field);
}
/** \internal
 *  \brief called in getConTblEntry.  
 *   Preempts a known case of 'invalid_datarequest','has_structure_tag','has_bitfield_tag'
 */
ReLnkTblEntry *relinkLink(char *ctx,char *srv,char *prp,char *dev,char **pkey,DTYPE *dout)
{ /* fix the DTYPE object if required */
  ReLnkTblEntry *r;
  char key[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+4],fld[FIELD_NAME_SIZE];
  int siz = 0;
  int fmt = CF_NULL;
  if (dout == NULL) return NULL;
  siz = dout->dArrayLength;
  fmt = dout->dFormat;
  if ((r=getReLnkLstItem(ctx,srv,dev,prp,siz,fmt)) == NULL) 
  {
    *pkey = NULL;
    return NULL;
  }
  if (r->key < (char *)0x100)
  { /* this sanity check is necessary because of what ? */
    *pkey = NULL;
    return NULL;
  }
  msglog(0,"re-link %.192s (%d %s %.16s -> %d %s %.16s) from re-link table",
    r->key,dout->dArrayLength,itof(dout->dFormat),dout->dTag,r->siz,itof(r->fmt),r->tag);
  dout->dFormat = r->fmt;
  strncpy(dout->dTag,r->tag,TAG_NAME_SIZE);
  if (r->fmt == CF_STRUCT)
  { /* a relinked struct ... */
    isStructProperty(r->tag,prp,fld); /* peel off the field from the property */
    r->dat = dout->data.bptr; /* record the original bound data pointer */
    dout->data.vptr = NULL;   /* signal an alloc within getConTblEntry() */
  }
  else if (r->sts == has_bitfield_tag)
  { /* peel the field from the property name */
    sprintf(key,"/%.32s/%.32s",ctx,srv);
    if (!isBitfieldProperty(key,r->tag,dev,fld)) isBitfieldProperty(key,r->tag,prp,fld);
  }
  else if (r->sts == invalid_datarequest)
  {
    if (r->siz == (int)dout->dArrayLength && r->fmt == dout->dFormat)
    { /* no reason to re-link ! */
      *pkey = NULL;
      return NULL;
    }
    r->dat = dout->data.bptr; /* record the original bound data pointer */
    dout->dArrayLength = r->siz;
    dout->data.vptr = NULL;   /* signal an alloc within getConTblEntry() */
  }
  *pkey = r->key;
  return r;
}
/** \internal
 *  \brief called (with a delay) due to response to one of
 *  'invalid_datarequest', 'has_structure_tag', 'has_bitfield_tag'
 */
int reAttachLink(int i,int isStructField)
{ 
  ReLnkTblEntry *r=NULL;
  ConTblEntry *c=conTbl[i],*cp=NULL;
  struct structStruct *s;
  char devName[FULL_NAME_MAXSIZE],fld[FIELD_NAME_SIZE];
  int cc = 0,nb,fixDefault,fldsiz,fldfmt;
  UINT32 doutsize, datamtu;

  fld[0] = 0;
  /* 'c' still contains original contents here */
  if ((r=getReLnkLstItem(c->ctxName,c->expName,c->devName,c->prpName,c->sizeOut,c->formatOut)) == NULL)
    ccerr(invalid_link);
  fixDefault = r->lnkFmt == CF_DEFAULT ? TRUE : FALSE;
  if (fixDefault) c->sizeOut = r->siz = 1;
  sprintf(devName,"/%.32s/%.32s/%.64s",c->ctxName,c->expName,c->devName);
  if (isStructField) 
  {
    acquireAndRegisterStruct(devName,r->tag,c->sizeOut);
    if (!isStructProperty(r->tag,c->prpName,fld)) ccerr(invalid_field);
    if ((s=findstructEx(r->tag,c->ctxName,c->expName)) == NULL) ccerr(invalid_structure_tag);
    if ((fldsiz=getFieldSize(s,fld)) > 1)
    { /* the structure field itself is an array */
      c->sizeOut = 1;
    }
    fldfmt = getFieldFormat(s,fld);
    c->sizeOut *= s->rawlen;
    if (fixDefault) r->lnkFmt = fldfmt;
    if (r->dat == NULL) 
    {
      r->dat = (BYTE *)c->dataPtr;
      if ((c->dataPtr =(char *)SystemCalloc(1,(size_t)c->sizeOut)) == NULL)
        ccerr(out_of_client_memory);
      if (c->sizeOut > c->allocatedBytes && c->bufOut != NULL)
      { /* reallocate this buffer as well */
        datamtu = PMTU(0,c->inetProtocol,MINIMUM_REQHDR_PROTOCOL); /* worst-case: small */
        c->numblks = (c->sizeOut+STATUS_SIZE)/datamtu + ((c->sizeOut+STATUS_SIZE)%datamtu ? 1 : 0);
        c->numblks *= 3; /* client could be using a small MTU ! */
        if (c->numblks)
        {
          SystemFree(c->bufOut); SystemFree(c->bufOutBlks);
          nb = c->numblks; nb += (8 - nb%8); 
          c->bufOut = (char *)SystemCalloc(1,(size_t)c->sizeOut+STATUS_SIZE);
          c->bufOutBlks = (BYTE *)SystemCalloc(nb,1);
        }
      }
      c->allocatedBytes = c->sizeOut;
      if (s->hasExtSpace) c->sizeOut = s->rawlen;
    }
  }
  else if (r->sts == has_bitfield_tag)
  {
    DBITFIELD *bf;
    sprintf(devName,"/%.32s/%.32s",c->ctxName,c->expName);
    acquireAndRegisterBitfield(devName,r->tag);
    if (!isBitfieldProperty(devName,r->tag,c->devName,fld) && 
        !isBitfieldProperty(devName,r->tag,c->prpName,fld))
      ccerr(invalid_field);
    if ((bf=findBitField(devName,r->tag)) == NULL) ccerr(code_failure);
    r->fmt = bf->fmt; /* set official bitfield format */
    if (fixDefault) r->lnkFmt = GetIntegerFormat((short)r->fmt);
    if (r->dat == NULL) 
    {
      r->dat = (BYTE *)c->dataPtr;
      doutsize = c->sizeOut*fmtsizeof(r->fmt);
      if ((c->dataPtr =(char *)SystemCalloc(1,(size_t)doutsize)) == NULL)
        ccerr(out_of_client_memory);
      c->allocatedBytes = doutsize;
    }
  }
  else if (r->sts == invalid_datarequest)
  {
    if (fixDefault) 
    {
      r->lnkFmt = GetIntegerFormat((short)r->fmt);
      r->lnkSiz = 1;
    }
    c->sizeOut = r->siz;
    if (r->dat == NULL) 
    {
      r->dat = (BYTE *)c->dataPtr;
      doutsize = r->siz*fmtsizeof(r->fmt);
      doutsize += getFormatHeaderSize(r->fmt);
      if ((c->dataPtr=(char *)SystemCalloc(1,(size_t)doutsize)) == NULL)
        ccerr(out_of_client_memory);
      if (doutsize > c->allocatedBytes && c->bufOut != NULL)
      { /* reallocate this buffer as well */
        datamtu = PMTU(0,c->inetProtocol,MINIMUM_REQHDR_PROTOCOL); /* worst-case: small */
        c->numblks = (doutsize+STATUS_SIZE)/datamtu + ((doutsize+STATUS_SIZE)%datamtu ? 1 : 0);
        c->numblks *= 3; /* client could be using a small MTU ! */
        if (c->numblks)
        {
          SystemFree(c->bufOut); SystemFree(c->bufOutBlks);
          nb = c->numblks; nb += (8 - nb%8);
          c->bufOut = (char *)SystemCalloc(1,(size_t)doutsize+STATUS_SIZE);
          c->bufOutBlks = (BYTE *)SystemCalloc(nb,1);
        }
      }
      c->allocatedBytes = doutsize;    
    }
  }
  c->formatOut = r->fmt;
  strncpy(c->strTagOut,r->tag,TAG_NAME_SIZE);
  c->rlkEntry = r;
  c->heartbeat = HBT_SUB_PENDING;
  if ((cp=getReLnkParent(c)) != NULL)
  {
    c->boundToId = cp->linkId;
    addDependentLink(c->linkId);
  }
err:
  return cc;
}
void flushReLnkLst(void)
{
  ReLnkTblEntry *r, *rn;
  ReLnkCbEntry *cb, *cbn;
  int i;

  for (i=0; i<LNKTBL_HASH_SIZE; i++)
  {
    for (r=gReLnkTbl[i]; r != NULL; r = rn)
    {
      rn = r->nxt;
      for (cb=r->cbLst; cb != NULL; cb=cbn)
      {
        cbn = cb->nxt; SystemFree(cb);
      }
      SystemFree(r);
    }
    gReLnkTbl[i] = NULL;
  }
  gReLnkTblSize = 0;
  for (i=1; i<nConnectionTableEntries; i++) 
  { /* don't forget these ... */
    conTbl[i]->rlkEntry = NULL;
  }
}
#endif /* !SMALL_TINE_LIB */
extern INADR fecAddrMcGrp; /* current assigned multicast group of addr re-acq. */
void checkAddressAcquisition(int fecIdx,int cc)
{
  int i, slot = -1;
  for (i=0; i<TO_TABLE_SIZE; i++)
  { /* look for this link the timeout table */
    if (slot == -1 && FecToTbl[i].i1val == 0) slot = i;
    if (FecToTbl[i].i1val == fecIdx) break;
  }
  if (cc == 0)
  { /* successful acquisition, clear counter if there : */
    if (i < TO_TABLE_SIZE) FecToTbl[i].i1val = FecToTbl[i].i2val = 0;
    return;
  }
  if (i == TO_TABLE_SIZE) 
  { /* target link not found in timeout table */
    if (slot >= 0)
    { /* but found a vacant slot */
      i = slot;
      FecToTbl[i].i1val = fecIdx;
    }
    else
    { /* lot's of timeouts! (no room in timout table) */
      return;
    }
  }
  if (FecToTbl[i].i2val++ >= TO_THRESHOLD  && fecAddrAcqId == -1)
  {
    feclog("FEC %.16s : force address resolution due to %d timeouts (%s)",
      FecTbl[fecIdx].fecName,FecToTbl[i].i2val,(cc == connection_timeout ? "synchronous" : "asynchronous"));
    FecTbl[fecIdx].inetProtocol = 0; /* forces address re-acquisition */
    FecToTbl[i].i1val = FecToTbl[i].i2val = 0; /* re-cycle this */
  }
}
void resetLinkTOcounters(int fecIdx)
{
  int i;
  ConTblEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
    if ((c=conTbl[i]) != NULL && c->fecIdx == fecIdx) c->tocounter = 0;
}
int allLinksAboveThreshold(int fecIdx)
{
  int i, bmod;
  ConTblEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if ((c=conTbl[i]) == NULL) continue;
    if (c->fecIdx != fecIdx) continue;
    if ((bmod=BASEMODE(c->mode)) <= CM_SINGLE || bmod == CM_REGISTER) continue;
    if (c->tocounter < TO_THRESHOLD) return 0;
  }
  return 1;
}
void setFecLinkErrorAlarm(int lid,int mode)
{ /* not exported -> can forgo certain checks ... */
  char *fullName;
  BYTE flags = almNEWALARM;
  ExportListStruct *el;
  ConTblEntry *cte = NULL;
  int code = encodeLinkErrorAlarm(lid);
  for (el=ExportList; el != NULL; el=el->next)
  { /* find the first 'reasonable' equipment module */
    if (!strcmp(el->EqmName,SRVEQM_NAME)) continue; /* don't attach to this one !*/
    if (!el->inidone || el->isidle) continue;
    break;
  }
  if (el == NULL || !AlarmServerInitialized) return;
  if ((cte=getConTblEntryPtr(lid)) == NULL) return;
  fullName = getLinkKey(cte);
  if (mode == CM_SINGLE)
  {
    if ((cte->access & CA_WRITE) == 0) return; /* don't set for sync READ calls */
    flags = almTRANSIENT;
  }
  _setAlarm(el->EqmName,0,code,(BYTE *)fullName,flags,makeDataTimeStamp(),NULL);
}
void setFileErrorAlarm(char *filename)
{
  BYTE flags = almNEWALARM;
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next)
  { /* find the first 'reasonable' equipment module */
    if (!strcmp(el->EqmName,SRVEQM_NAME)) continue; /* don't attach to this one !*/
    if (!el->inidone || el->isidle) continue;
    break;
  }
  if (el == NULL || !AlarmServerInitialized) return;
  _setAlarm(el->EqmName,0,file_error,(BYTE *)filename,flags,makeDataTimeStamp(),NULL);
}
void clearFileErrorAlarm(void)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next)
  { /* find the first 'reasonable' equipment module */
    if (!strcmp(el->EqmName,SRVEQM_NAME)) continue; /* don't attach to this one !*/
    if (!el->inidone || el->isidle) continue;
    break;
  }
  if (el == NULL || !AlarmServerInitialized) return;
  lasClearAlarm(el->EqmName,0,file_error,NULL);
}
void rmvRdrInfo(RdrTblEntry *r,int cc)
{
  if (r != NULL)
  { /* revert to original conditions */
    int k, fid;
    msglog(cc,"Redirected Link /%.32s/%.32s/%.64s[%.64s] to /%.32s/%.32s/%.64s[%.64s] has gone down : revert to original conditions",
      r->srcContext,r->srcServerName,r->srcDeviceName,r->srcProperty,
      r->dstContext,r->dstServerName,r->dstDeviceName,r->dstProperty);
    for (k=0; k<nConnectionTableEntries; k++) 
    { /* check everybody in the list against this guy */
      if (conTbl[k]->rdrLink == r)
      { /* the next reonnection attempt will not see a redirection and must re-learn */
        conTbl[k]->rdrLink = NULL;
        conTbl[k]->rdrId = 0;
        strncpy(conTbl[k]->ctxName,r->srcContext,CONTEXT_NAME_SIZE);
        strncpy(conTbl[k]->expName,r->srcServerName,EXPORT_NAME_SIZE);
        strncpy(conTbl[k]->devName,r->srcDeviceName,DEVICE_NAME_SIZE);
        strncpy(conTbl[k]->prpName,r->srcProperty,PROPERTY_NAME_SIZE);
        makeLinkKey(conTbl[k]->key,r->srcContext,r->srcServerName,r->srcDeviceName,r->srcProperty);
        fid = locateExport(r->srcServerName,r->srcContext,r->srcDeviceName,&conTbl[k]->srvIdx);
        if (fid >= 0) conTbl[k]->fecIdx = fid;
      }
    }
    removeRedirectedLink(r,cc);
  }
}
void reAcqAddr(ConTblEntry *c)
{
  int ismcast = (c->mode & CM_NETWORK);
  resetLinkTOcounters(c->fecIdx);
  c->blksin = c->bytesin = 0;
  feclog("Re-acquire address for connection %d: FEC %.16s",c->linkId,FecTbl[c->fecIdx].fecName);
  FecTbl[c->fecIdx].inetProtocol = 0; /* forces address re-acquisition */
  if (NameServerLoaded) 
  { /* ENS known: use asynchronous address re-acquisition */
    fecAddrMcGrp = ismcast ? c->mcastGrp : inadr_zero;
    asyncRelocateExport(c);
  }
  else 
  { /* follow synchronous procedure: looks at local cache as well */
    int fid = locateExport(SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmContext,c->devName,NULL);
    if (fid < 0) return;
    c->fecIdx = fid;
    if (ismcast) 
    {
      INADR inaddr = fixMcastHaddr((BYTE *)&FecTbl[c->fecIdx].netAdr,!(FecTbl[c->fecIdx].addrFlags & ADDRFLAG_MCAST_MODERN));
      if (memcmp(&c->mcastGrp,&inaddr,gInAdrLen))
      { /* new multicast group ! */
        detachPortFromMulticastGroup(mcastClnSck,(UINT16)gtMCastPort,c->mcastGrp,"server specific");
        attachPortToMulticastGroup(mcastClnSck,(UINT16)gtMCastPort,inaddr,"server specific");
        c->mcastGrp = inaddr;
      }
    }
  }
}
void clearFecLinkErrorAlarm(int linkId)
{ /* not exported -> can forgo certain checks ... */
  int code;
  ExportListStruct *el=ExportList;
  for (el=ExportList; el != NULL; el=el->next)
  { /* find the first 'reasonable' equipment module */
    if (!strcmp(el->EqmName,SRVEQM_NAME)) continue;
    if (!el->inidone || el->isidle) continue;
    break;
  }
  if (el == NULL || !AlarmServerInitialized) return;
  code = encodeLinkErrorAlarm(linkId);
  RemoveAlarm(el->EqmName,0,code);
}
UINT32 nPartialTransfers = 0;
#define BLACKLIST_FLUSH_INTERVAL 300
#define GRACEINTVL 100
#define TCP_IDLE_CHECK_INTERVAL 60
void checkConnections(void)
{
# define DTSECS(t,s) ((int)((t).tv_sec-(time_t)(s)))
# define DTMSECS(dt,t,c) ((dt)*1000 + ((t).tv_usec/1000-(int)(c)->timeMsecLastSync))
  int i, k, mode, bmode, to_multiplier, to_ceiling, hbt, delta_t, cc=link_timeout;
  int addrChng=FALSE, ismmf=FALSE, resetBadPacketCnts=FALSE;
  static time_t lasttime = 0, lastaddrreq = 0;
  int isTrace = FALSE;
  ConTblEntry *c;
  GrpTblEntry *g;
  FecAddrStruct fec;
  ExpDataStruct srv;
  IDLE_CONNECTION *ic=NULL;
  struct timeval tv_stamp;                
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  int pipe;
# endif
  INADR inaddr;

  gettimeofday(&tv_stamp,(struct timezone *)NULL);
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) != 0) return;
  for (i=0; i<nConnectionTableEntries; i++) 
  {
    if ((c=conTbl[i]) == NULL) continue;
    if (c->tineProtocol < 5) continue;
    isTrace = isTraceLink(i);
    if ((bmode=BASEMODE(c->mode)) == CM_CANCEL) 
    {
      c->tcpSck = 0;
      continue;
    }
    if (c->cancelPending && c->mcaEntry != NULL)
    { /* check parent MCA links for closeLink */
      if (WaitForSystemMutex(hMcaLinkMutex,gSystemTick) == 0)
      {
        if (c->cancelPending && (UINT32)tv_stamp.tv_sec > (UINT32)(c->timemark + 5))
        { /* cancelPending still set and enough time has expired */
          McaTblEntry *mca = c->mcaEntry;
          c->mcaEntry = NULL;
          c->cancelPending = FALSE;
          _closeLink(i);
          removeMcaLinkEntry(mca);
        }
        ReleaseSystemMutex(hMcaLinkMutex);
      }
    }
    if (c->heartbeat == HBT_NONE) continue; /* heartbeat = 0 -> REGISTER */
    hbt = (c->heartbeat == HBT_SUB_PENDING) ? (c->pollingInterval/1000) : c->heartbeat;
    if (bmode == CM_SINGLE && c->mode & CM_STREAM) continue; /* don't check on connected streams in sync mode */
    if (c->needsNotification) continue; /* let this notification happen normally */
    if (c->timeLastSync == 0) continue; /* not yet requested */
    if ((delta_t=DTSECS(tv_stamp,c->timeLastSync)) > MAXALLOWED_POLLINTVL) delta_t = MAXALLOWED_POLLINTVL/1000;
    delta_t = DTMSECS(delta_t,tv_stamp,c);
    if (delta_t < 0) delta_t = 0;
    if (c->fecIdx >= 0 && FecTbl[c->fecIdx].inetProtocol == 0 && c->heartbeat == HBT_SUB_PENDING)
    { /* don't get stuck without knowing the address and protocol ! */
      c->tocounter++;
      goto reacq;
    }
    if (bmode == CM_SINGLE && (UINT32)delta_t <= c->pollingInterval) continue;
    to_ceiling = (UINT32)1000*hbt + GRACEINTVL;
    if (c->isGrouped != 0 && (g=getGrpTblEntry((size_t)c->cbfcn)) != NULL)
    { /* belongs to a group: if synchronized find the best timeout ceiling */
      if (g->grpSyncLevel == GRP_SYNC_INSYNC &&
          g->grpUpdateInterval > 0) 
        to_ceiling = g->grpUpdateInterval;
    }
    if (bmode > CM_SINGLE && delta_t < to_ceiling) continue;
    if (c->heartbeat == HBT_SUB_PENDING) 
    { /* retry already signaled, but reset bookkeeping */
      c->blksin = c->bytesin = 0;
      continue;
    }
    /* timeout noted ! */
    cc = c->blksin > 0 ? remitted_data_lost : link_timeout;
    c->blksin = c->bytesin = 0;
    if (IsConnProt(c->inetProtocol) && gBadPacketCnts > BADPCKT_CEILING)
    {
      msglog(link_timeout,"%.96s timeout with %d bad packets: close TCP connection!",
        GetLinkName(i),gBadPacketCnts);
      CloseConnectionSocket(c->tcpSck, c->inetProtocol, TRUE); /* runs thru conTbl ! */
      c->tcpSck = 0;
      c->heartbeat = HBT_SUB_PENDING;
      resetBadPacketCnts = TRUE;
    }
    if (gIsRunningAsServer && gAutoLinkErrorAlarms)
    { /* set a relevant alarm */
      if (c->tocounter > TO_RETRY_THRESHOLD || bmode == CM_SINGLE)
      {
        setFecLinkErrorAlarm(i,bmode);
      }
    }
    if (tineDebug > 1)
#     ifdef OS_16BIT
      dbglog("%.192s : %ld >= %ld + %d",c->key,tv_stamp.tv_sec,c->timereq,(int)c->heartbeat);
#     else
      dbglog("%.192s : %d >= %d + %d",c->key,tv_stamp.tv_sec,c->timeLastSync,(int)c->heartbeat);
#     endif
    c->lastLnkErrSrc = lastLnkErrSrc = LNK_ERROR_LCL;
    if (c->linkStatus == link_blacklisted) cc = link_blacklisted;
    lastLnkErr = cc;
    strncpy(gLastStatusString,cc2str(cc),STATUS_SHORTSIZE);
    strncpy(c->lastStatusString,cc2str(cc),STATUS_SHORTSIZE);
    nConnectionTimeouts++;
    c->tocounter++;
    c->dataStatus = not_posted;
    c->linkStatus = cc;
    if ((c->access & CA_RETRY) && (c->tocounter % TO_RETRY_THRESHOLD))
    { /* RETRY bit is set && not retried (N.B. tocounter will increase to 30 before being reset!) */
      c->heartbeat = HBT_SUB_PENDING;
      c->mode |= CM_RETRY; /* always set this here (clear in prepOutgoingSubRequestHdr() */
      if (tineDebug > 1 || errorDebug) dbglog("retry %.192s : %d timeouts so far",c->key,c->tocounter);
      continue;
    }
    c->xferReason = CX_NULL;
    c->suppressCallback = FALSE;
    mode = c->mode; /* note existing transport mode */
    if (!strcmp(FecTbl[c->fecIdx].fecName,"FECSINK")) cc = address_unknown;
    ReleaseSystemMutex(hLinkTblMutex); /* never know what someone might do inside a callback ... */
    NotifyClient(i,cc);
    if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) != 0) return;
    if (bmode != CM_SINGLE) 
    { /* check again in case it was canceled inside the callback ... */
      bmode = BASEMODE(c->mode); 
    }
    if (c->bytesin > 0) nPartialTransfers++;
    c->dtimestamp = 0L; c->dtimestampUSEC = 0;
    c->blksin = 0; c->bytesin = 0;
    c->timeLastSync = (UINT32)tv_stamp.tv_sec;
    c->timeMsecLastSync = (UINT32)(tv_stamp.tv_usec/1000);
    if (c->boundToId > 0) continue;
    ismmf = (FecTbl[c->fecIdx].inetProtocol == MMF) ? TRUE : FALSE;
    if (bmode > CM_SINGLE) 
    { /* is a persistent link */
      c->heartbeat = HBT_SUB_PENDING;
      if (ismmf && c->tocounter < TO_THRESHOLD - 3) c->tocounter = TO_THRESHOLD - 3;
    }
    else if (bmode == CM_SINGLE)
    { /* a single link has timed out => cancel it ! */
      if (c->heartbeat == HBT_SUB_PENDING)
      { /* CM_SINGLE was re-submitted in callback */
        c->mode = mode;
        if (isTrace) traceLinkOutput(c, "check connections", "resubmitting SINGLE link");
      }
      else
      {
        c->heartbeat = HBT_NONE;
        c->mode = CM_CANCEL;
        if (isTrace) traceLinkOutput(c, "check connections", "link cancel (due to timeout)");
      }
      msglog(cc,"/%.32s/%.32s/%.64s[%.64s] : timed out",
        SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName);
      if (ismmf) c->tocounter = TO_THRESHOLD + 1;
      if (c->cbfcn != (TCBFCNP)_cbExecLink) checkAddressAcquisition(c->fecIdx,link_timeout);
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    if (FecTbl[c->fecIdx].inetProtocol == PIPE) 
    {
      pipe = FecTbl[c->fecIdx].lclAdr.pipe.sck;
      ClosePipeSocket(pipe);
      msglog(cc,"FEC %.16s : reset pipe due to timeout",FecTbl[c->fecIdx].fecName);
    }
#   endif
reacq:
    if (c->linkStatus == link_blacklisted) continue;
    if (strncmp(c->expName,ENS_TAG,EXPORT_NAME_SIZE)) /* not the name server */
    { /* locateExport will call checkConnections too */
      to_multiplier = bmode == CM_SINGLE ? 1 : 3;
      if (c->pollingInterval > 1000) to_multiplier /= (c->pollingInterval / 1000);
      if (ismmf) to_multiplier = 1;
      if (c->tocounter > TO_THRESHOLD*to_multiplier &&
          allLinksAboveThreshold(c->fecIdx) &&
          tv_stamp.tv_sec != lastaddrreq)
      {
        if (fecAddrAcqId != -1)
        {
          if (tv_stamp.tv_sec > lastaddrreq + 3) 
          {
            msglog(0,"reset address reacquisition semaphore");
            fecAddrAcqId = -1;
            fecAddrMcGrp = inadr_zero;
          }
        }
        else
        {
          reAcqAddr(c); /* resets the link counter to 0 */
          lastaddrreq = tv_stamp.tv_sec;
        }
      }
    }
    else
    { /* a link to the ENS ! */
      if (tineDebug || errorDebug) dbglog("%s : timed out!",ENS_TAG);
      if (c->tocounter >= TO_RETRY_THRESHOLD) c->tocounter = 0;
    }
  }
  for (ic=tcpIdleConnectionList; ic != NULL; )
  {
    if ((delta_t=DTSECS(tv_stamp,lasttime)) > MAXALLOWED_POLLINTVL) delta_t = MAXALLOWED_POLLINTVL/1000;
    ic->ttl -= delta_t*1000;
    if (ic->ttl < 0) 
    {
      msglog(0,"close idle TCP connection on socket %d",ic->sck);
      CloseConnectionSocket(ic->sck, ic->trp, TRUE); /* runs thru conTbl ! */
      break; /* continue scanning next time around */
    }
    else ic = ic->nxt;
  }
  ReleaseSystemMutex(hLinkTblMutex);
  for (i=0; i<nglobals; i++)
  { /* raise alarm for missing heartbeat */
    if (glbTbl[i]->disabled) continue;
    if (glbTbl[i]->maxPeriod < 0) continue; /* not from a known GLOBALS server */
    if (tv_stamp.tv_sec > glbTbl[i]->tv.tv_sec + glbTbl[i]->maxPeriod*GLOBALGRACE)
    { /* data is not coming in on time ! */
      if (glbTbl[i]->cbfcn != NULL || glbTbl[i]->conTblId > 0) NotifyGlobalClient(i,link_timeout);
      glbTbl[i]->hasNotifiedOnce = TRUE;
      glbTbl[i]->tv = tv_stamp;
      glbTbl[i]->status = link_timeout;
      if (glbTbl[i]->tocounter++ > TO_THRESHOLD * 3)
      { /* threshold counter exceeded ! ... */
        if (gIsRunningAsServer) 
          feclogEx(TINE_LOGLEVEL_ERROR,"missing global keyword %.64s (%d consecutive timeouts)",glbTbl[i]->keyword,glbTbl[i]->tocounter);
        glbTbl[i]->tocounter = 0;
        inaddr = glbTbl[i]->mcastGrp;
        /* always detach */
        detachPortFromMulticastGroup(glbClnSck,(UINT16)gtGCastPort,inaddr,"globals specific");
        glbTbl[i]->mcastGrp = inadr_zero; /* and zero it out in the globals table ! */
        glbTbl[i]->time = glbTbl[i]->timeUSEC = 0; /* and reset this (in case of VxWorks timestamp problems) */
        addrChng = FALSE;
        if (queryNameServerForAddress(glbTbl[i]->context,glbTbl[i]->server,&fec,&srv) == 0)
        { /* re-acquired the address from the ENS */
          if (memcmp(&glbTbl[i]->srvIp,&fec.netAdr,gInAdrLen)) addrChng = TRUE;
          inaddr = fixMcastHaddr((BYTE *)&fec.netAdr,!(fec.addrFlags & ADDRFLAG_MCAST_MODERN));
        }
        /* and always re-attach */
        attachPortToMulticastGroup(glbClnSck,(UINT16)gtGCastPort,inaddr,"globals specific");
        glbTbl[i]->mcastGrp = inaddr;
        if (addrChng)
        { /* noted address change of globals server */
          for (k=0; k<nglobals; k++)
          { /* anybody else pointing to the same server ? */
            if (i == k) continue;
            if (glbTbl[k]->status != link_timeout) continue;
            if (!adrcmp(&glbTbl[i]->srvIp,&glbTbl[k]->srvIp))
            { /* probably another globals variable from same source */
              glbTbl[k]->mcastGrp = inaddr;
              adrcpy(&glbTbl[k]->srvIp,&fec.netAdr);
              glbTbl[k]->tocounter = 0;
              feclog("multicast group for global %.32s keyword %.64s changed to %.40s",
                glbTbl[k]->context,glbTbl[k]->keyword,fec.strAdr);
            }
          }
          adrcpy(&glbTbl[i]->srvIp,&fec.netAdr);
          feclog("multicast group for global %.32s keyword %.64s changed to %.40s",
            glbTbl[i]->context,glbTbl[i]->keyword,fec.strAdr);
        }
        break; /* break out in order to stagger the 'reacquisition shuffle' */
      }
    }
  }
  if (resetBadPacketCnts) gBadPacketCnts = 0;
  if (tv_stamp.tv_sec != lasttime)
  { /* TRUE once per second */
#   if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
    if ((tv_stamp.tv_sec % BLACKLIST_FLUSH_INTERVAL) == 0) 
    {
      flushBlackLnkLst();
    }
    flushGroupCacheList(50);
#   endif
    if ((tv_stamp.tv_sec % TCP_IDLE_CHECK_INTERVAL) == 0)
    {
      checkTcpIdleConnections();
    }
    lasttime = tv_stamp.tv_sec;
  }
  return;
}
void checkRemotePipe(void)
{ /* TODO: do a better job of this */
# if defined(WIN32) && !defined(NO_SERVICE_PIPE)
  int len;
  char cmdbuf[CMDSIZE];
  if (ipcCmdSck == 0) return;
# if !defined(NO_SERVICE_PIPE)
  if ((len=GetCommandEx(ipcCmdSck,cmdbuf,CMDSIZE)) > 0) InterpretCommand(cmdbuf);
# endif
# endif
}
int gClientCycleIdle = FALSE;
UINT32 gTimeClientCycle = 0;
UINT32 gTimeClientIdleLastCycle = 0;
UINT32 gTimeClientFdIdle = 0;
TINE_EXPORT void SetClientCycleIdle(int value)
{
  gClientCycleIdle = value;
}
TINE_EXPORT int GetClientCycleIdle(void)
{
  return gClientCycleIdle;
}
void clientCycle(void)
{
  int saveClientCycle;
  ProbeTimeElapsed("client cycle entered");
  if (gClientCycleIdle)
  {
    millisleep(gSystemTick);
    return;
  }
  if (WaitForSystemMutex(hSystemClientMutex,0) == 0)
  {
    ProbeTimeElapsed("client cycle check activity");
    CheckConsumerData(FALSE); /* flush input on the socket */
    checkSubscriptions(FALSE); /* need to send anything out ? */
    CheckConsumerData(TRUE);  /* check input on socket again */
    if (needToAcknowledge)
    { /* last batch of updates requires an ACK */
      checkSubscriptions(FALSE);
      needToAcknowledge = FALSE;
    }
    checkConnections(); /* missing heartbeats ?   */
    ReleaseSystemMutex(hSystemClientMutex);
  }
# ifdef MULTITHREADED
  else if (gRunServerCycleInSeparateThread)
  { /* nothing else below will yield execution ! */
    YieldThread();
  }
# endif
  ProbeTimeElapsed("client cycle check notifications");
  checkClientNotification(); /* always check, irrespective of Semaphore */
  ProbeTimeElapsed("client cycle check local pipes");
  checkRemotePipe();
# if !defined(SMALL_TINE_LIB)
  if (accessLockLst != NULL) checkAccessLockItems();
# endif
  ProbeTimeElapsed("client cycle returning");
  return;
}
int isSckError(int cc)
{
  switch (cc)
  {
    case tcp_connect_error:
    case tcp_connection_closed:
    case udp_socket_error:
    case tcp_socket_error:
    case tcp_not_supported:
      return TRUE;
    default:
      return FALSE;
  }
}
void checkClientNotification(void)
{
  int i, bmode, needsToNotify, isSyncNotify;
  ConTblEntry *c;
# ifndef SMALL_TINE_LIB
  int isStructField;
# endif
  if (WaitForSystemMutex(hLinkCbMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if ((c=conTbl[i]) == NULL) continue;
#   if !defined(SMALL_TINE_LIB)
    if (c->relinked & RELINKED_NEEDED)
    {
      isStructField = (c->relinked & RELINKED_STRUCT) ? TRUE : FALSE;
      c->relinked = RELINKED_TRUE;
      c->linkStatus = reAttachLink(i,isStructField);
      if (c->linkStatus != invalid_field) continue;
    }
#   endif /* !SMALL_TINE_LIB */
    bmode = BASEMODE(c->mode);
    if (bmode == CM_REGISTER) c->needsNotification = FALSE;
    if (c->cbDelay > 0 && c->cbDelayCounter-- > 0) continue;
    isSyncNotify = (c->mode & CM_SYNCNOTIFY) == CM_SYNCNOTIFY ? TRUE : FALSE;
    needsToNotify = c->needsNotification;
    if (c->linkStatus == notification_pending) needsToNotify = FALSE;
    if (needsToNotify && isSckError(c->linkStatus))
    { /* don't fall into the following TCP check ! */
      isSyncNotify = FALSE;
    }
    if (bmode > CM_SINGLE && isSyncNotify)
    { /* allow notification from doProcessData() only: needsNotification is NOT thread-safe */
      if ((c->inetProtocol&(TCP|TCP6|STREAM|STREAM6))) needsToNotify = FALSE;
      if (c->tcpSck == 0) c->needsNotification = FALSE; else needsToNotify = FALSE;
    }
    if (needsToNotify) NotifyClient(i,c->linkStatus);
    if (c->cbDelay > 0) c->cbDelayCounter = c->cbDelay;
  }
  ReleaseSystemMutex(hLinkCbMutex);
}
void applyErrorValue(ConTblEntry *c,void *errValue)
{
  int i,fmt = c->formatOut,len = fmtsizeof(fmt);
  BYTE *ptr;
  if (errValue == NULL) return;
  if (len < 1 || len > 64) return;
  if (c->tocounter < TO_RETRY_THRESHOLD) return;
  for (i=0; i<(int)c->sizeOut; i++) 
  {
    ptr = (BYTE *)(c->dataPtr + (size_t)i*len);
    memcpy(ptr,errValue,len);
  }
}
int removeLinkFromGroup(int i)
{
  GrpTblEntry *gt;
  ConTblEntry *c;

  if (i < 0 || i >= nConnectionTableEntries) return invalid_index;
  if ((c=conTbl[i]) == NULL || c->mode == CM_CANCEL) return un_allocated;
  if (c->isGrouped != 0)
  {
    if ((gt=getGrpTblEntry((size_t)c->cbfcn)) != NULL)
    {
      gt->numPending--;
      gt->numInGrp--;
      msglog(0,"/%.32s/%.32s/%.64s[%.64s] : removing link [%d] from grouped callback",
        SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName,i);
      if (gt->members != NULL)
      { /* have an established members list: remove this one */
        GrpMember *grpMmbr=NULL, *prv=NULL;
        for (grpMmbr=gt->members; grpMmbr != NULL; grpMmbr=grpMmbr->nxt)
        { 
          if (grpMmbr->conTblId == i)
          { /* found it ! */
            if (grpMmbr == gt->members)
            { /* right at the head of the list ... */
              gt->members = grpMmbr->nxt;
            }
            else
            { /* splice */
              prv->nxt = grpMmbr->nxt;
            }
            SystemFree(grpMmbr);
            break;
          }
          prv = grpMmbr;
        }
      }
    }
    c->isGrouped = 0; c->grp = NULL;
  }
  return 0;
}
int addLinkToGroup(int i)
{
  GrpTblEntry *gt;
  ConTblEntry *c;

  if (i < 0 || i >= nConnectionTableEntries) return invalid_index;
  if ((c=conTbl[i]) == NULL || c->mode == CM_CANCEL) return un_allocated;
  if (c->isGrouped) return already_assigned;
  if ((gt=getGrpTblEntry((size_t)c->cbfcn)) == NULL) return code_failure;
  gt->numPending++;
  gt->numInGrp++;
  msglog(0,"%.192s : add link [%d] to grouped callback",c->key,i);
  c->isGrouped = 1;
  if (gt->members != NULL)
  { /* already have a members list: need to add this one */
    GrpMember *grpMmbr = (GrpMember *)calloc(1,sizeof(GrpMember));
    if (grpMmbr != NULL)
    {
      grpMmbr->conTblId = i; grpMmbr->glbTblId = -1;
      strncpy(grpMmbr->key,c->key,192);
      grpMmbr->dout.dFormat = c->formatOut;
      grpMmbr->dout.dArrayLength = c->sizeOut;
      strncpy(grpMmbr->dout.dTag,c->strTagOut,TAG_NAME_SIZE);
      grpMmbr->nxt = gt->members;
      gt->members = grpMmbr;
    }
  }
  c->grp = gt;
  return 0;
}
TINE_EXPORT int RemoveLinkFromGroup(int linkId)
{
  return removeLinkFromGroup(linkId);
}
TINE_EXPORT int AddLinkToGroup(int linkId)
{
  return addLinkToGroup(linkId);
}
void NotifyGlobalClient(int i,int cc)
{
  int k;
  GlobalListStruct *g = NULL;
  GrpTblEntry *gt = NULL;

  if (!(g=glbTbl[i])->isPending) return;
  g->isPending = FALSE;
  gDataTimeStamp = makeDataTimeStamp();
  if ((k=g->conTblId) > 0)
  { /* was initiated as a conTbl link entry */
    ConTblEntry *c = conTbl[--k];
    if (c->formatOut == CF_DEFAULT)
    {
      c->formatOut = g->format;
      c->sizeOut = g->size;
    }
    c->notifyPending = 1;
    c->linkStatus = cc;
    strncpy(c->lastStatusString,cc2str(cc),STATUS_SHORTSIZE);
    c->lastDataSize = c->sizeOut;
    c->dtimestamp = (UINT32)g->time;
    c->dtimestampUSEC = g->timeUSEC;
    c->systemstamp = g->sysStamp;
    c->datastamp = g->dataStamp;
    NotifyClient(k,cc);
    if (BASEMODE(c->mode) <= CM_SINGLE) 
    {
      c->mode = CM_CANCEL;
      c->glbTblId = 0;
      g->conTblId = 0;
    }
  }
  g->hasNotifiedOnce = TRUE;
  if (g->isGrouped)
  {
    for (gt=grpTbl; gt && gt->grpId != (size_t)g->cbfcn; gt = gt->next);
    if (gt == NULL) 
    {
      g->isGrouped = 0; g->grp = NULL;
    }
    else
    {
      gt->grpCc |= cc;
      if (--(gt->numPending) > 0) return;
      gt->numPending = gt->numInGrp;
      cc = gt->grpCc;
      gt->grpCc = 0;
    }
  }
  if (g->cbfcn != NULL) TCBCALL(g,cc);
  g->isPending = TRUE;
  if (g->isGrouped)
  { /* in case of a group, run thru the list */
    for (k=0; k<nglobals; k++) 
    { 
      if (glbTbl[k]->cbfcn == g->cbfcn) glbTbl[k]->isPending = TRUE;
    }
  }
  return;
}
int groupCanNotify(ConTblEntry *c,int *cc)
{
# define GRP_JITTER 50
  static time_t last_shuffle = 0;
  int i, upd, latency, latency_floor, offset, offset_floor;
  struct timeval tv, tvg;
  ConTblEntry *cte;
  GrpTblEntry *g = NULL;
  if ((g=c->grp) == NULL) 
  { /* not a group after all ... */
    c->isGrouped = 0;
  }
  else
  {
    gettimeofday(&tv,NULL);
    if (g->numPending == g->numInGrp)
    { /* initial callback in this set */
      g->grpBndWdthC = 0;  
      g->grpCc = *cc;
      if (g->grpSyncLevel <= GRP_SYNC_STARTSYNC)
      { /* absolute initial callback */
        g->grpHeadLnkId = c->linkId;
        g->grpUpdateInterval = c->pollingInterval;
      }
    }
    if (g->grpCc == 0) g->grpCc |= *cc;
chkGroupHead:
    if (c->linkId == g->grpHeadLnkId)
    { /* peg all of this on the group master */
      g->grpLastUpdate = g->grpTimeUpdate;
      g->grpTimeUpdate = tv;
      g->grpLastStamp = g->grpSysStamp;
      g->grpSysStamp = c->systemstamp;
    }
    latency = MSECS(tv,g->grpTimeUpdate);
    offset = c->syncOffset = c->systemstamp - g->grpSysStamp;
    if (g->grpSyncLevel > GRP_SYNC_CYCLEOFFSET)
    {
      if (c->linkId != g->grpHeadLnkId)
      { /* not the group head */
        if ((cte=conTbl[g->grpHeadLnkId]) != NULL)
        {
          if (cte->notifyPending)
          { /* members updating but group head is still pending */
            if (g->grpIncMono)
            { /* redefine the group head */
              if (c->systemstamp <= g->grpSysStamp)
              { /* skip this extraneous update */
                c->notifyPending = 1;
                return FALSE;
              }
              g->grpHeadLnkId = c->linkId;
              goto chkGroupHead;
            }
            if (latency < g->grpUpdateInterval - GRP_JITTER)
            { /* reject this: it'c comming in too fast (scheduled ?) */
              c->notifyPending = 1;
              return FALSE;
            }
            offset = latency = 0; /* as yet, no group head notification ! */
          }
          else if (cte->systemstamp > g->grpSysStamp) 
          { /* could be advancing underneath (e.g. scheduled property) */
            tvg.tv_sec = cte->timeLastSync; tvg.tv_usec = cte->timeMsecLastSync * 1000;
            g->grpSysStamp = cte->systemstamp;
            g->grpTimeUpdate = tvg;
          }
        }
      }
    }
    c->grpLatency = latency;
    if (g->grpSyncLevel == GRP_SYNC_NOTIFICATION)
    { /* find the best group order */
      if (c->systemstamp > g->grpSysStamp &&
          c->grpLatency > (int)c->pollingInterval/2)
      { /* re-group ! */
        msglog(0,"group synchronization: redefine group order");
        g->grpHeadLnkId = c->linkId;
        g->grpUpdateInterval = c->pollingInterval;
        g->numPending = g->numInGrp;
        g->grpTimeUpdate = tv;
        g->grpSysStamp = c->systemstamp;
        c->grpLatency = 0;
        g->grpSyncLevel = GRP_SYNC_CYCLEOFFSET;
        for (i=0; i<nConnectionTableEntries; i++)
        { /* reset the pending flags for the other members ? */
          if ((cte=conTbl[i]) == NULL || cte == c) continue;
          if (cte->grp == g) cte->notifyPending = 1;
        }
      }
    }
    if (--(g->numPending) > 0) return FALSE;
    /* now they're all in ... */
    latency = 0; latency_floor = 0;
    offset = 0; offset_floor = 0;
    if ((cte=conTbl[g->grpHeadLnkId]) != NULL)
    { /* re-sync these values in case a straggler held us up */
      tvg.tv_sec = cte->timeLastSync; tvg.tv_usec = cte->timeMsecLastSync * 1000;
      g->grpSysStamp = cte->systemstamp;
      g->grpTimeUpdate = tvg;
    }
    for (i=0; i<nConnectionTableEntries; i++)
    { /* find the real offsets */
      if ((cte=conTbl[i]) == NULL) continue;
      if (i == g->grpHeadLnkId) continue;
      if (cte->grp != g) continue;
      tv.tv_sec = cte->timeLastSync; tv.tv_usec = cte->timeMsecLastSync * 1000;
      cte->syncOffset = cte->systemstamp - g->grpSysStamp;
      if (g->grpIncMono && cte->syncOffset < 0)
      { /* grp is monotonically incrementing but some members are late : re-order ! */
        if (last_shuffle == 0)
        { /* but only every so often */
          cte->notifyPending = 1;
          g->numPending++;
        }
      }
      cte->grpLatency = MSECS(tv,g->grpTimeUpdate);
      if (cte->grpLatency < latency_floor) latency_floor = cte->grpLatency;
      if (cte->grpLatency > latency) latency = cte->grpLatency;
      if (cte->syncOffset < offset_floor) offset_floor = cte->syncOffset;
      if (cte->syncOffset > offset) offset = cte->syncOffset;
    }
    if (g->numPending > 0)
    { /* could happen due to group reshuffling */
      last_shuffle = tv.tv_sec; /* will prevent numPending being artificially incremented too often */
      return FALSE;
    }
    if (last_shuffle > 0 && tv.tv_sec > last_shuffle + 1200) 
    { /* if 'active' reset this after 10 minutes */
      last_shuffle = 0;
    }
    g->grpBndWdthT = latency - latency_floor;
    g->grpBndWdthC = offset - offset_floor;
    switch (g->grpSyncLevel)
    {
      case GRP_SYNC_STARTSYNC:
        g->grpSyncLevel = GRP_SYNC_NOTIFICATION;
        msglog(0,"group synchronization: find most effective group order");
        break;
      case GRP_SYNC_NOTIFICATION:
        g->grpSyncLevel = GRP_SYNC_CYCLEOFFSET;
        break;
      case GRP_SYNC_CYCLEOFFSET:
        g->grpSyncLevel = GRP_SYNC_INTERVAL;
        break;
      case GRP_SYNC_INTERVAL:
        g->grpSyncLevel = GRP_SYNC_INSYNC;
        msglog(0,"group synchronization: mark as synchronized");
      case GRP_SYNC_INSYNC:
        if ((cte=conTbl[g->grpHeadLnkId]) != NULL)
        {
          tv.tv_sec = cte->timeLastSync; tv.tv_usec = cte->timeMsecLastSync * 1000;
          upd = MSECS(tv,g->grpLastUpdate) + GRP_JITTER - 1;
          upd -= (upd%GRP_JITTER);
          if (upd < g->grpUpdateInterval)
          {
            g->grpUpdateInterval = upd;
            msglog(0,"group synchronization: fix effective interval to %d msec",g->grpUpdateInterval);
          }
        }
        break;
    }
    g->grpIncMono = (g->grpSysStamp - g->grpLastStamp == 1) ? TRUE : FALSE;
    if (BASEMODE(c->mode) <= CM_SINGLE) 
    { /* will be marked for cancellation so reduce the group size to 0 */
      g->numInGrp = 0;
    }
    g->numPending = g->numInGrp;
    *cc = g->grpCc;
    g->grpCc = 0;
  }
  return TRUE;
}
void copyDpdData(ConTblEntry *dst,ConTblEntry *src)
{
  int hsiz,dsiz,bsiz,*sptr=NULL;
  BYTE *bptr=NULL,*dbptr=NULL;

  if (WaitForSystemMutex(hLinkTblMutex, -1) == 0)
  {
    if (dst->mode != CM_CANCEL)
    {
      if (dst->dataPtr != NULL && dst->dataPtr != src->dataPtr)
      { /* somewhere to copy to */
        dsiz = src->sizeOut*fmtsizeof(src->formatOut);
        hsiz = getFormatHeaderSize(src->formatOut);
        dsiz += hsiz;
        if (isVariableLengthFormat(src->formatOut))
        { /* do some fancy foot work */
          if (hsiz > 0)
          { /* need to trap formats with headers and memcpy to right place */
            bptr = GetFormatBodyPtr(src->dataPtr,src->formatOut);
            dbptr = GetFormatBodyPtr(dst->dataPtr,src->formatOut);
          }
          if (bptr == NULL)
          { /* reasonably normal case */
            if (src->formatOut == CF_STRING)
            { /* dsiz should just refer to the number of string pointers */
              dsiz = src->lastDataSize * sizeof(char *);
              strncpy(dst->strTagOut,src->strTagOut,TAG_NAME_SIZE);
            }
            memcpy(dst->dataPtr,src->dataPtr,dsiz);
          }
          else
          { /* destination body does not follow the format header */
            memcpy(dst->dataPtr,src->dataPtr,hsiz);
            if (dbptr != bptr) memcpy(dbptr,bptr,dsiz-hsiz);
            bsiz = GetFormatBodyBufferSize(src->dataPtr,src->formatOut);
            sptr = (int *)GetFormatBodyBufferSizeAddr(dst->dataPtr,src->formatOut);
            memcpy(sptr,&bsiz,sizeof(int));
          }
        }
        else if (src->formatOut == CF_MDA)
        {
          DMDA *mdasrc, *mdadst;
          mdasrc = (DMDA *)src->dataPtr;
          mdadst = (DMDA *)dst->dataPtr;
          memcpy(mdadst,mdasrc,hsiz);
          dsiz = MIN(dst->sizeOut,src->sizeOut);
          memcpy(mdadst->data.vptr,mdasrc->data.vptr,dsiz);
        }
        else
        { /* normal case */
          memcpy(dst->dataPtr,src->dataPtr,dsiz);
        }
      }
      dst->lastDataSize = src->lastDataSize;
    }
    ReleaseSystemMutex(hLinkTblMutex);
  }
}
/**
 * \internal checks whether data should be marked as out of tolerance
 *  according to registered 'data tolerance factor'.
 *
 * \return TRUE if incoming data is Out Of Tolerance 
 * \note this routine is not exposed -> forgo check on 'c == NULL' ...
 */
int isOOT(ConTblEntry *c,size_t dsz)
{
  int i,n,m,fsz,boff;
  float delta = 0.0, epsilon = 0.0;  
  char *d1 = c->dataPtr;
  char *d2 = c->bufCmp;
  short fmt = c->formatOut;
  DataToleranceFactor *dtf = c->dtf;
  int oot = FALSE;
  
  if (d2 == NULL) return TRUE; /* no comparison buffer -> always out of tolerance */
  if (!isQueueableFormat(fmt))
  { /* these have headers and separate data bodies */
    int hsz = GetFormatHeaderSize(fmt);
    if (memcmp(d1, d2, hsz)) return TRUE; /* headers not equal */
    d1 = (char *)GetFormatBodyPtr(c->dataPtr, fmt);
    d2 = &d2[hsz];
    fmt = GetFormatDataType(fmt);
    dsz -= hsz;
  }
  if (dtf == NULL) return memcmp(d1,d2,dsz); /* no explicit tolerance factor, but NO_DUPLICATES specified */
  if ((fsz=fmtsizeof(fmt)) < 1) return FALSE; /* sanity check */
  if (dtf->flags == DTF_NONE) return TRUE; /* tolerance check turned off */
  boff = fsz * dtf->offset; /* dtf->offset can only be positive ! */
  n = dtf->offset + dtf->length; /* end of comparison */
  m = (int)dsz/fsz; /* how much came in ? */
  if (n > m) n = m; /* fewer bytes came in as registered */
  switch (fmt)
  {
    case CF_DOUBLE:
      for (i=dtf->offset; i<n && !oot; i++) 
      {
        delta = (float)fabs(((double *)d1)[i] - ((double *)d2)[i]);
        epsilon = (float)fabs(dtf->percent * ((double *)d2)[i]) + dtf->absolute;
        if (delta > epsilon) oot = TRUE;
      }
      break;
    case CF_FLOAT:
      for (i=dtf->offset; i<n && !oot; i++) 
      {
        delta = (float)fabs(((float *)d1)[i] - ((float *)d2)[i]);
        epsilon = (float)fabs(dtf->percent * ((float *)d2)[i]) + dtf->absolute;
        if (delta > epsilon) oot = TRUE;
      }
      break;
    case CF_INT16:
      for (i=dtf->offset; i<n && !oot; i++) 
      {
        delta = (float)abs(((SINT16 *)d1)[i] - ((SINT16 *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((SINT16 *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) oot = TRUE;
      }
      break;
    case CF_UINT16:
      for (i = dtf->offset; i<n && !oot; i++)
      {
        delta = (float)fabs((float)((UINT16 *)d1)[i] - (float)((UINT16 *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((UINT16 *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) oot = TRUE;
      }
      break;
    case CF_INT32:
      for (i=dtf->offset; i<n && !oot; i++) 
      {
        delta = (float)abs(((SINT32 *)d1)[i] - ((SINT32 *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((SINT32 *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) oot = TRUE;
      }
      break;
    case CF_UINT32:
      for (i = dtf->offset; i<n && !oot; i++)
      {
        delta = (float)fabs((float)((UINT32 *)d1)[i] - (float)((UINT32 *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((UINT32 *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) oot = TRUE;
      }
      break;
    case CF_BYTE:
    default:
      oot = memcmp(&d1[boff],&d2[boff],dsz-boff);
  }
  if (dtf->flags & DTF_NOTIFY)
  { /* flag requested ... */
    if (oot) c->dataStatus = out_of_tolerance;
    oot = TRUE; /* notification will not be suppressed ! */
  }
  return oot;
}
TINE_EXPORT void SetSuspendCallbacks(int value)
{
  if (WaitForSystemMutex(hAttrMutex,-1) == 0)
  { 
    if (value) 
    {
      _gSuspendCallbacksCounter++;
    }
    else
    {
      if (--_gSuspendCallbacksCounter > 0) value = TRUE;
      if (_gSuspendCallbacksCounter < 0) _gSuspendCallbacksCounter = 0; 
    }
    _gSuspendCallbacks = value;
    ReleaseSystemMutex(hAttrMutex); 
  }
}
TINE_EXPORT int GetSuspendCallbacks(void)
{
  return _gSuspendCallbacks;
}
TINE_EXPORT void ResetSuspendCallbacks(void)
{
  _gSuspendCallbacks = FALSE;
  _gSuspendCallbacksCounter = 0;
}
static int hasSuppressedCallback(TCBFCNP cbfcn)
{
  if (!_gSuspendCallbacks) return FALSE;
  if (cbfcn == NULL) return FALSE;
  if (cbfcn == (TCBFCNP)_cbExecLink) return FALSE;
  if (cbfcn == (TCBFCNP)_cbLnkLsn) return FALSE;
  if (cbfcn == (TCBFCNP)wcSingleCallback) return FALSE;
  if (cbfcn == (TCBFCNP)wcGroupCallback) return FALSE;
  return TRUE;
}
void NotifyClient(int i,int cc)
{
  ConTblEntry *c = getConTblEntryPtr(i), *dc;
  ConTblQueue *q = NULL;
  int bmode, fsiz, isBlackListed = FALSE, blacklistIt = FALSE, fireCallback=FALSE;
  int hasData, resetPending=TRUE, isGrouped=FALSE;
  int isTrace = FALSE;
  TCBFCNP cbfcn = NULL;
  RdrTblEntry *r;
# ifndef SMALL_TINE_LIB
  DpdTblEntry *lnk;
  int dlids[100], n, dlid, fid, sid=-1,ndlids=0; /* 100 dependent link ought to be enough for anybody ... */
# endif
  if (c == NULL) return;
  getLinkKey(c);
  _gIsInsideNofityClient = TRUE;
# ifndef SMALL_TINE_LIB
  if (cc == reset_mca_property && !IsMcaParentLink(c))
  { /* direct mca link is being signaled (let one callback thru) */
    msglog(reset_mca_property,"received reset mca property signal for %.192s",c->key);
    c->starttime = (UINT32)time(NULL);
    c->heartbeat = HBT_SUB_PENDING;
  }
# endif
  if (tineDebug > 2)
    dbglog("%.192s <%d>: %s",c->key,cc,c->notifyPending ? "notification" : "not pending");
  if (!c->notifyPending && c->queueLen == 0) 
  { /* notifyPending set initially or (persistent links) following a notification from this routine */
    c->needsNotification = FALSE;
    _gIsInsideNofityClient = FALSE;
    return;
  }
  if (gMarshallBooleanAsByte)
  { /* check for BOOLEANs that might need marshalling */
    if (WaitForSystemMutex(hLinkDataMutex, -1) != 0) return;
    if (c->dataMarshalled != NULL && c->formatOut == CF_BOOLEAN)
    { /* caller's data pointer needs to be updated */
      int i;
      for (i = 0; i < (int)c->sizeOut; i++) ((_Bool*)c->dataMarshalled)[i] = (_Bool)((int*)c->dataPtr)[i];
    }
    ReleaseSystemMutex(hLinkDataMutex);
  }
  isTrace = isTraceLink(i);
  if (_gSuspendCallbacks)
  {
    if (hasSuppressedCallback(c->cbfcn))
    {
      if (isTrace) traceLinkOutput(c, "fireCallback", "callback deferred");
      if (tineDebug > 1) dbglog("%.192s <%d>: deferred as callbacks are suspended",c->key,cc);
      c->needsNotification = TRUE;
      _gIsInsideNofityClient = FALSE;
      return;
    }
    if (tineDebug > 1) dbglog("%.192s <%d>: call updating with callbacks suspended",c->key,cc);
  }
  c->linkStatus = cc; /* ensure this is set properly */
  c->notifyPending = 0;
  c->needsNotification = FALSE;
  cbfcn = c->cbfcn;
  isGrouped = c->isGrouped;

  if (cc != 0)
  { /* non-zero status => do some checking ... */
    if (cc == illegal_property || cc == non_existent || 
        cc == non_existent_elem || cc == non_existent_fec)
    {
      FecAddrStruct *fec = NULL;
      if (c->fecIdx >= 0 || c->fecIdx < numFecTblEntries) fec = &FecTbl[c->fecIdx];
      if ((r=c->rdrLink) != NULL)
      {
        if (r->isGroupMember)
        {
          rmvRdrInfo(c->rdrLink,cc);
          reAcqAddr(c);
          _gIsInsideNofityClient = FALSE;
          if (tineDebug > 1) dbglog("%.192s <%d>: revert from group member to original conditions",c->key,cc);
          return;
        }
      }
      blacklistIt = c->isWildcardLink ? FALSE : TRUE;
      if (cc == non_existent_elem && c->tocounter > 0) blacklistIt = FALSE;
      if (cbfcn == (TCBFCNP)_cbFltrLnk && fec != NULL && strcmp(fec->fecName,GENS_FEC))
      { /* don't blacklist filter links (unless a GENS link) */
        blacklistIt = FALSE; 
      }
      if (blacklistIt)
      {
#       ifndef SMALL_TINE_LIB
        if (c->blcounter++ > 1)
        {
          appendBlackLnkLst(c->ctxName,c->expName,c->devName,c->prpName,cc);
          removeLinkFromGroup(i);
          isBlackListed = TRUE; /* but re-acquire the address for the next round ... */
          if (fec != NULL)
          {
            feclog("Re-acquire (%d) %.16s",i,fec->fecName);
            if (c->fecIdx > 0) fec->inetProtocol = 0; /* forces address re-acquisition */
            fid = locateExport(c->expName,c->ctxName,c->devName,&sid);
            if (fid != c->fecIdx && fid >= 0)
            {
              feclog("correct target FEC for /%.32s/%.32s : %.16s -> %.16s",
                c->ctxName,c->expName,fec->fecName,FecTbl[fid].fecName);
              c->fecIdx = fid;
            }
            if (sid != c->srvIdx && sid >= 0)
            {
              feclog("correct target EQM for /%.32s/%.32s : %.6s -> %.6s",
                c->ctxName,c->expName,SrvTbl[c->srvIdx].EqmName,SrvTbl[sid].EqmName);
              c->srvIdx = sid;
            }
          }
        }
#       else
        dbglog("%.192s cannot be resolved!",getLinkKey(c));
#       endif
      }
    }
    else
    { /* safe to clear this */
      c->blcounter = 0;
    }
    if (!(cc & CE_SENDDATA) && (c->rbModFlags & RB_USE_ERRORVALUE))
    {
      applyErrorValue(c,c->errValue);
    }
  }
  else
  { /* safe to clear this */
    c->blcounter = 0;
  }
  bmode = BASEMODE(c->mode); /* base link mode */
  fireCallback = (cbfcn == NULL || c->cancelPending || c->suppressCallback) ? FALSE : TRUE;
  if (isGrouped && !groupCanNotify(c,&cc)) resetPending = fireCallback = FALSE;
  LastCompletionDataType = c->formatOut;
  LastCompletionDataSize = c->lastDataSize;
  LastIncomingLinkId = i;
  gDataTimeStamp = PutDataTimeStampU(0,c->dtimestamp,c->dtimestampUSEC);
  if (bmode <= CM_SINGLE) c->mode = CM_CANCEL; /* can't do this often enough ! */
  if (fireCallback) 
  { /* call the primary callback */
    if (isTrace)
    {
      char txt[256];
      sprintf(txt,"needs to notify with status %d",cc);
      if (c->mcaIndex > 0) strcat(txt," (via MCA parent)");
      traceLinkOutput(c, "fireCallback", txt);
    }
    if (tineDebug > 2) dbglog("%.192s : fire callback",c->key);
    c->lastStatusReported = cc;
    if (cc == 0 && c->dataStatus == out_of_tolerance) cc = CE_SENDDATA + out_of_tolerance;
    TCBCALL(c,cc);
    if (c->bufCmp != NULL) memcpy(c->bufCmp,c->dataPtr,c->sizeBytesOut);
  }
  if (c->queueLen > 0) c->queueLen--; /* reduce the queue length -> usually sets it back to 0 */
  /* continue checking for dependencies and queues ... */
# if !defined(SMALL_TINE_LIB)
  if (!c->hasNotifiedOnce && gAutoLinkWatchdogs) attachLwdLink(i);
# endif
  hasData = HAS_DATA(cc);
  if (hasData) c->hasNotifiedOnce = TRUE;
  if (isBlackListed) c->linkStatus = link_blacklisted;
# ifndef SMALL_TINE_LIB
  if (WaitForSystemMutex(hDpdLinkMutex,gSystemTick*MUTEX_WAIT_AUGMENT) == 0)
  { /* make a note of the dependent link ids ... */
    for (lnk=c->dpdLink; lnk != NULL && ndlids < 100; lnk=lnk->nxt) dlids[ndlids++] = lnk->linkId;
    ReleaseSystemMutex(hDpdLinkMutex);
  }
  for (n=0; n<ndlids; n++)
  { /* check for dependent links ... */
    dlid = dlids[n];
    dc = getConTblEntryPtr(dlid);
    if (dc == NULL) continue;
    if (dc->mode == CM_CANCEL)
    { /* cleanup: look for any new stranded cancelations */
      removeDependentLink(dlid);
      continue;
    }
    if (hasData) copyDpdData(dc,c);
    dc->lastDataSize = c->lastDataSize;
    dc->dataStatus = c->dataStatus;
    dc->linkStatus = c->linkStatus;
    dc->xferReason = c->xferReason;
    dc->dtimestamp = c->dtimestamp;
    dc->dtimestampUSEC = c->dtimestampUSEC;
    dc->systemstamp = c->systemstamp;
    dc->datastamp = c->datastamp;
    dc->hasNotifiedOnce = c->hasNotifiedOnce;
    dc->suppressCallback = !isOOT(dc, dc->sizeBytesOut);
    dc->tocounter = c->tocounter;
    fireCallback = (dc->suppressCallback || dc->cbfcn == NULL) ? FALSE : TRUE;
    if (dc->isGrouped && !groupCanNotify(dc,&cc)) resetPending = fireCallback = FALSE;
    if (fireCallback) 
    {
      if (tineDebug > 1) dbglog("%.192s : fire dependent link callback",dc->key);
      if (cc == 0 && c->dataStatus == out_of_tolerance) cc = CE_SENDDATA + out_of_tolerance;
      TCBCALL(dc,cc);
      if (dc->bufCmp != NULL) memcpy(dc->bufCmp,dc->dataPtr,dc->sizeBytesOut);
    }
    if (bmode <= CM_SINGLE || BASEMODE(dc->mode) <= CM_SINGLE) 
    { /* cancel it here too */
      dc->mode = CM_CANCEL; 
    }
  }
# endif /* !SMALL_TINE_LIB */
  if (!resetPending) 
  {
    _gIsInsideNofityClient = FALSE;
    return;
  }
  for (i=0; i<nConnectionTableEntries; i++) 
  { /* run thru the list and check for queues and reset the notifyPending flag */
    if ((dc=getConTblEntryPtr(i)) == NULL) continue;
    if ((fsiz=fmtsizeof(dc->formatOut)) == 0) fsiz = 1;
    if (dc->cbfcn == cbfcn)
    { /* in case of grouped calls, make sure they're all in a wait state */
      if (c != dc && /* different connection entries with same callback */
          (!isGrouped || !dc->isGrouped)) /* not grouped ! */
        continue; /* so skip it (as they need to notify separately) */
      dc->notifyPending = 1;
      if (dc->queueLen > 0 && WaitForSystemMutex(hLinkQueueMutex,gSystemTick) == 0)
      { /* there's a queue  */
        while (dc->queueLen > 0)
        { /* -> work thru it */
          if ((q=dc->queue) == NULL) 
          { /* shouldn't be possible */
            dc->queueLen = 0;
            continue; 
          }
          fireCallback = FALSE;
          if (dc->mode == CM_CANCEL) dc->suppressCallback = TRUE;
          if (dc->cbfcn != NULL && q->dataOut != NULL)
          { /* prepare the data and call the callback */
            memcpy(dc->dataPtr,q->dataOut,q->dataSize);
            dc->lastDataSize = q->dataSize/fsiz;
            dc->dataStatus = q->dataStatus;
            dc->linkStatus = q->statusCode;
            dc->dtimestamp = q->timeStamp;
            dc->dtimestampUSEC = q->timeStampUSEC;
            dc->systemstamp = q->sysStamp;
            dc->datastamp = q->usrStamp;
            dc->suppressCallback = !isOOT(dc,dc->sizeBytesOut);
            if (!dc->suppressCallback) fireCallback = TRUE;
            SystemFree(dc->queue->dataOut);
          }
          dc->queueLen--;
          q = q->next;
          SystemFree(dc->queue);
          dc->queue = q;
          if (fireCallback) 
          {
            if (tineDebug > 1) dbglog("%.192s : fire queued callback",dc->key);
            cc = dc->linkStatus;
            if (cc == 0 && dc->dataStatus == out_of_tolerance) cc = CE_SENDDATA + out_of_tolerance;
            TCBCALL(dc,cc);
            if (dc->bufCmp != NULL) memcpy(dc->bufCmp,dc->dataPtr,dc->sizeBytesOut);
          }
        }
        ReleaseSystemMutex(hLinkQueueMutex);
      }
    }
  }
  assertHeapOkay("NotifyClient",c->devName,c->prpName);
  if ((c->linkStatus &0x0fff) == information_static) 
  {
    if (tineDebug) dbglog("%.192s : closing link to static property",c->key);
    _closeLink(c->linkId);
  }
  if (tineDebug > 2) dbglog("%.192s : return from notification",c->key);
  _gIsInsideNofityClient = FALSE;
}
int gClnRecvQueueDepth = DEFAULT_RECV_QUEUE_DEPTH;
void dbgbowout(char *txt,int id,ConTblEntry *c,UINT32 ts)
{
# ifdef OS_16BIT /* 16-bit OS needs the %ld, 64-bit OS can get confused with %ld ! */
  sprintf(dbgbuf,"discard subscription %d %ld",id,ts);
  if (c != NULL) sprintf(&dbgbuf[strlen(dbgbuf)],"(%.32s %.64s %.64s %ld)",c->expName,c->devName,c->prpName,c->dtimestamp);
# else
  sprintf(dbgbuf,"discard subscription %d %d",id,ts);
  if (c != NULL) sprintf(&dbgbuf[strlen(dbgbuf)],"(%.32s %.64s %.64s %d)",c->expName,c->devName,c->prpName,c->dtimestamp);
# endif
  sprintf(&dbgbuf[strlen(dbgbuf)],": %s",txt);
  dbglog(dbgbuf);
}
#define bowout(txt) {\
  if (tineDebug || errorDebug) dbgbowout(txt,i,c,subhdr.timestamp);\
  if (needXferMutex) ReleaseSystemMutex(hXferMutex), needXferMutex = FALSE;\
  goto bow;\
}
#define qbowout(txt,ok) {\
   ReleaseSystemMutex(hLinkQueueMutex);\
   if (!ok) c->queueLen--; c->blksin = 0; c->bytesin = 0;\
   if (tineDebug && !ok) dbgbowout(txt,i,c,subhdr.timestamp);\
   goto qbow;\
}
int prepRedirString(ConTblEntry *c,char *rdr,char *str)
{
  char *prpstr, *srvstr, *bpstr, *mpstr;
  switch (c->tineProtocol)
  {
    case 5:
      memset(rdr,0,STATUS_SIZE);
      strncpy(&rdr[PRP_REDIR_SRV_OFFSET],str,16);
      strncpy(&rdr[PRP_REDIR_PRP_OFFSET],&str[16],16);
      if (strlen(c->ctxName) > 0) strncpy(&rdr[PRP_REDIR_CTX_OFFSET],c->ctxName,CONTEXT_NAME_SIZE);
      strncpy(&rdr[PRP_REDIR_DEV_OFFSET],c->devName,DEVICE_NAME_SIZE);
      break;
    case 6:
    case 7:
      memcpy(rdr,str,STATUS_SIZE);
      if (rdr[PRP_REDIR_CTX_OFFSET] == 0) strncpy(&rdr[PRP_REDIR_CTX_OFFSET],c->ctxName,CONTEXT_NAME_SIZE);
      if (rdr[PRP_REDIR_DEV_OFFSET] == 0) strncpy(&rdr[PRP_REDIR_DEV_OFFSET],c->devName,DEVICE_NAME_SIZE);
      break;
    default:
      return illegal_protocol;
  }
  prpstr = &rdr[PRP_REDIR_PRP_OFFSET];
  if (strlen(prpstr) == 0) 
  { /* use the calling property name as is */
    strncpy(prpstr,c->prpName,PROPERTY_NAME_SIZE);
  }
  else
  { /* add the meta modifier back on if necessary */
    srvstr = &rdr[PRP_REDIR_SRV_OFFSET];
    if (!IsMetaProperty(prpstr,NULL,NULL,TRUE) && 
         IsMetaProperty(c->prpName,&bpstr,&mpstr,TRUE) &&
        !IsMetaToHistory(c->prpName,srvstr) &&
         strlen(prpstr) < PROPERTY_NAME_SIZE - 16)
    {
      strcat(prpstr,".");
      strncat(prpstr,mpstr,15);
    }
  }
  return 0;
}
TINE_EXPORT int LinkDataMemcpy(void *dst,void *src,size_t sizeInBytes)
{
  if (WaitForSystemMutex(hLinkDataMutex,-1) != 0) return mutex_error;
  memcpy(dst,src,sizeInBytes);
  ReleaseSystemMutex(hLinkDataMutex);
  return 0;
}
TINE_EXPORT int LockLinkData(void)
{
  return WaitForSystemMutex(hLinkDataMutex,-1);
}
TINE_EXPORT int UnlockLinkData(void)
{
  return ReleaseSystemMutex(hLinkDataMutex);
}
TINE_EXPORT int ReassignLinkData(int id,void *buffer,UINT32 length)
{
  int cc = 0;
  ConTblEntry *c;

  if (id < 1 || id >= nConnectionTableEntries) return invalid_index;
  if ((c=conTbl[id]) == NULL) return link_error;
  if (WaitForSystemMutex(hLinkDataMutex,-1) != 0) return mutex_error;
  if (c->allocatedBytes) ccerr(already_assigned);
  if (c->dataPtr == NULL && buffer != NULL) ccerr(data_size_mismatch);
  if (c->dataPtr != NULL && buffer == NULL) ccerr(data_size_mismatch);
  if (c->sizeBytesOut > length) ccerr(buffer_too_small);
  c->dataPtr = (char *)buffer;
  feclog("reassigned link data buffer for link %d : %.196s",id,c->key);
err:
  ReleaseSystemMutex(hLinkDataMutex);
  return cc;
}
extern struct structFormat *MDAHeaderStruct;
void prepIncomingData(BYTE *dst,BYTE *src,UINT32 dsiz,UINT32 rsiz,short fmt,short endianness,ConTblEntry *c)
{
  struct structStruct *s=NULL;
  UINT32 hsiz = 0, fsiz = getFormatSize(fmt);
  BYTE *bptr=NULL;
  int i, n, nsa, nstructs = 0;
  UINT32 p, xsiz, xoff;
  char **sa, *sb, key[128];

  UNUSED(endianness); /* use at some future date ... */
  if (dsiz == 0 || fsiz == 0) return;
  if (WaitForSystemMutex(hLinkDataMutex,-1) != 0) 
  {
    dbglog("cannot get link data mutex!");
    return;
  }
  if (rsiz == 0)
  { /* make sure the last data size is zeroed ! */
    if (c != NULL) c->lastDataSize = 0;
    goto out;
  }
  if (fmt == CF_STRUCT && c != NULL && (s=findstructEx(c->strTagOut,c->ctxName,c->expName)))
  { /* registered tagged structure */
    sprintf(key,"/%.32s/%.32s",c->ctxName,c->expName);
    nstructs = _taggedStructSwap(c->strTagOut,key,src,dst,RTOL,(int)rsiz/s->rawlen,s->size);
  }
  else if (isAdjustableLengthFormat(fmt))
  { /* a volatile data set (output data contains pointers) */
    UINT32 vtlCap = fmtsizeof(c->formatOut) * c->sizeOut;
    vtlCap += getFormatHeaderSize(c->formatOut);
    if (c == NULL) goto out;
    if ((n=atoi(c->strTagOut)) < 1) goto out;
    if (dsiz > vtlCap) 
    { /* re-assign the capacity */
      SystemFree(c->bufVtl);
    }
    if (c->bufVtl == NULL) c->bufVtl = (char *)SystemCalloc(1,(size_t)(vtlCap+STATUS_SIZE));
    if (c->bufVtl != NULL)
    { /* copy into the buffer seen by the client and reassign src pointer */
      memcpy(c->bufVtl,src,MIN(dsiz,rsiz));
      src = (BYTE *)c->bufVtl;
    }
    switch (fmt)
    {
      case CF_KEYVALUE:
      case CF_STRING:
        sa = (char **)dst; sb = (char *)src; nsa = 0;
        for (i=0,p=0; i<n; i++)
        { /* scan for terminating '0's*/
          sa[i] = &sb[p];
          p += (int)strlen(sa[i])+1;
          nsa++;
          if (p >= rsiz) break;
        }
        c->lastDataSize = nsa;
        goto out;
      case CF_AIMAGE:
        hsiz = getFormatHeaderSize(fmt);
        for (i=0; i<n; i++)
        {
          SwapBytes((char *)&((DIMAGE *)src)[i],1,fmt,RTOL);
          memcpy(&((DIMAGE *)dst)[i],&src[i*(hsiz+8)],hsiz);
          memcpy(&xoff,&src[i*(hsiz+8)+hsiz],4);
          memcpy(&xsiz,&src[i*(hsiz+8)+hsiz+4],4);
          ((DIMAGE *)dst)[i].frameBufferSize = xsiz;
          ((DIMAGE *)dst)[i].frameBuffer = &src[xoff];
          if (i == 0) n = xoff/(hsiz+8);
        }
        c->lastDataSize = i;
        goto out;
      case CF_ASPECTRUM:
        hsiz = getFormatHeaderSize(fmt);
        for (i=0; i<n; i++)
        {
          SwapBytes((char *)&((ASPECTRUM *)src)[i],1,fmt,RTOL);
          memcpy(&((ASPECTRUM *)dst)[i],&src[i*(hsiz+8)],hsiz);
          memcpy(&xoff,&src[i*(hsiz+8)+hsiz],4);
          memcpy(&xsiz,&src[i*(hsiz+8)+hsiz+4],4); 
          xsiz /= sizeof(float);
          ((ASPECTRUM *)dst)[i].spectBufferSize = xsiz;
          SwapBytes((char *)&src[xoff],xsiz,CF_FLOAT,RTOL);
          ((ASPECTRUM *)dst)[i].spectBuffer = (float *)&src[xoff];
          if (i == 0) n = xoff/(hsiz+8);
        }
        c->lastDataSize = i;
        goto out;
    }
  }
  else if (fmt == CF_HISTORY)
  {
    if (c == NULL) goto out;
    switch (ftoi(c->strTagOut))
    {
      case CF_IMAGE:
        fsiz = 0;
        c->lastDataSize = 0;
        while (hsiz < rsiz)
        { /* keep going until the end */
          memcpy(&fsiz,&src[HISTORY_HDR_SIZE+hsiz]+IMAGE_HDR_SIZE-sizeof(SINT32),sizeof(SINT32));
#         ifndef INTEL
          ISWAP(fsiz);
#         endif
          fsiz += HISTORY_HDR_SIZE + IMAGE_HDR_SIZE;
          if (fsiz > rsiz)
          { /* bail out */
            dbglog("invalid image frame save in history call!");
            goto out;
          }
          cfHistorySwap(c->strTagOut,&src[hsiz],&dst[hsiz],RTOL,fsiz);
          hsiz += fsiz;
          c->lastDataSize++;
        }
        goto out;
      case CF_STRING:
      case CF_KEYVALUE:
        c->lastDataSize = 0;
        fsiz = 0;
        while (fsiz < rsiz)
        { /* scan thru the bytes */
          hsiz = HISTORY_HDR_SIZE + (int)strlen((char *)&src[HISTORY_HDR_SIZE+fsiz])+1;
          cfHistorySwap(c->strTagOut,(BYTE *)&src[fsiz],(BYTE *)&dst[fsiz],RTOL,hsiz);
          fsiz += hsiz;
          c->lastDataSize++;
        }
        goto out;
      default: /* normal case */
        cfHistorySwap(c->strTagOut,src,dst,RTOL,rsiz);
        break;
    }
    fsiz = HISTORY_HDR_SIZE + getCarriedFormatSize(c->strTagOut);
  }
  else 
  { /* relatively normal format */
    if ((hsiz=GetFormatHeaderSize(fmt)) > 0)
    { /* need to trap formats with headers and memcpy to right place */
      bptr = GetFormatBodyPtr(dst,fmt);
      if (c == NULL)
      { /* a server and not a client getting incoming data : body follows header! */
        bptr = dst + GetFormatHeaderSize(fmt);
      }
    }
    if (fmt == CF_MDA)
    {
      DMDA *mda = (DMDA *)dst;
      fmt = ftoi(c->strTagOut);
      fsiz = fmtsizeof(fmt);
      _structSwap(MDAHeaderStruct,src,0,NULL,dst,RTOL);
      memcpy(mda->data.bptr,&src[MDA_HDR_SIZE],(rsiz-hsiz));
      SwapBytes(mda->data.cptr,rsiz/fsiz,fmt,RTOL);
    }
    else if (bptr == NULL)
    { /* normal case */
      memcpy(dst,src,MIN(dsiz,rsiz));
      SwapBytes((char *)dst,rsiz/fsiz,fmt,RTOL);
      if (dsiz > rsiz) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);
    }
    else
    { /* destination body does not follow the format header */
      memcpy(dst,src,hsiz);
      SwapBytes((char *)dst,hsiz,fmt,RTOL);
      if (rsiz > hsiz && bptr > (BYTE *)0x1000)
      { /* a modicum of protection against a bad bptr from above */
        memcpy(bptr,&src[hsiz],rsiz-hsiz);
        SwapBytes((char *)bptr,(rsiz-hsiz)/fsiz,GetFormatDataType(fmt),RTOL);
      }
    }
  }
  if (c != NULL) 
  { /* has a connection table entry */
    if (rsiz > hsiz)
    {
      rsiz -= hsiz;
      c->lastDataSize = rsiz/fsiz;
    }
    else
    {
      c->lastDataSize = 0;
    }
    if (s != NULL && s->rawlen > 0) c->lastDataSize = nstructs;
  }
out:
  ReleaseSystemMutex(hLinkDataMutex);
  return;
}
#ifndef SMALL_TINE_LIB
int assertBufferCapacity(ConTblEntry *c,int numblks,int mtu,int isVolatile)
{
  UINT32 ssiz = numblks*mtu+STATUS_SIZE; /* worst case */
  struct structStruct *ss;
  if (c->formatOut == CF_STRUCT && !isENSCall(c->key))
  {
    if ((ss=findstructEx(c->strTagOut,c->ctxName,c->expName)) == NULL) return 0;
    if (ss->hasExtSpace == 0) return 0;
  }
  if (isVolatile && c->sizeOut < ssiz) 
  { /* free and null the memory for subsequent re-allocation */
    c->sizeOut = ssiz;
    SystemFree(c->bufOut);
  }
  if (c->numblks < (UINT32)numblks)
  {
    int nb;
    SystemFree(c->bufOutBlks);
    nb = numblks; nb += (8 - nb%8);
    if (!c->bufOutBlks && !(c->bufOutBlks=(BYTE *)SystemCalloc(nb,1))) return out_of_client_memory;
  }
  c->numblks = (UINT32)numblks;
  /* so far : always return success */
  return 0;
}
int hasVolatileDataSize(int fmt,char *tag,char *ctx,char *srv)
{
  struct structStruct *ss;
  switch (fmt)
  {
    case CF_AIMAGE:
    case CF_ASPECTRUM:
    case CF_KEYVALUE:
    case CF_STRING: 
      return TRUE;
    case CF_STRUCT:
      if ((ss=findstructEx(tag,ctx,srv)) != NULL && ss->hasExtSpace)
      {
        return TRUE;
      }
    default: break;
  }
  return FALSE;
}
#endif

#define linkWaitingOnStartup(c) ((c)->tocounter == -1)
/**
 * \internal check the initial (single-delivery) response for contract coercion 
 * \param specs is returned from prepIncomingResponseHdr().
 * \param stream is the incoming bytestream positioned after the response header
 * \param padr is passed from doProcessData and points to the parent address
 * \param inetprot is the incomind protocol (UDP, TCP, etc.)
 * \return TRUE if the associated contract has been coerced and the calling routine
 * doProcessData() should return;  FALSE indicates that doProcessData() should continue.
 */
int requestIsCoerced(RspSpecs *specs, BYTE *stream, void *padr, int inetprot)
{
  DTYPE dout, din;
  char namstr[196], devstr[132], *pctx, *psrv, *pdev, *pprp;
  struct timeval tv_stamp;
  ConTblEntry *c = NULL;
  int cc, rc, stepsize, idx, i = specs->id;
# if !defined(SMALL_TINE_LIB)
  int pid;
# endif
  if (specs->starttime == BCAST_ID) return FALSE;
  gettimeofday(&tv_stamp, (struct timezone *)NULL);
  switch (specs->cc)
  {
    case illegal_protocol:
      if ((idx=getIncomingFecIndex(padr, inetprot)) < 0)
      {
        msglog(invalid_index, "illegal protocol : invalid fec index %d !", idx);
        ccerr(invalid_index);
      }
      for (i=0; i<nConnectionTableEntries; i++)
      {
        if ((c=conTbl[i]) != NULL && c->fecIdx == idx)
        {
          if (c->heartbeat == HBT_SUB_PENDING) continue;
          if (BASEMODE(c->mode) == CM_CANCEL) continue;
          switch (c->tineProtocol)
          {
            case 6: /* is a legacy server ? */
              msglog(illegal_protocol, "server %.32s appears to be Release 3 vintage", c->expName);
            case 7:
              c->tineProtocol--;
              if (FecTbl[c->fecIdx].tineProtocol > c->tineProtocol)
                FecTbl[c->fecIdx].tineProtocol = c->tineProtocol;
              c->heartbeat = HBT_SUB_PENDING;
              msglog(illegal_protocol,"%.32s %.64s %.64s : switch to protocol %d", c->expName, c->devName, c->prpName, c->tineProtocol);
              feclog("%.32s %.64s %.64s : switch to protocol %d", c->expName, c->devName, c->prpName, c->tineProtocol);
              if (c->mode & CM_MCAST)
              { /* join legacy mcast group */
                attachPortToMulticastGroup(mcastClnSck, (UINT16)gtMCastPort, tMCastAddrLegacy, "server common");
              }
              break;
            default:
              if (IsConnProt(c->inetProtocol) && c->tineProtocol == 5)
              { /* corrupt ? */
                msglog(illegal_protocol, "%.96s : protocol %d is not valid (closing connection: corrupt?)",
                  GetLinkName(i), c->tineProtocol);
                closeIPConnectionToFec(c->fecIdx, c->tcpSck);
                c->heartbeat = HBT_SUB_PENDING;
              }
              else
              {
                msglog(illegal_protocol, "%.96s : illegal protocol %d cannot be decremented (closing link!)",
                  GetLinkName(i), c->tineProtocol);
                c->mode = CM_CANCEL;
                c->heartbeat = HBT_NONE;
              }
              break;
          }
          c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
        }
      }
      return TRUE;
    case get_subscription_id:
    {
      int idx = getIncomingFecIndex(padr, inetprot);
      if (idx < 0) ccerr(invalid_index);
      if (i < 0 || i >= nConnectionTableEntries) ccerr(invalid_link);
      if ((c = conTbl[i]) == NULL) ccerr(invalid_link);
      stepsize = SUBRSP_STEPSIZE(c->tineProtocol);
      c->subId = SPEEK(*(short *)&stream[stepsize]);
      if (tineDebug) dbglog("BCAST sub id %d from %.16s", c->subId, FecTbl[idx].fecName);
      c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
      return TRUE;
    }
    default:
      break;
  }
  if (i < 0 || i >= nConnectionTableEntries) ccerr(non_existent);
  if ((c = conTbl[i]) == NULL) ccerr(invalid_link);
  if (c->mode == CM_CANCEL)
  {
    msglog(invalid_link, "incoming link %.192s has been canceled !", c->key);
    ccerr(invalid_link);
  }
  switch (specs->cc)
  { /* check coercive status codes up front (they should be single-delivery) */
    case call_redirection:
    case server_redirection:
    {
      RdrTblEntry *rdrLnk = NULL;
      int m, isGrpMem = FALSE, isWaitingOnStartup = FALSE, isWaiting = FALSE;
      if (specs->starttime != c->starttime)
      {
        msglog(invalid_link, "%.192s server redirection : invalid start time !", c->key);
        return TRUE;
      }
      stepsize = SUBRSP_STEPSIZE(c->tineProtocol);
      prepRedirString(c, namstr, (char *)&stream[stepsize]);
      pctx = &namstr[PRP_REDIR_CTX_OFFSET]; psrv = &namstr[PRP_REDIR_SRV_OFFSET];
      pdev = &namstr[PRP_REDIR_DEV_OFFSET]; pprp = &namstr[PRP_REDIR_PRP_OFFSET];
      if (strlen(pctx) > 0) sprintf(devstr, "/%.32s/%.32s/%.64s", pctx, psrv, pdev);
      else sprintf(devstr, "%.32s/%.64s", psrv, pdev);
      DTYPEZERO(dout); DTYPEZERO(din);
      dout.dArrayLength = getAssignedDataLength(c, FALSE);
      dout.dFormat = c->formatOut;
      memcpy(dout.dTag, c->strTagOut, TAG_NAME_SIZE);
      dout.data.cptr = c->dataPtr;
      din.dArrayLength = getAssignedDataLength(c, TRUE);
      din.dFormat = c->formatIn;
      memcpy(din.dTag, c->strTagIn, TAG_NAME_SIZE);
      din.data.cptr = c->dataIn;
      msglog(0, "redirect %.32s/%.64s %.64s to %.128s %.64s", c->expName, c->devName, c->prpName, devstr, pprp);
      if (c->fecIdx >= 0 && strcmp(FecTbl[c->fecIdx].fecName, GENS_FEC) == 0)
      { /* update local group db cache */
        isGrpMem = TRUE;
  #       if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
        addToGroupCacheList(c->ctxName, c->expName, psrv, c->devName);
  #       endif
      }
      isWaiting = c->isWaitingSyncNotification;
      isWaitingOnStartup = linkWaitingOnStartup(c);
      c->rdrId = i;
      c->notifyPending = 0;
      c->tcpSck = 0;
      m = c->mode;
      rdrLnk = appendRedirectedLinkList(c->ctxName, c->expName, c->prpName, c->devName, pctx, psrv, pdev, pprp, isGrpMem);
      if ((i=_attachLink(devstr, pprp, &dout, &din, c->access, (int)c->pollingInterval,
        (void *)c->cbfcn, m, c->cbId, c->cbRef, c->rdrId)) < 0)
      { /* error creating link will have called the callback once */
        rc = -i;
        msglog(rc, "cannot attach to redirection link %.132s[%.64s] : %.32s", devstr, pprp, cc2str(rc));
        c->linkStatus = rc;
        c->cbfcn = NULL;
        break;
      }
      else
      {
        if (isWaitingOnStartup) conTbl[i]->tocounter = -1;
        conTbl[i]->isWaitingSyncNotification = isWaiting;
        ReturnToClientCycle++;
      }
      c->rdrLink = rdrLnk;
      c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
      return TRUE;
    }
    case invalid_interval:
    {
      int intvl = specs->sysstamp;
      if (intvl < 10) intvl = 1000; /* huh ? */
      if (intvl > 30000) intvl = 30000; /* huh ? */
      msglog(0, "%.32s %.64s %.64s : adjust polling interval from %d to %d msec",
        c->expName, c->devName, c->prpName, c->pollingInterval, intvl);
      c->pollingInterval = intvl; /* contained in system stamp */
      c->heartbeat = HBT_SUB_PENDING;
      return TRUE;
    }
    case invalid_datarequest:
    {
      int fmt = specs->userstamp;
      int doutsize = specs->sysstamp;
      if (fmt == (int)c->formatOut && doutsize < (int)c->sizeOut)
      { /* this can be handled without any fancy footwork ... */
        msglog(0, "%.32s %.64s %.64s : adjust output data length from %d to %d",
          c->expName, c->devName, c->prpName, c->sizeOut, doutsize);
        c->sizeOut = (UINT32)doutsize;
        c->heartbeat = HBT_SUB_PENDING;
        return TRUE;
      }
#     ifndef SMALL_TINE_LIB
      c->relinked = RELINKED_NEEDED | RELINKED_INVALIDDATA;
      appendReLnkLst(c, c->strTagOut, fmt, doutsize, specs->cc);
      c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
      return TRUE;
#     else
      break;
#     endif
    }
#   if !defined(SMALL_TINE_LIB)
    case tcp_access_required:
      c->mode &= ~CM_NETWORK;
      c->mode |= CM_CONNECT;
      c->heartbeat = HBT_SUB_PENDING;
      c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
      return TRUE;
    case async_access_required:
      if (BASEMODE(c->mode) == CM_SINGLE  && c->cbfcn != (TCBFCNP)_cbExecLink) goto lnkLsnStart;
      break;
    case mcast_access_required:
      if (ListenForMulticasts() != 0) break;
      if (BASEMODE(c->mode) != CM_SINGLE)
      { /* a monitor link */
        INADR inaddr = fixMcastHaddr((BYTE *)&FecTbl[c->fecIdx].netAdr, !(FecTbl[c->fecIdx].addrFlags & ADDRFLAG_MCAST_MODERN));
        attachPortToMulticastGroup(mcastClnSck, (UINT16)gtMCastPort, inaddr, "server specific");
        c->mcastGrp = inaddr;
        c->mode &= ~(CM_CONNECT | CM_STREAM);
        c->mode |= CM_NETWORK;
        c->heartbeat = HBT_SUB_PENDING;
        c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
        return TRUE;
      }
      else if (c->cbfcn != (TCBFCNP)_cbExecLink)
      { /* an asynchronous CM_SINGLE link ! */
      lnkLsnStart:
        if (!allowCommonLinks) break;
        c->timeLastSync = (UINT32)time(NULL);
        c->mode = CM_CANCEL;
        DTYPEZERO(dout);
        dout.dFormat = c->formatOut;
        dout.dArrayLength = c->sizeOut;
        strncpy(dout.dTag, c->strTagOut, TAG_NAME_SIZE);
        dout.data.vptr = NULL;
        sprintf(devstr, "/%.32s/%.32s/%.64s", c->ctxName, c->expName, c->devName);
        pid = _attachLink(devstr, c->prpName, &dout, NULL, c->access, 1000, (void *)_cbLnkLsn,
          (CM_TIMER|CM_NETWORK), UNASSIGNED_CALLBACKID, NULL, UNASSIGNED_INDEX);
        if (pid == i)
        { /* something went wrong ! */
          c->mode = CM_CANCEL;
          break;
        }
        c->mode = CM_SINGLE;
        c->boundToId = pid;
        addDependentLink(i);
        return TRUE;
      }
      break;
    case has_structure_tag:
    case has_bitfield_tag:
    {
      int fmt;
      char *tag;
      stepsize = SUBRSP_STEPSIZE(c->tineProtocol);
      tag = (char *)&stream[stepsize];
      if (specs->cc == has_structure_tag)
      {
        fmt = CF_STRUCT;
        c->relinked = RELINKED_NEEDED | RELINKED_STRUCT;
      }
      else
      {
        fmt = GetBitfieldFormat(c->formatOut);
        c->relinked = RELINKED_NEEDED | RELINKED_BITFIELD;
      }
      appendReLnkLst(c, tag, fmt, c->sizeOut, specs->cc);
      c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
      return TRUE;
    }
    case property_is_mca_elevated:
    case property_is_mca:
    {
      McaTblEntry *mca;
      ConTblEntry *cmca;
      int mcaLength = specs->sysstamp; /* contained in system stamp */
      if (c->formatOut == CF_DEFAULT)
      { /* need to assign the format type as well */
        c->formatOut = LFMT(specs->fmt);
        c->sizeOut = 1; /* always true for an MCA element */
      }
      if (mcaLength < 1)
      { /* gotta be robust ! */
        rc = code_failure;
        break;
      }
      if (mcaLength < (int)specs->userstamp)
      { /* mca, but non-existent device */
        rc = illegal_device;
        break;
      }
      c->mcaIndex = specs->userstamp + 1; /* contained in user stamp */
      stepsize = SUBRSP_STEPSIZE(c->tineProtocol);
      pdev = "#0";
      if (!strncmp((char *)&stream[stepsize], MCADEV_TAG, 8))
      { /* post-release 4.1.5 will do it this way */
        pdev = (char *)&stream[stepsize + 8];
      }
      dout.dArrayLength = mcaLength;
      dout.dFormat = c->formatOut;
      memcpy(dout.dTag, c->strTagOut, TAG_NAME_SIZE);
      dout.data.vptr = NULL; /* signal to allocate */
      din.dArrayLength = c->sizeIn;
      din.dFormat = c->formatIn;
      memcpy(din.dTag, c->strTagIn, TAG_NAME_SIZE);
      din.data.cptr = c->dataIn;
      msglog(0, "reassign /%.32s/%.32s/%.64s %.64s to multi-channel array", c->ctxName, c->expName, c->devName, c->prpName);
      c->notifyPending = 0;
      c->queueDepth = 0;
      if (isTraceLink(i)) traceLinkOutput(c, "processData", "reassign to multi-channel array");
      c->tcpSck = 0;
      if (WaitForSystemMutex(hMcaLinkMutex, gSystemTick) != 0) return TRUE;
      mca = getMcaLinkEntry(c->ctxName, c->expName, c->prpName, pdev);
      if (mca == NULL)
      { /* not yet allocated */
        int isElvFmt = (specs->cc == property_is_mca_elevated) ? TRUE : FALSE;
        if ((mca=addMcaLinkEntry(c->ctxName, c->expName, c->prpName, pdev, mcaLength, isElvFmt)) == NULL)
        {
          msglog(code_failure, "cannot add mca link for /%.32s/%.32s %.64s (out of memory?)",
            c->ctxName, c->expName, c->prpName);
          rc = code_failure;
          ReleaseSystemMutex(hMcaLinkMutex);
          break;
        }
        if (isElvFmt) dout.dFormat = getElevatedMcaFormat(dout.dFormat);
        sprintf(devstr, "/%.32s/%.32s/%.64s", c->ctxName, c->expName, pdev);
        if ((mca->linkId = _attachLink(devstr, c->prpName, &dout, &din, c->access, (int)c->pollingInterval,
          (void *)_cbMcaXRef, c->mode, UNASSIGNED_CALLBACKID, mca, UNASSIGNED_INDEX)) < 0)
        {
          msglog(-mca->linkId, "cannot attach mca link for /%.32s/%.32s/%.64s %.64s (%.32s)",
            c->ctxName, c->expName, pdev, c->prpName, erlst[-mca->linkId]);
          rc = code_failure;
          ReleaseSystemMutex(hMcaLinkMutex);
          break;
        }
        conTbl[mca->linkId]->mcaEntry = mca;
      }
      else
      { /* do we need to up the polling rate ? */
        if (c->pollingInterval < conTbl[mca->linkId]->pollingInterval)
        {
          conTbl[mca->linkId]->pollingInterval = c->pollingInterval;
          conTbl[mca->linkId]->heartbeat = HBT_SUB_PENDING;
        }
        conTbl[mca->linkId]->cancelPending = FALSE;
      }
      /* bind the original link to the mca link */
      if (conTbl[mca->linkId]->hasNotifiedOnce)
      { /* make sure a callback ensues (for the late-comers) */
        cmca = conTbl[mca->linkId]; /* simplify notation by assignment */
        if (c->mcaIndex > 0 && c->mcaIndex <= (int)cmca->sizeOut)
        { /* update the data contents in case of 'waiting' */
          if (putMcaValues(mca, cmca, c))
          {
            c->linkStatus = cmca->linkStatus;
            c->lastLnkErrSrc = cmca->lastLnkErrSrc;
            strncpy(c->lastStatusString, cmca->lastStatusString, STATUS_SIZE);
          }
        }
        cmca->needsNotification = TRUE;
        cmca->suppressCallback = FALSE;
        c->notifyPending = 1;
        NotifyClient(i, c->linkStatus);
        if (linkWaitingOnStartup(c)) _cbExecLink(c->linkId, c->linkStatus); /* stop any 'waiting' */
      }
      c->tocounter = 0; 
      if (c->renewalThreshold > 0) conTbl[mca->linkId]->renewalThreshold = c->renewalThreshold;
      c->mcaLink = mca;
      c->boundToId = mca->linkId; /* now bound to the mca parent link */
      c->mode = CM_REGISTER;
      c->heartbeat = HBT_NONE;
      c->timeLastSync = (UINT32)tv_stamp.tv_sec; c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec / 1000;
      c->notifyPending = 1;
      if (gIsRunningAsServer && gAutoLinkErrorAlarms)
      { /* remove any link error alarms */
        clearFecLinkErrorAlarm(c->linkId);
      }
      addDependentLinkToMca(mca, i);
      if (c->dpdLink != NULL)
      { /* possible dependencies prior to MCA discovery */
        movDpdToMca(c, mca);
      }
      ReleaseSystemMutex(hMcaLinkMutex);
      return TRUE;
    }
  # endif /* !defined(SMALL_TINE_LIB) */
    default:
      break;
  }
  return FALSE; /* signal that the calling routine should continue */
err:
  if (cc && tineDebug)
  {
    rc = cc & 0xfff;
    if (strlen(gLastStatusString) == 0 && rc > 0 && rc < numErr)
      strncpy(gLastStatusString, erlst[rc], STATUS_SHORTSIZE);
    dbglog("doProcessData <%d> : %.32s", cc, gLastStatusString);
  }
  return TRUE; /* signal that the calling routine should return */
}


int gClnCommunicationLockout = FALSE;
UINT32 doProcessData(UINT32 bufsize,BYTE *buf,void *padr,int inetprot,char *ptag)
{
  short i,nb,fmt,cc=0,rc=0;
  short idx,prot,baseMode,renewalCutoff,syncNotify=FALSE,suppressNotify=FALSE,infoStatic=FALSE;
  short ismcast=FALSE,isVolatile=FALSE,hasBadPackets=FALSE;
  short needXferMutex=FALSE;
  int isTrace = FALSE;
  int fecidx,srvidx,extbytes;
  UINT16 stepsize,fmtsize;
  UINT32 extsize, extptr, stssize, msgsize, bytesread = 0, totalsize;
  UINT32 datasize,dcmpsize,dataPtr,doutsize,dretsize,ipeek;
  RspHdr rsphdr;
  RspSpecs specs;
  SubRspHdr subhdr;
  ConTblEntry *c = NULL;
  struct timeval tv_stamp;                
  BYTE *srcbuf;
# ifndef SMALL_TINE_LIB
  UINT32 tsize;
# endif

  if (gClnCommunicationLockout) return 0;
  gettimeofday(&tv_stamp,(struct timezone *)NULL);
  gInAdr.prot = inetprot;
  if ((padr != NULL) && (inetprot & (UDP|TCP|STREAM|PIPE|UDP6|TCP6|STREAM6))) 
  { /* those with an ip address ... */
    gInAdr.ip = *(SCKADR *)padr;
    memset(&gInAdr.ipx,0,sizeof(IpxAdr));
  }
  else
  { /* zero it out */
    memset(&gInAdr.ip,0,sizeof(SCKADR));
  }
  nArrivals++;
  prot = prepIncomingResponseHdr(&rsphdr, &specs, buf, bufsize);
  if (prot < 5 || prot > DEFAULT_TRANSPORT_PROTOCOL_LEVEL) ccerr(illegal_protocol);
  totalsize = rsphdr.totalSizeInBytes - rsphdr.headerSizeInBytes;
  buf += rsphdr.headerSizeInBytes;

  if (requestIsCoerced(&specs, buf, padr, inetprot)) return bytesread;

  memset(gLastStatusString,0,STATUS_SIZE);
  if (specs.starttime != BCAST_ID) 
  { /* PHDR_PROT_OFFSET only avail for prot > 5 */
    if ((i=specs.id) >= 0 && i < nConnectionTableEntries)
    {
      if ((c=conTbl[i]) == NULL) ccerr(invalid_link);
    }
    else
    {
      if (tineDebug || errorDebug) dbglog("link id %d not in connection table!", i);
      ccerr(un_allocated);
    }
  }
  stepsize = SUBRSP_STEPSIZE(prot);

  while (bytesread < totalsize)
  { /* work thru the payload */
    prepIncomingSubResponseHdr(prot,&subhdr,&buf[bytesread]);
    msgsize = (bufsize == 0) ? subhdr.msgSize : totalsize;
    if (msgsize == 0) 
    { /* corrupt contract coming in */
      if (tineDebug || errorDebug) dbglog("corrupt contract entry discarded");
      return bytesread;
    }
    stssize = subhdr.stssize;
    if (subhdr.starttime == BCAST_ID)   
    { /* BCAST or MCAST contract */
      ismcast = TRUE;
      if ((idx=getIncomingFecIndex(padr,inetprot)) < 0) 
      { /* is this multi/broad-cast coming from someone I know ? */
        return bytesread;
      }
      for (i=0; i<nConnectionTableEntries; i++)
      { /* is this multi/broad-cast something I've asked for ? */
        if (conTbl[i] == NULL) continue;
        if (BASEMODE(conTbl[i]->mode) == CM_CANCEL) continue;
        if (conTbl[i]->cancelPending) continue;
        if (conTbl[i]->fecIdx != idx) continue;
        if (conTbl[i]->subId != subhdr.subId) continue;
        /* YES: sync with local contract */
        subhdr.subId = i; 
        subhdr.starttime = conTbl[i]->starttime;
        break;
      }
      if (i == nConnectionTableEntries) goto bow;/* not interested in this one */
    }
    i = subhdr.subId;
    if (i < 0 || i >= nConnectionTableEntries) { hasBadPackets = TRUE; bowout("invalid id"); }
    if ((c=getConTblEntryPtr(i)) == NULL) bowout("link id not valid");
    if (c->starttime != subhdr.starttime) { hasBadPackets = TRUE; bowout("invalid starttime"); }
    if ((baseMode=BASEMODE(c->mode)) == CM_CANCEL) 
    { /* link entry has been canceled ! */
      if (c->counter > CTR_REMINDER)
      { /* the server missed the CANCEL request? => re-submit the cancel request */
        if (cc != invalid_link && !ismcast && c->tcpSck == (SOCKET)0) c->heartbeat = HBT_SUB_PENDING;
      }
      bowout("canceled");
    }
    if (c->rbModFlags&RB_USE_TIMESTAMP)
    { /* instruction to apply the local timestamp regardless of what comes in */
      subhdr.timestamp = (UINT32)tv_stamp.tv_sec;
      subhdr.timestampUSec = (UINT32)tv_stamp.tv_usec;
    }
    if (subhdr.timestamp < c->dtimestamp) bowout("stale");
    if (c->fecIdx < 0 || c->fecIdx >= numFecTblEntries) bowout("invalid fec cross-reference");
    if (rsphdr.endianness == BO_BIG_ENDIAN) FecTbl[c->fecIdx].addrFlags |= ADDRFLAG_BIGENDIAN;
#   ifndef SMALL_TINE_LIB
    isVolatile = hasVolatileDataSize(c->formatOut,c->strTagOut,c->ctxName,c->expName);
    if (isENSCall(c->key) || isVolatile)
    {
      if (assertBufferCapacity(c,subhdr.numblks,subhdr.mtu,isVolatile) != 0) bowout("no data buffer capacity");
    }
    if (subhdr.numblks > c->numblks) bowout("unexpectedly large incoming data set");
#   endif
    fmtsize = c->formatOut == CF_MDA ? getCarriedFormatSize(c->strTagOut) : fmtsizeof(c->formatOut);
    doutsize = msgsize - stepsize;
    datasize = fmtsize*c->sizeOut;
    datasize += getFormatHeaderSize(c->formatOut);
    dcmpsize = datasize;
    if (dcmpsize%2) dcmpsize++; /* even boundaries */
    if (tineDebug > 2 || (tineDebug && baseMode <= CM_SINGLE && subhdr.numblks <= 4))
    {
      char *fecprot = ptag != NULL ? ptag : getProtocolAsString(inetprot);
      dbglog("%s (p%d %d of %d)-> %.32s %.64s [%s] (+%d) %d vals (%d)<%d>",
	      fecprot,prot,subhdr.blknum,subhdr.numblks,
	      c->expName,c->prpName,gFecAddrBuffer,subhdr.counter,
	      (int)c->sizeOut,c->formatOut,
	      (int)(subhdr.retCode & 0xfff));
    }
    memset(gLastStatusString,0,STATUS_SIZE);
    if (c->lastDataSize == 0) c->lastDataSize = c->sizeOut; 
    extsize = 0; /* reset this each time */
    if (gIsRunningAsServer && gAutoLinkErrorAlarms)
    { /* remove any link error alarms (refer only to link_timeout) */
      clearFecLinkErrorAlarm(i);
    }
    switch (cc=subhdr.retCode)
    { /* deal with special status codes */
      case reacquire_address:
        c->heartbeat = HBT_SUB_PENDING;
        FecTbl[c->fecIdx].inetProtocol = 0; /* forces address re-acquisition */
        locateExport(SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmContext,c->devName,NULL);
        if (tineDebug) dbglog("received reacquire address for /%.32s/%.32s",
          SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName);
        goto bow;
      case non_existent_elem:
        if (strncmp(c->expName,ENS_TAG,EXPORT_NAME_SIZE))
        { /* not the name server target */
          c->tocounter++; /* similar enough to a link timeout */
          if (c->tocounter > TO_THRESHOLD*3 &&
              allLinksAboveThreshold(c->fecIdx) &&
              fecAddrAcqId == -1)
          {
            reAcqAddr(c);
            c->tocounter = 0;
          }
        }
        if (c->tocounter == 1)
        { /* the initial 'non_existent_elem' return: re-check the cached address */
          feclog("%.32s/%.64s %.64s -> %.32s [%.6s] %.16s [%s]: %.32s",
            c->expName,c->devName,c->prpName,
            SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmName,
            FecTbl[c->fecIdx].fecName,gFecAddrBuffer,cc2str(cc));
          if (strnicmp(c->expName,SrvTbl[c->srvIdx].ExportName,EXPORT_NAME_SIZE))
          { /* not equal! connection table is corrupted, try to recover */
            fecidx = locateExport(c->expName,c->ctxName,c->devName,&srvidx);
            if (srvidx >= 0 && fecidx >= 0)
            {
              feclog("correct /%.32s/%.32s entry %.6s @ %.16s -> %.6s @ %.16s",
                c->ctxName,c->expName,SrvTbl[c->srvIdx].EqmName,FecTbl[c->fecIdx].fecName,
                SrvTbl[srvidx].EqmName,FecTbl[fecidx].fecName);
              c->fecIdx = fecidx; 
              c->srvIdx = srvidx;
            }
          }
        }
      case property_is_mca_elevated:
      case property_is_mca:
      case call_redirection:
      case server_redirection:
        if (cc != non_existent_elem) 
        { /* remember we're falling thru from the top */
          cc = subhdr.retCode = rc;
          strcpy((char *)&buf[bytesread+msgsize-(int)stssize],"redirection error");
        }
      default: /* non-zero */
        if ((cc & 0x0fff) == information_static) infoStatic = TRUE;
        if (doutsize < stssize && c->blksin == 0)
        {
          c->bytesin = 0;
          bowout("partial error block");
        }
        if (!(subhdr.retCode & CE_SENDDATA)) c->lastDataSize = datasize = 0;
        if (datasize == 0) subhdr.blknum = 1; /* must be true */
        extsize = 0;
        /* doutsize = what came in in this sub hdr; datasize = expected total */
        extbytes = (int)(doutsize * subhdr.blknum) - (int)datasize;
        if (subhdr.blknum == subhdr.numblks || extbytes > 0)
        {
          if (c->linkStatus == link_blacklisted) cc = link_blacklisted;
		      memset(c->lastStatusString,0,STATUS_SIZE);
          if (extbytes > 0)
          { /* 2nd to last block had some room or everything fit into 1 */
            extsize = (UINT32)extbytes;
            extptr = 0;
          }
          else
          { /* last block contains the last piece (or all) of the error string */
            extsize = stssize;
            if (doutsize < stssize)
            { /* just the last piece */
              if (subhdr.numblks == 1) bowout("short data block");
              extsize = doutsize;
            }
            extptr = stssize - extsize;
            msglog(out_of_sequence, "partial status message : %d extended bytes beginning at %d", extsize, extptr);
          }
          if (extsize > stssize - extptr)
          {
            msglog(system_error, "incoming response claims %d extended bytes beginning at %d",extsize,extptr);
            bowout("corrupt status message");
          }
          memcpy(&c->lastStatusString[(int)extptr],&buf[bytesread+msgsize-(int)extsize],(size_t)extsize);
          doutsize -= (UINT16)extsize;
          c->bytesin += (UINT16)extsize;
        }
        memcpy(gLastStatusString,&c->lastStatusString,STATUS_SIZE);
        if (!(subhdr.retCode & CE_SENDDATA)) break;
      case 0:
        if (fmtsize == 0) break;
        if (cc == 0)
        { /* remember: we could be 'falling thru' ... */
          strncpy(gLastStatusString, "success", STATUS_SIZE);
          strncpy(c->lastStatusString, "success", STATUS_SIZE);
        }
        if (dcmpsize < doutsize)
        { /* the segment that just came is bigger than the total expected ! */
          if (cc == 0) strncpy(c->lastStatusString,erlst[data_size_mismatch],32);
          cc = data_size_mismatch;
          if (tineDebug || errorDebug) dbglog("%.192s data size mismatch: %lu vs %lu",c->key,datasize,doutsize);
        }
        else
        {
          needXferMutex = (c->heartbeat == HBT_SUB_PENDING) ? TRUE : FALSE;
          if (needXferMutex) if (WaitForSystemMutex(hXferMutex,gSystemTick) != 0) { needXferMutex = FALSE; bowout("can't get xfer mutex"); }
          if (subhdr.numblks == 1 && c->counter == subhdr.counter) bowout("phantom block");
          if (!c->blksin)
          { /* new set */
            if (subhdr.counter > c->counter)
            { /* I'm getting a renewal set: check for new renewal conditions */
              if (c->renewalMultiplier > 1 && subhdr.counter == CTR_RENEWAL)
                c->renewalMultiplier = 1;
            }
            if (c->renewalMultiplier < 2 && subhdr.counter > CTR_RENEWAL) 
            { /* the server has established its own renewal count */
              c->renewalMultiplier = subhdr.counter/CTR_RENEWAL;
            }
            c->counter = subhdr.counter;
            if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
          }
          if (c->counter != subhdr.counter)   
          { /* we're being passed */
            if (tineDebug || errorDebug) dbglog("%.192s transfer block out of sequence: %d vs %d",c->key,subhdr.counter,c->counter);
            nPartialTransfers++;
            if (subhdr.counter < c->counter || 
                subhdr.counter > c->counter + 30 ||
                subhdr.counter == CTR_RENEWAL * c->renewalMultiplier)
            { /* more recent set */
              if (tineDebug) dbglog("reset sequence counter");
              c->blksin = 0; c->bytesin = 0;
              c->counter = subhdr.counter;       
              if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
            }
            else bowout("old data block"); /* old stuff coming in */
          }
          if (needXferMutex) ReleaseSystemMutex(hXferMutex), needXferMutex = FALSE;
          if (subhdr.numblks > 1 || isVolatile)
          { /* multiple blocks : need to collect the data blocks */
            if (!c->bufOut && !(c->bufOut=(char *)SystemCalloc(1,(size_t)(datasize+STATUS_SIZE)))) bowout("no local memory");
            nb = c->numblks; nb += (8 - nb%8);
            if (!c->bufOutBlks && !(c->bufOutBlks=(BYTE *)SystemCalloc(nb,1))) bowout("no local memory");
            if (subhdr.blknum > c->numblks) bowout("unexpected block");
            if (c->bufOutBlks[subhdr.blknum-1]) bowout("phantom block");
            c->bufOutBlks[subhdr.blknum-1] = 1;
            dataPtr = (subhdr.blknum-1)*subhdr.mtu;
            if (doutsize > 0) memcpy(&c->bufOut[(int)dataPtr],&buf[bytesread+stepsize],(size_t)doutsize);
            srcbuf = (BYTE *)c->bufOut;
            ReturnToClientCycle = subhdr.numblks - c->blksin;
          }
          else 
          { /* it's all here in this single data block */
            dataPtr = 0; 
            srcbuf = &buf[bytesread+stepsize];
          }
          c->bytesin += doutsize;
          if (++(c->blksin) < subhdr.numblks) goto bow;
          ReturnToClientCycle = 0;
          dretsize = c->bytesin - (cc == 0 ? 0 : stssize);
          if (dretsize <= 0) 
          {
            c->lastDataSize = 0;
            break;
          }
          if (c->formatOut == CF_DEFAULT)
          { /* record the default format and size */
            c->formatOut = LFMT(subhdr.dataFormat); 
            if ((fmtsize=fmtsizeof(c->formatOut)) == 0) bowout("illegal format");
            if (c->formatOut == CF_MDA)
            {
              DMDA *mda = (DMDA *)c->dataPtr;
              memcpy(&ipeek,&srcbuf[MDA_FMT_OFFSET],4); 
#             if !defined(INTEL) && !defined(VMS)
              ISWAP(ipeek);
#             endif
              fmt = (short)ipeek;
              strncpy(c->strTagOut,itof(fmt),TAG_NAME_SIZE);
              fmtsize *= fmtsizeof(fmt);
              mda->data.bptr = (BYTE *)SystemCalloc(1,(size_t)(dretsize-getFormatHeaderSize(c->formatOut)));
            }
            c->sizeOut = (prot >= MINIMUM_REQHDR_PROTOCOL) ? 
                            subhdr.dataSize : /* official contract size */
                            (dretsize-getFormatHeaderSize(c->formatOut))/fmtsize; /* best guess */
          }
          if (c->dataPtr == NULL || srcbuf == NULL)
          {
            cc = code_failure;
            break;
          }
          if (c->queueDepth > 0 && c->queueLen++ > 0)
          { /* we've got another one and have not yet finished with this one -> queue it */
            ConTblQueue *q=NULL,**nxt=NULL;
            if (WaitForSystemMutex(hLinkQueueMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) 
            { /* can't get the link queue mutex ? */
              c->queueLen--;
              bowout("can't queue entry");
            }
#           ifndef SMALL_TINE_LIB
            c->needsNotification = TRUE; /* avoid a trap in case this is already fixed at FALSE */
            if (c->queueLen > c->queueDepth) qbowout("queue full",FALSE);
            if (c->formatOut == CF_STRING || c->isWildcardLink || c->cbfcn == (TCBFCNP)wcGroupCallback) qbowout("queue not allowed",FALSE);
            if ((q=(ConTblQueue *)SystemCalloc(1,sizeof(ConTblQueue))) == NULL) qbowout("no memory for queue",FALSE);
            if ((q->dataOut=(BYTE *)SystemCalloc(1,dretsize)) == NULL) qbowout("no memory for queue data",FALSE);
            q->dataStatus = subhdr.stscode;
            q->statusCode = cc;
            q->sysStamp = subhdr.sysstamp;
            q->usrStamp = subhdr.userstamp;
            q->xfrReason = subhdr.xferReason;
            q->timeStamp = subhdr.timestamp;
            q->timeStampUSEC = subhdr.timestampUSec;
            tsize = c->lastDataSize; /* save */
            prepIncomingData(q->dataOut,srcbuf,(UINT32)dretsize,(UINT32)dretsize,c->formatOut,rsphdr.endianness,c);
            q->dataSize = dretsize;
            c->lastDataSize = tsize; /* restore */
            for (nxt=&(c->queue); *nxt != NULL; nxt = &((*nxt)->next));
            *nxt = q;
#           endif /* !SMALL_TINE_LIB */
            qbowout("entry queued",TRUE);
          }
          if (c->isInsideCallback) bowout("is inside callback");
          prepIncomingData((BYTE *)c->dataPtr,srcbuf,(UINT32)datasize,(UINT32)dretsize,c->formatOut,rsphdr.endianness,c);
          c->dataStatus = subhdr.stscode;
          suppressNotify = !isOOT(c,datasize);
#         ifndef SMALL_TINE_LIB
          if (IsMcaParentLink(c)) suppressNotify = FALSE;
#         endif
          if (!c->hasNotifiedOnce) suppressNotify = FALSE; /* hasn't yet called in */ 
          if (c->linkStatus != 0) suppressNotify = FALSE; /* pre-existing link status not 0 */
          if (tineDebug > 1)
          {
#           ifdef OS_16BIT
            dbglog("%.192s [%s] (+%d) data set received (%lu values)",c->key,gFecAddrBuffer,subhdr.counter,c->lastDataSize);
#           else
            dbglog("%.192s [%s] (+%d) data set received (%u values)",c->key,gFecAddrBuffer,subhdr.counter,c->lastDataSize);
#           endif
            if (tineDebug > 3) DataDump("data : ",c->dataPtr,(int)c->sizeOut,c->formatOut,RTOL);
          }
        }
        break;
      case use_stream_transport:
	      bowout("use stream transport");
        break;
    }
    lastLnkErr = cc;
    c->lastLnkErrSrc = lastLnkErrSrc = LNK_ERROR_RMT;
    syncNotify = (c->mode & CM_SYNCNOTIFY) ? TRUE : FALSE;
    isTrace = isTraceLink(c->linkId);
    switch (baseMode)
    {
      default:
      case CM_SINGLE : 
        if (isTrace) traceLinkOutput(c, "processData", "mode SINGLE");
        if (c->wcLink == NULL)
        {
          c->mode = CM_CANCEL;  /* got reply */
          if (isTrace) traceLinkOutput(c, "processData", "cancel a SINGLE link following incoming data");
        }
        if (c->cbfcn != (TCBFCNP)_cbExecLink) checkAddressAcquisition(c->fecIdx,0);
		    break;
      case CM_EVENT:
        if (isTrace) traceLinkOutput(c, "processData", "mode EVENT");
        suppressNotify = (subhdr.xferReason & (CX_EVENT|CX_MESSAGE|CX_ERROR)) ? FALSE : TRUE;
      case CM_DATACHANGE: 
        if (isTrace) traceLinkOutput(c, "processData", "mode DATACHANGE");
        c->heartbeat = 2*HBT_MINIMUM + HBT_MAXIMUM;
		    break;
      case CM_TIMER   : 
        if (isTrace) traceLinkOutput(c, "processData", "mode TIMER");
        c->heartbeat = MAX((UINT32)HBT_MINIMUM,c->pollingInterval/1000);
		    break;
    }
    if (c->linkStatus == notification_pending) suppressNotify = FALSE; /* if waiting for the 1st one never suppress ! */
    if (cc == non_existent_elem)
    {
      if (baseMode > CM_SINGLE && (time_t)c->timeLastSync != tv_stamp.tv_sec) c->heartbeat = HBT_SUB_PENDING;
    }
    else 
    {
      if (linkWaitingOnStartup(c)) _cbExecLink(c->linkId,cc); /* stop waiting */
      c->tocounter = 0;
    }
    c->xferReason = subhdr.xferReason;
    c->linkStatus = cc;
    c->counter = subhdr.counter;
    c->dtimestamp = subhdr.timestamp;
    c->dtimestampUSEC = subhdr.timestampUSec;
    c->timeLastSync = (UINT32)tv_stamp.tv_sec;
    c->timeMsecLastSync = (UINT32)tv_stamp.tv_usec/1000;
    c->datastamp = subhdr.userstamp;
    c->systemstamp = subhdr.sysstamp;
    c->blksin = 0; c->bytesin = 0;
    if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
    if (SystemLocalDataTimeSynchronization) 
    { /* don't let the incoming data be too far in the future ! */
      gDataTimeStamp = makeDataTimeStamp();
      if (c->dtimestamp > (UINT32)(gDataTimeStamp) + 1) 
      {
#       ifdef OS_16BIT
        if (tineDebug) dbglog("adjusting incoming data timestamp for %.32s/%.64s[%.64s] from %ld to %ld",
#       else
        if (tineDebug) dbglog("adjusting incoming data timestamp for %.32s/%.64s[%.64s] from %d to %d",
#       endif
          c->expName,c->devName,c->prpName,c->dtimestamp,(SINT32)gDataTimeStamp);
        c->dtimestamp = (SINT32)gDataTimeStamp;
        c->dtimestampUSEC = 0;
      }
    }
    if (c->linkStatus != 0) suppressNotify = FALSE; /* this link status not 0 */
    if (c->linkStatus != c->lastStatusReported) suppressNotify = FALSE; /* link status changed since last callback */
    c->suppressCallback = (UINT8) suppressNotify;
    c->needsNotification = TRUE;
    c->notifyPending = 1;
    if (syncNotify) NotifyClient(i,cc);
qbow:
    if (baseMode > CM_SINGLE && subhdr.counter > 0 && c->linkStatus != link_blacklisted && !infoStatic)
    {
      renewalCutoff = getSubsRenewalCutoff(c);
      if (subhdr.counter < (renewalCutoff/2)) 
      { /* if no activity here : get serious ! */
        c->heartbeat = HBT_SUB_PENDING;
        needToAcknowledge = TRUE;
      }
      else if (subhdr.counter <= renewalCutoff) 
      { /* n.b. HBT_ACK_PENDING > HBT_SUB_PENDING ! */
        if (!c->hasAcknowledged)
        { /* only renew if not flagged */
          c->heartbeat = HBT_ACK_PENDING;
          needToAcknowledge = TRUE;
        }
        c->hasAcknowledged = TRUE;
      }
      else
      { /* counter is in the normal area: turn this flag off */
        c->hasAcknowledged = FALSE;
      }
    }
bow:
    if (hasBadPackets) gBadPacketCnts++;
    if ((bytesread += msgsize) <= 0) return bytesread;
  }
  return bytesread;
err:
  if (cc && tineDebug)
  {
    rc = cc & 0xfff;
    if (strlen(gLastStatusString) == 0 && rc > 0 && rc < numErr)
      strncpy(gLastStatusString,erlst[rc],STATUS_SHORTSIZE);
    dbglog("doProcessData <%d> : %.32s",cc,gLastStatusString);
  }
  return bytesread;
}
#ifndef SMALL_TINE_LIB
void mapAndCallbackData(ConTblEntry *c)
{
  ReLnkCbEntry *cb;
  ReLnkTblEntry *r = c->rlkEntry;
  int dlen, n;
  if (r == NULL) return;
  if (r->lnkFmt == c->formatOut)
  { /* then it's only a matter of array length */
    if (r->lnkSiz != (int)c->sizeOut)
    { /* by all rights, these should not be equal ! */
      dlen = fmtsizeof(r->lnkFmt) * MIN(r->lnkSiz,(int)c->sizeOut);
      dlen += getFormatHeaderSize(r->lnkFmt);
      memcpy(r->dat,c->dataPtr,dlen);
    }
  }
  else
  { /* a bit more complicated */
    n = MIN(r->lnkSiz,(int)c->sizeOut);
    reformat(r->dat,r->lnkFmt,(BYTE *)c->dataPtr,c->formatOut,n);
  }
  for (cb=r->cbLst; cb != NULL; cb=cb->nxt)
  {
    if (cb->lnkId == c->linkId)
    {
      TCBCALL(cb,c->linkStatus);
      break;
    }
  }
}
void _relinkXrefCb(int id,int cc)
{
  ReLnkTblEntry *r;
  ConTblEntry *c;
  UNUSED(cc);
  if (id < 0 || id >= nConnectionTableEntries) return;
  if ((c=conTbl[id]) == NULL) return;
  if ((r=getReLnkTblEntry(c)) == NULL) return;
  switch (r->sts)
  {
    case has_structure_tag:
      if (c->formatOut == CF_STRUCT) mapAndCallbackStructField(c);
      break;
    case has_bitfield_tag:
      mapAndCallbackBitfield(c);
      break;
    case invalid_datarequest:
      mapAndCallbackData(c);
      break;
  }
}
#endif /* !SMALL_TINE_LIB */
#ifndef GLOBAL_SYNCHRONIZATION_TOLERANCE
# define GLOBAL_SYNCHRONIZATION_TOLERANCE 0.1
#endif
double gSyncTolerance = GLOBAL_SYNCHRONIZATION_TOLERANCE;
double gSyncTimestamp = 0;
#define TIMESYNC_DEPTH 6
#define TIMESYNC_NOSYNC 0
#define TIMESYNC_TRACKED 1
#define TIMESYNC_MONOTONIC 2
#define TIMESYNC_MONTH (3600L * 24L * 30L)
double gMaximumAllowedTimeCorrection = (double)TIMESYNC_MONTH;
void cbGlobalSynchronization(int id, int cc)
{
  int applyCorrection = FALSE, validCorrection = TIMESYNC_NOSYNC;
  int i,k,p,n;
  double correction = 0, interim_correction, myTimeStamp;
  double delta_t = 0, delta_rb, delta_ts, last_dt = 0, last_rb;
  static int last_cc = -1, last_ptr = 0, ncorrections = 0;
  static double last_logged = 0;
  static double last_dataTimeOffset = 0, largest_correction = 0, last_correction = 0;
  static double last_dataTimeRbDeltas[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };
  static double last_dataTimeDtDeltas[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };
  static double last_dataTimeStamps[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };
  static double last_dataTimeRbTimes[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };

  UNUSED(id);
  if (!useGlobalSynchronization) 
  { /* has been turned off after the fact */
    return;
  }
  if (cc != last_cc)
  {
    last_cc = cc;
    feclogEx( ((cc)?TINE_LOGLEVEL_WARN:TINE_LOGLEVEL_INFO), "global synchronization : %.32s",cc2str(cc));
  }
  if (cc) return;

  last_rb = last_dataTimeRbTimes[last_ptr];
  last_dt = last_dataTimeStamps[last_ptr];

  if (gSyncTimestamp == last_dt) 
  { /* extra broadcast/multicast of the same */
    if (tineDebug > 1) dbglog("discard extra synchronization timestamp");
    return; 
  }
  myTimeStamp = makeDataTimeStamp();
  last_ptr = (last_ptr + 1) % TIMESYNC_DEPTH;
  last_dataTimeRbTimes[last_ptr] = myTimeStamp;               /* my clock's current timestamp */
  last_dataTimeRbDeltas[last_ptr] = myTimeStamp - last_rb;    /* time difference based on my clock */
  last_dataTimeStamps[last_ptr] = gSyncTimestamp;             /* remote clock's current timestamp */
  last_dataTimeDtDeltas[last_ptr] = gSyncTimestamp - last_dt; /* time difference based on remote clock */
  n = 0;
  for (i=0; i<TIMESYNC_DEPTH; i++)
  {
    k = (last_ptr + i) % TIMESYNC_DEPTH;
    p = (k - 1 + TIMESYNC_DEPTH) % TIMESYNC_DEPTH;
    if (last_dataTimeStamps[k] < 1000) return; /* not a valid UTC timestamp */
    if (last_dataTimeStamps[p] < 1000) return; /* not a valid UTC timestamp */
    delta_rb = last_dataTimeRbDeltas[k] - last_dataTimeRbDeltas[p];
    delta_ts = last_dataTimeDtDeltas[k] - last_dataTimeDtDeltas[p];
    if (k != last_ptr + 1)
    {
      if (delta_rb > 10.0 || delta_rb < -10.0) return; /* just getting started ? */
      if (delta_ts > 10.0 || delta_ts < -10.0) return;
    }
    delta_t = last_dataTimeRbDeltas[k] - last_dataTimeDtDeltas[k];
    if (validCorrection != TIMESYNC_MONOTONIC && fabs(delta_t) < 0.01) 
    { /* both local and remote clocks are tracking smoothly */
      validCorrection = TIMESYNC_TRACKED;
    }
    else if (validCorrection != TIMESYNC_TRACKED && fabs(delta_rb) < 0.01 && fabs(delta_ts) < 0.01)
    { /* both clocks are increasing monotonically */
      validCorrection = TIMESYNC_MONOTONIC;
    }
    else
    { /* these deltas should be really close if the data is coming in and being read at the correct rate */
      return;
    }
    interim_correction = last_dataTimeStamps[k] - last_dataTimeRbTimes[k];
    if (interim_correction > gSyncTimestamp) return; /* something that might happen on VxWorks */
    correction += interim_correction;
    n++;
    if (validCorrection == TIMESYNC_MONOTONIC && i == TIMESYNC_DEPTH-2) break;
  }
  if (n == 0) return;
  correction /= n;
  if (fabs(correction) > gMaximumAllowedTimeCorrection)
  { 
    dbglog("time correction %g would exceed %g seconds (%g months) -> ignored !",
      correction,gMaximumAllowedTimeCorrection,(gMaximumAllowedTimeCorrection/(double)TIMESYNC_MONTH));
    return;
  }
  if (correction > gSyncTimestamp && fabs(correction-gDataTimeStampOffset) < 1000)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "invalid time correction %g greater than time server clock %g",correction,gSyncTimestamp);
    return;
  }
  if (correction > gSyncTolerance || correction < -gSyncTolerance) applyCorrection = TRUE;
  else if (last_correction < gSyncTolerance && correction > gSyncTolerance/2) applyCorrection = TRUE;
  else if (last_correction > gSyncTolerance && correction < -gSyncTolerance/2) applyCorrection = TRUE;
  if (!applyCorrection) return;
  gDataTimeStampOffset += correction;
  last_correction = correction;
  if (tineDebug) 
    dbglog("synchronization correction (%s): %g sec (current offset %g sec)",
      (validCorrection == TIMESYNC_TRACKED ? "tracking" : "monotonic"),correction,gDataTimeStampOffset);
  ncorrections++;
  if (fabs(correction) > largest_correction) largest_correction = fabs(correction);
  if (myTimeStamp < last_logged + 3600) return;
  last_logged = gDataTimeStamp;
  if (fabs(gDataTimeStampOffset-last_dataTimeOffset) < 0.1) return;
  feclog("global synchronization offset : %g sec",gDataTimeStampOffset);
  feclog("largest correction : %g sec  (nr. corrections %d)",largest_correction,ncorrections);
  last_logged = gDataTimeStamp; ncorrections = 0; largest_correction = 0;
  last_dataTimeOffset = gDataTimeStampOffset;
}
void setDataTimeStampOffset(double value)
{
  gDataTimeStampOffset = value;
}
int SystemStartGlobalSynchronization(void)
{
  static int glbSyncStarted = FALSE, cc = 0;
  int id, retry_count = 0;
  DTYPE dout;

  if (glbSyncStarted) return cc;
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = &gSyncTimestamp;
  dout.dTag[0] = 0;
retryRecvGlobal:
  id = recvNetGlobalEx("SYSTIME",&dout,cbGlobalSynchronization,UNASSIGNED_CALLBACKID);
  if (id < 0) ccerr(-id);
err:
  if (cc == semaphore_busy && ++retry_count < 5) goto retryRecvGlobal;
  feclog("initialize global synchronization : %.32s",cc2str(cc));
  glbSyncStarted = TRUE;
  return cc;
}
void cbTimeSynchronization(int id,int cc)
{
  gDataTimeStamp = makeDataTimeStamp();
  cbGlobalSynchronization(id,cc);
}
TINE_EXPORT int SystemStartTimeSynchronization(void)
{
  int cc = 0;
  DTYPE dout;
  char timeSrv[72];

  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = &gSyncTimestamp;
  dout.dTag[0] = 0;
  sprintf(timeSrv,"/SITE/%.32s",TIME_SRV);
  cc = AttachLink(timeSrv,"SYSTIME",&dout,NULL,CA_READ,5000,cbTimeSynchronization,CM_TIMER);
  if (cc < 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not start time synchronization : %.32s",erlst[-cc]);
  return cc < 0 ? -cc : 0;
}
TINE_EXPORT int SystemGetSiteTimestamp(double *timeStamp)
{
  DTYPE dout;
  char timeSrv[72];

  if (timeStamp == NULL) return argument_list_error;
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = timeStamp;
  dout.dTag[0] = 0;
  sprintf(timeSrv,"/SITE/%.32s",TIME_SRV);
  return ExecLinkEx(timeSrv,"SYSTIME",&dout,NULL,CA_READ,500);
}
#if !defined(SMALL_TINE_LIB)
CycTrg *gCycTrgLst = NULL;
TINE_EXPORT int RegisterCycleTriggerFunction(CYCBFCNP fcn,char *eqm,char *prpLst,void *reference)
{
  CycTrg *ct;
  for (ct=gCycTrgLst; ct != NULL; ct=ct->nxt)
  {
    if (ct->fcn == fcn && ct->ref == reference) return already_assigned;
  }
  if ((ct=(CycTrg *)SystemCalloc(1,sizeof(CycTrg))) == NULL) return out_of_local_memory;
  ct->fcn = fcn;
  if (prpLst != NULL && eqm != NULL)
  {
    strncpy(ct->eqm,eqm,EQM_NAME_SIZE);
    strncpy(ct->prpLst,prpLst,256);
    ct->prpLst[255] = 0; /* terminate! */
  }
  ct->ref = reference;
  ct->nxt = gCycTrgLst;
  gCycTrgLst = ct;
  feclog("register new machine cycle trigger function @%x",fcn);
  return 0;
}
TINE_EXPORT int UnregisterCycleTriggerFunction(CYCBFCNP fcn,void *reference)
{
  CycTrg *ct,*ct0=gCycTrgLst;
  for (ct=ct0; ct != NULL; ct0=ct,ct=ct->nxt)
  {
    if (ct->fcn == fcn && ct->ref == reference)
    {
      if (ct0 != gCycTrgLst) ct0->nxt=ct->nxt; else gCycTrgLst = ct->nxt;
      SystemFree(ct);
      feclog("unregister machine cycle trigger function @%x",fcn);
      return 0;
    }
  }
  return un_allocated;
}
SINT32 gCycleNumber = 0;
SINT32 gCycleOffset = 0;
TINE_EXPORT void SetSystemStampOffset(int cycleOffset)
{
  gCycleOffset = cycleOffset;
}
TINE_EXPORT int GetSystemStampOffset(void)
{
  return gCycleOffset;
}
TINE_EXPORT void SetSystemStampDelay(int cycleDelay)
{
  if (cycleDelay < 0) cycleDelay = 0;
  gSystemDataStampDly = cycleDelay;
}
TINE_EXPORT int GetSystemStampDelay(void)
{
  return gSystemDataStampDly;
}
void applySystemDataStamp(void)
{
  CycTrg *ct;
  gSystemDataStamp = gSystemDataStampRb;
  for (ct=gCycTrgLst; ct != NULL; ct=ct->nxt)
  { /* now schedule if necessary */
    SystemScheduleProperty(ct->eqm,ct->prpLst);
  }
}
void cbCycleTrigger(int id,int cc)
{
  CycTrg *ct;
  static int lastCycleNumber = -1;
  UNUSED(id);
  if (!isUsingCycleTrigger) return;
  if (gCycleNumber == lastCycleNumber) return;
  lastCycleNumber = gCycleNumber;
  if (WaitForSystemMutex(hCycleTriggerMutex,-1) != 0) ccerr(mutex_error); 
  for (ct=gCycTrgLst; ct != NULL; ct=ct->nxt)
  { /* call all trigger routines prior to setting the cycle number */
    if (ct->fcn != NULL) ct->fcn(gCycleNumber,cc,ct->ref);
  }
  /* set the cycle number */
  gSystemDataStampRb = gCycleOffset + gCycleNumber;
  gettimeofday(&gSystemDataStampTs,(struct timezone *)NULL);/* Cycle time */
  ReleaseSystemMutex(hCycleTriggerMutex);
  if (gSystemDataStampDly <= 0) applySystemDataStamp();
err:
  if (cc && tineDebug > 0) dbglog("machine cycle synchronization : %.32s",cc2str(cc));
}
char gCyclerNumberKey[GLOBALS_KEY_SIZE];
int assertTriggerKeyExists(const char *txt,const char *target)
{
  int cc = 0;
  char key[GLOBALS_KEY_SIZE], dsrv[128], rb[64], *c, *ctx, *srv, *kwd;
  ExportListStruct *el;
  ExpDataStruct srvaddr;
  FecAddrStruct fecaddr;
  DTYPE dout, din;
  /* MUST be of the form /<context>/<server>/<keyword> */
  strncpy(key,target,GLOBALS_KEY_SIZE);
  if ((ctx=strtok_r(key,DELIMITER,&c)) == NULL) ccerr(invalid_keyword);
  if ((srv=strtok_r(NULL,DELIMITER,&c)) == NULL) ccerr(invalid_keyword);
  if ((kwd=strtok_r(NULL,DELIMITER,&c)) == NULL) ccerr(invalid_keyword);

  if (strlen(ctx) == 0) ccerr(illegal_context);
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strnicmp(el->EqmContext,ctx,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(el->EqmExportName,srv,EXPORT_NAME_SIZE))
    {
      feclogEx(TINE_LOGLEVEL_INFO, "this server is the %s for context %.32s : do not listen !",txt,ctx);
      return server_name_in_use;
    }
  }
  sprintf(dsrv,"/%.32s/%.32s",ctx,srv);
  gNoMsgLog = TRUE;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = 1; dout.dFormat = CF_NAME64; dout.data.cptr = rb;
  din.dArrayLength = 1; din.dFormat = CF_NAME64; din.data.cptr = kwd;
  switch (ExecLinkEx(dsrv,"PROPERTIES",&dout,&din,CA_READ,200))
  {
    default:
    case 0:
      if (GetServerAddr(ctx,srv,&srvaddr,&fecaddr) == 0)
      { /* this should always be successful here ! */
        if (!strncmp(srvaddr.EqmName,DOOCSEQM_TAG,EQM_NAME_SHORTSIZE)) ccerr(illegal_equipment_name);
      }
      break;
    case non_existent:
    case non_existent_elem:
    case non_existent_fec:
    case illegal_equipment_name:
    case address_unknown:
    case address_unresolved:
    case name_unknown:
    case no_such_computer:
      ccerr(non_existent);
  }
err:
  gNoMsgLog = FALSE;
  return cc;
}
int isUsingCycleTrigger = FALSE;
int SystemStartCycleTrigger(void)
{
  static int cycSyncStarted=FALSE, cc = 0;
  int id, retry_count = 0;
  DTYPE dout;
  char *ptr;

  if (gIgnoreCentralServices) return 0;
  if (isUsingCycleTrigger) return cc;
  if (cycSyncStarted) return cc;
  if ((ptr=getenv("TINE_CYCLE_KEY")) != NULL)
  { /* set by environment */
    strncpy(gCyclerNumberKey,ptr,GLOBALS_KEY_SIZE);
  }
  else
  { /* default systematics ... */
    sprintf(gCyclerNumberKey,"/%.32s/CYCLER/CycleNumber",gDeviceContext);
  }
  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &gCycleNumber;
  if ((cc=assertTriggerKeyExists("machine cycler",gCyclerNumberKey)) != 0) 
  {
    feclogEx(TINE_LOGLEVEL_INFO, "no machine cycler exists for context %.32s",gDeviceContext);
    ccerr(non_existent);
  }
retryRecvCycleTrigger:
  id = recvNetGlobalEx(gCyclerNumberKey,&dout,cbCycleTrigger,UNASSIGNED_CALLBACKID);
  if (id < 0) ccerr(-id);
err:
  if (cc == semaphore_busy && ++retry_count < 5) goto retryRecvCycleTrigger;
  if (cc == 0) 
  { /* there's no stopping us now ... */
    isUsingCycleTrigger = cycSyncStarted = TRUE;
  }
  else
  { /* no CYCLER available: inform any registered callbacks ... */
    CycTrg *ct;
    for (ct=gCycTrgLst; ct != NULL; ct=ct->nxt) if (ct->fcn != NULL) ct->fcn(0,cc,ct->ref);
  }
  feclog("initialize machine cycle trigger : %.32s",cc2str(cc));
  return cc;
}
StaTrg *staTrgLst = NULL;
void cbStateTrigger(int id,int cc,void *ref)
{
  static int to_count = 0;
  StaTrg *trg = (StaTrg *)ref;
  ExportListStruct *el;

  UNUSED(id);
  if (trg == NULL) return;
  if (cc) 
  { /* i/o problems */
    if (tineDebug > 0) dbglog("system state update : %.32s",cc2str(cc));
    if (to_count++ < 5) return;
    strcpy(trg->thisState.name,"unavailable");
  }
  else
  { /* back to normal */
    to_count = 0;
  }
  if (!strnicmp(trg->thisState.name,trg->prevState.name,sizeof(NAME64))) return;
  for (el=ExportList; el != NULL; el=el->next)
  { /* scan thru and look for those with the correct context (usually a list of 1) */
    if (!strnicmp(el->EqmContext,trg->ctx,CONTEXT_NAME_SIZE))
    {
      if (el->staChgCb != NULL) el->staChgCb(trg->prevState.name,trg->thisState.name,el->staChgRef);
    }
  }
  feclog("detected state change from %.64s to %.64s",trg->prevState.name,trg->thisState.name);
  strncpy(trg->prevState.name,trg->thisState.name,sizeof(NAME64));
}
int SystemStartStateTrigger(const char *context,const char *stateKey)
{
  int id, cc = 0, retry_count = 0;
  DTYPE dout;
  char key[GLOBALS_KEY_SIZE], *ptr;
  StaTrg *trg;

  if (context == NULL || strlen(context) == 0) context = gDeviceContext;
  for (trg=staTrgLst; trg != NULL; trg=trg->nxt)
  { /* scan thru the list to see if we're already doing this one .. */
    if (strnicmp(trg->ctx,(char *)context,CONTEXT_NAME_SIZE) == 0) break;
  }
  if (trg != NULL) return 0;
  if ((trg=(StaTrg *)calloc(1,sizeof(StaTrg))) == NULL) return out_of_local_memory;
  strncpy(trg->ctx,context,CONTEXT_NAME_SIZE);
  if (stateKey == NULL)
  { /* not given: do the default thing ... */
    if ((ptr=getenv("TINE_STATE_KEY")) != NULL)
    { /* set by environment */
      strncpy(key,ptr,GLOBALS_KEY_SIZE);
    }
    else
    { /* default systematics ... */
      sprintf(key,"/%.32s/%.32s/DeclaredState",context,GLOBALSRV);
    }
  }
  else
  { /* take what was passed ... */
    sprintf(key,"/%.32s/%.32s/%.64s",context,GLOBALSRV,stateKey);
  }
  dout.dArrayLength = 1;
  dout.dFormat = CF_NAME64;
  dout.data.vptr = &trg->thisState;
  dout.dTag[0] = 0;
  if (assertTriggerKeyExists("state server",key)) 
  {
    feclogEx(TINE_LOGLEVEL_WARN, "no state server key %.132s exists for context %.32s",key,context);
    return non_existent;
  }
retryRecvStateTrigger:
  id = recvNetGlobalEx2(key,&dout,cbStateTrigger,UNASSIGNED_CALLBACKID,(void *)trg);
  if (id < 0) ccerr(-id);
err:
  if (cc == semaphore_busy && ++retry_count < 5) goto retryRecvStateTrigger;
  feclog("initialize state trigger %.132s: %.32s",key,cc2str(cc));
  return cc;
}

GlobalListStruct *getGlobalFromKeyword(char *ctx,char *kwd)
{
  GlobalListStruct *g;
  int i;
  for (i=0; i<nglobals; i++)
  { /* now run thru the application supplied requested list */
    if ((g=glbTbl[i])->disabled) continue;
    if (ctx != NULL && strnicmp(ctx,g->context,CONTEXT_NAME_SIZE)) continue;
    if (!strcmp(kwd,g->keyword)) return g;
  }
  return NULL;
}
#endif /* !defined(SMALL_TINE_LIB) */
int prepIncomingGCastHeader(GlbDataHdr *ghdr,BYTE *stream,int protocol) 
{
  if (protocol == 4)
  { /* legacy: Release 4.xx globals */
    GlbDataHdr4 hdr;
#   ifdef INTEL
    memcpy(&hdr, stream, sizeof(GlbDataHdr4));
#   else
    _structSwap(GlbDataHdr4Struct,stream,0,NULL,(BYTE *)&hdr,RTOL);
#   endif
    strncpy(ghdr->name, hdr.name, KEYWORD_NAME_SIZE);
    ghdr->format = hdr.format;
    ghdr->size = hdr.size;
    ghdr->status = hdr.status;
    ghdr->dStamp = hdr.dStamp;
    ghdr->sysStamp = hdr.sysStamp;
    ghdr->dTimeStamp = hdr.dTimeStamp;
    ghdr->dTimeStampUSEC = hdr.dTimeStampUSEC + hdr.dTimeStampMSEC * 1000;
    return sizeof(GlbDataHdr4);
  }
  else
  { 
#   ifdef INTEL
    memcpy(ghdr, stream, sizeof(GlbDataHdr));
#   else
    _structSwap(GlbDataHdrStruct, stream, 0, NULL, (BYTE *)ghdr, RTOL);
#   endif
    return sizeof(GlbDataHdr);
  }
}
int ctxok(char *tgt, char *ctx)
{
  char *c, ctxbuf[CONTEXT_NAME_SIZE+4];
  if (tgt == NULL) return TRUE;
  if (!strnicmp(tgt,ctx,CONTEXT_NAME_SIZE)) return TRUE;
  strncpy(ctxbuf,ctx,CONTEXT_NAME_SIZE);
  if ((c=strrchr(ctxbuf,'.')) == NULL) return FALSE;
  *c++ = 0;
  if (isSystemContextDecoration(c)) return FALSE;
  if (!strnicmp(tgt,ctxbuf,CONTEXT_NAME_SIZE)) return TRUE;
  return FALSE;
}
void getGlobalUpdate(SCKADR *sckadr,BYTE *buf)
{
  GlbDataHdr gdhdr, *gd = &gdhdr; /* what is coming in over the net */
  GlbHdr ghdr;
  int num,size=REQHDR_SIZE;
  int cc=0,i,j,hprot,fmtsize,datasize,dretsize;
  GlobalListStruct *g = NULL; /* what this client wants to see */
  GlbPrmLst *gp;
  struct timeval tv_stamp;
  char kwdBuffer[KEYWORD_NAME_SIZE],*ctx,*kwd;
  INADR inadr;

  if (WaitForSystemMutex(hSystemGlobalsUpdateMutex,-1) != 0) return;
  if (gClientIsClosing) ccerr(not_initialized);
  memcpy(&inadr,getInAddr(sckadr),gInAdrLen);
  adrcpy(&inadr,getInAddr(sckadr));
  gp = getGlbPrmByIP(inadr);
  BRDcount++; /* increment statistics counter */
  hprot = prepIncomingRequestHdr(&ghdr, buf);
  switch (hprot)
  {
    case 4:
    case DEFAULT_TRANSPORT_PROTOCOL_LEVEL:
      break;
    default: 
      ccerr(illegal_protocol);
  }
  size = ghdr.headerSizeInBytes;
  num = ghdr.numberSegments;
  gettimeofday(&tv_stamp,(struct timezone *)NULL);
  for (i=0; i<num; i++)
  {
    size += prepIncomingGCastHeader(gd,&buf[size],hprot);
    strncpy(kwdBuffer,gd->name,KEYWORD_NAME_SIZE);
    if (kwdBuffer[0] == '/')
    {
      ctx = &kwdBuffer[1];
      if ((kwd=strchr(ctx,'/')) == NULL) continue; /* bad keyword ! */
      *kwd++ = 0;
    }
    else
    {
      ctx = NULL; kwd = kwdBuffer;
    }
    if ((fmtsize=fmtsizeof(gd->format)) < 0) ccerr(illegal_format);
    if (gp != NULL)
    { /* known globals parameter list */
      for (j=0; j<gp->num; j++)
      { /* run thru the incoming contents and store away */
        if (!ctxok(ctx,gp->ctx)) continue;
        if (strnicmp(kwd,gp->lst[j].name,64)) continue;
        gp->time = gd->dTimeStamp;
        gp->timeUSEC = gd->dTimeStampUSEC;
        gp->dataStamp = gd->dStamp;
        gp->sysStamp = gd->sysStamp;
        if (gp->data[j] == NULL && gd->size < 256) 
        {
          gp->fmt[j] = BFMT(gd->format);
          gp->siz[j] = (BYTE)gd->size;
          gp->data[j] = (BYTE *)SystemCalloc(gd->size,fmtsize);
        }
        if (gp->data[j] != NULL && !strnicmp(gp->lst[j].name,kwd,PROPERTY_NAME_SIZE))
          memcpy(gp->data[j],&buf[size],(size_t)(gd->size*fmtsize));
        break; /* found this one */
      }
    }
    for (j=0; j<nglobals; j++)
    { /* now run thru the application supplied requested list */
      if ((g=glbTbl[j])->disabled) continue;
      if (!ctxok(ctx,g->context)) continue;
      if (!strnicmp(kwd,g->keyword,PROPERTY_NAME_SIZE))
      {
        if (adrcmp(&inadr,&inadr_zero) && 
            adrcmp(&g->srvIp,&inadr_zero) && 
            adrcmp(&inadr,&g->srvIp)) continue;
        if (gd->dTimeStamp < (UINT32)g->time) continue;
        g->endianness = ghdr.endianness;
        g->tocounter = 0;
        g->sysStamp = gd->sysStamp;
        g->dataStamp = gd->dStamp;
        g->time = gd->dTimeStamp;
        g->timeUSEC = gd->dTimeStampUSEC;
        g->tv = tv_stamp;
        g->status = gd->status;
        gLastGlobalDataSize = g->lastsize = gd->size;
        dretsize = fmtsize*gd->size; 
        if (g->format == CF_DEFAULT)
        { /* fix the format and datasize */
          if (dretsize > g->size) 
          { /* not enough room */
            g->status = buffer_too_small;
            NotifyGlobalClient(j,buffer_too_small);
            ccerr(buffer_too_small);
          }
          g->format = gd->format;
          g->size = gd->size;
        }     
        datasize = fmtsizeof(g->format)*g->size;
        if (g->format != gd->format) 
        {
          if (g->refdata == NULL)
          {
            if ((g->refdata=(BYTE *)SystemCalloc(gd->size,fmtsize)) == NULL) 
            {
              if (g->cbfcn != NULL) (g->cbfcn)((int)g->cbId,out_of_client_memory);
              continue;
            }
          }
          prepIncomingData(g->refdata,&buf[size],dretsize,dretsize,gd->format,ghdr.endianness,NULL);
          cc = reformat(g->data,g->format,g->refdata,gd->format,gd->size);
        }
        else
        {
          prepIncomingData(g->data,&buf[size],datasize,dretsize,gd->format,ghdr.endianness,NULL);
          cc = 0;
        }
        if (cc == 0) cc = g->status;
        if (g->conTblId > 0) conTbl[g->conTblId-1]->lastLnkErrSrc = LNK_ERROR_RMT;
        if (cc != 0) cc |= CE_SENDCODE;
        NotifyGlobalClient(j,cc);
        /* don't break; in case multiple links to the same global */
      }
    }
    size += gd->size*fmtsize;
  }
  if (gp != NULL) gp->hasUpdated = TRUE;
err:
  ReleaseSystemMutex(hSystemGlobalsUpdateMutex);
  if (tineDebug > 2 || (cc && errorDebug))
  {
    sprintf(dbgbuf,"Received globals packet (header protocol %d)",hprot);
    if (cc) sprintf(&dbgbuf[strlen(dbgbuf)],"ERR: %.32s",cc2str(cc));
    dbglog(dbgbuf);
  }
  return;
}
int getGlobalFromLastUpdate(const char *keyword,INADR inadr,struct timeval *tv_stamp)
{
  GlobalListStruct *g = NULL; /* what this client wants to see */
  GlbPrmLst *gp = getGlbPrmByIP(inadr);
  BYTE *buf;
  int cc=0, i, j, fsiz, dretsize, datasize;
  short fmt;
  if (gp == NULL) return non_existent;
  for (i=0; i<gp->num; i++)
  {
    if (!strnicmp((char *)keyword,gp->lst[i].name,PROPERTY_NAME_SIZE))
    {
      for (j=0; j<nglobals; j++)
      { /* now run thru the application supplied requested list */
        if ((g=glbTbl[j])->disabled) continue;
        if (!strcmp(keyword,g->keyword))
        {
          if (adrcmp(&inadr,&g->srvIp)) continue;
          if ((buf=gp->data[i]) == NULL) continue;
          fmt = LFMT(gp->fmt[i]);          
          fsiz = fmtsizeof(fmt);
          gLastGlobalDataSize = gp->siz[i];
          dretsize = fsiz*gp->siz[i]; 
          if (g->format == CF_DEFAULT)
          { /* fix the format and datasize */
            if (dretsize > g->size) return buffer_too_small;
            g->format = fmt;
            g->size = gp->siz[i];
          }
          g->time = gp->time;
          g->timeUSEC = gp->timeUSEC;
          g->sysStamp = gp->sysStamp;
          g->dataStamp = gp->dataStamp;
          g->tv = *tv_stamp;
          g->status = 0;
          datasize = fmtsizeof(g->format)*g->size;
          if (g->format != fmt) 
          {
            if (g->refdata == NULL)
            {
              if ((g->refdata=(BYTE *)SystemCalloc(gp->siz[i],fsiz)) == NULL) 
              {
                if (g->cbfcn != NULL) (g->cbfcn)((int)g->cbId,out_of_client_memory);
                continue;
              }
            }
            prepIncomingData(g->refdata,buf,dretsize,dretsize,fmt,g->endianness,NULL);
            cc = reformat(g->data,g->format,g->refdata,fmt,gp->siz[i]);
          }
          else
          {
            prepIncomingData(g->data,buf,datasize,dretsize,fmt,g->endianness,NULL);
            cc = 0;
          }
          return cc;
        }
      }
    }
  }
  return non_existent;
}
void dumpGlobals(void)
{
  int i,dsize;
  double ts;
  char s[256],addr[64], tsstr[64];

  if (nglobals > 0)
  {
    ttyoutput("Globals received :");
    for (i=0; i<nglobals; i++)
    {
      if (glbTbl[i]->disabled) continue;
      if (!glbTbl[i]->isInitialLink) continue;
      dsize = MIN(glbTbl[i]->size,10);
      prepDataDump(s,(char *)glbTbl[i]->data,dsize,glbTbl[i]->format,FALSE);
      ttyoutput("%.64s (%d) : %.256s",glbTbl[i]->keyword, glbTbl[i]->status,s);
      inet_ntop(gAfInet,&glbTbl[i]->mcastGrp,addr,40);
      ttyoutput("  multicast group: %.40s",addr);
      inet_ntop(gAfInet,&glbTbl[i]->srvIp,addr,40);
      ttyoutput("  multicast source: %.40s",addr);
      ts = PutDataTimeStampU(0,glbTbl[i]->tv.tv_sec,glbTbl[i]->tv.tv_usec);
      ttyoutput("  last update: %s",GetDataTimeStringEx(ts,FALSE,tsstr));
      ts = PutDataTimeStampU(0,glbTbl[i]->time,glbTbl[i]->timeUSEC);
      ttyoutput("  last timestamp: %s",GetDataTimeStringEx(ts,FALSE,tsstr));
    }
  }
  if (ngcasts > 0)
  {
    if (gSndIpBroadcastsOnly)
      ttyoutput("Globals sent (broadcasting only):");
    else
      ttyoutput("Globals sent (multicast group %.40s):",gtMCastAddr);
    for (i=0; i<ngcasts; i++)
    {
      prepDataDump(s,(char *)gcastTbl[i]->data,gcastTbl[i]->size,gcastTbl[i]->format,FALSE);
      ttyoutput("%.64s (%d) : %.256s",gcastTbl[i]->keyword, gcastTbl[i]->status,s);
    }
  }
  if (nglobals == 0 && ngcasts == 0) ttyoutput("neither receiving nor sending globals!");
  return;
}
/*
 * \internal
 * Try to find a reasonable match from the link table.
 * The caller has to manage his callback IDs effectively
 * for this routine to return something sensible.
 * TODO: make all of the GetCompletion<> routines take
 * the callback function address as a parameter so as to
 * be able to dereference it properly!
 */
int getLinkIdFromCallbackId(int cbId)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    if (!isActiveTransferMode(conTbl[i]->mode))
    { /* not an active link, but is it bound ? */
      if (conTbl[i]->boundToId <= 0 && conTbl[i]->mcaLink == NULL)
        continue;
    }
    if (conTbl[i]->wcLink != NULL && 
        conTbl[i]->wcLink->cbId == cbId) 
    { /* a non-local wildcard link */
      return i;
    }
    else if (conTbl[i]->isWildcardLink)
    { /* otherwise a wildcard link */
      WcTblEntry *wc=conTbl[i]->wcLink;
      if (wc != NULL && wc->cbId == cbId) return wc->linkId;
    }
    else if (conTbl[i]->hasUserCallbackId &&
             conTbl[i]->cbId == (UINT32)cbId) 
    { /* look for supplied callback id */
      return i;
    }
  }
  if (cbId >= 0 && cbId < nConnectionTableEntries && !conTbl[cbId]->hasUserCallbackId)
    return cbId;
  return -1;
}
TINE_EXPORT GlobalListStruct *GetGlobalListElement(int id)
{
  static GlobalListStruct gls;
  if (id < 0 || id >= nglobals) return NULL;
  if (glbTbl[id]->disabled) return  NULL;
  memcpy(&gls,glbTbl[id],sizeof(GlobalListStruct));
  return &gls;
}
TINE_EXPORT int GetGlobalLinkId(char *keyword)
{
  char *kwd=NULL,*ctx=NULL,*srv=NULL,kwdbuf[128];
  int i = -1,cc;
  strncpy(kwdbuf,keyword,64); keyword[64] = 0;
  if ((cc=parseGlobalsKeyword(kwdbuf,&ctx,&srv,&kwd)) != 0) return -cc;
  while ((i=getGlobalTableIndex(i+1,ctx,srv,kwd)) != -1)
  { /* find the first active link */
    if (!glbTbl[i]->disabled) return i;
  }
  return i;
}
TINE_EXPORT int GetGlobalLinkIdFromCallbackId(int cbId)
{
  int i;
  for (i=0; i<nglobals; i++)
  {
    if (glbTbl[i]->cbId == (UINT32)cbId) return i;
  }
  return -1;
}
#ifndef SMALL_TINE_LIB
/** 
 * \internal for use when this link is identical to another
 * \param linkId is the link table id, which should already know its 'boundToId'.
 */
int addDependentLink(int linkId)
{
  ConTblEntry *c;
  DpdTblEntry *lnk;
  int parentId,cc=0;
  if (linkId < 0 || linkId > nConnectionTableEntries) return argument_list_error; /* not a valid link table entry ! */
  parentId = conTbl[linkId]->boundToId;
  if (parentId < 1 || parentId >= nConnectionTableEntries) return argument_list_error; /* not a valid link table entry ! */
  c = conTbl[parentId];
  if (c->isWaitingSyncNotification) return completion_waiting;
  if (WaitForSystemMutex(hDpdLinkMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return mutex_error;
  if ((lnk=(DpdTblEntry *)SystemCalloc(1,sizeof(DpdTblEntry))) == NULL) ccerr(out_of_client_memory);
  lnk->linkId = linkId;
  lnk->nxt = c->dpdLink;
  if (c->dpdLink != NULL) c->dpdLink->prv = lnk;
  c->dpdLink = lnk;
  if (c->hasNotifiedOnce) 
  { /* for the late-comers */
    if (c->mcaLink != NULL) 
    { /* dependent on an MCA single-link element (already dependent) */
      conTbl[c->mcaLink->linkId]->needsNotification = TRUE;
      conTbl[c->mcaLink->linkId]->suppressCallback = FALSE;
    }
    else 
    {
      c->needsNotification = TRUE; 
      c->suppressCallback = FALSE;
    }
  }
err:
  ReleaseSystemMutex(hDpdLinkMutex);
  msglog(cc,"add dependent link %d to parent (%d) %.192s",linkId,parentId,c->key);
  return cc;
}
void clearDependentLinks(int linkId)
{
  ConTblEntry *c;
  DpdTblEntry *lnk, *nlnk=NULL;
  if (linkId < 0 || linkId > nConnectionTableEntries) return; /* not a valid link table entry ! */
  c = conTbl[linkId];
  for (lnk=c->dpdLink; lnk != NULL; lnk=nlnk)
  {
    nlnk = lnk->nxt;
    free(lnk);
  }
  c->dpdLink = NULL;
}
void removeDependentLink(int linkId)
{ 
  ConTblEntry *c;
  DpdTblEntry *lnk;
  int parentId, found=FALSE;
  if (linkId < 0 || linkId > nConnectionTableEntries) return; /* not a valid link table entry ! */
  parentId = conTbl[linkId]->boundToId;
  if (parentId < 1 || parentId > nConnectionTableEntries) return; /* not a valid link table entry ! */
  c = conTbl[parentId];
  if ((lnk=c->dpdLink) == NULL) return; /* no dependency list ! */
  conTbl[linkId]->mode = CM_CANCEL;
  conTbl[linkId]->boundToId = 0;
  /* if (conTbl[linkId]->isInsideCallback) return; */ /* should not be necessary any longer ...*/
  if (WaitForSystemMutex(hDpdLinkMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return;
  for (lnk=c->dpdLink; lnk != NULL; lnk=lnk->nxt)
  {
    if (lnk->linkId == linkId)
    { /* got it */
      found = TRUE;
      if (lnk->prv != NULL) lnk->prv->nxt = lnk->nxt;
      if (lnk->nxt != NULL) lnk->nxt->prv = lnk->prv;
      if (lnk == c->dpdLink) c->dpdLink = lnk->nxt;
      free(lnk); 
      break;
    }
  }
  ReleaseSystemMutex(hDpdLinkMutex);
  if (found) msglog(0,"remove dependent link %d from parent (%d) %.192s",
    linkId,parentId,c->key);
  return;
}
int wcAllowedPending(WcTblEntry *wc)
{ /* internal use: skip the NULL pointer test */
  int i,n=0;
  for (i=0; i<wc->length; i++) if (wc->status[i] == link_error) n++;
  return n;
}
void wcMembersCallback(int id,int cc,void *ref)
{
  int wccc, lstid = id;
  int len, maxlen, fsiz;
  int i, k;
  WcTblEntry *wc = (WcTblEntry *)ref;
  ConTblEntry *c = NULL, *p = NULL;
  BYTE *mdata, *pdata;

  if (nWildCardTableEntries == 0) return; /* flushed ? */
  if (wc == NULL || wc->linkId < 0) return;
  if (lstid < 0 || lstid >= wc->length) return;
  if (wc->status[lstid] == not_ready) wc->numPending--;
  wc->status[lstid] = cc;
  switch (cc)
  {
    case illegal_format:
    case illegal_property:
    case link_blacklisted:
    case resources_exhausted:
      _closeLink(wc->listId[lstid]);
      break;
  }
  if (wc->numPending > wcAllowedPending(wc)) return;
  p = conTbl[wc->linkId];
  maxlen = p->sizeOut;
  pdata = (BYTE *)p->dataPtr;
  fsiz = fmtsizeof(p->formatOut);
	for (i=0, k=0; i<wc->length && k<maxlen; i++)
	{ /* each member link has a subset of the list to return ... */
    if (wc->status[i] == link_error || wc->listId[i] < 0) continue;
    c = conTbl[wc->listId[i]];
    if (c->formatOut != p->formatOut) ccerr(code_failure);
    mdata = (BYTE *)c->dataPtr;
	  len = c->lastDataSize;
    if (len >= maxlen - k) len = maxlen - k - 1;
    memcpy(&pdata[k*fsiz],mdata,(size_t)(len*fsiz));
    p->dtimestamp = c->dtimestamp;
    p->dtimestampUSEC = c->dtimestampUSEC;
    p->systemstamp = c->systemstamp;
    p->datastamp = c->datastamp;
	  k += len;
	}
  p->lastDataSize = k;
err:
  wc->numActive = wc->numPending = wc->length;
  wccc = cc;
 for (i=0; i<wc->length; i++)
  {
    if (wccc != 0 && wc->status[i] == 0) wccc = 0; /* at least one success */
    if (wc->status[i] != link_error) wc->status[i] = not_ready;
  }
  p->lastDataSize = k;
  p->linkStatus = wccc;
  p->dataStatus = wccc;
  LastCompletionDataSize = p->lastDataSize;
  GetErrorString(wccc,p->lastStatusString);
  if (wc->cbfcn != NULL) TCBCALL(wc,wccc);
  if (wc->mode <= CM_SINGLE || !wc->active) 
  {
    p->mode = wc->mode = CM_CANCEL;
    wc->active = FALSE;
  }
  if (linkWaitingOnStartup(p)) _cbExecLink(p->linkId,cc); /* stop waiting */
  p->tocounter = 0;

  return;
}
void wcGroupCallback(int id,int cc,void *ref)
{
  int wccc, lstid = id;
  int i, k, fmtlen, nmlen, noff, voff, soff;
  WcTblEntry *wc = (WcTblEntry *)ref;
  ConTblEntry *c = NULL;
  double ts;

  if (wc == NULL || wc->linkId < 0) return;
  if (wc->status[lstid] == not_ready) wc->numPending--;
  wc->status[lstid] = cc;
  switch (cc)
  {
    case illegal_format:
    case illegal_property:
    case link_blacklisted:
    case resources_exhausted:
      _closeLink(wc->listId[lstid]);
      break;
  }

  if (wc->numPending > wcAllowedPending(wc)) return;

  c = conTbl[wc->linkId]; /* the original link entry */
  k = 0; /* count the legal links */
  fmtlen = fmtsizeof(wc->dout.dFormat);
  switch (wc->dout.dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      nmlen = fmtlen - 16;
      for (i=0; i<wc->length && k<(int)c->sizeOut; i++)
      { 
        if (wc->status[i] == illegal_format) continue;
        if (wc->status[i] == illegal_property) continue;
        *(double *)(c->dataPtr + k*fmtlen + nmlen) = (double)((float *)wc->data)[i];
        *(double *)(c->dataPtr + k*fmtlen + nmlen + 8) = (double)wc->status[i];
        strncpy(c->dataPtr + k*fmtlen,wc->list[i].name,nmlen);
        k++;
      }
      break;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      if (wc->dout.dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      for (i=0; i<wc->length && k<(int)c->sizeOut; i++)
      { 
        if (wc->status[i] == illegal_format) continue;
        if (wc->status[i] == illegal_property) continue;
        strncpy(c->dataPtr + k*fmtlen + noff,wc->list[i].name,MIN(nmlen,64));
        memcpy(c->dataPtr + k*fmtlen + voff,&((UINT32 *)wc->data)[i],4);
        *(UINT32 *)(c->dataPtr + k*fmtlen + soff) = wc->status[i];
        k++;
      }
      break;
    default:
      if (wc->dout.dFormat <= CF_NAME32 && strcmp(c->devName,"*") == 0)
      {
        for (i=0; i<wc->length && k<(int)c->sizeOut; i++)
        {
          if (wc->status[i] == illegal_format) continue;
          if (wc->status[i] == illegal_property) continue;
          memcpy((BYTE *)c->dataPtr +(size_t)(k*fmtlen),wc->data+(size_t)(i*fmtlen),fmtlen);
          k++;
        }
      }
      else
      {
        for (i=0; i<wc->length; i++)
        {
          wc->status[i] = illegal_format;
          _closeLink(wc->listId[i]);
        }
      }
      break;
  }
  wc->numActive = wc->numPending = k;
  wccc = cc;
  for (i=0; i<wc->length; i++)
  {
    if (wc->status[i] == illegal_format) continue;
    if (wc->status[i] == illegal_property) continue;
    if (wccc != 0 && wc->status[i] == 0) wccc = 0; /* at least one success */
    if (wc->status[i] != link_error) wc->status[i] = not_ready;
  }
  c->lastDataSize = k;
  c->linkStatus = wccc;
  c->dataStatus = wccc;
  ts = MakeDataTimeStamp();
  c->dtimestamp = (int)ts;
  c->dtimestampUSEC = (int)((ts - c->dtimestamp) * 1000);
  LastCompletionDataSize = c->lastDataSize;
  GetErrorString(wccc,c->lastStatusString);
  if (wc->cbfcn != NULL && !hasSuppressedCallback(wc->cbfcn)) TCBCALL(wc,wccc);
  if (wc->mode <= CM_SINGLE) 
  { 
    c->mode = wc->mode = CM_CANCEL; 
    wc->active = FALSE; 
  }
  if (linkWaitingOnStartup(c)) _cbExecLink(c->linkId,cc); /* stop waiting */
  c->tocounter = 0;
  return;
}
void wcSingleCallback(int id,int cc,void *ref)
{
  ConTblEntry *c = NULL;
  WcTblEntry *wc = (WcTblEntry *)ref;
  char srv[70],dev[132];
  int fmt, fmtsiz, i, n, mask=0, rc=link_timeout;
  DTYPE dout, din;

  c = conTbl[id]; /* the parent link */
  if (c->wcLink == NULL) wc = NULL; /* flushed ? */
  if (wc == NULL || wc->linkId < 0) ccerr(code_failure);
  switch (cc)
  {
    case illegal_property:
      /* will be returned by the group server */
    case data_not_local:
      /* mark this entry as CM_REGISTER */
      c->heartbeat = HBT_NONE;
      wc->mode = c->mode == CM_CANCEL ? CM_SINGLE : c->mode;
      c->mode = CM_REGISTER;
      /* get the device list and length and allocate the relevant arrays */
      if (strlen(c->ctxName) > 0) sprintf(srv,"/%.32s/",c->ctxName); else srv[0] = 0;
      strcat(srv,c->expName);
      sprintf(dev,"%.64s/%.64s",srv,c->devName);
      /* first find out how many */
      dout.dArrayLength = 1; dout.dFormat = CF_INT32; 
      dout.data.lptr = (SINT32 *)&n;
      rc = ExecLinkEx(dev,"NDEVICES",&dout,NULL,CA_READ|CA_RETRY,800);
      if (rc != 0) ccerr(rc);
      /* now allocate the device tables */
      wc->length = n;
      /* NDEVICES returns the total number, so this is a worst case allocation ... */
      if ((wc->list=(NAME64 *)SystemCalloc(n,sizeof(NAME64))) == NULL) ccerr(out_of_client_memory);
      if ((wc->status=(int *)SystemCalloc(n,sizeof(int))) == NULL) ccerr(out_of_client_memory);
      if ((wc->listId=(int *)SystemCalloc(n,sizeof(int))) == NULL) ccerr(out_of_client_memory);
      if (wc->dout.dFormat == CF_DEFAULT)
      { /* a wild card call with unknown format + propertyQueryFunction returning illegal_property */
        c->formatOut = wc->dout.dFormat = CF_FLOAT; 
      }
      switch (wc->dout.dFormat)
      {
        case CF_NAME16DBLDBL:
        case CF_NAME32DBLDBL:
        case CF_NAME64DBLDBL:
        case CF_USTRING:
        case CF_NAME16FI:
          fmt = CF_FLOAT;
          break;
        case CF_NAME16II:
          fmt = CF_INT32;
          break;
        default:
          if (wc->dout.dFormat > CF_NAME32 && strcmp(c->devName,"*") != 0) ccerr(illegal_format);
          fmt = wc->dout.dFormat;
          break;
      }
      fmtsiz = fmtsizeof(fmt);
      if ((wc->data=(BYTE *)SystemCalloc(n,fmtsiz)) == NULL) ccerr(out_of_client_memory);   
      /* fill in the device list */
      if (c->sizeIn == 1 && c->formatIn == CF_INT32)
      { /* a wildcard mask is being used */
        memcpy(&mask,c->dataIn,sizeof(int));
        rc = GetWildcardDevices(dev,mask,wc->list,&n);
      }
      else
      { /* get all devices */
        rc = GetSystemDevices(dev,NULL,wc->list,&n);
      }
      if (rc != 0) ccerr(rc);
      /* if the wildcard contains characters, then n returned here is likely to be smaller than above */
      wc->length = n;
      dout.dArrayLength = 1;
      dout.dFormat = fmt;
      din.dArrayLength = c->sizeIn;
      din.dFormat = c->formatIn;
      din.data.vptr = c->dataIn;
      wc->numPending = wc->length;
      for (i=0; i<wc->length; i++)
      { /* start the individual links (these will be redirected by the source) */
        sprintf(dev,"%.67s/%.64s",srv,wc->list[i].name);
        dout.data.vptr = wc->data + (size_t)(i*fmtsiz);
        wc->status[i] = not_ready;
        wc->listId[i] = _attachLink(dev,c->prpName,&dout,&din,c->access,c->pollingInterval,
                                    (void *)wcGroupCallback,wc->mode,i,wc,UNASSIGNED_INDEX);
        if (wc->listId[i] < 0) wc->status[i] = link_error;
      }
      /* cross-reference back into client connection table */
      c->wcLink = wc;
      c->tocounter = -1; /* reset */
      break;
    default:
      if (wc->cbfcn != NULL && !hasSuppressedCallback(wc->cbfcn)) TCBCALL(wc,cc);
      if (wc->mode <= CM_SINGLE) 
      {
        c->mode = wc->mode = CM_CANCEL;
        wc->active = FALSE;
      }
      break;
  }
  return;
err:
  if (wc != NULL)
  {
    if (wc->cbfcn != NULL && !hasSuppressedCallback(wc->cbfcn)) TCBCALL(wc,cc);
    wc->mode = CM_CANCEL;
    wc->active = FALSE;
    _closeLink(wc->linkId);
  }
  msglog(cc,"wildcard call cannot get device information : %.32s",cc2str(cc));
  return;
}
int lwdPollingInterval = LWD_POLLING_RATE;
TINE_EXPORT int SetLinkWatchdogPollingInterval(int value)
{
  if (value < 50 || value > 10000) return out_of_range;
  feclog("set link watchdog polling interval to %d",value);
  lwdPollingInterval = value;
  return 0;
}
LnkWatchDog *wdLinks = NULL;
void lwdCloseLink(LnkWatchDog *lwd)
{
  LnkWatchDog *nlwd;
  if (lwd == NULL) return;
  if (tineDebug) dbglog("close link watchdog for server /%.32s/%.32s",conTbl[lwd->lnkId]->ctxName,conTbl[lwd->lnkId]->expName);
  _closeLink(lwd->lnkId);
  conTbl[lwd->lnkId]->cbRef = NULL; /* just in case callback is still called */
  if (WaitForSystemMutex(hLwdLinkMutex,gSystemTick) != 0) return;
  lwd->lnkId = -1;
  if (lwd == wdLinks) 
  { /* we're at the head */
    wdLinks = lwd->nxt;
  }
  else
  { /* we're in the middle */
    for (nlwd=wdLinks; nlwd != NULL; nlwd=nlwd->nxt)
    { /* find it */
      if (nlwd->nxt == lwd)
      { /* splice and leave */
        nlwd->nxt = nlwd->nxt->nxt;
        break;
      }
    }
  }
  SystemFree(lwd);
  ReleaseSystemMutex(hLwdLinkMutex);
}
void lwdCallback(int id,int cc,void *ref)
{
  LnkWatchDog *lwd=(LnkWatchDog *)ref;
  int i;
  UNUSED(id);
  if (lwd == NULL)
  { /* this should not be possible ! */
    feclogEx(TINE_LOGLEVEL_ERROR, "watchdog link without list reference (code error!)");
    return;
  }
  if (lwd->nlinks <= 0) 
  { /* nobody left to watch */
    lwdCloseLink(lwd);
    return;
  }
  if (lwd->lnkLastStatus == cc) return;
  lwd->lnkLastStatus = cc;
  switch (cc) 
  { 
    case 0:
      /* TODO: could see if srvStartTime has changed (due to SRVRESET) */
    default:
      break;
    case link_timeout:
    case connection_timeout:
      for (i=0; i<nConnectionTableEntries; i++)
      { /* scan thru and shrink the heartbeat for those being watched */
        if (conTbl[i] == NULL || conTbl[i]->wdLink != lwd) continue;
        conTbl[i]->heartbeat = HBT_MINIMUM;
        conTbl[i]->tocounter = TO_RETRY_THRESHOLD - 1;
        conTbl[i]->suppressCallback = FALSE;
      }
      break;
  }
  return;
}
int attachLwdLink(int linkId)
{
  DTYPE dout;
  char dev[128],*ctx,*srv;
  int i, bmode, cc = 0;
  LnkWatchDog *lwd = NULL;
  if (linkId < 0 || linkId >= nConnectionTableEntries) return -invalid_link;
  if (conTbl[linkId] == NULL) return -invalid_link;
  if (conTbl[linkId]->mcaIndex > 0) return -not_applicable;
  if (conTbl[linkId]->dpdLink != NULL) return -not_applicable;
  if (conTbl[linkId]->wdLink && conTbl[linkId]->cbfcn == (TCBFCNP)_cbMcaXRef) return -not_applicable;
  if (IsStockProperty(conTbl[linkId]->prpName)) return -not_applicable;
  if ((bmode=BASEMODE(conTbl[linkId]->mode)) != CM_DATACHANGE && bmode != CM_EVENT) return -not_applicable;
  if (bmode == CM_TIMER && conTbl[linkId]->pollingInterval < (UINT32)lwdPollingInterval * LWD_TIMER_DEADBAND) 
  { /* don't watch the timer links under most circumstances */
    return -not_applicable;
  }
  ctx = conTbl[linkId]->ctxName; srv = conTbl[linkId]->expName;
  if (conTbl[linkId]->rdrLink != NULL)
  { /* redirect the watchdog link as well */
    ctx = conTbl[linkId]->rdrLink->dstContext;
    srv = conTbl[linkId]->rdrLink->dstServerName;
  }
  if (WaitForSystemMutex(hLwdLinkMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) ccerr(-mutex_error);
  for (lwd=wdLinks; lwd != NULL; lwd=lwd->nxt)
  { /* already doing this ? */
    if ((i=lwd->lnkId) < 0 || i >= nConnectionTableEntries) continue;
    if (conTbl[i] == NULL || i == linkId) continue;
    if (strncmp(conTbl[i]->strTagOut,LWD_DTAG,TAG_NAME_SIZE)) continue;
    if (strncmp(conTbl[i]->ctxName,ctx,CONTEXT_NAME_SIZE)) continue;
    if (strncmp(conTbl[i]->expName,srv,EXPORT_NAME_SIZE)) continue;
    if (strncmp(conTbl[i]->prpName,LWD_PROPERTY,PROPERTY_NAME_SIZE)) continue;
    conTbl[linkId]->wdLink = lwd; /* link is being watched */
    lwd->nlinks++;
    ccerr(i); /* yep, already doing this ! */
  }
  if (bmode == CM_TIMER) 
  { /* if not already watching the server via a datachange link then don't watch ! */
    ccerr(-not_applicable);
  }
  if ((lwd=(LnkWatchDog *)SystemCalloc(1,sizeof(LnkWatchDog))) == NULL) ccerr(-out_of_client_memory);
  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  strncpy(dout.dTag,LWD_DTAG,TAG_NAME_SIZE);
  dout.data.ulptr = &lwd->srvStartTime;
  sprintf(dev,"/%.32s/%.32s",ctx,srv);
  lwd->lnkId = AttachLinkEx2(dev,LWD_PROPERTY,&dout,NULL,CA_READ,lwdPollingInterval,lwdCallback,CM_TIMER,UNASSIGNED_CALLBACKID,(void *)lwd);
  if ((i=lwd->lnkId) < 0)
  {
    SystemFree(lwd);
    ccerr(i);
  }
  lwd->nlinks = 1;
  lwd->nxt = wdLinks;
  wdLinks = lwd;
  conTbl[linkId]->wdLink = lwd; /* link is being watched */
  if (tineDebug) dbglog("add link watchdog for server %.128s",dev);
  ccerr(lwd->lnkId);
err:
  ReleaseSystemMutex(hLwdLinkMutex);
  return cc;
}
#endif /* not SMALL_TINE_LIB */
int makeValidPollingInterval(int pollingRate, int baseMode)
{
  if (pollingRate <= 0) return 1000;
  if (baseMode <= CM_SINGLE) return pollingRate;
  if (pollingRate < MinPollingRate) return MinPollingRate;
  if (pollingRate <= 100) return pollingRate;
  if (pollingRate < 170)  return 100; 
  if (pollingRate < 250)  return 200;
  if (pollingRate < 350)  return 300;
  if (pollingRate < 450)  return 400;
  if (pollingRate < 750)  return 500;
  if (pollingRate < 1000) return 1000;
  return (pollingRate/1000) * 1000;
}
int isENSCall(char *devName)
{
  if (strstr(devName,ENS_TAG) != NULL) return TRUE;
  if (!stricmp(devName,ENS_SRV)) return TRUE;
  if (!strnicmp(devName,"/SITE/ENS/",10)) return TRUE;
  if (!strnicmp(devName,"/SERVICE/ENS/",13)) return TRUE;
  return FALSE;
}
int parseGlobalsKeyword(char *key,char **context,char **server,char **keyword)
{ /* not exported, so we know that the args are always okay */
  char *c, *ctx, *srv, *kwd;
  if (key[0] == '/')
  { /* context given */
    ctx = &key[1];
    if ((c=strchr(ctx,'/')) == NULL) return invalid_property; /* badly formed -> context but no real keyword given */
    *c++ = 0; 
    srv = c;
    if ((c=strchr(srv,'/')) == NULL)
    { /* just context and keyword given */
      kwd = srv;
      srv = GLOBALSRV;
    }
    else
    { /* context, server, and keyword given */
      *c++ = 0;
      kwd = c;
    }
  }
  else
  { /* naked keyword */
    if ((c=strchr(key,'/')) != NULL) return invalid_property; /* badly formed -> server + keyword given with no context */
    ctx = "SITE";
    srv = GLOBALSRV;
    kwd = key;
  }
  *server = srv;
  *context = ctx;
  *keyword = kwd;
  return 0;
}
int getGlobalTableIndex(int start,char *context,char *server,char *keyword)
{
  int i;
  if (start < 0) start = 0;
  for (i=start; i<nglobals; i++) 
  {
    if (strnicmp(glbTbl[i]->context,context,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(glbTbl[i]->server,server,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(glbTbl[i]->keyword,keyword,PROPERTY_NAME_SIZE)) continue;
    return i;
  }
  return -1;
}
TINE_EXPORT int ListenForMulticasts(void) 
{
  return ListenForIPMulticasts();
}
TINE_EXPORT int ListenForNetGlobals(void) 
{
  return ListenForIPNetGlobals();
}
TINE_EXPORT char *GetRPCLastError(short cc)
{
  static char s[128];
  char *errsrc, errstr[64];
  if (lastLnkErrSrc < 0 || lastLnkErrSrc > 1) lastLnkErrSrc = 0;
  errsrc = gStatusSourceString[lastLnkErrSrc];
  if (cc == -1) cc = lastLnkErr;
  if (cc < 0) cc = -cc;
  cc = cc & 0x0fff;
  if ((cc >= 0 && cc < numErr) || strlen(gLastStatusString) == 0)
  {
    strncpy(errstr, cc2str(cc), 32);
  }
  else
  {
    strncpy(errstr, gLastStatusString, 64);
  }
  sprintf(s,"%.3s: %.64s",errsrc,errstr);
  return s;
}
TINE_EXPORT int AssignImageFrameBuffer(DIMAGE *img, BYTE *buf, int buflen)
{
  if (img == NULL || buf == NULL) return argument_list_error;
  img->frameBuffer = buf;
  img->frameBufferSize = (UINT32)buflen;
  return 0;
}
#ifndef SMALL_TINE_LIB
int isLinkToLocalProcess(int fid)
{
  if (fid < 0 || fid >= numFecTblEntries) return FALSE;
  if (FecTbl[fid].addrFlags & ADDRFLAG_LOCAL_PROCESS) return TRUE;
  return FALSE;
}
FilterLink *fltrLnkLst = NULL;
void _cbFltrLnk(int id, int cc,void *ref)
{
  FilterLink *fl = (FilterLink *)ref;
  ConTblEntry *c = conTbl[id];
  DTYPE d;
  int top = FALSE;
  static int last_cc = 0;
  if (cc == illegal_property) return; /* avoid startup problems by ignoring illegal_property */
  if (!HAS_DATA(cc)) ccerr(cc);
  DTYPEZERO(d);
  d.dFormat = c->formatOut;
  d.dArrayLength = c->sizeOut;
  d.data.vptr = c->dataPtr;
  if (isNumberFormat(c->formatOut))
  {
    switch (fl->fcmp)
    {
      case fl_eq: top = TRUE;
      case fl_neq:
      {
        int iv;
        _GetValuesAsLong(&d,&iv,1,TRUE);
        fl->valid = top ? (iv == (int)fl->fval) : (iv != (int)fl->fval);
        return;
      }
      case fl_gt: top = TRUE;
      case fl_lt:
      {
        double dv;
        d.dFormat = c->formatOut;
        d.dArrayLength = 1;
        d.data.vptr = c->dataPtr;
        if ((cc=_GetValuesAsDouble(&d,&dv,1,TRUE)) != 0) ccerr(cc);
        fl->valid = top ? (dv > fl->fval) : (dv < fl->fval);
        return;
      }
    }
    return;  
  }
  if (isSimpleStringFormat(c->formatOut))
  {
    char sv[128];
    if ((cc=GetValuesAsStringEx(&d,sv,CF_TEXT,128,0)) != 0) ccerr(cc);
    switch (fl->fcmp)
    {
      case fl_eq:  fl->valid = !strnicmp(sv,fl->fstr,64); return;
      case fl_neq: fl->valid = strnicmp(sv,fl->fstr,64); return;
      case fl_gt:  fl->valid = (strnicmp(sv,fl->fstr,64) > 0) ? TRUE : FALSE; return;
      case fl_lt:  fl->valid = (strnicmp(sv,fl->fstr,64) < 0) ? TRUE : FALSE; return;
    }
    return;  
  }
err:
  if (isLinkToLocalProcess(c->fecIdx)) 
  {
    dbglog("ignoring filter link error %.32s to local process",cc2str(cc));
    return;
  }
  fl->valid = TRUE;
  if (cc != last_cc)
  {
    feclogEx(TINE_LOGLEVEL_WARN, "filter link %s status now : %.32s",fl->key,cc2str(cc));
    last_cc = cc;
  }
}
FilterLink *getFilterLink(char *key,char *target,int comparator,int param)
{
  FilterLink *fl;
  int cc = 0;
  char dev[132],prp[64];
  DTYPE dout;

  strtrm(key); strtrm(target);
  for (fl=fltrLnkLst; fl != NULL; fl=fl->nxt)
  {
    if (strnicmp(fl->key,key,256)) continue;
    if (strnicmp(fl->fstr,target,256)) continue;
    if (fl->fcmp != comparator) continue;
    return fl;
  }
  /* it's new! */
  fl = (FilterLink *)SystemCalloc(1,sizeof(FilterLink));
  if (fl == NULL) ccerr(out_of_local_memory);
  strncpy(fl->key,key,256);
  strncpy(fl->fstr,target,64);
  fl->fval = strtod(target,NULL);
  fl->fcmp = comparator;
  fl->param1 = param;
  /* split the name and start the link */
  DTYPEZERO(dout);
  parseFullDeviceName(key,dev,prp,FALSE);
  dout.dFormat = CF_DEFAULT;
  dout.dArrayLength = 128;
  dout.data.bptr = fl->buf;
  fl->lid = AttachLinkEx2(dev,prp,&dout,NULL,CA_READ,1000,_cbFltrLnk,CM_TIMER,UNASSIGNED_CALLBACKID,fl);
  if (fl->lid < 0) ccerr(-fl->lid);
  fl->active = TRUE;
  fl->valid = TRUE;
  fl->nxt = fltrLnkLst;
  fltrLnkLst = fl;
  feclog("new filter %s : value %s %s is valid",key,getComparatorString(comparator),target);
  return fl;
err:
  if (fl != NULL) SystemFree(fl);
  feclogEx(TINE_LOGLEVEL_ERROR, "cannot create filter %s : %.32s",key,cc2str(cc));
  return NULL;
}
FilterLink *getFilterLinkFromParsableString(char *parsableString)
{
  int cmpr=0, cc, len=1, param = -1;
  char *target = NULL;
  char key[256], ps[512];
  char *c,*pc;

  if (parsableString == NULL || parsableString[0] == 0) return NULL;
  strncpy(ps, parsableString, 512); ps[511] = 0;
  if ((c=strchr(ps, '&')) != NULL)
  {
    *c++ = 0; /* terminate 'ps' here */
    switch (*c)
    {
      case 's': /* severity */
      case 't': /* tolerance */
        if ((c=strchr(c, '=')) != NULL)
        { /* parameter value given */
          c++;
          param = atoi(c);
        }
        break;
      default:
        break;
    }
  }
  if ((c=strstr(ps,"!=")) != NULL) { cmpr = fl_neq; len = 2; }
  else if ((c=strstr(ps,"==")) != NULL) { cmpr = fl_eq; len = 2; }
  else if ((c=strstr(ps,"=")) != NULL) cmpr = fl_eq;
  else if ((c=strstr(ps,"<")) != NULL) cmpr = fl_lt;
  else if ((c=strstr(ps,">")) != NULL) cmpr = fl_gt;
  else ccerr(parsing_error);
  *c = 0;
  c += len;
  strncpy(key,strtrm(ps),252); key[252] = 0;
  if ((pc=strchr(key,'[')) == NULL)
  { /* make all keys of the form /ctx/srv/dev[prp] */
    if ((pc=strrchr(key,'/')) == NULL) ccerr(parsing_error);
    *pc = '['; strcat(key,"]");
  }
  target = strtrm(c);
  return getFilterLink(key,target,cmpr,param);
err:
  feclogEx(TINE_LOGLEVEL_ERROR, "could not parse filter string %s : %.32s",parsableString,cc2str(cc));
  return NULL;
}
void rmvFilterLink(FilterLink *flnk)
{
  FilterLink *fl, *flprv=NULL;
  for (fl=fltrLnkLst; fl != NULL; fl=fl->nxt)
  {
    if (fl == flnk)
    {
      if (fl == fltrLnkLst) fltrLnkLst = fl->nxt;
      else flprv->nxt = fl->nxt;
      CloseLink(fl->lid);
      SystemFree(fl);
      return;
    }
    flprv = fl;
  }
}
void flushFilterLinks(void)
{
  FilterLink *fl, *flnxt=NULL;
  AWSLstEntry *aw;
  int i;
  for (fl=fltrLnkLst; fl != NULL; fl=flnxt)
  {
    if (gIsRunningAsServer)
    {
      for (i=0; i<nHistoryRecords; i++) hstTbl[i].fltr = NULL;
      for (aw=almWatchList; aw != NULL; aw = aw->nxt) aw->fltr = NULL;
    }
    flnxt = fl->nxt;
    CloseLink(fl->lid);
    SystemFree(fl);
  }
  fltrLnkLst = NULL;
}
int dumpFilterLinks(void)
{
  FilterLink *fl;
  int n = 0;
  ttyoutput(" Current Filters (local alarms/histories)\n");
  for (fl=fltrLnkLst; fl != NULL; fl=fl->nxt)
  {
    if (!fl->active) continue;
    ttyoutput("  %s %s %s : %s",fl->key,getComparatorString(fl->fcmp),fl->fstr,(fl->valid?"valid":"not valid"));
    n++;
  }
  if (n == 0) ttyoutput("  no active filters assigned");
  return 0;
}
int getFilterComparator(char *str)
{
  if (!strcmp(str,"==")) return fl_eq;
  if (!strcmp(str,"=")) return fl_eq;
  if (!strcmp(str,"!=")) return fl_neq;
  if (!strcmp(str,"<")) return fl_lt;
  if (!strcmp(str,">")) return fl_gt;
  return fl_eq;
}
char *getComparatorString(int cmp)
{
  switch (cmp)
  {
    case fl_eq: return "=";
    case fl_neq: return "!=";
    case fl_lt: return "<";
    case fl_gt: return ">";
  }
  return "?";
}
int putMcaValues(McaTblEntry *mca,ConTblEntry *cmca,ConTblEntry *c)
{
  DTYPE d;
  BYTE outbuf[256];
  int fsiz = fmtsizeof(cmca->formatOut), dsiz = fsiz-4;
  int cc, idx = c->mcaIndex-1;
  d.dArrayLength = 1;
  d.dFormat = cmca->formatOut;
  strncpy(d.dTag,cmca->strTagOut,TAG_NAME_SIZE);
  c->lastDataSize = 1;
  c->datastamp = cmca->datastamp;
  c->systemstamp = cmca->systemstamp;
  c->xferReason = cmca->xferReason;
  c->dtimestamp = cmca->dtimestamp;
  c->dtimestampUSEC = cmca->dtimestampUSEC;
  c->dataStatus = cmca->dataStatus;
  if (mca->isElevatedFormat)
  { /* an elevated format carries the status */
    d.data.bptr = outbuf;
    PutValuesFromAny(&d,cmca->dataPtr,cmca->formatOut,TRUE,fsiz,cmca->sizeOut,idx);
    switch (c->formatOut)
    {
      case CF_INT16: 
      {
        int ival;
        UINT16 sval;
        memcpy(&ival,outbuf,sizeof(int));
        sval = (UINT16)ival;
        memcpy(c->dataPtr,&sval,sizeof(short));
        /* *(UINT16 *)c->dataPtr = (UINT16)(*(int *)outbuf); << -Wstrict-aliasing complains */
        break;
      }
      case CF_DOUBLE: dsiz = fsiz - 8;
      default: memcpy(c->dataPtr,outbuf,dsiz); break;
    }
    if (cmca->linkStatus == 0)
    { /* the overall link status is okay ... */
      if (c->formatOut != CF_DOUBLE) 
      {
        memcpy(&cc,&outbuf[dsiz],sizeof(int));
        /* cc = *(int *)&outbuf[dsiz]; */
      }
      else 
      {
        double dval;
        memcpy(&dval,&outbuf[dsiz],sizeof(double));
        cc = (int)dval;
        /* cc = (int)(*(double *)&outbuf[dsiz]); */
      }
      if (cc & CE_SENDDATA) return TRUE; /* returned status should be treated as 'data status' */
      strncpy(c->lastStatusString,cc2str(cc),STATUS_SIZE);
      c->lastLnkErrSrc = c->linkStatus;
      c->linkStatus = cc;
      return FALSE;
    }
  }
  else
  {
    d.data.bptr = (BYTE *)c->dataPtr;
    PutValuesFromAny(&d,cmca->dataPtr,cmca->formatOut,TRUE,fsiz,cmca->sizeOut,idx);
  }
  return TRUE;
}
void _cbMcaXRef(int id, int cc,void *ref)
{
  McaTblEntry *mca;
  ConTblEntry *c = NULL, *cmca;
  McaLnkItem *item, *lastitem;
  int i, idx, backout=FALSE;
  if (WaitForSystemMutex(hMcaLinkMutex,-1) != 0) return;
  if ((mca=(McaTblEntry *)ref)== NULL)
  {
    if (tineDebug || errorDebug) dbglog("callback cross reference to MCA table not valid");
    ReleaseSystemMutex(hMcaLinkMutex);
    return;
  }
  cmca = conTbl[id];
  if (cmca->mode == CM_CANCEL) 
  { /* the reference won't be valid anyway */
    if (tineDebug) dbglog("callback to canceled MCA link ignored");
    ReleaseSystemMutex(hMcaLinkMutex);
    return; 
  }
  if (cc == reset_mca_property) 
  {
    msglog(cc,"received reset mca property signal for /%.32s/%.32s [%.64s]",
      mca->mcaContext,mca->mcaServerName,mca->mcaProperty);
    backout = TRUE;
  }
  if (cc == link_timeout && cmca->tocounter > TO_RETRY_THRESHOLD*5) backout = TRUE;
  if (backout)
  { /* server seems to have disappeared => revert to original conditions */
    msglog(cc,"Mapped MCA Link /%.32s/%.32s [%.64s] has gone down : revert to original conditions",
      mca->mcaContext,mca->mcaServerName,mca->mcaProperty);
    for (item=mca->links; item != NULL; item=item->nxt)
    {
      if ((i=item->linkId) > 0 && (c=conTbl[i]) != NULL)
      { 
        c->mcaIndex = 0;
        c->mcaLink = NULL;
        c->boundToId = item->boundToId; /* original link was bound ? */
        c->mode = cmca->mode;
        if (c->boundToId > 0)
        { /* re-insert into the dependency list */
          addDependentLink(c->linkId);
        }
        else
        { /* don't let these loose yet */
          c->heartbeat = HBT_SUB_PREPENDING; 
        }
      }
    }
    _closeLink(mca->linkId);
    removeMcaLinkEntry(mca);
    cmca->cbRef = NULL;
    /* backed out completely: now let go */
    for (i=0; i<nConnectionTableEntries; i++)
    {
      if (conTbl[i] != NULL && conTbl[i]->heartbeat == HBT_SUB_PREPENDING) 
        conTbl[i]->heartbeat = HBT_SUB_PENDING;
    }
    ReleaseSystemMutex(hMcaLinkMutex);
    return;
  }
  for (item=mca->links; item != NULL; item=item->nxt)
  {
    if ((i=item->linkId) > 0 && (c=conTbl[i]) != NULL && !item->isCanceled)
    { /* prepare the data element */
      idx = c->mcaIndex - 1;
      if (c->formatOut == CF_DEFAULT)
      {
        c->formatOut = cmca->formatOut;
        c->sizeOut = 1;
      }
      if (idx >= 0 && idx < (int)cmca->sizeOut)
      {
        if (putMcaValues(mca,cmca,c))
        { /* signal to use the incoming cc: will be the same as cmca->linkStatus */
          strncpy(c->lastStatusString,cmca->lastStatusString,STATUS_SIZE);
          c->linkStatus = cmca->linkStatus;
          c->lastLnkErrSrc = cmca->lastLnkErrSrc;
        }
        if (linkWaitingOnStartup(c)) _cbExecLink(c->linkId,cc); /* stop waiting */
        c->tocounter = 0;
        c->suppressCallback = !isOOT(c,c->sizeBytesOut);
        NotifyClient(i,c->linkStatus);
        if (c->mcaIndex == -1) 
        { /* the reference has been removed in the callback ! */
          ReleaseSystemMutex(hMcaLinkMutex);
          return;
        }
      }
    }
  }
  for (lastitem=item=mca->links; item != NULL; item=item->nxt)
  { /* scan for cancelations .... */
    if (item->isCanceled)
    {
      if (item == mca->links)
      { /* must be at the beginning */
        lastitem = mca->links = item->nxt;
      }
      else
      { /* splice the ends */
        lastitem->nxt = item->nxt;
      }
      SystemFree(item);
      mca->nlinks--;
      item = lastitem;
    }
    if ((lastitem=item) == NULL) break;
  }
  if (mca->nlinks <= 0 && mca->linkId > 0 && conTbl[mca->linkId] != NULL)
  { /* no active links left -> remove the mca link */
    if (time(NULL) > (time_t)(conTbl[mca->linkId]->starttime + 1))
    { /* but don't kill anybody just starting out !*/
      cmca->cancelPending = TRUE;
      cmca->timemark = (UINT32)time(NULL);
    }
  }
  ReleaseSystemMutex(hMcaLinkMutex);
}
#endif /* !SMALL_TINE_LIB */
void _cbXRef(int id, int cc)
{
  if (id < 0 || id >= nCbXRefTblEntries)
  {
    if (tineDebug || errorDebug) 
      dbglog("bad callback identifier (%d) in cross-reference table (%d entries)",id,nCbXRefTblEntries);
    return;
  }
  if (cbXRefTbl[id]->linkId < 0)
  {
    if (tineDebug || errorDebug) dbglog("callback identifier (%d) called but is no longer active",id);
    return;
  }
  if (cbXRefTbl[id]->cbfcn != NULL) TCBCALL(cbXRefTbl[id],cc);
}
int getCbXRefTblId(void)
{
  int i = code_failure, cc = resources_exhausted;

  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) != 0) return -semaphore_busy;
  if (cbXRefTbl == NULL && 
      (cbXRefTbl=(cbXRefTblEntry **)SystemCalloc(gConTblCapacity,sizeof(cbXRefTblEntry *))) == NULL)
    ccerr(out_of_client_memory);
  for (i=0; i<gConTblCapacity; i++)
  {
    if (cbXRefTbl[i] == NULL)
    {
      if ((cbXRefTbl[i]=(cbXRefTblEntry *)SystemCalloc(1,sizeof(cbXRefTblEntry))) == NULL) 
        ccerr(out_of_client_memory);
      nCbXRefTblEntries++;
      ccerr(0);
    }
    if (cbXRefTbl[i]->linkId == -1) ccerr(0); /* available slot */
  }
err:
  ReleaseSystemMutex(hLinkTblMutex);
  return cc != 0 ? -cc : i;
}
int isGlobalsLink(char *ctx,char *srv,char *prp)
{
  if (IsStockProperty(prp) == STOCK_PROP) return FALSE;
  if (stricmp(srv,GLOBALSRV) == 0 && isGlobalsParameter(ctx,srv,prp)) return TRUE;
  if (stricmp(srv,"CYCLER") != 0) return FALSE;
  if (strstr(prp,"CycleNumber") == NULL) return FALSE;
  return TRUE;
}
#ifndef SMALL_TINE_LIB
DTYPE *remapAdjustableInputData(DTYPE *din)
{
  int k,n,xsiz,xoff,hsiz,cc = 0,dinsize = 0;
  char *c, *sa = NULL;
  static DTYPE dinx;

  switch (din->dFormat)
  { /* scan thru the input for the total length */
    case CF_KEYVALUE:
    case CF_STRING:
      for (k=0; k<(int)din->dArrayLength; k++)
      { 
        if ((c = din->data.strptr[k]) == NULL) break;
        if (c == NULL) ccerr(-argument_list_error);
        dinsize += (int)strlen(c) + 1; /* add the terminating 0 */
      }
      if ((sa=(char *)SystemCalloc(1,dinsize)) == NULL) ccerr(out_of_client_memory);
      for (k=0,n=0; k<(int)din->dArrayLength; k++)
      { /* now fill in the string buffer */
        if ((c = din->data.strptr[k]) == NULL) break;
        strcpy(&sa[n],c);
        n += (int)strlen(c) + 1;
      }
      break;
    case CF_AIMAGE:
      hsiz = IMAGE_HDR_SIZE + 2*sizeof(SINT32);
      for (k=0; k<(int)din->dArrayLength; k++)
      { 
        dinsize += hsiz + ((DIMAGE *)din->data.vptr)[k].frameBufferSize;
      }
      if ((sa=(char *)SystemCalloc(1,dinsize)) == NULL) ccerr(out_of_client_memory);
      xoff = (int)din->dArrayLength * hsiz;
      for (k=0; k<(int)din->dArrayLength; k++)
      { /* now fill in the buffer */
        memcpy(&sa[k*hsiz],&((DIMAGE *)din->data.vptr)[k],IMAGE_HDR_SIZE);
        xsiz = ((DIMAGE *)din->data.vptr)[k].frameBufferSize;
        memcpy(&sa[k*hsiz+IMAGE_HDR_SIZE],&xoff,4);
        memcpy(&sa[k*hsiz+IMAGE_HDR_SIZE+4],&xsiz,4);
        memcpy(&sa[xoff],((DIMAGE *)din->data.vptr)[k].frameBuffer,xsiz);
        xoff += xsiz;
      }
      break;
    case CF_ASPECTRUM:
      hsiz = SPECTRUM_HDR_SIZE + 2*sizeof(SINT32);
      for (k=0; k<(int)din->dArrayLength; k++)
      { 
        dinsize += hsiz + ((ASPECTRUM *)din->data.vptr)[k].spectBufferSize * sizeof(float);
      }
      if ((sa=(char *)SystemCalloc(1,dinsize)) == NULL) ccerr(out_of_client_memory);
      xoff = (int)din->dArrayLength * hsiz;
      for (k=0; k<(int)din->dArrayLength; k++)
      { /* now fill in the buffer */
        memcpy(&sa[k*hsiz],&((ASPECTRUM *)din->data.vptr)[k],SPECTRUM_HDR_SIZE);
        xsiz = ((ASPECTRUM *)din->data.vptr)[k].spectBufferSize * sizeof(float);
        memcpy(&sa[k*hsiz+SPECTRUM_HDR_SIZE],&xoff,4);
        memcpy(&sa[k*hsiz+SPECTRUM_HDR_SIZE+4],&xsiz,4);
        memcpy(&sa[xoff],((ASPECTRUM *)din->data.vptr)[k].spectBuffer,xsiz);
        xoff += xsiz;
      }
      break;
  }
err:
  if (cc) return NULL;
  memcpy(&dinx,din,sizeof(DTYPE));
  dinx.data.vptr = sa;
  sprintf(dinx.dTag,"%d",din->dArrayLength);
  dinx.dArrayLength = dinsize;
  return &dinx;
}
#endif /* !SMALL_TINE_LIB */
#define tineUrlPrefix "tine:/"
#define tineUrlPrefixLength 6
int _attachLink(const char *pDevName,const char *devProp,DTYPE *dout,DTYPE *din,short access,
                int pollingRate,void *cbFcn,int mode,UINT32 cbId,void *reference,int reuseIndex)
{
  char fulname[FULL_NAME_MAXSIZE+4],devProperty[PROPERTY_NAME_SIZE+4],devName[FULL_NAME_MAXSIZE+4],devField[FIELD_NAME_SIZE+4];
  int i=-1,baseMode,modeFlags,dinsize=0,doutsize=0,cc=0,glid=-1;
  int isENS,isGrouped,isSyncGroup,isBfPrp=FALSE,isBf=FALSE,isGlb=FALSE;
  int isTrace = FALSE;
  int suppressDuplication,wait,useOnError,suppressQueue=FALSE,hasReference=FALSE;
# ifndef SMALL_TINE_LIB
  int pid,hsiz;
  DTYPE doutx, *dinx; /* internal copy for possible use with CF_STRING */
# endif
# ifdef MULTITHREADED
  int busy_retries=50;
# endif
  GrpTblEntry *gt = NULL;
  struct structStruct *s;
  char *c = NULL, *ctx = NULL,*srv = NULL,*dev = NULL,*ap = NULL,*pdev = (char *)pDevName, devNameBuffer[1048];
  char dsrv[2*CONTEXT_NAME_SIZE+4];
  struct timeval tv_stamp;
  BYTE *ivals =  NULL;
  _Bool *bvals = NULL;

  if (pdev == NULL || devProp == NULL) ccerr(argument_list_error);
  if (strnicmp(pdev,tineUrlPrefix,tineUrlPrefixLength) == 0) pdev += tineUrlPrefixLength;
  if (pollingRate < 0) ccerr(invalid_interval);
  if (pollingRate > MAXALLOWED_POLLINTVL) pollingRate = MAXALLOWED_POLLINTVL;
  if (reference != NULL) hasReference = TRUE;
  /* get rid of possible nonsense (client call should not be using these bits : */
  access &= ~(CA_REPEAT|CA_ALARM|CA_HIST|CA_FIRST|CA_NOCALLBACKS); 
  if (!(access & CA_NORETRY) && gAlwaysRetry) access |= CA_RETRY;
  gettimeofday(&tv_stamp,(struct timezone *)NULL);
  strncpy(devName,pdev,FULL_NAME_MAXSIZE);
  strncpy(devProperty,devProp,PROPERTY_NAME_SIZE);
  strntrm(devName,FULL_NAME_MAXSIZE);
  strntrm(devProperty,PROPERTY_NAME_SIZE);
  if (isBitfieldAccess(dout)) 
  { /* is a bitfield ... */
    isBf = TRUE;
    acquireAndRegisterBitfield(devName,dout->dTag);
    if ((cc=splitFullName(devName,dsrv,NULL)) != 0) ccerr(cc);
    if (isBitfieldProperty(dsrv,dout->dTag,devProperty,devField)) isBfPrp = TRUE;
  }
  strncpy(fulname,devName,FULL_NAME_MAXSIZE); fulname[FULL_NAME_MAXSIZE] = 0;
  if (getCdiLinkNameHook != NULL && (dev=getCdiLinkNameHook(devName,devNameBuffer)) != NULL)
  { /* this call is going to the CDI layer ... */
    cc = -not_implemented;
    if (cdiAttachLinkHook != NULL) 
    {
      if (hasReference)
      { /* use a callback interceptor */
        if ((i=getCbXRefTblId()) < 0) return i;
        cbXRefTbl[i]->cbfcn = (TCBFCNP)cbFcn;
        cbXRefTbl[i]->cbId = cbId;
        cbXRefTbl[i]->cbRef = reference;
        cbXRefTbl[i]->hasReference = TRUE;
        cbFcn = (void *)_cbXRef;
        cbId = i;
      }
      cc = cdiAttachLinkHook(dev,devProperty,dout,din,access,pollingRate,(TCBFCNP)cbFcn,mode,cbId);
      if (hasReference) cbXRefTbl[i]->linkId = cc;
    }
    if (cc != -server_redirection) return MAKE_CDI_LINKID(cc);
    strncpy(fulname,dev,64);
    dev = NULL; cc = 0;
  }
  lastLnkErrSrc = LNK_ERROR_LCL;
  isENS = isENSCall(devName);

  if ((ctx=strtok_r(fulname,DELIMITER,&c)) == NULL) ccerr(argument_list_error);
  if ((srv=strtok_r(NULL,DELIMITER,&c)) == NULL) 
  { /* only one given */
    srv = ctx; ctx = NULL; dev = &fulname[FULL_NAME_MAXSIZE];
  }
  else if ((dev=strtok_r(NULL,DELIMITER,&c)) == NULL) 
  { /* only two given */
    if (ctx == fulname) 
    { /* no leading delimiter */
      dev = srv; srv = ctx; ctx = NULL;
    }
    else dev = &fulname[FULL_NAME_MAXSIZE];
  }
  else if ((ap=strtok_r(NULL,"\0",&c)) != NULL) 
  { /* a '/' in the device name */
    dev[strlen(dev)] = '/';  /* put it back */
  }
  if (isGlobalsLink(ctx,srv,devProperty))
  {
    mode &= ~(CM_CONNECT|CM_STREAM);
    baseMode = BASEMODE(mode);
    if (baseMode != CM_SINGLE && baseMode != CM_REGISTER && baseMode != CM_RECEIVE)
    { /* not already in 'GLOBALS' mode */
      modeFlags = mode & 0xf400;
      mode = modeFlags | CM_RECEIVE;
    }
    isGlb = TRUE;
  }
  if (gUseSynchronousNotification) mode |= CM_SYNCNOTIFY;
  suppressDuplication = mode & CM_NODUPLICATES; 
  isGrouped = mode & CM_GROUPED;
  isSyncGroup = mode & CM_SYNCGROUP;
  if (isSyncGroup) isGrouped = TRUE;
  wait = mode & CM_WAIT;
  useOnError = mode & CM_USEONERROR; 
  mode &= ~(CM_REMOVEVOLATILE); /* strip off these bits */
  if ((mode & CM_MCAST) && !gListenForMCasts)
  {
    if ((cc=ListenForMulticasts()) != 0) ccerr(cc);
  }
  baseMode = BASEMODE(mode);
  pollingRate = makeValidPollingInterval(pollingRate,baseMode);
  if (baseMode == CM_EVENT && pollingRate < 30000) pollingRate = 30000;
  if (tineDebug) dbglog("Attach %.192s %.64s (%d msec): %s",devName,devProperty,pollingRate,modetoa[baseMode]);
  if (isGlb)
  { /* re-form devName ... */
    if (ctx != NULL) sprintf(devName,"/%.32s/%.32s/%.64s",ctx,srv,devProperty); 
    else sprintf(devName,"%.64s",devProperty);
    if ((glid=_recvGlobal(devName, dout, NULL, 0, NULL)) < 0)
    {
      msglog(glid, "Unable to subscribe to GLOBALS link %.192s %.64s: %.32s", devName, devProperty, cc2str(glid));
      return glid;
    }
  }
  if (baseMode > CM_EVENT  || baseMode < CM_CANCEL) ccerr(illegal_mode);
  if (baseMode == CM_DATACHANGE || baseMode == CM_EVENT || baseMode == CM_RECEIVE) isSyncGroup = isGrouped = 0;
  if (SystemSuppressHeartbeatNotification && (baseMode == CM_DATACHANGE)) suppressDuplication = TRUE;
  if (isGrouped) suppressQueue = TRUE;
  if (din)
  {
    din->dTimeStamp = 0;
    if (din->dFormat == CF_NULL) din->dArrayLength = 0;
    else if (din->dArrayLength == 0) din->dFormat = CF_NULL;
    dinsize = (int)(fmtsizeof(din->dFormat)*din->dArrayLength);
    if (dinsize == 0 && din->dFormat != CF_NULL)
    { /* sanity check against wild format code */
      din->dFormat = CF_NULL;
      din->dArrayLength = 0;
    }
    dinsize += getFormatHeaderSize(din->dFormat);
    if (dinsize > 0 && din->data.vptr == NULL) ccerr(argument_list_error);
    if (din->dFormat == CF_STRUCT && (s=findstructEx(din->dTag,ctx,srv)) != NULL) 
    { /* registered tagged structure */
      dinsize *= s->rawlen;
    }
    else if (isAdjustableLengthFormat(din->dFormat))
    { /* make sure the input is serialized 'properly' */
#     ifndef SMALL_TINE_LIB
      if ((dinx=remapAdjustableInputData(din)) == NULL) ccerr(out_of_client_memory);
      dinsize = dinx->dArrayLength;
      din = dinx; /* re-assign (dinx lives long enough to make it thru getConTblEntry) */
#     else
      ccerr(-illegal_format);
#     endif
    }
    else if (din->dFormat == CF_HISTORY)
    {
      if ((dinsize=getCarriedFormatSize(din->dTag)) < 0) ccerr(-dinsize);
      dinsize += HISTORY_HDR_SIZE;
      dinsize *= din->dArrayLength;
    }
    else if (din->dFormat == CF_MDA)
    {
      if ((dinsize=getCarriedFormatSize(din->dTag)) < 0) ccerr(-dinsize);
      dinsize *= din->dArrayLength;
      dinsize += MDA_HDR_SIZE;
    }
    else if (din->dFormat == CF_BOOLEAN)
    {
      if (gMarshallBooleanAsByte)
      { /* we've been passed a pointer to a smaller object than the one carried */
        bvals = (_Bool *)din->data.vptr;
        ivals = (BYTE *)alloca(dinsize);
        if (ivals == NULL) ccerr(out_of_local_memory);
        memcpy(ivals, din->data.vptr, din->dArrayLength);
        marshallBoolean(ivals, din->dArrayLength, LTOR);
        din->data.lptr = (SINT32 *)ivals;
      }
    }
    else if (din->dFormat == CF_NULL) 
    {
      memset(din->dTag,0,TAG_NAME_SIZE);
    }
  }
  if (dout)
  {
    dout->dTimeStamp = 0;
    if (dout->dFormat == CF_NULL) dout->dArrayLength = 0;
    else if (dout->dArrayLength == 0) dout->dFormat = CF_NULL;
    doutsize = (int)(fmtsizeof(dout->dFormat)*dout->dArrayLength);
    if (doutsize == 0 && dout->dFormat != CF_NULL)
    { /* sanity check against wild format code */
      dout->dFormat = CF_NULL;
      dout->dArrayLength = 0;
    }
    doutsize += getFormatHeaderSize(dout->dFormat);
    if (dout->dFormat == CF_STRUCT && (s=findstructEx(dout->dTag,ctx,srv)) != NULL) 
    { /* registered tagged structure */
      doutsize *= s->size;
    }
#   ifndef SMALL_TINE_LIB
    else if (isAdjustableLengthFormat(dout->dFormat))
    { /* take a good guess as to the total capacity */
      doutsize = (int)(fmtsizeof(dout->dFormat)*dout->dArrayLength);
      hsiz = getFormatHeaderSize(dout->dFormat) + 2*sizeof(SINT32);
      doutsize *= hsiz + getFormatStandardCapacity(dout->dFormat);
      memcpy(&doutx,dout,sizeof(DTYPE));
      sprintf(doutx.dTag,"%d",dout->dArrayLength);
      doutx.dArrayLength = doutsize;
      dout = &doutx; /* re-assign */
    }
    else if (dout->dFormat == CF_HISTORY)
    {
      if ((doutsize=getCarriedFormatSize(dout->dTag)) < 0) ccerr(-doutsize);
      doutsize += HISTORY_HDR_SIZE;
      doutsize *= dout->dArrayLength;
    }
#   endif
    else if (dout->dFormat == CF_NULL) 
    {
      memset(dout->dTag,0,TAG_NAME_SIZE);
    }
    else if (!isQueueableFormat(dout->dFormat))
    {
      suppressDuplication = FALSE;
      suppressQueue = TRUE;
      mode |= CM_SYNCNOTIFY;
    }
    else if (dout->dFormat == CF_MDA)
    { /* re-calculate the output bytes */
      if ((doutsize=getCarriedFormatSize(dout->dTag)) < 0) ccerr(-doutsize);
      doutsize *= dout->dArrayLength;
      doutsize += MDA_HDR_SIZE;
    }
    if (isBf && !isBfPrp)
    { /* maybe bitfield applied to the device ? */
      if (isBitfieldProperty(dsrv,dout->dTag,dev,devField)) isBfPrp = TRUE;
    }
  }
  if (dout == NULL || dout->dArrayLength == 0) access &= ~(CA_READ);
# ifdef MULTITHREADED
retryGetEntry:
# endif
  if ((i=getConTblEntry(dout,din,ctx,srv,dev,devProperty,access,reuseIndex)) < 0)
  { /* could not get a link table entry ! */
    if (i != -link_exists)
    { /* could be semaphore_busy or address_unknown, etc. */
#     ifdef MULTITHREADED
      if (i == -semaphore_busy) 
      {
        YieldThread();
        if (busy_retries-- > 0) goto retryGetEntry;
      }
#     endif
      lastLnkErr = -i; strncpy(gLastStatusString,erlst[lastLnkErr],STATUS_SHORTSIZE);
      msglog(-i,"Unable to subscribe to %.192s %.64s: %.32s",devName,devProperty,cc2str(-i));
      if (cbFcn != NULL && cbId != UNASSIGNED_CALLBACKID) 
      { /* in case of server_redirect */
        if (hasReference) ((XTCBFCNP)cbFcn)(cbId,-i,reference); else ((TCBFCNP)cbFcn)(cbId,-i); 
      }
      ccerr(-i);
    }
    if (dout != NULL && (i=dout->xferReason) >= 0 && i < nConnectionTableEntries)
    { /* latch on to the existing parent link ('i' reassigned) ... */
      if (BASEMODE(conTbl[i]->mode) == CM_SINGLE)  
      { /* was a SINGLE link */
        if (tineDebug) dbglog("attaching %.192s %.64s to link %d",devName,devProperty,i);
        if (dout->data.vptr != NULL) conTbl[i]->dataPtr = (char *)dout->data.vptr;
        else dout->data.vptr = (void *)conTbl[i]->dataPtr;
      }
    }
    /* note: link_exists only returned if 'allowCommonLinks' = FALSE (not the default) */
    ccerr(link_exists);
  }
  if (dout != NULL && dout->dFormat == CF_BOOLEAN && gMarshallBooleanAsByte)
  {
    conTbl[i]->dataMarshalled = dout->data.cptr;
    if (conTbl[i]->allocatedBytes)
    { /* dout->data.cptr was passed as NULL */
      conTbl[i]->dataMarshalled = conTbl[i]->bufVtl = (char *)SystemCalloc(dout->dArrayLength, sizeof(_Bool));
      if (conTbl[i]->dataMarshalled == NULL) ccerr(out_of_client_memory);
    }
    else
    { /* re-wire dataPtr */
      conTbl[i]->dataPtr = (char *)SystemCalloc(dout->dArrayLength, sizeof(SINT32));
      if (conTbl[i]->dataPtr == NULL) ccerr(out_of_client_memory);
      conTbl[i]->allocatedBytes = TRUE; /* get the cleanup right ! */
    }
  }
  makeLinkKey(conTbl[i]->key, conTbl[i]->ctxName, conTbl[i]->expName, conTbl[i]->devName, conTbl[i]->prpName);
  isTrace = isTraceLink(i);
  if (isTrace) traceLinkOutput(conTbl[i], "attachLink", "link has been made");

  if (cbId == UNASSIGNED_CALLBACKID) cbId = i; else conTbl[i]->hasUserCallbackId = TRUE;
  if (access & CA_FIXTIMESTAMP) conTbl[i]->rbModFlags |= RB_USE_TIMESTAMP;
  if (conTbl[i]->tineProtocol == 0) conTbl[i]->tineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
  if (useConnectedSockets) mode |= CM_CONNECT;
  if (mode & CM_NETWORK)
  { /* a request for multicasts must use UDP datagrams */
    mode &= ~(CM_CONNECT);
    FecTbl[conTbl[i]->fecIdx].inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP;
  }
  if (glid >= 0) 
  { /* is a re-mapped globals link */
    if (BASEMODE(mode) == CM_RECEIVE) mode = CM_REGISTER;
    conTbl[i]->glbTblId = glid + 1;
    if (conTbl[i]->boundToId == 0) 
    { /* not itself a dependent link */
      glbTbl[glid]->conTblId = i + 1;
    }
    else
    { /* already a dependent link: up the link count */
      glbTbl[glid]->tv = tv_stamp;
      glbTbl[glid]->nlnks++;
    }
  }
  conTbl[i]->lastLnkErrSrc = LNK_ERROR_LCL;
  conTbl[i]->pollingInterval  = pollingRate;
  conTbl[i]->mode      = mode;
  conTbl[i]->cbId      = cbId;
  conTbl[i]->timeLastSync   = 0; /* set in prepSubRequest */
  conTbl[i]->timeMsecLastSync = 0;
  conTbl[i]->heartbeat = mode == CM_REGISTER ? HBT_NONE : HBT_SUB_PREPENDING;
  if (glid >= 0) conTbl[i]->heartbeat = HBT_NONE;
  conTbl[i]->access    = access;
  conTbl[i]->notifyPending   = 1;
  conTbl[i]->needsNotification = FALSE;
  if (baseMode > CM_SINGLE) conTbl[i]->tocounter = 0;
  conTbl[i]->cbRef = reference;
  conTbl[i]->hasReference = hasReference;
  conTbl[i]->linkId    = i;
  conTbl[i]->sizeBytesOut = doutsize;
  if (suppressQueue) conTbl[i]->queueDepth = 0;
  conTbl[i]->cbfcn     = (TCBFCNP)cbFcn;
  if (dout == NULL) isBfPrp = useOnError = FALSE;
# ifndef SMALL_TINE_LIB
  if (isBfPrp) appendReLnkLst(conTbl[i],dout->dTag,dout->dFormat,conTbl[i]->sizeOut,has_bitfield_tag);
  if (conTbl[i]->rlkEntry != NULL)
  { /* is a re-linked link! */
    appendReLnkCb(conTbl[i]);
    conTbl[i]->cbfcn = (TCBFCNP)_relinkXrefCb;
    conTbl[i]->cbId = i;
    conTbl[i]->hasReference = FALSE;
  }
# endif
  if (suppressDuplication && doutsize > 0) conTbl[i]->bufCmp = (char *)SystemCalloc(1,(size_t)doutsize);
  if (useOnError && dout->dArrayLength > 0) putLnkErrorValue(i,dout->data.vptr);
  if (mode & CM_MCAST && baseMode > CM_SINGLE)
  { /* join the server's specific multicast group */
    INADR inaddr = fixMcastHaddr((BYTE *)&FecTbl[conTbl[i]->fecIdx].netAdr,!(FecTbl[conTbl[i]->fecIdx].addrFlags & ADDRFLAG_MCAST_MODERN));
    attachPortToMulticastGroup(mcastClnSck,(UINT16)gtMCastPort,inaddr,"server specific");
    conTbl[i]->mcastGrp = inaddr;
  }
# ifndef SMALL_TINE_LIB
  if ((pid=conTbl[i]->boundToId) > 0)
  { /* check if this is a dependent link */
    conTbl[i]->mode = CM_REGISTER;
    conTbl[i]->heartbeat = HBT_NONE;
    conTbl[i]->linkStatus = conTbl[pid]->linkStatus;
    conTbl[i]->lastDataSize = conTbl[pid]->lastDataSize;
    conTbl[i]->dtimestamp = conTbl[pid]->dtimestamp;
    conTbl[i]->dtimestampUSEC = conTbl[pid]->dtimestampUSEC;
    conTbl[i]->datastamp = conTbl[pid]->datastamp;
    conTbl[i]->systemstamp = conTbl[pid]->systemstamp;
    wait = FALSE;
    if (doutsize > 0)
    { /* copy the data */
      copyDpdData(conTbl[i],conTbl[pid]);
    }
    if (baseMode == CM_SINGLE && _cbExecLink != (TCBFCNP)cbFcn) 
    { /* caller does not want to be synchronous */
      conTbl[i]->needsNotification = TRUE;
      if (conTbl[i]->linkStatus != notification_pending) conTbl[i]->mode = CM_CANCEL;
    }
    else if (conTbl[pid]->hasNotifiedOnce)
    { /* only fire the callback if there's been at least one notification */
      if (cbFcn != NULL) 
      { /* fire the callback with the parent's data */
        if (hasReference) ((XTCBFCNP)cbFcn)(conTbl[i]->cbId,conTbl[i]->linkStatus,reference); 
        else ((TCBFCNP)cbFcn)(conTbl[i]->cbId,conTbl[i]->linkStatus); 
      }
      conTbl[pid]->needsNotification = FALSE;
    }
  }
  /* check for wildcard call */
  else if (strchr(devName,'*') != NULL  && !IsStockProperty(devProperty))
  { /* A wildcard link (love those ...) ! */
    if (dout != NULL && dout->dFormat == CF_DEFAULT)
    { /* catch this here */
      int ndevs = GetNumberOf(devName,"DEVICES");
      conTbl[i]->formatOut = dout->dFormat = CF_USTRING;
      dout->dArrayLength /= fmtsizeof(CF_USTRING);
      if (ndevs > 0)
      {
        if (ndevs > (int)dout->dArrayLength) ccerr(buffer_too_small);
        dout->dArrayLength = (UINT32)ndevs;
      }
      conTbl[i]->sizeOut = dout->dArrayLength;
    }
    if ((cc=assignWildcardLink(i)) != 0) ccerr(cc);
    if (access & CA_SYNC) wait = TRUE; /* have to let possible queries complete ! */
    if (conTbl[i]->mode == CM_REGISTER) mode = CM_REGISTER; /* if GENS wildcard */
  }
# endif /* SMALL_TINE_LIB */
  if (conTbl[i]->heartbeat == HBT_SUB_PREPENDING) 
  { /* let this loose now */
    conTbl[i]->heartbeat = HBT_SUB_PENDING;
  }
  if (isGrouped && (gt=getGrpTblEntry((size_t)cbFcn)) == NULL) isSyncGroup = isGrouped = 0;
  if (gt)
  {
    addLinkToGroup(i);
    if (isSyncGroup) gt->grpSyncLevel = GRP_SYNC_STARTSYNC;
  }
  if (wait && !conTbl[i]->hasNotifiedOnce)
  {
    conTbl[i]->tocounter = -1;
    if (WaitForSyncLink(i,(int)(2*conTbl[i]->pollingInterval)) != 0) conTbl[i]->linkStatus = link_timeout;
    if (!conTbl[i]->hasNotifiedOnce) NotifyClient(i,conTbl[i]->linkStatus);
  }
  else
  { /* in case single-threaded and the serverCycle() comes up next */
    ReturnToClientCycle++;
  }
err:
  if (bvals != NULL) din->data.vptr = bvals; /* restore the original reference pointer */
  if (cc || (mode != CM_REGISTER && !isENS))
  {
    msglog(cc,"AttachLink %.192s %.64s (%d bytes out %d bytes in) %s %s %d ms (cbId: %d, linkId: %d)",
      devName,devProperty,doutsize,dinsize, access & CA_WRITE ? "WRITE" : "READ",
      modetoa[BASEMODE(mode)],pollingRate,cbId,i);
  }
  return cc ? -cc : i;
}

TINE_EXPORT int gConTblCapacity = CONTBL_CAPACITY;
#if !defined(SMALL_TINE_LIB)
int gAllowAutoGrowConnectionTable = TRUE;
TINE_EXPORT void SetAllowAutoGrowConnectionTable(int value)
{
  gAllowAutoGrowConnectionTable = value;
  feclog("Auto-grow connection table set to %s",value ? "TRUE" : "FALSE");
}
/**
 * \note the FecTbl and SrvTbl are allocated with the original gConTblCapacity value
 * which is a 'worst case' scenario where every connection is to a unique FEC entry.
 * Most reasons for increasing the capacity involve multiple links to only a 'few' 
 * FECs.
 */
TINE_EXPORT int SetConnectionTableCapacity(int value)
{ /* gConTblCapacity established following allocation */
  int cc = 0;
  if (value < 10) return invalid_data;
# ifdef VXWORKS
  if (nConnectionTableEntries > 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"Client Connection Table already in use : reallocation not allowed");
    return already_assigned;
  }
# endif
# ifndef MULTITHREADED
  hLinkTblMutex = 0; /* need a recursive mutex flag */
# endif
  if (WaitForSystemMutex(hLinkTblMutex,-1) != 0) return mutex_error;
  if (value > gConTblCapacity)
  { /* only increase in size ... */
    if (conTbl != NULL)
    {
      ConTblEntry **newTbl = (ConTblEntry **)SystemCalloc(value,sizeof(ConTblEntry *));
      if (newTbl != NULL)
      { /* got new memory area: now copy the current pointers ... */
        ConTblEntry **oldTbl = conTbl; /* make a note ... */
        memcpy(newTbl, conTbl, gConTblCapacity*sizeof(ConTblEntry *));
        conTbl = newTbl; 
        feclog("Client Connection Table has been reallocated for %d entries",value);
        /* whatever thread is looking at 'conTbl' just prior or just before still see's something valid */
        YieldThread();
        /* should be safe to free the old table ... */
        SystemFree(oldTbl);
      }
      else
      {
        feclogEx(TINE_LOGLEVEL_ERROR,"Client Connection Table reallocation to %d entries failed",value);
        value = gConTblCapacity;
        cc = out_of_client_memory;
      }
    }
  }
  gConTblCapacity = value;
  ReleaseSystemMutex(hLinkTblMutex);
  return cc;
}
TINE_EXPORT int GetConnectionTableCapacity(void)
{
  return gConTblCapacity;
}
TINE_EXPORT void SetGlobalsTableCapacity(int value)
{ /* GlbTblCapacity established following allocation */
  if (value < 0) return;
  MaxNumGlobals = value;
}
TINE_EXPORT int GetGlobalsTableCapacity(void)
{
  return GlbTblCapacity;
}
AccessLockListItem *getAccessLockItem(char *context, char *server)
{
  AccessLockListItem *a=NULL;
  int cc = 0;

  if (WaitForSystemMutex(hLinkLstMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return NULL;
  for (a=accessLockLst; a != NULL; a = a->nxt)
  {
    if (strnicmp(context,a->ctx,CONTEXT_NAME_SIZE) != 0) continue;
    if (strnicmp(server,a->srv,EXPORT_NAME_SIZE) != 0) continue;
    ccerr(0); /* found it in the list */
  }
  if ((a=(AccessLockListItem *)SystemCalloc(1,sizeof(AccessLockListItem))) == NULL) ccerr(out_of_client_memory);
  strncpy(a->ctx,context,CONTEXT_NAME_SIZE);
  strncpy(a->srv,server,EXPORT_NAME_SIZE);
  a->lockLinkStatus = not_ready;
  a->nxt = accessLockLst;
  accessLockLst = a;
err:
  if (cc != 0) msglog(cc,"cannot get access lock : %.32s",cc2str(cc));
  ReleaseSystemMutex(hLinkLstMutex);
  return a;
}
#define LOCK_MATCHES(c,s,a) (\
  (!strnicmp(c,a->ctx,CONTEXT_NAME_SIZE) && !strnicmp(s,a->srv,EXPORT_NAME_SIZE)) ||\
  (c) == NULL || (s) == NULL) 
int rmvAccessLock(char *context, char *server,int wait)
{
  AccessLockListItem *a, *prv;
  int cc = not_locked;
  UNUSED(wait);
  if (WaitForSystemMutex(hLinkLstMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return mutex_error;
  for (prv=a=accessLockLst; a != NULL; prv=a, a=a->nxt)
  {
    if (LOCK_MATCHES(context,server,a))
    {
      a->lockLinkStatus = not_ready;
      cc = _setAccessLock(a->ctx,a->srv,LOCK_UNLOCKED,0);
      if (a == accessLockLst)
      {
        accessLockLst = a->nxt;
      }
      else
      {
        prv->nxt = a->nxt;
      }
      SystemFree(a);
      break;
    }
  }
  ReleaseSystemMutex(hLinkLstMutex);
  return cc;
}
void checkAccessLockItems(void)
{
  AccessLockListItem *a;
  time_t t = time(NULL);

  if (WaitForSystemMutex(hLinkLstMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return;
  for (a=accessLockLst; a != NULL; a = a->nxt)
  {
    if (a->lockLinkStatus == not_ready) continue;
    if ((a->lockType & 0xff) != LOCK_PERSISTENT) continue;
    if (t < a->lockDuration + a->lastSent - 5) continue;
    _setAccessLock(a->ctx,a->srv,a->lockType,MAX_LOCK_DURATION);
  }
  ReleaseSystemMutex(hLinkLstMutex);
}
void _cbAccessLock(int id, int cc)
{
  AccessLockListItem *a=NULL;
  if (WaitForSystemMutex(hLinkLstMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return;
  for (a=accessLockLst; a != NULL && a->lockLinkId != id; a = a->nxt);
  ReleaseSystemMutex(hLinkLstMutex);
  if (a == NULL)
  {
    if (tineDebug || errorDebug) dbglog("_cbAccessLock: cannot find access lock list item from link id %d !",id);
    return;
  }
  a->lockLinkStatus = cc;
  if (cc != 0)
  {
    feclogEx(TINE_LOGLEVEL_WARN, "access lock /%.32s/%.32s : %.32s",a->ctx,a->srv,cc2str(cc));
    return;
  }
}
int _tryAccessLock(char *context, char *server, int lockType)
{
  DTYPE din;
  char devsrv[128];
  short svals[2];
  int lt = lockType & 0xff;
  
  if (context == NULL || strlen(context) == 0) return invalid_parameter;
  if (server == NULL || strlen(server) == 0) return invalid_parameter;
  if (lt < 0 || lt > LOCK_ABORT) return invalid_parameter;

  sprintf(devsrv,"/%.32s/%.32s",context,server);
  svals[0] = lockType;
  svals[1] = 1;
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = 2;
  din.dFormat = CF_INT16;
  din.data.sptr = svals;

  return ExecLinkEx(devsrv,"ACCESSLOCK",NULL,&din,CA_WRITE|CA_NOCALLBACKS,200);
}
int _setAccessLock(char *context, char *server, int lockType,int lockDuration)
{
  AccessLockListItem *a;
  DTYPE din;
  char devsrv[128];
  short svals[2];
  int suppressCallbacks = (lockType & LOCK_NOCALLBACKS) ? TRUE : FALSE;
  int lt = lockType & 0xff;

  if (context == NULL || strlen(context) == 0) return invalid_parameter;
  if (server == NULL || strlen(server) == 0) return invalid_parameter;
  if (lt < 0 || lt > LOCK_ABORT) return invalid_parameter;
  if (lockDuration < 0) lockDuration = 0;
  if (lockDuration > MAX_LOCK_DURATION) lockDuration = MAX_LOCK_DURATION;
  if (lt == LOCK_PERSISTENT) lockDuration = MAX_LOCK_DURATION;

  a = getAccessLockItem(context,server);
  if (a == NULL) return code_failure;
  a->lockDuration = lockDuration;
  a->lockType = lockType;
  sprintf(devsrv,"/%.32s/%.32s",context,server);
  svals[0] = lockType;
  svals[1] = lockDuration;
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = 2;
  din.dFormat = CF_INT16;
  din.data.sptr = svals;
  if (suppressCallbacks)
  {
    a->lockLinkStatus = not_ready;
    a->lockLinkStatus = ExecLinkEx(devsrv,"ACCESSLOCK",NULL,&din,CA_WRITE|CA_NOCALLBACKS,200);
  }
  else
  {
    int lid = AttachLink(devsrv,"ACCESSLOCK",NULL,&din,CA_WRITE|CA_RETRY,500,_cbAccessLock,CM_SINGLE);
    if (lid < 0) a->lockLinkStatus = -lid; else  a->lockLinkId = lid;
    while (a->lockLinkStatus == not_ready) clientCycle();
  }
  a->lastSent = (int)time(NULL);
  return a->lockLinkStatus;
}
#endif /* SMALL_TINE_LIB*/

/* TINE Client API Calls follow: */
TINE_EXPORT int AttachLinkEx2(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*cbFcn)(int,int,void *),int mode,UINT32 cbId,void *reference)
{
  return _attachLink(devName,devProperty,dout,din,access,pollingRate,
                     (void *)cbFcn,mode,cbId,reference,UNASSIGNED_INDEX);
}
TINE_EXPORT int AttachLinkEx(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*cbFcn)(int,int),int mode,UINT32 cbId)
{
  return _attachLink(devName,devProperty,dout,din,access,pollingRate,
                     (void *)cbFcn,mode,cbId,NULL,UNASSIGNED_INDEX);
}
TINE_EXPORT int AttachLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*cbFcn)(int,int),int mode)
{
  return _attachLink(devName,devProperty,dout,din,access,pollingRate,
                     (void *)cbFcn,mode,UNASSIGNED_CALLBACKID,NULL,UNASSIGNED_INDEX);
}
TINE_EXPORT int CloseGlobalLink(int linkId)
{
  if (linkId < 0 || linkId >= nglobals) return invalid_link;
  if (glbTbl[linkId]->nlnks-- > 1) return 0;
  detachPortFromMulticastGroup(glbClnSck,(UINT16)gtGCastPort,glbTbl[linkId]->mcastGrp,"globals specific");
  glbTbl[linkId]->disabled = TRUE;
  if (glbTbl[linkId]->freeMemoryOnClose)
  {
    SystemFree(glbTbl[linkId]->data);
    glbTbl[linkId]->freeMemoryOnClose = FALSE;
  }
  glbTbl[linkId]->data = NULL;
  return 0;
}
void _closeLink(int linkId)
{
  ConTblEntry *c;
  McaTblEntry *mca;
  int i, m, lid, isInformationStatic;
# ifndef SMALL_TINE_LIB
  WcTblEntry *wc;
  DpdTblEntry *lnk;
# endif
  if (linkId < 0 || linkId >= nConnectionTableEntries) return; /* not a valid link table entry ! */
  if ((c=conTbl[linkId]) == NULL) return;
  if (!isENSCall(SrvTbl[c->srvIdx].ExportName)) msglog(0,"/%.32s/%.32s/%.64s[%.64s] : closing link [%d] (cbId: %d)",
     SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName,linkId,c->cbId);
  if (gIsRunningAsServer && gAutoLinkErrorAlarms)
  { /* remove any link error alarms */
    clearFecLinkErrorAlarm(linkId);
  }
  isInformationStatic = (c->linkStatus & 0x0fff) == information_static ? TRUE : FALSE;
  if (isInformationStatic && c->isInsideCallback)
  { /* callback has called close() ! */
    c->linkStatus = 0; /* avoid calling the post-callback _closeLink() */
    isInformationStatic = FALSE; /* don't need special tricks below */
  }
  if (c->isGrouped) removeLinkFromGroup(linkId);
  if (c->mcaIndex > 0)
  { /* this is an mca link element */
#   ifndef SMALL_TINE_LIB
    if (WaitForSystemMutex(hDpdLinkMutex,gSystemTick*MUTEX_WAIT_AUGMENT) == 0) 
    { /* check for dependencies ... */
      if (c->dpdLink != NULL && conTbl[c->boundToId] != NULL && conTbl[c->boundToId]->mode != CM_CANCEL)
      { /* this has dependencies itself: reassign them */
        lid = c->dpdLink->linkId;
        if (lid >= 0 && lid < nConnectionTableEntries)
        { /* check link id validity! */
          conTbl[lid]->mcaIndex = c->mcaIndex; /* dpdLink head takes over ! */
          conTbl[lid]->mcaLink = c->mcaLink;
          conTbl[lid]->dpdLink = c->dpdLink->nxt; 
          if (conTbl[lid]->dpdLink != NULL) conTbl[lid]->dpdLink->prv = NULL;
          conTbl[lid]->cbRef = c->cbRef;
          conTbl[lid]->hasReference = c->hasReference;
          if (lid != c->boundToId) conTbl[lid]->boundToId = c->boundToId;
          if (c->boundToId > 0) addDependentLinkToMca((McaTblEntry *)conTbl[c->boundToId]->cbRef,lid);
          for (lnk=conTbl[lid]->dpdLink; lnk != NULL; lnk=lnk->nxt)
          { /* run thru list and reassign boundtoId */
            if (lid != lnk->linkId) conTbl[lnk->linkId]->boundToId = lid;
          }
        }
      }
      ReleaseSystemMutex(hDpdLinkMutex);
    }
#   endif
    c->mcaIndex = 0;
    c->tocounter = 0;
    c->mode = CM_CANCEL;
    if ((i=c->boundToId) > 0 && i < nConnectionTableEntries && conTbl[i]->cbRef != NULL)
    { /* should always be true if mcaIndex > 0 */
      mca = (McaTblEntry *)conTbl[i]->cbRef;
      cancelDependentLinkInMca(mca,linkId);
      c->boundToId = 0;
      if (mca->nlinks <= 0) 
      { /* last one -> remove the mca link */
        rmvDependentLinkFromMca(mca,linkId);
        c->mcaIndex = -1; /* signal that the mca list is no longer valid */
        lid = mca->linkId;
        conTbl[lid]->cancelPending = TRUE;
        conTbl[lid]->timemark = (UINT32)time(NULL);
      }
    }
    c->cbRef = NULL;
    c->cbfcn = NULL;
    c->mcaLink = NULL;
    c->hasReference = FALSE;
    goto out;
  }
# ifndef SMALL_TINE_LIB
  if (c->wdLink != NULL) 
  {
    c->wdLink->nlinks--;
    if (c->wdLink->nlinks == 0) lwdCloseLink(c->wdLink);
    c->wdLink = NULL;
  }
  if ((i=c->boundToId) > 0)
  { /* this one is just bound to another link (don't close for real) */
    removeDependentLink(linkId); 
    if (conTbl[i]->dpdLink == NULL)
    { /* removed last dependency; is the parent still active ? */
      if (!conTbl[i]->cancelPending) goto out;
      /* can now cancel the parent link ! */
      c = conTbl[i]; 
    }
    else
    { /* there are still dependencies, don't do anything */
      goto out;
    }
  }
  if (c->dpdLink != NULL)
  { /* parent link is canceled with dependencies ! */
    c->cancelPending = TRUE;
    c->needsNotification = FALSE;
    if (c->allocatedBytes == 0)
    { /* must re-shuffle the data storage as the caller might free his memory! */
      if ((c->dataPtr=(char *)SystemCalloc(1,(size_t)c->sizeBytesOut)) != NULL)
      {
        c->allocatedBytes = c->sizeBytesOut;
      }
    }
    goto out;
  }
  if ((wc=c->wcLink) != NULL)
  { /* closing a wildcard link */
    if (wc->length > 0)
    { /* not a single-server link */
      for (i=0; i<wc->length; i++)
      { /* re-entrancy shouldn't be a problem here ...*/
        _closeLink(wc->listId[i]);
      }
      c->mode = CM_CANCEL; 
      c->heartbeat = HBT_NONE;
      c->tocounter = 0;
    }
    c->wcLink = NULL;
    /* mark as available and cleanup ... */
    wc->linkId = -1; 
    wc->active = FALSE;
  }
# endif /* !SMALL_TINE_LIB */
  if ((m=BASEMODE(c->mode)) > CM_SINGLE)
  {
    if (c->mode & CM_MCAST && adrcmp(&c->mcastGrp,&inadr_zero)) 
    {
      detachPortFromMulticastGroup(mcastClnSck,(UINT16)gtMCastPort,c->mcastGrp,"server specific");
      c->mcastGrp = inadr_zero;
    }
    if (c->mode & CM_MCAST) c->mode = CM_CANCEL | CM_MCAST;
    else c->mode = CM_CANCEL;
    if (c->tcpSck > 0) c->mode |= (c->inetProtocol&(TCP|TCP6) ? CM_CONNECT : CM_STREAM);
    if (m < CM_REGISTER) c->heartbeat = HBT_SUB_PENDING;
    c->blksin = c->bytesin = 0; 
    c->tocounter = 0;
  }
  if (c->linkStatus == address_unknown)
  { /* then no sense in trying to inform the server that I'm closing the link ! */
    c->heartbeat = HBT_NONE;
    c->mode = CM_CANCEL;
  }
  c->cbRef = NULL;
  c->mcaLink = NULL;
  c->needsNotification = FALSE;
  c->cbfcn = NULL;
  c->hasReference = FALSE;
  c->hasNotifiedOnce = FALSE;
  if (c->glbTblId > 0) 
  { /* mapped globals link ... */
    CloseGlobalLink(c->glbTblId-1);
    c->glbTblId = 0;
  }
out:
  if (isInformationStatic)
  {
    if (c->heartbeat != HBT_SUB_PENDING) 
    { /* if not passing thru prepSubRequest() flag link as 'registered' */
      c->mode = CM_REGISTER;
      c->heartbeat = HBT_NONE;
      c->linkStatus = 0;
    }
  }
  return;
}
TINE_EXPORT int CloseLink(int linkId)
{ /* explicit call to close the link ... */
  ConTblEntry *c;
  int cc, i;
  if (cdiCloseLinkHook != NULL)
  { /* there is a registered cdi hook */
    if ((i=GET_CDI_LINKID(linkId)) >= 0)
    { /* this one is a cdi link */
      cc = cdiCloseLinkHook(i);
      if (i > 0 && i < nCbXRefTblEntries) cbXRefTbl[i]->linkId = -1;
      return cc;
    }
  }
  if (linkId < 0 || linkId >= nConnectionTableEntries) return invalid_link;
  if ((c=conTbl[linkId]) == NULL) return invalid_index;
  c->cbfcn = NULL; /* ensure no further callbacks */
  if (c->mode == CM_CANCEL) return 0; /* already closed */
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) 
  {
    msglog(0,"unable to close link [%d] /%.32s/%.32s/%.64s[%.64s] : cannot get link table mutex",
     linkId,SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName);
    c->mode = CM_CANCEL; /* mark as canceled anyway */
    return mutex_error;
  }
  _closeLink(linkId);
  ReleaseSystemMutex(hLinkTblMutex);
  return 0;
}
TINE_EXPORT int CloseNetGlobal(const char *keyword)
{
  int i = -1, cc;
  char *kwd=NULL,*ctx=NULL,*srv=NULL,kwdbuf[128];

  strncpy(kwdbuf,keyword,64); kwdbuf[64] = 0;
  if ((cc=parseGlobalsKeyword(kwdbuf,&ctx,&srv,&kwd)) != 0) return cc;

  if (getGlobalTableIndex(0,ctx,srv,kwd) == -1) return invalid_keyword;
  while ((i=getGlobalTableIndex(i+1,ctx,srv,kwd)) != -1)
  {
    if ((cc=closeGlobalLink(i)) != 0) break;
  }
  return cc;
}
TINE_EXPORT int IsParentLink(int linkId)
{
  if (linkId < 0 || linkId >= nConnectionTableEntries) return FALSE;
  if (conTbl[linkId]->dpdLink != NULL) return TRUE;
  return FALSE;
}
void _cbLnkLsn(int id, int cc)
{
  UNUSED(cc); /* not interested in the completion code */
  conTbl[id]->idleCount++;
  if (conTbl[id]->dpdLink == NULL && conTbl[id]->idleCount > ALLOWED_IDLE_LISTENING_COUNT)
  { /* no dependencies and ~5 minutes of no interest ... */
    _closeLink(id);
  }
}
TINE_EXPORT int ExecLinkEx(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,UINT16 timeout)
{
  int cc=0,id,pid=0,retries=5;
  int mode,to,isAdjustable=0,keepLink=FALSE,wait=TRUE,allflags=FALSE,showall=FALSE,suspendCallbacks=FALSE;
  char *c,*dn=NULL,*cn=NULL,*sn=NULL,devNameBuffer[1048];

  if (devName == NULL || devProperty == NULL) return argument_list_error;
  if (tineDebug > 2) dbglog("%s[%.64s] synchronous link entered",devName,devProperty);
  strncpy(devNameBuffer,devName,1024);
  if (dout != NULL && dout->dArrayLength > 0 && devNameBuffer[0] == '/')
  { /* trap synchronous globals calls */
    cn = &devNameBuffer[1];
    if ((sn=strchr(cn,'/')) != NULL && !strnicmp(&sn[1],"GLOBALS",7) && (sn[8] == 0 || sn[8] == '/'))
    { /* this looks like one */
      *sn++ = 0; 
      if ((c=strchr(sn,'/')) != NULL) *c = 0;
      if (isGlobalsParameter(cn,sn,(char *)devProperty))
      { /* not all GLOBALS properties are globals parameters */
        BYTE *dptr = dout->data.bptr;
        if (access & CA_WRITE) return illegal_read_write;
        if (tineDebug > 2) dbglog("%s[%.64s] re-assigned to globals link",devName,devProperty);
        dout->data.bptr = NULL; /* this ensures proper data storage but will postpone a data copy */
        sprintf(devNameBuffer,"/%.32s/%.32s/%.64s",cn,sn,(char *)devProperty);
        if ((id=_recvGlobal(devNameBuffer,dout,NULL,0,NULL)) < 0) 
        { /* unexpected error, but don't disturb the data pointer */
          dout->data.bptr = dptr;
          return -id;
        }
        if (dptr != NULL) 
        { /* the caller could have intended the assignment */
          dout->data.bptr = dptr;
          glbcpy(dout,glbTbl[id]);
        }
        glbTbl[id]->freeMemoryOnClose = TRUE;

        return glbTbl[id]->status;
      }
    }
  }
  if (isBitfieldAccess(dout)) acquireAndRegisterBitfield(devName,dout->dTag);
  isAdjustable = access & CA_MUTABLE;
  showall = access & CA_CCSHOWALL;
  suspendCallbacks = access & CA_NOCALLBACKS;
  access &= ~(CA_MUTABLE|CA_CCSHOWALL|CA_NOCALLBACKS);
  if (!(access & CA_NORETRY) && (gAlwaysRetry || timeout >= 1000)) access |= CA_RETRY;
  if (getCdiLinkNameHook != NULL && (dn=getCdiLinkNameHook((char *)devName,devNameBuffer)) != NULL)
  {
    cc = not_implemented;
    if (cdiExecLinkHook != NULL) cc = cdiExecLinkHook(dn,(char *)devProperty,dout,din,access,timeout);
    if (tineDebug > 2) dbglog("%s[%.64s] re-assigned to CDI : %.32s",devName,devProperty,cc2str(cc));
    if (cc != server_redirection) return cc;
    cc = 0;
  }
  dn = (char *)devName; /* not a CDI hook ! */
  lastLnkErrSrc = LNK_ERROR_LCL;

  mode = CM_SINGLE;
  if (access & CA_CONNECT) mode |= CM_CONNECT;
  if (access & CA_STREAM) mode |= CM_STREAM;
  access &= ~(CA_CONNECT|CA_STREAM);
  mode |= CM_SYNCNOTIFY; /* always use synchronous notification on synchronous links */
  if (suspendCallbacks) SetSuspendCallbacks(TRUE);
retryLink:
  allflags = showall;
  if ((id=AttachLink(dn,devProperty,dout,din,access,timeout,_cbExecLink,mode)) < 0 ||
      (pid=conTbl[id]->boundToId) > 0)
  {
    if (id == -link_exists) 
    { /* this link exists in the table -> must have a parent */
      if (dout != NULL) pid = dout->xferReason;
      keepLink = TRUE;
    }
    if (pid > 0) 
    { /* get output from parent */
      if (!conTbl[pid]->hasNotifiedOnce)
      { /* but the parent hasn't been notified yet ! */
        conTbl[pid]->tocounter = -1;
        if (WaitForSyncLink(pid,(int)(2*conTbl[pid]->pollingInterval)) != 0) 
          conTbl[pid]->linkStatus = link_timeout;
      }
      if (dout != NULL)
      {
        if (tineDebug) dbglog("redirecting output for %s[%.64s] to link %d",dn,devProperty,pid);
        if (dout->data.vptr == NULL) dout->data.vptr = (void *)conTbl[pid]->dataPtr;
        if (isAdjustable) dout->dArrayLength = conTbl[pid]->lastDataSize;
        dout->dTimeStamp = PutDataTimeStampU(0,conTbl[pid]->dtimestamp,conTbl[pid]->dtimestampUSEC);
        dout->sysStamp = conTbl[pid]->systemstamp;
        dout->dStamp = conTbl[pid]->datastamp;
      }
#     ifndef SMALL_TINE_LIB
      if (!keepLink)
      { /* just pass the data along and remove this link */
        conTbl[id]->heartbeat = HBT_NONE;
        conTbl[id]->mode = CM_CANCEL; /* make sure it is canceled ! */
        removeDependentLink(id);
      }
#     endif
      ccerr(conTbl[pid]->linkStatus);
    }
#   ifdef MULTITHREADED
    if (id == -semaphore_busy && retries-- > 0) goto retryLink;
#   endif
    ccerr(-id);
  }
  if (dout != NULL) conTbl[id]->lastDataSize = dout->dArrayLength; /* in case CE_RENEGOTIATE is needed */
  to = (int)conTbl[id]->pollingInterval;
  if (access & CA_RETRY) to = to*2 + 100;
  if (mode & CM_STREAM) to *= 10; /* CM_STREAM should not time out ! */
  if (wait && WaitForSyncLink(id,to) != 0)
  { 
    conTbl[id]->lastLnkErrSrc = LNK_ERROR_LCL;
    conTbl[id]->dataStatus = not_posted;
    conTbl[id]->linkStatus = connection_timeout;
    strncpy(conTbl[id]->lastStatusString,erlst[connection_timeout],STATUS_SHORTSIZE);
    strncpy(gLastStatusString,conTbl[id]->lastStatusString,STATUS_SHORTSIZE);
    checkAddressAcquisition(conTbl[id]->fecIdx,connection_timeout);
  }
  if (conTbl[id]->linkStatus == 0) checkAddressAcquisition(conTbl[id]->fecIdx,0);
  conTbl[id]->heartbeat = HBT_NONE;
  if (conTbl[id]->mode != CM_CANCEL)
  {
    if (isTraceLink(id)) traceLinkOutput(conTbl[id], "ExecLink", "link canceled");
  }
  conTbl[id]->mode = CM_CANCEL; /* make sure it is canceled ! */
  cc = lastLnkErr = conTbl[id]->linkStatus;
  if (cc == 0) strcpy(conTbl[id]->lastStatusString,"success");
  if (cc != 0) allflags = TRUE;
  if (allflags && conTbl[id]->rdrId > 0) cc |= CE_REDIRECTED;
  if (dout != NULL)
  {
    dout->dTimeStamp = PutDataTimeStampU(0,conTbl[id]->dtimestamp,conTbl[id]->dtimestampUSEC);
    if (dout->dFormat == CF_DEFAULT)
    {
#     ifndef SMALL_TINE_LIB
      ReLnkTblEntry *r = conTbl[id]->rlkEntry;
      if (conTbl[id]->relinked && r != NULL)
      { /* has been re-linked */
        dout->dFormat = r->lnkFmt;
        dout->dArrayLength = r->lnkSiz;
      }
      else
#     endif
      {
        dout->dFormat = conTbl[id]->formatOut;
        dout->dArrayLength = conTbl[id]->sizeOut;
      }
      strncpy(dout->dTag,conTbl[id]->strTagOut,TAG_NAME_SIZE);
    }
    dout->dStamp = conTbl[id]->datastamp;
    dout->sysStamp = conTbl[id]->systemstamp;
    if (allflags && cc == 0 && dout->dArrayLength != conTbl[id]->lastDataSize) cc |= CE_RENEGOTIATE;
    if (isAdjustable) dout->dArrayLength = conTbl[id]->lastDataSize;
    dout->xferReason = conTbl[id]->xferReason;
  }
  if ((tineDebug || errorDebug) && cc) dbglog("ExecLink: <%d> %.192s",cc,gLastStatusString);
  if (cc == async_access_required || cc == mcast_access_required)
  {
    DTYPE dlsn;
    int lmode = CM_TIMER;
    if (isGlobalsLink(conTbl[id]->ctxName,conTbl[id]->expName,conTbl[id]->prpName)) goto err;
    if (cc == mcast_access_required) lmode |= CM_NETWORK;
    if (tineDebug > 2) dbglog("synchronous link /%.192s[%.64s] : %.32s",devName,devProperty,cc2str(cc));
    if (retries-- > 0 && conTbl[id]->boundToId == 0)
    {
      conTbl[id]->timeLastSync = conTbl[id]->starttime = 0;
      DTYPEZERO(dlsn);
      if (dout != NULL) memcpy(&dlsn,dout,sizeof(DTYPE)); 
      dlsn.data.vptr = NULL;
      _attachLink(devName,devProperty,&dlsn,din,access,1000,(void *)_cbLnkLsn,
        lmode,UNASSIGNED_CALLBACKID,NULL,UNASSIGNED_INDEX);
      goto retryLink;
    }
  }
  if (cc == connection_timeout && isENSCall((char *)devName) && retries-- > 0)
  {
    if (retries > numNameServers - 1) retries = numNameServers - 1;
    ToggleNameServer();
    goto retryLink;
  }
err:
  if (suspendCallbacks) SetSuspendCallbacks(FALSE);
  if (BASE_CC(cc) != 0 && !gNoMsgLog)
    msglog(cc,"ExecLink to %s[%.64s]: %.32s",devName,devProperty,cc2str(cc));
  if (tineDebug > 2) dbglog("%s[%.64s] sync link terminated: %.32s",devName,devProperty,cc2str(cc));
  return cc;
}
TINE_EXPORT int ExecLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char lclDevName[132],eqmn[EQM_NAME_SIZE],*en, *c;
  if (!strnicmp((char *)&devName[1],"LOCAL",5))
  { /* context is LOCAL or LOCALHOST */
    strncpy(lclDevName,devName,132); lclDevName[131] = 0;
    if ((en=strchr(&lclDevName[1],'/')) == NULL) return non_existent_elem;
    en++; /* trim off and find the eqm name */
    strncpy(eqmn,en,EQM_NAME_SIZE); eqmn[EQM_NAME_SHORTSIZE] = 0;
    if ((c=strchr(eqmn,'/')) != NULL) *c = 0;
#   ifndef SMALL_TINE_LIB
    if (getExportListItem(eqmn) != NULL)
    { /* really is a locally managed equipment module (could otherwise be a CDI link) */
      return ExecLocalLink(en,devProperty,dout,din,access);
    }
#   endif
  }
  return ExecLinkEx(devName,devProperty,dout,din,access,DEFAULT_CLIENT_POLLRATE);
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int FreeAccessLock(char *context, char *server)
{
  return rmvAccessLock(context,server,TRUE);
}
TINE_EXPORT int GetAccessLockStatus(char *context, char *server)
{
  AccessLockListItem *a;
  time_t t = time(NULL);
  int cc;
  if (context == NULL || server == NULL) return invalid_parameter;
  for (a=accessLockLst; a != NULL; a = a->nxt)
  {
    if (strnicmp(context,a->ctx,CONTEXT_NAME_SIZE) != 0) continue;
    if (strnicmp(server,a->srv,EXPORT_NAME_SIZE) != 0) continue;
  }
  if (a == NULL) return not_locked;
  cc = a->lockLinkStatus;
  if (cc == 0 && t > a->lockDuration + a->lastSent) cc = lock_expired;
  return cc;
}
TINE_EXPORT int GetClnRecvQueueDepth(void)
{
  return gClnRecvQueueDepth;
}
TINE_EXPORT int GetLinkQueueDepth(int linkId)
{
  if (!isValidConTblEntry(linkId)) return -invalid_link;
  return conTbl[linkId]->queueDepth;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT int GetCompletionDataSize(int i)
{
# ifndef SMALL_TINE_LIB
  ReLnkTblEntry *r;
# endif
  if (i < 0) return (int)LastCompletionDataSize;
  if (!isValidConTblEntry(i)) return -invalid_link;
# ifndef SMALL_TINE_LIB
  if ((r=conTbl[i]->rlkEntry) != NULL)
  { /* has been re-linked */
    return r->lnkSiz;
  }
# endif
  return (int)conTbl[i]->lastDataSize;
}
TINE_EXPORT int GetCompletionDataSizeFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
# ifndef SMALL_TINE_LIB
  ReLnkTblEntry *r;
  if (i > 0 && (r=conTbl[i]->rlkEntry) != NULL)
  { /* has been re-linked */
    return r->lnkSiz;
  }
# endif
  return i < 0 ? -invalid_link : (int)conTbl[i]->lastDataSize;
}
TINE_EXPORT int ResubmitDataLink(int id)
{
  if (!isValidConTblEntry(id)) return invalid_link;
  conTbl[id]->heartbeat = HBT_SUB_PENDING;
  return 0;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int GetCompletionDataType(int i)
{
  ReLnkTblEntry *r;
  if (i < 0) i = LastIncomingLinkId;
  if (i == -1) return (int)LastCompletionDataType;
  if (!isValidConTblEntry(i)) return -invalid_link;
  if ((r=conTbl[i]->rlkEntry) != NULL)
  { /* has been re-linked */
    return r->lnkFmt;
  }
  return (int)conTbl[i]->formatOut;
}
TINE_EXPORT int GetCompletionDataTypeFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return GetCompletionDataType(i);
}
TINE_EXPORT char *GetCompletionStatus(int i)
{
  if (i < 0) return gLastStatusString;
  if (i >= nConnectionTableEntries) return "invalid link id";
  if (conTbl[i] == NULL) return "invalid link id";
  return conTbl[i]->lastStatusString;
}
TINE_EXPORT int FillCompletionStatusString(int i,char *errstr)
{
  if (errstr == NULL) return argument_list_error;
  if (i == -1) strncpy(errstr,gLastStatusString,STATUS_SIZE);
  if (!isValidConTblEntry(i)) return invalid_link;
  strncpy(errstr,conTbl[i]->lastStatusString,STATUS_SIZE);
  return 0;
}
TINE_EXPORT short GetCompletionSource(int i)
{
  if (i < 0) return lastLnkErrSrc;
  if (i >= nConnectionTableEntries) return -1;
  if (conTbl[i] == NULL) return -un_allocated;
  return conTbl[i]->lastLnkErrSrc;
}
TINE_EXPORT char *GetCompletionStatusFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? NULL : conTbl[i]->lastStatusString;
}
TINE_EXPORT int FillCompletionStatusStringFromCallbackId(int id,char *errstr)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? invalid_link : FillCompletionStatusString(i,errstr);
}
TINE_EXPORT short GetCompletionSourceFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -invalid_link : conTbl[i]->lastLnkErrSrc;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT double GetCurrentDataTimeStamp(int i)
{
  if (!isValidConTblEntry(i)) return 0;
  return PutDataTimeStampU(0,conTbl[i]->dtimestamp,conTbl[i]->dtimestampUSEC);
}
TINE_EXPORT double GetCurrentDataTimeStampFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? 0 : PutDataTimeStampU(0,conTbl[i]->dtimestamp,conTbl[i]->dtimestampUSEC);
}
TINE_EXPORT int GetCurrentLinkStatus(int i)
{
  if (!isValidConTblEntry(i)) return invalid_link;
  return conTbl[i]->linkStatus;
}
TINE_EXPORT int GetCurrentLinkStatusFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -invalid_link : (int)conTbl[i]->linkStatus;
}
TINE_EXPORT int GetCurrentDataStatus(int i)
{
  if (!isValidConTblEntry(i)) return invalid_link;
  return conTbl[i]->dataStatus;
}
TINE_EXPORT int GetCurrentDataStatusFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -invalid_link : (int)conTbl[i]->dataStatus;
}
TINE_EXPORT int UpdateDataStampsFromLinkId(int linkId,DTYPE *dout)
{
  ConTblEntry *c;
  int cc = 0;
  if (dout == NULL) return argument_list_error;
  if (linkId < 0 || linkId >= nConnectionTableEntries) ccerr(invalid_index); /* not a valid link table entry ! */
  c = conTbl[linkId];
  dout->dTimeStamp = (double)c->dtimestamp + ((double)c->dtimestampUSEC)/1000000.0;
  dout->dStamp = c->datastamp;
  dout->sysStamp = c->systemstamp;
  dout->xferReason = c->xferReason;
err:
  return cc;

}
TINE_EXPORT int UpdateDataStampsFromCallbackId(int id,DTYPE *dout)
{
  int i = getLinkIdFromCallbackId(id);
  return UpdateDataStampsFromLinkId(i,dout);
}
TINE_EXPORT int GetDataFromLinkId(int linkId,DTYPE *dout,UINT16 *status)
{
  ConTblEntry *c;
  int cc = 0;
  if (dout == NULL) return argument_list_error;
  if (linkId < 0 || linkId >= nConnectionTableEntries) ccerr(invalid_index); /* not a valid link table entry ! */
  c = conTbl[linkId];
  makeOutputDTYPEFromConTblEntry(c, dout);
  if (status != NULL) *status = (UINT16)c->linkStatus;
err:
  return cc;
}
TINE_EXPORT int GetDataFromCallbackId(int id,DTYPE *dout,UINT16 *status)
{
  int i = getLinkIdFromCallbackId(id);
  return GetDataFromLinkId(i,dout,status);
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT void *GetLinkCallbackReference(int id)
{
  if (!isValidConTblEntry(id)) return NULL;
  return conTbl[id]->cbRef;
}
TINE_EXPORT double GetGlobalsTimeStamp(int linkId)
{
  if (linkId < 0 || linkId >= nglobals) return 0;
  return PutDataTimeStampU(0,glbTbl[linkId]->time,glbTbl[linkId]->timeUSEC);
}
TINE_EXPORT double GetGlobalsTimeStampFromCallbackId(int id)
{
  int i = getGlobalLinkIdFromCallbackId(id);
  return GetGlobalsTimeStamp(i);
}
TINE_EXPORT int GetLastGlobalDataSize(int id)
{
  if (id == -1) return gLastGlobalDataSize;
  if (id < 0 || id >= nglobals) return -invalid_link;
  return glbTbl[id]->lastsize;
}
TINE_EXPORT int GetLastGlobalDataSizeFromCallbackId(int id)
{
  int i = getGlobalLinkIdFromCallbackId(id);
  if (i == -1) return gLastGlobalDataSize;
  return glbTbl[i]->lastsize;
}
TINE_EXPORT int GetGlobalDataType(int id)
{
  if (id < 0 || id >= nglobals) return -invalid_link;
  return (short)glbTbl[id]->format;
}
TINE_EXPORT int GetGlobalDataTypeFromCallbackId(int id)
{
  int i = getGlobalLinkIdFromCallbackId(id);
  if (i == -1) return -invalid_index;
  return (short)glbTbl[i]->format;
}
TINE_EXPORT char *GetLinkName(int i)
{
  static char linkname[256];
  strcpy(linkname,"link not found");
  if (i >= 0 && i < nConnectionTableEntries)
  {
    sprintf(linkname,"/%.32s/%.32s/%.64s[%.64s]",conTbl[i]->ctxName,conTbl[i]->expName,
      conTbl[i]->devName,conTbl[i]->prpName);
  }
  return linkname;
}
TINE_EXPORT char *GetLinkNameFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  static char linkname[256];
  strcpy(linkname,"link not found");
  if (i >= 0) 
  {
    sprintf(linkname,"/%.32s/%.32s/%.64s[%.64s]",conTbl[i]->ctxName,conTbl[i]->expName,
      conTbl[i]->devName,conTbl[i]->prpName);
  }
  return linkname;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT char *GetLastLinkError(short cc, char *errstr)
{
  if (errstr == NULL) return "[null error string]";
  strncpy(errstr,GetRPCLastError(cc),64);
  return errstr;
}
TINE_EXPORT double GetSynchronizationTolerance(void) { return gSyncTolerance; }
TINE_EXPORT int GetSystemDataStamp(int linkId)
{
  if (linkId < 0) return gSystemDataStamp;
  if (linkId > nConnectionTableEntries-1) return -invalid_link;
  return conTbl[linkId]->systemstamp;
}
TINE_EXPORT int GetSystemDataStampFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -invalid_link : conTbl[i]->systemstamp;
}
TINE_EXPORT int GetGlobalsSystemStamp(int glbId)
{
  if (glbId < 0) return gSystemDataStamp;
  if (glbId > nglobals-1) return -invalid_link;
  return glbTbl[glbId]->sysStamp;
}
TINE_EXPORT int GetDataStamp(int linkId)
{
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return -invalid_link;
  return conTbl[linkId]->datastamp;
}
TINE_EXPORT int GetDataStampFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -invalid_link : conTbl[i]->datastamp;
}
TINE_EXPORT int GetGlobalsDataStamp(int glbId)
{
  if (glbId < 0) return 0;
  if (glbId > nglobals-1) return -invalid_link;
  return glbTbl[glbId]->dataStamp;
}
TINE_EXPORT UINT16 GetTransferFlag(int linkId)
{
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return CX_NULL;
  return conTbl[linkId]->xferReason;
}
TINE_EXPORT UINT16 GetTransferFlagFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? CX_NULL : conTbl[i]->xferReason;
}
TINE_EXPORT void SetSuppressHeartbeatNotification(int value)
{
  SystemSuppressHeartbeatNotification = value;
}
GlbPrmLst *glbPrmLst = NULL;
GlbPrmLst *getGlbPrmByIP(INADR inadr)
{
  GlbPrmLst *g;
  if (!adrcmp(&inadr,&inadr_zero)) return NULL;
  for (g=glbPrmLst; g != NULL; g=g->nxt)
  {
    if (!adrcmp(&g->srvIp,&inadr)) return g;
  }
  return NULL;
}
int isGlobalsParameter(char *ctx,char *srv,char *kwd)
{
  int i, cc = 0, rc=FALSE;
  char dev[128];
  DTYPE dout;
  GlbPrmLst *g;
  FecAddrStruct *fec;
  
  if (stricmp(srv,GLOBALSRV)) 
  { /* if not explicitly the globals server, then pass it thru */
    /* n.b.: internally this routine is always used where srv == GLOBALSRV */
    return TRUE;
  }
  if (!stricmp(kwd,"SYSTIME"))
  { /* this is always redirected from a globals server */
    return TRUE;
  }
  if (!stricmp(kwd,"GLOBALS") || !stricmp(kwd,"SRVGLOBALS"))
  { /* these are never globals parameters ! */
    return FALSE;
  }
  if (WaitForSystemMutex(hLinkLstMutex, -1) != 0) return FALSE;
  for (g=glbPrmLst; g != NULL; g=g->nxt)
  {
    if (!strnicmp(ctx,g->ctx,CONTEXT_NAME_SIZE)) break;
  }
  if (g == NULL)
  {
    int tries = 3;
    if ((g=(GlbPrmLst *)calloc(1,sizeof(GlbPrmLst))) == NULL) ccerr(out_of_local_memory);
    strncpy(g->ctx,ctx,CONTEXT_NAME_SIZE);
    DTYPEZERO(dout);
    dout.dArrayLength = GLBKWD_LIST_CAPACITY;
    dout.dFormat = CF_NAME64;
    dout.data.vptr = g->lst;
    sprintf(dev,"/%.32s/%.32s",ctx,GLOBALSRV);
    cc = connection_timeout;
    while (tries-- > 0 && cc == connection_timeout)
    {
      cc = ExecLinkEx(dev, "GLOBALS", &dout, NULL, CA_READ|CA_NOCALLBACKS|CA_MUTABLE, 200);
    }
    if (cc != 0) ccerr(cc);
    if ((fec=GetFecAddrFromCache(ctx,srv)) != NULL) memcpy(&g->srvIp,&fec->netAdr,gInAdrLen);
    g->num = dout.dArrayLength;
    g->nxt = glbPrmLst;
    glbPrmLst = g;
  }
  for (i=0; i<g->num; i++)
  {
    if (!strnicmp(g->lst[i].name, kwd, PROPERTY_NAME_SIZE)) break; 
  }
  if (i < g->num) rc = TRUE;
err:
  ReleaseSystemMutex(hLinkLstMutex);

  if (rc) msglog(cc,"isGlobalsParameter /%.32s/%.32s[%.64s] : TRUE", ctx, srv, kwd);
  if (cc && (tineDebug || errorDebug)) dbglog("isGlobalsParameter /%.32s/%.32s[%.64s] : %.32s", ctx, srv, kwd, cc2str(cc));
  return rc;
}
int glbcpy(DTYPE *d,GlobalListStruct *g)
{
  int cc = 0;
  UINT32 dsize = g->size*fmtsizeof(g->format);
  if (d->data.vptr == NULL) return 0;
  if (d->dFormat == CF_DEFAULT)
  { /* fix the format and datasize */
    if (d->dArrayLength < dsize) return buffer_too_small;
    d->dFormat = g->format;
    d->dArrayLength = g->size;
  }
  if (d->dArrayLength > (UINT32)g->size) d->dArrayLength = (UINT32)g->size;
  if (d->dFormat == g->format)
  {
    dsize = d->dArrayLength*fmtsizeof(d->dFormat);
    memcpy(d->data.vptr,g->data,dsize);
  }
  else
  {
    cc = reformat(d->data.bptr,d->dFormat,g->data,g->format,g->size);
  }
  return cc;
}
void flushGlobalsTable(void)
{
  if (glbTbl!= NULL)
  {
    int i;
    for (i=0; i<nglobals; i++)
    {
      if (glbTbl[i] == NULL) continue;
      SystemFree(glbTbl[i]->refdata);
      SystemFree(glbTbl[i]);
    }
    SystemFree(glbTbl);
  }
  glbTbl = NULL;
  nglobals = 0;
}
int _recvGlobal(const char *keyword,DTYPE *dout,void *cbFcn,UINT32 cbId,void *reference)
{
  static int initialized = 0;
  GrpTblEntry *gt = NULL;
  int cc = 0,i = -1,lognew=FALSE;
  int isGrouped=FALSE,isSyncGroup=FALSE,hasReference=FALSE,isSyncGet=FALSE,isActive=FALSE,isNetGlb=TRUE;
  UINT16 dsize;
  char *kwd=NULL,*ctx=NULL,*srv=NULL,kwdbuf[136];
  INADR inadr=inadr_zero;
  struct timeval tv_stamp;
  FecAddrStruct *fec;
  GlbPrmLst *g;

  if (keyword == NULL || dout == NULL) return -argument_list_error;
  if (WaitForSystemMutex(hSystemGlobalsUpdateMutex,gSystemTick) != 0) return -semaphore_busy;
  if (reference != NULL) hasReference = TRUE;
  if (!gListenForGlobals)
  {
    if ((cc=ListenForNetGlobals()) != 0) ccerr(cc);
  }
  if (!initialized)
  {
    initialized = TRUE;
    if ((glbTbl=(GlobalListStruct **)SystemCalloc(MaxNumGlobals,sizeof(GlobalListStruct *))) == NULL)
      ccerr(out_of_client_memory);
    GlbTblCapacity = MaxNumGlobals;
  }

  strncpy(kwdbuf,keyword,132); kwdbuf[132] = 0;
  if ((cc=parseGlobalsKeyword(kwdbuf,&ctx,&srv,&kwd)) != 0) ccerr(cc);
  if (!isGlobalsParameter(ctx,srv,kwd)) ccerr(invalid_property);
  isNetGlb = isGlobalsLink(ctx,srv,kwd);
  if (cbFcn != NULL)
  { /* check if cbId is exactly equal in the following : */
    isGrouped = (cbId == CM_GROUPED) ? TRUE : FALSE;
    if (cbId == (CM_GROUPED|CM_SYNCGROUP) || cbId == CM_SYNCGROUP)
    {
      isSyncGroup = isGrouped = TRUE;
    }
  }
  if (cbFcn == NULL && cbId != UNASSIGNED_CALLBACKID) isSyncGet = TRUE;
  gettimeofday(&tv_stamp,(struct timezone *)NULL);
  while ((i=getGlobalTableIndex(i+1,ctx,srv,kwd)) != -1) 
  { /* keyword is in the list  */
    isActive = !glbTbl[i]->disabled;
    if (!isSyncGet)
    { /* not a simple synchronous call to get a globals value from the soup */
      if (glbTbl[i]->data != dout->data.vptr) continue;
      if (glbTbl[i]->cbfcn != (TCBFCNP)cbFcn) continue;
      if (glbTbl[i]->hasReference != hasReference) continue;
    }
    if (isActive) 
    { /* is active ! */
      if ((cc=glbcpy(dout,glbTbl[i])) != 0) ccerr(cc);
      dout->dTimeStamp = (double)glbTbl[i]->time;
      if (glbTbl[i]->cbfcn != (TCBFCNP)cbFcn)
      { /* must be a different link ? */
        glbTbl[i]->tv = tv_stamp;
        glbTbl[i]->nlnks++;
      }
      ccerr(glbTbl[i]->status);
    }
    break;
  }
  if (i < 0)
  { /* entry (with data reference) is not in the list */
    i = nglobals;
  }
  if (nglobals >= GlbTblCapacity) 
  {
    msglog(resources_exhausted,"addition globals parameter would exceed globals table capacity (%d)",GlbTblCapacity);
    ccerr(resources_exhausted);
  }
  /* request for new global ... */
  if (glbTbl[i] == NULL && (glbTbl[i]=(GlobalListStruct *)SystemCalloc(1,sizeof(GlobalListStruct))) == NULL)
    ccerr(out_of_client_memory);
  glbTbl[i]->nlnks = 1; /* only one listener at the moment */
  if (cbId == UNASSIGNED_CALLBACKID) cbId = i; else glbTbl[i]->hasUserCallbackId = TRUE;
  strncpy(glbTbl[i]->context,ctx,CONTEXT_NAME_SIZE);
  strncpy(glbTbl[i]->server,srv,EXPORT_NAME_SIZE);
  strncpy(glbTbl[i]->keyword,kwd,KEYWORD_NAME_SIZE);
  dsize = MIN(dout->dArrayLength,UDP_BUFFER_SIZE);
  glbTbl[i]->size = (short)dsize;
  glbTbl[i]->format = dout->dFormat;
  glbTbl[i]->cbfcn = (TCBFCNP)cbFcn;
  glbTbl[i]->cbId = cbId;
  glbTbl[i]->maxPeriod = isNetGlb ? GlobalHeartbeat : -1; 
  glbTbl[i]->hasReference = hasReference;
  glbTbl[i]->cbRef = reference;
  glbTbl[i]->isInitialLink = isActive ? FALSE : TRUE;
  dsize = (UINT16)(dsize*fmtsizeof(dout->dFormat));
  if (dout->data.vptr == NULL && (dout->data.vptr=(BYTE *)SystemCalloc(1,dsize)) == NULL)
    ccerr(out_of_client_memory);
  glbTbl[i]->data = (BYTE *)dout->data.vptr;
  glbTbl[i]->tv = tv_stamp;

  fec = GetFecAddrFromCache(ctx,srv);
  if (fec != NULL)
  { /* It's a known address */
    adrcpy(&glbTbl[i]->srvIp,&fec->netAdr);
    if (fec->tineProtocol < 6)
    { /* legacy: one general globals server */
      attachPortToMulticastGroup(glbClnSck,(UINT16)gtGCastPort,tGCastAddrLegacy,"common globals");
      glbTbl[i]->mcastGrp = tGCastAddrLegacy;
    }
    else
    { /* context specific globals */
      INADR inaddr = fixMcastHaddr((BYTE *)&fec->netAdr,!(fec->addrFlags & ADDRFLAG_MCAST_MODERN));
      attachPortToMulticastGroup(glbClnSck,(UINT16)gtGCastPort,inaddr,"context globals");
      glbTbl[i]->mcastGrp = inaddr;
      feclog("join multicast group for global %.32s keyword %.64s : %.40s",
        glbTbl[i]->context,glbTbl[i]->keyword,fec->strAdr);
    }
    memcpy(&inadr,&fec->netAdr,SCKADR_LEN);
  }
# ifndef NOT_MULTICASTING
  else
  {
    cc = non_existent;
  }
# endif
  glbTbl[i]->isPending = TRUE;
  if (isGrouped && (gt=getGrpTblEntry((size_t)cbFcn)) == NULL) isSyncGroup = isGrouped = 0;
  glbTbl[i]->isGrouped = isGrouped;
  glbTbl[i]->grp = gt;
  if (isSyncGroup && gt != NULL) gt->grpSyncLevel = GRP_SYNC_STARTSYNC;
  feclog("Listening for global parameter : %.64s",glbTbl[i]->keyword);
  glbTbl[i]->disabled = FALSE;
  if (i == nglobals)
  {
    nglobals++;
    lognew = TRUE;
  }
err:
  ReleaseSystemMutex(hSystemGlobalsUpdateMutex);
  if (cc == 0)
  { /* notification will also need the update mutex */
    if (!glbTbl[i]->hasNotifiedOnce && (g=getGlbPrmByIP(inadr)) != NULL)
    { /* new globals call, but maybe its data has been cached ! */
      if (g->hasUpdated && getGlobalFromLastUpdate(kwd,inadr,&tv_stamp) == 0)
        NotifyGlobalClient(i,0);
    }
    if (glbTbl[i]->maxPeriod > 0)
    { /* a known GLOBALS server : can timeout ! */
      while (!ServerExitCondition && !glbTbl[i]->hasNotifiedOnce) clientCycle();
      dout->dTimeStamp = PutDataTimeStampU(0,glbTbl[i]->time,glbTbl[i]->timeUSEC);
      dout->dStamp = glbTbl[i]->dataStamp;
      dout->sysStamp = glbTbl[i]->sysStamp;
    }
  }
  else
  {
    if (i >= 0) NotifyGlobalClient(i,cc);
  }
  if (lognew) msglog(cc, "recvGlobal [%d] %.192s (cbId: %d)",i,keyword,cbId);
  return cc ? -cc : i;
}
TINE_EXPORT int recvNetGlobalEx2(const char *keyword,DTYPE *dout,void (*cbFcn)(int,int,void *),UINT32 cbId,void *reference)
{
  return _recvGlobal(keyword,dout,(void *)cbFcn,cbId,reference);
}
TINE_EXPORT int recvNetGlobalEx(const char *keyword,DTYPE *dout,void (*cbFcn)(int,int),UINT32 cbId)
{
  return _recvGlobal(keyword,dout,(void *)cbFcn,cbId,NULL);
}
TINE_EXPORT int recvNetGlobal(const char *keyword,DTYPE *dout,void (*cbFcn)(int,int))
{
  return _recvGlobal(keyword,dout,(void *)cbFcn,UNASSIGNED_CALLBACKID,NULL);
}
#if !defined(SMALL_TINE_LIB)
TINE_EXPORT void SetAutoLinkWatchdogs(int value)
{
  gAutoLinkWatchdogs = value;
}
TINE_EXPORT void SetAlwaysRetry(int value)
{
  gAlwaysRetry = value;
}
TINE_EXPORT int GetAlwaysRetry(void)
{
  return gAlwaysRetry;
}
TINE_EXPORT void UseMultiThreadedCycler(int value)
{
  useMultiThreadedCycler = value;
}
TINE_EXPORT void UseSynchronousNotification(int value)
{
  gUseSynchronousNotification = value;
}
TINE_EXPORT void SetGlobalsHeartbeat(int value)
{
  if (value > 0) GlobalHeartbeat = value;
}
TINE_EXPORT int GetGlobalsHeartbeat(void)
{
  return GlobalHeartbeat;
}
TINE_EXPORT int SetAccessLock(char *context, char *server, int lockType,int lockDuration)
{
  int cc = _tryAccessLock(context,server,lockType);
  if (cc != 0) return cc;
  return _setAccessLock(context,server,lockType,lockDuration);
}
TINE_EXPORT int GetAccessLockInformation(char *context,char *server,NAME32 *callerName,NAME32 *callerIp,NAME32 *timeLeft)
{
  DTYPE dout;
  char devsrv[128];
  NAME32 n32[3];
  int rc;
  
  if (context == NULL || strlen(context) == 0) return invalid_parameter;
  if (server == NULL || strlen(server) == 0) return invalid_parameter;
  if (callerName == NULL) return invalid_parameter;
  sprintf(devsrv,"/%.32s/%.32s",context,server);
  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = 3;
  dout.dFormat = CF_NAME32;
  dout.data.vptr = n32;

  rc = ExecLinkEx(devsrv,"ACCESSLOCK",&dout,NULL,CA_READ|CA_NOCALLBACKS,500);
  if (rc != 0) return rc;

  if (callerName != NULL) strncpy(callerName->name,n32[0].name,32);
  if (callerIp != NULL) strncpy(callerIp->name,n32[1].name,32);
  if (timeLeft != NULL) strncpy(timeLeft->name,n32[2].name,32);

  return 0;
}
TINE_EXPORT void SetClnRecvQueueDepth(int depth)
{
  gClnRecvQueueDepth = depth > 0 ? depth : 0;
}
TINE_EXPORT void SetLinkQueueDepth(int linkId,int depth)
{
  if (!isValidConTblEntry(linkId)) return;
  if (conTbl[linkId]->queueDepth == 0) return;
  conTbl[linkId]->queueDepth = depth > 0 ? depth : 0;
}
TINE_EXPORT void SetLinkCallbackDelay(int linkId,int cyclesToDelay)
{
  if (!isValidConTblEntry(linkId)) return;
  conTbl[linkId]->cbDelay = cyclesToDelay > 0 ? cyclesToDelay : 0;
}
TINE_EXPORT int GetLinkCallbackDelay(int linkId)
{
  if (!isValidConTblEntry(linkId)) return 0;
  return conTbl[linkId]->cbDelay;
}
TINE_EXPORT int ModifyLinkAttributes(int i,short access,int pollingRate,void (*cbFcn)(int,int),int mode,UINT32 cbId)
{
  int isGrouped,isSyncGroup=FALSE,wasGrouped,baseMode,cc=0;
  GrpTblEntry *gt = NULL;

  if (!isValidConTblEntry(i)) return invalid_index;
  if (conTbl[i]->mode == CM_CANCEL) return un_allocated;
  if ((wasGrouped=conTbl[i]->isGrouped) != 0)
  {
    if ((gt=getGrpTblEntry((size_t)conTbl[i]->cbfcn)) == NULL) wasGrouped = 0;
  }
  if (mode == 0)
  {
    mode = conTbl[i]->mode; /* keep it */
    isGrouped = conTbl[i]->isGrouped; /* keep it */
    if (gt != NULL && gt->grpSyncLevel != GRP_SYNC_NOSYNC) isSyncGroup = TRUE;
  }
  else
  {
    isGrouped = mode & CM_GROUPED; 
    isSyncGroup = mode & CM_SYNCGROUP;
    mode &= ~(CM_GROUPED|CM_SYNCGROUP);
    if ((mode & CM_MCAST) && !gListenForMCasts)
    {
      if ((cc=ListenForMulticasts()) != 0) return cc;
    }
    baseMode = BASEMODE(mode);
    if (baseMode > CM_EVENT  || baseMode < CM_CANCEL || baseMode == CM_RECEIVE) return illegal_mode;
    if (baseMode == CM_DATACHANGE || baseMode == CM_EVENT) isGrouped = 0;
  }
  if (pollingRate == 0) pollingRate = (int) conTbl[i]->pollingInterval; /* keep it */
  if (access == 0) access = conTbl[i]->access;                   /* keep it */
  if (cbId == UNASSIGNED_CALLBACKID) cbId = conTbl[i]->cbId;       /* keep it */
  if (cbFcn == NULL) cbFcn = (TCBFCNP)conTbl[i]->cbfcn;     /* keep it */

  conTbl[i]->pollingInterval = pollingRate;
  conTbl[i]->mode = mode;
  conTbl[i]->cbfcn = (TCBFCNP)cbFcn;
  conTbl[i]->cbId = cbId;
  conTbl[i]->heartbeat = mode == CM_REGISTER ? HBT_NONE : HBT_SUB_PENDING;
  conTbl[i]->access = access;
  conTbl[i]->notifyPending   = 1;

  if (!wasGrouped && isGrouped)
  {
    if ((gt=getGrpTblEntry((size_t)cbFcn)) != NULL)
    {
      gt->numPending++;
      gt->numInGrp++;
      conTbl[i]->isGrouped = 1;
      conTbl[i]->grp = gt;
      if (isSyncGroup) gt->grpSyncLevel = GRP_SYNC_STARTSYNC;
    }
  }
  else if (wasGrouped && !isGrouped)
  {
    if (gt != NULL)
    {
      gt->numPending--;
      gt->numInGrp--;
    }
    conTbl[i]->isGrouped = 0;
    conTbl[i]->grp = NULL;
  }
  return 0;
}
TINE_EXPORT int SetNotificationToleranceEx(int linkId, float toleranceAbsolute, float tolerancePercent, int offset, int length, int flags)
{
  if (toleranceAbsolute < 0 && tolerancePercent < 0) return 0; /* no reason to do anything */
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return invalid_link;
  conTbl[linkId]->mode |= CM_NODUPLICATES;
  if (toleranceAbsolute <= 0 && tolerancePercent <= 0) return 0; /* no reason to allocate a dtf */
  if (conTbl[linkId]->dtf == NULL)
  {
    conTbl[linkId]->dtf = (DataToleranceFactor *)SystemCalloc(1,sizeof(DataToleranceFactor));
  }
  if (conTbl[linkId]->dtf == NULL) return out_of_local_memory;
  conTbl[linkId]->dtf->absolute = toleranceAbsolute;
  conTbl[linkId]->dtf->percent = tolerancePercent;  
  if (length == 0) length = conTbl[linkId]->sizeOut;
  if (offset > 0 && offset < (int)conTbl[linkId]->sizeOut) conTbl[linkId]->dtf->offset = offset;
  if (length > (int)(conTbl[linkId]->sizeOut - conTbl[linkId]->dtf->offset)) length = (int)conTbl[linkId]->sizeOut - conTbl[linkId]->dtf->offset;
  conTbl[linkId]->dtf->length = length;
  if (flags < DTF_NONE) flags = DTF_SUPPRESS;
  conTbl[linkId]->dtf->flags = flags;
  if (conTbl[linkId]->bufCmp == NULL)
  { /* not yet allocated */
    if (conTbl[linkId]->sizeBytesOut > 0)
    {
      conTbl[linkId]->bufCmp = (char *)SystemCalloc(1,(size_t)conTbl[linkId]->sizeBytesOut);
    }
  }
  return 0;
}
TINE_EXPORT int SetNotificationTolerance(int linkId, float toleranceAbsolute, float tolerancePercent)
{
  return SetNotificationToleranceEx(linkId,toleranceAbsolute,tolerancePercent,0,0,DTF_SUPPRESS);
}
TINE_EXPORT void SetSynchronizationTolerance(double toleranceInSeconds)
{
  if (toleranceInSeconds < 0.1) return; /* this would be ridiculous */
  gSyncTolerance = toleranceInSeconds;
}
TINE_EXPORT void SetAutoLinkErrorAlarms(int value)
{
  gAutoLinkErrorAlarms = value;
}
TINE_EXPORT int GetAutoLinkErrorAlarms(void)
{
  return gAutoLinkErrorAlarms;
}
TINE_EXPORT void SetDoocsUserName(const char *usr)
{
  if (usr == NULL) 
  {
    gDoocsUserName[0] = 0;
    return;
  }
  strncpy(gDoocsUserName,usr,USERNAME_SIZE);
  gDoocsUserName[USERNAME_SIZE] = 0;
  msglog(0,"DOOCS username set to %.16s",gDoocsUserName);
  return;
}
TINE_EXPORT char *GetDoocsUserName(void)
{
  return gDoocsUserName;
}
#endif /* SMALL_TINE_LIB */
