/*
 * File hstlib.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file hstlib.c
 * \brief TINE Local History Server module plus API routines
 *
 * A TINE server can be instructed to maintain a local history of certain properties.  The
 * stored data can then be obtained with normal data acquition calls to the TINE Local History
 * Server.  By and large the instructions as to which properties are to be maintained in the
 * local history server and as to what filtering criteria are to be used in commiting the data
 * to disk are typically entered into a startup configuration file 'history.csv'.  As an 
 * alternative, see AppendHistoryInformation().  The local history server maintains two sets
 * of histories, a short-term history kept in main memory and a long-term history kept on disk.
 * Whereas the short-term history can be kept on a millisecond time raster, the long-term history
 * will not be written to disk any more often than once per second.
 *
 */
#ifndef NO_DOXYGEN
#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#ifdef WIN32
# include "direct.h"
# include <float.h>
#endif
#if defined(UNIX) || defined(MACOS)
# include <errno.h>
# include <sys/stat.h>
# include <dirent.h>
# ifndef FREEBSD
# include <alloca.h>
# endif
#endif
#ifdef VMS
# include <file.h>
# include <unixio.h>
#endif
#include "prolog.h"
#include "srvcore.h"
#include "srvip.h"
#include "dbtools.h"
#include "toolkit.h"
#include "hstlib.h"
#endif

#if defined(DOS)
# define O_createmode O_RDWR|O_CREAT|O_APPEND|O_BINARY
# define O_writemode O_RDWR|O_BINARY
# define O_readmode O_RDONLY|O_BINARY
# define O_mode O_RDONLY|O_BINARY
# define FO_mode "rt"
#elif defined(WIN32)
# define O_createmode O_RDWR|O_CREAT|O_APPEND|O_BINARY,_S_IREAD|_S_IWRITE
# define O_writemode O_RDWR|O_BINARY,_S_IREAD|_S_IWRITE
# define O_readmode O_RDONLY|O_BINARY
# define O_mode O_RDONLY|O_BINARY
# define FO_mode "rt"
#elif defined(VMS)
# define O_createmode O_RDWR|O_CREAT|O_APPEND
# define O_writemode O_RDWR
# define O_readmode O_RDONLY
# define O_mode O_RDONLY
# define FO_mode "r"
#else
# define O_createmode O_RDWR|O_CREAT|O_APPEND,S_USERWRITE
# define O_writemode O_RDWR,S_USERWRITE
# define O_readmode O_RDONLY,S_USERWRITE
# define O_mode O_RDONLY
# define FO_mode "r"
#endif

#define SEC_IN_HOUR 3600
#define SEC_IN_DAY (3600L * 24L)
#define AS_STAT_SIZE 16
#define IMPORTANT_RECS_DAY 1000

#if defined(UNIX) && !defined(HAS_CHSIZE)
#define chsize(f,s) ftruncate(f,s)
#endif
#ifdef NEEDS_CHSIZE
int chsize(int hFile,SINT32 size)
{
  BYTE *tmpHstDatBuf;
  int bytesWritten, i;
  int chunkSize = (size - AS_STAT_SIZE)/tmpWorkAreaSize;
  int chunksPerFile = 1+(size - AS_STAT_SIZE)/chunkSize;
  UINT32 statbuf[4];
# if defined(INTEL)
  tmpHstDatBuf = (BYTE *)alloca(chunkSize);
# else
  tmpHstDatBuf = tmpWorkArea;
# endif
  /* create the 16 bytes of file status */
  memset(statbuf,0,AS_STAT_SIZE);
  /* create or append to the standard file */
  memset(tmpHstDatBuf,0,chunkSize);
  if (write(hFile,(char *)statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) return no_file_space;
  for (i=0; i<chunksPerFile; i++)
  { /* fill in the worst case standard file (minimal fragmentation) */
    if ((bytesWritten=write(hFile,(char *)tmpHstDatBuf,chunkSize)) != chunkSize)
    {
      return no_file_space;
    }
  }
  return 0;
}
#endif

static int gAutoRemoveUnsynchronizedMcaHistories = FALSE;
#define HST_LATE_ADDITION_WINDOW 900 /* 15 minutes ! */
int gLclHstLateAdditionWindow = HST_LATE_ADDITION_WINDOW;
TINE_EXPORT void SetHistoryLateAddtionWindow(int value)
{
  gLclHstLateAdditionWindow = value;
  feclog("HIST: set local history late addition window to %d via API\n",value);
}
TINE_EXPORT int GetHistoryLateAdditionWindow(void)
{
  return gLclHstLateAdditionWindow;
}

int hstdbg = 0;
int hasHstFileAlarm(void)
{
  int i;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (hstTbl[i].lastArchiveSts) return TRUE;
  }
  return FALSE;
}
#ifndef VXWORKS
static int hstlog = FALSE;
#endif
static int excludePointsOfInterest = FALSE;
TINE_EXPORT void SetExcludePointsOfInterest(int value)
{
  excludePointsOfInterest = value;
  if (value) feclog("HIST: Points of interest file will not be kept according to API Call");
}
TINE_EXPORT int GetExcludePointsOfInterest(void)
{
  return excludePointsOfInterest;
}
int history_home_done = FALSE;
int gHstDeactivateOnError = FALSE;
TINE_EXPORT void SetHistoryDeactivateOnError(int value)
{
  gHstDeactivateOnError = value;
}
TINE_EXPORT int GetHistoryDeactivateOnError(void)
{
  return gHstDeactivateOnError;
}
char arcDbPath[HISTORY_HOME_SIZE] = "";

void validateHistoryHomePath(void)
{
# ifdef FS_DELIMITER
  int len;
  if ((len=(int)strlen(arcDbPath)) == 0) 
  {
    strcpy(arcDbPath,".");
    len = 1;
  }
  if (arcDbPath[len-1] != FS_DELIMITER && len < HISTORY_HOME_SIZE) arcDbPath[len] = FS_DELIMITER;
# endif
}
TINE_EXPORT void SetHistoryFilesRepository(char *path)
{
  if (path == NULL) return;
  history_home_done = TRUE;
  memset(arcDbPath,0,HISTORY_HOME_SIZE);
  strncpy(arcDbPath,path,HISTORY_HOME_SIZE-1);
  validateHistoryHomePath();
  feclog("HISTORY HOME repository %.128s set by API",arcDbPath);
}
TINE_EXPORT char *GetHistoryFilesRepository(void)
{
  return arcDbPath;
}
char arcDbPathSaved[HISTORY_HOME_SIZE] = "";
TINE_EXPORT void SetHistoryStaticFilesRepository(char *path)
{
  int len;
  if (path == NULL) return;
  memset(arcDbPathSaved,0,HISTORY_HOME_SIZE);
  strncpy(arcDbPathSaved,path,HISTORY_HOME_SIZE-1);
  if ((len=(int)strlen(arcDbPathSaved)) == 0) 
  {
    strcpy(arcDbPathSaved,".");
    len = 1;
  }
# ifdef FS_DELIMITER
  if (arcDbPathSaved[len-1] != FS_DELIMITER && len < 128) arcDbPathSaved[len] = FS_DELIMITER;
# endif
  feclog("HISTORY static repository set to %.128s by API",arcDbPathSaved);
}
TINE_EXPORT char *GetHistoryStaticFilesRepository(void)
{
  return arcDbPathSaved;
}
HstTblEntry *hstTbl = NULL;
int nHistoryRecords = 0;
int addHistoryCapacity = HISTORY_RECORD_CAPACITY_ADD;
TINE_EXPORT int maxHistoryRecords = MAX_HISTORY_RECORDS;
TINE_EXPORT void SetMaximumHistoryTableRecords(int value)
{
  if (value > 0) maxHistoryRecords = value;
}
TINE_EXPORT int minHistoryPollingRate = 1000;
TINE_EXPORT int useMonthlyHistoryFiles = FALSE;
TINE_EXPORT int useMinimalStorage = FALSE;
static int stdWorstCaseSpillOver = IMPORTANT_RECS_DAY;
TINE_EXPORT void SetStandardSpillOverSize(int value)
{
  if (value < 0) return;
  stdWorstCaseSpillOver = value;
}
TINE_EXPORT int GetStandardSpillOverSize(void)
{
  return stdWorstCaseSpillOver;
}
static int gScavengeFiles = FALSE;
static time_t gScavengeFileTime = 0;
static unsigned long minFreeDiskSpaceInBlocks = 0;
TINE_EXPORT void SetMinimumDiskSpaceInBlocks(UINT32 spaceInBlocks)
{
  minFreeDiskSpaceInBlocks = (long)spaceInBlocks;
  feclog("HIST: set minimum free disk free space to %lu from API",minFreeDiskSpaceInBlocks);
}

int gScavangerThreadBusy = FALSE;
#if defined(MULTITHREADED) && !defined(VXWORKS)
THRHANDLE hScvrThrd = 0;
THREAD_TYPE scvThreadTask(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;
  gScavengeFiles = FALSE;
  feclog("scavenger thread now active");
  if (minFreeDiskSpaceInBlocks > 0) 
  {
    scavengeOlderLTSFiles(minFreeDiskSpaceInBlocks,gScavengeFileTime);
  }
# if defined(WIN32)
  if (hScvrThrd != (THRHANDLE)0) CloseHandle(hScvrThrd);
# endif
  gScavangerThreadBusy = FALSE;
  return 0;
}
#endif
int useStandardFileSet = FALSE;
#define GetRecordPrefixLength(ms) ((ms) ? sizeof(SINT32) : (sizeof(double)+2*sizeof(SINT32)))

#define MakeScanFileName(nm,id) (\
  sprintf((nm),"%s*.%0x",GetFilePrefix(id),hstTbl[(id)].recordIndex)\
)

#define MakeTargetFileName(nm,yy,mm,dd,id) (\
  sprintf((nm),"%s%02d%02d%02d.%0x",GetFilePrefix(id),((yy)%100),(mm),\
  useMonthlyHistoryFiles ? 1 : (dd),hstTbl[(id)].recordIndex)\
)

#define HSTRECLEN(h) ((h)->storeSummary ? (int)sizeof(DSUMMARY) : (h)->recordLength)
#define HSTDEPLNG(h) ((h)->storeSummary ? 1 : (h)->depthLong)
#define HSTDEPDAYS(h) ((h)->storeSummary ? 0 : (h)->depthDays)

int nDaysInMonthTable[2][12] = {
  {31,28,31,30,31,30,31,31,30,31,30,31},
  {31,29,31,30,31,30,31,31,30,31,30,31},
};
#define nDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonthTable[0][mm] : nDaysInMonthTable[1][mm])

int populateHRS(char *fecname,DBLNKLIST *fecList,HRIS **hris);

TINE_EXPORT void SetUseStandardHistoryFiles(int value)
{
# ifndef VXWORKS
  useStandardFileSet = value;
  if (value) feclog("HIST: API set to allocate standard non-fragmented history files in history home");
# else
  printf("SetUseStandardHistoryFiles has no effect for %s",OS_TAG);
# endif
}
char *GetFilePrefix(int id)
{
  if (hstTbl[(id)].useStandardSet) return "ts";
  return useMinimalStorage ? "ar" : "ta";
}
int getHistoryTableIndex(HstTblEntry *hst,int nentries)
{
  int i;
  for (i=0; i<nentries; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,hst->c.EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,hst->c.EqmProperty,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,hst->c.EqmDeviceName,DEVICE_NAME_SIZE)) continue;
    if (hst->c.EqmSizeOut == 0 && hst->c.EqmFormatIn == 0)
    { /* quick search for a match vs. calling address */
      hst->c.EqmSizeOut = hstTbl[i].c.EqmSizeOut;
      hst->c.EqmFormatOut = hstTbl[i].c.EqmFormatOut;
      return i;
    }
    if (hstTbl[i].c.EqmSizeOut != hst->c.EqmSizeOut) continue;
    if (hstTbl[i].c.EqmFormatOut != hst->c.EqmFormatOut) continue;
    return i;
  }
  return -1;
}
int isInHistoryTable(HstTblEntry *hst,int nentries)
{
  if (getHistoryTableIndex(hst,nentries) < 0) return FALSE;
  feclog("HIST: ignoring double entry (%.6s)/%.64s[%.64s]",hst->c.EqmName,hst->c.EqmDeviceName,hst->c.EqmProperty);
  return TRUE;
}
int getHstTblEntryFromRecordIndex(int idx)
{
  int i;
  if (hstTbl == NULL) return -1;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (hstTbl[i].recordIndex == idx) return i;
  }
  return -1;
}
#ifndef SMALL_TINE_LIB
int prepIncomingHstFilter(BYTE *filterstream, FilterStruct *flst)
{
# ifdef INTEL
  memcpy(flst, filterstream, FILTERSTRUCT_SIZE);
# else
  _systemStructSwap(FilterStructStruct, filterstream, (BYTE *)flst, RTOL, 1, sizeof(FilterStruct));
# endif
  return 0;
}
int prepOutgoingHstFilter(BYTE *filterstream, int length)
{
  int i;
  char filterString[256];
  FilterStruct flst;
  HstTblEntry *hstEnt = NULL;

  for (i = 0; i < length; i++)
  { 
    hstEnt = &hstTbl[i];
    flst.filterIdx = getHistoryTableIndex(hstEnt, length);
    getHistoryFilterString(hstEnt->fltr, filterString);
    strncpy(flst.filterLinkStr, filterString, 256);
#   ifdef INTEL
    memcpy(&filterstream[i * FILTERSTRUCT_SIZE], &flst, FILTERSTRUCT_SIZE);
#   else
    _systemStructSwap(FilterStructStruct, &filterstream[i*FILTERSTRUCT_SIZE], (unsigned char *)&flst, LTOR, 1, sizeof(FilterStruct));
#   endif
  }
  
  return 0;
}
#ifndef VXWORKS
static unsigned long fsBlocksFree = 0;
unsigned long getAvailableDiskSpaceInBlocks(char *path)
{
# if defined(WIN32)
  ULARGE_INTEGER freeBytesCaller;
  ULARGE_INTEGER totalBytes;
  ULARGE_INTEGER totalFreeBytes;
  
  if (path == NULL) return 0;
  if (GetDiskFreeSpaceEx(path,&freeBytesCaller,&totalBytes,&totalFreeBytes) == 0) goto err;
  return (unsigned long)(freeBytesCaller.QuadPart / 1024);
# elif defined(UNIX)
# ifdef SOLARIS
  struct statvfs fsbuf;
  if (path == NULL) goto err;
  if (statvfs(path, &fsbuf) != 0) goto err;
# else
  struct statfs fsbuf;
  if (path == NULL) return 0;
  if (statfs(path, &fsbuf) != 0) goto err;
# endif
  return (unsigned long)fsbuf.f_bavail;
# else
  return minFreeDiskSpaceInBlocks;
# endif
err:
  feclog("HIST: cannot get available disk space : %s",SystemGetLastErrorString());
  return minFreeDiskSpaceInBlocks;
}
#endif
int getWorstCaseNumRecordsInFile(HstTblEntry *hst)
{
  double ndps;
  if (hst == NULL) return -argument_list_error;
  ndps = 1000.0 / (double)hst->archiveRate;
  ndps *= SEC_IN_DAY;
  ndps += stdWorstCaseSpillOver; /* buffer space for 'scheduled' storage */
  if (useMonthlyHistoryFiles)
  { /* how many times per month is possible ? */
    return (int)ndps * 31;
  }
  /* how many times per day is possible ? */
  return (int)ndps;
}
int getWorstCaseFileSpaceSizeBlocks(HstTblEntry *hst)
{
  int space = 0,recsPerFile,recSize;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  int depth = HSTDEPLNG(hst);
  if (depth <= 0) return 0;
  recsPerFile = getWorstCaseNumRecordsInFile(hst);
  recSize = HSTRECLEN(hst);
  recSize += recordPrefix;
  space = recsPerFile * recSize + AS_STAT_SIZE;
  space /= 1000; /* blocks */
  space *= (depth+1);
  return space;
}
#define MAX_NUM_OUTDATEDFILES 100
NAME32 *outdatedFiles = NULL;
int nOutdatedFiles = 0;
int nStdDailyFiles = 0;
int canRescanOutdatedFiles = FALSE;
int scanForOutdatedFiles(int index,int mdepth,int ddepth)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int n = 0, dd, mm, yy, fdd, fmm, fyy;
  char dirname[256],filename[256],fname[64],fileext[16],fn[64],str[128];
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif
  struct tm *tm, lt;
  time_t now = time(NULL);

  if (ServerExitCondition) return -not_running;
  nStdDailyFiles = 0;
  if (outdatedFiles == NULL) 
  {
    if ((outdatedFiles=(NAME32 *)SystemCalloc(MAX_NUM_OUTDATEDFILES,sizeof(NAME32))) == NULL)
      return -out_of_server_memory;
    nOutdatedFiles = MAX_NUM_OUTDATEDFILES;
  }
  if (index < 0 || index >= nHistoryRecords) return -invalid_index;
  if (mdepth <= 0 && ddepth <= 0) return 0;
  /* what year and month is it now ? */
  if ((tm=localtime(&now)) == NULL) return -code_failure; 
  lt = *tm;
  yy = lt.tm_year % 100;
  mm = lt.tm_mon - mdepth + 1;
  while (mm < 1) 
  {
    mm += 12;
    yy -= 1;
  }
  dd = lt.tm_mday - ddepth;
  if (dd < 1)
  {
    if (--mm < 1) { mm += 12; yy--; }
    dd = nDaysInMonth(lt.tm_year,mm-1) + dd;
  }

  MakeScanFileName(fname,index);
  sprintf(fileext,".%0x",hstTbl[index].recordIndex);

# if defined(UNIX)
  sprintf(dirname,"%.128s",arcDbPath);
  if ((dp=opendir(dirname)) == NULL) return 0;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name,fileext) == NULL) continue;
    strncpy(fn,dirp->d_name,32); fn[32] = 0;
    sprintf(filename,"%.128s%.32s",arcDbPath,dirp->d_name);
# elif defined(WIN32)
  sprintf(dirname,"%.128s%.32s",arcDbPath,fname);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (strstr(fd.name,fileext) == NULL) continue;
    strncpy(fn,fd.name,32); fn[32] = 0;
    sprintf(filename,"%.128s%.32s",arcDbPath,fd.name);
# endif
    strncpy(str,&fn[2],2); str[2] = 0;
    fyy = atoi(str);
    strncpy(str,&fn[4],2); str[2] = 0;
    fmm = atoi(str);
    strncpy(str,&fn[6],2); str[2] = 0;
    fdd = atoi(str);
    if (fyy > yy) continue;
    if (fyy == yy && fmm > mm) continue;
    if (fyy == yy && fmm == mm)
    {
      if (ddepth <= 0) continue;
      if (fdd >= dd) 
      {
        nStdDailyFiles++;
        continue;
      }
    }
    if (n >= MAX_NUM_OUTDATEDFILES) break;
    strncpy(outdatedFiles[n].name,fn,32);
    n++;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
  _findclose(hfp);
# endif
  canRescanOutdatedFiles = (n == MAX_NUM_OUTDATEDFILES) ? TRUE : FALSE;
  return n; 
# else
  index = index;
  mdepth = mdepth;
  return 0;
# endif /* NO_LTS */
}
int getOldestFile(char *path,char *filePattern,char *fileName,time_t fileTime)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  struct stat sbuf;
  char *c,dirname[128],pfn[128],fn[64],fpat[64];
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif
  int hasPattern = FALSE;
  time_t otime = time(NULL);
  fileName[0] = 0;
  if (filePattern != NULL && strlen(filePattern) > 0)
  {
    strncpy(fpat,filePattern,64); fpat[63] = 0;
    if ((c=strchr(fpat,'*')) != NULL) *c = 0;
    if (strlen(fpat) > 0) hasPattern = TRUE;
  }
# if defined(UNIX)
  strncpy(dirname,path,96); dirname[96] = 0;
  if ((dp=opendir(dirname)) == NULL) return FALSE;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (hasPattern && strstr(dirp->d_name,fpat) == NULL) continue;
    strncpy(fn,dirp->d_name,32); fn[32] = 0;
# elif defined(WIN32)
  sprintf(dirname,"%s*.*",path);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return FALSE;
  do
  {
    if (hasPattern && strstr(fd.name,fpat) == NULL) continue;
    strncpy(fn,fd.name,32); fn[32] = 0;
# endif
    if (!strcmp(fn,"..") || !strcmp(fn,".")) continue;
    sprintf(pfn,"%s%s",path,fn);
    if (stat(pfn,&sbuf) != 0) continue;
    if (sbuf.st_mtime >= fileTime) continue;
    if (sbuf.st_mtime < otime) 
    {
      otime = sbuf.st_mtime;
      strncpy(fileName,fn,64);
    }
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
  _findclose(hfp);
# endif
  return strlen(fileName) > 0 ? TRUE : FALSE;
# else
  path = path;
  filePattern = filePattern;
  fileName = fileName;
  return FALSE;
# endif /* NO_LTS */
}
int checkForStandardFilesEnvironment(void)
{
  char *ptr;
  if ((ptr=getenv("TINE_HISTORY_ALLOC_FILESET")) != NULL ||
      (ptr=getenv("HISTORY_ALLOC_FILESET")) != NULL)
  {
    if (strtobool(ptr))
    {
      feclog("HIST: environment variable set to allocate standard non-fragmented history files in history home");
      return TRUE;
    }
  }
  return FALSE;
}
int checkForStandardFiles(void)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  char dirname[160];
  int filesfound = FALSE;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

# if defined(UNIX)
  sprintf(dirname,"%.128s",arcDbPath);
  if ((dp=opendir(dirname)) == NULL) return FALSE;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strncmp(dirp->d_name,"ts",2) != 0) continue;
# elif defined(WIN32)
  sprintf(dirname,"%.128s*.*",arcDbPath);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (strncmp(fd.name,"ts",2) != 0) continue;
# endif
    filesfound = TRUE;
    break;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
  _findclose(hfp);
# endif
  if (filesfound) feclog("HIST: found standard non-fragmented history files in history home");
  return filesfound;
# else
  return FALSE;
# endif /* NO_LTS */
}
int checkForSavedFiles(int idx)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  char dirname[160];
  int filesfound = FALSE;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
  char *c, ext[8];
# elif defined(WIN32)
  struct _finddata_t fd;
  int ff = FALSE;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

# if defined(UNIX)
  sprintf(dirname, "%.128s", arcDbPathSaved);
  if ((dp=opendir(dirname)) == NULL) return FALSE;
  snprintf(ext, 8, "%x", idx);
  while ((dirp = readdir(dp)) != NULL)
  {
    if ((c=strrchr(dirp->d_name, '.')) == NULL) continue;
    if (strstr(c, ext) == NULL) continue;
# elif defined(WIN32)
  sprintf(dirname, "%.128s*.%x", arcDbPathSaved,idx);
  if ((hfp =_findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (strncmp(fd.name, "ts", 2) == 0) ff = TRUE;
    if (strncmp(fd.name, "ta", 2) == 0) ff = TRUE;
    if (strncmp(fd.name, "ar", 2) == 0) ff = TRUE;
    if (!ff) continue;
# endif
    filesfound = TRUE;
    break;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp, &fd) != -1);
  _findclose(hfp);
# endif
  if (filesfound) feclog("HIST: found long-term history files in saved area %.128s", arcDbPathSaved);
  return filesfound;
# else
  return FALSE;
# endif /* NO_LTS */
}
int resetStats(char *fname,int recsize)
{
  int cc = 0;
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int hFile = -1, hAltFile = -1;
  long fpos = 0;
  UINT32 statbuf[16];
  char fileName[256], altFileName[256], aname[64];
  BYTE *tmpHstDatBuf;
  time_t t = 0;

  if (fname == NULL || recsize == 0) ccerr(argument_list_error);

  sprintf(fileName,"%.128s%.32s",arcDbPath,fname);
  /* prepare alternate file name in order to re-use old-style data */
  strncpy(aname,fname,64);
  if (strncmp(aname,"ts",2) != 0) ccerr(invalid_parameter);
  strncpy(aname,"ta",2);
  sprintf(altFileName,"%.128s%.32s",arcDbPath,aname);
  if ((hAltFile=open(altFileName,O_readmode)) == -1)
  { /* also look in the 'static' directory for files to convert */
    sprintf(altFileName,"%.128s%.32s",arcDbPathSaved,aname);
    hAltFile = open(altFileName,O_readmode);
  }
  if ((hFile=open(fileName,O_writemode)) == -1)
  { /* the principal file */
    if (tineDebug || errorDebug) perror(fileName);
    ccerr(file_error);
  }
  if (hAltFile != -1)
  { /* there was an old guy -> re-use the data */
    feclog("HIST: convert usable data from alternate file %.256s to standard file %.256s",
      altFileName,fileName);
    lseek(hFile,AS_STAT_SIZE,SEEK_SET);
    tmpHstDatBuf = (BYTE *)alloca(recsize);
    if (tmpHstDatBuf == NULL) ccerr(out_of_local_memory);
    lseek(hAltFile,0,SEEK_SET);
    while(read(hAltFile,(char *)tmpHstDatBuf,recsize) == recsize)   
    { /* read through file */
      if (write(hFile,(char *)tmpHstDatBuf,recsize) != recsize) break;
      fpos += recsize;
    }
    t = time(NULL); /* give it the current time */
    close(hAltFile);
    /* now remove the old file */
    unlink(altFileName);
  }
  if (hstlog) feclog("HIST: adjust file position pointer for %.256s to %ld",fileName,fpos);
  statbuf[0] = (UINT32)fpos;
  statbuf[1] = (UINT32)t;
  statbuf[2] = recsize;
  statbuf[3] = 0;
  lseek(hFile,0L,SEEK_SET);
  if (write(hFile,(char *)statbuf,16) != 16) dbglog("HIST: %.125s -> %.32s",fileName,cc2str(no_file_space));

err:
  if (hFile != -1) close(hFile);
  if (cc != 0) feclog("HIST: reset std file stat header -> %.32s",cc2str(cc));
# else
  fname = fname;
# endif
  return cc;
}
int _stdFileSetProbe_ = FALSE; /* undocumented flag used by mkhstfiles */
/**
 * \internal
 *
 * \brief create the standard archive files 
 * according to depth in months, record size and archive rate.
 * 
 */
int makeStandardFiles(int index,int mdepth,int mday)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int hFile = -1, mm, yy, ptr, isCurrentFile = FALSE, wasFound = FALSE;
  int cc = 0, filesPerMonth = 1, recsPerFile, recsCapacity, reclen, recSize, needResetStats = FALSE;
  int hPoolFile = -1, poolSize = 0, poolCapacity = 0;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  char fileName[256], poolName[256], fname[64], tmp[256];
  UINT32 statbuf[4];
  time_t t=time(NULL), ftime;
  struct tm lt, ft, *tm;
  HstTblEntry *hst;

  if (ServerExitCondition) return not_running;
  if (index < 0 || index >= nHistoryRecords) return invalid_index;
  hst = &hstTbl[index];

  if ((tm=localtime(&t)) == NULL) return code_failure; 
  lt = *tm; 
  mm = lt.tm_mon - mdepth + 1;
  yy = lt.tm_year;

  if (mdepth == 0 && (useMonthlyHistoryFiles || mday == lt.tm_mday))
  {
    isCurrentFile = TRUE;
  }
  else
  {
    if (mm == (lt.tm_mon + 1) && mday > lt.tm_mday) 
    { /* possible recycled (future) file */
      needResetStats = FALSE; /* TODO: was TRUE, but this shouldn't be necessary -> remove it later */ 
    }
    while (mm < 1) { mm += 12; yy -= 1; }
  }

  MakeTargetFileName(fname,yy,mm,mday,index);
  sprintf(fileName,"%.128s%.32s",arcDbPath,fname);

  if (nOutdatedFiles > 0)
  { /* there are outdated files -> re-use them ! */
    if ((hFile=open(fileName,O_mode)) == -1)
    { /* not there, so rotate an old one */
      nOutdatedFiles--;
      if (hstlog) feclog("HIST: recycle history file %.32s to %.32s",outdatedFiles[nOutdatedFiles].name,fname);
      sprintf(tmp,"%.132s%.32s",arcDbPath,outdatedFiles[nOutdatedFiles].name);
      rename(tmp,fileName);
      if (hst->hasDataPool)
      { /* complex data type */
        sprintf(poolName,"%.128std%.30s",arcDbPath,&fname[2]);
        sprintf(tmp,"%.132std%.30s",arcDbPath,&outdatedFiles[nOutdatedFiles].name[2]);
        if (rename(tmp,poolName) != 0)
        { /* file problems ! */
          feclogEx(TINE_LOGLEVEL_ERROR, "could not rename data pool file %s to %s",tmp,fileName);
          perror("rename data pool file");
        }
      }
      if (nOutdatedFiles == 0 && canRescanOutdatedFiles) 
      { /* was maxed out: refill the outdated files buffer */
        nOutdatedFiles = scanForOutdatedFiles(index,HSTDEPLNG(hst),HSTDEPDAYS(hst));
        if (hstlog)
        {
          feclog("HIST: rescanning for outdated files for rec %d (0x%x) -> (%.6s)/%.64s[%.64s] : %d additional files",
            hst->recordIndex,hst->recordIndex,
            hst->c.EqmName,hst->c.EqmDeviceName,hst->c.EqmProperty,nOutdatedFiles);
        }
      }
    }
    else
    { /* was there ! */
      close(hFile);
    }
  }
  if ((hFile=open(fileName,O_createmode)) == -1)
  {
    if (tineDebug || errorDebug) perror(fileName);
    return file_error;
  }
  reclen = HSTRECLEN(hst);
  recsPerFile = getWorstCaseNumRecordsInFile(hst);
  recSize = reclen + recordPrefix;
  recsCapacity = recsPerFile * recSize + AS_STAT_SIZE;
  hst->recordCapacity = recsCapacity;
  if (hst->hasDataPool)
  { /* complex data type */
    sprintf(poolName,"%.128std%.30s",arcDbPath,&fname[2]);
    if ((hPoolFile=open(poolName,O_createmode)) == -1)
    {
      if (tineDebug || errorDebug) perror(poolName);
      return file_error;
    }
    poolSize = hst->poolEntryCapacity;
    poolCapacity = recsPerFile * poolSize + AS_STAT_SIZE;
    if ((ptr=lseek(hPoolFile,0L,SEEK_END)) > 0) 
    { /* found a file; is it big enough or too big? */
      wasFound = TRUE;
      if (ServerInitialized) 
      { /* if already there (e.g. made with mkhstfiles) then keep ! */
        ccerr(0);
      }
      if (ptr >= poolCapacity && ptr < poolCapacity + 0x1000) 
      { /* good enough */
        ccerr(0);
      }
      feclog("HIST: rec %d (0x%x) file pool capacity for %d records needs to be adjusted to %d records",
        hst->recordIndex,hst->recordIndex,(ptr-AS_STAT_SIZE)/recSize,recsPerFile);
    }
  }
  if ((ptr=lseek(hFile,0L,SEEK_END)) > 0) 
  { /* found a file; is it big enough or too big? */
    wasFound = TRUE;
    if (ServerInitialized) 
    { /* if already there (e.g. made with mkhstfiles) then keep ! */
      ccerr(0);
    }
    if (ptr >= recsCapacity && ptr < recsCapacity + 50) 
    { /* good enough */
      if (!hst->hasDataPool) ccerr(0);
    }
    feclog("HIST: rec %d (0x%x) file space capacity for %d records needs to be adjusted to %d records",
      hst->recordIndex,hst->recordIndex,(ptr-AS_STAT_SIZE)/recSize,recsPerFile);
  }
  if (tineDebug) dbglog("make standard local history file : %s",fileName);
  if (mday == 1)
  {
    feclog("HIST: creating file space for history rec %d (0x%x) (depth %d months)",
      hst->recordIndex,hst->recordIndex,mdepth);
    if (!useMonthlyHistoryFiles) 
    {
      filesPerMonth = 31;
      feclog("file space includes 31 separate files per month");
    }
    feclog("HIST: file space for history rec %d (0x%x) requires %d blocks",
      hst->recordIndex,hst->recordIndex,filesPerMonth*getWorstCaseFileSpaceSizeBlocks(hst));
  }
  if (_stdFileSetProbe_) ccerr(0); /* used by mkhstfiles utility */
  if (chsize(hFile,recsCapacity) != 0) ccerr(no_file_space);
  if (hst->hasDataPool && chsize(hPoolFile,poolCapacity) != 0) ccerr(no_file_space);
err:
  /* are there records written to the file ? */
  if (wasFound && cc == 0)
  { /* not the first startup !*/
    if (isCurrentFile)
    { /* ensure that file time is not too old */
      lseek(hFile,0L,SEEK_SET);
      if (read(hFile,(char *)&statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) 
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "HIST: error reading local history standard file : %s",fileName);
        return file_error;
      }
      ptr = statbuf[0];
      ftime = (time_t)statbuf[1];
      if (recSize != (int)statbuf[2]) 
      {
        feclog("HIST: needed to adjust record size in local history standard file : %s",fname);
        needResetStats = TRUE;
      }
      if ((tm=localtime(&ftime)) == NULL) return code_failure;
      if (ptr > 0 || ftime > 0)
      {
        ft = *tm;
        if (ft.tm_year != yy || ft.tm_mon != mm - 1) needResetStats = TRUE;
        if (!useMonthlyHistoryFiles && ft.tm_mday != mday) needResetStats = TRUE;
      }
    }
  }
  if (hFile != -1) 
  {
    close(hFile);
    if (!wasFound || needResetStats)
    { /* fill in position = 0, UTC = 0 */
      resetStats(fname,recSize);
    }
  }
  if (hPoolFile != -1) 
  { /* companion complex data file */
    close(hPoolFile);
    if (!wasFound || needResetStats)
    { /* fill in position = 0, UTC = 0 */
      fname[0] = 't'; fname[1] = 'd';
      resetStats(fname,recSize);
    }
  }
  return cc;
# else
  index = index;
  mdepth = mdepth;
  mday = mday;
  return 0;
# endif /* NO_LTS */
}
#endif /* !SMALL_TINE_LIB */
char *SystemGetLastErrorString(void)
{
  static char msgbuf[256];
# if defined(WIN32)
  LPVOID lpMsgBuf;
  DWORD lastErr = GetLastError();
  msgbuf[0] = 0;
  if (FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,lastErr,
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), 
        (LPTSTR) &lpMsgBuf,0,NULL) > 0)
  {
    strncpy(msgbuf,(char *)lpMsgBuf,256);
    LocalFree(lpMsgBuf);
  }
# elif defined(UNIX)
  strncpy(msgbuf,strerror(errno),256);
# else
  msgbuf[0] = 0;
# endif
  return msgbuf;
}
int rotateHistoryFile(char *srcfn,char *dstfn,HstTblEntry *hst)
{
# if !defined(NO_LTS) && !defined(VXWORKS)
  int cc = 0, reclen = HSTRECLEN(hst);
  char srcfile[256],dstfile[256],pifn[32];
  struct stat sbuf;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);

  if (srcfn == NULL || dstfn == NULL || hst == NULL) ccerr(argument_list_error);
  sprintf(dstfile,"%.128s%.32s",arcDbPath,dstfn);
  if (stat(dstfile,&sbuf) == 0) ccerr(0); /* already there */
  sprintf(srcfile,"%.128s%.32s",arcDbPath,srcfn);
  if (rename(srcfile,dstfile) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "HIST: could not rename %.32s to %.32s : %.128s",srcfn,dstfn,SystemGetLastErrorString());
    strncpy(hst->rmvFileName,srcfn,32);
    ccerr(file_error);
  }
  hst->rmvFileName[0] = 0;
  /* remove any possible points-of-interest files ... */
  strncpy(pifn,srcfn,32);
  pifn[0] = 'p'; pifn[1] = 'i';
  sprintf(srcfile,"%.128s%.32s",arcDbPath,pifn);
  unlink(srcfile);
err:
# ifndef SMALL_TINE_LIB
  resetStats(dstfn,reclen+recordPrefix);
# else
  reclen = reclen;
# endif
  if (cc) feclog("HIST: rotateHistoryFile %.256s -> %.256s : %.32s",srcfile,dstfile,cc2str(cc));
  return cc;
# else
  srcfn = srcfn;
  dstfn = dstfn;
  return 0;
# endif
}
long getFilePosition(int hFile,int useStdFileSet)
{
  int cc = 0;
  long pos = -1;
  UINT32 statbuf[4];
  if (useStdFileSet)
  {
    lseek(hFile,0L,SEEK_SET);
    if (read(hFile,(char *)&statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) ccerr(file_error);
    pos = (long)statbuf[0] + AS_STAT_SIZE;
  }
  else
  {
    pos = lseek(hFile,0L,SEEK_END);
  }
err:
  if (cc) rprtSckErr("getFilePosition","read",0,errno,0,NULL);
  return cc != 0 ? -cc : pos;
}
/**
 * \internal used only in 'starndard files'
 */
int setFilePosition(int hFile,long pos)
{
  int cc = 0;
  time_t t = time(NULL);
  UINT32 statbuf[4];
  lseek(hFile,0L,SEEK_SET);
  statbuf[0] = (UINT32)pos - AS_STAT_SIZE;
  statbuf[1] = (UINT32)t;
  if (write(hFile,(char *)&statbuf,AS_STAT_SIZE/2) != AS_STAT_SIZE/2) ccerr(file_error);
err:
  if (cc) rprtSckErr("setFilePosition","write",0,errno,0,NULL);
  return cc;
}
long getNumberOfRecords(int hFile,int recsize,int useStdFileSet)
{
  int cc = 0;
  long pos = 0, siz = 0;
  UINT32 statbuf[4];
  if (useStdFileSet)
  {
    lseek(hFile,0L,SEEK_SET);
    if (read(hFile,(char *)&statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) ccerr(file_error);
    pos = (long)statbuf[0];
    siz = (long)statbuf[2];
    if (siz != recsize) ccerr(invalid_data);
  }
  else
  {
    pos = lseek(hFile,0,SEEK_END);
    siz = recsize;
  }
err:
  return cc != 0 ? (long)(-cc) : pos/siz;
}
#if !defined(SMALL_TINE_LIB) && !defined(VXWORKS)
int scavengeOlderLTSFiles(unsigned long minDiskSpaceInBlocks,time_t currentFileTime)
{
  char fn[256],ofn[64];

  while (getAvailableDiskSpaceInBlocks(arcDbPath) < minDiskSpaceInBlocks)
  {
    if (getOldestFile(arcDbPath,"ar*",ofn,currentFileTime) == FALSE && 
        getOldestFile(arcDbPath,"ta*",ofn,currentFileTime) == FALSE   ) break;
    if (strlen(ofn) == 0) break;
    sprintf(fn,"%.128s%.32s",arcDbPath,ofn);
    if (unlink(fn) != 0) break;
    feclog("HIST: remove file %.256s to increase free disk space",fn);
    ofn[0] = 'p'; ofn[1] = 'i';
    sprintf(fn,"%.128s%.32s",arcDbPath,ofn);
    unlink(fn); /* also remove possible points-of-interest file */
  }
  gScavengeFiles = FALSE;
  return 0;
}
void doLtsScavenger(void)
{
# if defined(MULTITHREADED) && !defined(VXWORKS)
  /* launch a thread to do this activity (might be lengthy) */
  hScvrThrd = scavengerCreateThread(NULL);
# else
  /* gotta handle it here ... */
  scavengeOlderLTSFiles(minFreeDiskSpaceInBlocks,gScavengeFileTime);
# endif
}
#endif
int removeOutdatedLTSFile(char *fn,int dkeep,int depthDays)
{
  int filesRemoved = 0;
# ifndef NO_LTS
  int i;
  char fileName[256],lfn[32],pifn[32];
  static int stagger = -1;

  if (stagger < 0) stagger = (int)time(NULL); /* seed the start up */
  /* stagger file removal to avoid blocking the file system */
  stagger = (stagger + 1) % 100;
  if (stagger > 0) return -1; 

  for (i=1; i<32; i++)
  {
    if (depthDays > 0  && i >= dkeep) continue;
    strncpy(lfn,fn,6);
    sprintf(&lfn[6],"%02d",i);
    strcat(lfn,&fn[8]);
    sprintf(fileName,"%.128s%.32s",arcDbPath,lfn);
    if (unlink(fileName) == 0) filesRemoved++;
    /* remove any possible points-of-interest files ... */
    strncpy(pifn,lfn,32);
    pifn[0] = 'p'; pifn[1] = 'i';
    sprintf(fileName,"%.128s%.32s",arcDbPath,pifn);
    unlink(fileName);
  }
# endif /* NO_LTS */
  if (filesRemoved > 0) feclog("HIST: removed %d files from long term storage",filesRemoved);
  return filesRemoved;
}
int resetLTSFiles (HstTblEntry *hst)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int n = 0;
  char dirname[256], filename[256], fname[64], fileext[16], fn[64];
# if defined(UNIX)
  DIR* dp;
  struct dirent* dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

  if (ServerExitCondition) return -not_running;
  if (hst == NULL) return -invalid_index;
  sprintf(fname, "*.%0x", hst->recordIndex);
  sprintf(fileext, ".%0x", hst->recordIndex);

# if defined(UNIX)
  sprintf(dirname, "%.128s", arcDbPath);
  if ((dp = opendir(dirname)) == NULL) return 0;
  while ((dirp = readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name, fileext) == NULL) continue;
    strncpy(fn, dirp->d_name, 32); fn[32] = 0;
    sprintf(filename, "%.128s%.32s", arcDbPath, dirp->d_name);
# elif defined(WIN32)
  sprintf(dirname, "%.128s%.32s", arcDbPath, fname);
  if ((hfp = _findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (strstr(fd.name, fileext) == NULL) continue;
    strncpy(fn, fd.name, 32); fn[32] = 0;
    sprintf(filename, "%.128s%.32s", arcDbPath, fd.name);
# endif
    if (hst->useStandardSet)
    {
      resetStats(filename, hst->recordLength);
    }
    else
    {
      unlink(filename);
    }
    n++;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp, &fd) != -1);
  _findclose(hfp);
# endif
  return n;
# else
  UNUSED(hst);
  return 0;
# endif /* NO_LTS */
}
int checkFileIntegrity(HstTblEntry *hst)
{
  int hFile,segsiz,cc=0,mvFile=FALSE,reclen;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  long fpos;
  char fileName[256],mvName[256];

  sprintf(fileName,"%.128s%.32s",arcDbPath,hst->archiveFileName);
  if ((hFile=open(fileName,O_readmode)) == -1) 
  { /* doesn't exist yet:  this is okay */
    return 0;
  }
  if ((fpos=getFilePosition(hFile,hst->useStandardSet)) < 0) ccerr(file_error);
  if (hst->useStandardSet) fpos -= AS_STAT_SIZE;
  reclen = HSTRECLEN(hst);
  segsiz = reclen + recordPrefix;
  if ((fpos % segsiz) != 0)
  {
    feclog("HIST: file alignment issue with %.32s (file will be reinitialized)",hst->archiveFileName);
    if (hst->useStandardSet) setFilePosition(hFile,AS_STAT_SIZE); else mvFile = TRUE;
  }
err:
  close(hFile);
  if (mvFile) 
  {
    sprintf(mvName,"%.128s%.32s.defunct",arcDbPath,hst->archiveFileName);
    rename(fileName,mvName);
  }
  return cc;
}
int WriteRecordToLTS(HstTblEntry *hst,int insertTransition,int useAlternatePath)
{ /* write record to long-term-storage file */
# ifndef NO_LTS
  int hFile=-1,hPoolFile=-1,p,ptr,reclen,cc=0,isTransition=FALSE;
  long fpos, ppos=0;
  size_t dp, dlen=0;
  char fileName[256];
  BYTE cmpHstDatRec[16], *tmpHstDatBuf, *srcptr, *dstptr;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  UINT32 lTimeStamp;
  DSUMMARY ds;
  static int cc_lastlogged = 0;
  static time_t time_lastlogged = 0;

  if (WaitForSystemMutex(hDbMutex,-1) != 0) return mutex_error;
  sprintf(fileName,"%.128s%.32s",arcDbPath,hst->archiveFileName);
  if (useAlternatePath)
  {
    char dname[256];
    sprintf(fileName,"%.128s%.32s",arcDbPathSaved,hst->archiveFileName);
    strncpy(dname, arcDbPathSaved, 128);
    if (!makeDirectoryPath(dname, TRUE)) ccerr(file_error);
  }
  reclen = HSTRECLEN(hst);
  hFile = hst->useStandardSet ? open(fileName,O_writemode) : open(fileName,O_createmode);
  if (hFile == -1)
  {
#   ifndef SMALL_TINE_LIB
    int i;
    if (hst->useStandardSet && strlen(hst->rmvFileName) > 0)
    { /* evidently the last rotation attempt failed: keep on trying to rotate the file */
      cc = rotateHistoryFile(hst->rmvFileName,hst->archiveFileName,hst);
      if (cc != 0) 
      { /* maybe there's no longer a removal target ? */
        feclogEx(TINE_LOGLEVEL_WARN, "HIST: attempt to recover from history file rotation error by re-making file records where necessary");
        i = getHstTblEntryFromRecordIndex(hst->recordIndex);
        makeFilesForRecord(i);
      }
    }
#   endif
    if (tineDebug || errorDebug) perror(fileName);
    ccerr(file_error);
  }
  if ((fpos=getFilePosition(hFile,hst->useStandardSet)) < 0) ccerr(file_error);
  lseek(hFile,fpos,SEEK_SET);
  /* write the timestamp and record as one unit to avoid corruption */
  tmpHstDatBuf = (BYTE *)alloca(reclen+recordPrefix);
  if (tmpHstDatBuf == NULL) ccerr(out_of_local_memory);
  ptr = hst->dataPtr;
  if (insertTransition)
  { /* instructed to insert a transition point; double check the timestamp: */
    ptr = (hst->dataPtr - 1 + hst->depthShort) % hst->depthShort; /* back up */
    if (hst->lastArchiveTime > 100 && 
        hst->timestamp[ptr] > hst->lastArchiveTime &&
        hst->timestamp[ptr] < hst->timestamp[hst->dataPtr])
    { /* okay to insert a transition point */
      isTransition = TRUE;
    }
    else
    { /* no!  set it back */
      isTransition = FALSE;
      ptr = hst->dataPtr;
    }
  }
commitToDisk:
  if (!useMinimalStorage)
  {
    p = 0;
    memcpy(tmpHstDatBuf,(char *)&hst->timestamp[ptr],sizeof(double)); p += sizeof(double);
    memcpy(&tmpHstDatBuf[p],(char *)&hst->sysstamp[ptr],sizeof(SINT32)); p += sizeof(SINT32);
    memcpy(&tmpHstDatBuf[p],(char *)&hst->usrstamp[ptr],sizeof(SINT32)); p += sizeof(SINT32);
  }
  else
  {
    lTimeStamp = (UINT32)hst->timestamp[ptr];
    memcpy(tmpHstDatBuf,(char *)&lTimeStamp,sizeof(UINT32));
  }
  if (hst->hasDataPool)
  { /* complex data type (cannot be a data summary) */
    sprintf(fileName,"%.128std%.30s",arcDbPath,&hst->archiveFileName[2]);
    hPoolFile = hst->useStandardSet ? open(fileName,O_writemode) : open(fileName,O_createmode);
    if (hPoolFile == -1) 
    {
      if (tineDebug || errorDebug) perror(fileName);
      ccerr(file_error);
    }
    dp = getFilePosition(hPoolFile,hst->useStandardSet);
    memcpy(cmpHstDatRec,&dp,sizeof(size_t));
    dlen = ((DataPoolEntry *)hst->data)[ptr].length;
    memcpy(&cmpHstDatRec[sizeof(size_t)],&dlen,sizeof(size_t));
    srcptr = cmpHstDatRec;
    if ((ppos=getFilePosition(hPoolFile,hst->useStandardSet)) < 0) ccerr(file_error);
    lseek(hFile,ppos,SEEK_SET);
    if (write(hPoolFile,(char *)((DataPoolEntry *)hst->data)[ptr].data,(unsigned int)dlen) < 0) ccerr(no_file_space);
  }
  else
  { /* normal case */
    if (hst->storeSummary)
    { /* get the summary from ring buffer */
      int i = getHstTblEntryFromRecordIndex(hst->recordIndex);
      double si = (double)hst->archiveRate/1000.0;
      if ((cc=getRBSummary(i,si,&ds)) != 0) ccerr(cc);
      srcptr = (BYTE *)&ds;
    }
    else
    {
      srcptr = hst->data + ptr * hst->recordLength;
    }
  }
  memcpy(&tmpHstDatBuf[recordPrefix],srcptr,reclen);

  if (write(hFile,(char *)tmpHstDatBuf,reclen+recordPrefix) < 0) ccerr(no_file_space);
  if (isTransition)
  { /* move the pointer back to 'this' point and commit to disk */
    ptr = hst->dataPtr;
    isTransition = FALSE;
    goto commitToDisk;
  }  
  if (hst->useStandardSet)
  {
    time_t t = time(NULL);
    fpos += reclen+recordPrefix;
    if (fpos > hst->recordCapacity + AS_STAT_SIZE && t > time_lastlogged + 60)
    {
      feclogEx(TINE_LOGLEVEL_WARN, "HIST: record position %d in %.32s exceeds capacity %d",
        fpos,hst->archiveFileName,hst->recordCapacity);
      time_lastlogged = t;
    }
    if ((cc=setFilePosition(hFile,fpos)) != 0) ccerr(cc);
    if (hPoolFile != -1) setFilePosition(hPoolFile,ppos+(long)dlen);
  }
  /* note the last archived data in the ring buffer for tolerance comparisons */
  if (hst->hasDataPool)
  { /* complex data */
    srcptr = ((DataPoolEntry *)hst->data)[ptr].data;
    dstptr = ((DataPoolEntry *)hst->data)[hst->depthShort].data;
    ((DataPoolEntry *)hst->data)[hst->depthShort].length = ((DataPoolEntry *)hst->data)[ptr].length;
  }
  else
  { /* normal case */
    dstptr = hst->data + hst->depthShort * hst->recordLength;
  }
  memcpy(dstptr, srcptr, hst->recordLength);
  hst->lastArchiveTime = hst->timestamp[ptr];
err:
  hst->lastArchiveSts = cc;
  if (hFile != -1) close(hFile);
  if (hPoolFile != -1) close(hPoolFile);
  ReleaseSystemMutex(hDbMutex);
  if (cc) 
  {
    dbglog("HIST: write record to long-term storage %.32s : %.32s",hst->archiveFileName,cc2str(cc));
    if (cc != cc_lastlogged)
    {
      cc_lastlogged = cc;
      feclog("HIST: write record %.32s  to %.128s: %.32s",hst->archiveFileName,arcDbPath,cc2str(cc));
    }
    setFileErrorAlarm(hst->archiveFileName);
  }
# endif /* NO_LTS */
  return 0;
}
int WriteRecordToPIF(HstTblEntry *hst)
{ /* write record #i to points-of-interest file */
  HstTblEntry pifhst;
  memcpy(&pifhst,hst,sizeof(HstTblEntry));
  pifhst.archiveFileName[0] = 'p';
  pifhst.archiveFileName[1] = 'i';
  pifhst.useStandardSet = FALSE;
  return WriteRecordToLTS(&pifhst,FALSE,FALSE);
}
long locateRecordFromLTS(char *fileName,int nidx,HstTblEntry *hst,double dt)
{ /* p0,d0 < p1,d1 < p2,d2 */
# ifndef NO_LTS
  BYTE rb[64];
  SINT32 t0,t1,t2;
  double dt0,dt1,dt2;
  long nrecs;
  long p0,p1,p2;
  double f;
  int hFile,it=0,useStdSet,cc=0;
  long lts_offset, recsize;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);

  useStdSet = (hst->useStandardSet && strncmp(&fileName[nidx],"ts",2) == 0) ? TRUE : FALSE;
  recsize = HSTRECLEN(hst);
  recsize += recordPrefix;

openLTSFile:
  if ((hFile=open(fileName,O_readmode)) == -1) 
  { /* file name found */
    if (useStdSet)
    { /* flagged as using standard file set : try legacy file name */
      memcpy(&fileName[nidx],"ta",2);
      useStdSet = FALSE;
      goto openLTSFile;
    }
    return -no_such_file;
  }
  nrecs = getNumberOfRecords(hFile,(int)recsize,useStdSet);
  lts_offset = useStdSet ? AS_STAT_SIZE : 0;
  p0 = p1 = lts_offset; p2 = (nrecs-1)*recsize + lts_offset;
  lseek(hFile,p0,SEEK_SET);
  if (!useMinimalStorage)
  {
    if (read(hFile,(char *)rb,recordPrefix) < 0) ccerr(file_error);
    memcpy(&dt0,rb,sizeof(double));
    t0 = (SINT32)dt0;
  }
  else
  {
    if (read(hFile,(char *)&t0,recordPrefix) < 0) ccerr(file_error);
    dt0 = (double)t0;
  }
  if (dt < dt0) ccerr(0);
  lseek(hFile,p2,SEEK_SET); 
  if (!useMinimalStorage)
  {
    if (read(hFile,(char *)rb,recordPrefix) < 0) ccerr(file_error);
    memcpy(&dt2,rb,sizeof(double));
    t2 = (SINT32)dt2;
  }
  else
  {
    if (read(hFile,(char *)&t2,recordPrefix) < 0) ccerr(file_error);
    dt2 = (double)t2;
  }
  if (isinf(dt0) || isnan(dt0)) ccerr(file_error);
  if (isinf(dt2) || isnan(dt2)) ccerr(file_error);
  while (p2 > p0 + recsize)
  {
    if (dt > (dt1=dt2)) { p1 = p2; break; }
    while (dt1 > dt0)   /* seek <--  */
    {
      f = ((dt-dt0)/(dt1-dt0))*((double)(p2-p0)/recsize);
      if ((p1=p0+recsize*(long)f) >= p2) break;
	    lseek(hFile,p1,SEEK_SET);
      if (!useMinimalStorage)
      {
        if (read(hFile,(char *)rb,recordPrefix) < 0) ccerr(file_error);
        memcpy(&dt1,rb,sizeof(double));
        t1 = (SINT32)dt1;
      }
      else
      {
        if (read(hFile,(char *)&t1,recordPrefix) < 0) ccerr(file_error);
        dt1 = (double)t1;
      }
      if (isinf(dt1) || isnan(dt1)) ccerr(file_error);
      if (dt1 < dt) break;
      p2 = p1;
      t2 = t1;
      dt2 = dt1;
      it++;
    }
    dt0 = dt1; t0 = t1; p0 = p1;
    while (dt2 > dt1)  /* seek --> */
    {
      f = ((dt2-dt)/(dt2-dt1))*((double)(p2-p0)/recsize);
      if ((p1=p2-recsize*(long)f) <= p0) break;
	    lseek(hFile,p1,SEEK_SET); 
      if (!useMinimalStorage)
      {
        if (read(hFile,(char *)rb,recordPrefix) < 0) ccerr(file_error);
        memcpy(&dt1,rb,sizeof(double));
        t1 = (SINT32)dt1;
      }
      else
      {
        if (read(hFile,(char *)&t1,recordPrefix) < 0) ccerr(file_error);
        dt1 = (double)t1;
      }
      if (isinf(dt1) || isnan(dt1)) ccerr(file_error);
      if (dt1 > dt) break;
      p0 = p1; 
      t0 = t1;
      dt0 = dt1;
      it++;
    }
    p2 = p1; t2 = t1; dt2 = dt1;
  }
err: 
  close(hFile);
  if (cc == file_error) feclogEx(TINE_LOGLEVEL_ERROR, "HIST: error reading file %.256s",fileName);
  return cc == 0 ? p1 : -cc;
# else
  return 0;
# endif /* NO_LTS */
}
static ClnHdr hstcln;
static ClnLst hstcon;
static ContractListStruct hstcl;
#define PRE_DUMP_WINDOW 60
#define PRE_UPDATE_WINDOW 30
int gHstIsCycling = FALSE;
void historyRecordInit(HstTblEntry *hst)
{
  int keepPoi=FALSE;
  ExportPropertyListStruct *prp;
  char* dev = NULL;
  if (!strncmp(hst->c.EqmName, DOOCSEQM_TAG, EQM_NAME_SHORTSIZE)) dev = hst->c.EqmDeviceName;
  prp = GetPropertyListStruct(hst->c.EqmName,hst->c.EqmProperty,dev);
  if (hst->c.EqmSizeOut == 1) keepPoi = TRUE;
  if (prp != NULL) 
  {
    hst->range = prp->egu.max - prp->egu.min;
    if (prp->prpArrayType & AT_CHANNEL) 
    { /* an MCA record ... */
      char memnam[PROPERTY_NAME_SIZE+8];
      keepPoi = TRUE;
      sprintf(memnam,"%.56s.MEMBERS",hst->c.EqmProperty);
      if ((prp=GetPropertyListStruct(hst->c.EqmName,memnam,dev)) != NULL)
      { /* explicit member list not given directly by registered devices ! */
        int cc, flen = fmtsizeof(prp->prpFormat);
        DTYPE dout, din;
        DTYPEZERO(dout); DTYPEZERO(din);
        hstcl.exp = getExportListItem(hst->c.EqmName);
        hstcl.prp = prp;
        dout.dArrayLength = hst->c.EqmSizeOut;
        dout.dFormat = LFMT(prp->prpFormat);
        dout.data.vptr = SystemCalloc(hst->c.EqmSizeOut,flen);
        cc = callEqm(&hstcl, hst->c.EqmDeviceName, memnam, &dout, &din, CA_READ|CA_HIST);
        if (HAS_DATA(cc))
        { /* call to <P>.MEMBERS was successful */
          hst->memidx = (UINT16 *)SystemCalloc(hst->c.EqmSizeOut,sizeof(UINT16));
          if (hst->memidx != NULL)
          { /* now make the cross-reference list */
            int i, foff = LFMT(prp->prpFormat) == CF_USTRING ? 16 : 0;
            char *s;
            for (i=0; i<(int)hst->c.EqmSizeOut; i++)
            {
              s = dout.data.cptr + i * flen + foff;
              hst->memidx[i] = (UINT16)GetDeviceNumber(hst->c.EqmName,s);
            }
            feclog("HIST: assigned  %.64s MCA members index cross-reference (%d entries)", hst->c.EqmProperty, hst->c.EqmSizeOut);
          }
        }
        else
        {
          feclog("HIST: get  %.64s MCA members failed : %.32s", hst->c.EqmProperty, cc2str(cc));
        }
        SystemFree(dout.data.vptr);
      }
    }
  }
  if (excludePointsOfInterest || hst->keepPointsOfInterest == USE_POI_NONE)
  { /* explicit request to turn it off */
    hst->keepPointsOfInterest = 0; 
  }
  else if (hst->range <= 0)
  { /* nothing was registered -> turn it off! */
    hst->keepPointsOfInterest = 0; 
  }
  else if (keepPoi)
  {
    if (hst->aTolerance > hst->range * 0.001 ||
        hst->pTolerance > 0)
    {
      hst->keepPointsOfInterest = USE_POI_TOLERANCE;
    }
    else
    {
      hst->keepPointsOfInterest = USE_POI_RANGE;
    }
  }
  hst->recordInitialized = TRUE;
  return;
}
#ifndef SMALL_TINE_LIB
static float gPoiRangeFactor = (float)POI_RANGE_FACTOR;
static float gPoiToleranceFactor = (float)POI_TOLERANCE_FACTOR;
# define checkIsPoi(CF) \
  for (p=0; !poi && p<(int)hst->c.EqmSizeOut; p+=inc) {\
    t = useRange ? gPoiRangeFactor * hst->range :\
      gPoiToleranceFactor*(hst->aTolerance + hst->pTolerance*fabs((float)((CF *)d1)[p]));\
    if (fabs((float)((CF *)d0)[p] - (float)((CF *)d1)[p]) > t) poi = TRUE;\
  }
TINE_EXPORT void SetPointsOfInterestRangeFactor(float value)
{
  if (value < 0.10) value = (float)0.10;
  if (value > 100.0) value = (float)100.0;
  gPoiRangeFactor = value;
}
TINE_EXPORT void SetPointsOfInterestToleranceFactor(float value)
{
  if (value < 2.0) value = (float)2.0;
  gPoiToleranceFactor = value;
}
int isPointOfInterest(HstTblEntry *hst)
{
  BYTE *d0,*d1;
  int p,ptr,inc=1,poi=FALSE,useRange;
  double t;
  if (!hst->keepPointsOfInterest) return FALSE;
  useRange = hst->keepPointsOfInterest == USE_POI_RANGE ? TRUE : FALSE;
  ptr = hst->dataPtr;
  d0 = hst->data + ptr * hst->recordLength;
  if (ptr == 1 && hst->timestamp[hst->depthShort-1] == 0)
  { /* could be just starting up */
    return FALSE;
  }
  p = ptr == 0 ? hst->depthShort - 1 : ptr - 1;
  d1 = hst->data + p * hst->recordLength;
  if (useRange && hst->range <= 0)
  {
    useRange = FALSE;
    if ((hst->aTolerance + hst->pTolerance) == 0) return FALSE;
  }

  switch (LFMT(hst->c.EqmFormatOut))
  {
    case CF_DBLTIME:
    case CF_DBLDBL:
         inc = 2;
    case CF_DOUBLE:
         checkIsPoi(double);
         break;
    case CF_FLTFLT:
    case CF_FLTINT:
         inc = 2;
    case CF_FLOAT:
         checkIsPoi(float);
         break;
    case CF_INT16:
         checkIsPoi(SINT16);
         break;
    case CF_UINT16:
         checkIsPoi(UINT16);
         break;
    case CF_INTINT:
         inc = 2;
    case CF_INT32:
         checkIsPoi(SINT32);
         break;
    case CF_UINT32:
         checkIsPoi(SINT32);
         break;
    case CF_BYTE:
         checkIsPoi(BYTE);
         break;
    default:
       break;
  }
  return poi ? TRUE : FALSE;
}
#endif /* !SMALL_TINE_LIB */
#define checkNeedArchive(CF) \
  for (p=0; !doarchive && p<(int)hst->c.EqmSizeOut; p+=inc) {\
    t = hst->aTolerance + hst->pTolerance * fabs((float)((CF *)d1)[p]);\
    if (fabs((float)((CF *)d0)[p] - (float)((CF *)d1)[p]) > t) doarchive = LHS_APPLY_ARCHIVE_TRANSITION;\
  }

int needToArchiveRecord(int i)
{
  double t, ts = makeDataTimeStamp();
  BYTE *d0,*d1;
  HstTblEntry *hst;
  int p,ptr,inc=1,dlen,haslts,doarchive = LHS_APPLY_ARCHIVE_NONE;
  UINT32 delta_t;

  if (ServerExitCondition) return LHS_APPLY_ARCHIVE_NONE;
  if (i < 0 || i >= nHistoryRecords) return LHS_APPLY_ARCHIVE_NONE;
  hst = &hstTbl[i];
  haslts = (HSTDEPLNG(hst) > 0) || (HSTDEPDAYS(hst) > 0);
  if (hst->archiveRate < 0 || hst->heartbeat < 0 || !haslts) return LHS_APPLY_ARCHIVE_NONE; /* no long term storage */
  ptr = hst->dataPtr;

  if (hst->timestamp[ptr] <= hst->lastArchiveTime) return LHS_APPLY_ARCHIVE_NONE;/* data timestamp hasn't changed */
  if (useMinimalStorage && (SINT32)hst->timestamp[ptr] == (SINT32)hst->lastArchiveTime) return LHS_APPLY_ARCHIVE_NONE;
  if (hst->lastArchiveTime == 0) hst->lastArchiveTime = ts - hst->archiveRate/1000.0;
  delta_t = (UINT32)((ts - hst->lastArchiveTime)*1000);
  if (hst->isScheduled && !hst->storeSummary) delta_t = (UINT32)hst->archiveRate * 2;
  if (delta_t < (UINT32)hst->archiveRate) return LHS_APPLY_ARCHIVE_NONE;
  if (delta_t >= (UINT32)hst->heartbeat*1000) return LHS_APPLY_ARCHIVE_THIS; /* archive no matter what */
  if (hst->storeSummary) return LHS_APPLY_ARCHIVE_THIS; /* if summary: at this point return TRUE */
  d0 = hst->data + ptr * hst->recordLength;
  d1 = hst->data + hst->depthShort * hst->recordLength;

  switch (LFMT(hst->c.EqmFormatOut))
  {
    case CF_DBLTIME:
    case CF_DBLDBL:
      inc = 2;
    case CF_DOUBLE:
      checkNeedArchive(double);
      break;
    case CF_FLTFLT:
    case CF_FLTINT:
      inc = 2;
    case CF_FLOAT:
      checkNeedArchive(float);
      break;
    case CF_INT16:
      checkNeedArchive(SINT16);
      break;
    case CF_UINT16:
      checkNeedArchive(UINT16);
      break;
    case CF_INTINT:
      inc = 2;
    case CF_INT32:
      checkNeedArchive(SINT32);
      break;
    case CF_UINT32:
      checkNeedArchive(UINT32);
      break;
    case CF_BYTE:
      checkNeedArchive(BYTE);
      break;
    default:
      dlen = hst->recordLength;
      if (hst->hasDataPool)
      { /* complex data type */
        d0 = ((DataPoolEntry *)hst->data)[ptr].data;
        d1 = ((DataPoolEntry *)hst->data)[hst->depthShort].data;
        dlen = ((DataPoolEntry *)hst->data)[ptr].capacity;
      }
      if (!doarchive && memcmp(d0,d1,dlen)) doarchive = LHS_APPLY_ARCHIVE_TRANSITION;
      break;
  } 
  if (doarchive == LHS_APPLY_ARCHIVE_TRANSITION && hst->hasDataPool) doarchive = LHS_APPLY_ARCHIVE_THIS;
  return doarchive;
}
int gAllowOverloadedConcurrentLocalHistories = FALSE;
TINE_EXPORT void SetAllowOverloadedConcurrentLocalHistories(int value)
{
  gAllowOverloadedConcurrentLocalHistories = value;
}
TINE_EXPORT int GetAllowOverloadedConcurrentLocalHistories(void)
{
  return gAllowOverloadedConcurrentLocalHistories;
}
int isPrpInHistoryTable(char *prp)
{
  int i;
  if (hstTbl == NULL || prp == NULL) return FALSE;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (!strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) return TRUE;
  }
  return FALSE;
}
#ifndef SMALL_TINE_LIB
int makeFilesForRecord(int index)
{
  int i, cc, dd, dstart, dstop, depth, depthDays;
  HstTblEntry *hst;
  char scratch[64];
  time_t t = time(NULL);
  struct tm *tm = localtime(&t);
  if (ServerExitCondition) return not_running;
  if (index < 0 || index >= nHistoryRecords) return invalid_index;
  hst = &hstTbl[index];
  depth = HSTDEPLNG(hst); depthDays = HSTDEPDAYS(hst);
  if (depth < 0) return invalid_data;
  gServerCycleState = CYCLE_STATE_HST_MKFILES;
  _sendToDbg = 1;
  /* make sure the standard (this month) file is prepared first */
  nOutdatedFiles = scanForOutdatedFiles(index,depth,depthDays);
  if (nOutdatedFiles < 0) sprintf(scratch,"%.32s",cc2str(-nOutdatedFiles));
  else sprintf(scratch,"found %d recyclable files",nOutdatedFiles);
  feclog("HIST: check standard files for rec %d (0x%x) -> (%.6s)/%.64s[%.64s] : %.32s",
        hst->recordIndex,hst->recordIndex,
        hst->c.EqmName,hst->c.EqmDeviceName,hst->c.EqmProperty,scratch);
  depth++;
  dstart = 1; 
  dstop = useMonthlyHistoryFiles ? 1 : 31;
  if (depthDays > 0) 
  { /* different range: just keep so many days ... */
    dstart = tm->tm_mday - depthDays;
    dstop = tm->tm_mday;
    if (dstart <= 0)
    { /* start as near to the beginning as possible */
      if (nStdDailyFiles > 0) dstop += (depthDays - nStdDailyFiles); else dstop -= --dstart;
      if (dstop < tm->tm_mday) dstop = tm->tm_mday;
      dstart = 1;
    }
  }
  for (i=0; i<depth; i++) 
  {
    for (dd=dstart; dd<dstop+1; dd++)
    {
      if ((cc=makeStandardFiles(index,i,dd)) != 0)
      {
        feclog("HIST: make local history files : %s",cc2str(cc));
      }
    }
  }
  _sendToDbg = 0;
  return 0;
}
#endif /* !SMALL_TINE_LIB */
int isAllowedHistoryFormat(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_DEFAULT:
    case CF_NULL:
    case CF_HISTORY:
      return FALSE;
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_IMAGE: 
      return TRUE;
    case CF_ASPECTRUM:
    case CF_AIMAGE: /* n.b. later to be allowed !*/
      return FALSE; 
    default:
      break;
  }
  return TRUE;
}

_CSVCOLUMN csvHistoryInfoColumns[] = 
{
  { "INDEX","",0,CF_INT32,1,(BYTE *)OFFSETIN(HRIS,recordIndex),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,c.EqmName),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,c.EqmProperty),NULL,0 },
  { "DEVICE","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,c.EqmDeviceName),NULL,0 },
  { "DATA_LENGTH","1",0,CF_INT32,1,(BYTE *)OFFSETIN(HRIS,c.EqmSizeOut),NULL,0 },
  { "DATA_CAPACITY","0",0,CF_INT32,1,(BYTE *)OFFSETIN(HRIS,poolEntryCapacity),NULL,0 },
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(HRIS,fmt),NULL,0 },
  { "HEARTBEAT","900",0,CF_INT32,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(HRIS,heartbeat),NULL,0 },
  { "POLLING_RATE","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(HRIS,pollingRate),NULL,0 },
  { "ARCHIVE_RATE","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(HRIS,archiveRate),NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(HRIS,tolerance),NULL,0 },
  { "SHORT_DEPTH","0",0,CF_INT32,1,(BYTE *)OFFSETIN(HRIS,depthShort),NULL,0 },
  { "LONG_DEPTH","0",0,CF_TEXT,64,(BYTE *)OFFSETIN(HRIS,depthLong),NULL,0 },
  { "EXPORT_NAME","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,exportTarget),NULL,0 },
  { "REMOTE_SERVER","",0,CF_TEXT,96,(BYTE *)OFFSETIN(HRIS,redirection),NULL,0 },
# ifndef SMALL_TINE_LIB
  { "FILTER","",0,CF_TEXT,256,(BYTE *)OFFSETIN(HRIS,filterString),NULL, 0 },
# endif
};
CSVDB csvHistoryInfoDb =
{
  csvHistoryInfoColumns,
  sizeof(csvHistoryInfoColumns)/sizeof(csvHistoryInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HRIS),
  0
};
void parseLongDepth(char *depthStr,int *mm, int *dd)
{ /* internal: parameters never NULL */
  int m=0, d=0;
  char *c=NULL;
  if (depthStr != NULL && strlen(depthStr) > 0)
  {
    m = strtol(depthStr,&c,10);
    if (c != NULL)
    {
      while (*c == ' ') c++;
      if (!strnicmp(c,"day",3))
      { 
        d = m;
        m = 0;
      }
      else if ((c=strchr(c,'.')) != NULL)
      {
        c++; d = strtol(c,&c,10);
        m += d/31;
        d = d%31;
      }
    }
  }
  *mm = m; *dd = d;
  return;
}
int LoadHistoryInformation(char *fname,char *expName,char *eqm) /* read history record info from file */
{
  HRIS *hrs = NULL;
  HistorySpecification hspec;
  int i, n, sz, fmt, mm, dd, cc = 0, hstFlags = 0;
  char *c, *s1, *s2, hfn[128], scratch[64];
  
  if (gTineConfigurationCoded) ccerr(not_implemented);
  /* read in the details */
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  {
    feclog("HIST: looking in fec.xml for local history configurations ...");
    sz = populateHRS(gFecName,gFecInfoList,&hrs);
    if (sz < 0) ccerr(-sz);
  }
  else
# endif
  {
#   ifdef FS_DELIMITER
    if (eqm != NULL)
    {
      sprintf(hfn,"%.6s%c%s",eqm,FS_DELIMITER,fname);
      feclog("HIST: looking in %.128s%.32s for local history configurations ...",fecDbPath,hfn);
      if ((cc=csvReadFile(fecDbPath,hfn,&csvHistoryInfoDb,(void **)&hrs)) == 0) goto file_read;
    }
#   endif
    if (ignoreCommonDbFiles) ccerr(no_such_file);
    feclog("HIST: looking in %.128s%.32s for local history configurations ...",fecDbPath,fname);
    if ((cc=csvReadFile(fecDbPath,fname,&csvHistoryInfoDb,(void **)&hrs)) != 0) ccerr(cc);
file_read:
    sz = csvHistoryInfoDb.siz;
  }
  /* make adjustments to local table size */
  if (sz > maxHistoryRecords/2 && nHistoryRecords == 0) maxHistoryRecords = sz + addHistoryCapacity;
  /* allocate space for local table (allow extra space for new entries) */
  if (hstTbl == NULL && (hstTbl=(HstTblEntry *)SystemCalloc(maxHistoryRecords,sizeof(HstTblEntry))) == NULL)
    ccerr(out_of_local_memory);
  for (i=0,n=nHistoryRecords; i<sz && n<maxHistoryRecords; i++)
  { /* prepare local history record table */
    if (eqm != NULL && strncmp(hrs[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strlen(hrs[i].exportTarget) > 0 && expName != NULL)
    { /* legacy configuration trying to latch a registered export name */
      if (strnicmp(hrs[i].exportTarget,expName,EXPORT_NAME_SIZE)) continue; /* didn't match ! */
    }
    memset(&hspec,0,sizeof(HistorySpecification));
    if ((c=strstr(hrs[i].tolerance,"NOPOI")) != NULL)
    {
      *c = 0;
      hstFlags = USE_POI_NONE;
    }
    strncpy(scratch,hrs[i].tolerance,64); scratch[63] = 0;
    s1 = scratch;
    if ((s2=strchr(s1,'+')) != NULL) *s2++ = 0;
    if (strchr(s1, '%'))
    {
      hspec.pTolerance = (float)(atof(s1) / 100.0);
      if (s2 != NULL) hspec.aTolerance = (float)atof(s2);
    }
    else
    {
      hspec.aTolerance = (float)atof(s1);
      if (s2 != NULL && strchr(s2,'%')) hspec.pTolerance = (float)(atof(s2) / 100.0);
    }
    fmt = CF_NULL;
    if (strlen(hrs[i].fmt) > 0) 
    {
      fmt = ftoi(hrs[i].fmt);
      hrs[i].c.EqmFormatOut = BFMT(fmt);
      if (fmt == CF_STRUCT)
      {
        if ((c=strchr(hrs[i].fmt,'.')) == NULL)
        {
          feclogEx(TINE_LOGLEVEL_ERROR, "HIST: structure format given without tag!");
          continue;
        }
        strncpy(hrs[i].c.strTagOut,++c,TAG_NAME_SIZE);
      }
    }
    if (!isAllowedHistoryFormat(fmt)) 
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] : illegal format",
        hrs[i].recordIndex,hrs[i].recordIndex,
        hrs[i].c.EqmName,hrs[i].c.EqmDeviceName,hrs[i].c.EqmProperty);
      continue;
    }
    if (strlen(hrs[i].redirection) > 0) hspec.rhsServerName = hrs[i].redirection;
    hspec.pollingRate = hrs[i].pollingRate;
    hspec.archiveRate = hrs[i].archiveRate;
    hspec.depthShort = hrs[i].depthShort;
    parseLongDepth(hrs[i].depthLong,&mm,&dd);
    hspec.depthLong = MAKELONGDEPTH(mm,dd); 
    hspec.heartbeat = hrs[i].heartbeat;
    cc = addHistoryItem(n,hrs[i].c.EqmName,hrs[i].c.EqmProperty,hrs[i].c.EqmDeviceName,hrs[i].c.strTagOut,
            hrs[i].c.EqmSizeOut,LFMT(hrs[i].c.EqmFormatOut),hrs[i].recordIndex,&hspec,hstFlags);
    switch (cc)
    {
      case 0: break;
      case out_of_local_memory: ccerr(cc);
      default: continue;
    }
    nHistoryRecords++;
#   ifndef SMALL_TINE_LIB  
    if (strlen(hrs[i].filterString) > 0) 
    {
      if ((cc=ApplyHistoryFilter(hrs[i].recordIndex,hrs[i].filterString)) != 0)
      {
        feclogEx(TINE_LOGLEVEL_ERROR, "HIST: could not apply filter %.256s to record %d : %.32s",
          hrs[i].filterString,hrs[i].recordIndex,cc2str(cc));
      }
    }
#   endif
    n++; 
  }
  cc = 0; /* re-zero this for final log entry */
# ifndef SMALL_TINE_LIB  
  if (useStandardFileSet)
  { /* reading history file : true for one, true for all */
    feclog("HIST: using standard non-fragmented history files at initialization");
    for (i=0; i<nHistoryRecords; i++)
    {
      makeFilesForRecord(i);
    }
  }
# endif
  if (MaxPollingRate > minHistoryPollingRate)
  {
    if (gSystemCycleDeadband > (UINT32)minHistoryPollingRate) 
    {
      feclog("adjust cycle deadband (due to local history subsystem) from %d to %d msec",
        gSystemCycleDeadband,minHistoryPollingRate);
      gSystemCycleDeadband = (UINT32)minHistoryPollingRate;
    }
    MaxPollingRate = minHistoryPollingRate;
  }
err:
  SystemFree(hrs);
  if (cc && cc != no_such_file)
  {
    feclog("HIST: %s : %.32s",fname,cc2str(cc));
  }
  return cc;
}
void fixHistoryStaticRepository(void)
{
  static int done = FALSE;
  if (done) return;
# ifdef FS_DELIMITER
  if (strlen(arcDbPathSaved) == 0)
  { /* not set from API: use default */
    sprintf(arcDbPathSaved,"%.120sSAVED%c",arcDbPath,FS_DELIMITER);
  }
  feclog("HISTORY static repository : %s",arcDbPathSaved);
  #if !defined(SMALL_TINE_LIB) && !defined(VXWORKS)
  {
    int i;
    for (i = 0; i < nHistoryRecords; i++)
    {
      hstTbl[i].hasSavedFiles = checkForSavedFiles(hstTbl[i].recordIndex);
    }
  }
# endif
# endif
  done = TRUE;
}

void moveFiles(char *src,char *dst)
{
  char srcfn[264], dstfn[264];
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char dirname[132];
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif
  if (src == NULL || strlen(src) == 0) return;
  if (dst == NULL || strlen(dst) == 0) return;
  if (strcmp(src,dst) == 0) return;
# if defined(UNIX)
  if ((dp=opendir(dst)) == NULL)
  {
    makeDirectoryPath(dst,TRUE);
  }
  else
  {
    closedir(dp);
  }
  if ((dp=opendir((char *)src)) == NULL) return;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (!strcmp(dirp->d_name,".") || !strcmp(dirp->d_name,"..")) continue;
    snprintf(srcfn,256,"%.200s%c%s",src,FS_DELIMITER,dirp->d_name);
    snprintf(dstfn,256,"%.200s%c%s",dst,FS_DELIMITER,dirp->d_name);
    rename(srcfn,dstfn);
  }
  closedir(dp);
  remove(src);
# elif defined(WIN32)
  if (strlen(dst) == 0) dst = ".";
  snprintf(dirname,128,"%s\\*.*",dst);
  if ((hfp=_findfirst(dirname, &fd)) == -1)
  {
    makeDirectoryPath(dst,TRUE);
  }
  else
  {
    _findclose(hfp);
  }
  if (strlen(src) == 0) dst = ".";
  snprintf(dirname,128,"%s\\*.*",src);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return;
  do
  {
    if (!strcmp(fd.name,".") || !strcmp(fd.name,"..")) continue;
    snprintf(srcfn,256,"%.200s%c%s",src,FS_DELIMITER,fd.name);
    snprintf(dstfn,256,"%.200s%c%s",dst,FS_DELIMITER,fd.name);
    rename(srcfn,dstfn);
  } while (_findnext(hfp,&fd) != -1);
  _findclose(hfp);
  remove(src);
# else
  srcfn[0] = dstfn[0] = 0;
# endif
  return;
}
void fixHistoryRepository(void)
{
  static int done = FALSE;
  if (done) return;
# ifdef FS_DELIMITER
  if (strlen(arcDbPath) == 0)
  { /* could have been set from fec.xml or fecid.csv */
    char *ptr, *c, hsthome[256], lhsthome[256];
    memset(hsthome,0,128);
    if ((ptr=getenv("TINE_HISTORY_HOME")) == NULL && 
        (ptr=getenv("HISTORY_HOME")) == NULL && 
        (ptr=getenv("HISTORYDB")) == NULL) 
    { /* not set from environment -> use FEC_HOME */
      char fpath[256];
      if (strlen(fecDbPath) > 0) 
      {
        strncpy(fpath,fecDbPath,224);
        fpath[224] = 0;
      }
      else
      {
        sprintf(fpath,".%c",FS_DELIMITER);
      }  
      sprintf(hsthome,"%.128sHISTORY",fpath);
      sprintf(lhsthome,"%.128s..%cHISTORY",fecDbPath,FS_DELIMITER);
      if (hasFilesOfType(lhsthome,"*.*") && !hasFilesOfType(hsthome,"*.*"))
      { /* legacy default settings found and no new default directory... */
        moveFiles(lhsthome,hsthome);
      }
      ptr = hsthome;
    }
    if (*ptr == '\"') ptr++; 
    if ((c=strchr(ptr,'\"')) != NULL) *c = 0;
    strncpy(arcDbPath,ptr,120);
  }
  validateHistoryHomePath();
# endif
  done = TRUE;
}
void historyInit(char *expName,char *eqmName)
{
  ExportListStruct *el;
# if !defined(SMALL_TINE_LIB) && !defined(VXWORKS)
  char *ptr;
# endif
  if (ServerExitCondition) return;
  if (expName == NULL && eqmName == NULL) return;
  if ((el=getExportListItem(eqmName)) != NULL && el->hstdone) return;
  gServerCycleState = CYCLE_STATE_HST_INIT;
  feclog("HIST: initializing local history subsystem for module %s",eqmName == NULL ? "(null)" : eqmName);
  if (!history_home_done)
  {
    fixHistoryRepository();
    history_home_done = TRUE;
  }
  feclog("HISTORY HOME : %s",arcDbPath);
# if !defined(SMALL_TINE_LIB) && !defined(VXWORKS)
  fsBlocksFree = getAvailableDiskSpaceInBlocks(arcDbPath);
  feclog("HIST: file system has %lu free blocks",fsBlocksFree);
  if ((ptr=getenv("TINE_HISTORY_FREE_BLOCKS")) != NULL ||
      (ptr=getenv("HISTORY_FREE_BLOCKS")) != NULL) 
  { /* set from environment */
    minFreeDiskSpaceInBlocks = (unsigned long)atoi(ptr);
    feclog("HIST: set minimum free disk free space to %lu from environment",minFreeDiskSpaceInBlocks);
  }
  if (!useStandardFileSet)
  { /* not set from API; is there already a standard file set ? */
    useStandardFileSet = checkForStandardFiles();
  }
  if (!useStandardFileSet)
  { /* no file set seen; is there an environment variable ? */
    useStandardFileSet = checkForStandardFilesEnvironment();
  }
  if (!useStandardFileSet) feclog("HIST: not using standard non-fragmented history files");
  if ((ptr=getenv("TINE_HISTORY_EXTRA_LOGGING")) != NULL) hstlog = atoi(ptr);
  if ((ptr=getenv("TINE_HISTORY_NOPOI")) != NULL) 
  {
    if (strtobool(ptr)) 
    {
      excludePointsOfInterest = TRUE;
      feclog("HIST: Points of interest file will not be kept according to environment");
    }
  }
  if ((ptr=getenv("TINE_HISTORY_DEACTIVATE_ON_ERROR")) != NULL) 
  {
    if (strtobool(ptr))
    {
      gHstDeactivateOnError = TRUE;
      feclog("HIST: access error will deactivate record according to environment");
    }
  }
  if ((ptr=getenv("TINE_HISTORY_LATE_ADDITION_WINDOW")) != NULL) 
  {
    gLclHstLateAdditionWindow = atoi(ptr);
    feclog("HIST: set late addition window to %d seconds from environment",gLclHstLateAdditionWindow);
  }
  if ((ptr=getenv("TINE_HISTORY_AUTOREMOVE_UNSYNCHRONIZED")) != NULL)
  {
    gAutoRemoveUnsynchronizedMcaHistories = strtobool(ptr);
    feclog("HIST: set AutoRemoveUnsynchronizedHistories to %s from environment", 
      (gAutoRemoveUnsynchronizedMcaHistories ? "TRUE" : "FALSE"));
  }
# endif
  LoadHistoryInformation(HISTORY_FILENAME,expName,eqmName);
  if (nHistoryRecords > 0) makeDirectoryPath(arcDbPath,TRUE);
  fixHistoryStaticRepository();
  if (el != NULL) el->hstdone = TRUE;
}
void freeHstMemory(void)
{
  int i, k;
  if (hstTbl != NULL)
  {
    feclog("freeing local history table (%d records)",nHistoryRecords);
    for (i=0; i<nHistoryRecords; i++)
    {
      SystemFree(hstTbl[i].prd);
      SystemFree(hstTbl[i].timestamp);
      SystemFree(hstTbl[i].sysstamp);
      SystemFree(hstTbl[i].usrstamp);
      SystemFree(hstTbl[i].memidx);
      if (hstTbl[i].data != NULL) 
      {
        if (hstTbl[i].hasDataPool) 
        { /* complex data type */
          DataPoolEntry *dpe = (DataPoolEntry *)hstTbl[i].data;
          for (k=0; k<hstTbl[i].depthShort+1; k++)
            if (dpe[k].data != NULL) SystemFree(dpe[k].data);
        }
        SystemFree(hstTbl[i].data);
      }
    }
    SystemFree(hstTbl);
  }
  nHistoryRecords = 0;
  hstTbl = NULL;
  history_home_done = FALSE;
# ifndef SMALL_TINE_LIB
  SystemFree(outdatedFiles); /* return memory to the heap */  
  nOutdatedFiles = 0;
# endif
}
#ifndef SMALL_TINE_LIB
char *getStrToleranceForManifest(float aTol, float pTol)
{
  static char s[256] = {0};
  int len = 0;
  s[0] = 0;
  if (aTol > 0) sprintf(s, "%g", aTol);
  if (pTol > 0)
  {
    if ((len=(int)strlen(s)) > 0)
    {
      strcat(s, " + ");
      len += 3;
    }
    sprintf(&s[len], "%g%c", (pTol * 100.0), '%');
  }
  if (strlen(s) == 0) strcpy(s, "0");
  return s;
}
typedef struct 
{
  char fec[FEC_NAME_SIZE];
  char eqm[EQM_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char fmtStr[16];
  char depth[16];
  int port;
  int idx;
  int stdf;
  int siz;
  int fmt;
  int end;
} HSTMF;
#define hsthomemfhdr "FecName, LocalName, Port, Index, StdFiles, Property, Device, DataLength, Format, LongDepth, BigEndian"
_CSVCOLUMN csvHstMfColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(HSTMF,fec),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(HSTMF,eqm),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(HSTMF,prp),NULL,0 },
  { "DEVICE","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(HSTMF,dev),NULL,0 },
  { "PORT","",0,CF_INT32,1,(BYTE *)OFFSETIN(HSTMF,port),NULL,0 },
  { "INDEX","",0,CF_INT32,1,(BYTE *)OFFSETIN(HSTMF,idx),NULL,0 },
  { "STD_FILES","",0,CF_INT32,1,(BYTE *)OFFSETIN(HSTMF,stdf),NULL,0 },
  { "DATA_LENGTH","",0,CF_INT32,1,(BYTE *)OFFSETIN(HSTMF,siz),NULL,0 },
  { "FORMAT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(HSTMF,fmtStr),NULL,0 },
  { "LONG_DEPTH","",0,CF_TEXT,16,(BYTE *)OFFSETIN(HSTMF,depth),NULL,0 },
  { "BIG_ENDIAN","",0,CF_BOOLEAN,1,(BYTE *)OFFSETIN(HSTMF,end),NULL,0 },
};
CSVDB csvHstMfDb =
{
  csvHstMfColumns,
  sizeof(csvHstMfColumns)/sizeof(csvHstMfColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HSTMF),
  0
};
int gHstHomeMfUpdated = FALSE;
int updateHstHomeManifest(void)
{
  int cc=0;
# if defined(FS_DELIMITER) && !defined(FS_RDONLY)
  ExportListStruct *el;
  int i,r,depth,depthDays,n=0;
  int oldmask;
  struct stat sbuf;
  HSTMF *hstmf = NULL;
  FILE *fp=NULL;
  char tmpname[256], fname[256], depthStr[16];

  if (ServerExitCondition) return 0;
  if (nHistoryRecords == 0) return 0;
  oldmask = umask(0);
  sprintf(tmpname,"%.128s%s.tmp",arcDbPath,HISTORYHOME_MANIFEST);
  if (stat(tmpname,&sbuf) == 0)
  { /* the temp file exists! (other processes updating it?) */
    if (sbuf.st_mtime > (time(NULL) - 300))
    { /* file more recent than the last 5 minutes -> some other client app is updating */
      ccerr(operation_busy);
    }
    /* else: must be something left over from a crash */
  }
  if ((fp=fopen(tmpname,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"%s\n",hsthomemfhdr);
  if ((cc=csvReadFile(arcDbPath,HISTORYHOME_MANIFEST,&csvHstMfDb,(void **)&hstmf)) == 0)
  { 
    n = csvHstMfDb.siz;
  }
  feclog("update history home manifest");
  for (i=0; i<n; i++)
  { /* scan thru existing manifest entries */
    for (el=ExportList; el != NULL; el=el->next)
    { /* find entries from other FECs ... */
      if (hstmf[i].port == gPortOffset && !strnicmp(hstmf[i].eqm,el->EqmName,EQM_NAME_SHORTSIZE))
        break;
    }
    if (el != NULL) continue;
    fprintf(fp,"%.16s, %.6s, %d, %d, %d, %.64s, %.64s, %d, %.16s, %.16s, %s\n",
      hstmf[i].fec,hstmf[i].eqm,
      hstmf[i].port,hstmf[i].idx,hstmf[i].stdf,
      hstmf[i].prp,hstmf[i].dev,
      hstmf[i].siz,hstmf[i].fmtStr,hstmf[i].depth,
      (hstmf[i].end ? "TRUE" : "FALSE"));
    if ((r=getHstTblEntryFromRecordIndex(hstmf[i].idx)) >= 0)
    { /* looks like a collision ! */
      feclogEx(TINE_LOGLEVEL_WARN, "HIST: WARNING -> record %d (%.6s/%.64s[%.64s]) appears to conflict with local history of %.6s/%.64s[%.64s] on FEC %.16s",
        r,hstTbl[r].c.EqmName,hstTbl[r].c.EqmDeviceName,hstTbl[r].c.EqmProperty,
        hstmf[i].eqm,hstmf[i].dev,hstmf[i].prp,hstmf[i].fec);
    }    
  }
  for (i=0; i<nHistoryRecords; i++)
  { /* scan thru this FEC's history records */
    depth = HSTDEPLNG(&hstTbl[i]); depthDays = HSTDEPDAYS(&hstTbl[i]);
    if (depth <= 0 && depthDays <= 0) continue;
    if (depthDays > 0) sprintf(depthStr,".%d",depthDays); else depthStr[0] = 0;
    fprintf(fp,"%.16s, %.6s, %d, %d, %d, %.64s, %.64s, %d, %.16s, %d%s, %s\n",
      gFecName,hstTbl[i].c.EqmName,gPortOffset,
      hstTbl[i].recordIndex,hstTbl[i].useStandardSet,
      hstTbl[i].c.EqmProperty,hstTbl[i].c.EqmDeviceName,
      hstTbl[i].c.EqmSizeOut,itof(hstTbl[i].c.EqmFormatOut),depth,depthStr,
      (gEndianness ? "TRUE" : "FALSE"));
  }
err:
  if (fp != NULL) 
  {
    fclose(fp);
    sprintf(fname,"%.128s%s",arcDbPath,HISTORYHOME_MANIFEST);
    unlink(fname);
    rename(tmpname,fname);
  }
  SystemFree(hstmf);
  oldmask = umask(oldmask);
# endif
  return cc;
}
TINE_EXPORT void SetAutoRemoveUnsynchronizedHistories(int value)
{
  gAutoRemoveUnsynchronizedMcaHistories = value;
}
TINE_EXPORT int GetAutoRemoveUnsynchronizedHistories(void)
{
  return gAutoRemoveUnsynchronizedMcaHistories;
}
void chkMcaIntegrity(HstTblEntry *hst)
{
  ExportListStruct *el = getExportListItem(hst->c.EqmName);
  ExportPropertyListStruct *pl = GetPropertyListStruct(hst->c.EqmName, hst->c.EqmProperty, hst->c.EqmDeviceName);
  int insync = TRUE;
  char msg[256] = "";
  if (!gAutoRemoveUnsynchronizedMcaHistories) return;
  if (el == NULL || pl == NULL) return;
  if (hst->c.EqmSizeOut > 1 && pl->prpArrayType & AT_CHANNEL)
  {
    int i, n = 0;
    DEVMF *devs = NULL;
    GetDevicesFromManifest(hst->c.EqmName, hst->c.EqmProperty, &devs, &n);
    if (pl->numDevices > 0 && pl->devNames != NULL)
    {
      for (i=0; i < pl->numDevices && i < n && insync; i++)
      {
        if (strnicmp(pl->devNames[i].name, devs[i].name, DEVICE_NAME_SIZE))
        {
          sprintf(msg, "current channel %.64s != manifest channel %.64s", devs[i].name, pl->devNames[i].name);
          insync = FALSE;
        }
      }
    }
    else
    {
      for (i=0; i < el->EqmNumDevices && i < n && insync; i++)
      {
        if (strnicmp(el->EqmDevLst[i].name, devs[i].name, DEVICE_NAME_SIZE))
        {
          sprintf(msg, "current channel %.64s != manifest channel %.64s", devs[i].name,el->EqmDevLst[i].name);
          insync = FALSE;
        }
      }
    }
    SystemFree(devs);
  }
  if (!insync)
  { /* start over ... */
    resetLTSFiles(hst);
    feclogEx(TINE_LOGLEVEL_WARN, "HIST: reset local history files for %.6s/%.64s[%.64s] : %.256s", 
      hst->c.EqmName,hst->c.EqmDeviceName,hst->c.EqmProperty,msg);
  }
}
int dumpHistoryManifest(char *eqm,int to_stdout)
{
  int cc=0;
# if defined(FS_DELIMITER) && !defined(FS_RDONLY)
# define hstmfhdr "Index, LocalName, Property, Device, DataLength, Format, Heartbeat, PollingRate, ArchiveRate, Tolerance, ShortDepth, LongDepth"
  int i;
  char daysStr[16], fmtStr[64];
  FILE *fp = NULL;

  for (i=0; i<nHistoryRecords; i++)
  {
    if (eqm != NULL && strlen(eqm) > 0 && strncmp(eqm,hstTbl[i].c.EqmName,EQM_NAME_SHORTSIZE) != 0) continue;
    if (fp == NULL)
    {
      if ((fp=getMfFile(eqm,"w",HISTORY_MANIFEST)) == NULL) ccerr(file_error);
      fprintf(fp, "%s\n", hstmfhdr);
    }
    if (hstTbl[i].depthDays > 0) sprintf(daysStr,".%d",hstTbl[i].depthDays); else daysStr[0] = 0;
    strncpy(fmtStr, itof(hstTbl[i].c.EqmFormatOut), 16); fmtStr[16] = 0;
    if (strlen(hstTbl[i].c.strTagOut) > 0)
    {
      strcat(fmtStr, ".");
      strncat(fmtStr, hstTbl[i].c.strTagOut, TAG_NAME_SIZE);
    }
    fprintf(fp,"%d, %.6s, %.64s, %.64s, %d, %.16s, %d, %d, %d, %s, %d, %d%s\n",
      hstTbl[i].recordIndex,hstTbl[i].c.EqmName,hstTbl[i].c.EqmProperty,hstTbl[i].c.EqmDeviceName,
      hstTbl[i].c.EqmSizeOut,fmtStr,hstTbl[i].heartbeat,
      hstTbl[i].pollingRate,hstTbl[i].archiveRate,
      getStrToleranceForManifest(hstTbl[i].aTolerance,hstTbl[i].pTolerance),
      hstTbl[i].depthShort,hstTbl[i].depthLong,daysStr);
  }
  if (to_stdout) dbglog("local history manifest written");
err:
  if (fp != NULL) fclose(fp);
# endif
  return cc;
}
int findRecordIndexInHstManifest(char *eqm,char *prp,char *dev,int fmt,int siz)
{
  FILE *fp=getMfFile(eqm,"r",HISTORY_MANIFEST);
  int idx=-1,cc=0,done=0,eqm_col=-1,prp_col=-1,dev_col=-1,fmt_col=-1,siz_col=-1,idx_col=-1;
  char str[256],hdr[256],scratch[64];
  if (eqm == NULL || prp == NULL || dev == NULL) return argument_list_error;
  if (fp == NULL) ccerr(no_such_file);
  for (done=0; csvGetLine(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((eqm_col=findcol(hdr,"LOCALNAME")) < 0) ccerr(no_such_column);
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_column);
      if ((dev_col=findcol(hdr,"DEVICE")) < 0) ccerr(no_such_column);
      if ((fmt_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_column);
      if ((siz_col=findcol(hdr,"DATALENGTH")) < 0) ccerr(no_such_column);
      if ((idx_col=findcol(hdr,"INDEX")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (strncmp(eqm,colptr(eqm_col,str),EQM_NAME_SHORTSIZE)) continue;
    if (strncmp(prp,colptr(prp_col,str),PROPERTY_NAME_SIZE)) continue;
    if (strncmp(dev,colptr(dev_col,str),DEVICE_NAME_SIZE)) continue;
    strncpy(scratch,colptr(fmt_col,str),64);
    if (ftoi(scratch) != fmt) continue;
    strncpy(scratch,colptr(siz_col,str),64);
    if (atoi(scratch) != siz) continue;
    strncpy(scratch,colptr(idx_col,str),64);
    idx = atoi(scratch);
    break;
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc) 
  {
    feclogEx(TINE_LOGLEVEL_WARN, "HIST: cannot find record index for %.6s/%.64s[%.64s] %d %.32s value(s) in manifest: %.32s",
      eqm,dev,prp,siz,itof(fmt),cc2str(cc));
  }
  return idx;
}
void dumpHstManifests(int to_stdout)
{
  ExportListStruct *el;

  feclog("dump history manifests");
  for (el=ExportList; el != NULL; el=el->next)
  {
    dumpHistoryManifest(el->EqmName,to_stdout);
  }
  return;
}
int gHstMfDumped = FALSE;
#endif /* SMALL_TINE_LIB */
TINE_EXPORT int SaveShortTermData(char *eqm, char *prp, char *dev, HstRngType rng)
{
  HstTblEntry htt;
  int useSysStamp = FALSE;
  int idx, i, n, cp;
  if (eqm == NULL || prp == NULL) return argument_list_error;
  if (dev == NULL) dev = "#0";
  memset(&htt, 0, sizeof(htt));
  strncpy(htt.c.EqmName, eqm, EQM_NAME_SHORTSIZE);
  strncpy(htt.c.EqmProperty, prp, PROPERTY_NAME_SIZE);
  strncpy(htt.c.EqmDeviceName, dev, DEVICE_NAME_SIZE);
  if ((idx=getHistoryTableIndex(&htt, nHistoryRecords)) < 0) return un_allocated;
  if (WaitForSystemMutex(hHistoryCycleMutex, -1) != 0) return mutex_error;
  if (rng.startstamp > 0 && rng.stopstamp >= rng.startstamp) useSysStamp = TRUE;
  cp = hstTbl[idx].dataPtr; /* current pointer */
  n = (cp + 1) % hstTbl[idx].depthShort;
  /* move past the current position to start at the beginning */
  for (i = 0; i < hstTbl[idx].depthShort; i++)
  {
    n = (n + 1 + hstTbl[idx].depthShort) % hstTbl[idx].depthShort; /* advance */
    if (n == cp) break; /* wrapped around: don't include the current working pointer */
    if (hstTbl[idx].timestamp[n] < rng.starttime) continue;
    if (useSysStamp && hstTbl[idx].sysstamp[n] < rng.startstamp) continue;
    if (hstTbl[idx].timestamp[n] > rng.stoptime) break;
    if (useSysStamp && hstTbl[idx].sysstamp[n] > rng.stopstamp) break;
    hstTbl[idx].dataPtr = n; /* position the data pointer prior to 'write' */
    WriteRecordToLTS(&hstTbl[idx],FALSE,TRUE);
  }
  hstTbl[idx].dataPtr = cp; /* restore the current pointer */
  hstTbl[idx].hasSavedFiles = TRUE;
  ReleaseSystemMutex(hHistoryCycleMutex);
  return 0;
}
TINE_EXPORT int AppendHistoryData(char *eqm,char *prp,char *dev,DTYPE *dout)
{
  HstTblEntry htt;
  int i, p, n;
  BYTE *dst;
  if (eqm == NULL || prp == NULL || dout == NULL) return argument_list_error;
  if (dev == NULL) dev = "#0";
  strncpy(htt.c.EqmName,eqm,EQM_NAME_SHORTSIZE);
  strncpy(htt.c.EqmProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(htt.c.EqmDeviceName,dev,DEVICE_NAME_SIZE);
  htt.c.EqmSizeOut = dout->dArrayLength;
  htt.c.EqmFormatOut = BFMT(dout->dFormat);
  if ((i=getHistoryTableIndex(&htt,nHistoryRecords)) < 0) return un_allocated;
  p = hstTbl[i].dataPtr;
  n = (p + 1) % hstTbl[i].depthShort; /* move along */
  if (dout->dTimeStamp < hstTbl[i].timestamp[p]) return out_of_sequence;
  if (WaitForSystemMutex(hHistoryTableMutex,gSystemTick*MUTEX_WAIT_AUGMENT)) return semaphore_busy;
  hstTbl[i].timestamp[n] = dout->dTimeStamp;
  hstTbl[i].sysstamp[n] = dout->sysStamp;
  hstTbl[i].usrstamp[n] = dout->dStamp;
  dst = hstTbl[i].data + n*hstTbl[i].recordLength;
  memcpy(dst,dout->data.bptr,hstTbl[i].recordLength);
  hstTbl[i].isScheduled = TRUE; /* for all intents and purposes .... */
  hstTbl[i].dataPtr = n;
  if (needToArchiveRecord(i)) 
  {
    WriteRecordToLTS(&hstTbl[i],FALSE,FALSE);
#   ifndef SMALL_TINE_LIB
    if (isPointOfInterest(&hstTbl[i])) WriteRecordToPIF(&hstTbl[i]);
#   endif
  }
  ReleaseSystemMutex(hHistoryTableMutex);
  return 0;
}
int rewireDataPool(DTYPE *d,HstTblEntry *hst,int idx)
{
  int i, dsiz=0, ssiz=0;
  size_t pptr;
  BYTE *ptr;
  DIMAGE *img;
  DataPoolEntry *dpe, *ref;

  dpe = &((DataPoolEntry *)hst->data)[idx];
  ref = &((DataPoolEntry *)hst->data)[hst->depthShort];
  switch (d->dFormat)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      dsiz = d->dArrayLength*sizeof(size_t);
      for (i=0; i<(int)d->dArrayLength; i++)
      {
        dsiz += (int)strlen(d->data.strptr[i]) + 1;
      }
      break;
    case CF_IMAGE:
      img = (DIMAGE *)d->data.vptr;
      dsiz = img->frameHeader.appendedFrameSize + IMAGE_HDR_SIZE;
      break;
  }
  if (dsiz > dpe->capacity)
  { /* must increase capacity */
    int rasiz = dsiz + 0x1000; /* a bit of surplus */
    if ((ptr=realloc(ref->data, rasiz)) == NULL) return out_of_server_memory;
    ref->data = ptr;
    ref->capacity = rasiz;
    if ((ptr=realloc(dpe->data, rasiz)) == NULL) return out_of_server_memory;
    dpe->data = ptr;
    dpe->capacity = rasiz;
    d->data.bptr = dpe->data;
  }
  switch (d->dFormat)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      ptr = d->data.bptr + d->dArrayLength*sizeof(size_t);
      pptr = (size_t)ptr;
      for (i=0; i<(int)d->dArrayLength; i++)
      {
        ssiz = (int)strlen(d->data.strptr[i]) + 1;
        memcpy(ptr,d->data.strptr[i], ssiz);
        d->data.strptr[i] = (char *)NULL + pptr;
        ptr += ssiz;
        pptr += ssiz;
      }
      dpe->length = d->dArrayLength*sizeof(size_t) + dsiz;
      break;
    case CF_IMAGE:
      img = (DIMAGE *)d->data.vptr;
      ptr = d->data.bptr + IMAGE_HDR_SIZE;
      memcpy(ptr,img->frameBuffer,img->frameHeader.appendedFrameSize);
      dpe->length = dsiz;
      break;
  }
  return 0;
}
void historyCycle(int isScheduled)
{
# define nextHstCycle { if (recordToRemove == i) recordToRemove++; continue; }
  int i,n,p,cc;
  ExportListStruct *el;
  char sDevName[DEVICE_NAME_SIZE];
  char sPropName[PROPERTY_NAME_SIZE];
  struct tm *tm, lt;
  int depth, depthDays, hss, mxto = gSystemTick;
  UINT32 delta_t;
  double hts, lhts;
  static struct timeval lastCycleTime = {0, 0};
  static time_t lasttime = 0;
  static DTYPE din,dout;
  static int mday = -1;
  int yy,mm,dd,rmvrec,nstdfiles=0,keep_longterm=TRUE,append_longterm;
  char rmvfn[64];
  static int recordToRemove = 0;

  if (ServerExitCondition) return;
  if (ServerWaiting) return;
  if (isScheduled) mxto *= MUTEX_WAIT_AUGMENT;
  if (WaitForSystemMutex(hHistoryCycleMutex,mxto) != 0) return;
  gHistoryCycleNeedsScheduler = FALSE;
  if (!isScheduled)
  {
    if (MSECS(thisCycleTime,lastCycleTime) < (int)(minHistoryPollingRate-gSystemTick)) ccerr(0);
    lastCycleTime = thisCycleTime;
  }
  if (lasttime == 0) 
  { /* initialization */
    strncpy(hstcln.usr,"LOCAL_HISTORY",USERNAME_SIZE);
    hstcon.client = &hstcln;
    hstcl.clnLst = &hstcon;
    hstcl.nclients = 1;
    hstcl.threadId = GetThreadHandle();
    DTYPEZERO(din);
    gHstIsCycling = TRUE;
  }
  lasttime = thisCycleTime.tv_sec;
  if ((tm=localtime(&lasttime)) == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "HIST: historyCycle: cannot obtain localtime !");
    ccerr(code_failure);
  }
  lt = *tm;
  if (mday < 0) mday = lt.tm_mday; /* just starting up */
  for (i=0; i<nHistoryRecords; i++)
  {
    if ((el=getExportListItem(hstTbl[i].c.EqmName)) == NULL) nextHstCycle;
    if (!el->hasPropertyHandler || el->isidle) nextHstCycle;
    if (!hstTbl[i].recordInitialized)
    {
      historyRecordInit(&hstTbl[i]);
      chkMcaIntegrity(&hstTbl[i]);
    }
    if (hstTbl[i].depthShort <= 0) nextHstCycle;
    if (mday != lt.tm_mday)
    { /* it's the dawning of a new day (sort of) */
      MakeTargetFileName(hstTbl[i].archiveFileName,lt.tm_year,lt.tm_mon+1,lt.tm_mday,i);
      rmvrec = hstTbl[i].useStandardSet ? i : recordToRemove;
      depth = HSTDEPLNG(&hstTbl[rmvrec]);
      depthDays = HSTDEPDAYS(&hstTbl[rmvrec]);
      if (depthDays <= 0) depth++;
      mm = ((lt.tm_mon+12) - (depth%12))%12;
      yy = lt.tm_year - depth/12;
      if (mm > lt.tm_mon) yy--;
      dd = mday <= 0 ? lt.tm_mday : mday;
      if (depthDays > 0 && dd > 0)
      {
        dd -= depthDays;
        if (dd < 0 || lt.tm_mday < mday)
        {
          if (--mm < 0) { mm += 12; yy--; }
          if (dd < 0) dd = -dd;
          dd = nDaysInMonth(yy,mm) - dd;
        }
      }
      MakeTargetFileName(rmvfn,yy,mm+1,dd,rmvrec);
      depth = HSTDEPLNG(&hstTbl[i]);
      depthDays = HSTDEPDAYS(&hstTbl[i]);
      if (!hstTbl[i].useStandardSet)
      { /* if not using a standard rotation set, then remove the old ones */
        if ((depth+depthDays) > 0)
        { /* and there is a long term storage */
          if (recordToRemove >= i && recordToRemove < nHistoryRecords)
          { /* careful of modulo arithmetic here. uses months from 0 to 11 */
            gServerCycleState = CYCLE_STATE_HST_RMVFILES;
            if (removeOutdatedLTSFile(rmvfn,dd+1,depthDays) >= 0) recordToRemove++;
          }
          if (minFreeDiskSpaceInBlocks > 0 && 
              !gScavengeFiles &&
              lasttime > gScavengeFileTime + 2*SEC_IN_DAY + SEC_IN_HOUR) 
          { /* if this is set, then run thru the scavenger ... */
            gScavengeFiles = TRUE;
            gScavengeFileTime = lasttime - 2*SEC_IN_DAY;
          }
        }
        else
        { /* up the counter ... */
          if (recordToRemove >= i && recordToRemove < nHistoryRecords) recordToRemove++;
        }
      }
      else
      { /* rotate the outdated file into the current position */
        if (mday > 0 && (depth+depthDays) > 0)
        {
          rotateHistoryFile(rmvfn,hstTbl[i].archiveFileName,&hstTbl[i]);
        }
        nstdfiles++; /* count the standard files (not to be removed) */
      }
    }
    gServerDataTimeStamp = PutDataTimeStampU(gDataTimeStampOffset,thisCycleTime.tv_sec,(int)(thisCycleTime.tv_usec));
    p = n = hstTbl[i].dataPtr; /* where we are now */
    if ((lhts=hstTbl[i].timestamp[p]) == 0) lhts = gServerDataTimeStamp - (double)hstTbl[i].pollingRate/1000.0;
    delta_t = (UINT32)((gServerDataTimeStamp - lhts)*1000);
    if (hstTbl[i].isScheduled) delta_t = (UINT32)hstTbl[i].pollingRate * 2;
    if (delta_t < (UINT32)(hstTbl[i].pollingRate-gSystemTick)) continue;
    n = (hstTbl[i].dataPtr + 1) % hstTbl[i].depthShort; /* move along */
    dout.dArrayLength = hstTbl[i].c.EqmSizeOut;
    dout.dFormat = LFMT(hstTbl[i].c.EqmFormatOut);
    strncpy(dout.dTag,hstTbl[i].c.strTagOut,TAG_NAME_SIZE);
    dout.data.vptr = hstTbl[i].data + n*hstTbl[i].recordLength;
    if (hstTbl[i].hasDataPool)
    { /* complex data type */
      dout.data.vptr = ((DataPoolEntry *)hstTbl[i].data)[n].data;
    }
    dout.dTimeStamp = hts = gServerDataTimeStamp;
    dout.sysStamp = hss = gSystemDataStamp;
    dout.dStamp = 0;
    strncpy(sDevName,hstTbl[i].c.EqmDeviceName,DEVICE_NAME_SIZE);
    strncpy(sPropName,hstTbl[i].c.EqmProperty,PROPERTY_NAME_SIZE);
    hstcl.exp = el;
    hstcl.prp = GetPropertyListStruct(el->EqmName,sPropName,sDevName);
    hstcl.compStatus = not_signalled;
    hstTbl[i].lastAccessTime = gServerDataTimeStamp;
    gServerCycleState = CYCLE_STATE_HST_EQM;
    cc = callEqm(&hstcl,sDevName,sPropName,&dout,&din,CA_READ|CA_HIST);
    hstTbl[i].lastAccessSts = cc;
    if (cc != 0)
    {
      char errstr[64];
      if (tineDebug || hstdbg)
      {
        dbglog("HIST: (%d) %.64s %.64s : <%d> %.32s",
          i, sDevName, hstTbl[i].c.EqmProperty, cc, GetErrorText(cc, errstr, 32));
      }
      switch (cc)
      {
        case operation_busy:
        case not_ready: /* make sure we try again next clock tick */
          lastCycleTime.tv_sec = lastCycleTime.tv_usec = 0;
          break;
        default:
          if (gHstDeactivateOnError) 
          {
            hstTbl[i].depthShort = -1;
            feclogEx(TINE_LOGLEVEL_WARN, "HIST: de-activating record (%d) %.64s %.64s due to error <%d> : %.32s",
              i, sDevName, hstTbl[i].c.EqmProperty, cc, GetErrorText(cc, errstr, 32));
          }
        case link_timeout:
          hstTbl[i].isScheduled = FALSE;
          break;
      }
      continue;
    }
    if (hstTbl[i].hasDataPool && rewireDataPool(&dout,&hstTbl[i],n) != 0) continue;
    p = hstTbl[i].dataPtr;

    if ((lhts=hstTbl[i].timestamp[p]) == 0) lhts = gServerDataTimeStamp - (double)hstTbl[i].pollingRate/1000.0;
    delta_t = (UINT32)((gServerDataTimeStamp - lhts)*1000);
    if (hstTbl[i].isScheduled) delta_t = (UINT32)hstTbl[i].pollingRate * 2;
    if (delta_t < (UINT32)(hstTbl[i].pollingRate-gSystemTick)) continue;
    if (WaitForSystemMutex(hHistoryTableMutex,mxto) == 0)
    { /* update short-term and long-term information */
      hstTbl[i].timestamp[n] = dout.dTimeStamp != hts ? dout.dTimeStamp : gServerDataTimeStamp;
      hstTbl[i].sysstamp[n] = dout.sysStamp != hss ? dout.sysStamp : gSystemDataStamp;
      hstTbl[i].usrstamp[n] = dout.dStamp;
      hstTbl[i].dataPtr = n;
      /* TODO: check points-of-interest 1st and flag inside table entry */
#     ifndef SMALL_TINE_LIB
      keep_longterm = (hstTbl[i].fltr != NULL && !hstTbl[i].fltr->valid) ? FALSE : TRUE;
      if (hstdbg && hstTbl[i].fltr != NULL)
        dbglog("HIST: (%d) %.64s %.64s : filter is %s",
          i,sDevName,hstTbl[i].c.EqmProperty,(keep_longterm?"valid":"not valid"));
#     endif
      if (keep_longterm && (append_longterm=needToArchiveRecord(i)) != 0) 
      {
        int isTransition = append_longterm == LHS_APPLY_ARCHIVE_TRANSITION ? TRUE : FALSE;
        gServerCycleState = CYCLE_STATE_HST_WRITE;
        WriteRecordToLTS(&hstTbl[i],isTransition,FALSE);
#       ifndef SMALL_TINE_LIB
        if (isPointOfInterest(&hstTbl[i])) WriteRecordToPIF(&hstTbl[i]);
#       endif
      }
      ReleaseSystemMutex(hHistoryTableMutex);
    }
    hstTbl[i].isScheduled = FALSE;
  }
  if (recordToRemove >= nHistoryRecords - nstdfiles) 
  { /* finished removing old files */
    mday = lt.tm_mday;
    recordToRemove = 0;
  }
# if !defined(SMALL_TINE_LIB) && !defined(VXWORKS)
  if (!gHstMfDumped && lasttime > gSystemStartupTime + PRE_DUMP_WINDOW) 
  {
    dumpHstManifests(gStartupDebug);
    gHstMfDumped = TRUE;
  }
  if (!gHstHomeMfUpdated && lasttime > gSystemStartupTime + PRE_UPDATE_WINDOW) 
  {
    if (updateHstHomeManifest() != operation_busy) gHstHomeMfUpdated = TRUE;
  }
  if (gScavengeFiles && recordToRemove == 0) doLtsScavenger();
# endif
err:
  ReleaseSystemMutex(hHistoryCycleMutex);
  return;
}
double gHistTimeStamp = 0;
int gHistSysStamp = 0;
int gHistUsrStamp = 0;
#define setHistTimeStamp(t) gHistTimeStamp = (t)
#define setHistSysStamp(s) gHistSysStamp = (s)
#define setHistUsrStamp(u) gHistUsrStamp = (u)
#ifndef SMALL_TINE_LIB
int openMostRecentLTS(int i)
{
  char fn[16],fileName[256];
  time_t starttime = DATATIME_NOW;
  struct tm *lt = localtime((time_t *)&starttime);
  int hFile = -1, n = 0;
  int yy, mm, dd;
  yy = lt->tm_year; mm = lt->tm_mon + 1; dd = lt->tm_mday;
  do
  {
    MakeTargetFileName(fn,yy,mm,dd,i);
    sprintf(fileName,"%.128s%.16s",arcDbPath,fn);
    if ((hFile=open(fileName,O_readmode)) != -1) return hFile;
    if (useMonthlyHistoryFiles)
    { /* start backing up thru the months */
      if (--mm == 0) { mm = 12; --yy; }
      n++;
    }
    else
    { /* start backing up thru the days */
      if (--dd == 0)
      { /* previous month ... */
        if (--mm == 0) { mm = 12; --yy; }
        dd = nDaysInMonth(yy,mm-1);
        n++;
      }
    }
  } while (n <= hstTbl[i].depthLong);
  return -1;
}
int getLastLTSData(int i,int index,void *datbuf,int fmt,int bufsiz)
{
# ifndef NO_LTS
  int hFile = -1,number,cc = 0;
  SINT32 recsiz,reclen;
  long pstart;
  DTYPE d;
  int fmtsize;
  int recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  time_t timestamp;
  double dtimestamp, datatime = 0;
  BYTE *tmpHstDatBuf;

  if (bufsiz < 1) return buffer_too_small;
  if (hstTbl == NULL) return un_allocated;
  if (ServerExitCondition) return not_running;
  if (i < 0 || i >= nHistoryRecords) return invalid_index;
  if (hstTbl[i].depthLong <= 0 && hstTbl[i].depthDays <= 0) return invalid_data; /* unavailable for summary data */
  reclen = hstTbl[i].recordLength;
  if (reclen > WORK_SIZE) goto err;

  if ((hFile=openMostRecentLTS(i)) == -1) ccerr(file_error);
  recsiz = reclen + recordPrefix;
  datatime = 0;
  pstart = getFilePosition(hFile,hstTbl[i].useStandardSet);
  pstart -= recsiz;

  if (lseek(hFile,pstart,SEEK_SET) < 0) ccerr(file_error);
  tmpHstDatBuf = (BYTE *)alloca(reclen);
  if (tmpHstDatBuf == NULL) ccerr(out_of_local_memory);
  fmtsize = fmtsizeof(hstTbl[i].c.EqmFormatOut);
  d.dFormat = LFMT(hstTbl[i].c.EqmFormatOut);
  d.dArrayLength = MIN(bufsiz, (int)hstTbl[i].c.EqmSizeOut);
  number = (int)d.dArrayLength;
  d.data.bptr = tmpHstDatBuf + index*fmtsize;
  if (read(hFile,(char *)tmpHstDatBuf,recordPrefix) != recordPrefix) ccerr(file_error);
  if (!useMinimalStorage)
  {
    memcpy(&dtimestamp,tmpHstDatBuf,sizeof(double));
  }
  else
  {
    memcpy(&timestamp,tmpHstDatBuf,sizeof(SINT32));
    dtimestamp = (double)timestamp;
  }
  if (read(hFile,(char *)tmpHstDatBuf,reclen) != reclen) ccerr(file_error);
  datatime = dtimestamp;

  switch (fmt)
  {
    case CF_INT16:
      cc = _GetValuesAsShort(&d,(SINT16 *)datbuf, number, TRUE);
      break;
    case CF_UINT16:
      cc = _GetValuesAsUnsignedShort(&d, (UINT16 *)datbuf, number, TRUE);
      break;
    case CF_INT32:
      cc = _GetValuesAsLong(&d,(SINT32 *)datbuf, number, TRUE);
      break;
    case CF_UINT32:
      cc = _GetValuesAsUnsignedLong(&d, (UINT32 *)datbuf, number, TRUE);
      break;
    case CF_INT64:
      cc = _GetValuesAsDLong(&d, (SINT64 *)datbuf, number, TRUE);
      break;
    case CF_UINT64:
      cc = _GetValuesAsUnsignedDLong(&d, (UINT64 *)datbuf, number, TRUE);
      break;
    case CF_FLOAT:
      cc = _GetValuesAsFloat(&d,(float *)datbuf, number, TRUE);
      break;
    case CF_DOUBLE:
      cc = _GetValuesAsDouble(&d,(double *)datbuf, number, TRUE);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
    case CF_TEXT:
    case CF_STRING:
    case CF_USTRING:
      cc = GetValuesAsStringEx(&d,(char *)datbuf,fmt,number,0);
      break;
    default:
      if (getFormatHeaderSize(fmt) > 0) goto err;
      bufsiz *= fmtsize;
      if (fmt == d.dFormat && bufsiz >= reclen)
      { /* asking for exactly what was stored ... */
        memcpy(datbuf,tmpHstDatBuf,reclen);
        break;
      }
      goto err;
  }
err:
  if (hFile != -1) close(hFile);
  if (datatime > 0) setHistTimeStamp(datatime);
  return cc;
# else
  return 0;
# endif /* NO_LTS */
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT int getFormattedHistoryDataPoint(DTYPE *dst,DTYPE *src)
{
# ifndef SMALL_TINE_LIB
  UINT32 strsiz;
# endif
  int fmt = dst->dFormat;
  BYTE *data = dst->data.bptr;
  DTYPE srcx;

  if (fmt == CF_HISTORY)
  { /* fill in the history header */
    int dsiz;
    memcpy(data,&src->dTimeStamp, sizeof(double));
    memcpy(&data[8],&src->sysStamp,sizeof(SINT32));
    memcpy(&data[12], &src->sysStamp, sizeof(SINT32));
    /* reset the fmt and data ptr */
    fmt = ftoi(dst->dTag);
    data += HISTORY_HDR_SIZE;
    if (fmt == src->dFormat)
    { /* finished ! */
      switch (fmt)
      {
        case CF_KEYVALUE:
        case CF_STRING:
          dsiz = (int)strlen(src->data.cptr) + 1;
          break;
        case CF_IMAGE:
        case CF_SPECTRUM:
        case CF_GSPECTRUM:
          dsiz = getFormatHeaderSize(fmt) + fmtsizeof(fmt) * dst->dArrayLength;
          break;
        default:
          dsiz = getCarriedFormatSize(dst->dTag);
          if (dsiz < 0) return -dsiz;
          break;
      }
      memcpy(data,src->data.bptr,dsiz);
      return 0;
    }
  }
  switch (src->dFormat)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      memcpy(&srcx,src,sizeof(DTYPE));
      srcx.dFormat = CF_TEXT;
      srcx.dArrayLength = (UINT32)strlen(src->data.cptr) + 1;
      src = &srcx;
      break;
    case CF_IMAGE:
      {
        DIMAGE *img = (DIMAGE *)src->data.vptr;
        DTYPEZERO(srcx);
        srcx.dFormat = CF_INT32;
        srcx.dArrayLength = 1;
        srcx.data.ulptr = &img->frameHeader.frameNumber;
        srcx.dTimeStamp = src->dTimeStamp;
        srcx.dStamp = src->dStamp;
        srcx.sysStamp = src->sysStamp;
        src = &srcx;
        break;
      }
    case CF_SPECTRUM:
      {
        DSPECTRUM *spc = (DSPECTRUM *)&src->data.vptr;
        memcpy(&srcx,src,sizeof(DTYPE));
        srcx.dFormat = CF_FLOAT;
        srcx.dArrayLength = 1;
        srcx.data.fptr = &spc->d_spect_array[0];
        src = &srcx;
        break;
      }
    case CF_GSPECTRUM:
      {
        DGSPECTRUM *spc = (DGSPECTRUM *)&src->data.vptr;
        memcpy(&srcx, src, sizeof(DTYPE));
        srcx.dFormat = CF_FLOAT;
        srcx.dArrayLength = 1;
        srcx.data.fptr = spc->spectBuffer;
        src = &srcx;
        break;
      }
    default:
      if (isTimeFormat(src->dFormat))
      { /* rewire the src */
        memcpy(&srcx,src,sizeof(DTYPE));
        switch (src->dFormat)
        {
          case CF_DBLTIME:
            srcx.dFormat = CF_DOUBLE;
            srcx.dTimeStamp = src->data.dptr[1];
            break;
          case CF_NAME64TIME:
            srcx.dFormat = CF_NAME64;
            srcx.dTimeStamp = ((NAME64DBL *)src->data.vptr)[0].dval;
            break;
        }
        src = &srcx;
      }
      break;
  }
  switch (fmt)
  { /* requested data format */
    case CF_FLOAT:
      return _GetValuesAsFloat(src,(float *)data,1, TRUE);
    case CF_BOOLEAN:
    case CF_INT32:
    case CF_UINT32:
      return _GetValuesAsLong(src,(SINT32 *)data,1, TRUE);
#   ifdef SINT64
    case CF_INT64:  
    case CF_UINT64:
      return _GetValuesAsDLong(src,(SINT64 *)data,1, TRUE);
#   endif
    case CF_DOUBLE:
      return _GetValuesAsDouble(src,(double *)data,1, TRUE);
#   ifndef SMALL_TINE_LIB
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      strsiz = fmtsizeof(fmt);
      return GetValuesAsString(src,(char *)data,&strsiz);
#   endif
    case CF_FLTINT:
      ((FLTINT *)data)->ival = (UINT32)(src->dTimeStamp);
      return _GetValuesAsFloat(src,&((FLTINT *)data)->fval,1, TRUE);
    case CF_INTFLTINT: /* CF_TDS */
    {
      int sts;
      ((INTFLTINT *)data)->i1val = (UINT32)(src->dTimeStamp);
      if (GetIntValue(src,&sts) != 0) sts = 0;
      sts &= 0x0fff;
      ((INTFLTINT *)data)->i2val = sts;
      return _GetValuesAsFloat(src,&((INTFLTINT *)data)->fval,1, TRUE);
    }
    case CF_FLTINTINT:
      ((FLTINTINT *)data)->i1val = (UINT32)src->dTimeStamp;
      ((FLTINTINT *)data)->i2val = (UINT32)src->sysStamp;
      return _GetValuesAsFloat(src,&((FLTINTINT *)data)->fval,1, TRUE);
    case CF_INTINT:
      ((INTINT *)data)->i2val = (UINT32)src->dTimeStamp;
      return _GetValuesAsLong(src,&((INTINT *)data)->i1val,1, TRUE);
    case CF_DBLTIME:
    case CF_DBLDBL:
      ((DBLDBL *)data)->d2val = src->dTimeStamp;
      return _GetValuesAsDouble(src,&((DBLDBL *)data)->d1val,1, TRUE);
#   ifndef SMALL_TINE_LIB
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
      strsiz = fmtsizeof(fmt) - sizeof(UINT32);
      *(UINT32 *)(&data[strsiz]) = (UINT32)src->dTimeStamp;
      return GetValuesAsString(src,(char *)data,&strsiz);
    case CF_NAME64TIME:
    case CF_NAME64DBL:
      strsiz = fmtsizeof(fmt) - sizeof(double);
      memcpy(&data[strsiz],&src->dTimeStamp,sizeof(double));
      return GetValuesAsString(src,(char *)data,&strsiz);
    case CF_DBLDBLDBL:
	    ((DBLDBLDBL *)data)->d2val = src->dTimeStamp;
	    ((DBLDBLDBL *)data)->d3val = (double)src->sysStamp;
      return _GetValuesAsDouble(src,&((DBLDBLDBL *)data)->d1val,1, TRUE);
    case CF_NAME64DBLDBL:
	    ((NAME64DBLDBL *)data)->d1val = src->dTimeStamp;
	    ((NAME64DBLDBL *)data)->d2val = (double)src->sysStamp;
      return GetValuesAsString(src,(char *)data,&strsiz);
#   endif
    default: 
      return illegal_format;
  }
}
int getFormattedHistoryDataSet(DTYPE *dst,DTYPE *src,int fmtsiz)
{
  UINT32 dsiz;
  int i, dlen;

  switch(src->dFormat)
  {
    case CF_INT16:
    case CF_UINT16:
    case CF_INT32:
    case CF_UINT32:
#   ifdef SINT64
    case CF_INT64:
    case CF_UINT64:
#   endif
    case CF_FLOAT:
    case CF_DOUBLE:
      return PutValuesFromAny(dst,src->data.bptr,(short)src->dFormat,1,fmtsiz,(int)src->dArrayLength,0);
    case CF_HISTORY:
      return illegal_format;
    case CF_IMAGE:
      if (dst->dFormat != src->dFormat) return illegal_format;
      dsiz = getFormatHeaderSize(src->dFormat) + MIN(dst->dArrayLength,src->dArrayLength);
      if (dsiz <= 0) return illegal_format;
      memcpy(dst->data.bptr,src->data.bptr,dsiz);
      return 0;
    case CF_KEYVALUE:
    case CF_STRING:
      if (dst->dFormat != src->dFormat) return illegal_format;
      for (dsiz=0,i=0; i<(int)dst->dArrayLength; i++)
      {
        dlen = (int)strlen(src->data.strptr[i]) + 1;
        memcpy(dst->data.bptr + dsiz,src->data.strptr[i],dlen);
        dsiz += dlen;
      }
      memcpy(dst->data.bptr,src->data.strptr,sizeof(char *) * dst->dArrayLength);
      return 0;
    default:
      if (dst->dFormat != src->dFormat) return illegal_format;
      dsiz = MIN(dst->dArrayLength,src->dArrayLength);
      dsiz *= fmtsizeof(src->dFormat);
      dsiz += getFormatHeaderSize(src->dFormat);
      if (dsiz <= 0) return illegal_format;
      memcpy(dst->data.bptr,src->data.bptr,dsiz);
      return 0;
  }
}
int isNonCompoundDataType(int fmt)
{
  switch (fmt)
  {
    case CF_IMAGE:
    case CF_DBLTIME:
    case CF_NAME64TIME:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
#ifndef NO_LTS
static double dtEarliestPossible = 0;
#endif
int loadLTSData(HstTblEntry *hst,char *fn,HstReqType *req, HstRngType rng,BYTE *data,int fmt,char *tag,int len)
{
# ifndef NO_LTS
  int hFile = -1,nrecs,raster = 1;
  int isSnapshot=FALSE,hasPointsOfInterest=FALSE,isPoiFile=FALSE,useSysStamp=FALSE;
  int hPoolFile = -1, pidx = 0;
  size_t poolPos, poolLen = 0, plen = 0;
  SINT32 cnt = 0,recsiz, reclen;
  long pstart,pstop;
  char fileName[256];
  DTYPE d, dst;
  int number=0,cc=0,p=0,sizPool=0,dstamp,sysstamp=0,usrstamp=0,datasys=rng.startstamp;
  int fmtsize,poisize,depthInSeconds,indexLimit,pci=0,sizFile=0,nidx=(int)strlen(arcDbPath);
  int rp, rfmt, recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  int foffset=0, ffmt=CF_NULL;
  time_t timestamp;
  double dband, dtimestamp, dlaststamp = -1, datatime = 0, dtnow = makeDataTimeStamp();
  static double timebin = -1;
  BYTE *tmpHstDatBuf, *nxtptr=NULL, tsBuffer[16];
# if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
  struct stat sbuf;
  int n,sizPoi;
  char poiName[164], poi[32];
  BYTE *dpoi;
  HstTblEntry hstpoi;
  HstReqType poireq; 
  HstRngType poirng;
# endif
  if (rng.stoptime < rng.starttime) return 0;
  if (timebin < 0) timebin = useMinimalStorage ? 1.0 : 0.001;
  if (len < 1) return 0;
  if (hst->depthLong < 0) return 0;
  if (hst->depthLong == 0 && hst->depthDays == 0 && !hst->storeSummary) return 0;
  depthInSeconds = hst->depthShort * hst->pollingRate / 1000;
  indexLimit = hst->storeSummary ? 4 : (int)hst->c.EqmSizeOut;
  if ((dtnow - rng.starttime) < depthInSeconds) return 0;
  reclen = HSTRECLEN(hst);
  if (reclen > WORK_SIZE) goto err;
  if (req->index < 0 || req->index >= indexLimit) goto err;
  if ((rng.stoptime-rng.starttime) < 0.01) isSnapshot = TRUE;
  if (rng.startstamp > 0 && rng.startstamp == rng.stopstamp) isSnapshot = TRUE;
  if (rng.startstamp > 0 && rng.stopstamp >= rng.startstamp && !useMinimalStorage) useSysStamp = TRUE;
  dband = useMinimalStorage ? 1.0 : 0.5;
  if (isSnapshot) 
  { /* snapshot requests always have index = 0 */
    req->index = 0; req->sample = 0;
    if (!useSysStamp) datatime = rng.starttime - timebin;
  }
  sprintf(fileName,"%.128s%.32s",arcDbPath,fn);
# if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
  if (strnicmp(fn,"pi",2) != 0)
  { /* not looking at a points-of-interest file */
    if (stat(fileName,&sbuf) == 0) 
    { /* normal archive location */
      sizFile = sbuf.st_size;
    }
    else
    { /* not there: try 'saved' area */
      sprintf(fileName,"%.128s%.32s",arcDbPathSaved,fn);
      nidx = (int)strlen(arcDbPathSaved);
      if (stat(fileName,&sbuf) == 0) sizFile = sbuf.st_size;
    }
    strncpy(poi,fn,32); poi[0] = 'p'; poi[1] = 'i';
    sprintf(poiName,"%.128s%.32s",arcDbPath,poi);
    if (sizFile > 0 && stat(poiName,&sbuf) == 0) 
    { /* history file found: is there a poi file ? */
      sizPoi = sbuf.st_size;
      if (sizPoi > 0 && sizPoi < (sizFile/10))
      {
        pci = (sizPoi*100)/sizFile; /* estimate percent interesting */
        if (pci <= 0) pci = 1;
        hasPointsOfInterest = TRUE;
        memcpy(&hstpoi,hst,sizeof(HstTblEntry));
        hstpoi.useStandardSet = FALSE;
      }
    }
  }
  else
  {
    isPoiFile = TRUE;
  }
# endif /* !SMALL_TINE_LIB */
  if (WaitForSystemMutex(hDbMutex,-1) != 0) return 0;
  if ((pstart=locateRecordFromLTS(fileName,nidx,hst,rng.starttime)) < 0) goto err;
  if ((pstop=locateRecordFromLTS(fileName,nidx,hst,rng.stoptime)) < 0) goto err;
  recsiz = reclen + recordPrefix;
  nrecs = (int)((pstop-pstart)/recsiz);
  if (len == 1 && !isSnapshot) 
  { /* just after the number of points in the interval */
    if (isPoiFile) ccerr(not_applicable);
    setHistTimeStamp(rng.stoptime);
    number = MAX(nrecs,0);
    if (nrecs > 0 && dtEarliestPossible > rng.starttime) dtEarliestPossible = rng.starttime;
    ccerr(0);
  }
  if (nrecs == 0 && isPoiFile) ccerr(not_applicable);
  if (len < nrecs) raster += nrecs/(len+1);
  if (raster > 1 && isPoiFile) ccerr(not_applicable);
  if (req->sample > 0) raster = req->sample;
  if (isSnapshot) raster = 1;
  if (raster == 1) 
  { /* don't bother ... */
    hasPointsOfInterest = FALSE; 
    pci = 0;
  }
  else if (pci > 0)
  {
    poisize = (nrecs * pci)/100;
    if (pci == 1 || poisize <= 0) poisize = 1;
    if (poisize < len/3)
    { /* tolerate up to 33% of the caller's buffer designated 'interesting' */
      raster = 1 + nrecs/(len-poisize);
    }
    else
    { /* too much is 'interesting' -> turn it off! */
      hasPointsOfInterest = FALSE;
      pci = 0;
    }
  }
  if ((hFile=open(fileName,O_readmode)) == -1) goto err;
  if (hst->hasDataPool)
  { /* complex data type : open companion file as well */
    sprintf(fileName,"%.128std%.30s",arcDbPath,&fn[2]);
    if ((hPoolFile=open(fileName,O_readmode)) == -1) goto err;
    if (!isSnapshot && fmt != CF_HISTORY && LFMT(hst->c.EqmFormatOut) != CF_IMAGE) ccerr(illegal_format);
    pidx = req->index;
    req->index = 0;
  }
  number = 0; cnt = 0; 
  lseek(hFile,pstart,SEEK_SET);

# if defined(INTEL)
  tmpHstDatBuf = hst->hasDataPool ? tmpWorkArea : (BYTE *)alloca(reclen);
# else
  tmpHstDatBuf = tmpWorkArea;
# endif
  /* kluge summary format to CF_DOUBLE (only support the 1st 4 indices) */
  rfmt = hst->storeSummary ? CF_DOUBLE : LFMT(hst->c.EqmFormatOut);
  fmtsize = fmtsizeof(rfmt);

  d.dFormat = rfmt;
  d.dArrayLength = 1;
  /* 'normally' fmt is something like CF_DBLDBL, etc. */
  p = fmtsizeof(fmt); 
  if (fmt == CF_HISTORY) 
  { /* CF_HISTORY format requested ! */
    int cfmtsiz = getCarriedFormatSize(tag);
    if (cfmtsiz < 0) ccerr(-cfmtsiz);
    p = HISTORY_HDR_SIZE + cfmtsiz;
    strncpy(dst.dTag,tag,TAG_NAME_SIZE);
  }
  else if (!isSimpleFormat((short)rfmt) && !isNonCompoundDataType(rfmt))
  {
    cc = getFieldInFormat(rfmt,NULL,hst->c.strTagOut,req->fieldindex,&foffset,&ffmt);
    if (cc != 0) ccerr(cc);
    d.dFormat = ffmt;
    if (LFMT(hst->c.EqmFormatOut) == CF_STRUCT) fmtsize *= GetStructSize(hst->c.strTagOut);
  }
  d.data.bptr = tmpHstDatBuf + foffset + req->index * fmtsize;
  dst.dArrayLength = 1;
  dst.dFormat = fmt;

  while(read(hFile,(char *)tsBuffer,recordPrefix) == recordPrefix)   
  { /* read through file */
    if (!useMinimalStorage) 
    {
      memcpy(&dtimestamp,tsBuffer,sizeof(double)); rp = sizeof(double);
      timestamp = (time_t)dtimestamp;
      memcpy(&sysstamp,&tsBuffer[rp],sizeof(SINT32)); rp += sizeof(SINT32);
      memcpy(&usrstamp,&tsBuffer[rp],sizeof(SINT32));
    }
    else 
    {
      memcpy(&timestamp,tsBuffer,sizeof(SINT32));
      dtimestamp = (double)timestamp;
    }
    dstamp = rng.userstamp ? usrstamp : sysstamp;
#   if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
    /* check if there are points of interest to add ... */
    if (hasPointsOfInterest && dlaststamp > 0 && dtimestamp > dlaststamp-2)
    { 
      memcpy(&poireq,req,sizeof(HstReqType));
      memcpy(&poirng,&rng,sizeof(HstRngType));
      poireq.sample = 1;
      poirng.starttime = dlaststamp + dband;
      poirng.stoptime = dtimestamp - dband;
      dpoi = (BYTE *)data + p*number;
      if ((poisize=len-number-1) > 2) poisize = 2;
      /* TODO: should sysstart and sysstop be adjusted here? */
      n = loadLTSData(&hstpoi,poi,&poireq,poirng,(void *)dpoi,fmt,tag,poisize);
      if (n >= len - number) goto err;
      if (n > 0) number += n;
    }
#   endif
    cnt++;
    if (isSnapshot)
    {
      if (dtimestamp > rng.stoptime) rng.stoptime = dtimestamp;
      if (dstamp > rng.stopstamp) rng.stopstamp = dstamp;
      if (useSysStamp && dstamp < rng.startstamp)
      { /* targeting a system/user stamp in a snapshot and this isn't it! */
        if (lseek(hFile, reclen, SEEK_CUR) < 0) goto err;
        continue;
      }
    }
    if (dtimestamp > rng.stoptime) goto err;
    if (useSysStamp && dstamp > rng.stopstamp) goto err;
    if ((useSysStamp && dstamp < datasys) || (dtimestamp < datatime))
    { /* older point somehow got mixed in ? */
      if (lseek(hFile,reclen,SEEK_CUR) < 0) goto err;
      continue;
    }
    if (dtimestamp < datatime + timebin)
    { /* same timestamp read again ? */
      if (number > 0) number--;
    }
    if (read(hFile,(char *)tmpHstDatBuf,reclen) != reclen) goto err;
    if (ServerExitCondition) goto err;
    if (hst->hasDataPool)
    { /* complex data type (must use CF_HISTORY, len refers to bytes and not elements) */
      plen = sizeof(size_t);
      memcpy(&poolPos,tmpHstDatBuf,plen);
      memcpy(&poolLen,&tmpHstDatBuf[plen],plen);
      lseek(hPoolFile,(long)poolPos,SEEK_SET);
      if (read(hPoolFile,(char *)tmpHstDatBuf,(unsigned int)poolLen) != (int)poolLen) goto err;
      d.dArrayLength = (unsigned int)poolLen - getFormatHeaderSize(d.dFormat);
      if (!isVariableLengthFormat((short)fmt))
      {
        sizPool++;
      }
      else
      {
        sizPool += (int)poolLen;
        if (fmt == CF_HISTORY) sizPool += HISTORY_HDR_SIZE;
      }
      if (sizPool > len) goto err;
      if (isStringFormat((short)LFMT(hst->c.EqmFormatOut)))
      {
        memcpy(&plen,tmpHstDatBuf+pidx*sizeof(plen),sizeof(plen));
        d.data.cptr = (char *)tmpHstDatBuf + hst->c.EqmSizeOut * sizeof(size_t);
        d.data.cptr += plen;
        poolLen = strlen(d.data.cptr) + 1;
      }
    }
    datatime = dtimestamp;
    datasys = dstamp;
    /* prepare source data object ... */
    d.dTimeStamp = dtimestamp;
    d.sysStamp = sysstamp;
    d.dStamp = usrstamp;

    dst.dArrayLength = len;
    dst.data.vptr = nxtptr == NULL ? (BYTE *)data + p*number : nxtptr;

    if (isSnapshot)
    { /* number will be = 0 here */
      int fsz = fmtsizeof(hst->c.EqmFormatOut);
      if (LFMT(hst->c.EqmFormatOut) == CF_STRUCT) fsz *= GetStructSize(hst->c.strTagOut);
      d.dArrayLength = MIN(len,(int)hst->c.EqmSizeOut);
      number = (int)d.dArrayLength;
      cc = getFormattedHistoryDataSet(&dst,&d,fsz);
      goto err; /* we're done */
    }
    else
    {
      cc = getFormattedHistoryDataPoint(&dst,&d);
      if (cc != 0) goto err;
      if (req->sample < 0 && !isInRange(&d,rng.lwrLimit,rng.uprLimit)) continue;
      if (hst->hasDataPool)
      {
        size_t incr = !isVariableLengthFormat((short)fmt) ? (size_t)fmtsizeof(fmt) : poolLen + HISTORY_HDR_SIZE;
        nxtptr = dst.data.bptr + incr;
      }
    }
    number++;
    if (number >= len) goto err;
    if (raster > 1) 
    { /* move along through file */
      if (lseek(hFile,recsiz*(raster-1),SEEK_CUR) < 0) goto err;
    }
    dlaststamp = dtimestamp;
  }
err:
  if (hFile != -1) close(hFile);
  if (hPoolFile != -1) close(hPoolFile);
  ReleaseSystemMutex(hDbMutex);
  if (datatime > 0) setHistTimeStamp(datatime);
  if (sysstamp > 0) setHistSysStamp(sysstamp);
  if (usrstamp > 0) setHistUsrStamp(usrstamp);
  if (cc != 0) return 0;
  return number;
# else
  return 0;
# endif /* NO_LTS */
}
int IsLiteralDevice(char *eqm,char *prp,char *dev)
{
  int i;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE)) continue;
    break;
  }
  return i < nHistoryRecords ? TRUE: FALSE;
}
int getHstRecIdx(char *eqm,char *prp, char *dev,int *index,int *index_err)
{
  int i,literalDevice;
  if (hstTbl == NULL) return -un_allocated;
  if (ServerExitCondition) return -not_running;
  literalDevice = IsLiteralDevice(eqm,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE))
    {
      if ((*index) == 0 && !literalDevice) 
      { /* not literally stored: find index from device name */
        if (index_err != NULL) *index_err = device_not_connected;
        if (dev[0] != '#')
        { /* real device name given: is this an MCA ? */
          ExportPropertyListStruct *pl=GetPropertyListStruct(eqm,prp,dev);
          if (pl == NULL) return -illegal_property;
          if (hstTbl[i].memidx != NULL)
          { /* a members list -> use the index */
            int k;
            UINT16 tgt = (UINT16)GetDeviceNumber(hstTbl[i].c.EqmName,dev);
            for (k=0; k<(int)hstTbl[i].c.EqmSizeOut; k++) if (hstTbl[i].memidx[k] == tgt)
            {
              *index = k;
              return i;
            }
            msglog(device_not_connected, "did not find %.64s in in members list", dev);
          }
          if (hstTbl[i].c.EqmDeviceName[0] != '#' && (pl->prpArrayType & AT_CHANNEL) == 0) 
          { /* real name used in local history as well and this isn't an MCA ! */
            msglog(un_allocated, "getHstRecIdx: literally stored device not bound to MCA");
            return -un_allocated;
          }
        }
        *index = GetDeviceNumberEx(eqm,dev,prp);
      }
      else continue;
    }
    return i;
  }
  msglog(un_allocated, "getHstRecIdx: literally stored device not found!");
  return -un_allocated;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int setSTSData(char *eqm,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize)
{
  int i,n,fmtsiz,sttptr=0,cc_badindex=out_of_range;
  BYTE *dptr=NULL;

  if ((i=getHstRecIdx(eqm,prp,dev,&index,&cc_badindex)) < 0) return i;
  if (hstTbl[i].depthShort == 0) return -un_allocated;
  if (index < 0) return index;
  if (index >= (int)hstTbl[i].c.EqmSizeOut) return -cc_badindex;
  if (arraysize < 1) return -buffer_too_small;
  fmtsiz = fmtsizeof(hstTbl[i].c.EqmFormatOut);
  if (arraysize > hstTbl[i].depthShort) 
  {
    sttptr = arraysize - hstTbl[i].depthShort;
    arraysize = hstTbl[i].depthShort;
  }

  for (n=0; n < arraysize && n <= hstTbl[i].depthShort; n++)
  {
    dptr = hstTbl[i].data + index*fmtsiz + n*hstTbl[i].recordLength;
    switch (fmt)
    {
      case CF_FLTINT: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_FLOAT) return -illegal_format;
        *((float *)dptr) = ((FLTINT *)data)[n+sttptr].fval;
        hstTbl[i].timestamp[n] = (double)((FLTINT *)data)[n+sttptr].ival;
        break;
      case CF_INTINT: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_INT32) return -illegal_format;
        *((UINT32 *)dptr) = ((INTINT *)data)[n+sttptr].i1val;
        hstTbl[i].timestamp[n] = (double)((INTINT *)data)[n+sttptr].i2val;
        break;
      case CF_DBLTIME:
      case CF_DBLDBL:
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_DOUBLE) return -illegal_format;
        *((double *)dptr) = ((DBLDBL *)data)[n+sttptr].d1val;
        hstTbl[i].timestamp[n] = ((DBLDBL *)data)[n+sttptr].d2val;
        break;
      case CF_NAME16I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME16) return -illegal_format;
        strncpy(((NAME16 *)dptr)->name,((NAME16I *)data)[n+sttptr].name,16);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME32I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME32) return -illegal_format;
        strncpy(((NAME32 *)dptr)->name,((NAME32I *)data)[n+sttptr].name,32);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME48I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME48) return -illegal_format;
        strncpy(((NAME48 *)dptr)->name,((NAME48I *)data)[n+sttptr].name,48);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME64I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME64) return -illegal_format;
        strncpy(((NAME64 *)dptr)->name,((NAME64I *)data)[n+sttptr].name,64);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
    }
  }    
  hstTbl[i].dataPtr = (arraysize-1) % hstTbl[i].depthShort;
  return n;
}
#endif /* !SMALL_TINE_LIB */
int getRBSummary(int idx,double sampleInterval,DSUMMARY *ds)
{
  int n,np,cp,fmtsiz,npts,k,index=0,cc=0;
  double depthInSeconds, pps, si=1, ra, ave=0, del=0, max, min, v;
  double tnow = MakeDataTimeStamp();
  BYTE *dptr;
  DTYPE d;
  DTYPEZERO(d);
  d.dFormat = LFMT(hstTbl[idx].c.EqmFormatOut);
  d.dArrayLength = 1;
  if (d.dFormat == CF_STRUCT) ccerr(illegal_format);
  fmtsiz = fmtsizeof(d.dFormat);
  if (sampleInterval > 0) si = sampleInterval; /* same value was specified */
  pps = 1000.0 / (double)hstTbl[idx].pollingRate;
  depthInSeconds = (double)hstTbl[idx].depthShort / pps;
  if (si > depthInSeconds) si = depthInSeconds;
  if ((npts=(int)(pps * si)) < 1) ccerr(invalid_interval);
  if (npts > hstTbl[idx].depthShort) npts = hstTbl[idx].depthShort;

  cp = hstTbl[idx].dataPtr; /* current data pointer */
  tnow -= (double)hstTbl[idx].pollingRate/1000.0; /* backoff the current time by 1 */
  n = (cp + hstTbl[idx].depthShort - npts) % hstTbl[idx].depthShort; /* beginning of sampling interval */
  for (np=-1,k=0; n != cp; n=np)
  {
    np = (n + 1 + hstTbl[idx].depthShort) % hstTbl[idx].depthShort;
    if (hstTbl[idx].timestamp[n] < tnow - si) continue; 
    dptr = hstTbl[idx].data + index*fmtsiz + n*hstTbl[idx].recordLength;
    d.data.vptr = dptr;
    if ((cc=_GetValuesAsDouble(&d,&v,1, TRUE)) != 0) goto err;
    ave += v;
    if (k == 0) 
    {
      max = min = v;
    }
    else
    {
      if (v > max) max = v;
      else if (v < min) min = v;
      ra = ave/(double)(k+1);
      if (v > ra) del += v - ra; else del += ra - v;
    }
    k++;
  }
  if (k == 0) ccerr(no_data_in_range);
  ave /= (double)k;
  del /= (double)k;
  ds->ave = ave;
  ds->max = max;
  ds->min = min;
  ds->dev = del;
  ds->cnt = k;
  ds->ts = (UINT32)(si + (tnow - si)/2.0);
  ds->tMax = ds->ts;
  ds->tMin = ds->ts;
err:
  return cc;
}
int getSTSDataSummary(char *eqm,char *prp, char *dev,double sampleInterval,DSUMMARY *ds)
{
  int i,index=0,cc_badindex=out_of_range,cc=0;

  if ((i=getHstRecIdx(eqm,prp,dev,&index,&cc_badindex)) < 0) return -i;
  if (hstTbl[i].depthShort == 0) return getRemoteHistoryServer(eqm,prp);
  if (hstTbl[i].hasDataPool) return not_implemented;
  if (index < 0) return -index;
  if (index >= (int)hstTbl[i].c.EqmSizeOut) return cc_badindex;
  if (WaitForSystemMutex(hHistoryTableMutex,gSystemTick) != 0) return semaphore_busy;
  cc = getRBSummary(i,sampleInterval,ds);
  ReleaseSystemMutex(hHistoryTableMutex);
  return cc;
}
#define IN_SEQ(v,n,x) (!((v)<(n) || (v)>(x)))
int getSTSData(HstReqType *req, HstRngType rng,BYTE *data,int fmt,char *tag,int len)
{
  int cc=0,i,n,p,cp,number,rnumber=0,raster,cnt,fmtsiz,dlen=1,cc_badindex=out_of_range;
  int isSnapshot=FALSE,useSysStamp=FALSE,in_seq;
  int ptot=0,plen=0,sttidx=0,foffset=0,ffmt=CF_NULL,stfmt=CF_NULL;
  int sysstamp=0, lastreadsysstamp=0, usrstamp=0, lastreadusrstamp=0, dstamp=0, laststamp;
  size_t pidx=0;
  double dtimestamp=0,lastreadtimestamp=0;
  double depthInSeconds, pps;
  DTYPE d, dst;
  BYTE *dptr, *nxtptr=NULL;
  int index = req->index;

  if ((i=getHstRecIdx(req->eqm,req->prp,req->dev,&index,&cc_badindex)) < 0) return i;
  if (hstTbl[i].depthShort == 0) return -getRemoteHistoryServer(req->eqm,req->prp);
  if (index < 0) return index;
  if (index >= (int)hstTbl[i].c.EqmSizeOut) return -cc_badindex;
  if (len < 1) return -buffer_too_small;
  req->index = index; /* re-assign in case it was coerced inside getHstRecIdx */
  pps = 1000.0 / (double)hstTbl[i].pollingRate;
  depthInSeconds = (double)hstTbl[i].depthShort / pps;
  if (hstdbg) 
  {
    dbglog("HIST: looking for %.32s data from %10.4f to %10.4f in short term memory",
      req->prp,rng.starttime,rng.stoptime);
    if (hstTbl[i].hasDataPool) dbglog("HIST: record is using a data pool");
  }
  if (rng.starttime < rng.stoptime - depthInSeconds) 
  {
    rng.starttime = rng.stoptime - depthInSeconds;
    if (hstdbg) dbglog("HIST: adjust search start time to %10.4f",rng.starttime);
  }
  stfmt = LFMT(hstTbl[i].c.EqmFormatOut);
  d.dFormat = stfmt;
  d.dArrayLength = dlen;
  strncpy(d.dTag,hstTbl[i].c.strTagOut,TAG_NAME_SIZE);
  if (!isSimpleFormat(d.dFormat) && fmt != CF_HISTORY && !isNonCompoundDataType(d.dFormat))
  {
    cc = getFieldInFormat(d.dFormat,NULL,d.dTag,req->fieldindex,&foffset,&ffmt);
    if (cc != 0) ccerr(cc);
    d.dFormat = ffmt;
  }
  if ((raster=(int)(.75 + ((rng.stoptime-rng.starttime)*pps/(double)len))) < 1) raster = 1;
  if (req->sample > 0) raster = req->sample; /* as dictated from above */
  if (len == 1) 
  {
    raster = 1;
    sttidx = 0;
  }
  isSnapshot = ((rng.stoptime-rng.starttime) < .01) ? TRUE : FALSE;
  if (rng.startstamp > 0 && rng.startstamp == rng.stopstamp) isSnapshot = TRUE;
  if (isSnapshot && len > 1)
  { /* snapshot requests always have index = 0, raster = 1 */
    raster = 1;
    index = 0; 
  }
  if (rng.startstamp > 0 && rng.stopstamp >= rng.startstamp) useSysStamp = TRUE;
  if (hstdbg) dbglog("HIST: using sampling raster %d", raster);
  fmtsiz = stfmt == CF_STRUCT ? GetStructSize(hstTbl[i].c.strTagOut) : fmtsizeof(d.dFormat);
  if (WaitForSystemMutex(hHistoryTableMutex,gSystemTick) != 0) return -semaphore_busy;
  n = hstTbl[i].dataPtr;
  /* current working pointer (should be excluded from calls) */
  cp = (n + 1) % hstTbl[i].depthShort;
  /* move raster + sttidx past the current position to start at the beginning */
  n = (cp + raster + sttidx + hstTbl[i].depthShort) % hstTbl[i].depthShort;
  lastreadtimestamp = rng.starttime;
  lastreadsysstamp = rng.startstamp;
  lastreadusrstamp = rng.userstamp;
  for (number=sttidx,cnt=0; number < len &&
                            number <= hstTbl[i].depthShort &&
                            cnt < hstTbl[i].depthShort; cnt += raster)
  {
    if (n == cp) break; /* wrapped around: don't include the current working pointer */
    in_seq = TRUE; /* be optimistic */
    if (hstTbl[i].hasDataPool)
    { /* complex data type */
      if (fmt != CF_HISTORY && stfmt != CF_IMAGE && !isSnapshot) ccerr(illegal_format);
      dptr = ((DataPoolEntry *)hstTbl[i].data)[n].data;
      plen = ((DataPoolEntry *)hstTbl[i].data)[n].length;
      if (isStringFormat((short)stfmt))
      {
        memcpy(&pidx,dptr+index*sizeof(pidx),sizeof(pidx));
        dptr += hstTbl[i].c.EqmSizeOut * sizeof(size_t);
        dptr += pidx;
        plen = (int)strlen((char *)dptr) + 1;
      }
      else if (!isVariableLengthFormat((short)stfmt))
      {
        dlen = fmtsizeof(stfmt);
        plen = getFormatHeaderSize(d.dFormat);
      }
      else
      { /* probably CF_IMAGE */
        dlen = plen - getFormatHeaderSize(d.dFormat);
      }
      d.dArrayLength = dlen;
      if (plen <= 0) { in_seq = FALSE; goto advance; }
      ptot += plen;
      if (fmt == CF_HISTORY) ptot += HISTORY_HDR_SIZE;
      if (ptot > len) goto err; /* len is bytes in this case */
    }
    else
    { /* normal case */
      dptr = hstTbl[i].data + index*fmtsiz + n*hstTbl[i].recordLength;
    }
    dtimestamp = hstTbl[i].timestamp[n];
    sysstamp = hstTbl[i].sysstamp[n];
    usrstamp = hstTbl[i].usrstamp[n];
    d.data.vptr = dptr;
    d.dTimeStamp = dtimestamp;
    d.sysStamp = sysstamp;
    d.dStamp = usrstamp;
    dstamp = rng.userstamp ? usrstamp : sysstamp;
    laststamp = rng.userstamp ? lastreadusrstamp : lastreadsysstamp;
    /* TODO: scan for points of interest in ring buffer here */
advance:
    n = (n + raster + hstTbl[i].depthShort) % hstTbl[i].depthShort;
    if (isSnapshot)
    {
      if (dtimestamp > rng.stoptime) rng.stoptime = dtimestamp;
      if (dstamp > rng.stopstamp) rng.stopstamp = dstamp;
    }
    if (in_seq) in_seq = useSysStamp ? 
                         IN_SEQ(dstamp,laststamp,rng.stopstamp) : 
                         IN_SEQ(dtimestamp,lastreadtimestamp,rng.stoptime);
    if (!in_seq) 
    {
      if (plen > 0 && ptot > 0) ptot -= plen; /* roll back the bytes counter */
      continue;
    }
    if (hstdbg) dbglog("HIST: found record at %10.4f",dtimestamp);
    lastreadtimestamp = dtimestamp;
    lastreadsysstamp = sysstamp;
    lastreadusrstamp = usrstamp;
    rnumber++;
    if (!isSnapshot && len < 2) continue;
    p = fmtsizeof(fmt);
    if (fmt == CF_HISTORY)
    {
      int cfmtsiz = getCarriedFormatSize(tag);
      if (cfmtsiz < 0) ccerr(-cfmtsiz);
      p = HISTORY_HDR_SIZE + cfmtsiz;
      strncpy(dst.dTag,tag,TAG_NAME_SIZE);
    }
    dst.dArrayLength = dlen;
    dst.dFormat = fmt;
    dst.data.vptr = nxtptr == NULL ? (BYTE *)data + p*number : nxtptr;
    if (isSnapshot)
    { /* number = 0 here */
      int fsz = fmtsizeof(hstTbl[i].c.EqmFormatOut);
      if (LFMT(hstTbl[i].c.EqmFormatOut) == CF_STRUCT) fsz *= GetStructSize(hstTbl[i].c.strTagOut);
      d.dArrayLength = MIN(len,(int)hstTbl[i].c.EqmSizeOut);
      dst.dArrayLength = d.dArrayLength;
      number = (int)d.dArrayLength;
      cc = getFormattedHistoryDataSet(&dst,&d,fsz);
      goto err; /* we're done */
    }
    else
    {
      if (ffmt != CF_NULL)
      {
        d.data.bptr += foffset;
        d.dFormat = ffmt;
      }
      cc = getFormattedHistoryDataPoint(&dst,&d);
      if (cc != 0) goto err;
      if (req->sample < 0 && !isInRange(&d,rng.lwrLimit,rng.uprLimit)) continue;
      if (hstTbl[i].hasDataPool)
      {
        size_t incr = !isVariableLengthFormat((short)stfmt) ? fmtsizeof(stfmt) : plen + HISTORY_HDR_SIZE;
        nxtptr = dst.data.bptr + incr;
      }
    }
    number++;
  }
  if (hstdbg) dbglog("HIST: found %d records in time range (snapshot: %s)",
    rnumber, isSnapshot ? "TRUE" : "FALSE");
  if (!isSnapshot) assignNumberOfPoints(data,fmt,len,&number,&rnumber);
err:
  ReleaseSystemMutex(hHistoryTableMutex);
  if (cc != 0) return -cc;
  if (lastreadtimestamp > 0) setHistTimeStamp(lastreadtimestamp);
  if (lastreadsysstamp > 0) setHistSysStamp(lastreadsysstamp); 
  if (lastreadusrstamp > 0) setHistUsrStamp(lastreadusrstamp);
  return len < 2 ? rnumber : number;
}
int isInRange(DTYPE *d,float lwrLimit,float uprLimit)
{
  BYTE *data = d->data.bptr;
  float fval = 0;
  switch (d->dFormat)
  {
    case CF_INT8:
      fval = (float)d->data.bptr[0];
      break;
    case CF_INT16:
      fval = (float)d->data.sptr[0];
      break;
    case CF_UINT16:
      fval = (float)d->data.usptr[0];
      break;
    case CF_INT32:
      fval = (float)d->data.lptr[0];
      break;
    case CF_UINT32:
      fval = (float)d->data.ulptr[0];
      break;
#   ifdef SINT64
    case CF_UINT64:
#     ifdef OS_64BIT
      fval = (float)d->data.ullptr[0];
      break;
#     endif
    case CF_INT64:
      fval = (float)d->data.llptr[0];
      break;
#   endif
    case CF_FLOAT:
      fval = d->data.fptr[0];
      break;
    case CF_DOUBLE:
      fval = (float)d->data.dptr[0];
      break;
    case CF_FLTINT:
      fval = ((FLTINT *)data)->fval;
      break;
    case CF_FLTINTINT:
      fval = ((FLTINTINT *)data)->fval;
      break;
    case CF_INTINT:
      fval = (float)((INTINT *)data)->i1val;
      break;
    case CF_DBLTIME:
    case CF_DBLDBL:
      fval = (float)((DBLDBL *)data)->d1val;
      break;
    case CF_INTFLTINT:
      fval = (float)((INTFLTINT *)data)->fval;
      break;
    default:
      return FALSE;
  }
  return (fval < lwrLimit || fval > uprLimit) ? FALSE : TRUE;
}
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound)
{ /* used onlyl for calls for 'number in interval' */
  if (data == NULL || npoints == NULL || nfound == NULL) return;
  if (siz != 1) return;
  *npoints = *nfound;

  switch (fmt)
  {
    case CF_FLTINT:
      ((FLTINT *)data)[0].ival = (UINT32)(*npoints);
      ((FLTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_FLTINTINT:
      ((FLTINTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].i2val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_INTINT:
      ((INTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((INTINT *)data)[0].i2val = (UINT32)(*npoints);
      break;
    case CF_DBLTIME:
    case CF_DBLDBL:
      ((DBLDBL *)data)[0].d1val = (double)(*npoints);
      ((DBLDBL *)data)[0].d2val = (double)(*npoints);
      break;
    case CF_NAME16I:
      ((NAME16I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME32I:
      ((NAME32I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME48I:
      ((NAME48I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME64I:
      ((NAME64I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_UINT32:
    case CF_INT32:
      if (siz == 1) *(UINT32 *)data = (UINT32)(*nfound);
      break;
    case CF_FLOAT:
      if (siz == 1) *(float *)data = (float)(*nfound);
      break;
    case CF_INTFLTINT:
      if (siz == 1) 
      {
        ((INTFLTINT *)data)[0].i1val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].i2val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].fval = (float)(*nfound);
      }
      break;
    case CF_DOUBLE:
      if (siz == 1) *(double *)data = (double)(*nfound);
      break;
  }
}

int nDaysInTimeSpan(struct tm *s,struct tm *e)
{
  int n = 0, nx, y, d, ms, me, i;

  d = nDaysInMonth(e->tm_year,e->tm_mon);
  nx = s->tm_mday + (d-e->tm_mday-1);
  for (y=s->tm_year; y<e->tm_year+1; y++)
  {
    ms = y == s->tm_year ? s->tm_mon : 0;
    me = y == e->tm_year ? e->tm_mon : 11;
    for (i=ms; i<me+1; i++) 
    {
      n += nDaysInMonth(y,i);
    }
  }
  return n - nx;
}
time_t getStartTimeFromDepth(char *eqm,char *prp,char *dev,int depth)
{
  time_t t=DATATIME_NOW;
  int i,idx=0,std;
# ifndef NO_LTS
  int n,k,siz,recsize,recordPrefix = GetRecordPrefixLength(useMinimalStorage);
  struct tm *lt,it,itprev;
  char fn[16],fileName[256];
  FILE *fp = NULL;
# endif

  if ((i=getHstRecIdx(eqm,prp,dev,&idx,NULL)) < 0) return i;
  if (depth < 0)
  { /* signal for short-term only */
    depth = -depth;
    goto err;
  }
# ifndef NO_LTS
  /* if no long term storage then there's only what's in short term */
  if (HSTDEPLNG(&hstTbl[i]) <= 0) goto err;
  recsize = HSTRECLEN(&hstTbl[i]) + recordPrefix;
  /* start looking backwards from today to find a start point where at least 'depth' points are stored */
  lt = localtime(&t); 
  it = *lt; it.tm_sec = 0; it.tm_hour = 0; it.tm_min = 0; itprev = it;
  for (k=0,n=0; k<30 && n<depth; k++)
  {
    MakeTargetFileName(fn,it.tm_year,it.tm_mon+1,it.tm_mday,i);
    sprintf(fileName,"%.128s%.16s",arcDbPath,fn);
    if ((fp=fopen(fileName,"r")) != NULL)
    {
      fseek(fp,0,SEEK_END);
      siz = ftell(fp);
      fclose(fp);
      if (siz > 0) 
      {
        if (n + siz/recsize > depth) 
        {
          return n == 0 ? mktime(&it) : mktime(&itprev);
        }
        else 
        {
          n += siz/recsize;
        }
      }
    }
    itprev = it;
    if (it.tm_mday > 1) it.tm_mday--;
    else 
    {
      if (it.tm_mon > 0)
      {
        it.tm_mon--;
      }
      else
      {
        it.tm_year--;
        it.tm_mon = 11;
      }
      it.tm_mday = nDaysInMonth(it.tm_year,it.tm_mon) + 1;
    }  
  }
  if (n > 0) return mktime(&itprev);
# endif
err:
  std = depth*hstTbl[i].pollingRate/1000;
  if (hstTbl[i].depthShort < std) std = hstTbl[i].depthShort;
  return t - std;
}
int getLTSData(HstReqType *req, HstRngType rng,BYTE *data,int fmt,char *tag,int len)
{
# ifndef NO_LTS
  SINT32 ps;
  int recsread=0,ptr=0,recsfound=0,rnumber=0,recsize,i,asize,daystotal,daysinmonth,depthInSeconds;
  int finalPass=FALSE,cc_badindex=out_of_range,indexLimit,depth,depthDays;
  int isNumberQuery=FALSE,isSnapshot=FALSE,onePassOnly=FALSE;
  char fn[64], *ctag;
  struct tm st,et,*lt,it;
  int index = req->index;
  time_t itime = 0,t = DATATIME_NOW;
  time_t starttime = (time_t)rng.starttime;
  time_t endtime = (time_t)rng.stoptime;

  if ((i=getHstRecIdx(req->eqm,req->prp,req->dev,&index,&cc_badindex)) < 0) return i;
  if (hstTbl[i].depthShort <= 0) return -getRemoteHistoryServer(req->eqm,req->prp);
  depth = HSTDEPLNG(&hstTbl[i]);
  depthDays = HSTDEPDAYS(&hstTbl[i]);
  depthInSeconds = hstTbl[i].depthShort * hstTbl[i].pollingRate / 1000;
  if (starttime >= t - depthInSeconds * STS_DEADBAND) return 0;
  if (depth <= 0  && depthDays <= 0)
  {
    if (!hstTbl[i].hasSavedFiles) return -at_limit;
    depth = 0; /* nominally irrelevant as we'll be looking specifically in the saved area */
  }
  itime = t - (depth+1)*86400*30; /* as far back as possible */
  if (starttime < 1000 && starttime < itime) starttime = itime;
  indexLimit = hstTbl[i].storeSummary ? 4 : (int)hstTbl[i].c.EqmSizeOut;
  if (index < 0) return index;
  if (index >= indexLimit) return -cc_badindex;
  req->index = index; /* in case it was coerced inside getHstRecIdx() */
  if (len < 1) return -buffer_too_small;
  if (isVariableLengthFormat(ftoi(tag))) onePassOnly = TRUE;
  if ((rng.stoptime-rng.starttime) < 0.01) isSnapshot = TRUE;
  if (rng.startstamp > 0 && rng.startstamp == rng.stopstamp) isSnapshot = TRUE;
  if (fmt == CF_HISTORY)
  {
    int cfmtsiz = getCarriedFormatSize(tag);
    if (cfmtsiz < 0) return -cfmtsiz;
    recsize = HISTORY_HDR_SIZE + cfmtsiz;
  }
  else
  {
    recsize = fmtsizeof(fmt);
  }
  lt = localtime(&endtime); et = *lt;
  lt = localtime(&starttime); st = *lt;
  if (et.tm_mon == st.tm_mon && et.tm_mday == st.tm_mday)
  {
    asize = len;
    finalPass = TRUE;
  }
  else if (len == 1)
  { /* just a number query */
    asize = len;
    isNumberQuery = TRUE;
  }
  else
  { /* spans multiple files */
    asize = 0;
  }
  dtEarliestPossible = (double)t;
  daystotal = nDaysInTimeSpan(&st,&et);
  if (daystotal <= 0)
  { /* due to TZ weirdness from above ? */
    asize = len;
    finalPass = TRUE;
  }
  while (recsread < len)
  {
    if (asize < len || len == 1)
    {
      it = st;
      if (it.tm_year == et.tm_year && it.tm_mon == et.tm_mon && 
          (it.tm_mday == et.tm_mday || useMonthlyHistoryFiles))
      {
        asize = len;
        finalPass = TRUE;
      }
      else  
      { /* move along */
        it.tm_mday = (it.tm_mday+1)%(nDaysInMonth(it.tm_year,it.tm_mon)+1);
        if (it.tm_mday == 0)
        {
          it.tm_mday++;
          it.tm_mon = (it.tm_mon+1)%12;
          if (it.tm_mon == 0) it.tm_year++;
        }
        it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
        itime = mktime(&it);
        daysinmonth = (int)MAX((itime-starttime)/(3600*24),1);
        if (len > 1) 
        {
          if (isSimpleFormat((short)fmt) || isSnapshot) 
          { /* almost certainly a snapshot if fmt a simple type */
            asize = len;
            finalPass = TRUE;
          }
          else if (req->sample > 0)
          { /* specific sampling raster given */
            asize = len;
          }
          else 
          {
            asize += MAX((len * daysinmonth)/daystotal,1);
            if (asize >= len) finalPass = TRUE; /* just in case */
          }
        }
      }
    }
    if (asize > len) asize = len;
retrySnapshot:
    MakeTargetFileName(fn,st.tm_year,st.tm_mon+1,st.tm_mday,i);
    ctag = fmt == CF_HISTORY ? tag : hstTbl[i].c.strTagOut;
    ps = loadLTSData(&hstTbl[i],fn,req,rng,&data[ptr],fmt,ctag,asize-recsread);
    if (req->sample > 0) asize = 0; /* reset this for next pass */
    if (asize == 1) recsfound += (int)ps; else recsread += (int)ps;
    ptr = recsread * recsize;
    if (isSnapshot && ps == 0)
    { /* snapshot call and possibly at the end of the designated file (n.b. finalPass = TRUE) */
      st.tm_mday = (st.tm_mday+1)%(nDaysInMonth(st.tm_year,st.tm_mon)+1);
      if (st.tm_mday == 0)
      {
        st.tm_mday++;
        st.tm_mon = (st.tm_mon+1)%12;
        if (st.tm_mon == 0) st.tm_year++;
      }
      st.tm_hour = 0; st.tm_min = 0; st.tm_sec = 0;
      isSnapshot = 0; /* only do this once ! */
      finalPass = TRUE; /* just to be sure ! */
      goto retrySnapshot;
    }
    if (ps > 0 && onePassOnly) break;
    if (finalPass) break;
    starttime = itime;
    rng.starttime = (double)starttime;
    lt = localtime(&starttime); st = *lt;
  }
  rnumber = recsfound > 0 ? recsfound : recsread;
  if (len == 1) recsread = recsfound;

  if (!isSnapshot) assignNumberOfPoints(data,fmt,len,&recsread,&recsfound);
  if (isNumberQuery) setHistTimeStamp(dtEarliestPossible);
  return rnumber;
# else
  return 0;
# endif /* NO_LTS */
}
int getRemoteHistoryServer(char *eqm,char *prp)
{
  int i, cc=0;
  char rdr[STATUS_SIZE];
  ExportListStruct *el = getExportListItem(eqm);
  if (ServerExitCondition) return not_running;
  if (el == NULL) return non_existent_elem;
  if (nHistoryRecords <= 0 || hstTbl == NULL) ccerr(un_allocated);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) ccerr(un_allocated);
  if (hstTbl[i].prd == NULL) ccerr(not_defined);
  memcpy(el->EqmLastErrorString,hstTbl[i].prd->prpRedirection,STATUS_SIZE);
err:
  if (cc != 0)
  { /* try the central archive systematics */
    memset(rdr,0,STATUS_SIZE);
    strncpy(&rdr[PRP_REDIR_CTX_OFFSET],el->EqmContext,CONTEXT_NAME_SIZE);
    strncpy(&rdr[PRP_REDIR_SRV_OFFSET],"HISTORY",EXPORT_NAME_SIZE);
    if (strlen(el->EqmMaster) > 0)
      sprintf(&rdr[PRP_REDIR_PRP_OFFSET],"%.30s.%.32s",el->EqmMaster,prp);
    else
      sprintf(&rdr[PRP_REDIR_PRP_OFFSET],"%.30s.%.32s",el->EqmExportName,prp);
    memcpy(el->EqmLastErrorString,rdr,STATUS_SIZE);
  }
  return server_redirection;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int FindAvailableHistoryIndex(char *eqm,char *prp,char *dev,int fmt,int siz)
{
  int i;
  static int maxidx=0;
  if ((i=findRecordIndexInHstManifest(eqm,prp,dev,fmt,siz)) >= 0)
  { /* already known to the manifest */
    return i; 
  }
  for (i=0; i<nHistoryRecords; i++)
  {
    if (hstTbl[i].recordIndex > maxidx) maxidx = hstTbl[i].recordIndex;
  }
  if (maxidx < HRS_FREE_RECORD_INDEX_OFFSET) maxidx += HRS_FREE_RECORD_INDEX_OFFSET;
  return maxidx + 1;
}
TINE_EXPORT int ApplyHistoryFilter(int idx,char *parsableFilterString)
{
  FilterLink *fltr = getFilterLinkFromParsableString(parsableFilterString);
  int i,removeFilter=FALSE;
  char fltrStr[512];
  if (ServerExitCondition) return not_running;
  if (parsableFilterString == NULL) removeFilter = TRUE;
  else if (fltr == NULL) return parsing_error;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (hstTbl[i].recordIndex == idx)
    {
      if (removeFilter)
      {
        strcpy(fltrStr,"remove any applied filters");
        if (hstTbl[i].fltr != NULL)
        {
          rmvFilterLink(hstTbl[i].fltr);
        }
      }
      else
      {
        sprintf(fltrStr,"apply filter %.256s %.2s %.64s",
          fltr->key,getComparatorString(fltr->fcmp),fltr->fstr);
      }
      hstTbl[i].fltr = fltr;
      feclog("HIST: rec %d (0x%x) (%.6s)/%.64s[%.64s] -> %s",
            hstTbl[i].recordIndex,hstTbl[i].recordIndex,
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].c.EqmProperty,
            fltrStr);
      return 0;
    }
  }
  return invalid_index;
}
HistorySpecification defaultHSpec = 
{
  HRS_DEFAULT_POLL_RATE,HRS_DEFAULT_ARCHIVE_RATE,HRS_DEFAULT_DEPTHSHORT,HRS_DEFAULT_DEPTHLONG,
  HRS_DEFAULT_HEARTBEAT,HRS_DEFAULT_REL_TOLERANCE,HRS_DEFAULT_ABS_TOLERANCE,NULL,NULL
};
int gIsLclHstLateAddition = FALSE;
TINE_EXPORT int AppendHistoryInformationEx(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec,int hstFlags)
{
  int i = -1, cc = 0;
  time_t t = time(NULL);
  ExportPropertyListStruct *pl=NULL;
# if !defined(VXWORKS)
  HstTblEntry *htt;
# endif

  if (eqm == NULL || prp == NULL || dev == NULL || hspec == NULL) return argument_list_error;
  if (ServerExitCondition) return not_running;
  if (!isAllowedHistoryFormat(fmt)) ccerr(illegal_format);
  if ((pl=GetPropertyListStruct(eqm,prp,dev)) == NULL) return illegal_property;
  if (len < 1) ccerr(illegal_data_size);
  if (idx < 0) idx = FindAvailableHistoryIndex(eqm, prp, dev, fmt, len);
  historyInit(GetExportedName(eqm),eqm);
  makeDirectoryPath(arcDbPath,TRUE);

  if (hstTbl == NULL)
  {
    if ((hstTbl=(HstTblEntry *)SystemCalloc(maxHistoryRecords,sizeof(HstTblEntry))) == NULL)
      ccerr(out_of_local_memory);
  }
  if (WaitForSystemMutex(hHistoryCycleMutex, -1) != 0) return mutex_error;
  if ((i=nHistoryRecords) >= maxHistoryRecords)
  { /* just let this grow */
#   if !defined(VXWORKS)
    if ((htt=(HstTblEntry *)realloc(hstTbl,(maxHistoryRecords+addHistoryCapacity)*sizeof(HstTblEntry))) == NULL) ccerr(out_of_server_memory);
    hstTbl = htt;
    memset(&hstTbl[maxHistoryRecords],0,addHistoryCapacity*sizeof(HstTblEntry));
    maxHistoryRecords += addHistoryCapacity;
#   else
    feclogEx(TINE_LOGLEVEL_WARN, "HIST: added history element would exceed capacity (%d)",maxHistoryRecords);
    ccerr(resources_exhausted);
#   endif
  }
  if (!gIsLclHstLateAddition && (t - gSystemStartupTime > gLclHstLateAdditionWindow))
  { /* 15 minutes after server start */
    gIsLclHstLateAddition = TRUE;
    feclog("HIST: now marking appended history records as LATE");
  }
  if ((cc=addHistoryItem(i,eqm,prp,dev,pl->prpFormatTag.name,len,fmt,idx,hspec,hstFlags)) != 0) ccerr(cc);
  nHistoryRecords++;
  if (hstTbl[i].useStandardSet) makeFilesForRecord(i);
  if (MaxPollingRate > minHistoryPollingRate)
  {
    feclog("adjust cycle deadband (due to local history subsystem) from %d to %d msec",
      gSystemCycleDeadband,minHistoryPollingRate);
    if (gSystemCycleDeadband > (UINT32)minHistoryPollingRate) gSystemCycleDeadband = (UINT32)minHistoryPollingRate;
    MaxPollingRate = minHistoryPollingRate;
  }
  if (gHstMfDumped) 
  { /* this record is new and the history manifest has already been dumped: do it again! */
    dumpHistoryManifest(eqm,TRUE);
  }
err:
  ReleaseSystemMutex(hHistoryCycleMutex);
  if (cc != 0 && cc != already_assigned && i >= 0)
  {
    feclog("HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] : %.32s",
      hstTbl[i].recordIndex,hstTbl[i].recordIndex,
      hstTbl[i].c.EqmName,hstTbl[i].c.EqmDeviceName,hstTbl[i].c.EqmProperty,
      cc2str(cc));
  }
  return cc;
}
TINE_EXPORT int AppendHistoryInformation(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec)
{
  return AppendHistoryInformationEx(eqm,prp,dev,len,fmt,idx,hspec,0);
}
int addHistoryItem(int i, char *eqm,char *prp,char *dev,char *tag,int len,int fmt,int idx,HistorySpecification *hspec,int hstFlags)
{
# define MAX_LONGTERMSTORAGE 240
  int tblidx, fsiz, cc = 0;
  time_t t = time(NULL);
  struct tm lt, *tm = localtime(&t);
  if (tm == NULL) return code_failure;
  lt = *tm;
  strncpy(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE);
  strncpy(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE);
  hstTbl[i].c.EqmSizeOut = len;
  hstTbl[i].c.EqmFormatOut = BFMT(fmt);
  if (tag != NULL) strncpy(hstTbl[i].c.strTagOut,tag,TAG_NAME_SIZE);
  if ((tblidx=getHistoryTableIndex(&hstTbl[i],nHistoryRecords)) >= 0)
  { /* already in the history tables ! */
    i = tblidx;
    hstTbl[i].heartbeat = hspec->heartbeat;
    hstTbl[i].pollingRate = hspec->pollingRate;
    hstTbl[i].archiveRate = hspec->archiveRate;
    if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate *= 1000; /* was thinking in seconds ! */
    if (hstTbl[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = hstTbl[i].pollingRate;
    hstTbl[i].pTolerance = hspec->pTolerance;
    if (hstTbl[i].pTolerance > 1.0) hstTbl[i].pTolerance /= 100;
    hstTbl[i].aTolerance = hspec->aTolerance;
    hstTbl[i].depthLong = HSTLONGMONTHS(hspec->depthLong);
    if (hstTbl[i].depthLong > MAX_LONGTERMSTORAGE) 
    {
      feclogEx(TINE_LOGLEVEL_WARN, "HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] long term storage truncated from %d to %d months",
              hstTbl[i].recordIndex,hstTbl[i].recordIndex,
              hstTbl[i].c.EqmName,
              hstTbl[i].c.EqmDeviceName,
              hstTbl[i].c.EqmProperty,
              hstTbl[i].depthLong, MAX_LONGTERMSTORAGE
              );
      hstTbl[i].depthLong = MAX_LONGTERMSTORAGE;
    }
    hstTbl[i].depthDays = HSTLONGDAYS(hspec->depthLong);
    feclog("HIST: modifying history parameters for rec %d (0x%x) -> (%.6s)/%.64s[%.64s]",
            hstTbl[i].recordIndex,hstTbl[i].recordIndex,
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].c.EqmProperty);
    ccerr(already_assigned);
  }
  hstTbl[i].recordIndex = idx;
  if ((tblidx=getHstTblEntryFromRecordIndex(idx)) >= 0)
  { /* check for conflicts */
    feclogEx(TINE_LOGLEVEL_ERROR, "HIST: cannot add %.64s %.64s to history table, rec %d (0x%x) already assigned to %.64s %.64s",
      prp,dev,idx,idx,
      hstTbl[tblidx].c.EqmProperty,hstTbl[tblidx].c.EqmDeviceName);
    return already_assigned;
  }
  hstTbl[i].heartbeat = hspec->heartbeat;
  hstTbl[i].pollingRate = hspec->pollingRate;
  if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate *= 1000; /* was thinking in seconds ! */
  if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate = 1000;  /* don't allow some idiot to pass a zero ! */
  if (hstTbl[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = hstTbl[i].pollingRate;
  hstTbl[i].archiveRate = hspec->archiveRate;
  if (hstTbl[i].archiveRate < 20) hstTbl[i].archiveRate *= 1000; /* was thinking in seconds ! */
  if (hstTbl[i].archiveRate < 20) hstTbl[i].archiveRate = 1000; /* don't allow some idiot to pass a zero ! */
  if (hspec->pTolerance < 0)
  {
    hstTbl[i].keepPointsOfInterest = USE_POI_NONE;
    hspec->pTolerance = 0;
  }
  if (hspec->aTolerance < 0)
  {
    hstTbl[i].keepPointsOfInterest = USE_POI_NONE;
    hspec->aTolerance = 0;
  }
  if ((hstFlags & 0x0007) != 0) hstTbl[i].keepPointsOfInterest = hstFlags;
  hstTbl[i].pTolerance = hspec->pTolerance;
  if (hstTbl[i].pTolerance > 1.0) hstTbl[i].pTolerance /= 100;
  hstTbl[i].aTolerance = hspec->aTolerance;
  hstTbl[i].depthShort = hspec->depthShort;
  hstTbl[i].depthLong = HSTLONGMONTHS(hspec->depthLong);
  if (hstTbl[i].depthLong > 240) hstTbl[i].depthLong = 240;
  hstTbl[i].depthDays = HSTLONGDAYS(hspec->depthLong);
  fsiz = LFMT(hstTbl[i].c.EqmFormatOut) == CF_STRUCT ? 
           GetStructSize(hstTbl[i].c.strTagOut) : 
           fmtsizeof(hstTbl[i].c.EqmFormatOut);
  hstTbl[i].recordLength =
      (int)hstTbl[i].c.EqmSizeOut * fsiz + getFormatHeaderSize(fmt);
  if (hstTbl[i].recordLength == 0)
  {
    feclog("HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] record has no record size",
            hstTbl[i].recordIndex,hstTbl[i].recordIndex,
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].c.EqmProperty);
    ccerr(configuration_error);
  }
# ifdef OS_16BIT
  if (hstTbl[i].recordLength > 0x3fff)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] record too long (%d bytes)",
            hstTbl[i].recordIndex,hstTbl[i].recordIndex,
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].c.EqmProperty,
            hstTbl[i].recordLength);
    ccerr(record_too_long);
  }
# endif
  if (isVariableLengthFormat((short)fmt))
  { /* will generally receive less data than asked for */
    hstTbl[i].hasDataPool = TRUE;
    hstTbl[i].keepPointsOfInterest = FALSE;
    if (hstTbl[i].poolEntryCapacity == 0)
    { /* not yet assigned */
      hstTbl[i].poolEntryCapacity = hstTbl[i].recordLength;
      switch (fmt)
      {
        case CF_STRING:
        case CF_KEYVALUE:
          hstTbl[i].poolEntryCapacity *= STRING_STD_CAPACITY;
        default:
          if (hstTbl[i].poolEntryCapacity > 0x100000)
          { /* likely CF_IMAGE : start off small */
            hstTbl[i].poolEntryCapacity /= 10;
          }
          break;
      }
    }
    /* now re-assign the recordLength */
    hstTbl[i].recordLength = 2 * sizeof(size_t);
  }
  if (hstTbl[i].depthLong == 0 && hstTbl[i].depthDays == 0 && hstTbl[i].archiveRate > hstTbl[i].pollingRate)
  { /* signal to keep a data summary in long-term storage */
    if (len == 1 && isNumberFormat((short)fmt)) 
    {
      hstTbl[i].storeSummary = TRUE;
      hstTbl[i].keepPointsOfInterest = USE_POI_NONE;
    }
    else
    {
      feclogEx(TINE_LOGLEVEL_WARN, "HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] cannot store summary or array type or non-number type!",
              hstTbl[i].recordIndex,hstTbl[i].recordIndex,
              hstTbl[i].c.EqmName,
              hstTbl[i].c.EqmDeviceName,
              hstTbl[i].c.EqmProperty);
    }
  }
  hstTbl[i].useStandardSet = useStandardFileSet && !gIsLclHstLateAddition ? TRUE : FALSE;
  MakeTargetFileName(hstTbl[i].archiveFileName,lt.tm_year,lt.tm_mon+1,lt.tm_mday,i);
  if (hspec->rhsServerName != NULL && strlen(hspec->rhsServerName) > 0)
  {
    if ((hstTbl[i].prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) 
      ccerr(out_of_server_memory);
    ParseRedirector(hspec->rhsServerName,hstTbl[i].prd);
    if (!assertRedirectionValid("addHistoryItem",hstTbl[i].c.EqmName, hstTbl[i].prd->prpRedirection))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] redirection string %.192s is invalid!",
              hstTbl[i].recordIndex,hstTbl[i].recordIndex,
              hstTbl[i].c.EqmName,
              hstTbl[i].c.EqmDeviceName,
              hstTbl[i].c.EqmProperty,
              hspec->rhsServerName);
      SystemFree(hstTbl[i].prd);
    }
    else if (hspec->rhsPropertyName != NULL)
    {
      strncpy(&hstTbl[i].prd->prpRedirection[PRP_REDIR_PRP_OFFSET],hspec->rhsPropertyName,PROPERTY_NAME_SIZE);
    }
  }
  if (hstTbl[i].depthShort > 0)
  { /* normally true */
    if ((hstTbl[i].timestamp=(double *)SystemCalloc(hstTbl[i].depthShort,sizeof(double))) == NULL)
        ccerr(out_of_local_memory);
    if ((hstTbl[i].sysstamp=(SINT32 *)SystemCalloc(hstTbl[i].depthShort,sizeof(SINT32))) == NULL)
      ccerr(out_of_local_memory);
    if ((hstTbl[i].usrstamp=(SINT32 *)SystemCalloc(hstTbl[i].depthShort,sizeof(SINT32))) == NULL)
      ccerr(out_of_local_memory);
    if (hstTbl[i].hasDataPool)
    { /* variable length format (complex entry) */
      DataPoolEntry *dpe;
      int k;
      if ((dpe=(DataPoolEntry *)SystemCalloc(hstTbl[i].depthShort+1,sizeof(DataPoolEntry))) == NULL)
        ccerr(out_of_local_memory);
      for (k=0; k<hstTbl[i].depthShort+1; k++)
      { /* each ring buffer index with its own data block */
        if ((dpe[k].data=(BYTE *)SystemCalloc(hstTbl[i].poolEntryCapacity,1)) == NULL) ccerr(out_of_local_memory);
        dpe[k].capacity = hstTbl[i].poolEntryCapacity; /* initial capacity */
      }
      hstTbl[i].data = (BYTE *)dpe;
    }
    else
    { /* normal case (Note: the 'depthShort+1' refers to allocating extra space for the 'last' saved record */
      if ((hstTbl[i].data=(BYTE *)SystemCalloc(hstTbl[i].depthShort+1,hstTbl[i].recordLength)) == NULL)
         ccerr(out_of_local_memory);
    }
  }
  else
  { /* not kept locally then we MUST be redirecting this ... */
    if (hstTbl[i].prd == NULL)
    {
      feclog("HIST: rec %d (0x%x) -> (%.6s)/%.64s[%.64s] record has no depth and is not redirected !",
              hstTbl[i].recordIndex,hstTbl[i].recordIndex,
              hstTbl[i].c.EqmName,
              hstTbl[i].c.EqmDeviceName,
              hstTbl[i].c.EqmProperty);
      ccerr(configuration_error);
    }
    hstTbl[i].depthShort = HRS_REDIRECTION_FLAG; /* will show up in the table with a -1 as short term depth */
  }
  feclog("HIST: rec %d (0x%x) (%d of %d) -> (%.6s)/%.64s[%.64s] (%d s %d mon %d days, summary: %s)",
          hstTbl[i].recordIndex,hstTbl[i].recordIndex,
          i,maxHistoryRecords,
          hstTbl[i].c.EqmName,
          hstTbl[i].c.EqmDeviceName,
          hstTbl[i].c.EqmProperty,
          hstTbl[i].depthShort,
          hstTbl[i].depthLong,
          hstTbl[i].depthDays,
         (hstTbl[i].storeSummary ? "TRUE" : "FALSE"));
  checkFileIntegrity(&hstTbl[i]);
err:
  return cc;
}

TINE_EXPORT int GetLastStoredData(char *eqm, char *prp, char *dev, int index, BYTE *data, int fmt, int len)
{
  return GetLastStoredDataEx(eqm, prp, dev, index, data, fmt, len, NULL);
}
TINE_EXPORT int GetLastStoredDataEx(char *eqm,char *prp, char *dev,int index,BYTE *data,int fmt,int len,double *tstamp)
{
  int ptr = 0, i, cc;
  if ((i=getHstRecIdx(eqm,prp,dev,&index,NULL)) < 0) return i;
  if (hstTbl[i].depthLong <= 0 && hstTbl[i].depthDays <= 0) return 0;
  if (index < 0) return -index;
  if (index > (int)hstTbl[i].c.EqmSizeOut) return out_of_range;
  if (len < 1) return buffer_too_small;
  if ((cc=getLastLTSData(i, index, &data[ptr], fmt, len)) == 0)
  {
    if (tstamp != NULL) *tstamp = gHistTimeStamp;
  }
  return cc;
}
void getHistoryFilterString(FilterLink *fltr, char *filterString)
{
  if (fltr != NULL)
  {
    int c = fltr->fcmp;
    strncpy(filterString, fltr->key, 256);
    switch (c)
    {
    case fl_eq:
      strcat(filterString, "=");
      break;
    case fl_neq:
      strcat(filterString, "!=");
      break;
    case fl_lt:
      strcat(filterString, "<");
      break;
    case fl_gt:
      strcat(filterString, ">");
      break;
    default:
      break;
    }
    strncat(filterString, fltr->fstr, 64);
  }
  else
  {
    strcpy(filterString, "NONE");
  }
}
#endif /* SMALL_TINE_LIB */
