/*
 * File server.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file server.c
 * \brief TINE server engine plus API routines
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "prolog.h"
#include "srvcore.h"
#include "threader.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"
#include "almlib.h"
#include "hstlib.h"
#include "srvip.h"
#if defined(IPXNW) || defined(IPXBSD)
# include "srvipx.h"
#endif
#include "toolkit.h"
#include "ttycmd.h"
#if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
# include "shmalloc.h"
#endif

int canSendGlobals = CAN_SENDGLOBALS;
int ngcasts = 0, nglobals = 0;
int gBurstLimit = BURST_LIMIT;
int gTcpBurstLimit = 0;
int gCycleDelay = CYCLE_DELAY;
int gCycleMicroDelay = 0;
TINE_EXPORT int SystemPresetMemory = TRUE;
int gDelayInitTineServices = FALSE;
int gSystemUseDataProtection = FALSE;
int gInitTineServicesDone = FALSE;
int gIsGlobalsServer = FALSE;
int gOneTimeSyncPass = FALSE;
int gIsExitingEQPModules = FALSE;
int putCommandsInCmdlog = INCLUDE_IN_LOGFILE;
TINE_EXPORT void SetPutCommandsInCmdlog(int value)
{
  putCommandsInCmdlog = value;
}
TINE_EXPORT int GetPutCommandsInCmdlog(void)
{
  return putCommandsInCmdlog ? TRUE : FALSE;
}
extern int gResetUseLoopback;
extern int clientTimeout;
extern int clientTimeoutMsec;
#ifdef WIN32
void timeSinceLastCall(char *text,...);
#endif
#ifndef srvWorkArea
BYTE *srvWorkArea;
UINT32 srvWorkAreaSize = WORK_SIZE;
#else
BYTE *WorkArea;
UINT32 WorkAreaSize = WORK_SIZE;
#endif
BYTE *tmpWorkArea = NULL;
UINT32 tmpWorkAreaSize = WORK_SIZE;
 
void SurveyAlarms(struct timeval *tv);
void DataDump(char *txt,char *data,int siz,short fmt,short dir);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
void CleanupStrandedDatagrams(void);
void _SystemEngine(void);
int initTineServices(void);
DBLNKLIST *getFecInfoList(void);
int isIndependentWriteProperty(ExportListStruct *el, char *prpName);

int useCycleTrigger = TRUE;
int gEqmFcnAccessor = 0;
UINT32 gSystemCycleDeadband = DEFAULT_POLLRATE;
TINE_EXPORT UINT16 MinPollingRate = MIN_POLLRATE;
TINE_EXPORT UINT16 MaxPollingRate = DEFAULT_POLLRATE;
TINE_EXPORT UINT32 MaxSystemTransportSize = MAX_TRANSPORTSIZE;
int gCycleDeadbandFixed = FALSE;
int enforceMcaAcquisition = FALSE;
int gGCastMaxPayload = GCAST_BUFFER_SIZE; 
int AveBusyTime = 0;
struct timeval thisCycleTime;
int gRequireAcknowledgments = 1; /* global flag to force acknowledge in REFRESH mode */
short gNumEqmBkgTasks = 0;
int nExpiredContracts = 0;
static char gGlbType[USERTYPE_SIZE];
TINE_EXPORT void SetGlobalsType(const char *type)
{
  if (type != NULL) strncpy(gGlbType, type, USERTYPE_SIZE);
}
TINE_EXPORT char *GetGlobalsType(void)
{
  return gGlbType;
}
ContractProgress conprg[MAXLONGGRAMS];
int nBusyContractProgress = 0;
int gTransportCeiling = -1;
int nDataOutBuffers = NUM_DEFAULT_DATAOUT_BUFFERS;
TINE_EXPORT int gServerCycleState = CYCLE_STATE_NONE;
#ifndef SMALL_TINE_LIB
char gServerCycleStateStr[][64] = {
  "none",
  "idle",
  "initializing",
  "initializing final step",
  "delaying",
  "accepting requests",
  "inside scheduler",
  "calling equipment module/stock property",
  "acquiring property device names",
  "cleanup activities",
  "checking the alarm watch table",
  "processing alarms",
  "initializing local history subsystem",
  "making local history files",
  "calling equipment modules for local history",
  "writing local history information to disk",
  "removing outdated local history files",
  "sending export information to ENS",
  "checking access locks",
  "checking ACL group information",
  "shutting down"
};
TINE_EXPORT int GetServerCycleState(char *cycleStateString)
{
  if (cycleStateString != NULL) strncpy(cycleStateString,gServerCycleStateStr[gServerCycleState],64);
  return gServerCycleState;
}
#endif
TINE_EXPORT void SetUseCycleTrigger(int value)
{
  useCycleTrigger = value;
  feclog("Use Cycle Trigger set to %s",(useCycleTrigger?"ON":"OFF"));
}
TINE_EXPORT int GetUseCycleTrigger(void)
{
  return useCycleTrigger;
}
TINE_EXPORT void SetServerTransportCeiling(int value)
{
  if (value < (int)MaxSystemTransportSize) value = (int)MaxSystemTransportSize;
  gTransportCeiling = value;
  feclog("server transport ceiling set to %d bytes",gTransportCeiling);
}
TINE_EXPORT int GetServerTransportCeiling(void)
{
  return gTransportCeiling;
}
TINE_EXPORT void SetTcpBurstLimit(int value) { gTcpBurstLimit = value; }
TINE_EXPORT int GetTcpBurstLimit(void) { return gTcpBurstLimit; }

GlobalListStruct **gcastTbl = NULL;
IPXAddress IPXFecAddress;
int ExportsRegistered = FALSE;
int FecNameRegistered = FALSE;
FecInfoStruct gFecInfo;

#ifdef INTEL
int gEndianness = BO_LITTLE_ENDIAN;
#else
int gEndianness = BO_BIG_ENDIAN;
#endif
void chkEndian(void)
{
  short shortendian = 1;
  BYTE shortbytes[2]={0,0};
  char *msg;

  memcpy(shortbytes,&shortendian,2);
  if (shortbytes[0] == 1)
  {
#   ifndef INTEL
    msg = "MAJOR PROBLEM: little-endian host linked with big-endian library!";
    feclog(msg);
    if (dieFunction != NULL) dieFunction(msg);
    printf("%s\n",msg);
    if (gDieAnotherDay) { gDieAnotherDay = FALSE; return; }
    exit(1);
#   else
    msg = "host is little endian";
    feclog("%s %s",OS_TAG,msg);
    gEndianness = BO_LITTLE_ENDIAN;
#   endif
  }
  else
  {
#   ifdef INTEL
    msg = "MAJOR PROBLEM: big-endian host linked with link-endian library!";
    if (dieFunction != NULL) dieFunction(msg);
    printf("%s\n",msg);
    if (gDieAnotherDay) { gDieAnotherDay = FALSE; return; }
    exit(1);
#   else
    msg = "host is big endian";
    feclog("%s %s",OS_TAG,msg);
    gEndianness = BO_BIG_ENDIAN;
#   endif
  }
}
#ifdef DEBUG_UTILS
#ifdef WIN32
#include "Psapi.h"
#endif
size_t GetSystemCurrentRSS()
{
# if defined(WIN32)
	PROCESS_MEMORY_COUNTERS info;
	GetProcessMemoryInfo( GetCurrentProcess( ), &info, sizeof(info) );
	return (size_t)info.WorkingSetSize;
#elif defined(MACOS)
	struct mach_task_basic_info info;
	mach_msg_type_number_t infoCount = MACH_TASK_BASIC_INFO_COUNT;
	if ( task_info( mach_task_self( ), MACH_TASK_BASIC_INFO,
		(task_info_t)&info, &infoCount ) != KERN_SUCCESS )
		return (size_t)0L;		/* Can't access? */
	return (size_t)info.resident_size;
#elif defined(UNIX)
	long rss = 0L;
	FILE* fp = NULL;
	if ( (fp = fopen( "/proc/self/statm", "r" )) == NULL )
		return (size_t)0L;		/* Can't open? */
	if ( fscanf( fp, "%*s%ld", &rss ) != 1 )
	{
		fclose( fp );
		return (size_t)0L;		/* Can't read? */
	}
	fclose( fp );
	return (size_t)rss * (size_t)sysconf( _SC_PAGESIZE);
#else
	return (size_t)0L;			/* Unsupported. */
#endif
}
#endif
int gSystemDataStamp = 0;
int gSystemDataStampRb = 0;
int gSystemDataStampDly = 0;
struct timeval gSystemDataStampTs = {0, 0};
int SystemRunning = FALSE;
int EquipmentModulesInitialized = FALSE;
TINE_EXPORT int gRetardSingleContractRemoval = TRUE;
SINT32 findpoll(SINT32 p1,SINT32 p2)
{
  SINT32 p3;

  if (p1 < MinPollingRate) return p2 <= MinPollingRate ? MinPollingRate : p2;
  if (p2 < MinPollingRate || p1 == p2) return p1;
  if (p1 <= 1) return p1; 
  if (p2 <= 1) return p2;
  if (p2 < p1) while ((p3 = p1%p2) >= MinPollingRate) { p1 = p2; p2=p3; }
  else if (p1 < p2) while ((p3 = p2%p1) >= MinPollingRate) { p2 = p1; p1=p3; }
  return p1 > p2 ? p2 : p1;
}
ExportListStruct *ExportList = NULL;
#ifdef MULTITHREADED
THRHANDLE sendStreamCreateThread(ClnHdr *cln,BYTE *stream,UINT32 msgsize);
#endif
TINE_EXPORT int ServerIsInitialized(const char *eqm)
{
  ExportListStruct *el = getExportListItem((char *)eqm);
  if (el == NULL) return non_existent_elem;
  return el->inidone;
}
void InitEqpModules(void)
{
  ExportListStruct *el;

  feclog("startup initialization of equipment modules");
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->inidone) 
    {
      feclog("%.6s : already initialized",el->EqmName);
      continue;
    }
    memset(&el->accessLock,0,sizeof(AccessLock));
    if (el->eqmIni) TSKCALL(el,eqmIni);
    rmvAlarmsFromCAS(el->EqmName);
#   if defined(VXWORKS)
    if (!SystemRunning)
    { /* cannot risk a network resources problem reading remote files ! */
#   endif
    historyInit(el->EqmExportName,el->EqmName);
    GetAlarmWatchTable(el->EqmName);
    getAllowedUsers(el->EqmName);
    getAllowedNets(el->EqmName);
    getFollowersFromFile(el->EqmName);
#   if defined(VXWORKS)
    }
#   endif
#   ifdef MULTITHREADED
    eqmInitMutexSet((void *)el);
#   endif
    el->isidle = gEqmInitializedIdleState;
    el->inidone = TRUE;
    el->eqmBaseline = el->eqmStartupTime = time(NULL);
    setFailoverConfigInExportList(el, NULL, NULL);
    feclog("%.6s marked as initialized (idle state : %s)",el->EqmName,el->isidle ? "TRUE" : "FALSE");
  }
  return;
}
void devNamesInitialization(void)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int plen, idx;
  char prpName[64];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->hasPropertyDeviceNames)
    {
      for (idx=0; idx<PROPERTY_HASH_SIZE; idx++)
        for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
      {
        if (prp->isReserved) continue;
        if ((plen=(int)strlen(prp->prpName)) < 5) continue;
        if (strcmp(&prp->prpName[plen-4],".NAM")) continue;
        strncpy(prpName,prp->prpName,(size_t)(plen-4)); prpName[plen-4] = 0;
        AssignDeviceListFromProperty(el->EqmName,prpName);
      }
    }
  }
  acquireDeviceNamesFromProperties = FALSE;
}
int RestoreStockValues(ExportListStruct *el)
{
  int cc = 0;
  cc = RestoreStockPropertyValues(el,"ZPOSITION");
  cc = RestoreStockPropertyValues(el,"DEVMASK");
  cc = RestoreStockPropertyValues(el,"DEVREGION");
  cc = RestoreStockPropertyValues(el,"DEVLOCATION");
  cc = RestoreStockPropertyValues(el,"DEVDESCRIPTION");
  return cc;
}
int gExportListEmptyInitialization = FALSE;
int LastMinuteInitialization(void)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int idx;

  if (ExportList == NULL) 
  { /* e.g. SystemInit called with no configuation or registration */
    gExportListEmptyInitialization = TRUE;
    feclogEx(TINE_LOGLEVEL_WARN, "initializing with empty export list!");
  }
  else
  { /* the 'normal' state of affairs ... */
    gExportListEmptyInitialization = FALSE;
  }
  InitEqpModules();
  EquipmentModulesInitialized = TRUE;
# ifdef VXWORKS
  feclog("cycler running in task %s, current working directory : %s",
    SystemGetStartupCommand(),SystemGetStartupDirectory());
# endif
  if (gDelayInitTineServices) 
  {
    initTineServices();
  }
  else
  { /* in case a hook came in late ... */
    if (cdiSystemInitHook != 0) cdiSystemInitHook(); /* if CDI is in use ... */
  }

  for (el=ExportList; el != NULL; el=el->next)
  {
#   ifndef SMALL_TINE_LIB
    applyStoredPrpValues(el);
    applyRegionRules(el);
    RestoreStockValues(el);
    if (el->failoverType == FAILOVER_SLAVE) startFailoverMonitor(el->EqmName);
#   endif
    if (gStartupDebug)
    {
      int fgtty = foregroundTTY;
      foregroundTTY = TRUE;
      ttyoutput("Exporting the following properties for %.32s (%.6s):",el->EqmExportName,el->EqmName);
      for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
        for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
      {
        ttyoutput("\t%.64s (%s) -> %.64s <-",prp->prpName,itom(prp->prpAccessMode),prp->prpDescription);
      }
      foregroundTTY = fgtty;
    }
  }
  /* following eqm initialization flag the system as running */
  feclog("FEC: system marked as running");
  SystemRunning = TRUE; 
  return 1;  /* init done */
}
void dumpExportManifests(void)
{
# ifndef FS_RDONLY
  ExportListStruct *el;
  if (!strcmp(gFecName, GENS_FEC) || !strcmp(gFecName, ENS_FEC)) return;
  for (el = ExportList; el != NULL; el = el->next)
  {
    dumpDevicesManifest(el->EqmName, gStartupDebug);
    dumpPropertiesManifest(el->EqmName, gStartupDebug);
  }
# endif /* FS_RDONLY */
}
int RegisterLocalServiceModule(void)
{
  ExportListStruct *el;
  int cc;
  if ((cc=RegisterEquipmentModule(SRVEXP_NAME,SRVEQM_NAME,0,NULL,NULL,NULL,0,NULL)) != 0) return cc;
  if ((el=getExportListItem(SRVEQM_NAME)) == NULL) return non_existent_elem;
  el->registered = REGISTERED_ALL; /* don't allow this to be sent to the ENS ! */
  return 0;
}
void ExitEQPModules(void)
{
  ExportListStruct *el;
  gIsExitingEQPModules = TRUE;
  for (el=ExportList; el != NULL; el=el->next)
  {
    el->isterminating = TRUE;
    feclog("stopping equipment module %.6s",el->EqmName);
#   ifdef MULTITHREADED
    if (el->hEqmBkg != 0) millisleep(el->rateBkg);
    millisleep(gSystemTick);
    /* the steps above should be enough, but just in case: */
    if (el->hEqmBkg != 0) StopThread(el->hEqmBkg);
#   endif
    if (el->eqmExi) TSKCALL(el,eqmExi);  /* call exit routines */
  }
  gIsExitingEQPModules = FALSE;
  return;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int SetNumberDataOutputBuffers(int value)
{
# ifndef MULTITHREADED
  if (value < 1) value = 1;
  if (value > 2) value = 2;
# else
  value = 2;
# endif
  nDataOutBuffers = value;
  return nDataOutBuffers;
}
TINE_EXPORT int GetNumberDataOutputBuffers(void) { return nDataOutBuffers; }
TINE_EXPORT void FreeEQPModules(void)
{
  ExportListStruct *el=ExportList;
  int cc;
  if (WaitForSystemMutex(hSystemKernelMutex,-1) != 0) return;
  ExitEQPModules();
  if (ServerExitCondition == 0) ServerExitCondition = 1;
  if (el != NULL) dumpMsgTableToFile();
  _SystemReset(0);
  while (el != NULL)
  {
    feclog("freeing equipment module memory for %.6s",el->EqmName);
    if ((cc=RemoveEquipmentModule(el->EqmName)) != 0)
      feclogEx(TINE_LOGLEVEL_ERROR,"could not free EQM %.6s : %.32s",el->EqmName,cc2str(cc));
    el = ExportList;
  }
  ExportList = NULL;
  freeHstMemory();
  freeAlmMemory();
  freeSrPrpList();
  gEqmListSize = 0;
  gHstMfDumped = FALSE;
  gAlmMfDumped = FALSE;
  ReleaseSystemMutex(hSystemKernelMutex);
  feclog("equipment module memory returned to heap");
  return;
}
#endif /* SMALL_TINE_LIB */
int gAllowBkgTaskReentrancy = FALSE;
/**
 * \internal not relevant in multi-threaded builds (task is on its own thread)
 */
void checkTasklist(void)
{
  SINT32 msec;
  struct timeval tv;
  ExportListStruct *el;

  gettimeofday(&tv,(struct timezone *)NULL);

  for (el=ExportList; el != NULL; el=el->next) 
  {
    if (el->eqmBkg)
    {
      if (el->isidle) continue;
      if ((msec=MSECS(tv,el->lastBkg)) < el->rateBkg) continue;
      if (msec == 0 && el->rateBkg != 0) continue;
      if (!gAllowBkgTaskReentrancy) if (el->bkgbusy) continue;
      el->bkgbusy = TRUE;
      el->lastBkg = tv;
#     ifdef MULTITHREADED
      if (gSystemUseDataProtection) WaitForMutex(el->hDataProtectionMutex,-1);
#     endif
      TSKCALL(el,eqmBkg);
#     ifdef MULTITHREADED
      if (gSystemUseDataProtection) ReleaseSystemMutex(el->hDataProtectionMutex);
#     endif
      el->bkgbusy = FALSE;
      if (el->next != NULL) _SystemEngine();
    }
  }
  return;
}
int assertRedirectionValid(char *tag,char *eqm,char *rdr)
{
  char *ctx = &rdr[PRP_REDIR_CTX_OFFSET];
  char *srv = &rdr[PRP_REDIR_SRV_OFFSET];
  ExportListStruct *el;
  
  if (srv[0] == 0 && ctx[0] == 0) 
  { /* must be given */
    feclogEx(TINE_LOGLEVEL_ERROR,"[%s] redirection missing context and server name  : rejected !",tag);
    return FALSE; 
  }
  if (ctx[0] == 0) strncpy(ctx,gDeviceContext,CONTEXT_NAME_SIZE);
  for (el=ExportList; el != NULL; el=el->next)
  { /* redirecting to myself ? */
    if (strnicmp(ctx,el->EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(srv,el->EqmExportName,EXPORT_NAME_SIZE)) continue;
    if (eqm != NULL && strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    feclogEx(TINE_LOGLEVEL_ERROR, "[%s] redirection to /%.32s/%.32s is the local process : rejected !",tag,ctx,srv);
    return FALSE; 
  }
  return TRUE;
}
void lockToExclusiveRead(ClnHdr *cln)
{
  ContractListStruct *cl; /* improve readability by assignment */
  ClnLst *c;
  int i;
  for (i=0; i<ncontracts; i++)
  {
    if ((cl=ContractList[i])->expired == CONTRACT_AVAILABLE) continue;
    if (isExclusiveProperty(cl->exp,cl->contract.EqmProperty))
    {
      for (c=cl->clnLst; c != NULL; c=c->next)
      {
        if (c->client != cln) c->status.counter = 1;
      }
    }
  }
}

#define codeCarriesStatus(cc) ((cc) == server_redirection || (cc) == has_bitfield_tag || (cc) == has_structure_tag)
ClnHdr gLastAllowedCln = CLNHDR_STRUCT_INITIALIZER;
int hasPrpAcls = FALSE;
char aclStsErr[STATUS_SIZE];
int chkClnCredentials(ClnHdr *cln,ExportListStruct *el,ExportPropertyListStruct *prp,char *prpName,char *devName,int logit)
{ /* in case of meta-properties, prpName could be different from the prp pointer */
  int chkPrpAcl=FALSE,chkDevAcl=FALSE;
  char addr[64];
  if (el == NULL) return 0;
  memset(aclStsErr,0,STATUS_SIZE);
  if (prp != NULL && (prp->aclLst.nUsrs+prp->aclLst.nGrps+prp->aclLst.nNets) > 0) chkPrpAcl = TRUE;
  if (el->chkAcl & (DEV_CHECK_USR|DEV_CHECK_NET|DEV_CHECK_GRP)) chkDevAcl = TRUE;
  if (!memcmp(&gLastAllowedCln,cln,CLNHDR_CMP_SIZE) && !chkPrpAcl && !chkDevAcl)
  { /* cut to the chase */
    goto cleared; 
  }
  if ((el->chkAcl & EQM_CHECK_USR) && !isAllowedUser(el->EqmName,cln->usr))
  {
    snprintf(aclStsErr,STATUS_SIZE,"access denied: user %.16s refused access",cln->usr);
    if (logit) feclogEx(TINE_LOGLEVEL_INFO, "COMMAND (%.6s)/%.64s %.64s called by %.16s : user refused access",
                  el->EqmName,devName,prpName,cln->usr);
    return access_denied;
  }
# ifndef SMALL_TINE_LIB
  if (!isAllowedPropertyAccess(el->EqmName,prpName,cln))
  {
    snprintf(aclStsErr,STATUS_SIZE,"access denied: user %.16s property %.64s refused access",cln->usr,prpName);
    if (logit) feclogEx(TINE_LOGLEVEL_INFO, "COMMAND (%.6s)/%.64s %.64s called by %.16s : property refused access",
                  el->EqmName,devName,prpName,cln->usr);
    return access_denied;
  }
  if (!isAllowedDeviceAccess(el->EqmName,devName,cln))
  {
    snprintf(aclStsErr,STATUS_SIZE,"access denied: user %.16s device %.64s refused access",cln->usr,devName);
    if (logit) feclogEx(TINE_LOGLEVEL_INFO, "COMMAND (%.6s)/%.64s %.64s called by %.16s : device refused access",
                  el->EqmName,devName,prpName,cln->usr);
    return access_denied;
  }
# endif
  if ((el->chkAcl & EQM_CHECK_NET) && !IsMemberControlNets(el,cln)) 
  {
    inetStrAdr(&cln->adr.ip,addr,TRUE);
    snprintf(aclStsErr,STATUS_SIZE,"access denied: address %.64s refused access",addr);
    if (logit) 
    {
      feclogEx(TINE_LOGLEVEL_INFO, "COMMAND (%.6s)/%.64s %.64s called by %.16s from addr %.64s: network refused access",
             el->EqmName,devName,prpName,cln->usr,addr);
    }
    return access_denied;
  }
  if (strcmp("ACCESSLOCK",prpName) && !isMemberLockSet(el,cln))
  {
    inetStrAdr(&cln->adr.ip,addr,TRUE);
    snprintf(aclStsErr,STATUS_SIZE,"access denied: user %.16s address %.64s access lock refused access",cln->usr,addr);
    if (logit) 
    {
      feclogEx(TINE_LOGLEVEL_INFO, "COMMAND (%.6s)/%.64s %.64s called by %.16s from addr %.64s: access lock refused access",
                  el->EqmName,devName,prpName,cln->usr,addr);
    }
    return access_denied;
  }
  memcpy(&gLastAllowedCln,cln,sizeof(ClnHdr));
cleared:
  if (putCommandsInFeclog && !el->accessLock.lockType && logit) 
  { /* log it if this isn't a lock token renewal */
    inetStrAdr(&cln->adr.ip,addr,TRUE);
    feclog("COMMAND (%.6s)/%.64s[%.64s] called by %.16s from %.64s",
       el->EqmName,devName,prpName,cln->usr,addr);
  }
  return 0;
}
int isPastDeadband(char *eqm,char *prp,short access,ClnHdr *cln)
{
  PrpDbaItem *dba = getPrpDbaItem(eqm,prp);
  if (dba != NULL)
  { /* found something ... */
    if (!(dba->access & access)) return 0;
    if (MSECS(thisCycleTime,dba->lastAccess) < dba->deadband) return operation_busy;
    if (strncmp(dba->usr,cln->usr,USERNAME_SIZE) || memcmp(&dba->addr,&cln->adr.ip,sizeof(SCKADR)))
    { /* different client as last access: check against MAX_LOCK_DURATION */
      if (thisCycleTime.tv_sec - dba->lastAccess.tv_sec < MAX_LOCK_DURATION) 
        return access_locked;
    }
    strncpy(dba->usr,cln->usr,USERNAME_SIZE);
    dba->addr = cln->adr.ip;
    dba->lastAccess = thisCycleTime;
  }
  return 0;
}
int IsIllegalProperty(char *eqmName,char *prpName,char *dev,ClnHdr *cln,short access,
                      ExportListStruct **expLstElem,ExportPropertyListStruct **prpLstElem,char *stsString)
{
  int i,plen=0,ps,isXPrp=FALSE,chkAccess,logit=TRUE,isMeta=FALSE;
  short prpType,cc=0,rc;
  ExportPropertyListStruct *prp=NULL;
  ExportListStruct *el;
  char *bpstr,*mpstr,basePrpName[PROPERTY_NAME_SIZE],meta[16];

  stsString[0] = 0; meta[0] = 0;
  strncpy(basePrpName,prpName,PROPERTY_NAME_SIZE);
  if (IsMetaProperty(prpName,&bpstr,&mpstr,FALSE))
  {
    isMeta = TRUE;
    strncpy(meta,mpstr,15);
    strncpy(basePrpName,bpstr,48);
    if ((i=getAliasTableIndex(bpstr)) >= 0) 
    { /* meta properties based on property alias */
      sprintf(prpName,"%.48s.%.15s",gAliasTable[i].name,mpstr);
      strncpy(basePrpName,gAliasTable[i].name,PROPERTY_NAME_SIZE);
    }
  }
  prpType = EMPTY_PROP;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqmName,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL)
  { /* came up empty, try the alias name */
    for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqmName,el->EqmAlias,EQM_NAME_SHORTSIZE)) break;
  }
  if (el == NULL) 
  { /* one last chance: could be a FEC property */
    if (IsFecProperty(prpName)) ccerr(0);
    ccerr(non_existent_elem);
  }
  if (el->inidone == FALSE) ccerr(not_initialized);
  if (!HAS_PRPQRYFCN(el))
  { /* use default property list */
    if ((rc=lookupPropertyName(el,prpName,dev,&prp)) == server_redirection) ccerr(rc);
    if (rc == has_bitfield_tag || rc == has_structure_tag) ccerr(rc);
    if (rc != 0 && (rc=lookupPropertyName(el,basePrpName,dev,&prp)) != 0) 
    { /* failed first test */
      if (rc == server_redirection) 
      { /* a meta property needs to be redirected */
        strcat(&el->EqmLastErrorString[PRP_REDIR_PRP_OFFSET],".");
        strncat(&el->EqmLastErrorString[PRP_REDIR_PRP_OFFSET],meta,8);
        ccerr(rc);
      }
      for (prp=el->EqmPrpLst[PROPERTY_HASH_SIZE]; prp; prp=prp->next)
      { /* check for 'registered' wild cards (e.g. "PROP*" accepts any property beginning with "PROP") */
        for (ps = 0; ps < (int)strlen(prp->prpName) && prp->prpName[ps] == '*'; ps++);
        plen = (ps < (int)strlen(prp->prpName)) ? (int)strcspn(&prp->prpName[ps],"*") : 0;
        if (plen && !strnicmp(&basePrpName[ps],&prp->prpName[ps],plen)) break;
      }
      if (prp == NULL) 
      { /* Maybe it's a query */
        if ((prpType=IsStockProperty(basePrpName)) != STOCK_PROP) ccerr(illegal_property);
      }
      else prpType = WILDCARD_PROP;
    }
    else 
    { /*  is registered normally */
      prpType = TARGET_PROP;
    }
  }
  else  
  { /* uses a property query function, but check a few things up front ... */
    prpType = IsStockProperty(basePrpName);
    if (!isWildcardContract(dev))
    { /* not a wildcard request */
      if (lookupPropertyName(el,prpName,dev,&prp) == server_redirection) ccerr(server_redirection);
      if (strnicmp(prpName,basePrpName,PROPERTY_NAME_SIZE) && 
          lookupPropertyName(el,basePrpName,dev,&prp) == server_redirection) 
      {
        strcat(&el->EqmLastErrorString[PRP_REDIR_PRP_OFFSET],".");
        strncat(&el->EqmLastErrorString[PRP_REDIR_PRP_OFFSET],meta,8);
        ccerr(server_redirection);
      }
      if (prp != NULL)
      {
        if ((prp->prpAccessMode & CA_NETWORK) && (access & CA_NETWORK) != CA_NETWORK)
          ccerr(mcast_access_required);
        if (cln->inetProtocol != MMF && (prp->prpAccessMode & CA_CONNECT) && (access & CA_CONNECT) != CA_CONNECT)
          ccerr(tcp_access_required);
      }
      if (lookupRedirectionNameStub != NULL)
      {
        if ((cc=lookupRedirectionNameStub(el->EqmName,prpName,dev)) == server_redirection) ccerr(server_redirection);
        if (strnicmp(prpName,basePrpName,PROPERTY_NAME_SIZE) != 0)
        { /* prpName was a META-property ? */
          cc = lookupRedirectionNameStub(el->EqmName,basePrpName,dev);
        }
        if (cc < 0) cc = software_error; /* improper return code from the stub */
        if (cc != 0) ccerr(cc); /* likely to be server_redirection */
      }
    }
  }
  chkAccess = access&CA_WRITE ? TRUE : FALSE;
  if (chkAccess)
  {
    if (isWildcardContract(dev) || isWildcardContract(basePrpName)) return not_allowed;
  }
  isXPrp = isExclusiveProperty(el,basePrpName);
  if (isXPrp & PRP_XREAD_ALL) { chkAccess = TRUE; logit = FALSE; }
  if (cln == NULL) chkAccess = FALSE;
  if (strcmp("MESSAGE",basePrpName) == 0) chkAccess = FALSE;
  if (chkAccess) 
  {
    if ((cc=chkClnCredentials(cln,el,prp,basePrpName,dev,logit)) != 0) ccerr(cc);
  }
  if ((el->accessLock.lockType) && !isMemberLockSet(el,cln) && isXPrp)
  {
    ccerr(access_locked);
  }
  if (prp != NULL && prp->minAccessDeadband > 0)
  { /* gotta check the deadband */
    if ((cc=isPastDeadband(eqmName,prpName,access,cln)) != 0) ccerr(cc);
  }
err:
  if (codeCarriesStatus(cc) && el != NULL)
  {
    memcpy(stsString,el->EqmLastErrorString,STATUS_SIZE);
  }
  if (expLstElem != NULL) *expLstElem = el;
  if (prp != NULL && prpLstElem != NULL) 
  { /* this might come up empty if property-query function is in use */
    if (!gOneTimeSyncPass && !isMeta)
    { /* only the globals server will allow this pass */
      if (prp->prpAccessMode & CA_NETWORK)
      { /* property requires CA_NETWORK access */
        if ((access & CA_NETWORK) == 0) cc = mcast_access_required;
      }
      if (prp->prpAccessMode & CA_NOSYNC)
      { /* property requires asynchronous access */
        if ((access & CA_SINGLE) != 0) cc = async_access_required;
      }
      if (cln->inetProtocol != MMF && prp->prpAccessMode & CA_CONNECT)
      { /* property requires CA_CONNECT access */
        if ((access & CA_CONNECT) == 0) cc = tcp_access_required;
      }
    }
    if (isMeta && usesGateFunction(meta)) isMeta = FALSE;
    *prpLstElem = prp; /* tbs20160805: set this always - also when is meta property */
    if (!isMeta)
    {
      if (prp->sigfcn != NULL) sendPrpSignal(prp,CS_ACCESS,-1,cc);
    }
    prp->numCalls++;
  }
  if (gConSigfcn != NULL)
  { /* general signal */
    sendContractSignal(eqmName,dev,prpName,CS_ACCESS,cc);
  }
  return cc;
}
int gEnforceMcaAccess = TRUE;
int IsMcaElement(CONTRACT *con,ExportPropertyListStruct *prp,int *index,int *len,char **dev)
{
  int idx;
  char *devn;
  ExportListStruct *el;
  DeviceMcaGroup *mcaGrp;
  if (!gEnforceMcaAccess) return FALSE;
  if (con == NULL || prp == NULL || index == NULL || len == NULL) return FALSE;
  if (con->EqmSizeOut != 1) return FALSE;
  if (con->EqmAccess & CA_WRITE) return FALSE;
  if (prp->isReserved) return FALSE;
  if (!(prp->prpArrayType & AT_CHANNEL)) return FALSE;
  idx = GetValidDeviceNumber(con->EqmName,con->EqmDeviceName,con->EqmProperty,-1);
  if (idx < 0 || prp->prpSize < 1) return FALSE;
  if (prp->prpSize < 2) 
  { /* claims to be channel array but size == 1 -> try group info */
    if ((el=getExportListItem(con->EqmName)) == NULL || el->EqmDevLst == NULL) return FALSE;
    if ((mcaGrp=el->EqmDevLst[idx].mcaGrp) == NULL) return FALSE;
    if (mcaGrp->grpSize < 2) return FALSE;
    if (strlen(mcaGrp->grpName) == 0) return FALSE;
    *index = mcaGrp->grpIndex;
    *len = mcaGrp->grpSize;
    if (dev != NULL) *dev = mcaGrp->grpName;
    return TRUE;
  }
  *index = idx;
  *len = prp->prpSize;
  if ((devn=getFirstDevice(con->EqmName,con->EqmProperty)) == NULL) devn = "#0";
  if (dev != NULL) *dev = devn;
  return TRUE;
}
int renegotiateContract(ContractListStruct *cl,int cc)
{
  ClnLst *c;
  UINT32 doutsize = 0,retsize,fmtoutsize;
  int addext = FALSE;
  short fmt = (short)LFMT(cl->contract.EqmFormatOut);
  if (cl->deliveryPending)
  { /* mega-unlikely, but just in case: bail out here */
    return cl->dataSizeOut;
  }
  cc &= 0xfff; /* strip off info bits */
  if (cc == not_posted || cc == not_signalled || cc == not_ready) cc = 0;
  if ((fmtoutsize=fmtsizeof(cl->contract.EqmFormatOut)) != 0)
  {
    if (isAdjustableLengthFormat(fmt))
    {
      doutsize = cl->dataSizeOut;
    }
    else
    {
      if (fmt == CF_MDA) fmtoutsize = getCarriedFormatSize(cl->contract.strTagOut);
      doutsize = cl->dataSizeOut*fmtoutsize;
      doutsize += getFormatHeaderSize(fmt);
    }
    if (doutsize > cl->dataOutCapacity)
    { /* this would be a problem ! */
      dbglog("%.6s/%.64s[%.64s] attempted to increase output past capacity %d vs. %d",
        cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,
        doutsize,cl->dataOutCapacity);
      if (doutsize == cl->dataSizeOut) cl->dataSizeOut = cl->dataOutCapacity;
      doutsize = cl->dataOutCapacity;
    }
    if (cc) addext = TRUE;
  }
  retsize = doutsize;
  for (c=cl->clnLst; c != NULL; c=c->next)
  {
    if (addext) doutsize += STSSIZE(c->client->tineProtocol);
    c->status.bytesRemaining = (UINT32)doutsize;
    c->status.bytesTotal = (UINT32)doutsize;
    c->status.nextDataSize = c->status.bytesRemaining;
    if (!isJumboDelivery(c->client))
    { /* normal case */
      if (c->status.bytesRemaining > c->status.mtu) c->status.nextDataSize = c->status.mtu;
      c->status.numblks = (UINT16)(doutsize/c->status.mtu + (doutsize%c->status.mtu ? 1 : 0));
      if (doutsize == 0) c->status.numblks = 1;
    }
    doutsize = retsize; /* in case addext is TRUE */
  }
  return retsize;
}
void gateOutputData(BYTE *data,int size,short format,UINT32 gate,int outputBoolean)
{
  int i;
  UINT16 *sval = (UINT16 *)data;
  UINT32 *ival = (UINT32 *)data;
  if (data == NULL) return;
  switch (format)
  {
    case CF_BYTE:
      for (i=0; i<size; i++) 
      {
        data[i] &= gate;
        if (outputBoolean) data[i] = data[i] == gate ? 1 : 0;
      }
      break;
    case CF_INT16:
    case CF_UINT16:
      for (i=0; i<size; i++)
      {
        sval[i] &= gate;
        if (outputBoolean) sval[i] = sval[i] == gate ? 1 : 0;
      }
      break;
    case CF_INT32:
    case CF_UINT32:
      for (i=0; i<size; i++)
      {
        ival[i] &= gate;
        if (outputBoolean) ival[i] = ival[i] == gate ? 1 : 0;
      }
      break;
    default:
      return;
  }
}
int gEqmFcnContractTableId = -1;
int gFwdWildcardDeviceCalls = FALSE;
int gFwdWildcardPropertyCalls = FALSE;
TINE_EXPORT void SetForwardWildcardCalls(int type,int value)
{
  switch (type)
  {
    case TYPE_PRP: gFwdWildcardPropertyCalls = value; break;
    case TYPE_DEV: gFwdWildcardDeviceCalls = value; break;
    default: break;
  }
}
int fwdWildCardCall(int type,int fmt)
{
  if (!isSimpleFormat((short)fmt) && fmt != CF_USTRING) return FALSE;
  switch (type)
  {
    case TYPE_PRP: return gFwdWildcardPropertyCalls;
    case TYPE_DEV: return gFwdWildcardDeviceCalls;
  }
  return FALSE;
}
int EqmFcnWithWildcardDevice(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access)
{ /* arguments are never NULL */
  DTYPE drdb, dnull;
  char sDevName[DEVICE_NAME_SIZE];
  int start = 0, len = 0, isCont = FALSE, isNotReady = FALSE, chkMask = FALSE, chkOffline = FALSE;
  int cc = not_allowed, rc, i, k;
  int fmt, fmtlen, nmlen, noff, voff, soff;
  UINT32 devlistsize = 0;
  ExportPropertyListStruct *prp = GetPropertyListStruct(el->EqmName,DevProperty,DevName);
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;
  UINT32 mask = 0x8000000;
  NAME64 *devlist=NULL;
  BYTE *rb;
  float *frb;
  UINT32 *lrb;
  NAME64 *nrb = NULL;
  double wcTimestamp = 0;
  EQMFCNP hndlr= el->eqmFcn;
  void *ref = el->EqmRef;
# if !defined(SMALL_TINE_LIB)
  UINT32 nlen;
# endif
  if (prp == NULL) return illegal_property;
  if (access & CA_WRITE) return illegal_read_write;
  if (isSpectrumArray(el->EqmName,DevProperty)) return not_allowed;
  if (dout == NULL || dout->dArrayLength < 1) return not_allowed;
  if (!isLocal(el->EqmName,DevName,DevProperty)) return data_not_local;
  memset(&dnull,0,sizeof(DTYPE)); dnull.dFormat = CF_NULL;
# if defined(INTEL) && !defined(VXWORKS) && !defined(MACOS)
  rb = (BYTE *)alloca(dout->dArrayLength*sizeof(float));
# else
  rb = tmpWorkArea;
# endif
  if (prp->prpHndlr != NULL) 
  { /* re-wire the dispatch routine */
    hndlr = prp->prpHndlr;
    ref = prp->hdlrRef;
  }
  frb = (float *)rb; lrb = (UINT32 *)rb;
  if (prp->devNames != NULL)
  {
    devlist = prp->devNames;
    devlistsize = prp->numDevices;
  }
# if !defined(SMALL_TINE_LIB)
  else if (el->hasPropertyDeviceNames)
  {
    nlen = prp->prpSize;
    nrb = (NAME64 *)alloca(2*nlen*sizeof(NAME64)); 
    if (GetDeviceListFromProperty(el->EqmName,DevProperty,nrb,&nlen) == 0)
    {
      for (i=nlen-1; i>=0; i--) 
      {
        strncpy(nrb[i*2].name,nrb[i].name,DEVICE_NAME_SIZE);
        nrb[i*2].name[DEVICE_NAME_SIZE-1] = 0;
      }
      devlist = (NAME64 *)nrb;
      devlistsize = nlen;
    }
  }
# endif /* not SMALL_TINE_LIB */
  if (devlistsize == 0 && el->EqmDevLst != NULL)
  { /* normal state of affairs: use the registered device list */
    chkOffline = TRUE;
    devlist = (NAME64 *)el->EqmDevLst[0].name;
    devlistsize = el->EqmNumDevices - getNumReservedDevices(el);
    if (din->dArrayLength == 1 && din->dFormat == CF_INT32)
    { /* accept this as DEVMASK (SYS_MASK) input */
      mask = din->data.lptr[0];
      if (mask != 0) chkMask = TRUE;
    }
  }
  if ((prp->prpArrayType & AT_CHANNEL) == AT_CHANNEL)
  {
    len = devlistsize;
    isCont = isContiguous(el->EqmName,DevName,devlist,&start,&len);
  }
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      fmtlen = fmtsizeof(dout->dFormat);
      nmlen = fmtlen - 16;
      fmt = prp->prpFormat == CF_FLOAT || prp->prpFormat == CF_DOUBLE ? CF_FLOAT : CF_INT32;
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      if (isCont)
      { /* only a single call is required */
        drdb.data.fptr = frb;
        drdb.dArrayLength = MIN(dout->dArrayLength,(UINT32)len);
        strncpy(sDevName,devlist[start].name,DEVICE_NAME_SIZE);
        cc = EQMCALL(hndlr,sDevName,DevProperty,&drdb,&dnull,access,ref);
        if (cc != 0) return cc;
        wcTimestamp = gServerDataTimeStamp;
        for (i=0,k=0; i<(int)drdb.dArrayLength; i++)
        {
          if (chkMask && el->EqmDevLst[start+i].mask != 0 && (el->EqmDevLst[start+i].mask & mask) == 0) continue;
          if (chkOffline && (el->EqmDevLst[start+i].mask & DEVICE_OFFLINE)) continue;
          *(double *)(dout->data.bptr + k*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb[i] : (double)lrb[i];
          *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = 0;
          strncpy(dout->data.cptr + k*fmtlen,devlist[start+i].name,nmlen);
          k++;
        }
        dout->dArrayLength = (UINT32)k;
      }
      else
      {
        getWildcardTokens(DevName,beginToken,endToken,midToken);
        /* calculate these once outside the loop ... */
        blen = (int)strlen(beginToken);
        mlen = (int)strlen(midToken);
        elen = (int)strlen(endToken);
        for (i=0, k=0; i<(int)devlistsize && k<(int)dout->dArrayLength; i++)
        {
          if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;          
          if (chkMask && el->EqmDevLst[i].mask != 0 && (el->EqmDevLst[i].mask & mask) == 0) continue;
          if (chkOffline && (el->EqmDevLst[i].mask & DEVICE_OFFLINE)) continue;
          strncpy(sDevName,devlist[i].name,DEVICE_NAME_SIZE);
          drdb.data.fptr = &frb[k];
          drdb.dArrayLength = 1;
          cc = EQMCALL(hndlr,sDevName,DevProperty,&drdb,&dnull,access,ref);
          if (cc == illegal_format || cc == illegal_property) continue;
          if (cc == server_redirection) return data_not_local;
          if (cc == not_ready) isNotReady = TRUE;
          if (wcTimestamp < gServerDataTimeStamp) wcTimestamp = gServerDataTimeStamp;
          *(double *)(dout->data.bptr + k*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb[k] : (double)lrb[k];
          *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = (double)cc;
          strncpy(dout->data.cptr + k*fmtlen,devlist[start+i].name,nmlen);
          k++;
        }
        dout->dArrayLength = k;
      }
      if (wcTimestamp > 0) gServerDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      fmt = dout->dFormat == CF_NAME16II ? CF_INT32 : CF_FLOAT;
      fmtlen = fmtsizeof(dout->dFormat);
      if (dout->dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      if (isCont)
      {
        drdb.data.fptr = frb;
        drdb.dArrayLength = MIN(dout->dArrayLength,(UINT32)len);
        strncpy(sDevName,devlist[start].name,DEVICE_NAME_SIZE);
        cc = EQMCALL(hndlr,sDevName,DevProperty,&drdb,&dnull,access,ref);
        if (cc != 0) return cc;
        wcTimestamp = gServerDataTimeStamp;
        for (i=0,k=0; i<(int)drdb.dArrayLength; i++)
        {
          if (chkMask && el->EqmDevLst[start+i].mask != 0 && (el->EqmDevLst[start+i].mask & mask) == 0) continue;
          if (chkOffline && (el->EqmDevLst[start+i].mask & DEVICE_OFFLINE)) continue;
          memset(dout->data.bptr + k*fmtlen,0,fmtlen);
          memcpy(dout->data.bptr + k*fmtlen + voff,&frb[i],4);
          strncpy(dout->data.cptr + k*fmtlen + noff,devlist[start+i].name,MIN(nmlen,64));
          k++;
        }
        dout->dArrayLength = (UINT32)k;
      }
      else
      {
        getWildcardTokens(DevName,beginToken,endToken,midToken);
        /* calculate these once outside the loop ... */
        blen = (int)strlen(beginToken);
        mlen = (int)strlen(midToken);
        elen = (int)strlen(endToken);
        for (i=0, k=0; i<(int)devlistsize && k<(int)dout->dArrayLength; i++)
        {
          if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
          if (chkMask && el->EqmDevLst[i].mask != 0 && (el->EqmDevLst[i].mask & mask) == 0) continue;
          if (chkOffline && (el->EqmDevLst[i].mask & DEVICE_OFFLINE)) continue;
          strncpy(sDevName,devlist[i].name,DEVICE_NAME_SIZE);
          drdb.data.fptr = &frb[k];
          drdb.dArrayLength = 1;
          cc = EQMCALL(hndlr,sDevName,DevProperty,&drdb,&dnull,access,ref);
          if (cc == illegal_format || cc == illegal_property) continue;
          if (cc == server_redirection) return data_not_local;
          if (cc == not_ready) isNotReady = TRUE;
          if (wcTimestamp < gServerDataTimeStamp) wcTimestamp = gServerDataTimeStamp;
          memset(dout->data.bptr + k*fmtlen,0,fmtlen);
          memcpy(dout->data.bptr + k*fmtlen + voff,&frb[k],4);
          *(UINT32 *)(dout->data.bptr + k*fmtlen + soff) = (UINT32)cc;
          strncpy(dout->data.cptr + k*fmtlen + noff,devlist[i].name,nmlen);
          k++;
        }
        dout->dArrayLength = (UINT32)k;
      }
      if (wcTimestamp > 0) gServerDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    default:
      if (dout->dFormat <= CF_NAME32 && strcmp(DevName,"*") == 0)
      { /* primitive format allowed in this case (but no mask checking) */
        if (isCont) 
        {
          if (dout->dArrayLength > (UINT32)len) dout->dArrayLength = (UINT32)len;
          strncpy(sDevName,devlist[0].name,DEVICE_NAME_SIZE);
          return EQMCALL(hndlr,sDevName,DevProperty,dout,&dnull,access,ref);
        }
        drdb.dFormat = dout->dFormat;
        drdb.dTag[0] = 0;
        rc = not_posted;
        for (i=0; i<(int)devlistsize && i<(int)dout->dArrayLength; i++)
        { /* call one by one */
          strncpy(sDevName,devlist[i].name,DEVICE_NAME_SIZE);
          drdb.data.bptr = &dout->data.bptr[i*fmtsizeof(dout->dFormat)];
          drdb.dArrayLength = 1;
          if ((cc=EQMCALL(hndlr,sDevName,DevProperty,&drdb,&dnull,access,ref)) != 0)
          { /* not a success */
            switch (cc)
            {
              case not_ready: 
                isNotReady = TRUE; 
                break;
              case server_redirection: 
                return data_not_local;
              case illegal_property:
                continue;
              default:
                rc = cc;
                i = (int)devlistsize;
            }
          }
          else
          { /* at least one of the group was successful */
            rc = 0;
          }
          if (wcTimestamp < gServerDataTimeStamp) wcTimestamp = gServerDataTimeStamp;
        }
        if (rc == not_posted) rc = cc;
        dout->dArrayLength = i;
        if (wcTimestamp > 0) gServerDataTimeStamp = wcTimestamp;
        return isNotReady ? not_ready : rc;
      }
  }
  return cc;
}
int EqmFcnWithWildcardProperty(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access)
{
  DTYPE drdb;
  char sDevProp[PROPERTY_NAME_SIZE];
  int isNotReady = FALSE;
  int cc = not_allowed, i, k, devnr;
  int fmt, fmtlen, nmlen, noff, voff, soff;
  int prplistsize = 0, isLegacyPqs = FALSE;
  char beginToken[PROPERTY_NAME_SIZE], endToken[PROPERTY_NAME_SIZE], midToken[PROPERTY_NAME_SIZE];
  int blen = 0, elen = 0, mlen = 0;
  NAME64 *prplist=NULL;
  float frb;
  double wcTimestamp = 0;
  PrpQueryStruct *prpqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  EQMFCNP hndlr= el->eqmFcn;
  void *ref = el->EqmRef;
  ExportPropertyListStruct *pls = NULL;

  if (dout == NULL || dout->dArrayLength < 1) return not_allowed;
  if ((devnr=GetDeviceNumber(el->EqmName,DevName)) < 0) return illegal_equipment_number;

  if (el->EqmDevLst != NULL && el->EqmDevLst[devnr].prps != NULL)
  { /* a registered device property list ... */
    prplist = el->EqmDevLst[devnr].prps->prpList;
    prplistsize = (int)el->EqmDevLst[devnr].prps->prpListSize;
  }
# if !defined(SMALL_TINE_LIB)
  else if (HAS_PRPQRYFCN(el))
  { /* registered query fcn */
    if (el->EqmPropertyQueryFcn != NULL)
    { /* modern */
      prplistsize = el->EqmPropertyQueryFcn(DevName,&prpqs);
    }
    else
    { /* legacy */
      prplistsize = el->XEqmPropertyQueryFcn(DevName,&xpqs);
      isLegacyPqs = TRUE;
    }
    if (prplistsize < 0) return prplistsize;
    if (el->EqmWcPrpLstCapacity < (prplistsize * 2))
    { /* more properties registered => start over */
      SystemFree(el->EqmWcPrpLst);
    }
    if (el->EqmWcPrpLst == NULL)
    {
      if ((el->EqmWcPrpLst=(NAME64 *)SystemCalloc(prplistsize*2,sizeof(NAME64))) == NULL) return out_of_server_memory;
      el->EqmWcPrpLstCapacity = prplistsize * 2;
    }
    prplist = el->EqmWcPrpLst;
    if (!isLegacyPqs) for (i=0; i<prplistsize; i++) strncpy(prplist[i].name,prpqs[i].prpName,PROPERTY_NAME_SIZE);
    else for (i=0; i<prplistsize; i++) strncpy(prplist[i].name,xpqs[i].prpName,PROPERTY_NAME_SHORTSIZE);
  }
  else 
  {
    if (el->EqmWcPrpLstCapacity < el->EqmNumProperties)
    { /* more properties registered ? => start over */
      SystemFree(el->EqmWcPrpLst);
    }
    if (el->EqmWcPrpLst == NULL)
    {
      if ((el->EqmWcPrpLst=(NAME64 *)SystemCalloc(el->EqmNumProperties,sizeof(NAME64))) == NULL) return out_of_server_memory;
      el->EqmWcPrpLstCapacity = prplistsize = el->EqmNumProperties;
      if ((cc=GetRegisteredPropertyList(el->EqmName,el->EqmWcPrpLst,&prplistsize)) != 0) return cc;
    }
    if (dout->dArrayLength > (UINT32)el->EqmNumProperties) dout->dArrayLength = el->EqmNumProperties;
    prplistsize = el->EqmNumProperties;
    prplist = el->EqmWcPrpLst;
  }
# endif /* SMALL_TINE_LIB */
  if (prplistsize == 0) return illegal_property;
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      fmtlen = fmtsizeof(dout->dFormat);
      nmlen = fmtlen - 16;
      fmt = CF_FLOAT;
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      getWildcardTokens(DevProperty,beginToken,endToken,midToken);
      /* calculate these once outside the loop ... */
      blen = (int)strlen(beginToken);
      mlen = (int)strlen(midToken);
      elen = (int)strlen(endToken);
      for (i=0, k=0; i<prplistsize && k<(int)dout->dArrayLength; i++)
      {
        if (!wildcardMatches(prplist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
        strncpy(sDevProp,prplist[i].name,PROPERTY_NAME_SIZE);
        if (isSpectrumArray(el->EqmName,sDevProp)) continue;
        drdb.data.fptr = &frb;
        drdb.dArrayLength = 1;
        if ((pls=GetPropertyListStruct(el->EqmName,sDevProp,DevName)) == NULL) continue;
        if (pls->prpHndlr != NULL)
        { /* re-wire the dispatch routine */
          hndlr = pls->prpHndlr; ref = pls->hdlrRef;
        }
        else
        {
          hndlr= el->eqmFcn; ref = el->EqmRef;
        }
        cc = EQMCALL(hndlr,DevName,sDevProp,&drdb,din,access,ref);
        if (cc == illegal_format || cc == illegal_equipment_number || cc == dimension_error) continue;
        if (cc == not_ready) isNotReady = TRUE;
        if (wcTimestamp < gServerDataTimeStamp) wcTimestamp = gServerDataTimeStamp;
        *(double *)(dout->data.bptr + k*fmtlen + nmlen) = (double)frb;
        *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = (double)cc;
        strncpy(dout->data.cptr + k*fmtlen,prplist[i].name,nmlen);
        k++;
      }
      dout->dArrayLength = k;
      if (wcTimestamp > 0) gServerDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      fmtlen = fmtsizeof(dout->dFormat);
      fmt = dout->dFormat == CF_NAME16II ? CF_INT32 : CF_FLOAT;
      if (dout->dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      getWildcardTokens(DevProperty,beginToken,endToken,midToken);
      /* calculate these once outside the loop ... */
      blen = (int)strlen(beginToken);
      mlen = (int)strlen(midToken);
      elen = (int)strlen(endToken);
      for (i=0, k=0; i<prplistsize && k<(int)dout->dArrayLength; i++)
      {
        if (!wildcardMatches(prplist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
        strncpy(sDevProp,prplist[i].name,PROPERTY_NAME_SIZE);
        if (isSpectrumArray(el->EqmName,sDevProp)) continue;
        drdb.data.fptr = &frb;
        drdb.dArrayLength = 1;
        if ((pls=GetPropertyListStruct(el->EqmName,sDevProp,DevName)) == NULL) continue;
        if (pls->prpHndlr != NULL)
        { /* re-wire the dispatch routine */
          hndlr = pls->prpHndlr; ref = pls->hdlrRef;
        }
        else
        {
          hndlr= el->eqmFcn; ref = el->EqmRef;
        }
        cc = EQMCALL(hndlr,DevName,sDevProp,&drdb,din,access,ref);
        if (cc == illegal_format || cc == illegal_equipment_number || cc == dimension_error) continue;
        if (cc == not_ready) isNotReady = TRUE;
        if (wcTimestamp < gServerDataTimeStamp) wcTimestamp = gServerDataTimeStamp;
        memset(dout->data.bptr + k*fmtlen,0,fmtlen);
        memcpy(dout->data.bptr + k*fmtlen + voff,&frb,4);
        *(UINT32 *)(dout->data.bptr + k*fmtlen + soff) = (UINT32)cc;
        strncpy(dout->data.cptr + k*fmtlen + noff,prplist[i].name,nmlen);
        k++;
      }
      dout->dArrayLength = k;
      if (wcTimestamp > 0) gServerDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    default:
      break;
  }
  return cc;
}
int callStk(ContractListStruct *cl,char *dataOut,char *dataIn,int (*fcn)(ContractListStruct *,char *,char *))
{
  int cc=0,isWrite=(cl->contract.EqmAccess & CA_WRITE) ? TRUE : FALSE;
  
  cl->isStkCall = TRUE;
  if (isWrite)
  {
    if (gStockResourceBusy) return not_ready;
    if (WaitForSystemMutex(hStkAccessMutex,-1)) return not_ready;
    gStockResourceBusy = TRUE;
  }
  cl->dataSizeOut = cl->contract.EqmSizeOut;
  cc = (*fcn)(cl,dataIn,dataOut);
  assertHeapOkay("callStk",cl->contract.EqmDeviceName,cl->contract.EqmProperty);
  if (isWrite)
  {
    gStockResourceBusy = FALSE;
    ReleaseSystemMutex(hStkAccessMutex);
  }
  return cc;
}
int validateReturnCode(int cc)
{
  switch (cc)
  {
    case invalid_interval:
    case invalid_datarequest:
    case illegal_protocol:
    case get_subscription_id:
    case mcast_access_required:
    case tcp_access_required:
    case property_is_mca:
    case property_is_mca_elevated:
      return not_applicable;
    default:
      break;
  }
  return cc;
}
#define ALLOWED_BUSY_COUNTS 10
int callEqm(ContractListStruct *cl,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int cc, signal=CS_CALLED;
  ExportListStruct *el;
  int haveBusyMutex = FALSE;

  if (cl == NULL) return code_failure;
  if ((el=cl->exp) == NULL) return non_existent_elem;
  if (access & (CA_ALARM|CA_HIST|CA_QUERY)) 
  { /* called from one of the subsystems and NOT from doContract() */
    signal |= CS_PROCESSED;
  }
  if (access & CA_WRITE && !isIndependentWriteProperty(el,devProperty))
  { /* do not allow simultaneous WRITEs or possible WRITE and READ on same data set */
    if (el->isbusy) return not_ready;
    if (WaitForSystemMutex(el->hEqmBusyMutex,gSystemTick) != 0) 
    {
      if (tineDebug > 2) dbglog("can't get busy mutex (count : %d)",el->busycnt);
      return el->busycnt++ > ALLOWED_BUSY_COUNTS ? operation_busy : not_ready;
    }
    haveBusyMutex = TRUE;
    el->isbusy = TRUE;
    if (tineDebug > 2) dbglog("got eqm busy mutex");
  }
  if (tineDebug > 2) dbglog("%.256s called on thread: %u",getContractKey(NULL,cl->listId),cl->threadId);
# ifdef MULTITHREADED
  if (gSystemUseDataProtection) WaitForMutex(el->hDataProtectionMutex,-1);
# endif
  gEqmFcnContractTableId = cl->listId;
  el->isinside = TRUE;   
  
  if (isWildcardContract(devName) && !fwdWildCardCall(TYPE_DEV,dout->dFormat) &&
      !IsMetaProperty(devProperty,NULL,NULL,FALSE))
  {
    cc = EqmFcnWithWildcardDevice(el,devName,devProperty,dout,din,access);
  }
  else if (isWildcardContract(devProperty) && !fwdWildCardCall(TYPE_PRP,dout->dFormat))
  {
    cc = EqmFcnWithWildcardProperty(el,devName,devProperty,dout,din,access);
  }
  else
  {
    EQMFCNP hndlr = el->eqmFcn;
    void *ref = el->EqmRef;
    if (cl->prp != NULL && cl->prp->prpHndlr != NULL)
    {
      hndlr = cl->prp->prpHndlr;
      ref = cl->prp->hdlrRef;
    }
    /* din and dout are never NULL pointers !*/
    if (din->dFormat == CF_BOOLEAN) marshallBoolean(din->data.bptr, din->dArrayLength, RTOL);
    cc = EQMCALL(hndlr,devName,devProperty,dout,din,access,ref);
    if (dout->dFormat == CF_BOOLEAN) marshallBoolean(dout->data.bptr, dout->dArrayLength, LTOR);
    /* marshal the input back to original conditions (for contract comparisons) */
    if (din->dFormat == CF_BOOLEAN) marshallBoolean(din->data.bptr, din->dArrayLength, LTOR);
  }
  el->isinside = FALSE;
  assertHeapOkay("callEqm",devName,devProperty);
  if (gConSigfcn != NULL) 
  { /* general contract signal function is registered */
    sendContractSignal(el->EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,signal,cc);
  }
  if (cl->prp != NULL && cl->prp->sigfcn != NULL) 
  { /* property-specific signal handler is registered */
    sendPrpSignal(cl->prp,signal,cl->contractID,cc);
  }
  if (cc == not_signalled) cc = invalid_return_code;
# ifdef MULTITHREADED
  if (gSystemUseDataProtection) ReleaseSystemMutex(el->hDataProtectionMutex);
# endif
  if (access & CA_WRITE)
  {
    el->isbusy = FALSE; el->busycnt = 0;
    if (haveBusyMutex)
    {
      ReleaseSystemMutex(el->hEqmBusyMutex);
      if (tineDebug > 2) dbglog("release eqm busy mutex");
    }
  }
  cl->sysstamp = gSystemDataStamp;
  return validateReturnCode(cc);
}
#ifndef SMALL_TINE_LIB
int prepAdjustableArrayOutputPreCall(ContractListStruct *cl,DTYPE *dout)
{
  int tsiz,n = atoi(dout->dTag);
  if (n < 1) return dimension_error;
  /* remap the output object passed to the eqm handler */
  dout->dArrayLength = n;
  dout->dTag[0] = 0;
  switch (dout->dFormat)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      tsiz = sizeof(char *);
      break;
    case CF_AIMAGE:
    case CF_ASPECTRUM:
      tsiz = getFormatTypeSize(dout->dFormat);
      break;
    default:
      return illegal_format;
  }
  if (cl->dataOutArr == NULL)
  {
    if ((cl->dataOutArr=(BYTE *)SystemCalloc(n,tsiz)) == NULL) 
      return out_of_server_memory;
  }
  dout->data.vptr = cl->dataOutArr;
  return 0;
}
int prepAdjustableArrayInputPreCall(ContractListStruct *cl,DTYPE *din)
{
  int n = atoi(din->dTag);
  if (n < 1) return dimension_error;
  switch (din->dFormat)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      din->data.strptr = (char **)cl->dataPtr;
      break;
    case CF_AIMAGE:
    case CF_ASPECTRUM:
      /* nothing to do */
      break;
  }
  din->dArrayLength = n;
  din->dTag[0] = 0;
  return 0;
}
int prepAdjustedArrayOutputPostCall(ContractListStruct *cl,DTYPE *dout)
{
  int i, n, p=0, cc = 0;
  int hsiz, allocsize, xsiz=0, xoff, stdCapacity;
  char *sb, **sa = dout->data.strptr;
  BYTE *oldMemory=NULL;

  /* on entry: dArrayLength = number of items */
  n = (int)dout->dArrayLength;
  hsiz = getFormatHeaderSize(dout->dFormat);
  stdCapacity = getFormatStandardCapacity(dout->dFormat);
  switch (dout->dFormat)
  { /* first scan for the total adjusted space size */
    case CF_KEYVALUE:
    case CF_STRING:
      for (i=0; i<n; i++) xsiz += (sa[i] == NULL) ? 1 : (int)strlen(sa[i]) + 1;
      break;
    case CF_AIMAGE:
      for (i=0; i<n; i++) 
      {
        xsiz += hsiz + 2*sizeof(SINT32);
        xsiz += ((DIMAGE *)dout->data.vptr)[i].frameBufferSize;
      }
      break;
    case CF_ASPECTRUM:
      for (i=0; i<n; i++)
      {
        xsiz += hsiz + 2*sizeof(SINT32);
        xsiz += ((ASPECTRUM *)dout->data.vptr)[i].spectBufferSize * sizeof(float);
      }
      break;
  }
  /* check the required capacity ... */
  if (xsiz > (int)cl->dataOutCapacity)
  { /* returned string space has grown ! */
    if (!cl->unbuffered)
    { /* not managed from above so ... */
      oldMemory = cl->dataOut;
      /* add some extra bytes, as long as we're re-allocating ... */
      allocsize = xsiz + (stdCapacity - xsiz%stdCapacity);
      if ((cl->dataOut = (BYTE *)SystemCalloc(nDataOutBuffers,(size_t)allocsize)) == NULL)
        ccerr(out_of_server_memory);
      if (nDataOutBuffers > 1) cl->dataBuf = &cl->dataOut[allocsize];
      cl->dataOutCapacity = allocsize;
    }
  }
  /* move the returned data onto the contract */
  sb = (char *)cl->dataBuf;
  switch (dout->dFormat)
  { /* first scan for the total adjusted space size */
    case CF_KEYVALUE:
    case CF_STRING:
      for (i=0; i<n; i++)
      {
        if (sa[i] == NULL) sb[p] = 0; else strcpy(&sb[p],sa[i]);
        p += (int)strlen(&sb[p])+1;
      }
      break;
    case CF_AIMAGE:
      xoff = n * (hsiz + 8);
      for (i=0; i<n; i++)
      {
        memcpy(&sb[i*(hsiz+8)],&((DIMAGE *)dout->data.vptr)[i],hsiz);
        xsiz = ((DIMAGE *)dout->data.vptr)[i].frameBufferSize;
        memcpy(&sb[xoff],((DIMAGE *)dout->data.vptr)[i].frameBuffer,xsiz);
        memcpy(&sb[i*(hsiz+8)+hsiz],&xoff,4);
        memcpy(&sb[i*(hsiz+8)+hsiz+4],&xsiz,4);
        xoff += xsiz;
      }
      xsiz = xoff;
      break;
    case CF_ASPECTRUM:
      xoff = n * (hsiz + 8);
      for (i=0; i<n; i++)
      {
        memcpy(&sb[i*(hsiz+8)],&((ASPECTRUM *)dout->data.vptr)[i],hsiz);
        xsiz = ((ASPECTRUM *)dout->data.vptr)[i].spectBufferSize * sizeof(float);
        memcpy(&sb[xoff],((ASPECTRUM *)dout->data.vptr)[i].spectBuffer,xsiz);
        SwapBytes(&sb[xoff],xsiz/sizeof(float),CF_FLOAT,LTOR); /* swap the float array here */
        memcpy(&sb[i*(hsiz+8)+hsiz],&xoff,4);
        memcpy(&sb[i*(hsiz+8)+hsiz+4],&xsiz,4);
        xoff += xsiz;
      }
      xsiz = xoff;
      break;
  }
  SystemFree(oldMemory);
  /* force contract renegotiation */
  dout->dArrayLength = xsiz;
err:
  return cc;
}
#endif /* SMALL_TINE_LIB */
#ifndef DOS
/** 
 * \internal not used ! we should not need to marshall structs with BOOLEANs.
 */
int marshallStruct(DTYPE *d, int dir)
{
  struct structStruct *ss;
  struct structFormat *s;
  BYTE *dbuf = d->data.bptr;
  int i;
  if (!gMarshallBooleanAsByte) return 0;
  if ((ss=findstruct(d->dTag)) == NULL) return invalid_structure_tag;
  for (i = 0; i < (int)d->dArrayLength; i++)
  { /* look for any CF_BOOLEAN fields */
    dbuf = &d->data.bptr[i * ss->size];
    for (s = ss->s; s != NULL && s->siz; s = s->nxt)
    {
      if (s->fmt == CF_BOOLEAN)
      {
        marshallBoolean(&dbuf[s->addr], s->siz, dir);
      }
      if (s->fmt == CF_STRUCT)
      {
        DTYPE ds;
        ds.dArrayLength = s->siz;
        ds.dFormat = s->fmt;
        ds.data.vptr = &dbuf[s->addr];
        strncpy(ds.dTag, s->stag, TAG_NAME_SIZE);
        marshallStruct(&ds, dir);
      }
    }
  }
  return 0;
}
int prepStructOutputPostCall(ContractListStruct *cl,DTYPE *dout)
{
  int i, k, n, nfields = 0, dsiz=0, hsiz, cc = 0;
  int ssiz, allocsize;
  struct structStruct *ss;
  struct structFormat *s;
  BYTE *dbuf = dout->data.bptr, *ptr;

  if ((ss=findstruct(cl->contract.strTagOut)) == NULL) return invalid_structure_tag;
  if (!ss->hasExtSpace) 
  { /* nothing to prep */
    dout->dArrayLength *= ss->rawlen;
    return 0; 
  }
  /* on entry: dArrayLength = number of structs */
  ssiz = 0;
  for (i=0; i<(int)dout->dArrayLength; i++)
  { /* first scan for the total extended space size */
    dbuf = &dout->data.bptr[i*ss->size];
    for (s=ss->s, n=0; s != NULL && s->siz; s=s->nxt, n += dsiz)
    {
      dsiz = 0;
      if (isVariableLengthFormat((short)s->fmt))
      { /* scan thru for bytes not in the rawlen ... */
        hsiz = getFormatHeaderSize(s->fmt);
        for (k=0; k<s->siz; k++)
        {
          if (hsiz > 0)
          {
            ptr = (BYTE *)&dbuf[s->addr+k*getFormatTypeSize(s->fmt)];
            dsiz += getFormatBodyBufferSize(ptr,s->fmt);
          }
          else 
          { /* CF_STRING */
            ptr = (BYTE *)((size_t *)&dbuf[s->addr])[k];
            dsiz += (int)strlen((char *)ptr)+1;
          }
        }
      }
      nfields++;
    }
    ssiz += ss->rawlen + n;
  }
  if (ssiz > (int)cl->dataOutCapacity)
  { /* returned space has grown ! */
    if (!cl->unbuffered)
    { /* not managed from above so ... */
      SystemFree(cl->dataOut);
      /* add some extra bytes, as long as we're re-allocating ... */
      allocsize = ssiz + nfields * STRING_STD_CAPACITY + (STRING_STD_CAPACITY - ssiz%STRING_STD_CAPACITY);
      if ((cl->dataOut = (BYTE *)SystemCalloc(nDataOutBuffers,(size_t)allocsize)) == NULL)
        ccerr(out_of_server_memory);
      if (nDataOutBuffers > 1) cl->dataBuf = &cl->dataOut[allocsize];
      cl->dataOutCapacity = allocsize;
    }
  }
  /* set dArrayLength to total number of bytes */
  dout->dArrayLength = ssiz;
err:
  return cc;
}
#endif /* DOS */
int resizeStruct(struct structStruct *s,int newLength)
{ /* internal use: no null-pointer checks */
  int cc = 0,ssize = newLength*s->size, mxsize = (int)MaxSystemTransportSize;
  BYTE *ptr = NULL;
  if (newLength <= s->num) return 0; /* false alarm */
# ifndef DOS
  if (gTransportCeiling > 0) mxsize = gTransportCeiling;
  if (ssize > mxsize) return dimension_error; 
  if ((ptr=(BYTE *)realloc(s->sdata,ssize)) == NULL) ccerr(out_of_local_memory);
  s->sdata = ptr;
# endif
  s->num = newLength;
err:
  return cc;
}
int assignStructDataPtr(ContractListStruct *cl,DTYPE *d,BYTE *sptr,int dir)
{
  int cc=0,n=0;
  struct structStruct *ss;
  CONTRACT *con = &cl->contract;

  switch (dir) 
  {
    case RTOL: 
      d->data.bptr = sptr;
      if ((n=con->EqmSizeIn) == 0) ccerr(0);
      if ((ss=findstruct(con->strTagIn)) == NULL) 
      {
        if (strlen(con->strTagIn) > 0) ccerr(invalid_structure_tag);
        break;
      }
      n = n/ss->rawlen; 
      if (!ss->hasExtSpace)
      { /* 'normal struct: make the following checks ... */
        if (n > ss->num) if ((cc=resizeStruct(ss,n)) != 0) ccerr(cc);
        if (con->EqmSizeIn%ss->rawlen) ccerr(invalid_structure_size);
      }
      /* note asymmetry in EqmSizeIn (contains all bytes coming in) vs EqmSizeOut */
#     ifndef DOS
      n = _taggedStructSwap(con->strTagIn,NULL,sptr,ss->sdata,RTOL,n,ss->size);
      if (n < 0) ccerr(-n);
      d->data.bptr = ss->sdata;
#     endif
      break;
    case LTOR: 
      if ((n=con->EqmSizeOut) == 0) ccerr(0);
      if ((ss=findstruct(con->strTagOut)) == NULL) 
      {
        if (strlen(con->strTagOut) > 0) ccerr(invalid_structure_tag);
        break;
      }
      if ((n=n/ss->rawlen) > ss->num) if ((cc=resizeStruct(ss,n)) != 0) ccerr(cc);
      if (con->EqmSizeOut%ss->rawlen) ccerr(invalid_structure_size);
      n = d->dArrayLength; 
#     ifndef DOS
      if ((cc=prepStructOutputPostCall(cl,d)) != 0) ccerr(cc);
      _taggedStructSwap(con->strTagOut,NULL,cl->dataBuf,ss->sdata,LTOR,n,ss->size);
#     else
      d->dArrayLength *= s->rawlen;
#     endif
      break;
    case LTOL: 
      if ((n=con->EqmSizeOut) == 0) ccerr(0);
      if ((ss=findstruct(con->strTagOut)) == NULL) 
      {
        if (strlen(con->strTagOut) > 0) ccerr(invalid_structure_tag);
        break;
      }
      if ((n=n/ss->rawlen) > ss->num) if ((cc=resizeStruct(ss,n)) != 0) ccerr(cc);
      if (con->EqmSizeOut%ss->rawlen) ccerr(invalid_structure_size);
#     ifndef DOS
      memset(ss->sdata,0,(size_t)(n*ss->size));
      d->data.vptr = ss->sdata;
#     endif
      break;
  }
err:
  return cc != 0 ? -cc : n;
}
int isIndependentWriteProperty(ExportListStruct *el, char *prpName)
{
  int idx;
  ExportPropertyListStruct *prp = NULL;
  if (el == NULL || prpName == NULL) return FALSE;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp = el->EqmPrpLst[idx]; prp; prp = prp->next)
  {
    if (prp->isReserved) continue;
    if (!strnicmp(prp->prpName, prpName, PROPERTY_NAME_SIZE))
    {
      short acc = (short)(CA_WRITE|CA_INDEPENDENT);
      return (prp->prpAccessMode & acc) == acc ? TRUE : FALSE;
    }
  }
  return FALSE;
}
int isRegisteredMetaProperty(ExportListStruct *el,char *eqm,char *mprp,char *dev)
{
  PrpQueryStruct prpqsQuery, *prpqs = &prpqsQuery;
  XPropertyQueryStruct xpqsQuery, *xpqs = &xpqsQuery;
  int i,n, cc = FALSE;

  if (gRejectOverloadedMetaProperties) return FALSE;
  if (GetPropertyId(eqm,mprp) >= 0) ccerr(TRUE);
  if (el->XEqmPropertyQueryFcn != NULL)
  {
    xpqsQuery.prpAccess = CA_FIX;
    strncpy(xpqsQuery.prpName,mprp,PROPERTY_NAME_SHORTSIZE);
    n = el->XEqmPropertyQueryFcn(dev,&xpqs);
    for (i=0; i<n; i++) 
      if (!strnicmp(mprp,xpqs[i].prpName,PROPERTY_NAME_SHORTSIZE)) ccerr(TRUE);
    return FALSE;
  }
  if (el->EqmPropertyQueryFcn != NULL)
  {
    prpqsQuery.prpAccess = CA_FIX;
    strncpy(prpqsQuery.prpName,mprp,PROPERTY_NAME_SIZE);
    n = el->EqmPropertyQueryFcn(dev,&prpqs);
    for (i=0; i<n; i++) 
      if (!strnicmp(mprp,prpqs[i].prpName,PROPERTY_NAME_SIZE)) ccerr(TRUE);
    return FALSE;
  }
err:
  if (cc == TRUE && strstr(mprp,".HIST") != NULL && !gAllowOverloadedConcurrentLocalHistories)
  { /* a .HIST call that appears to be overloaded (doocs ?) */
    char *c, prp[PROPERTY_NAME_SIZE+8];
    strncpy(prp,mprp,PROPERTY_NAME_SIZE); prp[PROPERTY_NAME_SIZE] = 0;
    if ((c=strrchr(prp,'.')) != NULL) *c = 0;
    if (isPrpInHistoryTable(prp)) 
    { /* but .HIST is in the TINE history table ! */
      return FALSE; 
    }
  }
  return cc;
}
void zeroEqmStatus(ContractListStruct *cl)
{ /* internally used: cl never NULL ! */
  memset(cl->compString,0,STATUS_SIZE);
  memset(cl->exp->EqmLastErrorString,0,STATUS_SIZE);
}
int gRejectOverloadedMetaProperties = FALSE;
char *getContractKey(char *key,int clid)
{
  static char lclKeyBuf[256];
  if (clid < 0 || clid >= ncontracts || ContractList[clid] == NULL) return "--non existent--";
  if (key == NULL) key = lclKeyBuf;
  sprintf(key,"(%.6s)/%.64s[%.64s]",
    ContractList[clid]->contract.EqmName,
    ContractList[clid]->contract.EqmDeviceName,
    ContractList[clid]->contract.EqmProperty);
  return key;
}
int isLateResponding(ContractListStruct *cl)
{
  struct timeval tv;
  if (cl->expired) return FALSE; /* not active */
  if (cl->starttime.tv_sec == 0) return FALSE; /* not yet assigned */
  gettimeofday(&tv, (struct timezone*)NULL);/* Cycle time */
  if (MSECS(tv, cl->starttime) >= (int)(cl->pollingInterval - gSystemTick)) return TRUE;
  return FALSE;
}
int doContract(ContractListStruct *cl,int canUseSeparateThread)
{
  CONTRACT *con = &cl->contract;
  ClnLst *cln;
  ExportListStruct *el = cl->exp;
  BYTE *dataIn = cl->dataIn;
  BYTE *dataOut = cl->dataBuf;
  BYTE *bptr;
  DMDA mda_i,mda_o;
  DIMAGE img_i;
  DGSPECTRUM gsp_i;
  int i,k,istart,fmtsiz,hsiz=0,dsiz=1,cc=0;
  int isGated=GATE_NONE,isAdjustableOut=FALSE,wasCalled=FALSE,isVolatile=FALSE,needsRenegotiation=FALSE,hasData,memzero;
  short cf,access;
  UINT32 doutsize,cSizeOut,bsiz,dp,gate=0;
  char txtbuf[256],dbkey[256]="";
  double dts = gServerDataTimeStamp; /* gServerDataTimeStamp is set prior to call */
  DTYPE din,dout;
  char tDevName[DEVICE_NAME_MAXSIZE],tDevProperty[PROPERTY_NAME_SIZE+4],ctag[TAG_NAME_SIZE];

  if (tineDebug) getContractKey(dbkey,cl->listId);
  if (canUseSeparateThread && cl->runInSeparateThread)
  {
#   ifdef MULTITHREADED
    if (cl->threadId == 0) cl->returnCode = not_running;
    switch (cl->returnCode)
    {
    	case not_running:
    	  cl->returnCode = not_ready;
        cl->threadId = clCreateThread(cl);
       	if (tineDebug > 1) dbglog("launch contract %.256s in its own thread: 0x%lx",dbkey,cl->threadId);
    	case not_ready:
    	  if (tineDebug > 1) dbglog("contract %.256s has not yet completed on thread 0x%lx",dbkey,cl->threadId);
        if (!cl->expired) return not_ready; /* basically, this will always happen ... */
        dbglog("contract %.256s is expired but has not yet completed on thread 0x%lx", dbkey, cl->threadId);
        cl->returnCode = not_posted; /* contract expired: drop thru and get out */
    	default:
    	  if (tineDebug > 1) dbglog("contract %.256s has finished on thread 0x%lx",dbkey,cl->threadId);
#       ifdef WIN32
        CloseHandle(cl->threadId);
#       endif
        cl->threadId = 0;
    	  return cl->returnCode;
    }
#   endif
  }
  if (cl->threadId == 0) cl->threadId = GetThreadHandle();
  cSizeOut = con->EqmSizeOut; /* official contract output length */
  hsiz = getFormatHeaderSize(LFMT(con->EqmFormatOut));
  fmtsiz = fmtsizeof(con->EqmFormatOut);
  if (con->EqmFormatOut == BFMT(CF_MDA)) fmtsiz = getCarriedFormatSize(con->strTagOut);
  doutsize = cSizeOut*fmtsiz + hsiz;

  if (tineDebug > 3)
  {
    strncpy(txtbuf,dbkey,256); txtbuf[255] = 0;
    if (con->EqmSizeIn != 0)
      DataDump(txtbuf,(char *)dataIn,(int)con->EqmSizeIn,(short)LFMT(con->EqmFormatIn),(short)RTOL);
  }
  if (cl->compStatus & CE_SENDCODE) 
  { /* special signal from API */
    cc = cl->compStatus & 0x0fff;
    if (tineDebug > 1) dbglog("%.256s -> sending %32s signal to clients",dbkey,cc2str(cc));
    cl->compStatus = not_signalled; /* return to normal conditions after this pass */
    ccerr(cc);
  }
  if (cl->expired)
  {
    if (tineDebug > 1) dbglog("%.256s -> returning expired data",dbkey);
    cSizeOut = cl->dataSizeOut;
    ccerr(cl->returnCode);
  }
  /* check stock properties: */
  /* copy contract's property into temporary variable */
  if (strchr(con->EqmProperty,'*') != NULL && isSimpleFormat((short)LFMT(con->EqmFormatOut))) 
  { /* doocs likes to make query calls like this */
    strcpy(tDevProperty,"PROPERTIES");
  }
  else 
  {
    strncpy(tDevProperty,con->EqmProperty,PROPERTY_NAME_SIZE);
  }
  if ((i=GetStockPropertyId(tDevProperty)) >= 0)
  { /* zero out the output buffer if call is fresh */
    if (doutsize && !cl->callPending) memset(dataOut,0,(size_t)doutsize);
    cl->callPending = TRUE;
    cc = callStk(cl,(char *)dataOut,(char *)dataIn,gStockProperty[i].fcn);
    cl->callPending = FALSE;    
    if ((cln=cl->clnLst) != NULL && BASEMODE(cln->status.mode) != CM_SINGLE)
    { /* if a static stock property is being polled inform the caller ! */
      if (cc == 0 && IsStaticProperty(tDevProperty)) cc = CE_SENDDATA + information_static;
    }
    cSizeOut = cl->dataSizeOut;
    ccerr(cc);
  }
  if (el == NULL) return non_existent_elem;
  /* copy contract's property into temporary variable */
  strncpy(tDevProperty,con->EqmProperty,PROPERTY_NAME_SIZE);
  /* check stock modifiers: (allow overloads) */
  if ((i=IsMetaProperty(con->EqmProperty,NULL,NULL,FALSE)) > 0 &&
      !isRegisteredMetaProperty(el,con->EqmName,con->EqmProperty,con->EqmDeviceName) )
  { /* not an overridden meta property */
    i--; 
    if (doutsize && !cl->callPending) memset(dataOut,0,(size_t)doutsize);
    if ((isGated=IsGatedModifier(con,tDevProperty,&gate)) == 0) 
    { /* not gated: gated meta properties call the registered property */
      cl->callPending = TRUE;
      cc = callStk(cl,(char *)dataOut,(char *)dataIn,StockModifier[i].fcn);
      cl->callPending = FALSE;
      if ((cln=cl->clnLst) != NULL && BASEMODE(cln->status.mode) != CM_SINGLE)
      { /* if a static meta-property is being polled inform the caller ! */
        if (cc == 0 && IsStaticProperty(tDevProperty)) cc = CE_SENDDATA + information_static;
      }
      cSizeOut = cl->dataSizeOut;
      ccerr(cc);
    }
  }
  if (!el->hasPropertyHandler) return non_existent_function;
  if (el->isidle) return server_idle;
  if (cl->compStatus == not_signalled && el->isbusy) 
  {
    if (tineDebug > 2) dbglog("%.256s is %s (count %d)",dbkey,
      cl->compStatus == not_signalled ? "not signalled" : "busy", el->busycnt);
    if (el->busycnt < ALLOWED_BUSY_COUNTS) return not_ready; else return cl->compStatus = operation_busy;
  }
  /* copy contract's device name into local static variable */
  strncpy(tDevName,con->EqmDeviceName,DEVICE_NAME_SIZE); tDevName[DEVICE_NAME_SIZE] = 0;
  if (strlen(tDevName) == DEVICE_NAME_SIZE && tDevName[DEVICE_NAME_SIZE-1] == '&') 
  { /* long device name requested */
    tDevName[DEVICE_NAME_SIZE-1] = 0;
    if (cl->extStringSpace != NULL) 
    {
      strncat(tDevName,cl->extStringSpace,cl->extStringSize);
    }
    else
    {
      if (tineDebug) dbglog("non-terminated long device name !");
    }
    tDevName[DEVICE_NAME_MAXSIZE-1] = 0;
  }
  /* call equipment function or property handler: */
  if (el->hasPropertyHandler)
  { /* EQM function registered */
    UINT32 dal;
    if (doutsize > 0 && dataOut == NULL) ccerr(code_failure); /* sanity check */
    memzero = (SystemPresetMemory && doutsize > 0 && cl->compStatus == not_signalled && !cl->unbuffered);
    dout.dStamp = 0;
    dout.sysStamp = gSystemDataStamp;
    dout.dTimeStamp = dts;
    dout.dArrayLength = con->EqmSizeOut;
    dout.dFormat = LFMT(con->EqmFormatOut);
    dout.data.vptr = dataOut;
    strncpy(dout.dTag,con->strTagOut,TAG_NAME_SIZE);
    isAdjustableOut = isAdjustableLengthFormat(dout.dFormat);
    if (dout.dFormat == CF_STRUCT)
    {
      if ((i=assignStructDataPtr(cl,&dout,dataOut,LTOL)) < 0) return -i;
      dout.dArrayLength = i;
    }
#   ifndef SMALL_TINE_LIB
    else if (isAdjustableOut)
    {
      memzero = FALSE;
      if ((cc=prepAdjustableArrayOutputPreCall(cl,&dout)) != 0) ccerr(cc);
    }
    else if (dout.dFormat == CF_HISTORY)
    {
      if ((dsiz=getCarriedFormatSize(dout.dTag)) <= 0) ccerr(-dsiz);
      dsiz += HISTORY_HDR_SIZE;
      dout.dArrayLength /= dsiz;
    }
#   endif
    else if (dout.dFormat == CF_MDA)
    {
      memset(&mda_o,0,sizeof(DMDA));
      memzero = FALSE;
      dout.data.vptr = (BYTE *)&mda_o;
    }
    if (memzero)
    { /* clean out the working data buffer */
      memset(dataOut,0,doutsize);
    }
    din.dStamp = 0;
    din.sysStamp = gSystemDataStamp;
    din.dTimeStamp = dts;
    din.dArrayLength = con->EqmSizeIn;
    din.dFormat = LFMT(con->EqmFormatIn);
    din.data.vptr = dataIn;
    strncpy(din.dTag,con->strTagIn,TAG_NAME_SIZE);
    /* some last minute changes (TODO: consolidate this) */
    if (din.dFormat == CF_STRUCT)
    {
      if ((i=assignStructDataPtr(cl,&din,dataIn,RTOL)) < 0) return -i;
      din.dArrayLength = i;
    }
#   ifndef SMALL_TINE_LIB
    else if (isAdjustableLengthFormat(din.dFormat))
    {
      if ((cc=prepAdjustableArrayInputPreCall(cl,&din)) != 0) ccerr(cc);
    }
    else if (din.dFormat == CF_HISTORY)
    {
      if ((i=getCarriedFormatSize(din.dTag)) <= 0) ccerr(-i);
      i += HISTORY_HDR_SIZE;
      cfHistorySwap(din.dTag,din.data.bptr,dataIn,RTOL,din.dArrayLength);
      din.dArrayLength /= i;
    }
#   endif
    else if (din.dFormat == CF_MDA)
    {
      memcpy(&mda_i,cl->dataIn,MDA_HDR_SIZE);
      mda_i.data.bptr = &cl->dataIn[MDA_HDR_SIZE];
      din.data.vptr = &mda_i;
    }
    else if (din.dFormat == CF_IMAGE)
    {
      memcpy(&img_i, cl->dataIn, IMAGE_HDR_SIZE);
      img_i.frameBuffer = &cl->dataIn[IMAGE_HDR_SIZE];
      img_i.frameBufferSize = din.dArrayLength;
      din.data.vptr = &img_i;
    }
    else if (din.dFormat == CF_GSPECTRUM)
    {
      memcpy(&gsp_i, cl->dataIn, GSPECTRUM_HDR_SIZE);
      gsp_i.spectBuffer = (float *)&cl->dataIn[GSPECTRUM_HDR_SIZE];
      gsp_i.spectBufferSize = din.dArrayLength;
      din.data.vptr = &gsp_i;
    }
    access = con->EqmAccess;
    if (access & CA_WRITE && el->accessLock.lockType) access |= CA_LOCKED;
    if (cl->accessFlags & CA_REPEAT) access |= CA_REPEAT; /* apply this for the call */
    cf = dout.dFormat; /* record format prior to call (take no chances inside the eqm!) */
    strncpy(ctag,dout.dTag,TAG_NAME_SIZE); /* and record the data tag !*/
    /* call the equipment module : */
    dal = dout.dArrayLength; /* requested output array length pre-call */
    el->EqmNumCalls++;
    cl->callPending = TRUE;
    zeroEqmStatus(cl); /* clear the status string information */
    cc = callEqm(cl,tDevName,tDevProperty,&dout,&din,access);
    cl->callPending = FALSE;
    /* call completed! */
    wasCalled = TRUE;
    if (cc != 0 && !(cc & CE_SENDDATA)) ccerr(cc);
    if (dout.dArrayLength > dal)
    { /* caller set this to an illegal value: it could already be too late ... */
      feclogEx(TINE_LOGLEVEL_ERROR, 
        "equipment module %.6s/%.64s[%.64s] set return length %d past the requested length %d",
        con->EqmName,con->EqmDeviceName,con->EqmProperty,
        dout.dArrayLength,dal);
      dout.dArrayLength = dal; /* set it back to what was allcoated and hope for the best */
    }
#   ifndef SMALL_TINE_LIB
    if (access & CA_WRITE)
    { /* was a successful WRITE command (check save/restore conditions) */
      writeSrValues(el,tDevName,tDevProperty,&din); 
    }
#   endif /* SMALL_TINE_LIB */
    /* have a look at the returned output data : */
    if ((hsiz=GetFormatHeaderSize(cf)) > 0 && !isAdjustableOut)
    { /* if format has a header, map to outgoing  stream */
      if ((bptr=GetFormatBodyPtr(dataOut,cf)) != NULL)
      { /* body pointer does not follow the header */
        if ((bsiz=GetFormatBodyBufferSize(dataOut,cf)) < dout.dArrayLength) dout.dArrayLength = bsiz;
        memcpy(&dataOut[hsiz],bptr,(size_t)(dout.dArrayLength*fmtsizeof(GetFormatDataType(cf))));
      }
    }
    if (cf == CF_STRUCT)
    {
      if ((i=assignStructDataPtr(cl,&dout,dataOut,LTOR)) < 0) return -i;
      cSizeOut = dout.dArrayLength;
    }
#   ifndef SMALL_TINE_LIB
    else if (isAdjustableOut)
    { /* prepare the return buffer */
      cc = prepAdjustedArrayOutputPostCall(cl,&dout);
      isGated = GATE_NONE;
      cSizeOut = dout.dArrayLength;
    }
    else if (cf == CF_HISTORY)
    { /* a very special case */
      if (ftoi(ctag) == CF_IMAGE)
      { /* an even more special case (do these 1 at a time) */
        UINT32 isiz;
        dsiz = 0;
        for (k=0; k<(int)dout.dArrayLength; k++)
        {
          memcpy(&isiz,&dataOut[HISTORY_HDR_SIZE+dsiz]+IMAGE_HDR_SIZE-sizeof(SINT32),sizeof(SINT32));
          isiz += HISTORY_HDR_SIZE + IMAGE_HDR_SIZE;
          cfHistorySwap(ctag,&dataOut[dsiz],&dataOut[dsiz],LTOR,isiz);
          dsiz += isiz;
        }
      }
      else
      { /* a normal special case */
        cfHistorySwap(ctag,dataOut,dataOut,LTOR,dout.dArrayLength*dsiz);
      }
      dout.dArrayLength *= dsiz;
    }
#   endif
    else if (cf == CF_MDA)
    { /* another special case */
      UINT32 dlen = 1;
      if (mda_o.ndim > MAX_MDA_DIMS) ccerr(illegal_data_size);
      for (k=0; k<mda_o.ndim; k++) dlen *= mda_o.axis[k].len;
      if (dlen > cSizeOut) dlen = cSizeOut; /* sanity check */
      dout.dArrayLength = dlen;
      dataOut = cl->dataBuf; /* re-assign back to 'normal' */
      memcpy(dataOut,&mda_o,MDA_HDR_SIZE);
      memcpy(&dataOut[hsiz],mda_o.data.vptr,dlen*fmtsizeof(mda_o.fmt));
    }
    if (dout.dArrayLength < cSizeOut)
    {
#     ifndef SMALL_TINE_LIB
      isVolatile = hasVolatileDataSize(cf,ctag,NULL,NULL);
#     endif
      if (!SystemPresetMemory && !cl->unbuffered && !isVolatile)
      { /* zero out the remainder for future data comparisons */
        dp = hsiz + dout.dArrayLength*fmtsiz;
        memset(&dataOut[(int)dp],0,(size_t)(con->EqmSizeOut-dout.dArrayLength)*fmtsiz);
      }
      cSizeOut = dout.dArrayLength; 
    }
    if (!isGated && cl->dataSizeOut != cSizeOut)
    { /* last sent data length != the result of this call ! */
      needsRenegotiation = TRUE;
    }
    cl->dataSizeOut = cSizeOut; 
    /* static data ? */
    if (cl->accessFlags & CA_STATIC) cc = CE_SENDDATA + information_static;
    /* apply the timestamp : */
    if (dts != dout.dTimeStamp) 
    { /* a thread-safe way of setting this from the eqm */
      dts = dout.dTimeStamp; 
    }
    else 
    { /* a legacy way of setting this from the eqm */
      dts = gServerDataTimeStamp;
    }
    cl->dtimestamp = (SINT32)dts;
    cl->dtimestampUSEC = (int)((dts - (SINT32)dts) * 1000000);
    cl->datastamp = dout.dStamp;
    cl->sysstamp = dout.sysStamp; /* central servers will be setting this inside the dispatch */
  }
  switch (isGated)
  {
    case GATE_BOOLEAN:
      gateOutputData(dataOut,(int)cSizeOut,(short)LFMT(con->EqmFormatOut),gate,TRUE);
      break;
    case GATE_FILTER:
      gateOutputData(dataOut,(int)cSizeOut,(short)LFMT(con->EqmFormatOut),gate,FALSE);
      break;
    case GATE_ARRAY:
      if (hsiz > 0) break;
      if ((istart=GetDeviceNumber(el->EqmName,tDevName)) < 0) break;
      for (i=0,k=0; k<(int)cSizeOut && i<el->EqmNumDevices-istart; i++)
      { /* this will only be MCA types and nothing with a format header */
        if (el->EqmDevLst == NULL) continue;
        if (el->EqmDevLst[i + istart].isReserved) continue;
        if (el->EqmDevLst[i+istart].mask != 0 && (el->EqmDevLst[i+istart].mask & gate) == 0) continue;
        memcpy(&dataOut[k*fmtsiz],&dataOut[i*fmtsiz],fmtsiz);
        k++;
      }
      cl->dataSizeOut = cSizeOut = k;
      break;
  }
err:
  if (wasCalled)
  { /* the eqm was definitely called */
    if (gConSigfcn != NULL) 
    { /* general contract signal function is registered */
      sendContractSignal(el->EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,CS_PROCESSED,cc);
    }
    if (cl->prp != NULL && cl->prp->sigfcn != NULL) 
    { /* property-specific signal handler is registered */
      sendPrpSignal(cl->prp,CS_PROCESSED,cl->contractID,cc);
    }
  }
  if (con->EqmAccess&CA_FIRST) 
  {
    con->EqmAccess &= ~(CA_FIRST);
  }
  if (cc == not_ready) return cc;  /* do NOT pass go, do NOT collect $200 */
  cl->accessFlags &= ~(CA_REPEAT); /* turn this off here */
  if (cc < 0) cc = ambiguous;
  hasData = HAS_DATA(cc);
  if (tineDebug > 3 && hasData)
    DataDump(txtbuf,(char *)cl->dataBuf,(int)cSizeOut,(short)LFMT(con->EqmFormatOut),(short)LTOR);
  if (hasData && cl->dataSizeOut != cSizeOut)
  { /* data length has just been reduced (only possible in gated output above) */
    needsRenegotiation = TRUE;
  }
  if (cSizeOut != con->EqmSizeOut) 
  { /* last data length != requested data length */
    needsRenegotiation = TRUE;
  }
  if (needsRenegotiation)
  { /* don't disrupt the contract itself */
    /* con->EqmSizeOut = cSizeOut; */   
    cc |= CE_RENEGOTIATE;
  }
  return cc;
}
int initServerTransport(void)
{
  int cc = 0;
  char txt[128];
  feclog("initializing server transport");
  if ((cc=initControlStructs()) != 0) 
  {
    sprintf(txt,"Cannot map header structures");
    ccerr(cc);
  }
# if defined(IPXBSD) || defined(IPXNW)
  feclog("initializing ipx subsystem");
  if ((cc=initIPX()) != 0)
  {
    sprintf(txt,"IPX/SPX server failed to initialize");
    ccerr(cc);
  }
# endif
  feclog("initializing ip subsystem");
  if ((cc=initIPServer()) != 0)
  {
    sprintf(txt,"UDP/TCP server failed to initialize");
    ccerr(cc);
  }
  if (!gRunServerCycleInSeparateThread) clientTimeout = clientTimeoutMsec = 0;
  sprintf(txt,"server transport initialized");
err:
  feclog("%s : %.32s",txt,cc2str(cc));
  return cc;
}
int initTineServices(void)
{
  int cc = 0;
  char *usr;
  if (gInitTineServicesDone) return 0;
  if (!FecNameRegistered)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "aborting server initialization : fec name is unknown");
    return host_not_resolved;
  }
  feclog("initializing tine services");
  memset(&gLastAllowedCln,0,sizeof(ClnHdr));
  getSrvIPAddr();
  if (gRespondToServiceRequests) cc = RegisterLocalServiceModule();
  cc = initServerTransport();
  initAlarmServer();
  makeStockPropXRefTable();
  if (MinPollingRate < 10) MinPollingRate = 10; /* sanity check */
  if (!gCycleDeadbandFixed) gSystemCycleDeadband = MinPollingRate;
  if (cdiSystemInitHook != 0) cdiSystemInitHook(); /* if CDI is in use ... */
# if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
  if (gResetUseLoopback) gUseLoopback = FALSE;
  if (shmStart())
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "Cannot create shm window!");
    gUseLoopback = TRUE;
  }
# endif
# ifdef MULTITHREADED
  hSchdThrd = schCreateThread(0);
# endif
  initClient();
  gInitTineServicesDone = TRUE;
  gDelayInitTineServices = FALSE;
  usr = whoami(gUserName);
  feclog("Caller ID: %.16s, Owner %.16s",gUserName,usr);
  feclog("tine services initialized");
  return cc;
}
int initTineResources(void)
{
  int cc=0, val;
  BYTE *ver = SystemVersion();
  char *ptr;

  chkEndian();
  gSystemTick = getClockTickMilliseconds();
  initIpFromEnv();
  if (gSystemRunningStandAlone) gIgnoreCentralServices = TRUE;
  if ((ptr=getenv("TINE_BURSTLIMIT")) != NULL)
  {
    val = atoi(ptr);
    if (val < 10) val = 10;
    gBurstLimit = val;
  }
  if ((ptr=getenv("TINE_NETWORKADDRESS_RESOLUTION")) != NULL) 
  {
    gRespondToServiceRequests = strtobool(ptr);
  }
  if ((ptr=getenv("TINE_USE_GLOBAL_SYNCHRONIZATION")) != NULL) 
  {
    useGlobalSynchronization = strtobool(ptr);
    feclog("set global synchronization to %s via environment",useGlobalSynchronization ? "TRUE" : "FALSE");
  }
  if ((ptr=getenv("TINE_USE_CYCLE_TRIGGER")) != NULL) 
  {
    useCycleTrigger = strtobool(ptr);
    feclog("Use Cycle Trigger set to %s via environment",useCycleTrigger ? "TRUE" : "FALSE");
  }
  feclog("VERSION : %d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  feclog("BUILD ID: %d", tineBuildId);
  feclog("OS      : %s",OS_TAG);
  if (srvWorkAreaSize >= WORK_SIZE && 
      MaxSystemTransportSize > MAX_TRANSPORTSIZE &&
      MaxSystemTransportSize > srvWorkAreaSize) srvWorkAreaSize = MaxSystemTransportSize;
  if ((srvWorkArea=(BYTE *)SystemCalloc(1, (size_t)srvWorkAreaSize)) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_FATAL,"Work Space %u bytes: insufficient memory",(unsigned int)srvWorkAreaSize);
    return out_of_server_memory;
  }
  tmpWorkAreaSize = srvWorkAreaSize;
# ifdef MULTITHREADED
  feclog("Work Area: %u",srvWorkAreaSize);
  feclog("Temp Size: %u",tmpWorkAreaSize);
  feclog("FEC: is running multi-threaded");
# else
  if (srvWorkAreaSize <= WORK_SIZE && 
      MaxSystemTransportSize > srvWorkAreaSize) MaxSystemTransportSize = srvWorkAreaSize;
  feclog("Work Area: %u",(unsigned int)srvWorkAreaSize);
  feclog("Temp Size: %u",(unsigned int)tmpWorkAreaSize);
  feclog("Max Transport Size: %u",(unsigned int)MaxSystemTransportSize);
# if defined(INTEL) && !defined(VXWORKS) && !defined(MACOS)
  tmpWorkArea = srvWorkArea;
# endif
# endif
  feclog("FEC HOME : [%.128s]",fecDbPath);
  feclog("Allowed Log File Scanning : %s",logFileScanText[gLogfileAllowScan]);
  gSystemStartupTime = time(NULL);
  nconsumers = ncontracts = 0;
  if (tmpWorkArea == NULL && (tmpWorkArea=(BYTE *)SystemCalloc(1,tmpWorkAreaSize)) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "Temp Work Space: insufficient memory");
    return out_of_server_memory;
  }
  if (MaxNumClients <= 0) return 0; /* no server functionality required */
  if ((ptr=getenv("FEC_MAXIMUM_NUM_CONTRACTS")) != NULL)
  {
    val = atoi(ptr);
    SetContractListCapacity(val);
  }
  if ((ContractList=(ContractListStruct **)SystemCalloc(MaxNumContracts,sizeof(ContractListStruct *))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "Contract Table: insufficient memory");
    return out_of_server_memory;
  }
  else
  {
    ContractListCapacity = MaxNumContracts;
    feclog("Contract Table: reserved space for %d entries",ContractListCapacity);
  }
  if ((ClnTbl=(ClnHdr **)SystemCalloc(MaxNumClients,sizeof(ClnHdr *))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "Client Table: insufficient memory");
    return out_of_server_memory;
  }
  else
  {
    ClientListCapacity = MaxNumClients;
    feclog("Client Table: reserved space for %d entries",ClientListCapacity);
  }
  if (canSendGlobals)
  {
    if ((gcastTbl=(GlobalListStruct **)SystemCalloc(MaxNumGCasts,sizeof(GlobalListStruct *))) == NULL)
    {
      feclogEx(TINE_LOGLEVEL_FATAL, "Globals List: insufficient memory");
      return out_of_server_memory;
    }
    GCastTblCapacity = MaxNumGCasts;
  }
  getFecInfoList(); /* will establish XML readout or not */
  GetAliasTable(NULL);
  memset(&IPXFecAddress,0,sizeof(IPXAddress));
  getRegisteredExports(EXPORT_FILE,NULL);
  getFecName();
  getRegUsrsFromFile(NULL,NULL,NULL,&gDeniedUsersList,DENYUSER_FILE,"DENIED",&CheckDeniedList);
# if defined(VXWORKS)
  gDelayInitTineServices = TRUE;
# else
  if (!FecNameRegistered)
  {
    gDelayInitTineServices = TRUE;
    feclog("delaying system initialization");
  }
  else
  {
    gDelayInitTineServices = FALSE;
    cc = initTineServices();
  }
# endif /* VXWORKS */
  return cc;
}
TINE_EXPORT int ServerExitCondition = 0;
TINE_EXPORT void SetServerExitCondition(int value)
{
  if (value < 0) return;
  ServerExitCondition = value;
  if (value > 0 && gIsRunningAsServer)
  {
    WaitForSystemMutex(hHistoryCycleMutex,-1);
#   ifdef MULTITHREADED
    ReleaseSystemSemaphore(hSchedulerSemaphore);
#   endif
    millisleep(gSystemTick*2);
    CloseClients(TRUE);
    millisleep(gSystemTick*2);
    ReleaseSystemMutex(hHistoryCycleMutex);
  }
}
TINE_EXPORT void CloseServers(int closesockets)
{
  if (ServerExitCondition == 0) ServerExitCondition = 1;
  UNUSED(closesockets); /* always true */
# ifdef MULTITHREADED
  if (hSchedulerSemaphore != (SEMHANDLE)NULL) ReleaseSystemSemaphore(hSchedulerSemaphore);
# endif
# if defined(IPXNW) || defined(IPXBSD)
  closeIPXServer(); 
# endif
  closeIPServer();
# if defined(WIN32)
  stopPipeListener();
# if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
  shmStop();
# endif
# endif
  gInitTineServicesDone = FALSE;
  gInitIPServicesDone = FALSE;
  FecNameRegistered = FALSE;
  gInitIPClientDone = FALSE;
  SystemRunning = FALSE;
  if (gFecInfoList) freeFecListMemory(gFecInfoList); 
  gFecInfoList = NULL;
  fecXmlDone = FALSE;
  ServerInitialized = FALSE;
  return;
}
void quitServer(void)                         /* allow a clean exit */
{
  if (ServerExitCondition == 0) ServerExitCondition = 1;
  return;
}
WriteAccessInfo wrAccTbl[WRITEACCESS_TBL_SIZE];
int wrAccTblPtr = -1;

void appendWriteAccessTable(ContractListStruct *cl)
{
  WriteAccessInfo *wa;
  wrAccTblPtr = (wrAccTblPtr+1)%WRITEACCESS_TBL_SIZE;
  wa = &wrAccTbl[wrAccTblPtr];
  if (cl->clnLst != NULL && cl->clnLst->client != NULL) 
  {
    strncpy(wa->user,cl->clnLst->client->usr,USERNAME_SIZE);
    inetStrAdr(&cl->clnLst->client->adr.ip,wa->addr,FALSE);
    wa->port = inetPort(&cl->clnLst->client->adr.ip);
  }
  strncpy(wa->prp,cl->contract.EqmProperty,PROPERTY_NAME_SIZE);
  strncpy(wa->dev,cl->contract.EqmDeviceName,DEVICE_NAME_SIZE);
  strncpy(wa->eqm,cl->contract.EqmName,EQM_NAME_SIZE);
  wa->timestamp = gServerDataTimeStamp;
  return;
}

UINT32 gBurstLimitReachedCount = 0;
UINT32 gTimeServerCycle = 0;
UINT32 gTimeServerIdleLastCycle = 0;
UINT32 gTimeFdIdle = 0;
int gLastMinuteInitializationDone = FALSE;
int gExportManifestsDone = FALSE;
TINE_EXPORT int FecIsInitialized(void) { return gLastMinuteInitializationDone; }
TINE_EXPORT void ResetLastMinuteInitialization(void) { gLastMinuteInitializationDone = FALSE; }
TINE_EXPORT void resetLastMinuteInitialization(void) { ResetLastMinuteInitialization(); }
int gCycleCounts = 0, gMaxCycleCounts = 1;
int gServerCycleIdle = FALSE;
int gServerCycleNeedsScheduler = TRUE;
int gHistoryCycleNeedsScheduler = TRUE;
int gSchedulerInterval = 0;
static struct timeval lastScheduled = {0, 0};
int gHaltOnCycleHalt = FALSE;
int serverCycle(void)
{
  struct timeval tv;
  static struct timeval lastCycleTime;
  static int CycleCounts = 0;

  if (ClientListCapacity <= 0) return 0; /* no server functionality required */
  ProbeTimeElapsed("server Cycle entered");
  if (ServerExitCondition) 
  { 
    gServerCycleState = CYCLE_STATE_EXIT;
    if (ServerExitCondition++ > 3 && gHaltOnCycleHalt && !gIsExitingEQPModules) haltServer(); 
  }
  if (!ServerInitialized) return 0; /* no server functionality required yet */
  if (gServerCycleIdle)
  {
    gServerCycleState = CYCLE_STATE_IDLE;
    AcceptRequest(&thisCycleTime);/* accept console commands */
    return 0;
  }
  if (!gLastMinuteInitializationDone) 
  {
    int lmi_done = FALSE;
    gServerCycleState = CYCLE_STATE_INIT_FINAL;
    if (WaitForSystemMutex(hSystemInitMutex,gSystemTick) == 0)
    {
      gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
      lmi_done = LastMinuteInitialization();
      lastCycleTime = thisCycleTime;
      ReleaseSystemMutex(hSystemInitMutex);
    }
    if (!gSystemRunningStandAlone) 
    {
      if (useGlobalSynchronization) SystemStartGlobalSynchronization();
#     ifndef SMALL_TINE_LIB
      if (useCycleTrigger) SystemStartCycleTrigger();
#     endif
    }
    gLastMinuteInitializationDone = lmi_done;
  }
  if (WaitForSystemMutex(hSystemServerMutex,0) != 0) return not_ready;
  gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
  lastCycleTime = thisCycleTime;
  if (!gInitTineServicesDone) 
  { /* no server functionality required yet */
    ReleaseSystemMutex(hSystemServerMutex);
    return 0; 
  }
# ifndef SMALL_TINE_LIB
  if ((gSystemDataStampDly > 0) && /* is a delay value registered */
      (gSystemDataStamp != gSystemDataStampRb) && /* system stamp not yet applied */
      MSECS(thisCycleTime,gSystemDataStampTs) > gSystemDataStampDly) /* the delay period has elapsed */
  { /* apply system stamp */
    applySystemDataStamp();
  }
# endif
  gServerCycleNeedsScheduler = gSchedulerInterval > 0 ? TRUE : FALSE; /* turned on during the 'accept'! */
  ProbeTimeElapsed("server Cycle checking requests");
  tv = thisCycleTime; /* note this, it might change ! */
  gServerCycleState = CYCLE_STATE_ACCEPT;
  AcceptRequest(&thisCycleTime);/* accept consumer subscriptions and renewals   */
  gTimeServerIdleLastCycle += gTimeFdIdle;
  ProbeTimeElapsed("server Cycle finished checking requests");
  if (gSchedulerInterval <= 0 || MSECS(tv,lastScheduled) >= gSchedulerInterval)
  { /* is it time to service consumers? */
    if (MSECS(tv,lastScheduled) > (int)gSystemCycleDeadband) gServerCycleNeedsScheduler = TRUE;
    if (gServerCycleNeedsScheduler) doScheduler(&thisCycleTime,gSystemTick,SCHED_INTERN,NULL,NULL);
  }
  if (ngcasts) doGCast(&thisCycleTime);
  if (acquireDeviceNamesFromProperties) 
  {
    gServerCycleState = CYCLE_STATE_ACQNAMES;
    devNamesInitialization();
  }
  gServerCycleState = CYCLE_STATE_CLEANUP;
  CleanupTransfers(&thisCycleTime);
  ProbeTimeElapsed("server Cycle finished checking transfer cleanup");
  SurveyAlarms(&thisCycleTime);
  ProbeTimeElapsed("server Cycle finished checking alarms");
  if (nHistoryRecords) 
  {
    historyCycle(gHistoryCycleNeedsScheduler);
    ProbeTimeElapsed("server Cycle finished checking history");
  }
  if (!ExportsRegistered) 
  {
    gServerCycleState = CYCLE_STATE_SNDEXP;
    sndRegisteredExports(&thisCycleTime);
  }
  gServerCycleState = CYCLE_STATE_CHK_LOCKS;
  checkAccessLocks((time_t)thisCycleTime.tv_sec);
  gServerCycleState = CYCLE_STATE_CHK_GRP;
  checkGroupMembers((time_t)thisCycleTime.tv_sec);
  if (gRespondToServiceRequests) reAttachToServiceGroup();
  if (!gExportManifestsDone && thisCycleTime.tv_sec > gSystemStartupTime + 20)
  {
    dumpExportManifests();
    gExportManifestsDone = TRUE;
  }
  gTimeServerCycle += MSECS(thisCycleTime,lastCycleTime);
  lastCycleTime = thisCycleTime;
  CycleCounts++;
  if (gTimeServerCycle >= 1000)
  { /* have spent a cummulative 1000 msec in the serverCycle() */
    if (CycleCounts > gMaxCycleCounts) gMaxCycleCounts = CycleCounts;
    gCycleCounts = CycleCounts;
    if (!gIsCyclePollingMode)
    {
      AveBusyTime = (int)((gTimeServerCycle-gTimeServerIdleLastCycle)*100/gTimeServerCycle);
    }
    else
    {
      AveBusyTime = (gMaxCycleCounts-CycleCounts)*100/gMaxCycleCounts;
    }
    gTimeServerCycle = gTimeServerIdleLastCycle = 0;
    CycleCounts = 0;
  }
  ReleaseSystemMutex(hSystemServerMutex);
  ProbeTimeElapsed("server Cycle returning");
  if (ttimeprobe > 0) ttimeprobe--;
  return 0;
}
ContractListStruct **ContractList;
ClnHdr **ClnTbl; /* registered consumer list */
int nconsumers = 0, ncontracts = 0;
int StaleData = 0;
static char *getContractState(int state)
{
  switch (state)
  {
    case CONTRACT_ACTIVE: return "active";
    case CONTRACT_EXPIRED: return "expired";
    case CONTRACT_AVAILABLE: return "available";
    default: return "unknown";
  }
}
#ifndef SMALL_TINE_LIB
int isValidClient(ClnHdr *cln)
{
  int i;
  for (i=0; i<nconsumers; i++) if (ClnTbl[i] == cln) break;
  return i < nconsumers ? TRUE : FALSE;
}
int removePhantomClientFromContracts(ClnHdr *cln,char *tag,int needLock)
{
  ClnLst *c, *p;
  ContractListStruct *cl; 
  int i, n;
  char addr[64];
  if (ContractList == NULL)
  {
    ncontracts = 0;
    return code_failure;
  }
  if (needLock && WaitForSystemMutex(hClientTableMutex,-1) != 0) return mutex_error;
  for (i=0; i<ncontracts; i++)
  { /* scan thru all contracts */
    if ((cl=ContractList[i]) == NULL) continue;
    for (p=c=cl->clnLst; c != NULL; c=c->next)
    { /* search and destroy */
      if (c->client == cln)
      { /* found the scalliwag */
        if (c == cl->clnLst)
        { /* list head */
          cl->clnLst = c->next;
        }
        else
        { /* join the ends */
          p->next = c->next;
        }
        inetStrAdr(&c->client->adr.ip,addr,FALSE);
        feclogEx(TINE_LOGLEVEL_INFO, "remove phantom client %.16s at %.40s from %s contract %.256s : %s",
          c->client->usr,addr,getContractState(cl->expired),
          getContractKey(NULL,cl->listId),tag);
        if (cl->deliveryPending > 0) cl->deliveryPending--;
        break;
      }
      p = c;
    }
    /* now ensure consistency */
    for (c=cl->clnLst,n=0; c != NULL; c=c->next,n++);
    cl->nclients = n;
  }
  if (needLock) ReleaseSystemMutex(hClientTableMutex);
  return 0;
}
#endif
int removeClientFromContract(ClnLst *c,int tid) /* from contract table id : tid */
{
  ClnLst *p;
  int cc=0,n,wasPending = FALSE;

  if (tid < 0 || tid>=ncontracts || c == NULL) return invalid_index;
  if (WaitForSystemMutex(hClientTableMutex,-1) != 0) return mutex_error;
  if (tineDebug > 1) dbglog("remove consumer %.16s from contract %d",c->client->usr,tid);
  if ((p=ContractList[tid]->clnLst) == NULL) ccerr(code_failure);
  if (c->status.blknum < c->status.numblks) wasPending = TRUE;
  StaleData++; /*  Force removeClientTableEntry() */
  if (p == c) 
  { /* head of the list */
    ContractList[tid]->clnLst = c->next;
  }
  else
  { /* somewhere in the middle */
    while (p->next != NULL && p->next != c) p=p->next;
    if (p->next == NULL)
    { /* made it thru the list without finding the client ! */
      feclogEx(TINE_LOGLEVEL_INFO, "removed client not found in contract");
      ccerr(non_existent_client);
    }
    else
    { /* splice the ends */
      p->next = c->next;
    }
  }
  if (p == p->next) p->next = NULL; /* just in case ! */
  c->client->ncontracts = MAX(c->client->ncontracts-1,0);
  SystemFree(c);
  ContractList[tid]->nclients = MAX(ContractList[tid]->nclients-1,0);
  if (ContractList[tid]->clnLst == NULL) ContractList[tid]->nclients = 0;
  if (ContractList[tid]->nclients == 0) 
  {
    if (ContractList[tid]->clnLst != NULL) feclogEx(TINE_LOGLEVEL_WARN, "removed phantom user list from contract");
    ContractList[tid]->clnLst = NULL;
  }
  for (n=0, p=ContractList[tid]->clnLst; p != NULL; p=p->next, n++)
  { /* scan to the end of the list and guarantee this */
    if (n == ContractList[tid]->nclients - 1) 
    {
      if (p->next != NULL) feclogEx(TINE_LOGLEVEL_INFO, "removed phantom user from contract");
      p->next = NULL;
    }
  }
  if (wasPending) ContractList[tid]->deliveryPending = 0;
  if (ContractList[tid]->nclients > 0)
  { /* still a client list left: find the best polling rate */
    p = ContractList[tid]->clnLst;
    ContractList[tid]->pollingInterval = p->status.pollingInterval;
    for (p=p->next; p != NULL; p=p->next)
      ContractList[tid]->pollingInterval = findpoll(ContractList[tid]->pollingInterval,p->status.pollingInterval);
  }
err:
  ReleaseSystemMutex(hClientTableMutex);
  return cc;
}
int removeClientTableEntry(int tid)
{ /* return memory to heap */
  int i;
  if (tid < 0 || tid >= nconsumers) return invalid_index;
  if (ClnTbl[tid] == NULL) return code_failure;
  if (tineDebug > 1) dbglog("remove client %.16s",ClnTbl[tid]->usr);
# ifndef SMALL_TINE_LIB
  removePhantomClientFromContracts(ClnTbl[tid],"client removed",FALSE);
# endif
  SystemFree(ClnTbl[tid]);
  for (i=tid;i<nconsumers-1; i++) ClnTbl[i] = ClnTbl[i+1];
  ClnTbl[--nconsumers] = NULL;

  return 0;
}
int removeClient(ClnHdr *cln)
{
  int i;
  for (i=0; i<nconsumers; i++)
  {
    if (ClnTbl[i] == cln)
    {
      return removeClientTableEntry(i);
    }
  }
  return un_allocated;
}
int removeContractTableEntry(int tid)
{ /* return memory to heap */
  ContractListStruct *cl=ContractList[tid];
  ClnLst *c,*p;
  int isbusy;

  if (tid < 0 || tid>=ncontracts) return argument_list_error;
  if (cl == NULL) return code_failure;
  if (cl->compStatus == not_posted)
  {
    if (tineDebug) dbglog("remove contract %d (not yet posted)",tid);
  }
  /* isbusy is NOT thread-safe */
  isbusy = cl->isStkCall ? gStockResourceBusy : 
           cl->exp != NULL ? cl->exp->isinside : FALSE;
  if (cl->callPending || isbusy)
  { /* only in here, because there are no more attached clients */
    if (tineDebug) dbglog("remove contract %d (eqm still busy)",tid);
    if (cl->runInSeparateThread && cl->threadId != 0) 
    { /* shouldn't happen, but if it does: let this keep going around */
      if (tineDebug) dbglog("postpone remove contract %d", tid);
      return operation_busy;
    }
  }
  if (tineDebug > 1) dbglog("remove contract %d",tid);
  for (p=c=cl->clnLst; p != NULL; c=p)                
  { /* decrement consumer count: */
    c->client->ncontracts = MAX(c->client->ncontracts-1,0);
    StaleData++; /*  Force removeClientTableEntry() */
    p = c->next;
    SystemFree(c);
  }
  cl->clnLst = NULL;
  cl->expired = CONTRACT_AVAILABLE;
  SystemFree(cl->dataIn);
  SystemFree(cl->dataPtr);
  if (!cl->unbuffered) SystemFree(cl->dataOut);
  cl->dataOut = NULL;
  SystemFree(cl->dataOutArr);
  SystemFree(cl->extStringSpace);
  return 0;
}
static time_t rejectEarlierThan = 0;
#define FLUSH_WINDOW 5
void flushContractTable(void)
{
  int i;

  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i] == NULL) continue;
    ContractList[i]->contract.EqmAccess |= CA_LAST;
    doContract(ContractList[i],TRUE);
    while (ContractList[i]->nclients > 0)
    {
      if (removeClientFromContract(ContractList[i]->clnLst,i) != 0) break;
    }
    removeContractTableEntry(i);
  }
  ncontracts = 0; nExpiredContracts = 0;
  while (nconsumers > 0)
  {
    if (removeClientTableEntry(nconsumers-1) != 0) break;
  }
  nconsumers = 0;
  feclog("contract table flushed");
  rejectEarlierThan = time(NULL) + FLUSH_WINDOW;
}
UINT32 gContractMisses = 0;
UINT32 gContractDelays = 0;
UINT32 gClientMisses = 0;
UINT32 gClientReconnects = 0;
UINT32 gClientRetries = 0;
UINT32 gSystemTick = SYSTEM_TICK;
UINT32 gSynchronizeContracts = (UINT32)CONTRACT_SYNCHRONIZE;
int doScheduler(struct timeval *tv,int timeout,int source,ClnHdr *tgtcln,ContractListStruct *tgtcon)
{
  ContractListStruct *cl; /* improve readability by assignment */
  ClnLst *c;
  short cc,i,fmt,fmtoutsize;
  UINT32 doutsize;
  SINT32 delta_t;
  double dts;
  int isScheduled, isStale, isDataChange=FALSE, chkSignal=FALSE, 
      isExpressDelivery=FALSE, clockAdjusted=FALSE, applyTsOffset=TRUE;
  int nstale, preStaleCntr, msec, blimit, nactive=0;
  struct timeval tv_stamp;
  char dbkey[256]="";
  
# ifndef MULTITHREADED
  UNUSED(timeout);
# endif
  if (source == SCHED_INTERN) gServerCycleState = CYCLE_STATE_SCHED;
  ProbeTimeElapsed("doScheduler entered");
  if (tgtcon != NULL) isExpressDelivery = TRUE;
  if (!isExpressDelivery) 
  {
    if (WaitForSystemMutex(hSystemKernelMutex,timeout) != 0) return not_ready;
    lastScheduled = *tv;
  }
  for (i=0; i<ncontracts; i++)
  {
    if (isExpressDelivery) 
    { /* a target was specified */
      if (i > tgtcon->listId) break; /* due to a continue ? */
      i = tgtcon->listId;
    }
    if ((cl=ContractList[i])->expired == CONTRACT_AVAILABLE) continue;/* just a place-holder */
    if (tineDebug) getContractKey(dbkey,i);
    if (cl->expired == CONTRACT_ACTIVE) nactive++;
    if (source == SCHED_EXTERN && cl->runInSeparateThread) continue; /* would have grave consequences ! */
    if (cl->nclients == 0 && cl->expired == CONTRACT_ACTIVE) 
    { /* no clients, not marked as expired */
      if (cl->contract.EqmAccess & CA_LAST) 
      { /* going out of scope */
        if (tineDebug > 1) dbglog("[%d] %.256s going out of scope with no clients",i,dbkey);
        while (doContract(cl,TRUE) == not_ready) YieldThread(); 
      }
      if (removeContractTableEntry(i) == 0) nExpiredContracts++;
      continue;
    }
    if (cl->prp != NULL && cl->prp->sigfcn != NULL) chkSignal = TRUE;
    cc = isStale = isScheduled = 0;
    if ((fmtoutsize=fmtsizeof(cl->contract.EqmFormatOut)) > 0)
    {
      fmt = LFMT(cl->contract.EqmFormatOut);
      if (fmt == CF_MDA) fmtoutsize = getCarriedFormatSize(cl->contract.strTagOut);
      if (isAdjustableLengthFormat(fmt) || fmt == CF_STRUCT)
      { /* these types can make adjustments to the output capacity */
        doutsize = cl->dataOutCapacity;
      }
      else
      { /* some types might have 'headers' ... */
        doutsize = cl->contract.EqmSizeOut*fmtoutsize;
        doutsize += getFormatHeaderSize(fmt);
      }
    }
    else 
    { /* no output data */
      doutsize = 0;
    }
    if ((delta_t=MSECS(*tv,cl->lasttime)) < 0) 
    { /* system clock set back ! */
      cl->lasttime = *tv; 
      clockAdjusted = TRUE;
      if ((int)gDataTimeStampOffset != 0 && applyTsOffset)
      { /* is receiving TINE time updates and has an offset (probably negative) */
        gDataTimeStampOffset -= ((double)delta_t)/1000.0;
        feclog("synchronization correction (applied): system clock set back %d msec -> adjust data timestamp offset to %g",delta_t,gDataTimeStampOffset);
        applyTsOffset = FALSE;
      }
      delta_t = cl->pollingInterval; /* get past the next check */
    }
    if (delta_t < cl->pollingInterval)
    { /* no need to call the handler */
      if (isExpressDelivery && cl->returnCode & CE_SENDDATA)
      { /* but if this is true then ... */
        renegotiateContract(cl, cl->returnCode);
      }
      continue;
    }
    if (cl->expired == CONTRACT_EXPIRED && cl->nclients == 0)
    {
      if (delta_t > 3*cl->pollingInterval) removeContractTableEntry(i);
      continue;
    }
    if (delta_t < MAXALLOWED_POLLINTVL && delta_t >= 2*cl->pollingInterval)
    {
      msglog(0,"[%d] %.256s missed last schedule (%d vs %d)",
        i,dbkey,(int)delta_t,(int)(2*cl->pollingInterval));
      cl->misses++;
      gContractMisses++;
      if (chkSignal) sendPrpSignal(cl->prp,CS_LATE,cl->contractID,cl->compStatus);
      sendContractSignal(cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,CS_LATE,cl->compStatus);
    }
    if (cl->deliveryPending > 0) 
    { /* still sending out the last round (probably) */
      nstale = 0;
      for (c=cl->clnLst; c != NULL; c=c->next) 
      {
#       ifndef SMALL_TINE_LIB
        if (!isValidClient(c->client))
        {
          removePhantomClientFromContracts(c->client,"not found in client table",TRUE);
        }
        else
#       endif
        { /* count how many are still stale */
          if (c->status.Stale) nstale++;
        }
      }
      if (nstale > 0)
      { /* yes, we do seem to still be delivering data ... */
        if (tineDebug > 1) dbglog("[%d] %.256s still pending",i,dbkey);
        StaleData++;
        cl->delays++;
        gContractDelays++;
        if (chkSignal) sendPrpSignal(cl->prp,CS_PENDING,cl->contractID,cl->compStatus);
        sendContractSignal(cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,CS_PENDING,cl->compStatus);
        continue; 
      }
      /* if I'm here: then I'm not delivering anything to anybody ! */
      cl->deliveryPending = 0;
    }
    if (cl->starttime.tv_sec == 0 && cl->starttime.tv_usec == 0) cl->starttime = *tv;
    tv_stamp = *tv; 
    msec = (int)(tv->tv_usec/1000);
    /* Synchronize on 1000 msec boundaries if possible */
    if (!(cl->pollingInterval%1000) && 
        ((int)cl->dataSizeOut < gPacketMtu - SUBRSPHDR_SIZE) &&
        gSynchronizeContracts && 
        msec < (1000 - (int)gSystemTick)*2) tv_stamp.tv_usec = 0;
    dts = PutDataTimeStampU(gDataTimeStampOffset,tv->tv_sec,(int)tv->tv_usec);
    setDataTimeStamp(dts);
    if (cl->nclients == 1 && cl->clnLst->status.counter == 1 && !(cl->accessFlags&CA_REPEAT))
    { /* dangling client going out of scope */
      cl->contract.EqmAccess |= CA_LAST;
      if ((c=cl->clnLst) != NULL && BASEMODE(c->status.mode) != CM_SINGLE)
      { /* make sure this is handled in doDelivery (remove client from contract) */
        c->status.Stale = CX_TIMER; 
        StaleData++;
      }
    }
    ProbeTimeElapsed("doScheduler calling contract");
    if (source == SCHED_INTERN) gServerCycleState = CYCLE_STATE_EQM;
    switch (cc=doContract(cl,TRUE))
    {
      case 0: 
        break;
      case not_ready:
        /* call returned not_ready or still not_signalled */
        cl->contract.EqmAccess &= ~(CA_LAST); /* remove the CA_LAST bit */
        cl->accessFlags |= CA_REPEAT;
        msec = MSECS(*tv,cl->starttime);
        for (c=cl->clnLst; c != NULL; c=c->next) c->status.Stale = 0;
        cl->deliveryPending = 0;
        if (msec > cl->pollingInterval + 500)
        { /* thread still busy */
          if (tineDebug > 2) dbglog("%.256s not ready for %d msecs",dbkey,msec);
        }
        ProbeTimeElapsed("doScheduler contract returned not ready");
        continue; /* don't send out ever! : it has to complete in its own way ! */
      case access_denied:
        memcpy(cl->compString,cl->exp->EqmLastErrorString,STATUS_SIZE);
        goto err;
      case has_bitfield_tag:
      case has_structure_tag:
        if (cl->exp == NULL) ccerr(non_existent_elem);
        memcpy(cl->compString,cl->exp->EqmLastErrorString,STATUS_SIZE);
        goto err;
      case call_redirection:
      case server_redirection:
        if (cl->exp == NULL) ccerr(non_existent_elem);
        if (!assertRedirectionValid(dbkey,cl->contract.EqmName, cl->exp->EqmLastErrorString)) ccerr(configuration_error);
        memcpy(cl->compString,cl->exp->EqmLastErrorString,STATUS_SIZE);
      default: 
        if (doutsize > 0 && (cl->dataOut == NULL || cl->dataBuf == NULL))
        {
          cc &= 0x0fff;
          goto err;
        }
        if (cc & 0xf000) 
        { /* CE_RENEGOTIATE or CE_SENDDATA */
          doutsize = renegotiateContract(cl,(cc & 0xfff));
          if (doutsize == 0) cc &= ~CE_SENDDATA; /* no data to send */
          cc &= ~CE_RENEGOTIATE;
        }
        else goto err;
    }
    if (source == SCHED_INTERN) gServerCycleState = CYCLE_STATE_SCHED;
    if (cl->contract.EqmAccess & CA_WRITE)
    {
      appendWriteAccessTable(cl);
    }
    if (doutsize && cl->expired == CONTRACT_ACTIVE && 
        (cl->unbuffered || memcmp(cl->dataOut,cl->dataBuf,(size_t)doutsize)))
    { /* Data have changed: the cached data are dirty */
      if (!cl->unbuffered) memcpy(cl->dataOut,cl->dataBuf,(size_t)doutsize);
      /* for the adjustable types: this will also copy the serialized data to the dataOut area ! */
      isStale = isDataChange = 1;                       
    }                                          
    else
    { /* Cached data have not changed */
      if (MSECS(*tv,cl->lastmarked) > (HBT_MAXIMUM-1)*1000L) isStale = 1;
      isDataChange = 0;                               
    }
err:
    ProbeTimeElapsed("doScheduler contract finished");
    if (cc) isStale = 1;
    if (cl->lasttime.tv_sec == PRP_SCHEDULE_SIGNAL) isScheduled = TRUE;
    if (isStale) cl->lastmarked = isScheduled ? *tv : tv_stamp; /* Last marked as Stale */
    if (cc == not_ready)
    { /* not_ready for longer than the client wait period */
      if (tineDebug) dbglog("return code not_ready changed to operation_timeout");
      cc = operation_timeout;
    }
    else
    { /* Last polled */
      cl->lasttime = isScheduled ? *tv : tv_stamp;                   
      cl->compStatus = not_signalled;
    }
    cl->starttime.tv_sec = cl->starttime.tv_usec = 0;
    cl->returnCode = cc;

    for (c=cl->clnLst; c != NULL; c=c->next)
    {
      if (c->status.counter == 0)              
      { /* scheduled for removal */
        StaleData++; /* ensure pass thru doDelivery */
        continue; 
      }
      c->status.statusCode = cc;
      if (isScheduled) 
      { /* scheduled: flag the timestamp as needed */
        c->status.lasttime.tv_sec = c->status.lasttime.tv_usec = PRP_SCHEDULE_SIGNAL;
        c->status.Stale |= CX_EVENT;
      }
      else if (clockAdjusted) 
      { /* system clock set back : flag the timestamp as needed */
        c->status.lasttime.tv_sec = cl->lasttime.tv_sec;
        c->status.lasttime.tv_usec = cl->lasttime.tv_usec;
      }
      if (cc) c->status.Stale |= CX_ERROR;      /* Stale because of error code */
      if (c->status.mode == CM_DATACHANGE || c->status.mode == CM_EVENT)
      {
        if (isScheduled && c->status.mode == CM_EVENT) 
        { /* is an event and was requested as event */
          c->status.Stale |= CX_EVENT;
        }
        else if (isStale)                             
        { /* stale because of data change or system heartbeat */
          c->status.Stale |= (isDataChange ? CX_STALE : CX_HEARTBEAT);
          if (isDataChange && gRequireAcknowledgments && (c->client->inetProtocol&(UDP|UDP6)))
            c->status.counter = CTR_ACK_REQUEST; 
        }
      }
      else /* CM_TIMER or CM_SINGLE */
      { /* honor polling rate   */
        if (MSECS(*tv,c->status.lasttime) > c->status.pollingInterval - (SINT32)gSystemTick)
        { /* Stale because of polling rate */
          c->status.Stale |= (isDataChange ? CX_STALE : CX_TIMER);
        }
      }
      if (c->status.Stale)                     
      { /* consumer for this contract is to be refreshed */
        cl->signalPending = TRUE;
        cl->deliveryPending++;
        c->status.lasttime = isScheduled ? *tv : tv_stamp;
        StaleData++; /* Ensure pass thru doDelivery */
      }
      if (c->status.counter == CTR_ACK_REQUEST)    
      { /* get serious until an acknowledgment comes */
        c->status.mode = HBT_ACK_PENDING; /* Reset by renewal */
        c->status.pollingInterval = cl->pollingInterval;
      }
    }
    if (isExpressDelivery) break;
  }
  if (!isExpressDelivery && nExpiredContracts != ncontracts - nactive)
  { /* only do this check if NOT an express delivery */
    if (tineDebug) dbglog("adjust number expired contracts from %d to %d",nExpiredContracts,ncontracts-nactive);
    nExpiredContracts = ncontracts - nactive;
  }
  gettimeofday(tv,(struct timezone *)NULL);
  for (i=0; StaleData; i++)
  {
    preStaleCntr = StaleData;
    ProbeTimeElapsed("doScheduler delivering results");
    if ((blimit=doDelivery(&thisCycleTime,tgtcln,tgtcon,FALSE)) > 0)
    { /* doDelivery returns the burst limit to apply */
      if (i > blimit)
      { /* burst limit exceeded => impose a delay */
        while (doDelivery(&thisCycleTime,NULL,NULL,TRUE) < 0); /* but keep servicing the TCP clients */
        if (gCycleMicroDelay > 0) microsleep(gCycleMicroDelay);
        else if (gCycleDelay) millisleep(gCycleDelay);
        gBurstLimitReachedCount++;
        if (tineDebug) dbglog("server burst limit %d reached",gBurstLimit);
        ProbeTimeElapsed("doScheduler delivery burst limit reached");
        i = 0;
      }
    }
    else
    { /* TCP connections often do not use burst limits! re-establish the global Stale counter */
      StaleData = preStaleCntr;
      i = 0;
    }
  }
  if (gConSigfcn != NULL) chkSignal = TRUE;
  if (chkSignal)
  {
    for (i=0; i<ncontracts; i++)
    { /* gotta go thru the list and check */
      if (tgtcon != NULL) 
      { /* a target was specified */
        if (i > tgtcon->listId) break; /* due to a continue ? */
        i = tgtcon->listId;
      }
      if ((cl=ContractList[i])->expired == CONTRACT_AVAILABLE) continue;
      /* general signal */
      if (!cl->signalPending) continue;
      sendContractSignal(cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,CS_SENT,cl->returnCode);
      if (cl->prp != NULL && cl->prp->sigfcn != NULL) sendPrpSignal(cl->prp,CS_SENT,cl->contractID,cl->returnCode);
      cl->signalPending = FALSE;
    }
  }
  if (!isExpressDelivery) ReleaseSystemMutex(hSystemKernelMutex);
  ProbeTimeElapsed("doScheduler returning");
  return 0;
}

int prepOutgoingSubResponseHdr(ClnLst *cn, ContractListStruct *cls, UINT32 msgsize, BYTE *stream)
{
  switch (cn->client->tineProtocol)
  {
    case 5:
    { /* legacy: release 3.xx era */
      PrdrHdr5 PHdr5, *pPHdr5 = &PHdr5;
      pPHdr5->msgsize = msgsize;
      pPHdr5->retCode = cn->status.statusCode;
      pPHdr5->timestampMSEC = cls->dtimestampUSEC / 1000;
      pPHdr5->EqmFormat = cls->contract.EqmFormatOut;
      pPHdr5->timestamp = cls->dtimestamp;
      pPHdr5->ClnStarttime = cn->status.starttime;
      pPHdr5->subId = cn->status.id;
      pPHdr5->counter = (BYTE)cn->status.counter;
      pPHdr5->mtu = cn->status.mtu;
      pPHdr5->blknum = cn->status.blknum;
      pPHdr5->numblks = cn->status.numblks;
#     ifdef INTEL
      memcpy(stream, pPHdr5, sizeof(PrdrHdr5));
#     else
      _structSwap(PrdrHdr5Struct, (BYTE *)stream, 0, NULL, (BYTE *)pPHdr5, LTOR);
#     endif
      break;
    }
    case 6:
    { /* legacy: release 4.xx era */
      PrdrHdr PHdr, *pPHdr = &PHdr;
      pPHdr->msgsize = msgsize;
      pPHdr->retCode = cn->status.statusCode; /* call completion code */
      pPHdr->timestampUSec = cls->dtimestampUSEC;
      pPHdr->EqmFormat = cls->contract.EqmFormatOut;
      pPHdr->timestamp = cls->dtimestamp;
      pPHdr->ClnStarttime = cn->status.starttime;
      pPHdr->subId = cn->status.id;
      pPHdr->counter = (UINT16)cn->status.counter;
      pPHdr->mtu = cn->status.mtu;
      pPHdr->blknum = cn->status.blknum;
      pPHdr->numblks = cn->status.numblks;
      pPHdr->tineProtocol = cn->client->tineProtocol;
      pPHdr->xferReason = cn->status.Stale;
      pPHdr->stssize = cn->status.statusCode == 0 ? 0 : STATUS_SIZE;
      pPHdr->stscode =  cls->returnCode; /* return code of the contract execution */
      pPHdr->userstamp = cls->datastamp;   /* call applied data stamp */
      pPHdr->sysstamp = cls->sysstamp;  /* system data stamp */
      if (isJumboDelivery(cn->client)) pPHdr->blknum = pPHdr->numblks = 1;
#     ifdef INTEL
      memcpy(stream, pPHdr, sizeof(PrdrHdr));
#     else
      _structSwap(PrdrHdrStruct, (BYTE *)stream, 0, NULL, (BYTE *)pPHdr, LTOR);
#     endif
      break;
    }
    case 7:
    {
      SubRspHdr subhdr, *phdr = &subhdr;
      phdr->msgSize = msgsize;
      phdr->retCode = cn->status.statusCode; /* call completion code */
      phdr->timestampUSec = cls->dtimestampUSEC;
      phdr->dataFormat = cls->contract.EqmFormatOut;
      phdr->timestamp = cls->dtimestamp;
      phdr->starttime = cn->status.starttime;
      phdr->subId = cn->status.id;
      phdr->counter = (UINT16)cn->status.counter;
      phdr->mtu = cn->status.mtu;
      phdr->blknum = cn->status.blknum;
      phdr->numblks = cn->status.numblks;
      phdr->tineProtocol = cn->client->tineProtocol;
      phdr->xferReason = cn->status.Stale;
      phdr->stssize = cn->status.statusCode == 0 ? 0 : STATUS_SIZE;
      phdr->stscode = cls->returnCode; /* return code of the contract execution */
      phdr->userstamp = cls->datastamp;   /* call applied data stamp */
      phdr->sysstamp = cls->sysstamp;  /* system data stamp */
      phdr->dataSize = cls->contract.EqmSizeOut;
      phdr->msgSizeInElements = cls->dataSizeOut;
      if (isJumboDelivery(cn->client)) phdr->blknum = phdr->numblks = 1;
#     ifdef INTEL
      memcpy(stream, phdr, sizeof(SubRspHdr));
#     else
      _structSwap(SubRspHdrStruct, (BYTE *)stream, 0, NULL, (BYTE *)phdr, LTOR);
#     endif
      break;
    }
    default:
      return illegal_protocol;
  }
  return 0;
}
#ifndef CONSISTENCY_CHECK_THRESHOLD
# define CONSISTENCY_CHECK_THRESHOLD 10
#endif
#define maxDeadTime(m,p) ((m) == CM_DATACHANGE || (m) == CM_EVENT ? HBT_MAXIMUM + (p) : 2 * (p))
int isJumboDelivery(ClnHdr *cln)
{
  switch (cln->inetProtocol)
  {
    case STREAM6:
    case STREAM: return TRUE;
    case MMF: if (gUseMmfJumbos) return TRUE;
    default: return FALSE;
  }
}
int isJumboDeliveryAllowed(ClnHdr *cln,UINT32 siz)
{
  switch (cln->inetProtocol)
  {
    case STREAM6:
    case STREAM: return TRUE;
    case MMF: 
#     if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
      if (gUseMmfJumbos && siz <= gShmBufferSize) return TRUE;
#     else
      UNUSED(siz);
#     endif
    default: return FALSE;
  }
}
int codeNeedsSingleDelivery(int cc)
{
  switch(cc)
  {
    case server_redirection:
    case call_redirection:
    case has_structure_tag:
    case has_bitfield_tag:
      return TRUE;
    default:
      return FALSE;
  }
}
void prepOutgoingResponseHdr(ClnHdr *cln, RspHdr *hdr, BYTE *stream)
{
  if (cln == NULL) return;
  if (hdr->tineProtocol < MINIMUM_REQHDR_PROTOCOL)
  { /* map to modern request header */
    if (isJumboDelivery(cln))
    {
      UINT32 lmsgsize = hdr->totalSizeInBytes;
#     if !defined(INTEL) && !defined(VMS)
      ISWAP(lmsgsize);
#     endif
      memcpy(stream, &lmsgsize, 4);
    }
    else
    {
      *(UINT16 *)stream = (UINT16)SPEEK(hdr->totalSizeInBytes);
    }
  }
  else
  {
    hdr->headerSizeInBytes = sizeof(RspHdr);
    hdr->endianness = BO_LITTLE_ENDIAN;
    hdr->encoding = CE_ASCII;
    strncpy(hdr->fecName, gFecName, FEC_NAME_SIZE);
#   ifdef INTEL
    memcpy(stream, hdr, sizeof(RspHdr));
#   else
    _structSwap(RspHdrStructStruct, stream, 0, NULL, (BYTE *)hdr, LTOR);
#   endif
  }
  return;
}

int doDelivery(struct timeval *tv,ClnHdr *tgtcln,ContractListStruct *tgtcon,int tcponly)
{ /* Delivery system */
  static int consistencyCheckInterval = 0;
  ClnHdr *cln;
  ClnLst *c;
  ContractListStruct *cl; /* improve readability by assignment */
  ClnSts cs;
  RspHdr rsphdr;
  short i,k,n,nsegs,n2send,fsize,snddata;
  SINT32 extsize,extptr,dsize,stssize = 0;
  UINT32 msgsize,ssize=0,totalsize=0,hsize,stepsize,packetmtu;
  UINT32 dptr, delta_t;
  short fmt,basemode,nscan;
  BYTE *buf = srvWorkArea;
  char addr[64], dbkey[256]="";
  int actionTaken=FALSE,dataDelivered=FALSE,clientFound=FALSE,scanAll=FALSE,mark=FALSE,breakscan,clientDeliveryFinished;
  int swallowIt=FALSE, istcp=FALSE, isjumbo=FALSE, payloadReady=FALSE;
  int sendSingleDelivery=FALSE, useBurstLimit=gBurstLimit;
  if (tgtcln == NULL && ++consistencyCheckInterval == CONSISTENCY_CHECK_THRESHOLD)
  { /* do not scan all for express delivery items ! */
    consistencyCheckInterval = 0;
    scanAll = TRUE;
  }
  for (k=0; k<nconsumers; k++)
  { /* top most loop over all registered clients */
    if ((cln=ClnTbl[k]) == NULL) continue;
    if (tgtcln != NULL && tgtcln != cln) continue; /* find the targeted table entry */
    if (cln->ncontracts == 0 && nBusyContractProgress == 0)
    { /* client has no more contracts : remove it */
      removeClientTableEntry(k--);
      continue;
    }
    istcp = (cln->inetProtocol&(TCP|TCP6|STREAM|STREAM6)) ? TRUE : FALSE;
    if (tcponly && !istcp) continue;
    isjumbo = isJumboDelivery(cln);
    clientFound = tgtcln == NULL ? FALSE : TRUE;
    clientDeliveryFinished = TRUE; /* be optimistic */
    hsize = (UINT32)RETURNHDR_SIZE(cln->tineProtocol);
    totalsize = hsize;
    nscan = scanAll ? cln->ncontracts : ncontracts;
    nsegs = n2send = 0; sendSingleDelivery = FALSE;
    for (i=0,n=0,breakscan=FALSE; i<ncontracts && n<nscan && !breakscan; i++)
    { /* scan the contract table for things this client is interested in */
      if (tgtcon != NULL)
      { /* a target contract was specified */
        if (i > tgtcon->listId) break;
        i = tgtcon->listId;
      }
      cl = ContractList[i]; dataDelivered = FALSE;
      if (cl->expired == CONTRACT_AVAILABLE)
      { /* nothing to deliver */
        cl->deliveryPending = 0;
        continue;
      }
      for (c=cl->clnLst; c != NULL; c=c->next) if (c->client == cln) break;
      if (c == NULL) 
      { /* this contract does not service this client */
        continue;
      }
      if (istcp && cln->ep.sck == 0)
      { /*  TCP connection error -> remove it */
        c->status.counter = 0;
        isjumbo = FALSE;
      }
      if (tineDebug) getContractKey(dbkey,cl->listId);
      n++; snddata = TRUE; extptr = extsize = 0; clientFound = TRUE; mark = FALSE;
      basemode = c->status.mode;
      if (isServiceRequest(&cl->contract) && c->status.statusCode) swallowIt = TRUE;
      if (c->status.Stale)
      { /* marked as stale -> send to client */
        sendSingleDelivery = codeNeedsSingleDelivery(c->status.statusCode);
        if (n2send > 0 && sendSingleDelivery) 
        { /* e.g. late redirection: cannot pack this result with others ... */
          clientDeliveryFinished = FALSE; 
          continue;
        }
        if (istcp) useBurstLimit = gTcpBurstLimit;
        actionTaken = TRUE; dataDelivered = TRUE;
        stepsize = SUBRSP_STEPSIZE(c->client->tineProtocol);
        packetmtu = hsize + stepsize + c->status.mtu;
        snddata = HAS_DATA(c->status.statusCode) ? TRUE : FALSE;
        stssize = STSSIZE(c->client->tineProtocol);
        if (snddata)
        { /* status okay or returning data */
          dsize = c->status.nextDataSize;
          dptr  = c->status.nextDataPtr;
        }
        else
        { /* record current status settings and send error report */
          memcpy(&cs,&c->status,sizeof(ClnSts));
          extsize = stssize;
          dsize = extsize;
          dptr  = 0;
          c->status.numblks = 1;
          c->status.bytesRemaining = c->status.bytesTotal = dsize;
          c->status.nextDataSize = c->status.bytesRemaining;
        }
        msgsize = dsize + stepsize;
        if (!isjumbo && msgsize > packetmtu) { c->status.counter = 0; continue; }
        prepOutgoingSubResponseHdr(c, cl, msgsize, &buf[totalsize]);
        if (c->status.numblks > 1 && c->status.blknum == 1) mark = TRUE;
        if (snddata && (fsize=fmtsizeof(cl->contract.EqmFormatOut)) != 0)
        {
          int is_mda = LFMT(cl->contract.EqmFormatOut) == CF_MDA ? TRUE : FALSE;
          if (is_mda) fsize = getCarriedFormatSize(cl->contract.strTagOut);
          if ((c->status.statusCode & 0xfff))
          {
            extsize = (SINT32)(c->status.nextDataPtr + c->status.nextDataSize -
                               cl->contract.EqmSizeOut*fmtsizeof(cl->contract.EqmFormatOut));
            if (extsize > stssize) extsize = stssize; /* shouldn't be possible */
            if (extsize > dsize) 
            { /* err string remainder  */
              extsize = dsize;
              extptr  = stssize - extsize;
            }
            if (extsize > 0) dsize -= extsize; else extsize = 0;
            if (isjumbo) 
            { /* if CM_STREAM: always set to the following */
              extsize = stssize; extptr = 0; 
            }
          }
          if (!isjumbo)
          { /* can't be a CM_STREAM */
            ssize = dsize / fsize; 
            fmt = (short)LFMT(cl->contract.EqmFormatOut);
            if (isAdjustableLengthFormat(fmt)) ssize = (UINT16)atoi(cl->contract.strTagOut);
            if (c->status.blknum > 1) 
            { /* long contract : possible SPECTRUM or IMAGE or MDA */
              fmt = is_mda ? ftoi(cl->contract.strTagOut) : GetFormatDataType(fmt);
            }
            prepOutgoingData(&buf[totalsize+stepsize], &cl->dataOut[(int)dptr], (UINT32)dsize, fmt);
          }
          else
          { /* CM_STREAM: slightly different delivery strategy */
            ssize = c->status.nextDataSize / fsize; 
            fmt = (short)LFMT(cl->contract.EqmFormatOut);
            if (isAdjustableLengthFormat(fmt)) ssize = (UINT32)atoi(cl->contract.strTagOut);
            prepOutgoingData(&buf[totalsize+stepsize], &cl->dataOut[0], c->status.nextDataSize, fmt);
          }
        }
        if (!isjumbo && (totalsize + msgsize) > packetmtu) 
        { /* this data segment won't fit in this packet */
          clientDeliveryFinished = FALSE; 
          goto bok; 
        }
        totalsize += msgsize;
        if (tineDebug)
        { /* debug output */
          int dbgIt = (tineDebug > 2 || c->status.mode <= CM_SINGLE) ? TRUE : FALSE;
          if (tineDebug > 1 && c->status.blknum == c->status.numblks) dbgIt = TRUE;
          if (dbgIt)
          {
            getPeerAddrStr(addr,c->client);
            dbglog("[%d] %.256s <%d> %d bytes %d elems (blk %d of %d) to %.16s @ %.32s (+%d)",
              i,dbkey,c->status.statusCode,msgsize,ssize,c->status.blknum,c->status.numblks,
              c->client->usr,addr,c->status.counter);
          }
        }
        if (extsize > 0 && cl->exp != NULL)                      
        { /* insert err string */
          if (c->status.statusCode == server_redirection ||
              c->status.statusCode == call_redirection) 
          { /* insert redirection information and break out */
            if (c->client->tineProtocol < 6)
            {
              memcpy(cl->exp->EqmLastErrorString,&cl->compString[PRP_REDIR_SRV_OFFSET],16);
              memcpy(&cl->exp->EqmLastErrorString[16],&cl->compString[PRP_REDIR_PRP_OFFSET],16);
            }
            else
            {
              memcpy(cl->exp->EqmLastErrorString,cl->compString,STATUS_SIZE);
            }
            c->status.counter = 1;
          }
          else if (c->status.statusCode == has_structure_tag || c->status.statusCode == has_bitfield_tag)
          { /* insert the tag and break out */
            memcpy(cl->exp->EqmLastErrorString,cl->compString,TAG_NAME_SIZE);
          }
          else if ((c->status.statusCode & 0xfff) < numErr) 
          { /* insert system error code */
            switch (c->status.statusCode)
            {
              case access_denied:
                if (strlen(cl->compString) > 0) break;
              case data_not_local: c->status.counter = 1; 
              default: 
                SetEqmCompletion(cl->exp->EqmName,erlst[c->status.statusCode & 0xff]);
                break;
            }
          }
          else if (strlen(cl->exp->EqmLastErrorString) == 0)
          { /* is a user registered code ? */
            GetErrorText(c->status.statusCode,cl->exp->EqmLastErrorString,32);
          }
          if (!isjumbo)
          { /* normally not a stream connection */
            memcpy(&buf[totalsize-(int)extsize],&cl->exp->EqmLastErrorString[(int)extptr],(size_t)extsize);
          }
          else
          { /* slightly different delivery strategy */
            memcpy(&buf[hsize+stepsize+c->status.nextDataSize],&cl->exp->EqmLastErrorString[(int)extptr],(size_t)extsize);
          }
        }
        if (!snddata || c->status.numblks == c->status.blknum)
        { /* datablock finished and/or error code */
          n2send++;
          if (!snddata)
          { /* reapply status settings in case of temporary error for a persistent link */
            c->status.bytesRemaining = cs.bytesRemaining;
            c->status.bytesTotal = cs.bytesTotal;
            c->status.numblks = cs.numblks;
            c->status.nextDataSize = cs.nextDataSize;
          }
          c->status.Stale = 0;
          if (c->status.counter > 0 ) c->status.counter--;
          if (c->status.mode == CM_CANCEL) c->status.counter = 0;
          c->status.blknum = 1;
          c->status.nextDataPtr = 0;
          c->status.bytesRemaining = c->status.bytesTotal;
          ReturnToServerCycle = 0;
          cl->deliveryPending = 0;
          if ((delta_t=MSECS(*tv,c->status.lasttime)) > (UINT32)maxDeadTime(c->status.mode,c->status.pollingInterval))
          {
            if (tineDebug > 1)
            {
              dbglog("[%d] %.256s -> returning late data (%d > %d ms)",
                i,dbkey,(int)delta_t,c->status.pollingInterval);
            }
            c->status.misses++;
            gClientMisses++;
          }
        }
        else
        { /* more data segments to send ... */
          ReturnToServerCycle = c->status.numblks - c->status.blknum;
          c->status.blknum++;
          c->status.nextDataPtr += c->status.nextDataSize;
          c->status.bytesRemaining = c->status.bytesTotal - c->status.nextDataPtr;
          StaleData++;
        }
        if (!isjumbo) c->status.nextDataSize = MIN(c->status.mtu,c->status.bytesRemaining);
        if (c->status.numblks > 1) breakscan = TRUE;
      }
      nsegs = n;
      if (c->status.counter == 0) 
      {
        actionTaken = TRUE;
        if (removeClientFromContract(c,i) == 0) n--;
      }
      if (cl->nclients == 0) 
      { /* the last client has been removed */
        actionTaken = TRUE;
        if (cl->signalPending)
        {
          sendContractSignal(cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,CS_SENT,cl->returnCode);
          if (cl->prp != NULL && cl->prp->sigfcn != NULL) sendPrpSignal(cl->prp,CS_SENT,cl->contractID,cl->returnCode);
          cl->signalPending = FALSE;
        }
        if (gRetardSingleContractRemoval && basemode == CM_SINGLE) 
        { /* mark as expired, but keep in tables for ~ 1 second */
          if (cl->expired != CONTRACT_EXPIRED) nExpiredContracts++;
          cl->expired = CONTRACT_EXPIRED;
          cl->contract.EqmAccess &= ~(CA_FIRST|CA_LAST);
          cl->pollingInterval = DEFAULT_CLIENT_POLLRATE;
          cl->clnLst = NULL;
        }
        else 
        { /* get rid of it now ... */
          if (removeContractTableEntry(i) == 0) nExpiredContracts++;
        }
      }
      if (isjumbo && actionTaken) goto bok;
      if (!dataDelivered) cl->deliveryPending = 0;
      if (sendSingleDelivery)
      {
        clientDeliveryFinished = FALSE; 
        goto bok; 
      }
    } /* end contracts loop */
    if (scanAll && !breakscan && n != cln->ncontracts)
    { /* can do consistency check */
      if (tineDebug) 
      {
        inetStrAdr(&cln->adr.ip,addr,FALSE);
        dbglog("adjust number of contracts for client %.16s @ %.40s from %d to %d",
          cln->usr,addr,cln->ncontracts,n);
      }
      cln->ncontracts = n;
    }
#   ifndef SMALL_TINE_LIB
    if (!clientFound)
    { /* this client not found in any contract */
      removePhantomClientFromContracts(cln,"not found in contract tables",TRUE);
    }
#   endif
bok:if (cln != NULL)
    { /* could have been removed ? */
      rsphdr.tineProtocol = cln->tineProtocol;
      rsphdr.numberSegments = nsegs;
      if (totalsize > hsize)
      { /* normal case */
        rsphdr.totalSizeInBytes = totalsize;
        payloadReady = TRUE;
        if (isjumbo) clientDeliveryFinished = TRUE;
      }
    }
    if (payloadReady)
    {
      if (!swallowIt)
      {
        prepOutgoingResponseHdr(cln, &rsphdr, buf);
        sendToPeer(cln, buf, rsphdr.totalSizeInBytes, SRV_RSP_, mark);
      }
      actionTaken = TRUE;
      payloadReady = FALSE;
    }
    if (!clientDeliveryFinished) 
    { /* still more data to send to this client */
      k--;
    }
    else 
    {
      if (actionTaken && istcp) break;
    }
  }
  gettimeofday(tv,(struct timezone *)NULL);
  if (!actionTaken)
  { /* nothing done => nothing to do */
    if (!tcponly) StaleData = 0;
  }
  else 
  { /* there was someting to do */
    if (tcponly || isjumbo) return -1; /* special signal */
  }
  if (StaleData > 0 && !tcponly) StaleData--;
  return useBurstLimit;
}
int sendToPeer(ClnHdr *c,BYTE *payload,UINT32 msgsize,int msgtype,int mark)
{
  int cc=0, retry=3;
# if defined(MULTITHREADED)
  TCPBCKT *bckt;
  SEMHANDLE sem = (SEMHANDLE)0;
# else
  UNUSED(msgtype);
  UNUSED(mark);
# endif
  if (c == NULL) return argument_list_error;
  switch (c->inetProtocol)
  {
#   if defined(MULTITHREADED)
    case STREAM6:
    case STREAM:
      if (c->ep.sck == 0) ccerr(gClientIsClosing ? 0 : tcp_connect_error);
      if (msgtype == SRV_RSP_) 
      {
        cc = sendStreamCreateThread(c,payload,msgsize) == 0 ? runtime_error : 0;
        break;
      }
      /* else fall thru ... */
#   endif
    case TCP6:
    case TCP:
      if (c->ep.sck == 0) ccerr(gClientIsClosing ? 0 : tcp_connect_error);
#     if defined(MULTITHREADED)
      if (gRunTransportInSeparateThread && msgtype == SRV_RSP_) 
      {
        if ((bckt=findBucket(c->ep.sck,FALSE)) != NULL) sem = bckt->sndSem;
        if ((cc=putStreamParams(c,payload,msgsize,mark)) != 0) break;
        ReleaseSystemSemaphore(sem);
        cc = 0;
        break;
      }
#     endif
      /* else fall thru ... */
    case PIPE:
    case UDP6:
    case UDP:
      if (c->ep.sck == 0) ccerr(gClientIsClosing ? 0 : udp_socket_error);
      while (retry-- > 0)
      {
        if ((cc=sendIpData(c,payload,msgsize,msgtype,c->inetProtocol,TRUE,TRUE)) != cannot_select) break;
      }
      break;
#   if defined(IPXNW) || defined(IPXBSD)
    case IPX:
    case SPX:
      cc = sendIPXData(c,payload);
      break;
#   endif
#   if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
    case MMF:
      cc = sendShmData(c,payload,(int)msgsize,(UINT)(msgtype == SRV_RSP_ ? WM_RCV_LCL : WM_REQ_LCL));
      break;
#   endif
    default:
      cc = illegal_protocol;
      break;
  }
err:
  return cc;
}
void CleanupStrandedDatagrams(void)
{
  int i;
  nBusyContractProgress = 0;
  for (i=0; i<MAXLONGGRAMS; i++) if (conprg[i].busy)
  { 
    nBusyContractProgress++;
    if (thisCycleTime.tv_sec > (SINT32)conprg[i].blktime + 5)
    {
      SystemFree(conprg[i].data);
      memset(&conprg[i],0,sizeof(ContractProgress));
      if (tineDebug > 1) dbglog("stale long datagram %d",i);
    }
  }
}
int GetDefaultFormatAndSize(CONTRACT *con,char *status)
{
  int idx, rc, cc = illegal_property;
  ExportListStruct *el;
  ExportPropertyListStruct *pls=NULL;
  char *eqm, *prp, *meta;
  BYTE nfmt = BFMT(CF_NULL), *fmt = &nfmt;
  UINT32 *siz;

  if (con == NULL) ccerr(code_failure);

  eqm = con->EqmName; prp = con->EqmProperty;
  fmt = &con->EqmFormatOut; siz = &con->EqmSizeOut;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (lookupRedirectionNameStub != NULL) 
  {
    if ((cc=lookupRedirectionNameStub(el->EqmName,con->EqmProperty,con->EqmDeviceName)) < 0) cc = software_error;
    if (cc == server_redirection) memcpy(status,el->EqmLastErrorString,STATUS_SIZE);
    if (cc != 0) ccerr(cc); /* likely : cc = server_redirection, cc = 0 falls thru */
  }
chkprp:
  if ((rc=lookupPropertyName(el,prp,con->EqmDeviceName,&pls)) == server_redirection) 
  {
    if (el->EqmDevLst != NULL && 
        (idx=GetDeviceNumberEx(eqm,con->EqmDeviceName,NULL)) >= 0 && 
        idx < el->EqmNumDevices &&
        el->EqmDevLst[idx].rdr != NULL)
    {
      memcpy(status,el->EqmDevLst[idx].rdr,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE);
      ccerr(rc);
    }
    if (isPropertyRedirected(el,pls,con->EqmDeviceName)) 
    {
      memcpy(status,pls->prd,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE);
      ccerr(rc);
    }
  }
  if (rc == has_bitfield_tag || rc == has_structure_tag) 
  {
    if (pls != NULL) strncpy(status,pls->prpFormatTag.name,TAG_NAME_SIZE);
    ccerr(rc);
  }
  if (rc == 0 && pls != NULL)
  { /* usually this has been assigned (even with query functions) */
    if (isWildcardContract(con->EqmDeviceName) && isNumberFormat(pls->prpFormat))
    { /* an allowable wild-card contract */
      *fmt = BFMT(CF_USTRING);
      *siz = el->EqmNumDevices - getNumReservedDevices(el);
      ccerr(0);
    }
    *fmt = BFMT(pls->prpFormat);
    *siz = pls->prpSize;
    strncpy(con->strTagOut,pls->prpFormatTag.name,TAG_NAME_SIZE);
    if (pls->prpArrayType & AT_CHANNEL) *siz = 1;
    if (!isSimpleFormat(pls->prpFormat))
    { /* suggestion is not a simple format: look for an overload */
      while (pls->next != NULL)
      { /* run thru the remaining list looking for overloads */
        pls = pls->next;
        if (strnicmp(pls->prpName,prp,PROPERTY_NAME_SIZE)) continue;
        if (isSimpleFormat(pls->prpFormat))
        { /* found one */
          *fmt = BFMT(pls->prpFormat);
          break;
        }
      }

    }
    ccerr(0);
  }
  if ((idx=GetStockPropertyId(prp)) >= 0)
  {
    *fmt = BFMT(gStockProperty[idx].prpFormat);
    *siz = gStockProperty[idx].prpSize;
    if (isTaggableFormat(gStockProperty[idx].prpFormat)) 
      strncpy(con->strTagOut,gStockProperty[idx].prpStructTag,TAG_NAME_SIZE);
    ccerr(0);
  }
  if (IsMetaProperty(prp,&prp,&meta,FALSE))
  {
    short sfmt;
    int isiz;
    if (strchr(meta,'@') != NULL) goto chkprp;
    if (getMetaFormatAndSize(meta,&sfmt,&isiz) != ambiguous)
    { /* good guess here ... */
      *fmt = BFMT(sfmt);
      *siz = (UINT32)isiz;
      ccerr(0);
    }
    else
    { /* punt ... */
      *fmt = BFMT(CF_INT32);
      *siz = 1;
    }
    ccerr(0);
  }
err:
  if (*fmt == BFMT(CF_STRUCT) && cc == 0) 
  { /* there is currently no way to pass the structure tag back to the caller via CF_DEFAULT */
    return invalid_structure_tag;
  }
  return cc;
}
int sendMessageToCaller(int tineProtocol,ClnHdr *cln,SubReqHdr *s,int cc,BYTE *data,int len)
{
  static BYTE dbuf[REQHDR_SIZE + SUBRSPHDR_SIZE + STATUS_SIZE] = {0}; /* worst-case buffer size */
  ClnLst clnlst = {0};
  RspHdr rsphdr = {0};
  ContractListStruct cls;
  UINT16 msgsize,stssize,psize,hsize,subid;
  int sysst=gSystemDataStamp, datst=0;
  double dts;
  if (cln == NULL || s == NULL || data == NULL) return code_failure;
  memset(dbuf,0,REQHDR_SIZE+SUBRSPHDR_SIZE+STATUS_SIZE); /* removes a valgrind warning */
  hsize = (UINT16)RETURNHDR_SIZE(tineProtocol);
  switch (cc)
  {
    case get_subscription_id:
      if (tineProtocol < 5) ccerr(illegal_protocol);
      break;
    case property_is_mca_elevated:
    case property_is_mca:
      memcpy(&datst,data,4); /* index */
      memcpy(&sysst,data+4,4); /* length */
      memcpy(data,MCADEV_TAG,8);
      len = DEVICE_NAME_SIZE+8;
      break;
    case invalid_interval:
      sysst = MinPollingRate;
      len = 0;
      break;
    case invalid_datarequest:
      memcpy(&sysst,data,4); /* required length */
      memcpy(&datst,data+4,4); /* required format */      
      len = 0;
      break;
    case access_denied:
      data = strlen(aclStsErr) > 0 ? (BYTE *)aclStsErr : (BYTE *)cc2str(cc);
      len = STSSIZE(tineProtocol);
      break;
    case call_redirection:
    case server_redirection:
      if (assertRedirectionValid("sendMessageToCaller",s->contract.EqmName,(char *)data))
      {
        len = STSSIZE(tineProtocol);
        if (tineProtocol < 6)
        { /* try this and hope for the best */
          memcpy(data,&data[PRP_REDIR_SRV_OFFSET],16);
          memcpy(&data[16],&data[PRP_REDIR_PRP_OFFSET],16);
        }
        break;
      }
      else
      {
        data = NULL;
        cc = configuration_error;
      }
    default:
      if (cc > 0 && cc < numErr) 
      {
        if (cc == illegal_protocol && 
            tineProtocol > DEFAULT_TRANSPORT_PROTOCOL_LEVEL)
        { /* re-assign so that the message can go back */
          tineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
        }
        if (data == NULL) data = (BYTE *)cc2str(cc);
      }
      len = STSSIZE(tineProtocol);
      break;
  }
  if (tineDebug) 
  {
    switch (cc)
    {
      case illegal_property:
      case illegal_equipment_name:
      case non_existent_elem:
      case call_redirection:
      case server_redirection:
      case access_denied:
        dbglog("MSG: (%.6s)/%.64s[%.64s] -> %.32s",
          s->contract.EqmName,s->contract.EqmDeviceName,s->contract.EqmProperty,cc2str(cc));
        break;
      case get_subscription_id:
        memcpy(&subid,data,2);
        dbglog("MSG: (id = %d) -> %.32s",SPEEK(subid),cc2str(cc));
        break;
      case property_is_mca_elevated:
      case property_is_mca:
        dbglog("MSG: index %d of %d -> %.32s",datst,sysst,cc2str(cc));
        break;
      case invalid_interval:
        dbglog("MSG: poll interval set to %d -> %.32s",sysst,cc2str(cc));
        break;
      case invalid_datarequest:
        dbglog("MSG: output data length and type set to %d and %d -> %.32s",sysst,datst,cc2str(cc));
        break;
      default: 
        dbglog("MSG: -> %.32s",cc2str(cc));
        break;
    }
  }
  dts = PutDataTimeStampU(gDataTimeStampOffset, thisCycleTime.tv_sec, (int)(thisCycleTime.tv_usec));
  clnlst.status.starttime = s->starttime;
  clnlst.status.statusCode = cc;
  clnlst.status.id = s->id;
  clnlst.status.counter = 0;
  clnlst.status.blknum = 1;
  clnlst.status.numblks = 1;
  clnlst.client = cln;
  clnlst.client->tineProtocol = tineProtocol;
  cls.clnLst = &clnlst;
  cls.contract = s->contract;
  cls.returnCode = not_posted;
  cls.datastamp = datst;
  cls.sysstamp = sysst;
  cls.dtimestamp = (UINT32)dts;
  cls.dtimestampUSEC = (int)((dts - (SINT32)dts) * 1000000);
  stssize = STSSIZE(tineProtocol);
  psize = SUBRSP_STEPSIZE(tineProtocol);
  msgsize = psize + stssize;
  if ((cc=prepOutgoingSubResponseHdr(&clnlst,&cls, msgsize, &dbuf[hsize])) != 0)
  { /* seriously corrupt */
    return cc;
  }
  rsphdr.totalSizeInBytes = (UINT32)(msgsize + hsize);
  rsphdr.numberSegments = 1;
  rsphdr.tineProtocol = tineProtocol;
  prepOutgoingResponseHdr(cln, &rsphdr, dbuf);
  if (data != NULL && len > 0)
  {
    memcpy(&dbuf[hsize+psize],data,len);
  }
  sendToPeer(cln,dbuf,rsphdr.totalSizeInBytes,SRV_RSP_,FALSE);
err:
  return cc;
}
int getNextFreeContractID(void)
{
  static UINT16 cid = 0;
  int i;
  for (i=0,cid++; i<ncontracts; i++) 
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    if (ContractList[i]->contractID == cid) cid++;
  }
  return cid;
}
/**
 * \internal process incoming header stream
 * \param hdr (out) is the stream converted to the default protocol level
 * \param stream is the incoming byte stream
 * \return protocol level corresponding to the stream (not necessarily the default protocol level)
 */
int prepIncomingRequestHdr(ReqHdr *hdr, BYTE *stream)
{
  int p = (int)SPEEK(*(UINT16 *)&stream[SUB_PROT_LOCATION]); /* this is always the spot ! */
  if (p < MINIMUM_REQHDR_PROTOCOL)
  { /* map to modern request header */
    PktHdr h, *ph = &h;
#   ifdef INTEL
    memcpy(ph, stream, sizeof(PktHdr));
#   else
    _structSwap(PktHdrStructStruct, stream, 0, NULL, (BYTE *)ph, RTOL);
#   endif
    memset(hdr, 0, sizeof(ReqHdr));
    hdr->tineProtocol = h.tineProtocol;
    hdr->totalSizeInBytes = (UINT32)h.totalSize;
    hdr->numberSegments = h.number;
    hdr->pid = 0;
    hdr->encoding = CE_ASCII;
    hdr->endianness = BO_LITTLE_ENDIAN;
    hdr->headerSizeInBytes = (UINT32)sizeof(PktHdr);
    hdr->tineVersion = (h.tineProtocol<<16) | h.number;
    strncpy(hdr->userType, "LEGACY", USERTYPE_SIZE);
    strncpy(hdr->userName, h.userName, USERNAME_SIZE);
  }
  else
  {
#   ifdef INTEL
    memcpy(hdr, stream, sizeof(ReqHdr));
#   else
    _structSwap(ReqHdrStructStruct,stream, 0, NULL, (BYTE *)hdr, RTOL);
#   endif
    if (hdr->headerSizeInBytes < REQHDR_SIZE) return -1;
  }
  return p;
}

#define contract5to6(c6,c5) {\
  strncpy(c6.EqmProperty,c5.EqmProperty,PROPERTY_NAME_SHORTSIZE);\
  strncpy(c6.EqmDeviceName,c5.EqmDeviceName,DEVICE_NAME_SHORTSIZE);\
  strncpy(c6.EqmName,c5.EqmName,EQM_NAME_SIZE);\
  c6.EqmSizeIn = c5.EqmSizeIn;\
  c6.EqmSizeOut = c5.EqmSizeOut;\
  c6.hEqmName = c5.hEqmName;\
  c6.EqmAccess = c5.EqmAccess;\
  c6.EqmFormatIn = c5.EqmFormatIn;\
  c6.EqmFormatOut = c5.EqmFormatOut;\
  strncpy(c6.strTagIn,c5.strTagIn,TAG_NAME_SHORTSIZE);\
  strncpy(c6.strTagOut,c5.strTagOut,TAG_NAME_SHORTSIZE);\
}
int prepIncomingSubRequestHdr(int tineProtocol, SubReqHdr *hdr, BYTE *stream, int *swallowError)
{
  switch (tineProtocol)
  {
    case 5:
    { /* Release 3.xx era */
      SubInfoPkt5 sub5, *pSub5 = &sub5;
#     ifdef INTEL
      if (ForceStructAlignment)
      { /* NIOS for instance */
        memcpy(pSub5, stream, sizeof(SubInfoPkt5));
      }
      else
      { /* linux, windows, dos, etc. */
        pSub5 = (SubInfoPkt5 *)stream;
      }
#     else
      memset(pSub5, 0, sizeof(SubInfoPkt5));
      _structSwap(SubInfoPkt5Struct, stream, 0, NULL, (BYTE *)pSub5, RTOL);
#     endif
      memset(hdr, 0, sizeof(SubReqHdr));
      hdr->msgsize = pSub5->msgsize;
      hdr->mtu = pSub5->mtu;
      hdr->numblks = pSub5->numblks;
      hdr->blknum = pSub5->blknum;
      hdr->id = pSub5->id;
      hdr->mode = pSub5->mode;
      hdr->blkid = pSub5->blkid;
      hdr->pollingInterval = pSub5->pollingInterval;
      hdr->starttime = pSub5->starttime;
      contract5to6(hdr->contract, pSub5->contract);
      if (hdr->contract.EqmDeviceName[DEVICE_NAME_SHORTSIZE - 1] == '&')
      { /* legacy kluge to support channel access names as device names */
        hdr->contract.EqmDeviceName[DEVICE_NAME_SHORTSIZE - 1] = 0;
        strncat(hdr->contract.EqmDeviceName, &hdr->contract.EqmProperty[DEVICE_NAME_SHORTSIZE + 1], DEVICE_NAME_SHORTSIZE - 1);
        memset(&hdr->contract.EqmProperty[DEVICE_NAME_SHORTSIZE], 0, DEVICE_NAME_SHORTSIZE);
      }
      break;
    }
    case 6:
    { /* Release 4.xx era */
      SubInfoPkt sub, *pSub = &sub;
#     ifdef INTEL
      if (ForceStructAlignment)
      { /* NIOS for instance */
        memcpy(pSub, stream, sizeof(SubInfoPkt));
      }
      else
      { /* linux, windows, dos, etc. */
        pSub = (SubInfoPkt *)stream;
      }
#     else
      memset(pSub, 0, sizeof(SubInfoPkt));
      _structSwap(SubInfoPktStruct, stream, 0, NULL, (BYTE *)pSub, RTOL);
#     endif
      memset(hdr, 0, sizeof(SubReqHdr));
      hdr->msgsize = pSub->msgsize;
      hdr->extsize = pSub->extsize;
      hdr->mtu = pSub->mtu;
      hdr->blkid = pSub->blkid;
      hdr->numblks = pSub->numblks;
      hdr->blknum = pSub->blknum;
      hdr->id = pSub->id;
      hdr->mode = pSub->mode;
      hdr->pollingInterval = pSub->pollingInterval;
      hdr->starttime = pSub->starttime;
      memcpy(&hdr->contract, &pSub->contract, sizeof(CONTRACT));
      break;
    }
    case 7:
#     ifdef INTEL
      memcpy(hdr, stream, sizeof(SubReqHdr));
#     else
      memset(hdr, 0, sizeof(SubReqHdr));
      _structSwap(SubReqHdrStruct, stream, 0, NULL, (BYTE *)hdr, RTOL);
#     endif
    break;
  default:
    if (tineProtocol < 5 || tineProtocol > 8) *swallowError = TRUE;
    return illegal_protocol;
  }
  if (hdr->pollingInterval > MAXALLOWED_POLLINTVL) hdr->pollingInterval = MAXALLOWED_POLLINTVL;
  strntrm(hdr->contract.EqmProperty, 64);
  strntrm(hdr->contract.EqmDeviceName, 64);
  return 0;
}
int contractsEqual(CONTRACT *c1, CONTRACT *c2)
{
  if (c1 == NULL || c2 == NULL) return FALSE;
  if (memcmp(c1,c2,PROPERTY_NAME_SIZE+DEVICE_NAME_SIZE+EQM_NAME_SIZE+2*sizeof(UINT32))) return FALSE;
  if ((c1->EqmAccess & 0x03) != (c2->EqmAccess & 0x03)) return FALSE;
  if (c1->EqmFormatIn != c2->EqmFormatIn) return FALSE;
  if (c1->EqmFormatOut != c2->EqmFormatOut) return FALSE;
  if (strncmp(c1->strTagIn,c2->strTagIn,TAG_NAME_SIZE)) return FALSE;
  if (strncmp(c1->strTagOut,c2->strTagOut,TAG_NAME_SIZE)) return FALSE;
  return TRUE;
}
int cmpArrayInput(ContractListStruct *cl,BYTE *dbuf,size_t bufsize)
{
# ifndef SMALL_TINE_LIB
  int i, hsiz, tsiz, xoff, xsiz, n;
  char *sb = (char *)cl->dataIn;
# endif

  switch (LFMT(cl->contract.EqmFormatIn))
  {
#   ifndef SMALL_TINE_LIB
    case CF_ASPECTRUM:
      /* compares raw incoming input data with that stored in a contract */
      n = atoi(cl->contract.strTagIn);
      hsiz = SPECTRUM_HDR_SIZE;
      tsiz = sizeof(ASPECTRUM);
      for (i=0; i<n; i++)
      {
        if (memcmp(&sb[i*tsiz],&dbuf[i*(hsiz+8)],hsiz)) return TRUE;
        memcpy(&xoff,&dbuf[i*(hsiz+8)+hsiz],4); /* offset */
        memcpy(&xsiz,&dbuf[i*(hsiz+8)+hsiz+4],4); /* length */
        ISWAP(xoff); ISWAP(xsiz);
        if (((ASPECTRUM *)&sb[i*tsiz])->spectBufferSize != (UINT32)xsiz/sizeof(float)) return TRUE;
        if (memcmp(((ASPECTRUM *)&sb[i*tsiz])->spectBuffer,(float *)&sb[xoff],xsiz)) return TRUE;
      }
      break;  
    case CF_AIMAGE:
      /* compares raw incoming input data with that stored in a contract */
      n = atoi(cl->contract.strTagIn);
      hsiz = IMAGE_HDR_SIZE;
      tsiz = sizeof(DIMAGE);
      for (i=0; i<n; i++)
      {
        if (memcmp(&sb[i*tsiz],&dbuf[i*(hsiz+8)],hsiz)) return TRUE;
        memcpy(&xoff,&dbuf[i*(hsiz+8)+hsiz],4); /* offset */
        memcpy(&xsiz,&dbuf[i*(hsiz+8)+hsiz+4],4); /* length */
        if (((DIMAGE *)&sb[i*tsiz])->frameBufferSize != (UINT32)xsiz) return TRUE;
        if (memcmp(((DIMAGE *)&sb[i*tsiz])->frameBuffer,&sb[xoff],xsiz)) return TRUE;
      }
      break;  
#   endif
    default:
      return memcmp(cl->dataIn,dbuf,bufsize);
      break;  
  }
  return 0;
}
int prepArrayInput(ContractListStruct *cl,BYTE *dbuf,size_t bufsize)
{ /* map the input data as the dispatch routine will want to see it */
# ifndef SMALL_TINE_LIB
  int i, hsiz, tsiz, xoff, deloff, xsiz, n;
  UINT32 p = 0, blen = cl->contract.EqmSizeIn;
  char **sa, *sb = (char *)cl->dataIn;
# endif

  switch (LFMT(cl->contract.EqmFormatIn))
  {
#   ifndef SMALL_TINE_LIB
    case CF_KEYVALUE:
    case CF_STRING:
      memcpy(cl->dataIn,dbuf,(size_t)bufsize);
      n = atoi(cl->contract.strTagIn);
      if (cl->dataPtr == NULL && (cl->dataPtr=(BYTE **)SystemCalloc(n,sizeof(char *))) == NULL)
      { /* allocate string pointer array */
        return out_of_server_memory;
      }
      blen = cl->contract.EqmSizeIn;
      sa = (char **)cl->dataPtr;
      for (i=0; i<n && p<blen; i++)
      {
        sa[i] = &sb[p];
        p += (int)strlen(sa[i]) + 1;
      }
      break;
    case CF_ASPECTRUM:
      n = atoi(cl->contract.strTagIn);
      hsiz = SPECTRUM_HDR_SIZE;
      tsiz = sizeof(ASPECTRUM);
      deloff = (tsiz-hsiz-8); /* remap the extended byte offsets */
      for (i=0; i<n; i++)
      {
        memcpy(&sb[i*tsiz],&dbuf[i*(hsiz+8)],hsiz);
        memcpy(&xoff,&dbuf[i*(hsiz+8)+hsiz],4); /* offset */
        memcpy(&xsiz,&dbuf[i*(hsiz+8)+hsiz+4],4); /* length */
        memcpy(&sb[xoff+n*deloff],&dbuf[xoff],xsiz);
        xsiz /= sizeof(float);
        xoff += n*deloff;
        SwapBytes(&sb[xoff],xsiz,CF_FLOAT,RTOL);
        ((ASPECTRUM *)&sb[i*tsiz])->spectBufferSize = xsiz;
        ((ASPECTRUM *)&sb[i*tsiz])->spectBuffer = (float *)&sb[xoff];
      }
      break;  
    case CF_AIMAGE:
      n = atoi(cl->contract.strTagIn);
      hsiz = IMAGE_HDR_SIZE;
      tsiz = sizeof(DIMAGE);
      deloff = (tsiz-hsiz-8); /* remap the extended byte offsets */
      for (i=0; i<n; i++)
      {
        memcpy(&sb[i*tsiz],&dbuf[i*(hsiz+8)],hsiz);
        memcpy(&xoff,&dbuf[i*(hsiz+8)+hsiz],4); /* offset */
        memcpy(&xsiz,&dbuf[i*(hsiz+8)+hsiz+4],4); /* length */
        ((DIMAGE *)&sb[i*tsiz])->frameBufferSize = xsiz;
        memcpy(&sb[xoff+n*deloff],&dbuf[xoff],xsiz);
        xoff += n*deloff;
        ((DIMAGE *)&sb[i*tsiz])->frameBuffer = (BYTE *)&sb[xoff];
      }
      break;  
#   endif
    default:
      /* special data type will get massaged prior to callEqm */
      memcpy(cl->dataIn,dbuf,(size_t)bufsize);
      break;  
  }
  return 0;
}
int getMcaPrpCc(CONTRACT *con)
{
  ExportPropertyListStruct *pls;
  char *prp = con->EqmProperty;
  ExportListStruct *el = getExportListItem(con->EqmName);
  int idx = (int)(ElfHash((unsigned char *)prp) % PROPERTY_HASH_SIZE);
  short fmt = (short)LFMT(con->EqmFormatOut);
  short efmt = getElevatedMcaFormat(fmt);
  if (fmt == efmt) return property_is_mca;
  for (pls=el->EqmPrpLst[idx]; pls != NULL; pls=pls->next)
  {
    if (strnicmp(pls->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    if (pls->prpFormat == efmt) return property_is_mca_elevated;  
  }
  return property_is_mca;
}
int gSubRenewalLength = CTR_RENEWAL;
int gSubRenewalMultiplier = 1;
static int getThisSubscriptionRenewalMultiplier(ContractListStruct *cl,ClnLst *c)
{ /* internal use only: no parameter checking necessary */
  int rm = 0;
  if (GetPropertySubscriptionRenewalLength(cl->contract.EqmName,cl->contract.EqmProperty,&rm) != 0) rm = 0;
  if (rm == 0) rm = gSubRenewalMultiplier; else rm /= CTR_RENEWAL;
  rm = getEffectiveRenewalMultiplier(rm,c->status.pollingInterval);
  if (rm < 1) rm = 1;
  if (c->client->tineProtocol < 6 && rm > 4) rm = 4;  /* legacy client */
  return rm;
}
UINT32 gSingleLinkCount = 0;
#define replaceContractAliasNames(cn) { if (nAliasTableEntries > 0) {\
  if ((i=getAliasTableIndex((cn)->EqmProperty)) >= 0) strncpy((cn)->EqmProperty,gAliasTable[i].name,PROPERTY_NAME_SIZE);\
  if ((i=getAliasTableIndex((cn)->EqmDeviceName)) >= 0) strncpy((cn)->EqmDeviceName,gAliasTable[i].name,DEVICE_NAME_SIZE);\
  }\
}
#define ccout(o) { cc = o; goto out; }
int chkMcaHandshake(ReqHdr *h)
{
  if (enforceMcaAcquisition) return TRUE;
  if (h->tineProtocol > 5) return TRUE; /* anybody post R3 is allowed: forget about pre 4.0.11 clients */
  return FALSE;
}
int concmp(CONTRACT *c1,CONTRACT *c2)
{
# define csiz1 (PROPERTY_NAME_SIZE+DEVICE_NAME_SIZE+EQM_NAME_SIZE+8)
# define csiz2 (2+2*TAG_NAME_SIZE)
  if (memcmp(c1,c2,csiz1)) return -1;
  if (memcmp(&c1->EqmFormatIn,&c2->EqmFormatIn,csiz2)) return -1;
  if ((c1->EqmAccess & 0x03) != (c2->EqmAccess & 0x03)) return -1;
  return 0;
}
void dumpContractToScreen(ReqHdr *rh, SubReqHdr *hdr)
{
  dbglog("contract header from %.16s (protocol %d)\n\n", rh->userName, rh->tineProtocol);
  dbglog("msgsize: %d\n", hdr->msgsize);
  dbglog("extsize: %d\n", hdr->extsize);
  dbglog("mtu: %d\n", hdr->mtu);
  dbglog("blkid: %d\n", hdr->blkid);
  dbglog("numblks: %d\n", hdr->numblks);
  dbglog("blknum: %d\n", hdr->blknum);
  dbglog("id: %d\n", hdr->id);
  dbglog("mode: %d\n", hdr->mode);
  dbglog("pollingInterval: %d\n", hdr->pollingInterval);
  dbglog("starttime: %d\n", hdr->starttime);
  dbglog("contract :\n");
  dbglog("EqmProperty: %.64s\n", hdr->contract.EqmProperty);
  dbglog("EqmDeviceName: %.64s\n", hdr->contract.EqmDeviceName);
  dbglog("EqmName: %.6s\n", hdr->contract.EqmName);
  dbglog("EqmSizeIn: %d\n", hdr->contract.EqmSizeIn);
  dbglog("EqmSizeOut: %d\n", hdr->contract.EqmSizeOut);
  dbglog("EqmAccess: %d\n", hdr->contract.EqmAccess);
  dbglog("EqmFormatIn: %d\n", hdr->contract.EqmFormatIn);
  dbglog("EqmFormatOut: %d\n", hdr->contract.EqmFormatOut);
  dbglog("strTagIn: %.16s\n", hdr->contract.strTagIn);
  dbglog("strTagOut: %.16s\n", hdr->contract.strTagOut);
  return;
}
#define PutInCmdLog(c) (\
  putCommandsInCmdlog &&\
  ((c)->contract.EqmAccess & CA_WRITE) &&\
  strcmp((c)->contract.EqmProperty,"ACCESSLOCK") &&\
  strcmp((c)->contract.EqmProperty,"MESSAGE"))

int getConsumerRequest(ClnHdr *caller,BYTE *buf,int packetsize)
{
  BYTE *dbuf;
  ReqHdr rh;
  ClnLst tmpC,*c = &tmpC;
  ClnHdr *nwcln, *cln=caller, *tgtcln=NULL, newcln;
  ContractListStruct *cl=NULL, *tgtcon=NULL;
  int i, n, hprot, clid = -1, slot = -1, idx, dlen, nreqs=0, msecs;
  short subid,fmtinsize,fmtoutsize,cc=0,fmt,acc,bmode,renewalMultiplier;
  int isTrueMCast = 0, swallowOnErr = FALSE, resetClnCounters = FALSE,
    isNewContract = FALSE, isNewClient = FALSE, isAdjustable = FALSE,
    canRemoveContractOnError = TRUE, sendSingleDelivery = FALSE,
    chkResendData = FALSE, resendCurrentData = FALSE, isNewStarttime = FALSE,
    isXPrp,chkAccess,logit;
  UINT32 totalsize,datasize,bytesread;
  UINT32 dinsize,doutsize,dinallocsize=0,defsize,dataPtr,mxsize,datamtu;
  SubReqHdr subhdr;
  ExportPropertyListStruct *prp=NULL;
  ExportListStruct *el=NULL;
  int  extsiz = 0, ssiz, hdrsiz;
  char extstr[DEVICE_NAME_MAXSIZE],stsstr[STATUS_SIZE], *dnam;

  if (thisCycleTime.tv_sec < rejectEarlierThan) return not_ready;
  stsstr[0] = 0;

  hprot = prepIncomingRequestHdr(&rh, buf);
  if (hprot < 0) ccerr(illegal_protocol);
  bytesread = PAYLOADHDR_SIZE(hprot);
  totalsize = rh.totalSizeInBytes;
  caller->tineProtocol = hprot;

  if (totalsize <= 0) ccout(out_of_range);
  if (tineDebug) sprintf(&dbgbuf[strlen(dbgbuf)],"%.16s -> ",rh.userName);
  strncpy(cln->usr,rh.userName,USERNAME_SIZE);  /* always reset   */
  strncpy(cln->typ,rh.userType,USERTYPE_SIZE);
  cln->pid = rh.pid;
  /* map incoming tineProtocol to the most modern : at the moment 6 */
  if ((cc=prepIncomingSubRequestHdr(hprot, &subhdr, &buf[bytesread], &swallowOnErr)) != 0) ccout(cc);
  if (subhdr.extsize > 1024 || subhdr.numblks > subhdr.msgsize || subhdr.msgsize == 0)
  {
    if (tineDebug) dbglog("incoming subscription header appears to be corrupt!");
    swallowOnErr = TRUE;
    ccout(invalid_transport_size);
  }
  if (subhdr.extsize > 0 && subhdr.blknum == 1)
  {  /* extended (device name) string set (comes with first block) */
     extsiz = MIN(subhdr.extsize,DEVICE_NAME_MAXSIZE);
     memcpy(extstr,&buf[bytesread+SUBREQ_STEPSIZE(hprot)],extsiz);
     if (extsiz % 2) 
     { /* make sure even boundaries and properly terminated */
       extstr[extsiz] = 0;
       extsiz++;
     }
  }
  replaceContractAliasNames(&subhdr.contract);
  if (isServiceRequest(&subhdr.contract)) swallowOnErr = TRUE;
  while (bytesread < totalsize && bytesread < (UINT32)packetsize) 
  { /* unpack request(s) */
    nreqs++;
    memset(stsstr,0,STATUS_SIZE); clid = -1;
    isTrueMCast = FALSE;
    cln = caller; /* reset this in case of packed NETWORK request ! */
    bmode = BASEMODE(subhdr.mode);
    if (tineDebug && subhdr.numblks < 4)
    { /* output to screen can influence io buffering */
      sprintf(&dbgbuf[strlen(dbgbuf)],"(%.6s)/%.64s[%.64s] %d msec (%d of %d)[%s]",
        subhdr.contract.EqmName, subhdr.contract.EqmDeviceName,
        subhdr.contract.EqmProperty,(int)subhdr.pollingInterval,
        subhdr.blknum, subhdr.numblks,modetoa[bmode]);
      dbglog(dbgbuf);
      sprintf(dbgbuf,"\t\t ->");
    }
    if (gDeniedUsersList.nUsrs > 0 && isDeniedUser(rh.userName)) ccerr(access_denied);
    if (subhdr.contract.EqmSizeOut <= 0) subhdr.contract.EqmFormatOut = BFMT(CF_NULL); /* guarantee this! */
    if ((fmtoutsize=fmtsizeof(subhdr.contract.EqmFormatOut)) > 0)
    {
      fmt = LFMT(subhdr.contract.EqmFormatOut);
      if (fmt == CF_MDA) fmtoutsize = getCarriedFormatSize(subhdr.contract.strTagOut);
      if ((doutsize=subhdr.contract.EqmSizeOut*fmtoutsize) <= 0) ccerr(illegal_data_size);
      hdrsiz = getFormatHeaderSize(fmt); 
      n = 1; /* number outgoing adjustable items (if relevant) */
#     ifndef SMALL_TINE_LIB
      if (isAdjustableLengthFormat(fmt))
      { /* if an adjustable type, get the requested arraylength from the tag */
        n = atoi(subhdr.contract.strTagOut);
        /* re-assign the sizeOut according to the house rules */
        subhdr.contract.EqmSizeOut = n * getFormatStandardCapacity(fmt);
        doutsize = subhdr.contract.EqmSizeOut*fmtoutsize;
        hdrsiz = (fmt == CF_STRING || fmt == CF_KEYVALUE) ?  0 : getFormatTypeSize(fmt);
      }
#     endif
      doutsize += n * hdrsiz;
      mxsize = (UINT32)MaxSystemTransportSize;
#     ifdef MULTITHREADED
chkMaxSize:
#     endif
      if (isJumboDelivery(cln)) 
      {
        if (!isJumboDeliveryAllowed(cln,doutsize)) ccerr(buffer_too_small);
        mxsize -= SUBRSPHDR_SIZE+RSPHDR_SIZE+STATUS_SIZE;
      }
      if (doutsize > mxsize) 
      { /* incoming request size exceeds the size supported: make adjustments */
#       ifdef MULTITHREADED
        if (!IsStockProperty(subhdr.contract.EqmProperty) && 
            !IsMetaProperty(subhdr.contract.EqmProperty,NULL,NULL,FALSE))
        { /* stock properties always need to follow MaxSystemTransportSize (i.e. tmpWorkAreaSize) */
          if (gTransportCeiling > 0 && gTransportCeiling != (int)mxsize) 
          { /* multi-threaded servers allocate memory/contract in lieu of srvWorkArea */
            mxsize = (UINT32)gTransportCeiling;
            goto chkMaxSize;
          }
        }
#       endif
        subhdr.contract.EqmSizeOut = (mxsize - (n * hdrsiz)) / fmtoutsize;
        if (fmt == CF_STRUCT && (ssiz=GetStructSize(subhdr.contract.strTagOut)) > 0) 
        { /* if tagged struct, make sure of the formal boundaries when truncating */
          subhdr.contract.EqmSizeOut -= (subhdr.contract.EqmSizeOut % ssiz);
        }
        else if (fmt == CF_HISTORY && (ssiz=getCarriedFormatSize(subhdr.contract.strTagOut)) > 0)
        { /* if CF_HISTORY, make sure of the formal boundaries when truncating */
          ssiz += HISTORY_HDR_SIZE;
          subhdr.contract.EqmSizeOut -= (subhdr.contract.EqmSizeOut % ssiz);
        }
        if (subhdr.contract.EqmSizeOut <= 0) ccerr(invalid_transport_size);
        doutsize = subhdr.contract.EqmSizeOut*fmtoutsize;
        doutsize += n * hdrsiz;
      }
    }
    else 
    { /* either CF_NULL or not a valid format */
      if (subhdr.contract.EqmFormatOut != BFMT(CF_NULL)) ccerr(illegal_format);
      doutsize = 0;
      subhdr.contract.EqmFormatOut = BFMT(CF_NULL);
      subhdr.contract.EqmSizeOut = 0;
    }
    if (subhdr.contract.EqmSizeIn <= 0) subhdr.contract.EqmFormatIn = BFMT(CF_NULL); /* guarantee this! */
    if ((fmtinsize=fmtsizeof(subhdr.contract.EqmFormatIn)) != 0) 
    { /* data input */
      if (subhdr.blknum > subhdr.numblks) ccerr(link_error);
      fmt = LFMT(subhdr.contract.EqmFormatIn);
      if (fmt == CF_MDA) 
      { /* special case : has a carried datatype */
        fmtinsize = getCarriedFormatSize(subhdr.contract.strTagIn);
        if (fmtinsize <= 0) ccerr(illegal_data_size);
      }
      isAdjustable = isAdjustableLengthFormat(fmt);
      hdrsiz = getFormatHeaderSize(fmt); 
      n = 1; /* number of incoming adjustable elements (if relevant) */
      dinsize = subhdr.contract.EqmSizeIn*fmtinsize;
      if (isAdjustable)
      { /* the 'effective header' needs to be adjusted as well */
        if ((n=atoi(subhdr.contract.strTagIn)) <= 0) ccerr(illegal_format);
        hdrsiz = (fmt == CF_STRING || fmt == CF_KEYVALUE) ? 0 : getFormatTypeSize(fmt);
      }
      dinsize += n * hdrsiz;
      dbuf = &buf[bytesread+SUBREQ_STEPSIZE(hprot)+subhdr.extsize];
      dataPtr = ((UINT32)(subhdr.blknum-1))*subhdr.mtu;
      datasize = (UINT16)MIN(dinsize-dataPtr,(UINT32)subhdr.mtu);
      dinallocsize = ALLOCSIZE(dinsize);
      if (dinallocsize > (UINT32)MaxSystemTransportSize) ccerr(invalid_transport_size);
      if (dinsize > subhdr.mtu)
      { /* incoming long datagram */
        if (subhdr.numblks == 1) ccerr(data_size_mismatch); /* sanity check */
        for (i=0; i<MAXLONGGRAMS; i++)
          if (sckadr_eq(&conprg[i].sckadr,&cln->adr.ip) &&
              conprg[i].subId == subhdr.id &&
              conprg[i].SubsStarttime  == subhdr.starttime) break;
        if (i == MAXLONGGRAMS) for (i=0; conprg[i].busy && i<MAXLONGGRAMS; i++);
        if (i == MAXLONGGRAMS) 
        {
          if (tineDebug) dbglog("incoming data assembly table has reached capacity (%d)",i);
          ccerr(resources_exhausted);
        }
        if (++(conprg[i].blksin) <= subhdr.numblks)
        { /* track the blocks in */
          if (conprg[i].blksin == 1) 
          { /* starting out */
            conprg[i].blkid = subhdr.blkid;
            conprg[i].SubsStarttime = subhdr.starttime;
          }
          else if (conprg[i].blkid != subhdr.blkid || conprg[i].SubsStarttime != subhdr.starttime)
          { /* oops ! */
            if (tineDebug > 1)
              dbglog("re-sync long datagram %d: id %d to %d",i,conprg[i].blkid,subhdr.blkid);
            conprg[i].blksin = 1;
            conprg[i].blkid = subhdr.blkid;
            conprg[i].blktime = (UINT32)time(NULL);
            conprg[i].SubsStarttime = subhdr.starttime;
          }
          if (!conprg[i].busy)
          { /* starting out */
            if (tineDebug > 1) 
              dbglog("new long datagram %d: (%d) sub id %d %d bytes",i,subhdr.blkid,subhdr.id,(int)dinsize);
            conprg[i].busy = 1;
            conprg[i].sckadr = cln->adr.ip;
            conprg[i].subId = subhdr.id;
            conprg[i].blktime = (UINT32)time(NULL);
            conprg[i].SubsStarttime = subhdr.starttime;
            if ((conprg[i].data = (BYTE *)SystemCalloc(1, (size_t)dinallocsize)) == NULL)
            {
              conprg[i].busy = 0;
              ccerr(out_of_server_memory);
            }
          }
          if (tineDebug > 2) dbglog("copy long datagram %d: blk %d to location %d",i,subhdr.blknum,dataPtr);
          memcpy(&conprg[i].data[(int)dataPtr],dbuf,(size_t)datasize);
          ReturnToServerCycle = subhdr.numblks - conprg[i].blksin;
          if (conprg[i].blksin < subhdr.numblks) goto bow;
          conprg[i].busy = 2;
          dbuf = conprg[i].data;
          if (tineDebug > 1) dbglog("in-coming long datagram %d: (%d blocks) complete",i,subhdr.numblks);
        }
        else ReturnToServerCycle = 0;
      }
      if (fmt == CF_GSPECTRUM) n = (int)dinsize;
      else if (!isAdjustable) n = subhdr.contract.EqmSizeIn;
      PrepSubData(dbuf,n,subhdr.contract.EqmFormatIn,RTOL);
    }
    else 
    { /* no input data -> make sure the input tag is zeroed out */
      if (subhdr.contract.EqmFormatIn != BFMT(CF_NULL)) ccerr(illegal_format);
      memset(subhdr.contract.strTagIn,0,TAG_NAME_SIZE);
      subhdr.contract.EqmSizeIn = 0;
      dinsize = 0; dinallocsize = 0; 
      dbuf = NULL;
    }
    if (ServerWaiting) ccerr(not_initialized);
    if (subhdr.pollingInterval < MinPollingRate && bmode > CM_SINGLE) ccerr(invalid_interval);
    if (gIsGlobalsServer && bmode == CM_SINGLE && subhdr.pollingInterval == 100)
    { /* special allowance for _recvGlobal() calls */
      gOneTimeSyncPass = TRUE;
    }
    subhdr.pollingInterval = makeValidPollingInterval(subhdr.pollingInterval,BASEMODE(subhdr.mode));
    if (LFMT(subhdr.contract.EqmFormatOut) == CF_DEFAULT)
    { /* caller wants default settings */
      if ((cc=GetDefaultFormatAndSize(&subhdr.contract,stsstr)) != 0) ccerr(cc);
      if ((fmtoutsize=fmtsizeof(subhdr.contract.EqmFormatOut)) != 0)
      { /* default conditions do deliver a payload with a known format */
        fmt = LFMT(subhdr.contract.EqmFormatOut); 
        n = 1; /* number of outgoing adjustable elements requested */
        if (fmt == CF_MDA) fmtoutsize = getCarriedFormatSize(subhdr.contract.strTagOut);
#       ifndef SMALL_TINE_LIB
        if (isAdjustableLengthFormat(fmt))
        {
          n = subhdr.contract.EqmSizeOut;
          subhdr.contract.EqmSizeOut *= getFormatStandardCapacity(fmt);
          sprintf(subhdr.contract.strTagOut,"%d",n);
        }
#       endif
        if ((defsize=subhdr.contract.EqmSizeOut*fmtoutsize) <= 0) ccerr(illegal_data_size);
        hdrsiz = GetFormatHeaderSize(fmt);
        doutsize += n * hdrsiz;
        defsize += n * hdrsiz;
#       ifndef MULTITHREADED
        if (defsize > (UINT32)MaxSystemTransportSize) ccerr(work_area_full);
#       endif
        if (defsize > doutsize)           
        { /* caller can't buffer default */
          ccerr(buffer_too_small);
        }
      }
      else 
      { /* fmtoutsize is 0 => no output payload for this property ! */
        defsize = 0;
      }
      doutsize = defsize; /* for allocation purposes */
    }
    slot = -1; /* start search for available slot */
    canRemoveContractOnError = TRUE;
    for (clid=0; clid<ncontracts; clid++)            
    { /* is contract in list ? */
      if ((cl=ContractList[clid])->expired == CONTRACT_AVAILABLE) 
      { /* could be an available slot */
        if (slot < 0) slot = clid;
        continue;
      }
      if (!concmp(&cl->contract,&subhdr.contract))
      { /* straight-up memcmp : the access carried is only the lower BYTE */
        if (extsiz > 0)
        { /* extended string coming */
          if (cl->extStringSpace == NULL || 
              strncmp(cl->extStringSpace,extstr,extsiz)) continue;
        }
        if (!dbuf && !cl->dataIn) break; /* no data input */
        if (dbuf && cl->dataIn) 
        { /* input buffer has been allocated */
          if (!cmpArrayInput(cl,dbuf,(size_t)dinsize)) break;
        }
      }
    }
    if (clid == ncontracts)                    
    { /* not in list: add it */
      isNewContract = TRUE;
      if (slot >= 0) 
      { /* there is a free re-usable slot in the table */
        clid = slot;
        memset(ContractList[clid],0,sizeof(ContractListStruct));
        if (nExpiredContracts > 0) nExpiredContracts--;
      }
      else
      { /* no free slot: gotta expand the table */
        if (clid >= ContractListCapacity)
        {
          if (tineDebug) dbglog("contract table has reached capacity (%d)",ContractListCapacity);
          ccerr(resources_exhausted);
        }
        if ((ContractList[clid]=(ContractListStruct *)SystemCalloc(1,sizeof(ContractListStruct))) == NULL)
          ccerr(out_of_server_memory);
        ncontracts++;
      }
      cl = ContractList[clid];
      memcpy(&cl->contract,&subhdr.contract,sizeof(CONTRACT));
      if (extsiz > 0)
      {
        if ((cl->extStringSpace = (char *)SystemCalloc(1,DEVICE_NAME_MAXSIZE)) == NULL)
          ccerr(out_of_server_memory);
        cl->extStringSize = extsiz;
        strncpy(cl->extStringSpace,extstr,extsiz);
      }
      cl->listId = clid;
      acc = cl->contract.EqmAccess;
      if (subhdr.mode & CM_MCAST) acc |= CA_NETWORK; /* apply to the 'access' passed to IsIllegalProperty */
      if (cln->inetProtocol & (TCP|TCP6|STREAM|STREAM6)) acc |= CA_CONNECT; /* apply to the 'access' passed to IsIllegalProperty */
      if ((bmode == CM_SINGLE) && (cl->contract.EqmAccess & CA_SYNC) == 0) acc |= CA_SINGLE;
      if ((cc=IsIllegalProperty(cl->contract.EqmName,
                                cl->contract.EqmProperty,
                                cl->contract.EqmDeviceName,
                                cln,acc,&cl->exp,&cl->prp,stsstr)) != 0) ccerr(cc);
#     if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
      if (PutInCmdLog(cl)) 
      { /* log it if this isn't a lock token renewal */
        char datastrbuf[516],addr[64];
        UINT32 blen = 512;
        DTYPE d;
        inetStrAdr(&cln->adr.ip,addr,TRUE);
        if (dbuf != NULL)
        {
          d.dArrayLength = (UINT32)cl->contract.EqmSizeIn;
          d.dFormat = LFMT(cl->contract.EqmFormatIn);
          strncpy(d.dTag,cl->contract.strTagIn,TAG_NAME_SIZE);
          d.data.bptr = dbuf;
          datastrbuf[0] = 0;
          if (!isVariableLengthFormat(d.dFormat))
          {
            _GetValuesArrayAsString(&d, datastrbuf, &blen, " ", TRUE);
          }
          else if (d.dFormat == CF_STRING)
          {
            char *sb[32]; /* max this at 32 strings output */
            int ns=0, sl=0;
            for (ns=0; ns<32; ns++)
            {
              if (sl >= (int)d.dArrayLength) break;
              sb[ns] = (char *)&dbuf[sl];
              sl += (int)strlen(sb[ns]) + 1; /* skip over terminating 0 */
            }
            d.dArrayLength = ns;
            d.data.strptr = sb;
            _GetValuesArrayAsString(&d, datastrbuf, &blen, " ", TRUE);
          }
          if (strlen(datastrbuf) == 0) sprintf(datastrbuf,"%d %.16s value(s)",d.dArrayLength,itof(d.dFormat));
        }
        else
        {
          strcpy(datastrbuf,"none");
        }
#       if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
        srvlog("commands.log","COMMAND","(%.6s)/%.64s[%.64s] called by %.16s from %.64s; input: %.512s",
           cl->contract.EqmName,cl->contract.EqmDeviceName,cl->contract.EqmProperty,cln->usr,addr,datastrbuf);
#       endif
      }
#     endif
      if (bmode > CM_SINGLE)
      { /* persistent link -> check a few things */
        if (IsMcaElement(&cl->contract,cl->prp,&idx,&dlen,&dnam) && 
            !IsMetaProperty(cl->contract.EqmProperty,NULL,NULL,FALSE))
        { /* check for mca access */
          if (tineDebug) dbglog("mca element %d from %d is being accessed",idx,dlen);
          if (chkMcaHandshake(&rh))
          {
            if (tineDebug) dbglog("renegotiate contract with caller");
            memcpy(stsstr,&idx,4);
            memcpy(stsstr+4,&dlen,4);
            strncpy(stsstr+8,dnam,DEVICE_NAME_SIZE);
            ccerr(getMcaPrpCc(&cl->contract));
          }
        }
        if (cl->prp != NULL)
        {
          if (cl->prp->prpAccessMode & CA_STATIC) 
          { /* flagged as static */
            cl->accessFlags |= CA_STATIC;
          }
          if ((cl->prp->prpArrayType & AT_CHANNEL) && (cl->contract.EqmSizeOut > 1))
          { /* is being accessed as an MCA */
            if (cl->prp->mcaValidFloor > (time_t)subhdr.starttime) 
            {
              msglog(0,"send reset mca property (%.6s) %.64s reset signal to caller",
                cl->contract.EqmName,cl->contract.EqmProperty);
              ccerr(reset_mca_property);
            }
          }
        }
      }
      if (cl->prp != NULL && cl->prp->prpId < 0) cl->prp = NULL; /* remove static reference */
      if (cl->prp != NULL && cl->prp->enforceDataReq && !(acc&CA_WRITE))
      { /* a READ contract on a property with forced output */
        if (cl->prp->prpSize != subhdr.contract.EqmSizeOut ||
            cl->prp->prpFormat != LFMT(subhdr.contract.EqmFormatOut))
        { /* rejected ! */
          memcpy(stsstr,&cl->prp->prpSize,4);
          memcpy(stsstr+4,&cl->prp->prpFormat,4);
          ccerr(invalid_datarequest);
        }
      }
      cl->compStatus = not_signalled;
      cl->returnCode = not_posted;
      cl->runInSeparateThread = IsPropertyCalledInSeparateThread(cl);
      if (dinallocsize && (cl->dataIn = (BYTE *)SystemCalloc(1,(size_t)dinallocsize)) == NULL)
      {
        ccerr(out_of_server_memory);
      }
      if (doutsize > 0)
      { /* output payload requested */
        cl->dataBuf = GetPropertyBuffer(cl->contract.EqmName,cl->contract.EqmProperty);
        if (cl->dataBuf != NULL)
        { /* flagged as unbuffered and dataBuf is non NULL */
          cl->unbuffered = TRUE;
          cl->dataOut = cl->dataBuf;
        }
        else if ((cl->dataOut = (BYTE *)SystemCalloc(nDataOutBuffers,(size_t)doutsize)) == NULL)
        { /* multi-threaded => nDataOutBuffers = 2, otherwise srvWorkArea used for calls */
          if (dinsize) SystemFree(cl->dataIn);
          ccerr(out_of_server_memory);
        }
        if (cl->dataBuf == NULL)
        { /* normal case : buffered output payload */
          cl->dataBuf = nDataOutBuffers > 1 ? &cl->dataOut[doutsize] : srvWorkArea;
        }
        cl->dataOutCapacity = doutsize; /* initial contract entry: take what's known */
      }
      if (dinsize > 0 && dbuf) prepArrayInput(cl,dbuf,(size_t)dinsize);
      cl->contract.EqmAccess &= ~(CA_REPEAT); /* this should not inadvertently be sent ! */
      cl->contract.EqmAccess |= CA_FIRST; /* turn on FIRST bit */
      cl->contractID = getNextFreeContractID();
      cl->expired = CONTRACT_ACTIVE;
      cl->pollingInterval = subhdr.pollingInterval; /* reset this too ! */
      cl->lasttime.tv_sec = PRP_SCHEDULE_SIGNAL;
      if (tineDebug > 1)
      {
        dbglog("contract [%d] (%.6s)/%.64s[%.64s] created",
          cl->listId,subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,subhdr.contract.EqmProperty);
      }
    }
    else
    { /* contract was found in the list */
      if ((prp=GetPropertyListStruct(cl->contract.EqmName,cl->contract.EqmProperty,cl->contract.EqmDeviceName)) != NULL)
      {
        canRemoveContractOnError = FALSE;
        el = getExportListItem((char *)cl->contract.EqmName);
        isXPrp = isExclusiveProperty(el,cl->contract.EqmProperty);
        acc = cl->contract.EqmAccess;
        chkAccess = acc&CA_WRITE ? TRUE : FALSE;
        logit = TRUE;
        if (isXPrp & PRP_XREAD_ALL) { chkAccess = TRUE; logit = FALSE; }
        if (cln == NULL) chkAccess = FALSE;
        if (chkAccess) 
        {
          if ((cc=chkClnCredentials(cln,el,prp,cl->contract.EqmProperty,cl->contract.EqmDeviceName,logit)) != 0) ccerr(cc);
        }
        if (prp->minAccessDeadband > 0)
        { /* gotta check the deadband even if same contract shows up in list */
          if ((cc=isPastDeadband(cl->contract.EqmName,cl->contract.EqmProperty,acc,cln)) != 0) ccerr(cc);
        }
        if (prp->prpAccessMode & CA_CONNECT)
        { /* property requires CA_CONNECT access */
          if ((cln->inetProtocol & (MMF|TCP|TCP6|STREAM|STREAM6)) == 0) ccerr(tcp_access_required);
        }
        if (!gOneTimeSyncPass)
        { /* only the globals server will allow this pass */
          if ((prp->prpAccessMode & CA_NETWORK) && (subhdr.mode & CM_MCAST) == 0)
            ccerr(mcast_access_required);
          if ((prp->prpAccessMode & CA_NOSYNC) && (bmode == CM_SINGLE) && (cl->contract.EqmAccess & CA_SYNC) == 0)
            ccerr(async_access_required);
        }
        if ((bmode > CM_SINGLE) && (prp->prpArrayType & AT_CHANNEL) && (cl->contract.EqmSizeOut > 1))
        { /* is being accessed as an MCA */
          if (prp->mcaValidFloor > (time_t)subhdr.starttime) 
          {
            msglog(0,"send reset mca property %.6s %.64s reset signal to caller",
              cl->contract.EqmName,cl->contract.EqmProperty);
            ccerr(reset_mca_property);
          }
        }
      }
      if (tineDebug > 1)
      {
        dbglog("contract [%d] (%.6s)/%.64s[%.64s] (%s) attached",
          cl->listId,subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,subhdr.contract.EqmProperty,
          getContractState(cl->expired));
      }
    }
    if (cl->expired == CONTRACT_EXPIRED && !(subhdr.mode & CM_RETRY)) 
    { /* is a new link attaching to an expired contract ? */
      cl->expired = CONTRACT_ACTIVE;
      cl->contract.EqmAccess |= CA_FIRST; /* turn on FIRST bit */
      cl->lasttime.tv_sec = 0; /* set this far to the rear */
      if (nExpiredContracts > 0) nExpiredContracts--;  /* no longer expired */
      if (tineDebug > 1) dbglog("re-activating contract");
      if (IsMcaElement(&cl->contract,cl->prp,&idx,&dlen,&dnam) && 
          !IsMetaProperty(cl->contract.EqmProperty,NULL,NULL,FALSE))
      {
        if (tineDebug) dbglog("mca element %d from %d is being accessed",idx,dlen);
        if (BASEMODE(subhdr.mode) > CM_SINGLE && chkMcaHandshake(&rh))
        {
          if (tineDebug) dbglog("renegotiate contract with caller");
          memcpy(stsstr,&idx,4);
          memcpy(stsstr+4,&dlen,4);
          strncpy(stsstr+8,dnam,DEVICE_NAME_SIZE);
          ccerr(getMcaPrpCc(&cl->contract));
        }
      }
    }
#   if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
    if (cln->inetProtocol == MMF && subhdr.mode & CM_MCAST) 
    { /* a local client is asking for a network subscription */
      if (gForceShmDatagrams || getNetworkClientFromList() == NULL)
      { /* there's no network client going at the moment */ 
        subhdr.mode &= ~(CM_MCAST); /* strip this off for a more efficient MMF transfer */
      }
    }
#   endif
    if (subhdr.mode & CM_MCAST)
    { /* is broadcast or multicast request ? */
      isTrueMCast = subhdr.mode & CM_MCAST;
      subhdr.mode &= 0x000f; /* strip off */
      if (subhdr.mode == CM_CANCEL) ccerr(0); /* don't do anything */
      if (cln->tineProtocol != gTineProtocol) ccerr(not_accepted);
      if (IsMemberBcastNets(cln) || isTrueMCast)
      { /* send subscription ID to caller */
        nwcln = getNetworkClientFromList();
        if (nwcln != NULL && nwcln->tineProtocol != cln->tineProtocol)
        { /* NETWORK client is there but using a different protocol level as this client ! */
          cc = nwcln->tineProtocol > cln->tineProtocol ? not_accepted : illegal_protocol;
          ccerr(cc);
        }
        subid = SPEEK(cl->contractID);
        sendMessageToCaller(hprot,cln,&subhdr,get_subscription_id,(BYTE *)&subid,2);
        if (nwcln != NULL) 
        { /* already servicing the NETWORK client */
          cln = nwcln;
          if (tineDebug > 1) dbglog("additional NETWORK client attaching");
        }
        else
        { /* Network client not yet in list */
          memcpy(&newcln,cln,sizeof(ClnHdr)); /* don't dork the existing settings */
          cln = &newcln;
          if (cln->inetProtocol == PIPE || cln->inetProtocol == MMF)
          { /* switch from local address ... */
            cln->inetProtocol = (gAfInet == AF_INET6) ? UDP6 : UDP; 
            memcpy(getInAddr(&cln->adr.ip),getmyipaddr(),gInAdrLen);
            cln->ep.sck = udpSrvSck; /* reset this in case of windows MMF */
          }
          if (zerocmp(getInAddr(&cln->adr.ip),gInAdrLen))
          { /* add network to client list and replace cln */
            BYTE *badr = (BYTE *)&tMCastAddr;
            if (!isTrueMCast)
            {
              badr = getInAddr(&cln->adr.ip); 
              badr[gInAdrLen-1] = 0xff;
            }
            sckadr_in(&cln->adr.ip,badr,(UINT16)gtMCastPort,gAfInet);
            cln = LocateConsumerInList(&cln->adr.ip,NULL,NULL);
            if (tineDebug > 1) dbglog("add NETWORK to client list");
          }
#         if defined(IPXNW) || defined(IPXBSD)
          else if (*(UINT32 *)cln->IPXaddress.network != 0)
          { /* IPX Client ! -> add network to client list and replace cln */
            memset(&cln->ep.ipx.sckRmt,MCAST_SOCKET,2);
            memset(cln->adr.ipx.IpxNode,0xff,6);
            cln = LocateConsumerInList(NULL,&cln->adr.ipx,NULL);
          }
#         endif
          cln->ep.sck = newcln.ep.sck; cln->inetProtocol = newcln.inetProtocol;
          cln->tineProtocol = newcln.tineProtocol;
          strncpy(cln->usr,"NETWORK",USERNAME_SIZE);
          strncpy(cln->typ,"NETWORK",USERTYPE_SIZE);
          cln->pid = -1;
        }
        subhdr.id = cl->contractID;
        subhdr.starttime = BCAST_ID;
      }
    }
    if (WaitForSystemMutex(hClientTableMutex,-1) == 0) 
    { /* is client in contractList?    */
      for (c=cl->clnLst; c!=NULL && c->client!=cln; c=c->next);
      if (c == NULL) 
      { /* no, add to list */
        if (!IsStockProperty(cl->contract.EqmProperty))
        { /* add to followers list if contract to a registered property */
          addFollower(cl->contract.EqmName,cln->usr,USERNAME_SIZE);
        }
        if (BASEMODE(subhdr.mode) == CM_CANCEL) 
        { /* a new client cancelling his link makes no sense ! */
          /* TODO: this doesn't work ! apparently something else has to run its course !
             the TCP thread gets hung?
          swallowOnErr = TRUE;
          ccerr(invalid_link);
          */
        }
        isNewClient = TRUE;
        if ((c=(ClnLst *)SystemCalloc(1,sizeof(ClnLst))) == NULL) 
        {
          ReleaseSystemMutex(hClientTableMutex);
          ccerr(out_of_server_memory);
        }
        c->next = cl->clnLst;
        cl->clnLst = c;
        c->client = cln;
        cl->nclients++;
        c->client->ncontracts++;
        c->client->tineProtocol = hprot;
        c->status.statusCode = cl->returnCode;
        c->status.counter = 1;
        resetClnCounters = TRUE;
        if (tineDebug > 1) dbglog("append client %.16s to contract %d",cln->usr,cl->listId);
      }
      else if (!cl->deliveryPending)
      {
        if (cl->nclients == 0)
        { /* this shouldn't happen */
          if (tineDebug) 
          {
            dbglog("re-activate contract (%.6s)/%.64s[%.64s]",
              subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,subhdr.contract.EqmProperty);
          }
          cl->nclients++;
          cl->contract.EqmAccess &= ~(CA_LAST);
        }
        resetClnCounters = TRUE;
      }
      ReleaseSystemMutex(hClientTableMutex); 
    }
    isNewStarttime = c->status.starttime != subhdr.starttime ? TRUE : FALSE;
    c->status.pollingInterval = subhdr.pollingInterval;
    c->status.id = subhdr.id;
    c->status.starttime = subhdr.starttime;
    datamtu = PMTU((UINT16)gPacketMtu,c->client->inetProtocol,hprot);
    if (fmtoutsize > 0) datamtu -= datamtu%fmtoutsize;
    if (resetClnCounters)
    {
      c->status.blknum = 1;
      c->status.nextDataPtr = 0;
      c->status.bytesRemaining = doutsize;
      c->status.bytesTotal = doutsize;
      c->status.mtu = datamtu;
      c->status.nextDataSize = c->status.bytesRemaining;
      if (!isJumboDelivery(cln))
      { /* normal case */
        if (c->status.bytesRemaining > c->status.mtu) c->status.nextDataSize = c->status.mtu;
        c->status.numblks = doutsize/((UINT32)c->status.mtu) + (doutsize%((UINT32)c->status.mtu) ? 1 : 0);
        if (doutsize == 0) c->status.numblks = 1;
      }
      else
      { /* streams always get send in one go */
        c->status.numblks = 1;
      }
    }
    renewalMultiplier = getThisSubscriptionRenewalMultiplier(cl,c);
    if (c->status.mode != HBT_ACK_PENDING)
    { /* not a simple contract renewal : force Contract execution */
      int canDelayUpdate = (bmode == CM_DATACHANGE || bmode == CM_EVENT) ? TRUE : FALSE;
      if (bmode != CM_CANCEL) gServerCycleNeedsScheduler = TRUE;
      if (tineDebug > 3 && gServerCycleNeedsScheduler) dbglog("incoming request -> call scheduler");
      if (subhdr.starttime == BCAST_ID)
      { /* latching on to a NETWORK contract */
        if ((cl->contract.EqmAccess & CA_FIRST) == 0)
        { /* wait for next delivery cycle */
          if (tineDebug > 1) dbglog("additional network client to contract");
          if (cl->pollingInterval > 1000) c->status.lasttime.tv_sec -= cl->pollingInterval;
        }
        else
        {
          if (tineDebug > 1) dbglog("initial network client for contract");
        }
      }
      else
      { /* NOT a NETWORK contract ! */
        chkResendData = (subhdr.mode == CM_RETRY || c->status.counter == 0 || canDelayUpdate);
        if (cl->nclients > 1)
        { /* client is attaching to a contract */
          msecs = MSECS(thisCycleTime,cl->lasttime);
          if (bmode == CM_SINGLE && msecs <= (SINT32)subhdr.pollingInterval)            
          { /* have data within the desired timeout -> use them */
            chkResendData = TRUE;
          }
          else if (bmode == CM_TIMER && msecs < (SINT32)(subhdr.pollingInterval) - 50)
          { /* will be at least 50 msec before I look again ! */
            chkResendData = TRUE;
          }
        }
      }
      if (chkResendData && cl->returnCode != not_posted)
      { /* can send out the last acquired data */
        int contractValid = isNewClient;
        int renewal = CTR_RENEWAL*renewalMultiplier;
        if (!contractValid)
        { /* not a new client ! */
          if (subhdr.mode == (CM_RETRY|CM_SINGLE)) /* remove CM_SINGLE at a later date (java problem) */
          { /* flagged as a RETRY */
            contractValid = TRUE;
          }
          else if (c->status.counter == renewal)
          { /* possible ungraceful exit and restart of remote client */
            if (isNewStarttime) contractValid = TRUE;
          }
          else if (canDelayUpdate)
          { /* DATACHANGE or EVENT => there could be a long wait ... */
            if ((c->client->inetProtocol & (UDP|UDP6)) && 
                isNewStarttime && c->status.counter < renewal-1)
            { /* possible ungraceful exit and restart of remote client in DATACHANGE/EVENT mode */
              contractValid = TRUE;
            }
          }
        }
        if (contractValid) 
        { /* contract return code can be trusted */
          resendCurrentData = TRUE;
          c->status.Stale = CX_ATTACH; 
          c->status.lasttime = thisCycleTime;
          if (cl->dataSizeOut < cl->contract.EqmSizeOut)
            renegotiateContract(cl,cl->returnCode);
        }
      }
      if (bmode != CM_CANCEL && !canDelayUpdate &&
          c->status.counter > CTR_ACK_REQUEST*renewalMultiplier &&
          c->status.counter < CTR_RENEWAL*renewalMultiplier)
      {
        msglog(0,"reconnecting to (%.6s)/%.64s[%.64s] from %.16s (+%d)",
          subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,
          subhdr.contract.EqmProperty,rh.userName,c->status.counter);
        c->status.reconnects++;
        gClientReconnects++;
      }
      StaleData++;
    }
    if (subhdr.mode & CM_RETRY)
    {
      msglog(0,"retrying (%.6s)/%.64s[%.64s] by %.16s",
        subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,
        subhdr.contract.EqmProperty,rh.userName);
      renegotiateContract(cl, cl->returnCode);
      c->status.Stale = CX_RETRY;                     
      StaleData++;   /* if expired, doDelivery() will send the last payload again */
      c->status.isRetry = TRUE;
      gClientRetries++;
      subhdr.mode = BASEMODE(subhdr.mode);
      sendPrpSignal(cl->prp,CS_RETRY,-1,cc);
      sendContractSignal(subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,
                         subhdr.contract.EqmProperty,CS_RETRY,cc);
    }
    c->status.mode = subhdr.mode;
    switch (BASEMODE(subhdr.mode))
    {
      case CM_SINGLE :
        gSingleLinkCount++;
        c->status.counter = 1;
        if (cl->expired == CONTRACT_ACTIVE && cl->runInSeparateThread && !isNewClient)
        { /* impatient client */
          ccerr(0);
        }
        if (!(cl->expired == CONTRACT_EXPIRED && c->status.isRetry))
        { /* normal case: not expired and not a retry */
          if (!resendCurrentData) cl->lasttime.tv_sec = PRP_REQUEST_SIGNAL;  
        }
        else  
        { /* a RETRY or an expired contract */
          if (cl->expired == CONTRACT_EXPIRED) 
          { /* expired: should just resend last data, but reset FIRST and LAST just in case */
            cl->lasttime.tv_sec = (UINT32)time(NULL);
            cl->contract.EqmAccess |= (CA_FIRST|CA_LAST);
          }
          if (cl->dataSizeOut < cl->contract.EqmSizeOut)
            renegotiateContract(cl,cl->returnCode);
        }
        if (cl->pollingInterval == 0) cl->pollingInterval = c->status.pollingInterval;
        if (cl->nclients == 1)
        { /* turn on LAST bit */
          cl->contract.EqmAccess |= CA_LAST;
        }
        sendSingleDelivery = TRUE;
        tgtcon = cl;
        tgtcln = cln;
        break;
      case CM_TIMER:
      case CM_DATACHANGE:
      case CM_EVENT:
        if (c->status.counter < CTR_RENEWAL*renewalMultiplier - 1)
          c->status.counter = CTR_RENEWAL * renewalMultiplier;
        cl->pollingInterval = findpoll(subhdr.pollingInterval,cl->pollingInterval);
        break;
      default     :
      case CM_CANCEL :
        if (!strcmp(c->client->usr,"NETWORK")) break;
        if (cl->nclients == 1)
        {
          if (isNewContract)
          { /* cancelling a brand new link makes no sense ! */
            swallowOnErr = TRUE;
            ccerr(invalid_link);
          }
          if (c->status.counter) 
          { /* was active : turn on LAST bit */
            cl->contract.EqmAccess |= CA_LAST; 
          }
          cl->lasttime.tv_sec = PRP_CANCEL_SIGNAL;
          cl->deliveryPending = 0;
        }
        c->status.counter = 0;
        break;
    }
    if (tineDebug > 2)
      dbglog("link counter set to %d for %.16s (mode %d)",c->status.counter,c->client->usr,c->status.mode);
err:
    if (cc)
    { /* error condition */
      if (strlen(stsstr) == 0)
      { /* no stsstr yet known */
        switch (cc)
        {
          case server_redirection:
          case call_redirection:
          case property_is_mca:
          case property_is_mca_elevated:
          case invalid_datarequest:
            break; /* don't touch the contents of stsstr ! */
          default:
            strncpy(stsstr,cc2str(cc),STATUS_SHORTSIZE);
            break;
        }
      }
      if (!swallowOnErr) sendMessageToCaller(hprot,cln,&subhdr,cc,(BYTE *)stsstr,STATUS_SIZE);
      if (clid >= 0 && clid < ContractListCapacity && 
          cl->expired == CONTRACT_ACTIVE &&
          canRemoveContractOnError) 
      { /* had been so far as to have a valid contract entry ... */
        if (removeContractTableEntry(clid) == 0) nExpiredContracts++;
      }
      cc = 0;
    }
    for (i=0; i<MAXLONGGRAMS; i++) if (conprg[i].busy == 2)
    {
      SystemFree(conprg[i].data);
      memset(&conprg[i],0,sizeof(ContractProgress));
      if (tineDebug > 1) dbglog("discarding long datagram %d",i);
    }
bow:
    if (subhdr.msgsize <= 0) 
    { /* appears to be super corrupt ! */
      if (tineDebug || errorDebug) dbglog("corrupt message size : %d",subhdr.msgsize);
      break; 
    }
    if ((bytesread += subhdr.msgsize) < totalsize)
    {
      resendCurrentData = FALSE;
      if ((cc=prepIncomingSubRequestHdr(hprot, &subhdr, &buf[bytesread], &swallowOnErr)) != 0) ccout(cc);
      extsiz = 0;
      if (subhdr.extsize > 0)
      {  /* extended (device name) string set */
         extsiz = MIN(subhdr.extsize,DEVICE_NAME_MAXSIZE);
         memcpy(extstr,&buf[bytesread+SUBREQ_STEPSIZE(hprot)],extsiz);
         if (extsiz % 2) 
         {
           extstr[extsiz] = 0;
           extsiz++;
         }
      }
      else
      { /* this should be an even number of bytes */
        if (subhdr.msgsize % 2 != 0) break;
      }
      replaceContractAliasNames(&subhdr.contract);
    }
  } /* continue unpacking  */
  if (nreqs > 1) sendSingleDelivery = FALSE; 
  if (disallowSingleDelivery) sendSingleDelivery = FALSE;
  if (gServerCycleNeedsScheduler && sendSingleDelivery)
  {
    if (tineDebug > 1)
    {
      dbglog("send synchronous express contract (%.6s)/%.64s[%.64s] to %.16s (call contract dispatch : %s)",
        subhdr.contract.EqmName,subhdr.contract.EqmDeviceName,
        subhdr.contract.EqmProperty,rh.userName,tgtcon->lasttime.tv_sec == PRP_REQUEST_SIGNAL ? "YES" : "NO");
    }
    doScheduler(&thisCycleTime,gSystemTick,SCHED_RESPONSE,tgtcln,tgtcon);
    gServerCycleNeedsScheduler = FALSE;
    cln = NULL; /* null this here to avoid the check below */
  }
out:
  gOneTimeSyncPass = FALSE;
  if (cc && !swallowOnErr) 
  {
    if (strlen(stsstr) == 0) strncpy(stsstr,cc2str(cc),32);
    sendMessageToCaller(hprot,cln,&subhdr,cc,(BYTE *)stsstr,STATUS_SIZE);
  }
  if (cln != NULL && cln->ncontracts == 0) removeClient(cln);
  return cc;
}
TINE_EXPORT void SetSendLegacyGlobals(int value)
{
  gGlobalsProtocol = value ? 4 : DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
}
TINE_EXPORT int GetGlobalsProtocol(void)
{
  return gGlobalsProtocol;
}
int prepOutgoingGCastHeader(int protocol,GlbDataHdr *ghdr,BYTE *stream)
{
  switch (protocol)
  {
    case 4:
    {
      GlbDataHdr4 hdr;
      strncpy(hdr.name, ghdr->name, KEYWORD_NAME_SIZE);
      hdr.format = ghdr->format;
      hdr.size = ghdr->size;
      hdr.dStamp = ghdr->dStamp;
      hdr.sysStamp = ghdr->sysStamp;
      hdr.dTimeStamp = ghdr->dTimeStamp;
      hdr.dTimeStampUSEC = (short)(ghdr->dTimeStampUSEC % 1000);
      hdr.dTimeStampMSEC = ghdr->dTimeStampUSEC / 1000;
      hdr.status = ghdr->status;
#     ifdef INTEL
      memcpy(stream, &hdr, sizeof(GlbDataHdr4));
#     else
      _structSwap(GlbDataHdr4Struct, stream, 0, NULL, (BYTE *)&hdr, LTOR);
#     endif
      return sizeof(GlbDataHdr4);
    }
    default:
#     ifdef INTEL
      memcpy(stream, ghdr, sizeof(GlbDataHdr));
#     else
      _structSwap(GlbDataHdrStruct, stream, 0, NULL, (BYTE *)ghdr, LTOR);
#     endif
      return sizeof(GlbDataHdr);
  }
}
void doGCast(struct timeval *thisTime)
{
  BYTE *buf = srvWorkArea;
  int cc = 0,i,n,setsize,GlobalPacketSize,datasize,pkthdrsize;
  char *strptr;
  GlbDataHdr gdhdr, *gd = &gdhdr;
  int delta_t, usecs;
  double ts = MakeDataTimeStamp();
  time_t interim = thisTime->tv_sec;

  if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick) != 0) return;
  pkthdrsize = PAYLOADHDR_SIZE(gGlobalsProtocol);
  GlobalPacketSize = pkthdrsize;
  for (i=0, n=0; i<ngcasts; i++)
  {
    if (gcastTbl[i]->tv.tv_sec > 0 &&  /* not a scheduled broadcast */
        (gcastTbl[i]->minPeriod <= 0 || gcastTbl[i]->maxPeriod <= 0)) continue;
    delta_t = (int)MSECS(*thisTime,gcastTbl[i]->tv);
    if (delta_t < gcastTbl[i]->minPeriod) continue; /* too soon */
    if (delta_t < gcastTbl[i]->maxPeriod) 
    { /* within heartbeat */
      if (!datacmp(gcastTbl[i]->refdata,
            gcastTbl[i]->data,
            gcastTbl[i]->size,
            gcastTbl[i]->format,
            gcastTbl[i]->tolerance)) continue;
    }
    datasize = gcastTbl[i]->size*fmtsizeof(gcastTbl[i]->format);
    setsize = GLBDATAPKT_SIZE(gGlobalsProtocol) + datasize;
    if (setsize > gGCastMaxPayload - pkthdrsize)
    {
      if (tineDebug || errorDebug) dbglog("global %.64s data set too large",gcastTbl[i]->keyword);
      continue;
    }
    gcastTbl[i]->isBeingSent = TRUE;
    if (GlobalPacketSize + setsize > gGCastMaxPayload)
    { /* send what's there  */
      prepOutgoingRequestHdr(gGlobalsProtocol, GlobalPacketSize, n, gUserName, gGlbType, buf);
      cc = sendGCast(buf,GlobalPacketSize);
      GlobalPacketSize = pkthdrsize;
      n = 0;
    }
    memset(gd,0,sizeof(GlbDataHdr));
    strncpy(gd->name,gcastTbl[i]->keyword,KEYWORD_NAME_SIZE);
    gd->size = (UINT16)gcastTbl[i]->size;
    gd->format = gcastTbl[i]->format;
    gd->dTimeStamp = (int)ts;
    usecs = (int)((ts - (double)gd->dTimeStamp) * 1000000.0);
    gd->dTimeStampUSEC = usecs;
    gd->status = gcastTbl[i]->status;
    gd->dStamp = gcastTbl[i]->dataStamp;
    gd->sysStamp = gcastTbl[i]->sysStamp;
    gd->interval = gcastTbl[i]->maxPeriod;
    gcastTbl[i]->time = interim;
    gcastTbl[i]->tv = *thisTime;
    GlobalPacketSize += prepOutgoingGCastHeader(gGlobalsProtocol,gd,&buf[GlobalPacketSize]);
    if (gcastTbl[i]->data)
    {
      memcpy(gcastTbl[i]->refdata,gcastTbl[i]->data,datasize);
      prepOutgoingData(&buf[GlobalPacketSize],gcastTbl[i]->data,datasize,gcastTbl[i]->format);
    }
    if (tineDebug > 1 || (cc && errorDebug))
    {
      sprintf(dbgbuf,"GCAST: %.64s ",gcastTbl[i]->keyword);
      strptr = &dbgbuf[strlen(dbgbuf)];
      if (cc) sprintf(strptr,"(ERR: %.32s)",cc2str(cc));
      dbglog(dbgbuf);
      if (cc == 0 && tineDebug > 2)
      {
        DataDump("GCAST: ",(char *)gcastTbl[i]->data,gcastTbl[i]->size,gcastTbl[i]->format,LTOR);
      }
    }
    if (gcastTbl[i]->cbfcn) TCBCALL(gcastTbl[i],cc);
    GlobalPacketSize += datasize; n++;
    gcastTbl[i]->isBeingSent = FALSE; /* has at least been packaged */
  }
  if (n > 0) 
  { /* something left */
    prepOutgoingRequestHdr(gGlobalsProtocol, GlobalPacketSize, n, gUserName, gGlbType, buf);
    cc = sendGCast(buf,GlobalPacketSize);
  }
  ReleaseSystemMutex(hSystemKernelMutex);
  return;
}

char tExecLinkReturnedCompletion[STATUS_SIZE+4];
int tExecLinkReturnedDataSize = 0;
int gTineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
int gGlobalsProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
int tExecLink(char *fec,char *eqm,char *dev,char *prp,DTYPE *dout,DTYPE *din,short access,short timout)
{
  int fecindex,fmtsize,fmtsizeIn,hprot,retry = 2;
  short siz=0,fmt=CF_NULL,sizIn=0,fmtIn=CF_NULL;
  char *datOut=NULL,*datIn=NULL,*stagOut=NULL,*stagIn=NULL;
  short cc = 0, dsize, dsizeIn, dretsiz = 0;
  UINT32 totalsize, sstepsize, pstepsize, pkthdrsize, rsphdrsize;
  static BYTE lclWorkArea[IPX_BUFFER_SIZE];
  static UINT16 tid = 0;            /* trivial subscription id */
  SubRspHdr subhdr;
  ConTblEntry c;

  if (tineDebug) dbglog("trivial ExecLink: %.16s:%.6s/%.64s[%.64s]",fec,eqm,dev,prp);
  if (FecTbl == NULL) return not_initialized;
  if ((fecindex=getFecIndex(fec)) < 0) return -fecindex;
  hprot = FecTbl[fecindex].tineProtocol;
  if (hprot == 0) hprot = gTineProtocol;
  if (dout != NULL)
  {
    fmt = dout->dFormat;
    siz = dout->dArrayLength;
    datOut = dout->data.cptr;
    stagOut = dout->dTag;
  }
  if (din != NULL)
  {
    fmtIn = din->dFormat;
    sizIn = din->dArrayLength;
    datIn = din->data.cptr;
    stagIn = din->dTag;
  }
  fmtsize = fmtsizeof(fmt);
  fmtsizeIn = fmtsizeof(fmtIn);
  pkthdrsize = PAYLOADHDR_SIZE(hprot);
  if (sizIn > 0 && datIn == NULL) return argument_list_error;
  if (strlen(eqm) > EQM_NAME_SIZE || strlen(prp) > PROPERTY_NAME_SIZE) return string_too_long;
  if ((dsize = siz*fmtsize) < 0) return out_of_range;
  if ((dsizeIn = sizIn*fmtsizeIn) < 0) return out_of_range;
  if ((RSPHDR_SIZE+SUBRSPHDR_SIZE + dsize) > IPX_BUFFER_SIZE) return use_stream_transport;
  if ((pkthdrsize + SUBREQ_STEPSIZE(hprot) + dsizeIn) > IPX_BUFFER_SIZE) return use_stream_transport;
  if (timout < 1) timout = 1;
  if (hprot < 5 || hprot > DEFAULT_TRANSPORT_PROTOCOL_LEVEL) return illegal_protocol;

  if (WaitForSystemMutex(hTxMutex,-1) != 0) return semaphore_busy;
pRETRY:
  memset(&c,0,sizeof(ConTblEntry));
  c.mode = CM_SINGLE; 
  c.starttime = (UINT32)time(NULL);
  strncpy(c.prpName,prp,PROPERTY_NAME_SIZE);
  strncpy(c.devName,dev,DEVICE_NAME_SIZE);
  if (stagOut != NULL) strncpy(c.strTagOut,stagOut,TAG_NAME_SIZE);
  if (stagIn != NULL) strncpy(c.strTagIn,stagIn,TAG_NAME_SIZE);
  c.fecIdx = fecindex;
  c.sizeIn = sizIn; 
  c.formatIn = fmtIn;
  c.sizeOut = siz; 
  c.formatOut = fmt;
  c.access = access;
  c.datamtu = SUBMTU(FecTbl[fecindex].inetProtocol,hprot); /* input data mtu */
  c.blkidIn = 0;
  c.blknumIn = 1;
  c.numblksIn = 1;
  sstepsize = SUBREQ_STEPSIZE(hprot);
  pstepsize = SUBRSP_STEPSIZE(hprot);
tRETRY:
  c.tineProtocol = hprot;
  FecTbl[fecindex].tineProtocol = hprot;
  pkthdrsize = PAYLOADHDR_SIZE(hprot);
  rsphdrsize = RETURNHDR_SIZE(hprot);
  /* below assumes that a data payload has already been prepared (swapped prior to entering) */
  if (sizIn > 0)
  {
    memcpy(&lclWorkArea[pkthdrsize+sstepsize],datIn,dsizeIn);
  }
  prepOutgoingSubRequestHdr(hprot,&c,eqm,dsizeIn,tid,&lclWorkArea[pkthdrsize]);
  totalsize = pkthdrsize + sstepsize + dsizeIn;
  prepOutgoingRequestHdr(hprot, totalsize, 1, gUserName, GetUserType(), (BYTE *)lclWorkArea);
  switch (cc = tCall(fecindex,lclWorkArea,totalsize,timout))
  {
    case 0: 
      break;
    case illegal_protocol:
      if (hprot > 5)
      { /* minumum supported legacy protocol level = 5 */
        hprot--;
        goto pRETRY;
      }
    default:
      if (isTimeout(cc) && --retry > 0) goto tRETRY;
      ccerr(cc);
  }
  prepIncomingSubResponseHdr(hprot,&subhdr,&lclWorkArea[rsphdrsize]);
  if (subhdr.subId != tid) ccerr(connection_timeout);
  if ((cc= subhdr.retCode) == 0 || (cc & CE_SENDDATA))
  {
    if (dsize < (dretsiz= subhdr.msgSize-pstepsize)) ccerr(illegal_format);
    memcpy(datOut,&lclWorkArea[rsphdrsize+pstepsize],dretsiz);
  }
  else
  {
    strncpy(tExecLinkReturnedCompletion,(char *)&lclWorkArea[rsphdrsize+pstepsize],STSSIZE(hprot));
  }
err:
  tid = (tid + 1) % 10000;
  if (cc && (tineDebug || errorDebug))
  {
    dbglog("trivial ExecLink error: <%d> %s",cc, cc > 0 && cc < numErr ? cc2str(cc) : "unknown");
  }
  tExecLinkReturnedDataSize = fmtsize > 0 ? dretsiz / fmtsize : 0;
  ReleaseSystemMutex(hTxMutex); 
  return cc;
}
