/*
 * File srvdbase.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file srvdbase.c
 * \brief TINE server engine and registration API routines
 *
 * The API routines presented here deal almost exclusively with server-side registration calls.
 * All TINE servers must export their behavior in the form of properties, devices, server names,
 * etc.  The routines offered here provide ways to do this.  In most cases there are also 
 * alternatives based on startup configuration files such as fecid.csv, exports.csv, etc.  On
 * front-end systems where a local disk is available, this is the suggested registration 
 * method, since it avoids the hard coding of 'names' inside of the server programs themselves.
 *
 */
#ifndef NO_DOXYGEN
#include "project.def"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#ifdef WIN32
# include <direct.h>
extern int WSAinitialized;
#endif
#ifdef MACOS
# include <sys/types.h>
# include <sys/sysctl.h>
#endif
#if defined(UNIX) || defined(MACOS) || defined(VXWORKS)
# include <sys/stat.h>
# include <dirent.h>
#endif
#ifdef SOLARIS
# include <procfs.h>
# undef PS_NONE
#endif
#include "prolog.h"
#include "srvcore.h"
#include "threader.h"
#include "srvip.h"
#if defined(IPXNW) || defined(IPXBSD)
# include "srvipx.h"
#endif
#include "dbtools.h"
#include "toolkit.h"
#include "errors.h"
#include "hstlib.h"
#include "almlib.h"
#include "globals.h"
#include "ttycmd.h"
#if defined(MAKE_TINE32_DLL)
# include "shmalloc.h"
#endif
#endif /* NO_DOXYGEN */

time_t appmktime = 0;
char appmkversion[64] = "1.00.0000";
TINE_EXPORT char gFecName[FEC_NAME_SIZE+4] =  FECNAME_UNKNOWN;
TINE_EXPORT char gDeviceContext[CONTEXT_NAME_SIZE+4] = "";
TINE_EXPORT char gDeviceSubSystem[SUBSYSTEM_NAME_SIZE+4] = "";
TINE_EXPORT int gScanForNameLists = SCAN_FOR_NAME_LISTS;
TINE_EXPORT int gScanForDeviceLists = SCAN_FOR_DEVICE_LISTS;
extern char *vFeclogBuffer;

int gSystemExitCode = 0;
int gTineConfigurationCoded = TINE_CONFIGURATION_CODED;
int queryDeviceNamesFromProperties = FALSE;
int acquireDeviceNamesFromProperties = FALSE;
int isUsingXMLFecDatabase = FALSE;
int gSystemRunningStandAlone = FALSE;
int gEqmInitializedIdleState = FALSE;
int gIgnoreCentralServices = FALSE;
extern int logStructureEntries;
int seekAlternateExportsFile = FALSE; /* remove this at Release 4 time */
int gEqmListSize = 0;
int gEnsAccepted = FALSE;
int gEnsAddrResolutionInProgess = FALSE;
int gMarshallBooleanAsByte = FALSE;
TINE_EXPORT void SetMarshallBooleanAsByte(int value)
{
  gMarshallBooleanAsByte = value;
}
TINE_EXPORT int GetMarshallBooleanAsByte(void)
{
  return gMarshallBooleanAsByte;
}
/** \internal */
int updateFecManifest(ExportListStruct *el);
int renegotiateContract(ContractListStruct *cl,int cc);
int addPrpDbaItem(char *eqm,char *prp,int access,int deadband);
int isOverloadedWriteProperty(ExportListStruct *el, ExportPropertyListStruct *pl);

static ClnHdr qrycln;
static ClnLst qrycon;
static ContractListStruct qrycl;

void marshallBoolean(BYTE *values, int length, int dir)
{
  _Bool *bvals = (_Bool *)values;
  int *ivals = (int *)values;
  int i;

  if (!gMarshallBooleanAsByte) return;
  if (dir == RTOL)
  {
    for (i=0; i < length;  i++) bvals[i] = (_Bool)ivals[i];
  }
  else
  {
    for (i=length-1; i >= 0; i--) ivals[i] = (int)bvals[i];
  }

}
#ifndef SMALL_TINE_LIB
int getFilesOfType(const char *path,const char *pattern,NAME64 *lst,int lstsiz)
{
  int nfound=0;
# if defined(UNIX) || defined(VXWORKS)
  DIR *dp;
  struct dirent *dirp;
  char *c, substr[128];
# elif defined(WIN32)
  char dirname[132];
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

  if (pattern == NULL || strlen(pattern) == 0) return 0;
# if defined(UNIX) || defined(VXWORKS)
  if ((dp=opendir((char *)path)) == NULL) return 0;
  for (c=(char *)pattern; *c == '*'; c++);
  strncpy(substr,c,128);
  if ((c=strchr(substr,'*')) != NULL) *c = 0;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name,substr) == NULL) continue;
    if (lst != NULL && nfound < lstsiz) strncpy(lst[nfound].name,dirp->d_name,64);
    nfound++;
  }
  closedir(dp);
# elif defined(WIN32)
  _snprintf(dirname,128,"%s\\%s",path,pattern);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (lst != NULL && nfound < lstsiz) strncpy(lst[nfound].name,fd.name,64);
    nfound++;
  } while (_findnext(hfp,&fd) != -1);
  _findclose(hfp);
# endif
  return nfound;
}
int hasFilesOfType(const char *path,const char *pattern)
{
  int found=FALSE;
# if defined(UNIX) || defined(VXWORKS)
  DIR *dp;
  struct dirent *dirp;
  char *c, substr[128];
# elif defined(WIN32)
  int len;
  char dirname[128];
  struct _finddata_t fd;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif

  if (gTineConfigurationCoded) return FALSE;
  if (pattern == NULL || strlen(pattern) == 0) return FALSE;
# if defined(UNIX) || defined(VXWORKS)
  if ((dp=opendir((char *)path)) == NULL) return 0;
  for (c=(char *)pattern; *c == '*'; c++);
  strncpy(substr,c,128);
  if ((c=strchr(substr,'*')) != NULL) *c = 0;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name,substr) == NULL) continue;
    found = TRUE;
    break;
  }
  closedir(dp);
# elif defined(WIN32)
  if ((len=(int)strlen(path)) > 0)
  {
    if (path[len-1] != FS_DELIMITER) 
      snprintf(dirname,128,"%s\\%s",path,pattern);
    else
      snprintf(dirname,128,"%s%s",path,pattern);
  }
  else
    strncpy(dirname,pattern,128);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return FALSE;
  found = TRUE;
  _findclose(hfp);
# endif
  return found;
}

TINE_EXPORT void SetConfigurationCoded(int value)
{
  gTineConfigurationCoded = value;
}
TINE_EXPORT int GetConfigurationCoded(void)
{
  return gTineConfigurationCoded;
}
TINE_EXPORT void SetInitializeIdle(int value)
{
  gEqmInitializedIdleState = value;
}
TINE_EXPORT int IsStandAlone(void)
{
  return gSystemRunningStandAlone;
}
TINE_EXPORT int SetStandAlone(int value)
{
  if (gSystemStartupTime > 0)
  {
    feclogEx(TINE_LOGLEVEL_WARN, "setting stand-alone via API refused as it has not effect");
    return already_assigned;
  }
  gSystemRunningStandAlone = value;
  return 0;
}
TINE_EXPORT char *GetExportedContext(char *eqmName)
{
  ExportListStruct *el = getExportListItem(eqmName);
  char *ctx = el == NULL ? gDeviceContext : el->EqmContext;
  static char contextBuffer[CONTEXT_NAME_SIZE];
  strncpy(contextBuffer,ctx,CONTEXT_NAME_SIZE);
  return contextBuffer;
}
TINE_EXPORT void SetExportedContext(char *eqmName,char *context)
{
  ExportListStruct *el = getExportListItem(eqmName);
  if (context == NULL || strlen(context) == 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "SetExportedContext: invalid argument");
    return;
  }
  if (el == NULL)
  {
    strncpy(gDeviceContext,context,CONTEXT_NAME_SIZE);
    return;
  }
  strncpy(el->EqmContext,context,CONTEXT_NAME_SIZE);
  return;
}
TINE_EXPORT void SetDefaultDeviceRegion(char *eqmName,char *region)
{
  ExportListStruct *el = getExportListItem(eqmName);
  if (el != NULL) el->region = GetValidDeviceRegion(eqmName,region);
  return;
}
TINE_EXPORT int GetDefaultDeviceRegion(char *eqmName)
{
  ExportListStruct *el = getExportListItem(eqmName);
  if (el == NULL) return 0;
  return el->region;
}
TINE_EXPORT char *GetExportedSubSystem(char *eqmName)
{
  ExportListStruct *el = getExportListItem(eqmName);
  char *sub = el == NULL ? gDeviceSubSystem : el->EqmSubSystem;
  static char strBuffer[SUBSYSTEM_NAME_SIZE];
  UNUSED(eqmName); /* avoid warning until this is used */
  strncpy(strBuffer,sub,SUBSYSTEM_NAME_SIZE);
  return strBuffer;
}
TINE_EXPORT void SetExportedSubSystem(char *eqmName,char *subsystem)
{
  ExportListStruct *el = getExportListItem(eqmName);
  if (subsystem == NULL || strlen(subsystem) == 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "SetExportedSubSystem: invalid argument");
    return;
  }
  if (el == NULL)
  {
    strncpy(gDeviceSubSystem,subsystem,SUBSYSTEM_NAME_SIZE);
    return;
  }
  strncpy(el->EqmSubSystem,subsystem,SUBSYSTEM_NAME_SIZE);
  return;
}
TINE_EXPORT char *GetExportedFecName(void)
{
  static char fecNameBuffer[32];
  strncpy(fecNameBuffer,gFecName,FEC_NAME_SIZE);
  return fecNameBuffer;
}
TINE_EXPORT char *SystemGetStartupDirectory(void)
{
  static char buf[256];
  static int done=FALSE;
  if (done) return buf;
# if defined(WIN32) || defined(UNIX) || defined(VXWORKS)
  if (getcwd(buf,256) == NULL) strcpy(buf,"unavailable");
# else
  strcpy(buf,"unknown");
# endif
  done = TRUE;
  return buf;
}
TINE_EXPORT UINT32 SystemGetProcessId(void)
{
# if defined(WIN32)
  return (UINT32)GetCurrentProcessId();
# elif defined(UNIX) || defined(MACOS)
  return (UINT32)getpid();
# elif defined(VXWORKS)
  return (UINT32)taskIdSelf();
# else
  return 0;
# endif
}
TINE_EXPORT char *SystemGetStartupCommand(void)
{
# if defined(WIN32)
  return GetCommandLine();
# elif defined(SOLARIS)
  static int done = FALSE;
  int i, hp;
  static char buf[512];
  pid_t p = getpid();
  struct psinfo ps;
  if (done) return buf;
  sprintf(buf,"/proc/%d/psinfo",p);
  if ((hp=open(buf,O_RDONLY)) == -1) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot open proc info");
    return NULL;
  }
  if (read(hp,&ps,sizeof(struct psinfo)) == -1) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot read command line information");
    close(hp);
    return NULL;
  }
  memset(buf,0,512);
  for (i=0; i<ps.pr_argc; i++)
  {
    sprintf(&buf[strlen(buf)],"%s ",((char **)ps.pr_argv)[i]);
  }
  close(hp);
  done = TRUE;
  return buf;
# elif defined(MACOS)
  static int done = FALSE;
  static char buf[512];
  FILE *pfp = NULL;
  char s[64];
  if (done) return buf;
  memset(buf,0,512);
  sprintf(s,"/bin/ps -p %d -o args=",getpid());
  if ((pfp=popen(s,"r")) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot open pipe to /bin/ps");
    return NULL;
  }
  fgets(buf,255,pfp);
  pclose(pfp);
  done = TRUE;
  return buf;
# elif defined(UNIX)
  static int done = FALSE;
  int hp;
  static char buf[512];
  char *c;
  if (done) return buf;
  sprintf(buf,"/proc/self/cmdline");
  if ((hp=open(buf,O_RDONLY)) == -1) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot open proc info");
    return NULL;
  }
  memset(buf,0,512);
  if (read(hp,buf,512) == -1) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot read command line information");
    close(hp);
    return NULL;
  }
  for (c=buf; *c != 0; c += strlen(c)+1)
  { /* replace '0' with blank */
    buf[strlen(buf)] = ' ';
  }
  close(hp);
  done = TRUE;
  return buf;
# elif defined(VXWORKS)
  static int done = FALSE;
  int tskid;
  char *tskname;
  static char buf[512];
  if (done) return buf;
  tskid = taskIdSelf();
  tskname = taskName(tskid);
  strncpy(buf,tskname,512);
  return buf;
# else
  return NULL;
# endif
}
UINT32 getAvailableMemoryBytes(void)
{
# if defined(WIN32)
# ifdef SINT64
  SINT64 amem;
# else
  UINT32 amem;
# endif
# ifdef VS_LEGACY_BUILD
  MEMORYSTATUS stat;
  GlobalMemoryStatus(&stat);
  if ((amem=stat.dwAvailPhys) > 0xffffffff || amem < 0) amem = 0xffffffff;
# else
  MEMORYSTATUSEX statx;
  statx.dwLength = sizeof(statx);
  GlobalMemoryStatusEx(&statx);
  if ((amem=statx.ullAvailPhys) > 0xffffffff || amem < 0) amem = 0xffffffff;
# endif
  return (UINT32)amem;
# elif defined(MACOS)
  int mib[2];
  size_t len;
  UINT64 amem;
  mib[0] = CTL_HW;
  mib[1] = HW_MEMSIZE;
  len = sizeof(amem);
  sysctl(mib, 2, &amem, &len, NULL, 0);
  if (amem > 0xffffffff) amem = 0xffffffff;
  return (UINT32)amem;
# elif defined(UNIX)
  long pages = sysconf(_SC_PHYS_PAGES);
  long page_size = sysconf(_SC_PAGE_SIZE);
# ifdef SINT64
  SINT64 amem = (SINT64)pages * (SINT64)page_size;
# else
  long amem = pages * page_size;
# endif
  if (amem > 0xffffffff || amem < 0) amem = 0xffffffff;
  return (UINT32)amem;
# elif defined(VXWORKS)
  int amem = memFindMax();
  if (amem < 0) amem = 0x1000000;
  return (UINT32)amem;
# else /* punt */
  return (UINT32)0x1000000;
# endif
}
int gAutoAdjustWorkAreaSize = TRUE;
TINE_EXPORT void SetAutoAdjustWorkAreaSize(int value)
{
  gAutoAdjustWorkAreaSize = value;
}
int adjustWorkAreaSize(UINT32 newSizeInBytes)
{
  int cc = 0;
  BYTE *m;
  UINT32 memAvail = getAvailableMemoryBytes();
  if (newSizeInBytes > memAvail/2)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "requested work area size %u bytes exceeds half the available memory %u !",newSizeInBytes,memAvail);
    newSizeInBytes = memAvail/2;
  }
  if (WaitForSystemMutex(hSystemServerMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return mutex_error;
  if ((m=(BYTE *)realloc(srvWorkArea,(size_t)newSizeInBytes)) == NULL) ccerr(out_of_server_memory);
  srvWorkArea = m;
  srvWorkAreaSize = MaxSystemTransportSize = newSizeInBytes;
# if defined(INTEL) && !defined(VXORKS) && !defined(MACOS)
  tmpWorkArea = srvWorkArea;
  tmpWorkAreaSize = srvWorkAreaSize;
# endif
err:
  ReleaseSystemMutex(hSystemServerMutex);
  feclog("Work Space re-allocation to %u bytes: %.32s",newSizeInBytes,cc2str(cc));
  return cc;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT void _SystemFreeResources(int exitOnFree)
{
  if (!gSystemMutexSetInitialized) return; /* not applicable */
  if (nConnectionTableEntries || nglobals) CloseClients(TRUE);
  if (WaitForSystemMutex(hSystemKernelMutex,-1) != 0) return;
  CloseServers(TRUE);
  ExitEQPModules();
# if !defined(DOS) && !defined(NIOS)
  freeServerMemory();
# endif
# if defined(WIN32)
# if !defined(NO_SERVICE_PIPE)
  if (ipcCmdSck != 0) CloseHandle(((PIPEINST *)ipcCmdSck)->hPipeInst);
  ipcCmdSck = 0;
# endif
  WSACleanup();
  WSAinitialized = -1;
  feclog("Windows Socket Cleanup initiated");
# endif
# ifdef UNIX
  CloseIpcPipe(NULL);
# endif
  ReleaseSystemMutex(hSystemKernelMutex);
  if (exitOnFree) exit(gSystemExitCode);
}
TINE_EXPORT void haltServer(void)
{
  _SystemStop(TRUE);
}
void ParseRedirector(char *str,PrpRedirBlk *prd)
{
  char *c = str, *cstr = NULL, *sstr = NULL, *pstr = str, *dstr = NULL, *cp = NULL;

  memset(prd->prpRedirection,0,STATUS_SIZE);
  if (str == NULL) return;
  if (str[0] == '/')
  { /* context given ? */
    *c++ = 0; cstr = c;
    if ((c=strchr(cstr,'/')) == NULL) return; /* device server must be there !*/
    *c++ = 0;
  }
  sstr = c; /* device server */
  if ((c=strchr(sstr,'/')) != NULL)
  { /* device name given ? */
    char *cd = NULL;
    *c++ = 0; dstr = c;
    if ((cd=strstr(dstr,RDR_MATCHANYDEVICE)) != NULL)
    { /* instructions to match any device in a check */
      prd->matchAnyDevice = TRUE;
      *cd++ = 0; /* removed this decoration */
      c = cd; /* advance the str pointer */
    }
  }
  else
  { /* no, put the pointer back */
    c = sstr;
  }
  if ((pstr=strtok_r(c,"[",&cp)) != NULL) pstr = strtok_r(NULL,"]",&cp);
  if (cstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_CTX_OFFSET],cstr,CONTEXT_NAME_SIZE);
  if (sstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_SRV_OFFSET],sstr,EXPORT_NAME_SIZE);
  if (dstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_DEV_OFFSET],dstr,DEVICE_NAME_SIZE);
  if (pstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_PRP_OFFSET],pstr,PROPERTY_NAME_SIZE);
}
void parseEgu(PrpEgu *egu,char *str)
{
  char *c;

  if ((c=strstr(str,"!")) != NULL)
  { /* graph type */
    *c++ = 0;
    egu->graph = getGraphType(c);
  }
  if ((c=strchr(str,':')) != NULL)
  { /* prp min:max settings given */
    egu->min = (float)strtod(str,NULL); c++;
    egu->max = (float)strtod(c,&c); 
  }
  else
  { /* just the units */
    c = str;
  }
  while (*c == ' ') c++;
  if (*c != 0)
  {
    strncpy(egu->units,c,UNITS_SIZE);
    ensureAsciiCharacters(egu->units,UNITS_SIZE);
  }
}
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp)
{
  char *c=NULL, *c1=NULL, *c2=NULL, *tok[10];
  char *hsstr=NULL, *vsstr=NULL, *hplstr=NULL, *vplstr=NULL, *urlstr=NULL, *dscstr=NULL;
  int i, n = 0, len;
  if (str == NULL || prp == NULL) return;

  for (c=strtok(str,"]"); c != NULL && n < 10; c=strtok(NULL,"]"))
  { /* tokenize */
    if (*c != '[')
    { /* non-bracketed token ? */
      if (dscstr == NULL) dscstr = c; /* old-fashioned, but okay */
      continue; /* don't collect these */
    }
    c++; /* skip over the bracket */
    if (strlen(c) > 0) tok[n++] = c;
  }
  for (i=0; i<n; i++)
  { /* scan thru the tokens and make assignments */
    if (!strnicmp(tok[i],"vscale",6) || !strnicmp(tok[i],"scale",5))
    { /* tagged token for vertical scale */
      vsstr = strchr(tok[i],'='); vsstr++;
    }
    else if (!strnicmp(tok[i],"hscale",6) || !strnicmp(tok[i],"range",5))
    { /* tagged token for horizontal scale */
      hsstr = strchr(tok[i],'='); hsstr++;
    }
    else if (!strnicmp(tok[i],"url",3))
    { /* tagged token for url reference */
      urlstr = strchr(tok[i],'='); urlstr++;
    }
    else if (!strnicmp(tok[i],"desc",4))
    { /* tagged token for description */
      dscstr = strchr(tok[i],'='); dscstr++;
    }
    else if (!strnicmp(tok[i],"vplot",5) || !strnicmp(tok[i],"plot",4))
    { /* tagged token for vertical plot style */
      vplstr = strchr(tok[i],'='); vplstr++;
    }
    else if (!strnicmp(tok[i],"hplot",5))
    { /* tagged token for horizontal plot style */
      hplstr = strchr(tok[i],'='); hplstr++;
    }
    else if (c1 == NULL)
    { /* old-fashioned initial token = vertical scale */
      c1 = tok[i]; 
    }
    else if (c2 == NULL)
    { /* old-fashioned second token = horizontal scale */
      c2 = tok[i]; 
    }
  }  
  if (vsstr == NULL) vsstr = c1;
  if (hsstr == NULL) hsstr = c2;

  if (vsstr != NULL)
  {
    parseEgu(&prp->egu,vsstr);
  }
  if (vplstr != NULL)
  { /* plot type given separately */
    prp->egu.graph = getGraphType(vplstr);
  }
  if (hsstr != NULL)
  {
    parseEgu(&prp->xegu,hsstr);
  }
  if (hplstr != NULL)
  { /* plot type given separately */
    prp->xegu.graph = getGraphType(hplstr);
  }
  if (urlstr != NULL)
  {
    len = (int)strlen(urlstr);
    prp->prpUrl = (char *)SystemCalloc(len+1,1);
    if (prp->prpUrl != NULL) strncpy(prp->prpUrl,urlstr,len);
  }
  if (dscstr != NULL)
  {
    strncpy(prp->prpDescription,dscstr,PROPERTY_DESC_SIZE);
    ensureAsciiCharacters(prp->prpDescription,PROPERTY_DESC_SIZE);
  }
}
void makeEqpDevicesHashTable(ExportListStruct *el)
{
  HashXRefStruct *xref;
  char devname[DEVICE_NAME_SIZE+4];
  int cc=0,idx,i;

  if (el == NULL) return;
  if (el->EqmNumDevices < DEVICES_HASH_PAYOFF) return;
  if ((el->EqmDevXRef == NULL) &&
      (el->EqmDevXRef=(HashXRefStruct **)SystemCalloc(DEVICES_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return;
  if (el->EqmDevLst == NULL) return;
  devname[DEVICE_NAME_SIZE] = 0;
  for (i=0; i<el->EqmNumDevices; i++)
  {
    strncpy(devname,el->EqmDevLst[i].name,DEVICE_NAME_SIZE);
    if (strlen(devname) == 0) continue;
    if (el->EqmDevLst[i].isReserved) continue;
    idx = (int)(ElfHash((unsigned char *)devname) % DEVICES_HASH_SIZE);
    for (xref=el->EqmDevXRef[idx]; xref && xref->index != i; xref=xref->next);
    if (xref == NULL) /* it's a new one*/
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
      xref->index = i;
      xref->next = el->EqmDevXRef[idx];
      el->EqmDevXRef[idx] = xref;
    }
  }
err:
  if (cc) feclog("makeEqpDevicesHashTable: %.32s\n",cc2str(cc));
  return;
}

_CSVCOLUMN csvFecNameColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,fecName),NULL,0 },
  { "OS",OS_TAG,0,CF_TEXT,16,(BYTE *)OFFSETIN(FIDS,fecOs),NULL,0 },
  { "PORT_OFFSET","0",0,CF_INT32,1,(BYTE *)OFFSETIN(FIDS,fecPort),NULL,0 },
  { "CONTEXT","TEST",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,fecContext),NULL,0 },
  { "EXPORT_NAME","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,expNameTarget),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,eqmNameTarget),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,FEC_DESC_SIZE,(BYTE *)OFFSETIN(FIDS,fecDescription),NULL,0 },
  { "LOCATION","",0,CF_TEXT,FEC_LOCATION_SIZE,(BYTE *)OFFSETIN(FIDS,fecLocation),NULL,0 },
  { "HARDWARE","none",0,CF_TEXT,FEC_HDW_SIZE,(BYTE *)OFFSETIN(FIDS,fecHardware),NULL,0 },
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,fecSubsystem),NULL,0 },
  { "RESPONSIBLE","",0,CF_TEXT,FEC_RESP_SIZE,(BYTE *)OFFSETIN(FIDS,responsible),NULL,0 },
  { "HISTORY_HOME","",0,CF_TEXT,HISTORY_HOME_SIZE,(BYTE *)OFFSETIN(FIDS,historyHome),NULL,0 },
  { "MASTER","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,master),NULL,0 },
  { "SLAVE_MASTER","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,slaveMaster),NULL,0 },  
  { "MTU","0",0,CF_INT32,1,(BYTE *)OFFSETIN(FIDS,fecMtu),NULL,0 },
  { "BURST_LIMIT","0",0,CF_INT32,1,(BYTE *)OFFSETIN(FIDS,fecBurstLimit),NULL,0 },
  { "CYCLE_DELAY","0",0,CF_INT32,1,(BYTE *)OFFSETIN(FIDS,fecCycleDelay),NULL,0 },
  { "WORKAREA_SIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(FIDS,fecWorkAreaSize),NULL,0 },
};
CSVDB csvFecNameDb =
{
  csvFecNameColumns,
  sizeof(csvFecNameColumns)/sizeof(csvFecNameColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(FIDS),
  0
};
#ifndef SMALL_TINE_LIB
int getSrvAddress(DTYPE *sa,char *eqm,char *devsrv)
{ /* response to a 'network' address resolution query */
  ExportListStruct *el=NULL;
  int len;
  char lDeviceServer[EXPORT_NAME_SIZE+4], lEqmName[EQM_NAME_SIZE], lSubName[SUBSYSTEM_NAME_SIZE+4], lCxtName[CONTEXT_NAME_SIZE+4];
  if (sa == NULL) return argument_list_error;
  switch (sa->dFormat)
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      len = fmtsizeof(sa->dFormat);
      break;
    default:
      return illegal_format;
  }
  if (eqm != NULL && strlen(eqm) > 0)
  {
    for (el=ExportList; el != NULL; el=el->next) 
      if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  }
  else if (devsrv != NULL && strlen(devsrv) > 0)
  {
    for (el=ExportList; el != NULL; el=el->next) 
      if (!strnicmp(devsrv,el->EqmExportName,EXPORT_NAME_SIZE)) break;
  }
  lSubName[0] = lDeviceServer[0] = lEqmName[0] = lCxtName[0] = 0;
  if (el != NULL)
  {
    strncpy(lDeviceServer,el->EqmExportName,EXPORT_NAME_SIZE);
    memset(lEqmName,0,EQM_NAME_SIZE);
    strncpy(lEqmName,el->EqmName,EQM_NAME_SHORTSIZE);
    strncpy(lSubName,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);
    strncpy(lCxtName,el->EqmContext,CONTEXT_NAME_SIZE);
  }
  switch (sa->dArrayLength)
  {
    case 7: strcpy(sa->data.cptr + (size_t)6*len,(gMultiCastUseLegacyRules ? "LEGACY" : "MODERN"));
    case 6: strncpy(sa->data.cptr + (size_t)5*len,lSubName,SUBSYSTEM_NAME_SIZE);
    case 5: strncpy(sa->data.cptr + (size_t)4*len,lDeviceServer,MIN(EXPORT_NAME_SIZE,len));
    case 4: strncpy(sa->data.cptr + (size_t)3*len,lEqmName,EQM_NAME_SHORTSIZE);
    case 3: strncpy(sa->data.cptr + (size_t)2*len,lCxtName,MIN(CONTEXT_NAME_SIZE,len));
    case 2: strncpy(sa->data.cptr + len,gFecName,FEC_NAME_SIZE);
    case 1: sprintf(sa->data.cptr,"%d",gPortOffset);
      break;
    default: 
      return dimension_error;
  }
  return 0;
}
#endif
int validatePortOffset(int portOffset)
{
# ifndef DOS
  int p = portOffset;
  if (p < 0 || p > 55500)
  {
    sprintf(dbgbuf, "invalid port offset %d !", portOffset);
    p = 0;
    feclogEx(TINE_LOGLEVEL_FATAL, dbgbuf);
    if (dieFunction != NULL) dieFunction(dbgbuf);
    if (gDieAnotherDay) { gDieAnotherDay = FALSE; return portOffset; }
    dbglog(dbgbuf);
    feclogEx(TINE_LOGLEVEL_FATAL, "Aborting server process! (fix this problem and try again!)");
    exit(1);
  }
  if (gtUDPPort + p == gtMCastPort) 
  {
    p++;
    feclogEx(TINE_LOGLEVEL_WARN, "port offset %d collides with multicast port, corrected to %d",portOffset,p);
  }
  if (gtTCPPort + p == gtMCastPort) 
  {
    p++;
    feclogEx(TINE_LOGLEVEL_WARN, "port offset %d collides with multicast port, corrected to %d",portOffset,p);
  }
  return p;
# else
  return 0;
# endif
}
DBLNKLIST *getFecInfoList(void)
{
# ifndef SMALL_TINE_LIB
  int cc;
  if (gFecInfoList != NULL || fecXmlDone) return gFecInfoList;
  fecXmlDone = TRUE;
  gFecInfoList = XMLListCreate();
  if ((cc=XMLReadFile(fecDbPath,FECXML_FILE,gFecInfoList)) == 0)
  {
    isUsingXMLFecDatabase = TRUE;
    return gFecInfoList;
  }
  SystemFree(gFecInfoList);
# endif
  return NULL;
}
int assertNameIsValid(const char *name,const char *badchars,int length)
{
  char str[128];
  int i;
  if (name == NULL) return FALSE;
  if (length > 127) length = 127;
  strncpy(str,name,length);
  str[length] = 0;
  for (i = 0; i < length; i++) if ((BYTE)str[i] > 127) return FALSE;
  return (strpbrk(str,badchars) == NULL) ? TRUE : FALSE;

}
int getFecName(void)
{
  char *c,scratch[32];
  int i,siz=0,cc=0;
  FIDS *fids = NULL;
  ExportListStruct *el = NULL;

  if (FecNameRegistered) return 0;
  gLastMinuteInitializationDone = FALSE;
# ifndef SMALL_TINE_LIB
  if ((gFecInfoList=getFecInfoList()) != NULL)
  {
    feclog("scanning fec.xml for FEC name");
    siz = populateFIDS(gFecInfoList,&fids);
  }
  else
# endif
  {
    feclog("scanning %s for FEC name",FECID_FILE);
    if ((cc=csvReadFile(fecDbPath,FECID_FILE,&csvFecNameDb,(void **)&fids)) != 0) ccerr(cc);
    siz = csvFecNameDb.siz;
  }
  cc = name_not_found;
  for (i=0; i<siz; i++)
  {
    if (strlen(fids[i].expNameTarget) > 0)
    { /* target export name in fecid.csv file -> find a match */
      for (el=ExportList; el != NULL; el=el->next)
      {
        if (strlen(el->EqmContext) > 0 && strlen(fids[i].fecContext) > 0)
        { /* if both contexts known and I'm looking for a match then ... */
          if (strnicmp(fids[i].fecContext,el->EqmContext,CONTEXT_NAME_SIZE)) continue;
        }
        if (!strnicmp(fids[i].expNameTarget,el->EqmExportName,EXPORT_NAME_SIZE)) break;
      }
      if (el == NULL) continue;
      feclog("export entry %.32s found!",el->EqmExportName);
    }
    if (strlen(fids[i].eqmNameTarget) > 0)
    { /* target eqm (local) name in fecid.csv file -> find a match */
      for (el=ExportList; el != NULL; el=el->next)
      {
        if (!strnicmp(fids[i].eqmNameTarget,el->EqmName,EQM_NAME_SIZE)) break;
      }
      if (el == NULL) continue;
      feclog("local name entry %.6s found!",el->EqmName);
    }
    if (!assertNameIsValid(fids[i].fecName,RUDECHARS,FEC_NAME_SIZE))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "FEC name %.16s contains invalid characters!",fids[i].fecName);
      cc = invalid_name;
      continue;
    }
    strncpy(gFecName,fids[i].fecName,FEC_NAME_SIZE);
    strncpy(gDeviceContext,fids[i].fecContext,CONTEXT_NAME_SIZE);
    strncpy(gDeviceSubSystem,fids[i].fecSubsystem,SUBSYSTEM_NAME_SIZE);
    memset(&gFecInfo, 0, sizeof(gFecInfo));
    strncpy(gFecInfo.os,OS_TAG,FEC_OS_SIZE);
    strncpy(gFecInfo.desc,fids[i].fecDescription,FEC_DESC_SIZE);
    strncpy(gFecInfo.loc,fids[i].fecLocation,FEC_LOCATION_SIZE);
    strncpy(gFecInfo.hdw,fids[i].fecHardware,FEC_HDW_SIZE);
    strncpy(gFecInfo.resp,fids[i].responsible,FEC_RESP_SIZE);
    strncpy(gFecInfo.ver,GetSystemVersionString(),FEC_VERSION_SIZE);
    gPortOffset = validatePortOffset(fids[i].fecPort);
    SetDoocsUserName(whoami(NULL));
    strncpy(gUserName,gFecName,USERNAME_SIZE);
    SetUserType("FEC");
    if (strlen(fids[i].historyHome) > 0)
    {
      history_home_done = TRUE;
      strncpy(arcDbPath,fids[i].historyHome,HISTORY_HOME_SIZE);
      validateHistoryHomePath();
      feclog("HISTORY HOME repository %s set from FEC configuration file",arcDbPath);
    }
    if (fids[i].fecMtu > 0) SetPacketMTU(fids[i].fecMtu);
    if (fids[i].fecBurstLimit > 0) SetBurstLimit(fids[i].fecBurstLimit);
    if (fids[i].fecCycleDelay > 0) SetCycleDelay(fids[i].fecCycleDelay);
    if (fids[i].fecWorkAreaSize > 0) 
    {
#     ifdef MULTITHREADED
      SetServerTransportCeiling(fids[i].fecWorkAreaSize);
#     else
      SetWorkAreaSize(fids[i].fecWorkAreaSize);
#     endif
    }
    if( strlen(fids[i].master)>0) {
      /*possibly no eqms registered yet, saving values for later in static vars within setFailoverConfigInExportList*/
      setFailoverConfigInExportList(NULL, fids[i].master, fids[i].slaveMaster);
    }
    cc = 0;
    FecNameRegistered = TRUE;
    break;
  }
  if (!FecNameRegistered)
  { /* reset these if no match found ... */
    strcpy(gFecName,"*unknown*");
    gDeviceContext[0] = 0;
    feclog("%.128s%s : %.32s",fecDbPath,FECID_FILE,cc2str(cc));
    feclog("currently unable to register as fec name is unknown!");
    ccerr(name_unknown);
  }
  feclog("accepting %.16s as FEC name",gFecName);
  if (strlen(gDeviceSubSystem) == 0)
  { /* subsystem not explicitly given; is it to be parsed from the description ? */
    if ((c=strchr(gFecInfo.desc,'[')) != 0)
    {
      strncpy(scratch,++c,20);
      if ((c=strchr(scratch,']')) != 0) *c=0;
      strncpy(gDeviceSubSystem,scratch,16);
      feclog("assigning FEC subsystem %.16s from description",gDeviceSubSystem);
    }
  }
err:
  SystemFree(fids);
  return cc;
}
void checkAccessLocks(time_t t)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (!el->accessLock.lockType) continue; /* no active lock (normal case) */
    if (el->accessLock.lockDuration < 0) continue;  /* forever */
    if (t > (time_t)(el->accessLock.lockStart + el->accessLock.lockDuration))
      memset(&el->accessLock,0,sizeof(AccessLock)); /* reset everything */
  }
}
int isMemberLockSet(ExportListStruct *el,ClnHdr *cln)
{
  if (!el->accessLock.lockType) return TRUE; /* unlocked */
  if (strnicmp(el->accessLock.cln.usr,cln->usr,USERNAME_SIZE)) return FALSE;
  if (sckadrcmp(&el->accessLock.cln.adr.ip,&cln->adr.ip)) return FALSE;
  return TRUE;
}
int isExclusiveProperty(ExportListStruct *el,char *prp)
{
  ExportPropertyListStruct *p;
  if (IsStockProperty(prp)) return FALSE;
  if ((p=GetPropertyListStruct(el->EqmName,prp,NULL)) == NULL) return FALSE;
  return (int)p->hasExclusiveRead;
}
int CheckDeniedList = FALSE;
int MaxNumAllowedUsers = MAX_NUM_USERS;
AclInfoType gDeniedUsersList;
int nUsrGrps = 0;
int MaxNumUserGroups = MAX_NUM_USERGROUPS;
UsrGrpsLst *gUsrGrps = NULL;
UsrGrpsLst *getUserGroupFromCache(char *grp)
{
  UsrGrpsLst *g=NULL;
  for (g=gUsrGrps; g != NULL; g=g->nxt)
  { 
    if (!strnicmp(grp,g->name,GROUPNAME_SIZE)) break;
  }
  return g;
}
int addUserGroupToList(AclInfoType *acl,char *fn,char *grp,char *src)
{ /* used internally: acl never NULL */
  UsrGrpsLst *g=getUserGroupFromCache(grp);
  int i=acl->nGrps,cc=0;
  char *psrc = src;
  if (acl->grps == NULL)
  { /* not yet allocated -> assume 'worst-case' size */
    if ((acl->grps=(NAME16 *)SystemCalloc(acl->nUsrs,sizeof(NAME16))) == NULL)
      return out_of_server_memory;
  }
  if (g != NULL) ccerr(0); /* group already known */
  if ((g=(UsrGrpsLst *)SystemCalloc(1,sizeof(UsrGrpsLst))) == NULL) return out_of_server_memory;
  if (src == NULL || strlen(src) == 0) psrc = NULL;
  strncpy(g->name,grp,GROUPNAME_SIZE);
  cc = GetUserGroupMembers(fn,grp,psrc,&g->members,&g->nmembers);
  if (cc == 0) 
  { /* no problems */
    if (psrc == NULL) psrc = "registry";
    strncpy(g->type,psrc,GROUPSOURCE_SIZE);
    g->nxt = gUsrGrps;
    gUsrGrps = g;
    nUsrGrps++; 
  }
  else 
  { /* error: free the memory here */
    SystemFree(g);
  }
err:
  if (cc == 0)
  {
    strncpy(acl->grps[i].name,grp,GROUPNAME_SIZE); 
    acl->nGrps++;
  }
  return cc;
}
void checkGroupMembers(time_t t)
{
  static time_t last_checked = 0;
  ExportListStruct *el;
  UsrGrpsLst *g;
  int i, nm;
  if (last_checked == 0) last_checked = t; /* skip this the first time */
  if (t < last_checked + 3600) return;
  last_checked = t;
  for (el=ExportList; el != NULL; el=el->next)
  { /* run thru list of equipment modules ... */
    for (i=0; i<el->aclLst.nGrps; i++)
    { /* run thru list of known user groups */
      if ((g=getUserGroupFromCache(el->aclLst.grps[i].name)) != NULL)
      { /* should always be true */
        nm = g->nmembers;
        GetUserGroupMembers(el->usersFileName,g->name,g->type,&g->members,&g->nmembers);
        if (nm != g->nmembers) feclog("update group member list for %.16s",g->name);
      }
    }
  }
}
int isDeniedUser(char *usr)
{
  int i;
  for (i=0; i<gDeniedUsersList.nUsrs && strnicmp(usr,gDeniedUsersList.usrs[i].name,USERNAME_SIZE); i++);
  return i == gDeniedUsersList.nUsrs ? 0 : -1;
}
int isAllowedUser(char *eqm,char *usr)
{
  ExportListStruct *el = getExportListItem(eqm);
  UsrGrpsLst *g;
  int i, m;
  static char lastAllowedUser[USERNAME_SIZE+4];
  if (el == NULL) return TRUE;
  if (!strnicmp(usr,lastAllowedUser,USERNAME_SIZE)) return TRUE;
  for (i=0; i<el->aclLst.nUsrs; i++)
  { /* is it worth it to use a hash table ? */
    if (!strnicmp(usr,el->aclLst.usrs[i].name,USERNAME_SIZE)) 
    { /* help speed up next user's check */
      strncpy(lastAllowedUser,usr,USERNAME_SIZE);
      return TRUE;
    }
  }
  for (i=0; i<el->aclLst.nGrps; i++)
  { /* run thru list of known user groups */
    if ((g=getUserGroupFromCache(el->aclLst.grps[i].name)) != NULL)
    { /* should always be true */
      for (m=0; m<g->nmembers; m++)
      { /* run thru members list */
        if (!strnicmp(usr,g->members[m].name,USERNAME_SIZE)) 
        { /* help speed up next user's check */
          strncpy(lastAllowedUser,usr,USERNAME_SIZE);
          return TRUE;
        }
      }
    }
  }
  return FALSE;
}
TINE_EXPORT int RemoveRegisteredUser(char *eqm,NAME16 *userlist,int listsize)
{
  return rmvRegUserLst(eqm,userlist,listsize);
}
int rmvRegUserLst(char *eqm,NAME16 *userlist,int listsize)
{
  ExportListStruct *el = getExportListItem(eqm);
# if !defined(FS_RDONLY) && !defined(FS_NONE)
  FILE *fp = NULL, *tmpfp = NULL;
  int i,n=0,len,cc=0,acc=0;
  char *c,str[256],userfilename[256],tmpfilename[256];

  if (el == NULL) ccerr(non_existent_elem);
  if (userlist == NULL || listsize < 1) ccerr(argument_list_error);
  if (el->aclLst.nUsrs == 0) ccerr(0);
  if (isUsingXMLFecDatabase) ccerr(not_implemented);
  memset(&gLastAllowedCln,0,sizeof(ClnHdr)); /* reset this */
  for (i=0; i<listsize; i++) strntrm(userlist[i].name,16);
  for (i=0; i<listsize; i++)
  {
    if (!isAllowedUser(eqm,userlist[i].name)) userlist[i].name[0] = 0;
    if (strlen(userlist[i].name) > 0) n++;
  }
  if (n == 0) ccerr(0); /* users weren't in the list anyway ! */

  sprintf(userfilename,"%.128s%.32s",fecDbPath,el->usersFileName);
  strncpy(tmpfilename,userfilename,128); tmpfilename[128] = 0;
  if ((c=strstr(tmpfilename,".csv")) == NULL) ccerr(file_error);
  *c = 0; strcat(tmpfilename,".bak");
  unlink(tmpfilename);
  rename(userfilename,tmpfilename);

  if ((tmpfp=fopen(tmpfilename,"r")) == NULL) ccerr(no_such_file);
  if ((fp=fopen(userfilename,"w+")) == NULL) ccerr(no_such_file);

  while (fgets(str,255,tmpfp) != NULL)
  {
    if ((len=(int)strlen(str)-1) <= 0) continue;
    for (i=0; i<listsize; i++) if (!strnicmp(str,userlist[i].name,len)) break;
    if (i < listsize) continue; /* skip, i.e. remove, this one */
    fwrite(str,strlen(str),1,fp);
  }
  fclose(fp); fp = NULL;
  fclose(tmpfp); tmpfp = NULL;
  el->chkAcl &= ~(EQM_CHECK_USR|EQM_CHECK_GRP);
  getRegUsrsFromFile(eqm,NULL,NULL,&el->aclLst,el->usersFileName,"WRITE",&acc);
  el->chkAcl |= acc;
err:
  if (fp != NULL) fclose(fp);
  if (tmpfp != NULL) fclose(tmpfp);
  return cc;
# else
  int i, n;
  if (userlist == NULL || listsize == 0) return argument_list_error;
  for (i=0; i<listsize; i++) strntrm(userlist[i].name,16);
  for (i=0; i<el->aclLst.nUsrs; i++) 
  { /* clear those entries to be removed */
    for (n=0; n<listsize; n++) if (!strnicmp(el->aclLst.usrs[i].name,userlist[n].name,USERNAME_SIZE)) break;
    if (n < listsize) el->aclLst.usrs[i].name[0] = 0; /* clear this one */
  }
  for (i=0; i<el->aclLst.nUsrs; i++) 
  { /* now collapse the list */
    if (strlen(el->aclLst.usrs[i].name) == 0)
    {
      el->aclLst.nUsrs--;
      if (i < el->aclLst.nUsrs)
      {
        strncpy(el->aclLst.usrs[i].name,el->aclLst.usrs[i+1].name,USERNAME_SIZE);
        i--; 
      }
    }
  }
  return 0;
# endif
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int AppendRegisteredNetsList(const char *eqm,NAME64 *ipaddr,int addsiz)
{
  ExportListStruct *el = getExportListItem((char *)eqm);
  if (el == NULL) return non_existent_elem;
  return addRegNetsLst(el->netsFileName,NETS_ACCESS,&el->aclLst.nNets,&el->aclLst.nets,ipaddr,addsiz);
}
TINE_EXPORT int AppendRegisteredBCastNetsList(NAME64 *ipaddr,int addsiz)
{
  return addRegNetsLst(NULL,NETS_NETWORK,&sizIpBcstLst,&IpBcstLst,ipaddr,addsiz);
}
TINE_EXPORT int RemoveRegisteredNetsList(const char * eqm,NAME64 *ipaddr,int rmvsiz)
{
  ExportListStruct *el = getExportListItem((char *)eqm);
  if (el == NULL) return non_existent_elem;
  return rmvRegNetsLst(el->netsFileName,NETS_ACCESS,&el->aclLst.nNets,&el->aclLst.nets,ipaddr,rmvsiz);
}
TINE_EXPORT int RemoveRegisteredBCastNetsList(NAME64 *ipaddr,int rmvsiz)
{
  return rmvRegNetsLst(NULL,NETS_NETWORK,&sizIpBcstLst,&IpBcstLst,ipaddr,rmvsiz);
}
TINE_EXPORT int AppendRegisteredUserList(NAME16 *userlist,int listsize)
{
  UNUSED(userlist); 
  UNUSED(listsize);
  printf("this routine has been deprecated; please use AppendRegisteredUsers()\n");
  return function_deprecated;
}
TINE_EXPORT int GetRegisteredNetworks(const char *eqm, NAME64 *list, int *listsize)
{
  int i;
  ExportListStruct* el = getExportListItem((char*)eqm);
  SCKACL *lst;
  char ipstr[64];
  if (el == NULL) return non_existent_elem;
  if (list == NULL || listsize == NULL || *listsize < 1) return argument_list_error;
  lst = el->aclLst.nets;
  for (i = 0; i < *listsize && i < el->aclLst.nNets; i++)
  {
    inetStrAdr(&lst[i].addr, ipstr, FALSE);
    addIpSubStr(ipstr, lst[i]);
    strncpy(list[i].name, ipstr, 64);
  }
  *listsize = el->aclLst.nNets;
  return 0;
}
#endif /* !SMALL_TINE_LIB */
void dumpUsersToFile(NAME16 *usrlst,int lstsiz,FILE *fp)
{
  int i;
  for (i=0; i<lstsiz; i++)
  {
    if (strlen(usrlst[i].name) == 0) continue;
    fprintf(fp,"%.16s\n",usrlst[i].name);
  }
}
TINE_EXPORT int GetRegisteredUsers(const char *eqm,NAME16 *userlist,int *listsize)
{
  int i;
  ExportListStruct *el=getExportListItem((char *)eqm);
  if (el == NULL) return non_existent_elem;
  if (userlist == NULL || listsize == NULL || *listsize < 1) return argument_list_error;
  for (i=0; i<*listsize && i<el->aclLst.nUsrs; i++)
  {
    strncpy(userlist[i].name,el->aclLst.usrs[i].name,USERNAME_SIZE);
  }
  *listsize = el->aclLst.nUsrs;
  return 0;
}
TINE_EXPORT int AppendRegisteredUsers(const char *eqm,NAME16 *userlist,int listsize)
{
  int cc=0, i, n=0;
  ExportListStruct *el=getExportListItem((char *)eqm);
# if !defined(FS_RDONLY) && !defined(FS_NONE)
  /* normal case: there is a file system */
  FILE *fp = NULL;
  int acc = 0;
  char userfilename[256];
  if (el == NULL) return non_existent_elem;
  if (isUsingXMLFecDatabase) return not_implemented;
  memset(&gLastAllowedCln,0,sizeof(ClnHdr)); /* reset this */
  if (userlist != NULL && listsize > 0)
  { /* there's a list coming */
    for (i=0; i<listsize; i++) strntrm(userlist[i].name,16);
    for (i=0; i<listsize; i++)
    {
      if (isAllowedUser((char *)eqm,userlist[i].name)) userlist[i].name[0] = 0;
      if (strlen(userlist[i].name) > 0) n++;
    }
    if (n == 0) return 0; /* everybody's already in the list */
    /* re-write the file ... */
    sprintf(userfilename,"%.128s%.32s",fecDbPath,el->usersFileName);
    if ((fp=fopen(userfilename,"w")) == NULL) ccerr(no_such_file);
    fprintf(fp,"USERNAME\n");  /* single column header */
    dumpUsersToFile(el->aclLst.usrs,el->aclLst.nUsrs,fp);
    dumpUsersToFile(userlist,listsize,fp);
    fclose(fp);
  }
  el->chkAcl &= ~(EQM_CHECK_USR|EQM_CHECK_GRP);
  getRegUsrsFromFile((char *)eqm,NULL,NULL,&el->aclLst,el->usersFileName,"WRITE",&acc);
  el->chkAcl |= acc;
# else
  if (el == NULL) return non_existent_elem;
  if (el->aclLst.usrs == NULL)
  {
    if ((el->aclLst.usrs=(NAME16 *)SystemCalloc(MaxNumAllowedUsers,sizeof(NAME16))) == NULL)
      ccerr(out_of_server_memory);
    el->aclLst.nUsrs = 0;
  }
  if (userlist == NULL || listsize == 0) ccerr(argument_list_error);
  for (i=0; i<listsize; i++) strntrm(userlist[i].name,16);
  for (i=0; i<listsize; i++)
  {
    if (isAllowedUser((char *)eqm,userlist[i].name)) userlist[i].name[0] = 0;
    if (strlen(userlist[i].name) > 0) n++;
  }
  if (n == 0) return 0; /* everybody's already in the list */
  if (el->aclLst.nUsrs == MaxNumAllowedUsers) ccerr(resources_exhausted);
  n = el->aclLst.nUsrs;
  for (i=0; i<listsize && n<MaxNumAllowedUsers; i++)
  {
    if (strlen(userlist[i].name) == 0) continue;
    strncpy(el->aclLst.usrs[n++].name,userlist[i].name,USERNAME_SIZE);
  }
  el->aclLst.nUsrs = n;
# endif
err:
  return cc;
}
_CSVCOLUMN csvUserFileColumns[] = 
{
  { "USER_NAME","",0,CF_TEXT,USERNAME_SIZE,NULL,NULL,0 }
};
CSVDB csvUserFileDb =
{
  csvUserFileColumns,
  sizeof(csvUserFileColumns)/sizeof(csvUserFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME64),
  0
};
int getRegUsrsFromFile(char *eqm,char *dev,char *prp,AclInfoType *acl,char *fn,char *accstr,int *accflg)
{ 
  int i,n=0,sz=0,fl=0,cc=no_such_file;
  char usr[64],*grp=NULL,*src=NULL,*c;
  NAME64 *ul64 = NULL;
  char label[128] = "FEC";
  int chkPrp=FALSE,chkDev=FALSE;
  if (acl == NULL) return argument_list_error;
  if (gTineConfigurationCoded) return not_implemented;
  if (eqm != NULL && strlen(eqm) > 0) 
  {
    strncpy(label,eqm,EQM_NAME_SHORTSIZE);
  }
  if (prp != NULL && strlen(prp) > 0) 
  {
    chkPrp = TRUE;
    if (strlen(label) > 0) strcat(label,": property ");
    strncat(label,prp,PROPERTY_NAME_SIZE);
  }
  else if (dev != NULL && strlen(dev) > 0) 
  {
    chkDev = TRUE;
    if (strlen(label) > 0) strcat(label,": device ");
    strncat(label,dev,DEVICE_NAME_SIZE);
  }
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase && accflg != NULL)
  { /* give this a try */
    char *tag = !strcmp(accstr,"WRITE") ? USERS_ALLOWED_SETNAME : USERS_DENIED_SETNAME;
    sz = populateNameList(gFecInfoList,gFecName,eqm,dev,prp,tag,&ul64);
    cc = (sz < 0) ? -sz : 0;
    if (sz > 0) fl = CHECK_USR;
  }
# endif
  if (cc != 0)
  { /* try the csv file */
    if ((cc=csvReadFile(fecDbPath,fn,&csvUserFileDb,(void **)&ul64)) == 0)
    { /* file found */
      sz = csvUserFileDb.siz;
      fl = CHECK_USR;
    }
  }
  if (fl != 0)
  { /* file or list found */
    acl->nUsrs = sz;
    if (sz > 0)
    { /* list has members */
      if (ul64 == NULL) ccerr(code_failure);
      SystemFree(acl->usrs);
      SystemFree(acl->grps);
      if ((acl->usrs=(NAME16 *)SystemCalloc(sz,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
      for (i=0; i<sz; i++)
      { /* parse and spit out the list */
        strncpy(usr,ul64[i].name,64);
        if (usr[0] == '<')
        { /* user group specified */
          grp = &usr[1];
          if ((c=strchr(grp,'>')) != NULL) *c = 0;
          if ((c=strchr(grp,':')) != NULL)
          { /* group registry source specified */
            src = grp; *c++ = 0; grp = c;
          }
          if (addUserGroupToList(acl,fn,grp,src) == 0)
          {
            feclog("%s: group %.16s has %s access",label,grp,accstr);
          }
          fl |= CHECK_GRP;
          continue;
        }
        strncpy(acl->usrs[n].name,ul64[i].name,USERNAME_SIZE);
        feclog("%s: user %.16s has %s access",label,acl->usrs[n].name,accstr);
        n++;
      }
      acl->nUsrs = n; /* re-adjust in case of user groups */
    }
    else if (accflg != NULL)
    { /* no members and I'm checking global access! */
      strcpy(usr,!strcmp(accstr,"WRITE") ? "NONE" : "EVERYONE");
      feclog("%s: user access : %.16s has %s access",label,usr,accstr);
      dbglog("WARNING: empty users list found => %.16s has %s access to module %.6s",usr,accstr,eqm);
    }
  }
  else
  { /* file or list not found : everything is open */
    if (!chkPrp && !chkDev)
    { /* only enter in log file if checking general files */
      strcpy(usr,!strcmp(accstr,"WRITE") ? "EVERYONE" : "NONE");
      feclog("%s: user access : %.16s has %s access",label,usr,accstr);
    }
  }
  if (accflg != NULL) *accflg = fl;
err:
  SystemFree(ul64);
  return cc;
}
char *makeListFileName(char *eqm,int usesubdir,char *prefix,char *postfix)
{
  char subdir[32];
  static char fname[128];

  if (postfix == NULL) return NULL;
  if (usesubdir)
  {
#   ifdef FS_DELIMITER
    sprintf(subdir,"%.6s%c",eqm,FS_DELIMITER);
#   else
    return NULL;
#   endif
  }
  else
  {
    subdir[0] = 0;
  }
  if (prefix != NULL && strlen(prefix) >0)
  {
    sprintf(fname,"%.10s%.64s-%s",subdir,prefix,postfix);
  }
  else
  {
    sprintf(fname,"%s%s",subdir,postfix);
  }
  return fname;
}
int getAllowedUsers(char *eqm)
{
  int i,cc=0,acc=0;
  ExportListStruct *el=getExportListItem(eqm);
  char *fname;
  if (gTineConfigurationCoded) return not_implemented;
  if (el == NULL) return non_existent_elem;
  for (i=1; i>=0; i--)
  {
    if ((fname=makeListFileName(eqm,i,NULL,USERS_FILE)) == NULL) continue;
    strncpy(el->usersFileName,fname,32);
    el->chkAcl &= ~(EQM_CHECK_USR|EQM_CHECK_GRP);
    cc = getRegUsrsFromFile(eqm,NULL,NULL,&el->aclLst,el->usersFileName,"WRITE",&acc);
    el->chkAcl |= acc;
    if (cc == 0) break;
  }
  if (cc == no_such_file)
  { /* nada! but reset the prefered location */
    strncpy(el->usersFileName, makeListFileName(eqm,TRUE,NULL,USERS_FILE), 32);
  }
  else if (cc != 0)
  { /* something's amiss */
    feclogEx(TINE_LOGLEVEL_ERROR, "%.6s : could not read users file -> %.32s",eqm,cc2str(cc));
  }
  return cc;
}
void dumpRegisteredUsersList(AclInfoType *acl)
{
  int i;
  dbglog("\tWrite Access open to users:\n");
  for (i=0; i<acl->nUsrs; i++)
  {
    dbglog("\t  %.16s\n",acl->usrs[i].name);
  }
  if (acl->nGrps > 0)
  {
    dbglog("\tWrite Access open to groups:\n");
    for (i=0; i<acl->nGrps; i++)
    {
      dbglog("\t  %.16s\n",acl->grps[i].name);
    }
  }
}
void dumpRegisteredAdmins(void)
{
  int i;
  if (gNumFecAdmins == 0)
  {
    dbglog("There are no registered administrators !\n");
    return;
  }
  dbglog("Registered administrators:\n");
  for (i=0; i<gNumFecAdmins; i++)
  {
    dbglog("  %.16s",gFecAdmins[i].name);
  }
}
void dumpRegisteredUsers(void)
{
  int idx;
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  for (el=ExportList; el != NULL; el = el->next)
  {
    dbglog("Equipment Module : %.8s\n",el->EqmName);
    if (!(el->chkAcl & (EQM_CHECK_USR|EQM_CHECK_GRP)))
    {
      dbglog("\tWrite Access open to all users\n");
      continue;
    }
    dumpRegisteredUsersList(&el->aclLst);
#   ifndef SMALL_TINE_LIB
    if (el->chkAcl & (PRP_CHECK_USR|PRP_CHECK_GRP))
    {
      for (idx=0; idx<PROPERTY_HASH_SIZE; idx++)
        for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
      {
        if (prp->isReserved) continue;
        if (prp->aclLst.nUsrs > 0)
        {
          dbglog("    Property : %.64s\n",prp->prpName);
          dumpRegisteredUsersList(&prp->aclLst);
        }
      }
    }
    if (el->chkAcl & (DEV_CHECK_USR|DEV_CHECK_GRP))
    {
      for (idx=0; idx<el->EqmNumDevices && el->EqmDevLst != NULL; idx++)
      {
        if (el->EqmDevLst[idx].aclLst == NULL) continue;
        if (el->EqmDevLst[idx].aclLst->nUsrs > 0)
        {
          dbglog("    Device : %.64s\n",el->EqmDevLst[idx].name);
          dumpRegisteredUsersList(el->EqmDevLst[idx].aclLst);
        }
      }
    }
#   else
    prp = NULL;
    idx = 0;
#   endif
  }
}
void dumpRegisteredGroups(void)
{
  UsrGrpsLst *g=NULL;
  ExportListStruct *el;
  int i;
  for (el=ExportList; el != NULL; el = el->next)
  {
    dbglog("Equipment Module : %.8s\n",el->EqmName);
    if (el->aclLst.nGrps == 0) dbglog("\tno groups registered");
    if (!(el->chkAcl & EQM_CHECK_GRP))
    {
      dbglog("\tWrite Access open to all users\n");
      continue;
    }
    for (i=0; i<el->aclLst.nGrps; i++)
    { /* run thru list of known user groups */
      if ((g=getUserGroupFromCache(el->aclLst.grps[i].name)) != NULL)
      { /* should always be true */
        ttyoutput("\tWrite Access granted to %.64s group %.16s (%d members)",
          g->name,g->type,g->nmembers);
      }
    }
  }
}
int getUserGroupMembersFromFile(char *filename,NAME16 **members,int *nmembers)
{
  int cc=0, sz;

  cc = csvReadFile(fecDbPath,filename,&csvUserFileDb,(void **)members);
  sz = cc == 0 ? csvUserFileDb.siz : 0;
  if (*nmembers != 0) *nmembers = sz;
  return cc;
}
#if defined(WIN32) && defined(LINK_NETAPI32)
#include <lm.h>
int getWinUserGroupMembers(char *grp,char *src,NAME16 **members,int *nmembers)
{
  DWORD ner=0,ne=0;
# ifdef OS_64BIT
  PDWORD rh = 0;
# else
  DWORD rh = 0;
# endif
  LPGROUP_USERS_INFO_0 pBuf = NULL, pMem;
  wchar_t tsrv[64];
  wchar_t tgrp[64];
  int i, rc, cc=0, bsiz=10000; /* can accomodate 100 users */
  NAME16 *mems = NULL;

  if (src == NULL) src = "";
  memset(tsrv,0,64*sizeof(wchar_t));
  memset(tgrp,0,64*sizeof(wchar_t));
  mbstowcs(tsrv,src,strlen(src));
  mbstowcs(tgrp,grp,strlen(grp));

retryGetUsers:
# ifdef OS_64BIT
  switch (rc=NetGroupGetUsers(tsrv,tgrp,0,(LPBYTE*)&pBuf,bsiz,&ner,&ne,(PDWORD_PTR)&rh))
# else
  switch (rc=NetGroupGetUsers(tsrv,tgrp,0,(LPBYTE*)&pBuf,bsiz,&ner,&ne,&rh))
# endif
  {
    case NERR_Success:
      break;
    case ERROR_ACCESS_DENIED:
      ccerr(access_denied);
    case ERROR_MORE_DATA:
      if (bsiz < 20000) 
      { /* be reasonable */
        bsiz += 2000;
        goto retryGetUsers;
      }
      ccerr(buffer_too_small);
    case NERR_InvalidComputer:
      ccerr(no_such_computer);
    case NERR_UserNotFound:
      ccerr(non_existent);
    case ERROR_LOGON_FAILURE:
      ccerr(unauthorised_action);
    default:
      ccerr(code_failure);
  }
  if ((mems=(NAME16 *)SystemCalloc(ner,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
 
  pMem = pBuf;
  for (i=0,pMem=pBuf; i<(int)ner && pMem != NULL; i++,pMem++)
  {
    wcstombs(mems[i].name,pMem->grui0_name,16);
  }
  *members = mems;
err:
  *nmembers = ner;
  if (pBuf != NULL &&!IsBadReadPtr(pBuf->grui0_name,bsiz))
  { 
    NetApiBufferFree(pBuf);
  }
  if (cc != 0) feclogEx(TINE_LOGLEVEL_WARN, "NetGroupGetUsers returned %d",rc);
  return cc;
}
#elif defined(UNIX)
#include "grp.h"
int getUnixUserGroupMembers(char *grp,NAME16 **members,int *nmembers)
{
# define NICE_GRPLST_SIZE 256
  int n=0, cc=0;
  NAME16 mems[NICE_GRPLST_SIZE], *pmems=NULL;
  struct group *gp;
  char **cp;
  if (grp == NULL || members == NULL || nmembers == NULL) return argument_list_error;
  setgrent ();
  while ((gp=getgrent()) != NULL)
  { /* interate until we get a match */
    if (strcmp(gp->gr_name,grp)) continue;
    cp = gp->gr_mem;
    while (*cp && n<NICE_GRPLST_SIZE) strncpy(mems[n++].name,*cp++,16);
  }
  endgrent ();
  if (n == 0) ccerr(0);
  if ((pmems=(NAME16 *)SystemCalloc(n,sizeof(NAME16))) == NULL) 
  {
    n = 0;
    ccerr(out_of_server_memory);
  }
  memcpy(pmems,mems,n*sizeof(NAME16));
err:
  *nmembers = n;
  *members = pmems;
  return cc;
}
#endif /* WIN32 or UNIX */
TINE_EXPORT int GetUserGroupMembers(char *filename,char *grp,char *src,NAME16 **members,int *nmembers)
{
  int cc = not_implemented;
  char *tag = "local";
  if (grp == NULL || members == NULL || nmembers == NULL) ccerr(argument_list_error);
  /* clear the balance sheets ... */
  SystemFree(*members);
  if (src == NULL && filename != NULL)
  { /* no query source specified -> go local */
    char fn[128]="";
#   ifdef FS_DELIMITER
    char *c;
    strncpy(fn,filename,64); fn[64] = 0;
    if ((c=strrchr(filename,FS_DELIMITER)) != NULL) *c = 0;
#   endif
    strcat(fn,grp);
    strcat(fn,"-members.csv");
    cc = getUserGroupMembersFromFile(fn,members,nmembers);
    if (cc == 0) ccerr(cc);
  }
# if defined(WIN32) && defined(LINK_NETAPI32)
  if (src == NULL || strlen(src) == 0) src = getenv("USERDOMAIN");
  tag = src;
  ccerr(getWinUserGroupMembers(grp,src,members,nmembers));
# endif
# if defined(UNIX)
  if (src == NULL || strlen(src) == 0) src = getenv("services");
  tag = src;
  ccerr(getUnixUserGroupMembers(grp,members,nmembers));
# endif
err:
  if (cc == 0)
  {
    feclog("%s group %.16s found %d members",tag,grp,*nmembers);
  }
  else if (cc != no_such_file)
  {
    feclog("%s group %.16s : %.32s",tag,grp,cc2str(cc));
  }
  return cc;
}
_CSVCOLUMN csvFollowersColumns[] = 
{
  { "USER_NAME","",0,CF_TEXT,64,NULL,NULL,0 }
};
CSVDB csvFollowersDb =
{
  csvFollowersColumns,
  sizeof(csvFollowersColumns)/sizeof(csvFollowersColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME64),
  0
};
int getFollowersFromFile(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  NAME64 *cl64 = NULL;
# if !defined(FS_RDONLY) && !defined(FS_NONE) && defined(FS_DELIMITER)
  char fn[256],filename[64];
  int cc = 0, n = 0;
  if (el == NULL) return non_existent_elem;
  SystemFree(el->followers); el->nfollowers = 0;
  sprintf(filename,"%.6s%c%.16s",eqm,FS_DELIMITER,FOLLOWERS_FILE);
  snprintf(fn,256,"%.128s%.64s",fecDbPath,filename);
  if ((n=GetDataBaseSize(fn)) < 0) n = 0;
  if ((cl64=(NAME64 *)SystemCalloc(n+FOLLOWERS_BUFFER_SPACE,sizeof(NAME64))) == NULL) return out_of_server_memory;
  el->capfollowers = csvFollowersDb.siz = n + FOLLOWERS_BUFFER_SPACE;
  el->nfollowers = n;
  el->followers = cl64;
  if (n > 0 && (cc=csvReadFile(fecDbPath,filename,&csvFollowersDb,(void **)&cl64)) != 0)
  {
    el->nfollowers = FOLLOWERS_BUFFER_SPACE;
  }
  return cc;  
# else
  if (el == NULL) return non_existent_elem;
  SystemFree(el->followers); el->nfollowers = 0;
  if ((cl64=(NAME64 *)SystemCalloc(FOLLOWERS_BUFFER_SPACE,sizeof(NAME64))) == NULL) return out_of_server_memory;
  el->capfollowers = FOLLOWERS_BUFFER_SPACE;
  el->followers = cl64;
  return 0;
# endif
}
int writeFollowersToFile(char *eqm)
{
# if !defined(FS_RDONLY) && !defined(FS_NONE) && defined(FS_DELIMITER)
  FILE *fp = NULL;
  ExportListStruct *el = getExportListItem(eqm);
  char fn[256];
  int cc = 0, i = 0;
  int dirchkdone = FALSE;
  snprintf(fn,256,"%.128s%.6s%c%.16s",fecDbPath,eqm,FS_DELIMITER,FOLLOWERS_FILE);
  if (el == NULL) return non_existent_elem;
  if (nofeclog) return not_implemented;
# ifndef SMALL_TINE_LIB
fnDone:
# endif
  fp = fopen(fn,"wt");
# ifndef SMALL_TINE_LIB
  if (fp == NULL && !dirchkdone)
  {
    char dn[256];
    snprintf(dn,256,"%.128s%.6s%c",fecDbPath,eqm,FS_DELIMITER);
    dirchkdone = TRUE;
    if (makeDirectoryPath(dn,TRUE)) goto fnDone;
  }
# endif
  if (fp == NULL) ccerr(file_error);
  fprintf(fp,"USER_NAME\n");
  for (i=0; i<el->nfollowers; i++)
  {
    fprintf(fp,"%.64s\n",el->followers[i].name);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR,"could not open %.256s",fn);
  return cc;
# else
  eqm = eqm;
  return 0;
# endif
}
int gIgnoreFollowers = FALSE;
TINE_EXPORT void SetIgnoreFollowers(int value)
{
  gIgnoreFollowers = value;
}
int addFollower(char *eqm,char *follower,int length)
{
  int i,len=MIN(length,64);
  ExportListStruct *el = getExportListItem(eqm);
  if (gIgnoreFollowers) return 0;
  if (el == NULL) return non_existent_elem;
  if (!strcmp(eqm,ENS_TAG) || !strcmp(eqm,GENS_TAG)) return 0; /* don't do this for the ENS */
  if (el->followers == NULL) return not_initialized;
  for (i=0; i<el->nfollowers; i++) if (!strnicmp(el->followers[i].name,follower,len)) return already_assigned;
  if (el->nfollowers == el->capfollowers)
  { /* we're at capacity: realloc ! */
    NAME64 *tn = (NAME64 *)realloc(el->followers,(el->capfollowers+FOLLOWERS_BUFFER_SPACE)*sizeof(NAME64));
    if (tn == NULL) return out_of_server_memory;
    memset(&tn[el->nfollowers],0,FOLLOWERS_BUFFER_SPACE*sizeof(NAME64));
    el->followers = tn;
    el->capfollowers += FOLLOWERS_BUFFER_SPACE;
  }
  strncpy(el->followers[i].name,follower,len);
  el->nfollowers++;
  writeFollowersToFile(eqm);
  return 0;
}
int rmvFollower(char *eqm,char *follower)
{
  int i;
  ExportListStruct *el = getExportListItem(eqm);
  if (gIgnoreFollowers) return 0;
  if (el == NULL) return non_existent_elem;
  if (el->followers == NULL || el->nfollowers == 0) return not_initialized;
  for (i=0; i<el->nfollowers; i++) if (!strnicmp(el->followers[i].name,follower,64)) break;
  if (i == el->nfollowers) return name_not_found;
  for (; i<el->nfollowers-1; i++)
  { /* shore up the list */
    memcpy(el->followers[i].name,el->followers[i+1].name,64);
  }
  el->nfollowers--;
  writeFollowersToFile(eqm);
  return 0;
}
int dumpFollowerList(char *eqm)
{ 
  int i;
  ExportListStruct *el = getExportListItem(eqm);
  if (eqm == NULL || strlen(eqm) == 0) el = ExportList;
  if (el == NULL) return non_existent_elem;
  dbglog("known followers of %.6s:\n",el->EqmName);
  for (i=0; i<el->nfollowers; i++) 
  {
    dbglog("  %.16s",el->followers[i].name);
  }
  return 0;  
}
int RegisterExport(char *expName,char *eqmName,int numdevices)
{
  int i, cc = 0, scanFiles = TRUE;
  ExportListStruct *el;
  static int inside = 0;
  char expNameBuffer[128], *c, *exp = NULL, *ctx = NULL, *sub = NULL;

  if (inside) return -1; else inside = -1;
  if (ServerExitCondition > 0) ccerr(not_running);

  if (eqmName == NULL || strlen(eqmName) == 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "RegisterExport: missing local equipment module name");
    ccerr(argument_list_error);
  }
# if defined(VXWORKS)
  if (EquipmentModulesInitialized) scanFiles = FALSE;
# endif
  if (numdevices < 1) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "RegisterExport: invalid number of devices %d corrected to 1 device",numdevices);
    numdevices = 1;
  }
  feclog("RegisterExport : server (export) name -> %.32s, eqm (local) name -> %.6s, num devices -> %d",
    expName == NULL ? "(null)" : expName, eqmName, numdevices);
  if ((el=getExportListItem(eqmName)) != NULL) ccerr(0);        /* already in list */
  el = (ExportListStruct *)SystemCalloc(1,sizeof(ExportListStruct));
  if (el == NULL) ccerr(out_of_server_memory);
  strncpy(el->EqmName,eqmName,EQM_NAME_SHORTSIZE);
  if (expName == NULL || strlen(expName) == 0) 
  { /* registered without an export name -> find in config file ! */
    if (scanFiles)
    {
      feclog("looking for exports for local name %.6s",eqmName);
      GetAliasTable(eqmName);
      ccerr(getRegisteredExports(EXPORT_FILE,eqmName));
    }
    else
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "no export name for %6s given !",eqmName);
      ccerr(argument_list_error);
    }
  }
  strncpy(expNameBuffer,expName,128);
  if (expNameBuffer[0] == '/')
  { /* context given */
    ctx = &expNameBuffer[1];
    if ((exp=strchr(ctx,'/')) == NULL) ccerr(argument_list_error);
    *exp++ = 0;
    if ((sub=strchr(ctx,'[')) != NULL)
    { /* subsystem given */
      *sub++ = 0;
      if ((c=strchr(sub,']')) != NULL) *c = 0;
    }
  }
  else
  { /* just the export */
    exp = expNameBuffer;
  }
  if (strlen(exp) == 0) ccerr(argument_list_error);
  if (ctx == NULL) ctx = gDeviceContext;
  if (sub == NULL) sub = gDeviceSubSystem;
  if (!assertNameIsValid(exp,RUDECHARS,EXPORT_NAME_SIZE)) ccerr(invalid_name);
  if (!assertNameIsValid(ctx,RUDECHARS,CONTEXT_NAME_SIZE)) ccerr(invalid_name);
  strncpy(el->EqmExportName,exp,EXPORT_NAME_SIZE);
  strncpy(el->EqmContext,ctx,CONTEXT_NAME_SIZE);
  strncpy(el->EqmSubSystem,sub,SUBSYSTEM_NAME_SIZE);
  if ((i=getAliasTableIndexFromName(el->EqmName)) >= 0)
    strncpy(el->EqmAlias,gAliasTable[i].alias,EQM_NAME_SHORTSIZE);
  el->EqmDevLstCapacity = numdevices;
  el->EqmNumDevices = numdevices;
  el->EqmNumProperties = 0;
  el->exportId = gEqmListSize++;
  el->next = ExportList;
  ExportList = el;
# ifndef SMALL_TINE_LIB
  getRegionRules(eqmName);
# endif
  if (scanFiles) GetEqpDeviceNames(el->EqmName,NULL,&el->EqmDevLst,&numdevices);
  if (el->EqmDevLst != NULL && strlen(el->EqmDevLst[0].name) > 0) el->hasRegisteredDevicesNames = TRUE;
  makeEqpDevicesHashTable(el);
# ifdef WIN32
  if (seekAlternateExportsFile) 
  {
    getFecName();
    if (strlen(el->EqmContext) == 0) strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
    if (strlen(el->EqmSubSystem) == 0) strncpy(el->EqmSubSystem,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
    feclog("looking for exports for local name %.6s",eqmName);
    GetAliasTable(eqmName);
    getRegisteredExports(EXPORT_FILE,eqmName);
    if (el->EqmDevLst != NULL) freeEqmDevLst(el->EqmDevLst,el->EqmDevLstCapacity);
    GetEqpDeviceNames(el->EqmName,NULL,&el->EqmDevLst,&numdevices);
    if (el->EqmDevLst != NULL && el->EqmDevLst[0].name != NULL && strlen(el->EqmDevLst[0].name) > 0) 
    {
      el->hasRegisteredDevicesNames = TRUE;
      GetDeviceAclInfo(el->EqmName,-1);
    }
  }
# endif
  if (scanFiles) getAlarmInfoTable(eqmName);
  if (el->EqmDevLst == NULL) 
  { /* if there's still no device space at this stage then make some ... */
    makeEqmDevLst(&el->EqmDevLst,el->EqmDevLstCapacity);
  }
err:
  if (!FecNameRegistered && scanFiles) getFecName();
  inside = 0;
  return cc;
}
TINE_EXPORT int GetNumberEquipmentModules(void)
{
  int n=0;
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next) n++; 
  return n;
}

_CSVCOLUMN csvExportsInfoColumns[] = 
{
  { "CONTEXT","",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmContext),NULL,0 },
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(XIS,SubSystem),NULL,0 },
  { "EXPORT_NAME","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,ExpName),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmName),NULL,0 },
  { "NUM_DEVICES","0",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,DevSpace),NULL,0 },
  { "NUM_MODULES","0",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,altDevSpace),NULL,0 },
  { "REGION","",0,CF_TEXT,TAG_NAME_SIZE,(BYTE *)OFFSETIN(XIS,region),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(XIS,PrpName),NULL,0 },
  { "PROPERTY_SIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,PrpSizeOut),NULL,0 },
  { "PROPERTY_INSIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,PrpSizeIn),NULL,0 },
  { "PROPERTY_ID","-1",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,PrpId),NULL,0 },
  { "FORMAT","float",0,CF_TEXT,FORMAT_STR_SIZE,(BYTE *)OFFSETIN(XIS,fmtOut),NULL,0 },
  { "INFORMAT","",0,CF_TEXT,FORMAT_STR_SIZE,(BYTE *)OFFSETIN(XIS,fmtIn),NULL,0 },
  { "ACCESS","READ",0,CF_TEXT,ACCESS_STR_SIZE,(BYTE *)OFFSETIN(XIS,access),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,128,(BYTE *)OFFSETIN(XIS,description),NULL,0 },
  { "REDIRECTION","",0,CF_TEXT,PROPERTY_REDIR_SIZE,(BYTE *)OFFSETIN(XIS,redirection),NULL,0 },
# ifndef SMALL_TINE_LIB
  { "ACCESS_DEADBAND","0",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,deadband),NULL,0 },
  { "UNITS","",0,CF_TEXT,16,(BYTE *)OFFSETIN(XIS,units),NULL,0 },
  { "XUNITS","",0,CF_TEXT,16,(BYTE *)OFFSETIN(XIS,xunits),NULL,0 },
  { "MAX_VALUE","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(XIS,max),NULL,0 },
  { "MIN_VALUE","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(XIS,min),NULL,0 },
  { "XMAX_VALUE","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(XIS,xmax),NULL,0 },
  { "XMIN_VALUE","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(XIS,xmin),NULL,0 },
  { "GROUP","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmGroup),NULL,0 },
  { "GROUP_DEVICE_PREFIX","",0,CF_TEXT,XFIX_SIZE,(BYTE *)OFFSETIN(XIS,EqmGroupDevPrefix),NULL,0 },
  { "GROUP_DEVICE_POSTFIX","",0,CF_TEXT,XFIX_SIZE,(BYTE *)OFFSETIN(XIS,EqmGroupDevPostfix),NULL,0 },
  { "GROUP_INDEX","",0,CF_INT32,1,(BYTE *)OFFSETIN(XIS,EqmGroupIdx),NULL,0 },
  { "MASTER","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmMaster),NULL,0 },
  { "SLAVE_MASTER","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmSlaveMaster),NULL,0 },
# endif
};
CSVDB csvExportsInfoDb =
{
  csvExportsInfoColumns,
  sizeof(csvExportsInfoColumns)/sizeof(csvExportsInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(XIS),
  0
};
#ifndef SMALL_TINE_LIB
int gUsersScanFlag = 0;
int gNetsScanFlag = 0;
TINE_EXPORT int SetScanForUsersFiles(const char *eqm)
{
  ExportListStruct *el = getExportListItem((char *)eqm);
  if (el == NULL) 
  {
    gUsersScanFlag = 2;
  }
  else
  {
    el->usersScanFlag = 2;
  }
  return 0;
}
TINE_EXPORT int SetScanForNetsFiles(const char *eqm)
{
  ExportListStruct *el = getExportListItem((char *)eqm);
  if (el == NULL)
  {
    gNetsScanFlag = 2;
  }
  else
  {
    el->netsScanFlag = 2;
  }
  return 0;
}
int setAclFileInfo(char *eqm,int *usrFile,int *netFile)
{
  char path[256];
  int cc=0,uf=-1, nf=-1;
  ExportListStruct *el = getExportListItem(eqm);
  if (isUsingXMLFecDatabase) ccerr(not_applicable);
  if (el == NULL) ccerr(invalid_parameter);
  if (el->usersScanFlag == 0)
  {
    if (gUsersScanFlag != 0) 
    {
      el->usersScanFlag = gUsersScanFlag;
      goto netsScanCheck;
    }
    sprintf(path,"%.128s%.6s",fecDbPath,eqm);
    if (hasFilesOfType(path,"*-users.csv")) uf = 1;
    else if (hasFilesOfType(fecDbPath,"*-users.csv")) uf = 0;
    el->usersScanFlag = uf < 0 ? uf : uf + 1;
  }
netsScanCheck:
  if (el->netsScanFlag == 0)
  {
    if (gNetsScanFlag != 0) 
    {
      el->netsScanFlag = gNetsScanFlag;
      goto doneScanCheck;
    }
    if (hasFilesOfType(path,"*-ipnets.csv")) nf = 1;
    else if (hasFilesOfType(fecDbPath,"*-ipnets.csv")) nf = 0;
    el->netsScanFlag = nf < 0 ? nf : nf + 1;
  }
doneScanCheck:
  if (usrFile != NULL) *usrFile = el->usersScanFlag - 1;
  if (netFile != NULL) *netFile = el->netsScanFlag - 1;
err:
  return cc;
}
#endif /* !SMALL_TINE_LIB */
int ignoreCommonDbFiles = FALSE;
TINE_EXPORT void SetIgnoreCommonFiles(int value)
{
  ignoreCommonDbFiles = value;
}
TINE_EXPORT int GetIgnoreCommonFiles(void)
{
  return ignoreCommonDbFiles;
}
#ifndef SMALL_TINE_LIB
int logSaveRestoreProblem(char *txt,char *prp)
{
  feclogEx(TINE_LOGLEVEL_ERROR, "%.64s is not an allowed save-and-restored property : %s",prp,txt);
  return FALSE;
}
int isAllowedSaveRestoreTraceArray(short fmt)
{
  switch (fmt)
  { 
    case CF_STRUCT:
    case CF_SPECTRUM:
    case CF_GSPECTRUM:
    case CF_DBLINT:
    case CF_SHTINT:
    case CF_INTDBL:
    case CF_FLTDBL:
    case CF_BIT:
    case CF_HITTS:
    case CF_STRING:
    case CF_XML:
    case CF_KEYVALUE:
    case CF_IMAGE:
    case CF_ASPECTRUM:
    case CF_AIMAGE:
    case CF_HISTORY:
    case CF_MDA:
    case CF_DEFAULT:
    case CF_NULL:
      return FALSE;
    default:
      break;
  }
  return TRUE;
}
int isAllowedSaveRestoreFormat(short fmt)
{
  switch (fmt)
  { 
    case CF_DBLINT:
    case CF_SHTINT:
    case CF_INTDBL:
    case CF_FLTDBL:
    case CF_BIT:
    case CF_HITTS:
    case CF_GSPECTRUM:
    case CF_ASPECTRUM:
    case CF_AIMAGE:
    case CF_HISTORY:
    case CF_MDA:
    case CF_DEFAULT:
    case CF_NULL:
      return FALSE;
    default:
      break;
  }
  return TRUE;
}
int isAllowedSaveRestoreProperty(ExportListStruct *el, ExportPropertyListStruct *pl)
{
  UINT32 sizOut,sizIn,sizEx=1,sizTg=1,sizFmt=1,sizHdr=0;
  int n;
  if (pl->srBuf != NULL) return TRUE;
  if ((sizIn=pl->prpSizeIn) < 1) return logSaveRestoreProblem("no registered input",pl->prpName);
  if ((sizOut=pl->prpSize) < 1) return logSaveRestoreProblem("no registered output",pl->prpName);
  if (pl->prpFormat != pl->prpFormatIn) return logSaveRestoreProblem("input and output data types do not match",pl->prpName);
  if (!isAllowedSaveRestoreFormat(pl->prpFormat)) return logSaveRestoreProblem("not an allowed data type",pl->prpName);
  if (isOverloadedWriteProperty(el,pl))
  {
    ExportPropertyListStruct *p;
    for (p=pl; p != NULL; p=p->next)
    { /* remove information for the initial registration */
      SystemFree(p->srBuf);
      p->srBufSiz = 0;
    }
    return logSaveRestoreProblem("overloaded WRITE calls not allowed",pl->prpName);
  }
  sizFmt = fmtsizeof(pl->prpFormat);
  switch (pl->prpFormat)
  { /* some type specific checks ... */
    case CF_TEXT:
      if ((pl->prpArrayType & AT_CHANNEL) || sizIn != sizOut)
      {
        return logSaveRestoreProblem("text data size mismatch",pl->prpName);
      }
      sizIn = sizOut = 1;  /* fall into next 'if': an array of chars is 1 string */
      sizEx = pl->prpSize; /* reserve the right amount of space */
      break;
    case CF_STRING:
      sizFmt = sizeof(char *) + (STRING_STD_CAPACITY*2);
      break;
    case CF_STRUCT:
      if (strncmp(pl->prpFormatTag.name,pl->prpFormatTagIn.name,TAG_NAME_SIZE) != 0)
        return logSaveRestoreProblem("input and output structure tags do not match",pl->prpName);
      if ((sizTg=GetStructSize(pl->prpFormatTag.name)) == 0)
        return logSaveRestoreProblem("expected registered structure tag",pl->prpName);
      if (sizIn != 1) 
        return logSaveRestoreProblem("structure input size must be only 1 structure",pl->prpName);
      break;
    case CF_GSPECTRUM:
    case CF_SPECTRUM:
    case CF_IMAGE:
      if (sizIn <= 1)
        return logSaveRestoreProblem("SPECTRUM or IMAGE input must have size > 1!",pl->prpName);
      if (sizIn != sizOut)
        return logSaveRestoreProblem("SPECTRUM or IMAGE must have equal input and output size!",pl->prpName);
      break;
    default:
      break;
  }
  if (sizIn == 1)
  { /* property sets one at a time */
    if (sizOut == 1)
    { /* attribute style */
      n = (pl->numDevices > 0) ? pl->numDevices : el->EqmNumDevices;
      pl->srBufSiz = n * sizEx * sizTg;
    }
    else
    { /* part of a multi-channel array ? */
      if ((pl->prpArrayType & AT_CHANNEL) == 0 && pl->prpFormat != CF_STRUCT) 
        return logSaveRestoreProblem("array is not declared as multi-channel",pl->prpName);
      pl->srBufSiz = pl->prpSize * sizTg;
    }
  }
  else
  { /* property sets multiple values */
    if (sizIn != sizOut) logSaveRestoreProblem("input and output array sizes are not equal",pl->prpName);
    if (pl->prpArrayType & AT_CHANNEL)
    { /* a multi-channel array */
      pl->srBufSiz = pl->prpSize;
    }
    else if (isAllowedSaveRestoreTraceArray(pl->prpFormatIn))
    { /* an allowed trace array */
      n = (pl->numDevices > 0) ? pl->numDevices : el->EqmNumDevices;
      pl->srBufSiz = n * pl->prpSize;
    }
    else if ((sizHdr=getFormatHeaderSize(pl->prpFormatIn)) > 0)
    { /* e.g. CF_SPECTRUM or CF_IMAGE */
      n = (pl->numDevices > 0) ? pl->numDevices : el->EqmNumDevices;
      pl->srBufSiz = n * (pl->prpSize + sizHdr/sizFmt); /* sizFmt = 1 or 4 (for CF_SPECTRUM) */
    }
    else
    { /* not allowed ! */
      return logSaveRestoreProblem("non-channel array of data type not allowed!",pl->prpName);
    }
  }
  pl->srBuf = (BYTE *)SystemCalloc(pl->srBufSiz,sizFmt);
  if (pl->prpFormat == CF_STRING && pl->srBuf != NULL)
  { /* old Mr. String is always a difficult case ... */
    char *sp = (char *)pl->srBuf + pl->srBufSiz * sizeof(char *);
    for (n=0; n<(int)pl->srBufSiz; n++) 
    { /* assign the individual string pointers */
      ((char **)pl->srBuf)[n] = sp + n * (STRING_STD_CAPACITY*2);
    }
  }
  return pl->srBuf == NULL ? FALSE : TRUE;
}
srLst *gSrLst = NULL;
void addSrPrp(ExportListStruct *el,ExportPropertyListStruct *pl,int status)
{
  srLst *sr = gSrLst;
  if (status != 0 && status != no_such_file && status != string_expected) return;
  while (sr != NULL)
  {
    if (sr->el == el && sr->pl == pl) return;
    sr = sr->nxt;
  }
  if ((sr=(srLst *)SystemCalloc(1,sizeof(srLst))) == NULL) return;
  sr->el = el;
  sr->pl = pl;
  sr->ts = lastSaveRestoreFileTime;
  sr->nxt = gSrLst;
  if (status != 0) sr->uninitialized = TRUE;
  gSrLst = sr;
  lastSaveRestoreFileTime = 0;
  return;
}
void freeSrPrpList(void)
{
  srLst *sr=gSrLst, *srnxt=NULL;
  while (sr != NULL)
  {
    srnxt = sr->nxt;
    SystemFree(sr);
    sr = srnxt;
  }
  gSrLst = NULL;
}
srLst *getSaveRestoreItem(ExportListStruct *el,ExportPropertyListStruct *pl)
{
  srLst *sr = NULL;
  for (sr = gSrLst; sr != NULL; sr=sr->nxt) if (sr->el == el && sr->pl == pl) break;
  return sr;
}
char *getDevNameAtIndex(ExportListStruct *el,ExportPropertyListStruct *pl,int idx)
{
  static char dev[DEVICE_NAME_SIZE+4];
  dev[DEVICE_NAME_SIZE] = 0;

  if (pl->prpArrayType & AT_CHANNEL && pl->devNames == NULL)
  {
    AssignDeviceListFromProperty(el->EqmName,pl->prpName);
  }
  if (pl->devNames != NULL && idx < pl->numDevices) strncpy(dev,pl->devNames[idx].name,DEVICE_NAME_SIZE);
  else if (idx < el->EqmNumDevices && el->EqmDevLst != NULL) strncpy(dev,el->EqmDevLst[idx].name,DEVICE_NAME_SIZE);
  if (strlen(dev) == 0) sprintf(dev,"#%d",idx);

  return dev;
}
void *getDataPtrAtIndex(ExportPropertyListStruct *pl,BYTE *data,int siz,int idx)
{
  int sizEx=1,sizFmt=1;
  if (data == NULL) 
  {
    data = pl->srBuf;
    siz = (int)pl->srBufSiz;
  }
  if (data == NULL || siz == 0) return NULL; /* not given and not registered save/restore */
  if (idx < 0 || idx > siz) return NULL;
  sizFmt = fmtsizeof(pl->prpFormatIn);
  switch (pl->prpFormatIn)
  {
    case CF_STRUCT:
      sizEx = GetStructSize(pl->prpFormatTag.name);
      if (idx >= (int)(siz/sizEx)) return NULL;
      return (void *)(&data[idx * sizEx]);
    case CF_IMAGE:
    case CF_SPECTRUM:
    case CF_GSPECTRUM:
      sizEx = pl->prpSizeIn * sizFmt + getFormatHeaderSize(pl->prpFormatIn);
      if (idx >= (int)(siz/sizEx)) return NULL;
      return (void *)(&data[idx * sizEx]);
    case CF_HISTORY:
    case CF_MDA:
      break;
    case CF_STRING:
      sizFmt = sizeof(char *);
    default:
      if (isSrArray(pl))
      { /* CF_TEXT or an allowed array */
        sizEx = pl->prpSizeIn;
        if (idx >= (int)(siz/sizEx)) return NULL;
      }
      return (void *)(&data[idx * sizFmt * sizEx]);
  }
  return NULL;
}
int acqDefPrpValsFromEqm(char *eqmName,char *prpName,BYTE *data,int num,int devNr)
{
  ClnHdr cln;
  ClnLst con;
  ContractListStruct cl;
  ExportListStruct *el=getExportListItem(eqmName);
  ExportPropertyListStruct *pl=GetPropertyListStructEx(eqmName,prpName,NULL,CA_SAVERESTORE);
  char *dev;
  DTYPE dout, din;

  if (el == NULL) return non_existent_elem;
  if (pl == NULL) return illegal_property;
  if (!el->hasPropertyHandler) return not_initialized;
  DTYPEZERO(din);
  dout.dArrayLength = pl->prpSizeIn;
  dout.dFormat = pl->prpFormatIn;
  strncpy(dout.dTag,pl->prpFormatTagIn.name,16);
  if ((dout.data.vptr=getDataPtrAtIndex(pl,data,num,devNr)) == NULL) return invalid_property;
  if ((dev=getDevNameAtIndex(el,pl,devNr)) == NULL) return illegal_device;
  memset(&cln,0,sizeof(ClnHdr)); memset(&con,0,sizeof(ClnLst)); memset(&cl,0,sizeof(ContractListStruct));
  strncpy(cln.usr,"INITIALIZATION",USERNAME_SIZE);
  con.client = &cln;
  cl.exp = el;
  cl.prp = pl;
  cl.clnLst = &con;
  cl.nclients = 1;
  cl.compStatus = not_signalled;
  return callEqm(&cl,dev,prpName,&dout,&din,CA_READ|CA_SAVERESTORE);
}
TINE_EXPORT int ApplyStoredPropertyValues(const char *eqm,const char *property)
{
  int found = FALSE;
  srLst *sr = NULL;
  ExportListStruct *el=getExportListItem((char *)eqm);
  ExportPropertyListStruct *pl=GetPropertyListStructEx((char *)eqm,(char *)property,NULL,CA_SAVERESTORE);
  if (el == NULL) return non_existent_elem;
  if (pl == NULL) return illegal_property;
  if (!el->hasPropertyHandler) return not_initialized;
  for (sr=gSrLst; sr != NULL; sr=sr->nxt)
  {
    if (el != sr->el) continue;
    if (pl != sr->pl) continue;
    if (sr->restored && sr->status == 0) return already_assigned;
    sr->restored = FALSE;
    found = TRUE;
    break;
  }
  if (!found) return invalid_property;
  applyStoredPrpValues(el);
  return 0;
}
void applyStoredPrpValues(ExportListStruct *el)
{
  ClnHdr cln;
  ClnLst con;
  ContractListStruct cl;
  ExportPropertyListStruct *pl;
  srLst *sr = NULL;
  char *dev, *tag=NULL;
  DTYPE dout, din;
  int i, n=0, cc, ssiz;

  if (el == NULL || gSrLst == NULL) return;
  DTYPEZERO(dout); DTYPEZERO(din);
  memset(&cln,0,sizeof(ClnHdr)); memset(&con,0,sizeof(ClnLst)); memset(&cl,0,sizeof(ContractListStruct));
  strncpy(cln.usr,"INITIALIZATION",USERNAME_SIZE);
  con.client = &cln;
  cl.clnLst = &con;
  cl.nclients = 1;
  for (sr=gSrLst; sr != NULL; sr=sr->nxt)
  {
    if (sr->restored) continue;
    if (el != sr->el) continue;
    pl = sr->pl;
    if (pl == NULL || pl->srBuf == NULL) continue;
    cl.exp = el;
    cl.prp = pl;
    if ((n=pl->numDevices) == 0) n = pl->srBufSiz;
    if (pl->prpSizeIn > 1) n = 1; /* special case */
    if (isSrArray(pl)) n = pl->srBufSiz / pl->prpSize;
    if (pl->prpFormatIn == CF_STRUCT)
    {
      tag = pl->prpFormatTag.name;
      if ((ssiz=GetStructSize(tag)) < 1) ssiz = 1;
      n /= ssiz;
    }
    for (i=0; i<n; i++)
    {
      if (sr->uninitialized)
      {
        din.dArrayLength = 0;
        din.dFormat = CF_NULL;
        din.data.vptr = NULL;
        dout.dArrayLength = pl->prpSizeIn;
        dout.dFormat = pl->prpFormatIn;
        if (dout.dFormat == CF_STRUCT && tag != NULL) strncpy(dout.dTag,tag,TAG_NAME_SIZE);
        if ((dout.data.vptr=getDataPtrAtIndex(pl,NULL,0,i)) == NULL) continue;
        if ((dev=getDevNameAtIndex(el,pl,i)) == NULL) continue;
        cl.compStatus = not_signalled;
        cc = callEqm(&cl,dev,pl->prpName,&dout,&din,CA_READ|CA_SAVERESTORE);
        feclog("acquire restore point (%.6s)/%.64s[%.64s] : %.32s",el->EqmName,dev,pl->prpName,cc2str(cc));
        sr->status = cc;
        continue;
      }
      dout.dArrayLength = 0;
      dout.dFormat = CF_NULL;
      dout.data.vptr = NULL;
      din.dArrayLength = pl->prpSizeIn;
      din.dFormat = pl->prpFormatIn;
      if (din.dFormat == CF_STRUCT && tag != NULL) strncpy(din.dTag,tag,TAG_NAME_SIZE);
      din.dTimeStamp = (double)sr->ts;
      if ((din.data.vptr=getDataPtrAtIndex(pl,NULL,0,i)) == NULL) continue;
      if ((dev=getDevNameAtIndex(el,pl,i)) == NULL) continue;
      cl.compStatus = not_signalled;
      cc = callEqm(&cl,dev,pl->prpName,&dout,&din,CA_WRITE|CA_SAVERESTORE);
      feclog("restore (%.6s)/%.64s[%.64s] : %.32s",el->EqmName,dev,pl->prpName,cc2str(cc));
      sr->status = cc;
    }
    sr->restored = TRUE;
  }
}
void writeSrValues(ExportListStruct *el,char *dev,char *prp,DTYPE *din)
{
  int devnr,cc = 0,size,limit,sizFmt,sizHdr;
  void *ptr;
  ExportPropertyListStruct *pl=GetPropertyListStructEx(el->EqmName,prp,dev,CA_SAVERESTORE);
  if (pl == NULL || pl->srBuf == NULL) return;
  if ((devnr=GetDeviceNumberEx(el->EqmName,dev,prp)) < 0) ccerr(illegal_device);
  limit = size = pl->srBufSiz;
  sizFmt = fmtsizeof(pl->prpFormat);
  switch (pl->prpFormat)
  {
    case CF_TEXT:
      if ((cc=GetValuesAsStringEx(din,(char *)&pl->srBuf[devnr*pl->prpSize],CF_TEXT,pl->prpSize,0)) != 0)
        ccerr(cc);
      size = pl->srBufSiz/pl->prpSize;
      break;
    case CF_STRUCT:
      if (din->dFormat != pl->prpFormat) ccerr(illegal_format);
      if (strncmp(din->dTag,pl->prpFormatTag.name,TAG_NAME_SIZE) != 0) ccerr(invalid_structure_tag);
      size = pl->srBufSiz/pl->prpSize;  /* the struct size in bytes */
      if (GetStructSize(pl->prpFormatTag.name) != size) ccerr(invalid_structure_size);
      if ((ptr=getDataPtrAtIndex(pl,NULL,0,devnr)) == NULL) ccerr(illegal_device);
      memcpy(ptr,din->data.bptr,size);
      size = pl->prpSize; /* the number of structures (> 1 if multi-channel array) */
      break;
    case CF_GSPECTRUM:
    case CF_SPECTRUM:
    case CF_IMAGE:
      if (din->dFormat != pl->prpFormat) ccerr(illegal_format);
      if ((ptr=getDataPtrAtIndex(pl,NULL,0,devnr)) == NULL) ccerr(illegal_device);
      sizHdr = getFormatHeaderSize(pl->prpFormat);
      size = pl->prpSize*sizFmt + sizHdr; /* size in bytes of the spectrum/image */
      memcpy(ptr,din->data.bptr,size);
      size /= sizFmt; /* effective number of items per type */
      size = pl->srBufSiz/size; /* the number of spectrums or images */
      break;
    case CF_STRING:
      sizFmt = sizeof(char *); /* correct value to pass in the fallthru below */
    default:
      if (isSrArray(pl))
      { /* allowed array save-and-restore: (adjust size, position, and limit parameters) */
        devnr = devnr * pl->prpSizeIn;
        size = pl->srBufSiz/pl->prpSize;
        limit = pl->prpSize + devnr; /* note: this will work fine below as the formats are equal */
      }
      if ((cc=_GetValuesAsAny(din,pl->srBuf,pl->prpFormat,sizFmt,limit,devnr,TRUE)) != 0) ccerr(cc);
      break;
  }
  if ((cc=_SavePropertyValuesEx(el->EqmName,dev,prp,pl->srBuf,pl->prpFormat,size,TRUE)) != 0) ccerr(cc);
err:
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "save (%.6s)/%.64s[%.64s] values: %.32s",el->EqmName,dev,prp,cc2str(cc));
  return;
}
#endif /* SMALL_TINE_LIB */
int getRegisteredExports(char *fname,char *eqm)
{
  int i, k, idx, sz, ndevs, fsiz, ssiz, hsiz, cc=0;
  SINT32 prpId;
  UINT32 dSizeInBytes = 0, qSizeInBytes = 0, tSizeInBytes;
  char *c, *ctx, *sub, *tag=eqm, xfn[128], msg[256];
  XIS *xis = NULL;
  ExportPropertyListStruct *prp;
  ExportListStruct *el=NULL;
  PrpRedirBlk *prd = NULL;
# ifndef SMALL_TINE_LIB
  char *fn;
  int prpUsersFile = -1, prpNetsFile = -1, rc;

  if (gTineConfigurationCoded) return not_implemented;
  if (tag == NULL) tag = "FEC";
  if (isUsingXMLFecDatabase)
  {
    if (eqm == NULL) getFecName();
    fname = "fec.xml";
    sz = populateXIS(gFecName,eqm,gFecInfoList,&xis);
    if (sz <= 0) ccerr(-sz);
  }
  else
# endif
  {
#   ifdef FS_DELIMITER
    if (eqm != NULL)
    {
      sprintf(xfn,"%.6s%c%.120s",eqm,FS_DELIMITER,fname);
      if ((cc=csvReadFile(fecDbPath,xfn,&csvExportsInfoDb,(void **)&xis)) == 0) goto file_read;
    }
#   endif
    if (ignoreCommonDbFiles) ccerr(no_such_file);
    if ((cc=csvReadFile(fecDbPath,fname,&csvExportsInfoDb,(void **)&xis)) != 0) ccerr(cc);
file_read:
    sz = csvExportsInfoDb.siz;
  }
  feclog("get registered exports from file (%d entries)",sz);
  for (i=0; i<sz; i++)
  { /* prepare export list table */
    if (strlen(xis[i].ExpName) == 0 || strlen(xis[i].EqmName) == 0) 
    { /* not accepted */
      feclogEx(TINE_LOGLEVEL_ERROR, "%.128s : entry %d (property %.64s) was not assigned an export or local equipment name",
        fname,i,xis[i].PrpName);
      continue; 
    }
    if (strlen(xis[i].ExpName) > 0 && !assertNameIsValid(xis[i].ExpName, STRICTLY_ILLEGALCHARS, EXPORT_NAME_SIZE))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "%.128s : entry %d export name has illegal characters",
        fname, i, xis[i].ExpName);
      continue;
    }
    if (strlen(xis[i].EqmName) > 0 && !assertNameIsValid(xis[i].EqmName, STRICTLY_ILLEGALCHARS, EQM_NAME_SHORTSIZE))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "%.128s : entry %d local name has illegal characters",
        fname, i, xis[i].EqmName);
      continue;
    }
    if (strlen(xis[i].EqmContext) > 0 && !assertNameIsValid(xis[i].EqmContext, STRICTLY_ILLEGALCHARS, CONTEXT_NAME_SIZE))
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "%.128s : entry %d context has illegal characters",
        fname, i, xis[i].EqmContext);
      continue;
    }
    for (el=ExportList; el != NULL; el=el->next)
    { /* scan eqm registry, catch a few inconsistencies ... */
      ctx = strlen(xis[i].EqmContext) > 0 ? xis[i].EqmContext : gDeviceContext;
      if (!strncmp(xis[i].EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) 
      { /* same eqm name */
        if (strnicmp(xis[i].ExpName,el->EqmExportName,EXPORT_NAME_SIZE) ||
            strnicmp(ctx,el->EqmContext,CONTEXT_NAME_SIZE))
        { /* different export names */
          feclogEx(TINE_LOGLEVEL_ERROR, "configuration error: module %.6s exported as /%.32s/%.32s is attempting to export as /%.32s/%.32s",
            el->EqmName,el->EqmContext,el->EqmExportName,ctx,xis[i].ExpName);
        }
        break;
      }
      if (!strnicmp(xis[i].ExpName,el->EqmExportName,EXPORT_NAME_SIZE) &&
          !strnicmp(ctx,el->EqmContext,CONTEXT_NAME_SIZE))
      { /* same export name */
        if (strncmp(xis[i].EqmName,el->EqmName,EQM_NAME_SHORTSIZE))
        { /* different eqm names */
          feclogEx(TINE_LOGLEVEL_ERROR, "configuration error: module %.6s exported as /%.32s/%.32s is attempting to assign module name /%.6s",
            el->EqmName,el->EqmContext,el->EqmExportName,xis[i].EqmName);
        }
        break;
      }
    }
    if (el == NULL)               
    { /* new eqm module */
      el = (ExportListStruct *)SystemCalloc(1,sizeof(ExportListStruct));
      if (el == NULL) ccerr(out_of_server_memory);
      ctx = strlen(xis[i].EqmContext) > 0 ? xis[i].EqmContext : gDeviceContext;
      sub = strlen(xis[i].SubSystem) > 0 ? xis[i].SubSystem : gDeviceSubSystem;
      strncpy(el->EqmContext,ctx,CONTEXT_NAME_SIZE);
      strncpy(el->EqmSubSystem,sub,SUBSYSTEM_NAME_SIZE);
      el->exportId = gEqmListSize++;
      el->next = ExportList;
      ExportList = el;
    }
    if (strlen(el->EqmMaster) == 0)
    {
      setFailoverConfigInExportList(el, xis[i].EqmMaster, xis[i].EqmSlaveMaster);
    }
    if (strlen(el->EqmExportName) == 0)
    { /* find export name in config file */
      if (el->EqmDevLst != NULL) freeEqmDevLst(el->EqmDevLst,el->EqmDevLstCapacity);
      strncpy(el->EqmExportName,xis[i].ExpName,EXPORT_NAME_SIZE);
      strncpy(el->EqmName,xis[i].EqmName,EQM_NAME_SIZE);
      strncpy(el->EqmRegion,xis[i].region,TAG_NAME_SIZE);
#     ifndef SMALL_TINE_LIB
      getRegionRules(el->EqmName);
#     endif
      el->region = GetValidDeviceRegion(xis[i].EqmName,xis[i].region);
      el->EqmNumProperties = 0;
    }
    if (xis[i].DevSpace == 0) xis[i].DevSpace = xis[i].altDevSpace;
    if (xis[i].DevSpace == 0) xis[i].DevSpace = 1;
    if (!strnicmp(el->EqmExportName, xis[i].ExpName, EXPORT_NAME_SIZE))
    {
      if (xis[i].DevSpace > el->EqmNumDevices)
      { /* find the maximum number entered in the config file: */
        el->EqmDevLstCapacity = el->EqmNumDevices = xis[i].DevSpace;
      }
    }
#   ifndef SMALL_TINE_LIB
    if (isUsingXMLFecDatabase) 
    {
      prpUsersFile = prpNetsFile = 1;
    }
    else
    {
      setAclFileInfo(el->EqmName,&prpUsersFile,&prpNetsFile);
    }
#   endif
    if ((prp=(ExportPropertyListStruct *)SystemCalloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
    strncpy(prp->prpName,xis[i].PrpName,PROPERTY_NAME_SIZE);
    if ((k=getAliasTableIndexFromName(prp->prpName)) >= 0)
      strncpy(prp->prpAlias,gAliasTable[k].alias,PROPERTY_NAME_SIZE);
    prp->prpSize = xis[i].PrpSizeOut;
    ssiz = 1;
    if (!strnicmp(xis[i].fmtOut,"STRUCT.",7)) /* tag coming ... */
    {
      if ((c=strstr(xis[i].fmtOut,".")) == NULL) ccerr(code_failure);
      *c++ = 0;
      strncpy(prp->prpFormatTag.name,c,TAG_NAME_SIZE);
      if ((ssiz=GetStructSize(prp->prpFormatTag.name)) == 0) ssiz = 1;
    }
    if (!strnicmp(xis[i].fmtOut,"BITFIELD",8)) /* tag coming ? ... */
    {
      if ((c=strstr(xis[i].fmtOut,".")) != NULL)
      {
        *c++ = 0;
        strncpy(prp->prpFormatTag.name,c,TAG_NAME_SIZE);
      }
    }
    prp->prpFormat = ftoi(xis[i].fmtOut);
    hsiz = getFormatHeaderSize(prp->prpFormat);
    fsiz = fmtsizeof(prp->prpFormat) * ssiz;
    tSizeInBytes = prp->prpSize * fsiz + hsiz;
    if (tSizeInBytes > dSizeInBytes) dSizeInBytes = tSizeInBytes;
    prp->prpArrayType = getArrayType(xis[i].fmtOut);
    if (!(prp->prpArrayType & AT_DOUBLE))
    {
      if (prp->prpArrayType & AT_CHANNEL) prp->numDevices = prp->prpSize;
      if (prp->prpSize > 1) prp->prpArrayType |= AT_SINGLE;
      else prp->prpArrayType = AT_SCALAR;
    }
    prp->rowSize = getArrayRowLength(xis[i].fmtOut);
    if (prp->rowSize > 0) prp->numRows = (UINT16)(prp->prpSize/prp->rowSize); else prp->rowSize = (short)prp->prpSize;
    prp->prpAccessMode = mtoi(xis[i].access);
    if (strstr(xis[i].access,"XREAD") != NULL) 
    { /* explicit check on this sub-string */
      prp->hasExclusiveRead = (UINT16)((prp->prpAccessMode & CA_READ) ? PRP_XREAD_LOCK : PRP_XREAD_ALL);
      prp->prpAccessMode &= ~CA_XREAD; /* strip it back off (?) */
    }
    if (strstr(xis[i].access,"FORCEOUTPUT") != NULL) 
    { /* explicit check on this sub-string */
      prp->enforceDataReq = TRUE;
      prp->prpAccessMode &= ~CA_FORCEOUTPUT; /* strip it back off (?) */
    }
#   ifdef MULTITHREADED
    prp->runInSeparateThread = strstr(xis[i].access,"MT") == NULL ? FALSE : TRUE;
    if (prp->runInSeparateThread) useMultiThreadedEquipmentFunctions = TRUE;
#   endif
    prp->prpSizeIn = xis[i].PrpSizeIn;
    prp->prpFormatIn = CF_NULL;
    if (strlen(xis[i].fmtIn) > 0)
    {
      if (!strnicmp(xis[i].fmtIn,"STRUCT.",7)) /* tag coming ... */
      {
        if ((c=strstr(xis[i].fmtIn,".")) == NULL) ccerr(code_failure);
        *c++ = 0;
        strncpy(prp->prpFormatTagIn.name,c,TAG_NAME_SIZE);
      }
      prp->prpFormatIn = ftoi(xis[i].fmtIn);
    }
    else
    {
      if (prp->prpAccessMode & CA_WRITE) 
      {
        prp->prpFormatIn = prp->prpFormat;
        strncpy(prp->prpFormatTagIn.name,prp->prpFormatTag.name,TAG_NAME_SIZE);
        if (prp->prpSizeIn == 0 && prp->prpFormatIn != CF_NULL) prp->prpSizeIn = prp->prpSize;
      }
    }
    ParsePropertyDescription(xis[i].description,prp);
    if (strlen(xis[i].units) > 0) strncpy(prp->egu.units,xis[i].units,UNITS_SIZE);
    if (xis[i].max != 0 || xis[i].min != 0)
    {
      prp->egu.max = xis[i].max; prp->egu.min = xis[i].min;
    }
    if (xis[i].xmax != 0 || xis[i].xmin != 0)
    {
      prp->xegu.max = xis[i].xmax; prp->xegu.min = xis[i].xmin;
    }
    if (strlen(xis[i].xunits) > 0) strncpy(prp->xegu.units,xis[i].xunits,UNITS_SIZE);
    if (strlen(xis[i].EqmGroup) > 0)
    {
      strncpy(el->EqmGroup,xis[i].EqmGroup,EXPORT_NAME_SIZE);
      el->groupIndex = xis[i].EqmGroupIdx;
      strncpy(el->EqmGroupDevPrefix,xis[i].EqmGroupDevPrefix,XFIX_SIZE);
      strncpy(el->EqmGroupDevPostfix,xis[i].EqmGroupDevPostfix,XFIX_SIZE);
    }
    if (strlen(xis[i].redirection) > 0)
    {
      if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
      ParseRedirector(xis[i].redirection,prd);
      sprintf(msg, "Redirection for property %.64s", xis[i].PrpName);
      if (assertRedirectionValid(msg,el->EqmName, prd->prpRedirection))
      {
        prd->next = prp->prd;
        prp->prd = prd;
      }
      else
      {
        SystemFree(prd);
      }      
    }
    if (!assertNameIsValid(prp->prpName, STRICTLY_ILLEGALCHARS, PROPERTY_NAME_SIZE))
    {
      feclogEx(TINE_LOGLEVEL_WARN, "Property %s contains illegal characters", prp->prpName);
      SystemFree(prp);
      continue;
    }
    idx = (int)(ElfHash((unsigned char *)prp->prpName) % PROPERTY_HASH_SIZE);
    prpId = xis[i].PrpId >= 0 ?  xis[i].PrpId: el->EqmNumProperties;
    prp->prpId = MAKE_PROPERTY_ID(prpId,idx);
    prp->next = el->EqmPrpLst[idx];
    el->EqmPrpLst[idx] = prp;
    if ((k=(int)strlen(prp->prpName)) > 4 && 
         !strcmp(&prp->prpName[k-4],".NAM") && 
         strstr(prp->prpName,".NAM.NAM") == NULL) 
    {
      el->hasPropertyDeviceNames = TRUE;
      if (queryDeviceNamesFromProperties) acquireDeviceNamesFromProperties = TRUE;
    }
    /* prp->numDevices = xis[i].DevSpace; */
    ndevs = prp->prpSize;
    if (eqm == NULL) eqm = el->EqmName;
    if ((prp->devNames=GetEqpDeviceNames(eqm,prp->prpName,NULL,&ndevs)) != NULL)
    {
      prp->numDevices = ndevs;
      el->hasPropertyDeviceNames = TRUE;
    }
#   ifndef SMALL_TINE_LIB
    if (xis[i].deadband > 0) 
    {
      prp->minAccessDeadband = xis[i].deadband;
      addPrpDbaItem(el->EqmName,prp->prpName,mtoi(xis[i].access),xis[i].deadband);
    }
    if (prpUsersFile >= 0)
    {
      int acc = 0;
      fn = makeListFileName(eqm,prpUsersFile,prp->prpName,USERS_FILE);
      if (getRegUsrsFromFile(eqm,NULL,prp->prpName,&prp->aclLst,fn,"WRITE",&acc) != 0)
      {
        fn = makeListFileName(eqm,prpUsersFile,"property-default",USERS_FILE);
        getRegUsrsFromFile(eqm,NULL,prp->prpName,&prp->aclLst,fn,"WRITE",&acc);
      }
      el->chkAcl |= (acc<<4);
    }
    if (prpNetsFile >= 0)
    {
      fn = makeListFileName(eqm,prpNetsFile,prp->prpName,IPNETS_FILE);
      if (getRegNetsFromFile(eqm,NULL,prp->prpName,fn,NETS_ACCESS,&prp->aclLst.nNets,&prp->aclLst.nets) != 0)
      {
        fn = makeListFileName(eqm,prpNetsFile,"property-default",IPNETS_FILE);
        getRegNetsFromFile(eqm,NULL,prp->prpName,fn,NETS_ACCESS,&prp->aclLst.nNets,&prp->aclLst.nets);
      }
      if (prp->aclLst.nNets > 0) el->chkAcl |= PRP_CHECK_NET;
    }
    if (strstr(xis[i].access,"SAVERESTORE") != NULL) 
    { /* one last check ... */
      if (isAllowedSaveRestoreProperty(el,prp))
      { /* so far, so good ... */
        rc = _RestorePropertyValues(eqm,prp->prpName,prp->srBuf,prp->prpFormat,prp->srBufSiz,TRUE);
        addSrPrp(el,prp,rc);
      }
    }
#   endif
    el->EqmNumProperties++;
    if (strstr(xis[i].access,"HIST") != NULL) 
    { /* explicit check on this sub-string (following EqmNumProperties increment */
      AppendHistoryInformation(eqm,prp->prpName,"#0",prp->prpSize,prp->prpFormat,-1,&defaultHSpec);
    }
  } /* end of scan loop i over exports found in file */
  for (el = ExportList; el != NULL; el = el->next)
  { /* now scan for officially registered devices ... */
    ndevs = el->EqmDevLstCapacity;
    GetEqpDeviceNames(el->EqmName, NULL, &el->EqmDevLst, &ndevs);
    if (el->EqmDevLst != NULL && strlen(el->EqmDevLst[0].name) > 0)
    {
      el->hasRegisteredDevicesNames = TRUE;
      GetDeviceAclInfo(el->EqmName, -1);
    }
    makeEqpDevicesHashTable(el);
  }
  getAlarmInfo();
  if (!gAutoAdjustWorkAreaSize) ccerr(0);
  if (dSizeInBytes > MaxSystemTransportSize)
  { /* need to make adjustments */
#   if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
    adjustWorkAreaSize(dSizeInBytes+0x10000);
#   else
    feclogEx(TINE_LOGLEVEL_ERROR, "maximum property size (%d bytes) larger than registered work area (%d bytes)",dSizeInBytes,MaxSystemTransportSize);
#   endif
  }
  qSizeInBytes = el == NULL ? 0 : el->EqmNumProperties * PRPQUERYSTRUCT_SIZE;
  if (qSizeInBytes > MaxSystemTransportSize)
  { /* need to make adjustments */
#   if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
    feclog("adjust work area size to accommodate property information queries");
    adjustWorkAreaSize(qSizeInBytes+0x10000);
#   else
    feclog("simultaneous query for all information for %d properties larger than registered work area (%d bytes)",
      el->EqmNumProperties,MaxSystemTransportSize);
#   endif
  }
err:
  SystemFree(xis);
  feclog("exports file %s (%.6s): %.32s",fname, tag, cc ? cc2str(cc) : "found");
  if (cc) seekAlternateExportsFile = TRUE; /* if initial read from FEC_HOME fails */
  return cc;
}
void setFailoverConfigInExportList(ExportListStruct *el, char *master, char *slaveMaster) 
{
  /* static vars for saving master slaveMaster values when ExportList not yet available */
  static char staticFailoverMaster[EXPORT_NAME_SIZE] = "";
  static char staticFailoverSlaveMaster[EXPORT_NAME_SIZE] = "";
  /* */
  if( strlen(staticFailoverMaster)==0 && master!=NULL && strlen(master)>0) {
    strncpy(staticFailoverMaster,master,EXPORT_NAME_SIZE);
  }
  if( strlen(staticFailoverSlaveMaster)==0 && slaveMaster!=NULL && strlen(slaveMaster)>0) {
    strncpy(staticFailoverSlaveMaster,slaveMaster,EXPORT_NAME_SIZE);
  }
  if( el!=NULL && strlen(staticFailoverMaster)>0) {
    /* failover master or slave */
    strncpy(el->EqmMaster,staticFailoverMaster,EXPORT_NAME_SIZE);
    el->failoverType = FAILOVER_MASTER;
    el->failoverState = FAILOVER_MASTER;
    if (strlen(staticFailoverSlaveMaster) > 0)
    { /* a slave ! */
      strncpy(el->EqmSlaveMaster,staticFailoverSlaveMaster,EXPORT_NAME_SIZE);
      el->failoverType = FAILOVER_SLAVE;
      el->failoverState = FAILOVER_SLAVE;
      el->registered |= REGISTERED_MASTER;
    }
  }
}
ExportListStruct *getExportListItem(char *eqm)
{
  ExportListStruct *el;

  if (eqm == NULL) eqm = INTEQM_TAG; /* tbs20160805: return internal EQM instead of NULL */
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) return el;
  return NULL;
}
TINE_EXPORT ExportListStruct *GetExportListItem(char *eqm)
{
  return getExportListItem(eqm);
}
TINE_EXPORT int GetEquipmentModuleList(NAME8 *lst,int *lstSize)
{
  ExportListStruct *el;
  int i;
  if (lst == NULL || lstSize == NULL) return argument_list_error;
  for (el=ExportList,i=0; el != NULL && i<*lstSize; el=el->next,i++) 
  {
    strncpy(lst[i].name,el->EqmName,EQM_NAME_SHORTSIZE);
  }
  *lstSize = i;
  return 0;
}
ExportListStruct *getExportListItemFromId(int id)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next) if (el->exportId == id) return el;
  return NULL;
}
int getEqpPropertyList(char *eqm,char *filename,EqmDevicePropertyList **prpList)
{ /* used internally : arguments never NULL */
  NAME64 *prps = NULL;
  EqmDevicePropertyList *pl = NULL;
  int n=0,i=0,cc=0,done=0,name_col=-1,addcsv=FALSE;
  char *c,str[256],hdr[256],fn[256];
  FILE *fp = NULL;
  
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  {
    n = populateDEVPRPIS(gFecName,eqm,filename,gFecInfoList,&prps);
    if (n <= 0) ccerr(-n);
    if ((pl=(EqmDevicePropertyList *)SystemCalloc(1,sizeof(EqmDevicePropertyList))) == NULL) ccerr(out_of_server_memory);
  }
  else
# endif
  {
#   ifdef FS_DELIMITER
    if ((c=strrchr(filename,'.')) == NULL || stricmp(c,".csv") != 0) addcsv = TRUE; 
    if (eqm != NULL)
    {
      sprintf(fn,"%.128s%.6s%c%.32s",fecDbPath,eqm,FS_DELIMITER,filename);
      if (addcsv) strcat(fn,".csv");
      if ((n=GetDataBaseSize(fn)) >= 0) goto file_read;
    }
#   endif
    if (ignoreCommonDbFiles) ccerr(no_such_file);
    sprintf(fn,"%.128s%.32s",fecDbPath,filename);
    if (addcsv) strcat(fn,".csv");
    if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
file_read:
    if ((prps=(NAME64 *)SystemCalloc(n,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    if ((pl=(EqmDevicePropertyList *)SystemCalloc(1,sizeof(EqmDevicePropertyList))) == NULL) ccerr(out_of_server_memory);
    if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
    for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
    {
      strtrm(str);
      if (strlen(str) == 0) continue;          /* carriage return */
      if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
      if (!done)
      {
        strncpy(hdr,str,255);
        if ((name_col=findcol(hdr,"PROPERTY_NAME")) < 0) ccerr(no_such_column);
        done = 1;                              /* read header info */
        continue;
      }
      getColumnStr(name_col,str,prps[i++].name,64);
    }
  }
  strncpy(pl->filename,filename,64);
  pl->prpList = prps;
  pl->prpListSize = n;
err:
  *prpList = pl;
  if (fp) fclose(fp);
  if (cc == 0) feclog("property list %s : found with %d entries",filename,n);
  else if (cc != no_such_file) feclog("property list %s : %.32s",filename,cc2str(cc));
  return cc;
}
#ifndef SMALL_TINE_LIB
_CSVCOLUMN csvRegionRulesColumns[] =
{
  { "PATTERN","0",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(DeviceRegionItem,dev),NULL,0 },
  { "REGION","0",0,CF_TEXT,TAG_NAME_SIZE,(BYTE *)OFFSETIN(DeviceRegionItem,reg),NULL,0 },
  { "CODE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(DeviceRegionItem,code),NULL,0 },
};
CSVDB csvRegionRulesDb =
{
  csvRegionRulesColumns,
  sizeof(csvRegionRulesColumns)/sizeof(csvRegionRulesColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(DeviceRegionItem),
  0
};
int getRegionRulesFromFile(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  int cc;
  char *fn;
  if (el == NULL) return non_existent_elem;
  if (el->nRegionItems > 0) return already_assigned;
  fn = makeDbFileName(eqm,REGIONRULES_FILE,TRUE);
  if ((cc=csvReadFile(fecDbPath,fn,&csvRegionRulesDb,(void **)&el->regionsLst)) == 0)
  {
    el->nRegionItems = csvRegionRulesDb.siz;
    feclog("file %s found",fn);
  }
  return cc;
}
int dumpRegionRulesToFile(char *eqm)
{
# ifndef VXWORKS
  ExportListStruct *el = getExportListItem(eqm);
  int i, cc = 0;
  FILE *fp = NULL;
  char fn[256];

  if (el == NULL) ccerr(non_existent_elem);
  if (el->nRegionItems == 0 || el->regionsLst == NULL) ccerr(not_applicable);
  fixDbFileName(fn,eqm,REGIONRULES_FILE,FALSE);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Pattern, Region, Code\n");
  for (i=0; i<el->nRegionItems; i++) 
  {
    if (strlen(el->regionsLst[i].reg) > 0 && strcmp(el->regionsLst[i].reg,"0") != 0 && el->regionsLst[i].code >= 0)
      fprintf(fp,"%.64s, %.16s, %d\n",el->regionsLst[i].dev,el->regionsLst[i].reg,el->regionsLst[i].code);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
# else
  eqm = eqm;
  return not_implemented;
# endif
}
#define REGION_UNMODIFIED 0
#define REGION_MODIFIED 1
#define REGION_APPENDED 2
int chkInRegionsList(char *eqm,DeviceRegionItem *item)
{
  ExportListStruct *el = getExportListItem(eqm);
  int i, result=REGION_UNMODIFIED;
  if (el == NULL) return REGION_UNMODIFIED;
  for (i=0; i<el->nRegionItems; i++)
  {
    if (strnicmp(el->regionsLst[i].dev,item->dev,DEVICE_NAME_SIZE)) continue;
    if (strncmp(el->regionsLst[i].reg,item->reg,TAG_NAME_SIZE))
    {
      strncpy(el->regionsLst[i].reg,item->reg,TAG_NAME_SIZE);
      result = REGION_MODIFIED;
    }
    if (el->regionsLst[i].code != item->code)
    {
      el->regionsLst[i].code = item->code;
      result = REGION_MODIFIED;
    }
    return result;
  }
  return REGION_APPENDED;
}
#define DRLSTBUFCAPACITY 100
DeviceRegionItem *_drlstbuf=NULL;
int _drlstbufsiz = DRLSTBUFCAPACITY;
int _drlstsiz = 0;
int getRegionRulesFromCAS(char *eqm,DeviceRegionItem **lst,int *siz)
{
  ExportListStruct *el = getExportListItem(eqm);
  DeviceRegionItem *rlst=NULL;
  int n = 0, cc = 0;
  char devName[256], *casctx="", *ctx, *subs;
  DTYPE dout;
  static char last_ctx[CONTEXT_NAME_SIZE] = "";
  if (el == NULL) return non_existent_elem;
  if (!strcmp(eqm,SRVEQM_NAME)) return not_applicable;
  if (gSystemRunningStandAlone) return not_applicable;
  DTYPEZERO(dout);
  ctx = el->EqmContext; 
  if (strlen(ctx) == 0) ctx = gDeviceContext;
  if (strlen(ctx) == 0) return illegal_context; /* context still not known */
  if (_drlstbuf == NULL)
  { /* regions list buffer not yet allocated */
    if ((_drlstbuf=(DeviceRegionItem *)SystemCalloc(_drlstbufsiz,sizeof(DeviceRegionItem))) == NULL)
      return out_of_server_memory;
  }
  subs = el->EqmSubSystem;
  if (strlen(subs) == 0) subs = gDeviceSubSystem;
  casctx = getPrimaryContext(ctx,subs);
  if (!strnicmp(casctx,last_ctx,CONTEXT_NAME_SIZE))
  { /* return cached results ... */
    feclog("returning cached region rules for context %.32s",casctx);
    *siz = _drlstsiz;
    *lst = _drlstbuf;
    return 0;
  }
  sprintf(devName,"/%.32s/%.32s/%.32s",casctx,CAS_SRV,el->EqmExportName);
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &n;
  if ((cc=ExecLinkEx(devName,"NREGIONRULES",&dout,NULL,CA_READ,200)) != 0) 
  {
    n = 0;
    ccerr(cc);
  }
  if (n <= 0) ccerr(0);
  if (n == 0) ccerr(not_defined);
  feclog("acquire region rules for context %.32s from CAS",casctx);
  if (n > _drlstbufsiz)
  {
    if ((rlst=(DeviceRegionItem *)SystemCalloc(n,sizeof(DeviceRegionItem))) != NULL)
    {
      SystemFree(_drlstbuf);
      _drlstbuf = rlst;
      _drlstbufsiz = n;
    }
    else
    {
      feclogEx(TINE_LOGLEVEL_WARN, "could not allocate %d bytes for device regions buffer",n*sizeof(DeviceRegionItem));
      n = _drlstbufsiz;
    }
  }
  dout.dArrayLength = n;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"RgnRule");
  dout.data.vptr = _drlstbuf;
  if ((cc=ExecLink(devName,"REGIONRULES",&dout,NULL,CA_READ|CA_MUTABLE)) == 0)
  {
    n = (int)dout.dArrayLength;
    *siz = n;
    *lst = _drlstbuf;
  }
err:
  _drlstsiz = n > 0 ? n : 0;
  strncpy(last_ctx,casctx,CONTEXT_NAME_SIZE);
  return cc;
}
int GetValidDeviceRegion(char *eqm,char *regionTag)
{ /* just find the first region name match */
  int i;
  ExportListStruct *el = getExportListItem(eqm);
  if (regionTag == NULL || strlen(regionTag) == 0 || el == NULL) goto err;
  for (i=0; i<el->nRegionItems; i++)
  {
    if (strnicmp(regionTag,el->regionsLst[i].reg,TAG_NAME_SIZE)) continue;
    return el->regionsLst[i].code;
  }
err:
  return DEV_REGION_UNKNOWN;
}
char *GetRegionTag(char *eqm,int code)
{
  int i;
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) goto err;
  for (i=0; i<el->nRegionItems; i++)
  {
    if (el->regionsLst[i].code == code) return el->regionsLst[i].reg;
  }
err:
  return "unknown";
}
int sortDeviceRegions(char *eqm)
{ /* sorted as 'longer string first' */
  int i, ilen, k, klen, kidx, sort_needed=FALSE;
  DeviceRegionItem tmp;
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return FALSE;

  for (i=0; i<el->nRegionItems; i++)
  {
    if (strlen(el->regionsLst[i].reg) == 0 || !strcmp(el->regionsLst[i].reg,"0") || el->regionsLst[i].code < 0) 
    { /* bad entry : remove it */
      if (i < el->nRegionItems-1) el->regionsLst[i] = el->regionsLst[i+1];
      el->nRegionItems--; i--;
      sort_needed = TRUE;
    }
  }
  for (i=0; i<el->nRegionItems; i++)
  { /* assume these are almost sorted */
    ilen = (int)strlen(el->regionsLst[i].dev); 
    kidx = 0;
    for (k=i+1; k<el->nRegionItems; k++)
    {
      klen = (int)strlen(el->regionsLst[k].dev);
      if (klen > ilen)
      { 
        kidx = k;
        ilen = klen;
      }
    }
    if (kidx > i)
    { /* swap */
      tmp = el->regionsLst[kidx];
      memmove(&el->regionsLst[i+1],&el->regionsLst[i],(kidx-i)*sizeof(DeviceRegionItem));
      el->regionsLst[i] = tmp;
      sort_needed = TRUE;
    }
  }
  return sort_needed;
}
int GetRegionForDevice(char *eqm,char *regionTag,char *devName)
{
  ExportListStruct *el = getExportListItem(eqm);
  int i,code;
  if (el == NULL) return DEV_REGION_UNKNOWN;
  if ((code=GetValidDeviceRegion(eqm,regionTag)) != DEV_REGION_UNKNOWN) return code;
  if (devName == NULL) goto err;
  for (i=0; i<el->nRegionItems; i++)
  { /* sorted as 'longer string first' */
    if (wc_stricmp(el->regionsLst[i].dev,devName)) continue;
    return el->regionsLst[i].code;
  }
err:
  return (el != NULL) ? el->region : DEV_REGION_UNKNOWN;
}
int gIgnoreRegionRules = FALSE;
TINE_EXPORT void SetIgnoreRegionRules(int value)
{
  gIgnoreRegionRules = value;
}
int getRegionRules(char *eqm)
{
  int dump=FALSE, rlstsiz=0, i, n, nnew=0, cc;
  DeviceRegionItem *rlst=NULL, *tlst;    
  ExportListStruct *el = getExportListItem(eqm);

  if (el == NULL) return non_existent_elem;
  if (gIgnoreRegionRules || gIsCentralAlarmServer || gIgnoreCentralServices) return not_applicable;
  if (el->nRegionItems > 0) return already_assigned;
  if (strlen(el->EqmContext) == 0 && strlen(gDeviceContext) == 0) return illegal_context;
  if ((cc=getRegionRulesFromFile(eqm)) == 0)
  { /* a local rules db was found */
    if (sortDeviceRegions(eqm)) dump = TRUE;
  }
  if ((cc=getRegionRulesFromCAS(eqm,&rlst,&rlstsiz)) == 0)
  { /* CAS came back with results: merge and dump */
    if (el->nRegionItems == 0)
    { /* don't have anything from the local disk: use the CAS list directly */
      if (rlstsiz > 0)
      {
        if ((el->regionsLst=(DeviceRegionItem *)SystemCalloc(rlstsiz,sizeof(DeviceRegionItem))) == NULL)
        {
          feclogEx(TINE_LOGLEVEL_ERROR, "could allocate device regions area (%d items)!",rlstsiz);
          ccerr(out_of_local_memory);
        }
        memcpy(el->regionsLst,rlst,rlstsiz*sizeof(DeviceRegionItem));
      }
      el->nRegionItems = rlstsiz;
      sortDeviceRegions(eqm);
      dump = TRUE;
    }
    else
    { /* anything we don't know about ? */
      int *skip = (int *)alloca(rlstsiz*sizeof(int));
      if (skip == NULL) return out_of_local_memory;
      memset(skip,0,rlstsiz*sizeof(int));
      for (i=0; i<rlstsiz; i++)
      {
        switch (chkInRegionsList(eqm,&rlst[i]))
        {
          case REGION_UNMODIFIED:
            skip[i] = TRUE;
            break;
          case REGION_MODIFIED:
            skip[i] = TRUE;
            dump = TRUE;
            break;
          case REGION_APPENDED:
            nnew++;
            break;
        }
      }
      if (nnew > 0)
      { /* re-map the storage */
        if ((tlst=(DeviceRegionItem *)realloc(el->regionsLst,(el->nRegionItems+nnew)*sizeof(DeviceRegionItem))) == NULL)
        {
          feclogEx(TINE_LOGLEVEL_ERROR, "could not reallocate device regions area!");
          ccerr(out_of_local_memory);
        }
        n = el->nRegionItems;
        el->regionsLst = tlst; 
        el->nRegionItems += nnew;
        for (i=0; i<rlstsiz && n<el->nRegionItems; i++)
        {
          if (!skip[i] && strlen(rlst[i].dev) > 0)
          { /* append to the list */
            el->regionsLst[n++] = rlst[i];
          }
          else
          { /* skip this one (not a new entry) */
            el->nRegionItems--;
          }
        }
      }
      sortDeviceRegions(eqm);
    }
    sortDeviceRegions(eqm);
    dump = TRUE;
  }
  if (dump) dumpRegionRulesToFile(eqm);
err:
  return cc;
}
int applyRegionRules(ExportListStruct *el)
{
  int i;
  if (el == NULL) return non_existent_elem;
  if (gIgnoreRegionRules) return not_applicable;
  getRegionRules(el->EqmName);
  el->region = GetValidDeviceRegion(el->EqmName,el->EqmRegion);
  if (el->EqmDevLst == NULL) return not_initialized;
  for (i=0; i<el->EqmNumDevices; i++)
  {
    if (el->EqmDevLst[i].name == NULL || strlen(el->EqmDevLst[i].name) == 0) continue;
    el->EqmDevLst[i].region = GetRegionForDevice(el->EqmName,NULL,el->EqmDevLst[i].name);
  }
  return 0;
}
#else
int GetRegionForDevice(char *eqm,char *regionTag,char *devName)
{
  eqm = eqm; devName = devName;
  return atoi(regionTag);
}
int GetValidDeviceRegion(char *eqm,char *regionTag)
{
  eqm = eqm;
  return atoi(regionTag);
}
#endif /* SMALL_TINE_LIB */
_CSVCOLUMN csvDevicesColumns[] =
{
  { "DEVICE_NUMBER","0",0,CF_INT32,1,(BYTE *)OFFSETIN(DEVIS,devnum),NULL,0 },
  { "DEVICE_MASK","0",0,CF_INT32,1,(BYTE *)OFFSETIN(DEVIS,devmask),NULL,0 },
  { "DEVICE_OFFLINE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(DEVIS,devoffline),NULL,0 },
  { "DEVICE_REGION","",0,CF_TEXT,TAG_NAME_SIZE,(BYTE *)OFFSETIN(DEVIS,devregion),NULL,0 },
  { "DEVICE_ZPOS","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(DEVIS,zpos),NULL,0 },
  { "DEVICE_NAME","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(DEVIS,devname),NULL,0 },
  { "DEVICE_DESCRIPTION","",0,CF_TEXT,DEVICE_DESC_SIZE,(BYTE *)OFFSETIN(DEVIS,devdesc),NULL,0 },
  { "DEVICE_LOCATION","",0,CF_TEXT,DEVICE_LOCATION_SIZE,(BYTE *)OFFSETIN(DEVIS,devloc),NULL,0 },
  { "REDIRECTION","",0,CF_TEXT,DEVICE_REDIR_SIZE,(BYTE *)OFFSETIN(DEVIS,devrdr),NULL,0 },
  { "PROPERTY_LIST","",0,CF_TEXT,32,(BYTE *)OFFSETIN(DEVIS,prplst),NULL,0 },
};
CSVDB csvDevicesDb =
{
  csvDevicesColumns,
  sizeof(csvDevicesColumns)/sizeof(csvDevicesColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(DEVIS),
  0
};
int getDevicesFromFile(char *fname,DEVIS **devis,int *size)
{
  int cc;
  if ((cc=csvReadFile(fecDbPath,fname,&csvDevicesDb,(void **)devis)) != 0) ccerr(cc);
  if (size != NULL) *size = csvDevicesDb.siz;
  feclog("file %s found",fname);
err:
  return cc;
}
char *makeDbFileName(char *eqm,char *fname,int usesubdir)
{
  static char fn[64];
  char subdir[32];
  if (usesubdir && eqm != NULL)
  {
#   ifdef FS_DELIMITER
    sprintf(subdir,"%.6s%c",eqm,FS_DELIMITER);
#   else
    eqm = eqm;
    subdir[0] = 0;
#   endif
  }
  else
  {
    subdir[0] = 0;
  }
  sprintf(fn,"%s%s",subdir,fname);
  return fn;
}
char *makeDeviceFileName(char *eqm,int usesubdir,int decorate)
{
  char dname[32];

  if (decorate == 0)
  {
    strcpy(dname,"devices.csv");
  }
  else
  {
    memset(dname,0,sizeof(dname));
    strncpy(dname,eqm,EQM_NAME_SHORTSIZE);
    strlwr(dname);
#   if !defined(FS_SHORTNAMES)
    strcat(dname,"-devices.csv");
#   else
    strcat(dname,"-d.csv");
#   endif
  }
  return makeDbFileName(eqm,dname,usesubdir);
}
int GetDeviceAclInfo(char *eqm,int devnr)
{
# ifndef SMALL_TINE_LIB
  ExportListStruct *el = getExportListItem(eqm);
  char *fn;
  int cc = 0, devUsersFile = -1, devNetsFile = -1;
  int rc = 0, acc = 0;
  int i, istart, istop;

  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if (isUsingXMLFecDatabase) 
  {
    devUsersFile = devNetsFile = 1;
  }
  else
  {
    setAclFileInfo(eqm,&devUsersFile,&devNetsFile);
  }
  if (devnr < 0)
  { /* signal for the entire range of registered devices */
    istart = 0; istop = el->EqmNumDevices;
  }
  else
  {
    if (devnr >= el->EqmNumDevices) ccerr(illegal_device);
    istart = devnr, istop = devnr + 1;
  }
  if (devUsersFile >= 0)
  { /* there are files of this sort on the path */
    AclInfoType acl;
    for (i=istart; i<istop; i++)
    {
      memset(&acl,0,sizeof(AclInfoType));
      fn = makeListFileName(eqm,devUsersFile,el->EqmDevLst[i].name,USERS_FILE);
      if ((rc=getRegUsrsFromFile(eqm,el->EqmDevLst[i].name,NULL,&acl,fn,"WRITE",&acc)) != 0)
      { /* try the 'default' file */
        fn = makeListFileName(eqm,devUsersFile,"device-default",USERS_FILE);
        rc = getRegUsrsFromFile(eqm,el->EqmDevLst[i].name,NULL,&acl,fn,"WRITE",&acc);
      }
      if (rc == 0)
      { /* got the info */
        if (el->EqmDevLst[i].aclLst == NULL && 
            (el->EqmDevLst[i].aclLst=(AclInfoType *)SystemCalloc(1,sizeof(AclInfoType))) == NULL)
          ccerr(out_of_server_memory);
        memcpy(el->EqmDevLst[i].aclLst,&acl,sizeof(AclInfoType));
        el->chkAcl |= (acc<<8);
      }
    }
  }
  if (devNetsFile >= 0)
  { /* there are files of this sort on the path */
    int nr;
    SCKACL *lst;
    for (i=istart; i<istop; i++)
    {
      lst = NULL; nr = 0;
      fn = makeListFileName(eqm,devNetsFile,el->EqmDevLst[i].name,IPNETS_FILE);
      if ((rc=getRegNetsFromFile(eqm,el->EqmDevLst[i].name,NULL,fn,NETS_ACCESS,&nr,&lst)) != 0)
      { /* try the 'default' file */
        fn = makeListFileName(eqm,devNetsFile,"device-default",IPNETS_FILE);
        rc = getRegNetsFromFile(eqm,el->EqmDevLst[i].name,NULL,fn,NETS_ACCESS,&nr,&lst);
      }
      if (rc == 0)
      { /* got the info */
        if (el->EqmDevLst[i].aclLst == NULL && 
            (el->EqmDevLst[i].aclLst=(AclInfoType *)SystemCalloc(1,sizeof(AclInfoType))) == NULL)
          ccerr(out_of_server_memory);
        el->EqmDevLst[i].aclLst->nNets = nr;
        el->EqmDevLst[i].aclLst->nets = lst;
        el->chkAcl |= DEV_CHECK_NET;
      }
    }
  }
err:
  if (cc)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "GetDeviceAclInfo for module %.6s failed: %.32s",eqm,cc2str(cc));
  }
  return cc;
# else
  eqm = eqm;
  devnr = devnr;
  return 0;
# endif
}

FILE *getMfFile(char *eqm, char *mode, char *fileName)
{
# ifndef VXWORKS
  char mfn[256];
  FILE* fp = NULL;
  fixDbFileName(mfn, eqm, fileName,TRUE);
  if ((fp=fopen(mfn, mode)) != NULL) return fp;
  if (eqm != NULL)
  {
    fixDbFileName(mfn, NULL, fileName,TRUE);
    return fopen(mfn, mode);
  }
# else
  eqm = eqm;
  mode = mode;
  fileName = fileName;
# endif
  return NULL;
}
int dumpPropertiesManifest(char *eqm, int to_stdout)
{
  int cc = 0;
# if defined(FS_DELIMITER) && !defined(FS_RDONLY)
# define prpmfhdr "NAME, ID, DESCRIPTION, ALIAS, DEVICE_LIST"
  char membersfn[128];
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  FILE *fp = NULL, *memfp = NULL;
  int i, idx;
  if (el == NULL) ccerr(non_existent_elem);
  if ((fp=getMfFile(eqm, "w", PROPERTIES_MANIFEST)) == NULL) ccerr(file_error);
  if (HAS_PRPQRYFCN(el))
  {
    fprintf(fp, "Uses a property query function\n");
    ccerr(0);
  }
  fprintf(fp, "%s\n", prpmfhdr);
  for (idx = 0; idx < PROPERTY_HASH_SIZE + 1; idx++)
  {
    for (prp=el->EqmPrpLst[idx]; prp != NULL; prp = prp->next)
    {
      if (prp->numDevices > 0 && prp->devNames != NULL)
      {
        sprintf(membersfn, "%.64s-members.csv", prp->prpName);
        if ((memfp=getMfFile(eqm, "w", membersfn)) != NULL)
        {
          fprintf(memfp, "NAME, NUMBER\n");
          for (i=0; i < prp->numDevices; i++)
          {
            fprintf(memfp, "%.64s, %d\n", prp->devNames[i].name, i);
          }
          fclose(memfp);
        }
      }
      else
      {
        membersfn[0] = 0;
      }
      fprintf(fp, "%.64s, %d, %.64s, %.64s, %.128s\n",
        prp->prpName, prp->prpId, prp->prpDescription, prp->prpAlias, membersfn);
    }
  }

  if (to_stdout) dbglog("local property info written");
err:
  if (fp != NULL) fclose(fp);
# endif
  return cc;
}
int dumpDevicesManifest(char *eqm, int to_stdout)
{
  int cc = 0;
# if defined(FS_DELIMITER) && !defined(FS_RDONLY)
# define devmfhdr "NAME, NUMBER, MASK, REGION, ZPOS, DESCRIPTION, LOCATION"
  ExportListStruct *el = getExportListItem(eqm);
  ExportDeviceType *dv = NULL;
  FILE *fp = NULL;
  int i;
  if (el == NULL) ccerr(non_existent_elem);
  if ((fp = getMfFile(eqm, "w", DEVICES_MANIFEST)) == NULL) ccerr(file_error);
  fprintf(fp, "%s\n", devmfhdr);
  for (i=0; i < el->EqmNumDevices; i++) 
  {
    if (el->EqmDevLst == NULL) continue;
    if ((dv=&el->EqmDevLst[i]) == NULL) continue;
    if (strlen(dv->name) == 0) continue; /* don't dump the 'empties' */
    fprintf(fp, "%.64s, %d, 0x%x, %d, %g, %.64s, %.64s\n",
      dv->name, dv->number, dv->mask, dv->region, dv->zpos, PSTRP(dv->desc), PSTRP(dv->loc));
  }
  if (to_stdout) dbglog("local device info written");
err:
  if (fp != NULL) fclose(fp);
# endif
  return cc;
}
_CSVCOLUMN csvDevMfColumns[] =
{
  { "NAME","",0,CF_TEXT,64,(BYTE*)OFFSETIN(DEVMF,name),NULL,0 },
  { "NUMBER","",0,CF_INT32,1,(BYTE*)OFFSETIN(DEVMF,num),NULL,0 },
  { "MASK","",0,CF_INT32,1,(BYTE*)OFFSETIN(DEVMF,msk),NULL,0 },
  { "REGION","",0,CF_INT32,1,(BYTE*)OFFSETIN(DEVMF,rgn),NULL,0 },
  { "ZPOS","",0,CF_FLOAT,1,(BYTE*)OFFSETIN(DEVMF,zpos),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE*)OFFSETIN(DEVMF,dsc),NULL,0 },
  { "LOCATION","",0,CF_TEXT,64,(BYTE*)OFFSETIN(DEVMF,loc),NULL,0 },
};
CSVDB csvDevMfDb =
{
  csvDevMfColumns,
  sizeof(csvDevMfColumns) / sizeof(csvDevMfColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(DEVMF),
  0
};
_CSVCOLUMN csvPrpMfColumns[] =
{
  { "NAME","",0,CF_TEXT,64,(BYTE*)OFFSETIN(PRPMF,name),NULL,0 },
  { "ID","",0,CF_INT32,1,(BYTE*)OFFSETIN(PRPMF,id),NULL,0 },
  { "ALIAS","",0,CF_FLOAT,1,(BYTE*)OFFSETIN(PRPMF,alias),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE*)OFFSETIN(PRPMF,dsc),NULL,0 },
  { "DEVICE_LIST","",0,CF_TEXT,64,(BYTE*)OFFSETIN(PRPMF,memfn),NULL,0 },
};
CSVDB csvPrpMfDb =
{
  csvPrpMfColumns,
  sizeof(csvPrpMfColumns) / sizeof(csvPrpMfColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(PRPMF),
  0
};
int GetDevicesFromManifest(char *eqm, char *prp, DEVMF **devs, int *listSize)
{
  PRPMF *prps = NULL;
  int cc=0;
  char dn[256], *fn = DEVICES_MANIFEST;
  if (eqm == NULL || devs == NULL || listSize == NULL) return argument_list_error;
  sprintf(dn, "%.128s%.6s%ccache%c", fecDbPath, eqm, FS_DELIMITER, FS_DELIMITER);
  if (prp != NULL)
  {
    ExportPropertyListStruct *pl = GetPropertyListStruct(eqm, prp, NULL);
    if (pl != NULL && pl->devNames != NULL && pl->numDevices > 0)
    { /* this one has a specific list ... */
      int i, nprps = 0;
      if (GetPropertiesFromManifest(eqm, &prps, &nprps) == 0)
      {
        for (i = 0; i < nprps && strncmp(prp, prps[i].name, PROPERTY_NAME_SIZE); i++);
        if (i < nprps && strlen(prps[i].memfn) > 0) fn = prps[i].memfn;
      }
    }
  }
  if ((cc=csvReadFile(dn, fn, &csvDevMfDb, (void **)devs)) != 0) ccerr(cc);
  *listSize = csvDevMfDb.siz;
err:
  SystemFree(prps);
  return cc;
}
int GetPropertiesFromManifest(char *eqm, PRPMF **prps, int *listSize)
{
  int cc = 0;
  char dn[256];
  if (eqm == NULL || prps == NULL || listSize == NULL) return argument_list_error;

  sprintf(dn, "%.128s%.6s%ccache%c", fecDbPath, eqm, FS_DELIMITER, FS_DELIMITER);
  if ((cc = csvReadFile(dn, PROPERTIES_MANIFEST, &csvPrpMfDb, (void **)prps)) != 0) ccerr(cc);
  *listSize = csvPrpMfDb.siz;
err:
  return cc;
}
NAME64 *GetEqpDeviceNames(char *eqm,char *prpname,ExportDeviceType **devs,int *listSize)
{
  ExportDeviceType *devlst = NULL;
  NAME64 *nlst = NULL;
  int i,k,idx,cc=0,decorate=0,usesubdir=TRUE,size=0,num=0;
  char *fname,*tgt="[unknown]",msg[256];
  PrpRedirBlk prd;
  EqmDevicePropertyList *prplist = NULL;
  DEVIS *devis = NULL;
  ExportListStruct *el = NULL;

  if (listSize == NULL || *listSize == 0) return NULL;
  num = *listSize; 

  if (eqm != NULL && prpname == NULL)
  { /* common device list */
    if ((cc=makeEqmDevLst(&devlst,num)) != 0) ccerr(cc);
    nlst = (NAME64 *)devlst->name;
    if (gTineConfigurationCoded) ccerr(not_implemented);
#   ifndef SMALL_TINE_LIB
    if (isUsingXMLFecDatabase)
    {
      size = populateDEVIS(gFecName,eqm,gFecInfoList,&devis);
      if (size <= 0) ccerr(-size);
    }
    else
#   endif
    {
      if (!gScanForDeviceLists) ccerr(no_such_file);
retryEqmDeviceFile:
      fname = makeDeviceFileName(eqm,usesubdir,decorate);
      if ((cc=getDevicesFromFile(fname,&devis,&size)) == no_such_file && decorate++ == 0) goto retryEqmDeviceFile;
      if (cc == no_such_file && usesubdir)
      {
        usesubdir = FALSE;
        decorate = 0;
        goto retryEqmDeviceFile;
      }
      if (cc != 0) ccerr(cc);
      if (devis == NULL) ccerr(code_failure);
    }
    tgt = eqm;
  }
  else if (prpname != NULL)
  { /* property-specific device list */
    tgt = prpname;
    if (gTineConfigurationCoded) ccerr(not_implemented);
#   ifndef SMALL_TINE_LIB
    if (isUsingXMLFecDatabase)
    {
      size = populatePRPDEVIS(gFecName,eqm,prpname,gFecInfoList,&devis);
      if (size <= 0) ccerr(-size);
    }
    else
#   endif
    {
      if (!gScanForNameLists) ccerr(no_such_file);
retryPrpDeviceFile:
      fname = makeListFileName(eqm,usesubdir,prpname,"names.csv");
      cc = getDevicesFromFile(fname,&devis,&size);
      if (cc == no_such_file && usesubdir)
      {
        usesubdir = FALSE;
        goto retryPrpDeviceFile;
      }
      if (cc != 0) ccerr(cc);
    }
    if (devis == NULL) ccerr(code_failure);
    if (size != num) *listSize = num = size;
    if ((nlst=(NAME64 *)SystemCalloc(num,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
  }
  else
    ccerr(argument_list_error);

  for (i=0; i<size; i++)
  { /* device redirection ? property lists ? */
    idx = devis[i].devnum;
    if (idx < 0 || idx >= num) continue;
    strncpy(nlst[idx].name,devis[i].devname,DEVICE_NAME_SIZE);
    if (devlst != NULL) 
    {
      devlst[idx].number = idx;
      devlst[idx].mask = devis[i].devmask;
      if (devis[i].devoffline) devlst[idx].mask |= DEVICE_OFFLINE;
      devlst[idx].zpos = devis[i].zpos;
      devlst[idx].region = GetRegionForDevice(eqm,devis[i].devregion,devis[i].devname);
      if (strlen(devis[i].devdesc) > 0)
      {
        if (devlst[idx].desc == NULL &&
            (devlst[idx].desc=(char *)SystemCalloc(1,64)) == NULL) ccerr(out_of_server_memory);
        strncpy(devlst[idx].desc,devis[i].devdesc,64); 
      }
      if (strlen(devis[i].devloc) > 0)
      {
        if (devlst[idx].loc == NULL &&
            (devlst[idx].loc=(char *)SystemCalloc(1,64)) == NULL) ccerr(out_of_server_memory);
        strncpy(devlst[idx].loc,devis[i].devloc,64); 
      }
      if (strlen(devis[i].devrdr) > 0)
      {
        ParseRedirector(devis[i].devrdr,&prd); /* only the device server name is relevant */
        sprintf(msg, "Redirection for device %.64s", devis[i].devname);
        if (assertRedirectionValid(msg,eqm,prd.prpRedirection))
        {
          if (devlst[idx].rdr == NULL &&
              (devlst[idx].rdr=(char *)SystemCalloc(1,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE)) == NULL)
            ccerr(out_of_server_memory);
          strncpy(devlst[idx].rdr,&prd.prpRedirection[PRP_REDIR_CTX_OFFSET],CONTEXT_NAME_SIZE); 
          strncpy(&devlst[idx].rdr[CONTEXT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_SRV_OFFSET],EXPORT_NAME_SIZE); 
          strncpy(&devlst[idx].rdr[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_DEV_OFFSET],DEVICE_NAME_SIZE); 
        }
      }
      if (strlen(devis[i].prplst) > 0)
      {        
        prplist = NULL;
        for (k=0; k<size; k++) 
        { /* check if reference already in the list */
          if (devlst[k].prps == NULL) continue;
          if (strnicmp(devlst[k].prps->filename,devis[i].prplst,64)) continue;
          prplist = devlst[k].prps;
          break;
        }
        if (prplist == NULL) getEqpPropertyList(eqm,devis[i].prplst,&prplist);
        devlst[idx].prps = prplist;
        if (el == NULL) 
        { /* make sure this is set */
          if ((el=getExportListItem(eqm)) != NULL) el->hasDevQry = TRUE;
        }
      }
    }
  }
  if (devs != NULL) *devs = devlst;
err:
  SystemFree(devis);
  if (cc)
  { /* then devs not assigned ! */
    if (devlst != NULL) SystemFree(devlst[0].name);
    SystemFree(devlst);
  }
  if (cc != no_such_file && cc != not_implemented)
  {
    feclog("device file for %.64s : %.32s",tgt, cc ? cc2str(cc) : "found");
  }
  return nlst;
}
int isPropertyRedirected(ExportListStruct *el,ExportPropertyListStruct *pls,char *devName)
{
  PrpRedirBlk *prd = NULL;
  int redir=FALSE;
  if (pls->isReserved) return FALSE;
  if (pls->prd == NULL) return FALSE;
  for (prd=pls->prd; prd != NULL; prd=prd->next)
  { /* property is redirected, but is it true for the device name given ? */
    if (prd->matchAnyDevice) redir = TRUE;
    else if (prd->prpRedirection[PRP_REDIR_DEV_OFFSET] == 0) redir = TRUE;
    else if (!strnicmp(&prd->prpRedirection[PRP_REDIR_DEV_OFFSET],devName,DEVICE_NAME_SIZE)) redir = TRUE;
    if (redir)
    { /* there's a match, redirect it */
      memcpy(el->EqmLastErrorString,prd->prpRedirection,STATUS_SIZE);
      return TRUE;
    }
  }
  return FALSE;
}
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName,ExportPropertyListStruct **prp)
{
  ExportPropertyListStruct *pls;
  int idx;
  char *c,fld[FIELD_NAME_SIZE+8],prpnam[PROPERTY_NAME_SIZE+8];

  if (prp != NULL) *prp = NULL;
  if (el == NULL || prpName == NULL) return argument_list_error;
  if (!strcmp(prpName,"DEVICES") || !strcmp(prpName,"NDEVICES")) return not_accepted;
  if (el->EqmDevLst != NULL)
  { /* maybe there's a device-specific redirection list ? */
    if ((idx=GetDeviceNumberEx(el->EqmName,devName,NULL)) >= 0 && idx < el->EqmNumDevices)
    { /* given device was found  */
      if (el->EqmDevLst[idx].rdr != NULL && strlen(&el->EqmDevLst[idx].rdr[CONTEXT_NAME_SIZE]) > 0)
      { /* redirect, regardless of property */
        SetRedirectionParameters(el->EqmName,el->EqmDevLst[idx].rdr,
                                 &el->EqmDevLst[idx].rdr[CONTEXT_NAME_SIZE],
                                 &el->EqmDevLst[idx].rdr[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE],NULL);
        return server_redirection;
      }
    }
  }
  if ((pls=GetPropertyListStruct(el->EqmName,prpName,devName)) != NULL)
  {
    if (pls->isReserved) return not_registered;
    if (prp != NULL) *prp = pls;
    if (isPropertyRedirected(el,pls,devName)) return server_redirection;
    return 0;
  }
  /* last chance: maybe decorated with a struct or bitfield 'field' */
  strncpy(prpnam,prpName,PROPERTY_NAME_SIZE); prpnam[PROPERTY_NAME_SIZE] = 0;
  if ((c=strrchr(prpnam,'.')) != NULL)
  { /* check for single-field bitfield or struct requests */
    if (IsMetaProperty(prpnam,NULL,NULL,FALSE)) return not_accepted; /* don't let this thru here, wait for next pass */
    *c = 0;
    if ((pls=GetPropertyListStruct(el->EqmName,prpnam,devName)) != NULL)
    { /* so far, so good */
      if (isPropertyRedirected(el,pls,devName)) return server_redirection;
      strncpy(prpnam,prpName,PROPERTY_NAME_SIZE); /* copy the full string back */
      if (isBitfieldProperty(NULL,pls->prpFormatTag.name,prpnam,fld))
      { /* a recognized bitfield acquisition */
        if (prp != NULL) *prp = pls;
        memcpy(el->EqmLastErrorString,pls->prpFormatTag.name,TAG_NAME_SIZE);
        return has_bitfield_tag;
      }
      if (isStructProperty(pls->prpFormatTag.name,prpnam,fld))
      { /* a recognized tagged structure acquisition */
        if (prp != NULL) *prp = pls;
        memcpy(el->EqmLastErrorString,pls->prpFormatTag.name,TAG_NAME_SIZE);
        return has_structure_tag;
      }
    }
    if ((idx=GetStockPropertyId(prpnam)) >= 0)
    { /* maybe a stock property structure field ? */
      strncpy(prpnam,prpName,PROPERTY_NAME_SIZE); /* copy the full string back */
      if (isStructProperty(gStockProperty[idx].prpStructTag,prpnam,fld))
      { /* a recognized tagged structure acquisition */
        if (prp != NULL) *prp = NULL;
        memcpy(el->EqmLastErrorString,gStockProperty[idx].prpStructTag,TAG_NAME_SIZE);
        return has_structure_tag;
      }
    }
  }
  return illegal_property;
}
TINE_EXPORT int SetPropertySubscriptionRenewalLength(char *eqm,char *prpName,int value)
{
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,prpName,NULL);
  if (eqm == NULL || prpName == NULL) return argument_list_error;
  if (prp == NULL) return invalid_property;
  if (value < 20) return out_of_range;
  prp->contractRenewalLengh = value;
  feclog("renewal length for contracts to %.6s %.64s set to %d",eqm,prpName,value);
  return 0;
}
TINE_EXPORT int GetPropertySubscriptionRenewalLength(char *eqm,char *prpName,int *value)
{
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,prpName,NULL);
  if (eqm == NULL || prpName == NULL || value == NULL) return argument_list_error;
  if (prp == NULL) return invalid_property;
  *value = prp->contractRenewalLengh;
  return 0;
}
TINE_EXPORT BYTE *GetPropertyBuffer(char *eqmName,char *prpName)
{
# ifndef SMALL_TINE_LIB
  int idx;
  ExportPropertyListStruct *pls;
  ExportListStruct *el=getExportListItem(eqmName);

  if (el == NULL || prpName == NULL) return NULL;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (pls=el->EqmPrpLst[idx]; pls; pls=pls->next)
    if (!strnicmp(pls->prpName,prpName,PROPERTY_NAME_SIZE)) 
    { /* property is registered */
      return pls->prpBuf;
    }
# endif
  return NULL;
}
int setContractListBuffers(char *eqm,char *prp,BYTE *buf)
{
  int i;
  for (i=0; i<ncontracts; i++)            
  { /* is contract in list ? */
    if (ContractList[i] == NULL) continue;
    if (strncmp(ContractList[i]->contract.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strncmp(ContractList[i]->contract.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (ContractList[i]->deliveryPending > 0) return operation_busy;
    if (!ContractList[i]->unbuffered)
    {
      SystemFree(ContractList[i]->dataOut);
      ContractList[i]->dataBuf = NULL;
    }
    ContractList[i]->unbuffered = TRUE;
    ContractList[i]->dataOut = ContractList[i]->dataBuf = buf;
  }
  return 0;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int SetPropertyBuffer(char *eqmName,char *prpName,BYTE *buffer)
{
  int idx;
  ExportPropertyListStruct *prp;
  ExportListStruct *el=getExportListItem(eqmName);

  if (el == NULL) return non_existent_elem; 
  if (prpName == NULL) return argument_list_error;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
    if (!strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE)) 
    { /* property is registered */
      prp->prpBuf = buffer;
      return setContractListBuffers(eqmName,prpName,buffer);
    }
  return illegal_property;
}
#endif
char *lookupPropertyId(ExportListStruct *el,short id)
{
  ExportPropertyListStruct *prp;
  int idx;

  if (el == NULL) return (char *)NULL;
  idx = PROPERTY_HASH_INDEX(id);
  if (idx < 0 || idx >= PROPERTY_HASH_SIZE) return (char *)NULL;
  for (prp = el->EqmPrpLst[idx]; prp; prp = prp->next)
  {
    if (prp->isReserved) continue;
    if (prp->prpId == id) return prp->prpName;
  }
  return (char *)NULL;
}
int RegisterLocalAlias(char *eqmName,char *alias)
{
  int cc=0;
  ExportListStruct *el;

  if ((el=getExportListItem(eqmName)) == NULL) ccerr(non_existent_elem);
  strncpy(el->EqmAlias,alias,EQM_NAME_SHORTSIZE);
  feclog("Register local alias %.6s for %.6s",alias,eqmName);
err:
  return cc;
}

#ifndef SMALL_TINE_LIB
int foPollingInterval = FAILOVER_POLLING_RATE;
int foThreshold = FAILOVER_THRESHOLD;
int foRevertDeadband = 0;
TINE_EXPORT void SetFailoverPollingInterval(int pollingInterval)
{
  if (pollingInterval < 50) return;
  foPollingInterval = pollingInterval;
  feclog("failover polling interval set to %d msec",foPollingInterval);
}
TINE_EXPORT void SetFailoverThreshold(int errorCounts)
{
  if (errorCounts < 2) return;
  foThreshold = errorCounts;
  feclog("failover threshold set to %d error counts",foThreshold);
}
static SINT32 failoverMasterStarttime = 0;
void _cbFailoverMonitor(int id,int cc,void *ref)
{
  static int errcnt=0;
  static int last_cc=0;
  static int ismaster = FALSE;
  static SINT32 last_st = 0;
  ExportListStruct *el;
  UNUSED(id); /* avoid warning */
  if (last_cc != cc)
  {
    feclogEx(TINE_LOGLEVEL_INFO, "failover slave link status changed from %.32s to %.32s",erlst[last_cc],cc2str(cc));
    last_cc = cc;
  }
  if (cc != 0)
  { /* link down : do I need to take over ? */
    gServerCycleIdle = FALSE;
    errcnt++;
    if (errcnt > foThreshold && !ismaster)
    { /* need to take over as the master ! */
      if ((el=(ExportListStruct *)ref) != NULL)
      {        
        el->failoverState = (FAILOVER_MASTER | FAILOVER_SLAVE);
        el->registered &= ~REGISTERED_MASTER;
        ExportsRegistered = FALSE;
        ismaster = TRUE;
        clslog(el->EqmContext,"FAILOVER",gFecName,CLOG_PRIORITY_IMPORTANT,CLOG_STATUS_WARN,
          "%.32s is taking over as %.32s master",el->EqmExportName,el->EqmMaster);
      }
    }
    return;
  }
  if (ismaster)
  { /* have become the master, but now the real master is back ! */
    if ((el=(ExportListStruct *)ref) != NULL) {
      el->failoverState = FAILOVER_SLAVE;
    }
    gServerCycleIdle = TRUE;
    foRevertDeadband = thisCycleTime.tv_sec + 120;
    FlushContractTable();
    feclogEx(TINE_LOGLEVEL_INFO, "failover master is now available again: disconnect all clients");
    if (failoverMasterStarttime < last_st + 30)
    { /* appears to have been simply unresponsive for a while */
      if ((el=(ExportListStruct *)ref) != NULL && strlen(el->EqmSlaveMaster) > 0)
      {
        char devnam[FULL_NAME_MAXSIZE];
        sprintf(devnam,"/%.32s/%.32s",el->EqmContext,el->EqmSlaveMaster);
        cc = ExecLinkEx(devnam,"SRVREEXPORT",NULL,NULL,CA_WRITE|CA_NOCALLBACKS,100);
        feclog("sent SRVREEXPORT to %.64s : %.32s",devnam,cc2str(cc));        
      }
    }
  }
  if (gServerCycleIdle && foRevertDeadband > 0 && thisCycleTime.tv_sec > foRevertDeadband)
  { /* clients needing the real master have now had enough time */
    gServerCycleIdle = FALSE;
    foRevertDeadband = 0;
    feclogEx(TINE_LOGLEVEL_INFO, "revert to original failover slave conditions");
  }
  errcnt = 0;
  ismaster = FALSE;
  last_st = failoverMasterStarttime;
}
TINE_EXPORT int GetCurrentFailoverState(void) 
{
  ExportListStruct *el;
  int isFirstEqm=1;
  int currentEqmState=0;
  
  for(el=ExportList; el!=NULL; el=el->next) 
  {
    if (isFirstEqm) 
    {
      isFirstEqm = 0;
      currentEqmState = el->failoverState;
    } 
    else 
    {
      if (currentEqmState != el->failoverState) 
      {
        /* in case of differently configured eqms 
         * (one eqm master, one eqm slave): 
         * return indicator for inconsistency
         */
        return ~FAILOVER_NONE;
      }
    }
  }
  return currentEqmState;
}
void startFailoverMonitor(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  DTYPE dout;
  char devnam[FULL_NAME_MAXSIZE];
  int id, failoverErrorCount;
  char *ptr;
  static int done=FALSE;
  if (done) return;
  if (el == NULL || strlen(el->EqmSlaveMaster) == 0) return;
  sprintf(devnam,"/%.32s/%.32s",el->EqmContext,el->EqmSlaveMaster);
  dout.dArrayLength = 1; dout.dFormat = CF_INT32; dout.data.lptr = &failoverMasterStarttime; dout.dTag[0] = 0;
  id = AttachLinkEx2(devnam,LWD_PROPERTY,&dout,NULL,CA_READ,foPollingInterval,
                     _cbFailoverMonitor,CM_POLL,UNASSIGNED_CALLBACKID,(void *)el);
  if ((ptr=getenv("TINE_FAILOVER_DELAY")) != NULL && (failoverErrorCount=atoi(ptr)))
  { /* set by environment */    
    SetFailoverThreshold( failoverErrorCount);
    feclogEx(TINE_LOGLEVEL_INFO, "SetFailoverThreshold(%d) by environment TINE_FAILOVER_DELAY\n", failoverErrorCount);
  }  
  feclogEx(TINE_LOGLEVEL_INFO, "failover slave link to %.192s : %.32s",devnam,id < 0 ? erlst[-id] : "success");
  done = TRUE;
  return;
}
int joinEnsGroup(ExportListStruct *el)
{
  int cc=0,pass=0;
  char srvName[EXPORT_NAME_SIZE];
  NAME32 n32data[7];
  NAME16 *n16data = (NAME16 *)n32data;
  DTYPE d;

  if (el == NULL) return non_existent_elem;
  if (el->failoverType == FAILOVER_SLAVE) 
  { /* slaves don't join groups! */
    return 0; 
  }
  strncpy(srvName,el->EqmExportName,EXPORT_NAME_SIZE);
  if (el->failoverType == FAILOVER_MASTER)
    strncpy(srvName,el->EqmMaster,EXPORT_NAME_SIZE);
retryIllegalFormat:
  if (pass == 0)
  {
    strncpy(n32data[0].name,srvName,32); 
    strncpy(n32data[1].name,el->EqmGroup,32); 
    strncpy(n32data[2].name,el->EqmContext,32); 
    sprintf(n32data[3].name,"%d",el->groupIndex); 
    strncpy(n32data[4].name,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE); 
    strncpy(n32data[5].name,el->EqmGroupDevPrefix,XFIX_SIZE); 
    strncpy(n32data[6].name,el->EqmGroupDevPostfix,XFIX_SIZE); 
    d.dArrayLength = 7;
    d.dFormat = CF_NAME32;
    memset(d.dTag,0,TAG_NAME_SIZE);
    d.data.cptr = n32data[0].name;
  }
  else
  {
    strncpy(n16data[0].name,srvName,16); 
    strncpy(n16data[1].name,el->EqmGroup,16); 
    strncpy(n16data[2].name,el->EqmContext,16); 
    sprintf(n16data[3].name,"%d",el->groupIndex); 
    strncpy(n16data[4].name,el->EqmSubSystem,16); 
    strncpy(n16data[5].name,el->EqmGroupDevPrefix,XFIX_SIZE); 
    strncpy(n16data[6].name,el->EqmGroupDevPostfix,XFIX_SIZE); 
    d.dArrayLength = 7;
    d.dFormat = CF_NAME16;
    memset(d.dTag,0,TAG_NAME_SIZE);
    d.data.cptr = n16data[0].name;
  }
  cc = tExecLink(GENS_SRV,GENS_TAG,"#0","JOIN",&d,&d,CA_WRITE,500);
  if (cc == illegal_format && pass == 0)
  { /* legacy GENS ? */
    pass++;
    goto retryIllegalFormat;
  }
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "%.32s joining group %.32s failed : %.32s",srvName,el->EqmGroup,erlst[cc & 0xfff]);

  return cc;
}
#endif /* !SMALL_TINE_LIB */
/**
 * \internal used by server to register with the ENS
 */
int prepOutgoingAddr(int dlen,char *addrstream,FecAddrStruct *fec,ExpDataStruct *srv,FecInfoStruct *inf)
{ 
# define ADDRPKTSIZE4L (RPCFECSTRUCT_SIZE + RPCDATASTRUCT_SIZE + sizeof(RPCFecExStruct))
# define ADDRPKTSIZE4 (EXPDATASTRUCT_SIZE + FECDATASTRUCT_SIZE + sizeof(FecInfoStruct))
# define ADDRPKTSIZE6 (EXPDATASTRUCT_SIZE + FECADDRSTRUCT_SIZE + sizeof(FecInfoStruct))
  RPCFecStruct lfec;
  RPCDataStruct lsrv;
  RPCFecExStruct linf;
  FecDataStruct lfec4;
  int p;
  if (fec == NULL || srv == NULL) return argument_list_error;
  p = getProtocolFromAddressString(fec->strAdr);
  switch (dlen)
  {
    case ADDRPKTSIZE4L:
      memset(&lfec,0,sizeof(RPCFecStruct));
      strncpy(lfec.fecName,fec->fecName,FEC_NAME_SIZE);
      if (p == IPX)
      {
        memcpy(lfec.Network,&fec->netAdr.ipx,16);
      }
      else
      {
        strncpy(lfec.IP,fec->strAdr,ADDR_SIZE);
        memcpy(lfec.IPh_addr,fec->netAdr.ipv4,4);
      }
      lfec.Port[0] = (BYTE)fec->portOffset;
      lfec.Port[1] = (BYTE)(fec->portOffset >> 8);
      lfec.portOffset = (BYTE)fec->portOffset;
      lfec.Prot = fec->inetProtocol;
      memset(&lsrv,0,sizeof(RPCDataStruct));
      strncpy(lsrv.FecName,srv->FecName,FEC_NAME_SIZE);
      strncpy(lsrv.SubSystem,srv->SubSystem,SUBSYSTEM_NAME_SHORTSIZE);
      strncpy(lsrv.ExportName,srv->ExportName,EXPORT_NAME_SHORTSIZE);
      strncpy(lsrv.EqmName,srv->EqmName,EQM_NAME_SHORTSIZE+1);
      strncpy(lsrv.EqmContext,srv->EqmContext,CONTEXT_NAME_SHORTSIZE);
      memset(&linf,0,sizeof(RPCFecExStruct));
      strncpy(linf.os,inf->os,12);
      strncpy(linf.desc,inf->desc,64);
      strncpy(linf.loc,inf->loc,32);
      strncpy(linf.hdw,inf->hdw,32);
      strncpy(linf.resp,inf->resp,32);
      strncpy(linf.sub,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
#     ifdef INTEL
      memcpy(addrstream,&lsrv,RPCDATASTRUCT_SIZE);
      memcpy(&addrstream[RPCDATASTRUCT_SIZE],&lfec,RPCFECSTRUCT_SIZE);
      memcpy(&addrstream[RPCDATASTRUCT_SIZE+RPCFECSTRUCT_SIZE],&linf,sizeof(RPCFecExStruct));
#     else
      /* avoid alignment problems on some machines: */
	    _structSwap(RPCFecStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)&lfec,LTOR);
      memcpy(&addrstream[RPCDATASTRUCT_SIZE],addrstream,RPCFECSTRUCT_SIZE);
      _structSwap(RPCDataStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)&lsrv,LTOR);
      memcpy(&addrstream[RPCDATASTRUCT_SIZE+RPCFECSTRUCT_SIZE],&linf,sizeof(RPCFecExStruct));
#     endif
      break;
    case ADDRPKTSIZE4:
      toFecData(&lfec4,fec);
#     ifdef INTEL
      memcpy(&addrstream[EXPDATASTRUCT_SIZE],&lfec4,FECDATASTRUCT_SIZE);
#     else
	    _structSwap(FDSStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)&lfec4,LTOR);
      memcpy(&addrstream[EXPDATASTRUCT_SIZE],addrstream,FECDATASTRUCT_SIZE);
#     endif
      memcpy(addrstream,srv,EXPDATASTRUCT_SIZE);
      memcpy(&addrstream[EXPDATASTRUCT_SIZE+FECDATASTRUCT_SIZE],inf,sizeof(FecInfoStruct));
      break;
    case ADDRPKTSIZE6:
#     ifdef INTEL
      memcpy(&addrstream[EXPDATASTRUCT_SIZE],fec,FECADDRSTRUCT_SIZE);
#     else
	    _structSwap(FASStructStruct,(BYTE *)addrstream,0,NULL,(BYTE *)fec,LTOR);
      memcpy(&addrstream[EXPDATASTRUCT_SIZE],addrstream,FECADDRSTRUCT_SIZE);
#     endif
      memcpy(addrstream,srv,EXPDATASTRUCT_SIZE);
      memcpy(&addrstream[EXPDATASTRUCT_SIZE+FECADDRSTRUCT_SIZE],inf,sizeof(FecInfoStruct));
      break;
    default:
      return not_supported;
  }
  return 0;
}
char *sysCtxDec[] = { "TEST", "SIM", "EXT" };
#define nSysCtxDec (sizeof(sysCtxDec)/sizeof(sysCtxDec[0]))
int isSystemContextDecoration(char *dec)
{
  int i=0;
  if (dec == NULL) return FALSE;
  if (*dec == '.') dec++;
  for (i=0; i<(int)nSysCtxDec; i++) 
    if (!stricmp(dec,sysCtxDec[i])) return TRUE;
  return FALSE;
}
char *getPrimaryContext(char *ctx,char *sub)
{ 
  static char prmctx[CONTEXT_NAME_SIZE+4] = "";
  char *c;
  int cc = 0;
  if (ctx == NULL || sub == NULL) ccerr(argument_list_error);
  strncpy(prmctx, ctx, CONTEXT_NAME_SIZE); prmctx[CONTEXT_NAME_SIZE] = 0;
  if ((c=strchr(prmctx,'.')) == NULL) ccerr(0);
  if (strlen(sub) == 0) ccerr(0);
  *c++ = 0; /* remove decoration from context */
  if (strnicmp(c,sub,SUBSYSTEM_NAME_SIZE) != 0)
  { /* don't remove decoration from context */
    return ctx;
  }
err:
  if (cc != 0) msglog(cc,"unexpected configuration error");
  return prmctx;
}
void validateContextAndSubsystem(ExportListStruct *el)
{
  char *c, ctx[CONTEXT_NAME_SIZE], sub[SUBSYSTEM_NAME_SIZE];
  strncpy(ctx,el->EqmContext,CONTEXT_NAME_SIZE);
  strncpy(sub,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);
  if ((c=strchr(ctx,'.')) == NULL || isSystemContextDecoration(c))
  { /* definitely not subsystem-decorated */
    strncpy(sub,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);
  }
  else if (strlen(sub) == 0)
  { /* definitely decorated */
    *c++ = 0;
    strncpy(sub,c,SUBSYSTEM_NAME_SIZE);
  }
  if (strlen(sub) == 0) strncpy(sub,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
  if (strnicmp(ctx,el->EqmContext,CONTEXT_NAME_SIZE) != 0)
  { /* log this */
    feclog("given context %.32s will be treated as the %.16s sub-system decorated context %.32s",el->EqmContext,sub,ctx);
  }
  return;
}
int sndRegExp(ExportListStruct *el)
{
  static int fecInfoDone = FALSE;
  static FecAddrStruct fec;
  static FecInfoStruct finf;
  static int dlen = ADDRPKTSIZE6;
  static int toCount = 0, errCount = 0;
  static char *loc = NULL;
  static BYTE fecAddressBuf[ADDRPKTSIZE6];
  ExpDataStruct srv;
  FecInfoStruct inf;
  char envstr[64], *envptr;
  int cc=0,ismaster=FALSE;
  DTYPE d;

  if (toCount > 30 || errCount > 30)
  { /* Name Server not available, so don't try again */
    feclogEx(TINE_LOGLEVEL_ERROR, "ENS not available: mark servers as registered");
    el->registered = REGISTERED_ALL;
    ExportsRegistered = TRUE;
    return link_timeout;
  }
  if (!fecInfoDone)
  {
    if (gPortOffset > 55500)
    { /* one last check before we send this to the ENS */
      sprintf(dbgbuf, "assigned port offset %d is invalid: aborting process", gPortOffset);
      feclogEx(TINE_LOGLEVEL_FATAL, dbgbuf);
      if (dieFunction != NULL) dieFunction(dbgbuf);
      if (gDieAnotherDay) { gDieAnotherDay = FALSE; return port_not_available; }
      dbglog(dbgbuf);
      exit(1);
    }
    fecInfoDone = TRUE;
    memset(&fec,0,sizeof(FecAddrStruct));
    GetThisFecAddresses(&fec);
    strncpy(fec.fecName,gFecName,FEC_NAME_SIZE);
    fec.portOffset = gPortOffset;
    if (!gMultiCastUseLegacyRules) fec.addrFlags |= ADDRFLAG_MCAST_MODERN;
    sprintf(gFecInfo.ver,"%.16s",GetSystemVersionString());
    if ((loc=getenv("FEC_LOCATION")) != NULL) strncpy(gFecInfo.loc,loc,FEC_LOCATION_SIZE);
    memset(&finf,0,sizeof(FecInfoStruct));
    strncpy(finf.os,gFecInfo.os,FEC_OS_SIZE);
    strncpy(finf.desc,gFecInfo.desc,FEC_DESC_SIZE);
    strncpy(finf.loc,gFecInfo.loc,FEC_LOCATION_SIZE);
    strncpy(finf.ver,gFecInfo.ver,FEC_VERSION_SIZE);
    strncpy(finf.hdw,gFecInfo.hdw,FEC_HDW_SIZE);
    strncpy(finf.resp,gFecInfo.resp,FEC_RESP_SIZE);
  }
  memcpy(&inf,&finf,sizeof(FecInfoStruct));
  memset(&srv,0,sizeof(ExpDataStruct));
  if (el->registered & REGISTERED_ENS)
  { /* has been registered at the ENS */
#   ifndef SMALL_TINE_LIB
    if (el->failoverType == FAILOVER_MASTER)
    { /* still need to register as a failover master ? */
      if (el->registered & REGISTERED_MASTER) return 0;
      strncpy(inf.desc,FAILOVER_MASTER_DESC,FEC_DESC_SIZE);
      strncpy(srv.ExportName,el->EqmMaster,EXPORT_NAME_SIZE);
      ismaster = TRUE;
    }
    else if (el->failoverType == FAILOVER_SLAVE)
    { /* still need to register as a failover master ? */
      if (el->registered & REGISTERED_MASTER) return 0;
      strncpy(inf.desc,FAILOVER_SLAVE_DESC,FEC_DESC_SIZE);
      strncpy(srv.ExportName,el->EqmMaster,EXPORT_NAME_SIZE);
      ismaster = TRUE;
    }
    else
    { /* we're done, but set this bit ... */
      el->registered |= REGISTERED_MASTER;
      return 0;
    }
#   endif
  }
  else
  { /* has not been registered at the ENS yet */
    strncpy(srv.ExportName,el->EqmExportName,EXPORT_NAME_SIZE);
  }
  strncpy(srv.FecName,gFecName,FEC_NAME_SIZE);
  strncpy(srv.EqmName,el->EqmName,EQM_NAME_SHORTSIZE);

  sprintf(envstr,"%.6s_SUBSYSTEM",el->EqmName);
  if ((envptr=getenv(envstr)) != NULL) strncpy(el->EqmSubSystem,envptr,SUBSYSTEM_NAME_SIZE);

  if (strlen(gDeviceContext) == 0) 
  { /* don't let this be an empty string if possible */
    strncpy(gDeviceContext,el->EqmContext,CONTEXT_NAME_SIZE);
  }
  if (strlen(el->EqmContext) == 0) 
  { /* if they're both empty, then we're out of luck ! */
    strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
  }
  strncpy(srv.EqmContext,el->EqmContext,CONTEXT_NAME_SIZE);
  
  if (strlen(gDeviceSubSystem) == 0) 
  { /* don't let this be an empty string if possible */
    strncpy(gDeviceSubSystem,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);
  }
  if (strlen(el->EqmSubSystem) == 0) 
  { /* if they're both empty, then we're out of luck ! */
    strncpy(el->EqmSubSystem,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
  }
  strncpy(srv.SubSystem,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);

  validateContextAndSubsystem(el);

retrySend:
  prepOutgoingAddr(dlen,(char *)fecAddressBuf,&fec,&srv,&inf);
  DTYPEZERO(d);
  d.dArrayLength = dlen;
  d.dFormat = CF_STRUCT;
  d.data.bptr = fecAddressBuf;
  switch (cc=tExecLink(ENS_SRV,ENS_TAG,el->EqmContext,"EXPORT",&d,&d,CA_WRITE,500))
  {
    case 0:
      feclog("FEC %.16s, Server %.32s (%.6s) registered with equipment name server",
        gFecName,srv.ExportName,el->EqmName);
      feclog("Context is %.32s",strlen(el->EqmContext) > 0 ? el->EqmContext : "not given");
      if (ismaster) feclog("Server %.32s is a failover master server",srv.ExportName);
      toCount = errCount = 0;
      gEnsAccepted = TRUE;
      break;
    case access_denied:
      ToggleNameServer();
      if(errCount++ < 5) return cc;
      break;
    case out_of_range:
      if (dlen == ADDRPKTSIZE4)
      { /* this would be a really old ENS */
        dlen = ADDRPKTSIZE4L;
        goto retrySend;
      }
      break;
    case server_name_in_use:
      toCount = 0;
      el->EqmGroup[0] = 0;
      feclogEx(TINE_LOGLEVEL_ERROR, "FEC name %.16s or Server name /%.32s/%.32s is in use",gFecName,el->EqmContext,srv.ExportName);
      break;
    case operation_busy: 
      toCount = 0;
      ccerr(cc);
    case illegal_data_size:
      if (dlen == ADDRPKTSIZE6)
      { /* Legacy ENS */
        dlen = ADDRPKTSIZE4;
        goto retrySend;
      }
    default: 
      if (isTimeout(cc)) toCount++;
      errCount++;
      if (errCount < 30) return cc;
      feclogEx(TINE_LOGLEVEL_ERROR, "register FEC %.16s, Server %.32s : %.32s",gFecName,srv.ExportName,cc2str(cc));
      break;
  }
  el->registered |= REGISTERED_ENS;
  if (ismaster) el->registered |= REGISTERED_MASTER;
err:
  return cc;
}
int sndRegisteredExports(struct timeval *tv)
{
  int cc=0;
# if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
  int fecid,srvid;
# endif
  static time_t lastAttempt = 0;
  static int errCount = 0;
  ExportListStruct *el;

  if (!FecNameRegistered)
  { /* Fec Name not available, so don't try again */
    ExportsRegistered = TRUE;
    return database_not_loaded;
  }
  if (gSystemRunningStandAlone)
  {
    for (el=ExportList; el != NULL; el=el->next)
    {
      if (!strcmp(el->EqmExportName,"LOCAL")) el->registered = REGISTERED_ALL;
      if (el->registered) continue;
      if (strlen(gDeviceContext) == 0) strncpy(gDeviceContext,el->EqmContext,CONTEXT_NAME_SIZE);
      if (strlen(el->EqmContext) == 0) strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
#     if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
      if (addStandAloneEntryToAddressCache(el->EqmContext,el->EqmExportName,el->EqmName,&fecid,&srvid) == 0)
      {
        if ((cc=addAddressToCacheFile(gAddrCacheFilePath,srvid,fecid)) != 0)
        {
          feclogEx(TINE_LOGLEVEL_ERROR, "FEC %.16s could not be added as stand-alone to local address cache : %32s",gFecName,cc2str(cc));
          if (++errCount < 5) return cc;
        }
        feclog("FEC %.16s, Server %.32s (%.6s) registered stand alone",gFecName,el->EqmExportName,el->EqmName);
        feclog("Context is %.32s",strlen(el->EqmContext) > 0 ? el->EqmContext : "not given");
      }
#     endif
      el->registered = REGISTERED_ENS;
    }
    ExportsRegistered = TRUE;
    return 0;
  }
  if ((cc=GetNameServerAddress()) != 0)
  { /* Name Server Address not available, so don't try again */
    ExportsRegistered = TRUE;
    return cc;
  }
  if (tv->tv_sec < (long)lastAttempt + 1) return 0;
  lastAttempt = tv->tv_sec;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (!strcmp(el->EqmContext, "LOCALHOST")) el->registered = REGISTERED_ALL; /* e.g. local repeater */
    if (!strcmp(el->EqmExportName,"LOCAL")) el->registered = REGISTERED_ALL;
    if (!strncmp(el->EqmName,DOOCSEQM_TAG,6)) gDieOnAddrInUse = FALSE;
    if ((cc=sndRegExp(el)) != 0) ccerr(cc);
  }
# ifndef SMALL_TINE_LIB
  for (el=ExportList; el != NULL; el=el->next)
  { /* anybody want to join a group ? */
    if ((el->registered & (REGISTERED_ENS|REGISTERED_MASTER)) != (REGISTERED_ENS|REGISTERED_MASTER)) 
    { /* don't check unless both of the above bits are on */
      ccerr(not_ready);
    }
    if (strlen(el->EqmGroup) > 0)
    { /* keep trying for a while ... */
      if ((cc=joinEnsGroup(el)) != 0 && errCount++ < 30) ccerr(cc);
    }
    el->registered = REGISTERED_ALL;
  }
# endif
  ExportsRegistered = TRUE;
err:
  gettimeofday(tv,(struct timezone *)NULL);
  return cc;
}
int outputConsumerList(void)
{ 
  int k;
  char addr[64], *pstr, *pad;
  struct sockaddr_in *sa4;
  struct sockaddr_in6 *sa6;
  if (gAfInet == AF_INET6)
    ttyoutput("      CLIENT %55s     TYPE    PID PROTOCOL CONTRACTS","ADDRESS");
  else
    ttyoutput("      CLIENT %30s     TYPE    PID PROTOCOL CONTRACTS","ADDRESS");
  for (k=0; k<nconsumers; k++)
  {
    if (ClnTbl[k]->inetProtocol & (UDP6|TCP6|STREAM6))
    {
      sa6 = (struct sockaddr_in6 *)&ClnTbl[k]->adr.ip;
      inet_ntop(AF_INET6,getInAddr((SCKADR *)sa6),addr,40);
      sprintf(&addr[strlen(addr)],"|%d",ntohs(sa6->sin6_port));
    }
    else
    {
      sa4 = (struct sockaddr_in *)&ClnTbl[k]->adr.ip;
      ntoa(ClnTbl[k],addr);
      sprintf(&addr[strlen(addr)],":%d",ntohs(sa4->sin_port));
    }
    pstr = getProtocolAsString(ClnTbl[k]->inetProtocol);
    pad = k < 10 ? "  " : k < 100 ? " " : "";
    if (gAfInet == AF_INET6)
      ttyoutput("%s(%d) %-16.16s %45.45s %8.8s %6d %8.7s %9d",pad,
        k,ClnTbl[k]->usr,addr,ClnTbl[k]->typ,ClnTbl[k]->pid,pstr,ClnTbl[k]->ncontracts);
    else
      ttyoutput("%s(%d) %-16.16s %20.20s %8.8s %6d %8.6s %9d",pad,
        k,ClnTbl[k]->usr,addr,ClnTbl[k]->typ,ClnTbl[k]->pid,pstr,ClnTbl[k]->ncontracts);
  }
  return 0;
}
int outputConsumerStats(int i)
{ 
  char Buffer[256], addr[64];
  struct sockaddr_in *sa4;
  struct sockaddr_in6 *sa6;

  if (i < 0 || i >= nconsumers) return out_of_range;
  sa4 = (struct sockaddr_in *)&ClnTbl[i]->adr.ip;
  sa6 = (struct sockaddr_in6 *)&ClnTbl[i]->adr.ip;
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"client: %.16s\n",ClnTbl[i]->usr); 
  ttyoutput(Buffer); memset(Buffer,' ',80);
  switch (ClnTbl[i]->inetProtocol)
  {
    case IPX: 
      ntoa(ClnTbl[i],addr); 
      break;
    case STREAM:
    case TCP:
    case UDP: 
      ntoa(ClnTbl[i],addr);
      sprintf(&addr[strlen(addr)],":%d",ntohs(sa4->sin_port));
      break;
    case TCP6:
    case UDP6: 
      inet_ntop(AF_INET6,getInAddr((SCKADR *)sa6),addr,40);
      sprintf(&addr[strlen(addr)],"|%d",ntohs(sa6->sin6_port));
      break;
    case PIPE:
      sprintf(addr,"PIPE: %d",ClnTbl[i]->ep.pipe.sck); 
      break;
    case MMF:
#     ifdef WIN32
      sprintf(addr, "HWND: %zd (0x%zx)", (size_t)ClnTbl[i]->ep.wndMmf, (size_t)ClnTbl[i]->ep.wndMmf);
#     endif
      break;
    default:
      sprintf(addr,"unknown"); 
      break;

  }
  sprintf(&Buffer[0],"inet Protocol: %s, address : %s\n",getProtocolAsString(ClnTbl[i]->inetProtocol),addr); 
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"tine Protocol: %d, number contracts : %d\n",ClnTbl[i]->tineProtocol,ClnTbl[i]->ncontracts); 
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0], "application type: %.8s, pid : %d\n", ClnTbl[i]->typ, ClnTbl[i]->pid);
  ttyoutput(Buffer); memset(Buffer, ' ', 80);
  return 0;
}
int outputEqmList(void)
{
  ExportListStruct *el;
  char Buffer[256];
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," MODULE"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[10],"EXPORT NAME"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[30],"CONTEXT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[50],"SUBSYSTEM"); Buffer[strlen(Buffer)] = 0;
  ttyoutput(Buffer);
  for (el=ExportList; el != NULL; el=el->next)
  {
    memset(Buffer,' ',80);
    sprintf(&Buffer[0]," %.6s",el->EqmName); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[10],"%.32s",el->EqmExportName); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[30],"%.32s",el->EqmContext); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[50],"%.16s",el->EqmSubSystem); Buffer[strlen(Buffer)] = 0;
    ttyoutput(Buffer);
  }
  return 0;
}
int outputEqmProperties(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  char Buffer[256];
  NAME64 *prps;
  int cc, i, nprps = 100;
  if (el == NULL) return non_existent_elem;
  if ((prps=(NAME64 *)alloca(100*sizeof(NAME64))) == NULL) return out_of_server_memory;
  if ((cc=GetRegisteredPropertyList(eqm,prps,&nprps)) != 0) return cc;
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," Registered Properties for %.6s (1st 100):",eqm); Buffer[strlen(Buffer)] = 0;
  ttyoutput(Buffer);
  for (i=0; i<nprps; i++)
  {
    sprintf(&Buffer[0]," %.32s",prps[i].name);
    Buffer[strlen(Buffer)] = 0;
    ttyoutput(Buffer);
  }
  return 0;
}
int outputEqmDevices(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  char dstr[DEVICE_NAME_SIZE+8];
  int i;
  if (el == NULL) return non_existent_elem;
  ttyoutput(" Registered Devices for %.6s :",eqm);
  for (i=0; i<el->EqmNumDevices && el->EqmDevLst != NULL; i++)
  {
    if (el->EqmDevLst[i].isReserved) continue;
    sprintf(dstr,"device %d",i);
    ttyoutput(" %.64s",el->EqmDevLst ? el->EqmDevLst[i].name : dstr);
  }
  return 0;
}
int outputMcaDeviceMembers(char *eqm)
{
# define NMCAGRPS2DISPLAY 16 /* 16 groups ought to be enough ... */
  ExportListStruct *el = getExportListItem(eqm);
  char dstr[DEVICE_NAME_SIZE+8];
  char *grps[NMCAGRPS2DISPLAY]; 
  int i, k, n=0, output_done=FALSE;
  if (el == NULL) return non_existent_elem;
  if (el->EqmDevLst == NULL) return not_initialized;
  memset(grps,0,NMCAGRPS2DISPLAY*sizeof(char *));
  for (i=0; i<el->EqmNumDevices; i++)
  { /* find all registered device groups (up to NMCAGRPS2DISPLAY of them) */
    if (el->EqmDevLst[i].mcaGrp != NULL)
    { /* here's one ! */
      if (n == NMCAGRPS2DISPLAY) break; /* we're done */
      for (k=0; k<n && k<NMCAGRPS2DISPLAY; k++) 
        if (!strnicmp(grps[k],el->EqmDevLst[i].mcaGrp->grpName,DEVICE_NAME_SIZE)) break;
      if (k == n) grps[n++] = el->EqmDevLst[i].mcaGrp->grpName; /* new one */
    }
  }
  ttyoutput(" Equipment module %.6s has %d registered device groups",eqm,n);
  for (k=0; k<n; k++) 
  {
    output_done = FALSE;
    for (i=0; i<el->EqmNumDevices; i++)
    {
      if (el->EqmDevLst[i].mcaGrp != NULL)
      {
        if (strnicmp(grps[k],el->EqmDevLst[i].mcaGrp->grpName,DEVICE_NAME_SIZE)) continue;
        if (!output_done)
        {
          output_done = TRUE;
          ttyoutput("\n %d registered members for device group %.64s :", 
            el->EqmDevLst[i].mcaGrp->grpSize, grps[k]);
          ttyoutput("   Index : Device Name");
        }
        sprintf(dstr,"device %d",i);
        ttyoutput("   %5d : %.64s", 
          el->EqmDevLst[i].mcaGrp->grpIndex,
          el->EqmDevLst ? el->EqmDevLst[i].name : dstr);
      }
    }
  }
  return 0;
}
int outputEqmHistories(char *eqm)
{
  int i, p;
  char tag[16],daysStr[16],tsstr[64];
# ifndef SMALL_TINE_LIB
  FilterLink *fl;
# endif
  memset(tag,0,16);
  if (eqm != NULL) strncpy(tag,eqm,EQM_NAME_SHORTSIZE); else strcpy(tag,"ALL modules");
  ttyoutput(" Registered History Records for %s",tag);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (eqm != NULL && strnicmp(eqm,hstTbl[i].c.EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (hstTbl[i].depthDays > 0) sprintf(daysStr,", %d days",hstTbl[i].depthDays); else daysStr[0] = 0;
    ttyoutput(" [%3d] %.64s %.64s : %d %.16s values (%d deep, %d months%s)",
      hstTbl[i].recordIndex,hstTbl[i].c.EqmProperty,hstTbl[i].c.EqmDeviceName,
      hstTbl[i].c.EqmSizeOut,itof(hstTbl[i].c.EqmFormatOut),
      hstTbl[i].depthShort,hstTbl[i].depthLong,daysStr);
    ttyoutput("\tinitialized             : %s",hstTbl[i].recordInitialized ? "TRUE" : "FALSE");
    ttyoutput("\tscheduled               : %s",hstTbl[i].isScheduled ? "TRUE" : "FALSE");
    ttyoutput("\tkeeps points of interest: %s",hstTbl[i].keepPointsOfInterest ? "TRUE" : "FALSE");
    ttyoutput("\tpolling interval        : %d msec",hstTbl[i].pollingRate);
    ttyoutput("\tarchive intervals       : minimum %d msec, maximum %d sec",hstTbl[i].archiveRate,hstTbl[i].heartbeat);
    ttyoutput("\tarchive tolerance       : absolute %g, relative %g percent",hstTbl[i].aTolerance,hstTbl[i].pTolerance*100);
    ttyoutput("\trecord length           : %d bytes",hstTbl[i].recordLength);
    if (hstTbl[i].depthShort > 0)
    {
      ttyoutput("\tcurrent file name       : %.32s",hstTbl[i].archiveFileName);
      ttyoutput("\tlast access time        : %s",GetDataTimeStringEx(hstTbl[i].lastAccessTime,FALSE,tsstr));
      ttyoutput("\tlast access status      : %.32s",cc2str(hstTbl[i].lastAccessSts));
    }
    if (hstTbl[i].prd != NULL)
    {
      char *ctx,*srv,*dev,*prp;
      ctx = hstTbl[i].prd->prpRedirection;
      srv = &ctx[PRP_REDIR_SRV_OFFSET];
      dev = &ctx[PRP_REDIR_DEV_OFFSET];
      prp = &ctx[PRP_REDIR_PRP_OFFSET];
      ttyoutput("\tredirected to           : /%.32s/%.32s/%.64s[%.64s]",ctx,srv,dev,prp);
    }
    p = hstTbl[i].dataPtr; /* where we are now */
    ttyoutput("\tcurrent data timestamp  : %s",
      hstTbl[i].timestamp != NULL ? GetDataTimeStringEx(hstTbl[i].timestamp[p],FALSE,tsstr) : "(unavailable)");
    ttyoutput("\tlast archived           : %s",GetDataTimeStringEx(hstTbl[i].lastArchiveTime,FALSE,tsstr));
#   ifndef SMALL_TINE_LIB
    if ((fl=hstTbl[i].fltr) != NULL && fl->active)
    {
      ttyoutput("\tfilter: %s %s %s => %s",
        fl->key,getComparatorString(fl->fcmp),fl->fstr,(fl->valid?"valid":"not valid"));
    }
#   endif
  }
  return 0;
}
int outputEqmAlmWatch(char *eqm)
{
  AWSLstEntry *aw;
  char tag[16];
# ifndef SMALL_TINE_LIB
  FilterLink *fl;
# endif
  memset(tag,0,16);
  if (eqm != NULL) strncpy(tag,eqm,EQM_NAME_SHORTSIZE); else strcpy(tag,"ALL modules");
  ttyoutput(" Registered Alarm Watch Records for %.16s",tag);
  for (aw=almWatchList; aw != NULL; aw = aw->nxt)
  {
    if (eqm != NULL && strnicmp(eqm,aw->aws.eqm,EQM_NAME_SHORTSIZE)) continue;
    ttyoutput(" %.64s %.64s : %d %.16s values (hi: %g %g; lo: %g %g; norm: %d)",
      aw->aws.prp,aw->aws.dev,aw->aws.siz,itof(aw->aws.fmt),
      aw->aws.hiwarn,aw->aws.hi,aw->aws.lowarn,aw->aws.lo,aw->aws.normal);
#   ifndef SMALL_TINE_LIB
    if ((fl=aw->fltr) != NULL && fl->active)
    {
      ttyoutput("    filter: %s %s %s => %s",
        fl->key,getComparatorString(fl->fcmp),fl->fstr,(fl->valid?"valid":"not valid"));
    }
#   endif
  }
  return 0;
}
int getSizeInElements(UINT32 siz, int fmt, char *tag)
{
  int sz = (int)siz, ssz = 0;
  switch (LFMT(fmt))
  {
    case CF_STRUCT:
      if ((ssz=GetStructSize(tag)) > 0) sz /= ssz;
      break;
    case CF_KEYVALUE:
    case CF_STRING:
      if ((ssz=atoi(tag)) > 0) sz = ssz;
      break;
  }
  return sz;
}
int outputContractList(void)
{
  int i, n;
  char Buffer[256];
  ClnLst *c;
  CONTRACT *con;

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CONTRACT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[50],"POLL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[60],"TO\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    con = &ContractList[i]->contract;
    c   = ContractList[i]->clnLst;
    if (con->EqmFormatOut )
    sprintf(&Buffer[0],"[%d] %.6s %.64s <%.64s> (%d elements)",
      i,con->EqmName,con->EqmProperty,con->EqmDeviceName,
      getSizeInElements(con->EqmSizeOut,con->EqmFormatOut,con->strTagOut));
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[50],"%d msec",(int)ContractList[i]->pollingInterval);
    Buffer[strlen(Buffer)] = ' ';
    if (ContractList[i]->expired)
    {
      sprintf(&Buffer[60]," [expired]\n");
      ttyoutput(Buffer);
      memset(Buffer,' ',80);
    }
    for (n=0, c=ContractList[i]->clnLst; c != NULL; c=c->next, n++)
    {
      if (n < ContractList[i]->nclients)
      {
        sprintf(&Buffer[60],"%.16s\n",c->client->usr);
        ttyoutput(Buffer);
        memset(Buffer,' ',80);
      }
      else
      { /* something is not adding up here ! */
        dbglog("phantom user assigned to contract!");
      }
    }
  }
  return 0;
}
int outputContractStats(int i)
{
  char Buffer[256],*tagOut,*tagIn;
  ClnLst *c;
  CONTRACT *con;

  if (i < 0 || i >= ncontracts) return out_of_range;

  con = &ContractList[i]->contract;
  if (ContractList[i]->expired != CONTRACT_ACTIVE)
  {
    memset(Buffer,' ',80);
    sprintf(&Buffer[0]," contract entry is no longer active\n");
    ttyoutput(Buffer);
    return 0;
  }
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"\nCONTRACT:\n");
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"eqm: %.6s dev: %.64s prp: %.64s\n",con->EqmName,con->EqmDeviceName,con->EqmProperty);
  ttyoutput(Buffer); memset(Buffer,' ',80);
  tagOut = (strlen(con->strTagOut) > 0) ? con->strTagOut : "no tag";
  tagIn = (strlen(con->strTagIn) > 0) ? con->strTagIn : "no tag";
# ifdef OS_16BIT
  sprintf(&Buffer[0],"output: %d %s elements (%.16s)\n",
    getSizeInElements(con->EqmSizeOut,con->EqmFormatOut,con->strTagOut),itof(con->EqmFormatOut),tagOut);
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0], "output: %d elements sent\n",
    getSizeInElements(ContractList[i]->dataSizeOut,con->EqmFormatOut,con->strTagOut));
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"input: %d %s elements (%.16s)\n", 
    getSizeInElements(con->EqmSizeIn,con->EqmFormatIn,con->strTagIn),itof(con->EqmFormatIn),tagIn);
  ttyoutput(Buffer); memset(Buffer,' ',80);
# else
  sprintf(&Buffer[0],"output: %d %s elements (%.16s)\n",
    getSizeInElements(con->EqmSizeOut,con->EqmFormatOut,con->strTagOut),itof(con->EqmFormatOut),tagOut);
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0], "output: %d elements sent\n",
    getSizeInElements(ContractList[i]->dataSizeOut, con->EqmFormatOut, con->strTagOut));
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"input: %d %s elements (%.16s)\n", 
    getSizeInElements(con->EqmSizeIn,con->EqmFormatIn,con->strTagIn),itof(con->EqmFormatIn),tagIn);
  ttyoutput(Buffer); memset(Buffer,' ',80);
# endif
  sprintf(&Buffer[0],"current status: %d last return code: %d separate thread: %s\n\n",
    ContractList[i]->compStatus,ContractList[i]->returnCode,
    ContractList[i]->runInSeparateThread ? "YES" : "NO");
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"POLL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[12],"LAST SENT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[40],"LAST STALE"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[63],"MISSES DELAYS\n"); 
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"%d msec",(int)ContractList[i]->pollingInterval);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[12],"%ld sec %ld usec",
    ContractList[i]->lasttime.tv_sec,ContractList[i]->lasttime.tv_usec);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[40],"%ld sec %ld usec",
    ContractList[i]->lastmarked.tv_sec,ContractList[i]->lastmarked.tv_usec);
  Buffer[strlen(Buffer)] = ' ';
# ifdef OS_16BIT
  sprintf(&Buffer[63],"%6ld %6ld\n\n",
# else
  sprintf(&Buffer[63],"%6d %6d\n\n",
# endif
   ContractList[i]->misses,ContractList[i]->delays);
  ttyoutput(Buffer); memset(Buffer,' ',80);
  sprintf(&Buffer[0],"CLIENTS\n");
  ttyoutput(Buffer); memset(Buffer,' ',80);
  for (c=ContractList[i]->clnLst; c != NULL; c=c->next)
  {
    sprintf(&Buffer[0],"%s",c->client->usr); Buffer[strlen(Buffer)] = ' ';
#   ifdef OS_16BIT
    sprintf(&Buffer[9]," [%s] %d ms +%ld : %s %ld s %ld us %lu miss %lu rec\n",
#   else
    sprintf(&Buffer[9]," [%s] %d ms +%d : %s %ld s %ld us %u miss %u rec\n",
#   endif
            modetoa[BASEMODE(c->status.mode)],
            c->status.pollingInterval,
            c->status.counter,
            c->status.Stale ? "STALE" : "FRESH",
            c->status.lasttime.tv_sec,
            c->status.lasttime.tv_usec,
            c->status.misses,
            c->status.reconnects);
    ttyoutput(Buffer); memset(Buffer,' ',80);
  }
  return 0;
}
UINT32 CMDcount = 0;
int outputServerStats(void)
{
  ttyoutput("Running since %s",ctime(&gSystemStartupTime));
# ifdef OS_16BIT
# if defined(IPXNW) || defined(IPXBSD)
  ttyoutput("Total SPX requests  : %lu\n",SPXcount);
  ttyoutput("Total IPX requests  : %lu\n",IPXcount);
# endif
  ttyoutput("Total TCP requests  : %lu\n",TCPcount);
  ttyoutput("Total UDP requests  : %lu\n",UDPcount);
  ttyoutput("Total KBD commands  : %lu\n",CMDcount);
# else
# if defined(IPXNW) || defined(IPXBSD)
  ttyoutput("Total SPX requests  : %u\n",SPXcount);
  ttyoutput("Total IPX requests  : %u\n",IPXcount);
# endif
  ttyoutput("Total TCP requests  : %u\n",TCPcount);
  ttyoutput("Total UDP requests  : %u\n",UDPcount);
  ttyoutput("Total KBD commands  : %u\n",CMDcount);
# endif
  ttyoutput("CLN sck RCV Buffers : %d\n",max_clnrcvbuf);
  ttyoutput("CLN sck SND Buffers : %d\n",max_clnsndbuf);
  ttyoutput("SRV sck RCV Buffers : %d\n",max_srvrcvbuf);
  ttyoutput("SRV sck SND Buffers : %d\n",max_srvsndbuf);
  ttyoutput("UDP sck SND Buffers : %d\n",max_udpsndbuf);
  if (gEqmListSize) /* non-servers must cover these variables */
  {
#   ifdef OS_16BIT
    ttyoutput("Server Work Area    : %lu bytes\n",srvWorkAreaSize);
#   else
    ttyoutput("Server Work Area    : %u bytes\n",srvWorkAreaSize);
#   endif
    ttyoutput("Registered clients  : %d\n",nconsumers);
    ttyoutput("Registered contracts: %d\n",ncontracts-nExpiredContracts);
#   ifdef OS_16BIT
    ttyoutput("Contract misses     : %lu\n",gContractMisses);
    ttyoutput("Contract delays     : %lu\n",gContractDelays);
    ttyoutput("Client misses       : %lu\n",gClientMisses);
    ttyoutput("Client reconnects   : %lu\n",gClientReconnects);
    ttyoutput("Client retries      : %lu\n",gClientRetries);
    ttyoutput("Synchronous calls   : %lu\n",gSingleLinkCount);
    ttyoutput("Bursts              : %lu\n",gBurstLimitReachedCount);
#   else
    ttyoutput("Contract misses     : %u\n",gContractMisses);
    ttyoutput("Contract delays     : %u\n",gContractDelays);
    ttyoutput("Client misses       : %u\n",gClientMisses);
    ttyoutput("Client reconnects   : %u\n",gClientReconnects);
    ttyoutput("Client retries      : %u\n",gClientRetries);
    ttyoutput("Synchronous calls   : %u\n",gSingleLinkCount);
    ttyoutput("Bursts              : %u\n",gBurstLimitReachedCount);
#   endif
    ttyoutput("Server Cycle Idle   : %s\n",gServerCycleIdle ? "TRUE" : "FALSE");
  }
  if (nConnectionTableEntries)                 /* non-clients must cover these variables */
  {
#   ifdef OS_16BIT
    ttyoutput("Connection arrivals : %lu\n",nArrivals);
    ttyoutput("Connection timeouts : %lu\n",nConnectionTimeouts);
    ttyoutput("Incomplete transfers: %lu\n",nPartialTransfers);
    ttyoutput("Client Work Area    : %lu bytes\n",clnWorkAreaSize);
#   else
    ttyoutput("Connection arrivals : %u\n",nArrivals);
    ttyoutput("Connection timeouts : %u\n",nConnectionTimeouts);
    ttyoutput("Incomplete transfers: %u\n",nPartialTransfers);
    ttyoutput("Client Work Area    : %u bytes\n",clnWorkAreaSize);
#   endif
    ttyoutput("Client Cycle Idle   : %s\n",gClientCycleIdle ? "TRUE" : "FALSE");
  }
  ttyoutput("System Polling intvl: %d msec\n",(int)gSystemCycleDeadband);
  ttyoutput("System Idle time    : %d msec\n",gSystemCycleIdleTime);
  ttyoutput("CPU usage           : %d percent\n", AveBusyTime);
  ttyoutput("Average Cycles/sec  : %d Hz\n", gCycleCounts);
  ttyoutput("Max Cycles/sec      : %d Hz\n", gMaxCycleCounts);
  ttyoutput("Contract Data Stale : %d counts\n", StaleData);
  ttyoutput("Delivery Stale      : %d counts\n", ReturnToServerCycle);
  return 0;
}
#ifndef SMALL_TINE_LIB
int outputSettings(void)
{
  char *cwd;
  ttyoutput("Server Settings      :");
  ttyoutput("FEC name             : %.16s\n",gFecName);
  ttyoutput("Port offset          : %d\n",gPortOffset);
# ifdef OS_16BIT
  ttyoutput("Server Work Area     : %lu bytes\n",srvWorkAreaSize);
# else
  ttyoutput("Server Work Area     : %u bytes\n",srvWorkAreaSize);
# endif
  ttyoutput("System Cycle interval: %d msec\n",(int)gSystemCycleDeadband);
  ttyoutput("Min Polling interval : %d msec\n",(int)MinPollingRate);
  ttyoutput("Contract tbl capacity: %d items\n",(int)ContractListCapacity);
  ttyoutput("Contract renewal len : %d items\n",(int)gSubRenewalLength);
  ttyoutput("Client tbl capacity  : %d items\n",(int)ClientListCapacity);
  ttyoutput("Req ack. on change   : %s\n",(gRequireAcknowledgments ? "yes" : "no"));
  ttyoutput("Retard cont. removal : %s\n",(gRetardSingleContractRemoval ? "yes" : "no"));
  ttyoutput("INET family          : %s\n",(gAfInet == AF_INET6 ? "IPv6" : "IPv4"));
  ttyoutput("Server Burst Limit   : %d packets\n",(int)gBurstLimit);
  ttyoutput("Burst Cycle Delay    : %d msec %d usec\n",(int)gCycleDelay,(int)gCycleMicroDelay);
  ttyoutput("Server Packet MTU    : %d bytes\n",(int)gPacketMtu);
  ttyoutput("TCP maximum msg size : %d bytes (STREAM mode)\n",(int)gMaxTcpMsgSize);
  ttyoutput("TCP tbl capacity     : %d items\n",(int)tcpConTblSize);
  ttyoutput("TCP Burst Limit      : %d blocks (16-KB blocks; 0 => no burst limit)\n",(int)gTcpBurstLimit);
  ttyoutput("Server Send Buffers  : %d bytes\n",(int)max_srvsndbuf);
  ttyoutput("Server Recv Buffers  : %d bytes\n",(int)max_srvrcvbuf);
  ttyoutput("udp Send Buffers     : %d bytes\n",(int)max_udpsndbuf);
# if defined(MAKE_TINE32_DLL) && !defined(NO_TINE_PIPE)
  ttyoutput("shared mem buf size  : %d bytes\n",(int)gShmBufferSize);
  ttyoutput("shared mem jumbos    : %s\n",(gUseMmfJumbos ? "yes" : "no"));
# endif
  ttyoutput("Server Scheduling    : %s\n",(gSchedulePropertyLazy ? "lazy" : "eager"));
  ttyoutput("Scheduler interval   : %d msec\n",gSchedulerInterval);
  ttyoutput("Server tasks         : %s\n",(gAllowBkgTaskReentrancy ? "re-entrant" : "not re-entrant"));
# ifdef MULTITHREADED
  ttyoutput("Server cycle thread  : %s\n",(gRunServerCycleInSeparateThread ? "separate" : "common"));
  ttyoutput("Transport thread     : %s\n",(gRunTransportInSeparateThread ? "separate" : "common"));
  ttyoutput("Transport Ceiling    : %u bytes\n",gTransportCeiling);
# endif
  ttyoutput("Allow remote exit    : %s\n",(gAllowRemoteExit ? "yes" : "no"));
  ttyoutput("Allow remote init    : %s\n",(gAllowRemoteInit ? "yes" : "no"));
  ttyoutput("Allow remote reset   : %s\n",(gAllowRemoteReset ? "yes" : "no"));
  ttyoutput("Client Settings      :");
# ifdef OS_16BIT
  ttyoutput("Client Work Area     : %lu bytes\n",clnWorkAreaSize);
# else
  ttyoutput("Client Work Area     : %u bytes\n",clnWorkAreaSize);
# endif
  ttyoutput("Connect. tbl capacity: %d items\n",(int)gConTblCapacity);
  ttyoutput("Client Burst Limit   : %d packets\n",(int)gClientBurstLimit);
  ttyoutput("Client Send Buffers  : %d bytes\n",(int)max_clnsndbuf);
  ttyoutput("Client Recv Buffers  : %d bytes\n",(int)max_clnrcvbuf);
  ttyoutput("Client Recv Queue    : %d items\n",(int)gClnRecvQueueDepth);
  ttyoutput("use loopback addr    : %s\n",(gUseLoopback ? "yes" : "no"));
  ttyoutput("use watchdog links   : %s\n",(gAutoLinkWatchdogs ? "yes" : "no"));
  ttyoutput("allow common links   : %s\n",(allowCommonLinks ? "yes" : "no"));
  ttyoutput("retry on timeout     : %s\n",(gAlwaysRetry ? "yes" : "no"));
  ttyoutput("callbacks suspended  : %s\n",(GetSuspendCallbacks() ? "yes" : "no"));
  ttyoutput("Directory Settings   :");
  ttyoutput("TINE HOME            : %.128s\n",tineHomePath);
  ttyoutput("FEC HOME             : %.128s\n",fecDbPath);
  ttyoutput("FEC LOG              : %.128s\n",gFeclogPath);
  ttyoutput("HISTORY HOME         : %.128s\n",arcDbPath);
  if ((cwd=SystemGetStartupDirectory()) == NULL) cwd = "(unknown)";
  ttyoutput("WORKING DIRECTORY    : %.128s\n",cwd);

  return 0;
}
int dumpSettings(void) { return outputSettings(); }
#define rptEnvSetting(e) ttyoutput("%s : %s\n",e,((ptr=getenv(e)) != NULL ? ptr : "not set"))
#define rptEnvSetting2(e,ea) ttyoutput("%s : %s\n",e,((ptr=getenv(e)) != NULL ? ptr : (ptr=getenv(ea)) ? ptr : "not set"))
int dumpEnv(void)
{
  char *ptr;
  rptEnvSetting("TINE_HOME");
  rptEnvSetting("TINE_ENS");
  rptEnvSetting("TINE_STANDALONE");
  rptEnvSetting("TINE_CACHE");
  rptEnvSetting("TINE_BURSTLIMIT");
  rptEnvSetting("TINE_NETWORKADDRESS_RESOLUTION");
  rptEnvSetting("TINE_USE_GLOBAL_SYNCHRONIZATION");
  rptEnvSetting("TINE_USE_CYCLE_TRIGGER");
  rptEnvSetting("TINE_LEGACY_MULTICAST");
  rptEnvSetting("TINE_TRANSPORT");
  rptEnvSetting("TINE_QUEUEDEPTH");
  rptEnvSetting("TINE_CONNECTION_TABLE_MIN_CAPACITY");
  rptEnvSetting("TINE_CYCLE_KEY");
  rptEnvSetting("TINE_STATE_KEY");
  rptEnvSetting2("TINE_USE_LOOPBACK","USE_LOOPBACK");
  rptEnvSetting2("TINE_FORCE_SHARED_MEMORY","FORCE_SHARED_MEMORY");
  rptEnvSetting("TINE_MCAST_ADDR");
  rptEnvSetting("TINE_GCAST_ADDR");
  rptEnvSetting("TINE_MCAST_MASK");
  rptEnvSetting("TINE_GCAST_MASK");
  rptEnvSetting("TINE_GLOBALS_BROADCAST_ONLY");
  rptEnvSetting("TINE_RECV_BUFFER_SIZE");
  rptEnvSetting("TINE_SEND_BUFFER_SIZE");
  rptEnvSetting("TINE_DBG_PIPE");
  rptEnvSetting("TINE_START_CYCLER");
  rptEnvSetting("TINE_REMOTE_PIPE");
  rptEnvSetting("TINE_STD_TIME_STR");
  rptEnvSetting("TINE_DST_TIME_STR");
# ifdef WIN32
  rptEnvSetting("TINE_SHM_THREAD_PRIORITY");
# endif
  rptEnvSetting("TINE_CYCLE_THREAD_PRIORITY");
  rptEnvSetting("TINE_HISTORY_ALLOC_FILESET");
  rptEnvSetting2("TINE_HISTORY_HOME","HISTORY_HOME");
  rptEnvSetting2("TINE_HISTORY_FREE_BLOCKS","HISTORY_FREE_BLOCKS");
  rptEnvSetting("TINE_HISTORY_EXTRA_LOGGING");
  rptEnvSetting("TINE_HISTORY_NOPOI");
  rptEnvSetting("TINE_HISTORY_LATE_ADDITION_WINDOW");
  rptEnvSetting("TINE_HISTORY_DEACTIVATE_ON_ERROR");
  rptEnvSetting("TINE_HISTORY_AUTOREMOVE_UNSYNCHRONIZED");
  rptEnvSetting("TINE_ALARM_COLLAPSE_WINDOW");
  rptEnvSetting("TINE_ALARM_SURVEY_INTERVAL");
  rptEnvSetting("FEC_GCAST_ADDR");
  rptEnvSetting("FEC_PACKETMTU");
  rptEnvSetting("FEC_ADDRESS");
  rptEnvSetting("FEC_LOCATION");
  rptEnvSetting("FEC_HOME");
  rptEnvSetting2("FEC_LOG","FECLOG");
  rptEnvSetting2("FEC_LOGDEPTH","FECLOGDEPTH");
  rptEnvSetting2("FEC_POLLRATE","FECPOLLRATE");
  rptEnvSetting2("FEC_WORKAREASIZE","FECWORKAREASIZE");
  rptEnvSetting("FEC_REMOTE_MANAGEMENT");
  rptEnvSetting("FEC_DIE_INUSE");
  rptEnvSetting("FEC_DIE_SOCKETERROR");
  rptEnvSetting("FEC_FILES_ALLOW_SCAN");
  rptEnvSetting("FEC_LOG_COMMANDS");
  rptEnvSetting("FEC_LAZY_SCHEDULING");
  rptEnvSetting("FEC_MINIMUM_FREE_MAINBLOCKS");
  rptEnvSetting("FEC_MINIMUM_FREE_DISKBLOCKS");
  rptEnvSetting("FEC_MAXIMUM_NUM_CONTRACTS");
  return 0;
}
#endif /* !SMALL_TINE_LIB */
int outputServerTime(void)
{
  char tstr[64];
  time_t t;

  t = time(NULL); sprintf(tstr,"%s",ctime(&t)); tstr[24] = 0;
  ttyoutput("current local time: %s (UNIX: %ld)",tstr,t);
  gServerDataTimeStamp = makeDataTimeStamp();
  t = UTCtimestamp; sprintf(tstr,"%s",ctime(&t)); tstr[24] = 0;
  sprintf(&tstr[strlen(tstr)]," %d msec",UTCtimestampMSEC);
  ttyoutput("current data timestamp: %s (UNIX: %10.4f)",tstr,gServerDataTimeStamp);
  ttyoutput("current synchronization offset: %10.4f secs",gDataTimeStampOffset);
  ttyoutput("current system stamp (cycle/pulse number): %10d",gCycleNumber);
  ttyoutput("current assigned system stamp offset: %10d",gCycleOffset);
  return 0;
}
int dumpBuildInfo(void)
{
  return dumpVersion();
}
#define TINE_BUILD_ID 5746
int tineBuildId = TINE_BUILD_ID;
TINE_EXPORT char *GetSystemVersionString(void)
{
  static char vs[64] = { 0 };
  BYTE *ver;
  if (strlen(vs) == 0)
  {
    ver = SystemVersion();
    snprintf(vs,64,"%d.%d.%d:%d",ver[0],ver[1],ver[2]*256 + ver[3],tineBuildId);
  }
  return vs;
}
int dumpVersion(void)
{
  BYTE *ver = SystemVersion();
# if defined(OS_64BIT)
  char *osArch = "64";
# elif defined(OS_16BIT)
  char *osArch = "16";
# else
  char *osArch = "32";
# endif
# ifdef MULTITHREADED
  char *mt = "TRUE";
# else
  char *mt = "FALSE";
# endif
  ttyoutput("Library build information:");
  ttyoutput("TINE library version: %d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  ttyoutput("TINE library build date: %.24s",__DATE__);
  ttyoutput("TINE library build id: %d",tineBuildId);
  ttyoutput("Application version: %.64s",appmkversion);
  if (appmktime > 0) ttyoutput("Application build date: %.64s",ctime(&appmktime));
  ttyoutput("Architecture: %s %s bit, %s endian",
    OS_TAG,osArch,(gEndianness == 0 ? "little" : "big"));
  ttyoutput("Multithreaded: %s",mt);
  return 0;
}
int dumpMessageList(void)
{
  int i, j;
  for (i=0; i<MESSAGE_LOG_DEPTH; i++) 
  {
    j = (gMsgTblPtr - 1 - i + MESSAGE_LOG_DEPTH) % MESSAGE_LOG_DEPTH;
    if ((int)MsgTbl[j].ts == 0) break; /* table not yet filled */
    ttyoutput("%s %.128s",getDataTimeString(MsgTbl[j].ts,0),MsgTbl[j].msg);
  }
  return 0;
}
int dumpFecAddr(void)
{
  char addr[64];
  ttyoutput("FEC Name: %.16s",gFecName);
  inet_ntop(gAfInet,getmyipaddr(),addr,40);
  feclog("determined local host ip address to be : %.40s",addr);
  ttyoutput("FEC IP: %.40s",addr);
  ttyoutput("FEC Port: %d",gPortOffset);
  ttyoutput("FEC registered: %s",gEnsAccepted ? "TRUE" : "FALSE");
  return 0;
}
int dumpConsumerList(void) { return outputConsumerList(); }
int dumpContractList(void) { return outputContractList(); }
int dumpEqmList(void) { return outputEqmList(); }
int dumpEqmProperties(char *eqm) { return outputEqmProperties(eqm); }
int dumpEqmDevices(char *eqm) { return outputEqmDevices(eqm); }
int dumpMcaDeviceMembers(char *eqm) { return outputMcaDeviceMembers(eqm); }
int dumpEqmHistories(char *eqm) { return outputEqmHistories(eqm); }
int dumpEqmAlmWatch(char *eqm) { return outputEqmAlmWatch(eqm); }
int dumpContractStats(int i) { return outputContractStats(i); }
int dumpConsumerStats(int i) { return outputConsumerStats(i); }
int dumpServerStats(void) { return outputServerStats(); }
int dumpServerTime(void) { return outputServerTime(); }

ALIAS_TABLE *gAliasTable = NULL;
int nAliasTableEntries = 0;
HashXRefStruct **AliasXRefTable = NULL;

_CSVCOLUMN csvAliasFileColumns[] = 
{
  { "NAME","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALIAS_TABLE,name),NULL,0 },
  { "ALIAS","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALIAS_TABLE,alias),NULL,0 }
};
CSVDB csvAliasFileDb =
{
  csvAliasFileColumns,
  sizeof(csvAliasFileColumns)/sizeof(csvAliasFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ALIAS_TABLE),
  0
};
int maxAliasEntries = MAX_ALIAS_ENTRIES;
int GetAliasTable(char *eqm)
{
  int cc=0,sz=0;
  char afn[64], *tag=eqm;
  ALIAS_TABLE *at = NULL;

  if (nAliasTableEntries > 0) return already_assigned;
  if (gTineConfigurationCoded) return not_implemented;
  if (tag == NULL) tag = "FEC";
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase)
  {
    sz = populateAIS(gFecName,gFecInfoList,&at);
    if (sz <= 0)ccerr(-sz);
  }
  else
# endif
  {  
#   ifdef FS_DELIMITER
    if (eqm != NULL)
    { /* n.b. aliases belong to the whole fec, but someone might do it this way ... */
      sprintf(afn,"%.6s%c%s",eqm,FS_DELIMITER,ALIAS_FILE);
      if ((cc=csvReadFile(fecDbPath,afn,&csvAliasFileDb,(void **)&at)) == 0) goto file_read;
    }
#   endif
    if (ignoreCommonDbFiles) ccerr(no_such_file);
    if ((cc=csvReadFile(fecDbPath,ALIAS_FILE,&csvAliasFileDb,(void **)&at)) != 0) ccerr(cc);
file_read:
    sz = csvAliasFileDb.siz;
  }
  if (sz > maxAliasEntries) maxAliasEntries = sz * 2;
  if (gAliasTable == NULL && (gAliasTable=(ALIAS_TABLE *)SystemCalloc(maxAliasEntries,sizeof(ALIAS_TABLE))) == NULL)
    ccerr(out_of_local_memory);
  memcpy(gAliasTable,at,sz*sizeof(ALIAS_TABLE));
  nAliasTableEntries = sz;
  SystemFree(at);
  if (nAliasTableEntries > ALIAS_HASH_PAYOFF && (cc=makeAliasXRefTable()) != 0) ccerr(cc);
err:
  if (cc != 0 || sz != 0) feclogEx(TINE_LOGLEVEL_INFO, "alias table (%.6s) : %.32s", tag, cc ? cc2str(cc) : "found");
  return cc;
}
int dumpAliasTable(char *eqm)
{
  int i, cc=0;
  char fn[256];
  FILE *fp=NULL;

  if (nAliasTableEntries == 0 || gAliasTable == NULL) return 0;
# ifdef FS_DELIMITER
  if (eqm != NULL)
  { /* n.b. aliases belong to the whole fec, but someone might do it this way ... */
    sprintf(fn,"%.128s%.6s%c%s",fecDbPath,eqm,FS_DELIMITER,ALIAS_FILE);
    if ((fp=fopen(fn,"r")) != NULL)
    {
      fclose(fp);
      goto fnDone;
    }
  }
# endif
  if (ignoreCommonDbFiles) ccerr(no_such_file);
  sprintf(fn,"%.128s%s",fecDbPath,ALIAS_FILE);
fnDone:
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"NAME, ALIAS\n");
  for (i=0; i<nAliasTableEntries; i++)
  {
    fprintf(fp,"%.64s, %.64s\n",gAliasTable[i].name,gAliasTable[i].alias);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int addAliasToXRefTable(ALIAS_TABLE *ate,int index)
{
  int idx;
  HashXRefStruct *xref;

  if (AliasXRefTable == NULL &&
      (AliasXRefTable=(HashXRefStruct **)SystemCalloc(ALIAS_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_local_memory;
  idx = (int)(ElfHash((unsigned char *)ate->alias) % ALIAS_HASH_SIZE);
  for (xref=AliasXRefTable[idx]; xref && xref->index != index; xref = xref->next)
  {
    if (!strnicmp(ate->alias,gAliasTable[xref->index].alias,64)) break;
  }
  if (xref == NULL) 
  {
    if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
    xref->index = index;
    xref->next = AliasXRefTable[idx];
    AliasXRefTable[idx] = xref;
  }
  return 0;
}
int makeAliasXRefTable(void)
{
  int k,cc;

  for (k=0; k<nAliasTableEntries; k++)
  {
    if ((cc=addAliasToXRefTable(&gAliasTable[k],k)) != 0) return cc;
  }
  return 0;
}
char *getAliasFromTable(char *name)
{
  int i;
  if (name == NULL || nAliasTableEntries == 0) return NULL;
  for (i=0; i<nAliasTableEntries; i++) 
    if (!strnicmp(gAliasTable[i].name,name,64)) return gAliasTable[i].alias;
  return NULL;
}
char *resolveAliasFromTable(char *alias)
{
  int i;
  if (alias == NULL || nAliasTableEntries == 0) return NULL;
  for (i=0; i<nAliasTableEntries; i++) 
    if (!strnicmp(gAliasTable[i].alias,alias,64)) return gAliasTable[i].name;
  return NULL;
}
int addAliasToTable(char *name,char *alias)
{
  int i,cc;
  if (!assertNameIsValid(alias,STRICTLY_ILLEGALCHARS,64)) return invalid_name;
  if ((i=getAliasTableIndex(alias)) < 0)
  { /* not yet in table (normal case) */
    if (nAliasTableEntries == maxAliasEntries) return resources_exhausted;
    if (gAliasTable == NULL && (gAliasTable=(ALIAS_TABLE *)SystemCalloc(maxAliasEntries,sizeof(ALIAS_TABLE))) == NULL)
      return out_of_local_memory;
    if (nAliasTableEntries <= ALIAS_HASH_PAYOFF)
    { /* there were already 'some' aliases added via config file */
      if ((cc=makeAliasXRefTable()) != 0) return cc;
    }
    i = nAliasTableEntries;
    strncpy(gAliasTable[i].alias,alias,64);
    if ((cc=addAliasToXRefTable(&gAliasTable[i],i)) != 0) return cc;
    nAliasTableEntries++;
  }
  strncpy(gAliasTable[i].name,name,64);
  return 0;
}
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen)
{
  int i,idx;
  HashXRefStruct *xref;

  if (tbl == NULL || str == NULL) return -1;
  if (xrf != NULL && hshsiz > 0) /* this should be quick */
  {
    idx = (int)(ElfHash((unsigned char *)str) % hshsiz);
    for (xref=xrf[idx]; xref; xref = xref->next)
    {
      i = xref->index;
      if (!strnicmp(str,tbl+reclen*i+off,64)) return i;
    }
  }
  else
  {
    for (i=0; i<tblsiz; i++)
    {
      if (!strnicmp(str,tbl+reclen*i+off,64)) return i;
    }
  }
  return -1;
}
int isOverloadedWriteProperty(ExportListStruct *el, ExportPropertyListStruct *pl)
{
  char *prpName = pl->prpName;
  int idx, nwrite=0;
  if (pl->isReserved) return FALSE;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (pl=el->EqmPrpLst[idx]; pl != NULL; pl=pl->next)
  {
    if (strnicmp(pl->prpName,prpName,PROPERTY_NAME_SIZE)) continue;
    if ((pl->prpAccessMode & CA_WRITE)) nwrite++;
  }
  if (nwrite > 1) return TRUE;
  return FALSE;
}
/**
 * \intern Finds the 'first in the list' (i.e. last registered) 'default' property.
 *  Overloads are not considered.
 */
ExportPropertyListStruct *GetPropertyListStruct(char *eqm,char *prpName,char *devName)
{
  return GetPropertyListStructEx(eqm,prpName,devName,0);
}
ExportPropertyListStruct *GetPropertyListStructEx(char *eqm,char *prpName,char *devName,int flags)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  PrpQueryStruct prpqsQuery, *prpqs = &prpqsQuery;
  XPropertyQueryStruct xpqsQuery, *xpqs = &xpqsQuery;
  int idx, i, nprps;
  ExportPropertyListStruct *ppls = NULL;
  static ExportPropertyListStruct pqsprp, *last_pls=NULL;
  static ExportListStruct *last_el=NULL;
  static char last_prpName[PROPERTY_NAME_SIZE];
  static char last_devName[DEVICE_NAME_SIZE];
  if (el == NULL || prpName == NULL) return NULL;
  if (devName == NULL) devName = "*";
  if (WaitForSystemMutex(hMfMutex, -1)) return NULL; /* can make use of the manifest mutex */
  if (el == last_el && 
      !strnicmp(prpName,last_prpName,PROPERTY_NAME_SIZE) &&
      !strnicmp(devName,last_devName,DEVICE_NAME_SIZE))
  { /* don't repeat the request unnecessarily */
    if (last_pls != NULL)
    { /* check against NULL: the Karol work-around */
      if (!last_pls->isReserved) goto out;
    }
  }
  last_pls = NULL;
  last_el = el;
  strncpy(last_prpName,prpName,PROPERTY_NAME_SIZE);
  strncpy(last_devName,devName,DEVICE_NAME_SIZE);
  if (HAS_PRPQRYFCN(el))
  { /* registered query fcn => fix these fields and return */
    if (strcmp(devName, "*") == 0) goto out; /* last_pls = NULL !  */
    memset(&pqsprp,0,sizeof(ExportPropertyListStruct));
    pqsprp.prpFormat = CF_FLOAT;
    pqsprp.prpSize = el->EqmNumDevices;
    pqsprp.prpArrayType = AT_UNKNOWN;
    pqsprp.runInSeparateThread = FALSE;
    if (el->EqmPropertyQueryFcn != NULL)
    { /* modern query function */
      prpqs->prpAccess = CA_FIX;
      strncpy(prpqs->prpName,prpName,PROPERTY_NAME_SIZE);
      if ((nprps=el->EqmPropertyQueryFcn(devName,&prpqs)) > 0)
      { /* handled the call */
        for (i=0; i<nprps && strnicmp(prpName,prpqs[i].prpName,PROPERTY_NAME_SIZE); i++);
        if (i == nprps) goto out; /* last_pls = NULL !  */
        pqsprp.prpFormat = prpqs[i].prpFormat;
        pqsprp.prpSize = prpqs[i].prpSize;
        pqsprp.prpArrayType = prpqs[i].prpArrayType;
        if (prpqs[i].prpAccess != CA_FIX && prpqs[i].prpAccess&CA_MT) pqsprp.runInSeparateThread = TRUE;
        pqsprp.prpAccessMode = prpqs[i].prpAccess & 0x03;
        strncpy(pqsprp.prpFormatTag.name,prpqs[i].prpTag,TAG_NAME_SIZE);
        strncpy(pqsprp.prpFormatTagIn.name,prpqs[i].prpTagIn,TAG_NAME_SIZE);
      }
    }
#   ifndef SMALL_TINE_LIB
    else if (el->XEqmPropertyQueryFcn != NULL)
    { /* legacy query function */
      xpqs->prpAccess = CA_FIX;
      strncpy(xpqs->prpName,prpName,PROPERTY_NAME_SHORTSIZE);
      if ((nprps=el->XEqmPropertyQueryFcn(devName,&xpqs)) > 0)
      { /* handled the call */
        for (i=0; i<nprps && strnicmp(prpName,xpqs[i].prpName,PROPERTY_NAME_SIZE); i++);
        if (i == nprps) goto out; /* last_pls = NULL !  */
        pqsprp.prpFormat = xpqs[i].prpFormat;
        pqsprp.prpSize = xpqs[i].prpSize;
        pqsprp.prpArrayType = xpqs[i].prpArrayType;
        if (xpqs[i].prpAccess != CA_FIX && xpqs[i].prpAccess&CA_MT) pqsprp.runInSeparateThread = TRUE;
        pqsprp.prpAccessMode = xpqs[i].prpAccess & 0x03;
        strncpy(pqsprp.prpFormatTag.name,xpqs[i].prpTag,TAG_NAME_SHORTSIZE);
        strncpy(pqsprp.prpFormatTagIn.name,xpqs[i].prpTagIn,TAG_NAME_SHORTSIZE);
      }
    }
#   endif
    pqsprp.devNames = NULL;
    pqsprp.prd = NULL;
    pqsprp.sigfcn = NULL;
    pqsprp.aclLst.usrs = NULL;
    pqsprp.aclLst.nUsrs = 0;
    pqsprp.prpId = -1; /* not an officially registered property */
    last_pls = &pqsprp;
    goto out;
  }
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqmPrpLst[idx]; prp != NULL; prp=prp->next)
  {
    if (strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE)) continue;
    if (prp->isReserved) continue;
    if (flags & CA_SAVERESTORE)
    {
      if (!(prp->prpAccessMode & CA_WRITE)) continue;
      if (prp->prpSizeIn == 0) continue;
      if (prp->prpFormat != prp->prpFormatIn) continue;
      if (prp->prpSize != prp->prpSizeIn)
      { /* if true then: allowed are channel arrays or structures */
        if (!(prp->prpArrayType & AT_CHANNEL) && prp->prpFormat != CF_STRUCT) continue;
      }
    }
    if (flags & CA_WRITE)
    {
      if (!(prp->prpAccessMode & CA_WRITE)) continue;
    }
    break;
  }
  last_pls = prp;
out:
  ppls = last_pls; /* return a stack variable ! */
  ReleaseSystemMutex(hMfMutex);
  return ppls;
}
int isSpectrumArray(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,prpName,NULL);
  if (prp == NULL || !(prp->prpArrayType & AT_SPECTRUM)) return FALSE;
  if (prp->prpSize <= 1) return FALSE;
  return TRUE;
}
int isChannelArray(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,prpName,NULL);
  if (prp == NULL || !(prp->prpArrayType & AT_CHANNEL)) return FALSE;
  return TRUE;
}
TINE_EXPORT int wc_stricmp(const char *wcstr,const char *str)
{
  char beginToken[64], endToken[64], midToken[64];
  int blen,elen,mlen;
  if (wcstr == NULL || str == NULL) return -1;
  if (!strcmp(wcstr,"*")) return 0; /* always matches */
  if (strchr(wcstr,'*') == NULL) return stricmp((char *)wcstr,(char *)str); /* not a wildcard compare */
  getWildcardTokens((char *)wcstr,beginToken,endToken,midToken);
  blen = (int)strlen(beginToken); if (blen > 64) blen = 64;
  elen = (int)strlen(endToken); if (elen > 64) elen = 64;
  mlen = (int)strlen(midToken); if (mlen > 64) mlen = 64;
  return wildcardMatches((char *)str,beginToken,blen,endToken,elen,midToken,mlen) ? 0 : -1;
}
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken)
{
  char *s,*c,lclStr[128];

  if (beginToken == NULL || endToken == NULL || midToken == NULL) return;
  beginToken[0] = endToken[0] = midToken[0] = 0; /* start from scratch */
  if (wildcardString == NULL) return;
  strncpy(lclStr,wildcardString,64); lclStr[64] = 0;
  s = lclStr;
  if (*s == '*') 
  { /* no begin token */
    while (*s == '*') s++;
    if ((c=strchr(s,'*')) != NULL) 
    { /* there's a midToken ("*abc*") */
      *c = 0;
      strncpy(midToken,s,64);
    }
    else
    { /* there's an end token */
      strncpy(endToken,s,64);
    }
  }
  else
  {
    strncpy(beginToken,s,64); beginToken[63] = 0;
    if ((c=strchr(beginToken,'*')) != NULL) 
    { /* found begin Token end */
      *c++ = 0;
      s = c; /* re-position (could still be an end token */
      while (*s == '*') s++; /* is it "abc*xyz" ? */
      if (strlen(s) > 0) strncpy(endToken,s,64);
    }
  }
}
TINE_EXPORT char *stristr(const char *str,const char *tok)
{
  char s1[256], s2[256], *c;
  size_t len;
  if (str == NULL || tok == NULL) return NULL;
  strncpy(s1,str,255); strncpy(s2,tok,255);
  s1[255] = s2[255] = 0;
  strupr(s1); strupr(s2);
  if ((c=strstr(s1,s2)) == NULL) return NULL;
  len = (size_t)c - (size_t)s1;
  return (char *)(str + len);
}
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen)
{
  int dlen;
  if (devName == NULL || beginToken == NULL || endToken == NULL || midToken == NULL) return FALSE;
  if ((dlen=(int)strlen(devName)) == 0) return FALSE;
  if (beginLen > 0)
  { /* check against begin token */
    if (strnicmp(devName,beginToken,beginLen)) return FALSE;
  }
  if (endLen > 0 && endLen <= dlen)
  { /* check against end token */
    if (strnicmp(&devName[dlen-endLen],endToken,endLen)) return FALSE;
  }
  if (midLen > 0)
  { /* check against mid token */
    if (stristr(devName,midToken) == NULL) return FALSE;
  }
  return TRUE;
}
int isContiguous(char *eqm,char *devName,NAME64 *devlist,int *start,int *len)
{ /* called in the case of local multi-channel arrays */
  int i, n, inSequence = FALSE, finSequence = FALSE;
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;
  ExportListStruct *el=getExportListItem(eqm);

  if (el == NULL || devName == NULL || devlist == NULL || start == NULL || len == NULL) return FALSE;

  if (!strcmp(devName,"*"))  
  { /* channel array with '*' is continguous */
    *start = 0;
    return TRUE;
  }
  getWildcardTokens(devName,beginToken,endToken,midToken);

  /* calculate these once outside the loop ... */
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);

  for (i=0,n=0; i<*len; i++)
  {
    if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen))
    {
      if (inSequence) finSequence = TRUE;
      continue;
    }
    if (finSequence) return FALSE; /* found another match -> not contiguous ! */
    if (!inSequence) *start = i;
    inSequence = TRUE;
    n++;
  }
  *len = n;
  return inSequence;
}
int isLocal(char *eqm,char *devName,char *prpName)
{ /* not a registered multi-channel array */
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,prpName,NULL);
  PrpRedirBlk *prd = NULL;
  char beginToken[64], endToken[64], midToken[64];
  int i, blen = 0, elen = 0, mlen = 0;

  if (el == NULL || prp == NULL) return FALSE;
  getWildcardTokens(devName,beginToken,endToken,midToken);
  /* calculate these once outside the loop ... */
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);
  if (el->EqmDevLst != NULL) 
  { /* has device redirection list, check it */
    for (i=0; i<el->EqmNumDevices; i++)
    {
      if (el->EqmDevLst[i].rdr != NULL && strlen(&el->EqmDevLst[i].rdr[CONTEXT_NAME_SIZE]) > 0 &&
          wildcardMatches(el->EqmDevLst[i].name,beginToken,blen,endToken,elen,midToken,mlen))
        return FALSE;
    }
  }
  if (prp->prd != NULL)
  { /* has property redirector, check it */
    for (prd=prp->prd; prd != NULL; prd=prd->next)
    {
      if (wildcardMatches(&prd->prpRedirection[PRP_REDIR_DEV_OFFSET],beginToken,blen,endToken,elen,midToken,mlen))
        return FALSE;
    }
  }
  if (HAS_PRPQRYFCN(el))
  { /* has registered query function -> let it handle this */
  }
  return TRUE; /* appears to be local so-far */
}
TINE_EXPORT char *GetRegisteredExportName(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return NULL;
  return el->EqmExportName;
}
TINE_EXPORT char *GetRegisteredContext(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return NULL;
  return el->EqmContext;
}
TINE_EXPORT int GetRegisteredPropertyList(char *eqm,NAME64 *prpNames,int *nprps)
{
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *prp;
  int n,k,idx;

  if (el == NULL) return non_existent_elem;
  if (prpNames == NULL || nprps == NULL) return argument_list_error;
  n = *nprps;

  for (k=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && k<n; idx++)
  {
    for (prp=el->EqmPrpLst[idx]; prp && k<n; prp=prp->next)
    {
      if (prp->isReserved) continue;
      if (k > 0 && !strnicmp(prpNames[k-1].name,prp->prpName,PROPERTY_NAME_SIZE)) continue;
      strncpy(prpNames[k].name,prp->prpName,PROPERTY_NAME_SIZE);
      k++;
    }
  }
  *nprps = k;
  return 0;
}
void freeEqmDevType(ExportDeviceType *devlst, int idx, int len)
{
  ExportDeviceType *dev = &devlst[idx];
  ALARM *alm, *alm0;
  int i;
  dev->isReserved = TRUE;
  sprintf(dev->name, "reserved#%d", idx);
  SystemFree(dev->desc);
  SystemFree(dev->rdr);
  SystemFree(dev->loc);
  if (dev->aclLst != NULL)
  {
    SystemFree(dev->aclLst->nets);
    SystemFree(dev->aclLst->usrs);
    SystemFree(dev->aclLst->grps);
    SystemFree(dev->aclLst);
  }
  for (i=0; i<len; i++)
  { /* look for other using the same property list */
    if (i != idx && devlst[i].prps == dev->prps) break;
  }
  if (i == len && dev->prps != NULL)
  { /* the only user */
    SystemFree(dev->prps->prpList);
    SystemFree(dev->prps);
  }
  else
  {
    dev->prps = NULL;
  }
  SystemFree(dev->mcaGrp);
  if (dev->almLst != NULL)
  { /* any alarms still in the list ? */
    for (alm = dev->almLst; alm != NULL; )
    {
      alm0 = alm->next; SystemFree(alm); alm = alm0;
    }
    dev->almLst = NULL;
  }
}
void freeEqmDevLst(ExportDeviceType *devlst,int len)
{
  ALARM *alm, *alm0;
  int i, k;
  SystemFree(devlst[0].name); /* the NAME64 list */
  for (i=0; i<len; i++)
  {
    SystemFree(devlst[i].desc);
    SystemFree(devlst[i].rdr);
    SystemFree(devlst[i].loc);
    if (devlst[i].aclLst != NULL)
    {
      SystemFree(devlst[i].aclLst->nets);
      SystemFree(devlst[i].aclLst->usrs);
      SystemFree(devlst[i].aclLst->grps);
      SystemFree(devlst[i].aclLst);
    }
    if (devlst[i].prps != NULL) SystemFree(devlst[i].prps->prpList);
    for (k=0; k<len; k++)
    { /* NULL out any dependencies */
      if (k == i) continue;
      if (devlst[k].prps == devlst[i].prps) devlst[k].prps = NULL;
    }
    SystemFree(devlst[i].prps);
    SystemFree(devlst[i].mcaGrp);
    if (devlst[i].almLst != NULL)
    { /* any alarms still in the list ? */
      for (alm=devlst[i].almLst; alm != NULL; )
      {
        alm0 = alm->next; SystemFree(alm); alm = alm0;
      }
    }
  }
  SystemFree(devlst);
}
#if !defined(DOS) && !defined(NIOS)
void freePropertyOverloadReferences(ExportPropertyListStruct *prp)
{
  char *prpName = prp->prpName;
  while (prp != NULL)
  {
    if (!strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE))
    {
      prp->devNames = NULL;
      prp->aclLst.usrs = NULL;
      prp->aclLst.grps = NULL;
      prp->aclLst.nets = NULL;
    }
    prp = prp->next;
  }
}
TINE_EXPORT int RemoveEquipmentModule(const char *eqm)
{
  ExportListStruct *el, *el0=NULL;
  HashXRefStruct *xref, *xref0;
  ExportPropertyListStruct *prp, *prp0;
  ADSList *lst=NULL, *nxtl=NULL;
  PrpRedirBlk *prd, *prd0;
  int i, cc = 0;

  if (WaitForSystemMutex(hSystemServerMutex,gSystemTick*MUTEX_WAIT_AUGMENT) != 0) return semaphore_busy;
  for (el=ExportList; el != NULL; el=el->next)
  { /* locate and find its predecesor */
    if (!strncmp(el->EqmName,eqm,EQM_NAME_SHORTSIZE)) break;
    el0 = el;
  }
  if (el == NULL) ccerr(non_existent_elem);
  el->isterminating = TRUE;
# ifdef MULTITHREADED
  /* stop related threads */
  if (el->hEqmBkg != 0) millisleep(el->rateBkg);
  millisleep(gSystemTick);
# endif
  /* free property related memory */
  for (i=0; i<PROPERTY_HASH_SIZE; i++)
  {
    for (prp=el->EqmPrpLst[i]; prp != NULL; )
    {
      prp0 = prp->next;
      if (prp->prd != NULL) 
      {
        for (prd=prp->prd; prd != NULL; )
        {
          prd0 = prd->next; SystemFree(prd); prd = prd0;
        }
      }
      SystemFree(prp->devNames);
      SystemFree(prp->aclLst.usrs);
      SystemFree(prp->aclLst.grps);
      SystemFree(prp->aclLst.nets);
      SystemFree(prp->srBuf);
      SystemFree(prp->prpUrl);
      freePropertyOverloadReferences(prp);
      SystemFree(prp); prp = prp0;
    }
  }
  /* free device related memory */
  if (el->EqmDevXRef != NULL)
  {
    for (i=0; i<DEVICES_HASH_SIZE; i++)
    {
      for (xref=el->EqmDevXRef[i]; xref != NULL; )
      {
        xref0 = xref->next; SystemFree(xref); xref = xref0;
      }
    }
    SystemFree(el->EqmDevXRef);
  }
  if (el->EqmDevLst != NULL) freeEqmDevLst(el->EqmDevLst,el->EqmDevLstCapacity);
  SystemFree(el->EqmWcPrpLst);
  /* free alarm related memory */
  if (el->adsTable != NULL) 
  {
    for (lst=el->adsTable; lst !=NULL; )
    {
      nxtl = lst->next; SystemFree(lst); lst = nxtl;
    }
  }
  /* free ACL related memory */
  SystemFree(el->aclLst.usrs);
  SystemFree(el->aclLst.nets);
  /* free alarm region pattern list if any */
  SystemFree(el->regionsLst);
  /* free followers list */
  SystemFree(el->followers);
  if (el0 != NULL)
  { /* freed module not the first in the linked list => splice */
    el0->next = el->next;
    SystemFree(el);
  }
  else
  { /* freed module was the first in the list */
    ExportList = el->next;
    SystemFree(el);
  }
err:
  ReleaseSystemMutex(hSystemServerMutex);
  return cc;
}
void freeServerMemory(void)
{
  ExportListStruct *el=ExportList;
  HashXRefStruct *xref, *xref0;
  int i, cc;

  feclog("freeing allocated memory");
  flushRedirectedLinkList();
  flushReLnkLst();
  flushBlackLnkLst();
  flushFilterLinks();
  flushWcLst();

  if (cdiSystemFreeHook != NULL) cdiSystemFreeHook();
  SystemFree(tcpSckTbl);
  SystemFree(dbgSckTbl);
  SystemFree(IpBcstLst);
  SystemFree(vFeclogBuffer);
  rmvTcpClnSckList();
  if (gAliasTable != NULL)
  {
    for (i=0; AliasXRefTable != NULL && i<ALIAS_HASH_SIZE; i++)
    {
      for (xref=AliasXRefTable[i]; xref != NULL; )
      {
        xref0 = xref->next; SystemFree(xref); xref = xref0;
      }
    }
    SystemFree(gAliasTable);
  }
  if (prpDbaLst != NULL)
  {
    PrpDbaItem *dba, *ndba;
    for (dba=prpDbaLst; dba != NULL; dba=ndba)
    {
      ndba = dba->nxt; SystemFree(dba); 
    }
    prpDbaLst = NULL;
  }
  if (StockPropXRefTable != NULL)
  {
    for (i=0; i<STOCKPROP_HASH_SIZE; i++)
    {
      for (xref=StockPropXRefTable[i]; xref != NULL; )
      {
        xref0 = xref->next; SystemFree(xref); xref = xref0;
      }
    }
    SystemFree(StockPropXRefTable);
  }
  freeHstMemory();
  freeAlmMemory();
  freeSrPrpList();
  freeErrorTable();

  while (el != NULL)
  {
    if ((cc=RemoveEquipmentModule(el->EqmName)) != 0)
      feclogEx(TINE_LOGLEVEL_ERROR,"could not free EQM %.6s : %.32s",el->EqmName,cc2str(cc));
    el = ExportList;
  }
  gEqmListSize = 0;
  ExportList = NULL;
  if (gcastTbl != NULL)
  {
    for (i=0; i<ngcasts; i++)
    {
      SystemFree(gcastTbl[i]->refdata);
      SystemFree(gcastTbl[i]);
    }
    SystemFree(gcastTbl);
  }
# ifndef SMALL_TINE_LIB
  SystemFree(_drlstbuf);
  _drlstbufsiz = _drlstsiz = 0;
# endif
  if (ContractList != NULL)
  {
    ClnLst *c, *cnxt=NULL;
    for (i=0; i<ContractListCapacity; i++)
    {
      if (ContractList[i] == NULL) continue;
      for (c=ContractList[i]->clnLst; c != NULL; c = cnxt)
      {
        cnxt = c->next;
        SystemFree(c);
      }
      SystemFree(ContractList[i]->dataIn);
      SystemFree(ContractList[i]->dataPtr);
      if (!ContractList[i]->unbuffered) SystemFree(ContractList[i]->dataOut);
      SystemFree(ContractList[i]->dataOutArr);
      SystemFree(ContractList[i]);
    }
    SystemFree(ContractList);
  }
  if (ClnTbl != NULL)
  {
    for (i=0; i<ClientListCapacity; i++)
    {
      if (ClnTbl[i] == NULL) continue;
      SystemFree(ClnTbl[i]);
    }
    SystemFree(ClnTbl);
  }

  rmvAccessLock(NULL,NULL,FALSE);

  if (conTbl != NULL)
  {
    GrpTblEntry *gt, *gt0;
    GrpMember *gm, *gm0;
    nConnectionTableEntries = 0;
    for (gt=grpTbl; gt != NULL; gt=gt0)
    {
      gt0 = gt->next;
      for (gm=gt->members; gm != NULL; gm=gm0)
      {
        gm0 = gm->nxt;
        free(gm);
      }
      SystemFree(gt);
    }
    for (i=0; i<gConTblCapacity; i++)
    {
      if (conTbl[i] != NULL && (i > 1 || !NameServerLoaded))
      {
        SystemFree(conTbl[i]->dataIn);
        SystemFree(conTbl[i]->bufOut);
        SystemFree(conTbl[i]->bufOutBlks);
        SystemFree(conTbl[i]);
      }
    }
    SystemFree(conTbl);
  }

  if (glbPrmLst != NULL)
  {
    GlbPrmLst *g, *g0;
    for (g=glbPrmLst; g != NULL; g=g0)
    {
      for (i=0; i<g->num; i++) if (g->data[i] != NULL) SystemFree(g->data[i]);
      g0 = g->nxt; free(g);
    }
    glbPrmLst = NULL;
  }
  flushGlobalsTable();
  freeUserCommands();

  SystemFree(SrvTbl);
  SystemFree(FecTbl);
  SystemFree(NameServer);
  numSrvTblEntries = 0;
  numFecTblEntries = 0;

  SystemFree(gtMCastAddrLst);
  nMCastAddrLstEntries = 0;

  if (tmpWorkArea != NULL && tmpWorkArea != srvWorkArea) SystemFree(tmpWorkArea);
# ifndef srvWorkArea
  SystemFree(srvWorkArea);
# else
  SystemFree(WorkArea);
# endif
# ifndef clnWorkArea
  SystemFree(clnWorkArea);
# endif
# ifdef MULTITHREADED
  SystemFree(mcstWorkArea);
# endif
  FREESTRUCT(PktHdrStructStruct);
  FREESTRUCT(ReqHdrStructStruct);
  FREESTRUCT(PrdrHdrStruct);
  FREESTRUCT(PrdrHdr5Struct);
  FREESTRUCT(SubInfoPktStruct);
  FREESTRUCT(SubInfoPkt5Struct);
  FREESTRUCT(RPCDataStructStruct);
  FREESTRUCT(RPCFecStructStruct);
  FREESTRUCT(ServerStateStructStruct);
  FREESTRUCT(GlbDataHdr3Struct);
  FREESTRUCT(GlbDataHdr4Struct);
  FREESTRUCT(GlbDataHdrStruct);
  FREESTRUCT(SpectrumHeaderStruct);
  FREESTRUCT(ImageHeaderStruct);
  FREESTRUCT(MDAHeaderStruct);
  FREESTRUCT(RspHdrStructStruct);
  FREESTRUCT(SubReqHdrStruct);
  FREESTRUCT(SubRspHdrStruct);
  FREESTRUCT(GSpectrumHeaderStruct);

  freeStructList(TRUE);
  structList = NULL;
  
  rmvBitFieldList();

  return;
}
int GetDeviceListFromProperty(char *eqm,char *prp,NAME64 *devlst,UINT32 *devlen)
{
  int cc = 0,nid, i;
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl;
  char namProperty[128], dev[16];
  DTYPE dout,din;
  NAME16 *devlst16 = (NAME16 *)tmpWorkArea;
  static int done = 0;

  if (!done)
  {
    memset(&qrycln,0,sizeof(ClnHdr));
    memset(&qrycon,0,sizeof(ClnLst));
    strncpy(qrycln.usr,"LOCAL_QUERY",USERNAME_SIZE);
    qrycon.client = &qrycln;
    qrycl.clnLst = &qrycon;
    qrycl.nclients = 1;
    done = TRUE;
  }
  if (devlst == NULL || devlen == NULL) return argument_list_error;
  if (el == NULL || !el->hasPropertyHandler) return non_existent_elem;
  if (strlen(prp) > PROPERTY_NAME_SIZE - 4) return string_too_long;
  sprintf(namProperty,"%.60s.NAM",prp);
  if ((nid=GetPrpId(eqm,namProperty)) < 0) return illegal_property;
  if ((pl=GetPropertyListStruct(eqm,namProperty,NULL)) == NULL) return code_failure;
  din.dArrayLength = 0;
  din.dFormat = CF_NULL;
  din.dTag[0] = 0;
  din.data.vptr = NULL;
  dout.dArrayLength = MIN(pl->prpSize,*devlen);
  dout.dFormat = CF_NAME64;
  dout.dTag[0] = 0;
  dout.data.vptr = devlst;
  qrycl.exp = el;
  qrycl.compStatus = not_signalled;
  cc = callEqm(&qrycl,"#0",namProperty,&dout,&din,CA_READ|CA_QUERY);
  if (cc == illegal_format)
  { /* legacy eqm with new lib ? */
    dout.dArrayLength = MIN(pl->prpSize,*devlen);
    dout.dFormat = CF_NAME16;
    dout.data.vptr = devlst16;
    qrycl.compStatus = not_signalled;
    cc = callEqm(&qrycl,"#0",namProperty,&dout,&din,CA_READ|CA_QUERY);
    if (cc == 0)
    { /* this worked so re-map : */
      for (i=dout.dArrayLength-1; i >= 0; i--)
      {
        strncpy(dev,devlst16[i].name,16);
        memset(devlst[i].name,0,64);
        strncpy(devlst[i].name,dev,16);
      }
    }
  }
  if (dout.dArrayLength == 0) return illegal_data_size;
  *devlen = dout.dArrayLength;
  return cc;
}
/**
 * \internal NOT thread-safe 
 */
int resetPropertyDeviceList(char *eqm,char *prp)
{
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl=GetPropertyListStruct(eqm,prp,NULL);

  if (el == NULL || !el->hasPropertyHandler) return non_existent_elem;
  if (!el->hasPropertyDeviceNames) return not_applicable;
  if (pl == NULL) return illegal_property;
  if (strlen(prp) > PROPERTY_NAME_SIZE-4) return invalid_property;
  SystemFree(pl->devNames);
  for (; pl != NULL; pl=pl->next)
  { /* set for any overloads */
    if (strnicmp(pl->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    pl->devNames = NULL;
    pl->numDevices = 0;
  }
  return 0;
}
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp)
{
  int cc = 0,i;
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *p, *pl=GetPropertyListStruct(eqm,prp,NULL);
  static int inside = 0;
  NAME64 *namlst = (NAME64 *)tmpWorkArea;
  UINT32 devlen=1024;

  if (el == NULL || !el->hasPropertyHandler) return NULL;
  if (!el->hasPropertyDeviceNames) return NULL;
  if (pl == NULL) return NULL;
  if (strlen(prp) > PROPERTY_NAME_SIZE-4) return NULL;
  if (pl->devNames == NULL)
  {
    if (inside) return NULL;
    inside = TRUE;
    if ((cc=GetDeviceListFromProperty(eqm,prp,namlst,&devlen)) != 0) ccerr(cc);
    if ((pl->devNames=(NAME64 *)SystemCalloc(devlen,64)) == NULL) ccerr(out_of_server_memory);
    for (i=0; i<(int)devlen; i++) strncpy(pl->devNames[i].name,namlst[i].name,DEVICE_NAME_SIZE);
    pl->numDevices = devlen;
    for (p=pl->next; p != NULL; p=p->next)
    { /* set references for any overloads as well */
      if (strnicmp(p->prpName,prp,PROPERTY_NAME_SIZE)) continue;
      p->devNames = pl->devNames;
      p->numDevices = p->numDevices;
    }
  }
err:
  inside = FALSE;
  return cc ? NULL : pl;
}
int callEqmInSeparateThread(char *eqm,char *prp,int access)
{
  ExportPropertyListStruct *pl=GetPropertyListStruct(eqm,prp,NULL);
  if (access & (CA_HIST|CA_ALARM)) return FALSE;
  return pl != NULL ? pl->runInSeparateThread : FALSE;
}
TINE_EXPORT int AssignDeviceListToProperty(char *eqm,char *prp,NAME64 *devlst,int lstlen)
{
  int cc = 0,i;
  char lclname[DEVICE_NAME_SIZE+4] = {0};
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *p, *pl=GetPropertyListStruct(eqm,prp,NULL);

  if (el == NULL || !el->hasPropertyHandler || devlst == NULL || lstlen < 1) return argument_list_error;
  if (pl == NULL) return illegal_property;
  if (strlen(prp) > PROPERTY_NAME_SIZE-4) return string_too_long;
  if (pl->devNames != NULL && lstlen > pl->numDevices)
  { /* new list? and not enough room: free the old memory */
    SystemFree(pl->devNames);
  }
  if (pl->devNames == NULL)
  { /* allocate room for the list */
    if ((pl->devNames=(NAME64 *)SystemCalloc(lstlen,64)) == NULL) ccerr(out_of_server_memory); 
  }
  for (i = 0; i < lstlen; i++)
  {
    strncpy(lclname, devlst[i].name, DEVICE_NAME_SIZE);
    strtrm(lclname);
    strncpy(pl->devNames[i].name,lclname,DEVICE_NAME_SIZE);
  }
  pl->numDevices = lstlen;
  for (p=pl->next; p != NULL; p=p->next)
  { /* set references for any overloads as well */
    if (strnicmp(p->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    p->devNames = pl->devNames;
    p->numDevices = p->numDevices;
  }
  el->hasPropertyDeviceNames = TRUE;
  if (pl->srBuf != NULL)
  { /* reallocate the SR buffer */
    SystemFree(pl->srBuf);
    pl->srBufSiz = 0;
    isAllowedSaveRestoreProperty(el, pl);
    _RestorePropertyValues(eqm, pl->prpName, pl->srBuf, pl->prpFormat, pl->srBufSiz, TRUE);
  }
err:
  return cc;
}
TINE_EXPORT int GetRegisteredPropertyListStruct(char *eqm,char *prpName,ExportPropertyListStruct *prpLstStruct)
{
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,prpName,NULL);
  if (prp == NULL) return not_registered;
  if (prpLstStruct == NULL) return argument_list_error;
  memcpy(prpLstStruct,prp,sizeof(ExportPropertyListStruct));
  return 0;
}
TINE_EXPORT int RegisteredPropertyIsWritable(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetPropertyListStructEx(eqm,prpName,NULL,CA_WRITE);
  return prp == NULL ? FALSE : TRUE;
}
TINE_EXPORT int GetNumberRegisteredDevices(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return -argument_list_error;
  return el->EqmNumDevices - getNumReservedDevices(el);
}
TINE_EXPORT int GetSizeDeviceCapacity(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return -argument_list_error;
  return el->EqmDevLstCapacity;
}
#if !defined(VXWORKS)
TINE_EXPORT int SetSizeDeviceCapacity(char *eqm,int size)
{ /* re-assign memory */
  int i, addsize = 0, cc = 0;
  void *mptr;
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL || size == 0) return argument_list_error;
  if (el->EqmDevLstCapacity == 0) return not_initialized;
  if (size <= el->EqmDevLstCapacity) return already_assigned;
  if (el->EqmDevLstCapacity < DEVICES_HASH_PAYOFF && size >= DEVICES_HASH_PAYOFF)
  {
    if ((el->EqmDevXRef == NULL) &&
        (el->EqmDevXRef=(HashXRefStruct **)SystemCalloc(DEVICES_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
      ccerr(out_of_server_memory);
  }
  if (el->EqmDevLst != NULL)
  {
    NAME64 *nptr = (NAME64 *)el->EqmDevLst[0].name;
    addsize = size - el->EqmDevLstCapacity; /* must be positive (see check above) */
    if ((nptr=(NAME64 *)realloc(nptr,size*sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    el->EqmDevLst[0].name = (char *)nptr;
    if ((mptr=realloc(el->EqmDevLst,size*sizeof(ExportDeviceType))) == NULL) ccerr(out_of_server_memory);
    el->EqmDevLst = (ExportDeviceType *)mptr;
    memset(&el->EqmDevLst[el->EqmDevLstCapacity],0,addsize*sizeof(ExportDeviceType)); /* zero out the new stuff */
    for (i=1; i<size; i++) el->EqmDevLst[i].name = nptr[i].name; /* need to reassign pointers */
  }
  el->EqmDevLstCapacity = size;
err:
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR, "could not increase device capacity: %.32s",cc2str(cc));
  return cc;
}
#endif /* !defined(VXWORKS) */
TINE_EXPORT char *GetLocalName(char *exportName)
{
  ExportListStruct *el;
  if (exportName == NULL || strlen(exportName) == 0) return NULL;
  for (el=ExportList; el != NULL; el=el->next) 
    if (!strnicmp(exportName,el->EqmExportName,DEVICE_NAME_SIZE)) break;
  return el == NULL ? NULL : el->EqmName;
}
#else
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp)
{
  eqm = eqm; prp = prp;
  return NULL;
}
#endif /* !DOS && !NIOS */
void sendPrpSignal(ExportPropertyListStruct *prp,int signal,int conId,int sts)
{
  if (prp == NULL || prp->sigfcn == NULL) return;
  if (prp->sigmask & signal) prp->sigfcn(signal,conId,prp->prpId,sts,prp->sigref);
}
#ifndef SMALL_TINE_LIB
int isAllowedPropertyAccess(const char *eqm,const char *prp,ClnHdr *cln)
{
  ExportListStruct *el=getExportListItem((char *)eqm);
  ExportPropertyListStruct *pl=GetPropertyListStructEx((char *)eqm,(char *)prp,NULL,CA_WRITE);
  int i,usrOK=TRUE,netOK=TRUE;
  if (el == NULL || !el->hasPropertyHandler || pl == NULL || cln == NULL) return TRUE;
  if (pl->aclLst.usrs != NULL && pl->aclLst.nUsrs > 0)
  {
    for (i=0, usrOK=FALSE; i<pl->aclLst.nUsrs && !usrOK; i++)
    {
      if (!strnicmp(cln->usr,pl->aclLst.usrs[i].name,USERNAME_SIZE)) usrOK = TRUE;
    }
  }
  if (!usrOK) return FALSE;
  if (pl->aclLst.nets != NULL && pl->aclLst.nNets > 0)
  {    
    netOK = IsMemberIPNets(pl->aclLst.nNets,pl->aclLst.nets,&cln->adr.ip);
  }
  if (!netOK) return FALSE;
  return TRUE;
}
int isAllowedDeviceAccess(const char *eqm,const char *dev,ClnHdr *cln)
{
  ExportListStruct *el=getExportListItem((char *)eqm);
  int devnr = GetDeviceNumber((char *)eqm,(char *)dev);
  int i,usrOK=TRUE,netOK=TRUE;
  if (el == NULL || !el->hasPropertyHandler || cln == NULL) return TRUE;
  if (devnr < 0 || devnr >= el->EqmNumDevices) return TRUE;
  if ((el->chkAcl & (DEV_CHECK_USR|DEV_CHECK_GRP|DEV_CHECK_NET)) == 0) return TRUE;
  if (el->EqmDevLst == NULL || el->EqmDevLst[devnr].aclLst == NULL) return TRUE;
  if (el->EqmDevLst[devnr].aclLst->usrs != NULL && el->EqmDevLst[devnr].aclLst->nUsrs > 0)
  {
    for (i=0, usrOK=FALSE; i<el->EqmDevLst[devnr].aclLst->nUsrs && !usrOK; i++)
    {
      if (!strnicmp(cln->usr,el->EqmDevLst[devnr].aclLst->usrs[i].name,USERNAME_SIZE)) usrOK = TRUE;
    }
  }
  if (!usrOK) return FALSE;
  if (el->EqmDevLst[devnr].aclLst->nets != NULL && el->EqmDevLst[devnr].aclLst->nNets > 0)
  {    
    netOK = IsMemberIPNets(el->EqmDevLst[devnr].aclLst->nNets,el->EqmDevLst[devnr].aclLst->nets,&cln->adr.ip);
  }
  if (!netOK) return FALSE;
  return TRUE;
}
#endif /* !SMALL_TINE_LIB */
int (*lookupRedirectionNameStub)(char *eqm,char *prpName,char *devName) = NULL;
int GetPrpId(char *eqm,char *prpName)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
  {
    if (prp->isReserved) continue;
    if (!strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE)) return prp->prpId;
  }
  return -1;
}
TINE_EXPORT int hasDevicePropertyList(char *eqm,char *device)
{
  ExportListStruct *el = getExportListItem(eqm);
  int devnr;
  if (el == NULL || el->EqmDevLst == NULL) return FALSE;
  if (device == NULL) return el->hasDevQry;
  devnr = GetDeviceNumber(eqm,device);
  if (devnr < 0 || devnr >= el->EqmNumDevices) return FALSE;
  if (el->EqmDevLst[devnr].prps == NULL) return FALSE;
  return TRUE;
}
TINE_EXPORT int hasPropertyDeviceList(char *eqm,char *property)
{
  ExportPropertyListStruct *prp=GetPropertyListStruct(eqm,property,NULL);
  if (prp == NULL || prp->devNames == NULL) return FALSE;
  return TRUE;
}
TINE_EXPORT void SetRPCCompletion(char *errstr)
{
  printf("SetRPCCompletion has been deprecated: Please use SetEqmCompletion!\n");
  SetEqmCompletion(NULL,errstr);
}
TINE_EXPORT void SetRPCRedirection(char *rdrSrv,char *rdrPrp)
{
  printf("SetRPCRedirection has been deprecated: Please use SetRedirectionParameters!\n");
  SetRedirectionParameters(NULL,NULL,rdrSrv,NULL,rdrPrp);
}
#ifndef SMALL_TINE_LIB
/**
 * \internal
 */
typedef struct 
{
  char expName[EXPORT_NAME_SIZE];
  char cntName[CONTEXT_NAME_SIZE];
  char eqmName[EQM_NAME_SIZE];
  char fecName[FEC_NAME_SIZE];
  char version[16];
  char fecHome[STD_PATH_LENGTH];
  char started[32];
  char cmdline[128];
  char cwd[STD_PATH_LENGTH];
  UINT32 port;
} FECMF;
#define hdrFecMf "EXPORT, CONTEXT, EQM, FEC, PORT, VERSION, FEC_HOME, LAST_STARTED, COMMAND_LINE, WORKING_DIRECTORY\n"
_CSVCOLUMN csvFecMfColumns[] =
{
  { "EXPORT","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,expName),NULL,0 },
  { "CONTEXT","0",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,cntName),NULL,0 },
  { "EQM","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,eqmName),NULL,0 },
  { "FEC","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,fecName),NULL,0 },
  { "PORT","",0,CF_INT32,1,(BYTE *)OFFSETIN(FECMF,port),NULL,0 },
  { "VERSION","",0,CF_TEXT,16,(BYTE *)OFFSETIN(FECMF,version),NULL,0 },
  { "FEC_HOME","",0,CF_TEXT,STD_PATH_LENGTH,(BYTE *)OFFSETIN(FECMF,fecHome),NULL,0 },
  { "LAST_STARTED","",0,CF_TEXT,32,(BYTE *)OFFSETIN(FECMF,started),NULL,0 },
  { "COMMAND_LINE","",0,CF_TEXT,128,(BYTE *)OFFSETIN(FECMF,cmdline),NULL,0 },
  { "WORKING_DIRECTORY","",0,CF_TEXT,STD_PATH_LENGTH,(BYTE *)OFFSETIN(FECMF,cwd),NULL,0 },
};
CSVDB csvFecMfDb =
{
  csvFecMfColumns,
  sizeof(csvFecMfColumns)/sizeof(csvFecMfColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(FECMF),
  0
};
#define FECMF_TEMP_FILENAME "fecmf.tmp"
#define FECMF_LEGACY_FILENAME "fec.mf"
/**
 * \internal
 */
void writeFecEntryToManifest(FECMF *mfe,FILE *fp)
{ /* internal: don't check NULL pointers ... */
  char str[1028];
  char fhstr[128], clstr[128], wdstr[128];
  getValidCsvString(fhstr, mfe->fecHome, 128, 128);
  getValidCsvString(clstr, mfe->cmdline, 128, 128);
  getValidCsvString(wdstr, mfe->cwd, 128, 128);
  snprintf(str,1024,"%.32s, %.32s, %.6s, %.16s, %d, %.16s, %.128s, %.32s, %.128s, %.128s\n",
    mfe->expName,mfe->cntName,mfe->eqmName,mfe->fecName,
    mfe->port,mfe->version,fhstr,mfe->started,clstr,wdstr);
  fwrite(str,strlen(str),1,fp);
}
int updateFecManifest(ExportListStruct *el)
{
# if !defined(FS_RDONLY)
  FILE *fp=NULL;
  char *ctx,mffilename[160],tmpfilename[160],verstr[16],tstr[32],*cmdline,*cwd;
  time_t t = time(NULL);
  struct tm *tms;
  int cc = 0,siz,i,oldmask,found = FALSE;
  BYTE *ver = SystemVersion();
  FECMF *fecmf = NULL, thisfec;

  oldmask = umask(0);
  if (!FecNameRegistered) ccerr(name_unknown);
  if (el == NULL) ccerr(argument_list_error);
  establishAddrCacheFilePath();
  sprintf(verstr,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  if ((tms=localtime(&t)) != NULL)
  {
    sprintf(tstr,"%02d:%02d:%02d %02d.%02d.%04d (%d)",
      tms->tm_hour,tms->tm_min,tms->tm_sec,
      tms->tm_mday,tms->tm_mon+1,1900+tms->tm_year,
      SystemGetProcessId());
  }

  cmdline = SystemGetStartupCommand();
  cwd = SystemGetStartupDirectory();
  ctx = el->EqmContext;
  if (strlen(ctx) == 0) ctx = gDeviceContext;
  strncpy(thisfec.cntName,ctx,CONTEXT_NAME_SIZE);
  strncpy(thisfec.expName,el->EqmExportName,EXPORT_NAME_SIZE);
  strncpy(thisfec.eqmName,el->EqmName,EQM_NAME_SHORTSIZE);
  strncpy(thisfec.fecName,gFecName,FEC_NAME_SIZE);
  thisfec.port = gPortOffset;
  strncpy(thisfec.version,verstr,16);
  strncpy(thisfec.fecHome,fecDbPath,STD_PATH_LENGTH);
  strncpy(thisfec.started,tstr,32);
  if (cmdline != NULL) strncpy(thisfec.cmdline,cmdline,128); else thisfec.cmdline[0] = 0;
  if (cwd!= NULL) strncpy(thisfec.cwd,cwd,STD_PATH_LENGTH); else thisfec.cwd[0] = 0;

  sprintf(mffilename,"%.128s%s",gAddrCacheFilePath,FECMF_FILENAME);
  sprintf(tmpfilename,"%.128s%s",gAddrCacheFilePath,FECMF_TEMP_FILENAME);

  if ((cc=csvReadFile(gAddrCacheFilePath,FECMF_FILENAME,&csvFecMfDb,(void **)&fecmf)) != 0 &&
      (cc=csvReadFile(gAddrCacheFilePath,FECMF_LEGACY_FILENAME,&csvFecMfDb,(void **)&fecmf)) != 0)
  { /* doesn't exist yet -> create it */
    if ((fp=fopen(mffilename,"w")) == NULL) ccerr(file_error);
    fwrite(hdrFecMf,strlen(hdrFecMf),1,fp);
    writeFecEntryToManifest(&thisfec,fp);
    fclose(fp);
    return 0;
  }
  /* does exist: read it in ...*/
  siz = csvFecMfDb.siz;
  if ((fp=fopen(tmpfilename,"w+")) == NULL) ccerr(file_error);
  fwrite(hdrFecMf,strlen(hdrFecMf),1,fp);
  /* clean, edit and dump it back out */
  for (i=0; i<siz; i++)
  {
    if (!strnicmp(fecmf[i].expName,thisfec.expName,EXPORT_NAME_SIZE) &&
        !strnicmp(fecmf[i].cntName,thisfec.cntName,CONTEXT_NAME_SIZE))
    { /* update this entry */
      found = TRUE;
      memcpy(&fecmf[i],&thisfec,sizeof(FECMF));
    }
    if (strlen(fecmf[i].fecName) == 0) continue;
    if (strlen(fecmf[i].eqmName) == 0) continue;
    if (!strcmp(fecmf[i].fecName,FECNAME_UNKNOWN)) continue;
    writeFecEntryToManifest(&fecmf[i],fp);
  }
  if (!found)
  { /* not yet in the manifest ... */
    writeFecEntryToManifest(&thisfec,fp);
  }
  fclose(fp); fp = NULL;

  unlink(mffilename);
  rename(tmpfilename,mffilename);
# ifdef UNIX
  chmod(mffilename, S_ALLWRITE);
# endif
err:
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR, "could not update fec manifest : %.32s",cc2str(cc));
  if (fp) fclose(fp);
  SystemFree(fecmf);
  oldmask = umask(oldmask);
  return cc;
# else
  el = el;
  return 0;
# endif /* !FS_RDONLY */
}
TINE_EXPORT int GetPortOffset(const char *fecName)
{
  int cc = not_implemented, fp = 0, p = -1;
# if defined(WIN32) || defined(UNIX)
  int siz,i;
  FECMF *fecmf = NULL;
  time_t ts, fts=0;
  struct tm tms;
  char *c;
  if (fecName == NULL) ccerr(argument_list_error);
  establishAddrCacheFilePath();
  if ((cc=csvReadFile(gAddrCacheFilePath,FECMF_FILENAME,&csvFecMfDb,(void **)&fecmf)) != 0) ccerr(file_error);
  siz = csvFecMfDb.siz;
  for (i=0; i<siz; i++)
  {
    if (fecmf[i].port > (UINT32)fp) fp = (int)fecmf[i].port;
    if (!strnicmp(fecmf[i].fecName,(char *)fecName,FEC_NAME_SIZE))
    {
      c = fecmf[i].started;
      tms.tm_hour = strtol(c,&c,10); c++;
      tms.tm_min = strtol(c,&c,10); c++;
      tms.tm_sec = strtol(c,&c,10); c++;
      tms.tm_mon = strtol(c,&c,10) - 1; c++;
      tms.tm_mday = strtol(c,&c,10); c++;
      tms.tm_year = (strtol(c,&c,10)%100) + 100; c++;
      tms.tm_isdst = -1;
      ts = mktime(&tms);
      if (fts < ts)
      { /* find the most recent entry using this FEC name */
        if (fts > 0 && p != (int)fecmf[i].port)
        { /* an older entry matches the FEC name but not the port ! */
          feclog("Port Offset: old manifest entry claims fec %.16s has port %d instead of %d",
            fecName,p,fecmf[i].port);
        }
        fts = ts;
        p = (int)fecmf[i].port;
      }
    }
  }
  if (p != -1) 
  { /* found it in the manifest */
    fp = p;
    ccerr(0);
  }
  if (ServerInitialized) ccerr(name_unknown);
  /* not yet initialized and looking for a free port ... */
  fp++;
err:
  SystemFree(fecmf);
  if (cc == file_error)
  { /* as yet, no manifest: first-time server ? */
    fp = 1; /* start off at port = 1 */
    cc = 0;
  }
# else
  p = p;
  if (fecName == NULL) cc = argument_list_error;
# endif
  return cc ? -cc : fp;
}
TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecAddrStruct *fec,ExpDataStruct *srv)
{
  int i,p,clen,cc=0,pass=0;
  NAME32 n32[7];
  NAME16 n16[7];
  DTYPE dout;
  DTYPE din;

  memset(n32,0,7*sizeof(NAME32)); memset(n16,0,7*sizeof(NAME16));
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
retryIllegalFormat:
  if (pass == 0)
  {
    din.dFormat = CF_NAME32;
    din.dArrayLength = 3;
    din.data.vptr = n32;
    n32[0].name[0] = 0;
    n32[1].name[0] = 0;
    n32[2].name[0] = 0;
    if (context != NULL) strncpy(n32[0].name,context,32);
    if (eqmName != NULL) strncpy(n32[1].name,eqmName,32);
    if (exportName != NULL) strncpy(n32[2].name,exportName,32);
    dout.dFormat = CF_NAME32;
    dout.dArrayLength = 7;
    dout.data.vptr = (void *)n32;
  }
  else
  { 
    din.dFormat = CF_NAME16;
    din.dArrayLength = 3;
    din.data.vptr = n16;
    n16[0].name[0] = 0;
    n16[1].name[0] = 0;
    n16[2].name[0] = 0;
    if (context != NULL) strncpy(n16[0].name,context,16);
    if (eqmName != NULL) strncpy(n16[1].name,eqmName,16);
    if (exportName != NULL) strncpy(n16[2].name,exportName,16);
    dout.dFormat = CF_NAME16;
    dout.dArrayLength = 7;
    dout.data.vptr = (void *)n16;
  }
  for (i=0,cc=link_timeout; cc != 0 && i <5; i++)
  {
    cc = ExecLinkEx(SRVEXP_NAME,"SRVADDR",&dout,&din,(short)CA_READ,200);
    if (cc == illegal_format && pass == 0)
    { /* legacy server ? */
      pass++;
      goto retryIllegalFormat;
    }
  }
  if (cc == 0 && pass > 0)
  {
    for (i=0; i<7; i++) 
    {
      memcpy(n32[i].name,n16[i].name,16); n32[i].name[16] = 0;
    }
  }
  if (fec != NULL)
  {
    memset(fec,0,sizeof(FecAddrStruct));
    p = atoi(n32[0].name);
    strncpy(fec->fecName,n32[1].name,FEC_NAME_SIZE);
    fec->portOffset = p;
    clen = gInAdr.ip.ss_family == AF_INET ? 4 : 16;
    memcpy(&fec->netAdr,getInAddr(&gInAdr.ip),clen);
    inet_ntop(gInAdr.ip.ss_family,getInAddr(&gInAdr.ip),fec->strAdr,ADDR_SIZE_IPV6);
    p = !stricmp(n32[6].name,"MODERN") ? ADDRFLAG_MCAST_MODERN : 0;
    fec->addrFlags |= p;
  }
  if (srv != NULL)
  {
    strncpy(srv->FecName,n32[1].name,FEC_NAME_SIZE);
    strncpy(srv->EqmContext,n32[2].name,CONTEXT_NAME_SHORTSIZE);
    strncpy(srv->EqmName,n32[3].name,EQM_NAME_SHORTSIZE);
    strncpy(srv->ExportName,n32[4].name,EXPORT_NAME_SHORTSIZE);
    strncpy(srv->SubSystem,n32[5].name,SUBSYSTEM_NAME_SIZE);
  }
  return cc;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT int AssignPropertyHandlerEx(char *eqm, char *prp, XEQMFCNP hndlr, void *ref)
{
  ExportListStruct *el;
  ExportPropertyListStruct *pls;
  int cc = non_existent_property, idx;
  if ((el=getExportListItem(eqm)) == NULL) ccerr(non_existent_elem);
  idx = (int)(ElfHash((unsigned char *)prp) % PROPERTY_HASH_SIZE);
  for (pls=el->EqmPrpLst[idx]; pls != NULL; pls=pls->next)
  {
    if (pls->isReserved) continue;
    if (!strnicmp(pls->prpName,prp,PROPERTY_NAME_SIZE)) 
    { /* assign to all overloads ! */
      pls->prpHndlr = (EQMFCNP)hndlr;
      pls->hdlrRef = ref;
      cc = 0; /* clear 'non_existent_property' */
    }    
  }
  if (cc == 0) el->hasPropertyHandler = (hndlr != NULL) ? TRUE : (el->eqmFcn != NULL);
err:
  feclog("AssignPropertyHandler (eqm %.6s prop %.64s) : %.32s",
    eqm == NULL ?"(null)" : eqm, prp == NULL ? "(null)" : prp, cc2str(cc));
  return cc;
}
TINE_EXPORT int AssignPropertyHandler(char *eqm, char *prp, EQMFCNP hndlr)
{
  return AssignPropertyHandlerEx(eqm,prp,(XEQMFCNP)hndlr,NULL);
}
TINE_EXPORT int RegisterPropertyAndHandlerEx(
    char *eqm,
    XEQMFCNP hndlr,
    char *prp,DTYPE *dout,DTYPE *din,short acc,short atype,
    UINT16 rowlen,char *dsc,
    PrpEgu *egu,PrpEgu *xegu, 
    int pid,void *ref
) 
{ /* assign to the internal equipment module */
  int cc, rid;  
  if (eqm == NULL) eqm = INTEQM_TAG;
  if((rid=RegisterPropertyInformationEx(eqm,prp,dout,din,acc,atype,rowlen,dsc,egu,xegu,pid,NULL)) < 0 )
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "RegisterPropertyAndHandler: RegisterProperty %.64s failed : %.32s",prp == NULL ? "(null)" : prp, cc2str(-pid));
    return -rid;
  }  
  if ((cc=AssignPropertyHandlerEx(eqm, prp, hndlr, ref)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "RegisterPropertyAndHandler : AssignPropertyHandler(%s,%s) failed : %.32s", 
      eqm, prp == NULL ? "(null)" : prp, cc2str(cc));
    return cc;
  }
  return 0;
}
TINE_EXPORT int RegisterPropertyAndHandler(
    char *eqm,
    EQMFCNP hndlr,
    char *prp,DTYPE *dout,DTYPE *din,short acc,short atype,
    UINT16 rowlen,char *dsc,
    PrpEgu *egu,PrpEgu *xegu, 
    int pid
) 
{
  return RegisterPropertyAndHandlerEx(eqm,(XEQMFCNP)hndlr,prp,dout,din,acc,atype,rowlen,dsc,egu,xegu,pid,NULL);
}
TINE_EXPORT int RegisterServer(char *expName, int numdevices, char *deviceNameList[]) 
{
  int cc, i;  
  if (expName == NULL || numdevices == 0) ccerr(argument_list_error);
  if ((cc=RegisterEquipmentModule(expName, INTEQM_TAG, numdevices, NULL, NULL, NULL, 0, NULL)) != 0) ccerr(cc);
  if (deviceNameList == NULL) return 0; /* just numbers at this stage */
  for (i = 0; i < numdevices; i++) 
  {
    if ((cc=RegisterDeviceName(INTEQM_TAG,deviceNameList[i],i)) != 0) 
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "RegisterServer %.32s : RegisterDeviceName %.64s failed : %.32s", 
        expName, deviceNameList[i], cc2str(cc));
    }
  }
err:
  feclog("RegisterServer: %s (%d devices) : %.32s", expName == NULL ? "(null)" : expName, numdevices, cc2str(cc));
  return cc;
}
extern int gExportListEmptyInitialization;
int _registerEquipmentModule(
    char *expName,char *eqmName,int numdevices,
    void *fcn,void *ini,void *tsk,int rate,void *exi,void *ref
)
{
  int cc, idx;
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
# ifdef MULTITHREADED
  BkgThreadFcn *btf;
# endif
  fixFecRepository();
  strtrm(expName);
  if (numdevices == 0) numdevices = 1;
  if ((cc=RegisterExport(expName,eqmName,numdevices)) != 0) 
  {
    feclog("RegisterExport %.32s (%.6s) %d devices : %.32s",
      expName == NULL ? "(null)" : expName,
      eqmName == NULL ? "(null)" : eqmName,
      numdevices, cc2str(cc));
    return cc;
  }
  if ((el=getExportListItem(eqmName)) == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "_registerEquipmentModule (%.6s) : non existent element",
      eqmName == NULL ?"(null)" : eqmName);
    return non_existent_elem;
  }
  if (!strcmp(eqmName,ENS_TAG)) gIgnoreVolatileAddressCache = TRUE;
  if (!strcmp(eqmName,ENS_TAG) || 
      !strcmp(eqmName,GENS_TAG)||
      !strcmp(eqmName,CAS_TAG) ||
      !strcmp(gFecName,GENS_FEC))
  { /* don't bother with CAS alarm notification */
    el->almdone = TRUE;
  }
  if (ref != NULL) 
  {
    el->hasReference = TRUE;
    el->EqmRef = ref;
  }
  el->eqmFcn  = (EQMFCNP)fcn;
  el->hasPropertyHandler = (fcn != NULL);
  el->eqmBkg  = (EQMTSKP)tsk;

  if (rate <= 0) rate = 1000;
# if defined(MULTITHREADED)
  if (tsk != NULL && useMultiThreadedBackgroundTasks)
  {
    feclog("FEC: module %.6s is using multi-threaded background tasks",eqmName);
    if ((btf=(BkgThreadFcn *)SystemCalloc(1,sizeof(BkgThreadFcn))) == NULL) 
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "_registerEquipmentModule : cannot allocate background thread task!");
      return out_of_local_memory;
    }
    btf->rate = rate; btf->tsk = (EQMTSKP)tsk; btf->elRef = (void *)el;
    btf->hasReference = el->hasReference; btf->EqmRef = ref;
    el->bkgbusy = FALSE;
    el->hEqmBkg = bkgCreateThread(btf);
    el->eqmBkg  = NULL;     /* let the thread handle the task */
  }
# endif
  el->rateBkg = rate;
  if (!gIsCyclePollingMode && el->eqmBkg && rate && rate < MaxPollingRate)
  {
    if (gSystemCycleDeadband > (UINT32)rate) 
    {
      feclog("adjust cycle deadband (due to background task) from %d to %d msec",
        gSystemCycleDeadband,rate);
      gSystemCycleDeadband = (UINT32)rate;
    }
    MaxPollingRate = rate;
  }
  el->eqmIni = (EQMTSKP)ini;
  el->eqmExi = (EQMTSKP)exi;

  if (WaitForSystemMutex(hSystemInitMutex,-1) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "_registerEquipmentModule : cannot get system initialization mutex !");
    return mutex_error;
  }
  feclog("%12s task registration: %s %s %s",eqmName,
          ini ? "INIT" : "----",
          tsk ? "BCKG" : "----",
          exi ? "EXIT" : "----");

  if (EquipmentModulesInitialized)
  {
    feclog("%.6s : late equipment module registration",el->EqmName);
    if (gExportListEmptyInitialization)
    {
      gLastMinuteInitializationDone = FALSE;
      feclog("reset last-minute initialization");
    }
    if (el->eqmIni) TSKCALL(el,eqmIni);
    if (!gIsCentralAlarmServer) rmvAlarmsFromCAS(el->EqmName);
#   if !defined(VXWORKS)
    historyInit(el->EqmExportName,el->EqmName);
    GetAlarmWatchTable(el->EqmName);
    getAllowedUsers(el->EqmName);
    getAllowedNets(el->EqmName);
    getFollowersFromFile(el->EqmName);
#   endif
#   ifdef MULTITHREADED
    eqmInitMutexSet((void *)el);
#   endif
    el->isidle = gEqmInitializedIdleState;
    el->inidone = TRUE;
    el->eqmBaseline = el->eqmStartupTime = time(NULL);
    if (gLastMinuteInitializationDone && gStartupDebug)
    {
      int fgtty = foregroundTTY;
      foregroundTTY = TRUE;
      ttyoutput("Exporting the following properties for %.32s (%.6s) via configuration file:",el->EqmExportName,el->EqmName);
      for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
        for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
      {
        ttyoutput("\t%.64s (%s) -> %.64s <-",prp->prpName,itom(prp->prpAccessMode),prp->prpDescription);
      }
      foregroundTTY = fgtty;
    }
    feclog("%.6s marked as initialized (idle state : %s)",el->EqmName,el->isidle ? "TRUE" : "FALSE");
  }
# ifndef SMALL_TINE_LIB
  updateFecManifest(el);
# endif
  ExportsRegistered = FALSE;
  ReleaseSystemMutex(hSystemInitMutex);
  return 0;
}
/** \internal */
int IsPropertyCalledInSeparateThread(ContractListStruct *cl)
{ /* cl is never NULL here : no need to check */
# ifdef MULTITHREADED
  char *eqm = cl->contract.EqmName;
  char *prp = cl->contract.EqmProperty;
  char *dev = cl->contract.EqmDeviceName;
  int i;
  ExportPropertyListStruct *pl;
  if ((pl=GetPropertyListStruct(eqm,prp,dev)) != NULL) return pl->runInSeparateThread;
  if ((i=GetStockPropertyId(prp)) >= 0) return gStockProperty[i].runInSeparateThread;
  if (cl->exp == NULL) cl->exp = GetExportListItem(eqm);
  if ((i=IsMetaProperty(prp, NULL, NULL, FALSE)) > 0 &&
      !isRegisteredMetaProperty(cl->exp, eqm, prp, dev))
  { /* returns the meta table index + 1 ... */
    i--;
    return StockModifier[i].runInSeparateThread;
  }
# else
  UNUSED(cl);
# endif
  return FALSE;
}
TINE_EXPORT int ResetServerBaseline(const char *eqm)
{
  int cc = 0;
  ExportListStruct *el=getExportListItem((char *)eqm);
  if (eqm == NULL) ccerr(argument_list_error);
  if (el == NULL) ccerr(non_existent_elem);
  el->eqmBaseline = time(NULL);
err:
  return cc;
}

TINE_EXPORT int MaxNumContracts = CONTRACTLIST_CAPACITY;
TINE_EXPORT int MaxNumClients   = CLIENTLIST_CAPACITY;
int ContractListCapacity = CONTRACTLIST_CAPACITY;
int ClientListCapacity = CLIENTLIST_CAPACITY;
#ifndef SMALL_TINE_LIB
TINE_EXPORT int AssignDeviceAccessList(const char *eqm,const char *dev,NAME16 *users,int nusers)
{
  int cc = 0;
  ExportListStruct *el=getExportListItem((char *)eqm);
  int devnr = GetDeviceNumber((char *)eqm,(char *)dev);
  if (eqm == NULL) ccerr(argument_list_error);
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if (devnr < 0 || devnr > el->EqmNumDevices) ccerr(illegal_device);
  if (el->EqmDevLst[devnr].aclLst == NULL)
  { /* not yet allocated */
    if ((el->EqmDevLst[devnr].aclLst=(AclInfoType *)SystemCalloc(1,sizeof(AclInfoType))) == NULL) 
      ccerr(out_of_server_memory);
  }
  if (nusers > MaxNumAllowedUsers) nusers = MaxNumAllowedUsers;
  if (el->EqmDevLst[devnr].aclLst->usrs == NULL && 
      (el->EqmDevLst[devnr].aclLst->usrs=(NAME16 *)SystemCalloc(MaxNumAllowedUsers,sizeof(NAME16))) == NULL) 
    ccerr(out_of_server_memory);
  memcpy(el->EqmDevLst[devnr].aclLst->usrs,users,sizeof(NAME16)*nusers);
  el->EqmDevLst[devnr].aclLst->nUsrs = nusers;
  el->chkAcl |= DEV_CHECK_USR;
err:
  return cc;
}
TINE_EXPORT int AssignDeviceNetsList(const char *eqm,const char *dev,NAME64 *ipnets,int nipnets)
{ 
  int i, cc = 0, fam = gAfInet;
  INADR inadr;
  ExportListStruct *el=getExportListItem((char *)eqm);
  int devnr = GetDeviceNumber((char *)eqm,(char *)dev);
  if (el == NULL) ccerr(argument_list_error);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if (devnr < 0 || devnr > el->EqmNumDevices) ccerr(illegal_device);
  if (el->EqmDevLst[devnr].aclLst == NULL)
  { /* not yet allocated */
    if ((el->EqmDevLst[devnr].aclLst=(AclInfoType *)SystemCalloc(1,sizeof(AclInfoType))) == NULL) 
      ccerr(out_of_server_memory);
  }
  if (nipnets > MaxNumAllowedIPNets) nipnets = MaxNumAllowedIPNets;
  if (el->EqmDevLst[devnr].aclLst->nets == NULL && 
      (el->EqmDevLst[devnr].aclLst->nets =
        (SCKACL *)SystemCalloc(MaxNumAllowedIPNets,sizeof(SCKACL))) == NULL) 
    ccerr(out_of_server_memory);
  for (i=0; i<nipnets; i++)
  { /* port is irrelevant except for possible ipv4 broadcast lists */
    inet_pton(fam,ipnets[i].name,&inadr);
    sckadr_in(&el->EqmDevLst[devnr].aclLst->nets[i].addr,(BYTE *)&inadr,(UINT16)gtGCastPort,fam);
  }
  el->EqmDevLst[devnr].aclLst->nNets = nipnets;
  el->chkAcl |= DEV_CHECK_NET;
err:
  return cc;
}
TINE_EXPORT int AssignPropertyAccessList(const char *eqm,const char *prp,NAME16 *users,int nusers)
{
  int cc = 0;
  ExportListStruct *el=getExportListItem((char *)eqm);
  ExportPropertyListStruct *p, *pl=GetPropertyListStruct((char *)eqm,(char *)prp,NULL);
  if (el == NULL || !el->hasPropertyHandler) return argument_list_error;
  if (pl == NULL) ccerr(illegal_property);
  if (nusers > MaxNumAllowedUsers) nusers = MaxNumAllowedUsers;
  if (pl->aclLst.usrs == NULL && 
      (pl->aclLst.usrs=(NAME16 *)SystemCalloc(MaxNumAllowedUsers,sizeof(NAME16))) == NULL) 
    ccerr(out_of_server_memory);
  memcpy(pl->aclLst.usrs,users,sizeof(NAME16)*nusers);
  pl->aclLst.nUsrs = nusers;
  for (p=pl->next; p != NULL; p=p->next)
  { /* set references for any overloads as well */
    if (strnicmp(p->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    p->aclLst.nUsrs = pl->aclLst.nUsrs;
    p->aclLst.usrs = p->aclLst.usrs;
  }
err:
  return cc;
}
TINE_EXPORT int AssignPropertyNetsList(const char *eqm,const char *prp,NAME64 *ipnets,int nipnets)
{ 
  int i, cc = 0, fam = gAfInet;
  INADR inadr;
  ExportListStruct *el=getExportListItem((char *)eqm);
  ExportPropertyListStruct *p, *pl=GetPropertyListStruct((char *)eqm,(char *)prp,NULL);
  if (el == NULL || !el->hasPropertyHandler) return argument_list_error;
  if (pl == NULL) ccerr(illegal_property);
  if (nipnets > MaxNumAllowedIPNets) nipnets = MaxNumAllowedIPNets;
  if (pl->aclLst.nets == NULL && 
      (pl->aclLst.nets=(SCKACL *)SystemCalloc(MaxNumAllowedIPNets,sizeof(SCKACL))) == NULL) 
    ccerr(out_of_server_memory);
  for (i=0; i<nipnets; i++)
  { /* port is irrelevant */
    inet_pton(fam,ipnets[i].name,&inadr);
    sckadr_in(&pl->aclLst.nets[i].addr,(BYTE *)&inadr,(UINT16)gtGCastPort,fam);
  }
  pl->aclLst.nNets = nipnets;
  for (p=pl->next; p != NULL; p=p->next)
  { /* set references for any overloads as well */
    if (strnicmp(p->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    p->aclLst.nNets = pl->aclLst.nNets;
    p->aclLst.nets = p->aclLst.nets;
  }
err:
  return cc;
}
#endif /* SMALL_TINE_LIB */
CONSIG gConSigfcn = NULL;
int gConSigMask = 0;
void *gConSigRef = NULL;
void sendContractSignal(const char *eqm,const char *dev,const char *prp,int signal,int sts)
{
  if (gConSigfcn == NULL) return;
  if (gConSigMask & signal) gConSigfcn(eqm,dev,prp,sts,signal,gConSigRef);
}
TINE_EXPORT int RegisterPropertySignalFunction(const char *eqm,const char *prp,PRPSIG fcn,int mask,void *ref)
{
  int cc = 0;
  ExportListStruct *el=getExportListItem((char *)eqm);
  ExportPropertyListStruct *p, *pl=GetPropertyListStruct((char *)eqm,(char *)prp,NULL);
  if (el == NULL || !el->hasPropertyHandler) return argument_list_error;
  if (pl == NULL) ccerr(illegal_property);
  pl->sigfcn = mask != 0 ? fcn : NULL;
  pl->sigmask = mask;
  pl->sigref = ref;
  for (p=pl->next; p != NULL; p=p->next)
  { /* set same for any overloads */
    if (strnicmp(p->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    p->sigfcn = pl->sigfcn;
    p->sigmask = pl->sigmask;
    p->sigref = pl->sigref;
  }
err:
  return cc;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT void SetContractSignalFunction(CONSIG fcn,int mask,void *ref)
{
  gConSigfcn = fcn;
  gConSigMask = mask;
  gConSigRef = ref;
  return;
}
TINE_EXPORT int RegisterStateChangeCallback(SCCBFCNP fcn,const char *eqm,const char *stateKey, void *reference)
{
  ExportListStruct *el=getExportListItem((char *)eqm);
  char *ctx;
  if (el == NULL) return non_existent_elem;
  el->staChgCb = fcn;
  el->staChgRef = reference;
  ctx = getPrimaryContext(el->EqmContext,el->EqmSubSystem);
  return SystemStartStateTrigger(ctx,stateKey);
}
TINE_EXPORT int GetContractListCapacity(void) 
{ 
  return ContractListCapacity;
}
TINE_EXPORT void SetContractListCapacity(int value)
{ /* ContractListCapacity set following allocation*/
  if (value < 10) return;
  feclog("Set contract list capacity to %d",value);
  MaxNumContracts = value;
}
TINE_EXPORT int GetClientListCapacity(void) 
{ 
  return ClientListCapacity;
}
TINE_EXPORT void SetClientListCapacity(int value)
{ /* ClientListCapacity set following allocation*/
  if (value < 5) return;
  MaxNumClients = value;
  feclog("Set client list capacity to %d", value);
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT void FlushContractTable(void)
{
  flushContractTable();
}
TINE_EXPORT int AddFieldToStruct(char *tag,int addr,int size,int fmt,char *field)
{
  struct structStruct *s, *ss;
  char *c, *fld, subtag[68];
  int i, flen=0;

  if (isInvalidStructureTag(tag)) return invalid_structure_tag;
  if (field == NULL) return invalid_field;
  flen = (int)strlen(field);
  for (i=0; i<flen; i++) if ((BYTE)field[i] > 127) return invalid_field;
  if ((s=getTaggedStruct(tag)) == NULL) return out_of_local_memory;
  if (s->sealed) return struct_sealed;
  fld = field;
  if (fmt == CF_STRUCT)
  { /* nested structure */
    if ((c=strchr(field,'<')) != NULL) 
    {
      strncpy(subtag, &field[1], 64); subtag[64] = 0;
      if ((c=strchr(subtag,'>')) == NULL) return invalid_field;
      *c++ = 0; fld = c;
    }
    else
    {
      strncpy(subtag,field,64);
    }
    if ((ss=findstruct(subtag)) == NULL) return illegal_format;
    s->rawlen += size*ss->rawlen;
  }
  else
  {
    subtag[0] = 0;
    if (!isVariableLengthFormat((short)fmt)) 
    { /* normal case */
      s->rawlen += size*getFormatSize(fmt)+getFormatHeaderSize(fmt);
    }
    else
    { /* will have an extended space ! */
#     ifdef DOS
      return illegal_format;
#     else
      if (fmt == CF_IMAGE) size = 1; /* special case */
      s->rawlen += size*(getFormatHeaderSize(fmt)+2*sizeof(SINT32));
      s->hasExtSpace = TRUE;
#     endif
    }
  }
  return addtostruct(&s->s,addr,size,fmt,subtag,fld);
}
TINE_EXPORT int AssignNumDevices(char *eqm,int num)
{
  ExportListStruct *el;
  int cc=0;

  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (num < 0 || num >= el->EqmDevLstCapacity) ccerr(out_of_range);
  el->EqmNumDevices = num;      /* this is a one-way street  */
  makeEqpDevicesHashTable(el);  /* in case we're now beyond the HASH_PAYOFF */
# ifndef SMALL_TINE_LIB
  if (strlen(el->EqmGroup) > 0)
  { /* belongs to a group : notify the group server of any change in the device list */
    joinEnsGroup(el);
  }
# endif
err:
  return cc;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int AssignPropertyList(char *eqm,char *devname,char *listname,int listsize,NAME64 *list)
{
  int i, k, cc = 0;
  ExportListStruct *el;
  EqmDevicePropertyList *pl = NULL;

  if (eqm == NULL || devname == NULL || listname == NULL || list == NULL) ccerr(argument_list_error);
  if (listsize < 1) ccerr(illegal_data_size);

  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmDevLst == NULL) ccerr(code_failure);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqmDevLstCapacity) ccerr(device_not_connected);
  for (k=0; k<el->EqmDevLstCapacity; k++)
  {
    if (i == k) continue;
    if (el->EqmDevLst[k].prps == NULL) continue;
    if (strncmp(el->EqmDevLst[k].prps->filename,listname,64)) continue;
    if (el->EqmDevLst[k].prps->prpListSize != listsize)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "attempt to assign known property list %.64s to %.64s with inconsistent list size (%d vs. %d)",
        listname,devname,listsize,el->EqmDevLst[k].prps->prpListSize);
    }
    pl = el->EqmDevLst[k].prps;
    break;
  }
  if (pl == NULL)
  { /* a new one */
    if ((pl=(EqmDevicePropertyList *)SystemCalloc(listsize,sizeof(EqmDevicePropertyList))) == NULL) 
      ccerr(out_of_server_memory);
    if ((pl->prpList=(NAME64 *)SystemCalloc(listsize,sizeof(NAME64))) == NULL) 
      ccerr(out_of_server_memory);
    memcpy(pl->prpList,list,listsize*sizeof(NAME64));
    strncpy(pl->filename,listname,64);
    pl->prpListSize = listsize;
  }
  el->EqmDevLst[i].prps = pl;
  el->hasDevQry = TRUE;
err:
  return cc;
}
TINE_EXPORT int ExecLocalLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  ContractListStruct cl;
  ClnHdr cln;
  BYTE *datain = NULL,*dataout = NULL;
  char fulname[128],sts[STATUS_SIZE];
  char *en = NULL,*dn = NULL,*cp=NULL;
  int cc=0, dsiz=0, ssiz=0, n;

  strncpy(fulname,devName,128); fulname[127] = 0;
  if ((en=strtok_r(fulname,DELIMITER,&cp)) == NULL) ccerr(argument_list_error);
  if ((dn=strtok_r(NULL,DELIMITER,&cp)) == NULL) dn = &fulname[127];

  memset(&cln,0,sizeof(ClnHdr));
  strncpy(cln.usr,gFecName,USERNAME_SIZE);
  strncpy(cln.typ, "FEC", USERTYPE_SIZE);
  memset(&cl,0,sizeof(ContractListStruct));
  strncpy(cl.contract.EqmName,en,EQM_NAME_SHORTSIZE);
  strncpy(cl.contract.EqmDeviceName,dn,DEVICE_NAME_SIZE);
  strncpy(cl.contract.EqmProperty,devProperty,PROPERTY_NAME_SIZE);
  cl.exp = GetExportListItem(en);
  if (din != NULL)
  {
    cl.contract.EqmSizeIn = din->dArrayLength;
    cl.contract.EqmFormatIn = BFMT(din->dFormat);
    strncpy(cl.contract.strTagIn,din->dTag,TAG_NAME_SIZE);
    datain = din->data.bptr;
  }
  if (dout != NULL)
  {
    cl.contract.EqmSizeOut = dout->dArrayLength;
    cl.contract.EqmFormatOut = BFMT(dout->dFormat);
    strncpy(cl.contract.strTagOut,dout->dTag,TAG_NAME_SIZE);
    dataout = dout->data.bptr;
  }
  cl.contract.EqmAccess = (BYTE)access;

  if ((cc=IsIllegalProperty(cl.contract.EqmName,
        cl.contract.EqmProperty,cl.contract.EqmDeviceName,
        &cln,access,&cl.exp,NULL,sts)) != 0) ccerr(cc);

  setDataTimeStamp(makeDataTimeStamp());
  cl.dataOut = dataout;
  cl.dataIn = datain;
  cl.dataBuf = srvWorkArea;
  cl.runInSeparateThread = IsPropertyCalledInSeparateThread(&cl);
  /* TODO: doContract must be called on the SystemCycle thread if a mt call */
  if (cl.runInSeparateThread) ccerr(not_implemented);
  if (dout != NULL)
  {
    dsiz = cl.contract.EqmSizeOut*fmtsizeof(dout->dFormat);
    dsiz += getFormatHeaderSize(dout->dFormat);
    if (dout->dFormat == CF_STRUCT) ssiz = GetStructSize(dout->dTag);
    if (ssiz > 0) dsiz *= ssiz;
    cl.dataOutCapacity = dsiz;
    if (cl.dataOut == NULL) cl.dataOut = (BYTE *)SystemCalloc(1,dsiz);
    if (cl.dataOut == NULL) ccerr(out_of_client_memory);
    memset(cl.dataOut,0,dsiz); 
  }
  cc = doContract(&cl,TRUE);
  if (cc & 0xf000) 
  { /* CE_RENEGOTIATE or CE_SENDDATA */
    dsiz = renegotiateContract(&cl,(cc & 0xfff));
    if (dsiz == 0) cc &= ~CE_SENDDATA; /* no data to send */
    cc &= ~CE_RENEGOTIATE;
    if (dout != NULL)
    {
      n = dsiz/fmtsizeof(dout->dFormat);
      if (access & CA_MUTABLE) dout->dArrayLength = n;
    }
  }
  if (dsiz > 0 && cl.dataOut != NULL) memcpy(cl.dataOut,cl.dataBuf,(size_t)dsiz);
err:
  return cc;
}
int getKnownENSes(NAME32 *lst,int *siz)
{
  NAME32 eqmn;
  int n, cc;
  DTYPE dout, din;
  
  if (lst == NULL || siz == NULL) return argument_list_error;
  n = *siz;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = n;
  dout.dFormat = CF_NAME32;
  dout.data.vptr = lst;
  din.dArrayLength = 1;
  din.dFormat = CF_NAME16;
  din.data.vptr = &eqmn;
  strcpy(eqmn.name,"(ENSEQM)");
  if ((cc=ExecLinkEx("/SITE/ENS/SITE","SERVERS",&dout,&din,CA_READ|CA_MUTABLE,500)) == 0)
  {
    n = dout.dArrayLength;
  }
  *siz = n;
  return cc;
}
TINE_EXPORT int FindNameServerOnNetwork()
{
  int cc = 0;
  FecAddrStruct fec;
  ExpDataStruct srv;
  if (gotNameServerAddress) return 0;
  if ((cc=reserveFecAddressSpace()) != 0) ccerr(cc);
  gEnsAddrResolutionInProgess = TRUE;
  if ((cc=FindServerOnNetwork(NULL,ENS_TAG,NULL,&fec,&srv)) == 0)
  {
#   define MAX_NUMENS 4
    NAME32 lst[4];
    int i, k, n = 4;
    if ((NameServer=(FecAddrStruct *)SystemCalloc(MAX_NUMENS,sizeof(FecAddrStruct))) == NULL) ccerr(out_of_server_memory);
    memcpy(NameServer,&fec,sizeof(FecAddrStruct));
    numNameServers = 1;
    gotNameServerAddress = 1;
    addNameServerToTables();
    feclog("Using ENS %.16s at address %.40s",NameServer->fecName,NameServer->strAdr);
    if (getKnownENSes(lst,&n) == 0)
    {
      for (i=0; i<n; i++)
      {
        GetServerAddr("SITE",lst[i].name,&srv,&fec);
        for (k=0; k<numNameServers; k++) if (!memcmp(&fec.netAdr,&NameServer[k].netAdr,16)) break;
        if (k < numNameServers) continue;
        if (numNameServers >= MAX_NUMENS) break;
        memcpy(&NameServer[numNameServers],&fec,sizeof(FecAddrStruct));
        feclog("Append ENS %.16s at address %.40s to conigured ENS table",
          NameServer[numNameServers].fecName,NameServer[numNameServers].strAdr);
        numNameServers++;
      }
    }
  }
err:
  gEnsAddrResolutionInProgess = FALSE;
  feclog("FindNameServerOnNetwork : %.32s",cc2str(cc));
  return cc;
}
#endif
TINE_EXPORT char *GetCaller(char *eqm)
{
  static char name[USERNAME_SIZE+4];
  NAME16 un;
  int num = 1;

  if (GetCallerInfo(eqm,&un,NULL,NULL,NULL,&num) != 0) return NULL;
  strncpy(name,un.name,USERNAME_SIZE);
  return name;
}
TINE_EXPORT int GetCallerInfo(char *eqm,NAME16 *un,BYTE *ipx,UINT32 *ip,short *prot,int *num)
{
# define SIZE_INFOLIST_QRY_BUFFER 10
  static ClnInfo clnInfoList[SIZE_INFOLIST_QRY_BUFFER];
  char scratch[48];
  int cc, i, n, fam;
  UNUSED(ipx);
  if (num == NULL || *num < 1) return buffer_too_small;
  n = *num;
  if (*num > SIZE_INFOLIST_QRY_BUFFER) *num = SIZE_INFOLIST_QRY_BUFFER;
  if ((cc=GetCallerInformation(eqm,clnInfoList,num)) != 0) return cc;
  for (i=0; i<n && i<(*num); i++)
  {
    if (un != NULL) strncpy(un[i].name,clnInfoList[i].userName,USERNAME_SIZE);
    if (ip != NULL)
    {
      if ((fam=clnInfoList[i].addr.ss_family) != AF_INET) continue;
      inetStrAdr(&clnInfoList[i].addr,scratch,FALSE);
      inet_pton(fam,scratch,&ip[i]);
    }
    if (prot != NULL) prot[i] = clnInfoList[i].tineProtocol;
  }
  return 0;
}
TINE_EXPORT int GetCallerInformation(char *eqm,ClnInfo *clnInfoList,int *num)
{
  ExportListStruct *el = getExportListItem(eqm);
  THRHANDLE tid = GetThreadHandle();
  ClnLst *clnl = NULL;
  int i;

  if (num == NULL || *num < 1) return buffer_too_small;
  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;/* just a place-holder */
    if (!ContractList[i]->callPending) continue;
    if (ContractList[i]->threadId != tid) continue;
    if (ContractList[i]->exp != el) continue;
    break;
  }
  if (i != ncontracts)
  { /* found a contract list entry (normal case) */
    clnl = ContractList[i]->clnLst;
  }
  for (i=0; clnl != NULL; clnl=clnl->next,i++)
  { /* fill in the information ... */
    if (clnl->client == NULL) break;
    if (i < *num)
    {
      clnInfoList[i].accessMode = clnl->status.mode;
      clnInfoList[i].pollingRate = clnl->status.pollingInterval;
      clnInfoList[i].starttime = clnl->status.starttime;
      clnInfoList[i].counter = clnl->status.counter;
      memcpy(clnInfoList[i].userName,clnl->client->usr,USERNAME_SIZE);
      clnInfoList[i].addr = clnl->client->adr.ip;
      clnInfoList[i].ncontracts = clnl->client->ncontracts;
      clnInfoList[i].tineProtocol = clnl->client->tineProtocol;
      clnInfoList[i].inetProtocol = clnl->client->inetProtocol;
    }
  }
  if (i == 0)
  { /* must be a local (internal) call to the equipment module */
    clnInfoList[0].accessMode = CM_SINGLE;
    clnInfoList[0].pollingRate = 0;
    clnInfoList[0].starttime = (UINT32)time(NULL);
    clnInfoList[0].counter = 1;
    strncpy(clnInfoList[0].userName,"LOCAL",USERNAME_SIZE);
    memset(&clnInfoList[0].addr,0,sizeof(SCKADR));
    clnInfoList[0].ncontracts = 1;
    clnInfoList[0].tineProtocol = 0;
    clnInfoList[0].inetProtocol = 0;
    i = 1;
  }
  *num = i;
  return 0;
}
TINE_EXPORT void **GetContractDataReferenceByEqmName(char *eqm)
{
  return GetContractDataReference(getExportListItem(eqm));
}
TINE_EXPORT void **GetContractDataReference(ExportListStruct *el)
{
  THRHANDLE tid = GetThreadHandle();
  int i;
  if (el == NULL || !el->isinside) return NULL;
  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;/* just a place-holder */
    if (ContractList[i]->threadId != tid) continue;
    if (ContractList[i]->exp != el) continue;
    return &ContractList[i]->prpDataRef;
  }
  return NULL;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT char *GetExportedName(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return NULL;
  return el->EqmExportName;
}
TINE_EXPORT char *GetDeviceName(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  static char devname[DEVICE_NAME_SIZE+4];

  sprintf(devname,"DEVICE %d",devnr);
  if (el == NULL || el->EqmDevLst == NULL || 
      devnr < 0 || devnr >= el->EqmNumDevices) return devname;
  return el->EqmDevLst[devnr].name;
}
TINE_EXPORT char *GetDeviceDescription(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  static char lcldesc[64];
  
  if (el == NULL || devnr < 0 || devnr >= el->EqmNumDevices) return "no such device";
  if (el->EqmDevLst != NULL && el->EqmDevLst[devnr].desc != NULL)
  {
    return el->EqmDevLst[devnr].desc;
  }
  sprintf(lcldesc,"%.32s device %d",el->EqmExportName,devnr);
  return lcldesc;
}
TINE_EXPORT char *GetDeviceLocation(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  
  if (el == NULL || devnr < 0 || devnr >= el->EqmNumDevices) return "no such device";
  if (el->EqmDevLst != NULL && el->EqmDevLst[devnr].loc != NULL)
  {
    return el->EqmDevLst[devnr].loc;
  }
  return gFecInfo.loc;
}
TINE_EXPORT int GetDeviceMask(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  
  if (el == NULL || devnr < 0 || devnr >= el->EqmNumDevices) return 0;
  if (el->EqmDevLst != NULL) return el->EqmDevLst[devnr].mask;
  return 0;
}
TINE_EXPORT int GetDeviceOfflineStatus(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  
  if (el == NULL || devnr < 0 || devnr >= el->EqmNumDevices) return 0;
  if (el->EqmDevLst != NULL) return (el->EqmDevLst[devnr].mask & DEVICE_OFFLINE);
  return 0;
}
TINE_EXPORT float GetDeviceZPosition(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  
  if (el == NULL || devnr < 0 || devnr >= el->EqmNumDevices) return 0;
  if (el->EqmDevLst != NULL) return el->EqmDevLst[devnr].zpos;
  return 0;
}
TINE_EXPORT NAME64 *GetExportedDeviceList(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el != NULL) return (NAME64 *)el->EqmDevLst[0].name;
  return NULL;
}
#endif /* !SMALL_TINE_LIB */
char *getFirstDevice(char *eqm,char *prpname)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp;
  if (el == NULL) return NULL;
  if (prpname != NULL && (prp=AssignDeviceListFromProperty(eqm,prpname)) != NULL)
  {
    return prp->devNames[0].name;
  }
  if (el->EqmDevLst == NULL) return NULL;
  return el->EqmDevLst[0].name;
}
TINE_EXPORT int GetDeviceNumberEx(char *eqm,char *devname,char *prpname)
{
  return GetValidDeviceNumber(eqm,devname,prpname,0);
}
TINE_EXPORT int GetValidDeviceNumber(char *eqm,char *devname,char *prpname,int ceiling)
{
  char lclname[DEVICE_NAME_SIZE+4];
  int idx,i,n;
  HashXRefStruct *xref;
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp;

  if (el == NULL) return -non_existent_elem;
  if (devname == NULL || strlen(devname) == 0) return -argument_list_error;
  n = el->EqmNumDevices;
  if (prpname != NULL && (prp=AssignDeviceListFromProperty(eqm,prpname)) != NULL)
  { /* TODO: replace this too with a hash table ... */
    for (idx=0; idx<prp->numDevices; idx++) if (!strnicmp(devname,prp->devNames[idx].name,DEVICE_NAME_SIZE)) return idx;
    n = prp->numDevices;
  }
  if (ceiling < 0) 
  { /* use the registered information */
    ceiling = n;
  }
  memset(lclname,0,DEVICE_NAME_SIZE+4);
  strncpy(lclname,devname,DEVICE_NAME_SIZE);
  if (devname[0] == '#') 
  { /* ceiling == 0 -> just parse, convert and return */
    i = atoi(&lclname[1]);
    if (ceiling == 0) return i;
    if (ceiling > 0 && i < ceiling) return i;
    return -illegal_device;
  }
  if (el->EqmDevLst == NULL) return -not_registered;
  if (el->EqmNumDevices < DEVICES_HASH_PAYOFF)
  {
    for (i=0; i<el->EqmNumDevices; i++)
      if (!strnicmp(lclname,el->EqmDevLst[i].name,DEVICE_NAME_SIZE)) return i;
    return -device_not_connected;
  }
  if (el->EqmDevXRef == NULL) return -device_not_connected;
  idx = (int)(ElfHash((unsigned char *)lclname) % DEVICES_HASH_SIZE);
  for (xref=el->EqmDevXRef[idx]; xref; xref=xref->next)
    if (!strnicmp(lclname,el->EqmDevLst[xref->index].name,DEVICE_NAME_SIZE)) return xref->index;
  return -device_not_connected;
}
TINE_EXPORT int GetDeviceNumber(char *eqm,char *devname)
{
  return GetValidDeviceNumber(eqm,devname,NULL,0);
}
TINE_EXPORT int GetNumContracts(char *eqm)
{
  int i,n=0;
  if (eqm == NULL || strlen(eqm) == 0) return ncontracts - nExpiredContracts;
  for (i=0; i<ncontracts; i++)
   if (!strncmp(ContractList[i]->contract.EqmName,eqm,EQM_NAME_SHORTSIZE) && 
       ContractList[i]->expired == CONTRACT_ACTIVE) n++;
  return n;
}
TINE_EXPORT int GetNumCallers(char *eqm)
{
  return GetNumConsumers(eqm);
}
TINE_EXPORT int GetNumConsumers(char *eqm)
{
  int i,n=0;
  if (eqm == NULL || strlen(eqm) == 0) return nconsumers;
  for (i=0; i<ncontracts; i++)
   if (!strncmp(ContractList[i]->contract.EqmName,eqm,EQM_NAME_SHORTSIZE) &&
       ContractList[i]->nclients > n) n = ContractList[i]->nclients;
  return n;
}
TINE_EXPORT int GetNumCalls(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return -non_existent_elem;
  return el->EqmNumCalls;
}
TINE_EXPORT int GetPropertyId(char *eqm,char *prpName)
{
  int pid = GetPrpId(eqm,prpName);
  return pid < 0 ? -1 : (int)PROPERTY_NUM_INDEX(pid);
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int JoinEquipmentGroup(char *eqmName,char *groupname,int groupindex)
{
  ExportListStruct *el = getExportListItem(eqmName);

  if (el == NULL) return non_existent_elem;
  if (groupname == NULL || strlen(groupname) == 0) return argument_list_error;
  if (groupindex < 0) return out_of_range;
  strncpy(el->EqmGroup,groupname,EXPORT_NAME_SIZE);
  el->groupIndex = groupindex;
  return 0;
}
TINE_EXPORT int JoinEquipmentGroupEx(char *eqmName,char *groupname,int groupindex,char *devPrefix,char *devPostfix)
{
  ExportListStruct *el = getExportListItem(eqmName);

  if (el == NULL) return non_existent_elem;
  if (groupname == NULL || strlen(groupname) == 0) return argument_list_error;
  if (groupindex < 0) return out_of_range;
  strncpy(el->EqmGroup,groupname,EXPORT_NAME_SIZE);
  el->groupIndex = groupindex;
  if (devPrefix != NULL) strncpy(el->EqmGroupDevPrefix,devPrefix,XFIX_SIZE);
  if (devPostfix != NULL) strncpy(el->EqmGroupDevPostfix,devPostfix,XFIX_SIZE);
  return 0;
}
#endif
int makeEqmDevLst(ExportDeviceType **devlst,int len)
{ /* used internally : devlst never NULL */
  ExportDeviceType *dl=NULL;
  NAME64 *n64=NULL;
  int i, cc = 0;
  if ((dl=(ExportDeviceType *)SystemCalloc(len,sizeof(ExportDeviceType))) == NULL) ccerr(out_of_server_memory);
  if ((n64=(NAME64 *)SystemCalloc(len,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
  for (i = 0; i < len; i++)
  {
    dl[i].name = n64[i].name;
    dl[i].number = i;
  }
  *devlst = dl;
err:
  if (cc != 0)
  {
    SystemFree(dl);
    SystemFree(n64);
  }
  return cc;
}
ExportDeviceType *getDeviceRef(char *eqm,char *devname,int *sts)
{
  int i,cc = 0;
  ExportListStruct *el=getExportListItem(eqm);
  if (el == NULL) ccerr(non_existent_elem);
  if (devname == NULL) ccerr(argument_list_error);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqmNumDevices) ccerr(device_not_connected);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
err:
  *sts = cc;
  return cc ? NULL : &el->EqmDevLst[i];
}
TINE_EXPORT int RemoveDevice(char *eqm, char *devname)
{
  ExportListStruct *el = getExportListItem(eqm);
  HashXRefStruct *xref, *prvref, *nxtref = NULL;
  int cc = 0, idx, devnr;
  if (WaitForSystemMutex(hSystemKernelMutex, gSystemTick * 3) != 0) return semaphore_busy;
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  devnr = GetDeviceNumber(eqm, devname);
  if (devnr < 0 || devnr >= el->EqmNumDevices) ccerr(out_of_range);
  if (el->EqmDevXRef != NULL)
  { /* enough registered devices to maintain a hash table */
    idx = (int)(ElfHash((unsigned char *)devname) % DEVICES_HASH_SIZE);
    prvref = el->EqmDevXRef[idx];
    for (xref = el->EqmDevXRef[idx]; xref != NULL; xref = nxtref)
    { /* run thru the hash list */
      nxtref = xref->next;
      if (xref->index == devnr)
      { /* found it, now remove it ! */
        if (prvref == xref) prvref = el->EqmDevXRef[idx] = nxtref;
        else prvref->next = nxtref;
        SystemFree(xref);
        break;
      }
      prvref = xref;
    }
  }
  if (WaitForSystemMutex(hAlmTblMutex, -1) != 0) return mutex_error;
  freeEqmDevType(el->EqmDevLst,devnr,el->EqmDevLstCapacity);
  ReleaseSystemMutex(hAlmTblMutex);
err:
  ReleaseSystemMutex(hSystemKernelMutex);
  return cc;
}
TINE_EXPORT int RegisterDeviceName(char *eqm,char *devname,int devnr)
{
  ExportListStruct *el=getExportListItem(eqm);
  HashXRefStruct *xref, *prvref, *nxtref=NULL;
  char lclname[DEVICE_NAME_SIZE+4];
  int cc=0,idx,snr=-1;

  if (el == NULL) ccerr(non_existent_elem);
  if (devname == NULL) ccerr(argument_list_error);
  if (!assertNameIsValid(devname,STRICTLY_ILLEGALCHARS,DEVICE_NAME_SIZE)) ccerr(invalid_name);

  if (devnr < 0 || devnr >= el->EqmDevLstCapacity) ccerr(out_of_range);
  if (el->EqmDevLst == NULL) makeEqmDevLst(&el->EqmDevLst,el->EqmDevLstCapacity);
  memset(lclname,0,DEVICE_NAME_SIZE+4);
  if (strlen(devname) > DEVICE_NAME_SIZE) feclogEx(TINE_LOGLEVEL_WARN, "device name %s truncated to %d characters",devname,DEVICE_NAME_SIZE);
  strncpy(lclname,devname,DEVICE_NAME_SIZE);
  strtrm(lclname);
  snr = GetDeviceNumber((char *)eqm,(char *)lclname);
  if (lclname[0] != '#' && snr >= 0)
  { /* name already registered */
    if (snr == devnr)
    { /* same device number -> exit early */
      el->EqmDevLst[snr].isReserved = FALSE;
      return 0; 
    }
    sprintf(el->EqmDevLst[snr].name,"#%d",snr); /* no duplicate names ! */
  }
  strncpy(el->EqmDevLst[devnr].name,lclname,DEVICE_NAME_SIZE);
  if (devnr == 0 && strlen(el->EqmDevLst[0].name) > 0) el->hasRegisteredDevicesNames = TRUE;
  el->EqmDevLst[devnr].isReserved = FALSE;
  el->EqmDevLst[devnr].region = el->region; /* assign the default region initially */
  GetDeviceAclInfo(eqm,devnr);
  if (el->EqmDevLstCapacity < DEVICES_HASH_PAYOFF) return 0;
  if (el->EqmDevXRef == NULL) return 0;
  idx = (int)(ElfHash((unsigned char *)lclname) % DEVICES_HASH_SIZE);
  prvref = el->EqmDevXRef[idx];
  for (xref=el->EqmDevXRef[idx]; xref && xref->index != devnr; xref=nxtref)
  { /* run thru the hash list */
    nxtref = xref->next;
    if (xref->index == snr)
    { /* remove this stored device number ! */
      if (prvref == xref) prvref = el->EqmDevXRef[idx] = nxtref;
      else prvref->next = nxtref;
      SystemFree(xref);
      xref = prvref;
    }
    prvref = xref;
  }
  if (xref == NULL) 
  { /* it's a new one*/
    if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
    xref->index = devnr;
    xref->next = el->EqmDevXRef[idx];
    el->EqmDevXRef[idx] = xref;
  }
err:
  return cc;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int RedirectDeviceName(char *eqm,char *devname,char *rdr)
{
  int cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  PrpRedirBlk prd;
  char msg[256];
  if (dv == NULL) ccerr(cc);
  if (rdr == NULL) ccerr(argument_list_error);
  ParseRedirector(rdr,&prd); /* only the device context/server name is relevant */
  sprintf(msg, "Redirection for device %.64s", devname);
  if (!assertRedirectionValid(msg,eqm,prd.prpRedirection)) ccerr(not_accepted);
  SystemFree(dv->rdr);
  if ((dv->rdr=(char *)(SystemCalloc(1,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE))) == NULL)
    ccerr(out_of_server_memory);
  strncpy(dv->rdr,&prd.prpRedirection[PRP_REDIR_CTX_OFFSET],CONTEXT_NAME_SIZE); 
  strncpy(&dv->rdr[CONTEXT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_SRV_OFFSET],EXPORT_NAME_SIZE); 
  strncpy(&dv->rdr[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_DEV_OFFSET],DEVICE_NAME_SIZE); 
err:
  return cc;
}
TINE_EXPORT int SetDeviceDescription(char *eqm,char *devname,char *description)
{
  int slen, cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  if (description == NULL) ccerr(argument_list_error);
  if (dv != NULL)
  {
    slen = (int)strlen(description);
    if (dv->desc == NULL)
    {
      if (slen > 64) slen = 64;
      SystemFree(dv->desc);
      if ((dv->desc=(char *)SystemCalloc(1,slen)) == NULL) ccerr(out_of_server_memory);
    }
    strncpy(dv->desc,description,slen); 
  }
err:
  return cc;
}
TINE_EXPORT int RegisterDeviceDescription(char *eqm,char *devname,char *description)
{
  return SetDeviceDescription(eqm,devname,description);
}
TINE_EXPORT int SetDeviceLocation(char *eqm,char *devname,char *location)
{
  int slen, cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  if (location == NULL) ccerr(argument_list_error);
  if (dv != NULL)
  {
    slen = (int)strlen(location);
    if (dv->loc == NULL)
    {
      if (slen < 64) slen = 64;
      SystemFree(dv->loc);
      if ((dv->loc=(char *)SystemCalloc(1,slen)) == NULL) ccerr(out_of_server_memory);
    }
    strncpy(dv->loc,location,slen); 
  }
err:
  return cc;
}
TINE_EXPORT int SetDeviceMask(char *eqm,char *devname,int mask)
{
  int cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  if (dv != NULL) dv->mask = mask;
  return cc;
}
TINE_EXPORT int SetDeviceOfflineStatus(char *eqm,char *devname,int offline)
{
  int cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  if (dv != NULL)
  {
    if (offline) dv->mask |= DEVICE_OFFLINE; else dv->mask &= ~(DEVICE_OFFLINE);
  }
  return cc;
}
TINE_EXPORT int SetDeviceZPosition(char *eqm,char *devname,float zpos)
{
  int cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  if (dv != NULL) dv->zpos = zpos;
  return cc;
}
TINE_EXPORT int SetDeviceRegion(char *eqm,char *devname,char *region)
{
  int cc = 0;
  ExportDeviceType *dv = getDeviceRef(eqm,devname,&cc);
  if (dv != NULL) dv->region = GetValidDeviceRegion(eqm,region);
  return cc;
}
TINE_EXPORT int RegisterMultiChannelGroupDevice(char *eqm,char *grpname,char *devname,int grpindex,int grpsize)
{
  int i,cc = 0;
  ExportListStruct *el=getExportListItem(eqm);
  if (el == NULL) ccerr(non_existent_elem);
  if (devname == NULL || grpname == NULL) ccerr(argument_list_error);
  if (grpindex < 0 || grpindex >= el->EqmNumDevices) ccerr(invalid_index);
  if (grpsize < 2) ccerr(illegal_data_size);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqmNumDevices) ccerr(device_not_connected);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if (el->EqmDevLst[i].mcaGrp == NULL)
  {
    if ((el->EqmDevLst[i].mcaGrp=(DeviceMcaGroup *)SystemCalloc(1,sizeof(DeviceMcaGroup))) == NULL) 
      ccerr(out_of_server_memory);
  }
  strncpy(el->EqmDevLst[i].mcaGrp->grpName,grpname,DEVICE_NAME_SIZE);
  el->EqmDevLst[i].mcaGrp->grpIndex = grpindex;
  el->EqmDevLst[i].mcaGrp->grpSize = grpsize;
err:
  return cc;
}
TINE_EXPORT int GetDeviceGroups(char *eqm, NAME64 *groups, int *len)
{
  int i, k, max = 0, n = 0, cc = 0;
  ExportListStruct *el = getExportListItem(eqm);
  if (groups == NULL || len == NULL) ccerr(argument_list_error);
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if ((max = *len) <= 0) ccerr(dimension_error);
  memset(groups, 0, max * sizeof(NAME64));
  for (i = 0; i < el->EqmNumDevices && n < max; i++)
  {
    if (el->EqmDevLst[i].mcaGrp == NULL) continue;
    for (k = 0; k < n; k++) /* what have we got so far ? */
      if (!strnicmp(el->EqmDevLst[i].mcaGrp->grpName, groups[k].name, DEVICE_NAME_SIZE)) break;
    if (k < n) continue; /* already in the list */
    strncpy(groups[n++].name, el->EqmDevLst[i].mcaGrp->grpName, DEVICE_NAME_SIZE);
  }
  *len = n;
err:
  return cc;
}
TINE_EXPORT int GetGroupDeviceMembers(char *eqm, char *grpname, NAME64 *members, int *len)
{
  int i, idx, max = -1, n = 0, cc = 0;
  ExportListStruct *el = getExportListItem(eqm);
  if (grpname == NULL || members == NULL || len == NULL) ccerr(argument_list_error);
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqmDevLst == NULL) ccerr(not_initialized);
  if ((n = *len) <= 0) ccerr(dimension_error);
  memset(members, 0, n * sizeof(NAME64));
  for (i = 0; i < el->EqmNumDevices; i++)
  {
    if (el->EqmDevLst[i].mcaGrp == NULL) continue;
    if (strnicmp(el->EqmDevLst[i].mcaGrp->grpName, grpname, DEVICE_NAME_SIZE) != 0) continue;
    idx = el->EqmDevLst[i].mcaGrp->grpIndex;
    if (idx >= *len) continue;
    if (idx > max) max = idx;
    strncpy(members[idx].name, el->EqmDevLst[i].name, DEVICE_NAME_SIZE);
  }
  *len = max + 1;
err:
  return cc;
}
TINE_EXPORT int SetFailoverMaster(char *eqm,char *masterName)
{
  int cc = 0;
  ExportListStruct *el=getExportListItem(eqm);
  if (el == NULL) ccerr(non_existent_elem);
  strncpy(el->EqmMaster,masterName,EXPORT_NAME_SIZE);
  el->failoverType = FAILOVER_MASTER;
  /* TODO: if registered late -> register as master */
err:
  return cc;
}
TINE_EXPORT int SetFailoverSlave(char *eqm,char *masterName,char *slaveMaster)
{
  int cc = 0;
  ExportListStruct *el=getExportListItem(eqm);
  if (el == NULL) ccerr(non_existent_elem);
  strncpy(el->EqmMaster,masterName,EXPORT_NAME_SIZE);
  strncpy(el->EqmSlaveMaster,slaveMaster,EXPORT_NAME_SIZE);
  el->failoverType = FAILOVER_SLAVE;
  el->registered |= REGISTERED_MASTER;
  /* TODO: if registered late -> start slave activity */
err:
  return cc;
}
int gNumFecAdmins = 0;
NAME16 *gFecAdmins = NULL;
int isAdmin(const char *nam)
{
  int i;
  if (gNumFecAdmins == 0) return TRUE;
  for (i=0; i<gNumFecAdmins && strnicmp((char *)nam,gFecAdmins[i].name,USERNAME_SIZE); i++);
  return gNumFecAdmins - i;
}
int getAdmins(void)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256],fn[256];

  sprintf(fn,"%.128s%.32s",fecDbPath,ADMINS_FILE);
  if (gFecAdmins != NULL)
  {
    free(gFecAdmins);
    gNumFecAdmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((gFecAdmins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  gNumFecAdmins = n;
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<gNumFecAdmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0 && 
          (usr_col=findcol(hdr,"USERNAME")) < 0 ) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(gFecAdmins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
  feclog("found %d registered administrators: allow remote management",gNumFecAdmins);
  SetAllowRemoteManagement(TRUE);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
TINE_EXPORT int AssertIsAdministrator(const char *usr)
{
  if (usr == NULL) return FALSE;
  if (isAdmin(usr)) return TRUE;
  feclogEx(TINE_LOGLEVEL_INFO, "user %.16s is not an authorized administrator",usr);
  return FALSE;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT int RegisterEquipmentModule(
    char *expName,
    char *eqmName,
    int numdevices,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short),
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
)
{
  return _registerEquipmentModule(expName,eqmName,numdevices,(void *)fcn,(void *)ini,(void *)tsk,rate,(void *)exi,NULL);
}
TINE_EXPORT int RegisterEquipmentModuleEx(
    char *expName,
    char *eqmName,
    int numdevices,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short,void *),
    void (*ini)(void *),
    void (*tsk)(void *),
    int rate,
    void (*exi)(void *),
    void *reference
)
{
  return _registerEquipmentModule(expName,eqmName,numdevices,(void *)fcn,(void *)ini,(void *)tsk,rate,(void *)exi,reference);
}
TINE_EXPORT int RegisterFecInformation(char *name,char *sub,char *cntxt,char *dsc,char *loc,char *hdw,char *resp,UINT16 poff)
{
  if (name == NULL) return argument_list_error;
  if (FecNameRegistered)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "FEC Name %.16s already registered for this server",gFecName);
    if (strncmp(name,gFecName,FEC_NAME_SIZE) == 0) return 0; /* don't panic, but stop here */
    return not_allowed;
  }
  if (!assertNameIsValid(name,RUDECHARS,FEC_NAME_SIZE)) return invalid_name;
  if (strncmp(name,ENS_FEC,FEC_NAME_SIZE) == 0) isEquipmentNameServer(TRUE);
  if (strncmp(name,GENS_FEC,FEC_NAME_SIZE) == 0) isGroupNameServer(TRUE);
  strncpy(gFecName,name,FEC_NAME_SIZE);
  memset(&gFecInfo, 0, sizeof(gFecInfo));
  strncpy(gFecInfo.ver,GetSystemVersionString(),FEC_VERSION_SIZE);
  strncpy(gFecInfo.os,OS_TAG,FEC_OS_SIZE);
  if (dsc != NULL) strncpy(gFecInfo.desc,dsc,FEC_DESC_SIZE);
  if (loc != NULL) strncpy(gFecInfo.loc,loc,FEC_LOCATION_SIZE);
  if (hdw != NULL) strncpy(gFecInfo.hdw,hdw,FEC_HDW_SIZE);
  if (resp != NULL) strncpy(gFecInfo.resp,resp,FEC_RESP_SIZE);
  SetDoocsUserName(whoami(NULL));
  strncpy(gUserName,gFecName,USERNAME_SIZE);
  SetUserType("FEC");
  gPortOffset = validatePortOffset((int)poff);
  if (cntxt != NULL) strncpy(gDeviceContext,cntxt,CONTEXT_NAME_SIZE);
  if (sub != NULL) strncpy(gDeviceSubSystem,sub,SUBSYSTEM_NAME_SIZE);
  FecNameRegistered = TRUE;
  feclog("FEC Name %.16s registered from API",name);
  return 0;
}
TINE_EXPORT int RegisterFecNameEx(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff,char *context)
{
  char *c,sub[SUBSYSTEM_NAME_SIZE+4] = "";
  UNUSED(os); /* avoid warning */
  if ((c=strchr(desc,'[')) != NULL) 
  {
    strncpy(sub, ++c, SUBSYSTEM_NAME_SIZE); sub[SUBSYSTEM_NAME_SIZE] = 0;
    if ((c=strchr(sub,']')) != NULL) *c = 0;
  }
  return RegisterFecInformation(name,sub,context,desc,loc,hdw,resp,poff);
}
TINE_EXPORT int RegisterFecName(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff)
{
  return RegisterFecNameEx(name,desc,os,loc,hdw,resp,poff,NULL);
}
TINE_EXPORT int RegisterPropertyAlias(char *eqm,char *property,char *alias)
{
  int cc;
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,property,NULL);
  if (prp == NULL) return non_existent_property;
  if (alias == NULL || strlen(alias) == 0) return argument_list_error;
  if ((cc=addAliasToTable(property,alias)) != 0) return cc;
  strncpy(prp->prpAlias,alias,PROPERTY_NAME_SIZE);
  for (prp=prp->next; prp != NULL; prp=prp->next)
  { /* set same for any overloads */
    if (strnicmp(prp->prpName,property,PROPERTY_NAME_SIZE)) continue;
    strncpy(prp->prpAlias,alias,PROPERTY_NAME_SIZE);
  }
  return 0;
}
PrpDbaItem *prpDbaLst = NULL;
TINE_EXPORT int RegisterPropertyAccessDeadband(char *eqm,char *property,int access,int deadbandInMilliSeconds)
{
  ExportPropertyListStruct *prp = GetPropertyListStruct(eqm,property,NULL);
  PrpDbaItem *dba = NULL;
  if (prp == NULL) return non_existent_property;
  if (deadbandInMilliSeconds < 0) deadbandInMilliSeconds = 0; /* turn it off */
  prp->minAccessDeadband = (UINT16)deadbandInMilliSeconds;
  for (prp=prp->next; prp != NULL; prp=prp->next)
  { /* set same for any overloads */
    if (strnicmp(prp->prpName,property,PROPERTY_NAME_SIZE)) continue;
    prp->minAccessDeadband = (UINT16)deadbandInMilliSeconds;
  }
  if ((dba=getPrpDbaItem(eqm,property)) != NULL)
  { /* already in the list ! */
    dba->deadband = deadbandInMilliSeconds;
    dba->access = access;
    return 0;
  }
  if (deadbandInMilliSeconds > 0) return addPrpDbaItem(eqm,property,access,deadbandInMilliSeconds);
  return 0;
}
int addPrpDbaItem(char *eqm,char *prp,int access,int deadband)
{
  PrpDbaItem *dba = NULL;
  if ((dba=getPrpDbaItem(eqm,prp)) != NULL)
  { /* already in the list ! */
    dba->access = access;
    dba->deadband = deadband;
    return 0;
  }
  if ((dba=(PrpDbaItem *)SystemCalloc(1,sizeof(PrpDbaItem))) == NULL) return out_of_server_memory;
  if (access == 0) access = CA_WRITE;
  strncpy(dba->eqm,eqm,EQM_NAME_SIZE);
  strncpy(dba->prp,prp,PROPERTY_NAME_SIZE);
  dba->access = access;
  dba->deadband = deadband;
  dba->nxt = prpDbaLst;
  prpDbaLst = dba;
  return 0;
}
PrpDbaItem *getPrpDbaItem(char *eqm,char *prp)
{
  PrpDbaItem *dba = NULL;
  for (dba=prpDbaLst; dba != NULL; dba=dba->nxt)
  {
    if (strncmp(eqm,dba->eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(prp,dba->prp,PROPERTY_NAME_SIZE)) continue;
    return dba;
  }
  return NULL;
}
int dumpPrpDbaList(void)
{
  PrpDbaItem *dba = NULL;
  char addr[64], tsstr[64], *usr;
  double ts;
  if (prpDbaLst == NULL)
  {
    ttyoutput("there are no registered deadband access properties!");
    return 0;
  }
  ttyoutput("current deadband access:");
  ttyoutput("    (-EQM--)         Property : deadband, applies to, last access");
  for (dba=prpDbaLst; dba != NULL; dba=dba->nxt)
  {
    inetStrAdr(&dba->addr,addr,FALSE);
    usr = dba->usr;
    ts = (double)dba->lastAccess.tv_sec + ((double)dba->lastAccess.tv_usec)/1000000.0;
    ttyoutput("    (%.6s) %16.64s : %5d ms, %11s, %.26s %.16s %.40s",
      dba->eqm,dba->prp,dba->deadband,itom((short)dba->access),
      GetDataTimeStringEx(ts,FALSE,tsstr),usr,addr);
  }
  return 0;
}
int prpRedirectionExists(PrpRedirBlk *prd,char *rdr)
{
  PrpRedirBlk *p;
  for (p=prd; p != NULL; p=p->next)
  {
    if (memcmp(p->prpRedirection,rdr,64)) continue;
    if (strlen(&p->prpRedirection[PRP_REDIR_DEV_OFFSET]) == 0) return TRUE; /* all devices get redirected ! */
    if (strnicmp(&p->prpRedirection[PRP_REDIR_DEV_OFFSET],&rdr[PRP_REDIR_DEV_OFFSET],DEVICE_NAME_SIZE)) 
    { /* different specific device */
      continue;
    }
    if ((strlen(&p->prpRedirection[PRP_REDIR_PRP_OFFSET]) > 0) &&
         strnicmp(&p->prpRedirection[PRP_REDIR_PRP_OFFSET],&rdr[PRP_REDIR_PRP_OFFSET],PROPERTY_NAME_SIZE)) 
    { /* different specific property */
      continue;
    }
    return TRUE;
  }
  return FALSE;
}
int applyRedirectionInfoToProperty(char *eqm,ExportPropertyListStruct *pl,char *rdr)
{
  PrpRedirBlk *prd;
  char strbuf[256], msg[256];
  int cc = 0;
  if (pl == NULL || pl->isReserved) return not_registered;
  if (rdr != NULL && strlen(rdr) != 0)
  { /* redirection string given */
    strncpy(strbuf,rdr,255);
    if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
    ParseRedirector(strbuf,prd);
    sprintf(msg, "Redirection for property %.64s",pl->prpName);
    if (!prpRedirectionExists(pl->prd,prd->prpRedirection) &&
         assertRedirectionValid(msg,eqm, prd->prpRedirection))
    { /* redirection doesn't already exist and is valid */
      prd->next = pl->prd;
      pl->prd = prd;
    }
    else
    {
      SystemFree(prd);
    }
  }
err:
  return cc;
}
/* Appends property list for eqm ...*/
TINE_EXPORT int RegisterPropertyInformationEx(char *eqm,char *prp,DTYPE *dout,DTYPE *din,
                  short acc,short atype,UINT16 rowlen,char *dsc,PrpEgu *egu,PrpEgu *xegu,int pid,char *rdr) 
{
  int i,idx,cc=0,isNew,fsiz,ssiz=1,hsiz=0,plsize,restorePrpValues=FALSE;
  SINT32 prpId = pid, prpId0 = -1;
  UINT32 dSizeInBytes = 0, qSizeInBytes = 0;
  ExportPropertyListStruct *pl = NULL;
  ExportListStruct *el = getExportListItem(eqm);
  char strbuf[256];
# ifndef SMALL_TINE_LIB
  char *fn;
  int prpUsersFile = -1, prpNetsFile = -1, rc;
# endif
  if (el == NULL) ccerr(non_existent_elem);
  if (!assertNameIsValid(prp,STRICTLY_ILLEGALCHARS,PROPERTY_NAME_SIZE)) ccerr(invalid_name);
  if ((i=(int)strlen(prp)) > 4 && !strcmp(&prp[i-4],".NAM") && strstr(prp,".NAM.NAM") == NULL) 
  {
    el->hasPropertyDeviceNames = TRUE;
    if (queryDeviceNamesFromProperties) acquireDeviceNamesFromProperties = TRUE;
  }
  if (strchr(prp,'*')) idx = PROPERTY_HASH_SIZE;
  else idx = (int)(ElfHash((unsigned char *)prp) % PROPERTY_HASH_SIZE);

  for (pl=el->EqmPrpLst[idx]; pl; pl=pl->next) if (!strnicmp(pl->prpName,prp,PROPERTY_NAME_SIZE)) break;

  for (isNew=TRUE; pl != NULL; pl=pl->next) 
  { /* follow this hash index and check for overloads/multiple registrations */
    if (strnicmp(pl->prpName,prp,PROPERTY_NAME_SIZE)) continue;
    pl->isReserved = FALSE;
    prpId0 = pl->prpId;
    if (dout != NULL)
    {
      if (dout->dArrayLength != pl->prpSize || dout->dFormat != pl->prpFormat) continue;
      if (dout->dFormat == CF_STRUCT && strncmp(dout->dTag,pl->prpFormatTag.name,TAG_NAME_SIZE)) continue;
    }
    else
    {
      if (pl->prpSize > 0) continue;
    }
    if (din != NULL)
    {
      if (din->dArrayLength != pl->prpSizeIn || din->dFormat != pl->prpFormatIn) continue;
      if (din->dFormat == CF_STRUCT && strncmp(din->dTag,pl->prpFormatTagIn.name,TAG_NAME_SIZE)) continue;
    }
    else
    {
      if (pl->prpSizeIn > 0) continue;
    }
    if (acc != pl->prpAccessMode) continue;
    isNew = FALSE; /* found a match ! */
    break;
  }  
  if (!isNew) 
  { /* a 're-register': update a few things */
    pl->prpArrayType = atype;
    if (rowlen > 0)
    {
      pl->rowSize = rowlen;
      pl->numRows = (UINT16)(pl->prpSize/rowlen);
    }
    memset(strbuf,0,256);
    if (dsc != NULL) strncpy(strbuf,dsc,255);
    ParsePropertyDescription(strbuf,pl);
    applyRedirectionInfoToProperty(eqm,pl,rdr);
    ccerr(0);
  }
# ifndef SMALL_TINE_LIB
  if (isUsingXMLFecDatabase) 
  {
    prpUsersFile = prpNetsFile = 1;
  }
  else
  {
    setAclFileInfo(eqm,&prpUsersFile,&prpNetsFile);
  }
# endif
  if ((pl=(ExportPropertyListStruct *)SystemCalloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
  if (strlen(prp) > PROPERTY_NAME_SIZE) feclogEx(TINE_LOGLEVEL_WARN, "property %s truncated to %d characters",prp,PROPERTY_NAME_SIZE);
  strncpy(pl->prpName,prp,PROPERTY_NAME_SIZE);
  if ((i=getAliasTableIndexFromName(prp)) >= 0)
    strncpy(pl->prpAlias,gAliasTable[i].alias,PROPERTY_NAME_SIZE);
  pl->next = el->EqmPrpLst[idx];
  el->EqmPrpLst[idx] = pl;
  if (prpId < 0) prpId = el->EqmNumProperties;
  pl->prpId = prpId0 >= 0 ? prpId0 : MAKE_PROPERTY_ID(prpId,idx);
  el->EqmNumProperties++;

  if (dout != NULL)
  {
    int isDoubleArray = (atype & AT_DOUBLE);
    pl->prpSize = dout->dArrayLength;
    pl->prpFormat = dout->dFormat; 
    pl->prpArrayType = atype;
    if (!isDoubleArray)
    {
      if (atype & AT_CHANNEL)
      {
        pl->numDevices = pl->prpSize;
        if (acc & CA_SAVERESTORE && pl->numDevices < el->EqmNumDevices)
          feclog("%.6s: Save-and-Restore for property %.64s will honor %d channel(s) rather than all %d devices due to registration",
            eqm, prp, pl->numDevices, el->EqmNumDevices);
      }
      if (pl->prpSize > 1)
      { /* not a double array, then it's a single array */
        pl->prpArrayType |= AT_SINGLE;
      }
      else if ((pl->prpArrayType & AT_SCALAR) == 0)
      {
        if (pl->prpArrayType != 0)
        {
          feclog("%.6s: Changing array type for property %.64s from %s to SCALAR due to output size = 1",
            eqm, prp, GetArrayTypeAsString(pl->prpArrayType));
        }
        pl->prpArrayType = AT_SCALAR;
      }
    }
    if (strlen(dout->dTag)) strncpy(pl->prpFormatTag.name,dout->dTag,TAG_NAME_SIZE);
    if (isDoubleArray && rowlen > 0)
    {
      pl->rowSize = rowlen;
      pl->numRows = (UINT16)(pl->prpSize / rowlen);
    }
    else if (!isDoubleArray && pl->prpSize >= 65536)
    { /* ultra large: IMAGE ? */
      pl->rowSize = 0;
      pl->numRows = 0;
    }
    else 
    {
      pl->rowSize = (UINT16)pl->prpSize;
      pl->numRows = 1;
    }
    if (pl->prpFormat == CF_STRUCT)
    {
      if ((ssiz=GetStructSize(pl->prpFormatTag.name)) == 0) ssiz = 1;
    }
    fsiz = fmtsizeof(pl->prpFormat);
    hsiz = getFormatHeaderSize(pl->prpFormat);
    dSizeInBytes = pl->prpSize * (fsiz*ssiz) + hsiz;
  }
  else
  {
    pl->prpFormat = CF_NULL;
  }
  if (din != NULL)
  {
    pl->prpSizeIn = din->dArrayLength;
    pl->prpFormatIn = din->dFormat;
    if (strlen(din->dTag)) strncpy(pl->prpFormatTagIn.name,din->dTag,TAG_NAME_SIZE);
  }
  else
  {
    pl->prpFormatIn = CF_NULL;
  }
  if (acc & CA_XREAD)
  {
    pl->hasExclusiveRead = (UINT16)((acc & CA_READ) ? PRP_XREAD_LOCK : PRP_XREAD_ALL);
    acc &= ~CA_XREAD; /* strip it back off */
  }
  if (acc & CA_SAVERESTORE)
  {
#   ifndef SMALL_TINE_LIB
    restorePrpValues = isAllowedSaveRestoreProperty(el,pl);
#   endif
    acc &= ~CA_SAVERESTORE; /* strip it back off */
  }
  if (acc & CA_HIST)
  { /* quick addition to local history subsystem (use defaults) */
    AppendHistoryInformation(eqm,prp,"#0",pl->prpSize,pl->prpFormat,-1,&defaultHSpec);
    acc &= ~CA_HIST; /* strip it back off */
  }
  if (acc & CA_FORCEOUTPUT)
  { /* use conract coercion to force clients to use the registered output data length */
    pl->enforceDataReq = TRUE;
    acc &= ~CA_FORCEOUTPUT; /* strip it back off */
  }
  pl->prpAccessMode = acc;
  memset(strbuf,0,256);
  if (dsc != NULL) strncpy(strbuf,dsc,255);
  ParsePropertyDescription(strbuf,pl);
  applyRedirectionInfoToProperty(eqm,pl,rdr);
  if (egu != NULL)
  { /* info here supercedes that parsed from the description */
    memcpy(&pl->egu,egu,sizeof(PrpEgu));
  }
  if (xegu != NULL)
  { /* info here supercedes that parsed from the description */
    memcpy(&pl->xegu,xegu,sizeof(PrpEgu));
  }
  plsize = pl->prpSize;
  if ((pl->devNames=GetEqpDeviceNames(eqm,pl->prpName,NULL,&plsize)) != NULL)
  {
    pl->numDevices = plsize;
    el->hasPropertyDeviceNames = TRUE;
  }
# ifndef SMALL_TINE_LIB
  if (prpUsersFile >= 0)
  {
    int acc = 0;
    fn = makeListFileName(eqm,prpUsersFile,pl->prpName,USERS_FILE);
    if (getRegUsrsFromFile(eqm,NULL,pl->prpName,&pl->aclLst,fn,"WRITE",&acc) != 0)
    {
      fn = makeListFileName(eqm,prpUsersFile,"property-default",USERS_FILE);
      getRegUsrsFromFile(eqm,NULL,pl->prpName,&pl->aclLst,fn,"WRITE",&acc);
    }
    el->chkAcl |= (acc<<4);
  }
  if (prpNetsFile >= 0)
  {
    fn = makeListFileName(eqm,prpNetsFile,pl->prpName,IPNETS_FILE);
    if (getRegNetsFromFile(eqm,NULL,pl->prpName,fn,NETS_ACCESS,&pl->aclLst.nNets,&pl->aclLst.nets) != 0)
    {
      fn = makeListFileName(eqm,prpNetsFile,"property-default",IPNETS_FILE);
      getRegNetsFromFile(eqm,NULL,pl->prpName,fn,NETS_ACCESS,&pl->aclLst.nNets,&pl->aclLst.nets);
    }
  }
  if (restorePrpValues)
  { /* so far, so good ... */
    rc = _RestorePropertyValues(eqm,pl->prpName,pl->srBuf,pl->prpFormat,pl->srBufSiz,TRUE);
    addSrPrp(el,pl,rc);
  }
# endif
  if (!gAutoAdjustWorkAreaSize) ccerr(0);
  if (dSizeInBytes > MaxSystemTransportSize)
  { /* need to make adjustments */
#   if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
    adjustWorkAreaSize(dSizeInBytes+0x10000);
#   else
    feclogEx(TINE_LOGLEVEL_WARN, "Property %.64s size (%d bytes) larger than registered work area (%d bytes)",
      prp,dSizeInBytes,MaxSystemTransportSize);
#   endif
  }
  qSizeInBytes = el->EqmNumProperties * PRPQUERYSTRUCT_SIZE;
  if (qSizeInBytes > MaxSystemTransportSize)
  { /* need to make adjustments */
#   if !defined(VXWORKS) && !defined(SMALL_TINE_LIB)
    feclog("adjust work area size to accommodate property information queries");
    adjustWorkAreaSize(qSizeInBytes+0x10000);
#   else
    feclogEx(TINE_LOGLEVEL_WARN, "simultaneous query for all information for %d properties larger than registered work area (%d bytes)",
      el->EqmNumProperties,MaxSystemTransportSize);
#   endif
  }
err:
  return cc ? -cc : (int)PROPERTY_NUM_INDEX(pl->prpId);
}
TINE_EXPORT int RegisterPropertyInformation(char *eqm,char *prp,DTYPE *dout,DTYPE *din,
                        short acc,short atype,UINT16 rowlen,char *dsc,int pid,char *rdr) 
{
  return RegisterPropertyInformationEx(eqm,prp,dout,din,acc,atype,rowlen,dsc,NULL,NULL,pid,rdr);
}
/* removes property from the eqm (but does not free memory !) */
TINE_EXPORT int RemoveProperty(char *eqm, char *property)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;
  if (el == NULL) return non_existent_elem;
  if (HAS_PRPQRYFCN(el)) return not_accepted;
  idx = (int)(ElfHash((unsigned char *)property) % PROPERTY_HASH_SIZE);
  for (prp = el->EqmPrpLst[idx]; prp != NULL; prp = prp->next)
  { /* find all the overloads ... */
    if (strnicmp(prp->prpName, property, PROPERTY_NAME_SIZE)) continue;
    prp->isReserved = TRUE;
  }
  return 0;
}
TINE_EXPORT int RedirectProperty(char *eqm,char *property,char *rdr) 
{
  char strbuf[256],msg[256];
  ExportPropertyListStruct *prp;
  ExportListStruct *el=getExportListItem(eqm);
  PrpRedirBlk *prd;
  int cc, idx;
  if (rdr == NULL || strlen(rdr) == 0) return argument_list_error; /* must be given ! */
  if (el == NULL) return non_existent_elem;

  idx = (int)(ElfHash((unsigned char *)property) % PROPERTY_HASH_SIZE);
  for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next) if (!strnicmp(prp->prpName,property,PROPERTY_NAME_SIZE)) break;
  if (prp == NULL) return illegal_property;
  if (prp->isReserved) return not_registered;

  strncpy(strbuf,rdr,255);
  if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) return out_of_server_memory;
  ParseRedirector(strbuf,prd);
  sprintf(msg, "RedirectProperty %.64s", property);
  if (assertRedirectionValid(msg,eqm,prd->prpRedirection))
  {
    prd->next = prp->prd;
    prp->prd = prd;
    cc = 0;
  }
  else
  {
    SystemFree(prd);
    cc = not_accepted;
  }
  return cc;
}
TINE_EXPORT int RegisterPropertyEx(char *eqm,char *property,int siz,short fmt,short acc,char *dsc,int pid) /* Appends property list for eqm */
{
  DTYPE dout,din,*dinPtr = NULL;
  dout.dArrayLength = siz;
  dout.dFormat = fmt;
  dout.dTag[0] = 0;
  if (acc & CA_WRITE)
  {
    memset(&din,0,sizeof(DTYPE));
    din.dArrayLength = siz;
    din.dFormat = fmt;
    dinPtr = &din;
  }
  return RegisterPropertyInformation(eqm,property,&dout,dinPtr,acc,AT_UNKNOWN,0,dsc,pid,NULL);
}
TINE_EXPORT int RegisterProperty(char *eqm,char *property,int siz,short fmt,short acc,char *dsc) /* Appends property list for eqm */
{
  return RegisterPropertyEx(eqm,property,siz,fmt,acc,dsc,-1);
}
TINE_EXPORT short RegisterPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,PrpQueryStruct **prpqs),int numprops)
{
  short cc=0;
  ExportListStruct *el=getExportListItem(eqm);
  char *peqm = eqm == NULL ? "(null)" : eqm;

  if (el == NULL) ccerr(non_existent_elem);
  el->EqmPropertyQueryFcn = fcn;
  el->EqmNumProperties = numprops;
err:
  feclog("%.6s Registered property query function : %.32s", peqm, cc2str(cc));
  return cc;
}
#ifdef MULTITHREADED
TINE_EXPORT int SetCallPropertyInSeparateThread(char *eqm,char *property,int value)
{
  ExportPropertyListStruct *p, *pl = GetPropertyListStruct(eqm,property,NULL);
  if (pl == NULL) return illegal_property;
  pl->runInSeparateThread = (short)value;
  for (p=pl->next; p != NULL; p=p->next)
  { /* set same for any overloads */
    if (strnicmp(p->prpName,property,PROPERTY_NAME_SIZE)) continue;
    p->runInSeparateThread = pl->runInSeparateThread;
    if (p->runInSeparateThread == (short)(CA_WRITE|CA_INDEPENDENT))
    { /* explicit instructions */
      p->prpAccessMode |= (short)CA_INDEPENDENT;
    }
  }
  if (value) useMultiThreadedEquipmentFunctions = TRUE;
  return 0;
}
#endif
#ifndef SMALL_TINE_LIB
TINE_EXPORT short RegisterXPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,XPropertyQueryStruct **xpqs),int numprops)
{
  short cc=0;
  ExportListStruct *el=getExportListItem(eqm);
  char *peqm = eqm == NULL ? "(null)" : eqm;

  if (el == NULL) ccerr(non_existent_elem);
  el->XEqmPropertyQueryFcn = fcn;
  el->EqmNumProperties = numprops;
err:
  feclog("%.6s Registered (legacy) property query function : %.32s", peqm, cc2str(cc));
  return cc;
}
#endif
TINE_EXPORT int SealTaggedStruct(char *tag,int size,int number)
{
  int cc = 0;
  struct structStruct *s;
  struct structFormat *sf;

  if (isInvalidStructureTag(tag)) ccerr(invalid_structure_tag);
  if ((s=findstruct(tag)) == NULL) ccerr(un_allocated);
  for (sf=s->s; sf != NULL; sf=sf->nxt) sf->owner = s;
  if (!s->hasExtSpace && size < s->rawlen) size = s->rawlen; /* trap this legacy problem */
  s->num = number;
  s->size = size;
# ifndef DOS
  if ((s->sdata=(BYTE *)SystemCalloc(number,s->size)) == NULL) ccerr(out_of_local_memory);
# else
  s->rawlen = s->size;
# endif
  s->sealed = TRUE;
err:
  if (logStructureEntries)
  {
    feclog("SealTaggedStruct %.16s (size %d bytes, reserved space %d structures) : %.32s",tag,size,number,cc2str(cc));
  }
  return cc;
}
/**
 * \internal
 */
#ifndef SMALL_TINE_LIB
void mapAndCallbackStructField(ConTblEntry *c)
{
  ReLnkCbEntry *cb;
  ReLnkTblEntry *r = c->rlkEntry;
  struct structStruct *s;
  struct structFormat *sf;
  DTYPE d;
  int n, siz = (int)c->lastDataSize;
  BYTE *dptr = (BYTE *)c->dataPtr;

  if (r == NULL || dptr == NULL) return;
  if ((s=findstructEx(c->strTagOut,c->ctxName,c->expName)) == NULL) return;
  n = ((int)c->sizeOut)/s->rawlen;
  if (!s->hasExtSpace && r->lnkSiz != r->siz && r->siz != n) return;
  for (sf=s->s; sf != NULL && strnicmp(r->field,sf->field,FIELD_NAME_SIZE); sf = sf->nxt);
  if (sf == NULL) return;
  d.dArrayLength = r->lnkSiz;
  d.dFormat = r->lnkFmt;
  d.data.bptr = r->dat;
  dptr += sf->addr;
  if (d.dFormat == CF_STRING)
  {
    char **sa, *sb;
    int i, p;
    sa = (char **)dptr; sb = sa[0];
    for (i=0,p=0; i<n; i++)
    { /* scan for terminating '0's*/
      sa[i] = &sb[p];
      p += (int)strlen(sa[i])+1;
      if (sb[p] == 0) break;
      if (p >= r->siz) break;
    }
    siz = i+1;
  }
  PutValuesFromAny(&d,dptr,(short)sf->fmt,1,s->size,siz,0);
  for (cb=r->cbLst; cb != NULL; cb=cb->nxt)
  {
    if (cb->lnkId == c->linkId)
    {
      TCBCALL(cb,c->linkStatus);
      break;
    }
  }
  return;
}
#endif
TINE_EXPORT int AssignDataStampsToGlobal(char *keyword,int dataStamp,int sysStamp)
{
  int i;
  for (i=0; i<ngcasts; i++) if (!strcmp(gcastTbl[i]->keyword,keyword)) break;
  if (i == ngcasts) return invalid_keyword;
  gcastTbl[i]->dataStamp = dataStamp;
  gcastTbl[i]->sysStamp = sysStamp;
  return 0;
}
TINE_EXPORT int sendNetGlobalEx(char *context,char *keyword,DTYPE *din,void (*callback)(int,int),int minPeriod,int maxPeriod,double tolerance)
{
  char fullname[128];
  if (context == NULL) context = gDeviceContext;
  if (keyword == NULL) return argument_list_error;
  sprintf(fullname,"/%.32s/%.64s",context,keyword);
  return sendNetGlobal(fullname,din,callback,minPeriod,maxPeriod,tolerance);
}
TINE_EXPORT int sendNetGlobal(char *keyword,DTYPE *din,void (*callback)(int,int),int minPeriod,int maxPeriod,double tolerance)
{
  int i=0, cc=0, isRegistered=FALSE;
  UINT16 dsize;
  double ts = MakeDataTimeStamp();

  if (!ServerInitialized) return not_ready;
  if (!canSendGlobals)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "sending globals not enabled !");
    ccerr(un_allocated);
  }
  if (keyword == NULL || din == NULL || din->data.vptr == NULL) ccerr(argument_list_error);
  for (i=0; i<ngcasts; i++) if (!strcmp(gcastTbl[i]->keyword,keyword)) break;
  if (i == ngcasts)
  {
    if (ngcasts >= GCastTblCapacity) ccerr(resources_exhausted);
    if ((gcastTbl[i]=(GlobalListStruct *)SystemCalloc(1,sizeof(GlobalListStruct))) == NULL)
      ccerr(out_of_client_memory);
    if (minPeriod <= 20) minPeriod *= 1000; /* convert to msec */
    if (maxPeriod <= 20) maxPeriod *= 1000;
    strncpy(gcastTbl[i]->keyword,keyword,KEYWORD_NAME_SIZE);
    gcastTbl[i]->size = (short)din->dArrayLength;
    gcastTbl[i]->lastsize = gcastTbl[i]->size; /* note the initial registration size */
    gcastTbl[i]->format = din->dFormat;
    gcastTbl[i]->cbId = i;
    gcastTbl[i]->cbfcn = (TCBFCNP)callback;
    gcastTbl[i]->maxPeriod = maxPeriod;
    gcastTbl[i]->minPeriod = minPeriod;
    gcastTbl[i]->tolerance = tolerance;
    dsize = gcastTbl[i]->size*fmtsizeof(gcastTbl[i]->format);
    if ((gcastTbl[i]->refdata=(BYTE *)SystemCalloc(1,dsize)) == NULL) ccerr(out_of_server_memory);
    gcastTbl[i]->time = (time_t)ts;
    feclog("GLOBAL %.64s (%d bytes) : %.32s",gcastTbl[i]->keyword,dsize,cc ? cc2str(cc) : "registered");
    if (minPeriod > 0 && minPeriod < MaxPollingRate)
    {
      if (gSystemCycleDeadband > (UINT32)minPeriod) 
      {
        feclog("adjust cycle deadband (due to net global send) from %d to %d msec",
          gSystemCycleDeadband,minPeriod);
        gSystemCycleDeadband = (UINT32)minPeriod;
      }
      MaxPollingRate = minPeriod;
      feclog("adjust maximum polling interval to %d msec",MaxPollingRate);
    }
    ngcasts++;
  }
  else
  {
    if ((short)din->dArrayLength > gcastTbl[i]->lastsize) ccerr(dimension_error);
    if (din->dFormat != gcastTbl[i]->format) ccerr(illegal_format);
    if (gcastTbl[i]->isBeingSent) ccerr(semaphore_busy);
    gcastTbl[i]->size = (short)din->dArrayLength;
    isRegistered = TRUE;
  }
  dsize = gcastTbl[i]->size*fmtsizeof(gcastTbl[i]->format);
  gcastTbl[i]->data = (BYTE *)din->data.vptr;
  gcastTbl[i]->sysStamp = din->sysStamp > 0 ? din->sysStamp : gSystemDataStamp;
  gcastTbl[i]->dataStamp = din->dStamp;
  if (isRegistered)
  { /* send it out immediately if already registered */
    gcastTbl[i]->tv.tv_sec = 0;
    gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
    doGCast(&thisCycleTime);
  }
err:
  return cc ? -cc : i;
}
TINE_EXPORT void SetAppDate(time_t appdate)
{
  appmktime = appdate;
  return;
}
TINE_EXPORT void SetAppVersion(int major,int minor,int revision)
{
  sprintf(appmkversion,"%d.%02d.%04d",major,minor,revision);
  return;
}
/* this is a documented API call (see tinetype.h) */
TINE_EXPORT int SetNameServerAddress(char *ens)
{ /* comma separated list of addresses */
  int i, n = 1, cc = 0, p;
  char s[256],*c, *c1, *ip, *po, *cp=NULL;

  if (gotNameServerAddress) return 0;
  if (ens == NULL) ccerr(argument_list_error);
  strncpy(s,ens,256);
  for (i=0; i<256; i++) if (s[i] == ',') n++;
  if ((NameServer=(FecAddrStruct *)SystemCalloc(n,sizeof(FecAddrStruct))) == NULL) ccerr(out_of_server_memory);
  for (i=0, c=strtok_r(s,",",&cp); i<n && c != NULL; i++, c=strtok_r(NULL,",",&cp))
  {
    ip = c; po = NULL;
    if ((c1=strchr(c,':')) != NULL)
    {
      *c1++ = 0; po = c1;
    }
    sprintf(NameServer[i].fecName,"ENS#%d",i);
    strncpy(NameServer[i].strAdr,ip,ADDR_SIZE_IPV6);
    p = getProtocolFromAddressString(NameServer[i].strAdr);
    fixInetProtocol(&NameServer[i],p);
    NameServer[i].inetProtocol = p;
    if (po != NULL) 
    {
      NameServer[i].portOffset = atoi(po);
    }
  }
  numNameServers = i;
  gotNameServerAddress = 1;
err:
  return cc;
}
TINE_EXPORT void SetEqmCompletion(char *eqm,char *errstr)
{
  ExportListStruct *el=getExportListItem(eqm);
  
  if (el == NULL) 
  {
    if (ExportList != NULL && ExportList->next == NULL) el = ExportList; else return;
  }
  memset(el->EqmLastErrorString,0,STATUS_SIZE);
  if (errstr != NULL && strlen(errstr) > 0)
  {
    strncpy(el->EqmLastErrorString,errstr,STATUS_SIZE-1);
  }
}
TINE_EXPORT void SetRedirectionParameters(char *eqm,char *rdrCnt,char *rdrSrv,char *rdrDev,char *rdrPrp)
{
  ExportListStruct *el=getExportListItem(eqm);
  
  if (el == NULL) 
  {
    if (ExportList != NULL && ExportList->next == NULL) 
    { /* there's only one equipment module being managed: use it */
      el = ExportList; 
    }
    else 
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "Equipment module %.6s not registered",(eqm == NULL ? "NULL" : eqm));
      return;
    }
  }
  memset(el->EqmLastErrorString,0,STATUS_SIZE);
  if (rdrCnt != NULL) strncpy(&el->EqmLastErrorString[PRP_REDIR_CTX_OFFSET],rdrCnt,CONTEXT_NAME_SIZE);
  if (rdrSrv != NULL) strncpy(&el->EqmLastErrorString[PRP_REDIR_SRV_OFFSET],rdrSrv,EXPORT_NAME_SIZE);
  if (rdrDev != NULL) strncpy(&el->EqmLastErrorString[PRP_REDIR_DEV_OFFSET],rdrDev,DEVICE_NAME_SIZE);
  if (rdrPrp != NULL) strncpy(&el->EqmLastErrorString[PRP_REDIR_PRP_OFFSET],rdrPrp,PROPERTY_NAME_SIZE);
}

TINE_EXPORT void SetlookupRedirectionNameStub(int (*fcn)(char *eqm,char *prpName,char *devName))
{
  lookupRedirectionNameStub = fcn;
}
/**
 * \internal
 */
int getBitShiftFromMask(UINT32 mask,int fmt)
{
  int shf = 0, b = 1, mshf;

  mshf = fmtsizeof(fmt) * 8 - 1;
  while ((mask & b) == 0)
  {
    if (++shf == mshf) break;
    b *= 2;
  }
  return shf;
}
/**
 * \internal
 */
TINE_EXPORT DBITFIELD *findBitField(char *srv,char *tag)
{
# define CDI_LOCALHOST "/LOCALHOST/CDI"
  static char *CDI_BITFIELD_SRV = "thisfec";
  DBITFIELD *bf;
  if (tag == NULL || strlen(tag) == 0) return NULL;
  if (srv != NULL && !stricmp(srv,CDI_LOCALHOST))
  { /* CDI uses this in bitfield registration */
    srv = CDI_BITFIELD_SRV;
  }
  for (bf=gBitFieldList; bf; bf=bf->next)
  {
     if (strncmp(bf->name,tag,TAG_NAME_SIZE)) continue;
     if (srv != NULL && strncmp(bf->srvName,srv,2*CONTEXT_NAME_SIZE+2)) continue;
     break;
  }
  return bf;
}
void rmvBitFieldList(void)
{
  DBITFIELD *bf=gBitFieldList,*bfn=NULL;
  struct bitfieldFormat *bff, *bffn=NULL;

  while (bf != NULL)
  {
    bfn = bf->next;
    for (bff=bf->bfs; bff != NULL; bff=bffn)
    {
      bffn = bff->nxt;
      SystemFree(bff);
    }
    SystemFree(bf);
    bf = bfn;
  }
  gBitFieldList = NULL;
}

/**
 * \internal
 */
int queryBitfieldFormat(char *srv,char *tag,NAME16I *bffq,int *bffqsiz)
{
  int i,n,nq,cc=0;
  DBITFIELD *bf;
  struct bitfieldFormat *bff;
  
  if (bffqsiz == NULL) ccerr(argument_list_error);
  nq = *bffqsiz;
  if ((bf=findBitField(srv,tag)) == NULL) ccerr(invalid_structure_tag);
  for (bff=bf->bfs,n=0; bff != NULL; bff=bff->nxt,n++); /* count them first (reverse ordered!) */
  for (i=0,bff=bf->bfs; i<nq && i<n && bff != NULL; i++, bff=bff->nxt)
  {
    strncpy(bffq[n-i-1].name,bff->field,16);
    bffq[n-i-1].ival = bff->bitmask;
  }
  if (i == nq) i--;
  strncpy(bffq[i].name,tag,16);
  bffq[i].ival = bf->fmt;
  *bffqsiz = n+1;
err:
  return cc;
}
/**
 * \internal
 */
UINT32 getBitfieldValue(int fmt,BYTE *dptr)
{
  switch (fmt)
  {
    case CF_BYTE:
    case CF_BITFIELD8:   return (UINT32)(*dptr);
    case CF_INT16:
    case CF_UINT16:
    case CF_BITFIELD16: return (UINT32)(*(UINT16 *)dptr);
    case CF_INT32:
    case CF_UINT32:
    case CF_BITFIELD32: return *(UINT32 *)dptr;
    default: break;
  }
  return (UINT32)(*dptr);
}
/**
 * \internal
 */
#ifndef SMALL_TINE_LIB
void mapAndCallbackBitfield(ConTblEntry *c)
{
  ReLnkCbEntry *cb;
  ReLnkTblEntry *r = c->rlkEntry;
  UINT32 v;
  char srv[CONTEXT_NAME_SIZE*2+4];
  int i, siz = (int)c->lastDataSize;
  BYTE *dptr = (BYTE *)c->dataPtr;

  if (r == NULL || dptr == NULL) return;
  if (strlen(c->strTagOut) == 0) return;

  sprintf(srv,"/%.32s/%.32s",c->ctxName,c->expName);
  if (r->siz != (int)c->sizeOut) return;
  for (i=0; i<siz; i++)
  {
    switch (r->lnkFmt)
    {
      case CF_BYTE:
      case CF_BITFIELD8:   
        v = (UINT32)(dptr[i]);
        v = getFieldFromBitfield(srv,r->tag,r->field,v);
        dptr[i] = (BYTE)v;
        break;
      case CF_INT16:
      case CF_UINT16:
      case CF_BITFIELD16:
        v = (UINT32)((UINT16 *)dptr)[i];
        v = getFieldFromBitfield(srv,r->tag,r->field,v);
        ((UINT16 *)dptr)[i] = (UINT16)v;
        break;
      case CF_INT32:
      case CF_UINT32:
      case CF_BITFIELD32:
        v = ((UINT32 *)dptr)[i];
        v = getFieldFromBitfield(srv,r->tag,r->field,v);
        ((UINT32 *)dptr)[i] = v;
        break;
    }
  }
  for (cb=r->cbLst; cb != NULL; cb=cb->nxt)
  {
    if (cb->lnkId == c->linkId)
    {
      TCBCALL(cb,c->linkStatus);
      break;
    }
  }
  return;
}
#endif
/**
 * \internal
 */
int acquireAndRegisterBitfield(const char *devName, char *tag)
{
  int i, n, fmt, cc=0;
  DTYPE dout,din;
  NAME16I n16i[32];
  char srv[2*CONTEXT_NAME_SIZE+4];

  if (devName == NULL || tag == NULL) ccerr(argument_list_error);
  if ((cc=splitFullName(devName,srv,NULL)) != 0) ccerr(cc);
  if (findBitField(srv,tag) != NULL) return already_assigned;

  dout.dFormat = CF_NAME16I;
  dout.dArrayLength = 32;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = n16i;

  din.dFormat = CF_TEXT;
  din.dArrayLength = (int)strlen(tag);
  memset(din.dTag,0,TAG_NAME_SIZE);
  din.data.cptr = tag;

  if ((cc = ExecLinkEx(devName,"BITFIELDFORMAT",&dout,&din,CA_READ|CA_MUTABLE|CA_RETRY,200)) != 0)
    ccerr(cc);

  n = dout.dArrayLength;
  fmt = n16i[n-1].ival;
  OpenBitField(srv,tag,fmt);
  for (i=0; i<n-1; i++)
  {
    addFieldToBitField(srv,tag,n16i[i].ival,n16i[i].name); 
  }
err:
  if (cc) feclog("acquireAndRegisterBitfield %.64s %.16s : %.32s",devName,tag,cc2str(cc));
  return cc;
}
TINE_EXPORT int AcquireAndRegisterBitfield(const char *devName, char *tag)
{
  return acquireAndRegisterBitfield(devName,tag);
}
TINE_EXPORT int AddFieldToBitField(char *srv,char *tag,UINT32 mask,char *field)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  if (bf == NULL) return invalid_structure_tag;
  if (field == NULL || strlen(field) == 0 || mask == 0) return argument_list_error;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    if (!strnicmp(field,bff->field,FIELD_NAME_SIZE)) return name_already_exists;
    if (mask == bff->bitmask) return already_assigned;
  }
  bff = (struct bitfieldFormat *)SystemCalloc(1,sizeof(struct bitfieldFormat));
  if (bff == NULL) return out_of_local_memory;
  bff->bitmask = mask;
  bff->bitshift = getBitShiftFromMask(mask,bf->fmt);
  if (strlen(field) > FIELD_NAME_SIZE) 
    feclog("truncating bitfield %.16s field name %s to %d characters",bf->name,field,FIELD_NAME_SIZE);
  strncpy(bff->field,field,FIELD_NAME_SIZE);
  bff->nxt = bf->bfs;
  bf->bfs = bff;
  return 0;
}
TINE_EXPORT int GetBitfieldAsString(char *srv,char *tag,UINT32 value,char *strbuf,int buflen)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  int n;
  if (bf == NULL) return invalid_structure_tag;
  if (strbuf == NULL) return argument_list_error;
  for (bff=bf->bfs, n = 0; bff != NULL; bff=bff->nxt, n++);
  if (buflen < n * 32) return buffer_too_small;
  n = 0;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    sprintf(&strbuf[n],"%.16s : %d\n",bff->field,getFieldFromBitfield(srv,tag,bff->field,value));
    n = (int)strlen(strbuf);
  }
  return 0;
}
TINE_EXPORT int GetFieldFromBitfield(char *srv,char *tag,char *field,UINT32 value)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  UINT32 bfvalue;
  if (bf == NULL) return -invalid_structure_tag;
  if (field == NULL || strlen(field) == 0) return -argument_list_error;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    if (!strnicmp(field,bff->field,FIELD_NAME_SIZE)) break;
  }
  if (bff == NULL) return -invalid_field;
  bfvalue = bff->bitmask & value;
  bfvalue = bfvalue >> bff->bitshift;
  return bfvalue;
}
TINE_EXPORT int OpenBitField(char *srv,char *tag,int fmt)
{
  DBITFIELD *bf = findBitField(srv,tag);

  if (bf != NULL) return 0;
  if (!isIntegerFormat((short)fmt))
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "OpenBitField: cannot register bitfield with format %s",itof(fmt));
    return illegal_format;
  }
  if (fmt == CF_BYTE) fmt = CF_BITFIELD8;
  if (fmt == CF_INT16 || fmt == CF_UINT16) fmt = CF_BITFIELD16;
  if (fmt == CF_INT32 || fmt == CF_UINT32) fmt = CF_BITFIELD32;
  if ((bf=(DBITFIELD *)SystemCalloc(1,sizeof(DBITFIELD))) != NULL)
  { /* didn't find it -> need to allocated it */
    if (strlen(tag) > TAG_NAME_SIZE) feclog("truncating bitfield tag name %.16s to %d characters",tag,TAG_NAME_SIZE);
    strncpy(bf->name,tag,TAG_NAME_SIZE);
    if (srv != NULL) strncpy(bf->srvName,srv,2*CONTEXT_NAME_SIZE+4);
    bf->fmt = fmt;
    bf->next = gBitFieldList;
    gBitFieldList = bf;
    if (tineDebug) dbglog("OpenBitField : New bitfield tag registered: %.16s",tag);
  }
  else
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "OpenBitField: cannot allocate bit field memory!");
    return out_of_local_memory;
  }
  return 0;
}
TINE_EXPORT void SetGCastTableCapacity(int value)
{ /* GCastTblCapacity established following allocation */
  if (value < 0) return;
  MaxNumGCasts = value;
}
TINE_EXPORT int GetGCastTableCapacity(void)
{
  return GCastTblCapacity;
}
TINE_EXPORT void SetRejectOverloadedMetaProperties(int value)
{
  gRejectOverloadedMetaProperties = value;
}
TINE_EXPORT int GetRejectOverloadedMetaProperties(void)
{
  return gRejectOverloadedMetaProperties;
}
TINE_EXPORT void SetEnforceMultiChannelAccess(int value)
{
  gEnforceMcaAccess = value;
}
TINE_EXPORT int GetEnforceMultiChannelAccess(void)
{
  return gEnforceMcaAccess;
}
TINE_EXPORT void SetAllowRemoteManagement(int value)
{
  gAllowRemoteReset = value;
  gAllowRemoteInit = value;
  gAllowRemoteExit = value;
}
TINE_EXPORT int GetAllowedRemoteManagement(void)
{
  return gAllowRemoteExit|gAllowRemoteInit|gAllowRemoteReset;
}
char *logFileScanText[3] = { "NONE", "FEC directories", "FULL" };
TINE_EXPORT void SetLogFileAllowScan(int value)
{
  if (value < LOG_SCAN_NONE) value = LOG_SCAN_NONE;
  if (value > LOG_SCAN_FULL) value = LOG_SCAN_FULL;
  gLogfileAllowScan = value;
}
TINE_EXPORT int GetLogFileAllowScan(void)
{
  return gLogfileAllowScan;
}
TINE_EXPORT void SetSchedulerInterval(int value)
{
  gSchedulerInterval = value;
}
TINE_EXPORT int GetSchedulerInterval(void)
{
  return gSchedulerInterval;
}
TINE_EXPORT void SetServerCycleIdle(int value)
{
  gServerCycleIdle = value;
}
TINE_EXPORT int GetServerCycleIdle(void)
{
  return gServerCycleIdle;
}
void resetLinkCounters(void)
{
  int i;
  ClnLst *cln;
  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i] == NULL) continue;
    if (ContractList[i]->nclients == 0) continue;
    for (cln=ContractList[i]->clnLst; cln!=NULL; cln=cln->next)
    {
      if (cln == NULL) continue;
      if (BASEMODE(cln->status.mode) <= CM_SINGLE) continue;
      if (cln->status.counter <= gSubRenewalLength) continue;
      cln->status.counter = gSubRenewalLength;
    }
  }
}
TINE_EXPORT void SetSystemSubscriptionRenewalLength(int value)
{
  int doreset = FALSE;
  if (value < CTR_RENEWAL) value = CTR_RENEWAL;
  if (value > CTR_MAXIMUM) value = CTR_MAXIMUM;
  if (value < gSubRenewalLength) doreset = TRUE;
  gSubRenewalLength = value;
  gSubRenewalMultiplier = gSubRenewalLength/CTR_RENEWAL;
  if (gSubRenewalMultiplier < 1) gSubRenewalMultiplier = 1;
  if (gSubRenewalMultiplier > 500) gSubRenewalMultiplier = 500;
  if (doreset) resetLinkCounters();
  feclog("set subscription renewal length to %d",gSubRenewalLength);
}
TINE_EXPORT int GetSystemSubscriptionRenewalLength(void)
{
  return gSubRenewalLength;
}
TINE_EXPORT void SetSystemCycleDeadband(int value)
{
  feclog("set system cycle deadband from %d to %d",gSystemCycleDeadband,value);
  gSystemCycleDeadband = (UINT32)value;
  gCycleDeadbandFixed = TRUE; /* has been set via API call */
  if (gSystemCycleDeadband > MinPollingRate) 
  {
    feclog("adjust minimum allowed polling interval from %d to %d",MinPollingRate,gSystemCycleDeadband);
    MinPollingRate = gSystemCycleDeadband;
  }
}
TINE_EXPORT int GetSystemCycleDeadband(void)
{
  return (int)gSystemCycleDeadband;
}
TINE_EXPORT void SetMinimumAllowedPollingInterval(int value)
{
  if (value < 10) value = 10;
  if (value > 1000) value = 1000;
  MinPollingRate = value;
  feclog("minimum allowed polling interval set to %d",value);
  if (gSystemCycleDeadband > MinPollingRate) 
  {
    feclog("adjust system cycle deadband from %d to %d",gSystemCycleDeadband,MinPollingRate);
    gSystemCycleDeadband = MinPollingRate;
    gCycleDeadbandFixed = TRUE;
  }
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int GetMinimumAllowedPollingInterval(void)
{
  return MinPollingRate;
}
TINE_EXPORT void SetEnforceMcaAcquisition(int value)
{
  enforceMcaAcquisition = value;
}
TINE_EXPORT int GetEnforceMcaAcquisition(void)
{
  return enforceMcaAcquisition;
}
TINE_EXPORT void SetSystemSynchronizeContracts(int value)
{
  gSynchronizeContracts = value;
}
TINE_EXPORT int GetSystemSynchronizeContracts(void)
{
  return gSynchronizeContracts;
}
TINE_EXPORT void SetRetardSingleContractRemoval(int value)
{
  gRetardSingleContractRemoval = value;
}
TINE_EXPORT int GetRetardSingleContractRemoval(void)
{
  return gRetardSingleContractRemoval;
}
TINE_EXPORT void SetAllowBackgroundTaskReentrancy(int value) 
{ 
  gAllowBkgTaskReentrancy = value; 
}
TINE_EXPORT int GetAllowBackgroundTaskReentrancy(void) 
{ 
  return gAllowBkgTaskReentrancy; 
}
TINE_EXPORT int SetSystemAlias(char *alias, char *name)
{
  return addAliasToTable(name,alias);
}
TINE_EXPORT char *GetSystemAlias(char *name)
{
  return getAliasFromTable(name);
}
TINE_EXPORT char *ResolveSystemAlias(char *alias)
{
  return resolveAliasFromTable(alias);
}
TINE_EXPORT void SetAllowNetworkAddressResolution(int value)
{
  gAllowNetworkAddressResolution = value;
  gIgnoreVolatileAddressCache = value;
  feclog("set allow network address resolution to %s",value ? "TRUE" : "FALSE");
}
TINE_EXPORT int GetAllowNetworkAddressResolution(void)
{
  return gAllowNetworkAddressResolution;
}

SystemAttribute systemAttributes[] = 
{
  { "SubscriptionRenewalLength", &gSubRenewalLength, SetSystemSubscriptionRenewalLength, CF_INT32 },
  { "RejectOverloadedMetaProperties", &gRejectOverloadedMetaProperties, SetRejectOverloadedMetaProperties, CF_INT32 },
  { "EnforceMcaAccess", &gEnforceMcaAccess, SetEnforceMultiChannelAccess, CF_INT32 },
  { "ServerCycleIdle", &gServerCycleIdle, SetServerCycleIdle, CF_INT32 },
  { "SchedulerInterval", &gSchedulerInterval, SetSchedulerInterval, CF_INT32 },
  { "MinimumAllowedPollingInterval", &MinPollingRate, SetMinimumAllowedPollingInterval, CF_INT32 },
  { "RetartSingleContractRemoval", &gRetardSingleContractRemoval, SetRetardSingleContractRemoval, CF_INT32 },
  { "AllowBackgroundTaskReentrancy", &gAllowBkgTaskReentrancy, SetAllowBackgroundTaskReentrancy, CF_INT32 },
  { "AutoLinkErrorAlarms", &gAutoLinkErrorAlarms, SetAutoLinkErrorAlarms, CF_INT32 },
  { "SynchronizationTolerance", &gSyncTolerance, (ISETFCN)SetSynchronizationTolerance, CF_DOUBLE },
  { "ClientRecvQueueDepth", &gClnRecvQueueDepth, SetClnRecvQueueDepth, CF_INT32 },
  { "UseSynchronousNotification", &gUseSynchronousNotification, UseSynchronousNotification, CF_INT32 },
  { "SystemCycleDeadband", &gSystemCycleDeadband, SetSystemCycleDeadband, CF_INT32 },
  { "TcpConnectionTableSize", &tcpConTblSize, SetTcpConnectionTableSize, CF_INT32 },
  { "PortOffset", &gPortOffset, NULL, CF_UINT16 },
  { "FecName", &gFecName, NULL, CF_NAME16 },
  { "Context", &gDeviceContext, NULL, CF_NAME32 },
  { "SubSystem", &gDeviceSubSystem, NULL, CF_NAME16 },
};
int nSystemAttributes = sizeof(systemAttributes)/sizeof(SystemAttribute);
TINE_EXPORT int GetSystemAttributeList(NAME64I *lst, int *length)
{
  int i, len;
  if (lst == NULL || length == NULL) return invalid_parameter;
  len = *length;
  if (len <= 0) return buffer_too_small;
  if (len > nSystemAttributes) len = nSystemAttributes;
  for (i=0; i < len; i++)
  {
    strncpy(lst[i].name, systemAttributes[i].name, 32);
    lst[i].ival = systemAttributes[i].type;
  }
  *length = nSystemAttributes;
  return 0;
}
/** 
 * \note by definition the Attribute setting function takes a signed integer
 */
TINE_EXPORT int SetSystemAttribute(char *attribute,void *value,int format)
{
  int i, ival=0, cc=invalid_name;
  double dval=0.0;
  if (attribute == NULL || value == NULL) return invalid_parameter;
  for (i=0; i<nSystemAttributes; i++)
  { /* scan thru list */
    if (!strnicmp(attribute,systemAttributes[i].name,32))
    { /* found it */
      if (systemAttributes[i].fcn == NULL) return illegal_read_write;
      if (tineDebug) dbglog("set attribute %.32s via API",systemAttributes[i].name);
      switch (format)
      {
        case CF_INT16:
          ival = (int)(*(short *)value);
          break;
        case CF_UINT16:
          ival = (int)(*(UINT16 *)value);
          break;
        case CF_INT32:
          ival = *(int *)value;
          break;
        case CF_UINT32:
          ival = (int)(*(UINT32 *)value);
          break;
        case CF_FLOAT:
          dval = (double)(*(float *)value);
          break;
        case CF_DOUBLE:
          dval = *(double *)value;
          break;
        default:
          return illegal_format;
      }
      switch (systemAttributes[i].type)
      {
        case CF_INT32:
          systemAttributes[i].fcn(ival);
          return 0;
        case CF_DOUBLE:
          ((DSETFCN)(systemAttributes[i].fcn))(dval);
          return 0;
        default:
          return illegal_format;
      }
    }
  }
  return cc;
}
TINE_EXPORT int GetSystemAttribute(char *attribute,void *value, int format)
{
  int i=0, ival=0, slen=0, cc = invalid_name;
  double dval=0.0;
  char *sval = NULL;
  if (attribute == NULL || value == NULL) return invalid_parameter;
  for (i=0; i<nSystemAttributes; i++)
  { /* scan thru list */
    if (!strnicmp(attribute,systemAttributes[i].name,32))
    { /* found it */
      if (tineDebug) dbglog("get attribute %.32s via API",systemAttributes[i].name);
      switch (systemAttributes[i].type)
      {
        case CF_UINT16:
          ival = (int)(*(UINT16 *)systemAttributes[i].attr);
          break;
        case CF_INT16:
          ival = (int)(*(short *)systemAttributes[i].attr);
          break;
        case CF_UINT32:
          ival = (int)(*(UINT32 *)systemAttributes[i].attr);
          break;
        case CF_INT32:
          ival = *(int*)systemAttributes[i].attr;
          break;
        case CF_DOUBLE:
          dval = *(double *)systemAttributes[i].attr;
          break;
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          sval = (char *)systemAttributes[i].attr;
          slen = fmtsizeof(systemAttributes[i].type);
          break;
        default:
          return illegal_format;
      }
      switch (format)
      {
        case CF_UINT16:
          *(UINT16 *)value = (UINT16)ival;
          return 0;
        case CF_INT16:
          *(short *)value = (short)ival;
          return 0;
        case CF_UINT32:
          *(UINT32 *)value = ival;
          return 0;
        case CF_INT32:
          *(int *)value = ival;
          return 0;
        case CF_FLOAT:
          *(float *)value = (float)dval;
          return 0;
        case CF_DOUBLE:
          *(double *)value = dval;
          return 0;
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
        case CF_TEXT:
          strncpy((char *)value, sval, slen);
          return 0;
        default:
          return illegal_format;
      }
    }
  }
  return cc;
}
#endif /* !SMALL_TINE_LIB */
