/*
 * File srvquery.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file srvquery.c
 * \brief query routines for the stock properties
 * \internal
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "prolog.h"
#include "srvcore.h"
#include "globals.h"
#include "errors.h"
#include "dbtools.h"
#include "almlib.h"
#include "hstlib.h"
#include "srvip.h"
#if defined(IPXNW) || defined(IPXBSD)
# include "srvipx.h"
#endif
#include "ttycmd.h"
#include "toolkit.h"

extern int gHaltOnCycleHalt;
extern int _gReinitIdleTime;
int getStringFromInputField(CONTRACT *con,char *dataIn,char *strbuf,int buflen)
{
  int len,fmt;
  if (con == NULL || strbuf == NULL || dataIn == NULL || buflen < 1) return argument_list_error;
  if ((len=(int)con->EqmSizeIn) == 0) return illegal_read_write;
  memset(strbuf,0,buflen);
  fmt = LFMT(con->EqmFormatIn);
  switch (fmt)
  {
    case CF_TEXT:
      len = MIN(len,buflen);
      if (len < 1) return illegal_read_write;
      strncpy(strbuf,dataIn,len);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      len = MIN(fmtsizeof(fmt),buflen);
      strncpy(strbuf,dataIn,len);
      break;
    default: 
      return illegal_format;
  }
  return 0;
}
int getStringFormatLength(short fmt,int siz,int maxText)
{
  int cc = 0, slen = 0;
  switch (LFMT(fmt))
  {
    case CF_NAME8: 
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      slen = fmtsizeof(fmt);
      if (maxText > 0 && siz != 1) ccerr(dimension_error);
      break;
    case CF_TEXT:
      if (maxText == 0) ccerr(illegal_format);
      if (siz < 1) ccerr(dimension_error);
      slen = MIN(siz,maxText);
      break;
    default: 
      cc = illegal_format;
      break;
  }
err:
  return cc ? -cc : slen;
}
int T_numQuery(CONTRACT *con,char *data,int num)
{
  if (con->EqmSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_UINT16:
    case CF_INT16: *(UINT16 *)data = (UINT16)num; break;
    case CF_UINT32:
    case CF_INT32: *(UINT32 *)data = (UINT32)num; break;
    case CF_FLOAT: *(float *)data = (float)num; break;
    case CF_DOUBLE: *(double *)data = (double) num; break;
    default: return illegal_format;
  }
  return 0;
}
int getInputInt32(CONTRACT *con,char *dataIn,int *v)
{
  DTYPE din;
  DTYPEZERO(din);
  din.dArrayLength = con->EqmSizeIn;
  din.dFormat = LFMT(con->EqmFormatIn);
  din.data.vptr = dataIn;
  return GetValuesAsLong(&din,v,1);
}
void touchStkMetaData(ContractListStruct *cl)
{
  cl->dtimestamp = (SINT32)gServerDataTimeStamp;
  cl->dtimestampUSEC = (int)((gServerDataTimeStamp - (SINT32)gServerDataTimeStamp) * 1000000);
  cl->sysstamp = gSystemDataStamp;
}
void assignStkMetaData(ContractListStruct *cl,double timeStamp,int sysStamp,int usrStamp, int retSize)
{ /* local use only: don't bother checking NULL pointer */
  if (retSize >= 0) cl->dataSizeOut = retSize;
  if (timeStamp > 1.0) 
  {
    cl->dtimestamp = (SINT32)timeStamp;
    cl->dtimestampUSEC = (int)((timeStamp - (SINT32)timeStamp) * 1000000);
  }
  if (sysStamp > 0) cl->sysstamp = sysStamp;
  if (usrStamp > 0) cl->datastamp = usrStamp;
  return;
}
int nstructListQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int n;
  int cc,includeSystem=FALSE;
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (stricmp(con->EqmDeviceName,"ALL")) includeSystem = TRUE;
  cc = getRegisteredStructNames(NULL,&n,LFMT(con->EqmFormatOut),FALSE,includeSystem);
  if (cc != 0) return cc;
  return T_numQuery(con,dataOut,n);
}
int structListQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int n = (int)con->EqmSizeOut;
  int cc,includeKeys=FALSE,includeSystem=FALSE;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (LFMT(con->EqmFormatOut) == CF_NAME64) includeKeys = TRUE;
  if (!stricmp(con->EqmDeviceName,"ALL")) includeSystem = TRUE;
  cc = getRegisteredStructNames(dataOut,&n,LFMT(con->EqmFormatOut),includeKeys,includeSystem);
  if (n < (int)con->EqmSizeOut) assignStkMetaData(cl,0,0,0,n);
  return cc;
}
int taggedStructQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int cc,rs = (int)(cl->dataSizeOut);
  char tag[32];
  CONTRACT *con=&cl->contract;
  if (rs == 0) return illegal_read_write;
  if ((cc=getStringFromInputField(con,dataIn,tag,TAG_NAME_SIZE)) != 0) return cc;
  touchStkMetaData(cl);
  tag[TAG_NAME_SIZE] = 0;
  cc = queryStructFormat(tag,(BYTE *)dataOut,LFMT(con->EqmFormatOut),&rs);
  if (rs < (int)con->EqmSizeOut) assignStkMetaData(cl,0,0,0,rs);
  return cc;
}
int bitfieldStructQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int cc,rs = (int)(cl->dataSizeOut);
  char tag[32];
  CONTRACT *con=&cl->contract;
  if (rs == 0) return illegal_read_write;
  if ((cc=getStringFromInputField(con,dataIn,tag,TAG_NAME_SIZE)) != 0) return cc;
  touchStkMetaData(cl);
  tag[TAG_NAME_SIZE] = 0;
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME16I:
      cc = queryBitfieldFormat(NULL,tag,(NAME16I *)dataOut,&rs);
      break;
    default:
      return illegal_format;
  }
  if (rs < (int)con->EqmSizeOut) assignStkMetaData(cl,0,0,0,rs);
  return cc;
}
#ifndef SMALL_TINE_LIB
int getEnsList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int len,i,rs = (int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (rs == 0) return illegal_read_write;
  if ((len=getStringFormatLength(con->EqmFormatOut,con->EqmSizeOut,0)) < 0) return -len;
  for (i=0; i<(int)con->EqmSizeOut && i<numNameServers; i++)
  {
    strncpy(dataOut + i*len,NameServer[i].strAdr,len);
  }
  if (i < (int)con->EqmSizeOut) assignStkMetaData(cl,0,0,0,i);
  return 0;
}
int getServerAliasList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int len,i,j,rs = (int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (rs == 0) return illegal_read_write;
  if ((len=getStringFormatLength(con->EqmFormatOut,con->EqmSizeOut,0)) < 0) return -len;
  for (i=0,j=0; i<(int)con->EqmSizeOut && j<nAliasTableEntries; i += 2,j++)
  {
    strncpy(dataOut + i*len,gAliasTable[j].name,len);
    strncpy(dataOut + (i+1)*len,gAliasTable[j].alias,len);
  }
  if (i < (int)con->EqmSizeOut) assignStkMetaData(cl,0,0,0,i);
  return 0;
}
int addAliasToTable(char *name,char *alias);
int dumpAliasTable(char *eqm);
int addAlias(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int cc, nlen;
  char name[64], alias[64];
  CONTRACT *con=&cl->contract;
  UNUSED(dataOut);
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqmAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  if (con->EqmSizeIn != 2) return dimension_error;
  touchStkMetaData(cl);
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      nlen = fmtsizeof(con->EqmFormatIn);
      memset(name,0,64); memset(alias,0,64);
      strncpy(name,dataIn,nlen);
      strncpy(alias,dataIn+nlen,nlen);
      break;
    default: 
      return illegal_format;
  }
  if ((cc=addAliasToTable(name,alias)) != 0) return cc;
  if ((cc=dumpAliasTable(con->EqmName)) != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot update alias file: %s",cc2str(cc));
  return 0;
}
#endif
int getNumReservedDevices(ExportListStruct *el)
{
  int i, n = 0;
  if (el != NULL && el->EqmDevLst != NULL)
  {
    for (i = 0, n = 0; i < el->EqmNumDevices; i++) 
      if (el->EqmDevLst[i].isReserved) n++;
  }
  return n;
}
int ndeviceQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  ExportPropertyListStruct *pl = NULL;
  int n;
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  n = el->EqmNumDevices - getNumReservedDevices(el);
  if (con->EqmSizeIn > 0)
  { /* property as input */
    if (!isSimpleStringFormat((short)LFMT(con->EqmFormatIn))) return illegal_format;
    if ((pl=GetPropertyListStruct(con->EqmName,dataIn,con->EqmDeviceName)) == NULL)
      return illegal_property;
    if (pl->numDevices > 0) n = pl->numDevices;
  }
  return T_numQuery(con,dataOut,n);
}
int deviceGroupQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con = &cl->contract;
  ExportListStruct *el = cl->exp;
  int rs = (int)(cl->dataSizeOut);
  int cc = 0, n = rs;
  NAME64 *mems = (NAME64 *)dataOut;
  UNUSED(dataIn);
  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME64) return illegal_format;
  touchStkMetaData(cl);
  cc = GetGroupDeviceMembers(el->EqmName, con->EqmDeviceName, mems, &n);
  if (n < rs) assignStkMetaData(cl, 0, 0, 0, n);
  return cc;
}
int deviceGroupsQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con = &cl->contract;
  ExportListStruct *el = cl->exp;
  int rs = (int)(cl->dataSizeOut);
  int cc = 0, n = rs;
  NAME64 *grps = (NAME64 *)dataOut;
  UNUSED(dataIn);
  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME64) return illegal_format;
  touchStkMetaData(cl);
  cc = GetDeviceGroups(el->EqmName, grps, &n);
  if (n < rs) assignStkMetaData(cl, 0, 0, 0, n);
  return cc;
}
int deviceQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int i,k,slen,tlen=0,start,mask=0,rs=(int)(cl->dataSizeOut);
  int cc = 0, n = 0;
  char btok[DEVICE_NAME_SIZE], etok[DEVICE_NAME_SIZE], mtok[DEVICE_NAME_SIZE];
  int blen=0,elen=0,mlen=0,chkWildCard=FALSE,chkMask=FALSE,chkOn=FALSE;

  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  touchStkMetaData(cl);
  start = con->EqmDeviceName[0] == '#' ? atoi(&con->EqmDeviceName[1]) : 0;
  if (start < 0) start = 0;

  if (strchr(con->EqmDeviceName,'*') != NULL)
  {
    chkOn = TRUE;
    if (con->EqmSizeIn == 1)
    {
      int fmt = LFMT(con->EqmFormatIn);
      if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
      memcpy(&mask,dataIn,sizeof(int));
      if (mask != 0) chkMask = TRUE;
    }
    if (strcmp(con->EqmDeviceName,"*") != 0)
    { /* device query with non-trivial wildcard string */
      getWildcardTokens(con->EqmDeviceName,btok,etok,mtok);
      blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
      chkWildCard = TRUE;
    }
  }
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME8: 
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      tlen = fmtsizeof(con->EqmFormatOut); 
      break;
#   ifndef SMALL_TINE_LIB
    case CF_STRING:
      if ((n=atoi(con->strTagOut)) < 1) return illegal_format;
      break;
#   endif
    case CF_STRUCT:
    { /* get all relevant info for a device in one fell swoop */
      ExportDeviceType *lst = el->EqmDevLst;
      DeviceInfoStruct dis;
      if (strcmp(con->strTagOut, "DEVQSr5") != 0) return invalid_structure_tag;
      if ((i=GetDeviceNumber(el->EqmName, con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
      memset(&dis, 0, sizeof(DeviceInfoStruct));
      strncpy(dis.ctx, el->EqmContext, CONTEXT_NAME_SIZE);
      strncpy(dis.srv, el->EqmExportName, EXPORT_NAME_SIZE);
      if (lst[i].name != NULL) strncpy(dis.name, lst[i].name, DEVICE_NAME_SIZE);
      if (lst[i].desc != NULL) strncpy(dis.desc, lst[i].desc, DEVICE_DESC_SIZE);
      if (lst[i].loc != NULL) strncpy(dis.loc, lst[i].loc, DEVICE_LOCATION_SIZE);
      dis.mask = lst[i].mask;
      dis.region = lst[i].region;
      dis.zpos = lst[i].zpos;
      dis.number = i;
      memcpy(dataOut, &dis, sizeof(DeviceInfoStruct));
      assignStkMetaData(cl, 0, 0, 0, sizeof(DeviceInfoStruct));
      return 0;
    }
    default: 
      return illegal_format;
  }
  for (i=0,k=0; k<rs && i+start<el->EqmNumDevices; i++)
  {
    if (el->EqmDevLst && (slen=(int)strlen(el->EqmDevLst[i+start].name)) > 0)
    {
      if (el->EqmDevLst[i + start].isReserved) continue;
      if (chkWildCard && 
          !wildcardMatches(el->EqmDevLst[i+start].name,btok,blen,etok,elen,mtok,mlen))
        continue;
      if (chkMask && el->EqmDevLst[i+start].mask != 0 && (el->EqmDevLst[i+start].mask & mask) == 0) continue;
      if (chkOn && (el->EqmDevLst[i+start].mask & DEVICE_OFFLINE)) continue;
#     ifndef SMALL_TINE_LIB
      if (n > 0)
      { /* must be CF_STRING format ! */
        if (slen > DEVICE_NAME_SIZE) slen = DEVICE_NAME_SIZE;
        if (k + slen + 1 > rs) break;
        strncpy(&dataOut[k],el->EqmDevLst[i+start].name,DEVICE_NAME_SIZE);
        k += slen; dataOut[k] = 0;
      }
      else
#     endif
      { 
        strncpy(&dataOut[k*tlen],el->EqmDevLst[i+start].name,tlen);
      }
    }
    else 
    {
      if (chkWildCard) continue;
#     ifndef SMALL_TINE_LIB
      if (n > 0)
      {
        if (k + 8 > rs) break;
        k += sprintf(&dataOut[k],"#%d",i); dataOut[k] = 0;
      }
      else
#     endif
      {
        sprintf(&dataOut[k*tlen],"#%d",i);
      }
      if (k == 0) k = 1;
      break;
    }
    k++;
  }
  if (el->hasPropertyDeviceNames) 
  {
    cc = has_query_function | CE_SENDDATA;
  }
  if (k < rs || n > 0) assignStkMetaData(cl,0,0,0,k);
  return cc;
}
#ifndef SMALL_TINE_LIB
int deviceLocQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int i,slen,rs=(int)(cl->dataSizeOut);
  char *c,*en,*pen,devloc[DEVICE_LOCATION_SIZE+8],rdrstr[256],msg[256];
  PrpRedirBlk prb;

  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_NAME64) return illegal_format;
    if (el->EqmDevLst == NULL) return not_initialized;
    if (el->EqmDevLst[i].loc == NULL &&
        (el->EqmDevLst[i].loc=(char *)SystemCalloc(1,64)) == NULL) return out_of_server_memory;
    strncpy(el->EqmDevLst[i].loc,dataIn,DEVICE_LOCATION_SIZE);
    SaveStockPropertyValues(el,con->EqmProperty);
    if (rs == 0) return 0;
  }
  if (rs == 0) return illegal_read_write;
  if ((slen=getStringFormatLength(con->EqmFormatOut,rs,64)) < 0) return -slen;
  if (el->EqmDevLst != NULL && el->EqmDevLst[i].loc != NULL) 
  {
    strncpy(devloc,el->EqmDevLst[i].loc,DEVICE_LOCATION_SIZE);
    if (devloc[0] == '<')
    { /* redirection character */
      memset(rdrstr,0,256);
      c = devloc;
      devloc[DEVICE_LOCATION_SIZE] = 0;
      memset(&prb,0,sizeof(PrpRedirBlk));
      while (*(++c) == ' ');
      while (*c == '$')
      { /* take the environment */
        en = ++c;
        while (*c != '/' && *c != 0) c++;
        if (strlen(c) > 0) *c++ = 0;
        if ((pen=getenv(en)) == NULL) 
        {
          if (tineDebug) dbglog("device location env variable %s is not set",en);
          strncpy(devloc,el->EqmDevLst[i].loc,DEVICE_LOCATION_SIZE);
          goto err;
        }
        sprintf(rdrstr,"%.72s",pen);
        if (strlen(c) > 0) strcat(rdrstr,"/");
      }
      if (strlen(c) > 0) 
      { /* take what's left literally */
        strcat(rdrstr,c);
      }
      ParseRedirector(rdrstr,&prb);
      if (prb.prpRedirection[PRP_REDIR_PRP_OFFSET] == 0)
      { /* property not set: try this (let the stock property 'DEVLOCATION' thru ?) */
        strcpy(&prb.prpRedirection[PRP_REDIR_PRP_OFFSET],"Location");
      }
      sprintf(msg, "deviceLocQuery for device %.64s",con->EqmDeviceName);
      if (!assertRedirectionValid(msg,con->EqmName,prb.prpRedirection))
      {
        if (tineDebug) dbglog("device location redirection to %.96s is not valid",rdrstr);
        strncpy(devloc,el->EqmDevLst[i].loc,DEVICE_LOCATION_SIZE);
        goto err;
      }
      memcpy(el->EqmLastErrorString,prb.prpRedirection,STATUS_SIZE);
      return server_redirection;
    }
    strncpy(devloc,el->EqmDevLst[i].loc,slen);
  }
  else
  { /* no registered device location : use the fec location */
    strncpy(devloc,gFecInfo.loc,slen);
  }
err:
  strncpy(dataOut,devloc,slen);
  return 0;
}
int deviceRegionQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int i,k,n,code,rs=(int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  if (el == NULL) return non_existent_elem;
  if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
  if (el->EqmDevLst == NULL) return not_initialized;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_UINT32: code = (int)(*(UINT32 *)dataIn); break;
      case CF_INT32: code = *(SINT32 *)dataIn; break;
      case CF_NAME64: code = GetValidDeviceRegion(con->EqmName,dataIn); break;
      default: return illegal_format;
    }
    el->EqmDevLst[i].region = code;
    SaveStockPropertyValues(el,"DEVREGION");
  }
  if (rs > 0)
  { /* deliver multi-channel if required ... */
    n = el->EqmNumDevices;
    if (rs > n) { assignStkMetaData(cl,0,0,0,n); rs = n; }
    for (k=0; k<rs; k++)
    {
      switch (LFMT(con->EqmFormatOut))
      {
        case CF_UINT32: ((UINT32 *)dataOut)[k] = el->EqmDevLst[(i + k) % n].region; break;
        case CF_INT32: ((SINT32 *)dataOut)[k] = el->EqmDevLst[(i+k)%n].region; break;
        case CF_NAME64: strncpy(((NAME64 *)dataOut)[k].name,GetRegionTag(con->EqmName,el->EqmDevLst[(i+k)%n].region),64); break;
        default: return illegal_format;
      }
    }
  }
  return 0;
}
int deviceOnlineQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int i, k, n,rs=(int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;

  if (el == NULL) return non_existent_elem;
  if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
  if (el->EqmDevLst == NULL) return not_initialized;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    int fmt = LFMT(con->EqmFormatIn);
    if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (*(SINT32 *)dataIn)
    {
      el->EqmDevLst[i].mask &= ~(DEVICE_OFFLINE);
    }
    else
    {
      el->EqmDevLst[i].mask |= DEVICE_OFFLINE;
    }
    SaveStockPropertyValues(el,"DEVMASK");
  }
  if (rs > 0)
  { /* deliver multi-channel if required ... */
    int fmt = LFMT(con->EqmFormatOut);
    if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
    n = el->EqmNumDevices;
    if (rs > n) { assignStkMetaData(cl,0,0,0,n); rs = n; }
    for (k=0; k<rs; k++)
    {
      ((SINT32 *)dataOut)[k] = (el->EqmDevLst[(i+k)%n].mask & DEVICE_OFFLINE) ? 0 : 1;
    }
  }
  return 0;
}
int deviceMaskQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int i,k,n,offlin,rs=(int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  if (el == NULL) return non_existent_elem;
  if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
  if (el->EqmDevLst == NULL) return not_initialized;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    int fmt = LFMT(con->EqmFormatIn);
    if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    offlin = (el->EqmDevLst[i].mask & DEVICE_OFFLINE);
    el->EqmDevLst[i].mask = *(SINT32 *)dataIn;
    if (offlin) el->EqmDevLst[i].mask |= DEVICE_OFFLINE; /* add this back */
    SaveStockPropertyValues(el,"DEVMASK");
  }
  if (rs > 0)
  { /* deliver multi-channel if required ... */
    int fmt = LFMT(con->EqmFormatOut);
    if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
    n = el->EqmNumDevices;
    if (rs > n) { assignStkMetaData(cl,0,0,0,n); rs = n; }
    for (k=0; k<rs; k++)
    {
      ((SINT32 *)dataOut)[k] = el->EqmDevLst[(i+k)%n].mask;
      ((SINT32 *)dataOut)[k] &= ~(DEVICE_OFFLINE); /* don't deliver the offline bit */
    }
  }
  return 0;
}
int deviceZPosQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int i, k, n,rs=(int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  if (el == NULL) return non_existent_elem;
  if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
  if (el->EqmDevLst == NULL) return not_initialized;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_FLOAT: el->EqmDevLst[i].zpos = *(float *)dataIn; break;
      case CF_DOUBLE: el->EqmDevLst[i].zpos = (float)(*(double *)dataIn); break;
      case CF_INT32: el->EqmDevLst[i].zpos = (float)(*(SINT32 *)dataIn); break;
      case CF_UINT32: el->EqmDevLst[i].zpos = (float)(*(UINT32 *)dataIn); break;
      default: return illegal_format;
    }
    SaveStockPropertyValues(el,con->EqmProperty);
  }
  if (rs > 0)
  { /* deliver multi-channel if required ... */
    n = el->EqmNumDevices;
    if (rs > n) { assignStkMetaData(cl,0,0,0,n); rs = n; }
    for (k=0; k<rs ; k++)
    {
      switch (LFMT(con->EqmFormatOut))
      {
        case CF_FLOAT: ((float *)dataOut)[k] = el->EqmDevLst[(i+k)%n].zpos; break;
        case CF_DOUBLE: ((double *)dataOut)[k] = (double)el->EqmDevLst[(i+k)%n].zpos; break;
        case CF_INT32: ((SINT32 *)dataOut)[k] = (SINT32)el->EqmDevLst[(i+k)%n].zpos; break;
        case CF_UINT32: ((UINT32 *)dataOut)[k] = (UINT32)el->EqmDevLst[(i + k) % n].zpos; break;
        default: return illegal_format;
      }
    }
  }
  return 0;
}
int deviceDescQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int i,k,istop,slen,rs=(int)(cl->dataSizeOut);
  short fmt = LFMT(con->EqmFormatOut);
  char devdsc[64], *devp=devdsc;

  UNUSED(dataIn);
  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) < 0 || i >= el->EqmNumDevices) return illegal_device;
  if (el->EqmDevLst == NULL) return not_initialized;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_NAME64) return illegal_format;
    if (el->EqmDevLst[i].desc == NULL &&
        (el->EqmDevLst[i].desc=(char *)SystemCalloc(1,DEVICE_DESC_SIZE)) == NULL) return out_of_server_memory;
    strncpy(el->EqmDevLst[i].desc,dataIn,DEVICE_DESC_SIZE);
    SaveStockPropertyValues(el,con->EqmProperty);
  }
  if (!isSimpleStringFormat(fmt)) return illegal_format;
  istop = (fmt == CF_TEXT) ? i+1 : MIN(i + (int)con->EqmSizeOut,el->EqmNumDevices);
  slen = (fmt == CF_TEXT) ? MIN((int)con->EqmSizeOut,64) : fmtsizeof(fmt);
  for (k=i; k<istop; k++)
  {
    if (el->EqmDevLst != NULL && el->EqmDevLst[k].desc != NULL) 
    {
      devp = el->EqmDevLst[k].desc;
    }
    else
    {
      sprintf(devp,"%.32s device #%d",el->EqmExportName,k);
    }
    strncpy(dataOut+(k-i)*slen,devp,slen);
  }
  if (fmt != CF_TEXT) assignStkMetaData(cl,0,0,0,istop - i);
  return 0;
}
#endif
int stockpropsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int i,k,n,fmt,len,slen,rs=(int)(cl->dataSizeOut);
  PropertyQueryStruct *pqs = (PropertyQueryStruct *)tmpWorkArea;
  PropertyQueryStructEx *pqsx = (PropertyQueryStructEx *)tmpWorkArea;
  XPropertyQueryStruct *xpqs = (XPropertyQueryStruct *)tmpWorkArea;
  PrpQueryStruct *prpqs = (PrpQueryStruct *)tmpWorkArea;
  int tgtprplen=0,fecPrpsOnly=FALSE,eqmPrpsOnly=FALSE;
  char *c,tgtprp[PROPERTY_NAME_SIZE+4];
  CONTRACT *con=&cl->contract;
  touchStkMetaData(cl);
  if ((len=(int)con->EqmSizeIn) > 0)
  {
    memset(tgtprp,0,PROPERTY_NAME_SIZE+4);
    if ((len=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -len;
    tgtprplen = 32;
    strncpy(tgtprp,dataIn,len);
    if ((c=strchr(tgtprp,'*')) != NULL)
    {
      *c = 0;
      tgtprplen = (int)strlen(tgtprp);
    }
    else if (stricmp(tgtprp,"FECONLY") == 0)
    {
      fecPrpsOnly = TRUE;
      tgtprplen = 0;
    }
    else if (stricmp(tgtprp,"EQMONLY") == 0)
    {
      eqmPrpsOnly = TRUE;
      tgtprplen = 0;
    }
  }
  switch (fmt=LFMT(con->EqmFormatOut))
  {
    case CF_NAME64:
    case CF_NAME48:
    case CF_NAME32:
    case CF_NAME16:
      if (rs == 0) return illegal_read_write;
      len = fmtsizeof(fmt);
      for (i=0,k=0; i<rs && i<gNrStockProperties; i++)
      {
        if (fecPrpsOnly && !gStockProperty[i].isFecProperty) continue;
        if (eqmPrpsOnly && gStockProperty[i].isFecProperty) continue;
        if (tgtprplen && strnicmp(gStockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strnicmp(dataOut+(k-1)*len,gStockProperty[i].prpName,len)) continue;
        strncpy(dataOut+k*len,gStockProperty[i].prpName,len);
        k++;
      }
      if (k < rs) assignStkMetaData(cl,0,0,0,k);
      return 0;
    case CF_STRING:
      if ((n=atoi(con->strTagOut)) < 1) return illegal_format;
      for (i=0,k=0; i<gNrStockProperties; i++)
      {
        if (fecPrpsOnly && !gStockProperty[i].isFecProperty) continue;
        if (eqmPrpsOnly && gStockProperty[i].isFecProperty) continue;
        if (tgtprplen && strncmp(gStockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        slen = (int)strlen(gStockProperty[i].prpName);
        if (slen > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
        if (k + slen + 1 > rs) break;
        strncpy(&dataOut[k],gStockProperty[i].prpName,PROPERTY_NAME_SIZE);
        k += slen; dataOut[k++] = 0;
      }
      break;
    case CF_USTRING:
      if (rs == 0) return illegal_read_write;
      for (i=0,k=0; i<rs && i<gNrStockProperties; i++)
      {
        if (fecPrpsOnly && !gStockProperty[i].isFecProperty) continue;
        if (eqmPrpsOnly && gStockProperty[i].isFecProperty) continue;
        if (tgtprplen && strncmp(gStockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strncmp(((DUSTRING *)dataOut)[k-1].str,gStockProperty[i].prpName,32)) continue;
        strncpy(((DUSTRING *)dataOut)[k].str,gStockProperty[i].prpName,32);
        strncpy(&((DUSTRING *)dataOut)[k].str[32],gStockProperty[i].prpDescription,48);
        ((DUSTRING *)dataOut)[k].ival = (UINT32)gStockProperty[i].prpSize;
        ((DUSTRING *)dataOut)[k].tm = (UINT32)gStockProperty[i].prpFormat;
        k++;
      }
      if (k < rs)  assignStkMetaData(cl,0,0,0,k);
      return 0;
    case CF_STRUCT:
      for (i=0,k=0; i<gNrStockProperties; i++)
      {
        if (fecPrpsOnly && !gStockProperty[i].isFecProperty) continue;
        if (eqmPrpsOnly && gStockProperty[i].isFecProperty) continue;
        if (tgtprplen && strncmp(gStockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        memset(&prpqs[k],0,sizeof(PrpQueryStruct));
        strncpy(prpqs[k].prpName,gStockProperty[i].prpName,PROPERTY_NAME_SIZE);
        strncpy(prpqs[k].prpDescription,gStockProperty[i].prpDescription,64);
        prpqs[k].prpSize = gStockProperty[i].prpSize;
        prpqs[k].prpSizeIn = gStockProperty[i].prpSizeIn;
        prpqs[k].prpFormat = BFMT(gStockProperty[i].prpFormat);
        prpqs[k].prpFormatIn = BFMT(gStockProperty[i].prpFormatIn);
        prpqs[k].prpAccess = (BYTE)gStockProperty[i].prpAccessMode;
        prpqs[k].prpArrayType = gStockProperty[i].prpArrayType;
        if (isTaggableFormat(prpqs[k].prpFormat)) strncpy(prpqs[k].prpTag,gStockProperty[i].prpStructTag,TAG_NAME_SIZE);
        if (isTaggableFormat(prpqs[k].prpFormatIn)) strncpy(prpqs[k].prpTagIn,gStockProperty[i].prpStructTag,TAG_NAME_SIZE);
        k++;
      }
      if (tgtprplen) for (i=0; i<k; i++) prpqs[i].prpNumOverloads = k;
      if (rs%PRPQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"PRPQSr4",TAG_NAME_SIZE))
      {
        n = (UINT16)rs/PRPQUERYSTRUCT_SIZE;
        if (k > n) k = n;
        PrepOutgoingPRPQS(prpqs,dataOut,k);
        if (k < n)  assignStkMetaData(cl,0,0,0,k*PRPQUERYSTRUCT_SIZE);
        return 0;
      }
#     ifndef SMALL_TINE_LIB
      if (rs%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",TAG_NAME_SIZE))
      {
        n = (UINT16)rs/XPROPERTYQUERYSTRUCT_SIZE;
        if (k > n) k = n;
        for (i=0; i<n; i++) prpqs2xpqs(i);
        PrepOutgoingXPQS(xpqs,dataOut,k);
        if (k < n) assignStkMetaData(cl,0,0,0,k*XPROPERTYQUERYSTRUCT_SIZE);
        return 0;
      }
      if (rs%PROPERTYQUERYSTRUCTEX_SIZE == 0)
      {
        n = (int)(rs/PROPERTYQUERYSTRUCTEX_SIZE);
        if (k > n) k = n;
        for (i=0; i<n; i++) prpqs2pqsx(i);
        PrepOutgoingPQSX(pqsx,dataOut,k);
        if (k < n) assignStkMetaData(cl,0,0,0,k*PROPERTYQUERYSTRUCTEX_SIZE);
        return 0;
      }
#     endif
      if (con->EqmSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
      {
        n = (int)(rs/PROPERTYQUERYSTRUCT_SIZE);
        if (k > n) k = n;
        for (i=0; i<n; i++) prpqs2pqs(i);
        PrepOutgoingPQS(pqs,dataOut,k);
        if (k < n) assignStkMetaData(cl,0,0,0,k*PROPERTYQUERYSTRUCT_SIZE);
        return 0;
      }
      return illegal_format;
    default:
      break;
  }
  return illegal_format;
}
int nstockpropsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return T_numQuery(con,dataOut,gNrStockProperties);
}
int historyRecordDepth(char *prpName,UINT16 *shortDepth,UINT16 *longDepth)
{
  int i;
  for (i=0; i<nHistoryRecords && strnicmp(prpName,hstTbl[i].c.EqmProperty,PROPERTY_NAME_SIZE); i++);
  if (i < nHistoryRecords) 
  {
    if (shortDepth != NULL) *shortDepth = (UINT16)hstTbl[i].depthShort;
    if (longDepth != NULL) *longDepth = (UINT16)hstTbl[i].depthLong;
    return hstTbl[i].depthShort;
  }
  return 0;
}
int gAppendProperyAliasNamesToQueries = TRUE;
int lclPrpQueryFcn(ExportListStruct *el,char *dev,PrpQueryStruct *prpqs,int num,int uniqueList,int useAliasName)
{
  ExportPropertyListStruct *prp;
  int idx,k,i,n,mark,m,devnr=-1;
  /* char *bp=NULL,*mp=NULL; */

  if (dev != NULL) devnr = GetValidDeviceNumber(el->EqmName,dev,NULL,-1);
  for (i=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && i<num; idx++)
  {
    mark = i; /* mark the start of this hash index */
    for (prp=el->EqmPrpLst[idx]; prp && i<num; prp=prp->next)
    {
      if (prp->isReserved) continue;
      if (el->EqmDevLst != NULL && devnr >=0 && el->EqmDevLst[devnr].prps != NULL) 
      { /* there's a device-specific property list */
        if (el->EqmDevLst == NULL) continue;
        for (k=0; k<el->EqmDevLst[devnr].prps->prpListSize; k++) 
          if (!strnicmp(el->EqmDevLst[devnr].prps->prpList[k].name,prp->prpName,PROPERTY_NAME_SIZE)) break;
        if (k == el->EqmDevLst[devnr].prps->prpListSize) continue; /* not found in list */
      }
      if (uniqueList && i > mark)
      {
        for (m=mark; m<i; m++) if (!strnicmp(prpqs[m].prpName,prp->prpName,PROPERTY_NAME_SIZE)) break;
        if (m < i) continue; /* duplicate (overloaded?) name */
      }
      memset(&prpqs[i],0,sizeof(PrpQueryStruct));
      strncpy(prpqs[i].prpName,prp->prpName,PROPERTY_NAME_SIZE);
      strncpy(prpqs[i].prpDescription,prp->prpDescription,PROPERTY_DESC_SIZE);
      prpqs[i].prpSize = prp->prpSize;
      prpqs[i].prpSizeIn = prp->prpSizeIn;
      prpqs[i].prpFormat = BFMT(prp->prpFormat);
      prpqs[i].prpFormatIn = BFMT(prp->prpFormatIn);
      strncpy(prpqs[i].prpTag,prp->prpFormatTag.name,TAG_NAME_SIZE);
      strncpy(prpqs[i].prpTagIn,prp->prpFormatTagIn.name,TAG_NAME_SIZE);
      prpqs[i].prpGraphType = prp->egu.graph;
      prpqs[i].prpMaxValue = (float)prp->egu.max;
      prpqs[i].prpMinValue = (float)prp->egu.min;
      if (prp->prd != NULL) memcpy(prpqs[i].prpRedirection,prp->prd->prpRedirection,STATUS_SIZE);
      strncpy(prpqs[i].prpUnits,prp->egu.units,16);
      prpqs[i].prpNumOverloads++;
      prpqs[i].prpAccess = (BYTE)prp->prpAccessMode;
      historyRecordDepth(prp->prpName,&prpqs[i].prpHistoryDepthShort,&prpqs[i].prpHistoryDepthLong);
      prpqs[i].prpArrayType = prp->prpArrayType;
      prpqs[i].numRows = prp->numRows;
      prpqs[i].rowSize = prp->rowSize;
      prpqs[i].rngMaxValue = prp->xegu.max;
      prpqs[i].rngMinValue = prp->xegu.min;
      strncpy(prpqs[i].rngUnits,prp->xegu.units,16);
      n = i++;
      if (gAppendProperyAliasNamesToQueries && !useAliasName)
      { /* add the alias names to the query if desired */
        if (prp->prpAlias[0] != 0  && i > 0)
        { /* yes, there is an alias registered */
          if (GetPrpId(el->EqmName,prp->prpAlias) < 0)
          { /* and the alias is not already some other registered property */
            prpqs[i] = prpqs[n];
            strncpy(prpqs[i].prpName,prp->prpAlias,PROPERTY_NAME_SIZE);
            i++;
          }
        }
      }
    }
  }
  return i;
}
#ifndef SMALL_TINE_LIB
int getSrvAddressParams(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  DTYPE d;
  char *tgtcon,*tgteqm,*tgtexp;
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int len = 0;
  touchStkMetaData(cl);
  if (isServiceRequest(con))
  {
    if (con->EqmSizeIn != 3) return illegal_read_write;
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
       len = fmtsizeof(LFMT(con->EqmFormatIn));
       break;
      default:
        return illegal_format;
    }
    tgtcon = dataIn; tgteqm = dataIn+len; tgtexp = dataIn+len*2;
    if (strlen(tgteqm) == 0 && strlen(tgtexp) == 0) return invalid_data;
    if (strlen(tgtcon) && strnicmp(tgtcon,gDeviceContext,MIN(CONTEXT_NAME_SIZE,len))) return invalid_data;
    for (el=ExportList; el != NULL; el=el->next)
    {
      if (strlen(tgteqm) && strncmp(tgteqm,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
      if (strlen(tgtexp) && strnicmp(tgtexp,el->EqmExportName,MIN(EXPORT_NAME_SIZE,len))) continue;
      break; /* this looks like me ! */
    }
    if (el == NULL) return invalid_data;
  }
  if (el == NULL) return non_existent_elem;
  d.dArrayLength = con->EqmSizeOut;
  d.dFormat = LFMT(con->EqmFormatOut);
  d.dTag[0] = 0;
  d.data.cptr = dataOut;
  return getSrvAddress(&d,el->EqmName,NULL);
}
#endif
/* TODO: this uses a 16-char IP address => not good for IPv6 ! */
int getLastAccessList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  WriteAccessInfo *wap;
  WriteAccessQuery waq, *waqp;
  int i, n, idx, slen = 0, isstruct=FALSE, rs = (int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      slen = fmtsizeof(con->EqmFormatOut);
      break;
    case CF_STRUCT:
      if (strncmp(con->strTagOut,"WRACCTBL",TAG_NAME_SIZE)) return invalid_structure_tag;
      isstruct = TRUE;
      break;
    default: 
      return illegal_format;
  }
  if (wrAccTblPtr < 0) 
  { /* no commands in the table */
    wrAccTblPtr = 0; 
    memset(&wrAccTbl[wrAccTblPtr],0,WRITEACCESS_STRUCT_SIZE);
  }
  if (!isstruct)
  { /* one of the string types ... */
    wap = &wrAccTbl[wrAccTblPtr];
    switch (rs)
    {
      case 5: strncpy(dataOut + slen*4,getDataTimeString(wap->timestamp,-1),slen);
      case 4: strncpy(dataOut + slen*3,wap->dev,slen);
      case 3: strncpy(dataOut + slen*2,wap->prp,slen);
      case 2: strncpy(dataOut + slen*1,wap->addr,ADDR_SIZE);
        if (slen > 16 && wap->port != 0) sprintf(dataOut+slen*1+strlen(wap->addr),":%d",wap->port);
      case 1: strncpy(dataOut,wap->user,USERNAME_SIZE);
        return 0;
      default:
        return dimension_error;
    }
  }
  idx = wrAccTblPtr; n = 0;
  for (i=0; i<rs && i<WRITEACCESS_TBL_SIZE; i++)
  {
    if (wrAccTbl[idx].timestamp == 0)
    { /* we're finished looking ... */
      assignStkMetaData(cl,0,0,0,n*WRITEACCESS_QUERY_SIZE);
      break;
    }
    if (strncmp(wrAccTbl[idx].eqm,con->EqmName,EQM_NAME_SHORTSIZE) == 0)
    { /* TODO:  accomodate the IPv6 address */
      memcpy(&waq,&wrAccTbl[idx],WRITEACCESS_QUERY_SIZE);
      memcpy(waq.user,wrAccTbl[idx].user,USERNAME_SIZE);
      memcpy(waq.addr,wrAccTbl[idx].addr,ADDR_SIZE);
      memcpy(waq.prp,wrAccTbl[idx].prp,PROPERTY_NAME_SIZE);
      memcpy(waq.dev,wrAccTbl[idx].dev,DEVICE_NAME_SIZE);
      waq.timestamp = wrAccTbl[idx].timestamp;
#     if !defined(INTEL) && !defined(NONALIGNED) && !defined(VMS_IEEE)
      DSWAP(waq.timestamp);
#     endif
      waqp = (WriteAccessQuery *)(dataOut + n*WRITEACCESS_QUERY_SIZE);
      memcpy(waqp,&waq,WRITEACCESS_QUERY_SIZE);
      n++;
    }
    idx = (idx-1+WRITEACCESS_TBL_SIZE)%WRITEACCESS_TBL_SIZE;
  }
  return 0;
}
int propStatsQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  NAME64I *n64 = (NAME64I *)tmpWorkArea;
  int cacheSize = tmpWorkAreaSize / sizeof(NAME64I);
  CONTRACT *con = &cl->contract;
  ExportListStruct *el = cl->exp;
  ExportPropertyListStruct *prp;
  int i, k, m, n = 0, idx, num, len;
  int devnr = -1, mark = -1, rs = (int)(cl->dataSizeOut);
  int cc = 0, tgtprplen = 0;
  char tgtprp[PROPERTY_NAME_SIZE + 4];

  if (HAS_PRPQRYFCN(el)) return not_implemented;  
  if (rs == 0) return illegal_read_write;
  if (dataOut == NULL) return argument_list_error;
  touchStkMetaData(cl);
  devnr = GetValidDeviceNumber(el->EqmName, con->EqmDeviceName, NULL, -1);
  num = el->EqmNumProperties;

  if ((len=(int)con->EqmSizeIn) > 0)
  {
    memset(tgtprp, 0, PROPERTY_NAME_SIZE + 4);
    if ((len = getStringFormatLength(con->EqmFormatIn, con->EqmSizeIn, 64)) < 0) return -len;
    strncpy(tgtprp, dataIn, len);
    if (nAliasTableEntries > 0)
    { /* is the incoming name an alias ? */
      int idx = getAliasTableIndex(tgtprp);
      if (idx >= 0)
      {
        if (!IsMetaProperty(gAliasTable[idx].name, NULL, NULL, FALSE))
        {
          strncpy(tgtprp, gAliasTable[idx].name, PROPERTY_NAME_SIZE);
        }
      }
    }
    tgtprplen = (int)strlen(tgtprp);
  }
  for (i = 0, idx = 0; idx<PROPERTY_HASH_SIZE + 1 && i<num && i<cacheSize; idx++)
  {
    mark = i; /* mark the start of this hash index */
    for (prp = el->EqmPrpLst[idx]; prp && i<num && i<cacheSize; prp = prp->next)
    {
      if (prp->isReserved) continue;
      if (tgtprplen > 0 && strnicmp(tgtprp, prp->prpName, 64)) continue;
      strncpy(n64[i].name, prp->prpName, 64);
      if (el->EqmDevLst != NULL && devnr >= 0 && el->EqmDevLst[devnr].prps != NULL)
      { /* there's a device-specific property list */
        if (el->EqmDevLst == NULL) continue;
        for (k = 0; k<el->EqmDevLst[devnr].prps->prpListSize; k++)
          if (!strnicmp(el->EqmDevLst[devnr].prps->prpList[k].name, prp->prpName, PROPERTY_NAME_SIZE)) break;
        if (k == el->EqmDevLst[devnr].prps->prpListSize) continue; /* not found in list */
      }
      if (i > mark)
      {
        for (m = mark; m<i; m++) if (!strnicmp(n64[m].name, prp->prpName, PROPERTY_NAME_SIZE)) break;
        if (m < i)
        { /* duplicate (overloaded?) name */
          n64[m].ival += (int)prp->numCalls;
          continue; 
        }
      }
      n64[i].ival = (int)prp->numCalls;
      n = ++i;
    }
  }
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_INT32:
      for (i=0; i<n; i++) ((int *)dataOut)[i] = n64[i].ival;
      break;
    case CF_NAME64I:
      for (i = 0; i < n; i++)
      {
        ((NAME64I *)dataOut)[i] = n64[i];
      }
      break;
    default:
      return illegal_format;
  }
  assignStkMetaData(cl, 0, 0, 0, n);
  return cc;
}
int propsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  static BYTE *tmpDataCache = (BYTE *)NULL;
  static int tmpDataCacheSize = 0;
  int idx,j=0,k,n,num,len;
# ifndef SMALL_TINE_LIB
  int slen;
# endif
  short fmt;
  CONTRACT *con=&cl->contract;
  ExportListStruct *el;
  PropertyQueryStruct *pqs = (PropertyQueryStruct *)tmpDataCache;
  PropertyQueryStructEx *pqsx = (PropertyQueryStructEx *)tmpDataCache;
  XPropertyQueryStruct *xpqs = (XPropertyQueryStruct *)tmpDataCache;
  PrpQueryStruct *prpqs = (PrpQueryStruct *)tmpDataCache;
  int cc = 0,tgtprplen = 0,useAliasName=FALSE,usepqs=FALSE,usepqsx=FALSE,uselist=FALSE,hasdevgroup=FALSE;
  char tgtprp[PROPERTY_NAME_SIZE+4];
  char btok[PROPERTY_NAME_SIZE], etok[PROPERTY_NAME_SIZE], mtok[PROPERTY_NAME_SIZE];
  int blen = 0, elen = 0, mlen = 0, chkWildCard = FALSE, rs = (int)(cl->dataSizeOut);

  if (tmpDataCache == NULL)
  { /* on INTEL machines, tmpWorkArea points to the srvWorkArea */
    tmpDataCache = tmpWorkArea;
    tmpDataCacheSize = tmpWorkAreaSize;
  }
  if (rs == 0) return illegal_read_write;
  if (dataOut == NULL) return argument_list_error;
  touchStkMetaData(cl);
  if (strcmp(con->EqmProperty,"*") && strchr(con->EqmProperty,'*') != NULL)
  { /* property names query with non-trivial wildcard string */
    getWildcardTokens(con->EqmProperty,btok,etok,mtok);
    blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
    chkWildCard = TRUE;
  }
  if ((len=(int)con->EqmSizeIn) > 0)
  {
    memset(tgtprp,0,PROPERTY_NAME_SIZE+4);
    if ((len=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -len;
    strncpy(tgtprp,dataIn,len);
    if (nAliasTableEntries > 0) 
    { /* is the incoming name an alias ? */
      if ((j=getAliasTableIndex(tgtprp)) >= 0) 
      {
        if (!IsMetaProperty(gAliasTable[j].name,NULL,NULL,FALSE))
        {
          strncpy(tgtprp,gAliasTable[j].name,PROPERTY_NAME_SIZE);
          useAliasName = TRUE;
        }
      }
    }
    tgtprplen = (int)strlen(tgtprp);
  }
  fmt = LFMT(con->EqmFormatOut); /* requested output format */
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (strlen(el->EqmGroup) > 0) hasdevgroup = TRUE;
    len = tmpDataCacheSize/PRPQUERYSTRUCT_SIZE; /* worst case */
    if (el->EqmPropertyQueryFcn != NULL)           
    { /* registered query fcn */
      prpqs = NULL;
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (num < 0) return -num;
      if (prpqs == NULL) return illegal_property;
      if (num > 0 && tmpDataCache != (BYTE *)prpqs)
      { /* cache area not pointing to the query struct area */
        if (num > len)
        { /* not enough room -> re-allocate it */
          if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
            return out_of_server_memory;
          tmpDataCacheSize = num * sizeof(PrpQueryStruct);
          feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
        }
        memcpy(tmpDataCache,prpqs,num*sizeof(PrpQueryStruct));
        prpqs = (PrpQueryStruct *)tmpDataCache;
      }
      usepqs = TRUE;
    }
#   ifndef SMALL_TINE_LIB
    else if (el->XEqmPropertyQueryFcn != NULL)        
    { /* registered ext query fcn (legacy) */
      xpqs = NULL;
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (num < 0) return -num;
      if (xpqs == NULL) return illegal_property;
      if (num > 0 && tmpDataCache != (BYTE *)xpqs)
      { /* not in-situ */
        if (num > len)
        { /* not enough room -> re-allocate it */
          if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
            return out_of_server_memory;
          tmpDataCacheSize = num * sizeof(PrpQueryStruct);
          feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
        }
        prpqs = (PrpQueryStruct *)tmpDataCache;
      }
      for (k=0; k<num; k++) xpqs2prpqs(k);
      usepqsx = TRUE;
    }
#   endif
    else                                                      
    { /* handle it here */
      if ((num=(el->EqmNumProperties*2)) > len) 
      { /* allow room for alias entries (len = tmpDataCacheSize/PRPQUERYSTRUCT_SIZE) */
        if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
          return out_of_server_memory;
        tmpDataCacheSize = num * sizeof(PrpQueryStruct);
        feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
      }
      prpqs = (PrpQueryStruct *)tmpDataCache;
      num = lclPrpQueryFcn(el,con->EqmDeviceName,prpqs,num,isStringFormat(fmt),useAliasName);
      if (el->hasDevQry) uselist = TRUE;
    }
    if (num <= 0) return -num;
    /* re-assign these just in case they've changed */
    pqs = (PropertyQueryStruct *)tmpDataCache;
    pqsx = (PropertyQueryStructEx *)tmpDataCache;
    xpqs = (XPropertyQueryStruct *)tmpDataCache;
    prpqs = (PrpQueryStruct *)tmpDataCache;

    /* do we have to collapse the query results ? */
    if (chkWildCard)
    { /* checking a wildcard match */
      for (idx=0,k=0; idx<num && k<num; idx++)
      {
        if (!wildcardMatches(prpqs[idx].prpName,btok,blen,etok,elen,mtok,mlen))
          continue;
        if (idx != k) memcpy(&prpqs[k],&prpqs[idx],sizeof(PrpQueryStruct));
        k++;
      }
      num = k;
    }
    else if (tgtprplen > 0)
    { /* checking for input target property */
      for (idx=0,k=0; idx<num && k<num; idx++)
      {
        if (strnicmp(prpqs[idx].prpName,tgtprp,PROPERTY_NAME_SIZE)) continue;
        if (idx != k) memcpy(&prpqs[k],&prpqs[idx],sizeof(PrpQueryStruct));
        if (useAliasName) strncpy(prpqs[k].prpName,gAliasTable[j].alias,PROPERTY_NAME_SIZE);
        k++;
      }
      for (j=0; j<k; j++) prpqs[j].prpNumOverloads = k;
      num = k;
    }
    switch (fmt)
    {
      case CF_STRUCT:
        if (num == 0) return non_existent_property;
        if (rs%PRPQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"PRPQSr4",TAG_NAME_SIZE))
        {
          n = rs/PRPQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          PrepOutgoingPRPQS(prpqs,dataOut,num);
          if (num < n) assignStkMetaData(cl,0,0,0,num*PRPQUERYSTRUCT_SIZE);
          return 0;
        }
#       ifndef SMALL_TINE_LIB
        if (rs%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",TAG_NAME_SIZE))
        {
          n = rs/XPROPERTYQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          for (k=0; k<num; k++) prpqs2xpqs(k);
          PrepOutgoingXPQS(xpqs,dataOut,num);
          if (num < n) assignStkMetaData(cl,0,0,0,num*XPROPERTYQUERYSTRUCT_SIZE);
          return 0;
        }
        if (rs%PROPERTYQUERYSTRUCTEX_SIZE == 0)
        {
          n = rs/PROPERTYQUERYSTRUCTEX_SIZE;
          if (num > n) num = n;
          for (k=0; k<num; k++) prpqs2pqsx(k);
          PrepOutgoingPQSX(pqsx,dataOut,num);
          if (num < n) assignStkMetaData(cl,0,0,0,num*PROPERTYQUERYSTRUCTEX_SIZE);
          return 0;
        }
#       endif
        if (rs%PROPERTYQUERYSTRUCT_SIZE == 0)
        {
          n = rs/PROPERTYQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          if (usepqs && (chkWildCard || tgtprplen > 0))
          {
            pqs = (PropertyQueryStruct *)tmpDataCache;
            usepqs = FALSE;
          }
          if (!usepqs) for (k=0; k<num; k++) prpqs2pqs(k);
          PrepOutgoingPQS(pqs,dataOut,num);
          if (num < n) assignStkMetaData(cl,0,0,0,num*PROPERTYQUERYSTRUCT_SIZE);
          return 0;
        }
        return illegal_format;
      case CF_NAME64:
      case CF_NAME48:
      case CF_NAME32:
      case CF_NAME16:
        n = rs;
        if (num > n) num = n;
        len = fmtsizeof(fmt);
        for (idx=0,k=0; idx<num && k<n; idx++)
        { /* check for up to 2 overloads */
          if (k > 0 && !strnicmp(dataOut+(k-1)*len,prpqs[idx].prpName,len)) continue;
          if (k > 1 && !strnicmp(dataOut+(k-2)*len,prpqs[idx].prpName,len)) continue;
          strncpy(dataOut+k*len,prpqs[idx].prpName,len);
          k++;
        }
        if (k == 0) return non_existent_property;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        if (k < n) assignStkMetaData(cl,0,0,0,k);
        return cc;
#     ifndef SMALL_TINE_LIB
      case CF_STRING:
        if ((n=atoi(con->strTagOut)) < 1) return illegal_format;
        if (num > n) num = n;
        for (idx=0,k=0; idx<num; idx++)
        { 
          slen = (int)strlen(prpqs[idx].prpName);
          if (slen > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],prpqs[idx].prpName,PROPERTY_NAME_SIZE);
          k += slen; dataOut[k++] = 0;
        }
        if (k == 0) return non_existent_property;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        assignStkMetaData(cl,0,0,0,k);
        return cc;
      case CF_USTRING:
        n = rs;
        if (num > n) num = n;
        for (idx=0,k=0; idx<num && k<n; idx++)
        { /* check for up to 2 overloads */
          if (k > 0 && !strncmp(((DUSTRING *)dataOut)[k-1].str,prpqs[idx].prpName,32)) continue;
          if (k > 1 && !strncmp(((DUSTRING *)dataOut)[k-2].str,prpqs[idx].prpName,32)) continue;
          strncpy(((DUSTRING *)dataOut)[k].str,prpqs[idx].prpName,32);
          strncpy(&((DUSTRING *)dataOut)[k].str[32],prpqs[idx].prpDescription,48);
          ((DUSTRING *)dataOut)[k].ival = (UINT32)prpqs[idx].prpSize;
          ((DUSTRING *)dataOut)[k].tm = (UINT32)prpqs[idx].prpFormat;
          k++;
        }
        if (k == 0) return non_existent_property;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        if (k < n) assignStkMetaData(cl,0,0,0,k);
        return cc;
#     endif
      default:
        return illegal_format;
    }
  }
  return non_existent_elem;
}
int npropsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el;
  PrpQueryStruct *prpqs = NULL;
  int i, n=0, nx=0;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    n = el->EqmNumProperties;
    if ((i=GetDeviceNumber(con->EqmName,con->EqmDeviceName)) >= el->EqmNumDevices) i = -1;
    if (i >= 0)
    { /* specific device name given */
      if (el->EqmPropertyQueryFcn != NULL)           
      { /* registered query fcn */
        nx = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
        if (nx > 0) n = nx;
      }
      else if (el->EqmDevLst != NULL && el->EqmDevLst[i].prps != NULL)
      { /* registered device list */
        nx = el->EqmDevLst[i].prps->prpListSize;
        if (nx > 0) n = nx;
      }
    }
    return T_numQuery(con,dataOut,n);
  }
  return non_existent_elem;
}
#ifndef SMALL_TINE_LIB
#define pqs2ustr(p,u) {\
   strncpy(((DUSTRING *)(u))->str,(p)->prpName,32);\
   strncpy(&((DUSTRING *)(u))->str[32],(p)->prpDescription,48);\
   ((DUSTRING *)(u))->ival = (p)->prpSize;\
   ((DUSTRING *)(u))->tm = (p)->prpFormat;\
 }
TINE_EXPORT int gIgnoreMetaPropertyQueries = FALSE;
int metapropsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  static BYTE *tmpDataCache = (BYTE *)NULL;
  static int tmpDataCacheSize = 0;
  int idx,j=0,k,n,num,len,rs = (int)(cl->dataSizeOut);
  int slen;
  short fmt;
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  PrpQueryStruct mprpqs, *prpqs = (PrpQueryStruct *)tmpDataCache;
  int tgtprplen = 0,useAliasName=FALSE,uselist=FALSE;
  char tgtprp[PROPERTY_NAME_SIZE+4], lastprp[PROPERTY_NAME_SIZE+4];
  char *c, btok[PROPERTY_NAME_SIZE], etok[PROPERTY_NAME_SIZE], mtok[PROPERTY_NAME_SIZE];
  int blen = 0, elen = 0, mlen = 0, chkWildCard = FALSE;
  int addHst=TRUE,addDsc=TRUE,addOnl=TRUE,addEgu=TRUE,addMax=TRUE,addXEgu=TRUE,addXMax=TRUE;

  if (gIgnoreMetaPropertyQueries)
  {
    assignStkMetaData(cl,0,0,0,0);
    return 0;
  }
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  if (tmpDataCache == NULL)
  { /* on INTEL machines, tmpWorkArea points to the srvWorkArea */
    tmpDataCache = tmpWorkArea;
    tmpDataCacheSize = tmpWorkAreaSize;
  }
  if (rs == 0) return illegal_read_write;

  if (strcmp(con->EqmProperty,"*") && strchr(con->EqmProperty,'*') != NULL)
  { /* property names query with non-trivial wildcard string */
    getWildcardTokens(con->EqmProperty,btok,etok,mtok);
    blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
    chkWildCard = TRUE;
  }
  if (con->EqmSizeIn > 0)
  {
    memset(tgtprp,0,PROPERTY_NAME_SIZE+4);
    if ((len=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -len;
    strncpy(tgtprp,dataIn,len);
    if (nAliasTableEntries > 0)
    { /* is the incoming name an alias ? */
      if ((j=getAliasTableIndex(tgtprp)) >= 0) 
      {
        if (!IsMetaProperty(gAliasTable[j].name,NULL,NULL,FALSE))
        {
          strncpy(tgtprp,gAliasTable[j].name,PROPERTY_NAME_SIZE);
          useAliasName = TRUE;
        }
      }
    }
    if (tgtprp[0] == '.')
    { /* signal for list of desired meta property info */
      addHst = addDsc = addOnl = addEgu = addMax = addXEgu = addXMax = FALSE;
      for (c=strtok(tgtprp,","); c != NULL; c=strtok(NULL,","))
      {
        if (!strcmp(".HIST",c) || !strcmp(".HST",c)) addHst = TRUE; 
        else if (!strcmp(".DESC",c) || !strcmp(".DSC",c)) addDsc = TRUE; 
        else if (!strcmp(".EGU",c)) addEgu = TRUE; 
        else if (!strcmp(".XEGU",c)) addXEgu = TRUE; 
        else if (!strcmp(".ONLINE",c)) addOnl = TRUE; 
        else if (!strcmp(".MAX",c)) addMax = TRUE; 
        else if (!strcmp(".XMAX",c)) addXMax = TRUE; 
      }
    }
    else
    {
      tgtprplen = (int)strlen(tgtprp);
    }
  }
  fmt = LFMT(con->EqmFormatOut); /* requested output format */
  len = tmpDataCacheSize/PRPQUERYSTRUCT_SIZE; /* worst case */

  if ((num=(el->EqmNumProperties*2)) > len) 
  { /* allow room for alias entries (len = tmpDataCacheSize/PRPQUERYSTRUCT_SIZE) */
    if ((tmpDataCache=(BYTE *)calloc(num,sizeof(PrpQueryStruct))) == NULL)
      return out_of_server_memory;
    tmpDataCacheSize = num * sizeof(PrpQueryStruct);
    feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
  }
  prpqs = (PrpQueryStruct *)tmpDataCache;
  num = lclPrpQueryFcn(el,con->EqmDeviceName,prpqs,num,isStringFormat(fmt),useAliasName);
  if (el->hasDevQry) uselist = TRUE;
  if (num <= 0) return -num;
  /* re-assign these just in case they've changed */
  prpqs = (PrpQueryStruct *)tmpDataCache;

  /* do we have to collapse the query results ? */
  if (chkWildCard)
  { /* checking a wildcard match */
    for (idx=0,k=0; idx<num && k<num; idx++)
    {
      if (!wildcardMatches(prpqs[idx].prpName,btok,blen,etok,elen,mtok,mlen))
        continue;
      if (idx != k) memcpy(&prpqs[k],&prpqs[idx],sizeof(PrpQueryStruct));
      k++;
    }
    num = k;
  }
  else if (tgtprplen > 0)
  { /* checking for input target property */
    for (idx=0,k=0; idx<num && k<num; idx++)
    {
      if (strnicmp(prpqs[idx].prpName,tgtprp,PROPERTY_NAME_SIZE)) continue;
      if (idx != k) memcpy(&prpqs[k],&prpqs[idx],sizeof(PrpQueryStruct));
      if (useAliasName) strncpy(prpqs[k].prpName,gAliasTable[j].alias,PROPERTY_NAME_SIZE);
      k++;
    }
    for (j=0; j<k; j++) prpqs[j].prpNumOverloads = k;
    num = k;
  }
  lastprp[0] = 0;
  switch (fmt)
  {
    case CF_NAME64:
    case CF_NAME48:
    case CF_NAME32:
    case CF_NAME16:
    case CF_USTRING:
    case CF_STRING:
    case CF_STRUCT:
      if (num == 0) return non_existent_property;
      n = rs;
      len = fmtsizeof(fmt);
      if (fmt == CF_STRUCT)
      {
        if (strncmp(con->strTagOut,"PRPQSr4",TAG_NAME_SIZE)) return invalid_structure_tag;
        n /= PRPQUERYSTRUCT_SIZE;
        len = PRPQUERYSTRUCT_SIZE;
      }
      if (fmt == CF_STRING)
      {
        if (atoi(con->strTagOut) < 1) return illegal_format;
      }
      break;
    default:
      return illegal_format;
  }
  memset(&mprpqs,0,sizeof(PrpQueryStruct));
  for (idx=0,k=0; idx<num && k<n; idx++)
  {
    if (k > 0 && !strnicmp(lastprp,prpqs[idx].prpName,PROPERTY_NAME_SIZE)) continue;
    strncpy(lastprp,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
    mprpqs.numRows = 1;
    mprpqs.prpAccess = CA_READ;
    if (addDsc && strlen(prpqs[idx].prpDescription) > 0 && k<n)
    {
      strcpy(mprpqs.prpDescription,"description of the property");
      mprpqs.prpFormat = BFMT(CF_CHAR);
      mprpqs.prpFormatIn = BFMT(CF_NULL);
      mprpqs.rowSize = mprpqs.prpSize = PROPERTY_DESC_SIZE;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-5) strcat(mprpqs.prpName,".DESC");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
    if (addHst && prpqs[idx].prpHistoryDepthShort > 0 && k<n)
    { /* prpHistoryDepthShort is unsigned : HRS_REDIRECTION_FLAG will show up as positive here */
      strcpy(mprpqs.prpDescription,"history of the property");
      mprpqs.prpFormat = BFMT(CF_DBLDBL);
      mprpqs.prpFormatIn = BFMT(CF_INT32);
      mprpqs.rowSize = mprpqs.prpSize = 1000;
      mprpqs.prpSizeIn = 2;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-5) strcat(mprpqs.prpName,".HIST");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
    if (addOnl && (prpqs[idx].prpArrayType & AT_CHANNEL) && !uselist && k<n)
    {
      strcpy(mprpqs.prpDescription,prpqs[idx].prpDescription);
      mprpqs.prpFormat = prpqs[idx].prpFormat;
      mprpqs.prpFormatIn = prpqs[idx].prpFormatIn;
      mprpqs.rowSize = mprpqs.prpSize = prpqs[idx].prpSize;
      mprpqs.prpSizeIn = prpqs[idx].prpSizeIn;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-7) strcat(mprpqs.prpName,".ONLINE");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
    if (addEgu && strlen(prpqs[idx].prpUnits) > 0 && k<n)
    {
      strcpy(mprpqs.prpDescription,"engineering units of the property");
      mprpqs.prpFormat = BFMT(CF_USTRING);
      mprpqs.prpFormatIn = BFMT(CF_NULL);
      mprpqs.rowSize = mprpqs.prpSize = 1;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-4) strcat(mprpqs.prpName,".EGU");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
    if (addXEgu && strlen(prpqs[idx].rngUnits) > 0 && k<n)
    {
      strcpy(mprpqs.prpDescription,"x-axis engineering units of the property");
      mprpqs.prpFormat = BFMT(CF_USTRING);
      mprpqs.prpFormatIn = BFMT(CF_NULL);
      mprpqs.rowSize = mprpqs.prpSize = 1;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-5) strcat(mprpqs.prpName,".XEGU");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
    mprpqs.prpAccess = CA_READ|CA_WRITE;
    if (addMax && prpqs[idx].prpMaxValue > 0 && k<n)
    {
      strcpy(mprpqs.prpDescription,"maximum value of the property");
      mprpqs.prpFormat = BFMT(CF_FLOAT);
      mprpqs.prpFormatIn = BFMT(CF_FLOAT);
      mprpqs.rowSize = mprpqs.prpSize = 1;
      mprpqs.prpSizeIn = 1;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-4) strcat(mprpqs.prpName,".MAX");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
      if (k == n) break;
      strcpy(mprpqs.prpDescription,"minimum value of the property");
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-4) strcat(mprpqs.prpName,".MIN");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
    if (addXMax && prpqs[idx].rngMaxValue > 0 && k<n)
    {
      strcpy(mprpqs.prpDescription,"x-axis maximum value of the property");
      mprpqs.prpFormat = BFMT(CF_FLOAT);
      mprpqs.prpFormatIn = BFMT(CF_FLOAT);
      mprpqs.rowSize = mprpqs.prpSize = 1;
      mprpqs.prpSizeIn = 1;
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-5) strcat(mprpqs.prpName,".XMAX");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
      if (k == n) break;
      strcpy(mprpqs.prpDescription,"x-axis minimum value of the property");
      strncpy(mprpqs.prpName,prpqs[idx].prpName,PROPERTY_NAME_SIZE);
      if (strlen(mprpqs.prpName) < PROPERTY_NAME_SIZE-5) strcat(mprpqs.prpName,".XMIN");
      switch (fmt)
      {
        case CF_STRUCT: PrepOutgoingPRPQS(&mprpqs,&dataOut[k*len],1); break;
        case CF_USTRING: pqs2ustr(&mprpqs,&dataOut[k*len]);break;
        case CF_STRING:
          if ((slen=(int)strlen(mprpqs.prpName)) > PROPERTY_NAME_SIZE) slen = PROPERTY_NAME_SIZE;
          if (k + slen + 1 > rs) break;
          strncpy(&dataOut[k],mprpqs.prpName,PROPERTY_NAME_SIZE); k += slen; dataOut[k] = 0;
          break;
        default: strncpy(dataOut+k*len,mprpqs.prpName,len); break;
      }
      k++;
    }
  }
  if (k < n) 
  {
    if (fmt == CF_STRUCT) k *= PRPQUERYSTRUCT_SIZE;
    assignStkMetaData(cl,0,0,0,k);
  }
  return 0;
}
#endif /* SMALL_TINE_LIB */
int failoverStateQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int n = GetCurrentFailoverState();
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return T_numQuery(con,dataOut,n);
}
int alarmsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  AMS *ams;
  AMRB amrb;
  int n,nalm,minsev=0,devnr,isLegacyCall=FALSE,rs = (int)(cl->dataSizeOut);
  int sizeIn = (int)con->EqmSizeIn, sizeOut = rs;
  UINT32 startt, stopt;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  touchStkMetaData(cl);
  if (sizeIn > 0)
  { 
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_STRUCT:
        if (sizeIn != AMRB_SIZE) return dimension_error;
        prepIncomingAMRB(&amrb,(BYTE *)dataIn);
        stopt = amrb.almStopTime;
        startt = amrb.almStartTime;
        minsev = amrb.almMinSeverity;
        break;
      case CF_INT32:
      case CF_UINT32:
        if (sizeIn < 2) return dimension_error;
        startt = ((UINT32 *)dataIn)[0];
        stopt = ((UINT32 *)dataIn)[1];
        if (sizeIn > 2) minsev = ((UINT32 *)dataIn)[2];
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    stopt = 0x7fffffffL;
    startt = 0;
    minsev = 0;
  }
  if (strcmp(con->strTagOut,"AMSr4") != 0)
  { /* old-style request (could be untagged) */
    if (rs%AMS5_SIZE) return illegal_read_write;
    if (con->EqmSizeOut > tmpWorkAreaSize) sizeOut = (int)tmpWorkAreaSize;
    n = sizeOut/AMS5_SIZE;
    isLegacyCall = TRUE;
  }
  else
  {
    if (con->EqmSizeOut > tmpWorkAreaSize) sizeOut = (int)tmpWorkAreaSize;
    n = sizeOut/AMS_SIZE;
  }
  ams = (AMS *)tmpWorkArea;
  if (con->EqmDeviceName[0] == '#') 
  { /* explicit starting offset */
    devnr = -atoi(&con->EqmDeviceName[1]);
    if (devnr == 0) devnr = -1; /* all */
  }
  else if (!strcmp(con->EqmDeviceName,"*"))
  { /* signal all alarms (normal call from CAS) */
    devnr = -1; 
  }
  else
  { /* the targeted device */
    devnr = GetDeviceNumber(con->EqmName,con->EqmDeviceName);
    if (devnr < 0) return -devnr;
  }
  nalm = getAlarmsEx(con->EqmName,ams,(short)n,(short)minsev,(time_t)startt,(time_t)stopt,(short)devnr);
  if (nalm < 0) return -nalm;
  if (nalm > n) nalm = n; /* not even allowed by a called to getAlarmsEx ! */
  prepOutgoingAMS((BYTE *)dataOut,ams,con->strTagOut,nalm);  
  if (nalm < n) assignStkMetaData(cl,0,0,0,nalm * (isLegacyCall ? AMS5_SIZE : AMS_SIZE));
  return 0;
}
int nalarmsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  short sc,ntim,nsev;
  UINT32 ts;
  int devnr, nalm, fmt;
  CONTRACT *con=&cl->contract;

  UNUSED(dataIn);
  touchStkMetaData(cl);
  fmt = LFMT(con->EqmFormatOut);
  if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
  if (con->EqmDeviceName[0] == '#') 
  { /* number given : force to all */
    devnr = -1; /* all */
  }
  else if (!strcmp(con->EqmDeviceName,"*") || strlen(con->EqmDeviceName) == 0)
  { /* signal all alarms (normal call from CAS) */
    devnr = -1; 
  }
  else
  { /* the targeted device */
    devnr = GetDeviceNumber(con->EqmName,con->EqmDeviceName);
    if (devnr < 0) return -devnr;
  }
  nalm = numAlarmsEx(con->EqmName,devnr,&ts,&sc,&ntim,&nsev);
  switch (con->EqmSizeOut)
  {
    case 6:
      ((UINT32 *)dataOut)[5] = (UINT32)numAlarmDefs(con->EqmName);
    case 5:
      ((UINT32 *)dataOut)[4] = (UINT32)nsev;
    case 4:
      ((UINT32 *)dataOut)[3] = (UINT32)ntim;
    case 3:
      ((UINT32 *)dataOut)[2] = (UINT32)sc;
    case 2:
      ((UINT32 *)dataOut)[1] = ts;
    case 1:
      *(UINT32 *)dataOut = (UINT32)nalm;
      break;
    default: 
      return illegal_read_write;
  }
  if (!gRequireReadFromCAS) 
  { /* CAS has persistent link  to NALARMS ! */
    if ((gRequireReadFromCAS=isCASInUserList(con->EqmName)) != FALSE) 
      feclog("ALM: Central Alarm Server attached");
  }
  return 0;
}
int alarmFlush(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  UNUSED(dataIn);
  UNUSED(dataOut);
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  return RemoveAllAlarms(el->EqmName);
}
int alarmWindowQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  int cc,v=0,isWrite=FALSE;
  CONTRACT *con=&cl->contract;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    isWrite = TRUE;
    if ((cc=getInputInt32(con,dataIn,&v)) != 0) return cc;    
  }
  if (!stricmp(con->EqmProperty,"ALMCOLLAPSE"))
  {
    if (isWrite) if ((cc=SetAlarmCollapseWindow(v)) != 0) return cc;
    if (con->EqmSizeOut > 0) return T_numQuery(con,dataOut,gAlmCollapseWindow);
    return 0;
  }
  if (!stricmp(con->EqmProperty,"ALMOSCILLATION"))
  {
    if (isWrite) if ((cc=SetAlarmOscillationWindow(v)) != 0) return cc;
    if (con->EqmSizeOut > 0) return T_numQuery(con,dataOut,gAlmOscillationWindow);
    return 0;
  }
  return illegal_property;
}
#ifndef SMALL_TINE_LIB
int nalarmWatchQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return T_numQuery(con,dataOut,nAlmWatchList);
}
int alarmWatchQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  AWSLstEntry *lst = NULL;
  AWLQS aws;
  int cc,n,rs = (int)(cl->dataSizeOut);
  ExportListStruct *el;
  ExportPropertyListStruct *pl;
  char sts[STATUS_SIZE];

  touchStkMetaData(cl);
  if (con->EqmAccess&CA_WRITE)
  { /* write 1 alarm watch entry */
    if (LFMT(con->EqmFormatIn) != CF_STRUCT) return illegal_format;
    if ((cc=prepIncomingAWS((BYTE *)dataIn,&aws,con->strTagIn,1)) != 0) return cc;
    /* enforce the local eqm name : */
    strncpy(aws.eqm,con->EqmName,EQM_NAME_SIZE);
    for (lst=almWatchList; lst != NULL; lst=lst->nxt)
    { /* find the entry */
      if (strncmp(lst->aws.eqm,aws.eqm,EQM_NAME_SIZE)) continue;
      if (strnicmp(lst->aws.dev,aws.dev,DEVICE_NAME_SIZE)) continue;
      if (strnicmp(lst->aws.prp,aws.prp,PROPERTY_NAME_SIZE)) continue;
      /* it's in the current list, so enforce the following */
      aws.siz = lst->aws.siz;
      aws.fmt = lst->aws.fmt;
      aws.atyp = lst->aws.atyp;
      aws.code = lst->code;
      aws.codeHigh = lst->codeHigh;
      aws.codeLow = lst->codeLow;
      aws.codeLowWarn = lst->codeLowWarn;
      break;
    }
    if (lst == NULL)
    { /* was not in the current list */
      if ((cc=IsIllegalProperty(aws.eqm,aws.prp,aws.dev,NULL,CA_READ,&el,NULL,sts)) != 0) return cc;
      if ((lst=(AWSLstEntry *)SystemCalloc(1,sizeof(AWSLstEntry))) == NULL) return out_of_server_memory;
      if ((pl=GetPropertyListStruct(aws.eqm,aws.prp,NULL)) != NULL)
      { /* does not use a property query function (good news!) */
        if (aws.siz <= 0 || aws.siz > (int)pl->prpSize) aws.siz = pl->prpSize;
        aws.fmt = pl->prpFormat;
        aws.atyp = pl->prpArrayType;
      }
      /* add it to the list */
      memcpy(&lst->aws,&aws,AWLQS_CPSIZE);
      lst->nxt = almWatchList;
      almWatchList = lst;
      nAlmWatchList++;
      dumpAlmManifest(ALM_MF_WAT,TRUE);
    }
    else
    { /* apply the changes */
      memcpy(&lst->aws,&aws,AWLQS_CPSIZE);
    }
  }
  if (rs > 0)
  {
    int aws_size = strcmp(con->strTagOut, "AWSr4") == 0 ? AWS_SIZE : AWLQS_SIZE;
    if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
    n = rs/aws_size;
    if (nAlmWatchList > 0)
    {
      if ((cc=prepOutgoingAWS((BYTE *)dataOut,almWatchList,con->strTagOut,n)) != 0) return cc;
    }
    if (nAlmWatchList < n) assignStkMetaData(cl,0,0,0,nAlmWatchList*aws_size);
  }
  return 0;
}
int localAnnotatedCMTQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con=&cl->contract;
  FILE *file;
  time_t utcTime = 0;
  int n=0,rs = (int)(cl->dataSizeOut);
  char annotationDir[132];
  char fileName[132];
  char fileTime[16];
  char outputText[512];

  memset(annotationDir, 0, 128);
  sprintf(annotationDir, "%.115s%.12s%c", GetHistoryFilesRepository(), "ANNOTATIONS", FS_DELIMITER);

  if (strpbrk(con->EqmDeviceName, ":")) utcTime = (time_t)GetTimeStampFromString(con->EqmDeviceName);
  else utcTime = atol(con->EqmDeviceName);

  if (utcTime <= 0) return illegal_device;

  touchStkMetaData(cl);
  sprintf(fileTime, "%d.cmt", (int)utcTime);
  memset(fileName, 0, 128);
  sprintf(fileName, "%.112s%.16s", annotationDir, fileTime);

  if (con->EqmAccess&CA_WRITE)
  {
    if (LFMT(con->EqmFormatIn) != CF_TEXT) return illegal_format;
    makeDirectoryPath(annotationDir, TRUE);
    file = fopen(fileName, "r");
    if (file)
    {
      fclose(file);
      return file_already_exists;
    }
    file = fopen(fileName, "w");
    if (file == NULL) return file_error;
    fprintf(file, "%s", dataIn);
    fclose(file);
  }
  if (rs > 0)
  { /* only check if return output is requested ! */
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    file = fopen(fileName, "r");
    if (file)
    {
      csvGetLine(outputText, 512, (FILE*)file);
      fclose(file);
      strncpy(dataOut, outputText, rs);
      n = (int)strlen(outputText);
    }
    else
    {
      return illegal_device;
    }
    if (n < rs) assignStkMetaData(cl,0,0,0,n);
  }
  return 0;
}
int localAnnotatedCMTSQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con = &cl->contract;
  FILE *file;
  NAME64 fileList[100];
  char annotationDir[132];
  char filePath[192];
  char fileContent[132];
  char outPutData[132];
  char tsstr[64];
  char *tmp_ptr;
  int n = 0, k = 0, i;
  int slen = 0;
  int rs = (int)(cl->dataSizeOut);
  int fileCount = 0;
  int inRange = 0;
  SINT32 fromTime = 0;
  SINT32 toTime = 0;
  SINT32 tmp = 0;
  SINT32 fNlong = 0;

  switch (LFMT(con->EqmFormatIn))
  {
    case CF_INT32:
      if (con->EqmSizeIn < 2) return illegal_data_size;
      fromTime = ((SINT32 *)dataIn)[0];
      toTime = ((SINT32 *)dataIn)[1];
      break;
    case CF_DOUBLE:
      if (con->EqmSizeIn < 2) return illegal_data_size;
      fromTime = (SINT32)((double *)dataIn)[0];
      toTime = (SINT32)((double *)dataIn)[1];
      break;
    case CF_NULL: /* no input: last 24 hours ... */
      toTime = (SINT32)time(NULL);
      fromTime = toTime - 86400;
      break;
    default:
      return illegal_format;
  }
  if (fromTime > toTime)
  {
    tmp = fromTime;
    fromTime = toTime;
    toTime = tmp;
  }

  touchStkMetaData(cl);
  memset(annotationDir, 0, 128);
  sprintf(annotationDir, "%.115s%.12s%c", GetHistoryFilesRepository(), "ANNOTATIONS", FS_DELIMITER);

  fileCount = getFilesOfType(annotationDir, "*.cmt", fileList, 100);
  
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME64:
      for (i = 0; i < fileCount; i++)
      {
        fNlong = strtol(fileList[i].name, &tmp_ptr, 10);
        if (fNlong >= fromTime && fNlong <= toTime)
        {
          strncpy(((NAME64 *)dataOut)[inRange].name, GetDataTimeStringEx(fNlong,FALSE,tsstr), 64);
          inRange++;
        }
      }
      break;
    case CF_DOUBLE:
      for (i = 0; i < fileCount; i++)
      {
        fNlong = strtol(fileList[i].name, &tmp_ptr, 10);
        if (fNlong >= fromTime && fNlong <= toTime)
        {
          ((double *)dataOut)[inRange] = (double)strtol(fileList[i].name, &tmp_ptr, 10);
          inRange++;
        }
      }
      break;
    case CF_INT32:
      for (i = 0; i < fileCount; i++)
      {
        fNlong = strtol(fileList[i].name, &tmp_ptr, 10);
        if (fNlong >= fromTime && fNlong <= toTime)
        {
          ((SINT32 *)dataOut)[inRange] = strtol(fileList[i].name, &tmp_ptr, 10);
          inRange++;
        }
      }
      break;
    case CF_STRING:
      if ((n = atoi(con->strTagOut)) < 1) return illegal_format;

      for (i = 0; i < fileCount; i++)
      {
        fNlong = strtol(fileList[i].name, &tmp_ptr, 10);
        if (fNlong >= fromTime && fNlong <= toTime)
        {
          sprintf(filePath, "%.128s%.64s", annotationDir, fileList[i].name);
          file = fopen(filePath, "r");
          if (file)
          {
            csvGetLine(fileContent, 128, (FILE*)file);
            fclose(file);
            memset(outPutData, 0, 132);
            snprintf(outPutData, 128, "%d : %s", fNlong, fileContent);
            if (k + slen + 1 > rs) break;
            strncpy(&dataOut[k], outPutData, 128);
            slen = (int)strlen(outPutData);
            k += slen;
            dataOut[k++] = 0;
          }
          else
          {
            return file_error;
          }
        }
      }
      inRange = k;
      break;
    default:
      return illegal_format;
  }
  if (inRange < rs || n > 0) assignStkMetaData(cl,0,0,0,inRange);
  return 0;
}
#endif
int alarmDefsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{ 
  CONTRACT *con=&cl->contract;
  ADSList *adslst = NULL, *lst;
  ADS adsIn;
  int n, cc, isLegacyCall=FALSE,rs = (int)(cl->dataSizeOut);
  short num = 0;
  if ((num=getAlarmDefs(con->EqmName,&adslst)) < 0) return -num;
  if (num == 0) return un_allocated;
  touchStkMetaData(cl);
  if (con->EqmAccess&CA_WRITE)
  { /* someone want's to change something (e.g. severity) */
    if (LFMT(con->EqmFormatIn) != CF_STRUCT) return illegal_format;
    if (strcmp(con->strTagIn,"ADSr4") != 0) return invalid_structure_tag;
    n = (int)con->EqmSizeIn/ADS_SIZE;
    if (n != 1) return dimension_error;
    if ((cc=prepIncomingADS((BYTE *)dataIn,&adsIn)) != 0) return cc;
    for (lst=adslst; lst != NULL; lst=lst->next)
    { /* try to match the input alarm definition */
      if (lst->ads.alarmCode == adsIn.alarmCode)
      { /* got it ! */
        memcpy(&lst->ads,&adsIn,sizeof(ADS));
        break;
      }
    }
    if (lst == NULL) return un_allocated;
  }
  if (rs == 0) return 0;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (strcmp(con->strTagOut,"ADSr4") != 0)
  { /* old-style request (could be untagged) */
    if (rs%ADS5_SIZE) return illegal_read_write;
    n = rs/ADS5_SIZE;
    isLegacyCall = TRUE;
  }
  else
  {
    n = rs/ADS_SIZE;
  }
  prepOutgoingADS((BYTE *)dataOut,adslst,con->strTagOut,n);

  if (num < n) assignStkMetaData(cl,0,0,0,num * (isLegacyCall ? ADS5_SIZE : ADS_SIZE));

  return 0;
}
int nalarmDefsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return T_numQuery(con,dataOut,numAlarmDefs(con->EqmName));
}
int msgDisplay(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int len = (int)con->EqmSizeIn;
  char *un;
  static int demiseNoted = FALSE;
  UNUSED(dataOut);
  if (len == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatIn) != CF_TEXT) return illegal_format;
  touchStkMetaData(cl);
  memset(dbgbuf,0,256);
  strncpy(dbgbuf,dataIn,MIN(len,255));
  _sendToDbg = 1;
  feclog(dbgbuf);
  msglog(0,dbgbuf);
  un = GetCaller(con->EqmName);
  if (strstr(dbgbuf," in use by ") != NULL || strstr(dbgbuf," can only be used by ") != NULL)
  {
    if (gDieOnAddrInUse && !strcmp(un,"ENS"))
    {
      if (!demiseNoted)
      {
        demiseNoted = TRUE;
        if (dieFunction != NULL) dieFunction(dbgbuf);
        if (gDieAnotherDay) { gDieAnotherDay = FALSE; return 0; }
        feclogEx(TINE_LOGLEVEL_FATAL, "stopping server process (as per configuration: SetDieOnAddressInUse(FALSE); to prevent this)");
      }
      gHaltOnCycleHalt = TRUE;
      gSystemExitCode = 1;
      quitServer();
    }
  }
  else if (strstr(dbgbuf,"FEC has an alias ") != NULL)
  {
    if (gDieOnFecIsAlias && !strcmp(un,"ENS"))
    {
      if (!demiseNoted)
      {
        demiseNoted = TRUE;
        if (dieFunction != NULL) dieFunction(dbgbuf);
        if (gDieAnotherDay) { gDieAnotherDay = FALSE; return 0; }
        feclog("stopping server process (as per configuration: SetDieOnFecIsAlias(TRUE); has been configured)");
      }
      gHaltOnCycleHalt = TRUE;
      gSystemExitCode = 1;
      quitServer();
    }
  }
  _sendToDbg = 0;
  return 0;
}
#ifndef SMALL_TINE_LIB
int almFilterOuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con = &cl->contract;
  ADSList *adslst = NULL, *lst;
  FilterStruct flstIn;
  int n, cc, rs = (int)(cl->dataSizeOut);
  short num = 0;
  if ((num = getAlarmDefs(con->EqmName, &adslst)) < 0) return -num;
  if (num == 0) return un_allocated;
  touchStkMetaData(cl);
  if (con->EqmAccess&CA_WRITE)
  {
    if (LFMT(con->EqmFormatIn) != CF_STRUCT) return illegal_format;
    if (strcmp(con->strTagIn, "Filt") != 0) return invalid_structure_tag;
    n = (int)con->EqmSizeIn / FILTERSTRUCT_SIZE;
    if (n != 1) return dimension_error;
    if ((cc = prepIncomingAlmFilter((BYTE *)dataIn, &flstIn)) != 0) return cc;
    for (lst = adslst; lst != NULL; lst = lst->next)
    {
      if (lst->ads.alarmCode == flstIn.filterIdx) 
      {
        char oldFilter[256], newFilter[256];
        getAlarmFilterString(lst->fltr, oldFilter);
        strncpy(newFilter, flstIn.filterLinkStr, 256);
        if (strcmp(oldFilter, "NONE") == 0)
        {
          lst->fltr = getFilterLinkFromParsableString(newFilter);
        }
        else if (strcmp(oldFilter, newFilter) != 0)
        {
          rmvFilterLink(lst->fltr);
          lst->fltr = getFilterLinkFromParsableString(newFilter);
        }
        break;
      }
    }
    if (lst == NULL) return un_allocated;
    dumpAlmManifest(ALM_MF_ALL, TRUE);
  }
  if (rs == 0) return 0;
  if (rs > num*FILTERSTRUCT_SIZE) rs = num*FILTERSTRUCT_SIZE;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (strcmp(con->strTagOut, "Filt") != 0) return invalid_structure_tag;
  n = rs / FILTERSTRUCT_SIZE;
  prepOutgoingAlmFilter((BYTE *)dataOut,adslst,n);
  assignStkMetaData(cl, 0, 0, 0, n * FILTERSTRUCT_SIZE);

  return 0;
}
int hstFilterQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con = &cl->contract;
  FilterStruct flstIn;
  int i, n, cc, rs = (int)(cl->dataSizeOut);
  short num = 0;
  if ((num = nHistoryRecords) < 0) return -num;
  if (num == 0) return un_allocated;
  touchStkMetaData(cl);
  if (con->EqmAccess&CA_WRITE)
  {
    if (LFMT(con->EqmFormatIn) != CF_STRUCT) return illegal_format;
    if (strcmp(con->strTagOut, "Filt") != 0) return invalid_structure_tag;
    n = (int)con->EqmSizeIn / FILTERSTRUCT_SIZE;
    if (n != 1) return dimension_error;
    if ((cc = prepIncomingHstFilter((BYTE *)dataIn, &flstIn)) != 0) return cc;
    for (i = 0; i < num; i++)
    {
      if (i == (int)flstIn.filterIdx)
      {
        char oldFilter[256], newFilter[256];
        getHistoryFilterString(hstTbl[i].fltr, oldFilter);
        strncpy(newFilter, flstIn.filterLinkStr, 256);
        if (strcmp(oldFilter, "NONE") == 0)
        {
          hstTbl[i].fltr = getFilterLinkFromParsableString(newFilter);
        }
        else if (strcmp(oldFilter, newFilter) != 0)
        {
          rmvFilterLink(hstTbl[i].fltr);
          hstTbl[i].fltr = getFilterLinkFromParsableString(newFilter);
        }
        break;
      }
    }
    dumpHistoryManifest(hstTbl[i].c.EqmName,TRUE);
  }
  if (rs == 0) return 0;
  if (rs > num*FILTERSTRUCT_SIZE) rs = num*FILTERSTRUCT_SIZE;
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (strcmp(con->strTagOut, "Filt") != 0) return invalid_structure_tag;
  n = rs / FILTERSTRUCT_SIZE;
  prepOutgoingHstFilter((BYTE *)dataOut,n);
  assignStkMetaData(cl, 0, 0, 0, n*FILTERSTRUCT_SIZE);

  return 0;
}
#endif /* SMALL_TINE_LIB */
int activityQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ActivityQueryStruct aqs;

  UNUSED(dataIn);
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqmSizeOut != ACTIVITYQUERYSTRUCT_SIZE) return illegal_read_write;
  touchStkMetaData(cl);
  strncpy(aqs.FecName,gFecName,FEC_NAME_SIZE);
  aqs.numCalls = GetNumCalls(con->EqmName);
  aqs.localtime = (UINT32)time(NULL);
  aqs.starttime = (UINT32)gSystemStartupTime;
  aqs.systemPollingRate = gSystemCycleDeadband;
  aqs.numBkgTsks = gNumEqmBkgTasks;
  aqs.numTotalContracts = ncontracts - nExpiredContracts;
  aqs.numTotalClients = nconsumers;
  aqs.numTargetContracts = GetNumContracts(con->EqmName);
  aqs.numTargetClients = GetNumConsumers(con->EqmName);
  aqs.numConnections = nConnectionTableEntries;
  aqs.numConnectionTimeouts = nConnectionTimeouts;
  aqs.numConnectionArrivals = nArrivals;
  aqs.numUDPpkts = UDPcount;
  aqs.numTCPpkts = TCPcount;
  aqs.numIPXpkts = IPXcount;
  aqs.numSPXpkts = SPXcount;

  PrepOutgoingAQS(&aqs,dataOut);
  return 0;
}
int contractsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ContractQueryStruct *cqs = NULL;
  Contract5QueryStruct *cqs5 = NULL;
  ContractListStruct *cli;
  int i, n, ncon=0,rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (!strcmp(con->strTagOut,"CTQSr4") || rs%CONTRACT_QUERY_SIZE == 0)
  {
    if ((ncon=rs/CONTRACT_QUERY_SIZE) == 0) return dimension_error;
    cqs = (ContractQueryStruct *)dataOut;
  }
  if (!strcmp(con->strTagOut,"CONQS") || rs%CONTRACT5_QUERY_SIZE == 0)
  {
    if ((ncon=rs/CONTRACT5_QUERY_SIZE) == 0) return dimension_error;
    cqs5 = (Contract5QueryStruct *)dataOut;
  }
  for (i=0,n=0; n<ncon && i<ncontracts; i++)
  {
    cli = ContractList[i];
    if (cli->expired) continue;
    if (cqs != NULL)
    {
      memcpy(&cqs[n].contract,&cli->contract,sizeof(CONTRACT));
      cqs[n].pollingInterval = cli->pollingInterval;
      cqs[n].nclients = cli->nclients;
      PrepOutgoingCONQS(cqs[n]);
    }
    else if (cqs5 != NULL)
    {
      memcpy(cqs5[n].contract.EqmProperty,cli->contract.EqmProperty,PROPERTY_NAME_SHORTSIZE);
      memcpy(cqs5[n].contract.EqmDeviceName,cli->contract.EqmDeviceName,DEVICE_NAME_SHORTSIZE);
      memcpy(cqs5[n].contract.EqmName,cli->contract.EqmName,EQM_NAME_SIZE);
      cqs5[n].contract.EqmSizeIn = cli->contract.EqmSizeIn;
      cqs5[n].contract.EqmSizeOut = cli->contract.EqmSizeOut;
      memcpy(&cqs5[n].contract.hEqmName,&cli->contract.hEqmName,4);
      memcpy(cqs5[n].contract.strTagIn,cli->contract.strTagIn,TAG_NAME_SHORTSIZE);
      memcpy(cqs5[n].contract.strTagOut,cli->contract.strTagOut,TAG_NAME_SHORTSIZE);
      cqs5[n].pollingInterval = cli->pollingInterval;
      cqs5[n].nclients = cli->nclients;
      PrepOutgoingCONQS(cqs5[n]);
    }
    n++;
  }
  assignStkMetaData(cl,0,0,0,n * (cqs == NULL ? CONTRACT5_QUERY_SIZE : CONTRACT_QUERY_SIZE));
  return 0;
}
int contblQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ConTblInfo *contbl;
  int n,rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (rs%CONTBL_QUERY_SIZE) return illegal_read_write;
  if ((n=rs/CONTBL_QUERY_SIZE) == 0) return dimension_error;
  if (con->EqmSizeOut > tmpWorkAreaSize) n = tmpWorkAreaSize/CONTBL_QUERY_SIZE;
  contbl = (ConTblInfo *)tmpWorkArea;
  GetConnectionTable(contbl,&n);
  PrepOutgoingConTblInfo(contbl,dataOut,n);
  assignStkMetaData(cl,0,0,0,n*CONTBL_QUERY_SIZE);
  return 0;
}
int rmvRegisteredUser(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqmSizeIn, lstsiz = 0;
  UNUSED(dataOut);
  touchStkMetaData(cl);
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqmAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,USERNAME_SIZE);
      break;
    case CF_TEXT:
      memset(usr.name,0,USERNAME_SIZE);
      if (len < 1) return illegal_read_write;
      if (len > USERNAME_SIZE) len = USERNAME_SIZE;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    default: return illegal_format;
  }
  return rmvRegUserLst(con->EqmName,usrlst,lstsiz);
}
int addRegisteredUser(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqmSizeIn, lstsiz = 0;
  UNUSED(dataOut);
  if (el == NULL) return non_existent_elem;
  /* If there's no list now, don't let someone make one over the net (dangerous!) */
  if (el->aclLst.nUsrs == 0) return not_allowed; 
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqmAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  touchStkMetaData(cl);
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,USERNAME_SIZE);
      break;
    case CF_TEXT:
      memset(usr.name,0,USERNAME_SIZE);
      if (len < 1) return illegal_read_write;
      if (len > USERNAME_SIZE) len = USERNAME_SIZE;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    case CF_NULL:
      /* force a re-read of users.csv */
      break;
    default: 
      return illegal_format;
  }
  return AppendRegisteredUsers(con->EqmName,usrlst,lstsiz);
}
#ifndef SMALL_TINE_LIB
int prepIPNetLst(CONTRACT *con,char *dataIn,int *siz,NAME64 **lst)
{
  NAME64 *iplst = NULL;
  char *c;
  int i, lstsiz = 0, nlen, len = (int)con->EqmSizeIn;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  switch (LFMT(con->EqmFormatIn))
  {
    case CF_NAME16:
    case CF_NAME32:
      iplst = (NAME64 *)tmpWorkArea;
      if (len > (int)tmpWorkAreaSize/64) len = tmpWorkAreaSize/64;
      lstsiz = len;
      nlen = fmtsizeof(LFMT(con->EqmFormatIn));
      for (i=0; i<len; i++) 
      {
        strncpy(iplst[i].name,dataIn+i*nlen,nlen);
        iplst[i].name[nlen] = 0;
      }
      break;
    case CF_NAME64:
      iplst = (NAME64 *)dataIn;
      lstsiz = len;
      break;
    case CF_TEXT:
      memset(tmpWorkArea,0,64);
      iplst = (NAME64 *)tmpWorkArea;
      if (len < 1) return illegal_read_write;
      if (len > 64) len = 64;
      strncpy(iplst->name,dataIn,len);
      if ((c=strchr(iplst->name,'\n')) != NULL) *c = 0;
      if ((c=strchr(iplst->name,'\t')) != NULL) *c = 0;
      if ((c=strchr(iplst->name,' ')) != NULL) *c = 0;
      lstsiz = 1;
      break;
    case CF_NULL:
      /* force a file re-read */
      break;
    default: 
      return illegal_format;
  }
  *lst = iplst;
  *siz = lstsiz;
  return 0;
}
int addRegisteredIPNet(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  NAME64 *iplst = NULL;
  int cc,lstsiz = 0;
  UNUSED(dataOut);
  if (el == NULL) return non_existent_elem;
  /* If there's no list now, don't let someone make one over the net (dangerous!) */
  if (el->aclLst.nNets == 0) return not_allowed; 
  touchStkMetaData(cl);
  if ((cc=prepIPNetLst(con,dataIn,&lstsiz,&iplst)) != 0) return cc;
  return addRegNetsLst(el->netsFileName,NETS_ACCESS,&el->aclLst.nNets,&el->aclLst.nets,iplst,lstsiz);
}
int rmvRegisteredIPNet(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  NAME64 *iplst = NULL;
  int cc,lstsiz = 0;
  UNUSED(dataOut);
  if (el == NULL) return non_existent_elem;
  /* If there's no list there's nothing to remove */
  if (el->aclLst.nNets == 0) return not_allowed; 
  if ((cc=prepIPNetLst(con,dataIn,&lstsiz,&iplst)) != 0) return cc;
  touchStkMetaData(cl);
  return rmvRegNetsLst(el->netsFileName,NETS_ACCESS,&el->aclLst.nNets,&el->aclLst.nets,iplst,lstsiz);
}
int getNumRegisteredIPNets(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int n;
  int slen,devnr;
  char str[72];
  ExportPropertyListStruct *prp;
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  n = el->aclLst.nNets;
  if (con->EqmSizeIn > 0)
  {
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -slen;
    strncpy(str,dataIn,slen);
    if ((prp=GetPropertyListStruct(con->EqmName,str,NULL)) != NULL)
    { /* input is a property */
      n = prp->aclLst.nNets;
    }
    else if ((devnr=GetDeviceNumber(con->EqmName,str)) >= 0 && 
             (devnr < el->EqmNumDevices) && el->EqmDevLst[devnr].aclLst != NULL)
    { /* input is a device */
      if (el->EqmDevLst != NULL) n = el->EqmDevLst[devnr].aclLst->nNets;
    }
  }
  return T_numQuery(con,dataOut,n);
}
int getRegisteredIPNetList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int i,n,slen,fmt=LFMT(con->EqmFormatOut),rs = (int)(cl->dataSizeOut);
  char ipstr[64];
  SCKACL *lst;
  int devnr;
  char str[72];
  ExportPropertyListStruct *prp;
  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  if (!isSimpleStringFormat((short)fmt)) return illegal_format;
  touchStkMetaData(cl);
  n = el->aclLst.nNets;
  lst = el->aclLst.nets;
  if (con->EqmSizeIn > 0)
  {
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -slen;
    strncpy(str,dataIn,slen);
    if ((prp=GetPropertyListStruct(con->EqmName,str,NULL)) != NULL)
    { /* input is a property */
      n = prp->aclLst.nNets;
      lst = prp->aclLst.nets;
    }
    else if ((devnr=GetDeviceNumber(con->EqmName,str)) >= 0 && 
             (devnr < el->EqmNumDevices) && el->EqmDevLst[devnr].aclLst != NULL)
    { /* input is a device */
      if (el->EqmDevLst != NULL)
      {
        n = el->EqmDevLst[devnr].aclLst->nNets;
        lst = el->EqmDevLst[devnr].aclLst->nets;
      }
    }
  }
  slen = fmt == CF_TEXT ? rs : fmtsizeof(fmt);
  for (i=0; i<rs && i<n; i++)
  {
    inetStrAdr(&lst[i].addr,ipstr,FALSE);
    addIpSubStr(ipstr,lst[i]);
    strncpy(dataOut + i*slen,ipstr,slen);
  }
  assignStkMetaData(cl,0,0,0,i);
  return 0;
}
int getNumRegisteredIPXNets(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
# if defined(IPXNW) || defined(IPXBSD)
  return T_numQuery(con,dataOut,NrOfIPXNets);
# else
  UNUSED(dataOut);
  UNUSED(con);
  return ipx_not_supported;
# endif
}
int getRegisteredIPXNetList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
# if defined(IPXNW) || defined(IPXBSD)
  CONTRACT *con=&cl->contract;
  int i,j,rs = (int)(cl->dataSizeOut);
  char ipx[32];
  UNUSED(dataIn);
  if (rs == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  touchStkMetaData(cl);
  for (i=0; i<rs && i<NrOfIPXNets; i++)
  {
    for (j=0;j<4;j++) sprintf(&ipx[j*2],"%02X",IPXNetList[i].net[j]);
    strncpy(((NAME16 *)dataOut)[i].name,ipx,16);
  }
  return 0;
# else
  UNUSED(dataIn);
  UNUSED(dataOut);
  UNUSED(cl);
  return ipx_not_supported;
# endif
}
#endif /* !SMALL_TINE_LIB */
int getNumRegisteredUsers(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int n;
# ifndef SMALL_TINE_LIB
  char str[72];
  int slen, devnr;
  ExportPropertyListStruct *prp;
# else
  UNUSED(dataIn);
# endif
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  n = el->aclLst.nUsrs;
# ifndef SMALL_TINE_LIB
  if (con->EqmSizeIn > 0)
  {
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -slen;
    strncpy(str,dataIn,slen);
    if ((prp=GetPropertyListStruct(con->EqmName,str,NULL)) != NULL)
    { /* input is a property */
      n = prp->aclLst.nUsrs;
    }
    else if ((devnr=GetDeviceNumber(con->EqmName,str)) >= 0 && 
             (devnr < el->EqmNumDevices) && el->EqmDevLst[devnr].aclLst != NULL)
    { /* input is a device */
      if (el->EqmDevLst != NULL) n = el->EqmDevLst[devnr].aclLst->nUsrs;
    }
  }
# endif
  return T_numQuery(con,dataOut,n);
}
int getFecLog(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (rs > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    readFecLog(dataOut,rs);
  }
  return 0;
}
int isAllowedLogFileAccess(char *fn)
{
  int plen;
  if (gLogfileAllowScan == LOG_SCAN_FULL) return TRUE;
  if (gLogfileAllowScan == LOG_SCAN_NONE)
  {
    if (!stricmp(fn,"fec.log")) return TRUE;
    return FALSE;
  }
  if (!strncmp(fn,"./",2) || !isAbsolutePath(fn)) return TRUE;
  plen = (int)strlen(gFeclogPath);
  if (plen > 0 && !strncmp(gFeclogPath,fn,plen)) return TRUE;
  if (stricmp(gFeclogPath,fecDbPath))
  { /* FEC_LOG != FEC_HOME */
    plen = (int)strlen(fecDbPath);
    if (plen > 0 && !strncmp(fecDbPath,fn,plen)) return TRUE;
  }
  return FALSE;
}
int getFecLogFile(ContractListStruct *cl,char *dataIn,char *dataOut)
{ /* targets log (text) files */
  char fn[128];
  int slen, cc,rs = (int)(cl->dataSizeOut);
  CONTRACT *con=&cl->contract;
  double fts;
  touchStkMetaData(cl);
  if (con->EqmDeviceName[0] != '#') 
  {
    strncpy(fn,con->EqmDeviceName,DEVICE_NAME_SIZE); fn[DEVICE_NAME_SIZE] = 0;
  }
  if (con->EqmSizeIn > 0)
  {
    if (con->EqmAccess & CA_WRITE)
    { /* write the text file submitted to disk */
      if (con->EqmSizeIn == 0) return illegal_read_write;
      if (!isAllowedLogFileAccess(fn)) return not_allowed;
      if ((cc=writeFecLogFile(fn,dataIn,(int)con->EqmSizeIn)) != 0) return cc;
    }
    else
    { /* alternate way of supplying a (long) file name */
      if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,128)) < 0) return -slen;
      strncpy(fn,dataIn,slen); fn[slen] = 0;
    }
  }
  if (rs > 0)
  {
    int ts = 0;
    int siz = getFecLogFileLength(fn, &ts);
    if (siz < 0) return -siz;
    fts = (double)ts;
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_INT32:
      case CF_UINT32:
        rs = 1;
        *(int *)dataOut = siz;
        break;
      case CF_TEXT:
        if ((cc=readFecLogFile(fn, dataOut, rs)) != 0) return cc;
        rs = (int)strlen(dataOut);
        break;
      default:
        return illegal_format;
    }
    if (!isAllowedLogFileAccess(fn)) return not_allowed;
    assignStkMetaData(cl,fts,0,0,(UINT32)rs);
  }
  return 0;
}
int getFecFile(ContractListStruct *cl,char *dataIn,char *dataOut)
{ /* targets binary files */
  CONTRACT *con=&cl->contract;
  char fn[128];
  int slen, cc, sz,rs = (int)(cl->dataSizeOut);

  if (con->EqmDeviceName[0] == '#') return invalid_name;
  touchStkMetaData(cl);
  strncpy(fn,con->EqmDeviceName,DEVICE_NAME_SIZE); fn[DEVICE_NAME_SIZE] = 0;
  if (con->EqmSizeIn > 0)
  {
    if (con->EqmAccess & CA_WRITE)
    { /* write the text file submitted to disk */
      if (!isAllowedLogFileAccess(fn)) return not_allowed;
      if ((cc=writeFecFile(fn,dataIn,(int)con->EqmSizeIn)) != 0) return cc;
    }
    else
    { /* alternate way of supplying a (long) file name */
      if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,128)) < 0) return -slen;
      strncpy(fn,dataIn,slen); fn[slen] = 0;
    }
  }
  if (rs > 0)
  {
    if (fmtsizeof(con->EqmFormatOut) != 1) return illegal_format;
    sz = rs;
    if (!isAllowedLogFileAccess(fn)) return not_allowed;
    if ((cc=readFecFile(fn,dataOut,&sz)) != 0) return cc;
    assignStkMetaData(cl,0,0,0,sz);
  }
  return 0;
}
int getSelfTestFile(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int bytesread,rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (rs > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    bytesread = readSelfTest(con->EqmName,dataOut,rs);
    if (bytesread == 0) return no_such_file;
    assignStkMetaData(cl,0,0,0,bytesread);
  }
  return 0;
}
#ifndef SMALL_TINE_LIB
int getFecLogFileList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int cc, n, slen = 0, rs = (int)(cl->dataSizeOut);
  char path[256];
  touchStkMetaData(cl);
  if (rs > 0)
  {
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,128)) < 0) slen = 0;
    if (slen > 0) 
    {
      strncpy(path,dataIn,slen); 
    }
    else
    { /* alternate way of supplying search path */
      if (con->EqmDeviceName[0] != '#') 
      { /* but check against my registered device list */
        if (GetDeviceNumber(con->EqmName,con->EqmDeviceName) < 0)
        { /* okay, it's not a registered device name (e.g. from a simple query via the instant client) */
          strncpy(path,con->EqmDeviceName,DEVICE_NAME_SIZE); path[DEVICE_NAME_SIZE] = 0;
          slen = (int)strlen(path);
        }
      }
    }
    path[slen] = 0; 
    applyFecLogPath(path,128);
    if (!isAllowedLogFileAccess(path)) return not_allowed;
    n = rs;
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_NAME64:
        if ((cc=scanFecLogFiles(path,NULL,(NAME64 *)dataOut,CF_NAME64,&n,0)) != 0) return cc;
        break;
      case CF_NAME64I:
        if ((cc=scanFecLogFiles(path,NULL,(NAME64I *)dataOut,CF_NAME64I,&n,0)) != 0) return cc;
        break;
      default:
        return illegal_format;
    }
    assignStkMetaData(cl,0,0,0,n);
  }
  return 0;
}
int setLogDepth(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int rs = (int)(cl->dataSizeOut);
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0 && !nofeclog)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_INT16) return illegal_format;
    FeclogDepth = *(short *)dataIn;
  }
  if (rs > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_INT16) return illegal_format;
    *(short *)dataOut = (short)FeclogDepth;
  }
  return 0;
}
int setLogCommands(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int rs = (int)(cl->dataSizeOut);
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_INT16) return illegal_format;
    putCommandsInFeclog = *(short *)dataIn;
  }
  if (rs > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_INT16) return illegal_format;
    *(short *)dataOut = (short)putCommandsInFeclog;
  }
  return 0;
}
int setDebugLevel(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0)
  {
    if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqmFormatIn) != CF_INT16) return illegal_format;
    tineDebug = *(short *)dataIn;
  }
  if (con->EqmSizeOut > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_INT16) return illegal_format;
    *(short *)dataOut = tineDebug;
  }
  return 0;
}
int getAppVersion(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  DTYPE d;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  memset(&d,0,sizeof(DTYPE));
  d.dArrayLength = con->EqmSizeOut;
  d.dFormat = LFMT(con->EqmFormatOut);
  d.data.cptr = dataOut;
  return PutValuesFromString(&d,appmkversion,CF_TEXT,(int)strlen(appmkversion),0);
}
#endif /* !SMALL_TINE_LIB */
int getTINEVersion(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  DTYPE d;
  char *vs;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  memset(&d,0,sizeof(DTYPE));
  d.dArrayLength = con->EqmSizeOut;
  d.dFormat = LFMT(con->EqmFormatOut);
  d.data.cptr = dataOut;
  vs = GetSystemVersionString();
  return PutValuesFromString(&d,vs,CF_TEXT,(int)strlen(vs),0);
}
int getServerOS(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  DTYPE d;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  memset(&d,0,sizeof(DTYPE));
  d.dArrayLength = con->EqmSizeOut;
  d.dFormat = LFMT(con->EqmFormatOut);
  d.data.cptr = dataOut;
  return PutValuesFromString(&d,OS_TAG,CF_TEXT,(int)strlen(OS_TAG),0);
}
int getServerTimeStr(CONTRACT *con,char *data,time_t t)
{
  int len,fmt;
  char s[32];
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  fmt = LFMT(con->EqmFormatOut);
  switch (fmt)
  {
    case CF_INT32:
      *(UINT32 *)data = (UINT32)t;
      return 0;
    case CF_NAME16:
      len = 16;
      break;
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      len = 32;
      break;
    case CF_TEXT:
      break;
    default:
      return illegal_format;
  }
  strncpy(s,ctime(&t),24); s[24] = 0;
  if (len < 16) return dimension_error;
  if (len > 24) 
  {
    strncpy(data,s,24);
  }
  else
  {
    strncpy(data,&s[4],15); data[15] = 0;
  }
  return 0;
}
int getServerBaseline(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  time_t st = el == NULL ? gSystemStartupTime : el->eqmBaseline;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return getServerTimeStr(con,dataOut,st);
}
int getServerStarttime(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  time_t st = el == NULL ? gSystemStartupTime : el->eqmStartupTime;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return getServerTimeStr(con,dataOut,st);
}
#ifndef SMALL_TINE_LIB
int getAppDate(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return getServerTimeStr(con,dataOut,appmktime);
}
#ifndef FS_RDONLY
int getManifest(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  char fn[256];
  int cc,rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (rs > 0)
  {
    if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
    sprintf(fn,"%.128s%s",gAddrCacheFilePath,FECMF_FILENAME);
    if ((cc=readFecLogFile(fn,dataOut,rs)) != 0) return cc;
  }
  return 0;
}
int getSrvManifestPath(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int len, slen;
  UNUSED(dataIn);
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  touchStkMetaData(cl);
  if ((slen=(int)strlen(gAddrCacheFilePath)) < len) len = slen;
  strncpy(dataOut,gAddrCacheFilePath,len);
  return 0;
}
#endif
int getSrvCmdLine(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  char *cmdln;
  int len, slen;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  if ((cmdln=SystemGetStartupCommand()) == NULL) cmdln = "(unknown)";
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  if ((slen=(int)strlen(cmdln)) < len) len = slen;
  strncpy(dataOut,cmdln,len);
  return 0;
}
int getSrvCwd(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  char *cwd;
  int len, slen;
  UNUSED(dataIn);
  if ((cwd=SystemGetStartupDirectory()) == NULL) cwd = "(unknown)";
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  touchStkMetaData(cl);
  if ((slen=(int)strlen(cwd)) < len) len = slen;
  strncpy(dataOut,cwd,len);
  return 0;
}
int getSrvPid(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UINT32 pid=SystemGetProcessId();
  UNUSED(dataIn);
  touchStkMetaData(cl);
  return T_numQuery(con,dataOut,(int)pid);
}
#endif /* !SMALL_TINE_LIB */
int getErrorDescription(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT* con = &cl->contract;
  int len, slen=0, code=0;
  char errstr[256];
  if (con->EqmSizeIn > 0)
  {
    if (con->EqmAccess & CA_WRITE) return illegal_read_write;
    switch (LFMT(con->EqmFormatIn))
    {
    case CF_UINT16:
    case CF_INT16:
      code = (int)((SINT16*)dataIn)[0];
      break;
    case CF_UINT32:
    case CF_INT32:
      code = (int)((SINT32*)dataIn)[0];
      break;
    default:
      return illegal_format;
    }
  }
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  if (code == 0)
  { /* return list */
    int i=0, n;
    ErrTblItem *eti = NULL;
    NAME32I *n32i = (NAME32I *)dataOut;
    if (LFMT(con->EqmFormatOut) != CF_NAME32I) return illegal_format;
    if (len > numRegisteredErrors) len = numRegisteredErrors;
    for (eti=gRegisteredErrors,n=0,i=len-1; eti != NULL; eti = eti->nxt, n++)
    {
      if (i < 0) break;
      if (n < numRegisteredErrors - len) continue;
      n32i[i].ival = eti->code;
      strncpy(n32i[i].name, eti->msg, 32);
      i--;
    }
    assignStkMetaData(cl, 0, 0, 0, n - numRegisteredErrors + len);
  }
  else
  { /* targeted error code */
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_NAME64:
      case CF_NAME32:
        if (len != 1) return dimension_error;
        slen = 32;
        break;
      case CF_TEXT:
        slen = len;
        break;
      default:
        return illegal_format;
    }
    if (slen == 0) return invalid_datarequest;
    GetErrorText(code, errstr, slen);
    strncpy(dataOut, errstr, slen);
  }
  touchStkMetaData(cl);
  return 0;
}
int getServerDescription(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int len;
  UNUSED(dataIn);
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  touchStkMetaData(cl);
  strncpy(dataOut,gFecInfo.desc,len);
  return 0;
}
int getServerLocation(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int len;
  UNUSED(dataIn);
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  touchStkMetaData(cl);
  strncpy(dataOut,gFecInfo.loc,len);
  return 0;
}
int getServerSubsystem(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int len;
  char *subs = el == NULL ? gDeviceSubSystem : el->EqmSubSystem;
  UNUSED(dataIn);
  if (LFMT(con->EqmFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqmSizeOut) < 1) return illegal_read_write;
  touchStkMetaData(cl);
  strncpy(dataOut,subs,len);
  return 0;
}
int getServerSckErrors(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UINT32 sckErrBuffer[16];
  int i;
  int fmt = LFMT(con->EqmFormatOut);
  UNUSED(dataIn);
  if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
  touchStkMetaData(cl);
  sckErrBuffer[0] = sendtoErrcnt;
  sckErrBuffer[1] = selectErrcnt;
  sckErrBuffer[2] = selectFailcnt;
  for (i=0; i<(int)con->EqmSizeOut && i<3; i++) ((UINT32 *)dataOut)[i] = sckErrBuffer[i];
  return 0;
}
int getServerStats(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  UINT32 statsBuffer[16];
  int i;
  int fmt = LFMT(con->EqmFormatOut);
  UNUSED(dataIn);
  if (fmt != CF_INT32 && fmt != CF_UINT32) return illegal_format;
  touchStkMetaData(cl);
  statsBuffer[0] = AveBusyTime;
  statsBuffer[1] = gCycleCounts;
  statsBuffer[2] = gMaxCycleCounts;
  statsBuffer[3] = gSingleLinkCount;
  statsBuffer[4] = gClientMisses;
  statsBuffer[5] = gClientReconnects;
  statsBuffer[6] = gClientRetries;
  statsBuffer[7] = gContractMisses;
  statsBuffer[8] = gContractDelays;
  statsBuffer[9] = gBurstLimitReachedCount;
  statsBuffer[10] = (UINT32)(gDataTimeStampOffset * 1000);
  for (i=0; i<(int)con->EqmSizeOut && i<11; i++) ((UINT32 *)dataOut)[i] = statsBuffer[i];
  return 0;
}
int getServerSettings(ContractListStruct *cl,char *dataIn,char *dataOut)
{
# ifndef DOS
  CONTRACT *con=&cl->contract;
  ServerSettingsStruct sss;

  UNUSED(dataIn);
  touchStkMetaData(cl);
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (strcmp(con->strTagOut,"SrvSetQy")) return invalid_structure_tag;
  if (con->EqmSizeOut != SERVERSETTINGSSTRUCT_SIZE) return dimension_error;
  memset(&sss,0,sizeof(ServerSettingsStruct));
  sss.srvWorkArea = srvWorkAreaSize;
  sss.cycDeadBand = gSystemCycleDeadband;
  sss.minPollInterval = MinPollingRate;
  sss.conTblCapacity = ContractListCapacity;
  sss.conRenewalLen = gSubRenewalLength;
  sss.clnTblCapacity = ClientListCapacity;
  sss.ackOnChange = gRequireAcknowledgments ? 1 : 0;
  sss.srvBurstLimit = gBurstLimit;
  sss.srvBurstCycDly = gCycleDelay;
  sss.srvPacketMTU = gPacketMtu;
  sss.tcpMaxMsgSize = gMaxTcpMsgSize;
  sss.tcpTblCapacity = tcpConTblSize;
  sss.srvSendBuffers = max_srvsndbuf;
  sss.srvRecvBuffers = max_srvrcvbuf;
  sss.srvLazySched = gSchedulePropertyLazy ? 1 : 0;
  sss.srvBkgTskRentr = gAllowBkgTaskReentrancy ? 1 : 0;
  sss.srvCycInSepThrd = gRunServerCycleInSeparateThread ? 1 : 0;
  sss.tcpTraInSepThrd = gRunTransportInSeparateThread ? 1 : 0;
  sss.clnWorkArea = clnWorkAreaSize;
  sss.lnkTblCapacity = gConTblCapacity;
  sss.clnBurstLimit = gClientBurstLimit;
  sss.clnSendBuffers = max_clnsndbuf;
  sss.clnRecvBuffers = max_clnrcvbuf;
  sss.clnRecvQueue = gClnRecvQueueDepth;
  sss.useLoopbackAddr = gUseLoopback ? 1 : 0;
  sss.useWdogLinks = gAutoLinkWatchdogs ? 1 : 0;
  sss.allowCommonLnks = allowCommonLinks ? 1 : 0;
  sss.retryOnTimeout = gAlwaysRetry ? 1 : 0;
  sss.srvSchedIntvl = gSchedulerInterval;
  sss.srvRetrdRemovl = gRetardSingleContractRemoval ? 1 : 0;
  sss.srvAllowRemote = (gAllowRemoteExit|gAllowRemoteInit|gAllowRemoteReset) ? 1 : 0;

  PrepOutgoingSrvSettings(&sss,dataOut);
  return 0;
# else
  UNUSED(cl);
  UNUSED(dataOut);
  UNUSED(dataIn);
  UNUSED(dataTimeStamp);
  UNUSED(retSize);
  return not_implemented;
# endif
}
int getDeniedUserList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int i,rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  if (rs == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  touchStkMetaData(cl);
  for (i=0; i<rs && i<gDeniedUsersList.nUsrs; i++)
  {
    strncpy(((NAME16 *)dataOut)[i].name,gDeniedUsersList.usrs[i].name,USERNAME_SIZE);
  }
  assignStkMetaData(cl,0,0,0,i);
  return 0;
}
int getAdminsUserList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int i,rs = (int)(cl->dataSizeOut);
  UNUSED(dataIn);
  if (rs == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  touchStkMetaData(cl);
  for (i=0; i<(int)con->EqmSizeOut && i<gNumFecAdmins; i++)
  {
    strncpy(((NAME16 *)dataOut)[i].name,gFecAdmins[i].name,USERNAME_SIZE);
  }
  assignStkMetaData(cl,0,0,0,i);
  return 0;
}
int getRegisteredUserList(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int i,n,rs = (int)(cl->dataSizeOut);
  NAME16 *lst;
# ifndef SMALL_TINE_LIB
  int slen,devnr;
  char str[72];
  ExportPropertyListStruct *prp;
# else
  UNUSED(dataIn);
# endif
  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  if (LFMT(con->EqmFormatOut) != CF_NAME16) return illegal_format;
  touchStkMetaData(cl);
  n = el->aclLst.nUsrs;
  lst = el->aclLst.usrs;
# ifndef SMALL_TINE_LIB
  if (con->EqmSizeIn > 0)
  {
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -slen;
    strncpy(str,dataIn,slen);
    if ((prp=GetPropertyListStruct(con->EqmName,str,NULL)) != NULL)
    { /* input is a property */
      n = prp->aclLst.nUsrs;
      lst = prp->aclLst.usrs;
    }
    else if ((devnr=GetDeviceNumber(con->EqmName,str)) >= 0 && 
             (devnr < el->EqmNumDevices) && el->EqmDevLst[devnr].aclLst != NULL)
    { /* input is a device */
      if (el->EqmDevLst != NULL)
      {
        n = el->EqmDevLst[devnr].aclLst->nUsrs;
        lst = el->EqmDevLst[devnr].aclLst->usrs;
      }
    }
  }
# endif
  for (i=0; i<rs && i<n; i++)
  {
    strncpy(((NAME16 *)dataOut)[i].name,lst[i].name,USERNAME_SIZE);
  }
  assignStkMetaData(cl,0,0,0,i);
  return 0;
}
void prepOutgoingCLNQS(void *tgt,ClnHdr *src)
{
  CLNQS cqs;
  memset(&cqs,0,sizeof(CLNQS));
  strncpy(cqs.userName,src->usr,USERNAME_SIZE);
  memcpy(&cqs.IPaddress,&src->adr.ip,sizeof(struct sockaddr_in));
  cqs.IPaddress.sin_port = inetPort(&src->adr.ip); /* straighten this out for he caller */
  cqs.ncontracts = src->ncontracts;
  cqs.tineProtocol = src->tineProtocol;
# ifdef DOS
  memcpy(tgt,&cqs,sizeof(CLNQS));
# else
   _structSwap(LClientQueryStructStruct,(BYTE *)tgt,0,NULL,(BYTE *)&cqs,LTOR);
# endif
}
void prepOutgoingClnInfoQS(void *tgt, ClnHdr *src)
{
  ClnInfoQueryStruct cqs;
  memset(&cqs, 0, sizeof(ClnInfoQueryStruct));
  strncpy(cqs.userName, src->usr, USERNAME_SIZE);
  inet_ntop(src->adr.ip.ss_family, getInAddr(&(src->adr.ip)), cqs.userAddr, ADDR_SIZE_IPV6);
  strncpy(cqs.userType, src->typ, USERTYPE_SIZE);
  cqs.pid = src->pid;
  cqs.port = inetPort(&src->adr.ip); /* straighten this out for he caller */
  cqs.ncontracts = src->ncontracts;
  cqs.tineProtocol = src->tineProtocol;
# ifdef DOS
  memcpy(tgt, &cqs, sizeof(ClnInfoQueryStruct));
# else
  _structSwap(ClnInfoQueryStructStruct, (BYTE *)tgt, 0, NULL, (BYTE *)&cqs, LTOR);
# endif
}
int clientsQuery(ContractListStruct *cl, char *dataIn, char *dataOut)
{
  CONTRACT *con = &cl->contract;
  ClnHdr *cln;
  int i, ssiz, ncln, rs = (int)(cl->dataSizeOut);

  UNUSED(dataIn);
  if (LFMT(con->EqmFormatOut) != CF_STRUCT) return illegal_format;
  if (!strcmp(con->strTagOut, "CLNQS"))
  {
    ssiz = CLIENT_QUERY_SIZE;
  }
  else if (!strcmp(con->strTagOut, "ClnInfoQS"))
  {
    ssiz = sizeof(ClnInfoQueryStruct);
  }
  else
  {
    return invalid_structure_tag;
  }
  if ((ncln=rs/ssiz) == 0) return dimension_error;
  touchStkMetaData(cl);
  for (i = 0; i < ncln && i < nconsumers; i++)
  {
    if (ClnTbl[i] == NULL) continue;
    cln = ClnTbl[i];
    if (ssiz == CLIENT_QUERY_SIZE)
    { /* legacy request */
      prepOutgoingCLNQS(&dataOut[i*ssiz], cln);
    }
    else
    {
      prepOutgoingClnInfoQS(&dataOut[i*ssiz], cln);
    }
  }
  assignStkMetaData(cl, 0, 0, 0, i*ssiz);
  return 0;
}
int followersQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int i, rs = (int)(cl->dataSizeOut);

  if (el == NULL) return non_existent_elem;
  if ((con->EqmAccess&CA_WRITE))
  {
    int cc=0, slen=0, p=0, rmv=FALSE;
    char fllr[70];
    if (!(con->EqmAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        slen = fmtsizeof(con->EqmFormatIn);
        break;
      case CF_TEXT:
        slen = con->EqmSizeIn;
      default:
        return illegal_format;
    }
    if (*dataIn == '-') p = 1, rmv = TRUE;
    strncpy(fllr,&dataIn[p],slen); fllr[64] = 0;
    cc = rmv ? rmvFollower(el->EqmName,fllr) : addFollower(el->EqmName,fllr,64);
    if (cc != 0) return cc;
  }    
  if (rs == 0) return 0;
  if (LFMT(con->EqmFormatOut) != CF_NAME64) return illegal_format;
  touchStkMetaData(cl);
  for (i=0; i<el->nfollowers && i<rs; i++)
  {
    strncpy(((NAME64 *)dataOut)[i].name,el->followers[i].name,64);
  }
  assignStkMetaData(cl,0,0,0,i);
  return 0;
}
int gateFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{ /* keep up apperances ... */
  UNUSED(dataIn);
  UNUSED(dataOut);
  UNUSED(cl);
  return 0;
}
int EGUQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int idx,num,len,nlen,isWrite=FALSE,isRangeQuery=FALSE,isMax=FALSE,isMin=FALSE,rs = (int)(cl->dataSizeOut);
  ExportPropertyListStruct *prp,prpTmp;
  ExportListStruct *el;
  PrpQueryStruct *prpqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  char *c, Property[PROPERTY_NAME_SIZE+4];
  DUSTRING *ustr;

  touchStkMetaData(cl);
  isWrite = (con->EqmAccess&CA_WRITE);
  len = rs;
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".XEGU")) != NULL ||
      (c=strstr(Property,".XMIN")) != NULL ||
      (c=strstr(Property,".XMAX")) != NULL )
  { /* a range query */
    isRangeQuery = TRUE;
  }
  else if ((c=strstr(Property,".EGU")) == NULL &&
           (c=strstr(Property,".MIN")) == NULL &&
           (c=strstr(Property,".MAX")) == NULL) 
  { /* can't be ! */
    return illegal_property;
  }
  if (strstr(c,"MIN") != NULL) isMin = TRUE;
  if (strstr(c,"MAX") != NULL) isMax = TRUE;
  *c = 0;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL)
    { /* registered query fcn */
      prp = NULL;
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (xpqs == NULL) return illegal_property;
      if (num <= 0) return -num;
      if (isWrite) return illegal_read_write;
      for (idx=0; idx<num; idx++)
      { /* Is .EGU query managed by query function ? */
        if (!strnicmp(con->EqmProperty,xpqs[idx].prpName,PROPERTY_NAME_SHORTSIZE)) return ambiguous;
        if (strnicmp(Property,xpqs[idx].prpName,PROPERTY_NAME_SHORTSIZE)) continue;
        strncpy(prpTmp.egu.units,xpqs[idx].prpUnits,UNITS_SIZE);
        prpTmp.egu.min = xpqs[idx].prpMinValue;
        prpTmp.egu.max = xpqs[idx].prpMaxValue;
        prpTmp.egu.graph = xpqs[idx].prpGraphType;
        strncpy(prpTmp.xegu.units,xpqs[idx].rngUnits,UNITS_SIZE);
        prpTmp.xegu.min = xpqs[idx].rngMinValue;
        prpTmp.xegu.max = xpqs[idx].rngMaxValue;
        prpTmp.xegu.graph = xpqs[idx].prpGraphType;
        prp = &prpTmp;
      }
      if (prp != NULL) goto checkformat;
      return illegal_property;
    }
    if (el->EqmPropertyQueryFcn != NULL) 
    { /* registered query fcn */
      prp = NULL;
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (prpqs == NULL) return illegal_property;
      if (num <= 0) return -num;
      if (isWrite) return illegal_read_write;
      for (idx=0; idx<num; idx++)
      { /* Is .EGU query managed by query function ? */
        if (!strnicmp(con->EqmProperty,prpqs[idx].prpName,PROPERTY_NAME_SIZE)) return ambiguous;
        if (strnicmp(Property,prpqs[idx].prpName,PROPERTY_NAME_SIZE)) continue;
        strncpy(prpTmp.egu.units,prpqs[idx].prpUnits,UNITS_SIZE);
        prpTmp.egu.min = prpqs[idx].prpMinValue;
        prpTmp.egu.max = prpqs[idx].prpMaxValue;
        prpTmp.egu.graph = prpqs[idx].prpGraphType;
        strncpy(prpTmp.xegu.units,prpqs[idx].rngUnits,UNITS_SIZE);
        prpTmp.xegu.min = prpqs[idx].rngMinValue;
        prpTmp.xegu.max = prpqs[idx].rngMaxValue;
        prpTmp.xegu.graph = prpqs[idx].prpGraphType;
        prp = &prpTmp;
      }
      if (prp != NULL) goto checkformat;
      return illegal_property;
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
    {
      if (prp->isReserved) continue;
      if (strnicmp(Property,prp->prpName,PROPERTY_NAME_SIZE)) continue;
      goto checkformat;
    }
    if (prp == NULL)
    {
      memset(&prpTmp,0,sizeof(ExportPropertyListStruct));
      prp = &prpTmp;
    }
checkformat:
    if (isWrite && con->EqmSizeIn > 0)
    { /* WRITE (non-query fcns only) requests are volatile ! */
      switch (LFMT(con->EqmFormatIn))
      {
        case CF_USTRING:
          if (isMin || isMax) return illegal_format;
          ustr = (DUSTRING *)dataIn;
          prp->egu.graph =  (int)ustr->ival;
          if (isRangeQuery) 
          {
            prp->xegu.min = ustr->f1val; 
            prp->xegu.max = ustr->f2val;
            strncpy(prp->xegu.units,ustr->str,UNITS_SIZE);
          }
          else
          {
            prp->egu.min = ustr->f1val; 
            prp->egu.max = ustr->f2val;
            strncpy(prp->egu.units,ustr->str,UNITS_SIZE);
          }
          if (len == 0) return 0;
          break;
        case CF_NAME64:
        case CF_NAME32:
        case CF_NAME16:
        case CF_TEXT:
          if (isMin || isMax) return illegal_format;
          nlen = (int)con->EqmSizeIn * fmtsizeof(con->EqmFormatIn);
          if (nlen < 1) return illegal_read_write;
          if (nlen > UNITS_SIZE) nlen = UNITS_SIZE;
          if (isRangeQuery) 
          {
            strncpy(prp->xegu.units,dataIn,nlen);
          }
          else
          {
            strncpy(prp->egu.units,dataIn,nlen);
          }
          if (len == 0) return 0;
          break;
        case CF_FLOAT:
          if (isRangeQuery) 
          {
            if (isMin) prp->xegu.min = ((float *)dataOut)[0];
            else if (isMax) prp->xegu.max = ((float *)dataOut)[0];
            else
            {
              if (con->EqmSizeIn != 2) return illegal_read_write;
              prp->xegu.min = ((float *)dataOut)[0]; 
              prp->xegu.max = ((float *)dataOut)[1];
            }
          }
          else
          {
            if (isMin) prp->egu.min = ((float *)dataOut)[0];
            else if (isMax) prp->egu.max = ((float *)dataOut)[0];
            else
            {
              if (con->EqmSizeIn != 2) return illegal_read_write;
              prp->egu.min = ((float *)dataOut)[0]; 
              prp->egu.max = ((float *)dataOut)[1];
            }
          }
          if (len == 0) return 0;
          break;
        default:
          return illegal_format;
      }
    }    
    if (len == 0) return illegal_read_write;
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_USTRING:
        if (isMin || isMax) return illegal_format;
        ustr = (DUSTRING *)dataOut;
        ustr->ival  = (UINT32)prp->egu.graph;
        ustr->f1val = (float)(isRangeQuery ? prp->xegu.min : prp->egu.min);
        ustr->f2val = (float)(isRangeQuery ? prp->xegu.max : prp->egu.max);
        ustr->tm = (UINT32)gSystemStartupTime;
        strncpy(ustr->str,(isRangeQuery ? prp->xegu.units : prp->egu.units),UNITS_SIZE);
        return 0;
      case CF_NAME64:
      case CF_NAME32:
      case CF_NAME16:
        if (len != 1) return dimension_error;
      case CF_TEXT:
        if (isMin || isMax) return illegal_format;
        nlen = len * fmtsizeof(con->EqmFormatOut);
        if (nlen < 1) return illegal_read_write;
        if (nlen > UNITS_SIZE) nlen = UNITS_SIZE;
        strncpy((char *)dataOut,(isRangeQuery ? prp->xegu.units : prp->egu.units),nlen);
        return 0;
      case CF_FLOAT:
        if (isMin) 
        {
          ((float *)dataOut)[0] = isRangeQuery ? prp->xegu.min : prp->egu.min;
        }
        else if (isMax) 
        {
          ((float *)dataOut)[0] = isRangeQuery ? prp->xegu.max : prp->egu.max;
        }
        else
        {
          if (len != 2) return illegal_read_write;
          ((float *)dataOut)[0] = (float)(isRangeQuery ? prp->xegu.min : prp->egu.min);
          ((float *)dataOut)[1] = (float)(isRangeQuery ? prp->xegu.max : prp->egu.max);
        }
        if (isMin || isMax) assignStkMetaData(cl,0,0,0,1);
        return 0;
      default:
        return illegal_format;
    }
  }
  return non_existent_elem;
}

extern int _gNeedWarmRestart;
TINE_EXPORT int gAllowRemoteReset = DEFAULT_REMOTE_MANAGMENT;
TINE_EXPORT int gAllowRemoteInit = DEFAULT_REMOTE_MANAGMENT;
TINE_EXPORT int gAllowRemoteExit = DEFAULT_REMOTE_MANAGMENT;
#ifndef SMALL_TINE_LIB
int srvExit(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int code = 0;
  char *un;
  UNUSED(dataOut);
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0) 
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_INT16:
        code = (int)((SINT16 *)dataIn)[0];
        break;
      case CF_UINT16:
        code = (int)((UINT16 *)dataIn)[0];
        break;
      case CF_INT32:
        code = (int)((SINT32 *)dataIn)[0];
        break;
      default:
        return illegal_format;
    }
  }
  if (!gAllowRemoteExit) return not_accepted;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  un = GetCaller(con->EqmName);
  if (!AssertIsAdministrator(un))
  {
    char es[256];
    sprintf(es,"access denied: user %.16s is not an administrator",un);
    SetEqmCompletion(con->EqmName,es);
    return access_denied;
  }
  feclogEx(TINE_LOGLEVEL_INFO, "remote exit from %s",un);
  gHaltOnCycleHalt = TRUE;
  gSystemExitCode = code;
  quitServer();
  return 0;
}
int srvReset(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  char *un;
  int idleTime = _gReinitIdleTime;
  UNUSED(dataOut);
  touchStkMetaData(cl);
  if (con->EqmSizeIn > 0) 
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_INT16:
        idleTime = (int)((SINT16 *)dataIn)[0];
        break;
      case CF_UINT16:
        idleTime = (int)((UINT16 *)dataIn)[0];
        break;
      case CF_INT32:
        idleTime = (int)((SINT32 *)dataIn)[0];
        break;
      case CF_UINT32:
        idleTime = (int)((UINT32 *)dataIn)[0];
        break;
      default:
        return illegal_format;
    }
    if (idleTime < 0) return out_of_range;
  }
  if (!gAllowRemoteReset) return not_accepted;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  un = GetCaller(con->EqmName);
  if (!AssertIsAdministrator(un))
  {
    char es[256];
    sprintf(es,"access denied: user %.16s is not an administrator",un);
    SetEqmCompletion(con->EqmName,es);
    return access_denied;
  }
  feclogEx(TINE_LOGLEVEL_INFO, "remote reset from %s",un);
  _gReinitIdleTime = idleTime;
  _gNeedWarmRestart = TRUE;
  return 0;
}
int srvInit(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  char *un;
  UNUSED(dataIn);
  UNUSED(dataOut);
  if (!gAllowRemoteInit) return not_accepted;
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  un = GetCaller(con->EqmName);
  if (!AssertIsAdministrator(un))
  {
    char es[256];
    sprintf(es,"access denied: user %.16s is not an administrator",un);
    SetEqmCompletion(con->EqmName,es);
    return access_denied;
  }
  feclogEx(TINE_LOGLEVEL_INFO, "remote init from %s",un);
  if (el->eqmIni) TSKCALL(el,eqmIni);
  return 0;
}
int srvReex(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  UNUSED(dataIn);
  UNUSED(dataOut);
  if ((con->EqmAccess&CA_WRITE)) 
  {
    ExportsRegistered = FALSE;
    el->registered = 0;
    feclogEx(TINE_LOGLEVEL_INFO, "signal to re-export modules from %.16s",GetCaller(con->EqmName));
  }
  else
  {
    return illegal_read_write;
  }
  return 0;
}
int srvIdle(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int idleState = 0;
  if (el == NULL) return non_existent_elem;
  touchStkMetaData(cl);
  if ((con->EqmAccess&CA_WRITE)) 
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_INT16:
        idleState = (int)((SINT16 *)dataIn)[0];
        break;
      case CF_UINT16:
        idleState = (int)((UINT16 *)dataIn)[0];
        break;
      case CF_BOOLEAN:
      case CF_INT32:
        idleState = (int)((SINT32 *)dataIn)[0];
        break;
      case CF_UINT32:
        idleState = (int)((UINT32 *)dataIn)[0];
        break;
      default:
        return illegal_format;
    }
    el->isidle = idleState;
    feclogEx(TINE_LOGLEVEL_INFO, "set remote idle state to %s from %.16s",idleState ? "TRUE" : "FALSE",GetCaller(con->EqmName));
  }
  if (con->EqmSizeOut > 0)
  {
    switch (LFMT(con->EqmFormatOut))
    {
      case CF_INT16:
        ((SINT16 *)dataOut)[0] = el->isidle;
        break;
      case CF_UINT16:
        ((UINT16 *)dataOut)[0] = (UINT16)el->isidle;
        break;
      case CF_BOOLEAN:
      case CF_INT32:
        ((SINT32 *)dataOut)[0] = el->isidle;
        break;
      case CF_UINT32:
        ((UINT32 *)dataOut)[0] = (UINT32)el->isidle;
        break;
      default:
        return illegal_format;
    }
  }
  return 0;
}
#endif
int lockAccess(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int lockType, lockDuration, nlen = 0, lt, dt,rs = (int)(cl->dataSizeOut);
  int actionTaken = FALSE, passed = FALSE;
  char *addr;
  ClnHdr *cln = cl->clnLst->client;
  touchStkMetaData(cl);
  if (con->EqmAccess&CA_WRITE)
  {
    if (con->EqmSizeIn != 2) return dimension_error;
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_INT16:
        lockType = (int)((SINT16 *)dataIn)[0];
        lockDuration = (int)((SINT16 *)dataIn)[1];
        break;
      case CF_UINT16:
        lockType = (int)((UINT16 *)dataIn)[0];
        lockDuration = (int)((UINT16 *)dataIn)[1];
        break;
      case CF_INT32:
        lockType = (int)((SINT32 *)dataIn)[0];
        lockDuration = (int)((SINT32 *)dataIn)[1];
        break;
      case CF_UINT32:
        lockType = (int)((UINT32 *)dataIn)[0];
        lockDuration = (int)((UINT32 *)dataIn)[1];
        break;
      default:
        return illegal_format;
    }
    lt = lockType & 0xff;
    if (lt < 0 || lt > LOCK_ABORT) return out_of_range;
    if (lockDuration < 0) lockDuration = 0;
    if (lockDuration > MAX_LOCK_DURATION) lockDuration = MAX_LOCK_DURATION;
    if (isMemberLockSet(el,cln)) passed = TRUE;
    else if ((el->accessLock.lockType & 0xff) != LOCK_PERSISTENT) passed = TRUE;
    if (!passed) return access_denied;
    if (!putCommandsInFeclog && el->accessLock.lockType != lockType)
    {
      feclogEx(TINE_LOGLEVEL_INFO, "ACCESSLOCK %d (%d seconds) called from %s",lockType,lockDuration,GetCaller(con->EqmName));
    }
    if (lt == LOCK_ABORT || lt == LOCK_UNLOCKED)
    {
      memset(&el->accessLock,0,sizeof(AccessLock));
    }
    else
    {
      memcpy(&el->accessLock.cln,cln,sizeof(ClnHdr));
      el->accessLock.lockType = lockType;
      el->accessLock.lockDuration = lockDuration;
      el->accessLock.lockStart = (UINT32)time(NULL);
      lockToExclusiveRead(cln);
    }
    actionTaken = TRUE;
  }
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      nlen = fmtsizeof(con->EqmFormatOut);
      switch (rs)
      {
        case 3:
          dt = el->accessLock.lockDuration + el->accessLock.lockStart - (UINT32)time(NULL);
          if (dt < 0) dt = 0;
          sprintf(&dataOut[2*nlen],"%d sec remaining",dt);
        case 2:
          addr = &dataOut[nlen];
          inetStrAdr(&el->accessLock.cln.adr.ip,addr,TRUE);
        case 1:
          strncpy(dataOut,el->accessLock.cln.usr,USERNAME_SIZE);
          actionTaken = TRUE;
        default:
          break;
      }
      break;
    case CF_NULL:
      break;
    default:
      if (!actionTaken) return illegal_format;
      break; 
  }
  return actionTaken ? 0 : illegal_read_write;
}
#ifndef SMALL_TINE_LIB
int globalsRecordsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int i,slen,flen,n=0,addfmt = CF_NULL,rs = (int)(cl->dataSizeOut);

  UNUSED(dataIn);
  touchStkMetaData(cl);
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      slen = fmtsizeof(con->EqmFormatOut);
      break;
    case CF_NAME64DBLDBL:
      slen = 64;
      addfmt = CF_DOUBLE;
      break;
    case CF_NAME16II:
      slen = 16;
      addfmt = CF_INT32;
      break;
    default: 
      return illegal_format;
  }
  flen = fmtsizeof(con->EqmFormatOut);
  for (i=0; i<ngcasts && n<rs; i++)
  {
    strncpy(&dataOut[n*flen],gcastTbl[i]->keyword,slen);
    switch (addfmt)
    {
      case CF_DOUBLE:
        *(double *)&dataOut[n*flen+64] = (double)gcastTbl[i]->format;
        *(double *)&dataOut[n*flen+72] = (double)gcastTbl[i]->size;
        break;
      case CF_INT32:
        *(SINT32 *)&dataOut[n*flen+16] = (SINT32)gcastTbl[i]->format;
        *(SINT32 *)&dataOut[n*flen+20] = (SINT32)gcastTbl[i]->size;
        break;
      default:
        break;
    }
    n++;
  }
  assignStkMetaData(cl,0,0,0,n);
  return 0;
}
#endif
int nhistoryRecordsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int i,n=0;
  UNUSED(dataIn);
  touchStkMetaData(cl);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strcmp(con->EqmName,hstTbl[i].c.EqmName)) continue;
    n++;
  }
  return T_numQuery(con,dataOut,n);
}
int historyRecordsQuery(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int i,slen=16,nreq=0,n=0,rs = (int)(cl->dataSizeOut);
  char tgt[PROPERTY_NAME_SIZE];
  HRS *hrs=NULL;
  if (dataOut == NULL) return argument_list_error;
  touchStkMetaData(cl);
  memset(tgt,0,PROPERTY_NAME_SIZE);
  if (con->EqmSizeIn > 0)
  {
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -slen;
    strncpy(tgt,dataIn,slen);
  }
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      slen = fmtsizeof(con->EqmFormatOut);
      nreq = rs;
      break;
    case CF_STRUCT:
      if (strcmp(con->strTagOut,"HRSr4")) return invalid_structure_tag;
      if (rs%HRS_SIZE) return invalid_structure_size;
      nreq = rs/HRS_SIZE;
      hrs = (HRS *)tmpWorkArea;
      break;
    default: 
      return illegal_format;
  }
  for (i=0; i<nHistoryRecords && n<nreq; i++)
  {
    if (strcmp(con->EqmName,hstTbl[i].c.EqmName)) continue;
    if (tgt[0] != 0 && strnicmp(hstTbl[i].c.EqmProperty,tgt,PROPERTY_NAME_SIZE)) continue;
    if (hrs == NULL)
    { /* just a names list */
      strncpy(&dataOut[n*slen],hstTbl[i].c.EqmProperty,slen);
    }
    else
    { /* a history record structure list */
      strncpy(hrs[n].property,hstTbl[i].c.EqmProperty,PROPERTY_NAME_SIZE);
      strncpy(hrs[n].device,hstTbl[i].c.EqmDeviceName,DEVICE_NAME_SIZE);
      hrs[n].datasiz = hstTbl[i].c.EqmSizeOut;
      hrs[n].datafmt = LFMT(hstTbl[i].c.EqmFormatOut);
      hrs[n].archiveRate = hstTbl[i].archiveRate;
      hrs[n].pollingRate = hstTbl[i].pollingRate;
      hrs[n].depthShort = hstTbl[i].depthShort;
      hrs[n].depthLong = hstTbl[i].depthLong;
      hrs[n].heartbeat = hstTbl[i].heartbeat;
      hrs[n].pTolerance = hstTbl[i].pTolerance;
      hrs[n].aTolerance = hstTbl[i].aTolerance;
    }
    n++;
  }
  if (hrs != NULL) 
  {
    PrepOutgoingHRS(hrs,dataOut,n);
    n *= HRS_SIZE;
  }
  assignStkMetaData(cl,0,0,0,n);
  return 0;
}
#ifndef SMALL_TINE_LIB
int historyRecordsAppend(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  HRS hrs;
  HistorySpecification hspec, *phspec = &hspec;
  ExportPropertyListStruct *pl=NULL;
  int cc,n,slen,idx,siz,fmt;
  char prp[PROPERTY_NAME_SIZE+4],*pprp=NULL,*pdev=NULL;

  UNUSED(dataOut);
  touchStkMetaData(cl);
  if (!(con->EqmAccess&CA_WRITE)) return illegal_read_write;
  if (LFMT(con->EqmFormatIn) == CF_STRUCT)
  { /* full specification given */
    if (con->EqmSizeIn%HRS_SIZE) return illegal_read_write;
    n = (int)con->EqmSizeIn/HRS_SIZE;
    if (n != 1) return dimension_error;
    memset(&hrs,0,HRS_SIZE);
    PrepIncomingHRS(&hrs,dataIn);
    hspec.archiveRate = hrs.archiveRate;
    hspec.aTolerance = hrs.aTolerance;
    hspec.depthLong = hrs.depthLong;
    hspec.depthShort = hrs.depthShort;
    hspec.heartbeat = hrs.heartbeat;
    hspec.pollingRate = hrs.pollingRate;
    hspec.pTolerance = hrs.pTolerance;
    hspec.rhsPropertyName = NULL;
    hspec.rhsServerName = NULL;
    siz = hrs.datasiz;
    fmt = hrs.datafmt;
    pprp = hrs.property;
    pdev = hrs.device;
  }
  else
  { /* only property name specified: user defaults ... */
    if ((slen=getStringFormatLength(con->EqmFormatIn,con->EqmSizeIn,64)) < 0) return -slen;
    strncpy(prp,dataIn,slen); prp[PROPERTY_NAME_SIZE] = 0;
    if ((pl=GetPropertyListStruct(con->EqmName,prp,con->EqmDeviceName)) == NULL)
      return illegal_property;
    pprp = prp;
    pdev = con->EqmDeviceName;
    siz = pl->prpSize;
    fmt = pl->prpFormat;
    phspec = &defaultHSpec;
  }
  idx = FindAvailableHistoryIndex(con->EqmName,pprp,pdev,fmt,siz);
  gIsLclHstLateAddition = TRUE;
  cc = AppendHistoryInformation(con->EqmName,pprp,pdev,siz,fmt,idx,phspec);  
  return cc;
}
#endif
int NAMQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  ExportListStruct *el = cl->exp;
  int cc,i,k,slen,start=0,chkOn=FALSE,rs = (int)(cl->dataSizeOut);
  UINT32 mask = 0;
  ExportPropertyListStruct *prp;
  char *c,tgtprp[PROPERTY_NAME_SIZE+4];
  UNUSED(dataIn);
  strncpy(tgtprp,con->EqmProperty,PROPERTY_NAME_SIZE);
  if (el == NULL) return non_existent_elem;
  if (rs == 0) return illegal_read_write;
  touchStkMetaData(cl);
  if ((slen=getStringFormatLength(con->EqmFormatOut,rs,0)) < 0) return -slen;
  if ((c=strstr(tgtprp,".NAM")) != NULL) *c = 0;
  if ((c=strstr(tgtprp,".DMASK.")) != NULL)
  { /* a double meta property: get the names of those with the given device mask */
    *c = 0;
    c += 7;
    mask = strtol(c,NULL,0);
    if (strchr(con->EqmDeviceName,'*') != NULL) con->EqmDeviceName[0] = 0;
    goto EqmDeviceListCheck;
  }
  else if ((c=strstr(tgtprp,".ONLINE")) != NULL)
  { /* a double meta property: get the names of those not marked off-line */
    *c = 0;
    c += 7;
    chkOn = TRUE;
    if (strchr(con->EqmDeviceName,'*') != NULL) con->EqmDeviceName[0] = 0;
    goto EqmDeviceListCheck;
  }
  if ((prp=AssignDeviceListFromProperty(con->EqmName,tgtprp)) != NULL)
  {
    if (strlen(con->EqmDeviceName) > 0)
    { /* device list at a starting point ... */
      if (con->EqmDeviceName[0] == '#') 
      { /* explicit */
        start = atoi(&con->EqmDeviceName[1]);
      }
      else for (i=0; i<prp->numDevices; i++)
      { /* a start device name given */
        if (!strnicmp(con->EqmDeviceName,prp->devNames[i].name,DEVICE_NAME_SIZE)) 
        {
          start = i;
          break;
        }
      }
      /* if not found then start remains = 0 */
    }
    for (i=start,k=0; k<rs && i<prp->numDevices; i++, k++)
    {
      strncpy(&dataOut[k*slen],prp->devNames[i].name,slen);
    }
    assignStkMetaData(cl,0,0,0,k);
    return 0;
  }
EqmDeviceListCheck:
  cc = deviceQuery(cl,dataIn,dataOut);
  if (cc & CE_SENDDATA) cc = 0;
  if (cc == 0 && mask != 0 && el->EqmDevLst != NULL)
  {
    for (i=0,k=0; i<rs; i++)
    { /* this will only be MCA types and nothing with a format header */
      if (el->EqmDevLst == NULL) continue;
      if (el->EqmDevLst[i].mask != 0 && (el->EqmDevLst[i].mask & mask) == 0) continue;
      if (chkOn && (el->EqmDevLst[i].mask & DEVICE_OFFLINE)) continue;
      if (k < i) memcpy(&dataOut[k*slen],&dataOut[i*slen],slen);
      k++;
    }
    assignStkMetaData(cl,0,0,0,k);
  }
  return cc;
}
int DESCQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int idx,len,off,num,fmt,rs = (int)(cl->dataSizeOut);
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  XPropertyQueryStruct *xpqs = NULL;
  PrpQueryStruct *prpqs = NULL;
  char *c, Property[PROPERTY_NAME_SIZE+4];

  UNUSED(dataIn);
  touchStkMetaData(cl);
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".DESC")) == NULL &&
      (c=strstr(Property,".DSC")) == NULL) return illegal_property;
  *c = 0;
  off = 0;
  if ((len=rs) < 1) return dimension_error;
  switch (fmt=LFMT(con->EqmFormatOut))
  {
    case CF_TEXT:
      if (len > 64) len = 64;
      break;
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      assignStkMetaData(cl,0,0,0,1);
      len = fmtsizeof(fmt);
      break;
    case CF_USTRING:
      assignStkMetaData(cl,0,0,0,1);
      len = 64;
      off = 16;
      break;
    default:  
      return illegal_format;
  }

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL)
    { /* registered query fcn */
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (xpqs == NULL) return illegal_property;
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        if (strnicmp(xpqs[idx].prpName,Property,PROPERTY_NAME_SHORTSIZE)) continue;
        strncpy(&dataOut[off],xpqs[idx].prpDescription,len);
        return 0;
      }
    }
    if (el->EqmPropertyQueryFcn != NULL)
    { /* registered query fcn */
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (prpqs == NULL) return illegal_property;
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        if (strnicmp(prpqs[idx].prpName,Property,PROPERTY_NAME_SIZE)) continue;
        strncpy(&dataOut[off],prpqs[idx].prpDescription,MIN(len,PROPERTY_NAME_SIZE));
        return 0;
      }
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
    {
      if (prp->isReserved) continue;
      if (strnicmp(prp->prpName,Property,PROPERTY_NAME_SIZE)) continue;
      strncpy(&dataOut[off],prp->prpDescription,len);
      return 0;
    }
    return illegal_property;
  }
  return non_existent_elem;
}
#ifndef SMALL_TINE_LIB
int URLQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  int idx,len,off,num,fmt,rs = (int)(cl->dataSizeOut);
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  XPropertyQueryStruct *xpqs = NULL;
  PrpQueryStruct *prpqs = NULL;
  char *c, *url="none", Property[PROPERTY_NAME_SIZE+4];

  UNUSED(dataIn);
  touchStkMetaData(cl);
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".URL")) == NULL) return illegal_property;
  *c = 0;
  off = 0;
  if ((len=rs) < 1) return dimension_error;
  switch (fmt=LFMT(con->EqmFormatOut))
  {
    case CF_TEXT:
      break;
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      assignStkMetaData(cl,0,0,0,1);
      len = fmtsizeof(fmt);
      break;
    case CF_USTRING:
      assignStkMetaData(cl,0,0,0,1);
      len = 80;
      off = 16;
      break;
    default:  
      return illegal_format;
  }

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL)
    { /* registered query fcn */
      num = el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs);
      if (xpqs == NULL) return illegal_property;
      if (num <= 0) return -num;
      strncpy(&dataOut[off],url,len);
      return 0;
    }
    if (el->EqmPropertyQueryFcn != NULL)
    { /* registered query fcn */
      num = el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs);
      if (prpqs == NULL) return illegal_property;
      if (num <= 0) return -num;
      strncpy(&dataOut[off],url,len);
      return 0;
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqmPrpLst[idx]; prp; prp=prp->next)
    {
      if (prp->isReserved) continue;
      if (strnicmp(prp->prpName,Property,PROPERTY_NAME_SIZE)) continue;
      if (prp->prpUrl != NULL) url = prp->prpUrl;
      strncpy(&dataOut[off],url,len);
      return 0;
    }
    return illegal_property;
  }
  return non_existent_elem;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT int GetLocalHistoryData(char *eqm,char *prp, char *dev,DTYPE *dout,DTYPE *din)
{
  ContractListStruct cl = {0};
  ClnLst cln = {0};
  ClnHdr chdr = CLNHDR_STRUCT_INITIALIZER;
  int cc;
  strncpy(chdr.usr, "LOCALHOST", USERNAME_SIZE);
  chdr.ncontracts = 1;
  chdr.tineProtocol = 7;
  cln.client = &chdr;
  cln.status.mode = CM_SINGLE;
  cl.clnLst = &cln;
  cl.nclients = 1;
  strncpy(cl.contract.EqmName,eqm,EQM_NAME_SHORTSIZE);
  strncpy(cl.contract.EqmProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(cl.contract.EqmDeviceName,dev,DEVICE_NAME_SIZE);
  cl.contract.EqmSizeOut = dout->dArrayLength;
  cl.contract.EqmFormatOut = BFMT(dout->dFormat);
  cl.contract.EqmSizeIn = din->dArrayLength;
  cl.contract.EqmFormatIn = BFMT(din->dFormat);
  cl.dataOutCapacity = cl.dataSizeOut = dout->dArrayLength;
  cl.dataBuf = dout->data.bptr;
  cc = HISTQueryFunction(&cl,din->data.cptr,dout->data.cptr);
  if (cc == 0)
  {
    dout->dTimeStamp = (double)cl.dtimestamp + (double)(cl.dtimestampUSEC/1000000.0);
    dout->sysStamp = cl.sysstamp;
    dout->dStamp = cl.datastamp;
    dout->dArrayLength = cl.dataSizeOut;
  }
  return cc;
}
int HRBQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  DSUMMARY ds;
  char *ext, prp[PROPERTY_NAME_SIZE];
  double si = 1;
  double dvals[4];
  int cc;

  strncpy(prp,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((ext=strrchr(prp,'.')) == NULL) return code_failure;
  touchStkMetaData(cl);
  *ext++ = 0;
  if (con->EqmSizeIn > 0)
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_INT32:
        si = (double)(*(SINT32 *)dataIn);
        break;
      case CF_UINT32:
        si = (double)(*(UINT32 *)dataIn);
        break;
      case CF_FLOAT:
        si = (double)(*(float *)dataIn);
        break;
      case CF_DOUBLE:
        si = (*(double *)dataIn);
        break;
      default:
        return illegal_format;
    }
  }
  if (con->EqmSizeOut < 1) return dimension_error;
  cc = getSTSDataSummary(con->EqmName,prp,con->EqmDeviceName,si,&ds);
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_STRUCT:
      if (strcmp(ext,"RBSUMMARY")) return illegal_property;
      if (strncmp(con->strTagOut,"DSUMMARY",TAG_NAME_SIZE)) return invalid_structure_tag;
      break; /* we're done */
    default:
      if (cc == 0)
      { /* so far, so good */
        DTYPE d;
        DTYPEZERO(d);
        memcpy(dvals,&ds.ave,4*sizeof(double));
        d.dArrayLength = con->EqmSizeOut;
        d.dFormat = LFMT(con->EqmFormatOut);
        d.data.dptr = (double *)dataOut;
        if (strcmp(ext,"RBSUMMARY") == 0) 
        {
          if (d.dArrayLength > 4) d.dArrayLength = 4;
          assignStkMetaData(cl,0,0,0,d.dArrayLength);
          return PutValuesFromDouble(&d,dvals,4);
        }
        assignStkMetaData(cl,0,0,0,1);
        if (strcmp(ext,"RBAVE") == 0) return PutValuesFromDouble(&d,dvals,1);
        if (strcmp(ext,"RBMAX") == 0) return PutValuesFromDouble(&d,&dvals[2],1);
        if (strcmp(ext,"RBMIN") == 0) return PutValuesFromDouble(&d,&dvals[3],1);
        return illegal_property;
      }
      break;
  }
  return cc;
}
double assignKeyValue(const char *key,char **kvset,int nkv)
{
  int k;
  char *c, sbuf[128];
  for (k = 0; k < nkv; k++)
  {
    strncpy(sbuf, kvset[k], 128);
    if ((c=strchr(sbuf, '=')) == NULL && (c=strchr(sbuf, ':')) == NULL) continue;
    *c++ = 0;
    if (stricmp(sbuf, key)) continue;
    if (stristr(key, "time") != NULL)
    {
      return parseTineTimeString(c);
    }
    return strtod(c, NULL);
  }
  return 0;
}
#define assertIsSynchronous(c) if ((c) == NULL || BASEMODE((c)->status.mode) != CM_SINGLE) return not_allowed
int HISTQueryFunction(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  SINT32 numread=0, nr=0;
  char *c, Property[PROPERTY_NAME_SIZE];
  time_t stop=DATATIME_NOW; /* default values */
  double dstop = (double)stop, dstart = 0;
  float lwr=0,upr=0;
  int n,slen,fmt=CF_NULL,fsiz=0,cc=0,rc=0,index=0,fieldindex=0,sample=0,rs=(int)(cl->dataSizeOut),retSize=rs;
  int sysstop=0,sysstart=0;
  int isSnapshot=FALSE,callSTS=TRUE,isWrite=FALSE;
  int onePassOnly=FALSE,keepSizeInBytes=FALSE,useusrstamp=FALSE;
  SINT32 *lptr;
  UINT32 isiz = 0;
  double *dptr;
  ExportListStruct *el;
  PrpQueryStruct *prpqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  BYTE *dataptr = (BYTE *)dataOut;
  HstReqType hreq;
  HstRngType hrng;
  touchStkMetaData(cl);
  assertIsSynchronous(cl->clnLst);
  for (el=ExportList; el != NULL; el=el->next)
  { /* check for server redirection from possible property query functions ... */
    if (strncmp(con->EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (el->XEqmPropertyQueryFcn != NULL &&
        el->XEqmPropertyQueryFcn(con->EqmDeviceName,&xpqs) == -server_redirection) ccerr(server_redirection);
    if (el->EqmPropertyQueryFcn != NULL &&
        el->EqmPropertyQueryFcn(con->EqmDeviceName,&prpqs) == -server_redirection) ccerr(server_redirection);
    break;
  }
  /* proceed ...*/
  isWrite = (con->EqmAccess&CA_WRITE);
  strncpy(Property,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(Property,".AR")) != NULL)  
  { /* pass along to central archiver */
    *c = 0;
    return getRemoteHistoryServer(con->EqmName,Property);
  }
  if (strstr(Property,".HIST@") != NULL || strstr(Property,".HST@") != NULL)
    isSnapshot = TRUE;
  if ((c=strstr(Property,".HIST")) == NULL &&
      (c=strstr(Property,".HST")) == NULL) return illegal_property;
  *c = 0;
  if (rs == 0) return illegal_read_write;
  if ((c=strstr(Property,".RB")) != NULL)  
  {
    if (!strcmp(c,".RBAVE")) { index = 0; *c = 0; callSTS = FALSE; }
    else if (!strcmp(c,".RBMAX")) { index = 2; *c = 0; callSTS = FALSE; }
    else if (!strcmp(c,".RBMIN")) { index = 3; *c = 0; callSTS = FALSE; }
  }
  switch (fmt = LFMT(con->EqmFormatOut))
  {
    case CF_HISTORY:
      if (isVariableLengthFormat(ftoi(con->strTagOut))) 
      {
        onePassOnly = TRUE;
        keepSizeInBytes = TRUE;
        break;
      }
      if ((fsiz=getCarriedFormatSize(con->strTagOut)) <= 0) ccerr(-fsiz);
      fsiz += HISTORY_HDR_SIZE;
      rs /= fsiz; /* remap -> downsize */
      break;
    case CF_STRUCT:
      fsiz = GetStructSize(con->strTagOut);
      break;
    default:
      fsiz = fmtsizeof(fmt);
      break;
  }
  n = rs;
  if (con->EqmSizeIn > 0)
  {
    switch (LFMT(con->EqmFormatIn))
    {
      case CF_INT32:
        lptr = (SINT32 *)dataIn;
        switch (con->EqmSizeIn)
        {
          case 8: fieldindex = (int)lptr[7];
                  if (fieldindex < 0) fieldindex = 0;
          case 7: if ((int)lptr[6] != 0) useusrstamp = TRUE;
          case 6: sysstop = (int)lptr[5];
                  if (gCycleNumber == 0) sysstop = 0;
                  if (sysstop > (SINT32)gCycleNumber) sysstop = gCycleNumber;
          case 5: sysstart = (int)lptr[4];
                  if (gCycleNumber == 0) sysstart = 0;
          case 4: sample = (int)lptr[3];
          case 3: index = (int)lptr[2];
          case 2: if (lptr[1] < (SINT32)stop) dstop = (double)lptr[1];
          case 1: dstart = (double)lptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_DOUBLE:
        dptr = (double *)dataIn;
        switch (con->EqmSizeIn)
        {
          case 8: fieldindex = (int)dptr[7];
                  if (fieldindex < 0) fieldindex = 0;
          case 7: if ((int)dptr[6] != 0) useusrstamp = TRUE;
          case 6: sysstop = (int)dptr[5];
                  if (gCycleNumber == 0) sysstop = 0;
                  if (sysstop > (SINT32)gCycleNumber) sysstop = gCycleNumber;
          case 5: sysstart = (int)dptr[4];
                  if (gCycleNumber == 0) sysstart = 0;
          case 4: sample = (int)dptr[3];
          case 3: index = (int)dptr[2];
          case 2: if (dptr[1] < stop) dstop = dptr[1];
          case 1: dstart = dptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_CHAR:
      case CF_STRING:
      case CF_KEYVALUE:
        { /* n.b. sizeIn => size in bytes/chars */
          int nkv = 0;
          char *kvb[10], *sbuf, **kv;
          if (LFMT(con->EqmFormatIn) == CF_CHAR)
          { /* inpt = single string (array of characters: tokenize it ... */
            char *tok = "\n,";
            if ((sbuf=alloca(con->EqmSizeIn + 1)) == NULL) ccerr(out_of_local_memory);
            strncpy(sbuf, dataIn, con->EqmSizeIn);
            for (c = strtok(sbuf, tok); c != NULL; c = strtok(NULL, tok))
            {
              if (strlen(c) == 0) continue;
              kvb[nkv++] = c;
            }
            kv = kvb;
          }
          else
          { /* Stock/Meta properties don't call prepAdjustableArrayInputPreCall, prepAdjustableArrayInputPostCall */
            nkv = atoi(con->strTagIn);
            kv = (char **)cl->dataPtr;
          }
          if (kv == NULL) ccerr(invalid_data);
          dstart = assignKeyValue("starttime", kv, nkv);
          dstop = assignKeyValue("stoptime", kv, nkv);
          index = (int)assignKeyValue("index", kv, nkv);
          sample = (int)assignKeyValue("sample", kv, nkv);
          sysstart = (int)assignKeyValue("startstamp", kv, nkv);
          sysstop = (int)assignKeyValue("stopstamp", kv, nkv);
          fieldindex = (int)assignKeyValue("fieldindex", kv, nkv);
          useusrstamp = (int)assignKeyValue("userstamp", kv, nkv);
        }
        break;
      case CF_FWINDOW:
        dstart = (double)((FWINDOW *)dataIn)[0].i1val;
        dstop = (double)((FWINDOW *)dataIn)[0].i2val;
        lwr = ((FWINDOW *)dataIn)[0].f1val;
        upr = ((FWINDOW *)dataIn)[0].f2val;
        if (lwr < upr) sample = -1;
        break;
      case CF_WINDOW:
        dstart = (double)((DWINDOW *)dataIn)[0].t1;
        dstop = (double)((DWINDOW *)dataIn)[0].t2;
        sample = ((DWINDOW *)dataIn)[0].i2val;
        break;
      default:
        return illegal_format;
    }
    if (dstart < 0)
    { /* signal to take the last 'dstart' values from short-term memory only */
      dstop = (double)stop;
      dstart = (double)getStartTimeFromDepth(con->EqmName,Property,con->EqmDeviceName,(int)dstart);
    }
    else if ((dstop-dstart) < 0.01) 
    { /* also apply this check here */
      isSnapshot = TRUE;
    }
  }
  else
  {
    if (n == 1) 
    { /* the last stored value */
      dstart = dstop; 
    }
    else
    {
      dstart = (double)getStartTimeFromDepth(con->EqmName,Property,con->EqmDeviceName,n);
    }
    if (dstart < 0) ccerr((int)(-dstart));
  }
  gHistTimeStamp = gServerDataTimeStamp;
  if (dstart < 0) dstart = 0;
  if (dstop < dstart) dstop = (double)time(NULL);
  if (isSnapshot) 
  { 
    if (dstart < 1.0) return argument_list_error;
    /* if (fmt > CF_FLOAT) return illegal_format; <- was once necessary ? */
    if (sysstart > 0) sysstop = sysstart; else dstop = dstart;
    /* snapshot read doesn't use sample, index or recsread ... */
    sample = 0;
    index = 0;
    lwr = upr = 0.0;
  }
  hreq.eqm = con->EqmName; hreq.prp = Property; hreq.dev = con->EqmDeviceName;
  hreq.index = index; hreq.sample = sample; hreq.fieldindex = fieldindex;
  hrng.lwrLimit = lwr;  hrng.uprLimit = upr;
  hrng.startstamp = sysstart; hrng.stopstamp = sysstop;
  hrng.starttime = dstart; hrng.stoptime = dstop;
  hrng.userstamp = useusrstamp;
  if (isWrite)
  {
    cc = SaveShortTermData(con->EqmName, Property, con->EqmDeviceName, hrng);
    if (cc != 0) return cc;
  }
retryLTS:
  numread = getLTSData(&hreq, hrng,dataptr,fmt,con->strTagOut,n);
  if (numread < 0) 
  {
    if (numread == -un_allocated && 
        (c=strstr(Property,".SCH")) != NULL &&
        (slen=(int)strlen(c)) > 3 && slen < 7)
    { /* decorated with .SCH, .SCHD or .SCHED -> try unscheduled property */
      *c = 0; goto retryLTS;
    }
    if (numread == -at_limit)
    {
      rc = at_limit; 
      cc = 0; 
      numread = 0;
    }
    else
    {
      msglog(-numread, "HISTQueryFunction: LTSData for %.64s (dev = %.64s, index = %d) -> %.32s",
        hreq.prp, hreq.dev, hreq.index, cc2str(-numread));
      ccerr((int)(-numread));
    }
  }
  if (numread > 0) 
  { /* found at least 1 record in long-term storage */
    if (!onePassOnly && !isSnapshot && numread < rs - 1)
    { /* there's still room */
      dataptr += numread * fsiz;
      dstart = gHistTimeStamp + 1.0;
      n -= numread;
      if (n < 2) callSTS = FALSE;
      hrng.starttime = dstart; /* reasign */
    }
    else 
    {
      ccerr(0);
    }
  }
  if (!isSnapshot && (dstop-dstart) < 0.01) callSTS = FALSE;
  if (callSTS)
  {
    int i = 0;
    do
    { /* possible semaphore issue */
      hreq.index = index; /* re-assign in case it got changed */
      nr = getSTSData(&hreq, hrng,dataptr,fmt,con->strTagOut,n);
    } while (nr == -semaphore_busy && ++i < 3);
    if (nr == -semaphore_busy) nr = 0;
    if (nr < 0)
    {
      msglog(-nr, "HISTQueryFunction: STSData for %.64s (dev = %.64s, index = %d) -> %.32s",
        hreq.prp, hreq.dev, hreq.index, cc2str(-nr));
      ccerr((int)(-nr));
    }
  }
  numread += nr;
  if (!keepSizeInBytes && numread < rs) { retSize = (UINT32)numread; rs = numread; }
err:
  if (fmt == CF_HISTORY && cc == 0) 
  {
    switch (ftoi(con->strTagOut))
    { 
      case CF_IMAGE: /* very special case */
        fsiz = 0;
        for (n=0; n<numread && (fsiz < rs); n++)
        { /* scan thru the bytes */
          memcpy(&isiz,&dataOut[HISTORY_HDR_SIZE+fsiz]+IMAGE_HDR_SIZE-sizeof(SINT32),sizeof(SINT32));
          isiz += HISTORY_HDR_SIZE + IMAGE_HDR_SIZE;
          cfHistorySwap(con->strTagOut,(BYTE *)&dataOut[fsiz],(BYTE *)&dataOut[fsiz],LTOR,isiz);
          fsiz += isiz;
        }
        retSize = (UINT32)fsiz;
        break;
      case CF_STRING:
      case CF_KEYVALUE:
        fsiz = 0;
        for (n=0; n<numread && (fsiz < rs); n++)
        { /* scan thru the bytes */
          cfHistorySwap(con->strTagOut,(BYTE *)&dataOut[fsiz],(BYTE *)&dataOut[fsiz],LTOR,1);
          fsiz += HISTORY_HDR_SIZE + (int)strlen((char *)&dataOut[HISTORY_HDR_SIZE+fsiz])+1;
        }
        retSize = (UINT32)fsiz;
        break;
      default: /* normal case */
        rs *= fsiz; /* remap -> upsize */
        cfHistorySwap(con->strTagOut,(BYTE *)dataOut,(BYTE *)dataOut,LTOR,rs);
        retSize = (UINT32)rs;
    }
  }
  assignStkMetaData(cl,gHistTimeStamp,gHistSysStamp,gHistUsrStamp,retSize);
  if (!isSnapshot && rs == 1 && rc != 0) 
  { /* a number query and there's no LTS and the range exceeds the STS */
    cc = rc | CE_SENDDATA;
  }
  return cc;
}
#ifdef WIN32
int gCntNotPosted = 0;
int getNotPostedCount(ContractListStruct *cl,char *dataIn,char *dataOut)
{
  CONTRACT *con=&cl->contract;
  dataIn = dataIn;
  if (con->EqmSizeOut != 1) return illegal_read_write;
  touchStkMetaData(cl);
  switch (LFMT(con->EqmFormatOut))
  {
    case CF_INT16:
    case CF_UINT16:
      *(SINT16 *)dataOut = (SINT16)gCntNotPosted; break;
    case CF_INT32:
    case CF_UINT32:
      *(SINT32 *)dataOut = gCntNotPosted; break;
    case CF_FLOAT:
      *(float *)dataOut = (float)gCntNotPosted; break;
    case CF_DOUBLE:
      *(double *)dataOut = (double)gCntNotPosted; break;
    default: return illegal_format;
  }
  return 0;
}
#endif
StockPropertyStructType gStockProperty[] =
{
     {"PROPS",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPS",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
     {"PROPS",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPS",20,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPS",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPS",10,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE,FALSE},
#    endif
     {"PROPS",12,1,CF_STRUCT,CF_NAME64,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPS",12,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE,FALSE},
     {"NPROPS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery,FALSE,FALSE},
     {"PROPERTIES",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPERTIES",12,1,CF_STRUCT,CF_NAME64,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPERTIES",12,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PRPQSr4","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPERTIES",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
     {"PROPERTIES",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPERTIES",20,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPERTIES",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE,FALSE},
     {"PROPERTIES",10,PROPERTY_NAME_SIZE,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery,FALSE,FALSE},
     {"METAPROPERTIES",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","valid meta properties of exported Properties",metapropsQuery,FALSE,FALSE},
     {"METAPROPERTIES",12,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PRPQSr4","valid meta properties of exported Properties",metapropsQuery,FALSE,FALSE},
     {"METAPROPS",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","valid meta properties of exported Properties",metapropsQuery,FALSE,FALSE},
     {"METAPROPS",12,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PRPQSr4","valid meta properties of exported Properties",metapropsQuery,FALSE,FALSE},
#    endif
     {"PROPERTY.STATS",512,0,CF_NAME64I,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties call statistics",propStatsQuery,FALSE,FALSE },
     {"NPROPERTIES",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery,FALSE,FALSE},
     {"DEVICES",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","Device Names",deviceQuery,FALSE,FALSE},
     {"DEVICES",1,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"DEVQSr5","Device Name Information",deviceQuery,FALSE,FALSE},
     {"NDEVICES",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of Devices",ndeviceQuery,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
     {"DEVDESCRIPTION",1,0,CF_NAME64,CF_NAME64,CA_READ|CA_WRITE,AT_CHANNEL,"","Description for the specified device",deviceDescQuery,FALSE,FALSE},
     {"DEVLOCATION",1,1,CF_NAME64,CF_NAME64,CA_READ|CA_WRITE,AT_SINGLE,"","Location for the specified device",deviceLocQuery,FALSE,FALSE},
     {"DEVMASK",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_SINGLE,"","system mask for the specified device",deviceMaskQuery,FALSE,FALSE},
     {"DEVGROUP.MEMBERS",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","members list for the specified device group",deviceGroupQuery,FALSE,FALSE},
     {"DEVGROUPS",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","list registered device groups",deviceGroupsQuery,FALSE,FALSE},
#    ifndef DOOCSSVR
     {"SYS_MASK",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_SINGLE,"","system mask for the specified device (alias for DEVMASK)",deviceMaskQuery,FALSE,FALSE},
#    endif
     {"ZPOSITION",1,1,CF_FLOAT,CF_FLOAT,CA_READ|CA_WRITE,AT_SINGLE,"","Z position for the specified device",deviceZPosQuery,FALSE,FALSE},
     {"DEVONLINE",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_SINGLE,"","on-line flag for the specified device",deviceOnlineQuery,FALSE,FALSE},
     {"DEVREGION",1,1,CF_NAME64,CF_NAME64,CA_READ|CA_WRITE,AT_SINGLE,"","assigned region tag for the specified device",deviceRegionQuery,FALSE,FALSE},
     {"DEVREGION",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_SINGLE,"","assigned region code for the specified device",deviceRegionQuery,FALSE,FALSE},
#    endif
     {"FAILOVERSTATE",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","Current Failover State",failoverStateQuery,FALSE,FALSE},
     {"ALARMS",256,2,CF_STRUCT,CF_INT32,CA_READ,AT_SINGLE,"AMSr4","Total Current Alarm List",alarmsQuery,FALSE,FALSE},
     {"ALARMS",512,2,CF_STRUCT,CF_INT32,CA_READ,AT_SINGLE,"AMS","Total Current Alarm List",alarmsQuery,FALSE,FALSE},
     {"NALARMS",5,0,CF_INT32,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Current Alarm Information",nalarmsQuery,FALSE,FALSE},
     {"ALMDEFS",48,1,CF_STRUCT,CF_STRUCT,CA_READ|CA_WRITE,AT_SINGLE,"ADSr4","Defined Alarm Descriptions",alarmDefsQuery,FALSE,FALSE},
     {"ALMDEFS",32,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"ADS","Defined Alarm Descriptions",alarmDefsQuery,FALSE,FALSE},
     {"NALMDEFS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of Defined Alarms",nalarmDefsQuery,FALSE,FALSE},
     {"ALARMSEXT",256,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMSr4","Specified Current Alarm List",alarmsQuery,FALSE,FALSE}, 
     {"ALARMSEXT",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Specified Current Alarm List",alarmsQuery,FALSE,FALSE}, 
     {"ALMCOLLAPSE",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_SCALAR,"","Current Alarm Collapse Window",alarmWindowQuery,FALSE,FALSE},
     {"ALMOSCILLATION",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_SCALAR,"","Current Alarm Oscillation Window",alarmWindowQuery,FALSE,FALSE},
     {"ALARMSFLUSH",0,0,CF_NULL,CF_NULL,CA_WRITE,AT_UNKNOWN,"","Remove all alarms",alarmFlush,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
     {"ALMFILTERS",16,1,CF_STRUCT,CF_STRUCT,CA_READ | CA_WRITE,AT_SINGLE,"Filt","Current Alarm filters",almFilterOuery,FALSE,FALSE},
     {"HSTFILTERS",16,1,CF_STRUCT,CF_STRUCT,CA_READ | CA_WRITE,AT_SINGLE,"Filt","Current History filters",hstFilterQuery,FALSE,FALSE},
     {"NALMWATCH",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of alarms in watch table",nalarmWatchQuery,FALSE,FALSE},
     {"ALMWATCHTBL",32,0,CF_STRUCT,CF_STRUCT,CA_READ|CA_WRITE,AT_SINGLE,"AWSr4","Current Alarm Watch Table",alarmWatchQuery,FALSE,FALSE},
     {"ALMWATCHTBL",32,0,CF_STRUCT,CF_STRUCT,CA_READ | CA_WRITE,AT_SINGLE,"AWSr5","Current Alarm Watch Table",alarmWatchQuery,FALSE,FALSE},
     {"HISTORY.CMT",512,512,CF_TEXT,CF_TEXT,CA_READ|CA_WRITE,AT_SINGLE,"","set/get local annotated record (device=Timestamp)",localAnnotatedCMTQuery,FALSE,FALSE},
     {"HISTORY.CMTS",100,2,CF_NAME64,CF_INT32,CA_READ,AT_SINGLE,"","get local annotated records (from to)",localAnnotatedCMTSQuery,FALSE,FALSE},
     {"HISTORY.CMTS",100,2,CF_STRING,CF_INT32,CA_READ,AT_SINGLE,"","get local annotated records (from to)",localAnnotatedCMTSQuery,FALSE,FALSE},
     {"HISTORY.CMTS",100,2,CF_INT32,CF_INT32,CA_READ,AT_SINGLE,"","get local annotated records (from to)",localAnnotatedCMTSQuery,FALSE,FALSE},
#    endif
     {"MESSAGE",0,256,CF_NULL,CF_TEXT,CA_WRITE,AT_SINGLE,"","Accept Message",msgDisplay,FALSE,FALSE},
     {"ACTIVITY",1,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AQS","Current Server Activity",activityQuery,FALSE,FALSE},
     {"ACTIVITY",68,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"","Current Server Activity",activityQuery,FALSE,FALSE},
     {"CONTRACTS",12,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CTQSr4","Current Contract List",contractsQuery,FALSE,FALSE},
     {"CONTRACTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CONQS","Current Contract List",contractsQuery,FALSE,FALSE},
     {"CONTRACTS",920,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Contract List",contractsQuery,FALSE,FALSE},
     {"CLIENTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"ClnInfoQS","Current Client List",clientsQuery,FALSE,FALSE },
     {"CLIENTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CLNQS","Current Client List",clientsQuery,FALSE,FALSE},
     {"FOLLOWERS",64,1,CF_NAME64,CF_NAME64,CA_READ|CA_WRITE,AT_UNKNOWN,"","Current Followers List",followersQuery,FALSE,FALSE},
     {"CONNECTIONS",12,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CONTBLr4","Current Connection Table",contblQuery,FALSE,TRUE},
#    ifndef SMALL_TINE_LIB
     {"SRVEXIT",0,1,CF_NULL,CF_INT32,CA_WRITE,AT_UNKNOWN,"","Stop Server (when enabled)",srvExit,FALSE,TRUE},
#    endif
     {"ACCESSLOCK",0,2,CF_NULL,CF_INT16,CA_WRITE,AT_UNKNOWN,"","lock type/lock duration",lockAccess,FALSE,FALSE},
     {"ACCESSLOCK",3,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","lock owner/address",lockAccess,FALSE,FALSE},
     {"ADDUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add User to Access List",addRegisteredUser,FALSE,FALSE},
     {"DELUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Remove User from Access List",rmvRegisteredUser,FALSE,FALSE},
     {"USERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read User Access List",getRegisteredUserList,FALSE,FALSE},
     {"DENIEDUSERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read Denied User List",getDeniedUserList,FALSE,FALSE},
     {"NUSERS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of Users with Access",getNumRegisteredUsers,FALSE,FALSE},
     {"SRVADMINS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Registered Administrators",getAdminsUserList,FALSE,FALSE},
     {"STOCKPROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE,TRUE},
     {"STOCKPROPS",12,64,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PRPQSr4","Stock Properties",stockpropsQuery,FALSE,TRUE},
     {"STOCKPROPS",68*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE,TRUE},
#    ifndef SMALL_TINE_LIB
     {"STOCKPROPS",188*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE,TRUE},
     {"STOCKPROPS",188*10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery,FALSE,TRUE},
#    endif
     {"STOCKPROPS",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Stock Properties",stockpropsQuery,FALSE,TRUE},
     {"NSTOCKPROPS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of Stock Properties",nstockpropsQuery,FALSE,TRUE},
#    ifndef SMALL_TINE_LIB
     {"IPNETS",256,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Read IP Access List",getRegisteredIPNetList,FALSE,FALSE},
     {"NIPNETS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of IP Nets with Access",getNumRegisteredIPNets,FALSE,FALSE},
     {"ADDIPNET",0,32,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add IP Net to Access List",addRegisteredIPNet,FALSE,FALSE},
     {"DELIPNET",0,32,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Delete IP Net from Access List",rmvRegisteredIPNet,FALSE,FALSE},
     {"IPXNETS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read IPX Access List",getRegisteredIPXNetList,FALSE,FALSE},
     {"NIPXNETS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of IPX Nets with Access",getNumRegisteredIPXNets,FALSE,FALSE},
     {"DEBUGLEVEL",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","get debug level",setDebugLevel,FALSE,TRUE},
     {"DEBUGLEVEL",1,1,CF_INT16,CF_INT16,CA_READ|CA_WRITE,AT_SCALAR,"","get/set debug level",setDebugLevel,FALSE,TRUE},
     {"LOGCOMMANDS",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","read en/dis-able command logging flag",setLogCommands,FALSE,TRUE},
     {"LOGCOMMANDS",1,1,CF_INT16,CF_INT16,CA_READ|CA_WRITE,AT_SCALAR,"","en/dis-able command logging",setLogCommands,FALSE,TRUE},
     {"LOGDEPTH",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","get log depth in lines",setLogDepth,FALSE,TRUE},
     {"LOGDEPTH",1,1,CF_INT16,CF_INT16,CA_READ|CA_WRITE,AT_SCALAR,"","get/set log depth in lines",setLogDepth,FALSE,TRUE},
#    endif
     {"LOGFILE",8000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get log file",getFecLog,TRUE,TRUE},
     {"SRVVERSION",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","TINE Server Version Number",getTINEVersion,FALSE,TRUE},
     {"SRVOS",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","General Server OS",getServerOS,FALSE,TRUE},
     {"SRVDESC",64,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Description",getServerDescription,FALSE,TRUE},
     {"SRVLOCATION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Location",getServerLocation,FALSE,TRUE},
     {"SRVSUBSYSTEM",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Subsystem",getServerSubsystem,FALSE,TRUE},
     {"SRVSTATS",16,0,CF_INT32,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Server Statistics counters",getServerStats,FALSE,TRUE},
     {"SRVSCKERRS",3,0,CF_INT32,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Server Socket error counters",getServerSckErrors,FALSE,TRUE},
     {"SRVSETTINGS",1,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"SrvSetQy","Server Configuration Settings",getServerSettings,FALSE,TRUE},
#    ifndef SMALL_TINE_LIB
     {"SRVALIASLIST",512,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","Alias List (Name/Alias pairs)",getServerAliasList,FALSE,FALSE},
     {"ADDALIAS",0,2,CF_NULL,CF_NAME64,CA_WRITE,AT_SINGLE,"","Add Alias to alias List (Name/Alias pair)",addAlias,FALSE,FALSE},
#    endif
     {"NHISTORIES",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of history properties",nhistoryRecordsQuery,FALSE,FALSE},
     {"HISTORIES",32,0,CF_NAME64,CF_NULL,CA_READ,AT_SINGLE,"","history properties",historyRecordsQuery,FALSE,FALSE},
     {"HISTORIES",32,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"HRSr4","history property information",historyRecordsQuery,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
     {"ADDHISTORY",0,1,CF_NULL,CF_STRUCT,CA_WRITE,AT_SINGLE,"HRSr4","add/edit a history record",historyRecordsAppend,FALSE,FALSE},
#    endif
     {"STRUCTFORMAT",512,1,CF_NAME64DBLDBL,CF_NAME16,CA_READ,AT_SINGLE,"","Registered structure information",taggedStructQuery,FALSE,TRUE},
     {"STRUCTURES",512,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","List of registered structures",structListQuery,FALSE,TRUE},
     {"NSTRUCTURES",1,0,CF_INT16,CF_NULL,CA_READ,AT_SCALAR,"","Number of registered structures",nstructListQuery,FALSE,TRUE},
     {"BITFIELDFORMAT",512,1,CF_NAME16I,CF_NAME16,CA_READ,AT_SINGLE,"","Registered bitfield information",bitfieldStructQuery,FALSE,TRUE},
     {"SRVSTARTTIME",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime,FALSE,TRUE},
     {"SRVSTARTTIME",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime,FALSE,TRUE},
     {"SRVBASELINE",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Baseline",getServerBaseline,FALSE,TRUE},
     {"SRVBASELINE",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","Server Baseline Time",getServerBaseline,FALSE,TRUE},
#    ifndef SMALL_TINE_LIB
     {"APPDATE",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile Time",getAppDate,FALSE,TRUE},
     {"APPDATE",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile  Time",getAppDate,FALSE,TRUE},
     {"APPVERSION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Version Number",getAppVersion,FALSE,TRUE},
     {"ENSADDR",5,0,CF_NAME64,CF_NULL,CA_READ,AT_UNKNOWN,"","known ens address list",getEnsList,FALSE,TRUE},
#    endif
     {"SRVSELFTEST",32000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get self-test file",getSelfTestFile,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
     {"SRVLOGFILES",32,0,CF_NAME64,CF_NULL,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList,FALSE,TRUE},
     {"SRVLOGFILES",32,64,CF_NAME64,CF_TEXT,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList,FALSE,TRUE},
     {"SRVLOGFILES",32,64,CF_NAME64I,CF_TEXT,CA_READ,AT_SCALAR,"","get log file list + file lengths",getFecLogFileList,FALSE,TRUE},
     {"SRVLOGFILE",32000,0,CF_TEXT,CF_TEXT,CA_READ|CA_WRITE,AT_SCALAR,"","get specific log file",getFecLogFile,TRUE,TRUE},
     {"SRVLOGFILE",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","get specific log file length",getFecLogFile,TRUE,TRUE},
     {"SRVBINFILE",32000,0,CF_BYTE,CF_TEXT,CA_READ|CA_WRITE,AT_SCALAR,"","get specific binary file",getFecFile,TRUE,TRUE},
#    endif
     {"SRVERRORTEXT",1,1,CF_NAME32,CF_INT32,CA_READ,AT_SCALAR,"","get registered error text",getErrorDescription,FALSE,TRUE },
     {"SRVERRORTEXT",128,1,CF_TEXT,CF_INT32,CA_READ,AT_SCALAR,"","get registered error long text",getErrorDescription,FALSE,TRUE },
     {"SRVERRORTEXT",32,0,CF_NAME32I,CF_NULL,CA_READ,AT_COLLECTION,"","get registered error information list",getErrorDescription,FALSE,TRUE },
     {"SRVLASTACCESS",5,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","get most recent access",getLastAccessList,FALSE,TRUE},
     {"SRVLASTACCESS",100,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"WRACCTBL","get command history",getLastAccessList,FALSE,FALSE},
     {"SRVCOMMANDS",100,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"WRACCTBL","get command history",getLastAccessList,FALSE,FALSE},
#    ifndef SMALL_TINE_LIB
#    ifndef FS_RDONLY
     {"MANIFEST",8000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get manifest file",getManifest,TRUE,TRUE},
     {"SRVMANIFESTPATH",132,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","location of server's manifest",getSrvManifestPath,FALSE,TRUE},
#    endif
     {"SRVCMDLINE",132,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","full command line used to start the server",getSrvCmdLine,FALSE,TRUE},
     {"SRVCWD",132,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","working directory in play when the server started",getSrvCwd,FALSE,TRUE},
     {"SRVADDR",5,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","get server address parameters",getSrvAddressParams,FALSE,TRUE},
     {"SRVPID",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","server's current process id",getSrvPid,FALSE,TRUE},
     {"SRVRESET",0,1,CF_NULL,CF_INT32,CA_WRITE,AT_UNKNOWN,"","Reset Server (when enabled)",srvReset,FALSE,TRUE},
     {"SRVINIT",0,0,CF_NULL,CF_NULL,CA_WRITE,AT_UNKNOWN,"","re-initialize Server (when enabled)",srvInit,FALSE,FALSE},
     {"SRVIDLE",1,1,CF_INT32,CF_INT32,CA_READ|CA_WRITE,AT_UNKNOWN,"","set/get server idle state",srvIdle,FALSE,FALSE},
     {"SRVREEXPORT",0,0,CF_NULL,CF_NULL,CA_WRITE,AT_UNKNOWN,"","re-export equipment modules with the ENS",srvReex,FALSE,FALSE},
     {"SRVGLOBALS",50,0,CF_NAME64DBLDBL,CF_NULL,CA_READ,AT_SINGLE,"","server globals table",globalsRecordsQuery,FALSE,TRUE},
#    endif
#    ifdef WIN32
     {"SRVNOTPOSTED",1,0,CF_INT32,CF_NULL,CA_READ,AT_SCALAR,"","Number Contracts not posted",getNotPostedCount,FALSE,FALSE},
#    endif
};
TINE_EXPORT int gNrStockProperties = (sizeof(gStockProperty)/sizeof(StockPropertyStruct));

HashXRefStruct **StockPropXRefTable = NULL;

int makeStockPropXRefTable(void)
{
  int k,idx;
  HashXRefStruct *xref;

  if ((StockPropXRefTable=(HashXRefStruct **)SystemCalloc(STOCKPROP_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_server_memory;
  for (k=0; k<gNrStockProperties; k++)
  {
    idx = (int)(ElfHash((unsigned char *)gStockProperty[k].prpName) % STOCKPROP_HASH_SIZE);
    for (xref=StockPropXRefTable[idx]; xref && xref->index != k; xref = xref->next)
    {
      if (!strnicmp(gStockProperty[k].prpName,gStockProperty[xref->index].prpName,PROPERTY_NAME_SHORTSIZE)) break;
    }
    if (xref == NULL) 
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
      xref->index = k;
      xref->next = StockPropXRefTable[idx];
      StockPropXRefTable[idx] = xref;
    }
  }
  return 0;
}
StockModifierStructType StockModifier[] =
{
  {".EGU",EGUQueryFunction,FALSE,FALSE},
  {".XEGU",EGUQueryFunction,FALSE,FALSE},
  {".MIN",EGUQueryFunction,FALSE,FALSE},
  {".MAX",EGUQueryFunction,FALSE,FALSE},
  {".XMIN",EGUQueryFunction,FALSE,FALSE},
  {".XMAX",EGUQueryFunction,FALSE,FALSE},
  {".HIST",HISTQueryFunction,TRUE,FALSE},
  {".HST",HISTQueryFunction,TRUE,FALSE},
  {".HIST@",HISTQueryFunction,TRUE,FALSE},
  {".HST@",HISTQueryFunction,TRUE,FALSE},
  {".ARCH",HISTQueryFunction,TRUE,FALSE},
  {".ARCH@",HISTQueryFunction,TRUE,FALSE},
  {".RBMAX",HRBQueryFunction,FALSE,FALSE},
  {".RBMIN",HRBQueryFunction,FALSE,FALSE},
  {".RBAVE",HRBQueryFunction,FALSE,FALSE},
  {".RBSUMMARY",HRBQueryFunction,FALSE,FALSE},
# ifndef SMALL_TINE_LIB
  {".ARC",HISTQueryFunction,TRUE,FALSE},
  {".ARC@",HISTQueryFunction,TRUE,FALSE},
  {".AR",HISTQueryFunction,TRUE,FALSE},
  {".AR@",HISTQueryFunction,TRUE,FALSE},
# endif
  {".NAM",NAMQueryFunction,FALSE,FALSE},
  {".DESC",DESCQueryFunction,FALSE,FALSE},
# ifndef SMALL_TINE_LIB
  {".DSC",DESCQueryFunction,FALSE,FALSE},
  {".URL",URLQueryFunction,FALSE,FALSE},
# endif
  {".BIT.",gateFunction,FALSE,TRUE},
  {".MASK.",gateFunction,FALSE,TRUE},
  {".GATE.",gateFunction,FALSE,TRUE},
  {".DMASK.",gateFunction,FALSE,TRUE},
  {".ONLINE",gateFunction,FALSE,FALSE},
};
int NrStockModifiers = (sizeof(StockModifier)/sizeof(StockModifierStruct));

int usesGateFunction(char *meta)
{
  int i;
  for (i=0; i<NrStockModifiers; i++)
  { /* run thru the list */
    if (strstr(StockModifier[i].modName,meta) != NULL)
    { /* this is a good enough check for our purposes */
      if (StockModifier[i].fcn == gateFunction) return TRUE;
      break;
    }
  }
  return FALSE;
}
int getMetaFormatAndSize(char *meta,short *fmt,int *siz)
{ /* internal usage: limited null pointer checking ... */
  if (meta == NULL) return ambiguous;
  if (strchr(meta,'@') != NULL) return ambiguous;
  if (strstr(meta,"DMASK") || !strcmp(meta,"ONLINE")) return ambiguous;
  if (!strcmp(meta,"HIST") || !strcmp(meta,"HST") || !strncmp(meta,"AR",2))
  {
    *fmt = CF_DBLDBL;
    *siz = 1000;
  }
  else if (!strcmp(meta,"EGU") || !strcmp(meta,"DESC") || !strcmp(meta,"DSC") || !strcmp(meta,"URL"))
  {
    *fmt = CF_NAME64;
    *siz = 1;
  }
  else if (!strcmp(meta,"MAX") || !strcmp(meta,"MIN") || !strcmp(meta,"XMAX") || !strcmp(meta,"XMIN"))
  {
    *fmt = CF_FLOAT;
    *siz = 1;
    return illegal_format;
  }
  else if (!strncmp(meta,"NAM",3))
  {
    *fmt = CF_NAME64;
    *siz = 1000;
    return illegal_data_size;
  }
  if (!strcmp(meta,"RBSUMMARY"))
  {
    *fmt = CF_DOUBLE;
    *siz = 4;
  }
  else if (!strncmp(meta,"RB",2))
  {
    *fmt = CF_DOUBLE;
    *siz = 1;
  }
  else
  {
    *fmt = CF_INT32;
    *siz = 1;
  }
  return 0;
}
short IsFecProperty(char *Property)
{
  int i = GetStockPropertyId(Property);
  if (i < 0) return FALSE;
  return gStockProperty[i].isFecProperty;
}
TINE_EXPORT short IsStockProperty(char *Property)
{
  if (Property == NULL) return EMPTY_PROP;
  if (strchr(Property,'*') != NULL) return STOCK_PROP; /* maps to "PROPERTIES" */
  return GetStockPropertyId(Property) < 0 ? EMPTY_PROP : STOCK_PROP;
}
short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty,int localOnly)
{
  int i;
  char *c, *ce;
  static char metaPropertyString[PROPERTY_NAME_SIZE+8];

  strncpy(metaPropertyString,Property,PROPERTY_NAME_SIZE);
  metaPropertyString[PROPERTY_NAME_SIZE] = 0;
  if ((c=strrchr(metaPropertyString,'.')) == NULL) return 0;
  if (strstr(Property,".NAM.NAM") != NULL) return 0; /* TODO: these are not yet meta-properties */
  for (i=0; i<NrStockModifiers; i++)
  {
    if (localOnly && !strncmp(StockModifier[i].modName,".AR",3)) continue; /* these don't count ! */
    if (StockModifier[i].isEmbedded)
    { /* e.g. .BIT.2 */
      if ((ce=strstr(metaPropertyString,StockModifier[i].modName)) == NULL) continue;
      c = ce;
    }
    else
    { /* normal case */
      if (stricmp(c,StockModifier[i].modName) != 0) continue;
      if (stricmp(c,".NAM") == 0)
      { /* special cases */
        if ((ce=strstr(metaPropertyString,".DMASK.")) != NULL) *ce = 0;
        if ((ce=strstr(metaPropertyString,".ONLINE")) != NULL) *ce = 0;
      }
      if (!stricmp(c,".HIST") || !stricmp(c,".HST"))
      {
        if ((ce=strstr(metaPropertyString,".RBAVE")) != NULL) *ce = 0;
        else if ((ce=strstr(metaPropertyString,".RBMAX")) != NULL) *ce = 0;
        else if ((ce=strstr(metaPropertyString,".RBMIN")) != NULL) *ce = 0;
      }
    }
    *c++ = 0;
    if (baseProperty != NULL) *baseProperty = metaPropertyString;
    if (metaProperty != NULL) *metaProperty = c;
    return (i+1);
  }
  return 0;
}
short IsStaticProperty(char *Property)
{
  char *p, prpstr[PROPERTY_NAME_SIZE+8];
  if (Property == NULL) return FALSE;
  if (!stricmp(Property,"SRVDESC")) return TRUE;
  if (!stricmp(Property,"SRVOS")) return TRUE;
  if (!stricmp(Property,"SRVADDR")) return TRUE;
  if (!stricmp(Property,"SRVPID")) return TRUE;
  if (!stricmp(Property,"STOCKPROPS")) return TRUE;
  if (!stricmp(Property,"NSTOCKPROPS")) return TRUE;
  if (!stricmp(Property,"APPDATE")) return TRUE;
  if (!stricmp(Property,"APPVERSION")) return TRUE;
  strncpy(prpstr,Property,PROPERTY_NAME_SIZE); 
  prpstr[PROPERTY_NAME_SIZE] = 0;
  if ((p=strrchr(prpstr,'.')) != NULL)
  {
    if (!stricmp(p,".EGU")) return TRUE;
    if (!stricmp(p,".XEGU")) return TRUE;
    if (!stricmp(p,".MIN")) return TRUE;
    if (!stricmp(p,".XMIN")) return TRUE;
    if (!stricmp(p,".MAX")) return TRUE;
    if (!stricmp(p,".XMAX")) return TRUE;
    if (!stricmp(p,".DESC")) return TRUE;
    if (!stricmp(p,".DSC")) return TRUE;
    if (!stricmp(p,".URL")) return TRUE;
  }
  return FALSE;
}
static int isGateableInteger(int fmt)
{
  switch (fmt)
  {
    case CF_BYTE:
    case CF_INT16:
    case CF_UINT16:
    case CF_INT32:
    case CF_UINT32:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
#define isGateableArray(p) ((p) != NULL && ((p)->prpArrayType & AT_CHANNEL) && (p)->devNames == NULL)
short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate)
{
  char *c;
  int bit;
  ExportPropertyListStruct *pl;
  
  if (baseProperty == NULL || gate == NULL) return GATE_NONE;
  strncpy(baseProperty,con->EqmProperty,PROPERTY_NAME_SIZE);
  if ((c=strstr(baseProperty,".BIT.")) != NULL)
  {
    *c = 0;
    c += 5;
    bit = atoi(c);
    if (bit < 0 || bit >= 8*fmtsizeof(con->EqmFormatOut)) return FALSE;
    *gate = 0x01 << bit;
    return isGateableInteger(LFMT(con->EqmFormatOut)) ? GATE_BOOLEAN : GATE_NONE;
  }
  if ((c=strstr(baseProperty,".MASK.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return isGateableInteger(LFMT(con->EqmFormatOut)) ? GATE_FILTER : GATE_NONE;
  }
  if ((c=strstr(baseProperty,".GATE.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return isGateableInteger(LFMT(con->EqmFormatOut)) ? GATE_BOOLEAN : GATE_NONE;
  }
  if ((c=strstr(baseProperty,".DMASK.")) != NULL)
  {
    if (strstr(&c[1],".NAM")) return GATE_NONE;
    *c = 0;
    c += 7;
    *gate = strtol(c,NULL,0);
    pl = GetPropertyListStruct(con->EqmName,baseProperty,con->EqmDeviceName);
    if (pl == NULL || pl->isReserved) return GATE_NONE;
    return isGateableArray(pl) ? GATE_ARRAY : GATE_NONE;
  }
  if ((c=strstr(baseProperty,".ONLINE")) != NULL)
  {
    if (strstr(&c[1],".NAM")) return GATE_NONE;
    *c = 0;
    c += 7;
    *gate = ~(DEVICE_OFFLINE);
    pl = GetPropertyListStruct(con->EqmName,baseProperty,con->EqmDeviceName);
    if (pl == NULL || pl->isReserved) return GATE_NONE;
    return isGateableArray(pl) ? GATE_ARRAY : GATE_NONE;
  }
  return GATE_NONE;
}
