/*
 * File syslib.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file syslib.c
 * \brief TINE System API routines:
 *
 * The API routines offered here generally refer to the TINE engine and have relevance for
 * both client-side and server-side programming.  Some routines such as _SystemCycle() are
 * required.  Others such as _SystemInit() are required for server operation.  Many routines
 * are optional and offer ways of setting and getting system timestamps for instance or making
 * entries into the server's log file.
 *
 */
#ifndef NO_DOXYGEN
#ifdef WIN32
/* needed for the WaitableTimer in nanosleep() re-enactment */
#ifndef VS_LEGACY_BUILD
# include "project.def"
#endif
# ifndef _WIN32_WINNT
# define _WIN32_WINNT 0x0400 
# endif
# include "windows.h"
# include "winbase.h"
# ifndef VS_LEGACY_BUILD
# include "psapi.h"
# endif
#endif
#include "project.def"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <math.h>
#if defined(UNIX) || defined(MACOS)
# include <dirent.h>
# include <pwd.h>
# ifdef _GNU_SOURCE
#  include <pthread.h>
# endif
#endif
#if defined(MACOS)
# include <mach/clock.h>
# include <mach/mach.h>
#endif
#if defined(SOLARIS)
# ifndef CLOCK_MONOTONIC
#   define CLOCK_MONOTONIC CLOCK_REALTIME
# endif
#endif
#ifdef DOS
# include <signal.h>
# include <dos.h>
#endif
#if defined(WIN32) && !defined(UNDER_CE)
# include "io.h"
#endif
#include "prolog.h"
#include "srvcore.h"
#include "dbtools.h"
#include "errors.h"
#include "globals.h"
#include "hstlib.h"
#include "almlib.h"
#include "srvip.h"
#include "toolkit.h"
#include "ttycmd.h"
#include "threader.h"
#ifdef DOS
# include "watchdog.h"
#endif
#endif /* NO_DOXYGEN */

void FecSynchronize(void);
void initWatchDog(void);
void watchDog(void);
int rdate(char *tsHost);
time_t *tnull = NULL;
char dbgbuf[DBGBUF_SIZE];
TINE_EXPORT int gStartupDebug = 1;
int _gNeedWarmRestart = FALSE;
int _gReinitIdleTime = 0;
void _reinit(int idleTime);
int getAdmins(void);
extern int gInitTineServicesDone;
extern int gDelayInitTineServices;
extern int gClientIsClosing;
extern int DataTagging;
TINE_EXPORT int tineDebug = 0;
int errorDebug = 0;
int _sendToDbg = 0;
int NGdebug = 0; /* empty variable to avoid breaking code from previous Releases */
#if !defined(IPXNW) && !defined(IPXBSD)
BYTE ipxSrvSckB[2] = { 0, 0 };
#endif
int gCharEncoding = CE_ASCII;
/**
 * \internal
 * \brief Login or user name if a pure client, otherwise the fec name
 */
char gUserName[20] = "";
FILE *dbgfp = NULL;
WORD watchdogBaseAddress = 0x0000;
int MaxNumGlobals = GLBTBL_CAPACITY;
int MaxNumGCasts = GLBTBL_CAPACITY;
int GlbTblCapacity = GLBTBL_CAPACITY;
int GCastTblCapacity = GLBTBL_CAPACITY;
#ifndef GLOBAL_SYNCHRONIZATION
# define GLOBAL_SYNCHRONIZATION TRUE
#endif

#ifndef STD_TIME_STR
# define STD_TIME_STR "CET"
#endif
#ifndef DST_TIME_STR
# define DST_TIME_STR "CDT"
#endif
#ifndef STD_TIMEZONE_STR
# define STD_TIMEZONE_STR "TZ=GMT-1GDT"
#endif
#ifndef DLT_TIMEZONE_STR
# define DLT_TIMEZONE_STR "TZ=GMT-1GDT"
#endif
#ifndef STD_TIME_UTC_OFFSET
# define STD_TIME_UTC_OFFSET 0
#endif
#ifndef FECLOG_DEPTH
# define FECLOG_DEPTH 1000
#endif

char *ltzname[2] = { STD_TIME_STR, DST_TIME_STR };
char fecDbPath[STD_PATH_LENGTH] = "";
char *vFeclogBuffer = NULL;
int vFeclogBufferPos = 0;
int leftFeclogTime = FALSE;
int gLogfileAllowScan = LOG_SCAN_FEC;
char *getTimeServerHost(void);
int METFindDaylight(time_t t);
int NoDstFindDaylight(time_t t);
#ifdef VXWORKS
int (*findDaylightHook)(time_t) = NoDstFindDaylight;
#else
int (*findDaylightHook)(time_t) = METFindDaylight;
#endif
int heapCorrupted = FALSE;
int gUseMmfJumbos = FALSE;

TINE_EXPORT void (*cdiSystemInitHook)(void) = NULL;
TINE_EXPORT int (*cdiCloseLinkHook)(int) = NULL;
TINE_EXPORT char *(*getCdiLinkNameHook)(char *,char *) = NULL;
TINE_EXPORT int (*cdiExecLinkHook)(char *,char *,DTYPE *,DTYPE *,short,UINT16) = NULL;
TINE_EXPORT int (*cdiAttachLinkHook)(char *,char *,DTYPE *,DTYPE *,short,int,TCBFCNP,int,UINT32) = NULL;
TINE_EXPORT void (*cdiSystemFreeHook)(void) = NULL;

/**
 * \internal compare memory versus 0
 * \return 0 if 'cnt' bytes are all 0 else the location of the first non-zero byte
 */
int zerocmp(void *v,int cnt)
{
  BYTE *b = (BYTE *)v;
  int i;
  if (v == NULL) return -1;
  for (i=0;i<cnt;i++) if (b[i] != 0) return i+1;
  return 0;
}
int gCatchConsoleBreak = FALSE;
TINE_EXPORT void SetCatchConsoleBreak(int value)
{
  gCatchConsoleBreak = value;
}
TINE_EXPORT int GetCatchConsoleBreak(void)
{
  return gCatchConsoleBreak;
}
int gDieOnAddrInUse = TRUE;
TINE_EXPORT void SetDieOnAddressInUse(int value)
{
  gDieOnAddrInUse = value;
}
TINE_EXPORT int GetDieOnAddressInUse(void)
{
  return gDieOnAddrInUse;
}
int gDieAnotherDay = FALSE;
TINE_EXPORT void SetDieAnotherDay(int value)
{
  gDieAnotherDay = value;
}
int gDieOnSocketErrors = TRUE;
DIEFCNP dieFunction = NULL;
TINE_EXPORT void SetDieFunction(DIEFCNP fcn)
{
  dieFunction = fcn;
}
TINE_EXPORT void SetDieOnSocketError(int value)
{
  gDieOnSocketErrors = value;
}
TINE_EXPORT int GetDieOnSocketError(void)
{
  return gDieOnSocketErrors;
}
int gDieOnFecIsAlias = FALSE;
TINE_EXPORT void SetDieOnFecIsAlias(int value)
{
  gDieOnFecIsAlias = value;
}
TINE_EXPORT int GetDieOnFecIsAlias(void)
{
  return gDieOnFecIsAlias;
}
int gIsRunningAsServer = FALSE;
TINE_EXPORT void SetRunAsServer(void)
{
  gIsRunningAsServer = TRUE;
}
void fixLocalTimeSettings(void)
{
  char *ptr=NULL;
  if ((ptr=getenv("TINE_STD_TIME_STR")) != NULL ||
      (ptr=getenv("STD_TIME_STR")) != NULL)
  {
    ltzname[0] = ptr;
  }
  if ((ptr=getenv("TINE_DST_TIME_STR")) != NULL ||
      (ptr=getenv("DST_TIME_STR")) != NULL)
  {
    ltzname[1] = ptr;
  }
  if (!strcmp(ltzname[0],ltzname[1])) /* no dst ! */
  {
    findDaylightHook = NoDstFindDaylight;
  }
}
TINE_EXPORT UINT32 ElfHash(const unsigned char *key)
{
  UINT32 g, h = 0;
  char ukey; /* make all hashing convert to ucase */
  while (*key)
  {
    ukey = toupper(*key);
    h = (h << 4) + ukey;
    g = h & 0xF0000000L;
    if (g) h ^= g >> 24;
    h &= ~g;
    key++;
  }
  return h;
}
TINE_EXPORT int GetFormatSize(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_DOUBLE  : return 8;
    case CF_UINT16:
    case CF_INT16   : return 2;
    case CF_BYTE    : return 1;
    case CF_UINT32:
    case CF_INT32    : return 4;
    case CF_FLOAT   : return 4;
    case CF_STRUCT  : return 1;
    case CF_UINT64:
    case CF_INT64   : return 8;
    case CF_CHAR8   : return 8;
    case CF_CHAR16  : return 16;
    case CF_CHAR32  : return 32;
    case CF_CHAR48  : return 48;
    case CF_CHAR64  : return 64;
    case CF_CHAR16FI: return 24;
    case CF_SI      : return 6;
    case CF_FI      : return 8;
    case CF_II      : return 8;
    case CF_DI      : return 12;
    case CF_TDS     : return 12;
    case CF_HITTS   : return 6;
    case CF_CHAR8I  : return 12;
    case CF_CHAR16I : return 20;
    case CF_CHAR32I : return 36;
    case CF_CHAR48I : return 52;
    case CF_CHAR64I : return 68;
    case CF_USTRING : return 96;
    case CF_FFI     : return 12;
    case CF_FII     : return 12;
    case CF_IFFF    : return 16;
    case CF_IIII    : return 16;
    case CF_TTII    : return 16;
    case CF_IIFF    : return 16;
    case CF_FIFI    : return 16;
    case CF_XY      : return 8;
    case CF_TDSI    : return 12;
    case CF_ASPECTRUM: 
    case CF_GSPECTRUM:
    case CF_SPECTRUM: return 4;
    case CF_TEXT    : 
    case CF_KEYVALUE:
    case CF_STRING  : 
    case CF_XML     : return 1;
    case CF_BOOLEAN : return 4;
    case CF_BIT     : return 4;
    case CF_LNGDBL  : return 12;
    case CF_FLTDBL  : return 12;
    case CF_DBLTIME :
    case CF_DBLDBL  : return 16;
    case CF_NAMEDBLDBL : return 32;
    case CF_NAME16II : return 24;
    case CF_III     : return 12;
    case CF_NAME32DBLDBL : return 48;
    case CF_NAME64TIME:
    case CF_NAME64DBL : return 72;
    case CF_NAME64DBLDBL : return 80;
    case CF_IFFFNAME : return 32;
    case CF_DBLDBLDBL : return 24;
    case CF_NAME64DBLDBLDBL: return 88;
    case CF_BITFIELD8 : return 1;
    case CF_BITFIELD16 : return 2;
    case CF_BITFIELD32 : return 4;
    case CF_AIMAGE: return 1;
    case CF_IMAGE: return 1;
    case CF_HISTORY: return 1;
    case CF_MDA: return 1; /* actual format size is given by the embedded format */
    case CF_DEFAULT : return 1;
    default : break;
  }
  return 0;
}
int isIntegerFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_INT8:
    case CF_INT16:
    case CF_INT32:
    case CF_INT64:
    case CF_UINT16:
    case CF_UINT32:
    case CF_UINT64:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
TINE_EXPORT int IsIntegerFormat(int fmt)
{
  return isIntegerFormat((short)fmt);
}
int isNumberFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_INT8:
    case CF_INT16:
    case CF_INT32:
    case CF_INT64:
    case CF_UINT16:
    case CF_UINT32:
    case CF_UINT64:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_FLOAT:
    case CF_DOUBLE:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
TINE_EXPORT int IsNumberFormat(int fmt)
{
  return isNumberFormat((short)fmt);
}
int isSimpleFormat(short fmt)
{
  fmt = LFMT(fmt);
  if (fmt == CF_STRUCT) return FALSE;
  if (fmt <= CF_NAME16) return TRUE;
  if (fmt == CF_NAME32) return TRUE;
  if (fmt == CF_NAME48) return TRUE;
  if (fmt == CF_NAME64) return TRUE;
  if (fmt == CF_BOOLEAN) return TRUE;
  if (fmt >= CF_UINT16 && fmt <= CF_UINT64) return TRUE;
  return FALSE;
}
TINE_EXPORT int IsSimpleFormat(int fmt)
{
  return isSimpleFormat((short)fmt);
}
int isSimpleStringFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_TEXT:
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
TINE_EXPORT int IsSimpleStringFormat(int fmt)
{
  return isSimpleStringFormat((short)fmt);
}
int isStringFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_TEXT:
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_USTRING:
    case CF_UNAME:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
TINE_EXPORT int IsStringFormat(int fmt)
{
  return isStringFormat((short)fmt);
}
TINE_EXPORT int IsTaggableFormat(short fmt) {
	return isTaggableFormat(fmt);
}
int isTaggableFormat(short fmt)
{
  if (DataTagging) return TRUE;
  switch (LFMT(fmt))
  {
    case CF_STRUCT:
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
    case CF_HISTORY:
    case CF_AIMAGE:
    case CF_ASPECTRUM:
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_MDA:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
TINE_EXPORT int IsVariableLengthFormat(short fmt) 
{
	return isVariableLengthFormat(fmt);
}
int isVariableLengthFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_AIMAGE:
    case CF_IMAGE:
    case CF_GSPECTRUM:
    case CF_ASPECTRUM:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
int isQueueableFormat(short fmt)
{
  switch (LFMT(fmt))
  {
  case CF_AIMAGE:
  case CF_IMAGE:
  case CF_GSPECTRUM:
  case CF_ASPECTRUM:
    return FALSE;
  default:
    break;
  }
  return TRUE;
}
int isAdjustableLengthFormat(short fmt)
{
  switch (fmt)
  {
    case CF_KEYVALUE:
    case CF_STRING:
    case CF_AIMAGE:
    case CF_ASPECTRUM:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
int isTimeFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_DBLTIME:
    case CF_NAME64TIME:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
short getElevatedMcaFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_UINT16:
    case CF_INT16:
    case CF_UINT32:
    case CF_INT32: return CF_INTINT;
    case CF_FLOAT: return CF_FLTINT;
    case CF_DOUBLE: return CF_DBLDBL;
    case CF_NAME16: return CF_NAME16I;
    case CF_NAME32: return CF_NAME32I;
    case CF_NAME64: return CF_NAME64I;
    default: break;
  }
  return fmt;
}
short getPrimaryFormat(short fmt)
{
  switch (LFMT(fmt))
  {
    case CF_IIII:
    case CF_III:
    case CF_INTINT: return CF_INT32;
    case CF_FIFI:
    case CF_FFI:
    case CF_FII:
    case CF_XY:
    case CF_FLTINT: return CF_FLOAT;
    case CF_DBLDBLDBL:
    case CF_DBLTIME:
    case CF_DBLDBL: return CF_DOUBLE;
    case CF_NAME16FI:
    case CF_NAME16II:
    case CF_NAME16I: return CF_NAME16;
    case CF_NAME32I: return CF_NAME32;
    case CF_NAME48I: return CF_NAME48;
    case CF_NAME64DBLDBLDBL:
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL: 
    case CF_NAME64DBL:
    case CF_NAME64I: return CF_NAME64;
    default: break;
  }
  return fmt;
}
TINE_EXPORT int GetNumberSegments(int fmt)
{
  if (isSimpleFormat((short)fmt)) return 1;
  if (GetFormatHeaderSize(fmt) > 0) return 1;
  switch (LFMT(fmt))
  {
    case CF_BITFIELD8:
    case CF_BITFIELD16:
    case CF_BITFIELD32:
    case CF_BITFIELD64:
      return 1;
    case CF_NAME8I:
    case CF_NAME16I:
    case CF_NAME32I:
    case CF_NAME48I:
    case CF_NAME64I:
    case CF_NAME64DBL:
    case CF_FLTINT:
    case CF_INTINT:
    case CF_DBLDBL:
    case CF_FLTFLT:
      return 2;
    case CF_NAME16FI:
    case CF_NAME16II:
    case CF_INTFLTINT:
    case CF_FLTINTINT:
    case CF_FLTFLTINT:
    case CF_INTINTINT:
    case CF_DBLDBLDBL:
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      return 3;
    case CF_ADDRESS:
    case CF_FILTER:
    case CF_WINDOW:
    case CF_FWINDOW:
    case CF_FLTINTFLTINT:
    case CF_NAME64DBLDBLDBL:
      return 4;
    case CF_USTRING:
    case CF_UNAME:
      return 5;
    default:
      break;
  }
  return 1;
}
TINE_EXPORT int GetFormatHeaderSize(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_SPECTRUM: 
    case CF_ASPECTRUM: return SPECTRUM_HDR_SIZE;
    case CF_GSPECTRUM: return GSPECTRUM_HDR_SIZE;
    case CF_IMAGE:
    case CF_AIMAGE: return IMAGE_HDR_SIZE;
    case CF_MDA: return MDA_HDR_SIZE;
    default: return 0;
  }
}
TINE_EXPORT int GetFormatDataType(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_SPECTRUM:
    case CF_GSPECTRUM:
    case CF_ASPECTRUM: return CF_FLOAT;
    case CF_IMAGE:
    case CF_AIMAGE: return CF_BYTE;
    default: return fmt;
  }
}
TINE_EXPORT BYTE *GetFormatBodyPtr(void *dst, int fmt)
{
  if (dst == NULL) return NULL;
  switch (LFMT(fmt))
  {
    case CF_IMAGE:
    case CF_AIMAGE: return ((DIMAGE *)dst)->frameBuffer;
    case CF_ASPECTRUM: return (BYTE *)((ASPECTRUM *)dst)->spectBuffer;
    case CF_GSPECTRUM: return (BYTE *)((DGSPECTRUM *)dst)->spectBuffer;
    default: return NULL;
  }
}
TINE_EXPORT BYTE **GetFormatBodyPtrAddr(void *dst, int fmt)
{
  if (dst == NULL) return NULL;
  switch (LFMT(fmt))
  {
    case CF_IMAGE:
    case CF_AIMAGE: return &(((DIMAGE *)dst)->frameBuffer);
    case CF_ASPECTRUM: return (BYTE **)&(((ASPECTRUM *)dst)->spectBuffer);
    case CF_GSPECTRUM: return (BYTE **)&((DGSPECTRUM *)dst)->spectBuffer;
    default: return NULL;
  }
}
TINE_EXPORT UINT32 GetFormatBodyBufferSize(void *dst, int fmt)
{
  if (dst == NULL) return 0;
  switch (LFMT(fmt))
  {
    case CF_IMAGE:
    case CF_AIMAGE: return ((DIMAGE *)dst)->frameBufferSize;
    case CF_ASPECTRUM: return ((ASPECTRUM *)dst)->spectBufferSize;
    case CF_GSPECTRUM: return ((DGSPECTRUM *)dst)->spectBufferSize;
    default: return 0;
  }
}
TINE_EXPORT BYTE *GetFormatBodyBufferSizeAddr(void *dst, int fmt)
{
  if (dst == NULL) return NULL;
  switch (LFMT(fmt))
  {
    case CF_IMAGE:
    case CF_AIMAGE: return (BYTE *)&((DIMAGE *)dst)->frameBufferSize;
    case CF_ASPECTRUM: return (BYTE *)&((ASPECTRUM *)dst)->spectBufferSize;
    case CF_GSPECTRUM: return (BYTE *)&((DGSPECTRUM *)dst)->spectBufferSize;
    default: return NULL;
  }
}
TINE_EXPORT int GetFormatTypeSize(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_IMAGE:
    case CF_AIMAGE: return (int)sizeof(DIMAGE);
    case CF_SPECTRUM: return (int)sizeof(DSPECTRUM);
    case CF_ASPECTRUM: return (int)sizeof(ASPECTRUM);
    case CF_GSPECTRUM: return (int)sizeof(DGSPECTRUM);
    default: return fmtsizeof(fmt);
  }
}
TINE_EXPORT int GetFormatStandardCapacity(int fmt)
{
  switch (LFMT(fmt))
  {
    case CF_IMAGE:
    case CF_AIMAGE: return IMAGE_STD_FRAMESIZE;
    case CF_SPECTRUM: 
    case CF_ASPECTRUM: return SPECTRUM_STD_SIZE;
    case CF_GSPECTRUM: return GSPECTRUM_STD_SIZE;
    case CF_STRING: 
    case CF_KEYVALUE: return STRING_STD_CAPACITY;
    default: return 1;
  }
}
TINE_EXPORT short GetFormatFromString(char *fmt)
{
  return ftoi(fmt);
}
short ftoi(char *fmt)
{
  char *c,fmtstr[64];
  short ifmt;
  if (fmt == NULL) return CF_NULL;
  strncpy(fmtstr,fmt,64);
  if ((c=strchr(fmtstr,'.')) != NULL) *c = 0;
  strupr(fmtstr);
  if ((ifmt=getSimpleFormatType(fmtstr)) != CF_NULL) return ifmt;
  return getCompoundFormatType(fmtstr);
}
TINE_EXPORT char *GetFormatAsString(int fmt)
{
  return itof(fmt);
}
char *itof(int fmt)
{
  static char s[64];
  switch(LFMT(fmt))  
  {                 
    case CF_NULL: return "NULL";
    case CF_DEFAULT: return "DEFAULT";
    case CF_BYTE: return "BYTE";
    case CF_INT16: return "INT16";
    case CF_UINT16: return "UINT16";
    case CF_INT32: return "INT32";
    case CF_UINT32: return "UINT32";
    case CF_FLOAT: return "float";
    case CF_DOUBLE: return "double";
    case CF_TEXT: return "char";
    case CF_STRUCT: return "STRUCT";
    case CF_STRING: return "STRING";
    case CF_KEYVALUE: return "KEYVALUE";
    case CF_XML: return "xml";
    case CF_INT64: return "INT64";
    case CF_NAME8: return "NAME8";
    case CF_NAME16: return "NAME16";
    case CF_NAME32: return "NAME32";
    case CF_NAME48: return "NAME48";
    case CF_NAME64: return "NAME64";
    case CF_INTINT: return "INTINT";
    case CF_FLTINT: return "FLTINT";
    case CF_DBLTIME: return "DBLTIME";
    case CF_DBLDBL: return "DBLDBL";
    case CF_DBLDBLDBL: return "DBLDBLDBL";
    case CF_NAME8I: return "NAME8I";
    case CF_NAME16I: return "NAME16I";
    case CF_NAME32I: return "NAME32I";
    case CF_NAME48I: return "NAME48I";
    case CF_NAME64I: return "NAME64I";
    case CF_NAME16FI: return "NAME16FLTINT";
    case CF_NAME16II: return "NAME16INTINT";
    case CF_NAME16DBLDBL: return "NAME16DBLDBL";
    case CF_NAME32DBLDBL: return "NAME32DBLDBL";
    case CF_NAME64TIME: return "NAME64TIME";
    case CF_NAME64DBL: return "NAME64DBL";
    case CF_NAME64DBLDBL: return "NAME64DBLDBL";
    case CF_NAME64DBLDBLDBL: return "NAME64DBLDBLDBL";
    case CF_INTFLTINT: return "INTFLTINT";
    case CF_FLTINTINT: return "FLTINTINT";
    case CF_FLTFLTINT: return "FLTFLTINT";
    case CF_INTINTINT: return "INTINTINT";
    case CF_FILTER: return "FILTER";
    case CF_ADDRESS: return "ADDRESS";
    case CF_USTRING: return "USTRING";
    case CF_UNAME: return "UNAME";
    case CF_SPECTRUM: return "SPECTRUM";
    case CF_ASPECTRUM: return "ASPECTRUM";
    case CF_GSPECTRUM: return "GSPECTRUM";
    case CF_POINT: return "POINT";
    case CF_WINDOW: return "WINDOW";
    case CF_BOOLEAN: return "BOOLEAN";
    case CF_IIFF: return "IIFF";
    case CF_FIFI: return "FIFI";
    case CF_BITFIELD8: return "BITFIELD8";
    case CF_BITFIELD16: return "BITFIELD16";
    case CF_BITFIELD32: return "BITFIELD32";
    case CF_AIMAGE: return "AIMAGE";
    case CF_IMAGE: return "IMAGE";
    case CF_HISTORY: return "HISTORY";
    case CF_MDA: return "MDA";
    default: break;
  }
  sprintf(s,"%d",LFMT(fmt));
  return s;
}
int getCarriedFormatSize(char *tag)
{
  int fmt = ftoi(tag);
  struct structStruct *s;
  switch (fmt)
  {
    case CF_NULL:
      if ((s=findstruct(tag)) == NULL) return -invalid_structure_tag;
      return s->size;
    case CF_STRUCT:
    case CF_HISTORY:
      break;
    case CF_AIMAGE:
    case CF_IMAGE:
      return IMAGE_HDR_SIZE + IMAGE_STD_FRAMESIZE;
    case CF_ASPECTRUM:
    case CF_SPECTRUM:
      return SPECTRUM_HDR_SIZE + SPECTRUM_STD_SIZE;
    case CF_GSPECTRUM:
      return GSPECTRUM_HDR_SIZE + GSPECTRUM_STD_SIZE;
    case CF_STRING:
    case CF_KEYVALUE:
      return STRING_STD_CAPACITY;
    default:
      return fmtsizeof(fmt);
  }
  return -illegal_format;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int GetFieldInFormat(int fmt, char *dev, char *tag, int fieldIndex, int *foffset, int *ffmt)
{
  return getFieldInFormat(fmt,dev,tag,fieldIndex,foffset,ffmt);
}
int getFieldInFormat(int fmt, char *dev, char *tag, int fieldIndex, int *foffset, int *ffmt)
{
  if (fieldIndex < 0) return invalid_index;
  fmt = LFMT(fmt);
  if (isBitfieldFormat(fmt)) fmt = GetIntegerFormat((short)fmt);
  if (isSimpleFormat((short)fmt)) 
  {
    if (fieldIndex > 0) return invalid_index; 
    *foffset = 0;
    *ffmt = fmt;
    return 0;
  }
  switch (fmt)
  { /* the doublets ... */
    case CF_INTINT:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 4;
      *ffmt = CF_INT32;
      return 0;
    case CF_FLTINT:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 4;
      *ffmt = fieldIndex == 0 ? CF_FLOAT : CF_INT32;
      return 0;
    case CF_FLTFLT:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 4;
      *ffmt = CF_FLOAT;
      return 0;
    case CF_DBLDBL:
    case CF_DBLTIME:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 8;
      *ffmt = CF_DOUBLE;
      return 0;
    case CF_NAME8I:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 8;
      *ffmt = fieldIndex == 0 ? CF_NAME8 : CF_INT32;
      return 0;
    case CF_NAME16I:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 16;
      *ffmt = fieldIndex == 0 ? CF_NAME16 : CF_INT32;
      return 0;
    case CF_NAME32I:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 32;
      *ffmt = fieldIndex == 0 ? CF_NAME32 : CF_INT32;
      return 0;
    case CF_NAME48I:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 48;
      *ffmt = fieldIndex == 0 ? CF_NAME48 : CF_INT32;
      return 0;
    case CF_NAME64I:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 64;
      *ffmt = fieldIndex == 0 ? CF_NAME64 : CF_INT32;
      return 0;
    case CF_NAME64TIME:
    case CF_NAME64DBL:
      if (fieldIndex > 1) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 64;
      *ffmt = fieldIndex == 0 ? CF_NAME64 : CF_DOUBLE;
      return 0;
    /* the triplets ... */
    case CF_NAME16FI:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : fieldIndex == 1 ? 16 : 20;
      *ffmt = fieldIndex == 0 ? CF_NAME16 : fieldIndex == 1 ? CF_FLOAT : CF_INT32;
      return 0;
    case CF_NAME16II:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : fieldIndex == 1 ? 16 : 20;
      *ffmt = fieldIndex == 0 ? CF_NAME16 : CF_INT32;
      return 0;
    case CF_NAME16DBLDBL:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : fieldIndex == 1 ? 16 : 24;
      *ffmt = fieldIndex == 0 ? CF_NAME16 : CF_DOUBLE;
      return 0;
    case CF_NAME32DBLDBL:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : fieldIndex == 1 ? 32 : 40;
      *ffmt = fieldIndex == 0 ? CF_NAME32 : CF_DOUBLE;
      return 0;
    case CF_NAME64DBLDBL:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : fieldIndex == 1 ? 64 : 72;
      *ffmt = fieldIndex == 0 ? CF_NAME64 : CF_DOUBLE;
      return 0;
    case CF_INTINTINT:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = CF_INT32;
      return 0;
    case CF_FLTINTINT:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = fieldIndex == 0 ? CF_FLOAT : CF_INT32;
      return 0;
    case CF_FLTFLTINT:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = fieldIndex == 2 ? CF_INT32 : CF_FLOAT;
      return 0;
    case CF_DBLDBLDBL:
      if (fieldIndex > 2) return invalid_index;
      *foffset = fieldIndex*8;
      *ffmt = CF_DOUBLE;
      return 0;
    /* the quads ... */
    case CF_TTII: 
    case CF_INTINTINTINT:
      if (fieldIndex > 3) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = CF_INT32;
      return 0;
    case CF_FLTINTFLTINT:
      if (fieldIndex > 3) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = (fieldIndex == 0 || fieldIndex == 2) ? CF_FLOAT : CF_INT32;
      return 0;
    case CF_INTINTFLTFLT:
      if (fieldIndex > 3) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = (fieldIndex == 2 || fieldIndex == 3) ? CF_FLOAT : CF_INT32;
      return 0;
    case CF_INTFLTFLTFLT:
      if (fieldIndex > 3) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = fieldIndex == 0 ? CF_INT32 : CF_FLOAT;
      return 0;
    case CF_NAME64DBLDBLDBL:
      if (fieldIndex > 3) return invalid_index;
      *foffset = fieldIndex == 0 ? 0 : 64 + (fieldIndex-1)*8;
      *ffmt = fieldIndex == 0 ? CF_NAME64 : CF_DOUBLE;
      return 0;
    /* the quints ... */
    case CF_USTRING:
      if (fieldIndex > 4) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = (fieldIndex == 0 || fieldIndex == 3) ? CF_INT32 : (fieldIndex < 4) ? CF_FLOAT : CF_TEXT;
      return 0;
    case CF_UNAME:
      if (fieldIndex > 4) return invalid_index;
      *foffset = fieldIndex*4;
      *ffmt = fieldIndex == 0 ? CF_INT32 : (fieldIndex < 4) ? CF_FLOAT : CF_NAME16;
      return 0;
    case CF_STRUCT:
      {
       /* get the struct from the tag and run thru the fields */
       struct structStruct *ss;
       struct structFormat *sf;
       int i, nfields = 0;
       if (dev != NULL)
       {
         char *c, key[128];
         strncpy(key,dev,128);
         c = strpbrk(&key[1],DELIMITER); 
         if ((c=strpbrk(++c,DELIMITER)) != NULL) *c = 0; /* lop off the end */
         ss = findstructEx(tag,key);
       }
       else
       {
         ss = findstruct(tag);
       }
       if (ss == NULL) return invalid_structure_tag;
       for (sf=ss->s; sf != NULL; sf=sf->nxt) nfields++;
       if (fieldIndex >= nfields) return invalid_index;
       for (sf=ss->s,i=nfields-1; sf != NULL; sf=sf->nxt,i--)
       { /* the link list is reverse ordered  */
         if (i == fieldIndex)
         {
           if (!isSimpleFormat((short)sf->fmt)) return not_implemented;
           *foffset = sf->addr;
           *ffmt = sf->fmt;
           return 0;
         }
       }
       return invalid_structure_tag;
      }
    default:
      break;
  }
  return illegal_format;
}
#endif /* SMALL_TINE_LIB */
short getArrayType(char *fmt)
{
  char *c,typstr[64];
  short itype = AT_UNKNOWN;
  if (fmt == NULL) return itype;
  if ((c=strchr(fmt,'.')) != NULL)
  { /* possibly as a decorated format */
    c++;
  }
  else if (strnicmp(fmt,"DOUBLE",6) != 0)
  { /* take it straight up ... */
    c = fmt;
  }
  else
  {
    return itype;
  }
  strncpy(typstr,c,64); strupr(typstr);
  if (strstr(typstr,"BOOLEAN") != NULL) itype = AT_BOOLEAN;
  if (strstr(typstr,"DOUBLE") != NULL) itype = AT_DOUBLE;
  if (strstr(typstr,"CHANNEL") != NULL) itype |= AT_CHANNEL;
  else if (strstr(typstr,"SPECTRUM") != NULL) itype |= AT_SPECTRUM;
  else if (strstr(typstr,"TRACE") != NULL) itype |= AT_SPECTRUM;
  else if (strstr(typstr,"WAVEFORM") != NULL) itype |= AT_SPECTRUM;
  else if (strstr(typstr,"UNSTRUCTURED") != NULL) itype |= AT_UNSTRUCTURED;
  else if (strstr(typstr,"BLOB") != NULL) itype |= AT_UNSTRUCTURED;
  else if (strstr(typstr,"IMAGE") != NULL) itype |= AT_UNSTRUCTURED;
  else if (strstr(typstr,"STRUCTURED") != NULL) itype |= AT_STRUCTURED;
  else if (strstr(typstr,"TEXT") != NULL) itype |= AT_SCALAR;
  else if (strstr(typstr,"XML") != NULL) itype |= AT_SCALAR;
  return itype;
}
short getArrayRowLength(char *fmt)
{
  char *c,*ch,typstr[128];
  if (fmt == NULL) return 0;
  strncpy(typstr,fmt,128); strupr(typstr);
  if ((c=strstr(typstr,".DOUBLE.")) == NULL) return 0;
  c += 8; 
  if ((ch=strchr(c,'X')) != NULL) c = ++ch;
  return atoi(c);
}
TINE_EXPORT short GetArrayRowLengthFromString(char *fmt)
{
  return getArrayRowLength(fmt);
}
TINE_EXPORT short GetArrayTypeFromString(char *fmt)
{
  return getArrayType(fmt);
}
TINE_EXPORT char *GetArrayTypeAsString(int type)
{
  switch (type)
  {
    case AT_SCALAR: return "SCALAR";
    case AT_SINGLE: return "SINGLE";
    case AT_DOUBLE: return "DOUBLE";
    case AT_BOOLEAN: return "BOOLEAN";
    case AT_CHANNEL: return "CHANNEL";
    case AT_SPECTRUM: return "SPECTRUM";
    case AT_STRUCTURED: return "STRUCTURED";
    case AT_UNSTRUCTURED: return "UNSTRUCTURED";
    case AT_UNKNOWN:
    default:
      break;
  }
  return "UNKNOWN";
}
short mtoi(char *mod)
{ /* can't tokenize using "|+" due to original parsing conditions */
  short accessMode = CA_NULL;
  char *c, modStr[128];
  
  if (mod == NULL) return accessMode;
  strncpy(modStr,mod,128); modStr[127] = 0; strupr(modStr);
  if ((c=strstr(modStr,"XREAD")) != NULL)
  { /* check for this first in case of "READXREAD" (instead of "READ|XREAD") */
    accessMode |= CA_XREAD;
    *c = 0; c += 5;
    strcat(modStr,c);
  }
  if (strstr(modStr,"READ")) accessMode |= CA_READ;
  if (strstr(modStr,"WRITE")) accessMode |= CA_WRITE;
  if (strstr(modStr,"NETWORK")) accessMode |= (CA_READ|CA_NETWORK);
  if (strstr(modStr, "CONNECT")) accessMode |= CA_CONNECT;
  if (strstr(modStr,"ASYNC"))
  {
    accessMode &= ~(CA_WRITE); /* don't allow forced asynchronous WRITEs ! */
    accessMode = (CA_READ|CA_NOSYNC);
  }
  else if (strstr(modStr,"SYNC"))
  {
    accessMode |= (CA_READ|CA_SYNC);
  }
  if (strstr(modStr,"STATIC")) 
  {
    accessMode &= ~(CA_WRITE|CA_NETWORK|CA_NOSYNC); /* illogical ! */
    accessMode |= (CA_READ|CA_STATIC);
  }
  if (strstr(modStr,"FORCEOUTPUT")) 
  {
    accessMode |= (CA_READ|CA_FORCEOUTPUT);
  }
  if (strstr(modStr, "INDEPENDENT")) accessMode |= CA_INDEPENDENT;
  return accessMode;
}
char *itom(short accessMode)
{
  static char accessModeStr[64];

  accessModeStr[0] = 0;
  if (accessMode & CA_READ) strcat(accessModeStr,strlen(accessModeStr)?"|READ":"READ");
  if (accessMode & CA_XREAD) strcat(accessModeStr,strlen(accessModeStr)?"|XREAD":"XREAD");
  if (accessMode & CA_WRITE) strcat(accessModeStr,strlen(accessModeStr)?"|WRITE":"WRITE");
  if (accessMode & CA_CONNECT) strcat(accessModeStr, strlen(accessModeStr) ? "|CONNECT" : "CONNECT");
  if (accessMode & CA_NETWORK) strcat(accessModeStr,strlen(accessModeStr)?"|NETWORK":"READ|NETWORK");
  if (accessMode & CA_NOSYNC) strcat(accessModeStr,strlen(accessModeStr)?"|ASYNC":"READ|ASYNC");
  if (accessMode & CA_STATIC) strcat(accessModeStr,strlen(accessModeStr)?"|STATIC":"READ|STATIC");
  if (accessMode & CA_FORCEOUTPUT) strcat(accessModeStr,strlen(accessModeStr)?"|FORCEOUTPUT":"READ|FORCEOUTPUT");
  if (accessMode & CA_INDEPENDENT) strcat(accessModeStr, strlen(accessModeStr) ? "|INDEPENDENT" : "WRITE|INDEPENDENT");
  if (strlen(accessModeStr) == 0) strcat(accessModeStr,"NO");

  return accessModeStr;
}
TINE_EXPORT short GetAccessMode(char *mod)
{
  return mtoi(mod);
}
TINE_EXPORT char *GetAccessAsString(int accessMode)
{
  return itom((short)accessMode);
}
short getTransferMode(char *mod)
{
  char modstr[512];
  int m = 0;
  if (mod == NULL) return CM_NULL;
  strncpy(modstr,mod,512); modstr[511] = 0;
  strupr(modstr);
  m = getAccessMode(modstr);
  if (strstr(modstr,"CONNECT") != NULL) m |= CM_CONNECT;
  if (strstr(modstr,"NODUPLICATES") != NULL) m |= CM_NODUPLICATES;
  if (strstr(modstr,"SEEKPATTERN") != NULL) m |= CM_SEEKPATTERN;
  if (strstr(modstr,"MISSPATTERN") != NULL) m |= CM_MISSPATTERN;
  if (strstr(modstr,"STREAM") != NULL) m |= CM_STREAM;
  if (strstr(modstr,"BCAST") != NULL) m |= CM_BCAST;
  if (strstr(modstr,"MCAST") != NULL) m |= CM_MCAST;
  if (strstr(modstr,"NETWORK") != NULL) m |= CM_NETWORK;
  if (strstr(modstr,"MCAST") != NULL) m |= CM_MCAST;
  if (strstr(modstr,"USEONERROR") != NULL) m |= CM_USEONERROR;
  if (strstr(modstr,"GROUPED") != NULL) m |= CM_GROUPED;
  if (strstr(modstr,"SYNCGROUP") != NULL) m |= CM_SYNCGROUP;
  if (strstr(modstr,"WAIT") != NULL) m |= CM_WAIT;
  if (strstr(modstr,"SYNCNOTIFY") != NULL) m |= CM_SYNCNOTIFY;
  return m;
}
TINE_EXPORT short GetTransferMode(char *mod)
{
  return getTransferMode(mod);
}
TINE_EXPORT char *GetModeAsString(int transferMode)
{
  static char mstr[128];
  strncpy(mstr,modetoa[BASEMODE(transferMode)],16); mstr[16] = 0;
  if (transferMode & CM_CONNECT) strcat(mstr,"|CONNECT");
  else if (transferMode & CM_STREAM) strcat(mstr,"|STREAM");
  else if (transferMode & CM_NETWORK) strcat(mstr,"|NETWORK");
  if (transferMode & CM_GROUPED) strcat(mstr,"|GROUPED");
  if (transferMode & CM_SYNCNOTIFY) strcat(mstr,"|SYNCNOTIFY");
  if (transferMode & CM_WAIT) strcat(mstr,"|WAIT");
  return mstr;
}
TINE_EXPORT int isActiveTransferMode(short mode)
{
  switch (BASEMODE(mode))
  {
    case CM_SINGLE:
    case CM_DATACHANGE:
    case CM_TIMER:
    case CM_EVENT:
    case CM_RECEIVE:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
TINE_EXPORT int MakeMDA(DMDA *mda,char *cmt,int fmt,void *data,int ndim,...)
{
  int i,cc=0;
  MDX *pmdx;
  va_list argp;

  if (mda == NULL) return argument_list_error;
  va_start(argp,ndim);
  if (ndim < 1) ccerr(dimension_error);
  if (ndim > MAX_MDA_DIMS)
  {
    msglog(0,"multi-dimensional array number dimensions truncated to %d",MAX_MDA_DIMS);
    ndim = MAX_MDA_DIMS;
  }
  memset(mda,0,sizeof(DMDA));
  for (i=0; i<ndim; i++)
  {
    if ((pmdx=va_arg(argp,MDX *)) == NULL) ccerr(argument_list_error);
    mda->axis[i] = *pmdx;
  }
  strncpy(mda->cmt,cmt,80);
  mda->fmt = fmt;
  mda->ndim = ndim;
  mda->data.vptr = data;
err:
  va_end(argp);  
  return cc;
}
TINE_EXPORT void *GetMDAElemRef(DMDA *mda,...)
{
  int cc=0,p=0,rlen=1,fsiz,i,idx[MAX_MDA_DIMS];
  BYTE *dp=NULL;
  va_list argp;

  if (mda == NULL) return NULL;
  va_start(argp,mda);
  for (i=0; i<mda->ndim; i++)
  {
    if ((idx[i]=va_arg(argp,int)) < 0) ccerr(argument_list_error);
  }
  for (i=mda->ndim-1; i >= 0; i--)
  {
    p += idx[i] * rlen;
    rlen *= mda->axis[i].len;
  }
  fsiz = fmtsizeof(mda->fmt);
  dp = mda->data.bptr + p * fsiz;
err:
  va_end(argp);  
  if (cc != 0) msglog(cc,"Get MDA element reference : %.32s",cc2str(cc));
  return (void *)dp;
}
int MDAtoDTYPE(DTYPE *d,DMDA *mda,int bufferlength)
{
  int i,fsiz,len;
  if (d == NULL || mda == NULL) return argument_list_error;
  if (mda->ndim < 1) return invalid_data;
  d->dFormat = CF_MDA;
  if ((fsiz=fmtsizeof(mda->fmt)) < 0) return illegal_format;
  if (bufferlength < 1)
  {
    len = 1;
    for (i=0; i<mda->ndim; i++) len *= mda->axis[i].len;
    if (len == 0) return dimension_error;
    bufferlength = len;
  }
  d->dArrayLength = bufferlength;
  d->data.vptr = mda;
  if (mda->data.vptr == NULL)
  { /* no data buffer as yet allocated */
    if ((mda->data.bptr = (BYTE *)SystemCalloc(1,bufferlength)) == NULL)
      return out_of_local_memory;
  }
  strncpy(d->dTag,itof(mda->fmt),TAG_NAME_SIZE);
  return 0;
}
time_t gSystemStartupTime = 0;
double gDataTimeStamp; /* System Timestamp  */
double gServerDataTimeStamp; /* System Timestamp for server thread */
double gDataTimeStampOffset = 0.0; /* for external synchronization */
TINE_EXPORT double GetDataTimeStampOffset(void)
{
  return gDataTimeStampOffset;
}
/**
 * \internal
 *
 * \brief Determine if daylight time is set according to rules in force
 * in Germany as of 2000.
 * 
 * This function assumes daylight time from May 25 to October 25.  This
 * routine is only called on some OSes which do not have a time service
 * capable of supplying this information (DOS, NIOS, VxWorks, etc.).
 * 
 * \param t the UTC time for which the check on DST is to be made
 *
 * \todo make static
 */
int findDaylight(time_t t)
{
  return (findDaylightHook != NULL) ? findDaylightHook(t) : 0;
}
void hardDelay(int msec)
{
  struct timeval tv0,tv;
  gettimeofday(&tv0,(struct timezone *)NULL);
  tv = tv0;
  while (MSECS(tv,tv0) < msec) gettimeofday(&tv,(struct timezone *)NULL);
}
#ifndef FEC_SYNCHRONIZE_CLOCK
# define FEC_SYNCHRONIZE_CLOCK FALSE
#endif
short SychronizeFecClock = FEC_SYNCHRONIZE_CLOCK;
#ifdef DOS
void GetCommand(void);
#else
int readvFecLog(char *textbuf,int sizeInBytes);
int vfeclog(char *text);
#endif
void DataDump(char *txt,char *data,int siz,short fmt,short dir)
{
  ttyoutput("%s %d entries %s",txt,siz,dir == RTOL ? "received" : "returned");
  prepDataDump(dbgbuf,data,siz,fmt,TRUE);
}
void prepDataDump(char *buf,char *data,int siz,short fmt,int output)
{
  int i,k;
  char *sbuf;
  int ncols=16, nblanks=1;

  buf[0] = 0;
  switch (fmt)
  {
    case CF_TEXT:
    case CF_GSPECTRUM:
    case CF_ASPECTRUM:
    case CF_SPECTRUM:
    case CF_AIMAGE:
    case CF_IMAGE:
    case CF_HISTORY:
    case CF_MDA:
      dataValueToString(buf,(BYTE *)data,0,fmt,NULL,MIN(siz,DBGBUF_SIZE),TRUE);
      if (output) ttyoutput(buf);
      return;
    default:
      break;
  }
  if (!isSimpleFormat(fmt))
  {
    ncols = 4; nblanks = 2;
  }
  for (i=0; i<siz; i++)
  {
    sbuf = &buf[strlen(buf)];
    dataValueToString(sbuf,(BYTE *)data,i,fmt,NULL,0,TRUE);
    if (output && (i+1)%ncols == 0) 
    { 
      ttyoutput(buf); buf[0] = 0; 
    }
    else
    {
      for (k=0; k<nblanks; k++) strcat(buf," ");
    }
  }
  if (output && strlen(buf) > 0) ttyoutput(buf);
  return;
}
#define checkDataOOT(CF) \
  for (i=0; i<siz; i++) {\
    if (((CF *)d1)[i] < ((CF *)d2)[i] - t ||\
        ((CF *)d1)[i] > ((CF *)d2)[i] + t) return TRUE;\
  }

/** 
 * \intern  compares array d1[siz] against d2[siz] with given format and absolute tolerance
 */
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t)
{ 
  int fmtsize,i;

  fmtsize = fmtsizeof(fmt);
  if (!memcmp(d1,d2,siz*fmtsize)) return FALSE;
  if (t == 0) return TRUE;
  /* data have changed, but by how much? */
  switch (fmt)
  {
   case CF_DOUBLE:
       checkDataOOT(double);
       return FALSE;
   case CF_INT16:
       checkDataOOT(SINT16);
       return FALSE;
   case CF_UINT16:
       checkDataOOT(UINT16);
       return FALSE;
   case CF_INT32:
       checkDataOOT(SINT32);
       return FALSE;
   case CF_UINT32:
       checkDataOOT(UINT32);
       return FALSE;
   case CF_UINT64:
#      ifdef OS_64BIT
       checkDataOOT(UINT64);
       return FALSE;
#      endif
   case CF_INT64:
       checkDataOOT(SINT64);
       return FALSE;
   case CF_FLOAT:
       checkDataOOT(float);
       return FALSE;
   case CF_BYTE:
       checkDataOOT(BYTE);
       return FALSE;
  }
  return TRUE;
}
/** 
 * \internal (but EXPORTED?)
 */
TINE_EXPORT int reformat(BYTE *dstData,int dstFmt,BYTE *srcData,int srcFmt,int num)
{
  DTYPE d;
  UINT32 n = (UINT32)num;

  if (dstFmt == srcFmt) return  0;
  d.dArrayLength = num;
  d.dFormat = srcFmt;
  d.data.bptr = srcData;
  switch (dstFmt)
  {
    case CF_INT16: return _GetValuesAsShort(&d,(SINT16 *)dstData,num, TRUE);
    case CF_UINT16: return _GetValuesAsUnsignedShort(&d, (UINT16 *)dstData, num, TRUE);
    case CF_INT32: return _GetValuesAsLong(&d,(SINT32 *)dstData,num, TRUE);
    case CF_UINT32: return _GetValuesAsUnsignedLong(&d, (UINT32 *)dstData, num, TRUE);
    case CF_BOOLEAN: return _GetValuesAsLong(&d, (SINT32*)dstData, num, TRUE);
    case CF_FLOAT: return _GetValuesAsFloat(&d,(float *)dstData,num, TRUE);
    case CF_DOUBLE: return _GetValuesAsDouble(&d,(double *)dstData,num, TRUE);
#   ifndef SMALL_TINE_LIB
    case CF_TEXT: return GetValuesAsString(&d,(char *)dstData,&n);
    case CF_NAME8: 
      if (srcFmt == CF_TEXT) d.dArrayLength = n = fmtsizeof(dstFmt);
      return GetValuesAsString(&d,((NAME8 *)dstData)->name,&n);
    case CF_NAME16: 
      if (srcFmt == CF_TEXT) d.dArrayLength = n = fmtsizeof(dstFmt);
      return GetValuesAsString(&d,((NAME16 *)dstData)->name,&n);
    case CF_NAME32: 
      if (srcFmt == CF_TEXT) d.dArrayLength = n = fmtsizeof(dstFmt);
      return GetValuesAsString(&d,((NAME32 *)dstData)->name,&n);
    case CF_NAME48: 
      if (srcFmt == CF_TEXT) d.dArrayLength = n = fmtsizeof(dstFmt);
      return GetValuesAsString(&d,((NAME48 *)dstData)->name,&n);
    case CF_NAME64: 
      if (srcFmt == CF_TEXT) d.dArrayLength = n = fmtsizeof(dstFmt);
      return GetValuesAsString(&d,((NAME64 *)dstData)->name,&n);
#   endif
  }
  return illegal_format;
}
#if defined(DOS)|defined(VMS)|defined(WIN32)
TINE_EXPORT struct timeval *gettimeofday(struct timeval *t,struct timezone *tz)
{
  struct timeb tb;
  ftime(&tb);
  tz = tz;
  t->tv_sec  = (long)tb.time;
  t->tv_usec = ((UINT32)tb.millitm)*1000L;
  return t;
}
TINE_EXPORT int nanosleep(const struct timespec *req, struct timespec *rem)
{
# ifdef WIN32
  static int done=FALSE;
  static HANDLE hTmr = 0;
  int n100s;
  LARGE_INTEGER li;
  DWORD to;
  if (req == NULL) return -1;
  if (!done)
  {
    if ((hTmr=CreateWaitableTimer(NULL,FALSE,NULL)) == 0) return -1;
    done = TRUE;
  }
  n100s = (int)(req->tv_sec * 10000000 + (req->tv_nsec / 100));
  li.QuadPart = -n100s;
  if ((to = n100s/10000) == 0) to = 1;
  if (SetWaitableTimer(hTmr,&li,0,NULL,NULL,FALSE) == 0) return -1;
  switch (WaitForSingleObject(hTmr,to))
  {
    case WAIT_TIMEOUT:
      rem = rem; /* just ignore this in windows */
    case WAIT_OBJECT_0: 
      return 0;
    default: 
      break;
  }
# else
  req = req; rem = rem;
# endif
  return -1;
}
#elif defined(VXWORKS)
#ifndef HAVE_GETTIMEOFDAY
struct timeval *gettimeofday(struct timeval *t,struct timezone *tz)
{
  struct timespec ts;
  clock_gettime(CLOCK_REALTIME,&ts);
  t->tv_sec  = ts.tv_sec;
  t->tv_usec = ts.tv_nsec/1000;
  return t;
}
#endif
int millisleep(int msecs)
{
# ifdef USE_TASKDELAY
  int ticks = (sysClkRateGet() * msecs)/1000;
  if (ticks == 0 && msecs > 0) ticks = 1;
  return taskDelay(ticks);
# else
  struct timespec ts;
  ts.tv_sec = msecs/1000;
  ts.tv_nsec = (msecs%1000) * 1000000;
  return nanosleep(&ts,NULL);
# endif
}
extern time_t getUnixTime(char *tsHost);
int rdate(char *tsHost)
{
  SINT32 utc = 0;
  struct timespec ts;
  double dts = 0;

  if (SystemGetSiteTimestamp(&dts) == 0)
  {
    utc =  (SINT32)dts;
  }
  else 
  {
    if (tsHost == NULL) return not_running;
    utc = getUnixTime(tsHost);
    if (utc == 0) return not_running;
  }
  ts.tv_sec = (SINT32)dts;
  ts.tv_nsec = 0;
  if (clock_settime(CLOCK_REALTIME,&ts) != 0)
  {
    perror("clock_settime");
    return code_failure;
  }
  return 0;
}
#elif defined(NEED_CLOCK_GETTIME)
int clock_gettime(int clk_id, struct timespec *tp)
{
  struct timeval tv;
  int cc;
  clk_id = clk_id;
  cc = gettimeofday(&tv,NULL);
  if (tp != NULL)
  {
    tp->tv_sec = tv.tv_sec;
    tp->tv_nsec = tv.tv_usec * 1000;
  }
  return cc;
}
#endif
int getClockTickMilliseconds(void)
{
  static int tickMilliseconds = 0;
  int tps = 0;
  if (tickMilliseconds != 0) return tickMilliseconds;
# if defined(VXWORKS)
  tps = sysClkRateGet();
# elif defined(WIN32) || defined(UNIX)
  tps = CLOCKS_PER_SEC;
# elif defined(DOS)
  tps = CLK_TCK;
# endif
  tickMilliseconds = tps == 0 ? SYSTEM_TICK : 1000/tps;
  if (tickMilliseconds < SYSTEM_TICK) tickMilliseconds = SYSTEM_TICK;
  return tickMilliseconds;
}
TINE_EXPORT int microsleep(int usecs)
{
  struct timeval tv0,tv1;
  struct timespec ts, trem;
  int rc, msecs, usecs_slept = 0;
# if defined(WIN32) && defined(SINT64)
  static LARGE_INTEGER cps = { 0, 0};
  LARGE_INTEGER cts0,cts1;
  SINT64 ncts;
  int yld = (usecs > 4) ? TRUE : FALSE;
  if (usecs <= 0) return 0;
  if (usecs > 50) goto nsleep;
  if (usecs > 1000) goto msleep;
  if (cps.QuadPart == 0 && QueryPerformanceFrequency(&cps) == 0) return 0;
  if (QueryPerformanceCounter(&cts0) == 0) return 0;
  ncts = (cps.QuadPart * usecs) / 1000000;
  do
  {
    if (yld) YieldThread();
    if (QueryPerformanceCounter(&cts1) == 0) return 0;
  } while (cts1.QuadPart < cts0.QuadPart + ncts);
  usecs_slept = (int)(((cts1.QuadPart-cts0.QuadPart) * 1000000) / cps.QuadPart);
  return usecs_slept;
# elif defined(MACOS)
  clock_serv_t cclock;
  mach_timespec_t ts0, ts1;
  int yld = (usecs > 4) ? TRUE : FALSE;
  if (usecs <= 0) return 0;
  if (usecs > 50) goto nsleep;
  if (usecs > 10000) goto msleep;
  host_get_clock_service(mach_host_self(), CALENDAR_CLOCK, &cclock); 
  clock_get_time(cclock, &ts0);
  do
  {
    if (yld) YieldThread();
    clock_get_time(cclock, &ts1);
    usecs_slept = (ts1.tv_sec - ts0.tv_sec) * 1000000;
    usecs_slept += (ts1.tv_nsec - ts0.tv_nsec) / 1000;
  } while (usecs_slept < usecs);  
  mach_port_deallocate(mach_task_self(), cclock); 
  return usecs_slept;
# elif defined(UNIX)
  struct timespec ts0, ts1;
  int yld = (usecs > 4) ? TRUE : FALSE;
  if (usecs <= 0) return 0;
  if (usecs > 50) goto nsleep;
  if (usecs > 10000) goto msleep;
  clock_gettime(CLOCK_MONOTONIC, &ts0);
  do
  {
    if (yld) YieldThread();
    clock_gettime(CLOCK_MONOTONIC, &ts1);
    usecs_slept = (ts1.tv_sec - ts0.tv_sec) * 1000000;
    usecs_slept += (ts1.tv_nsec - ts0.tv_nsec) / 1000;
  } while (usecs_slept < usecs);
  return usecs_slept;
# elif defined(VXWORKS)
  if (usecs <= 0) return 0;
  if (usecs > 10000) goto msleep;
  goto nsleep;
# else
  goto msleep;
# endif
nsleep:
  ts.tv_sec = usecs/1000000;
  ts.tv_nsec = (usecs%1000000) * 1000;
  rc = nanosleep(&ts,&trem);
  usecs_slept = usecs;
  if (rc != 0) 
  {
    usecs_slept -= (int)((trem.tv_sec-ts.tv_sec) * 1000000);
    usecs_slept -= (trem.tv_nsec - ts.tv_nsec) /1000;
  }
  return usecs_slept;
msleep:
  msecs = usecs / 1000;
  if (msecs < 1) msecs = 1;
  gettimeofday(&tv0,(struct timezone *)NULL);
  millisleep(msecs);
  gettimeofday(&tv1,(struct timezone *)NULL);
  usecs_slept = (tv1.tv_sec - tv0.tv_sec) * 1000000;
  usecs_slept += (tv1.tv_usec - tv0.tv_usec);
  return usecs_slept;
}
int NoDstFindDaylight(time_t t) 
{
  UNUSED(t);
  return 0;
}
int METFindDaylight(time_t t)
{
  int dl;
  struct tm lt, *tm = localtime(&t);
  if (tm == NULL) return 0;
  lt = *tm;
# if defined(UNIX) || defined(WIN32)
  if (lt.tm_isdst > 0) return 1;
  if (lt.tm_isdst == 0) return 0;
# endif
  if      (lt.tm_mon<2 || lt.tm_mon>9)       dl = 0;  /* winter */
  else if (lt.tm_mon>2 && lt.tm_mon<9)       dl = 1;  /* summer */
  else if (lt.tm_mon==2)                               /* March */
  { if      (lt.tm_mday < 25)                  dl = 0;
    else if (lt.tm_wday > lt.tm_mday-25)      dl = 0;
    else if (lt.tm_wday == 0 && lt.tm_hour<2) dl = 0; /* Sunday */
    else dl = 1;
  }
  else                                                  /* October */
  { if      (lt.tm_mday < 25)                  dl = 1;
    else if (lt.tm_wday > lt.tm_mday-25)      dl = 1;
    else if (lt.tm_wday == 0 && lt.tm_hour<2) dl = 1; /* Sunday */
    else dl = 0;
  }
  return dl;
}
#ifndef DOS
/**
 * \internal
 * \brief Find the name of the owner of this process AND the caller name.
 *
 * For a client process the caller = logged in user. For a FEC the
 * caller will be the FEC name.
 * 
 * \param user is the caller 'user name' identity
 * \return the loggen in (process owner) 'user name' identity
 */
char *whoami(char *user)
{
  char *up = NULL;
  static char usrnam[USERNAME_SIZE+4] = "";
# ifdef WIN32
  DWORD n = USERNAME_SIZE;
# endif
# ifdef UNIX
  FILE *upipe = NULL;
  struct passwd *passwd;
# endif

  if (strlen(usrnam) > 0) goto out;
  if ((up=getenv("CONTROLUSER")) != NULL) strncpy(usrnam,up,USERNAME_SIZE); 
  if (up == NULL)
  { /* normal case: no backdoor user*/
#   if defined(WIN32)
    GetUserName(usrnam,&n);
#   elif defined(UNIX)
    passwd = getpwuid(getuid());
    if (passwd != NULL) 
    { /* can this ever fail ?*/
      strncpy(usrnam,passwd->pw_name,USERNAME_SIZE);
    }
    else
    {
      if ((upipe=popen("whoami","r")) && fgets(usrnam,sizeof(usrnam),upipe))
        usrnam[strcspn(usrnam,"\n")] = 0; /* strip CR from whoami return */
      if (upipe != NULL) pclose(upipe);
    }
#   elif defined(DOS)
    if ((up=getenv("LOGINAME")) != NULL ||  /* Novell env. */
        (up=getenv("LOGINNAME")) != NULL || /*      "      */
        (up=getenv("USERNAME")) != NULL ||  /* Windows-like */
        (up=getenv("USER")) != NULL)        /* unix-like */
    {
      strncpy(usrnam,up,USERNAME_SIZE); 
    }
#   elif defined(VMS)
    strncpy(usrnam,"VAX_USER",USERNAME_SIZE);
#   elif defined(VXWORKS)
    strncpy(usrnam,"VXW_USER",USERNAME_SIZE);
#   elif defined(NIOS)
    strncpy(usrnam,"NIOS_USER",USERNAME_SIZE);
#   endif
  }
  strupr(usrnam);
  if (strlen(usrnam) == 0) strcpy(usrnam,"UNKNOWN");
out:
  if (strlen(gUserName) == 0) 
  { /* caller id not yet set (not a FEC?) */
    strncpy(gUserName,usrnam,USERNAME_SIZE); 
  }
  if (user != NULL && user != gUserName) strncpy(user,gUserName,USERNAME_SIZE);
  if (tineDebug) dbglog("caller is %.16s, logged in user: %.16s",gUserName,usrnam);
  return usrnam; /* process owner */
}
#endif
int isAbsolutePath(char *fname)
{
  if (fname == NULL || strlen(fname) < 2) return FALSE;
  if (fname[1] == ':') return TRUE;
  if (fname[0] == '\\') return TRUE;
  if (fname[0] == '/') return TRUE;
  if (strncmp(fname,"./",2) == 0) return TRUE;
  return FALSE;
}
#define NotALogFile(f) (\
  strstr(f,".log") == NULL &&\
  strstr(f,".txt") == NULL &&\
  strstr(f,".ini") == NULL &&\
  strstr(f,".dat") == NULL &&\
  strstr(f,".cfg") == NULL &&\
  strstr(f,".csv") == NULL)

#if defined(UNIX) || defined(WIN32)
static char tmptok[][4] = { "#", "~", "lck", "tmp", "swp" };
static int numtmptok = sizeof(tmptok)/sizeof(tmptok[0]);
static int isTmpFile(char *fn)
{
  int i,p,len = (int)strlen(fn);
  for (i=0; i<numtmptok; i++)
  {
    if ((p=len-(int)strlen(tmptok[i])) < 0) continue;
    if (stricmp(&fn[p],tmptok[i]) == 0) return TRUE;
  }
  return FALSE;
}
#endif
char gFeclogPath[STD_PATH_LENGTH];
int gFeclogPathEstablished = FALSE;
void assertPathDelimiter(char *path,int maxlen)
{
  int len = (int)strlen(path);
  if (len > maxlen) len = maxlen;
# ifdef FS_DELIMITER
  if (len == 0) return; /* current working directory ! */
  if (path[len-1] != FS_DELIMITER) path[len] = FS_DELIMITER;
# endif
}
TINE_EXPORT int SetFecLogPath(char *path)
{
  if (path == NULL || strlen(path) == 0)
  {
    memset(gFeclogPath,0,STD_PATH_LENGTH);
    gFeclogPathEstablished = TRUE;
    return 0;
  }
  if (strlen(path) > STD_PATH_LENGTH) return string_too_long;
  memcpy(gFeclogPath,path,STD_PATH_LENGTH);
  assertPathDelimiter(gFeclogPath,STD_PATH_LENGTH);
  gFeclogPathEstablished = TRUE;
  return 0;
}
void applyFecLogPath(char *path,int pathlen)
{
  char root[256];
  int len = (int)strlen(gFeclogPath);
  if (path == NULL) return;
  if (isAbsolutePath(path)) return; /* nothing to apply */
  if (len + (int)strlen(path) > pathlen) 
  {
    dbglog("%s + %s too long!",gFeclogPath,path);
    return;
  }
  strncpy(root,gFeclogPath,STD_PATH_LENGTH);
# if defined(UNIX) || defined(VXWORKS)
  if (len == 0) strcpy(root,"./");
# else
  if (len == 0) return; /* nothing to apply */
# endif
  strncat(root,path,128);
  strncpy(path,root,256); /* copy it back */
  return;
}
#ifndef SMALL_TINE_LIB
#define MAX_FILE_SCAN_LIMIT 100
int getFecLogFileLength(char *fname,int *mtime)
{
# ifndef VXWORKS
  struct stat sbuf;
  int useAbsolutePath = isAbsolutePath(fname);
  char fn[256];
  if (useAbsolutePath) strncpy(fn, fname, 128); else snprintf(fn, 128, "%s%s", gFeclogPath, fname);
  if (stat(fn, &sbuf) == 0)
  {
    if (mtime != NULL) *mtime = (int)sbuf.st_mtime;
    return (int)sbuf.st_size;
  }
  return -file_error;
# else
  return -not_supported;
# endif
}
int scanFecLogFiles(char *path,char *subpath,void *namebuf,int fmt,int *numfiles,int numoffset)
{
  int n = numoffset;
# ifndef VXWORKS
  int scancounter = 0;
  struct stat sbuf;
# ifdef UNIX
  char fn[128],tfn[128],ffn[264];
# ifdef FECLOG_SCANDIR
  char spath[64];
# endif
  DIR *dp;
  struct dirent *dirp;
# ifdef FECLOG_SCANDIR
  int nread = 0;
# endif
# endif
# if defined(DOS)
  char fn[128],fnp[128],tfn[128],spath[64],ffn[264];
  struct find_t fb;
  int nread = 0;
# elif defined(UNDER_CE)
  int nread = 0;
# elif defined(WIN32)
  char fn[128],fnp[128],tfn[128],spath[64],ffn[264];
  struct _finddata_t fp;
  int nread = 0;
# ifndef OS_64BIT
  long hfp;
# else
  intptr_t hfp;
# endif
# endif
  int appendFileLength = FALSE;
  NAME64 *n64 = (NAME64 *)namebuf;
  NAME64I *n64i = (NAME64I *)namebuf;

  if (namebuf == NULL || numfiles == NULL || *numfiles == 0) return argument_list_error;
  switch (fmt)
  {
    case CF_NAME64I:
      appendFileLength = TRUE;
    case CF_NAME64:
      break;
    default:
      return illegal_format;
  }
# if defined(UNIX)
  if (path == NULL) path = "./";
  if (strlen(path) > 0 && path[strlen(path) - 1] != '/') strcat(path, "/");
  strncpy(fn, path, 128);
  if ((dp=opendir(fn)) == NULL) return no_such_file;
  memset(tfn,0,128);
  while ((dirp=readdir(dp)) != NULL)
  {
    if (scancounter++ > MAX_FILE_SCAN_LIMIT) break;
    if (!strcmp(dirp->d_name,".") || !strcmp(dirp->d_name,"..")) continue;
#   ifdef FECLOG_SCANDIR
    if (dirp->d_type == DT_DIR)
    {
      if (strlen(path) + strlen(dirp->d_name) > 128) continue;
      if (subpath != NULL) strncpy(spath,subpath,60); else spath[0] = 0;
      strcat(path,dirp->d_name);
      strncpy(fn,path,128);
      strcat(spath,dirp->d_name); strcat(spath,"/");
      nread = *numfiles;
      scanFecLogFiles(fn,spath,namebuf,fmt,&nread,n);
      n = nread;
    }
#   endif
    strncpy(tfn,dirp->d_name,64); strlwr(tfn);
    if (NotALogFile(tfn)) continue;
    if (isTmpFile(tfn)) continue;
    if (n >= *numfiles) break;
    if (subpath != NULL && strlen(subpath) + strlen(dirp->d_name) < 64)
    {
      snprintf(tfn,64,"%.48s%s",subpath,dirp->d_name);
    }
    else
    {
      strncpy(tfn,dirp->d_name,64);
    }
    if (appendFileLength)
    {
      strncpy(n64i[n].name, tfn, 64);
      snprintf(ffn,256,"%s/%s",path,tfn);
      if (stat(ffn,&sbuf) == 0) n64i[n].ival = (int)sbuf.st_size;
    }
    else
    {
      strncpy(n64[n].name, tfn, 64);
    }
    n++;
  }
  closedir(dp);
# elif defined(UNDER_CE)
  /* do something reasonable later ... */
  *numfiles = nread;
# elif defined(WIN32)
  if (strlen(path) > 0 && path[strlen(path)-1] != '\\') strcat(path,"\\");
  if (subpath != NULL) strncpy(spath,subpath,48);
  sprintf(fn,"%s*.*",path);
  if ((hfp=_findfirst(fn, &fp)) == -1) return no_such_file;
  do
  {
    if (scancounter++ > MAX_FILE_SCAN_LIMIT) break;
    if (!strcmp(fp.name,".") || !strcmp(fp.name,"..")) continue;
    if (fp.attrib & _A_SUBDIR)
    {
      if (subpath != NULL) strncpy(spath,subpath,60); else spath[0] = 0;
      strncpy(fnp,path,128);
      if (strlen(fnp) + strlen(fp.name) > 128) continue;
      if (strlen(spath) + strlen(fp.name) > 60) continue;
      strcat(fnp,fp.name); strcat(fnp,"\\");
      strcat(spath,fp.name); strcat(spath,"\\");
      nread = *numfiles;
      scanFecLogFiles(fnp,spath,namebuf,fmt,&nread,n);
      n = nread;
    }
    strncpy(tfn,fp.name,64); strlwr(tfn);
    if (NotALogFile(tfn)) continue;
    if (isTmpFile(tfn)) continue;
    if (n >= *numfiles) break;
    if (subpath != NULL && strlen(subpath) + strlen(fp.name) < 64)
    {
      snprintf(tfn,64,"%.48s%s",subpath,fp.name);
    }
    else
    {
      strncpy(tfn,fp.name,32);
    }
    if (appendFileLength)
    {
      strncpy(n64i[n].name, tfn, 64);
      snprintf(ffn,256,"%s%s", path, tfn);
      if (stat(ffn, &sbuf) == 0)
      {
        n64i[n].ival = (int)sbuf.st_size;
      }
    }
    else
    {
      strncpy(n64[n].name, tfn, 64);
    }
    n++;
  } while (_findnext(hfp,&fp) != -1);
  _findclose(hfp);
# elif defined(DOS)
  if (strlen(path) > 0 && path[strlen(path)-1] != '\\') strcat(path,"\\");
  sprintf(fn,"%s*.*",path);
  if (_dos_findfirst(fn,_A_NORMAL,&fb) != 0) return no_such_file;
  do
  {
    if (scancounter++ > MAX_FILE_SCAN_LIMIT) break;
    if (!strcmp(fb.name,".") || !strcmp(fb.name,"..")) continue;
    if (fb.attrib & FA_DIREC)
    {
      if (subpath != NULL) strncpy(spath,subpath,60); else spath[0] = 0;
      strncpy(fnp,path,128);
      if (strlen(fnp) + strlen(fb.name) > 128) continue;
      if (strlen(spath) + strlen(fb.name) > 60) continue;
      strcat(fnp,fb.name); strcat(fnp,"\\");
      strcat(spath,fb.name); strcat(spath,"\\");
      nread = *numfiles;
      scanFecLogFiles(fnp,spath,namebuf,fmt,&nread,n);
      n = nread;
    }
    strncpy(tfn,fb.name,64); strlwr(tfn);
    if (NotALogFile(tfn)) continue;
    if (n >= *numfiles) break;
    if (subpath != NULL && strlen(subpath) + strlen(fb.name) < 64)
    {
      sprintf(tfn,"%s%s",subpath,fb.name);
    }
    else
    {
      strncpy(tfn,fb.name,13);
    }
    if (appendFileLength)
    {
      strncpy(n64i[n].name, tfn, 64);
      snprintf(ffn, 256, "%s%s", path, tfn);
      if (stat(ffn, &sbuf) == 0) n64i[n].ival = (int)sbuf.st_size;
    }
    else
    {
      strncpy(n64[n].name, tfn, 64);
    }
    n++;
  } while (_dos_findnext(&fb) == 0);
# endif /* UNIX, WIN32 or DOS */
# else
  if (namebuf == NULL || numfiles == NULL || *numfiles == 0) return argument_list_error;
# endif /* !VXWORKS */
  *numfiles = n;
  return 0;
}
#endif /* !SMALL_TINE_LIB */
int readFecLog(char *textbuf,int sizeInBytes)
{
# ifndef DOS
  if (vFeclogBuffer != NULL) return readvFecLog(textbuf,sizeInBytes);
# endif
  return readFecLogFile("fec.log",textbuf,sizeInBytes);
}
int readFecLogFile(char *fname,char *textbuf,int sizeInBytes)
{
  int cc=0,useAbsolutePath=0;
  FILE *fp = NULL, *fpb = NULL;
  char fn[128],bakname[96], *c;
  int sz=0,szb=0,szt=0,pos=0;
# ifdef DOS
  static char *fmod = "r+t";
# else
  static char *fmod = "r";
# endif

  if (nofeclog) return 0;
  if (fname == NULL) return no_such_file;
  if (sizeInBytes <= 0) return dimension_error;
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else snprintf(fn,128,"%s%s",gFeclogPath,fname);
  if ((fp=fopen(fn,fmod)) == NULL) goto checkbak;
  fseek(fp,0,SEEK_END); sz = ftell(fp);
checkbak:
  if (strstr(fname,".log") == NULL) goto filesread; /* something other than a log file */
  if (sz < sizeInBytes)  /* look for back file */
  {
    szt = 0;
    strncpy(bakname,fname,95); bakname[95] = 0;
    if ((c=strstr(bakname,".log")) != NULL) *c = 0;
    strcat(bakname,".bak");
    if (useAbsolutePath) strncpy(fn,bakname,128); else snprintf(fn,128,"%.96s%s",gFeclogPath,bakname);
    if ((fpb=fopen(fn,fmod)) == NULL) goto filesread;
    szt = sizeInBytes - sz;
    fseek(fpb,0,SEEK_END); szb = ftell(fpb);
    if (szb > szt) pos = szb - szt;
    fseek(fpb,pos,SEEK_SET);
    szt = (int)fread(textbuf,1,szt,fpb);
    pos = 0;
    fclose(fpb);
  }
filesread:
  if (fp)
  {
    if (sz > sizeInBytes) pos = sz - sizeInBytes;
    fseek(fp,pos,SEEK_SET);
    szb = (int)fread(&textbuf[szt],1,sizeInBytes-szt,fp);
    fclose(fp);
    if (szt + szb < sizeInBytes) textbuf[szt+szb] = 0;
  }
  if ((sz + szb) == 0) cc = no_such_file;
  if (cc) dbglog("%s : %s",fn,cc2str(cc));
  return cc;
}
int writeFecLogFile(char *fname,char *textbuf,int sizeInBytes)
{
  int cc=0,useAbsolutePath=0;
# ifndef FS_RDONLY
  FILE *fp = NULL;
  char fn[128],fb[128], *c;
# ifdef DOS
  static char *fmod = "w+t";
# else
  static char *fmod = "w";
# endif

  if (fname == NULL) return no_such_file;
  if (sizeInBytes <= 0) return dimension_error;
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else snprintf(fn,128,"%s%s",gFeclogPath,fname);
  strncpy(fb,fn,123); fb[123] = 0;
  if ((c=strchr(fb,'.')) != NULL) *c = 0;
  strcat(fb,".bak");
  remove(fb);
  rename(fn,fb);
  if ((fp=fopen(fn,fmod)) == NULL) ccerr(file_error);
  fwrite(textbuf,1,sizeInBytes,fp);
  fclose(fp);
# else 
  useAbsolutePath = 0;
  ccerr(not_implemented); /* no way to write a log file */
# endif
err:
# ifndef FS_RDONLY
  if (cc) dbglog("%s : %s",fn,cc2str(cc));
# endif
  return cc;
}
int readFecFile(char *fname,char *buf,int *sizeInBytes)
{
  int cc=0,useAbsolutePath=0;
  FILE *fp = NULL;
  char fn[132] = {0};
  int sz=0,szb=0;
  static char *fmod = "r";

  if (fname == NULL || buf == NULL || sizeInBytes == NULL) return argument_list_error;
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else snprintf(fn,128,"%s%s",gFeclogPath,fname);
  if ((fp=fopen(fn,fmod)) == NULL) return no_such_file;
  fseek(fp,0,SEEK_END); sz = ftell(fp);
  szb = *sizeInBytes;
  if (sz < szb) szb = sz;
  fseek(fp,0,SEEK_SET);
  szb = (int)fread(buf,1,szb,fp);
  fclose(fp);
  *sizeInBytes = szb;
  if (szb == 0) cc = file_error;
  if (cc) dbglog("%s : %s",fn,cc2str(cc));
  return cc;
}
int writeFecFile(char *fname,char *buf,int sizeInBytes)
{
  int cc=0,useAbsolutePath=0;
# ifndef FS_RDONLY
  FILE *fp = NULL;
  char fn[128];
  static char *fmod = "w";

  if (fname == NULL) return no_such_file;
  if (sizeInBytes <= 0) return dimension_error;
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else sprintf(fn,"%s%s",gFeclogPath,fname);
  if ((fp=fopen(fn,fmod)) == NULL) ccerr(file_error);
  fwrite(buf,1,sizeInBytes,fp);
  fclose(fp);
# else 
  useAbsolutePath = 0;
  ccerr(not_implemented); /* no way to write a log file */
# endif
err:
# ifndef FS_RDONLY
  if (cc) dbglog("%s : %s",fn,cc2str(cc));
# endif
  return cc;
}
extern char TimeServerHost[64];
char *TimeServerHostPtr=TIME_SERVER_HOST;
char *getTimeServerHost(void)
{
  char *ptr=NULL;
  if ((ptr=getenv("RDATE")) != NULL)
  {
    if (!strcmp(ptr,"NONE")) return NULL;
    strncpy(TimeServerHost,ptr,64);
  }
  else
  {
    if (TimeServerHostPtr != NULL) strncpy(TimeServerHost,TimeServerHostPtr,64);
    else return NULL;
  }
  return TimeServerHost;
}
int file_exists(char *filename)
{
# ifndef VXWORKS
  struct stat sbuf;
  return stat(filename, &sbuf) == 0 ? TRUE : FALSE;
# else
  UNUSED(filename);
  return FALSE;
# endif
}
/**
 * \internal helper routine to establish a full database file name
 * \param ffn should point to a buffer of min length 256
 */
int fixDbFileName(char *ffn,char *eqm,char *file, int ismff)
{
  char dn[256];
  if (ffn == NULL || file == NULL) return argument_list_error;
# ifdef FS_DELIMITER
  if (eqm != NULL)
  { 
    if (ismff)
    { /* remove anything found in the non-cache area */
      char *fn = !strcmp(file, ALARM_WATCH_MANIFEST) ? ALARM_WATCH_MANIFEST_LEGACY : file;
      sprintf(dn, "%.128s%.6s%c%s", fecDbPath, eqm, FS_DELIMITER, fn);
      if (file_exists(dn)) unlink(dn);
      sprintf(dn, "%.128s%.6s%ccache", fecDbPath, eqm, FS_DELIMITER);
    }
    else
    {
      sprintf(dn, "%.128s%.6s", fecDbPath, eqm);
    }
    if (!makeDirectoryPath(dn,TRUE)) goto ezPath;
    sprintf(ffn,"%.144s%c%s",dn,FS_DELIMITER,file);
    return 0;
  }
# else
  eqm = NULL;
  goto ezPath;
# endif
ezPath:
  sprintf(ffn,"%.128s%s",fecDbPath,file);
  return 0;
}
int readSelfTest(char *eqm,char *textbuf,int sizeInBytes)
{
  int sz=0, pass=0;
  FILE *fp = NULL;
  char fn[256], *ext;
# ifdef DOS
  static char *fmod = "r+t";
# else
  static char *fmod = "r";
# endif
  textbuf[0] = 0;
openSelfTestFile:
  ext = pass == 0 ? ".csv" : ".xml";
  fixDbFileName(fn,eqm,SELFTEST_FILE,FALSE);
  if (eqm != NULL) goto fnDone;
  if (ignoreCommonDbFiles && pass) goto err;
fnDone:
  strcat(fn,ext);
  if ((fp=fopen(fn,fmod)) == NULL) 
  { /* not found */
    if (pass == 0) 
    { /* try other extension */
      pass++;
      goto openSelfTestFile;
    }
    if (eqm != NULL)
    { /* try FEC common directory */
      pass = 0;
      eqm = NULL;
      goto openSelfTestFile;
    }
    goto err;
  }
  sz = (int)fread(textbuf,1,sizeInBytes,fp);
  fclose(fp);
err:
  return sz;
}
#ifdef RDATE
int FecSettime(void)
{
  static time_t lastlogged = 0;
  time_t t;
  int delt,log = 0;

  t = time(NULL);
  rdate(TimeServerHostPtr);
  delt = (int)(time(NULL) - t);
  if (abs(delt) > 2) log = 1;
  if (t > lastlogged + (3600*24)) log = 1;
  if (tineDebug)
  {
    ttyoutput("time synchronized to server\n>");
    FecShowtime();
  }
  if (log)
  {
    feclog("synchronized (%d sec)",delt);
    lastlogged = t;
  }
  return 0;
}
void FecSynchronize(void)
{
  time_t t = time(NULL);

  if ((t % 3600) == 0)     /* sychronize once per hour */
  {
    FecSettime();
  }
}
void FecShowtime(void)
{
  char tstr[32];
  time_t t = time(NULL);

  memset(tstr,0,32);
  strncpy(tstr,ctime(&t),19);
  ttyoutput("current FEC time: %s (UNIX: %ld)\n>",&tstr[4],t);
}
#endif /* RDATE */
#ifdef NIOS
int nr_setclock(SINT32 tgtTime);
extern time_t getUnixTime(char *tsHost);
int rdate(char *tsHost)
{
  int isdst;                         /* daylight saving time flag   */
  time_t utime=time(NULL);
  SINT32 dt;

  isdst = findDaylight(utime);
  putenv(isdst ? DLT_TIMEZONE_STR : STD_TIMEZONE_STR);
  tzset();

  if ((utime=getUnixTime(tsHost)) == 0) return -1;
  if ((dt=nr_setclock(utime)) != 0) 
  {
    feclog("adjust system clock by %d seconds",dt);
  }
  return 0;
}
void FecShowtime(void)
{
  char tstr[32];
  time_t t = time(NULL);

  memset(tstr,0,32);
  strncpy(tstr,ctime(&t),19);
  ttyoutput("current FEC time: %s (UNIX: %d)",&tstr[4],t);
}
#endif /* NIOS */

#ifdef DOS

int hConn,hControl;
WORD userConnNr;                     /* IPX connection number of PC */
extern time_t getUnixTime(char *tsHost);
int rdate(char *tsHost)
{
  struct dostime_t t, t0;
  struct dosdate_t d, d0;
  struct tm *lt;
  time_t utime = getUnixTime(tsHost);
  int isdst;                         /* daylight saving time flag   */

  isdst = findDaylight(utime);
  putenv(isdst ? DLT_TIMEZONE_STR : STD_TIMEZONE_STR);
  tzset();

  if (utime == 0) return -1;

  lt = localtime(&utime);

  _dos_gettime(&t); _dos_getdate(&d);
  memcpy(&d0,&d,sizeof(struct dosdate_t));
  memcpy(&t0,&t,sizeof(struct dostime_t));

  d.year  = 1900 + lt->tm_year;
  d.month = 1 + lt->tm_mon;
  d.day   = lt->tm_mday;
  t.hour  = lt->tm_hour;
  t.minute= lt->tm_min;
  t.second= lt->tm_sec;

  if (memcmp(&d,&d0,sizeof(struct dosdate_t)) ) _dos_setdate(&d);
  if (memcmp(&t,&t0,sizeof(struct dostime_t)) ) _dos_settime(&t);

  return 0;
}
int DiskErrorCount = 0;
#pragma argsused
int NetHandler(int errval,int ax,int bp,int si)
{
  if (ax > 0x04)hardretn(2);
  DiskErrorCount++;
  return 2;
}
UINT32 far *ColdBoot = (UINT32 far *)0x00400072L;
int F_boot(void)
{
  feclog("Booting ...");
  *ColdBoot = 0x1234;
  asm DB 0EAh,0,0,0FFh,0FFh
  return 0;
}

#define WD_PFORTE 0
#define WD_BECKER 1
#ifndef WD_CARD
# define WD_CARD WD_PFORTE
#endif

#define WATCH_INTERVAL  5
#define WD_RESET(i)  ((i)+0x0003)
#define WD_POLL(i)   ((i)+0x0002)
#define WD_STATUS(i) ((i)+0x0001)

int watchdogEnabled = 0;
int watchdogType = WD_CARD;
void GetWatchDogBaseAddress(void)
{
  char *c,str[80];
  if ((c=getenv("WD_TYPE")) != NULL)
  {
    if (strstr(c,"BECKER") != NULL) watchdogType = WD_BECKER;
  }
  sprintf(str,"WD_TYPE : [%s]",c != NULL ? c : "");
  feclog(str);
  if ((c=getenv("WD_BASEADDR")) != NULL || (c=getenv("WATCHDOG")) != NULL)
  {
    watchdogBaseAddress = (WORD)strtol(c,NULL,0);
  }
  sprintf(str,"WD_BASEADDR : [%s]",c != NULL ? c : "");
  feclog(str);
}
void initWatchDog(void)
{
  char str[256];
  WORD baddr = watchdogBaseAddress;

  if (watchdogBaseAddress == 0) return;
  if (baddr < 0x240 || baddr > 0x400)
  {
    sprintf(str,"Illegal watchdog port : %x",baddr);
    watchdogBaseAddress = 0;
  }
  else
  {
    sprintf(str,"Initializing WATCHDOG at base address %x",baddr);
    if (watchdogType == WD_PFORTE)
    {
      outport(baddr,0x1111);           /* set pattern */
      outport(baddr+4,0x01);           /* enable watchdog */
      watchdogEnabled = 1;
    }
    else if (watchdogType == WD_BECKER)
    {
      outportb(WD_STATUS(baddr),0x01);    /* contact watchdog */
      outportb(watchdogBaseAddress,0x00); /* enable watchdog */
      watchdogEnabled = 1;
    }
  }
  feclog(str);
  return;
}
extern struct timeval thisCycleTime;
void watchDog(void)
{
  static unsigned int status, temp;
  static time_t t0 = 0;

  if (t0 == 0) initWatchDog();
  if (watchdogBaseAddress == 0) return;
  if (time(NULL) < t0 + WATCH_INTERVAL ||
      !watchdogBaseAddress ||
      !watchdogEnabled) return;
  t0 = time(NULL);

  if (watchdogType == WD_PFORTE)
  {
    outport(watchdogBaseAddress+2,0x1111);          /* feed watchdog */
    status = 1;
  }
  else if (watchdogType == WD_BECKER)
  {
    outportb(WD_POLL(watchdogBaseAddress),0x01);    /* feed watchdog */
    status = inportb(watchdogBaseAddress);
    temp = inportb(watchdogBaseAddress+1);
  }
  if (tineDebug>1)
  {
    dbglog("Watchdog %s (%x)",status == 1 ? "active" : "inactive",status);
    dbglog("Temperature %d (%x)",temp,temp);
  }
  return;
}
void enableWatchDog(void)
{
  if (!watchdogBaseAddress || watchdogEnabled) return;
  if (watchdogType == WD_PFORTE)
  {
    outport(watchdogBaseAddress,0x1111);           /* set pattern */
    outport(watchdogBaseAddress+4,0x01);           /* enable watchdog */
  }
  else if (watchdogType == WD_BECKER)
  {
    outportb(watchdogBaseAddress,0x00);           /* enable watchdog */
  }
  watchdogEnabled = 1;
  return;
}
void disableWatchDog(void)
{
  if (!watchdogBaseAddress || !watchdogEnabled) return;
  if (watchdogType == WD_PFORTE)
  {
    outport(watchdogBaseAddress+4,0x00);   /* disable watchdog */
  }
  else if (watchdogType == WD_BECKER)
  {
    outportb(watchdogBaseAddress,0x01);    /* disable watchdog */
  }
  watchdogEnabled = 0;
  return;
}

#else  /* DOS not defined */

int vfeclog(char *text)
{
  int i, len, maxlen = FeclogDepth*80;
  char str[256] = {0};
  if (vFeclogBuffer == NULL || text == NULL) return 0;
  strncpy(str,text,255);
  len = (int)strlen(str);
  for (i=0; i<len; i++) vFeclogBuffer[(vFeclogBufferPos+i)%maxlen] = str[i];
  vFeclogBufferPos = (vFeclogBufferPos + len) % maxlen;
  return 0;
}
int readvFecLog(char *textbuf,int sizeInBytes)
{
  int endptr = vFeclogBufferPos, startptr = 0;
  int i, maxlen = FeclogDepth*80;
  if (vFeclogBuffer[vFeclogBufferPos] != 0)
  {
    endptr = (vFeclogBufferPos + maxlen - 1)%maxlen;
    startptr = vFeclogBufferPos;
  }
  if (sizeInBytes < (endptr-startptr+maxlen)%maxlen)
    startptr = (endptr - sizeInBytes + maxlen) % maxlen;
  for (i=0; i<sizeInBytes; i++)
  {
    if (i >= maxlen) break;
    if ((textbuf[i]=vFeclogBuffer[(startptr+i)%maxlen]) == 0) break;
  }
  return 0;
}
#ifndef WIN32 /* or DOS */
char *_strdate(char *buf)
{
  time_t t;
  struct tm lt;
  time(&t);  
  t += (SINT32)gDataTimeStampOffset;
  lt = *localtime(&t);
  sprintf(buf,"%02d/%02d/%02d",(lt.tm_mon)+1,lt.tm_mday,lt.tm_year%100);
  return buf;
}
char *_strtime(char *buf)
{
  time_t t;
  struct tm lt;
  time(&t);  
  t += (SINT32)gDataTimeStampOffset;
  lt = *localtime(&t);
  sprintf(buf,"%02d:%02d:%02d",lt.tm_hour,lt.tm_min,lt.tm_sec);
  return buf;
}
char *strupr(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c >= 'a' && *c <= 'z') *c -= 32;
  return s;
}
char *strlwr(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c >= 'A' && *c <= 'Z') *c += 32;
  return s;
}
char *strrev(char *s)
{
  char c;
  register int i,len = (int)strlen(s);
  for (i=0; i<len/2; i++) { c = s[i]; s[i] = s[len-1-i]; s[len-1-i] = c; }
  return s;
}
#ifndef NIOS /* or WIN32 or DOS */
int stricmp(const char *s1,const char *s2)
{
  char *c1,*c2;
  for (c1=(char *)s1,c2=(char *)s2; *c1 != 0 && *c2 != 0; c1++,c2++)
    if (toupper(*c1) != toupper(*c2)) break;
  return toupper(*c1) - toupper(*c2);
}
int strnicmp(const char *s1,const char *s2,int n)
{
  char *c1,*c2;
  int i;
  for (c1=(char *)s1,c2=(char *)s2,i=0; *c1 != 0 && *c2 != 0; c1++,c2++,i++)
  {
    if (toupper(*c1) != toupper(*c2)) break;
    if (i >= n-1) break;
  }
  return toupper(*c1) - toupper(*c2);
}
#endif /* not defined NIOS, DOS, WIN32 */
#endif /* not defined WIN32, DOS */

#endif /* else DOS not defined */

#ifdef VXWORKS
void VXWbkgtsk(void (*tsk)(void),int rate,int ref)
{
  int ticks = (sysClkRateGet() * rate) / 1000;
  void (*bkgtsk)(void) = tsk;
  int ready = FALSE;
  ExportListStruct *el = getExportListItemFromId(ref);

  if (ticks < 1) ticks = 1;
  if (gStartupDebug)
  {
    dbglog("background task polling rate : %d msec (%d ticks)",rate,ticks);
  }
  if (el == NULL) ready = TRUE;
  while (!ServerExitCondition)
  {
    if (!ready) ready = el->inidone;
    if (ServerInitialized && ready) bkgtsk();
    taskDelay(ticks);
  }
}
/* a not-thread-safe cover routine for VxWorks ! */
void *alloca(size_t size)
{
  if (tmpWorkArea == NULL) tmpWorkArea = (BYTE *)SystemCalloc(1, tmpWorkAreaSize);
  if (size > tmpWorkAreaSize) return NULL;
  return tmpWorkArea;
}
#endif /* VXWORKS */

#ifdef MULTITHREADED
TINE_EXPORT int useMultiThreadedBackgroundTasks = TRUE;
TINE_EXPORT void SetUseMultiThreadedBackgroundTasks(int value)
{
  useMultiThreadedBackgroundTasks = value;
}
TINE_EXPORT int GetUseMultiThreadedBackgroundTasks(void)
{
  return useMultiThreadedBackgroundTasks;
}
TINE_EXPORT int useMultiThreadedEquipmentFunctions = FALSE;
TINE_EXPORT void SetUseMultiThreadedEquipmentFunctions(int value)
{
  useMultiThreadedEquipmentFunctions = value;
}
TINE_EXPORT int GetUseMultiThreadedEquipmentFunctions(void)
{
  return useMultiThreadedEquipmentFunctions;
}
TINE_EXPORT int useMultiThreadedStockFunctions = TRUE;
TINE_EXPORT void SetUseMultiThreadedStockFunctions(int value)
{
  useMultiThreadedStockFunctions = value;
}
TINE_EXPORT int GetUseMultiThreadedStockFunctions(void)
{
  return useMultiThreadedStockFunctions;
}
int getMsgSizeFromStream(BYTE *stream)
{
  int siz;
  if (stream == NULL) return 0;
  memcpy(&siz,stream,4);
# ifndef INTEL
  ISWAP(siz);
# endif
  return siz;
}
int getStreamCapacity(int msgsiz)
{
  int n = (msgsiz+TCP_BUFFER_SIZE)%TCP_BUFFER_SIZE;
  return n*TCP_BUFFER_SIZE;
}
THRHANDLE hGlbThread = 0;
THRHANDLE hDscEvnThread = 0;
THRHANDLE hMcstThread = 0;
THRHANDLE hSchdThrd = 0;

static int mcstRecvThreadActive = FALSE;
static int glbRecvThreadActive = FALSE;
static int srvRunningPriority = STD_THREAD_PRIORITY;
static int clnRunningPriority = STD_THREAD_PRIORITY;
static int bkgRunningPriority = STD_THREAD_PRIORITY;
static int cycle_priority_fixed=FALSE;
#if defined(UNIX)
#if !defined(PTHREAD_MUTEX_RECURSIVE) && defined(PTHREAD_MUTEX_RECURSIVE_NP)
# define PTHREAD_MUTEX_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP
#endif
pthread_attr_t gSendStreamThreadAttr, *gPtrSendStreamThreadAttr = &gSendStreamThreadAttr;
pthread_attr_t gRecvStreamThreadAttr, *gPtrRecvStreamThreadAttr = &gRecvStreamThreadAttr;
sem_t gSendStreamSemaphore;
SEMHANDLE hSendStreamSemaphore = NULL;
pthread_attr_t gClnCycleThreadAttr, *gPtrClnCycleThreadAttr = &gClnCycleThreadAttr;
pthread_attr_t gSrvCycleThreadAttr, *gPtrSrvCycleThreadAttr = &gSrvCycleThreadAttr;
pthread_attr_t gSchedulerThreadAttr, *gPtrSchedulerThreadAttr = &gSchedulerThreadAttr;
pthread_attr_t gSckSendThreadAttr, *gPtrSckSendThreadAttr = &gSckSendThreadAttr;
pthread_attr_t gSckRecvThreadAttr, *gPtrSckRecvThreadAttr = &gSckRecvThreadAttr;
pthread_attr_t gGlbRecvThreadAttr, *gPtrGlbRecvThreadAttr = &gGlbRecvThreadAttr;
pthread_attr_t gDcsEvnRecvThreadAttr, *gPtrDcsEvnRecvThreadAttr = &gDcsEvnRecvThreadAttr;
pthread_attr_t gMcstRecvThreadAttr, *gPtrMcstRecvThreadAttr = &gMcstRecvThreadAttr;
pthread_attr_t gTineGenericThreadAttr, *gPtrTineGenericThreadAttr = &gTineGenericThreadAttr;
pthread_attr_t gReinitThreadAttr, *gPtrReinitThreadAttr = &gReinitThreadAttr;
sem_t gSchedulerSemaphore;
SEMHANDLE hSchedulerSemaphore = NULL;
sem_t gNotificationSemaphore;
SEMHANDLE hNotificationSemaphore = NULL;
pthread_attr_t gBkgThreadAttr, *gPtrBkgThreadAttr = &gBkgThreadAttr;
pthread_attr_t gClThreadAttr, *gPtrClThreadAttr = &gClThreadAttr;
pthread_mutex_t gSystemInitMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSystemInitMutex = &gSystemInitMutex;
pthread_mutex_t gSystemServerMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSystemServerMutex = &gSystemServerMutex;
pthread_mutex_t gSystemKernelMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSystemKernelMutex = &gSystemKernelMutex;
pthread_mutex_t gSystemClientMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSystemClientMutex = &gSystemClientMutex;
pthread_mutex_t gSystemGlobalsUpdateMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSystemGlobalsUpdateMutex = &gSystemGlobalsUpdateMutex;
pthread_mutex_t gSystemMCastUpdateMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSystemMCastUpdateMutex = &gSystemMCastUpdateMutex;
pthread_mutex_t gLinkCbMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLinkCbMutex = &gLinkCbMutex;
pthread_mutex_t gLinkTblMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLinkTblMutex = &gLinkTblMutex;
pthread_mutex_t gXferMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hXferMutex = &gXferMutex;
pthread_mutex_t gSndMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSndMutex = &gSndMutex;
pthread_mutex_t gDpdLinkMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hDpdLinkMutex = &gDpdLinkMutex;
pthread_mutex_t gMcaLinkMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hMcaLinkMutex = &gMcaLinkMutex;
pthread_mutex_t gLwdLinkMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLwdLinkMutex = &gLwdLinkMutex;
pthread_mutex_t gLinkDataMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLinkDataMutex = &gLinkDataMutex;
pthread_mutex_t gLinkLstMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLinkLstMutex = &gLinkLstMutex;
pthread_mutex_t gCycleTriggerMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hCycleTriggerMutex = &gCycleTriggerMutex;
pthread_mutex_t gLogMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLogMutex = &gLogMutex;
pthread_mutex_t gDbMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hDbMutex = &gDbMutex;
pthread_mutex_t gCsvMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hCsvMutex = &gCsvMutex;
pthread_mutex_t gMfMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hMfMutex = &gMfMutex;
pthread_mutex_t gLinkQueueMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hLinkQueueMutex = &gLinkQueueMutex;
pthread_mutex_t gHistoryTableMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hHistoryTableMutex = &gHistoryTableMutex;
pthread_mutex_t gHistoryCycleMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hHistoryCycleMutex = &gHistoryCycleMutex;
pthread_mutex_t gClientTableMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hClientTableMutex = &gClientTableMutex;
pthread_mutex_t gContractTableMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hContractTableMutex  = &gContractTableMutex;
pthread_mutex_t gAlmTblMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hAlmTblMutex = &gAlmTblMutex;
pthread_mutex_t gTcpBucketMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hTcpBucketMutex = &gTcpBucketMutex;
pthread_mutex_t gEnsMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hEnsMutex = &gEnsMutex;
pthread_mutex_t gTxMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hTxMutex = &gTxMutex;
pthread_mutex_t gSrvTblMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSrvTblMutex = &gSrvTblMutex;
pthread_mutex_t gStkAccessMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hStkAccessMutex = &gStkAccessMutex;
pthread_mutex_t gAttrMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hAttrMutex = &gAttrMutex;
pthread_mutex_t gGeneralMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hGeneralMutex = &gGeneralMutex;
pthread_mutex_t gDataProtectionMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t gEqmBusyMutex = PTHREAD_MUTEX_INITIALIZER;
int setAttrParams(const char *tag,pthread_attr_t *attr,int priority)
{
  char txt[64];
  struct sched_param param;
  int cc = 0;
  if (attr == NULL) return invalid_parameter;
  sprintf(txt,"set %s thread priority",tag);
  param.sched_priority = priority;
  if (pthread_attr_setinheritsched(attr, PTHREAD_EXPLICIT_SCHED) != 0) ccerr(thread_error);
  if (pthread_attr_setschedpolicy(attr, SCHED_RR) != 0) ccerr(thread_error);
  if (pthread_attr_setschedparam(attr, &param) != 0) ccerr(thread_error);
  feclog("%s to %d",txt,priority);
err:
  if (cc && cc != EALREADY) rprtSckErr("setAttrParams",txt,0,errno,0,NULL);
  return cc;
}
void SetThreadName(char *name)
{
# ifndef NO_THREAD_NAMES
# ifdef MACOS
  pthread_setname_np(name);
# else
# if defined(__GLIBC__) && __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 12 && defined(_GNU_SOURCE)
  pthread_t tid = pthread_self();
  pthread_setname_np(tid,name);
# endif /* GLIBC OK */
# endif /* MACOS defined */
# else
  name = name;
# endif /* NO_THREAD_NAMES not defined */
}
int initAttr(const char *tag,pthread_attr_t *pattr,int priority)
{
  int cc = 0;
  if ((cc=pthread_attr_init(pattr)) != 0) ccerr(cc);
  if (priority != STD_THREAD_PRIORITY) cc = setAttrParams(tag,pattr,priority);
err:
  return cc;
}
int fixSystemCyclePriority(void)
{
  int cc;
  char txt[64];
  struct sched_param param;
  cycle_priority_fixed = TRUE;
  if (clnThreadPriority == STD_THREAD_PRIORITY) return 0;
  sprintf(txt,"set system cycle thread priority");
  param.sched_priority = clnThreadPriority;
  if (pthread_setschedparam(pthread_self(), SCHED_RR, &param) != 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot set system cycle thread priority (are you not running as root?)");
    bkgThreadPriority = srvThreadPriority = clnThreadPriority = STD_THREAD_PRIORITY;
    ccerr(thread_error);
  }
  clnRunningPriority = clnThreadPriority;
  feclog("%s to %d",txt,clnThreadPriority);
err:
  if (cc) rprtSckErr("fixSystemCyclePriority",txt,0,errno,0,NULL);
  return cc;
}
static THRHANDLE cycThrd = (THRHANDLE)0;
int isAllowedSystemCycle(void)
{
  if (cycThrd == (THRHANDLE)0) 
  { /* as yet unknown */
    if (WaitForSystemMutex(hSystemClientMutex,-1) != 0) return FALSE;
    if (cycThrd == (THRHANDLE)0) cycThrd = pthread_self();
    ReleaseSystemMutex(hSystemClientMutex);
  }
  return cycThrd == pthread_self() ? TRUE : FALSE;
}
void resetSystemCycleThread(void)
{
  cycThrd = (THRHANDLE)0;
}
THRHANDLE bkgCreateThread(BkgThreadFcn *tsk)
{
  int cc;
  pthread_t tid;
  cc = initAttr("background",gPtrBkgThreadAttr,bkgThreadPriority);
  pthread_attr_setdetachstate(gPtrBkgThreadAttr, PTHREAD_CREATE_DETACHED);
  bkgRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : bkgThreadPriority;
  cc = pthread_create(&tid,gPtrBkgThreadAttr,bkgThreadTask,(void *)tsk);
  if (cc != 0) feclog("could not create background thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE sendStreamCreateThread(ClnHdr *cln,BYTE *stream,UINT32 msgsize)
{
  int cc=0,haveMutex=FALSE,cap=getStreamCapacity(msgsize),ptcc;
  pthread_t tid;
  TCPBCKT *bckt=NULL;
  initAttr("stream",gPtrSendStreamThreadAttr,srvThreadPriority);
  if ((bckt=findBucket(cln->ep.sck,FALSE)) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot create sendStreamCreateThread : socket %d no has no buffer space",cln->ep.sck);
    ccerr(non_existent);
  }
  if (bckt->sndSem == NULL) 
  {
    if ((bckt->sndSem=CreateSystemSemaphore("sckStrm")) == NULL) return 0;
    ReleaseSystemSemaphore(bckt->sndSem);
  }
  if (bckt->sndMtx == NULL)
  {
    pthread_mutex_init(&bckt->psndMtx, NULL);
    bckt->sndMtx = &bckt->psndMtx;
    if (InitMutex(&bckt->sndMtx) != 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream mutex : %s",strerror(errno));
      bckt->sndMtx = NULL;
      return 0;
    }
  }
  if (WaitForSemaphore(bckt->sndSem,0) != 0) return 0;
  if (WaitForMutex(bckt->sndMtx,gSystemTick) != 0) ccerr(semaphore_busy);
  haveMutex = TRUE;
  if (msgsize == 0) ccerr(invalid_data);
  if (bckt->sndPrm.payload == NULL)
  {
    memcpy(&bckt->sndPrm.dst,cln,sizeof(ClnHdr));
    if ((bckt->sndPrm.payload=(BYTE *)SystemCalloc(cap,1)) == NULL) ccerr(out_of_server_memory);
    bckt->sndPrm.capacity = cap;
  }
  if (msgsize > (UINT32)bckt->sndPrm.capacity)
  {
    BYTE *data=NULL;
    if ((data=(BYTE *)realloc(bckt->sndPrm.payload,cap)) == NULL) ccerr(out_of_server_memory);
    bckt->sndPrm.payload = data;
    bckt->sndPrm.capacity = cap;
  }
  bckt->sndPrm.length = (int)msgsize;
  memcpy(bckt->sndPrm.payload,stream,msgsize);
  pthread_attr_setdetachstate(gPtrSendStreamThreadAttr, PTHREAD_CREATE_DETACHED);
  ptcc = pthread_create(&tid,gPtrSendStreamThreadAttr,sendStreamThreadTask,(void *)bckt);
  ReleaseSystemMutex(bckt->sndMtx); haveMutex = FALSE;
  if (ptcc == 0) return (THRHANDLE)tid;
err:
  if (haveMutex) ReleaseSystemMutex(bckt->sndMtx);
  ReleaseSystemSemaphore(bckt->sndSem);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot send stream : %.32s",cc2str(cc));
  else feclogEx(TINE_LOGLEVEL_ERROR, "could not create send stream thread : %s",strerror(errno));
  return 0;
}
THRHANDLE recvStreamCreateThread(SOCKET sck,int type)
{
  INTINT *thrdinf = (INTINT *)SystemCalloc(1, sizeof(INTINT));
  int cc = 0;
  pthread_t tid = 0;
  if (thrdinf == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "could not allocate memory for recv stream thread for socket %d", sck);
    ccerr(out_of_local_memory);
  }
  initAttr("recv stream",gPtrRecvStreamThreadAttr,clnThreadPriority);
  thrdinf->i1val = sck; thrdinf->i2val = type;
  pthread_attr_setdetachstate(gPtrRecvStreamThreadAttr, PTHREAD_CREATE_DETACHED);
  cc = pthread_create(&tid,gPtrRecvStreamThreadAttr,recvStreamThreadTask,(void *)thrdinf);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create recv stream thread for socket %d: %s",sck,strerror(errno));
err:
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE clnCycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc;
  if (!gNeedToStartClnCycleThread) return (THRHANDLE)0;
  gNeedToStartClnCycleThread = FALSE;
  cc = initAttr("client cycle",gPtrClnCycleThreadAttr,clnThreadPriority);
  clnRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : clnThreadPriority;
  cc = pthread_create(&tid,gPtrClnCycleThreadAttr,clnThreadTask,lpvThreadParam);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create client cycle thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE srvCycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc;
  if (!gNeedToStartSrvCycleThread) return (THRHANDLE)0;
  gNeedToStartSrvCycleThread = FALSE;
  cc = initAttr("server cycle",gPtrSrvCycleThreadAttr,srvThreadPriority);
  srvRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : srvThreadPriority;
  cc = pthread_create(&tid,gPtrSrvCycleThreadAttr,srvThreadTask,lpvThreadParam);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create server cycle thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE scavengerCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc;
  if (gScavangerThreadBusy) return (THRHANDLE)0;
  gScavangerThreadBusy = TRUE;
  pthread_attr_init(gPtrTineGenericThreadAttr);
  pthread_attr_setdetachstate(gPtrTineGenericThreadAttr, PTHREAD_CREATE_DETACHED);
  cc = pthread_create(&tid, gPtrTineGenericThreadAttr,scvThreadTask,lpvThreadParam);
  if (cc != 0) feclog("could not create scavanger thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE reinitCreateThread(void)
{
  int cc;
  pthread_t tid;
  cc = initAttr("reinit",gPtrReinitThreadAttr,srvThreadPriority);
  cc = pthread_create(&tid,gPtrReinitThreadAttr,reinitThreadTask,NULL);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create re-init thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
int initNotifierSemaphore(void)
{
  if (hNotificationSemaphore == NULL) 
  {
    if ((hNotificationSemaphore=CreateSystemSemaphore("notifier")) == NULL) return semaphore_error;
  }
  return 0;
}
THRHANDLE schCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc;
  if (hSchedulerSemaphore == NULL) 
  {
    if ((hSchedulerSemaphore=CreateSystemSemaphore("sched")) == NULL) return 0;
  }
  initAttr("scheduler",gPtrSchedulerThreadAttr,srvThreadPriority);
  cc = pthread_create(&tid,gPtrSchedulerThreadAttr,schThreadTask,lpvThreadParam);
  if (cc != 0) feclog("could not create scheduler thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE sckSendCreateThread(SOCKET sck,int type,SEMHANDLE *semHandle)
{
  pthread_t tid;
  int cc;
  SEMHANDLE hSem;
  SckThreadInfo *stinf;
  if (semHandle == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "parameter error (NULL semaphore pointer)");
    return (THRHANDLE)0;
  }
  hSem = *semHandle;
  if (hSem == NULL) 
  {
    if ((hSem=CreateSystemSemaphore("sckSend")) == NULL) return 0;
    *semHandle = hSem;  
  }
  if ((stinf=SystemCalloc(1, sizeof(SckThreadInfo))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for SckSend Thread");
    return (THRHANDLE)0;
  }
  stinf->sck = sck;
  stinf->type = type;
  stinf->sem = hSem;
  initAttr("tcp send",gPtrSckSendThreadAttr,srvThreadPriority);
  pthread_attr_setdetachstate(gPtrSckSendThreadAttr, PTHREAD_CREATE_DETACHED);
  cc = pthread_create(&tid,gPtrSckSendThreadAttr,sckSendThreadTask,(void *)stinf);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create tcp send thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE sckRecvCreateThread(SOCKET sck,int type)
{
  pthread_t tid;
  int cc;
  SckThreadInfo *stinf;
  if ((stinf=SystemCalloc(1, sizeof(SckThreadInfo))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for SckSend Thread");
    return (THRHANDLE)0;
  }
  stinf->sck = sck;
  stinf->type = type;
  stinf->sem = 0;
  initAttr("tcp recv",gPtrSckRecvThreadAttr,clnThreadPriority);
  pthread_attr_setdetachstate(gPtrSckRecvThreadAttr, PTHREAD_CREATE_DETACHED);
  cc = pthread_create(&tid,gPtrSckRecvThreadAttr,sckRecvThreadTask,(void *)stinf);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create tcp recv thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE glbRecvCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc;
  if (glbRecvThreadActive) return (THRHANDLE)0;
  glbRecvThreadActive = TRUE;
  initAttr("globals",gPtrGlbRecvThreadAttr,clnThreadPriority);
  cc = pthread_create(&tid,gPtrGlbRecvThreadAttr,glbRecvThreadTask,lpvThreadParam);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create globals thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
THRHANDLE mcstRecvCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc;
  if (mcstRecvThreadActive) return (THRHANDLE)0;
  mcstRecvThreadActive = TRUE;
  initAttr("multicast",gPtrMcstRecvThreadAttr,clnThreadPriority);
  cc = pthread_create(&tid,gPtrMcstRecvThreadAttr,mcstRecvThreadTask,lpvThreadParam);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create multicast thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
void YieldThread(void)
{
  sched_yield();
}
THRHANDLE GetThreadHandle()
{
  return (THRHANDLE)pthread_self();
}
SEMHANDLE CreateSystemSemaphore(char *name)
{
  SEMHANDLE sem;
# ifndef FREEBSD
  sem = (SEMHANDLE)SystemCalloc(1,sizeof(sem_t));
# else
  char semName[32];
  sprintf(semName,"/%s_%d",name,(int)getpid());
  sem = sem_open(semName,O_CREAT,S_IRWXU|S_IRWXG|S_IRWXO,0);
  if (sem == SEM_FAILED)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "could not create %s semaphore : %s",name,strerror(errno));
    return NULL;
  }
# endif
  if (sem_init(sem,0,0) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "could not initialize %s semaphore : %s",name,strerror(errno));
    return NULL;
  }
  return sem;
}
int InitMutex(MXHANDLE *mx)
{
  pthread_mutexattr_t attr;
  if (mx == NULL) return -1;
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  return pthread_mutex_init(*mx, &attr);
}
int CloseSystemMutex(MXHANDLE mx)
{
  if (mx == (MXHANDLE)0) return -1;
  if (!gSystemMutexSetInitialized) return -1;
  return pthread_mutex_destroy(mx);
}
int eqmInitMutexSet(void *lpvMutexParam)
{
  int cc;
  ExportListStruct *el = (ExportListStruct *)lpvMutexParam;
  if (el == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "eqmInitMutexSet: non existent element");
    return argument_list_error;
  }
  feclog("%s : intializing Equipment module mutex set",el->EqmName);
  if (_gNeedWarmRestart)
  { /* a warm restart : just exit */
    return 0;
  }
  if (el->hDataProtectionMutex != 0) 
  {
    feclogEx(TINE_LOGLEVEL_INFO, "%s : eqmInitMutexSet: mutex set already initialized",el->EqmName);
    return already_assigned;
  }
  el->DataProtectionMutex = gDataProtectionMutex;
  el->hDataProtectionMutex = &el->DataProtectionMutex;
  if ((cc=InitMutex(&el->hDataProtectionMutex)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s : Cannot create Data Protection Mutex: error %d",el->EqmName,cc);
    return cc;
  }
  el->EqmBusyMutex = gEqmBusyMutex;
  el->hEqmBusyMutex = &el->EqmBusyMutex;
  if ((cc=InitMutex(&el->hEqmBusyMutex)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s : Cannot create Eqm Busy Mutex: error %d",el->EqmName,cc);
    return cc;
  }
  feclog("%s : Equipment module mutex set initialized",el->EqmName);
  return 0;
}
THRHANDLE clCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  ContractListStruct *cl = (ContractListStruct *)lpvThreadParam;
  int cc;
  static int done = FALSE;
  if (cl == NULL) return 0;
  if (!done)
  {
    done = TRUE;
    initAttr("conlst",gPtrClThreadAttr,srvThreadPriority);
#   ifdef FORCE_PTHREAD_SCOPE_SYSTEM
    if ((cc=pthread_attr_setscope(gPtrClThreadAttr,PTHREAD_SCOPE_SYSTEM)) != 0) 
    {
      feclog("Cannot set contract list thread scope !");
    }
#   endif
    pthread_attr_setdetachstate(gPtrClThreadAttr, PTHREAD_CREATE_DETACHED);
  }
  cc = pthread_create(&tid,gPtrClThreadAttr,clThreadTask,lpvThreadParam);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "could not create contract list thread : %s",strerror(errno));
  return cc ? 0 : (THRHANDLE)tid;
}
int WaitForMutex(MXHANDLE mx,int to)
{
# ifndef NO_MUTEX_TIMEDLOCK
  struct timespec ts;
  UINT64 nsecs;
# endif
  int cc = -1;
  struct timeval tv0, tv1;
  gettimeofday(&tv0,(struct timezone *)NULL); tv1 = tv0;
  if (mx == (MXHANDLE)0) return 0;
  if (to < 0)
  {
    int tries = 0;
    while ((cc=pthread_mutex_lock(mx)) != 0)
    {
      if (tries++ == 0) feclogEx(TINE_LOGLEVEL_WARN,"%s : %.128s", getSystemMutexName(mx), strerror(errno));
      if (tries > 10)
      {
        feclogEx(TINE_LOGLEVEL_WARN, "abandon wait after %d retries", tries);
        break;
      }
      sched_yield();
    }
    return cc;
  }
# ifndef NO_MUTEX_TIMEDLOCK
  /* make sure both tv1 and tv0 get used in order to avoid a warning */
  ts.tv_sec = tv0.tv_sec + to/1000;
  nsecs = tv1.tv_usec * 1000UL + (to % 1000) * 1000000;
  if (nsecs >= 1000000000)
  { /* leads to an EINVAL if > 1E9 */
    ts.tv_sec += 1;
    nsecs -= 1000000000;
  }
  ts.tv_nsec = nsecs;
  cc = pthread_mutex_timedlock(mx,&ts);
# else
  while ((UINT32)MSECS(tv1,tv0) <= to)
  {
    if ((cc=pthread_mutex_trylock(mx)) == 0 || to == 0) return cc;
    millisleep(1);
    gettimeofday(&tv1,(struct timezone *)NULL);
  }
# endif
  return cc;
}
int WaitForSemaphore(SEMHANDLE mx,int to)
{
# ifdef NO_MUTEX_TIMEDLOCK
  int cc = -1;
  struct timeval tv0, tv1;
# else
  struct timespec ts;
# endif
  if (to < 0) return sem_wait(mx);
# ifndef NO_MUTEX_TIMEDLOCK
  ts.tv_sec = to/1000;
  ts.tv_nsec = (to%1000) * 1000000;
  return sem_timedwait(mx,&ts);
# else
  gettimeofday(&tv0,(struct timezone *)NULL); tv1 = tv0;
  while ((UINT32)MSECS(tv1,tv0) <= to)
  {
    if ((cc=sem_trywait(mx)) == 0 || to == 0) return cc;
    sched_yield();
    gettimeofday(&tv1,(struct timezone *)NULL);
  }
  return cc;
# endif
}
int FreeSystemSemaphore(SEMHANDLE mx)
{
  int cc;
  if (mx == (SEMHANDLE)0) return 0;
  cc = sem_destroy(mx);
# ifndef FREEBSD
  SystemFree(mx);
# endif
  return cc ? 0 : errno;
}
#elif defined(WIN32)
DWORD dwBkgThreadId,dwClThreadId,dwSendStreamThreadId,
      dwRecvStreamThreadId,dwClnThreadId,dwSrvThreadId,dwSchThreadId,
      dwScvThreadId,dwReinitThreadId;
SEMHANDLE hSendStreamSemaphore;
SEMHANDLE hSchedulerSemaphore;
SEMHANDLE hNotificationSemaphore;
MXHANDLE hSystemInitMutex;
MXHANDLE hSystemServerMutex;
MXHANDLE hSystemKernelMutex;
MXHANDLE hCycleTriggerMutex;
MXHANDLE hSystemClientMutex;
MXHANDLE hSystemGlobalsUpdateMutex;
MXHANDLE hSystemMCastUpdateMutex;
MXHANDLE hLinkCbMutex;
MXHANDLE hTcpBucketMutex;
MXHANDLE hEnsMutex;
MXHANDLE hTxMutex;
MXHANDLE hAttrMutex;
MXHANDLE hGeneralMutex;
MXHANDLE hSrvTblMutex;
MXHANDLE hStkAccessMutex;
MXHANDLE hLinkTblMutex;
MXHANDLE hXferMutex;
MXHANDLE hSndMutex;
MXHANDLE hDpdLinkMutex;
MXHANDLE hMcaLinkMutex;
MXHANDLE hLwdLinkMutex;
MXHANDLE hLinkDataMutex;
MXHANDLE hLinkLstMutex;
MXHANDLE hLinkQueueMutex;
MXHANDLE hClientTableMutex;
MXHANDLE hContractTableMutex;
MXHANDLE hLogMutex;
MXHANDLE hDbMutex;
MXHANDLE hCsvMutex;
MXHANDLE hMfMutex;
MXHANDLE hHistoryTableMutex;
MXHANDLE hHistoryCycleMutex;
MXHANDLE hAlmTblMutex;

#ifndef NO_THREAD_NAMES
typedef struct tagTHREADNAME_INFO
{
   DWORD dwType; /* must be 0x1000 */
   LPCSTR szName; /* pointer to name (in user addr space) */
   DWORD dwThreadID; /* thread ID (-1=caller thread) */
   DWORD dwFlags; /* reserved for future use, must be zero */
} THREADNAME_INFO;

void winSetThreadName(DWORD dwThreadID, LPCSTR szThreadName)
{
   THREADNAME_INFO info;
   info.dwType = 0x1000;
   info.szName = szThreadName;
   info.dwThreadID = dwThreadID;
   info.dwFlags = 0;

   __try
   {
      RaiseException( 0x406D1388, 0, sizeof(info)/sizeof(DWORD), (ULONG_PTR *)&info );
   }
   __except(EXCEPTION_CONTINUE_EXECUTION)
   {
   }
}
#else
void winSetThreadName(DWORD dwThreadID, LPCSTR szThreadName)
{
  dwThreadID = dwThreadID;
  szThreadName = szThreadName;
}
#endif 
void SetThreadName(char *name)
{
  winSetThreadName(-1,name);
}
int initAttr(const char *tag,HANDLE hThr,int priority)
{
  int cc = 0;
  char txt[64];
  if (priority == STD_THREAD_PRIORITY) return 0;
  if (hThr == NULL) ccerr(thread_error);
  sprintf(txt,"set %s thread priority",tag);
  if (SetThreadPriority(hThr,priority) == 0) ccerr(thread_error);
  feclog("%s to %d",txt,priority);
err:
  if (cc) rprtSckErr("initAttr",txt,0,errno,0,NULL);
  return cc;
}
int fixSystemCyclePriority(void)
{
  int cc;
  HANDLE hThr = GetCurrentThread();
  cycle_priority_fixed = TRUE;
  cc = initAttr("system cycle",hThr,clnThreadPriority);
  clnRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : clnThreadPriority;
  return cc;
}
static THRHANDLE cycThrd = (THRHANDLE)0;
int isAllowedSystemCycle(void)
{
  static HANDLE cycThrd = 0;
  if (cycThrd == (THRHANDLE)0) 
  { /* as yet unknown */
    if (WaitForSystemMutex(hSystemClientMutex,-1) != 0) return FALSE;
    if (cycThrd == (THRHANDLE)0) cycThrd = GetCurrentThread();
    ReleaseSystemMutex(hSystemClientMutex);
  }
  return cycThrd == GetCurrentThread() ? TRUE : FALSE;
}
void resetSystemCycleThread(void)
{
  if (cycThrd != (THRHANDLE)0) CloseHandle(cycThrd);
  cycThrd = (THRHANDLE)0;
}
THRHANDLE bkgCreateThread(BkgThreadFcn *tsk)
{
  int cc;
  HANDLE hThr = CreateThread(NULL,0,bkgThreadTask,(LPVOID)tsk,0,&dwBkgThreadId);
  cc = initAttr("background",hThr,bkgThreadPriority);
  bkgRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : bkgThreadPriority;
  return (THRHANDLE)hThr;
}
THRHANDLE sendStreamCreateThread(ClnHdr *cln,BYTE *stream,UINT32 msgsize)
{
  THRHANDLE tid;
  int cc=0, haveMutex=FALSE, cap=getStreamCapacity(msgsize);
  TCPBCKT *bckt=NULL;
  if ((bckt=findBucket(cln->ep.sck,FALSE)) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot create sendStreamCreateThread : socket %d no has no buffer space",cln->ep.sck);
    return 0;
  }
  if (bckt->sndSem == NULL) 
  {
    if ((bckt->sndSem=CreateSystemSemaphore("sckStrm")) == NULL) return 0;
    ReleaseSystemSemaphore(bckt->sndSem);
  }
  if (bckt->sndMtx == NULL && InitMutex(&bckt->sndMtx) != 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream mutex : <%d>",GetLastError());
    return 0;
  }
  if (WaitForSemaphore(bckt->sndSem,0) != 0) return 0;
  if (WaitForMutex(bckt->sndMtx,gSystemTick) != 0) ccerr(semaphore_busy);
  haveMutex = TRUE;
  if (msgsize == 0) ccerr(invalid_data);
  if (bckt->sndPrm.payload == NULL)
  {
    memcpy(&bckt->sndPrm.dst,cln,sizeof(ClnHdr));
    if ((bckt->sndPrm.payload=(BYTE *)SystemCalloc(1,cap)) == NULL) ccerr(out_of_server_memory);
    bckt->sndPrm.capacity = cap;
  }
  if (msgsize > (UINT32)bckt->sndPrm.capacity)
  {
    BYTE *data=NULL;
    if ((data=(BYTE *)realloc(bckt->sndPrm.payload,cap)) == NULL) ccerr(out_of_server_memory);
    bckt->sndPrm.payload = data;
    bckt->sndPrm.capacity = cap;
  }
  bckt->sndPrm.length = (int)msgsize;
  memcpy(bckt->sndPrm.payload,stream,msgsize);
  tid = (THRHANDLE)CreateThread(NULL,0,sendStreamThreadTask,(LPVOID)bckt,0,&dwSendStreamThreadId);
  initAttr("stream",(HANDLE)tid,srvThreadPriority);
  ReleaseSystemMutex(bckt->sndMtx); haveMutex = FALSE;
  if (tid != 0) return tid;
err:
  if (haveMutex) ReleaseSystemMutex(bckt->sndMtx);
  ReleaseSystemSemaphore(bckt->sndSem);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot send stream : %.32s",cc2str(cc));
  else feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream thread: <%d>",GetLastError());
  return 0;
}
THRHANDLE recvStreamCreateThread(SOCKET sck,int type)
{
  INTINT *thrdinf = (INTINT *)SystemCalloc(1,sizeof(INTINT));
  THRHANDLE tid = (THRHANDLE)0;
  int cc = 0;
  if (thrdinf == NULL) ccerr(out_of_local_memory);
  thrdinf->i1val = (int)sck; thrdinf->i2val = type;
  tid = (THRHANDLE)CreateThread(NULL,0,recvStreamThreadTask,(LPVOID)thrdinf,0,&dwRecvStreamThreadId);
  if (tid == 0) ccerr(GetLastError());
  initAttr("stream",(HANDLE)tid,clnThreadPriority);
err:
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream thread for socket %d: <%d>",sck,cc);
  return tid;
}
void YieldThread(void)
{
  Sleep(0);
}
THRHANDLE GetThreadHandle()
{
  return (THRHANDLE)GetCurrentThread();
}
SEMHANDLE CreateSystemSemaphore(char *name)
{
  SEMHANDLE sem = CreateSemaphore(NULL,0,1,NULL);
  if (sem == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot create %s semaphore : <%d>",name,GetLastError());
    return NULL;
  }
  return sem;
}
int InitMutex(MXHANDLE *mx)
{
  if (mx == NULL) return -1;
  *mx = CreateMutex(NULL,FALSE,NULL);
  if (*mx == NULL) return -2;
  return 0;
}
int WaitForMutex(MXHANDLE mx,int to)
{
  int cc;
  if (mx == (MXHANDLE)0) return 0;
  cc = WaitForSingleObject(mx,(to < 0 ? INFINITE : (UINT32)to));
  if (cc == WAIT_FAILED) cc = GetLastError();
  return cc == WAIT_OBJECT_0 ? 0 : cc;
}
int WaitForSemaphore(SEMHANDLE mx,int to)
{
  return WaitForMutex((MXHANDLE)mx,to);
}
int FreeSystemSemaphore(SEMHANDLE mx)
{
  int cc;
  if (mx == (SEMHANDLE)0) return 0;
  cc = CloseHandle(mx);
  return (cc == 0) ? GetLastError() : 0;
}
int CloseSystemMutex(MXHANDLE mx)
{
  if (!gSystemMutexSetInitialized) return not_initialized;
  return CloseHandle(mx) == 0 ? mutex_error : 0;
}
int eqmInitMutexSet(void *lpvMutexParam)
{
  int cc;
  ExportListStruct *el = (ExportListStruct *)lpvMutexParam;
  if (el == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "eqmInitMutexSet: non existent element");
    return argument_list_error;
  }
  feclog("%s : intializing Equipment module mutex set",el->EqmName);
  if (_gNeedWarmRestart)
  { /* a warm restart : just exit */
    return 0;
  }
  if (el->hDataProtectionMutex != 0) 
  {
    feclogEx(TINE_LOGLEVEL_INFO, "%s : eqmInitMutexSet: mutex set already initialized",el->EqmName);
    return already_assigned;
  }
  if ((cc=InitMutex(&el->hDataProtectionMutex)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s : Cannot create Data Protection Mutex: error %d",el->EqmName,cc);
    return cc;
  }
  if ((cc=InitMutex(&el->hEqmBusyMutex)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s : Cannot create Eqm Busy Mutex: error %d",el->EqmName,cc);
    return cc;
  }
  feclog("%s : Equipment module mutex set initialized",el->EqmName);
  return 0;
}
THRHANDLE clCreateThread(THREAD_PARAM lpvThreadParam)
{
  THRHANDLE tid;
  ContractListStruct *cl = (ContractListStruct *)lpvThreadParam;
  if (cl == NULL) return 0;
  tid = (THRHANDLE)CreateThread(NULL,0,clThreadTask,lpvThreadParam,0,&dwClThreadId);
  initAttr("conlst",(HANDLE)tid,srvThreadPriority);
  return tid;
}
THRHANDLE clnCycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  int cc;
  THRHANDLE tid;
  if (!gNeedToStartClnCycleThread) return (THRHANDLE)0;
  gNeedToStartClnCycleThread = TRUE;
  tid = (THRHANDLE)CreateThread(NULL,0,clnThreadTask,lpvThreadParam,0,&dwClnThreadId);
  cc = initAttr("client cycle",(HANDLE)tid,clnThreadPriority);
  clnRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : clnThreadPriority;
  return tid;
}
THRHANDLE srvCycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  int cc;
  THRHANDLE tid;
  if (!gNeedToStartSrvCycleThread) return (THRHANDLE)0;
  gNeedToStartSrvCycleThread = FALSE;
  tid = (THRHANDLE)CreateThread(NULL,0,srvThreadTask,lpvThreadParam,0,&dwSrvThreadId);
  cc = initAttr("server cycle",(HANDLE)tid,srvThreadPriority);
  srvRunningPriority = (cc != 0) ? STD_THREAD_PRIORITY : srvThreadPriority;
  return tid;
}
THRHANDLE scavengerCreateThread(THREAD_PARAM lpvThreadParam)
{
  if (gScavangerThreadBusy) return (THRHANDLE)0;
  gScavangerThreadBusy = TRUE;
  return (THRHANDLE)CreateThread(NULL,0,scvThreadTask,lpvThreadParam,0,&dwScvThreadId);
}
THRHANDLE reinitCreateThread(void)
{
  return (THRHANDLE)CreateThread(NULL,0,reinitThreadTask,NULL,0,&dwReinitThreadId);
}
int initNotifierSemaphore(void)
{
  if (hNotificationSemaphore == NULL) 
  {
    if ((hNotificationSemaphore=CreateSystemSemaphore("notifier")) == NULL) return semaphore_error;
  }
  return 0;
}
THRHANDLE schCreateThread(THREAD_PARAM lpvThreadParam)
{
  THRHANDLE hThr;
  if (hSchedulerSemaphore == NULL) 
  {
    hSchedulerSemaphore = CreateSemaphore(NULL,0,1,NULL);
    if (hSchedulerSemaphore == NULL)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot create scheduler semaphore : <%d>",GetLastError());
      return (THRHANDLE)0;
    }
  }
  hThr = (THRHANDLE)CreateThread(NULL,0,schThreadTask,lpvThreadParam,0,&dwSchThreadId);
  initAttr("scheduler",(HANDLE)hThr,srvThreadPriority);
  return hThr;
}
THRHANDLE sckSendCreateThread(SOCKET sck,int type,SEMHANDLE *semHandle)
{
  THRHANDLE hThr;
  SEMHANDLE hSem;
  DWORD dwThrId;
  SckThreadInfo *stinf;
  if (semHandle == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "parameter error (NULL semaphore pointer)");
    return (THRHANDLE)0;
  }
  hSem = *semHandle;
  if (hSem == NULL) 
  {
    if ((hSem=CreateSystemSemaphore("sckSend")) == NULL) return (THRHANDLE)0;
    *semHandle = hSem;
  }
  if ((stinf=SystemCalloc(1,sizeof(SckThreadInfo))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for SckSend Thread");
    return (THRHANDLE)0;
  }
  stinf->sck = sck;
  stinf->type = type;
  stinf->sem = hSem;
  hThr = (THRHANDLE)CreateThread(NULL,0,sckSendThreadTask,(LPVOID)stinf,0,&dwThrId);
  initAttr("tcp send",(HANDLE)hThr,srvThreadPriority);
  return hThr;
}
THRHANDLE sckRecvCreateThread(SOCKET sck,int type)
{
  THRHANDLE hThr;
  DWORD dwThrId;
  SckThreadInfo *stinf;
  if ((stinf=SystemCalloc(1, sizeof(SckThreadInfo))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for SckSend Thread");
    return (THRHANDLE)0;
  }
  stinf->sck = sck;
  stinf->type = type;
  stinf->sem = 0;
  hThr = (THRHANDLE)CreateThread(NULL,0,sckRecvThreadTask,(LPVOID)stinf,0,&dwThrId);
  initAttr("tcp recv",(HANDLE)hThr,clnThreadPriority);
  return hThr;
}
THRHANDLE glbRecvCreateThread(THREAD_PARAM lpvThreadParam)
{
  THRHANDLE hThr;
  DWORD dwThrId;
  if (glbRecvThreadActive) return (THRHANDLE)0;
  glbRecvThreadActive = TRUE;
  hThr = (THRHANDLE)CreateThread(NULL,0,glbRecvThreadTask,lpvThreadParam,0,&dwThrId);
  initAttr("globals",(HANDLE)hThr,clnThreadPriority);
  return hThr;
}
THRHANDLE mcstRecvCreateThread(THREAD_PARAM lpvThreadParam)
{
  THRHANDLE hThr;
  DWORD dwThrId;
  if (mcstRecvThreadActive) return (THRHANDLE)0;
  mcstRecvThreadActive = TRUE;
  hThr = (THRHANDLE)CreateThread(NULL,0,mcstRecvThreadTask,lpvThreadParam,0,&dwThrId);
  initAttr("multicast",(HANDLE)hThr,clnThreadPriority);
  return hThr;
}
#elif defined(VXWORKS)
DWORD dwBkgThreadId,dwClThreadId;
SEMHANDLE hSendStreamSemaphore = NULL;
SEMHANDLE hSchedulerSemaphore = NULL;
SEMHANDLE hNotificationSemaphore = NULL;
MXHANDLE hSystemInitMutex;
MXHANDLE hSystemServerMutex;
MXHANDLE hSystemKernelMutex;
MXHANDLE hCycleTriggerMutex;
MXHANDLE hSystemClientMutex;
MXHANDLE hSystemGlobalsUpdateMutex;
MXHANDLE hSystemMCastUpdateMutex;
MXHANDLE hLinkCbMutex;
MXHANDLE hTcpBucketMutex;
MXHANDLE hEnsMutex;
MXHANDLE hTxMutex;
MXHANDLE hAttrMutex;
MXHANDLE hGeneralMutex;
MXHANDLE hSrvTblMutex;
MXHANDLE hStkAccessMutex;
MXHANDLE hLinkTblMutex;
MXHANDLE hXferMutex;
MXHANDLE hSndMutex;
MXHANDLE hDpdLinkMutex;
MXHANDLE hMcaLinkMutex;
MXHANDLE hLwdLinkMutex;
MXHANDLE hLinkDataMutex;
MXHANDLE hLinkLstMutex;
MXHANDLE hLinkQueueMutex;
MXHANDLE hClientTableMutex;
MXHANDLE hContractTableMutex;
MXHANDLE hLogMutex;
MXHANDLE hDbMutex;
MXHANDLE hCsvMutex;
MXHANDLE hMfMutex;
MXHANDLE hHistoryTableMutex;
MXHANDLE hHistoryCycleMutex;
MXHANDLE hAlmTblMutex;
int gVxwDefaultTaskPriority = VXWTSK_PRIORITY;
int gVxwDefaultTaskOptions = VXWTSK_FLAGS;
int gVxwDefaultTaskStack = VXWTSK_STACK;
int validateTaskPriority(int p)
{
  if (p < 1) p = 1;
  if (p > 1000) p = 1000;
  return p;
}
void SetThreadName(char *name)
{ /* vxworks tasks have already been named */
  name = name;
}
THRHANDLE vxwCreateThread(char *tname,FUNCPTR thr,int lpvThreadParam,int priority)
{
  THRHANDLE h;
  char str[128];
  int tskid = taskIdSelf();
  int tskp = gVxwDefaultTaskPriority, tsko = gVxwDefaultTaskOptions;
  int tcks = sysClkRateGet()/100;
  taskPriorityGet(tskid,&tskp);
  taskOptionsGet(tskid,&tsko);
  if (tskp == 1) 
  { /* in case started from the shell prompt */
    tskp = gVxwDefaultTaskPriority;
    tsko = gVxwDefaultTaskOptions;
  }
  if (priority != STD_THREAD_PRIORITY) 
  { /* an 'offset' priority was passed */
    tskp = validateTaskPriority(tskp + priority);
    feclog("set %s task priority to %d",tname,tskp);
  }
# if (_WRS_VXWORKS_MAJOR > 5)
  tsko &= ~(VX_DEALLOC_STACK|VX_DEALLOC_EXC_STACK|VX_DEALLOC_TCB);
# endif
  if (tcks < 1) tcks = 1;
  kernelTimeSlice(tcks);
  h = (THRHANDLE)taskSpawn(tname,tskp,tsko,gVxwDefaultTaskStack,thr,lpvThreadParam,0,0,0,0,0,0,0,0,0);
  if (h == ERROR) 
  {
    sprintf(str,"taskSpawn %.32s",tname);
    perror(str);
  }
  return h;
}
int fixSystemCyclePriority(void)
{ /* nothing to do here for vxworks */
  cycle_priority_fixed = TRUE;
  return 0;
}
static THRHANDLE cycThrd = (THRHANDLE)0;
int isAllowedSystemCycle(void)
{ /* fix this if there's ever a need .... */
  return TRUE;
}
void resetSystemCycleThread(void)
{
  cycThrd = (THRHANDLE)0;
}
THRHANDLE bkgCreateThread(BkgThreadFcn *tsk)
{
  static int cnt=0;
  char tname[32],ename[8];
  if (tsk == NULL) return -1;
  cnt++;
  strncpy(ename,((ExportListStruct *)tsk->elRef)->EqmName,6); ename[6] = 0;
  if (tsk->elRef != NULL) sprintf(tname,"t%.6sbkg",strupr(ename));
  else sprintf(tname,"t%dbkg",cnt);
  return vxwCreateThread(tname,(FUNCPTR)bkgThreadTask,(int)tsk,bkgThreadPriority);
}
THRHANDLE sendStreamCreateThread(ClnHdr *cln,BYTE *stream,UINT32 msgsize)
{
  char tname[32];
  THRHANDLE tid;
  int cc=0, haveMutex=FALSE, cap=getStreamCapacity(msgsize);
  TCPBCKT *bckt=NULL;
  if ((bckt=findBucket(cln->ep.sck,FALSE)) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot create sendStreamCreateThread : socket %d no has no buffer space",cln->ep.sck);
    ccerr(non_existent);
  }
  sprintf(tname,"tTCPSTRM");
  if (bckt->sndSem == NULL) 
  {
    bckt->sndSem = semBCreate(SEM_Q_FIFO,SEM_FULL);
    if (bckt->sndSem == NULL)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream semaphore : <%d>",errno);
      return semaphore_error;
    }
    ReleaseSystemSemaphore(bckt->sndSem);
  }
  if (bckt->sndMtx == NULL && InitMutex(&bckt->sndMtx) != 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream mutex : <%d>",errno);
    return 0;
  }
  if (WaitForSemaphore(bckt->sndSem,0) != 0) return 0;
  if (WaitForMutex(bckt->sndMtx,gSystemTick) != 0) ccerr(semaphore_busy);
  haveMutex = TRUE;
  if (msgsize == 0) ccerr(invalid_data);
  if (bckt->sndPrm.payload == NULL)
  {
    memcpy(&bckt->sndPrm.dst,cln,sizeof(ClnHdr));
    if ((bckt->sndPrm.payload=(BYTE *)SystemCalloc(cap,1)) == NULL) ccerr(out_of_server_memory);
    bckt->sndPrm.capacity = cap;
  }
  if (msgsize > (UINT32)bckt->sndPrm.capacity)
  {
    BYTE *data=NULL;
    if ((data=(BYTE *)realloc(bckt->sndPrm.payload,cap)) == NULL) ccerr(out_of_server_memory);
    bckt->sndPrm.payload = data;
    bckt->sndPrm.capacity = cap;
  }
  bckt->sndPrm.length = (int)msgsize;
  memcpy(bckt->sndPrm.payload,stream,msgsize);
  tid = vxwCreateThread(tname,(FUNCPTR)sendStreamThreadTask,(int)bckt,srvThreadPriority);
  ReleaseSystemMutex(bckt->sndMtx); haveMutex = FALSE;
  if (tid != (THRHANDLE)0) return tid;
err:
  if (haveMutex) ReleaseSystemMutex(bckt->sndMtx);
  ReleaseSystemSemaphore(bckt->sndSem);
  if (cc != 0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot send stream : %.32s",cc2str(cc));
  else feclogEx(TINE_LOGLEVEL_ERROR, "cannot create stream thread");
  return 0;
}
THRHANDLE recvStreamCreateThread(SOCKET sck,int type)
{
  INTINT *thrdinf = (INTINT *)SystemCalloc(1, sizeof(INTINT));
  char tname[32];
  THRHANDLE tid = (THRHANDLE)0;
  if (thrdinf == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for recv stream thread for socket %d",sck);
    return tid;
  }
  sprintf(tname,"tTCPrcv:%d",sck);
  thrdinf->i1val = sck; thrdinf->i2val = type;
  tid = vxwCreateThread(tname,(FUNCPTR)recvStreamThreadTask,(int)thrdinf,clnThreadPriority);
  if (tid == (THRHANDLE)0) feclogEx(TINE_LOGLEVEL_ERROR, "cannot create recv stream thread for socket %d",sck);
  return tid;
}
void YieldThread(void)
{
  taskDelay(0);
}
THRHANDLE GetThreadHandle()
{
  return (THRHANDLE)taskIdSelf();
}
int InitMutex(MXHANDLE *mx)
{ /* use recusive mutex semaphores in general */
  if (mx == NULL) return -1;
  *mx = semMCreate(SEM_Q_FIFO);
  if (*mx == NULL) return -2;
  return 0;
}
int WaitForMutex(MXHANDLE mx,int to)
{ /* timeout is in msec and semTake() takes ticks */
  int cc, tps, twait = 0;
  if (mx == (MXHANDLE)0) return 0;
  if (to < 0)
  {
    twait = WAIT_FOREVER;
  }
  else if (to > 0) 
  {
    tps = sysClkRateGet();
    twait = (tps * to)/1000;
    if (twait == 0) twait = 1;
  }
  cc = semTake(mx,twait);
  return cc == OK ? 0 : cc;
}
int vxwReleaseSemaphore(MXHANDLE mx)
{
  int cc = semGive(mx);
  taskDelay(0);
  return cc;
}
int WaitForSemaphore(SEMHANDLE mx,int to)
{ 
  return WaitForMutex((MXHANDLE)mx,to);
}
int CloseSystemMutex(MXHANDLE mx)
{
  int cc = semDelete(mx);
  return cc == OK ? 0 : cc;
}
int FreeSystemSemaphore(SEMHANDLE mx)
{
  if (mx == (SEMHANDLE)0) return 0;
  return CloseSystemMutex(mx);
}
int eqmInitMutexSet(void *lpvMutexParam)
{
  int cc;
  ExportListStruct *el = (ExportListStruct *)lpvMutexParam;
  if (el == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "eqmInitMutexSet: non existent element");
    return argument_list_error;
  }
  feclog("%s : intializing Equipment module mutex set",el->EqmName);
  if (_gNeedWarmRestart)
  { /* a warm restart : just exit */
    return 0;
  }
  if (el->hDataProtectionMutex != 0) 
  {
    feclogEx(TINE_LOGLEVEL_INFO, "%s : eqmInitMutexSet: mutex set already initialized",el->EqmName);
    return already_assigned;
  }
  if ((cc=InitMutex(&el->hDataProtectionMutex)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s : Cannot create Data Protection Mutex: error %d",el->EqmName,cc);
    return cc;
  }
  if ((cc=InitMutex(&el->hEqmBusyMutex)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "%s : Cannot create Eqm Busy Mutex: error %d",el->EqmName,cc);
    return cc;
  }
  feclog("%s : Equipment module mutex set initialized",el->EqmName);
  return 0;
}
THRHANDLE clCreateThread(THREAD_PARAM lpvThreadParam)
{
  ContractListStruct *cl = (ContractListStruct *)lpvThreadParam;
  char tname[32];
  if (cl == NULL) return 0;
  sprintf(tname,"tCL.%d",cl->contractID);
  return vxwCreateThread(tname,(FUNCPTR)clThreadTask,(int)lpvThreadParam,srvThreadPriority);
}
THRHANDLE clnCycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  if (!gNeedToStartClnCycleThread) return (THRHANDLE)0;
  gNeedToStartClnCycleThread = TRUE;
  return vxwCreateThread("tClnCycThr",(FUNCPTR)clnThreadTask,(int)lpvThreadParam,clnThreadPriority);
}
THRHANDLE srvCycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  if (!gNeedToStartSrvCycleThread) return (THRHANDLE)0;
  gNeedToStartSrvCycleThread = FALSE;
  return vxwCreateThread("tSrvCycThr",(FUNCPTR)srvThreadTask,(int)lpvThreadParam,srvThreadPriority);
}
THRHANDLE reinitCreateThread(void)
{
  return vxwCreateThread("tReinitThr",(FUNCPTR)reinitThreadTask,0,STD_THREAD_PRIORITY);
}
int initNotifierSemaphore(void)
{
  if (hNotificationSemaphore == NULL) 
  {
    hNotificationSemaphore = semBCreate(SEM_Q_FIFO,SEM_FULL);
    if (hNotificationSemaphore == NULL)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot create notifier semaphore : <%d>",errno);
      return semaphore_error;
    }
  }
  return 0;
}
THRHANDLE schCreateThread(THREAD_PARAM lpvThreadParam)
{
  if (hSchedulerSemaphore == NULL) 
  {
    hSchedulerSemaphore = semBCreate(SEM_Q_FIFO,SEM_FULL);
    if (hSchedulerSemaphore == NULL)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot create scheduler semaphore : <%d>",errno);
      return semaphore_error;
    }
  }
  gNeedToStartSrvCycleThread = FALSE;
  return vxwCreateThread("tSchdlrThr",(FUNCPTR)schThreadTask,(int)lpvThreadParam,srvThreadPriority);
}
THRHANDLE sckSendCreateThread(SOCKET sck,int type,SEMHANDLE *semHandle)
{
  char thrName[32];
  SEMHANDLE hSem;
  SckThreadInfo *stinf;
  if (semHandle == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "parameter error (NULL semaphore pointer)");
    return (THRHANDLE)0;
  }
  hSem = *semHandle;
  if (hSem == NULL) 
  {
    hSem = semBCreate(SEM_Q_FIFO,SEM_FULL);
    if (hSem == NULL)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot create socket send semaphore : <%d>",errno);
      return (THRHANDLE)0;
    }
    *semHandle = hSem;
  }
  if ((stinf=SystemCalloc(1, sizeof(SckThreadInfo))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for SckSend Task");
    return (THRHANDLE)0;
  }
  stinf->sck = sck;
  stinf->type = type;
  stinf->sem = hSem;
  sprintf(thrName,"tSckSnd_%d",sck);
  return vxwCreateThread(thrName,(FUNCPTR)sckSendThreadTask,(int)(stinf),srvThreadPriority);
}
THRHANDLE sckRecvCreateThread(SOCKET sck,int type)
{
  char thrName[32];
  SckThreadInfo *stinf;
  if ((stinf=SystemCalloc(1, sizeof(SckThreadInfo))) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot allocate memory for SckSend Thread");
    return (THRHANDLE)0;
  }
  stinf->sck = sck;
  stinf->type = type;
  stinf->sem = 0;
  sprintf(thrName,"tSckRcv_%d",sck);
  return vxwCreateThread(thrName,(FUNCPTR)sckRecvThreadTask,(int)(stinf),clnThreadPriority);
}
THRHANDLE glbRecvCreateThread(THREAD_PARAM lpvThreadParam)
{
  if (glbRecvThreadActive) return (THRHANDLE)0;
  glbRecvThreadActive = TRUE;
  return vxwCreateThread("tGlbRecv",(FUNCPTR)glbRecvThreadTask,(int)lpvThreadParam,clnThreadPriority);
}
THRHANDLE mcstRecvCreateThread(THREAD_PARAM lpvThreadParam)
{
  if (mcstRecvThreadActive) return (THRHANDLE)0;
  mcstRecvThreadActive = TRUE;
  return vxwCreateThread("tMcstRecv",(FUNCPTR)mcstRecvThreadTask,(int)lpvThreadParam,clnThreadPriority);
}
#endif /* UNIX or WIN32 or VXWORKS */
void stopGlobalsThread(SOCKET sck)
{
  static char stopSignal = 0;
  int tolen = getInAddrLen(tGCastSckAddr.ss_family);
  if (sck == glbClnSck)
  {
    if (sck != (SOCKET)0) sendto(sck, &stopSignal, 1, 0, (struct sockaddr*)&tGCastSckAddr, tolen);
    return;
  }
  if (sck == mcastClnSck)
  {
    SCKADR tMCastSckAddr;
    sckadr_in(&tMCastSckAddr, (BYTE*)&tMCastAddr, (UINT16)gtMCastPort, tGCastSckAddr.ss_family);
    if (sck != (SOCKET)0) sendto(sck, &stopSignal, 1, 0, (struct sockaddr*)&tMCastSckAddr, tolen);
    return;
  }
  return;
}
int dumpThreadPriorities(void)
{
  ttyoutput("globals       : %d",clnRunningPriority);
  ttyoutput("multicast     : %d",clnRunningPriority);
  ttyoutput("tcp recv      : %d",clnRunningPriority);
  ttyoutput("stream recv   : %d",clnRunningPriority);
  if (gRunServerCycleInSeparateThread)
  {
    ttyoutput("server cycle  : %d",srvRunningPriority);
    ttyoutput("client cycle  : %d",clnRunningPriority);
  }
  else
  {
    ttyoutput("system cycle  : %d",clnRunningPriority);
  }
  if (ExportList != NULL)
  {
    ttyoutput("tcp send      : %d",srvRunningPriority);
    ttyoutput("stream send   : %d",srvRunningPriority);
    ttyoutput("scheduler     : %d",srvRunningPriority);
    ttyoutput("stock handler : %d",srvRunningPriority);
    ttyoutput("eqm handler   : %d",srvRunningPriority);
    ttyoutput("background io : %d",bkgRunningPriority);
  }
  return 0;
}
int srvThreadPriority = STD_THREAD_PRIORITY;
int clnThreadPriority = STD_THREAD_PRIORITY;
int bkgThreadPriority = STD_THREAD_PRIORITY;
int _setThreadPriority(const char *tag,int *tgt,int priority,int revoke)
{
  if (revoke)
  {
    feclogEx(TINE_LOGLEVEL_WARN, "%s thread active! Call to set priority from %d to %d refused",tag,*tgt,priority);
    return srvThreadPriority;
  }
  if (priority == *tgt) return priority;
  if (priority < MIN_THREAD_PRIORITY) priority = MIN_THREAD_PRIORITY;
  if (priority > MAX_THREAD_PRIORITY) priority = MAX_THREAD_PRIORITY;
  *tgt = priority;
  feclog("%s thread priority set from %d to %d",tag,*tgt,priority);
  return priority;
}
TINE_EXPORT int SetServerThreadPriority(int priority)
{
  return _setThreadPriority("Server",&srvThreadPriority,priority,gSrvThreadRunning);
}
TINE_EXPORT int GetServerThreadPriority(void)
{
  return srvThreadPriority;
}
TINE_EXPORT int SetClientThreadPriority(int priority)
{
  return _setThreadPriority("Client",&clnThreadPriority,priority,gClnThreadRunning);
}
TINE_EXPORT int GetClientThreadPriority(void)
{
  return clnThreadPriority;
}
TINE_EXPORT int SetBackgroundThreadPriority(int priority)
{
  return _setThreadPriority("Background",&bkgThreadPriority,priority,gSrvThreadRunning);
}
TINE_EXPORT int GetBackgroundThreadPriority(void)
{
  return bkgThreadPriority;
}
TINE_EXPORT int SetKernelPriority(int priority)
{
  int cc=not_accepted;
  if ((_setThreadPriority("Server",&srvThreadPriority,priority,gSrvThreadRunning)) != priority) goto err;
  if ((_setThreadPriority("Client",&clnThreadPriority,priority,gClnThreadRunning)) != priority) goto err;
  if ((_setThreadPriority("Background",&bkgThreadPriority,priority,gSrvThreadRunning)) != priority) goto err;
  cc = 0;
err:
  return cc;
}
int gSystemMutexSetInitialized = FALSE;
char *getSystemMutexName(MXHANDLE mx)
{
  ExportListStruct *el;
  if (mx == NULL) return "null mutex handle!";
  if (mx == hSystemInitMutex) return "SystemInitMutex";
  if (mx == hSystemServerMutex) return "SystemServerMutex";
  if (mx == hSystemKernelMutex) return "SystemKernelMutex";
  if (mx == hCycleTriggerMutex) return "CycleTriggerMutex";
  if (mx == hSystemClientMutex) return "SystemClientMutex";
  if (mx == hSystemGlobalsUpdateMutex) return "SystemGlobalsUpdateMutex";
  if (mx == hSystemMCastUpdateMutex) return "SystemMCastUpdateMutex";
  if (mx == hLinkCbMutex) return "LinkCbMutex";
  if (mx == hLinkTblMutex) return "LinkTblMutex";
  if (mx == hXferMutex) return "XferMutex";
  if (mx == hSndMutex) return "SndToPeerMutex";
  if (mx == hDpdLinkMutex) return "DpdLinkMutex";
  if (mx == hMcaLinkMutex) return "McaLinkMutex";
  if (mx == hLwdLinkMutex) return "LwdLinkMutex";
  if (mx == hLinkDataMutex) return "LinkDataMutex";
  if (mx == hLinkLstMutex) return "LinkLstMutex";
  if (mx == hTcpBucketMutex) return "TcpBucketMutex";
  if (mx == hEnsMutex) return "EnsMutex";
  if (mx == hTxMutex) return "TxMutex";
  if (mx == hAttrMutex) return "AttrMutex";
  if (mx == hGeneralMutex) return "GeneralMutex";
  if (mx == hSrvTblMutex) return "SrvTblMutex";
  if (mx == hStkAccessMutex) return "StkAccessMutex";
  if (mx == hLinkQueueMutex) return "LinkQueueMutex";
  if (mx == hClientTableMutex) return "ClientTableMutex";
  if (mx == hContractTableMutex) return "ContractTableMutex";
  if (mx == hLogMutex) return "LogMutex";
  if (mx == hDbMutex) return "DbMutex";
  if (mx == hCsvMutex) return "CsvMutex";
  if (mx == hMfMutex) return "MfMutex";
  if (mx == hHistoryTableMutex) return "HistoryTableMutex";
  if (mx == hHistoryCycleMutex) return "HistoryCycleMutex";
  if (mx == hAlmTblMutex) return "AlmTableMutex";
  for (el=ExportList; el != NULL; el=el->next) 
  {
    if (mx == el->hDataProtectionMutex) return "DataProtectionMutex";
    if (mx == el->hEqmBusyMutex) return "EqmBusyMutex";
  }
  return "unknown mutex";
}
int isAllowedMutexError(MXHANDLE mx)
{
  if (mx == hLogMutex) return TRUE;
  if (mx == hDbMutex) return TRUE;
  return FALSE;
}
int InitSystemMutexSet(void)
{
  int cc=0;
  MXHANDLE mx=0;
  if (!gSystemMutexSetInitialized)
  {
    initNotifierSemaphore();
    mx = hSystemInitMutex; if ((cc=InitMutex(&hSystemInitMutex)) != 0) ccerr(cc);
    mx = hSystemServerMutex; if ((cc=InitMutex(&hSystemServerMutex)) != 0) ccerr(cc);
    mx = hSystemKernelMutex; if ((cc=InitMutex(&hSystemKernelMutex)) != 0) ccerr(cc);
    mx = hCycleTriggerMutex; if ((cc=InitMutex(&hCycleTriggerMutex)) != 0) ccerr(cc);
    mx = hSystemClientMutex; if ((cc=InitMutex(&hSystemClientMutex)) != 0) ccerr(cc);
    mx = hSystemGlobalsUpdateMutex; if ((cc=InitMutex(&hSystemGlobalsUpdateMutex)) != 0) ccerr(cc);
    mx = hSystemMCastUpdateMutex; if ((cc=InitMutex(&hSystemMCastUpdateMutex)) != 0) ccerr(cc);
    mx = hLinkCbMutex; if ((cc=InitMutex(&hLinkCbMutex)) != 0) ccerr(cc);
    mx = hLinkTblMutex; if ((cc=InitMutex(&hLinkTblMutex)) != 0) ccerr(cc);
    mx = hXferMutex; if ((cc=InitMutex(&hXferMutex)) != 0) ccerr(cc);
    mx = hSndMutex; if ((cc=InitMutex(&hSndMutex)) != 0) ccerr(cc);
    mx = hDpdLinkMutex; if ((cc=InitMutex(&hDpdLinkMutex)) != 0) ccerr(cc);
    mx = hMcaLinkMutex; if ((cc=InitMutex(&hMcaLinkMutex)) != 0) ccerr(cc);
    mx = hLwdLinkMutex; if ((cc=InitMutex(&hLwdLinkMutex)) != 0) ccerr(cc);
    mx = hLinkDataMutex; if ((cc=InitMutex(&hLinkDataMutex)) != 0) ccerr(cc);
    mx = hLinkLstMutex; if ((cc=InitMutex(&hLinkLstMutex)) != 0) ccerr(cc);
    mx = hTcpBucketMutex; if ((cc=InitMutex(&hTcpBucketMutex)) != 0) ccerr(cc);
    mx = hEnsMutex; if ((cc=InitMutex(&hEnsMutex)) != 0) ccerr(cc);
    mx = hTxMutex; if ((cc=InitMutex(&hTxMutex)) != 0) ccerr(cc);
    mx = hAttrMutex; if ((cc=InitMutex(&hAttrMutex)) != 0) ccerr(cc);
    mx = hGeneralMutex; if ((cc=InitMutex(&hGeneralMutex)) != 0) ccerr(cc);
    mx = hSrvTblMutex; if ((cc=InitMutex(&hSrvTblMutex)) != 0) ccerr(cc);
    mx = hStkAccessMutex; if ((cc=InitMutex(&hStkAccessMutex)) != 0) ccerr(cc);
    mx = hLinkQueueMutex; if ((cc=InitMutex(&hLinkQueueMutex)) != 0) ccerr(cc);
    mx = hClientTableMutex; if ((cc=InitMutex(&hClientTableMutex)) != 0) ccerr(cc);
    mx = hContractTableMutex; if ((cc=InitMutex(&hContractTableMutex)) != 0) ccerr(cc);
    mx = hLogMutex; if ((cc=InitMutex(&hLogMutex)) != 0) ccerr(cc);
    mx = hDbMutex; if ((cc=InitMutex(&hDbMutex)) != 0) ccerr(cc);
    mx = hCsvMutex; if ((cc=InitMutex(&hCsvMutex)) != 0) ccerr(cc);
    mx = hMfMutex; if ((cc=InitMutex(&hMfMutex)) != 0) ccerr(cc);
    mx = hHistoryTableMutex; if ((cc=InitMutex(&hHistoryTableMutex)) != 0) ccerr(cc);
    mx = hHistoryCycleMutex; if ((cc=InitMutex(&hHistoryCycleMutex)) != 0) ccerr(cc);
    mx = hAlmTblMutex; if ((cc=InitMutex(&hAlmTblMutex)) != 0) ccerr(cc);
    gSystemMutexSetInitialized = TRUE;
    gStockResourceBusy = FALSE;
    feclog("system mutexes initialized");
  }
err:
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR, "init Mutex %s error %d",mx == 0 ? "stkMutexes" : getSystemMutexName(mx));
  return cc;
}
#if defined(UNIX)
# define RESETMUTEX(m) CloseSystemMutex(m)
#else
# define RESETMUTEX(m) { CloseSystemMutex(m); m = 0; }
#endif
int CloseClientMutexSet(void)
{
  if (hGlbThread != (THRHANDLE)0) WaitForThreadTerminationEx(hGlbThread,2000);
  if (hMcstThread != (THRHANDLE)0) WaitForThreadTerminationEx(hMcstThread,2000);
  RESETMUTEX(hSystemClientMutex);
  RESETMUTEX(hSystemGlobalsUpdateMutex);
  hGlbThread = (THRHANDLE)0;
  hMcstThread = (THRHANDLE)0;
  return 0;
}
int WaitForThreadTerminationEx(THRHANDLE hThr,int timeout)
{
# if defined(WIN32)
  int cc = 0;
  DWORD threadExitCode;
  if (hThr == 0) return 0;
  WaitForSingleObject(hThr,timeout < 0 ? INFINITE : (UINT32)timeout);
  if (GetExitCodeThread(hThr,&threadExitCode) != 0 && 
      threadExitCode == STILL_ACTIVE) cc = (int)threadExitCode;
  CloseHandle(hThr);
  return cc;
# elif defined(UNIX)
  if (hThr == 0) return 0;
  /* TODO: there's no timed join, but there are kluges */
  UNUSED(timeout);
  return pthread_join(hThr,NULL);
# else
  return 0;
# endif
}
int WaitForThreadTermination(THRHANDLE hThr)
{
  return WaitForThreadTerminationEx(hThr,-1);
}
static void ThreadExit(int exitCode)
{
# if defined(WIN32)
  ExitThread((DWORD)exitCode);
# elif defined(UNIX)
  pthread_exit(&exitCode);
# else
  exitCode = exitCode;
# endif
}
int CloseSystemMutexSet(void)
{
  ExportListStruct *el;

  if (!gSystemMutexSetInitialized) return not_initialized;
  FreeSystemSemaphore(hSendStreamSemaphore); hSendStreamSemaphore = (SEMHANDLE)0;
  FreeSystemSemaphore(hSchedulerSemaphore); hSchedulerSemaphore = (SEMHANDLE)0;
  FreeSystemSemaphore(hNotificationSemaphore); hNotificationSemaphore = (SEMHANDLE)0;

  RESETMUTEX(hSystemMCastUpdateMutex);
  RESETMUTEX(hLinkCbMutex);
  RESETMUTEX(hSystemInitMutex);
  RESETMUTEX(hSystemServerMutex);
  RESETMUTEX(hSystemKernelMutex);
  RESETMUTEX(hCycleTriggerMutex);
  RESETMUTEX(hLinkTblMutex);
  RESETMUTEX(hXferMutex);
  RESETMUTEX(hSndMutex);
  RESETMUTEX(hDpdLinkMutex);
  RESETMUTEX(hMcaLinkMutex);
  RESETMUTEX(hLwdLinkMutex);
  RESETMUTEX(hLinkDataMutex);
  RESETMUTEX(hLinkLstMutex);
  RESETMUTEX(hTcpBucketMutex);
  RESETMUTEX(hEnsMutex);
  RESETMUTEX(hTxMutex);
  RESETMUTEX(hAttrMutex);
  RESETMUTEX(hGeneralMutex);
  RESETMUTEX(hSrvTblMutex);
  RESETMUTEX(hStkAccessMutex);
  RESETMUTEX(hLinkQueueMutex);
  RESETMUTEX(hClientTableMutex);
  RESETMUTEX(hContractTableMutex);
  RESETMUTEX(hLogMutex);
  RESETMUTEX(hDbMutex);
  RESETMUTEX(hCsvMutex);
  RESETMUTEX(hMfMutex);
  RESETMUTEX(hHistoryTableMutex);
  RESETMUTEX(hHistoryCycleMutex);
  RESETMUTEX(hAlmTblMutex);

  for (el=ExportList; el != NULL; el=el->next) 
  {
    if (el->hEqmBusyMutex != 0) RESETMUTEX(el->hEqmBusyMutex);
    RESETMUTEX(el->hDataProtectionMutex);
  }
  gSystemMutexSetInitialized = FALSE;
  gStockResourceBusy = FALSE;
  return 0;
}
int ReleaseSystemMutex(MXHANDLE mx) 
{ 
  int cc = 0;
  if (mx == (MXHANDLE)0) return 0; /* already freed (?) */
# if defined(UNIX)
  cc = ReleaseMutex(mx);
# elif defined(VXWORKS)
  if ((cc=ReleaseMutex(mx)) == OK) cc = 0;
# elif defined(WIN32)
  if (ReleaseMutex(mx) == 0) cc = GetLastError();
# else
  ReleaseMutex(mx);
# endif
  if (cc != 0 && !ServerExitCondition)
  { /* at exit: don't log error: various threads will try to cleanup ... */
    if (!isAllowedMutexError(mx)) feclogEx(TINE_LOGLEVEL_ERROR, "ReleaseMutex %s failed : %d",getSystemMutexName(mx),cc);
  }
  return cc;
}
int ReleaseSystemSemaphore(SEMHANDLE mx) 
{ 
  int cc = 0;
  if (mx == (SEMHANDLE)0) return 0; /* already freed (?) */
# if defined(UNIX)
  cc = ReleaseSemaphore(mx);
# elif defined(VXWORKS)
  if ((cc=ReleaseSemaphore(mx)) == OK) cc = 0;
# elif defined(WIN32)
  if (ReleaseSemaphore(mx,1,NULL) == 0) cc = GetLastError();
  if (cc == ERROR_TOO_MANY_POSTS) cc = 0; /* don't fret */
# else
  ReleaseMutex(mx);
# endif
  if (cc != 0 && !ServerExitCondition)
  { /* at exit: don't log error: various threads will try to cleanup ... */
    if (tineDebug > 1 || errorDebug) dbglog("ReleaseSemaphore failed : %d",cc);
  }
  return cc;
}
THREAD_TYPE sendStreamThreadTask(THREAD_PARAM lpvThreadParam)
{ /* send payload as a TCP/IP Stream */
  TCPBCKT *bckt = (TCPBCKT *)lpvThreadParam;
  if (bckt == NULL) return 0;
  SetThreadName("SndStream");
  if (WaitForMutex(bckt->sndMtx,-1) == 0)
  {
    sendStreamData(&bckt->sndPrm.dst,bckt->sndPrm.payload,bckt->sndPrm.length);
    ReleaseSystemMutex(bckt->sndMtx);
  }
  ReleaseSystemSemaphore(bckt->sndSem);
  return 0;
}
void freeBucketThreadMemory(TCPBCKT *bckt)
{
  if (bckt == NULL) return;
  msglog(0, "freeing bucket thread memory for socket %d", bckt->sck);
  if (bckt->sndSem != (SEMHANDLE)0) FreeSystemSemaphore(bckt->sndSem);
  bckt->sndSem = (SEMHANDLE)0;
# ifdef WIN32
  if (bckt->sndThr != (THRHANDLE)0) CloseHandle(bckt->sndThr);
  bckt->sndThr = (THRHANDLE)0;
  if (bckt->rcvThr != (THRHANDLE)0) CloseHandle(bckt->rcvThr);
  bckt->rcvThr = (THRHANDLE)0;
# endif
  SystemFree(bckt->buf);
  SystemFree(bckt->clrbuf);
  SystemFree(bckt->sdTbl);
  SystemFree(bckt->sndPrm.payload);
  SystemFree(bckt);
}
THREAD_TYPE recvStreamThreadTask(THREAD_PARAM lpvThreadParam)
{ /* recv payload as a TCP/IP Stream */
  INTINT *thrdinf = (INTINT *)lpvThreadParam;
  SOCKET sck;
  int cc = 0, i, nread, nready;
  TCPBCKT *bckt=NULL;
  TCPSCKT *sckt=NULL;
  struct pollfd fds[1];
  
  SetThreadName("RcvStream");
  if (thrdinf != NULL) 
  {
    sck = thrdinf->i1val;
    if ((bckt=findBucket(sck,FALSE)) == NULL)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot start recvStreamThreadTask : socket %d no has no buffer space",sck);
      ccerr(non_existent);
    }
    bckt->sdThrState = SD_THREAD_RUNNING;
    while (bckt->sdThrState == SD_THREAD_RUNNING)
    { /* block on recv until a socket error happens */
      if (bckt == NULL) goto recvErr; /* can happen if reassigned */
      if (bckt->sdThrState == SD_THREAD_STOPPING) break;
      fds[0].fd = sck; fds[0].events = POLLIN; fds[0].revents = 0;
      if ((nready=poll(fds,1,1000)) < 0 || pollerr(0)) goto recvErr;
      if (nready == 0) continue;
      if ((nread=recv(sck,(char *)bckt->clrbuf,bckt->clrlen,0)) > 0)
      {
        processTCPDataStream(&bckt,nread);
        TCPcount++;
      }
      else
      {
        if (nread < 0) 
        { /* an error on the socket ! -> bail out */
recvErr:  
          sckt = getTcpClnSck(sck, -1, -1);
          if (!ServerExitCondition) for (i=0; i<nConnectionTableEntries; i++)
          { /* are any clients using this socket? */
            if (conTbl[i] == NULL) continue;
            if (conTbl[i]->tcpSck == sck)
            {
              conTbl[i]->linkStatus = net_read_error;
              if (BASEMODE(conTbl[i]->mode) == CM_SINGLE)
              { /* checkConnections will not cover this case */
                conTbl[i]->needsNotification = TRUE;
                conTbl[i]->mode = CM_CANCEL;
              }
            }
          }
          if (sckt != NULL) closeIPConnectionToFec(sckt->idx, sck);
          ccerr(tcp_socket_error);
        }
      }
    }
  }
  bckt->sdThrState = SD_THREAD_STOPPED;
err:
  removeBucketFromList(bckt);
  freeBucketThreadMemory(bckt);
  SystemFree(thrdinf);
  if (cc && !ServerExitCondition) feclogEx(TINE_LOGLEVEL_WARN, "recvStreamThreadTask: %.32s",cc2str(cc));
  return 0;
}
THREAD_TYPE schThreadTask(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;
  SetThreadName("SchdTask");
  while (!ServerExitCondition)
  {
    WaitForSemaphore(hSchedulerSemaphore,-1);
    doScheduler(&thisCycleTime,gSystemTick,SCHED_EXTERN,NULL,NULL);
  }
# if defined(WIN32)
  if (hSchdThrd != (THRHANDLE)0) CloseHandle(hSchdThrd);
# endif
  return 0;
}
int gSdtBusyCeiling = 0;
TINE_EXPORT void SetStreamTransportRetryLimit(int value)
{
  gSdtBusyCeiling = value;
}
TINE_EXPORT int GetStreamTransportRetryLimit(void) 
{ 
  return gSdtBusyCeiling; 
}
int streamQueueOverwriteEager = FALSE;
TINE_EXPORT void SetStreamQueueOverwriteEager(int value)
{
  streamQueueOverwriteEager = value;
}
TINE_EXPORT int GetStreamQueueOverwriteEager(void)
{
  return streamQueueOverwriteEager;
}
void resetStreamQueueBusyHeadToTail(TCPBCKT *bckt,int idx)
{
  int i;
  for (i=idx; i != bckt->sdTblTail; i = (i+1) % bckt->sdTblSize)
  { /* move to the tail and clear the busy flags */
    bckt->sdTbl[i].isbusy = FALSE;
    bckt->sdTbl[i].isready = FALSE;
  }
}
int putStreamParams(ClnHdr *c,BYTE *payload,UINT32 msgsize, int mark)
{
  TCPBCKT *bckt=NULL;
  StreamDataTable *sdt;
  int i, busycnt = 0, cc = 0;
  if ((bckt=findBucket(c->ep.sck,TRUE)) == NULL) return un_allocated;
  i = bckt->sdTblHead;
  sdt = &bckt->sdTbl[i];
  if (sdt->isbusy)
  { /* trouble: the queue is full of unsent data */
    if (bckt->sdTblMark > 0) 
    { /* at least there's a previous mark */
      if (tineDebug) dbglog("stream data queue full (socket %d): reset next transmission mark",c->ep.sck);
      /* reset the head */
      i = bckt->sdTblHead = bckt->sdTblMark - 1;
      resetStreamQueueBusyHeadToTail(bckt,i);
      sdt = &bckt->sdTbl[i];
      bckt->sdTblWaitMark = TRUE;
    }
    else
    {
      if (tineDebug) dbglog("stream data queue full (socket %d) with no next transmission mark !!",c->ep.sck);
      while (sdt->isbusy && busycnt++ < gSdtBusyCeiling) 
      {
        if (tineDebug) dbglog("stream data queue full (socket %d) -> wait %d msec",c->ep.sck,gSystemTick);
        millisleep(gSystemTick);
      }
      ccerr(resources_exhausted);
    }
  }
  if (mark) 
  { /* mark the start of a new data set (add 1 to avoid a '0' problem) */
    if (!sdt->isbusy && bckt->sdTblMark > 0 && streamQueueOverwriteEager) 
    { /* there is already a mark */
      if ((bckt->sdTblMark - bckt->sdTblTail + bckt->sdTblSize)%bckt->sdTblSize > 5)
      { /* looks like we're far enough along */
        i = bckt->sdTblHead = bckt->sdTblMark - 1;
        resetStreamQueueBusyHeadToTail(bckt,i);
        sdt = &bckt->sdTbl[i];
      }
    }
    bckt->sdTblMark = i + 1;
    bckt->sdTblWaitMark = FALSE;
  }
  if (bckt->sdTblWaitMark) ccerr(not_accepted);
  sdt->isbusy = TRUE;
  sdt->dst = *c;
  memcpy(sdt->payload,payload,msgsize);
  sdt->msgsize = (int)msgsize;
  sdt->isready = TRUE;
err:
  if (cc == 0) 
  { /* advance the head */
    bckt->sdTblHead = (bckt->sdTblHead + 1)%bckt->sdTblSize;
  }
  ReleaseSystemMutex(hTcpBucketMutex);
  return cc;
}
int freeStreamParams(SOCKET sck)
{
  int idx, cc=0;
  TCPBCKT *bckt=NULL;
  StreamDataTable *sdt;

  if ((bckt=findBucket(sck, TRUE)) == NULL)
  {
    msglog(non_existent, "could not find bucket for socket %d", sck);
    return un_allocated;
  }
  idx = bckt->sdTblTail;
  sdt = &bckt->sdTbl[idx];
  sdt->isready = FALSE;
  sdt->isbusy = FALSE;
  bckt->sdTblTail = (bckt->sdTblTail + 1)%bckt->sdTblSize;
  ReleaseSystemMutex(hTcpBucketMutex);
  return cc;
}
THREAD_TYPE sckSendThreadTask(THREAD_PARAM lpvThreadParam)
{
  SckThreadInfo *stinf = (SckThreadInfo *)lpvThreadParam;
  SEMHANDLE hSem;
  SOCKET sck;
  int cc=0,type;
  StreamDataTable *sdt;
  TCPBCKT *bckt=NULL;

  if (stinf == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "invalid information passed to socket send thread");
    return 0;
  }
  SetThreadName("SndTcp");
  hSem = stinf->sem;
  sck = stinf->sck;
  type = stinf->type;
  if ((bckt=findBucket(sck,FALSE)) == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot start sckSendThreadTask : socket %d no has no buffer space",sck); 
    return 0;
  }
  bckt->sdThrState = SD_THREAD_RUNNING;
  while (!ServerExitCondition && bckt->sdThrState == SD_THREAD_RUNNING)
  {
    WaitForSemaphore(hSem,-1);
    if (bckt->sdThrState == SD_THREAD_STOPPING) ccerr(0);
    while (bckt->sdTbl[bckt->sdTblTail].isready)
    { /* something needs to be sent out */
      if (bckt->sdThrState != SD_THREAD_RUNNING) ccerr(0);
      sdt = &bckt->sdTbl[bckt->sdTblTail];
      if (bckt->sdTblTail == bckt->sdTblMark - 1) 
      { /* clear the mark */
        bckt->sdTblMark = 0;
      }
      if ((cc=sendIpData(&sdt->dst,sdt->payload,sdt->msgsize,SRV_RSP_,type,FALSE,FALSE)) != 0)
      { /* wasn't delivered */
        if (cc == cannot_select) continue; /* no write buffers ? */
        if (cc == semaphore_busy) continue; /* hSndMutex locked in another thread */
        removeBucketFromList(bckt);
        ccerr(cc); /* the connection to peer was closed !? */
      }
      if (bckt->sdThrState != SD_THREAD_RUNNING) break;
      freeStreamParams(bckt->sck); /* advances the tail */
    }
  }
err:
  bckt->sdThrState = SD_THREAD_STOPPED;
  removeBucketFromList(bckt);
  freeBucketThreadMemory(bckt);
  SystemFree(stinf);
  return 0;
}
THREAD_TYPE sckRecvThreadTask(THREAD_PARAM lpvThreadParam)
{ /* recv payload as a TCP/IP connection on separate thread */
  SckThreadInfo *stinf = (SckThreadInfo *)lpvThreadParam;
  SOCKET sck;
  int type, nread=-1, nready, cc=0;
  TCPBCKT *bckt=NULL;
  TCPSCKT *sckt=NULL;
  struct pollfd fds[1];
  
  if (stinf == NULL) 
  {
    feclog("invalid information passed to socket send thread");
    return 0;
  }
  SetThreadName("RcvTcp");
  sck = stinf->sck;
  type = stinf->type;
  if ((bckt=findBucket(sck,FALSE)) == NULL)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot start sckRecvThreadTask : socket %d no has no buffer space",sck); 
    return 0;
  }
  if (tineDebug) dbglog("start %s sck recv thread ",getProtocolAsString(type));
  bckt->sdThrState = SD_THREAD_RUNNING;
  while (bckt->sdThrState == SD_THREAD_RUNNING)
  { /* block on recv until a socket error happens */
    if (bckt == NULL) goto recvErr; /* can happen if reassigned */
    if (bckt->sdThrState == SD_THREAD_STOPPING) break;
    fds[0].fd = sck; fds[0].events = POLLIN; fds[0].revents = 0;
    if ((nready=poll(fds, 1, 1000)) < 0 || pollerr(0))
    {
      if (errno == EINTR) continue;
      rprtSckErr("sckRecvThreadTask", "poll", sck, -1, 0, NULL);
      goto recvErr;
    }
    if (nready == 0) continue;
    if ((nread=recv(sck,(char *)bckt->clrbuf,bckt->clrlen,0)) > 0)
    {
      SCKADR sckadr;
      _socklen_t plen = getInAddrLen(getInetFamily(bckt->inetprot));
      if (getpeername(sck,(struct sockaddr *)&sckadr,&plen) < 0)
      {
        rprtSckErr("sckRecvThreadTask","getpeername",sck,-1,0,NULL);
        goto recvErr;
      }
      processTCPDataPacket(sck,bckt->clrbuf,nread,sckadr,TCP);
      TCPcount++;
    }
    else
    {
      if (nread <= 0) 
      { /* an error on the socket ! -> bail out */
recvErr:rprtSckErr("sckRecvThreadTask","recv",sck,-1,1,NULL);
        if (errno == EINTR) continue;
        sckt = getTcpClnSck(sck, -1, -1);
        if (sckt != NULL) closeIPConnectionToFec(sckt->idx, sck);
        ccerr(tcp_socket_error);
      }
    }
  }  
err:
  if (cc) feclogEx(TINE_LOGLEVEL_ERROR, "recv thread error on socket %d : %.32s",sck,cc2str(cc));
  bckt->sdThrState = SD_THREAD_STOPPED;
  removeBucketFromList(bckt);
  freeBucketThreadMemory(bckt);
  SystemFree(stinf);
  return 0;
}
#ifndef MCAST_THREAD_DELAY
# define MCAST_THREAD_DELAY 10
#endif
BYTE glbWorkArea[GCAST_BUFFER_SIZE];
THREAD_TYPE glbRecvThreadTask(THREAD_PARAM lpvThreadParam)
{ /* recv globals payload on separate thread */
  int mode = 0;
  SCKADR addr;
  BYTE *buf = glbWorkArea;
  int nready, nbytes;
  struct pollfd fds[1];

  _socklen_t len = sizeof(addr);
  lpvThreadParam = lpvThreadParam;
  glbRecvThreadActive = TRUE;
  if (soioctl(glbClnSck,FIONBIO,(char *)&mode)) dbglog("could not reset globals socket to blocking mode");
  SetThreadName("RcvGlb");
  while (!gClientIsClosing)
  { /* stay here until closing time */
    fds[0].fd = glbClnSck; fds[0].events = POLLIN; fds[0].revents = 0;
    if ((nready=poll(fds,1,MCAST_THREAD_DELAY)) < 0 || pollerr(0))
    {
      if (errno == EINTR) continue;
      goto sckErr;
    }
    if (nready == 0) continue;
    if ((nbytes=recvfrom(glbClnSck, (char *)buf, GCAST_BUFFER_SIZE, 0, (struct sockaddr *)&addr, &len)) <= 0)
    {
      if (errno == EINTR) continue;
      goto sckErr;
    }
    if (nbytes == 1 && *buf == 0)
    { /* received stop character: if gClientIsClosing == TRUE then we leave ! */
      if (tineDebug) dbglog("RcvGlb: received stop signal");
      continue; 
    }
    getGlobalUpdate(&addr,buf);
  }
sckErr:
  if (!ServerExitCondition && !gClientIsClosing) rprtSckErr("glbRecvThreadTask","recvfrom",glbClnSck,-1,0,NULL);
  glbRecvThreadActive = FALSE;
  if (tineDebug) dbglog("RcvGlb: stopping thread task");
  return 0;
}
THREAD_TYPE mcstRecvThreadTask(THREAD_PARAM lpvThreadParam)
{ /* recv mcast payload on separate thread */
  int p, mode = 0;
  SCKADR sckadr;
  int nready, nbytes;
  struct pollfd fds[1];

  _socklen_t len = sizeof(sckadr);
  lpvThreadParam = lpvThreadParam;
  mcstRecvThreadActive = TRUE;
  if (soioctl(mcastClnSck,FIONBIO,(char *)&mode)) dbglog("could not reset mcast socket to blocking mode");
  SetThreadName("RcvMcast");
  p = (gAfInet == AF_INET6) ? UDP6 : UDP;
  while (!gClientIsClosing)
  { /* stay here until closing time */
    fds[0].fd = mcastClnSck; fds[0].events = POLLIN;
    if ((nready=poll(fds,1,MCAST_THREAD_DELAY)) < 0 || pollerr(0))
    {
      if (errno == EINTR) continue;
      goto sckErr;
    }
    if (nready == 0) continue;
    if ((nbytes=recvfrom(mcastClnSck, (char *)mcstWorkArea, clnWorkAreaSize, 0, (struct sockaddr *)&sckadr, &len)) <= 0)
    {
      if (errno == EINTR) continue;
      goto sckErr;
    }
    if (nbytes == 1 && *mcstWorkArea == 0)
    { /* received stop character: if gClientIsClosing == TRUE then we leave ! */
      if (tineDebug) dbglog("RcvMcast: received stop signal");
      continue; 
    }
    doProcessData(0,mcstWorkArea,&sckadr,p,"MCAST");
  }
sckErr:
  if (!ServerExitCondition && !gClientIsClosing) rprtSckErr("mcstRecvThreadTask","recvfrom",mcastClnSck,-1,0,NULL);
  mcstRecvThreadActive = FALSE;
  if (tineDebug) dbglog("RcvMcast: stopping thread task");
  return 0;
}
THREAD_TYPE clThreadTask(THREAD_PARAM lpvThreadParam)
{
  ContractListStruct *cl = (ContractListStruct *)lpvThreadParam;
  SetThreadName("cntHdr");
  cl->returnCode = not_ready;
  cl->returnCode = doContract(cl,FALSE);
  ThreadExit(0);
  return 0;
}
void bkgThreadCleanup(void *ref)
{
  BkgThreadFcn btf = *(BkgThreadFcn *)ref;
  ExportListStruct *el = (ExportListStruct*)btf.elRef;
  if (el != NULL) el->hEqmBkg = (THRHANDLE)0;
  if (gSystemUseDataProtection && el != NULL) ReleaseSystemMutex(el->hDataProtectionMutex);
}
THREAD_TYPE bkgThreadTask(THREAD_PARAM lpvThreadParam)
{
  int dly;
  struct timeval tv1,tv2;
  int ready = FALSE;
  BkgThreadFcn btf;
  ExportListStruct *el;

  if (lpvThreadParam == NULL) return 0;
  SetThreadName("BkgTask");
  btf = *(BkgThreadFcn *)lpvThreadParam;
  if ((el=(ExportListStruct *)btf.elRef) == NULL) 
  { /* task not assigned to eqm ! */
    ready = TRUE;
  }
  while (!ServerExitCondition)
  {
    if (el->isterminating) break;
    if (!ready) ready = el->inidone;
    gettimeofday(&tv1,(struct timezone *)NULL);
    if (gInitTineServicesDone && ready && !el->isidle) 
    { /* do the registered background task */
      if (!gAllowBkgTaskReentrancy) while (el->bkgbusy) millisleep(gSystemTick);
      el->bkgbusy = TRUE;
      if (gSystemUseDataProtection && el != NULL) WaitForMutex(el->hDataProtectionMutex,-1);
      TSKCALL((&btf),tsk);
      if (gSystemUseDataProtection && el != NULL) ReleaseSystemMutex(el->hDataProtectionMutex);
      el->bkgbusy = FALSE;
      if (el->isterminating || ServerExitCondition) break;
    }
    gettimeofday(&tv2,(struct timezone *)NULL);
    dly = btf.rate - MSECS(tv2,tv1);
    if (dly < 0) dly = 0;
    if (dly > btf.rate) dly = btf.rate; /* clock set back ? */
    millisleep(dly);
  }
# if defined(WIN32)
  if (el->hEqmBkg != (THRHANDLE)0) CloseHandle(el->hEqmBkg);
# endif
  el->hEqmBkg = (THRHANDLE)0;
  feclogEx(TINE_LOGLEVEL_INFO, "FEC: background thread terminating (ServerExitCondition : %d)",ServerExitCondition);
  return 0;
}
int gClnThreadRunning = FALSE;
THREAD_TYPE clnThreadTask(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;
  SetThreadName("ClnCycle");
  feclog("client cycle thread starting");
  gClnThreadRunning = TRUE;
  while (!ServerExitCondition)
  {
    if (nConnectionTableEntries == 0) millisleep(gSystemTick);
    clientCycle();
  }
  feclog("client cycle thread terminating");
  return 0;
}
int gSrvThreadRunning = FALSE;
THREAD_TYPE srvThreadTask(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;
  SetThreadName("SrvCycle");
  feclog("server cycle thread starting");
  gSrvThreadRunning = TRUE;
  while (ServerExitCondition < 5)
  {
    if (gEqmListSize == 0) millisleep(gSystemTick);
    serverCycle();
  }
  gNeedToStartSrvCycleThread = TRUE; /* in case a restart is needed */
  feclogEx(TINE_LOGLEVEL_INFO, "FEC: server thread terminating");
  return 0;
}
THREAD_TYPE reinitThreadTask(THREAD_PARAM lpvThreadParam)
{ 
  lpvThreadParam = lpvThreadParam;
  _reinit(_gReinitIdleTime);  
  return 0;
}
#else /* !MULTITHREADED */
int hSystemClientMutex = 0;
int hSystemGlobalsUpdateMutex = 0;
int hSystemMCastUpdateMutex = 0;
int hSystemInitMutex = 0;
int hSystemServerMutex = 0;
int hSystemKernelMutex = 0;
int hCycleTriggerMutex = 0;
int hLinkCbMutex = 0;
int hLinkTblMutex = 0;
int hXferMutex = 0;
int hSndMutex = 0;
int hDpdLinkMutex = 0;
int hMcaLinkMutex = 0;
int hLwdLinkMutex = 0;
int hLinkDataMutex = 0;
int hLinkLstMutex = 0;
int hTcpBucketMutex = 0;
int hEnsMutex = 0;
int hTxMutex = 0;
int hAttrMutex = 0;
int hGeneralMutex = 0;
int hSrvTblMutex = 0;
int hStkAccessMutex = 0;
int hLinkQueueMutex = 0;
int hClientTableMutex = 0;
int hContractTableMutex = 0;
int hLogMutex = 0;
int hDbMutex = 0;
int hCsvMutex = 0;
int hMfMutex = 0;
int hHistoryTableMutex = 0;
int hHistoryCycleMutex = 0;
int hAlmTblMutex = 0;
int gSystemMutexSetInitialized = FALSE;
int InitSystemMutexSet(void)
{
  if (gSystemMutexSetInitialized) return 0;
  gStockResourceBusy = FALSE;
  hSystemClientMutex = 0;
  hSystemGlobalsUpdateMutex = 0;
  hSystemMCastUpdateMutex = 0;
  hSystemInitMutex = 0;
  hSystemServerMutex = 0;
  hSystemKernelMutex = 0;
  hCycleTriggerMutex = 0;
  hLinkCbMutex = 0;
  hLinkTblMutex = 0;
  hXferMutex = 0;
  hSndMutex = 0;
  hDpdLinkMutex = 0;
  hMcaLinkMutex = 0;
  hLwdLinkMutex = 0;
  hLinkDataMutex = 0;
  hLinkLstMutex = 0;
  hTcpBucketMutex = 0;
  hEnsMutex = 0;
  hTxMutex = 0;
  hAttrMutex = 0;
  hGeneralMutex = 0;
  hSrvTblMutex = 0;
  hStkAccessMutex = 0;
  hLinkQueueMutex = 0;
  hClientTableMutex = 0;
  hContractTableMutex = 0;
  hLogMutex = 0;
  hDbMutex = 0;
  hCsvMutex = 0;
  hMfMutex = 0;
  hHistoryTableMutex = 0;
  hHistoryCycleMutex = 0;
  hAlmTblMutex = 0;
  gSystemMutexSetInitialized = TRUE;
  return 0;
}
int CloseClientMutexSet(void)
{
  hSystemClientMutex = 0;
  hSystemGlobalsUpdateMutex = 0;
  hSystemMCastUpdateMutex = 0;
  return 0;
}
int CloseSystemMutexSet(void)
{
  if (!gSystemMutexSetInitialized) return not_initialized;
  _SystemReset(0);
  gSystemMutexSetInitialized = FALSE;
  return 0;
}
/* get around some recursive issues in single-threaded builds */
int hMutexSuspendedState = 0;
void YieldThread(void) { return; }
THRHANDLE GetThreadHandle()
{
  return 0;
}
#endif /* MULTITHREADED */
MsgType MsgTbl[MESSAGE_LOG_DEPTH];
int gMsgTblPtr = 0;
int msglog(int status,char *text,...)
{
  int lptr = (gMsgTblPtr+MESSAGE_LOG_DEPTH-1)%MESSAGE_LOG_DEPTH, mlen=128, dbg=tineDebug;
  char str[512], *c;
  static int initialized=FALSE;
  va_list args;
# if defined(UNIX) || defined(WIN32)
  va_start(args,text);
  vsnprintf(str,256,text,args); str[256] = 0;
# else
  if (strlen(text) > 128) return string_too_long;
  va_start(args,text);
  vsprintf(str,text,args);
# endif
  va_end(args);
  if (!initialized)
  {
    ClearMsgTable();
    initialized = TRUE;
  }
  if ((c=strstr(MsgTbl[lptr].msg," -#")) != NULL) mlen = (int)(c - MsgTbl[lptr].msg);
  if (mlen > 128) mlen = 128;
  if (!strncmp(MsgTbl[lptr].msg,str,mlen))
  { /* last message = this message */
    char add[32];
    int slen = (int)strlen(str), alen;
    if (c != NULL)
    {
      int cnt;
      c += 3;
      cnt = (int)strtol(c,&c,10);
      snprintf(add,32," -#%4d%s",++cnt,c);
    }
    else
    {
      char *tss = getDataTimeString(MsgTbl[lptr].ts,0);
      snprintf(add,32," -#   2 times since %.8s",&tss[9]);
    }
    add[31] = 0; alen = (int)strlen(add);
    if (slen > 127 - alen) slen = 127 - alen;
    strncat(&str[slen],add,alen);
    gMsgTblPtr = lptr; /* back up */
  }
  MsgTbl[gMsgTblPtr].ts = makeDataTimeStamp();
  MsgTbl[gMsgTblPtr].status = status;
  strncpy(MsgTbl[gMsgTblPtr].msg,str,128);
  gMsgTblPtr = (gMsgTblPtr+1) % MESSAGE_LOG_DEPTH; /* move along */
  if (_sendToDbg < 0) return 0; /* no console output */
  if (errorDebug && status != 0) dbg = 1;
  if (dbg || _sendToDbg) dbglog(str);
  return 0;
}
TINE_EXPORT int GetMsgTable(MsgType *msgtbl,int *nummsg)
{
  int i, j, cc = 0;
  
  if (nummsg == NULL || *nummsg <= 0) ccerr(argument_list_error);
  if (*nummsg > MESSAGE_LOG_DEPTH) *nummsg = MESSAGE_LOG_DEPTH;
  for (i=0; i<*nummsg; i++) 
  {
    j = (gMsgTblPtr - 1 - i + MESSAGE_LOG_DEPTH) % MESSAGE_LOG_DEPTH;
    if ((int)MsgTbl[j].ts == 0) break; /* table not yet filled */
    if (MsgTbl[j].ts == msgtbl[i].ts) break; /* no change */
    strncpy(msgtbl[i].msg,MsgTbl[j].msg,128);
    msgtbl[i].ts = MsgTbl[j].ts;
    msgtbl[i].status = MsgTbl[j].status;
  }
  *nummsg = i;
err:
  return cc;
}
TINE_EXPORT void ClearMsgTable(void)
{
	memset(MsgTbl, 0, sizeof(MsgTbl));
	gMsgTblPtr = 0;
}
TINE_EXPORT int GetMessageList(char *lstbuf,int bufsize)
{
  int i, k, n, nbytes=0;
  char msgstr[256], *stsstr;

  if (lstbuf == NULL || bufsize < 256) return argument_list_error;
  n = bufsize / 256;
  if (n < 1) return buffer_too_small;
  lstbuf[0] = 0;

  for (i=0; i<MESSAGE_LOG_DEPTH; i++) 
  {
    if (nbytes > bufsize - 256) break;
    k = (gMsgTblPtr - 1 - i + MESSAGE_LOG_DEPTH) % MESSAGE_LOG_DEPTH;
    if ((int)MsgTbl[k].ts == 0) break; /* table not yet filled */
    stsstr = MsgTbl[k].status == 0 ? "INFO" : "WARN";
    sprintf(msgstr,"%s [%s] %.128s\n",getDataTimeString(MsgTbl[k].ts,0),stsstr,MsgTbl[k].msg);
    strcat(lstbuf,msgstr); 
    nbytes = (int)strlen(lstbuf);
  }
  return 0;
}
TINE_EXPORT void dumpMsgTableToFile(void)
{
# define MSGFILENAME "messages.log"
  int i, k, cc=0;
  char fn[256],str[256];
  FILE *fp=NULL;
  if (!gIsRunningAsServer) return;
  if (WaitForSystemMutex(hLogMutex,-1) != 0) return;
  sprintf(fn,"%.128s%s",gFeclogPath,MSGFILENAME);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"System Messages: last %d message table entries (recent at top)\n",MESSAGE_LOG_DEPTH);
  for (i=0; i<MESSAGE_LOG_DEPTH; i++) 
  {
    k = (gMsgTblPtr - 1 - i + MESSAGE_LOG_DEPTH) % MESSAGE_LOG_DEPTH;
    if ((int)MsgTbl[k].ts == 0) break; /* table not yet filled */
    cc = MsgTbl[k].status;
    sprintf(str,"%s [%.32s] %.128s",
      getDataTimeString(MsgTbl[k].ts,0),(cc == 0 ? "normal" : cc2str(cc)),MsgTbl[k].msg);
    fprintf(fp,"%.192s\n",str);
  }
err:
  if (fp != NULL) fclose(fp);
  ReleaseSystemMutex(hLogMutex);
  return;
}
int gStockResourceBusy = FALSE;
/**
 * \internal used in checkTasklist() and only relevant in single-threaded builds
 */
void _SystemEngine(void)
{
  if (gEqmListSize && !gRunServerCycleInSeparateThread) 
  { /* check remote tasks  */
    serverCycle(); 
  }
  if (nConnectionTableEntries) 
  { /* Check our client list if any */
    clientCycle(); 
  }
}
TINE_EXPORT int ServerInitialized = FALSE;
TINE_EXPORT int GetServerInitialization(void)
{
  return ServerInitialized;
}
TINE_EXPORT int ServerWaiting = FALSE;
TINE_EXPORT int GetServerWaiting(void)
{
  return ServerWaiting;
}
TINE_EXPORT void SetServerWaiting(int value)
{
  ServerWaiting = value;
}
TINE_EXPORT int GetServerIdleState(char *eqm)
{
  ExportListStruct *el=getExportListItem(eqm);
  return el != NULL ? el->isidle : FALSE;
}
TINE_EXPORT void SetServerIdleState(char *eqm,int value)
{
  ExportListStruct *el=getExportListItem(eqm);
  if (el == NULL) return;
  el->isidle = value;
  gEqmInitializedIdleState = value;
  return;
}
#ifdef UNIX
int millisleep(int msec)
{
# ifdef USE_USLEEP
  return usleep(msec*1000);
# else
  struct timespec ts;
  if (msec <= 0) return 0;
  ts.tv_sec = msec/1000;
  ts.tv_nsec = (msec%1000) * 1000000;
  return nanosleep(&ts,NULL);
# endif
}
#endif
int isMutexBusy(MXHANDLE mx)
{
# ifdef MULTITHREADED
  if (WaitForSystemMutex(mx,gSystemTick) == 0)
  {
    ReleaseSystemMutex(mx);
    return FALSE;
  }
  return TRUE;
# else
  return mx;
# endif
}
int dumpSemaphores(void)
{
  ttyoutput("system kernel        : %s",isMutexBusy(hSystemKernelMutex) ? "BUSY" : "FREE");
  ttyoutput("recv globals         : %s",isMutexBusy(hSystemGlobalsUpdateMutex) ? "BUSY" : "FREE");
  ttyoutput("server init          : %s",isMutexBusy(hSystemInitMutex) ? "BUSY" : "FREE");
  ttyoutput("server cycle         : %s",isMutexBusy(hSystemServerMutex) ? "BUSY" : "FREE");
  ttyoutput("client table         : %s",isMutexBusy(hClientTableMutex) ? "BUSY" : "FREE");
  ttyoutput("contract table       : %s",isMutexBusy(hContractTableMutex) ? "BUSY" : "FREE");
  ttyoutput("addr table list      : %s",isMutexBusy(hSrvTblMutex) ? "BUSY" : "FREE");
  ttyoutput("link table           : %s",isMutexBusy(hLinkTblMutex) ? "BUSY" : "FREE");
  ttyoutput("stock property access: %s",isMutexBusy(hStkAccessMutex) ? "BUSY" : "FREE");
  ttyoutput("client transfer      : %s",isMutexBusy(hXferMutex) ? "BUSY" : "FREE");
  ttyoutput("send to peer         : %s",isMutexBusy(hSndMutex) ? "BUSY" : "FREE");
  ttyoutput("link table queue     : %s",isMutexBusy(hLinkQueueMutex) ? "BUSY" : "FREE");
  ttyoutput("link table list      : %s",isMutexBusy(hLinkLstMutex) ? "BUSY" : "FREE");
  ttyoutput("link dependency list : %s",isMutexBusy(hDpdLinkMutex) ? "BUSY" : "FREE");
  ttyoutput("mca link list        : %s",isMutexBusy(hMcaLinkMutex) ? "BUSY" : "FREE");
  ttyoutput("link watchdog list   : %s",isMutexBusy(hLwdLinkMutex) ? "BUSY" : "FREE");
  ttyoutput("link data protection : %s",isMutexBusy(hLinkDataMutex) ? "BUSY" : "FREE");
  ttyoutput("history table        : %s",isMutexBusy(hHistoryTableMutex) ? "BUSY" : "FREE");
  ttyoutput("history cycle        : %s",isMutexBusy(hHistoryCycleMutex) ? "BUSY" : "FREE");
  ttyoutput("alarm table          : %s",isMutexBusy(hAlmTblMutex) ? "BUSY" : "FREE");
  ttyoutput("client cycle         : %s",isMutexBusy(hSystemClientMutex) ? "BUSY" : "FREE");
  ttyoutput("recv mcast updates   : %s",isMutexBusy(hSystemMCastUpdateMutex) ? "BUSY" : "FREE");
  ttyoutput("link callback        : %s",isMutexBusy(hLinkCbMutex) ? "BUSY" : "FREE");
  ttyoutput("cycle trigger        : %s",isMutexBusy(hCycleTriggerMutex) ? "BUSY" : "FREE");
  ttyoutput("tcp bucket           : %s",isMutexBusy(hTcpBucketMutex) ? "BUSY" : "FREE");
  ttyoutput("ens access           : %s",isMutexBusy(hEnsMutex) ? "BUSY" : "FREE");
  ttyoutput("trivial sync link    : %s",isMutexBusy(hTxMutex) ? "BUSY" : "FREE");
  ttyoutput("attribute access     : %s",isMutexBusy(hAttrMutex) ? "BUSY" : "FREE");
  ttyoutput("system calloc        : %s",isMutexBusy(hGeneralMutex) ? "BUSY" : "FREE");
  ttyoutput("log file             : %s",isMutexBusy(hLogMutex) ? "BUSY" : "FREE");
  ttyoutput("db file              : %s",isMutexBusy(hDbMutex) ? "BUSY" : "FREE");
  ttyoutput("csv file             : %s",isMutexBusy(hCsvMutex) ? "BUSY" : "FREE");
  ttyoutput("manifest             : %s",isMutexBusy(hMfMutex) ? "BUSY" : "FREE");
  return 0;
}
extern int gFecHomeEstablished;
void fixFecRepository(void)
{
  char *ptr;
  int v;
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  if (gIsEquipmentNameServer) return;

  if (!gFecHomeEstablished)
  { /* only allow this if not set per API before hand */
    if ((ptr=getenv("FEC_HOME")) != NULL || (ptr=getenv("FECDB")) != NULL)
    { /* also allow legacy ENV setting */
      memset(fecDbPath,0,STD_PATH_LENGTH);
      strncpy(fecDbPath,ptr,STD_PATH_LENGTH);
      assertPathDelimiter(fecDbPath,STD_PATH_LENGTH);
    }
  }
  if (!gFeclogPathEstablished)
  { /* only allow this if not set per API before hand */
    memset(gFeclogPath,0,STD_PATH_LENGTH);
    if ((ptr=getenv("FEC_LOG")) != NULL || (ptr=getenv("FECLOG")) != NULL) 
    { /* also allow legacy ENV setting */
      strncpy(gFeclogPath,ptr,STD_PATH_LENGTH);
      assertPathDelimiter(gFeclogPath,STD_PATH_LENGTH);
    }
    else
    { /* then use the established FEC_HOME */
      strncpy(gFeclogPath,fecDbPath,STD_PATH_LENGTH);
    }
  }
  if ((ptr = getenv("FEC_LOGDEPTH")) != NULL ||
    (ptr = getenv("FECLOGDEPTH")) != NULL ||
    (ptr = getenv("LOGDEPTH")) != NULL)
  {
    FeclogDepth = atoi(ptr);
    if (FeclogDepth < 100) FeclogDepth = 100;
  }
  if ((ptr=getenv("FEC_POLLRATE")) != NULL || (ptr=getenv("FECPOLLRATE")) != NULL) 
  {
    SetMinimumAllowedPollingInterval(atoi(ptr));
  }
  if ((ptr=getenv("FEC_WORKAREASIZE")) != NULL || (ptr=getenv("FECWORKAREASIZE")) != NULL)
  {
    MaxSystemTransportSize = atoi(ptr);
    if (MaxSystemTransportSize < 4096) MaxSystemTransportSize = 4096;
  }
  if ((ptr=getenv("FEC_REMOTE_MANAGEMENT")) != NULL)
  {
    v = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
    SetAllowRemoteManagement(v);
  }
  if ((ptr=getenv("FEC_DIE_INUSE")) != NULL)
  {
    v = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
    SetDieOnAddressInUse(v);
  }
  if ((ptr=getenv("FEC_DIE_SOCKETERROR")) != NULL)
  {
    v = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
    SetDieOnSocketError(v);
  }
  if ((ptr=getenv("FEC_FILES_ALLOW_SCAN")) != NULL)
  {
    if (!stricmp(ptr,"TRUE")) v = LOG_SCAN_FEC;
    else if (!stricmp(ptr,"FULL")) v = LOG_SCAN_FULL;
    else v = LOG_SCAN_NONE;
    SetLogFileAllowScan(v);
  }  
  if ((ptr=getenv("FEC_LAZY_SCHEDULING")) != NULL)
  {
    gSchedulePropertyLazy = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
  }
  if ((ptr=getenv("FEC_MINIMUM_FREE_MAINBLOCKS")) != NULL)
  {
    UINT32 minbytes = (UINT32)strtol(ptr, NULL, 10);
    minbytes *= 1024;
    SetAvailableMemoryThreshold(minbytes);
  }
  if ((ptr=getenv("FEC_MINIMUM_FREE_DISKBLOCKS")) != NULL)
  { /* e.g. set FEC_MINIMUM_FREE_DISKBLOCKS=C:\|100000 */
    char *c, path[256];
    UINT32 minblocks;
    strncpy(path, ptr, 255); path[255] = 0;
    if ((c=strchr(path, '|')) != NULL) *c++ = 0;
    if (c != NULL)
    {
      minblocks = (UINT32)strtol(c, NULL, 10);
      SetFreeBlocksAlarmThreshold(path, minblocks);
    }
  }
}
TINE_EXPORT double getDataTimeStamp(void)
{
  return gDataTimeStamp;
}
TINE_EXPORT UINT32 setWorkAreaSize(UINT32 size)
{
  if (!ServerInitialized &&  /* server already initialized => too late!*/
       size > MIN_WORK_SIZE) /* small setting not allowed */
    MaxSystemTransportSize = (UINT32)size;
  feclog("work area set to %d bytes",size);
  if (ServerInitialized) feclogEx(TINE_LOGLEVEL_WARN, "work area size setting has no effect");
  gIsRunningAsServer = TRUE; /* only a server would be calling setWorkAreaSize() */
  return (UINT32)MaxSystemTransportSize; /* return the working variable */
}
TINE_EXPORT double makeDataTimeStamp(void) 
{ 
  struct timeval tv;
  double ts;
  gettimeofday(&tv,(struct timezone *)NULL);
  ts = (double)tv.tv_sec;
  ts += (double)(tv.tv_usec)/1000000.0;
  ts += gDataTimeStampOffset;
  return ts;
}
#ifndef SMALL_TINE_LIB
void cbClsLog(int id, int cc)
{
  UNUSED(id);
  switch (cc)
  { 
    case 0: /* entry accepted */
      break;
    default: /* something went wrong */
      feclogEx(TINE_LOGLEVEL_ERROR, "clslog missed entry: %s",cc2str(cc));
      break;
  }
  return;
}
TINE_EXPORT int clslog(char *context,char *tag,char *logger,int priority,int status,char *text,...)
{ /* send logging entry to central logger */
  static int clsUnknown = FALSE;
  int id, cc=0;
  double ts = makeDataTimeStamp();
  char str[512],*ctx=context,*usr=logger;
  CLOG clog;
  DTYPE din;
  va_list args;

# if defined(UNIX) || defined(WIN32)
  va_start(args,text);
  vsnprintf(str,256,text,args); str[256] = 0;
# else
  if (strlen(text) > 256) ccerr(string_too_long);
  va_start(args,text);
  vsprintf(str,text,args);
# endif
  va_end(args);

  feclog(str);

  if (clsUnknown) return address_unknown;

  if (ctx == NULL || strlen(ctx) == 0) ctx = gDeviceContext;
  if (usr == NULL || strlen(usr) == 0) usr = gUserName;
  if (ctx == NULL) ctx = "UNKNOWN";
  if (tag == NULL) tag = "UNKNOWN";
  memset(&clog, 0, sizeof(CLOG));
  strncpy(clog.context,ctx,32);
  strncpy(clog.logger,usr,32);
  strncpy(clog.tag,tag,32);
  strncpy(clog.text,str,256);
  clog.priority = priority;
  clog.status = status;
  clog.timeSent = ts;

  din.dArrayLength = 1;
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"CLOG");
  din.data.vptr = &clog;

  id = AttachLink("/SITE/CLOG/Logger","NewEntry",NULL,&din,CA_WRITE|CA_RETRY,500,cbClsLog,CM_SINGLE);
  if (id < 0) ccerr(-id);
err:
  if (cc)
  {
    dbglog("clslog : %s",cc2str(cc));
    clsUnknown = TRUE;
  }
  return cc;
}
#endif /* !SMALL_TINE_LIB */
#if defined(VXWORKS) || defined(NIOS)
TINE_EXPORT int nofeclog = TRUE;
#else
TINE_EXPORT int nofeclog = FALSE;
#endif
TINE_EXPORT int gErrorLogLevel = TINE_LOGLEVEL_WARN;
TINE_EXPORT int putCommandsInFeclog = INCLUDE_IN_LOGFILE;
TINE_EXPORT void SetPutCommandsInFeclog(int value)
{
  putCommandsInFeclog = value;
}
TINE_EXPORT int GetPutCommandsInFeclog(void)
{
  return putCommandsInFeclog ? TRUE : FALSE;
}
TINE_EXPORT int FeclogDepth = FECLOG_DEPTH;
TINE_EXPORT void SetFeclogDepth(int value)
{
  if (value < 100) value = 100;
  FeclogDepth = value;
  feclog("fec log depth set to %d",value);
}
TINE_EXPORT int GetFeclogDepth(void)
{
  return FeclogDepth;
}
TINE_EXPORT int useGlobalSynchronization = GLOBAL_SYNCHRONIZATION;
TINE_EXPORT void SetUseGlobalSynchronization(int value)
{
  if (ServerInitialized)
  {
    feclog("attempt to set global synchronization to %s after server initialization !",value ? "TRUE" : "FALSE");
    return;
  }
  useGlobalSynchronization = value;
  feclog("set global synchronization to %s via API",value ? "TRUE" : "FALSE");
}
TINE_EXPORT int GetUseGlobalSynchronization(void)
{
  return useGlobalSynchronization;
}
TINE_EXPORT int openFecLog(void)
{
# if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
  char fn[256];
  int tries, fd = -1;

  if (!makeDirectoryPath(gFeclogPath,FALSE))
  {
    strncpy(gFeclogPath,fecDbPath,STD_PATH_LENGTH);
  }
  sprintf(fn,"%.128s%s",gFeclogPath,"fec.log");
  for (tries=0; tries < 3; tries++) 
  {
    if ((fd=open(fn,O_RDWR|O_APPEND)) != -1) break;
    microsleep(200);
  }
  return fd;
# else
  return -1;
# endif
}
int feclogWriteFile(char *log_text, char *filenameNoExtension)
{
  static int cleared = FALSE;
  int tries,len,cc=0;
  static SINT32 flen = -1;
  FILE *fp = NULL;
  char fn[256],fb[256];
# ifdef DOS
  static char *fmod = "a+t";
# else
  static char *fmod = "a+";
# endif
# if defined(VXWORKS)
  fmod = fmod;
# endif
  if (!gIsRunningAsServer) return 0;
  fn[0] = 0;
  if (!cleared)
  {
    cleared = TRUE;
    fixFecRepository();
    InitSystemMutexSet();
  }
  if (!gSystemMutexSetInitialized) return not_initialized; /* not applicable : resources have been freed ? */
  if (WaitForSystemMutex(hLogMutex,-1) != 0) return semaphore_error;

  if (log_text[strlen(log_text)-1] != '\n') strcat(log_text,"\n");
# ifndef DOS
  if (vFeclogBuffer != NULL) ccerr(vfeclog(log_text));
# endif

# ifndef FS_RDONLY
  sprintf(fn,"%.128s%s.%s",gFeclogPath,filenameNoExtension,"log");
  for (tries=0; tries < 3; tries++) 
  {
    if ((fp=fopen(fn,fmod)) != NULL) break;
    if (strncmp(gFeclogPath,fecDbPath,STD_PATH_LENGTH) && !makeDirectoryPath(gFeclogPath,FALSE)) 
    { /* an exotic LOG file directory is not yet there ? */
      strncpy(gFeclogPath,fecDbPath,STD_PATH_LENGTH);
      sprintf(fn,"%.128s%s.%s",gFeclogPath,filenameNoExtension,"log");
    }
    microsleep(200);
  }
  if (fp == NULL) ccerr(file_error);
  fwrite(log_text,len=(int)strlen(log_text),1,fp);
  if (flen == -1) flen = ftell(fp);
  if ((flen += len) < FeclogDepth*2*80) ccerr(0);     /* finished */
  fclose(fp); fp = NULL; flen = 0;
  sprintf(fb,"%.128s%s.%s",gFeclogPath,filenameNoExtension,"bak");
  remove(fb);
  rename(fn,fb);
# else 
  fb[0] = 0; len = 0; flen = flen; tries = tries; /* avoid warnings in this case */
  ccerr(not_implemented); /* no way to write a log file */
# endif
err:
  hasLogFileAlarm = FALSE;
  if (fp) fclose(fp);
  if (cc == file_error) 
  { 
    setFileErrorAlarm(fn); 
    hasLogFileAlarm = TRUE; 
  }
  ReleaseSystemMutex(hLogMutex);
  return cc;
}
TINE_EXPORT int feclogEx(int logLevel, char *text,...)
{
  int cc=0;
  double ts = makeDataTimeStamp();
  char str[512];
  va_list args;
  /* precondition */
  if (!gIsRunningAsServer) return 0;
  if( logLevel<0 || logLevel>=GetLogLevelListSize()) ccerr(invalid_parameter);
  /* prepare log text */
  sprintf(str,"%s[%.16s][%s] ",getDataTimeString(ts,0),gFecName,logLevel2str(logLevel));
# if defined(UNIX) || defined(WIN32)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) 
  {
    if (strchr(text,'%') == NULL)
    { /* then truncate */
      strncpy(&str[strlen(str)],text,256); str[256] = 0;
    }
    ccerr(string_too_long);
  }
  else
  {
    va_start(args,text);
    vsprintf(&str[strlen(str)],text,args);
    va_end(args);
  }
# endif
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  /* write file(s) */
  cc = feclogWriteFile(str, "fec");  
  if (logLevel>=gErrorLogLevel && !nofeclog)
  {
    cc = feclogWriteFile( str, "error");
  }  
err:
  if (cc) sprintf(str,"cannot log entry : %s",cc2str(cc));
  if (_sendToDbg < 0 || gSufferInSilence) return cc; /* signal to not dump to the console */
  if (tineDebug || _sendToDbg) dbglog(str);
  return cc;
}
int hasLogFileAlarm = FALSE;
TINE_EXPORT int feclog(char *text,...)
{
  char str[512];
  va_list args;
  /* preconditions */
  if (text == NULL) return argument_list_error;
  if (!gIsRunningAsServer) return 0;
  /* prepare log text */
# if defined(UNIX) || defined(WIN32)
  va_start(args,text);
  vsnprintf(str,256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) 
  {
    if (strchr(text,'%') == NULL)
    { /* then truncate */
      strncpy(str,text,256); str[256] = 0;
    }
    sprintf(str, "cannot log entry : %s", cc2str(string_too_long));
    if (tineDebug) dbglog(str);
    return string_too_long;
  }
  else
  {
    va_start(args,text);
    vsprintf(str,text,args);
    va_end(args);
  }
# endif  
  /* continue preparing log text and write to file */
  return feclogEx(TINE_LOGLEVEL_INFO, str);
}
#if !defined(VXWORKS) && !defined(DOS) && !defined(FS_RDONLY)
TINE_EXPORT int srvlog(char *filename,char *tag,char *text,...)
{
  static int cleared = FALSE;
  int len,cc=0;
  double ts = makeDataTimeStamp();
  char *rn, str[800] = {0}, fn[256] = {0}, fb[256] = {0};
  FILE *fp = NULL;
  va_list args;
  if (filename == NULL || text == NULL) return argument_list_error;
  if (tag == NULL) tag = "Logger";
  if (!gIsRunningAsServer) return 0;
  if (!cleared)
  {
    cleared = TRUE;
    fixFecRepository();
    InitSystemMutexSet();
  }
  fn[0] = 0;
  if (WaitForSystemMutex(hLogMutex,-1) != 0) return semaphore_error;
  sprintf(str,"%.64s[%.32s] ",getDataTimeString(ts,0),tag);
# if defined(UNIX) || defined(WIN32)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],768,text,args); str[768] = 0;
  va_end(args);
# else
  ccerr(not_implemented);
# endif
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  if ((rn=strrchr(filename,FS_DELIMITER)) != NULL)
  { /* absolute path/file name given */
    strncpy(fn,filename,96);
    if (strchr(rn,'.') == NULL) strcat(fn,".log");
  }
  else
  { /* just file name: use the canonical location */
    sprintf(fn,"%.128s%.96s",gFeclogPath,filename);
    if (strchr(fn,'.') == NULL) strcat(fn,".log");
  }
  if ((fp=fopen(fn,"a+")) == NULL) ccerr(file_error);
  fwrite(str,len=(int)strlen(str),1,fp);
  if (ftell(fp) + len < FeclogDepth*2*80) ccerr(0); /* finished */
  fclose(fp); fp = NULL;
  sprintf(fb,"%.132s.bak",fn);
  remove(fb);
  rename(fn,fb);
err:
  hasLogFileAlarm = FALSE;
  if (fp) fclose(fp);
  if (cc) sprintf(str,"cannot log entry : %s",cc2str(cc));
  if (cc == file_error) 
  { 
    setFileErrorAlarm(fn); 
    hasLogFileAlarm = TRUE; 
  }
  ReleaseSystemMutex(hLogMutex);
  if (tineDebug) dbglog(str);
  return cc;
}
#endif
TINE_EXPORT char *GetDataTimeString(double ts, int useLongStringFormat)
{
  static char tsstr[64]; /* static: call returns pointer to tsstr as a curtesy ! */
  return GetDataTimeStringEx(ts, useLongStringFormat, tsstr);
}
TINE_EXPORT char *GetDataTimeStringEx(double ts,int useLongStringFormat,char *tsstr)
{
  time_t t = (time_t)ts;
  UINT32 ms = (UINT32)((ts-(double)t) * 1000.0);
  int isdst = findDaylight(t);
  struct tm *tms;

  if (tsstr == NULL) return GetDataTimeString(ts, useLongStringFormat);
  t += (isdst+1) * STD_TIME_UTC_OFFSET;

  if (useLongStringFormat)
  {
    strncpy(tsstr,ctime(&t),24);
    strncpy(&tsstr[28],&tsstr[20],4); tsstr[32] = 0;
#   ifdef OS_16BIT
    sprintf(&tsstr[19],".%03ld",ms);
#   else
    sprintf(&tsstr[19],".%03d",ms);
#   endif
    sprintf(&tsstr[23]," %s",ltzname[isdst]);
    tsstr[27] = ' ';
  }
  else
  {
    if ((tms=localtime(&t)) != NULL)
    {
      sprintf(tsstr,"%02d.%02d.%02d %02d:%02d:%02d.%03d %s",
        tms->tm_mday,tms->tm_mon+1,tms->tm_year%100,
        tms->tm_hour,tms->tm_min,tms->tm_sec,
        (int)ms,ltzname[isdst]);
    }
    else
    {
      strcpy(tsstr,"invalid time");
    }
  }
  return tsstr;
}
TINE_EXPORT struct timeval *GetDataTimeStampAsTimeval(void)
{
  static struct timeval tv;
  tv.tv_sec = UTCtimestamp;
  tv.tv_usec = (UTCtimestampMSEC) * 1000;
  return &tv;
}
TINE_EXPORT double GetTimeStampFromString(char *timestr)
{
  time_t t=time(NULL);
  int i, yr, len, ndigits=0, ndecimals=0, isnumber=TRUE;
  double ms=0.0;
  char ts[256];
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16],msec[8];
  struct tm tmts, *tnow;

  strncpy(ts,timestr,256); ts[255] = 0;
  strtrm(ts);
  hr[0] = 0;
  tnow = localtime(&t);
  if ((len=(int)strlen(ts)) < 5) isnumber = FALSE;
  if (len == 0) return 0;
  strupr(ts);
  if (!strcmp(ts,"NOW")) return makeDataTimeStamp();
  for (i=0; i<len && isnumber; i++)
  {
    if (isdigit(ts[i])) ndigits++;
    if (ts[i] == '.') 
    {
      if (i < 6 || ndecimals++ > 0) isnumber = FALSE;
    }
  }
  if (len != (ndigits + ndecimals)) isnumber = FALSE;
  if (isnumber) return strtod(ts,NULL);
  if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year,"_")) != NULL || (c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0; msec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
          if ((c=strstr(sec,".")) != NULL)
          {
            *c++ = 0;
            strncpy(msec,c,8);
          }
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
      ms = atoi(msec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    yr = atoi(year);
    if (yr < 70) yr += 100;
    if (yr > 1900) yr -= 1900;
    tmts.tm_year = yr;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return (double)t + ms/1000.0;
}
TINE_EXPORT double GetDataTimeStamp(void)
{
  return gDataTimeStamp;
}
int gFecHomeEstablished = FALSE;
#ifndef SMALL_TINE_LIB
TINE_EXPORT char *GetFecLogPath(void)
{
  return gFeclogPath;
}
TINE_EXPORT char *GetFecHome(void)
{
  return fecDbPath;
}
TINE_EXPORT int SetFecHome(char *fecHomePath)
{
 if (gInitTineServicesDone) return already_assigned;
  if (gIsEquipmentNameServer) return already_assigned;
  if (gIsGroupNameServer) return already_assigned;
  memset(fecDbPath,0,STD_PATH_LENGTH);
  strncpy(fecDbPath,fecHomePath,STD_PATH_LENGTH);
  assertPathDelimiter(fecDbPath,STD_PATH_LENGTH);
  gFecHomeEstablished = TRUE;
  feclog("set FEC_HOME to %s",fecDbPath);
  return 0;
}
TINE_EXPORT char *GetSystemUserName(void)
{
  return gUserName;
}
TINE_EXPORT int GetSystemLogging(void)
{
  return nofeclog ? FALSE : TRUE;
}
#endif /* !SMALL_TINE_LIB */
TINE_EXPORT double MakeDataTimeStamp(void)
{
  return makeDataTimeStamp();
}
TINE_EXPORT double PutDataTimeStamp(double toffset, time_t tsec, int tmsec)
{
  double ts;
  ts = (double)tsec;
  ts += (double)tmsec/1000.0;
  ts += toffset;
  return ts;
}
TINE_EXPORT double PutDataTimeStampU(double toffset, time_t tsec, int tusec)
{
  double ts;
  ts = (double)tsec;
  ts += (double)tusec/1000000.0;
  ts += toffset;
  return ts;
}
TINE_EXPORT int SetBurstLimit(int npackets)
{
  if (npackets < 1) npackets = 1;
  gBurstLimit = npackets;
  feclog("burst limit set to %d packets",npackets);
  return gBurstLimit;
}
TINE_EXPORT int GetBurstLimit(void)
{
  return gBurstLimit;
}
TINE_EXPORT int SetCycleDelay(int msecs)
{
  if (msecs < 0) msecs = 0;
  gCycleDelay = msecs;
  return gCycleDelay;
}
TINE_EXPORT int GetCycleDelay(void)
{
  return gCycleDelay;
}
TINE_EXPORT int SetCycleMicroDelay(int usecs)
{
  if (usecs > 0)
  {
    gCycleMicroDelay = usecs;
  }
  else
  { /* reset (turn it off) */
    gCycleMicroDelay = 0;
  }
  return gCycleMicroDelay;
}
TINE_EXPORT int GetCycleMicroDelay(void)
{
  return gCycleMicroDelay;
}
TINE_EXPORT void SetDataTimeStamp(double ts)
{
  gServerDataTimeStamp = ts; /* this should be thread-safe */
  gDataTimeStamp = ts;       /* this isn't */
}
TINE_EXPORT void SetSystemLogging(int value)
{
  nofeclog = value ? FALSE : TRUE ;
}
TINE_EXPORT UINT32 SetWorkAreaSize(UINT32 size) { return setWorkAreaSize(size); }
TINE_EXPORT UINT32 GetWorkAreaSize(void) { return MaxSystemTransportSize; }
#ifdef WIN32
TINE_EXPORT void setDataTimeStamp(double ts) { SetDataTimeStamp(ts); }
#endif
TINE_EXPORT int _SystemCycle(int chkcmd)
{
  int nullpass = TRUE;
  ProbeTimeElapsed("SystemCycle entered");
# ifdef MULTITHREADED
  if (!isAllowedSystemCycle()) { millisleep(gSystemTick); return 0; }
  if (!cycle_priority_fixed) fixSystemCyclePriority();
# endif
# ifdef UNIX
  foregroundTTY = chkcmd ? IsInForeground() : FALSE;
# endif
# ifdef MULTITHREADED
  if (gRunServerCycleInSeparateThread)
  {
    if (gNeedToStartSrvCycleThread) srvCycleCreateThread(0);
    if (gIsRunningAsServer && !gInitTineServicesDone) millisleep(1);
  }
  else
# endif
  {
    if ((gEqmListSize > 0) || FecNameRegistered || gDelayInitTineServices) 
    { /* only run thru server cycle if it makes sense */
      serverCycle();  
      nullpass = FALSE;
    }
  }
  checkTasklist(); /* run through the task list */
  if (nConnectionTableEntries || glbClnSck) 
  { /* only run thru client cycle if it makes sense */
    clientCycle(); 
    nullpass = FALSE;
  }
# ifdef RDATE
  if (SychronizeFecClock) FecSynchronize(); /* Synchronize FEC */
# endif
# ifdef DOS
  if (watchdogBaseAddress) watchDog(); /* hardware watchdog */
# endif
# if defined(DOS) || defined(WIN32) || defined(NIOS)
  if (chkcmd) GetCommand(); /* simple command line interpreter */
# endif
# ifndef SMALL_TINE_LIB
  if (_gNeedWarmRestart) _SystemReInit();
# endif
  if (nullpass) millisleep(gSystemTick);
  ProbeTimeElapsed("SystemCycle returning");
  if (!gIsRunningAsServer && ttimeprobe > 0) ttimeprobe--;
  return (ReturnToServerCycle | ReturnToClientCycle); /* need to return */
}
TINE_EXPORT void _SystemDelay(int msec)
{
  int forever = msec < 0 ? 1 : 0;
# ifndef MULTITHREADED
  struct timeval tv0,tv;
  gettimeofday(&tv0,(struct timezone *)NULL);
  tv = tv0;
  while (forever || (MSECS(tv,tv0) < msec))
  {
    SystemCycle(TRUE);
    gettimeofday(&tv,(struct timezone *)NULL);
  }
# else
  if (forever) msec = 1000;
  do
  {
    millisleep(msec);
  } while (forever);
# endif
}
static int gEqmSyncLockCount = 0;
TINE_EXPORT int LockEquipmentModules(void)
{
  if (gEqmSyncLockCount) return already_assigned;
  if (WaitForSystemMutex(hSystemServerMutex,-1) != 0) return semaphore_busy;
  gEqmSyncLockCount++;
  return 0;
}
TINE_EXPORT int UnlockEquipmentModules(void)
{
  if (gEqmSyncLockCount == 0) return 0;
  if (ReleaseSystemMutex(hSystemServerMutex) != 0) return semaphore_error;
  gEqmSyncLockCount = 0;
  return 0;
}
TINE_EXPORT void _SystemStop(int exitOnFree)
{
# ifdef MULTITHREADED
  if (gIsRunningAsServer)
  { /* if we're not a server then don't worry about last minute initialization */
    int i;
    for (i=0; i<1000 && !gLastMinuteInitializationDone; i++)
    { /* called while we're sill initializing !? */
      millisleep(gSystemTick);
      YieldThread();
    }
    if (!gLastMinuteInitializationDone)
    { /* did some crazy person call SystemStop inside a eqm initialization routine ? */
      feclogEx(TINE_LOGLEVEL_FATAL, "Force TINE Server stop due to initialization/termination deadlock.");
      exit(0);
    }
  }
# else
  ReleaseSystemMutex(hSystemServerMutex);
# endif
  feclogEx(TINE_LOGLEVEL_INFO, "Stopping TINE process ...");
  ServerExitCondition = 4;
  _SystemFreeResources(exitOnFree);
}
SYSTSKP _preSysIniFcn = NULL;
SYSTSKP _postSysIniFcn = NULL;
SYSTSKP _postSysCycFcn = NULL;
TINE_EXPORT void SetPreSystemInitFunction(SYSTSKP fcn)
{
  _preSysIniFcn = fcn;
}
TINE_EXPORT void SetPostSystemInitFunction(SYSTSKP fcn)
{
  _postSysIniFcn = fcn;
}
TINE_EXPORT void SetSystemCleanupFunction(SYSTSKP fcn)
{
  _postSysCycFcn = fcn;
}
TINE_EXPORT int _SystemInit(int chkcmd)
{
  int cc;
  char *ptr = getenv("FEC_LOG_COMMANDS");
  gIsRunningAsServer = TRUE;
  ServerExitCondition = 0;
  if (ServerInitialized) 
  { /* already initialized: just return */
    feclogEx(TINE_LOGLEVEL_INFO, "SystemInit: redundant call to SystemInit() ignored");
    return 0; 
  }
  gServerCycleState = CYCLE_STATE_INIT;
  if (ptr != NULL) 
  {
    if (stricmp(ptr,"FALSE") == 0 || stricmp(ptr,"NO") == 0) putCommandsInFeclog = FALSE;
  }
  if (_preSysIniFcn) _preSysIniFcn();
# ifdef UNIX
  foregroundTTY = chkcmd;
# else
  chkcmd = chkcmd;
# endif
  InitSystemMutexSet();
  if (WaitForSystemMutex(hSystemServerMutex,-1) != 0) 
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "SystemInit: cannot acquire SystemServerMutex => cannot initialize");
    return semaphore_error;
  }
  fixLocalTimeSettings();
  fixFecRepository();
  getAdmins();
# ifndef DOS
  if (nofeclog && FeclogDepth > 0)
  {
    vFeclogBuffer = (char *)SystemCalloc(FeclogDepth,80);
  }
# else
  signal(SIGFPE,SIG_IGN);            /* catch Floating domain errors */
  harderr(NetHandler);               /* start error handler          */
  whoami(gUserName);                 /* find out who we are */
  GetWatchDogBaseAddress();          /* check environment for watchdog */
# endif
# if defined(RDATE)
  TimeServerHostPtr = getTimeServerHost();
  rdate(TimeServerHostPtr);          /* synchronize */
# elif !defined(VXWORKS)
  tzset();
# endif
  /* initialize memory and network resources */
  if ((cc=initTineResources()) != 0) ccerr(cc);
  getAlmSettings();
# ifdef MULTITHREADED
  gNeedToStartSrvCycleThread = gRunServerCycleInSeparateThread;
# endif
  ServerInitialized = TRUE;
err:
  ReleaseSystemMutex(hSystemServerMutex);
  if (_postSysIniFcn) _postSysIniFcn();
  feclog("SystemInit: %.32s",cc2str(cc));
  return cc;
}
#ifndef SMALL_TINE_LIB
void _reinit(int idleTime)
{
  if (idleTime > 0) 
  {
    feclog("pausing re-initialization for %d seconds",idleTime);
    millisleep(idleTime*1000);
  }
  feclog("re-initializing FEC");
  ServerExitCondition = 0;
  if (_postSysIniFcn) _postSysIniFcn();
  SetServerCycleIdle(FALSE);
  gLastMinuteInitializationDone = FALSE;
}
TINE_EXPORT void SystemReInit(int idleTime)
{
  _gReinitIdleTime = idleTime;
  _gNeedWarmRestart = TRUE;
}
void _SystemReInit(void)
{
# ifndef DOS
  static int reinitCountdown = 5;
  if (reinitCountdown-- > 0) return;
  _gNeedWarmRestart = FALSE;
  FlushContractTable();
  SetServerCycleIdle(TRUE);
  FreeEQPModules();
  freeUserCommands();
  ServerExitCondition = 0;
  EquipmentModulesInitialized = FALSE;
  if (_postSysCycFcn) _postSysCycFcn();
  reinitCountdown = 5;
# ifdef MULTITHREADED
  gNeedToStartSrvCycleThread = TRUE;
  if (_gReinitIdleTime > 0) 
  {
    reinitCreateThread();
    return;
  }
# endif
  _reinit(_gReinitIdleTime);
# else
  printf("no warm restart possible in DOS\n");
# endif
}
#endif
TINE_EXPORT int _SystemReset(int level)
{
  ExportListStruct *el;
  int i;
  if (gIsRunningAsServer) feclogEx(TINE_LOGLEVEL_INFO, "SystemReset called (level %d)",level);
  ReturnToClientCycle = 0;          /* no immediate need */
  ReturnToServerCycle = 0;
  if (conTbl != NULL)
  {
    CloseClientLinks(); /* close all client links */
    for (i = 1; i < gConTblCapacity; i++)
    { /* free connection table memory (post ENS entry) */
      freeConTblEntryMemory(i);
      SystemFree(conTbl[i]);
    }
  }
  nConnectionTableEntries = (NameServerLoaded && !ServerExitCondition) ? 1 : 0; /* reset number of connection table entries */
  /* clear all semaphores */
  if (WaitForSystemMutex(hSystemClientMutex,gSystemTick) == 0) ReleaseSystemMutex(hSystemClientMutex); 
  if (WaitForSystemMutex(hSystemGlobalsUpdateMutex,gSystemTick) == 0) ReleaseSystemMutex(hSystemGlobalsUpdateMutex);
  if (WaitForSystemMutex(hLinkCbMutex,gSystemTick) == 0) ReleaseSystemMutex(hLinkCbMutex); 
  if (WaitForSystemMutex(hSystemMCastUpdateMutex,gSystemTick) == 0) ReleaseSystemMutex(hSystemMCastUpdateMutex);
  if (WaitForSystemMutex(hSystemInitMutex,gSystemTick) == 0) ReleaseSystemMutex(hSystemInitMutex); 
  if (WaitForSystemMutex(hSystemServerMutex,gSystemTick) == 0) ReleaseSystemMutex(hSystemServerMutex); 
  if (WaitForSystemMutex(hSystemKernelMutex,gSystemTick) == 0) ReleaseSystemMutex(hSystemKernelMutex); 
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) == 0) ReleaseSystemMutex(hLinkTblMutex);
  if (WaitForSystemMutex(hXferMutex,gSystemTick) == 0) ReleaseSystemMutex(hXferMutex);
  if (WaitForSystemMutex(hSndMutex,gSystemTick) == 0) ReleaseSystemMutex(hSndMutex);
  if (WaitForSystemMutex(hDpdLinkMutex,gSystemTick) == 0) ReleaseSystemMutex(hDpdLinkMutex);
  if (WaitForSystemMutex(hMcaLinkMutex,gSystemTick) == 0) ReleaseSystemMutex(hMcaLinkMutex);
  if (WaitForSystemMutex(hLwdLinkMutex,gSystemTick) == 0) ReleaseSystemMutex(hLwdLinkMutex); 
  if (WaitForSystemMutex(hLinkDataMutex,gSystemTick) == 0) ReleaseSystemMutex(hLinkDataMutex); 
  if (WaitForSystemMutex(hLinkLstMutex,gSystemTick) == 0) ReleaseSystemMutex(hLinkLstMutex); 
  if (WaitForSystemMutex(hTcpBucketMutex,gSystemTick) == 0) ReleaseSystemMutex(hTcpBucketMutex); 
  if (WaitForSystemMutex(hEnsMutex,gSystemTick) == 0) ReleaseSystemMutex(hEnsMutex); 
  if (WaitForSystemMutex(hTxMutex,gSystemTick) == 0) ReleaseSystemMutex(hTxMutex); 
  if (WaitForSystemMutex(hAttrMutex,gSystemTick) == 0) ReleaseSystemMutex(hAttrMutex); 
  if (WaitForSystemMutex(hGeneralMutex,gSystemTick) == 0) ReleaseSystemMutex(hGeneralMutex); 
  if (WaitForSystemMutex(hSrvTblMutex,gSystemTick) == 0) ReleaseSystemMutex(hSrvTblMutex); 
  if (WaitForSystemMutex(hStkAccessMutex,gSystemTick) == 0) ReleaseSystemMutex(hStkAccessMutex); 
  if (WaitForSystemMutex(hLinkQueueMutex,gSystemTick) == 0) ReleaseSystemMutex(hLinkQueueMutex); 
  if (WaitForSystemMutex(hClientTableMutex,gSystemTick) == 0) ReleaseSystemMutex(hClientTableMutex); 
  if (WaitForSystemMutex(hContractTableMutex,gSystemTick) == 0) ReleaseSystemMutex(hContractTableMutex); 
  if (WaitForSystemMutex(hLogMutex,gSystemTick) == 0) ReleaseSystemMutex(hLogMutex); 
  if (WaitForSystemMutex(hCsvMutex,gSystemTick) == 0) ReleaseSystemMutex(hCsvMutex); 
  if (WaitForSystemMutex(hMfMutex,gSystemTick) == 0) ReleaseSystemMutex(hMfMutex); 
  if (WaitForSystemMutex(hHistoryTableMutex,gSystemTick) == 0) ReleaseSystemMutex(hHistoryTableMutex);
  if (WaitForSystemMutex(hHistoryCycleMutex,gSystemTick) == 0) ReleaseSystemMutex(hHistoryCycleMutex); 
  if (WaitForSystemMutex(hAlmTblMutex,gSystemTick) == 0) ReleaseSystemMutex(hAlmTblMutex); 
# ifdef MULTITHREADED
  resetSystemCycleThread();
  for (el=ExportList; el != NULL; el=el->next) 
  {
    if (el->hEqmBusyMutex != NULL) ReleaseSystemMutex(el->hEqmBusyMutex);
    if (el->hDataProtectionMutex != NULL) 
    {
      if (WaitForSystemMutex(el->hDataProtectionMutex,gSystemTick) == 0) 
        ReleaseSystemMutex(el->hDataProtectionMutex);
    }
  }
# endif
  isNameServerRequest = FALSE;
  if (level > 0)
  {
    for (el=ExportList; el != NULL; el=el->next) 
    {
      el->inidone = FALSE;
      el->hstdone = FALSE;
    }
    gLastMinuteInitializationDone = FALSE;
  }
  gSystemStartupTime = time(NULL);
  return 0;
}
TINE_EXPORT void SetSystemReinitIdleTime(int idleTimeInSeconds)
{
  if (idleTimeInSeconds < 0) idleTimeInSeconds = 0;
  _gReinitIdleTime = idleTimeInSeconds;
}
TINE_EXPORT BYTE *SystemVersion(void)
{
  static BYTE ver[4];
  ver[0] = MAJOR_VERSION; ver[1] = MINOR_VERSION;
  ver[2] = VERSION_REVISION/256; ver[3] = VERSION_REVISION%256;
  return ver;
}
int gSchedulePropertyLazy = FALSE;
TINE_EXPORT void SetSystemSchedulePropertyLazy(int value)
{
  gSchedulePropertyLazy = value;
}
TINE_EXPORT int GetSystemSchedulePropertyLazy(void)
{
  return gSchedulePropertyLazy;
}
TINE_EXPORT int _SystemSchedulePropertyEx(char *eqm, char *prp,int scope)
{
  int i,cc=0,dn=-1;
  int setCompletion=FALSE,needsScheduler=FALSE,needsHistoryCycle=FALSE,needsAlarmCycle=FALSE,hasSeparateThread=FALSE;
  ContractListStruct *cl; /* improve readability by assignment */
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl;
  char *c,*p=NULL,*d=NULL,prpstr[4096];
  AWSLstEntry *aw;

  if (eqm == NULL || prp == NULL) return argument_list_error;
  if (el == NULL) return non_existent_elem;
  if (ServerWaiting) return not_initialized;
  if (!el->inidone) return not_initialized;
  if (el->isidle) return server_idle;
  if (scope & CA_SENDCODE)
  { /* property is scheduled to send the given return code to the caller */
    cc = scope & 0x0fff;
    setCompletion = TRUE;
    scope = CA_NETWORK;
  }
  strncpy(prpstr,prp,4096);
  for (p=strtok(prpstr,",|"); p != NULL; p=strtok(NULL,",|"))
  {
    if ((d=strchr(p,'<')) != NULL) 
    { /* device name also specified! */
      *d++ = 0;
      if ((c=strchr(d,'>')) != NULL) *c = 0;
      if (strchr(d,'*') != NULL) d = NULL; /* same as any device */
      dn = GetDeviceNumberEx(eqm,d,p);
    }
    if ((pl=GetPropertyListStruct(eqm,p,NULL)) == NULL)
    {
      if (tineDebug > 1 || errorDebug) dbglog("scheduled property %.64s not found !",p);
      /* continue; */
    }
    else
    {
      if (pl->runInSeparateThread) hasSeparateThread = TRUE;
    }
    if (scope & CA_NETWORK)
    { /* all listening clients in the connection table get scheduled */
      for (i=0; i<ncontracts; i++)
      {
        if ((cl=ContractList[i])->expired) continue;
        if (strncmp(cl->contract.EqmName,eqm,EQM_NAME_SIZE)) continue;
        if (cl->contract.EqmAccess & (CA_WRITE|CA_LAST)) continue;
        if (d != NULL && dn != GetDeviceNumberEx(eqm,cl->contract.EqmDeviceName,p)) continue;
        if (cl->nclients > 0 && !strnicmp(cl->contract.EqmProperty,p,PROPERTY_NAME_SIZE))
        { /* guarantee immediate execution of any contract using this property */
          if (setCompletion) cl->compStatus = CE_SENDCODE | cc;
          cl->lasttime.tv_sec = cl->lasttime.tv_usec = PRP_SCHEDULE_SIGNAL;
          needsScheduler = TRUE;
        }
      }
    }
    if (scope & CA_HIST)
    { /* the local history subsystem gets scheduled */
      for (i=0; i<nHistoryRecords; i++)
      {
        if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SIZE)) continue;
        if (strnicmp(hstTbl[i].c.EqmProperty,p,PROPERTY_NAME_SIZE)) continue;
        if (d != NULL && dn != GetDeviceNumberEx(eqm,hstTbl[i].c.EqmDeviceName,p)) continue;
        needsHistoryCycle = TRUE;
        hstTbl[i].isScheduled = TRUE;
      }
    }
    if (scope & CA_ALARM)
    { /* the local alarm subsystem gets scheduled */
      for (aw=almWatchList; aw != NULL; aw = aw->nxt)
      {
        if (strncmp(aw->aws.eqm,eqm,EQM_NAME_SHORTSIZE)) continue;
        if (strnicmp(aw->aws.prp,p,PROPERTY_NAME_SIZE)) continue;
        if (d != NULL && dn != GetDeviceNumberEx(eqm,aw->aws.dev,p)) continue;
        needsAlarmCycle = TRUE;
        break;
      }
    }
  }
  if (gSchedulePropertyLazy) 
  {
#   ifdef MULTITHREADED
    if (needsScheduler && hSchedulerSemaphore) 
    {
      ReleaseSystemSemaphore(hSchedulerSemaphore);
      millisleep(0);
    }
#   endif
    if (!gHistoryCycleNeedsScheduler) gHistoryCycleNeedsScheduler = needsHistoryCycle;
    return 0;
  }
  gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
  if (hasSeparateThread) needsAlarmCycle = needsHistoryCycle = FALSE;
  if (needsScheduler) doScheduler(&thisCycleTime,gSystemTick,SCHED_EXTERN,NULL,NULL); /* call the contract */
  if (needsHistoryCycle) historyCycle(TRUE);
  if (needsAlarmCycle && nAlmWatchList > 0) checkAlarmsInWatchTable();

  return 0;
}
TINE_EXPORT int _SystemScheduleProperty(char *eqm, char *prp)
{
  return _SystemSchedulePropertyEx(eqm,prp,CA_NETWORK|CA_HIST|CA_ALARM);
}
TINE_EXPORT int _SystemFireEvent(char *eqm, char *prp)
{
  return _SystemSchedulePropertyEx(eqm,prp,CA_NETWORK|CA_HIST|CA_ALARM);
}
TINE_EXPORT int ResetMultiChannelProperty(char *eqm, char *prp)
{
  ExportPropertyListStruct *pl;
  char *p=NULL,prpstr[4096];

  if (eqm == NULL || prp == NULL) return argument_list_error;
  if (getExportListItem(eqm) == NULL) return non_existent_elem;
  strncpy(prpstr,prp,4096);
  for (p=strtok(prpstr,",|"); p != NULL; p=strtok(NULL,",|"))
  {
    msglog(0,"set mca property reset valid floor for %.8s %.64s",eqm,p);
    for (pl=GetPropertyListStruct(eqm,p,NULL); pl != NULL; pl=pl->next) 
    { /* reset ALL overloads ... */
      if (strnicmp(pl->prpName,p,PROPERTY_NAME_SIZE)) continue;
      pl->mcaValidFloor = time(NULL);
    }
  }
  _SystemSchedulePropertyEx(eqm,prp,CA_SENDCODE|reset_mca_property);
  return 0;
}
TINE_EXPORT int SetSystemUserName(char *usr)
{
  if (usr == NULL) return argument_list_error;
  strncpy(gUserName,usr,USERNAME_SIZE);
  return 0;
}
TINE_EXPORT char *GetSystemErrorString(short cc,char *errstr)
{ /* assume errstr can hold 32 characters or we crash and burn ! */
  if (errstr == NULL) return "[null pointer passed!]";
  if (cc < 0 || cc >= numErr) 
  {
    strcpy(errstr,"[not a system error code!]");
  }
  else
  {
    strcpy(errstr,cc2str(cc));
  }
  return errstr;
}
int ttimeprobe = FALSE;
TINE_EXPORT void SystemTimeProbe(const char *tag)
{
  static struct timeval t0 = {0,0};
  struct timeval t;
  int dms;
  char *ptag = tag == NULL ? "" : (char *)tag;
  gettimeofday(&t,(struct timezone *)NULL);
  if (t0.tv_sec == 0 || !strcmp(ptag,"probestart")) 
  { /* initial probe */
    t0 = t;
    return;
  }
  dms = (t.tv_sec - t0.tv_sec)*1000 + (t.tv_usec-t0.tv_usec)/1000;
  t0 = t;
  ttyoutput("%.64s: [%d] %d sec %d msec since last probe\n",ptag,ttimeprobe,(dms/1000),(dms%1000));
}
TINE_EXPORT UINT32 SystemGetMemory(void)
{
  UINT64 memory = 0;
# if defined(WIN32) && !defined(VS_LEGACY_BUILD) && !defined(MINGW)
  HANDLE hProcess;
  PROCESS_MEMORY_COUNTERS pmc;
  static DWORD pid = 0;
  
  if (pid == 0) pid = GetCurrentProcessId();
  hProcess = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ,FALSE, pid);
  if (NULL == hProcess) return 0;
  if (GetProcessMemoryInfo(hProcess, &pmc, sizeof(pmc)))
  {
    memory = pmc.PagefileUsage;
  }
  CloseHandle(hProcess);
# elif defined(UNIX) && !defined(MACOS) && !defined(SOLARIS)
  FILE *fp;
  UINT64 res, share, text, lib, dat, dt;
  static int page_size = 0;
  if (page_size <= 0) page_size = getpagesize();

  fp = fopen("/proc/self/statm", "r");
  if (fp == NULL) return 0;
# ifdef OS_64BIT
  fscanf(fp, "%ld %ld %ld %ld %ld %ld %ld", &memory, &res, &share, &text, &lib, &dat, &dt);
# else
  fscanf(fp, "%lld %lld %lld %lld %lld %lld %lld", &memory, &res, &share, &text, &lib, &dat, &dt);
# endif
  fclose(fp);

  memory *= page_size;
# endif
  return (UINT32)memory;
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT void SetSystemUseDataProtection(int value)
{
  gSystemUseDataProtection = value;
}
TINE_EXPORT int GetSystemUseDataProtection(void)
{
  return gSystemUseDataProtection;
}
TINE_EXPORT void SetSystemRequireAcknowledgments(int value)
{
  gRequireAcknowledgments = value;
}
TINE_EXPORT int GetSystemRequireAcknowledgments(void)
{
  return gRequireAcknowledgments;
}
TINE_EXPORT void SetSystemDataStamp(int value)
{
  gSystemDataStamp = value;
}
TINE_EXPORT int MakeSystemDataStamp(void)
{
  return gSystemDataStamp;
}
TINE_EXPORT int IsServerRunning(void)
{
  return SystemRunning;
}
TINE_EXPORT int GetContractId(const char *eqm)
{
  ExportListStruct *el=getExportListItem((char *)eqm);
  THRHANDLE tid = GetThreadHandle();
  int i;
  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;/* just a place-holder */
    if (ContractList[i]->threadId != tid) continue;
    if (ContractList[i]->exp != el) continue;
    return ContractList[i]->contractID;
  }
  return -1;
}
TINE_EXPORT int GetLastContractId(void)
{
  if (gEqmFcnContractTableId == -1) return -1;
  return ContractList[gEqmFcnContractTableId]->contractID;
}
TINE_EXPORT int GetLastContractTableId(void)
{
  return gEqmFcnContractTableId;
}
TINE_EXPORT int GetContractAccessRate(int id)
{
  if (id < 0) id = gEqmFcnContractTableId;
  if (id < 0 || id >= ncontracts) return 0;
  return ContractList[id]->pollingInterval;
}
int gIsDynamicLoad = FALSE;
TINE_EXPORT int TineLoadedDynamically(void)
{
  return gIsDynamicLoad;
}
#endif /* !SMALL_TINE_LIB */
