/*
 * File toolkit.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file toolkit.c
 * \brief TINE API tools
 *
 * Many optional routines are offered to make programming easier.  Below are API routines which
 * are useful both for client-side programming and server-side programming.  Some query calls
 * such as GetDeviceNames(), etc. are clearly suitable for information gathering.  The set of
 * of data manipulation routines such at GetValuesAsFloat() or PutValuesFromLongEx() are 
 * designed for easy integration into server code, where format conversion, array-wrapping, 
 * array segmenting, etc. are automatically provided.  Helper routines such as sendPMTrigger()
 * are also provided.
 */
#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <errno.h>
#include "prolog.h"
#include "srvcore.h"
#include "srvip.h"
#include "swaplib.h"
#include "dbtools.h"
#include "toolkit.h"

#define QUERY_TO 500
#define MAP_STRING_SRC 0x01
#define MAP_STRING_DST 0x02
int mapStringArrays(DTYPE *ddst,DTYPE *dsrc,int mapping,int offset);
/**
 * \internal
 */
DBITFIELD *gBitFieldList = NULL;

char feclogfile[80] = "";

static char *months[12] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
TINE_EXPORT time_t parseCtimeStr(char *cTimeStr)
{
  struct tm it;
  char *c = NULL;
  int i;

  for (i=0; i<12; i++) if ((c=strstr(cTimeStr,months[i])) != NULL) break;
  if (c == NULL) return (time_t)0;
  it.tm_mon = i; c += 3;
  it.tm_mday = atoi(c); c += 4;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);
  /* if this string comes from 'date' then there's a TZ field in front of the year */
  c = &cTimeStr[strlen(cTimeStr)-5];
  it.tm_year = atoi(c) - 1900;
  it.tm_isdst = -1;
  return mktime(&it);
}
TINE_EXPORT time_t parseCDateAndTimeStr(char *cDateStr,char *cTimeStr)
{
  struct tm it;
  char *c = NULL;
  int i;

  for (i=0; i<12; i++) if ((c=strstr(cDateStr,months[i])) != NULL) break;
  if (c == NULL) return (time_t)0;
  it.tm_mon = i; c += 3;
  it.tm_mday = atoi(c); c += 4;
  it.tm_year = atoi(c) - 1900;
  if ((c=cTimeStr) == NULL) return (time_t)0;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);  c += 3;
  it.tm_isdst = -1;
  return mktime(&it);
}
TINE_EXPORT double parseTineTimeString(const char *timeStr)
{
  time_t t = time(NULL);
  double dt = (double)t, fs = 0;
  int i, nd = 0, len;
  char *c, day[16], mon[80], year[80], ts[128];
  char hr[64], mn[32], sec[16];
  struct tm tmts, *tnow;
  memset(&tmts, 0, sizeof(struct tm));
  strncpy(ts, timeStr, 128);
  hr[0] = 0;
  tnow = localtime(&t);
  len = (int)strlen(ts);
  strupr(ts);
  for (i = 0; i < len && nd < 2; i++)
  { /* can contain a decimal point for fractions of a second */
    if (ts[i] == '.') nd++; else if (!isdigit(ts[i])) break;
  }
  if (i == len)
  {
    dt = strtod(ts,NULL);
  }
  else if (!strcmp(ts, "NOW"))
  {
    dt = (double)time(NULL);
  }
  else if ((c=strchr(ts, '.')) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day, ts, 16);
    strncpy(mon, c, 64);
    if ((c=strchr(mon, '.')) != NULL)
    {
      *c++ = 0;
      strncpy(year, c, 64);
      if ((c=strchr(year, ' ')) != NULL || (c=strchr(year, '_')) != NULL)
      {
        *c++ = 0;
        strncpy(hr, c, 64);
      }
    }
    else
    {
      sprintf(year, "%d", 1900 + tnow->tm_year);
      if ((c=strchr(mon, ' ')) != NULL || (c=strchr(mon, '_')) != NULL)
      {
        *c++ = 0;
        strncpy(hr, c, 64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c= strchr(hr, ':')) != NULL)
      {
        *c++ = 0;
        strncpy(mn, c, 32);
        if ((c= strchr(mn, ':')) != NULL)
        {
          *c++ = 0;
          strncpy(sec, c, 16);
          if ((c= strchr(sec, '.')) != NULL)
          {
            int ms, nn;
            char *ce;
            *c++ = 0;
            ms = strtol(c,&ce,10);
            *ce = 0;
            nn = (int)strlen(c);
            fs = (double)ms / pow(10,nn);
          }
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min = atoi(mn);
      tmts.tm_sec = atoi(sec);
    }
    tmts.tm_mon = (atoi(mon) + 11) % 12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year < 50) tmts.tm_year += 100; /* given with 2 digits */
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900; /* 'year' is relative to 1900 */
    tmts.tm_isdst = -1;
    dt = (double)mktime(&tmts);
    if (fs > 0) dt += fs;
  }
  return dt;
}
#ifndef SMALL_TINE_LIB
#define NSORT(t,d,n) {\
  t val;\
  int i, k, m;\
  for (i=1; i<(int)(n); i++) {\
    if ((val=((t *)d)[i]) < ((t *)d)[i-1]) {\
      for (k=0; k<i && val>((t *)d)[k]; k++);\
      for (m=i; m>k; m--) ((t *)d)[m] = ((t *)d)[m-1];\
      ((t *)d)[k] = val;\
    }\
  }\
}
#define NFSORT(t,f,d,n) {\
  t val;\
  int i, k, m;\
  for (i=1; i<(int)(n); i++) {\
    val=((t *)d)[i];\
    if (((t *)d)[i].f < ((t *)d)[i-1].f) {\
      for (k=0; k<i && val.f>((t *)d)[k].f; k++);\
      for (m=i; m>k; m--) ((t *)d)[m] = ((t *)d)[m-1];\
      ((t *)d)[k] = val;\
    }\
  }\
}
#define SSORT(t,d,n,s) {\
  t val;\
  int i, k, m;\
  for (i=1; i<(int)(n); i++) {\
    val=((t *)d)[i];\
    if (strnicmp(val.name,((t *)d)[i-1].name,s) < 0) {\
      for (k=0; k<i && strnicmp(val.name,((t *)d)[k].name,s) > 0; k++);\
      for (m=i; m>k; m--) ((t *)d)[m] = ((t *)d)[m-1];\
      ((t *)d)[k] = val;\
    }\
  }\
}
int sortDataValues(DTYPE *d)
{
  if (d == NULL) return invalid_parameter;
  if (d->dArrayLength < 2) return 0;
  switch (d->dFormat)
  {
    case CF_INT16:
      NSORT(SINT16,d->data.vptr,d->dArrayLength);
      break;    
    case CF_UINT16:
      NSORT(UINT16, d->data.vptr, d->dArrayLength);
      break;
    case CF_INT32:
      NSORT(SINT32,d->data.vptr,d->dArrayLength);
      break;    
    case CF_UINT32:
      NSORT(UINT32, d->data.vptr, d->dArrayLength);
      break;
    case CF_FLOAT:
      NSORT(float,d->data.vptr,d->dArrayLength);
      break;    
    case CF_DOUBLE:
      NSORT(double,d->data.vptr,d->dArrayLength);
      break;    
    case CF_INTINT:
      NFSORT(INTINT,i1val,d->data.vptr,d->dArrayLength);
      break;   
    case CF_DBLDBL:
      NFSORT(DBLDBL,d1val,d->data.vptr,d->dArrayLength);
      break;   
   case CF_NAME16:
      SSORT(NAME16,d->data.vptr,d->dArrayLength,fmtsizeof(d->dFormat));
      break;      
   case CF_NAME32:
      SSORT(NAME32,d->data.vptr,d->dArrayLength,fmtsizeof(d->dFormat));
      break;      
   case CF_NAME64:
      SSORT(NAME64,d->data.vptr,d->dArrayLength,fmtsizeof(d->dFormat));
      break;      
    default:
      return illegal_format;
  }
  return 0;
}
TINE_EXPORT int GetArchiveAnnotation(char *ctx,char *srv,char *prp,double target,char *buf,int bufsiz,int timeout)
{
  DTYPE dout;
  char dev[256], *tprp=NULL;
  int isHstSrv = TRUE;

  if (ctx == NULL || buf == NULL) return argument_list_error;
  if (bufsiz < 64) return buffer_too_small;
  if (timeout < 100) timeout = 100;
  if (srv == NULL || strlen(srv) == 0) srv = "HISTORY";
  if (stricmp(srv,"HISTORY")) isHstSrv = FALSE;
  if (isHstSrv)
  {
    if (prp == NULL) prp = "State";
    sprintf(dev,"/%.32s/%.32s/%.64s@%d",ctx,srv,prp,(int)target);
    tprp = "RECORD.CMT";
  }
  else
  {
    sprintf(dev,"/%.32s/%.32s/%d",ctx,srv,(int)target);
    tprp = "HISTORY.CMT";
  }
  DTYPEZERO(dout);
  dout.dFormat = CF_TEXT;
  dout.dArrayLength = bufsiz;
  dout.data.cptr = buf;
  return ExecLinkEx(dev,tprp,&dout,NULL,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,(UINT16)timeout);
}
TINE_EXPORT int GetArchiveAnnotationTimes(char *ctx, char *srv, char *prp, double start, double stop, void *buf, int buffmt, int *bufsiz, int timeout)
{
  DTYPE dout, din;
  char dev[256], *tprp=NULL;
  int startstop[2];
  int isHstSrv = TRUE, cc = 0, siz;
  
  if (ctx == NULL || buf == NULL || bufsiz == NULL) return invalid_parameter;
  if ((siz = *bufsiz) < 1) return dimension_error;
  if (timeout < 100) timeout = 100;
  if (srv == NULL || strlen(srv) == 0) srv = "HISTORY";
  if (stricmp(srv,"HISTORY")) isHstSrv = FALSE;
  if (isHstSrv)
  {
    if (prp == NULL) prp = "State";
    sprintf(dev,"/%.32s/%.32s/%.64s",ctx,srv,prp);
    tprp = "RECORD.CMTS";
  }
  else
  {
    sprintf(dev,"/%.32s/%.32s",ctx,srv);
    tprp = "HISTORY.CMTS";
  }
  DTYPEZERO(dout);
  DTYPEZERO(din);  
  startstop[0] = (int)start;
  startstop[1] = (int)stop;

  if (start > 0 && stop > start)
  {
    din.dArrayLength = 2;
    din.dFormat = CF_INT32;
    din.data.vptr = startstop;
  }
  dout.dArrayLength = siz;
  dout.dFormat = buffmt;
  dout.data.vptr = buf;
  cc = ExecLinkEx(dev,tprp,&dout,&din,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,(UINT16)timeout);
  if (cc == 0)
  {
    *bufsiz = dout.dArrayLength;
    sortDataValues(&dout);    
  }
  return cc;
}
TINE_EXPORT int SetArchiveAnnotation(char *ctx, char *srv, char *prp, double target, char *text, int keepAlive, int timeout)
{
  DTYPE dout, din;
  char dev[256], *tprp=NULL, *keep = keepAlive ? "[KEEPALIVE]" : "";
  int isHstSrv = TRUE;

  if (ctx == NULL || text == NULL) return invalid_parameter;
  if (srv == NULL || strlen(srv) == 0) srv = "HISTORY";
  if (timeout < 100) timeout = 100;
  if (target <= 0) target = (double)time(NULL);
  DTYPEZERO(dout); DTYPEZERO(din);
  din.dFormat = CF_TEXT;
  din.dArrayLength = (UINT32)strlen(text);
  din.data.cptr = text;
  if (stricmp(srv,"HISTORY")) isHstSrv = FALSE;
  if (isHstSrv)
  {
    if (prp == NULL) prp = "State";
    sprintf(dev,"/%.32s/%.32s/%.64s@%d%s",ctx,srv,prp,(int)target,keep);
    tprp = "RECORD.CMT";
  }
  else
  {
    sprintf(dev,"/%.32s/%.32s/%d",ctx,srv,(int)target);
    tprp = "HISTORY.CMT";
  }
  return ExecLinkEx(dev,tprp,NULL,&din,CA_WRITE|CA_NOCALLBACKS,(UINT16)timeout);
}
TINE_EXPORT int PropertyDataExchange(void *data,short fmt,short siz,
                        DTYPE *dout, DTYPE *din,short access)
{
  BYTE *dst=NULL,*src=NULL;
  int nbytes=0;
  if (din->dArrayLength > 0 && data != NULL)
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    switch(fmt)
    {
      case CF_INT16: 
        return GetValuesAsShort(din,(SINT16 *)data,siz);
      case CF_UINT16:
        return GetValuesAsUnsignedShort(din, (UINT16 *)data, siz);
      case CF_INT32:
        return GetValuesAsLong(din,(SINT32 *)data,siz);
      case CF_UINT32:
        return GetValuesAsUnsignedLong(din, (UINT32 *)data, siz);
      case CF_FLOAT:
        return GetValuesAsFloat(din,(float *)data,siz);
      case CF_DOUBLE:
        return GetValuesAsDouble(din,(double *)data,siz);
      default:
        if (din->dFormat != fmt) return illegal_format;
        dst = (BYTE *)data; src = din->data.bptr; nbytes = (int)(din->dArrayLength*fmtsizeof(fmt));
        goto err;
    }
  }
  if (dout->dArrayLength > 0 && data != NULL)
  {
    switch(fmt)
    {
      case CF_INT16: 
        return PutValuesFromShort(dout,(SINT16 *)data,siz);
      case CF_UINT16:
        return PutValuesFromUnsignedShort(dout, (UINT16 *)data, siz);
      case CF_INT32:
        return PutValuesFromLong(dout,(SINT32 *)data,siz);
      case CF_UINT32:
        return PutValuesFromUnsignedLong(dout, (UINT32 *)data, siz);
      case CF_FLOAT:
        return PutValuesFromFloat(dout,(float *)data,siz);
      case CF_DOUBLE:
        return PutValuesFromDouble(dout,(double *)data,siz);
      default:
        if (dout->dFormat != fmt) return illegal_format;
        dst = dout->data.bptr; src = (BYTE *)data; nbytes = (int)(dout->dArrayLength*fmtsizeof(fmt));
        goto err;
    }
  }
err:
  if (dst != NULL && src != NULL && nbytes > 0) memcpy(dst,src,nbytes);
  return 0;
}
time_t getTimeStampFromLogEntry(char *str)
{
  struct tm tms;
  tms.tm_mday = atoi(str);
  tms.tm_mon = atoi(&str[3]) - 1;
  tms.tm_year = 100 + atoi(&str[6]);
  tms.tm_hour = atoi(&str[9]);
  tms.tm_min = atoi(&str[12]);
  tms.tm_sec = atoi(&str[15]);
  tms.tm_isdst = memcmp(&str[22],ltzname[1],3) == 0 ? 1 : 0;
  return mktime(&tms);
}
int findLogFileEntry(FILE *fp,int fpos,time_t timestamp)
{
  int pos=fpos;
  time_t t=0;
  char str[256];

  fseek(fp,pos,SEEK_SET);
  while (csvGetLine(str,256,fp) != NULL)
  {
    t = getTimeStampFromLogEntry(str);
    if (t > timestamp) return pos;
    pos += (int)strlen(str);
  }
  return pos;
}
int readFecLogFileEx(char *fname,char *textbuf,time_t start,time_t stop,int sizeInBytes)
{
  int cc=0,done=0,useAbsolutePath=0;
  FILE *fp = NULL, *fpb = NULL;
  char str[256],fn[256],bakname[96], *c;
  int sz=0,pos=0;
  int pos1=0,pos2=0;
  static char *fmod = "r";

  if (nofeclog) return 0;
  if (fname == NULL) return no_such_file;
  if (stop <= start) stop = time(NULL);
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else sprintf(fn,"%.128s%s",gFeclogPath,fname);
  if ((fp=fopen(fn,fmod)) == NULL) goto checkbak;
  pos1 = findLogFileEntry(fp,0,start);
  pos2 = findLogFileEntry(fp,pos1,stop);
  pos = 0;
checkbak:
  if (pos1 == 0)  /* look for back file */
  {
    strncpy(bakname,fname,95); bakname[95] = 0;
    if ((c=strstr(bakname,".log")) != NULL) *c = 0;
    strcat(bakname,".bak");
    if (useAbsolutePath) strncpy(fn,bakname,128); else sprintf(fn,"%.128s%s",gFeclogPath,bakname);
    if ((fpb=fopen(fn,fmod)) == NULL) goto filesread;
    pos1 = findLogFileEntry(fpb,0,start);
    pos2 = findLogFileEntry(fpb,pos1,stop);
    fseek(fpb,pos1,SEEK_SET);
    while (csvGetLine(str,256,fp) != NULL)
    {
      if ((sz=(int)strlen(str)) + pos > sizeInBytes) 
      {
        done = TRUE;
        break;
      }
      strncpy(&textbuf[pos],str,sz);
      if ((pos += sz) > pos2) 
      {
        done = TRUE;
        break;
      }
    }
    fclose(fpb);
    pos1 = 0;
  }
filesread:
  if (fp)
  {
    pos2 = findLogFileEntry(fp,pos1,stop);
    fseek(fp,pos1,SEEK_SET);
    while (csvGetLine(str,256,fp) != NULL && !done)
    {
      if ((sz=(int)strlen(str)) + pos > sizeInBytes) 
      {
        done = TRUE;
        break;
      }
      strncpy(&textbuf[pos],str,sz);
      if ((pos += sz) > pos2) 
      {
        done = TRUE;
        break;
      }
    }
    fclose(fp);
  }
  textbuf[pos] = 0;
  if (sz == 0) cc = no_such_file;
  if (cc) dbglog("%s : %s",fn,cc2str(cc));
  return cc;
}
int GetTagList(char *context,char *tagtype,int *tagfmt,void *taglist,int *num)
{
  int cc = 0, fmt, retryfmt;
  short n = 0;
  char *c,prop[32],srv[64],*sub = NULL;
  DTYPE dout,din, *dinptr = NULL;

  if (tagfmt == NULL) return argument_list_error;
  if (num == NULL || *num <= 0) return buffer_too_small;
  switch (*tagfmt)
  {
    case CF_NAME16:
    case CF_NAME32:
      retryfmt = CF_NAME16;
      break;
    case CF_NAME64:
      retryfmt = CF_NAME32;
      break;
    default:
      return illegal_format;
  }
  strncpy(srv,ENS_TAG,16);
  /* is a context given ? */
  if (context != NULL && strlen(context) > 0) 
  { /* is a subsystem given ? */
    if ((c=strchr(context,'/')) != NULL)
    { /* has the form <context>/<subsystem> */
      *c++ = 0; sub = c;
    }
    if (strlen(context) > 0) 
    { /* might have been given as /<subsystem> ! */
      strcat(srv,"/");
      strncat(srv,context,CONTEXT_NAME_SIZE);
    }
  }
  if (sub != NULL)
  {
    memset(&din,0,sizeof(DTYPE));
    din.dFormat = CF_TEXT;
    din.dArrayLength = (UINT32)strlen(sub);
    din.data.cptr = sub;
    dinptr = &din;
  }
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_INT16;
  dout.dArrayLength = 1;
  dout.data.vptr = &n;
  sprintf(prop,"N%s",tagtype);
  if ((cc=ExecLinkEx(srv,prop,&dout,dinptr,CA_READ|CA_RETRY|CA_NOCALLBACKS,QUERY_TO)) != 0) return cc;
  if (n > *num) n = *num;
  fmt = *tagfmt;
  if (taglist != NULL && n > 0)
  { /* Get Tags from name server (synchronous call) */
retryIllegalFmt:
    dout.dFormat = fmt;
    dout.dArrayLength = n;
    dout.data.vptr = taglist;
    if ((cc=ExecLinkEx(srv,tagtype,&dout,dinptr,CA_READ|CA_RETRY|CA_NOCALLBACKS,QUERY_TO)) != 0) 
    {
      if (cc == illegal_format && fmt != retryfmt) 
      {
        fmt = retryfmt;
        goto retryIllegalFmt;
      }
      return cc;
    }
  }
  *tagfmt = fmt;
  *num = n;
  return 0;
}
TINE_EXPORT int GetDevicePropertyInformationEx(char *srv,char *prp,PropertyQueryStructEx *srvProps,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int cc = 0;
  short n;
  BYTE *data = (BYTE *)srvProps;
  NAME32 n32;
# if !defined(INTEL) || defined(VXWORKS) || defined(MACOS)
  if (tmpWorkArea == NULL && (tmpWorkArea=(BYTE *)SystemCalloc(1,tmpWorkAreaSize)) == NULL)
  {
    return out_of_local_memory;
  }
  data = tmpWorkArea;
# endif
  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if (prp != NULL)
  {
    DTYPEZERO(din);
    strncpy(n32.name,prp,32);
    din.dArrayLength = 1;
    din.dFormat = CF_NAME32;
    din.data.vptr = &n32;
    dinptr = &din;
  }
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n*PROPERTYQUERYSTRUCTEX_SIZE;
  dout.data.vptr = data;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,dinptr,CA_READ|CA_RETRY|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO)) != 0) return cc;
  *num = dout.dArrayLength;
# ifndef INTEL
  if (*num > 0) PrepIncomingPQSX(srvProps,data,*num);
# endif
  return 0;
}
/**
 * \internal
 */
int parseFullDeviceName(char *fullname,char *dev,char *prp,int isHistoryCall)
{
  char *c,tmp[256];
  int isBracketed = FALSE;

  strncpy(tmp,fullname,256);
  if ((c=strchr(tmp,'[')) != 0)
  { /* fullname given as "/<context>/<server>/<device>[<property>]" */
    *c++ = 0;
    strncpy(prp,c,PROPERTY_NAME_SIZE);
    if ((c=strchr(prp,']')) != 0) *c = 0;
    isBracketed = TRUE;
  }
  if (!isBracketed)
  { /* fullname given as "/<context>/<server>/<device>/<property>" */
    strrev(tmp);
    if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
    *c++ = 0;
    strncpy(prp,tmp,PROPERTY_NAME_SIZE); strrev(prp);
    strncpy(dev,c,STATUS_SIZE); strrev(dev);
  }
  else
  {
    strncpy(dev,tmp,132);
  }
  if (isHistoryCall)
  {
    if (strstr(dev,"HISTORY") == NULL && /* not a central archive call */
      strstr(prp,".HIST") == NULL && strstr(prp,".HST") == NULL) /* meta property not yet appended */
    {
      if (strlen(prp) < PROPERTY_NAME_SIZE-4) strcat(prp,".HST");
    }
  }
  return 0;
}
/**
 * \internal
 */
void augmentNameSize(int fmt,char *list,int siz,int stepsize)
{
  int i,fromlen,tolen;
  char scratch[64];
  if (list == NULL || siz <= 0) return;
  if (stepsize <= 0) stepsize = 1;
  switch (fmt)
  {
    case CF_NAME16:
      fromlen = 16; tolen = 32*stepsize; break;
    case CF_NAME32:
      fromlen = 32; tolen = 64; break;
    default: return;
  }
  for (i=siz-1; i >= 0; i--)
  {
    memset(scratch,0,64);
    memcpy(scratch,list + i*fromlen, fromlen);
    memcpy(list + i*tolen,scratch, tolen);
  }
}
#endif /* not SMALL_TINE_LIB */

int GetIntValue(DTYPE *d,SINT32 *ival)
{
  if (d == NULL || ival == NULL) return argument_list_error;
  if (d->dArrayLength < 1) return dimension_error;
  switch (d->dFormat)
  {
    case CF_FLTINT: 
      *ival = (SINT32)((FLTINT *)d->data.vptr)->ival;
      return 0;
    case CF_INTINT: 
      *ival = (SINT32)((INTINT *)d->data.vptr)->i2val;
      return 0;
    case CF_DBLDBL: 
      *ival = (SINT32)((DBLDBL *)d->data.vptr)->d2val;
      return 0;
    case CF_FLTFLTINT: 
      *ival = (SINT32)((FLTFLTINT *)d->data.vptr)->ival;
      return 0;
    case CF_INTINTINT: 
      *ival = (SINT32)((INTINTINT *)d->data.vptr)->i3val;
      return 0;
    case CF_NAME16FI:
      *ival = (SINT32)((NAME16FI *)d->data.vptr)->ival;
      return 0;
    case CF_NAME16II:
      *ival = (SINT32)((NAME16II *)d->data.vptr)->i2val;
      return 0;
    case CF_INTFLTINT:
      *ival = (SINT32)((INTFLTINT *)d->data.vptr)->i2val;
      return 0;
    case CF_FLTINTINT:
      *ival = (SINT32)((FLTINTINT *)d->data.vptr)->i2val;
      return 0;
    default:
      break;
  }
  return illegal_format;
}
TINE_EXPORT short GetBitfieldFormat(short fmt)
{
  switch (fmt)
  {
    case CF_INT8:
    case CF_BITFIELD8: return CF_BITFIELD8;
    case CF_INT16:
    case CF_UINT16:
    case CF_BITFIELD16: return CF_BITFIELD16;
    case CF_DEFAULT:
    case CF_INT32:
    case CF_UINT32:
    case CF_BITFIELD32: return CF_BITFIELD32;
    case CF_INT64:
    case CF_BITFIELD64: return CF_BITFIELD64;
  }
  return fmt;
}
TINE_EXPORT short GetIntegerFormat(short fmt)
{
  switch (fmt)
  {
    case CF_INT8:
    case CF_BITFIELD8: return CF_INT8;
    case CF_INT16:
    case CF_UINT16:
    case CF_BITFIELD16: return CF_INT16;
    case CF_DEFAULT:
    case CF_INT32:
    case CF_UINT32:
    case CF_BITFIELD32: return CF_INT32;
    case CF_INT64:
    case CF_BITFIELD64: return CF_INT64;
  }
  return fmt;
}
int isBitfieldAccess(DTYPE *d)
{
  if (d == NULL) return FALSE;
  if (d->dFormat < CF_BITFIELD8 || d->dFormat > CF_BITFIELD32) return FALSE;
  if (d->dTag[0] == 0) return FALSE;
  return TRUE;
}
/**
 * \internal
 */
int isBitfieldProperty(char *srv,char *tag,char *prp,char *fld)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  int i, p;

  if (prp == NULL || bf == NULL) return FALSE;
  if ((p=(int)strlen(prp)) == 0) return FALSE;
  p = MIN(p,PROPERTY_NAME_SIZE) - 1;
  for (i=p; i>0; i--) if (prp[i] == '.') break;
  if (i == 0 || i == p) return FALSE;
  strncpy(fld,&prp[i+1],FIELD_NAME_SIZE); 
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    if (!strnicmp(bff->field,fld,FIELD_NAME_SIZE)) 
    { /* got it: now peel off the field */
      prp[i] = 0;
      return TRUE;
    }
  }
  return FALSE;  
}
/**
 * \internal
 */
int isTaggedStructAccess(DTYPE *d)
{
  if (d == NULL) return FALSE;
  if (d->dFormat != CF_STRUCT) return FALSE;
  if (d->dTag[0] == 0) return FALSE;
  return TRUE;
}
/**
 * \internal
 */
int isStructProperty(char *tag,char *prp,char *fld)
{
  struct structStruct *ss = findstruct(tag);
  struct structFormat *sf;
  int i, p;

  if (prp == NULL || ss == NULL) return FALSE;
  if ((p=(int)strlen(prp)) == 0) return FALSE;
  p = MIN(p,PROPERTY_NAME_SIZE) - 1;
  for (i=p; i>0; i--) if (prp[i] == '.') break;
  if (i == 0 || i == p) return FALSE;
  strncpy(fld,&prp[i+1],FIELD_NAME_SIZE); prp[i] = 0;
  sf = ss->s;
  for (sf=ss->s; sf != NULL; sf=sf->nxt)
  {
    if (!strnicmp(sf->field,fld,FIELD_NAME_SIZE)) return TRUE;
  }
  return FALSE;  
}
/**
 * \internal
 */
int acquireAndRegisterStruct(char *devName, char *tag,int num)
{
  int i, n, siz, fmt, len, slen, offset, cc = 0, usettag = FALSE;
# if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
  int readFromCache = FALSE;
# endif
  DTYPE dout,din;
  NAME64DBLDBL n64dd[64];
  INTINT ii[64];
  char *c, *ptag, stag[TAG_NAME_SIZE+8], key[128];
  struct structStruct *s;
  struct structFormat *sf;
# define ttag "_tEmPoRaRy_"

  if (devName == NULL || tag == NULL) ccerr(argument_list_error);
  if (strlen(devName) == 0 || strlen(tag) == 0) ccerr(invalid_parameter);
  strncpy(key,devName,128);
  c = strpbrk(&key[1],DELIMITER); 
  if ((c=strpbrk(++c,DELIMITER)) != NULL) *c = 0; /* lop off the end */
  if (findstructEx(tag,key) != NULL) return already_assigned;

  dout.dFormat = CF_NAME64DBLDBL;
  dout.dArrayLength = 64;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = n64dd;

retryFormat:
  din.dFormat = CF_TEXT;
  din.dArrayLength = (UINT32)strlen(tag);
  memset(din.dTag,0,TAG_NAME_SIZE);
  din.data.cptr = tag;

  if ((cc=ExecLinkEx(devName,"STRUCTFORMAT",&dout,&din,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,200)) != 0) 
  {
    int rc = cc;
    if (cc == illegal_format && dout.dFormat == CF_NAME64DBLDBL)
    { /* legacy ? */
      dout.dFormat = CF_INTINT;
      dout.data.vptr = ii;
      goto retryFormat;
    }
    dout.dFormat = CF_NAME64DBLDBL;
    dout.data.vptr = n64dd;
#   if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
    if ((cc=readStructCacheFile(key,tag,&dout)) == 0)
    {
      feclog("acquire struct %.16s from local cache",tag);
      readFromCache = TRUE;
    }
    else
#   endif
    {
      feclogEx(TINE_LOGLEVEL_ERROR,"acquire struct %.16s : %.32s (remote), %.32s (local)",tag,cc2str(rc),cc2str(cc));
      ccerr(cc);
    }
  }
  if (dout.dFormat == CF_INTINT)
  { /* covert to modern */
    for (i=0; i<64; i++)
    {
      n64dd[i].d1val = (double)ii[i].i1val;
      n64dd[1].d2val = (double)ii[i].i2val;
      if (ii[i].i2val == CF_NULL)
      {
        strncpy(n64dd[i].name,tag,TAG_NAME_SIZE);
        break;
      }
      sprintf(n64dd[i].name,"field_%d",i);
    }
  }

  n = dout.dArrayLength;
  siz = (int)n64dd[n-1].d1val;
  offset = 0;
  s = findstruct(tag);
  while (s != NULL)
  { /* same struct tag from another server */
    usettag = TRUE;
    for (i=n-2,sf=s->s; i >= 0 && sf != NULL; i--, sf=sf->nxt)
    {
      if (sf->siz != (int)n64dd[i].d1val) break;
      if (sf->fmt != (int)n64dd[i].d2val) break;
      if (strnicmp(sf->field,n64dd[i].name,FIELD_NAME_SIZE)) break;
    }
    if (i < 0)
    { /* a match! -> add the server to the list */
      addKeyToStruct(s,key);
      return 0;
    }
    s = findstructEx(tag,STRUCT_LIST_MARKER,s->next);
  }
  ptag = usettag ? ttag : tag;
  for (i=0; i<n-1; i++)
  {
    len = (int)n64dd[i].d1val;
    fmt = (int)n64dd[i].d2val;
    slen = 1;
    if (fmt == CF_STRUCT)
    { /* nested structs -> find struct tag ... */
      if (n64dd[i].name[0] != '<') ccerr(illegal_format);
      strncpy(stag,&n64dd[i].name[1],TAG_NAME_SIZE);
      if ((c=strchr(stag,'>')) == NULL) ccerr(illegal_format);
      *c = 0;
      if ((cc=acquireAndRegisterStruct(devName,stag,len)) != 0) ccerr(cc);
      slen = GetStructSize(stag);
    }
    addFieldToStruct(ptag,offset,len,fmt,n64dd[i].name);
    offset += slen * len * fmtsizeof(fmt);
  }
  if ((cc=sealTaggedStruct(ptag,siz,num)) != 0) ccerr(cc);
  if ((s=findstruct(ptag)) == NULL) ccerr(invalid_structure_tag);
  addKeyToStruct(s,key);
  if (usettag) strncpy(s->name,tag,TAG_NAME_SIZE);
# if !defined(FS_RDONLY) && !defined(SMALL_TINE_LIB)
  if (!readFromCache) writeStructCacheFile(key,tag);
# endif
err:
  if (cc && cc != already_assigned) 
    msglog(cc,"acquireAndRegisterStruct %.64s %.16s : %.32s",devName,tag,cc2str(cc));
  return cc;
}
TINE_EXPORT int tokenizeFullName(char *fullName,char **ctx,char **srv,char **dev)
{
  int cc = 0;
  char *c;

  if (fullName == NULL) ccerr(argument_list_error);
  if (ctx == NULL || srv == NULL || dev == NULL) ccerr(argument_list_error);
  if (fullName[0] == '/') 
  {
    *ctx = &fullName[1];
    if ((c=strchr(*ctx,'/')) == NULL) ccerr(invalid_name);
    *c++ = 0;
    *srv = c;
  }
  else
  {
    *ctx = "DEFAULT";
    *srv = fullName;
  }
  if ((c=strchr(*srv,'/')) == NULL)
  {
    *dev = "";
  }
  else
  {
    *c++ = 0;
    *dev = c;
  }
err:
  return cc;
}
/**
 * TODO:
 * This split logic shows up elsewhere in other forms !
 * Make everyone use the same routine !
 */
int splitFullName(const char *fullName,char *srvName,char *devName)
{
  char *c0,*c1,dev[128];
  int cc = 0;
  if (fullName == NULL) ccerr(argument_list_error);
  strncpy(dev,fullName,128); strtrm(dev);
  if ((c0=strchr(&dev[1],'/')) == NULL) 
  { /* just the device name was given */
    if (srvName != NULL) srvName[0] = 0;
    if (devName != NULL) strncpy(devName,fullName,DEVICE_NAME_SIZE);
    ccerr(0);
  }
  c0++;
  if ((c1=strchr(c0,'/')) != NULL) *c1++ = 0;
  if (dev[0] == '/') 
  { /* normal case: context given */
    if (srvName != NULL) strncpy(srvName,dev,2*CONTEXT_NAME_SIZE+2);
    if (c1 == NULL) c1 = "";
    if (devName != NULL) strncpy(devName,c1,DEVICE_NAME_SIZE);
  }
  else
  { /* no context given */
    *c0++ = 0;
    if (srvName != NULL) strncpy(srvName,dev,2*CONTEXT_NAME_SIZE+2);
    if (devName != NULL) strncpy(devName,c0,DEVICE_NAME_SIZE);
  }
err:
  return cc;
}
#endif /* NO_DOXYGEN */

int fecLogFileDepth = 500;
#if !defined(SMALL_TINE_LIB)
TINE_EXPORT void AppendHistoryLog(char *text,...)
{
# ifndef FS_RDONLY
  double ts = makeDataTimeStamp();
  FILE *fp = NULL, *fpb = NULL;
  SINT32 flen;
  int k, cc = 0;
  char str[512],tmpfilename[128];
  va_list args;

  if (strlen(feclogfile) == 0) sprintf(feclogfile,"%s.log",gFecName);
  if ((fp=fopen(feclogfile,"a")) == NULL)
  {
    perror("LOG: fopen");
    ccerr(file_error);
  }
  sprintf(str,"%s[%.16s] ",getDataTimeString(ts,0),gFecName);
# if defined(UNIX) || defined(WIN32)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) 
  {
    if (strchr(text,'%') == NULL)
    { /* then truncate */
      strncpy(&str[strlen(str)],text,256); str[256] = 0;
    }
    ccerr(string_too_long);
  }
  else
  {
    va_start(args,text);
    vsprintf(&str[strlen(str)],text,args);
    va_end(args);
  }
# endif
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  fprintf(fp,"%s",str);

  fseek(fp, 0L, SEEK_END);
  flen = ftell(fp);

  if (flen > 2*fecLogFileDepth*80)
  {
    fclose(fp);
    strncpy(tmpfilename,feclogfile,79); tmpfilename[79] = 0;
    strncpy(&tmpfilename[strlen(tmpfilename)-3],"BAK",3);
    rename(feclogfile,tmpfilename);
    if ((fp=fopen(feclogfile,"w")) == NULL) ccerr(file_error);
    if ((fpb=fopen(tmpfilename,"r")) == NULL) ccerr(no_such_file);
    for (k=0; csvGetLine(str,256,fpb) != NULL; k++)
    {
      if (k < fecLogFileDepth) continue;
      fwrite(str,strlen(str),1,fp);
    }

  }
err:
  if (fp) fclose(fp);
  if (fpb) { fclose(fpb); remove(tmpfilename); }
  if (cc) dbglog("AppendHistoryLog : %s",cc2str(cc));
# endif
  return;
}
TINE_EXPORT int AcquireAndRegisterStruct(char *devName, char *tag,int num)
{
  return acquireAndRegisterStruct(devName,tag,num);
}
TINE_EXPORT int GetStructureAsString(char *tag,void *structref,char *strbuf,int buflen)
{
  int i, n, cc=0, slen=0, nfields=0;
  struct structStruct *ss = findstruct(tag);
  struct structFormat *sf, **sffa;
  char *c=strbuf, ws[16], wsi[16];
  BYTE *structbytes = (BYTE *)structref;
  DTYPE d;
  static int level=0;

  if (ss == NULL) return not_registered;
  if (structref == NULL || strbuf == NULL) return argument_list_error;
  for (sf=ss->s; sf != NULL; sf=sf->nxt) nfields++;
  if (buflen < nfields*32) return buffer_too_small;
  DTYPEZERO(d);
  sffa = (struct structFormat **)alloca(nfields*sizeof(size_t));
  if (sffa == NULL) return out_of_local_memory;
  level++;
  memset(ws,0,16); memset(wsi,0,16);
  for (i=0; i<level*2 && i<16; i++)
  {
    if (i > 1) ws[i-2] = ' ';
    wsi[i] = ' ';
  }
  for (sf=ss->s,i=nfields-1; sf != NULL; sf=sf->nxt,i--)
  { /* the link list is reverse ordered => prepare the proper order */
    sffa[i] = sf;
  }
  sprintf(strbuf,"%s{\n",ws); slen = (int)strlen(strbuf); c = &strbuf[slen]; 
  for (i=0; i<nfields; i++)
  {
    if (buflen - slen < 32) break; /* a modicum of string length protection */
    d.dArrayLength = sffa[i]->siz;
    d.dFormat = sffa[i]->fmt;
    strncpy(d.dTag,sffa[i]->stag,TAG_NAME_SIZE);
    d.data.bptr = &structbytes[sffa[i]->addr];
    sprintf(c,"%s%.16s: ",wsi,sffa[i]->field);
    slen = (int)strlen(strbuf); c = &strbuf[slen];
    n = buflen - slen;
    if (d.dFormat == CF_STRUCT)
    {
      *c++ = '\n'; *c = 0; slen++;
      if ((cc= GetStructureAsString(d.dTag,d.data.bptr,c,n)) != 0) ccerr(cc);
    }
    else
    {
      GetValuesArrayAsString(&d,c,(UINT32 *)&n,", ");
    }
    slen = (int)strlen(strbuf); c = &strbuf[slen];
  }
  sprintf(c,"%s}\n",ws); slen = (int)strlen(strbuf); c = &strbuf[slen];
err:
  level--;
  return cc;
}
TINE_EXPORT int GetArchivedData(char *devsrv,int index,time_t start,time_t stop,FLTINT *fiDataArray,short *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = index;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  dout.data.vptr = fiDataArray;
  din.dArrayLength = 3;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
TINE_EXPORT int GetArchivedTraceDataAsFloat(char *devsrv,time_t start,time_t stop,float *fDataArray,int num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = num;
  dout.dFormat = CF_FLOAT;
  dout.data.vptr = fDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  return ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_NOCALLBACKS,200);
}
TINE_EXPORT int GetArchivedDataAsSnapshot(char *devsrv,time_t *target,float *fDataArray,int size)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (size <= 0) return buffer_too_small;
  if (target == NULL) return argument_list_error;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)(*target);
  startstopArray[1] = (UINT32)(*target);
  dout.dArrayLength = size;
  dout.dFormat = CF_FLOAT;
  dout.data.vptr = fDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_NOCALLBACKS,200)) != 0) *target = 0;
  else *target = (time_t)gDataTimeStamp;

  return cc;
}
TINE_EXPORT int GetArchivedDataAsFloat(char *devsrv,time_t start,time_t stop,FLTINT *fiDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  dout.data.vptr = fiDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_NOCALLBACKS|CA_MUTABLE,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
BYTE *_arch_rb_buf_ = NULL;
int _arch_rb_buf_capacity = 0;
TINE_EXPORT void FreeArchivedDataReadbackBuffer(void)
{
  SystemFree(_arch_rb_buf_);
  _arch_rb_buf_capacity = 0;
}
TINE_EXPORT int GetArchivedDataAsAny(char *devsrv,time_t start,time_t stop,HstHdr *dataHdr,BYTE *data,int dataFmt,char *dataTag, int *num)
{
  return GetArchivedDataAsAnyEx(devsrv,start,stop,0,0,dataHdr,data,dataFmt,dataTag,num);
}
TINE_EXPORT int GetArchivedDataAsAnyEx(char *devsrv,time_t start,time_t stop,int index,int sampleRaster,HstHdr *dataHdr,BYTE *data,int dataFmt,char *dataTag, int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[4];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE],*tag;
  int i, fsiz, rsiz, cc;
  DIMAGE *img;
  BYTE *rb;
  char **str;

  if (data == NULL) return argument_list_error;
  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = index > 0 ? (UINT32)index : 0;
  startstopArray[3] = sampleRaster > 0 ? (UINT32)sampleRaster : 0;
  dout.dArrayLength = *num;
  dout.dFormat = CF_HISTORY;
  tag = dataFmt == CF_STRUCT ? dataTag : itof(dataFmt);
  if ((fsiz=getCarriedFormatSize(tag)) < 0) return -fsiz;
  strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  switch (dataFmt)
  { 
    case CF_IMAGE:
      /* special case: look at the first one */
      img = (DIMAGE *)data;
      fsiz = IMAGE_HDR_SIZE + img[0].frameBufferSize;
      rsiz = fsiz + HISTORY_HDR_SIZE;
      break;
    case CF_STRING:
    case CF_KEYVALUE:
      rsiz = STRING_STD_CAPACITY + HISTORY_HDR_SIZE;
      break;
    default:
      rsiz = fsiz + HISTORY_HDR_SIZE;
      break;
  }   

  if (_arch_rb_buf_ == NULL)
  {
    if ((_arch_rb_buf_=(BYTE *)SystemCalloc(*num,rsiz)) == NULL) return -out_of_local_memory;
    _arch_rb_buf_capacity = *num * rsiz;
  }
  else if (*num * rsiz > _arch_rb_buf_capacity)
  {
    if ((rb=realloc(_arch_rb_buf_,*num * rsiz)) != NULL)
    {
      _arch_rb_buf_ = rb;
      _arch_rb_buf_capacity = *num * rsiz;
    }
  }
  rb = _arch_rb_buf_;
  dout.data.vptr = rb;

  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_NOCALLBACKS|CA_MUTABLE,200)) != 0) 
  { /* error condition ! */
    *num = 0;
  }
  else 
  { /* so far, so good */
    if (dataFmt == CF_IMAGE)
    { /* very special case */
      img = (DIMAGE *)data;
      rsiz = fsiz = 0;
      for (i=0; i<(int)dout.dArrayLength && i<*num; i++)
      {
        memcpy(&dataHdr[i],rb+rsiz,HISTORY_HDR_SIZE);
        rsiz += HISTORY_HDR_SIZE;
        memcpy(&img[i],rb+rsiz,IMAGE_HDR_SIZE);
        rsiz += IMAGE_HDR_SIZE;
        fsiz = img[i].frameHeader.appendedFrameSize;
        memcpy(img[i].frameBuffer,rb+rsiz,MIN((int)img[i].frameBufferSize,fsiz));
        rsiz += fsiz;
      }
      *num = i;
    }
    else if (dataFmt == CF_STRING || dataFmt == CF_KEYVALUE)
    {
      rsiz = fsiz = 0;
      str = (char **)data;
      for (i=0; i<(int)dout.dArrayLength && i<*num; i++)
      {
        memcpy(&dataHdr[i],rb+rsiz,HISTORY_HDR_SIZE);
        rsiz += HISTORY_HDR_SIZE;
        str[i] = (char *)rb+rsiz;
        fsiz = (int)strlen(str[i]) + 1;
        rsiz += fsiz;
      }
      *num = i;      
    }
    else
    { /* normal case */
      for (i=0; i<(int)dout.dArrayLength; i++)
      {
        memcpy(&dataHdr[i],rb+i*rsiz,HISTORY_HDR_SIZE);
        memcpy(data+i*fsiz,rb+i*rsiz+HISTORY_HDR_SIZE,fsiz);
      }
      *num = (int)dout.dArrayLength;
    }
  }
  return cc;
}
TINE_EXPORT int GetArchivedDataAsText(char *devsrv,time_t start,time_t stop,NAME32I *niDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  DTYPEZERO(dout); DTYPEZERO(din);
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_NAME32I;
  dout.data.vptr = niDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_NOCALLBACKS|CA_MUTABLE,200)) != 0) *num = 0;
  else *num = (int)dout.dArrayLength;

  return cc;
}
TINE_EXPORT int PutArchivedData(char *devsrv, BYTE *data, int dataFmt, int num, double dataTime, int sysStamp, int usrStamp)
{
  DTYPE din;
  BYTE *payload;
  char *c, aServer[STATUS_SIZE], prp[PROPERTY_NAME_SIZE];
  double ts = dataTime;
  int sst = sysStamp, ust = usrStamp;
  int cc, n, fsz;

  if (devsrv == NULL || data == NULL || dataFmt == CF_NULL || num == 0) return argument_list_error;
  if ((cc = parseFullDeviceName(devsrv, aServer, prp, FALSE)) != 0) return cc;
  if (aServer[0] != '/') return invalid_parameter;
  if ((c=strchr(&aServer[1], '/')) != NULL) *c = 0;
  if (strlen(aServer) > CONTEXT_NAME_SIZE) return illegal_context;
  strcat(aServer, "/ARCHIVER/#0");
  DTYPEZERO(din);
  if ((fsz = fmtsizeof(dataFmt)) > 16) return illegal_format;
  n = num + 16 / fsz;
  if ((payload=(BYTE *)malloc(n*fsz)) == NULL) return out_of_local_memory;
  memcpy(payload, data, num*fsz);
  DSWAP(ts);
  memcpy(&payload[num*fsz], &ts, 8);
  ISWAP(sst);
  memcpy(&payload[num*fsz+8], &sst, 4);
  ISWAP(ust);
  memcpy(&payload[num*fsz + 12], &ust, 4);
  din.dArrayLength = n;
  din.data.bptr = payload;
  din.dFormat = dataFmt;
  return ExecLinkEx(aServer, prp, NULL, &din, CA_WRITE | CA_NOCALLBACKS, 200);
}
TINE_EXPORT int GetDeviceContexts(NAME16 *clist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(NULL,"CONTEXTS",&fmt,clist,num);
  return cc;
}
TINE_EXPORT int GetDeviceContextsFromFile(NAME16 *clist,int *num)
{
  FILE  *fp=NULL;
  int i=0,j,n,cc=0,done=0,con_col=-1;
  char fname[256],str[256],hdr[256],scratch[32];

  if (num == NULL || clist == NULL) ccerr(argument_list_error);
  sprintf(fname,"%.128s%s",tineHomePath,EQPDB_FILE);
  if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
  n = *num;
  for (done=0; csvGetLine(str,256,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((con_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(con_col,str),16);
    for (j=0; j<i && strncmp(scratch,clist[j].name,16); j++);
    if (j < i) continue;
    strncpy(clist[i].name,scratch,16);
    if (++i > n) break;
  }
  *num = i;
  fclose(fp);
  if (cc) ccerr(cc);
err:
  return cc;
}
TINE_EXPORT int GetMyServerAddress(char *eqmName,char *expName,char *ctxName,char *fecName,int *port)
{
  int i,cc=0,pass=0;
  NAME32 n32[5];
  NAME16 n16[5];
  DTYPE dout, din, *dinptr=NULL;

  if (WaitForSystemMutex(hEnsMutex,gSystemTick) != 0) return semaphore_busy;
retryIllegalFormat:
  if (pass == 0)
  {
    if (eqmName != NULL && strlen(eqmName) > 0)
    {
      strncpy(n32[0].name,eqmName,32);
      din.dArrayLength = 1;
      din.dFormat = CF_NAME32;
      din.data.cptr = n32[0].name;
      memset(din.dTag,0,TAG_NAME_SIZE);
      dinptr = &din;
    }
    dout.dArrayLength = 5;
    dout.dFormat = CF_NAME32;
    dout.data.cptr = n32[0].name;
    memset(dout.dTag,0,TAG_NAME_SIZE);
  }
  else
  {
    if (eqmName != NULL && strlen(eqmName) > 0)
    {
      strncpy(n16[0].name,eqmName,16);
      din.dArrayLength = 1;
      din.dFormat = CF_NAME16;
      din.data.cptr = n16[0].name;
      memset(din.dTag,0,TAG_NAME_SIZE);
      dinptr = &din;
    }
    dout.dArrayLength = 5;
    dout.dFormat = CF_NAME16;
    dout.data.cptr = n16[0].name;
    memset(dout.dTag,0,TAG_NAME_SIZE);
  }
  for (i=0,cc=link_timeout; cc != 0 && i <5; i++)
  {
    cc = tExecLink(ENS_SRV,ENS_TAG,"#0","WHOAMI",&dout,dinptr,CA_READ|CA_RETRY,200);
    if (cc == illegal_format && pass == 0)
    {
      pass++;
      goto retryIllegalFormat;
    }
  }
  if (cc == 0)
  {
    if (pass > 0)
    {
      for (i=0; i<5; i++) 
      {
        memcpy(n32[i].name,n16[i].name,16); n32[i].name[16] = 0;
      }
    }
    if (expName != NULL) strncpy(expName,n32[0].name,EXPORT_NAME_SIZE);
    if (fecName != NULL) strncpy(fecName,n32[2].name,FEC_NAME_SIZE);
    if (ctxName != NULL) strncpy(ctxName,n32[3].name,CONTEXT_NAME_SIZE);
    if (port != NULL) *port = atoi(n32[4].name);
  }
  ReleaseSystemMutex(hEnsMutex); 
  return cc;
}
TINE_EXPORT int GetDevicePropertyEGU(char *srv,char *prp,float *max,float *min,char *egu)
{
  DTYPE dout;
  DUSTRING ustr;
  int cc = 0;
  char uprp[36];

  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_USTRING;
  dout.dArrayLength = 1;
  dout.data.vptr = &ustr;
  strncpy(uprp,prp,28); uprp[28] = 0; strcat(uprp,".EGU");
  if ((cc=ExecLinkEx(srv,uprp,&dout,NULL,CA_READ|CA_NOCALLBACKS,QUERY_TO)) != 0) return cc;
  if (max != NULL) *min = ustr.f1val;
  if (min != NULL) *max = ustr.f2val;
  if (egu != NULL) strncpy(egu,ustr.str,80);
  return 0;
}
TINE_EXPORT int GetDevicePropertyInformation(char *srv,PropertyQueryStruct *srvProps,int *num)
{
  DTYPE dout;
  int n, cc = 0;
# ifndef INTEL
  int i;
# endif
  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if ((n % PROPERTYQUERYSTRUCTEX_SIZE) == 0) n++; /* avoid magic numbers */
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n*PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,NULL,CA_READ|CA_NOCALLBACKS|CA_MUTABLE,QUERY_TO)) != 0) return cc;
# ifndef INTEL
  for (i=0; i<n; i++) SSWAP(srvProps[i].prpSize);
# endif
  *num = dout.dArrayLength/PROPERTYQUERYSTRUCT_SIZE;
  return 0;
}
TINE_EXPORT int GetDevicePropertyInformationX(char *srv,char *prp,XPropertyQueryStruct *srvProps,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int cc = 0;
  short n;
  BYTE *data = (BYTE *)srvProps;
  NAME32 n32;

  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if (prp != NULL)
  {
    DTYPEZERO(din);
    strncpy(n32.name,prp,32);
    din.dArrayLength = 1;
    din.dFormat = CF_NAME32;
    din.data.vptr = &n32;
    dinptr = &din;
  }
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n;
  strncpy(dout.dTag,"XPQS",TAG_NAME_SIZE);
  dout.data.vptr = data;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,dinptr,CA_READ|CA_NOCALLBACKS|CA_MUTABLE,QUERY_TO)) != 0) return cc;
  *num = dout.dArrayLength;
  return 0;
}
TINE_EXPORT int GetDeviceServers(NAME16 *dslist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(NULL,"SERVERS",&fmt,dslist,num);
  return cc;
}
TINE_EXPORT int GetDeviceServersEx(char *context,NAME16 *dslist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(context,"SERVERS",&fmt,dslist,num);
  return cc;
}
TINE_EXPORT int GetDeviceServersFromFile(char *context,NAME16 *dslist,int *num)
{
  FILE  *fp=NULL;
  int i=0,n,cc=0,done=0,nam_col=-1,con_col=-1;
  char fname[256],str[256],hdr[256],ctx[CONTEXT_NAME_SIZE+4];
  
  if (num == NULL || dslist == NULL) ccerr(argument_list_error);
  sprintf(fname,"%.128s%s",tineHomePath,EQPDB_FILE);
  if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
  if (context != NULL)
  {
    if (strlen(context) == 0 || stricmp(context,"DEFAULT") == 0)
      context = NULL;
  }
  n = *num;
  for (done=0; csvGetLine(str,256,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((con_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    strncpy(ctx,colptr(con_col,str),16);
    if (context != NULL && strnicmp(ctx,context,CONTEXT_NAME_SIZE)) continue;
    strncpy(dslist[i].name,colptr(nam_col,str),16);
    if (++i > n) break;
  }
  *num = i;
  fclose(fp);
  if (cc) ccerr(cc);
err:
  return cc;
}
TINE_EXPORT int GetServersEx(char *context,NAME16 *slist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(context,"FECS",&fmt,slist,num);
  return cc;
}
TINE_EXPORT int GetServers(NAME16 *slist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(NULL,"FECS",&fmt,slist,num);
  return cc;
}
TINE_EXPORT int GetSystemContexts(NAME32 *clist,int *num)
{
  int fmt = CF_NAME32;
  int cc = GetTagList(NULL,"CONTEXTS",&fmt,clist,num);
  if (fmt != CF_NAME32 && cc == 0) augmentNameSize(CF_NAME16,(char *)clist,*num,1);
  return cc;
}
TINE_EXPORT int GetSystemGroupMembers(char *context,char *group,NAME32 *mlist,int *num)
{
  int n, cc = 0;
  char dev[128];
  DTYPE dout, din;
  if (context == NULL || group == NULL || mlist == NULL || num == NULL) return invalid_parameter;
  DTYPEZERO(dout); DTYPEZERO(din);
  n = *num;
  din.dArrayLength = (UINT32)strlen(context); din.dFormat = CF_TEXT; din.data.cptr = context;
  dout.dArrayLength = n; dout.dFormat = CF_NAME32; dout.data.vptr = mlist;
  sprintf(dev,"/SITE/GENS/%.32s",group);
  cc = ExecLinkEx(dev,"MEMBERS",&dout,&din,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,500);
  *num = dout.dArrayLength;
  return cc;
}
TINE_EXPORT int GetNumberOf(char *srv,char *prp)
{
  DTYPE dout;
  int cc;
  short n;
  char tmpprp[128];

  if (srv == NULL || prp == NULL) return -argument_list_error;
  if (!IsStockProperty(prp)) return -illegal_property;
  if (prp[0] == 'N' || prp[0] == 'n') 
  {
    strncpy(tmpprp,prp,PROPERTY_NAME_SIZE);
  }
  else
  {
    sprintf(tmpprp,"N%.64s",prp);
  }
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_INT16;
  dout.dArrayLength = 1;
  dout.data.vptr = &n;
  cc = ExecLinkEx(srv,tmpprp,&dout,NULL,CA_READ|CA_NOCALLBACKS,QUERY_TO);
  return cc ? -cc : (int)n;
}
TINE_EXPORT int GetSystemDevices(char *srv,char *prp, NAME64 *dlist,int *num)
{
  DTYPE dout;
  int i,cc = 0,fmt = CF_NAME64, prpInpt = FALSE;
  short n;
  ExportListStruct *el = getExportListItem(srv);
  NAME64 *lcldevlst;

  if (srv == NULL || dlist == NULL || num == NULL) return argument_list_error;
  if ((n = *num) <= 0) return buffer_too_small;
  if (prp != NULL && strlen(prp) > 0) prpInpt = TRUE;
  if (el != NULL) /* local eqm name given */
  {
    lcldevlst = GetExportedDeviceList(srv);
    for (i=0; i<el->EqmNumDevices && i<n; i++) 
      strncpy(dlist[i].name,lcldevlst[i].name,64);
    *num = i;
  }
  else
  {
    ExpDataStruct srvaddr;
    char tmpprp[64],tmpsrv[132], *ctx=NULL, *exp, *c;
    int chkMembers=FALSE;
    DTYPEZERO(dout);
    if (prpInpt && strlen(prp) < PROPERTY_NAME_SIZE-4)
    {
      sprintf(tmpprp,"%s.NAM",prp);
    }
    else
    {
      strcpy(tmpprp,"DEVICES");
    }
    strncpy(tmpsrv,srv,128);
    if (tmpsrv[0] == '/') 
    {
      ctx = &tmpsrv[1];
      if ((exp=strchr(ctx,'/')) == NULL) return non_existent_server;
      *exp++ = 0;
    }
    else
    {
      exp = tmpsrv;
    }
    if ((c=strchr(exp,'/')) != NULL) *c = 0;
    if ((cc=GetServerAddr(ctx,exp,&srvaddr,NULL)) != 0) return cc;
    if (!strncmp(srvaddr.EqmName,DOOCSEQM_TAG,EQM_NAME_SHORTSIZE) && prpInpt)
    {
      strcpy(tmpprp,"DEVICES");
      chkMembers = TRUE;
    }
retryFormat:
    dout.dFormat = fmt;
    dout.dArrayLength = n;
    dout.data.vptr = dlist;
    cc = ExecLinkEx(srv,tmpprp,&dout,NULL,CA_READ|CA_CCSHOWALL|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO);
    cc &= ~CE_RENEGOTIATE; /* don't care about this flag */
    if (cc & CE_REDIRECTED)
    { /* redirected by device ? */
      /* TODO: finish this : strcpy(tmpprp,"DEVICES"); */
    }
    if (cc != 0 && !(cc & CE_SENDDATA)) 
    {
      if (cc == illegal_format && fmt == CF_NAME64)
      {
        fmt = CF_NAME32;
        goto retryFormat;
      }
      return cc;
    }
    if (fmt != CF_NAME16 && strlen(dlist[0].name) == 0)
    { /* 'success' but no names : e.g. a release 3.31 java server */
      fmt = CF_NAME16;
      goto retryFormat;
    }
    if (HAS_DATA(cc))
    {
      switch (fmt)
      {
        case CF_NAME64: /* normal release 4 server */
          break;
        case CF_NAME32:
          augmentNameSize(CF_NAME32,(char *)dlist,n,1);
          break;
        case CF_NAME16:
          augmentNameSize(CF_NAME16,(char *)dlist,n,2);
          break;
      }
    }
    if (n == (int)dout.dArrayLength)
    { /* coincidence ? */
      if (dlist[n-1].name[0] == 0)
      { /* return did not truncate the call ! */
        for (i=0; i<n; i++) if (dlist[i].name[0] == 0) break;
        n = i;
      }
    }
    else
    {
      n = dout.dArrayLength;
    }
#   ifndef VXWORKS
    if (chkMembers)
    { /* doocs server ? */
      DUSTRING *unames;
      PrpQueryStruct pqs[1], *ppqs = pqs;
      int np = 1, nd = 0, k;
      char *c;
      strncpy(tmpsrv, srv, 128);
      if ((c=strstr(tmpsrv,"/#0")) != NULL) *c = 0;
      strcat(tmpsrv, "/");
      for (i=0; i<n; i++) if (!strncmp(dlist[i].name,"DEVGRP",6))
      {
        strncat(tmpsrv,dlist[i].name,64);
        if (GetSystemPropertyInformation(tmpsrv,prp,&ppqs,&np) == 0)
        { /* this group device supports the target property */
          nd = pqs[0].prpSize;
          if ((unames=(DUSTRING *)alloca(nd * sizeof(DUSTRING))) == NULL) break;
          dout.dFormat = CF_USTRING;
          dout.dArrayLength = pqs[0].prpSize;
          dout.data.vptr = unames;
          sprintf(tmpprp,"%s.MEMBERS",prp);
          if (ExecLinkEx(tmpsrv,tmpprp,&dout,NULL,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO) != 0) break;
          n = dout.dArrayLength;
          for (k=0; k<n; k++) strncpy(dlist[k].name,unames[k].str,64);
          break;
        }
      }
      if (i == n)
      { /* did not find a device group!  try this: */
        strcat(tmpsrv, "*");
        if ((unames=(DUSTRING *)alloca(n * sizeof(DUSTRING))) != NULL)
        {
          dout.dFormat = CF_USTRING;
          dout.dArrayLength = n;
          dout.data.vptr = unames;
          if (ExecLinkEx(tmpsrv,prp,&dout,NULL,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO) == 0)
          {
            n = dout.dArrayLength;
            for (k=0; k<n; k++) strncpy(dlist[k].name,unames[k].str,64);
          }
        }
      }
    }
#   endif
    *num = n;
  }
  return cc;
}
TINE_EXPORT int GetWildcardDevices(char *srv,int mask, NAME64 *dlist,int *num)
{
  DTYPE dout, din;
  int i,k,n,cc = 0,fmt = CF_NAME64,chkMask = FALSE;
  ExportListStruct *el = getExportListItem(srv);
  NAME64 *lcldevlst;

  if (srv == NULL || dlist == NULL || num == NULL) return argument_list_error;
  if ((n = *num) <= 0) return buffer_too_small;
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  if (mask != 0) 
  {
    din.dArrayLength = 1;
    din.dFormat = CF_INT32;
    din.data.lptr = &mask;
    chkMask = TRUE;
  }
  else
  {
    din.dFormat = CF_NULL;
  }
  if (el != NULL) 
  { /* local eqm name given */
    lcldevlst = GetExportedDeviceList(srv);
    for (i=0,k=0; i<el->EqmNumDevices && i<n; i++)
    {
      if (chkMask && el->EqmDevLst[i].mask > 0 && (el->EqmDevLst[i].mask & mask) == 0) continue;
      strncpy(dlist[k].name,lcldevlst[i].name,64);
      k++;
    }
    *num = k;
  }
  else
  {
    dout.dFormat = fmt;
    dout.dArrayLength = n;
    dout.data.vptr = dlist;
    cc = ExecLinkEx(srv,"DEVICES",&dout,&din,CA_READ|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO);
    if (cc & CE_SENDDATA) cc = 0;
    if (cc != 0) return cc;
    *num = dout.dArrayLength;
  }
  return 0;
}
TINE_EXPORT int GetSystemFecs(char *context,NAME16 *slist,int *num)
{
  int fmt = CF_NAME16;
  int cc = GetTagList(context,"FECS",&fmt,slist,num);
  return cc;
}
TINE_EXPORT int GetStockProperties(NAME64 *plist, int *num)
{
  int i, k, n;
  if (plist == NULL || num == NULL) return argument_list_error;
  if ((n = *num) <= 0) return buffer_too_small;
  for (i = 0, k = 0; i<n && i<gNrStockProperties; i++)
  {
    if (k > 0 && !strncmp(plist[k-1].name, gStockProperty[i].prpName, PROPERTY_NAME_SIZE)) continue;
    strncpy(plist[k].name, gStockProperty[i].prpName, PROPERTY_NAME_SIZE);
    k++;
  }
  *num = k;
  return 0;
}
TINE_EXPORT int GetSystemProperties(char *srv, NAME64 *plist, int *num)
{
  return GetSystemPropertiesEx(srv, plist, num, PRP_REGISTERED);
}
TINE_EXPORT int GetSystemPropertiesEx(char *srv,NAME64 *plist,int *num,int flags)
{
  DTYPE dout;
  int cc = 0, fmt = CF_NAME64;
  short n, nr = 0, i;
  char *tgtprp[] = { "PROPERTIES", "STOCKPROPS", "METAPROPS" };

  if (srv == NULL || plist == NULL || num == NULL) return argument_list_error;
  if ((n = *num) <= 0) return buffer_too_small;
  if (flags == 0) flags = PRP_REGISTERED;
  for (i = 0; i < 3 && n > nr; i++)
  {
    if ((flags & (1<<i)) == 0) continue;
retryFormat:
    memset(&dout, 0, sizeof(DTYPE));
    dout.dFormat = fmt;
    dout.dArrayLength = n - nr;
    dout.data.cptr = (char *)&plist[nr];
    if ((cc = ExecLinkEx(srv, tgtprp[i], &dout, NULL, CA_READ | CA_RETRY | CA_MUTABLE | CA_NOCALLBACKS, QUERY_TO)) != 0 && !(cc & CE_SENDDATA))
    {
      if (cc == illegal_format && fmt == CF_NAME64)
      {
        fmt = CF_NAME32;
        goto retryFormat;
      }
      return cc;
    }
    if (fmt != CF_NAME64 && cc == 0) augmentNameSize(CF_NAME32, (char *)&plist[nr],(int)dout.dArrayLength, 1);
    nr += dout.dArrayLength;
  }
  *num = nr;
  return cc;
}
TINE_EXPORT int GetSystemPropertyInformation(char *srv,char *prp,PrpQueryStruct **pqs,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int i, cc = 0, siz, isMeta=FALSE;
  short n, fmt;
  PrpQueryStruct *prpqs;
  XPropertyQueryStruct *xpqs;
  char *bpstr, *mpstr=NULL, *tprp=prp, *tgtprp="PROPERTIES";
  NAME64 n64;

  if (srv == NULL || num == NULL || pqs == NULL) return invalid_parameter;
  DTYPEZERO(dout); DTYPEZERO(din);
  if ((n = *num) <= 0)
  { /* caller wants me to find out how many */
    if (prp == NULL)
    { /* general property info query */
      dout.dFormat = CF_INT16;
      dout.dArrayLength = 1;
      dout.data.vptr = &n;
      if ((cc=ExecLinkEx(srv,"NPROPERTIES",&dout,dinptr,CA_READ|CA_RETRY|CA_NOCALLBACKS,QUERY_TO)) != 0) return cc;
    }
    else
    { /* targeted: just get the overloads */
      n = 10; /* a 'nice' query size */
    }
  }
  if ((prpqs = *pqs) == NULL)
  { /* caller wants me to allocate this */
    if ((prpqs=(PrpQueryStruct *)SystemCalloc(n,sizeof(PrpQueryStruct))) == NULL) 
      return out_of_local_memory;
  }
  if (prp != NULL)
  {
    if (IsStockProperty(prp)) 
    {
      int k;
      for (i=0,k=0; i<gNrStockProperties && k<n; i++)
      {
        if (strncmp(gStockProperty[i].prpName,prp,PROPERTY_NAME_SIZE)) continue;
        memset(&prpqs[k],0,sizeof(PrpQueryStruct));
        strncpy(prpqs[k].prpName,gStockProperty[i].prpName,PROPERTY_NAME_SIZE);
        strncpy(prpqs[k].prpDescription,gStockProperty[i].prpDescription,64);
        prpqs[k].prpSize = gStockProperty[i].prpSize;
        prpqs[k].prpSizeIn = gStockProperty[i].prpSizeIn;
        prpqs[k].prpFormat = BFMT(gStockProperty[i].prpFormat);
        prpqs[k].prpFormatIn = BFMT(gStockProperty[i].prpFormatIn);
        prpqs[k].prpAccess = (BYTE)gStockProperty[i].prpAccessMode;
        prpqs[k].prpArrayType = gStockProperty[i].prpArrayType;
        if (isTaggableFormat(prpqs[k].prpFormat)) strncpy(prpqs[k].prpTag,gStockProperty[i].prpStructTag,TAG_NAME_SIZE);
        if (isTaggableFormat(prpqs[k].prpFormatIn)) strncpy(prpqs[k].prpTagIn,gStockProperty[i].prpStructTag,TAG_NAME_SIZE);
        k++;
      }
      if (!strcmp(prp,"DEVICES") || !strcmp(prp,"PROPERTIES") || !strcmp(prp,"PROPS"))
      { /* may as well get the 'real' number */
        if ((n=GetNumberOf(srv,prp)) > 0) prpqs->prpSize = n;
      }
      *pqs = prpqs;
      *num = k;
      ccerr(0);
    }
    if (IsMetaProperty(prp,&bpstr,&mpstr,FALSE))
    {
      isMeta = TRUE;
      tprp = bpstr;
      n = 1;
    }
    strncpy(n64.name,tprp,64);
    din.dArrayLength = 1;
    din.dFormat = CF_NAME64;
    din.data.vptr = &n64;
    dinptr = &din;
  }
pinfocall:
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n;
  strncpy(dout.dTag,"PRPQSr4",TAG_NAME_SIZE);
  dout.data.vptr = (void *)prpqs;
  cc = ExecLinkEx(srv,tgtprp,&dout,dinptr,CA_READ|CA_RETRY|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO);
  if (HAS_DATA(cc))
  { /* modern server understood the query */
    *num = dout.dArrayLength;
    *pqs = prpqs;
    ccerr(0);
  }
  if (cc == non_existent_property && isMeta)
  { /* maybe not a meta-property call after all ? */
    isMeta = FALSE;
    strncpy(n64.name,prp,64);
    din.data.vptr = &n64;
    din.dArrayLength = 1;
    din.dFormat = CF_NAME64;
    goto pinfocall;
  }
  if (cc == illegal_format || cc == invalid_structure_tag)
  { /* legacy server ? */
    if ((xpqs=(XPropertyQueryStruct *)SystemCalloc(n,sizeof(XPropertyQueryStruct))) == NULL) 
      ccerr(out_of_local_memory);
    *num = n;
    cc = GetDevicePropertyInformationX(srv,prp,xpqs,num);
    if (HAS_DATA(cc))
    {
      for (i=0; i<n; i++) xpqs2prpqs(i);
      SystemFree(xpqs);
      *pqs = prpqs;
      ccerr(0);
    }
  }
err:
  if (cc == 0 && isMeta)
  {
    switch (getMetaFormatAndSize(mpstr,&fmt,&siz))
    {
      case ambiguous:
        break;
      case illegal_format: /* returned size is okay */
        prpqs->prpSize = (UINT32)siz;
        break;
      case illegal_data_size: /* returned format is okay */
        prpqs->prpFormat = BFMT(fmt);
        if (prpqs->prpArrayType & AT_CHANNEL)
        {
          prpqs->prpSize = (UINT32)siz;
        }
        else
        {
          n = GetNumberOf(srv,"DEVICES");
          prpqs->prpSize = n > 0 ? n : 100;
        }
        break;
      default:
        prpqs->prpSize = (UINT32)siz;
        prpqs->prpFormat = BFMT(fmt);
        break;
    }
  }
  return cc;
}
TINE_EXPORT int GetSystemServers(char *context,NAME32 *dslist,int *num)
{
  int fmt = CF_NAME32;
  int cc = GetTagList(context,"SERVERS",&fmt,dslist,num);
  if (fmt != CF_NAME32 && cc == 0) augmentNameSize(CF_NAME16,(char *)dslist,*num,1);
  return cc;
}
TINE_EXPORT int GetTargetPropertyInformation(char *srv,char *prp,int *fmt,int *siz,char *dsc)
{
  int cc, n = 1;
  PrpQueryStruct pqs[1], *ppqs = pqs;

  cc = GetSystemPropertyInformation(srv,prp,&ppqs,&n);
  if (!HAS_DATA(cc)) return cc;
  if (fmt != NULL) *fmt = LFMT(pqs[0].prpFormat);
  if (siz != NULL) *siz = pqs[0].prpSize;
  if (dsc != NULL) strncpy(dsc,pqs[0].prpDescription,PROPERTY_DESC_SIZE);
  return 0;
}

FILE *openSettingsFile(const char *eqm,const char *prp,char *attr,char *filename)
{
  int cc=0,chkdir=FALSE;
  char fn[512],dn[256]="";
  FILE *fp=NULL;
  if (attr == NULL || prp == NULL) ccerr(argument_list_error);
# ifdef FS_DELIMITER
  if (eqm != NULL)
  { 
    sprintf(dn,"%.128s%.6s%c",fecDbPath,eqm,FS_DELIMITER);
    sprintf(fn,"%.192s%.64s-settings.csv",dn,prp);
    chkdir = TRUE;
    goto fnDone;
  }
# endif
  if (ignoreCommonDbFiles) ccerr(no_such_file);
  sprintf(fn,"%.128s%.64s-settings.csv",fecDbPath,prp);
fnDone:
  fp = fopen(fn,attr);
# ifndef SMALL_TINE_LIB
  if (fp == NULL && chkdir && strchr(attr,'w') != NULL)
  {
    chkdir = FALSE;
    if (makeDirectoryPath(dn,TRUE)) goto fnDone;
  }
# endif
  if (fp == NULL)
  {
    if (!IsStockProperty((char *)prp))
    {
      rprtSckErr("openSettingsFile", fn, 0, errno, 0, NULL);
    }
    ccerr(file_error);
  }
  if (filename != NULL) strncpy(filename,fn,256);
err:
  if (cc)
  {
    if (fp != NULL) fclose(fp);
    fp = NULL;
  }
  return fp;
}
time_t lastSaveRestoreFileTime = 0;
TINE_EXPORT int RestorePropertyValues(const char *eqmName, const char *prpName, void *values, short format, int size)
{
  return _RestorePropertyValues(eqmName, prpName, values, format, size, FALSE);
}
int _RestorePropertyValues(const char *eqmName,const char *prpName,void *values,short format,int size,int internal)
{
  int devnr=0,n=0,cc=0,done=0,val_col=-1,dev_col=-1,sizFmt=fmtsizeof(format),sizItem=1,sizHdr=0,hFile;
  FILE *fp = NULL;
  char *c, scratch[256], str[512], fn[256];
  ExportPropertyListStruct *pls=NULL;
  ExportListStruct *el = getExportListItem((char *)eqmName);
# ifndef VXWORKS
  struct stat sbuf;
# endif
  if (el == NULL) return non_existent_elem;
  if (WaitForSystemMutex(hDbMutex,-1) != 0) return semaphore_error;
  if (prpName == NULL || values == NULL || size <= 0 || sizFmt <= 0) ccerr(argument_list_error);
  if ((fp=openSettingsFile(eqmName,prpName,"r",fn)) == NULL) ccerr(no_such_file);
  if ((pls=GetPropertyListStructEx((char *)eqmName,(char *)prpName,NULL,CA_SAVERESTORE)) == NULL) ccerr(illegal_property);
  switch (format)
  {
    case CF_STRUCT: 
      sizItem = GetStructSize(pls->prpFormatTag.name);
      break;
    case CF_GSPECTRUM:
    case CF_SPECTRUM:
    case CF_IMAGE:
      sizHdr = getFormatHeaderSize(format);
      sizItem = pls->prpSizeIn + sizHdr/sizFmt; /* effective num elements for below */
      break;
    case CF_TEXT:
      memset(values, 0, size*sizFmt); /* zero the full sr buffer */
      size = el->EqmNumDevices;
      sizItem = pls->prpSizeIn;
      break;
    default: 
      sizItem = isAllowedSaveRestoreTraceArray(format) ? pls->prpSizeIn : 1; 
      break;
  }
  if (format != CF_STRING) 
  { /* do not zero out the string pointers ! */
    memset(values,0,size*sizFmt);
  }
  scratch[64] = 0;
  while (csvGetLine(str,256,fp) != NULL)
  {
    if (n >= size) break;
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carriage return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      val_col = findcol(str,"VALUES");
      dev_col = findcol(str,"DEVICES");
      done = TRUE;
      continue;
    }
    devnr = n;
    if (dev_col >= 0)
    {
      getColumnStr(dev_col,str,scratch,64);
      devnr = GetDeviceNumberEx((char *)eqmName,scratch,(char *)prpName);
      if (devnr < 0) devnr = n; else n = devnr;
      if (devnr >= size) ccerr(illegal_device);
    }
    if (val_col >= 0) 
    { /* values (or file reference) given */
      getColumnStr(val_col,str,scratch,256); 
      if (!stricmp(scratch,"<$FILE"))
      { /* prepare string for stringToDataValue (fn must end in '.csv') */
        if ((c=strrchr(fn,'.')) != NULL && !strcmp(c,".csv")) *c = 0;
        sprintf(scratch,"<%.240s.%d",fn,devnr);
      }
    }
    else 
    { /* default value */
      strcpy(scratch,"0");
    }
    if ((cc=stringToDataValue((BYTE *)values,scratch,devnr*sizItem,format,NULL,sizItem,internal)) != 0) 
    { /* could be a missing file for array properties ... */
      int num = isAllowedSaveRestoreTraceArray(format) ? size : size * sizItem;
      cc = acqDefPrpValsFromEqm((char *)eqmName,(char *)prpName,(BYTE *)values,num,devnr);
    }
    n++;
  }
  if (n == 0) ccerr(string_expected);
  hFile = fileno(fp);
# ifndef VXWORKS
  fstat(hFile,&sbuf);
  lastSaveRestoreFileTime = sbuf.st_mtime;
# else
  lastSaveRestoreFileTime = time(NULL);
# endif
err:
  if (fp != NULL) fclose(fp);
  ReleaseSystemMutex(hDbMutex);
  if (cc != 0 && cc != no_such_file)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"cannot restore property settings for %.6s %.64s : %.32s",eqmName,prpName,cc2str(cc));
  }
  return cc;
}
int dataValuesToFile(char *filename,BYTE *data,int nbytes)
{
  FILE *fp = NULL;
  int cc = 0;
# if defined(WIN32) || defined(DOS)
  char *fmod="w+b";
# else
  char *fmod="w";
# endif
  if ((fp=fopen(filename,fmod)) == NULL) ccerr(file_error);
  if (fwrite(data,nbytes,1,fp) != 1) ccerr(file_error);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int RestoreStockPropertyValues(ExportListStruct *el,char *prpName)
{
  int id,i=0,cc=0,done=0,val_col=-1,dev_col=-1,off=0,lim=1,deref=FALSE;
  short fmt;
  FILE *fp = NULL;
  char scratch[256], str[512], fn[256];
  ExportDeviceType *devLst;
  BYTE *dptr;

  if (el == NULL || prpName == NULL) return invalid_parameter;
  if ((id=GetStockPropertyId(prpName)) < 0) return invalid_property;
  if (WaitForSystemMutex(hDbMutex,-1) != 0) return semaphore_error;
  if (!strncmp(prpName,"ZPOSITION",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat;
    off = (int)OFFSETIN(ExportDeviceType,zpos);
  }
  else if (!strncmp(prpName,"DEVMASK",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat;
    off = (int)OFFSETIN(ExportDeviceType,mask);
  }
  else if (!strncmp(prpName,"DEVREGION",PROPERTY_NAME_SIZE))
  {
    fmt = CF_INT32;
    off = (int)OFFSETIN(ExportDeviceType,region);
  }
  else if (!strncmp(prpName,"DEVLOCATION",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat; /* CF_NAME64 */
    off = (int)OFFSETIN(ExportDeviceType,loc);
    deref = TRUE;
  }
  else if (!strncmp(prpName,"DEVDESCRIPTION",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat; /* CF_NAME64 */
    off = (int)OFFSETIN(ExportDeviceType,desc);
    deref = TRUE;
  }
  else ccerr(not_implemented);
  if ((fp=openSettingsFile(el->EqmName,prpName,"r",fn)) == NULL) ccerr(no_such_file);
  scratch[64] = 0;
  if ((devLst=el->EqmDevLst) == NULL) ccerr(not_initialized);
  while (csvGetLine(str,256,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carraige return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      if ((val_col=findcol(str,"VALUES")) < 0) ccerr(no_such_column);
      if ((dev_col=findcol(str,"DEVICES")) < 0) ccerr(no_such_column);
      done = TRUE;
      continue;
    }
    getColumnStr(dev_col,str,scratch,64);
    if ((i=GetDeviceNumber((char *)el->EqmName,scratch)) < 0) continue;
    if (i >= el->EqmNumDevices) ccerr(illegal_device);
    getColumnStr(val_col,str,scratch,256);
    dptr = (BYTE *)devLst + i*sizeof(ExportDeviceType) + off;
    if (deref) 
    { /* the above list location pointer points to a pointer */
      BYTE *pdptr = (BYTE *)(*(size_t *)dptr);
      if (pdptr == NULL) 
      {
        if ((pdptr=(BYTE *)SystemCalloc(1,fmtsizeof(fmt))) == NULL) ccerr(out_of_server_memory);
        memcpy(dptr,&pdptr,sizeof(size_t));
        pdptr = (BYTE *)(*(size_t *)dptr);
      }
      dptr = pdptr;
    }
    stringToDataValue(dptr,scratch,0,fmt,NULL,lim,TRUE);
  }
err:
  if (fp != NULL) fclose(fp);
  ReleaseSystemMutex(hDbMutex);
  if (cc != 0 && cc != no_such_file)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot restore property settings for %.6s %.64s : %.32s",el->EqmName,prpName,cc2str(cc));
  }
  return cc;
}
int SaveStockPropertyValues(ExportListStruct *el,char *prpName)
{
  int cc=0,i,id,off=0,lim=1,deref=FALSE;
  short fmt;
  FILE *fp=NULL;
  char s[DBGBUF_SIZE],fn[256];
  ExportDeviceType *devLst;
  BYTE *dptr;
  if (el == NULL || prpName == NULL) return invalid_parameter;
  if ((id=GetStockPropertyId(prpName)) < 0) return invalid_property;
  if (WaitForSystemMutex(hDbMutex,-1) != 0) return semaphore_error;
  if (!strncmp(prpName,"ZPOSITION",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat;
    off = (int)OFFSETIN(ExportDeviceType,zpos);
  }
  else if (!strncmp(prpName,"DEVMASK",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat;
    off = (int)OFFSETIN(ExportDeviceType,mask);
  }
  else if (!strncmp(prpName,"DEVREGION",PROPERTY_NAME_SIZE))
  {
    fmt = CF_INT32;
    off = (int)OFFSETIN(ExportDeviceType,region);
  }
  else if (!strncmp(prpName,"DEVLOCATION",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat;
    off = (int)OFFSETIN(ExportDeviceType,loc);
    deref = TRUE;
  }
  else if (!strncmp(prpName,"DEVDESCRIPTION",PROPERTY_NAME_SIZE))
  {
    fmt = gStockProperty[id].prpFormat;
    off = (int)OFFSETIN(ExportDeviceType,desc);
    deref = TRUE;
  }
  else ccerr(not_implemented);
  if ((fp=openSettingsFile(el->EqmName,prpName,"w",fn)) == NULL) ccerr(no_such_file);
  devLst = el->EqmDevLst;
  fprintf(fp,"DEVICES, VALUES\n");
  for (i=0; i<el->EqmNumDevices; i++)
  {
    dptr = (BYTE *)devLst + i*sizeof(ExportDeviceType) + off;
    if (deref) dptr = (BYTE *)(*(size_t *)dptr);
    if (dataValueToString(s,dptr,0,fmt,NULL,lim,TRUE) == NULL) ccerr(illegal_format);
    fprintf(fp,"%.64s, %.255s\n",devLst[i].name,s);
  }
err:
  if (fp != NULL) fclose(fp);
  ReleaseSystemMutex(hDbMutex);
  if (cc != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot save property settings for %.6s %.64s : %.32s",el->EqmName,prpName,cc2str(cc));
  }
  return cc;
}
TINE_EXPORT int SavePropertyValuesEx(const char *eqmName, const char *devName, const char *prpName, void *values, short format, int size)
{
  return _SavePropertyValuesEx(eqmName, devName, prpName, values, format, size, FALSE);
}
int _SavePropertyValuesEx(const char *eqmName,const char *devName,const char *prpName,void *values,short format,int size,int internal)
{
  int i, devnr, cc=0, sizEx=1, sizHdr=0, fsiz, limit=DBGBUF_SIZE-1;
  int isSrRegistered=FALSE, devColumn=FALSE, uninitialized=FALSE, hasSrFile=FALSE, srBaseFileNameDone=FALSE;
  FILE *fp=NULL;
  char *dev=NULL;
  ExportListStruct *el=getExportListItem((char *)eqmName);
  ExportPropertyListStruct *pl=GetPropertyListStructEx((char *)eqmName,(char *)prpName,(char *)devName,CA_SAVERESTORE);
  srLst *sr=getSaveRestoreItem(el,pl);
  char *c,s[DBGBUF_SIZE],fn[256];

  if (WaitForSystemMutex(hDbMutex,-1) != 0) return semaphore_error;
  if (pl == NULL && GetPropertyId((char *)eqmName,(char *)prpName) >= 0) ccerr(illegal_read_write);
  if (el == NULL || pl == NULL || values == NULL || size <= 0) ccerr(argument_list_error);
  if (pl->srBuf != NULL) isSrRegistered = TRUE;
  if ((fp=openSettingsFile(eqmName,pl->prpName,"w",fn)) == NULL) ccerr(file_error);
  sizHdr = getFormatHeaderSize(pl->prpFormat);
  hasSrFile = (isSrArray(pl) || pl->prpFormat == CF_STRUCT || sizHdr > 0);
  devnr = GetDeviceNumberEx((char *)eqmName,(char *)devName,(char *)prpName);
  if (devnr < 0) devnr = 0;
  if (isSrRegistered) 
  { /* then I must have met the criteria */
    devColumn = TRUE;
    if (!hasSrFile && (void *)pl->srBuf != values)
    { /* called as an API call with different buffer ! */
      void *ptr = getDataPtrAtIndex(pl,NULL,0,devnr);
      if (ptr != NULL)
      {
        memcpy(ptr,values,size*fmtsizeof(pl->prpFormat));
        values = pl->srBuf; /* remap this for below ... */
      }
    }
  }
  else
  { /* called as an API call */
    if (size == (int)pl->prpSize && devName != NULL)
    { /* specific request */
      if (!stricmp((char *)devName,SR_ADD_DEVICES) || hasSrFile)
        devColumn = TRUE;
    }
  }
  if (hasSrFile || devColumn)
  { /* get the size right for the loop below */
    size = el->EqmNumDevices;
    if (pl->numDevices > 0) size = pl->numDevices;
  }
  if (sr != NULL) uninitialized = sr->uninitialized;
  fprintf(fp,devColumn ? "DEVICES, VALUES\n" : "VALUES\n");
  for (i=0; i<size; i++)
  {
    if (devColumn)
    { /* registered as a save-restore property */
      if ((dev=getDevNameAtIndex(el,pl,i)) == NULL) continue;
      if (format == CF_TEXT) 
      { /* first check TEXT (hasSrFile shows a false positive) */
        sizEx = pl->prpSizeIn;
        limit = MIN(sizEx,DBGBUF_SIZE-1);
      }
      else if (hasSrFile)
      { /* allowed array save-and-restore */
        if (i == devnr || uninitialized)
        { /* the targeted device */
          if (!srBaseFileNameDone)
          { /* some first-time thru the loop stuff : */
            fsiz = fmtsizeof(pl->prpFormatIn);
            switch (pl->prpFormat)
            {
              case CF_STRUCT: 
                sizEx = GetStructSize(pl->prpFormatTag.name); 
                break;
              case CF_GSPECTRUM:
              case CF_SPECTRUM:
              case CF_IMAGE: 
                sizEx = (int)pl->prpSizeIn + sizHdr / fsiz; /* effective num of elements */
                break;
              case CF_BOOLEAN:
                if (!internal) marshallBoolean(values, pl->prpSize, LTOR);
              default:
                sizEx = (int)pl->prpSizeIn; 
                break;
            }
            limit = sizEx;
            fsiz = fmtsizeof(pl->prpFormatIn);
            if ((c=strrchr(fn, '.')) != NULL) *c = 0;
            srBaseFileNameDone = TRUE;
          }
          sprintf(s,"%.240s.%d",fn,i);
          if ((cc=dataValuesToFile(s,&((BYTE *)values)[i*sizEx*fsiz],limit*fsiz)) != 0) ccerr(cc);
        }
        fprintf(fp,"%.64s, <$FILE\n",dev);
        continue;
      }
    }
    if (dataValueToString(s,(BYTE *)values,i*sizEx,format,NULL,limit,internal) == NULL) ccerr(illegal_format);
    if (devColumn) fprintf(fp,"%.64s, %.255s\n",dev,s); else fprintf(fp,"%.255s\n",s);
  }
  if (uninitialized) sr->uninitialized = FALSE;
err:
  if (fp != NULL) fclose(fp);
  ReleaseSystemMutex(hDbMutex);
  if (cc != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot save property settings for %.6s %.64s : %.32s",eqmName,prpName,cc2str(cc));
  }
  return cc;
}
TINE_EXPORT int SavePropertyValues(const char *eqmName,const char *prpName,void *values,short format,int size)
{
  return _SavePropertyValuesEx(eqmName,NULL,prpName,values,format,size,FALSE);
}
TINE_EXPORT int AssertRangeValid(const char *eqmName,const char *prpName,DTYPE *din,int enforceLimits)
{  /* any error returns 'TRUE' */
  ExportPropertyListStruct prp;
  if (eqmName == NULL || prpName == NULL || din == NULL) return TRUE;
  if (GetRegisteredPropertyListStruct((char *)eqmName,(char *)prpName,&prp) != 0) return TRUE;
  if (din->dArrayLength != 1) return TRUE;
  switch (din->dFormat)
  {
    case CF_BYTE:
      if (din->data.bptr[0] < (BYTE)prp.egu.min) 
      {
        if (enforceLimits) din->data.bptr[0] = (BYTE)prp.egu.min; else return FALSE;
      }
      if (din->data.bptr[0] > (BYTE)prp.egu.max) 
      {
        if (enforceLimits) din->data.bptr[0] = (BYTE)prp.egu.max; else return FALSE;
      }
      break;
    case CF_INT16:
      if (din->data.sptr[0] < (SINT16)prp.egu.min) 
      {
        if (enforceLimits) din->data.sptr[0] = (SINT16)prp.egu.min; else return FALSE;
      }
      if (din->data.sptr[0] > (SINT16)prp.egu.max) 
      {
        if (enforceLimits) din->data.sptr[0] = (SINT16)prp.egu.max; else return FALSE;
      }
      break;
    case CF_UINT16:
      if (din->data.usptr[0] < (UINT16)prp.egu.min)
      {
        if (enforceLimits) din->data.usptr[0] = (UINT16)prp.egu.min; else return FALSE;
      }
      if (din->data.usptr[0] > (UINT16)prp.egu.max)
      {
        if (enforceLimits) din->data.usptr[0] = (UINT16)prp.egu.max; else return FALSE;
      }
      break;
    case CF_INT32:
      if (din->data.lptr[0] < (SINT32)prp.egu.min) 
      {
        if (enforceLimits) din->data.lptr[0] = (SINT32)prp.egu.min; else return FALSE;
      }
      if (din->data.lptr[0] > (SINT32)prp.egu.max) 
      {
        if (enforceLimits) din->data.lptr[0] = (SINT32)prp.egu.max; else return FALSE;
      }
      break;
    case CF_UINT32:
      if (din->data.ulptr[0] < (UINT32)prp.egu.min)
      {
        if (enforceLimits) din->data.ulptr[0] = (UINT32)prp.egu.min; else return FALSE;
      }
      if (din->data.ulptr[0] > (UINT32)prp.egu.max)
      {
        if (enforceLimits) din->data.ulptr[0] = (UINT32)prp.egu.max; else return FALSE;
      }
      break;
#   ifdef SINT64
    case CF_INT64:
      if (din->data.llptr[0] < (SINT64)prp.egu.min) 
      {
        if (enforceLimits) din->data.llptr[0] = (SINT64)prp.egu.min; else return FALSE;
      }
      if (din->data.llptr[0] > (SINT64)prp.egu.max) 
      {
        if (enforceLimits) din->data.llptr[0] = (SINT64)prp.egu.max; else return FALSE;
      }
      break;
    case CF_UINT64:
      if (din->data.ullptr[0] < (UINT64)prp.egu.min)
      {
        if (enforceLimits) din->data.ullptr[0] = (UINT64)prp.egu.min; else return FALSE;
      }
      if (din->data.ullptr[0] > (UINT64)prp.egu.max)
      {
        if (enforceLimits) din->data.ullptr[0] = (UINT64)prp.egu.max; else return FALSE;
      }
      break;
#   endif
    case CF_FLOAT:
      if (din->data.fptr[0] < prp.egu.min) 
      {
        if (enforceLimits) din->data.fptr[0] = prp.egu.min; else return FALSE;
      }
      if (din->data.fptr[0] > prp.egu.max) 
      {
        if (enforceLimits) din->data.fptr[0] = prp.egu.max; else return FALSE;
      }
      break;
    case CF_DOUBLE:
      if (din->data.dptr[0] < (double)prp.egu.min) 
      {
        if (enforceLimits) din->data.dptr[0] = (double)prp.egu.min; else return FALSE;
      }
      if (din->data.dptr[0] > (double)prp.egu.max) 
      {
        if (enforceLimits) din->data.dptr[0] = (double)prp.egu.max; else return FALSE;
      }
      break;
    default:
      break;
  }
  return TRUE;
}
TINE_EXPORT int GetDeviceProperties(char *srv,NAME32 *props,int *num)
{
  DTYPE dout;
  int cc = 0;
  short n;

  if ((n = *num) <= 0) return buffer_too_small;
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_NAME32;
  dout.dArrayLength = n;
  dout.data.cptr = (char *)props;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,NULL,CA_READ|CA_RETRY|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO)) != 0 && !(cc & CE_SENDDATA)) return cc;
  *num = dout.dArrayLength;
  return 0;
}
TINE_EXPORT int GetDeviceNamesEx(char *srv,char *prp, NAME16 *devs,int *num)
{
  DTYPE dout;
  int i,cc = 0;
  short n;
  char tmpprp[64];
  ExportListStruct *el = getExportListItem(srv);
  NAME64 *lcldevlst;

  if ((n = *num) <= 0) return buffer_too_small;
  if (prp != NULL && strlen(prp) < 28)
  {
    sprintf(tmpprp,"%s.NAM",prp);
  }
  else
  {
    strcpy(tmpprp,"DEVICES");
  }
  if (el != NULL) /* local eqm name given */
  {
    lcldevlst = GetExportedDeviceList(srv);
    for (i=0; i<el->EqmNumDevices && i<n; i++) 
      strncpy(devs[i].name,lcldevlst[i].name,16);
    *num = el->EqmNumDevices;
  }
  else
  {
    memset(&dout,0,sizeof(DTYPE));
    dout.dFormat = CF_NAME16;
    dout.dArrayLength = n;
    dout.data.vptr = devs;
    if ((cc=ExecLinkEx(srv,tmpprp,&dout,NULL,CA_READ|CA_RETRY|CA_MUTABLE|CA_NOCALLBACKS,QUERY_TO)) != 0 && !(cc & CE_SENDDATA)) return cc;
    *num = dout.dArrayLength;
  }
  return 0;
}
TINE_EXPORT int GetDeviceNames(char *srv,NAME16 *devs,int *num)
{
  return GetDeviceNamesEx(srv,NULL,devs,num);
}
char *valuesAsStringSeparator = "\n";
TINE_EXPORT int SetValuesAsStringSeparator(char* str)
{
	if (str == NULL || strlen(str) == 0 || strlen(str) > 2) 
		return invalid_parameter;
	valuesAsStringSeparator = str;
	return 0;
}
TINE_EXPORT int GetValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  return GetValuesArrayAsString(d,str,dsiz,NULL);
}

#define NumbersToString(CF,FS,FSS,v) \
for (i = 0; i<n; i++) {\
  memcpy(&v, d->data.bptr + i*sizeof(CF), sizeof(CF));\
  if (valueOnly) { sprintf(str, FS, (CF)v); goto err; }\
  sprintf(&str[len], FSS, (CF)v, sep(i));\
  if ((len = (int)strlen(str)) > siz - 16) break;\
}
#define NamesToString(CF,FS,FSS) \
  for (i=0; i<n; i++) {\
    if (valueOnly) { sprintf(str, FS, ((CF *)d->data.vptr)[i].name); goto err; }\
    sprintf(&str[len], FSS, ((CF *)d->data.vptr)[i].name, sep(i));\
    if ((len = (int)strlen(str)) > siz - fmtsizeof(d->dFormat) - 16) break;\
  }
TINE_EXPORT int GetValuesArrayAsString(DTYPE *d, char *str, UINT32 *dsiz, char *separator)
{
  return _GetValuesArrayAsString(d, str, dsiz, separator, FALSE);
}
int _GetValuesArrayAsString(DTYPE *d,char *str,UINT32 *dsiz,char *separator,int internal)
{
# define sep(i) ((i)<(n-1) ? separator == NULL ? valuesAsStringSeparator : separator : "\n" )
  char *ssep;
  int i,n,len,siz,cc=0;
  int valueOnly = FALSE;
  time_t t;
  SINT16 sval;
  SINT32 ival;
  float fval;
  double dval;
# ifdef SINT64
  SINT64 llval;
# endif

  if (d == NULL || d->data.vptr == NULL) return argument_list_error;
  if (str == NULL || dsiz == NULL) return un_allocated;
  if ((siz = (int)(*dsiz)) < 16) return dimension_error;
  if (siz <= 64) valueOnly = TRUE;

  if (separator != NULL) valueOnly = FALSE;
  str[0] = 0; len = 0;
  n = (int)d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_KEYVALUE:
    case CF_STRING:
      for (i=0; i<n; i++) 
      {
        if (((char **)d->data.vptr)[i] == NULL) return un_allocated;
        if (valueOnly) { snprintf(str,siz,"%s",((char **)d->data.vptr)[i]); goto err; }
        snprintf(&str[len],siz,"%s%s",((char **)d->data.vptr)[i],sep(i));
        if ((len=(int)strlen(str)) > siz - 32) break;
      }
      ccerr(0);
    case CF_TEXT:
      len = MIN(siz,n);
      strncpy(str,d->data.cptr,len);
      if (len < siz) str[len] = 0;
      n = (int)strlen(str);
      if (!valueOnly && n < siz) str[n++] = '\n'; 
      if (n < siz) str[n] = 0;
      ccerr(0);
    case CF_BYTE:   
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%x",d->data.bptr[i]); goto err; }
        sprintf(&str[len],"%x%s",d->data.bptr[i],sep(i));
        if ((len=(int)strlen(str)) > siz - 16) break;
      }
      ccerr(0);
    case CF_INT16: 
      NumbersToString(SINT16, "%d","%d%s",sval);
      ccerr(0);
    case CF_UINT16:
      NumbersToString(UINT16, "%d", "%d%s", sval);
      ccerr(0);
    case CF_INT32:
#     ifdef OS_16BIT
      NumbersToString(SINT32, "%ld", "%ld%s", ival);
#     else
      NumbersToString(SINT32, "%d", "%d%s", ival);
#     endif
      ccerr(0);
    case CF_UINT32:
#     ifdef OS_16BIT
      NumbersToString(UINT32, "%ld", "%ld%s", ival);
#     else
      NumbersToString(UINT32, "%d", "%d%s", ival);
#     endif
      ccerr(0);
    case CF_BOOLEAN:
      for (i=0; i<n; i++) 
      {
        if (internal || !gMarshallBooleanAsByte)
        {
          memcpy(&ival, d->data.bptr + i * sizeof(SINT32), sizeof(SINT32));
        }
        else
        {
          ival = (int)d->data.bptr[i];
        }
        if (valueOnly) { sprintf(str,"%s",(ival ? "TRUE" : "FALSE")); goto err; }
        sprintf(&str[len],"%s%s",(ival ? "TRUE" : "FALSE"),sep(i));
        if ((len=(int)strlen(str)) > siz - 16) break;
      }
      ccerr(0);
#   ifdef SINT64
    case CF_INT64:
#     ifdef WIN32
      NumbersToString(SINT64, "%I64d", "%I64d%s", llval);
#     elif defined(OS_64BIT)
      NumbersToString(SINT64, "%ld", "%ld%s", llval);
#     else
      NumbersToString(SINT64, "%lld", "%lld%s", llval);
#     endif
      ccerr(0);
    case CF_UINT64:
#     ifdef WIN32
      NumbersToString(UINT64, "%I64d", "%I64d%s", llval);
#     elif defined(OS_64BIT)
      NumbersToString(UINT64, "%ld", "%ld%s", llval);
#     else
      NumbersToString(UINT64, "%lld", "%lld%s", llval);
#     endif
      ccerr(0);
#   endif
    case CF_FLOAT:
      NumbersToString(float, "%g", "%g%s", fval);
      ccerr(0);
    case CF_DOUBLE:
      NumbersToString(double, "%g", "%g%s", dval);
      ccerr(0);
    case CF_NAME8:
      NamesToString(NAME8, "%.8s", "%.8s%s");
      ccerr(0);
    case CF_NAME16:
      NamesToString(NAME16, "%.16s", "%.16s%s");
      ccerr(0);
    case CF_NAME32:
      NamesToString(NAME32, "%.32s", "%.32s%s");
      ccerr(0);
    case CF_NAME48:
      NamesToString(NAME48, "%.48s", "%.48s%s");
      ccerr(0);
    case CF_NAME64:
      NamesToString(NAME64, "%.64s", "%.64s%s");
      ccerr(0);
    case CF_NAME16I:
    {
      NAME16I nv;
      for (i=0; i<n; i++) 
      {
        memcpy(&nv,d->data.bptr+i*sizeof(NAME16I),sizeof(NAME16I));
        if (valueOnly) { sprintf(str,"%.16s",nv.name); goto err; }
        sprintf(&str[len],"%.16s %d%s",nv.name,nv.ival,sep(i));
        if ((len=(int)strlen(str)) > siz - 64) break;
      }
      ccerr(0);
    }
    case CF_NAME32I:
    {
      NAME32I nv;
      for (i=0; i<n; i++) 
      {
        memcpy(&nv,d->data.bptr+i*sizeof(NAME32I),sizeof(NAME32I));
        if (valueOnly) { sprintf(str,"%.32s",nv.name); goto err; }
        sprintf(&str[len],"%.32s %d%s",nv.name,nv.ival,sep(i));
        if ((len=(int)strlen(str)) > siz - 64) break;
      }
      ccerr(0);
    }
    case CF_NAME48I:
    {
      NAME48I nv;
      for (i=0; i<n; i++) 
      {
        memcpy(&nv,d->data.bptr+i*sizeof(NAME48I),sizeof(NAME48I));
        if (valueOnly) { sprintf(str,"%.48s",nv.name); goto err; }
        sprintf(&str[len],"%.48s %d%s",nv.name,nv.ival,sep(i));
        if ((len=(int)strlen(str)) > siz - 64) break;
      }
      ccerr(0);
    }
    case CF_NAME64I:
    {
      NAME64I nv;
      for (i=0; i<n; i++) 
      {
        memcpy(&nv,d->data.bptr+i*sizeof(NAME64I),sizeof(NAME64I));
        if (valueOnly) { sprintf(str,"%.64s",nv.name); goto err; }
        sprintf(&str[len],"%.64s %d%s",nv.name,nv.ival,sep(i));
        if ((len=(int)strlen(str)) > siz - 80) break;
      }
      ccerr(0);
    }
    case CF_NAME64TIME:
    case CF_NAME64DBL:
    {
      NAME64DBL nv;
      for (i=0; i<n; i++) 
      {
        memcpy(&nv,d->data.bptr+i*sizeof(NAME64DBL),sizeof(NAME64DBL));
        if (valueOnly) { sprintf(str,"%.64s",nv.name); goto err; }
        sprintf(&str[len],"%.64s %10.4f%s",nv.name,nv.dval,sep(i));
        if ((len=(int)strlen(str)) > siz - 80) break;
      }
      ccerr(0);
    }
    case CF_FLTINT:
    {
      FLTINT v;
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        memcpy(&v,d->data.bptr+i*sizeof(FLTINT),sizeof(FLTINT));
        sprintf(&str[len],"%g %d%s",v.fval,v.ival,sep(i));
        if ((len=(int)strlen(str)) > siz - 32) break;
      }
      ccerr(0);
    }
    case CF_LNGINT:
    {
      LNGINT v;
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        memcpy(&v,d->data.bptr+i*sizeof(LNGINT),sizeof(LNGINT));
        sprintf(&str[len],"%d %d%s",v.lval,v.ival,sep(i));
        if ((len=(int)strlen(str)) > siz - 32) break;
      }
      ccerr(0);
    }
    case CF_DBLTIME:
    case CF_DBLDBL:
    {
      DBLDBL v;
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        memcpy(&v,d->data.bptr+i*sizeof(DBLDBL),sizeof(DBLDBL));
        sprintf(&str[len],"%g %10.4f%s",v.d1val,v.d2val,sep(i));
        if ((len=(int)strlen(str)) > siz - 32) break;
      }
      ccerr(0);
    }
    case CF_INTFLTINT:
    {
      INTFLTINT v;
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        memcpy(&v,d->data.bptr+i*sizeof(INTFLTINT),sizeof(INTFLTINT));
        sprintf(&str[len],"%d %g %d%s",v.i1val,v.fval,v.i2val,sep(i));
        if ((len=(int)strlen(str)) > siz - 32) break;
      }
      ccerr(0);
    }
    case CF_USTRING:
    {
      DUSTRING v;
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        memcpy(&v,d->data.bptr+i*sizeof(DUSTRING),sizeof(DUSTRING));
        t = (time_t)v.tm;
        sprintf(&str[len],"%d %g %g %.24s %.80s%s",
          v.ival,v.f1val,v.f2val,ctime(&t),v.str,sep(i));
        if ((len=(int)strlen(str)) > siz - 128) break;
      }
      ccerr(0);
    }
    case CF_STRUCT:
      if (valueOnly) goto err;
      GetStructureAsString(d->dTag,d->data.vptr,str,siz);
      ccerr(0);
    case CF_SPECTRUM:
    { /* TODO: make use of hdrSPECTRUM or USPECTRUM + dynamic assessment of data length (avoid large stack) */
      DSPECTRUM v;
      if (valueOnly) goto err;
      if (siz < 256) return buffer_too_small;
      memcpy(&v,d->data.bptr,sizeof(DSPECTRUM));
      sprintf(&str[len],"%.80s\n",v.comment);
      t = (time_t)v.tm; len = (int)strlen(str);
      sprintf(&str[len],"%.24s (%ld)\n",ctime(&t),(long)t); len = (int)strlen(str);
      sprintf(&str[len],"start : %f\n",v.s_start); len = (int)strlen(str);
      sprintf(&str[len],"inc   : %f\n",v.s_inc); len = (int)strlen(str);
      sprintf(&str[len],"status: %d\n",v.status); len = (int)strlen(str);
      sprintf(&str[len],"values: "); len = (int)strlen(str);
      if (len > siz - 64) break;
      if (n > 10) n = 10;
      for (i=0; i<n && i<10; i++) 
      {
        ssep = separator == NULL ? i<(n-1) ? ", " : i<9 ? "\n" : ", ...\n" : separator;
        sprintf(&str[len],"%f%s",v.d_spect_array[i],ssep);
        len = (int)strlen(str);
      }
      ccerr(0);
    }
    case CF_ASPECTRUM:
    {
      ASPECTRUM asp;
      int k,slen;
      if (valueOnly) goto err;
      if (siz < 256*n) return buffer_too_small;
      for (i=0; i<n; i++)
      {
        memcpy(&asp,d->data.bptr+i*sizeof(ASPECTRUM),sizeof(ASPECTRUM));
        sprintf(&str[len],"%.80s\n",asp.comment);
        t = (time_t)asp.tm; len = (int)strlen(str);
        sprintf(&str[len],"%.24s (%ld)\n",ctime(&t),(long)t); len = (int)strlen(str);
        sprintf(&str[len],"start : %f\n",asp.s_start); len = (int)strlen(str);
        sprintf(&str[len],"inc   : %f\n",asp.s_inc); len = (int)strlen(str);
        sprintf(&str[len],"status: %d\n",asp.status); len = (int)strlen(str);
        sprintf(&str[len],"values: "); len = (int)strlen(str);
        if (len > siz - 64) break;
        slen = asp.spectBufferSize;
        if (slen > 10) slen = 10;
        for (k=0; k<slen && k<10; k++) 
        {
          ssep = separator == NULL ? k<(slen-1) ? ", " : k<9 ? "\n" : ", ...\n" : separator;
          memcpy(&fval,((BYTE *)asp.spectBuffer)+k*sizeof(float),sizeof(float));
          sprintf(&str[len],"%f%s",fval,ssep);
          len = (int)strlen(str);
        }
        sprintf(&str[len],"\n"); len = (int)strlen(str);
      }
      ccerr(0);
    }
    case CF_GSPECTRUM:
    { /* TODO: make use of hdrSPECTRUM or USPECTRUM + dynamic assessment of data length (avoid large stack) */
      DGSPECTRUM v;
      if (valueOnly) goto err;
      if (siz < 256) return buffer_too_small;
      memcpy(&v, d->data.bptr, sizeof(DGSPECTRUM));
      sprintf(&str[len], "%.80s\n", v.comment);
      sprintf(&str[len], "id    : %d\n", v.id); len = (int)strlen(str);
      t = (time_t)v.tm; len = (int)strlen(str);
      sprintf(&str[len], "%.24s (%ld)\n", ctime(&t), (long)t); len = (int)strlen(str);
      sprintf(&str[len], "start : %f\n", v.s_start); len = (int)strlen(str);
      sprintf(&str[len], "inc   : %f\n", v.s_inc); len = (int)strlen(str);
      sprintf(&str[len], "status: %d\n", v.status); len = (int)strlen(str);
      sprintf(&str[len], "groups: %d\n", v.groups); len = (int)strlen(str);
      sprintf(&str[len], "group_size: %d\n", v.group_size); len = (int)strlen(str);
      sprintf(&str[len], "group_inc: %f\n", v.group_inc); len = (int)strlen(str);
      sprintf(&str[len], "stat_mask: %d\n", v.stat_mask); len = (int)strlen(str);
      /* TODO: fill in later ...*/
      sprintf(&str[len], "values: "); len = (int)strlen(str);
      if (len > siz - 64) break;
      if (n > 10) n = 10;
      for (i = 0; i<n && i<10; i++)
      {
        ssep = separator == NULL ? i<(n - 1) ? ", " : i<9 ? "\n" : ", ...\n" : separator;
        sprintf(&str[len], "%f%s", v.spectBuffer[i], ssep);
        len = (int)strlen(str);
      }
      ccerr(0);
    }
    default:
      if (siz > 18) sprintf(str,"%.16s values",GetFormatAsString(d->dFormat));
      return illegal_format;
  }
err:
  *dsiz = len;
  return cc;
}
TINE_EXPORT int GetValuesAsStringEx(DTYPE *d,char *str,int fmt,int num,int offset)
{
  int i,n,len;
  int stlen = 0, slen, snum;
  int dtlen = 0, dlen, dnum = num, doff=0;
  int mapStringArray=0;
  DTYPE ddst;
  char *dst, *src;

  if (d == NULL || str == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  src = d->data.cptr;
  snum = d->dArrayLength;
  dst = (char *)str;
  switch (d->dFormat)
  { /* source fmt */
    case CF_KEYVALUE:
    case CF_STRING:
      mapStringArray |= MAP_STRING_SRC;
      break;
    case CF_BYTE:
    case CF_CHAR:
      slen = d->dArrayLength;
      snum = 1;
      stlen = d->dArrayLength;
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      slen = fmtsizeof(d->dFormat);
      stlen = slen;
      break;
    case CF_USTRING:
      slen = 80;
      stlen = fmtsizeof(d->dFormat);
      src += 16; /* skip over the header */
      break;
    case CF_UNAME:
      slen = 16;
      stlen = fmtsizeof(d->dFormat);
      src += 16; /* skip over the header */
      break;
    default:
      return illegal_format;      
  }
  switch (fmt)
  { /* destination format */
    case CF_KEYVALUE:
    case CF_STRING:
      mapStringArray |= MAP_STRING_DST;
      break;
    case CF_CHAR:
      dlen = num;
      dnum = 1;
      dtlen = num;
      offset = 0;
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dlen = fmtsizeof(fmt);
      dtlen = dlen;
      break;
    case CF_NAME64TIME:
    case CF_NAME64DBL:
      dlen = fmtsizeof(fmt);
      dtlen = 64;
      break;
    case CF_USTRING:
      dlen = 80;
      dtlen = fmtsizeof(fmt);
      doff = 16; /* skip over the header */
      break;
    case CF_UNAME:
      dlen = 16;
      dtlen = fmtsizeof(fmt);
      doff = 16; /* skip over the header */
      break;
    default:
      return illegal_format;
  }
  if (offset < 0 || offset >= dnum) return dimension_error;
  if (mapStringArray)
  {
    memset(&ddst,0,sizeof(DTYPE));
    ddst.dArrayLength = fmt == CF_CHAR ? dlen : MIN(dnum,snum);
    ddst.dFormat = fmt;
    ddst.data.strptr = &((char **)dst)[offset];
    return mapStringArrays(&ddst,d,mapStringArray,0);
  }
  n = MIN(snum,(dnum-offset));
  len = MIN(dtlen,stlen);
  if (src == dst)
  { /* in-situ */
    if (dtlen > stlen)
    { /* smaller to larger */
      for (i=n-1; i>=0; i--)
      {
        if (i == 0 && offset == 0)
        { /* don't wipe out the first entry */
          memset(src+stlen,0,stlen);
          break;
        }
        memset(dst + (i+offset)*dtlen,0,dtlen);
        strncpy(dst + doff + (i+offset)*dtlen, src + i*stlen, len);
      }
    }
    else
    { /* larger to smaller */
      for (i=0; i<n; i++)
      {
        strncpy(dst + doff + (i+offset)*dtlen, src + i*stlen, len);
      }
    }
  }
  else
  { /* not in-situ : just do it */
    for (i=0; i<n; i++)
    {
      memset(dst + (i+offset)*dtlen,0,dtlen);
      strncpy(dst + doff + (i+offset)*dtlen, src + i*stlen, len);
    }
  }
  return 0;
}

TINE_EXPORT int GetValuesAsDBLDBL(DTYPE *d,DBLDBL *ddval,int num)
{
  int i,n;

  if (d == NULL || ddval == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_FLTFLT: 
      for (i=0; i<n; i++) 
      {
        ddval[i].d1val = (double)((FLTFLT *)d->data.vptr)[i].f1val; 
        ddval[i].d2val = (double)((FLTFLT *)d->data.vptr)[i].f2val; 
      }
      return 0;
    case CF_FI : 
      for (i=0; i<n; i++) 
      {
        ddval[i].d1val = (double)((FLTINT *)d->data.vptr)[i].fval; 
        ddval[i].d2val = (double)((FLTINT *)d->data.vptr)[i].ival; 
      }
      return 0;
    case CF_II : 
      for (i=0; i<n; i++) 
      {
        ddval[i].d1val = (double)((INTINT *)d->data.vptr)[i].i1val; 
        ddval[i].d2val = (double)((INTINT *)d->data.vptr)[i].i2val; 
      }
      return 0;
    case CF_DBLTIME : 
    case CF_DBLDBL : 
      if (ddval == (DBLDBL *)d->data.vptr) return 0;
		  memcpy(ddval,d->data.vptr,n*sizeof(DBLDBL));
      return 0;
    default: 
      return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsNAME64DBL(DTYPE *d,NAME64DBL *ndval,int num)
{
  int i,n, flen, nlen;
  char *nptr;
  SINT32 *iptr;

  if (d == NULL || ndval == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_CHAR8I  : 
    case CF_CHAR16I : 
    case CF_CHAR32I :
    case CF_CHAR48I : 
    case CF_CHAR64I : 
      flen = fmtsizeof(d->dFormat);
      nlen = flen - sizeof(SINT32);
      nptr = d->data.cptr;
      for (i=0; i<n; i++) 
      {
        nptr += flen; 
        iptr = (SINT32 *)(nptr + nlen);
        memset(ndval[i].name,0,64);
        strncpy(ndval[i].name,nptr,nlen);
        ndval[i].dval = (double)(*iptr);
      }
      return 0;
    case CF_NAME64TIME:
    case CF_NAME64DBL:
      if (ndval == (NAME64DBL *)d->data.vptr) return 0;
		  memcpy(ndval,d->data.vptr,n*sizeof(NAME64DBL));
      return 0;
    default:
      return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsNAME64DBLDBL(DTYPE *d,NAME64DBLDBL *nddval,int num)
{
  int i,n;

  if (d == NULL || nddval == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_CHAR16FI: 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME16FI *)d->data.vptr)[i].name,16);
        nddval[i].d1val = (double)((NAME16FI *)d->data.vptr)[i].fval; 
        nddval[i].d2val = (double)((NAME16FI *)d->data.vptr)[i].ival; 
      }
      return 0;
    case CF_NAME16II : 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME16II *)d->data.vptr)[i].name,16);
        nddval[i].d1val = (double)((NAME16II *)d->data.vptr)[i].i1val; 
        nddval[i].d2val = (double)((NAME16II *)d->data.vptr)[i].i2val; 
      }
      return 0;
    case CF_NAME16DBLDBL : 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME16DBLDBL *)d->data.vptr)[i].name,16);
        memcpy(&nddval[i].d1val,&((NAME16DBLDBL *)d->data.vptr)[i].d1val,2*sizeof(double));
      }
      return 0;
    case CF_NAME32DBLDBL : 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME32DBLDBL *)d->data.vptr)[i].name,32);
        memcpy(&nddval[i].d1val,&((NAME32DBLDBL *)d->data.vptr)[i].d1val,2*sizeof(double));
      }
      return 0;
    case CF_NAME64DBLDBL : 
      if (nddval == (NAME64DBLDBL *)d->data.vptr) return 0;
		  memcpy(nddval,d->data.vptr,n*sizeof(NAME64DBLDBL));
      return 0;
    default:
      return illegal_format;
  }
}
#else
TINE_EXPORT int GetValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  return illegal_format;
}
#endif /* not SMALL_TINE_LIB */
#define getValuesAsType(dt,v,nr,f,CF) {\
  if (dt == NULL || v == NULL) return argument_list_error;\
  if (dt->dArrayLength < 1 || nr < 1) return dimension_error;\
  n = MIN((int)dt->dArrayLength, nr);\
  if (dt->dFormat == f) {\
    if (v == (CF *)dt->data.bptr) return 0;\
    memcpy(v, d->data.bptr, n*sizeof(CF));\
    return 0;\
  }\
  switch (dt->dFormat)\
  {\
  case CF_BITFIELD8: case CF_BYTE:\
    for (i = n - 1; i >= 0; i--) v[i] = (CF)dt->data.bptr[i];\
    return 0;\
  case CF_BITFIELD16:  case CF_UINT16: case CF_INT16:\
    for (i = n - 1; i >= 0; i--) v[i] = (CF)dt->data.sptr[i];\
    return 0;\
  case CF_BOOLEAN:\
    for (i = n - 1; i >= 0; i--) v[i] = (CF)((internal || !gMarshallBooleanAsByte) ? dt->data.lptr[i] : dt->data.bptr[i]);\
    return 0;\
  case CF_BITFIELD32: case CF_UINT32: case CF_INT32:\
    for (i = n - 1; i >= 0; i--) v[i] = (CF)dt->data.lptr[i];\
    return 0;\
  case CF_FLOAT:\
    for (i = 0; i<n; i++) v[i] = (CF)dt->data.fptr[i];\
    return 0;\
  case CF_DOUBLE:\
    for (i = 0; i<n; i++) v[i] = (CF)dt->data.dptr[i];\
    return 0;\
  case CF_FLTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((FLTINT *)dt->data.vptr)[i].fval;\
    return 0;\
  case CF_INTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((INTINT *)dt->data.vptr)[i].i1val;\
    return 0;\
  case CF_DBLTIME: case CF_DBLDBL:\
    for (i = 0; i<n; i++) v[i] = (CF)((DBLDBL *)dt->data.vptr)[i].d1val;\
    return 0;\
  case CF_FLTFLT:\
    for (i = 0; i<n; i++) v[i] = (CF)((FLTFLT *)dt->data.vptr)[i].f1val;\
    return 0;\
  case CF_FLTFLTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((FLTFLTINT *)dt->data.vptr)[i].f1val;\
    return 0;\
  case CF_INTINTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((INTINTINT *)dt->data.vptr)[i].i1val;\
    return 0;\
  case CF_INTINTINTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((DADDRESS *)dt->data.vptr)[i].i1val;\
    return 0;\
  case CF_NAME16FI:\
    for (i = 0; i<n; i++) v[i] = (CF)((NAME16FI *)dt->data.vptr)[i].fval;\
    return 0;\
  case CF_NAME16II:\
    for (i = 0; i<n; i++) v[i] = (CF)((NAME16II *)dt->data.vptr)[i].i1val;\
    return 0;\
  case CF_INTFLTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((INTFLTINT *)dt->data.vptr)[i].fval;\
    return 0;\
  case CF_FLTINTINT:\
    for (i = 0; i<n; i++) v[i] = (CF)((FLTINTINT *)dt->data.vptr)[i].fval;\
    return 0;\
  case CF_NAME16DBLDBL:\
    for (i = 0; i<n; i++) v[i] = (CF)((NAME16DBLDBL *)dt->data.vptr)[i].d1val;\
    return 0;\
  case CF_NAME32DBLDBL:\
    for (i = 0; i<n; i++) v[i] = (CF)((NAME32DBLDBL *)dt->data.vptr)[i].d1val;\
    return 0;\
  case CF_NAME64DBLDBL:\
    for (i = 0; i<n; i++) v[i] = (CF)((NAME64DBLDBL *)dt->data.vptr)[i].d1val;\
    return 0;\
  case CF_SPECTRUM:\
    for (i = 0; i<n; i++) v[i] = (CF)((DSPECTRUM *)dt->data.vptr)->d_spect_array[i];\
    return 0;\
  case CF_GSPECTRUM:\
    for (i = 0; i<n; i++) v[i] = (CF)((DGSPECTRUM *)dt->data.vptr)->spectBuffer[i];\
    return 0;\
  default: break;\
} }
#ifdef SINT64
#ifdef OS_64BIT
#define getLongValuesAsType(dt,v,nr,CF) {\
  switch (dt->dFormat)\
  {\
    case CF_UINT64:\
      for (i = 0; i<n; i++) v[i] = (CF)(dt->data.ullptr)[i];\
      return 0;\
    case CF_INT64:\
      for (i = 0; i<n; i++) v[i] = (CF)(dt->data.llptr)[i];\
      return 0;\
   } }
#else
#define getLongValuesAsType(dt,v,nr,CF) {\
  switch (dt->dFormat)\
  {\
    case CF_UINT64:\
    case CF_INT64:\
      for (i = 0; i<n; i++) v[i] = (CF)(dt->data.llptr)[i];\
      return 0;\
   } }
#endif /* OS_64BIT is defined */
#endif /* SINT64 is defined */
TINE_EXPORT int GetValuesAsByte(DTYPE *d, BYTE *bval, int num)
{
  return _GetValuesAsByte(d, bval, num, FALSE);
}
int _GetValuesAsByte(DTYPE *d, BYTE *bval, int num, int internal)
{
  int i, n;
  getValuesAsType(d, bval, num, CF_BYTE, BYTE);
# ifdef SINT64
  getLongValuesAsType(d, bval, num, BYTE);
# endif
  return illegal_format;
}
TINE_EXPORT int GetValuesAsShort(DTYPE *d, short *sval, int num)
{
  return _GetValuesAsShort(d, sval, num, FALSE);
}
int _GetValuesAsShort(DTYPE *d, short *sval, int num, int internal)
{
  int i, n;
  getValuesAsType(d, sval, num, CF_INT16, SINT16);
# ifdef SINT64
  getLongValuesAsType(d, sval, num, SINT16);
# endif
  return illegal_format;
}
TINE_EXPORT int GetValuesAsLong(DTYPE *d, SINT32 *lval, int num)
{
  return _GetValuesAsLong(d, lval, num, FALSE);
}
int _GetValuesAsLong(DTYPE *d,SINT32 *lval,int num, int internal)
{
  int i,n;
  getValuesAsType(d, lval, num, CF_INT32, SINT32);
# ifdef SINT64
  getLongValuesAsType(d, lval, num, SINT32);
# endif
  return illegal_format;
}
TINE_EXPORT int GetValuesAsUnsignedLong(DTYPE *d, UINT32 *lval, int num)
{
  return _GetValuesAsUnsignedLong(d, lval, num, FALSE);
}
int _GetValuesAsUnsignedLong(DTYPE *d,UINT32 *lval,int num, int internal)
{
  int i,n;
  getValuesAsType(d, lval, num, CF_UINT32, UINT32);
# ifdef SINT64
  getLongValuesAsType(d, lval, num, UINT32);
# endif
  return illegal_format;
}
TINE_EXPORT int GetValuesAsUnsignedShort(DTYPE *d, UINT16 *val, int num)
{
  return _GetValuesAsUnsignedShort(d, val, num, FALSE);
}
int _GetValuesAsUnsignedShort(DTYPE *d, UINT16 *val, int num, int internal)
{
  int i, n;
  getValuesAsType(d, val, num, CF_UINT16, UINT16);
# ifdef SINT64
  getLongValuesAsType(d, val, num, UINT16);
# endif
  return illegal_format;
}
#ifdef SINT64
TINE_EXPORT int GetValuesAsDLong(DTYPE *d, SINT64 *llval, int num)
{
  return _GetValuesAsDLong(d, llval, num, FALSE);
}
int _GetValuesAsDLong(DTYPE *d,SINT64 *llval,int num, int internal)
{
  int i, n;
  getValuesAsType(d, llval, num, CF_INT64, SINT64);
  getLongValuesAsType(d, llval, num, SINT64);
  return illegal_format;
}
TINE_EXPORT int GetValuesAsUnsignedDLong(DTYPE *d, UINT64 *llval, int num)
{
  return _GetValuesAsUnsignedDLong(d, llval, num, FALSE);
}
int _GetValuesAsUnsignedDLong(DTYPE *d,UINT64 *llval,int num, int internal)
{
  int i,n;
  getValuesAsType(d, llval, num, CF_UINT64, UINT64);
  getLongValuesAsType(d, llval, num, UINT64);
  return illegal_format;
}
#endif /* SINT64 */
TINE_EXPORT int GetValuesAsFloat(DTYPE *d, float *fval, int num)
{
  return _GetValuesAsFloat(d, fval, num, FALSE);
}
int _GetValuesAsFloat(DTYPE *d, float *fval, int num, int internal)
{
  int i, n;
  getValuesAsType(d, fval, num, CF_FLOAT, float);
# ifdef SINT64
  getLongValuesAsType(d, fval, num, float);
# endif
  return illegal_format;
}
TINE_EXPORT int GetValuesAsDouble(DTYPE *d, double *dval, int num)
{
  return _GetValuesAsDouble(d, dval, num, FALSE);
}
int _GetValuesAsDouble(DTYPE *d,double *dval,int num, int internal)
{
  int i,n;
  getValuesAsType(d, dval, num, CF_DOUBLE, double);
# ifdef SINT64
  getLongValuesAsType(d, dval, num, double);
# endif
  return illegal_format;
}
#ifdef WIN32
TINE_EXPORT int getValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  return GetValuesAsString(d,str,dsiz);
}
TINE_EXPORT int getValuesAsShort(DTYPE *d,short *s,int n)
{
  return GetValuesAsShort(d,s,n);
}
TINE_EXPORT int getValuesAsFloat(DTYPE *d,float *s,int n)
{
  return GetValuesAsFloat(d,s,n);
}
TINE_EXPORT int getValuesAsDouble(DTYPE *d,double *s,int n) 
{
  return GetValuesAsDouble(d,s,n);
}
TINE_EXPORT int getValuesAsLong(DTYPE *d,SINT32 *s,int n) 
{
  return GetValuesAsLong(d,s,n);
}
#endif

#define putValuesFromType(dt,v,nr,off,f) {\
  if (dt == NULL || v == NULL) return argument_list_error;\
  if ((dt->dFormat != CF_TEXT) && (dt->dArrayLength < 1 || nr < 1)) return dimension_error;\
  if (off < 0 || off >= nr) return dimension_error;\
  n = MIN((int)dt->dArrayLength, nr);\
  if (f == dt->dFormat && off == 0) memcpy(dt->data.vptr,v,n * fmtsizeof(f));\
  switch (dt->dFormat)\
  {\
    case CF_BYTE: case CF_BITFIELD8:\
      for (i = 0; i<n; i++) dt->data.bptr[i] = (BYTE)v[(i + off) % nr];\
      return 0;\
    case CF_UINT16: case CF_BITFIELD16:\
      for (i = 0; i<n; i++) dt->data.usptr[i] = (UINT16)v[(i + off) % nr];\
      return 0;\
    case CF_INT16:\
      for (i = 0; i<n; i++) dt->data.sptr[i] = (SINT16)v[(i + off) % nr];\
      return 0;\
    case CF_BOOLEAN:\
      for (i = 0; i<n; i++)\
      {\
        if (internal || !gMarshallBooleanAsByte)\
          dt->data.ulptr[i] = (UINT32)v[(i + off) % nr];\
        else\
          dt->data.bptr[i] = (_Bool)v[(i + off) % nr];\
      }\
      return 0;\
    case CF_UINT32: case CF_BITFIELD32:\
      for (i = 0; i<n; i++) dt->data.ulptr[i] = (UINT32)v[(i + off) % nr];\
      return 0;\
    case CF_INT32:\
      for (i = 0; i<n; i++) dt->data.lptr[i] = (SINT32)v[(i + off) % nr];\
      return 0;\
    case CF_FLOAT:\
      for (i = 0; i<n; i++) dt->data.fptr[i] = (float)v[(i + off) % nr];\
      return 0;\
    case CF_DBLTIME: case CF_DBLDBL:\
       for (i = 0; i<n; i++) ((DBLDBL *)dt->data.vptr)[i].d1val = (double)v[(i + off) % nr];\
       return 0;\
    case CF_DOUBLE:\
       for (i = 0; i<n; i++) dt->data.dptr[i] = (double)v[(i + off) % nr];\
       return 0;\
    case CF_TEXT:\
       dl.dArrayLength = nr; dl.dFormat = f; dl.data.vptr = v;\
       return GetValuesAsString(&dl, dt->data.cptr, &dt->dArrayLength);\
    default: break;\
  } }

#define putLongValuesFromType(dt,v,nr,off) {\
  switch (dt->dFormat)\
  {\
    case CF_UINT64:\
      for (i = 0; i<n; i++) dt->data.ullptr[i] = (UINT64)v[(i + off) % nr];\
      return 0;\
    case CF_INT64:\
      for (i = 0; i<n; i++) dt->data.llptr[i] = (SINT64)v[(i + off) % nr];\
      return 0;\
   } }
TINE_EXPORT int PutValuesFromDoubleEx(DTYPE *d, double *dval, int num, int offset)
{
  return _PutValuesFromDoubleEx(d, dval, num, offset, FALSE);
}
int _PutValuesFromDoubleEx(DTYPE *d,double *dval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, dval, num, offset, CF_DOUBLE);
# ifdef SINT64
  putLongValuesFromType(d, dval, num, offset);
# endif
  return illegal_format;
}
TINE_EXPORT int PutValuesFromFloatEx(DTYPE *d, float *fval, int num, int offset)
{
  return _PutValuesFromFloatEx(d, fval, num, offset, FALSE);
}
int _PutValuesFromFloatEx(DTYPE *d,float *fval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, fval, num, offset, CF_FLOAT);
# ifdef SINT64
  putLongValuesFromType(d, fval, num, offset);
# endif
  return illegal_format;
}
TINE_EXPORT int PutValuesFromLongEx(DTYPE *d, SINT32 *lval, int num, int offset)
{
  return _PutValuesFromLongEx(d, lval, num, offset, FALSE);
}
int _PutValuesFromLongEx(DTYPE *d,SINT32 *lval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, lval, num, offset, CF_INT32);
# ifdef SINT64
  putLongValuesFromType(d, lval, num, offset);
# endif
  return illegal_format;
}
TINE_EXPORT int PutValuesFromUnsignedLongEx(DTYPE *d, UINT32 *lval, int num, int offset)
{
  return _PutValuesFromUnsignedLongEx(d, lval, num, offset, FALSE);
}
int _PutValuesFromUnsignedLongEx(DTYPE *d,UINT32 *lval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, lval, num, offset, CF_UINT32);
# ifdef SINT64
  putLongValuesFromType(d, lval, num, offset);
# endif
  return illegal_format;
}
#ifdef SINT64
TINE_EXPORT int PutValuesFromDLongEx(DTYPE *d, SINT64 *llval, int num, int offset)
{
  return _PutValuesFromDLongEx(d, llval, num, offset, FALSE);
}
int _PutValuesFromDLongEx(DTYPE *d,SINT64 *llval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, llval, num, offset, CF_INT64);
  putLongValuesFromType(d, llval, num, offset);
  return illegal_format;
}
TINE_EXPORT int PutValuesFromUnsignedDLongEx(DTYPE *d, UINT64 *llval, int num, int offset)
{
  return _PutValuesFromUnsignedDLongEx(d, llval, num, offset, FALSE);
}
int _PutValuesFromUnsignedDLongEx(DTYPE *d, UINT64 *llval, int num, int offset, int internal)
{
  int i, n;
  DTYPE dl;
# ifdef OS_64BIT
  UINT64 *ullval = llval;
# else
  SINT64 *ullval = (SINT64 *)llval;
# endif
  putValuesFromType(d, ullval, num, offset, CF_UINT64);
  putLongValuesFromType(d, llval, num, offset);
  return illegal_format;
}
#endif /* SINT64 */
TINE_EXPORT int PutValuesFromShortEx(DTYPE *d, short *sval, int num, int offset)
{
  return _PutValuesFromShortEx(d, sval, num, offset, FALSE);
}
int _PutValuesFromShortEx(DTYPE *d,short *sval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, sval, num, offset, CF_INT16);
# ifdef SINT64
  putLongValuesFromType(d, sval, num, offset);
# endif
  return illegal_format;
}
TINE_EXPORT int PutValuesFromUnsignedShortEx(DTYPE *d, UINT16 *sval, int num, int offset)
{
  return _PutValuesFromUnsignedShortEx(d, sval, num, offset, FALSE);
}
int _PutValuesFromUnsignedShortEx(DTYPE *d,UINT16 *sval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, sval, num, offset, CF_UINT16);
# ifdef SINT64
  putLongValuesFromType(d, sval, num, offset);
# endif
  return illegal_format;
}
TINE_EXPORT int PutValuesFromByteEx(DTYPE *d, BYTE *bval, int num, int offset)
{
  return _PutValuesFromByteEx(d, bval, num, offset, FALSE);
}
int _PutValuesFromByteEx(DTYPE *d,BYTE *bval,int num,int offset,int internal)
{
  int i,n;
  DTYPE dl;
  putValuesFromType(d, bval, num, offset, CF_BYTE);
# ifdef SINT64
  putLongValuesFromType(d, bval, num, offset);
# endif
  return illegal_format;
}
int mapStringArrays(DTYPE *ddst,DTYPE *dsrc,int mapping,int offset)
{
  int i, n = MIN((int)ddst->dArrayLength,(int)dsrc->dArrayLength-offset);
  int flen, slen = 0, soff = 0, tptr = 0;
  char *ssrc, *tsrc;
  if (tmpWorkArea == NULL && (tmpWorkArea=(BYTE *)SystemCalloc(1,tmpWorkAreaSize)) == NULL)
  {
    return out_of_local_memory;
  }
  tsrc = (char *)tmpWorkArea;
  switch (mapping)
  {
    case MAP_STRING_SRC:
      flen = fmtsizeof(ddst->dFormat); soff = 0;
      switch (ddst->dFormat)
      {
        case CF_TEXT:
          slen = ddst->dArrayLength;
          n = 1;
          break;
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          slen = flen;
          break;
        case CF_NAME64TIME:
        case CF_NAME64DBL:
          slen = 64;
          break;
        case CF_USTRING:
          slen = 80;
          soff = 16; /* skip over the header */
          break;
        case CF_UNAME:
          slen = 16;
          soff = 16; /* skip over the header */
          break;
        default:
          return illegal_format;
      }
      for (i=0; i<n; i++)
      {
        memset(ddst->data.bptr + i*flen,0,flen);
        strncpy(ddst->data.cptr + soff + i*flen, dsrc->data.strptr[i+offset], slen);
      }
      return 0;
    case MAP_STRING_DST:
      flen = fmtsizeof(dsrc->dFormat); soff = 0;
      switch (dsrc->dFormat)
      {
        case CF_TEXT:
          slen = dsrc->dArrayLength;
          n = 1;
          break;
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          slen = flen;
          break;
        case CF_NAME64TIME:
        case CF_NAME64DBL:
          slen = 64;
          break;
        case CF_USTRING:
          slen = 80;
          soff = 16;
          break;
        case CF_UNAME:
          slen = 16;
          soff = 16;
          break;
        default:
          return illegal_format;
      }
      for (i=0; i<n; i++)
      {
        ssrc = &dsrc->data.cptr[soff+i*flen];
        if (strlen(ssrc) > (size_t)slen)
        { /* no visible string termination => make a temporary terminated string */
          if (tptr > (int)tmpWorkAreaSize - slen) break;
          strncpy(&tsrc[tptr],ssrc,slen); tsrc[tptr+slen] = 0;
          ssrc = &tsrc[tptr];
          tptr += slen + 1;
        }
        ddst->data.strptr[i] = ssrc;
      }
      return 0;
    case MAP_STRING_SRC|MAP_STRING_DST:
      for (i=0; i<n; i++) 
        ddst->data.strptr[i] = dsrc->data.strptr[i+offset];
      return 0;
    default:
      return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromString(DTYPE *d,char *str,int fmt,int num,int offset)
{
  int i,n,len;
  int stlen = 0, slen, snum = num, soff = 0;
  int dtlen = 0, dlen, dnum, doff=0;
  int mapStringArray=0;
  DTYPE dsrc;
  char *dst, *src;

  if (d == NULL || str == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  dst = d->data.cptr;
  src = (char *)str;
  switch (fmt)
  { /* source fmt (USTRING, UNAME should the 2nd param to (char *) */
    case CF_KEYVALUE:
    case CF_STRING:
      mapStringArray |= MAP_STRING_SRC;
      break;
    case CF_TEXT:
      slen = num;
      snum = 1;
      stlen = num;
      offset = 0;
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      slen = fmtsizeof(fmt);
      stlen = slen;
      break;
    case CF_USTRING:
      slen = 80;
      stlen = fmtsizeof(fmt);
      soff = 16;
      break;
    case CF_UNAME:
      slen = 16;
      stlen = fmtsizeof(fmt);
      soff = 16;
      break;
    default:
      return illegal_format;      
  }
  if (offset < 0 || offset >= snum) return dimension_error;
  dnum = d->dArrayLength;
  switch (d->dFormat)
  { /* destination format */
    case CF_KEYVALUE:
    case CF_STRING:
      mapStringArray |= MAP_STRING_DST;
      break;
    case CF_TEXT:
      dlen = d->dArrayLength;
      dnum = 1;
      dtlen = d->dArrayLength;
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dlen = fmtsizeof(d->dFormat);
      dtlen = dlen;
      break;
    case CF_NAME64TIME:
    case CF_NAME64DBL:
      dlen = fmtsizeof(d->dFormat);
      dtlen = 64;
      break;
    case CF_USTRING:
      dlen = 80;
      dtlen = fmtsizeof(d->dFormat);
      doff = 16; /* skip over the header */
      break;
    case CF_UNAME:
      dlen = 16;
      dtlen = fmtsizeof(d->dFormat);
      doff = 16; /* skip over the header */
      break;
    default:
      return illegal_format;
  }
  if (mapStringArray)
  {
    memset(&dsrc,0,sizeof(DTYPE));
    dsrc.dArrayLength = num;
    dsrc.dFormat = fmt;
    dsrc.data.bptr = (BYTE *)(str + stlen*offset);
    return mapStringArrays(d,&dsrc,mapStringArray,offset);
  }
  n = MIN(dnum,(snum-offset));
  len = MIN(dtlen,stlen);
  if (src == dst)
  { /* in-situ */
    if (dtlen > stlen)
    { /* smaller to larger */
      for (i=n-1; i>=0; i--)
      {
        if (i == 0 && offset == 0)
        { /* don't wipe out the first entry */
          memset(src+stlen,0,stlen);
          break;
        }
        memset(dst + i*dtlen,0,dtlen);
        strncpy(dst + doff + i*dtlen, src + soff + (i+offset)*stlen, len);
      }
    }
    else
    { /* larger to smaller */
      for (i=0; i<n; i++)
      {
        strncpy(dst + doff + i*dtlen, src + soff + (i+offset)*stlen, len);
      }
    }
  }
  else
  { /* not in-situ : just do it */
    for (i=0; i<n; i++)
    {
      memset(dst + i*dtlen,0,dtlen);
      strncpy(dst + doff + i*dtlen, src + soff + (i+offset)*stlen, len);
    }
  }
  return 0;
}
static int _formatsEqual(short fmt1, short fmt2)
{
  fmt1 = LFMT(fmt1);
  fmt2 = LFMT(fmt2);
  if (fmt1 == fmt2) return TRUE;
  switch (fmt1)
  {
    case CF_DBLDBL:
      if (fmt2 == CF_DBLTIME) return TRUE;
      return FALSE; /* fmt2 = fmt1 handled above */
    case CF_DBLTIME:
      if (fmt2 == CF_DBLDBL) return TRUE;
      return FALSE; /* fmt2 = fmt1 handled above */
    case CF_NAME64DBL:
      if (fmt2 == CF_NAME64TIME) return TRUE;
      return FALSE; /* fmt2 = fmt1 handled above */
    case CF_NAME64TIME:
      if (fmt2 == CF_NAME64DBL) return TRUE;
      return FALSE; /* fmt2 = fmt1 handled above */
  }
  return FALSE;
}
TINE_EXPORT int PutValuesFromAny(DTYPE *d,void *val,short fmt,int sgn,int objectSizeInBytes,int num,int offset)
{
  int i,n,slen,cc=0,dfsiz,sfsiz;
  DTYPE d0;
  void *v;

  if (d == NULL || val == NULL) return invalid_parameter;
  dfsiz = fmtsizeof(d->dFormat); sfsiz = fmtsizeof(fmt);
  if (objectSizeInBytes < sfsiz) return invalid_parameter;
  if (objectSizeInBytes == sfsiz) 
  { /* pass it on directly : let the memcpy()'s do their stuff ... */
    switch (fmt)
    {
      case CF_INT8:
        return PutValuesFromByteEx(d,(BYTE *)val,num,offset);
      case CF_UINT16:
      case CF_INT16:
        return sgn ? PutValuesFromShortEx(d,(SINT16 *)val,num,offset) : PutValuesFromUnsignedShortEx(d,(UINT16 *)val,num,offset);
      case CF_UINT32:
      case CF_INT32:
        return sgn ? PutValuesFromLongEx(d,(SINT32 *)val,num,offset) : PutValuesFromUnsignedLongEx(d,(UINT32 *)val,num,offset);
#     ifdef SINT64
      case CF_UINT64:
      case CF_INT64:
        return PutValuesFromDLongEx(d,(SINT64 *)val,num,offset);
#     endif
      case CF_FLOAT:
        return PutValuesFromFloatEx(d,(float *)val,num,offset);
      case CF_DOUBLE:
        return PutValuesFromDoubleEx(d,(double *)val,num,offset);
      case CF_STRING:
      default:
        if (isSimpleStringFormat(fmt) || fmt == CF_STRING)
        {
          return PutValuesFromString(d,(char *)val,fmt,num,offset);
        }
        if (!_formatsEqual(fmt,d->dFormat)) return illegal_format;
        for (i=0; i<(int)d->dArrayLength; i++) 
        {
          memcpy(d->data.bptr+i*sfsiz,(BYTE *)val+sfsiz*((i+offset)%num),sfsiz);
        }
        return 0;
    }
  }
  n = MIN((int)d->dArrayLength,num);
  d0.dFormat = d->dFormat;
  d0.dArrayLength = 1;
  for (i=0; i<n && cc == 0; i++)
  {
    d0.data.bptr = d->data.bptr + i*dfsiz;
    v = (void *)((BYTE *)val + objectSizeInBytes*((i+offset)%num));
    switch (fmt)
    {
      case CF_INT8:
        cc = PutValuesFromByteEx(&d0,(BYTE *)v,1,0); 
        break;
      case CF_UINT16:
      case CF_INT16:
        cc = sgn ? PutValuesFromShortEx(&d0,(SINT16 *)v,1,0) : PutValuesFromUnsignedShortEx(&d0,(UINT16 *)v,1,0);
        break;
      case CF_UINT32:
      case CF_INT32:
        cc = sgn ? PutValuesFromLongEx(&d0,(SINT32 *)v,1,0) : PutValuesFromUnsignedLongEx(&d0,(UINT32 *)v,1,0);
        break;
#     ifdef SINT64
      case CF_UINT64:
      case CF_INT64:
        cc = sgn ? PutValuesFromDLongEx(&d0,(SINT64 *)v,1,0) : PutValuesFromUnsignedDLongEx(&d0, (UINT64 *)v, 1, 0);
        break;
#     endif
      case CF_FLOAT:
        cc = PutValuesFromFloatEx(&d0,(float *)v,1,0);
        break;
      case CF_DOUBLE:
        cc = PutValuesFromDoubleEx(&d0,(double *)v,1,0);
        break;
      case CF_STRING:
        slen = n;
        i = n;
      default:
        if (isSimpleStringFormat(fmt) || fmt == CF_STRING)
        {
          slen = (fmt == CF_TEXT || fmt == CF_STRING) ? n : 1;
          d0.dArrayLength = n;
          PutValuesFromString(&d0,(char *)v,fmt,slen,0);
          if (fmt == CF_TEXT) i = n;
          break;
        }
        if (!_formatsEqual(fmt, d->dFormat)) return illegal_format;
        memcpy(d0.data.bptr,(BYTE *)v,sfsiz);
        break;
    }
  }
  return cc;
}
TINE_EXPORT int GetValuesAsAny(DTYPE *d, void *val, short fmt, int objectSizeInBytes, int num, int offset)
{
  return _GetValuesAsAny(d, val, fmt, objectSizeInBytes, num, offset, FALSE);
}
int _GetValuesAsAny(DTYPE *d,void *val,short fmt,int objectSizeInBytes,int num,int offset,int internal)
{ /* val is the target (destination) => curious mixup between sf and df below : sf is the destination format */
  int i,n,cc=0,dfsiz,sfsiz;
  DTYPE d0;
  void *v;

  if (d == NULL || val == NULL) return invalid_parameter;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  dfsiz = fmtsizeof(d->dFormat); if (d->dFormat == CF_STRING) dfsiz = sizeof(char *);
  sfsiz = fmtsizeof(fmt); if (fmt == CF_STRING) sfsiz = sizeof(char *);
  if (objectSizeInBytes < sfsiz) return invalid_parameter;
  if (objectSizeInBytes == sfsiz) 
  { /* pass it on directly : let the memcpy()'s do their stuff ... */
    switch (fmt)
    {
      case CF_INT8:
        return _GetValuesAsByte(d,&((BYTE *)val)[offset],num-offset,internal);
      case CF_UINT16:
        return _GetValuesAsUnsignedShort(d, &((UINT16 *)val)[offset], num - offset, internal);
      case CF_INT16:
        return _GetValuesAsShort(d,&((SINT16 *)val)[offset],num-offset, internal);
      case CF_UINT32:
        return _GetValuesAsUnsignedLong(d, &((UINT32 *)val)[offset], num - offset, internal);
      case CF_INT32:
        return _GetValuesAsLong(d,&((SINT32 *)val)[offset],num-offset, internal);
#     ifdef SINT64
      case CF_UINT64:
        return _GetValuesAsUnsignedDLong(d, &((UINT64 *)val)[offset], num - offset, internal);
      case CF_INT64:
        return _GetValuesAsDLong(d,&((SINT64 *)val)[offset],num-offset, internal);
#     endif
      case CF_BOOLEAN:
        if (internal || !gMarshallBooleanAsByte)
          return _GetValuesAsLong(d, &((SINT32*)val)[offset], num - offset, internal);
        else
          return GetValuesAsByte(d, &((BYTE*)val)[offset], num - offset);
      case CF_FLOAT:
        return _GetValuesAsFloat(d,&((float *)val)[offset],num-offset, internal);
      case CF_DOUBLE:
        return _GetValuesAsDouble(d,&((double *)val)[offset],num-offset, internal);
      case CF_STRING:
      default:
        if (isSimpleStringFormat(fmt) || fmt == CF_STRING)
        {
          return GetValuesAsStringEx(d,(char *)val,fmt,num,offset);
        }
        if (!_formatsEqual(fmt, d->dFormat)) return illegal_format;
        for (i=0; i<(int)d->dArrayLength && i<num-offset; i++) 
        {
          memcpy((BYTE *)val+sfsiz*(i+offset),d->data.bptr+i*sfsiz,sfsiz);
        }
        return 0;
    }
  }
  DTYPEZERO(d0);
  if (fmt == CF_STRUCT)
  {
    struct structStruct *ss = findstruct(d->dTag);
    if (d->dFormat != CF_STRUCT) return illegal_format;
    if (ss == NULL) return invalid_structure_tag;
    if (ss->size != objectSizeInBytes) return invalid_structure_size;
    dfsiz = objectSizeInBytes;
    strncpy(d0.dTag,d->dTag,TAG_NAME_SIZE);
  }
  n = MIN((int)d->dArrayLength,num);
  d0.dFormat = d->dFormat;
  d0.dArrayLength = 1;
  for (i=0; i<n && cc == 0; i++)
  {
    d0.data.bptr = d->data.bptr + i*dfsiz;
    v = (void *)((BYTE *)val + objectSizeInBytes*((i+offset)%num));
    switch (fmt)
    {
      case CF_INT8:
        cc = _GetValuesAsByte(&d0,(BYTE *)v,1, internal);
        break;
      case CF_UINT16:
        cc = _GetValuesAsUnsignedShort(&d0, (UINT16 *)v, 1, internal);
        break;
      case CF_INT16:
        cc = _GetValuesAsShort(&d0,(SINT16 *)v,1, internal);
        break;
      case CF_UINT32:
        cc = _GetValuesAsUnsignedLong(&d0, (UINT32 *)v, 1, internal);
        break;
      case CF_INT32:
        cc = _GetValuesAsLong(&d0,(SINT32 *)v,1, internal);
        break;
#     ifdef SINT64
      case CF_UINT64:
        cc = _GetValuesAsUnsignedDLong(&d0, (UINT64 *)v, 1, internal);
        break;
      case CF_INT64:
        cc = _GetValuesAsDLong(&d0,(SINT64 *)v,1, internal);
        break;
#     endif
      case CF_BOOLEAN:
        if (internal || !gMarshallBooleanAsByte)
          return  _GetValuesAsLong(&d0, (SINT32*)v, 1, internal);
        else
          return GetValuesAsByte(&d0, (BYTE*)v, 1);
      case CF_FLOAT:
        cc = _GetValuesAsFloat(&d0,(float *)v,1, internal);
        break;
      case CF_DOUBLE:
        cc = _GetValuesAsDouble(&d0,(double *)v,1, internal);
        break;
      case CF_STRING:
      default:
        if (isSimpleStringFormat(fmt) || fmt == CF_STRING)
        {
          cc = GetValuesAsStringEx(&d0,(char *)v,fmt,num,offset);
          break;
        }
        if (!_formatsEqual(fmt, d->dFormat)) return illegal_format;
        memcpy((BYTE *)v,d0.data.bptr,dfsiz);
    }
  }
  return cc;
}
TINE_EXPORT int AssignDataToDTYPE(DTYPE *d, void *ref, int offset, int format, int length, char *tag)
{
  BYTE *dp = (BYTE *)ref;
  int fsiz;
  if (d == NULL || dp == NULL) return argument_list_error;
  if (offset < 0 || length < 1) return invalid_data;
  if (getFormatHeaderSize(format) > 0) return not_supported;
  DTYPEZERO(*d);
  d->dFormat = format;
  if (d->dFormat == CF_NULL) return 0;
  fsiz = fmtsizeof(d->dFormat);
  if (d->dFormat == CF_STRUCT) 
  {
    if (tag == NULL) return invalid_structure_tag;
    strncpy(d->dTag,tag,TAG_NAME_SIZE);
    if ((fsiz=GetStructSize(tag)) < 0) return invalid_structure_tag;
  }
  d->dArrayLength = length;
  d->data.bptr = dp + fsiz * offset;
  return 0;
}
#ifdef WIN32
TINE_EXPORT int putValuesFromShortEx(DTYPE *d,short *s,int n,int o)
{
  return PutValuesFromShortEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromFloatEx(DTYPE *d,float *s,int n,int o)
{
  return PutValuesFromFloatEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromDoubleEx(DTYPE *d,double *s,int n,int o)
{
  return PutValuesFromDoubleEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromLongEx(DTYPE *d,SINT32 *s,int n,int o)
{
  return PutValuesFromLongEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromUnsignedLongEx(DTYPE *d,UINT32 *s,int n,int o)
{
  return PutValuesFromUnsignedLongEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromUnsignedShortEx(DTYPE *d,UINT16 *s,int n,int o)
{
  return PutValuesFromUnsignedShortEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromString(DTYPE *d,char *s,int f,int n,int o)
{
  return PutValuesFromString(d,s,f,n,o);
}
#endif
int _sndEvntTrgr(char *dev,char *cmt,int lvl,SINT32 tot,SINT32 rngStrt,SINT32 rngStop,int rngMax,int usrStmp)
{
  DTYPE din, dout;
  char triggerDevName[256];
  char cmtbuf[512];
  int cc, ivals[6];

  if (dev == NULL || strlen(dev) == 0) return argument_list_error;
  if (strchr(dev,'/') == 0)
  { /* only the device name is given -> use my known context */
    if (strlen(dev) > DEVICE_NAME_SIZE) return illegal_device;
    if (strlen(gDeviceContext) == 0) return illegal_context;
    sprintf(triggerDevName,"/%.32s/EVENTSTORE/%.64s",gDeviceContext,dev);
  }
  else 
  { /* full name specified */
    strncpy(triggerDevName,dev,256);
  }
  ivals[0] = lvl;
  ivals[1] = tot;
  ivals[2] = rngStrt;
  ivals[3] = rngStop;
  ivals[4] = rngMax;
  ivals[5] = usrStmp;
  DTYPEZERO(din);
  din.dFormat = CF_INT32;
  din.dArrayLength = rngMax < 0 ? 2 : (usrStmp == 0 ? 5 : 6);
  din.data.lptr = ivals;
  DTYPEZERO(dout);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  dout.data.lptr = &tot;
  if ((cc=ExecLinkEx(triggerDevName,"TRIGGER",&dout,&din,CA_WRITE|CA_NOCALLBACKS,200)) != 0) return cc;

  /* send comment buffer */
  strncpy(cmtbuf,cmt,480); cmtbuf[480] = 0;
  din.dFormat = CF_TEXT;
  din.data.cptr = cmtbuf;
  sprintf(&cmtbuf[strlen(cmtbuf)],"@%d",tot);
  din.dArrayLength = (UINT32)strlen(cmtbuf);
  if ((cc=ExecLinkEx(triggerDevName,"ANNOTATE",NULL,&din,CA_WRITE|CA_NOCALLBACKS,200)) != 0) return cc;
  return 0;
}
TINE_EXPORT int SendEventTriggerEx(char *dev,char *cmt,int triggerLevel,
                                   int triggerTime,int rangeStart,int rangeStop,int rangeMax, int options)
{
  int usrStmp = (options & EVNT_TRIGGER_USE_USERSTAMP) ? 1 : 0;
  return _sndEvntTrgr(dev,cmt,triggerLevel,triggerTime,rangeStart,rangeStop,rangeMax,usrStmp);
}
TINE_EXPORT int sendEventTrigger(char *dev,char *cmt,short triggerLevel)
{
  return _sndEvntTrgr(dev,cmt,(int)triggerLevel,0,0,0,-1,0);
}
TINE_EXPORT int SendEventTrigger(char *dev,char *cmt,short triggerLevel)
{
  return _sndEvntTrgr(dev,cmt,(int)triggerLevel,0,0,0,-1,0);
}
int _useCsvStringLiteral=FALSE;
char *getValidCsvString(char *dst,char *src,int dstlen,int srclen)
{
  char *c, buf[DBGBUF_SIZE];
  if (dst == NULL || src == NULL) return NULL;
  if (dstlen > DBGBUF_SIZE) dstlen = DBGBUF_SIZE;
  dstlen--; /* leave room for terminating '0' */
  if (dstlen > srclen) srclen = dstlen;
  if (src == dst) 
  { /* same pointers : make use of local buffer */
    strncpy(buf,src,dstlen);
    src = buf;
  }
  c = strtrm(src);
  if (!_useCsvStringLiteral && strpbrk(c,", \t") != NULL && c[0] != '"')
  {
    if (srclen > dstlen-2) srclen = dstlen-2;
    dst[0] = 0;
    strcat(dst,"\"");
    strncat(dst,c,srclen); 
    strcat(dst,"\"");
  }
  else
  {
    strncpy(dst,c,srclen); 
  }
  dst[dstlen] = 0;
  if ((c=strchr(dst,'"')) != NULL  && dstlen > 2)
  { /* need to have matched pairs! */
    int n = 0;
    while (c != NULL)
    {
      c++; n++;
      c = strchr(c,'"');
    }
    if (n % 2) dst[dstlen - 1] = '"';
  }
  return dst;
}
/**
 * \brief returns the supplied buffer IF the data format fully converts the data into a string
 *  else returns NULL (but fills in some diagnostic information in the supplied buffer)
 *
 * used in prepDumpData() as well as some save-and-restore routines, etc. where a string
 * representation of data is needed.
 *
 * \param sbuf is the string buffer to hold the converted string
 * \param data is a pointer to the data (value or array) which is to be converted to a string
 * \param idx is the starting array index if data points to an array 
 * \param fmt is the TINE format code describing the data
 * \param tag is the data tag if the given data represents a tagged format (e.g. CF_STRUCT)
 * \param limit is the maximum array length, i.e. buffer length of the data passed.
 *
 * \return the pointer passed as 'sbuf' if successful, else NULL
 */
TINE_EXPORT char *DataValueToString(char *sbuf, BYTE *data, int idx, short fmt, char *tag, int limit)
{
  return dataValueToString(sbuf, data, idx, fmt, tag, limit, FALSE);
}
char *dataValueToString(char *sbuf,BYTE *data,int idx,short fmt,char *tag,int limit,int internal)
{
  int nsiz,fsiz;
  float  fdat;
  SINT32 idat;
  SINT16  sdat;
  double ddat;
# ifndef SMALL_TINE_LIB
  UINT32 udat;
  double ddat2, ddat3;
  float  fdat2, fdat3;
  SINT32 idat2, idat3, idat4;
  time_t tdat;
# endif
# ifdef SINT64
  SINT64 lldat;
# endif
  if (data == NULL) { sbuf[0] = 0; return sbuf; }
  fsiz = fmtsizeof(fmt);
  switch (fmt)
  { /* handle possible alignment problems by memcpy() */
    case CF_BYTE:
      sprintf(sbuf,"%d",((BYTE *)data)[idx]);
      break;
    case CF_UINT16:
      memcpy(&sdat, (SINT16 *)&data[idx*fsiz], fsiz);
      sprintf(sbuf, "%u", (UINT16)sdat);
      break;
    case CF_INT16:
      memcpy(&sdat,(SINT16 *)&data[idx*fsiz],fsiz);
      sprintf(sbuf,"%d",sdat);
      break;
    case CF_UINT32:
      memcpy(&idat, (UINT32 *)&data[idx*fsiz], fsiz);
#     ifdef OS_16BIT
      sprintf(sbuf, "%lu", (UINT32)idat);
#     else
      sprintf(sbuf, "%u", (UINT32)idat);
#     endif
      break;
    case CF_INT32:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],fsiz);
#     ifdef OS_16BIT
      sprintf(sbuf,"%ld",idat);
#     else
      sprintf(sbuf,"%d",idat);
#     endif
      break;
#   ifdef SINT64
    case CF_UINT64:
      memcpy(&lldat, (UINT64 *)&data[idx*fsiz], fsiz);
#     if defined(WIN32)
      sprintf(sbuf, "%I64u", (UINT64)lldat);
#     elif !defined(OS_64BIT)
      sprintf(sbuf, "%llu", (UINT64)lldat);
#     else
      sprintf(sbuf, "%lu", (UINT64)lldat);
#     endif
      break;
    case CF_INT64:
      memcpy(&lldat,(SINT64 *)&data[idx*fsiz],fsiz);
#     if defined(WIN32)
      sprintf(sbuf, "%I64d", lldat);
#     elif !defined(OS_64BIT)
      sprintf(sbuf,"%lld",lldat);
#     else
      sprintf(sbuf,"%ld",lldat);
#     endif
      break;
#   endif    
    case CF_BOOLEAN:
      if (internal || !gMarshallBooleanAsByte)
        memcpy(&idat, (SINT32 *)&data[idx * fsiz], fsiz);
      else
        idat = (int)((_Bool *)data)[idx];
      strcpy(sbuf,(idat ? "true" : "false"));
      break;
    case CF_FLOAT:
      memcpy(&fdat,(float *)&data[idx*fsiz],fsiz);
      sprintf(sbuf,"%g",fdat);
      break;
    case CF_DOUBLE:
      memcpy(&ddat,(double *)&data[idx*fsiz],fsiz);
      sprintf(sbuf,"%.10g",ddat);
      break;
    case CF_TEXT:
      nsiz = limit > 0 ? limit : DBGBUF_SIZE-1;
      getValidCsvString(sbuf,(char *)&data[idx],DBGBUF_SIZE,nsiz);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      getValidCsvString(sbuf,(char *)&data[idx*fsiz],DBGBUF_SIZE,fsiz);
      break;
#   ifndef SMALL_TINE_LIB /* for DOS, etc. ... */
    case CF_II:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&idat2,(SINT32 *)&data[idx*fsiz+4],4);
      sprintf(sbuf,"%d %d",idat,idat2);
      break;
    case CF_III:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&idat2,(SINT32 *)&data[idx*fsiz+4],4);
      memcpy(&idat3,(SINT32 *)&data[idx*fsiz+8],4);
      sprintf(sbuf,"%d %d %d",idat,idat2,idat3);
      break;
    case CF_IIII:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&idat2,(SINT32 *)&data[idx*fsiz+4],4);
      memcpy(&idat3,(SINT32 *)&data[idx*fsiz+8],4);
      memcpy(&idat4,(SINT32 *)&data[idx*fsiz+12],4);
      sprintf(sbuf,"%d %d %d %d",idat,idat2,idat3,idat4);
      break;
   case CF_FLTINT:
      memcpy(&fdat,(float *)&data[idx*fsiz],4);
      memcpy(&idat,(SINT32 *)&data[idx*fsiz+4],4);
      sprintf(sbuf,"%g %d",fdat,idat);
      break;
   case CF_FLTFLT:
      memcpy(&fdat,(float *)&data[idx*fsiz],4);
      memcpy(&fdat2,(float *)&data[idx*fsiz+4],4);
      sprintf(sbuf,"%g %g",fdat,fdat2);
      break;
   case CF_DBLTIME:
   case CF_DBLDBL:
      memcpy(&ddat2,(double *)&data[idx*fsiz],8);
      memcpy(&ddat,(double *)&data[idx*fsiz+8],8);
      sprintf(sbuf,"%g %10.4f",ddat2,ddat);
      break;
   case CF_DBLDBLDBL:
      memcpy(&ddat3,(double *)&data[idx*fsiz],8);
      memcpy(&ddat2,(double *)&data[idx*fsiz+8],8);
      memcpy(&ddat,(double *)&data[idx*fsiz+16],8);
      sprintf(sbuf,"%g %g %g",ddat3,ddat2,ddat);
      break;
   case CF_NAME64TIME:
      memcpy(&ddat,(double *)&data[idx*fsiz+64],8);
      sprintf(sbuf,"%.64s %10.4f",(char *)&data[idx*fsiz],ddat);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_NAME64DBL:
      memcpy(&ddat,(double *)&data[idx*fsiz+64],8);
      sprintf(sbuf,"%.64s %g",(char *)&data[idx*fsiz],ddat);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_TDS:
      memcpy(&udat,(UINT32 *)&data[idx*fsiz],4);
      memcpy(&fdat,(float *)&data[idx*fsiz+4],4);
      memcpy(&idat,(SINT32 *)&data[idx*fsiz+8],4);
      sprintf(sbuf,"%d %g %d",udat,fdat,idat);
      break;
   case CF_NAME8I:
   case CF_NAME16I:
   case CF_NAME32I:
   case CF_NAME48I:
   case CF_NAME64I:
      nsiz = fsiz - 4;
      memcpy(&idat,&data[idx*fsiz+nsiz],4);
      strncpy(sbuf,(char *)&data[idx*fsiz],nsiz);
      sprintf(&sbuf[strlen(sbuf)]," %d",idat);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_NAME16FI:
      memcpy(&fdat,&data[idx*fsiz+16],4);
      memcpy(&idat,&data[idx*fsiz+20],4);
      strncpy(sbuf,(char *)&data[idx*fsiz],16);
      sprintf(&sbuf[strlen(sbuf)]," %g %d",fdat,idat);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_NAME16II:
      memcpy(&idat,&data[idx*fsiz+16],4);
      memcpy(&idat2,&data[idx*fsiz+20],4);
      strncpy(sbuf,(char *)&data[idx*fsiz],16);
      sprintf(&sbuf[strlen(sbuf)]," %d %d",idat,idat2);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_NAME16DBLDBL:
   case CF_NAME32DBLDBL:
   case CF_NAME64DBLDBL:
      nsiz = fsiz - 16;
      memcpy(&ddat,&data[idx*fsiz+nsiz],8);
      memcpy(&ddat2,&data[idx*fsiz+nsiz+8],8);
      strncpy(sbuf,(char *)&data[idx*fsiz],nsiz);
      sprintf(&sbuf[strlen(sbuf)]," %g %g",ddat,ddat2);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_FLTFLTINT:
      memcpy(&fdat,(float *)&data[idx*fsiz],4);
      memcpy(&fdat2,(float *)&data[idx*fsiz+4],4);
      memcpy(&idat,(SINT32 *)&data[idx*fsiz+8],4);
      sprintf(sbuf,"%g %g %d",fdat,fdat2,idat);
      break;
   case CF_FLTINTINT:
      memcpy(&fdat,(float *)&data[idx*fsiz],4);
      memcpy(&idat,(SINT32 *)&data[idx*fsiz+4],4);
      memcpy(&idat2,(SINT32 *)&data[idx*fsiz+8],4);
      sprintf(sbuf,"%g %d %d",fdat,idat,idat2);
      break;
   case CF_IFFF:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&fdat,(float *)&data[idx*fsiz+4],4);
      memcpy(&fdat2,(float *)&data[idx*fsiz+8],4);
      memcpy(&fdat3,(float *)&data[idx*fsiz+12],4);
      sprintf(sbuf,"%d %g %g %g",idat,fdat,fdat2,fdat3);
      break;
   case CF_INTINTFLTFLT:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&idat2,(SINT32 *)&data[idx*fsiz+4],4);
      memcpy(&fdat,(float *)&data[idx*fsiz+8],4);
      memcpy(&fdat2,(float *)&data[idx*fsiz+12],4);
      sprintf(sbuf,"%d %d %g %g",idat,idat2,fdat2,fdat2);
      break;
   case CF_FLTINTFLTINT:
      memcpy(&fdat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&idat,(SINT32 *)&data[idx*fsiz+4],4);
      memcpy(&fdat2,(float *)&data[idx*fsiz+8],4);
      memcpy(&idat2,(float *)&data[idx*fsiz+12],4);
      sprintf(sbuf,"%g %d %g %d",fdat,idat,fdat2,idat2);
      break;
   case CF_USTRING:
      memcpy(&idat,(SINT32 *)&data[idx*fsiz],4);
      memcpy(&fdat,(float *)&data[idx*fsiz+4],4);
      memcpy(&fdat2,(float *)&data[idx*fsiz+8],4);
      memcpy(&idat2,(SINT32 *)&data[idx*fsiz+12],4);
      sprintf(sbuf,"%d %g %g %d %.80s",idat,fdat,fdat2,idat2,(char *)&data[idx*fsiz+16]);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
   case CF_STRING:
     {
      char *sptr = ((char **)data)[idx];
      if (sptr == NULL) sptr = "";
      snprintf(sbuf,DBGBUF_SIZE,"%s",sptr);
      getValidCsvString(sbuf,sbuf,DBGBUF_SIZE,DBGBUF_SIZE);
      break;
     }
   case CF_SPECTRUM:
      tdat = (time_t)(*((UINT32 *)&data[80]));
      memcpy(&fdat,(float *)&data[84],4);
      memcpy(&fdat2,(float *)&data[88],4);
      memcpy(&idat,(SINT32 *)&data[92],4);
      sprintf(sbuf,"%.80s\n%.24s\nstart: %g\ninc: %g\nstatus: %d\n",
         (char *)&data[0],ctime(&tdat),fdat,fdat2,idat);
      return NULL;
   case CF_STRUCT:
      if (tag == NULL) tag = "structure";
      sprintf(sbuf,"%.16s data",tag);
      return NULL;
   case CF_GSPECTRUM:
   case CF_ASPECTRUM:
      strcpy(sbuf,"spectrum array data");
      return NULL;
   case CF_AIMAGE:
   case CF_IMAGE:
      strcpy(sbuf,"image data");
      return NULL;
   case CF_HISTORY:
      strcpy(sbuf,"history data");
      return NULL;
#  endif /* SMALL_TINE_LIB */
   default:
      strncpy(sbuf,GetFormatAsString(fmt),32);
      return NULL;
  }
  return sbuf; /* as a courtesy */
}
TINE_EXPORT char *GetDataAsString(BYTE* data, int idx, short fmt, char* tag)
{
  return _GetDataAsString(data, idx, fmt, tag, FALSE);
}
char *_GetDataAsString(BYTE *data, int idx, short fmt, char *tag,int internal)
{
  static char sbuf[512];
  dataValueToString(sbuf, data, idx, fmt, tag, 128, internal);
  return sbuf;
}
int isAllowedSrFile(short fmt)
{
  switch (fmt)
  {
    case CF_TEXT: return FALSE;
    case CF_STRUCT: return TRUE;
    case CF_IMAGE: return TRUE;
    case CF_SPECTRUM: return TRUE;
    default: break;
  }
  return isAllowedSaveRestoreTraceArray(fmt);
}
/**
 * \brief parses input string into the designated array element of the format type given.
 *
 * the input string can be a reference to a binary file (for array data) which is simply
 * read into the supplied data buffer.
 * it is assumed the the data array aligment is not an issue (as the buffer passed will
 * have been allocated to the specifications of the format type passed.
 *
 * \param data is the data buffer to receive the data 
 * \param sbuf is the string containing either the data to be parsed or a pipe reference
 * to a binary file containing a array of (primitive) data
 * \param idx is the index into the data parameter in which to copy the parsed or read data
 * \param fmt is the TINE format of the data
 * \param tag is a structure tag (not yet implemented)
 * \param limit is the array length or string limit if applicable
 *
 * used in some save-and-restore routines, etc.
 *
 * \note string contained in sbuf is 'trimmed' prior to call
 * 
 * \return 0 upon success, otherwise a TINE error code
 */
TINE_EXPORT int StringToDataValue(BYTE *data, char *sbuf, int idx, short fmt, char *tag, int limit)
{
  return stringToDataValue(data, sbuf, idx, fmt, tag, limit, FALSE);
}
int stringToDataValue(BYTE *data,char *sbuf,int idx,short fmt,char *tag,int limit,int internal)
{
  int nsiz,fsiz = fmtsizeof(fmt);
# ifndef SMALL_TINE_LIB
  char *c, *fn;
  FILE *fp=NULL;
# if defined(WIN32) || defined(DOS)
  char *fmod="r+b";
# else
  char *fmod="r";
# endif
  int cc = 0;
  UINT32 szb;

  if (data == NULL) return un_allocated;
  if (sbuf[0] == '<' && isAllowedSrFile(fmt) && limit > 1)
  { /* a pipe from a binary file (array data) */
    fn = &sbuf[1];
    if ((fp=fopen(fn,fmod)) == NULL) ccerr(no_such_file);
    fseek(fp,0,SEEK_END); szb = ftell(fp); fseek(fp,0,SEEK_SET);
    if (szb != (UINT32)(fsiz * limit)) ccerr(invalid_data);
    ccerr((fread(&data[idx*fsiz],1,szb,fp) == szb) ? 0 : file_error);
err:
    if (fp != NULL) fclose(fp);
    if (!internal) 
    {
      switch (fmt)
      {
        case CF_BOOLEAN:
          marshallBoolean(&data[idx * fsiz], limit, RTOL);
          break;
      }
    }
    return cc;
  }
# endif
  switch (fmt)
  { 
    case CF_BYTE:
      ((BYTE *)data)[idx] = (BYTE)atoi(sbuf);
      break;
    case CF_INT16:
      ((SINT16 *)data)[idx] = (SINT16)strtol(sbuf,NULL,0);
      break;
    case CF_INT32:
      ((SINT32 *)data)[idx] = (SINT32)strtol(sbuf,NULL,0);
      break;
#   ifdef SINT64
    case CF_INT64:
#     ifdef OS_64BIT
      ((SINT64 *)data)[idx] = (SINT64)strtol(sbuf,NULL,0);
#     else
      ((SINT64 *)data)[idx] = (SINT64)strtoll(sbuf,NULL,0);
#     endif
      break;
#   endif    
    case CF_BOOLEAN:
      if (internal || !gMarshallBooleanAsByte)
        ((SINT32 *)data)[idx] = stricmp(sbuf,"true") == 0 ? TRUE : FALSE;
      else
        ((_Bool *)data)[idx] = stricmp(sbuf, "true") == 0 ? (_Bool)TRUE : (_Bool)FALSE;
      break;
    case CF_FLOAT:
      ((float *)data)[idx] = (float)strtod(sbuf,NULL);
      break;
    case CF_DOUBLE:
      ((double *)data)[idx] = strtod(sbuf,NULL);
      break;
    case CF_TEXT:
      nsiz = limit > 0 ? limit : 255;
      strncpy((char *)&data[idx],sbuf,nsiz);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      strncpy((char *)&data[idx*fsiz],sbuf,fsiz);
      break;
#   ifndef SMALL_TINE_LIB /* for DOS, etc. ... */
    case CF_II:
      c = sbuf;
      ((INTINT *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
      ((INTINT *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      break;
    case CF_III:
      c = sbuf;
      ((INTINTINT *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
      ((INTINTINT *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      ((INTINTINT *)data)[idx].i3val = (SINT32)strtol(c,&c,0);
      break;
    case CF_IIII:
      c = sbuf;
      ((DADDRESS *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
      ((DADDRESS *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      ((DADDRESS *)data)[idx].i3val = (SINT32)strtol(c,&c,0);
      ((DADDRESS *)data)[idx].i4val = (SINT32)strtol(c,&c,0);
      break;
   case CF_FLTINT:
      c = sbuf;
      ((FLTINT *)data)[idx].fval = (float)strtod(c,&c);
      ((FLTINT *)data)[idx].ival = (SINT32)strtol(c,&c,0);
      break;
   case CF_FLTFLT:
      c = sbuf;
      ((FLTFLT *)data)[idx].f1val = (float)strtod(c,&c);
      ((FLTFLT *)data)[idx].f2val = (float)strtod(c,&c);
      break;
   case CF_DBLTIME:
   case CF_DBLDBL:
      c = sbuf;
      ((DBLDBL *)data)[idx].d1val = strtod(c,&c);
      ((DBLDBL *)data)[idx].d2val = strtod(c,&c);
      break;
   case CF_DBLDBLDBL:
      c = sbuf;
      ((DBLDBLDBL *)data)[idx].d1val = strtod(c,&c);
      ((DBLDBLDBL *)data)[idx].d2val = strtod(c,&c);
      ((DBLDBLDBL *)data)[idx].d3val = strtod(c,&c);
      break;
   case CF_NAME64TIME:
   case CF_NAME64DBL:
      if ((c=strrchr(sbuf,' ')) != NULL)
      { /* get the tail end first !*/
        *c++ = 0;
        *(double *)&data[idx*fsiz+64] = strtod(c,&c);
      }
      strncpy((char *)&data[idx*fsiz],sbuf,64);
      break;
   case CF_TDS:
      c = sbuf;
      ((INTFLTINT *)data)[idx].i1val = (UINT32)strtol(c,&c,0);
      ((INTFLTINT *)data)[idx].fval = (float)strtod(c,&c);
      ((INTFLTINT *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      break;
   case CF_NAME8I:
   case CF_NAME16I:
   case CF_NAME32I:
   case CF_NAME48I:
   case CF_NAME64I:
      nsiz = fsiz - 4;
      if ((c=strrchr(sbuf,' ')) != NULL)
      { /* get the tail end first !*/
        *c++ = 0;
        *(SINT32 *)&data[idx*fsiz+nsiz] = strtol(c,NULL,0);
      }
      strncpy((char *)&data[idx*fsiz],sbuf,nsiz);
      break;
   case CF_NAME16FI:
      if ((c=strrchr(sbuf,' ')) != NULL)
      { /* get the tail end first !*/
        *c++ = 0;
        ((NAME16FI *)data)[idx].ival = (SINT32)strtol(c,&c,0);
        if ((c=strrchr(sbuf,' ')) != NULL)
        {
          *c++ = 0;
          ((NAME16FI *)data)[idx].fval = (float)strtod(c,&c);
        }
      }
      strncpy(((NAME16FI *)data)[idx].name,sbuf,16);
      break;
   case CF_NAME16II:
      if ((c=strrchr(sbuf,' ')) != NULL)
      { /* get the tail end first !*/
        *c++ = 0;
        ((NAME16II *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
        strtrm(sbuf);
        if ((c=strrchr(sbuf,' ')) != NULL)
        {
          *c++ = 0;
          ((NAME16II *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
        }
      }
      strncpy(((NAME16FI *)data)[idx].name,sbuf,16);
      break;
   case CF_NAME16DBLDBL:
   case CF_NAME32DBLDBL:
   case CF_NAME64DBLDBL:
      nsiz = fsiz - 16;
      if ((c=strrchr(sbuf,' ')) != NULL)
      { /* get the tail end first !*/
        *c++ = 0;
        *(double *)&data[idx*fsiz+nsiz+8] = strtod(c,&c);
        strtrm(sbuf);
        if ((c=strrchr(sbuf,' ')) != NULL)
        {
          *c++ = 0;
          *(double *)&data[idx*fsiz+nsiz] = strtod(c,&c);
        }
      }
      strncpy((char *)&data[idx*fsiz],sbuf,nsiz);
      break;
   case CF_FLTFLTINT:
      c = sbuf;
      ((FLTFLTINT *)data)[idx].f1val = (float)strtod(c,&c);
      ((FLTFLTINT *)data)[idx].f2val = (float)strtod(c,&c);
      ((FLTFLTINT *)data)[idx].ival = (SINT32)strtol(c,&c,0);
      break;
   case CF_FLTINTINT:
      c = sbuf;
      ((FLTINTINT *)data)[idx].fval = (float)strtod(c,&c);
      ((FLTINTINT *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
      ((FLTINTINT *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      break;
   case CF_IFFF:
      c = sbuf;
      ((DFILTER *)data)[idx].ival = (SINT32)strtol(c,&c,0);
      ((DFILTER *)data)[idx].f1val = (float)strtod(c,&c);
      ((DFILTER *)data)[idx].f2val = (float)strtod(c,&c);
      ((DFILTER *)data)[idx].f3val = (float)strtod(c,&c);
      break;
   case CF_INTINTFLTFLT:
      c = sbuf;
      ((FWINDOW *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
      ((FWINDOW *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      ((FWINDOW *)data)[idx].f1val = (float)strtod(c,&c);
      ((FWINDOW *)data)[idx].f2val = (float)strtod(c,&c);
      break;
   case CF_FLTINTFLTINT:
      c = sbuf;
      ((FLTINTFLTINT *)data)[idx].f1val = (float)strtod(c,&c);
      ((FLTINTFLTINT *)data)[idx].i1val = (SINT32)strtol(c,&c,0);
      ((FLTINTFLTINT *)data)[idx].f2val = (float)strtod(c,&c);
      ((FLTINTFLTINT *)data)[idx].i2val = (SINT32)strtol(c,&c,0);
      break;
   case CF_USTRING:
      c = sbuf;
      ((DUSTRING *)data)[idx].ival = (SINT32)strtol(c,&c,0);
      ((DUSTRING *)data)[idx].f1val = (float)strtod(c,&c);
      ((DUSTRING *)data)[idx].f2val = (float)strtod(c,&c);
      ((DUSTRING *)data)[idx].tm = (SINT32)strtol(c,&c,0);
      while (*c == ' ') c++;
      strncpy(((DUSTRING *)data)[idx].str,c,80);
      break;
   case CF_STRING:
      if (strlen(sbuf) >= STRING_STD_CAPACITY*2) 
        msglog(string_too_long,"stingToDataValue: truncating string value to %d chars",STRING_STD_CAPACITY*2);
      strncpy(((char **)data)[idx],sbuf,STRING_STD_CAPACITY*2);
      break;
   case CF_MDA:
      return illegal_format;
   case CF_STRUCT:
      UNUSED(tag);
      return illegal_format;
   case CF_SPECTRUM:
   case CF_GSPECTRUM:
   case CF_ASPECTRUM:
      return illegal_format;
   case CF_AIMAGE:
   case CF_IMAGE:
      return illegal_format;
   case CF_HISTORY:
      return illegal_format;
#  endif /* SMALL_TINE_LIB */
   default:
      return illegal_format;
  }
  return 0;
}
/** 
 * \internal
 * \param s is the string to check for and replace non-ascii characters
 * \param len is the maximum length the string s is allowed to have.
 */
TINE_EXPORT void ensureAsciiCharacters(char *s, int len)
{
# define MAX_ASCII_STRING_LENGTH 256
  char *c, buf[MAX_ASCII_STRING_LENGTH];
  int isascii = FALSE, isutf8 = FALSE, docopy=FALSE, i = 0;
  if (s == NULL) return;
  if (len > MAX_ASCII_STRING_LENGTH) len = MAX_ASCII_STRING_LENGTH;
  for (c = s; *c != '\0' && i < len; c++)
  {
    isascii = *c < 0 ? FALSE : TRUE;
    isutf8 = ((BYTE)*c == 0xc3) ? TRUE : FALSE;
    if (isascii)
    {
      buf[i++] = *c;
    }
    else if (isutf8)
    {
      c++;
      docopy = TRUE;
      switch ((BYTE)*c)
      {
        case 0x84:
          buf[i++] = 'A';
          buf[i++] = 'E';
          break;
        case 0xa4:
          buf[i++] = 'a';
          buf[i++] = 'e';
          break;
        case 0x96:
          buf[i++] = 'O';
          buf[i++] = 'E';
          break;
        case 0xb6:
          buf[i++] = 'o';
          buf[i++] = 'e';
          break;
        case 0x9c:
          buf[i++] = 'U';
          buf[i++] = 'E';
          break;
        case 0xbc:
          buf[i++] = 'u';
          buf[i++] = 'e';
          break;
        case 0x9f:
          buf[i++] = 's';
          buf[i++] = 's';
          break;
        default:
          buf[i++] = *c;
          break;
      }
    }
    else
    { /* assume it's windows-1252 */
      docopy = TRUE;
      switch ((int)*c)
      {
        case -60:
          buf[i++] = 'A';
          buf[i++] = 'E';
          break;
        case -28:
          buf[i++] = 'a';
          buf[i++] = 'e';
          break;
        case -42:
          buf[i++] = 'O';
          buf[i++] = 'E';
          break;
        case -10:
          buf[i++] = 'o';
          buf[i++] = 'e';
          break;
        case -36:
          buf[i++] = 'U';
          buf[i++] = 'E';
          break;
        case -4:
          buf[i++] = 'u';
          buf[i++] = 'e';
          break;
        case -33:
          buf[i++] = 's';
          buf[i++] = 's';
          break;
        default:
          buf[i++] = *c;
          break;
        }
    }
  }
  if (i < MAX_ASCII_STRING_LENGTH) buf[i] = 0;
  if (i >= len) i = len;
  if (docopy) strncpy(s, buf, i);
}
#ifdef NEED_STRTOLL
/*-
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifdef WIN32
#define LLONG_MIN  (-9223372036854775807 - 1)
#define LLONG_MAX  9223372036854775807
SINT64 _strtoi64(const char *nptr, char **endptr, int base)
#else
# ifndef LLONG_MIN
#   define LLONG_MIN  (-9223372036854775807LL - 1)
# endif
# ifndef LLONG_MAX
#   define LLONG_MAX  9223372036854775807LL
# endif
SINT64 strtoll(const char *nptr, char **endptr, int base)
#endif
{
  const char *s;
  SINT64 acc, cutoff;
  int c;
  int neg, any, cutlim;
	/* endptr may be NULL */
#ifdef __GNUC__
  /* This outrageous construct just to shut up a GCC warning. */
  (void) &acc; (void) &cutoff;
#endif
  s = nptr;
  do { c = (unsigned char) *s++; } while (isspace(c));
  if (c == '-') 
  {
    neg = 1;
    c = *s++;
  } 
  else 
  {
    neg = 0;
    if (c == '+') c = *s++;
  }
  if ((base == 0 || base == 16) && c == '0' && (*s == 'x' || *s == 'X')) 
  {
    c = s[1];
    s += 2;
    base = 16;
  }
  if (base == 0) base = c == '0' ? 8 : 10;
 /*
  * Compute the cutoff value between legal numbers and illegal
  * numbers.  That is the largest legal value, divided by the
  * base.  An input number that is greater than this value, if
  * followed by a legal input character, is too big.  One that
  * is equal to this value may be valid or not; the limit
  * between valid and invalid numbers is then based on the last
  * digit.  For instance, if the range for long longs is
  * [-9223372036854775808..9223372036854775807] and the input base
  * is 10, cutoff will be set to 922337203685477580 and cutlim to
  * either 7 (neg==0) or 8 (neg==1), meaning that if we have
  * accumulated a value > 922337203685477580, or equal but the
  * next digit is > 7 (or 8), the number is too big, and we will
  * return a range error.
  *
  * Set any if any `digits' consumed; make it negative to indicate
  * overflow.
  */
  cutoff = neg ? LLONG_MIN : LLONG_MAX;
  cutlim = (int)(cutoff % base);
  cutoff /= base;
  if (neg) 
  {
    if (cutlim > 0) 
    {
      cutlim -= base;
      cutoff += 1;
    }
    cutlim = -cutlim;
  }
  for (acc = 0, any = 0;; c = (unsigned char) *s++) 
  {
    if (isdigit(c)) c -= '0';
    else if (isalpha(c)) c -= isupper(c) ? 'A' - 10 : 'a' - 10;
    else break;
    if (c >= base) break;
    if (any < 0) continue;
    if (neg) 
    {
      if (acc < cutoff || (acc == cutoff && c > cutlim)) 
      {
        any = -1;
        acc = LLONG_MIN;
        errno = ERANGE;
      } 
      else 
      {
        any = 1;
        acc *= base;
        acc -= c;
      }
	  } 
    else 
    {
      if (acc > cutoff || (acc == cutoff && c > cutlim)) 
      {
        any = -1;
        acc = LLONG_MAX;
        errno = ERANGE;
      } 
      else 
      {
        any = 1;
        acc *= base;
        acc += c;
      }
    }
  }
  if (endptr != 0) *endptr = (char *)(any ? s - 1 : nptr);
  return acc;
}
#endif
